# -*- coding: utf-8 -*-
# Spearmint
#
# Academic and Non-Commercial Research Use Software License and Terms
# of Use
#
# Spearmint is a software package to perform Bayesian optimization
# according to specific algorithms (the “Software”).  The Software is
# designed to automatically run experiments (thus the code name
# 'spearmint') in a manner that iteratively adjusts a number of
# parameters so as to minimize some objective in as few runs as
# possible.
#
# The Software was developed by Ryan P. Adams, Michael Gelbart, and
# Jasper Snoek and at Harvard University, Kevin Swersky and Richard
# Zemel at the University of Toronto (“Toronto”), and Hugo Larochelle
# at the Université de Sherbrooke (“Sherbrooke”), which assigned its
# rights in the Software to Socpra Sciences et Génie
# S.E.C. (“Socpra”). Pursuant to an inter-institutional agreement
# between the parties, it is distributed for free academic and
# non-commercial research use by the President and Fellows of Harvard
# College (“Harvard”).
#
# Using the Software indicates your agreement to be bound by the terms
# of this Software Use Agreement (“Agreement”). Absent your agreement
# to the terms below, you (the “End User”) have no rights to hold or
# use the Software whatsoever.
#
# Harvard agrees to grant hereunder the limited non-exclusive license
# to End User for the use of the Software in the performance of End
# User’s internal, non-commercial research and academic use at End
# User’s academic or not-for-profit research institution
# (“Institution”) on the following terms and conditions:
#
# 1.  NO REDISTRIBUTION. The Software remains the property Harvard,
# Toronto and Socpra, and except as set forth in Section 4, End User
# shall not publish, distribute, or otherwise transfer or make
# available the Software to any other party.
#
# 2.  NO COMMERCIAL USE. End User shall not use the Software for
# commercial purposes and any such use of the Software is expressly
# prohibited. This includes, but is not limited to, use of the
# Software in fee-for-service arrangements, core facilities or
# laboratories or to provide research services to (or in collaboration
# with) third parties for a fee, and in industry-sponsored
# collaborative research projects where any commercial rights are
# granted to the sponsor. If End User wishes to use the Software for
# commercial purposes or for any other restricted purpose, End User
# must execute a separate license agreement with Harvard.
#
# Requests for use of the Software for commercial purposes, please
# contact:
#
# Office of Technology Development
# Harvard University
# Smith Campus Center, Suite 727E
# 1350 Massachusetts Avenue
# Cambridge, MA 02138 USA
# Telephone: (617) 495-3067
# Facsimile: (617) 495-9568
# E-mail: otd@harvard.edu
#
# 3.  OWNERSHIP AND COPYRIGHT NOTICE. Harvard, Toronto and Socpra own
# all intellectual property in the Software. End User shall gain no
# ownership to the Software. End User shall not remove or delete and
# shall retain in the Software, in any modifications to Software and
# in any Derivative Works, the copyright, trademark, or other notices
# pertaining to Software as provided with the Software.
#
# 4.  DERIVATIVE WORKS. End User may create and use Derivative Works,
# as such term is defined under U.S. copyright laws, provided that any
# such Derivative Works shall be restricted to non-commercial,
# internal research and academic use at End User’s Institution. End
# User may distribute Derivative Works to other Institutions solely
# for the performance of non-commercial, internal research and
# academic use on terms substantially similar to this License and
# Terms of Use.
#
# 5.  FEEDBACK. In order to improve the Software, comments from End
# Users may be useful. End User agrees to provide Harvard with
# feedback on the End User’s use of the Software (e.g., any bugs in
# the Software, the user experience, etc.).  Harvard is permitted to
# use such information provided by End User in making changes and
# improvements to the Software without compensation or an accounting
# to End User.
#
# 6.  NON ASSERT. End User acknowledges that Harvard, Toronto and/or
# Sherbrooke or Socpra may develop modifications to the Software that
# may be based on the feedback provided by End User under Section 5
# above. Harvard, Toronto and Sherbrooke/Socpra shall not be
# restricted in any way by End User regarding their use of such
# information.  End User acknowledges the right of Harvard, Toronto
# and Sherbrooke/Socpra to prepare, publish, display, reproduce,
# transmit and or use modifications to the Software that may be
# substantially similar or functionally equivalent to End User’s
# modifications and/or improvements if any.  In the event that End
# User obtains patent protection for any modification or improvement
# to Software, End User agrees not to allege or enjoin infringement of
# End User’s patent against Harvard, Toronto or Sherbrooke or Socpra,
# or any of the researchers, medical or research staff, officers,
# directors and employees of those institutions.
#
# 7.  PUBLICATION & ATTRIBUTION. End User has the right to publish,
# present, or share results from the use of the Software.  In
# accordance with customary academic practice, End User will
# acknowledge Harvard, Toronto and Sherbrooke/Socpra as the providers
# of the Software and may cite the relevant reference(s) from the
# following list of publications:
#
# Practical Bayesian Optimization of Machine Learning Algorithms
# Jasper Snoek, Hugo Larochelle and Ryan Prescott Adams
# Neural Information Processing Systems, 2012
#
# Multi-Task Bayesian Optimization
# Kevin Swersky, Jasper Snoek and Ryan Prescott Adams
# Advances in Neural Information Processing Systems, 2013
#
# Input Warping for Bayesian Optimization of Non-stationary Functions
# Jasper Snoek, Kevin Swersky, Richard Zemel and Ryan Prescott Adams
# Preprint, arXiv:1402.0929, http://arxiv.org/abs/1402.0929, 2013
#
# Bayesian Optimization and Semiparametric Models with Applications to
# Assistive Technology Jasper Snoek, PhD Thesis, University of
# Toronto, 2013
#
# 8.  NO WARRANTIES. THE SOFTWARE IS PROVIDED "AS IS." TO THE FULLEST
# EXTENT PERMITTED BY LAW, HARVARD, TORONTO AND SHERBROOKE AND SOCPRA
# HEREBY DISCLAIM ALL WARRANTIES OF ANY KIND (EXPRESS, IMPLIED OR
# OTHERWISE) REGARDING THE SOFTWARE, INCLUDING BUT NOT LIMITED TO ANY
# IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OWNERSHIP, AND NON-INFRINGEMENT.  HARVARD, TORONTO AND
# SHERBROOKE AND SOCPRA MAKE NO WARRANTY ABOUT THE ACCURACY,
# RELIABILITY, COMPLETENESS, TIMELINESS, SUFFICIENCY OR QUALITY OF THE
# SOFTWARE.  HARVARD, TORONTO AND SHERBROOKE AND SOCPRA DO NOT WARRANT
# THAT THE SOFTWARE WILL OPERATE WITHOUT ERROR OR INTERRUPTION.
#
# 9.  LIMITATIONS OF LIABILITY AND REMEDIES. USE OF THE SOFTWARE IS AT
# END USER’S OWN RISK. IF END USER IS DISSATISFIED WITH THE SOFTWARE,
# ITS EXCLUSIVE REMEDY IS TO STOP USING IT.  IN NO EVENT SHALL
# HARVARD, TORONTO OR SHERBROOKE OR SOCPRA BE LIABLE TO END USER OR
# ITS INSTITUTION, IN CONTRACT, TORT OR OTHERWISE, FOR ANY DIRECT,
# INDIRECT, SPECIAL, INCIDENTAL, CONSEQUENTIAL, PUNITIVE OR OTHER
# DAMAGES OF ANY KIND WHATSOEVER ARISING OUT OF OR IN CONNECTION WITH
# THE SOFTWARE, EVEN IF HARVARD, TORONTO OR SHERBROOKE OR SOCPRA IS
# NEGLIGENT OR OTHERWISE AT FAULT, AND REGARDLESS OF WHETHER HARVARD,
# TORONTO OR SHERBROOKE OR SOCPRA IS ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGES.
#
# 10. INDEMNIFICATION. To the extent permitted by law, End User shall
# indemnify, defend and hold harmless Harvard, Toronto and Sherbrooke
# and Socpra, their corporate affiliates, current or future directors,
# trustees, officers, faculty, medical and professional staff,
# employees, students and agents and their respective successors,
# heirs and assigns (the "Indemnitees"), against any liability,
# damage, loss or expense (including reasonable attorney's fees and
# expenses of litigation) incurred by or imposed upon the Indemnitees
# or any one of them in connection with any claims, suits, actions,
# demands or judgments arising from End User’s breach of this
# Agreement or its Institution’s use of the Software except to the
# extent caused by the gross negligence or willful misconduct of
# Harvard, Toronto or Sherbrooke or Socpra. This indemnification
# provision shall survive expiration or termination of this Agreement.
#
# 11. GOVERNING LAW. This Agreement shall be construed and governed by
# the laws of the Commonwealth of Massachusetts regardless of
# otherwise applicable choice of law standards.
#
# 12. NON-USE OF NAME.  Nothing in this License and Terms of Use shall
# be construed as granting End Users or their Institutions any rights
# or licenses to use any trademarks, service marks or logos associated
# with the Software.  You may not use the terms “Harvard” or
# “University of Toronto” or “Université de Sherbrooke” or “Socpra
# Sciences et Génie S.E.C.” (or a substantially similar term) in any
# way that is inconsistent with the permitted uses described
# herein. You agree not to use any name or emblem of Harvard, Toronto
# or Sherbrooke, or any of their subdivisions for any purpose, or to
# falsely suggest any relationship between End User (or its
# Institution) and Harvard, Toronto and/or Sherbrooke, or in any
# manner that would infringe or violate any of their rights.
#
# 13. End User represents and warrants that it has the legal authority
# to enter into this License and Terms of Use on behalf of itself and
# its Institution.

import numpy as np
import cPickle as pickle

# Numba autojit might be nice.  Currently asplodes.
def sobol(num_points, num_dims):
    """Get Sobol sequence with num_points for num_dims."""

    # Total number of bits we need for a sequence of length num_points.
    num_bits = int(np.ceil(np.log2(num_points)))

    # Compute the c_i sequence.
    X = to_binary(np.arange(num_points, dtype=np.uint32), num_bits)
    C = num_bits - np.max((1-X) * (np.arange(num_bits)+1), axis=1) + 1
    
    # Direction numbers
    V = np.zeros((num_dims, num_bits), dtype=np.uint32)

    # Direction numbers for first dimension.
    V[0,:] = 1 << np.arange(31, 31-num_bits, -1, dtype=np.uint32)

    Z = np.zeros((num_points, num_dims))

    # Get the parameters for the Sobol sequence.
    # The first entry in this array (index 0) is the second dimension.
    params = get_params()

    # Loop over dimensions
    for dd in xrange(1,num_dims):
        s = params[dd-1]['s']
        a = params[dd-1]['a']
        m = params[dd-1]['m']

        # Direction numbers for dd-th dimension.
        if (num_bits <= s):
            V[dd,:] = m << np.arange(31, 31-num_bits, -1, dtype=np.uint32)
        else:
            V[dd,:s] = m << np.arange(31, 31-s, -1, dtype=np.uint32)
            for s0 in xrange(s, num_bits):
                V[dd,s0] = V[dd,s0-s] ^ (V[dd,s0-s] >> s)
                for s1 in xrange(s-1):
                    V[dd,s0] = V[dd,s0] ^ (((a >> (s-2-s1)) & 1) * V[dd,s0-s1-1])

    X = np.zeros((num_points, num_dims), dtype=np.uint32)

    # Wish we could do this without recursion.
    # Fancy loop unrolling?
    for nn in xrange(1,num_points):
        X[nn,:] = X[nn-1,:] ^ V[:,C[nn-1]-1]

    Z = X.astype('double') / float(1<<32)

    return Z

def to_binary(X, bits):
    return 1 & (X[:,np.newaxis]/2**np.arange(bits-1,-1,-1, dtype=np.uint32))

# These are the parameters for the Sobol sequence.
# This is hilarious.
params = """(lp1
(dp2
S'a'
I0
sS's'
I1
sS'm'
(lp3
I1
asS'd'
I2
sa(dp4
S'a'
I1
sS's'
I2
sS'm'
(lp5
I1
aI3
asS'd'
I3
sa(dp6
S'a'
I1
sS's'
I3
sS'm'
(lp7
I1
aI3
aI1
asS'd'
I4
sa(dp8
S'a'
I2
sS's'
I3
sS'm'
(lp9
I1
aI1
aI1
asS'd'
I5
sa(dp10
S'a'
I1
sS's'
I4
sS'm'
(lp11
I1
aI1
aI3
aI3
asS'd'
I6
sa(dp12
S'a'
I4
sS's'
I4
sS'm'
(lp13
I1
aI3
aI5
aI13
asS'd'
I7
sa(dp14
S'a'
I2
sS's'
I5
sS'm'
(lp15
I1
aI1
aI5
aI5
aI17
asS'd'
I8
sa(dp16
S'a'
I4
sS's'
I5
sS'm'
(lp17
I1
aI1
aI5
aI5
aI5
asS'd'
I9
sa(dp18
S'a'
I7
sS's'
I5
sS'm'
(lp19
I1
aI1
aI7
aI11
aI19
asS'd'
I10
sa(dp20
S'a'
I11
sS's'
I5
sS'm'
(lp21
I1
aI1
aI5
aI1
aI1
asS'd'
I11
sa(dp22
S'a'
I13
sS's'
I5
sS'm'
(lp23
I1
aI1
aI1
aI3
aI11
asS'd'
I12
sa(dp24
S'a'
I14
sS's'
I5
sS'm'
(lp25
I1
aI3
aI5
aI5
aI31
asS'd'
I13
sa(dp26
S'a'
I1
sS's'
I6
sS'm'
(lp27
I1
aI3
aI3
aI9
aI7
aI49
asS'd'
I14
sa(dp28
S'a'
I13
sS's'
I6
sS'm'
(lp29
I1
aI1
aI1
aI15
aI21
aI21
asS'd'
I15
sa(dp30
S'a'
I16
sS's'
I6
sS'm'
(lp31
I1
aI3
aI1
aI13
aI27
aI49
asS'd'
I16
sa(dp32
S'a'
I19
sS's'
I6
sS'm'
(lp33
I1
aI1
aI1
aI15
aI7
aI5
asS'd'
I17
sa(dp34
S'a'
I22
sS's'
I6
sS'm'
(lp35
I1
aI3
aI1
aI15
aI13
aI25
asS'd'
I18
sa(dp36
S'a'
I25
sS's'
I6
sS'm'
(lp37
I1
aI1
aI5
aI5
aI19
aI61
asS'd'
I19
sa(dp38
S'a'
I1
sS's'
I7
sS'm'
(lp39
I1
aI3
aI7
aI11
aI23
aI15
aI103
asS'd'
I20
sa(dp40
S'a'
I4
sS's'
I7
sS'm'
(lp41
I1
aI3
aI7
aI13
aI13
aI15
aI69
asS'd'
I21
sa(dp42
S'a'
I7
sS's'
I7
sS'm'
(lp43
I1
aI1
aI3
aI13
aI7
aI35
aI63
asS'd'
I22
sa(dp44
S'a'
I8
sS's'
I7
sS'm'
(lp45
I1
aI3
aI5
aI9
aI1
aI25
aI53
asS'd'
I23
sa(dp46
S'a'
I14
sS's'
I7
sS'm'
(lp47
I1
aI3
aI1
aI13
aI9
aI35
aI107
asS'd'
I24
sa(dp48
S'a'
I19
sS's'
I7
sS'm'
(lp49
I1
aI3
aI1
aI5
aI27
aI61
aI31
asS'd'
I25
sa(dp50
S'a'
I21
sS's'
I7
sS'm'
(lp51
I1
aI1
aI5
aI11
aI19
aI41
aI61
asS'd'
I26
sa(dp52
S'a'
I28
sS's'
I7
sS'm'
(lp53
I1
aI3
aI5
aI3
aI3
aI13
aI69
asS'd'
I27
sa(dp54
S'a'
I31
sS's'
I7
sS'm'
(lp55
I1
aI1
aI7
aI13
aI1
aI19
aI1
asS'd'
I28
sa(dp56
S'a'
I32
sS's'
I7
sS'm'
(lp57
I1
aI3
aI7
aI5
aI13
aI19
aI59
asS'd'
I29
sa(dp58
S'a'
I37
sS's'
I7
sS'm'
(lp59
I1
aI1
aI3
aI9
aI25
aI29
aI41
asS'd'
I30
sa(dp60
S'a'
I41
sS's'
I7
sS'm'
(lp61
I1
aI3
aI5
aI13
aI23
aI1
aI55
asS'd'
I31
sa(dp62
S'a'
I42
sS's'
I7
sS'm'
(lp63
I1
aI3
aI7
aI3
aI13
aI59
aI17
asS'd'
I32
sa(dp64
S'a'
I50
sS's'
I7
sS'm'
(lp65
I1
aI3
aI1
aI3
aI5
aI53
aI69
asS'd'
I33
sa(dp66
S'a'
I55
sS's'
I7
sS'm'
(lp67
I1
aI1
aI5
aI5
aI23
aI33
aI13
asS'd'
I34
sa(dp68
S'a'
I56
sS's'
I7
sS'm'
(lp69
I1
aI1
aI7
aI7
aI1
aI61
aI123
asS'd'
I35
sa(dp70
S'a'
I59
sS's'
I7
sS'm'
(lp71
I1
aI1
aI7
aI9
aI13
aI61
aI49
asS'd'
I36
sa(dp72
S'a'
I62
sS's'
I7
sS'm'
(lp73
I1
aI3
aI3
aI5
aI3
aI55
aI33
asS'd'
I37
sa(dp74
S'a'
I14
sS's'
I8
sS'm'
(lp75
I1
aI3
aI1
aI15
aI31
aI13
aI49
aI245
asS'd'
I38
sa(dp76
S'a'
I21
sS's'
I8
sS'm'
(lp77
I1
aI3
aI5
aI15
aI31
aI59
aI63
aI97
asS'd'
I39
sa(dp78
S'a'
I22
sS's'
I8
sS'm'
(lp79
I1
aI3
aI1
aI11
aI11
aI11
aI77
aI249
asS'd'
I40
sa(dp80
S'a'
I38
sS's'
I8
sS'm'
(lp81
I1
aI3
aI1
aI11
aI27
aI43
aI71
aI9
asS'd'
I41
sa(dp82
S'a'
I47
sS's'
I8
sS'm'
(lp83
I1
aI1
aI7
aI15
aI21
aI11
aI81
aI45
asS'd'
I42
sa(dp84
S'a'
I49
sS's'
I8
sS'm'
(lp85
I1
aI3
aI7
aI3
aI25
aI31
aI65
aI79
asS'd'
I43
sa(dp86
S'a'
I50
sS's'
I8
sS'm'
(lp87
I1
aI3
aI1
aI1
aI19
aI11
aI3
aI205
asS'd'
I44
sa(dp88
S'a'
I52
sS's'
I8
sS'm'
(lp89
I1
aI1
aI5
aI9
aI19
aI21
aI29
aI157
asS'd'
I45
sa(dp90
S'a'
I56
sS's'
I8
sS'm'
(lp91
I1
aI3
aI7
aI11
aI1
aI33
aI89
aI185
asS'd'
I46
sa(dp92
S'a'
I67
sS's'
I8
sS'm'
(lp93
I1
aI3
aI3
aI3
aI15
aI9
aI79
aI71
asS'd'
I47
sa(dp94
S'a'
I70
sS's'
I8
sS'm'
(lp95
I1
aI3
aI7
aI11
aI15
aI39
aI119
aI27
asS'd'
I48
sa(dp96
S'a'
I84
sS's'
I8
sS'm'
(lp97
I1
aI1
aI3
aI1
aI11
aI31
aI97
aI225
asS'd'
I49
sa(dp98
S'a'
I97
sS's'
I8
sS'm'
(lp99
I1
aI1
aI1
aI3
aI23
aI43
aI57
aI177
asS'd'
I50
sa(dp100
S'a'
I103
sS's'
I8
sS'm'
(lp101
I1
aI3
aI7
aI7
aI17
aI17
aI37
aI71
asS'd'
I51
sa(dp102
S'a'
I115
sS's'
I8
sS'm'
(lp103
I1
aI3
aI1
aI5
aI27
aI63
aI123
aI213
asS'd'
I52
sa(dp104
S'a'
I122
sS's'
I8
sS'm'
(lp105
I1
aI1
aI3
aI5
aI11
aI43
aI53
aI133
asS'd'
I53
sa(dp106
S'a'
I8
sS's'
I9
sS'm'
(lp107
I1
aI3
aI5
aI5
aI29
aI17
aI47
aI173
aI479
asS'd'
I54
sa(dp108
S'a'
I13
sS's'
I9
sS'm'
(lp109
I1
aI3
aI3
aI11
aI3
aI1
aI109
aI9
aI69
asS'd'
I55
sa(dp110
S'a'
I16
sS's'
I9
sS'm'
(lp111
I1
aI1
aI1
aI5
aI17
aI39
aI23
aI5
aI343
asS'd'
I56
sa(dp112
S'a'
I22
sS's'
I9
sS'm'
(lp113
I1
aI3
aI1
aI5
aI25
aI15
aI31
aI103
aI499
asS'd'
I57
sa(dp114
S'a'
I25
sS's'
I9
sS'm'
(lp115
I1
aI1
aI1
aI11
aI11
aI17
aI63
aI105
aI183
asS'd'
I58
sa(dp116
S'a'
I44
sS's'
I9
sS'm'
(lp117
I1
aI1
aI5
aI11
aI9
aI29
aI97
aI231
aI363
asS'd'
I59
sa(dp118
S'a'
I47
sS's'
I9
sS'm'
(lp119
I1
aI1
aI5
aI15
aI19
aI45
aI41
aI7
aI383
asS'd'
I60
sa(dp120
S'a'
I52
sS's'
I9
sS'm'
(lp121
I1
aI3
aI7
aI7
aI31
aI19
aI83
aI137
aI221
asS'd'
I61
sa(dp122
S'a'
I55
sS's'
I9
sS'm'
(lp123
I1
aI1
aI1
aI3
aI23
aI15
aI111
aI223
aI83
asS'd'
I62
sa(dp124
S'a'
I59
sS's'
I9
sS'm'
(lp125
I1
aI1
aI5
aI13
aI31
aI15
aI55
aI25
aI161
asS'd'
I63
sa(dp126
S'a'
I62
sS's'
I9
sS'm'
(lp127
I1
aI1
aI3
aI13
aI25
aI47
aI39
aI87
aI257
asS'd'
I64
sa(dp128
S'a'
I67
sS's'
I9
sS'm'
(lp129
I1
aI1
aI1
aI11
aI21
aI53
aI125
aI249
aI293
asS'd'
I65
sa(dp130
S'a'
I74
sS's'
I9
sS'm'
(lp131
I1
aI1
aI7
aI11
aI11
aI7
aI57
aI79
aI323
asS'd'
I66
sa(dp132
S'a'
I81
sS's'
I9
sS'm'
(lp133
I1
aI1
aI5
aI5
aI17
aI13
aI81
aI3
aI131
asS'd'
I67
sa(dp134
S'a'
I82
sS's'
I9
sS'm'
(lp135
I1
aI1
aI7
aI13
aI23
aI7
aI65
aI251
aI475
asS'd'
I68
sa(dp136
S'a'
I87
sS's'
I9
sS'm'
(lp137
I1
aI3
aI5
aI1
aI9
aI43
aI3
aI149
aI11
asS'd'
I69
sa(dp138
S'a'
I91
sS's'
I9
sS'm'
(lp139
I1
aI1
aI3
aI13
aI31
aI13
aI13
aI255
aI487
asS'd'
I70
sa(dp140
S'a'
I94
sS's'
I9
sS'm'
(lp141
I1
aI3
aI3
aI1
aI5
aI63
aI89
aI91
aI127
asS'd'
I71
sa(dp142
S'a'
I103
sS's'
I9
sS'm'
(lp143
I1
aI1
aI3
aI3
aI1
aI19
aI123
aI127
aI237
asS'd'
I72
sa(dp144
S'a'
I104
sS's'
I9
sS'm'
(lp145
I1
aI1
aI5
aI7
aI23
aI31
aI37
aI243
aI289
asS'd'
I73
sa(dp146
S'a'
I109
sS's'
I9
sS'm'
(lp147
I1
aI1
aI5
aI11
aI17
aI53
aI117
aI183
aI491
asS'd'
I74
sa(dp148
S'a'
I122
sS's'
I9
sS'm'
(lp149
I1
aI1
aI1
aI5
aI1
aI13
aI13
aI209
aI345
asS'd'
I75
sa(dp150
S'a'
I124
sS's'
I9
sS'm'
(lp151
I1
aI1
aI3
aI15
aI1
aI57
aI115
aI7
aI33
asS'd'
I76
sa(dp152
S'a'
I137
sS's'
I9
sS'm'
(lp153
I1
aI3
aI1
aI11
aI7
aI43
aI81
aI207
aI175
asS'd'
I77
sa(dp154
S'a'
I138
sS's'
I9
sS'm'
(lp155
I1
aI3
aI1
aI1
aI15
aI27
aI63
aI255
aI49
asS'd'
I78
sa(dp156
S'a'
I143
sS's'
I9
sS'm'
(lp157
I1
aI3
aI5
aI3
aI27
aI61
aI105
aI171
aI305
asS'd'
I79
sa(dp158
S'a'
I145
sS's'
I9
sS'm'
(lp159
I1
aI1
aI5
aI3
aI1
aI3
aI57
aI249
aI149
asS'd'
I80
sa(dp160
S'a'
I152
sS's'
I9
sS'm'
(lp161
I1
aI1
aI3
aI5
aI5
aI57
aI15
aI13
aI159
asS'd'
I81
sa(dp162
S'a'
I157
sS's'
I9
sS'm'
(lp163
I1
aI1
aI1
aI11
aI7
aI11
aI105
aI141
aI225
asS'd'
I82
sa(dp164
S'a'
I167
sS's'
I9
sS'm'
(lp165
I1
aI3
aI3
aI5
aI27
aI59
aI121
aI101
aI271
asS'd'
I83
sa(dp166
S'a'
I173
sS's'
I9
sS'm'
(lp167
I1
aI3
aI5
aI9
aI11
aI49
aI51
aI59
aI115
asS'd'
I84
sa(dp168
S'a'
I176
sS's'
I9
sS'm'
(lp169
I1
aI1
aI7
aI1
aI23
aI45
aI125
aI71
aI419
asS'd'
I85
sa(dp170
S'a'
I181
sS's'
I9
sS'm'
(lp171
I1
aI1
aI3
aI5
aI23
aI5
aI105
aI109
aI75
asS'd'
I86
sa(dp172
S'a'
I182
sS's'
I9
sS'm'
(lp173
I1
aI1
aI7
aI15
aI7
aI11
aI67
aI121
aI453
asS'd'
I87
sa(dp174
S'a'
I185
sS's'
I9
sS'm'
(lp175
I1
aI3
aI7
aI3
aI9
aI13
aI31
aI27
aI449
asS'd'
I88
sa(dp176
S'a'
I191
sS's'
I9
sS'm'
(lp177
I1
aI3
aI1
aI15
aI19
aI39
aI39
aI89
aI15
asS'd'
I89
sa(dp178
S'a'
I194
sS's'
I9
sS'm'
(lp179
I1
aI1
aI1
aI1
aI1
aI33
aI73
aI145
aI379
asS'd'
I90
sa(dp180
S'a'
I199
sS's'
I9
sS'm'
(lp181
I1
aI3
aI1
aI15
aI15
aI43
aI29
aI13
aI483
asS'd'
I91
sa(dp182
S'a'
I218
sS's'
I9
sS'm'
(lp183
I1
aI1
aI7
aI3
aI19
aI27
aI85
aI131
aI431
asS'd'
I92
sa(dp184
S'a'
I220
sS's'
I9
sS'm'
(lp185
I1
aI3
aI3
aI3
aI5
aI35
aI23
aI195
aI349
asS'd'
I93
sa(dp186
S'a'
I227
sS's'
I9
sS'm'
(lp187
I1
aI3
aI3
aI7
aI9
aI27
aI39
aI59
aI297
asS'd'
I94
sa(dp188
S'a'
I229
sS's'
I9
sS'm'
(lp189
I1
aI1
aI3
aI9
aI11
aI17
aI13
aI241
aI157
asS'd'
I95
sa(dp190
S'a'
I230
sS's'
I9
sS'm'
(lp191
I1
aI3
aI7
aI15
aI25
aI57
aI33
aI189
aI213
asS'd'
I96
sa(dp192
S'a'
I234
sS's'
I9
sS'm'
(lp193
I1
aI1
aI7
aI1
aI9
aI55
aI73
aI83
aI217
asS'd'
I97
sa(dp194
S'a'
I236
sS's'
I9
sS'm'
(lp195
I1
aI3
aI3
aI13
aI19
aI27
aI23
aI113
aI249
asS'd'
I98
sa(dp196
S'a'
I241
sS's'
I9
sS'm'
(lp197
I1
aI3
aI5
aI3
aI23
aI43
aI3
aI253
aI479
asS'd'
I99
sa(dp198
S'a'
I244
sS's'
I9
sS'm'
(lp199
I1
aI1
aI5
aI5
aI11
aI5
aI45
aI117
aI217
asS'd'
I100
sa(dp200
S'a'
I253
sS's'
I9
sS'm'
(lp201
I1
aI3
aI3
aI7
aI29
aI37
aI33
aI123
aI147
asS'd'
I101
sa(dp202
S'a'
I4
sS's'
I10
sS'm'
(lp203
I1
aI3
aI1
aI15
aI5
aI5
aI37
aI227
aI223
aI459
asS'd'
I102
sa(dp204
S'a'
I13
sS's'
I10
sS'm'
(lp205
I1
aI1
aI7
aI5
aI5
aI39
aI63
aI255
aI135
aI487
asS'd'
I103
sa(dp206
S'a'
I19
sS's'
I10
sS'm'
(lp207
I1
aI3
aI1
aI7
aI9
aI7
aI87
aI249
aI217
aI599
asS'd'
I104
sa(dp208
S'a'
I22
sS's'
I10
sS'm'
(lp209
I1
aI1
aI3
aI13
aI9
aI47
aI7
aI225
aI363
aI247
asS'd'
I105
sa(dp210
S'a'
I50
sS's'
I10
sS'm'
(lp211
I1
aI3
aI7
aI13
aI19
aI13
aI9
aI67
aI9
aI737
asS'd'
I106
sa(dp212
S'a'
I55
sS's'
I10
sS'm'
(lp213
I1
aI3
aI5
aI5
aI19
aI59
aI7
aI41
aI319
aI677
asS'd'
I107
sa(dp214
S'a'
I64
sS's'
I10
sS'm'
(lp215
I1
aI1
aI5
aI3
aI31
aI63
aI15
aI43
aI207
aI789
asS'd'
I108
sa(dp216
S'a'
I69
sS's'
I10
sS'm'
(lp217
I1
aI1
aI7
aI9
aI13
aI39
aI3
aI47
aI497
aI169
asS'd'
I109
sa(dp218
S'a'
I98
sS's'
I10
sS'm'
(lp219
I1
aI3
aI1
aI7
aI21
aI17
aI97
aI19
aI415
aI905
asS'd'
I110
sa(dp220
S'a'
I107
sS's'
I10
sS'm'
(lp221
I1
aI3
aI7
aI1
aI3
aI31
aI71
aI111
aI165
aI127
asS'd'
I111
sa(dp222
S'a'
I115
sS's'
I10
sS'm'
(lp223
I1
aI1
aI5
aI11
aI1
aI61
aI83
aI119
aI203
aI847
asS'd'
I112
sa(dp224
S'a'
I121
sS's'
I10
sS'm'
(lp225
I1
aI3
aI3
aI13
aI9
aI61
aI19
aI97
aI47
aI35
asS'd'
I113
sa(dp226
S'a'
I127
sS's'
I10
sS'm'
(lp227
I1
aI1
aI7
aI7
aI15
aI29
aI63
aI95
aI417
aI469
asS'd'
I114
sa(dp228
S'a'
I134
sS's'
I10
sS'm'
(lp229
I1
aI3
aI1
aI9
aI25
aI9
aI71
aI57
aI213
aI385
asS'd'
I115
sa(dp230
S'a'
I140
sS's'
I10
sS'm'
(lp231
I1
aI3
aI5
aI13
aI31
aI47
aI101
aI57
aI39
aI341
asS'd'
I116
sa(dp232
S'a'
I145
sS's'
I10
sS'm'
(lp233
I1
aI1
aI3
aI3
aI31
aI57
aI125
aI173
aI365
aI551
asS'd'
I117
sa(dp234
S'a'
I152
sS's'
I10
sS'm'
(lp235
I1
aI3
aI7
aI1
aI13
aI57
aI67
aI157
aI451
aI707
asS'd'
I118
sa(dp236
S'a'
I158
sS's'
I10
sS'm'
(lp237
I1
aI1
aI1
aI7
aI21
aI13
aI105
aI89
aI429
aI965
asS'd'
I119
sa(dp238
S'a'
I161
sS's'
I10
sS'm'
(lp239
I1
aI1
aI5
aI9
aI17
aI51
aI45
aI119
aI157
aI141
asS'd'
I120
sa(dp240
S'a'
I171
sS's'
I10
sS'm'
(lp241
I1
aI3
aI7
aI7
aI13
aI45
aI91
aI9
aI129
aI741
asS'd'
I121
sa(dp242
S'a'
I181
sS's'
I10
sS'm'
(lp243
I1
aI3
aI7
aI1
aI23
aI57
aI67
aI141
aI151
aI571
asS'd'
I122
sa(dp244
S'a'
I194
sS's'
I10
sS'm'
(lp245
I1
aI1
aI3
aI11
aI17
aI47
aI93
aI107
aI375
aI157
asS'd'
I123
sa(dp246
S'a'
I199
sS's'
I10
sS'm'
(lp247
I1
aI3
aI3
aI5
aI11
aI21
aI43
aI51
aI169
aI915
asS'd'
I124
sa(dp248
S'a'
I203
sS's'
I10
sS'm'
(lp249
I1
aI1
aI5
aI3
aI15
aI55
aI101
aI67
aI455
aI625
asS'd'
I125
sa(dp250
S'a'
I208
sS's'
I10
sS'm'
(lp251
I1
aI3
aI5
aI9
aI1
aI23
aI29
aI47
aI345
aI595
asS'd'
I126
sa(dp252
S'a'
I227
sS's'
I10
sS'm'
(lp253
I1
aI3
aI7
aI7
aI5
aI49
aI29
aI155
aI323
aI589
asS'd'
I127
sa(dp254
S'a'
I242
sS's'
I10
sS'm'
(lp255
I1
aI3
aI3
aI7
aI5
aI41
aI127
aI61
aI261
aI717
asS'd'
I128
sa(dp256
S'a'
I251
sS's'
I10
sS'm'
(lp257
I1
aI3
aI7
aI7
aI17
aI23
aI117
aI67
aI129
aI1009
asS'd'
I129
sa(dp258
S'a'
I253
sS's'
I10
sS'm'
(lp259
I1
aI1
aI3
aI13
aI11
aI39
aI21
aI207
aI123
aI305
asS'd'
I130
sa(dp260
S'a'
I265
sS's'
I10
sS'm'
(lp261
I1
aI1
aI3
aI9
aI29
aI3
aI95
aI47
aI231
aI73
asS'd'
I131
sa(dp262
S'a'
I266
sS's'
I10
sS'm'
(lp263
I1
aI3
aI1
aI9
aI1
aI29
aI117
aI21
aI441
aI259
asS'd'
I132
sa(dp264
S'a'
I274
sS's'
I10
sS'm'
(lp265
I1
aI3
aI1
aI13
aI21
aI39
aI125
aI211
aI439
aI723
asS'd'
I133
sa(dp266
S'a'
I283
sS's'
I10
sS'm'
(lp267
I1
aI1
aI7
aI3
aI17
aI63
aI115
aI89
aI49
aI773
asS'd'
I134
sa(dp268
S'a'
I289
sS's'
I10
sS'm'
(lp269
I1
aI3
aI7
aI13
aI11
aI33
aI101
aI107
aI63
aI73
asS'd'
I135
sa(dp270
S'a'
I295
sS's'
I10
sS'm'
(lp271
I1
aI1
aI5
aI5
aI13
aI57
aI63
aI135
aI437
aI177
asS'd'
I136
sa(dp272
S'a'
I301
sS's'
I10
sS'm'
(lp273
I1
aI1
aI3
aI7
aI27
aI63
aI93
aI47
aI417
aI483
asS'd'
I137
sa(dp274
S'a'
I316
sS's'
I10
sS'm'
(lp275
I1
aI1
aI3
aI1
aI23
aI29
aI1
aI191
aI49
aI23
asS'd'
I138
sa(dp276
S'a'
I319
sS's'
I10
sS'm'
(lp277
I1
aI1
aI3
aI15
aI25
aI55
aI9
aI101
aI219
aI607
asS'd'
I139
sa(dp278
S'a'
I324
sS's'
I10
sS'm'
(lp279
I1
aI3
aI1
aI7
aI7
aI19
aI51
aI251
aI393
aI307
asS'd'
I140
sa(dp280
S'a'
I346
sS's'
I10
sS'm'
(lp281
I1
aI3
aI3
aI3
aI25
aI55
aI17
aI75
aI337
aI3
asS'd'
I141
sa(dp282
S'a'
I352
sS's'
I10
sS'm'
(lp283
I1
aI1
aI1
aI13
aI25
aI17
aI65
aI45
aI479
aI413
asS'd'
I142
sa(dp284
S'a'
I361
sS's'
I10
sS'm'
(lp285
I1
aI1
aI7
aI7
aI27
aI49
aI99
aI161
aI213
aI727
asS'd'
I143
sa(dp286
S'a'
I367
sS's'
I10
sS'm'
(lp287
I1
aI3
aI5
aI1
aI23
aI5
aI43
aI41
aI251
aI857
asS'd'
I144
sa(dp288
S'a'
I382
sS's'
I10
sS'm'
(lp289
I1
aI3
aI3
aI7
aI11
aI61
aI39
aI87
aI383
aI835
asS'd'
I145
sa(dp290
S'a'
I395
sS's'
I10
sS'm'
(lp291
I1
aI1
aI3
aI15
aI13
aI7
aI29
aI7
aI505
aI923
asS'd'
I146
sa(dp292
S'a'
I398
sS's'
I10
sS'm'
(lp293
I1
aI3
aI7
aI1
aI5
aI31
aI47
aI157
aI445
aI501
asS'd'
I147
sa(dp294
S'a'
I400
sS's'
I10
sS'm'
(lp295
I1
aI1
aI3
aI7
aI1
aI43
aI9
aI147
aI115
aI605
asS'd'
I148
sa(dp296
S'a'
I412
sS's'
I10
sS'm'
(lp297
I1
aI3
aI3
aI13
aI5
aI1
aI119
aI211
aI455
aI1001
asS'd'
I149
sa(dp298
S'a'
I419
sS's'
I10
sS'm'
(lp299
I1
aI1
aI3
aI5
aI13
aI19
aI3
aI243
aI75
aI843
asS'd'
I150
sa(dp300
S'a'
I422
sS's'
I10
sS'm'
(lp301
I1
aI3
aI7
aI7
aI1
aI19
aI91
aI249
aI357
aI589
asS'd'
I151
sa(dp302
S'a'
I426
sS's'
I10
sS'm'
(lp303
I1
aI1
aI1
aI9
aI1
aI25
aI109
aI197
aI279
aI411
asS'd'
I152
sa(dp304
S'a'
I428
sS's'
I10
sS'm'
(lp305
I1
aI3
aI1
aI15
aI23
aI57
aI59
aI135
aI191
aI75
asS'd'
I153
sa(dp306
S'a'
I433
sS's'
I10
sS'm'
(lp307
I1
aI1
aI5
aI15
aI29
aI21
aI39
aI253
aI383
aI349
asS'd'
I154
sa(dp308
S'a'
I446
sS's'
I10
sS'm'
(lp309
I1
aI3
aI3
aI5
aI19
aI45
aI61
aI151
aI199
aI981
asS'd'
I155
sa(dp310
S'a'
I454
sS's'
I10
sS'm'
(lp311
I1
aI3
aI5
aI13
aI9
aI61
aI107
aI141
aI141
aI1
asS'd'
I156
sa(dp312
S'a'
I457
sS's'
I10
sS'm'
(lp313
I1
aI3
aI1
aI11
aI27
aI25
aI85
aI105
aI309
aI979
asS'd'
I157
sa(dp314
S'a'
I472
sS's'
I10
sS'm'
(lp315
I1
aI3
aI3
aI11
aI19
aI7
aI115
aI223
aI349
aI43
asS'd'
I158
sa(dp316
S'a'
I493
sS's'
I10
sS'm'
(lp317
I1
aI1
aI7
aI9
aI21
aI39
aI123
aI21
aI275
aI927
asS'd'
I159
sa(dp318
S'a'
I505
sS's'
I10
sS'm'
(lp319
I1
aI1
aI7
aI13
aI15
aI41
aI47
aI243
aI303
aI437
asS'd'
I160
sa(dp320
S'a'
I508
sS's'
I10
sS'm'
(lp321
I1
aI1
aI1
aI7
aI7
aI3
aI15
aI99
aI409
aI719
asS'd'
I161
sa(dp322
S'a'
I2
sS's'
I11
sS'm'
(lp323
I1
aI3
aI3
aI15
aI27
aI49
aI113
aI123
aI113
aI67
aI469
asS'd'
I162
sa(dp324
S'a'
I11
sS's'
I11
sS'm'
(lp325
I1
aI3
aI7
aI11
aI3
aI23
aI87
aI169
aI119
aI483
aI199
asS'd'
I163
sa(dp326
S'a'
I21
sS's'
I11
sS'm'
(lp327
I1
aI1
aI5
aI15
aI7
aI17
aI109
aI229
aI179
aI213
aI741
asS'd'
I164
sa(dp328
S'a'
I22
sS's'
I11
sS'm'
(lp329
I1
aI1
aI5
aI13
aI11
aI17
aI25
aI135
aI403
aI557
aI1433
asS'd'
I165
sa(dp330
S'a'
I35
sS's'
I11
sS'm'
(lp331
I1
aI3
aI1
aI1
aI1
aI61
aI67
aI215
aI189
aI945
aI1243
asS'd'
I166
sa(dp332
S'a'
I49
sS's'
I11
sS'm'
(lp333
I1
aI1
aI7
aI13
aI17
aI33
aI9
aI221
aI429
aI217
aI1679
asS'd'
I167
sa(dp334
S'a'
I50
sS's'
I11
sS'm'
(lp335
I1
aI1
aI3
aI11
aI27
aI3
aI15
aI93
aI93
aI865
aI1049
asS'd'
I168
sa(dp336
S'a'
I56
sS's'
I11
sS'm'
(lp337
I1
aI3
aI7
aI7
aI25
aI41
aI121
aI35
aI373
aI379
aI1547
asS'd'
I169
sa(dp338
S'a'
I61
sS's'
I11
sS'm'
(lp339
I1
aI3
aI3
aI9
aI11
aI35
aI45
aI205
aI241
aI9
aI59
asS'd'
I170
sa(dp340
S'a'
I70
sS's'
I11
sS'm'
(lp341
I1
aI3
aI1
aI7
aI3
aI51
aI7
aI177
aI53
aI975
aI89
asS'd'
I171
sa(dp342
S'a'
I74
sS's'
I11
sS'm'
(lp343
I1
aI1
aI3
aI5
aI27
aI1
aI113
aI231
aI299
aI759
aI861
asS'd'
I172
sa(dp344
S'a'
I79
sS's'
I11
sS'm'
(lp345
I1
aI3
aI3
aI15
aI25
aI29
aI5
aI255
aI139
aI891
aI2031
asS'd'
I173
sa(dp346
S'a'
I84
sS's'
I11
sS'm'
(lp347
I1
aI3
aI1
aI1
aI13
aI9
aI109
aI193
aI419
aI95
aI17
asS'd'
I174
sa(dp348
S'a'
I88
sS's'
I11
sS'm'
(lp349
I1
aI1
aI7
aI9
aI3
aI7
aI29
aI41
aI135
aI839
aI867
asS'd'
I175
sa(dp350
S'a'
I103
sS's'
I11
sS'm'
(lp351
I1
aI1
aI7
aI9
aI25
aI49
aI123
aI217
aI113
aI909
aI215
asS'd'
I176
sa(dp352
S'a'
I104
sS's'
I11
sS'm'
(lp353
I1
aI1
aI7
aI3
aI23
aI15
aI43
aI133
aI217
aI327
aI901
asS'd'
I177
sa(dp354
S'a'
I112
sS's'
I11
sS'm'
(lp355
I1
aI1
aI3
aI3
aI13
aI53
aI63
aI123
aI477
aI711
aI1387
asS'd'
I178
sa(dp356
S'a'
I115
sS's'
I11
sS'm'
(lp357
I1
aI1
aI3
aI15
aI7
aI29
aI75
aI119
aI181
aI957
aI247
asS'd'
I179
sa(dp358
S'a'
I117
sS's'
I11
sS'm'
(lp359
I1
aI1
aI1
aI11
aI27
aI25
aI109
aI151
aI267
aI99
aI1461
asS'd'
I180
sa(dp360
S'a'
I122
sS's'
I11
sS'm'
(lp361
I1
aI3
aI7
aI15
aI5
aI5
aI53
aI145
aI11
aI725
aI1501
asS'd'
I181
sa(dp362
S'a'
I134
sS's'
I11
sS'm'
(lp363
I1
aI3
aI7
aI1
aI9
aI43
aI71
aI229
aI157
aI607
aI1835
asS'd'
I182
sa(dp364
S'a'
I137
sS's'
I11
sS'm'
(lp365
I1
aI3
aI3
aI13
aI25
aI1
aI5
aI27
aI471
aI349
aI127
asS'd'
I183
sa(dp366
S'a'
I146
sS's'
I11
sS'm'
(lp367
I1
aI1
aI1
aI1
aI23
aI37
aI9
aI221
aI269
aI897
aI1685
asS'd'
I184
sa(dp368
S'a'
I148
sS's'
I11
sS'm'
(lp369
I1
aI1
aI3
aI3
aI31
aI29
aI51
aI19
aI311
aI553
aI1969
asS'd'
I185
sa(dp370
S'a'
I157
sS's'
I11
sS'm'
(lp371
I1
aI3
aI7
aI5
aI5
aI55
aI17
aI39
aI475
aI671
aI1529
asS'd'
I186
sa(dp372
S'a'
I158
sS's'
I11
sS'm'
(lp373
I1
aI1
aI7
aI1
aI1
aI35
aI47
aI27
aI437
aI395
aI1635
asS'd'
I187
sa(dp374
S'a'
I162
sS's'
I11
sS'm'
(lp375
I1
aI1
aI7
aI3
aI13
aI23
aI43
aI135
aI327
aI139
aI389
asS'd'
I188
sa(dp376
S'a'
I164
sS's'
I11
sS'm'
(lp377
I1
aI3
aI7
aI3
aI9
aI25
aI91
aI25
aI429
aI219
aI513
asS'd'
I189
sa(dp378
S'a'
I168
sS's'
I11
sS'm'
(lp379
I1
aI1
aI3
aI5
aI13
aI29
aI119
aI201
aI277
aI157
aI2043
asS'd'
I190
sa(dp380
S'a'
I173
sS's'
I11
sS'm'
(lp381
I1
aI3
aI5
aI3
aI29
aI57
aI13
aI17
aI167
aI739
aI1031
asS'd'
I191
sa(dp382
S'a'
I185
sS's'
I11
sS'm'
(lp383
I1
aI3
aI3
aI5
aI29
aI21
aI95
aI27
aI255
aI679
aI1531
asS'd'
I192
sa(dp384
S'a'
I186
sS's'
I11
sS'm'
(lp385
I1
aI3
aI7
aI15
aI9
aI5
aI21
aI71
aI61
aI961
aI1201
asS'd'
I193
sa(dp386
S'a'
I191
sS's'
I11
sS'm'
(lp387
I1
aI3
aI5
aI13
aI15
aI57
aI33
aI93
aI459
aI867
aI223
asS'd'
I194
sa(dp388
S'a'
I193
sS's'
I11
sS'm'
(lp389
I1
aI1
aI1
aI15
aI17
aI43
aI127
aI191
aI67
aI177
aI1073
asS'd'
I195
sa(dp390
S'a'
I199
sS's'
I11
sS'm'
(lp391
I1
aI1
aI1
aI15
aI23
aI7
aI21
aI199
aI75
aI293
aI1611
asS'd'
I196
sa(dp392
S'a'
I213
sS's'
I11
sS'm'
(lp393
I1
aI3
aI7
aI13
aI15
aI39
aI21
aI149
aI65
aI741
aI319
asS'd'
I197
sa(dp394
S'a'
I214
sS's'
I11
sS'm'
(lp395
I1
aI3
aI7
aI11
aI23
aI13
aI101
aI89
aI277
aI519
aI711
asS'd'
I198
sa(dp396
S'a'
I220
sS's'
I11
sS'm'
(lp397
I1
aI3
aI7
aI15
aI19
aI27
aI85
aI203
aI441
aI97
aI1895
asS'd'
I199
sa(dp398
S'a'
I227
sS's'
I11
sS'm'
(lp399
I1
aI3
aI1
aI3
aI29
aI25
aI21
aI155
aI11
aI191
aI197
asS'd'
I200
sa(dp400
S'a'
I236
sS's'
I11
sS'm'
(lp401
I1
aI1
aI7
aI5
aI27
aI11
aI81
aI101
aI457
aI675
aI1687
asS'd'
I201
sa(dp402
S'a'
I242
sS's'
I11
sS'm'
(lp403
I1
aI3
aI1
aI5
aI25
aI5
aI65
aI193
aI41
aI567
aI781
asS'd'
I202
sa(dp404
S'a'
I251
sS's'
I11
sS'm'
(lp405
I1
aI3
aI1
aI5
aI11
aI15
aI113
aI77
aI411
aI695
aI1111
asS'd'
I203
sa(dp406
S'a'
I256
sS's'
I11
sS'm'
(lp407
I1
aI1
aI3
aI9
aI11
aI53
aI119
aI171
aI55
aI297
aI509
asS'd'
I204
sa(dp408
S'a'
I259
sS's'
I11
sS'm'
(lp409
I1
aI1
aI1
aI1
aI11
aI39
aI113
aI139
aI165
aI347
aI595
asS'd'
I205
sa(dp410
S'a'
I265
sS's'
I11
sS'm'
(lp411
I1
aI3
aI7
aI11
aI9
aI17
aI101
aI13
aI81
aI325
aI1733
asS'd'
I206
sa(dp412
S'a'
I266
sS's'
I11
sS'm'
(lp413
I1
aI3
aI1
aI1
aI21
aI43
aI115
aI9
aI113
aI907
aI645
asS'd'
I207
sa(dp414
S'a'
I276
sS's'
I11
sS'm'
(lp415
I1
aI1
aI7
aI3
aI9
aI25
aI117
aI197
aI159
aI471
aI475
asS'd'
I208
sa(dp416
S'a'
I292
sS's'
I11
sS'm'
(lp417
I1
aI3
aI1
aI9
aI11
aI21
aI57
aI207
aI485
aI613
aI1661
asS'd'
I209
sa(dp418
S'a'
I304
sS's'
I11
sS'm'
(lp419
I1
aI1
aI7
aI7
aI27
aI55
aI49
aI223
aI89
aI85
aI1523
asS'd'
I210
sa(dp420
S'a'
I310
sS's'
I11
sS'm'
(lp421
I1
aI1
aI5
aI3
aI19
aI41
aI45
aI51
aI447
aI299
aI1355
asS'd'
I211
sa(dp422
S'a'
I316
sS's'
I11
sS'm'
(lp423
I1
aI3
aI1
aI13
aI1
aI33
aI117
aI143
aI313
aI187
aI1073
asS'd'
I212
sa(dp424
S'a'
I319
sS's'
I11
sS'm'
(lp425
I1
aI1
aI7
aI7
aI5
aI11
aI65
aI97
aI377
aI377
aI1501
asS'd'
I213
sa(dp426
S'a'
I322
sS's'
I11
sS'm'
(lp427
I1
aI3
aI1
aI1
aI21
aI35
aI95
aI65
aI99
aI23
aI1239
asS'd'
I214
sa(dp428
S'a'
I328
sS's'
I11
sS'm'
(lp429
I1
aI1
aI5
aI9
aI3
aI37
aI95
aI167
aI115
aI425
aI867
asS'd'
I215
sa(dp430
S'a'
I334
sS's'
I11
sS'm'
(lp431
I1
aI3
aI3
aI13
aI1
aI37
aI27
aI189
aI81
aI679
aI773
asS'd'
I216
sa(dp432
S'a'
I339
sS's'
I11
sS'm'
(lp433
I1
aI1
aI3
aI11
aI1
aI61
aI99
aI233
aI429
aI969
aI49
asS'd'
I217
sa(dp434
S'a'
I341
sS's'
I11
sS'm'
(lp435
I1
aI1
aI1
aI7
aI25
aI63
aI99
aI165
aI245
aI793
aI1143
asS'd'
I218
sa(dp436
S'a'
I345
sS's'
I11
sS'm'
(lp437
I1
aI1
aI5
aI11
aI11
aI43
aI55
aI65
aI71
aI283
aI273
asS'd'
I219
sa(dp438
S'a'
I346
sS's'
I11
sS'm'
(lp439
I1
aI1
aI5
aI5
aI9
aI3
aI101
aI251
aI355
aI379
aI1611
asS'd'
I220
sa(dp440
S'a'
I362
sS's'
I11
sS'm'
(lp441
I1
aI1
aI1
aI15
aI21
aI63
aI85
aI99
aI49
aI749
aI1335
asS'd'
I221
sa(dp442
S'a'
I367
sS's'
I11
sS'm'
(lp443
I1
aI1
aI5
aI13
aI27
aI9
aI121
aI43
aI255
aI715
aI289
asS'd'
I222
sa(dp444
S'a'
I372
sS's'
I11
sS'm'
(lp445
I1
aI3
aI1
aI5
aI27
aI19
aI17
aI223
aI77
aI571
aI1415
asS'd'
I223
sa(dp446
S'a'
I375
sS's'
I11
sS'm'
(lp447
I1
aI1
aI5
aI3
aI13
aI59
aI125
aI251
aI195
aI551
aI1737
asS'd'
I224
sa(dp448
S'a'
I376
sS's'
I11
sS'm'
(lp449
I1
aI3
aI3
aI15
aI13
aI27
aI49
aI105
aI389
aI971
aI755
asS'd'
I225
sa(dp450
S'a'
I381
sS's'
I11
sS'm'
(lp451
I1
aI3
aI5
aI15
aI23
aI43
aI35
aI107
aI447
aI763
aI253
asS'd'
I226
sa(dp452
S'a'
I385
sS's'
I11
sS'm'
(lp453
I1
aI3
aI5
aI11
aI21
aI3
aI17
aI39
aI497
aI407
aI611
asS'd'
I227
sa(dp454
S'a'
I388
sS's'
I11
sS'm'
(lp455
I1
aI1
aI7
aI13
aI15
aI31
aI113
aI17
aI23
aI507
aI1995
asS'd'
I228
sa(dp456
S'a'
I392
sS's'
I11
sS'm'
(lp457
I1
aI1
aI7
aI15
aI3
aI15
aI31
aI153
aI423
aI79
aI503
asS'd'
I229
sa(dp458
S'a'
I409
sS's'
I11
sS'm'
(lp459
I1
aI1
aI7
aI9
aI19
aI25
aI23
aI171
aI505
aI923
aI1989
asS'd'
I230
sa(dp460
S'a'
I415
sS's'
I11
sS'm'
(lp461
I1
aI1
aI5
aI9
aI21
aI27
aI121
aI223
aI133
aI87
aI697
asS'd'
I231
sa(dp462
S'a'
I416
sS's'
I11
sS'm'
(lp463
I1
aI1
aI5
aI5
aI9
aI19
aI107
aI99
aI319
aI765
aI1461
asS'd'
I232
sa(dp464
S'a'
I421
sS's'
I11
sS'm'
(lp465
I1
aI1
aI3
aI3
aI19
aI25
aI3
aI101
aI171
aI729
aI187
asS'd'
I233
sa(dp466
S'a'
I428
sS's'
I11
sS'm'
(lp467
I1
aI1
aI3
aI1
aI13
aI23
aI85
aI93
aI291
aI209
aI37
asS'd'
I234
sa(dp468
S'a'
I431
sS's'
I11
sS'm'
(lp469
I1
aI1
aI1
aI15
aI25
aI25
aI77
aI253
aI333
aI947
aI1073
asS'd'
I235
sa(dp470
S'a'
I434
sS's'
I11
sS'm'
(lp471
I1
aI1
aI3
aI9
aI17
aI29
aI55
aI47
aI255
aI305
aI2037
asS'd'
I236
sa(dp472
S'a'
I439
sS's'
I11
sS'm'
(lp473
I1
aI3
aI3
aI9
aI29
aI63
aI9
aI103
aI489
aI939
aI1523
asS'd'
I237
sa(dp474
S'a'
I446
sS's'
I11
sS'm'
(lp475
I1
aI3
aI7
aI15
aI7
aI31
aI89
aI175
aI369
aI339
aI595
asS'd'
I238
sa(dp476
S'a'
I451
sS's'
I11
sS'm'
(lp477
I1
aI3
aI7
aI13
aI25
aI5
aI71
aI207
aI251
aI367
aI665
asS'd'
I239
sa(dp478
S'a'
I453
sS's'
I11
sS'm'
(lp479
I1
aI3
aI3
aI3
aI21
aI25
aI75
aI35
aI31
aI321
aI1603
asS'd'
I240
sa(dp480
S'a'
I457
sS's'
I11
sS'm'
(lp481
I1
aI1
aI1
aI9
aI11
aI1
aI65
aI5
aI11
aI329
aI535
asS'd'
I241
sa(dp482
S'a'
I458
sS's'
I11
sS'm'
(lp483
I1
aI1
aI5
aI3
aI19
aI13
aI17
aI43
aI379
aI485
aI383
asS'd'
I242
sa(dp484
S'a'
I471
sS's'
I11
sS'm'
(lp485
I1
aI3
aI5
aI13
aI13
aI9
aI85
aI147
aI489
aI787
aI1133
asS'd'
I243
sa(dp486
S'a'
I475
sS's'
I11
sS'm'
(lp487
I1
aI3
aI1
aI1
aI5
aI51
aI37
aI129
aI195
aI297
aI1783
asS'd'
I244
sa(dp488
S'a'
I478
sS's'
I11
sS'm'
(lp489
I1
aI1
aI3
aI15
aI19
aI57
aI59
aI181
aI455
aI697
aI2033
asS'd'
I245
sa(dp490
S'a'
I484
sS's'
I11
sS'm'
(lp491
I1
aI3
aI7
aI1
aI27
aI9
aI65
aI145
aI325
aI189
aI201
asS'd'
I246
sa(dp492
S'a'
I493
sS's'
I11
sS'm'
(lp493
I1
aI3
aI1
aI15
aI31
aI23
aI19
aI5
aI485
aI581
aI539
asS'd'
I247
sa(dp494
S'a'
I494
sS's'
I11
sS'm'
(lp495
I1
aI1
aI7
aI13
aI11
aI15
aI65
aI83
aI185
aI847
aI831
asS'd'
I248
sa(dp496
S'a'
I499
sS's'
I11
sS'm'
(lp497
I1
aI3
aI5
aI7
aI7
aI55
aI73
aI15
aI303
aI511
aI1905
asS'd'
I249
sa(dp498
S'a'
I502
sS's'
I11
sS'm'
(lp499
I1
aI3
aI5
aI9
aI7
aI21
aI45
aI15
aI397
aI385
aI597
asS'd'
I250
sa(dp500
S'a'
I517
sS's'
I11
sS'm'
(lp501
I1
aI3
aI7
aI3
aI23
aI13
aI73
aI221
aI511
aI883
aI1265
asS'd'
I251
sa(dp502
S'a'
I518
sS's'
I11
sS'm'
(lp503
I1
aI1
aI3
aI11
aI1
aI51
aI73
aI185
aI33
aI975
aI1441
asS'd'
I252
sa(dp504
S'a'
I524
sS's'
I11
sS'm'
(lp505
I1
aI3
aI3
aI9
aI19
aI59
aI21
aI39
aI339
aI37
aI143
asS'd'
I253
sa(dp506
S'a'
I527
sS's'
I11
sS'm'
(lp507
I1
aI1
aI7
aI1
aI31
aI33
aI19
aI167
aI117
aI635
aI639
asS'd'
I254
sa(dp508
S'a'
I555
sS's'
I11
sS'm'
(lp509
I1
aI1
aI1
aI3
aI5
aI13
aI59
aI83
aI355
aI349
aI1967
asS'd'
I255
sa(dp510
S'a'
I560
sS's'
I11
sS'm'
(lp511
I1
aI1
aI1
aI5
aI19
aI3
aI53
aI133
aI97
aI863
aI983
asS'd'
I256
sa(dp512
S'a'
I565
sS's'
I11
sS'm'
(lp513
I1
aI3
aI1
aI13
aI9
aI41
aI91
aI105
aI173
aI97
aI625
asS'd'
I257
sa(dp514
S'a'
I569
sS's'
I11
sS'm'
(lp515
I1
aI1
aI5
aI3
aI7
aI49
aI115
aI133
aI71
aI231
aI1063
asS'd'
I258
sa(dp516
S'a'
I578
sS's'
I11
sS'm'
(lp517
I1
aI1
aI7
aI5
aI17
aI43
aI47
aI45
aI497
aI547
aI757
asS'd'
I259
sa(dp518
S'a'
I580
sS's'
I11
sS'm'
(lp519
I1
aI3
aI5
aI15
aI21
aI61
aI123
aI191
aI249
aI31
aI631
asS'd'
I260
sa(dp520
S'a'
I587
sS's'
I11
sS'm'
(lp521
I1
aI3
aI7
aI9
aI17
aI7
aI11
aI185
aI127
aI169
aI1951
asS'd'
I261
sa(dp522
S'a'
I589
sS's'
I11
sS'm'
(lp523
I1
aI1
aI5
aI13
aI11
aI11
aI9
aI49
aI29
aI125
aI791
asS'd'
I262
sa(dp524
S'a'
I590
sS's'
I11
sS'm'
(lp525
I1
aI1
aI1
aI15
aI31
aI41
aI13
aI167
aI273
aI429
aI57
asS'd'
I263
sa(dp526
S'a'
I601
sS's'
I11
sS'm'
(lp527
I1
aI3
aI5
aI3
aI27
aI7
aI35
aI209
aI65
aI265
aI1393
asS'd'
I264
sa(dp528
S'a'
I607
sS's'
I11
sS'm'
(lp529
I1
aI3
aI1
aI13
aI31
aI19
aI53
aI143
aI135
aI9
aI1021
asS'd'
I265
sa(dp530
S'a'
I611
sS's'
I11
sS'm'
(lp531
I1
aI1
aI7
aI13
aI31
aI5
aI115
aI153
aI143
aI957
aI623
asS'd'
I266
sa(dp532
S'a'
I614
sS's'
I11
sS'm'
(lp533
I1
aI1
aI5
aI11
aI25
aI19
aI29
aI31
aI297
aI943
aI443
asS'd'
I267
sa(dp534
S'a'
I617
sS's'
I11
sS'm'
(lp535
I1
aI3
aI3
aI5
aI21
aI11
aI127
aI81
aI479
aI25
aI699
asS'd'
I268
sa(dp536
S'a'
I618
sS's'
I11
sS'm'
(lp537
I1
aI1
aI3
aI11
aI25
aI31
aI97
aI19
aI195
aI781
aI705
asS'd'
I269
sa(dp538
S'a'
I625
sS's'
I11
sS'm'
(lp539
I1
aI1
aI5
aI5
aI31
aI11
aI75
aI207
aI197
aI885
aI2037
asS'd'
I270
sa(dp540
S'a'
I628
sS's'
I11
sS'm'
(lp541
I1
aI1
aI1
aI11
aI9
aI23
aI29
aI231
aI307
aI17
aI1497
asS'd'
I271
sa(dp542
S'a'
I635
sS's'
I11
sS'm'
(lp543
I1
aI1
aI5
aI11
aI11
aI43
aI111
aI233
aI307
aI523
aI1259
asS'd'
I272
sa(dp544
S'a'
I641
sS's'
I11
sS'm'
(lp545
I1
aI1
aI7
aI5
aI1
aI21
aI107
aI229
aI343
aI933
aI217
asS'd'
I273
sa(dp546
S'a'
I647
sS's'
I11
sS'm'
(lp547
I1
aI1
aI1
aI11
aI3
aI21
aI125
aI131
aI405
aI599
aI1469
asS'd'
I274
sa(dp548
S'a'
I654
sS's'
I11
sS'm'
(lp549
I1
aI3
aI5
aI5
aI9
aI39
aI33
aI81
aI389
aI151
aI811
asS'd'
I275
sa(dp550
S'a'
I659
sS's'
I11
sS'm'
(lp551
I1
aI1
aI7
aI7
aI7
aI1
aI59
aI223
aI265
aI529
aI2021
asS'd'
I276
sa(dp552
S'a'
I662
sS's'
I11
sS'm'
(lp553
I1
aI3
aI1
aI3
aI9
aI23
aI85
aI181
aI47
aI265
aI49
asS'd'
I277
sa(dp554
S'a'
I672
sS's'
I11
sS'm'
(lp555
I1
aI3
aI5
aI11
aI19
aI23
aI9
aI7
aI157
aI299
aI1983
asS'd'
I278
sa(dp556
S'a'
I675
sS's'
I11
sS'm'
(lp557
I1
aI3
aI1
aI5
aI15
aI5
aI21
aI105
aI29
aI339
aI1041
asS'd'
I279
sa(dp558
S'a'
I682
sS's'
I11
sS'm'
(lp559
I1
aI1
aI1
aI1
aI5
aI33
aI65
aI85
aI111
aI705
aI479
asS'd'
I280
sa(dp560
S'a'
I684
sS's'
I11
sS'm'
(lp561
I1
aI1
aI1
aI7
aI9
aI35
aI77
aI87
aI151
aI321
aI101
asS'd'
I281
sa(dp562
S'a'
I689
sS's'
I11
sS'm'
(lp563
I1
aI1
aI5
aI7
aI17
aI1
aI51
aI197
aI175
aI811
aI1229
asS'd'
I282
sa(dp564
S'a'
I695
sS's'
I11
sS'm'
(lp565
I1
aI3
aI3
aI15
aI23
aI37
aI85
aI185
aI239
aI543
aI731
asS'd'
I283
sa(dp566
S'a'
I696
sS's'
I11
sS'm'
(lp567
I1
aI3
aI1
aI7
aI7
aI55
aI111
aI109
aI289
aI439
aI243
asS'd'
I284
sa(dp568
S'a'
I713
sS's'
I11
sS'm'
(lp569
I1
aI1
aI7
aI11
aI17
aI53
aI35
aI217
aI259
aI853
aI1667
asS'd'
I285
sa(dp570
S'a'
I719
sS's'
I11
sS'm'
(lp571
I1
aI3
aI1
aI9
aI1
aI63
aI87
aI17
aI73
aI565
aI1091
asS'd'
I286
sa(dp572
S'a'
I724
sS's'
I11
sS'm'
(lp573
I1
aI1
aI3
aI3
aI11
aI41
aI1
aI57
aI295
aI263
aI1029
asS'd'
I287
sa(dp574
S'a'
I733
sS's'
I11
sS'm'
(lp575
I1
aI1
aI5
aI1
aI27
aI45
aI109
aI161
aI411
aI421
aI1395
asS'd'
I288
sa(dp576
S'a'
I734
sS's'
I11
sS'm'
(lp577
I1
aI3
aI5
aI11
aI25
aI35
aI47
aI191
aI339
aI417
aI1727
asS'd'
I289
sa(dp578
S'a'
I740
sS's'
I11
sS'm'
(lp579
I1
aI1
aI5
aI15
aI21
aI1
aI93
aI251
aI351
aI217
aI1767
asS'd'
I290
sa(dp580
S'a'
I747
sS's'
I11
sS'm'
(lp581
I1
aI3
aI3
aI11
aI3
aI7
aI75
aI155
aI313
aI211
aI491
asS'd'
I291
sa(dp582
S'a'
I749
sS's'
I11
sS'm'
(lp583
I1
aI3
aI3
aI5
aI11
aI9
aI101
aI161
aI453
aI913
aI1067
asS'd'
I292
sa(dp584
S'a'
I752
sS's'
I11
sS'm'
(lp585
I1
aI1
aI3
aI1
aI15
aI45
aI127
aI141
aI163
aI727
aI1597
asS'd'
I293
sa(dp586
S'a'
I755
sS's'
I11
sS'm'
(lp587
I1
aI3
aI3
aI7
aI1
aI33
aI63
aI73
aI73
aI341
aI1691
asS'd'
I294
sa(dp588
S'a'
I762
sS's'
I11
sS'm'
(lp589
I1
aI3
aI5
aI13
aI15
aI39
aI53
aI235
aI77
aI99
aI949
asS'd'
I295
sa(dp590
S'a'
I770
sS's'
I11
sS'm'
(lp591
I1
aI1
aI5
aI13
aI31
aI17
aI97
aI13
aI215
aI301
aI1927
asS'd'
I296
sa(dp592
S'a'
I782
sS's'
I11
sS'm'
(lp593
I1
aI1
aI7
aI1
aI1
aI37
aI91
aI93
aI441
aI251
aI1131
asS'd'
I297
sa(dp594
S'a'
I784
sS's'
I11
sS'm'
(lp595
I1
aI3
aI7
aI9
aI25
aI5
aI105
aI69
aI81
aI943
aI1459
asS'd'
I298
sa(dp596
S'a'
I787
sS's'
I11
sS'm'
(lp597
I1
aI3
aI7
aI11
aI31
aI43
aI13
aI209
aI27
aI1017
aI501
asS'd'
I299
sa(dp598
S'a'
I789
sS's'
I11
sS'm'
(lp599
I1
aI1
aI7
aI15
aI1
aI33
aI31
aI233
aI161
aI507
aI387
asS'd'
I300
sa(dp600
S'a'
I793
sS's'
I11
sS'm'
(lp601
I1
aI3
aI3
aI5
aI5
aI53
aI33
aI177
aI503
aI627
aI1927
asS'd'
I301
sa(dp602
S'a'
I796
sS's'
I11
sS'm'
(lp603
I1
aI1
aI7
aI11
aI7
aI61
aI119
aI31
aI457
aI229
aI1875
asS'd'
I302
sa(dp604
S'a'
I803
sS's'
I11
sS'm'
(lp605
I1
aI1
aI5
aI15
aI19
aI5
aI53
aI201
aI157
aI885
aI1057
asS'd'
I303
sa(dp606
S'a'
I805
sS's'
I11
sS'm'
(lp607
I1
aI3
aI7
aI9
aI1
aI35
aI51
aI113
aI249
aI425
aI1009
asS'd'
I304
sa(dp608
S'a'
I810
sS's'
I11
sS'm'
(lp609
I1
aI3
aI5
aI7
aI21
aI53
aI37
aI155
aI119
aI345
aI631
asS'd'
I305
sa(dp610
S'a'
I815
sS's'
I11
sS'm'
(lp611
I1
aI3
aI5
aI7
aI15
aI31
aI109
aI69
aI503
aI595
aI1879
asS'd'
I306
sa(dp612
S'a'
I824
sS's'
I11
sS'm'
(lp613
I1
aI3
aI3
aI1
aI25
aI35
aI65
aI131
aI403
aI705
aI503
asS'd'
I307
sa(dp614
S'a'
I829
sS's'
I11
sS'm'
(lp615
I1
aI3
aI7
aI7
aI19
aI33
aI11
aI153
aI45
aI633
aI499
asS'd'
I308
sa(dp616
S'a'
I830
sS's'
I11
sS'm'
(lp617
I1
aI3
aI3
aI5
aI11
aI3
aI29
aI93
aI487
aI33
aI703
asS'd'
I309
sa(dp618
S'a'
I832
sS's'
I11
sS'm'
(lp619
I1
aI1
aI3
aI15
aI21
aI53
aI107
aI179
aI387
aI927
aI1757
asS'd'
I310
sa(dp620
S'a'
I841
sS's'
I11
sS'm'
(lp621
I1
aI1
aI3
aI7
aI21
aI45
aI51
aI147
aI175
aI317
aI361
asS'd'
I311
sa(dp622
S'a'
I847
sS's'
I11
sS'm'
(lp623
I1
aI1
aI1
aI7
aI7
aI13
aI15
aI243
aI269
aI795
aI1965
asS'd'
I312
sa(dp624
S'a'
I849
sS's'
I11
sS'm'
(lp625
I1
aI1
aI3
aI5
aI19
aI33
aI57
aI115
aI443
aI537
aI627
asS'd'
I313
sa(dp626
S'a'
I861
sS's'
I11
sS'm'
(lp627
I1
aI3
aI3
aI9
aI3
aI39
aI25
aI61
aI185
aI717
aI1049
asS'd'
I314
sa(dp628
S'a'
I871
sS's'
I11
sS'm'
(lp629
I1
aI3
aI7
aI3
aI7
aI37
aI107
aI153
aI7
aI269
aI1581
asS'd'
I315
sa(dp630
S'a'
I878
sS's'
I11
sS'm'
(lp631
I1
aI1
aI7
aI3
aI7
aI41
aI91
aI41
aI145
aI489
aI1245
asS'd'
I316
sa(dp632
S'a'
I889
sS's'
I11
sS'm'
(lp633
I1
aI1
aI5
aI9
aI7
aI7
aI105
aI81
aI403
aI407
aI283
asS'd'
I317
sa(dp634
S'a'
I892
sS's'
I11
sS'm'
(lp635
I1
aI1
aI7
aI9
aI27
aI55
aI29
aI77
aI193
aI963
aI949
asS'd'
I318
sa(dp636
S'a'
I901
sS's'
I11
sS'm'
(lp637
I1
aI1
aI5
aI3
aI25
aI51
aI107
aI63
aI403
aI917
aI815
asS'd'
I319
sa(dp638
S'a'
I908
sS's'
I11
sS'm'
(lp639
I1
aI1
aI7
aI3
aI7
aI61
aI19
aI51
aI457
aI599
aI535
asS'd'
I320
sa(dp640
S'a'
I920
sS's'
I11
sS'm'
(lp641
I1
aI3
aI7
aI1
aI23
aI51
aI105
aI153
aI239
aI215
aI1847
asS'd'
I321
sa(dp642
S'a'
I923
sS's'
I11
sS'm'
(lp643
I1
aI1
aI3
aI5
aI27
aI23
aI79
aI49
aI495
aI45
aI1935
asS'd'
I322
sa(dp644
S'a'
I942
sS's'
I11
sS'm'
(lp645
I1
aI1
aI1
aI11
aI11
aI47
aI55
aI133
aI495
aI999
aI1461
asS'd'
I323
sa(dp646
S'a'
I949
sS's'
I11
sS'm'
(lp647
I1
aI1
aI3
aI15
aI27
aI51
aI93
aI17
aI355
aI763
aI1675
asS'd'
I324
sa(dp648
S'a'
I950
sS's'
I11
sS'm'
(lp649
I1
aI3
aI1
aI3
aI1
aI3
aI79
aI119
aI499
aI17
aI995
asS'd'
I325
sa(dp650
S'a'
I954
sS's'
I11
sS'm'
(lp651
I1
aI1
aI1
aI1
aI15
aI43
aI45
aI17
aI167
aI973
aI799
asS'd'
I326
sa(dp652
S'a'
I961
sS's'
I11
sS'm'
(lp653
I1
aI1
aI1
aI3
aI27
aI49
aI89
aI29
aI483
aI913
aI2023
asS'd'
I327
sa(dp654
S'a'
I968
sS's'
I11
sS'm'
(lp655
I1
aI1
aI3
aI3
aI5
aI11
aI75
aI7
aI41
aI851
aI611
asS'd'
I328
sa(dp656
S'a'
I971
sS's'
I11
sS'm'
(lp657
I1
aI3
aI1
aI3
aI7
aI57
aI39
aI123
aI257
aI283
aI507
asS'd'
I329
sa(dp658
S'a'
I973
sS's'
I11
sS'm'
(lp659
I1
aI3
aI3
aI11
aI27
aI23
aI113
aI229
aI187
aI299
aI133
asS'd'
I330
sa(dp660
S'a'
I979
sS's'
I11
sS'm'
(lp661
I1
aI1
aI3
aI13
aI9
aI63
aI101
aI77
aI451
aI169
aI337
asS'd'
I331
sa(dp662
S'a'
I982
sS's'
I11
sS'm'
(lp663
I1
aI3
aI7
aI3
aI3
aI59
aI45
aI195
aI229
aI415
aI409
asS'd'
I332
sa(dp664
S'a'
I986
sS's'
I11
sS'm'
(lp665
I1
aI3
aI5
aI3
aI11
aI19
aI71
aI93
aI43
aI857
aI369
asS'd'
I333
sa(dp666
S'a'
I998
sS's'
I11
sS'm'
(lp667
I1
aI3
aI7
aI9
aI19
aI33
aI115
aI19
aI241
aI703
aI247
asS'd'
I334
sa(dp668
S'a'
I1001
sS's'
I11
sS'm'
(lp669
I1
aI3
aI5
aI11
aI5
aI35
aI21
aI155
aI463
aI1005
aI1073
asS'd'
I335
sa(dp670
S'a'
I1010
sS's'
I11
sS'm'
(lp671
I1
aI3
aI7
aI3
aI25
aI15
aI109
aI83
aI93
aI69
aI1189
asS'd'
I336
sa(dp672
S'a'
I1012
sS's'
I11
sS'm'
(lp673
I1
aI3
aI5
aI7
aI5
aI21
aI93
aI133
aI135
aI167
aI903
asS'd'
I337
sa(dp674
S'a'
I41
sS's'
I12
sS'm'
(lp675
I1
aI1
aI7
aI7
aI3
aI59
aI121
aI161
aI285
aI815
aI1769
aI3705
asS'd'
I338
sa(dp676
S'a'
I52
sS's'
I12
sS'm'
(lp677
I1
aI3
aI1
aI1
aI3
aI47
aI103
aI171
aI381
aI609
aI185
aI373
asS'd'
I339
sa(dp678
S'a'
I61
sS's'
I12
sS'm'
(lp679
I1
aI3
aI3
aI15
aI23
aI33
aI107
aI131
aI441
aI445
aI689
aI2059
asS'd'
I340
sa(dp680
S'a'
I62
sS's'
I12
sS'm'
(lp681
I1
aI3
aI3
aI11
aI7
aI53
aI101
aI167
aI435
aI803
aI1255
aI3781
asS'd'
I341
sa(dp682
S'a'
I76
sS's'
I12
sS'm'
(lp683
I1
aI1
aI5
aI11
aI15
aI59
aI41
aI19
aI135
aI835
aI1263
aI505
asS'd'
I342
sa(dp684
S'a'
I104
sS's'
I12
sS'm'
(lp685
I1
aI1
aI7
aI11
aI21
aI49
aI23
aI219
aI127
aI961
aI1065
aI385
asS'd'
I343
sa(dp686
S'a'
I117
sS's'
I12
sS'm'
(lp687
I1
aI3
aI5
aI15
aI7
aI47
aI117
aI217
aI45
aI731
aI1639
aI733
asS'd'
I344
sa(dp688
S'a'
I131
sS's'
I12
sS'm'
(lp689
I1
aI1
aI7
aI11
aI27
aI57
aI91
aI87
aI81
aI35
aI1269
aI1007
asS'd'
I345
sa(dp690
S'a'
I143
sS's'
I12
sS'm'
(lp691
I1
aI1
aI3
aI11
aI15
aI37
aI53
aI219
aI193
aI937
aI1899
aI3733
asS'd'
I346
sa(dp692
S'a'
I145
sS's'
I12
sS'm'
(lp693
I1
aI3
aI5
aI3
aI13
aI11
aI27
aI19
aI199
aI393
aI965
aI2195
asS'd'
I347
sa(dp694
S'a'
I157
sS's'
I12
sS'm'
(lp695
I1
aI3
aI1
aI3
aI5
aI1
aI37
aI173
aI413
aI1023
aI553
aI409
asS'd'
I348
sa(dp696
S'a'
I167
sS's'
I12
sS'm'
(lp697
I1
aI3
aI1
aI7
aI15
aI29
aI123
aI95
aI255
aI373
aI1799
aI3841
asS'd'
I349
sa(dp698
S'a'
I171
sS's'
I12
sS'm'
(lp699
I1
aI3
aI5
aI13
aI21
aI57
aI51
aI17
aI511
aI195
aI1157
aI1831
asS'd'
I350
sa(dp700
S'a'
I176
sS's'
I12
sS'm'
(lp701
I1
aI1
aI1
aI15
aI29
aI19
aI7
aI73
aI295
aI519
aI587
aI3523
asS'd'
I351
sa(dp702
S'a'
I181
sS's'
I12
sS'm'
(lp703
I1
aI1
aI5
aI13
aI13
aI35
aI115
aI191
aI123
aI535
aI717
aI1661
asS'd'
I352
sa(dp704
S'a'
I194
sS's'
I12
sS'm'
(lp705
I1
aI3
aI3
aI5
aI23
aI21
aI47
aI251
aI379
aI921
aI1119
aI297
asS'd'
I353
sa(dp706
S'a'
I217
sS's'
I12
sS'm'
(lp707
I1
aI3
aI3
aI9
aI29
aI53
aI121
aI201
aI135
aI193
aI523
aI2943
asS'd'
I354
sa(dp708
S'a'
I236
sS's'
I12
sS'm'
(lp709
I1
aI1
aI1
aI7
aI29
aI45
aI125
aI9
aI99
aI867
aI425
aI601
asS'd'
I355
sa(dp710
S'a'
I239
sS's'
I12
sS'm'
(lp711
I1
aI3
aI1
aI9
aI13
aI15
aI67
aI181
aI109
aI293
aI1305
aI3079
asS'd'
I356
sa(dp712
S'a'
I262
sS's'
I12
sS'm'
(lp713
I1
aI3
aI3
aI9
aI5
aI35
aI15
aI209
aI305
aI87
aI767
aI2795
asS'd'
I357
sa(dp714
S'a'
I283
sS's'
I12
sS'm'
(lp715
I1
aI3
aI3
aI11
aI27
aI57
aI113
aI123
aI179
aI643
aI149
aI523
asS'd'
I358
sa(dp716
S'a'
I286
sS's'
I12
sS'm'
(lp717
I1
aI1
aI3
aI15
aI11
aI17
aI67
aI223
aI63
aI657
aI335
aI3309
asS'd'
I359
sa(dp718
S'a'
I307
sS's'
I12
sS'm'
(lp719
I1
aI1
aI1
aI9
aI25
aI29
aI109
aI159
aI39
aI513
aI571
aI1761
asS'd'
I360
sa(dp720
S'a'
I313
sS's'
I12
sS'm'
(lp721
I1
aI1
aI3
aI1
aI5
aI63
aI75
aI19
aI455
aI601
aI123
aI691
asS'd'
I361
sa(dp722
S'a'
I319
sS's'
I12
sS'm'
(lp723
I1
aI1
aI1
aI3
aI21
aI5
aI45
aI169
aI377
aI513
aI1951
aI2565
asS'd'
I362
sa(dp724
S'a'
I348
sS's'
I12
sS'm'
(lp725
I1
aI1
aI3
aI11
aI3
aI33
aI119
aI69
aI253
aI907
aI805
aI1449
asS'd'
I363
sa(dp726
S'a'
I352
sS's'
I12
sS'm'
(lp727
I1
aI1
aI5
aI13
aI31
aI15
aI17
aI7
aI499
aI61
aI687
aI1867
asS'd'
I364
sa(dp728
S'a'
I357
sS's'
I12
sS'm'
(lp729
I1
aI3
aI7
aI11
aI17
aI33
aI73
aI77
aI299
aI243
aI641
aI2345
asS'd'
I365
sa(dp730
S'a'
I391
sS's'
I12
sS'm'
(lp731
I1
aI1
aI7
aI11
aI9
aI35
aI31
aI235
aI359
aI647
aI379
aI1161
asS'd'
I366
sa(dp732
S'a'
I398
sS's'
I12
sS'm'
(lp733
I1
aI3
aI3
aI15
aI31
aI25
aI5
aI67
aI33
aI45
aI437
aI4067
asS'd'
I367
sa(dp734
S'a'
I400
sS's'
I12
sS'm'
(lp735
I1
aI1
aI3
aI11
aI7
aI17
aI37
aI87
aI333
aI253
aI1517
aI2921
asS'd'
I368
sa(dp736
S'a'
I412
sS's'
I12
sS'm'
(lp737
I1
aI1
aI7
aI15
aI7
aI15
aI107
aI189
aI153
aI769
aI1521
aI3427
asS'd'
I369
sa(dp738
S'a'
I415
sS's'
I12
sS'm'
(lp739
I1
aI3
aI5
aI13
aI5
aI61
aI113
aI37
aI293
aI393
aI113
aI43
asS'd'
I370
sa(dp740
S'a'
I422
sS's'
I12
sS'm'
(lp741
I1
aI1
aI1
aI15
aI29
aI43
aI107
aI31
aI167
aI147
aI301
aI1021
asS'd'
I371
sa(dp742
S'a'
I440
sS's'
I12
sS'm'
(lp743
I1
aI1
aI1
aI13
aI3
aI1
aI35
aI93
aI195
aI181
aI2027
aI1491
asS'd'
I372
sa(dp744
S'a'
I460
sS's'
I12
sS'm'
(lp745
I1
aI3
aI3
aI3
aI13
aI33
aI77
aI199
aI153
aI221
aI1699
aI3671
asS'd'
I373
sa(dp746
S'a'
I465
sS's'
I12
sS'm'
(lp747
I1
aI3
aI5
aI13
aI7
aI49
aI123
aI155
aI495
aI681
aI819
aI809
asS'd'
I374
sa(dp748
S'a'
I468
sS's'
I12
sS'm'
(lp749
I1
aI3
aI5
aI15
aI27
aI61
aI117
aI189
aI183
aI887
aI617
aI4053
asS'd'
I375
sa(dp750
S'a'
I515
sS's'
I12
sS'm'
(lp751
I1
aI1
aI1
aI7
aI31
aI59
aI125
aI235
aI389
aI369
aI447
aI1039
asS'd'
I376
sa(dp752
S'a'
I536
sS's'
I12
sS'm'
(lp753
I1
aI3
aI5
aI1
aI5
aI39
aI115
aI89
aI249
aI377
aI431
aI3747
asS'd'
I377
sa(dp754
S'a'
I539
sS's'
I12
sS'm'
(lp755
I1
aI1
aI1
aI5
aI7
aI47
aI59
aI157
aI77
aI445
aI699
aI3439
asS'd'
I378
sa(dp756
S'a'
I551
sS's'
I12
sS'm'
(lp757
I1
aI1
aI3
aI5
aI11
aI21
aI19
aI75
aI11
aI599
aI1575
aI735
asS'd'
I379
sa(dp758
S'a'
I558
sS's'
I12
sS'm'
(lp759
I1
aI3
aI5
aI3
aI19
aI13
aI41
aI69
aI199
aI143
aI1761
aI3215
asS'd'
I380
sa(dp760
S'a'
I563
sS's'
I12
sS'm'
(lp761
I1
aI3
aI5
aI7
aI19
aI43
aI25
aI41
aI41
aI11
aI1647
aI2783
asS'd'
I381
sa(dp762
S'a'
I570
sS's'
I12
sS'm'
(lp763
I1
aI3
aI1
aI9
aI19
aI45
aI111
aI97
aI405
aI399
aI457
aI3219
asS'd'
I382
sa(dp764
S'a'
I595
sS's'
I12
sS'm'
(lp765
I1
aI1
aI3
aI1
aI23
aI15
aI65
aI121
aI59
aI985
aI829
aI2259
asS'd'
I383
sa(dp766
S'a'
I598
sS's'
I12
sS'm'
(lp767
I1
aI1
aI3
aI7
aI17
aI13
aI107
aI229
aI75
aI551
aI1299
aI2363
asS'd'
I384
sa(dp768
S'a'
I617
sS's'
I12
sS'm'
(lp769
I1
aI1
aI5
aI5
aI21
aI57
aI23
aI199
aI509
aI139
aI2007
aI3875
asS'd'
I385
sa(dp770
S'a'
I647
sS's'
I12
sS'm'
(lp771
I1
aI3
aI1
aI11
aI19
aI53
aI15
aI229
aI215
aI741
aI695
aI823
asS'd'
I386
sa(dp772
S'a'
I654
sS's'
I12
sS'm'
(lp773
I1
aI3
aI7
aI1
aI29
aI3
aI17
aI163
aI417
aI559
aI549
aI319
asS'd'
I387
sa(dp774
S'a'
I678
sS's'
I12
sS'm'
(lp775
I1
aI3
aI1
aI13
aI17
aI9
aI47
aI133
aI365
aI7
aI1937
aI1071
asS'd'
I388
sa(dp776
S'a'
I713
sS's'
I12
sS'm'
(lp777
I1
aI3
aI5
aI7
aI19
aI37
aI55
aI163
aI301
aI249
aI689
aI2327
asS'd'
I389
sa(dp778
S'a'
I738
sS's'
I12
sS'm'
(lp779
I1
aI3
aI5
aI13
aI11
aI23
aI61
aI205
aI257
aI377
aI615
aI1457
asS'd'
I390
sa(dp780
S'a'
I747
sS's'
I12
sS'm'
(lp781
I1
aI3
aI5
aI1
aI23
aI37
aI13
aI75
aI331
aI495
aI579
aI3367
asS'd'
I391
sa(dp782
S'a'
I750
sS's'
I12
sS'm'
(lp783
I1
aI1
aI1
aI9
aI1
aI23
aI49
aI129
aI475
aI543
aI883
aI2531
asS'd'
I392
sa(dp784
S'a'
I757
sS's'
I12
sS'm'
(lp785
I1
aI3
aI1
aI5
aI23
aI59
aI51
aI35
aI343
aI695
aI219
aI369
asS'd'
I393
sa(dp786
S'a'
I772
sS's'
I12
sS'm'
(lp787
I1
aI3
aI3
aI1
aI27
aI17
aI63
aI97
aI71
aI507
aI1929
aI613
asS'd'
I394
sa(dp788
S'a'
I803
sS's'
I12
sS'm'
(lp789
I1
aI1
aI5
aI1
aI21
aI31
aI11
aI109
aI247
aI409
aI1817
aI2173
asS'd'
I395
sa(dp790
S'a'
I810
sS's'
I12
sS'm'
(lp791
I1
aI1
aI3
aI15
aI23
aI9
aI7
aI209
aI301
aI23
aI147
aI1691
asS'd'
I396
sa(dp792
S'a'
I812
sS's'
I12
sS'm'
(lp793
I1
aI1
aI7
aI5
aI5
aI19
aI37
aI229
aI249
aI277
aI1115
aI2309
asS'd'
I397
sa(dp794
S'a'
I850
sS's'
I12
sS'm'
(lp795
I1
aI1
aI1
aI5
aI5
aI63
aI5
aI249
aI285
aI431
aI343
aI2467
asS'd'
I398
sa(dp796
S'a'
I862
sS's'
I12
sS'm'
(lp797
I1
aI1
aI1
aI11
aI7
aI45
aI35
aI75
aI505
aI537
aI29
aI2919
asS'd'
I399
sa(dp798
S'a'
I906
sS's'
I12
sS'm'
(lp799
I1
aI3
aI5
aI15
aI11
aI39
aI15
aI63
aI263
aI9
aI199
aI445
asS'd'
I400
sa(dp800
S'a'
I908
sS's'
I12
sS'm'
(lp801
I1
aI3
aI3
aI3
aI27
aI63
aI53
aI171
aI227
aI63
aI1049
aI827
asS'd'
I401
sa(dp802
S'a'
I929
sS's'
I12
sS'm'
(lp803
I1
aI1
aI3
aI13
aI7
aI11
aI115
aI183
aI179
aI937
aI1785
aI381
asS'd'
I402
sa(dp804
S'a'
I930
sS's'
I12
sS'm'
(lp805
I1
aI3
aI1
aI11
aI13
aI15
aI107
aI81
aI53
aI295
aI1785
aI3757
asS'd'
I403
sa(dp806
S'a'
I954
sS's'
I12
sS'm'
(lp807
I1
aI3
aI3
aI13
aI11
aI5
aI109
aI243
aI3
aI505
aI323
aI1373
asS'd'
I404
sa(dp808
S'a'
I964
sS's'
I12
sS'm'
(lp809
I1
aI3
aI3
aI11
aI21
aI51
aI17
aI177
aI381
aI937
aI1263
aI3889
asS'd'
I405
sa(dp810
S'a'
I982
sS's'
I12
sS'm'
(lp811
I1
aI3
aI5
aI9
aI27
aI25
aI85
aI193
aI143
aI573
aI1189
aI2995
asS'd'
I406
sa(dp812
S'a'
I985
sS's'
I12
sS'm'
(lp813
I1
aI3
aI5
aI11
aI13
aI9
aI81
aI21
aI159
aI953
aI91
aI1751
asS'd'
I407
sa(dp814
S'a'
I991
sS's'
I12
sS'm'
(lp815
I1
aI1
aI3
aI3
aI27
aI61
aI11
aI253
aI391
aI333
aI1105
aI635
asS'd'
I408
sa(dp816
S'a'
I992
sS's'
I12
sS'm'
(lp817
I1
aI3
aI3
aI15
aI9
aI57
aI95
aI81
aI419
aI735
aI251
aI1141
asS'd'
I409
sa(dp818
S'a'
I1067
sS's'
I12
sS'm'
(lp819
I1
aI1
aI5
aI9
aI31
aI39
aI59
aI13
aI319
aI807
aI1241
aI2433
asS'd'
I410
sa(dp820
S'a'
I1070
sS's'
I12
sS'm'
(lp821
I1
aI3
aI3
aI5
aI27
aI13
aI107
aI141
aI423
aI937
aI2027
aI3233
asS'd'
I411
sa(dp822
S'a'
I1096
sS's'
I12
sS'm'
(lp823
I1
aI3
aI3
aI9
aI9
aI25
aI125
aI23
aI443
aI835
aI1245
aI847
asS'd'
I412
sa(dp824
S'a'
I1099
sS's'
I12
sS'm'
(lp825
I1
aI1
aI7
aI15
aI17
aI17
aI83
aI107
aI411
aI285
aI847
aI1571
asS'd'
I413
sa(dp826
S'a'
I1116
sS's'
I12
sS'm'
(lp827
I1
aI1
aI3
aI13
aI29
aI61
aI37
aI81
aI349
aI727
aI1453
aI1957
asS'd'
I414
sa(dp828
S'a'
I1143
sS's'
I12
sS'm'
(lp829
I1
aI3
aI7
aI11
aI31
aI13
aI59
aI77
aI273
aI591
aI1265
aI1533
asS'd'
I415
sa(dp830
S'a'
I1165
sS's'
I12
sS'm'
(lp831
I1
aI1
aI7
aI7
aI13
aI17
aI25
aI25
aI187
aI329
aI347
aI1473
asS'd'
I416
sa(dp832
S'a'
I1178
sS's'
I12
sS'm'
(lp833
I1
aI3
aI7
aI7
aI5
aI51
aI37
aI99
aI221
aI153
aI503
aI2583
asS'd'
I417
sa(dp834
S'a'
I1184
sS's'
I12
sS'm'
(lp835
I1
aI3
aI1
aI13
aI19
aI27
aI11
aI69
aI181
aI479
aI1183
aI3229
asS'd'
I418
sa(dp836
S'a'
I1202
sS's'
I12
sS'm'
(lp837
I1
aI3
aI3
aI13
aI23
aI21
aI103
aI147
aI323
aI909
aI947
aI315
asS'd'
I419
sa(dp838
S'a'
I1213
sS's'
I12
sS'm'
(lp839
I1
aI3
aI1
aI3
aI23
aI1
aI31
aI59
aI93
aI513
aI45
aI2271
asS'd'
I420
sa(dp840
S'a'
I1221
sS's'
I12
sS'm'
(lp841
I1
aI3
aI5
aI1
aI7
aI43
aI109
aI59
aI231
aI41
aI1515
aI2385
asS'd'
I421
sa(dp842
S'a'
I1240
sS's'
I12
sS'm'
(lp843
I1
aI3
aI1
aI5
aI31
aI57
aI49
aI223
aI283
aI1013
aI11
aI701
asS'd'
I422
sa(dp844
S'a'
I1246
sS's'
I12
sS'm'
(lp845
I1
aI1
aI5
aI1
aI19
aI53
aI55
aI31
aI31
aI299
aI495
aI693
asS'd'
I423
sa(dp846
S'a'
I1252
sS's'
I12
sS'm'
(lp847
I1
aI3
aI3
aI9
aI5
aI33
aI77
aI253
aI427
aI791
aI731
aI1019
asS'd'
I424
sa(dp848
S'a'
I1255
sS's'
I12
sS'm'
(lp849
I1
aI3
aI7
aI11
aI1
aI9
aI119
aI203
aI53
aI877
aI1707
aI3499
asS'd'
I425
sa(dp850
S'a'
I1267
sS's'
I12
sS'm'
(lp851
I1
aI1
aI3
aI7
aI13
aI39
aI55
aI159
aI423
aI113
aI1653
aI3455
asS'd'
I426
sa(dp852
S'a'
I1293
sS's'
I12
sS'm'
(lp853
I1
aI1
aI3
aI5
aI21
aI47
aI51
aI59
aI55
aI411
aI931
aI251
asS'd'
I427
sa(dp854
S'a'
I1301
sS's'
I12
sS'm'
(lp855
I1
aI3
aI7
aI3
aI31
aI25
aI81
aI115
aI405
aI239
aI741
aI455
asS'd'
I428
sa(dp856
S'a'
I1305
sS's'
I12
sS'm'
(lp857
I1
aI1
aI5
aI1
aI31
aI3
aI101
aI83
aI479
aI491
aI1779
aI2225
asS'd'
I429
sa(dp858
S'a'
I1332
sS's'
I12
sS'm'
(lp859
I1
aI3
aI3
aI3
aI9
aI37
aI107
aI161
aI203
aI503
aI767
aI3435
asS'd'
I430
sa(dp860
S'a'
I1349
sS's'
I12
sS'm'
(lp861
I1
aI3
aI7
aI9
aI1
aI27
aI61
aI119
aI233
aI39
aI1375
aI4089
asS'd'
I431
sa(dp862
S'a'
I1384
sS's'
I12
sS'm'
(lp863
I1
aI1
aI5
aI9
aI1
aI31
aI45
aI51
aI369
aI587
aI383
aI2813
asS'd'
I432
sa(dp864
S'a'
I1392
sS's'
I12
sS'm'
(lp865
I1
aI3
aI7
aI5
aI31
aI7
aI49
aI119
aI487
aI591
aI1627
aI53
asS'd'
I433
sa(dp866
S'a'
I1402
sS's'
I12
sS'm'
(lp867
I1
aI1
aI7
aI1
aI9
aI47
aI1
aI223
aI369
aI711
aI1603
aI1917
asS'd'
I434
sa(dp868
S'a'
I1413
sS's'
I12
sS'm'
(lp869
I1
aI3
aI5
aI3
aI21
aI37
aI111
aI17
aI483
aI739
aI1193
aI2775
asS'd'
I435
sa(dp870
S'a'
I1417
sS's'
I12
sS'm'
(lp871
I1
aI3
aI3
aI7
aI17
aI11
aI51
aI117
aI455
aI191
aI1493
aI3821
asS'd'
I436
sa(dp872
S'a'
I1423
sS's'
I12
sS'm'
(lp873
I1
aI1
aI5
aI9
aI23
aI39
aI99
aI181
aI343
aI485
aI99
aI1931
asS'd'
I437
sa(dp874
S'a'
I1451
sS's'
I12
sS'm'
(lp875
I1
aI3
aI1
aI7
aI29
aI49
aI31
aI71
aI489
aI527
aI1763
aI2909
asS'd'
I438
sa(dp876
S'a'
I1480
sS's'
I12
sS'm'
(lp877
I1
aI1
aI5
aI11
aI5
aI5
aI73
aI189
aI321
aI57
aI1191
aI3685
asS'd'
I439
sa(dp878
S'a'
I1491
sS's'
I12
sS'm'
(lp879
I1
aI1
aI5
aI15
aI13
aI45
aI125
aI207
aI371
aI415
aI315
aI983
asS'd'
I440
sa(dp880
S'a'
I1503
sS's'
I12
sS'm'
(lp881
I1
aI3
aI3
aI5
aI25
aI59
aI33
aI31
aI239
aI919
aI1859
aI2709
asS'd'
I441
sa(dp882
S'a'
I1504
sS's'
I12
sS'm'
(lp883
I1
aI3
aI5
aI13
aI27
aI61
aI23
aI115
aI61
aI413
aI1275
aI3559
asS'd'
I442
sa(dp884
S'a'
I1513
sS's'
I12
sS'm'
(lp885
I1
aI3
aI7
aI15
aI5
aI59
aI101
aI81
aI47
aI967
aI809
aI3189
asS'd'
I443
sa(dp886
S'a'
I1538
sS's'
I12
sS'm'
(lp887
I1
aI1
aI5
aI11
aI31
aI15
aI39
aI25
aI173
aI505
aI809
aI2677
asS'd'
I444
sa(dp888
S'a'
I1544
sS's'
I12
sS'm'
(lp889
I1
aI1
aI5
aI9
aI19
aI13
aI95
aI89
aI511
aI127
aI1395
aI2935
asS'd'
I445
sa(dp890
S'a'
I1547
sS's'
I12
sS'm'
(lp891
I1
aI1
aI5
aI5
aI31
aI45
aI9
aI57
aI91
aI303
aI1295
aI3215
asS'd'
I446
sa(dp892
S'a'
I1555
sS's'
I12
sS'm'
(lp893
I1
aI3
aI3
aI3
aI19
aI15
aI113
aI187
aI217
aI489
aI1285
aI1803
asS'd'
I447
sa(dp894
S'a'
I1574
sS's'
I12
sS'm'
(lp895
I1
aI1
aI3
aI1
aI13
aI29
aI57
aI139
aI255
aI197
aI537
aI2183
asS'd'
I448
sa(dp896
S'a'
I1603
sS's'
I12
sS'm'
(lp897
I1
aI3
aI1
aI15
aI11
aI7
aI53
aI255
aI467
aI9
aI757
aI3167
asS'd'
I449
sa(dp898
S'a'
I1615
sS's'
I12
sS'm'
(lp899
I1
aI3
aI3
aI15
aI21
aI13
aI9
aI189
aI359
aI323
aI49
aI333
asS'd'
I450
sa(dp900
S'a'
I1618
sS's'
I12
sS'm'
(lp901
I1
aI3
aI7
aI11
aI7
aI37
aI21
aI119
aI401
aI157
aI1659
aI1069
asS'd'
I451
sa(dp902
S'a'
I1629
sS's'
I12
sS'm'
(lp903
I1
aI1
aI5
aI7
aI17
aI33
aI115
aI229
aI149
aI151
aI2027
aI279
asS'd'
I452
sa(dp904
S'a'
I1634
sS's'
I12
sS'm'
(lp905
I1
aI1
aI5
aI15
aI5
aI49
aI77
aI155
aI383
aI385
aI1985
aI945
asS'd'
I453
sa(dp906
S'a'
I1636
sS's'
I12
sS'm'
(lp907
I1
aI3
aI7
aI3
aI7
aI55
aI85
aI41
aI357
aI527
aI1715
aI1619
asS'd'
I454
sa(dp908
S'a'
I1639
sS's'
I12
sS'm'
(lp909
I1
aI1
aI3
aI1
aI21
aI45
aI115
aI21
aI199
aI967
aI1581
aI3807
asS'd'
I455
sa(dp910
S'a'
I1657
sS's'
I12
sS'm'
(lp911
I1
aI1
aI3
aI7
aI21
aI39
aI117
aI191
aI169
aI73
aI413
aI3417
asS'd'
I456
sa(dp912
S'a'
I1667
sS's'
I12
sS'm'
(lp913
I1
aI1
aI1
aI13
aI1
aI31
aI57
aI195
aI231
aI321
aI367
aI1027
asS'd'
I457
sa(dp914
S'a'
I1681
sS's'
I12
sS'm'
(lp915
I1
aI3
aI7
aI3
aI11
aI29
aI47
aI161
aI71
aI419
aI1721
aI437
asS'd'
I458
sa(dp916
S'a'
I1697
sS's'
I12
sS'm'
(lp917
I1
aI1
aI7
aI3
aI11
aI9
aI43
aI65
aI157
aI1
aI1851
aI823
asS'd'
I459
sa(dp918
S'a'
I1704
sS's'
I12
sS'm'
(lp919
I1
aI1
aI1
aI5
aI21
aI15
aI31
aI101
aI293
aI299
aI127
aI1321
asS'd'
I460
sa(dp920
S'a'
I1709
sS's'
I12
sS'm'
(lp921
I1
aI1
aI7
aI1
aI27
aI1
aI11
aI229
aI241
aI705
aI43
aI1475
asS'd'
I461
sa(dp922
S'a'
I1722
sS's'
I12
sS'm'
(lp923
I1
aI3
aI7
aI1
aI5
aI15
aI73
aI183
aI193
aI55
aI1345
aI49
asS'd'
I462
sa(dp924
S'a'
I1730
sS's'
I12
sS'm'
(lp925
I1
aI3
aI3
aI3
aI19
aI3
aI55
aI21
aI169
aI663
aI1675
aI137
asS'd'
I463
sa(dp926
S'a'
I1732
sS's'
I12
sS'm'
(lp927
I1
aI1
aI1
aI13
aI7
aI21
aI69
aI67
aI373
aI965
aI1273
aI2279
asS'd'
I464
sa(dp928
S'a'
I1802
sS's'
I12
sS'm'
(lp929
I1
aI1
aI7
aI7
aI21
aI23
aI17
aI43
aI341
aI845
aI465
aI3355
asS'd'
I465
sa(dp930
S'a'
I1804
sS's'
I12
sS'm'
(lp931
I1
aI3
aI5
aI5
aI25
aI5
aI81
aI101
aI233
aI139
aI359
aI2057
asS'd'
I466
sa(dp932
S'a'
I1815
sS's'
I12
sS'm'
(lp933
I1
aI1
aI3
aI11
aI15
aI39
aI55
aI3
aI471
aI765
aI1143
aI3941
asS'd'
I467
sa(dp934
S'a'
I1826
sS's'
I12
sS'm'
(lp935
I1
aI1
aI7
aI15
aI9
aI57
aI81
aI79
aI215
aI433
aI333
aI3855
asS'd'
I468
sa(dp936
S'a'
I1832
sS's'
I12
sS'm'
(lp937
I1
aI1
aI5
aI5
aI19
aI45
aI83
aI31
aI209
aI363
aI701
aI1303
asS'd'
I469
sa(dp938
S'a'
I1843
sS's'
I12
sS'm'
(lp939
I1
aI3
aI7
aI5
aI1
aI13
aI55
aI163
aI435
aI807
aI287
aI2031
asS'd'
I470
sa(dp940
S'a'
I1849
sS's'
I12
sS'm'
(lp941
I1
aI3
aI3
aI7
aI3
aI3
aI17
aI197
aI39
aI169
aI489
aI1769
asS'd'
I471
sa(dp942
S'a'
I1863
sS's'
I12
sS'm'
(lp943
I1
aI1
aI3
aI5
aI29
aI43
aI87
aI161
aI289
aI339
aI1233
aI2353
asS'd'
I472
sa(dp944
S'a'
I1905
sS's'
I12
sS'm'
(lp945
I1
aI3
aI3
aI9
aI21
aI9
aI77
aI1
aI453
aI167
aI1643
aI2227
asS'd'
I473
sa(dp946
S'a'
I1928
sS's'
I12
sS'm'
(lp947
I1
aI1
aI7
aI1
aI15
aI7
aI67
aI33
aI193
aI241
aI1031
aI2339
asS'd'
I474
sa(dp948
S'a'
I1933
sS's'
I12
sS'm'
(lp949
I1
aI3
aI1
aI11
aI1
aI63
aI45
aI65
aI265
aI661
aI849
aI1979
asS'd'
I475
sa(dp950
S'a'
I1939
sS's'
I12
sS'm'
(lp951
I1
aI3
aI1
aI13
aI19
aI49
aI3
aI11
aI159
aI213
aI659
aI2839
asS'd'
I476
sa(dp952
S'a'
I1976
sS's'
I12
sS'm'
(lp953
I1
aI3
aI5
aI11
aI9
aI29
aI27
aI227
aI253
aI449
aI1403
aI3427
asS'd'
I477
sa(dp954
S'a'
I1996
sS's'
I12
sS'm'
(lp955
I1
aI1
aI3
aI1
aI7
aI3
aI77
aI143
aI277
aI779
aI1499
aI475
asS'd'
I478
sa(dp956
S'a'
I2013
sS's'
I12
sS'm'
(lp957
I1
aI1
aI1
aI5
aI11
aI23
aI87
aI131
aI393
aI849
aI193
aI3189
asS'd'
I479
sa(dp958
S'a'
I2014
sS's'
I12
sS'm'
(lp959
I1
aI3
aI5
aI11
aI3
aI3
aI89
aI9
aI449
aI243
aI1501
aI1739
asS'd'
I480
sa(dp960
S'a'
I2020
sS's'
I12
sS'm'
(lp961
I1
aI3
aI1
aI9
aI29
aI29
aI113
aI15
aI65
aI611
aI135
aI3687
asS'd'
I481
sa(dp962
S'a'
I13
sS's'
I13
sS'm'
(lp963
I1
aI1
aI1
aI9
aI21
aI19
aI39
aI151
aI395
aI501
aI1339
aI959
aI2725
asS'd'
I482
sa(dp964
S'a'
I19
sS's'
I13
sS'm'
(lp965
I1
aI3
aI7
aI1
aI7
aI35
aI45
aI33
aI119
aI225
aI1631
aI1695
aI1459
asS'd'
I483
sa(dp966
S'a'
I26
sS's'
I13
sS'm'
(lp967
I1
aI1
aI1
aI3
aI25
aI55
aI37
aI79
aI167
aI907
aI1075
aI271
aI4059
asS'd'
I484
sa(dp968
S'a'
I41
sS's'
I13
sS'm'
(lp969
I1
aI3
aI5
aI13
aI5
aI13
aI53
aI165
aI437
aI67
aI1705
aI3177
aI8095
asS'd'
I485
sa(dp970
S'a'
I50
sS's'
I13
sS'm'
(lp971
I1
aI3
aI3
aI13
aI27
aI57
aI95
aI55
aI443
aI245
aI1945
aI1725
aI1929
asS'd'
I486
sa(dp972
S'a'
I55
sS's'
I13
sS'm'
(lp973
I1
aI3
aI1
aI9
aI5
aI33
aI109
aI35
aI99
aI827
aI341
aI2401
aI2411
asS'd'
I487
sa(dp974
S'a'
I69
sS's'
I13
sS'm'
(lp975
I1
aI1
aI5
aI9
aI7
aI33
aI43
aI39
aI87
aI799
aI635
aI3481
aI7159
asS'd'
I488
sa(dp976
S'a'
I70
sS's'
I13
sS'm'
(lp977
I1
aI3
aI1
aI1
aI31
aI15
aI45
aI27
aI337
aI113
aI987
aI2065
aI2529
asS'd'
I489
sa(dp978
S'a'
I79
sS's'
I13
sS'm'
(lp979
I1
aI1
aI5
aI9
aI5
aI15
aI105
aI123
aI479
aI289
aI1609
aI2177
aI4629
asS'd'
I490
sa(dp980
S'a'
I82
sS's'
I13
sS'm'
(lp981
I1
aI3
aI5
aI11
aI31
aI47
aI97
aI87
aI385
aI195
aI1041
aI651
aI3271
asS'd'
I491
sa(dp982
S'a'
I87
sS's'
I13
sS'm'
(lp983
I1
aI1
aI3
aI7
aI17
aI3
aI101
aI55
aI87
aI629
aI1687
aI1387
aI2745
asS'd'
I492
sa(dp984
S'a'
I93
sS's'
I13
sS'm'
(lp985
I1
aI3
aI5
aI5
aI7
aI21
aI9
aI237
aI313
aI549
aI1107
aI117
aI6183
asS'd'
I493
sa(dp986
S'a'
I94
sS's'
I13
sS'm'
(lp987
I1
aI1
aI3
aI9
aI9
aI5
aI55
aI201
aI487
aI851
aI1103
aI2993
aI4055
asS'd'
I494
sa(dp988
S'a'
I97
sS's'
I13
sS'm'
(lp989
I1
aI1
aI5
aI9
aI31
aI19
aI59
aI7
aI363
aI381
aI1167
aI2057
aI5715
asS'd'
I495
sa(dp990
S'a'
I100
sS's'
I13
sS'm'
(lp991
I1
aI3
aI3
aI15
aI23
aI63
aI19
aI227
aI387
aI827
aI487
aI1049
aI7471
asS'd'
I496
sa(dp992
S'a'
I112
sS's'
I13
sS'm'
(lp993
I1
aI3
aI1
aI5
aI23
aI25
aI61
aI245
aI363
aI863
aI963
aI3583
aI6475
asS'd'
I497
sa(dp994
S'a'
I121
sS's'
I13
sS'm'
(lp995
I1
aI1
aI5
aI1
aI5
aI27
aI81
aI85
aI275
aI49
aI235
aI3291
aI1195
asS'd'
I498
sa(dp996
S'a'
I134
sS's'
I13
sS'm'
(lp997
I1
aI1
aI5
aI7
aI23
aI53
aI85
aI107
aI511
aI779
aI1265
aI1093
aI7859
asS'd'
I499
sa(dp998
S'a'
I138
sS's'
I13
sS'm'
(lp999
I1
aI3
aI3
aI1
aI9
aI21
aI75
aI219
aI59
aI485
aI1739
aI3845
aI1109
asS'd'
I500
sa(dp1000
S'a'
I148
sS's'
I13
sS'm'
(lp1001
I1
aI3
aI5
aI1
aI13
aI41
aI19
aI143
aI293
aI391
aI2023
aI1791
aI4399
asS'd'
I501
sa(dp1002
S'a'
I151
sS's'
I13
sS'm'
(lp1003
I1
aI3
aI7
aI15
aI21
aI13
aI21
aI195
aI215
aI413
aI523
aI2099
aI2341
asS'd'
I502
sa(dp1004
S'a'
I157
sS's'
I13
sS'm'
(lp1005
I1
aI1
aI1
aI3
aI29
aI51
aI47
aI57
aI135
aI575
aI943
aI1673
aI541
asS'd'
I503
sa(dp1006
S'a'
I161
sS's'
I13
sS'm'
(lp1007
I1
aI3
aI5
aI1
aI9
aI13
aI113
aI175
aI447
aI115
aI657
aI4077
aI5973
asS'd'
I504
sa(dp1008
S'a'
I179
sS's'
I13
sS'm'
(lp1009
I1
aI1
aI1
aI11
aI17
aI41
aI37
aI95
aI297
aI579
aI911
aI2207
aI2387
asS'd'
I505
sa(dp1010
S'a'
I181
sS's'
I13
sS'm'
(lp1011
I1
aI3
aI5
aI3
aI23
aI11
aI23
aI231
aI93
aI667
aI711
aI1563
aI7961
asS'd'
I506
sa(dp1012
S'a'
I188
sS's'
I13
sS'm'
(lp1013
I1
aI1
aI7
aI3
aI17
aI59
aI13
aI181
aI141
aI991
aI1817
aI457
aI1711
asS'd'
I507
sa(dp1014
S'a'
I196
sS's'
I13
sS'm'
(lp1015
I1
aI3
aI3
aI5
aI31
aI59
aI81
aI205
aI245
aI537
aI1049
aI997
aI1815
asS'd'
I508
sa(dp1016
S'a'
I203
sS's'
I13
sS'm'
(lp1017
I1
aI3
aI7
aI5
aI17
aI13
aI9
aI79
aI17
aI185
aI5
aI2211
aI6263
asS'd'
I509
sa(dp1018
S'a'
I206
sS's'
I13
sS'm'
(lp1019
I1
aI3
aI7
aI13
aI7
aI53
aI61
aI145
aI13
aI285
aI1203
aI947
aI2933
asS'd'
I510
sa(dp1020
S'a'
I223
sS's'
I13
sS'm'
(lp1021
I1
aI1
aI7
aI3
aI31
aI19
aI69
aI217
aI47
aI441
aI1893
aI673
aI4451
asS'd'
I511
sa(dp1022
S'a'
I224
sS's'
I13
sS'm'
(lp1023
I1
aI1
aI1
aI1
aI25
aI9
aI23
aI225
aI385
aI629
aI603
aI3747
aI4241
asS'd'
I512
sa(dp1024
S'a'
I227
sS's'
I13
sS'm'
(lp1025
I1
aI3
aI1
aI9
aI5
aI37
aI31
aI237
aI431
aI79
aI1521
aI459
aI2523
asS'd'
I513
sa(dp1026
S'a'
I230
sS's'
I13
sS'm'
(lp1027
I1
aI3
aI7
aI3
aI9
aI43
aI105
aI179
aI5
aI225
aI799
aI1777
aI4893
asS'd'
I514
sa(dp1028
S'a'
I239
sS's'
I13
sS'm'
(lp1029
I1
aI1
aI3
aI1
aI29
aI45
aI29
aI159
aI267
aI247
aI455
aI847
aI3909
asS'd'
I515
sa(dp1030
S'a'
I241
sS's'
I13
sS'm'
(lp1031
I1
aI1
aI3
aI7
aI25
aI21
aI121
aI57
aI467
aI275
aI719
aI1521
aI7319
asS'd'
I516
sa(dp1032
S'a'
I248
sS's'
I13
sS'm'
(lp1033
I1
aI3
aI1
aI3
aI11
aI35
aI119
aI123
aI81
aI979
aI1187
aI3623
aI4293
asS'd'
I517
sa(dp1034
S'a'
I253
sS's'
I13
sS'm'
(lp1035
I1
aI1
aI1
aI7
aI15
aI25
aI121
aI235
aI25
aI487
aI873
aI1787
aI1977
asS'd'
I518
sa(dp1036
S'a'
I268
sS's'
I13
sS'm'
(lp1037
I1
aI1
aI1
aI11
aI3
aI7
aI17
aI135
aI345
aI353
aI383
aI4011
aI2573
asS'd'
I519
sa(dp1038
S'a'
I274
sS's'
I13
sS'm'
(lp1039
I1
aI3
aI7
aI15
aI27
aI13
aI97
aI123
aI65
aI675
aI951
aI1285
aI6559
asS'd'
I520
sa(dp1040
S'a'
I283
sS's'
I13
sS'm'
(lp1041
I1
aI3
aI7
aI3
aI7
aI1
aI71
aI19
aI325
aI765
aI337
aI1197
aI2697
asS'd'
I521
sa(dp1042
S'a'
I286
sS's'
I13
sS'm'
(lp1043
I1
aI3
aI5
aI1
aI31
aI37
aI11
aI71
aI169
aI283
aI83
aI3801
aI7083
asS'd'
I522
sa(dp1044
S'a'
I289
sS's'
I13
sS'm'
(lp1045
I1
aI1
aI3
aI15
aI17
aI29
aI83
aI65
aI275
aI679
aI1749
aI4007
aI7749
asS'd'
I523
sa(dp1046
S'a'
I301
sS's'
I13
sS'm'
(lp1047
I1
aI1
aI3
aI1
aI21
aI11
aI41
aI95
aI237
aI361
aI1819
aI2783
aI2383
asS'd'
I524
sa(dp1048
S'a'
I302
sS's'
I13
sS'm'
(lp1049
I1
aI3
aI7
aI11
aI29
aI57
aI111
aI187
aI465
aI145
aI605
aI1987
aI8109
asS'd'
I525
sa(dp1050
S'a'
I316
sS's'
I13
sS'm'
(lp1051
I1
aI1
aI3
aI3
aI19
aI15
aI55
aI83
aI357
aI1001
aI643
aI1517
aI6529
asS'd'
I526
sa(dp1052
S'a'
I319
sS's'
I13
sS'm'
(lp1053
I1
aI3
aI1
aI5
aI29
aI35
aI73
aI23
aI77
aI619
aI1523
aI1725
aI8145
asS'd'
I527
sa(dp1054
S'a'
I324
sS's'
I13
sS'm'
(lp1055
I1
aI1
aI5
aI5
aI19
aI23
aI7
aI197
aI449
aI337
aI717
aI2921
aI315
asS'd'
I528
sa(dp1056
S'a'
I331
sS's'
I13
sS'm'
(lp1057
I1
aI3
aI5
aI9
aI7
aI63
aI117
aI97
aI97
aI813
aI1925
aI2817
aI1579
asS'd'
I529
sa(dp1058
S'a'
I333
sS's'
I13
sS'm'
(lp1059
I1
aI1
aI1
aI11
aI31
aI7
aI25
aI235
aI231
aI133
aI1007
aI1371
aI1553
asS'd'
I530
sa(dp1060
S'a'
I345
sS's'
I13
sS'm'
(lp1061
I1
aI1
aI7
aI5
aI19
aI7
aI47
aI171
aI267
aI243
aI1331
aI567
aI6033
asS'd'
I531
sa(dp1062
S'a'
I351
sS's'
I13
sS'm'
(lp1063
I1
aI1
aI5
aI1
aI7
aI49
aI55
aI89
aI109
aI735
aI1455
aI3193
aI6239
asS'd'
I532
sa(dp1064
S'a'
I358
sS's'
I13
sS'm'
(lp1065
I1
aI1
aI1
aI7
aI1
aI61
aI9
aI103
aI3
aI929
aI1481
aI2927
aI2957
asS'd'
I533
sa(dp1066
S'a'
I375
sS's'
I13
sS'm'
(lp1067
I1
aI1
aI5
aI13
aI17
aI21
aI75
aI49
aI255
aI1019
aI1161
aI2133
aI1177
asS'd'
I534
sa(dp1068
S'a'
I379
sS's'
I13
sS'm'
(lp1069
I1
aI3
aI1
aI3
aI13
aI15
aI41
aI247
aI211
aI409
aI1163
aI523
aI2635
asS'd'
I535
sa(dp1070
S'a'
I381
sS's'
I13
sS'm'
(lp1071
I1
aI3
aI7
aI7
aI21
aI59
aI91
aI149
aI479
aI391
aI681
aI2311
aI6249
asS'd'
I536
sa(dp1072
S'a'
I386
sS's'
I13
sS'm'
(lp1073
I1
aI1
aI5
aI11
aI27
aI53
aI21
aI211
aI197
aI815
aI719
aI1605
aI255
asS'd'
I537
sa(dp1074
S'a'
I403
sS's'
I13
sS'm'
(lp1075
I1
aI1
aI3
aI3
aI9
aI33
aI59
aI3
aI323
aI1
aI101
aI1135
aI8105
asS'd'
I538
sa(dp1076
S'a'
I405
sS's'
I13
sS'm'
(lp1077
I1
aI3
aI3
aI1
aI29
aI5
aI17
aI141
aI51
aI991
aI841
aI327
aI3859
asS'd'
I539
sa(dp1078
S'a'
I419
sS's'
I13
sS'm'
(lp1079
I1
aI3
aI1
aI5
aI11
aI19
aI23
aI89
aI175
aI173
aI165
aI2881
aI1881
asS'd'
I540
sa(dp1080
S'a'
I426
sS's'
I13
sS'm'
(lp1081
I1
aI1
aI1
aI15
aI13
aI51
aI87
aI39
aI495
aI611
aI1341
aI1531
aI7029
asS'd'
I541
sa(dp1082
S'a'
I428
sS's'
I13
sS'm'
(lp1083
I1
aI1
aI3
aI11
aI13
aI55
aI75
aI185
aI57
aI61
aI1917
aI2051
aI5965
asS'd'
I542
sa(dp1084
S'a'
I439
sS's'
I13
sS'm'
(lp1085
I1
aI1
aI5
aI5
aI7
aI53
aI11
aI217
aI213
aI933
aI921
aI3607
aI5175
asS'd'
I543
sa(dp1086
S'a'
I440
sS's'
I13
sS'm'
(lp1087
I1
aI3
aI3
aI5
aI17
aI53
aI103
aI251
aI369
aI781
aI1319
aI3717
aI4439
asS'd'
I544
sa(dp1088
S'a'
I446
sS's'
I13
sS'm'
(lp1089
I1
aI3
aI5
aI13
aI1
aI39
aI25
aI235
aI321
aI773
aI251
aI3111
aI6397
asS'd'
I545
sa(dp1090
S'a'
I451
sS's'
I13
sS'm'
(lp1091
I1
aI1
aI7
aI3
aI31
aI5
aI25
aI29
aI325
aI385
aI1313
aI127
aI4705
asS'd'
I546
sa(dp1092
S'a'
I454
sS's'
I13
sS'm'
(lp1093
I1
aI1
aI5
aI15
aI15
aI27
aI15
aI85
aI239
aI243
aI1633
aI3473
aI2621
asS'd'
I547
sa(dp1094
S'a'
I458
sS's'
I13
sS'm'
(lp1095
I1
aI3
aI3
aI3
aI9
aI19
aI113
aI13
aI137
aI165
aI25
aI2957
aI7549
asS'd'
I548
sa(dp1096
S'a'
I465
sS's'
I13
sS'm'
(lp1097
I1
aI3
aI1
aI3
aI11
aI21
aI3
aI97
aI417
aI183
aI1205
aI1437
aI247
asS'd'
I549
sa(dp1098
S'a'
I468
sS's'
I13
sS'm'
(lp1099
I1
aI1
aI7
aI3
aI17
aI21
aI125
aI55
aI67
aI387
aI385
aI2323
aI887
asS'd'
I550
sa(dp1100
S'a'
I472
sS's'
I13
sS'm'
(lp1101
I1
aI3
aI5
aI5
aI29
aI11
aI103
aI223
aI233
aI641
aI133
aI415
aI1297
asS'd'
I551
sa(dp1102
S'a'
I475
sS's'
I13
sS'm'
(lp1103
I1
aI3
aI3
aI11
aI1
aI9
aI5
aI189
aI235
aI1007
aI1363
aI3985
aI889
asS'd'
I552
sa(dp1104
S'a'
I477
sS's'
I13
sS'm'
(lp1105
I1
aI3
aI7
aI9
aI23
aI19
aI19
aI183
aI269
aI403
aI1643
aI3559
aI5189
asS'd'
I553
sa(dp1106
S'a'
I496
sS's'
I13
sS'm'
(lp1107
I1
aI3
aI7
aI3
aI29
aI45
aI17
aI69
aI475
aI149
aI1291
aI2689
aI7625
asS'd'
I554
sa(dp1108
S'a'
I502
sS's'
I13
sS'm'
(lp1109
I1
aI3
aI7
aI3
aI27
aI37
aI41
aI73
aI253
aI1001
aI431
aI1111
aI7887
asS'd'
I555
sa(dp1110
S'a'
I508
sS's'
I13
sS'm'
(lp1111
I1
aI1
aI7
aI5
aI3
aI7
aI87
aI143
aI289
aI495
aI631
aI3011
aI6151
asS'd'
I556
sa(dp1112
S'a'
I517
sS's'
I13
sS'm'
(lp1113
I1
aI1
aI1
aI13
aI5
aI45
aI17
aI167
aI23
aI975
aI801
aI1975
aI6833
asS'd'
I557
sa(dp1114
S'a'
I521
sS's'
I13
sS'm'
(lp1115
I1
aI3
aI1
aI11
aI7
aI21
aI39
aI23
aI213
aI429
aI1301
aI2059
aI197
asS'd'
I558
sa(dp1116
S'a'
I527
sS's'
I13
sS'm'
(lp1117
I1
aI3
aI3
aI15
aI3
aI57
aI121
aI133
aI29
aI711
aI1961
aI2497
aI189
asS'd'
I559
sa(dp1118
S'a'
I530
sS's'
I13
sS'm'
(lp1119
I1
aI1
aI3
aI5
aI11
aI55
aI115
aI137
aI233
aI673
aI985
aI2849
aI5911
asS'd'
I560
sa(dp1120
S'a'
I532
sS's'
I13
sS'm'
(lp1121
I1
aI1
aI7
aI15
aI29
aI45
aI1
aI241
aI329
aI323
aI925
aI2821
aI3331
asS'd'
I561
sa(dp1122
S'a'
I542
sS's'
I13
sS'm'
(lp1123
I1
aI1
aI5
aI7
aI13
aI31
aI81
aI105
aI199
aI145
aI195
aI1365
aI5119
asS'd'
I562
sa(dp1124
S'a'
I552
sS's'
I13
sS'm'
(lp1125
I1
aI3
aI7
aI11
aI3
aI55
aI11
aI31
aI117
aI343
aI1265
aI1837
aI2451
asS'd'
I563
sa(dp1126
S'a'
I555
sS's'
I13
sS'm'
(lp1127
I1
aI1
aI3
aI7
aI29
aI57
aI61
aI179
aI429
aI591
aI177
aI1945
aI2159
asS'd'
I564
sa(dp1128
S'a'
I560
sS's'
I13
sS'm'
(lp1129
I1
aI3
aI5
aI11
aI23
aI49
aI101
aI137
aI339
aI323
aI1035
aI1749
aI7737
asS'd'
I565
sa(dp1130
S'a'
I566
sS's'
I13
sS'm'
(lp1131
I1
aI3
aI1
aI13
aI21
aI35
aI55
aI79
aI19
aI269
aI1055
aI2651
aI7083
asS'd'
I566
sa(dp1132
S'a'
I575
sS's'
I13
sS'm'
(lp1133
I1
aI3
aI3
aI11
aI9
aI9
aI95
aI167
aI437
aI361
aI1185
aI4083
aI603
asS'd'
I567
sa(dp1134
S'a'
I577
sS's'
I13
sS'm'
(lp1135
I1
aI1
aI1
aI7
aI31
aI61
aI77
aI65
aI489
aI657
aI691
aI2423
aI4147
asS'd'
I568
sa(dp1136
S'a'
I589
sS's'
I13
sS'm'
(lp1137
I1
aI3
aI5
aI7
aI21
aI37
aI87
aI191
aI311
aI453
aI2013
aI829
aI2619
asS'd'
I569
sa(dp1138
S'a'
I590
sS's'
I13
sS'm'
(lp1139
I1
aI1
aI5
aI9
aI17
aI47
aI35
aI101
aI5
aI813
aI1157
aI1279
aI7365
asS'd'
I570
sa(dp1140
S'a'
I602
sS's'
I13
sS'm'
(lp1141
I1
aI1
aI5
aI3
aI11
aI35
aI113
aI199
aI369
aI721
aI901
aI1471
aI7801
asS'd'
I571
sa(dp1142
S'a'
I607
sS's'
I13
sS'm'
(lp1143
I1
aI3
aI1
aI5
aI9
aI61
aI83
aI157
aI391
aI739
aI1957
aI2123
aI4341
asS'd'
I572
sa(dp1144
S'a'
I608
sS's'
I13
sS'm'
(lp1145
I1
aI3
aI5
aI11
aI19
aI19
aI111
aI225
aI383
aI219
aI997
aI717
aI7505
asS'd'
I573
sa(dp1146
S'a'
I611
sS's'
I13
sS'm'
(lp1147
I1
aI3
aI1
aI11
aI13
aI63
aI35
aI127
aI209
aI831
aI501
aI3017
aI3507
asS'd'
I574
sa(dp1148
S'a'
I613
sS's'
I13
sS'm'
(lp1149
I1
aI3
aI7
aI9
aI29
aI7
aI11
aI163
aI81
aI563
aI1445
aI3215
aI6377
asS'd'
I575
sa(dp1150
S'a'
I625
sS's'
I13
sS'm'
(lp1151
I1
aI3
aI7
aI11
aI25
aI3
aI39
aI195
aI491
aI45
aI839
aI4021
aI4899
asS'd'
I576
sa(dp1152
S'a'
I644
sS's'
I13
sS'm'
(lp1153
I1
aI3
aI7
aI15
aI13
aI5
aI67
aI143
aI117
aI505
aI1281
aI3679
aI5695
asS'd'
I577
sa(dp1154
S'a'
I651
sS's'
I13
sS'm'
(lp1155
I1
aI3
aI7
aI9
aI9
aI19
aI21
aI221
aI147
aI763
aI683
aI2211
aI589
asS'd'
I578
sa(dp1156
S'a'
I654
sS's'
I13
sS'm'
(lp1157
I1
aI1
aI3
aI5
aI21
aI47
aI53
aI109
aI299
aI807
aI1153
aI1209
aI7961
asS'd'
I579
sa(dp1158
S'a'
I656
sS's'
I13
sS'm'
(lp1159
I1
aI3
aI7
aI11
aI9
aI31
aI45
aI43
aI505
aI647
aI1127
aI2681
aI4917
asS'd'
I580
sa(dp1160
S'a'
I662
sS's'
I13
sS'm'
(lp1161
I1
aI1
aI5
aI15
aI31
aI41
aI63
aI113
aI399
aI727
aI673
aI2587
aI5259
asS'd'
I581
sa(dp1162
S'a'
I668
sS's'
I13
sS'm'
(lp1163
I1
aI1
aI1
aI13
aI17
aI53
aI35
aI99
aI57
aI243
aI1447
aI1919
aI2831
asS'd'
I582
sa(dp1164
S'a'
I681
sS's'
I13
sS'm'
(lp1165
I1
aI3
aI7
aI11
aI23
aI51
aI13
aI9
aI49
aI449
aI997
aI3073
aI4407
asS'd'
I583
sa(dp1166
S'a'
I682
sS's'
I13
sS'm'
(lp1167
I1
aI3
aI5
aI7
aI23
aI33
aI89
aI41
aI415
aI53
aI697
aI1113
aI1489
asS'd'
I584
sa(dp1168
S'a'
I689
sS's'
I13
sS'm'
(lp1169
I1
aI1
aI3
aI7
aI1
aI13
aI29
aI13
aI255
aI749
aI77
aI3463
aI1761
asS'd'
I585
sa(dp1170
S'a'
I696
sS's'
I13
sS'm'
(lp1171
I1
aI3
aI3
aI7
aI13
aI15
aI93
aI191
aI309
aI869
aI739
aI1041
aI3053
asS'd'
I586
sa(dp1172
S'a'
I699
sS's'
I13
sS'm'
(lp1173
I1
aI3
aI5
aI13
aI5
aI19
aI109
aI211
aI347
aI839
aI893
aI2947
aI7735
asS'd'
I587
sa(dp1174
S'a'
I707
sS's'
I13
sS'm'
(lp1175
I1
aI3
aI1
aI13
aI27
aI3
aI119
aI157
aI485
aI99
aI1703
aI3895
aI573
asS'd'
I588
sa(dp1176
S'a'
I709
sS's'
I13
sS'm'
(lp1177
I1
aI3
aI7
aI11
aI1
aI23
aI123
aI105
aI31
aI359
aI275
aI1775
aI3685
asS'd'
I589
sa(dp1178
S'a'
I714
sS's'
I13
sS'm'
(lp1179
I1
aI3
aI3
aI5
aI27
aI11
aI125
aI3
aI413
aI199
aI2043
aI2895
aI2945
asS'd'
I590
sa(dp1180
S'a'
I716
sS's'
I13
sS'm'
(lp1181
I1
aI3
aI3
aI3
aI15
aI49
aI121
aI159
aI233
aI543
aI193
aI4007
aI321
asS'd'
I591
sa(dp1182
S'a'
I719
sS's'
I13
sS'm'
(lp1183
I1
aI1
aI3
aI5
aI9
aI47
aI87
aI1
aI51
aI1011
aI1595
aI2239
aI6467
asS'd'
I592
sa(dp1184
S'a'
I727
sS's'
I13
sS'm'
(lp1185
I1
aI3
aI7
aI9
aI1
aI33
aI87
aI137
aI469
aI749
aI1413
aI805
aI6817
asS'd'
I593
sa(dp1186
S'a'
I734
sS's'
I13
sS'm'
(lp1187
I1
aI3
aI1
aI13
aI19
aI45
aI95
aI227
aI29
aI677
aI1275
aI3395
aI4451
asS'd'
I594
sa(dp1188
S'a'
I738
sS's'
I13
sS'm'
(lp1189
I1
aI1
aI7
aI5
aI7
aI63
aI33
aI71
aI443
aI561
aI1311
aI3069
aI6943
asS'd'
I595
sa(dp1190
S'a'
I743
sS's'
I13
sS'm'
(lp1191
I1
aI1
aI1
aI13
aI9
aI37
aI23
aI69
aI13
aI415
aI1479
aI1197
aI861
asS'd'
I596
sa(dp1192
S'a'
I747
sS's'
I13
sS'm'
(lp1193
I1
aI3
aI3
aI13
aI27
aI21
aI13
aI233
aI105
aI777
aI345
aI2443
aI1105
asS'd'
I597
sa(dp1194
S'a'
I757
sS's'
I13
sS'm'
(lp1195
I1
aI1
aI7
aI11
aI23
aI13
aI21
aI147
aI221
aI549
aI73
aI2729
aI6279
asS'd'
I598
sa(dp1196
S'a'
I769
sS's'
I13
sS'm'
(lp1197
I1
aI1
aI7
aI7
aI25
aI27
aI15
aI45
aI227
aI39
aI75
aI1191
aI3563
asS'd'
I599
sa(dp1198
S'a'
I770
sS's'
I13
sS'm'
(lp1199
I1
aI1
aI5
aI7
aI13
aI49
aI99
aI167
aI227
aI13
aI353
aI1047
aI8075
asS'd'
I600
sa(dp1200
S'a'
I776
sS's'
I13
sS'm'
(lp1201
I1
aI1
aI3
aI13
aI31
aI9
aI27
aI7
aI461
aI737
aI1559
aI3243
aI53
asS'd'
I601
sa(dp1202
S'a'
I790
sS's'
I13
sS'm'
(lp1203
I1
aI3
aI1
aI1
aI21
aI41
aI97
aI165
aI171
aI821
aI587
aI2137
aI2293
asS'd'
I602
sa(dp1204
S'a'
I799
sS's'
I13
sS'm'
(lp1205
I1
aI3
aI1
aI11
aI17
aI41
aI29
aI187
aI87
aI599
aI1467
aI1395
aI5931
asS'd'
I603
sa(dp1206
S'a'
I805
sS's'
I13
sS'm'
(lp1207
I1
aI1
aI1
aI9
aI9
aI49
aI89
aI205
aI409
aI453
aI61
aI1923
aI1257
asS'd'
I604
sa(dp1208
S'a'
I809
sS's'
I13
sS'm'
(lp1209
I1
aI3
aI7
aI3
aI9
aI43
aI89
aI143
aI431
aI83
aI1243
aI1795
aI3599
asS'd'
I605
sa(dp1210
S'a'
I812
sS's'
I13
sS'm'
(lp1211
I1
aI3
aI5
aI13
aI3
aI25
aI59
aI219
aI43
aI223
aI797
aI2651
aI6015
asS'd'
I606
sa(dp1212
S'a'
I820
sS's'
I13
sS'm'
(lp1213
I1
aI1
aI5
aI15
aI7
aI55
aI65
aI207
aI213
aI311
aI1287
aI1269
aI6467
asS'd'
I607
sa(dp1214
S'a'
I827
sS's'
I13
sS'm'
(lp1215
I1
aI3
aI7
aI11
aI21
aI57
aI31
aI183
aI351
aI857
aI911
aI1683
aI7155
asS'd'
I608
sa(dp1216
S'a'
I829
sS's'
I13
sS'm'
(lp1217
I1
aI3
aI5
aI11
aI27
aI1
aI21
aI47
aI387
aI383
aI1593
aI115
aI3805
asS'd'
I609
sa(dp1218
S'a'
I835
sS's'
I13
sS'm'
(lp1219
I1
aI3
aI1
aI1
aI13
aI23
aI87
aI173
aI181
aI619
aI1653
aI3931
aI6073
asS'd'
I610
sa(dp1220
S'a'
I841
sS's'
I13
sS'm'
(lp1221
I1
aI1
aI7
aI5
aI17
aI43
aI37
aI61
aI307
aI621
aI1785
aI55
aI115
asS'd'
I611
sa(dp1222
S'a'
I844
sS's'
I13
sS'm'
(lp1223
I1
aI3
aI7
aI15
aI25
aI61
aI123
aI15
aI237
aI671
aI1473
aI467
aI1907
asS'd'
I612
sa(dp1224
S'a'
I856
sS's'
I13
sS'm'
(lp1225
I1
aI1
aI7
aI5
aI29
aI57
aI75
aI237
aI85
aI699
aI159
aI3577
aI4771
asS'd'
I613
sa(dp1226
S'a'
I859
sS's'
I13
sS'm'
(lp1227
I1
aI1
aI1
aI11
aI25
aI19
aI51
aI1
aI147
aI31
aI895
aI2617
aI625
asS'd'
I614
sa(dp1228
S'a'
I862
sS's'
I13
sS'm'
(lp1229
I1
aI3
aI7
aI5
aI29
aI15
aI115
aI175
aI395
aI391
aI1141
aI1827
aI1181
asS'd'
I615
sa(dp1230
S'a'
I865
sS's'
I13
sS'm'
(lp1231
I1
aI3
aI5
aI7
aI17
aI7
aI11
aI193
aI89
aI243
aI561
aI3787
aI4551
asS'd'
I616
sa(dp1232
S'a'
I885
sS's'
I13
sS'm'
(lp1233
I1
aI3
aI1
aI11
aI7
aI57
aI7
aI125
aI403
aI947
aI1261
aI409
aI8083
asS'd'
I617
sa(dp1234
S'a'
I890
sS's'
I13
sS'm'
(lp1235
I1
aI1
aI5
aI13
aI21
aI63
aI115
aI233
aI231
aI921
aI1747
aI3635
aI2519
asS'd'
I618
sa(dp1236
S'a'
I905
sS's'
I13
sS'm'
(lp1237
I1
aI1
aI5
aI11
aI3
aI27
aI15
aI91
aI505
aI591
aI1451
aI3881
aI2997
asS'd'
I619
sa(dp1238
S'a'
I916
sS's'
I13
sS'm'
(lp1239
I1
aI1
aI3
aI11
aI21
aI9
aI109
aI153
aI317
aI533
aI593
aI3967
aI2797
asS'd'
I620
sa(dp1240
S'a'
I925
sS's'
I13
sS'm'
(lp1241
I1
aI3
aI3
aI13
aI9
aI57
aI121
aI245
aI219
aI867
aI967
aI791
aI7095
asS'd'
I621
sa(dp1242
S'a'
I935
sS's'
I13
sS'm'
(lp1243
I1
aI1
aI1
aI9
aI29
aI21
aI99
aI35
aI375
aI959
aI329
aI4087
aI7171
asS'd'
I622
sa(dp1244
S'a'
I939
sS's'
I13
sS'm'
(lp1245
I1
aI1
aI1
aI9
aI11
aI17
aI17
aI97
aI89
aI135
aI631
aI3809
aI3253
asS'd'
I623
sa(dp1246
S'a'
I942
sS's'
I13
sS'm'
(lp1247
I1
aI1
aI1
aI15
aI21
aI51
aI91
aI249
aI459
aI801
aI757
aI2353
aI2033
asS'd'
I624
sa(dp1248
S'a'
I949
sS's'
I13
sS'm'
(lp1249
I1
aI3
aI5
aI9
aI23
aI29
aI77
aI53
aI399
aI767
aI1817
aI2171
aI1629
asS'd'
I625
sa(dp1250
S'a'
I953
sS's'
I13
sS'm'
(lp1251
I1
aI1
aI3
aI5
aI29
aI5
aI43
aI121
aI17
aI859
aI1479
aI3785
aI6641
asS'd'
I626
sa(dp1252
S'a'
I956
sS's'
I13
sS'm'
(lp1253
I1
aI1
aI3
aI7
aI7
aI61
aI45
aI109
aI371
aI833
aI91
aI153
aI4553
asS'd'
I627
sa(dp1254
S'a'
I961
sS's'
I13
sS'm'
(lp1255
I1
aI1
aI3
aI11
aI7
aI55
aI81
aI123
aI389
aI139
aI1933
aI891
aI1789
asS'd'
I628
sa(dp1256
S'a'
I968
sS's'
I13
sS'm'
(lp1257
I1
aI3
aI7
aI15
aI25
aI17
aI93
aI165
aI503
aI717
aI1553
aI1475
aI1627
asS'd'
I629
sa(dp1258
S'a'
I976
sS's'
I13
sS'm'
(lp1259
I1
aI1
aI1
aI13
aI13
aI63
aI13
aI225
aI357
aI571
aI33
aI4073
aI3795
asS'd'
I630
sa(dp1260
S'a'
I988
sS's'
I13
sS'm'
(lp1261
I1
aI1
aI3
aI11
aI1
aI31
aI107
aI145
aI407
aI961
aI501
aI2987
aI103
asS'd'
I631
sa(dp1262
S'a'
I995
sS's'
I13
sS'm'
(lp1263
I1
aI1
aI7
aI1
aI23
aI63
aI49
aI193
aI173
aI281
aI25
aI2465
aI5927
asS'd'
I632
sa(dp1264
S'a'
I997
sS's'
I13
sS'm'
(lp1265
I1
aI1
aI7
aI1
aI1
aI1
aI85
aI77
aI273
aI693
aI349
aI1239
aI4503
asS'd'
I633
sa(dp1266
S'a'
I1007
sS's'
I13
sS'm'
(lp1267
I1
aI1
aI5
aI11
aI7
aI61
aI9
aI121
aI25
aI357
aI1443
aI405
aI7827
asS'd'
I634
sa(dp1268
S'a'
I1015
sS's'
I13
sS'm'
(lp1269
I1
aI1
aI7
aI13
aI11
aI53
aI11
aI207
aI145
aI211
aI1703
aI1081
aI2117
asS'd'
I635
sa(dp1270
S'a'
I1016
sS's'
I13
sS'm'
(lp1271
I1
aI1
aI3
aI11
aI27
aI23
aI19
aI9
aI297
aI279
aI1481
aI2273
aI6387
asS'd'
I636
sa(dp1272
S'a'
I1027
sS's'
I13
sS'm'
(lp1273
I1
aI3
aI3
aI5
aI15
aI45
aI3
aI41
aI305
aI87
aI1815
aI3461
aI5349
asS'd'
I637
sa(dp1274
S'a'
I1036
sS's'
I13
sS'm'
(lp1275
I1
aI3
aI3
aI13
aI9
aI37
aI79
aI125
aI259
aI561
aI1087
aI4091
aI793
asS'd'
I638
sa(dp1276
S'a'
I1039
sS's'
I13
sS'm'
(lp1277
I1
aI3
aI5
aI7
aI31
aI55
aI7
aI145
aI347
aI929
aI589
aI2783
aI5905
asS'd'
I639
sa(dp1278
S'a'
I1041
sS's'
I13
sS'm'
(lp1279
I1
aI1
aI7
aI15
aI3
aI25
aI1
aI181
aI13
aI243
aI653
aI2235
aI7445
asS'd'
I640
sa(dp1280
S'a'
I1048
sS's'
I13
sS'm'
(lp1281
I1
aI3
aI5
aI5
aI17
aI53
aI65
aI7
aI33
aI583
aI1363
aI1313
aI2319
asS'd'
I641
sa(dp1282
S'a'
I1053
sS's'
I13
sS'm'
(lp1283
I1
aI3
aI3
aI7
aI27
aI47
aI97
aI201
aI187
aI321
aI63
aI1515
aI7917
asS'd'
I642
sa(dp1284
S'a'
I1054
sS's'
I13
sS'm'
(lp1285
I1
aI1
aI3
aI5
aI23
aI9
aI3
aI165
aI61
aI19
aI1789
aI3783
aI3037
asS'd'
I643
sa(dp1286
S'a'
I1058
sS's'
I13
sS'm'
(lp1287
I1
aI3
aI1
aI13
aI15
aI43
aI125
aI191
aI67
aI273
aI1551
aI2227
aI5253
asS'd'
I644
sa(dp1288
S'a'
I1075
sS's'
I13
sS'm'
(lp1289
I1
aI1
aI1
aI13
aI25
aI53
aI107
aI33
aI299
aI249
aI1475
aI2233
aI907
asS'd'
I645
sa(dp1290
S'a'
I1082
sS's'
I13
sS'm'
(lp1291
I1
aI3
aI5
aI1
aI23
aI37
aI85
aI17
aI207
aI643
aI665
aI2933
aI5199
asS'd'
I646
sa(dp1292
S'a'
I1090
sS's'
I13
sS'm'
(lp1293
I1
aI1
aI7
aI7
aI25
aI57
aI59
aI41
aI15
aI751
aI751
aI1749
aI7053
asS'd'
I647
sa(dp1294
S'a'
I1109
sS's'
I13
sS'm'
(lp1295
I1
aI3
aI3
aI1
aI13
aI25
aI127
aI93
aI281
aI613
aI875
aI2223
aI6345
asS'd'
I648
sa(dp1296
S'a'
I1110
sS's'
I13
sS'm'
(lp1297
I1
aI1
aI5
aI3
aI29
aI55
aI79
aI249
aI43
aI317
aI533
aI995
aI1991
asS'd'
I649
sa(dp1298
S'a'
I1119
sS's'
I13
sS'm'
(lp1299
I1
aI3
aI3
aI15
aI17
aI49
aI79
aI31
aI193
aI233
aI1437
aI2615
aI819
asS'd'
I650
sa(dp1300
S'a'
I1126
sS's'
I13
sS'm'
(lp1301
I1
aI1
aI5
aI15
aI25
aI3
aI123
aI145
aI377
aI9
aI455
aI1191
aI3953
asS'd'
I651
sa(dp1302
S'a'
I1130
sS's'
I13
sS'm'
(lp1303
I1
aI3
aI5
aI3
aI15
aI19
aI41
aI231
aI81
aI393
aI3
aI19
aI2409
asS'd'
I652
sa(dp1304
S'a'
I1135
sS's'
I13
sS'm'
(lp1305
I1
aI1
aI3
aI1
aI27
aI43
aI113
aI179
aI7
aI853
aI947
aI2731
aI297
asS'd'
I653
sa(dp1306
S'a'
I1137
sS's'
I13
sS'm'
(lp1307
I1
aI1
aI1
aI11
aI29
aI39
aI53
aI191
aI443
aI689
aI529
aI3329
aI7431
asS'd'
I654
sa(dp1308
S'a'
I1140
sS's'
I13
sS'm'
(lp1309
I1
aI3
aI7
aI5
aI3
aI29
aI19
aI67
aI441
aI113
aI949
aI2769
aI4169
asS'd'
I655
sa(dp1310
S'a'
I1149
sS's'
I13
sS'm'
(lp1311
I1
aI3
aI5
aI11
aI11
aI55
aI85
aI169
aI215
aI815
aI803
aI2345
aI3967
asS'd'
I656
sa(dp1312
S'a'
I1156
sS's'
I13
sS'm'
(lp1313
I1
aI1
aI7
aI9
aI5
aI45
aI111
aI5
aI419
aI375
aI303
aI1725
aI4489
asS'd'
I657
sa(dp1314
S'a'
I1159
sS's'
I13
sS'm'
(lp1315
I1
aI3
aI5
aI15
aI29
aI43
aI79
aI19
aI23
aI417
aI381
aI541
aI4923
asS'd'
I658
sa(dp1316
S'a'
I1160
sS's'
I13
sS'm'
(lp1317
I1
aI1
aI3
aI15
aI3
aI31
aI117
aI39
aI117
aI305
aI1227
aI1223
aI143
asS'd'
I659
sa(dp1318
S'a'
I1165
sS's'
I13
sS'm'
(lp1319
I1
aI1
aI5
aI9
aI5
aI47
aI87
aI239
aI181
aI353
aI1561
aI3313
aI1921
asS'd'
I660
sa(dp1320
S'a'
I1173
sS's'
I13
sS'm'
(lp1321
I1
aI3
aI3
aI1
aI3
aI15
aI53
aI221
aI441
aI987
aI1997
aI2529
aI8059
asS'd'
I661
sa(dp1322
S'a'
I1178
sS's'
I13
sS'm'
(lp1323
I1
aI1
aI7
aI11
aI15
aI57
aI111
aI139
aI137
aI883
aI1881
aI2823
aI5661
asS'd'
I662
sa(dp1324
S'a'
I1183
sS's'
I13
sS'm'
(lp1325
I1
aI3
aI5
aI5
aI21
aI11
aI5
aI13
aI27
aI973
aI587
aI1331
aI1373
asS'd'
I663
sa(dp1326
S'a'
I1184
sS's'
I13
sS'm'
(lp1327
I1
aI1
aI7
aI11
aI29
aI51
aI93
aI29
aI217
aI221
aI55
aI2477
aI1979
asS'd'
I664
sa(dp1328
S'a'
I1189
sS's'
I13
sS'm'
(lp1329
I1
aI3
aI3
aI13
aI3
aI11
aI49
aI75
aI379
aI371
aI1441
aI793
aI7633
asS'd'
I665
sa(dp1330
S'a'
I1194
sS's'
I13
sS'm'
(lp1331
I1
aI1
aI1
aI13
aI19
aI45
aI89
aI249
aI91
aI649
aI1695
aI915
aI5619
asS'd'
I666
sa(dp1332
S'a'
I1211
sS's'
I13
sS'm'
(lp1333
I1
aI3
aI1
aI7
aI7
aI29
aI1
aI77
aI313
aI895
aI519
aI771
aI295
asS'd'
I667
sa(dp1334
S'a'
I1214
sS's'
I13
sS'm'
(lp1335
I1
aI3
aI1
aI15
aI5
aI3
aI1
aI57
aI331
aI109
aI485
aI2853
aI6831
asS'd'
I668
sa(dp1336
S'a'
I1216
sS's'
I13
sS'm'
(lp1337
I1
aI1
aI1
aI15
aI17
aI3
aI35
aI99
aI245
aI971
aI839
aI2509
aI2803
asS'd'
I669
sa(dp1338
S'a'
I1225
sS's'
I13
sS'm'
(lp1339
I1
aI3
aI3
aI3
aI9
aI37
aI57
aI251
aI325
aI317
aI529
aI1313
aI6379
asS'd'
I670
sa(dp1340
S'a'
I1231
sS's'
I13
sS'm'
(lp1341
I1
aI1
aI1
aI15
aI25
aI59
aI1
aI119
aI95
aI15
aI795
aI2375
aI6463
asS'd'
I671
sa(dp1342
S'a'
I1239
sS's'
I13
sS'm'
(lp1343
I1
aI3
aI1
aI5
aI1
aI49
aI117
aI21
aI47
aI179
aI863
aI85
aI1669
asS'd'
I672
sa(dp1344
S'a'
I1243
sS's'
I13
sS'm'
(lp1345
I1
aI3
aI7
aI3
aI9
aI37
aI19
aI221
aI455
aI973
aI571
aI1427
aI817
asS'd'
I673
sa(dp1346
S'a'
I1246
sS's'
I13
sS'm'
(lp1347
I1
aI1
aI1
aI15
aI17
aI9
aI67
aI213
aI127
aI887
aI1299
aI2913
aI7451
asS'd'
I674
sa(dp1348
S'a'
I1249
sS's'
I13
sS'm'
(lp1349
I1
aI3
aI1
aI13
aI27
aI27
aI41
aI43
aI171
aI623
aI691
aI391
aI4885
asS'd'
I675
sa(dp1350
S'a'
I1259
sS's'
I13
sS'm'
(lp1351
I1
aI3
aI1
aI13
aI17
aI17
aI123
aI239
aI143
aI227
aI1151
aI519
aI6543
asS'd'
I676
sa(dp1352
S'a'
I1273
sS's'
I13
sS'm'
(lp1353
I1
aI3
aI7
aI5
aI7
aI63
aI97
aI39
aI101
aI555
aI1057
aI381
aI7891
asS'd'
I677
sa(dp1354
S'a'
I1274
sS's'
I13
sS'm'
(lp1355
I1
aI3
aI5
aI1
aI3
aI27
aI85
aI129
aI161
aI875
aI1945
aI3541
aI695
asS'd'
I678
sa(dp1356
S'a'
I1281
sS's'
I13
sS'm'
(lp1357
I1
aI3
aI3
aI5
aI21
aI59
aI25
aI183
aI35
aI25
aI987
aI1459
aI181
asS'd'
I679
sa(dp1358
S'a'
I1287
sS's'
I13
sS'm'
(lp1359
I1
aI3
aI5
aI13
aI1
aI15
aI127
aI237
aI349
aI337
aI1491
aI2383
aI7811
asS'd'
I680
sa(dp1360
S'a'
I1294
sS's'
I13
sS'm'
(lp1361
I1
aI3
aI5
aI5
aI31
aI5
aI109
aI51
aI409
aI733
aI1395
aI3207
aI6049
asS'd'
I681
sa(dp1362
S'a'
I1296
sS's'
I13
sS'm'
(lp1363
I1
aI1
aI5
aI7
aI13
aI35
aI113
aI25
aI263
aI389
aI299
aI2521
aI1783
asS'd'
I682
sa(dp1364
S'a'
I1305
sS's'
I13
sS'm'
(lp1365
I1
aI3
aI7
aI11
aI15
aI47
aI97
aI73
aI55
aI75
aI113
aI2695
aI1023
asS'd'
I683
sa(dp1366
S'a'
I1306
sS's'
I13
sS'm'
(lp1367
I1
aI3
aI1
aI1
aI3
aI13
aI69
aI211
aI289
aI483
aI1335
aI787
aI677
asS'd'
I684
sa(dp1368
S'a'
I1318
sS's'
I13
sS'm'
(lp1369
I1
aI1
aI3
aI3
aI17
aI7
aI37
aI77
aI505
aI137
aI1113
aI345
aI2975
asS'd'
I685
sa(dp1370
S'a'
I1332
sS's'
I13
sS'm'
(lp1371
I1
aI1
aI1
aI13
aI3
aI11
aI95
aI199
aI453
aI109
aI479
aI3725
aI239
asS'd'
I686
sa(dp1372
S'a'
I1335
sS's'
I13
sS'm'
(lp1373
I1
aI1
aI7
aI15
aI19
aI53
aI3
aI145
aI359
aI863
aI347
aI3833
aI3043
asS'd'
I687
sa(dp1374
S'a'
I1336
sS's'
I13
sS'm'
(lp1375
I1
aI1
aI7
aI15
aI25
aI63
aI127
aI129
aI125
aI195
aI155
aI2211
aI8153
asS'd'
I688
sa(dp1376
S'a'
I1341
sS's'
I13
sS'm'
(lp1377
I1
aI1
aI7
aI13
aI9
aI49
aI121
aI115
aI73
aI119
aI1851
aI727
aI47
asS'd'
I689
sa(dp1378
S'a'
I1342
sS's'
I13
sS'm'
(lp1379
I1
aI3
aI3
aI13
aI13
aI11
aI71
aI7
aI45
aI591
aI133
aI2407
aI5563
asS'd'
I690
sa(dp1380
S'a'
I1362
sS's'
I13
sS'm'
(lp1381
I1
aI1
aI1
aI13
aI23
aI29
aI87
aI89
aI501
aI71
aI1759
aI1119
aI687
asS'd'
I691
sa(dp1382
S'a'
I1364
sS's'
I13
sS'm'
(lp1383
I1
aI1
aI7
aI7
aI13
aI7
aI13
aI183
aI53
aI951
aI1877
aI3991
aI6771
asS'd'
I692
sa(dp1384
S'a'
I1368
sS's'
I13
sS'm'
(lp1385
I1
aI3
aI7
aI11
aI7
aI1
aI27
aI47
aI61
aI21
aI919
aI961
aI1091
asS'd'
I693
sa(dp1386
S'a'
I1378
sS's'
I13
sS'm'
(lp1387
I1
aI3
aI5
aI5
aI1
aI27
aI1
aI5
aI63
aI157
aI1297
aI1049
aI5893
asS'd'
I694
sa(dp1388
S'a'
I1387
sS's'
I13
sS'm'
(lp1389
I1
aI3
aI7
aI9
aI19
aI33
aI17
aI133
aI425
aI797
aI1721
aI153
aI119
asS'd'
I695
sa(dp1390
S'a'
I1389
sS's'
I13
sS'm'
(lp1391
I1
aI3
aI3
aI7
aI13
aI37
aI1
aI215
aI509
aI1003
aI61
aI2353
aI7511
asS'd'
I696
sa(dp1392
S'a'
I1397
sS's'
I13
sS'm'
(lp1393
I1
aI1
aI7
aI1
aI29
aI19
aI31
aI79
aI199
aI555
aI1209
aI1603
aI6089
asS'd'
I697
sa(dp1394
S'a'
I1401
sS's'
I13
sS'm'
(lp1395
I1
aI3
aI1
aI1
aI5
aI31
aI111
aI127
aI333
aI429
aI1863
aI3925
aI5411
asS'd'
I698
sa(dp1396
S'a'
I1408
sS's'
I13
sS'm'
(lp1397
I1
aI1
aI7
aI5
aI5
aI5
aI123
aI191
aI47
aI993
aI269
aI4051
aI2111
asS'd'
I699
sa(dp1398
S'a'
I1418
sS's'
I13
sS'm'
(lp1399
I1
aI1
aI5
aI15
aI1
aI9
aI87
aI5
aI47
aI463
aI865
aI1813
aI7357
asS'd'
I700
sa(dp1400
S'a'
I1425
sS's'
I13
sS'm'
(lp1401
I1
aI3
aI1
aI3
aI23
aI63
aI123
aI83
aI511
aI777
aI63
aI1285
aI4537
asS'd'
I701
sa(dp1402
S'a'
I1426
sS's'
I13
sS'm'
(lp1403
I1
aI3
aI3
aI7
aI27
aI25
aI31
aI65
aI441
aI529
aI1815
aI1893
aI323
asS'd'
I702
sa(dp1404
S'a'
I1431
sS's'
I13
sS'm'
(lp1405
I1
aI3
aI7
aI5
aI11
aI19
aI7
aI5
aI397
aI811
aI755
aI2883
aI4217
asS'd'
I703
sa(dp1406
S'a'
I1435
sS's'
I13
sS'm'
(lp1407
I1
aI3
aI1
aI13
aI9
aI21
aI13
aI7
aI271
aI539
aI1769
aI3243
aI5325
asS'd'
I704
sa(dp1408
S'a'
I1441
sS's'
I13
sS'm'
(lp1409
I1
aI1
aI7
aI1
aI31
aI13
aI47
aI131
aI181
aI457
aI1559
aI2663
aI6653
asS'd'
I705
sa(dp1410
S'a'
I1444
sS's'
I13
sS'm'
(lp1411
I1
aI3
aI3
aI7
aI29
aI55
aI25
aI203
aI419
aI91
aI437
aI1159
aI5691
asS'd'
I706
sa(dp1412
S'a'
I1462
sS's'
I13
sS'm'
(lp1413
I1
aI1
aI3
aI13
aI29
aI19
aI71
aI217
aI337
aI329
aI501
aI939
aI2205
asS'd'
I707
sa(dp1414
S'a'
I1471
sS's'
I13
sS'm'
(lp1415
I1
aI1
aI3
aI1
aI1
aI27
aI17
aI201
aI97
aI285
aI1269
aI4043
aI2207
asS'd'
I708
sa(dp1416
S'a'
I1474
sS's'
I13
sS'm'
(lp1417
I1
aI1
aI1
aI1
aI3
aI41
aI13
aI199
aI141
aI129
aI1515
aI3129
aI5969
asS'd'
I709
sa(dp1418
S'a'
I1483
sS's'
I13
sS'm'
(lp1419
I1
aI3
aI3
aI9
aI3
aI17
aI119
aI41
aI271
aI933
aI877
aI701
aI2197
asS'd'
I710
sa(dp1420
S'a'
I1485
sS's'
I13
sS'm'
(lp1421
I1
aI1
aI1
aI7
aI15
aI47
aI3
aI195
aI115
aI821
aI725
aI843
aI6071
asS'd'
I711
sa(dp1422
S'a'
I1494
sS's'
I13
sS'm'
(lp1423
I1
aI3
aI5
aI15
aI17
aI33
aI85
aI65
aI297
aI571
aI1123
aI2743
aI5727
asS'd'
I712
sa(dp1424
S'a'
I1497
sS's'
I13
sS'm'
(lp1425
I1
aI1
aI5
aI11
aI27
aI15
aI37
aI235
aI415
aI293
aI1439
aI2739
aI4171
asS'd'
I713
sa(dp1426
S'a'
I1516
sS's'
I13
sS'm'
(lp1427
I1
aI3
aI7
aI7
aI1
aI55
aI71
aI35
aI307
aI11
aI401
aI1881
aI933
asS'd'
I714
sa(dp1428
S'a'
I1522
sS's'
I13
sS'm'
(lp1429
I1
aI3
aI1
aI11
aI21
aI37
aI3
aI177
aI119
aI339
aI559
aI3991
aI3437
asS'd'
I715
sa(dp1430
S'a'
I1534
sS's'
I13
sS'm'
(lp1431
I1
aI3
aI3
aI9
aI17
aI17
aI97
aI119
aI301
aI169
aI157
aI3267
aI2261
asS'd'
I716
sa(dp1432
S'a'
I1543
sS's'
I13
sS'm'
(lp1433
I1
aI3
aI3
aI9
aI29
aI3
aI111
aI101
aI355
aI869
aI375
aI2609
aI7377
asS'd'
I717
sa(dp1434
S'a'
I1552
sS's'
I13
sS'm'
(lp1435
I1
aI3
aI5
aI9
aI7
aI21
aI123
aI99
aI343
aI693
aI1927
aI1605
aI4923
asS'd'
I718
sa(dp1436
S'a'
I1557
sS's'
I13
sS'm'
(lp1437
I1
aI1
aI3
aI5
aI13
aI31
aI99
aI17
aI75
aI385
aI1539
aI1553
aI7077
asS'd'
I719
sa(dp1438
S'a'
I1558
sS's'
I13
sS'm'
(lp1439
I1
aI3
aI3
aI5
aI31
aI35
aI107
aI11
aI407
aI1019
aI1317
aI3593
aI7203
asS'd'
I720
sa(dp1440
S'a'
I1567
sS's'
I13
sS'm'
(lp1441
I1
aI3
aI3
aI13
aI17
aI33
aI99
aI245
aI401
aI957
aI157
aI1949
aI1571
asS'd'
I721
sa(dp1442
S'a'
I1568
sS's'
I13
sS'm'
(lp1443
I1
aI3
aI1
aI11
aI27
aI15
aI11
aI109
aI429
aI307
aI1911
aI2701
aI861
asS'd'
I722
sa(dp1444
S'a'
I1574
sS's'
I13
sS'm'
(lp1445
I1
aI1
aI5
aI13
aI13
aI35
aI55
aI255
aI311
aI957
aI1803
aI2673
aI5195
asS'd'
I723
sa(dp1446
S'a'
I1592
sS's'
I13
sS'm'
(lp1447
I1
aI1
aI1
aI11
aI19
aI3
aI89
aI37
aI211
aI783
aI1355
aI3567
aI7135
asS'd'
I724
sa(dp1448
S'a'
I1605
sS's'
I13
sS'm'
(lp1449
I1
aI1
aI5
aI5
aI21
aI49
aI79
aI17
aI509
aI331
aI183
aI3831
aI855
asS'd'
I725
sa(dp1450
S'a'
I1606
sS's'
I13
sS'm'
(lp1451
I1
aI3
aI7
aI5
aI29
aI19
aI85
aI109
aI105
aI523
aI845
aI3385
aI7477
asS'd'
I726
sa(dp1452
S'a'
I1610
sS's'
I13
sS'm'
(lp1453
I1
aI1
aI1
aI7
aI25
aI17
aI125
aI131
aI53
aI757
aI253
aI2989
aI2939
asS'd'
I727
sa(dp1454
S'a'
I1617
sS's'
I13
sS'm'
(lp1455
I1
aI3
aI3
aI9
aI19
aI23
aI105
aI39
aI351
aI677
aI211
aI401
aI8103
asS'd'
I728
sa(dp1456
S'a'
I1623
sS's'
I13
sS'm'
(lp1457
I1
aI3
aI5
aI1
aI5
aI11
aI17
aI3
aI405
aI469
aI1569
aI2865
aI3133
asS'd'
I729
sa(dp1458
S'a'
I1630
sS's'
I13
sS'm'
(lp1459
I1
aI1
aI3
aI13
aI15
aI5
aI117
aI179
aI139
aI145
aI477
aI1137
aI2537
asS'd'
I730
sa(dp1460
S'a'
I1634
sS's'
I13
sS'm'
(lp1461
I1
aI1
aI7
aI9
aI5
aI21
aI9
aI93
aI211
aI963
aI1207
aI3343
aI4911
asS'd'
I731
sa(dp1462
S'a'
I1640
sS's'
I13
sS'm'
(lp1463
I1
aI1
aI1
aI9
aI13
aI43
aI17
aI53
aI81
aI793
aI1571
aI2523
aI3683
asS'd'
I732
sa(dp1464
S'a'
I1643
sS's'
I13
sS'm'
(lp1465
I1
aI3
aI3
aI13
aI25
aI21
aI5
aI59
aI489
aI987
aI1941
aI171
aI6009
asS'd'
I733
sa(dp1466
S'a'
I1648
sS's'
I13
sS'm'
(lp1467
I1
aI3
aI3
aI7
aI1
aI39
aI89
aI171
aI403
aI467
aI1767
aI3423
aI2791
asS'd'
I734
sa(dp1468
S'a'
I1651
sS's'
I13
sS'm'
(lp1469
I1
aI1
aI3
aI9
aI19
aI49
aI91
aI125
aI163
aI1013
aI89
aI2849
aI6785
asS'd'
I735
sa(dp1470
S'a'
I1653
sS's'
I13
sS'm'
(lp1471
I1
aI1
aI5
aI9
aI9
aI11
aI15
aI241
aI43
aI297
aI1719
aI1541
aI1821
asS'd'
I736
sa(dp1472
S'a'
I1670
sS's'
I13
sS'm'
(lp1473
I1
aI3
aI7
aI15
aI29
aI23
aI103
aI239
aI191
aI33
aI1043
aI3649
aI6579
asS'd'
I737
sa(dp1474
S'a'
I1676
sS's'
I13
sS'm'
(lp1475
I1
aI3
aI3
aI9
aI21
aI51
aI123
aI55
aI223
aI645
aI1463
aI4021
aI5891
asS'd'
I738
sa(dp1476
S'a'
I1684
sS's'
I13
sS'm'
(lp1477
I1
aI1
aI5
aI7
aI3
aI41
aI27
aI235
aI391
aI303
aI2021
aI3187
aI7607
asS'd'
I739
sa(dp1478
S'a'
I1687
sS's'
I13
sS'm'
(lp1479
I1
aI1
aI1
aI9
aI5
aI49
aI49
aI29
aI377
aI251
aI1887
aI1017
aI1301
asS'd'
I740
sa(dp1480
S'a'
I1691
sS's'
I13
sS'm'
(lp1481
I1
aI1
aI3
aI3
aI13
aI41
aI27
aI47
aI223
aI23
aI517
aI3227
aI6731
asS'd'
I741
sa(dp1482
S'a'
I1693
sS's'
I13
sS'm'
(lp1483
I1
aI1
aI7
aI1
aI31
aI25
aI47
aI9
aI511
aI623
aI2047
aI1263
aI1511
asS'd'
I742
sa(dp1484
S'a'
I1698
sS's'
I13
sS'm'
(lp1485
I1
aI1
aI3
aI15
aI15
aI23
aI53
aI1
aI261
aI595
aI85
aI241
aI7047
asS'd'
I743
sa(dp1486
S'a'
I1709
sS's'
I13
sS'm'
(lp1487
I1
aI3
aI3
aI11
aI17
aI5
aI81
aI73
aI149
aI781
aI2035
aI3163
aI4247
asS'd'
I744
sa(dp1488
S'a'
I1715
sS's'
I13
sS'm'
(lp1489
I1
aI3
aI7
aI7
aI29
aI59
aI49
aI79
aI397
aI901
aI1105
aI2191
aI6277
asS'd'
I745
sa(dp1490
S'a'
I1722
sS's'
I13
sS'm'
(lp1491
I1
aI3
aI3
aI11
aI13
aI27
aI25
aI173
aI107
aI73
aI1265
aI585
aI5251
asS'd'
I746
sa(dp1492
S'a'
I1732
sS's'
I13
sS'm'
(lp1493
I1
aI1
aI7
aI15
aI29
aI23
aI73
aI229
aI235
aI887
aI1469
aI4073
aI2591
asS'd'
I747
sa(dp1494
S'a'
I1735
sS's'
I13
sS'm'
(lp1495
I1
aI1
aI3
aI9
aI17
aI15
aI83
aI173
aI207
aI879
aI1701
aI1509
aI11
asS'd'
I748
sa(dp1496
S'a'
I1747
sS's'
I13
sS'm'
(lp1497
I1
aI1
aI3
aI5
aI5
aI37
aI65
aI161
aI39
aI421
aI1153
aI2007
aI5355
asS'd'
I749
sa(dp1498
S'a'
I1749
sS's'
I13
sS'm'
(lp1499
I1
aI1
aI7
aI11
aI23
aI37
aI5
aI11
aI9
aI499
aI17
aI157
aI5747
asS'd'
I750
sa(dp1500
S'a'
I1754
sS's'
I13
sS'm'
(lp1501
I1
aI3
aI7
aI13
aI25
aI9
aI49
aI7
aI39
aI945
aI1349
aI1759
aI1441
asS'd'
I751
sa(dp1502
S'a'
I1777
sS's'
I13
sS'm'
(lp1503
I1
aI1
aI5
aI3
aI21
aI15
aI113
aI81
aI265
aI837
aI333
aI3625
aI6133
asS'd'
I752
sa(dp1504
S'a'
I1784
sS's'
I13
sS'm'
(lp1505
I1
aI3
aI1
aI11
aI13
aI27
aI73
aI109
aI297
aI327
aI299
aI3253
aI6957
asS'd'
I753
sa(dp1506
S'a'
I1790
sS's'
I13
sS'm'
(lp1507
I1
aI1
aI3
aI13
aI19
aI39
aI123
aI73
aI65
aI5
aI1061
aI2187
aI5055
asS'd'
I754
sa(dp1508
S'a'
I1795
sS's'
I13
sS'm'
(lp1509
I1
aI1
aI3
aI1
aI11
aI31
aI21
aI115
aI453
aI857
aI711
aI495
aI549
asS'd'
I755
sa(dp1510
S'a'
I1801
sS's'
I13
sS'm'
(lp1511
I1
aI3
aI7
aI7
aI15
aI29
aI79
aI103
aI47
aI713
aI1735
aI3121
aI6321
asS'd'
I756
sa(dp1512
S'a'
I1802
sS's'
I13
sS'm'
(lp1513
I1
aI1
aI5
aI5
aI29
aI9
aI97
aI33
aI471
aI705
aI329
aI1501
aI1349
asS'd'
I757
sa(dp1514
S'a'
I1812
sS's'
I13
sS'm'
(lp1515
I1
aI3
aI3
aI1
aI21
aI9
aI111
aI209
aI71
aI47
aI491
aI2143
aI1797
asS'd'
I758
sa(dp1516
S'a'
I1828
sS's'
I13
sS'm'
(lp1517
I1
aI3
aI3
aI3
aI11
aI39
aI21
aI135
aI445
aI259
aI607
aI3811
aI5449
asS'd'
I759
sa(dp1518
S'a'
I1831
sS's'
I13
sS'm'
(lp1519
I1
aI1
aI7
aI9
aI11
aI25
aI113
aI251
aI395
aI317
aI317
aI91
aI1979
asS'd'
I760
sa(dp1520
S'a'
I1837
sS's'
I13
sS'm'
(lp1521
I1
aI3
aI1
aI9
aI3
aI21
aI103
aI133
aI389
aI943
aI1235
aI1749
aI7063
asS'd'
I761
sa(dp1522
S'a'
I1838
sS's'
I13
sS'm'
(lp1523
I1
aI1
aI3
aI7
aI1
aI11
aI5
aI15
aI497
aI477
aI479
aI3079
aI6969
asS'd'
I762
sa(dp1524
S'a'
I1840
sS's'
I13
sS'm'
(lp1525
I1
aI1
aI3
aI3
aI15
aI39
aI105
aI131
aI475
aI465
aI181
aI865
aI3813
asS'd'
I763
sa(dp1526
S'a'
I1845
sS's'
I13
sS'm'
(lp1527
I1
aI1
aI7
aI9
aI19
aI63
aI123
aI131
aI415
aI525
aI457
aI2471
aI3135
asS'd'
I764
sa(dp1528
S'a'
I1863
sS's'
I13
sS'm'
(lp1529
I1
aI3
aI7
aI15
aI25
aI35
aI123
aI45
aI341
aI805
aI485
aI4049
aI7065
asS'd'
I765
sa(dp1530
S'a'
I1864
sS's'
I13
sS'm'
(lp1531
I1
aI1
aI1
aI5
aI29
aI9
aI47
aI227
aI51
aI867
aI1873
aI1593
aI2271
asS'd'
I766
sa(dp1532
S'a'
I1867
sS's'
I13
sS'm'
(lp1533
I1
aI1
aI7
aI15
aI31
aI9
aI71
aI117
aI285
aI711
aI837
aI1435
aI6275
asS'd'
I767
sa(dp1534
S'a'
I1870
sS's'
I13
sS'm'
(lp1535
I1
aI3
aI1
aI1
aI5
aI19
aI79
aI25
aI301
aI415
aI1871
aI645
aI3251
asS'd'
I768
sa(dp1536
S'a'
I1877
sS's'
I13
sS'm'
(lp1537
I1
aI3
aI1
aI3
aI17
aI51
aI99
aI185
aI447
aI43
aI523
aI219
aI429
asS'd'
I769
sa(dp1538
S'a'
I1881
sS's'
I13
sS'm'
(lp1539
I1
aI3
aI1
aI13
aI29
aI13
aI51
aI93
aI7
aI995
aI757
aI3017
aI6865
asS'd'
I770
sa(dp1540
S'a'
I1884
sS's'
I13
sS'm'
(lp1541
I1
aI1
aI3
aI15
aI7
aI25
aI75
aI17
aI155
aI981
aI1231
aI1229
aI1995
asS'd'
I771
sa(dp1542
S'a'
I1903
sS's'
I13
sS'm'
(lp1543
I1
aI3
aI5
aI3
aI27
aI45
aI71
aI73
aI225
aI763
aI377
aI1139
aI2863
asS'd'
I772
sa(dp1544
S'a'
I1917
sS's'
I13
sS'm'
(lp1545
I1
aI1
aI3
aI1
aI1
aI39
aI69
aI113
aI29
aI371
aI1051
aI793
aI3749
asS'd'
I773
sa(dp1546
S'a'
I1918
sS's'
I13
sS'm'
(lp1547
I1
aI1
aI3
aI13
aI23
aI61
aI27
aI183
aI307
aI431
aI1345
aI2757
aI4031
asS'd'
I774
sa(dp1548
S'a'
I1922
sS's'
I13
sS'm'
(lp1549
I1
aI3
aI7
aI5
aI5
aI59
aI117
aI197
aI303
aI721
aI877
aI723
aI1601
asS'd'
I775
sa(dp1550
S'a'
I1924
sS's'
I13
sS'm'
(lp1551
I1
aI3
aI5
aI1
aI27
aI33
aI99
aI237
aI485
aI711
aI665
aI3077
aI5105
asS'd'
I776
sa(dp1552
S'a'
I1928
sS's'
I13
sS'm'
(lp1553
I1
aI1
aI3
aI1
aI13
aI9
aI103
aI201
aI23
aI951
aI2029
aI165
aI2093
asS'd'
I777
sa(dp1554
S'a'
I1931
sS's'
I13
sS'm'
(lp1555
I1
aI3
aI5
aI13
aI5
aI29
aI55
aI85
aI221
aI677
aI611
aI3613
aI4567
asS'd'
I778
sa(dp1556
S'a'
I1951
sS's'
I13
sS'm'
(lp1557
I1
aI1
aI1
aI1
aI7
aI61
aI9
aI233
aI261
aI561
aI953
aI4023
aI2443
asS'd'
I779
sa(dp1558
S'a'
I1952
sS's'
I13
sS'm'
(lp1559
I1
aI3
aI3
aI13
aI1
aI17
aI103
aI71
aI223
aI213
aI833
aI1747
aI6999
asS'd'
I780
sa(dp1560
S'a'
I1957
sS's'
I13
sS'm'
(lp1561
I1
aI3
aI5
aI15
aI25
aI53
aI57
aI187
aI25
aI695
aI1207
aI4089
aI2877
asS'd'
I781
sa(dp1562
S'a'
I1958
sS's'
I13
sS'm'
(lp1563
I1
aI1
aI7
aI1
aI7
aI31
aI87
aI129
aI493
aI519
aI1555
aI1155
aI4637
asS'd'
I782
sa(dp1564
S'a'
I1964
sS's'
I13
sS'm'
(lp1565
I1
aI1
aI1
aI15
aI21
aI17
aI23
aI29
aI19
aI255
aI927
aI1791
aI3093
asS'd'
I783
sa(dp1566
S'a'
I1967
sS's'
I13
sS'm'
(lp1567
I1
aI1
aI3
aI9
aI17
aI33
aI95
aI129
aI175
aI461
aI287
aI2633
aI2325
asS'd'
I784
sa(dp1568
S'a'
I1970
sS's'
I13
sS'm'
(lp1569
I1
aI3
aI5
aI7
aI23
aI19
aI63
aI209
aI249
aI583
aI1373
aI2039
aI2225
asS'd'
I785
sa(dp1570
S'a'
I1972
sS's'
I13
sS'm'
(lp1571
I1
aI3
aI3
aI5
aI5
aI19
aI79
aI241
aI459
aI355
aI1455
aI3313
aI3639
asS'd'
I786
sa(dp1572
S'a'
I1994
sS's'
I13
sS'm'
(lp1573
I1
aI1
aI7
aI9
aI21
aI41
aI97
aI119
aI129
aI769
aI1541
aI3495
aI7741
asS'd'
I787
sa(dp1574
S'a'
I2002
sS's'
I13
sS'm'
(lp1575
I1
aI1
aI7
aI11
aI9
aI29
aI35
aI255
aI141
aI937
aI1763
aI41
aI1393
asS'd'
I788
sa(dp1576
S'a'
I2007
sS's'
I13
sS'm'
(lp1577
I1
aI3
aI7
aI1
aI13
aI51
aI61
aI157
aI177
aI847
aI1829
aI3539
aI285
asS'd'
I789
sa(dp1578
S'a'
I2008
sS's'
I13
sS'm'
(lp1579
I1
aI1
aI1
aI15
aI21
aI13
aI9
aI55
aI397
aI19
aI1495
aI1255
aI7235
asS'd'
I790
sa(dp1580
S'a'
I2023
sS's'
I13
sS'm'
(lp1581
I1
aI1
aI7
aI7
aI25
aI37
aI53
aI237
aI319
aI197
aI269
aI1205
aI1485
asS'd'
I791
sa(dp1582
S'a'
I2030
sS's'
I13
sS'm'
(lp1583
I1
aI1
aI5
aI15
aI23
aI17
aI35
aI247
aI323
aI807
aI233
aI3681
aI4407
asS'd'
I792
sa(dp1584
S'a'
I2035
sS's'
I13
sS'm'
(lp1585
I1
aI1
aI3
aI7
aI9
aI59
aI85
aI105
aI493
aI763
aI1639
aI391
aI1451
asS'd'
I793
sa(dp1586
S'a'
I2038
sS's'
I13
sS'm'
(lp1587
I1
aI3
aI3
aI9
aI15
aI33
aI5
aI253
aI129
aI625
aI1527
aI2793
aI6057
asS'd'
I794
sa(dp1588
S'a'
I2042
sS's'
I13
sS'm'
(lp1589
I1
aI3
aI1
aI1
aI7
aI47
aI21
aI161
aI235
aI83
aI397
aI3563
aI5953
asS'd'
I795
sa(dp1590
S'a'
I2047
sS's'
I13
sS'm'
(lp1591
I1
aI3
aI7
aI11
aI3
aI41
aI25
aI117
aI375
aI779
aI1297
aI3715
aI8117
asS'd'
I796
sa(dp1592
S'a'
I2051
sS's'
I13
sS'm'
(lp1593
I1
aI1
aI3
aI7
aI31
aI19
aI103
aI173
aI475
aI189
aI2035
aI2921
aI1107
asS'd'
I797
sa(dp1594
S'a'
I2058
sS's'
I13
sS'm'
(lp1595
I1
aI1
aI7
aI3
aI25
aI7
aI93
aI255
aI307
aI113
aI1893
aI2233
aI6919
asS'd'
I798
sa(dp1596
S'a'
I2060
sS's'
I13
sS'm'
(lp1597
I1
aI3
aI5
aI15
aI9
aI57
aI79
aI143
aI165
aI5
aI1389
aI193
aI693
asS'd'
I799
sa(dp1598
S'a'
I2071
sS's'
I13
sS'm'
(lp1599
I1
aI3
aI5
aI1
aI29
aI45
aI91
aI49
aI189
aI461
aI439
aI1283
aI7835
asS'd'
I800
sa(dp1600
S'a'
I2084
sS's'
I13
sS'm'
(lp1601
I1
aI1
aI3
aI13
aI11
aI61
aI41
aI231
aI373
aI695
aI395
aI915
aI5393
asS'd'
I801
sa(dp1602
S'a'
I2087
sS's'
I13
sS'm'
(lp1603
I1
aI3
aI7
aI11
aI5
aI51
aI67
aI53
aI483
aI95
aI1943
aI247
aI5653
asS'd'
I802
sa(dp1604
S'a'
I2099
sS's'
I13
sS'm'
(lp1605
I1
aI3
aI7
aI5
aI5
aI57
aI45
aI235
aI137
aI793
aI1069
aI1661
aI1557
asS'd'
I803
sa(dp1606
S'a'
I2108
sS's'
I13
sS'm'
(lp1607
I1
aI3
aI5
aI3
aI25
aI55
aI103
aI177
aI81
aI861
aI1151
aI143
aI7655
asS'd'
I804
sa(dp1608
S'a'
I2111
sS's'
I13
sS'm'
(lp1609
I1
aI1
aI3
aI1
aI21
aI41
aI67
aI131
aI253
aI431
aI1269
aI3181
aI3429
asS'd'
I805
sa(dp1610
S'a'
I2120
sS's'
I13
sS'm'
(lp1611
I1
aI3
aI1
aI1
aI21
aI7
aI77
aI221
aI257
aI663
aI71
aI2949
aI2481
asS'd'
I806
sa(dp1612
S'a'
I2128
sS's'
I13
sS'm'
(lp1613
I1
aI3
aI5
aI3
aI3
aI23
aI45
aI107
aI299
aI739
aI1013
aI3
aI3165
asS'd'
I807
sa(dp1614
S'a'
I2138
sS's'
I13
sS'm'
(lp1615
I1
aI1
aI5
aI1
aI3
aI37
aI109
aI37
aI243
aI983
aI1221
aI1691
aI3869
asS'd'
I808
sa(dp1616
S'a'
I2143
sS's'
I13
sS'm'
(lp1617
I1
aI1
aI5
aI5
aI31
aI7
aI5
aI193
aI397
aI867
aI1495
aI3435
aI7441
asS'd'
I809
sa(dp1618
S'a'
I2144
sS's'
I13
sS'm'
(lp1619
I1
aI1
aI1
aI1
aI17
aI59
aI97
aI233
aI389
aI597
aI1013
aI1631
aI483
asS'd'
I810
sa(dp1620
S'a'
I2153
sS's'
I13
sS'm'
(lp1621
I1
aI1
aI1
aI11
aI7
aI41
aI107
aI53
aI111
aI125
aI1513
aI1921
aI7647
asS'd'
I811
sa(dp1622
S'a'
I2156
sS's'
I13
sS'm'
(lp1623
I1
aI3
aI3
aI3
aI31
aI29
aI117
aI3
aI365
aI971
aI1139
aI2123
aI5913
asS'd'
I812
sa(dp1624
S'a'
I2162
sS's'
I13
sS'm'
(lp1625
I1
aI1
aI1
aI13
aI23
aI3
aI1
aI167
aI475
aI639
aI1811
aI3841
aI3081
asS'd'
I813
sa(dp1626
S'a'
I2167
sS's'
I13
sS'm'
(lp1627
I1
aI1
aI5
aI3
aI5
aI47
aI65
aI123
aI275
aI783
aI95
aI119
aI7591
asS'd'
I814
sa(dp1628
S'a'
I2178
sS's'
I13
sS'm'
(lp1629
I1
aI3
aI1
aI15
aI13
aI33
aI93
aI237
aI467
aI431
aI705
aI4013
aI4035
asS'd'
I815
sa(dp1630
S'a'
I2183
sS's'
I13
sS'm'
(lp1631
I1
aI3
aI5
aI1
aI19
aI7
aI101
aI231
aI155
aI737
aI1381
aI3343
aI2051
asS'd'
I816
sa(dp1632
S'a'
I2202
sS's'
I13
sS'm'
(lp1633
I1
aI1
aI5
aI9
aI15
aI49
aI45
aI163
aI433
aI765
aI2031
aI201
aI2589
asS'd'
I817
sa(dp1634
S'a'
I2211
sS's'
I13
sS'm'
(lp1635
I1
aI3
aI7
aI9
aI19
aI41
aI31
aI89
aI93
aI623
aI105
aI745
aI4409
asS'd'
I818
sa(dp1636
S'a'
I2214
sS's'
I13
sS'm'
(lp1637
I1
aI1
aI5
aI1
aI11
aI45
aI127
aI85
aI389
aI439
aI829
aI477
aI7965
asS'd'
I819
sa(dp1638
S'a'
I2223
sS's'
I13
sS'm'
(lp1639
I1
aI3
aI3
aI15
aI13
aI41
aI1
aI207
aI435
aI585
aI311
aI1725
aI2737
asS'd'
I820
sa(dp1640
S'a'
I2225
sS's'
I13
sS'm'
(lp1641
I1
aI3
aI3
aI3
aI13
aI49
aI21
aI31
aI197
aI799
aI1411
aI2959
aI7133
asS'd'
I821
sa(dp1642
S'a'
I2232
sS's'
I13
sS'm'
(lp1643
I1
aI3
aI1
aI3
aI7
aI43
aI9
aI141
aI133
aI579
aI1059
aI93
aI957
asS'd'
I822
sa(dp1644
S'a'
I2237
sS's'
I13
sS'm'
(lp1645
I1
aI3
aI7
aI1
aI15
aI51
aI23
aI213
aI381
aI851
aI699
aI2261
aI3419
asS'd'
I823
sa(dp1646
S'a'
I2257
sS's'
I13
sS'm'
(lp1647
I1
aI3
aI5
aI9
aI25
aI35
aI67
aI141
aI35
aI409
aI1423
aI365
aI1645
asS'd'
I824
sa(dp1648
S'a'
I2260
sS's'
I13
sS'm'
(lp1649
I1
aI3
aI3
aI11
aI15
aI33
aI27
aI181
aI93
aI87
aI1761
aI3511
aI1353
asS'd'
I825
sa(dp1650
S'a'
I2267
sS's'
I13
sS'm'
(lp1651
I1
aI3
aI5
aI3
aI25
aI63
aI111
aI137
aI321
aI819
aI705
aI1547
aI7271
asS'd'
I826
sa(dp1652
S'a'
I2274
sS's'
I13
sS'm'
(lp1653
I1
aI3
aI1
aI1
aI5
aI57
aI99
aI59
aI411
aI757
aI1371
aI3953
aI3695
asS'd'
I827
sa(dp1654
S'a'
I2276
sS's'
I13
sS'm'
(lp1655
I1
aI3
aI5
aI11
aI11
aI21
aI25
aI147
aI239
aI455
aI709
aI953
aI7175
asS'd'
I828
sa(dp1656
S'a'
I2285
sS's'
I13
sS'm'
(lp1657
I1
aI3
aI3
aI15
aI5
aI53
aI91
aI205
aI341
aI63
aI723
aI1565
aI7135
asS'd'
I829
sa(dp1658
S'a'
I2288
sS's'
I13
sS'm'
(lp1659
I1
aI1
aI7
aI15
aI11
aI21
aI99
aI79
aI63
aI593
aI2007
aI3629
aI5271
asS'd'
I830
sa(dp1660
S'a'
I2293
sS's'
I13
sS'm'
(lp1661
I1
aI3
aI3
aI1
aI9
aI21
aI45
aI175
aI453
aI435
aI1855
aI2649
aI6959
asS'd'
I831
sa(dp1662
S'a'
I2294
sS's'
I13
sS'm'
(lp1663
I1
aI1
aI3
aI15
aI15
aI33
aI121
aI121
aI251
aI431
aI1127
aI3305
aI4199
asS'd'
I832
sa(dp1664
S'a'
I2297
sS's'
I13
sS'm'
(lp1665
I1
aI1
aI1
aI9
aI31
aI15
aI71
aI29
aI345
aI391
aI1159
aI2809
aI345
asS'd'
I833
sa(dp1666
S'a'
I2303
sS's'
I13
sS'm'
(lp1667
I1
aI3
aI7
aI1
aI23
aI29
aI95
aI151
aI327
aI727
aI647
aI1623
aI2971
asS'd'
I834
sa(dp1668
S'a'
I2308
sS's'
I13
sS'm'
(lp1669
I1
aI1
aI7
aI7
aI9
aI29
aI79
aI91
aI127
aI909
aI1293
aI1315
aI5315
asS'd'
I835
sa(dp1670
S'a'
I2311
sS's'
I13
sS'm'
(lp1671
I1
aI1
aI5
aI11
aI13
aI37
aI89
aI73
aI149
aI477
aI1909
aI3343
aI525
asS'd'
I836
sa(dp1672
S'a'
I2318
sS's'
I13
sS'm'
(lp1673
I1
aI3
aI5
aI7
aI5
aI59
aI55
aI255
aI223
aI459
aI2027
aI237
aI4205
asS'd'
I837
sa(dp1674
S'a'
I2323
sS's'
I13
sS'm'
(lp1675
I1
aI1
aI1
aI7
aI27
aI11
aI95
aI65
aI325
aI835
aI907
aI3801
aI3787
asS'd'
I838
sa(dp1676
S'a'
I2332
sS's'
I13
sS'm'
(lp1677
I1
aI1
aI1
aI11
aI27
aI33
aI99
aI175
aI51
aI913
aI331
aI1851
aI4133
asS'd'
I839
sa(dp1678
S'a'
I2341
sS's'
I13
sS'm'
(lp1679
I1
aI3
aI5
aI5
aI13
aI37
aI31
aI99
aI273
aI409
aI1827
aI3845
aI5491
asS'd'
I840
sa(dp1680
S'a'
I2345
sS's'
I13
sS'm'
(lp1681
I1
aI1
aI3
aI7
aI23
aI19
aI107
aI85
aI283
aI523
aI509
aI451
aI421
asS'd'
I841
sa(dp1682
S'a'
I2348
sS's'
I13
sS'm'
(lp1683
I1
aI3
aI5
aI7
aI13
aI9
aI51
aI81
aI87
aI619
aI61
aI2803
aI5271
asS'd'
I842
sa(dp1684
S'a'
I2354
sS's'
I13
sS'm'
(lp1685
I1
aI1
aI1
aI15
aI9
aI45
aI35
aI219
aI401
aI271
aI953
aI649
aI6847
asS'd'
I843
sa(dp1686
S'a'
I2368
sS's'
I13
sS'm'
(lp1687
I1
aI1
aI7
aI11
aI9
aI45
aI17
aI219
aI169
aI837
aI1483
aI1605
aI2901
asS'd'
I844
sa(dp1688
S'a'
I2377
sS's'
I13
sS'm'
(lp1689
I1
aI1
aI7
aI7
aI21
aI43
aI37
aI33
aI291
aI359
aI71
aI2899
aI7037
asS'd'
I845
sa(dp1690
S'a'
I2380
sS's'
I13
sS'm'
(lp1691
I1
aI3
aI3
aI13
aI31
aI53
aI37
aI15
aI149
aI949
aI551
aI3445
aI5455
asS'd'
I846
sa(dp1692
S'a'
I2383
sS's'
I13
sS'm'
(lp1693
I1
aI3
aI1
aI5
aI19
aI45
aI81
aI223
aI193
aI439
aI2047
aI3879
aI789
asS'd'
I847
sa(dp1694
S'a'
I2388
sS's'
I13
sS'm'
(lp1695
I1
aI1
aI7
aI3
aI11
aI63
aI35
aI61
aI255
aI563
aI459
aI2991
aI3359
asS'd'
I848
sa(dp1696
S'a'
I2395
sS's'
I13
sS'm'
(lp1697
I1
aI1
aI5
aI9
aI13
aI49
aI47
aI185
aI239
aI221
aI1533
aI3635
aI2045
asS'd'
I849
sa(dp1698
S'a'
I2397
sS's'
I13
sS'm'
(lp1699
I1
aI3
aI7
aI3
aI25
aI37
aI127
aI223
aI51
aI357
aI483
aI3837
aI6873
asS'd'
I850
sa(dp1700
S'a'
I2401
sS's'
I13
sS'm'
(lp1701
I1
aI1
aI7
aI9
aI31
aI37
aI113
aI31
aI387
aI833
aI1243
aI1543
aI5535
asS'd'
I851
sa(dp1702
S'a'
I2411
sS's'
I13
sS'm'
(lp1703
I1
aI3
aI1
aI9
aI23
aI59
aI119
aI221
aI73
aI185
aI2007
aI2885
aI2563
asS'd'
I852
sa(dp1704
S'a'
I2413
sS's'
I13
sS'm'
(lp1705
I1
aI1
aI1
aI13
aI7
aI33
aI53
aI179
aI67
aI185
aI1541
aI1807
aI4659
asS'd'
I853
sa(dp1706
S'a'
I2419
sS's'
I13
sS'm'
(lp1707
I1
aI3
aI1
aI11
aI31
aI37
aI23
aI215
aI269
aI357
aI207
aI645
aI4219
asS'd'
I854
sa(dp1708
S'a'
I2435
sS's'
I13
sS'm'
(lp1709
I1
aI3
aI3
aI13
aI19
aI27
aI107
aI55
aI91
aI71
aI1695
aI1815
aI89
asS'd'
I855
sa(dp1710
S'a'
I2442
sS's'
I13
sS'm'
(lp1711
I1
aI1
aI3
aI15
aI3
aI19
aI35
aI247
aI49
aI529
aI1523
aI3317
aI6151
asS'd'
I856
sa(dp1712
S'a'
I2455
sS's'
I13
sS'm'
(lp1713
I1
aI1
aI7
aI7
aI23
aI25
aI107
aI139
aI483
aI503
aI1277
aI243
aI7879
asS'd'
I857
sa(dp1714
S'a'
I2472
sS's'
I13
sS'm'
(lp1715
I1
aI3
aI3
aI13
aI3
aI15
aI11
aI197
aI135
aI839
aI985
aI275
aI5527
asS'd'
I858
sa(dp1716
S'a'
I2478
sS's'
I13
sS'm'
(lp1717
I1
aI3
aI5
aI3
aI25
aI47
aI95
aI21
aI113
aI307
aI1001
aI3065
aI295
asS'd'
I859
sa(dp1718
S'a'
I2490
sS's'
I13
sS'm'
(lp1719
I1
aI1
aI3
aI9
aI19
aI19
aI99
aI213
aI363
aI449
aI735
aI2851
aI2521
asS'd'
I860
sa(dp1720
S'a'
I2507
sS's'
I13
sS'm'
(lp1721
I1
aI1
aI3
aI9
aI5
aI49
aI63
aI61
aI157
aI857
aI497
aI2801
aI6987
asS'd'
I861
sa(dp1722
S'a'
I2509
sS's'
I13
sS'm'
(lp1723
I1
aI1
aI1
aI9
aI1
aI41
aI109
aI119
aI499
aI939
aI867
aI3675
aI8023
asS'd'
I862
sa(dp1724
S'a'
I2517
sS's'
I13
sS'm'
(lp1725
I1
aI3
aI1
aI1
aI13
aI33
aI109
aI123
aI289
aI3
aI1271
aI2773
aI4265
asS'd'
I863
sa(dp1726
S'a'
I2524
sS's'
I13
sS'm'
(lp1727
I1
aI3
aI1
aI11
aI9
aI57
aI83
aI221
aI95
aI43
aI1189
aI457
aI7133
asS'd'
I864
sa(dp1728
S'a'
I2528
sS's'
I13
sS'm'
(lp1729
I1
aI1
aI7
aI3
aI11
aI49
aI33
aI219
aI229
aI289
aI685
aI3359
aI4495
asS'd'
I865
sa(dp1730
S'a'
I2531
sS's'
I13
sS'm'
(lp1731
I1
aI3
aI1
aI3
aI19
aI43
aI67
aI193
aI41
aI771
aI407
aI81
aI3891
asS'd'
I866
sa(dp1732
S'a'
I2538
sS's'
I13
sS'm'
(lp1733
I1
aI1
aI7
aI11
aI5
aI29
aI51
aI175
aI297
aI539
aI1
aI2245
aI6439
asS'd'
I867
sa(dp1734
S'a'
I2545
sS's'
I13
sS'm'
(lp1735
I1
aI3
aI7
aI15
aI21
aI33
aI117
aI183
aI511
aI489
aI1283
aI3281
aI5979
asS'd'
I868
sa(dp1736
S'a'
I2546
sS's'
I13
sS'm'
(lp1737
I1
aI3
aI7
aI5
aI9
aI3
aI125
aI147
aI359
aI549
aI369
aI3049
aI2405
asS'd'
I869
sa(dp1738
S'a'
I2555
sS's'
I13
sS'm'
(lp1739
I1
aI3
aI5
aI7
aI19
aI5
aI65
aI97
aI483
aI377
aI1523
aI1457
aI2995
asS'd'
I870
sa(dp1740
S'a'
I2557
sS's'
I13
sS'm'
(lp1741
I1
aI1
aI5
aI1
aI11
aI21
aI41
aI113
aI277
aI131
aI1475
aI1043
aI2367
asS'd'
I871
sa(dp1742
S'a'
I2564
sS's'
I13
sS'm'
(lp1743
I1
aI3
aI3
aI1
aI15
aI17
aI101
aI69
aI443
aI865
aI817
aI1421
aI5231
asS'd'
I872
sa(dp1744
S'a'
I2573
sS's'
I13
sS'm'
(lp1745
I1
aI1
aI3
aI3
aI3
aI55
aI95
aI99
aI75
aI195
aI1929
aI3931
aI5855
asS'd'
I873
sa(dp1746
S'a'
I2579
sS's'
I13
sS'm'
(lp1747
I1
aI3
aI1
aI3
aI19
aI23
aI93
aI213
aI241
aI551
aI1307
aI585
aI7729
asS'd'
I874
sa(dp1748
S'a'
I2592
sS's'
I13
sS'm'
(lp1749
I1
aI3
aI1
aI11
aI23
aI15
aI53
aI249
aI467
aI519
aI95
aI741
aI409
asS'd'
I875
sa(dp1750
S'a'
I2598
sS's'
I13
sS'm'
(lp1751
I1
aI1
aI1
aI15
aI29
aI37
aI43
aI203
aI233
aI877
aI77
aI1933
aI2729
asS'd'
I876
sa(dp1752
S'a'
I2607
sS's'
I13
sS'm'
(lp1753
I1
aI3
aI7
aI11
aI27
aI39
aI43
aI161
aI255
aI15
aI1463
aI833
aI495
asS'd'
I877
sa(dp1754
S'a'
I2612
sS's'
I13
sS'm'
(lp1755
I1
aI1
aI7
aI11
aI3
aI53
aI81
aI67
aI375
aI823
aI1903
aI3061
aI395
asS'd'
I878
sa(dp1756
S'a'
I2619
sS's'
I13
sS'm'
(lp1757
I1
aI1
aI1
aI1
aI15
aI37
aI93
aI233
aI247
aI501
aI1321
aI3275
aI5409
asS'd'
I879
sa(dp1758
S'a'
I2621
sS's'
I13
sS'm'
(lp1759
I1
aI3
aI3
aI7
aI7
aI11
aI5
aI105
aI139
aI983
aI1239
aI531
aI3881
asS'd'
I880
sa(dp1760
S'a'
I2627
sS's'
I13
sS'm'
(lp1761
I1
aI1
aI5
aI3
aI19
aI49
aI107
aI227
aI361
aI101
aI355
aI2649
aI7383
asS'd'
I881
sa(dp1762
S'a'
I2633
sS's'
I13
sS'm'
(lp1763
I1
aI1
aI7
aI5
aI25
aI41
aI101
aI121
aI209
aI293
aI1937
aI2259
aI5557
asS'd'
I882
sa(dp1764
S'a'
I2636
sS's'
I13
sS'm'
(lp1765
I1
aI1
aI3
aI7
aI7
aI1
aI9
aI13
aI463
aI1019
aI995
aI3159
aI107
asS'd'
I883
sa(dp1766
S'a'
I2642
sS's'
I13
sS'm'
(lp1767
I1
aI3
aI5
aI11
aI5
aI35
aI127
aI97
aI261
aI789
aI807
aI807
aI6257
asS'd'
I884
sa(dp1768
S'a'
I2654
sS's'
I13
sS'm'
(lp1769
I1
aI1
aI7
aI5
aI11
aI13
aI45
aI91
aI417
aI101
aI1973
aI3645
aI2107
asS'd'
I885
sa(dp1770
S'a'
I2660
sS's'
I13
sS'm'
(lp1771
I1
aI1
aI3
aI7
aI5
aI63
aI57
aI49
aI203
aI157
aI115
aI1393
aI8117
asS'd'
I886
sa(dp1772
S'a'
I2669
sS's'
I13
sS'm'
(lp1773
I1
aI3
aI5
aI5
aI3
aI43
aI15
aI155
aI127
aI489
aI1165
aI3701
aI4867
asS'd'
I887
sa(dp1774
S'a'
I2675
sS's'
I13
sS'm'
(lp1775
I1
aI1
aI7
aI7
aI29
aI29
aI69
aI215
aI415
aI367
aI371
aI1901
aI6075
asS'd'
I888
sa(dp1776
S'a'
I2684
sS's'
I13
sS'm'
(lp1777
I1
aI1
aI1
aI3
aI11
aI33
aI89
aI149
aI433
aI705
aI1437
aI1597
aI505
asS'd'
I889
sa(dp1778
S'a'
I2694
sS's'
I13
sS'm'
(lp1779
I1
aI3
aI5
aI1
aI13
aI37
aI19
aI119
aI5
aI581
aI2037
aI1633
aI2099
asS'd'
I890
sa(dp1780
S'a'
I2703
sS's'
I13
sS'm'
(lp1781
I1
aI3
aI7
aI13
aI5
aI49
aI103
aI245
aI215
aI515
aI133
aI2007
aI1933
asS'd'
I891
sa(dp1782
S'a'
I2706
sS's'
I13
sS'm'
(lp1783
I1
aI3
aI1
aI9
aI1
aI3
aI25
aI197
aI253
aI387
aI1683
aI2267
aI221
asS'd'
I892
sa(dp1784
S'a'
I2712
sS's'
I13
sS'm'
(lp1785
I1
aI3
aI5
aI15
aI21
aI9
aI73
aI201
aI405
aI999
aI437
aI3877
aI6045
asS'd'
I893
sa(dp1786
S'a'
I2715
sS's'
I13
sS'm'
(lp1787
I1
aI1
aI3
aI1
aI31
aI55
aI25
aI83
aI421
aI395
aI1807
aI2129
aI7797
asS'd'
I894
sa(dp1788
S'a'
I2722
sS's'
I13
sS'm'
(lp1789
I1
aI1
aI3
aI1
aI23
aI21
aI121
aI183
aI125
aI347
aI143
aI3685
aI4317
asS'd'
I895
sa(dp1790
S'a'
I2727
sS's'
I13
sS'm'
(lp1791
I1
aI3
aI3
aI3
aI17
aI45
aI17
aI223
aI267
aI795
aI1815
aI1309
aI155
asS'd'
I896
sa(dp1792
S'a'
I2734
sS's'
I13
sS'm'
(lp1793
I1
aI1
aI1
aI15
aI17
aI59
aI5
aI133
aI15
aI715
aI1503
aI153
aI2887
asS'd'
I897
sa(dp1794
S'a'
I2742
sS's'
I13
sS'm'
(lp1795
I1
aI1
aI1
aI1
aI27
aI13
aI119
aI77
aI243
aI995
aI1851
aI3719
aI4695
asS'd'
I898
sa(dp1796
S'a'
I2745
sS's'
I13
sS'm'
(lp1797
I1
aI3
aI1
aI5
aI31
aI49
aI43
aI165
aI49
aI609
aI1265
aI1141
aI505
asS'd'
I899
sa(dp1798
S'a'
I2751
sS's'
I13
sS'm'
(lp1799
I1
aI1
aI7
aI13
aI11
aI63
aI21
aI253
aI229
aI585
aI1543
aI3719
aI4141
asS'd'
I900
sa(dp1800
S'a'
I2766
sS's'
I13
sS'm'
(lp1801
I1
aI3
aI7
aI11
aI23
aI27
aI17
aI131
aI295
aI895
aI1493
aI1411
aI3247
asS'd'
I901
sa(dp1802
S'a'
I2768
sS's'
I13
sS'm'
(lp1803
I1
aI1
aI5
aI9
aI29
aI7
aI97
aI15
aI113
aI445
aI859
aI1483
aI1121
asS'd'
I902
sa(dp1804
S'a'
I2780
sS's'
I13
sS'm'
(lp1805
I1
aI3
aI1
aI9
aI13
aI49
aI99
aI107
aI323
aI201
aI681
aI3071
aI5281
asS'd'
I903
sa(dp1806
S'a'
I2790
sS's'
I13
sS'm'
(lp1807
I1
aI1
aI1
aI15
aI9
aI19
aI61
aI161
aI7
aI87
aI587
aI2199
aI2811
asS'd'
I904
sa(dp1808
S'a'
I2794
sS's'
I13
sS'm'
(lp1809
I1
aI3
aI3
aI15
aI15
aI19
aI95
aI45
aI299
aI829
aI981
aI3479
aI487
asS'd'
I905
sa(dp1810
S'a'
I2796
sS's'
I13
sS'm'
(lp1811
I1
aI1
aI1
aI9
aI3
aI37
aI7
aI19
aI227
aI13
aI397
aI513
aI1257
asS'd'
I906
sa(dp1812
S'a'
I2801
sS's'
I13
sS'm'
(lp1813
I1
aI1
aI5
aI15
aI15
aI13
aI17
aI111
aI135
aI929
aI1145
aI811
aI1801
asS'd'
I907
sa(dp1814
S'a'
I2804
sS's'
I13
sS'm'
(lp1815
I1
aI3
aI1
aI3
aI27
aI57
aI31
aI19
aI279
aI103
aI693
aI631
aI3409
asS'd'
I908
sa(dp1816
S'a'
I2807
sS's'
I13
sS'm'
(lp1817
I1
aI1
aI1
aI1
aI15
aI13
aI67
aI83
aI23
aI799
aI1735
aI2063
aI3363
asS'd'
I909
sa(dp1818
S'a'
I2816
sS's'
I13
sS'm'
(lp1819
I1
aI3
aI3
aI7
aI3
aI1
aI61
aI31
aI41
aI533
aI2025
aI4067
aI6963
asS'd'
I910
sa(dp1820
S'a'
I2821
sS's'
I13
sS'm'
(lp1821
I1
aI1
aI5
aI7
aI17
aI27
aI81
aI79
aI107
aI205
aI29
aI97
aI4883
asS'd'
I911
sa(dp1822
S'a'
I2831
sS's'
I13
sS'm'
(lp1823
I1
aI1
aI1
aI5
aI19
aI49
aI91
aI201
aI283
aI949
aI651
aI3819
aI5073
asS'd'
I912
sa(dp1824
S'a'
I2834
sS's'
I13
sS'm'
(lp1825
I1
aI1
aI7
aI9
aI11
aI13
aI73
aI197
aI37
aI219
aI1931
aI3369
aI6017
asS'd'
I913
sa(dp1826
S'a'
I2839
sS's'
I13
sS'm'
(lp1827
I1
aI1
aI7
aI15
aI11
aI7
aI75
aI205
aI7
aI819
aI399
aI661
aI6487
asS'd'
I914
sa(dp1828
S'a'
I2845
sS's'
I13
sS'm'
(lp1829
I1
aI3
aI3
aI3
aI27
aI37
aI95
aI41
aI307
aI165
aI1077
aI3485
aI563
asS'd'
I915
sa(dp1830
S'a'
I2852
sS's'
I13
sS'm'
(lp1831
I1
aI3
aI5
aI3
aI21
aI49
aI57
aI179
aI109
aI627
aI1789
aI431
aI2941
asS'd'
I916
sa(dp1832
S'a'
I2856
sS's'
I13
sS'm'
(lp1833
I1
aI1
aI7
aI5
aI11
aI19
aI43
aI137
aI149
aI679
aI1543
aI245
aI1381
asS'd'
I917
sa(dp1834
S'a'
I2861
sS's'
I13
sS'm'
(lp1835
I1
aI3
aI5
aI5
aI15
aI3
aI69
aI81
aI135
aI159
aI1363
aI3401
aI6355
asS'd'
I918
sa(dp1836
S'a'
I2873
sS's'
I13
sS'm'
(lp1837
I1
aI3
aI5
aI1
aI9
aI61
aI49
aI53
aI319
aI25
aI1647
aI1297
aI615
asS'd'
I919
sa(dp1838
S'a'
I2874
sS's'
I13
sS'm'
(lp1839
I1
aI3
aI5
aI11
aI31
aI43
aI9
aI101
aI71
aI919
aI335
aI3147
aI5823
asS'd'
I920
sa(dp1840
S'a'
I2888
sS's'
I13
sS'm'
(lp1841
I1
aI3
aI1
aI1
aI15
aI5
aI29
aI109
aI511
aI945
aI867
aI3677
aI6915
asS'd'
I921
sa(dp1842
S'a'
I2893
sS's'
I13
sS'm'
(lp1843
I1
aI3
aI3
aI15
aI17
aI49
aI91
aI111
aI215
aI29
aI1879
aI97
aI2505
asS'd'
I922
sa(dp1844
S'a'
I2894
sS's'
I13
sS'm'
(lp1845
I1
aI3
aI1
aI13
aI19
aI61
aI11
aI111
aI163
aI777
aI533
aI1113
aI5339
asS'd'
I923
sa(dp1846
S'a'
I2902
sS's'
I13
sS'm'
(lp1847
I1
aI1
aI7
aI9
aI17
aI55
aI117
aI91
aI455
aI289
aI557
aI913
aI4455
asS'd'
I924
sa(dp1848
S'a'
I2917
sS's'
I13
sS'm'
(lp1849
I1
aI3
aI1
aI7
aI25
aI19
aI123
aI37
aI1
aI277
aI717
aI2965
aI4469
asS'd'
I925
sa(dp1850
S'a'
I2921
sS's'
I13
sS'm'
(lp1851
I1
aI3
aI7
aI3
aI19
aI23
aI87
aI235
aI209
aI457
aI2041
aI2893
aI1805
asS'd'
I926
sa(dp1852
S'a'
I2922
sS's'
I13
sS'm'
(lp1853
I1
aI3
aI3
aI5
aI5
aI43
aI23
aI61
aI351
aI791
aI59
aI2009
aI2909
asS'd'
I927
sa(dp1854
S'a'
I2929
sS's'
I13
sS'm'
(lp1855
I1
aI1
aI3
aI7
aI5
aI1
aI27
aI231
aI385
aI257
aI1261
aI2701
aI1807
asS'd'
I928
sa(dp1856
S'a'
I2935
sS's'
I13
sS'm'
(lp1857
I1
aI3
aI1
aI1
aI27
aI19
aI87
aI253
aI131
aI685
aI1743
aI3983
aI2651
asS'd'
I929
sa(dp1858
S'a'
I2946
sS's'
I13
sS'm'
(lp1859
I1
aI3
aI7
aI11
aI21
aI17
aI11
aI81
aI191
aI641
aI1821
aI3005
aI7251
asS'd'
I930
sa(dp1860
S'a'
I2951
sS's'
I13
sS'm'
(lp1861
I1
aI3
aI3
aI5
aI15
aI31
aI41
aI213
aI55
aI931
aI1953
aI49
aI6037
asS'd'
I931
sa(dp1862
S'a'
I2957
sS's'
I13
sS'm'
(lp1863
I1
aI1
aI7
aI15
aI7
aI27
aI65
aI223
aI113
aI79
aI1875
aI911
aI5445
asS'd'
I932
sa(dp1864
S'a'
I2960
sS's'
I13
sS'm'
(lp1865
I1
aI3
aI7
aI7
aI23
aI55
aI51
aI167
aI495
aI25
aI1585
aI3447
aI799
asS'd'
I933
sa(dp1866
S'a'
I2966
sS's'
I13
sS'm'
(lp1867
I1
aI1
aI3
aI7
aI27
aI15
aI95
aI193
aI337
aI415
aI975
aI3085
aI967
asS'd'
I934
sa(dp1868
S'a'
I2972
sS's'
I13
sS'm'
(lp1869
I1
aI1
aI7
aI15
aI19
aI7
aI93
aI41
aI433
aI551
aI401
aI3169
aI3971
asS'd'
I935
sa(dp1870
S'a'
I2976
sS's'
I13
sS'm'
(lp1871
I1
aI1
aI7
aI11
aI13
aI15
aI53
aI69
aI433
aI59
aI1117
aI3359
aI6231
asS'd'
I936
sa(dp1872
S'a'
I2979
sS's'
I13
sS'm'
(lp1873
I1
aI1
aI7
aI3
aI23
aI5
aI115
aI201
aI225
aI109
aI1903
aI3897
aI6265
asS'd'
I937
sa(dp1874
S'a'
I2985
sS's'
I13
sS'm'
(lp1875
I1
aI1
aI1
aI11
aI17
aI1
aI39
aI143
aI361
aI659
aI1105
aI23
aI4923
asS'd'
I938
sa(dp1876
S'a'
I3000
sS's'
I13
sS'm'
(lp1877
I1
aI1
aI1
aI9
aI27
aI57
aI85
aI227
aI261
aI119
aI1881
aI3965
aI6999
asS'd'
I939
sa(dp1878
S'a'
I3003
sS's'
I13
sS'm'
(lp1879
I1
aI3
aI7
aI7
aI15
aI7
aI107
aI17
aI315
aI49
aI1591
aI905
aI7789
asS'd'
I940
sa(dp1880
S'a'
I3013
sS's'
I13
sS'm'
(lp1881
I1
aI3
aI1
aI7
aI29
aI3
aI47
aI237
aI157
aI769
aI839
aI3199
aI3195
asS'd'
I941
sa(dp1882
S'a'
I3018
sS's'
I13
sS'm'
(lp1883
I1
aI1
aI3
aI15
aI25
aI39
aI63
aI15
aI111
aI857
aI881
aI1505
aI7671
asS'd'
I942
sa(dp1884
S'a'
I3020
sS's'
I13
sS'm'
(lp1885
I1
aI1
aI7
aI1
aI3
aI35
aI41
aI215
aI99
aI895
aI1025
aI1483
aI4707
asS'd'
I943
sa(dp1886
S'a'
I3025
sS's'
I13
sS'm'
(lp1887
I1
aI3
aI5
aI1
aI1
aI31
aI25
aI247
aI113
aI841
aI397
aI1825
aI6969
asS'd'
I944
sa(dp1888
S'a'
I3042
sS's'
I13
sS'm'
(lp1889
I1
aI1
aI3
aI5
aI19
aI41
aI49
aI243
aI225
aI973
aI241
aI175
aI1041
asS'd'
I945
sa(dp1890
S'a'
I3047
sS's'
I13
sS'm'
(lp1891
I1
aI1
aI1
aI7
aI15
aI15
aI105
aI141
aI83
aI75
aI1675
aI3523
aI5219
asS'd'
I946
sa(dp1892
S'a'
I3048
sS's'
I13
sS'm'
(lp1893
I1
aI1
aI7
aI5
aI13
aI27
aI47
aI199
aI445
aI841
aI959
aI1157
aI2209
asS'd'
I947
sa(dp1894
S'a'
I3051
sS's'
I13
sS'm'
(lp1895
I1
aI3
aI5
aI15
aI23
aI31
aI31
aI81
aI85
aI33
aI785
aI2639
aI7799
asS'd'
I948
sa(dp1896
S'a'
I3054
sS's'
I13
sS'm'
(lp1897
I1
aI1
aI5
aI13
aI21
aI3
aI47
aI99
aI235
aI943
aI1731
aI2467
aI7891
asS'd'
I949
sa(dp1898
S'a'
I3056
sS's'
I13
sS'm'
(lp1899
I1
aI1
aI1
aI3
aI17
aI53
aI85
aI219
aI73
aI131
aI1339
aI875
aI1191
asS'd'
I950
sa(dp1900
S'a'
I3065
sS's'
I13
sS'm'
(lp1901
I1
aI1
aI5
aI7
aI17
aI63
aI113
aI7
aI185
aI557
aI749
aI3563
aI4973
asS'd'
I951
sa(dp1902
S'a'
I3073
sS's'
I13
sS'm'
(lp1903
I1
aI3
aI3
aI15
aI15
aI21
aI43
aI111
aI155
aI689
aI345
aI423
aI3597
asS'd'
I952
sa(dp1904
S'a'
I3074
sS's'
I13
sS'm'
(lp1905
I1
aI1
aI5
aI1
aI15
aI29
aI93
aI5
aI361
aI713
aI695
aI3937
aI425
asS'd'
I953
sa(dp1906
S'a'
I3083
sS's'
I13
sS'm'
(lp1907
I1
aI3
aI7
aI7
aI13
aI41
aI115
aI175
aI315
aI937
aI123
aI2841
aI4457
asS'd'
I954
sa(dp1908
S'a'
I3086
sS's'
I13
sS'm'
(lp1909
I1
aI1
aI3
aI11
aI25
aI5
aI103
aI53
aI423
aI811
aI657
aI399
aI7257
asS'd'
I955
sa(dp1910
S'a'
I3091
sS's'
I13
sS'm'
(lp1911
I1
aI1
aI1
aI1
aI1
aI13
aI101
aI211
aI383
aI325
aI97
aI1703
aI4429
asS'd'
I956
sa(dp1912
S'a'
I3097
sS's'
I13
sS'm'
(lp1913
I1
aI3
aI7
aI9
aI31
aI45
aI83
aI157
aI509
aI701
aI841
aI1105
aI3643
asS'd'
I957
sa(dp1914
S'a'
I3109
sS's'
I13
sS'm'
(lp1915
I1
aI1
aI1
aI7
aI1
aI9
aI69
aI17
aI129
aI281
aI1161
aI2945
aI7693
asS'd'
I958
sa(dp1916
S'a'
I3116
sS's'
I13
sS'm'
(lp1917
I1
aI3
aI7
aI1
aI11
aI29
aI51
aI143
aI77
aI433
aI1723
aI2317
aI5641
asS'd'
I959
sa(dp1918
S'a'
I3124
sS's'
I13
sS'm'
(lp1919
I1
aI1
aI1
aI1
aI21
aI43
aI13
aI67
aI177
aI505
aI1629
aI1267
aI4885
asS'd'
I960
sa(dp1920
S'a'
I3128
sS's'
I13
sS'm'
(lp1921
I1
aI1
aI3
aI11
aI27
aI63
aI111
aI47
aI233
aI781
aI453
aI1679
aI3209
asS'd'
I961
sa(dp1922
S'a'
I3153
sS's'
I13
sS'm'
(lp1923
I1
aI1
aI3
aI13
aI29
aI27
aI119
aI141
aI493
aI971
aI461
aI1159
aI633
asS'd'
I962
sa(dp1924
S'a'
I3160
sS's'
I13
sS'm'
(lp1925
I1
aI1
aI3
aI15
aI23
aI5
aI79
aI215
aI163
aI149
aI1805
aI2399
aI61
asS'd'
I963
sa(dp1926
S'a'
I3165
sS's'
I13
sS'm'
(lp1927
I1
aI3
aI5
aI13
aI19
aI5
aI1
aI39
aI409
aI561
aI709
aI829
aI1357
asS'd'
I964
sa(dp1928
S'a'
I3172
sS's'
I13
sS'm'
(lp1929
I1
aI3
aI3
aI13
aI19
aI43
aI9
aI177
aI449
aI447
aI73
aI2107
aI5669
asS'd'
I965
sa(dp1930
S'a'
I3175
sS's'
I13
sS'm'
(lp1931
I1
aI3
aI5
aI1
aI23
aI13
aI63
aI109
aI203
aI593
aI829
aI4017
aI6881
asS'd'
I966
sa(dp1932
S'a'
I3184
sS's'
I13
sS'm'
(lp1933
I1
aI1
aI5
aI7
aI3
aI9
aI53
aI175
aI391
aI169
aI1283
aI3793
aI4451
asS'd'
I967
sa(dp1934
S'a'
I3193
sS's'
I13
sS'm'
(lp1935
I1
aI1
aI5
aI7
aI29
aI43
aI9
aI5
aI209
aI77
aI927
aI2941
aI8145
asS'd'
I968
sa(dp1936
S'a'
I3196
sS's'
I13
sS'm'
(lp1937
I1
aI3
aI5
aI15
aI17
aI49
aI5
aI143
aI131
aI771
aI1685
aI925
aI2175
asS'd'
I969
sa(dp1938
S'a'
I3200
sS's'
I13
sS'm'
(lp1939
I1
aI1
aI3
aI11
aI27
aI27
aI27
aI159
aI161
aI1015
aI1587
aI4049
aI1983
asS'd'
I970
sa(dp1940
S'a'
I3203
sS's'
I13
sS'm'
(lp1941
I1
aI3
aI1
aI3
aI23
aI57
aI119
aI67
aI481
aI577
aI389
aI3319
aI5325
asS'd'
I971
sa(dp1942
S'a'
I3205
sS's'
I13
sS'm'
(lp1943
I1
aI3
aI5
aI1
aI19
aI39
aI87
aI61
aI329
aI657
aI1773
aI31
aI1707
asS'd'
I972
sa(dp1944
S'a'
I3209
sS's'
I13
sS'm'
(lp1945
I1
aI1
aI3
aI1
aI5
aI25
aI15
aI241
aI131
aI815
aI1751
aI3029
aI8039
asS'd'
I973
sa(dp1946
S'a'
I3224
sS's'
I13
sS'm'
(lp1947
I1
aI3
aI3
aI13
aI27
aI13
aI77
aI87
aI437
aI57
aI621
aI1031
aI7891
asS'd'
I974
sa(dp1948
S'a'
I3239
sS's'
I13
sS'm'
(lp1949
I1
aI3
aI1
aI13
aI23
aI51
aI117
aI37
aI331
aI745
aI605
aI3179
aI4713
asS'd'
I975
sa(dp1950
S'a'
I3251
sS's'
I13
sS'm'
(lp1951
I1
aI1
aI5
aI5
aI19
aI17
aI99
aI167
aI87
aI721
aI737
aI789
aI2165
asS'd'
I976
sa(dp1952
S'a'
I3254
sS's'
I13
sS'm'
(lp1953
I1
aI3
aI5
aI13
aI1
aI51
aI119
aI211
aI165
aI299
aI1327
aI3053
aI3343
asS'd'
I977
sa(dp1954
S'a'
I3265
sS's'
I13
sS'm'
(lp1955
I1
aI1
aI5
aI15
aI29
aI45
aI17
aI129
aI67
aI345
aI1553
aI2705
aI7369
asS'd'
I978
sa(dp1956
S'a'
I3266
sS's'
I13
sS'm'
(lp1957
I1
aI1
aI1
aI9
aI23
aI7
aI13
aI209
aI7
aI407
aI317
aI3077
aI7287
asS'd'
I979
sa(dp1958
S'a'
I3275
sS's'
I13
sS'm'
(lp1959
I1
aI1
aI1
aI5
aI9
aI59
aI89
aI3
aI487
aI451
aI505
aI2499
aI7563
asS'd'
I980
sa(dp1960
S'a'
I3280
sS's'
I13
sS'm'
(lp1961
I1
aI3
aI1
aI7
aI21
aI1
aI21
aI203
aI101
aI417
aI1389
aI2751
aI1397
asS'd'
I981
sa(dp1962
S'a'
I3283
sS's'
I13
sS'm'
(lp1963
I1
aI3
aI7
aI13
aI7
aI31
aI3
aI247
aI349
aI485
aI1259
aI549
aI6321
asS'd'
I982
sa(dp1964
S'a'
I3286
sS's'
I13
sS'm'
(lp1965
I1
aI1
aI7
aI7
aI27
aI33
aI107
aI197
aI293
aI729
aI1753
aI2571
aI103
asS'd'
I983
sa(dp1966
S'a'
I3301
sS's'
I13
sS'm'
(lp1967
I1
aI3
aI5
aI9
aI25
aI35
aI5
aI253
aI137
aI213
aI2041
aI3387
aI1809
asS'd'
I984
sa(dp1968
S'a'
I3302
sS's'
I13
sS'm'
(lp1969
I1
aI1
aI7
aI13
aI15
aI35
aI67
aI83
aI295
aI175
aI839
aI2831
aI839
asS'd'
I985
sa(dp1970
S'a'
I3305
sS's'
I13
sS'm'
(lp1971
I1
aI3
aI3
aI11
aI3
aI17
aI55
aI141
aI247
aI991
aI117
aI3799
aI1221
asS'd'
I986
sa(dp1972
S'a'
I3319
sS's'
I13
sS'm'
(lp1973
I1
aI1
aI5
aI1
aI11
aI37
aI87
aI233
aI457
aI653
aI899
aI2933
aI3105
asS'd'
I987
sa(dp1974
S'a'
I3323
sS's'
I13
sS'm'
(lp1975
I1
aI1
aI3
aI15
aI3
aI31
aI67
aI167
aI437
aI9
aI651
aI1109
aI1139
asS'd'
I988
sa(dp1976
S'a'
I3326
sS's'
I13
sS'm'
(lp1977
I1
aI1
aI3
aI1
aI7
aI63
aI67
aI17
aI11
aI883
aI1855
aI1941
aI4751
asS'd'
I989
sa(dp1978
S'a'
I3331
sS's'
I13
sS'm'
(lp1979
I1
aI3
aI7
aI9
aI19
aI33
aI113
aI117
aI495
aI39
aI1795
aI2561
aI5519
asS'd'
I990
sa(dp1980
S'a'
I3348
sS's'
I13
sS'm'
(lp1981
I1
aI1
aI7
aI5
aI1
aI3
aI103
aI37
aI201
aI223
aI1101
aI877
aI6483
asS'd'
I991
sa(dp1982
S'a'
I3351
sS's'
I13
sS'm'
(lp1983
I1
aI1
aI5
aI9
aI29
aI49
aI51
aI33
aI439
aI917
aI861
aI1321
aI2135
asS'd'
I992
sa(dp1984
S'a'
I3358
sS's'
I13
sS'm'
(lp1985
I1
aI1
aI3
aI3
aI1
aI5
aI17
aI93
aI217
aI619
aI613
aI1357
aI6095
asS'd'
I993
sa(dp1986
S'a'
I3368
sS's'
I13
sS'm'
(lp1987
I1
aI3
aI1
aI11
aI3
aI21
aI5
aI41
aI15
aI175
aI843
aI2937
aI6849
asS'd'
I994
sa(dp1988
S'a'
I3374
sS's'
I13
sS'm'
(lp1989
I1
aI3
aI3
aI7
aI9
aI57
aI55
aI127
aI79
aI287
aI445
aI2205
aI7989
asS'd'
I995
sa(dp1990
S'a'
I3376
sS's'
I13
sS'm'
(lp1991
I1
aI1
aI7
aI13
aI23
aI17
aI93
aI129
aI157
aI135
aI1747
aI1813
aI4183
asS'd'
I996
sa(dp1992
S'a'
I3379
sS's'
I13
sS'm'
(lp1993
I1
aI1
aI1
aI5
aI31
aI59
aI99
aI33
aI425
aI329
aI887
aI367
aI1761
asS'd'
I997
sa(dp1994
S'a'
I3385
sS's'
I13
sS'm'
(lp1995
I1
aI1
aI7
aI9
aI17
aI53
aI77
aI139
aI435
aI387
aI49
aI3649
aI1773
asS'd'
I998
sa(dp1996
S'a'
I3386
sS's'
I13
sS'm'
(lp1997
I1
aI3
aI3
aI15
aI21
aI57
aI45
aI161
aI331
aI719
aI273
aI3479
aI4173
asS'd'
I999
sa(dp1998
S'a'
I3396
sS's'
I13
sS'm'
(lp1999
I1
aI1
aI3
aI9
aI3
aI3
aI105
aI201
aI373
aI877
aI919
aI1263
aI6649
asS'd'
I1000
sa(dp2000
S'a'
I3420
sS's'
I13
sS'm'
(lp2001
I1
aI3
aI1
aI15
aI13
aI43
aI13
aI99
aI73
aI163
aI353
aI3569
aI5601
asS'd'
I1001
sa(dp2002
S'a'
I3423
sS's'
I13
sS'm'
(lp2003
I1
aI3
aI7
aI3
aI5
aI9
aI69
aI177
aI449
aI47
aI781
aI1125
aI4245
asS'd'
I1002
sa(dp2004
S'a'
I3430
sS's'
I13
sS'm'
(lp2005
I1
aI1
aI1
aI5
aI3
aI45
aI1
aI123
aI409
aI903
aI205
aI2057
aI7637
asS'd'
I1003
sa(dp2006
S'a'
I3433
sS's'
I13
sS'm'
(lp2007
I1
aI3
aI5
aI9
aI19
aI47
aI87
aI135
aI481
aI799
aI101
aI3409
aI2241
asS'd'
I1004
sa(dp2008
S'a'
I3434
sS's'
I13
sS'm'
(lp2009
I1
aI3
aI1
aI13
aI3
aI25
aI15
aI27
aI181
aI967
aI669
aI2577
aI7249
asS'd'
I1005
sa(dp2010
S'a'
I3439
sS's'
I13
sS'm'
(lp2011
I1
aI1
aI7
aI3
aI31
aI5
aI103
aI53
aI1
aI911
aI1209
aI3697
aI6685
asS'd'
I1006
sa(dp2012
S'a'
I3442
sS's'
I13
sS'm'
(lp2013
I1
aI1
aI3
aI1
aI5
aI5
aI49
aI135
aI281
aI747
aI761
aI2973
aI7963
asS'd'
I1007
sa(dp2014
S'a'
I3444
sS's'
I13
sS'm'
(lp2015
I1
aI3
aI3
aI5
aI19
aI61
aI125
aI199
aI299
aI515
aI1365
aI369
aI7027
asS'd'
I1008
sa(dp2016
S'a'
I3453
sS's'
I13
sS'm'
(lp2017
I1
aI3
aI1
aI7
aI5
aI41
aI63
aI229
aI283
aI571
aI147
aI447
aI657
asS'd'
I1009
sa(dp2018
S'a'
I3464
sS's'
I13
sS'm'
(lp2019
I1
aI3
aI1
aI11
aI5
aI15
aI55
aI7
aI259
aI61
aI27
aI1429
aI5631
asS'd'
I1010
sa(dp2020
S'a'
I3477
sS's'
I13
sS'm'
(lp2021
I1
aI1
aI5
aI1
aI3
aI53
aI51
aI253
aI155
aI553
aI1293
aI3735
aI6567
asS'd'
I1011
sa(dp2022
S'a'
I3478
sS's'
I13
sS'm'
(lp2023
I1
aI3
aI5
aI9
aI5
aI41
aI21
aI159
aI101
aI785
aI1981
aI3799
aI7693
asS'd'
I1012
sa(dp2024
S'a'
I3482
sS's'
I13
sS'm'
(lp2025
I1
aI3
aI7
aI7
aI9
aI3
aI95
aI105
aI129
aI213
aI1215
aI1027
aI5699
asS'd'
I1013
sa(dp2026
S'a'
I3487
sS's'
I13
sS'm'
(lp2027
I1
aI1
aI3
aI3
aI29
aI13
aI9
aI253
aI449
aI321
aI341
aI2879
aI171
asS'd'
I1014
sa(dp2028
S'a'
I3497
sS's'
I13
sS'm'
(lp2029
I1
aI3
aI7
aI11
aI21
aI11
aI75
aI35
aI43
aI965
aI675
aI2217
aI7175
asS'd'
I1015
sa(dp2030
S'a'
I3500
sS's'
I13
sS'm'
(lp2031
I1
aI1
aI5
aI15
aI31
aI5
aI29
aI137
aI311
aI751
aI47
aI1367
aI5921
asS'd'
I1016
sa(dp2032
S'a'
I3505
sS's'
I13
sS'm'
(lp2033
I1
aI1
aI3
aI15
aI17
aI1
aI45
aI69
aI55
aI649
aI835
aI569
aI7615
asS'd'
I1017
sa(dp2034
S'a'
I3506
sS's'
I13
sS'm'
(lp2035
I1
aI3
aI1
aI13
aI31
aI7
aI23
aI15
aI391
aI145
aI1845
aI1825
aI1403
asS'd'
I1018
sa(dp2036
S'a'
I3511
sS's'
I13
sS'm'
(lp2037
I1
aI1
aI3
aI15
aI5
aI9
aI79
aI77
aI105
aI399
aI1933
aI2503
aI4781
asS'd'
I1019
sa(dp2038
S'a'
I3512
sS's'
I13
sS'm'
(lp2039
I1
aI3
aI1
aI3
aI17
aI47
aI19
aI13
aI107
aI475
aI759
aI2933
aI3761
asS'd'
I1020
sa(dp2040
S'a'
I3515
sS's'
I13
sS'm'
(lp2041
I1
aI1
aI7
aI11
aI3
aI7
aI121
aI209
aI397
aI877
aI293
aI847
aI7039
asS'd'
I1021
sa(dp2042
S'a'
I3525
sS's'
I13
sS'm'
(lp2043
I1
aI1
aI1
aI15
aI29
aI45
aI5
aI109
aI335
aI461
aI143
aI931
aI4045
asS'd'
I1022
sa(dp2044
S'a'
I3532
sS's'
I13
sS'm'
(lp2045
I1
aI3
aI1
aI7
aI11
aI57
aI73
aI89
aI201
aI173
aI803
aI3953
aI5205
asS'd'
I1023
sa(dp2046
S'a'
I3538
sS's'
I13
sS'm'
(lp2047
I1
aI1
aI5
aI11
aI11
aI33
aI37
aI29
aI263
aI1019
aI657
aI1453
aI7807
asS'd'
I1024
sa(dp2048
S'a'
I3540
sS's'
I13
sS'm'
(lp2049
I1
aI3
aI3
aI13
aI31
aI25
aI37
aI47
aI261
aI607
aI1703
aI2603
aI417
asS'd'
I1025
sa(dp2050
S'a'
I3547
sS's'
I13
sS'm'
(lp2051
I1
aI1
aI1
aI1
aI31
aI61
aI45
aI115
aI275
aI239
aI1989
aI1897
aI4329
asS'd'
I1026
sa(dp2052
S'a'
I3549
sS's'
I13
sS'm'
(lp2053
I1
aI3
aI5
aI3
aI31
aI3
aI11
aI173
aI335
aI579
aI1193
aI2219
aI7875
asS'd'
I1027
sa(dp2054
S'a'
I3560
sS's'
I13
sS'm'
(lp2055
I1
aI1
aI7
aI9
aI29
aI45
aI13
aI67
aI399
aI177
aI1293
aI3865
aI2225
asS'd'
I1028
sa(dp2056
S'a'
I3571
sS's'
I13
sS'm'
(lp2057
I1
aI1
aI7
aI11
aI11
aI51
aI121
aI227
aI469
aI905
aI929
aI2635
aI4165
asS'd'
I1029
sa(dp2058
S'a'
I3577
sS's'
I13
sS'm'
(lp2059
I1
aI3
aI7
aI9
aI13
aI39
aI55
aI167
aI23
aI147
aI1603
aI2083
aI4645
asS'd'
I1030
sa(dp2060
S'a'
I3583
sS's'
I13
sS'm'
(lp2061
I1
aI1
aI3
aI15
aI27
aI53
aI11
aI155
aI157
aI629
aI259
aI3009
aI4605
asS'd'
I1031
sa(dp2062
S'a'
I3590
sS's'
I13
sS'm'
(lp2063
I1
aI3
aI1
aI7
aI15
aI47
aI51
aI1
aI259
aI603
aI887
aI2833
aI6581
asS'd'
I1032
sa(dp2064
S'a'
I3593
sS's'
I13
sS'm'
(lp2065
I1
aI3
aI5
aI3
aI1
aI47
aI91
aI43
aI361
aI571
aI29
aI1453
aI4269
asS'd'
I1033
sa(dp2066
S'a'
I3594
sS's'
I13
sS'm'
(lp2067
I1
aI1
aI3
aI9
aI11
aI51
aI55
aI23
aI415
aI277
aI1423
aI3475
aI1527
asS'd'
I1034
sa(dp2068
S'a'
I3599
sS's'
I13
sS'm'
(lp2069
I1
aI1
aI3
aI11
aI29
aI49
aI101
aI75
aI299
aI709
aI805
aI4037
aI4389
asS'd'
I1035
sa(dp2070
S'a'
I3601
sS's'
I13
sS'm'
(lp2071
I1
aI1
aI7
aI3
aI23
aI1
aI37
aI51
aI379
aI771
aI1301
aI3717
aI6673
asS'd'
I1036
sa(dp2072
S'a'
I3602
sS's'
I13
sS'm'
(lp2073
I1
aI1
aI5
aI3
aI23
aI11
aI125
aI177
aI375
aI665
aI951
aI1577
aI2603
asS'd'
I1037
sa(dp2074
S'a'
I3613
sS's'
I13
sS'm'
(lp2075
I1
aI1
aI1
aI1
aI1
aI5
aI71
aI255
aI21
aI459
aI467
aI2083
aI5415
asS'd'
I1038
sa(dp2076
S'a'
I3623
sS's'
I13
sS'm'
(lp2077
I1
aI1
aI5
aI13
aI23
aI29
aI109
aI157
aI363
aI971
aI549
aI647
aI1177
asS'd'
I1039
sa(dp2078
S'a'
I3630
sS's'
I13
sS'm'
(lp2079
I1
aI1
aI3
aI9
aI7
aI15
aI101
aI3
aI365
aI213
aI745
aI1155
aI6099
asS'd'
I1040
sa(dp2080
S'a'
I3638
sS's'
I13
sS'm'
(lp2081
I1
aI3
aI5
aI15
aI15
aI19
aI47
aI179
aI303
aI521
aI1279
aI219
aI2415
asS'd'
I1041
sa(dp2082
S'a'
I3649
sS's'
I13
sS'm'
(lp2083
I1
aI3
aI3
aI13
aI27
aI11
aI83
aI165
aI369
aI989
aI261
aI3933
aI4809
asS'd'
I1042
sa(dp2084
S'a'
I3655
sS's'
I13
sS'm'
(lp2085
I1
aI1
aI3
aI11
aI31
aI59
aI1
aI185
aI53
aI703
aI1471
aI2935
aI1107
asS'd'
I1043
sa(dp2086
S'a'
I3662
sS's'
I13
sS'm'
(lp2087
I1
aI3
aI3
aI7
aI25
aI3
aI81
aI27
aI93
aI521
aI433
aI2859
aI5861
asS'd'
I1044
sa(dp2088
S'a'
I3667
sS's'
I13
sS'm'
(lp2089
I1
aI3
aI3
aI11
aI29
aI15
aI49
aI167
aI315
aI927
aI543
aI3473
aI4307
asS'd'
I1045
sa(dp2090
S'a'
I3669
sS's'
I13
sS'm'
(lp2091
I1
aI3
aI1
aI3
aI29
aI33
aI53
aI15
aI183
aI691
aI703
aI1311
aI3393
asS'd'
I1046
sa(dp2092
S'a'
I3676
sS's'
I13
sS'm'
(lp2093
I1
aI3
aI5
aI13
aI23
aI49
aI3
aI11
aI1
aI357
aI1407
aI415
aI7211
asS'd'
I1047
sa(dp2094
S'a'
I3683
sS's'
I13
sS'm'
(lp2095
I1
aI3
aI7
aI15
aI1
aI25
aI91
aI113
aI323
aI371
aI189
aI925
aI1181
asS'd'
I1048
sa(dp2096
S'a'
I3700
sS's'
I13
sS'm'
(lp2097
I1
aI3
aI3
aI3
aI17
aI59
aI119
aI199
aI115
aI223
aI877
aI2193
aI193
asS'd'
I1049
sa(dp2098
S'a'
I3709
sS's'
I13
sS'm'
(lp2099
I1
aI1
aI1
aI5
aI5
aI35
aI31
aI59
aI437
aI411
aI37
aI2405
aI3797
asS'd'
I1050
sa(dp2100
S'a'
I3710
sS's'
I13
sS'm'
(lp2101
I1
aI3
aI1
aI13
aI9
aI37
aI1
aI241
aI59
aI157
aI1785
aI1223
aI563
asS'd'
I1051
sa(dp2102
S'a'
I3713
sS's'
I13
sS'm'
(lp2103
I1
aI3
aI5
aI13
aI3
aI21
aI25
aI95
aI15
aI745
aI85
aI701
aI5361
asS'd'
I1052
sa(dp2104
S'a'
I3723
sS's'
I13
sS'm'
(lp2105
I1
aI3
aI7
aI1
aI31
aI33
aI111
aI195
aI35
aI913
aI2013
aI2951
aI6611
asS'd'
I1053
sa(dp2106
S'a'
I3725
sS's'
I13
sS'm'
(lp2107
I1
aI3
aI5
aI1
aI19
aI3
aI75
aI119
aI111
aI409
aI951
aI1457
aI4957
asS'd'
I1054
sa(dp2108
S'a'
I3728
sS's'
I13
sS'm'
(lp2109
I1
aI3
aI1
aI15
aI19
aI59
aI3
aI155
aI237
aI657
aI1967
aI3323
aI6235
asS'd'
I1055
sa(dp2110
S'a'
I3734
sS's'
I13
sS'm'
(lp2111
I1
aI1
aI5
aI1
aI3
aI19
aI45
aI105
aI377
aI881
aI167
aI2255
aI4483
asS'd'
I1056
sa(dp2112
S'a'
I3737
sS's'
I13
sS'm'
(lp2113
I1
aI1
aI7
aI7
aI13
aI13
aI99
aI89
aI201
aI279
aI161
aI2483
aI6001
asS'd'
I1057
sa(dp2114
S'a'
I3738
sS's'
I13
sS'm'
(lp2115
I1
aI1
aI7
aI3
aI13
aI17
aI97
aI129
aI137
aI377
aI1519
aI183
aI3725
asS'd'
I1058
sa(dp2116
S'a'
I3744
sS's'
I13
sS'm'
(lp2117
I1
aI1
aI7
aI9
aI9
aI5
aI45
aI135
aI115
aI181
aI1685
aI3505
aI4387
asS'd'
I1059
sa(dp2118
S'a'
I3750
sS's'
I13
sS'm'
(lp2119
I1
aI1
aI1
aI1
aI19
aI35
aI69
aI113
aI305
aI419
aI949
aI2969
aI247
asS'd'
I1060
sa(dp2120
S'a'
I3762
sS's'
I13
sS'm'
(lp2121
I1
aI1
aI5
aI13
aI23
aI61
aI13
aI139
aI501
aI811
aI67
aI1501
aI6493
asS'd'
I1061
sa(dp2122
S'a'
I3764
sS's'
I13
sS'm'
(lp2123
I1
aI1
aI3
aI13
aI15
aI41
aI27
aI217
aI293
aI13
aI145
aI2631
aI6991
asS'd'
I1062
sa(dp2124
S'a'
I3774
sS's'
I13
sS'm'
(lp2125
I1
aI3
aI3
aI13
aI15
aI37
aI71
aI123
aI285
aI49
aI627
aI1283
aI5993
asS'd'
I1063
sa(dp2126
S'a'
I3776
sS's'
I13
sS'm'
(lp2127
I1
aI3
aI3
aI11
aI9
aI25
aI11
aI1
aI203
aI353
aI1347
aI1999
aI2799
asS'd'
I1064
sa(dp2128
S'a'
I3786
sS's'
I13
sS'm'
(lp2129
I1
aI3
aI5
aI1
aI7
aI49
aI101
aI231
aI499
aI63
aI1977
aI2207
aI7829
asS'd'
I1065
sa(dp2130
S'a'
I3800
sS's'
I13
sS'm'
(lp2131
I1
aI1
aI7
aI1
aI17
aI15
aI115
aI139
aI381
aI943
aI623
aI4037
aI2971
asS'd'
I1066
sa(dp2132
S'a'
I3803
sS's'
I13
sS'm'
(lp2133
I1
aI1
aI3
aI5
aI13
aI55
aI23
aI87
aI139
aI795
aI1669
aI1375
aI1185
asS'd'
I1067
sa(dp2134
S'a'
I3809
sS's'
I13
sS'm'
(lp2135
I1
aI3
aI3
aI5
aI5
aI45
aI97
aI253
aI241
aI333
aI645
aI555
aI7867
asS'd'
I1068
sa(dp2136
S'a'
I3816
sS's'
I13
sS'm'
(lp2137
I1
aI3
aI5
aI1
aI1
aI1
aI89
aI27
aI407
aI509
aI1433
aI609
aI2355
asS'd'
I1069
sa(dp2138
S'a'
I3821
sS's'
I13
sS'm'
(lp2139
I1
aI3
aI7
aI1
aI27
aI29
aI5
aI157
aI495
aI811
aI1293
aI1143
aI827
asS'd'
I1070
sa(dp2140
S'a'
I3827
sS's'
I13
sS'm'
(lp2141
I1
aI1
aI3
aI3
aI25
aI49
aI127
aI111
aI191
aI3
aI845
aI1383
aI2521
asS'd'
I1071
sa(dp2142
S'a'
I3829
sS's'
I13
sS'm'
(lp2143
I1
aI1
aI5
aI7
aI5
aI51
aI101
aI155
aI237
aI461
aI831
aI3091
aI3851
asS'd'
I1072
sa(dp2144
S'a'
I3836
sS's'
I13
sS'm'
(lp2145
I1
aI3
aI7
aI1
aI29
aI35
aI105
aI91
aI285
aI705
aI131
aI395
aI6011
asS'd'
I1073
sa(dp2146
S'a'
I3842
sS's'
I13
sS'm'
(lp2147
I1
aI3
aI5
aI3
aI13
aI21
aI83
aI173
aI221
aI827
aI1775
aI1931
aI6727
asS'd'
I1074
sa(dp2148
S'a'
I3844
sS's'
I13
sS'm'
(lp2149
I1
aI1
aI3
aI5
aI3
aI25
aI95
aI115
aI205
aI569
aI1447
aI933
aI6425
asS'd'
I1075
sa(dp2150
S'a'
I3847
sS's'
I13
sS'm'
(lp2151
I1
aI1
aI7
aI9
aI31
aI3
aI17
aI175
aI145
aI447
aI1321
aI1069
aI6527
asS'd'
I1076
sa(dp2152
S'a'
I3853
sS's'
I13
sS'm'
(lp2153
I1
aI1
aI3
aI3
aI23
aI1
aI79
aI51
aI421
aI419
aI873
aI3939
aI1801
asS'd'
I1077
sa(dp2154
S'a'
I3861
sS's'
I13
sS'm'
(lp2155
I1
aI1
aI5
aI1
aI3
aI39
aI15
aI85
aI169
aI669
aI919
aI397
aI5579
asS'd'
I1078
sa(dp2156
S'a'
I3871
sS's'
I13
sS'm'
(lp2157
I1
aI3
aI5
aI1
aI21
aI61
aI87
aI217
aI251
aI619
aI1091
aI4009
aI229
asS'd'
I1079
sa(dp2158
S'a'
I3872
sS's'
I13
sS'm'
(lp2159
I1
aI1
aI1
aI11
aI23
aI55
aI85
aI121
aI363
aI867
aI315
aI447
aI3373
asS'd'
I1080
sa(dp2160
S'a'
I3881
sS's'
I13
sS'm'
(lp2161
I1
aI3
aI3
aI13
aI29
aI19
aI89
aI85
aI137
aI469
aI1873
aI2765
aI3975
asS'd'
I1081
sa(dp2162
S'a'
I3890
sS's'
I13
sS'm'
(lp2163
I1
aI3
aI7
aI13
aI19
aI63
aI61
aI77
aI67
aI361
aI11
aI1787
aI4703
asS'd'
I1082
sa(dp2164
S'a'
I3892
sS's'
I13
sS'm'
(lp2165
I1
aI1
aI3
aI11
aI7
aI15
aI127
aI105
aI179
aI857
aI1671
aI3647
aI3389
asS'd'
I1083
sa(dp2166
S'a'
I3909
sS's'
I13
sS'm'
(lp2167
I1
aI1
aI1
aI7
aI19
aI21
aI99
aI161
aI499
aI519
aI1287
aI2973
aI479
asS'd'
I1084
sa(dp2168
S'a'
I3921
sS's'
I13
sS'm'
(lp2169
I1
aI1
aI3
aI13
aI29
aI51
aI95
aI251
aI55
aI519
aI1955
aI2881
aI5951
asS'd'
I1085
sa(dp2170
S'a'
I3934
sS's'
I13
sS'm'
(lp2171
I1
aI1
aI3
aI11
aI23
aI63
aI121
aI237
aI175
aI311
aI701
aI1539
aI2383
asS'd'
I1086
sa(dp2172
S'a'
I3938
sS's'
I13
sS'm'
(lp2173
I1
aI1
aI7
aI5
aI5
aI45
aI73
aI97
aI5
aI153
aI715
aI2037
aI3353
asS'd'
I1087
sa(dp2174
S'a'
I3947
sS's'
I13
sS'm'
(lp2175
I1
aI1
aI1
aI3
aI13
aI7
aI67
aI173
aI425
aI843
aI1497
aI2729
aI5193
asS'd'
I1088
sa(dp2176
S'a'
I3950
sS's'
I13
sS'm'
(lp2177
I1
aI1
aI7
aI1
aI23
aI3
aI119
aI11
aI77
aI141
aI1905
aI2269
aI4269
asS'd'
I1089
sa(dp2178
S'a'
I3952
sS's'
I13
sS'm'
(lp2179
I1
aI1
aI7
aI15
aI1
aI23
aI79
aI251
aI439
aI603
aI405
aI2449
aI6383
asS'd'
I1090
sa(dp2180
S'a'
I3964
sS's'
I13
sS'm'
(lp2181
I1
aI3
aI7
aI11
aI29
aI27
aI47
aI255
aI47
aI661
aI1967
aI1007
aI3689
asS'd'
I1091
sa(dp2182
S'a'
I3974
sS's'
I13
sS'm'
(lp2183
I1
aI3
aI7
aI5
aI19
aI39
aI35
aI115
aI417
aI373
aI291
aI329
aI603
asS'd'
I1092
sa(dp2184
S'a'
I3980
sS's'
I13
sS'm'
(lp2185
I1
aI3
aI1
aI9
aI11
aI33
aI27
aI193
aI207
aI423
aI1311
aI1369
aI7307
asS'd'
I1093
sa(dp2186
S'a'
I3983
sS's'
I13
sS'm'
(lp2187
I1
aI1
aI3
aI11
aI9
aI29
aI83
aI17
aI497
aI493
aI329
aI3141
aI5935
asS'd'
I1094
sa(dp2188
S'a'
I3986
sS's'
I13
sS'm'
(lp2189
I1
aI3
aI1
aI5
aI31
aI51
aI29
aI171
aI51
aI493
aI1621
aI3501
aI4091
asS'd'
I1095
sa(dp2190
S'a'
I3995
sS's'
I13
sS'm'
(lp2191
I1
aI1
aI5
aI9
aI21
aI43
aI105
aI207
aI245
aI363
aI1191
aI699
aI1139
asS'd'
I1096
sa(dp2192
S'a'
I3998
sS's'
I13
sS'm'
(lp2193
I1
aI1
aI3
aI11
aI19
aI5
aI81
aI119
aI247
aI169
aI1337
aI45
aI6565
asS'd'
I1097
sa(dp2194
S'a'
I4001
sS's'
I13
sS'm'
(lp2195
I1
aI3
aI1
aI11
aI3
aI51
aI3
aI101
aI159
aI11
aI253
aI299
aI5043
asS'd'
I1098
sa(dp2196
S'a'
I4002
sS's'
I13
sS'm'
(lp2197
I1
aI3
aI1
aI5
aI11
aI53
aI85
aI39
aI57
aI645
aI2007
aI1039
aI3627
asS'd'
I1099
sa(dp2198
S'a'
I4004
sS's'
I13
sS'm'
(lp2199
I1
aI3
aI5
aI3
aI17
aI61
aI97
aI165
aI415
aI357
aI283
aI601
aI5505
asS'd'
I1100
sa(dp2200
S'a'
I4008
sS's'
I13
sS'm'
(lp2201
I1
aI3
aI7
aI3
aI9
aI51
aI49
aI85
aI3
aI227
aI137
aI309
aI243
asS'd'
I1101
sa(dp2202
S'a'
I4011
sS's'
I13
sS'm'
(lp2203
I1
aI1
aI5
aI3
aI11
aI59
aI11
aI131
aI409
aI703
aI455
aI123
aI6727
asS'd'
I1102
sa(dp2204
S'a'
I4016
sS's'
I13
sS'm'
(lp2205
I1
aI3
aI7
aI9
aI25
aI49
aI21
aI171
aI287
aI379
aI667
aI313
aI713
asS'd'
I1103
sa(dp2206
S'a'
I4033
sS's'
I13
sS'm'
(lp2207
I1
aI1
aI3
aI9
aI7
aI35
aI47
aI3
aI367
aI581
aI1627
aI1665
aI3905
asS'd'
I1104
sa(dp2208
S'a'
I4036
sS's'
I13
sS'm'
(lp2209
I1
aI3
aI1
aI1
aI29
aI57
aI35
aI55
aI255
aI653
aI823
aI2197
aI6179
asS'd'
I1105
sa(dp2210
S'a'
I4040
sS's'
I13
sS'm'
(lp2211
I1
aI3
aI7
aI15
aI17
aI15
aI117
aI83
aI359
aI163
aI115
aI2999
aI5373
asS'd'
I1106
sa(dp2212
S'a'
I4053
sS's'
I13
sS'm'
(lp2213
I1
aI1
aI5
aI3
aI21
aI61
aI35
aI97
aI71
aI687
aI207
aI2917
aI1049
asS'd'
I1107
sa(dp2214
S'a'
I4058
sS's'
I13
sS'm'
(lp2215
I1
aI1
aI1
aI15
aI13
aI15
aI125
aI81
aI263
aI661
aI417
aI3243
aI1669
asS'd'
I1108
sa(dp2216
S'a'
I4081
sS's'
I13
sS'm'
(lp2217
I1
aI1
aI7
aI3
aI3
aI19
aI111
aI193
aI443
aI339
aI659
aI1211
aI1557
asS'd'
I1109
sa(dp2218
S'a'
I4091
sS's'
I13
sS'm'
(lp2219
I1
aI3
aI1
aI3
aI27
aI3
aI3
aI173
aI391
aI213
aI803
aI3281
aI3207
asS'd'
I1110
sa(dp2220
S'a'
I4094
sS's'
I13
sS'm'
(lp2221
I1
aI1
aI5
aI15
aI19
aI1
aI7
aI211
aI157
aI603
aI403
aI1387
aI1583
asS'd'
I1111
sa(dp2222
S'a'
I21
sS's'
I14
sS'm'
(lp2223
I1
aI3
aI5
aI13
aI17
aI53
aI125
aI13
aI339
aI723
aI521
aI413
aI5801
aI10451
asS'd'
I1112
sa(dp2224
S'a'
I28
sS's'
I14
sS'm'
(lp2225
I1
aI1
aI3
aI13
aI29
aI9
aI99
aI77
aI141
aI609
aI1533
aI983
aI2039
aI51
asS'd'
I1113
sa(dp2226
S'a'
I41
sS's'
I14
sS'm'
(lp2227
I1
aI1
aI3
aI11
aI21
aI55
aI5
aI51
aI423
aI309
aI525
aI3715
aI3025
aI15055
asS'd'
I1114
sa(dp2228
S'a'
I47
sS's'
I14
sS'm'
(lp2229
I1
aI1
aI3
aI7
aI9
aI21
aI77
aI171
aI239
aI341
aI1653
aI1093
aI2273
aI10723
asS'd'
I1115
sa(dp2230
S'a'
I61
sS's'
I14
sS'm'
(lp2231
I1
aI1
aI1
aI15
aI31
aI15
aI23
aI35
aI317
aI869
aI1917
aI1415
aI4313
aI3687
asS'd'
I1116
sa(dp2232
S'a'
I84
sS's'
I14
sS'm'
(lp2233
I1
aI1
aI1
aI5
aI21
aI25
aI99
aI167
aI439
aI453
aI473
aI431
aI6665
aI4989
asS'd'
I1117
sa(dp2234
S'a'
I87
sS's'
I14
sS'm'
(lp2235
I1
aI1
aI7
aI9
aI31
aI47
aI81
aI83
aI345
aI43
aI1363
aI1885
aI3155
aI3185
asS'd'
I1118
sa(dp2236
S'a'
I93
sS's'
I14
sS'm'
(lp2237
I1
aI3
aI7
aI1
aI31
aI17
aI61
aI185
aI341
aI129
aI547
aI717
aI2071
aI9991
asS'd'
I1119
sa(dp2238
S'a'
I94
sS's'
I14
sS'm'
(lp2239
I1
aI3
aI1
aI13
aI23
aI61
aI77
aI217
aI455
aI77
aI1263
aI1601
aI3501
aI14953
asS'd'
I1120
sa(dp2240
S'a'
I103
sS's'
I14
sS'm'
(lp2241
I1
aI1
aI7
aI7
aI19
aI19
aI1
aI229
aI431
aI943
aI1069
aI1949
aI1289
aI15729
asS'd'
I1121
sa(dp2242
S'a'
I117
sS's'
I14
sS'm'
(lp2243
I1
aI1
aI3
aI5
aI1
aI35
aI97
aI251
aI487
aI459
aI1265
aI1739
aI165
aI10365
asS'd'
I1122
sa(dp2244
S'a'
I121
sS's'
I14
sS'm'
(lp2245
I1
aI3
aI5
aI3
aI11
aI25
aI79
aI175
aI383
aI545
aI187
aI197
aI4329
aI3363
asS'd'
I1123
sa(dp2246
S'a'
I134
sS's'
I14
sS'm'
(lp2247
I1
aI1
aI3
aI3
aI29
aI9
aI63
aI55
aI175
aI277
aI431
aI2549
aI2629
aI6409
asS'd'
I1124
sa(dp2248
S'a'
I137
sS's'
I14
sS'm'
(lp2249
I1
aI1
aI3
aI15
aI17
aI21
aI79
aI139
aI99
aI135
aI1763
aI1805
aI3471
aI5439
asS'd'
I1125
sa(dp2250
S'a'
I157
sS's'
I14
sS'm'
(lp2251
I1
aI1
aI3
aI9
aI9
aI15
aI35
aI119
aI289
aI835
aI769
aI3843
aI4119
aI4421
asS'd'
I1126
sa(dp2252
S'a'
I161
sS's'
I14
sS'm'
(lp2253
I1
aI1
aI1
aI5
aI19
aI19
aI67
aI199
aI307
aI815
aI1367
aI1231
aI3927
aI6593
asS'd'
I1127
sa(dp2254
S'a'
I205
sS's'
I14
sS'm'
(lp2255
I1
aI1
aI3
aI1
aI29
aI51
aI121
aI209
aI431
aI47
aI1115
aI907
aI2535
aI9755
asS'd'
I1128
sa(dp2256
S'a'
I206
sS's'
I14
sS'm'
(lp2257
I1
aI1
aI3
aI5
aI17
aI1
aI5
aI119
aI121
aI223
aI1719
aI1291
aI3947
aI15891
asS'd'
I1129
sa(dp2258
S'a'
I211
sS's'
I14
sS'm'
(lp2259
I1
aI3
aI1
aI15
aI29
aI25
aI3
aI131
aI373
aI307
aI645
aI3513
aI1289
aI1987
asS'd'
I1130
sa(dp2260
S'a'
I214
sS's'
I14
sS'm'
(lp2261
I1
aI3
aI3
aI11
aI29
aI45
aI105
aI179
aI331
aI465
aI891
aI1315
aI403
aI3057
asS'd'
I1131
sa(dp2262
S'a'
I218
sS's'
I14
sS'm'
(lp2263
I1
aI1
aI5
aI13
aI17
aI59
aI77
aI127
aI485
aI855
aI1147
aI3093
aI891
aI9869
asS'd'
I1132
sa(dp2264
S'a'
I234
sS's'
I14
sS'm'
(lp2265
I1
aI1
aI1
aI7
aI23
aI27
aI31
aI203
aI285
aI463
aI827
aI685
aI1349
aI15051
asS'd'
I1133
sa(dp2266
S'a'
I236
sS's'
I14
sS'm'
(lp2267
I1
aI1
aI1
aI5
aI29
aI5
aI107
aI195
aI31
aI425
aI19
aI2865
aI3869
aI11153
asS'd'
I1134
sa(dp2268
S'a'
I248
sS's'
I14
sS'm'
(lp2269
I1
aI1
aI7
aI5
aI7
aI47
aI1
aI73
aI307
aI347
aI393
aI2205
aI7709
aI15121
asS'd'
I1135
sa(dp2270
S'a'
I262
sS's'
I14
sS'm'
(lp2271
I1
aI1
aI1
aI13
aI15
aI61
aI25
aI131
aI113
aI369
aI1995
aI2527
aI4475
aI1745
asS'd'
I1136
sa(dp2272
S'a'
I299
sS's'
I14
sS'm'
(lp2273
I1
aI1
aI1
aI1
aI31
aI63
aI21
aI253
aI307
aI281
aI859
aI3319
aI6721
aI2891
asS'd'
I1137
sa(dp2274
S'a'
I304
sS's'
I14
sS'm'
(lp2275
I1
aI1
aI3
aI11
aI1
aI17
aI5
aI183
aI301
aI979
aI651
aI1685
aI6343
aI10067
asS'd'
I1138
sa(dp2276
S'a'
I319
sS's'
I14
sS'm'
(lp2277
I1
aI1
aI5
aI15
aI23
aI45
aI99
aI145
aI263
aI507
aI1381
aI3425
aI2215
aI1815
asS'd'
I1139
sa(dp2278
S'a'
I322
sS's'
I14
sS'm'
(lp2279
I1
aI3
aI1
aI5
aI11
aI63
aI85
aI203
aI411
aI881
aI1369
aI1237
aI4657
aI6541
asS'd'
I1140
sa(dp2280
S'a'
I334
sS's'
I14
sS'm'
(lp2281
I1
aI3
aI3
aI13
aI17
aI53
aI121
aI201
aI269
aI983
aI215
aI3187
aI7121
aI6111
asS'd'
I1141
sa(dp2282
S'a'
I355
sS's'
I14
sS'm'
(lp2283
I1
aI3
aI5
aI15
aI15
aI5
aI13
aI143
aI3
aI313
aI1677
aI1093
aI3295
aI3387
asS'd'
I1142
sa(dp2284
S'a'
I357
sS's'
I14
sS'm'
(lp2285
I1
aI1
aI3
aI13
aI3
aI23
aI73
aI17
aI257
aI965
aI239
aI1271
aI2803
aI7327
asS'd'
I1143
sa(dp2286
S'a'
I358
sS's'
I14
sS'm'
(lp2287
I1
aI3
aI5
aI13
aI9
aI57
aI115
aI37
aI41
aI467
aI135
aI1403
aI3811
aI4741
asS'd'
I1144
sa(dp2288
S'a'
I369
sS's'
I14
sS'm'
(lp2289
I1
aI3
aI7
aI15
aI9
aI33
aI39
aI203
aI351
aI367
aI1355
aI1403
aI3685
aI4757
asS'd'
I1145
sa(dp2290
S'a'
I372
sS's'
I14
sS'm'
(lp2291
I1
aI3
aI5
aI11
aI31
aI3
aI113
aI123
aI203
aI421
aI1821
aI3151
aI2375
aI4419
asS'd'
I1146
sa(dp2292
S'a'
I375
sS's'
I14
sS'm'
(lp2293
I1
aI1
aI1
aI7
aI21
aI63
aI99
aI23
aI133
aI79
aI991
aI1755
aI4989
aI4709
asS'd'
I1147
sa(dp2294
S'a'
I388
sS's'
I14
sS'm'
(lp2295
I1
aI3
aI5
aI1
aI25
aI63
aI113
aI239
aI49
aI443
aI173
aI1261
aI3201
aI10599
asS'd'
I1148
sa(dp2296
S'a'
I400
sS's'
I14
sS'm'
(lp2297
I1
aI3
aI3
aI13
aI3
aI25
aI101
aI169
aI23
aI585
aI327
aI1327
aI111
aI10059
asS'd'
I1149
sa(dp2298
S'a'
I415
sS's'
I14
sS'm'
(lp2299
I1
aI3
aI3
aI5
aI19
aI1
aI33
aI89
aI437
aI213
aI1933
aI1741
aI2603
aI5625
asS'd'
I1150
sa(dp2300
S'a'
I446
sS's'
I14
sS'm'
(lp2301
I1
aI3
aI1
aI3
aI15
aI15
aI25
aI139
aI73
aI335
aI237
aI2461
aI3101
aI14951
asS'd'
I1151
sa(dp2302
S'a'
I451
sS's'
I14
sS'm'
(lp2303
I1
aI3
aI5
aI1
aI31
aI15
aI31
aI187
aI177
aI659
aI1339
aI3767
aI4975
aI7123
asS'd'
I1152
sa(dp2304
S'a'
I458
sS's'
I14
sS'm'
(lp2305
I1
aI3
aI1
aI3
aI25
aI19
aI47
aI89
aI107
aI107
aI649
aI683
aI3123
aI11085
asS'd'
I1153
sa(dp2306
S'a'
I471
sS's'
I14
sS'm'
(lp2307
I1
aI3
aI7
aI9
aI15
aI21
aI101
aI25
aI11
aI625
aI1555
aI675
aI3893
aI5805
asS'd'
I1154
sa(dp2308
S'a'
I484
sS's'
I14
sS'm'
(lp2309
I1
aI1
aI1
aI5
aI7
aI49
aI123
aI21
aI439
aI369
aI207
aI535
aI4619
aI14665
asS'd'
I1155
sa(dp2310
S'a'
I501
sS's'
I14
sS'm'
(lp2311
I1
aI1
aI5
aI7
aI1
aI25
aI103
aI185
aI99
aI239
aI1093
aI1561
aI6177
aI4039
asS'd'
I1156
sa(dp2312
S'a'
I502
sS's'
I14
sS'm'
(lp2313
I1
aI3
aI7
aI5
aI29
aI21
aI43
aI103
aI343
aI973
aI1561
aI2975
aI7467
aI7947
asS'd'
I1157
sa(dp2314
S'a'
I517
sS's'
I14
sS'm'
(lp2315
I1
aI1
aI7
aI9
aI19
aI3
aI13
aI23
aI461
aI813
aI1191
aI985
aI559
aI3317
asS'd'
I1158
sa(dp2316
S'a'
I545
sS's'
I14
sS'm'
(lp2317
I1
aI3
aI5
aI5
aI27
aI31
aI79
aI15
aI365
aI901
aI1949
aI117
aI3619
aI13311
asS'd'
I1159
sa(dp2318
S'a'
I569
sS's'
I14
sS'm'
(lp2319
I1
aI3
aI5
aI7
aI5
aI33
aI67
aI199
aI425
aI189
aI1691
aI3099
aI815
aI1677
asS'd'
I1160
sa(dp2320
S'a'
I617
sS's'
I14
sS'm'
(lp2321
I1
aI1
aI7
aI11
aI13
aI29
aI73
aI137
aI265
aI601
aI445
aI3893
aI2511
aI8047
asS'd'
I1161
sa(dp2322
S'a'
I618
sS's'
I14
sS'm'
(lp2323
I1
aI1
aI3
aI1
aI13
aI5
aI57
aI101
aI357
aI391
aI335
aI601
aI1359
aI1065
asS'd'
I1162
sa(dp2324
S'a'
I623
sS's'
I14
sS'm'
(lp2325
I1
aI1
aI1
aI1
aI25
aI57
aI27
aI115
aI31
aI873
aI611
aI2125
aI447
aI13585
asS'd'
I1163
sa(dp2326
S'a'
I625
sS's'
I14
sS'm'
(lp2327
I1
aI3
aI3
aI13
aI27
aI17
aI73
aI11
aI359
aI33
aI1153
aI271
aI4537
aI15141
asS'd'
I1164
sa(dp2328
S'a'
I637
sS's'
I14
sS'm'
(lp2329
I1
aI3
aI7
aI3
aI11
aI63
aI103
aI61
aI59
aI629
aI1629
aI3279
aI3919
aI3177
asS'd'
I1165
sa(dp2330
S'a'
I661
sS's'
I14
sS'm'
(lp2331
I1
aI1
aI5
aI15
aI3
aI63
aI85
aI193
aI381
aI165
aI175
aI3247
aI2501
aI4209
asS'd'
I1166
sa(dp2332
S'a'
I668
sS's'
I14
sS'm'
(lp2333
I1
aI1
aI5
aI15
aI1
aI33
aI59
aI219
aI487
aI193
aI1557
aI703
aI2907
aI7953
asS'd'
I1167
sa(dp2334
S'a'
I684
sS's'
I14
sS'm'
(lp2335
I1
aI1
aI7
aI3
aI9
aI3
aI105
aI95
aI389
aI991
aI21
aI3841
aI6983
aI285
asS'd'
I1168
sa(dp2336
S'a'
I695
sS's'
I14
sS'm'
(lp2337
I1
aI1
aI1
aI1
aI1
aI31
aI25
aI137
aI117
aI67
aI1283
aI1963
aI6591
aI15541
asS'd'
I1169
sa(dp2338
S'a'
I716
sS's'
I14
sS'm'
(lp2339
I1
aI3
aI5
aI11
aI7
aI15
aI127
aI89
aI453
aI777
aI1827
aI2311
aI7661
aI11833
asS'd'
I1170
sa(dp2340
S'a'
I719
sS's'
I14
sS'm'
(lp2341
I1
aI1
aI7
aI13
aI19
aI29
aI79
aI165
aI223
aI453
aI2039
aI3961
aI6467
aI5481
asS'd'
I1171
sa(dp2342
S'a'
I722
sS's'
I14
sS'm'
(lp2343
I1
aI3
aI3
aI7
aI17
aI41
aI43
aI157
aI323
aI3
aI1001
aI2109
aI4513
aI12127
asS'd'
I1172
sa(dp2344
S'a'
I731
sS's'
I14
sS'm'
(lp2345
I1
aI1
aI5
aI9
aI31
aI57
aI3
aI217
aI113
aI271
aI1663
aI1367
aI6949
aI8165
asS'd'
I1173
sa(dp2346
S'a'
I738
sS's'
I14
sS'm'
(lp2347
I1
aI1
aI7
aI15
aI27
aI35
aI81
aI235
aI61
aI205
aI525
aI311
aI6357
aI2527
asS'd'
I1174
sa(dp2348
S'a'
I747
sS's'
I14
sS'm'
(lp2349
I1
aI3
aI1
aI9
aI19
aI29
aI71
aI207
aI321
aI1011
aI1615
aI1333
aI3459
aI6681
asS'd'
I1175
sa(dp2350
S'a'
I755
sS's'
I14
sS'm'
(lp2351
I1
aI3
aI7
aI7
aI3
aI57
aI41
aI19
aI25
aI397
aI565
aI1837
aI7625
aI11813
asS'd'
I1176
sa(dp2352
S'a'
I761
sS's'
I14
sS'm'
(lp2353
I1
aI3
aI3
aI1
aI27
aI47
aI31
aI79
aI441
aI961
aI1255
aI423
aI2405
aI913
asS'd'
I1177
sa(dp2354
S'a'
I767
sS's'
I14
sS'm'
(lp2355
I1
aI3
aI3
aI13
aI3
aI29
aI69
aI227
aI85
aI201
aI395
aI3199
aI3869
aI13099
asS'd'
I1178
sa(dp2356
S'a'
I775
sS's'
I14
sS'm'
(lp2357
I1
aI3
aI3
aI7
aI29
aI61
aI99
aI7
aI27
aI227
aI945
aI873
aI475
aI4363
asS'd'
I1179
sa(dp2358
S'a'
I782
sS's'
I14
sS'm'
(lp2359
I1
aI3
aI5
aI13
aI19
aI21
aI57
aI149
aI217
aI443
aI565
aI453
aI5487
aI10981
asS'd'
I1180
sa(dp2360
S'a'
I787
sS's'
I14
sS'm'
(lp2361
I1
aI3
aI3
aI1
aI9
aI27
aI47
aI191
aI35
aI395
aI1429
aI4079
aI6871
aI8013
asS'd'
I1181
sa(dp2362
S'a'
I794
sS's'
I14
sS'm'
(lp2363
I1
aI3
aI5
aI15
aI5
aI43
aI9
aI79
aI279
aI563
aI1125
aI985
aI8117
aI4099
asS'd'
I1182
sa(dp2364
S'a'
I803
sS's'
I14
sS'm'
(lp2365
I1
aI3
aI5
aI1
aI13
aI41
aI21
aI117
aI287
aI667
aI701
aI1483
aI8167
aI13283
asS'd'
I1183
sa(dp2366
S'a'
I812
sS's'
I14
sS'm'
(lp2367
I1
aI3
aI1
aI3
aI15
aI15
aI59
aI5
aI383
aI509
aI1657
aI3977
aI7697
aI10941
asS'd'
I1184
sa(dp2368
S'a'
I817
sS's'
I14
sS'm'
(lp2369
I1
aI3
aI1
aI1
aI17
aI29
aI19
aI23
aI377
aI45
aI981
aI1631
aI3557
aI6749
asS'd'
I1185
sa(dp2370
S'a'
I824
sS's'
I14
sS'm'
(lp2371
I1
aI3
aI3
aI9
aI9
aI51
aI9
aI193
aI345
aI361
aI1679
aI3333
aI713
aI5387
asS'd'
I1186
sa(dp2372
S'a'
I829
sS's'
I14
sS'm'
(lp2373
I1
aI3
aI5
aI5
aI17
aI45
aI97
aI17
aI385
aI349
aI105
aI2245
aI7295
aI14393
asS'd'
I1187
sa(dp2374
S'a'
I850
sS's'
I14
sS'm'
(lp2375
I1
aI3
aI7
aI3
aI19
aI51
aI35
aI99
aI79
aI301
aI1563
aI399
aI5879
aI14675
asS'd'
I1188
sa(dp2376
S'a'
I866
sS's'
I14
sS'm'
(lp2377
I1
aI1
aI7
aI15
aI13
aI53
aI55
aI203
aI417
aI161
aI2033
aI1845
aI6763
aI3585
asS'd'
I1189
sa(dp2378
S'a'
I871
sS's'
I14
sS'm'
(lp2379
I1
aI1
aI3
aI3
aI7
aI23
aI11
aI43
aI241
aI309
aI1453
aI3147
aI2619
aI3163
asS'd'
I1190
sa(dp2380
S'a'
I877
sS's'
I14
sS'm'
(lp2381
I1
aI1
aI1
aI11
aI17
aI1
aI17
aI137
aI443
aI465
aI993
aI3217
aI7879
aI14607
asS'd'
I1191
sa(dp2382
S'a'
I920
sS's'
I14
sS'm'
(lp2383
I1
aI1
aI7
aI13
aI29
aI49
aI71
aI217
aI291
aI793
aI135
aI21
aI2503
aI11091
asS'd'
I1192
sa(dp2384
S'a'
I935
sS's'
I14
sS'm'
(lp2385
I1
aI3
aI1
aI11
aI31
aI51
aI121
aI227
aI377
aI157
aI1457
aI1317
aI5625
aI6217
asS'd'
I1193
sa(dp2386
S'a'
I959
sS's'
I14
sS'm'
(lp2387
I1
aI1
aI3
aI7
aI23
aI61
aI47
aI93
aI79
aI617
aI1805
aI2403
aI5513
aI16335
asS'd'
I1194
sa(dp2388
S'a'
I979
sS's'
I14
sS'm'
(lp2389
I1
aI3
aI5
aI11
aI23
aI25
aI41
aI11
aI495
aI587
aI1223
aI3107
aI1469
aI15223
asS'd'
I1195
sa(dp2390
S'a'
I992
sS's'
I14
sS'm'
(lp2391
I1
aI3
aI7
aI7
aI9
aI1
aI1
aI49
aI23
aI723
aI1761
aI3717
aI7375
aI10875
asS'd'
I1196
sa(dp2392
S'a'
I1010
sS's'
I14
sS'm'
(lp2393
I1
aI3
aI3
aI11
aI25
aI37
aI57
aI63
aI309
aI603
aI183
aI285
aI1663
aI5627
asS'd'
I1197
sa(dp2394
S'a'
I1012
sS's'
I14
sS'm'
(lp2395
I1
aI3
aI7
aI11
aI19
aI25
aI25
aI201
aI391
aI257
aI529
aI1645
aI1
aI15111
asS'd'
I1198
sa(dp2396
S'a'
I1015
sS's'
I14
sS'm'
(lp2397
I1
aI3
aI3
aI9
aI11
aI43
aI91
aI65
aI5
aI959
aI301
aI1015
aI6343
aI3453
asS'd'
I1199
sa(dp2398
S'a'
I1033
sS's'
I14
sS'm'
(lp2399
I1
aI3
aI3
aI11
aI17
aI17
aI103
aI37
aI77
aI973
aI575
aI439
aI49
aI3639
asS'd'
I1200
sa(dp2400
S'a'
I1036
sS's'
I14
sS'm'
(lp2401
I1
aI1
aI5
aI7
aI1
aI15
aI107
aI237
aI231
aI967
aI923
aI1101
aI6715
aI1713
asS'd'
I1201
sa(dp2402
S'a'
I1053
sS's'
I14
sS'm'
(lp2403
I1
aI3
aI1
aI15
aI9
aI33
aI29
aI211
aI245
aI601
aI1783
aI887
aI1209
aI11785
asS'd'
I1202
sa(dp2404
S'a'
I1057
sS's'
I14
sS'm'
(lp2405
I1
aI3
aI3
aI7
aI21
aI43
aI27
aI89
aI27
aI141
aI865
aI367
aI1379
aI4063
asS'd'
I1203
sa(dp2406
S'a'
I1069
sS's'
I14
sS'm'
(lp2407
I1
aI3
aI7
aI7
aI15
aI17
aI15
aI15
aI131
aI649
aI1955
aI3289
aI3983
aI10689
asS'd'
I1204
sa(dp2408
S'a'
I1072
sS's'
I14
sS'm'
(lp2409
I1
aI3
aI1
aI5
aI17
aI7
aI125
aI69
aI359
aI981
aI1345
aI933
aI5281
aI7113
asS'd'
I1205
sa(dp2410
S'a'
I1075
sS's'
I14
sS'm'
(lp2411
I1
aI1
aI5
aI9
aI17
aI7
aI41
aI207
aI497
aI1015
aI493
aI891
aI3563
aI3541
asS'd'
I1206
sa(dp2412
S'a'
I1087
sS's'
I14
sS'm'
(lp2413
I1
aI3
aI5
aI11
aI27
aI3
aI47
aI31
aI303
aI1007
aI2047
aI2203
aI6257
aI8369
asS'd'
I1207
sa(dp2414
S'a'
I1089
sS's'
I14
sS'm'
(lp2415
I1
aI1
aI1
aI15
aI25
aI15
aI89
aI51
aI217
aI357
aI1133
aI1917
aI213
aI3365
asS'd'
I1208
sa(dp2416
S'a'
I1137
sS's'
I14
sS'm'
(lp2417
I1
aI1
aI5
aI13
aI29
aI23
aI123
aI207
aI429
aI805
aI819
aI2357
aI6313
aI11019
asS'd'
I1209
sa(dp2418
S'a'
I1166
sS's'
I14
sS'm'
(lp2419
I1
aI1
aI3
aI7
aI19
aI15
aI41
aI73
aI279
aI11
aI1089
aI3107
aI7737
aI15953
asS'd'
I1210
sa(dp2420
S'a'
I1174
sS's'
I14
sS'm'
(lp2421
I1
aI3
aI5
aI7
aI7
aI15
aI41
aI73
aI493
aI457
aI1731
aI1139
aI2513
aI12373
asS'd'
I1211
sa(dp2422
S'a'
I1180
sS's'
I14
sS'm'
(lp2423
I1
aI3
aI5
aI9
aI17
aI5
aI55
aI155
aI173
aI1005
aI529
aI3175
aI7667
aI4747
asS'd'
I1212
sa(dp2424
S'a'
I1204
sS's'
I14
sS'm'
(lp2425
I1
aI1
aI7
aI7
aI5
aI21
aI105
aI31
aI205
aI847
aI1033
aI3167
aI2347
aI8499
asS'd'
I1213
sa(dp2426
S'a'
I1211
sS's'
I14
sS'm'
(lp2427
I1
aI3
aI5
aI3
aI11
aI17
aI59
aI189
aI179
aI1007
aI33
aI3287
aI4813
aI8177
asS'd'
I1214
sa(dp2428
S'a'
I1219
sS's'
I14
sS'm'
(lp2429
I1
aI3
aI3
aI13
aI27
aI47
aI47
aI171
aI413
aI875
aI1081
aI1259
aI7139
aI8645
asS'd'
I1215
sa(dp2430
S'a'
I1236
sS's'
I14
sS'm'
(lp2431
I1
aI3
aI5
aI7
aI25
aI21
aI51
aI29
aI361
aI293
aI51
aI1119
aI1453
aI5283
asS'd'
I1216
sa(dp2432
S'a'
I1255
sS's'
I14
sS'm'
(lp2433
I1
aI3
aI7
aI7
aI29
aI55
aI103
aI199
aI511
aI341
aI1957
aI3987
aI2855
aI1279
asS'd'
I1217
sa(dp2434
S'a'
I1264
sS's'
I14
sS'm'
(lp2435
I1
aI1
aI1
aI9
aI23
aI51
aI61
aI63
aI391
aI37
aI55
aI3771
aI6517
aI15913
asS'd'
I1218
sa(dp2436
S'a'
I1306
sS's'
I14
sS'm'
(lp2437
I1
aI1
aI1
aI9
aI3
aI19
aI13
aI147
aI453
aI855
aI1321
aI189
aI5043
aI11215
asS'd'
I1219
sa(dp2438
S'a'
I1330
sS's'
I14
sS'm'
(lp2439
I1
aI3
aI3
aI13
aI23
aI3
aI87
aI155
aI401
aI981
aI607
aI3413
aI995
aI6473
asS'd'
I1220
sa(dp2440
S'a'
I1341
sS's'
I14
sS'm'
(lp2441
I1
aI3
aI1
aI9
aI29
aI47
aI95
aI123
aI421
aI353
aI1867
aI2609
aI2569
aI14083
asS'd'
I1221
sa(dp2442
S'a'
I1344
sS's'
I14
sS'm'
(lp2443
I1
aI1
aI5
aI13
aI25
aI39
aI29
aI111
aI125
aI545
aI1493
aI2371
aI6361
aI6307
asS'd'
I1222
sa(dp2444
S'a'
I1347
sS's'
I14
sS'm'
(lp2445
I1
aI3
aI3
aI11
aI13
aI31
aI87
aI75
aI27
aI393
aI921
aI3655
aI3343
aI16349
asS'd'
I1223
sa(dp2446
S'a'
I1349
sS's'
I14
sS'm'
(lp2447
I1
aI1
aI5
aI9
aI19
aI19
aI7
aI129
aI223
aI715
aI433
aI1627
aI4463
aI2951
asS'd'
I1224
sa(dp2448
S'a'
I1361
sS's'
I14
sS'm'
(lp2449
I1
aI1
aI7
aI1
aI31
aI13
aI49
aI33
aI89
aI43
aI1529
aI725
aI3809
aI3427
asS'd'
I1225
sa(dp2450
S'a'
I1380
sS's'
I14
sS'm'
(lp2451
I1
aI1
aI7
aI3
aI1
aI27
aI45
aI9
aI309
aI875
aI659
aI2661
aI553
aI7069
asS'd'
I1226
sa(dp2452
S'a'
I1390
sS's'
I14
sS'm'
(lp2453
I1
aI1
aI7
aI15
aI13
aI37
aI61
aI19
aI125
aI683
aI1227
aI2255
aI1455
aI9339
asS'd'
I1227
sa(dp2454
S'a'
I1404
sS's'
I14
sS'm'
(lp2455
I1
aI3
aI5
aI7
aI19
aI7
aI71
aI21
aI465
aI645
aI1885
aI873
aI7405
aI1913
asS'd'
I1228
sa(dp2456
S'a'
I1435
sS's'
I14
sS'm'
(lp2457
I1
aI3
aI1
aI11
aI11
aI35
aI79
aI61
aI79
aI57
aI1603
aI3719
aI6323
aI16371
asS'd'
I1229
sa(dp2458
S'a'
I1444
sS's'
I14
sS'm'
(lp2459
I1
aI1
aI7
aI1
aI29
aI57
aI85
aI21
aI205
aI37
aI2045
aI683
aI4901
aI8223
asS'd'
I1230
sa(dp2460
S'a'
I1453
sS's'
I14
sS'm'
(lp2461
I1
aI1
aI5
aI13
aI31
aI31
aI65
aI131
aI259
aI535
aI967
aI3943
aI2605
aI2089
asS'd'
I1231
sa(dp2462
S'a'
I1461
sS's'
I14
sS'm'
(lp2463
I1
aI1
aI7
aI9
aI27
aI61
aI39
aI243
aI207
aI41
aI1909
aI3279
aI1331
aI4635
asS'd'
I1232
sa(dp2464
S'a'
I1462
sS's'
I14
sS'm'
(lp2465
I1
aI3
aI3
aI5
aI11
aI63
aI105
aI19
aI169
aI95
aI773
aI3175
aI1869
aI1797
asS'd'
I1233
sa(dp2466
S'a'
I1465
sS's'
I14
sS'm'
(lp2467
I1
aI3
aI3
aI15
aI13
aI33
aI107
aI197
aI153
aI795
aI1477
aI105
aI4965
aI991
asS'd'
I1234
sa(dp2468
S'a'
I1468
sS's'
I14
sS'm'
(lp2469
I1
aI3
aI7
aI11
aI11
aI37
aI23
aI149
aI197
aI3
aI1035
aI3857
aI553
aI1059
asS'd'
I1235
sa(dp2470
S'a'
I1474
sS's'
I14
sS'm'
(lp2471
I1
aI3
aI1
aI3
aI17
aI29
aI89
aI189
aI193
aI59
aI1477
aI3517
aI2565
aI7739
asS'd'
I1236
sa(dp2472
S'a'
I1483
sS's'
I14
sS'm'
(lp2473
I1
aI1
aI1
aI9
aI23
aI3
aI25
aI163
aI469
aI305
aI1791
aI3393
aI6141
aI8119
asS'd'
I1237
sa(dp2474
S'a'
I1488
sS's'
I14
sS'm'
(lp2475
I1
aI3
aI5
aI7
aI7
aI41
aI19
aI101
aI179
aI487
aI1071
aI2761
aI8043
aI5103
asS'd'
I1238
sa(dp2476
S'a'
I1493
sS's'
I14
sS'm'
(lp2477
I1
aI1
aI7
aI9
aI1
aI21
aI101
aI103
aI349
aI85
aI1841
aI1033
aI4473
aI3563
asS'd'
I1239
sa(dp2478
S'a'
I1500
sS's'
I14
sS'm'
(lp2479
I1
aI1
aI3
aI13
aI23
aI61
aI39
aI27
aI479
aI13
aI45
aI1371
aI7897
aI10637
asS'd'
I1240
sa(dp2480
S'a'
I1509
sS's'
I14
sS'm'
(lp2481
I1
aI1
aI5
aI9
aI17
aI61
aI71
aI55
aI355
aI99
aI1695
aI3053
aI839
aI959
asS'd'
I1241
sa(dp2482
S'a'
I1510
sS's'
I14
sS'm'
(lp2483
I1
aI1
aI3
aI1
aI7
aI27
aI87
aI221
aI327
aI241
aI461
aI3177
aI5933
aI8299
asS'd'
I1242
sa(dp2484
S'a'
I1514
sS's'
I14
sS'm'
(lp2485
I1
aI3
aI7
aI9
aI5
aI41
aI111
aI245
aI447
aI263
aI1363
aI1767
aI6331
aI3355
asS'd'
I1243
sa(dp2486
S'a'
I1519
sS's'
I14
sS'm'
(lp2487
I1
aI3
aI3
aI13
aI15
aI11
aI15
aI169
aI429
aI149
aI1965
aI2477
aI7733
aI2499
asS'd'
I1244
sa(dp2488
S'a'
I1528
sS's'
I14
sS'm'
(lp2489
I1
aI1
aI5
aI15
aI15
aI47
aI25
aI33
aI469
aI701
aI773
aI2747
aI1533
aI14633
asS'd'
I1245
sa(dp2490
S'a'
I1533
sS's'
I14
sS'm'
(lp2491
I1
aI3
aI1
aI5
aI19
aI57
aI37
aI75
aI423
aI11
aI685
aI2487
aI1779
aI8797
asS'd'
I1246
sa(dp2492
S'a'
I1540
sS's'
I14
sS'm'
(lp2493
I1
aI3
aI1
aI5
aI19
aI41
aI67
aI99
aI333
aI991
aI953
aI3221
aI939
aI4197
asS'd'
I1247
sa(dp2494
S'a'
I1550
sS's'
I14
sS'm'
(lp2495
I1
aI3
aI1
aI15
aI11
aI39
aI25
aI1
aI159
aI679
aI465
aI1611
aI5799
aI2537
asS'd'
I1248
sa(dp2496
S'a'
I1567
sS's'
I14
sS'm'
(lp2497
I1
aI1
aI5
aI11
aI5
aI37
aI37
aI7
aI101
aI703
aI235
aI23
aI2209
aI12799
asS'd'
I1249
sa(dp2498
S'a'
I1571
sS's'
I14
sS'm'
(lp2499
I1
aI1
aI7
aI3
aI11
aI23
aI71
aI215
aI45
aI269
aI1539
aI3625
aI5773
aI6889
asS'd'
I1250
sa(dp2500
S'a'
I1573
sS's'
I14
sS'm'
(lp2501
I1
aI3
aI5
aI15
aI27
aI33
aI105
aI109
aI205
aI653
aI821
aI435
aI1087
aI2495
asS'd'
I1251
sa(dp2502
S'a'
I1578
sS's'
I14
sS'm'
(lp2503
I1
aI1
aI3
aI5
aI11
aI39
aI53
aI213
aI41
aI385
aI1425
aI25
aI5553
aI12523
asS'd'
I1252
sa(dp2504
S'a'
I1598
sS's'
I14
sS'm'
(lp2505
I1
aI3
aI5
aI15
aI29
aI49
aI13
aI253
aI505
aI407
aI985
aI2569
aI6727
aI4761
asS'd'
I1253
sa(dp2506
S'a'
I1606
sS's'
I14
sS'm'
(lp2507
I1
aI1
aI1
aI3
aI29
aI17
aI69
aI47
aI25
aI819
aI1145
aI2479
aI1183
aI3343
asS'd'
I1254
sa(dp2508
S'a'
I1618
sS's'
I14
sS'm'
(lp2509
I1
aI3
aI1
aI15
aI25
aI61
aI43
aI55
aI279
aI579
aI361
aI355
aI6101
aI3143
asS'd'
I1255
sa(dp2510
S'a'
I1630
sS's'
I14
sS'm'
(lp2511
I1
aI3
aI5
aI11
aI3
aI59
aI125
aI101
aI451
aI495
aI1711
aI3443
aI3625
aI15579
asS'd'
I1256
sa(dp2512
S'a'
I1634
sS's'
I14
sS'm'
(lp2513
I1
aI3
aI1
aI11
aI25
aI61
aI49
aI219
aI23
aI795
aI481
aI3609
aI3691
aI15419
asS'd'
I1257
sa(dp2514
S'a'
I1640
sS's'
I14
sS'm'
(lp2515
I1
aI3
aI7
aI5
aI9
aI59
aI49
aI233
aI345
aI143
aI181
aI3587
aI3041
aI1219
asS'd'
I1258
sa(dp2516
S'a'
I1643
sS's'
I14
sS'm'
(lp2517
I1
aI3
aI7
aI13
aI9
aI31
aI39
aI137
aI261
aI919
aI1367
aI3145
aI4659
aI5875
asS'd'
I1259
sa(dp2518
S'a'
I1654
sS's'
I14
sS'm'
(lp2519
I1
aI1
aI3
aI3
aI27
aI43
aI95
aI65
aI301
aI915
aI31
aI451
aI7743
aI7277
asS'd'
I1260
sa(dp2520
S'a'
I1679
sS's'
I14
sS'm'
(lp2521
I1
aI3
aI1
aI5
aI23
aI37
aI53
aI31
aI203
aI453
aI71
aI1585
aI6011
aI16369
asS'd'
I1261
sa(dp2522
S'a'
I1688
sS's'
I14
sS'm'
(lp2523
I1
aI1
aI5
aI1
aI15
aI47
aI91
aI227
aI297
aI45
aI1415
aI3647
aI7811
aI14015
asS'd'
I1262
sa(dp2524
S'a'
I1698
sS's'
I14
sS'm'
(lp2525
I1
aI1
aI1
aI1
aI29
aI27
aI93
aI121
aI169
aI69
aI1361
aI2907
aI1867
aI7017
asS'd'
I1263
sa(dp2526
S'a'
I1703
sS's'
I14
sS'm'
(lp2527
I1
aI3
aI1
aI7
aI23
aI53
aI77
aI41
aI25
aI873
aI1333
aI3889
aI3239
aI1771
asS'd'
I1264
sa(dp2528
S'a'
I1704
sS's'
I14
sS'm'
(lp2529
I1
aI1
aI1
aI7
aI31
aI27
aI87
aI81
aI167
aI343
aI1981
aI2499
aI7749
aI15747
asS'd'
I1265
sa(dp2530
S'a'
I1722
sS's'
I14
sS'm'
(lp2531
I1
aI3
aI5
aI13
aI1
aI17
aI97
aI37
aI81
aI645
aI1167
aI3547
aI7769
aI10731
asS'd'
I1266
sa(dp2532
S'a'
I1735
sS's'
I14
sS'm'
(lp2533
I1
aI1
aI7
aI5
aI9
aI17
aI31
aI55
aI151
aI463
aI1041
aI2303
aI4015
aI3737
asS'd'
I1267
sa(dp2534
S'a'
I1750
sS's'
I14
sS'm'
(lp2535
I1
aI1
aI3
aI11
aI31
aI9
aI81
aI213
aI95
aI215
aI2031
aI2129
aI4299
aI3021
asS'd'
I1268
sa(dp2536
S'a'
I1753
sS's'
I14
sS'm'
(lp2537
I1
aI1
aI1
aI3
aI25
aI25
aI115
aI229
aI101
aI441
aI783
aI1729
aI7905
aI2375
asS'd'
I1269
sa(dp2538
S'a'
I1760
sS's'
I14
sS'm'
(lp2539
I1
aI1
aI5
aI9
aI3
aI19
aI73
aI35
aI379
aI493
aI1333
aI1647
aI13
aI197
asS'd'
I1270
sa(dp2540
S'a'
I1789
sS's'
I14
sS'm'
(lp2541
I1
aI1
aI7
aI9
aI3
aI55
aI99
aI43
aI281
aI9
aI73
aI2477
aI8183
aI11055
asS'd'
I1271
sa(dp2542
S'a'
I1792
sS's'
I14
sS'm'
(lp2543
I1
aI3
aI7
aI13
aI25
aI19
aI27
aI195
aI469
aI175
aI355
aI1861
aI7255
aI15377
asS'd'
I1272
sa(dp2544
S'a'
I1802
sS's'
I14
sS'm'
(lp2545
I1
aI1
aI3
aI11
aI15
aI19
aI115
aI31
aI413
aI835
aI697
aI879
aI6515
aI13465
asS'd'
I1273
sa(dp2546
S'a'
I1819
sS's'
I14
sS'm'
(lp2547
I1
aI3
aI3
aI15
aI3
aI61
aI105
aI201
aI151
aI739
aI49
aI3963
aI2573
aI3303
asS'd'
I1274
sa(dp2548
S'a'
I1825
sS's'
I14
sS'm'
(lp2549
I1
aI3
aI5
aI7
aI23
aI5
aI11
aI215
aI19
aI591
aI509
aI2887
aI1631
aI4391
asS'd'
I1275
sa(dp2550
S'a'
I1828
sS's'
I14
sS'm'
(lp2551
I1
aI3
aI3
aI3
aI25
aI1
aI109
aI5
aI363
aI545
aI1745
aI503
aI827
aI4677
asS'd'
I1276
sa(dp2552
S'a'
I1832
sS's'
I14
sS'm'
(lp2553
I1
aI1
aI3
aI15
aI1
aI45
aI121
aI141
aI497
aI745
aI1825
aI2041
aI2561
aI8153
asS'd'
I1277
sa(dp2554
S'a'
I1845
sS's'
I14
sS'm'
(lp2555
I1
aI3
aI1
aI11
aI29
aI7
aI71
aI241
aI7
aI39
aI1379
aI2479
aI7483
aI7195
asS'd'
I1278
sa(dp2556
S'a'
I1846
sS's'
I14
sS'm'
(lp2557
I1
aI1
aI7
aI11
aI3
aI27
aI39
aI97
aI339
aI217
aI1409
aI1569
aI4761
aI1567
asS'd'
I1279
sa(dp2558
S'a'
I1857
sS's'
I14
sS'm'
(lp2559
I1
aI1
aI5
aI15
aI11
aI53
aI87
aI213
aI297
aI923
aI393
aI717
aI3297
aI16123
asS'd'
I1280
sa(dp2560
S'a'
I1869
sS's'
I14
sS'm'
(lp2561
I1
aI1
aI1
aI11
aI27
aI41
aI121
aI49
aI225
aI379
aI1305
aI319
aI2461
aI5445
asS'd'
I1281
sa(dp2562
S'a'
I1897
sS's'
I14
sS'm'
(lp2563
I1
aI1
aI5
aI5
aI25
aI3
aI121
aI23
aI47
aI843
aI1679
aI1427
aI6393
aI4199
asS'd'
I1282
sa(dp2564
S'a'
I1906
sS's'
I14
sS'm'
(lp2565
I1
aI1
aI5
aI13
aI17
aI3
aI17
aI25
aI161
aI487
aI121
aI361
aI1375
aI10745
asS'd'
I1283
sa(dp2566
S'a'
I1908
sS's'
I14
sS'm'
(lp2567
I1
aI1
aI7
aI3
aI3
aI37
aI7
aI245
aI107
aI107
aI745
aI2415
aI2131
aI11419
asS'd'
I1284
sa(dp2568
S'a'
I1911
sS's'
I14
sS'm'
(lp2569
I1
aI1
aI5
aI3
aI3
aI23
aI67
aI91
aI281
aI387
aI465
aI905
aI883
aI9775
asS'd'
I1285
sa(dp2570
S'a'
I1934
sS's'
I14
sS'm'
(lp2571
I1
aI3
aI7
aI15
aI25
aI55
aI123
aI49
aI23
aI983
aI1903
aI2589
aI2073
aI7823
asS'd'
I1286
sa(dp2572
S'a'
I1962
sS's'
I14
sS'm'
(lp2573
I1
aI1
aI5
aI11
aI25
aI17
aI63
aI229
aI267
aI175
aI1759
aI1947
aI479
aI11089
asS'd'
I1287
sa(dp2574
S'a'
I1967
sS's'
I14
sS'm'
(lp2575
I1
aI3
aI7
aI3
aI11
aI37
aI83
aI95
aI415
aI1003
aI1175
aI2361
aI2117
aI9809
asS'd'
I1288
sa(dp2576
S'a'
I1972
sS's'
I14
sS'm'
(lp2577
I1
aI3
aI1
aI9
aI5
aI39
aI51
aI129
aI249
aI161
aI1981
aI2755
aI8057
aI13641
asS'd'
I1289
sa(dp2578
S'a'
I1975
sS's'
I14
sS'm'
(lp2579
I1
aI1
aI7
aI1
aI15
aI47
aI9
aI197
aI199
aI549
aI1091
aI2853
aI2331
aI4535
asS'd'
I1290
sa(dp2580
S'a'
I1999
sS's'
I14
sS'm'
(lp2581
I1
aI3
aI3
aI13
aI15
aI21
aI23
aI111
aI463
aI719
aI1667
aI377
aI5039
aI10723
asS'd'
I1291
sa(dp2582
S'a'
I2004
sS's'
I14
sS'm'
(lp2583
I1
aI1
aI3
aI7
aI23
aI47
aI39
aI47
aI307
aI949
aI1651
aI2525
aI5835
aI1425
asS'd'
I1292
sa(dp2584
S'a'
I2011
sS's'
I14
sS'm'
(lp2585
I1
aI3
aI3
aI9
aI23
aI47
aI111
aI39
aI251
aI1001
aI179
aI3985
aI535
aI15435
asS'd'
I1293
sa(dp2586
S'a'
I2013
sS's'
I14
sS'm'
(lp2587
I1
aI1
aI3
aI13
aI5
aI45
aI51
aI123
aI205
aI651
aI1583
aI1691
aI1631
aI11975
asS'd'
I1294
sa(dp2588
S'a'
I2037
sS's'
I14
sS'm'
(lp2589
I1
aI1
aI7
aI9
aI1
aI29
aI59
aI27
aI389
aI497
aI1459
aI1633
aI521
aI14037
asS'd'
I1295
sa(dp2590
S'a'
I2051
sS's'
I14
sS'm'
(lp2591
I1
aI1
aI3
aI3
aI3
aI23
aI35
aI247
aI371
aI729
aI931
aI681
aI1777
aI8353
asS'd'
I1296
sa(dp2592
S'a'
I2063
sS's'
I14
sS'm'
(lp2593
I1
aI3
aI3
aI1
aI19
aI15
aI17
aI191
aI495
aI643
aI319
aI37
aI5691
aI7049
asS'd'
I1297
sa(dp2594
S'a'
I2066
sS's'
I14
sS'm'
(lp2595
I1
aI3
aI5
aI11
aI5
aI31
aI123
aI243
aI335
aI573
aI113
aI209
aI4825
aI7783
asS'd'
I1298
sa(dp2596
S'a'
I2094
sS's'
I14
sS'm'
(lp2597
I1
aI3
aI7
aI7
aI29
aI19
aI25
aI191
aI89
aI515
aI55
aI3013
aI4523
aI12913
asS'd'
I1299
sa(dp2598
S'a'
I2128
sS's'
I14
sS'm'
(lp2599
I1
aI1
aI3
aI3
aI15
aI3
aI35
aI37
aI339
aI7
aI697
aI359
aI4553
aI1431
asS'd'
I1300
sa(dp2600
S'a'
I2154
sS's'
I14
sS'm'
(lp2601
I1
aI3
aI1
aI1
aI9
aI15
aI33
aI77
aI161
aI13
aI255
aI1187
aI6587
aI11715
asS'd'
I1301
sa(dp2602
S'a'
I2164
sS's'
I14
sS'm'
(lp2603
I1
aI3
aI7
aI7
aI25
aI57
aI61
aI171
aI231
aI43
aI1219
aI903
aI5623
aI4781
asS'd'
I1302
sa(dp2604
S'a'
I2198
sS's'
I14
sS'm'
(lp2605
I1
aI1
aI5
aI15
aI29
aI47
aI117
aI23
aI213
aI907
aI1423
aI369
aI4529
aI9651
asS'd'
I1303
sa(dp2606
S'a'
I2217
sS's'
I14
sS'm'
(lp2607
I1
aI1
aI5
aI7
aI15
aI55
aI105
aI249
aI401
aI37
aI1885
aI3779
aI3441
aI9057
asS'd'
I1304
sa(dp2608
S'a'
I2220
sS's'
I14
sS'm'
(lp2609
I1
aI1
aI5
aI3
aI3
aI27
aI49
aI89
aI335
aI561
aI1235
aI3251
aI2731
aI12711
asS'd'
I1305
sa(dp2610
S'a'
I2223
sS's'
I14
sS'm'
(lp2611
I1
aI1
aI1
aI15
aI29
aI49
aI37
aI173
aI25
aI743
aI1321
aI821
aI5899
aI9213
asS'd'
I1306
sa(dp2612
S'a'
I2238
sS's'
I14
sS'm'
(lp2613
I1
aI1
aI7
aI3
aI1
aI41
aI61
aI209
aI275
aI925
aI521
aI3029
aI1569
aI9277
asS'd'
I1307
sa(dp2614
S'a'
I2245
sS's'
I14
sS'm'
(lp2615
I1
aI3
aI5
aI13
aI17
aI1
aI11
aI171
aI441
aI119
aI1589
aI299
aI157
aI11439
asS'd'
I1308
sa(dp2616
S'a'
I2252
sS's'
I14
sS'm'
(lp2617
I1
aI1
aI5
aI9
aI13
aI33
aI27
aI77
aI363
aI939
aI1103
aI2135
aI1759
aI5429
asS'd'
I1309
sa(dp2618
S'a'
I2258
sS's'
I14
sS'm'
(lp2619
I1
aI3
aI7
aI1
aI17
aI39
aI49
aI201
aI49
aI803
aI2003
aI1193
aI7415
aI13847
asS'd'
I1310
sa(dp2620
S'a'
I2264
sS's'
I14
sS'm'
(lp2621
I1
aI1
aI5
aI5
aI17
aI49
aI39
aI19
aI311
aI801
aI1441
aI3263
aI7973
aI14181
asS'd'
I1311
sa(dp2622
S'a'
I2280
sS's'
I14
sS'm'
(lp2623
I1
aI1
aI3
aI9
aI9
aI27
aI59
aI89
aI81
aI473
aI1369
aI3121
aI7929
aI10905
asS'd'
I1312
sa(dp2624
S'a'
I2285
sS's'
I14
sS'm'
(lp2625
I1
aI3
aI3
aI5
aI17
aI35
aI35
aI239
aI379
aI431
aI501
aI3561
aI2059
aI9679
asS'd'
I1313
sa(dp2626
S'a'
I2293
sS's'
I14
sS'm'
(lp2627
I1
aI3
aI5
aI15
aI25
aI29
aI113
aI179
aI269
aI891
aI301
aI2017
aI7513
aI9379
asS'd'
I1314
sa(dp2628
S'a'
I2294
sS's'
I14
sS'm'
(lp2629
I1
aI3
aI1
aI11
aI17
aI35
aI49
aI149
aI135
aI661
aI1691
aI3169
aI3765
aI9003
asS'd'
I1315
sa(dp2630
S'a'
I2298
sS's'
I14
sS'm'
(lp2631
I1
aI3
aI7
aI15
aI5
aI21
aI53
aI241
aI475
aI271
aI683
aI2351
aI2181
aI6333
asS'd'
I1316
sa(dp2632
S'a'
I2303
sS's'
I14
sS'm'
(lp2633
I1
aI1
aI7
aI13
aI25
aI33
aI71
aI153
aI221
aI507
aI2017
aI2401
aI7545
aI8489
asS'd'
I1317
sa(dp2634
S'a'
I2306
sS's'
I14
sS'm'
(lp2635
I1
aI1
aI7
aI5
aI1
aI49
aI87
aI1
aI179
aI331
aI1597
aI3713
aI809
aI11109
asS'd'
I1318
sa(dp2636
S'a'
I2311
sS's'
I14
sS'm'
(lp2637
I1
aI3
aI1
aI5
aI5
aI61
aI93
aI39
aI479
aI977
aI1099
aI1291
aI7049
aI2797
asS'd'
I1319
sa(dp2638
S'a'
I2326
sS's'
I14
sS'm'
(lp2639
I1
aI3
aI1
aI13
aI19
aI41
aI57
aI77
aI5
aI117
aI125
aI115
aI3969
aI1345
asS'd'
I1320
sa(dp2640
S'a'
I2354
sS's'
I14
sS'm'
(lp2641
I1
aI1
aI1
aI9
aI15
aI9
aI57
aI7
aI219
aI41
aI767
aI23
aI5771
aI14175
asS'd'
I1321
sa(dp2642
S'a'
I2373
sS's'
I14
sS'm'
(lp2643
I1
aI3
aI7
aI9
aI17
aI61
aI1
aI59
aI227
aI349
aI63
aI189
aI3871
aI7919
asS'd'
I1322
sa(dp2644
S'a'
I2380
sS's'
I14
sS'm'
(lp2645
I1
aI3
aI5
aI5
aI9
aI29
aI33
aI203
aI413
aI701
aI1129
aI2103
aI1889
aI8377
asS'd'
I1323
sa(dp2646
S'a'
I2385
sS's'
I14
sS'm'
(lp2647
I1
aI1
aI3
aI1
aI9
aI17
aI69
aI115
aI123
aI1001
aI1
aI2893
aI3957
aI8593
asS'd'
I1324
sa(dp2648
S'a'
I2392
sS's'
I14
sS'm'
(lp2649
I1
aI1
aI3
aI1
aI31
aI41
aI83
aI91
aI113
aI195
aI1121
aI2665
aI6815
aI1189
asS'd'
I1325
sa(dp2650
S'a'
I2401
sS's'
I14
sS'm'
(lp2651
I1
aI1
aI1
aI13
aI3
aI59
aI13
aI123
aI95
aI103
aI1689
aI2809
aI5049
aI4055
asS'd'
I1326
sa(dp2652
S'a'
I2402
sS's'
I14
sS'm'
(lp2653
I1
aI1
aI1
aI15
aI21
aI41
aI11
aI167
aI375
aI589
aI207
aI1631
aI1597
aI8091
asS'd'
I1327
sa(dp2654
S'a'
I2408
sS's'
I14
sS'm'
(lp2655
I1
aI3
aI5
aI5
aI1
aI33
aI57
aI89
aI157
aI921
aI1353
aI2777
aI461
aI14567
asS'd'
I1328
sa(dp2656
S'a'
I2419
sS's'
I14
sS'm'
(lp2657
I1
aI1
aI5
aI1
aI25
aI5
aI51
aI247
aI1
aI577
aI463
aI3741
aI303
aI16059
asS'd'
I1329
sa(dp2658
S'a'
I2450
sS's'
I14
sS'm'
(lp2659
I1
aI1
aI7
aI5
aI13
aI7
aI17
aI87
aI51
aI987
aI835
aI93
aI5203
aI3973
asS'd'
I1330
sa(dp2660
S'a'
I2452
sS's'
I14
sS'm'
(lp2661
I1
aI1
aI7
aI7
aI3
aI27
aI7
aI1
aI135
aI171
aI231
aI3349
aI4459
aI2925
asS'd'
I1331
sa(dp2662
S'a'
I2477
sS's'
I14
sS'm'
(lp2663
I1
aI1
aI5
aI5
aI9
aI51
aI71
aI153
aI115
aI315
aI265
aI2207
aI4127
aI12631
asS'd'
I1332
sa(dp2664
S'a'
I2509
sS's'
I14
sS'm'
(lp2665
I1
aI1
aI3
aI15
aI23
aI59
aI35
aI121
aI425
aI921
aI1255
aI2123
aI5811
aI15937
asS'd'
I1333
sa(dp2666
S'a'
I2510
sS's'
I14
sS'm'
(lp2667
I1
aI3
aI7
aI7
aI11
aI21
aI45
aI57
aI269
aI395
aI555
aI783
aI6677
aI2889
asS'd'
I1334
sa(dp2668
S'a'
I2515
sS's'
I14
sS'm'
(lp2669
I1
aI3
aI5
aI7
aI31
aI19
aI73
aI35
aI465
aI349
aI1429
aI863
aI4707
aI6121
asS'd'
I1335
sa(dp2670
S'a'
I2524
sS's'
I14
sS'm'
(lp2671
I1
aI3
aI3
aI9
aI25
aI27
aI119
aI159
aI195
aI949
aI19
aI73
aI4511
aI15711
asS'd'
I1336
sa(dp2672
S'a'
I2527
sS's'
I14
sS'm'
(lp2673
I1
aI3
aI3
aI7
aI9
aI59
aI47
aI57
aI91
aI749
aI1579
aI1297
aI2445
aI5167
asS'd'
I1337
sa(dp2674
S'a'
I2531
sS's'
I14
sS'm'
(lp2675
I1
aI3
aI3
aI3
aI31
aI57
aI19
aI203
aI61
aI927
aI1477
aI2863
aI1305
aI11673
asS'd'
I1338
sa(dp2676
S'a'
I2552
sS's'
I14
sS'm'
(lp2677
I1
aI3
aI7
aI11
aI29
aI13
aI3
aI111
aI351
aI79
aI1863
aI2213
aI3273
aI7049
asS'd'
I1339
sa(dp2678
S'a'
I2561
sS's'
I14
sS'm'
(lp2679
I1
aI3
aI3
aI9
aI7
aI23
aI47
aI237
aI121
aI877
aI441
aI119
aI2723
aI3989
asS'd'
I1340
sa(dp2680
S'a'
I2567
sS's'
I14
sS'm'
(lp2681
I1
aI3
aI3
aI11
aI17
aI23
aI63
aI177
aI231
aI363
aI1451
aI33
aI2169
aI7251
asS'd'
I1341
sa(dp2682
S'a'
I2571
sS's'
I14
sS'm'
(lp2683
I1
aI1
aI5
aI11
aI31
aI41
aI93
aI229
aI39
aI1009
aI1061
aI433
aI2393
aI15401
asS'd'
I1342
sa(dp2684
S'a'
I2586
sS's'
I14
sS'm'
(lp2685
I1
aI1
aI5
aI15
aI31
aI37
aI25
aI135
aI135
aI897
aI33
aI3713
aI7663
aI8079
asS'd'
I1343
sa(dp2686
S'a'
I2588
sS's'
I14
sS'm'
(lp2687
I1
aI1
aI5
aI7
aI17
aI49
aI43
aI89
aI411
aI731
aI1431
aI3893
aI1635
aI7063
asS'd'
I1344
sa(dp2688
S'a'
I2595
sS's'
I14
sS'm'
(lp2689
I1
aI1
aI1
aI13
aI29
aI27
aI5
aI77
aI283
aI913
aI789
aI817
aI3309
aI475
asS'd'
I1345
sa(dp2690
S'a'
I2607
sS's'
I14
sS'm'
(lp2691
I1
aI1
aI3
aI1
aI19
aI21
aI67
aI77
aI423
aI551
aI5
aI1057
aI5469
aI7859
asS'd'
I1346
sa(dp2692
S'a'
I2621
sS's'
I14
sS'm'
(lp2693
I1
aI1
aI5
aI1
aI1
aI21
aI99
aI237
aI215
aI759
aI1505
aI1983
aI1517
aI8923
asS'd'
I1347
sa(dp2694
S'a'
I2630
sS's'
I14
sS'm'
(lp2695
I1
aI3
aI5
aI7
aI19
aI61
aI73
aI215
aI165
aI127
aI205
aI259
aI7755
aI15395
asS'd'
I1348
sa(dp2696
S'a'
I2639
sS's'
I14
sS'm'
(lp2697
I1
aI1
aI5
aI9
aI15
aI23
aI17
aI111
aI471
aI751
aI1923
aI775
aI6901
aI13095
asS'd'
I1349
sa(dp2698
S'a'
I2653
sS's'
I14
sS'm'
(lp2699
I1
aI1
aI7
aI1
aI25
aI5
aI63
aI141
aI461
aI687
aI1589
aI1559
aI7719
aI11349
asS'd'
I1350
sa(dp2700
S'a'
I2670
sS's'
I14
sS'm'
(lp2701
I1
aI1
aI1
aI3
aI11
aI63
aI11
aI27
aI253
aI439
aI297
aI1315
aI829
aI3765
asS'd'
I1351
sa(dp2702
S'a'
I2672
sS's'
I14
sS'm'
(lp2703
I1
aI3
aI1
aI1
aI9
aI47
aI127
aI179
aI173
aI809
aI241
aI35
aI7355
aI5049
asS'd'
I1352
sa(dp2704
S'a'
I2700
sS's'
I14
sS'm'
(lp2705
I1
aI3
aI1
aI11
aI19
aI63
aI93
aI1
aI205
aI977
aI303
aI3409
aI6529
aI10927
asS'd'
I1353
sa(dp2706
S'a'
I2711
sS's'
I14
sS'm'
(lp2707
I1
aI3
aI7
aI9
aI31
aI63
aI41
aI79
aI477
aI91
aI1801
aI3487
aI6885
aI13341
asS'd'
I1354
sa(dp2708
S'a'
I2715
sS's'
I14
sS'm'
(lp2709
I1
aI1
aI3
aI7
aI15
aI59
aI9
aI101
aI459
aI247
aI549
aI2855
aI5765
aI7785
asS'd'
I1355
sa(dp2710
S'a'
I2748
sS's'
I14
sS'm'
(lp2711
I1
aI1
aI7
aI3
aI13
aI59
aI71
aI123
aI93
aI517
aI1453
aI2389
aI4429
aI5053
asS'd'
I1356
sa(dp2712
S'a'
I2751
sS's'
I14
sS'm'
(lp2713
I1
aI1
aI5
aI3
aI19
aI21
aI77
aI53
aI81
aI879
aI1653
aI1637
aI3667
aI2623
asS'd'
I1357
sa(dp2714
S'a'
I2753
sS's'
I14
sS'm'
(lp2715
I1
aI1
aI1
aI15
aI17
aI57
aI65
aI53
aI407
aI765
aI417
aI497
aI5009
aI2175
asS'd'
I1358
sa(dp2716
S'a'
I2754
sS's'
I14
sS'm'
(lp2717
I1
aI3
aI3
aI7
aI31
aI13
aI5
aI203
aI263
aI17
aI119
aI1607
aI6773
aI11195
asS'd'
I1359
sa(dp2718
S'a'
I2760
sS's'
I14
sS'm'
(lp2719
I1
aI3
aI3
aI13
aI19
aI13
aI13
aI147
aI93
aI735
aI689
aI781
aI655
aI6853
asS'd'
I1360
sa(dp2720
S'a'
I2774
sS's'
I14
sS'm'
(lp2721
I1
aI1
aI1
aI1
aI1
aI25
aI63
aI159
aI493
aI987
aI71
aI1249
aI5859
aI11717
asS'd'
I1361
sa(dp2722
S'a'
I2784
sS's'
I14
sS'm'
(lp2723
I1
aI1
aI1
aI15
aI13
aI23
aI61
aI61
aI5
aI947
aI1853
aI3331
aI467
aI8081
asS'd'
I1362
sa(dp2724
S'a'
I2793
sS's'
I14
sS'm'
(lp2725
I1
aI1
aI3
aI9
aI19
aI61
aI65
aI189
aI95
aI309
aI283
aI1725
aI5683
aI15463
asS'd'
I1363
sa(dp2726
S'a'
I2804
sS's'
I14
sS'm'
(lp2727
I1
aI1
aI7
aI5
aI9
aI33
aI35
aI75
aI475
aI831
aI1445
aI1485
aI5047
aI9631
asS'd'
I1364
sa(dp2728
S'a'
I2811
sS's'
I14
sS'm'
(lp2729
I1
aI1
aI3
aI15
aI11
aI23
aI59
aI87
aI433
aI221
aI685
aI3113
aI4095
aI13819
asS'd'
I1365
sa(dp2730
S'a'
I2822
sS's'
I14
sS'm'
(lp2731
I1
aI1
aI7
aI15
aI25
aI29
aI67
aI17
aI349
aI353
aI1321
aI563
aI57
aI533
asS'd'
I1366
sa(dp2732
S'a'
I2826
sS's'
I14
sS'm'
(lp2733
I1
aI3
aI3
aI3
aI5
aI43
aI109
aI217
aI15
aI185
aI1895
aI1015
aI1831
aI10623
asS'd'
I1367
sa(dp2734
S'a'
I2836
sS's'
I14
sS'm'
(lp2735
I1
aI1
aI7
aI1
aI1
aI47
aI81
aI185
aI59
aI691
aI191
aI3709
aI1535
aI13347
asS'd'
I1368
sa(dp2736
S'a'
I2839
sS's'
I14
sS'm'
(lp2737
I1
aI1
aI5
aI1
aI23
aI57
aI83
aI217
aI457
aI771
aI1877
aI2789
aI8143
aI4797
asS'd'
I1369
sa(dp2738
S'a'
I2840
sS's'
I14
sS'm'
(lp2739
I1
aI1
aI3
aI7
aI23
aI35
aI79
aI49
aI227
aI205
aI1523
aI3873
aI4843
aI10505
asS'd'
I1370
sa(dp2740
S'a'
I2893
sS's'
I14
sS'm'
(lp2741
I1
aI1
aI1
aI1
aI17
aI43
aI121
aI95
aI205
aI35
aI189
aI2061
aI1693
aI13273
asS'd'
I1371
sa(dp2742
S'a'
I2901
sS's'
I14
sS'm'
(lp2743
I1
aI1
aI1
aI15
aI31
aI49
aI83
aI249
aI433
aI497
aI1949
aI1845
aI5215
aI5971
asS'd'
I1372
sa(dp2744
S'a'
I2902
sS's'
I14
sS'm'
(lp2745
I1
aI3
aI1
aI1
aI21
aI53
aI73
aI211
aI265
aI929
aI923
aI279
aI3621
aI9469
asS'd'
I1373
sa(dp2746
S'a'
I2905
sS's'
I14
sS'm'
(lp2747
I1
aI3
aI7
aI7
aI1
aI57
aI13
aI45
aI467
aI705
aI371
aI1345
aI1647
aI3411
asS'd'
I1374
sa(dp2748
S'a'
I2912
sS's'
I14
sS'm'
(lp2749
I1
aI3
aI1
aI11
aI27
aI29
aI117
aI163
aI143
aI669
aI489
aI3913
aI7891
aI9031
asS'd'
I1375
sa(dp2750
S'a'
I2915
sS's'
I14
sS'm'
(lp2751
I1
aI3
aI7
aI15
aI27
aI15
aI77
aI217
aI107
aI839
aI1517
aI1543
aI357
aI10365
asS'd'
I1376
sa(dp2752
S'a'
I2918
sS's'
I14
sS'm'
(lp2753
I1
aI1
aI1
aI5
aI31
aI17
aI107
aI245
aI345
aI939
aI1453
aI3645
aI6865
aI16173
asS'd'
I1377
sa(dp2754
S'a'
I2939
sS's'
I14
sS'm'
(lp2755
I1
aI3
aI5
aI5
aI9
aI61
aI43
aI97
aI453
aI917
aI945
aI2143
aI5473
aI5611
asS'd'
I1378
sa(dp2756
S'a'
I2965
sS's'
I14
sS'm'
(lp2757
I1
aI1
aI5
aI11
aI3
aI33
aI71
aI97
aI137
aI549
aI1605
aI3839
aI4883
aI2677
asS'd'
I1379
sa(dp2758
S'a'
I2966
sS's'
I14
sS'm'
(lp2759
I1
aI3
aI1
aI11
aI29
aI23
aI85
aI47
aI225
aI633
aI1613
aI1297
aI1415
aI15813
asS'd'
I1380
sa(dp2760
S'a'
I2975
sS's'
I14
sS'm'
(lp2761
I1
aI1
aI3
aI3
aI9
aI19
aI57
aI107
aI79
aI449
aI1951
aI753
aI6317
aI10377
asS'd'
I1381
sa(dp2762
S'a'
I2988
sS's'
I14
sS'm'
(lp2763
I1
aI1
aI1
aI5
aI21
aI3
aI39
aI187
aI299
aI517
aI1313
aI741
aI7259
aI4197
asS'd'
I1382
sa(dp2764
S'a'
I2993
sS's'
I14
sS'm'
(lp2765
I1
aI1
aI5
aI13
aI1
aI39
aI39
aI41
aI381
aI123
aI1257
aI3185
aI493
aI3723
asS'd'
I1383
sa(dp2766
S'a'
I3006
sS's'
I14
sS'm'
(lp2767
I1
aI3
aI7
aI7
aI3
aI37
aI15
aI161
aI129
aI169
aI555
aI3605
aI4287
aI15831
asS'd'
I1384
sa(dp2768
S'a'
I3017
sS's'
I14
sS'm'
(lp2769
I1
aI3
aI7
aI15
aI15
aI23
aI81
aI163
aI257
aI791
aI505
aI1903
aI2703
aI11919
asS'd'
I1385
sa(dp2770
S'a'
I3031
sS's'
I14
sS'm'
(lp2771
I1
aI3
aI7
aI7
aI27
aI63
aI17
aI147
aI111
aI851
aI1533
aI1365
aI5359
aI3315
asS'd'
I1386
sa(dp2772
S'a'
I3038
sS's'
I14
sS'm'
(lp2773
I1
aI3
aI7
aI1
aI15
aI5
aI61
aI143
aI385
aI261
aI1019
aI1705
aI1737
aI14485
asS'd'
I1387
sa(dp2774
S'a'
I3041
sS's'
I14
sS'm'
(lp2775
I1
aI3
aI5
aI5
aI25
aI17
aI49
aI229
aI431
aI567
aI1613
aI3431
aI2139
aI2981
asS'd'
I1388
sa(dp2776
S'a'
I3042
sS's'
I14
sS'm'
(lp2777
I1
aI3
aI5
aI11
aI17
aI57
aI71
aI241
aI31
aI1007
aI1695
aI2965
aI149
aI14125
asS'd'
I1389
sa(dp2778
S'a'
I3051
sS's'
I14
sS'm'
(lp2779
I1
aI1
aI3
aI11
aI7
aI49
aI39
aI101
aI5
aI501
aI1491
aI3061
aI225
aI12255
asS'd'
I1390
sa(dp2780
S'a'
I3073
sS's'
I14
sS'm'
(lp2781
I1
aI3
aI5
aI7
aI17
aI35
aI37
aI97
aI415
aI15
aI1349
aI997
aI2949
aI4511
asS'd'
I1391
sa(dp2782
S'a'
I3088
sS's'
I14
sS'm'
(lp2783
I1
aI3
aI1
aI5
aI25
aI35
aI99
aI183
aI161
aI59
aI1363
aI515
aI3767
aI3641
asS'd'
I1392
sa(dp2784
S'a'
I3097
sS's'
I14
sS'm'
(lp2785
I1
aI1
aI7
aI15
aI7
aI15
aI127
aI137
aI281
aI67
aI139
aI2315
aI3517
aI13371
asS'd'
I1393
sa(dp2786
S'a'
I3098
sS's'
I14
sS'm'
(lp2787
I1
aI1
aI5
aI15
aI23
aI49
aI19
aI79
aI425
aI805
aI1035
aI429
aI7707
aI14195
asS'd'
I1394
sa(dp2788
S'a'
I3103
sS's'
I14
sS'm'
(lp2789
I1
aI3
aI5
aI3
aI21
aI25
aI123
aI11
aI425
aI475
aI961
aI2995
aI7405
aI5449
asS'd'
I1395
sa(dp2790
S'a'
I3104
sS's'
I14
sS'm'
(lp2791
I1
aI1
aI7
aI1
aI21
aI1
aI75
aI231
aI451
aI595
aI719
aI2369
aI5907
aI1227
asS'd'
I1396
sa(dp2792
S'a'
I3146
sS's'
I14
sS'm'
(lp2793
I1
aI1
aI1
aI9
aI21
aI57
aI45
aI255
aI19
aI79
aI481
aI3363
aI3451
aI8399
asS'd'
I1397
sa(dp2794
S'a'
I3148
sS's'
I14
sS'm'
(lp2795
I1
aI1
aI7
aI13
aI31
aI49
aI95
aI69
aI483
aI427
aI37
aI4047
aI7057
aI9111
asS'd'
I1398
sa(dp2796
S'a'
I3153
sS's'
I14
sS'm'
(lp2797
I1
aI3
aI3
aI11
aI3
aI61
aI87
aI79
aI499
aI91
aI771
aI1987
aI2017
aI3381
asS'd'
I1399
sa(dp2798
S'a'
I3159
sS's'
I14
sS'm'
(lp2799
I1
aI3
aI1
aI7
aI5
aI57
aI1
aI121
aI155
aI225
aI501
aI477
aI6555
aI9863
asS'd'
I1400
sa(dp2800
S'a'
I3182
sS's'
I14
sS'm'
(lp2801
I1
aI3
aI7
aI11
aI27
aI49
aI83
aI213
aI61
aI283
aI1599
aI3205
aI2525
aI8553
asS'd'
I1401
sa(dp2802
S'a'
I3187
sS's'
I14
sS'm'
(lp2803
I1
aI1
aI1
aI9
aI9
aI49
aI3
aI51
aI141
aI33
aI301
aI2167
aI587
aI15067
asS'd'
I1402
sa(dp2804
S'a'
I3189
sS's'
I14
sS'm'
(lp2805
I1
aI1
aI1
aI11
aI7
aI55
aI99
aI81
aI191
aI553
aI953
aI3753
aI6731
aI1093
asS'd'
I1403
sa(dp2806
S'a'
I3199
sS's'
I14
sS'm'
(lp2807
I1
aI1
aI3
aI3
aI11
aI59
aI57
aI235
aI297
aI197
aI853
aI1411
aI3799
aI7527
asS'd'
I1404
sa(dp2808
S'a'
I3239
sS's'
I14
sS'm'
(lp2809
I1
aI3
aI5
aI3
aI7
aI7
aI5
aI201
aI393
aI95
aI91
aI3273
aI6285
aI10661
asS'd'
I1405
sa(dp2810
S'a'
I3263
sS's'
I14
sS'm'
(lp2811
I1
aI1
aI5
aI7
aI17
aI57
aI87
aI3
aI413
aI915
aI659
aI369
aI3593
aI14429
asS'd'
I1406
sa(dp2812
S'a'
I3271
sS's'
I14
sS'm'
(lp2813
I1
aI3
aI7
aI1
aI31
aI31
aI45
aI115
aI417
aI427
aI745
aI4087
aI953
aI1119
asS'd'
I1407
sa(dp2814
S'a'
I3275
sS's'
I14
sS'm'
(lp2815
I1
aI3
aI7
aI3
aI29
aI43
aI45
aI221
aI41
aI641
aI451
aI173
aI2999
aI12103
asS'd'
I1408
sa(dp2816
S'a'
I3278
sS's'
I14
sS'm'
(lp2817
I1
aI1
aI3
aI11
aI25
aI57
aI117
aI201
aI135
aI787
aI1525
aI3879
aI3247
aI8907
asS'd'
I1409
sa(dp2818
S'a'
I3280
sS's'
I14
sS'm'
(lp2819
I1
aI1
aI7
aI11
aI3
aI35
aI69
aI157
aI331
aI615
aI573
aI2169
aI3575
aI289
asS'd'
I1410
sa(dp2820
S'a'
I3283
sS's'
I14
sS'm'
(lp2821
I1
aI3
aI3
aI13
aI15
aI51
aI67
aI127
aI265
aI495
aI103
aI3145
aI2685
aI15919
asS'd'
I1411
sa(dp2822
S'a'
I3290
sS's'
I14
sS'm'
(lp2823
I1
aI3
aI5
aI11
aI31
aI27
aI65
aI57
aI153
aI465
aI1163
aI467
aI4103
aI4713
asS'd'
I1412
sa(dp2824
S'a'
I3311
sS's'
I14
sS'm'
(lp2825
I1
aI3
aI7
aI3
aI23
aI31
aI9
aI51
aI239
aI417
aI1597
aI229
aI2865
aI15199
asS'd'
I1413
sa(dp2826
S'a'
I3316
sS's'
I14
sS'm'
(lp2827
I1
aI3
aI5
aI3
aI11
aI45
aI123
aI217
aI31
aI765
aI1009
aI2001
aI3645
aI9407
asS'd'
I1414
sa(dp2828
S'a'
I3343
sS's'
I14
sS'm'
(lp2829
I1
aI3
aI3
aI9
aI5
aI23
aI117
aI83
aI237
aI1017
aI251
aI1187
aI2631
aI5151
asS'd'
I1415
sa(dp2830
S'a'
I3346
sS's'
I14
sS'm'
(lp2831
I1
aI1
aI1
aI7
aI23
aI55
aI97
aI141
aI501
aI305
aI467
aI4061
aI2369
aI15973
asS'd'
I1416
sa(dp2832
S'a'
I3357
sS's'
I14
sS'm'
(lp2833
I1
aI1
aI7
aI5
aI31
aI51
aI125
aI191
aI219
aI495
aI37
aI3337
aI813
aI241
asS'd'
I1417
sa(dp2834
S'a'
I3358
sS's'
I14
sS'm'
(lp2835
I1
aI3
aI1
aI1
aI11
aI39
aI93
aI109
aI285
aI147
aI1297
aI737
aI4051
aI7223
asS'd'
I1418
sa(dp2836
S'a'
I3361
sS's'
I14
sS'm'
(lp2837
I1
aI3
aI1
aI15
aI13
aI17
aI57
aI171
aI463
aI163
aI609
aI1681
aI7583
aI9231
asS'd'
I1419
sa(dp2838
S'a'
I3362
sS's'
I14
sS'm'
(lp2839
I1
aI3
aI1
aI1
aI23
aI5
aI51
aI5
aI205
aI415
aI419
aI989
aI4239
aI10943
asS'd'
I1420
sa(dp2840
S'a'
I3364
sS's'
I14
sS'm'
(lp2841
I1
aI1
aI3
aI15
aI3
aI13
aI65
aI145
aI387
aI59
aI395
aI1067
aI4143
aI5649
asS'd'
I1421
sa(dp2842
S'a'
I3386
sS's'
I14
sS'm'
(lp2843
I1
aI3
aI1
aI13
aI9
aI59
aI121
aI127
aI95
aI71
aI1541
aI1423
aI1753
aI8041
asS'd'
I1422
sa(dp2844
S'a'
I3418
sS's'
I14
sS'm'
(lp2845
I1
aI1
aI3
aI15
aI7
aI5
aI69
aI167
aI181
aI991
aI1189
aI4017
aI5935
aI6669
asS'd'
I1423
sa(dp2846
S'a'
I3424
sS's'
I14
sS'm'
(lp2847
I1
aI3
aI5
aI7
aI23
aI41
aI53
aI21
aI47
aI261
aI1231
aI2011
aI133
aI2247
asS'd'
I1424
sa(dp2848
S'a'
I3433
sS's'
I14
sS'm'
(lp2849
I1
aI1
aI1
aI5
aI17
aI47
aI77
aI19
aI331
aI609
aI1893
aI3965
aI3123
aI9093
asS'd'
I1425
sa(dp2850
S'a'
I3434
sS's'
I14
sS'm'
(lp2851
I1
aI3
aI1
aI3
aI9
aI39
aI103
aI231
aI249
aI75
aI373
aI107
aI1823
aI10801
asS'd'
I1426
sa(dp2852
S'a'
I3436
sS's'
I14
sS'm'
(lp2853
I1
aI3
aI3
aI7
aI1
aI51
aI35
aI111
aI137
aI879
aI1221
aI225
aI4285
aI2287
asS'd'
I1427
sa(dp2854
S'a'
I3463
sS's'
I14
sS'm'
(lp2855
I1
aI1
aI7
aI9
aI23
aI17
aI75
aI245
aI409
aI163
aI395
aI3731
aI7111
aI6845
asS'd'
I1428
sa(dp2856
S'a'
I3467
sS's'
I14
sS'm'
(lp2857
I1
aI1
aI3
aI13
aI29
aI47
aI75
aI153
aI497
aI621
aI1691
aI3187
aI2125
aI10533
asS'd'
I1429
sa(dp2858
S'a'
I3477
sS's'
I14
sS'm'
(lp2859
I1
aI1
aI7
aI7
aI9
aI7
aI55
aI159
aI255
aI417
aI1335
aI643
aI3843
aI3733
asS'd'
I1430
sa(dp2860
S'a'
I3484
sS's'
I14
sS'm'
(lp2861
I1
aI3
aI3
aI1
aI21
aI41
aI7
aI21
aI5
aI679
aI1655
aI95
aI5699
aI5785
asS'd'
I1431
sa(dp2862
S'a'
I3505
sS's'
I14
sS'm'
(lp2863
I1
aI1
aI1
aI13
aI19
aI7
aI85
aI7
aI195
aI357
aI1097
aI2893
aI2913
aI9635
asS'd'
I1432
sa(dp2864
S'a'
I3508
sS's'
I14
sS'm'
(lp2865
I1
aI1
aI5
aI9
aI25
aI33
aI41
aI155
aI39
aI655
aI1993
aI3117
aI3639
aI7977
asS'd'
I1433
sa(dp2866
S'a'
I3515
sS's'
I14
sS'm'
(lp2867
I1
aI1
aI1
aI13
aI3
aI63
aI121
aI247
aI151
aI673
aI609
aI285
aI2299
aI7663
asS'd'
I1434
sa(dp2868
S'a'
I3532
sS's'
I14
sS'm'
(lp2869
I1
aI3
aI7
aI11
aI17
aI13
aI49
aI253
aI245
aI21
aI273
aI993
aI911
aI863
asS'd'
I1435
sa(dp2870
S'a'
I3553
sS's'
I14
sS'm'
(lp2871
I1
aI1
aI5
aI5
aI23
aI1
aI121
aI95
aI225
aI9
aI1237
aI1183
aI6461
aI559
asS'd'
I1436
sa(dp2872
S'a'
I3554
sS's'
I14
sS'm'
(lp2873
I1
aI3
aI7
aI13
aI3
aI7
aI121
aI151
aI233
aI561
aI281
aI3583
aI897
aI1767
asS'd'
I1437
sa(dp2874
S'a'
I3568
sS's'
I14
sS'm'
(lp2875
I1
aI1
aI7
aI7
aI9
aI47
aI107
aI41
aI25
aI569
aI1697
aI2299
aI6231
aI12209
asS'd'
I1438
sa(dp2876
S'a'
I3573
sS's'
I14
sS'm'
(lp2877
I1
aI3
aI7
aI7
aI27
aI43
aI59
aI37
aI31
aI51
aI503
aI149
aI4043
aI11847
asS'd'
I1439
sa(dp2878
S'a'
I3587
sS's'
I14
sS'm'
(lp2879
I1
aI3
aI3
aI11
aI5
aI1
aI119
aI181
aI47
aI641
aI685
aI4017
aI637
aI16251
asS'd'
I1440
sa(dp2880
S'a'
I3589
sS's'
I14
sS'm'
(lp2881
I1
aI3
aI3
aI7
aI11
aI1
aI101
aI7
aI239
aI747
aI307
aI1721
aI5979
aI4367
asS'd'
I1441
sa(dp2882
S'a'
I3596
sS's'
I14
sS'm'
(lp2883
I1
aI3
aI5
aI7
aI1
aI63
aI19
aI151
aI469
aI333
aI1587
aI2453
aI897
aI4711
asS'd'
I1442
sa(dp2884
S'a'
I3608
sS's'
I14
sS'm'
(lp2885
I1
aI3
aI1
aI5
aI11
aI61
aI21
aI253
aI91
aI993
aI1347
aI1993
aI5607
aI13031
asS'd'
I1443
sa(dp2886
S'a'
I3620
sS's'
I14
sS'm'
(lp2887
I1
aI3
aI5
aI5
aI1
aI39
aI65
aI71
aI189
aI389
aI1437
aI1055
aI6439
aI3989
asS'd'
I1444
sa(dp2888
S'a'
I3630
sS's'
I14
sS'm'
(lp2889
I1
aI1
aI3
aI3
aI19
aI15
aI93
aI3
aI339
aI165
aI1675
aI3953
aI2145
aI12113
asS'd'
I1445
sa(dp2890
S'a'
I3644
sS's'
I14
sS'm'
(lp2891
I1
aI1
aI3
aI13
aI13
aI45
aI5
aI175
aI211
aI993
aI705
aI2761
aI3023
aI13633
asS'd'
I1446
sa(dp2892
S'a'
I3649
sS's'
I14
sS'm'
(lp2893
I1
aI1
aI3
aI1
aI19
aI39
aI121
aI29
aI287
aI87
aI281
aI3491
aI7107
aI13007
asS'd'
I1447
sa(dp2894
S'a'
I3664
sS's'
I14
sS'm'
(lp2895
I1
aI1
aI7
aI1
aI29
aI49
aI103
aI187
aI39
aI923
aI51
aI1533
aI3249
aI4399
asS'd'
I1448
sa(dp2896
S'a'
I3679
sS's'
I14
sS'm'
(lp2897
I1
aI1
aI5
aI5
aI5
aI43
aI25
aI107
aI453
aI955
aI115
aI57
aI4589
aI14573
asS'd'
I1449
sa(dp2898
S'a'
I3680
sS's'
I14
sS'm'
(lp2899
I1
aI1
aI3
aI5
aI21
aI45
aI103
aI99
aI183
aI987
aI1207
aI1697
aI8033
aI13703
asS'd'
I1450
sa(dp2900
S'a'
I3685
sS's'
I14
sS'm'
(lp2901
I1
aI1
aI5
aI7
aI11
aI23
aI9
aI17
aI261
aI749
aI1957
aI935
aI6283
aI8625
asS'd'
I1451
sa(dp2902
S'a'
I3686
sS's'
I14
sS'm'
(lp2903
I1
aI1
aI1
aI9
aI9
aI51
aI69
aI225
aI265
aI323
aI1161
aI2993
aI7305
aI2249
asS'd'
I1452
sa(dp2904
S'a'
I3698
sS's'
I14
sS'm'
(lp2905
I1
aI3
aI1
aI9
aI23
aI19
aI57
aI205
aI503
aI489
aI1499
aI3277
aI817
aI11931
asS'd'
I1453
sa(dp2906
S'a'
I3714
sS's'
I14
sS'm'
(lp2907
I1
aI3
aI3
aI5
aI1
aI7
aI49
aI1
aI313
aI123
aI643
aI2027
aI1469
aI3585
asS'd'
I1454
sa(dp2908
S'a'
I3726
sS's'
I14
sS'm'
(lp2909
I1
aI3
aI7
aI11
aI27
aI47
aI95
aI111
aI27
aI213
aI465
aI3693
aI3661
aI7531
asS'd'
I1455
sa(dp2910
S'a'
I3737
sS's'
I14
sS'm'
(lp2911
I1
aI1
aI7
aI9
aI3
aI37
aI115
aI189
aI31
aI613
aI1393
aI1229
aI4767
aI12425
asS'd'
I1456
sa(dp2912
S'a'
I3767
sS's'
I14
sS'm'
(lp2913
I1
aI1
aI3
aI3
aI25
aI17
aI99
aI47
aI161
aI931
aI959
aI1293
aI7095
aI8325
asS'd'
I1457
sa(dp2914
S'a'
I3782
sS's'
I14
sS'm'
(lp2915
I1
aI1
aI1
aI7
aI23
aI9
aI11
aI51
aI205
aI419
aI479
aI1497
aI2493
aI13921
asS'd'
I1458
sa(dp2916
S'a'
I3786
sS's'
I14
sS'm'
(lp2917
I1
aI3
aI1
aI9
aI17
aI29
aI51
aI79
aI159
aI435
aI477
aI413
aI3815
aI5589
asS'd'
I1459
sa(dp2918
S'a'
I3793
sS's'
I14
sS'm'
(lp2919
I1
aI3
aI7
aI5
aI7
aI23
aI99
aI43
aI169
aI665
aI403
aI1163
aI4337
aI1335
asS'd'
I1460
sa(dp2920
S'a'
I3796
sS's'
I14
sS'm'
(lp2921
I1
aI3
aI1
aI5
aI25
aI27
aI125
aI249
aI421
aI267
aI1259
aI4089
aI59
aI9377
asS'd'
I1461
sa(dp2922
S'a'
I3805
sS's'
I14
sS'm'
(lp2923
I1
aI3
aI3
aI1
aI27
aI37
aI91
aI17
aI123
aI597
aI1749
aI3449
aI6503
aI11043
asS'd'
I1462
sa(dp2924
S'a'
I3815
sS's'
I14
sS'm'
(lp2925
I1
aI3
aI7
aI7
aI23
aI41
aI19
aI245
aI109
aI569
aI547
aI1917
aI7943
aI2697
asS'd'
I1463
sa(dp2926
S'a'
I3841
sS's'
I14
sS'm'
(lp2927
I1
aI3
aI7
aI7
aI9
aI1
aI123
aI105
aI329
aI435
aI2013
aI2745
aI347
aI11045
asS'd'
I1464
sa(dp2928
S'a'
I3847
sS's'
I14
sS'm'
(lp2929
I1
aI1
aI1
aI13
aI29
aI53
aI51
aI67
aI105
aI89
aI1887
aI3543
aI963
aI8159
asS'd'
I1465
sa(dp2930
S'a'
I3853
sS's'
I14
sS'm'
(lp2931
I1
aI1
aI5
aI3
aI5
aI27
aI41
aI67
aI67
aI883
aI973
aI1785
aI901
aI14969
asS'd'
I1466
sa(dp2932
S'a'
I3862
sS's'
I14
sS'm'
(lp2933
I1
aI3
aI3
aI13
aI17
aI11
aI117
aI115
aI163
aI939
aI79
aI641
aI4365
aI2267
asS'd'
I1467
sa(dp2934
S'a'
I3875
sS's'
I14
sS'm'
(lp2935
I1
aI1
aI3
aI3
aI9
aI5
aI41
aI123
aI149
aI9
aI1533
aI3939
aI5995
aI12701
asS'd'
I1468
sa(dp2936
S'a'
I3902
sS's'
I14
sS'm'
(lp2937
I1
aI1
aI1
aI15
aI31
aI1
aI101
aI229
aI191
aI965
aI61
aI2671
aI4177
aI15779
asS'd'
I1469
sa(dp2938
S'a'
I3904
sS's'
I14
sS'm'
(lp2939
I1
aI1
aI5
aI15
aI1
aI25
aI49
aI185
aI33
aI697
aI1043
aI2639
aI7819
aI3171
asS'd'
I1470
sa(dp2940
S'a'
I3916
sS's'
I14
sS'm'
(lp2941
I1
aI3
aI5
aI13
aI19
aI9
aI111
aI49
aI47
aI847
aI1865
aI717
aI5287
aI13417
asS'd'
I1471
sa(dp2942
S'a'
I3947
sS's'
I14
sS'm'
(lp2943
I1
aI3
aI7
aI11
aI5
aI61
aI63
aI111
aI171
aI735
aI2003
aI73
aI5701
aI647
asS'd'
I1472
sa(dp2944
S'a'
I3949
sS's'
I14
sS'm'
(lp2945
I1
aI3
aI1
aI11
aI1
aI49
aI121
aI79
aI431
aI671
aI1241
aI1161
aI2057
aI263
asS'd'
I1473
sa(dp2946
S'a'
I3955
sS's'
I14
sS'm'
(lp2947
I1
aI3
aI3
aI1
aI1
aI23
aI75
aI15
aI117
aI641
aI313
aI1525
aI2041
aI1409
asS'd'
I1474
sa(dp2948
S'a'
I3962
sS's'
I14
sS'm'
(lp2949
I1
aI3
aI5
aI11
aI15
aI57
aI13
aI67
aI139
aI131
aI1339
aI2419
aI7945
aI11877
asS'd'
I1475
sa(dp2950
S'a'
I3971
sS's'
I14
sS'm'
(lp2951
I1
aI3
aI1
aI1
aI19
aI39
aI97
aI83
aI297
aI595
aI1611
aI5
aI4753
aI3435
asS'd'
I1476
sa(dp2952
S'a'
I3980
sS's'
I14
sS'm'
(lp2953
I1
aI3
aI1
aI9
aI7
aI49
aI125
aI101
aI383
aI717
aI63
aI2295
aI3873
aI13461
asS'd'
I1477
sa(dp2954
S'a'
I3985
sS's'
I14
sS'm'
(lp2955
I1
aI1
aI3
aI3
aI15
aI29
aI89
aI77
aI269
aI689
aI229
aI1207
aI7311
aI8663
asS'd'
I1478
sa(dp2956
S'a'
I3998
sS's'
I14
sS'm'
(lp2957
I1
aI1
aI1
aI1
aI1
aI61
aI25
aI255
aI203
aI233
aI271
aI987
aI2277
aI8735
asS'd'
I1479
sa(dp2958
S'a'
I4001
sS's'
I14
sS'm'
(lp2959
I1
aI1
aI5
aI7
aI21
aI27
aI63
aI79
aI337
aI133
aI1453
aI3633
aI6157
aI15875
asS'd'
I1480
sa(dp2960
S'a'
I4002
sS's'
I14
sS'm'
(lp2961
I1
aI3
aI1
aI7
aI7
aI55
aI31
aI81
aI203
aI709
aI1743
aI1677
aI4247
aI11411
asS'd'
I1481
sa(dp2962
S'a'
I4016
sS's'
I14
sS'm'
(lp2963
I1
aI1
aI3
aI3
aI29
aI51
aI37
aI17
aI487
aI325
aI1393
aI1433
aI3467
aI2851
asS'd'
I1482
sa(dp2964
S'a'
I4021
sS's'
I14
sS'm'
(lp2965
I1
aI1
aI7
aI9
aI3
aI41
aI99
aI177
aI241
aI869
aI739
aI2729
aI4585
aI14801
asS'd'
I1483
sa(dp2966
S'a'
I4026
sS's'
I14
sS'm'
(lp2967
I1
aI1
aI3
aI1
aI9
aI43
aI97
aI65
aI99
aI295
aI1693
aI2083
aI3241
aI4073
asS'd'
I1484
sa(dp2968
S'a'
I4043
sS's'
I14
sS'm'
(lp2969
I1
aI1
aI1
aI9
aI5
aI39
aI67
aI119
aI235
aI543
aI795
aI2773
aI3195
aI6273
asS'd'
I1485
sa(dp2970
S'a'
I4079
sS's'
I14
sS'm'
(lp2971
I1
aI1
aI5
aI5
aI21
aI41
aI89
aI1
aI85
aI81
aI57
aI2951
aI1531
aI10101
asS'd'
I1486
sa(dp2972
S'a'
I4102
sS's'
I14
sS'm'
(lp2973
I1
aI1
aI1
aI7
aI3
aI35
aI127
aI69
aI39
aI265
aI1643
aI2973
aI267
aI12365
asS'd'
I1487
sa(dp2974
S'a'
I4106
sS's'
I14
sS'm'
(lp2975
I1
aI3
aI1
aI1
aI21
aI57
aI99
aI205
aI119
aI477
aI1771
aI1989
aI2761
aI12573
asS'd'
I1488
sa(dp2976
S'a'
I4119
sS's'
I14
sS'm'
(lp2977
I1
aI1
aI3
aI13
aI1
aI59
aI93
aI125
aI279
aI935
aI1877
aI2061
aI4845
aI7835
asS'd'
I1489
sa(dp2978
S'a'
I4126
sS's'
I14
sS'm'
(lp2979
I1
aI1
aI7
aI9
aI7
aI45
aI69
aI99
aI273
aI35
aI1579
aI2137
aI7175
aI6999
asS'd'
I1490
sa(dp2980
S'a'
I4147
sS's'
I14
sS'm'
(lp2981
I1
aI1
aI7
aI7
aI29
aI21
aI127
aI91
aI9
aI383
aI787
aI1783
aI601
aI5047
asS'd'
I1491
sa(dp2982
S'a'
I4149
sS's'
I14
sS'm'
(lp2983
I1
aI1
aI7
aI13
aI7
aI29
aI35
aI219
aI43
aI581
aI2043
aI2211
aI6169
aI12173
asS'd'
I1492
sa(dp2984
S'a'
I4164
sS's'
I14
sS'm'
(lp2985
I1
aI3
aI5
aI13
aI29
aI29
aI39
aI63
aI411
aI645
aI415
aI2383
aI1989
aI11411
asS'd'
I1493
sa(dp2986
S'a'
I4174
sS's'
I14
sS'm'
(lp2987
I1
aI1
aI7
aI9
aI15
aI9
aI87
aI95
aI321
aI709
aI543
aI3831
aI2453
aI4167
asS'd'
I1494
sa(dp2988
S'a'
I4181
sS's'
I14
sS'm'
(lp2989
I1
aI3
aI1
aI5
aI31
aI25
aI5
aI85
aI239
aI487
aI1613
aI3937
aI4661
aI3535
asS'd'
I1495
sa(dp2990
S'a'
I4185
sS's'
I14
sS'm'
(lp2991
I1
aI3
aI5
aI11
aI27
aI41
aI3
aI201
aI39
aI91
aI1997
aI237
aI5639
aI14703
asS'd'
I1496
sa(dp2992
S'a'
I4188
sS's'
I14
sS'm'
(lp2993
I1
aI1
aI7
aI3
aI27
aI49
aI87
aI71
aI473
aI247
aI1007
aI47
aI475
aI5413
asS'd'
I1497
sa(dp2994
S'a'
I4202
sS's'
I14
sS'm'
(lp2995
I1
aI3
aI7
aI15
aI9
aI57
aI81
aI149
aI287
aI333
aI1911
aI3417
aI1081
aI8995
asS'd'
I1498
sa(dp2996
S'a'
I4228
sS's'
I14
sS'm'
(lp2997
I1
aI1
aI5
aI1
aI3
aI63
aI43
aI151
aI97
aI431
aI961
aI1019
aI5153
aI2407
asS'd'
I1499
sa(dp2998
S'a'
I4232
sS's'
I14
sS'm'
(lp2999
I1
aI1
aI5
aI5
aI27
aI21
aI127
aI161
aI507
aI311
aI129
aI3489
aI1133
aI3241
asS'd'
I1500
sa(dp3000
S'a'
I4246
sS's'
I14
sS'm'
(lp3001
I1
aI3
aI7
aI15
aI21
aI33
aI117
aI83
aI497
aI667
aI1399
aI931
aI1695
aI8171
asS'd'
I1501
sa(dp3002
S'a'
I4252
sS's'
I14
sS'm'
(lp3003
I1
aI1
aI1
aI13
aI3
aI39
aI53
aI27
aI193
aI993
aI671
aI1871
aI7579
aI11457
asS'd'
I1502
sa(dp3004
S'a'
I4256
sS's'
I14
sS'm'
(lp3005
I1
aI1
aI5
aI11
aI7
aI39
aI81
aI107
aI195
aI387
aI849
aI395
aI1317
aI6487
asS'd'
I1503
sa(dp3006
S'a'
I4286
sS's'
I14
sS'm'
(lp3007
I1
aI3
aI3
aI3
aI3
aI15
aI45
aI127
aI279
aI111
aI331
aI357
aI4637
aI4697
asS'd'
I1504
sa(dp3008
S'a'
I4303
sS's'
I14
sS'm'
(lp3009
I1
aI1
aI3
aI9
aI21
aI49
aI47
aI97
aI61
aI101
aI181
aI1867
aI1201
aI14099
asS'd'
I1505
sa(dp3010
S'a'
I4306
sS's'
I14
sS'm'
(lp3011
I1
aI1
aI5
aI11
aI25
aI19
aI51
aI51
aI101
aI451
aI545
aI101
aI7497
aI9141
asS'd'
I1506
sa(dp3012
S'a'
I4311
sS's'
I14
sS'm'
(lp3013
I1
aI1
aI1
aI3
aI13
aI53
aI119
aI81
aI377
aI245
aI765
aI251
aI3757
aI16045
asS'd'
I1507
sa(dp3014
S'a'
I4317
sS's'
I14
sS'm'
(lp3015
I1
aI1
aI1
aI3
aI5
aI61
aI65
aI37
aI331
aI925
aI1439
aI3219
aI2843
aI11397
asS'd'
I1508
sa(dp3016
S'a'
I4342
sS's'
I14
sS'm'
(lp3017
I1
aI3
aI5
aI9
aI23
aI31
aI95
aI155
aI83
aI641
aI1129
aI135
aI477
aI1623
asS'd'
I1509
sa(dp3018
S'a'
I4346
sS's'
I14
sS'm'
(lp3019
I1
aI1
aI3
aI9
aI9
aI61
aI93
aI11
aI331
aI585
aI799
aI1417
aI1533
aI463
asS'd'
I1510
sa(dp3020
S'a'
I4377
sS's'
I14
sS'm'
(lp3021
I1
aI1
aI7
aI7
aI21
aI51
aI61
aI29
aI467
aI935
aI11
aI3357
aI1087
aI12337
asS'd'
I1511
sa(dp3022
S'a'
I4401
sS's'
I14
sS'm'
(lp3023
I1
aI3
aI3
aI11
aI1
aI39
aI103
aI153
aI351
aI893
aI1823
aI835
aI2149
aI4203
asS'd'
I1512
sa(dp3024
S'a'
I4407
sS's'
I14
sS'm'
(lp3025
I1
aI1
aI1
aI9
aI31
aI13
aI61
aI235
aI369
aI359
aI835
aI2067
aI2697
aI15289
asS'd'
I1513
sa(dp3026
S'a'
I4414
sS's'
I14
sS'm'
(lp3027
I1
aI1
aI7
aI1
aI15
aI1
aI107
aI27
aI201
aI451
aI1521
aI313
aI3195
aI3847
asS'd'
I1514
sa(dp3028
S'a'
I4422
sS's'
I14
sS'm'
(lp3029
I1
aI1
aI5
aI13
aI1
aI27
aI63
aI137
aI355
aI489
aI2039
aI1015
aI2519
aI13797
asS'd'
I1515
sa(dp3030
S'a'
I4431
sS's'
I14
sS'm'
(lp3031
I1
aI1
aI7
aI9
aI29
aI33
aI23
aI197
aI49
aI555
aI1087
aI3447
aI7299
aI15513
asS'd'
I1516
sa(dp3032
S'a'
I4434
sS's'
I14
sS'm'
(lp3033
I1
aI3
aI5
aI11
aI7
aI37
aI55
aI63
aI443
aI573
aI1715
aI631
aI3405
aI6155
asS'd'
I1517
sa(dp3034
S'a'
I4436
sS's'
I14
sS'm'
(lp3035
I1
aI3
aI3
aI3
aI31
aI35
aI51
aI167
aI225
aI617
aI2007
aI2555
aI6819
aI12709
asS'd'
I1518
sa(dp3036
S'a'
I4443
sS's'
I14
sS'm'
(lp3037
I1
aI1
aI1
aI13
aI15
aI5
aI73
aI85
aI109
aI43
aI1067
aI3941
aI1125
aI10269
asS'd'
I1519
sa(dp3038
S'a'
I4459
sS's'
I14
sS'm'
(lp3039
I1
aI1
aI7
aI11
aI17
aI3
aI127
aI145
aI279
aI19
aI1007
aI3287
aI4751
aI12507
asS'd'
I1520
sa(dp3040
S'a'
I4461
sS's'
I14
sS'm'
(lp3041
I1
aI3
aI7
aI3
aI19
aI1
aI117
aI111
aI193
aI435
aI47
aI1801
aI529
aI8547
asS'd'
I1521
sa(dp3042
S'a'
I4462
sS's'
I14
sS'm'
(lp3043
I1
aI3
aI3
aI13
aI1
aI19
aI101
aI19
aI469
aI187
aI207
aI1243
aI8153
aI3273
asS'd'
I1522
sa(dp3044
S'a'
I4473
sS's'
I14
sS'm'
(lp3045
I1
aI3
aI1
aI5
aI11
aI51
aI69
aI189
aI453
aI775
aI241
aI3331
aI4067
aI14759
asS'd'
I1523
sa(dp3046
S'a'
I4497
sS's'
I14
sS'm'
(lp3047
I1
aI1
aI1
aI1
aI23
aI55
aI113
aI133
aI497
aI731
aI391
aI2777
aI3529
aI955
asS'd'
I1524
sa(dp3048
S'a'
I4504
sS's'
I14
sS'm'
(lp3049
I1
aI3
aI1
aI11
aI5
aI49
aI59
aI35
aI261
aI949
aI325
aI3595
aI7433
aI11099
asS'd'
I1525
sa(dp3050
S'a'
I4507
sS's'
I14
sS'm'
(lp3051
I1
aI3
aI5
aI9
aI13
aI37
aI103
aI219
aI329
aI865
aI1787
aI2497
aI7249
aI9877
asS'd'
I1526
sa(dp3052
S'a'
I4525
sS's'
I14
sS'm'
(lp3053
I1
aI3
aI7
aI9
aI11
aI33
aI19
aI255
aI191
aI935
aI1115
aI1901
aI1577
aI9623
asS'd'
I1527
sa(dp3054
S'a'
I4534
sS's'
I14
sS'm'
(lp3055
I1
aI1
aI5
aI7
aI29
aI23
aI77
aI43
aI283
aI143
aI1211
aI73
aI2835
aI10235
asS'd'
I1528
sa(dp3056
S'a'
I4538
sS's'
I14
sS'm'
(lp3057
I1
aI1
aI7
aI3
aI3
aI27
aI35
aI173
aI453
aI425
aI1225
aI3023
aI2159
aI8433
asS'd'
I1529
sa(dp3058
S'a'
I4548
sS's'
I14
sS'm'
(lp3059
I1
aI1
aI1
aI5
aI27
aI21
aI35
aI25
aI71
aI145
aI1545
aI523
aI4527
aI7655
asS'd'
I1530
sa(dp3060
S'a'
I4552
sS's'
I14
sS'm'
(lp3061
I1
aI1
aI5
aI3
aI13
aI49
aI61
aI157
aI113
aI775
aI763
aI1785
aI225
aI11851
asS'd'
I1531
sa(dp3062
S'a'
I4560
sS's'
I14
sS'm'
(lp3063
I1
aI1
aI3
aI1
aI31
aI57
aI97
aI229
aI291
aI777
aI213
aI4067
aI921
aI8203
asS'd'
I1532
sa(dp3064
S'a'
I4575
sS's'
I14
sS'm'
(lp3065
I1
aI1
aI5
aI1
aI25
aI13
aI125
aI123
aI263
aI207
aI119
aI3111
aI3841
aI843
asS'd'
I1533
sa(dp3066
S'a'
I4599
sS's'
I14
sS'm'
(lp3067
I1
aI1
aI7
aI7
aI25
aI57
aI81
aI129
aI31
aI133
aI1869
aI2949
aI5563
aI14965
asS'd'
I1534
sa(dp3068
S'a'
I4612
sS's'
I14
sS'm'
(lp3069
I1
aI3
aI3
aI7
aI3
aI51
aI33
aI127
aI281
aI425
aI1253
aI405
aI7941
aI8799
asS'd'
I1535
sa(dp3070
S'a'
I4619
sS's'
I14
sS'm'
(lp3071
I1
aI1
aI3
aI9
aI3
aI63
aI93
aI173
aI255
aI609
aI49
aI111
aI7785
aI15865
asS'd'
I1536
sa(dp3072
S'a'
I4640
sS's'
I14
sS'm'
(lp3073
I1
aI1
aI1
aI3
aI17
aI59
aI113
aI55
aI155
aI789
aI1335
aI177
aI3071
aI1851
asS'd'
I1537
sa(dp3074
S'a'
I4643
sS's'
I14
sS'm'
(lp3075
I1
aI3
aI7
aI15
aI15
aI23
aI35
aI35
aI131
aI623
aI47
aI437
aI1337
aI9891
asS'd'
I1538
sa(dp3076
S'a'
I4677
sS's'
I14
sS'm'
(lp3077
I1
aI3
aI7
aI5
aI29
aI57
aI39
aI31
aI111
aI271
aI59
aI1473
aI949
aI3899
asS'd'
I1539
sa(dp3078
S'a'
I4687
sS's'
I14
sS'm'
(lp3079
I1
aI1
aI3
aI11
aI17
aI19
aI41
aI229
aI259
aI691
aI1455
aI3023
aI7455
aI9711
asS'd'
I1540
sa(dp3080
S'a'
I4723
sS's'
I14
sS'm'
(lp3081
I1
aI3
aI5
aI11
aI29
aI13
aI9
aI165
aI499
aI355
aI1415
aI1395
aI7595
aI15571
asS'd'
I1541
sa(dp3082
S'a'
I4730
sS's'
I14
sS'm'
(lp3083
I1
aI3
aI1
aI9
aI5
aI5
aI25
aI247
aI185
aI241
aI1325
aI3133
aI7471
aI2649
asS'd'
I1542
sa(dp3084
S'a'
I4736
sS's'
I14
sS'm'
(lp3085
I1
aI3
aI3
aI11
aI17
aI29
aI57
aI61
aI51
aI203
aI993
aI1837
aI3785
aI15163
asS'd'
I1543
sa(dp3086
S'a'
I4741
sS's'
I14
sS'm'
(lp3087
I1
aI1
aI7
aI7
aI21
aI57
aI79
aI165
aI277
aI133
aI93
aI1055
aI7169
aI15685
asS'd'
I1544
sa(dp3088
S'a'
I4763
sS's'
I14
sS'm'
(lp3089
I1
aI1
aI5
aI3
aI5
aI17
aI25
aI177
aI95
aI323
aI367
aI1359
aI4915
aI6409
asS'd'
I1545
sa(dp3090
S'a'
I4765
sS's'
I14
sS'm'
(lp3091
I1
aI1
aI1
aI1
aI11
aI25
aI115
aI45
aI373
aI221
aI1483
aI591
aI6561
aI4527
asS'd'
I1546
sa(dp3092
S'a'
I4770
sS's'
I14
sS'm'
(lp3093
I1
aI3
aI5
aI3
aI5
aI23
aI69
aI77
aI313
aI473
aI1037
aI4045
aI3969
aI5445
asS'd'
I1547
sa(dp3094
S'a'
I4781
sS's'
I14
sS'm'
(lp3095
I1
aI3
aI1
aI5
aI1
aI15
aI73
aI83
aI439
aI463
aI203
aI361
aI6835
aI1061
asS'd'
I1548
sa(dp3096
S'a'
I4808
sS's'
I14
sS'm'
(lp3097
I1
aI1
aI3
aI11
aI21
aI5
aI89
aI233
aI405
aI253
aI773
aI3901
aI6085
aI5677
asS'd'
I1549
sa(dp3098
S'a'
I4822
sS's'
I14
sS'm'
(lp3099
I1
aI1
aI3
aI9
aI15
aI53
aI71
aI29
aI101
aI599
aI1073
aI705
aI4507
aI12779
asS'd'
I1550
sa(dp3100
S'a'
I4828
sS's'
I14
sS'm'
(lp3101
I1
aI1
aI3
aI1
aI3
aI9
aI27
aI97
aI207
aI859
aI417
aI735
aI2179
aI5071
asS'd'
I1551
sa(dp3102
S'a'
I4831
sS's'
I14
sS'm'
(lp3103
I1
aI1
aI1
aI3
aI13
aI63
aI65
aI125
aI195
aI611
aI649
aI2221
aI3143
aI143
asS'd'
I1552
sa(dp3104
S'a'
I4842
sS's'
I14
sS'm'
(lp3105
I1
aI3
aI3
aI15
aI17
aI57
aI99
aI119
aI243
aI407
aI1229
aI813
aI5245
aI1893
asS'd'
I1553
sa(dp3106
S'a'
I4855
sS's'
I14
sS'm'
(lp3107
I1
aI1
aI1
aI5
aI27
aI27
aI49
aI13
aI313
aI287
aI473
aI2629
aI3509
aI11371
asS'd'
I1554
sa(dp3108
S'a'
I4859
sS's'
I14
sS'm'
(lp3109
I1
aI1
aI7
aI7
aI23
aI3
aI75
aI59
aI245
aI689
aI1215
aI2375
aI3325
aI1593
asS'd'
I1555
sa(dp3110
S'a'
I4867
sS's'
I14
sS'm'
(lp3111
I1
aI3
aI1
aI5
aI21
aI51
aI43
aI107
aI91
aI611
aI1405
aI677
aI2087
aI9565
asS'd'
I1556
sa(dp3112
S'a'
I4870
sS's'
I14
sS'm'
(lp3113
I1
aI3
aI7
aI11
aI9
aI27
aI81
aI101
aI449
aI201
aI1507
aI2217
aI6767
aI8059
asS'd'
I1557
sa(dp3114
S'a'
I4881
sS's'
I14
sS'm'
(lp3115
I1
aI1
aI3
aI9
aI13
aI41
aI21
aI195
aI421
aI315
aI347
aI2621
aI2359
aI9247
asS'd'
I1558
sa(dp3116
S'a'
I4893
sS's'
I14
sS'm'
(lp3117
I1
aI1
aI5
aI7
aI31
aI45
aI77
aI229
aI455
aI575
aI1087
aI1147
aI2273
aI13773
asS'd'
I1559
sa(dp3118
S'a'
I4910
sS's'
I14
sS'm'
(lp3119
I1
aI1
aI1
aI1
aI9
aI5
aI87
aI19
aI207
aI545
aI1435
aI495
aI1299
aI4947
asS'd'
I1560
sa(dp3120
S'a'
I4917
sS's'
I14
sS'm'
(lp3121
I1
aI1
aI3
aI3
aI15
aI9
aI63
aI67
aI219
aI735
aI1911
aI2361
aI6503
aI11977
asS'd'
I1561
sa(dp3122
S'a'
I4929
sS's'
I14
sS'm'
(lp3123
I1
aI3
aI1
aI9
aI31
aI27
aI103
aI153
aI81
aI939
aI461
aI2753
aI697
aI537
asS'd'
I1562
sa(dp3124
S'a'
I4939
sS's'
I14
sS'm'
(lp3125
I1
aI3
aI3
aI9
aI21
aI53
aI49
aI211
aI415
aI817
aI321
aI3775
aI2921
aI9473
asS'd'
I1563
sa(dp3126
S'a'
I4947
sS's'
I14
sS'm'
(lp3127
I1
aI1
aI7
aI3
aI23
aI55
aI15
aI51
aI435
aI1013
aI73
aI3967
aI4575
aI13099
asS'd'
I1564
sa(dp3128
S'a'
I4949
sS's'
I14
sS'm'
(lp3129
I1
aI1
aI3
aI7
aI5
aI27
aI43
aI225
aI267
aI21
aI1261
aI603
aI6913
aI4421
asS'd'
I1565
sa(dp3130
S'a'
I4954
sS's'
I14
sS'm'
(lp3131
I1
aI1
aI7
aI13
aI25
aI31
aI101
aI109
aI237
aI91
aI1587
aI1987
aI2795
aI6991
asS'd'
I1566
sa(dp3132
S'a'
I4972
sS's'
I14
sS'm'
(lp3133
I1
aI1
aI3
aI13
aI23
aI51
aI91
aI89
aI287
aI39
aI1513
aI463
aI6135
aI10469
asS'd'
I1567
sa(dp3134
S'a'
I4975
sS's'
I14
sS'm'
(lp3135
I1
aI3
aI3
aI1
aI9
aI43
aI125
aI157
aI369
aI495
aI1849
aI785
aI6357
aI6557
asS'd'
I1568
sa(dp3136
S'a'
I5000
sS's'
I14
sS'm'
(lp3137
I1
aI3
aI1
aI13
aI5
aI25
aI107
aI139
aI367
aI239
aI1671
aI1239
aI7027
aI5291
asS'd'
I1569
sa(dp3138
S'a'
I5005
sS's'
I14
sS'm'
(lp3139
I1
aI3
aI5
aI13
aI11
aI13
aI35
aI177
aI45
aI939
aI251
aI59
aI333
aI13105
asS'd'
I1570
sa(dp3140
S'a'
I5029
sS's'
I14
sS'm'
(lp3141
I1
aI3
aI5
aI7
aI29
aI57
aI109
aI227
aI435
aI739
aI423
aI1941
aI3345
aI12731
asS'd'
I1571
sa(dp3142
S'a'
I5039
sS's'
I14
sS'm'
(lp3143
I1
aI3
aI3
aI9
aI23
aI51
aI19
aI207
aI69
aI99
aI955
aI519
aI7305
aI2415
asS'd'
I1572
sa(dp3144
S'a'
I5044
sS's'
I14
sS'm'
(lp3145
I1
aI1
aI5
aI13
aI17
aI1
aI67
aI201
aI61
aI403
aI1059
aI2915
aI2419
aI12773
asS'd'
I1573
sa(dp3146
S'a'
I5051
sS's'
I14
sS'm'
(lp3147
I1
aI3
aI1
aI11
aI17
aI19
aI25
aI27
aI207
aI299
aI143
aI1955
aI5669
aI2301
asS'd'
I1574
sa(dp3148
S'a'
I5056
sS's'
I14
sS'm'
(lp3149
I1
aI1
aI5
aI3
aI25
aI57
aI45
aI255
aI489
aI1011
aI1699
aI2637
aI5279
aI12211
asS'd'
I1575
sa(dp3150
S'a'
I5073
sS's'
I14
sS'm'
(lp3151
I1
aI3
aI3
aI15
aI7
aI47
aI113
aI33
aI511
aI907
aI1815
aI1741
aI2091
aI13857
asS'd'
I1576
sa(dp3152
S'a'
I5096
sS's'
I14
sS'm'
(lp3153
I1
aI3
aI3
aI5
aI5
aI27
aI95
aI3
aI353
aI253
aI947
aI393
aI1815
aI14551
asS'd'
I1577
sa(dp3154
S'a'
I5128
sS's'
I14
sS'm'
(lp3155
I1
aI1
aI5
aI11
aI29
aI19
aI63
aI117
aI293
aI861
aI2039
aI9
aI5999
aI6909
asS'd'
I1578
sa(dp3156
S'a'
I5134
sS's'
I14
sS'm'
(lp3157
I1
aI3
aI7
aI3
aI15
aI63
aI107
aI173
aI509
aI817
aI99
aI2825
aI131
aI7917
asS'd'
I1579
sa(dp3158
S'a'
I5161
sS's'
I14
sS'm'
(lp3159
I1
aI3
aI1
aI1
aI29
aI49
aI33
aI153
aI119
aI777
aI1315
aI3581
aI5675
aI4043
asS'd'
I1580
sa(dp3160
S'a'
I5179
sS's'
I14
sS'm'
(lp3161
I1
aI3
aI5
aI15
aI13
aI11
aI17
aI147
aI327
aI305
aI367
aI3237
aI5423
aI13757
asS'd'
I1581
sa(dp3162
S'a'
I5193
sS's'
I14
sS'm'
(lp3163
I1
aI1
aI5
aI13
aI1
aI39
aI35
aI29
aI25
aI751
aI1365
aI2243
aI8181
aI7063
asS'd'
I1582
sa(dp3164
S'a'
I5199
sS's'
I14
sS'm'
(lp3165
I1
aI3
aI7
aI11
aI25
aI53
aI11
aI111
aI289
aI755
aI1201
aI691
aI3679
aI3725
asS'd'
I1583
sa(dp3166
S'a'
I5202
sS's'
I14
sS'm'
(lp3167
I1
aI1
aI1
aI11
aI11
aI37
aI33
aI211
aI395
aI691
aI1817
aI861
aI6485
aI12077
asS'd'
I1584
sa(dp3168
S'a'
I5204
sS's'
I14
sS'm'
(lp3169
I1
aI3
aI3
aI11
aI21
aI3
aI111
aI171
aI305
aI561
aI1501
aI2011
aI7841
aI10931
asS'd'
I1585
sa(dp3170
S'a'
I5218
sS's'
I14
sS'm'
(lp3171
I1
aI3
aI7
aI9
aI9
aI59
aI109
aI113
aI31
aI915
aI103
aI1861
aI2779
aI10619
asS'd'
I1586
sa(dp3172
S'a'
I5247
sS's'
I14
sS'm'
(lp3173
I1
aI1
aI1
aI1
aI7
aI25
aI61
aI97
aI103
aI723
aI1145
aI3105
aI371
aI339
asS'd'
I1587
sa(dp3174
S'a'
I5260
sS's'
I14
sS'm'
(lp3175
I1
aI1
aI7
aI13
aI17
aI9
aI113
aI51
aI233
aI209
aI1117
aI211
aI6969
aI2347
asS'd'
I1588
sa(dp3176
S'a'
I5271
sS's'
I14
sS'm'
(lp3177
I1
aI1
aI5
aI9
aI25
aI43
aI21
aI217
aI327
aI735
aI197
aI1063
aI799
aI801
asS'd'
I1589
sa(dp3178
S'a'
I5301
sS's'
I14
sS'm'
(lp3179
I1
aI1
aI7
aI13
aI9
aI13
aI73
aI33
aI415
aI923
aI863
aI1999
aI5383
aI8119
asS'd'
I1590
sa(dp3180
S'a'
I5305
sS's'
I14
sS'm'
(lp3181
I1
aI3
aI1
aI5
aI7
aI33
aI51
aI185
aI289
aI967
aI1277
aI1011
aI767
aI15505
asS'd'
I1591
sa(dp3182
S'a'
I5319
sS's'
I14
sS'm'
(lp3183
I1
aI3
aI3
aI13
aI21
aI11
aI105
aI235
aI343
aI1021
aI2009
aI2251
aI3865
aI6923
asS'd'
I1592
sa(dp3184
S'a'
I5326
sS's'
I14
sS'm'
(lp3185
I1
aI3
aI5
aI9
aI29
aI11
aI33
aI17
aI149
aI155
aI1739
aI3039
aI7015
aI2401
asS'd'
I1593
sa(dp3186
S'a'
I5328
sS's'
I14
sS'm'
(lp3187
I1
aI3
aI7
aI7
aI17
aI13
aI89
aI177
aI297
aI267
aI545
aI3861
aI329
aI13267
asS'd'
I1594
sa(dp3188
S'a'
I5333
sS's'
I14
sS'm'
(lp3189
I1
aI3
aI5
aI15
aI27
aI33
aI1
aI231
aI181
aI557
aI447
aI379
aI7845
aI1295
asS'd'
I1595
sa(dp3190
S'a'
I5364
sS's'
I14
sS'm'
(lp3191
I1
aI1
aI5
aI13
aI3
aI63
aI59
aI33
aI263
aI877
aI1867
aI1383
aI641
aI7139
asS'd'
I1596
sa(dp3192
S'a'
I5376
sS's'
I14
sS'm'
(lp3193
I1
aI3
aI7
aI5
aI13
aI51
aI9
aI113
aI223
aI605
aI1189
aI4063
aI6925
aI9563
asS'd'
I1597
sa(dp3194
S'a'
I5399
sS's'
I14
sS'm'
(lp3195
I1
aI1
aI1
aI13
aI5
aI35
aI83
aI107
aI295
aI231
aI265
aI5
aI4087
aI6407
asS'd'
I1598
sa(dp3196
S'a'
I5416
sS's'
I14
sS'm'
(lp3197
I1
aI1
aI5
aI1
aI7
aI25
aI95
aI137
aI97
aI987
aI1753
aI2781
aI1369
aI6903
asS'd'
I1599
sa(dp3198
S'a'
I5421
sS's'
I14
sS'm'
(lp3199
I1
aI1
aI5
aI13
aI19
aI61
aI77
aI229
aI193
aI165
aI811
aI249
aI79
aI10719
asS'd'
I1600
sa(dp3200
S'a'
I5427
sS's'
I14
sS'm'
(lp3201
I1
aI3
aI7
aI7
aI27
aI9
aI119
aI193
aI459
aI43
aI1989
aI2959
aI3595
aI6341
asS'd'
I1601
sa(dp3202
S'a'
I5429
sS's'
I14
sS'm'
(lp3203
I1
aI1
aI5
aI11
aI5
aI43
aI35
aI33
aI25
aI581
aI897
aI351
aI4201
aI3971
asS'd'
I1602
sa(dp3204
S'a'
I5430
sS's'
I14
sS'm'
(lp3205
I1
aI1
aI7
aI11
aI21
aI29
aI53
aI45
aI359
aI197
aI313
aI3825
aI6717
aI4077
asS'd'
I1603
sa(dp3206
S'a'
I5434
sS's'
I14
sS'm'
(lp3207
I1
aI1
aI1
aI15
aI3
aI45
aI99
aI133
aI357
aI315
aI1159
aI241
aI2463
aI11253
asS'd'
I1604
sa(dp3208
S'a'
I5441
sS's'
I14
sS'm'
(lp3209
I1
aI1
aI7
aI11
aI9
aI33
aI111
aI85
aI443
aI601
aI447
aI337
aI6471
aI7029
asS'd'
I1605
sa(dp3210
S'a'
I5451
sS's'
I14
sS'm'
(lp3211
I1
aI3
aI7
aI9
aI13
aI33
aI25
aI31
aI9
aI729
aI1763
aI4077
aI7575
aI7877
asS'd'
I1606
sa(dp3212
S'a'
I5465
sS's'
I14
sS'm'
(lp3213
I1
aI3
aI5
aI13
aI13
aI37
aI29
aI103
aI53
aI229
aI591
aI1073
aI1323
aI14405
asS'd'
I1607
sa(dp3214
S'a'
I5466
sS's'
I14
sS'm'
(lp3215
I1
aI1
aI5
aI1
aI17
aI33
aI15
aI183
aI473
aI297
aI2003
aI93
aI4955
aI1787
asS'd'
I1608
sa(dp3216
S'a'
I5471
sS's'
I14
sS'm'
(lp3217
I1
aI1
aI5
aI13
aI5
aI29
aI113
aI161
aI267
aI451
aI1193
aI149
aI273
aI11809
asS'd'
I1609
sa(dp3218
S'a'
I5477
sS's'
I14
sS'm'
(lp3219
I1
aI1
aI1
aI9
aI17
aI39
aI47
aI233
aI165
aI373
aI955
aI2891
aI7523
aI7235
asS'd'
I1610
sa(dp3220
S'a'
I5492
sS's'
I14
sS'm'
(lp3221
I1
aI1
aI1
aI3
aI7
aI21
aI115
aI205
aI153
aI449
aI339
aI2073
aI1077
aI5749
asS'd'
I1611
sa(dp3222
S'a'
I5495
sS's'
I14
sS'm'
(lp3223
I1
aI1
aI7
aI13
aI9
aI39
aI117
aI187
aI37
aI753
aI227
aI3519
aI7391
aI5751
asS'd'
I1612
sa(dp3224
S'a'
I5505
sS's'
I14
sS'm'
(lp3225
I1
aI1
aI1
aI9
aI5
aI19
aI41
aI161
aI141
aI195
aI1719
aI3321
aI5
aI12877
asS'd'
I1613
sa(dp3226
S'a'
I5515
sS's'
I14
sS'm'
(lp3227
I1
aI3
aI7
aI11
aI21
aI13
aI83
aI55
aI299
aI75
aI1905
aI3765
aI4685
aI12297
asS'd'
I1614
sa(dp3228
S'a'
I5525
sS's'
I14
sS'm'
(lp3229
I1
aI1
aI7
aI3
aI3
aI23
aI111
aI243
aI187
aI297
aI1061
aI2515
aI977
aI9555
asS'd'
I1615
sa(dp3230
S'a'
I5529
sS's'
I14
sS'm'
(lp3231
I1
aI3
aI7
aI3
aI29
aI11
aI103
aI177
aI225
aI875
aI1649
aI1401
aI6383
aI8309
asS'd'
I1616
sa(dp3232
S'a'
I5532
sS's'
I14
sS'm'
(lp3233
I1
aI3
aI5
aI3
aI3
aI41
aI71
aI3
aI373
aI757
aI701
aI2825
aI1521
aI13217
asS'd'
I1617
sa(dp3234
S'a'
I5539
sS's'
I14
sS'm'
(lp3235
I1
aI1
aI5
aI3
aI11
aI5
aI103
aI227
aI209
aI723
aI1543
aI3895
aI6345
aI7901
asS'd'
I1618
sa(dp3236
S'a'
I5541
sS's'
I14
sS'm'
(lp3237
I1
aI1
aI5
aI1
aI9
aI51
aI77
aI67
aI359
aI937
aI557
aI993
aI3871
aI3577
asS'd'
I1619
sa(dp3238
S'a'
I5556
sS's'
I14
sS'm'
(lp3239
I1
aI3
aI7
aI1
aI1
aI15
aI121
aI239
aI29
aI113
aI1123
aI3877
aI6941
aI14129
asS'd'
I1620
sa(dp3240
S'a'
I5566
sS's'
I14
sS'm'
(lp3241
I1
aI1
aI5
aI1
aI27
aI61
aI83
aI113
aI185
aI601
aI947
aI3933
aI381
aI13869
asS'd'
I1621
sa(dp3242
S'a'
I5568
sS's'
I14
sS'm'
(lp3243
I1
aI1
aI5
aI3
aI5
aI37
aI97
aI31
aI81
aI367
aI747
aI1811
aI5313
aI14151
asS'd'
I1622
sa(dp3244
S'a'
I5574
sS's'
I14
sS'm'
(lp3245
I1
aI3
aI5
aI9
aI27
aI61
aI87
aI31
aI185
aI521
aI837
aI959
aI5001
aI3957
asS'd'
I1623
sa(dp3246
S'a'
I5595
sS's'
I14
sS'm'
(lp3247
I1
aI3
aI5
aI3
aI11
aI61
aI37
aI19
aI107
aI749
aI1345
aI3829
aI6701
aI4315
asS'd'
I1624
sa(dp3248
S'a'
I5602
sS's'
I14
sS'm'
(lp3249
I1
aI3
aI1
aI15
aI13
aI45
aI101
aI113
aI243
aI963
aI1861
aI3283
aI1419
aI12131
asS'd'
I1625
sa(dp3250
S'a'
I5611
sS's'
I14
sS'm'
(lp3251
I1
aI1
aI7
aI1
aI11
aI63
aI17
aI117
aI271
aI819
aI677
aI669
aI1991
aI12511
asS'd'
I1626
sa(dp3252
S'a'
I5616
sS's'
I14
sS'm'
(lp3253
I1
aI1
aI1
aI13
aI13
aI33
aI41
aI73
aI187
aI537
aI993
aI3147
aI1013
aI16063
asS'd'
I1627
sa(dp3254
S'a'
I5622
sS's'
I14
sS'm'
(lp3255
I1
aI3
aI1
aI1
aI25
aI21
aI107
aI81
aI117
aI917
aI113
aI349
aI4475
aI9149
asS'd'
I1628
sa(dp3256
S'a'
I5628
sS's'
I14
sS'm'
(lp3257
I1
aI1
aI1
aI11
aI21
aI21
aI29
aI251
aI125
aI681
aI141
aI2893
aI5843
aI14359
asS'd'
I1629
sa(dp3258
S'a'
I5655
sS's'
I14
sS'm'
(lp3259
I1
aI3
aI3
aI1
aI5
aI41
aI85
aI163
aI387
aI29
aI1593
aI221
aI2769
aI10809
asS'd'
I1630
sa(dp3260
S'a'
I5662
sS's'
I14
sS'm'
(lp3261
I1
aI3
aI5
aI11
aI1
aI17
aI69
aI127
aI273
aI449
aI1855
aI2971
aI7031
aI10583
asS'd'
I1631
sa(dp3262
S'a'
I5675
sS's'
I14
sS'm'
(lp3263
I1
aI1
aI5
aI7
aI1
aI61
aI9
aI211
aI123
aI563
aI111
aI1883
aI5801
aI2191
asS'd'
I1632
sa(dp3264
S'a'
I5689
sS's'
I14
sS'm'
(lp3265
I1
aI1
aI3
aI11
aI11
aI51
aI1
aI81
aI405
aI803
aI2017
aI161
aI5429
aI731
asS'd'
I1633
sa(dp3266
S'a'
I5722
sS's'
I14
sS'm'
(lp3267
I1
aI1
aI7
aI9
aI15
aI55
aI65
aI51
aI459
aI485
aI1539
aI3135
aI2929
aI7867
asS'd'
I1634
sa(dp3268
S'a'
I5724
sS's'
I14
sS'm'
(lp3269
I1
aI1
aI7
aI11
aI3
aI45
aI15
aI7
aI331
aI417
aI1813
aI4009
aI1341
aI10965
asS'd'
I1635
sa(dp3270
S'a'
I5728
sS's'
I14
sS'm'
(lp3271
I1
aI1
aI1
aI5
aI9
aI29
aI89
aI121
aI277
aI509
aI1989
aI1293
aI4787
aI16097
asS'd'
I1636
sa(dp3272
S'a'
I5731
sS's'
I14
sS'm'
(lp3273
I1
aI1
aI3
aI9
aI17
aI45
aI97
aI197
aI339
aI943
aI1377
aI2947
aI5833
aI7
asS'd'
I1637
sa(dp3274
S'a'
I5746
sS's'
I14
sS'm'
(lp3275
I1
aI1
aI7
aI9
aI15
aI61
aI75
aI233
aI401
aI705
aI825
aI2521
aI3787
aI14387
asS'd'
I1638
sa(dp3276
S'a'
I5764
sS's'
I14
sS'm'
(lp3277
I1
aI1
aI7
aI15
aI25
aI57
aI3
aI43
aI361
aI459
aI1551
aI1859
aI6787
aI2293
asS'd'
I1639
sa(dp3278
S'a'
I5771
sS's'
I14
sS'm'
(lp3279
I1
aI3
aI3
aI11
aI11
aI35
aI91
aI65
aI43
aI509
aI1829
aI1149
aI4801
aI4109
asS'd'
I1640
sa(dp3280
S'a'
I5781
sS's'
I14
sS'm'
(lp3281
I1
aI3
aI5
aI9
aI15
aI3
aI81
aI109
aI231
aI481
aI417
aI2505
aI315
aI6693
asS'd'
I1641
sa(dp3282
S'a'
I5801
sS's'
I14
sS'm'
(lp3283
I1
aI1
aI3
aI9
aI3
aI7
aI107
aI221
aI297
aI543
aI149
aI579
aI927
aI79
asS'd'
I1642
sa(dp3284
S'a'
I5809
sS's'
I14
sS'm'
(lp3285
I1
aI3
aI1
aI11
aI17
aI3
aI81
aI137
aI157
aI587
aI741
aI1277
aI2631
aI3953
asS'd'
I1643
sa(dp3286
S'a'
I5810
sS's'
I14
sS'm'
(lp3287
I1
aI1
aI7
aI5
aI13
aI43
aI117
aI19
aI495
aI185
aI1105
aI605
aI5249
aI11099
asS'd'
I1644
sa(dp3288
S'a'
I5812
sS's'
I14
sS'm'
(lp3289
I1
aI1
aI7
aI9
aI23
aI55
aI91
aI213
aI21
aI779
aI857
aI2047
aI7813
aI10053
asS'd'
I1645
sa(dp3290
S'a'
I5841
sS's'
I14
sS'm'
(lp3291
I1
aI1
aI1
aI1
aI27
aI7
aI39
aI181
aI63
aI519
aI1073
aI3147
aI4111
aI363
asS'd'
I1646
sa(dp3292
S'a'
I5848
sS's'
I14
sS'm'
(lp3293
I1
aI3
aI7
aI9
aI15
aI61
aI7
aI139
aI495
aI805
aI1545
aI3789
aI2411
aI3989
asS'd'
I1647
sa(dp3294
S'a'
I5853
sS's'
I14
sS'm'
(lp3295
I1
aI1
aI3
aI1
aI25
aI11
aI23
aI241
aI167
aI607
aI479
aI153
aI7787
aI13929
asS'd'
I1648
sa(dp3296
S'a'
I5854
sS's'
I14
sS'm'
(lp3297
I1
aI3
aI5
aI15
aI29
aI35
aI45
aI71
aI457
aI297
aI883
aI3021
aI5361
aI15427
asS'd'
I1649
sa(dp3298
S'a'
I5858
sS's'
I14
sS'm'
(lp3299
I1
aI3
aI1
aI7
aI29
aI27
aI93
aI241
aI427
aI89
aI1185
aI37
aI3863
aI14095
asS'd'
I1650
sa(dp3300
S'a'
I5892
sS's'
I14
sS'm'
(lp3301
I1
aI3
aI1
aI5
aI5
aI45
aI51
aI15
aI235
aI889
aI1649
aI2331
aI2713
aI10943
asS'd'
I1651
sa(dp3302
S'a'
I5907
sS's'
I14
sS'm'
(lp3303
I1
aI1
aI3
aI11
aI11
aI15
aI71
aI85
aI135
aI163
aI139
aI1147
aI1043
aI3195
asS'd'
I1652
sa(dp3304
S'a'
I5910
sS's'
I14
sS'm'
(lp3305
I1
aI3
aI5
aI13
aI3
aI43
aI71
aI131
aI473
aI933
aI569
aI2491
aI7751
aI1865
asS'd'
I1653
sa(dp3306
S'a'
I5913
sS's'
I14
sS'm'
(lp3307
I1
aI1
aI7
aI9
aI21
aI37
aI105
aI227
aI329
aI509
aI1319
aI307
aI1557
aI14625
asS'd'
I1654
sa(dp3308
S'a'
I5920
sS's'
I14
sS'm'
(lp3309
I1
aI1
aI3
aI13
aI15
aI1
aI25
aI93
aI335
aI953
aI769
aI4039
aI369
aI10727
asS'd'
I1655
sa(dp3310
S'a'
I5929
sS's'
I14
sS'm'
(lp3311
I1
aI3
aI7
aI5
aI17
aI21
aI59
aI89
aI437
aI679
aI437
aI1543
aI7663
aI5005
asS'd'
I1656
sa(dp3312
S'a'
I5949
sS's'
I14
sS'm'
(lp3313
I1
aI1
aI7
aI15
aI27
aI49
aI125
aI13
aI397
aI877
aI1087
aI2191
aI4711
aI9065
asS'd'
I1657
sa(dp3314
S'a'
I5952
sS's'
I14
sS'm'
(lp3315
I1
aI1
aI7
aI5
aI15
aI47
aI115
aI125
aI187
aI31
aI1003
aI2575
aI5397
aI3883
asS'd'
I1658
sa(dp3316
S'a'
I5955
sS's'
I14
sS'm'
(lp3317
I1
aI1
aI7
aI11
aI15
aI1
aI127
aI207
aI383
aI707
aI183
aI1053
aI3123
aI14071
asS'd'
I1659
sa(dp3318
S'a'
I5962
sS's'
I14
sS'm'
(lp3319
I1
aI3
aI3
aI1
aI31
aI53
aI15
aI19
aI477
aI245
aI777
aI1613
aI5813
aI7443
asS'd'
I1660
sa(dp3320
S'a'
I5975
sS's'
I14
sS'm'
(lp3321
I1
aI3
aI1
aI11
aI23
aI59
aI65
aI23
aI493
aI157
aI1389
aI2833
aI4535
aI3907
asS'd'
I1661
sa(dp3322
S'a'
I5985
sS's'
I14
sS'm'
(lp3323
I1
aI1
aI7
aI1
aI19
aI7
aI51
aI135
aI327
aI441
aI1841
aI3091
aI3451
aI14381
asS'd'
I1662
sa(dp3324
S'a'
I5997
sS's'
I14
sS'm'
(lp3325
I1
aI1
aI7
aI7
aI3
aI37
aI29
aI249
aI437
aI319
aI1693
aI945
aI7639
aI5923
asS'd'
I1663
sa(dp3326
S'a'
I5998
sS's'
I14
sS'm'
(lp3327
I1
aI3
aI7
aI15
aI7
aI61
aI81
aI127
aI383
aI99
aI23
aI3833
aI3973
aI7651
asS'd'
I1664
sa(dp3328
S'a'
I6012
sS's'
I14
sS'm'
(lp3329
I1
aI3
aI1
aI7
aI7
aI21
aI119
aI185
aI243
aI619
aI1363
aI2033
aI4835
aI5089
asS'd'
I1665
sa(dp3330
S'a'
I6016
sS's'
I14
sS'm'
(lp3331
I1
aI3
aI1
aI1
aI3
aI27
aI63
aI145
aI271
aI735
aI695
aI3981
aI3049
aI5433
asS'd'
I1666
sa(dp3332
S'a'
I6026
sS's'
I14
sS'm'
(lp3333
I1
aI3
aI3
aI1
aI3
aI29
aI79
aI211
aI279
aI819
aI501
aI3665
aI1455
aI10455
asS'd'
I1667
sa(dp3334
S'a'
I6036
sS's'
I14
sS'm'
(lp3335
I1
aI1
aI3
aI3
aI31
aI61
aI113
aI5
aI411
aI91
aI489
aI3257
aI5939
aI6715
asS'd'
I1668
sa(dp3336
S'a'
I6040
sS's'
I14
sS'm'
(lp3337
I1
aI1
aI5
aI1
aI23
aI11
aI103
aI89
aI377
aI441
aI43
aI967
aI3383
aI8717
asS'd'
I1669
sa(dp3338
S'a'
I6045
sS's'
I14
sS'm'
(lp3339
I1
aI1
aI5
aI13
aI29
aI39
aI97
aI189
aI197
aI621
aI1755
aI333
aI6783
aI9711
asS'd'
I1670
sa(dp3340
S'a'
I6055
sS's'
I14
sS'm'
(lp3341
I1
aI1
aI5
aI13
aI27
aI17
aI97
aI197
aI351
aI799
aI335
aI765
aI5329
aI12549
asS'd'
I1671
sa(dp3342
S'a'
I6059
sS's'
I14
sS'm'
(lp3343
I1
aI1
aI5
aI11
aI29
aI17
aI9
aI211
aI127
aI633
aI1187
aI3965
aI4145
aI12667
asS'd'
I1672
sa(dp3344
S'a'
I6088
sS's'
I14
sS'm'
(lp3345
I1
aI1
aI7
aI5
aI27
aI29
aI65
aI115
aI287
aI325
aI461
aI5
aI899
aI2027
asS'd'
I1673
sa(dp3346
S'a'
I6115
sS's'
I14
sS'm'
(lp3347
I1
aI1
aI1
aI5
aI27
aI17
aI31
aI13
aI231
aI627
aI1163
aI649
aI1693
aI9975
asS'd'
I1674
sa(dp3348
S'a'
I6124
sS's'
I14
sS'm'
(lp3349
I1
aI3
aI1
aI15
aI7
aI49
aI113
aI123
aI427
aI603
aI347
aI2785
aI7129
aI4645
asS'd'
I1675
sa(dp3350
S'a'
I6127
sS's'
I14
sS'm'
(lp3351
I1
aI1
aI3
aI7
aI1
aI33
aI113
aI105
aI411
aI939
aI205
aI3965
aI4361
aI4649
asS'd'
I1676
sa(dp3352
S'a'
I6132
sS's'
I14
sS'm'
(lp3353
I1
aI1
aI1
aI1
aI5
aI21
aI35
aI159
aI275
aI929
aI1193
aI3205
aI4787
aI3515
asS'd'
I1677
sa(dp3354
S'a'
I6146
sS's'
I14
sS'm'
(lp3355
I1
aI1
aI1
aI5
aI1
aI21
aI29
aI191
aI275
aI233
aI1239
aI515
aI4349
aI14989
asS'd'
I1678
sa(dp3356
S'a'
I6158
sS's'
I14
sS'm'
(lp3357
I1
aI1
aI5
aI11
aI27
aI43
aI111
aI83
aI153
aI577
aI1537
aI149
aI231
aI839
asS'd'
I1679
sa(dp3358
S'a'
I6169
sS's'
I14
sS'm'
(lp3359
I1
aI3
aI5
aI13
aI21
aI19
aI57
aI69
aI87
aI163
aI271
aI3535
aI1057
aI8517
asS'd'
I1680
sa(dp3360
S'a'
I6206
sS's'
I14
sS'm'
(lp3361
I1
aI3
aI3
aI13
aI17
aI49
aI65
aI45
aI457
aI241
aI391
aI2033
aI2507
aI7771
asS'd'
I1681
sa(dp3362
S'a'
I6228
sS's'
I14
sS'm'
(lp3363
I1
aI1
aI5
aI7
aI11
aI19
aI79
aI133
aI341
aI761
aI27
aI3905
aI4137
aI14363
asS'd'
I1682
sa(dp3364
S'a'
I6237
sS's'
I14
sS'm'
(lp3365
I1
aI3
aI3
aI13
aI19
aI1
aI11
aI139
aI249
aI245
aI1393
aI2151
aI2857
aI1665
asS'd'
I1683
sa(dp3366
S'a'
I6244
sS's'
I14
sS'm'
(lp3367
I1
aI1
aI3
aI15
aI11
aI7
aI127
aI47
aI385
aI1007
aI713
aI2235
aI5489
aI8755
asS'd'
I1684
sa(dp3368
S'a'
I6247
sS's'
I14
sS'm'
(lp3369
I1
aI3
aI5
aI13
aI19
aI21
aI21
aI167
aI405
aI655
aI1653
aI889
aI7367
aI4177
asS'd'
I1685
sa(dp3370
S'a'
I6256
sS's'
I14
sS'm'
(lp3371
I1
aI1
aI5
aI3
aI19
aI63
aI99
aI39
aI89
aI415
aI951
aI2863
aI6569
aI3797
asS'd'
I1686
sa(dp3372
S'a'
I6281
sS's'
I14
sS'm'
(lp3373
I1
aI1
aI1
aI13
aI31
aI29
aI119
aI35
aI311
aI839
aI1749
aI941
aI7487
aI2385
asS'd'
I1687
sa(dp3374
S'a'
I6282
sS's'
I14
sS'm'
(lp3375
I1
aI3
aI7
aI3
aI17
aI3
aI97
aI143
aI465
aI345
aI1457
aI2201
aI5329
aI359
asS'd'
I1688
sa(dp3376
S'a'
I6284
sS's'
I14
sS'm'
(lp3377
I1
aI3
aI7
aI11
aI1
aI15
aI3
aI115
aI335
aI567
aI1749
aI1811
aI3491
aI15939
asS'd'
I1689
sa(dp3378
S'a'
I6296
sS's'
I14
sS'm'
(lp3379
I1
aI1
aI3
aI13
aI3
aI21
aI7
aI141
aI149
aI571
aI1877
aI473
aI2143
aI9569
asS'd'
I1690
sa(dp3380
S'a'
I6299
sS's'
I14
sS'm'
(lp3381
I1
aI3
aI3
aI11
aI23
aI61
aI47
aI179
aI297
aI453
aI181
aI3405
aI2981
aI13409
asS'd'
I1691
sa(dp3382
S'a'
I6302
sS's'
I14
sS'm'
(lp3383
I1
aI3
aI1
aI13
aI1
aI43
aI5
aI201
aI371
aI1003
aI367
aI2709
aI7675
aI14973
asS'd'
I1692
sa(dp3384
S'a'
I6325
sS's'
I14
sS'm'
(lp3385
I1
aI3
aI3
aI15
aI29
aI17
aI19
aI241
aI495
aI317
aI1135
aI2227
aI6457
aI4783
asS'd'
I1693
sa(dp3386
S'a'
I6349
sS's'
I14
sS'm'
(lp3387
I1
aI3
aI3
aI7
aI29
aI9
aI57
aI95
aI261
aI531
aI1717
aI3389
aI7991
aI3793
asS'd'
I1694
sa(dp3388
S'a'
I6352
sS's'
I14
sS'm'
(lp3389
I1
aI1
aI1
aI5
aI31
aI43
aI73
aI119
aI499
aI589
aI1529
aI3337
aI4097
aI15641
asS'd'
I1695
sa(dp3390
S'a'
I6362
sS's'
I14
sS'm'
(lp3391
I1
aI1
aI7
aI9
aI29
aI43
aI127
aI91
aI243
aI979
aI1325
aI2835
aI2787
aI9445
asS'd'
I1696
sa(dp3392
S'a'
I6383
sS's'
I14
sS'm'
(lp3393
I1
aI1
aI7
aI5
aI9
aI3
aI115
aI199
aI219
aI901
aI747
aI1077
aI3197
aI2443
asS'd'
I1697
sa(dp3394
S'a'
I6386
sS's'
I14
sS'm'
(lp3395
I1
aI3
aI5
aI1
aI3
aI43
aI7
aI117
aI297
aI313
aI1043
aI1579
aI5099
aI13289
asS'd'
I1698
sa(dp3396
S'a'
I6395
sS's'
I14
sS'm'
(lp3397
I1
aI1
aI7
aI11
aI29
aI33
aI15
aI121
aI131
aI579
aI317
aI1871
aI1121
aI11653
asS'd'
I1699
sa(dp3398
S'a'
I6397
sS's'
I14
sS'm'
(lp3399
I1
aI1
aI5
aI9
aI25
aI25
aI43
aI89
aI355
aI1011
aI1385
aI2901
aI6387
aI1653
asS'd'
I1700
sa(dp3400
S'a'
I6415
sS's'
I14
sS'm'
(lp3401
I1
aI1
aI1
aI9
aI5
aI47
aI61
aI165
aI85
aI757
aI1397
aI1177
aI1657
aI4899
asS'd'
I1701
sa(dp3402
S'a'
I6424
sS's'
I14
sS'm'
(lp3403
I1
aI1
aI3
aI9
aI11
aI49
aI15
aI139
aI261
aI613
aI931
aI1299
aI2777
aI2835
asS'd'
I1702
sa(dp3404
S'a'
I6429
sS's'
I14
sS'm'
(lp3405
I1
aI1
aI1
aI5
aI3
aI55
aI83
aI227
aI125
aI581
aI1607
aI1171
aI6681
aI14463
asS'd'
I1703
sa(dp3406
S'a'
I6439
sS's'
I14
sS'm'
(lp3407
I1
aI3
aI5
aI13
aI5
aI55
aI3
aI247
aI493
aI155
aI1073
aI3743
aI5719
aI4019
asS'd'
I1704
sa(dp3408
S'a'
I6451
sS's'
I14
sS'm'
(lp3409
I1
aI1
aI7
aI1
aI11
aI23
aI13
aI75
aI399
aI847
aI499
aI1643
aI6977
aI3699
asS'd'
I1705
sa(dp3410
S'a'
I6489
sS's'
I14
sS'm'
(lp3411
I1
aI3
aI1
aI9
aI11
aI41
aI47
aI131
aI313
aI627
aI481
aI2469
aI3281
aI979
asS'd'
I1706
sa(dp3412
S'a'
I6496
sS's'
I14
sS'm'
(lp3413
I1
aI3
aI5
aI13
aI29
aI3
aI65
aI101
aI11
aI29
aI1807
aI153
aI1487
aI16109
asS'd'
I1707
sa(dp3414
S'a'
I6502
sS's'
I14
sS'm'
(lp3415
I1
aI1
aI5
aI9
aI13
aI31
aI83
aI195
aI351
aI355
aI467
aI3871
aI3085
aI4441
asS'd'
I1708
sa(dp3416
S'a'
I6511
sS's'
I14
sS'm'
(lp3417
I1
aI3
aI5
aI3
aI19
aI21
aI111
aI179
aI143
aI361
aI1619
aI1547
aI3409
aI6905
asS'd'
I1709
sa(dp3418
S'a'
I6514
sS's'
I14
sS'm'
(lp3419
I1
aI1
aI5
aI9
aI31
aI1
aI93
aI199
aI491
aI135
aI1627
aI2559
aI1389
aI14561
asS'd'
I1710
sa(dp3420
S'a'
I6520
sS's'
I14
sS'm'
(lp3421
I1
aI3
aI3
aI9
aI25
aI53
aI3
aI105
aI39
aI445
aI259
aI1045
aI1129
aI9153
asS'd'
I1711
sa(dp3422
S'a'
I6523
sS's'
I14
sS'm'
(lp3423
I1
aI1
aI5
aI9
aI19
aI63
aI71
aI9
aI73
aI435
aI1377
aI4015
aI1821
aI6453
asS'd'
I1712
sa(dp3424
S'a'
I6529
sS's'
I14
sS'm'
(lp3425
I1
aI3
aI7
aI13
aI19
aI13
aI37
aI247
aI391
aI23
aI1491
aI1257
aI6395
aI237
asS'd'
I1713
sa(dp3426
S'a'
I6532
sS's'
I14
sS'm'
(lp3427
I1
aI1
aI3
aI3
aI19
aI55
aI109
aI23
aI227
aI747
aI729
aI2221
aI727
aI2209
asS'd'
I1714
sa(dp3428
S'a'
I6547
sS's'
I14
sS'm'
(lp3429
I1
aI1
aI5
aI11
aI25
aI21
aI75
aI37
aI219
aI355
aI1005
aI1895
aI7039
aI5225
asS'd'
I1715
sa(dp3430
S'a'
I6549
sS's'
I14
sS'm'
(lp3431
I1
aI3
aI5
aI13
aI11
aI43
aI9
aI67
aI87
aI797
aI1077
aI245
aI4521
aI11845
asS'd'
I1716
sa(dp3432
S'a'
I6598
sS's'
I14
sS'm'
(lp3433
I1
aI3
aI5
aI3
aI15
aI29
aI127
aI237
aI277
aI373
aI1859
aI3083
aI587
aI1123
asS'd'
I1717
sa(dp3434
S'a'
I6601
sS's'
I14
sS'm'
(lp3435
I1
aI1
aI7
aI15
aI13
aI7
aI103
aI53
aI13
aI965
aI1497
aI775
aI3439
aI1501
asS'd'
I1718
sa(dp3436
S'a'
I6610
sS's'
I14
sS'm'
(lp3437
I1
aI3
aI3
aI15
aI17
aI13
aI97
aI169
aI67
aI953
aI189
aI2739
aI1459
aI10543
asS'd'
I1719
sa(dp3438
S'a'
I6622
sS's'
I14
sS'm'
(lp3439
I1
aI1
aI5
aI1
aI17
aI39
aI15
aI127
aI327
aI989
aI1471
aI3235
aI2801
aI15311
asS'd'
I1720
sa(dp3440
S'a'
I6632
sS's'
I14
sS'm'
(lp3441
I1
aI1
aI1
aI15
aI5
aI37
aI55
aI155
aI47
aI463
aI1851
aI3467
aI2765
aI9359
asS'd'
I1721
sa(dp3442
S'a'
I6655
sS's'
I14
sS'm'
(lp3443
I1
aI3
aI3
aI15
aI1
aI13
aI93
aI239
aI291
aI115
aI365
aI61
aI395
aI15853
asS'd'
I1722
sa(dp3444
S'a'
I6665
sS's'
I14
sS'm'
(lp3445
I1
aI1
aI5
aI1
aI19
aI27
aI61
aI95
aI105
aI369
aI1557
aI961
aI6917
aI3621
asS'd'
I1723
sa(dp3446
S'a'
I6666
sS's'
I14
sS'm'
(lp3447
I1
aI3
aI3
aI9
aI7
aI35
aI115
aI53
aI111
aI345
aI1145
aI1687
aI3401
aI12107
asS'd'
I1724
sa(dp3448
S'a'
I6695
sS's'
I14
sS'm'
(lp3449
I1
aI1
aI1
aI5
aI7
aI31
aI63
aI19
aI373
aI79
aI1369
aI3037
aI2835
aI4439
asS'd'
I1725
sa(dp3450
S'a'
I6701
sS's'
I14
sS'm'
(lp3451
I1
aI3
aI7
aI9
aI11
aI17
aI29
aI33
aI331
aI447
aI1981
aI3333
aI6535
aI6557
asS'd'
I1726
sa(dp3452
S'a'
I6709
sS's'
I14
sS'm'
(lp3453
I1
aI3
aI3
aI5
aI11
aI41
aI29
aI43
aI365
aI279
aI1919
aI945
aI179
aI1987
asS'd'
I1727
sa(dp3454
S'a'
I6710
sS's'
I14
sS'm'
(lp3455
I1
aI3
aI1
aI13
aI7
aI7
aI25
aI33
aI103
aI367
aI1267
aI763
aI5691
aI8643
asS'd'
I1728
sa(dp3456
S'a'
I6741
sS's'
I14
sS'm'
(lp3457
I1
aI3
aI1
aI5
aI11
aI15
aI3
aI213
aI511
aI211
aI1069
aI4047
aI3335
aI12729
asS'd'
I1729
sa(dp3458
S'a'
I6745
sS's'
I14
sS'm'
(lp3459
I1
aI1
aI3
aI1
aI5
aI11
aI27
aI201
aI361
aI537
aI679
aI3709
aI293
aI2997
asS'd'
I1730
sa(dp3460
S'a'
I6758
sS's'
I14
sS'm'
(lp3461
I1
aI1
aI3
aI1
aI25
aI15
aI19
aI185
aI369
aI577
aI1625
aI655
aI2363
aI3861
asS'd'
I1731
sa(dp3462
S'a'
I6767
sS's'
I14
sS'm'
(lp3463
I1
aI1
aI5
aI5
aI1
aI47
aI61
aI45
aI411
aI597
aI955
aI1007
aI3775
aI5809
asS'd'
I1732
sa(dp3464
S'a'
I6772
sS's'
I14
sS'm'
(lp3465
I1
aI1
aI5
aI3
aI27
aI51
aI101
aI167
aI429
aI333
aI1703
aI3541
aI2947
aI3695
asS'd'
I1733
sa(dp3466
S'a'
I6782
sS's'
I14
sS'm'
(lp3467
I1
aI3
aI5
aI5
aI1
aI53
aI17
aI63
aI141
aI215
aI1223
aI3129
aI635
aI15919
asS'd'
I1734
sa(dp3468
S'a'
I6797
sS's'
I14
sS'm'
(lp3469
I1
aI3
aI3
aI1
aI23
aI31
aI25
aI11
aI195
aI241
aI995
aI3941
aI573
aI13855
asS'd'
I1735
sa(dp3470
S'a'
I6800
sS's'
I14
sS'm'
(lp3471
I1
aI3
aI3
aI7
aI17
aI13
aI71
aI203
aI465
aI479
aI1857
aI1493
aI8067
aI7113
asS'd'
I1736
sa(dp3472
S'a'
I6843
sS's'
I14
sS'm'
(lp3473
I1
aI1
aI5
aI3
aI11
aI57
aI9
aI59
aI225
aI691
aI425
aI2423
aI6031
aI6631
asS'd'
I1737
sa(dp3474
S'a'
I6845
sS's'
I14
sS'm'
(lp3475
I1
aI3
aI7
aI1
aI29
aI57
aI103
aI123
aI401
aI807
aI471
aI2759
aI5113
aI15937
asS'd'
I1738
sa(dp3476
S'a'
I6860
sS's'
I14
sS'm'
(lp3477
I1
aI3
aI1
aI1
aI3
aI1
aI67
aI123
aI157
aI655
aI519
aI323
aI1853
aI15041
asS'd'
I1739
sa(dp3478
S'a'
I6865
sS's'
I14
sS'm'
(lp3479
I1
aI1
aI7
aI5
aI11
aI11
aI105
aI135
aI247
aI689
aI1141
aI2347
aI7113
aI9347
asS'd'
I1740
sa(dp3480
S'a'
I6878
sS's'
I14
sS'm'
(lp3481
I1
aI1
aI3
aI11
aI15
aI37
aI87
aI3
aI209
aI575
aI1521
aI3863
aI3893
aI211
asS'd'
I1741
sa(dp3482
S'a'
I6887
sS's'
I14
sS'm'
(lp3483
I1
aI3
aI1
aI3
aI29
aI55
aI115
aI31
aI19
aI195
aI985
aI3275
aI363
aI9801
asS'd'
I1742
sa(dp3484
S'a'
I6888
sS's'
I14
sS'm'
(lp3485
I1
aI1
aI3
aI9
aI13
aI31
aI57
aI251
aI201
aI275
aI1751
aI389
aI1463
aI13159
asS'd'
I1743
sa(dp3486
S'a'
I6901
sS's'
I14
sS'm'
(lp3487
I1
aI3
aI5
aI15
aI19
aI51
aI127
aI255
aI397
aI243
aI29
aI3007
aI7845
aI4687
asS'd'
I1744
sa(dp3488
S'a'
I6906
sS's'
I14
sS'm'
(lp3489
I1
aI1
aI7
aI15
aI9
aI37
aI39
aI217
aI509
aI137
aI1123
aI3361
aI6323
aI5323
asS'd'
I1745
sa(dp3490
S'a'
I6940
sS's'
I14
sS'm'
(lp3491
I1
aI3
aI7
aI5
aI25
aI3
aI93
aI203
aI345
aI581
aI261
aI2811
aI4829
aI6977
asS'd'
I1746
sa(dp3492
S'a'
I6947
sS's'
I14
sS'm'
(lp3493
I1
aI1
aI7
aI1
aI15
aI41
aI51
aI227
aI447
aI893
aI1209
aI3865
aI5229
aI4277
asS'd'
I1747
sa(dp3494
S'a'
I6953
sS's'
I14
sS'm'
(lp3495
I1
aI1
aI1
aI5
aI31
aI19
aI23
aI195
aI359
aI853
aI595
aI337
aI2503
aI16371
asS'd'
I1748
sa(dp3496
S'a'
I6954
sS's'
I14
sS'm'
(lp3497
I1
aI3
aI7
aI5
aI5
aI13
aI89
aI157
aI351
aI777
aI151
aI3565
aI4219
aI7423
asS'd'
I1749
sa(dp3498
S'a'
I6959
sS's'
I14
sS'm'
(lp3499
I1
aI1
aI1
aI5
aI7
aI1
aI9
aI89
aI175
aI909
aI1523
aI2295
aI7949
aI6739
asS'd'
I1750
sa(dp3500
S'a'
I6961
sS's'
I14
sS'm'
(lp3501
I1
aI3
aI5
aI15
aI27
aI17
aI11
aI235
aI19
aI105
aI457
aI465
aI3819
aI11335
asS'd'
I1751
sa(dp3502
S'a'
I6964
sS's'
I14
sS'm'
(lp3503
I1
aI3
aI1
aI13
aI3
aI41
aI85
aI221
aI451
aI613
aI543
aI2265
aI6831
aI1725
asS'd'
I1752
sa(dp3504
S'a'
I6991
sS's'
I14
sS'm'
(lp3505
I1
aI1
aI7
aI7
aI3
aI29
aI9
aI197
aI455
aI665
aI343
aI1811
aI5395
aI393
asS'd'
I1753
sa(dp3506
S'a'
I6993
sS's'
I14
sS'm'
(lp3507
I1
aI1
aI3
aI13
aI29
aI55
aI71
aI95
aI475
aI615
aI2029
aI123
aI413
aI16127
asS'd'
I1754
sa(dp3508
S'a'
I6999
sS's'
I14
sS'm'
(lp3509
I1
aI1
aI5
aI9
aI15
aI61
aI9
aI51
aI105
aI271
aI511
aI2801
aI693
aI11839
asS'd'
I1755
sa(dp3510
S'a'
I7016
sS's'
I14
sS'm'
(lp3511
I1
aI1
aI7
aI13
aI29
aI9
aI105
aI59
aI377
aI635
aI717
aI4033
aI6963
aI10541
asS'd'
I1756
sa(dp3512
S'a'
I7029
sS's'
I14
sS'm'
(lp3513
I1
aI1
aI1
aI13
aI7
aI13
aI59
aI17
aI335
aI355
aI77
aI3665
aI7003
aI9521
asS'd'
I1757
sa(dp3514
S'a'
I7036
sS's'
I14
sS'm'
(lp3515
I1
aI3
aI1
aI1
aI23
aI43
aI51
aI209
aI151
aI365
aI1021
aI2859
aI3937
aI2899
asS'd'
I1758
sa(dp3516
S'a'
I7045
sS's'
I14
sS'm'
(lp3517
I1
aI1
aI3
aI3
aI31
aI41
aI111
aI107
aI171
aI433
aI1233
aI505
aI2971
aI6927
asS'd'
I1759
sa(dp3518
S'a'
I7076
sS's'
I14
sS'm'
(lp3519
I1
aI3
aI7
aI13
aI17
aI25
aI127
aI195
aI257
aI551
aI1867
aI2145
aI3695
aI14567
asS'd'
I1760
sa(dp3520
S'a'
I7083
sS's'
I14
sS'm'
(lp3521
I1
aI1
aI5
aI13
aI13
aI45
aI39
aI195
aI55
aI991
aI1981
aI1043
aI5875
aI581
asS'd'
I1761
sa(dp3522
S'a'
I7094
sS's'
I14
sS'm'
(lp3523
I1
aI3
aI3
aI11
aI25
aI31
aI91
aI153
aI415
aI449
aI1301
aI563
aI7755
aI10671
asS'd'
I1762
sa(dp3524
S'a'
I7097
sS's'
I14
sS'm'
(lp3525
I1
aI1
aI3
aI5
aI31
aI63
aI1
aI157
aI229
aI949
aI971
aI137
aI6589
aI8387
asS'd'
I1763
sa(dp3526
S'a'
I7123
sS's'
I14
sS'm'
(lp3527
I1
aI3
aI7
aI15
aI25
aI7
aI89
aI133
aI73
aI497
aI1361
aI613
aI455
aI1005
asS'd'
I1764
sa(dp3528
S'a'
I7130
sS's'
I14
sS'm'
(lp3529
I1
aI3
aI3
aI1
aI13
aI5
aI119
aI93
aI175
aI511
aI1923
aI763
aI7573
aI7545
asS'd'
I1765
sa(dp3530
S'a'
I7139
sS's'
I14
sS'm'
(lp3531
I1
aI1
aI3
aI15
aI27
aI59
aI49
aI205
aI497
aI485
aI117
aI2523
aI4495
aI15153
asS'd'
I1766
sa(dp3532
S'a'
I7145
sS's'
I14
sS'm'
(lp3533
I1
aI3
aI7
aI9
aI15
aI47
aI111
aI31
aI363
aI11
aI475
aI2931
aI6813
aI1259
asS'd'
I1767
sa(dp3534
S'a'
I7146
sS's'
I14
sS'm'
(lp3535
I1
aI1
aI5
aI5
aI1
aI35
aI95
aI225
aI17
aI991
aI809
aI2601
aI6455
aI13803
asS'd'
I1768
sa(dp3536
S'a'
I7178
sS's'
I14
sS'm'
(lp3537
I1
aI1
aI5
aI5
aI15
aI1
aI1
aI171
aI433
aI887
aI1813
aI3431
aI2471
aI7803
asS'd'
I1769
sa(dp3538
S'a'
I7186
sS's'
I14
sS'm'
(lp3539
I1
aI3
aI3
aI15
aI1
aI15
aI43
aI179
aI15
aI949
aI1881
aI1027
aI6989
aI8955
asS'd'
I1770
sa(dp3540
S'a'
I7192
sS's'
I14
sS'm'
(lp3541
I1
aI3
aI7
aI13
aI1
aI3
aI49
aI183
aI373
aI175
aI1733
aI913
aI929
aI1065
asS'd'
I1771
sa(dp3542
S'a'
I7198
sS's'
I14
sS'm'
(lp3543
I1
aI3
aI5
aI7
aI15
aI51
aI107
aI115
aI323
aI357
aI167
aI2069
aI7541
aI9601
asS'd'
I1772
sa(dp3544
S'a'
I7222
sS's'
I14
sS'm'
(lp3545
I1
aI1
aI3
aI5
aI5
aI21
aI31
aI107
aI21
aI299
aI1937
aI43
aI3673
aI8155
asS'd'
I1773
sa(dp3546
S'a'
I7269
sS's'
I14
sS'm'
(lp3547
I1
aI3
aI5
aI11
aI9
aI55
aI35
aI113
aI29
aI99
aI161
aI1607
aI8141
aI4951
asS'd'
I1774
sa(dp3548
S'a'
I7270
sS's'
I14
sS'm'
(lp3549
I1
aI3
aI7
aI15
aI25
aI7
aI113
aI179
aI213
aI19
aI1717
aI1027
aI2021
aI11263
asS'd'
I1775
sa(dp3550
S'a'
I7276
sS's'
I14
sS'm'
(lp3551
I1
aI1
aI5
aI1
aI31
aI33
aI85
aI111
aI67
aI95
aI2013
aI2217
aI871
aI5329
asS'd'
I1776
sa(dp3552
S'a'
I7287
sS's'
I14
sS'm'
(lp3553
I1
aI1
aI1
aI7
aI7
aI63
aI67
aI145
aI495
aI419
aI1945
aI3437
aI6255
aI151
asS'd'
I1777
sa(dp3554
S'a'
I7307
sS's'
I14
sS'm'
(lp3555
I1
aI3
aI5
aI7
aI17
aI37
aI97
aI187
aI215
aI399
aI1603
aI2195
aI5923
aI769
asS'd'
I1778
sa(dp3556
S'a'
I7315
sS's'
I14
sS'm'
(lp3557
I1
aI1
aI3
aI9
aI25
aI1
aI119
aI193
aI385
aI861
aI2005
aI2769
aI675
aI767
asS'd'
I1779
sa(dp3558
S'a'
I7334
sS's'
I14
sS'm'
(lp3559
I1
aI3
aI1
aI15
aI19
aI7
aI5
aI227
aI173
aI383
aI289
aI461
aI579
aI3689
asS'd'
I1780
sa(dp3560
S'a'
I7340
sS's'
I14
sS'm'
(lp3561
I1
aI3
aI1
aI11
aI1
aI37
aI93
aI239
aI465
aI891
aI1479
aI921
aI4439
aI15265
asS'd'
I1781
sa(dp3562
S'a'
I7351
sS's'
I14
sS'm'
(lp3563
I1
aI1
aI1
aI13
aI27
aI61
aI99
aI69
aI279
aI655
aI1853
aI1593
aI6319
aI9003
asS'd'
I1782
sa(dp3564
S'a'
I7352
sS's'
I14
sS'm'
(lp3565
I1
aI1
aI1
aI11
aI5
aI7
aI19
aI7
aI387
aI303
aI321
aI931
aI5809
aI16029
asS'd'
I1783
sa(dp3566
S'a'
I7357
sS's'
I14
sS'm'
(lp3567
I1
aI1
aI1
aI15
aI21
aI55
aI43
aI107
aI217
aI687
aI19
aI3225
aI3419
aI9991
asS'd'
I1784
sa(dp3568
S'a'
I7360
sS's'
I14
sS'm'
(lp3569
I1
aI1
aI7
aI5
aI7
aI55
aI79
aI41
aI317
aI357
aI859
aI1205
aI191
aI9395
asS'd'
I1785
sa(dp3570
S'a'
I7363
sS's'
I14
sS'm'
(lp3571
I1
aI1
aI3
aI11
aI3
aI43
aI7
aI133
aI115
aI995
aI1205
aI1055
aI4153
aI10481
asS'd'
I1786
sa(dp3572
S'a'
I7384
sS's'
I14
sS'm'
(lp3573
I1
aI1
aI7
aI11
aI31
aI57
aI53
aI9
aI459
aI223
aI1969
aI3513
aI7033
aI8505
asS'd'
I1787
sa(dp3574
S'a'
I7396
sS's'
I14
sS'm'
(lp3575
I1
aI1
aI3
aI7
aI17
aI11
aI115
aI255
aI281
aI97
aI1685
aI2039
aI2845
aI11637
asS'd'
I1788
sa(dp3576
S'a'
I7403
sS's'
I14
sS'm'
(lp3577
I1
aI3
aI7
aI1
aI23
aI41
aI69
aI199
aI53
aI105
aI657
aI1453
aI4429
aI1101
asS'd'
I1789
sa(dp3578
S'a'
I7406
sS's'
I14
sS'm'
(lp3579
I1
aI3
aI1
aI5
aI11
aI33
aI91
aI131
aI191
aI73
aI823
aI117
aI1053
aI127
asS'd'
I1790
sa(dp3580
S'a'
I7425
sS's'
I14
sS'm'
(lp3581
I1
aI3
aI7
aI11
aI7
aI3
aI21
aI65
aI187
aI103
aI1393
aI1797
aI6673
aI1409
asS'd'
I1791
sa(dp3582
S'a'
I7437
sS's'
I14
sS'm'
(lp3583
I1
aI3
aI7
aI1
aI31
aI25
aI25
aI161
aI299
aI275
aI417
aI2267
aI6861
aI1255
asS'd'
I1792
sa(dp3584
S'a'
I7445
sS's'
I14
sS'm'
(lp3585
I1
aI3
aI5
aI13
aI5
aI11
aI61
aI155
aI115
aI1001
aI747
aI889
aI3235
aI5709
asS'd'
I1793
sa(dp3586
S'a'
I7450
sS's'
I14
sS'm'
(lp3587
I1
aI3
aI7
aI7
aI7
aI1
aI97
aI177
aI507
aI273
aI1781
aI3455
aI5123
aI15607
asS'd'
I1794
sa(dp3588
S'a'
I7455
sS's'
I14
sS'm'
(lp3589
I1
aI1
aI7
aI5
aI1
aI7
aI59
aI49
aI147
aI343
aI97
aI3517
aI5611
aI8705
asS'd'
I1795
sa(dp3590
S'a'
I7461
sS's'
I14
sS'm'
(lp3591
I1
aI1
aI5
aI13
aI21
aI29
aI13
aI21
aI503
aI515
aI1217
aI3905
aI5513
aI15849
asS'd'
I1796
sa(dp3592
S'a'
I7466
sS's'
I14
sS'm'
(lp3593
I1
aI3
aI1
aI9
aI9
aI39
aI65
aI111
aI385
aI757
aI583
aI2225
aI2039
aI2817
asS'd'
I1797
sa(dp3594
S'a'
I7488
sS's'
I14
sS'm'
(lp3595
I1
aI3
aI3
aI15
aI23
aI17
aI63
aI169
aI503
aI949
aI849
aI461
aI6799
aI669
asS'd'
I1798
sa(dp3596
S'a'
I7494
sS's'
I14
sS'm'
(lp3597
I1
aI1
aI1
aI3
aI1
aI41
aI63
aI159
aI251
aI457
aI521
aI1653
aI623
aI3287
asS'd'
I1799
sa(dp3598
S'a'
I7515
sS's'
I14
sS'm'
(lp3599
I1
aI1
aI7
aI3
aI9
aI1
aI41
aI37
aI441
aI921
aI1415
aI2955
aI5841
aI1451
asS'd'
I1800
sa(dp3600
S'a'
I7517
sS's'
I14
sS'm'
(lp3601
I1
aI1
aI5
aI11
aI23
aI29
aI89
aI185
aI413
aI357
aI1131
aI2369
aI3835
aI6233
asS'd'
I1801
sa(dp3602
S'a'
I7521
sS's'
I14
sS'm'
(lp3603
I1
aI1
aI5
aI15
aI27
aI35
aI17
aI73
aI315
aI911
aI1761
aI797
aI5349
aI3219
asS'd'
I1802
sa(dp3604
S'a'
I7536
sS's'
I14
sS'm'
(lp3605
I1
aI3
aI7
aI11
aI21
aI9
aI119
aI233
aI249
aI901
aI189
aI3625
aI2691
aI16201
asS'd'
I1803
sa(dp3606
S'a'
I7546
sS's'
I14
sS'm'
(lp3607
I1
aI3
aI3
aI13
aI29
aI61
aI105
aI145
aI187
aI79
aI609
aI321
aI4289
aI3933
asS'd'
I1804
sa(dp3608
S'a'
I7569
sS's'
I14
sS'm'
(lp3609
I1
aI3
aI1
aI15
aI19
aI63
aI13
aI185
aI115
aI219
aI1021
aI1205
aI4273
aI11521
asS'd'
I1805
sa(dp3610
S'a'
I7591
sS's'
I14
sS'm'
(lp3611
I1
aI1
aI3
aI3
aI23
aI31
aI93
aI153
aI87
aI947
aI1039
aI469
aI4047
aI8869
asS'd'
I1806
sa(dp3612
S'a'
I7592
sS's'
I14
sS'm'
(lp3613
I1
aI1
aI1
aI1
aI9
aI1
aI85
aI3
aI15
aI995
aI455
aI2769
aI6781
aI16203
asS'd'
I1807
sa(dp3614
S'a'
I7598
sS's'
I14
sS'm'
(lp3615
I1
aI1
aI3
aI3
aI13
aI7
aI55
aI215
aI185
aI367
aI765
aI441
aI4497
aI1521
asS'd'
I1808
sa(dp3616
S'a'
I7612
sS's'
I14
sS'm'
(lp3617
I1
aI1
aI1
aI5
aI1
aI31
aI13
aI95
aI417
aI735
aI975
aI3407
aI4871
aI16133
asS'd'
I1809
sa(dp3618
S'a'
I7623
sS's'
I14
sS'm'
(lp3619
I1
aI1
aI3
aI3
aI5
aI43
aI111
aI107
aI419
aI515
aI1075
aI3597
aI1187
aI4143
asS'd'
I1810
sa(dp3620
S'a'
I7632
sS's'
I14
sS'm'
(lp3621
I1
aI1
aI3
aI13
aI31
aI51
aI83
aI163
aI489
aI887
aI863
aI599
aI9
aI13861
asS'd'
I1811
sa(dp3622
S'a'
I7637
sS's'
I14
sS'm'
(lp3623
I1
aI3
aI3
aI3
aI19
aI27
aI91
aI115
aI103
aI969
aI593
aI3667
aI1867
aI15433
asS'd'
I1812
sa(dp3624
S'a'
I7644
sS's'
I14
sS'm'
(lp3625
I1
aI3
aI3
aI13
aI7
aI25
aI47
aI141
aI57
aI553
aI1785
aI1709
aI7453
aI2209
asS'd'
I1813
sa(dp3626
S'a'
I7657
sS's'
I14
sS'm'
(lp3627
I1
aI3
aI1
aI13
aI11
aI13
aI71
aI219
aI5
aI451
aI2043
aI1605
aI6439
aI12203
asS'd'
I1814
sa(dp3628
S'a'
I7665
sS's'
I14
sS'm'
(lp3629
I1
aI3
aI1
aI13
aI5
aI57
aI61
aI223
aI401
aI413
aI321
aI1365
aI619
aI12477
asS'd'
I1815
sa(dp3630
S'a'
I7672
sS's'
I14
sS'm'
(lp3631
I1
aI3
aI1
aI5
aI25
aI57
aI89
aI211
aI195
aI455
aI1165
aI3979
aI6313
aI5751
asS'd'
I1816
sa(dp3632
S'a'
I7682
sS's'
I14
sS'm'
(lp3633
I1
aI1
aI1
aI9
aI31
aI23
aI71
aI145
aI89
aI285
aI1593
aI1171
aI5685
aI15459
asS'd'
I1817
sa(dp3634
S'a'
I7699
sS's'
I14
sS'm'
(lp3635
I1
aI3
aI7
aI7
aI9
aI41
aI65
aI251
aI65
aI137
aI1577
aI3027
aI5555
aI2865
asS'd'
I1818
sa(dp3636
S'a'
I7702
sS's'
I14
sS'm'
(lp3637
I1
aI1
aI5
aI13
aI27
aI5
aI125
aI21
aI171
aI647
aI983
aI2921
aI6623
aI5695
asS'd'
I1819
sa(dp3638
S'a'
I7724
sS's'
I14
sS'm'
(lp3639
I1
aI1
aI1
aI13
aI15
aI9
aI117
aI197
aI123
aI953
aI1191
aI3657
aI5757
aI15957
asS'd'
I1820
sa(dp3640
S'a'
I7749
sS's'
I14
sS'm'
(lp3641
I1
aI1
aI3
aI7
aI29
aI13
aI5
aI175
aI395
aI127
aI679
aI255
aI6055
aI7639
asS'd'
I1821
sa(dp3642
S'a'
I7753
sS's'
I14
sS'm'
(lp3643
I1
aI3
aI7
aI15
aI15
aI51
aI77
aI147
aI319
aI147
aI1775
aI3983
aI3175
aI5723
asS'd'
I1822
sa(dp3644
S'a'
I7754
sS's'
I14
sS'm'
(lp3645
I1
aI3
aI3
aI3
aI7
aI11
aI119
aI41
aI43
aI153
aI975
aI679
aI3081
aI10359
asS'd'
I1823
sa(dp3646
S'a'
I7761
sS's'
I14
sS'm'
(lp3647
I1
aI1
aI5
aI13
aI3
aI7
aI65
aI67
aI63
aI399
aI1561
aI2789
aI2083
aI12289
asS'd'
I1824
sa(dp3648
S'a'
I7771
sS's'
I14
sS'm'
(lp3649
I1
aI1
aI7
aI3
aI19
aI53
aI103
aI67
aI35
aI865
aI161
aI93
aI2533
aI3851
asS'd'
I1825
sa(dp3650
S'a'
I7777
sS's'
I14
sS'm'
(lp3651
I1
aI1
aI1
aI11
aI31
aI9
aI29
aI189
aI199
aI817
aI1571
aI395
aI345
aI3777
asS'd'
I1826
sa(dp3652
S'a'
I7784
sS's'
I14
sS'm'
(lp3653
I1
aI3
aI5
aI11
aI31
aI3
aI9
aI67
aI277
aI735
aI181
aI2777
aI3009
aI7233
asS'd'
I1827
sa(dp3654
S'a'
I7804
sS's'
I14
sS'm'
(lp3655
I1
aI1
aI3
aI3
aI17
aI7
aI17
aI3
aI375
aI933
aI237
aI3919
aI5409
aI3355
asS'd'
I1828
sa(dp3656
S'a'
I7807
sS's'
I14
sS'm'
(lp3657
I1
aI3
aI3
aI5
aI9
aI27
aI19
aI77
aI221
aI3
aI1965
aI309
aI3001
aI15977
asS'd'
I1829
sa(dp3658
S'a'
I7808
sS's'
I14
sS'm'
(lp3659
I1
aI1
aI5
aI1
aI3
aI33
aI35
aI133
aI37
aI709
aI627
aI1705
aI2525
aI4307
asS'd'
I1830
sa(dp3660
S'a'
I7818
sS's'
I14
sS'm'
(lp3661
I1
aI1
aI7
aI3
aI25
aI21
aI105
aI55
aI375
aI681
aI881
aI1299
aI5879
aI459
asS'd'
I1831
sa(dp3662
S'a'
I7835
sS's'
I14
sS'm'
(lp3663
I1
aI3
aI7
aI1
aI13
aI7
aI113
aI103
aI313
aI515
aI1041
aI3683
aI4619
aI5093
asS'd'
I1832
sa(dp3664
S'a'
I7842
sS's'
I14
sS'm'
(lp3665
I1
aI1
aI3
aI7
aI19
aI43
aI83
aI37
aI39
aI133
aI1759
aI1171
aI1521
aI13717
asS'd'
I1833
sa(dp3666
S'a'
I7865
sS's'
I14
sS'm'
(lp3667
I1
aI1
aI7
aI13
aI7
aI35
aI15
aI155
aI293
aI1001
aI157
aI3883
aI405
aI1797
asS'd'
I1834
sa(dp3668
S'a'
I7868
sS's'
I14
sS'm'
(lp3669
I1
aI1
aI3
aI3
aI13
aI19
aI125
aI49
aI333
aI387
aI339
aI1815
aI4503
aI7359
asS'd'
I1835
sa(dp3670
S'a'
I7880
sS's'
I14
sS'm'
(lp3671
I1
aI1
aI3
aI13
aI19
aI19
aI105
aI225
aI151
aI27
aI1251
aI885
aI4815
aI7863
asS'd'
I1836
sa(dp3672
S'a'
I7883
sS's'
I14
sS'm'
(lp3673
I1
aI1
aI1
aI5
aI7
aI59
aI17
aI145
aI77
aI117
aI1355
aI1429
aI2301
aI16177
asS'd'
I1837
sa(dp3674
S'a'
I7891
sS's'
I14
sS'm'
(lp3675
I1
aI3
aI3
aI13
aI5
aI31
aI119
aI167
aI459
aI727
aI1799
aI2537
aI695
aI13637
asS'd'
I1838
sa(dp3676
S'a'
I7897
sS's'
I14
sS'm'
(lp3677
I1
aI3
aI3
aI3
aI27
aI51
aI107
aI85
aI267
aI57
aI1279
aI823
aI6247
aI3603
asS'd'
I1839
sa(dp3678
S'a'
I7907
sS's'
I14
sS'm'
(lp3679
I1
aI1
aI7
aI15
aI29
aI17
aI67
aI197
aI215
aI465
aI109
aI3461
aI5269
aI15287
asS'd'
I1840
sa(dp3680
S'a'
I7910
sS's'
I14
sS'm'
(lp3681
I1
aI1
aI3
aI5
aI11
aI15
aI123
aI53
aI293
aI797
aI1105
aI1777
aI6509
aI217
asS'd'
I1841
sa(dp3682
S'a'
I7924
sS's'
I14
sS'm'
(lp3683
I1
aI3
aI3
aI13
aI3
aI5
aI109
aI53
aI203
aI693
aI871
aI135
aI369
aI11149
asS'd'
I1842
sa(dp3684
S'a'
I7933
sS's'
I14
sS'm'
(lp3685
I1
aI3
aI5
aI15
aI17
aI43
aI81
aI235
aI119
aI817
aI1777
aI261
aI8049
aI4251
asS'd'
I1843
sa(dp3686
S'a'
I7934
sS's'
I14
sS'm'
(lp3687
I1
aI1
aI3
aI7
aI7
aI13
aI87
aI99
aI481
aI931
aI1507
aI651
aI5267
aI8281
asS'd'
I1844
sa(dp3688
S'a'
I7942
sS's'
I14
sS'm'
(lp3689
I1
aI3
aI1
aI13
aI27
aI43
aI77
aI225
aI341
aI163
aI933
aI429
aI4943
aI7781
asS'd'
I1845
sa(dp3690
S'a'
I7948
sS's'
I14
sS'm'
(lp3691
I1
aI1
aI7
aI1
aI1
aI49
aI85
aI211
aI449
aI479
aI1395
aI787
aI5653
aI14891
asS'd'
I1846
sa(dp3692
S'a'
I7959
sS's'
I14
sS'm'
(lp3693
I1
aI1
aI5
aI9
aI25
aI13
aI49
aI85
aI125
aI85
aI1281
aI3365
aI4305
aI11791
asS'd'
I1847
sa(dp3694
S'a'
I7984
sS's'
I14
sS'm'
(lp3695
I1
aI3
aI1
aI13
aI3
aI31
aI117
aI39
aI43
aI151
aI663
aI669
aI1571
aI5207
asS'd'
I1848
sa(dp3696
S'a'
I7994
sS's'
I14
sS'm'
(lp3697
I1
aI3
aI7
aI15
aI17
aI7
aI79
aI163
aI37
aI841
aI1799
aI1787
aI4501
aI3785
asS'd'
I1849
sa(dp3698
S'a'
I7999
sS's'
I14
sS'm'
(lp3699
I1
aI1
aI3
aI9
aI1
aI23
aI67
aI191
aI449
aI931
aI1521
aI2705
aI887
aI7037
asS'd'
I1850
sa(dp3700
S'a'
I8014
sS's'
I14
sS'm'
(lp3701
I1
aI1
aI1
aI1
aI5
aI13
aI55
aI161
aI419
aI577
aI1703
aI2589
aI2651
aI2873
asS'd'
I1851
sa(dp3702
S'a'
I8021
sS's'
I14
sS'm'
(lp3703
I1
aI3
aI3
aI3
aI5
aI19
aI37
aI169
aI69
aI1003
aI1755
aI3101
aI1469
aI8583
asS'd'
I1852
sa(dp3704
S'a'
I8041
sS's'
I14
sS'm'
(lp3705
I1
aI1
aI1
aI1
aI11
aI33
aI105
aI79
aI283
aI91
aI299
aI835
aI3193
aI5593
asS'd'
I1853
sa(dp3706
S'a'
I8049
sS's'
I14
sS'm'
(lp3707
I1
aI3
aI3
aI13
aI25
aI21
aI81
aI213
aI465
aI475
aI331
aI457
aI61
aI9511
asS'd'
I1854
sa(dp3708
S'a'
I8050
sS's'
I14
sS'm'
(lp3709
I1
aI1
aI3
aI11
aI1
aI11
aI77
aI95
aI455
aI949
aI1999
aI1833
aI1275
aI5631
asS'd'
I1855
sa(dp3710
S'a'
I8068
sS's'
I14
sS'm'
(lp3711
I1
aI1
aI1
aI1
aI15
aI25
aI51
aI137
aI275
aI451
aI1179
aI3595
aI5177
aI7105
asS'd'
I1856
sa(dp3712
S'a'
I8080
sS's'
I14
sS'm'
(lp3713
I1
aI3
aI3
aI3
aI3
aI59
aI79
aI143
aI393
aI583
aI349
aI3039
aI7079
aI14245
asS'd'
I1857
sa(dp3714
S'a'
I8095
sS's'
I14
sS'm'
(lp3715
I1
aI1
aI7
aI9
aI21
aI11
aI123
aI105
aI53
aI297
aI803
aI4025
aI5421
aI14527
asS'd'
I1858
sa(dp3716
S'a'
I8102
sS's'
I14
sS'm'
(lp3717
I1
aI3
aI7
aI11
aI21
aI15
aI103
aI109
aI311
aI321
aI1217
aI2777
aI5457
aI1823
asS'd'
I1859
sa(dp3718
S'a'
I8106
sS's'
I14
sS'm'
(lp3719
I1
aI3
aI5
aI11
aI19
aI31
aI79
aI89
aI295
aI413
aI817
aI499
aI3699
aI14411
asS'd'
I1860
sa(dp3720
S'a'
I8120
sS's'
I14
sS'm'
(lp3721
I1
aI1
aI1
aI5
aI11
aI3
aI81
aI13
aI315
aI841
aI1543
aI411
aI6883
aI6347
asS'd'
I1861
sa(dp3722
S'a'
I8133
sS's'
I14
sS'm'
(lp3723
I1
aI3
aI3
aI11
aI23
aI43
aI23
aI131
aI17
aI517
aI995
aI2687
aI7443
aI15085
asS'd'
I1862
sa(dp3724
S'a'
I8134
sS's'
I14
sS'm'
(lp3725
I1
aI1
aI1
aI1
aI11
aI57
aI73
aI9
aI123
aI905
aI1763
aI1789
aI3701
aI7131
asS'd'
I1863
sa(dp3726
S'a'
I8143
sS's'
I14
sS'm'
(lp3727
I1
aI1
aI3
aI5
aI9
aI53
aI99
aI229
aI43
aI207
aI625
aI1583
aI6727
aI15249
asS'd'
I1864
sa(dp3728
S'a'
I8162
sS's'
I14
sS'm'
(lp3729
I1
aI1
aI7
aI7
aI17
aI39
aI91
aI1
aI297
aI711
aI225
aI513
aI7391
aI291
asS'd'
I1865
sa(dp3730
S'a'
I8168
sS's'
I14
sS'm'
(lp3731
I1
aI1
aI7
aI11
aI7
aI55
aI111
aI129
aI423
aI521
aI1807
aI3015
aI1449
aI12321
asS'd'
I1866
sa(dp3732
S'a'
I8179
sS's'
I14
sS'm'
(lp3733
I1
aI3
aI7
aI3
aI13
aI9
aI125
aI187
aI11
aI485
aI647
aI275
aI3495
aI11989
asS'd'
I1867
sa(dp3734
S'a'
I1
sS's'
I15
sS'm'
(lp3735
I1
aI1
aI3
aI11
aI11
aI25
aI49
aI33
aI361
aI105
aI271
aI3841
aI4837
aI2437
aI30181
asS'd'
I1868
sa(dp3736
S'a'
I8
sS's'
I15
sS'm'
(lp3737
I1
aI3
aI5
aI1
aI27
aI15
aI119
aI35
aI159
aI273
aI1489
aI3157
aI5433
aI3337
aI26859
asS'd'
I1869
sa(dp3738
S'a'
I11
sS's'
I15
sS'm'
(lp3739
I1
aI3
aI5
aI13
aI23
aI31
aI97
aI145
aI41
aI605
aI1455
aI59
aI5389
aI5527
aI14447
asS'd'
I1870
sa(dp3740
S'a'
I22
sS's'
I15
sS'm'
(lp3741
I1
aI1
aI7
aI9
aI7
aI41
aI61
aI193
aI353
aI879
aI1805
aI581
aI5447
aI11177
aI7331
asS'd'
I1871
sa(dp3742
S'a'
I26
sS's'
I15
sS'm'
(lp3743
I1
aI1
aI7
aI11
aI29
aI19
aI55
aI207
aI361
aI759
aI63
aI2255
aI2119
aI14671
aI21783
asS'd'
I1872
sa(dp3744
S'a'
I47
sS's'
I15
sS'm'
(lp3745
I1
aI3
aI1
aI13
aI17
aI7
aI73
aI179
aI103
aI23
aI917
aI1205
aI4925
aI1691
aI5419
asS'd'
I1873
sa(dp3746
S'a'
I59
sS's'
I15
sS'm'
(lp3747
I1
aI3
aI5
aI3
aI15
aI3
aI9
aI109
aI227
aI861
aI867
aI3529
aI1535
aI489
aI22873
asS'd'
I1874
sa(dp3748
S'a'
I64
sS's'
I15
sS'm'
(lp3749
I1
aI3
aI3
aI9
aI15
aI15
aI95
aI193
aI385
aI997
aI1525
aI1865
aI1425
aI4079
aI14771
asS'd'
I1875
sa(dp3750
S'a'
I67
sS's'
I15
sS'm'
(lp3751
I1
aI1
aI3
aI5
aI5
aI29
aI49
aI171
aI171
aI623
aI1167
aI3743
aI1809
aI12009
aI7043
asS'd'
I1876
sa(dp3752
S'a'
I73
sS's'
I15
sS'm'
(lp3753
I1
aI3
aI7
aI5
aI23
aI11
aI87
aI183
aI299
aI555
aI1857
aI489
aI3505
aI9161
aI28763
asS'd'
I1877
sa(dp3754
S'a'
I82
sS's'
I15
sS'm'
(lp3755
I1
aI3
aI5
aI9
aI19
aI21
aI85
aI127
aI337
aI439
aI1183
aI1891
aI1877
aI4373
aI10451
asS'd'
I1878
sa(dp3756
S'a'
I97
sS's'
I15
sS'm'
(lp3757
I1
aI3
aI7
aI13
aI27
aI17
aI29
aI83
aI463
aI385
aI1167
aI3453
aI4523
aI4759
aI9321
asS'd'
I1879
sa(dp3758
S'a'
I103
sS's'
I15
sS'm'
(lp3759
I1
aI1
aI3
aI7
aI21
aI59
aI65
aI83
aI177
aI763
aI317
aI2913
aI7527
aI5967
aI17167
asS'd'
I1880
sa(dp3760
S'a'
I110
sS's'
I15
sS'm'
(lp3761
I1
aI1
aI7
aI15
aI13
aI27
aI49
aI35
aI253
aI101
aI1699
aI355
aI2181
aI10859
aI24221
asS'd'
I1881
sa(dp3762
S'a'
I115
sS's'
I15
sS'm'
(lp3763
I1
aI1
aI5
aI1
aI17
aI17
aI81
aI91
aI349
aI655
aI1373
aI2225
aI945
aI899
aI31801
asS'd'
I1882
sa(dp3764
S'a'
I122
sS's'
I15
sS'm'
(lp3765
I1
aI3
aI7
aI11
aI5
aI1
aI81
aI53
aI215
aI587
aI167
aI4045
aI5671
aI5597
aI13529
asS'd'
I1883
sa(dp3766
S'a'
I128
sS's'
I15
sS'm'
(lp3767
I1
aI3
aI5
aI15
aI1
aI9
aI59
aI235
aI315
aI195
aI909
aI2237
aI505
aI10415
aI28145
asS'd'
I1884
sa(dp3768
S'a'
I138
sS's'
I15
sS'm'
(lp3769
I1
aI1
aI1
aI3
aI9
aI31
aI41
aI43
aI275
aI921
aI25
aI671
aI5737
aI11241
aI4193
asS'd'
I1885
sa(dp3770
S'a'
I146
sS's'
I15
sS'm'
(lp3771
I1
aI3
aI3
aI13
aI29
aI13
aI95
aI213
aI317
aI995
aI1489
aI3779
aI3043
aI8569
aI28823
asS'd'
I1886
sa(dp3772
S'a'
I171
sS's'
I15
sS'm'
(lp3773
I1
aI1
aI7
aI5
aI9
aI49
aI125
aI241
aI87
aI153
aI1673
aI3849
aI7253
aI1715
aI11627
asS'd'
I1887
sa(dp3774
S'a'
I174
sS's'
I15
sS'm'
(lp3775
I1
aI1
aI3
aI9
aI27
aI27
aI19
aI223
aI63
aI463
aI1095
aI1395
aI6643
aI11589
aI2145
asS'd'
I1888
sa(dp3776
S'a'
I176
sS's'
I15
sS'm'
(lp3777
I1
aI1
aI3
aI15
aI21
aI17
aI45
aI23
aI357
aI11
aI1307
aI1791
aI2481
aI2123
aI24341
asS'd'
I1889
sa(dp3778
S'a'
I182
sS's'
I15
sS'm'
(lp3779
I1
aI3
aI5
aI15
aI31
aI53
aI117
aI51
aI433
aI193
aI1239
aI3329
aI2403
aI12745
aI32219
asS'd'
I1890
sa(dp3780
S'a'
I194
sS's'
I15
sS'm'
(lp3781
I1
aI1
aI5
aI9
aI7
aI27
aI9
aI115
aI417
aI579
aI83
aI173
aI4717
aI15665
aI27463
asS'd'
I1891
sa(dp3782
S'a'
I208
sS's'
I15
sS'm'
(lp3783
I1
aI3
aI5
aI7
aI9
aI9
aI31
aI35
aI249
aI567
aI331
aI905
aI5101
aI14817
aI14255
asS'd'
I1892
sa(dp3784
S'a'
I211
sS's'
I15
sS'm'
(lp3785
I1
aI3
aI7
aI3
aI1
aI61
aI29
aI129
aI119
aI421
aI1597
aI2987
aI3041
aI7629
aI23451
asS'd'
I1893
sa(dp3786
S'a'
I220
sS's'
I15
sS'm'
(lp3787
I1
aI1
aI7
aI9
aI13
aI1
aI99
aI105
aI107
aI509
aI989
aI2259
aI1009
aI6827
aI8903
asS'd'
I1894
sa(dp3788
S'a'
I229
sS's'
I15
sS'm'
(lp3789
I1
aI3
aI5
aI15
aI11
aI29
aI85
aI29
aI265
aI105
aI2035
aI3349
aI3543
aI13903
aI10213
asS'd'
I1895
sa(dp3790
S'a'
I230
sS's'
I15
sS'm'
(lp3791
I1
aI3
aI1
aI1
aI25
aI19
aI53
aI139
aI467
aI485
aI491
aI3067
aI7353
aI13861
aI25819
asS'd'
I1896
sa(dp3792
S'a'
I239
sS's'
I15
sS'm'
(lp3793
I1
aI1
aI5
aI3
aI3
aI43
aI41
aI185
aI45
aI463
aI351
aI2823
aI2519
aI6705
aI11395
asS'd'
I1897
sa(dp3794
S'a'
I254
sS's'
I15
sS'm'
(lp3795
I1
aI3
aI7
aI13
aI11
aI15
aI87
aI221
aI427
aI673
aI1631
aI599
aI3259
aI10691
aI31283
asS'd'
I1898
sa(dp3796
S'a'
I265
sS's'
I15
sS'm'
(lp3797
I1
aI3
aI5
aI11
aI9
aI9
aI15
aI49
aI275
aI335
aI1613
aI3587
aI5309
aI14849
aI26475
asS'd'
I1899
sa(dp3798
S'a'
I285
sS's'
I15
sS'm'
(lp3799
I1
aI3
aI7
aI9
aI29
aI13
aI79
aI225
aI381
aI781
aI1411
aI2761
aI7157
aI14983
aI19717
asS'd'
I1900
sa(dp3800
S'a'
I290
sS's'
I15
sS'm'
(lp3801
I1
aI1
aI7
aI11
aI29
aI25
aI117
aI183
aI101
aI651
aI653
aI3157
aI445
aI14389
aI23293
asS'd'
I1901
sa(dp3802
S'a'
I319
sS's'
I15
sS'm'
(lp3803
I1
aI1
aI1
aI3
aI5
aI33
aI73
aI155
aI473
aI387
aI591
aI2045
aI5965
aI16299
aI31499
asS'd'
I1902
sa(dp3804
S'a'
I324
sS's'
I15
sS'm'
(lp3805
I1
aI3
aI1
aI7
aI11
aI33
aI29
aI21
aI491
aI937
aI729
aI4075
aI975
aI2461
aI18991
asS'd'
I1903
sa(dp3806
S'a'
I327
sS's'
I15
sS'm'
(lp3807
I1
aI3
aI7
aI15
aI29
aI39
aI105
aI111
aI173
aI943
aI69
aI295
aI8175
aI13037
aI26131
asS'd'
I1904
sa(dp3808
S'a'
I333
sS's'
I15
sS'm'
(lp3809
I1
aI1
aI5
aI15
aI7
aI5
aI97
aI147
aI105
aI887
aI443
aI2595
aI5889
aI10753
aI1619
asS'd'
I1905
sa(dp3810
S'a'
I357
sS's'
I15
sS'm'
(lp3811
I1
aI3
aI3
aI15
aI11
aI45
aI87
aI207
aI353
aI909
aI1847
aI323
aI2283
aI12885
aI16415
asS'd'
I1906
sa(dp3812
S'a'
I364
sS's'
I15
sS'm'
(lp3813
I1
aI1
aI5
aI3
aI19
aI33
aI43
aI79
aI115
aI653
aI359
aI2873
aI4609
aI12439
aI6339
asS'd'
I1907
sa(dp3814
S'a'
I395
sS's'
I15
sS'm'
(lp3815
I1
aI3
aI7
aI9
aI17
aI61
aI49
aI227
aI291
aI69
aI1753
aI3899
aI483
aI3187
aI29041
asS'd'
I1908
sa(dp3816
S'a'
I397
sS's'
I15
sS'm'
(lp3817
I1
aI3
aI5
aI3
aI25
aI35
aI61
aI211
aI393
aI199
aI691
aI1779
aI6295
aI13371
aI15817
asS'd'
I1909
sa(dp3818
S'a'
I405
sS's'
I15
sS'm'
(lp3819
I1
aI3
aI7
aI5
aI7
aI23
aI37
aI91
aI245
aI915
aI579
aI867
aI6193
aI1063
aI17363
asS'd'
I1910
sa(dp3820
S'a'
I409
sS's'
I15
sS'm'
(lp3821
I1
aI3
aI7
aI7
aI23
aI51
aI41
aI63
aI375
aI3
aI159
aI1889
aI4419
aI1687
aI17977
asS'd'
I1911
sa(dp3822
S'a'
I419
sS's'
I15
sS'm'
(lp3823
I1
aI1
aI1
aI7
aI13
aI11
aI53
aI43
aI317
aI325
aI1749
aI2423
aI4123
aI8595
aI20773
asS'd'
I1912
sa(dp3824
S'a'
I422
sS's'
I15
sS'm'
(lp3825
I1
aI1
aI7
aI7
aI9
aI9
aI61
aI113
aI437
aI213
aI1407
aI645
aI4345
aI807
aI30411
asS'd'
I1913
sa(dp3826
S'a'
I431
sS's'
I15
sS'm'
(lp3827
I1
aI3
aI3
aI11
aI17
aI39
aI17
aI113
aI391
aI385
aI581
aI2023
aI7449
aI10153
aI22033
asS'd'
I1914
sa(dp3828
S'a'
I433
sS's'
I15
sS'm'
(lp3829
I1
aI1
aI3
aI5
aI29
aI31
aI101
aI215
aI379
aI377
aI1113
aI2855
aI7147
aI14377
aI25515
asS'd'
I1915
sa(dp3830
S'a'
I436
sS's'
I15
sS'm'
(lp3831
I1
aI3
aI5
aI5
aI13
aI3
aI121
aI125
aI227
aI969
aI11
aI1115
aI5657
aI9209
aI6117
asS'd'
I1916
sa(dp3832
S'a'
I440
sS's'
I15
sS'm'
(lp3833
I1
aI3
aI7
aI15
aI29
aI17
aI33
aI123
aI317
aI301
aI749
aI1365
aI5619
aI605
aI1613
asS'd'
I1917
sa(dp3834
S'a'
I453
sS's'
I15
sS'm'
(lp3835
I1
aI3
aI1
aI15
aI7
aI53
aI125
aI249
aI219
aI655
aI105
aI2825
aI1649
aI12783
aI19777
asS'd'
I1918
sa(dp3836
S'a'
I460
sS's'
I15
sS'm'
(lp3837
I1
aI1
aI7
aI1
aI25
aI53
aI19
aI53
aI157
aI373
aI1855
aI495
aI5065
aI9465
aI2313
asS'd'
I1919
sa(dp3838
S'a'
I471
sS's'
I15
sS'm'
(lp3839
I1
aI3
aI5
aI13
aI3
aI57
aI57
aI161
aI431
aI415
aI1859
aI1033
aI6349
aI1577
aI31579
asS'd'
I1920
sa(dp3840
S'a'
I478
sS's'
I15
sS'm'
(lp3841
I1
aI1
aI7
aI5
aI23
aI63
aI29
aI221
aI13
aI965
aI1997
aI2265
aI1583
aI10491
aI9551
asS'd'
I1921
sa(dp3842
S'a'
I482
sS's'
I15
sS'm'
(lp3843
I1
aI1
aI3
aI13
aI31
aI25
aI23
aI61
aI285
aI5
aI2005
aI879
aI795
aI13299
aI19685
asS'd'
I1922
sa(dp3844
S'a'
I488
sS's'
I15
sS'm'
(lp3845
I1
aI1
aI7
aI1
aI21
aI45
aI121
aI89
aI263
aI543
aI1333
aI2711
aI219
aI10823
aI26139
asS'd'
I1923
sa(dp3846
S'a'
I524
sS's'
I15
sS'm'
(lp3847
I1
aI1
aI3
aI3
aI27
aI13
aI19
aI117
aI161
aI457
aI1541
aI295
aI4953
aI12125
aI14503
asS'd'
I1924
sa(dp3848
S'a'
I529
sS's'
I15
sS'm'
(lp3849
I1
aI3
aI5
aI3
aI7
aI63
aI13
aI247
aI439
aI681
aI977
aI2537
aI6923
aI10323
aI7349
asS'd'
I1925
sa(dp3850
S'a'
I535
sS's'
I15
sS'm'
(lp3851
I1
aI3
aI5
aI9
aI3
aI51
aI81
aI251
aI349
aI983
aI581
aI2515
aI2281
aI2849
aI31915
asS'd'
I1926
sa(dp3852
S'a'
I536
sS's'
I15
sS'm'
(lp3853
I1
aI3
aI5
aI3
aI11
aI63
aI47
aI137
aI303
aI627
aI91
aI2269
aI7097
aI2145
aI31059
asS'd'
I1927
sa(dp3854
S'a'
I539
sS's'
I15
sS'm'
(lp3855
I1
aI1
aI3
aI15
aI13
aI17
aI53
aI27
aI133
aI13
aI117
aI1837
aI4103
aI5843
aI29153
asS'd'
I1928
sa(dp3856
S'a'
I563
sS's'
I15
sS'm'
(lp3857
I1
aI1
aI5
aI13
aI21
aI33
aI37
aI253
aI465
aI209
aI309
aI49
aI3209
aI15677
aI14569
asS'd'
I1929
sa(dp3858
S'a'
I566
sS's'
I15
sS'm'
(lp3859
I1
aI1
aI7
aI15
aI13
aI21
aI33
aI203
aI499
aI141
aI1155
aI3893
aI1663
aI2115
aI27459
asS'd'
I1930
sa(dp3860
S'a'
I572
sS's'
I15
sS'm'
(lp3861
I1
aI3
aI5
aI11
aI21
aI9
aI39
aI157
aI257
aI273
aI1257
aI1831
aI515
aI7969
aI20133
asS'd'
I1931
sa(dp3862
S'a'
I577
sS's'
I15
sS'm'
(lp3863
I1
aI1
aI3
aI13
aI19
aI29
aI15
aI189
aI103
aI219
aI1395
aI517
aI7425
aI6585
aI15865
asS'd'
I1932
sa(dp3864
S'a'
I587
sS's'
I15
sS'm'
(lp3865
I1
aI1
aI5
aI11
aI21
aI31
aI49
aI151
aI39
aI537
aI1783
aI3449
aI6915
aI223
aI11587
asS'd'
I1933
sa(dp3866
S'a'
I592
sS's'
I15
sS'm'
(lp3867
I1
aI3
aI3
aI11
aI7
aI63
aI69
aI31
aI27
aI911
aI1903
aI2821
aI7977
aI12949
aI32257
asS'd'
I1934
sa(dp3868
S'a'
I602
sS's'
I15
sS'm'
(lp3869
I1
aI1
aI7
aI9
aI25
aI45
aI23
aI233
aI511
aI595
aI1383
aI1721
aI6789
aI12055
aI21179
asS'd'
I1935
sa(dp3870
S'a'
I623
sS's'
I15
sS'm'
(lp3871
I1
aI1
aI7
aI13
aI1
aI27
aI123
aI49
aI439
aI683
aI501
aI641
aI1947
aI6111
aI25423
asS'd'
I1936
sa(dp3872
S'a'
I635
sS's'
I15
sS'm'
(lp3873
I1
aI3
aI3
aI5
aI1
aI23
aI57
aI241
aI243
aI593
aI2039
aI1617
aI2209
aI5171
aI9675
asS'd'
I1937
sa(dp3874
S'a'
I638
sS's'
I15
sS'm'
(lp3875
I1
aI1
aI1
aI7
aI5
aI19
aI83
aI55
aI481
aI125
aI177
aI1021
aI1139
aI11403
aI23099
asS'd'
I1938
sa(dp3876
S'a'
I654
sS's'
I15
sS'm'
(lp3877
I1
aI1
aI3
aI5
aI29
aI39
aI33
aI217
aI461
aI907
aI733
aI3795
aI4811
aI12939
aI27715
asS'd'
I1939
sa(dp3878
S'a'
I656
sS's'
I15
sS'm'
(lp3879
I1
aI3
aI7
aI3
aI7
aI11
aI39
aI165
aI495
aI147
aI999
aI1827
aI817
aI603
aI9293
asS'd'
I1940
sa(dp3880
S'a'
I659
sS's'
I15
sS'm'
(lp3881
I1
aI3
aI7
aI15
aI25
aI53
aI35
aI15
aI431
aI733
aI1213
aI2907
aI8087
aI3939
aI27363
asS'd'
I1941
sa(dp3882
S'a'
I665
sS's'
I15
sS'm'
(lp3883
I1
aI3
aI7
aI13
aI13
aI9
aI33
aI27
aI485
aI183
aI455
aI3341
aI2555
aI4985
aI8793
asS'd'
I1942
sa(dp3884
S'a'
I675
sS's'
I15
sS'm'
(lp3885
I1
aI1
aI1
aI15
aI25
aI47
aI75
aI21
aI205
aI15
aI1639
aI3067
aI1295
aI11693
aI16903
asS'd'
I1943
sa(dp3886
S'a'
I677
sS's'
I15
sS'm'
(lp3887
I1
aI1
aI1
aI15
aI3
aI31
aI93
aI57
aI43
aI185
aI251
aI1899
aI7885
aI10829
aI3609
asS'd'
I1944
sa(dp3888
S'a'
I687
sS's'
I15
sS'm'
(lp3889
I1
aI1
aI3
aI1
aI29
aI9
aI69
aI223
aI221
aI537
aI365
aI3411
aI5771
aI15279
aI5309
asS'd'
I1945
sa(dp3890
S'a'
I696
sS's'
I15
sS'm'
(lp3891
I1
aI1
aI7
aI5
aI1
aI5
aI125
aI243
aI213
aI1003
aI1571
aI3355
aI3981
aI8781
aI25993
asS'd'
I1946
sa(dp3892
S'a'
I701
sS's'
I15
sS'm'
(lp3893
I1
aI1
aI1
aI13
aI7
aI19
aI53
aI243
aI301
aI75
aI1183
aI2723
aI6687
aI13
aI16581
asS'd'
I1947
sa(dp3894
S'a'
I704
sS's'
I15
sS'm'
(lp3895
I1
aI3
aI1
aI13
aI17
aI51
aI91
aI239
aI437
aI191
aI1065
aI2495
aI5755
aI3405
aI8299
asS'd'
I1948
sa(dp3896
S'a'
I710
sS's'
I15
sS'm'
(lp3897
I1
aI1
aI5
aI5
aI11
aI59
aI21
aI169
aI299
aI123
aI1845
aI2199
aI2157
aI14461
aI10327
asS'd'
I1949
sa(dp3898
S'a'
I721
sS's'
I15
sS'm'
(lp3899
I1
aI3
aI7
aI7
aI19
aI47
aI51
aI179
aI41
aI19
aI1347
aI2325
aI8063
aI5993
aI15653
asS'd'
I1950
sa(dp3900
S'a'
I728
sS's'
I15
sS'm'
(lp3901
I1
aI1
aI1
aI9
aI25
aI27
aI7
aI133
aI223
aI533
aI719
aI353
aI7093
aI8285
aI10375
asS'd'
I1951
sa(dp3902
S'a'
I738
sS's'
I15
sS'm'
(lp3903
I1
aI3
aI5
aI15
aI31
aI5
aI67
aI39
aI441
aI495
aI977
aI3699
aI1435
aI11385
aI14567
asS'd'
I1952
sa(dp3904
S'a'
I740
sS's'
I15
sS'm'
(lp3905
I1
aI1
aI3
aI15
aI15
aI39
aI25
aI33
aI91
aI523
aI249
aI4035
aI769
aI5181
aI9691
asS'd'
I1953
sa(dp3906
S'a'
I749
sS's'
I15
sS'm'
(lp3907
I1
aI1
aI3
aI3
aI3
aI57
aI83
aI187
aI423
aI165
aI161
aI3453
aI2241
aI981
aI8429
asS'd'
I1954
sa(dp3908
S'a'
I758
sS's'
I15
sS'm'
(lp3909
I1
aI1
aI7
aI15
aI1
aI17
aI57
aI189
aI283
aI11
aI823
aI3505
aI7025
aI11879
aI15441
asS'd'
I1955
sa(dp3910
S'a'
I761
sS's'
I15
sS'm'
(lp3911
I1
aI1
aI3
aI11
aI1
aI41
aI7
aI255
aI385
aI339
aI607
aI1405
aI1473
aI13697
aI9491
asS'd'
I1956
sa(dp3912
S'a'
I772
sS's'
I15
sS'm'
(lp3913
I1
aI1
aI7
aI15
aI5
aI9
aI91
aI99
aI211
aI233
aI51
aI2663
aI1165
aI9283
aI18495
asS'd'
I1957
sa(dp3914
S'a'
I776
sS's'
I15
sS'm'
(lp3915
I1
aI1
aI3
aI7
aI21
aI37
aI13
aI91
aI39
aI27
aI1021
aI2813
aI5937
aI6645
aI3403
asS'd'
I1958
sa(dp3916
S'a'
I782
sS's'
I15
sS'm'
(lp3917
I1
aI3
aI1
aI1
aI29
aI29
aI5
aI69
aI399
aI665
aI1407
aI3921
aI2653
aI11753
aI18925
asS'd'
I1959
sa(dp3918
S'a'
I789
sS's'
I15
sS'm'
(lp3919
I1
aI3
aI7
aI15
aI13
aI41
aI39
aI1
aI437
aI549
aI161
aI2315
aI5631
aI8335
aI22661
asS'd'
I1960
sa(dp3920
S'a'
I810
sS's'
I15
sS'm'
(lp3921
I1
aI1
aI3
aI1
aI7
aI17
aI115
aI61
aI69
aI955
aI475
aI3763
aI8035
aI927
aI17893
asS'd'
I1961
sa(dp3922
S'a'
I812
sS's'
I15
sS'm'
(lp3923
I1
aI3
aI1
aI13
aI21
aI59
aI81
aI145
aI463
aI145
aI1941
aI2777
aI7453
aI14229
aI11281
asS'd'
I1962
sa(dp3924
S'a'
I818
sS's'
I15
sS'm'
(lp3925
I1
aI1
aI1
aI15
aI15
aI11
aI27
aI165
aI461
aI395
aI1645
aI3611
aI7463
aI12379
aI26787
asS'd'
I1963
sa(dp3926
S'a'
I830
sS's'
I15
sS'm'
(lp3927
I1
aI1
aI7
aI9
aI29
aI19
aI27
aI123
aI21
aI149
aI1643
aI4001
aI7207
aI6769
aI4647
asS'd'
I1964
sa(dp3928
S'a'
I832
sS's'
I15
sS'm'
(lp3929
I1
aI1
aI1
aI11
aI13
aI9
aI103
aI139
aI185
aI587
aI591
aI1113
aI2223
aI11667
aI32671
asS'd'
I1965
sa(dp3930
S'a'
I852
sS's'
I15
sS'm'
(lp3931
I1
aI3
aI1
aI1
aI31
aI13
aI19
aI93
aI229
aI125
aI1471
aI2369
aI3055
aI10277
aI28563
asS'd'
I1966
sa(dp3932
S'a'
I855
sS's'
I15
sS'm'
(lp3933
I1
aI3
aI7
aI5
aI7
aI53
aI99
aI175
aI161
aI851
aI617
aI4027
aI2357
aI11199
aI1931
asS'd'
I1967
sa(dp3934
S'a'
I859
sS's'
I15
sS'm'
(lp3935
I1
aI3
aI5
aI11
aI3
aI31
aI111
aI179
aI237
aI845
aI539
aI1057
aI259
aI3417
aI26637
asS'd'
I1968
sa(dp3936
S'a'
I865
sS's'
I15
sS'm'
(lp3937
I1
aI1
aI5
aI3
aI21
aI49
aI125
aI119
aI463
aI403
aI737
aI1811
aI3941
aI13015
aI29081
asS'd'
I1969
sa(dp3938
S'a'
I877
sS's'
I15
sS'm'
(lp3939
I1
aI3
aI5
aI13
aI5
aI29
aI69
aI251
aI313
aI357
aI663
aI1097
aI3307
aI12845
aI28495
asS'd'
I1970
sa(dp3940
S'a'
I895
sS's'
I15
sS'm'
(lp3941
I1
aI3
aI3
aI5
aI29
aI17
aI89
aI15
aI411
aI409
aI2013
aI757
aI4085
aI12521
aI11131
asS'd'
I1971
sa(dp3942
S'a'
I901
sS's'
I15
sS'm'
(lp3943
I1
aI1
aI1
aI15
aI7
aI51
aI3
aI193
aI493
aI133
aI381
aI2027
aI227
aI6635
aI12931
asS'd'
I1972
sa(dp3944
S'a'
I902
sS's'
I15
sS'm'
(lp3945
I1
aI1
aI1
aI15
aI7
aI23
aI99
aI203
aI323
aI1007
aI1465
aI2887
aI2215
aI1787
aI22069
asS'd'
I1973
sa(dp3946
S'a'
I906
sS's'
I15
sS'm'
(lp3947
I1
aI1
aI5
aI9
aI29
aI59
aI77
aI151
aI509
aI313
aI415
aI3977
aI5431
aI8019
aI8571
asS'd'
I1974
sa(dp3948
S'a'
I916
sS's'
I15
sS'm'
(lp3949
I1
aI3
aI1
aI15
aI19
aI13
aI57
aI217
aI87
aI119
aI25
aI1149
aI5667
aI3765
aI6959
asS'd'
I1975
sa(dp3950
S'a'
I920
sS's'
I15
sS'm'
(lp3951
I1
aI3
aI7
aI13
aI19
aI31
aI119
aI3
aI457
aI117
aI905
aI361
aI1483
aI12405
aI27005
asS'd'
I1976
sa(dp3952
S'a'
I949
sS's'
I15
sS'm'
(lp3953
I1
aI3
aI5
aI11
aI15
aI35
aI61
aI77
aI119
aI51
aI1753
aI2765
aI1091
aI10573
aI23595
asS'd'
I1977
sa(dp3954
S'a'
I962
sS's'
I15
sS'm'
(lp3955
I1
aI3
aI3
aI7
aI1
aI35
aI17
aI93
aI197
aI511
aI1253
aI3031
aI2739
aI15127
aI15147
asS'd'
I1978
sa(dp3956
S'a'
I964
sS's'
I15
sS'm'
(lp3957
I1
aI3
aI3
aI1
aI11
aI55
aI55
aI107
aI161
aI75
aI129
aI2195
aI2023
aI4877
aI25797
asS'd'
I1979
sa(dp3958
S'a'
I967
sS's'
I15
sS'm'
(lp3959
I1
aI3
aI5
aI7
aI23
aI19
aI113
aI167
aI167
aI271
aI1303
aI125
aI5057
aI1323
aI5165
asS'd'
I1980
sa(dp3960
S'a'
I981
sS's'
I15
sS'm'
(lp3961
I1
aI1
aI5
aI3
aI21
aI31
aI11
aI119
aI215
aI483
aI1535
aI407
aI6485
aI15401
aI30297
asS'd'
I1981
sa(dp3962
S'a'
I982
sS's'
I15
sS'm'
(lp3963
I1
aI3
aI5
aI9
aI21
aI5
aI77
aI95
aI443
aI247
aI913
aI605
aI365
aI7465
aI19707
asS'd'
I1982
sa(dp3964
S'a'
I985
sS's'
I15
sS'm'
(lp3965
I1
aI3
aI1
aI7
aI17
aI59
aI9
aI35
aI391
aI767
aI1493
aI475
aI4725
aI7529
aI31579
asS'd'
I1983
sa(dp3966
S'a'
I991
sS's'
I15
sS'm'
(lp3967
I1
aI3
aI3
aI7
aI31
aI21
aI61
aI31
aI421
aI179
aI273
aI771
aI5745
aI10575
aI32765
asS'd'
I1984
sa(dp3968
S'a'
I1007
sS's'
I15
sS'm'
(lp3969
I1
aI3
aI5
aI15
aI27
aI13
aI125
aI55
aI423
aI1021
aI497
aI3521
aI6903
aI15111
aI8285
asS'd'
I1985
sa(dp3970
S'a'
I1016
sS's'
I15
sS'm'
(lp3971
I1
aI1
aI5
aI9
aI13
aI31
aI105
aI93
aI421
aI709
aI643
aI1079
aI1533
aI9149
aI10799
asS'd'
I1986
sa(dp3972
S'a'
I1024
sS's'
I15
sS'm'
(lp3973
I1
aI3
aI1
aI11
aI19
aI29
aI53
aI199
aI319
aI247
aI655
aI3039
aI6411
aI12267
aI14245
asS'd'
I1987
sa(dp3974
S'a'
I1051
sS's'
I15
sS'm'
(lp3975
I1
aI3
aI1
aI11
aI9
aI57
aI5
aI91
aI469
aI149
aI259
aI329
aI5433
aI6941
aI15093
asS'd'
I1988
sa(dp3976
S'a'
I1060
sS's'
I15
sS'm'
(lp3977
I1
aI3
aI1
aI5
aI5
aI51
aI59
aI25
aI455
aI367
aI1623
aI441
aI3155
aI11695
aI20767
asS'd'
I1989
sa(dp3978
S'a'
I1070
sS's'
I15
sS'm'
(lp3979
I1
aI3
aI7
aI7
aI11
aI49
aI113
aI95
aI91
aI389
aI605
aI1973
aI2051
aI2315
aI22229
asS'd'
I1990
sa(dp3980
S'a'
I1072
sS's'
I15
sS'm'
(lp3981
I1
aI3
aI5
aI3
aI19
aI11
aI99
aI135
aI433
aI781
aI1473
aI885
aI1105
aI3573
aI3739
asS'd'
I1991
sa(dp3982
S'a'
I1084
sS's'
I15
sS'm'
(lp3983
I1
aI3
aI1
aI11
aI3
aI25
aI9
aI227
aI433
aI723
aI317
aI139
aI6627
aI8067
aI28439
asS'd'
I1992
sa(dp3984
S'a'
I1089
sS's'
I15
sS'm'
(lp3985
I1
aI1
aI1
aI9
aI9
aI9
aI5
aI63
aI241
aI215
aI1991
aI2949
aI3943
aI775
aI31511
asS'd'
I1993
sa(dp3986
S'a'
I1095
sS's'
I15
sS'm'
(lp3987
I1
aI1
aI3
aI7
aI17
aI49
aI35
aI167
aI131
aI107
aI1295
aI2465
aI4577
aI11147
aI29833
asS'd'
I1994
sa(dp3988
S'a'
I1114
sS's'
I15
sS'm'
(lp3989
I1
aI1
aI5
aI1
aI5
aI25
aI119
aI129
aI391
aI743
aI1069
aI2957
aI349
aI6891
aI13635
asS'd'
I1995
sa(dp3990
S'a'
I1123
sS's'
I15
sS'm'
(lp3991
I1
aI3
aI1
aI7
aI9
aI31
aI63
aI253
aI215
aI51
aI1347
aI2361
aI3125
aI13049
aI28461
asS'd'
I1996
sa(dp3992
S'a'
I1132
sS's'
I15
sS'm'
(lp3993
I1
aI1
aI7
aI9
aI3
aI31
aI21
aI163
aI255
aI47
aI259
aI535
aI5461
aI3349
aI30649
asS'd'
I1997
sa(dp3994
S'a'
I1154
sS's'
I15
sS'm'
(lp3995
I1
aI3
aI3
aI13
aI17
aI33
aI87
aI47
aI243
aI709
aI929
aI3943
aI3107
aI3421
aI13721
asS'd'
I1998
sa(dp3996
S'a'
I1156
sS's'
I15
sS'm'
(lp3997
I1
aI3
aI5
aI11
aI25
aI61
aI61
aI173
aI397
aI735
aI2005
aI3355
aI8121
aI11593
aI27697
asS'd'
I1999
sa(dp3998
S'a'
I1163
sS's'
I15
sS'm'
(lp3999
I1
aI3
aI5
aI15
aI17
aI43
aI63
aI231
aI275
aI311
aI1277
aI2669
aI7307
aI2099
aI9755
asS'd'
I2000
sa(dp4000
S'a'
I1171
sS's'
I15
sS'm'
(lp4001
I1
aI3
aI5
aI3
aI25
aI43
aI71
aI191
aI9
aI121
aI1873
aI3747
aI7491
aI14055
aI24293
asS'd'
I2001
sa(dp4002
S'a'
I1202
sS's'
I15
sS'm'
(lp4003
I1
aI3
aI5
aI13
aI17
aI35
aI113
aI113
aI385
aI941
aI39
aI2705
aI1225
aI5167
aI1373
asS'd'
I2002
sa(dp4004
S'a'
I1228
sS's'
I15
sS'm'
(lp4005
I1
aI3
aI5
aI5
aI7
aI35
aI19
aI105
aI487
aI71
aI139
aI627
aI4187
aI3183
aI713
asS'd'
I2003
sa(dp4006
S'a'
I1239
sS's'
I15
sS'm'
(lp4007
I1
aI1
aI5
aI13
aI29
aI29
aI103
aI5
aI157
aI869
aI1675
aI423
aI6689
aI10697
aI5303
asS'd'
I2004
sa(dp4008
S'a'
I1255
sS's'
I15
sS'm'
(lp4009
I1
aI1
aI5
aI1
aI29
aI31
aI61
aI111
aI473
aI963
aI685
aI1483
aI2383
aI8109
aI8495
asS'd'
I2005
sa(dp4010
S'a'
I1256
sS's'
I15
sS'm'
(lp4011
I1
aI1
aI5
aI3
aI19
aI13
aI95
aI113
aI217
aI59
aI1353
aI1647
aI3617
aI3271
aI2321
asS'd'
I2006
sa(dp4012
S'a'
I1262
sS's'
I15
sS'm'
(lp4013
I1
aI3
aI5
aI7
aI25
aI35
aI59
aI131
aI309
aI445
aI415
aI93
aI1453
aI8789
aI30201
asS'd'
I2007
sa(dp4014
S'a'
I1270
sS's'
I15
sS'm'
(lp4015
I1
aI1
aI5
aI1
aI5
aI43
aI71
aI241
aI123
aI189
aI831
aI3469
aI8093
aI6187
aI32721
asS'd'
I2008
sa(dp4016
S'a'
I1279
sS's'
I15
sS'm'
(lp4017
I1
aI3
aI7
aI5
aI25
aI31
aI123
aI171
aI319
aI379
aI889
aI2365
aI4881
aI12225
aI16609
asS'd'
I2009
sa(dp4018
S'a'
I1308
sS's'
I15
sS'm'
(lp4019
I1
aI3
aI1
aI11
aI27
aI43
aI121
aI63
aI291
aI591
aI811
aI1995
aI4777
aI2083
aI31385
asS'd'
I2010
sa(dp4020
S'a'
I1322
sS's'
I15
sS'm'
(lp4021
I1
aI1
aI5
aI11
aI27
aI53
aI85
aI187
aI461
aI823
aI703
aI399
aI6925
aI11517
aI28697
asS'd'
I2011
sa(dp4022
S'a'
I1329
sS's'
I15
sS'm'
(lp4023
I1
aI1
aI3
aI5
aI13
aI11
aI33
aI121
aI93
aI717
aI1275
aI3877
aI4247
aI5845
aI26909
asS'd'
I2012
sa(dp4024
S'a'
I1330
sS's'
I15
sS'm'
(lp4025
I1
aI3
aI1
aI9
aI7
aI5
aI47
aI199
aI367
aI561
aI185
aI2855
aI5997
aI2699
aI7581
asS'd'
I2013
sa(dp4026
S'a'
I1336
sS's'
I15
sS'm'
(lp4027
I1
aI1
aI5
aI9
aI23
aI11
aI71
aI201
aI61
aI729
aI1011
aI3529
aI663
aI1413
aI25675
asS'd'
I2014
sa(dp4028
S'a'
I1341
sS's'
I15
sS'm'
(lp4029
I1
aI3
aI7
aI13
aI27
aI21
aI11
aI127
aI281
aI487
aI1217
aI3129
aI5541
aI3129
aI17783
asS'd'
I2015
sa(dp4030
S'a'
I1347
sS's'
I15
sS'm'
(lp4031
I1
aI1
aI5
aI9
aI1
aI29
aI85
aI193
aI213
aI743
aI1473
aI611
aI391
aI9405
aI21137
asS'd'
I2016
sa(dp4032
S'a'
I1349
sS's'
I15
sS'm'
(lp4033
I1
aI3
aI3
aI3
aI31
aI63
aI37
aI147
aI39
aI351
aI79
aI3069
aI2441
aI8901
aI8777
asS'd'
I2017
sa(dp4034
S'a'
I1359
sS's'
I15
sS'm'
(lp4035
I1
aI1
aI7
aI7
aI25
aI49
aI55
aI47
aI441
aI343
aI1267
aI1123
aI5917
aI14395
aI10579
asS'd'
I2018
sa(dp4036
S'a'
I1367
sS's'
I15
sS'm'
(lp4037
I1
aI1
aI7
aI1
aI13
aI55
aI55
aI123
aI103
aI773
aI125
aI2145
aI4743
aI13347
aI2589
asS'd'
I2019
sa(dp4038
S'a'
I1368
sS's'
I15
sS'm'
(lp4039
I1
aI3
aI7
aI3
aI9
aI33
aI25
aI183
aI469
aI213
aI291
aI75
aI6725
aI6847
aI26745
asS'd'
I2020
sa(dp4040
S'a'
I1390
sS's'
I15
sS'm'
(lp4041
I1
aI3
aI3
aI7
aI15
aI43
aI7
aI79
aI171
aI21
aI1767
aI2537
aI4285
aI12007
aI24039
asS'd'
I2021
sa(dp4042
S'a'
I1413
sS's'
I15
sS'm'
(lp4043
I1
aI3
aI7
aI13
aI9
aI61
aI125
aI23
aI227
aI879
aI215
aI1635
aI2835
aI883
aI15939
asS'd'
I2022
sa(dp4044
S'a'
I1414
sS's'
I15
sS'm'
(lp4045
I1
aI1
aI5
aI13
aI25
aI45
aI63
aI43
aI183
aI829
aI149
aI989
aI987
aI3819
aI12181
asS'd'
I2023
sa(dp4046
S'a'
I1437
sS's'
I15
sS'm'
(lp4047
I1
aI1
aI3
aI7
aI19
aI27
aI35
aI83
aI135
aI459
aI785
aI131
aI2655
aI3329
aI3009
asS'd'
I2024
sa(dp4048
S'a'
I1441
sS's'
I15
sS'm'
(lp4049
I1
aI1
aI7
aI5
aI11
aI41
aI9
aI219
aI475
aI985
aI1329
aI3787
aI1975
aI4679
aI8627
asS'd'
I2025
sa(dp4050
S'a'
I1462
sS's'
I15
sS'm'
(lp4051
I1
aI1
aI7
aI3
aI1
aI17
aI91
aI155
aI3
aI763
aI1879
aI233
aI215
aI2955
aI25993
asS'd'
I2026
sa(dp4052
S'a'
I1465
sS's'
I15
sS'm'
(lp4053
I1
aI1
aI1
aI11
aI25
aI11
aI23
aI227
aI453
aI775
aI1935
aI3833
aI4583
aI269
aI705
asS'd'
I2027
sa(dp4054
S'a'
I1480
sS's'
I15
sS'm'
(lp4055
I1
aI3
aI3
aI11
aI7
aI25
aI105
aI21
aI449
aI555
aI1275
aI3475
aI5503
aI15617
aI813
asS'd'
I2028
sa(dp4056
S'a'
I1486
sS's'
I15
sS'm'
(lp4057
I1
aI3
aI7
aI13
aI31
aI37
aI25
aI255
aI233
aI663
aI1155
aI1563
aI4775
aI7449
aI29949
asS'd'
I2029
sa(dp4058
S'a'
I1504
sS's'
I15
sS'm'
(lp4059
I1
aI1
aI3
aI1
aI23
aI51
aI51
aI137
aI63
aI809
aI349
aI2789
aI6953
aI10605
aI18959
asS'd'
I2030
sa(dp4060
S'a'
I1509
sS's'
I15
sS'm'
(lp4061
I1
aI3
aI3
aI13
aI21
aI45
aI15
aI161
aI393
aI229
aI437
aI2967
aI4019
aI3893
aI21305
asS'd'
I2031
sa(dp4062
S'a'
I1514
sS's'
I15
sS'm'
(lp4063
I1
aI1
aI3
aI7
aI5
aI11
aI15
aI211
aI287
aI131
aI1847
aI2569
aI7881
aI15669
aI31037
asS'd'
I2032
sa(dp4064
S'a'
I1522
sS's'
I15
sS'm'
(lp4065
I1
aI3
aI3
aI15
aI27
aI19
aI85
aI251
aI221
aI639
aI665
aI3729
aI5771
aI7873
aI28005
asS'd'
I2033
sa(dp4066
S'a'
I1528
sS's'
I15
sS'm'
(lp4067
I1
aI3
aI7
aI15
aI15
aI47
aI93
aI215
aI343
aI85
aI1401
aI1375
aI2949
aI13661
aI25453
asS'd'
I2034
sa(dp4068
S'a'
I1552
sS's'
I15
sS'm'
(lp4069
I1
aI1
aI1
aI9
aI7
aI51
aI53
aI217
aI471
aI389
aI551
aI1141
aI1767
aI2237
aI17797
asS'd'
I2035
sa(dp4070
S'a'
I1555
sS's'
I15
sS'm'
(lp4071
I1
aI1
aI7
aI9
aI3
aI29
aI65
aI29
aI223
aI591
aI1719
aI1049
aI7643
aI3853
aI29867
asS'd'
I2036
sa(dp4072
S'a'
I1571
sS's'
I15
sS'm'
(lp4073
I1
aI1
aI1
aI11
aI13
aI41
aI85
aI29
aI451
aI387
aI1783
aI3733
aI8033
aI4711
aI31643
asS'd'
I2037
sa(dp4074
S'a'
I1578
sS's'
I15
sS'm'
(lp4075
I1
aI3
aI1
aI11
aI11
aI57
aI75
aI153
aI7
aI373
aI2011
aI271
aI469
aI3267
aI18969
asS'd'
I2038
sa(dp4076
S'a'
I1585
sS's'
I15
sS'm'
(lp4077
I1
aI1
aI5
aI3
aI19
aI43
aI7
aI243
aI385
aI293
aI923
aI843
aI4895
aI469
aI8421
asS'd'
I2039
sa(dp4078
S'a'
I1588
sS's'
I15
sS'm'
(lp4079
I1
aI3
aI1
aI15
aI29
aI47
aI17
aI125
aI471
aI927
aI349
aI3859
aI3059
aI11483
aI14791
asS'd'
I2040
sa(dp4080
S'a'
I1603
sS's'
I15
sS'm'
(lp4081
I1
aI3
aI1
aI11
aI17
aI17
aI111
aI109
aI9
aI213
aI1313
aI3903
aI4411
aI4329
aI28277
asS'd'
I2041
sa(dp4082
S'a'
I1609
sS's'
I15
sS'm'
(lp4083
I1
aI3
aI3
aI15
aI1
aI55
aI47
aI69
aI143
aI789
aI1149
aI3833
aI5053
aI6949
aI10569
asS'd'
I2042
sa(dp4084
S'a'
I1617
sS's'
I15
sS'm'
(lp4085
I1
aI3
aI5
aI7
aI11
aI15
aI79
aI83
aI123
aI937
aI1115
aI2775
aI3041
aI11869
aI21167
asS'd'
I2043
sa(dp4086
S'a'
I1620
sS's'
I15
sS'm'
(lp4087
I1
aI3
aI7
aI13
aI9
aI47
aI45
aI221
aI139
aI923
aI1661
aI1379
aI2485
aI7233
aI6035
asS'd'
I2044
sa(dp4088
S'a'
I1629
sS's'
I15
sS'm'
(lp4089
I1
aI1
aI3
aI3
aI11
aI55
aI77
aI3
aI87
aI693
aI1991
aI1145
aI2783
aI16207
aI24569
asS'd'
I2045
sa(dp4090
S'a'
I1636
sS's'
I15
sS'm'
(lp4091
I1
aI1
aI5
aI11
aI3
aI35
aI91
aI9
aI391
aI927
aI101
aI1839
aI3755
aI10345
aI16907
asS'd'
I2046
sa(dp4092
S'a'
I1648
sS's'
I15
sS'm'
(lp4093
I1
aI3
aI5
aI3
aI5
aI49
aI79
aI91
aI205
aI443
aI1369
aI197
aI2537
aI11219
aI17765
asS'd'
I2047
sa(dp4094
S'a'
I1667
sS's'
I15
sS'm'
(lp4095
I1
aI1
aI3
aI15
aI9
aI7
aI25
aI25
aI357
aI247
aI477
aI421
aI7679
aI5987
aI30079
asS'd'
I2048
sa(dp4096
S'a'
I1669
sS's'
I15
sS'm'
(lp4097
I1
aI1
aI5
aI3
aI29
aI5
aI89
aI117
aI481
aI491
aI371
aI389
aI7101
aI2253
aI23617
asS'd'
I2049
sa(dp4098
S'a'
I1682
sS's'
I15
sS'm'
(lp4099
I1
aI1
aI5
aI13
aI29
aI59
aI17
aI181
aI511
aI291
aI1991
aI3499
aI8177
aI5559
aI30045
asS'd'
I2050
sa(dp4100
S'a'
I1697
sS's'
I15
sS'm'
(lp4101
I1
aI3
aI3
aI11
aI23
aI31
aI117
aI217
aI241
aI115
aI749
aI945
aI1897
aI12253
aI8473
asS'd'
I2051
sa(dp4102
S'a'
I1704
sS's'
I15
sS'm'
(lp4103
I1
aI1
aI7
aI15
aI25
aI47
aI31
aI1
aI165
aI311
aI635
aI3629
aI1593
aI8305
aI30033
asS'd'
I2052
sa(dp4104
S'a'
I1709
sS's'
I15
sS'm'
(lp4105
I1
aI3
aI5
aI9
aI3
aI17
aI101
aI237
aI379
aI503
aI49
aI929
aI1687
aI3865
aI26723
asS'd'
I2053
sa(dp4106
S'a'
I1727
sS's'
I15
sS'm'
(lp4107
I1
aI3
aI5
aI5
aI15
aI41
aI1
aI239
aI53
aI215
aI1733
aI827
aI579
aI4089
aI6579
asS'd'
I2054
sa(dp4108
S'a'
I1730
sS's'
I15
sS'm'
(lp4109
I1
aI3
aI1
aI15
aI15
aI21
aI35
aI21
aI403
aI257
aI1475
aI2403
aI4705
aI11553
aI203
asS'd'
I2055
sa(dp4110
S'a'
I1732
sS's'
I15
sS'm'
(lp4111
I1
aI3
aI5
aI11
aI9
aI53
aI113
aI9
aI447
aI511
aI543
aI3141
aI7389
aI11249
aI431
asS'd'
I2056
sa(dp4112
S'a'
I1741
sS's'
I15
sS'm'
(lp4113
I1
aI3
aI5
aI9
aI9
aI11
aI55
aI93
aI325
aI411
aI305
aI2573
aI6871
aI12339
aI6435
asS'd'
I2057
sa(dp4114
S'a'
I1744
sS's'
I15
sS'm'
(lp4115
I1
aI3
aI3
aI7
aI31
aI27
aI21
aI113
aI99
aI853
aI365
aI589
aI3731
aI10875
aI12767
asS'd'
I2058
sa(dp4116
S'a'
I1759
sS's'
I15
sS'm'
(lp4117
I1
aI3
aI1
aI7
aI15
aI27
aI31
aI17
aI275
aI93
aI1161
aI2619
aI1329
aI7307
aI587
asS'd'
I2059
sa(dp4118
S'a'
I1765
sS's'
I15
sS'm'
(lp4119
I1
aI3
aI5
aI9
aI17
aI47
aI49
aI237
aI27
aI193
aI1237
aI591
aI5151
aI5521
aI31583
asS'd'
I2060
sa(dp4120
S'a'
I1766
sS's'
I15
sS'm'
(lp4121
I1
aI3
aI5
aI3
aI13
aI1
aI27
aI87
aI43
aI977
aI305
aI3293
aI2475
aI14571
aI18321
asS'd'
I2061
sa(dp4122
S'a'
I1778
sS's'
I15
sS'm'
(lp4123
I1
aI1
aI5
aI7
aI15
aI13
aI101
aI1
aI291
aI807
aI1711
aI2277
aI5573
aI11051
aI13133
asS'd'
I2062
sa(dp4124
S'a'
I1780
sS's'
I15
sS'm'
(lp4125
I1
aI3
aI3
aI1
aI9
aI3
aI65
aI81
aI415
aI733
aI1527
aI2747
aI6069
aI159
aI7095
asS'd'
I2063
sa(dp4126
S'a'
I1783
sS's'
I15
sS'm'
(lp4127
I1
aI3
aI3
aI15
aI27
aI1
aI71
aI49
aI231
aI851
aI2039
aI613
aI1899
aI2537
aI14511
asS'd'
I2064
sa(dp4128
S'a'
I1797
sS's'
I15
sS'm'
(lp4129
I1
aI1
aI1
aI11
aI3
aI41
aI55
aI23
aI247
aI1011
aI581
aI2363
aI2745
aI1337
aI20931
asS'd'
I2065
sa(dp4130
S'a'
I1807
sS's'
I15
sS'm'
(lp4131
I1
aI1
aI3
aI11
aI17
aI61
aI67
aI255
aI143
aI357
aI945
aI3407
aI5817
aI4155
aI23851
asS'd'
I2066
sa(dp4132
S'a'
I1821
sS's'
I15
sS'm'
(lp4133
I1
aI3
aI5
aI3
aI23
aI1
aI75
aI247
aI265
aI413
aI1899
aI2565
aI6629
aI15655
aI16117
asS'd'
I2067
sa(dp4134
S'a'
I1832
sS's'
I15
sS'm'
(lp4135
I1
aI1
aI1
aI9
aI11
aI49
aI11
aI189
aI223
aI177
aI1457
aI1931
aI163
aI15905
aI17297
asS'd'
I2068
sa(dp4136
S'a'
I1835
sS's'
I15
sS'm'
(lp4137
I1
aI3
aI7
aI13
aI17
aI1
aI111
aI189
aI343
aI961
aI427
aI2507
aI2393
aI8653
aI6353
asS'd'
I2069
sa(dp4138
S'a'
I1849
sS's'
I15
sS'm'
(lp4139
I1
aI3
aI7
aI13
aI23
aI61
aI59
aI51
aI313
aI963
aI791
aI3681
aI5637
aI3965
aI9263
asS'd'
I2070
sa(dp4140
S'a'
I1850
sS's'
I15
sS'm'
(lp4141
I1
aI3
aI7
aI7
aI21
aI53
aI127
aI141
aI499
aI859
aI337
aI2835
aI3195
aI4351
aI32369
asS'd'
I2071
sa(dp4142
S'a'
I1863
sS's'
I15
sS'm'
(lp4143
I1
aI1
aI7
aI5
aI1
aI5
aI53
aI63
aI497
aI535
aI35
aI305
aI4395
aI9757
aI13193
asS'd'
I2072
sa(dp4144
S'a'
I1867
sS's'
I15
sS'm'
(lp4145
I1
aI1
aI5
aI13
aI13
aI31
aI59
aI229
aI211
aI745
aI1453
aI3677
aI3005
aI7703
aI23907
asS'd'
I2073
sa(dp4146
S'a'
I1869
sS's'
I15
sS'm'
(lp4147
I1
aI3
aI5
aI5
aI7
aI63
aI17
aI197
aI493
aI861
aI499
aI3015
aI6349
aI1815
aI7437
asS'd'
I2074
sa(dp4148
S'a'
I1872
sS's'
I15
sS'm'
(lp4149
I1
aI1
aI1
aI13
aI13
aI37
aI29
aI189
aI253
aI1017
aI321
aI3145
aI407
aI7547
aI17099
asS'd'
I2075
sa(dp4150
S'a'
I1887
sS's'
I15
sS'm'
(lp4151
I1
aI3
aI3
aI3
aI23
aI53
aI69
aI77
aI175
aI17
aI1831
aI841
aI3851
aI1295
aI32107
asS'd'
I2076
sa(dp4152
S'a'
I1888
sS's'
I15
sS'm'
(lp4153
I1
aI3
aI7
aI13
aI13
aI39
aI107
aI237
aI389
aI729
aI635
aI3717
aI3041
aI3169
aI14987
asS'd'
I2077
sa(dp4154
S'a'
I1897
sS's'
I15
sS'm'
(lp4155
I1
aI1
aI3
aI1
aI25
aI7
aI69
aI35
aI495
aI49
aI659
aI2783
aI6051
aI13875
aI23927
asS'd'
I2078
sa(dp4156
S'a'
I1906
sS's'
I15
sS'm'
(lp4157
I1
aI3
aI7
aI5
aI5
aI25
aI49
aI7
aI193
aI493
aI93
aI657
aI1515
aI13975
aI14155
asS'd'
I2079
sa(dp4158
S'a'
I1917
sS's'
I15
sS'm'
(lp4159
I1
aI3
aI1
aI1
aI11
aI15
aI113
aI45
aI21
aI595
aI731
aI3397
aI4117
aI9711
aI16625
asS'd'
I2080
sa(dp4160
S'a'
I1927
sS's'
I15
sS'm'
(lp4161
I1
aI3
aI3
aI9
aI19
aI19
aI59
aI7
aI105
aI579
aI599
aI2859
aI97
aI14717
aI15361
asS'd'
I2081
sa(dp4162
S'a'
I1939
sS's'
I15
sS'm'
(lp4163
I1
aI1
aI1
aI5
aI27
aI49
aI113
aI5
aI367
aI563
aI1397
aI2805
aI3021
aI3111
aI20671
asS'd'
I2082
sa(dp4164
S'a'
I1941
sS's'
I15
sS'm'
(lp4165
I1
aI3
aI3
aI15
aI27
aI51
aI99
aI167
aI109
aI365
aI1959
aI1523
aI6959
aI14405
aI18191
asS'd'
I2083
sa(dp4166
S'a'
I1948
sS's'
I15
sS'm'
(lp4167
I1
aI3
aI1
aI5
aI21
aI51
aI125
aI67
aI123
aI45
aI1657
aI51
aI4825
aI14081
aI31049
asS'd'
I2084
sa(dp4168
S'a'
I1970
sS's'
I15
sS'm'
(lp4169
I1
aI1
aI5
aI7
aI21
aI59
aI21
aI249
aI77
aI793
aI1687
aI2561
aI2241
aI4321
aI7477
asS'd'
I2085
sa(dp4170
S'a'
I1979
sS's'
I15
sS'm'
(lp4171
I1
aI1
aI1
aI7
aI15
aI35
aI71
aI29
aI267
aI611
aI1813
aI1823
aI7039
aI3299
aI9919
asS'd'
I2086
sa(dp4172
S'a'
I1982
sS's'
I15
sS'm'
(lp4173
I1
aI3
aI7
aI11
aI21
aI59
aI109
aI213
aI371
aI785
aI659
aI1687
aI4827
aI6017
aI19619
asS'd'
I2087
sa(dp4174
S'a'
I2002
sS's'
I15
sS'm'
(lp4175
I1
aI1
aI3
aI11
aI27
aI17
aI1
aI55
aI367
aI939
aI333
aI127
aI5105
aI2405
aI28139
asS'd'
I2088
sa(dp4176
S'a'
I2020
sS's'
I15
sS'm'
(lp4177
I1
aI1
aI7
aI13
aI5
aI35
aI59
aI133
aI509
aI573
aI625
aI3857
aI7935
aI5279
aI3727
asS'd'
I2089
sa(dp4178
S'a'
I2024
sS's'
I15
sS'm'
(lp4179
I1
aI1
aI1
aI7
aI11
aI47
aI127
aI157
aI19
aI403
aI151
aI1143
aI7407
aI8985
aI32521
asS'd'
I2090
sa(dp4180
S'a'
I2032
sS's'
I15
sS'm'
(lp4181
I1
aI3
aI1
aI1
aI5
aI13
aI105
aI123
aI63
aI139
aI1569
aI1983
aI563
aI7175
aI27705
asS'd'
I2091
sa(dp4182
S'a'
I2053
sS's'
I15
sS'm'
(lp4183
I1
aI1
aI3
aI13
aI9
aI35
aI105
aI227
aI145
aI21
aI1369
aI57
aI393
aI2921
aI18511
asS'd'
I2092
sa(dp4184
S'a'
I2060
sS's'
I15
sS'm'
(lp4185
I1
aI3
aI1
aI7
aI17
aI61
aI99
aI187
aI261
aI281
aI437
aI2219
aI5999
aI1857
aI18001
asS'd'
I2093
sa(dp4186
S'a'
I2063
sS's'
I15
sS'm'
(lp4187
I1
aI3
aI3
aI5
aI1
aI59
aI67
aI45
aI451
aI439
aI2005
aI3607
aI3
aI7167
aI14227
asS'd'
I2094
sa(dp4188
S'a'
I2066
sS's'
I15
sS'm'
(lp4189
I1
aI3
aI3
aI3
aI29
aI19
aI25
aI251
aI275
aI733
aI1749
aI4021
aI871
aI3227
aI13701
asS'd'
I2095
sa(dp4190
S'a'
I2075
sS's'
I15
sS'm'
(lp4191
I1
aI3
aI3
aI13
aI27
aI53
aI57
aI243
aI491
aI521
aI1921
aI1037
aI5013
aI5703
aI15261
asS'd'
I2096
sa(dp4192
S'a'
I2078
sS's'
I15
sS'm'
(lp4193
I1
aI3
aI1
aI11
aI13
aI57
aI1
aI15
aI123
aI533
aI785
aI335
aI1423
aI14269
aI3483
asS'd'
I2097
sa(dp4194
S'a'
I2081
sS's'
I15
sS'm'
(lp4195
I1
aI3
aI7
aI13
aI15
aI55
aI5
aI139
aI385
aI47
aI1981
aI1291
aI7397
aI12925
aI29445
asS'd'
I2098
sa(dp4196
S'a'
I2091
sS's'
I15
sS'm'
(lp4197
I1
aI1
aI7
aI1
aI23
aI23
aI59
aI93
aI117
aI57
aI63
aI3047
aI4849
aI11637
aI25311
asS'd'
I2099
sa(dp4198
S'a'
I2096
sS's'
I15
sS'm'
(lp4199
I1
aI1
aI7
aI13
aI19
aI37
aI25
aI203
aI477
aI447
aI1345
aI3485
aI2099
aI13347
aI11621
asS'd'
I2100
sa(dp4200
S'a'
I2102
sS's'
I15
sS'm'
(lp4201
I1
aI1
aI7
aI3
aI11
aI23
aI81
aI17
aI41
aI735
aI1149
aI3253
aI7665
aI8291
aI22293
asS'd'
I2101
sa(dp4202
S'a'
I2106
sS's'
I15
sS'm'
(lp4203
I1
aI1
aI5
aI3
aI15
aI9
aI57
aI167
aI463
aI493
aI747
aI1947
aI6471
aI1111
aI31619
asS'd'
I2102
sa(dp4204
S'a'
I2116
sS's'
I15
sS'm'
(lp4205
I1
aI1
aI5
aI15
aI7
aI15
aI107
aI205
aI325
aI167
aI1749
aI927
aI3589
aI6127
aI7617
asS'd'
I2103
sa(dp4206
S'a'
I2120
sS's'
I15
sS'm'
(lp4207
I1
aI1
aI1
aI13
aI21
aI25
aI83
aI147
aI411
aI399
aI1423
aI2279
aI3661
aI7591
aI17429
asS'd'
I2104
sa(dp4208
S'a'
I2125
sS's'
I15
sS'm'
(lp4209
I1
aI1
aI1
aI9
aI5
aI17
aI69
aI205
aI243
aI647
aI473
aI1717
aI1977
aI10725
aI2913
asS'd'
I2105
sa(dp4210
S'a'
I2134
sS's'
I15
sS'm'
(lp4211
I1
aI1
aI3
aI5
aI5
aI37
aI103
aI15
aI485
aI641
aI1761
aI3755
aI6997
aI10985
aI11773
asS'd'
I2106
sa(dp4212
S'a'
I2178
sS's'
I15
sS'm'
(lp4213
I1
aI1
aI5
aI13
aI9
aI51
aI87
aI195
aI97
aI807
aI1801
aI961
aI6341
aI4307
aI29105
asS'd'
I2107
sa(dp4214
S'a'
I2180
sS's'
I15
sS'm'
(lp4215
I1
aI3
aI1
aI13
aI9
aI35
aI83
aI61
aI387
aI817
aI951
aI3993
aI7831
aI8479
aI23941
asS'd'
I2108
sa(dp4216
S'a'
I2187
sS's'
I15
sS'm'
(lp4217
I1
aI1
aI7
aI11
aI19
aI47
aI75
aI37
aI91
aI337
aI953
aI1169
aI163
aI2259
aI24713
asS'd'
I2109
sa(dp4218
S'a'
I2189
sS's'
I15
sS'm'
(lp4219
I1
aI1
aI1
aI11
aI13
aI15
aI83
aI171
aI159
aI87
aI619
aI2973
aI2653
aI13725
aI12499
asS'd'
I2110
sa(dp4220
S'a'
I2190
sS's'
I15
sS'm'
(lp4221
I1
aI3
aI5
aI3
aI5
aI63
aI119
aI25
aI343
aI269
aI553
aI2183
aI959
aI3825
aI22189
asS'd'
I2111
sa(dp4222
S'a'
I2208
sS's'
I15
sS'm'
(lp4223
I1
aI1
aI5
aI15
aI5
aI37
aI89
aI109
aI497
aI1013
aI265
aI669
aI1859
aI2647
aI3445
asS'd'
I2112
sa(dp4224
S'a'
I2214
sS's'
I15
sS'm'
(lp4225
I1
aI3
aI3
aI9
aI21
aI21
aI15
aI245
aI107
aI649
aI367
aI1601
aI7279
aI15783
aI4943
asS'd'
I2113
sa(dp4226
S'a'
I2237
sS's'
I15
sS'm'
(lp4227
I1
aI3
aI3
aI15
aI5
aI41
aI125
aI113
aI159
aI161
aI1191
aI3491
aI3531
aI55
aI20857
asS'd'
I2114
sa(dp4228
S'a'
I2252
sS's'
I15
sS'm'
(lp4229
I1
aI3
aI5
aI9
aI21
aI57
aI21
aI195
aI99
aI193
aI1915
aI2923
aI6349
aI15085
aI24929
asS'd'
I2115
sa(dp4230
S'a'
I2257
sS's'
I15
sS'm'
(lp4231
I1
aI1
aI1
aI11
aI31
aI11
aI73
aI141
aI361
aI621
aI1021
aI2067
aI5115
aI12665
aI26845
asS'd'
I2116
sa(dp4232
S'a'
I2260
sS's'
I15
sS'm'
(lp4233
I1
aI1
aI1
aI3
aI29
aI11
aI43
aI61
aI209
aI923
aI1753
aI1937
aI843
aI205
aI8367
asS'd'
I2117
sa(dp4234
S'a'
I2264
sS's'
I15
sS'm'
(lp4235
I1
aI1
aI1
aI5
aI15
aI33
aI119
aI209
aI215
aI973
aI1775
aI815
aI6693
aI7957
aI14517
asS'd'
I2118
sa(dp4236
S'a'
I2270
sS's'
I15
sS'm'
(lp4237
I1
aI1
aI1
aI5
aI17
aI57
aI27
aI147
aI489
aI59
aI1439
aI2279
aI445
aI11791
aI19739
asS'd'
I2119
sa(dp4238
S'a'
I2279
sS's'
I15
sS'm'
(lp4239
I1
aI3
aI1
aI7
aI11
aI55
aI1
aI83
aI305
aI17
aI1909
aI405
aI2325
aI5293
aI28559
asS'd'
I2120
sa(dp4240
S'a'
I2288
sS's'
I15
sS'm'
(lp4241
I1
aI3
aI3
aI7
aI11
aI27
aI103
aI157
aI455
aI1005
aI2033
aI3145
aI1919
aI15723
aI25197
asS'd'
I2121
sa(dp4242
S'a'
I2305
sS's'
I15
sS'm'
(lp4243
I1
aI1
aI5
aI11
aI15
aI51
aI37
aI131
aI503
aI1007
aI1795
aI2421
aI1335
aI7413
aI21741
asS'd'
I2122
sa(dp4244
S'a'
I2312
sS's'
I15
sS'm'
(lp4245
I1
aI1
aI3
aI1
aI23
aI63
aI69
aI83
aI419
aI283
aI583
aI123
aI7725
aI2243
aI8403
asS'd'
I2123
sa(dp4246
S'a'
I2317
sS's'
I15
sS'm'
(lp4247
I1
aI1
aI5
aI5
aI27
aI45
aI109
aI17
aI299
aI65
aI351
aI947
aI1165
aI10723
aI2053
asS'd'
I2124
sa(dp4248
S'a'
I2323
sS's'
I15
sS'm'
(lp4249
I1
aI1
aI3
aI3
aI23
aI61
aI115
aI253
aI1
aI931
aI1481
aI3187
aI441
aI14735
aI27207
asS'd'
I2125
sa(dp4250
S'a'
I2329
sS's'
I15
sS'm'
(lp4251
I1
aI1
aI5
aI3
aI25
aI11
aI83
aI141
aI359
aI343
aI901
aI1629
aI731
aI12841
aI14357
asS'd'
I2126
sa(dp4252
S'a'
I2335
sS's'
I15
sS'm'
(lp4253
I1
aI1
aI3
aI9
aI7
aI45
aI97
aI3
aI299
aI217
aI663
aI1527
aI6379
aI4527
aI26147
asS'd'
I2127
sa(dp4254
S'a'
I2342
sS's'
I15
sS'm'
(lp4255
I1
aI1
aI7
aI9
aI11
aI53
aI9
aI203
aI337
aI713
aI1517
aI719
aI4587
aI11443
aI26905
asS'd'
I2128
sa(dp4256
S'a'
I2345
sS's'
I15
sS'm'
(lp4257
I1
aI1
aI7
aI9
aI11
aI41
aI125
aI213
aI237
aI377
aI361
aI3231
aI4223
aI3263
aI12655
asS'd'
I2129
sa(dp4258
S'a'
I2365
sS's'
I15
sS'm'
(lp4259
I1
aI3
aI7
aI7
aI7
aI33
aI99
aI19
aI117
aI273
aI985
aI107
aI3831
aI10135
aI19423
asS'd'
I2130
sa(dp4260
S'a'
I2371
sS's'
I15
sS'm'
(lp4261
I1
aI1
aI5
aI15
aI25
aI41
aI13
aI125
aI449
aI169
aI1149
aI4021
aI5663
aI3077
aI19163
asS'd'
I2131
sa(dp4262
S'a'
I2378
sS's'
I15
sS'm'
(lp4263
I1
aI3
aI5
aI9
aI25
aI57
aI47
aI103
aI269
aI51
aI1805
aI2503
aI6687
aI8065
aI12045
asS'd'
I2132
sa(dp4264
S'a'
I2385
sS's'
I15
sS'm'
(lp4265
I1
aI3
aI5
aI7
aI3
aI35
aI87
aI225
aI189
aI229
aI931
aI3293
aI1347
aI1427
aI3269
asS'd'
I2133
sa(dp4266
S'a'
I2395
sS's'
I15
sS'm'
(lp4267
I1
aI1
aI1
aI3
aI5
aI31
aI61
aI19
aI247
aI9
aI1667
aI343
aI559
aI2703
aI3763
asS'd'
I2134
sa(dp4268
S'a'
I2404
sS's'
I15
sS'm'
(lp4269
I1
aI3
aI5
aI15
aI31
aI19
aI57
aI187
aI109
aI121
aI1287
aI2269
aI659
aI16235
aI1273
asS'd'
I2135
sa(dp4270
S'a'
I2414
sS's'
I15
sS'm'
(lp4271
I1
aI1
aI1
aI3
aI5
aI47
aI59
aI243
aI255
aI97
aI1959
aI1723
aI1347
aI3019
aI26989
asS'd'
I2136
sa(dp4272
S'a'
I2426
sS's'
I15
sS'm'
(lp4273
I1
aI3
aI3
aI15
aI29
aI35
aI75
aI67
aI497
aI731
aI193
aI3307
aI3579
aI12005
aI7209
asS'd'
I2137
sa(dp4274
S'a'
I2428
sS's'
I15
sS'm'
(lp4275
I1
aI1
aI5
aI9
aI13
aI35
aI79
aI213
aI51
aI983
aI1927
aI1793
aI5037
aI5463
aI965
asS'd'
I2138
sa(dp4276
S'a'
I2441
sS's'
I15
sS'm'
(lp4277
I1
aI1
aI7
aI11
aI5
aI41
aI7
aI83
aI15
aI411
aI1775
aI3515
aI6755
aI3249
aI16425
asS'd'
I2139
sa(dp4278
S'a'
I2456
sS's'
I15
sS'm'
(lp4279
I1
aI3
aI5
aI1
aI19
aI61
aI3
aI19
aI395
aI819
aI1331
aI179
aI5225
aI5333
aI3601
asS'd'
I2140
sa(dp4280
S'a'
I2466
sS's'
I15
sS'm'
(lp4281
I1
aI1
aI3
aI9
aI7
aI5
aI87
aI15
aI387
aI609
aI1465
aI277
aI987
aI8377
aI903
asS'd'
I2141
sa(dp4282
S'a'
I2468
sS's'
I15
sS'm'
(lp4283
I1
aI1
aI1
aI3
aI15
aI11
aI123
aI107
aI355
aI333
aI285
aI1801
aI6989
aI1549
aI25791
asS'd'
I2142
sa(dp4284
S'a'
I2475
sS's'
I15
sS'm'
(lp4285
I1
aI1
aI7
aI13
aI27
aI13
aI73
aI111
aI481
aI227
aI1091
aI365
aI5713
aI5087
aI27217
asS'd'
I2143
sa(dp4286
S'a'
I2489
sS's'
I15
sS'm'
(lp4287
I1
aI3
aI3
aI15
aI1
aI55
aI95
aI213
aI377
aI405
aI139
aI1867
aI2175
aI4217
aI28813
asS'd'
I2144
sa(dp4288
S'a'
I2495
sS's'
I15
sS'm'
(lp4289
I1
aI3
aI5
aI11
aI21
aI43
aI109
aI155
aI181
aI901
aI1951
aI507
aI4389
aI10815
aI3141
asS'd'
I2145
sa(dp4290
S'a'
I2497
sS's'
I15
sS'm'
(lp4291
I1
aI1
aI1
aI15
aI17
aI11
aI43
aI215
aI501
aI19
aI259
aI3479
aI6381
aI6927
aI31247
asS'd'
I2146
sa(dp4292
S'a'
I2510
sS's'
I15
sS'm'
(lp4293
I1
aI3
aI5
aI15
aI19
aI61
aI75
aI41
aI391
aI95
aI865
aI1441
aI7993
aI13979
aI24663
asS'd'
I2147
sa(dp4294
S'a'
I2512
sS's'
I15
sS'm'
(lp4295
I1
aI3
aI1
aI3
aI21
aI15
aI115
aI213
aI1
aI645
aI777
aI1517
aI2543
aI11223
aI3633
asS'd'
I2148
sa(dp4296
S'a'
I2522
sS's'
I15
sS'm'
(lp4297
I1
aI3
aI5
aI3
aI9
aI57
aI39
aI211
aI407
aI65
aI1795
aI2805
aI2799
aI8691
aI1987
asS'd'
I2149
sa(dp4298
S'a'
I2533
sS's'
I15
sS'm'
(lp4299
I1
aI1
aI3
aI13
aI17
aI55
aI47
aI113
aI29
aI139
aI1301
aI3303
aI1129
aI13947
aI29821
asS'd'
I2150
sa(dp4300
S'a'
I2543
sS's'
I15
sS'm'
(lp4301
I1
aI1
aI3
aI13
aI5
aI35
aI97
aI151
aI477
aI409
aI1397
aI3399
aI4421
aI15929
aI6163
asS'd'
I2151
sa(dp4302
S'a'
I2551
sS's'
I15
sS'm'
(lp4303
I1
aI3
aI1
aI9
aI21
aI51
aI99
aI133
aI149
aI763
aI623
aI173
aI4311
aI11081
aI1095
asS'd'
I2152
sa(dp4304
S'a'
I2552
sS's'
I15
sS'm'
(lp4305
I1
aI3
aI7
aI15
aI13
aI3
aI99
aI3
aI195
aI907
aI1335
aI1355
aI7977
aI5773
aI32383
asS'd'
I2153
sa(dp4306
S'a'
I2557
sS's'
I15
sS'm'
(lp4307
I1
aI1
aI3
aI9
aI17
aI43
aI43
aI217
aI475
aI917
aI1373
aI1677
aI4871
aI9619
aI16657
asS'd'
I2154
sa(dp4308
S'a'
I2567
sS's'
I15
sS'm'
(lp4309
I1
aI3
aI3
aI7
aI31
aI31
aI55
aI11
aI73
aI693
aI25
aI417
aI1195
aI6225
aI32279
asS'd'
I2155
sa(dp4310
S'a'
I2581
sS's'
I15
sS'm'
(lp4311
I1
aI3
aI5
aI9
aI21
aI57
aI127
aI149
aI79
aI379
aI1609
aI2543
aI6473
aI16033
aI27191
asS'd'
I2156
sa(dp4312
S'a'
I2586
sS's'
I15
sS'm'
(lp4313
I1
aI1
aI5
aI1
aI13
aI9
aI81
aI153
aI297
aI789
aI1749
aI2819
aI3961
aI11231
aI24927
asS'd'
I2157
sa(dp4314
S'a'
I2597
sS's'
I15
sS'm'
(lp4315
I1
aI3
aI5
aI3
aI23
aI61
aI45
aI43
aI43
aI133
aI1481
aI1543
aI2991
aI13739
aI10287
asS'd'
I2158
sa(dp4316
S'a'
I2601
sS's'
I15
sS'm'
(lp4317
I1
aI1
aI3
aI9
aI25
aI43
aI31
aI177
aI337
aI193
aI1083
aI1
aI991
aI9725
aI8379
asS'd'
I2159
sa(dp4318
S'a'
I2622
sS's'
I15
sS'm'
(lp4319
I1
aI3
aI5
aI11
aI13
aI33
aI65
aI83
aI421
aI149
aI409
aI2443
aI7423
aI8847
aI29599
asS'd'
I2160
sa(dp4320
S'a'
I2633
sS's'
I15
sS'm'
(lp4321
I1
aI1
aI5
aI11
aI11
aI1
aI23
aI225
aI77
aI585
aI1505
aI2525
aI739
aI10915
aI25733
asS'd'
I2161
sa(dp4322
S'a'
I2636
sS's'
I15
sS'm'
(lp4323
I1
aI3
aI7
aI13
aI7
aI55
aI3
aI223
aI415
aI521
aI1865
aI2349
aI5663
aI7455
aI16569
asS'd'
I2162
sa(dp4324
S'a'
I2642
sS's'
I15
sS'm'
(lp4325
I1
aI1
aI7
aI13
aI1
aI45
aI121
aI49
aI463
aI99
aI1061
aI2559
aI5087
aI13389
aI11035
asS'd'
I2163
sa(dp4326
S'a'
I2644
sS's'
I15
sS'm'
(lp4327
I1
aI3
aI7
aI11
aI31
aI51
aI35
aI235
aI385
aI1023
aI1771
aI2013
aI5437
aI4877
aI22119
asS'd'
I2164
sa(dp4328
S'a'
I2653
sS's'
I15
sS'm'
(lp4329
I1
aI3
aI3
aI11
aI21
aI3
aI11
aI119
aI81
aI737
aI1093
aI2377
aI4055
aI1121
aI15767
asS'd'
I2165
sa(dp4330
S'a'
I2667
sS's'
I15
sS'm'
(lp4331
I1
aI1
aI5
aI13
aI9
aI3
aI83
aI217
aI387
aI249
aI1047
aI1861
aI4103
aI15367
aI24545
asS'd'
I2166
sa(dp4332
S'a'
I2669
sS's'
I15
sS'm'
(lp4333
I1
aI3
aI3
aI1
aI5
aI37
aI43
aI183
aI383
aI463
aI937
aI1165
aI1481
aI959
aI17047
asS'd'
I2167
sa(dp4334
S'a'
I2672
sS's'
I15
sS'm'
(lp4335
I1
aI1
aI3
aI5
aI7
aI43
aI127
aI243
aI81
aI1021
aI165
aI753
aI4711
aI12965
aI22049
asS'd'
I2168
sa(dp4336
S'a'
I2675
sS's'
I15
sS'm'
(lp4337
I1
aI1
aI5
aI5
aI3
aI61
aI65
aI53
aI425
aI89
aI5
aI1467
aI1395
aI9579
aI8961
asS'd'
I2169
sa(dp4338
S'a'
I2682
sS's'
I15
sS'm'
(lp4339
I1
aI3
aI7
aI13
aI11
aI35
aI123
aI21
aI83
aI689
aI667
aI1203
aI5959
aI15697
aI26885
asS'd'
I2170
sa(dp4340
S'a'
I2687
sS's'
I15
sS'm'
(lp4341
I1
aI1
aI5
aI13
aI9
aI49
aI41
aI101
aI291
aI339
aI1067
aI657
aI4453
aI1137
aI21131
asS'd'
I2171
sa(dp4342
S'a'
I2691
sS's'
I15
sS'm'
(lp4343
I1
aI3
aI3
aI3
aI17
aI61
aI11
aI213
aI27
aI805
aI1691
aI1057
aI6011
aI11941
aI18883
asS'd'
I2172
sa(dp4344
S'a'
I2698
sS's'
I15
sS'm'
(lp4345
I1
aI3
aI1
aI7
aI3
aI51
aI5
aI63
aI121
aI3
aI245
aI2631
aI3737
aI16121
aI26803
asS'd'
I2173
sa(dp4346
S'a'
I2708
sS's'
I15
sS'm'
(lp4347
I1
aI3
aI1
aI1
aI23
aI51
aI79
aI19
aI161
aI107
aI609
aI3489
aI3389
aI4035
aI2427
asS'd'
I2174
sa(dp4348
S'a'
I2712
sS's'
I15
sS'm'
(lp4349
I1
aI3
aI1
aI1
aI17
aI11
aI101
aI101
aI373
aI63
aI1641
aI285
aI1333
aI165
aI14025
asS'd'
I2175
sa(dp4350
S'a'
I2718
sS's'
I15
sS'm'
(lp4351
I1
aI1
aI1
aI5
aI1
aI51
aI83
aI137
aI45
aI1019
aI821
aI867
aI6055
aI10443
aI9857
asS'd'
I2176
sa(dp4352
S'a'
I2722
sS's'
I15
sS'm'
(lp4353
I1
aI3
aI1
aI5
aI17
aI23
aI25
aI181
aI429
aI495
aI317
aI3219
aI5963
aI13945
aI9969
asS'd'
I2177
sa(dp4354
S'a'
I2736
sS's'
I15
sS'm'
(lp4355
I1
aI3
aI7
aI3
aI3
aI15
aI123
aI191
aI369
aI177
aI1697
aI2113
aI3889
aI5201
aI21839
asS'd'
I2178
sa(dp4356
S'a'
I2741
sS's'
I15
sS'm'
(lp4357
I1
aI3
aI1
aI11
aI21
aI39
aI51
aI139
aI271
aI605
aI1007
aI3513
aI3365
aI3781
aI6799
asS'd'
I2179
sa(dp4358
S'a'
I2756
sS's'
I15
sS'm'
(lp4359
I1
aI1
aI7
aI5
aI13
aI19
aI47
aI165
aI249
aI405
aI255
aI1295
aI4513
aI14395
aI5587
asS'd'
I2180
sa(dp4360
S'a'
I2765
sS's'
I15
sS'm'
(lp4361
I1
aI1
aI3
aI7
aI5
aI17
aI99
aI1
aI393
aI31
aI621
aI797
aI6113
aI16003
aI32043
asS'd'
I2181
sa(dp4362
S'a'
I2774
sS's'
I15
sS'm'
(lp4363
I1
aI3
aI5
aI13
aI11
aI21
aI65
aI81
aI147
aI443
aI775
aI3671
aI7029
aI11749
aI3339
asS'd'
I2182
sa(dp4364
S'a'
I2799
sS's'
I15
sS'm'
(lp4365
I1
aI3
aI7
aI1
aI23
aI33
aI99
aI177
aI161
aI577
aI1729
aI617
aI3465
aI11787
aI17577
asS'd'
I2183
sa(dp4366
S'a'
I2804
sS's'
I15
sS'm'
(lp4367
I1
aI1
aI5
aI7
aI15
aI15
aI53
aI193
aI97
aI255
aI1223
aI545
aI5153
aI873
aI24525
asS'd'
I2184
sa(dp4368
S'a'
I2825
sS's'
I15
sS'm'
(lp4369
I1
aI3
aI5
aI1
aI7
aI57
aI47
aI121
aI383
aI835
aI1709
aI2363
aI4731
aI12163
aI7001
asS'd'
I2185
sa(dp4370
S'a'
I2826
sS's'
I15
sS'm'
(lp4371
I1
aI3
aI3
aI11
aI19
aI33
aI63
aI99
aI387
aI95
aI783
aI1009
aI6373
aI4021
aI7685
asS'd'
I2186
sa(dp4372
S'a'
I2840
sS's'
I15
sS'm'
(lp4373
I1
aI1
aI1
aI15
aI25
aI33
aI73
aI135
aI335
aI785
aI935
aI1927
aI5847
aI10501
aI7719
asS'd'
I2187
sa(dp4374
S'a'
I2843
sS's'
I15
sS'm'
(lp4375
I1
aI1
aI5
aI3
aI27
aI45
aI71
aI215
aI489
aI157
aI1189
aI2577
aI6901
aI10219
aI3025
asS'd'
I2188
sa(dp4376
S'a'
I2846
sS's'
I15
sS'm'
(lp4377
I1
aI1
aI7
aI7
aI21
aI3
aI97
aI225
aI101
aI159
aI293
aI2789
aI7955
aI14829
aI1209
asS'd'
I2189
sa(dp4378
S'a'
I2849
sS's'
I15
sS'm'
(lp4379
I1
aI3
aI1
aI5
aI23
aI41
aI83
aI63
aI361
aI195
aI1707
aI2081
aI5363
aI6327
aI179
asS'd'
I2190
sa(dp4380
S'a'
I2867
sS's'
I15
sS'm'
(lp4381
I1
aI1
aI3
aI1
aI21
aI51
aI59
aI67
aI175
aI363
aI825
aI2971
aI3321
aI8837
aI11805
asS'd'
I2191
sa(dp4382
S'a'
I2876
sS's'
I15
sS'm'
(lp4383
I1
aI3
aI7
aI1
aI19
aI3
aI15
aI21
aI429
aI675
aI1589
aI2615
aI2575
aI1537
aI7139
asS'd'
I2192
sa(dp4384
S'a'
I2891
sS's'
I15
sS'm'
(lp4385
I1
aI3
aI3
aI5
aI21
aI29
aI17
aI115
aI345
aI397
aI523
aI1699
aI7043
aI11173
aI3023
asS'd'
I2193
sa(dp4386
S'a'
I2902
sS's'
I15
sS'm'
(lp4387
I1
aI1
aI5
aI7
aI19
aI63
aI99
aI175
aI91
aI433
aI153
aI3749
aI517
aI13667
aI7423
asS'd'
I2194
sa(dp4388
S'a'
I2912
sS's'
I15
sS'm'
(lp4389
I1
aI3
aI7
aI3
aI25
aI23
aI53
aI149
aI65
aI551
aI1231
aI365
aI6637
aI15137
aI16319
asS'd'
I2195
sa(dp4390
S'a'
I2917
sS's'
I15
sS'm'
(lp4391
I1
aI3
aI7
aI13
aI5
aI45
aI11
aI151
aI323
aI31
aI1749
aI409
aI6753
aI10503
aI14991
asS'd'
I2196
sa(dp4392
S'a'
I2927
sS's'
I15
sS'm'
(lp4393
I1
aI3
aI7
aI3
aI5
aI21
aI29
aI117
aI321
aI341
aI1811
aI3619
aI4337
aI12255
aI8629
asS'd'
I2197
sa(dp4394
S'a'
I2941
sS's'
I15
sS'm'
(lp4395
I1
aI3
aI7
aI3
aI7
aI3
aI5
aI221
aI407
aI671
aI1763
aI3669
aI2353
aI8175
aI23489
asS'd'
I2198
sa(dp4396
S'a'
I2965
sS's'
I15
sS'm'
(lp4397
I1
aI1
aI3
aI7
aI11
aI55
aI53
aI185
aI247
aI35
aI1823
aI513
aI1379
aI11827
aI20069
asS'd'
I2199
sa(dp4398
S'a'
I2970
sS's'
I15
sS'm'
(lp4399
I1
aI3
aI3
aI5
aI29
aI51
aI73
aI191
aI185
aI961
aI881
aI2019
aI5651
aI1019
aI15587
asS'd'
I2200
sa(dp4400
S'a'
I2982
sS's'
I15
sS'm'
(lp4401
I1
aI3
aI7
aI13
aI7
aI55
aI59
aI5
aI417
aI829
aI453
aI2339
aI587
aI13283
aI797
asS'd'
I2201
sa(dp4402
S'a'
I2993
sS's'
I15
sS'm'
(lp4403
I1
aI3
aI7
aI3
aI11
aI41
aI75
aI85
aI65
aI149
aI1583
aI529
aI2707
aI11479
aI7109
asS'd'
I2202
sa(dp4404
S'a'
I3018
sS's'
I15
sS'm'
(lp4405
I1
aI3
aI7
aI9
aI13
aI57
aI37
aI243
aI91
aI613
aI665
aI171
aI1631
aI13737
aI2377
asS'd'
I2203
sa(dp4406
S'a'
I3023
sS's'
I15
sS'm'
(lp4407
I1
aI1
aI3
aI7
aI5
aI43
aI97
aI53
aI477
aI793
aI999
aI3647
aI2555
aI7371
aI19295
asS'd'
I2204
sa(dp4408
S'a'
I3025
sS's'
I15
sS'm'
(lp4409
I1
aI1
aI7
aI1
aI1
aI9
aI99
aI253
aI317
aI817
aI1559
aI2081
aI2529
aI14611
aI15997
asS'd'
I2205
sa(dp4410
S'a'
I3026
sS's'
I15
sS'm'
(lp4411
I1
aI3
aI3
aI1
aI5
aI41
aI57
aI121
aI387
aI441
aI709
aI1511
aI7045
aI8409
aI13297
asS'd'
I2206
sa(dp4412
S'a'
I3028
sS's'
I15
sS'm'
(lp4413
I1
aI1
aI1
aI13
aI29
aI57
aI63
aI183
aI327
aI473
aI1943
aI213
aI3973
aI16289
aI2739
asS'd'
I2207
sa(dp4414
S'a'
I3032
sS's'
I15
sS'm'
(lp4415
I1
aI3
aI7
aI9
aI25
aI15
aI75
aI185
aI335
aI881
aI1041
aI3339
aI4471
aI6823
aI21121
asS'd'
I2208
sa(dp4416
S'a'
I3053
sS's'
I15
sS'm'
(lp4417
I1
aI3
aI3
aI13
aI23
aI3
aI57
aI117
aI511
aI927
aI771
aI3229
aI949
aI15487
aI11963
asS'd'
I2209
sa(dp4418
S'a'
I3054
sS's'
I15
sS'm'
(lp4419
I1
aI1
aI3
aI7
aI27
aI19
aI55
aI207
aI331
aI705
aI1945
aI797
aI7125
aI10493
aI16585
asS'd'
I2210
sa(dp4420
S'a'
I3065
sS's'
I15
sS'm'
(lp4421
I1
aI3
aI1
aI1
aI29
aI7
aI91
aI93
aI459
aI93
aI1501
aI1927
aI6415
aI16255
aI9823
asS'd'
I2211
sa(dp4422
S'a'
I3071
sS's'
I15
sS'm'
(lp4423
I1
aI1
aI5
aI5
aI31
aI11
aI97
aI179
aI505
aI807
aI877
aI4003
aI4377
aI8851
aI4239
asS'd'
I2212
sa(dp4424
S'a'
I3076
sS's'
I15
sS'm'
(lp4425
I1
aI1
aI3
aI5
aI11
aI25
aI17
aI131
aI23
aI95
aI311
aI1429
aI2029
aI13091
aI23739
asS'd'
I2213
sa(dp4426
S'a'
I3088
sS's'
I15
sS'm'
(lp4427
I1
aI1
aI3
aI11
aI13
aI27
aI33
aI127
aI481
aI117
aI1127
aI1619
aI6493
aI8507
aI6615
asS'd'
I2214
sa(dp4428
S'a'
I3107
sS's'
I15
sS'm'
(lp4429
I1
aI3
aI1
aI13
aI19
aI27
aI89
aI101
aI27
aI235
aI1579
aI1701
aI4421
aI16037
aI16239
asS'd'
I2215
sa(dp4430
S'a'
I3146
sS's'
I15
sS'm'
(lp4431
I1
aI3
aI1
aI15
aI1
aI15
aI3
aI117
aI317
aI475
aI1691
aI2423
aI5519
aI1703
aI2969
asS'd'
I2216
sa(dp4432
S'a'
I3148
sS's'
I15
sS'm'
(lp4433
I1
aI1
aI3
aI1
aI13
aI15
aI19
aI37
aI237
aI467
aI1321
aI453
aI2169
aI13313
aI31499
asS'd'
I2217
sa(dp4434
S'a'
I3159
sS's'
I15
sS'm'
(lp4435
I1
aI1
aI3
aI15
aI29
aI55
aI31
aI199
aI85
aI285
aI967
aI367
aI3941
aI151
aI20587
asS'd'
I2218
sa(dp4436
S'a'
I3165
sS's'
I15
sS'm'
(lp4437
I1
aI3
aI7
aI15
aI7
aI13
aI31
aI35
aI117
aI543
aI1179
aI3441
aI3039
aI11225
aI30229
asS'd'
I2219
sa(dp4438
S'a'
I3170
sS's'
I15
sS'm'
(lp4439
I1
aI1
aI3
aI15
aI3
aI43
aI1
aI63
aI353
aI395
aI1775
aI3493
aI5175
aI13193
aI25343
asS'd'
I2220
sa(dp4440
S'a'
I3179
sS's'
I15
sS'm'
(lp4441
I1
aI3
aI3
aI15
aI17
aI25
aI57
aI205
aI411
aI83
aI1877
aI2093
aI5599
aI12115
aI8751
asS'd'
I2221
sa(dp4442
S'a'
I3182
sS's'
I15
sS'm'
(lp4443
I1
aI1
aI1
aI11
aI15
aI9
aI115
aI99
aI85
aI887
aI987
aI4015
aI7077
aI3739
aI21505
asS'd'
I2222
sa(dp4444
S'a'
I3205
sS's'
I15
sS'm'
(lp4445
I1
aI3
aI1
aI11
aI25
aI39
aI127
aI37
aI329
aI273
aI1531
aI3211
aI7115
aI15501
aI26575
asS'd'
I2223
sa(dp4446
S'a'
I3212
sS's'
I15
sS'm'
(lp4447
I1
aI1
aI5
aI13
aI15
aI3
aI3
aI101
aI431
aI645
aI493
aI723
aI8083
aI1423
aI14879
asS'd'
I2224
sa(dp4448
S'a'
I3218
sS's'
I15
sS'm'
(lp4449
I1
aI3
aI3
aI5
aI31
aI35
aI37
aI131
aI259
aI849
aI325
aI3403
aI3627
aI3295
aI30885
asS'd'
I2225
sa(dp4450
S'a'
I3220
sS's'
I15
sS'm'
(lp4451
I1
aI3
aI7
aI1
aI9
aI3
aI31
aI201
aI379
aI907
aI1005
aI3333
aI7457
aI2533
aI30357
asS'd'
I2226
sa(dp4452
S'a'
I3223
sS's'
I15
sS'm'
(lp4453
I1
aI3
aI1
aI9
aI7
aI7
aI95
aI103
aI121
aI157
aI895
aI2683
aI5839
aI12403
aI14327
asS'd'
I2227
sa(dp4454
S'a'
I3227
sS's'
I15
sS'm'
(lp4455
I1
aI3
aI7
aI3
aI13
aI5
aI55
aI233
aI3
aI855
aI859
aI1115
aI3883
aI8041
aI3353
asS'd'
I2228
sa(dp4456
S'a'
I3233
sS's'
I15
sS'm'
(lp4457
I1
aI1
aI5
aI9
aI3
aI55
aI99
aI79
aI263
aI831
aI1579
aI205
aI5673
aI1999
aI14879
asS'd'
I2229
sa(dp4458
S'a'
I3234
sS's'
I15
sS'm'
(lp4459
I1
aI3
aI1
aI5
aI17
aI25
aI85
aI19
aI189
aI141
aI877
aI667
aI4461
aI11915
aI23247
asS'd'
I2230
sa(dp4460
S'a'
I3254
sS's'
I15
sS'm'
(lp4461
I1
aI1
aI5
aI5
aI1
aI35
aI15
aI219
aI469
aI725
aI1793
aI3683
aI3661
aI15627
aI30197
asS'd'
I2231
sa(dp4462
S'a'
I3263
sS's'
I15
sS'm'
(lp4463
I1
aI1
aI7
aI5
aI27
aI3
aI41
aI153
aI431
aI487
aI759
aI1345
aI6735
aI9937
aI26277
asS'd'
I2232
sa(dp4464
S'a'
I3268
sS's'
I15
sS'm'
(lp4465
I1
aI1
aI1
aI11
aI11
aI13
aI41
aI121
aI265
aI465
aI1447
aI5
aI3407
aI1907
aI10037
asS'd'
I2233
sa(dp4466
S'a'
I3272
sS's'
I15
sS'm'
(lp4467
I1
aI3
aI5
aI9
aI15
aI63
aI5
aI7
aI407
aI83
aI365
aI3687
aI7721
aI6973
aI16967
asS'd'
I2234
sa(dp4468
S'a'
I3277
sS's'
I15
sS'm'
(lp4469
I1
aI1
aI7
aI7
aI5
aI41
aI75
aI155
aI417
aI565
aI1199
aI1111
aI2823
aI10703
aI22561
asS'd'
I2235
sa(dp4470
S'a'
I3292
sS's'
I15
sS'm'
(lp4471
I1
aI3
aI7
aI5
aI7
aI43
aI39
aI185
aI105
aI327
aI1977
aI1137
aI3261
aI10583
aI11661
asS'd'
I2236
sa(dp4472
S'a'
I3295
sS's'
I15
sS'm'
(lp4473
I1
aI3
aI7
aI7
aI19
aI19
aI103
aI137
aI169
aI273
aI1357
aI3413
aI7647
aI10531
aI32489
asS'd'
I2237
sa(dp4474
S'a'
I3296
sS's'
I15
sS'm'
(lp4475
I1
aI1
aI3
aI13
aI13
aI3
aI81
aI23
aI161
aI295
aI735
aI2031
aI1027
aI15513
aI20165
asS'd'
I2238
sa(dp4476
S'a'
I3301
sS's'
I15
sS'm'
(lp4477
I1
aI1
aI5
aI1
aI15
aI1
aI91
aI35
aI375
aI207
aI1417
aI1115
aI2237
aI11749
aI8509
asS'd'
I2239
sa(dp4478
S'a'
I3306
sS's'
I15
sS'm'
(lp4479
I1
aI3
aI7
aI3
aI25
aI51
aI49
aI219
aI195
aI417
aI1523
aI3953
aI5739
aI7499
aI27071
asS'd'
I2240
sa(dp4480
S'a'
I3313
sS's'
I15
sS'm'
(lp4481
I1
aI1
aI3
aI11
aI23
aI29
aI19
aI81
aI421
aI633
aI513
aI547
aI7545
aI29
aI11909
asS'd'
I2241
sa(dp4482
S'a'
I3346
sS's'
I15
sS'm'
(lp4483
I1
aI1
aI1
aI7
aI13
aI61
aI33
aI243
aI221
aI231
aI111
aI879
aI2861
aI1795
aI27531
asS'd'
I2242
sa(dp4484
S'a'
I3367
sS's'
I15
sS'm'
(lp4485
I1
aI3
aI7
aI3
aI19
aI21
aI1
aI141
aI159
aI605
aI969
aI3013
aI6583
aI2447
aI19919
asS'd'
I2243
sa(dp4486
S'a'
I3371
sS's'
I15
sS'm'
(lp4487
I1
aI3
aI7
aI3
aI31
aI9
aI91
aI83
aI29
aI873
aI929
aI43
aI2253
aI12539
aI23951
asS'd'
I2244
sa(dp4488
S'a'
I3373
sS's'
I15
sS'm'
(lp4489
I1
aI1
aI5
aI3
aI31
aI15
aI87
aI105
aI319
aI973
aI1489
aI3417
aI3377
aI15749
aI2357
asS'd'
I2245
sa(dp4490
S'a'
I3374
sS's'
I15
sS'm'
(lp4491
I1
aI1
aI3
aI15
aI7
aI23
aI3
aI81
aI383
aI419
aI713
aI997
aI6873
aI593
aI285
asS'd'
I2246
sa(dp4492
S'a'
I3376
sS's'
I15
sS'm'
(lp4493
I1
aI3
aI3
aI1
aI29
aI13
aI29
aI101
aI441
aI693
aI2039
aI2951
aI5921
aI12129
aI12053
asS'd'
I2247
sa(dp4494
S'a'
I3382
sS's'
I15
sS'm'
(lp4495
I1
aI1
aI3
aI15
aI9
aI29
aI97
aI117
aI421
aI433
aI1017
aI125
aI3607
aI9415
aI6843
asS'd'
I2248
sa(dp4496
S'a'
I3388
sS's'
I15
sS'm'
(lp4497
I1
aI3
aI5
aI9
aI11
aI13
aI75
aI155
aI413
aI75
aI109
aI1599
aI6161
aI16115
aI12621
asS'd'
I2249
sa(dp4498
S'a'
I3391
sS's'
I15
sS'm'
(lp4499
I1
aI3
aI3
aI3
aI11
aI13
aI49
aI225
aI401
aI599
aI1815
aI1643
aI7853
aI13305
aI25195
asS'd'
I2250
sa(dp4500
S'a'
I3403
sS's'
I15
sS'm'
(lp4501
I1
aI3
aI7
aI5
aI15
aI11
aI27
aI95
aI387
aI931
aI549
aI2179
aI3397
aI15883
aI16563
asS'd'
I2251
sa(dp4502
S'a'
I3406
sS's'
I15
sS'm'
(lp4503
I1
aI1
aI7
aI3
aI9
aI39
aI121
aI5
aI453
aI27
aI1747
aI657
aI2593
aI1289
aI12577
asS'd'
I2252
sa(dp4504
S'a'
I3413
sS's'
I15
sS'm'
(lp4505
I1
aI3
aI7
aI5
aI25
aI25
aI109
aI49
aI185
aI985
aI631
aI803
aI3865
aI8955
aI17901
asS'd'
I2253
sa(dp4506
S'a'
I3420
sS's'
I15
sS'm'
(lp4507
I1
aI1
aI3
aI13
aI3
aI59
aI47
aI49
aI139
aI275
aI1471
aI2995
aI5593
aI14011
aI18741
asS'd'
I2254
sa(dp4508
S'a'
I3427
sS's'
I15
sS'm'
(lp4509
I1
aI1
aI5
aI15
aI29
aI11
aI97
aI225
aI245
aI291
aI1873
aI2365
aI767
aI3419
aI14943
asS'd'
I2255
sa(dp4510
S'a'
I3441
sS's'
I15
sS'm'
(lp4511
I1
aI3
aI3
aI5
aI15
aI17
aI19
aI209
aI359
aI891
aI1375
aI2003
aI7247
aI5299
aI28841
asS'd'
I2256
sa(dp4512
S'a'
I3453
sS's'
I15
sS'm'
(lp4513
I1
aI3
aI7
aI7
aI9
aI55
aI105
aI35
aI77
aI47
aI1023
aI13
aI2901
aI847
aI10265
asS'd'
I2257
sa(dp4514
S'a'
I3464
sS's'
I15
sS'm'
(lp4515
I1
aI3
aI7
aI7
aI7
aI5
aI65
aI233
aI141
aI277
aI1333
aI2357
aI443
aI7257
aI21979
asS'd'
I2258
sa(dp4516
S'a'
I3469
sS's'
I15
sS'm'
(lp4517
I1
aI3
aI5
aI11
aI13
aI63
aI41
aI87
aI193
aI737
aI1085
aI2317
aI7869
aI10149
aI12163
asS'd'
I2259
sa(dp4518
S'a'
I3481
sS's'
I15
sS'm'
(lp4519
I1
aI3
aI1
aI1
aI7
aI57
aI75
aI235
aI461
aI857
aI155
aI2679
aI5925
aI2565
aI10881
asS'd'
I2260
sa(dp4520
S'a'
I3488
sS's'
I15
sS'm'
(lp4521
I1
aI1
aI7
aI15
aI13
aI41
aI63
aI135
aI433
aI387
aI1943
aI2249
aI5469
aI11679
aI28661
asS'd'
I2261
sa(dp4522
S'a'
I3497
sS's'
I15
sS'm'
(lp4523
I1
aI3
aI3
aI13
aI5
aI3
aI103
aI161
aI367
aI649
aI789
aI1179
aI4163
aI5699
aI16787
asS'd'
I2262
sa(dp4524
S'a'
I3503
sS's'
I15
sS'm'
(lp4525
I1
aI3
aI7
aI7
aI31
aI13
aI45
aI141
aI113
aI769
aI1035
aI457
aI6709
aI14989
aI27311
asS'd'
I2263
sa(dp4526
S'a'
I3511
sS's'
I15
sS'm'
(lp4527
I1
aI1
aI3
aI1
aI1
aI43
aI119
aI145
aI111
aI593
aI1139
aI417
aI637
aI4437
aI17285
asS'd'
I2264
sa(dp4528
S'a'
I3515
sS's'
I15
sS'm'
(lp4529
I1
aI3
aI5
aI9
aI9
aI33
aI19
aI99
aI201
aI685
aI1793
aI2621
aI6857
aI8769
aI5623
asS'd'
I2265
sa(dp4530
S'a'
I3525
sS's'
I15
sS'm'
(lp4531
I1
aI3
aI5
aI5
aI23
aI43
aI27
aI189
aI325
aI415
aI215
aI1253
aI3599
aI1215
aI10093
asS'd'
I2266
sa(dp4532
S'a'
I3529
sS's'
I15
sS'm'
(lp4533
I1
aI1
aI3
aI13
aI11
aI35
aI113
aI173
aI503
aI19
aI1459
aI503
aI5363
aI3967
aI13945
asS'd'
I2267
sa(dp4534
S'a'
I3547
sS's'
I15
sS'm'
(lp4535
I1
aI1
aI5
aI11
aI31
aI49
aI13
aI173
aI199
aI623
aI1231
aI2495
aI6581
aI7957
aI25321
asS'd'
I2268
sa(dp4536
S'a'
I3550
sS's'
I15
sS'm'
(lp4537
I1
aI3
aI1
aI9
aI23
aI3
aI79
aI149
aI505
aI937
aI1839
aI3701
aI1673
aI8589
aI8031
asS'd'
I2269
sa(dp4538
S'a'
I3573
sS's'
I15
sS'm'
(lp4539
I1
aI3
aI3
aI5
aI21
aI27
aI107
aI11
aI505
aI407
aI177
aI3593
aI4729
aI12773
aI11685
asS'd'
I2270
sa(dp4540
S'a'
I3583
sS's'
I15
sS'm'
(lp4541
I1
aI3
aI1
aI11
aI29
aI49
aI79
aI53
aI61
aI895
aI2035
aI563
aI5613
aI6065
aI6207
asS'd'
I2271
sa(dp4542
S'a'
I3594
sS's'
I15
sS'm'
(lp4543
I1
aI1
aI3
aI7
aI1
aI53
aI3
aI215
aI99
aI865
aI1749
aI3533
aI4305
aI1243
aI28463
asS'd'
I2272
sa(dp4544
S'a'
I3607
sS's'
I15
sS'm'
(lp4545
I1
aI1
aI1
aI13
aI31
aI59
aI115
aI53
aI403
aI909
aI847
aI103
aI4967
aI10623
aI30073
asS'd'
I2273
sa(dp4546
S'a'
I3613
sS's'
I15
sS'm'
(lp4547
I1
aI1
aI7
aI5
aI27
aI1
aI119
aI83
aI457
aI81
aI395
aI811
aI6221
aI14337
aI541
asS'd'
I2274
sa(dp4548
S'a'
I3624
sS's'
I15
sS'm'
(lp4549
I1
aI1
aI5
aI5
aI5
aI53
aI83
aI117
aI269
aI327
aI875
aI101
aI3343
aI715
aI26339
asS'd'
I2275
sa(dp4550
S'a'
I3630
sS's'
I15
sS'm'
(lp4551
I1
aI1
aI1
aI11
aI31
aI39
aI121
aI147
aI305
aI383
aI1211
aI1897
aI7647
aI11687
aI18907
asS'd'
I2276
sa(dp4552
S'a'
I3635
sS's'
I15
sS'm'
(lp4553
I1
aI3
aI3
aI15
aI23
aI53
aI17
aI85
aI395
aI503
aI61
aI1745
aI4713
aI4641
aI13787
asS'd'
I2277
sa(dp4554
S'a'
I3642
sS's'
I15
sS'm'
(lp4555
I1
aI1
aI7
aI7
aI27
aI1
aI105
aI29
aI287
aI37
aI959
aI975
aI4427
aI4705
aI10175
asS'd'
I2278
sa(dp4556
S'a'
I3644
sS's'
I15
sS'm'
(lp4557
I1
aI3
aI3
aI5
aI7
aI63
aI57
aI199
aI27
aI107
aI1095
aI3923
aI6969
aI713
aI11619
asS'd'
I2279
sa(dp4558
S'a'
I3650
sS's'
I15
sS'm'
(lp4559
I1
aI3
aI5
aI1
aI5
aI49
aI85
aI45
aI449
aI45
aI49
aI3419
aI1109
aI455
aI15917
asS'd'
I2280
sa(dp4560
S'a'
I3679
sS's'
I15
sS'm'
(lp4561
I1
aI1
aI1
aI5
aI13
aI15
aI39
aI27
aI467
aI85
aI1537
aI3055
aI1977
aI8829
aI25231
asS'd'
I2281
sa(dp4562
S'a'
I3690
sS's'
I15
sS'm'
(lp4563
I1
aI1
aI1
aI15
aI1
aI47
aI23
aI121
aI147
aI547
aI1865
aI1491
aI779
aI3515
aI12667
asS'd'
I2282
sa(dp4564
S'a'
I3698
sS's'
I15
sS'm'
(lp4565
I1
aI3
aI3
aI1
aI19
aI5
aI77
aI101
aI1
aI721
aI1149
aI2967
aI4925
aI11889
aI16655
asS'd'
I2283
sa(dp4566
S'a'
I3704
sS's'
I15
sS'm'
(lp4567
I1
aI1
aI1
aI7
aI1
aI35
aI95
aI239
aI127
aI855
aI1031
aI455
aI7631
aI6039
aI21983
asS'd'
I2284
sa(dp4568
S'a'
I3707
sS's'
I15
sS'm'
(lp4569
I1
aI3
aI7
aI9
aI23
aI43
aI75
aI105
aI335
aI223
aI1825
aI3217
aI413
aI7473
aI30005
asS'd'
I2285
sa(dp4570
S'a'
I3713
sS's'
I15
sS'm'
(lp4571
I1
aI1
aI5
aI15
aI29
aI9
aI43
aI145
aI223
aI523
aI511
aI323
aI5955
aI11141
aI22533
asS'd'
I2286
sa(dp4572
S'a'
I3754
sS's'
I15
sS'm'
(lp4573
I1
aI1
aI3
aI1
aI13
aI61
aI93
aI133
aI461
aI233
aI383
aI693
aI7347
aI3165
aI27493
asS'd'
I2287
sa(dp4574
S'a'
I3756
sS's'
I15
sS'm'
(lp4575
I1
aI3
aI7
aI1
aI13
aI45
aI113
aI207
aI53
aI1007
aI815
aI1145
aI2937
aI289
aI22195
asS'd'
I2288
sa(dp4576
S'a'
I3761
sS's'
I15
sS'm'
(lp4577
I1
aI3
aI5
aI5
aI19
aI17
aI113
aI89
aI19
aI1023
aI1625
aI3277
aI697
aI5187
aI15433
asS'd'
I2289
sa(dp4578
S'a'
I3776
sS's'
I15
sS'm'
(lp4579
I1
aI1
aI3
aI13
aI21
aI15
aI15
aI197
aI409
aI391
aI1993
aI2475
aI3189
aI4431
aI29585
asS'd'
I2290
sa(dp4580
S'a'
I3781
sS's'
I15
sS'm'
(lp4581
I1
aI1
aI5
aI5
aI31
aI7
aI111
aI231
aI187
aI543
aI45
aI3863
aI3811
aI4573
aI4437
asS'd'
I2291
sa(dp4582
S'a'
I3788
sS's'
I15
sS'm'
(lp4583
I1
aI3
aI3
aI7
aI19
aI7
aI123
aI23
aI79
aI513
aI189
aI3663
aI1291
aI13257
aI8949
asS'd'
I2292
sa(dp4584
S'a'
I3791
sS's'
I15
sS'm'
(lp4585
I1
aI1
aI5
aI13
aI3
aI53
aI109
aI133
aI157
aI223
aI651
aI3059
aI6055
aI14455
aI26903
asS'd'
I2293
sa(dp4586
S'a'
I3794
sS's'
I15
sS'm'
(lp4587
I1
aI1
aI7
aI1
aI23
aI63
aI59
aI229
aI17
aI199
aI643
aI637
aI7631
aI13647
aI7399
asS'd'
I2294
sa(dp4588
S'a'
I3806
sS's'
I15
sS'm'
(lp4589
I1
aI1
aI1
aI3
aI1
aI51
aI119
aI67
aI335
aI543
aI913
aI3565
aI4795
aI13405
aI7463
asS'd'
I2295
sa(dp4590
S'a'
I3841
sS's'
I15
sS'm'
(lp4591
I1
aI1
aI5
aI3
aI31
aI5
aI91
aI97
aI23
aI223
aI837
aI1353
aI1929
aI12043
aI10039
asS'd'
I2296
sa(dp4592
S'a'
I3848
sS's'
I15
sS'm'
(lp4593
I1
aI3
aI5
aI7
aI19
aI3
aI79
aI171
aI301
aI687
aI1545
aI355
aI4709
aI12965
aI16797
asS'd'
I2297
sa(dp4594
S'a'
I3851
sS's'
I15
sS'm'
(lp4595
I1
aI3
aI5
aI11
aI11
aI49
aI111
aI123
aI251
aI569
aI1605
aI401
aI5439
aI13519
aI8847
asS'd'
I2298
sa(dp4596
S'a'
I3856
sS's'
I15
sS'm'
(lp4597
I1
aI3
aI1
aI3
aI3
aI53
aI7
aI55
aI369
aI633
aI181
aI4037
aI2993
aI15815
aI8661
asS'd'
I2299
sa(dp4598
S'a'
I3868
sS's'
I15
sS'm'
(lp4599
I1
aI1
aI1
aI13
aI31
aI29
aI75
aI167
aI279
aI597
aI539
aI1791
aI8013
aI4387
aI9717
asS'd'
I2300
sa(dp4600
S'a'
I3875
sS's'
I15
sS'm'
(lp4601
I1
aI1
aI5
aI7
aI17
aI15
aI99
aI183
aI211
aI49
aI225
aI3143
aI4537
aI13141
aI23375
asS'd'
I2301
sa(dp4602
S'a'
I3882
sS's'
I15
sS'm'
(lp4603
I1
aI1
aI3
aI5
aI3
aI59
aI25
aI149
aI467
aI69
aI1939
aI1007
aI2765
aI4693
aI29815
asS'd'
I2302
sa(dp4604
S'a'
I3884
sS's'
I15
sS'm'
(lp4605
I1
aI3
aI1
aI3
aI17
aI33
aI119
aI189
aI447
aI251
aI879
aI177
aI5395
aI13487
aI9587
asS'd'
I2303
sa(dp4606
S'a'
I3889
sS's'
I15
sS'm'
(lp4607
I1
aI3
aI3
aI7
aI15
aI31
aI115
aI3
aI21
aI817
aI475
aI1849
aI6041
aI12541
aI18701
asS'd'
I2304
sa(dp4608
S'a'
I3892
sS's'
I15
sS'm'
(lp4609
I1
aI1
aI5
aI13
aI31
aI33
aI7
aI115
aI361
aI587
aI1919
aI1007
aI3537
aI7493
aI19357
asS'd'
I2305
sa(dp4610
S'a'
I3919
sS's'
I15
sS'm'
(lp4611
I1
aI3
aI7
aI13
aI23
aI35
aI15
aI111
aI123
aI633
aI805
aI1983
aI2109
aI14477
aI4985
asS'd'
I2306
sa(dp4612
S'a'
I3921
sS's'
I15
sS'm'
(lp4613
I1
aI3
aI3
aI11
aI25
aI13
aI11
aI205
aI97
aI893
aI927
aI1291
aI4007
aI13593
aI29693
asS'd'
I2307
sa(dp4614
S'a'
I3958
sS's'
I15
sS'm'
(lp4615
I1
aI3
aI5
aI15
aI9
aI13
aI121
aI89
aI215
aI823
aI1389
aI1581
aI8095
aI4707
aI16061
asS'd'
I2308
sa(dp4616
S'a'
I3961
sS's'
I15
sS'm'
(lp4617
I1
aI3
aI1
aI3
aI23
aI39
aI83
aI23
aI47
aI941
aI1419
aI2389
aI5699
aI7519
aI5829
asS'd'
I2309
sa(dp4618
S'a'
I3973
sS's'
I15
sS'm'
(lp4619
I1
aI3
aI1
aI9
aI23
aI43
aI79
aI237
aI93
aI203
aI695
aI225
aI5645
aI3591
aI16775
asS'd'
I2310
sa(dp4620
S'a'
I3977
sS's'
I15
sS'm'
(lp4621
I1
aI3
aI5
aI3
aI15
aI19
aI89
aI129
aI375
aI125
aI225
aI1323
aI2267
aI11607
aI17937
asS'd'
I2311
sa(dp4622
S'a'
I3985
sS's'
I15
sS'm'
(lp4623
I1
aI3
aI3
aI1
aI31
aI37
aI93
aI133
aI377
aI959
aI707
aI621
aI7179
aI15493
aI30287
asS'd'
I2312
sa(dp4624
S'a'
I3991
sS's'
I15
sS'm'
(lp4625
I1
aI3
aI7
aI13
aI5
aI13
aI15
aI1
aI37
aI525
aI1641
aI2829
aI6139
aI4069
aI19187
asS'd'
I2313
sa(dp4626
S'a'
I4004
sS's'
I15
sS'm'
(lp4627
I1
aI3
aI3
aI9
aI17
aI3
aI67
aI97
aI375
aI845
aI403
aI973
aI3919
aI2275
aI31627
asS'd'
I2314
sa(dp4628
S'a'
I4007
sS's'
I15
sS'm'
(lp4629
I1
aI1
aI3
aI3
aI25
aI7
aI91
aI67
aI271
aI465
aI481
aI3477
aI5229
aI241
aI8411
asS'd'
I2315
sa(dp4630
S'a'
I4019
sS's'
I15
sS'm'
(lp4631
I1
aI1
aI1
aI11
aI1
aI41
aI109
aI115
aI75
aI787
aI309
aI2887
aI179
aI9073
aI13895
asS'd'
I2316
sa(dp4632
S'a'
I4045
sS's'
I15
sS'm'
(lp4633
I1
aI3
aI3
aI15
aI11
aI31
aI113
aI91
aI303
aI907
aI1933
aI2167
aI7799
aI11821
aI20659
asS'd'
I2317
sa(dp4634
S'a'
I4054
sS's'
I15
sS'm'
(lp4635
I1
aI3
aI1
aI15
aI27
aI17
aI21
aI41
aI99
aI137
aI1397
aI929
aI5819
aI11977
aI6201
asS'd'
I2318
sa(dp4636
S'a'
I4057
sS's'
I15
sS'm'
(lp4637
I1
aI1
aI7
aI13
aI21
aI29
aI47
aI239
aI287
aI305
aI899
aI2711
aI1723
aI3315
aI199
asS'd'
I2319
sa(dp4638
S'a'
I4058
sS's'
I15
sS'm'
(lp4639
I1
aI1
aI1
aI3
aI31
aI21
aI101
aI149
aI107
aI761
aI1197
aI1703
aI4803
aI8411
aI10649
asS'd'
I2320
sa(dp4640
S'a'
I4070
sS's'
I15
sS'm'
(lp4641
I1
aI1
aI5
aI15
aI23
aI45
aI109
aI221
aI85
aI619
aI169
aI1013
aI3305
aI9451
aI26189
asS'd'
I2321
sa(dp4642
S'a'
I4101
sS's'
I15
sS'm'
(lp4643
I1
aI3
aI5
aI13
aI7
aI57
aI19
aI153
aI231
aI627
aI565
aI1595
aI6309
aI5037
aI25505
asS'd'
I2322
sa(dp4644
S'a'
I4113
sS's'
I15
sS'm'
(lp4645
I1
aI1
aI7
aI7
aI1
aI45
aI43
aI79
aI271
aI59
aI219
aI2255
aI1785
aI7919
aI24061
asS'd'
I2323
sa(dp4646
S'a'
I4114
sS's'
I15
sS'm'
(lp4647
I1
aI3
aI7
aI5
aI31
aI57
aI57
aI231
aI33
aI227
aI531
aI679
aI1141
aI85
aI19777
asS'd'
I2324
sa(dp4648
S'a'
I4119
sS's'
I15
sS'm'
(lp4649
I1
aI1
aI3
aI15
aI11
aI59
aI59
aI169
aI459
aI693
aI907
aI1191
aI3783
aI12809
aI6263
asS'd'
I2325
sa(dp4650
S'a'
I4129
sS's'
I15
sS'm'
(lp4651
I1
aI1
aI7
aI13
aI19
aI21
aI105
aI65
aI267
aI141
aI1547
aI781
aI7295
aI13565
aI17775
asS'd'
I2326
sa(dp4652
S'a'
I4141
sS's'
I15
sS'm'
(lp4653
I1
aI3
aI3
aI5
aI31
aI63
aI97
aI155
aI477
aI661
aI329
aI797
aI2539
aI4061
aI10537
asS'd'
I2327
sa(dp4654
S'a'
I4142
sS's'
I15
sS'm'
(lp4655
I1
aI3
aI3
aI7
aI11
aI17
aI119
aI89
aI71
aI103
aI1043
aI413
aI6035
aI12829
aI11559
asS'd'
I2328
sa(dp4656
S'a'
I4147
sS's'
I15
sS'm'
(lp4657
I1
aI3
aI1
aI9
aI5
aI19
aI53
aI185
aI103
aI629
aI2015
aI1257
aI5163
aI10581
aI13449
asS'd'
I2329
sa(dp4658
S'a'
I4149
sS's'
I15
sS'm'
(lp4659
I1
aI1
aI1
aI5
aI23
aI35
aI25
aI129
aI179
aI959
aI677
aI2249
aI6315
aI12151
aI3459
asS'd'
I2330
sa(dp4660
S'a'
I4150
sS's'
I15
sS'm'
(lp4661
I1
aI1
aI1
aI1
aI9
aI47
aI93
aI45
aI35
aI45
aI265
aI2065
aI6225
aI25
aI27135
asS'd'
I2331
sa(dp4662
S'a'
I4164
sS's'
I15
sS'm'
(lp4663
I1
aI3
aI1
aI11
aI21
aI53
aI127
aI163
aI311
aI667
aI597
aI1561
aI4515
aI23
aI9551
asS'd'
I2332
sa(dp4664
S'a'
I4168
sS's'
I15
sS'm'
(lp4665
I1
aI1
aI3
aI3
aI7
aI47
aI105
aI211
aI241
aI95
aI389
aI899
aI6001
aI8129
aI19889
asS'd'
I2333
sa(dp4666
S'a'
I4186
sS's'
I15
sS'm'
(lp4667
I1
aI1
aI3
aI15
aI29
aI45
aI9
aI27
aI483
aI799
aI269
aI1811
aI4493
aI7109
aI22149
asS'd'
I2334
sa(dp4668
S'a'
I4198
sS's'
I15
sS'm'
(lp4669
I1
aI1
aI3
aI3
aI29
aI5
aI57
aI205
aI187
aI615
aI1677
aI3987
aI4577
aI8799
aI16311
asS'd'
I2335
sa(dp4670
S'a'
I4207
sS's'
I15
sS'm'
(lp4671
I1
aI1
aI5
aI3
aI15
aI5
aI91
aI101
aI319
aI445
aI1261
aI2039
aI4071
aI8249
aI11611
asS'd'
I2336
sa(dp4672
S'a'
I4221
sS's'
I15
sS'm'
(lp4673
I1
aI3
aI7
aI11
aI19
aI17
aI1
aI185
aI153
aI579
aI1001
aI2031
aI2295
aI16335
aI24771
asS'd'
I2337
sa(dp4674
S'a'
I4225
sS's'
I15
sS'm'
(lp4675
I1
aI3
aI3
aI15
aI13
aI45
aI93
aI185
aI319
aI667
aI1085
aI93
aI577
aI11551
aI11355
asS'd'
I2338
sa(dp4676
S'a'
I4231
sS's'
I15
sS'm'
(lp4677
I1
aI1
aI7
aI13
aI3
aI61
aI45
aI191
aI51
aI981
aI1151
aI2715
aI2503
aI4147
aI4587
asS'd'
I2339
sa(dp4678
S'a'
I4238
sS's'
I15
sS'm'
(lp4679
I1
aI1
aI3
aI3
aI27
aI17
aI71
aI141
aI57
aI981
aI1033
aI333
aI4639
aI15885
aI1039
asS'd'
I2340
sa(dp4680
S'a'
I4243
sS's'
I15
sS'm'
(lp4681
I1
aI3
aI3
aI15
aI21
aI55
aI33
aI123
aI357
aI893
aI829
aI4045
aI5027
aI11727
aI13357
asS'd'
I2341
sa(dp4682
S'a'
I4249
sS's'
I15
sS'm'
(lp4683
I1
aI1
aI1
aI9
aI31
aI47
aI27
aI223
aI311
aI205
aI179
aI3411
aI4019
aI10997
aI28115
asS'd'
I2342
sa(dp4684
S'a'
I4250
sS's'
I15
sS'm'
(lp4685
I1
aI3
aI5
aI1
aI3
aI39
aI15
aI7
aI501
aI641
aI735
aI295
aI2005
aI12641
aI19779
asS'd'
I2343
sa(dp4686
S'a'
I4252
sS's'
I15
sS'm'
(lp4687
I1
aI3
aI3
aI1
aI15
aI1
aI75
aI243
aI329
aI267
aI1323
aI2285
aI5389
aI11881
aI15737
asS'd'
I2344
sa(dp4688
S'a'
I4259
sS's'
I15
sS'm'
(lp4689
I1
aI1
aI3
aI3
aI13
aI17
aI101
aI99
aI209
aI939
aI1147
aI3221
aI5159
aI3435
aI183
asS'd'
I2345
sa(dp4690
S'a'
I4279
sS's'
I15
sS'm'
(lp4691
I1
aI1
aI1
aI1
aI27
aI43
aI29
aI179
aI179
aI659
aI807
aI313
aI4165
aI963
aI11317
asS'd'
I2346
sa(dp4692
S'a'
I4285
sS's'
I15
sS'm'
(lp4693
I1
aI1
aI3
aI13
aI9
aI51
aI125
aI245
aI381
aI555
aI1383
aI3887
aI2045
aI12829
aI12029
asS'd'
I2347
sa(dp4694
S'a'
I4288
sS's'
I15
sS'm'
(lp4695
I1
aI1
aI1
aI9
aI29
aI39
aI55
aI127
aI235
aI617
aI1553
aI3133
aI7735
aI14725
aI16733
asS'd'
I2348
sa(dp4696
S'a'
I4303
sS's'
I15
sS'm'
(lp4697
I1
aI1
aI3
aI5
aI15
aI9
aI47
aI217
aI89
aI987
aI1083
aI1045
aI4745
aI12915
aI13719
asS'd'
I2349
sa(dp4698
S'a'
I4312
sS's'
I15
sS'm'
(lp4699
I1
aI3
aI3
aI7
aI23
aI3
aI35
aI79
aI45
aI435
aI1549
aI2645
aI2831
aI10359
aI10041
asS'd'
I2350
sa(dp4700
S'a'
I4322
sS's'
I15
sS'm'
(lp4701
I1
aI1
aI7
aI15
aI31
aI61
aI25
aI223
aI511
aI319
aI487
aI1677
aI739
aI7097
aI18417
asS'd'
I2351
sa(dp4702
S'a'
I4327
sS's'
I15
sS'm'
(lp4703
I1
aI1
aI7
aI5
aI19
aI21
aI123
aI237
aI299
aI367
aI1341
aI1449
aI2949
aI8629
aI11051
asS'd'
I2352
sa(dp4704
S'a'
I4336
sS's'
I15
sS'm'
(lp4705
I1
aI3
aI7
aI7
aI31
aI53
aI125
aI33
aI257
aI719
aI1297
aI895
aI5095
aI10237
aI12309
asS'd'
I2353
sa(dp4706
S'a'
I4359
sS's'
I15
sS'm'
(lp4707
I1
aI3
aI1
aI5
aI31
aI59
aI73
aI211
aI97
aI209
aI1289
aI4033
aI6143
aI14275
aI7997
asS'd'
I2354
sa(dp4708
S'a'
I4384
sS's'
I15
sS'm'
(lp4709
I1
aI1
aI5
aI7
aI31
aI5
aI75
aI105
aI389
aI985
aI9
aI4033
aI1185
aI7821
aI19083
asS'd'
I2355
sa(dp4710
S'a'
I4387
sS's'
I15
sS'm'
(lp4711
I1
aI1
aI1
aI15
aI11
aI39
aI73
aI253
aI275
aI813
aI25
aI3441
aI2493
aI5873
aI3739
asS'd'
I2356
sa(dp4712
S'a'
I4401
sS's'
I15
sS'm'
(lp4713
I1
aI3
aI7
aI1
aI31
aI19
aI119
aI5
aI109
aI397
aI1329
aI3347
aI5941
aI12449
aI2533
asS'd'
I2357
sa(dp4714
S'a'
I4407
sS's'
I15
sS'm'
(lp4715
I1
aI1
aI1
aI1
aI5
aI59
aI61
aI175
aI435
aI985
aI65
aI3781
aI5425
aI15073
aI16361
asS'd'
I2358
sa(dp4716
S'a'
I4428
sS's'
I15
sS'm'
(lp4717
I1
aI3
aI5
aI7
aI31
aI13
aI53
aI87
aI69
aI305
aI1455
aI273
aI2197
aI4277
aI24423
asS'd'
I2359
sa(dp4718
S'a'
I4436
sS's'
I15
sS'm'
(lp4719
I1
aI3
aI3
aI15
aI13
aI13
aI91
aI171
aI71
aI583
aI15
aI3599
aI6801
aI10041
aI26097
asS'd'
I2360
sa(dp4720
S'a'
I4450
sS's'
I15
sS'm'
(lp4721
I1
aI3
aI3
aI5
aI5
aI13
aI91
aI225
aI63
aI69
aI1795
aI341
aI461
aI5015
aI9471
asS'd'
I2361
sa(dp4722
S'a'
I4452
sS's'
I15
sS'm'
(lp4723
I1
aI3
aI7
aI5
aI21
aI55
aI109
aI39
aI459
aI925
aI229
aI2855
aI5807
aI2117
aI31739
asS'd'
I2362
sa(dp4724
S'a'
I4459
sS's'
I15
sS'm'
(lp4725
I1
aI1
aI3
aI3
aI1
aI5
aI17
aI177
aI401
aI727
aI1555
aI3097
aI1243
aI5933
aI14579
asS'd'
I2363
sa(dp4726
S'a'
I4461
sS's'
I15
sS'm'
(lp4727
I1
aI1
aI7
aI3
aI19
aI19
aI37
aI87
aI105
aI73
aI197
aI4067
aI6237
aI10553
aI9207
asS'd'
I2364
sa(dp4728
S'a'
I4470
sS's'
I15
sS'm'
(lp4729
I1
aI1
aI3
aI15
aI1
aI55
aI119
aI115
aI441
aI3
aI1003
aI1631
aI197
aI12929
aI25385
asS'd'
I2365
sa(dp4730
S'a'
I4483
sS's'
I15
sS'm'
(lp4731
I1
aI3
aI7
aI11
aI31
aI1
aI119
aI49
aI467
aI647
aI685
aI2771
aI3689
aI11049
aI26787
asS'd'
I2366
sa(dp4732
S'a'
I4485
sS's'
I15
sS'm'
(lp4733
I1
aI1
aI1
aI11
aI19
aI19
aI21
aI73
aI459
aI935
aI615
aI371
aI1099
aI14407
aI10375
asS'd'
I2367
sa(dp4734
S'a'
I4486
sS's'
I15
sS'm'
(lp4735
I1
aI3
aI5
aI13
aI15
aI3
aI107
aI179
aI259
aI677
aI1101
aI315
aI7673
aI14639
aI11241
asS'd'
I2368
sa(dp4736
S'a'
I4492
sS's'
I15
sS'm'
(lp4737
I1
aI1
aI7
aI9
aI15
aI21
aI93
aI25
aI349
aI23
aI1087
aI27
aI5691
aI12997
aI29301
asS'd'
I2369
sa(dp4738
S'a'
I4497
sS's'
I15
sS'm'
(lp4739
I1
aI3
aI3
aI5
aI7
aI43
aI1
aI195
aI69
aI753
aI1315
aI2629
aI3259
aI5595
aI19439
asS'd'
I2370
sa(dp4740
S'a'
I4514
sS's'
I15
sS'm'
(lp4741
I1
aI3
aI5
aI5
aI31
aI9
aI75
aI217
aI217
aI197
aI1925
aI2033
aI3585
aI15219
aI20251
asS'd'
I2371
sa(dp4742
S'a'
I4533
sS's'
I15
sS'm'
(lp4743
I1
aI1
aI5
aI11
aI17
aI31
aI3
aI209
aI315
aI49
aI949
aI2267
aI4611
aI4375
aI16431
asS'd'
I2372
sa(dp4744
S'a'
I4537
sS's'
I15
sS'm'
(lp4745
I1
aI1
aI7
aI9
aI17
aI35
aI13
aI115
aI119
aI553
aI1527
aI2857
aI3599
aI391
aI25101
asS'd'
I2373
sa(dp4746
S'a'
I4546
sS's'
I15
sS'm'
(lp4747
I1
aI3
aI3
aI15
aI13
aI59
aI17
aI177
aI301
aI719
aI909
aI1663
aI5033
aI1129
aI529
asS'd'
I2374
sa(dp4748
S'a'
I4551
sS's'
I15
sS'm'
(lp4749
I1
aI1
aI7
aI5
aI15
aI13
aI99
aI157
aI379
aI975
aI1019
aI2251
aI3807
aI10621
aI351
asS'd'
I2375
sa(dp4750
S'a'
I4555
sS's'
I15
sS'm'
(lp4751
I1
aI3
aI3
aI13
aI5
aI57
aI5
aI31
aI361
aI981
aI883
aI3723
aI2259
aI5151
aI11783
asS'd'
I2376
sa(dp4752
S'a'
I4560
sS's'
I15
sS'm'
(lp4753
I1
aI1
aI1
aI13
aI1
aI43
aI125
aI19
aI77
aI509
aI1817
aI3795
aI1863
aI8401
aI27253
asS'd'
I2377
sa(dp4754
S'a'
I4569
sS's'
I15
sS'm'
(lp4755
I1
aI1
aI5
aI7
aI19
aI41
aI21
aI151
aI89
aI189
aI769
aI1937
aI4497
aI13607
aI24691
asS'd'
I2378
sa(dp4756
S'a'
I4576
sS's'
I15
sS'm'
(lp4757
I1
aI1
aI1
aI9
aI21
aI9
aI1
aI195
aI31
aI907
aI1719
aI1549
aI809
aI13629
aI16597
asS'd'
I2379
sa(dp4758
S'a'
I4582
sS's'
I15
sS'm'
(lp4759
I1
aI1
aI1
aI3
aI21
aI61
aI103
aI219
aI311
aI849
aI523
aI21
aI4533
aI6367
aI3935
asS'd'
I2380
sa(dp4760
S'a'
I4586
sS's'
I15
sS'm'
(lp4761
I1
aI1
aI7
aI9
aI7
aI33
aI77
aI19
aI489
aI933
aI1729
aI1813
aI6741
aI10701
aI7
asS'd'
I2381
sa(dp4762
S'a'
I4609
sS's'
I15
sS'm'
(lp4763
I1
aI1
aI1
aI5
aI23
aI53
aI43
aI63
aI453
aI209
aI1313
aI2847
aI2641
aI13783
aI14983
asS'd'
I2382
sa(dp4764
S'a'
I4610
sS's'
I15
sS'm'
(lp4765
I1
aI3
aI7
aI7
aI15
aI45
aI83
aI241
aI509
aI659
aI213
aI221
aI5205
aI6019
aI18945
asS'd'
I2383
sa(dp4766
S'a'
I4612
sS's'
I15
sS'm'
(lp4767
I1
aI1
aI5
aI9
aI25
aI43
aI37
aI9
aI191
aI505
aI765
aI295
aI953
aI1045
aI11203
asS'd'
I2384
sa(dp4768
S'a'
I4649
sS's'
I15
sS'm'
(lp4769
I1
aI3
aI7
aI11
aI5
aI49
aI45
aI177
aI379
aI695
aI355
aI1711
aI7747
aI497
aI7597
asS'd'
I2385
sa(dp4770
S'a'
I4652
sS's'
I15
sS'm'
(lp4771
I1
aI1
aI5
aI13
aI23
aI47
aI101
aI145
aI301
aI207
aI195
aI2225
aI8093
aI15345
aI14097
asS'd'
I2386
sa(dp4772
S'a'
I4672
sS's'
I15
sS'm'
(lp4773
I1
aI3
aI7
aI13
aI9
aI9
aI55
aI223
aI343
aI921
aI1825
aI3281
aI2627
aI855
aI27651
asS'd'
I2387
sa(dp4774
S'a'
I4677
sS's'
I15
sS'm'
(lp4775
I1
aI1
aI7
aI1
aI21
aI1
aI67
aI149
aI433
aI111
aI577
aI3675
aI495
aI9043
aI23613
asS'd'
I2388
sa(dp4776
S'a'
I4684
sS's'
I15
sS'm'
(lp4777
I1
aI3
aI1
aI13
aI9
aI39
aI37
aI73
aI117
aI559
aI1131
aI2511
aI7599
aI8393
aI24747
asS'd'
I2389
sa(dp4778
S'a'
I4690
sS's'
I15
sS'm'
(lp4779
I1
aI3
aI3
aI7
aI11
aI15
aI85
aI229
aI7
aI21
aI1649
aI739
aI375
aI13991
aI27053
asS'd'
I2390
sa(dp4780
S'a'
I4695
sS's'
I15
sS'm'
(lp4781
I1
aI1
aI5
aI5
aI15
aI41
aI49
aI117
aI173
aI825
aI1343
aI377
aI1789
aI12519
aI30667
asS'd'
I2391
sa(dp4782
S'a'
I4696
sS's'
I15
sS'm'
(lp4783
I1
aI1
aI7
aI15
aI9
aI11
aI97
aI99
aI347
aI729
aI9
aI1703
aI1177
aI5189
aI9061
asS'd'
I2392
sa(dp4784
S'a'
I4702
sS's'
I15
sS'm'
(lp4785
I1
aI1
aI5
aI11
aI15
aI25
aI99
aI63
aI89
aI675
aI561
aI215
aI8111
aI3955
aI24635
asS'd'
I2393
sa(dp4786
S'a'
I4705
sS's'
I15
sS'm'
(lp4787
I1
aI1
aI1
aI1
aI7
aI53
aI99
aI193
aI233
aI731
aI733
aI1883
aI7783
aI14413
aI14003
asS'd'
I2394
sa(dp4788
S'a'
I4717
sS's'
I15
sS'm'
(lp4789
I1
aI3
aI5
aI7
aI31
aI23
aI45
aI153
aI337
aI293
aI443
aI2301
aI5135
aI7455
aI13123
asS'd'
I2395
sa(dp4790
S'a'
I4726
sS's'
I15
sS'm'
(lp4791
I1
aI3
aI1
aI3
aI23
aI53
aI23
aI165
aI53
aI875
aI1543
aI1035
aI4247
aI5101
aI28445
asS'd'
I2396
sa(dp4792
S'a'
I4736
sS's'
I15
sS'm'
(lp4793
I1
aI1
aI1
aI15
aI13
aI41
aI77
aI93
aI205
aI743
aI1101
aI1413
aI2371
aI7183
aI12337
asS'd'
I2397
sa(dp4794
S'a'
I4753
sS's'
I15
sS'm'
(lp4795
I1
aI1
aI3
aI15
aI17
aI63
aI25
aI101
aI147
aI149
aI1207
aI3525
aI2661
aI9539
aI11145
asS'd'
I2398
sa(dp4796
S'a'
I4754
sS's'
I15
sS'm'
(lp4797
I1
aI3
aI1
aI9
aI17
aI5
aI3
aI35
aI389
aI909
aI1017
aI2803
aI5243
aI13025
aI8851
asS'd'
I2399
sa(dp4798
S'a'
I4756
sS's'
I15
sS'm'
(lp4799
I1
aI1
aI7
aI15
aI19
aI27
aI69
aI91
aI71
aI547
aI1421
aI831
aI6969
aI5517
aI28233
asS'd'
I2400
sa(dp4800
S'a'
I4775
sS's'
I15
sS'm'
(lp4801
I1
aI1
aI3
aI3
aI17
aI45
aI55
aI63
aI263
aI819
aI1211
aI2739
aI655
aI13269
aI22281
asS'd'
I2401
sa(dp4802
S'a'
I4801
sS's'
I15
sS'm'
(lp4803
I1
aI3
aI1
aI5
aI23
aI13
aI81
aI251
aI83
aI551
aI491
aI1029
aI3561
aI357
aI23393
asS'd'
I2402
sa(dp4804
S'a'
I4819
sS's'
I15
sS'm'
(lp4805
I1
aI3
aI1
aI13
aI25
aI27
aI93
aI143
aI407
aI403
aI1395
aI1733
aI3187
aI1917
aI31453
asS'd'
I2403
sa(dp4806
S'a'
I4828
sS's'
I15
sS'm'
(lp4807
I1
aI1
aI7
aI13
aI3
aI21
aI85
aI113
aI483
aI461
aI1343
aI561
aI2081
aI10857
aI24253
asS'd'
I2404
sa(dp4808
S'a'
I4838
sS's'
I15
sS'm'
(lp4809
I1
aI1
aI1
aI1
aI11
aI11
aI53
aI135
aI25
aI163
aI1729
aI617
aI1533
aI10881
aI16041
asS'd'
I2405
sa(dp4810
S'a'
I4852
sS's'
I15
sS'm'
(lp4811
I1
aI1
aI5
aI1
aI3
aI49
aI125
aI139
aI77
aI891
aI815
aI3431
aI4875
aI12513
aI4595
asS'd'
I2406
sa(dp4812
S'a'
I4856
sS's'
I15
sS'm'
(lp4813
I1
aI1
aI1
aI1
aI27
aI63
aI111
aI109
aI421
aI425
aI345
aI1613
aI5447
aI1357
aI32413
asS'd'
I2407
sa(dp4814
S'a'
I4873
sS's'
I15
sS'm'
(lp4815
I1
aI3
aI5
aI3
aI17
aI5
aI37
aI171
aI259
aI281
aI1003
aI2901
aI3241
aI15557
aI21415
asS'd'
I2408
sa(dp4816
S'a'
I4887
sS's'
I15
sS'm'
(lp4817
I1
aI1
aI5
aI11
aI15
aI55
aI75
aI199
aI493
aI215
aI1625
aI2345
aI7873
aI2325
aI11003
asS'd'
I2409
sa(dp4818
S'a'
I4891
sS's'
I15
sS'm'
(lp4819
I1
aI3
aI7
aI1
aI21
aI33
aI23
aI5
aI495
aI941
aI1185
aI475
aI5799
aI15161
aI10677
asS'd'
I2410
sa(dp4820
S'a'
I4904
sS's'
I15
sS'm'
(lp4821
I1
aI1
aI5
aI9
aI31
aI37
aI37
aI29
aI217
aI389
aI297
aI3097
aI7319
aI2601
aI15307
asS'd'
I2411
sa(dp4822
S'a'
I4912
sS's'
I15
sS'm'
(lp4823
I1
aI3
aI7
aI5
aI7
aI45
aI111
aI167
aI297
aI275
aI1669
aI2489
aI1511
aI15753
aI1289
asS'd'
I2412
sa(dp4824
S'a'
I4921
sS's'
I15
sS'm'
(lp4825
I1
aI3
aI1
aI7
aI3
aI45
aI19
aI11
aI189
aI199
aI1227
aI2647
aI1897
aI9077
aI17189
asS'd'
I2413
sa(dp4826
S'a'
I4936
sS's'
I15
sS'm'
(lp4827
I1
aI1
aI1
aI13
aI15
aI39
aI19
aI179
aI147
aI341
aI283
aI3029
aI7599
aI8937
aI18761
asS'd'
I2414
sa(dp4828
S'a'
I4941
sS's'
I15
sS'm'
(lp4829
I1
aI3
aI3
aI9
aI3
aI11
aI41
aI255
aI365
aI835
aI921
aI389
aI919
aI15223
aI14541
asS'd'
I2415
sa(dp4830
S'a'
I4942
sS's'
I15
sS'm'
(lp4831
I1
aI1
aI3
aI3
aI5
aI37
aI29
aI203
aI313
aI271
aI1207
aI487
aI3711
aI3811
aI26757
asS'd'
I2416
sa(dp4832
S'a'
I4963
sS's'
I15
sS'm'
(lp4833
I1
aI3
aI7
aI9
aI19
aI53
aI49
aI139
aI351
aI537
aI1681
aI1595
aI5399
aI13839
aI28685
asS'd'
I2417
sa(dp4834
S'a'
I4984
sS's'
I15
sS'm'
(lp4835
I1
aI3
aI1
aI1
aI15
aI35
aI21
aI37
aI247
aI891
aI1855
aI1243
aI3137
aI10381
aI30379
asS'd'
I2418
sa(dp4836
S'a'
I4990
sS's'
I15
sS'm'
(lp4837
I1
aI3
aI7
aI5
aI9
aI47
aI91
aI25
aI479
aI337
aI781
aI3545
aI1045
aI9491
aI22853
asS'd'
I2419
sa(dp4838
S'a'
I5005
sS's'
I15
sS'm'
(lp4839
I1
aI1
aI5
aI15
aI19
aI31
aI81
aI5
aI117
aI923
aI565
aI2443
aI7383
aI1795
aI11685
asS'd'
I2420
sa(dp4840
S'a'
I5013
sS's'
I15
sS'm'
(lp4841
I1
aI3
aI3
aI5
aI17
aI15
aI21
aI245
aI489
aI889
aI2047
aI2737
aI7445
aI14785
aI13401
asS'd'
I2421
sa(dp4842
S'a'
I5020
sS's'
I15
sS'm'
(lp4843
I1
aI1
aI1
aI15
aI19
aI45
aI67
aI117
aI299
aI607
aI953
aI743
aI6863
aI12123
aI6701
asS'd'
I2422
sa(dp4844
S'a'
I5039
sS's'
I15
sS'm'
(lp4845
I1
aI1
aI3
aI1
aI1
aI43
aI19
aI129
aI345
aI861
aI209
aI2387
aI7205
aI7131
aI8235
asS'd'
I2423
sa(dp4846
S'a'
I5048
sS's'
I15
sS'm'
(lp4847
I1
aI3
aI5
aI1
aI1
aI13
aI75
aI99
aI333
aI157
aI23
aI1217
aI1857
aI15479
aI16031
asS'd'
I2424
sa(dp4848
S'a'
I5062
sS's'
I15
sS'm'
(lp4849
I1
aI3
aI3
aI11
aI7
aI61
aI119
aI89
aI491
aI401
aI227
aI1739
aI3807
aI16003
aI2875
asS'd'
I2425
sa(dp4850
S'a'
I5080
sS's'
I15
sS'm'
(lp4851
I1
aI3
aI7
aI15
aI13
aI55
aI3
aI159
aI405
aI593
aI975
aI361
aI2563
aI6061
aI28087
asS'd'
I2426
sa(dp4852
S'a'
I5085
sS's'
I15
sS'm'
(lp4853
I1
aI1
aI3
aI13
aI19
aI5
aI5
aI9
aI119
aI41
aI33
aI1111
aI4443
aI4663
aI28841
asS'd'
I2427
sa(dp4854
S'a'
I5086
sS's'
I15
sS'm'
(lp4855
I1
aI1
aI7
aI7
aI25
aI59
aI125
aI255
aI49
aI947
aI1673
aI2947
aI6369
aI2267
aI8813
asS'd'
I2428
sa(dp4856
S'a'
I5095
sS's'
I15
sS'm'
(lp4857
I1
aI1
aI5
aI15
aI25
aI25
aI111
aI193
aI217
aI193
aI821
aI2779
aI69
aI2957
aI27043
asS'd'
I2429
sa(dp4858
S'a'
I5096
sS's'
I15
sS'm'
(lp4859
I1
aI3
aI5
aI7
aI21
aI19
aI51
aI157
aI203
aI487
aI1745
aI1875
aI911
aI14071
aI7557
asS'd'
I2430
sa(dp4860
S'a'
I5102
sS's'
I15
sS'm'
(lp4861
I1
aI1
aI5
aI9
aI3
aI15
aI55
aI73
aI313
aI245
aI1061
aI1929
aI3035
aI607
aI11563
asS'd'
I2431
sa(dp4862
S'a'
I5107
sS's'
I15
sS'm'
(lp4863
I1
aI1
aI5
aI7
aI3
aI57
aI105
aI121
aI461
aI43
aI803
aI1801
aI4059
aI2157
aI17547
asS'd'
I2432
sa(dp4864
S'a'
I5141
sS's'
I15
sS'm'
(lp4865
I1
aI3
aI7
aI7
aI19
aI11
aI1
aI121
aI499
aI841
aI601
aI3515
aI2969
aI13697
aI8917
asS'd'
I2433
sa(dp4866
S'a'
I5145
sS's'
I15
sS'm'
(lp4867
I1
aI3
aI3
aI3
aI13
aI35
aI113
aI231
aI391
aI689
aI697
aI2871
aI7387
aI715
aI27005
asS'd'
I2434
sa(dp4868
S'a'
I5148
sS's'
I15
sS'm'
(lp4869
I1
aI1
aI1
aI13
aI19
aI5
aI17
aI43
aI175
aI291
aI987
aI1917
aI7635
aI15655
aI10689
asS'd'
I2435
sa(dp4870
S'a'
I5157
sS's'
I15
sS'm'
(lp4871
I1
aI1
aI7
aI15
aI19
aI37
aI121
aI243
aI125
aI623
aI1231
aI29
aI2325
aI5147
aI21435
asS'd'
I2436
sa(dp4872
S'a'
I5158
sS's'
I15
sS'm'
(lp4873
I1
aI3
aI5
aI15
aI25
aI27
aI57
aI187
aI77
aI401
aI1489
aI2977
aI5415
aI3381
aI2551
asS'd'
I2437
sa(dp4874
S'a'
I5162
sS's'
I15
sS'm'
(lp4875
I1
aI1
aI1
aI7
aI1
aI1
aI85
aI27
aI115
aI559
aI9
aI2365
aI711
aI5733
aI2819
asS'd'
I2438
sa(dp4876
S'a'
I5172
sS's'
I15
sS'm'
(lp4877
I1
aI3
aI1
aI15
aI9
aI29
aI61
aI113
aI169
aI349
aI591
aI1061
aI6041
aI7613
aI23691
asS'd'
I2439
sa(dp4878
S'a'
I5182
sS's'
I15
sS'm'
(lp4879
I1
aI1
aI5
aI1
aI13
aI45
aI49
aI227
aI345
aI563
aI87
aI3597
aI3961
aI7205
aI8441
asS'd'
I2440
sa(dp4880
S'a'
I5184
sS's'
I15
sS'm'
(lp4881
I1
aI1
aI1
aI5
aI3
aI21
aI121
aI183
aI463
aI83
aI1365
aI539
aI1485
aI10063
aI24867
asS'd'
I2441
sa(dp4882
S'a'
I5193
sS's'
I15
sS'm'
(lp4883
I1
aI3
aI5
aI5
aI3
aI61
aI101
aI237
aI41
aI147
aI1907
aI3049
aI7583
aI8283
aI6099
asS'd'
I2442
sa(dp4884
S'a'
I5199
sS's'
I15
sS'm'
(lp4885
I1
aI3
aI1
aI15
aI31
aI57
aI19
aI155
aI445
aI805
aI1793
aI207
aI1975
aI3357
aI14281
asS'd'
I2443
sa(dp4886
S'a'
I5201
sS's'
I15
sS'm'
(lp4887
I1
aI1
aI7
aI13
aI9
aI39
aI27
aI73
aI165
aI345
aI543
aI4095
aI133
aI10469
aI11573
asS'd'
I2444
sa(dp4888
S'a'
I5204
sS's'
I15
sS'm'
(lp4889
I1
aI1
aI7
aI15
aI17
aI57
aI99
aI81
aI359
aI367
aI1057
aI1173
aI4225
aI15127
aI2615
asS'd'
I2445
sa(dp4890
S'a'
I5211
sS's'
I15
sS'm'
(lp4891
I1
aI3
aI5
aI3
aI31
aI23
aI113
aI111
aI495
aI947
aI1625
aI1195
aI2053
aI1509
aI1347
asS'd'
I2446
sa(dp4892
S'a'
I5223
sS's'
I15
sS'm'
(lp4893
I1
aI1
aI5
aI5
aI9
aI47
aI25
aI63
aI455
aI107
aI771
aI3815
aI3827
aI16287
aI11615
asS'd'
I2447
sa(dp4894
S'a'
I5230
sS's'
I15
sS'm'
(lp4895
I1
aI1
aI7
aI9
aI17
aI61
aI51
aI215
aI63
aI123
aI1253
aI3927
aI721
aI9647
aI3283
asS'd'
I2448
sa(dp4896
S'a'
I5232
sS's'
I15
sS'm'
(lp4897
I1
aI1
aI5
aI15
aI11
aI17
aI83
aI255
aI473
aI107
aI681
aI763
aI7855
aI8043
aI31503
asS'd'
I2449
sa(dp4898
S'a'
I5253
sS's'
I15
sS'm'
(lp4899
I1
aI3
aI1
aI7
aI7
aI31
aI37
aI5
aI253
aI155
aI2017
aI609
aI1421
aI14927
aI25241
asS'd'
I2450
sa(dp4900
S'a'
I5257
sS's'
I15
sS'm'
(lp4901
I1
aI3
aI3
aI13
aI31
aI25
aI21
aI241
aI431
aI193
aI681
aI2265
aI5091
aI11479
aI21443
asS'd'
I2451
sa(dp4902
S'a'
I5260
sS's'
I15
sS'm'
(lp4903
I1
aI3
aI5
aI5
aI15
aI9
aI49
aI255
aI157
aI995
aI631
aI1995
aI3605
aI9085
aI24245
asS'd'
I2452
sa(dp4904
S'a'
I5284
sS's'
I15
sS'm'
(lp4905
I1
aI3
aI3
aI7
aI19
aI31
aI85
aI153
aI493
aI951
aI451
aI1587
aI6609
aI3681
aI13205
asS'd'
I2453
sa(dp4906
S'a'
I5306
sS's'
I15
sS'm'
(lp4907
I1
aI1
aI5
aI1
aI17
aI41
aI107
aI231
aI307
aI361
aI575
aI3239
aI3443
aI16159
aI20625
asS'd'
I2454
sa(dp4908
S'a'
I5331
sS's'
I15
sS'm'
(lp4909
I1
aI1
aI7
aI9
aI31
aI49
aI93
aI79
aI181
aI117
aI1241
aI3645
aI4901
aI12599
aI13247
asS'd'
I2455
sa(dp4910
S'a'
I5334
sS's'
I15
sS'm'
(lp4911
I1
aI3
aI3
aI9
aI7
aI31
aI127
aI201
aI11
aI199
aI1851
aI23
aI5667
aI8159
aI20951
asS'd'
I2456
sa(dp4912
S'a'
I5364
sS's'
I15
sS'm'
(lp4913
I1
aI3
aI3
aI7
aI3
aI37
aI29
aI189
aI65
aI461
aI769
aI321
aI6577
aI16223
aI16865
asS'd'
I2457
sa(dp4914
S'a'
I5367
sS's'
I15
sS'm'
(lp4915
I1
aI1
aI5
aI11
aI1
aI13
aI91
aI167
aI33
aI111
aI1445
aI1047
aI2479
aI12623
aI22893
asS'd'
I2458
sa(dp4916
S'a'
I5371
sS's'
I15
sS'm'
(lp4917
I1
aI1
aI3
aI1
aI3
aI1
aI47
aI185
aI329
aI903
aI1651
aI3005
aI907
aI1255
aI8303
asS'd'
I2459
sa(dp4918
S'a'
I5382
sS's'
I15
sS'm'
(lp4919
I1
aI3
aI5
aI13
aI19
aI31
aI5
aI233
aI265
aI769
aI1303
aI2503
aI2229
aI14019
aI20257
asS'd'
I2460
sa(dp4920
S'a'
I5386
sS's'
I15
sS'm'
(lp4921
I1
aI3
aI7
aI3
aI27
aI11
aI67
aI195
aI5
aI661
aI125
aI3761
aI7211
aI16043
aI7267
asS'd'
I2461
sa(dp4922
S'a'
I5399
sS's'
I15
sS'm'
(lp4923
I1
aI1
aI1
aI3
aI27
aI13
aI115
aI25
aI473
aI417
aI1751
aI2223
aI2099
aI5913
aI14273
asS'd'
I2462
sa(dp4924
S'a'
I5400
sS's'
I15
sS'm'
(lp4925
I1
aI3
aI7
aI15
aI13
aI53
aI99
aI115
aI225
aI737
aI1621
aI539
aI4131
aI471
aI31865
asS'd'
I2463
sa(dp4926
S'a'
I5409
sS's'
I15
sS'm'
(lp4927
I1
aI1
aI5
aI5
aI25
aI19
aI39
aI207
aI153
aI569
aI1755
aI2477
aI3065
aI7383
aI29919
asS'd'
I2464
sa(dp4928
S'a'
I5415
sS's'
I15
sS'm'
(lp4929
I1
aI3
aI5
aI11
aI13
aI59
aI33
aI3
aI435
aI273
aI701
aI3819
aI7291
aI11803
aI26111
asS'd'
I2465
sa(dp4930
S'a'
I5416
sS's'
I15
sS'm'
(lp4931
I1
aI1
aI3
aI9
aI29
aI19
aI71
aI59
aI93
aI1019
aI887
aI83
aI4675
aI7541
aI26821
asS'd'
I2466
sa(dp4932
S'a'
I5424
sS's'
I15
sS'm'
(lp4933
I1
aI3
aI1
aI3
aI21
aI53
aI71
aI73
aI43
aI321
aI1581
aI1399
aI4043
aI12995
aI16825
asS'd'
I2467
sa(dp4934
S'a'
I5436
sS's'
I15
sS'm'
(lp4935
I1
aI3
aI7
aI15
aI3
aI13
aI37
aI11
aI93
aI873
aI1193
aI3481
aI451
aI15869
aI17879
asS'd'
I2468
sa(dp4936
S'a'
I5454
sS's'
I15
sS'm'
(lp4937
I1
aI3
aI1
aI11
aI31
aI19
aI101
aI57
aI129
aI753
aI853
aI463
aI6757
aI11083
aI8667
asS'd'
I2469
sa(dp4938
S'a'
I5462
sS's'
I15
sS'm'
(lp4939
I1
aI3
aI5
aI15
aI25
aI41
aI25
aI197
aI235
aI609
aI905
aI993
aI3233
aI1935
aI24661
asS'd'
I2470
sa(dp4940
S'a'
I5468
sS's'
I15
sS'm'
(lp4941
I1
aI3
aI1
aI5
aI21
aI7
aI53
aI107
aI473
aI77
aI1135
aI1045
aI4933
aI5615
aI15931
asS'd'
I2471
sa(dp4942
S'a'
I5481
sS's'
I15
sS'm'
(lp4943
I1
aI3
aI7
aI11
aI3
aI9
aI105
aI183
aI151
aI527
aI425
aI975
aI4073
aI913
aI2793
asS'd'
I2472
sa(dp4944
S'a'
I5505
sS's'
I15
sS'm'
(lp4945
I1
aI1
aI7
aI13
aI19
aI61
aI81
aI9
aI413
aI851
aI1723
aI1113
aI1453
aI8635
aI3353
asS'd'
I2473
sa(dp4946
S'a'
I5511
sS's'
I15
sS'm'
(lp4947
I1
aI3
aI7
aI15
aI19
aI53
aI83
aI31
aI441
aI343
aI575
aI935
aI4543
aI1303
aI12567
asS'd'
I2474
sa(dp4948
S'a'
I5518
sS's'
I15
sS'm'
(lp4949
I1
aI1
aI1
aI5
aI29
aI19
aI119
aI75
aI3
aI591
aI845
aI649
aI1717
aI13695
aI26905
asS'd'
I2475
sa(dp4950
S'a'
I5530
sS's'
I15
sS'm'
(lp4951
I1
aI1
aI7
aI9
aI5
aI53
aI127
aI191
aI15
aI773
aI1433
aI2899
aI21
aI4977
aI17839
asS'd'
I2476
sa(dp4952
S'a'
I5532
sS's'
I15
sS'm'
(lp4953
I1
aI1
aI5
aI9
aI21
aI9
aI99
aI115
aI397
aI99
aI725
aI3835
aI973
aI1219
aI21159
asS'd'
I2477
sa(dp4954
S'a'
I5539
sS's'
I15
sS'm'
(lp4955
I1
aI3
aI5
aI3
aI7
aI39
aI29
aI93
aI303
aI913
aI981
aI3549
aI5225
aI10907
aI393
asS'd'
I2478
sa(dp4956
S'a'
I5553
sS's'
I15
sS'm'
(lp4957
I1
aI3
aI3
aI11
aI9
aI25
aI105
aI101
aI1
aI867
aI389
aI2241
aI773
aI14123
aI10015
asS'd'
I2479
sa(dp4958
S'a'
I5573
sS's'
I15
sS'm'
(lp4959
I1
aI1
aI5
aI1
aI1
aI37
aI117
aI213
aI391
aI779
aI1851
aI1485
aI1277
aI5607
aI819
asS'd'
I2480
sa(dp4960
S'a'
I5580
sS's'
I15
sS'm'
(lp4961
I1
aI3
aI7
aI1
aI3
aI5
aI43
aI47
aI483
aI367
aI749
aI1693
aI4961
aI15257
aI3775
asS'd'
I2481
sa(dp4962
S'a'
I5597
sS's'
I15
sS'm'
(lp4963
I1
aI3
aI3
aI1
aI27
aI11
aI21
aI83
aI437
aI379
aI1041
aI393
aI5611
aI2421
aI31739
asS'd'
I2482
sa(dp4964
S'a'
I5602
sS's'
I15
sS'm'
(lp4965
I1
aI3
aI5
aI7
aI19
aI1
aI79
aI63
aI53
aI201
aI1159
aI2501
aI6327
aI11317
aI9537
asS'd'
I2483
sa(dp4966
S'a'
I5608
sS's'
I15
sS'm'
(lp4967
I1
aI3
aI5
aI13
aI9
aI37
aI61
aI217
aI427
aI913
aI1311
aI3503
aI5473
aI10583
aI19723
asS'd'
I2484
sa(dp4968
S'a'
I5611
sS's'
I15
sS'm'
(lp4969
I1
aI1
aI3
aI9
aI11
aI29
aI121
aI175
aI141
aI515
aI925
aI837
aI6011
aI10419
aI32157
asS'd'
I2485
sa(dp4970
S'a'
I5613
sS's'
I15
sS'm'
(lp4971
I1
aI3
aI5
aI9
aI27
aI57
aI97
aI175
aI365
aI367
aI1737
aI3845
aI1257
aI12243
aI2201
asS'd'
I2486
sa(dp4972
S'a'
I5625
sS's'
I15
sS'm'
(lp4973
I1
aI3
aI3
aI9
aI23
aI1
aI53
aI123
aI127
aI333
aI1335
aI707
aI5747
aI6541
aI9809
asS'd'
I2487
sa(dp4974
S'a'
I5632
sS's'
I15
sS'm'
(lp4975
I1
aI3
aI1
aI9
aI17
aI37
aI101
aI41
aI91
aI61
aI433
aI979
aI4345
aI12351
aI10829
asS'd'
I2488
sa(dp4976
S'a'
I5635
sS's'
I15
sS'm'
(lp4977
I1
aI3
aI3
aI13
aI3
aI21
aI15
aI49
aI257
aI99
aI1793
aI2987
aI5233
aI11625
aI28069
asS'd'
I2489
sa(dp4978
S'a'
I5638
sS's'
I15
sS'm'
(lp4979
I1
aI1
aI7
aI11
aI21
aI13
aI89
aI11
aI135
aI153
aI783
aI2893
aI6815
aI12007
aI15605
asS'd'
I2490
sa(dp4980
S'a'
I5652
sS's'
I15
sS'm'
(lp4981
I1
aI3
aI7
aI13
aI5
aI61
aI73
aI5
aI269
aI699
aI925
aI2925
aI5919
aI5841
aI24875
asS'd'
I2491
sa(dp4982
S'a'
I5659
sS's'
I15
sS'm'
(lp4983
I1
aI3
aI5
aI5
aI25
aI45
aI43
aI93
aI15
aI927
aI1253
aI319
aI1173
aI14559
aI20221
asS'd'
I2492
sa(dp4984
S'a'
I5677
sS's'
I15
sS'm'
(lp4985
I1
aI1
aI3
aI3
aI27
aI45
aI9
aI103
aI447
aI627
aI1239
aI3869
aI2169
aI49
aI17917
asS'd'
I2493
sa(dp4986
S'a'
I5686
sS's'
I15
sS'm'
(lp4987
I1
aI3
aI7
aI7
aI11
aI9
aI1
aI1
aI1
aI527
aI825
aI3295
aI623
aI2095
aI10537
asS'd'
I2494
sa(dp4988
S'a'
I5689
sS's'
I15
sS'm'
(lp4989
I1
aI3
aI3
aI11
aI21
aI11
aI59
aI165
aI33
aI743
aI1461
aI1535
aI6393
aI1301
aI17823
asS'd'
I2495
sa(dp4990
S'a'
I5698
sS's'
I15
sS'm'
(lp4991
I1
aI1
aI7
aI3
aI19
aI43
aI47
aI245
aI469
aI551
aI1447
aI1963
aI169
aI1481
aI31925
asS'd'
I2496
sa(dp4992
S'a'
I5703
sS's'
I15
sS'm'
(lp4993
I1
aI1
aI3
aI1
aI11
aI21
aI51
aI7
aI251
aI199
aI1153
aI767
aI6417
aI3417
aI30171
asS'd'
I2497
sa(dp4994
S'a'
I5707
sS's'
I15
sS'm'
(lp4995
I1
aI3
aI7
aI1
aI31
aI5
aI41
aI103
aI447
aI263
aI211
aI2029
aI8021
aI4705
aI10579
asS'd'
I2498
sa(dp4996
S'a'
I5731
sS's'
I15
sS'm'
(lp4997
I1
aI1
aI3
aI5
aI17
aI25
aI55
aI75
aI393
aI107
aI2017
aI2389
aI1685
aI14021
aI9161
asS'd'
I2499
sa(dp4998
S'a'
I5738
sS's'
I15
sS'm'
(lp4999
I1
aI1
aI1
aI9
aI13
aI1
aI75
aI237
aI205
aI461
aI689
aI2531
aI2839
aI13925
aI23351
asS'd'
I2500
sa(dp5000
S'a'
I5743
sS's'
I15
sS'm'
(lp5001
I1
aI3
aI7
aI1
aI23
aI39
aI33
aI189
aI157
aI571
aI239
aI1053
aI1559
aI1685
aI23059
asS'd'
I2501
sa(dp5002
S'a'
I5748
sS's'
I15
sS'm'
(lp5003
I1
aI3
aI3
aI3
aI27
aI61
aI71
aI121
aI49
aI157
aI1341
aI1707
aI2417
aI11689
aI26507
asS'd'
I2502
sa(dp5004
S'a'
I5758
sS's'
I15
sS'm'
(lp5005
I1
aI3
aI7
aI7
aI19
aI63
aI47
aI53
aI95
aI791
aI1467
aI1273
aI2045
aI755
aI8555
asS'd'
I2503
sa(dp5006
S'a'
I5762
sS's'
I15
sS'm'
(lp5007
I1
aI1
aI3
aI15
aI27
aI33
aI21
aI253
aI317
aI153
aI1509
aI1765
aI3809
aI601
aI5907
asS'd'
I2504
sa(dp5008
S'a'
I5768
sS's'
I15
sS'm'
(lp5009
I1
aI3
aI5
aI15
aI11
aI17
aI97
aI91
aI165
aI199
aI1751
aI2135
aI1315
aI3077
aI29995
asS'd'
I2505
sa(dp5010
S'a'
I5773
sS's'
I15
sS'm'
(lp5011
I1
aI3
aI1
aI5
aI3
aI33
aI93
aI49
aI39
aI743
aI341
aI2549
aI7603
aI3369
aI30889
asS'd'
I2506
sa(dp5012
S'a'
I5776
sS's'
I15
sS'm'
(lp5013
I1
aI1
aI3
aI13
aI3
aI5
aI87
aI63
aI293
aI785
aI1591
aI675
aI3915
aI2209
aI18201
asS'd'
I2507
sa(dp5014
S'a'
I5815
sS's'
I15
sS'm'
(lp5015
I1
aI3
aI3
aI11
aI3
aI15
aI69
aI231
aI241
aI127
aI429
aI2201
aI8173
aI12549
aI25745
asS'd'
I2508
sa(dp5016
S'a'
I5841
sS's'
I15
sS'm'
(lp5017
I1
aI1
aI5
aI11
aI15
aI39
aI3
aI29
aI125
aI685
aI643
aI1385
aI829
aI7347
aI28793
asS'd'
I2509
sa(dp5018
S'a'
I5847
sS's'
I15
sS'm'
(lp5019
I1
aI1
aI7
aI15
aI27
aI15
aI59
aI237
aI299
aI773
aI1097
aI3875
aI6503
aI7129
aI28495
asS'd'
I2510
sa(dp5020
S'a'
I5860
sS's'
I15
sS'm'
(lp5021
I1
aI3
aI5
aI13
aI9
aI17
aI31
aI227
aI69
aI443
aI1633
aI525
aI1659
aI14681
aI15209
asS'd'
I2511
sa(dp5022
S'a'
I5870
sS's'
I15
sS'm'
(lp5023
I1
aI3
aI5
aI5
aI13
aI51
aI69
aI173
aI111
aI433
aI279
aI2145
aI2091
aI9741
aI24881
asS'd'
I2512
sa(dp5024
S'a'
I5875
sS's'
I15
sS'm'
(lp5025
I1
aI3
aI1
aI7
aI7
aI35
aI55
aI51
aI357
aI99
aI1789
aI333
aI2073
aI10151
aI14527
asS'd'
I2513
sa(dp5026
S'a'
I5877
sS's'
I15
sS'm'
(lp5027
I1
aI3
aI3
aI7
aI13
aI41
aI101
aI87
aI425
aI701
aI1143
aI2733
aI6473
aI8667
aI17419
asS'd'
I2514
sa(dp5028
S'a'
I5884
sS's'
I15
sS'm'
(lp5029
I1
aI1
aI5
aI5
aI25
aI29
aI63
aI31
aI385
aI537
aI563
aI607
aI6723
aI9251
aI6531
asS'd'
I2515
sa(dp5030
S'a'
I5892
sS's'
I15
sS'm'
(lp5031
I1
aI3
aI5
aI5
aI9
aI63
aI111
aI131
aI239
aI723
aI705
aI2805
aI6579
aI12691
aI17521
asS'd'
I2516
sa(dp5032
S'a'
I5902
sS's'
I15
sS'm'
(lp5033
I1
aI3
aI1
aI7
aI31
aI55
aI101
aI225
aI477
aI271
aI611
aI3179
aI7859
aI9835
aI2165
asS'd'
I2517
sa(dp5034
S'a'
I5910
sS's'
I15
sS'm'
(lp5035
I1
aI1
aI3
aI3
aI5
aI15
aI81
aI127
aI391
aI333
aI419
aI1091
aI5997
aI12315
aI31521
asS'd'
I2518
sa(dp5036
S'a'
I5916
sS's'
I15
sS'm'
(lp5037
I1
aI3
aI5
aI15
aI23
aI7
aI35
aI109
aI181
aI295
aI825
aI419
aI969
aI15753
aI9365
asS'd'
I2519
sa(dp5038
S'a'
I5919
sS's'
I15
sS'm'
(lp5039
I1
aI3
aI5
aI5
aI25
aI23
aI69
aI177
aI325
aI359
aI1577
aI619
aI6233
aI11753
aI8103
asS'd'
I2520
sa(dp5040
S'a'
I5935
sS's'
I15
sS'm'
(lp5041
I1
aI3
aI5
aI11
aI31
aI13
aI79
aI61
aI241
aI1011
aI1961
aI949
aI6211
aI497
aI7099
asS'd'
I2521
sa(dp5042
S'a'
I5937
sS's'
I15
sS'm'
(lp5043
I1
aI3
aI5
aI3
aI25
aI19
aI67
aI235
aI337
aI1015
aI1485
aI355
aI3653
aI12735
aI14503
asS'd'
I2522
sa(dp5044
S'a'
I5944
sS's'
I15
sS'm'
(lp5045
I1
aI3
aI5
aI7
aI31
aI23
aI35
aI231
aI147
aI15
aI263
aI1995
aI431
aI5941
aI18931
asS'd'
I2523
sa(dp5046
S'a'
I5947
sS's'
I15
sS'm'
(lp5047
I1
aI3
aI3
aI7
aI1
aI35
aI37
aI7
aI85
aI443
aI715
aI743
aI2189
aI12537
aI17427
asS'd'
I2524
sa(dp5048
S'a'
I5958
sS's'
I15
sS'm'
(lp5049
I1
aI1
aI3
aI1
aI7
aI41
aI1
aI209
aI121
aI929
aI661
aI3999
aI955
aI5123
aI31115
asS'd'
I2525
sa(dp5050
S'a'
I5962
sS's'
I15
sS'm'
(lp5051
I1
aI1
aI3
aI5
aI11
aI43
aI127
aI125
aI107
aI293
aI273
aI2071
aI3003
aI11631
aI7769
asS'd'
I2526
sa(dp5052
S'a'
I5969
sS's'
I15
sS'm'
(lp5053
I1
aI1
aI1
aI13
aI13
aI29
aI39
aI217
aI111
aI779
aI1287
aI1675
aI4201
aI4869
aI20403
asS'd'
I2527
sa(dp5054
S'a'
I5981
sS's'
I15
sS'm'
(lp5055
I1
aI1
aI3
aI15
aI25
aI53
aI25
aI135
aI389
aI925
aI1971
aI663
aI7545
aI2673
aI7725
asS'd'
I2528
sa(dp5056
S'a'
I5995
sS's'
I15
sS'm'
(lp5057
I1
aI1
aI5
aI13
aI3
aI59
aI97
aI91
aI357
aI45
aI947
aI3031
aI8095
aI6269
aI13975
asS'd'
I2529
sa(dp5058
S'a'
I5998
sS's'
I15
sS'm'
(lp5059
I1
aI1
aI5
aI15
aI25
aI31
aI1
aI171
aI375
aI939
aI507
aI3591
aI1089
aI13605
aI2813
asS'd'
I2530
sa(dp5060
S'a'
I6003
sS's'
I15
sS'm'
(lp5061
I1
aI1
aI3
aI7
aI25
aI21
aI41
aI131
aI147
aI737
aI9
aI1603
aI1859
aI11573
aI28397
asS'd'
I2531
sa(dp5062
S'a'
I6010
sS's'
I15
sS'm'
(lp5063
I1
aI3
aI3
aI9
aI21
aI9
aI59
aI27
aI169
aI875
aI711
aI1389
aI2899
aI7937
aI4173
asS'd'
I2532
sa(dp5064
S'a'
I6016
sS's'
I15
sS'm'
(lp5065
I1
aI1
aI5
aI9
aI13
aI29
aI71
aI39
aI51
aI337
aI1067
aI2661
aI1203
aI5967
aI19249
asS'd'
I2533
sa(dp5066
S'a'
I6025
sS's'
I15
sS'm'
(lp5067
I1
aI3
aI7
aI1
aI17
aI21
aI43
aI79
aI181
aI741
aI1901
aI3445
aI7171
aI2109
aI1589
asS'd'
I2534
sa(dp5068
S'a'
I6031
sS's'
I15
sS'm'
(lp5069
I1
aI1
aI3
aI9
aI23
aI37
aI105
aI51
aI227
aI775
aI1265
aI2987
aI2197
aI13903
aI28891
asS'd'
I2535
sa(dp5070
S'a'
I6036
sS's'
I15
sS'm'
(lp5071
I1
aI1
aI1
aI13
aI23
aI47
aI111
aI41
aI93
aI261
aI75
aI2155
aI4301
aI11517
aI16101
asS'd'
I2536
sa(dp5072
S'a'
I6039
sS's'
I15
sS'm'
(lp5073
I1
aI1
aI3
aI3
aI27
aI27
aI123
aI125
aI501
aI775
aI413
aI1065
aI7607
aI15065
aI26013
asS'd'
I2537
sa(dp5074
S'a'
I6045
sS's'
I15
sS'm'
(lp5075
I1
aI3
aI7
aI3
aI27
aI11
aI59
aI87
aI207
aI743
aI1765
aI2969
aI913
aI8101
aI11583
asS'd'
I2538
sa(dp5076
S'a'
I6049
sS's'
I15
sS'm'
(lp5077
I1
aI3
aI3
aI1
aI23
aI7
aI113
aI17
aI285
aI993
aI695
aI2399
aI5019
aI4779
aI28917
asS'd'
I2539
sa(dp5078
S'a'
I6052
sS's'
I15
sS'm'
(lp5079
I1
aI3
aI1
aI5
aI11
aI51
aI49
aI139
aI213
aI435
aI1475
aI2209
aI6695
aI12981
aI9851
asS'd'
I2540
sa(dp5080
S'a'
I6067
sS's'
I15
sS'm'
(lp5081
I1
aI3
aI5
aI7
aI1
aI63
aI31
aI151
aI173
aI767
aI1453
aI1497
aI6911
aI9597
aI25551
asS'd'
I2541
sa(dp5082
S'a'
I6074
sS's'
I15
sS'm'
(lp5083
I1
aI1
aI7
aI7
aI21
aI53
aI39
aI159
aI389
aI231
aI309
aI359
aI7701
aI14819
aI5175
asS'd'
I2542
sa(dp5084
S'a'
I6087
sS's'
I15
sS'm'
(lp5085
I1
aI1
aI1
aI1
aI11
aI47
aI83
aI29
aI247
aI89
aI369
aI2727
aI3103
aI14421
aI17369
asS'd'
I2543
sa(dp5086
S'a'
I6101
sS's'
I15
sS'm'
(lp5087
I1
aI3
aI1
aI5
aI25
aI25
aI111
aI245
aI239
aI755
aI113
aI1765
aI3583
aI917
aI403
asS'd'
I2544
sa(dp5088
S'a'
I6121
sS's'
I15
sS'm'
(lp5089
I1
aI3
aI3
aI3
aI5
aI59
aI85
aI151
aI463
aI591
aI743
aI3767
aI121
aI2927
aI11031
asS'd'
I2545
sa(dp5090
S'a'
I6129
sS's'
I15
sS'm'
(lp5091
I1
aI3
aI5
aI9
aI11
aI39
aI77
aI161
aI275
aI233
aI1991
aI2683
aI6545
aI2423
aI32113
asS'd'
I2546
sa(dp5092
S'a'
I6142
sS's'
I15
sS'm'
(lp5093
I1
aI3
aI5
aI11
aI5
aI57
aI13
aI229
aI329
aI757
aI1863
aI3959
aI4243
aI7265
aI15599
asS'd'
I2547
sa(dp5094
S'a'
I6151
sS's'
I15
sS'm'
(lp5095
I1
aI1
aI1
aI1
aI1
aI23
aI19
aI67
aI453
aI593
aI2011
aI1813
aI4695
aI8903
aI9623
asS'd'
I2548
sa(dp5096
S'a'
I6157
sS's'
I15
sS'm'
(lp5097
I1
aI3
aI3
aI7
aI1
aI29
aI103
aI255
aI493
aI647
aI1709
aI4065
aI4199
aI949
aI28829
asS'd'
I2549
sa(dp5098
S'a'
I6166
sS's'
I15
sS'm'
(lp5099
I1
aI1
aI7
aI9
aI3
aI55
aI53
aI33
aI5
aI223
aI423
aI3347
aI7647
aI7211
aI25157
asS'd'
I2550
sa(dp5100
S'a'
I6170
sS's'
I15
sS'm'
(lp5101
I1
aI3
aI5
aI13
aI3
aI43
aI79
aI255
aI471
aI573
aI1007
aI2119
aI6731
aI10047
aI23179
asS'd'
I2551
sa(dp5102
S'a'
I6175
sS's'
I15
sS'm'
(lp5103
I1
aI1
aI1
aI3
aI7
aI39
aI55
aI61
aI53
aI377
aI435
aI401
aI3307
aI12621
aI14299
asS'd'
I2552
sa(dp5104
S'a'
I6186
sS's'
I15
sS'm'
(lp5105
I1
aI3
aI3
aI7
aI21
aI31
aI67
aI17
aI243
aI425
aI747
aI2995
aI1389
aI2557
aI18415
asS'd'
I2553
sa(dp5106
S'a'
I6203
sS's'
I15
sS'm'
(lp5107
I1
aI3
aI1
aI3
aI3
aI39
aI75
aI11
aI447
aI249
aI1135
aI1011
aI1657
aI10767
aI19501
asS'd'
I2554
sa(dp5108
S'a'
I6217
sS's'
I15
sS'm'
(lp5109
I1
aI3
aI1
aI11
aI17
aI51
aI117
aI129
aI17
aI143
aI785
aI103
aI5049
aI14703
aI28479
asS'd'
I2555
sa(dp5110
S'a'
I6231
sS's'
I15
sS'm'
(lp5111
I1
aI3
aI7
aI5
aI13
aI17
aI75
aI255
aI75
aI661
aI1175
aI477
aI1811
aI1479
aI15783
asS'd'
I2556
sa(dp5112
S'a'
I6241
sS's'
I15
sS'm'
(lp5113
I1
aI3
aI7
aI9
aI11
aI57
aI101
aI77
aI431
aI247
aI997
aI3657
aI5117
aI6815
aI3841
asS'd'
I2557
sa(dp5114
S'a'
I6242
sS's'
I15
sS'm'
(lp5115
I1
aI1
aI5
aI1
aI17
aI21
aI101
aI183
aI209
aI69
aI299
aI1585
aI6381
aI12983
aI10053
asS'd'
I2558
sa(dp5116
S'a'
I6248
sS's'
I15
sS'm'
(lp5117
I1
aI1
aI7
aI3
aI5
aI13
aI21
aI63
aI83
aI857
aI749
aI1251
aI5363
aI9629
aI16563
asS'd'
I2559
sa(dp5118
S'a'
I6256
sS's'
I15
sS'm'
(lp5119
I1
aI3
aI3
aI9
aI3
aI59
aI9
aI45
aI55
aI489
aI137
aI2423
aI2661
aI12111
aI4375
asS'd'
I2560
sa(dp5120
S'a'
I6265
sS's'
I15
sS'm'
(lp5121
I1
aI1
aI5
aI9
aI23
aI9
aI41
aI177
aI447
aI671
aI1631
aI3115
aI4215
aI14435
aI8743
asS'd'
I2561
sa(dp5122
S'a'
I6275
sS's'
I15
sS'm'
(lp5123
I1
aI3
aI7
aI11
aI19
aI23
aI15
aI221
aI413
aI783
aI1247
aI2343
aI4397
aI3145
aI32043
asS'd'
I2562
sa(dp5124
S'a'
I6277
sS's'
I15
sS'm'
(lp5125
I1
aI3
aI3
aI1
aI31
aI55
aI31
aI87
aI333
aI849
aI1777
aI343
aI5199
aI1507
aI11621
asS'd'
I2563
sa(dp5126
S'a'
I6302
sS's'
I15
sS'm'
(lp5127
I1
aI3
aI7
aI3
aI17
aI57
aI63
aI63
aI111
aI977
aI631
aI3019
aI2953
aI14273
aI29209
asS'd'
I2564
sa(dp5128
S'a'
I6315
sS's'
I15
sS'm'
(lp5129
I1
aI3
aI1
aI13
aI9
aI39
aI87
aI15
aI397
aI185
aI701
aI1487
aI3807
aI13727
aI19883
asS'd'
I2565
sa(dp5130
S'a'
I6318
sS's'
I15
sS'm'
(lp5131
I1
aI3
aI7
aI1
aI17
aI57
aI57
aI157
aI119
aI181
aI899
aI353
aI3603
aI15041
aI7421
asS'd'
I2566
sa(dp5132
S'a'
I6330
sS's'
I15
sS'm'
(lp5133
I1
aI1
aI7
aI3
aI29
aI13
aI29
aI191
aI105
aI373
aI961
aI1991
aI5531
aI6793
aI29497
asS'd'
I2567
sa(dp5134
S'a'
I6343
sS's'
I15
sS'm'
(lp5135
I1
aI3
aI3
aI11
aI7
aI61
aI65
aI39
aI215
aI187
aI191
aI1651
aI2481
aI3951
aI24965
asS'd'
I2568
sa(dp5136
S'a'
I6347
sS's'
I15
sS'm'
(lp5137
I1
aI1
aI7
aI5
aI25
aI11
aI105
aI23
aI257
aI771
aI1359
aI2837
aI7821
aI12223
aI28033
asS'd'
I2569
sa(dp5138
S'a'
I6350
sS's'
I15
sS'm'
(lp5139
I1
aI3
aI5
aI11
aI3
aI3
aI23
aI139
aI407
aI885
aI1679
aI2979
aI8149
aI14281
aI12487
asS'd'
I2570
sa(dp5140
S'a'
I6352
sS's'
I15
sS'm'
(lp5141
I1
aI3
aI7
aI3
aI21
aI45
aI13
aI85
aI249
aI1015
aI2023
aI1429
aI965
aI7091
aI31721
asS'd'
I2571
sa(dp5142
S'a'
I6371
sS's'
I15
sS'm'
(lp5143
I1
aI1
aI1
aI13
aI19
aI5
aI119
aI47
aI91
aI285
aI211
aI2607
aI4287
aI9197
aI455
asS'd'
I2572
sa(dp5144
S'a'
I6383
sS's'
I15
sS'm'
(lp5145
I1
aI3
aI1
aI1
aI9
aI59
aI25
aI137
aI121
aI287
aI577
aI3325
aI2365
aI8823
aI5033
asS'd'
I2573
sa(dp5146
S'a'
I6386
sS's'
I15
sS'm'
(lp5147
I1
aI3
aI3
aI13
aI25
aI63
aI99
aI43
aI15
aI855
aI245
aI3189
aI59
aI5181
aI21299
asS'd'
I2574
sa(dp5148
S'a'
I6405
sS's'
I15
sS'm'
(lp5149
I1
aI3
aI5
aI11
aI7
aI9
aI41
aI157
aI359
aI773
aI1347
aI2049
aI4589
aI13731
aI32133
asS'd'
I2575
sa(dp5150
S'a'
I6409
sS's'
I15
sS'm'
(lp5151
I1
aI1
aI7
aI11
aI31
aI37
aI83
aI105
aI183
aI375
aI79
aI1821
aI1989
aI15199
aI22207
asS'd'
I2576
sa(dp5152
S'a'
I6410
sS's'
I15
sS'm'
(lp5153
I1
aI1
aI5
aI3
aI23
aI37
aI127
aI9
aI467
aI651
aI993
aI69
aI6943
aI4093
aI20871
asS'd'
I2577
sa(dp5154
S'a'
I6433
sS's'
I15
sS'm'
(lp5155
I1
aI1
aI3
aI15
aI31
aI49
aI123
aI149
aI211
aI371
aI1825
aI3011
aI485
aI1251
aI17343
asS'd'
I2578
sa(dp5156
S'a'
I6436
sS's'
I15
sS'm'
(lp5157
I1
aI1
aI1
aI15
aI11
aI33
aI127
aI251
aI89
aI317
aI1869
aI219
aI2275
aI14201
aI27063
asS'd'
I2579
sa(dp5158
S'a'
I6439
sS's'
I15
sS'm'
(lp5159
I1
aI1
aI5
aI5
aI19
aI5
aI81
aI35
aI233
aI95
aI9
aI863
aI725
aI11095
aI16217
asS'd'
I2580
sa(dp5160
S'a'
I6463
sS's'
I15
sS'm'
(lp5161
I1
aI1
aI1
aI15
aI23
aI47
aI51
aI43
aI169
aI637
aI865
aI57
aI1509
aI1683
aI7587
asS'd'
I2581
sa(dp5162
S'a'
I6468
sS's'
I15
sS'm'
(lp5163
I1
aI3
aI1
aI3
aI7
aI7
aI117
aI187
aI273
aI303
aI717
aI3091
aI2083
aI3315
aI647
asS'd'
I2582
sa(dp5164
S'a'
I6477
sS's'
I15
sS'm'
(lp5165
I1
aI1
aI5
aI15
aI13
aI27
aI23
aI227
aI145
aI547
aI1783
aI987
aI6895
aI7135
aI11023
asS'd'
I2583
sa(dp5166
S'a'
I6496
sS's'
I15
sS'm'
(lp5167
I1
aI1
aI5
aI11
aI21
aI39
aI57
aI203
aI477
aI17
aI985
aI1729
aI4297
aI7483
aI13263
asS'd'
I2584
sa(dp5168
S'a'
I6511
sS's'
I15
sS'm'
(lp5169
I1
aI3
aI7
aI9
aI3
aI49
aI71
aI45
aI143
aI967
aI39
aI583
aI2123
aI5165
aI17437
asS'd'
I2585
sa(dp5170
S'a'
I6516
sS's'
I15
sS'm'
(lp5171
I1
aI1
aI1
aI9
aI21
aI51
aI71
aI163
aI441
aI709
aI397
aI445
aI6167
aI7753
aI11513
asS'd'
I2586
sa(dp5172
S'a'
I6519
sS's'
I15
sS'm'
(lp5173
I1
aI1
aI7
aI7
aI27
aI35
aI5
aI181
aI449
aI53
aI621
aI3401
aI5263
aI4557
aI9141
asS'd'
I2587
sa(dp5174
S'a'
I6523
sS's'
I15
sS'm'
(lp5175
I1
aI1
aI5
aI7
aI7
aI37
aI83
aI111
aI485
aI881
aI465
aI3371
aI5603
aI371
aI29393
asS'd'
I2588
sa(dp5176
S'a'
I6530
sS's'
I15
sS'm'
(lp5177
I1
aI3
aI1
aI15
aI7
aI47
aI41
aI245
aI377
aI823
aI309
aI3929
aI2159
aI13917
aI13365
asS'd'
I2589
sa(dp5178
S'a'
I6539
sS's'
I15
sS'm'
(lp5179
I1
aI3
aI7
aI7
aI7
aI29
aI25
aI141
aI19
aI611
aI79
aI2689
aI109
aI12321
aI8345
asS'd'
I2590
sa(dp5180
S'a'
I6547
sS's'
I15
sS'm'
(lp5181
I1
aI1
aI1
aI13
aI3
aI53
aI113
aI151
aI381
aI791
aI137
aI3185
aI3567
aI211
aI597
asS'd'
I2591
sa(dp5182
S'a'
I6589
sS's'
I15
sS'm'
(lp5183
I1
aI1
aI3
aI9
aI7
aI53
aI87
aI89
aI491
aI861
aI467
aI3763
aI2025
aI4187
aI9637
asS'd'
I2592
sa(dp5184
S'a'
I6592
sS's'
I15
sS'm'
(lp5185
I1
aI1
aI7
aI1
aI27
aI33
aI71
aI41
aI63
aI1011
aI741
aI1135
aI175
aI3739
aI21493
asS'd'
I2593
sa(dp5186
S'a'
I6601
sS's'
I15
sS'm'
(lp5187
I1
aI3
aI3
aI5
aI9
aI19
aI55
aI175
aI325
aI55
aI1193
aI1423
aI2049
aI9633
aI17515
asS'd'
I2594
sa(dp5188
S'a'
I6610
sS's'
I15
sS'm'
(lp5189
I1
aI1
aI3
aI1
aI27
aI55
aI69
aI103
aI401
aI707
aI825
aI399
aI6799
aI13199
aI6295
asS'd'
I2595
sa(dp5190
S'a'
I6616
sS's'
I15
sS'm'
(lp5191
I1
aI3
aI7
aI3
aI19
aI63
aI25
aI151
aI17
aI159
aI1673
aI615
aI6317
aI13261
aI26267
asS'd'
I2596
sa(dp5192
S'a'
I6619
sS's'
I15
sS'm'
(lp5193
I1
aI3
aI7
aI9
aI27
aI1
aI77
aI129
aI423
aI647
aI707
aI2579
aI3525
aI6723
aI31615
asS'd'
I2597
sa(dp5194
S'a'
I6626
sS's'
I15
sS'm'
(lp5195
I1
aI3
aI3
aI7
aI7
aI31
aI35
aI241
aI309
aI369
aI895
aI3683
aI4795
aI11319
aI451
asS'd'
I2598
sa(dp5196
S'a'
I6635
sS's'
I15
sS'm'
(lp5197
I1
aI3
aI5
aI7
aI17
aI7
aI117
aI141
aI267
aI713
aI569
aI1915
aI4369
aI7793
aI30853
asS'd'
I2599
sa(dp5198
S'a'
I6637
sS's'
I15
sS'm'
(lp5199
I1
aI3
aI7
aI1
aI29
aI61
aI81
aI73
aI413
aI13
aI1977
aI3229
aI5853
aI8451
aI15539
asS'd'
I2600
sa(dp5200
S'a'
I6638
sS's'
I15
sS'm'
(lp5201
I1
aI3
aI7
aI1
aI5
aI45
aI109
aI21
aI431
aI487
aI2019
aI2647
aI927
aI16015
aI10711
asS'd'
I2601
sa(dp5202
S'a'
I6652
sS's'
I15
sS'm'
(lp5203
I1
aI3
aI1
aI3
aI11
aI19
aI37
aI183
aI451
aI377
aI269
aI3993
aI3229
aI4899
aI26561
asS'd'
I2602
sa(dp5204
S'a'
I6656
sS's'
I15
sS'm'
(lp5205
I1
aI3
aI1
aI11
aI5
aI19
aI121
aI55
aI57
aI117
aI687
aI83
aI3047
aI1367
aI17595
asS'd'
I2603
sa(dp5206
S'a'
I6662
sS's'
I15
sS'm'
(lp5207
I1
aI3
aI1
aI7
aI17
aI31
aI41
aI219
aI239
aI963
aI199
aI2895
aI5599
aI7639
aI17201
asS'd'
I2604
sa(dp5208
S'a'
I6689
sS's'
I15
sS'm'
(lp5209
I1
aI3
aI3
aI5
aI27
aI53
aI71
aI183
aI509
aI771
aI1809
aI1539
aI2229
aI4893
aI17115
asS'd'
I2605
sa(dp5210
S'a'
I6699
sS's'
I15
sS'm'
(lp5211
I1
aI1
aI3
aI9
aI9
aI9
aI13
aI49
aI265
aI643
aI1929
aI859
aI497
aI9797
aI27771
asS'd'
I2606
sa(dp5212
S'a'
I6710
sS's'
I15
sS'm'
(lp5213
I1
aI3
aI7
aI11
aI19
aI39
aI115
aI139
aI207
aI903
aI963
aI1849
aI4403
aI6229
aI10021
asS'd'
I2607
sa(dp5214
S'a'
I6714
sS's'
I15
sS'm'
(lp5215
I1
aI3
aI7
aI13
aI3
aI57
aI99
aI223
aI503
aI423
aI1755
aI807
aI1885
aI213
aI18723
asS'd'
I2608
sa(dp5216
S'a'
I6719
sS's'
I15
sS'm'
(lp5217
I1
aI3
aI7
aI15
aI11
aI15
aI111
aI193
aI243
aI599
aI593
aI3385
aI5393
aI15073
aI17777
asS'd'
I2609
sa(dp5218
S'a'
I6739
sS's'
I15
sS'm'
(lp5219
I1
aI1
aI5
aI3
aI19
aI63
aI121
aI207
aI99
aI435
aI1961
aI2747
aI6405
aI3971
aI23481
asS'd'
I2610
sa(dp5220
S'a'
I6751
sS's'
I15
sS'm'
(lp5221
I1
aI3
aI5
aI13
aI9
aI29
aI79
aI131
aI415
aI49
aI229
aI1003
aI3263
aI12975
aI15987
asS'd'
I2611
sa(dp5222
S'a'
I6775
sS's'
I15
sS'm'
(lp5223
I1
aI1
aI3
aI7
aI1
aI41
aI127
aI155
aI29
aI73
aI963
aI659
aI2741
aI3465
aI2595
asS'd'
I2612
sa(dp5224
S'a'
I6779
sS's'
I15
sS'm'
(lp5225
I1
aI1
aI3
aI5
aI23
aI23
aI93
aI233
aI113
aI521
aI427
aI1557
aI6917
aI12953
aI22441
asS'd'
I2613
sa(dp5226
S'a'
I6788
sS's'
I15
sS'm'
(lp5227
I1
aI1
aI5
aI13
aI5
aI25
aI85
aI191
aI387
aI69
aI955
aI243
aI4473
aI9813
aI21711
asS'd'
I2614
sa(dp5228
S'a'
I6798
sS's'
I15
sS'm'
(lp5229
I1
aI3
aI3
aI7
aI1
aI53
aI95
aI65
aI231
aI995
aI539
aI2103
aI5513
aI14087
aI28655
asS'd'
I2615
sa(dp5230
S'a'
I6815
sS's'
I15
sS'm'
(lp5231
I1
aI3
aI5
aI3
aI17
aI13
aI19
aI227
aI197
aI91
aI1437
aI1121
aI3307
aI6903
aI3297
asS'd'
I2616
sa(dp5232
S'a'
I6819
sS's'
I15
sS'm'
(lp5233
I1
aI1
aI5
aI11
aI31
aI29
aI109
aI171
aI257
aI783
aI861
aI9
aI4895
aI1859
aI10909
asS'd'
I2617
sa(dp5234
S'a'
I6825
sS's'
I15
sS'm'
(lp5235
I1
aI1
aI7
aI13
aI5
aI47
aI61
aI5
aI363
aI351
aI1525
aI823
aI2883
aI12435
aI17629
asS'd'
I2618
sa(dp5236
S'a'
I6826
sS's'
I15
sS'm'
(lp5237
I1
aI1
aI5
aI11
aI9
aI3
aI69
aI159
aI371
aI477
aI1223
aI1973
aI2757
aI413
aI31223
asS'd'
I2619
sa(dp5238
S'a'
I6836
sS's'
I15
sS'm'
(lp5239
I1
aI1
aI3
aI5
aI23
aI45
aI43
aI195
aI423
aI829
aI1673
aI1563
aI6633
aI14775
aI21097
asS'd'
I2620
sa(dp5240
S'a'
I6843
sS's'
I15
sS'm'
(lp5241
I1
aI1
aI3
aI3
aI13
aI9
aI107
aI209
aI49
aI609
aI1047
aI3691
aI7483
aI4269
aI7557
asS'd'
I2621
sa(dp5242
S'a'
I6845
sS's'
I15
sS'm'
(lp5243
I1
aI1
aI3
aI15
aI3
aI43
aI73
aI161
aI53
aI813
aI325
aI3439
aI7009
aI8691
aI11711
asS'd'
I2622
sa(dp5244
S'a'
I6858
sS's'
I15
sS'm'
(lp5245
I1
aI1
aI3
aI3
aI23
aI45
aI99
aI61
aI407
aI15
aI1515
aI1557
aI953
aI8567
aI13729
asS'd'
I2623
sa(dp5246
S'a'
I6868
sS's'
I15
sS'm'
(lp5247
I1
aI1
aI5
aI9
aI31
aI35
aI117
aI57
aI227
aI923
aI1373
aI1811
aI3405
aI11979
aI10149
asS'd'
I2624
sa(dp5248
S'a'
I6877
sS's'
I15
sS'm'
(lp5249
I1
aI1
aI3
aI9
aI15
aI53
aI105
aI209
aI153
aI67
aI1477
aI667
aI3077
aI4911
aI3871
asS'd'
I2625
sa(dp5250
S'a'
I6881
sS's'
I15
sS'm'
(lp5251
I1
aI1
aI3
aI3
aI21
aI53
aI93
aI101
aI183
aI1023
aI3
aI3041
aI5815
aI9043
aI5801
asS'd'
I2626
sa(dp5252
S'a'
I6891
sS's'
I15
sS'm'
(lp5253
I1
aI3
aI3
aI5
aI17
aI49
aI127
aI161
aI321
aI869
aI1369
aI923
aI3235
aI711
aI30007
asS'd'
I2627
sa(dp5254
S'a'
I6896
sS's'
I15
sS'm'
(lp5255
I1
aI1
aI3
aI3
aI15
aI17
aI97
aI229
aI389
aI159
aI1075
aI2001
aI7905
aI15191
aI14693
asS'd'
I2628
sa(dp5256
S'a'
I6899
sS's'
I15
sS'm'
(lp5257
I1
aI1
aI5
aI11
aI5
aI5
aI121
aI173
aI95
aI173
aI1883
aI3915
aI1439
aI9981
aI24375
asS'd'
I2629
sa(dp5258
S'a'
I6901
sS's'
I15
sS'm'
(lp5259
I1
aI3
aI3
aI1
aI31
aI53
aI29
aI189
aI37
aI623
aI217
aI949
aI3959
aI7189
aI25427
asS'd'
I2630
sa(dp5260
S'a'
I6908
sS's'
I15
sS'm'
(lp5261
I1
aI3
aI5
aI9
aI21
aI45
aI101
aI23
aI355
aI729
aI797
aI2317
aI2931
aI7433
aI29175
asS'd'
I2631
sa(dp5262
S'a'
I6914
sS's'
I15
sS'm'
(lp5263
I1
aI3
aI7
aI1
aI1
aI63
aI63
aI155
aI237
aI865
aI1169
aI43
aI7335
aI6445
aI7979
asS'd'
I2632
sa(dp5264
S'a'
I6916
sS's'
I15
sS'm'
(lp5265
I1
aI3
aI7
aI7
aI11
aI51
aI37
aI199
aI503
aI991
aI319
aI3013
aI7885
aI12837
aI32419
asS'd'
I2633
sa(dp5266
S'a'
I6923
sS's'
I15
sS'm'
(lp5267
I1
aI3
aI7
aI7
aI27
aI31
aI101
aI243
aI37
aI811
aI1909
aI109
aI6455
aI7903
aI11821
asS'd'
I2634
sa(dp5268
S'a'
I6925
sS's'
I15
sS'm'
(lp5269
I1
aI1
aI3
aI13
aI23
aI21
aI89
aI99
aI243
aI605
aI1017
aI1871
aI1101
aI12825
aI8227
asS'd'
I2635
sa(dp5270
S'a'
I6928
sS's'
I15
sS'm'
(lp5271
I1
aI3
aI3
aI13
aI19
aI3
aI51
aI59
aI501
aI605
aI385
aI2189
aI3229
aI7981
aI31407
asS'd'
I2636
sa(dp5272
S'a'
I6931
sS's'
I15
sS'm'
(lp5273
I1
aI1
aI1
aI1
aI25
aI11
aI127
aI215
aI295
aI237
aI1245
aI3657
aI7803
aI3897
aI655
asS'd'
I2637
sa(dp5274
S'a'
I6934
sS's'
I15
sS'm'
(lp5275
I1
aI1
aI7
aI7
aI5
aI9
aI63
aI129
aI143
aI417
aI795
aI3409
aI2847
aI5887
aI3093
asS'd'
I2638
sa(dp5276
S'a'
I6937
sS's'
I15
sS'm'
(lp5277
I1
aI3
aI3
aI13
aI7
aI57
aI67
aI57
aI5
aI847
aI1185
aI3349
aI4841
aI11457
aI8857
asS'd'
I2639
sa(dp5278
S'a'
I6938
sS's'
I15
sS'm'
(lp5279
I1
aI1
aI3
aI3
aI9
aI53
aI51
aI43
aI85
aI437
aI13
aI2543
aI3651
aI15493
aI767
asS'd'
I2640
sa(dp5280
S'a'
I6949
sS's'
I15
sS'm'
(lp5281
I1
aI1
aI7
aI9
aI1
aI49
aI97
aI115
aI133
aI1011
aI1399
aI2653
aI7765
aI13999
aI12097
asS'd'
I2641
sa(dp5282
S'a'
I6956
sS's'
I15
sS'm'
(lp5283
I1
aI1
aI5
aI1
aI3
aI27
aI123
aI107
aI389
aI401
aI1759
aI1333
aI1371
aI5277
aI14865
asS'd'
I2642
sa(dp5284
S'a'
I6973
sS's'
I15
sS'm'
(lp5285
I1
aI1
aI5
aI1
aI13
aI23
aI3
aI123
aI137
aI821
aI399
aI1671
aI3095
aI3121
aI31387
asS'd'
I2643
sa(dp5286
S'a'
I6976
sS's'
I15
sS'm'
(lp5287
I1
aI1
aI5
aI3
aI7
aI35
aI57
aI237
aI509
aI753
aI1783
aI2815
aI6495
aI13283
aI7091
asS'd'
I2644
sa(dp5288
S'a'
I6981
sS's'
I15
sS'm'
(lp5289
I1
aI1
aI7
aI11
aI5
aI37
aI77
aI109
aI7
aI969
aI1087
aI3705
aI1695
aI14223
aI28959
asS'd'
I2645
sa(dp5290
S'a'
I6988
sS's'
I15
sS'm'
(lp5291
I1
aI3
aI1
aI11
aI25
aI5
aI25
aI163
aI179
aI185
aI671
aI1031
aI4537
aI11601
aI9323
asS'd'
I2646
sa(dp5292
S'a'
I6999
sS's'
I15
sS'm'
(lp5293
I1
aI1
aI3
aI7
aI17
aI25
aI49
aI221
aI183
aI619
aI1953
aI343
aI4523
aI14883
aI6833
asS'd'
I2647
sa(dp5294
S'a'
I7016
sS's'
I15
sS'm'
(lp5295
I1
aI3
aI7
aI5
aI27
aI19
aI59
aI153
aI11
aI807
aI513
aI3019
aI6875
aI5307
aI8405
asS'd'
I2648
sa(dp5296
S'a'
I7027
sS's'
I15
sS'm'
(lp5297
I1
aI1
aI1
aI13
aI25
aI41
aI21
aI109
aI321
aI135
aI497
aI1235
aI5177
aI5167
aI18609
asS'd'
I2649
sa(dp5298
S'a'
I7029
sS's'
I15
sS'm'
(lp5299
I1
aI1
aI7
aI5
aI21
aI53
aI25
aI197
aI411
aI503
aI1009
aI1921
aI4305
aI2633
aI31415
asS'd'
I2650
sa(dp5300
S'a'
I7055
sS's'
I15
sS'm'
(lp5301
I1
aI3
aI5
aI1
aI25
aI45
aI27
aI227
aI271
aI903
aI639
aI3805
aI657
aI8683
aI29585
asS'd'
I2651
sa(dp5302
S'a'
I7058
sS's'
I15
sS'm'
(lp5303
I1
aI1
aI5
aI3
aI9
aI49
aI37
aI35
aI351
aI491
aI851
aI2983
aI31
aI5619
aI6919
asS'd'
I2652
sa(dp5304
S'a'
I7074
sS's'
I15
sS'm'
(lp5305
I1
aI1
aI5
aI3
aI11
aI49
aI33
aI153
aI393
aI1017
aI1561
aI2795
aI4435
aI12589
aI22349
asS'd'
I2653
sa(dp5306
S'a'
I7083
sS's'
I15
sS'm'
(lp5307
I1
aI1
aI1
aI15
aI17
aI29
aI49
aI245
aI217
aI359
aI1133
aI393
aI3317
aI415
aI16407
asS'd'
I2654
sa(dp5308
S'a'
I7093
sS's'
I15
sS'm'
(lp5309
I1
aI1
aI3
aI5
aI3
aI9
aI95
aI63
aI319
aI319
aI1009
aI19
aI6453
aI16279
aI6975
asS'd'
I2655
sa(dp5310
S'a'
I7100
sS's'
I15
sS'm'
(lp5311
I1
aI1
aI5
aI9
aI3
aI25
aI67
aI95
aI369
aI237
aI285
aI2409
aI671
aI5143
aI121
asS'd'
I2656
sa(dp5312
S'a'
I7105
sS's'
I15
sS'm'
(lp5313
I1
aI1
aI3
aI1
aI9
aI49
aI35
aI87
aI317
aI185
aI445
aI2263
aI7923
aI10183
aI26615
asS'd'
I2657
sa(dp5314
S'a'
I7112
sS's'
I15
sS'm'
(lp5315
I1
aI3
aI3
aI11
aI9
aI59
aI29
aI135
aI129
aI477
aI353
aI3571
aI1057
aI16329
aI23523
asS'd'
I2658
sa(dp5316
S'a'
I7118
sS's'
I15
sS'm'
(lp5317
I1
aI1
aI1
aI15
aI13
aI11
aI19
aI5
aI133
aI827
aI1799
aI1893
aI1939
aI1101
aI12147
asS'd'
I2659
sa(dp5318
S'a'
I7120
sS's'
I15
sS'm'
(lp5319
I1
aI1
aI3
aI3
aI15
aI49
aI33
aI185
aI511
aI1013
aI41
aI3499
aI6235
aI7643
aI16725
asS'd'
I2660
sa(dp5320
S'a'
I7129
sS's'
I15
sS'm'
(lp5321
I1
aI1
aI5
aI11
aI27
aI45
aI89
aI157
aI63
aI137
aI2047
aI1275
aI4995
aI625
aI6111
asS'd'
I2661
sa(dp5322
S'a'
I7166
sS's'
I15
sS'm'
(lp5323
I1
aI3
aI7
aI11
aI3
aI1
aI121
aI1
aI341
aI33
aI1895
aI3033
aI3929
aI10257
aI21037
asS'd'
I2662
sa(dp5324
S'a'
I7207
sS's'
I15
sS'm'
(lp5325
I1
aI3
aI3
aI11
aI7
aI11
aI117
aI5
aI115
aI287
aI335
aI3415
aI5397
aI15065
aI19121
asS'd'
I2663
sa(dp5326
S'a'
I7216
sS's'
I15
sS'm'
(lp5327
I1
aI3
aI3
aI13
aI21
aI25
aI15
aI125
aI277
aI125
aI801
aI3761
aI2623
aI11333
aI16867
asS'd'
I2664
sa(dp5328
S'a'
I7226
sS's'
I15
sS'm'
(lp5329
I1
aI3
aI5
aI11
aI19
aI33
aI21
aI71
aI499
aI747
aI1515
aI185
aI1759
aI14623
aI895
asS'd'
I2665
sa(dp5330
S'a'
I7234
sS's'
I15
sS'm'
(lp5331
I1
aI3
aI7
aI1
aI29
aI35
aI9
aI203
aI277
aI299
aI1509
aI2017
aI2897
aI14175
aI1643
asS'd'
I2666
sa(dp5332
S'a'
I7236
sS's'
I15
sS'm'
(lp5333
I1
aI3
aI5
aI11
aI7
aI47
aI111
aI197
aI459
aI941
aI1619
aI2119
aI2191
aI11049
aI6811
asS'd'
I2667
sa(dp5334
S'a'
I7246
sS's'
I15
sS'm'
(lp5335
I1
aI1
aI5
aI9
aI7
aI43
aI103
aI115
aI87
aI269
aI1235
aI77
aI5887
aI1611
aI29041
asS'd'
I2668
sa(dp5336
S'a'
I7248
sS's'
I15
sS'm'
(lp5337
I1
aI1
aI5
aI7
aI1
aI61
aI83
aI225
aI179
aI81
aI1145
aI2403
aI1485
aI8967
aI20607
asS'd'
I2669
sa(dp5338
S'a'
I7254
sS's'
I15
sS'm'
(lp5339
I1
aI3
aI3
aI1
aI25
aI47
aI27
aI157
aI359
aI803
aI1683
aI1995
aI6445
aI13113
aI17899
asS'd'
I2670
sa(dp5340
S'a'
I7263
sS's'
I15
sS'm'
(lp5341
I1
aI3
aI1
aI7
aI21
aI37
aI43
aI119
aI245
aI49
aI1581
aI2275
aI3311
aI4087
aI29765
asS'd'
I2671
sa(dp5342
S'a'
I7273
sS's'
I15
sS'm'
(lp5343
I1
aI1
aI3
aI13
aI5
aI33
aI49
aI191
aI455
aI105
aI665
aI3855
aI3207
aI2671
aI32203
asS'd'
I2672
sa(dp5344
S'a'
I7274
sS's'
I15
sS'm'
(lp5345
I1
aI3
aI1
aI1
aI25
aI63
aI19
aI217
aI17
aI353
aI947
aI1951
aI4097
aI9041
aI11921
asS'd'
I2673
sa(dp5346
S'a'
I7293
sS's'
I15
sS'm'
(lp5347
I1
aI3
aI1
aI7
aI21
aI31
aI113
aI97
aI347
aI993
aI1799
aI3831
aI3711
aI6193
aI1235
asS'd'
I2674
sa(dp5348
S'a'
I7297
sS's'
I15
sS'm'
(lp5349
I1
aI1
aI1
aI5
aI3
aI63
aI11
aI203
aI425
aI445
aI1361
aI531
aI1265
aI1755
aI11685
asS'd'
I2675
sa(dp5350
S'a'
I7310
sS's'
I15
sS'm'
(lp5351
I1
aI3
aI1
aI7
aI13
aI29
aI23
aI85
aI57
aI467
aI1835
aI133
aI7961
aI4175
aI2445
asS'd'
I2676
sa(dp5352
S'a'
I7315
sS's'
I15
sS'm'
(lp5353
I1
aI1
aI1
aI15
aI23
aI27
aI37
aI5
aI123
aI913
aI1293
aI1633
aI3113
aI5413
aI26407
asS'd'
I2677
sa(dp5354
S'a'
I7317
sS's'
I15
sS'm'
(lp5355
I1
aI1
aI5
aI13
aI27
aI1
aI121
aI151
aI303
aI931
aI375
aI3679
aI1863
aI12301
aI30907
asS'd'
I2678
sa(dp5356
S'a'
I7331
sS's'
I15
sS'm'
(lp5357
I1
aI3
aI1
aI9
aI31
aI9
aI49
aI203
aI177
aI937
aI1503
aI933
aI5867
aI12533
aI13621
asS'd'
I2679
sa(dp5358
S'a'
I7338
sS's'
I15
sS'm'
(lp5359
I1
aI3
aI3
aI15
aI1
aI41
aI23
aI191
aI191
aI931
aI837
aI3553
aI2611
aI4735
aI18105
asS'd'
I2680
sa(dp5360
S'a'
I7340
sS's'
I15
sS'm'
(lp5361
I1
aI1
aI5
aI7
aI27
aI49
aI51
aI111
aI435
aI195
aI1229
aI711
aI7145
aI14571
aI31707
asS'd'
I2681
sa(dp5362
S'a'
I7346
sS's'
I15
sS'm'
(lp5363
I1
aI1
aI7
aI7
aI3
aI41
aI59
aI203
aI291
aI903
aI1727
aI2757
aI1463
aI6287
aI31535
asS'd'
I2682
sa(dp5364
S'a'
I7355
sS's'
I15
sS'm'
(lp5365
I1
aI1
aI7
aI13
aI23
aI5
aI75
aI3
aI207
aI525
aI411
aI2133
aI2231
aI477
aI7155
asS'd'
I2683
sa(dp5366
S'a'
I7366
sS's'
I15
sS'm'
(lp5367
I1
aI3
aI5
aI7
aI13
aI19
aI111
aI225
aI489
aI83
aI1177
aI4077
aI4617
aI14413
aI7133
asS'd'
I2684
sa(dp5368
S'a'
I7383
sS's'
I15
sS'm'
(lp5369
I1
aI3
aI1
aI7
aI9
aI59
aI3
aI113
aI379
aI803
aI1289
aI3347
aI4127
aI6669
aI14867
asS'd'
I2685
sa(dp5370
S'a'
I7389
sS's'
I15
sS'm'
(lp5371
I1
aI3
aI7
aI3
aI31
aI37
aI87
aI79
aI399
aI749
aI995
aI1611
aI3137
aI12543
aI31955
asS'd'
I2686
sa(dp5372
S'a'
I7393
sS's'
I15
sS'm'
(lp5373
I1
aI1
aI5
aI7
aI21
aI59
aI49
aI45
aI511
aI639
aI1033
aI2169
aI3265
aI15001
aI10745
asS'd'
I2687
sa(dp5374
S'a'
I7396
sS's'
I15
sS'm'
(lp5375
I1
aI1
aI5
aI1
aI25
aI19
aI23
aI203
aI11
aI883
aI1031
aI4087
aI5059
aI11321
aI21675
asS'd'
I2688
sa(dp5376
S'a'
I7400
sS's'
I15
sS'm'
(lp5377
I1
aI3
aI7
aI5
aI11
aI27
aI33
aI205
aI163
aI289
aI501
aI3505
aI1515
aI1895
aI15889
asS'd'
I2689
sa(dp5378
S'a'
I7414
sS's'
I15
sS'm'
(lp5379
I1
aI3
aI1
aI1
aI23
aI7
aI39
aI239
aI29
aI119
aI1499
aI2071
aI6495
aI12107
aI5339
asS'd'
I2690
sa(dp5380
S'a'
I7417
sS's'
I15
sS'm'
(lp5381
I1
aI3
aI1
aI1
aI23
aI29
aI55
aI181
aI327
aI905
aI427
aI1033
aI427
aI3687
aI5367
asS'd'
I2691
sa(dp5382
S'a'
I7426
sS's'
I15
sS'm'
(lp5383
I1
aI3
aI3
aI7
aI21
aI27
aI115
aI127
aI393
aI855
aI1291
aI2121
aI381
aI9995
aI29757
asS'd'
I2692
sa(dp5384
S'a'
I7432
sS's'
I15
sS'm'
(lp5385
I1
aI3
aI5
aI1
aI25
aI13
aI15
aI183
aI269
aI1005
aI1531
aI3451
aI3975
aI9479
aI23695
asS'd'
I2693
sa(dp5386
S'a'
I7452
sS's'
I15
sS'm'
(lp5387
I1
aI3
aI7
aI7
aI19
aI31
aI111
aI97
aI33
aI179
aI1343
aI2069
aI977
aI5043
aI9129
asS'd'
I2694
sa(dp5388
S'a'
I7468
sS's'
I15
sS'm'
(lp5389
I1
aI3
aI1
aI5
aI17
aI57
aI99
aI129
aI379
aI829
aI837
aI1845
aI3613
aI7351
aI19291
asS'd'
I2695
sa(dp5390
S'a'
I7488
sS's'
I15
sS'm'
(lp5391
I1
aI3
aI3
aI5
aI31
aI23
aI119
aI229
aI135
aI389
aI9
aI705
aI6697
aI15441
aI5303
asS'd'
I2696
sa(dp5392
S'a'
I7491
sS's'
I15
sS'm'
(lp5393
I1
aI1
aI1
aI11
aI25
aI31
aI105
aI95
aI5
aI931
aI789
aI375
aI7543
aI9957
aI28627
asS'd'
I2697
sa(dp5394
S'a'
I7494
sS's'
I15
sS'm'
(lp5395
I1
aI1
aI7
aI15
aI21
aI17
aI19
aI103
aI389
aI545
aI1725
aI2867
aI4251
aI3829
aI6907
asS'd'
I2698
sa(dp5396
S'a'
I7497
sS's'
I15
sS'm'
(lp5397
I1
aI3
aI7
aI7
aI15
aI37
aI97
aI65
aI337
aI409
aI1649
aI2869
aI7929
aI8905
aI21989
asS'd'
I2699
sa(dp5398
S'a'
I7515
sS's'
I15
sS'm'
(lp5399
I1
aI3
aI5
aI3
aI11
aI15
aI69
aI29
aI353
aI207
aI233
aI411
aI2047
aI10303
aI31655
asS'd'
I2700
sa(dp5400
S'a'
I7531
sS's'
I15
sS'm'
(lp5401
I1
aI3
aI3
aI7
aI27
aI43
aI125
aI107
aI69
aI981
aI215
aI1955
aI3589
aI597
aI12703
asS'd'
I2701
sa(dp5402
S'a'
I7552
sS's'
I15
sS'm'
(lp5403
I1
aI1
aI7
aI9
aI25
aI13
aI109
aI73
aI227
aI663
aI1115
aI285
aI471
aI3359
aI15787
asS'd'
I2702
sa(dp5404
S'a'
I7562
sS's'
I15
sS'm'
(lp5405
I1
aI3
aI7
aI5
aI1
aI45
aI7
aI79
aI441
aI149
aI701
aI1457
aI6595
aI14829
aI20865
asS'd'
I2703
sa(dp5406
S'a'
I7564
sS's'
I15
sS'm'
(lp5407
I1
aI3
aI7
aI15
aI15
aI47
aI83
aI239
aI295
aI23
aI1085
aI813
aI1209
aI3573
aI2855
asS'd'
I2704
sa(dp5408
S'a'
I7569
sS's'
I15
sS'm'
(lp5409
I1
aI1
aI3
aI15
aI13
aI7
aI59
aI67
aI255
aI537
aI1841
aI3857
aI6821
aI15175
aI13997
asS'd'
I2705
sa(dp5410
S'a'
I7582
sS's'
I15
sS'm'
(lp5411
I1
aI3
aI1
aI1
aI9
aI57
aI59
aI21
aI21
aI41
aI1693
aI2805
aI7953
aI1389
aI14105
asS'd'
I2706
sa(dp5412
S'a'
I7585
sS's'
I15
sS'm'
(lp5413
I1
aI3
aI5
aI15
aI19
aI49
aI107
aI117
aI99
aI607
aI145
aI53
aI1863
aI9383
aI12029
asS'd'
I2707
sa(dp5414
S'a'
I7588
sS's'
I15
sS'm'
(lp5415
I1
aI3
aI3
aI13
aI1
aI39
aI5
aI141
aI503
aI265
aI281
aI1785
aI2673
aI6597
aI6333
asS'd'
I2708
sa(dp5416
S'a'
I7592
sS's'
I15
sS'm'
(lp5417
I1
aI1
aI5
aI3
aI3
aI19
aI3
aI181
aI169
aI269
aI955
aI2399
aI3157
aI11053
aI8563
asS'd'
I2709
sa(dp5418
S'a'
I7597
sS's'
I15
sS'm'
(lp5419
I1
aI3
aI3
aI13
aI11
aI1
aI95
aI43
aI179
aI507
aI443
aI209
aI3239
aI14239
aI21829
asS'd'
I2710
sa(dp5420
S'a'
I7603
sS's'
I15
sS'm'
(lp5421
I1
aI1
aI7
aI9
aI3
aI17
aI99
aI179
aI445
aI479
aI1897
aI1507
aI5753
aI4757
aI2135
asS'd'
I2711
sa(dp5422
S'a'
I7610
sS's'
I15
sS'm'
(lp5423
I1
aI3
aI3
aI1
aI9
aI51
aI29
aI13
aI295
aI291
aI927
aI85
aI5707
aI7447
aI32319
asS'd'
I2712
sa(dp5424
S'a'
I7624
sS's'
I15
sS'm'
(lp5425
I1
aI1
aI1
aI3
aI13
aI11
aI21
aI157
aI213
aI327
aI1071
aI591
aI2639
aI15405
aI6617
asS'd'
I2713
sa(dp5426
S'a'
I7642
sS's'
I15
sS'm'
(lp5427
I1
aI3
aI5
aI1
aI7
aI25
aI55
aI47
aI495
aI681
aI727
aI2707
aI2955
aI705
aI7489
asS'd'
I2714
sa(dp5428
S'a'
I7647
sS's'
I15
sS'm'
(lp5429
I1
aI1
aI3
aI9
aI17
aI3
aI73
aI67
aI465
aI367
aI1473
aI3195
aI7825
aI5299
aI1817
asS'd'
I2715
sa(dp5430
S'a'
I7653
sS's'
I15
sS'm'
(lp5431
I1
aI1
aI1
aI1
aI19
aI31
aI77
aI253
aI71
aI599
aI1601
aI871
aI2243
aI6699
aI13013
asS'd'
I2716
sa(dp5432
S'a'
I7654
sS's'
I15
sS'm'
(lp5433
I1
aI1
aI7
aI9
aI21
aI1
aI71
aI115
aI5
aI65
aI767
aI925
aI7901
aI10761
aI19431
asS'd'
I2717
sa(dp5434
S'a'
I7666
sS's'
I15
sS'm'
(lp5435
I1
aI3
aI1
aI7
aI23
aI31
aI31
aI15
aI105
aI391
aI585
aI2995
aI2635
aI10607
aI24951
asS'd'
I2718
sa(dp5436
S'a'
I7668
sS's'
I15
sS'm'
(lp5437
I1
aI3
aI3
aI1
aI19
aI25
aI71
aI211
aI41
aI197
aI787
aI225
aI6781
aI813
aI10117
asS'd'
I2719
sa(dp5438
S'a'
I7684
sS's'
I15
sS'm'
(lp5439
I1
aI3
aI3
aI3
aI17
aI29
aI3
aI153
aI231
aI643
aI1151
aI447
aI3699
aI9625
aI26677
asS'd'
I2720
sa(dp5440
S'a'
I7705
sS's'
I15
sS'm'
(lp5441
I1
aI1
aI5
aI9
aI1
aI25
aI71
aI21
aI395
aI297
aI557
aI3841
aI233
aI1877
aI4569
asS'd'
I2721
sa(dp5442
S'a'
I7732
sS's'
I15
sS'm'
(lp5443
I1
aI1
aI3
aI13
aI1
aI45
aI115
aI61
aI5
aI937
aI173
aI2109
aI2927
aI9599
aI9155
asS'd'
I2722
sa(dp5444
S'a'
I7741
sS's'
I15
sS'm'
(lp5445
I1
aI1
aI3
aI3
aI15
aI21
aI61
aI121
aI253
aI285
aI1083
aI3545
aI5537
aI6773
aI2629
asS'd'
I2723
sa(dp5446
S'a'
I7749
sS's'
I15
sS'm'
(lp5447
I1
aI3
aI3
aI15
aI13
aI63
aI33
aI77
aI49
aI849
aI1795
aI2771
aI5481
aI9833
aI603
asS'd'
I2724
sa(dp5448
S'a'
I7750
sS's'
I15
sS'm'
(lp5449
I1
aI1
aI7
aI5
aI1
aI39
aI113
aI237
aI225
aI1005
aI1687
aI2297
aI3213
aI2605
aI14669
asS'd'
I2725
sa(dp5450
S'a'
I7759
sS's'
I15
sS'm'
(lp5451
I1
aI1
aI3
aI1
aI11
aI1
aI39
aI23
aI67
aI441
aI1235
aI2545
aI3139
aI15901
aI29243
asS'd'
I2726
sa(dp5452
S'a'
I7764
sS's'
I15
sS'm'
(lp5453
I1
aI3
aI1
aI3
aI15
aI49
aI39
aI57
aI311
aI345
aI525
aI223
aI4923
aI6311
aI25275
asS'd'
I2727
sa(dp5454
S'a'
I7777
sS's'
I15
sS'm'
(lp5455
I1
aI1
aI5
aI7
aI9
aI13
aI69
aI11
aI349
aI423
aI1773
aI1055
aI1001
aI9359
aI17025
asS'd'
I2728
sa(dp5456
S'a'
I7790
sS's'
I15
sS'm'
(lp5457
I1
aI1
aI1
aI13
aI15
aI63
aI89
aI207
aI335
aI591
aI1223
aI2701
aI55
aI12471
aI13127
asS'd'
I2729
sa(dp5458
S'a'
I7817
sS's'
I15
sS'm'
(lp5459
I1
aI1
aI3
aI5
aI15
aI19
aI83
aI67
aI407
aI113
aI1961
aI779
aI5803
aI12417
aI21751
asS'd'
I2730
sa(dp5460
S'a'
I7826
sS's'
I15
sS'm'
(lp5461
I1
aI3
aI3
aI1
aI21
aI53
aI81
aI95
aI405
aI427
aI1047
aI2443
aI4153
aI5843
aI22511
asS'd'
I2731
sa(dp5462
S'a'
I7831
sS's'
I15
sS'm'
(lp5463
I1
aI1
aI7
aI7
aI7
aI25
aI115
aI155
aI453
aI537
aI741
aI2379
aI2343
aI16035
aI19587
asS'd'
I2732
sa(dp5464
S'a'
I7859
sS's'
I15
sS'm'
(lp5465
I1
aI3
aI3
aI11
aI27
aI21
aI111
aI121
aI503
aI437
aI803
aI3399
aI5303
aI10163
aI18199
asS'd'
I2733
sa(dp5466
S'a'
I7871
sS's'
I15
sS'm'
(lp5467
I1
aI1
aI5
aI13
aI19
aI27
aI7
aI81
aI259
aI545
aI965
aI743
aI4533
aI8813
aI21253
asS'd'
I2734
sa(dp5468
S'a'
I7873
sS's'
I15
sS'm'
(lp5469
I1
aI1
aI5
aI5
aI1
aI59
aI37
aI11
aI105
aI343
aI75
aI1319
aI6317
aI9593
aI1699
asS'd'
I2735
sa(dp5470
S'a'
I7876
sS's'
I15
sS'm'
(lp5471
I1
aI3
aI1
aI9
aI13
aI9
aI115
aI131
aI387
aI1023
aI253
aI693
aI5191
aI12777
aI10565
asS'd'
I2736
sa(dp5472
S'a'
I7900
sS's'
I15
sS'm'
(lp5473
I1
aI3
aI1
aI15
aI7
aI35
aI111
aI195
aI287
aI305
aI533
aI1901
aI3363
aI10085
aI30791
asS'd'
I2737
sa(dp5474
S'a'
I7904
sS's'
I15
sS'm'
(lp5475
I1
aI1
aI3
aI9
aI27
aI51
aI21
aI77
aI413
aI925
aI717
aI791
aI4147
aI585
aI5649
asS'd'
I2738
sa(dp5476
S'a'
I7913
sS's'
I15
sS'm'
(lp5477
I1
aI3
aI3
aI5
aI25
aI59
aI79
aI249
aI185
aI567
aI71
aI1997
aI7373
aI2327
aI18637
asS'd'
I2739
sa(dp5478
S'a'
I7916
sS's'
I15
sS'm'
(lp5479
I1
aI3
aI3
aI11
aI15
aI21
aI97
aI99
aI391
aI57
aI1973
aI29
aI7451
aI2529
aI25737
asS'd'
I2740
sa(dp5480
S'a'
I7922
sS's'
I15
sS'm'
(lp5481
I1
aI3
aI7
aI5
aI7
aI59
aI93
aI5
aI287
aI469
aI1639
aI3637
aI5465
aI14431
aI32265
asS'd'
I2741
sa(dp5482
S'a'
I7946
sS's'
I15
sS'm'
(lp5483
I1
aI1
aI3
aI11
aI3
aI1
aI71
aI75
aI427
aI299
aI811
aI3697
aI3529
aI5433
aI26957
asS'd'
I2742
sa(dp5484
S'a'
I7953
sS's'
I15
sS'm'
(lp5485
I1
aI3
aI1
aI9
aI19
aI59
aI37
aI255
aI165
aI1005
aI19
aI2851
aI4309
aI455
aI9485
asS'd'
I2743
sa(dp5486
S'a'
I7956
sS's'
I15
sS'm'
(lp5487
I1
aI1
aI1
aI5
aI1
aI55
aI15
aI233
aI133
aI47
aI1831
aI713
aI2601
aI1017
aI3201
asS'd'
I2744
sa(dp5488
S'a'
I7963
sS's'
I15
sS'm'
(lp5489
I1
aI1
aI5
aI5
aI21
aI55
aI127
aI69
aI377
aI41
aI25
aI2295
aI7595
aI4733
aI11615
asS'd'
I2745
sa(dp5490
S'a'
I7979
sS's'
I15
sS'm'
(lp5491
I1
aI1
aI5
aI3
aI23
aI5
aI7
aI181
aI161
aI775
aI1095
aI2271
aI6637
aI14489
aI6873
asS'd'
I2746
sa(dp5492
S'a'
I7981
sS's'
I15
sS'm'
(lp5493
I1
aI3
aI5
aI9
aI9
aI15
aI5
aI133
aI357
aI21
aI127
aI2685
aI6299
aI4363
aI17573
asS'd'
I2747
sa(dp5494
S'a'
I7984
sS's'
I15
sS'm'
(lp5495
I1
aI3
aI3
aI9
aI13
aI39
aI51
aI223
aI201
aI401
aI1839
aI2461
aI7633
aI6039
aI10445
asS'd'
I2748
sa(dp5496
S'a'
I7989
sS's'
I15
sS'm'
(lp5497
I1
aI1
aI5
aI1
aI9
aI21
aI19
aI249
aI227
aI359
aI255
aI2895
aI4117
aI2073
aI27687
asS'd'
I2749
sa(dp5498
S'a'
I7999
sS's'
I15
sS'm'
(lp5499
I1
aI1
aI5
aI15
aI5
aI61
aI113
aI161
aI95
aI3
aI877
aI2775
aI293
aI6655
aI4023
asS'd'
I2750
sa(dp5500
S'a'
I8001
sS's'
I15
sS'm'
(lp5501
I1
aI3
aI7
aI1
aI7
aI55
aI73
aI39
aI295
aI403
aI985
aI2315
aI1667
aI13525
aI1453
asS'd'
I2751
sa(dp5502
S'a'
I8021
sS's'
I15
sS'm'
(lp5503
I1
aI1
aI5
aI1
aI27
aI1
aI85
aI195
aI11
aI713
aI1841
aI3895
aI3131
aI2193
aI17607
asS'd'
I2752
sa(dp5504
S'a'
I8056
sS's'
I15
sS'm'
(lp5505
I1
aI3
aI5
aI13
aI25
aI1
aI119
aI97
aI239
aI167
aI1393
aI1753
aI6989
aI12155
aI12509
asS'd'
I2753
sa(dp5506
S'a'
I8080
sS's'
I15
sS'm'
(lp5507
I1
aI1
aI7
aI15
aI31
aI21
aI41
aI255
aI425
aI445
aI165
aI2097
aI5627
aI4971
aI13207
asS'd'
I2754
sa(dp5508
S'a'
I8083
sS's'
I15
sS'm'
(lp5509
I1
aI1
aI1
aI15
aI13
aI33
aI81
aI105
aI453
aI197
aI13
aI1547
aI7381
aI8709
aI15103
asS'd'
I2755
sa(dp5510
S'a'
I8089
sS's'
I15
sS'm'
(lp5511
I1
aI1
aI3
aI11
aI11
aI33
aI107
aI123
aI483
aI367
aI121
aI995
aI1911
aI8205
aI22577
asS'd'
I2756
sa(dp5512
S'a'
I8090
sS's'
I15
sS'm'
(lp5513
I1
aI1
aI1
aI9
aI9
aI43
aI71
aI49
aI273
aI431
aI1705
aI3313
aI4259
aI16291
aI14345
asS'd'
I2757
sa(dp5514
S'a'
I8114
sS's'
I15
sS'm'
(lp5515
I1
aI1
aI1
aI7
aI3
aI1
aI43
aI213
aI97
aI547
aI1559
aI1149
aI2791
aI3751
aI887
asS'd'
I2758
sa(dp5516
S'a'
I8128
sS's'
I15
sS'm'
(lp5517
I1
aI1
aI3
aI15
aI25
aI47
aI49
aI251
aI425
aI35
aI295
aI3767
aI6305
aI9633
aI5045
asS'd'
I2759
sa(dp5518
S'a'
I8133
sS's'
I15
sS'm'
(lp5519
I1
aI3
aI3
aI1
aI5
aI55
aI91
aI245
aI27
aI981
aI331
aI555
aI6553
aI11017
aI15289
asS'd'
I2760
sa(dp5520
S'a'
I8145
sS's'
I15
sS'm'
(lp5521
I1
aI1
aI3
aI7
aI1
aI23
aI23
aI155
aI223
aI565
aI1005
aI3211
aI3847
aI7479
aI3643
asS'd'
I2761
sa(dp5522
S'a'
I8155
sS's'
I15
sS'm'
(lp5523
I1
aI1
aI5
aI1
aI17
aI7
aI47
aI95
aI35
aI779
aI1685
aI2099
aI7505
aI15425
aI18089
asS'd'
I2762
sa(dp5524
S'a'
I8161
sS's'
I15
sS'm'
(lp5525
I1
aI3
aI3
aI7
aI3
aI63
aI83
aI151
aI211
aI147
aI611
aI1171
aI1681
aI7687
aI13423
asS'd'
I2763
sa(dp5526
S'a'
I8182
sS's'
I15
sS'm'
(lp5527
I1
aI3
aI3
aI1
aI3
aI27
aI107
aI117
aI497
aI537
aI195
aI3075
aI2753
aI1665
aI19399
asS'd'
I2764
sa(dp5528
S'a'
I8186
sS's'
I15
sS'm'
(lp5529
I1
aI1
aI1
aI7
aI23
aI5
aI103
aI209
aI117
aI845
aI1243
aI1283
aI4253
aI9723
aI20937
asS'd'
I2765
sa(dp5530
S'a'
I8191
sS's'
I15
sS'm'
(lp5531
I1
aI3
aI1
aI1
aI5
aI49
aI7
aI13
aI419
aI125
aI287
aI1599
aI8161
aI1275
aI24661
asS'd'
I2766
sa(dp5532
S'a'
I8192
sS's'
I15
sS'm'
(lp5533
I1
aI3
aI3
aI3
aI13
aI63
aI23
aI183
aI39
aI979
aI1301
aI2349
aI905
aI15805
aI30151
asS'd'
I2767
sa(dp5534
S'a'
I8195
sS's'
I15
sS'm'
(lp5535
I1
aI1
aI3
aI9
aI17
aI11
aI97
aI189
aI189
aI511
aI1779
aI2077
aI6891
aI11623
aI23949
asS'd'
I2768
sa(dp5536
S'a'
I8201
sS's'
I15
sS'm'
(lp5537
I1
aI1
aI7
aI11
aI13
aI45
aI15
aI37
aI11
aI853
aI915
aI1569
aI6103
aI10633
aI3137
asS'd'
I2769
sa(dp5538
S'a'
I8207
sS's'
I15
sS'm'
(lp5539
I1
aI3
aI3
aI5
aI15
aI61
aI91
aI255
aI131
aI821
aI1755
aI1501
aI2663
aI1747
aI941
asS'd'
I2770
sa(dp5540
S'a'
I8210
sS's'
I15
sS'm'
(lp5541
I1
aI1
aI3
aI7
aI19
aI19
aI65
aI95
aI499
aI239
aI2023
aI3185
aI4649
aI3861
aI3767
asS'd'
I2771
sa(dp5542
S'a'
I8228
sS's'
I15
sS'm'
(lp5543
I1
aI3
aI5
aI15
aI15
aI63
aI55
aI93
aI127
aI303
aI171
aI1763
aI4991
aI9479
aI9917
asS'd'
I2772
sa(dp5544
S'a'
I8249
sS's'
I15
sS'm'
(lp5545
I1
aI3
aI7
aI5
aI31
aI53
aI111
aI35
aI433
aI163
aI1903
aI3991
aI3585
aI643
aI21941
asS'd'
I2773
sa(dp5546
S'a'
I8252
sS's'
I15
sS'm'
(lp5547
I1
aI3
aI1
aI9
aI27
aI39
aI67
aI89
aI487
aI349
aI587
aI1723
aI4311
aI11321
aI25785
asS'd'
I2774
sa(dp5548
S'a'
I8258
sS's'
I15
sS'm'
(lp5549
I1
aI3
aI5
aI7
aI1
aI63
aI23
aI237
aI507
aI689
aI1341
aI441
aI1721
aI843
aI20335
asS'd'
I2775
sa(dp5550
S'a'
I8267
sS's'
I15
sS'm'
(lp5551
I1
aI1
aI3
aI3
aI31
aI63
aI83
aI103
aI25
aI799
aI1379
aI1817
aI3809
aI12285
aI16673
asS'd'
I2776
sa(dp5552
S'a'
I8270
sS's'
I15
sS'm'
(lp5553
I1
aI1
aI5
aI3
aI25
aI29
aI99
aI193
aI21
aI549
aI33
aI3109
aI4135
aI10071
aI32355
asS'd'
I2777
sa(dp5554
S'a'
I8275
sS's'
I15
sS'm'
(lp5555
I1
aI3
aI1
aI7
aI13
aI27
aI83
aI189
aI121
aI167
aI379
aI1503
aI7955
aI13189
aI313
asS'd'
I2778
sa(dp5556
S'a'
I8284
sS's'
I15
sS'm'
(lp5557
I1
aI3
aI5
aI15
aI25
aI19
aI83
aI87
aI257
aI237
aI709
aI1169
aI1561
aI7117
aI4785
asS'd'
I2779
sa(dp5558
S'a'
I8293
sS's'
I15
sS'm'
(lp5559
I1
aI1
aI1
aI7
aI9
aI55
aI21
aI5
aI439
aI367
aI403
aI2311
aI6243
aI8349
aI13127
asS'd'
I2780
sa(dp5560
S'a'
I8298
sS's'
I15
sS'm'
(lp5561
I1
aI3
aI7
aI3
aI5
aI35
aI51
aI67
aI453
aI767
aI29
aI3293
aI6665
aI11459
aI2799
asS'd'
I2781
sa(dp5562
S'a'
I8305
sS's'
I15
sS'm'
(lp5563
I1
aI3
aI3
aI3
aI5
aI19
aI59
aI7
aI367
aI683
aI783
aI1317
aI7119
aI6129
aI19525
asS'd'
I2782
sa(dp5564
S'a'
I8317
sS's'
I15
sS'm'
(lp5565
I1
aI1
aI5
aI5
aI5
aI19
aI61
aI67
aI381
aI291
aI875
aI2179
aI2481
aI9325
aI11253
asS'd'
I2783
sa(dp5566
S'a'
I8328
sS's'
I15
sS'm'
(lp5567
I1
aI3
aI5
aI5
aI7
aI47
aI107
aI9
aI141
aI667
aI1989
aI821
aI3909
aI1733
aI10187
asS'd'
I2784
sa(dp5568
S'a'
I8336
sS's'
I15
sS'm'
(lp5569
I1
aI1
aI7
aI7
aI31
aI61
aI1
aI71
aI477
aI689
aI1539
aI3617
aI8105
aI6535
aI3293
asS'd'
I2785
sa(dp5570
S'a'
I8345
sS's'
I15
sS'm'
(lp5571
I1
aI1
aI5
aI5
aI23
aI9
aI103
aI197
aI241
aI249
aI297
aI3607
aI6217
aI1673
aI30103
asS'd'
I2786
sa(dp5572
S'a'
I8351
sS's'
I15
sS'm'
(lp5573
I1
aI3
aI1
aI5
aI23
aI15
aI115
aI105
aI365
aI51
aI825
aI2687
aI359
aI16325
aI15083
asS'd'
I2787
sa(dp5574
S'a'
I8367
sS's'
I15
sS'm'
(lp5575
I1
aI1
aI3
aI11
aI29
aI45
aI65
aI251
aI169
aI189
aI1243
aI2345
aI1345
aI14471
aI25631
asS'd'
I2788
sa(dp5576
S'a'
I8379
sS's'
I15
sS'm'
(lp5577
I1
aI1
aI5
aI9
aI7
aI63
aI81
aI167
aI309
aI539
aI1169
aI3949
aI4193
aI12047
aI1491
asS'd'
I2789
sa(dp5578
S'a'
I8381
sS's'
I15
sS'm'
(lp5579
I1
aI3
aI1
aI9
aI29
aI33
aI89
aI167
aI67
aI73
aI1885
aI477
aI5745
aI13365
aI6819
asS'd'
I2790
sa(dp5580
S'a'
I8382
sS's'
I15
sS'm'
(lp5581
I1
aI3
aI7
aI9
aI9
aI49
aI95
aI13
aI157
aI997
aI1725
aI935
aI7543
aI6349
aI18277
asS'd'
I2791
sa(dp5582
S'a'
I8393
sS's'
I15
sS'm'
(lp5583
I1
aI1
aI5
aI5
aI11
aI59
aI97
aI17
aI303
aI469
aI93
aI2761
aI7395
aI9021
aI24299
asS'd'
I2792
sa(dp5584
S'a'
I8402
sS's'
I15
sS'm'
(lp5585
I1
aI1
aI7
aI3
aI27
aI63
aI71
aI99
aI407
aI139
aI711
aI2589
aI4715
aI5405
aI3277
asS'd'
I2793
sa(dp5586
S'a'
I8414
sS's'
I15
sS'm'
(lp5587
I1
aI3
aI7
aI3
aI11
aI15
aI49
aI57
aI271
aI493
aI1165
aI2839
aI8191
aI2609
aI14759
asS'd'
I2794
sa(dp5588
S'a'
I8417
sS's'
I15
sS'm'
(lp5589
I1
aI1
aI1
aI7
aI21
aI15
aI71
aI245
aI413
aI473
aI1321
aI1165
aI1027
aI6983
aI12867
asS'd'
I2795
sa(dp5590
S'a'
I8420
sS's'
I15
sS'm'
(lp5591
I1
aI1
aI5
aI3
aI15
aI21
aI19
aI197
aI401
aI627
aI2047
aI2761
aI5807
aI5751
aI28025
asS'd'
I2796
sa(dp5592
S'a'
I8429
sS's'
I15
sS'm'
(lp5593
I1
aI1
aI3
aI3
aI5
aI57
aI19
aI209
aI341
aI165
aI489
aI455
aI231
aI14385
aI12457
asS'd'
I2797
sa(dp5594
S'a'
I8435
sS's'
I15
sS'm'
(lp5595
I1
aI3
aI3
aI11
aI13
aI63
aI79
aI129
aI17
aI315
aI1881
aI1069
aI177
aI12013
aI29567
asS'd'
I2798
sa(dp5596
S'a'
I8438
sS's'
I15
sS'm'
(lp5597
I1
aI1
aI3
aI7
aI31
aI29
aI51
aI235
aI475
aI375
aI617
aI437
aI6379
aI8505
aI23079
asS'd'
I2799
sa(dp5598
S'a'
I8450
sS's'
I15
sS'm'
(lp5599
I1
aI1
aI3
aI7
aI27
aI3
aI3
aI137
aI203
aI959
aI363
aI371
aI2899
aI13491
aI22979
asS'd'
I2800
sa(dp5600
S'a'
I8452
sS's'
I15
sS'm'
(lp5601
I1
aI3
aI3
aI3
aI9
aI1
aI57
aI7
aI363
aI537
aI713
aI2417
aI509
aI7747
aI22135
asS'd'
I2801
sa(dp5602
S'a'
I8459
sS's'
I15
sS'm'
(lp5603
I1
aI3
aI3
aI3
aI13
aI21
aI79
aI121
aI487
aI921
aI113
aI281
aI2853
aI14855
aI19747
asS'd'
I2802
sa(dp5604
S'a'
I8470
sS's'
I15
sS'm'
(lp5605
I1
aI1
aI1
aI11
aI3
aI53
aI89
aI123
aI307
aI585
aI567
aI1925
aI505
aI15935
aI20419
asS'd'
I2803
sa(dp5606
S'a'
I8486
sS's'
I15
sS'm'
(lp5607
I1
aI1
aI3
aI3
aI15
aI45
aI77
aI197
aI499
aI683
aI1405
aI3573
aI981
aI14135
aI19763
asS'd'
I2804
sa(dp5608
S'a'
I8490
sS's'
I15
sS'm'
(lp5609
I1
aI1
aI1
aI11
aI27
aI31
aI61
aI191
aI29
aI601
aI373
aI2011
aI6193
aI3599
aI4387
asS'd'
I2805
sa(dp5610
S'a'
I8500
sS's'
I15
sS'm'
(lp5611
I1
aI3
aI5
aI9
aI7
aI13
aI1
aI193
aI469
aI603
aI1315
aI3329
aI3761
aI8355
aI10425
asS'd'
I2806
sa(dp5612
S'a'
I8524
sS's'
I15
sS'm'
(lp5613
I1
aI1
aI3
aI9
aI29
aI61
aI103
aI17
aI117
aI251
aI2029
aI2963
aI3763
aI16117
aI6627
asS'd'
I2807
sa(dp5614
S'a'
I8536
sS's'
I15
sS'm'
(lp5615
I1
aI3
aI1
aI3
aI7
aI51
aI91
aI145
aI497
aI657
aI871
aI3707
aI5905
aI10449
aI14901
asS'd'
I2808
sa(dp5616
S'a'
I8552
sS's'
I15
sS'm'
(lp5617
I1
aI1
aI3
aI1
aI3
aI53
aI23
aI149
aI461
aI333
aI1809
aI1315
aI1815
aI8223
aI13297
asS'd'
I2809
sa(dp5618
S'a'
I8558
sS's'
I15
sS'm'
(lp5619
I1
aI1
aI1
aI7
aI15
aI31
aI3
aI47
aI443
aI829
aI1305
aI893
aI4191
aI9681
aI32661
asS'd'
I2810
sa(dp5620
S'a'
I8570
sS's'
I15
sS'm'
(lp5621
I1
aI3
aI1
aI3
aI27
aI43
aI51
aI221
aI295
aI825
aI649
aI2953
aI6203
aI8237
aI20253
asS'd'
I2811
sa(dp5622
S'a'
I8576
sS's'
I15
sS'm'
(lp5623
I1
aI3
aI1
aI3
aI9
aI35
aI41
aI195
aI249
aI225
aI387
aI3789
aI1499
aI2559
aI28413
asS'd'
I2812
sa(dp5624
S'a'
I8582
sS's'
I15
sS'm'
(lp5625
I1
aI1
aI5
aI15
aI19
aI29
aI13
aI115
aI333
aI787
aI787
aI723
aI2987
aI6227
aI10865
asS'd'
I2813
sa(dp5626
S'a'
I8594
sS's'
I15
sS'm'
(lp5627
I1
aI3
aI5
aI13
aI5
aI59
aI5
aI251
aI79
aI387
aI11
aI3167
aI6619
aI13317
aI18979
asS'd'
I2814
sa(dp5628
S'a'
I8606
sS's'
I15
sS'm'
(lp5629
I1
aI1
aI7
aI11
aI31
aI51
aI43
aI1
aI189
aI519
aI1945
aI2129
aI4365
aI14059
aI3139
asS'd'
I2815
sa(dp5630
S'a'
I8619
sS's'
I15
sS'm'
(lp5631
I1
aI1
aI7
aI5
aI31
aI9
aI43
aI19
aI151
aI533
aI1061
aI3849
aI6871
aI6941
aI14935
asS'd'
I2816
sa(dp5632
S'a'
I8621
sS's'
I15
sS'm'
(lp5633
I1
aI3
aI7
aI5
aI19
aI57
aI7
aI129
aI25
aI353
aI17
aI1739
aI6513
aI399
aI28835
asS'd'
I2817
sa(dp5634
S'a'
I8624
sS's'
I15
sS'm'
(lp5635
I1
aI3
aI5
aI15
aI25
aI15
aI37
aI125
aI39
aI239
aI271
aI65
aI2189
aI10449
aI11815
asS'd'
I2818
sa(dp5636
S'a'
I8633
sS's'
I15
sS'm'
(lp5637
I1
aI3
aI7
aI15
aI19
aI57
aI47
aI245
aI509
aI945
aI385
aI3987
aI3585
aI14711
aI9655
asS'd'
I2819
sa(dp5638
S'a'
I8641
sS's'
I15
sS'm'
(lp5639
I1
aI1
aI3
aI13
aI21
aI31
aI13
aI81
aI9
aI489
aI1321
aI63
aI1363
aI2219
aI19541
asS'd'
I2820
sa(dp5640
S'a'
I8653
sS's'
I15
sS'm'
(lp5641
I1
aI1
aI5
aI7
aI3
aI57
aI25
aI147
aI23
aI553
aI889
aI307
aI6429
aI15807
aI12861
asS'd'
I2821
sa(dp5642
S'a'
I8654
sS's'
I15
sS'm'
(lp5643
I1
aI1
aI3
aI15
aI29
aI21
aI99
aI237
aI151
aI881
aI675
aI3625
aI1159
aI11759
aI21347
asS'd'
I2822
sa(dp5644
S'a'
I8662
sS's'
I15
sS'm'
(lp5645
I1
aI1
aI7
aI1
aI9
aI13
aI111
aI239
aI235
aI609
aI1569
aI3271
aI2837
aI13807
aI7301
asS'd'
I2823
sa(dp5646
S'a'
I8675
sS's'
I15
sS'm'
(lp5647
I1
aI3
aI1
aI15
aI7
aI59
aI27
aI81
aI129
aI9
aI647
aI3595
aI1877
aI1067
aI1859
asS'd'
I2824
sa(dp5648
S'a'
I8689
sS's'
I15
sS'm'
(lp5649
I1
aI3
aI7
aI1
aI3
aI25
aI119
aI57
aI145
aI441
aI1045
aI789
aI215
aI1265
aI9369
asS'd'
I2825
sa(dp5650
S'a'
I8695
sS's'
I15
sS'm'
(lp5651
I1
aI3
aI7
aI3
aI17
aI25
aI87
aI211
aI441
aI229
aI223
aI2795
aI7241
aI7007
aI20575
asS'd'
I2826
sa(dp5652
S'a'
I8702
sS's'
I15
sS'm'
(lp5653
I1
aI1
aI3
aI1
aI13
aI1
aI55
aI227
aI389
aI141
aI1097
aI2487
aI7603
aI4161
aI5025
asS'd'
I2827
sa(dp5654
S'a'
I8706
sS's'
I15
sS'm'
(lp5655
I1
aI1
aI3
aI5
aI15
aI29
aI29
aI145
aI233
aI209
aI891
aI89
aI8097
aI2897
aI26685
asS'd'
I2828
sa(dp5656
S'a'
I8720
sS's'
I15
sS'm'
(lp5657
I1
aI1
aI3
aI1
aI29
aI53
aI19
aI95
aI161
aI359
aI435
aI3313
aI4955
aI7965
aI21015
asS'd'
I2829
sa(dp5658
S'a'
I8729
sS's'
I15
sS'm'
(lp5659
I1
aI3
aI5
aI9
aI19
aI3
aI109
aI77
aI29
aI937
aI1663
aI125
aI2453
aI1069
aI20639
asS'd'
I2830
sa(dp5660
S'a'
I8739
sS's'
I15
sS'm'
(lp5661
I1
aI3
aI7
aI13
aI5
aI23
aI43
aI231
aI347
aI591
aI1963
aI2491
aI4045
aI16029
aI8149
asS'd'
I2831
sa(dp5662
S'a'
I8741
sS's'
I15
sS'm'
(lp5663
I1
aI1
aI5
aI1
aI13
aI3
aI75
aI211
aI419
aI929
aI901
aI3453
aI8121
aI799
aI8897
asS'd'
I2832
sa(dp5664
S'a'
I8751
sS's'
I15
sS'm'
(lp5665
I1
aI1
aI7
aI15
aI11
aI11
aI123
aI111
aI309
aI415
aI1071
aI975
aI2009
aI12945
aI19617
asS'd'
I2833
sa(dp5666
S'a'
I8759
sS's'
I15
sS'm'
(lp5667
I1
aI1
aI1
aI7
aI31
aI35
aI81
aI255
aI89
aI643
aI451
aI513
aI497
aI11751
aI24215
asS'd'
I2834
sa(dp5668
S'a'
I8766
sS's'
I15
sS'm'
(lp5669
I1
aI3
aI5
aI5
aI25
aI17
aI5
aI165
aI139
aI929
aI1927
aI1353
aI7427
aI9719
aI17087
asS'd'
I2835
sa(dp5670
S'a'
I8777
sS's'
I15
sS'm'
(lp5671
I1
aI3
aI5
aI1
aI21
aI55
aI79
aI85
aI333
aI847
aI1305
aI851
aI5057
aI8361
aI18269
asS'd'
I2836
sa(dp5672
S'a'
I8783
sS's'
I15
sS'm'
(lp5673
I1
aI3
aI7
aI15
aI27
aI17
aI55
aI125
aI395
aI223
aI271
aI781
aI1639
aI10569
aI11143
asS'd'
I2837
sa(dp5674
S'a'
I8786
sS's'
I15
sS'm'
(lp5675
I1
aI1
aI7
aI9
aI7
aI33
aI127
aI85
aI209
aI339
aI483
aI241
aI2523
aI14951
aI6855
asS'd'
I2838
sa(dp5676
S'a'
I8795
sS's'
I15
sS'm'
(lp5677
I1
aI1
aI3
aI9
aI5
aI19
aI9
aI183
aI435
aI343
aI1105
aI3139
aI7617
aI1311
aI267
asS'd'
I2839
sa(dp5678
S'a'
I8802
sS's'
I15
sS'm'
(lp5679
I1
aI1
aI5
aI1
aI15
aI53
aI11
aI63
aI113
aI241
aI855
aI3123
aI4777
aI3495
aI23345
asS'd'
I2840
sa(dp5680
S'a'
I8814
sS's'
I15
sS'm'
(lp5681
I1
aI3
aI1
aI5
aI19
aI29
aI119
aI205
aI167
aI683
aI289
aI1629
aI4977
aI8981
aI6867
asS'd'
I2841
sa(dp5682
S'a'
I8821
sS's'
I15
sS'm'
(lp5683
I1
aI3
aI1
aI1
aI31
aI63
aI95
aI159
aI267
aI231
aI863
aI3385
aI5315
aI7267
aI13757
asS'd'
I2842
sa(dp5684
S'a'
I8828
sS's'
I15
sS'm'
(lp5685
I1
aI3
aI5
aI11
aI19
aI21
aI53
aI41
aI125
aI179
aI533
aI1279
aI3759
aI7073
aI13905
asS'd'
I2843
sa(dp5686
S'a'
I8831
sS's'
I15
sS'm'
(lp5687
I1
aI3
aI5
aI9
aI17
aI7
aI27
aI67
aI97
aI809
aI1423
aI2743
aI2859
aI16121
aI329
asS'd'
I2844
sa(dp5688
S'a'
I8837
sS's'
I15
sS'm'
(lp5689
I1
aI3
aI1
aI15
aI1
aI41
aI59
aI155
aI509
aI51
aI1827
aI3739
aI3879
aI13369
aI30821
asS'd'
I2845
sa(dp5690
S'a'
I8842
sS's'
I15
sS'm'
(lp5691
I1
aI3
aI3
aI7
aI21
aI31
aI7
aI13
aI347
aI919
aI1225
aI497
aI5051
aI3769
aI20211
asS'd'
I2846
sa(dp5692
S'a'
I8855
sS's'
I15
sS'm'
(lp5693
I1
aI3
aI7
aI13
aI31
aI9
aI127
aI195
aI123
aI387
aI3
aI3593
aI6623
aI9827
aI29319
asS'd'
I2847
sa(dp5694
S'a'
I8856
sS's'
I15
sS'm'
(lp5695
I1
aI1
aI3
aI9
aI7
aI27
aI95
aI211
aI287
aI189
aI1683
aI1999
aI7641
aI14983
aI4699
asS'd'
I2848
sa(dp5696
S'a'
I8868
sS's'
I15
sS'm'
(lp5697
I1
aI1
aI5
aI3
aI7
aI21
aI29
aI189
aI101
aI423
aI885
aI3275
aI6569
aI11023
aI22265
asS'd'
I2849
sa(dp5698
S'a'
I8877
sS's'
I15
sS'm'
(lp5699
I1
aI3
aI5
aI3
aI9
aI33
aI79
aI75
aI327
aI975
aI287
aI3025
aI2157
aI7301
aI24447
asS'd'
I2850
sa(dp5700
S'a'
I8890
sS's'
I15
sS'm'
(lp5701
I1
aI3
aI3
aI15
aI31
aI27
aI63
aI1
aI71
aI119
aI1151
aI517
aI6131
aI11055
aI179
asS'd'
I2851
sa(dp5702
S'a'
I8892
sS's'
I15
sS'm'
(lp5703
I1
aI3
aI7
aI11
aI23
aI15
aI101
aI247
aI349
aI735
aI673
aI997
aI6451
aI229
aI32103
asS'd'
I2852
sa(dp5704
S'a'
I8900
sS's'
I15
sS'm'
(lp5705
I1
aI3
aI5
aI15
aI7
aI1
aI51
aI135
aI207
aI741
aI1831
aI1235
aI4747
aI11915
aI22009
asS'd'
I2853
sa(dp5706
S'a'
I8909
sS's'
I15
sS'm'
(lp5707
I1
aI3
aI1
aI13
aI9
aI31
aI19
aI221
aI465
aI681
aI627
aI2595
aI5617
aI14201
aI30355
asS'd'
I2854
sa(dp5708
S'a'
I8912
sS's'
I15
sS'm'
(lp5709
I1
aI1
aI3
aI1
aI13
aI49
aI55
aI155
aI11
aI885
aI1275
aI3591
aI2217
aI6659
aI30885
asS'd'
I2855
sa(dp5710
S'a'
I8921
sS's'
I15
sS'm'
(lp5711
I1
aI1
aI7
aI11
aI27
aI57
aI93
aI95
aI243
aI63
aI1405
aI2049
aI7689
aI15943
aI18503
asS'd'
I2856
sa(dp5712
S'a'
I8922
sS's'
I15
sS'm'
(lp5713
I1
aI1
aI7
aI7
aI5
aI11
aI47
aI189
aI467
aI631
aI1665
aI2717
aI4285
aI2087
aI1435
asS'd'
I2857
sa(dp5714
S'a'
I8927
sS's'
I15
sS'm'
(lp5715
I1
aI1
aI3
aI11
aI7
aI27
aI127
aI3
aI231
aI757
aI435
aI2545
aI3537
aI9127
aI19915
asS'd'
I2858
sa(dp5716
S'a'
I8943
sS's'
I15
sS'm'
(lp5717
I1
aI1
aI5
aI13
aI5
aI29
aI85
aI127
aI339
aI875
aI497
aI1573
aI6553
aI11983
aI18029
asS'd'
I2859
sa(dp5718
S'a'
I8948
sS's'
I15
sS'm'
(lp5719
I1
aI3
aI1
aI1
aI21
aI3
aI15
aI91
aI231
aI683
aI1529
aI2651
aI4147
aI13437
aI23861
asS'd'
I2860
sa(dp5720
S'a'
I8951
sS's'
I15
sS'm'
(lp5721
I1
aI3
aI1
aI7
aI27
aI17
aI19
aI179
aI243
aI223
aI1037
aI1501
aI5935
aI2259
aI25185
asS'd'
I2861
sa(dp5722
S'a'
I8958
sS's'
I15
sS'm'
(lp5723
I1
aI1
aI3
aI15
aI11
aI19
aI127
aI27
aI483
aI219
aI583
aI2555
aI531
aI3451
aI17875
asS'd'
I2862
sa(dp5724
S'a'
I8984
sS's'
I15
sS'm'
(lp5725
I1
aI1
aI1
aI13
aI31
aI39
aI89
aI149
aI363
aI741
aI1355
aI4067
aI3171
aI6783
aI1799
asS'd'
I2863
sa(dp5726
S'a'
I8994
sS's'
I15
sS'm'
(lp5727
I1
aI1
aI3
aI11
aI25
aI51
aI45
aI235
aI379
aI123
aI1701
aI725
aI1991
aI7471
aI9833
asS'd'
I2864
sa(dp5728
S'a'
I9000
sS's'
I15
sS'm'
(lp5729
I1
aI1
aI5
aI13
aI15
aI47
aI13
aI201
aI263
aI57
aI375
aI2963
aI7475
aI15929
aI13775
asS'd'
I2865
sa(dp5730
S'a'
I9013
sS's'
I15
sS'm'
(lp5731
I1
aI1
aI3
aI1
aI29
aI29
aI11
aI161
aI345
aI253
aI97
aI255
aI7267
aI2379
aI3933
asS'd'
I2866
sa(dp5732
S'a'
I9018
sS's'
I15
sS'm'
(lp5733
I1
aI3
aI1
aI15
aI3
aI47
aI11
aI69
aI347
aI747
aI795
aI2401
aI3367
aI2383
aI6125
asS'd'
I2867
sa(dp5734
S'a'
I9020
sS's'
I15
sS'm'
(lp5735
I1
aI1
aI7
aI3
aI1
aI49
aI101
aI47
aI71
aI761
aI1503
aI2619
aI191
aI8895
aI873
asS'd'
I2868
sa(dp5736
S'a'
I9031
sS's'
I15
sS'm'
(lp5737
I1
aI3
aI3
aI5
aI25
aI41
aI93
aI85
aI427
aI109
aI1675
aI2409
aI4317
aI9233
aI30283
asS'd'
I2869
sa(dp5738
S'a'
I9035
sS's'
I15
sS'm'
(lp5739
I1
aI1
aI3
aI9
aI11
aI3
aI67
aI159
aI425
aI751
aI887
aI1415
aI403
aI15977
aI10739
asS'd'
I2870
sa(dp5740
S'a'
I9045
sS's'
I15
sS'm'
(lp5741
I1
aI1
aI5
aI13
aI9
aI1
aI9
aI103
aI481
aI601
aI931
aI1957
aI5763
aI7095
aI27141
asS'd'
I2871
sa(dp5742
S'a'
I9052
sS's'
I15
sS'm'
(lp5743
I1
aI1
aI3
aI15
aI29
aI13
aI43
aI33
aI297
aI269
aI1041
aI1411
aI3461
aI12043
aI10045
asS'd'
I2872
sa(dp5744
S'a'
I9056
sS's'
I15
sS'm'
(lp5745
I1
aI3
aI5
aI3
aI3
aI3
aI5
aI7
aI185
aI753
aI133
aI1561
aI5595
aI13777
aI25795
asS'd'
I2873
sa(dp5746
S'a'
I9059
sS's'
I15
sS'm'
(lp5747
I1
aI3
aI5
aI5
aI1
aI19
aI29
aI145
aI163
aI149
aI619
aI2603
aI7757
aI10035
aI10189
asS'd'
I2874
sa(dp5748
S'a'
I9066
sS's'
I15
sS'm'
(lp5749
I1
aI3
aI7
aI15
aI27
aI15
aI111
aI173
aI135
aI117
aI157
aI2601
aI7919
aI12111
aI22795
asS'd'
I2875
sa(dp5750
S'a'
I9076
sS's'
I15
sS'm'
(lp5751
I1
aI3
aI1
aI1
aI29
aI27
aI65
aI31
aI101
aI715
aI289
aI3643
aI2335
aI6789
aI23397
asS'd'
I2876
sa(dp5752
S'a'
I9089
sS's'
I15
sS'm'
(lp5753
I1
aI3
aI1
aI3
aI11
aI45
aI71
aI109
aI321
aI423
aI1695
aI169
aI3075
aI12423
aI11391
asS'd'
I2877
sa(dp5754
S'a'
I9129
sS's'
I15
sS'm'
(lp5755
I1
aI1
aI3
aI9
aI13
aI51
aI35
aI121
aI203
aI279
aI433
aI2725
aI7951
aI2105
aI27333
asS'd'
I2878
sa(dp5756
S'a'
I9132
sS's'
I15
sS'm'
(lp5757
I1
aI1
aI1
aI15
aI23
aI31
aI25
aI105
aI501
aI441
aI1511
aI3133
aI2811
aI10595
aI21779
asS'd'
I2879
sa(dp5758
S'a'
I9147
sS's'
I15
sS'm'
(lp5759
I1
aI1
aI5
aI13
aI7
aI1
aI97
aI193
aI121
aI993
aI1347
aI1903
aI1883
aI6583
aI24535
asS'd'
I2880
sa(dp5760
S'a'
I9164
sS's'
I15
sS'm'
(lp5761
I1
aI1
aI7
aI9
aI7
aI29
aI17
aI41
aI101
aI447
aI1289
aI387
aI1891
aI2723
aI26091
asS'd'
I2881
sa(dp5762
S'a'
I9167
sS's'
I15
sS'm'
(lp5763
I1
aI1
aI3
aI3
aI3
aI53
aI81
aI81
aI177
aI165
aI195
aI3413
aI8177
aI3817
aI8453
asS'd'
I2882
sa(dp5764
S'a'
I9185
sS's'
I15
sS'm'
(lp5765
I1
aI3
aI7
aI15
aI15
aI31
aI23
aI31
aI337
aI439
aI1773
aI63
aI5351
aI5491
aI1767
asS'd'
I2883
sa(dp5766
S'a'
I9195
sS's'
I15
sS'm'
(lp5767
I1
aI3
aI1
aI11
aI5
aI15
aI23
aI75
aI437
aI553
aI429
aI2705
aI3625
aI13851
aI19865
asS'd'
I2884
sa(dp5768
S'a'
I9197
sS's'
I15
sS'm'
(lp5769
I1
aI3
aI3
aI9
aI13
aI15
aI33
aI235
aI215
aI415
aI1737
aI1409
aI2101
aI14623
aI14717
asS'd'
I2885
sa(dp5770
S'a'
I9210
sS's'
I15
sS'm'
(lp5771
I1
aI3
aI7
aI7
aI13
aI51
aI101
aI217
aI175
aI813
aI1639
aI4009
aI1625
aI4991
aI17525
asS'd'
I2886
sa(dp5772
S'a'
I9217
sS's'
I15
sS'm'
(lp5773
I1
aI1
aI5
aI13
aI23
aI33
aI29
aI175
aI39
aI673
aI557
aI3239
aI5129
aI11049
aI27227
asS'd'
I2887
sa(dp5774
S'a'
I9229
sS's'
I15
sS'm'
(lp5775
I1
aI3
aI7
aI13
aI1
aI37
aI33
aI139
aI493
aI891
aI1883
aI2525
aI5741
aI15795
aI5875
asS'd'
I2888
sa(dp5776
S'a'
I9248
sS's'
I15
sS'm'
(lp5777
I1
aI3
aI1
aI15
aI15
aI27
aI127
aI111
aI147
aI363
aI725
aI3077
aI4341
aI9131
aI24635
asS'd'
I2889
sa(dp5778
S'a'
I9254
sS's'
I15
sS'm'
(lp5779
I1
aI1
aI7
aI3
aI17
aI25
aI59
aI135
aI177
aI635
aI73
aI3455
aI3083
aI6009
aI13033
asS'd'
I2890
sa(dp5780
S'a'
I9263
sS's'
I15
sS'm'
(lp5781
I1
aI1
aI1
aI5
aI15
aI53
aI93
aI161
aI215
aI459
aI1087
aI179
aI2235
aI8885
aI15309
asS'd'
I2891
sa(dp5782
S'a'
I9266
sS's'
I15
sS'm'
(lp5783
I1
aI1
aI7
aI13
aI7
aI17
aI75
aI173
aI449
aI855
aI103
aI2739
aI3421
aI11811
aI18805
asS'd'
I2892
sa(dp5784
S'a'
I9268
sS's'
I15
sS'm'
(lp5785
I1
aI1
aI7
aI9
aI5
aI11
aI53
aI75
aI247
aI249
aI1201
aI953
aI2455
aI4589
aI6027
asS'd'
I2893
sa(dp5786
S'a'
I9290
sS's'
I15
sS'm'
(lp5787
I1
aI1
aI5
aI13
aI27
aI51
aI119
aI39
aI137
aI11
aI1435
aI3773
aI3889
aI6081
aI11829
asS'd'
I2894
sa(dp5788
S'a'
I9310
sS's'
I15
sS'm'
(lp5789
I1
aI1
aI5
aI5
aI5
aI35
aI1
aI197
aI501
aI185
aI1039
aI1563
aI6421
aI14373
aI25655
asS'd'
I2895
sa(dp5790
S'a'
I9316
sS's'
I15
sS'm'
(lp5791
I1
aI1
aI3
aI13
aI31
aI55
aI115
aI183
aI483
aI655
aI1351
aI3203
aI725
aI3299
aI22579
asS'd'
I2896
sa(dp5792
S'a'
I9338
sS's'
I15
sS'm'
(lp5793
I1
aI3
aI5
aI11
aI31
aI31
aI83
aI59
aI395
aI21
aI1881
aI2821
aI2251
aI11781
aI26265
asS'd'
I2897
sa(dp5794
S'a'
I9340
sS's'
I15
sS'm'
(lp5795
I1
aI3
aI7
aI13
aI21
aI19
aI103
aI21
aI403
aI443
aI1951
aI55
aI985
aI15983
aI15087
asS'd'
I2898
sa(dp5796
S'a'
I9343
sS's'
I15
sS'm'
(lp5797
I1
aI1
aI5
aI15
aI29
aI11
aI51
aI53
aI255
aI183
aI1475
aI1491
aI259
aI387
aI10303
asS'd'
I2899
sa(dp5798
S'a'
I9344
sS's'
I15
sS'm'
(lp5799
I1
aI3
aI5
aI7
aI21
aI37
aI45
aI39
aI479
aI637
aI1325
aI3753
aI3319
aI7403
aI31759
asS'd'
I2900
sa(dp5800
S'a'
I9350
sS's'
I15
sS'm'
(lp5801
I1
aI1
aI3
aI5
aI7
aI43
aI89
aI53
aI269
aI187
aI995
aI141
aI119
aI8139
aI29699
asS'd'
I2901
sa(dp5802
S'a'
I9354
sS's'
I15
sS'm'
(lp5803
I1
aI1
aI1
aI5
aI1
aI53
aI3
aI23
aI379
aI223
aI1889
aI4035
aI1437
aI12425
aI9051
asS'd'
I2902
sa(dp5804
S'a'
I9359
sS's'
I15
sS'm'
(lp5805
I1
aI3
aI1
aI13
aI3
aI31
aI61
aI43
aI249
aI449
aI901
aI1921
aI3495
aI8599
aI5263
asS'd'
I2903
sa(dp5806
S'a'
I9361
sS's'
I15
sS'm'
(lp5807
I1
aI1
aI3
aI5
aI3
aI25
aI35
aI133
aI25
aI597
aI915
aI3663
aI5147
aI11831
aI24269
asS'd'
I2904
sa(dp5808
S'a'
I9364
sS's'
I15
sS'm'
(lp5809
I1
aI1
aI1
aI9
aI21
aI27
aI93
aI93
aI217
aI299
aI1881
aI3647
aI4825
aI7989
aI24121
asS'd'
I2905
sa(dp5810
S'a'
I9368
sS's'
I15
sS'm'
(lp5811
I1
aI3
aI1
aI15
aI5
aI15
aI49
aI129
aI315
aI631
aI2037
aI1567
aI4043
aI15589
aI30905
asS'd'
I2906
sa(dp5812
S'a'
I9371
sS's'
I15
sS'm'
(lp5813
I1
aI3
aI3
aI7
aI25
aI5
aI123
aI51
aI47
aI471
aI1563
aI3947
aI7975
aI3681
aI9611
asS'd'
I2907
sa(dp5814
S'a'
I9373
sS's'
I15
sS'm'
(lp5815
I1
aI3
aI7
aI15
aI1
aI17
aI73
aI245
aI465
aI95
aI95
aI1159
aI1319
aI4675
aI8841
asS'd'
I2908
sa(dp5816
S'a'
I9389
sS's'
I15
sS'm'
(lp5817
I1
aI1
aI3
aI15
aI5
aI51
aI35
aI71
aI423
aI651
aI753
aI173
aI2131
aI15799
aI29601
asS'd'
I2909
sa(dp5818
S'a'
I9390
sS's'
I15
sS'm'
(lp5819
I1
aI1
aI1
aI1
aI3
aI53
aI83
aI187
aI445
aI827
aI1549
aI979
aI5363
aI1701
aI2149
asS'd'
I2910
sa(dp5820
S'a'
I9409
sS's'
I15
sS'm'
(lp5821
I1
aI1
aI7
aI9
aI3
aI15
aI65
aI161
aI37
aI233
aI771
aI3749
aI727
aI6857
aI17175
asS'd'
I2911
sa(dp5822
S'a'
I9443
sS's'
I15
sS'm'
(lp5823
I1
aI1
aI7
aI7
aI27
aI29
aI107
aI247
aI249
aI353
aI773
aI3677
aI7273
aI5419
aI29397
asS'd'
I2912
sa(dp5824
S'a'
I9445
sS's'
I15
sS'm'
(lp5825
I1
aI3
aI3
aI7
aI31
aI49
aI87
aI159
aI145
aI497
aI1715
aI2115
aI5035
aI6431
aI7245
asS'd'
I2913
sa(dp5826
S'a'
I9446
sS's'
I15
sS'm'
(lp5827
I1
aI3
aI3
aI5
aI7
aI31
aI51
aI117
aI101
aI617
aI557
aI2551
aI6589
aI13295
aI31975
asS'd'
I2914
sa(dp5828
S'a'
I9452
sS's'
I15
sS'm'
(lp5829
I1
aI1
aI3
aI3
aI15
aI27
aI125
aI163
aI169
aI893
aI1771
aI25
aI5787
aI10267
aI10297
asS'd'
I2915
sa(dp5830
S'a'
I9490
sS's'
I15
sS'm'
(lp5831
I1
aI1
aI1
aI5
aI9
aI47
aI85
aI65
aI289
aI783
aI1105
aI4035
aI4111
aI2589
aI24575
asS'd'
I2916
sa(dp5832
S'a'
I9492
sS's'
I15
sS'm'
(lp5833
I1
aI3
aI3
aI13
aI23
aI33
aI7
aI49
aI301
aI531
aI1713
aI2755
aI5543
aI8153
aI24099
asS'd'
I2917
sa(dp5834
S'a'
I9495
sS's'
I15
sS'm'
(lp5835
I1
aI1
aI5
aI9
aI7
aI39
aI101
aI67
aI417
aI923
aI757
aI1537
aI5553
aI12233
aI20881
asS'd'
I2918
sa(dp5836
S'a'
I9508
sS's'
I15
sS'm'
(lp5837
I1
aI1
aI5
aI1
aI19
aI7
aI25
aI123
aI125
aI183
aI573
aI3317
aI6867
aI871
aI17631
asS'd'
I2919
sa(dp5838
S'a'
I9523
sS's'
I15
sS'm'
(lp5839
I1
aI1
aI3
aI15
aI19
aI13
aI117
aI41
aI129
aI715
aI1525
aI2257
aI2179
aI10807
aI23271
asS'd'
I2920
sa(dp5840
S'a'
I9543
sS's'
I15
sS'm'
(lp5841
I1
aI3
aI1
aI5
aI25
aI53
aI19
aI169
aI289
aI569
aI1135
aI1967
aI7001
aI15883
aI15113
asS'd'
I2921
sa(dp5842
S'a'
I9558
sS's'
I15
sS'm'
(lp5843
I1
aI3
aI7
aI15
aI7
aI37
aI127
aI147
aI415
aI313
aI1541
aI1889
aI3763
aI16199
aI12681
asS'd'
I2922
sa(dp5844
S'a'
I9567
sS's'
I15
sS'm'
(lp5845
I1
aI1
aI3
aI9
aI1
aI35
aI95
aI137
aI237
aI951
aI899
aI3177
aI6073
aI10655
aI31687
asS'd'
I2923
sa(dp5846
S'a'
I9580
sS's'
I15
sS'm'
(lp5847
I1
aI1
aI5
aI5
aI29
aI57
aI45
aI253
aI297
aI529
aI1553
aI467
aI8035
aI15675
aI21691
asS'd'
I2924
sa(dp5848
S'a'
I9585
sS's'
I15
sS'm'
(lp5849
I1
aI3
aI7
aI15
aI25
aI41
aI59
aI81
aI87
aI985
aI1001
aI2369
aI661
aI7551
aI11829
asS'd'
I2925
sa(dp5850
S'a'
I9591
sS's'
I15
sS'm'
(lp5851
I1
aI1
aI7
aI9
aI27
aI21
aI7
aI233
aI309
aI67
aI701
aI2737
aI4261
aI2467
aI15691
asS'd'
I2926
sa(dp5852
S'a'
I9611
sS's'
I15
sS'm'
(lp5853
I1
aI3
aI7
aI1
aI19
aI55
aI47
aI155
aI333
aI101
aI517
aI1991
aI4619
aI10435
aI27241
asS'd'
I2927
sa(dp5854
S'a'
I9613
sS's'
I15
sS'm'
(lp5855
I1
aI1
aI7
aI3
aI23
aI35
aI7
aI125
aI157
aI537
aI933
aI3281
aI4975
aI8969
aI27581
asS'd'
I2928
sa(dp5856
S'a'
I9614
sS's'
I15
sS'm'
(lp5857
I1
aI1
aI3
aI7
aI19
aI53
aI81
aI103
aI461
aI435
aI777
aI335
aI5261
aI12249
aI9695
asS'd'
I2929
sa(dp5858
S'a'
I9621
sS's'
I15
sS'm'
(lp5859
I1
aI3
aI1
aI7
aI19
aI9
aI75
aI245
aI355
aI37
aI1855
aI1339
aI3107
aI7251
aI16543
asS'd'
I2930
sa(dp5860
S'a'
I9631
sS's'
I15
sS'm'
(lp5861
I1
aI1
aI1
aI3
aI5
aI35
aI39
aI223
aI113
aI423
aI1423
aI713
aI6113
aI349
aI24147
asS'd'
I2931
sa(dp5862
S'a'
I9642
sS's'
I15
sS'm'
(lp5863
I1
aI3
aI1
aI1
aI15
aI31
aI11
aI75
aI499
aI345
aI1253
aI2629
aI2551
aI7483
aI25395
asS'd'
I2932
sa(dp5864
S'a'
I9656
sS's'
I15
sS'm'
(lp5865
I1
aI1
aI3
aI11
aI25
aI25
aI3
aI211
aI185
aI45
aI1865
aI1805
aI3303
aI11091
aI529
asS'd'
I2933
sa(dp5866
S'a'
I9661
sS's'
I15
sS'm'
(lp5867
I1
aI3
aI1
aI1
aI9
aI21
aI7
aI165
aI107
aI641
aI1083
aI2805
aI2099
aI5855
aI18477
asS'd'
I2934
sa(dp5868
S'a'
I9667
sS's'
I15
sS'm'
(lp5869
I1
aI3
aI5
aI3
aI9
aI21
aI77
aI103
aI505
aI277
aI335
aI797
aI3869
aI2957
aI1979
asS'd'
I2935
sa(dp5870
S'a'
I9694
sS's'
I15
sS'm'
(lp5871
I1
aI3
aI5
aI15
aI31
aI23
aI77
aI247
aI303
aI891
aI1261
aI3233
aI3495
aI13111
aI13185
asS'd'
I2936
sa(dp5872
S'a'
I9715
sS's'
I15
sS'm'
(lp5873
I1
aI3
aI5
aI11
aI11
aI35
aI49
aI229
aI149
aI931
aI881
aI775
aI2949
aI3141
aI29157
asS'd'
I2937
sa(dp5874
S'a'
I9722
sS's'
I15
sS'm'
(lp5875
I1
aI1
aI3
aI5
aI19
aI57
aI23
aI95
aI347
aI221
aI195
aI3561
aI1481
aI2063
aI3979
asS'd'
I2938
sa(dp5876
S'a'
I9738
sS's'
I15
sS'm'
(lp5877
I1
aI3
aI5
aI3
aI13
aI1
aI23
aI173
aI431
aI29
aI421
aI3235
aI2751
aI4447
aI28283
asS'd'
I2939
sa(dp5878
S'a'
I9745
sS's'
I15
sS'm'
(lp5879
I1
aI1
aI5
aI13
aI23
aI3
aI1
aI9
aI327
aI855
aI1251
aI2997
aI6129
aI4223
aI11555
asS'd'
I2940
sa(dp5880
S'a'
I9758
sS's'
I15
sS'm'
(lp5881
I1
aI3
aI7
aI13
aI29
aI21
aI37
aI229
aI217
aI353
aI1239
aI3955
aI491
aI12183
aI14777
asS'd'
I2941
sa(dp5882
S'a'
I9764
sS's'
I15
sS'm'
(lp5883
I1
aI1
aI5
aI5
aI1
aI33
aI103
aI187
aI183
aI939
aI1873
aI2633
aI6143
aI15405
aI17353
asS'd'
I2942
sa(dp5884
S'a'
I9782
sS's'
I15
sS'm'
(lp5885
I1
aI1
aI1
aI9
aI21
aI27
aI71
aI129
aI499
aI279
aI1181
aI4053
aI2485
aI1961
aI30603
asS'd'
I2943
sa(dp5886
S'a'
I9791
sS's'
I15
sS'm'
(lp5887
I1
aI1
aI3
aI15
aI21
aI37
aI45
aI201
aI221
aI187
aI727
aI1241
aI6171
aI1383
aI22277
asS'd'
I2944
sa(dp5888
S'a'
I9793
sS's'
I15
sS'm'
(lp5889
I1
aI3
aI7
aI5
aI21
aI17
aI67
aI177
aI323
aI601
aI633
aI865
aI6131
aI10329
aI8689
asS'd'
I2945
sa(dp5890
S'a'
I9794
sS's'
I15
sS'm'
(lp5891
I1
aI3
aI5
aI9
aI15
aI45
aI71
aI43
aI359
aI651
aI103
aI403
aI3249
aI11769
aI6567
asS'd'
I2946
sa(dp5892
S'a'
I9805
sS's'
I15
sS'm'
(lp5893
I1
aI3
aI3
aI13
aI3
aI23
aI101
aI145
aI367
aI999
aI1489
aI3673
aI2959
aI10855
aI16029
asS'd'
I2947
sa(dp5894
S'a'
I9808
sS's'
I15
sS'm'
(lp5895
I1
aI3
aI7
aI3
aI13
aI43
aI123
aI87
aI55
aI1015
aI141
aI2917
aI6567
aI16025
aI25555
asS'd'
I2948
sa(dp5896
S'a'
I9811
sS's'
I15
sS'm'
(lp5897
I1
aI3
aI1
aI3
aI17
aI7
aI21
aI161
aI41
aI889
aI1315
aI1897
aI639
aI15451
aI3049
asS'd'
I2949
sa(dp5898
S'a'
I9817
sS's'
I15
sS'm'
(lp5899
I1
aI3
aI5
aI15
aI27
aI33
aI55
aI17
aI81
aI431
aI325
aI909
aI3547
aI10121
aI17815
asS'd'
I2950
sa(dp5900
S'a'
I9824
sS's'
I15
sS'm'
(lp5901
I1
aI1
aI3
aI1
aI15
aI37
aI43
aI137
aI203
aI191
aI1129
aI1585
aI435
aI3177
aI769
asS'd'
I2951
sa(dp5902
S'a'
I9836
sS's'
I15
sS'm'
(lp5903
I1
aI3
aI7
aI11
aI21
aI23
aI125
aI41
aI17
aI951
aI465
aI3691
aI3465
aI13247
aI13779
asS'd'
I2952
sa(dp5904
S'a'
I9851
sS's'
I15
sS'm'
(lp5905
I1
aI3
aI3
aI1
aI31
aI23
aI43
aI101
aI405
aI739
aI1061
aI2955
aI5643
aI16137
aI8763
asS'd'
I2953
sa(dp5906
S'a'
I9853
sS's'
I15
sS'm'
(lp5907
I1
aI1
aI5
aI1
aI19
aI33
aI99
aI109
aI203
aI65
aI395
aI2775
aI1373
aI2557
aI5875
asS'd'
I2954
sa(dp5908
S'a'
I9854
sS's'
I15
sS'm'
(lp5909
I1
aI3
aI3
aI3
aI27
aI51
aI79
aI63
aI331
aI365
aI1071
aI1661
aI4549
aI8561
aI1719
asS'd'
I2955
sa(dp5910
S'a'
I9877
sS's'
I15
sS'm'
(lp5911
I1
aI3
aI3
aI9
aI3
aI17
aI53
aI161
aI141
aI489
aI1325
aI1709
aI1381
aI5093
aI171
asS'd'
I2956
sa(dp5912
S'a'
I9881
sS's'
I15
sS'm'
(lp5913
I1
aI1
aI7
aI15
aI9
aI3
aI95
aI237
aI197
aI949
aI7
aI1837
aI729
aI10111
aI6637
asS'd'
I2957
sa(dp5914
S'a'
I9923
sS's'
I15
sS'm'
(lp5915
I1
aI1
aI3
aI3
aI19
aI31
aI57
aI173
aI483
aI861
aI1001
aI1919
aI3389
aI11777
aI20693
asS'd'
I2958
sa(dp5916
S'a'
I9935
sS's'
I15
sS'm'
(lp5917
I1
aI3
aI1
aI9
aI27
aI13
aI113
aI177
aI75
aI925
aI949
aI119
aI4759
aI7775
aI23033
asS'd'
I2959
sa(dp5918
S'a'
I9937
sS's'
I15
sS'm'
(lp5919
I1
aI1
aI7
aI15
aI23
aI15
aI65
aI61
aI137
aI653
aI1843
aI323
aI379
aI15157
aI29885
asS'd'
I2960
sa(dp5920
S'a'
I9954
sS's'
I15
sS'm'
(lp5921
I1
aI3
aI3
aI7
aI29
aI3
aI11
aI205
aI347
aI745
aI1477
aI3929
aI5749
aI4735
aI29435
asS'd'
I2961
sa(dp5922
S'a'
I9959
sS's'
I15
sS'm'
(lp5923
I1
aI3
aI5
aI9
aI1
aI11
aI111
aI15
aI7
aI69
aI45
aI3607
aI1099
aI9203
aI21301
asS'd'
I2962
sa(dp5924
S'a'
I9963
sS's'
I15
sS'm'
(lp5925
I1
aI3
aI3
aI3
aI23
aI3
aI83
aI173
aI73
aI485
aI681
aI1867
aI3839
aI11823
aI13339
asS'd'
I2963
sa(dp5926
S'a'
I9968
sS's'
I15
sS'm'
(lp5927
I1
aI1
aI3
aI11
aI31
aI43
aI107
aI127
aI465
aI389
aI1595
aI427
aI1571
aI5885
aI29569
asS'd'
I2964
sa(dp5928
S'a'
I9973
sS's'
I15
sS'm'
(lp5929
I1
aI1
aI7
aI9
aI27
aI25
aI117
aI27
aI287
aI391
aI279
aI3247
aI35
aI12973
aI5483
asS'd'
I2965
sa(dp5930
S'a'
I9974
sS's'
I15
sS'm'
(lp5931
I1
aI3
aI7
aI11
aI19
aI55
aI45
aI127
aI245
aI945
aI305
aI3907
aI2455
aI3163
aI31
asS'd'
I2966
sa(dp5932
S'a'
I9980
sS's'
I15
sS'm'
(lp5933
I1
aI1
aI7
aI11
aI15
aI17
aI65
aI15
aI37
aI207
aI1447
aI3027
aI2281
aI6557
aI16717
asS'd'
I2967
sa(dp5934
S'a'
I9983
sS's'
I15
sS'm'
(lp5935
I1
aI1
aI1
aI13
aI5
aI27
aI33
aI213
aI29
aI603
aI1171
aI3235
aI2255
aI2017
aI30999
asS'd'
I2968
sa(dp5936
S'a'
I9985
sS's'
I15
sS'm'
(lp5937
I1
aI3
aI1
aI5
aI11
aI1
aI73
aI233
aI69
aI125
aI397
aI297
aI3337
aI6191
aI31055
asS'd'
I2969
sa(dp5938
S'a'
I10003
sS's'
I15
sS'm'
(lp5939
I1
aI1
aI1
aI15
aI1
aI1
aI65
aI145
aI201
aI917
aI1891
aI2999
aI4069
aI10413
aI15819
asS'd'
I2970
sa(dp5940
S'a'
I10010
sS's'
I15
sS'm'
(lp5941
I1
aI3
aI5
aI13
aI15
aI51
aI115
aI167
aI311
aI375
aI1069
aI2595
aI3337
aI753
aI11903
asS'd'
I2971
sa(dp5942
S'a'
I10034
sS's'
I15
sS'm'
(lp5943
I1
aI1
aI3
aI1
aI1
aI23
aI69
aI125
aI147
aI915
aI1945
aI411
aI979
aI13863
aI30443
asS'd'
I2972
sa(dp5944
S'a'
I10040
sS's'
I15
sS'm'
(lp5945
I1
aI3
aI1
aI11
aI5
aI1
aI93
aI23
aI135
aI93
aI1689
aI23
aI3519
aI4491
aI24673
asS'd'
I2973
sa(dp5946
S'a'
I10063
sS's'
I15
sS'm'
(lp5947
I1
aI1
aI7
aI3
aI11
aI59
aI93
aI153
aI487
aI475
aI1191
aI1455
aI5963
aI8259
aI18811
asS'd'
I2974
sa(dp5948
S'a'
I10077
sS's'
I15
sS'm'
(lp5949
I1
aI1
aI3
aI1
aI13
aI15
aI55
aI71
aI433
aI33
aI491
aI1835
aI5695
aI10509
aI347
asS'd'
I2975
sa(dp5950
S'a'
I10081
sS's'
I15
sS'm'
(lp5951
I1
aI1
aI1
aI15
aI19
aI1
aI23
aI47
aI235
aI101
aI1057
aI901
aI5477
aI7079
aI30885
asS'd'
I2976
sa(dp5952
S'a'
I10082
sS's'
I15
sS'm'
(lp5953
I1
aI1
aI5
aI13
aI11
aI43
aI119
aI77
aI441
aI121
aI783
aI827
aI1757
aI12751
aI31593
asS'd'
I2977
sa(dp5954
S'a'
I10084
sS's'
I15
sS'm'
(lp5955
I1
aI3
aI7
aI11
aI19
aI17
aI37
aI225
aI329
aI231
aI515
aI1541
aI7371
aI6355
aI10905
asS'd'
I2978
sa(dp5956
S'a'
I10088
sS's'
I15
sS'm'
(lp5957
I1
aI1
aI5
aI13
aI7
aI11
aI35
aI215
aI345
aI577
aI147
aI2803
aI3291
aI4631
aI5329
asS'd'
I2979
sa(dp5958
S'a'
I10091
sS's'
I15
sS'm'
(lp5959
I1
aI1
aI3
aI9
aI21
aI55
aI113
aI251
aI25
aI221
aI1445
aI3385
aI1589
aI4109
aI29897
asS'd'
I2980
sa(dp5960
S'a'
I10105
sS's'
I15
sS'm'
(lp5961
I1
aI1
aI5
aI7
aI9
aI45
aI5
aI33
aI331
aI285
aI1101
aI3131
aI2713
aI5653
aI3823
asS'd'
I2981
sa(dp5962
S'a'
I10111
sS's'
I15
sS'm'
(lp5963
I1
aI3
aI7
aI7
aI5
aI39
aI43
aI167
aI481
aI629
aI777
aI1827
aI4653
aI403
aI4781
asS'd'
I2982
sa(dp5964
S'a'
I10118
sS's'
I15
sS'm'
(lp5965
I1
aI3
aI3
aI7
aI31
aI33
aI31
aI159
aI313
aI673
aI1425
aI663
aI5819
aI1297
aI26627
asS'd'
I2983
sa(dp5966
S'a'
I10127
sS's'
I15
sS'm'
(lp5967
I1
aI3
aI3
aI1
aI19
aI61
aI117
aI93
aI373
aI491
aI1031
aI757
aI4185
aI771
aI7265
asS'd'
I2984
sa(dp5968
S'a'
I10135
sS's'
I15
sS'm'
(lp5969
I1
aI1
aI7
aI9
aI3
aI45
aI65
aI223
aI437
aI41
aI1139
aI2733
aI5963
aI2709
aI25429
asS'd'
I2985
sa(dp5970
S'a'
I10169
sS's'
I15
sS'm'
(lp5971
I1
aI3
aI5
aI11
aI21
aI27
aI31
aI127
aI255
aI761
aI1865
aI1319
aI6583
aI9235
aI10717
asS'd'
I2986
sa(dp5972
S'a'
I10172
sS's'
I15
sS'm'
(lp5973
I1
aI1
aI1
aI5
aI21
aI1
aI63
aI43
aI413
aI557
aI567
aI2893
aI8017
aI2307
aI29525
asS'd'
I2987
sa(dp5974
S'a'
I10183
sS's'
I15
sS'm'
(lp5975
I1
aI1
aI7
aI3
aI31
aI1
aI15
aI235
aI215
aI395
aI1971
aI469
aI5275
aI431
aI5349
asS'd'
I2988
sa(dp5976
S'a'
I10190
sS's'
I15
sS'm'
(lp5977
I1
aI1
aI1
aI13
aI25
aI59
aI71
aI245
aI389
aI279
aI1293
aI89
aI6551
aI10285
aI14495
asS'd'
I2989
sa(dp5978
S'a'
I10192
sS's'
I15
sS'm'
(lp5979
I1
aI1
aI5
aI5
aI9
aI63
aI17
aI229
aI425
aI939
aI877
aI3689
aI7229
aI6707
aI30771
asS'd'
I2990
sa(dp5980
S'a'
I10211
sS's'
I15
sS'm'
(lp5981
I1
aI3
aI7
aI7
aI11
aI29
aI43
aI41
aI25
aI237
aI1585
aI3735
aI2617
aI7541
aI26243
asS'd'
I2991
sa(dp5982
S'a'
I10218
sS's'
I15
sS'm'
(lp5983
I1
aI1
aI7
aI9
aI21
aI5
aI69
aI231
aI361
aI39
aI1695
aI3043
aI2973
aI5487
aI12857
asS'd'
I2992
sa(dp5984
S'a'
I10228
sS's'
I15
sS'm'
(lp5985
I1
aI1
aI5
aI3
aI17
aI63
aI91
aI217
aI407
aI133
aI1373
aI4021
aI1737
aI10043
aI4561
asS'd'
I2993
sa(dp5986
S'a'
I10235
sS's'
I15
sS'm'
(lp5987
I1
aI3
aI7
aI9
aI31
aI13
aI101
aI231
aI175
aI457
aI89
aI2167
aI2725
aI8505
aI375
asS'd'
I2994
sa(dp5988
S'a'
I10242
sS's'
I15
sS'm'
(lp5989
I1
aI1
aI3
aI15
aI31
aI11
aI27
aI211
aI347
aI291
aI1881
aI3091
aI3307
aI5117
aI13341
asS'd'
I2995
sa(dp5990
S'a'
I10244
sS's'
I15
sS'm'
(lp5991
I1
aI3
aI5
aI5
aI13
aI25
aI5
aI197
aI237
aI135
aI635
aI1175
aI5085
aI14737
aI10807
asS'd'
I2996
sa(dp5992
S'a'
I10271
sS's'
I15
sS'm'
(lp5993
I1
aI3
aI3
aI9
aI7
aI63
aI107
aI127
aI147
aI477
aI1813
aI2619
aI8089
aI2651
aI26549
asS'd'
I2997
sa(dp5994
S'a'
I10278
sS's'
I15
sS'm'
(lp5995
I1
aI1
aI5
aI11
aI15
aI45
aI27
aI133
aI45
aI621
aI707
aI2679
aI5929
aI19
aI9599
asS'd'
I2998
sa(dp5996
S'a'
I10296
sS's'
I15
sS'm'
(lp5997
I1
aI3
aI7
aI9
aI21
aI37
aI41
aI255
aI69
aI1009
aI1999
aI367
aI6177
aI10017
aI3549
asS'd'
I2999
sa(dp5998
S'a'
I10299
sS's'
I15
sS'm'
(lp5999
I1
aI1
aI1
aI15
aI19
aI55
aI73
aI189
aI423
aI983
aI1811
aI2551
aI4765
aI12077
aI18205
asS'd'
I3000
sa(dp6000
S'a'
I10307
sS's'
I15
sS'm'
(lp6001
I1
aI1
aI5
aI7
aI17
aI13
aI25
aI225
aI463
aI471
aI631
aI1811
aI5797
aI3235
aI32253
asS'd'
I3001
sa(dp6002
S'a'
I10309
sS's'
I15
sS'm'
(lp6003
I1
aI3
aI7
aI1
aI29
aI7
aI123
aI187
aI331
aI735
aI1757
aI1115
aI2077
aI15725
aI2183
asS'd'
I3002
sa(dp6004
S'a'
I10310
sS's'
I15
sS'm'
(lp6005
I1
aI3
aI7
aI9
aI17
aI61
aI111
aI93
aI21
aI1003
aI1905
aI3719
aI2111
aI11845
aI6427
asS'd'
I3003
sa(dp6006
S'a'
I10314
sS's'
I15
sS'm'
(lp6007
I1
aI3
aI7
aI7
aI17
aI21
aI51
aI59
aI115
aI723
aI2039
aI2833
aI5969
aI5737
aI18407
asS'd'
I3004
sa(dp6008
S'a'
I10316
sS's'
I15
sS'm'
(lp6009
I1
aI3
aI3
aI13
aI9
aI47
aI95
aI233
aI13
aI281
aI1049
aI619
aI405
aI16205
aI20097
asS'd'
I3005
sa(dp6010
S'a'
I10321
sS's'
I15
sS'm'
(lp6011
I1
aI3
aI7
aI13
aI9
aI41
aI11
aI171
aI453
aI713
aI587
aI1669
aI2489
aI10277
aI18599
asS'd'
I3006
sa(dp6012
S'a'
I10328
sS's'
I15
sS'm'
(lp6013
I1
aI3
aI3
aI13
aI21
aI41
aI123
aI173
aI511
aI399
aI859
aI1515
aI5773
aI12535
aI26289
asS'd'
I3007
sa(dp6014
S'a'
I10338
sS's'
I15
sS'm'
(lp6015
I1
aI1
aI7
aI15
aI11
aI3
aI113
aI111
aI73
aI7
aI1191
aI2573
aI7713
aI465
aI27615
asS'd'
I3008
sa(dp6016
S'a'
I10343
sS's'
I15
sS'm'
(lp6017
I1
aI1
aI7
aI15
aI5
aI5
aI39
aI11
aI489
aI13
aI1041
aI1639
aI7879
aI11899
aI6899
asS'd'
I3009
sa(dp6018
S'a'
I10344
sS's'
I15
sS'm'
(lp6019
I1
aI1
aI5
aI9
aI27
aI31
aI15
aI237
aI401
aI795
aI1675
aI2361
aI7333
aI12507
aI14627
asS'd'
I3010
sa(dp6020
S'a'
I10347
sS's'
I15
sS'm'
(lp6021
I1
aI3
aI1
aI7
aI21
aI53
aI31
aI81
aI189
aI683
aI1283
aI419
aI7585
aI9207
aI15053
asS'd'
I3011
sa(dp6022
S'a'
I10352
sS's'
I15
sS'm'
(lp6023
I1
aI3
aI5
aI11
aI21
aI1
aI49
aI251
aI403
aI489
aI1235
aI429
aI4855
aI4081
aI17575
asS'd'
I3012
sa(dp6024
S'a'
I10364
sS's'
I15
sS'm'
(lp6025
I1
aI3
aI1
aI15
aI29
aI33
aI77
aI53
aI105
aI731
aI749
aI2677
aI3967
aI7967
aI18723
asS'd'
I3013
sa(dp6026
S'a'
I10373
sS's'
I15
sS'm'
(lp6027
I1
aI3
aI3
aI11
aI9
aI47
aI11
aI95
aI137
aI923
aI599
aI1585
aI1969
aI9625
aI19171
asS'd'
I3014
sa(dp6028
S'a'
I10386
sS's'
I15
sS'm'
(lp6029
I1
aI1
aI1
aI5
aI7
aI7
aI85
aI49
aI339
aI883
aI261
aI2125
aI3831
aI9797
aI16395
asS'd'
I3015
sa(dp6030
S'a'
I10391
sS's'
I15
sS'm'
(lp6031
I1
aI3
aI3
aI3
aI5
aI9
aI33
aI99
aI75
aI889
aI101
aI2099
aI6635
aI11511
aI21573
asS'd'
I3016
sa(dp6032
S'a'
I10398
sS's'
I15
sS'm'
(lp6033
I1
aI1
aI5
aI11
aI1
aI11
aI79
aI49
aI7
aI131
aI471
aI1235
aI3287
aI14777
aI12053
asS'd'
I3017
sa(dp6034
S'a'
I10408
sS's'
I15
sS'm'
(lp6035
I1
aI1
aI5
aI15
aI9
aI9
aI83
aI15
aI21
aI899
aI1785
aI2349
aI3471
aI6723
aI1405
asS'd'
I3018
sa(dp6036
S'a'
I10413
sS's'
I15
sS'm'
(lp6037
I1
aI3
aI5
aI11
aI1
aI7
aI121
aI223
aI509
aI859
aI1037
aI491
aI5529
aI481
aI17029
asS'd'
I3019
sa(dp6038
S'a'
I10422
sS's'
I15
sS'm'
(lp6039
I1
aI1
aI7
aI5
aI17
aI35
aI91
aI171
aI113
aI65
aI423
aI2371
aI5105
aI12827
aI31087
asS'd'
I3020
sa(dp6040
S'a'
I10445
sS's'
I15
sS'm'
(lp6041
I1
aI1
aI3
aI3
aI21
aI47
aI55
aI11
aI159
aI51
aI263
aI2113
aI661
aI9147
aI28929
asS'd'
I3021
sa(dp6042
S'a'
I10460
sS's'
I15
sS'm'
(lp6043
I1
aI1
aI1
aI9
aI19
aI7
aI43
aI223
aI207
aI787
aI543
aI2141
aI4247
aI7369
aI29031
asS'd'
I3022
sa(dp6044
S'a'
I10463
sS's'
I15
sS'm'
(lp6045
I1
aI1
aI7
aI11
aI11
aI51
aI121
aI9
aI211
aI905
aI687
aI889
aI1827
aI13395
aI3507
asS'd'
I3023
sa(dp6046
S'a'
I10464
sS's'
I15
sS'm'
(lp6047
I1
aI3
aI1
aI7
aI15
aI23
aI5
aI139
aI469
aI569
aI33
aI3477
aI5391
aI13665
aI17011
asS'd'
I3024
sa(dp6048
S'a'
I10474
sS's'
I15
sS'm'
(lp6049
I1
aI1
aI1
aI15
aI29
aI29
aI29
aI201
aI63
aI1019
aI97
aI1671
aI9
aI4617
aI19833
asS'd'
I3025
sa(dp6050
S'a'
I10476
sS's'
I15
sS'm'
(lp6051
I1
aI1
aI5
aI15
aI25
aI5
aI67
aI225
aI189
aI919
aI1471
aI1451
aI5017
aI16189
aI31555
asS'd'
I3026
sa(dp6052
S'a'
I10487
sS's'
I15
sS'm'
(lp6053
I1
aI3
aI5
aI5
aI15
aI51
aI89
aI221
aI149
aI863
aI43
aI2381
aI1767
aI8037
aI5319
asS'd'
I3027
sa(dp6054
S'a'
I10494
sS's'
I15
sS'm'
(lp6055
I1
aI3
aI3
aI1
aI15
aI17
aI5
aI77
aI69
aI27
aI1883
aI63
aI5987
aI1497
aI3723
asS'd'
I3028
sa(dp6056
S'a'
I10499
sS's'
I15
sS'm'
(lp6057
I1
aI3
aI7
aI11
aI7
aI5
aI113
aI229
aI123
aI709
aI1531
aI641
aI6655
aI14923
aI22947
asS'd'
I3029
sa(dp6058
S'a'
I10506
sS's'
I15
sS'm'
(lp6059
I1
aI3
aI1
aI13
aI21
aI15
aI45
aI175
aI81
aI499
aI1113
aI587
aI7573
aI11689
aI15651
asS'd'
I3030
sa(dp6060
S'a'
I10513
sS's'
I15
sS'm'
(lp6061
I1
aI3
aI1
aI1
aI29
aI43
aI101
aI37
aI131
aI757
aI465
aI743
aI2737
aI8063
aI23967
asS'd'
I3031
sa(dp6062
S'a'
I10516
sS's'
I15
sS'm'
(lp6063
I1
aI1
aI7
aI13
aI9
aI21
aI39
aI177
aI51
aI691
aI2047
aI1519
aI6137
aI5271
aI8703
asS'd'
I3032
sa(dp6064
S'a'
I10523
sS's'
I15
sS'm'
(lp6065
I1
aI1
aI3
aI3
aI5
aI55
aI63
aI21
aI3
aI317
aI461
aI527
aI2673
aI16211
aI6721
asS'd'
I3033
sa(dp6066
S'a'
I10539
sS's'
I15
sS'm'
(lp6067
I1
aI3
aI5
aI5
aI5
aI47
aI7
aI241
aI387
aI589
aI323
aI203
aI7241
aI14761
aI13287
asS'd'
I3034
sa(dp6068
S'a'
I10549
sS's'
I15
sS'm'
(lp6069
I1
aI3
aI5
aI3
aI23
aI63
aI55
aI61
aI231
aI1023
aI1315
aI1181
aI243
aI7389
aI25639
asS'd'
I3035
sa(dp6070
S'a'
I10550
sS's'
I15
sS'm'
(lp6071
I1
aI1
aI7
aI13
aI31
aI43
aI41
aI81
aI127
aI887
aI1513
aI4055
aI1361
aI2443
aI6963
asS'd'
I3036
sa(dp6072
S'a'
I10567
sS's'
I15
sS'm'
(lp6073
I1
aI1
aI1
aI5
aI7
aI43
aI43
aI33
aI323
aI911
aI1373
aI3053
aI6503
aI513
aI6457
asS'd'
I3037
sa(dp6074
S'a'
I10576
sS's'
I15
sS'm'
(lp6075
I1
aI1
aI7
aI11
aI25
aI61
aI21
aI149
aI205
aI349
aI1433
aI1587
aI149
aI7275
aI5465
asS'd'
I3038
sa(dp6076
S'a'
I10625
sS's'
I15
sS'm'
(lp6077
I1
aI3
aI5
aI5
aI11
aI9
aI31
aI217
aI119
aI511
aI209
aI3325
aI2023
aI2877
aI463
asS'd'
I3039
sa(dp6078
S'a'
I10635
sS's'
I15
sS'm'
(lp6079
I1
aI3
aI5
aI15
aI21
aI47
aI89
aI41
aI347
aI849
aI1375
aI3311
aI807
aI11443
aI27643
asS'd'
I3040
sa(dp6080
S'a'
I10643
sS's'
I15
sS'm'
(lp6081
I1
aI1
aI5
aI7
aI29
aI43
aI123
aI191
aI321
aI373
aI447
aI2145
aI1221
aI2071
aI12689
asS'd'
I3041
sa(dp6082
S'a'
I10656
sS's'
I15
sS'm'
(lp6083
I1
aI3
aI5
aI15
aI1
aI21
aI43
aI141
aI461
aI779
aI1109
aI2915
aI909
aI8585
aI19859
asS'd'
I3042
sa(dp6084
S'a'
I10671
sS's'
I15
sS'm'
(lp6085
I1
aI3
aI3
aI11
aI5
aI17
aI57
aI13
aI393
aI661
aI1761
aI2455
aI43
aI8593
aI20505
asS'd'
I3043
sa(dp6086
S'a'
I10676
sS's'
I15
sS'm'
(lp6087
I1
aI3
aI5
aI1
aI31
aI47
aI65
aI249
aI77
aI513
aI851
aI2381
aI3447
aI693
aI7729
asS'd'
I3044
sa(dp6088
S'a'
I10683
sS's'
I15
sS'm'
(lp6089
I1
aI3
aI5
aI15
aI31
aI19
aI83
aI47
aI369
aI697
aI1815
aI819
aI7573
aI9245
aI8013
asS'd'
I3045
sa(dp6090
S'a'
I10685
sS's'
I15
sS'm'
(lp6091
I1
aI3
aI5
aI5
aI11
aI25
aI27
aI151
aI107
aI339
aI299
aI3869
aI3393
aI5661
aI2947
asS'd'
I3046
sa(dp6092
S'a'
I10688
sS's'
I15
sS'm'
(lp6093
I1
aI1
aI3
aI1
aI1
aI59
aI85
aI57
aI175
aI465
aI239
aI3115
aI7157
aI7035
aI11463
asS'd'
I3047
sa(dp6094
S'a'
I10697
sS's'
I15
sS'm'
(lp6095
I1
aI1
aI7
aI5
aI31
aI41
aI53
aI149
aI121
aI743
aI189
aI159
aI5289
aI2945
aI1179
asS'd'
I3048
sa(dp6096
S'a'
I10700
sS's'
I15
sS'm'
(lp6097
I1
aI3
aI3
aI15
aI23
aI51
aI83
aI25
aI159
aI163
aI61
aI713
aI4529
aI5253
aI1603
asS'd'
I3049
sa(dp6098
S'a'
I10712
sS's'
I15
sS'm'
(lp6099
I1
aI3
aI5
aI11
aI7
aI29
aI15
aI177
aI507
aI695
aI1305
aI1863
aI7525
aI3063
aI27433
asS'd'
I3050
sa(dp6100
S'a'
I10724
sS's'
I15
sS'm'
(lp6101
I1
aI1
aI3
aI11
aI5
aI41
aI115
aI227
aI409
aI951
aI591
aI4003
aI7717
aI4369
aI15637
asS'd'
I3051
sa(dp6102
S'a'
I10728
sS's'
I15
sS'm'
(lp6103
I1
aI1
aI7
aI11
aI23
aI55
aI71
aI135
aI51
aI411
aI2003
aI2375
aI6823
aI1711
aI4443
asS'd'
I3052
sa(dp6104
S'a'
I10734
sS's'
I15
sS'm'
(lp6105
I1
aI3
aI1
aI3
aI31
aI47
aI31
aI233
aI243
aI3
aI313
aI1649
aI6955
aI13679
aI32327
asS'd'
I3053
sa(dp6106
S'a'
I10739
sS's'
I15
sS'm'
(lp6107
I1
aI1
aI3
aI11
aI29
aI9
aI1
aI79
aI247
aI677
aI685
aI3107
aI5987
aI9675
aI29523
asS'd'
I3054
sa(dp6108
S'a'
I10762
sS's'
I15
sS'm'
(lp6109
I1
aI1
aI1
aI7
aI25
aI31
aI39
aI241
aI483
aI839
aI1143
aI437
aI2317
aI2437
aI173
asS'd'
I3055
sa(dp6110
S'a'
I10772
sS's'
I15
sS'm'
(lp6111
I1
aI1
aI5
aI1
aI17
aI19
aI83
aI57
aI39
aI479
aI715
aI1911
aI1091
aI10937
aI22145
asS'd'
I3056
sa(dp6112
S'a'
I10781
sS's'
I15
sS'm'
(lp6113
I1
aI1
aI7
aI1
aI27
aI45
aI35
aI55
aI477
aI719
aI217
aI3349
aI7717
aI6853
aI9699
asS'd'
I3057
sa(dp6114
S'a'
I10800
sS's'
I15
sS'm'
(lp6115
I1
aI3
aI1
aI11
aI9
aI39
aI25
aI223
aI303
aI713
aI151
aI2611
aI4629
aI5855
aI31729
asS'd'
I3058
sa(dp6116
S'a'
I10805
sS's'
I15
sS'm'
(lp6117
I1
aI1
aI1
aI11
aI13
aI35
aI53
aI39
aI167
aI779
aI1673
aI1221
aI6281
aI15113
aI32027
asS'd'
I3059
sa(dp6118
S'a'
I10827
sS's'
I15
sS'm'
(lp6119
I1
aI1
aI5
aI9
aI19
aI63
aI89
aI113
aI199
aI107
aI1015
aI835
aI2879
aI9499
aI25597
asS'd'
I3060
sa(dp6120
S'a'
I10830
sS's'
I15
sS'm'
(lp6121
I1
aI1
aI7
aI3
aI19
aI37
aI15
aI23
aI449
aI641
aI1811
aI3407
aI6775
aI6283
aI31157
asS'd'
I3061
sa(dp6122
S'a'
I10837
sS's'
I15
sS'm'
(lp6123
I1
aI1
aI3
aI1
aI19
aI15
aI31
aI99
aI511
aI897
aI1693
aI2093
aI955
aI15897
aI26693
asS'd'
I3062
sa(dp6124
S'a'
I10841
sS's'
I15
sS'm'
(lp6125
I1
aI1
aI5
aI1
aI5
aI15
aI47
aI19
aI441
aI541
aI1621
aI3877
aI6407
aI15991
aI1931
asS'd'
I3063
sa(dp6126
S'a'
I10847
sS's'
I15
sS'm'
(lp6127
I1
aI3
aI5
aI9
aI21
aI61
aI15
aI77
aI265
aI351
aI879
aI3835
aI6555
aI2349
aI23235
asS'd'
I3064
sa(dp6128
S'a'
I10848
sS's'
I15
sS'm'
(lp6129
I1
aI1
aI5
aI11
aI25
aI37
aI29
aI181
aI341
aI641
aI1213
aI1319
aI6359
aI6231
aI32573
asS'd'
I3065
sa(dp6130
S'a'
I10857
sS's'
I15
sS'm'
(lp6131
I1
aI1
aI1
aI7
aI1
aI37
aI87
aI123
aI33
aI913
aI111
aI2613
aI4895
aI12595
aI26633
asS'd'
I3066
sa(dp6132
S'a'
I10866
sS's'
I15
sS'm'
(lp6133
I1
aI3
aI5
aI3
aI27
aI11
aI45
aI89
aI183
aI241
aI1355
aI783
aI3343
aI239
aI8643
asS'd'
I3067
sa(dp6134
S'a'
I10868
sS's'
I15
sS'm'
(lp6135
I1
aI3
aI7
aI7
aI9
aI35
aI67
aI187
aI233
aI577
aI1445
aI3063
aI6039
aI16233
aI1453
asS'd'
I3068
sa(dp6136
S'a'
I10872
sS's'
I15
sS'm'
(lp6137
I1
aI1
aI3
aI13
aI27
aI11
aI23
aI15
aI95
aI63
aI1931
aI911
aI8149
aI6833
aI3051
asS'd'
I3069
sa(dp6138
S'a'
I10887
sS's'
I15
sS'm'
(lp6139
I1
aI3
aI3
aI5
aI29
aI49
aI125
aI117
aI47
aI143
aI423
aI3215
aI3605
aI3677
aI17155
asS'd'
I3070
sa(dp6140
S'a'
I10899
sS's'
I15
sS'm'
(lp6141
I1
aI3
aI1
aI1
aI31
aI1
aI123
aI195
aI83
aI893
aI1947
aI339
aI2927
aI7183
aI15443
asS'd'
I3071
sa(dp6142
S'a'
I10901
sS's'
I15
sS'm'
(lp6143
I1
aI1
aI7
aI13
aI31
aI15
aI91
aI207
aI39
aI275
aI439
aI2617
aI3093
aI11041
aI24997
asS'd'
I3072
sa(dp6144
S'a'
I10915
sS's'
I15
sS'm'
(lp6145
I1
aI1
aI5
aI3
aI3
aI41
aI13
aI67
aI361
aI497
aI25
aI3807
aI3551
aI9681
aI21043
asS'd'
I3073
sa(dp6146
S'a'
I10924
sS's'
I15
sS'm'
(lp6147
I1
aI3
aI3
aI3
aI11
aI27
aI103
aI59
aI427
aI327
aI1705
aI29
aI8127
aI1641
aI20847
asS'd'
I3074
sa(dp6148
S'a'
I10929
sS's'
I15
sS'm'
(lp6149
I1
aI3
aI7
aI5
aI3
aI37
aI81
aI137
aI225
aI101
aI187
aI3067
aI2491
aI12687
aI16227
asS'd'
I3075
sa(dp6150
S'a'
I10942
sS's'
I15
sS'm'
(lp6151
I1
aI3
aI5
aI15
aI15
aI33
aI69
aI223
aI225
aI771
aI1917
aI2293
aI2889
aI12083
aI23971
asS'd'
I3076
sa(dp6152
S'a'
I10971
sS's'
I15
sS'm'
(lp6153
I1
aI1
aI3
aI5
aI11
aI9
aI121
aI81
aI203
aI425
aI1189
aI2011
aI3041
aI3247
aI739
asS'd'
I3077
sa(dp6154
S'a'
I10992
sS's'
I15
sS'm'
(lp6155
I1
aI3
aI1
aI1
aI13
aI9
aI39
aI169
aI437
aI571
aI1481
aI3355
aI3895
aI8975
aI31031
asS'd'
I3078
sa(dp6156
S'a'
I10995
sS's'
I15
sS'm'
(lp6157
I1
aI3
aI1
aI11
aI1
aI43
aI35
aI35
aI293
aI11
aI657
aI1415
aI5021
aI14463
aI17945
asS'd'
I3079
sa(dp6158
S'a'
I11002
sS's'
I15
sS'm'
(lp6159
I1
aI1
aI5
aI5
aI13
aI47
aI91
aI15
aI159
aI23
aI971
aI3575
aI757
aI13477
aI31757
asS'd'
I3080
sa(dp6160
S'a'
I11010
sS's'
I15
sS'm'
(lp6161
I1
aI1
aI7
aI1
aI5
aI63
aI69
aI27
aI71
aI129
aI123
aI3767
aI89
aI7865
aI1189
asS'd'
I3081
sa(dp6162
S'a'
I11027
sS's'
I15
sS'm'
(lp6163
I1
aI3
aI3
aI5
aI23
aI1
aI83
aI3
aI487
aI491
aI217
aI2583
aI3889
aI15009
aI9227
asS'd'
I3082
sa(dp6164
S'a'
I11029
sS's'
I15
sS'm'
(lp6165
I1
aI3
aI5
aI15
aI25
aI1
aI73
aI107
aI245
aI191
aI1449
aI571
aI1403
aI6953
aI17457
asS'd'
I3083
sa(dp6166
S'a'
I11045
sS's'
I15
sS'm'
(lp6167
I1
aI3
aI3
aI3
aI27
aI19
aI25
aI105
aI207
aI857
aI1161
aI3657
aI2107
aI7955
aI517
asS'd'
I3084
sa(dp6168
S'a'
I11057
sS's'
I15
sS'm'
(lp6169
I1
aI3
aI3
aI9
aI21
aI29
aI5
aI103
aI219
aI35
aI3
aI1635
aI4815
aI15797
aI29839
asS'd'
I3085
sa(dp6170
S'a'
I11070
sS's'
I15
sS'm'
(lp6171
I1
aI1
aI7
aI7
aI3
aI63
aI75
aI77
aI13
aI57
aI603
aI2333
aI7761
aI14397
aI10875
asS'd'
I3086
sa(dp6172
S'a'
I11092
sS's'
I15
sS'm'
(lp6173
I1
aI3
aI7
aI13
aI3
aI11
aI5
aI255
aI1
aI947
aI1695
aI1927
aI7447
aI7407
aI20797
asS'd'
I3087
sa(dp6174
S'a'
I11099
sS's'
I15
sS'm'
(lp6175
I1
aI1
aI5
aI1
aI1
aI21
aI105
aI73
aI429
aI973
aI1801
aI3943
aI6161
aI1309
aI3359
asS'd'
I3088
sa(dp6176
S'a'
I11106
sS's'
I15
sS'm'
(lp6177
I1
aI1
aI3
aI15
aI27
aI9
aI9
aI129
aI117
aI545
aI9
aI1983
aI6351
aI10925
aI27337
asS'd'
I3089
sa(dp6178
S'a'
I11115
sS's'
I15
sS'm'
(lp6179
I1
aI3
aI3
aI5
aI5
aI5
aI13
aI155
aI503
aI875
aI1243
aI2259
aI3445
aI11953
aI6517
asS'd'
I3090
sa(dp6180
S'a'
I11120
sS's'
I15
sS'm'
(lp6181
I1
aI1
aI7
aI3
aI29
aI21
aI121
aI147
aI413
aI423
aI1887
aI2429
aI2765
aI16335
aI3071
asS'd'
I3091
sa(dp6182
S'a'
I11126
sS's'
I15
sS'm'
(lp6183
I1
aI1
aI7
aI9
aI5
aI53
aI41
aI137
aI463
aI583
aI1627
aI1731
aI6675
aI3703
aI8177
asS'd'
I3092
sa(dp6184
S'a'
I11153
sS's'
I15
sS'm'
(lp6185
I1
aI3
aI5
aI11
aI31
aI29
aI67
aI159
aI425
aI25
aI1457
aI139
aI5019
aI701
aI7357
asS'd'
I3093
sa(dp6186
S'a'
I11190
sS's'
I15
sS'm'
(lp6187
I1
aI3
aI1
aI5
aI25
aI15
aI123
aI123
aI245
aI859
aI249
aI2175
aI2137
aI5765
aI4873
asS'd'
I3094
sa(dp6188
S'a'
I11199
sS's'
I15
sS'm'
(lp6189
I1
aI1
aI3
aI5
aI23
aI1
aI111
aI111
aI111
aI469
aI1473
aI1777
aI3579
aI13503
aI2569
asS'd'
I3095
sa(dp6190
S'a'
I11222
sS's'
I15
sS'm'
(lp6191
I1
aI1
aI7
aI3
aI17
aI23
aI51
aI23
aI499
aI135
aI713
aI3317
aI807
aI9589
aI11349
asS'd'
I3096
sa(dp6192
S'a'
I11225
sS's'
I15
sS'm'
(lp6193
I1
aI1
aI1
aI15
aI9
aI51
aI75
aI159
aI359
aI773
aI1521
aI2913
aI5901
aI3047
aI14649
asS'd'
I3097
sa(dp6194
S'a'
I11226
sS's'
I15
sS'm'
(lp6195
I1
aI1
aI3
aI1
aI13
aI61
aI117
aI195
aI49
aI267
aI57
aI1769
aI3621
aI9415
aI29443
asS'd'
I3098
sa(dp6196
S'a'
I11231
sS's'
I15
sS'm'
(lp6197
I1
aI3
aI7
aI11
aI3
aI25
aI33
aI31
aI315
aI191
aI359
aI3399
aI2481
aI13831
aI20205
asS'd'
I3099
sa(dp6198
S'a'
I11244
sS's'
I15
sS'm'
(lp6199
I1
aI3
aI3
aI5
aI31
aI43
aI35
aI125
aI291
aI51
aI1469
aI3857
aI1707
aI2641
aI32137
asS'd'
I3100
sa(dp6200
S'a'
I11259
sS's'
I15
sS'm'
(lp6201
I1
aI3
aI5
aI1
aI25
aI11
aI113
aI137
aI211
aI159
aI1667
aI939
aI6439
aI5337
aI32059
asS'd'
I3101
sa(dp6202
S'a'
I11261
sS's'
I15
sS'm'
(lp6203
I1
aI3
aI3
aI11
aI31
aI61
aI99
aI49
aI383
aI343
aI395
aI51
aI6931
aI16039
aI5901
asS'd'
I3102
sa(dp6204
S'a'
I11270
sS's'
I15
sS'm'
(lp6205
I1
aI1
aI3
aI5
aI9
aI63
aI63
aI49
aI405
aI915
aI1505
aI2141
aI6749
aI7799
aI17313
asS'd'
I3103
sa(dp6206
S'a'
I11273
sS's'
I15
sS'm'
(lp6207
I1
aI3
aI7
aI11
aI15
aI11
aI49
aI161
aI155
aI869
aI121
aI301
aI6561
aI4279
aI15233
asS'd'
I3104
sa(dp6208
S'a'
I11300
sS's'
I15
sS'm'
(lp6209
I1
aI1
aI5
aI13
aI19
aI13
aI103
aI59
aI503
aI293
aI701
aI2527
aI5327
aI13927
aI5025
asS'd'
I3105
sa(dp6210
S'a'
I11307
sS's'
I15
sS'm'
(lp6211
I1
aI1
aI7
aI1
aI1
aI37
aI55
aI155
aI485
aI399
aI855
aI2677
aI5927
aI9657
aI2795
asS'd'
I3106
sa(dp6212
S'a'
I11318
sS's'
I15
sS'm'
(lp6213
I1
aI1
aI1
aI5
aI19
aI15
aI121
aI69
aI385
aI75
aI1567
aI2649
aI5601
aI12981
aI15903
asS'd'
I3107
sa(dp6214
S'a'
I11332
sS's'
I15
sS'm'
(lp6215
I1
aI1
aI1
aI11
aI19
aI21
aI45
aI59
aI505
aI737
aI15
aI1383
aI1177
aI8375
aI15557
asS'd'
I3108
sa(dp6216
S'a'
I11335
sS's'
I15
sS'm'
(lp6217
I1
aI1
aI7
aI13
aI29
aI19
aI123
aI127
aI469
aI773
aI733
aI3289
aI8087
aI5803
aI27897
asS'd'
I3109
sa(dp6218
S'a'
I11341
sS's'
I15
sS'm'
(lp6219
I1
aI3
aI3
aI11
aI19
aI55
aI101
aI67
aI485
aI939
aI607
aI1521
aI6161
aI12235
aI16499
asS'd'
I3110
sa(dp6220
S'a'
I11347
sS's'
I15
sS'm'
(lp6221
I1
aI3
aI5
aI13
aI29
aI31
aI31
aI9
aI453
aI151
aI1055
aI3873
aI405
aI12877
aI29829
asS'd'
I3111
sa(dp6222
S'a'
I11354
sS's'
I15
sS'm'
(lp6223
I1
aI3
aI5
aI1
aI17
aI1
aI17
aI131
aI107
aI1003
aI1749
aI1849
aI6207
aI2153
aI21275
asS'd'
I3112
sa(dp6224
S'a'
I11360
sS's'
I15
sS'm'
(lp6225
I1
aI3
aI7
aI15
aI7
aI25
aI51
aI143
aI51
aI517
aI1841
aI1771
aI5389
aI4633
aI11123
asS'd'
I3113
sa(dp6226
S'a'
I11369
sS's'
I15
sS'm'
(lp6227
I1
aI3
aI7
aI11
aI23
aI7
aI89
aI95
aI403
aI361
aI835
aI585
aI2783
aI8091
aI5141
asS'd'
I3114
sa(dp6228
S'a'
I11372
sS's'
I15
sS'm'
(lp6229
I1
aI3
aI1
aI9
aI1
aI53
aI115
aI11
aI493
aI587
aI305
aI3605
aI1711
aI4169
aI20013
asS'd'
I3115
sa(dp6230
S'a'
I11378
sS's'
I15
sS'm'
(lp6231
I1
aI3
aI7
aI3
aI17
aI59
aI55
aI251
aI49
aI759
aI1227
aI3685
aI7765
aI1445
aI20385
asS'd'
I3116
sa(dp6232
S'a'
I11396
sS's'
I15
sS'm'
(lp6233
I1
aI1
aI5
aI7
aI29
aI55
aI19
aI157
aI129
aI927
aI893
aI1235
aI1955
aI8153
aI2865
asS'd'
I3117
sa(dp6234
S'a'
I11405
sS's'
I15
sS'm'
(lp6235
I1
aI3
aI1
aI15
aI21
aI35
aI81
aI53
aI175
aI939
aI1635
aI3597
aI747
aI14011
aI12867
asS'd'
I3118
sa(dp6236
S'a'
I11417
sS's'
I15
sS'm'
(lp6237
I1
aI3
aI7
aI1
aI27
aI61
aI91
aI73
aI405
aI677
aI603
aI2715
aI7099
aI941
aI24523
asS'd'
I3119
sa(dp6238
S'a'
I11424
sS's'
I15
sS'm'
(lp6239
I1
aI3
aI5
aI9
aI13
aI45
aI35
aI167
aI57
aI483
aI735
aI2777
aI7847
aI6257
aI13109
asS'd'
I3120
sa(dp6240
S'a'
I11427
sS's'
I15
sS'm'
(lp6241
I1
aI3
aI5
aI7
aI1
aI3
aI97
aI13
aI159
aI533
aI1791
aI1061
aI981
aI10795
aI26165
asS'd'
I3121
sa(dp6242
S'a'
I11430
sS's'
I15
sS'm'
(lp6243
I1
aI1
aI5
aI13
aI27
aI5
aI125
aI25
aI251
aI221
aI1909
aI197
aI6987
aI11537
aI15287
asS'd'
I3122
sa(dp6244
S'a'
I11439
sS's'
I15
sS'm'
(lp6245
I1
aI3
aI5
aI5
aI27
aI15
aI1
aI131
aI375
aI923
aI81
aI3153
aI6071
aI2515
aI23729
asS'd'
I3123
sa(dp6246
S'a'
I11442
sS's'
I15
sS'm'
(lp6247
I1
aI3
aI3
aI9
aI9
aI23
aI71
aI13
aI465
aI261
aI937
aI1549
aI5993
aI11325
aI15065
asS'd'
I3124
sa(dp6248
S'a'
I11448
sS's'
I15
sS'm'
(lp6249
I1
aI3
aI1
aI3
aI7
aI63
aI17
aI129
aI435
aI23
aI215
aI2251
aI1561
aI9703
aI26483
asS'd'
I3125
sa(dp6250
S'a'
I11461
sS's'
I15
sS'm'
(lp6251
I1
aI1
aI3
aI1
aI5
aI53
aI77
aI109
aI9
aI605
aI371
aI2081
aI6023
aI7145
aI15837
asS'd'
I3126
sa(dp6252
S'a'
I11468
sS's'
I15
sS'm'
(lp6253
I1
aI3
aI7
aI11
aI27
aI39
aI115
aI47
aI259
aI337
aI1857
aI3465
aI1549
aI7747
aI8525
asS'd'
I3127
sa(dp6254
S'a'
I11471
sS's'
I15
sS'm'
(lp6255
I1
aI3
aI7
aI7
aI29
aI29
aI75
aI77
aI29
aI661
aI899
aI3137
aI2661
aI15271
aI28093
asS'd'
I3128
sa(dp6256
S'a'
I11473
sS's'
I15
sS'm'
(lp6257
I1
aI1
aI1
aI3
aI3
aI3
aI11
aI219
aI39
aI757
aI1465
aI249
aI7445
aI7013
aI15187
asS'd'
I3129
sa(dp6258
S'a'
I11476
sS's'
I15
sS'm'
(lp6259
I1
aI3
aI7
aI15
aI15
aI1
aI39
aI245
aI427
aI1003
aI1493
aI1913
aI6435
aI14787
aI13481
asS'd'
I3130
sa(dp6260
S'a'
I11480
sS's'
I15
sS'm'
(lp6261
I1
aI1
aI7
aI3
aI3
aI37
aI5
aI97
aI343
aI833
aI1379
aI1551
aI5403
aI1843
aI5877
asS'd'
I3131
sa(dp6262
S'a'
I11489
sS's'
I15
sS'm'
(lp6263
I1
aI3
aI1
aI1
aI3
aI17
aI17
aI163
aI339
aI691
aI1707
aI1845
aI5941
aI4259
aI24531
asS'd'
I3132
sa(dp6264
S'a'
I11499
sS's'
I15
sS'm'
(lp6265
I1
aI1
aI1
aI1
aI27
aI21
aI85
aI221
aI71
aI949
aI1753
aI391
aI6349
aI15901
aI20811
asS'd'
I3133
sa(dp6266
S'a'
I11516
sS's'
I15
sS'm'
(lp6267
I1
aI1
aI1
aI5
aI31
aI19
aI45
aI99
aI469
aI783
aI1747
aI3807
aI5889
aI9485
aI13715
asS'd'
I3134
sa(dp6268
S'a'
I11522
sS's'
I15
sS'm'
(lp6269
I1
aI3
aI1
aI9
aI23
aI21
aI39
aI25
aI421
aI713
aI461
aI2857
aI5023
aI5341
aI6409
asS'd'
I3135
sa(dp6270
S'a'
I11531
sS's'
I15
sS'm'
(lp6271
I1
aI3
aI7
aI5
aI25
aI19
aI59
aI147
aI387
aI857
aI375
aI3103
aI1261
aI13697
aI25675
asS'd'
I3136
sa(dp6272
S'a'
I11539
sS's'
I15
sS'm'
(lp6273
I1
aI3
aI5
aI5
aI31
aI21
aI49
aI251
aI463
aI441
aI473
aI3487
aI3915
aI11151
aI17721
asS'd'
I3137
sa(dp6274
S'a'
I11546
sS's'
I15
sS'm'
(lp6275
I1
aI1
aI3
aI9
aI15
aI47
aI81
aI219
aI143
aI141
aI81
aI1705
aI5847
aI3437
aI30521
asS'd'
I3138
sa(dp6276
S'a'
I11551
sS's'
I15
sS'm'
(lp6277
I1
aI1
aI7
aI3
aI25
aI19
aI97
aI41
aI77
aI105
aI1337
aI695
aI7589
aI8587
aI7509
asS'd'
I3139
sa(dp6278
S'a'
I11564
sS's'
I15
sS'm'
(lp6279
I1
aI1
aI5
aI13
aI3
aI11
aI61
aI19
aI139
aI667
aI963
aI1567
aI5715
aI7079
aI15967
asS'd'
I3140
sa(dp6280
S'a'
I11582
sS's'
I15
sS'm'
(lp6281
I1
aI1
aI5
aI5
aI5
aI29
aI67
aI57
aI477
aI173
aI1163
aI727
aI823
aI15635
aI17705
asS'd'
I3141
sa(dp6282
S'a'
I11589
sS's'
I15
sS'm'
(lp6283
I1
aI3
aI7
aI11
aI13
aI39
aI57
aI193
aI73
aI617
aI535
aI1623
aI4581
aI4451
aI2589
asS'd'
I3142
sa(dp6284
S'a'
I11593
sS's'
I15
sS'm'
(lp6285
I1
aI1
aI5
aI5
aI9
aI27
aI75
aI127
aI325
aI413
aI1669
aI1749
aI8045
aI16199
aI12237
asS'd'
I3143
sa(dp6286
S'a'
I11601
sS's'
I15
sS'm'
(lp6287
I1
aI1
aI3
aI1
aI17
aI23
aI27
aI189
aI319
aI953
aI347
aI909
aI4401
aI12791
aI25077
asS'd'
I3144
sa(dp6288
S'a'
I11608
sS's'
I15
sS'm'
(lp6289
I1
aI1
aI3
aI3
aI17
aI51
aI37
aI79
aI301
aI607
aI885
aI1169
aI3275
aI3327
aI20013
asS'd'
I3145
sa(dp6290
S'a'
I11617
sS's'
I15
sS'm'
(lp6291
I1
aI3
aI5
aI3
aI21
aI9
aI99
aI213
aI387
aI889
aI575
aI3591
aI5377
aI2981
aI23989
asS'd'
I3146
sa(dp6292
S'a'
I11630
sS's'
I15
sS'm'
(lp6293
I1
aI3
aI3
aI13
aI11
aI7
aI23
aI255
aI279
aI853
aI453
aI2377
aI8123
aI15393
aI9669
asS'd'
I3147
sa(dp6294
S'a'
I11663
sS's'
I15
sS'm'
(lp6295
I1
aI3
aI1
aI7
aI11
aI9
aI109
aI35
aI405
aI449
aI1967
aI2943
aI3485
aI5031
aI14273
asS'd'
I3148
sa(dp6296
S'a'
I11666
sS's'
I15
sS'm'
(lp6297
I1
aI3
aI3
aI1
aI25
aI27
aI43
aI115
aI435
aI675
aI1937
aI1477
aI4831
aI9417
aI7017
asS'd'
I3149
sa(dp6298
S'a'
I11668
sS's'
I15
sS'm'
(lp6299
I1
aI1
aI7
aI13
aI19
aI45
aI83
aI241
aI487
aI215
aI1453
aI209
aI4061
aI1765
aI15623
asS'd'
I3150
sa(dp6300
S'a'
I11677
sS's'
I15
sS'm'
(lp6301
I1
aI1
aI7
aI7
aI21
aI31
aI95
aI9
aI287
aI1005
aI1933
aI3405
aI6913
aI7733
aI18975
asS'd'
I3151
sa(dp6302
S'a'
I11682
sS's'
I15
sS'm'
(lp6303
I1
aI1
aI1
aI11
aI13
aI11
aI25
aI39
aI283
aI57
aI255
aI2809
aI5767
aI6127
aI6705
asS'd'
I3152
sa(dp6304
S'a'
I11687
sS's'
I15
sS'm'
(lp6305
I1
aI3
aI1
aI11
aI1
aI51
aI73
aI181
aI261
aI215
aI385
aI2777
aI5169
aI12431
aI23563
asS'd'
I3153
sa(dp6306
S'a'
I11696
sS's'
I15
sS'm'
(lp6307
I1
aI3
aI3
aI9
aI9
aI39
aI123
aI197
aI501
aI679
aI109
aI3369
aI4817
aI8855
aI7997
asS'd'
I3154
sa(dp6308
S'a'
I11713
sS's'
I15
sS'm'
(lp6309
I1
aI1
aI5
aI1
aI29
aI61
aI15
aI183
aI453
aI999
aI1211
aI3217
aI8035
aI5153
aI19975
asS'd'
I3155
sa(dp6310
S'a'
I11728
sS's'
I15
sS'm'
(lp6311
I1
aI3
aI7
aI11
aI11
aI21
aI51
aI45
aI379
aI793
aI289
aI309
aI1229
aI7159
aI581
asS'd'
I3156
sa(dp6312
S'a'
I11747
sS's'
I15
sS'm'
(lp6313
I1
aI1
aI3
aI9
aI17
aI11
aI75
aI67
aI289
aI191
aI1083
aI2949
aI6063
aI6611
aI21595
asS'd'
I3157
sa(dp6314
S'a'
I11750
sS's'
I15
sS'm'
(lp6315
I1
aI3
aI7
aI3
aI27
aI45
aI59
aI193
aI485
aI277
aI27
aI1219
aI2389
aI15875
aI6273
asS'd'
I3158
sa(dp6316
S'a'
I11754
sS's'
I15
sS'm'
(lp6317
I1
aI1
aI5
aI3
aI31
aI29
aI65
aI197
aI115
aI163
aI9
aI783
aI5573
aI2833
aI12603
asS'd'
I3159
sa(dp6318
S'a'
I11759
sS's'
I15
sS'm'
(lp6319
I1
aI1
aI3
aI7
aI5
aI53
aI115
aI181
aI175
aI749
aI1335
aI1151
aI2131
aI12467
aI15811
asS'd'
I3160
sa(dp6320
S'a'
I11761
sS's'
I15
sS'm'
(lp6321
I1
aI1
aI1
aI9
aI27
aI39
aI11
aI1
aI443
aI677
aI777
aI1857
aI7459
aI3177
aI3875
asS'd'
I3161
sa(dp6322
S'a'
I11764
sS's'
I15
sS'm'
(lp6323
I1
aI1
aI7
aI7
aI17
aI3
aI23
aI161
aI105
aI603
aI1991
aI3845
aI465
aI11467
aI2077
asS'd'
I3162
sa(dp6324
S'a'
I11767
sS's'
I15
sS'm'
(lp6325
I1
aI1
aI3
aI13
aI5
aI23
aI39
aI35
aI399
aI795
aI265
aI207
aI1823
aI15069
aI31839
asS'd'
I3163
sa(dp6326
S'a'
I11797
sS's'
I15
sS'm'
(lp6327
I1
aI1
aI1
aI1
aI29
aI61
aI89
aI193
aI41
aI99
aI315
aI1021
aI6109
aI12507
aI19973
asS'd'
I3164
sa(dp6328
S'a'
I11804
sS's'
I15
sS'm'
(lp6329
I1
aI1
aI5
aI3
aI13
aI57
aI119
aI251
aI215
aI695
aI1521
aI4081
aI2481
aI657
aI855
asS'd'
I3165
sa(dp6330
S'a'
I11808
sS's'
I15
sS'm'
(lp6331
I1
aI1
aI7
aI3
aI25
aI5
aI3
aI133
aI111
aI385
aI773
aI1027
aI4327
aI3031
aI3537
asS'd'
I3166
sa(dp6332
S'a'
I11831
sS's'
I15
sS'm'
(lp6333
I1
aI3
aI7
aI5
aI5
aI27
aI43
aI117
aI479
aI83
aI1421
aI2791
aI6551
aI6231
aI10353
asS'd'
I3167
sa(dp6334
S'a'
I11832
sS's'
I15
sS'm'
(lp6335
I1
aI1
aI1
aI13
aI3
aI29
aI35
aI71
aI85
aI821
aI1671
aI3057
aI797
aI13683
aI7025
asS'd'
I3168
sa(dp6336
S'a'
I11849
sS's'
I15
sS'm'
(lp6337
I1
aI3
aI7
aI1
aI1
aI47
aI117
aI233
aI141
aI993
aI1381
aI2551
aI1031
aI11765
aI18429
asS'd'
I3169
sa(dp6338
S'a'
I11855
sS's'
I15
sS'm'
(lp6339
I1
aI3
aI1
aI3
aI13
aI3
aI77
aI29
aI35
aI807
aI1109
aI695
aI5605
aI5477
aI449
asS'd'
I3170
sa(dp6340
S'a'
I11863
sS's'
I15
sS'm'
(lp6341
I1
aI1
aI1
aI15
aI21
aI37
aI117
aI105
aI273
aI311
aI1287
aI1415
aI1221
aI1847
aI19487
asS'd'
I3171
sa(dp6342
S'a'
I11880
sS's'
I15
sS'm'
(lp6343
I1
aI3
aI1
aI11
aI21
aI61
aI107
aI225
aI335
aI501
aI1995
aI2399
aI5475
aI12613
aI18877
asS'd'
I3172
sa(dp6344
S'a'
I11883
sS's'
I15
sS'm'
(lp6345
I1
aI3
aI3
aI1
aI31
aI41
aI27
aI205
aI103
aI837
aI639
aI2007
aI2759
aI12471
aI1457
asS'd'
I3173
sa(dp6346
S'a'
I11885
sS's'
I15
sS'm'
(lp6347
I1
aI1
aI7
aI13
aI29
aI39
aI71
aI245
aI105
aI235
aI1277
aI1515
aI6129
aI15947
aI26653
asS'd'
I3174
sa(dp6348
S'a'
I11898
sS's'
I15
sS'm'
(lp6349
I1
aI1
aI7
aI5
aI7
aI13
aI87
aI251
aI315
aI1017
aI587
aI2917
aI5911
aI2919
aI29715
asS'd'
I3175
sa(dp6350
S'a'
I11916
sS's'
I15
sS'm'
(lp6351
I1
aI1
aI1
aI3
aI7
aI19
aI81
aI243
aI177
aI917
aI2023
aI2365
aI7465
aI4901
aI29841
asS'd'
I3176
sa(dp6352
S'a'
I11924
sS's'
I15
sS'm'
(lp6353
I1
aI3
aI5
aI15
aI1
aI31
aI15
aI147
aI285
aI1003
aI1757
aI47
aI6925
aI1197
aI19633
asS'd'
I3177
sa(dp6354
S'a'
I11928
sS's'
I15
sS'm'
(lp6355
I1
aI1
aI5
aI7
aI27
aI25
aI47
aI209
aI85
aI403
aI1399
aI2331
aI3663
aI595
aI13407
asS'd'
I3178
sa(dp6356
S'a'
I11947
sS's'
I15
sS'm'
(lp6357
I1
aI3
aI5
aI9
aI7
aI25
aI7
aI139
aI389
aI817
aI1153
aI1421
aI5735
aI9577
aI10269
asS'd'
I3179
sa(dp6358
S'a'
I11955
sS's'
I15
sS'm'
(lp6359
I1
aI1
aI1
aI9
aI5
aI61
aI49
aI117
aI389
aI541
aI433
aI1405
aI2575
aI223
aI7265
asS'd'
I3180
sa(dp6360
S'a'
I11961
sS's'
I15
sS'm'
(lp6361
I1
aI1
aI5
aI7
aI15
aI1
aI81
aI207
aI435
aI843
aI835
aI3797
aI7637
aI5333
aI31115
asS'd'
I3181
sa(dp6362
S'a'
I11962
sS's'
I15
sS'm'
(lp6363
I1
aI3
aI7
aI11
aI13
aI3
aI47
aI249
aI301
aI715
aI2015
aI3049
aI8155
aI10989
aI26051
asS'd'
I3182
sa(dp6364
S'a'
I11982
sS's'
I15
sS'm'
(lp6365
I1
aI1
aI7
aI7
aI3
aI33
aI119
aI113
aI381
aI575
aI367
aI41
aI3317
aI11727
aI4351
asS'd'
I3183
sa(dp6366
S'a'
I11990
sS's'
I15
sS'm'
(lp6367
I1
aI3
aI3
aI13
aI9
aI3
aI51
aI37
aI173
aI137
aI533
aI1827
aI631
aI10047
aI6267
asS'd'
I3184
sa(dp6368
S'a'
I12010
sS's'
I15
sS'm'
(lp6369
I1
aI3
aI3
aI11
aI17
aI39
aI61
aI245
aI13
aI139
aI1281
aI1319
aI1233
aI13629
aI32269
asS'd'
I3185
sa(dp6370
S'a'
I12018
sS's'
I15
sS'm'
(lp6371
I1
aI1
aI1
aI7
aI15
aI17
aI91
aI109
aI163
aI609
aI11
aI3251
aI7653
aI14035
aI31755
asS'd'
I3186
sa(dp6372
S'a'
I12027
sS's'
I15
sS'm'
(lp6373
I1
aI3
aI3
aI15
aI13
aI21
aI55
aI231
aI385
aI133
aI1833
aI2637
aI6935
aI14303
aI26745
asS'd'
I3187
sa(dp6374
S'a'
I12029
sS's'
I15
sS'm'
(lp6375
I1
aI1
aI1
aI7
aI17
aI41
aI125
aI141
aI89
aI823
aI1411
aI3637
aI6211
aI13323
aI6111
asS'd'
I3188
sa(dp6376
S'a'
I12035
sS's'
I15
sS'm'
(lp6377
I1
aI1
aI1
aI11
aI1
aI21
aI9
aI43
aI97
aI685
aI1223
aI1491
aI121
aI1793
aI2397
asS'd'
I3189
sa(dp6378
S'a'
I12055
sS's'
I15
sS'm'
(lp6379
I1
aI3
aI5
aI5
aI17
aI17
aI5
aI223
aI129
aI865
aI1839
aI1137
aI6391
aI4377
aI9233
asS'd'
I3190
sa(dp6380
S'a'
I12062
sS's'
I15
sS'm'
(lp6381
I1
aI3
aI7
aI15
aI21
aI55
aI5
aI77
aI341
aI637
aI1853
aI1435
aI1195
aI9283
aI21257
asS'd'
I3191
sa(dp6382
S'a'
I12068
sS's'
I15
sS'm'
(lp6383
I1
aI3
aI5
aI1
aI9
aI49
aI43
aI211
aI127
aI655
aI1397
aI1235
aI5279
aI2351
aI24229
asS'd'
I3192
sa(dp6384
S'a'
I12071
sS's'
I15
sS'm'
(lp6385
I1
aI3
aI5
aI3
aI25
aI29
aI13
aI229
aI25
aI675
aI837
aI2753
aI2125
aI9863
aI11293
asS'd'
I3193
sa(dp6386
S'a'
I12072
sS's'
I15
sS'm'
(lp6387
I1
aI3
aI5
aI7
aI23
aI43
aI127
aI1
aI163
aI237
aI337
aI3019
aI7747
aI16227
aI2881
asS'd'
I3194
sa(dp6388
S'a'
I12086
sS's'
I15
sS'm'
(lp6389
I1
aI3
aI5
aI5
aI25
aI9
aI43
aI171
aI421
aI521
aI1885
aI337
aI7873
aI6347
aI13181
asS'd'
I3195
sa(dp6390
S'a'
I12097
sS's'
I15
sS'm'
(lp6391
I1
aI3
aI5
aI5
aI7
aI47
aI107
aI173
aI163
aI191
aI625
aI3389
aI2833
aI7945
aI24787
asS'd'
I3196
sa(dp6392
S'a'
I12098
sS's'
I15
sS'm'
(lp6393
I1
aI1
aI7
aI3
aI27
aI57
aI27
aI209
aI253
aI815
aI301
aI1633
aI3945
aI5051
aI28851
asS'd'
I3197
sa(dp6394
S'a'
I12100
sS's'
I15
sS'm'
(lp6395
I1
aI3
aI7
aI9
aI9
aI51
aI103
aI213
aI437
aI189
aI1857
aI1331
aI5551
aI10641
aI27405
asS'd'
I3198
sa(dp6396
S'a'
I12112
sS's'
I15
sS'm'
(lp6397
I1
aI1
aI5
aI5
aI15
aI1
aI25
aI105
aI117
aI347
aI161
aI3369
aI3589
aI12903
aI23559
asS'd'
I3199
sa(dp6398
S'a'
I12118
sS's'
I15
sS'm'
(lp6399
I1
aI1
aI1
aI5
aI3
aI61
aI93
aI51
aI81
aI281
aI1383
aI745
aI4137
aI2005
aI3635
asS'd'
I3200
sa(dp6400
S'a'
I12133
sS's'
I15
sS'm'
(lp6401
I1
aI3
aI7
aI5
aI13
aI57
aI111
aI211
aI303
aI477
aI359
aI4019
aI6779
aI5129
aI22035
asS'd'
I3201
sa(dp6402
S'a'
I12134
sS's'
I15
sS'm'
(lp6403
I1
aI1
aI1
aI7
aI17
aI29
aI113
aI113
aI201
aI531
aI749
aI739
aI2879
aI3315
aI18733
asS'd'
I3202
sa(dp6404
S'a'
I12137
sS's'
I15
sS'm'
(lp6405
I1
aI3
aI7
aI13
aI21
aI55
aI21
aI183
aI359
aI75
aI377
aI2211
aI4281
aI14317
aI28307
asS'd'
I3203
sa(dp6406
S'a'
I12161
sS's'
I15
sS'm'
(lp6407
I1
aI3
aI7
aI1
aI21
aI1
aI49
aI213
aI317
aI75
aI113
aI1741
aI7963
aI12785
aI11571
asS'd'
I3204
sa(dp6408
S'a'
I12162
sS's'
I15
sS'm'
(lp6409
I1
aI3
aI7
aI9
aI11
aI31
aI29
aI101
aI261
aI141
aI715
aI2727
aI8187
aI2075
aI32433
asS'd'
I3205
sa(dp6410
S'a'
I12171
sS's'
I15
sS'm'
(lp6411
I1
aI3
aI7
aI3
aI23
aI9
aI17
aI143
aI385
aI211
aI593
aI241
aI6567
aI10777
aI6677
asS'd'
I3206
sa(dp6412
S'a'
I12174
sS's'
I15
sS'm'
(lp6413
I1
aI1
aI3
aI15
aI3
aI17
aI117
aI99
aI91
aI793
aI989
aI2421
aI5643
aI16103
aI9759
asS'd'
I3207
sa(dp6414
S'a'
I12185
sS's'
I15
sS'm'
(lp6415
I1
aI3
aI7
aI11
aI23
aI43
aI107
aI35
aI421
aI431
aI743
aI853
aI7939
aI12169
aI4199
asS'd'
I3208
sa(dp6416
S'a'
I12204
sS's'
I15
sS'm'
(lp6417
I1
aI1
aI1
aI11
aI21
aI53
aI17
aI203
aI123
aI395
aI59
aI929
aI255
aI7585
aI10945
asS'd'
I3209
sa(dp6418
S'a'
I12212
sS's'
I15
sS'm'
(lp6419
I1
aI3
aI3
aI15
aI17
aI57
aI57
aI133
aI67
aI71
aI1685
aI903
aI4079
aI15423
aI26495
asS'd'
I3210
sa(dp6420
S'a'
I12215
sS's'
I15
sS'm'
(lp6421
I1
aI1
aI1
aI15
aI3
aI47
aI95
aI39
aI405
aI407
aI1557
aI3001
aI6225
aI15187
aI5663
asS'd'
I3211
sa(dp6422
S'a'
I12216
sS's'
I15
sS'm'
(lp6423
I1
aI3
aI5
aI5
aI13
aI47
aI33
aI61
aI375
aI1023
aI1981
aI2773
aI2375
aI11321
aI17731
asS'd'
I3212
sa(dp6424
S'a'
I12253
sS's'
I15
sS'm'
(lp6425
I1
aI3
aI5
aI9
aI17
aI59
aI117
aI95
aI493
aI149
aI1269
aI2865
aI369
aI2109
aI24601
asS'd'
I3213
sa(dp6426
S'a'
I12260
sS's'
I15
sS'm'
(lp6427
I1
aI3
aI5
aI13
aI17
aI63
aI67
aI247
aI95
aI721
aI67
aI305
aI6179
aI15399
aI32559
asS'd'
I3214
sa(dp6428
S'a'
I12277
sS's'
I15
sS'm'
(lp6429
I1
aI1
aI5
aI1
aI3
aI21
aI41
aI15
aI453
aI475
aI2017
aI3193
aI5903
aI897
aI4237
asS'd'
I3215
sa(dp6430
S'a'
I12289
sS's'
I15
sS'm'
(lp6431
I1
aI1
aI5
aI3
aI15
aI41
aI1
aI141
aI441
aI575
aI155
aI3791
aI7711
aI11231
aI24611
asS'd'
I3216
sa(dp6432
S'a'
I12295
sS's'
I15
sS'm'
(lp6433
I1
aI3
aI7
aI1
aI17
aI53
aI27
aI169
aI31
aI437
aI963
aI1793
aI7777
aI1917
aI29311
asS'd'
I3217
sa(dp6434
S'a'
I12314
sS's'
I15
sS'm'
(lp6435
I1
aI3
aI3
aI13
aI9
aI27
aI77
aI87
aI329
aI885
aI749
aI1713
aI6013
aI6921
aI629
asS'd'
I3218
sa(dp6436
S'a'
I12323
sS's'
I15
sS'm'
(lp6437
I1
aI3
aI5
aI13
aI3
aI7
aI53
aI27
aI353
aI267
aI925
aI2141
aI439
aI15175
aI30851
asS'd'
I3219
sa(dp6438
S'a'
I12325
sS's'
I15
sS'm'
(lp6439
I1
aI3
aI3
aI13
aI17
aI57
aI35
aI101
aI265
aI901
aI1825
aI2159
aI6149
aI5967
aI24023
asS'd'
I3220
sa(dp6440
S'a'
I12335
sS's'
I15
sS'm'
(lp6441
I1
aI1
aI5
aI11
aI13
aI51
aI99
aI111
aI193
aI415
aI1541
aI2929
aI5045
aI3147
aI12587
asS'd'
I3221
sa(dp6442
S'a'
I12349
sS's'
I15
sS'm'
(lp6443
I1
aI3
aI7
aI11
aI15
aI9
aI33
aI17
aI511
aI815
aI299
aI1077
aI6171
aI10451
aI15039
asS'd'
I3222
sa(dp6444
S'a'
I12358
sS's'
I15
sS'm'
(lp6445
I1
aI1
aI1
aI15
aI25
aI63
aI51
aI137
aI449
aI951
aI1051
aI1101
aI4725
aI2463
aI7355
asS'd'
I3223
sa(dp6446
S'a'
I12372
sS's'
I15
sS'm'
(lp6447
I1
aI1
aI1
aI7
aI27
aI63
aI29
aI179
aI317
aI521
aI1459
aI827
aI6599
aI13459
aI15439
asS'd'
I3224
sa(dp6448
S'a'
I12376
sS's'
I15
sS'm'
(lp6449
I1
aI3
aI3
aI15
aI17
aI31
aI37
aI191
aI229
aI245
aI181
aI941
aI5761
aI1849
aI31599
asS'd'
I3225
sa(dp6450
S'a'
I12379
sS's'
I15
sS'm'
(lp6451
I1
aI1
aI1
aI9
aI27
aI45
aI67
aI239
aI481
aI615
aI1667
aI3751
aI8141
aI10013
aI2125
asS'd'
I3226
sa(dp6452
S'a'
I12386
sS's'
I15
sS'm'
(lp6453
I1
aI1
aI1
aI1
aI13
aI51
aI117
aI135
aI73
aI151
aI1291
aI2541
aI1411
aI3767
aI26949
asS'd'
I3227
sa(dp6454
S'a'
I12395
sS's'
I15
sS'm'
(lp6455
I1
aI3
aI1
aI9
aI7
aI11
aI21
aI187
aI243
aI857
aI1951
aI865
aI7273
aI2041
aI8155
asS'd'
I3228
sa(dp6456
S'a'
I12416
sS's'
I15
sS'm'
(lp6457
I1
aI1
aI3
aI3
aI19
aI33
aI89
aI115
aI455
aI137
aI707
aI1867
aI4221
aI2433
aI9119
asS'd'
I3229
sa(dp6458
S'a'
I12421
sS's'
I15
sS'm'
(lp6459
I1
aI1
aI3
aI11
aI5
aI3
aI121
aI1
aI71
aI951
aI603
aI3873
aI723
aI3285
aI19289
asS'd'
I3230
sa(dp6460
S'a'
I12440
sS's'
I15
sS'm'
(lp6461
I1
aI3
aI7
aI15
aI21
aI1
aI117
aI17
aI455
aI519
aI731
aI3003
aI5741
aI9557
aI29163
asS'd'
I3231
sa(dp6462
S'a'
I12452
sS's'
I15
sS'm'
(lp6463
I1
aI1
aI3
aI13
aI25
aI5
aI43
aI147
aI209
aI895
aI255
aI1231
aI241
aI487
aI15593
asS'd'
I3232
sa(dp6464
S'a'
I12455
sS's'
I15
sS'm'
(lp6465
I1
aI1
aI3
aI13
aI7
aI1
aI89
aI187
aI217
aI927
aI2029
aI3521
aI2777
aI8103
aI22819
asS'd'
I3233
sa(dp6466
S'a'
I12456
sS's'
I15
sS'm'
(lp6467
I1
aI1
aI7
aI11
aI7
aI33
aI3
aI73
aI5
aI489
aI227
aI2259
aI7031
aI6425
aI26135
asS'd'
I3234
sa(dp6468
S'a'
I12462
sS's'
I15
sS'm'
(lp6469
I1
aI3
aI3
aI7
aI31
aI19
aI97
aI201
aI455
aI819
aI945
aI2771
aI8083
aI8711
aI2835
asS'd'
I3235
sa(dp6470
S'a'
I12467
sS's'
I15
sS'm'
(lp6471
I1
aI1
aI1
aI5
aI15
aI45
aI43
aI157
aI245
aI967
aI877
aI2289
aI4499
aI9891
aI18827
asS'd'
I3236
sa(dp6472
S'a'
I12479
sS's'
I15
sS'm'
(lp6473
I1
aI3
aI1
aI7
aI21
aI59
aI123
aI63
aI231
aI485
aI1781
aI1211
aI4597
aI5269
aI1607
asS'd'
I3237
sa(dp6474
S'a'
I12505
sS's'
I15
sS'm'
(lp6475
I1
aI1
aI1
aI13
aI23
aI39
aI105
aI55
aI249
aI991
aI1625
aI3089
aI3825
aI4275
aI29139
asS'd'
I3238
sa(dp6476
S'a'
I12521
sS's'
I15
sS'm'
(lp6477
I1
aI3
aI3
aI1
aI29
aI29
aI55
aI169
aI13
aI895
aI1355
aI1101
aI6063
aI12935
aI23215
asS'd'
I3239
sa(dp6478
S'a'
I12535
sS's'
I15
sS'm'
(lp6479
I1
aI1
aI5
aI5
aI31
aI49
aI99
aI137
aI209
aI1017
aI1179
aI3931
aI637
aI14131
aI19285
asS'd'
I3240
sa(dp6480
S'a'
I12547
sS's'
I15
sS'm'
(lp6481
I1
aI1
aI1
aI1
aI3
aI11
aI119
aI11
aI215
aI337
aI243
aI3883
aI3807
aI7335
aI11901
asS'd'
I3241
sa(dp6482
S'a'
I12556
sS's'
I15
sS'm'
(lp6483
I1
aI3
aI7
aI3
aI7
aI27
aI71
aI225
aI219
aI367
aI1213
aI2739
aI1185
aI10175
aI21313
asS'd'
I3242
sa(dp6484
S'a'
I12561
sS's'
I15
sS'm'
(lp6485
I1
aI3
aI7
aI13
aI7
aI49
aI23
aI223
aI61
aI1011
aI797
aI1335
aI6711
aI5961
aI5605
asS'd'
I3243
sa(dp6486
S'a'
I12568
sS's'
I15
sS'm'
(lp6487
I1
aI3
aI3
aI11
aI19
aI37
aI1
aI149
aI39
aI661
aI929
aI2125
aI2299
aI5181
aI28083
asS'd'
I3244
sa(dp6488
S'a'
I12578
sS's'
I15
sS'm'
(lp6489
I1
aI3
aI3
aI13
aI13
aI9
aI67
aI21
aI463
aI279
aI529
aI523
aI6705
aI11011
aI31695
asS'd'
I3245
sa(dp6490
S'a'
I12583
sS's'
I15
sS'm'
(lp6491
I1
aI3
aI1
aI5
aI13
aI1
aI123
aI3
aI291
aI625
aI1949
aI2713
aI5917
aI10343
aI13627
asS'd'
I3246
sa(dp6492
S'a'
I12595
sS's'
I15
sS'm'
(lp6493
I1
aI1
aI3
aI9
aI27
aI41
aI3
aI207
aI103
aI265
aI811
aI549
aI6109
aI313
aI8889
asS'd'
I3247
sa(dp6494
S'a'
I12604
sS's'
I15
sS'm'
(lp6495
I1
aI3
aI3
aI13
aI23
aI43
aI99
aI33
aI279
aI463
aI955
aI793
aI4113
aI10615
aI16957
asS'd'
I3248
sa(dp6496
S'a'
I12610
sS's'
I15
sS'm'
(lp6497
I1
aI1
aI5
aI7
aI11
aI49
aI79
aI45
aI17
aI937
aI359
aI1037
aI1099
aI3003
aI31561
asS'd'
I3249
sa(dp6498
S'a'
I12621
sS's'
I15
sS'm'
(lp6499
I1
aI1
aI1
aI7
aI3
aI45
aI111
aI35
aI109
aI983
aI53
aI4057
aI7349
aI3599
aI2209
asS'd'
I3250
sa(dp6500
S'a'
I12622
sS's'
I15
sS'm'
(lp6501
I1
aI3
aI7
aI11
aI9
aI43
aI27
aI9
aI85
aI529
aI1497
aI347
aI759
aI12449
aI11373
asS'd'
I3251
sa(dp6502
S'a'
I12624
sS's'
I15
sS'm'
(lp6503
I1
aI1
aI3
aI9
aI17
aI1
aI49
aI31
aI367
aI813
aI1385
aI2025
aI773
aI4679
aI4543
asS'd'
I3252
sa(dp6504
S'a'
I12629
sS's'
I15
sS'm'
(lp6505
I1
aI1
aI5
aI15
aI15
aI9
aI43
aI97
aI239
aI995
aI1037
aI841
aI4167
aI12113
aI23765
asS'd'
I3253
sa(dp6506
S'a'
I12630
sS's'
I15
sS'm'
(lp6507
I1
aI3
aI5
aI9
aI29
aI53
aI123
aI49
aI221
aI113
aI1157
aI73
aI6087
aI1363
aI11029
asS'd'
I3254
sa(dp6508
S'a'
I12639
sS's'
I15
sS'm'
(lp6509
I1
aI3
aI1
aI13
aI3
aI15
aI69
aI199
aI279
aI919
aI5
aI161
aI4817
aI15031
aI121
asS'd'
I3255
sa(dp6510
S'a'
I12640
sS's'
I15
sS'm'
(lp6511
I1
aI3
aI1
aI9
aI3
aI31
aI117
aI77
aI393
aI241
aI645
aI3181
aI1067
aI15879
aI2037
asS'd'
I3256
sa(dp6512
S'a'
I12650
sS's'
I15
sS'm'
(lp6513
I1
aI3
aI3
aI15
aI3
aI63
aI57
aI33
aI117
aI789
aI941
aI1301
aI5865
aI12693
aI3523
asS'd'
I3257
sa(dp6514
S'a'
I12679
sS's'
I15
sS'm'
(lp6515
I1
aI1
aI5
aI13
aI3
aI61
aI51
aI151
aI175
aI305
aI95
aI1557
aI6567
aI7841
aI13903
asS'd'
I3258
sa(dp6516
S'a'
I12680
sS's'
I15
sS'm'
(lp6517
I1
aI3
aI3
aI5
aI15
aI25
aI127
aI79
aI245
aI767
aI645
aI3933
aI1357
aI12579
aI4067
asS'd'
I3259
sa(dp6518
S'a'
I12698
sS's'
I15
sS'm'
(lp6519
I1
aI3
aI5
aI11
aI21
aI31
aI13
aI251
aI127
aI231
aI1795
aI2627
aI1191
aI3363
aI23543
asS'd'
I3260
sa(dp6520
S'a'
I12716
sS's'
I15
sS'm'
(lp6521
I1
aI1
aI3
aI5
aI7
aI49
aI121
aI57
aI131
aI481
aI1879
aI525
aI5225
aI337
aI1957
asS'd'
I3261
sa(dp6522
S'a'
I12721
sS's'
I15
sS'm'
(lp6523
I1
aI1
aI5
aI13
aI9
aI55
aI27
aI37
aI211
aI125
aI119
aI3373
aI251
aI12357
aI13975
asS'd'
I3262
sa(dp6524
S'a'
I12722
sS's'
I15
sS'm'
(lp6525
I1
aI3
aI3
aI15
aI1
aI51
aI91
aI119
aI233
aI993
aI203
aI1635
aI1167
aI6327
aI29119
asS'd'
I3263
sa(dp6526
S'a'
I12731
sS's'
I15
sS'm'
(lp6527
I1
aI1
aI7
aI1
aI13
aI5
aI23
aI253
aI121
aI989
aI1105
aI3321
aI3221
aI6073
aI21185
asS'd'
I3264
sa(dp6528
S'a'
I12742
sS's'
I15
sS'm'
(lp6529
I1
aI1
aI3
aI15
aI13
aI49
aI121
aI247
aI247
aI133
aI485
aI1067
aI7875
aI411
aI7647
asS'd'
I3265
sa(dp6530
S'a'
I12745
sS's'
I15
sS'm'
(lp6531
I1
aI3
aI7
aI13
aI31
aI37
aI127
aI241
aI145
aI133
aI53
aI267
aI2029
aI3703
aI16123
asS'd'
I3266
sa(dp6532
S'a'
I12751
sS's'
I15
sS'm'
(lp6533
I1
aI3
aI1
aI15
aI15
aI9
aI15
aI89
aI35
aI367
aI401
aI61
aI1953
aI7873
aI17861
asS'd'
I3267
sa(dp6534
S'a'
I12759
sS's'
I15
sS'm'
(lp6535
I1
aI1
aI1
aI1
aI1
aI41
aI71
aI249
aI213
aI779
aI1385
aI1767
aI999
aI15151
aI16647
asS'd'
I3268
sa(dp6536
S'a'
I12763
sS's'
I15
sS'm'
(lp6537
I1
aI3
aI7
aI13
aI31
aI23
aI123
aI235
aI343
aI949
aI309
aI3777
aI3587
aI853
aI19779
asS'd'
I3269
sa(dp6538
S'a'
I12769
sS's'
I15
sS'm'
(lp6539
I1
aI1
aI3
aI13
aI29
aI35
aI5
aI37
aI63
aI757
aI303
aI1579
aI3443
aI243
aI11873
asS'd'
I3270
sa(dp6540
S'a'
I12781
sS's'
I15
sS'm'
(lp6541
I1
aI3
aI1
aI9
aI19
aI49
aI81
aI53
aI11
aI901
aI1857
aI147
aI3103
aI14019
aI21
asS'd'
I3271
sa(dp6542
S'a'
I12793
sS's'
I15
sS'm'
(lp6543
I1
aI3
aI7
aI13
aI3
aI39
aI99
aI99
aI45
aI91
aI1567
aI551
aI3129
aI4809
aI29057
asS'd'
I3272
sa(dp6544
S'a'
I12799
sS's'
I15
sS'm'
(lp6545
I1
aI3
aI7
aI3
aI3
aI27
aI17
aI231
aI377
aI381
aI1479
aI2525
aI2595
aI2799
aI25737
asS'd'
I3273
sa(dp6546
S'a'
I12815
sS's'
I15
sS'm'
(lp6547
I1
aI3
aI5
aI15
aI15
aI25
aI103
aI215
aI301
aI59
aI1417
aI981
aI7579
aI12731
aI22329
asS'd'
I3274
sa(dp6548
S'a'
I12824
sS's'
I15
sS'm'
(lp6549
I1
aI1
aI1
aI13
aI5
aI31
aI61
aI31
aI349
aI925
aI1301
aI685
aI435
aI11567
aI10715
asS'd'
I3275
sa(dp6550
S'a'
I12836
sS's'
I15
sS'm'
(lp6551
I1
aI1
aI7
aI9
aI19
aI57
aI109
aI1
aI37
aI377
aI1015
aI2273
aI6741
aI3191
aI15949
asS'd'
I3276
sa(dp6552
S'a'
I12845
sS's'
I15
sS'm'
(lp6553
I1
aI3
aI3
aI13
aI3
aI23
aI103
aI127
aI11
aI59
aI1847
aI1175
aI425
aI3423
aI20643
asS'd'
I3277
sa(dp6554
S'a'
I12853
sS's'
I15
sS'm'
(lp6555
I1
aI3
aI3
aI7
aI3
aI11
aI105
aI141
aI55
aI217
aI1427
aI477
aI667
aI9403
aI11905
asS'd'
I3278
sa(dp6556
S'a'
I12854
sS's'
I15
sS'm'
(lp6557
I1
aI3
aI3
aI5
aI3
aI27
aI11
aI187
aI495
aI907
aI1925
aI445
aI6639
aI8159
aI15225
asS'd'
I3279
sa(dp6558
S'a'
I12857
sS's'
I15
sS'm'
(lp6559
I1
aI3
aI1
aI5
aI27
aI31
aI77
aI213
aI73
aI343
aI1123
aI3609
aI2431
aI15329
aI32165
asS'd'
I3280
sa(dp6560
S'a'
I12866
sS's'
I15
sS'm'
(lp6561
I1
aI1
aI7
aI5
aI1
aI11
aI105
aI139
aI485
aI1007
aI709
aI3509
aI5231
aI11717
aI31433
asS'd'
I3281
sa(dp6562
S'a'
I12872
sS's'
I15
sS'm'
(lp6563
I1
aI1
aI3
aI15
aI23
aI49
aI95
aI169
aI399
aI1019
aI19
aI2013
aI5311
aI7951
aI22609
asS'd'
I3282
sa(dp6564
S'a'
I12875
sS's'
I15
sS'm'
(lp6565
I1
aI3
aI1
aI7
aI13
aI3
aI29
aI203
aI209
aI701
aI1791
aI2615
aI5351
aI4237
aI12565
asS'd'
I3283
sa(dp6566
S'a'
I12878
sS's'
I15
sS'm'
(lp6567
I1
aI3
aI1
aI15
aI27
aI11
aI91
aI31
aI205
aI205
aI1683
aI901
aI5129
aI6049
aI11865
asS'd'
I3284
sa(dp6568
S'a'
I12880
sS's'
I15
sS'm'
(lp6569
I1
aI1
aI7
aI7
aI27
aI59
aI21
aI3
aI209
aI79
aI769
aI4013
aI2041
aI2645
aI11561
asS'd'
I3285
sa(dp6570
S'a'
I12885
sS's'
I15
sS'm'
(lp6571
I1
aI3
aI7
aI11
aI5
aI45
aI39
aI243
aI185
aI871
aI795
aI1845
aI8043
aI6285
aI20991
asS'd'
I3286
sa(dp6572
S'a'
I12901
sS's'
I15
sS'm'
(lp6573
I1
aI1
aI5
aI7
aI13
aI7
aI15
aI165
aI349
aI179
aI789
aI1269
aI3787
aI5429
aI26567
asS'd'
I3287
sa(dp6574
S'a'
I12902
sS's'
I15
sS'm'
(lp6575
I1
aI3
aI3
aI13
aI31
aI23
aI75
aI41
aI177
aI735
aI1889
aI4039
aI3519
aI15003
aI965
asS'd'
I3288
sa(dp6576
S'a'
I12920
sS's'
I15
sS'm'
(lp6577
I1
aI3
aI1
aI7
aI15
aI57
aI15
aI139
aI27
aI469
aI1003
aI691
aI7893
aI9643
aI30983
asS'd'
I3289
sa(dp6578
S'a'
I12926
sS's'
I15
sS'm'
(lp6579
I1
aI3
aI1
aI13
aI23
aI27
aI3
aI237
aI233
aI875
aI99
aI883
aI6167
aI5463
aI6245
asS'd'
I3290
sa(dp6580
S'a'
I12929
sS's'
I15
sS'm'
(lp6581
I1
aI1
aI5
aI13
aI25
aI57
aI79
aI51
aI147
aI619
aI1147
aI279
aI6583
aI1939
aI477
asS'd'
I3291
sa(dp6582
S'a'
I12939
sS's'
I15
sS'm'
(lp6583
I1
aI3
aI5
aI5
aI31
aI61
aI125
aI163
aI213
aI699
aI583
aI3865
aI615
aI9707
aI11651
asS'd'
I3292
sa(dp6584
S'a'
I12941
sS's'
I15
sS'm'
(lp6585
I1
aI1
aI5
aI1
aI5
aI21
aI93
aI239
aI31
aI641
aI777
aI27
aI5247
aI8993
aI21053
asS'd'
I3293
sa(dp6586
S'a'
I12950
sS's'
I15
sS'm'
(lp6587
I1
aI3
aI7
aI9
aI1
aI13
aI61
aI57
aI503
aI453
aI83
aI3271
aI2845
aI1121
aI18639
asS'd'
I3294
sa(dp6588
S'a'
I12953
sS's'
I15
sS'm'
(lp6589
I1
aI1
aI7
aI5
aI29
aI53
aI13
aI219
aI379
aI441
aI821
aI3179
aI4877
aI2535
aI7557
asS'd'
I3295
sa(dp6590
S'a'
I12992
sS's'
I15
sS'm'
(lp6591
I1
aI1
aI7
aI13
aI9
aI53
aI17
aI183
aI265
aI393
aI587
aI2753
aI6453
aI7135
aI24737
asS'd'
I3296
sa(dp6592
S'a'
I13002
sS's'
I15
sS'm'
(lp6593
I1
aI1
aI1
aI13
aI11
aI23
aI73
aI109
aI393
aI1013
aI559
aI755
aI7291
aI6631
aI26509
asS'd'
I3297
sa(dp6594
S'a'
I13010
sS's'
I15
sS'm'
(lp6595
I1
aI3
aI1
aI5
aI5
aI15
aI107
aI103
aI355
aI307
aI1559
aI837
aI5413
aI5285
aI17489
asS'd'
I3298
sa(dp6596
S'a'
I13058
sS's'
I15
sS'm'
(lp6597
I1
aI1
aI5
aI7
aI17
aI21
aI21
aI23
aI109
aI709
aI1947
aI3585
aI3629
aI4669
aI949
asS'd'
I3299
sa(dp6598
S'a'
I13072
sS's'
I15
sS'm'
(lp6599
I1
aI3
aI7
aI1
aI9
aI33
aI85
aI147
aI467
aI259
aI1913
aI199
aI7399
aI9551
aI22387
asS'd'
I3300
sa(dp6600
S'a'
I13084
sS's'
I15
sS'm'
(lp6601
I1
aI3
aI5
aI11
aI15
aI53
aI23
aI41
aI249
aI515
aI1161
aI2467
aI1299
aI7449
aI2613
asS'd'
I3301
sa(dp6602
S'a'
I13087
sS's'
I15
sS'm'
(lp6603
I1
aI1
aI5
aI5
aI5
aI29
aI91
aI139
aI487
aI545
aI321
aI3035
aI4545
aI6747
aI21673
asS'd'
I3302
sa(dp6604
S'a'
I13091
sS's'
I15
sS'm'
(lp6605
I1
aI1
aI3
aI13
aI23
aI49
aI95
aI103
aI25
aI119
aI469
aI2515
aI2551
aI841
aI25089
asS'd'
I3303
sa(dp6606
S'a'
I13097
sS's'
I15
sS'm'
(lp6607
I1
aI1
aI5
aI7
aI11
aI31
aI31
aI197
aI245
aI715
aI257
aI4043
aI8099
aI11531
aI5617
asS'd'
I3304
sa(dp6608
S'a'
I13108
sS's'
I15
sS'm'
(lp6609
I1
aI1
aI3
aI3
aI19
aI7
aI9
aI179
aI103
aI995
aI191
aI179
aI3843
aI5215
aI27639
asS'd'
I3305
sa(dp6610
S'a'
I13123
sS's'
I15
sS'm'
(lp6611
I1
aI3
aI1
aI7
aI23
aI59
aI25
aI65
aI399
aI211
aI1453
aI3511
aI7203
aI16015
aI32197
asS'd'
I3306
sa(dp6612
S'a'
I13149
sS's'
I15
sS'm'
(lp6613
I1
aI3
aI3
aI5
aI9
aI35
aI109
aI67
aI197
aI449
aI643
aI519
aI5751
aI15551
aI11331
asS'd'
I3307
sa(dp6614
S'a'
I13150
sS's'
I15
sS'm'
(lp6615
I1
aI3
aI5
aI3
aI1
aI17
aI53
aI201
aI265
aI351
aI467
aI911
aI1117
aI7183
aI20371
asS'd'
I3308
sa(dp6616
S'a'
I13163
sS's'
I15
sS'm'
(lp6617
I1
aI1
aI7
aI7
aI27
aI17
aI93
aI81
aI227
aI619
aI1191
aI735
aI815
aI12615
aI2719
asS'd'
I3309
sa(dp6618
S'a'
I13166
sS's'
I15
sS'm'
(lp6619
I1
aI3
aI1
aI15
aI19
aI3
aI83
aI75
aI343
aI297
aI1019
aI3469
aI4383
aI13299
aI29755
asS'd'
I3310
sa(dp6620
S'a'
I13178
sS's'
I15
sS'm'
(lp6621
I1
aI1
aI5
aI3
aI13
aI55
aI119
aI169
aI85
aI595
aI299
aI2469
aI5625
aI2877
aI16117
asS'd'
I3311
sa(dp6622
S'a'
I13180
sS's'
I15
sS'm'
(lp6623
I1
aI1
aI3
aI5
aI15
aI17
aI61
aI161
aI47
aI393
aI143
aI867
aI5517
aI9495
aI12795
asS'd'
I3312
sa(dp6624
S'a'
I13184
sS's'
I15
sS'm'
(lp6625
I1
aI3
aI5
aI1
aI27
aI31
aI113
aI125
aI251
aI687
aI969
aI1473
aI2245
aI6355
aI13655
asS'd'
I3313
sa(dp6626
S'a'
I13204
sS's'
I15
sS'm'
(lp6627
I1
aI1
aI1
aI5
aI5
aI37
aI29
aI133
aI443
aI899
aI277
aI2353
aI7223
aI4459
aI19159
asS'd'
I3314
sa(dp6628
S'a'
I13238
sS's'
I15
sS'm'
(lp6629
I1
aI1
aI3
aI9
aI19
aI27
aI53
aI145
aI195
aI167
aI2045
aI447
aI1803
aI1895
aI8431
asS'd'
I3315
sa(dp6630
S'a'
I13242
sS's'
I15
sS'm'
(lp6631
I1
aI1
aI3
aI9
aI5
aI27
aI91
aI147
aI233
aI451
aI475
aI27
aI4629
aI16181
aI16437
asS'd'
I3316
sa(dp6632
S'a'
I13249
sS's'
I15
sS'm'
(lp6633
I1
aI3
aI5
aI3
aI29
aI17
aI53
aI167
aI433
aI689
aI1131
aI2985
aI1553
aI11697
aI6993
asS'd'
I3317
sa(dp6634
S'a'
I13255
sS's'
I15
sS'm'
(lp6635
I1
aI3
aI3
aI13
aI21
aI43
aI69
aI229
aI399
aI525
aI179
aI237
aI7017
aI5703
aI17653
asS'd'
I3318
sa(dp6636
S'a'
I13269
sS's'
I15
sS'm'
(lp6637
I1
aI1
aI3
aI15
aI13
aI39
aI75
aI163
aI229
aI875
aI197
aI477
aI3667
aI15501
aI15801
asS'd'
I3319
sa(dp6638
S'a'
I13270
sS's'
I15
sS'm'
(lp6639
I1
aI1
aI7
aI15
aI15
aI51
aI81
aI187
aI487
aI673
aI865
aI1915
aI1009
aI5935
aI8097
asS'd'
I3320
sa(dp6640
S'a'
I13274
sS's'
I15
sS'm'
(lp6641
I1
aI3
aI5
aI5
aI7
aI3
aI63
aI77
aI495
aI815
aI391
aI2321
aI1007
aI15661
aI30715
asS'd'
I3321
sa(dp6642
S'a'
I13285
sS's'
I15
sS'm'
(lp6643
I1
aI1
aI7
aI3
aI17
aI25
aI83
aI173
aI173
aI911
aI1373
aI2957
aI4549
aI15977
aI17695
asS'd'
I3322
sa(dp6644
S'a'
I13289
sS's'
I15
sS'm'
(lp6645
I1
aI1
aI7
aI13
aI13
aI23
aI77
aI147
aI497
aI1003
aI1687
aI1795
aI1393
aI1881
aI8479
asS'd'
I3323
sa(dp6646
S'a'
I13298
sS's'
I15
sS'm'
(lp6647
I1
aI3
aI7
aI11
aI27
aI43
aI97
aI25
aI61
aI457
aI203
aI2573
aI5943
aI15021
aI4003
asS'd'
I3324
sa(dp6648
S'a'
I13307
sS's'
I15
sS'm'
(lp6649
I1
aI3
aI3
aI13
aI9
aI37
aI37
aI25
aI219
aI889
aI1535
aI2791
aI4531
aI13679
aI12663
asS'd'
I3325
sa(dp6650
S'a'
I13312
sS's'
I15
sS'm'
(lp6651
I1
aI1
aI3
aI1
aI17
aI7
aI51
aI123
aI89
aI887
aI1467
aI1645
aI3767
aI6383
aI30837
asS'd'
I3326
sa(dp6652
S'a'
I13335
sS's'
I15
sS'm'
(lp6653
I1
aI3
aI3
aI1
aI21
aI47
aI5
aI151
aI83
aI257
aI569
aI2711
aI637
aI12569
aI16893
asS'd'
I3327
sa(dp6654
S'a'
I13345
sS's'
I15
sS'm'
(lp6655
I1
aI3
aI7
aI1
aI31
aI37
aI73
aI3
aI115
aI919
aI1817
aI2483
aI4811
aI15245
aI4375
asS'd'
I3328
sa(dp6656
S'a'
I13357
sS's'
I15
sS'm'
(lp6657
I1
aI1
aI1
aI5
aI1
aI39
aI39
aI231
aI9
aI733
aI455
aI3383
aI4777
aI7235
aI12631
asS'd'
I3329
sa(dp6658
S'a'
I13366
sS's'
I15
sS'm'
(lp6659
I1
aI1
aI7
aI9
aI13
aI25
aI55
aI25
aI73
aI59
aI1699
aI929
aI755
aI1279
aI5583
asS'd'
I3330
sa(dp6660
S'a'
I13372
sS's'
I15
sS'm'
(lp6661
I1
aI3
aI5
aI3
aI9
aI49
aI79
aI55
aI479
aI179
aI1159
aI4079
aI3503
aI11603
aI12361
asS'd'
I3331
sa(dp6662
S'a'
I13380
sS's'
I15
sS'm'
(lp6663
I1
aI1
aI5
aI9
aI21
aI45
aI31
aI163
aI377
aI817
aI219
aI147
aI2581
aI12769
aI30783
asS'd'
I3332
sa(dp6664
S'a'
I13384
sS's'
I15
sS'm'
(lp6665
I1
aI3
aI1
aI7
aI15
aI27
aI39
aI189
aI493
aI259
aI1663
aI1213
aI961
aI11089
aI16079
asS'd'
I3333
sa(dp6666
S'a'
I13395
sS's'
I15
sS'm'
(lp6667
I1
aI1
aI5
aI9
aI5
aI41
aI13
aI153
aI313
aI337
aI1027
aI1267
aI4249
aI13071
aI27043
asS'd'
I3334
sa(dp6668
S'a'
I13407
sS's'
I15
sS'm'
(lp6669
I1
aI3
aI7
aI3
aI13
aI11
aI23
aI255
aI51
aI527
aI317
aI3217
aI5037
aI12723
aI17411
asS'd'
I3335
sa(dp6670
S'a'
I13408
sS's'
I15
sS'm'
(lp6671
I1
aI1
aI5
aI1
aI25
aI57
aI83
aI97
aI233
aI513
aI1283
aI2675
aI4111
aI4111
aI32141
asS'd'
I3336
sa(dp6672
S'a'
I13413
sS's'
I15
sS'm'
(lp6673
I1
aI3
aI3
aI15
aI25
aI33
aI103
aI81
aI155
aI189
aI139
aI1179
aI2691
aI15119
aI13959
asS'd'
I3337
sa(dp6674
S'a'
I13414
sS's'
I15
sS'm'
(lp6675
I1
aI3
aI3
aI1
aI25
aI55
aI67
aI19
aI19
aI9
aI437
aI579
aI4273
aI10733
aI7125
asS'd'
I3338
sa(dp6676
S'a'
I13417
sS's'
I15
sS'm'
(lp6677
I1
aI1
aI1
aI7
aI23
aI41
aI47
aI5
aI435
aI749
aI595
aI199
aI3941
aI7199
aI4795
asS'd'
I3339
sa(dp6678
S'a'
I13437
sS's'
I15
sS'm'
(lp6679
I1
aI3
aI1
aI15
aI5
aI49
aI35
aI9
aI199
aI703
aI1769
aI3269
aI5689
aI13063
aI22771
asS'd'
I3340
sa(dp6680
S'a'
I13441
sS's'
I15
sS'm'
(lp6681
I1
aI1
aI5
aI5
aI21
aI55
aI125
aI55
aI63
aI149
aI1167
aI3577
aI1051
aI3921
aI20545
asS'd'
I3341
sa(dp6682
S'a'
I13447
sS's'
I15
sS'm'
(lp6683
I1
aI3
aI7
aI13
aI29
aI53
aI107
aI193
aI163
aI339
aI1335
aI1573
aI5809
aI5681
aI29487
asS'd'
I3342
sa(dp6684
S'a'
I13456
sS's'
I15
sS'm'
(lp6685
I1
aI1
aI1
aI9
aI17
aI9
aI91
aI177
aI211
aI453
aI1807
aI1881
aI6051
aI225
aI6021
asS'd'
I3343
sa(dp6686
S'a'
I13459
sS's'
I15
sS'm'
(lp6687
I1
aI1
aI1
aI13
aI15
aI1
aI29
aI43
aI181
aI105
aI1945
aI2313
aI6429
aI2901
aI6221
asS'd'
I3344
sa(dp6688
S'a'
I13461
sS's'
I15
sS'm'
(lp6689
I1
aI3
aI5
aI11
aI29
aI55
aI115
aI115
aI187
aI1013
aI697
aI1885
aI121
aI12387
aI32443
asS'd'
I3345
sa(dp6690
S'a'
I13466
sS's'
I15
sS'm'
(lp6691
I1
aI1
aI1
aI7
aI19
aI51
aI21
aI107
aI55
aI125
aI1655
aI2281
aI3293
aI15749
aI27521
asS'd'
I3346
sa(dp6692
S'a'
I13484
sS's'
I15
sS'm'
(lp6693
I1
aI1
aI7
aI9
aI19
aI9
aI81
aI93
aI139
aI401
aI193
aI73
aI5159
aI9323
aI6019
asS'd'
I3347
sa(dp6694
S'a'
I13487
sS's'
I15
sS'm'
(lp6695
I1
aI1
aI7
aI9
aI15
aI51
aI115
aI69
aI247
aI599
aI1163
aI2251
aI1211
aI8827
aI15581
asS'd'
I3348
sa(dp6696
S'a'
I13489
sS's'
I15
sS'm'
(lp6697
I1
aI1
aI7
aI9
aI5
aI39
aI75
aI185
aI321
aI911
aI849
aI843
aI6791
aI10407
aI10513
asS'd'
I3349
sa(dp6698
S'a'
I13492
sS's'
I15
sS'm'
(lp6699
I1
aI1
aI5
aI5
aI15
aI9
aI21
aI47
aI459
aI681
aI2001
aI1545
aI5939
aI7073
aI29043
asS'd'
I3350
sa(dp6700
S'a'
I13496
sS's'
I15
sS'm'
(lp6701
I1
aI3
aI1
aI11
aI13
aI25
aI53
aI97
aI487
aI797
aI567
aI3757
aI5597
aI6313
aI18531
asS'd'
I3351
sa(dp6702
S'a'
I13510
sS's'
I15
sS'm'
(lp6703
I1
aI1
aI3
aI3
aI29
aI55
aI11
aI219
aI325
aI591
aI2015
aI383
aI2595
aI11855
aI22501
asS'd'
I3352
sa(dp6704
S'a'
I13531
sS's'
I15
sS'm'
(lp6705
I1
aI1
aI1
aI5
aI15
aI57
aI33
aI125
aI323
aI749
aI1843
aI4019
aI2075
aI6673
aI6957
asS'd'
I3353
sa(dp6706
S'a'
I13538
sS's'
I15
sS'm'
(lp6707
I1
aI1
aI5
aI7
aI19
aI7
aI47
aI239
aI51
aI107
aI1081
aI467
aI5493
aI7617
aI10355
asS'd'
I3354
sa(dp6708
S'a'
I13543
sS's'
I15
sS'm'
(lp6709
I1
aI3
aI1
aI1
aI11
aI3
aI67
aI199
aI175
aI421
aI935
aI309
aI4449
aI6363
aI9183
asS'd'
I3355
sa(dp6710
S'a'
I13547
sS's'
I15
sS'm'
(lp6711
I1
aI1
aI1
aI7
aI9
aI33
aI3
aI219
aI481
aI513
aI417
aI1267
aI2863
aI765
aI18431
asS'd'
I3356
sa(dp6712
S'a'
I13572
sS's'
I15
sS'm'
(lp6713
I1
aI3
aI1
aI1
aI19
aI1
aI89
aI109
aI415
aI105
aI487
aI3241
aI7465
aI9233
aI16307
asS'd'
I3357
sa(dp6714
S'a'
I13581
sS's'
I15
sS'm'
(lp6715
I1
aI1
aI3
aI13
aI9
aI43
aI25
aI231
aI383
aI789
aI1855
aI691
aI3465
aI2387
aI11715
asS'd'
I3358
sa(dp6716
S'a'
I13590
sS's'
I15
sS'm'
(lp6717
I1
aI3
aI3
aI3
aI13
aI39
aI63
aI107
aI33
aI265
aI437
aI117
aI3179
aI5543
aI28179
asS'd'
I3359
sa(dp6718
S'a'
I13605
sS's'
I15
sS'm'
(lp6719
I1
aI3
aI3
aI13
aI21
aI5
aI31
aI111
aI321
aI425
aI253
aI3501
aI3209
aI15429
aI18383
asS'd'
I3360
sa(dp6720
S'a'
I13612
sS's'
I15
sS'm'
(lp6721
I1
aI3
aI5
aI9
aI1
aI27
aI117
aI187
aI433
aI459
aI1999
aI1069
aI4857
aI8591
aI26343
asS'd'
I3361
sa(dp6722
S'a'
I13624
sS's'
I15
sS'm'
(lp6723
I1
aI1
aI7
aI3
aI15
aI43
aI11
aI193
aI391
aI341
aI1203
aI1259
aI7265
aI1783
aI13161
asS'd'
I3362
sa(dp6724
S'a'
I13630
sS's'
I15
sS'm'
(lp6725
I1
aI1
aI7
aI1
aI5
aI15
aI45
aI143
aI193
aI985
aI1105
aI3483
aI2421
aI9687
aI22347
asS'd'
I3363
sa(dp6726
S'a'
I13632
sS's'
I15
sS'm'
(lp6727
I1
aI3
aI7
aI13
aI21
aI17
aI79
aI231
aI487
aI663
aI1101
aI1025
aI5779
aI14681
aI29181
asS'd'
I3364
sa(dp6728
S'a'
I13638
sS's'
I15
sS'm'
(lp6729
I1
aI1
aI3
aI9
aI15
aI19
aI55
aI219
aI27
aI963
aI1513
aI1017
aI3907
aI12279
aI32655
asS'd'
I3365
sa(dp6730
S'a'
I13661
sS's'
I15
sS'm'
(lp6731
I1
aI3
aI7
aI3
aI31
aI27
aI17
aI1
aI51
aI861
aI529
aI1225
aI6395
aI15787
aI5231
asS'd'
I3366
sa(dp6732
S'a'
I13665
sS's'
I15
sS'm'
(lp6733
I1
aI3
aI3
aI11
aI27
aI7
aI101
aI143
aI21
aI191
aI1437
aI2393
aI4097
aI14319
aI6977
asS'd'
I3367
sa(dp6734
S'a'
I13668
sS's'
I15
sS'm'
(lp6735
I1
aI3
aI3
aI3
aI25
aI35
aI105
aI141
aI433
aI269
aI1469
aI2939
aI5387
aI7373
aI7863
asS'd'
I3368
sa(dp6736
S'a'
I13686
sS's'
I15
sS'm'
(lp6737
I1
aI3
aI7
aI5
aI5
aI21
aI23
aI11
aI217
aI357
aI1847
aI101
aI1161
aI5297
aI14879
asS'd'
I3369
sa(dp6738
S'a'
I13699
sS's'
I15
sS'm'
(lp6739
I1
aI3
aI1
aI3
aI25
aI23
aI81
aI217
aI505
aI161
aI1667
aI1343
aI1751
aI2463
aI26431
asS'd'
I3370
sa(dp6740
S'a'
I13701
sS's'
I15
sS'm'
(lp6741
I1
aI1
aI3
aI1
aI17
aI51
aI125
aI205
aI385
aI351
aI731
aI2911
aI2749
aI2689
aI27031
asS'd'
I3371
sa(dp6742
S'a'
I13708
sS's'
I15
sS'm'
(lp6743
I1
aI1
aI5
aI5
aI5
aI17
aI31
aI171
aI477
aI671
aI167
aI1797
aI8047
aI10123
aI4325
asS'd'
I3372
sa(dp6744
S'a'
I13716
sS's'
I15
sS'm'
(lp6745
I1
aI1
aI7
aI1
aI11
aI23
aI123
aI161
aI99
aI1007
aI765
aI1965
aI5395
aI16193
aI17751
asS'd'
I3373
sa(dp6746
S'a'
I13725
sS's'
I15
sS'm'
(lp6747
I1
aI3
aI1
aI9
aI13
aI11
aI111
aI217
aI31
aI753
aI377
aI2267
aI7893
aI7195
aI24999
asS'd'
I3374
sa(dp6748
S'a'
I13730
sS's'
I15
sS'm'
(lp6749
I1
aI3
aI1
aI9
aI21
aI53
aI127
aI121
aI151
aI395
aI1447
aI1411
aI5179
aI12043
aI27607
asS'd'
I3375
sa(dp6750
S'a'
I13742
sS's'
I15
sS'm'
(lp6751
I1
aI1
aI5
aI3
aI11
aI37
aI97
aI139
aI113
aI835
aI229
aI3741
aI827
aI5527
aI5779
asS'd'
I3376
sa(dp6752
S'a'
I13747
sS's'
I15
sS'm'
(lp6753
I1
aI1
aI7
aI7
aI27
aI55
aI11
aI55
aI429
aI269
aI1179
aI233
aI1053
aI10225
aI16703
asS'd'
I3377
sa(dp6754
S'a'
I13749
sS's'
I15
sS'm'
(lp6755
I1
aI1
aI1
aI3
aI15
aI9
aI67
aI119
aI95
aI753
aI511
aI2507
aI3953
aI6403
aI27635
asS'd'
I3378
sa(dp6756
S'a'
I13753
sS's'
I15
sS'm'
(lp6757
I1
aI3
aI3
aI7
aI27
aI57
aI25
aI27
aI249
aI515
aI193
aI4043
aI2017
aI751
aI10949
asS'd'
I3379
sa(dp6758
S'a'
I13754
sS's'
I15
sS'm'
(lp6759
I1
aI3
aI1
aI9
aI31
aI57
aI67
aI21
aI177
aI573
aI1835
aI2015
aI6201
aI2383
aI31087
asS'd'
I3380
sa(dp6760
S'a'
I13771
sS's'
I15
sS'm'
(lp6761
I1
aI1
aI5
aI1
aI19
aI3
aI89
aI243
aI69
aI387
aI1905
aI3465
aI2775
aI7713
aI30081
asS'd'
I3381
sa(dp6762
S'a'
I13773
sS's'
I15
sS'm'
(lp6763
I1
aI1
aI3
aI3
aI9
aI59
aI15
aI89
aI85
aI605
aI881
aI263
aI2551
aI797
aI16541
asS'd'
I3382
sa(dp6764
S'a'
I13782
sS's'
I15
sS'm'
(lp6765
I1
aI3
aI7
aI11
aI25
aI41
aI59
aI139
aI405
aI571
aI1147
aI2963
aI4175
aI12901
aI6309
asS'd'
I3383
sa(dp6766
S'a'
I13785
sS's'
I15
sS'm'
(lp6767
I1
aI3
aI1
aI5
aI29
aI29
aI11
aI243
aI183
aI281
aI807
aI1
aI7079
aI10079
aI13865
asS'd'
I3384
sa(dp6768
S'a'
I13798
sS's'
I15
sS'm'
(lp6769
I1
aI3
aI7
aI5
aI5
aI1
aI89
aI55
aI423
aI157
aI675
aI1849
aI241
aI6467
aI15459
asS'd'
I3385
sa(dp6770
S'a'
I13802
sS's'
I15
sS'm'
(lp6771
I1
aI1
aI7
aI11
aI15
aI63
aI89
aI109
aI501
aI549
aI317
aI3043
aI1151
aI3895
aI19851
asS'd'
I3386
sa(dp6772
S'a'
I13809
sS's'
I15
sS'm'
(lp6773
I1
aI3
aI1
aI15
aI7
aI23
aI97
aI97
aI225
aI737
aI1117
aI3325
aI209
aI14169
aI10813
asS'd'
I3387
sa(dp6774
S'a'
I13828
sS's'
I15
sS'm'
(lp6775
I1
aI3
aI7
aI13
aI13
aI39
aI91
aI153
aI395
aI879
aI1041
aI3753
aI5577
aI1985
aI25247
asS'd'
I3388
sa(dp6776
S'a'
I13832
sS's'
I15
sS'm'
(lp6777
I1
aI1
aI1
aI3
aI17
aI15
aI113
aI143
aI101
aI901
aI1119
aI1819
aI3577
aI3441
aI31511
asS'd'
I3389
sa(dp6778
S'a'
I13840
sS's'
I15
sS'm'
(lp6779
I1
aI3
aI1
aI11
aI15
aI27
aI21
aI37
aI287
aI121
aI451
aI1353
aI2173
aI299
aI18791
asS'd'
I3390
sa(dp6780
S'a'
I13850
sS's'
I15
sS'm'
(lp6781
I1
aI3
aI3
aI5
aI23
aI1
aI49
aI145
aI315
aI769
aI99
aI1385
aI5961
aI9121
aI1465
asS'd'
I3391
sa(dp6782
S'a'
I13861
sS's'
I15
sS'm'
(lp6783
I1
aI3
aI3
aI13
aI21
aI39
aI39
aI233
aI271
aI113
aI615
aI2257
aI3765
aI5921
aI313
asS'd'
I3392
sa(dp6784
S'a'
I13874
sS's'
I15
sS'm'
(lp6785
I1
aI3
aI7
aI7
aI25
aI45
aI11
aI237
aI83
aI203
aI929
aI1819
aI2679
aI11583
aI30091
asS'd'
I3393
sa(dp6786
S'a'
I13876
sS's'
I15
sS'm'
(lp6787
I1
aI1
aI1
aI7
aI21
aI63
aI85
aI251
aI133
aI991
aI1515
aI2547
aI6051
aI7279
aI3569
asS'd'
I3394
sa(dp6788
S'a'
I13886
sS's'
I15
sS'm'
(lp6789
I1
aI3
aI7
aI15
aI11
aI19
aI87
aI17
aI313
aI283
aI1021
aI2743
aI4855
aI13741
aI17955
asS'd'
I3395
sa(dp6790
S'a'
I13897
sS's'
I15
sS'm'
(lp6791
I1
aI1
aI7
aI13
aI29
aI13
aI61
aI93
aI81
aI91
aI995
aI907
aI4847
aI2599
aI20041
asS'd'
I3396
sa(dp6792
S'a'
I13900
sS's'
I15
sS'm'
(lp6793
I1
aI1
aI3
aI11
aI13
aI45
aI103
aI33
aI243
aI109
aI2029
aI121
aI231
aI16179
aI13741
asS'd'
I3397
sa(dp6794
S'a'
I13915
sS's'
I15
sS'm'
(lp6795
I1
aI3
aI5
aI9
aI9
aI5
aI73
aI225
aI199
aI723
aI611
aI1909
aI2345
aI10257
aI9909
asS'd'
I3398
sa(dp6796
S'a'
I13927
sS's'
I15
sS'm'
(lp6797
I1
aI1
aI3
aI11
aI7
aI5
aI33
aI89
aI431
aI899
aI803
aI3173
aI6131
aI16097
aI20561
asS'd'
I3399
sa(dp6798
S'a'
I13951
sS's'
I15
sS'm'
(lp6799
I1
aI3
aI3
aI7
aI7
aI47
aI23
aI47
aI411
aI69
aI239
aI661
aI5591
aI10457
aI24245
asS'd'
I3400
sa(dp6800
S'a'
I13955
sS's'
I15
sS'm'
(lp6801
I1
aI1
aI5
aI15
aI25
aI35
aI87
aI23
aI115
aI939
aI1579
aI119
aI4001
aI13791
aI9729
asS'd'
I3401
sa(dp6802
S'a'
I13962
sS's'
I15
sS'm'
(lp6803
I1
aI3
aI5
aI11
aI25
aI45
aI29
aI195
aI369
aI237
aI735
aI155
aI123
aI4415
aI32255
asS'd'
I3402
sa(dp6804
S'a'
I13969
sS's'
I15
sS'm'
(lp6805
I1
aI3
aI3
aI9
aI13
aI53
aI15
aI77
aI313
aI75
aI529
aI925
aI5679
aI14585
aI19889
asS'd'
I3403
sa(dp6806
S'a'
I13979
sS's'
I15
sS'm'
(lp6807
I1
aI1
aI7
aI15
aI15
aI27
aI105
aI13
aI31
aI669
aI563
aI1809
aI4321
aI7797
aI4177
asS'd'
I3404
sa(dp6808
S'a'
I13988
sS's'
I15
sS'm'
(lp6809
I1
aI1
aI5
aI9
aI3
aI29
aI111
aI177
aI33
aI235
aI1951
aI1561
aI141
aI4803
aI16327
asS'd'
I3405
sa(dp6810
S'a'
I13998
sS's'
I15
sS'm'
(lp6811
I1
aI1
aI1
aI7
aI9
aI41
aI1
aI149
aI95
aI933
aI115
aI1619
aI771
aI8189
aI8781
asS'd'
I3406
sa(dp6812
S'a'
I14000
sS's'
I15
sS'm'
(lp6813
I1
aI1
aI5
aI3
aI13
aI41
aI33
aI159
aI355
aI159
aI1243
aI1439
aI6571
aI14397
aI31321
asS'd'
I3407
sa(dp6814
S'a'
I14005
sS's'
I15
sS'm'
(lp6815
I1
aI1
aI7
aI11
aI9
aI15
aI91
aI145
aI457
aI255
aI1449
aI611
aI1449
aI2521
aI28949
asS'd'
I3408
sa(dp6816
S'a'
I14027
sS's'
I15
sS'm'
(lp6817
I1
aI3
aI7
aI5
aI27
aI57
aI35
aI99
aI447
aI287
aI743
aI1163
aI4379
aI7361
aI3831
asS'd'
I3409
sa(dp6818
S'a'
I14037
sS's'
I15
sS'm'
(lp6819
I1
aI3
aI3
aI7
aI15
aI53
aI41
aI83
aI133
aI571
aI1739
aI531
aI2921
aI11527
aI21941
asS'd'
I3410
sa(dp6820
S'a'
I14051
sS's'
I15
sS'm'
(lp6821
I1
aI1
aI1
aI13
aI9
aI27
aI39
aI113
aI429
aI447
aI595
aI3171
aI5245
aI4095
aI14847
asS'd'
I3411
sa(dp6822
S'a'
I14054
sS's'
I15
sS'm'
(lp6823
I1
aI1
aI3
aI7
aI19
aI19
aI21
aI101
aI489
aI1011
aI265
aI3899
aI3225
aI11701
aI5193
asS'd'
I3412
sa(dp6824
S'a'
I14060
sS's'
I15
sS'm'
(lp6825
I1
aI3
aI7
aI3
aI15
aI25
aI103
aI213
aI441
aI215
aI1483
aI263
aI3561
aI7915
aI7969
asS'd'
I3413
sa(dp6826
S'a'
I14063
sS's'
I15
sS'm'
(lp6827
I1
aI3
aI3
aI3
aI11
aI47
aI97
aI29
aI489
aI867
aI1347
aI2155
aI4871
aI8001
aI18305
asS'd'
I3414
sa(dp6828
S'a'
I14071
sS's'
I15
sS'm'
(lp6829
I1
aI3
aI1
aI9
aI25
aI15
aI61
aI17
aI343
aI775
aI1765
aI3803
aI4577
aI8437
aI12605
asS'd'
I3415
sa(dp6830
S'a'
I14078
sS's'
I15
sS'm'
(lp6831
I1
aI1
aI5
aI3
aI11
aI39
aI69
aI23
aI23
aI65
aI1967
aI2429
aI1703
aI6671
aI14981
asS'd'
I3416
sa(dp6832
S'a'
I14080
sS's'
I15
sS'm'
(lp6833
I1
aI1
aI5
aI15
aI23
aI59
aI125
aI51
aI225
aI439
aI2019
aI2589
aI7781
aI13111
aI2911
asS'd'
I3417
sa(dp6834
S'a'
I14085
sS's'
I15
sS'm'
(lp6835
I1
aI1
aI1
aI3
aI1
aI31
aI37
aI245
aI203
aI305
aI821
aI367
aI5211
aI9791
aI21777
asS'd'
I3418
sa(dp6836
S'a'
I14086
sS's'
I15
sS'm'
(lp6837
I1
aI1
aI5
aI9
aI9
aI31
aI97
aI25
aI271
aI83
aI343
aI2461
aI1805
aI14383
aI10059
asS'd'
I3419
sa(dp6838
S'a'
I14095
sS's'
I15
sS'm'
(lp6839
I1
aI1
aI5
aI13
aI15
aI33
aI127
aI109
aI137
aI963
aI961
aI1647
aI7881
aI8133
aI22359
asS'd'
I3420
sa(dp6840
S'a'
I14138
sS's'
I15
sS'm'
(lp6841
I1
aI1
aI3
aI7
aI25
aI31
aI123
aI241
aI283
aI1
aI1781
aI23
aI971
aI6485
aI127
asS'd'
I3421
sa(dp6842
S'a'
I14145
sS's'
I15
sS'm'
(lp6843
I1
aI1
aI5
aI15
aI15
aI27
aI25
aI145
aI395
aI679
aI979
aI571
aI1585
aI14787
aI7465
asS'd'
I3422
sa(dp6844
S'a'
I14158
sS's'
I15
sS'm'
(lp6845
I1
aI1
aI5
aI7
aI13
aI11
aI7
aI131
aI511
aI597
aI379
aI1513
aI6267
aI16039
aI1503
asS'd'
I3423
sa(dp6846
S'a'
I14166
sS's'
I15
sS'm'
(lp6847
I1
aI1
aI1
aI13
aI15
aI49
aI73
aI217
aI353
aI577
aI1913
aI1127
aI961
aI11557
aI24993
asS'd'
I3424
sa(dp6848
S'a'
I14179
sS's'
I15
sS'm'
(lp6849
I1
aI3
aI3
aI9
aI7
aI3
aI105
aI141
aI377
aI687
aI1917
aI485
aI983
aI11149
aI23303
asS'd'
I3425
sa(dp6850
S'a'
I14181
sS's'
I15
sS'm'
(lp6851
I1
aI1
aI3
aI15
aI11
aI7
aI117
aI179
aI505
aI67
aI1817
aI913
aI5757
aI1981
aI1637
asS'd'
I3426
sa(dp6852
S'a'
I14188
sS's'
I15
sS'm'
(lp6853
I1
aI1
aI1
aI7
aI5
aI29
aI3
aI43
aI223
aI295
aI1895
aI3425
aI5355
aI5155
aI17197
asS'd'
I3427
sa(dp6854
S'a'
I14193
sS's'
I15
sS'm'
(lp6855
I1
aI1
aI7
aI9
aI21
aI59
aI121
aI245
aI73
aI233
aI1527
aI869
aI4145
aI7995
aI6473
asS'd'
I3428
sa(dp6856
S'a'
I14200
sS's'
I15
sS'm'
(lp6857
I1
aI1
aI5
aI13
aI17
aI21
aI89
aI179
aI495
aI669
aI453
aI2603
aI5969
aI6161
aI4743
asS'd'
I3429
sa(dp6858
S'a'
I14203
sS's'
I15
sS'm'
(lp6859
I1
aI1
aI7
aI11
aI25
aI21
aI103
aI131
aI391
aI249
aI1633
aI2603
aI2207
aI8987
aI15487
asS'd'
I3430
sa(dp6860
S'a'
I14215
sS's'
I15
sS'm'
(lp6861
I1
aI3
aI7
aI9
aI13
aI45
aI99
aI251
aI115
aI597
aI1505
aI2421
aI1231
aI10015
aI24295
asS'd'
I3431
sa(dp6862
S'a'
I14224
sS's'
I15
sS'm'
(lp6863
I1
aI1
aI5
aI5
aI31
aI49
aI17
aI67
aI463
aI813
aI1491
aI3309
aI7881
aI8109
aI7289
asS'd'
I3432
sa(dp6864
S'a'
I14230
sS's'
I15
sS'm'
(lp6865
I1
aI3
aI1
aI15
aI23
aI35
aI123
aI21
aI169
aI499
aI95
aI603
aI1829
aI7865
aI26313
asS'd'
I3433
sa(dp6866
S'a'
I14233
sS's'
I15
sS'm'
(lp6867
I1
aI1
aI7
aI1
aI9
aI29
aI45
aI65
aI95
aI97
aI673
aI3673
aI2969
aI2317
aI22209
asS'd'
I3434
sa(dp6868
S'a'
I14236
sS's'
I15
sS'm'
(lp6869
I1
aI1
aI3
aI7
aI29
aI33
aI121
aI17
aI331
aI487
aI1901
aI1951
aI5383
aI9375
aI4029
asS'd'
I3435
sa(dp6870
S'a'
I14246
sS's'
I15
sS'm'
(lp6871
I1
aI3
aI7
aI9
aI25
aI43
aI91
aI147
aI141
aI401
aI1647
aI2697
aI4645
aI7179
aI31857
asS'd'
I3436
sa(dp6872
S'a'
I14267
sS's'
I15
sS'm'
(lp6873
I1
aI3
aI5
aI11
aI9
aI31
aI127
aI105
aI39
aI883
aI1635
aI919
aI5069
aI2875
aI24519
asS'd'
I3437
sa(dp6874
S'a'
I14282
sS's'
I15
sS'm'
(lp6875
I1
aI1
aI5
aI9
aI1
aI63
aI73
aI135
aI95
aI503
aI385
aI3903
aI545
aI12635
aI27569
asS'd'
I3438
sa(dp6876
S'a'
I14287
sS's'
I15
sS'm'
(lp6877
I1
aI1
aI3
aI11
aI27
aI31
aI47
aI173
aI55
aI339
aI1255
aI1947
aI793
aI14133
aI13963
asS'd'
I3439
sa(dp6878
S'a'
I14301
sS's'
I15
sS'm'
(lp6879
I1
aI1
aI3
aI15
aI17
aI33
aI113
aI249
aI401
aI743
aI1307
aI3123
aI627
aI1253
aI13285
asS'd'
I3440
sa(dp6880
S'a'
I14323
sS's'
I15
sS'm'
(lp6881
I1
aI1
aI3
aI1
aI9
aI7
aI39
aI65
aI281
aI107
aI833
aI193
aI2987
aI12267
aI31335
asS'd'
I3441
sa(dp6882
S'a'
I14325
sS's'
I15
sS'm'
(lp6883
I1
aI1
aI7
aI3
aI15
aI21
aI99
aI211
aI39
aI179
aI587
aI1169
aI6455
aI8225
aI2049
asS'd'
I3442
sa(dp6884
S'a'
I14329
sS's'
I15
sS'm'
(lp6885
I1
aI3
aI5
aI13
aI5
aI13
aI123
aI1
aI223
aI273
aI731
aI2967
aI4793
aI4229
aI26031
asS'd'
I3443
sa(dp6886
S'a'
I14339
sS's'
I15
sS'm'
(lp6887
I1
aI1
aI1
aI1
aI3
aI17
aI7
aI23
aI225
aI757
aI743
aI1257
aI2047
aI12509
aI25467
asS'd'
I3444
sa(dp6888
S'a'
I14342
sS's'
I15
sS'm'
(lp6889
I1
aI1
aI7
aI15
aI29
aI3
aI15
aI113
aI227
aI675
aI1295
aI2777
aI2921
aI5485
aI2577
asS'd'
I3445
sa(dp6890
S'a'
I14351
sS's'
I15
sS'm'
(lp6891
I1
aI3
aI7
aI13
aI19
aI21
aI85
aI129
aI45
aI599
aI317
aI1513
aI4953
aI10383
aI25253
asS'd'
I3446
sa(dp6892
S'a'
I14356
sS's'
I15
sS'm'
(lp6893
I1
aI1
aI7
aI11
aI13
aI47
aI127
aI67
aI219
aI131
aI905
aI2005
aI851
aI15243
aI5777
asS'd'
I3447
sa(dp6894
S'a'
I14359
sS's'
I15
sS'm'
(lp6895
I1
aI1
aI5
aI3
aI23
aI57
aI57
aI189
aI153
aI37
aI955
aI2049
aI1295
aI15119
aI27213
asS'd'
I3448
sa(dp6896
S'a'
I14370
sS's'
I15
sS'm'
(lp6897
I1
aI3
aI7
aI11
aI13
aI61
aI3
aI241
aI269
aI789
aI1595
aI2369
aI4843
aI11347
aI21543
asS'd'
I3449
sa(dp6898
S'a'
I14402
sS's'
I15
sS'm'
(lp6899
I1
aI1
aI5
aI5
aI25
aI21
aI19
aI237
aI3
aI605
aI1343
aI3965
aI3511
aI7889
aI27759
asS'd'
I3450
sa(dp6900
S'a'
I14411
sS's'
I15
sS'm'
(lp6901
I1
aI3
aI1
aI15
aI21
aI15
aI123
aI5
aI345
aI945
aI283
aI1313
aI335
aI2085
aI19505
asS'd'
I3451
sa(dp6902
S'a'
I14421
sS's'
I15
sS'm'
(lp6903
I1
aI1
aI3
aI3
aI5
aI21
aI123
aI89
aI67
aI11
aI1247
aI1155
aI287
aI13455
aI5693
asS'd'
I3452
sa(dp6904
S'a'
I14431
sS's'
I15
sS'm'
(lp6905
I1
aI3
aI3
aI13
aI1
aI53
aI101
aI27
aI387
aI379
aI19
aI751
aI2445
aI11737
aI975
asS'd'
I3453
sa(dp6906
S'a'
I14435
sS's'
I15
sS'm'
(lp6907
I1
aI3
aI3
aI3
aI9
aI29
aI81
aI117
aI443
aI145
aI1619
aI1813
aI8125
aI5829
aI28617
asS'd'
I3454
sa(dp6908
S'a'
I14442
sS's'
I15
sS'm'
(lp6909
I1
aI1
aI5
aI15
aI27
aI15
aI83
aI83
aI61
aI715
aI1655
aI1631
aI3457
aI2727
aI2163
asS'd'
I3455
sa(dp6910
S'a'
I14447
sS's'
I15
sS'm'
(lp6911
I1
aI3
aI1
aI5
aI11
aI11
aI121
aI7
aI135
aI883
aI927
aI1817
aI6839
aI12361
aI24119
asS'd'
I3456
sa(dp6912
S'a'
I14456
sS's'
I15
sS'm'
(lp6913
I1
aI3
aI7
aI11
aI23
aI59
aI39
aI165
aI109
aI355
aI1303
aI381
aI5697
aI275
aI3771
asS'd'
I3457
sa(dp6914
S'a'
I14459
sS's'
I15
sS'm'
(lp6915
I1
aI3
aI5
aI11
aI11
aI5
aI81
aI157
aI55
aI435
aI613
aI127
aI4087
aI3791
aI21627
asS'd'
I3458
sa(dp6916
S'a'
I14472
sS's'
I15
sS'm'
(lp6917
I1
aI3
aI7
aI15
aI13
aI37
aI83
aI195
aI207
aI771
aI51
aI3685
aI6389
aI1229
aI11101
asS'd'
I3459
sa(dp6918
S'a'
I14477
sS's'
I15
sS'm'
(lp6919
I1
aI3
aI7
aI13
aI31
aI3
aI9
aI13
aI487
aI95
aI77
aI809
aI5809
aI12887
aI29933
asS'd'
I3460
sa(dp6920
S'a'
I14490
sS's'
I15
sS'm'
(lp6921
I1
aI1
aI3
aI7
aI25
aI9
aI13
aI29
aI353
aI659
aI1785
aI3825
aI3729
aI13109
aI12973
asS'd'
I3461
sa(dp6922
S'a'
I14496
sS's'
I15
sS'm'
(lp6923
I1
aI1
aI1
aI5
aI21
aI3
aI97
aI1
aI245
aI917
aI29
aI1429
aI8141
aI7569
aI32493
asS'd'
I3462
sa(dp6924
S'a'
I14501
sS's'
I15
sS'm'
(lp6925
I1
aI3
aI1
aI9
aI19
aI13
aI13
aI109
aI377
aI1007
aI1737
aI1939
aI1419
aI1145
aI5065
asS'd'
I3463
sa(dp6926
S'a'
I14505
sS's'
I15
sS'm'
(lp6927
I1
aI1
aI7
aI9
aI27
aI57
aI53
aI69
aI423
aI43
aI1629
aI1003
aI1473
aI10809
aI5659
asS'd'
I3464
sa(dp6928
S'a'
I14513
sS's'
I15
sS'm'
(lp6929
I1
aI1
aI1
aI9
aI1
aI45
aI11
aI231
aI351
aI155
aI663
aI2783
aI3491
aI5725
aI25207
asS'd'
I3465
sa(dp6930
S'a'
I14520
sS's'
I15
sS'm'
(lp6931
I1
aI1
aI1
aI3
aI15
aI25
aI77
aI89
aI231
aI813
aI657
aI2603
aI4885
aI1383
aI14499
asS'd'
I3466
sa(dp6932
S'a'
I14534
sS's'
I15
sS'm'
(lp6933
I1
aI3
aI5
aI5
aI9
aI21
aI101
aI181
aI449
aI491
aI737
aI803
aI659
aI11771
aI545
asS'd'
I3467
sa(dp6934
S'a'
I14562
sS's'
I15
sS'm'
(lp6935
I1
aI3
aI7
aI9
aI7
aI19
aI27
aI199
aI265
aI329
aI1031
aI1235
aI3191
aI10071
aI16281
asS'd'
I3468
sa(dp6936
S'a'
I14576
sS's'
I15
sS'm'
(lp6937
I1
aI1
aI7
aI11
aI27
aI55
aI3
aI127
aI503
aI1003
aI1041
aI1953
aI5835
aI4851
aI13485
asS'd'
I3469
sa(dp6938
S'a'
I14579
sS's'
I15
sS'm'
(lp6939
I1
aI1
aI7
aI15
aI5
aI45
aI97
aI61
aI221
aI497
aI1949
aI3163
aI4707
aI8441
aI1437
asS'd'
I3470
sa(dp6940
S'a'
I14585
sS's'
I15
sS'm'
(lp6941
I1
aI3
aI5
aI1
aI3
aI35
aI107
aI9
aI473
aI971
aI227
aI2225
aI3999
aI3095
aI18879
asS'd'
I3471
sa(dp6942
S'a'
I14586
sS's'
I15
sS'm'
(lp6943
I1
aI1
aI1
aI9
aI21
aI59
aI21
aI1
aI41
aI435
aI575
aI491
aI1839
aI1095
aI9727
asS'd'
I3472
sa(dp6944
S'a'
I14606
sS's'
I15
sS'm'
(lp6945
I1
aI3
aI5
aI9
aI13
aI29
aI123
aI251
aI465
aI701
aI1105
aI829
aI573
aI11503
aI11861
asS'd'
I3473
sa(dp6946
S'a'
I14627
sS's'
I15
sS'm'
(lp6947
I1
aI3
aI3
aI13
aI27
aI59
aI29
aI111
aI225
aI973
aI561
aI1481
aI835
aI9261
aI13831
asS'd'
I3474
sa(dp6948
S'a'
I14630
sS's'
I15
sS'm'
(lp6949
I1
aI1
aI1
aI7
aI17
aI3
aI97
aI211
aI333
aI315
aI571
aI3523
aI7305
aI6461
aI20139
asS'd'
I3475
sa(dp6950
S'a'
I14634
sS's'
I15
sS'm'
(lp6951
I1
aI3
aI7
aI11
aI31
aI21
aI105
aI247
aI113
aI863
aI1767
aI381
aI4623
aI8935
aI7911
asS'd'
I3476
sa(dp6952
S'a'
I14636
sS's'
I15
sS'm'
(lp6953
I1
aI1
aI5
aI7
aI29
aI45
aI17
aI155
aI69
aI17
aI655
aI1983
aI6385
aI6177
aI7961
asS'd'
I3477
sa(dp6954
S'a'
I14647
sS's'
I15
sS'm'
(lp6955
I1
aI3
aI3
aI15
aI31
aI15
aI63
aI81
aI309
aI115
aI393
aI3445
aI689
aI13963
aI18887
asS'd'
I3478
sa(dp6956
S'a'
I14653
sS's'
I15
sS'm'
(lp6957
I1
aI1
aI5
aI1
aI19
aI39
aI127
aI61
aI357
aI53
aI195
aI2745
aI7853
aI5753
aI3669
asS'd'
I3479
sa(dp6958
S'a'
I14659
sS's'
I15
sS'm'
(lp6959
I1
aI3
aI7
aI7
aI17
aI51
aI57
aI145
aI451
aI365
aI1517
aI909
aI4265
aI10737
aI9579
asS'd'
I3480
sa(dp6960
S'a'
I14671
sS's'
I15
sS'm'
(lp6961
I1
aI1
aI3
aI13
aI3
aI37
aI121
aI103
aI257
aI47
aI1685
aI2951
aI5753
aI15379
aI8899
asS'd'
I3481
sa(dp6962
S'a'
I14674
sS's'
I15
sS'm'
(lp6963
I1
aI1
aI5
aI7
aI31
aI63
aI61
aI197
aI97
aI773
aI133
aI1517
aI3093
aI14879
aI22941
asS'd'
I3482
sa(dp6964
S'a'
I14701
sS's'
I15
sS'm'
(lp6965
I1
aI1
aI5
aI1
aI3
aI9
aI27
aI53
aI97
aI663
aI1915
aI409
aI471
aI1391
aI24853
asS'd'
I3483
sa(dp6966
S'a'
I14716
sS's'
I15
sS'm'
(lp6967
I1
aI1
aI1
aI7
aI21
aI53
aI69
aI5
aI187
aI571
aI2023
aI997
aI323
aI12059
aI7071
asS'd'
I3484
sa(dp6968
S'a'
I14719
sS's'
I15
sS'm'
(lp6969
I1
aI3
aI3
aI1
aI7
aI59
aI55
aI157
aI101
aI123
aI1301
aI3709
aI4673
aI3897
aI28791
asS'd'
I3485
sa(dp6970
S'a'
I14720
sS's'
I15
sS'm'
(lp6971
I1
aI3
aI7
aI5
aI5
aI23
aI39
aI139
aI365
aI415
aI1481
aI3415
aI6323
aI11109
aI5719
asS'd'
I3486
sa(dp6972
S'a'
I14725
sS's'
I15
sS'm'
(lp6973
I1
aI3
aI5
aI3
aI5
aI11
aI23
aI143
aI243
aI229
aI183
aI3367
aI3187
aI8151
aI28351
asS'd'
I3487
sa(dp6974
S'a'
I14730
sS's'
I15
sS'm'
(lp6975
I1
aI3
aI7
aI9
aI5
aI37
aI29
aI23
aI437
aI827
aI985
aI2879
aI7611
aI1391
aI19087
asS'd'
I3488
sa(dp6976
S'a'
I14743
sS's'
I15
sS'm'
(lp6977
I1
aI3
aI3
aI5
aI7
aI9
aI5
aI143
aI217
aI757
aI1697
aI2459
aI453
aI8679
aI4513
asS'd'
I3489
sa(dp6978
S'a'
I14747
sS's'
I15
sS'm'
(lp6979
I1
aI3
aI5
aI5
aI11
aI33
aI3
aI143
aI293
aI921
aI185
aI2461
aI5547
aI12247
aI28591
asS'd'
I3490
sa(dp6980
S'a'
I14786
sS's'
I15
sS'm'
(lp6981
I1
aI3
aI7
aI5
aI3
aI53
aI43
aI179
aI235
aI417
aI1307
aI1367
aI3695
aI12809
aI1807
asS'd'
I3491
sa(dp6982
S'a'
I14788
sS's'
I15
sS'm'
(lp6983
I1
aI3
aI1
aI11
aI15
aI43
aI115
aI229
aI157
aI25
aI687
aI3347
aI271
aI5777
aI8557
asS'd'
I3492
sa(dp6984
S'a'
I14792
sS's'
I15
sS'm'
(lp6985
I1
aI3
aI7
aI5
aI27
aI37
aI55
aI135
aI209
aI47
aI1603
aI957
aI5785
aI11141
aI10407
asS'd'
I3493
sa(dp6986
S'a'
I14795
sS's'
I15
sS'm'
(lp6987
I1
aI1
aI1
aI15
aI17
aI17
aI103
aI29
aI489
aI493
aI119
aI1707
aI3463
aI1815
aI32055
asS'd'
I3494
sa(dp6988
S'a'
I14809
sS's'
I15
sS'm'
(lp6989
I1
aI3
aI7
aI11
aI17
aI13
aI115
aI145
aI77
aI515
aI1911
aI477
aI5997
aI8731
aI3143
asS'd'
I3495
sa(dp6990
S'a'
I14831
sS's'
I15
sS'm'
(lp6991
I1
aI3
aI1
aI13
aI31
aI41
aI73
aI91
aI231
aI1
aI455
aI2023
aI4691
aI3613
aI16329
asS'd'
I3496
sa(dp6992
S'a'
I14834
sS's'
I15
sS'm'
(lp6993
I1
aI1
aI5
aI15
aI15
aI39
aI17
aI117
aI131
aI657
aI1939
aI2245
aI2575
aI195
aI25209
asS'd'
I3497
sa(dp6994
S'a'
I14850
sS's'
I15
sS'm'
(lp6995
I1
aI3
aI7
aI15
aI5
aI51
aI69
aI141
aI499
aI931
aI1165
aI2119
aI1703
aI10867
aI28443
asS'd'
I3498
sa(dp6996
S'a'
I14855
sS's'
I15
sS'm'
(lp6997
I1
aI1
aI1
aI15
aI13
aI45
aI45
aI103
aI115
aI177
aI651
aI2545
aI1417
aI5349
aI3385
asS'd'
I3499
sa(dp6998
S'a'
I14859
sS's'
I15
sS'm'
(lp6999
I1
aI3
aI3
aI1
aI1
aI41
aI117
aI15
aI225
aI861
aI843
aI2775
aI4543
aI6275
aI14671
asS'd'
I3500
sa(dp7000
S'a'
I14864
sS's'
I15
sS'm'
(lp7001
I1
aI3
aI5
aI15
aI5
aI35
aI87
aI193
aI341
aI55
aI1131
aI945
aI6865
aI11271
aI18705
asS'd'
I3501
sa(dp7002
S'a'
I14876
sS's'
I15
sS'm'
(lp7003
I1
aI3
aI5
aI9
aI13
aI35
aI71
aI197
aI79
aI351
aI3
aI3939
aI1105
aI12455
aI28921
asS'd'
I3502
sa(dp7004
S'a'
I14889
sS's'
I15
sS'm'
(lp7005
I1
aI3
aI1
aI13
aI9
aI23
aI89
aI165
aI59
aI257
aI1369
aI161
aI6255
aI2997
aI19175
asS'd'
I3503
sa(dp7006
S'a'
I14890
sS's'
I15
sS'm'
(lp7007
I1
aI3
aI5
aI3
aI5
aI41
aI107
aI231
aI111
aI207
aI1865
aI2079
aI5891
aI2487
aI5863
asS'd'
I3504
sa(dp7008
S'a'
I14898
sS's'
I15
sS'm'
(lp7009
I1
aI3
aI7
aI15
aI3
aI3
aI105
aI235
aI263
aI991
aI367
aI1885
aI1769
aI7805
aI11909
asS'd'
I3505
sa(dp7010
S'a'
I14909
sS's'
I15
sS'm'
(lp7011
I1
aI3
aI3
aI5
aI15
aI59
aI67
aI247
aI77
aI367
aI1641
aI1959
aI1921
aI5939
aI17355
asS'd'
I3506
sa(dp7012
S'a'
I14917
sS's'
I15
sS'm'
(lp7013
I1
aI1
aI7
aI1
aI3
aI53
aI37
aI5
aI221
aI779
aI1353
aI1633
aI2769
aI6355
aI8505
asS'd'
I3507
sa(dp7014
S'a'
I14924
sS's'
I15
sS'm'
(lp7015
I1
aI1
aI7
aI13
aI11
aI13
aI73
aI227
aI115
aI523
aI355
aI3127
aI7545
aI8409
aI22335
asS'd'
I3508
sa(dp7016
S'a'
I14929
sS's'
I15
sS'm'
(lp7017
I1
aI1
aI5
aI11
aI21
aI15
aI91
aI115
aI427
aI683
aI461
aI2433
aI6313
aI4595
aI24401
asS'd'
I3509
sa(dp7018
S'a'
I14942
sS's'
I15
sS'm'
(lp7019
I1
aI3
aI7
aI5
aI29
aI21
aI57
aI215
aI423
aI717
aI1455
aI705
aI6835
aI4503
aI26077
asS'd'
I3510
sa(dp7020
S'a'
I14951
sS's'
I15
sS'm'
(lp7021
I1
aI1
aI1
aI15
aI3
aI33
aI25
aI227
aI381
aI477
aI1023
aI2751
aI2229
aI631
aI16903
asS'd'
I3511
sa(dp7022
S'a'
I14969
sS's'
I15
sS'm'
(lp7023
I1
aI3
aI1
aI11
aI9
aI17
aI59
aI73
aI53
aI671
aI251
aI1729
aI7593
aI12473
aI22533
asS'd'
I3512
sa(dp7024
S'a'
I14970
sS's'
I15
sS'm'
(lp7025
I1
aI3
aI3
aI1
aI3
aI35
aI37
aI173
aI459
aI143
aI135
aI3871
aI2689
aI8007
aI4379
asS'd'
I3513
sa(dp7026
S'a'
I14972
sS's'
I15
sS'm'
(lp7027
I1
aI3
aI5
aI9
aI23
aI19
aI43
aI45
aI493
aI509
aI1851
aI1615
aI5675
aI13793
aI6973
asS'd'
I3514
sa(dp7028
S'a'
I14982
sS's'
I15
sS'm'
(lp7029
I1
aI3
aI3
aI15
aI5
aI17
aI77
aI85
aI451
aI753
aI579
aI1057
aI4851
aI6017
aI4195
asS'd'
I3515
sa(dp7030
S'a'
I14988
sS's'
I15
sS'm'
(lp7031
I1
aI3
aI3
aI5
aI31
aI29
aI81
aI159
aI103
aI391
aI15
aI899
aI4623
aI5957
aI31961
asS'd'
I3516
sa(dp7032
S'a'
I14994
sS's'
I15
sS'm'
(lp7033
I1
aI1
aI1
aI7
aI17
aI57
aI81
aI17
aI177
aI633
aI49
aI2793
aI5229
aI5995
aI9491
asS'd'
I3517
sa(dp7034
S'a'
I15005
sS's'
I15
sS'm'
(lp7035
I1
aI1
aI7
aI15
aI17
aI19
aI65
aI57
aI189
aI239
aI1229
aI929
aI2681
aI12845
aI29311
asS'd'
I3518
sa(dp7036
S'a'
I15016
sS's'
I15
sS'm'
(lp7037
I1
aI3
aI1
aI11
aI13
aI47
aI61
aI203
aI383
aI875
aI943
aI139
aI4217
aI8279
aI1047
asS'd'
I3519
sa(dp7038
S'a'
I15024
sS's'
I15
sS'm'
(lp7039
I1
aI3
aI7
aI13
aI23
aI7
aI1
aI69
aI47
aI537
aI1325
aI3101
aI685
aI14057
aI19953
asS'd'
I3520
sa(dp7040
S'a'
I15030
sS's'
I15
sS'm'
(lp7041
I1
aI3
aI3
aI1
aI1
aI7
aI39
aI77
aI47
aI755
aI527
aI2985
aI5433
aI15095
aI27741
asS'd'
I3521
sa(dp7042
S'a'
I15048
sS's'
I15
sS'm'
(lp7043
I1
aI1
aI7
aI5
aI23
aI57
aI79
aI155
aI81
aI937
aI1071
aI3929
aI1655
aI3831
aI17351
asS'd'
I3522
sa(dp7044
S'a'
I15054
sS's'
I15
sS'm'
(lp7045
I1
aI3
aI7
aI1
aI3
aI41
aI13
aI235
aI207
aI487
aI1883
aI2247
aI1231
aI2751
aI15615
asS'd'
I3523
sa(dp7046
S'a'
I15066
sS's'
I15
sS'm'
(lp7047
I1
aI1
aI7
aI1
aI21
aI57
aI95
aI191
aI119
aI483
aI283
aI2221
aI5665
aI14819
aI26097
asS'd'
I3524
sa(dp7048
S'a'
I15071
sS's'
I15
sS'm'
(lp7049
I1
aI3
aI1
aI1
aI9
aI59
aI27
aI51
aI393
aI31
aI925
aI715
aI7705
aI14885
aI28767
asS'd'
I3525
sa(dp7050
S'a'
I15072
sS's'
I15
sS'm'
(lp7051
I1
aI1
aI3
aI3
aI3
aI61
aI109
aI131
aI113
aI249
aI1331
aI2521
aI2973
aI6375
aI20093
asS'd'
I3526
sa(dp7052
S'a'
I15075
sS's'
I15
sS'm'
(lp7053
I1
aI3
aI7
aI9
aI31
aI37
aI125
aI245
aI237
aI245
aI111
aI379
aI7495
aI15531
aI2325
asS'd'
I3527
sa(dp7054
S'a'
I15119
sS's'
I15
sS'm'
(lp7055
I1
aI3
aI7
aI13
aI21
aI21
aI57
aI21
aI449
aI969
aI417
aI2999
aI509
aI639
aI7797
asS'd'
I3528
sa(dp7056
S'a'
I15121
sS's'
I15
sS'm'
(lp7057
I1
aI3
aI7
aI7
aI7
aI29
aI11
aI175
aI55
aI705
aI891
aI863
aI3021
aI10071
aI10267
asS'd'
I3529
sa(dp7058
S'a'
I15133
sS's'
I15
sS'm'
(lp7059
I1
aI1
aI3
aI13
aI19
aI17
aI127
aI57
aI449
aI579
aI337
aI899
aI1235
aI11269
aI4245
asS'd'
I3530
sa(dp7060
S'a'
I15138
sS's'
I15
sS'm'
(lp7061
I1
aI1
aI1
aI11
aI29
aI61
aI35
aI75
aI249
aI683
aI287
aI45
aI3277
aI7521
aI2073
asS'd'
I3531
sa(dp7062
S'a'
I15143
sS's'
I15
sS'm'
(lp7063
I1
aI3
aI5
aI5
aI15
aI25
aI77
aI63
aI63
aI801
aI1387
aI1533
aI2185
aI10899
aI28381
asS'd'
I3532
sa(dp7064
S'a'
I15170
sS's'
I15
sS'm'
(lp7065
I1
aI3
aI1
aI1
aI21
aI49
aI3
aI249
aI419
aI575
aI87
aI3749
aI2523
aI16125
aI9483
asS'd'
I3533
sa(dp7066
S'a'
I15194
sS's'
I15
sS'm'
(lp7067
I1
aI1
aI1
aI11
aI21
aI43
aI85
aI211
aI449
aI439
aI1495
aI1841
aI4765
aI15253
aI1467
asS'd'
I3534
sa(dp7068
S'a'
I15212
sS's'
I15
sS'm'
(lp7069
I1
aI3
aI3
aI15
aI3
aI37
aI31
aI243
aI187
aI995
aI1103
aI2723
aI1523
aI15967
aI28649
asS'd'
I3535
sa(dp7070
S'a'
I15223
sS's'
I15
sS'm'
(lp7071
I1
aI1
aI5
aI11
aI9
aI11
aI17
aI87
aI335
aI125
aI1079
aI1657
aI1237
aI8059
aI29833
asS'd'
I3536
sa(dp7072
S'a'
I15229
sS's'
I15
sS'm'
(lp7073
I1
aI3
aI1
aI3
aI3
aI41
aI35
aI37
aI33
aI61
aI505
aI3203
aI5
aI101
aI8571
asS'd'
I3537
sa(dp7074
S'a'
I15254
sS's'
I15
sS'm'
(lp7075
I1
aI1
aI3
aI11
aI9
aI11
aI85
aI235
aI261
aI473
aI109
aI2127
aI5745
aI6389
aI7431
asS'd'
I3538
sa(dp7076
S'a'
I15263
sS's'
I15
sS'm'
(lp7077
I1
aI1
aI5
aI15
aI3
aI55
aI77
aI97
aI17
aI193
aI1267
aI3063
aI6531
aI9797
aI8639
asS'd'
I3539
sa(dp7078
S'a'
I15270
sS's'
I15
sS'm'
(lp7079
I1
aI1
aI5
aI5
aI25
aI41
aI79
aI83
aI485
aI697
aI149
aI1023
aI89
aI6115
aI15227
asS'd'
I3540
sa(dp7080
S'a'
I15273
sS's'
I15
sS'm'
(lp7081
I1
aI1
aI3
aI15
aI1
aI9
aI73
aI251
aI33
aI599
aI1017
aI353
aI4305
aI16033
aI29663
asS'd'
I3541
sa(dp7082
S'a'
I15287
sS's'
I15
sS'm'
(lp7083
I1
aI3
aI7
aI15
aI3
aI1
aI89
aI39
aI125
aI337
aI1445
aI3131
aI3685
aI9849
aI25829
asS'd'
I3542
sa(dp7084
S'a'
I15299
sS's'
I15
sS'm'
(lp7085
I1
aI3
aI7
aI3
aI19
aI1
aI63
aI179
aI349
aI135
aI185
aI2977
aI2527
aI15087
aI18133
asS'd'
I3543
sa(dp7086
S'a'
I15301
sS's'
I15
sS'm'
(lp7087
I1
aI1
aI3
aI3
aI23
aI7
aI91
aI221
aI325
aI723
aI345
aI81
aI8077
aI5501
aI8453
asS'd'
I3544
sa(dp7088
S'a'
I15306
sS's'
I15
sS'm'
(lp7089
I1
aI1
aI3
aI9
aI7
aI3
aI13
aI173
aI479
aI161
aI1989
aI3255
aI2069
aI6717
aI559
asS'd'
I3545
sa(dp7090
S'a'
I15313
sS's'
I15
sS'm'
(lp7091
I1
aI3
aI3
aI5
aI9
aI61
aI93
aI203
aI277
aI367
aI1141
aI981
aI4745
aI12625
aI21003
asS'd'
I3546
sa(dp7092
S'a'
I15320
sS's'
I15
sS'm'
(lp7093
I1
aI3
aI5
aI5
aI27
aI17
aI5
aI211
aI403
aI701
aI5
aI3091
aI4611
aI5615
aI23667
asS'd'
I3547
sa(dp7094
S'a'
I15323
sS's'
I15
sS'm'
(lp7095
I1
aI1
aI3
aI1
aI21
aI61
aI125
aI77
aI57
aI463
aI1499
aI791
aI2087
aI2805
aI18829
asS'd'
I3548
sa(dp7096
S'a'
I15329
sS's'
I15
sS'm'
(lp7097
I1
aI3
aI5
aI3
aI11
aI41
aI125
aI231
aI119
aI837
aI831
aI1331
aI7439
aI2381
aI3759
asS'd'
I3549
sa(dp7098
S'a'
I15332
sS's'
I15
sS'm'
(lp7099
I1
aI3
aI1
aI11
aI19
aI59
aI117
aI107
aI443
aI699
aI315
aI1491
aI2581
aI15871
aI17159
asS'd'
I3550
sa(dp7100
S'a'
I15341
sS's'
I15
sS'm'
(lp7101
I1
aI3
aI5
aI11
aI5
aI9
aI121
aI35
aI209
aI877
aI527
aI3493
aI4657
aI16093
aI17589
asS'd'
I3551
sa(dp7102
S'a'
I15359
sS's'
I15
sS'm'
(lp7103
I1
aI1
aI7
aI15
aI9
aI43
aI119
aI29
aI381
aI479
aI1443
aI3171
aI5053
aI9625
aI21161
asS'd'
I3552
sa(dp7104
S'a'
I15361
sS's'
I15
sS'm'
(lp7105
I1
aI1
aI3
aI5
aI15
aI21
aI31
aI223
aI83
aI399
aI1529
aI3605
aI6343
aI10469
aI10099
asS'd'
I3553
sa(dp7106
S'a'
I15364
sS's'
I15
sS'm'
(lp7107
I1
aI1
aI3
aI5
aI5
aI45
aI23
aI123
aI353
aI971
aI85
aI3069
aI3245
aI6569
aI13241
asS'd'
I3554
sa(dp7108
S'a'
I15367
sS's'
I15
sS'm'
(lp7109
I1
aI1
aI1
aI3
aI25
aI49
aI5
aI77
aI491
aI881
aI993
aI1195
aI7677
aI5709
aI10807
asS'd'
I3555
sa(dp7110
S'a'
I15379
sS's'
I15
sS'm'
(lp7111
I1
aI3
aI3
aI3
aI5
aI49
aI127
aI255
aI183
aI583
aI1599
aI987
aI7281
aI7149
aI28507
asS'd'
I3556
sa(dp7112
S'a'
I15391
sS's'
I15
sS'm'
(lp7113
I1
aI1
aI5
aI1
aI13
aI55
aI55
aI157
aI197
aI25
aI1971
aI3161
aI3903
aI8919
aI13563
asS'd'
I3557
sa(dp7114
S'a'
I15415
sS's'
I15
sS'm'
(lp7115
I1
aI3
aI7
aI9
aI3
aI37
aI79
aI193
aI25
aI103
aI843
aI2651
aI6341
aI2653
aI24337
asS'd'
I3558
sa(dp7116
S'a'
I15416
sS's'
I15
sS'm'
(lp7117
I1
aI1
aI7
aI3
aI25
aI49
aI99
aI139
aI45
aI211
aI2033
aI2331
aI7037
aI7177
aI1755
asS'd'
I3559
sa(dp7118
S'a'
I15419
sS's'
I15
sS'm'
(lp7119
I1
aI3
aI7
aI3
aI5
aI19
aI127
aI135
aI403
aI221
aI141
aI1065
aI3935
aI2745
aI25979
asS'd'
I3560
sa(dp7120
S'a'
I15433
sS's'
I15
sS'm'
(lp7121
I1
aI1
aI3
aI3
aI31
aI23
aI111
aI37
aI261
aI7
aI835
aI2379
aI7927
aI8181
aI23751
asS'd'
I3561
sa(dp7122
S'a'
I15469
sS's'
I15
sS'm'
(lp7123
I1
aI3
aI7
aI15
aI1
aI39
aI79
aI3
aI103
aI427
aI1917
aI809
aI5039
aI689
aI1939
asS'd'
I3562
sa(dp7124
S'a'
I15478
sS's'
I15
sS'm'
(lp7125
I1
aI1
aI1
aI15
aI29
aI37
aI39
aI243
aI149
aI353
aI763
aI3405
aI5751
aI9441
aI6653
asS'd'
I3563
sa(dp7126
S'a'
I15481
sS's'
I15
sS'm'
(lp7127
I1
aI3
aI3
aI11
aI1
aI57
aI125
aI151
aI445
aI423
aI841
aI2265
aI5017
aI15863
aI13057
asS'd'
I3564
sa(dp7128
S'a'
I15482
sS's'
I15
sS'm'
(lp7129
I1
aI3
aI5
aI13
aI11
aI49
aI61
aI159
aI211
aI917
aI561
aI1903
aI3985
aI11117
aI28969
asS'd'
I3565
sa(dp7130
S'a'
I15498
sS's'
I15
sS'm'
(lp7131
I1
aI3
aI5
aI13
aI29
aI5
aI35
aI51
aI91
aI291
aI9
aI3713
aI3341
aI4551
aI12085
asS'd'
I3566
sa(dp7132
S'a'
I15505
sS's'
I15
sS'm'
(lp7133
I1
aI3
aI3
aI1
aI1
aI39
aI111
aI141
aI319
aI179
aI1709
aI1605
aI5063
aI13279
aI10003
asS'd'
I3567
sa(dp7134
S'a'
I15517
sS's'
I15
sS'm'
(lp7135
I1
aI1
aI3
aI9
aI7
aI59
aI91
aI41
aI343
aI475
aI1669
aI2311
aI5141
aI12661
aI25847
asS'd'
I3568
sa(dp7136
S'a'
I15518
sS's'
I15
sS'm'
(lp7137
I1
aI3
aI5
aI9
aI9
aI11
aI49
aI221
aI1
aI243
aI791
aI229
aI503
aI373
aI19189
asS'd'
I3569
sa(dp7138
S'a'
I15527
sS's'
I15
sS'm'
(lp7139
I1
aI1
aI5
aI11
aI17
aI13
aI45
aI57
aI215
aI491
aI1601
aI2183
aI3713
aI429
aI22007
asS'd'
I3570
sa(dp7140
S'a'
I15528
sS's'
I15
sS'm'
(lp7141
I1
aI1
aI3
aI11
aI31
aI61
aI23
aI237
aI261
aI955
aI1085
aI1541
aI2601
aI909
aI7749
asS'd'
I3571
sa(dp7142
S'a'
I15545
sS's'
I15
sS'm'
(lp7143
I1
aI1
aI3
aI9
aI13
aI11
aI121
aI173
aI177
aI551
aI1757
aI2745
aI2265
aI4611
aI743
asS'd'
I3572
sa(dp7144
S'a'
I15548
sS's'
I15
sS'm'
(lp7145
I1
aI1
aI3
aI15
aI23
aI43
aI107
aI239
aI463
aI369
aI1857
aI1073
aI1247
aI1029
aI22557
asS'd'
I3573
sa(dp7146
S'a'
I15554
sS's'
I15
sS'm'
(lp7147
I1
aI1
aI3
aI11
aI23
aI35
aI89
aI93
aI41
aI941
aI1141
aI2339
aI1423
aI8007
aI28685
asS'd'
I3574
sa(dp7148
S'a'
I15565
sS's'
I15
sS'm'
(lp7149
I1
aI3
aI5
aI13
aI29
aI7
aI79
aI15
aI59
aI145
aI1237
aI2215
aI1257
aI12621
aI31101
asS'd'
I3575
sa(dp7150
S'a'
I15577
sS's'
I15
sS'm'
(lp7151
I1
aI1
aI3
aI7
aI13
aI55
aI57
aI229
aI205
aI1009
aI341
aI3901
aI5189
aI957
aI32587
asS'd'
I3576
sa(dp7152
S'a'
I15580
sS's'
I15
sS'm'
(lp7153
I1
aI3
aI7
aI11
aI1
aI1
aI41
aI7
aI365
aI407
aI1609
aI1423
aI6483
aI5171
aI32519
asS'd'
I3577
sa(dp7154
S'a'
I15587
sS's'
I15
sS'm'
(lp7155
I1
aI3
aI7
aI3
aI17
aI31
aI125
aI27
aI125
aI335
aI1395
aI2639
aI329
aI2549
aI14449
asS'd'
I3578
sa(dp7156
S'a'
I15601
sS's'
I15
sS'm'
(lp7157
I1
aI3
aI3
aI7
aI19
aI45
aI11
aI73
aI123
aI179
aI1685
aI3385
aI2379
aI3387
aI16793
asS'd'
I3579
sa(dp7158
S'a'
I15604
sS's'
I15
sS'm'
(lp7159
I1
aI3
aI7
aI5
aI31
aI25
aI47
aI153
aI121
aI453
aI935
aI3953
aI2081
aI12145
aI24979
asS'd'
I3580
sa(dp7160
S'a'
I15611
sS's'
I15
sS'm'
(lp7161
I1
aI1
aI7
aI13
aI25
aI11
aI65
aI3
aI277
aI237
aI1129
aI1801
aI4165
aI9065
aI18747
asS'd'
I3581
sa(dp7162
S'a'
I15616
sS's'
I15
sS'm'
(lp7163
I1
aI1
aI7
aI7
aI13
aI5
aI37
aI253
aI507
aI645
aI1355
aI3401
aI6707
aI6329
aI11237
asS'd'
I3582
sa(dp7164
S'a'
I15619
sS's'
I15
sS'm'
(lp7165
I1
aI1
aI3
aI15
aI17
aI49
aI3
aI233
aI407
aI451
aI69
aI3859
aI3171
aI12303
aI21031
asS'd'
I3583
sa(dp7166
S'a'
I15625
sS's'
I15
sS'm'
(lp7167
I1
aI1
aI3
aI3
aI9
aI53
aI119
aI117
aI401
aI903
aI1449
aI3639
aI4083
aI2095
aI22085
asS'd'
I3584
sa(dp7168
S'a'
I15633
sS's'
I15
sS'm'
(lp7169
I1
aI3
aI7
aI15
aI5
aI61
aI117
aI193
aI137
aI431
aI195
aI4019
aI3047
aI5049
aI14281
asS'd'
I3585
sa(dp7170
S'a'
I15674
sS's'
I15
sS'm'
(lp7171
I1
aI1
aI1
aI15
aI17
aI19
aI29
aI83
aI449
aI257
aI1105
aI1949
aI1749
aI3459
aI6343
asS'd'
I3586
sa(dp7172
S'a'
I15681
sS's'
I15
sS'm'
(lp7173
I1
aI1
aI1
aI15
aI23
aI39
aI61
aI219
aI109
aI365
aI863
aI1813
aI6673
aI15999
aI5101
asS'd'
I3587
sa(dp7174
S'a'
I15691
sS's'
I15
sS'm'
(lp7175
I1
aI1
aI5
aI5
aI13
aI11
aI37
aI151
aI365
aI719
aI1233
aI2425
aI1285
aI1721
aI1205
asS'd'
I3588
sa(dp7176
S'a'
I15693
sS's'
I15
sS'm'
(lp7177
I1
aI3
aI3
aI3
aI7
aI53
aI109
aI153
aI45
aI425
aI1741
aI1229
aI4405
aI8071
aI25155
asS'd'
I3589
sa(dp7178
S'a'
I15696
sS's'
I15
sS'm'
(lp7179
I1
aI3
aI1
aI1
aI1
aI13
aI39
aI49
aI413
aI77
aI1367
aI2553
aI5563
aI7659
aI3467
asS'd'
I3590
sa(dp7180
S'a'
I15712
sS's'
I15
sS'm'
(lp7181
I1
aI1
aI5
aI9
aI3
aI49
aI23
aI11
aI445
aI121
aI1505
aI877
aI4137
aI1809
aI2429
asS'd'
I3591
sa(dp7182
S'a'
I15717
sS's'
I15
sS'm'
(lp7183
I1
aI1
aI1
aI11
aI21
aI13
aI93
aI33
aI493
aI805
aI775
aI2939
aI2961
aI13625
aI31879
asS'd'
I3592
sa(dp7184
S'a'
I15724
sS's'
I15
sS'm'
(lp7185
I1
aI1
aI7
aI5
aI1
aI59
aI63
aI131
aI373
aI23
aI337
aI2107
aI5315
aI4889
aI22851
asS'd'
I3593
sa(dp7186
S'a'
I15727
sS's'
I15
sS'm'
(lp7187
I1
aI1
aI3
aI13
aI21
aI47
aI15
aI131
aI353
aI793
aI1891
aI1757
aI5793
aI1147
aI23697
asS'd'
I3594
sa(dp7188
S'a'
I15730
sS's'
I15
sS'm'
(lp7189
I1
aI3
aI5
aI13
aI7
aI59
aI25
aI135
aI259
aI109
aI1835
aI429
aI8153
aI7355
aI145
asS'd'
I3595
sa(dp7190
S'a'
I15746
sS's'
I15
sS'm'
(lp7191
I1
aI3
aI3
aI13
aI9
aI47
aI121
aI89
aI89
aI635
aI1079
aI2353
aI4803
aI11369
aI12653
asS'd'
I3596
sa(dp7192
S'a'
I15751
sS's'
I15
sS'm'
(lp7193
I1
aI3
aI5
aI9
aI23
aI39
aI49
aI231
aI105
aI603
aI613
aI2021
aI6073
aI11819
aI10595
asS'd'
I3597
sa(dp7194
S'a'
I15760
sS's'
I15
sS'm'
(lp7195
I1
aI3
aI7
aI7
aI7
aI19
aI19
aI155
aI347
aI387
aI1459
aI3793
aI619
aI14437
aI2455
asS'd'
I3598
sa(dp7196
S'a'
I15770
sS's'
I15
sS'm'
(lp7197
I1
aI1
aI1
aI15
aI21
aI35
aI19
aI185
aI483
aI425
aI479
aI3429
aI5403
aI10791
aI14219
asS'd'
I3599
sa(dp7198
S'a'
I15782
sS's'
I15
sS'm'
(lp7199
I1
aI1
aI3
aI11
aI5
aI51
aI105
aI63
aI493
aI677
aI1457
aI2865
aI5619
aI9321
aI19583
asS'd'
I3600
sa(dp7200
S'a'
I15791
sS's'
I15
sS'm'
(lp7201
I1
aI1
aI3
aI3
aI23
aI1
aI77
aI177
aI263
aI289
aI1567
aI3837
aI5359
aI3269
aI16023
asS'd'
I3601
sa(dp7202
S'a'
I15796
sS's'
I15
sS'm'
(lp7203
I1
aI1
aI7
aI3
aI13
aI61
aI79
aI77
aI51
aI953
aI1417
aI795
aI4467
aI2981
aI25131
asS'd'
I3602
sa(dp7204
S'a'
I15808
sS's'
I15
sS'm'
(lp7205
I1
aI1
aI5
aI13
aI23
aI13
aI29
aI185
aI337
aI7
aI149
aI3609
aI8119
aI9545
aI16579
asS'd'
I3603
sa(dp7206
S'a'
I15814
sS's'
I15
sS'm'
(lp7207
I1
aI3
aI1
aI5
aI23
aI9
aI123
aI15
aI99
aI55
aI1021
aI3709
aI1521
aI15189
aI22193
asS'd'
I3604
sa(dp7208
S'a'
I15825
sS's'
I15
sS'm'
(lp7209
I1
aI3
aI7
aI9
aI13
aI41
aI39
aI45
aI49
aI181
aI1587
aI3213
aI1037
aI14775
aI3333
asS'd'
I3605
sa(dp7210
S'a'
I15828
sS's'
I15
sS'm'
(lp7211
I1
aI1
aI1
aI7
aI29
aI55
aI59
aI31
aI411
aI601
aI191
aI283
aI3211
aI7951
aI7919
asS'd'
I3606
sa(dp7212
S'a'
I15835
sS's'
I15
sS'm'
(lp7213
I1
aI1
aI7
aI7
aI21
aI47
aI7
aI193
aI343
aI831
aI1267
aI3289
aI1015
aI13093
aI2717
asS'd'
I3607
sa(dp7214
S'a'
I15844
sS's'
I15
sS'm'
(lp7215
I1
aI3
aI7
aI1
aI17
aI9
aI97
aI19
aI279
aI827
aI1699
aI3573
aI3137
aI3535
aI17791
asS'd'
I3608
sa(dp7216
S'a'
I15847
sS's'
I15
sS'm'
(lp7217
I1
aI1
aI5
aI11
aI27
aI15
aI103
aI135
aI35
aI625
aI1575
aI97
aI7013
aI13353
aI19333
asS'd'
I3609
sa(dp7218
S'a'
I15853
sS's'
I15
sS'm'
(lp7219
I1
aI3
aI3
aI7
aI17
aI13
aI49
aI135
aI435
aI743
aI1799
aI2655
aI4839
aI2893
aI31153
asS'd'
I3610
sa(dp7220
S'a'
I15856
sS's'
I15
sS'm'
(lp7221
I1
aI1
aI5
aI1
aI3
aI41
aI1
aI195
aI53
aI803
aI1575
aI2939
aI3873
aI10495
aI5211
asS'd'
I3611
sa(dp7222
S'a'
I15877
sS's'
I15
sS'm'
(lp7223
I1
aI3
aI1
aI15
aI19
aI19
aI37
aI59
aI355
aI483
aI685
aI3899
aI4645
aI15127
aI3479
asS'd'
I3612
sa(dp7224
S'a'
I15878
sS's'
I15
sS'm'
(lp7225
I1
aI1
aI5
aI3
aI25
aI9
aI9
aI229
aI101
aI631
aI1165
aI4091
aI3723
aI10655
aI9463
asS'd'
I3613
sa(dp7226
S'a'
I15887
sS's'
I15
sS'm'
(lp7227
I1
aI3
aI5
aI15
aI5
aI13
aI91
aI61
aI19
aI469
aI1675
aI3331
aI3121
aI3435
aI4111
asS'd'
I3614
sa(dp7228
S'a'
I15908
sS's'
I15
sS'm'
(lp7229
I1
aI1
aI7
aI1
aI31
aI61
aI23
aI83
aI165
aI551
aI1097
aI3825
aI5385
aI4723
aI3635
asS'd'
I3615
sa(dp7230
S'a'
I15917
sS's'
I15
sS'm'
(lp7231
I1
aI3
aI7
aI15
aI9
aI31
aI11
aI121
aI503
aI855
aI561
aI1647
aI1229
aI1147
aI15997
asS'd'
I3616
sa(dp7232
S'a'
I15923
sS's'
I15
sS'm'
(lp7233
I1
aI3
aI7
aI13
aI21
aI47
aI41
aI195
aI197
aI719
aI1263
aI3609
aI7515
aI2659
aI30713
asS'd'
I3617
sa(dp7234
S'a'
I15930
sS's'
I15
sS'm'
(lp7235
I1
aI1
aI1
aI7
aI31
aI61
aI101
aI101
aI479
aI571
aI605
aI301
aI6633
aI15587
aI23665
asS'd'
I3618
sa(dp7236
S'a'
I15937
sS's'
I15
sS'm'
(lp7237
I1
aI3
aI7
aI3
aI25
aI39
aI35
aI225
aI135
aI463
aI53
aI709
aI5129
aI4135
aI10421
asS'd'
I3619
sa(dp7238
S'a'
I15958
sS's'
I15
sS'm'
(lp7239
I1
aI1
aI5
aI13
aI19
aI55
aI107
aI15
aI163
aI287
aI673
aI899
aI5197
aI4619
aI3465
asS'd'
I3620
sa(dp7240
S'a'
I15977
sS's'
I15
sS'm'
(lp7241
I1
aI3
aI3
aI5
aI21
aI49
aI15
aI105
aI283
aI877
aI1875
aI1079
aI3431
aI13053
aI26599
asS'd'
I3621
sa(dp7242
S'a'
I15991
sS's'
I15
sS'm'
(lp7243
I1
aI1
aI7
aI1
aI1
aI1
aI95
aI113
aI119
aI575
aI1159
aI2325
aI6895
aI12177
aI4369
asS'd'
I3622
sa(dp7244
S'a'
I16007
sS's'
I15
sS'm'
(lp7245
I1
aI1
aI1
aI11
aI25
aI25
aI83
aI207
aI301
aI729
aI1947
aI2321
aI3621
aI15707
aI11303
asS'd'
I3623
sa(dp7246
S'a'
I16011
sS's'
I15
sS'm'
(lp7247
I1
aI1
aI5
aI5
aI7
aI63
aI83
aI105
aI211
aI175
aI1817
aI2883
aI5385
aI7437
aI24865
asS'd'
I3624
sa(dp7248
S'a'
I16014
sS's'
I15
sS'm'
(lp7249
I1
aI3
aI7
aI5
aI23
aI39
aI19
aI211
aI151
aI295
aI573
aI223
aI5065
aI6345
aI23187
asS'd'
I3625
sa(dp7250
S'a'
I16021
sS's'
I15
sS'm'
(lp7251
I1
aI1
aI7
aI11
aI15
aI31
aI89
aI123
aI57
aI695
aI685
aI1799
aI659
aI9929
aI22933
asS'd'
I3626
sa(dp7252
S'a'
I16022
sS's'
I15
sS'm'
(lp7253
I1
aI1
aI7
aI7
aI19
aI17
aI27
aI137
aI117
aI141
aI1481
aI869
aI7061
aI3073
aI19671
asS'd'
I3627
sa(dp7254
S'a'
I16028
sS's'
I15
sS'm'
(lp7255
I1
aI3
aI3
aI11
aI9
aI19
aI123
aI93
aI39
aI517
aI883
aI3769
aI2267
aI8089
aI6617
asS'd'
I3628
sa(dp7256
S'a'
I16035
sS's'
I15
sS'm'
(lp7257
I1
aI3
aI1
aI7
aI9
aI61
aI51
aI241
aI319
aI853
aI1239
aI899
aI105
aI1677
aI29351
asS'd'
I3629
sa(dp7258
S'a'
I16041
sS's'
I15
sS'm'
(lp7259
I1
aI1
aI7
aI15
aI13
aI59
aI85
aI175
aI223
aI87
aI905
aI3175
aI3405
aI3489
aI18475
asS'd'
I3630
sa(dp7260
S'a'
I16056
sS's'
I15
sS'm'
(lp7261
I1
aI1
aI1
aI15
aI1
aI55
aI79
aI97
aI315
aI605
aI851
aI4015
aI3689
aI9371
aI31523
asS'd'
I3631
sa(dp7262
S'a'
I16069
sS's'
I15
sS'm'
(lp7263
I1
aI1
aI5
aI15
aI1
aI39
aI91
aI27
aI211
aI881
aI1375
aI2307
aI5791
aI10185
aI23093
asS'd'
I3632
sa(dp7264
S'a'
I16076
sS's'
I15
sS'm'
(lp7265
I1
aI3
aI1
aI5
aI3
aI17
aI59
aI219
aI105
aI623
aI21
aI2843
aI3427
aI4799
aI3793
asS'd'
I3633
sa(dp7266
S'a'
I16081
sS's'
I15
sS'm'
(lp7267
I1
aI3
aI3
aI7
aI21
aI55
aI17
aI29
aI397
aI93
aI1981
aI4047
aI935
aI5971
aI14589
asS'd'
I3634
sa(dp7268
S'a'
I16087
sS's'
I15
sS'm'
(lp7269
I1
aI1
aI3
aI9
aI5
aI57
aI63
aI27
aI373
aI815
aI167
aI205
aI367
aI4945
aI30041
asS'd'
I3635
sa(dp7270
S'a'
I16088
sS's'
I15
sS'm'
(lp7271
I1
aI1
aI5
aI9
aI7
aI3
aI69
aI35
aI197
aI309
aI1729
aI3735
aI1523
aI10427
aI26253
asS'd'
I3636
sa(dp7272
S'a'
I16110
sS's'
I15
sS'm'
(lp7273
I1
aI1
aI3
aI7
aI7
aI49
aI35
aI189
aI297
aI311
aI2025
aI305
aI3863
aI14393
aI2533
asS'd'
I3637
sa(dp7274
S'a'
I16112
sS's'
I15
sS'm'
(lp7275
I1
aI3
aI3
aI9
aI17
aI31
aI5
aI17
aI167
aI601
aI909
aI3149
aI2533
aI12123
aI25325
asS'd'
I3638
sa(dp7276
S'a'
I16117
sS's'
I15
sS'm'
(lp7277
I1
aI3
aI5
aI3
aI11
aI41
aI69
aI199
aI79
aI611
aI133
aI3519
aI5955
aI4609
aI27403
asS'd'
I3639
sa(dp7278
S'a'
I16150
sS's'
I15
sS'm'
(lp7279
I1
aI3
aI3
aI13
aI3
aI17
aI53
aI165
aI361
aI797
aI1447
aI869
aI6707
aI6541
aI32249
asS'd'
I3640
sa(dp7280
S'a'
I16153
sS's'
I15
sS'm'
(lp7281
I1
aI3
aI1
aI1
aI29
aI47
aI17
aI45
aI473
aI199
aI1595
aI3095
aI3635
aI6965
aI21859
asS'd'
I3641
sa(dp7282
S'a'
I16160
sS's'
I15
sS'm'
(lp7283
I1
aI1
aI3
aI9
aI1
aI15
aI59
aI163
aI91
aI811
aI1087
aI1707
aI6743
aI12643
aI29901
asS'd'
I3642
sa(dp7284
S'a'
I16166
sS's'
I15
sS'm'
(lp7285
I1
aI1
aI1
aI3
aI19
aI21
aI7
aI209
aI121
aI821
aI709
aI1085
aI5333
aI7689
aI28355
asS'd'
I3643
sa(dp7286
S'a'
I16172
sS's'
I15
sS'm'
(lp7287
I1
aI3
aI1
aI15
aI5
aI27
aI115
aI31
aI37
aI79
aI1347
aI155
aI3709
aI13251
aI32151
asS'd'
I3644
sa(dp7288
S'a'
I16190
sS's'
I15
sS'm'
(lp7289
I1
aI3
aI7
aI15
aI27
aI27
aI127
aI231
aI137
aI205
aI1665
aI1461
aI299
aI2797
aI879
asS'd'
I3645
sa(dp7290
S'a'
I16195
sS's'
I15
sS'm'
(lp7291
I1
aI1
aI1
aI7
aI13
aI3
aI127
aI13
aI253
aI481
aI1435
aI1895
aI2665
aI7611
aI17761
asS'd'
I3646
sa(dp7292
S'a'
I16204
sS's'
I15
sS'm'
(lp7293
I1
aI1
aI3
aI7
aI7
aI21
aI71
aI247
aI301
aI183
aI1785
aI331
aI4835
aI2251
aI4493
asS'd'
I3647
sa(dp7294
S'a'
I16216
sS's'
I15
sS'm'
(lp7295
I1
aI3
aI7
aI9
aI9
aI1
aI77
aI169
aI103
aI647
aI1959
aI1847
aI5803
aI3421
aI15915
asS'd'
I3648
sa(dp7296
S'a'
I16222
sS's'
I15
sS'm'
(lp7297
I1
aI3
aI1
aI7
aI19
aI17
aI81
aI45
aI263
aI549
aI1607
aI2177
aI1117
aI14427
aI16451
asS'd'
I3649
sa(dp7298
S'a'
I16228
sS's'
I15
sS'm'
(lp7299
I1
aI1
aI7
aI15
aI27
aI25
aI27
aI27
aI33
aI813
aI1667
aI253
aI2749
aI927
aI29707
asS'd'
I3650
sa(dp7300
S'a'
I16245
sS's'
I15
sS'm'
(lp7301
I1
aI1
aI7
aI3
aI17
aI29
aI13
aI67
aI417
aI303
aI19
aI3809
aI7225
aI12775
aI3933
asS'd'
I3651
sa(dp7302
S'a'
I16255
sS's'
I15
sS'm'
(lp7303
I1
aI1
aI1
aI11
aI13
aI41
aI77
aI217
aI281
aI659
aI1099
aI3047
aI1619
aI525
aI4313
asS'd'
I3652
sa(dp7304
S'a'
I16265
sS's'
I15
sS'm'
(lp7305
I1
aI3
aI3
aI9
aI23
aI47
aI5
aI33
aI219
aI531
aI77
aI2307
aI1893
aI8335
aI8281
asS'd'
I3653
sa(dp7306
S'a'
I16273
sS's'
I15
sS'm'
(lp7307
I1
aI3
aI7
aI3
aI3
aI35
aI27
aI249
aI159
aI495
aI431
aI3001
aI1475
aI11505
aI15693
asS'd'
I3654
sa(dp7308
S'a'
I16276
sS's'
I15
sS'm'
(lp7309
I1
aI1
aI5
aI9
aI21
aI49
aI43
aI159
aI465
aI959
aI179
aI993
aI121
aI11569
aI21027
asS'd'
I3655
sa(dp7310
S'a'
I16283
sS's'
I15
sS'm'
(lp7311
I1
aI3
aI1
aI5
aI1
aI61
aI9
aI221
aI231
aI55
aI191
aI2829
aI3331
aI8911
aI15109
asS'd'
I3656
sa(dp7312
S'a'
I16295
sS's'
I15
sS'm'
(lp7313
I1
aI1
aI7
aI1
aI7
aI35
aI67
aI97
aI159
aI191
aI935
aI3151
aI6397
aI10751
aI1835
asS'd'
I3657
sa(dp7314
S'a'
I16304
sS's'
I15
sS'm'
(lp7315
I1
aI1
aI1
aI7
aI15
aI39
aI127
aI163
aI437
aI333
aI829
aI753
aI8151
aI13239
aI523
asS'd'
I3658
sa(dp7316
S'a'
I16313
sS's'
I15
sS'm'
(lp7317
I1
aI1
aI3
aI13
aI9
aI25
aI73
aI155
aI445
aI239
aI2035
aI15
aI5243
aI15531
aI1733
asS'd'
I3659
sa(dp7318
S'a'
I16319
sS's'
I15
sS'm'
(lp7319
I1
aI3
aI7
aI15
aI5
aI25
aI3
aI55
aI117
aI57
aI783
aI1509
aI7043
aI13159
aI8557
asS'd'
I3660
sa(dp7320
S'a'
I16328
sS's'
I15
sS'm'
(lp7321
I1
aI3
aI5
aI1
aI21
aI55
aI89
aI119
aI199
aI79
aI161
aI1597
aI3263
aI3335
aI5757
asS'd'
I3661
sa(dp7322
S'a'
I16345
sS's'
I15
sS'm'
(lp7323
I1
aI3
aI7
aI5
aI27
aI23
aI85
aI113
aI111
aI211
aI389
aI1513
aI2759
aI7945
aI931
asS'd'
I3662
sa(dp7324
S'a'
I16355
sS's'
I15
sS'm'
(lp7325
I1
aI1
aI1
aI7
aI1
aI5
aI17
aI177
aI357
aI619
aI5
aI2583
aI621
aI2973
aI28845
asS'd'
I3663
sa(dp7326
S'a'
I16364
sS's'
I15
sS'm'
(lp7327
I1
aI3
aI7
aI13
aI11
aI21
aI47
aI99
aI421
aI279
aI1541
aI1305
aI4571
aI6127
aI20735
asS'd'
I3664
sa(dp7328
S'a'
I16372
sS's'
I15
sS'm'
(lp7329
I1
aI3
aI5
aI5
aI23
aI43
aI19
aI137
aI425
aI409
aI1625
aI2671
aI4385
aI3197
aI25753
asS'd'
I3665
sa(dp7330
S'a'
I16375
sS's'
I15
sS'm'
(lp7331
I1
aI1
aI7
aI5
aI27
aI17
aI57
aI15
aI383
aI181
aI951
aI2115
aI5237
aI1495
aI9671
asS'd'
I3666
sa(dp7332
S'a'
I16382
sS's'
I15
sS'm'
(lp7333
I1
aI3
aI3
aI11
aI9
aI1
aI53
aI127
aI375
aI499
aI1487
aI121
aI1465
aI3175
aI24337
asS'd'
I3667
sa(dp7334
S'a'
I22
sS's'
I16
sS'm'
(lp7335
I1
aI3
aI7
aI11
aI29
aI35
aI67
aI129
aI221
aI439
aI1159
aI3501
aI7741
aI8885
aI11381
aI20707
asS'd'
I3668
sa(dp7336
S'a'
I28
sS's'
I16
sS'm'
(lp7337
I1
aI3
aI5
aI11
aI29
aI59
aI23
aI117
aI343
aI637
aI1825
aI1687
aI2823
aI11641
aI3311
aI23603
asS'd'
I3669
sa(dp7338
S'a'
I31
sS's'
I16
sS'm'
(lp7339
I1
aI1
aI5
aI11
aI1
aI35
aI103
aI155
aI233
aI575
aI1761
aI503
aI4175
aI6105
aI29737
aI32681
asS'd'
I3670
sa(dp7340
S'a'
I41
sS's'
I16
sS'm'
(lp7341
I1
aI3
aI3
aI1
aI5
aI63
aI27
aI71
aI245
aI433
aI1779
aI2475
aI5479
aI4705
aI10795
aI34247
asS'd'
I3671
sa(dp7342
S'a'
I94
sS's'
I16
sS'm'
(lp7343
I1
aI3
aI5
aI7
aI29
aI45
aI117
aI5
aI393
aI849
aI843
aI3131
aI6995
aI9979
aI28907
aI30115
asS'd'
I3672
sa(dp7344
S'a'
I107
sS's'
I16
sS'm'
(lp7345
I1
aI3
aI5
aI9
aI27
aI29
aI69
aI5
aI395
aI561
aI1531
aI409
aI2779
aI8785
aI16405
aI27315
asS'd'
I3673
sa(dp7346
S'a'
I151
sS's'
I16
sS'm'
(lp7347
I1
aI3
aI1
aI9
aI15
aI29
aI85
aI3
aI331
aI19
aI1941
aI567
aI6957
aI747
aI1627
aI11347
asS'd'
I3674
sa(dp7348
S'a'
I158
sS's'
I16
sS'm'
(lp7349
I1
aI1
aI3
aI9
aI27
aI45
aI47
aI127
aI133
aI921
aI1817
aI2231
aI6333
aI14371
aI12799
aI9831
asS'd'
I3675
sa(dp7350
S'a'
I167
sS's'
I16
sS'm'
(lp7351
I1
aI1
aI5
aI15
aI31
aI7
aI125
aI13
aI455
aI159
aI331
aI3629
aI4705
aI11261
aI3657
aI36307
asS'd'
I3676
sa(dp7352
S'a'
I174
sS's'
I16
sS'm'
(lp7353
I1
aI1
aI5
aI9
aI11
aI53
aI51
aI35
aI87
aI885
aI1975
aI3899
aI1013
aI7667
aI32385
aI33635
asS'd'
I3677
sa(dp7354
S'a'
I203
sS's'
I16
sS'm'
(lp7355
I1
aI1
aI1
aI3
aI7
aI45
aI107
aI177
aI193
aI765
aI731
aI139
aI5563
aI623
aI16485
aI54999
asS'd'
I3678
sa(dp7356
S'a'
I208
sS's'
I16
sS'm'
(lp7357
I1
aI1
aI5
aI9
aI17
aI53
aI117
aI69
aI385
aI587
aI1483
aI149
aI2769
aI3013
aI18183
aI10173
asS'd'
I3679
sa(dp7358
S'a'
I214
sS's'
I16
sS'm'
(lp7359
I1
aI1
aI5
aI11
aI5
aI3
aI25
aI153
aI351
aI749
aI801
aI3077
aI3209
aI11189
aI25241
aI14115
asS'd'
I3680
sa(dp7360
S'a'
I223
sS's'
I16
sS'm'
(lp7361
I1
aI1
aI7
aI9
aI1
aI47
aI41
aI247
aI135
aI163
aI899
aI1517
aI5647
aI10595
aI32531
aI12497
asS'd'
I3681
sa(dp7362
S'a'
I227
sS's'
I16
sS'm'
(lp7363
I1
aI3
aI5
aI11
aI5
aI61
aI111
aI215
aI251
aI279
aI825
aI2155
aI3527
aI173
aI10973
aI59257
asS'd'
I3682
sa(dp7364
S'a'
I266
sS's'
I16
sS'm'
(lp7365
I1
aI3
aI5
aI11
aI25
aI15
aI71
aI83
aI135
aI231
aI1415
aI3761
aI7513
aI8337
aI28979
aI43615
asS'd'
I3683
sa(dp7366
S'a'
I268
sS's'
I16
sS'm'
(lp7367
I1
aI3
aI5
aI13
aI19
aI5
aI55
aI165
aI141
aI119
aI1891
aI2255
aI4735
aI16217
aI26195
aI50527
asS'd'
I3684
sa(dp7368
S'a'
I274
sS's'
I16
sS'm'
(lp7369
I1
aI1
aI7
aI15
aI23
aI59
aI59
aI191
aI1
aI855
aI453
aI2619
aI5013
aI14749
aI24335
aI44339
asS'd'
I3685
sa(dp7370
S'a'
I279
sS's'
I16
sS'm'
(lp7371
I1
aI1
aI1
aI13
aI15
aI41
aI51
aI147
aI229
aI495
aI1191
aI867
aI1525
aI581
aI29713
aI26391
asS'd'
I3686
sa(dp7372
S'a'
I302
sS's'
I16
sS'm'
(lp7373
I1
aI1
aI1
aI9
aI29
aI5
aI59
aI127
aI105
aI417
aI301
aI2249
aI6335
aI3513
aI17373
aI52977
asS'd'
I3687
sa(dp7374
S'a'
I310
sS's'
I16
sS'm'
(lp7375
I1
aI1
aI3
aI7
aI21
aI27
aI109
aI143
aI63
aI347
aI1429
aI2889
aI2597
aI10243
aI9913
aI22687
asS'd'
I3688
sa(dp7376
S'a'
I322
sS's'
I16
sS'm'
(lp7377
I1
aI3
aI5
aI5
aI7
aI3
aI125
aI147
aI313
aI351
aI1163
aI415
aI5615
aI5571
aI7089
aI55621
asS'd'
I3689
sa(dp7378
S'a'
I328
sS's'
I16
sS'm'
(lp7379
I1
aI3
aI3
aI3
aI31
aI43
aI101
aI93
aI9
aI671
aI135
aI333
aI2169
aI11169
aI7403
aI50707
asS'd'
I3690
sa(dp7380
S'a'
I336
sS's'
I16
sS'm'
(lp7381
I1
aI1
aI7
aI13
aI15
aI33
aI125
aI155
aI227
aI827
aI1047
aI2441
aI3007
aI10881
aI19969
aI63805
asS'd'
I3691
sa(dp7382
S'a'
I370
sS's'
I16
sS'm'
(lp7383
I1
aI3
aI3
aI5
aI31
aI33
aI29
aI249
aI159
aI797
aI1475
aI841
aI6933
aI6417
aI25629
aI61865
asS'd'
I3692
sa(dp7384
S'a'
I398
sS's'
I16
sS'm'
(lp7385
I1
aI3
aI3
aI15
aI11
aI55
aI11
aI117
aI149
aI911
aI1589
aI3133
aI6477
aI6123
aI10471
aI41099
asS'd'
I3693
sa(dp7386
S'a'
I421
sS's'
I16
sS'm'
(lp7387
I1
aI3
aI3
aI9
aI27
aI37
aI1
aI119
aI509
aI969
aI831
aI3771
aI2093
aI13621
aI31737
aI43269
asS'd'
I3694
sa(dp7388
S'a'
I436
sS's'
I16
sS'm'
(lp7389
I1
aI1
aI1
aI1
aI9
aI23
aI119
aI109
aI487
aI753
aI1673
aI2163
aI3349
aI4741
aI29971
aI3407
asS'd'
I3695
sa(dp7390
S'a'
I440
sS's'
I16
sS'm'
(lp7391
I1
aI3
aI3
aI7
aI25
aI7
aI67
aI9
aI461
aI631
aI651
aI2271
aI5663
aI2621
aI3953
aI20975
asS'd'
I3696
sa(dp7392
S'a'
I451
sS's'
I16
sS'm'
(lp7393
I1
aI1
aI5
aI11
aI13
aI31
aI29
aI255
aI371
aI517
aI845
aI3649
aI1187
aI10061
aI22887
aI58417
asS'd'
I3697
sa(dp7394
S'a'
I454
sS's'
I16
sS'm'
(lp7395
I1
aI3
aI5
aI13
aI29
aI1
aI11
aI137
aI151
aI249
aI167
aI1243
aI997
aI11023
aI11875
aI42315
asS'd'
I3698
sa(dp7396
S'a'
I463
sS's'
I16
sS'm'
(lp7397
I1
aI1
aI5
aI5
aI5
aI55
aI103
aI71
aI255
aI1023
aI209
aI1005
aI2147
aI11527
aI17863
aI6661
asS'd'
I3699
sa(dp7398
S'a'
I465
sS's'
I16
sS'm'
(lp7399
I1
aI1
aI3
aI3
aI31
aI39
aI7
aI151
aI353
aI775
aI1313
aI1257
aI4197
aI2625
aI9571
aI27269
asS'd'
I3700
sa(dp7400
S'a'
I494
sS's'
I16
sS'm'
(lp7401
I1
aI1
aI1
aI3
aI7
aI17
aI3
aI127
aI501
aI503
aI1879
aI2329
aI3049
aI10603
aI2111
aI33189
asS'd'
I3701
sa(dp7402
S'a'
I508
sS's'
I16
sS'm'
(lp7403
I1
aI3
aI3
aI7
aI13
aI59
aI93
aI13
aI375
aI483
aI1991
aI2257
aI3003
aI1699
aI4339
aI51827
asS'd'
I3702
sa(dp7404
S'a'
I532
sS's'
I16
sS'm'
(lp7405
I1
aI3
aI7
aI15
aI27
aI41
aI59
aI225
aI405
aI807
aI1545
aI2581
aI1173
aI14137
aI3413
aI39299
asS'd'
I3703
sa(dp7406
S'a'
I555
sS's'
I16
sS'm'
(lp7407
I1
aI1
aI1
aI3
aI9
aI23
aI37
aI123
aI465
aI1023
aI1065
aI1455
aI5107
aI3839
aI20451
aI11461
asS'd'
I3704
sa(dp7408
S'a'
I563
sS's'
I16
sS'm'
(lp7409
I1
aI1
aI1
aI11
aI19
aI55
aI91
aI121
aI317
aI199
aI215
aI3031
aI7223
aI11891
aI21463
aI64921
asS'd'
I3705
sa(dp7410
S'a'
I577
sS's'
I16
sS'm'
(lp7411
I1
aI3
aI7
aI11
aI19
aI5
aI5
aI115
aI399
aI219
aI71
aI1465
aI281
aI14451
aI26807
aI42541
asS'd'
I3706
sa(dp7412
S'a'
I580
sS's'
I16
sS'm'
(lp7413
I1
aI3
aI5
aI13
aI3
aI33
aI75
aI35
aI19
aI559
aI761
aI947
aI7479
aI15325
aI31453
aI20561
asS'd'
I3707
sa(dp7414
S'a'
I584
sS's'
I16
sS'm'
(lp7415
I1
aI3
aI3
aI13
aI23
aI47
aI99
aI73
aI331
aI353
aI401
aI1737
aI6235
aI13781
aI5547
aI56443
asS'd'
I3708
sa(dp7416
S'a'
I607
sS's'
I16
sS'm'
(lp7417
I1
aI3
aI3
aI13
aI21
aI37
aI41
aI205
aI87
aI399
aI51
aI3175
aI7403
aI12875
aI21129
aI7079
asS'd'
I3709
sa(dp7418
S'a'
I608
sS's'
I16
sS'm'
(lp7419
I1
aI3
aI5
aI11
aI15
aI47
aI33
aI39
aI465
aI871
aI277
aI2351
aI695
aI1953
aI24293
aI20595
asS'd'
I3710
sa(dp7420
S'a'
I665
sS's'
I16
sS'm'
(lp7421
I1
aI1
aI7
aI11
aI13
aI15
aI115
aI59
aI469
aI715
aI191
aI1927
aI905
aI13463
aI29005
aI46789
asS'd'
I3711
sa(dp7422
S'a'
I675
sS's'
I16
sS'm'
(lp7423
I1
aI3
aI5
aI9
aI13
aI55
aI79
aI17
aI265
aI887
aI905
aI3985
aI6907
aI3379
aI20055
aI58569
asS'd'
I3712
sa(dp7424
S'a'
I692
sS's'
I16
sS'm'
(lp7425
I1
aI1
aI7
aI11
aI21
aI29
aI23
aI109
aI17
aI427
aI1623
aI2219
aI3857
aI3709
aI25033
aI63823
asS'd'
I3713
sa(dp7426
S'a'
I707
sS's'
I16
sS'm'
(lp7427
I1
aI3
aI5
aI15
aI19
aI27
aI113
aI15
aI25
aI63
aI1885
aI2693
aI5301
aI9385
aI14137
aI26097
asS'd'
I3714
sa(dp7428
S'a'
I737
sS's'
I16
sS'm'
(lp7429
I1
aI3
aI3
aI11
aI17
aI5
aI73
aI143
aI79
aI957
aI461
aI1709
aI4909
aI2285
aI18113
aI8401
asS'd'
I3715
sa(dp7430
S'a'
I750
sS's'
I16
sS'm'
(lp7431
I1
aI1
aI3
aI7
aI9
aI9
aI101
aI127
aI137
aI755
aI1359
aI1965
aI83
aI13335
aI27763
aI7941
asS'd'
I3716
sa(dp7432
S'a'
I757
sS's'
I16
sS'm'
(lp7433
I1
aI1
aI1
aI3
aI13
aI61
aI95
aI61
aI295
aI615
aI555
aI2163
aI8155
aI14043
aI21465
aI46741
asS'd'
I3717
sa(dp7434
S'a'
I800
sS's'
I16
sS'm'
(lp7435
I1
aI1
aI1
aI13
aI29
aI19
aI111
aI17
aI373
aI153
aI1703
aI2199
aI7209
aI15845
aI1879
aI7493
asS'd'
I3718
sa(dp7436
S'a'
I805
sS's'
I16
sS'm'
(lp7437
I1
aI3
aI1
aI13
aI21
aI51
aI49
aI51
aI255
aI151
aI207
aI1915
aI7629
aI2705
aI8739
aI7467
asS'd'
I3719
sa(dp7438
S'a'
I809
sS's'
I16
sS'm'
(lp7439
I1
aI3
aI7
aI5
aI21
aI21
aI23
aI193
aI467
aI739
aI519
aI2315
aI2953
aI10633
aI9163
aI6007
asS'd'
I3720
sa(dp7440
S'a'
I837
sS's'
I16
sS'm'
(lp7441
I1
aI3
aI1
aI5
aI23
aI19
aI23
aI247
aI93
aI297
aI1089
aI2349
aI4683
aI13609
aI7615
aI18647
asS'd'
I3721
sa(dp7442
S'a'
I865
sS's'
I16
sS'm'
(lp7443
I1
aI1
aI3
aI3
aI21
aI39
aI19
aI71
aI93
aI1
aI133
aI3531
aI7503
aI2819
aI24211
aI1739
asS'd'
I3722
sa(dp7444
S'a'
I949
sS's'
I16
sS'm'
(lp7445
I1
aI3
aI5
aI13
aI9
aI43
aI31
aI111
aI493
aI739
aI705
aI2715
aI3613
aI11877
aI27945
aI46053
asS'd'
I3723
sa(dp7446
S'a'
I950
sS's'
I16
sS'm'
(lp7447
I1
aI1
aI7
aI13
aI27
aI59
aI103
aI129
aI53
aI531
aI1379
aI1441
aI5341
aI14937
aI5079
aI39881
asS'd'
I3724
sa(dp7448
S'a'
I956
sS's'
I16
sS'm'
(lp7449
I1
aI1
aI3
aI3
aI11
aI63
aI91
aI95
aI433
aI393
aI715
aI809
aI591
aI4141
aI17417
aI54107
asS'd'
I3725
sa(dp7450
S'a'
I961
sS's'
I16
sS'm'
(lp7451
I1
aI3
aI5
aI1
aI7
aI25
aI25
aI175
aI205
aI803
aI183
aI1441
aI1279
aI2753
aI20001
aI56677
asS'd'
I3726
sa(dp7452
S'a'
I1016
sS's'
I16
sS'm'
(lp7453
I1
aI1
aI5
aI3
aI13
aI23
aI77
aI25
aI133
aI137
aI1907
aI1313
aI2463
aI14339
aI13
aI57757
asS'd'
I3727
sa(dp7454
S'a'
I1030
sS's'
I16
sS'm'
(lp7455
I1
aI1
aI5
aI9
aI23
aI35
aI1
aI119
aI111
aI61
aI403
aI1815
aI1985
aI5651
aI10883
aI55943
asS'd'
I3728
sa(dp7456
S'a'
I1072
sS's'
I16
sS'm'
(lp7457
I1
aI3
aI1
aI7
aI21
aI43
aI115
aI7
aI107
aI719
aI759
aI1521
aI467
aI8735
aI29785
aI63821
asS'd'
I3729
sa(dp7458
S'a'
I1119
sS's'
I16
sS'm'
(lp7459
I1
aI1
aI3
aI13
aI19
aI17
aI51
aI141
aI399
aI569
aI703
aI2221
aI2809
aI13355
aI1907
aI15837
asS'd'
I3730
sa(dp7460
S'a'
I1130
sS's'
I16
sS'm'
(lp7461
I1
aI1
aI5
aI15
aI15
aI53
aI57
aI31
aI481
aI69
aI1439
aI4049
aI6727
aI11307
aI20683
aI63517
asS'd'
I3731
sa(dp7462
S'a'
I1135
sS's'
I16
sS'm'
(lp7463
I1
aI1
aI1
aI3
aI13
aI27
aI9
aI255
aI363
aI131
aI1745
aI2489
aI6451
aI6585
aI12873
aI35405
asS'd'
I3732
sa(dp7464
S'a'
I1137
sS's'
I16
sS'm'
(lp7465
I1
aI3
aI5
aI1
aI17
aI31
aI113
aI135
aI449
aI915
aI1017
aI2317
aI6821
aI5483
aI30707
aI45279
asS'd'
I3733
sa(dp7466
S'a'
I1144
sS's'
I16
sS'm'
(lp7467
I1
aI3
aI5
aI1
aI13
aI47
aI25
aI53
aI413
aI545
aI1777
aI3049
aI7527
aI9689
aI25935
aI9919
asS'd'
I3734
sa(dp7468
S'a'
I1149
sS's'
I16
sS'm'
(lp7469
I1
aI3
aI7
aI11
aI17
aI39
aI13
aI131
aI295
aI517
aI1755
aI2977
aI6267
aI12351
aI8957
aI17765
asS'd'
I3735
sa(dp7470
S'a'
I1180
sS's'
I16
sS'm'
(lp7471
I1
aI1
aI7
aI5
aI27
aI57
aI47
aI21
aI125
aI429
aI1169
aI1717
aI5455
aI16359
aI29065
aI6671
asS'd'
I3736
sa(dp7472
S'a'
I1214
sS's'
I16
sS'm'
(lp7473
I1
aI1
aI5
aI5
aI21
aI15
aI79
aI241
aI83
aI515
aI859
aI2351
aI3125
aI7465
aI30475
aI19759
asS'd'
I3737
sa(dp7474
S'a'
I1221
sS's'
I16
sS'm'
(lp7475
I1
aI3
aI1
aI9
aI11
aI5
aI81
aI11
aI7
aI221
aI141
aI3329
aI3435
aI323
aI18999
aI54735
asS'd'
I3738
sa(dp7476
S'a'
I1234
sS's'
I16
sS'm'
(lp7477
I1
aI1
aI1
aI15
aI7
aI57
aI87
aI251
aI63
aI561
aI929
aI1367
aI2511
aI14527
aI9335
aI38775
asS'd'
I3739
sa(dp7478
S'a'
I1239
sS's'
I16
sS'm'
(lp7479
I1
aI3
aI3
aI9
aI23
aI37
aI59
aI105
aI179
aI515
aI235
aI2445
aI433
aI13039
aI27005
aI48829
asS'd'
I3740
sa(dp7480
S'a'
I1249
sS's'
I16
sS'm'
(lp7481
I1
aI1
aI1
aI1
aI23
aI37
aI103
aI31
aI89
aI921
aI1687
aI831
aI387
aI10237
aI1241
aI19295
asS'd'
I3741
sa(dp7482
S'a'
I1250
sS's'
I16
sS'm'
(lp7483
I1
aI3
aI3
aI7
aI25
aI23
aI57
aI251
aI309
aI579
aI603
aI807
aI7383
aI8579
aI4025
aI16757
asS'd'
I3742
sa(dp7484
S'a'
I1267
sS's'
I16
sS'm'
(lp7485
I1
aI1
aI3
aI15
aI23
aI59
aI29
aI33
aI467
aI641
aI1271
aI2915
aI2549
aI14767
aI26557
aI43483
asS'd'
I3743
sa(dp7486
S'a'
I1273
sS's'
I16
sS'm'
(lp7487
I1
aI1
aI7
aI13
aI1
aI57
aI23
aI129
aI321
aI75
aI189
aI4087
aI5011
aI4355
aI25759
aI37153
asS'd'
I3744
sa(dp7488
S'a'
I1342
sS's'
I16
sS'm'
(lp7489
I1
aI1
aI5
aI1
aI21
aI57
aI25
aI183
aI37
aI669
aI259
aI1381
aI877
aI10245
aI16643
aI61035
asS'd'
I3745
sa(dp7490
S'a'
I1344
sS's'
I16
sS'm'
(lp7491
I1
aI1
aI7
aI5
aI11
aI11
aI85
aI141
aI393
aI957
aI1745
aI2243
aI1681
aI5583
aI16527
aI12017
asS'd'
I3746
sa(dp7492
S'a'
I1373
sS's'
I16
sS'm'
(lp7493
I1
aI1
aI5
aI15
aI23
aI31
aI5
aI169
aI287
aI527
aI1831
aI2937
aI7533
aI9739
aI24305
aI2239
asS'd'
I3747
sa(dp7494
S'a'
I1378
sS's'
I16
sS'm'
(lp7495
I1
aI1
aI7
aI1
aI7
aI13
aI3
aI243
aI189
aI309
aI607
aI3659
aI6369
aI7649
aI24255
aI55373
asS'd'
I3748
sa(dp7496
S'a'
I1408
sS's'
I16
sS'm'
(lp7497
I1
aI1
aI1
aI3
aI3
aI59
aI103
aI209
aI287
aI913
aI1223
aI1063
aI7715
aI6073
aI26697
aI25671
asS'd'
I3749
sa(dp7498
S'a'
I1417
sS's'
I16
sS'm'
(lp7499
I1
aI3
aI7
aI5
aI19
aI19
aI117
aI191
aI275
aI637
aI991
aI2199
aI2921
aI10553
aI21211
aI25981
asS'd'
I3750
sa(dp7500
S'a'
I1418
sS's'
I16
sS'm'
(lp7501
I1
aI3
aI3
aI5
aI29
aI59
aI17
aI13
aI127
aI57
aI1405
aI3181
aI2237
aI1795
aI21419
aI43421
asS'd'
I3751
sa(dp7502
S'a'
I1448
sS's'
I16
sS'm'
(lp7503
I1
aI1
aI1
aI15
aI25
aI41
aI11
aI117
aI463
aI425
aI305
aI1441
aI4307
aI7967
aI17529
aI4043
asS'd'
I3752
sa(dp7504
S'a'
I1454
sS's'
I16
sS'm'
(lp7505
I1
aI3
aI5
aI5
aI19
aI53
aI69
aI73
aI453
aI611
aI1583
aI1721
aI6303
aI10561
aI18527
aI48973
asS'd'
I3753
sa(dp7506
S'a'
I1510
sS's'
I16
sS'm'
(lp7507
I1
aI1
aI7
aI11
aI15
aI61
aI87
aI69
aI463
aI771
aI819
aI469
aI8165
aI8897
aI29657
aI55161
asS'd'
I3754
sa(dp7508
S'a'
I1513
sS's'
I16
sS'm'
(lp7509
I1
aI1
aI5
aI1
aI15
aI25
aI23
aI47
aI287
aI457
aI1219
aI473
aI4127
aI3349
aI9425
aI41541
asS'd'
I3755
sa(dp7510
S'a'
I1522
sS's'
I16
sS'm'
(lp7511
I1
aI3
aI7
aI5
aI17
aI17
aI33
aI161
aI239
aI231
aI241
aI1297
aI4879
aI12761
aI20939
aI65261
asS'd'
I3756
sa(dp7512
S'a'
I1543
sS's'
I16
sS'm'
(lp7513
I1
aI3
aI3
aI9
aI19
aI53
aI95
aI89
aI117
aI333
aI1815
aI2217
aI7779
aI8213
aI4667
aI58395
asS'd'
I3757
sa(dp7514
S'a'
I1550
sS's'
I16
sS'm'
(lp7515
I1
aI3
aI3
aI9
aI17
aI7
aI41
aI99
aI371
aI797
aI729
aI2851
aI2003
aI4463
aI20793
aI54315
asS'd'
I3758
sa(dp7516
S'a'
I1552
sS's'
I16
sS'm'
(lp7517
I1
aI3
aI5
aI5
aI23
aI39
aI19
aI235
aI163
aI365
aI141
aI791
aI455
aI2761
aI9115
aI53351
asS'd'
I3759
sa(dp7518
S'a'
I1588
sS's'
I16
sS'm'
(lp7519
I1
aI3
aI3
aI3
aI9
aI27
aI29
aI139
aI165
aI867
aI2023
aI1333
aI3771
aI10451
aI9141
aI41177
asS'd'
I3760
sa(dp7520
S'a'
I1592
sS's'
I16
sS'm'
(lp7521
I1
aI1
aI3
aI7
aI3
aI11
aI125
aI157
aI355
aI519
aI187
aI3381
aI1151
aI1629
aI25247
aI42797
asS'd'
I3761
sa(dp7522
S'a'
I1597
sS's'
I16
sS'm'
(lp7523
I1
aI3
aI3
aI3
aI21
aI25
aI37
aI155
aI257
aI311
aI961
aI1945
aI1383
aI5679
aI7857
aI7183
asS'd'
I3762
sa(dp7524
S'a'
I1606
sS's'
I16
sS'm'
(lp7525
I1
aI3
aI3
aI5
aI29
aI11
aI49
aI125
aI171
aI605
aI1923
aI2781
aI2555
aI5063
aI5075
aI43301
asS'd'
I3763
sa(dp7526
S'a'
I1610
sS's'
I16
sS'm'
(lp7527
I1
aI3
aI5
aI9
aI27
aI1
aI27
aI149
aI253
aI205
aI1299
aI2901
aI2891
aI975
aI7641
aI8115
asS'd'
I3764
sa(dp7528
S'a'
I1617
sS's'
I16
sS'm'
(lp7529
I1
aI3
aI5
aI3
aI31
aI7
aI49
aI215
aI81
aI791
aI1485
aI837
aI5051
aI1947
aI7521
aI25723
asS'd'
I3765
sa(dp7530
S'a'
I1623
sS's'
I16
sS'm'
(lp7531
I1
aI3
aI5
aI7
aI23
aI25
aI69
aI13
aI3
aI859
aI441
aI3577
aI1687
aI6559
aI8687
aI46757
asS'd'
I3766
sa(dp7532
S'a'
I1657
sS's'
I16
sS'm'
(lp7533
I1
aI1
aI1
aI9
aI1
aI59
aI3
aI31
aI251
aI187
aI617
aI2607
aI4635
aI6121
aI8565
aI8871
asS'd'
I3767
sa(dp7534
S'a'
I1697
sS's'
I16
sS'm'
(lp7535
I1
aI3
aI3
aI9
aI29
aI37
aI127
aI87
aI153
aI633
aI1691
aI2729
aI3167
aI3219
aI21237
aI25573
asS'd'
I3768
sa(dp7536
S'a'
I1729
sS's'
I16
sS'm'
(lp7537
I1
aI1
aI5
aI13
aI19
aI63
aI93
aI235
aI299
aI621
aI405
aI663
aI6639
aI12265
aI9303
aI42719
asS'd'
I3769
sa(dp7538
S'a'
I1735
sS's'
I16
sS'm'
(lp7539
I1
aI1
aI3
aI9
aI25
aI11
aI9
aI231
aI101
aI335
aI1793
aI1497
aI7069
aI4171
aI30199
aI63
asS'd'
I3770
sa(dp7540
S'a'
I1769
sS's'
I16
sS'm'
(lp7541
I1
aI1
aI1
aI1
aI5
aI19
aI17
aI217
aI165
aI413
aI925
aI1409
aI6559
aI14537
aI22057
aI44331
asS'd'
I3771
sa(dp7542
S'a'
I1778
sS's'
I16
sS'm'
(lp7543
I1
aI1
aI3
aI7
aI11
aI51
aI45
aI217
aI57
aI795
aI951
aI2933
aI6705
aI137
aI30525
aI9679
asS'd'
I3772
sa(dp7544
S'a'
I1826
sS's'
I16
sS'm'
(lp7545
I1
aI1
aI3
aI15
aI27
aI47
aI35
aI125
aI363
aI619
aI1027
aI2861
aI3923
aI10459
aI16789
aI27277
asS'd'
I3773
sa(dp7546
S'a'
I1858
sS's'
I16
sS'm'
(lp7547
I1
aI1
aI7
aI7
aI13
aI37
aI33
aI29
aI385
aI851
aI143
aI119
aI7345
aI4251
aI25121
aI31609
asS'd'
I3774
sa(dp7548
S'a'
I1870
sS's'
I16
sS'm'
(lp7549
I1
aI3
aI1
aI1
aI17
aI25
aI119
aI7
aI365
aI397
aI601
aI2087
aI6903
aI15345
aI14671
aI37889
asS'd'
I3775
sa(dp7550
S'a'
I1875
sS's'
I16
sS'm'
(lp7551
I1
aI3
aI1
aI13
aI19
aI51
aI41
aI139
aI133
aI723
aI25
aI2621
aI1257
aI7037
aI9527
aI50037
asS'd'
I3776
sa(dp7552
S'a'
I1922
sS's'
I16
sS'm'
(lp7553
I1
aI1
aI5
aI11
aI5
aI59
aI119
aI75
aI397
aI545
aI1095
aI585
aI3271
aI1049
aI123
aI33029
asS'd'
I3777
sa(dp7554
S'a'
I1924
sS's'
I16
sS'm'
(lp7555
I1
aI1
aI7
aI11
aI9
aI27
aI21
aI197
aI177
aI31
aI453
aI2457
aI2733
aI7787
aI1923
aI24639
asS'd'
I3778
sa(dp7556
S'a'
I1933
sS's'
I16
sS'm'
(lp7557
I1
aI1
aI7
aI13
aI29
aI13
aI91
aI91
aI243
aI279
aI601
aI1699
aI7169
aI4727
aI7815
aI29099
asS'd'
I3779
sa(dp7558
S'a'
I1972
sS's'
I16
sS'm'
(lp7559
I1
aI3
aI7
aI5
aI1
aI35
aI27
aI235
aI163
aI913
aI1479
aI769
aI7179
aI1983
aI25977
aI55373
asS'd'
I3780
sa(dp7560
S'a'
I1979
sS's'
I16
sS'm'
(lp7561
I1
aI3
aI5
aI11
aI9
aI33
aI99
aI141
aI301
aI109
aI1785
aI129
aI1707
aI5181
aI4797
aI9979
asS'd'
I3781
sa(dp7562
S'a'
I1987
sS's'
I16
sS'm'
(lp7563
I1
aI1
aI1
aI13
aI3
aI47
aI89
aI43
aI293
aI87
aI1689
aI3885
aI7747
aI5607
aI477
aI31887
asS'd'
I3782
sa(dp7564
S'a'
I1994
sS's'
I16
sS'm'
(lp7565
I1
aI1
aI5
aI1
aI9
aI21
aI73
aI37
aI45
aI621
aI1855
aI3691
aI4899
aI2191
aI13459
aI23543
asS'd'
I3783
sa(dp7566
S'a'
I2008
sS's'
I16
sS'm'
(lp7567
I1
aI1
aI1
aI1
aI7
aI39
aI61
aI125
aI341
aI905
aI213
aI1755
aI241
aI13407
aI8791
aI10165
asS'd'
I3784
sa(dp7568
S'a'
I2023
sS's'
I16
sS'm'
(lp7569
I1
aI1
aI1
aI1
aI19
aI31
aI79
aI19
aI55
aI875
aI1017
aI1787
aI4879
aI533
aI15029
aI52295
asS'd'
I3785
sa(dp7570
S'a'
I2029
sS's'
I16
sS'm'
(lp7571
I1
aI3
aI1
aI1
aI9
aI59
aI113
aI71
aI113
aI649
aI561
aI71
aI5253
aI783
aI7389
aI19361
asS'd'
I3786
sa(dp7572
S'a'
I2053
sS's'
I16
sS'm'
(lp7573
I1
aI1
aI1
aI11
aI5
aI39
aI61
aI225
aI291
aI907
aI795
aI1099
aI597
aI11829
aI15137
aI42865
asS'd'
I3787
sa(dp7574
S'a'
I2081
sS's'
I16
sS'm'
(lp7575
I1
aI3
aI1
aI5
aI25
aI11
aI71
aI155
aI271
aI309
aI1981
aI1253
aI463
aI1133
aI20833
aI48625
asS'd'
I3788
sa(dp7576
S'a'
I2087
sS's'
I16
sS'm'
(lp7577
I1
aI3
aI5
aI9
aI7
aI41
aI87
aI241
aI457
aI899
aI1493
aI3675
aI3025
aI10607
aI22569
aI52813
asS'd'
I3789
sa(dp7578
S'a'
I2094
sS's'
I16
sS'm'
(lp7579
I1
aI3
aI7
aI13
aI7
aI37
aI37
aI103
aI281
aI915
aI1259
aI4049
aI559
aI173
aI4123
aI63767
asS'd'
I3790
sa(dp7580
S'a'
I2111
sS's'
I16
sS'm'
(lp7581
I1
aI3
aI7
aI15
aI13
aI57
aI9
aI51
aI39
aI549
aI1431
aI2887
aI1081
aI4643
aI16331
aI14221
asS'd'
I3791
sa(dp7582
S'a'
I2113
sS's'
I16
sS'm'
(lp7583
I1
aI3
aI5
aI7
aI13
aI1
aI101
aI125
aI25
aI713
aI1423
aI513
aI3323
aI9951
aI7163
aI20969
asS'd'
I3792
sa(dp7584
S'a'
I2114
sS's'
I16
sS'm'
(lp7585
I1
aI1
aI7
aI15
aI11
aI25
aI25
aI3
aI47
aI531
aI1529
aI471
aI6191
aI10051
aI29671
aI49085
asS'd'
I3793
sa(dp7586
S'a'
I2123
sS's'
I16
sS'm'
(lp7587
I1
aI1
aI3
aI5
aI23
aI51
aI117
aI141
aI55
aI275
aI761
aI1923
aI6267
aI2291
aI3701
aI26615
asS'd'
I3794
sa(dp7588
S'a'
I2190
sS's'
I16
sS'm'
(lp7589
I1
aI1
aI7
aI9
aI15
aI19
aI111
aI65
aI137
aI373
aI1753
aI3591
aI1137
aI11639
aI28591
aI27265
asS'd'
I3795
sa(dp7590
S'a'
I2231
sS's'
I16
sS'm'
(lp7591
I1
aI3
aI1
aI15
aI29
aI5
aI67
aI13
aI425
aI961
aI453
aI2481
aI1407
aI3479
aI23303
aI30407
asS'd'
I3796
sa(dp7592
S'a'
I2276
sS's'
I16
sS'm'
(lp7593
I1
aI1
aI5
aI3
aI19
aI39
aI39
aI123
aI351
aI77
aI1339
aI1765
aI3767
aI1907
aI13627
aI23877
asS'd'
I3797
sa(dp7594
S'a'
I2285
sS's'
I16
sS'm'
(lp7595
I1
aI3
aI5
aI9
aI23
aI7
aI103
aI177
aI221
aI197
aI561
aI2121
aI7231
aI12053
aI30127
aI29849
asS'd'
I3798
sa(dp7596
S'a'
I2297
sS's'
I16
sS'm'
(lp7597
I1
aI1
aI5
aI7
aI15
aI1
aI3
aI123
aI197
aI493
aI171
aI2425
aI3865
aI4061
aI31883
aI2491
asS'd'
I3799
sa(dp7598
S'a'
I2336
sS's'
I16
sS'm'
(lp7599
I1
aI1
aI3
aI13
aI29
aI33
aI99
aI67
aI327
aI969
aI1793
aI1871
aI1839
aI13059
aI7605
aI16797
asS'd'
I3800
sa(dp7600
S'a'
I2345
sS's'
I16
sS'm'
(lp7601
I1
aI3
aI5
aI11
aI25
aI53
aI25
aI93
aI303
aI623
aI1889
aI1471
aI1213
aI14459
aI8527
aI25095
asS'd'
I3801
sa(dp7602
S'a'
I2353
sS's'
I16
sS'm'
(lp7603
I1
aI1
aI1
aI13
aI15
aI3
aI115
aI3
aI289
aI743
aI1855
aI359
aI2375
aI13765
aI19711
aI40765
asS'd'
I3802
sa(dp7604
S'a'
I2363
sS's'
I16
sS'm'
(lp7605
I1
aI1
aI7
aI11
aI27
aI51
aI85
aI163
aI219
aI871
aI637
aI2011
aI5981
aI587
aI17521
aI17333
asS'd'
I3803
sa(dp7606
S'a'
I2368
sS's'
I16
sS'm'
(lp7607
I1
aI3
aI5
aI1
aI21
aI59
aI49
aI39
aI305
aI513
aI2017
aI285
aI5817
aI13123
aI27765
aI46741
asS'd'
I3804
sa(dp7608
S'a'
I2373
sS's'
I16
sS'm'
(lp7609
I1
aI3
aI3
aI7
aI21
aI39
aI71
aI163
aI423
aI845
aI783
aI397
aI7319
aI10677
aI13407
aI47471
asS'd'
I3805
sa(dp7610
S'a'
I2391
sS's'
I16
sS'm'
(lp7611
I1
aI3
aI7
aI5
aI21
aI59
aI99
aI179
aI473
aI687
aI1393
aI723
aI2245
aI2933
aI25943
aI7769
asS'd'
I3806
sa(dp7612
S'a'
I2402
sS's'
I16
sS'm'
(lp7613
I1
aI1
aI5
aI9
aI5
aI45
aI71
aI189
aI165
aI555
aI643
aI2289
aI3133
aI12319
aI22209
aI1533
asS'd'
I3807
sa(dp7614
S'a'
I2413
sS's'
I16
sS'm'
(lp7615
I1
aI1
aI3
aI9
aI7
aI43
aI1
aI155
aI323
aI169
aI339
aI2561
aI4049
aI4953
aI5289
aI8783
asS'd'
I3808
sa(dp7616
S'a'
I2422
sS's'
I16
sS'm'
(lp7617
I1
aI3
aI1
aI11
aI15
aI5
aI25
aI201
aI267
aI891
aI561
aI501
aI575
aI15147
aI1743
aI45237
asS'd'
I3809
sa(dp7618
S'a'
I2425
sS's'
I16
sS'm'
(lp7619
I1
aI3
aI5
aI13
aI25
aI27
aI105
aI205
aI165
aI795
aI975
aI943
aI7413
aI10299
aI14839
aI54895
asS'd'
I3810
sa(dp7620
S'a'
I2461
sS's'
I16
sS'm'
(lp7621
I1
aI1
aI5
aI1
aI17
aI43
aI69
aI103
aI449
aI917
aI103
aI945
aI513
aI709
aI11647
aI28065
asS'd'
I3811
sa(dp7622
S'a'
I2462
sS's'
I16
sS'm'
(lp7623
I1
aI1
aI3
aI15
aI23
aI51
aI23
aI7
aI159
aI743
aI177
aI3457
aI415
aI1775
aI25353
aI36385
asS'd'
I3812
sa(dp7624
S'a'
I2490
sS's'
I16
sS'm'
(lp7625
I1
aI3
aI5
aI13
aI9
aI63
aI121
aI19
aI165
aI449
aI1523
aI1959
aI6901
aI12281
aI29149
aI45999
asS'd'
I3813
sa(dp7626
S'a'
I2492
sS's'
I16
sS'm'
(lp7627
I1
aI3
aI7
aI11
aI17
aI19
aI9
aI155
aI373
aI753
aI1313
aI2205
aI3571
aI16317
aI16151
aI15325
asS'd'
I3814
sa(dp7628
S'a'
I2510
sS's'
I16
sS'm'
(lp7629
I1
aI3
aI3
aI7
aI15
aI43
aI65
aI183
aI407
aI123
aI1151
aI375
aI3461
aI6673
aI12985
aI21005
asS'd'
I3815
sa(dp7630
S'a'
I2564
sS's'
I16
sS'm'
(lp7631
I1
aI3
aI7
aI7
aI9
aI1
aI87
aI247
aI489
aI123
aI1677
aI1947
aI7961
aI13497
aI27919
aI28993
asS'd'
I3816
sa(dp7632
S'a'
I2573
sS's'
I16
sS'm'
(lp7633
I1
aI3
aI3
aI7
aI19
aI21
aI95
aI227
aI217
aI133
aI69
aI1535
aI699
aI3521
aI29255
aI34733
asS'd'
I3817
sa(dp7634
S'a'
I2598
sS's'
I16
sS'm'
(lp7635
I1
aI3
aI5
aI3
aI7
aI57
aI45
aI251
aI407
aI81
aI1259
aI2425
aI2217
aI13097
aI12773
aI14643
asS'd'
I3818
sa(dp7636
S'a'
I2627
sS's'
I16
sS'm'
(lp7637
I1
aI1
aI1
aI11
aI23
aI37
aI13
aI229
aI467
aI591
aI1521
aI469
aI3763
aI2289
aI14233
aI24053
asS'd'
I3819
sa(dp7638
S'a'
I2633
sS's'
I16
sS'm'
(lp7639
I1
aI3
aI5
aI1
aI27
aI53
aI105
aI5
aI85
aI765
aI1973
aI2597
aI5725
aI1063
aI18145
aI961
asS'd'
I3820
sa(dp7640
S'a'
I2647
sS's'
I16
sS'm'
(lp7641
I1
aI3
aI7
aI1
aI21
aI47
aI115
aI95
aI403
aI3
aI1593
aI3379
aI7371
aI15553
aI12503
aI57979
asS'd'
I3821
sa(dp7642
S'a'
I2660
sS's'
I16
sS'm'
(lp7643
I1
aI1
aI3
aI1
aI1
aI35
aI121
aI29
aI379
aI245
aI919
aI2673
aI3503
aI14197
aI31193
aI8355
asS'd'
I3822
sa(dp7644
S'a'
I2664
sS's'
I16
sS'm'
(lp7645
I1
aI3
aI5
aI11
aI19
aI49
aI97
aI7
aI195
aI1013
aI1671
aI3415
aI2009
aI13389
aI4837
aI27453
asS'd'
I3823
sa(dp7646
S'a'
I2678
sS's'
I16
sS'm'
(lp7647
I1
aI1
aI5
aI13
aI9
aI15
aI115
aI97
aI463
aI449
aI303
aI2681
aI1215
aI12559
aI15685
aI21321
asS'd'
I3824
sa(dp7648
S'a'
I2684
sS's'
I16
sS'm'
(lp7649
I1
aI3
aI5
aI13
aI23
aI5
aI113
aI193
aI419
aI301
aI1121
aI317
aI5503
aI4683
aI25519
aI65
asS'd'
I3825
sa(dp7650
S'a'
I2691
sS's'
I16
sS'm'
(lp7651
I1
aI3
aI3
aI7
aI15
aI29
aI45
aI97
aI323
aI475
aI143
aI1173
aI4033
aI8939
aI31849
aI3575
asS'd'
I3826
sa(dp7652
S'a'
I2759
sS's'
I16
sS'm'
(lp7653
I1
aI1
aI7
aI7
aI21
aI1
aI101
aI143
aI197
aI409
aI855
aI1753
aI5211
aI3763
aI11139
aI37309
asS'd'
I3827
sa(dp7654
S'a'
I2768
sS's'
I16
sS'm'
(lp7655
I1
aI1
aI3
aI13
aI25
aI33
aI55
aI45
aI381
aI349
aI991
aI535
aI4823
aI3701
aI31629
aI48037
asS'd'
I3828
sa(dp7656
S'a'
I2773
sS's'
I16
sS'm'
(lp7657
I1
aI3
aI1
aI11
aI17
aI51
aI27
aI57
aI409
aI551
aI949
aI365
aI8093
aI10831
aI19697
aI39437
asS'd'
I3829
sa(dp7658
S'a'
I2794
sS's'
I16
sS'm'
(lp7659
I1
aI3
aI5
aI3
aI31
aI33
aI81
aI49
aI91
aI865
aI469
aI2115
aI377
aI8237
aI31907
aI38239
asS'd'
I3830
sa(dp7660
S'a'
I2813
sS's'
I16
sS'm'
(lp7661
I1
aI1
aI3
aI7
aI29
aI59
aI57
aI17
aI121
aI889
aI1557
aI1797
aI5001
aI14209
aI21355
aI59739
asS'd'
I3831
sa(dp7662
S'a'
I2831
sS's'
I16
sS'm'
(lp7663
I1
aI1
aI5
aI9
aI11
aI45
aI89
aI87
aI397
aI785
aI525
aI1593
aI5251
aI12449
aI23579
aI54265
asS'd'
I3832
sa(dp7664
S'a'
I2843
sS's'
I16
sS'm'
(lp7665
I1
aI3
aI5
aI11
aI5
aI31
aI19
aI47
aI207
aI331
aI91
aI1691
aI5171
aI53
aI15945
aI33349
asS'd'
I3833
sa(dp7666
S'a'
I2846
sS's'
I16
sS'm'
(lp7667
I1
aI1
aI1
aI15
aI11
aI41
aI91
aI177
aI505
aI871
aI815
aI3673
aI5631
aI9915
aI1133
aI37861
asS'd'
I3834
sa(dp7668
S'a'
I2849
sS's'
I16
sS'm'
(lp7669
I1
aI3
aI5
aI5
aI25
aI63
aI53
aI231
aI55
aI51
aI481
aI303
aI1859
aI11973
aI28557
aI22045
asS'd'
I3835
sa(dp7670
S'a'
I2856
sS's'
I16
sS'm'
(lp7671
I1
aI1
aI5
aI3
aI27
aI11
aI37
aI91
aI363
aI411
aI1131
aI3369
aI377
aI6585
aI7353
aI42949
asS'd'
I3836
sa(dp7672
S'a'
I2893
sS's'
I16
sS'm'
(lp7673
I1
aI3
aI1
aI9
aI31
aI63
aI83
aI23
aI405
aI941
aI119
aI1471
aI2509
aI15507
aI29239
aI49613
asS'd'
I3837
sa(dp7674
S'a'
I2901
sS's'
I16
sS'm'
(lp7675
I1
aI1
aI5
aI1
aI11
aI63
aI117
aI237
aI407
aI231
aI1425
aI71
aI8005
aI4023
aI9029
aI59819
asS'd'
I3838
sa(dp7676
S'a'
I2924
sS's'
I16
sS'm'
(lp7677
I1
aI1
aI5
aI7
aI1
aI9
aI43
aI87
aI351
aI63
aI1075
aI3381
aI5447
aI2437
aI24983
aI26905
asS'd'
I3839
sa(dp7678
S'a'
I2942
sS's'
I16
sS'm'
(lp7679
I1
aI3
aI7
aI5
aI5
aI35
aI33
aI89
aI251
aI819
aI1735
aI2625
aI6363
aI6837
aI27603
aI26669
asS'd'
I3840
sa(dp7680
S'a'
I2975
sS's'
I16
sS'm'
(lp7681
I1
aI3
aI7
aI13
aI29
aI63
aI51
aI245
aI371
aI791
aI907
aI3499
aI3033
aI8443
aI20023
aI1805
asS'd'
I3841
sa(dp7682
S'a'
I2979
sS's'
I16
sS'm'
(lp7683
I1
aI1
aI5
aI7
aI13
aI15
aI109
aI197
aI451
aI709
aI929
aI3193
aI5727
aI11185
aI29479
aI1671
asS'd'
I3842
sa(dp7684
S'a'
I2985
sS's'
I16
sS'm'
(lp7685
I1
aI1
aI7
aI13
aI19
aI23
aI97
aI9
aI359
aI635
aI777
aI39
aI893
aI2531
aI13563
aI19295
asS'd'
I3843
sa(dp7686
S'a'
I3020
sS's'
I16
sS'm'
(lp7687
I1
aI1
aI5
aI1
aI31
aI63
aI55
aI7
aI157
aI877
aI991
aI1317
aI1595
aI2019
aI21435
aI52255
asS'd'
I3844
sa(dp7688
S'a'
I3025
sS's'
I16
sS'm'
(lp7689
I1
aI1
aI5
aI3
aI19
aI37
aI23
aI13
aI335
aI431
aI483
aI615
aI2431
aI505
aI26245
aI63323
asS'd'
I3845
sa(dp7690
S'a'
I3028
sS's'
I16
sS'm'
(lp7691
I1
aI3
aI7
aI5
aI5
aI9
aI37
aI65
aI303
aI423
aI1907
aI2661
aI7213
aI2975
aI29045
aI16243
asS'd'
I3846
sa(dp7692
S'a'
I3051
sS's'
I16
sS'm'
(lp7693
I1
aI3
aI1
aI5
aI13
aI37
aI115
aI217
aI227
aI159
aI707
aI1387
aI943
aI4935
aI5503
aI35171
asS'd'
I3847
sa(dp7694
S'a'
I3127
sS's'
I16
sS'm'
(lp7695
I1
aI3
aI7
aI9
aI19
aI15
aI87
aI233
aI453
aI159
aI169
aI1077
aI2129
aI413
aI19773
aI629
asS'd'
I3848
sa(dp7696
S'a'
I3142
sS's'
I16
sS'm'
(lp7697
I1
aI1
aI5
aI15
aI29
aI39
aI37
aI243
aI233
aI365
aI1843
aI2219
aI1255
aI15287
aI603
aI13511
asS'd'
I3849
sa(dp7698
S'a'
I3145
sS's'
I16
sS'm'
(lp7699
I1
aI1
aI3
aI3
aI31
aI53
aI33
aI125
aI497
aI597
aI127
aI1829
aI3905
aI2611
aI4263
aI40971
asS'd'
I3850
sa(dp7700
S'a'
I3156
sS's'
I16
sS'm'
(lp7701
I1
aI3
aI5
aI9
aI11
aI47
aI71
aI215
aI383
aI321
aI1445
aI135
aI5953
aI8791
aI22073
aI16537
asS'd'
I3851
sa(dp7702
S'a'
I3165
sS's'
I16
sS'm'
(lp7703
I1
aI3
aI3
aI13
aI15
aI7
aI7
aI133
aI401
aI459
aI1117
aI3165
aI4105
aI11943
aI22431
aI56821
asS'd'
I3852
sa(dp7704
S'a'
I3196
sS's'
I16
sS'm'
(lp7705
I1
aI1
aI7
aI9
aI31
aI39
aI19
aI7
aI19
aI401
aI79
aI3641
aI6815
aI1489
aI7537
aI49467
asS'd'
I3853
sa(dp7706
S'a'
I3199
sS's'
I16
sS'm'
(lp7707
I1
aI3
aI7
aI7
aI17
aI11
aI91
aI205
aI251
aI321
aI515
aI3521
aI311
aI3169
aI271
aI34749
asS'd'
I3854
sa(dp7708
S'a'
I3217
sS's'
I16
sS'm'
(lp7709
I1
aI3
aI3
aI7
aI29
aI15
aI5
aI153
aI83
aI603
aI1373
aI997
aI4939
aI9811
aI243
aI5375
asS'd'
I3855
sa(dp7710
S'a'
I3218
sS's'
I16
sS'm'
(lp7711
I1
aI1
aI3
aI11
aI21
aI47
aI25
aI221
aI237
aI177
aI535
aI2819
aI6213
aI7877
aI26795
aI36609
asS'd'
I3856
sa(dp7712
S'a'
I3253
sS's'
I16
sS'm'
(lp7713
I1
aI3
aI7
aI3
aI31
aI1
aI69
aI73
aI47
aI653
aI139
aI1649
aI7183
aI1293
aI26507
aI38415
asS'd'
I3857
sa(dp7714
S'a'
I3258
sS's'
I16
sS'm'
(lp7715
I1
aI1
aI1
aI13
aI17
aI41
aI23
aI73
aI115
aI509
aI787
aI3733
aI1871
aI171
aI29967
aI39941
asS'd'
I3858
sa(dp7716
S'a'
I3260
sS's'
I16
sS'm'
(lp7717
I1
aI3
aI5
aI1
aI9
aI7
aI61
aI23
aI105
aI381
aI1421
aI2887
aI3717
aI643
aI26375
aI57991
asS'd'
I3859
sa(dp7718
S'a'
I3289
sS's'
I16
sS'm'
(lp7719
I1
aI3
aI3
aI3
aI19
aI3
aI101
aI117
aI393
aI83
aI1255
aI3331
aI6481
aI8661
aI20855
aI28875
asS'd'
I3860
sa(dp7720
S'a'
I3314
sS's'
I16
sS'm'
(lp7721
I1
aI3
aI5
aI11
aI21
aI13
aI111
aI193
aI51
aI899
aI159
aI1989
aI7931
aI10511
aI3933
aI447
asS'd'
I3861
sa(dp7722
S'a'
I3326
sS's'
I16
sS'm'
(lp7723
I1
aI1
aI5
aI15
aI23
aI35
aI49
aI139
aI397
aI145
aI597
aI1847
aI7077
aI715
aI20227
aI42183
asS'd'
I3862
sa(dp7724
S'a'
I3331
sS's'
I16
sS'm'
(lp7725
I1
aI3
aI3
aI3
aI17
aI3
aI87
aI233
aI35
aI317
aI337
aI237
aI6901
aI3439
aI20033
aI10307
asS'd'
I3863
sa(dp7726
S'a'
I3371
sS's'
I16
sS'm'
(lp7727
I1
aI3
aI5
aI3
aI11
aI35
aI13
aI171
aI7
aI963
aI1443
aI1501
aI7617
aI963
aI25453
aI62589
asS'd'
I3864
sa(dp7728
S'a'
I3381
sS's'
I16
sS'm'
(lp7729
I1
aI1
aI1
aI5
aI11
aI9
aI39
aI175
aI409
aI411
aI1407
aI2743
aI4255
aI989
aI15823
aI1707
asS'd'
I3865
sa(dp7730
S'a'
I3396
sS's'
I16
sS'm'
(lp7731
I1
aI1
aI7
aI13
aI27
aI55
aI63
aI239
aI355
aI417
aI2007
aI2299
aI2921
aI1637
aI10687
aI60615
asS'd'
I3866
sa(dp7732
S'a'
I3441
sS's'
I16
sS'm'
(lp7733
I1
aI1
aI7
aI9
aI5
aI61
aI57
aI73
aI263
aI307
aI2003
aI1763
aI639
aI5885
aI14709
aI16985
asS'd'
I3867
sa(dp7734
S'a'
I3442
sS's'
I16
sS'm'
(lp7735
I1
aI1
aI3
aI3
aI21
aI55
aI19
aI249
aI509
aI533
aI1361
aI1397
aI2777
aI15523
aI4389
aI13339
asS'd'
I3868
sa(dp7736
S'a'
I3460
sS's'
I16
sS'm'
(lp7737
I1
aI3
aI5
aI15
aI9
aI3
aI91
aI237
aI451
aI299
aI1541
aI4083
aI879
aI7859
aI21585
aI14833
asS'd'
I3869
sa(dp7738
S'a'
I3477
sS's'
I16
sS'm'
(lp7739
I1
aI1
aI7
aI3
aI31
aI47
aI49
aI231
aI123
aI391
aI1633
aI2567
aI5577
aI1631
aI27951
aI22913
asS'd'
I3870
sa(dp7740
S'a'
I3491
sS's'
I16
sS'm'
(lp7741
I1
aI3
aI7
aI13
aI11
aI13
aI1
aI111
aI183
aI87
aI839
aI1915
aI5523
aI3677
aI13065
aI38225
asS'd'
I3871
sa(dp7742
S'a'
I3493
sS's'
I16
sS'm'
(lp7743
I1
aI1
aI3
aI7
aI15
aI15
aI63
aI241
aI167
aI345
aI653
aI701
aI4725
aI12911
aI11545
aI24475
asS'd'
I3872
sa(dp7744
S'a'
I3543
sS's'
I16
sS'm'
(lp7745
I1
aI1
aI3
aI7
aI25
aI15
aI49
aI235
aI331
aI639
aI965
aI1117
aI7147
aI3789
aI3309
aI20255
asS'd'
I3873
sa(dp7746
S'a'
I3549
sS's'
I16
sS'm'
(lp7747
I1
aI3
aI5
aI7
aI7
aI63
aI93
aI241
aI253
aI31
aI951
aI3723
aI3359
aI7303
aI191
aI36427
asS'd'
I3874
sa(dp7748
S'a'
I3550
sS's'
I16
sS'm'
(lp7749
I1
aI3
aI7
aI9
aI9
aI59
aI5
aI107
aI181
aI413
aI1269
aI3121
aI1929
aI11921
aI8931
aI47459
asS'd'
I3875
sa(dp7750
S'a'
I3553
sS's'
I16
sS'm'
(lp7751
I1
aI3
aI1
aI15
aI25
aI27
aI13
aI47
aI295
aI111
aI1287
aI2551
aI4887
aI4145
aI17063
aI42037
asS'd'
I3876
sa(dp7752
S'a'
I3563
sS's'
I16
sS'm'
(lp7753
I1
aI3
aI3
aI13
aI17
aI17
aI21
aI17
aI491
aI845
aI1463
aI1305
aI1375
aI16149
aI19331
aI25043
asS'd'
I3877
sa(dp7754
S'a'
I3568
sS's'
I16
sS'm'
(lp7755
I1
aI3
aI5
aI1
aI27
aI5
aI93
aI139
aI283
aI711
aI1141
aI1743
aI5001
aI8851
aI19351
aI12275
asS'd'
I3878
sa(dp7756
S'a'
I3604
sS's'
I16
sS'm'
(lp7757
I1
aI1
aI1
aI1
aI23
aI25
aI51
aI63
aI429
aI735
aI201
aI3785
aI6677
aI16375
aI19681
aI17857
asS'd'
I3879
sa(dp7758
S'a'
I3632
sS's'
I16
sS'm'
(lp7759
I1
aI3
aI3
aI3
aI9
aI63
aI71
aI147
aI463
aI465
aI1163
aI1045
aI6967
aI12537
aI31853
aI38391
asS'd'
I3880
sa(dp7760
S'a'
I3650
sS's'
I16
sS'm'
(lp7761
I1
aI3
aI7
aI1
aI5
aI51
aI79
aI239
aI389
aI3
aI601
aI3787
aI7635
aI16295
aI1681
aI63971
asS'd'
I3881
sa(dp7762
S'a'
I3662
sS's'
I16
sS'm'
(lp7763
I1
aI3
aI1
aI3
aI5
aI31
aI103
aI89
aI321
aI971
aI783
aI3685
aI1155
aI10353
aI2167
aI35423
asS'd'
I3882
sa(dp7764
S'a'
I3674
sS's'
I16
sS'm'
(lp7765
I1
aI1
aI5
aI15
aI25
aI19
aI93
aI59
aI361
aI217
aI1141
aI597
aI5877
aI15961
aI1593
aI22925
asS'd'
I3883
sa(dp7766
S'a'
I3685
sS's'
I16
sS'm'
(lp7767
I1
aI3
aI1
aI9
aI25
aI59
aI69
aI89
aI477
aI89
aI487
aI237
aI5625
aI9579
aI30421
aI21883
asS'd'
I3884
sa(dp7768
S'a'
I3686
sS's'
I16
sS'm'
(lp7769
I1
aI1
aI3
aI7
aI1
aI5
aI13
aI225
aI9
aI981
aI1081
aI1407
aI6855
aI15215
aI21713
aI62313
asS'd'
I3885
sa(dp7770
S'a'
I3700
sS's'
I16
sS'm'
(lp7771
I1
aI1
aI7
aI15
aI11
aI13
aI119
aI109
aI151
aI245
aI1195
aI3741
aI755
aI8047
aI15431
aI21001
asS'd'
I3886
sa(dp7772
S'a'
I3703
sS's'
I16
sS'm'
(lp7773
I1
aI3
aI7
aI3
aI17
aI47
aI107
aI137
aI99
aI255
aI1597
aI3281
aI5779
aI13487
aI15061
aI19199
asS'd'
I3887
sa(dp7774
S'a'
I3704
sS's'
I16
sS'm'
(lp7775
I1
aI1
aI3
aI3
aI9
aI39
aI77
aI227
aI511
aI839
aI1375
aI3887
aI25
aI14763
aI13259
aI217
asS'd'
I3888
sa(dp7776
S'a'
I3723
sS's'
I16
sS'm'
(lp7777
I1
aI3
aI5
aI7
aI17
aI3
aI87
aI61
aI439
aI287
aI709
aI4085
aI4251
aI8945
aI28203
aI24011
asS'd'
I3889
sa(dp7778
S'a'
I3743
sS's'
I16
sS'm'
(lp7779
I1
aI3
aI1
aI1
aI29
aI25
aI49
aI101
aI209
aI359
aI285
aI1593
aI4161
aI2943
aI23225
aI6381
asS'd'
I3890
sa(dp7780
S'a'
I3753
sS's'
I16
sS'm'
(lp7781
I1
aI1
aI3
aI13
aI1
aI45
aI87
aI7
aI491
aI399
aI905
aI1403
aI4791
aI7419
aI14355
aI47767
asS'd'
I3891
sa(dp7782
S'a'
I3756
sS's'
I16
sS'm'
(lp7783
I1
aI1
aI7
aI15
aI13
aI25
aI111
aI197
aI297
aI301
aI499
aI4007
aI2235
aI7681
aI4641
aI32447
asS'd'
I3892
sa(dp7784
S'a'
I3759
sS's'
I16
sS'm'
(lp7785
I1
aI1
aI3
aI3
aI27
aI41
aI97
aI83
aI405
aI353
aI1609
aI201
aI1503
aI10673
aI29377
aI20445
asS'd'
I3893
sa(dp7786
S'a'
I3762
sS's'
I16
sS'm'
(lp7787
I1
aI1
aI7
aI3
aI9
aI47
aI65
aI191
aI207
aI545
aI377
aI3011
aI7361
aI3467
aI14073
aI46769
asS'd'
I3894
sa(dp7788
S'a'
I3771
sS's'
I16
sS'm'
(lp7789
I1
aI1
aI7
aI5
aI7
aI39
aI9
aI91
aI187
aI949
aI1829
aI161
aI3689
aI4145
aI32675
aI23263
asS'd'
I3895
sa(dp7790
S'a'
I3776
sS's'
I16
sS'm'
(lp7791
I1
aI1
aI5
aI9
aI29
aI9
aI83
aI113
aI77
aI673
aI613
aI3645
aI6671
aI8583
aI27701
aI18615
asS'd'
I3896
sa(dp7792
S'a'
I3779
sS's'
I16
sS'm'
(lp7793
I1
aI3
aI5
aI9
aI29
aI13
aI127
aI247
aI285
aI845
aI463
aI539
aI4441
aI1867
aI12469
aI16213
asS'd'
I3897
sa(dp7794
S'a'
I3839
sS's'
I16
sS'm'
(lp7795
I1
aI3
aI7
aI15
aI1
aI29
aI47
aI157
aI239
aI595
aI563
aI1103
aI3431
aI2849
aI28125
aI19969
asS'd'
I3898
sa(dp7796
S'a'
I3856
sS's'
I16
sS'm'
(lp7797
I1
aI1
aI1
aI15
aI25
aI13
aI1
aI131
aI57
aI257
aI2021
aI169
aI7603
aI10721
aI21675
aI63171
asS'd'
I3899
sa(dp7798
S'a'
I3871
sS's'
I16
sS'm'
(lp7799
I1
aI3
aI5
aI3
aI5
aI19
aI31
aI57
aI275
aI381
aI775
aI681
aI1145
aI12237
aI5141
aI29375
asS'd'
I3900
sa(dp7800
S'a'
I3887
sS's'
I16
sS'm'
(lp7801
I1
aI3
aI5
aI13
aI27
aI13
aI47
aI201
aI267
aI581
aI1563
aI3845
aI951
aI7209
aI27253
aI19755
asS'd'
I3901
sa(dp7802
S'a'
I3896
sS's'
I16
sS'm'
(lp7803
I1
aI3
aI5
aI15
aI19
aI35
aI57
aI17
aI61
aI273
aI967
aI3029
aI1747
aI1753
aI31321
aI23711
asS'd'
I3902
sa(dp7804
S'a'
I3901
sS's'
I16
sS'm'
(lp7805
I1
aI1
aI1
aI5
aI13
aI13
aI7
aI177
aI335
aI393
aI1401
aI1411
aI4703
aI8259
aI1281
aI39835
asS'd'
I3903
sa(dp7806
S'a'
I3916
sS's'
I16
sS'm'
(lp7807
I1
aI1
aI3
aI15
aI25
aI27
aI27
aI121
aI183
aI105
aI663
aI1375
aI6987
aI7151
aI13763
aI39323
asS'd'
I3904
sa(dp7808
S'a'
I3919
sS's'
I16
sS'm'
(lp7809
I1
aI3
aI7
aI5
aI15
aI1
aI81
aI129
aI455
aI163
aI675
aI81
aI3735
aI14409
aI7269
aI16425
asS'd'
I3905
sa(dp7810
S'a'
I3937
sS's'
I16
sS'm'
(lp7811
I1
aI3
aI3
aI11
aI13
aI7
aI79
aI157
aI165
aI663
aI229
aI3539
aI1837
aI6485
aI30729
aI42157
asS'd'
I3906
sa(dp7812
S'a'
I3943
sS's'
I16
sS'm'
(lp7813
I1
aI1
aI5
aI15
aI9
aI9
aI9
aI47
aI133
aI863
aI43
aI1461
aI511
aI13991
aI24781
aI19221
asS'd'
I3907
sa(dp7814
S'a'
I3955
sS's'
I16
sS'm'
(lp7815
I1
aI3
aI1
aI7
aI31
aI33
aI103
aI13
aI159
aI689
aI1353
aI4025
aI6051
aI7683
aI1741
aI30047
asS'd'
I3908
sa(dp7816
S'a'
I3961
sS's'
I16
sS'm'
(lp7817
I1
aI1
aI3
aI11
aI5
aI45
aI71
aI219
aI475
aI585
aI1207
aI3163
aI4661
aI4713
aI12729
aI30445
asS'd'
I3909
sa(dp7818
S'a'
I3988
sS's'
I16
sS'm'
(lp7819
I1
aI3
aI7
aI5
aI5
aI53
aI101
aI227
aI129
aI521
aI91
aI1129
aI4683
aI11235
aI24697
aI45055
asS'd'
I3910
sa(dp7820
S'a'
I3997
sS's'
I16
sS'm'
(lp7821
I1
aI1
aI3
aI13
aI1
aI43
aI7
aI1
aI73
aI857
aI1713
aI185
aI1685
aI2369
aI24187
aI40419
asS'd'
I3911
sa(dp7822
S'a'
I4011
sS's'
I16
sS'm'
(lp7823
I1
aI1
aI7
aI7
aI21
aI7
aI13
aI177
aI503
aI1003
aI1091
aI2411
aI1433
aI9063
aI13901
aI3329
asS'd'
I3912
sa(dp7824
S'a'
I4026
sS's'
I16
sS'm'
(lp7825
I1
aI1
aI7
aI1
aI7
aI41
aI99
aI203
aI325
aI249
aI1763
aI545
aI2981
aI14125
aI7815
aI11385
asS'd'
I3913
sa(dp7826
S'a'
I4033
sS's'
I16
sS'm'
(lp7827
I1
aI3
aI7
aI11
aI3
aI11
aI95
aI137
aI325
aI701
aI1177
aI1631
aI4483
aI2955
aI30229
aI25577
asS'd'
I3914
sa(dp7828
S'a'
I4045
sS's'
I16
sS'm'
(lp7829
I1
aI1
aI7
aI7
aI17
aI45
aI77
aI103
aI143
aI97
aI1963
aI3635
aI1539
aI10491
aI23483
aI22767
asS'd'
I3915
sa(dp7830
S'a'
I4060
sS's'
I16
sS'm'
(lp7831
I1
aI1
aI7
aI15
aI7
aI5
aI81
aI63
aI243
aI55
aI39
aI207
aI2315
aI8285
aI8155
aI11631
asS'd'
I3916
sa(dp7832
S'a'
I4063
sS's'
I16
sS'm'
(lp7833
I1
aI3
aI5
aI15
aI23
aI19
aI115
aI9
aI125
aI851
aI161
aI3767
aI3513
aI1855
aI11139
aI1719
asS'd'
I3917
sa(dp7834
S'a'
I4064
sS's'
I16
sS'm'
(lp7835
I1
aI3
aI7
aI11
aI11
aI23
aI15
aI13
aI235
aI5
aI1039
aI1425
aI6485
aI5539
aI8967
aI64809
asS'd'
I3918
sa(dp7836
S'a'
I4126
sS's'
I16
sS'm'
(lp7837
I1
aI3
aI5
aI7
aI19
aI11
aI83
aI135
aI45
aI905
aI1081
aI1857
aI3185
aI13555
aI21365
aI38143
asS'd'
I3919
sa(dp7838
S'a'
I4136
sS's'
I16
sS'm'
(lp7839
I1
aI1
aI5
aI1
aI25
aI27
aI119
aI109
aI167
aI847
aI1539
aI2653
aI797
aI11185
aI23501
aI22389
asS'd'
I3920
sa(dp7840
S'a'
I4167
sS's'
I16
sS'm'
(lp7841
I1
aI1
aI7
aI7
aI11
aI3
aI51
aI97
aI277
aI557
aI207
aI3645
aI825
aI8521
aI26653
aI60071
asS'd'
I3921
sa(dp7842
S'a'
I4173
sS's'
I16
sS'm'
(lp7843
I1
aI3
aI3
aI15
aI17
aI35
aI57
aI7
aI267
aI549
aI97
aI243
aI1137
aI10311
aI6737
aI19077
asS'd'
I3922
sa(dp7844
S'a'
I4188
sS's'
I16
sS'm'
(lp7845
I1
aI1
aI1
aI15
aI23
aI33
aI27
aI203
aI415
aI1023
aI1145
aI1881
aI7715
aI4413
aI3727
aI5185
asS'd'
I3923
sa(dp7846
S'a'
I4195
sS's'
I16
sS'm'
(lp7847
I1
aI1
aI3
aI3
aI13
aI47
aI63
aI13
aI75
aI505
aI595
aI2911
aI4029
aI14187
aI23151
aI42877
asS'd'
I3924
sa(dp7848
S'a'
I4226
sS's'
I16
sS'm'
(lp7849
I1
aI1
aI5
aI15
aI23
aI5
aI11
aI65
aI147
aI675
aI1961
aI2177
aI727
aI15077
aI23759
aI10195
asS'd'
I3925
sa(dp7850
S'a'
I4291
sS's'
I16
sS'm'
(lp7851
I1
aI3
aI5
aI9
aI9
aI39
aI69
aI229
aI341
aI627
aI1331
aI3139
aI3921
aI9219
aI14887
aI4659
asS'd'
I3926
sa(dp7852
S'a'
I4298
sS's'
I16
sS'm'
(lp7853
I1
aI1
aI7
aI3
aI1
aI35
aI49
aI71
aI165
aI83
aI719
aI2771
aI6475
aI7821
aI16709
aI4449
asS'd'
I3927
sa(dp7854
S'a'
I4308
sS's'
I16
sS'm'
(lp7855
I1
aI3
aI5
aI5
aI23
aI15
aI3
aI57
aI465
aI77
aI121
aI3767
aI6841
aI13601
aI12035
aI14075
asS'd'
I3928
sa(dp7856
S'a'
I4312
sS's'
I16
sS'm'
(lp7857
I1
aI1
aI7
aI3
aI3
aI23
aI45
aI131
aI287
aI941
aI713
aI415
aI6865
aI14209
aI29555
aI55493
asS'd'
I3929
sa(dp7858
S'a'
I4336
sS's'
I16
sS'm'
(lp7859
I1
aI3
aI5
aI11
aI29
aI35
aI55
aI75
aI225
aI779
aI569
aI1795
aI1377
aI12765
aI19081
aI47287
asS'd'
I3930
sa(dp7860
S'a'
I4371
sS's'
I16
sS'm'
(lp7861
I1
aI3
aI7
aI3
aI31
aI47
aI127
aI89
aI157
aI737
aI1395
aI3615
aI7923
aI14731
aI15797
aI40061
asS'd'
I3931
sa(dp7862
S'a'
I4378
sS's'
I16
sS'm'
(lp7863
I1
aI1
aI1
aI11
aI21
aI37
aI21
aI59
aI9
aI141
aI193
aI3095
aI3435
aI12371
aI26931
aI61861
asS'd'
I3932
sa(dp7864
S'a'
I4384
sS's'
I16
sS'm'
(lp7865
I1
aI1
aI3
aI7
aI13
aI51
aI15
aI153
aI77
aI1013
aI651
aI3949
aI6229
aI14297
aI1039
aI46139
asS'd'
I3933
sa(dp7866
S'a'
I4393
sS's'
I16
sS'm'
(lp7867
I1
aI3
aI3
aI13
aI7
aI43
aI95
aI61
aI217
aI3
aI549
aI739
aI123
aI3661
aI15375
aI13919
asS'd'
I3934
sa(dp7868
S'a'
I4421
sS's'
I16
sS'm'
(lp7869
I1
aI3
aI5
aI9
aI13
aI37
aI101
aI89
aI55
aI413
aI1089
aI775
aI7575
aI13063
aI31393
aI29583
asS'd'
I3935
sa(dp7870
S'a'
I4425
sS's'
I16
sS'm'
(lp7871
I1
aI1
aI3
aI9
aI25
aI63
aI119
aI143
aI499
aI145
aI603
aI2067
aI4713
aI13457
aI14053
aI117
asS'd'
I3936
sa(dp7872
S'a'
I4439
sS's'
I16
sS'm'
(lp7873
I1
aI1
aI5
aI9
aI7
aI23
aI57
aI253
aI115
aI591
aI2003
aI63
aI7615
aI11493
aI28519
aI47087
asS'd'
I3937
sa(dp7874
S'a'
I4440
sS's'
I16
sS'm'
(lp7875
I1
aI1
aI7
aI3
aI7
aI53
aI121
aI33
aI233
aI645
aI1093
aI1697
aI7213
aI2603
aI10743
aI51303
asS'd'
I3938
sa(dp7876
S'a'
I4500
sS's'
I16
sS'm'
(lp7877
I1
aI3
aI5
aI7
aI13
aI31
aI17
aI125
aI93
aI969
aI159
aI1529
aI7165
aI7371
aI8707
aI56953
asS'd'
I3939
sa(dp7878
S'a'
I4514
sS's'
I16
sS'm'
(lp7879
I1
aI3
aI3
aI1
aI13
aI9
aI91
aI25
aI171
aI843
aI1635
aI2043
aI1043
aI15893
aI11409
aI53689
asS'd'
I3940
sa(dp7880
S'a'
I4523
sS's'
I16
sS'm'
(lp7881
I1
aI3
aI5
aI7
aI13
aI19
aI89
aI97
aI203
aI923
aI1109
aI2061
aI463
aI11703
aI8925
aI56015
asS'd'
I3941
sa(dp7882
S'a'
I4534
sS's'
I16
sS'm'
(lp7883
I1
aI3
aI5
aI11
aI5
aI21
aI79
aI237
aI195
aI649
aI717
aI211
aI919
aI12855
aI3045
aI39659
asS'd'
I3942
sa(dp7884
S'a'
I4593
sS's'
I16
sS'm'
(lp7885
I1
aI1
aI1
aI15
aI13
aI19
aI21
aI69
aI393
aI257
aI1263
aI309
aI3209
aI8403
aI24467
aI6467
asS'd'
I3943
sa(dp7886
S'a'
I4615
sS's'
I16
sS'm'
(lp7887
I1
aI1
aI1
aI11
aI7
aI27
aI59
aI117
aI379
aI353
aI943
aI2513
aI3869
aI4567
aI12989
aI13139
asS'd'
I3944
sa(dp7888
S'a'
I4630
sS's'
I16
sS'm'
(lp7889
I1
aI1
aI1
aI3
aI13
aI43
aI11
aI15
aI149
aI237
aI1555
aI71
aI2357
aI15773
aI21419
aI40571
asS'd'
I3945
sa(dp7890
S'a'
I4636
sS's'
I16
sS'm'
(lp7891
I1
aI3
aI1
aI9
aI19
aI23
aI59
aI215
aI15
aI921
aI1729
aI249
aI3785
aI7171
aI1233
aI3449
asS'd'
I3946
sa(dp7892
S'a'
I4645
sS's'
I16
sS'm'
(lp7893
I1
aI1
aI1
aI7
aI7
aI37
aI63
aI205
aI75
aI599
aI951
aI2513
aI3347
aI2497
aI8019
aI5433
asS'd'
I3947
sa(dp7894
S'a'
I4684
sS's'
I16
sS'm'
(lp7895
I1
aI3
aI3
aI15
aI27
aI17
aI25
aI201
aI23
aI699
aI1525
aI465
aI1115
aI12299
aI14747
aI40363
asS'd'
I3948
sa(dp7896
S'a'
I4687
sS's'
I16
sS'm'
(lp7897
I1
aI1
aI1
aI3
aI29
aI59
aI115
aI233
aI107
aI815
aI291
aI3821
aI7325
aI7381
aI21445
aI33917
asS'd'
I3949
sa(dp7898
S'a'
I4723
sS's'
I16
sS'm'
(lp7899
I1
aI3
aI1
aI11
aI11
aI33
aI107
aI171
aI421
aI893
aI587
aI3373
aI4101
aI3885
aI25383
aI12035
asS'd'
I3950
sa(dp7900
S'a'
I4735
sS's'
I16
sS'm'
(lp7901
I1
aI3
aI3
aI7
aI5
aI23
aI43
aI51
aI357
aI77
aI1327
aI2995
aI1321
aI1571
aI26419
aI23603
asS'd'
I3951
sa(dp7902
S'a'
I4746
sS's'
I16
sS'm'
(lp7903
I1
aI3
aI7
aI9
aI27
aI57
aI101
aI51
aI215
aI215
aI469
aI303
aI723
aI2903
aI30569
aI42631
asS'd'
I3952
sa(dp7904
S'a'
I4779
sS's'
I16
sS'm'
(lp7905
I1
aI3
aI3
aI13
aI1
aI7
aI63
aI205
aI143
aI321
aI1439
aI253
aI2667
aI1271
aI11761
aI55631
asS'd'
I3953
sa(dp7906
S'a'
I4782
sS's'
I16
sS'm'
(lp7907
I1
aI1
aI7
aI9
aI3
aI7
aI7
aI15
aI503
aI875
aI1619
aI1715
aI5047
aI5665
aI5503
aI17745
asS'd'
I3954
sa(dp7908
S'a'
I4793
sS's'
I16
sS'm'
(lp7909
I1
aI1
aI7
aI15
aI19
aI49
aI65
aI31
aI245
aI371
aI377
aI2963
aI6185
aI5519
aI10743
aI33231
asS'd'
I3955
sa(dp7910
S'a'
I4796
sS's'
I16
sS'm'
(lp7911
I1
aI1
aI7
aI3
aI25
aI27
aI115
aI51
aI299
aI451
aI285
aI1709
aI6153
aI14881
aI17861
aI22071
asS'd'
I3956
sa(dp7912
S'a'
I4813
sS's'
I16
sS'm'
(lp7913
I1
aI3
aI1
aI5
aI21
aI21
aI127
aI185
aI325
aI995
aI213
aI3279
aI4769
aI15943
aI2589
aI29567
asS'd'
I3957
sa(dp7914
S'a'
I4850
sS's'
I16
sS'm'
(lp7915
I1
aI3
aI7
aI5
aI21
aI9
aI63
aI59
aI159
aI743
aI663
aI2965
aI97
aI8993
aI25633
aI29033
asS'd'
I3958
sa(dp7916
S'a'
I4867
sS's'
I16
sS'm'
(lp7917
I1
aI3
aI7
aI13
aI3
aI35
aI59
aI101
aI21
aI659
aI1531
aI3995
aI795
aI2143
aI21749
aI52715
asS'd'
I3959
sa(dp7918
S'a'
I4874
sS's'
I16
sS'm'
(lp7919
I1
aI3
aI3
aI15
aI27
aI29
aI95
aI1
aI501
aI425
aI417
aI2351
aI7877
aI4127
aI3633
aI23347
asS'd'
I3960
sa(dp7920
S'a'
I4881
sS's'
I16
sS'm'
(lp7921
I1
aI3
aI5
aI7
aI7
aI49
aI55
aI19
aI329
aI467
aI425
aI1609
aI6987
aI16123
aI26879
aI42883
asS'd'
I3961
sa(dp7922
S'a'
I4894
sS's'
I16
sS'm'
(lp7923
I1
aI1
aI1
aI15
aI17
aI21
aI13
aI13
aI85
aI7
aI677
aI3739
aI5491
aI6299
aI29957
aI55765
asS'd'
I3962
sa(dp7924
S'a'
I4904
sS's'
I16
sS'm'
(lp7925
I1
aI1
aI1
aI7
aI31
aI21
aI1
aI5
aI193
aI659
aI979
aI3409
aI3151
aI6615
aI7445
aI8151
asS'd'
I3963
sa(dp7926
S'a'
I4927
sS's'
I16
sS'm'
(lp7927
I1
aI3
aI1
aI1
aI11
aI61
aI27
aI205
aI263
aI805
aI955
aI3469
aI1233
aI1609
aI15329
aI13353
asS'd'
I3964
sa(dp7928
S'a'
I4929
sS's'
I16
sS'm'
(lp7929
I1
aI3
aI3
aI9
aI3
aI29
aI59
aI75
aI149
aI557
aI663
aI3887
aI3369
aI3397
aI10611
aI9511
asS'd'
I3965
sa(dp7930
S'a'
I4989
sS's'
I16
sS'm'
(lp7931
I1
aI1
aI7
aI13
aI29
aI21
aI101
aI139
aI99
aI411
aI569
aI2343
aI6901
aI1685
aI20599
aI49543
asS'd'
I3966
sa(dp7932
S'a'
I5000
sS's'
I16
sS'm'
(lp7933
I1
aI3
aI3
aI15
aI11
aI3
aI87
aI89
aI5
aI293
aI291
aI1405
aI1489
aI9877
aI32505
aI32263
asS'd'
I3967
sa(dp7934
S'a'
I5020
sS's'
I16
sS'm'
(lp7935
I1
aI1
aI5
aI5
aI19
aI45
aI89
aI5
aI381
aI253
aI1339
aI707
aI4645
aI14177
aI29441
aI8965
asS'd'
I3968
sa(dp7936
S'a'
I5036
sS's'
I16
sS'm'
(lp7937
I1
aI3
aI7
aI15
aI27
aI45
aI25
aI177
aI81
aI229
aI1339
aI2143
aI6547
aI6841
aI23449
aI14813
asS'd'
I3969
sa(dp7938
S'a'
I5041
sS's'
I16
sS'm'
(lp7939
I1
aI1
aI1
aI3
aI27
aI23
aI81
aI157
aI53
aI513
aI1435
aI277
aI2353
aI3545
aI21461
aI51479
asS'd'
I3970
sa(dp7940
S'a'
I5059
sS's'
I16
sS'm'
(lp7941
I1
aI3
aI1
aI3
aI3
aI17
aI75
aI139
aI283
aI881
aI1157
aI2081
aI937
aI14549
aI10215
aI13053
asS'd'
I3971
sa(dp7942
S'a'
I5074
sS's'
I16
sS'm'
(lp7943
I1
aI1
aI7
aI9
aI25
aI27
aI27
aI133
aI21
aI559
aI225
aI613
aI6931
aI11785
aI23413
aI35757
asS'd'
I3972
sa(dp7944
S'a'
I5090
sS's'
I16
sS'm'
(lp7945
I1
aI1
aI3
aI13
aI19
aI9
aI65
aI49
aI453
aI779
aI621
aI1151
aI1807
aI13269
aI6515
aI17113
asS'd'
I3973
sa(dp7946
S'a'
I5110
sS's'
I16
sS'm'
(lp7947
I1
aI1
aI1
aI13
aI21
aI49
aI39
aI79
aI119
aI401
aI903
aI493
aI3779
aI7389
aI29425
aI28091
asS'd'
I3974
sa(dp7948
S'a'
I5134
sS's'
I16
sS'm'
(lp7949
I1
aI3
aI1
aI3
aI23
aI57
aI59
aI213
aI463
aI839
aI1201
aI1951
aI5197
aI13035
aI29657
aI35517
asS'd'
I3975
sa(dp7950
S'a'
I5152
sS's'
I16
sS'm'
(lp7951
I1
aI3
aI7
aI7
aI3
aI49
aI29
aI181
aI367
aI101
aI1277
aI3329
aI3563
aI10373
aI29757
aI62555
asS'd'
I3976
sa(dp7952
S'a'
I5176
sS's'
I16
sS'm'
(lp7953
I1
aI3
aI1
aI7
aI31
aI31
aI117
aI213
aI373
aI57
aI1095
aI2733
aI3431
aI3915
aI7665
aI44459
asS'd'
I3977
sa(dp7954
S'a'
I5181
sS's'
I16
sS'm'
(lp7955
I1
aI1
aI7
aI5
aI9
aI25
aI47
aI117
aI355
aI495
aI1367
aI2579
aI5617
aI787
aI27655
aI18885
asS'd'
I3978
sa(dp7956
S'a'
I5204
sS's'
I16
sS'm'
(lp7957
I1
aI1
aI1
aI3
aI9
aI39
aI113
aI87
aI107
aI477
aI891
aI2273
aI4239
aI7521
aI147
aI1737
asS'd'
I3979
sa(dp7958
S'a'
I5218
sS's'
I16
sS'm'
(lp7959
I1
aI1
aI1
aI3
aI13
aI61
aI81
aI225
aI113
aI441
aI889
aI1915
aI3897
aI15179
aI4053
aI5925
asS'd'
I3980
sa(dp7960
S'a'
I5242
sS's'
I16
sS'm'
(lp7961
I1
aI1
aI5
aI3
aI15
aI47
aI59
aI187
aI475
aI197
aI1381
aI33
aI4605
aI1487
aI14359
aI33769
asS'd'
I3981
sa(dp7962
S'a'
I5253
sS's'
I16
sS'm'
(lp7963
I1
aI3
aI7
aI15
aI23
aI45
aI53
aI215
aI129
aI465
aI795
aI53
aI7077
aI9517
aI2663
aI55397
asS'd'
I3982
sa(dp7964
S'a'
I5260
sS's'
I16
sS'm'
(lp7965
I1
aI1
aI7
aI13
aI25
aI49
aI105
aI255
aI245
aI153
aI1093
aI2123
aI2823
aI5125
aI17483
aI49003
asS'd'
I3983
sa(dp7966
S'a'
I5281
sS's'
I16
sS'm'
(lp7967
I1
aI1
aI1
aI13
aI31
aI5
aI7
aI243
aI255
aI231
aI1663
aI1007
aI7573
aI405
aI29183
aI11367
asS'd'
I3984
sa(dp7968
S'a'
I5282
sS's'
I16
sS'm'
(lp7969
I1
aI1
aI5
aI13
aI15
aI15
aI115
aI91
aI63
aI1013
aI1713
aI1945
aI6397
aI14213
aI24417
aI40807
asS'd'
I3985
sa(dp7970
S'a'
I5313
sS's'
I16
sS'm'
(lp7971
I1
aI1
aI5
aI3
aI19
aI49
aI91
aI25
aI43
aI601
aI25
aI2405
aI1973
aI629
aI497
aI12625
asS'd'
I3986
sa(dp7972
S'a'
I5316
sS's'
I16
sS'm'
(lp7973
I1
aI1
aI3
aI5
aI13
aI45
aI11
aI81
aI251
aI545
aI1155
aI1409
aI7187
aI847
aI2835
aI32909
asS'd'
I3987
sa(dp7974
S'a'
I5326
sS's'
I16
sS'm'
(lp7975
I1
aI3
aI1
aI13
aI27
aI57
aI1
aI103
aI465
aI809
aI1727
aI3721
aI3347
aI3791
aI17247
aI8377
asS'd'
I3988
sa(dp7976
S'a'
I5340
sS's'
I16
sS'm'
(lp7977
I1
aI3
aI3
aI15
aI25
aI31
aI91
aI91
aI119
aI205
aI1431
aI703
aI5327
aI7323
aI30415
aI61955
asS'd'
I3989
sa(dp7978
S'a'
I5347
sS's'
I16
sS'm'
(lp7979
I1
aI3
aI5
aI11
aI19
aI39
aI79
aI243
aI109
aI463
aI1869
aI2133
aI4139
aI10461
aI14793
aI24025
asS'd'
I3990
sa(dp7980
S'a'
I5354
sS's'
I16
sS'm'
(lp7981
I1
aI3
aI5
aI7
aI23
aI41
aI5
aI7
aI249
aI3
aI1743
aI489
aI4921
aI397
aI30955
aI22207
asS'd'
I3991
sa(dp7982
S'a'
I5368
sS's'
I16
sS'm'
(lp7983
I1
aI3
aI5
aI15
aI3
aI7
aI115
aI19
aI217
aI231
aI1183
aI3723
aI5055
aI12967
aI7855
aI5067
asS'd'
I3992
sa(dp7984
S'a'
I5394
sS's'
I16
sS'm'
(lp7985
I1
aI3
aI3
aI3
aI11
aI31
aI113
aI41
aI429
aI797
aI557
aI1199
aI1819
aI1933
aI9917
aI32229
asS'd'
I3993
sa(dp7986
S'a'
I5396
sS's'
I16
sS'm'
(lp7987
I1
aI1
aI5
aI3
aI13
aI63
aI31
aI183
aI465
aI915
aI723
aI3227
aI4125
aI2813
aI26313
aI34287
asS'd'
I3994
sa(dp7988
S'a'
I5400
sS's'
I16
sS'm'
(lp7989
I1
aI1
aI7
aI5
aI31
aI55
aI117
aI243
aI37
aI885
aI85
aI1067
aI3895
aI15655
aI28527
aI32109
asS'd'
I3995
sa(dp7990
S'a'
I5405
sS's'
I16
sS'm'
(lp7991
I1
aI3
aI7
aI15
aI17
aI43
aI43
aI173
aI119
aI187
aI1161
aI599
aI4595
aI1681
aI11981
aI681
asS'd'
I3996
sa(dp7992
S'a'
I5439
sS's'
I16
sS'm'
(lp7993
I1
aI1
aI7
aI7
aI29
aI47
aI25
aI93
aI411
aI103
aI783
aI1029
aI1927
aI3569
aI27647
aI8281
asS'd'
I3997
sa(dp7994
S'a'
I5442
sS's'
I16
sS'm'
(lp7995
I1
aI3
aI3
aI9
aI19
aI57
aI31
aI183
aI141
aI889
aI157
aI2267
aI5701
aI6273
aI25739
aI34039
asS'd'
I3998
sa(dp7996
S'a'
I5459
sS's'
I16
sS'm'
(lp7997
I1
aI3
aI5
aI1
aI29
aI35
aI105
aI165
aI505
aI299
aI1149
aI2397
aI2013
aI11591
aI15917
aI4791
asS'd'
I3999
sa(dp7998
S'a'
I5478
sS's'
I16
sS'm'
(lp7999
I1
aI3
aI3
aI9
aI7
aI35
aI47
aI77
aI69
aI335
aI585
aI1131
aI531
aI8597
aI307
aI55985
asS'd'
I4000
sa(dp8000
S'a'
I5484
sS's'
I16
sS'm'
(lp8001
I1
aI3
aI7
aI1
aI29
aI9
aI25
aI155
aI149
aI845
aI567
aI3735
aI3501
aI9365
aI12025
aI19131
asS'd'
I4001
sa(dp8002
S'a'
I5508
sS's'
I16
sS'm'
(lp8003
I1
aI3
aI5
aI3
aI11
aI31
aI25
aI9
aI411
aI519
aI1611
aI1441
aI1487
aI10049
aI14373
aI24605
asS'd'
I4002
sa(dp8004
S'a'
I5523
sS's'
I16
sS'm'
(lp8005
I1
aI3
aI3
aI5
aI3
aI7
aI101
aI107
aI339
aI155
aI1843
aI2529
aI443
aI8177
aI28655
aI8151
asS'd'
I4003
sa(dp8006
S'a'
I5545
sS's'
I16
sS'm'
(lp8007
I1
aI1
aI7
aI5
aI29
aI37
aI73
aI131
aI125
aI451
aI947
aI797
aI5053
aI10155
aI30801
aI27235
asS'd'
I4004
sa(dp8008
S'a'
I5565
sS's'
I16
sS'm'
(lp8009
I1
aI1
aI7
aI13
aI19
aI47
aI101
aI45
aI495
aI457
aI1293
aI1971
aI5495
aI12737
aI17687
aI26123
asS'd'
I4005
sa(dp8010
S'a'
I5566
sS's'
I16
sS'm'
(lp8011
I1
aI1
aI7
aI7
aI11
aI11
aI75
aI177
aI251
aI553
aI1883
aI3379
aI1429
aI12227
aI10301
aI16467
asS'd'
I4006
sa(dp8012
S'a'
I5580
sS's'
I16
sS'm'
(lp8013
I1
aI3
aI3
aI9
aI3
aI61
aI95
aI35
aI97
aI551
aI233
aI2045
aI4873
aI9109
aI10019
aI64523
asS'd'
I4007
sa(dp8014
S'a'
I5608
sS's'
I16
sS'm'
(lp8015
I1
aI3
aI1
aI5
aI11
aI3
aI15
aI177
aI301
aI573
aI2029
aI191
aI5551
aI12083
aI27287
aI57235
asS'd'
I4008
sa(dp8016
S'a'
I5613
sS's'
I16
sS'm'
(lp8017
I1
aI3
aI5
aI1
aI21
aI9
aI121
aI169
aI347
aI187
aI57
aI3163
aI5609
aI1921
aI17581
aI28351
asS'd'
I4009
sa(dp8018
S'a'
I5647
sS's'
I16
sS'm'
(lp8019
I1
aI3
aI3
aI1
aI31
aI51
aI15
aI45
aI429
aI245
aI573
aI1595
aI5343
aI7519
aI17009
aI1299
asS'd'
I4010
sa(dp8020
S'a'
I5661
sS's'
I16
sS'm'
(lp8021
I1
aI1
aI7
aI3
aI13
aI47
aI109
aI65
aI389
aI867
aI963
aI145
aI1089
aI9749
aI19625
aI43121
asS'd'
I4011
sa(dp8022
S'a'
I5671
sS's'
I16
sS'm'
(lp8023
I1
aI3
aI1
aI7
aI21
aI61
aI77
aI67
aI143
aI579
aI625
aI2007
aI6343
aI4259
aI21233
aI237
asS'd'
I4012
sa(dp8024
S'a'
I5678
sS's'
I16
sS'm'
(lp8025
I1
aI3
aI5
aI9
aI27
aI15
aI107
aI91
aI399
aI895
aI645
aI2301
aI439
aI6789
aI18299
aI47285
asS'd'
I4013
sa(dp8026
S'a'
I5680
sS's'
I16
sS'm'
(lp8027
I1
aI3
aI3
aI5
aI13
aI11
aI43
aI199
aI505
aI409
aI25
aI2057
aI479
aI6031
aI9561
aI51613
asS'd'
I4014
sa(dp8028
S'a'
I5685
sS's'
I16
sS'm'
(lp8029
I1
aI1
aI7
aI13
aI15
aI55
aI105
aI53
aI171
aI925
aI1849
aI2881
aI6951
aI13069
aI865
aI41019
asS'd'
I4015
sa(dp8030
S'a'
I5689
sS's'
I16
sS'm'
(lp8031
I1
aI3
aI1
aI9
aI17
aI31
aI45
aI23
aI411
aI185
aI189
aI2123
aI2583
aI12713
aI12681
aI2231
asS'd'
I4016
sa(dp8032
S'a'
I5692
sS's'
I16
sS'm'
(lp8033
I1
aI3
aI7
aI9
aI3
aI63
aI11
aI253
aI177
aI127
aI545
aI3293
aI4449
aI15995
aI10223
aI33529
asS'd'
I4017
sa(dp8034
S'a'
I5724
sS's'
I16
sS'm'
(lp8035
I1
aI1
aI5
aI11
aI13
aI7
aI53
aI161
aI421
aI551
aI697
aI627
aI3879
aI1639
aI24419
aI3337
asS'd'
I4018
sa(dp8036
S'a'
I5745
sS's'
I16
sS'm'
(lp8037
I1
aI1
aI7
aI7
aI27
aI7
aI37
aI205
aI429
aI407
aI1133
aI3563
aI2921
aI6173
aI11173
aI3009
asS'd'
I4019
sa(dp8038
S'a'
I5755
sS's'
I16
sS'm'
(lp8039
I1
aI3
aI3
aI15
aI31
aI39
aI117
aI81
aI337
aI729
aI567
aI2299
aI1481
aI3189
aI1795
aI40299
asS'd'
I4020
sa(dp8040
S'a'
I5757
sS's'
I16
sS'm'
(lp8041
I1
aI3
aI5
aI15
aI15
aI47
aI91
aI127
aI275
aI55
aI951
aI3423
aI2879
aI6115
aI1549
aI2287
asS'd'
I4021
sa(dp8042
S'a'
I5786
sS's'
I16
sS'm'
(lp8043
I1
aI3
aI3
aI11
aI17
aI3
aI127
aI207
aI141
aI889
aI185
aI1095
aI4567
aI1371
aI30545
aI54445
asS'd'
I4022
sa(dp8044
S'a'
I5792
sS's'
I16
sS'm'
(lp8045
I1
aI1
aI7
aI3
aI25
aI11
aI11
aI139
aI43
aI1
aI1977
aI397
aI5775
aI6913
aI13249
aI46767
asS'd'
I4023
sa(dp8046
S'a'
I5810
sS's'
I16
sS'm'
(lp8047
I1
aI1
aI7
aI7
aI27
aI13
aI31
aI251
aI191
aI1015
aI161
aI3719
aI5321
aI13013
aI25187
aI51881
asS'd'
I4024
sa(dp8048
S'a'
I5824
sS's'
I16
sS'm'
(lp8049
I1
aI1
aI1
aI1
aI3
aI3
aI13
aI19
aI423
aI349
aI1955
aI2103
aI6395
aI3315
aI23809
aI25925
asS'd'
I4025
sa(dp8050
S'a'
I5869
sS's'
I16
sS'm'
(lp8051
I1
aI3
aI5
aI13
aI3
aI59
aI41
aI167
aI423
aI93
aI1299
aI2623
aI5829
aI8537
aI8701
aI43757
asS'd'
I4026
sa(dp8052
S'a'
I5872
sS's'
I16
sS'm'
(lp8053
I1
aI3
aI5
aI11
aI9
aI19
aI127
aI119
aI329
aI819
aI7
aI3809
aI5305
aI3643
aI27369
aI61827
asS'd'
I4027
sa(dp8054
S'a'
I5895
sS's'
I16
sS'm'
(lp8055
I1
aI3
aI1
aI15
aI25
aI43
aI55
aI159
aI205
aI911
aI983
aI2825
aI3751
aI7845
aI12023
aI18431
asS'd'
I4028
sa(dp8056
S'a'
I5923
sS's'
I16
sS'm'
(lp8057
I1
aI3
aI3
aI13
aI11
aI1
aI65
aI133
aI479
aI181
aI679
aI981
aI3317
aI6241
aI11693
aI9619
asS'd'
I4029
sa(dp8058
S'a'
I5925
sS's'
I16
sS'm'
(lp8059
I1
aI3
aI3
aI3
aI21
aI25
aI117
aI183
aI127
aI73
aI703
aI1469
aI257
aI11229
aI10167
aI50847
asS'd'
I4030
sa(dp8060
S'a'
I5926
sS's'
I16
sS'm'
(lp8061
I1
aI1
aI5
aI13
aI5
aI5
aI113
aI15
aI231
aI269
aI989
aI465
aI3267
aI15239
aI29503
aI42855
asS'd'
I4031
sa(dp8062
S'a'
I5944
sS's'
I16
sS'm'
(lp8063
I1
aI3
aI3
aI15
aI9
aI63
aI79
aI27
aI21
aI709
aI1969
aI3761
aI1015
aI13619
aI4205
aI40591
asS'd'
I4032
sa(dp8064
S'a'
I5986
sS's'
I16
sS'm'
(lp8065
I1
aI1
aI7
aI11
aI29
aI3
aI5
aI45
aI107
aI131
aI1287
aI3551
aI849
aI2003
aI27451
aI47103
asS'd'
I4033
sa(dp8066
S'a'
I6012
sS's'
I16
sS'm'
(lp8067
I1
aI3
aI5
aI11
aI3
aI47
aI59
aI53
aI369
aI249
aI915
aI2475
aI7539
aI763
aI7063
aI63971
asS'd'
I4034
sa(dp8068
S'a'
I6015
sS's'
I16
sS'm'
(lp8069
I1
aI1
aI5
aI1
aI7
aI53
aI45
aI127
aI321
aI341
aI635
aI2277
aI1383
aI10951
aI29055
aI45087
asS'd'
I4035
sa(dp8070
S'a'
I6046
sS's'
I16
sS'm'
(lp8071
I1
aI3
aI7
aI3
aI5
aI1
aI119
aI79
aI487
aI93
aI25
aI491
aI4085
aI6403
aI27779
aI8753
asS'd'
I4036
sa(dp8072
S'a'
I6049
sS's'
I16
sS'm'
(lp8073
I1
aI1
aI1
aI3
aI9
aI59
aI49
aI141
aI323
aI703
aI1175
aI423
aI4323
aI10083
aI4289
aI28931
asS'd'
I4037
sa(dp8074
S'a'
I6061
sS's'
I16
sS'm'
(lp8075
I1
aI3
aI3
aI15
aI31
aI31
aI73
aI15
aI187
aI653
aI91
aI1707
aI1431
aI9861
aI19071
aI8137
asS'd'
I4038
sa(dp8076
S'a'
I6067
sS's'
I16
sS'm'
(lp8077
I1
aI1
aI1
aI5
aI27
aI63
aI93
aI1
aI329
aI353
aI863
aI473
aI7681
aI10653
aI15819
aI8495
asS'd'
I4039
sa(dp8078
S'a'
I6099
sS's'
I16
sS'm'
(lp8079
I1
aI1
aI1
aI5
aI25
aI57
aI119
aI167
aI219
aI319
aI231
aI1065
aI6217
aI5131
aI1513
aI49281
asS'd'
I4040
sa(dp8080
S'a'
I6121
sS's'
I16
sS'm'
(lp8081
I1
aI3
aI7
aI3
aI17
aI3
aI113
aI91
aI201
aI179
aI1907
aI3423
aI821
aI12927
aI24827
aI49403
asS'd'
I4041
sa(dp8082
S'a'
I6155
sS's'
I16
sS'm'
(lp8083
I1
aI1
aI5
aI7
aI19
aI63
aI75
aI151
aI387
aI489
aI777
aI2049
aI1151
aI1351
aI25687
aI4143
asS'd'
I4042
sa(dp8084
S'a'
I6163
sS's'
I16
sS'm'
(lp8085
I1
aI3
aI5
aI7
aI9
aI37
aI9
aI3
aI87
aI385
aI1667
aI2139
aI7527
aI16133
aI30023
aI28783
asS'd'
I4043
sa(dp8086
S'a'
I6203
sS's'
I16
sS'm'
(lp8087
I1
aI1
aI5
aI9
aI11
aI55
aI95
aI73
aI413
aI867
aI589
aI2901
aI3021
aI413
aI5955
aI38921
asS'd'
I4044
sa(dp8088
S'a'
I6208
sS's'
I16
sS'm'
(lp8089
I1
aI3
aI5
aI15
aI1
aI17
aI17
aI7
aI485
aI323
aI519
aI2325
aI2255
aI4211
aI20661
aI28931
asS'd'
I4045
sa(dp8090
S'a'
I6231
sS's'
I16
sS'm'
(lp8091
I1
aI1
aI5
aI13
aI21
aI19
aI85
aI189
aI167
aI645
aI1475
aI3095
aI7123
aI3351
aI7961
aI20967
asS'd'
I4046
sa(dp8092
S'a'
I6241
sS's'
I16
sS'm'
(lp8093
I1
aI1
aI7
aI13
aI3
aI47
aI13
aI213
aI237
aI291
aI285
aI1465
aI1765
aI12361
aI32651
aI54205
asS'd'
I4047
sa(dp8094
S'a'
I6254
sS's'
I16
sS'm'
(lp8095
I1
aI3
aI7
aI13
aI13
aI27
aI71
aI35
aI67
aI117
aI647
aI2359
aI3295
aI8445
aI24761
aI29379
asS'd'
I4048
sa(dp8096
S'a'
I6262
sS's'
I16
sS'm'
(lp8097
I1
aI1
aI1
aI3
aI3
aI19
aI23
aI37
aI5
aI1019
aI5
aI1605
aI2291
aI14015
aI9311
aI39751
asS'd'
I4049
sa(dp8098
S'a'
I6266
sS's'
I16
sS'm'
(lp8099
I1
aI3
aI3
aI3
aI31
aI1
aI65
aI159
aI221
aI675
aI1061
aI971
aI2333
aI8265
aI14361
aI3263
asS'd'
I4050
sa(dp8100
S'a'
I6275
sS's'
I16
sS'm'
(lp8101
I1
aI1
aI3
aI7
aI3
aI5
aI81
aI17
aI101
aI991
aI753
aI2883
aI4977
aI4409
aI1757
aI26803
asS'd'
I4051
sa(dp8102
S'a'
I6278
sS's'
I16
sS'm'
(lp8103
I1
aI1
aI5
aI9
aI13
aI25
aI45
aI43
aI199
aI967
aI829
aI713
aI4547
aI7223
aI6497
aI53895
asS'd'
I4052
sa(dp8104
S'a'
I6292
sS's'
I16
sS'm'
(lp8105
I1
aI1
aI7
aI5
aI23
aI15
aI89
aI179
aI509
aI147
aI315
aI133
aI111
aI15577
aI23427
aI5229
asS'd'
I4053
sa(dp8106
S'a'
I6329
sS's'
I16
sS'm'
(lp8107
I1
aI3
aI3
aI7
aI27
aI7
aI113
aI65
aI315
aI135
aI1309
aI1179
aI5755
aI7513
aI6815
aI5137
asS'd'
I4054
sa(dp8108
S'a'
I6355
sS's'
I16
sS'm'
(lp8109
I1
aI1
aI3
aI7
aI1
aI13
aI29
aI155
aI477
aI721
aI71
aI865
aI3897
aI3331
aI30641
aI65471
asS'd'
I4055
sa(dp8110
S'a'
I6357
sS's'
I16
sS'm'
(lp8111
I1
aI1
aI7
aI3
aI29
aI45
aI83
aI3
aI461
aI109
aI1545
aI1365
aI6633
aI16137
aI23859
aI5995
asS'd'
I4056
sa(dp8112
S'a'
I6374
sS's'
I16
sS'm'
(lp8113
I1
aI3
aI1
aI1
aI3
aI33
aI77
aI83
aI459
aI187
aI879
aI3731
aI6939
aI6707
aI23409
aI24245
asS'd'
I4057
sa(dp8114
S'a'
I6380
sS's'
I16
sS'm'
(lp8115
I1
aI3
aI5
aI5
aI13
aI43
aI127
aI41
aI29
aI735
aI1391
aI2947
aI4873
aI4627
aI15
aI41719
asS'd'
I4058
sa(dp8116
S'a'
I6423
sS's'
I16
sS'm'
(lp8117
I1
aI3
aI1
aI3
aI17
aI17
aI51
aI93
aI189
aI227
aI449
aI2809
aI825
aI2009
aI9563
aI41435
asS'd'
I4059
sa(dp8118
S'a'
I6427
sS's'
I16
sS'm'
(lp8119
I1
aI3
aI3
aI11
aI25
aI47
aI113
aI173
aI141
aI919
aI677
aI117
aI5293
aI815
aI23749
aI19789
asS'd'
I4060
sa(dp8120
S'a'
I6430
sS's'
I16
sS'm'
(lp8121
I1
aI1
aI1
aI13
aI15
aI61
aI121
aI223
aI53
aI389
aI489
aI1527
aI4771
aI8975
aI8005
aI14275
asS'd'
I4061
sa(dp8122
S'a'
I6436
sS's'
I16
sS'm'
(lp8123
I1
aI1
aI3
aI15
aI31
aI57
aI111
aI145
aI279
aI991
aI489
aI3239
aI7647
aI473
aI31279
aI33447
asS'd'
I4062
sa(dp8124
S'a'
I6443
sS's'
I16
sS'm'
(lp8125
I1
aI1
aI7
aI5
aI31
aI13
aI75
aI185
aI395
aI611
aI609
aI159
aI7931
aI9887
aI4115
aI53121
asS'd'
I4063
sa(dp8126
S'a'
I6445
sS's'
I16
sS'm'
(lp8127
I1
aI3
aI5
aI5
aI13
aI39
aI103
aI237
aI77
aI913
aI511
aI1583
aI6763
aI14523
aI4247
aI63403
asS'd'
I4064
sa(dp8128
S'a'
I6458
sS's'
I16
sS'm'
(lp8129
I1
aI1
aI1
aI15
aI11
aI5
aI43
aI43
aI297
aI827
aI747
aI5
aI3785
aI15021
aI11291
aI36743
asS'd'
I4065
sa(dp8130
S'a'
I6478
sS's'
I16
sS'm'
(lp8131
I1
aI1
aI7
aI9
aI9
aI53
aI113
aI95
aI403
aI53
aI1335
aI4033
aI8147
aI11963
aI6523
aI23675
asS'd'
I4066
sa(dp8132
S'a'
I6490
sS's'
I16
sS'm'
(lp8133
I1
aI1
aI5
aI9
aI27
aI29
aI69
aI79
aI327
aI409
aI1147
aI1579
aI2625
aI12227
aI30933
aI9057
asS'd'
I4067
sa(dp8134
S'a'
I6508
sS's'
I16
sS'm'
(lp8135
I1
aI1
aI1
aI7
aI1
aI33
aI29
aI173
aI5
aI517
aI437
aI2035
aI57
aI12825
aI22095
aI65519
asS'd'
I4068
sa(dp8136
S'a'
I6519
sS's'
I16
sS'm'
(lp8137
I1
aI1
aI3
aI7
aI27
aI29
aI53
aI79
aI429
aI707
aI589
aI2605
aI339
aI7039
aI19319
aI17649
asS'd'
I4069
sa(dp8138
S'a'
I6520
sS's'
I16
sS'm'
(lp8139
I1
aI3
aI3
aI11
aI9
aI57
aI43
aI117
aI39
aI193
aI1427
aI2553
aI6877
aI7653
aI29041
aI44865
asS'd'
I4070
sa(dp8140
S'a'
I6530
sS's'
I16
sS'm'
(lp8141
I1
aI3
aI3
aI7
aI23
aI45
aI29
aI151
aI265
aI739
aI365
aI3565
aI6447
aI9761
aI24021
aI679
asS'd'
I4071
sa(dp8142
S'a'
I6541
sS's'
I16
sS'm'
(lp8143
I1
aI3
aI5
aI1
aI1
aI43
aI73
aI55
aI131
aI175
aI959
aI659
aI7315
aI15145
aI18301
aI14865
asS'd'
I4072
sa(dp8144
S'a'
I6556
sS's'
I16
sS'm'
(lp8145
I1
aI1
aI3
aI5
aI15
aI15
aI91
aI113
aI359
aI241
aI1627
aI1069
aI1761
aI211
aI32671
aI58833
asS'd'
I4073
sa(dp8146
S'a'
I6607
sS's'
I16
sS'm'
(lp8147
I1
aI3
aI3
aI7
aI29
aI27
aI79
aI53
aI409
aI81
aI693
aI3137
aI7385
aI11007
aI28459
aI28621
asS'd'
I4074
sa(dp8148
S'a'
I6612
sS's'
I16
sS'm'
(lp8149
I1
aI1
aI7
aI5
aI29
aI7
aI67
aI195
aI77
aI773
aI1361
aI2153
aI4459
aI7301
aI5129
aI17797
asS'd'
I4075
sa(dp8150
S'a'
I6626
sS's'
I16
sS'm'
(lp8151
I1
aI3
aI3
aI7
aI25
aI27
aI91
aI223
aI115
aI91
aI1645
aI2167
aI1955
aI9601
aI22127
aI13055
asS'd'
I4076
sa(dp8152
S'a'
I6632
sS's'
I16
sS'm'
(lp8153
I1
aI3
aI7
aI3
aI27
aI53
aI67
aI249
aI51
aI151
aI663
aI3231
aI895
aI6777
aI3037
aI56755
asS'd'
I4077
sa(dp8154
S'a'
I6649
sS's'
I16
sS'm'
(lp8155
I1
aI1
aI5
aI1
aI25
aI63
aI71
aI179
aI375
aI301
aI1127
aI2125
aI783
aI14481
aI7293
aI47883
asS'd'
I4078
sa(dp8156
S'a'
I6666
sS's'
I16
sS'm'
(lp8157
I1
aI1
aI3
aI9
aI25
aI3
aI39
aI69
aI1
aI85
aI1271
aI1571
aI1953
aI5077
aI20369
aI44827
asS'd'
I4079
sa(dp8158
S'a'
I6674
sS's'
I16
sS'm'
(lp8159
I1
aI3
aI1
aI13
aI11
aI61
aI77
aI59
aI127
aI475
aI1609
aI3553
aI2553
aI15589
aI9351
aI59787
asS'd'
I4080
sa(dp8160
S'a'
I6733
sS's'
I16
sS'm'
(lp8161
I1
aI3
aI1
aI5
aI21
aI7
aI61
aI27
aI199
aI653
aI1243
aI2481
aI5369
aI12903
aI30229
aI39453
asS'd'
I4081
sa(dp8162
S'a'
I6782
sS's'
I16
sS'm'
(lp8163
I1
aI3
aI7
aI3
aI13
aI15
aI107
aI153
aI501
aI573
aI863
aI3179
aI6019
aI15177
aI16075
aI43767
asS'd'
I4082
sa(dp8164
S'a'
I6786
sS's'
I16
sS'm'
(lp8165
I1
aI1
aI7
aI1
aI23
aI55
aI17
aI35
aI5
aI137
aI1707
aI1377
aI6857
aI15361
aI27299
aI61871
asS'd'
I4083
sa(dp8166
S'a'
I6798
sS's'
I16
sS'm'
(lp8167
I1
aI3
aI5
aI7
aI27
aI17
aI87
aI213
aI95
aI125
aI1239
aI3923
aI4193
aI11049
aI12783
aI45017
asS'd'
I4084
sa(dp8168
S'a'
I6821
sS's'
I16
sS'm'
(lp8169
I1
aI1
aI5
aI15
aI9
aI55
aI11
aI217
aI7
aI249
aI369
aI205
aI4251
aI13785
aI24781
aI48929
asS'd'
I4085
sa(dp8170
S'a'
I6840
sS's'
I16
sS'm'
(lp8171
I1
aI3
aI7
aI1
aI15
aI35
aI33
aI107
aI503
aI563
aI1591
aI3487
aI7495
aI1767
aI24791
aI31281
asS'd'
I4086
sa(dp8172
S'a'
I6846
sS's'
I16
sS'm'
(lp8173
I1
aI3
aI1
aI11
aI3
aI15
aI47
aI193
aI289
aI253
aI909
aI1105
aI5119
aI1415
aI7737
aI4341
asS'd'
I4087
sa(dp8174
S'a'
I6865
sS's'
I16
sS'm'
(lp8175
I1
aI1
aI1
aI3
aI23
aI33
aI53
aI187
aI469
aI573
aI835
aI2049
aI791
aI1177
aI31051
aI30955
asS'd'
I4088
sa(dp8176
S'a'
I6884
sS's'
I16
sS'm'
(lp8177
I1
aI3
aI3
aI11
aI15
aI51
aI77
aI143
aI369
aI991
aI1103
aI1293
aI6019
aI6361
aI26301
aI20741
asS'd'
I4089
sa(dp8178
S'a'
I6891
sS's'
I16
sS'm'
(lp8179
I1
aI1
aI1
aI5
aI17
aI19
aI85
aI135
aI113
aI593
aI579
aI1063
aI7173
aI2491
aI9355
aI19223
asS'd'
I4090
sa(dp8180
S'a'
I6925
sS's'
I16
sS'm'
(lp8181
I1
aI1
aI5
aI15
aI25
aI51
aI107
aI193
aI31
aI1
aI1693
aI125
aI6223
aI14619
aI22683
aI26321
asS'd'
I4091
sa(dp8182
S'a'
I6938
sS's'
I16
sS'm'
(lp8183
I1
aI1
aI7
aI1
aI17
aI45
aI87
aI39
aI87
aI499
aI1185
aI2763
aI3989
aI2863
aI24555
aI33817
asS'd'
I4092
sa(dp8184
S'a'
I6967
sS's'
I16
sS'm'
(lp8185
I1
aI3
aI1
aI11
aI31
aI5
aI121
aI231
aI185
aI793
aI255
aI2785
aI5261
aI3687
aI21711
aI3941
asS'd'
I4093
sa(dp8186
S'a'
I6988
sS's'
I16
sS'm'
(lp8187
I1
aI3
aI7
aI15
aI5
aI59
aI73
aI227
aI365
aI937
aI893
aI2155
aI4385
aI14345
aI6813
aI28461
asS'd'
I4094
sa(dp8188
S'a'
I6996
sS's'
I16
sS'm'
(lp8189
I1
aI1
aI5
aI7
aI7
aI23
aI7
aI239
aI431
aI45
aI1015
aI1663
aI1893
aI5035
aI24075
aI2119
asS'd'
I4095
sa(dp8190
S'a'
I7009
sS's'
I16
sS'm'
(lp8191
I1
aI3
aI5
aI1
aI3
aI15
aI63
aI103
aI119
aI801
aI1681
aI3463
aI6083
aI6453
aI11379
aI8205
asS'd'
I4096
sa(dp8192
S'a'
I7016
sS's'
I16
sS'm'
(lp8193
I1
aI3
aI7
aI9
aI21
aI61
aI9
aI9
aI433
aI541
aI603
aI3905
aI3787
aI10187
aI3643
aI21319
asS'd'
I4097
sa(dp8194
S'a'
I7030
sS's'
I16
sS'm'
(lp8195
I1
aI3
aI5
aI3
aI11
aI1
aI101
aI243
aI363
aI559
aI561
aI1163
aI455
aI4657
aI1147
aI39961
asS'd'
I4098
sa(dp8196
S'a'
I7043
sS's'
I16
sS'm'
(lp8197
I1
aI3
aI5
aI13
aI17
aI37
aI57
aI47
aI483
aI571
aI1579
aI1137
aI8125
aI12271
aI23279
aI1615
asS'd'
I4099
sa(dp8198
S'a'
I7045
sS's'
I16
sS'm'
(lp8199
I1
aI3
aI5
aI1
aI21
aI5
aI13
aI155
aI75
aI705
aI389
aI2855
aI6345
aI2279
aI12627
aI49451
asS'd'
I4100
sa(dp8200
S'a'
I7052
sS's'
I16
sS'm'
(lp8201
I1
aI1
aI3
aI9
aI15
aI51
aI73
aI99
aI445
aI227
aI1705
aI2175
aI8111
aI9381
aI31555
aI48491
asS'd'
I4101
sa(dp8202
S'a'
I7073
sS's'
I16
sS'm'
(lp8203
I1
aI3
aI3
aI5
aI9
aI63
aI107
aI51
aI461
aI979
aI1033
aI421
aI4807
aI11707
aI13261
aI26015
asS'd'
I4102
sa(dp8204
S'a'
I7142
sS's'
I16
sS'm'
(lp8205
I1
aI1
aI5
aI3
aI13
aI53
aI117
aI249
aI57
aI851
aI1391
aI3245
aI35
aI16043
aI24399
aI63667
asS'd'
I4103
sa(dp8206
S'a'
I7153
sS's'
I16
sS'm'
(lp8207
I1
aI3
aI1
aI11
aI23
aI33
aI57
aI125
aI385
aI495
aI871
aI199
aI4269
aI2269
aI22897
aI23597
asS'd'
I4104
sa(dp8208
S'a'
I7168
sS's'
I16
sS'm'
(lp8209
I1
aI3
aI5
aI15
aI29
aI11
aI77
aI21
aI479
aI369
aI723
aI3721
aI1121
aI9463
aI19775
aI54525
asS'd'
I4105
sa(dp8210
S'a'
I7174
sS's'
I16
sS'm'
(lp8211
I1
aI3
aI5
aI7
aI7
aI45
aI29
aI153
aI395
aI223
aI1917
aI3713
aI5087
aI10827
aI1383
aI36823
asS'd'
I4106
sa(dp8212
S'a'
I7183
sS's'
I16
sS'm'
(lp8213
I1
aI3
aI1
aI3
aI31
aI19
aI111
aI55
aI275
aI923
aI917
aI2925
aI673
aI6579
aI18783
aI5137
asS'd'
I4107
sa(dp8214
S'a'
I7195
sS's'
I16
sS'm'
(lp8215
I1
aI3
aI1
aI15
aI25
aI31
aI59
aI255
aI31
aI697
aI1751
aI381
aI299
aI295
aI14037
aI40953
asS'd'
I4108
sa(dp8216
S'a'
I7204
sS's'
I16
sS'm'
(lp8217
I1
aI3
aI1
aI7
aI15
aI23
aI69
aI219
aI351
aI183
aI1083
aI2227
aI6249
aI9385
aI13167
aI2901
asS'd'
I4109
sa(dp8218
S'a'
I7214
sS's'
I16
sS'm'
(lp8219
I1
aI3
aI7
aI1
aI5
aI61
aI117
aI13
aI67
aI521
aI41
aI2929
aI3937
aI1215
aI25593
aI32627
asS'd'
I4110
sa(dp8220
S'a'
I7222
sS's'
I16
sS'm'
(lp8221
I1
aI3
aI5
aI1
aI9
aI47
aI63
aI39
aI371
aI657
aI491
aI2243
aI4049
aI10517
aI12409
aI40597
asS'd'
I4111
sa(dp8222
S'a'
I7267
sS's'
I16
sS'm'
(lp8223
I1
aI3
aI7
aI15
aI17
aI3
aI77
aI13
aI275
aI225
aI487
aI2055
aI1293
aI15927
aI31773
aI18275
asS'd'
I4112
sa(dp8224
S'a'
I7269
sS's'
I16
sS'm'
(lp8225
I1
aI1
aI5
aI13
aI11
aI57
aI113
aI27
aI191
aI363
aI1341
aI3487
aI8031
aI13801
aI7563
aI40675
asS'd'
I4113
sa(dp8226
S'a'
I7279
sS's'
I16
sS'm'
(lp8227
I1
aI1
aI3
aI3
aI27
aI31
aI103
aI143
aI271
aI305
aI2033
aI3623
aI4219
aI9307
aI7501
aI8959
asS'd'
I4114
sa(dp8228
S'a'
I7293
sS's'
I16
sS'm'
(lp8229
I1
aI1
aI1
aI13
aI1
aI3
aI27
aI97
aI475
aI143
aI333
aI2997
aI1807
aI4231
aI27437
aI59717
asS'd'
I4115
sa(dp8230
S'a'
I7312
sS's'
I16
sS'm'
(lp8231
I1
aI3
aI7
aI5
aI5
aI3
aI69
aI233
aI309
aI511
aI1429
aI1887
aI2745
aI4969
aI17595
aI5451
asS'd'
I4116
sa(dp8232
S'a'
I7327
sS's'
I16
sS'm'
(lp8233
I1
aI1
aI7
aI3
aI23
aI17
aI115
aI89
aI265
aI467
aI257
aI2027
aI5331
aI1195
aI4451
aI8621
asS'd'
I4117
sa(dp8234
S'a'
I7334
sS's'
I16
sS'm'
(lp8235
I1
aI1
aI7
aI13
aI29
aI35
aI117
aI155
aI99
aI327
aI853
aI3595
aI2997
aI10745
aI21619
aI26549
asS'd'
I4118
sa(dp8236
S'a'
I7337
sS's'
I16
sS'm'
(lp8237
I1
aI3
aI3
aI13
aI1
aI13
aI75
aI151
aI67
aI271
aI1609
aI1117
aI4293
aI4645
aI12005
aI55983
asS'd'
I4119
sa(dp8238
S'a'
I7343
sS's'
I16
sS'm'
(lp8239
I1
aI1
aI1
aI13
aI1
aI61
aI101
aI63
aI161
aI261
aI1759
aI567
aI665
aI2339
aI9157
aI55603
asS'd'
I4120
sa(dp8240
S'a'
I7346
sS's'
I16
sS'm'
(lp8241
I1
aI1
aI7
aI11
aI25
aI19
aI71
aI27
aI255
aI435
aI227
aI4087
aI4309
aI14903
aI14513
aI30207
asS'd'
I4121
sa(dp8242
S'a'
I7372
sS's'
I16
sS'm'
(lp8243
I1
aI3
aI3
aI3
aI11
aI41
aI1
aI67
aI383
aI403
aI45
aI1521
aI1535
aI3353
aI27361
aI7549
asS'd'
I4122
sa(dp8244
S'a'
I7387
sS's'
I16
sS'm'
(lp8245
I1
aI1
aI1
aI1
aI13
aI51
aI31
aI137
aI147
aI907
aI19
aI3639
aI3739
aI877
aI15005
aI60357
asS'd'
I4123
sa(dp8246
S'a'
I7390
sS's'
I16
sS'm'
(lp8247
I1
aI1
aI3
aI11
aI11
aI23
aI29
aI173
aI105
aI873
aI1727
aI2761
aI2015
aI7491
aI17491
aI41065
asS'd'
I4124
sa(dp8248
S'a'
I7396
sS's'
I16
sS'm'
(lp8249
I1
aI1
aI5
aI3
aI31
aI1
aI119
aI53
aI11
aI731
aI393
aI4031
aI4421
aI15715
aI6431
aI18089
asS'd'
I4125
sa(dp8250
S'a'
I7423
sS's'
I16
sS'm'
(lp8251
I1
aI1
aI3
aI5
aI15
aI37
aI55
aI147
aI307
aI521
aI711
aI3085
aI5989
aI8081
aI23351
aI35259
asS'd'
I4126
sa(dp8252
S'a'
I7428
sS's'
I16
sS'm'
(lp8253
I1
aI3
aI5
aI13
aI21
aI19
aI47
aI107
aI447
aI713
aI1655
aI2809
aI4741
aI2105
aI9255
aI103
asS'd'
I4127
sa(dp8254
S'a'
I7437
sS's'
I16
sS'm'
(lp8255
I1
aI3
aI1
aI3
aI17
aI47
aI63
aI125
aI343
aI763
aI1777
aI607
aI5625
aI9517
aI7221
aI27257
asS'd'
I4128
sa(dp8256
S'a'
I7466
sS's'
I16
sS'm'
(lp8257
I1
aI1
aI7
aI5
aI31
aI13
aI67
aI255
aI41
aI947
aI99
aI721
aI7367
aI11427
aI1357
aI12383
asS'd'
I4129
sa(dp8258
S'a'
I7474
sS's'
I16
sS'm'
(lp8259
I1
aI1
aI7
aI3
aI23
aI27
aI73
aI185
aI189
aI545
aI1877
aI3169
aI5419
aI15873
aI29059
aI23983
asS'd'
I4130
sa(dp8260
S'a'
I7476
sS's'
I16
sS'm'
(lp8261
I1
aI1
aI3
aI1
aI5
aI13
aI81
aI45
aI79
aI717
aI819
aI3539
aI7561
aI7319
aI5113
aI27273
asS'd'
I4131
sa(dp8262
S'a'
I7483
sS's'
I16
sS'm'
(lp8263
I1
aI3
aI7
aI9
aI21
aI25
aI71
aI247
aI41
aI583
aI771
aI3745
aI1883
aI5717
aI755
aI14549
asS'd'
I4132
sa(dp8264
S'a'
I7518
sS's'
I16
sS'm'
(lp8265
I1
aI1
aI3
aI7
aI23
aI25
aI87
aI143
aI499
aI515
aI1753
aI1229
aI173
aI10629
aI30579
aI29643
asS'd'
I4133
sa(dp8266
S'a'
I7527
sS's'
I16
sS'm'
(lp8267
I1
aI3
aI1
aI13
aI29
aI33
aI31
aI69
aI503
aI117
aI1717
aI101
aI7647
aI1567
aI28677
aI3079
asS'd'
I4134
sa(dp8268
S'a'
I7545
sS's'
I16
sS'm'
(lp8269
I1
aI3
aI1
aI15
aI29
aI45
aI57
aI81
aI171
aI813
aI505
aI3647
aI3913
aI5557
aI2477
aI42429
asS'd'
I4135
sa(dp8270
S'a'
I7572
sS's'
I16
sS'm'
(lp8271
I1
aI1
aI5
aI13
aI21
aI13
aI81
aI5
aI471
aI221
aI1563
aI1741
aI7269
aI16327
aI22687
aI5291
asS'd'
I4136
sa(dp8272
S'a'
I7586
sS's'
I16
sS'm'
(lp8273
I1
aI3
aI5
aI3
aI23
aI41
aI107
aI61
aI95
aI79
aI467
aI1533
aI739
aI6791
aI26911
aI20309
asS'd'
I4137
sa(dp8274
S'a'
I7597
sS's'
I16
sS'm'
(lp8275
I1
aI3
aI7
aI7
aI3
aI53
aI71
aI163
aI459
aI975
aI687
aI1115
aI5241
aI299
aI26361
aI38583
asS'd'
I4138
sa(dp8276
S'a'
I7630
sS's'
I16
sS'm'
(lp8277
I1
aI3
aI1
aI9
aI3
aI63
aI7
aI133
aI421
aI325
aI823
aI1175
aI6201
aI5707
aI31539
aI34645
asS'd'
I4139
sa(dp8278
S'a'
I7653
sS's'
I16
sS'm'
(lp8279
I1
aI3
aI7
aI5
aI27
aI27
aI107
aI239
aI247
aI257
aI1367
aI3685
aI7839
aI11693
aI3237
aI13085
asS'd'
I4140
sa(dp8280
S'a'
I7657
sS's'
I16
sS'm'
(lp8281
I1
aI1
aI1
aI3
aI27
aI41
aI51
aI69
aI271
aI809
aI1453
aI519
aI1301
aI2367
aI637
aI5267
asS'd'
I4141
sa(dp8282
S'a'
I7671
sS's'
I16
sS'm'
(lp8283
I1
aI3
aI7
aI13
aI19
aI17
aI3
aI69
aI369
aI447
aI1685
aI4075
aI6143
aI11387
aI5411
aI29825
asS'd'
I4142
sa(dp8284
S'a'
I7672
sS's'
I16
sS'm'
(lp8285
I1
aI1
aI3
aI1
aI25
aI61
aI79
aI163
aI1
aI905
aI1969
aI2735
aI7709
aI16121
aI20441
aI4543
asS'd'
I4143
sa(dp8286
S'a'
I7715
sS's'
I16
sS'm'
(lp8287
I1
aI3
aI7
aI5
aI15
aI29
aI7
aI245
aI343
aI803
aI1719
aI3993
aI983
aI2925
aI10393
aI6053
asS'd'
I4144
sa(dp8288
S'a'
I7717
sS's'
I16
sS'm'
(lp8289
I1
aI3
aI1
aI7
aI17
aI55
aI63
aI29
aI441
aI387
aI885
aI3269
aI1977
aI1927
aI3657
aI47043
asS'd'
I4145
sa(dp8290
S'a'
I7732
sS's'
I16
sS'm'
(lp8291
I1
aI1
aI3
aI1
aI17
aI59
aI51
aI9
aI173
aI327
aI1185
aI3241
aI3785
aI10907
aI19429
aI22209
asS'd'
I4146
sa(dp8292
S'a'
I7735
sS's'
I16
sS'm'
(lp8293
I1
aI1
aI3
aI13
aI21
aI57
aI125
aI5
aI359
aI437
aI1231
aI2441
aI5813
aI9991
aI283
aI52555
asS'd'
I4147
sa(dp8294
S'a'
I7753
sS's'
I16
sS'm'
(lp8295
I1
aI3
aI1
aI7
aI15
aI19
aI39
aI125
aI405
aI381
aI1757
aI4075
aI5565
aI2065
aI295
aI8867
asS'd'
I4148
sa(dp8296
S'a'
I7811
sS's'
I16
sS'm'
(lp8297
I1
aI3
aI3
aI11
aI7
aI33
aI95
aI19
aI253
aI141
aI1093
aI1787
aI7495
aI5229
aI15923
aI44157
asS'd'
I4149
sa(dp8298
S'a'
I7817
sS's'
I16
sS'm'
(lp8299
I1
aI3
aI7
aI15
aI1
aI49
aI69
aI163
aI85
aI345
aI901
aI2329
aI8003
aI9915
aI2209
aI33979
asS'd'
I4150
sa(dp8300
S'a'
I7825
sS's'
I16
sS'm'
(lp8301
I1
aI1
aI1
aI9
aI23
aI51
aI125
aI163
aI257
aI681
aI565
aI945
aI6375
aI8679
aI5985
aI28919
asS'd'
I4151
sa(dp8302
S'a'
I7832
sS's'
I16
sS'm'
(lp8303
I1
aI3
aI5
aI7
aI11
aI23
aI123
aI231
aI377
aI121
aI1519
aI2061
aI2957
aI14537
aI17625
aI37773
asS'd'
I4152
sa(dp8304
S'a'
I7838
sS's'
I16
sS'm'
(lp8305
I1
aI3
aI5
aI1
aI17
aI43
aI89
aI119
aI455
aI279
aI1857
aI3405
aI5225
aI13035
aI6055
aI30861
asS'd'
I4153
sa(dp8306
S'a'
I7841
sS's'
I16
sS'm'
(lp8307
I1
aI3
aI7
aI15
aI31
aI63
aI25
aI225
aI3
aI527
aI355
aI1435
aI5763
aI15203
aI26579
aI45659
asS'd'
I4154
sa(dp8308
S'a'
I7844
sS's'
I16
sS'm'
(lp8309
I1
aI1
aI1
aI3
aI27
aI43
aI71
aI193
aI135
aI5
aI683
aI925
aI7023
aI7711
aI2807
aI56003
asS'd'
I4155
sa(dp8310
S'a'
I7859
sS's'
I16
sS'm'
(lp8311
I1
aI1
aI1
aI11
aI3
aI3
aI109
aI29
aI109
aI683
aI419
aI3295
aI1961
aI5953
aI8887
aI1523
asS'd'
I4156
sa(dp8312
S'a'
I7861
sS's'
I16
sS'm'
(lp8313
I1
aI3
aI3
aI11
aI17
aI39
aI19
aI225
aI103
aI249
aI81
aI3875
aI4515
aI3969
aI24745
aI60031
asS'd'
I4157
sa(dp8314
S'a'
I7873
sS's'
I16
sS'm'
(lp8315
I1
aI1
aI3
aI3
aI3
aI23
aI15
aI149
aI305
aI399
aI1347
aI1001
aI597
aI10003
aI22123
aI29919
asS'd'
I4158
sa(dp8316
S'a'
I7880
sS's'
I16
sS'm'
(lp8317
I1
aI3
aI5
aI1
aI23
aI35
aI123
aI7
aI283
aI283
aI759
aI3061
aI2011
aI7771
aI32201
aI40667
asS'd'
I4159
sa(dp8318
S'a'
I7897
sS's'
I16
sS'm'
(lp8319
I1
aI3
aI7
aI15
aI23
aI5
aI81
aI51
aI357
aI79
aI133
aI285
aI425
aI7743
aI13499
aI18983
asS'd'
I4160
sa(dp8320
S'a'
I7904
sS's'
I16
sS'm'
(lp8321
I1
aI3
aI3
aI5
aI17
aI37
aI75
aI221
aI473
aI111
aI335
aI683
aI7353
aI2283
aI13457
aI61171
asS'd'
I4161
sa(dp8322
S'a'
I7910
sS's'
I16
sS'm'
(lp8323
I1
aI3
aI1
aI7
aI13
aI45
aI13
aI223
aI149
aI209
aI727
aI3553
aI2573
aI177
aI855
aI44139
asS'd'
I4162
sa(dp8324
S'a'
I7960
sS's'
I16
sS'm'
(lp8325
I1
aI1
aI3
aI15
aI23
aI5
aI103
aI139
aI17
aI29
aI1961
aI3021
aI5745
aI12963
aI30669
aI44171
asS'd'
I4163
sa(dp8326
S'a'
I7969
sS's'
I16
sS'm'
(lp8327
I1
aI3
aI1
aI1
aI3
aI33
aI67
aI203
aI29
aI785
aI71
aI1693
aI4487
aI10221
aI24523
aI51223
asS'd'
I4164
sa(dp8328
S'a'
I7970
sS's'
I16
sS'm'
(lp8329
I1
aI1
aI5
aI7
aI7
aI27
aI17
aI183
aI229
aI669
aI1675
aI3751
aI3233
aI10677
aI7509
aI52313
asS'd'
I4165
sa(dp8330
S'a'
I7976
sS's'
I16
sS'm'
(lp8331
I1
aI1
aI5
aI5
aI25
aI35
aI21
aI163
aI483
aI399
aI195
aI3465
aI6349
aI545
aI18861
aI31759
asS'd'
I4166
sa(dp8332
S'a'
I7979
sS's'
I16
sS'm'
(lp8333
I1
aI3
aI1
aI5
aI15
aI39
aI13
aI157
aI71
aI841
aI447
aI3625
aI53
aI12745
aI2719
aI27617
asS'd'
I4167
sa(dp8334
S'a'
I8007
sS's'
I16
sS'm'
(lp8335
I1
aI1
aI5
aI5
aI3
aI45
aI113
aI121
aI263
aI887
aI709
aI2189
aI3811
aI1409
aI10451
aI48777
asS'd'
I4168
sa(dp8336
S'a'
I8041
sS's'
I16
sS'm'
(lp8337
I1
aI1
aI5
aI15
aI9
aI41
aI31
aI95
aI377
aI215
aI437
aI3633
aI433
aI11935
aI15283
aI55451
asS'd'
I4169
sa(dp8338
S'a'
I8047
sS's'
I16
sS'm'
(lp8339
I1
aI1
aI7
aI7
aI13
aI23
aI79
aI3
aI451
aI409
aI1103
aI1771
aI553
aI3175
aI28703
aI49357
asS'd'
I4170
sa(dp8340
S'a'
I8052
sS's'
I16
sS'm'
(lp8341
I1
aI3
aI1
aI1
aI13
aI33
aI95
aI133
aI419
aI851
aI955
aI3985
aI5869
aI14219
aI253
aI46883
asS'd'
I4171
sa(dp8342
S'a'
I8061
sS's'
I16
sS'm'
(lp8343
I1
aI3
aI3
aI3
aI23
aI55
aI91
aI207
aI281
aI355
aI361
aI261
aI6837
aI4401
aI25455
aI25313
asS'd'
I4172
sa(dp8344
S'a'
I8078
sS's'
I16
sS'm'
(lp8345
I1
aI3
aI5
aI9
aI27
aI9
aI107
aI51
aI69
aI555
aI835
aI3541
aI1827
aI5737
aI31225
aI55619
asS'd'
I4173
sa(dp8346
S'a'
I8128
sS's'
I16
sS'm'
(lp8347
I1
aI1
aI1
aI11
aI27
aI51
aI79
aI85
aI447
aI447
aI9
aI2803
aI377
aI4347
aI2183
aI61257
asS'd'
I4174
sa(dp8348
S'a'
I8146
sS's'
I16
sS'm'
(lp8349
I1
aI1
aI1
aI3
aI23
aI21
aI51
aI217
aI297
aI135
aI573
aI689
aI4947
aI14143
aI26247
aI43061
asS'd'
I4175
sa(dp8350
S'a'
I8162
sS's'
I16
sS'm'
(lp8351
I1
aI3
aI1
aI7
aI15
aI13
aI27
aI151
aI463
aI707
aI43
aI3641
aI4999
aI3783
aI9083
aI22085
asS'd'
I4176
sa(dp8352
S'a'
I8250
sS's'
I16
sS'm'
(lp8353
I1
aI3
aI3
aI5
aI3
aI1
aI15
aI119
aI343
aI605
aI105
aI2939
aI2259
aI889
aI21759
aI34073
asS'd'
I4177
sa(dp8354
S'a'
I8270
sS's'
I16
sS'm'
(lp8355
I1
aI1
aI1
aI7
aI3
aI63
aI103
aI1
aI235
aI317
aI263
aI2701
aI7331
aI15921
aI17295
aI613
asS'd'
I4178
sa(dp8356
S'a'
I8294
sS's'
I16
sS'm'
(lp8357
I1
aI1
aI7
aI3
aI19
aI3
aI5
aI17
aI105
aI491
aI755
aI233
aI5397
aI12365
aI16325
aI59377
asS'd'
I4179
sa(dp8358
S'a'
I8324
sS's'
I16
sS'm'
(lp8359
I1
aI3
aI3
aI15
aI15
aI27
aI37
aI151
aI481
aI949
aI1473
aI233
aI1925
aI15471
aI24957
aI3241
asS'd'
I4180
sa(dp8360
S'a'
I8351
sS's'
I16
sS'm'
(lp8361
I1
aI1
aI7
aI5
aI9
aI61
aI49
aI91
aI169
aI179
aI701
aI3957
aI473
aI15087
aI6109
aI25083
asS'd'
I4181
sa(dp8362
S'a'
I8357
sS's'
I16
sS'm'
(lp8363
I1
aI3
aI3
aI11
aI27
aI43
aI5
aI33
aI69
aI705
aI733
aI2675
aI2677
aI4235
aI11109
aI15557
asS'd'
I4182
sa(dp8364
S'a'
I8361
sS's'
I16
sS'm'
(lp8365
I1
aI3
aI1
aI3
aI17
aI19
aI101
aI119
aI289
aI341
aI1395
aI563
aI6859
aI10359
aI10077
aI26905
asS'd'
I4183
sa(dp8366
S'a'
I8364
sS's'
I16
sS'm'
(lp8367
I1
aI1
aI1
aI15
aI21
aI47
aI41
aI145
aI439
aI583
aI1755
aI1977
aI5235
aI15961
aI21315
aI60577
asS'd'
I4184
sa(dp8368
S'a'
I8393
sS's'
I16
sS'm'
(lp8369
I1
aI1
aI5
aI3
aI9
aI59
aI71
aI143
aI27
aI1007
aI313
aI1567
aI1685
aI11063
aI28889
aI44253
asS'd'
I4185
sa(dp8370
S'a'
I8396
sS's'
I16
sS'm'
(lp8371
I1
aI1
aI5
aI5
aI29
aI29
aI43
aI127
aI13
aI585
aI1245
aI187
aI2697
aI8791
aI19561
aI6463
asS'd'
I4186
sa(dp8372
S'a'
I8407
sS's'
I16
sS'm'
(lp8373
I1
aI1
aI3
aI11
aI29
aI39
aI127
aI75
aI23
aI521
aI421
aI3115
aI139
aI5429
aI23341
aI58035
asS'd'
I4187
sa(dp8374
S'a'
I8413
sS's'
I16
sS'm'
(lp8375
I1
aI1
aI3
aI1
aI27
aI35
aI27
aI9
aI185
aI653
aI887
aI2715
aI3775
aI1753
aI22105
aI62095
asS'd'
I4188
sa(dp8376
S'a'
I8414
sS's'
I16
sS'm'
(lp8377
I1
aI1
aI5
aI5
aI5
aI63
aI23
aI31
aI317
aI1001
aI1751
aI1185
aI7933
aI525
aI30501
aI15565
asS'd'
I4189
sa(dp8378
S'a'
I8432
sS's'
I16
sS'm'
(lp8379
I1
aI1
aI1
aI5
aI9
aI27
aI79
aI91
aI453
aI995
aI1041
aI3213
aI8027
aI5855
aI7435
aI64079
asS'd'
I4190
sa(dp8380
S'a'
I8435
sS's'
I16
sS'm'
(lp8381
I1
aI1
aI3
aI11
aI1
aI51
aI27
aI195
aI139
aI41
aI1891
aI1437
aI1033
aI11671
aI3235
aI35083
asS'd'
I4191
sa(dp8382
S'a'
I8441
sS's'
I16
sS'm'
(lp8383
I1
aI3
aI1
aI3
aI11
aI25
aI33
aI249
aI373
aI497
aI1631
aI293
aI3657
aI10741
aI15831
aI59545
asS'd'
I4192
sa(dp8384
S'a'
I8447
sS's'
I16
sS'm'
(lp8385
I1
aI1
aI1
aI1
aI15
aI63
aI13
aI165
aI113
aI559
aI1615
aI3579
aI1993
aI1907
aI22335
aI47791
asS'd'
I4193
sa(dp8386
S'a'
I8449
sS's'
I16
sS'm'
(lp8387
I1
aI1
aI3
aI15
aI13
aI49
aI63
aI235
aI31
aI811
aI1729
aI221
aI5143
aI11547
aI30029
aI52003
asS'd'
I4194
sa(dp8388
S'a'
I8456
sS's'
I16
sS'm'
(lp8389
I1
aI1
aI5
aI13
aI29
aI61
aI25
aI221
aI421
aI221
aI583
aI373
aI2341
aI7493
aI13981
aI54141
asS'd'
I4195
sa(dp8390
S'a'
I8485
sS's'
I16
sS'm'
(lp8391
I1
aI1
aI5
aI11
aI21
aI49
aI71
aI249
aI237
aI369
aI1273
aI1067
aI4411
aI409
aI7219
aI52933
asS'd'
I4196
sa(dp8392
S'a'
I8504
sS's'
I16
sS'm'
(lp8393
I1
aI3
aI1
aI1
aI13
aI23
aI53
aI15
aI137
aI553
aI401
aI2247
aI5591
aI14021
aI445
aI20433
asS'd'
I4197
sa(dp8394
S'a'
I8512
sS's'
I16
sS'm'
(lp8395
I1
aI1
aI7
aI7
aI7
aI23
aI19
aI189
aI119
aI643
aI847
aI2123
aI5343
aI11839
aI4575
aI60461
asS'd'
I4198
sa(dp8396
S'a'
I8532
sS's'
I16
sS'm'
(lp8397
I1
aI1
aI5
aI5
aI11
aI19
aI111
aI219
aI185
aI499
aI595
aI723
aI3519
aI10891
aI27603
aI29261
asS'd'
I4199
sa(dp8398
S'a'
I8551
sS's'
I16
sS'm'
(lp8399
I1
aI3
aI3
aI1
aI9
aI13
aI95
aI227
aI459
aI133
aI1535
aI3481
aI7187
aI14797
aI16511
aI6737
asS'd'
I4200
sa(dp8400
S'a'
I8560
sS's'
I16
sS'm'
(lp8401
I1
aI1
aI7
aI7
aI19
aI57
aI117
aI7
aI455
aI941
aI455
aI797
aI6313
aI10071
aI18651
aI25013
asS'd'
I4201
sa(dp8402
S'a'
I8566
sS's'
I16
sS'm'
(lp8403
I1
aI3
aI7
aI3
aI25
aI25
aI79
aI19
aI383
aI971
aI1625
aI2803
aI2461
aI633
aI32317
aI48407
asS'd'
I4202
sa(dp8404
S'a'
I8581
sS's'
I16
sS'm'
(lp8405
I1
aI1
aI7
aI7
aI25
aI43
aI93
aI135
aI155
aI685
aI2001
aI3007
aI7315
aI15555
aI30401
aI36291
asS'd'
I4203
sa(dp8406
S'a'
I8609
sS's'
I16
sS'm'
(lp8407
I1
aI1
aI1
aI5
aI13
aI33
aI123
aI221
aI341
aI105
aI1075
aI3125
aI5323
aI14293
aI29875
aI52215
asS'd'
I4204
sa(dp8408
S'a'
I8639
sS's'
I16
sS'm'
(lp8409
I1
aI1
aI3
aI9
aI29
aI51
aI25
aI59
aI171
aI563
aI1695
aI2845
aI6067
aI10671
aI2909
aI33977
asS'd'
I4205
sa(dp8410
S'a'
I8641
sS's'
I16
sS'm'
(lp8411
I1
aI3
aI3
aI7
aI25
aI21
aI39
aI65
aI485
aI949
aI1773
aI2775
aI6019
aI14587
aI6715
aI54793
asS'd'
I4206
sa(dp8412
S'a'
I8671
sS's'
I16
sS'm'
(lp8413
I1
aI1
aI7
aI11
aI17
aI57
aI125
aI17
aI111
aI167
aI289
aI3939
aI7357
aI2289
aI1717
aI45225
asS'd'
I4207
sa(dp8414
S'a'
I8695
sS's'
I16
sS'm'
(lp8415
I1
aI1
aI7
aI7
aI21
aI55
aI3
aI139
aI471
aI747
aI1437
aI1353
aI7657
aI13133
aI14193
aI38191
asS'd'
I4208
sa(dp8416
S'a'
I8701
sS's'
I16
sS'm'
(lp8417
I1
aI3
aI5
aI7
aI25
aI57
aI55
aI17
aI315
aI159
aI437
aI933
aI7493
aI6025
aI2775
aI24287
asS'd'
I4209
sa(dp8418
S'a'
I8711
sS's'
I16
sS'm'
(lp8419
I1
aI1
aI7
aI1
aI15
aI45
aI67
aI191
aI355
aI681
aI1763
aI1237
aI105
aI1425
aI26089
aI42879
asS'd'
I4210
sa(dp8420
S'a'
I8739
sS's'
I16
sS'm'
(lp8421
I1
aI1
aI5
aI13
aI13
aI53
aI25
aI127
aI103
aI155
aI935
aI2561
aI475
aI4341
aI30541
aI43835
asS'd'
I4211
sa(dp8422
S'a'
I8763
sS's'
I16
sS'm'
(lp8423
I1
aI1
aI5
aI15
aI27
aI59
aI99
aI173
aI189
aI41
aI105
aI3287
aI4071
aI15005
aI18301
aI34487
asS'd'
I4212
sa(dp8424
S'a'
I8778
sS's'
I16
sS'm'
(lp8425
I1
aI1
aI5
aI11
aI21
aI9
aI57
aI115
aI495
aI561
aI579
aI397
aI3049
aI2007
aI5041
aI37345
asS'd'
I4213
sa(dp8426
S'a'
I8783
sS's'
I16
sS'm'
(lp8427
I1
aI1
aI5
aI11
aI15
aI11
aI101
aI241
aI69
aI483
aI1007
aI4069
aI5221
aI5323
aI20177
aI24397
asS'd'
I4214
sa(dp8428
S'a'
I8785
sS's'
I16
sS'm'
(lp8429
I1
aI1
aI1
aI7
aI29
aI15
aI119
aI161
aI21
aI517
aI847
aI2217
aI4487
aI4817
aI24053
aI21683
asS'd'
I4215
sa(dp8430
S'a'
I8797
sS's'
I16
sS'm'
(lp8431
I1
aI3
aI1
aI3
aI3
aI51
aI85
aI63
aI345
aI799
aI1699
aI3961
aI7109
aI3931
aI28173
aI46851
asS'd'
I4216
sa(dp8432
S'a'
I8802
sS's'
I16
sS'm'
(lp8433
I1
aI1
aI5
aI7
aI15
aI25
aI97
aI139
aI331
aI969
aI1129
aI2451
aI3107
aI12235
aI12949
aI29837
asS'd'
I4217
sa(dp8434
S'a'
I8816
sS's'
I16
sS'm'
(lp8435
I1
aI3
aI7
aI1
aI19
aI21
aI51
aI155
aI295
aI565
aI29
aI2107
aI341
aI14611
aI15281
aI50727
asS'd'
I4218
sa(dp8436
S'a'
I8828
sS's'
I16
sS'm'
(lp8437
I1
aI3
aI1
aI11
aI3
aI37
aI13
aI217
aI429
aI217
aI301
aI1217
aI5655
aI13845
aI32465
aI23559
asS'd'
I4219
sa(dp8438
S'a'
I8837
sS's'
I16
sS'm'
(lp8439
I1
aI3
aI3
aI9
aI9
aI57
aI79
aI231
aI433
aI703
aI699
aI3813
aI7035
aI5885
aI19185
aI52551
asS'd'
I4220
sa(dp8440
S'a'
I8852
sS's'
I16
sS'm'
(lp8441
I1
aI1
aI1
aI5
aI19
aI17
aI31
aI209
aI49
aI515
aI279
aI909
aI5881
aI2673
aI23635
aI23101
asS'd'
I4221
sa(dp8442
S'a'
I8859
sS's'
I16
sS'm'
(lp8443
I1
aI1
aI5
aI7
aI3
aI3
aI119
aI139
aI245
aI775
aI1009
aI1157
aI1405
aI9737
aI17671
aI62981
asS'd'
I4222
sa(dp8444
S'a'
I8889
sS's'
I16
sS'm'
(lp8445
I1
aI3
aI7
aI11
aI17
aI21
aI105
aI21
aI67
aI871
aI233
aI3607
aI571
aI9141
aI9751
aI28093
asS'd'
I4223
sa(dp8446
S'a'
I8900
sS's'
I16
sS'm'
(lp8447
I1
aI1
aI3
aI13
aI5
aI53
aI91
aI181
aI143
aI375
aI1113
aI705
aI837
aI10505
aI11459
aI57753
asS'd'
I4224
sa(dp8448
S'a'
I8903
sS's'
I16
sS'm'
(lp8449
I1
aI3
aI5
aI11
aI9
aI19
aI107
aI229
aI305
aI107
aI1027
aI691
aI4677
aI8987
aI7931
aI951
asS'd'
I4225
sa(dp8450
S'a'
I8909
sS's'
I16
sS'm'
(lp8451
I1
aI1
aI7
aI9
aI9
aI17
aI39
aI195
aI103
aI315
aI517
aI123
aI7167
aI7039
aI3571
aI40469
asS'd'
I4226
sa(dp8452
S'a'
I8910
sS's'
I16
sS'm'
(lp8453
I1
aI1
aI1
aI5
aI1
aI21
aI121
aI53
aI427
aI111
aI717
aI1065
aI2843
aI5873
aI28411
aI42443
asS'd'
I4227
sa(dp8454
S'a'
I8922
sS's'
I16
sS'm'
(lp8455
I1
aI1
aI3
aI11
aI27
aI7
aI37
aI255
aI429
aI363
aI997
aI2429
aI6871
aI1271
aI29375
aI62897
asS'd'
I4228
sa(dp8456
S'a'
I8924
sS's'
I16
sS'm'
(lp8457
I1
aI3
aI3
aI13
aI23
aI23
aI123
aI119
aI213
aI51
aI771
aI1603
aI1621
aI1497
aI23667
aI13443
asS'd'
I4229
sa(dp8458
S'a'
I8955
sS's'
I16
sS'm'
(lp8459
I1
aI1
aI3
aI13
aI17
aI21
aI81
aI17
aI211
aI815
aI1751
aI3875
aI4001
aI3927
aI6185
aI28753
asS'd'
I4230
sa(dp8460
S'a'
I8958
sS's'
I16
sS'm'
(lp8461
I1
aI3
aI1
aI5
aI13
aI41
aI23
aI187
aI475
aI353
aI1307
aI543
aI5077
aI3459
aI20553
aI29119
asS'd'
I4231
sa(dp8462
S'a'
I8980
sS's'
I16
sS'm'
(lp8463
I1
aI1
aI1
aI7
aI1
aI39
aI3
aI247
aI375
aI101
aI81
aI1515
aI1079
aI15307
aI18865
aI63115
asS'd'
I4232
sa(dp8464
S'a'
I8994
sS's'
I16
sS'm'
(lp8465
I1
aI1
aI5
aI9
aI23
aI7
aI61
aI45
aI379
aI553
aI435
aI1805
aI4147
aI2289
aI22081
aI40041
asS'd'
I4233
sa(dp8466
S'a'
I9006
sS's'
I16
sS'm'
(lp8467
I1
aI1
aI7
aI3
aI17
aI13
aI107
aI169
aI119
aI981
aI1825
aI3329
aI7779
aI12245
aI28367
aI6749
asS'd'
I4234
sa(dp8468
S'a'
I9017
sS's'
I16
sS'm'
(lp8469
I1
aI3
aI3
aI13
aI29
aI13
aI93
aI155
aI331
aI507
aI1073
aI279
aI6615
aI14077
aI3005
aI10171
asS'd'
I4235
sa(dp8470
S'a'
I9032
sS's'
I16
sS'm'
(lp8471
I1
aI1
aI5
aI7
aI29
aI23
aI81
aI181
aI321
aI921
aI1531
aI2607
aI7291
aI1255
aI29889
aI30095
asS'd'
I4236
sa(dp8472
S'a'
I9040
sS's'
I16
sS'm'
(lp8473
I1
aI1
aI1
aI5
aI7
aI1
aI9
aI231
aI203
aI559
aI243
aI3999
aI3649
aI7939
aI14729
aI34771
asS'd'
I4237
sa(dp8474
S'a'
I9061
sS's'
I16
sS'm'
(lp8475
I1
aI3
aI7
aI3
aI17
aI29
aI79
aI251
aI305
aI641
aI1125
aI1155
aI7139
aI6721
aI43
aI34927
asS'd'
I4238
sa(dp8476
S'a'
I9066
sS's'
I16
sS'm'
(lp8477
I1
aI1
aI5
aI13
aI21
aI39
aI55
aI103
aI143
aI487
aI849
aI1111
aI1105
aI8483
aI5417
aI10521
asS'd'
I4239
sa(dp8478
S'a'
I9071
sS's'
I16
sS'm'
(lp8479
I1
aI1
aI5
aI5
aI19
aI5
aI111
aI49
aI95
aI917
aI843
aI2539
aI6831
aI9019
aI16045
aI59363
asS'd'
I4240
sa(dp8480
S'a'
I9076
sS's'
I16
sS'm'
(lp8481
I1
aI3
aI3
aI11
aI7
aI45
aI87
aI51
aI49
aI615
aI603
aI1623
aI5351
aI11939
aI21945
aI40539
asS'd'
I4241
sa(dp8482
S'a'
I9086
sS's'
I16
sS'm'
(lp8483
I1
aI1
aI5
aI9
aI9
aI33
aI113
aI37
aI163
aI853
aI1313
aI4021
aI965
aI11465
aI8573
aI24425
asS'd'
I4242
sa(dp8484
S'a'
I9104
sS's'
I16
sS'm'
(lp8485
I1
aI3
aI3
aI9
aI17
aI19
aI121
aI95
aI93
aI441
aI1951
aI3335
aI6279
aI16087
aI14763
aI60771
asS'd'
I4243
sa(dp8486
S'a'
I9150
sS's'
I16
sS'm'
(lp8487
I1
aI3
aI3
aI9
aI13
aI15
aI19
aI129
aI257
aI641
aI533
aI1667
aI5959
aI2259
aI10439
aI29745
asS'd'
I4244
sa(dp8488
S'a'
I9161
sS's'
I16
sS'm'
(lp8489
I1
aI1
aI7
aI7
aI11
aI31
aI45
aI247
aI233
aI101
aI899
aI2033
aI7803
aI11423
aI30645
aI7723
asS'd'
I4245
sa(dp8490
S'a'
I9164
sS's'
I16
sS'm'
(lp8491
I1
aI3
aI5
aI11
aI23
aI3
aI69
aI79
aI319
aI125
aI1463
aI2047
aI7311
aI5819
aI445
aI13725
asS'd'
I4246
sa(dp8492
S'a'
I9185
sS's'
I16
sS'm'
(lp8493
I1
aI1
aI1
aI3
aI7
aI43
aI83
aI89
aI467
aI709
aI1993
aI3773
aI7805
aI305
aI15701
aI51101
asS'd'
I4247
sa(dp8494
S'a'
I9188
sS's'
I16
sS'm'
(lp8495
I1
aI1
aI7
aI5
aI19
aI53
aI77
aI237
aI27
aI853
aI1003
aI2041
aI5739
aI4663
aI9783
aI23761
asS'd'
I4248
sa(dp8496
S'a'
I9212
sS's'
I16
sS'm'
(lp8497
I1
aI1
aI3
aI7
aI19
aI31
aI71
aI33
aI479
aI693
aI1503
aI9
aI2779
aI1481
aI9413
aI36227
asS'd'
I4249
sa(dp8498
S'a'
I9230
sS's'
I16
sS'm'
(lp8499
I1
aI3
aI1
aI11
aI9
aI23
aI1
aI99
aI247
aI33
aI1987
aI1577
aI8029
aI7785
aI29947
aI38751
asS'd'
I4250
sa(dp8500
S'a'
I9242
sS's'
I16
sS'm'
(lp8501
I1
aI1
aI1
aI3
aI15
aI57
aI23
aI53
aI431
aI553
aI1433
aI2447
aI1871
aI10701
aI30961
aI12281
asS'd'
I4251
sa(dp8502
S'a'
I9247
sS's'
I16
sS'm'
(lp8503
I1
aI3
aI5
aI9
aI11
aI49
aI123
aI91
aI87
aI155
aI301
aI3339
aI6183
aI15895
aI17309
aI45927
asS'd'
I4252
sa(dp8504
S'a'
I9260
sS's'
I16
sS'm'
(lp8505
I1
aI1
aI1
aI9
aI9
aI41
aI79
aI123
aI261
aI547
aI1931
aI2553
aI7405
aI14431
aI24079
aI20769
asS'd'
I4253
sa(dp8506
S'a'
I9280
sS's'
I16
sS'm'
(lp8507
I1
aI3
aI1
aI3
aI31
aI17
aI17
aI137
aI419
aI591
aI1693
aI3977
aI861
aI16025
aI189
aI60995
asS'd'
I4254
sa(dp8508
S'a'
I9300
sS's'
I16
sS'm'
(lp8509
I1
aI3
aI7
aI11
aI19
aI47
aI107
aI243
aI87
aI135
aI507
aI189
aI1397
aI3841
aI22999
aI50781
asS'd'
I4255
sa(dp8510
S'a'
I9313
sS's'
I16
sS'm'
(lp8511
I1
aI3
aI5
aI5
aI15
aI11
aI19
aI239
aI133
aI295
aI673
aI2389
aI4753
aI4363
aI21669
aI25579
asS'd'
I4256
sa(dp8512
S'a'
I9325
sS's'
I16
sS'm'
(lp8513
I1
aI3
aI5
aI7
aI19
aI43
aI55
aI129
aI239
aI89
aI1731
aI1381
aI5483
aI11773
aI9201
aI17745
asS'd'
I4257
sa(dp8514
S'a'
I9343
sS's'
I16
sS'm'
(lp8515
I1
aI3
aI1
aI13
aI3
aI15
aI77
aI131
aI417
aI845
aI1953
aI2871
aI1789
aI10343
aI11363
aI20699
asS'd'
I4258
sa(dp8516
S'a'
I9349
sS's'
I16
sS'm'
(lp8517
I1
aI3
aI7
aI1
aI9
aI43
aI55
aI223
aI239
aI317
aI1951
aI2725
aI209
aI3853
aI2201
aI6839
asS'd'
I4259
sa(dp8518
S'a'
I9354
sS's'
I16
sS'm'
(lp8519
I1
aI3
aI1
aI3
aI7
aI35
aI29
aI21
aI149
aI779
aI467
aI65
aI811
aI4859
aI14021
aI38429
asS'd'
I4260
sa(dp8520
S'a'
I9367
sS's'
I16
sS'm'
(lp8521
I1
aI3
aI7
aI1
aI19
aI31
aI97
aI9
aI11
aI415
aI689
aI1513
aI2475
aI5039
aI5669
aI65103
asS'd'
I4261
sa(dp8522
S'a'
I9368
sS's'
I16
sS'm'
(lp8523
I1
aI3
aI3
aI3
aI11
aI25
aI37
aI247
aI189
aI911
aI429
aI2395
aI3653
aI79
aI28115
aI23513
asS'd'
I4262
sa(dp8524
S'a'
I9455
sS's'
I16
sS'm'
(lp8525
I1
aI1
aI5
aI5
aI5
aI27
aI25
aI45
aI291
aI455
aI741
aI2259
aI8131
aI11779
aI14693
aI58729
asS'd'
I4263
sa(dp8526
S'a'
I9458
sS's'
I16
sS'm'
(lp8527
I1
aI3
aI3
aI7
aI21
aI33
aI67
aI49
aI153
aI491
aI1811
aI1955
aI925
aI15555
aI13801
aI48717
asS'd'
I4264
sa(dp8528
S'a'
I9469
sS's'
I16
sS'm'
(lp8529
I1
aI3
aI1
aI3
aI11
aI53
aI105
aI129
aI457
aI225
aI497
aI1123
aI973
aI2901
aI26655
aI3643
asS'd'
I4265
sa(dp8530
S'a'
I9481
sS's'
I16
sS'm'
(lp8531
I1
aI1
aI7
aI13
aI29
aI49
aI71
aI37
aI321
aI865
aI735
aI357
aI7629
aI6011
aI28221
aI39041
asS'd'
I4266
sa(dp8532
S'a'
I9482
sS's'
I16
sS'm'
(lp8533
I1
aI3
aI5
aI3
aI19
aI59
aI65
aI199
aI69
aI391
aI1735
aI3075
aI287
aI16213
aI3211
aI22425
asS'd'
I4267
sa(dp8534
S'a'
I9495
sS's'
I16
sS'm'
(lp8535
I1
aI1
aI1
aI5
aI15
aI61
aI67
aI255
aI5
aI689
aI759
aI155
aI7245
aI5881
aI21685
aI3863
asS'd'
I4268
sa(dp8536
S'a'
I9526
sS's'
I16
sS'm'
(lp8537
I1
aI1
aI3
aI11
aI23
aI63
aI69
aI241
aI359
aI735
aI371
aI603
aI2707
aI15833
aI31795
aI14901
asS'd'
I4269
sa(dp8538
S'a'
I9530
sS's'
I16
sS'm'
(lp8539
I1
aI1
aI1
aI7
aI19
aI33
aI83
aI19
aI13
aI667
aI317
aI3891
aI5497
aI8213
aI4913
aI22387
asS'd'
I4270
sa(dp8540
S'a'
I9558
sS's'
I16
sS'm'
(lp8541
I1
aI3
aI5
aI9
aI13
aI21
aI11
aI187
aI249
aI647
aI349
aI605
aI2199
aI5033
aI29773
aI48953
asS'd'
I4271
sa(dp8542
S'a'
I9562
sS's'
I16
sS'm'
(lp8543
I1
aI3
aI3
aI11
aI3
aI3
aI93
aI235
aI441
aI933
aI383
aI2007
aI4015
aI4175
aI3937
aI20623
asS'd'
I4272
sa(dp8544
S'a'
I9573
sS's'
I16
sS'm'
(lp8545
I1
aI3
aI7
aI13
aI3
aI61
aI87
aI219
aI263
aI651
aI1343
aI2709
aI31
aI16249
aI4749
aI28909
asS'd'
I4273
sa(dp8546
S'a'
I9583
sS's'
I16
sS'm'
(lp8547
I1
aI3
aI1
aI1
aI17
aI19
aI101
aI107
aI499
aI127
aI13
aI2123
aI5597
aI3751
aI14527
aI12009
asS'd'
I4274
sa(dp8548
S'a'
I9595
sS's'
I16
sS'm'
(lp8549
I1
aI3
aI5
aI13
aI27
aI57
aI1
aI195
aI107
aI947
aI1475
aI2831
aI6449
aI16117
aI20555
aI61513
asS'd'
I4275
sa(dp8550
S'a'
I9597
sS's'
I16
sS'm'
(lp8551
I1
aI3
aI1
aI9
aI9
aI47
aI89
aI187
aI401
aI299
aI637
aI197
aI1235
aI12655
aI25025
aI24443
asS'd'
I4276
sa(dp8552
S'a'
I9616
sS's'
I16
sS'm'
(lp8553
I1
aI1
aI3
aI5
aI9
aI57
aI33
aI41
aI451
aI983
aI391
aI2707
aI705
aI13913
aI28843
aI34091
asS'd'
I4277
sa(dp8554
S'a'
I9638
sS's'
I16
sS'm'
(lp8555
I1
aI3
aI5
aI3
aI29
aI19
aI61
aI31
aI349
aI253
aI1267
aI3345
aI2151
aI11309
aI19623
aI62407
asS'd'
I4278
sa(dp8556
S'a'
I9649
sS's'
I16
sS'm'
(lp8557
I1
aI1
aI1
aI3
aI11
aI37
aI31
aI59
aI1
aI253
aI103
aI2811
aI1871
aI4239
aI26311
aI32507
asS'd'
I4279
sa(dp8558
S'a'
I9662
sS's'
I16
sS'm'
(lp8559
I1
aI1
aI5
aI7
aI7
aI7
aI69
aI63
aI281
aI901
aI1785
aI2131
aI4265
aI253
aI13997
aI30177
asS'd'
I4280
sa(dp8560
S'a'
I9691
sS's'
I16
sS'm'
(lp8561
I1
aI3
aI1
aI11
aI3
aI27
aI111
aI67
aI411
aI751
aI241
aI293
aI6271
aI4187
aI22119
aI63737
asS'd'
I4281
sa(dp8562
S'a'
I9700
sS's'
I16
sS'm'
(lp8563
I1
aI3
aI5
aI5
aI27
aI19
aI45
aI203
aI81
aI59
aI1839
aI935
aI4847
aI1869
aI12037
aI30971
asS'd'
I4282
sa(dp8564
S'a'
I9703
sS's'
I16
sS'm'
(lp8565
I1
aI1
aI3
aI9
aI19
aI25
aI9
aI27
aI373
aI875
aI1735
aI689
aI2957
aI7873
aI29771
aI4093
asS'd'
I4283
sa(dp8566
S'a'
I9710
sS's'
I16
sS'm'
(lp8567
I1
aI1
aI7
aI11
aI13
aI39
aI11
aI129
aI53
aI433
aI1731
aI899
aI5855
aI10221
aI24165
aI50205
asS'd'
I4284
sa(dp8568
S'a'
I9721
sS's'
I16
sS'm'
(lp8569
I1
aI3
aI1
aI15
aI25
aI31
aI85
aI49
aI325
aI299
aI183
aI287
aI2425
aI15353
aI25999
aI59129
asS'd'
I4285
sa(dp8570
S'a'
I9724
sS's'
I16
sS'm'
(lp8571
I1
aI1
aI5
aI5
aI17
aI25
aI23
aI5
aI287
aI677
aI591
aI981
aI429
aI15297
aI14573
aI61095
asS'd'
I4286
sa(dp8572
S'a'
I9727
sS's'
I16
sS'm'
(lp8573
I1
aI1
aI5
aI15
aI5
aI15
aI67
aI195
aI209
aI341
aI1621
aI1379
aI3031
aI5469
aI31563
aI49291
asS'd'
I4287
sa(dp8574
S'a'
I9743
sS's'
I16
sS'm'
(lp8575
I1
aI1
aI1
aI1
aI25
aI33
aI61
aI201
aI15
aI125
aI969
aI1965
aI2021
aI10253
aI23801
aI28165
asS'd'
I4288
sa(dp8576
S'a'
I9779
sS's'
I16
sS'm'
(lp8577
I1
aI1
aI5
aI5
aI13
aI17
aI5
aI245
aI11
aI133
aI287
aI1929
aI4331
aI15919
aI29663
aI10243
asS'd'
I4289
sa(dp8578
S'a'
I9785
sS's'
I16
sS'm'
(lp8579
I1
aI1
aI7
aI9
aI19
aI33
aI39
aI191
aI489
aI631
aI69
aI1883
aI2183
aI14993
aI32715
aI62217
asS'd'
I4290
sa(dp8580
S'a'
I9811
sS's'
I16
sS'm'
(lp8581
I1
aI3
aI3
aI13
aI23
aI61
aI103
aI193
aI501
aI431
aI437
aI417
aI6557
aI11701
aI27577
aI42943
asS'd'
I4291
sa(dp8582
S'a'
I9820
sS's'
I16
sS'm'
(lp8583
I1
aI3
aI3
aI9
aI9
aI27
aI69
aI247
aI469
aI841
aI733
aI813
aI2673
aI7145
aI5385
aI44917
asS'd'
I4292
sa(dp8584
S'a'
I9827
sS's'
I16
sS'm'
(lp8585
I1
aI1
aI7
aI9
aI5
aI13
aI19
aI71
aI323
aI923
aI1885
aI3031
aI4507
aI13787
aI14149
aI1483
asS'd'
I4293
sa(dp8586
S'a'
I9851
sS's'
I16
sS'm'
(lp8587
I1
aI3
aI1
aI5
aI1
aI15
aI89
aI229
aI301
aI733
aI1343
aI2415
aI6463
aI1875
aI9293
aI6037
asS'd'
I4294
sa(dp8588
S'a'
I9854
sS's'
I16
sS'm'
(lp8589
I1
aI3
aI1
aI5
aI29
aI57
aI67
aI121
aI311
aI441
aI1079
aI1963
aI7137
aI6745
aI9893
aI22811
asS'd'
I4295
sa(dp8590
S'a'
I9863
sS's'
I16
sS'm'
(lp8591
I1
aI1
aI7
aI7
aI25
aI13
aI27
aI61
aI331
aI361
aI481
aI3783
aI3061
aI7827
aI18885
aI27583
asS'd'
I4296
sa(dp8592
S'a'
I9884
sS's'
I16
sS'm'
(lp8593
I1
aI3
aI1
aI5
aI17
aI47
aI19
aI83
aI309
aI65
aI1573
aI3437
aI5623
aI12691
aI21075
aI27789
asS'd'
I4297
sa(dp8594
S'a'
I9894
sS's'
I16
sS'm'
(lp8595
I1
aI1
aI7
aI9
aI13
aI51
aI7
aI209
aI131
aI111
aI1143
aI53
aI7277
aI9297
aI20869
aI33121
asS'd'
I4298
sa(dp8596
S'a'
I9903
sS's'
I16
sS'm'
(lp8597
I1
aI1
aI3
aI9
aI13
aI17
aI57
aI89
aI239
aI281
aI775
aI333
aI5631
aI2805
aI10195
aI9665
asS'd'
I4299
sa(dp8598
S'a'
I9908
sS's'
I16
sS'm'
(lp8599
I1
aI1
aI3
aI7
aI19
aI39
aI71
aI79
aI63
aI551
aI103
aI3169
aI2761
aI13929
aI20751
aI18951
asS'd'
I4300
sa(dp8600
S'a'
I9912
sS's'
I16
sS'm'
(lp8601
I1
aI1
aI7
aI11
aI5
aI23
aI7
aI249
aI447
aI483
aI433
aI3117
aI1421
aI14991
aI5397
aI19813
asS'd'
I4301
sa(dp8602
S'a'
I9925
sS's'
I16
sS'm'
(lp8603
I1
aI3
aI1
aI13
aI3
aI9
aI109
aI241
aI181
aI33
aI853
aI3939
aI3751
aI2151
aI28375
aI64443
asS'd'
I4302
sa(dp8604
S'a'
I9926
sS's'
I16
sS'm'
(lp8605
I1
aI1
aI7
aI7
aI3
aI33
aI65
aI211
aI251
aI631
aI1819
aI3913
aI3353
aI12975
aI19117
aI51515
asS'd'
I4303
sa(dp8606
S'a'
I9971
sS's'
I16
sS'm'
(lp8607
I1
aI1
aI1
aI13
aI3
aI21
aI9
aI203
aI223
aI229
aI1399
aI117
aI6297
aI11535
aI16383
aI12541
asS'd'
I4304
sa(dp8608
S'a'
I9973
sS's'
I16
sS'm'
(lp8609
I1
aI1
aI5
aI7
aI25
aI9
aI53
aI27
aI497
aI103
aI1915
aI2179
aI3679
aI11375
aI18907
aI63385
asS'd'
I4305
sa(dp8610
S'a'
I9977
sS's'
I16
sS'm'
(lp8611
I1
aI3
aI1
aI5
aI1
aI53
aI91
aI169
aI87
aI387
aI377
aI1121
aI7241
aI5133
aI1949
aI13433
asS'd'
I4306
sa(dp8612
S'a'
I10021
sS's'
I16
sS'm'
(lp8613
I1
aI1
aI1
aI3
aI27
aI35
aI61
aI189
aI241
aI445
aI287
aI325
aI127
aI2363
aI30663
aI43557
asS'd'
I4307
sa(dp8614
S'a'
I10039
sS's'
I16
sS'm'
(lp8615
I1
aI3
aI1
aI3
aI17
aI47
aI59
aI237
aI213
aI979
aI807
aI85
aI4621
aI9915
aI13631
aI55657
asS'd'
I4308
sa(dp8616
S'a'
I10048
sS's'
I16
sS'm'
(lp8617
I1
aI3
aI5
aI7
aI27
aI5
aI101
aI89
aI495
aI675
aI1181
aI2295
aI1913
aI3731
aI32639
aI58297
asS'd'
I4309
sa(dp8618
S'a'
I10051
sS's'
I16
sS'm'
(lp8619
I1
aI3
aI3
aI11
aI5
aI41
aI49
aI229
aI93
aI659
aI927
aI3425
aI4083
aI11859
aI10603
aI20631
asS'd'
I4310
sa(dp8620
S'a'
I10065
sS's'
I16
sS'm'
(lp8621
I1
aI3
aI5
aI11
aI31
aI51
aI67
aI69
aI253
aI497
aI1665
aI1985
aI5439
aI15999
aI4175
aI62175
asS'd'
I4311
sa(dp8622
S'a'
I10071
sS's'
I16
sS'm'
(lp8623
I1
aI1
aI7
aI11
aI1
aI21
aI95
aI19
aI205
aI513
aI1329
aI1821
aI1251
aI2381
aI32623
aI23923
asS'd'
I4312
sa(dp8624
S'a'
I10072
sS's'
I16
sS'm'
(lp8625
I1
aI1
aI5
aI13
aI3
aI1
aI29
aI175
aI315
aI865
aI599
aI1695
aI1391
aI2313
aI31035
aI19159
asS'd'
I4313
sa(dp8626
S'a'
I10101
sS's'
I16
sS'm'
(lp8627
I1
aI3
aI3
aI1
aI3
aI45
aI109
aI93
aI481
aI285
aI869
aI3441
aI3715
aI1355
aI9581
aI50173
asS'd'
I4314
sa(dp8628
S'a'
I10106
sS's'
I16
sS'm'
(lp8629
I1
aI3
aI7
aI7
aI15
aI35
aI107
aI107
aI315
aI213
aI281
aI2073
aI4689
aI5877
aI31
aI40967
asS'd'
I4315
sa(dp8630
S'a'
I10130
sS's'
I16
sS'm'
(lp8631
I1
aI1
aI3
aI11
aI11
aI3
aI73
aI41
aI79
aI37
aI481
aI1993
aI931
aI7677
aI11321
aI45735
asS'd'
I4316
sa(dp8632
S'a'
I10136
sS's'
I16
sS'm'
(lp8633
I1
aI1
aI7
aI1
aI15
aI21
aI65
aI237
aI263
aI849
aI863
aI4039
aI3171
aI13381
aI30373
aI51639
asS'd'
I4317
sa(dp8634
S'a'
I10148
sS's'
I16
sS'm'
(lp8635
I1
aI1
aI1
aI3
aI21
aI57
aI113
aI3
aI487
aI41
aI1277
aI633
aI2839
aI4977
aI14537
aI31749
asS'd'
I4318
sa(dp8636
S'a'
I10155
sS's'
I16
sS'm'
(lp8637
I1
aI1
aI5
aI1
aI1
aI55
aI71
aI181
aI147
aI895
aI1839
aI2157
aI3187
aI6403
aI30367
aI48915
asS'd'
I4319
sa(dp8638
S'a'
I10157
sS's'
I16
sS'm'
(lp8639
I1
aI1
aI5
aI3
aI9
aI17
aI19
aI127
aI115
aI875
aI831
aI2439
aI7475
aI1921
aI18657
aI27709
asS'd'
I4320
sa(dp8640
S'a'
I10160
sS's'
I16
sS'm'
(lp8641
I1
aI3
aI3
aI13
aI29
aI11
aI35
aI81
aI291
aI483
aI625
aI3957
aI6017
aI12543
aI18773
aI2471
asS'd'
I4321
sa(dp8642
S'a'
I10166
sS's'
I16
sS'm'
(lp8643
I1
aI3
aI3
aI13
aI11
aI39
aI7
aI85
aI493
aI209
aI819
aI3277
aI4275
aI8997
aI22943
aI33199
asS'd'
I4322
sa(dp8644
S'a'
I10183
sS's'
I16
sS'm'
(lp8645
I1
aI1
aI1
aI7
aI11
aI25
aI1
aI57
aI505
aI135
aI1713
aI3051
aI5893
aI10711
aI10681
aI12235
asS'd'
I4323
sa(dp8646
S'a'
I10192
sS's'
I16
sS'm'
(lp8647
I1
aI3
aI5
aI11
aI23
aI33
aI13
aI107
aI289
aI251
aI235
aI1747
aI4097
aI12237
aI17559
aI5063
asS'd'
I4324
sa(dp8648
S'a'
I10225
sS's'
I16
sS'm'
(lp8649
I1
aI3
aI3
aI9
aI19
aI17
aI23
aI45
aI297
aI147
aI1301
aI2057
aI7871
aI9971
aI1965
aI62449
asS'd'
I4325
sa(dp8650
S'a'
I10241
sS's'
I16
sS'm'
(lp8651
I1
aI1
aI7
aI3
aI17
aI21
aI19
aI203
aI289
aI897
aI1967
aI3519
aI3261
aI8199
aI24181
aI23273
asS'd'
I4326
sa(dp8652
S'a'
I10247
sS's'
I16
sS'm'
(lp8653
I1
aI1
aI7
aI11
aI1
aI17
aI25
aI63
aI509
aI969
aI47
aI1329
aI701
aI5227
aI419
aI14839
asS'd'
I4327
sa(dp8654
S'a'
I10284
sS's'
I16
sS'm'
(lp8655
I1
aI3
aI5
aI11
aI1
aI41
aI81
aI157
aI395
aI97
aI1919
aI3043
aI6015
aI15
aI23733
aI55485
asS'd'
I4328
sa(dp8656
S'a'
I10304
sS's'
I16
sS'm'
(lp8657
I1
aI1
aI3
aI11
aI17
aI37
aI17
aI181
aI179
aI297
aI1007
aI1559
aI6275
aI11645
aI7535
aI28941
asS'd'
I4329
sa(dp8658
S'a'
I10322
sS's'
I16
sS'm'
(lp8659
I1
aI3
aI7
aI15
aI5
aI47
aI31
aI237
aI215
aI563
aI207
aI1867
aI6635
aI6857
aI11837
aI22865
asS'd'
I4330
sa(dp8660
S'a'
I10331
sS's'
I16
sS'm'
(lp8661
I1
aI3
aI1
aI7
aI7
aI39
aI51
aI179
aI57
aI987
aI893
aI2715
aI1045
aI5799
aI19805
aI54275
asS'd'
I4331
sa(dp8662
S'a'
I10333
sS's'
I16
sS'm'
(lp8663
I1
aI1
aI7
aI15
aI25
aI9
aI127
aI243
aI323
aI1013
aI929
aI2891
aI7549
aI1071
aI17663
aI15247
asS'd'
I4332
sa(dp8664
S'a'
I10340
sS's'
I16
sS'm'
(lp8665
I1
aI1
aI1
aI5
aI25
aI23
aI101
aI9
aI371
aI89
aI1749
aI3559
aI8071
aI13887
aI14807
aI42825
asS'd'
I4333
sa(dp8666
S'a'
I10347
sS's'
I16
sS'm'
(lp8667
I1
aI3
aI3
aI11
aI21
aI41
aI3
aI77
aI3
aI709
aI1745
aI3615
aI4141
aI5275
aI28329
aI59739
asS'd'
I4334
sa(dp8668
S'a'
I10357
sS's'
I16
sS'm'
(lp8669
I1
aI1
aI7
aI13
aI1
aI11
aI73
aI183
aI363
aI241
aI863
aI3983
aI3235
aI293
aI649
aI441
asS'd'
I4335
sa(dp8670
S'a'
I10371
sS's'
I16
sS'm'
(lp8671
I1
aI1
aI5
aI3
aI13
aI27
aI13
aI191
aI57
aI639
aI1803
aI2353
aI3143
aI12763
aI5771
aI36155
asS'd'
I4336
sa(dp8672
S'a'
I10386
sS's'
I16
sS'm'
(lp8673
I1
aI1
aI5
aI3
aI1
aI53
aI85
aI45
aI283
aI823
aI1213
aI3261
aI4599
aI13267
aI4613
aI12657
asS'd'
I4337
sa(dp8674
S'a'
I10404
sS's'
I16
sS'm'
(lp8675
I1
aI3
aI5
aI15
aI27
aI49
aI59
aI123
aI357
aI527
aI337
aI2751
aI3999
aI8525
aI12501
aI40621
asS'd'
I4338
sa(dp8676
S'a'
I10414
sS's'
I16
sS'm'
(lp8677
I1
aI1
aI1
aI7
aI1
aI55
aI85
aI17
aI447
aI599
aI1315
aI2313
aI1649
aI907
aI25647
aI3251
asS'd'
I4339
sa(dp8678
S'a'
I10422
sS's'
I16
sS'm'
(lp8679
I1
aI3
aI5
aI13
aI9
aI1
aI37
aI121
aI143
aI1
aI631
aI2273
aI1673
aI3649
aI27533
aI28731
asS'd'
I4340
sa(dp8680
S'a'
I10448
sS's'
I16
sS'm'
(lp8681
I1
aI1
aI7
aI13
aI9
aI31
aI57
aI249
aI397
aI815
aI501
aI895
aI1217
aI11387
aI8635
aI65193
asS'd'
I4341
sa(dp8682
S'a'
I10451
sS's'
I16
sS'm'
(lp8683
I1
aI1
aI5
aI5
aI9
aI35
aI95
aI193
aI133
aI513
aI1929
aI3841
aI3063
aI2383
aI24413
aI51185
asS'd'
I4342
sa(dp8684
S'a'
I10473
sS's'
I16
sS'm'
(lp8685
I1
aI1
aI1
aI13
aI3
aI49
aI45
aI191
aI15
aI181
aI1819
aI3741
aI1227
aI12775
aI9461
aI44951
asS'd'
I4343
sa(dp8686
S'a'
I10479
sS's'
I16
sS'm'
(lp8687
I1
aI1
aI1
aI1
aI3
aI45
aI121
aI19
aI269
aI829
aI517
aI3913
aI183
aI11019
aI24969
aI21973
asS'd'
I4344
sa(dp8688
S'a'
I10501
sS's'
I16
sS'm'
(lp8689
I1
aI1
aI5
aI11
aI31
aI39
aI125
aI189
aI401
aI57
aI1557
aI1727
aI1415
aI4025
aI30353
aI36589
asS'd'
I4345
sa(dp8690
S'a'
I10530
sS's'
I16
sS'm'
(lp8691
I1
aI1
aI3
aI9
aI3
aI55
aI125
aI187
aI409
aI499
aI1853
aI2781
aI4323
aI16159
aI16345
aI34659
asS'd'
I4346
sa(dp8692
S'a'
I10542
sS's'
I16
sS'm'
(lp8693
I1
aI3
aI5
aI11
aI31
aI5
aI125
aI137
aI197
aI475
aI2003
aI617
aI1289
aI8365
aI28981
aI57537
asS'd'
I4347
sa(dp8694
S'a'
I10544
sS's'
I16
sS'm'
(lp8695
I1
aI1
aI1
aI5
aI19
aI29
aI83
aI127
aI311
aI177
aI1635
aI2187
aI5377
aI12841
aI7591
aI6095
asS'd'
I4348
sa(dp8696
S'a'
I10571
sS's'
I16
sS'm'
(lp8697
I1
aI3
aI5
aI5
aI21
aI39
aI65
aI197
aI115
aI411
aI1457
aI3011
aI7021
aI14119
aI61
aI21107
asS'd'
I4349
sa(dp8698
S'a'
I10628
sS's'
I16
sS'm'
(lp8699
I1
aI3
aI3
aI9
aI19
aI57
aI99
aI23
aI451
aI507
aI973
aI415
aI7123
aI11367
aI29767
aI23763
asS'd'
I4350
sa(dp8700
S'a'
I10643
sS's'
I16
sS'm'
(lp8701
I1
aI1
aI5
aI7
aI29
aI23
aI47
aI11
aI267
aI873
aI591
aI373
aI7097
aI3783
aI23289
aI5547
asS'd'
I4351
sa(dp8702
S'a'
I10673
sS's'
I16
sS'm'
(lp8703
I1
aI1
aI5
aI15
aI7
aI7
aI7
aI91
aI389
aI841
aI1995
aI459
aI7013
aI3109
aI23615
aI21519
asS'd'
I4352
sa(dp8704
S'a'
I10683
sS's'
I16
sS'm'
(lp8705
I1
aI3
aI1
aI1
aI13
aI25
aI87
aI235
aI193
aI201
aI713
aI1633
aI3117
aI13609
aI17211
aI573
asS'd'
I4353
sa(dp8706
S'a'
I10736
sS's'
I16
sS'm'
(lp8707
I1
aI1
aI1
aI9
aI27
aI53
aI105
aI39
aI217
aI781
aI997
aI661
aI6243
aI6427
aI17739
aI62239
asS'd'
I4354
sa(dp8708
S'a'
I10795
sS's'
I16
sS'm'
(lp8709
I1
aI1
aI7
aI3
aI3
aI49
aI75
aI125
aI239
aI195
aI215
aI2983
aI1185
aI4743
aI12069
aI55509
asS'd'
I4355
sa(dp8710
S'a'
I10797
sS's'
I16
sS'm'
(lp8711
I1
aI1
aI5
aI15
aI31
aI11
aI9
aI91
aI253
aI361
aI571
aI1589
aI2425
aI4279
aI3765
aI46519
asS'd'
I4356
sa(dp8712
S'a'
I10815
sS's'
I16
sS'm'
(lp8713
I1
aI1
aI3
aI3
aI21
aI49
aI49
aI213
aI399
aI253
aI1565
aI2447
aI453
aI7957
aI24799
aI58503
asS'd'
I4357
sa(dp8714
S'a'
I10817
sS's'
I16
sS'm'
(lp8715
I1
aI1
aI7
aI1
aI5
aI59
aI81
aI97
aI209
aI477
aI17
aI3085
aI3655
aI599
aI24011
aI14981
asS'd'
I4358
sa(dp8716
S'a'
I10842
sS's'
I16
sS'm'
(lp8717
I1
aI3
aI3
aI13
aI19
aI49
aI7
aI35
aI111
aI169
aI629
aI1587
aI5345
aI13699
aI21187
aI30199
asS'd'
I4359
sa(dp8718
S'a'
I10844
sS's'
I16
sS'm'
(lp8719
I1
aI1
aI3
aI13
aI19
aI59
aI73
aI127
aI475
aI509
aI9
aI2661
aI711
aI15835
aI31429
aI33885
asS'd'
I4360
sa(dp8720
S'a'
I10863
sS's'
I16
sS'm'
(lp8721
I1
aI3
aI5
aI3
aI31
aI15
aI43
aI185
aI29
aI897
aI1041
aI1057
aI6285
aI13925
aI4023
aI25741
asS'd'
I4361
sa(dp8722
S'a'
I10899
sS's'
I16
sS'm'
(lp8723
I1
aI3
aI5
aI11
aI1
aI33
aI63
aI155
aI175
aI501
aI1147
aI3395
aI3285
aI16237
aI22315
aI50945
asS'd'
I4362
sa(dp8724
S'a'
I10902
sS's'
I16
sS'm'
(lp8725
I1
aI3
aI3
aI3
aI5
aI13
aI77
aI227
aI85
aI139
aI139
aI1
aI7147
aI2023
aI32705
aI38753
asS'd'
I4363
sa(dp8726
S'a'
I10917
sS's'
I16
sS'm'
(lp8727
I1
aI3
aI5
aI11
aI31
aI59
aI35
aI179
aI489
aI537
aI1537
aI2877
aI4937
aI10825
aI2445
aI34907
asS'd'
I4364
sa(dp8728
S'a'
I10953
sS's'
I16
sS'm'
(lp8729
I1
aI3
aI7
aI11
aI17
aI17
aI95
aI223
aI165
aI925
aI829
aI3971
aI1
aI7393
aI8825
aI25705
asS'd'
I4365
sa(dp8730
S'a'
I10967
sS's'
I16
sS'm'
(lp8731
I1
aI1
aI1
aI1
aI25
aI17
aI25
aI143
aI385
aI907
aI1381
aI1823
aI3819
aI8475
aI5321
aI12037
asS'd'
I4366
sa(dp8732
S'a'
I11002
sS's'
I16
sS'm'
(lp8733
I1
aI1
aI5
aI11
aI7
aI47
aI97
aI85
aI105
aI23
aI263
aI1329
aI1905
aI12121
aI29635
aI41249
asS'd'
I4367
sa(dp8734
S'a'
I11024
sS's'
I16
sS'm'
(lp8735
I1
aI1
aI7
aI11
aI1
aI51
aI13
aI13
aI5
aI143
aI83
aI3831
aI959
aI6083
aI16997
aI59887
asS'd'
I4368
sa(dp8736
S'a'
I11029
sS's'
I16
sS'm'
(lp8737
I1
aI3
aI3
aI13
aI25
aI9
aI31
aI5
aI215
aI791
aI767
aI1733
aI2715
aI14283
aI25795
aI54249
asS'd'
I4369
sa(dp8738
S'a'
I11030
sS's'
I16
sS'm'
(lp8739
I1
aI3
aI7
aI5
aI11
aI19
aI125
aI81
aI71
aI131
aI1869
aI1111
aI6763
aI5275
aI23095
aI1139
asS'd'
I4370
sa(dp8740
S'a'
I11043
sS's'
I16
sS'm'
(lp8741
I1
aI3
aI3
aI9
aI25
aI43
aI119
aI49
aI133
aI217
aI521
aI1367
aI5867
aI6829
aI29871
aI60383
asS'd'
I4371
sa(dp8742
S'a'
I11087
sS's'
I16
sS'm'
(lp8743
I1
aI1
aI7
aI9
aI5
aI11
aI59
aI157
aI279
aI301
aI481
aI3273
aI7943
aI3273
aI27783
aI17271
asS'd'
I4372
sa(dp8744
S'a'
I11106
sS's'
I16
sS'm'
(lp8745
I1
aI3
aI3
aI13
aI11
aI57
aI13
aI5
aI435
aI169
aI541
aI517
aI2359
aI9121
aI27911
aI15679
asS'd'
I4373
sa(dp8746
S'a'
I11130
sS's'
I16
sS'm'
(lp8747
I1
aI1
aI3
aI9
aI9
aI55
aI65
aI113
aI21
aI807
aI373
aI2825
aI1527
aI15565
aI8339
aI7227
asS'd'
I4374
sa(dp8748
S'a'
I11156
sS's'
I16
sS'm'
(lp8749
I1
aI3
aI5
aI9
aI1
aI1
aI49
aI255
aI477
aI821
aI1647
aI713
aI6841
aI3187
aI22649
aI51469
asS'd'
I4375
sa(dp8750
S'a'
I11176
sS's'
I16
sS'm'
(lp8751
I1
aI3
aI3
aI11
aI13
aI43
aI63
aI139
aI71
aI809
aI993
aI2429
aI833
aI6545
aI10987
aI39567
asS'd'
I4376
sa(dp8752
S'a'
I11221
sS's'
I16
sS'm'
(lp8753
I1
aI1
aI1
aI9
aI19
aI23
aI47
aI199
aI191
aI827
aI391
aI837
aI1209
aI2493
aI24071
aI46589
asS'd'
I4377
sa(dp8754
S'a'
I11267
sS's'
I16
sS'm'
(lp8755
I1
aI1
aI5
aI13
aI25
aI51
aI39
aI43
aI103
aI899
aI1729
aI2389
aI2965
aI189
aI3063
aI50609
asS'd'
I4378
sa(dp8756
S'a'
I11282
sS's'
I16
sS'm'
(lp8757
I1
aI1
aI3
aI1
aI5
aI29
aI105
aI201
aI503
aI199
aI507
aI205
aI2389
aI695
aI15233
aI50353
asS'd'
I4379
sa(dp8758
S'a'
I11294
sS's'
I16
sS'm'
(lp8759
I1
aI3
aI1
aI7
aI19
aI53
aI45
aI21
aI89
aI545
aI1885
aI765
aI6673
aI13485
aI9987
aI2609
asS'd'
I4380
sa(dp8760
S'a'
I11332
sS's'
I16
sS'm'
(lp8761
I1
aI3
aI7
aI13
aI17
aI7
aI59
aI23
aI159
aI309
aI1407
aI2483
aI1807
aI15733
aI5603
aI52989
asS'd'
I4381
sa(dp8762
S'a'
I11353
sS's'
I16
sS'm'
(lp8763
I1
aI1
aI1
aI11
aI13
aI19
aI123
aI185
aI413
aI745
aI361
aI2391
aI6697
aI2281
aI11999
aI13175
asS'd'
I4382
sa(dp8764
S'a'
I11369
sS's'
I16
sS'm'
(lp8765
I1
aI3
aI5
aI5
aI11
aI49
aI123
aI173
aI325
aI667
aI895
aI1067
aI8121
aI10577
aI30561
aI17391
asS'd'
I4383
sa(dp8766
S'a'
I11372
sS's'
I16
sS'm'
(lp8767
I1
aI1
aI5
aI5
aI23
aI21
aI77
aI223
aI281
aI161
aI141
aI345
aI3879
aI11393
aI26907
aI53805
asS'd'
I4384
sa(dp8768
S'a'
I11375
sS's'
I16
sS'm'
(lp8769
I1
aI3
aI3
aI5
aI3
aI47
aI17
aI109
aI185
aI139
aI957
aI1417
aI4553
aI6101
aI29537
aI34821
asS'd'
I4385
sa(dp8770
S'a'
I11413
sS's'
I16
sS'm'
(lp8771
I1
aI1
aI5
aI3
aI29
aI49
aI89
aI61
aI45
aI593
aI269
aI1483
aI2971
aI991
aI21239
aI29301
asS'd'
I4386
sa(dp8772
S'a'
I11429
sS's'
I16
sS'm'
(lp8773
I1
aI1
aI3
aI13
aI29
aI45
aI33
aI253
aI291
aI783
aI737
aI2363
aI2651
aI8627
aI21785
aI58419
asS'd'
I4387
sa(dp8774
S'a'
I11430
sS's'
I16
sS'm'
(lp8775
I1
aI3
aI7
aI15
aI29
aI15
aI81
aI185
aI363
aI681
aI1737
aI3789
aI4365
aI3295
aI23205
aI4457
asS'd'
I4388
sa(dp8776
S'a'
I11444
sS's'
I16
sS'm'
(lp8777
I1
aI3
aI3
aI11
aI15
aI39
aI67
aI167
aI197
aI357
aI871
aI2201
aI5377
aI6299
aI20873
aI59283
asS'd'
I4389
sa(dp8778
S'a'
I11462
sS's'
I16
sS'm'
(lp8779
I1
aI3
aI3
aI5
aI9
aI15
aI127
aI49
aI21
aI719
aI357
aI425
aI5507
aI9639
aI275
aI47503
asS'd'
I4390
sa(dp8780
S'a'
I11465
sS's'
I16
sS'm'
(lp8781
I1
aI1
aI7
aI1
aI13
aI63
aI111
aI121
aI21
aI481
aI247
aI3147
aI5783
aI8947
aI20809
aI42039
asS'd'
I4391
sa(dp8782
S'a'
I11471
sS's'
I16
sS'm'
(lp8783
I1
aI1
aI3
aI3
aI31
aI33
aI9
aI69
aI187
aI517
aI2029
aI2205
aI7661
aI4757
aI27525
aI9665
asS'd'
I4392
sa(dp8784
S'a'
I11473
sS's'
I16
sS'm'
(lp8785
I1
aI1
aI1
aI13
aI7
aI41
aI103
aI161
aI233
aI221
aI693
aI213
aI4609
aI7771
aI28703
aI17407
asS'd'
I4393
sa(dp8786
S'a'
I11495
sS's'
I16
sS'm'
(lp8787
I1
aI3
aI7
aI15
aI31
aI47
aI27
aI111
aI479
aI1003
aI1687
aI347
aI2179
aI11861
aI8169
aI31941
asS'd'
I4394
sa(dp8788
S'a'
I11499
sS's'
I16
sS'm'
(lp8789
I1
aI1
aI3
aI5
aI5
aI63
aI25
aI125
aI199
aI147
aI589
aI3565
aI3449
aI8331
aI8923
aI31207
asS'd'
I4395
sa(dp8790
S'a'
I11519
sS's'
I16
sS'm'
(lp8791
I1
aI1
aI3
aI11
aI19
aI25
aI77
aI99
aI299
aI19
aI1641
aI2193
aI4299
aI3635
aI20621
aI267
asS'd'
I4396
sa(dp8792
S'a'
I11562
sS's'
I16
sS'm'
(lp8793
I1
aI3
aI7
aI11
aI9
aI59
aI7
aI167
aI1
aI775
aI29
aI1935
aI3723
aI11835
aI2887
aI65285
asS'd'
I4397
sa(dp8794
S'a'
I11576
sS's'
I16
sS'm'
(lp8795
I1
aI1
aI7
aI13
aI5
aI47
aI101
aI155
aI235
aI93
aI465
aI3581
aI1837
aI7675
aI10789
aI45167
asS'd'
I4398
sa(dp8796
S'a'
I11582
sS's'
I16
sS'm'
(lp8797
I1
aI1
aI3
aI5
aI9
aI59
aI121
aI109
aI59
aI821
aI1625
aI343
aI1591
aI3875
aI13729
aI56381
asS'd'
I4399
sa(dp8798
S'a'
I11596
sS's'
I16
sS'm'
(lp8799
I1
aI1
aI1
aI9
aI27
aI53
aI93
aI215
aI133
aI561
aI39
aI2591
aI1041
aI11913
aI24493
aI37921
asS'd'
I4400
sa(dp8800
S'a'
I11602
sS's'
I16
sS'm'
(lp8801
I1
aI1
aI1
aI7
aI23
aI7
aI81
aI107
aI219
aI943
aI563
aI1083
aI5803
aI5687
aI32559
aI62727
asS'd'
I4401
sa(dp8802
S'a'
I11611
sS's'
I16
sS'm'
(lp8803
I1
aI3
aI7
aI7
aI21
aI53
aI3
aI5
aI231
aI601
aI1561
aI103
aI3837
aI2675
aI5263
aI23375
asS'd'
I4402
sa(dp8804
S'a'
I11627
sS's'
I16
sS'm'
(lp8805
I1
aI1
aI3
aI13
aI15
aI27
aI89
aI7
aI251
aI887
aI951
aI3001
aI5687
aI4921
aI5091
aI59337
asS'd'
I4403
sa(dp8806
S'a'
I11682
sS's'
I16
sS'm'
(lp8807
I1
aI3
aI7
aI15
aI25
aI53
aI19
aI155
aI185
aI503
aI547
aI1917
aI7633
aI15167
aI14177
aI46761
asS'd'
I4404
sa(dp8808
S'a'
I11687
sS's'
I16
sS'm'
(lp8809
I1
aI3
aI5
aI15
aI21
aI49
aI13
aI163
aI471
aI281
aI1141
aI3013
aI6847
aI9261
aI15955
aI9397
asS'd'
I4405
sa(dp8810
S'a'
I11691
sS's'
I16
sS'm'
(lp8811
I1
aI3
aI3
aI3
aI1
aI21
aI19
aI239
aI479
aI609
aI65
aI2735
aI5337
aI6293
aI19351
aI34135
asS'd'
I4406
sa(dp8812
S'a'
I11733
sS's'
I16
sS'm'
(lp8813
I1
aI1
aI7
aI1
aI9
aI1
aI119
aI23
aI411
aI535
aI101
aI1597
aI2379
aI2421
aI31471
aI36473
asS'd'
I4407
sa(dp8814
S'a'
I11747
sS's'
I16
sS'm'
(lp8815
I1
aI3
aI1
aI11
aI31
aI63
aI17
aI225
aI45
aI409
aI59
aI3943
aI8043
aI11759
aI10667
aI58793
asS'd'
I4408
sa(dp8816
S'a'
I11759
sS's'
I16
sS'm'
(lp8817
I1
aI1
aI3
aI3
aI9
aI49
aI61
aI239
aI245
aI765
aI1945
aI3567
aI5355
aI14799
aI7141
aI59511
asS'd'
I4409
sa(dp8818
S'a'
I11778
sS's'
I16
sS'm'
(lp8819
I1
aI3
aI7
aI9
aI3
aI33
aI103
aI99
aI35
aI799
aI1347
aI2253
aI8189
aI14177
aI13479
aI13749
asS'd'
I4410
sa(dp8820
S'a'
I11852
sS's'
I16
sS'm'
(lp8821
I1
aI3
aI1
aI15
aI15
aI51
aI7
aI179
aI471
aI265
aI1571
aI2983
aI701
aI15133
aI7885
aI29977
asS'd'
I4411
sa(dp8822
S'a'
I11857
sS's'
I16
sS'm'
(lp8823
I1
aI1
aI5
aI9
aI15
aI37
aI49
aI213
aI113
aI729
aI1115
aI2727
aI2635
aI8433
aI11145
aI46779
asS'd'
I4412
sa(dp8824
S'a'
I11858
sS's'
I16
sS'm'
(lp8825
I1
aI3
aI5
aI11
aI7
aI3
aI115
aI151
aI133
aI723
aI7
aI4063
aI5807
aI9845
aI25829
aI29315
asS'd'
I4413
sa(dp8826
S'a'
I11893
sS's'
I16
sS'm'
(lp8827
I1
aI3
aI5
aI9
aI25
aI55
aI17
aI135
aI145
aI379
aI1603
aI3459
aI5773
aI6545
aI17509
aI25847
asS'd'
I4414
sa(dp8828
S'a'
I11907
sS's'
I16
sS'm'
(lp8829
I1
aI1
aI7
aI11
aI1
aI61
aI113
aI147
aI489
aI775
aI1347
aI2199
aI299
aI9589
aI19931
aI1335
asS'd'
I4415
sa(dp8830
S'a'
I11928
sS's'
I16
sS'm'
(lp8831
I1
aI3
aI1
aI3
aI1
aI7
aI27
aI243
aI355
aI425
aI1215
aI3723
aI3489
aI9285
aI12739
aI24797
asS'd'
I4416
sa(dp8832
S'a'
I11931
sS's'
I16
sS'm'
(lp8833
I1
aI3
aI5
aI11
aI15
aI25
aI57
aI221
aI247
aI647
aI259
aI1665
aI7055
aI2835
aI16411
aI42999
asS'd'
I4417
sa(dp8834
S'a'
I11933
sS's'
I16
sS'm'
(lp8835
I1
aI1
aI3
aI7
aI9
aI25
aI61
aI233
aI73
aI235
aI1539
aI1865
aI5671
aI1329
aI5767
aI43039
asS'd'
I4418
sa(dp8836
S'a'
I11967
sS's'
I16
sS'm'
(lp8837
I1
aI1
aI7
aI9
aI21
aI11
aI123
aI7
aI41
aI407
aI1485
aI1723
aI585
aI10597
aI16215
aI63403
asS'd'
I4419
sa(dp8838
S'a'
I11976
sS's'
I16
sS'm'
(lp8839
I1
aI1
aI5
aI7
aI27
aI9
aI123
aI101
aI273
aI673
aI1141
aI3841
aI4041
aI13169
aI8221
aI12915
asS'd'
I4420
sa(dp8840
S'a'
I11989
sS's'
I16
sS'm'
(lp8841
I1
aI3
aI1
aI13
aI3
aI17
aI105
aI17
aI237
aI321
aI855
aI2821
aI2467
aI3503
aI15187
aI1689
asS'd'
I4421
sa(dp8842
S'a'
I12003
sS's'
I16
sS'm'
(lp8843
I1
aI1
aI5
aI5
aI19
aI23
aI9
aI205
aI87
aI153
aI1543
aI1193
aI6619
aI6845
aI8459
aI37533
asS'd'
I4422
sa(dp8844
S'a'
I12024
sS's'
I16
sS'm'
(lp8845
I1
aI1
aI7
aI15
aI13
aI29
aI79
aI9
aI203
aI211
aI239
aI2349
aI3447
aI7797
aI19311
aI58071
asS'd'
I4423
sa(dp8846
S'a'
I12030
sS's'
I16
sS'm'
(lp8847
I1
aI3
aI5
aI11
aI5
aI49
aI71
aI151
aI333
aI895
aI1095
aI2471
aI2477
aI14493
aI16711
aI14393
asS'd'
I4424
sa(dp8848
S'a'
I12037
sS's'
I16
sS'm'
(lp8849
I1
aI1
aI5
aI13
aI17
aI19
aI25
aI149
aI111
aI631
aI763
aI2535
aI3631
aI1809
aI8163
aI18037
asS'd'
I4425
sa(dp8850
S'a'
I12044
sS's'
I16
sS'm'
(lp8851
I1
aI3
aI3
aI13
aI23
aI61
aI25
aI179
aI351
aI247
aI1021
aI2413
aI2585
aI3731
aI5435
aI52723
asS'd'
I4426
sa(dp8852
S'a'
I12052
sS's'
I16
sS'm'
(lp8853
I1
aI1
aI5
aI11
aI1
aI39
aI65
aI59
aI21
aI927
aI1989
aI2823
aI4857
aI15521
aI30495
aI16067
asS'd'
I4427
sa(dp8854
S'a'
I12059
sS's'
I16
sS'm'
(lp8855
I1
aI3
aI3
aI7
aI17
aI5
aI17
aI125
aI379
aI875
aI1565
aI2435
aI933
aI6809
aI20129
aI26339
asS'd'
I4428
sa(dp8856
S'a'
I12075
sS's'
I16
sS'm'
(lp8857
I1
aI1
aI7
aI5
aI3
aI57
aI51
aI213
aI455
aI663
aI2007
aI3995
aI4889
aI9527
aI23507
aI3261
asS'd'
I4429
sa(dp8858
S'a'
I12083
sS's'
I16
sS'm'
(lp8859
I1
aI3
aI7
aI5
aI1
aI29
aI85
aI151
aI165
aI123
aI1425
aI2851
aI4427
aI7683
aI21085
aI28925
asS'd'
I4430
sa(dp8860
S'a'
I12117
sS's'
I16
sS'm'
(lp8861
I1
aI1
aI5
aI3
aI11
aI33
aI127
aI3
aI41
aI591
aI1539
aI3823
aI125
aI421
aI9051
aI55025
asS'd'
I4431
sa(dp8862
S'a'
I12138
sS's'
I16
sS'm'
(lp8863
I1
aI3
aI1
aI5
aI7
aI9
aI69
aI35
aI59
aI477
aI1207
aI1245
aI6423
aI11329
aI26535
aI37621
asS'd'
I4432
sa(dp8864
S'a'
I12146
sS's'
I16
sS'm'
(lp8865
I1
aI3
aI1
aI1
aI15
aI35
aI17
aI123
aI303
aI193
aI1489
aI2587
aI1883
aI4063
aI1921
aI60413
asS'd'
I4433
sa(dp8866
S'a'
I12202
sS's'
I16
sS'm'
(lp8867
I1
aI1
aI5
aI1
aI7
aI61
aI39
aI247
aI139
aI1015
aI757
aI823
aI4757
aI9307
aI32287
aI37241
asS'd'
I4434
sa(dp8868
S'a'
I12230
sS's'
I16
sS'm'
(lp8869
I1
aI1
aI7
aI15
aI3
aI5
aI85
aI93
aI457
aI999
aI1331
aI919
aI5271
aI11687
aI24233
aI38803
asS'd'
I4435
sa(dp8870
S'a'
I12254
sS's'
I16
sS'm'
(lp8871
I1
aI3
aI3
aI9
aI5
aI43
aI37
aI13
aI505
aI603
aI1857
aI2675
aI2017
aI9481
aI10873
aI54755
asS'd'
I4436
sa(dp8872
S'a'
I12304
sS's'
I16
sS'm'
(lp8873
I1
aI1
aI5
aI15
aI13
aI3
aI7
aI239
aI471
aI835
aI553
aI413
aI4029
aI8613
aI15533
aI58987
asS'd'
I4437
sa(dp8874
S'a'
I12316
sS's'
I16
sS'm'
(lp8875
I1
aI3
aI1
aI5
aI19
aI27
aI21
aI43
aI57
aI755
aI1245
aI2805
aI3799
aI2013
aI21145
aI10317
asS'd'
I4438
sa(dp8876
S'a'
I12337
sS's'
I16
sS'm'
(lp8877
I1
aI3
aI5
aI13
aI9
aI23
aI35
aI5
aI315
aI169
aI399
aI2641
aI1525
aI10561
aI11917
aI33009
asS'd'
I4439
sa(dp8878
S'a'
I12347
sS's'
I16
sS'm'
(lp8879
I1
aI3
aI5
aI7
aI23
aI53
aI101
aI105
aI451
aI207
aI1271
aI3067
aI6725
aI15525
aI7951
aI1283
asS'd'
I4440
sa(dp8880
S'a'
I12367
sS's'
I16
sS'm'
(lp8881
I1
aI1
aI5
aI5
aI27
aI21
aI77
aI143
aI213
aI443
aI149
aI2667
aI5269
aI10359
aI25287
aI5843
asS'd'
I4441
sa(dp8882
S'a'
I12398
sS's'
I16
sS'm'
(lp8883
I1
aI3
aI5
aI5
aI25
aI27
aI109
aI157
aI459
aI767
aI765
aI2667
aI1833
aI3781
aI9077
aI64321
asS'd'
I4442
sa(dp8884
S'a'
I12421
sS's'
I16
sS'm'
(lp8885
I1
aI3
aI3
aI13
aI31
aI25
aI97
aI237
aI279
aI431
aI1713
aI809
aI1989
aI10431
aI5867
aI42197
asS'd'
I4443
sa(dp8886
S'a'
I12428
sS's'
I16
sS'm'
(lp8887
I1
aI3
aI7
aI3
aI9
aI5
aI5
aI191
aI73
aI521
aI787
aI903
aI3073
aI2067
aI24741
aI57029
asS'd'
I4444
sa(dp8888
S'a'
I12446
sS's'
I16
sS'm'
(lp8889
I1
aI3
aI3
aI1
aI3
aI41
aI125
aI53
aI125
aI781
aI865
aI3677
aI6279
aI357
aI10667
aI1127
asS'd'
I4445
sa(dp8890
S'a'
I12449
sS's'
I16
sS'm'
(lp8891
I1
aI1
aI5
aI11
aI25
aI19
aI99
aI121
aI359
aI73
aI607
aI2149
aI5739
aI15669
aI29457
aI57549
asS'd'
I4446
sa(dp8892
S'a'
I12467
sS's'
I16
sS'm'
(lp8893
I1
aI1
aI5
aI3
aI23
aI5
aI35
aI55
aI369
aI239
aI329
aI3057
aI3757
aI12523
aI5017
aI52185
asS'd'
I4447
sa(dp8894
S'a'
I12479
sS's'
I16
sS'm'
(lp8895
I1
aI3
aI3
aI13
aI17
aI61
aI69
aI165
aI267
aI323
aI2007
aI2025
aI4423
aI15975
aI31897
aI37013
asS'd'
I4448
sa(dp8896
S'a'
I12502
sS's'
I16
sS'm'
(lp8897
I1
aI3
aI7
aI13
aI19
aI19
aI87
aI111
aI389
aI611
aI1523
aI963
aI4671
aI12569
aI21839
aI10919
asS'd'
I4449
sa(dp8898
S'a'
I12521
sS's'
I16
sS'm'
(lp8899
I1
aI1
aI1
aI3
aI1
aI27
aI13
aI227
aI29
aI457
aI221
aI127
aI5335
aI16247
aI19559
aI19185
asS'd'
I4450
sa(dp8900
S'a'
I12553
sS's'
I16
sS'm'
(lp8901
I1
aI3
aI5
aI7
aI29
aI21
aI23
aI157
aI197
aI87
aI1591
aI1829
aI407
aI15885
aI14933
aI1997
asS'd'
I4451
sa(dp8902
S'a'
I12568
sS's'
I16
sS'm'
(lp8903
I1
aI1
aI1
aI9
aI3
aI35
aI43
aI187
aI195
aI325
aI197
aI2905
aI7323
aI1563
aI7659
aI45185
asS'd'
I4452
sa(dp8904
S'a'
I12573
sS's'
I16
sS'm'
(lp8905
I1
aI1
aI1
aI15
aI3
aI23
aI105
aI33
aI73
aI495
aI1409
aI2583
aI1099
aI1041
aI28955
aI60293
asS'd'
I4453
sa(dp8906
S'a'
I12592
sS's'
I16
sS'm'
(lp8907
I1
aI3
aI7
aI13
aI25
aI19
aI99
aI137
aI139
aI719
aI529
aI1147
aI5813
aI11551
aI30183
aI14593
asS'd'
I4454
sa(dp8908
S'a'
I12597
sS's'
I16
sS'm'
(lp8909
I1
aI3
aI3
aI5
aI17
aI25
aI73
aI193
aI309
aI887
aI1655
aI1641
aI2091
aI12087
aI21881
aI1145
asS'd'
I4455
sa(dp8910
S'a'
I12601
sS's'
I16
sS'm'
(lp8911
I1
aI3
aI1
aI1
aI27
aI41
aI13
aI151
aI83
aI645
aI327
aI1795
aI2717
aI12433
aI22751
aI9823
asS'd'
I4456
sa(dp8912
S'a'
I12615
sS's'
I16
sS'm'
(lp8913
I1
aI1
aI3
aI7
aI1
aI43
aI77
aI229
aI59
aI499
aI1883
aI135
aI3461
aI9821
aI219
aI6111
asS'd'
I4457
sa(dp8914
S'a'
I12619
sS's'
I16
sS'm'
(lp8915
I1
aI3
aI3
aI3
aI23
aI7
aI17
aI67
aI361
aI565
aI1621
aI3253
aI7973
aI281
aI3209
aI48215
asS'd'
I4458
sa(dp8916
S'a'
I12694
sS's'
I16
sS'm'
(lp8917
I1
aI1
aI3
aI7
aI31
aI15
aI97
aI141
aI197
aI883
aI1689
aI269
aI7487
aI10403
aI18903
aI58147
asS'd'
I4459
sa(dp8918
S'a'
I12697
sS's'
I16
sS'm'
(lp8919
I1
aI3
aI3
aI3
aI23
aI9
aI87
aI125
aI359
aI527
aI1251
aI637
aI1145
aI12721
aI14693
aI6021
asS'd'
I4460
sa(dp8920
S'a'
I12698
sS's'
I16
sS'm'
(lp8921
I1
aI1
aI3
aI5
aI13
aI43
aI105
aI173
aI205
aI859
aI1237
aI1227
aI1409
aI15513
aI25317
aI30745
asS'd'
I4461
sa(dp8922
S'a'
I12713
sS's'
I16
sS'm'
(lp8923
I1
aI3
aI3
aI15
aI31
aI43
aI125
aI149
aI145
aI109
aI975
aI1167
aI7629
aI8373
aI5923
aI64117
asS'd'
I4462
sa(dp8924
S'a'
I12722
sS's'
I16
sS'm'
(lp8925
I1
aI3
aI1
aI15
aI3
aI33
aI89
aI173
aI379
aI615
aI1401
aI1567
aI4453
aI7461
aI32555
aI64369
asS'd'
I4463
sa(dp8926
S'a'
I12733
sS's'
I16
sS'm'
(lp8927
I1
aI3
aI7
aI11
aI27
aI23
aI45
aI7
aI15
aI21
aI1663
aI3327
aI5611
aI8535
aI27669
aI25525
asS'd'
I4464
sa(dp8928
S'a'
I12736
sS's'
I16
sS'm'
(lp8929
I1
aI1
aI3
aI15
aI15
aI61
aI127
aI145
aI151
aI41
aI629
aI767
aI5801
aI3395
aI1157
aI30033
asS'd'
I4465
sa(dp8930
S'a'
I12790
sS's'
I16
sS'm'
(lp8931
I1
aI1
aI1
aI5
aI9
aI63
aI73
aI9
aI299
aI237
aI369
aI1295
aI4869
aI6821
aI19961
aI32129
asS'd'
I4466
sa(dp8932
S'a'
I12794
sS's'
I16
sS'm'
(lp8933
I1
aI1
aI5
aI13
aI19
aI7
aI119
aI35
aI319
aI405
aI1255
aI1299
aI4311
aI14999
aI24567
aI4001
asS'd'
I4467
sa(dp8934
S'a'
I12803
sS's'
I16
sS'm'
(lp8935
I1
aI1
aI1
aI13
aI9
aI39
aI13
aI207
aI355
aI691
aI37
aI3137
aI6073
aI16179
aI28661
aI41
asS'd'
I4468
sa(dp8936
S'a'
I12815
sS's'
I16
sS'm'
(lp8937
I1
aI1
aI3
aI7
aI21
aI3
aI123
aI27
aI187
aI183
aI769
aI2367
aI2957
aI7065
aI17855
aI60805
asS'd'
I4469
sa(dp8938
S'a'
I12829
sS's'
I16
sS'm'
(lp8939
I1
aI1
aI1
aI1
aI19
aI31
aI71
aI85
aI303
aI617
aI1007
aI283
aI8087
aI11079
aI11463
aI65295
asS'd'
I4470
sa(dp8940
S'a'
I12833
sS's'
I16
sS'm'
(lp8941
I1
aI1
aI3
aI13
aI25
aI63
aI61
aI187
aI401
aI465
aI1485
aI801
aI3649
aI7763
aI8495
aI54479
asS'd'
I4471
sa(dp8942
S'a'
I12840
sS's'
I16
sS'm'
(lp8943
I1
aI3
aI7
aI3
aI13
aI51
aI41
aI119
aI311
aI699
aI1113
aI3631
aI1981
aI3259
aI25871
aI45659
asS'd'
I4472
sa(dp8944
S'a'
I12875
sS's'
I16
sS'm'
(lp8945
I1
aI3
aI7
aI13
aI31
aI27
aI57
aI181
aI325
aI107
aI1745
aI635
aI3941
aI3305
aI14563
aI29855
asS'd'
I4473
sa(dp8946
S'a'
I12877
sS's'
I16
sS'm'
(lp8947
I1
aI3
aI7
aI15
aI5
aI55
aI5
aI147
aI365
aI485
aI1841
aI3673
aI6513
aI11121
aI5725
aI18027
asS'd'
I4474
sa(dp8948
S'a'
I12916
sS's'
I16
sS'm'
(lp8949
I1
aI3
aI5
aI11
aI13
aI45
aI35
aI79
aI109
aI683
aI1171
aI3015
aI2163
aI4823
aI4365
aI42931
asS'd'
I4475
sa(dp8950
S'a'
I12923
sS's'
I16
sS'm'
(lp8951
I1
aI1
aI7
aI7
aI13
aI45
aI57
aI39
aI297
aI985
aI1559
aI487
aI5071
aI2657
aI9401
aI41889
asS'd'
I4476
sa(dp8952
S'a'
I12935
sS's'
I16
sS'm'
(lp8953
I1
aI1
aI5
aI9
aI29
aI33
aI79
aI237
aI509
aI537
aI549
aI3657
aI7141
aI15189
aI30853
aI36089
asS'd'
I4477
sa(dp8954
S'a'
I12949
sS's'
I16
sS'm'
(lp8955
I1
aI3
aI5
aI7
aI31
aI15
aI75
aI73
aI237
aI273
aI865
aI743
aI2607
aI7611
aI18441
aI12703
asS'd'
I4478
sa(dp8956
S'a'
I12954
sS's'
I16
sS'm'
(lp8957
I1
aI1
aI1
aI9
aI27
aI9
aI35
aI137
aI265
aI181
aI1341
aI1945
aI5615
aI161
aI18369
aI4791
asS'd'
I4479
sa(dp8958
S'a'
I12956
sS's'
I16
sS'm'
(lp8959
I1
aI3
aI7
aI11
aI27
aI29
aI29
aI43
aI489
aI177
aI1489
aI2927
aI623
aI14571
aI22447
aI46905
asS'd'
I4480
sa(dp8960
S'a'
I12959
sS's'
I16
sS'm'
(lp8961
I1
aI3
aI3
aI3
aI19
aI41
aI107
aI53
aI239
aI263
aI1433
aI1655
aI7991
aI7405
aI2735
aI25519
asS'd'
I4481
sa(dp8962
S'a'
I12978
sS's'
I16
sS'm'
(lp8963
I1
aI3
aI5
aI7
aI19
aI37
aI73
aI243
aI215
aI445
aI1781
aI3223
aI187
aI8443
aI18185
aI45093
asS'd'
I4482
sa(dp8964
S'a'
I13001
sS's'
I16
sS'm'
(lp8965
I1
aI1
aI3
aI13
aI9
aI57
aI111
aI111
aI221
aI505
aI1261
aI3045
aI1655
aI16247
aI21033
aI17993
asS'd'
I4483
sa(dp8966
S'a'
I13010
sS's'
I16
sS'm'
(lp8967
I1
aI1
aI7
aI5
aI7
aI55
aI77
aI5
aI131
aI969
aI1481
aI2883
aI2645
aI3003
aI5601
aI37063
asS'd'
I4484
sa(dp8968
S'a'
I13043
sS's'
I16
sS'm'
(lp8969
I1
aI1
aI5
aI15
aI29
aI55
aI39
aI197
aI349
aI29
aI341
aI67
aI1977
aI425
aI4063
aI42705
asS'd'
I4485
sa(dp8970
S'a'
I13049
sS's'
I16
sS'm'
(lp8971
I1
aI1
aI7
aI13
aI1
aI57
aI81
aI81
aI129
aI681
aI1407
aI2465
aI3627
aI2325
aI31649
aI18449
asS'd'
I4486
sa(dp8972
S'a'
I13058
sS's'
I16
sS'm'
(lp8973
I1
aI3
aI5
aI5
aI23
aI59
aI35
aI217
aI393
aI155
aI1315
aI105
aI2285
aI5167
aI27997
aI55193
asS'd'
I4487
sa(dp8974
S'a'
I13112
sS's'
I16
sS'm'
(lp8975
I1
aI1
aI7
aI3
aI11
aI59
aI53
aI179
aI319
aI819
aI947
aI3881
aI765
aI4219
aI16405
aI48055
asS'd'
I4488
sa(dp8976
S'a'
I13140
sS's'
I16
sS'm'
(lp8977
I1
aI1
aI3
aI9
aI23
aI9
aI67
aI67
aI137
aI523
aI1585
aI2441
aI167
aI5217
aI12031
aI14297
asS'd'
I4489
sa(dp8978
S'a'
I13149
sS's'
I16
sS'm'
(lp8979
I1
aI1
aI5
aI13
aI31
aI63
aI121
aI91
aI439
aI917
aI203
aI1519
aI4363
aI2391
aI25755
aI26677
asS'd'
I4490
sa(dp8980
S'a'
I13163
sS's'
I16
sS'm'
(lp8981
I1
aI1
aI3
aI5
aI25
aI31
aI11
aI95
aI339
aI817
aI35
aI3923
aI7365
aI10537
aI5521
aI54579
asS'd'
I4491
sa(dp8982
S'a'
I13187
sS's'
I16
sS'm'
(lp8983
I1
aI3
aI7
aI1
aI3
aI33
aI47
aI13
aI139
aI445
aI1357
aI3907
aI7495
aI8789
aI26589
aI43479
asS'd'
I4492
sa(dp8984
S'a'
I13196
sS's'
I16
sS'm'
(lp8985
I1
aI1
aI1
aI11
aI5
aI45
aI61
aI13
aI167
aI287
aI931
aI881
aI5713
aI12937
aI12951
aI21597
asS'd'
I4493
sa(dp8986
S'a'
I13237
sS's'
I16
sS'm'
(lp8987
I1
aI3
aI5
aI1
aI29
aI23
aI7
aI117
aI503
aI897
aI733
aI1113
aI7205
aI11507
aI561
aI34011
asS'd'
I4494
sa(dp8988
S'a'
I13244
sS's'
I16
sS'm'
(lp8989
I1
aI3
aI5
aI7
aI3
aI51
aI21
aI147
aI35
aI259
aI689
aI3801
aI2481
aI9673
aI4065
aI595
asS'd'
I4495
sa(dp8990
S'a'
I13264
sS's'
I16
sS'm'
(lp8991
I1
aI3
aI3
aI9
aI9
aI29
aI5
aI191
aI393
aI979
aI1627
aI937
aI75
aI2339
aI24007
aI30555
asS'd'
I4496
sa(dp8992
S'a'
I13279
sS's'
I16
sS'm'
(lp8993
I1
aI1
aI5
aI7
aI9
aI35
aI111
aI23
aI113
aI563
aI1689
aI1575
aI6127
aI9919
aI2555
aI52261
asS'd'
I4497
sa(dp8994
S'a'
I13292
sS's'
I16
sS'm'
(lp8995
I1
aI3
aI1
aI5
aI31
aI21
aI117
aI159
aI473
aI279
aI1281
aI311
aI159
aI3343
aI27761
aI59989
asS'd'
I4498
sa(dp8996
S'a'
I13295
sS's'
I16
sS'm'
(lp8997
I1
aI1
aI5
aI1
aI23
aI31
aI67
aI241
aI401
aI69
aI933
aI777
aI267
aI12411
aI23767
aI9047
asS'd'
I4499
sa(dp8998
S'a'
I13307
sS's'
I16
sS'm'
(lp8999
I1
aI1
aI5
aI1
aI15
aI49
aI99
aI15
aI267
aI913
aI1581
aI3713
aI651
aI14275
aI10103
aI57619
asS'd'
I4500
sa(dp9000
S'a'
I13312
sS's'
I16
sS'm'
(lp9001
I1
aI3
aI5
aI9
aI19
aI23
aI95
aI5
aI449
aI153
aI1195
aI1315
aI2347
aI12683
aI10865
aI50703
asS'd'
I4501
sa(dp9002
S'a'
I13317
sS's'
I16
sS'm'
(lp9003
I1
aI1
aI1
aI13
aI17
aI17
aI115
aI31
aI135
aI725
aI795
aI1695
aI6199
aI4179
aI5223
aI48457
asS'd'
I4502
sa(dp9004
S'a'
I13327
sS's'
I16
sS'm'
(lp9005
I1
aI3
aI5
aI15
aI31
aI15
aI3
aI247
aI385
aI269
aI1665
aI581
aI2809
aI6333
aI7459
aI14815
asS'd'
I4503
sa(dp9006
S'a'
I13348
sS's'
I16
sS'm'
(lp9007
I1
aI3
aI7
aI5
aI15
aI35
aI117
aI85
aI11
aI621
aI1523
aI981
aI511
aI14113
aI4673
aI22683
asS'd'
I4504
sa(dp9008
S'a'
I13390
sS's'
I16
sS'm'
(lp9009
I1
aI1
aI7
aI1
aI27
aI61
aI119
aI249
aI431
aI147
aI173
aI423
aI1353
aI4747
aI12761
aI32947
asS'd'
I4505
sa(dp9010
S'a'
I13413
sS's'
I16
sS'm'
(lp9011
I1
aI3
aI7
aI1
aI23
aI39
aI15
aI153
aI219
aI359
aI1233
aI169
aI2817
aI11043
aI12435
aI30135
asS'd'
I4506
sa(dp9012
S'a'
I13417
sS's'
I16
sS'm'
(lp9013
I1
aI1
aI5
aI1
aI1
aI55
aI39
aI1
aI151
aI865
aI125
aI2351
aI6315
aI1165
aI20163
aI43991
asS'd'
I4507
sa(dp9014
S'a'
I13418
sS's'
I16
sS'm'
(lp9015
I1
aI1
aI3
aI9
aI25
aI41
aI115
aI221
aI129
aI265
aI1887
aI4057
aI7523
aI13591
aI5735
aI59645
asS'd'
I4508
sa(dp9016
S'a'
I13451
sS's'
I16
sS'm'
(lp9017
I1
aI1
aI5
aI5
aI19
aI15
aI9
aI77
aI511
aI627
aI153
aI2015
aI247
aI15949
aI9715
aI45411
asS'd'
I4509
sa(dp9018
S'a'
I13475
sS's'
I16
sS'm'
(lp9019
I1
aI1
aI7
aI7
aI17
aI17
aI107
aI183
aI39
aI647
aI1339
aI3915
aI4937
aI537
aI27011
aI58937
asS'd'
I4510
sa(dp9020
S'a'
I13482
sS's'
I16
sS'm'
(lp9021
I1
aI3
aI3
aI13
aI3
aI3
aI69
aI201
aI431
aI65
aI683
aI121
aI7017
aI2791
aI16713
aI62555
asS'd'
I4511
sa(dp9022
S'a'
I13510
sS's'
I16
sS'm'
(lp9023
I1
aI3
aI7
aI3
aI7
aI41
aI117
aI237
aI23
aI757
aI545
aI3899
aI1837
aI5555
aI7891
aI29151
asS'd'
I4512
sa(dp9024
S'a'
I13527
sS's'
I16
sS'm'
(lp9025
I1
aI1
aI1
aI3
aI27
aI15
aI39
aI195
aI353
aI299
aI1431
aI2279
aI1795
aI13773
aI24915
aI49701
asS'd'
I4513
sa(dp9026
S'a'
I13543
sS's'
I16
sS'm'
(lp9027
I1
aI1
aI5
aI5
aI7
aI7
aI125
aI5
aI401
aI523
aI1967
aI2471
aI7279
aI7559
aI12025
aI60599
asS'd'
I4514
sa(dp9028
S'a'
I13547
sS's'
I16
sS'm'
(lp9029
I1
aI1
aI1
aI13
aI13
aI59
aI13
aI249
aI465
aI847
aI1483
aI975
aI7729
aI2773
aI15745
aI51237
asS'd'
I4515
sa(dp9030
S'a'
I13627
sS's'
I16
sS'm'
(lp9031
I1
aI1
aI7
aI9
aI31
aI41
aI115
aI141
aI247
aI355
aI1109
aI3239
aI6495
aI4515
aI30145
aI47705
asS'd'
I4516
sa(dp9032
S'a'
I13650
sS's'
I16
sS'm'
(lp9033
I1
aI1
aI3
aI13
aI29
aI41
aI69
aI179
aI45
aI271
aI1909
aI3095
aI7199
aI14049
aI9903
aI33383
asS'd'
I4517
sa(dp9034
S'a'
I13683
sS's'
I16
sS'm'
(lp9035
I1
aI1
aI3
aI13
aI7
aI45
aI105
aI105
aI243
aI121
aI67
aI3169
aI237
aI137
aI4175
aI54325
asS'd'
I4518
sa(dp9036
S'a'
I13696
sS's'
I16
sS'm'
(lp9037
I1
aI3
aI3
aI11
aI19
aI51
aI93
aI155
aI79
aI579
aI1621
aI1251
aI7639
aI15875
aI25815
aI56063
asS'd'
I4519
sa(dp9038
S'a'
I13702
sS's'
I16
sS'm'
(lp9039
I1
aI3
aI3
aI9
aI27
aI27
aI77
aI45
aI217
aI965
aI1045
aI875
aI4515
aI11261
aI27859
aI757
asS'd'
I4520
sa(dp9040
S'a'
I13713
sS's'
I16
sS'm'
(lp9041
I1
aI1
aI3
aI11
aI17
aI7
aI81
aI37
aI299
aI765
aI977
aI3371
aI3163
aI13267
aI18345
aI9239
asS'd'
I4521
sa(dp9042
S'a'
I13739
sS's'
I16
sS'm'
(lp9043
I1
aI1
aI1
aI3
aI15
aI23
aI115
aI11
aI183
aI511
aI557
aI3253
aI153
aI8465
aI22659
aI42143
asS'd'
I4522
sa(dp9044
S'a'
I13749
sS's'
I16
sS'm'
(lp9045
I1
aI1
aI5
aI13
aI17
aI61
aI127
aI219
aI225
aI981
aI615
aI731
aI4069
aI12637
aI11601
aI38257
asS'd'
I4523
sa(dp9046
S'a'
I13767
sS's'
I16
sS'm'
(lp9047
I1
aI1
aI5
aI3
aI29
aI3
aI73
aI79
aI393
aI779
aI823
aI2473
aI3811
aI4417
aI9399
aI50011
asS'd'
I4524
sa(dp9048
S'a'
I13774
sS's'
I16
sS'm'
(lp9049
I1
aI3
aI3
aI9
aI21
aI35
aI61
aI99
aI115
aI657
aI629
aI1129
aI2355
aI12665
aI459
aI40831
asS'd'
I4525
sa(dp9050
S'a'
I13781
sS's'
I16
sS'm'
(lp9051
I1
aI3
aI1
aI7
aI25
aI61
aI53
aI249
aI15
aI649
aI665
aI595
aI1441
aI8035
aI5381
aI7147
asS'd'
I4526
sa(dp9052
S'a'
I13795
sS's'
I16
sS'm'
(lp9053
I1
aI3
aI1
aI7
aI19
aI9
aI27
aI207
aI205
aI461
aI1069
aI4039
aI6549
aI2333
aI29
aI50067
asS'd'
I4527
sa(dp9054
S'a'
I13821
sS's'
I16
sS'm'
(lp9055
I1
aI1
aI5
aI3
aI15
aI7
aI115
aI205
aI71
aI73
aI53
aI71
aI6049
aI15293
aI17041
aI20313
asS'd'
I4528
sa(dp9056
S'a'
I13825
sS's'
I16
sS'm'
(lp9057
I1
aI1
aI7
aI7
aI9
aI7
aI119
aI99
aI357
aI729
aI2041
aI3355
aI5333
aI1263
aI30521
aI64867
asS'd'
I4529
sa(dp9058
S'a'
I13838
sS's'
I16
sS'm'
(lp9059
I1
aI1
aI1
aI7
aI31
aI63
aI63
aI159
aI281
aI295
aI913
aI2161
aI8033
aI13789
aI17711
aI14915
asS'd'
I4530
sa(dp9060
S'a'
I13852
sS's'
I16
sS'm'
(lp9061
I1
aI1
aI7
aI9
aI29
aI55
aI69
aI129
aI453
aI361
aI1883
aI17
aI1765
aI111
aI10311
aI55019
asS'd'
I4531
sa(dp9062
S'a'
I13879
sS's'
I16
sS'm'
(lp9063
I1
aI1
aI5
aI9
aI21
aI15
aI31
aI57
aI291
aI915
aI945
aI1775
aI5905
aI9073
aI3271
aI15571
asS'd'
I4532
sa(dp9064
S'a'
I13888
sS's'
I16
sS'm'
(lp9065
I1
aI1
aI1
aI7
aI21
aI11
aI1
aI9
aI167
aI143
aI1535
aI1267
aI6675
aI13037
aI19513
aI52027
asS'd'
I4533
sa(dp9066
S'a'
I13897
sS's'
I16
sS'm'
(lp9067
I1
aI3
aI3
aI7
aI31
aI45
aI57
aI105
aI63
aI121
aI631
aI679
aI3873
aI16333
aI32069
aI64725
asS'd'
I4534
sa(dp9068
S'a'
I13906
sS's'
I16
sS'm'
(lp9069
I1
aI1
aI1
aI9
aI23
aI41
aI29
aI207
aI489
aI319
aI905
aI3147
aI4195
aI2697
aI5281
aI1771
asS'd'
I4535
sa(dp9070
S'a'
I13939
sS's'
I16
sS'm'
(lp9071
I1
aI1
aI1
aI9
aI25
aI33
aI57
aI201
aI405
aI111
aI407
aI3489
aI449
aI8601
aI1273
aI42105
asS'd'
I4536
sa(dp9072
S'a'
I13962
sS's'
I16
sS'm'
(lp9073
I1
aI1
aI1
aI3
aI19
aI45
aI123
aI159
aI237
aI173
aI781
aI787
aI7537
aI15453
aI25567
aI53729
asS'd'
I4537
sa(dp9074
S'a'
I13981
sS's'
I16
sS'm'
(lp9075
I1
aI1
aI7
aI3
aI29
aI9
aI89
aI207
aI345
aI685
aI1701
aI2859
aI8065
aI9289
aI2459
aI28367
asS'd'
I4538
sa(dp9076
S'a'
I13985
sS's'
I16
sS'm'
(lp9077
I1
aI3
aI1
aI3
aI31
aI41
aI55
aI241
aI81
aI1001
aI595
aI1725
aI853
aI11427
aI20617
aI1717
asS'd'
I4539
sa(dp9078
S'a'
I14020
sS's'
I16
sS'm'
(lp9079
I1
aI1
aI3
aI3
aI9
aI45
aI121
aI69
aI177
aI1017
aI211
aI2753
aI6923
aI1387
aI32063
aI45337
asS'd'
I4540
sa(dp9080
S'a'
I14030
sS's'
I16
sS'm'
(lp9081
I1
aI1
aI5
aI15
aI21
aI23
aI95
aI61
aI485
aI403
aI619
aI3035
aI4881
aI13423
aI17815
aI35221
asS'd'
I4541
sa(dp9082
S'a'
I14041
sS's'
I16
sS'm'
(lp9083
I1
aI1
aI3
aI3
aI3
aI59
aI23
aI69
aI77
aI309
aI227
aI2877
aI3739
aI3539
aI20083
aI23415
asS'd'
I4542
sa(dp9084
S'a'
I14047
sS's'
I16
sS'm'
(lp9085
I1
aI3
aI7
aI3
aI17
aI43
aI95
aI239
aI223
aI353
aI1237
aI3239
aI1369
aI7245
aI32401
aI63889
asS'd'
I4543
sa(dp9086
S'a'
I14048
sS's'
I16
sS'm'
(lp9087
I1
aI1
aI1
aI5
aI25
aI63
aI123
aI3
aI291
aI819
aI793
aI241
aI5619
aI8871
aI18341
aI2757
asS'd'
I4544
sa(dp9088
S'a'
I14066
sS's'
I16
sS'm'
(lp9089
I1
aI3
aI7
aI15
aI3
aI21
aI17
aI97
aI395
aI333
aI909
aI3783
aI3635
aI751
aI24227
aI44281
asS'd'
I4545
sa(dp9090
S'a'
I14092
sS's'
I16
sS'm'
(lp9091
I1
aI3
aI7
aI13
aI29
aI49
aI123
aI195
aI191
aI159
aI211
aI1887
aI3047
aI4871
aI2607
aI32425
asS'd'
I4546
sa(dp9092
S'a'
I14120
sS's'
I16
sS'm'
(lp9093
I1
aI1
aI7
aI7
aI15
aI57
aI91
aI255
aI267
aI897
aI441
aI1581
aI953
aI6951
aI17275
aI50229
asS'd'
I4547
sa(dp9094
S'a'
I14126
sS's'
I16
sS'm'
(lp9095
I1
aI3
aI7
aI15
aI1
aI35
aI91
aI219
aI7
aI117
aI119
aI2687
aI1215
aI9517
aI10849
aI28347
asS'd'
I4548
sa(dp9096
S'a'
I14131
sS's'
I16
sS'm'
(lp9097
I1
aI1
aI1
aI1
aI21
aI55
aI67
aI221
aI503
aI883
aI1037
aI2965
aI1485
aI8557
aI28091
aI25459
asS'd'
I4549
sa(dp9098
S'a'
I14143
sS's'
I16
sS'm'
(lp9099
I1
aI3
aI5
aI9
aI19
aI3
aI73
aI123
aI95
aI307
aI1339
aI3669
aI5077
aI12049
aI523
aI21457
asS'd'
I4550
sa(dp9100
S'a'
I14146
sS's'
I16
sS'm'
(lp9101
I1
aI3
aI1
aI13
aI3
aI1
aI111
aI97
aI371
aI697
aI1989
aI3715
aI7875
aI6841
aI7009
aI17809
asS'd'
I4551
sa(dp9102
S'a'
I14152
sS's'
I16
sS'm'
(lp9103
I1
aI1
aI1
aI9
aI25
aI21
aI19
aI43
aI329
aI531
aI491
aI1147
aI1469
aI12841
aI29651
aI29517
asS'd'
I4552
sa(dp9104
S'a'
I14155
sS's'
I16
sS'm'
(lp9105
I1
aI1
aI5
aI1
aI15
aI63
aI101
aI197
aI477
aI245
aI341
aI61
aI3803
aI10001
aI5519
aI19083
asS'd'
I4553
sa(dp9106
S'a'
I14157
sS's'
I16
sS'm'
(lp9107
I1
aI3
aI7
aI5
aI13
aI43
aI7
aI143
aI291
aI531
aI1727
aI871
aI7091
aI5737
aI24285
aI51017
asS'd'
I4554
sa(dp9108
S'a'
I14188
sS's'
I16
sS'm'
(lp9109
I1
aI3
aI5
aI1
aI17
aI13
aI15
aI85
aI361
aI153
aI989
aI1367
aI4705
aI3599
aI4441
aI52471
asS'd'
I4555
sa(dp9110
S'a'
I14206
sS's'
I16
sS'm'
(lp9111
I1
aI1
aI7
aI13
aI21
aI43
aI111
aI29
aI299
aI439
aI1929
aI283
aI5901
aI14113
aI3929
aI55843
asS'd'
I4556
sa(dp9112
S'a'
I14243
sS's'
I16
sS'm'
(lp9113
I1
aI3
aI3
aI9
aI31
aI59
aI63
aI43
aI91
aI171
aI733
aI1359
aI425
aI8505
aI19777
aI54723
asS'd'
I4557
sa(dp9114
S'a'
I14278
sS's'
I16
sS'm'
(lp9115
I1
aI1
aI5
aI7
aI31
aI1
aI97
aI253
aI331
aI307
aI1749
aI2115
aI2535
aI9945
aI11013
aI14231
asS'd'
I4558
sa(dp9116
S'a'
I14308
sS's'
I16
sS'm'
(lp9117
I1
aI1
aI5
aI11
aI13
aI7
aI109
aI237
aI301
aI383
aI683
aI1603
aI6393
aI2437
aI12101
aI1767
asS'd'
I4559
sa(dp9118
S'a'
I14317
sS's'
I16
sS'm'
(lp9119
I1
aI1
aI3
aI11
aI15
aI61
aI119
aI199
aI109
aI265
aI1431
aI1729
aI3409
aI10129
aI16945
aI34681
asS'd'
I4560
sa(dp9120
S'a'
I14335
sS's'
I16
sS'm'
(lp9121
I1
aI3
aI7
aI9
aI13
aI59
aI121
aI73
aI29
aI513
aI279
aI457
aI7985
aI15199
aI10185
aI33621
asS'd'
I4561
sa(dp9122
S'a'
I14354
sS's'
I16
sS'm'
(lp9123
I1
aI3
aI7
aI7
aI23
aI17
aI27
aI65
aI387
aI487
aI1803
aI2789
aI461
aI11201
aI7001
aI40229
asS'd'
I4562
sa(dp9124
S'a'
I14356
sS's'
I16
sS'm'
(lp9125
I1
aI1
aI3
aI15
aI9
aI13
aI55
aI127
aI33
aI513
aI1055
aI643
aI505
aI3005
aI6121
aI64147
asS'd'
I4563
sa(dp9126
S'a'
I14379
sS's'
I16
sS'm'
(lp9127
I1
aI3
aI5
aI15
aI5
aI11
aI77
aI233
aI175
aI691
aI171
aI2491
aI6915
aI14195
aI7845
aI36499
asS'd'
I4564
sa(dp9128
S'a'
I14381
sS's'
I16
sS'm'
(lp9129
I1
aI1
aI5
aI11
aI19
aI45
aI103
aI207
aI99
aI645
aI1739
aI1517
aI5907
aI16035
aI14559
aI44007
asS'd'
I4565
sa(dp9130
S'a'
I14384
sS's'
I16
sS'm'
(lp9131
I1
aI3
aI7
aI15
aI21
aI27
aI53
aI107
aI89
aI291
aI983
aI3527
aI1025
aI2985
aI13747
aI32715
asS'd'
I4566
sa(dp9132
S'a'
I14389
sS's'
I16
sS'm'
(lp9133
I1
aI1
aI3
aI15
aI23
aI17
aI27
aI209
aI77
aI959
aI813
aI3597
aI5809
aI693
aI10325
aI16855
asS'd'
I4567
sa(dp9134
S'a'
I14390
sS's'
I16
sS'm'
(lp9135
I1
aI1
aI7
aI11
aI23
aI53
aI123
aI99
aI211
aI935
aI1371
aI1657
aI4699
aI2683
aI27933
aI21451
asS'd'
I4568
sa(dp9136
S'a'
I14414
sS's'
I16
sS'm'
(lp9137
I1
aI3
aI3
aI3
aI17
aI21
aI93
aI201
aI423
aI351
aI1639
aI227
aI5719
aI13111
aI5993
aI44615
asS'd'
I4569
sa(dp9138
S'a'
I14425
sS's'
I16
sS'm'
(lp9139
I1
aI1
aI7
aI3
aI13
aI49
aI59
aI255
aI213
aI147
aI1441
aI3593
aI6419
aI15657
aI1947
aI62713
asS'd'
I4570
sa(dp9140
S'a'
I14462
sS's'
I16
sS'm'
(lp9141
I1
aI3
aI1
aI7
aI7
aI41
aI79
aI135
aI275
aI585
aI925
aI3139
aI4351
aI1827
aI23533
aI63031
asS'd'
I4571
sa(dp9142
S'a'
I14472
sS's'
I16
sS'm'
(lp9143
I1
aI1
aI7
aI3
aI11
aI1
aI13
aI149
aI29
aI897
aI1043
aI2393
aI3931
aI6741
aI19973
aI46303
asS'd'
I4572
sa(dp9144
S'a'
I14508
sS's'
I16
sS'm'
(lp9145
I1
aI1
aI1
aI13
aI13
aI57
aI9
aI253
aI149
aI67
aI1531
aI4049
aI3013
aI13947
aI3371
aI35317
asS'd'
I4573
sa(dp9146
S'a'
I14511
sS's'
I16
sS'm'
(lp9147
I1
aI3
aI1
aI1
aI15
aI19
aI11
aI125
aI179
aI383
aI1273
aI1551
aI6441
aI6579
aI19659
aI31005
asS'd'
I4574
sa(dp9148
S'a'
I14519
sS's'
I16
sS'm'
(lp9149
I1
aI1
aI3
aI15
aI29
aI37
aI11
aI199
aI273
aI663
aI549
aI3167
aI2049
aI8815
aI30719
aI47905
asS'd'
I4575
sa(dp9150
S'a'
I14528
sS's'
I16
sS'm'
(lp9151
I1
aI1
aI1
aI15
aI7
aI9
aI113
aI231
aI155
aI27
aI419
aI1303
aI4493
aI5633
aI5743
aI51335
asS'd'
I4576
sa(dp9152
S'a'
I14561
sS's'
I16
sS'm'
(lp9153
I1
aI3
aI5
aI13
aI21
aI35
aI7
aI63
aI391
aI637
aI2011
aI841
aI5933
aI10563
aI7593
aI34767
asS'd'
I4577
sa(dp9154
S'a'
I14574
sS's'
I16
sS'm'
(lp9155
I1
aI3
aI1
aI15
aI19
aI13
aI89
aI127
aI507
aI715
aI1305
aI2989
aI7551
aI1953
aI26101
aI54913
asS'd'
I4578
sa(dp9156
S'a'
I14588
sS's'
I16
sS'm'
(lp9157
I1
aI1
aI5
aI1
aI1
aI33
aI101
aI211
aI173
aI761
aI177
aI2721
aI6949
aI15801
aI6639
aI21405
asS'd'
I4579
sa(dp9158
S'a'
I14594
sS's'
I16
sS'm'
(lp9159
I1
aI3
aI1
aI13
aI15
aI23
aI113
aI177
aI43
aI743
aI57
aI3875
aI7833
aI13619
aI17637
aI5547
asS'd'
I4580
sa(dp9160
S'a'
I14606
sS's'
I16
sS'm'
(lp9161
I1
aI1
aI3
aI13
aI21
aI7
aI123
aI83
aI391
aI731
aI597
aI2595
aI1949
aI14619
aI17141
aI60595
asS'd'
I4581
sa(dp9162
S'a'
I14614
sS's'
I16
sS'm'
(lp9163
I1
aI3
aI7
aI13
aI31
aI55
aI15
aI43
aI17
aI855
aI233
aI1411
aI1063
aI12977
aI22159
aI59185
asS'd'
I4582
sa(dp9164
S'a'
I14639
sS's'
I16
sS'm'
(lp9165
I1
aI3
aI7
aI7
aI17
aI53
aI67
aI37
aI245
aI941
aI1213
aI1965
aI6635
aI10189
aI12979
aI63503
asS'd'
I4583
sa(dp9166
S'a'
I14641
sS's'
I16
sS'm'
(lp9167
I1
aI1
aI5
aI15
aI31
aI23
aI15
aI175
aI177
aI643
aI1705
aI3541
aI2009
aI12005
aI27281
aI16821
asS'd'
I4584
sa(dp9168
S'a'
I14648
sS's'
I16
sS'm'
(lp9169
I1
aI3
aI1
aI13
aI27
aI37
aI1
aI171
aI255
aI445
aI171
aI3555
aI8169
aI399
aI20975
aI36195
asS'd'
I4585
sa(dp9170
S'a'
I14668
sS's'
I16
sS'm'
(lp9171
I1
aI3
aI7
aI11
aI13
aI15
aI123
aI65
aI173
aI317
aI1991
aI2093
aI8073
aI12831
aI15455
aI30175
asS'd'
I4586
sa(dp9172
S'a'
I14673
sS's'
I16
sS'm'
(lp9173
I1
aI3
aI1
aI7
aI5
aI53
aI59
aI219
aI407
aI501
aI875
aI2627
aI1335
aI14387
aI25523
aI28337
asS'd'
I4587
sa(dp9174
S'a'
I14679
sS's'
I16
sS'm'
(lp9175
I1
aI3
aI3
aI13
aI13
aI41
aI93
aI125
aI41
aI461
aI887
aI1085
aI3393
aI11945
aI16329
aI43531
asS'd'
I4588
sa(dp9176
S'a'
I14695
sS's'
I16
sS'm'
(lp9177
I1
aI3
aI1
aI11
aI21
aI39
aI1
aI185
aI429
aI285
aI443
aI1165
aI451
aI10903
aI31511
aI50555
asS'd'
I4589
sa(dp9178
S'a'
I14702
sS's'
I16
sS'm'
(lp9179
I1
aI1
aI7
aI5
aI11
aI25
aI61
aI171
aI493
aI733
aI215
aI1871
aI7783
aI14113
aI2061
aI58961
asS'd'
I4590
sa(dp9180
S'a'
I14704
sS's'
I16
sS'm'
(lp9181
I1
aI1
aI7
aI7
aI27
aI23
aI53
aI45
aI131
aI301
aI275
aI3855
aI4883
aI6303
aI25269
aI12697
asS'd'
I4591
sa(dp9182
S'a'
I14740
sS's'
I16
sS'm'
(lp9183
I1
aI3
aI5
aI7
aI11
aI15
aI71
aI101
aI377
aI803
aI1369
aI3741
aI633
aI10571
aI30659
aI31101
asS'd'
I4592
sa(dp9184
S'a'
I14754
sS's'
I16
sS'm'
(lp9185
I1
aI1
aI5
aI15
aI19
aI53
aI3
aI153
aI411
aI685
aI1405
aI109
aI5755
aI13311
aI3713
aI24579
asS'd'
I4593
sa(dp9186
S'a'
I14774
sS's'
I16
sS'm'
(lp9187
I1
aI1
aI3
aI3
aI27
aI7
aI89
aI39
aI5
aI853
aI1757
aI2927
aI2889
aI9735
aI17959
aI39301
asS'd'
I4594
sa(dp9188
S'a'
I14792
sS's'
I16
sS'm'
(lp9189
I1
aI3
aI1
aI3
aI13
aI41
aI57
aI71
aI187
aI285
aI825
aI1807
aI7261
aI2645
aI21861
aI1839
asS'd'
I4595
sa(dp9190
S'a'
I14797
sS's'
I16
sS'm'
(lp9191
I1
aI3
aI3
aI5
aI15
aI21
aI23
aI7
aI341
aI981
aI891
aI721
aI7221
aI3137
aI28725
aI54993
asS'd'
I4596
sa(dp9192
S'a'
I14806
sS's'
I16
sS'm'
(lp9193
I1
aI1
aI5
aI3
aI3
aI61
aI59
aI97
aI205
aI359
aI185
aI3361
aI7637
aI15473
aI6351
aI62097
asS'd'
I4597
sa(dp9194
S'a'
I14812
sS's'
I16
sS'm'
(lp9195
I1
aI1
aI1
aI9
aI13
aI11
aI123
aI71
aI199
aI251
aI535
aI371
aI1605
aI12107
aI13833
aI2019
asS'd'
I4598
sa(dp9196
S'a'
I14856
sS's'
I16
sS'm'
(lp9197
I1
aI1
aI7
aI13
aI27
aI1
aI43
aI73
aI409
aI601
aI1481
aI649
aI3293
aI12257
aI23377
aI17225
asS'd'
I4599
sa(dp9198
S'a'
I14876
sS's'
I16
sS'm'
(lp9199
I1
aI1
aI7
aI11
aI15
aI55
aI99
aI45
aI261
aI461
aI1507
aI3575
aI4425
aI9895
aI20191
aI61863
asS'd'
I4600
sa(dp9200
S'a'
I14900
sS's'
I16
sS'm'
(lp9201
I1
aI3
aI7
aI1
aI3
aI7
aI19
aI85
aI139
aI691
aI1685
aI137
aI7547
aI16143
aI14193
aI52479
asS'd'
I4601
sa(dp9202
S'a'
I14910
sS's'
I16
sS'm'
(lp9203
I1
aI3
aI5
aI9
aI17
aI61
aI7
aI189
aI31
aI641
aI1381
aI3999
aI4909
aI8463
aI31761
aI54493
asS'd'
I4602
sa(dp9204
S'a'
I14912
sS's'
I16
sS'm'
(lp9205
I1
aI1
aI5
aI15
aI17
aI15
aI69
aI111
aI55
aI1011
aI1859
aI2643
aI6043
aI5125
aI15875
aI611
asS'd'
I4603
sa(dp9206
S'a'
I14915
sS's'
I16
sS'm'
(lp9207
I1
aI1
aI3
aI5
aI3
aI33
aI73
aI227
aI327
aI369
aI189
aI1841
aI5625
aI1179
aI18651
aI34951
asS'd'
I4604
sa(dp9208
S'a'
I14922
sS's'
I16
sS'm'
(lp9209
I1
aI3
aI7
aI13
aI1
aI17
aI109
aI149
aI89
aI889
aI799
aI3423
aI2599
aI14525
aI12763
aI23855
asS'd'
I4605
sa(dp9210
S'a'
I14939
sS's'
I16
sS'm'
(lp9211
I1
aI1
aI3
aI15
aI5
aI63
aI87
aI7
aI63
aI171
aI1215
aI557
aI3009
aI16305
aI23517
aI40101
asS'd'
I4606
sa(dp9212
S'a'
I14946
sS's'
I16
sS'm'
(lp9213
I1
aI1
aI3
aI3
aI29
aI31
aI79
aI183
aI401
aI813
aI41
aI1111
aI5669
aI15697
aI7813
aI10215
asS'd'
I4607
sa(dp9214
S'a'
I14955
sS's'
I16
sS'm'
(lp9215
I1
aI1
aI5
aI7
aI9
aI25
aI25
aI57
aI343
aI375
aI535
aI3405
aI1909
aI3201
aI2417
aI52285
asS'd'
I4608
sa(dp9216
S'a'
I14966
sS's'
I16
sS'm'
(lp9217
I1
aI1
aI5
aI9
aI25
aI19
aI33
aI183
aI29
aI991
aI1045
aI2249
aI2933
aI12525
aI13943
aI10423
asS'd'
I4609
sa(dp9218
S'a'
I14976
sS's'
I16
sS'm'
(lp9219
I1
aI3
aI1
aI7
aI3
aI45
aI49
aI37
aI429
aI67
aI821
aI1289
aI7311
aI16165
aI25861
aI57715
asS'd'
I4610
sa(dp9220
S'a'
I14986
sS's'
I16
sS'm'
(lp9221
I1
aI1
aI7
aI3
aI19
aI3
aI33
aI153
aI505
aI683
aI611
aI1691
aI6421
aI15517
aI19161
aI49013
asS'd'
I4611
sa(dp9222
S'a'
I14993
sS's'
I16
sS'm'
(lp9223
I1
aI3
aI7
aI7
aI21
aI21
aI85
aI55
aI293
aI199
aI1671
aI1881
aI7147
aI8241
aI16173
aI51873
asS'd'
I4612
sa(dp9224
S'a'
I15012
sS's'
I16
sS'm'
(lp9225
I1
aI3
aI1
aI15
aI3
aI61
aI97
aI191
aI435
aI511
aI1599
aI2705
aI1897
aI2607
aI1801
aI48583
asS'd'
I4613
sa(dp9226
S'a'
I15041
sS's'
I16
sS'm'
(lp9227
I1
aI1
aI5
aI3
aI9
aI23
aI23
aI185
aI401
aI947
aI33
aI385
aI7491
aI14129
aI14561
aI13759
asS'd'
I4614
sa(dp9228
S'a'
I15053
sS's'
I16
sS'm'
(lp9229
I1
aI3
aI5
aI15
aI19
aI21
aI37
aI233
aI149
aI673
aI29
aI1315
aI3487
aI6705
aI28283
aI43135
asS'd'
I4615
sa(dp9230
S'a'
I15056
sS's'
I16
sS'm'
(lp9231
I1
aI1
aI1
aI11
aI9
aI35
aI101
aI255
aI345
aI829
aI689
aI2747
aI2145
aI2101
aI24863
aI35529
asS'd'
I4616
sa(dp9232
S'a'
I15059
sS's'
I16
sS'm'
(lp9233
I1
aI3
aI7
aI9
aI5
aI5
aI53
aI247
aI157
aI729
aI1621
aI129
aI2485
aI5371
aI11115
aI47771
asS'd'
I4617
sa(dp9234
S'a'
I15110
sS's'
I16
sS'm'
(lp9235
I1
aI1
aI3
aI9
aI29
aI29
aI13
aI229
aI87
aI281
aI1119
aI1085
aI4423
aI1667
aI27067
aI50397
asS'd'
I4618
sa(dp9236
S'a'
I15116
sS's'
I16
sS'm'
(lp9237
I1
aI1
aI3
aI7
aI11
aI29
aI77
aI85
aI121
aI495
aI501
aI3209
aI3531
aI2307
aI11367
aI34135
asS'd'
I4619
sa(dp9238
S'a'
I15133
sS's'
I16
sS'm'
(lp9239
I1
aI1
aI7
aI9
aI3
aI37
aI33
aI209
aI493
aI869
aI367
aI3221
aI1643
aI3353
aI22851
aI4313
asS'd'
I4620
sa(dp9240
S'a'
I15134
sS's'
I16
sS'm'
(lp9241
I1
aI1
aI1
aI7
aI15
aI53
aI27
aI17
aI29
aI345
aI821
aI1831
aI1963
aI10089
aI5101
aI32689
asS'd'
I4621
sa(dp9242
S'a'
I15137
sS's'
I16
sS'm'
(lp9243
I1
aI1
aI3
aI9
aI9
aI61
aI31
aI215
aI497
aI591
aI1301
aI157
aI3329
aI13877
aI9017
aI34673
asS'd'
I4622
sa(dp9244
S'a'
I15147
sS's'
I16
sS'm'
(lp9245
I1
aI1
aI5
aI1
aI29
aI49
aI93
aI139
aI279
aI167
aI143
aI279
aI6145
aI6247
aI519
aI8869
asS'd'
I4623
sa(dp9246
S'a'
I15182
sS's'
I16
sS'm'
(lp9247
I1
aI3
aI3
aI1
aI25
aI41
aI81
aI219
aI505
aI335
aI1941
aI2963
aI413
aI775
aI4181
aI55269
asS'd'
I4624
sa(dp9248
S'a'
I15203
sS's'
I16
sS'm'
(lp9249
I1
aI1
aI1
aI11
aI27
aI23
aI91
aI9
aI497
aI811
aI1469
aI1999
aI5377
aI2943
aI17635
aI11151
asS'd'
I4625
sa(dp9250
S'a'
I15215
sS's'
I16
sS'm'
(lp9251
I1
aI1
aI5
aI15
aI17
aI23
aI15
aI235
aI271
aI749
aI1873
aI3805
aI5405
aI7421
aI24339
aI14351
asS'd'
I4626
sa(dp9252
S'a'
I15245
sS's'
I16
sS'm'
(lp9253
I1
aI3
aI7
aI1
aI5
aI61
aI81
aI9
aI269
aI43
aI1391
aI2711
aI6973
aI11299
aI2263
aI8715
asS'd'
I4627
sa(dp9254
S'a'
I15246
sS's'
I16
sS'm'
(lp9255
I1
aI1
aI5
aI13
aI11
aI1
aI69
aI235
aI25
aI227
aI63
aI2591
aI913
aI12555
aI6263
aI38981
asS'd'
I4628
sa(dp9256
S'a'
I15264
sS's'
I16
sS'm'
(lp9257
I1
aI3
aI1
aI7
aI17
aI7
aI97
aI251
aI149
aI959
aI1895
aI1179
aI4031
aI15975
aI20313
aI64067
asS'd'
I4629
sa(dp9258
S'a'
I15269
sS's'
I16
sS'm'
(lp9259
I1
aI3
aI7
aI15
aI3
aI17
aI85
aI229
aI149
aI925
aI585
aI3755
aI2359
aI13131
aI12665
aI28861
asS'd'
I4630
sa(dp9260
S'a'
I15296
sS's'
I16
sS'm'
(lp9261
I1
aI3
aI3
aI9
aI5
aI31
aI107
aI93
aI347
aI851
aI1249
aI2161
aI6095
aI8315
aI20259
aI39527
asS'd'
I4631
sa(dp9262
S'a'
I15314
sS's'
I16
sS'm'
(lp9263
I1
aI3
aI7
aI7
aI21
aI63
aI85
aI241
aI501
aI627
aI1211
aI1713
aI6907
aI4229
aI7557
aI55719
asS'd'
I4632
sa(dp9264
S'a'
I15323
sS's'
I16
sS'm'
(lp9265
I1
aI1
aI1
aI13
aI19
aI43
aI21
aI177
aI13
aI353
aI679
aI511
aI5565
aI993
aI25345
aI25087
asS'd'
I4633
sa(dp9266
S'a'
I15364
sS's'
I16
sS'm'
(lp9267
I1
aI3
aI3
aI15
aI21
aI15
aI87
aI83
aI381
aI547
aI1429
aI2417
aI2425
aI2097
aI20889
aI12353
asS'd'
I4634
sa(dp9268
S'a'
I15386
sS's'
I16
sS'm'
(lp9269
I1
aI3
aI1
aI11
aI23
aI21
aI69
aI147
aI427
aI271
aI1829
aI525
aI2951
aI10773
aI32425
aI17685
asS'd'
I4635
sa(dp9270
S'a'
I15391
sS's'
I16
sS'm'
(lp9271
I1
aI3
aI1
aI7
aI15
aI55
aI21
aI131
aI195
aI927
aI635
aI3505
aI3817
aI14883
aI1149
aI10079
asS'd'
I4636
sa(dp9272
S'a'
I15436
sS's'
I16
sS'm'
(lp9273
I1
aI3
aI3
aI9
aI23
aI15
aI45
aI147
aI249
aI87
aI377
aI1551
aI4343
aI15373
aI2895
aI44973
asS'd'
I4637
sa(dp9274
S'a'
I15460
sS's'
I16
sS'm'
(lp9275
I1
aI3
aI1
aI7
aI31
aI63
aI67
aI107
aI109
aI1019
aI815
aI231
aI8135
aI559
aI8175
aI21735
asS'd'
I4638
sa(dp9276
S'a'
I15464
sS's'
I16
sS'm'
(lp9277
I1
aI1
aI5
aI7
aI7
aI63
aI103
aI133
aI167
aI883
aI1647
aI2827
aI6015
aI8541
aI16963
aI37129
asS'd'
I4639
sa(dp9278
S'a'
I15469
sS's'
I16
sS'm'
(lp9279
I1
aI3
aI5
aI9
aI27
aI25
aI59
aI147
aI29
aI943
aI865
aI1233
aI2165
aI15259
aI2235
aI25831
asS'd'
I4640
sa(dp9280
S'a'
I15470
sS's'
I16
sS'm'
(lp9281
I1
aI1
aI5
aI7
aI25
aI5
aI67
aI89
aI493
aI111
aI359
aI1115
aI7963
aI6545
aI7749
aI29179
asS'd'
I4641
sa(dp9282
S'a'
I15477
sS's'
I16
sS'm'
(lp9283
I1
aI3
aI7
aI5
aI19
aI17
aI89
aI195
aI337
aI115
aI1417
aI3837
aI4761
aI1959
aI16205
aI61597
asS'd'
I4642
sa(dp9284
S'a'
I15488
sS's'
I16
sS'm'
(lp9285
I1
aI1
aI5
aI11
aI25
aI43
aI3
aI111
aI491
aI897
aI1541
aI909
aI4751
aI739
aI7827
aI64485
asS'd'
I4643
sa(dp9286
S'a'
I15494
sS's'
I16
sS'm'
(lp9287
I1
aI1
aI5
aI15
aI19
aI61
aI39
aI111
aI451
aI419
aI1657
aI2427
aI4589
aI5577
aI23967
aI19259
asS'd'
I4644
sa(dp9288
S'a'
I15548
sS's'
I16
sS'm'
(lp9289
I1
aI3
aI3
aI1
aI31
aI15
aI7
aI131
aI329
aI847
aI537
aI1775
aI3833
aI5683
aI17267
aI16389
asS'd'
I4645
sa(dp9290
S'a'
I15551
sS's'
I16
sS'm'
(lp9291
I1
aI1
aI7
aI1
aI9
aI29
aI13
aI25
aI409
aI513
aI1695
aI2175
aI5099
aI727
aI5723
aI43547
asS'd'
I4646
sa(dp9292
S'a'
I15560
sS's'
I16
sS'm'
(lp9293
I1
aI1
aI5
aI7
aI13
aI11
aI29
aI123
aI127
aI193
aI1647
aI3157
aI2149
aI16209
aI19909
aI14473
asS'd'
I4647
sa(dp9294
S'a'
I15563
sS's'
I16
sS'm'
(lp9295
I1
aI1
aI1
aI15
aI15
aI37
aI125
aI157
aI487
aI143
aI1891
aI2895
aI4397
aI10685
aI1463
aI55027
asS'd'
I4648
sa(dp9296
S'a'
I15604
sS's'
I16
sS'm'
(lp9297
I1
aI3
aI7
aI1
aI1
aI15
aI115
aI105
aI479
aI529
aI1095
aI2687
aI4483
aI15027
aI15487
aI7113
asS'd'
I4649
sa(dp9298
S'a'
I15607
sS's'
I16
sS'm'
(lp9299
I1
aI1
aI3
aI9
aI23
aI63
aI113
aI211
aI155
aI931
aI175
aI3037
aI2359
aI10413
aI24561
aI21099
asS'd'
I4650
sa(dp9300
S'a'
I15616
sS's'
I16
sS'm'
(lp9301
I1
aI3
aI3
aI11
aI5
aI15
aI13
aI37
aI257
aI447
aI203
aI545
aI2467
aI9979
aI17543
aI62703
asS'd'
I4651
sa(dp9302
S'a'
I15631
sS's'
I16
sS'm'
(lp9303
I1
aI1
aI3
aI7
aI17
aI31
aI83
aI91
aI79
aI265
aI1415
aI2347
aI5337
aI7615
aI27739
aI33841
asS'd'
I4652
sa(dp9304
S'a'
I15699
sS's'
I16
sS'm'
(lp9305
I1
aI3
aI5
aI7
aI17
aI63
aI37
aI153
aI347
aI909
aI1945
aI7
aI2071
aI15195
aI32083
aI26713
asS'd'
I4653
sa(dp9306
S'a'
I15701
sS's'
I16
sS'm'
(lp9307
I1
aI1
aI3
aI11
aI19
aI51
aI69
aI21
aI323
aI635
aI443
aI1685
aI6275
aI13787
aI20921
aI45553
asS'd'
I4654
sa(dp9308
S'a'
I15708
sS's'
I16
sS'm'
(lp9309
I1
aI3
aI3
aI7
aI15
aI35
aI67
aI247
aI257
aI429
aI2029
aI523
aI3219
aI3893
aI26677
aI53273
asS'd'
I4655
sa(dp9310
S'a'
I15739
sS's'
I16
sS'm'
(lp9311
I1
aI1
aI7
aI9
aI9
aI3
aI119
aI121
aI445
aI149
aI1539
aI1887
aI2995
aI14867
aI809
aI48065
asS'd'
I4656
sa(dp9312
S'a'
I15746
sS's'
I16
sS'm'
(lp9313
I1
aI3
aI5
aI13
aI15
aI27
aI75
aI9
aI217
aI35
aI1363
aI2383
aI4357
aI1153
aI20565
aI62277
asS'd'
I4657
sa(dp9314
S'a'
I15772
sS's'
I16
sS'm'
(lp9315
I1
aI1
aI7
aI1
aI21
aI1
aI11
aI53
aI331
aI765
aI407
aI453
aI2725
aI11199
aI645
aI14915
asS'd'
I4658
sa(dp9316
S'a'
I15793
sS's'
I16
sS'm'
(lp9317
I1
aI1
aI5
aI1
aI29
aI11
aI5
aI159
aI395
aI53
aI323
aI1347
aI5529
aI8525
aI24003
aI20535
asS'd'
I4659
sa(dp9318
S'a'
I15832
sS's'
I16
sS'm'
(lp9319
I1
aI3
aI3
aI15
aI9
aI19
aI87
aI181
aI391
aI639
aI703
aI611
aI997
aI359
aI2471
aI58465
asS'd'
I4660
sa(dp9320
S'a'
I15837
sS's'
I16
sS'm'
(lp9321
I1
aI3
aI5
aI9
aI27
aI9
aI117
aI47
aI223
aI509
aI1537
aI1235
aI3885
aI6767
aI17131
aI63421
asS'd'
I4661
sa(dp9322
S'a'
I15866
sS's'
I16
sS'm'
(lp9323
I1
aI1
aI5
aI1
aI15
aI15
aI113
aI67
aI477
aI597
aI1795
aI3065
aI4565
aI3609
aI16419
aI19667
asS'd'
I4662
sa(dp9324
S'a'
I15899
sS's'
I16
sS'm'
(lp9325
I1
aI1
aI7
aI11
aI1
aI63
aI33
aI211
aI271
aI773
aI499
aI2309
aI1303
aI14015
aI30377
aI53195
asS'd'
I4663
sa(dp9326
S'a'
I15911
sS's'
I16
sS'm'
(lp9327
I1
aI1
aI7
aI11
aI5
aI23
aI17
aI183
aI321
aI315
aI203
aI3371
aI907
aI9331
aI21031
aI33765
asS'd'
I4664
sa(dp9328
S'a'
I15918
sS's'
I16
sS'm'
(lp9329
I1
aI3
aI3
aI7
aI7
aI53
aI111
aI69
aI441
aI283
aI195
aI2415
aI7293
aI7659
aI2731
aI5417
asS'd'
I4665
sa(dp9330
S'a'
I15952
sS's'
I16
sS'm'
(lp9331
I1
aI3
aI5
aI15
aI3
aI61
aI5
aI241
aI427
aI463
aI1729
aI653
aI7671
aI10979
aI7247
aI36931
asS'd'
I4666
sa(dp9332
S'a'
I15962
sS's'
I16
sS'm'
(lp9333
I1
aI3
aI1
aI9
aI3
aI5
aI105
aI117
aI465
aI853
aI2005
aI3925
aI4419
aI4441
aI3701
aI50747
asS'd'
I4667
sa(dp9334
S'a'
I15967
sS's'
I16
sS'm'
(lp9335
I1
aI1
aI3
aI7
aI1
aI3
aI3
aI149
aI65
aI405
aI299
aI99
aI481
aI14323
aI11565
aI6227
asS'd'
I4668
sa(dp9336
S'a'
I15973
sS's'
I16
sS'm'
(lp9337
I1
aI3
aI7
aI5
aI29
aI3
aI19
aI3
aI253
aI895
aI879
aI2435
aI2151
aI10673
aI11013
aI43055
asS'd'
I4669
sa(dp9338
S'a'
I15977
sS's'
I16
sS'm'
(lp9339
I1
aI3
aI1
aI11
aI15
aI57
aI127
aI197
aI319
aI913
aI1039
aI811
aI7767
aI5791
aI31725
aI8733
asS'd'
I4670
sa(dp9340
S'a'
I16016
sS's'
I16
sS'm'
(lp9341
I1
aI1
aI7
aI3
aI13
aI25
aI25
aI81
aI229
aI185
aI39
aI2789
aI579
aI4973
aI28617
aI60871
asS'd'
I4671
sa(dp9342
S'a'
I16035
sS's'
I16
sS'm'
(lp9343
I1
aI1
aI7
aI3
aI25
aI17
aI41
aI7
aI103
aI269
aI55
aI2651
aI7579
aI10935
aI8917
aI14323
asS'd'
I4672
sa(dp9344
S'a'
I16044
sS's'
I16
sS'm'
(lp9345
I1
aI3
aI7
aI7
aI13
aI7
aI99
aI205
aI293
aI877
aI1893
aI3013
aI2389
aI6021
aI2645
aI18175
asS'd'
I4673
sa(dp9346
S'a'
I16067
sS's'
I16
sS'm'
(lp9347
I1
aI1
aI3
aI7
aI9
aI39
aI59
aI187
aI191
aI761
aI339
aI3381
aI2921
aI5197
aI16963
aI43019
asS'd'
I4674
sa(dp9348
S'a'
I16082
sS's'
I16
sS'm'
(lp9349
I1
aI3
aI3
aI13
aI7
aI23
aI41
aI203
aI311
aI981
aI323
aI1675
aI6689
aI579
aI3885
aI64475
asS'd'
I4675
sa(dp9350
S'a'
I16084
sS's'
I16
sS'm'
(lp9351
I1
aI3
aI5
aI15
aI21
aI39
aI35
aI193
aI167
aI1009
aI493
aI829
aI6571
aI1299
aI13061
aI1163
asS'd'
I4676
sa(dp9352
S'a'
I16098
sS's'
I16
sS'm'
(lp9353
I1
aI1
aI3
aI5
aI3
aI19
aI123
aI123
aI111
aI599
aI193
aI3419
aI7173
aI5595
aI12449
aI52247
asS'd'
I4677
sa(dp9354
S'a'
I16107
sS's'
I16
sS'm'
(lp9355
I1
aI3
aI5
aI11
aI9
aI25
aI65
aI49
aI239
aI953
aI481
aI3455
aI4335
aI305
aI22469
aI11949
asS'd'
I4678
sa(dp9356
S'a'
I16142
sS's'
I16
sS'm'
(lp9357
I1
aI1
aI3
aI7
aI15
aI1
aI13
aI77
aI147
aI49
aI1445
aI931
aI3405
aI15951
aI15215
aI26451
asS'd'
I4679
sa(dp9358
S'a'
I16149
sS's'
I16
sS'm'
(lp9359
I1
aI3
aI1
aI1
aI21
aI53
aI17
aI7
aI247
aI243
aI805
aI795
aI489
aI13351
aI13493
aI30937
asS'd'
I4680
sa(dp9360
S'a'
I16165
sS's'
I16
sS'm'
(lp9361
I1
aI3
aI7
aI5
aI5
aI13
aI39
aI115
aI397
aI757
aI423
aI2559
aI1677
aI9449
aI24563
aI869
asS'd'
I4681
sa(dp9362
S'a'
I16172
sS's'
I16
sS'm'
(lp9363
I1
aI1
aI3
aI11
aI23
aI9
aI27
aI233
aI165
aI853
aI1721
aI599
aI551
aI11657
aI27623
aI40119
asS'd'
I4682
sa(dp9364
S'a'
I16178
sS's'
I16
sS'm'
(lp9365
I1
aI1
aI3
aI1
aI3
aI47
aI75
aI207
aI113
aI417
aI1317
aI2215
aI2395
aI1841
aI23125
aI50401
asS'd'
I4683
sa(dp9366
S'a'
I16197
sS's'
I16
sS'm'
(lp9367
I1
aI3
aI3
aI1
aI13
aI55
aI103
aI55
aI351
aI785
aI1665
aI3603
aI7705
aI4811
aI21129
aI38115
asS'd'
I4684
sa(dp9368
S'a'
I16201
sS's'
I16
sS'm'
(lp9369
I1
aI1
aI1
aI5
aI5
aI49
aI93
aI189
aI317
aI701
aI1545
aI1017
aI4133
aI7655
aI19827
aI52155
asS'd'
I4685
sa(dp9370
S'a'
I16215
sS's'
I16
sS'm'
(lp9371
I1
aI3
aI3
aI13
aI17
aI37
aI7
aI249
aI139
aI529
aI235
aI3801
aI7871
aI459
aI15127
aI13231
asS'd'
I4686
sa(dp9372
S'a'
I16221
sS's'
I16
sS'm'
(lp9373
I1
aI3
aI7
aI5
aI7
aI63
aI99
aI241
aI131
aI455
aI1287
aI3539
aI8029
aI12661
aI23313
aI54029
asS'd'
I4687
sa(dp9374
S'a'
I16226
sS's'
I16
sS'm'
(lp9375
I1
aI3
aI1
aI3
aI29
aI63
aI51
aI227
aI497
aI685
aI1351
aI449
aI7851
aI10815
aI17379
aI62097
asS'd'
I4688
sa(dp9376
S'a'
I16232
sS's'
I16
sS'm'
(lp9377
I1
aI3
aI1
aI11
aI25
aI61
aI73
aI29
aI467
aI533
aI855
aI853
aI5557
aI10953
aI18307
aI27135
asS'd'
I4689
sa(dp9378
S'a'
I16246
sS's'
I16
sS'm'
(lp9379
I1
aI1
aI7
aI3
aI13
aI49
aI63
aI171
aI177
aI283
aI1365
aI3087
aI5445
aI15109
aI12523
aI25193
asS'd'
I4690
sa(dp9380
S'a'
I16261
sS's'
I16
sS'm'
(lp9381
I1
aI3
aI5
aI15
aI9
aI39
aI95
aI81
aI417
aI199
aI1653
aI3673
aI2663
aI8101
aI12199
aI22759
asS'd'
I4691
sa(dp9382
S'a'
I16279
sS's'
I16
sS'm'
(lp9383
I1
aI1
aI3
aI9
aI29
aI15
aI29
aI215
aI21
aI721
aI245
aI1197
aI7251
aI5721
aI6735
aI7751
asS'd'
I4692
sa(dp9384
S'a'
I16280
sS's'
I16
sS'm'
(lp9385
I1
aI3
aI5
aI5
aI21
aI7
aI81
aI61
aI157
aI707
aI819
aI1689
aI4203
aI5559
aI25483
aI43325
asS'd'
I4693
sa(dp9386
S'a'
I16290
sS's'
I16
sS'm'
(lp9387
I1
aI1
aI7
aI13
aI15
aI51
aI47
aI197
aI269
aI921
aI353
aI2865
aI6227
aI537
aI20015
aI53823
asS'd'
I4694
sa(dp9388
S'a'
I16314
sS's'
I16
sS'm'
(lp9389
I1
aI1
aI3
aI5
aI13
aI25
aI91
aI221
aI111
aI587
aI1119
aI2343
aI4651
aI4641
aI15915
aI36323
asS'd'
I4695
sa(dp9390
S'a'
I16345
sS's'
I16
sS'm'
(lp9391
I1
aI1
aI7
aI11
aI1
aI45
aI7
aI215
aI483
aI545
aI731
aI3041
aI3121
aI8681
aI20651
aI4069
asS'd'
I4696
sa(dp9392
S'a'
I16355
sS's'
I16
sS'm'
(lp9393
I1
aI3
aI7
aI13
aI13
aI27
aI109
aI65
aI103
aI805
aI299
aI2069
aI6017
aI14565
aI20505
aI16161
asS'd'
I4697
sa(dp9394
S'a'
I16361
sS's'
I16
sS'm'
(lp9395
I1
aI1
aI7
aI5
aI11
aI33
aI105
aI213
aI237
aI583
aI1033
aI2333
aI845
aI6493
aI505
aI2563
asS'd'
I4698
sa(dp9396
S'a'
I16393
sS's'
I16
sS'm'
(lp9397
I1
aI1
aI5
aI7
aI3
aI5
aI11
aI173
aI373
aI341
aI269
aI177
aI3175
aI9413
aI601
aI13591
asS'd'
I4699
sa(dp9398
S'a'
I16394
sS's'
I16
sS'm'
(lp9399
I1
aI1
aI5
aI13
aI7
aI57
aI61
aI187
aI121
aI405
aI665
aI111
aI7535
aI3355
aI14051
aI511
asS'd'
I4700
sa(dp9400
S'a'
I16438
sS's'
I16
sS'm'
(lp9401
I1
aI1
aI1
aI3
aI3
aI29
aI15
aI253
aI227
aI123
aI333
aI1343
aI7313
aI1955
aI17741
aI30831
asS'd'
I4701
sa(dp9402
S'a'
I16450
sS's'
I16
sS'm'
(lp9403
I1
aI3
aI5
aI1
aI5
aI47
aI65
aI183
aI199
aI839
aI925
aI2711
aI4609
aI201
aI15177
aI29817
asS'd'
I4702
sa(dp9404
S'a'
I16507
sS's'
I16
sS'm'
(lp9405
I1
aI3
aI7
aI9
aI21
aI63
aI5
aI163
aI265
aI581
aI1795
aI3937
aI4641
aI2073
aI32225
aI60831
asS'd'
I4703
sa(dp9406
S'a'
I16523
sS's'
I16
sS'm'
(lp9407
I1
aI1
aI1
aI5
aI7
aI47
aI125
aI203
aI511
aI841
aI1937
aI3431
aI1495
aI12827
aI5893
aI19265
asS'd'
I4704
sa(dp9408
S'a'
I16533
sS's'
I16
sS'm'
(lp9409
I1
aI1
aI5
aI5
aI9
aI49
aI17
aI247
aI391
aI241
aI3
aI2253
aI6319
aI89
aI4449
aI6371
asS'd'
I4705
sa(dp9410
S'a'
I16603
sS's'
I16
sS'm'
(lp9411
I1
aI3
aI1
aI1
aI31
aI7
aI51
aI61
aI461
aI391
aI273
aI1609
aI5825
aI16029
aI3851
aI39213
asS'd'
I4706
sa(dp9412
S'a'
I16648
sS's'
I16
sS'm'
(lp9413
I1
aI3
aI3
aI7
aI29
aI21
aI65
aI75
aI317
aI925
aI1319
aI3827
aI965
aI5685
aI17007
aI64365
asS'd'
I4707
sa(dp9414
S'a'
I16653
sS's'
I16
sS'm'
(lp9415
I1
aI3
aI1
aI5
aI23
aI23
aI109
aI59
aI31
aI659
aI635
aI2209
aI857
aI9847
aI2507
aI18325
asS'd'
I4708
sa(dp9416
S'a'
I16672
sS's'
I16
sS'm'
(lp9417
I1
aI1
aI1
aI1
aI17
aI51
aI53
aI77
aI461
aI147
aI229
aI2821
aI2405
aI1259
aI1121
aI17429
asS'd'
I4709
sa(dp9418
S'a'
I16682
sS's'
I16
sS'm'
(lp9419
I1
aI3
aI5
aI3
aI31
aI3
aI57
aI157
aI321
aI731
aI1609
aI2139
aI899
aI12599
aI19803
aI51083
asS'd'
I4710
sa(dp9420
S'a'
I16709
sS's'
I16
sS'm'
(lp9421
I1
aI1
aI3
aI11
aI27
aI43
aI73
aI209
aI431
aI587
aI1247
aI2803
aI3593
aI1351
aI18701
aI33423
asS'd'
I4711
sa(dp9422
S'a'
I16713
sS's'
I16
sS'm'
(lp9423
I1
aI3
aI5
aI13
aI27
aI19
aI67
aI245
aI339
aI879
aI2035
aI1801
aI5845
aI3883
aI22057
aI15771
asS'd'
I4712
sa(dp9424
S'a'
I16719
sS's'
I16
sS'm'
(lp9425
I1
aI1
aI3
aI11
aI11
aI55
aI93
aI51
aI57
aI127
aI1325
aI3975
aI3989
aI2347
aI18831
aI2979
asS'd'
I4713
sa(dp9426
S'a'
I16733
sS's'
I16
sS'm'
(lp9427
I1
aI1
aI1
aI13
aI17
aI1
aI17
aI103
aI303
aI777
aI1973
aI2943
aI7439
aI8953
aI27227
aI10229
asS'd'
I4714
sa(dp9428
S'a'
I16740
sS's'
I16
sS'm'
(lp9429
I1
aI3
aI3
aI15
aI1
aI41
aI53
aI219
aI415
aI399
aI995
aI205
aI7719
aI10937
aI31879
aI755
asS'd'
I4715
sa(dp9430
S'a'
I16761
sS's'
I16
sS'm'
(lp9431
I1
aI3
aI7
aI9
aI13
aI7
aI99
aI93
aI419
aI1019
aI1605
aI161
aI3831
aI9147
aI7877
aI1333
asS'd'
I4716
sa(dp9432
S'a'
I16767
sS's'
I16
sS'm'
(lp9433
I1
aI3
aI7
aI15
aI5
aI51
aI37
aI115
aI259
aI549
aI353
aI2067
aI7657
aI1283
aI20333
aI2325
asS'd'
I4717
sa(dp9434
S'a'
I16771
sS's'
I16
sS'm'
(lp9435
I1
aI1
aI3
aI3
aI23
aI33
aI63
aI233
aI363
aI719
aI1099
aI471
aI3079
aI10577
aI19063
aI31535
asS'd'
I4718
sa(dp9436
S'a'
I16788
sS's'
I16
sS'm'
(lp9437
I1
aI3
aI7
aI15
aI23
aI19
aI109
aI105
aI497
aI881
aI1055
aI537
aI4607
aI239
aI22785
aI60201
asS'd'
I4719
sa(dp9438
S'a'
I16811
sS's'
I16
sS'm'
(lp9439
I1
aI3
aI3
aI5
aI19
aI11
aI1
aI207
aI163
aI437
aI713
aI667
aI1427
aI7505
aI28055
aI43101
asS'd'
I4720
sa(dp9440
S'a'
I16814
sS's'
I16
sS'm'
(lp9441
I1
aI3
aI5
aI5
aI25
aI45
aI75
aI9
aI109
aI545
aI573
aI2685
aI1013
aI2973
aI18619
aI55945
asS'd'
I4721
sa(dp9442
S'a'
I16816
sS's'
I16
sS'm'
(lp9443
I1
aI1
aI1
aI3
aI27
aI27
aI39
aI33
aI285
aI453
aI613
aI2707
aI2155
aI4355
aI21105
aI7969
asS'd'
I4722
sa(dp9444
S'a'
I16828
sS's'
I16
sS'm'
(lp9445
I1
aI3
aI3
aI9
aI1
aI31
aI71
aI101
aI491
aI409
aI65
aI1479
aI5743
aI525
aI2863
aI53657
asS'd'
I4723
sa(dp9446
S'a'
I16834
sS's'
I16
sS'm'
(lp9447
I1
aI1
aI3
aI1
aI17
aI63
aI55
aI11
aI125
aI447
aI275
aI2243
aI6827
aI5753
aI32401
aI13819
asS'd'
I4724
sa(dp9448
S'a'
I16863
sS's'
I16
sS'm'
(lp9449
I1
aI1
aI3
aI9
aI21
aI47
aI5
aI127
aI285
aI471
aI1681
aI945
aI6141
aI5651
aI10273
aI30811
asS'd'
I4725
sa(dp9450
S'a'
I16864
sS's'
I16
sS'm'
(lp9451
I1
aI3
aI3
aI1
aI13
aI53
aI91
aI3
aI255
aI429
aI107
aI2937
aI2971
aI10251
aI15009
aI37477
asS'd'
I4726
sa(dp9452
S'a'
I16879
sS's'
I16
sS'm'
(lp9453
I1
aI1
aI7
aI13
aI21
aI63
aI73
aI3
aI63
aI491
aI101
aI1981
aI7457
aI879
aI6243
aI22275
asS'd'
I4727
sa(dp9454
S'a'
I16888
sS's'
I16
sS'm'
(lp9455
I1
aI3
aI1
aI1
aI11
aI43
aI121
aI101
aI293
aI639
aI645
aI2723
aI2075
aI3629
aI22105
aI18199
asS'd'
I4728
sa(dp9456
S'a'
I16904
sS's'
I16
sS'm'
(lp9457
I1
aI1
aI3
aI1
aI31
aI9
aI69
aI97
aI511
aI663
aI1147
aI1237
aI389
aI255
aI8661
aI38533
asS'd'
I4729
sa(dp9458
S'a'
I16909
sS's'
I16
sS'm'
(lp9459
I1
aI3
aI3
aI7
aI3
aI13
aI23
aI71
aI197
aI439
aI2003
aI1771
aI8073
aI1549
aI29089
aI5409
asS'd'
I4730
sa(dp9460
S'a'
I16921
sS's'
I16
sS'm'
(lp9461
I1
aI3
aI1
aI1
aI9
aI23
aI1
aI221
aI159
aI699
aI593
aI3385
aI3869
aI10105
aI22097
aI34753
asS'd'
I4731
sa(dp9462
S'a'
I16934
sS's'
I16
sS'm'
(lp9463
I1
aI1
aI7
aI1
aI29
aI47
aI41
aI137
aI333
aI357
aI325
aI3151
aI6641
aI3823
aI8763
aI28327
asS'd'
I4732
sa(dp9464
S'a'
I16951
sS's'
I16
sS'm'
(lp9465
I1
aI3
aI1
aI7
aI19
aI19
aI39
aI225
aI477
aI619
aI583
aI229
aI6177
aI9615
aI1203
aI13711
asS'd'
I4733
sa(dp9466
S'a'
I16983
sS's'
I16
sS'm'
(lp9467
I1
aI1
aI3
aI13
aI9
aI41
aI127
aI147
aI13
aI301
aI861
aI2019
aI3517
aI1147
aI21587
aI42387
asS'd'
I4734
sa(dp9468
S'a'
I16999
sS's'
I16
sS'm'
(lp9469
I1
aI1
aI5
aI11
aI9
aI63
aI11
aI121
aI251
aI199
aI483
aI2287
aI4667
aI3887
aI10611
aI6019
asS'd'
I4735
sa(dp9470
S'a'
I17000
sS's'
I16
sS'm'
(lp9471
I1
aI1
aI3
aI13
aI23
aI19
aI89
aI73
aI355
aI399
aI749
aI687
aI5707
aI11443
aI817
aI38967
asS'd'
I4736
sa(dp9472
S'a'
I17006
sS's'
I16
sS'm'
(lp9473
I1
aI3
aI5
aI9
aI3
aI23
aI115
aI207
aI373
aI541
aI73
aI1285
aI7245
aI12505
aI5787
aI61207
asS'd'
I4737
sa(dp9474
S'a'
I17020
sS's'
I16
sS'm'
(lp9475
I1
aI3
aI5
aI15
aI27
aI37
aI115
aI203
aI195
aI793
aI1577
aI1283
aI7299
aI4025
aI5319
aI5375
asS'd'
I4738
sa(dp9476
S'a'
I17030
sS's'
I16
sS'm'
(lp9477
I1
aI1
aI7
aI15
aI25
aI19
aI61
aI11
aI215
aI771
aI1057
aI451
aI1965
aI13693
aI25617
aI42657
asS'd'
I4739
sa(dp9478
S'a'
I17033
sS's'
I16
sS'm'
(lp9479
I1
aI3
aI3
aI7
aI1
aI19
aI23
aI217
aI175
aI901
aI2009
aI2989
aI5111
aI5027
aI4805
aI37843
asS'd'
I4740
sa(dp9480
S'a'
I17044
sS's'
I16
sS'm'
(lp9481
I1
aI3
aI1
aI11
aI11
aI37
aI3
aI131
aI459
aI769
aI201
aI3979
aI3009
aI8691
aI27005
aI32175
asS'd'
I4741
sa(dp9482
S'a'
I17051
sS's'
I16
sS'm'
(lp9483
I1
aI3
aI5
aI7
aI27
aI27
aI117
aI23
aI403
aI1003
aI1501
aI785
aI6313
aI10187
aI5085
aI30751
asS'd'
I4742
sa(dp9484
S'a'
I17072
sS's'
I16
sS'm'
(lp9485
I1
aI1
aI7
aI3
aI11
aI41
aI73
aI151
aI19
aI657
aI131
aI1901
aI3879
aI14995
aI24085
aI56621
asS'd'
I4743
sa(dp9486
S'a'
I17078
sS's'
I16
sS'm'
(lp9487
I1
aI3
aI7
aI15
aI23
aI3
aI61
aI199
aI67
aI483
aI1961
aI3583
aI5937
aI5749
aI16685
aI11831
asS'd'
I4744
sa(dp9488
S'a'
I17084
sS's'
I16
sS'm'
(lp9489
I1
aI1
aI3
aI15
aI25
aI15
aI97
aI9
aI299
aI641
aI883
aI2901
aI123
aI1523
aI7055
aI15609
asS'd'
I4745
sa(dp9490
S'a'
I17087
sS's'
I16
sS'm'
(lp9491
I1
aI3
aI5
aI5
aI31
aI55
aI19
aI45
aI239
aI543
aI2005
aI1041
aI1711
aI11059
aI19927
aI17475
asS'd'
I4746
sa(dp9492
S'a'
I17090
sS's'
I16
sS'm'
(lp9493
I1
aI1
aI3
aI9
aI5
aI59
aI105
aI209
aI323
aI613
aI1963
aI2227
aI2947
aI11761
aI21375
aI13265
asS'd'
I4747
sa(dp9494
S'a'
I17123
sS's'
I16
sS'm'
(lp9495
I1
aI3
aI3
aI15
aI1
aI5
aI117
aI37
aI93
aI243
aI305
aI2299
aI5163
aI9205
aI28761
aI35987
asS'd'
I4748
sa(dp9496
S'a'
I17132
sS's'
I16
sS'm'
(lp9497
I1
aI1
aI1
aI5
aI5
aI29
aI13
aI147
aI457
aI187
aI1729
aI1547
aI7745
aI13155
aI7833
aI57449
asS'd'
I4749
sa(dp9498
S'a'
I17140
sS's'
I16
sS'm'
(lp9499
I1
aI3
aI3
aI13
aI1
aI51
aI49
aI253
aI23
aI389
aI1611
aI3045
aI5909
aI3947
aI25105
aI3327
asS'd'
I4750
sa(dp9500
S'a'
I17149
sS's'
I16
sS'm'
(lp9501
I1
aI3
aI1
aI11
aI15
aI47
aI19
aI15
aI231
aI57
aI763
aI1879
aI1765
aI14861
aI22893
aI19437
asS'd'
I4751
sa(dp9502
S'a'
I17157
sS's'
I16
sS'm'
(lp9503
I1
aI1
aI3
aI15
aI1
aI19
aI85
aI65
aI139
aI629
aI361
aI3513
aI3807
aI799
aI8349
aI29247
asS'd'
I4752
sa(dp9504
S'a'
I17164
sS's'
I16
sS'm'
(lp9505
I1
aI3
aI3
aI13
aI9
aI11
aI65
aI201
aI471
aI89
aI355
aI121
aI3947
aI10775
aI3599
aI6041
asS'd'
I4753
sa(dp9506
S'a'
I17170
sS's'
I16
sS'm'
(lp9507
I1
aI3
aI7
aI3
aI5
aI53
aI33
aI167
aI431
aI129
aI1449
aI3263
aI7691
aI12569
aI7551
aI41101
asS'd'
I4754
sa(dp9508
S'a'
I17179
sS's'
I16
sS'm'
(lp9509
I1
aI1
aI3
aI15
aI9
aI41
aI5
aI239
aI361
aI773
aI955
aI3663
aI6051
aI12889
aI5841
aI59615
asS'd'
I4755
sa(dp9510
S'a'
I17237
sS's'
I16
sS'm'
(lp9511
I1
aI1
aI7
aI5
aI5
aI33
aI97
aI9
aI495
aI845
aI1953
aI3533
aI5715
aI15055
aI25211
aI9351
asS'd'
I4756
sa(dp9512
S'a'
I17248
sS's'
I16
sS'm'
(lp9513
I1
aI3
aI1
aI11
aI25
aI37
aI83
aI153
aI289
aI739
aI353
aI1121
aI7641
aI2081
aI28439
aI38085
asS'd'
I4757
sa(dp9514
S'a'
I17260
sS's'
I16
sS'm'
(lp9515
I1
aI3
aI1
aI1
aI27
aI9
aI63
aI135
aI395
aI641
aI1759
aI3727
aI4371
aI5193
aI2783
aI12389
asS'd'
I4758
sa(dp9516
S'a'
I17272
sS's'
I16
sS'm'
(lp9517
I1
aI3
aI3
aI15
aI3
aI9
aI5
aI153
aI111
aI675
aI1957
aI3817
aI4269
aI10787
aI3413
aI34199
asS'd'
I4759
sa(dp9518
S'a'
I17275
sS's'
I16
sS'm'
(lp9519
I1
aI3
aI5
aI9
aI23
aI23
aI97
aI137
aI255
aI249
aI333
aI2329
aI1055
aI13769
aI13109
aI33443
asS'd'
I4760
sa(dp9520
S'a'
I17287
sS's'
I16
sS'm'
(lp9521
I1
aI1
aI1
aI15
aI7
aI37
aI99
aI219
aI483
aI755
aI263
aI3523
aI6179
aI4817
aI29873
aI12771
asS'd'
I4761
sa(dp9522
S'a'
I17294
sS's'
I16
sS'm'
(lp9523
I1
aI1
aI3
aI5
aI23
aI7
aI77
aI97
aI105
aI631
aI175
aI1911
aI7271
aI1009
aI24081
aI61207
asS'd'
I4762
sa(dp9524
S'a'
I17296
sS's'
I16
sS'm'
(lp9525
I1
aI3
aI5
aI3
aI1
aI31
aI71
aI91
aI265
aI669
aI1839
aI3989
aI8127
aI15001
aI1419
aI8895
asS'd'
I4763
sa(dp9526
S'a'
I17305
sS's'
I16
sS'm'
(lp9527
I1
aI3
aI1
aI13
aI27
aI51
aI93
aI155
aI49
aI991
aI653
aI203
aI3863
aI5363
aI31969
aI36083
asS'd'
I4764
sa(dp9528
S'a'
I17312
sS's'
I16
sS'm'
(lp9529
I1
aI3
aI3
aI7
aI31
aI27
aI21
aI73
aI21
aI675
aI407
aI1215
aI2963
aI6799
aI15259
aI13125
asS'd'
I4765
sa(dp9530
S'a'
I17321
sS's'
I16
sS'm'
(lp9531
I1
aI3
aI5
aI13
aI5
aI53
aI19
aI215
aI243
aI487
aI689
aI2519
aI6393
aI3987
aI30847
aI37919
asS'd'
I4766
sa(dp9532
S'a'
I17367
sS's'
I16
sS'm'
(lp9533
I1
aI3
aI3
aI7
aI5
aI31
aI115
aI231
aI255
aI955
aI2023
aI1487
aI6575
aI9873
aI22585
aI29951
asS'd'
I4767
sa(dp9534
S'a'
I17368
sS's'
I16
sS'm'
(lp9535
I1
aI3
aI5
aI11
aI11
aI57
aI109
aI203
aI417
aI29
aI1311
aI893
aI1047
aI2413
aI9305
aI38219
asS'd'
I4768
sa(dp9536
S'a'
I17378
sS's'
I16
sS'm'
(lp9537
I1
aI3
aI1
aI7
aI23
aI51
aI113
aI3
aI105
aI915
aI1145
aI3431
aI7331
aI3323
aI31669
aI21485
asS'd'
I4769
sa(dp9538
S'a'
I17433
sS's'
I16
sS'm'
(lp9539
I1
aI1
aI7
aI13
aI9
aI29
aI119
aI205
aI403
aI1023
aI257
aI863
aI2983
aI1895
aI16539
aI23385
asS'd'
I4770
sa(dp9540
S'a'
I17455
sS's'
I16
sS'm'
(lp9541
I1
aI1
aI7
aI13
aI27
aI21
aI47
aI139
aI341
aI509
aI1107
aI2197
aI3649
aI14301
aI30789
aI48783
asS'd'
I4771
sa(dp9542
S'a'
I17457
sS's'
I16
sS'm'
(lp9543
I1
aI3
aI3
aI7
aI25
aI19
aI99
aI11
aI309
aI919
aI1809
aI3015
aI1587
aI3779
aI1289
aI30207
asS'd'
I4772
sa(dp9544
S'a'
I17508
sS's'
I16
sS'm'
(lp9545
I1
aI3
aI5
aI11
aI9
aI43
aI57
aI171
aI9
aI151
aI173
aI2301
aI7723
aI2083
aI319
aI52883
asS'd'
I4773
sa(dp9546
S'a'
I17559
sS's'
I16
sS'm'
(lp9547
I1
aI1
aI3
aI1
aI3
aI13
aI63
aI11
aI231
aI117
aI1257
aI237
aI581
aI13871
aI15501
aI8741
asS'd'
I4774
sa(dp9548
S'a'
I17560
sS's'
I16
sS'm'
(lp9549
I1
aI3
aI5
aI9
aI13
aI63
aI55
aI155
aI291
aI721
aI123
aI929
aI3351
aI11651
aI12513
aI1779
asS'd'
I4775
sa(dp9550
S'a'
I17582
sS's'
I16
sS'm'
(lp9551
I1
aI3
aI7
aI3
aI31
aI5
aI61
aI81
aI465
aI639
aI1363
aI3157
aI2401
aI9513
aI32559
aI34477
asS'd'
I4776
sa(dp9552
S'a'
I17596
sS's'
I16
sS'm'
(lp9553
I1
aI3
aI1
aI15
aI27
aI25
aI3
aI117
aI277
aI13
aI707
aI3825
aI7287
aI10181
aI30127
aI57247
asS'd'
I4777
sa(dp9554
S'a'
I17599
sS's'
I16
sS'm'
(lp9555
I1
aI1
aI7
aI11
aI21
aI21
aI53
aI17
aI407
aI851
aI1191
aI285
aI6815
aI1595
aI25507
aI8099
asS'd'
I4778
sa(dp9556
S'a'
I17613
sS's'
I16
sS'm'
(lp9557
I1
aI3
aI5
aI9
aI9
aI61
aI83
aI61
aI65
aI671
aI63
aI311
aI6897
aI15327
aI29809
aI4899
asS'd'
I4779
sa(dp9558
S'a'
I17619
sS's'
I16
sS'm'
(lp9559
I1
aI1
aI7
aI1
aI21
aI45
aI99
aI235
aI477
aI461
aI1119
aI4087
aI2057
aI14861
aI31969
aI24357
asS'd'
I4780
sa(dp9560
S'a'
I17622
sS's'
I16
sS'm'
(lp9561
I1
aI1
aI7
aI9
aI31
aI9
aI65
aI123
aI281
aI273
aI1059
aI1625
aI6265
aI9635
aI11563
aI45053
asS'd'
I4781
sa(dp9562
S'a'
I17655
sS's'
I16
sS'm'
(lp9563
I1
aI3
aI3
aI7
aI1
aI41
aI15
aI23
aI43
aI727
aI1271
aI1741
aI765
aI13265
aI4145
aI1335
asS'd'
I4782
sa(dp9564
S'a'
I17661
sS's'
I16
sS'm'
(lp9565
I1
aI1
aI3
aI7
aI17
aI55
aI107
aI231
aI263
aI197
aI659
aI3621
aI2789
aI5171
aI28635
aI13595
asS'd'
I4783
sa(dp9566
S'a'
I17698
sS's'
I16
sS'm'
(lp9567
I1
aI1
aI5
aI1
aI27
aI23
aI13
aI83
aI431
aI507
aI1571
aI1573
aI1733
aI12171
aI8181
aI30843
asS'd'
I4784
sa(dp9568
S'a'
I17712
sS's'
I16
sS'm'
(lp9569
I1
aI3
aI7
aI11
aI1
aI53
aI107
aI39
aI497
aI579
aI453
aI1339
aI1415
aI10317
aI2741
aI34599
asS'd'
I4785
sa(dp9570
S'a'
I17715
sS's'
I16
sS'm'
(lp9571
I1
aI3
aI3
aI5
aI31
aI41
aI49
aI41
aI33
aI665
aI1783
aI87
aI317
aI6603
aI22603
aI22879
asS'd'
I4786
sa(dp9572
S'a'
I17721
sS's'
I16
sS'm'
(lp9573
I1
aI3
aI1
aI15
aI5
aI47
aI41
aI87
aI231
aI853
aI1615
aI2299
aI4643
aI9249
aI15641
aI14323
asS'd'
I4787
sa(dp9574
S'a'
I17722
sS's'
I16
sS'm'
(lp9575
I1
aI3
aI7
aI9
aI5
aI45
aI55
aI153
aI31
aI247
aI67
aI2335
aI6057
aI4379
aI27579
aI38437
asS'd'
I4788
sa(dp9576
S'a'
I17742
sS's'
I16
sS'm'
(lp9577
I1
aI1
aI5
aI7
aI9
aI3
aI73
aI81
aI479
aI909
aI1097
aI3945
aI4485
aI7815
aI22855
aI20825
asS'd'
I4789
sa(dp9578
S'a'
I17778
sS's'
I16
sS'm'
(lp9579
I1
aI3
aI1
aI15
aI19
aI43
aI97
aI57
aI339
aI167
aI135
aI1777
aI7681
aI9715
aI13863
aI6347
asS'd'
I4790
sa(dp9580
S'a'
I17818
sS's'
I16
sS'm'
(lp9581
I1
aI1
aI1
aI1
aI5
aI53
aI33
aI123
aI449
aI165
aI1283
aI2977
aI5919
aI12929
aI32073
aI61851
asS'd'
I4791
sa(dp9582
S'a'
I17836
sS's'
I16
sS'm'
(lp9583
I1
aI1
aI5
aI15
aI27
aI27
aI19
aI157
aI267
aI651
aI1319
aI3841
aI7739
aI9947
aI16801
aI41325
asS'd'
I4792
sa(dp9584
S'a'
I17841
sS's'
I16
sS'm'
(lp9585
I1
aI3
aI7
aI9
aI19
aI7
aI83
aI95
aI401
aI293
aI437
aI1983
aI119
aI7553
aI11097
aI11925
asS'd'
I4793
sa(dp9586
S'a'
I17856
sS's'
I16
sS'm'
(lp9587
I1
aI1
aI3
aI5
aI21
aI1
aI53
aI201
aI385
aI843
aI1801
aI99
aI2697
aI9659
aI28789
aI31417
asS'd'
I4794
sa(dp9588
S'a'
I17883
sS's'
I16
sS'm'
(lp9589
I1
aI1
aI5
aI5
aI29
aI57
aI103
aI89
aI77
aI597
aI1849
aI3433
aI4267
aI11167
aI3841
aI44023
asS'd'
I4795
sa(dp9590
S'a'
I17896
sS's'
I16
sS'm'
(lp9591
I1
aI1
aI7
aI1
aI21
aI47
aI113
aI253
aI249
aI431
aI1899
aI2859
aI7345
aI5725
aI14805
aI19999
asS'd'
I4796
sa(dp9592
S'a'
I17901
sS's'
I16
sS'm'
(lp9593
I1
aI3
aI3
aI5
aI1
aI11
aI77
aI213
aI359
aI665
aI1855
aI2743
aI2407
aI14677
aI17957
aI63257
asS'd'
I4797
sa(dp9594
S'a'
I17926
sS's'
I16
sS'm'
(lp9595
I1
aI3
aI7
aI13
aI23
aI29
aI127
aI183
aI275
aI849
aI1005
aI3159
aI3867
aI13029
aI7527
aI13035
asS'd'
I4798
sa(dp9596
S'a'
I17937
sS's'
I16
sS'm'
(lp9597
I1
aI1
aI1
aI15
aI29
aI47
aI81
aI225
aI77
aI879
aI1279
aI1929
aI1457
aI2025
aI32229
aI2847
asS'd'
I4799
sa(dp9598
S'a'
I17992
sS's'
I16
sS'm'
(lp9599
I1
aI1
aI1
aI3
aI29
aI45
aI37
aI189
aI217
aI53
aI281
aI1785
aI1929
aI763
aI5875
aI34303
asS'd'
I4800
sa(dp9600
S'a'
I17995
sS's'
I16
sS'm'
(lp9601
I1
aI3
aI1
aI9
aI21
aI61
aI21
aI149
aI215
aI13
aI1221
aI5
aI7153
aI14089
aI3119
aI33115
asS'd'
I4801
sa(dp9602
S'a'
I17998
sS's'
I16
sS'm'
(lp9603
I1
aI3
aI7
aI11
aI7
aI57
aI89
aI185
aI485
aI649
aI1765
aI747
aI2983
aI6791
aI25015
aI13627
asS'd'
I4802
sa(dp9604
S'a'
I18021
sS's'
I16
sS'm'
(lp9605
I1
aI1
aI1
aI9
aI11
aI53
aI77
aI203
aI425
aI237
aI487
aI2317
aI1047
aI8277
aI23405
aI30445
asS'd'
I4803
sa(dp9606
S'a'
I18039
sS's'
I16
sS'm'
(lp9607
I1
aI1
aI3
aI5
aI7
aI29
aI39
aI195
aI109
aI381
aI655
aI931
aI4469
aI16215
aI10627
aI64171
asS'd'
I4804
sa(dp9608
S'a'
I18067
sS's'
I16
sS'm'
(lp9609
I1
aI3
aI1
aI3
aI5
aI9
aI89
aI131
aI509
aI275
aI489
aI3161
aI3701
aI11951
aI6579
aI42839
asS'd'
I4805
sa(dp9610
S'a'
I18122
sS's'
I16
sS'm'
(lp9611
I1
aI3
aI7
aI13
aI15
aI37
aI65
aI91
aI305
aI433
aI1815
aI169
aI3117
aI47
aI30331
aI34863
asS'd'
I4806
sa(dp9612
S'a'
I18129
sS's'
I16
sS'm'
(lp9613
I1
aI3
aI3
aI9
aI21
aI3
aI21
aI113
aI25
aI833
aI1579
aI4021
aI3481
aI55
aI20833
aI46277
asS'd'
I4807
sa(dp9614
S'a'
I18130
sS's'
I16
sS'm'
(lp9615
I1
aI1
aI1
aI5
aI19
aI37
aI61
aI229
aI61
aI363
aI817
aI1235
aI6235
aI7261
aI29917
aI43057
asS'd'
I4808
sa(dp9616
S'a'
I18142
sS's'
I16
sS'm'
(lp9617
I1
aI3
aI1
aI9
aI7
aI59
aI119
aI189
aI341
aI945
aI633
aI3683
aI2589
aI15453
aI4989
aI40055
asS'd'
I4809
sa(dp9618
S'a'
I18148
sS's'
I16
sS'm'
(lp9619
I1
aI1
aI1
aI5
aI25
aI63
aI61
aI73
aI207
aI205
aI1011
aI2857
aI8137
aI2121
aI26731
aI46011
asS'd'
I4810
sa(dp9620
S'a'
I18163
sS's'
I16
sS'm'
(lp9621
I1
aI3
aI7
aI11
aI13
aI59
aI107
aI57
aI49
aI555
aI441
aI1771
aI4939
aI12107
aI8263
aI16243
asS'd'
I4811
sa(dp9622
S'a'
I18192
sS's'
I16
sS'm'
(lp9623
I1
aI3
aI5
aI13
aI15
aI49
aI89
aI217
aI83
aI225
aI2001
aI2727
aI4651
aI619
aI16473
aI47525
asS'd'
I4812
sa(dp9624
S'a'
I18211
sS's'
I16
sS'm'
(lp9625
I1
aI3
aI5
aI9
aI5
aI63
aI115
aI91
aI337
aI757
aI703
aI559
aI1683
aI14875
aI30769
aI30331
asS'd'
I4813
sa(dp9626
S'a'
I18228
sS's'
I16
sS'm'
(lp9627
I1
aI3
aI1
aI15
aI3
aI3
aI119
aI79
aI487
aI519
aI523
aI1177
aI7105
aI12343
aI24671
aI31711
asS'd'
I4814
sa(dp9628
S'a'
I18264
sS's'
I16
sS'm'
(lp9629
I1
aI1
aI7
aI15
aI25
aI63
aI87
aI23
aI59
aI277
aI849
aI953
aI4567
aI11309
aI26181
aI749
asS'd'
I4815
sa(dp9630
S'a'
I18347
sS's'
I16
sS'm'
(lp9631
I1
aI3
aI7
aI15
aI5
aI33
aI21
aI127
aI3
aI239
aI839
aI997
aI7253
aI8183
aI19779
aI4185
asS'd'
I4816
sa(dp9632
S'a'
I18372
sS's'
I16
sS'm'
(lp9633
I1
aI1
aI5
aI15
aI25
aI37
aI99
aI51
aI465
aI137
aI1339
aI541
aI4319
aI9883
aI17425
aI30743
asS'd'
I4817
sa(dp9634
S'a'
I18409
sS's'
I16
sS'm'
(lp9635
I1
aI3
aI3
aI5
aI13
aI7
aI3
aI249
aI365
aI749
aI527
aI3675
aI3005
aI12905
aI9621
aI899
asS'd'
I4818
sa(dp9636
S'a'
I18412
sS's'
I16
sS'm'
(lp9637
I1
aI3
aI3
aI7
aI29
aI31
aI69
aI21
aI365
aI1021
aI1329
aI2623
aI3549
aI5491
aI21293
aI63771
asS'd'
I4819
sa(dp9638
S'a'
I18418
sS's'
I16
sS'm'
(lp9639
I1
aI1
aI5
aI9
aI5
aI35
aI53
aI247
aI193
aI17
aI227
aI381
aI5233
aI821
aI3991
aI4439
asS'd'
I4820
sa(dp9640
S'a'
I18423
sS's'
I16
sS'm'
(lp9641
I1
aI1
aI7
aI15
aI5
aI59
aI27
aI167
aI489
aI335
aI1565
aI2999
aI1777
aI5009
aI519
aI57967
asS'd'
I4821
sa(dp9642
S'a'
I18433
sS's'
I16
sS'm'
(lp9643
I1
aI1
aI1
aI11
aI25
aI47
aI23
aI155
aI419
aI863
aI1905
aI355
aI1089
aI5871
aI10149
aI53341
asS'd'
I4822
sa(dp9644
S'a'
I18439
sS's'
I16
sS'm'
(lp9645
I1
aI1
aI7
aI7
aI29
aI55
aI127
aI83
aI33
aI309
aI2017
aI1021
aI5987
aI1101
aI13657
aI60587
asS'd'
I4823
sa(dp9646
S'a'
I18445
sS's'
I16
sS'm'
(lp9647
I1
aI1
aI1
aI7
aI3
aI1
aI9
aI75
aI257
aI407
aI659
aI529
aI2087
aI14759
aI14483
aI36425
asS'd'
I4824
sa(dp9648
S'a'
I18451
sS's'
I16
sS'm'
(lp9649
I1
aI3
aI7
aI3
aI11
aI29
aI113
aI255
aI301
aI799
aI1171
aI3721
aI135
aI3467
aI7109
aI50339
asS'd'
I4825
sa(dp9650
S'a'
I18467
sS's'
I16
sS'm'
(lp9651
I1
aI1
aI1
aI7
aI21
aI25
aI15
aI31
aI61
aI491
aI57
aI189
aI2033
aI4363
aI31295
aI16313
asS'd'
I4826
sa(dp9652
S'a'
I18502
sS's'
I16
sS'm'
(lp9653
I1
aI1
aI5
aI1
aI5
aI17
aI33
aI77
aI483
aI469
aI355
aI2245
aI4165
aI459
aI30411
aI29507
asS'd'
I4827
sa(dp9654
S'a'
I18514
sS's'
I16
sS'm'
(lp9655
I1
aI1
aI3
aI13
aI1
aI27
aI29
aI85
aI491
aI787
aI1157
aI1299
aI4743
aI14795
aI32587
aI12807
asS'd'
I4828
sa(dp9656
S'a'
I18547
sS's'
I16
sS'm'
(lp9657
I1
aI3
aI3
aI1
aI23
aI45
aI35
aI129
aI3
aI55
aI969
aI2387
aI3929
aI10397
aI19879
aI2723
asS'd'
I4829
sa(dp9658
S'a'
I18575
sS's'
I16
sS'm'
(lp9659
I1
aI1
aI1
aI7
aI19
aI3
aI9
aI23
aI497
aI347
aI2039
aI913
aI5925
aI7965
aI5789
aI40949
asS'd'
I4830
sa(dp9660
S'a'
I18593
sS's'
I16
sS'm'
(lp9661
I1
aI1
aI7
aI1
aI29
aI61
aI89
aI3
aI133
aI647
aI1585
aI2661
aI1875
aI1859
aI3937
aI12707
asS'd'
I4831
sa(dp9662
S'a'
I18613
sS's'
I16
sS'm'
(lp9663
I1
aI3
aI3
aI7
aI25
aI11
aI41
aI149
aI427
aI463
aI901
aI2433
aI2617
aI13511
aI3443
aI39867
asS'd'
I4832
sa(dp9664
S'a'
I18620
sS's'
I16
sS'm'
(lp9665
I1
aI1
aI1
aI5
aI27
aI33
aI103
aI251
aI201
aI1023
aI1979
aI3745
aI6365
aI14945
aI22153
aI55535
asS'd'
I4833
sa(dp9666
S'a'
I18637
sS's'
I16
sS'm'
(lp9667
I1
aI3
aI1
aI15
aI23
aI25
aI57
aI215
aI111
aI181
aI385
aI1123
aI3095
aI7085
aI31863
aI37393
asS'd'
I4834
sa(dp9668
S'a'
I18640
sS's'
I16
sS'm'
(lp9669
I1
aI3
aI5
aI13
aI17
aI35
aI27
aI159
aI255
aI241
aI413
aI1823
aI5329
aI1825
aI28489
aI58763
asS'd'
I4835
sa(dp9670
S'a'
I18712
sS's'
I16
sS'm'
(lp9671
I1
aI3
aI1
aI15
aI3
aI33
aI97
aI27
aI409
aI889
aI409
aI2315
aI4743
aI14827
aI3037
aI57149
asS'd'
I4836
sa(dp9672
S'a'
I18728
sS's'
I16
sS'm'
(lp9673
I1
aI1
aI3
aI5
aI19
aI63
aI93
aI51
aI233
aI715
aI1571
aI1101
aI2751
aI14805
aI25683
aI13323
asS'd'
I4837
sa(dp9674
S'a'
I18742
sS's'
I16
sS'm'
(lp9675
I1
aI3
aI7
aI3
aI21
aI15
aI117
aI179
aI263
aI229
aI199
aI2597
aI3999
aI3037
aI3801
aI4775
asS'd'
I4838
sa(dp9676
S'a'
I18748
sS's'
I16
sS'm'
(lp9677
I1
aI3
aI3
aI15
aI1
aI15
aI49
aI91
aI383
aI21
aI1955
aI773
aI1213
aI5971
aI19525
aI8699
asS'd'
I4839
sa(dp9678
S'a'
I18753
sS's'
I16
sS'm'
(lp9679
I1
aI3
aI1
aI15
aI29
aI49
aI11
aI101
aI261
aI761
aI709
aI3811
aI4055
aI15675
aI32305
aI15173
asS'd'
I4840
sa(dp9680
S'a'
I18756
sS's'
I16
sS'm'
(lp9681
I1
aI1
aI1
aI3
aI9
aI41
aI127
aI23
aI413
aI461
aI237
aI1595
aI2257
aI2971
aI31845
aI61485
asS'd'
I4841
sa(dp9682
S'a'
I18771
sS's'
I16
sS'm'
(lp9683
I1
aI1
aI1
aI11
aI23
aI13
aI63
aI21
aI23
aI209
aI1317
aI1071
aI3619
aI7275
aI9343
aI21455
asS'd'
I4842
sa(dp9684
S'a'
I18814
sS's'
I16
sS'm'
(lp9685
I1
aI1
aI5
aI9
aI31
aI35
aI41
aI249
aI477
aI569
aI1175
aI1571
aI4679
aI10337
aI3507
aI23415
asS'd'
I4843
sa(dp9686
S'a'
I18818
sS's'
I16
sS'm'
(lp9687
I1
aI3
aI5
aI11
aI29
aI3
aI117
aI65
aI301
aI913
aI1709
aI1765
aI1801
aI15513
aI31495
aI38131
asS'd'
I4844
sa(dp9688
S'a'
I18827
sS's'
I16
sS'm'
(lp9689
I1
aI3
aI5
aI11
aI27
aI3
aI71
aI195
aI81
aI313
aI505
aI3941
aI3797
aI2031
aI24151
aI65085
asS'd'
I4845
sa(dp9690
S'a'
I18835
sS's'
I16
sS'm'
(lp9691
I1
aI1
aI1
aI5
aI13
aI17
aI59
aI151
aI191
aI489
aI1267
aI3207
aI4495
aI15145
aI12789
aI46313
asS'd'
I4846
sa(dp9692
S'a'
I18842
sS's'
I16
sS'm'
(lp9693
I1
aI3
aI1
aI7
aI29
aI9
aI25
aI31
aI33
aI527
aI1939
aI4089
aI333
aI757
aI8895
aI25331
asS'd'
I4847
sa(dp9694
S'a'
I18854
sS's'
I16
sS'm'
(lp9695
I1
aI1
aI1
aI1
aI9
aI27
aI11
aI205
aI211
aI141
aI1171
aI1881
aI4029
aI10587
aI30103
aI39661
asS'd'
I4848
sa(dp9696
S'a'
I18858
sS's'
I16
sS'm'
(lp9697
I1
aI1
aI3
aI3
aI23
aI3
aI23
aI175
aI355
aI753
aI183
aI1331
aI6403
aI3369
aI29891
aI11109
asS'd'
I4849
sa(dp9698
S'a'
I18895
sS's'
I16
sS'm'
(lp9699
I1
aI3
aI7
aI3
aI17
aI25
aI95
aI145
aI135
aI525
aI1073
aI1841
aI3951
aI2027
aI23053
aI19699
asS'd'
I4850
sa(dp9700
S'a'
I18914
sS's'
I16
sS'm'
(lp9701
I1
aI1
aI5
aI3
aI11
aI43
aI117
aI67
aI299
aI885
aI1095
aI2777
aI8185
aI14331
aI29543
aI655
asS'd'
I4851
sa(dp9702
S'a'
I18925
sS's'
I16
sS'm'
(lp9703
I1
aI3
aI7
aI7
aI3
aI59
aI127
aI147
aI323
aI713
aI99
aI1179
aI6395
aI1821
aI12673
aI35587
asS'd'
I4852
sa(dp9704
S'a'
I18933
sS's'
I16
sS'm'
(lp9705
I1
aI3
aI5
aI3
aI7
aI11
aI33
aI87
aI99
aI967
aI1443
aI1585
aI6215
aI15125
aI30747
aI21513
asS'd'
I4853
sa(dp9706
S'a'
I18937
sS's'
I16
sS'm'
(lp9707
I1
aI3
aI7
aI11
aI23
aI5
aI91
aI171
aI229
aI601
aI833
aI3157
aI1691
aI15081
aI20607
aI7643
asS'd'
I4854
sa(dp9708
S'a'
I18944
sS's'
I16
sS'm'
(lp9709
I1
aI1
aI3
aI1
aI5
aI1
aI39
aI59
aI157
aI7
aI601
aI2079
aI3045
aI3693
aI26511
aI13245
asS'd'
I4855
sa(dp9710
S'a'
I18973
sS's'
I16
sS'm'
(lp9711
I1
aI3
aI5
aI9
aI9
aI27
aI83
aI135
aI185
aI379
aI2027
aI1407
aI7409
aI7363
aI26471
aI35907
asS'd'
I4856
sa(dp9712
S'a'
I19001
sS's'
I16
sS'm'
(lp9713
I1
aI3
aI7
aI15
aI29
aI49
aI1
aI69
aI383
aI915
aI183
aI3809
aI4511
aI8751
aI4715
aI7033
asS'd'
I4857
sa(dp9714
S'a'
I19012
sS's'
I16
sS'm'
(lp9715
I1
aI3
aI3
aI3
aI1
aI17
aI71
aI233
aI243
aI933
aI1165
aI3089
aI3565
aI6521
aI13203
aI13065
asS'd'
I4858
sa(dp9716
S'a'
I19016
sS's'
I16
sS'm'
(lp9717
I1
aI1
aI5
aI9
aI9
aI55
aI53
aI129
aI331
aI943
aI587
aI2573
aI2247
aI15101
aI4987
aI62983
asS'd'
I4859
sa(dp9718
S'a'
I19027
sS's'
I16
sS'm'
(lp9719
I1
aI3
aI1
aI13
aI11
aI43
aI45
aI127
aI129
aI857
aI669
aI321
aI3915
aI4477
aI26973
aI19911
asS'd'
I4860
sa(dp9720
S'a'
I19040
sS's'
I16
sS'm'
(lp9721
I1
aI3
aI1
aI13
aI15
aI3
aI83
aI23
aI13
aI441
aI953
aI2373
aI3539
aI4895
aI26327
aI61961
asS'd'
I4861
sa(dp9722
S'a'
I19074
sS's'
I16
sS'm'
(lp9723
I1
aI1
aI5
aI13
aI23
aI11
aI125
aI83
aI339
aI901
aI1809
aI2691
aI3789
aI15007
aI10569
aI65399
asS'd'
I4862
sa(dp9724
S'a'
I19085
sS's'
I16
sS'm'
(lp9725
I1
aI3
aI1
aI1
aI17
aI27
aI105
aI199
aI435
aI245
aI1227
aI3029
aI3911
aI1021
aI2931
aI24957
asS'd'
I4863
sa(dp9726
S'a'
I19093
sS's'
I16
sS'm'
(lp9727
I1
aI3
aI1
aI11
aI17
aI5
aI123
aI39
aI413
aI627
aI1149
aI3925
aI6635
aI8959
aI31387
aI65047
asS'd'
I4864
sa(dp9728
S'a'
I19100
sS's'
I16
sS'm'
(lp9729
I1
aI3
aI5
aI1
aI23
aI41
aI93
aI217
aI21
aI115
aI1311
aI3901
aI2559
aI2925
aI30755
aI7575
asS'd'
I4865
sa(dp9730
S'a'
I19107
sS's'
I16
sS'm'
(lp9731
I1
aI1
aI3
aI9
aI13
aI11
aI63
aI171
aI135
aI983
aI1679
aI395
aI7349
aI5153
aI26405
aI40589
asS'd'
I4866
sa(dp9732
S'a'
I19128
sS's'
I16
sS'm'
(lp9733
I1
aI3
aI7
aI13
aI27
aI47
aI53
aI169
aI85
aI871
aI1087
aI619
aI7399
aI9719
aI6349
aI59211
asS'd'
I4867
sa(dp9734
S'a'
I19141
sS's'
I16
sS'm'
(lp9735
I1
aI3
aI3
aI15
aI31
aI45
aI3
aI33
aI11
aI879
aI929
aI1977
aI1939
aI1049
aI16159
aI41515
asS'd'
I4868
sa(dp9736
S'a'
I19142
sS's'
I16
sS'm'
(lp9737
I1
aI3
aI5
aI11
aI9
aI27
aI13
aI23
aI127
aI747
aI1121
aI2497
aI8141
aI8601
aI12431
aI3243
asS'd'
I4869
sa(dp9738
S'a'
I19156
sS's'
I16
sS'm'
(lp9739
I1
aI3
aI7
aI15
aI23
aI43
aI23
aI225
aI283
aI13
aI1837
aI2089
aI6435
aI10121
aI22307
aI58767
asS'd'
I4870
sa(dp9740
S'a'
I19169
sS's'
I16
sS'm'
(lp9741
I1
aI1
aI5
aI11
aI17
aI3
aI41
aI221
aI143
aI669
aI261
aI1367
aI7813
aI15265
aI32751
aI62007
asS'd'
I4871
sa(dp9742
S'a'
I19176
sS's'
I16
sS'm'
(lp9743
I1
aI1
aI1
aI1
aI5
aI45
aI41
aI161
aI327
aI185
aI1403
aI485
aI2831
aI10025
aI12555
aI47661
asS'd'
I4872
sa(dp9744
S'a'
I19222
sS's'
I16
sS'm'
(lp9745
I1
aI3
aI7
aI1
aI31
aI55
aI87
aI83
aI439
aI929
aI653
aI4095
aI5443
aI7361
aI27801
aI12979
asS'd'
I4873
sa(dp9746
S'a'
I19226
sS's'
I16
sS'm'
(lp9747
I1
aI3
aI1
aI7
aI1
aI57
aI11
aI145
aI55
aI269
aI711
aI1889
aI8023
aI7171
aI3247
aI35691
asS'd'
I4874
sa(dp9748
S'a'
I19247
sS's'
I16
sS'm'
(lp9749
I1
aI1
aI1
aI3
aI15
aI1
aI15
aI131
aI479
aI797
aI815
aI2343
aI1603
aI10775
aI21341
aI20825
asS'd'
I4875
sa(dp9750
S'a'
I19259
sS's'
I16
sS'm'
(lp9751
I1
aI3
aI5
aI9
aI3
aI27
aI31
aI117
aI441
aI177
aI215
aI3991
aI3197
aI8397
aI19195
aI3883
asS'd'
I4876
sa(dp9752
S'a'
I19262
sS's'
I16
sS'm'
(lp9753
I1
aI1
aI5
aI13
aI1
aI19
aI13
aI27
aI157
aI1001
aI43
aI251
aI7997
aI7495
aI16515
aI44287
asS'd'
I4877
sa(dp9754
S'a'
I19291
sS's'
I16
sS'm'
(lp9755
I1
aI1
aI3
aI5
aI17
aI57
aI117
aI53
aI413
aI905
aI551
aI149
aI7531
aI14885
aI32493
aI34961
asS'd'
I4878
sa(dp9756
S'a'
I19309
sS's'
I16
sS'm'
(lp9757
I1
aI3
aI3
aI7
aI27
aI1
aI7
aI13
aI259
aI21
aI189
aI451
aI6171
aI3603
aI12053
aI45619
asS'd'
I4879
sa(dp9758
S'a'
I19324
sS's'
I16
sS'm'
(lp9759
I1
aI1
aI7
aI11
aI5
aI41
aI59
aI119
aI419
aI93
aI1399
aI629
aI1269
aI3789
aI17035
aI61583
asS'd'
I4880
sa(dp9760
S'a'
I19334
sS's'
I16
sS'm'
(lp9761
I1
aI3
aI5
aI11
aI1
aI11
aI59
aI83
aI473
aI273
aI839
aI3111
aI3081
aI10159
aI6143
aI16297
asS'd'
I4881
sa(dp9762
S'a'
I19338
sS's'
I16
sS'm'
(lp9763
I1
aI1
aI5
aI15
aI25
aI15
aI17
aI63
aI275
aI927
aI189
aI89
aI6595
aI4399
aI27201
aI57205
asS'd'
I4882
sa(dp9764
S'a'
I19343
sS's'
I16
sS'm'
(lp9765
I1
aI1
aI7
aI3
aI31
aI17
aI63
aI203
aI321
aI655
aI1751
aI2985
aI3291
aI11567
aI15135
aI49747
asS'd'
I4883
sa(dp9766
S'a'
I19376
sS's'
I16
sS'm'
(lp9767
I1
aI3
aI5
aI13
aI27
aI25
aI89
aI39
aI299
aI833
aI1269
aI271
aI6481
aI3045
aI7203
aI20279
asS'd'
I4884
sa(dp9768
S'a'
I19408
sS's'
I16
sS'm'
(lp9769
I1
aI3
aI1
aI1
aI29
aI13
aI13
aI37
aI33
aI563
aI1825
aI3257
aI3153
aI963
aI25801
aI15861
asS'd'
I4885
sa(dp9770
S'a'
I19413
sS's'
I16
sS'm'
(lp9771
I1
aI3
aI5
aI11
aI15
aI7
aI49
aI117
aI479
aI221
aI579
aI2995
aI1673
aI14927
aI2869
aI35547
asS'd'
I4886
sa(dp9772
S'a'
I19420
sS's'
I16
sS'm'
(lp9773
I1
aI3
aI1
aI11
aI31
aI11
aI77
aI161
aI183
aI187
aI1967
aI3037
aI4441
aI10547
aI1443
aI8619
asS'd'
I4887
sa(dp9774
S'a'
I19441
sS's'
I16
sS'm'
(lp9775
I1
aI1
aI3
aI11
aI27
aI41
aI83
aI179
aI293
aI421
aI1395
aI1673
aI6375
aI9801
aI14265
aI18117
asS'd'
I4888
sa(dp9776
S'a'
I19444
sS's'
I16
sS'm'
(lp9777
I1
aI1
aI3
aI7
aI9
aI19
aI55
aI235
aI499
aI637
aI1121
aI3583
aI8007
aI3749
aI20903
aI6179
asS'd'
I4889
sa(dp9778
S'a'
I19454
sS's'
I16
sS'm'
(lp9779
I1
aI3
aI7
aI13
aI9
aI55
aI125
aI77
aI395
aI9
aI2005
aI2247
aI1609
aI6805
aI13099
aI26367
asS'd'
I4890
sa(dp9780
S'a'
I19461
sS's'
I16
sS'm'
(lp9781
I1
aI3
aI5
aI13
aI9
aI41
aI49
aI133
aI443
aI995
aI209
aI341
aI8013
aI11037
aI29663
aI21161
asS'd'
I4891
sa(dp9782
S'a'
I19473
sS's'
I16
sS'm'
(lp9783
I1
aI3
aI1
aI1
aI1
aI47
aI45
aI243
aI161
aI433
aI129
aI39
aI4885
aI8777
aI6395
aI16953
asS'd'
I4892
sa(dp9784
S'a'
I19479
sS's'
I16
sS'm'
(lp9785
I1
aI3
aI3
aI15
aI11
aI13
aI41
aI113
aI279
aI657
aI763
aI2411
aI7115
aI463
aI10759
aI50493
asS'd'
I4893
sa(dp9786
S'a'
I19489
sS's'
I16
sS'm'
(lp9787
I1
aI1
aI5
aI5
aI31
aI5
aI25
aI181
aI385
aI445
aI625
aI313
aI4983
aI11763
aI6065
aI63835
asS'd'
I4894
sa(dp9788
S'a'
I19591
sS's'
I16
sS'm'
(lp9789
I1
aI3
aI3
aI15
aI13
aI25
aI103
aI5
aI205
aI223
aI1327
aI73
aI677
aI6751
aI2071
aI24963
asS'd'
I4895
sa(dp9790
S'a'
I19605
sS's'
I16
sS'm'
(lp9791
I1
aI1
aI7
aI15
aI21
aI61
aI21
aI11
aI47
aI775
aI113
aI991
aI1943
aI1459
aI18049
aI45025
asS'd'
I4896
sa(dp9792
S'a'
I19616
sS's'
I16
sS'm'
(lp9793
I1
aI3
aI3
aI1
aI11
aI43
aI27
aI89
aI49
aI711
aI173
aI181
aI1261
aI2751
aI21321
aI5467
asS'd'
I4897
sa(dp9794
S'a'
I19619
sS's'
I16
sS'm'
(lp9795
I1
aI3
aI3
aI7
aI17
aI7
aI57
aI61
aI87
aI973
aI985
aI1849
aI559
aI7319
aI11457
aI46071
asS'd'
I4898
sa(dp9796
S'a'
I19653
sS's'
I16
sS'm'
(lp9797
I1
aI1
aI1
aI1
aI9
aI37
aI99
aI157
aI423
aI189
aI1355
aI3983
aI6357
aI10825
aI26517
aI45815
asS'd'
I4899
sa(dp9798
S'a'
I19654
sS's'
I16
sS'm'
(lp9799
I1
aI1
aI3
aI11
aI23
aI33
aI57
aI55
aI59
aI831
aI339
aI725
aI359
aI14859
aI17523
aI36149
asS'd'
I4900
sa(dp9800
S'a'
I19681
sS's'
I16
sS'm'
(lp9801
I1
aI1
aI5
aI5
aI27
aI29
aI47
aI147
aI153
aI801
aI1737
aI3617
aI5447
aI8011
aI30631
aI7921
asS'd'
I4901
sa(dp9802
S'a'
I19711
sS's'
I16
sS'm'
(lp9803
I1
aI1
aI5
aI1
aI11
aI43
aI35
aI105
aI69
aI453
aI1023
aI875
aI6755
aI6015
aI12449
aI50235
asS'd'
I4902
sa(dp9804
S'a'
I19719
sS's'
I16
sS'm'
(lp9805
I1
aI3
aI1
aI5
aI29
aI31
aI43
aI89
aI369
aI891
aI1447
aI353
aI8103
aI2555
aI1197
aI64005
asS'd'
I4903
sa(dp9806
S'a'
I19726
sS's'
I16
sS'm'
(lp9807
I1
aI3
aI3
aI9
aI21
aI33
aI117
aI205
aI473
aI289
aI1699
aI2361
aI7083
aI13001
aI24127
aI48611
asS'd'
I4904
sa(dp9808
S'a'
I19738
sS's'
I16
sS'm'
(lp9809
I1
aI3
aI3
aI15
aI3
aI23
aI79
aI139
aI475
aI511
aI181
aI1331
aI6137
aI2653
aI14071
aI16947
asS'd'
I4905
sa(dp9810
S'a'
I19767
sS's'
I16
sS'm'
(lp9811
I1
aI3
aI3
aI1
aI11
aI47
aI51
aI217
aI305
aI599
aI1609
aI237
aI4537
aI5377
aI717
aI13269
asS'd'
I4906
sa(dp9812
S'a'
I19819
sS's'
I16
sS'm'
(lp9813
I1
aI1
aI7
aI3
aI19
aI31
aI1
aI173
aI509
aI817
aI785
aI1223
aI5585
aI8911
aI643
aI44575
asS'd'
I4907
sa(dp9814
S'a'
I19864
sS's'
I16
sS'm'
(lp9815
I1
aI1
aI3
aI11
aI5
aI11
aI31
aI129
aI269
aI369
aI1833
aI2885
aI441
aI11517
aI2323
aI26735
asS'd'
I4908
sa(dp9816
S'a'
I19867
sS's'
I16
sS'm'
(lp9817
I1
aI1
aI5
aI9
aI7
aI51
aI31
aI21
aI5
aI157
aI541
aI2939
aI4569
aI1679
aI17467
aI27995
asS'd'
I4909
sa(dp9818
S'a'
I19885
sS's'
I16
sS'm'
(lp9819
I1
aI1
aI7
aI3
aI21
aI33
aI85
aI213
aI41
aI851
aI1947
aI3205
aI5065
aI6079
aI30789
aI63677
asS'd'
I4910
sa(dp9820
S'a'
I19894
sS's'
I16
sS'm'
(lp9821
I1
aI1
aI5
aI3
aI9
aI53
aI3
aI179
aI157
aI407
aI537
aI1193
aI4645
aI13791
aI28153
aI11349
asS'd'
I4911
sa(dp9822
S'a'
I19900
sS's'
I16
sS'm'
(lp9823
I1
aI1
aI7
aI13
aI25
aI61
aI9
aI151
aI263
aI143
aI1583
aI2859
aI6617
aI8861
aI4163
aI40695
asS'd'
I4912
sa(dp9824
S'a'
I19903
sS's'
I16
sS'm'
(lp9825
I1
aI1
aI1
aI1
aI7
aI25
aI19
aI207
aI335
aI1019
aI1919
aI3893
aI831
aI12421
aI4667
aI38967
asS'd'
I4913
sa(dp9826
S'a'
I19941
sS's'
I16
sS'm'
(lp9827
I1
aI3
aI5
aI3
aI5
aI51
aI81
aI9
aI425
aI333
aI451
aI2569
aI2771
aI12145
aI26065
aI14385
asS'd'
I4914
sa(dp9828
S'a'
I19951
sS's'
I16
sS'm'
(lp9829
I1
aI1
aI3
aI7
aI3
aI49
aI17
aI147
aI327
aI331
aI1197
aI7
aI3703
aI15247
aI9723
aI52819
asS'd'
I4915
sa(dp9830
S'a'
I19959
sS's'
I16
sS'm'
(lp9831
I1
aI3
aI3
aI7
aI27
aI21
aI117
aI229
aI255
aI603
aI195
aI1049
aI6243
aI13593
aI14553
aI8267
asS'd'
I4916
sa(dp9832
S'a'
I19966
sS's'
I16
sS'm'
(lp9833
I1
aI1
aI5
aI15
aI9
aI53
aI1
aI187
aI79
aI151
aI321
aI1883
aI6105
aI13879
aI8201
aI53213
asS'd'
I4917
sa(dp9834
S'a'
I20009
sS's'
I16
sS'm'
(lp9835
I1
aI1
aI1
aI7
aI21
aI27
aI103
aI147
aI351
aI901
aI1927
aI2145
aI2685
aI453
aI15173
aI7371
asS'd'
I4918
sa(dp9836
S'a'
I20018
sS's'
I16
sS'm'
(lp9837
I1
aI1
aI3
aI5
aI21
aI27
aI125
aI77
aI395
aI27
aI827
aI3617
aI6033
aI1511
aI29461
aI18907
asS'd'
I4919
sa(dp9838
S'a'
I20020
sS's'
I16
sS'm'
(lp9839
I1
aI3
aI5
aI3
aI3
aI27
aI75
aI129
aI441
aI831
aI1213
aI2499
aI5829
aI12181
aI7991
aI58167
asS'd'
I4920
sa(dp9840
S'a'
I20038
sS's'
I16
sS'm'
(lp9841
I1
aI1
aI1
aI9
aI3
aI33
aI85
aI135
aI45
aI405
aI1731
aI551
aI1251
aI7895
aI3975
aI41621
asS'd'
I4921
sa(dp9842
S'a'
I20049
sS's'
I16
sS'm'
(lp9843
I1
aI3
aI5
aI7
aI19
aI19
aI25
aI7
aI477
aI569
aI1089
aI2705
aI6157
aI10279
aI14029
aI36229
asS'd'
I4922
sa(dp9844
S'a'
I20066
sS's'
I16
sS'm'
(lp9845
I1
aI3
aI7
aI3
aI5
aI19
aI99
aI137
aI67
aI361
aI2021
aI2891
aI1957
aI14961
aI22673
aI45707
asS'd'
I4923
sa(dp9846
S'a'
I20108
sS's'
I16
sS'm'
(lp9847
I1
aI3
aI7
aI1
aI21
aI11
aI81
aI225
aI151
aI235
aI1761
aI3875
aI7427
aI11213
aI27023
aI17945
asS'd'
I4924
sa(dp9848
S'a'
I20130
sS's'
I16
sS'm'
(lp9849
I1
aI1
aI7
aI1
aI3
aI1
aI3
aI123
aI39
aI769
aI1467
aI1907
aI7833
aI2099
aI19
aI54653
asS'd'
I4925
sa(dp9850
S'a'
I20132
sS's'
I16
sS'm'
(lp9851
I1
aI1
aI1
aI3
aI25
aI35
aI33
aI111
aI407
aI497
aI1527
aI3999
aI3083
aI6221
aI8319
aI56959
asS'd'
I4926
sa(dp9852
S'a'
I20167
sS's'
I16
sS'm'
(lp9853
I1
aI1
aI3
aI15
aI21
aI49
aI113
aI11
aI191
aI801
aI1835
aI3413
aI3379
aI13129
aI3655
aI23885
asS'd'
I4927
sa(dp9854
S'a'
I20219
sS's'
I16
sS'm'
(lp9855
I1
aI3
aI1
aI5
aI21
aI57
aI87
aI133
aI409
aI325
aI569
aI2099
aI8143
aI315
aI23287
aI21905
asS'd'
I4928
sa(dp9856
S'a'
I20227
sS's'
I16
sS'm'
(lp9857
I1
aI1
aI5
aI5
aI21
aI43
aI25
aI169
aI265
aI123
aI81
aI2683
aI6137
aI7341
aI16383
aI26435
asS'd'
I4929
sa(dp9858
S'a'
I20263
sS's'
I16
sS'm'
(lp9859
I1
aI3
aI1
aI5
aI23
aI17
aI125
aI173
aI3
aI829
aI693
aI751
aI8021
aI3701
aI32643
aI35405
asS'd'
I4930
sa(dp9860
S'a'
I20267
sS's'
I16
sS'm'
(lp9861
I1
aI1
aI3
aI1
aI31
aI13
aI1
aI195
aI435
aI487
aI961
aI1681
aI1233
aI6001
aI3113
aI29515
asS'd'
I4931
sa(dp9862
S'a'
I20272
sS's'
I16
sS'm'
(lp9863
I1
aI1
aI7
aI5
aI9
aI41
aI81
aI137
aI257
aI337
aI1837
aI145
aI4191
aI6313
aI9991
aI25541
asS'd'
I4932
sa(dp9864
S'a'
I20289
sS's'
I16
sS'm'
(lp9865
I1
aI1
aI5
aI13
aI29
aI13
aI1
aI117
aI501
aI991
aI571
aI793
aI1433
aI6005
aI19
aI61135
asS'd'
I4933
sa(dp9866
S'a'
I20296
sS's'
I16
sS'm'
(lp9867
I1
aI1
aI7
aI1
aI9
aI43
aI65
aI69
aI297
aI331
aI1777
aI1843
aI6477
aI13943
aI1301
aI51001
asS'd'
I4934
sa(dp9868
S'a'
I20307
sS's'
I16
sS'm'
(lp9869
I1
aI1
aI1
aI3
aI7
aI35
aI23
aI211
aI33
aI649
aI255
aI1831
aI635
aI9965
aI16679
aI14531
asS'd'
I4935
sa(dp9870
S'a'
I20316
sS's'
I16
sS'm'
(lp9871
I1
aI1
aI1
aI13
aI23
aI57
aI113
aI247
aI321
aI401
aI1761
aI4001
aI1823
aI14457
aI5251
aI4965
asS'd'
I4936
sa(dp9872
S'a'
I20323
sS's'
I16
sS'm'
(lp9873
I1
aI1
aI5
aI5
aI31
aI5
aI53
aI103
aI297
aI575
aI1577
aI2217
aI977
aI14415
aI16953
aI14793
asS'd'
I4937
sa(dp9874
S'a'
I20335
sS's'
I16
sS'm'
(lp9875
I1
aI1
aI5
aI7
aI9
aI19
aI25
aI29
aI121
aI563
aI1707
aI901
aI6167
aI10799
aI11897
aI31187
asS'd'
I4938
sa(dp9876
S'a'
I20344
sS's'
I16
sS'm'
(lp9877
I1
aI1
aI5
aI9
aI17
aI39
aI89
aI29
aI251
aI259
aI411
aI819
aI6037
aI4601
aI11481
aI46141
asS'd'
I4939
sa(dp9878
S'a'
I20354
sS's'
I16
sS'm'
(lp9879
I1
aI1
aI1
aI15
aI23
aI9
aI65
aI95
aI189
aI93
aI1485
aI2417
aI5183
aI5513
aI26623
aI42637
asS'd'
I4940
sa(dp9880
S'a'
I20360
sS's'
I16
sS'm'
(lp9881
I1
aI1
aI5
aI5
aI3
aI3
aI113
aI161
aI463
aI225
aI1089
aI2023
aI2341
aI14931
aI28097
aI56365
asS'd'
I4941
sa(dp9882
S'a'
I20368
sS's'
I16
sS'm'
(lp9883
I1
aI1
aI5
aI9
aI9
aI3
aI109
aI141
aI27
aI473
aI107
aI4023
aI3279
aI7595
aI13289
aI35649
asS'd'
I4942
sa(dp9884
S'a'
I20390
sS's'
I16
sS'm'
(lp9885
I1
aI1
aI5
aI3
aI9
aI37
aI73
aI153
aI487
aI57
aI2035
aI3583
aI239
aI2183
aI10903
aI171
asS'd'
I4943
sa(dp9886
S'a'
I20402
sS's'
I16
sS'm'
(lp9887
I1
aI3
aI3
aI15
aI23
aI39
aI87
aI217
aI451
aI597
aI1817
aI2883
aI145
aI11341
aI16015
aI16765
asS'd'
I4944
sa(dp9888
S'a'
I20413
sS's'
I16
sS'm'
(lp9889
I1
aI1
aI7
aI5
aI19
aI61
aI45
aI37
aI473
aI883
aI277
aI2801
aI13
aI7021
aI3851
aI53223
asS'd'
I4945
sa(dp9890
S'a'
I20425
sS's'
I16
sS'm'
(lp9891
I1
aI3
aI3
aI9
aI1
aI35
aI97
aI237
aI279
aI541
aI1911
aI661
aI7603
aI8183
aI22071
aI37317
asS'd'
I4946
sa(dp9892
S'a'
I20428
sS's'
I16
sS'm'
(lp9893
I1
aI3
aI3
aI11
aI11
aI63
aI101
aI71
aI227
aI259
aI1545
aI2779
aI3859
aI4859
aI18957
aI31749
asS'd'
I4947
sa(dp9894
S'a'
I20434
sS's'
I16
sS'm'
(lp9895
I1
aI3
aI3
aI1
aI27
aI29
aI91
aI215
aI381
aI607
aI1701
aI1709
aI247
aI12403
aI29943
aI59533
asS'd'
I4948
sa(dp9896
S'a'
I20443
sS's'
I16
sS'm'
(lp9897
I1
aI1
aI7
aI1
aI11
aI23
aI47
aI141
aI37
aI881
aI1443
aI3921
aI3281
aI7417
aI1549
aI50653
asS'd'
I4949
sa(dp9898
S'a'
I20488
sS's'
I16
sS'm'
(lp9899
I1
aI1
aI7
aI11
aI23
aI61
aI17
aI39
aI373
aI871
aI1107
aI1875
aI1419
aI3981
aI1333
aI11485
asS'd'
I4950
sa(dp9900
S'a'
I20502
sS's'
I16
sS'm'
(lp9901
I1
aI1
aI7
aI11
aI21
aI51
aI127
aI145
aI129
aI425
aI1263
aI1989
aI699
aI7317
aI24827
aI15049
asS'd'
I4951
sa(dp9902
S'a'
I20505
sS's'
I16
sS'm'
(lp9903
I1
aI1
aI1
aI11
aI9
aI59
aI59
aI67
aI329
aI841
aI905
aI467
aI1905
aI895
aI29711
aI25585
asS'd'
I4952
sa(dp9904
S'a'
I20535
sS's'
I16
sS'm'
(lp9905
I1
aI1
aI1
aI15
aI3
aI39
aI11
aI205
aI297
aI383
aI445
aI2139
aI2935
aI2399
aI22237
aI20355
asS'd'
I4953
sa(dp9906
S'a'
I20541
sS's'
I16
sS'm'
(lp9907
I1
aI3
aI7
aI7
aI17
aI9
aI17
aI205
aI369
aI1019
aI1703
aI755
aI5507
aI14749
aI16461
aI14519
asS'd'
I4954
sa(dp9908
S'a'
I20554
sS's'
I16
sS'm'
(lp9909
I1
aI3
aI7
aI3
aI5
aI31
aI97
aI35
aI43
aI249
aI773
aI4033
aI6085
aI1241
aI24743
aI22415
asS'd'
I4955
sa(dp9910
S'a'
I20577
sS's'
I16
sS'm'
(lp9911
I1
aI3
aI7
aI3
aI17
aI11
aI45
aI203
aI251
aI669
aI1463
aI1897
aI7913
aI2315
aI30307
aI15431
asS'd'
I4956
sa(dp9912
S'a'
I20583
sS's'
I16
sS'm'
(lp9913
I1
aI1
aI5
aI5
aI7
aI53
aI83
aI13
aI1
aI841
aI423
aI1059
aI3951
aI14209
aI11113
aI13931
asS'd'
I4957
sa(dp9914
S'a'
I20602
sS's'
I16
sS'm'
(lp9915
I1
aI3
aI3
aI5
aI5
aI15
aI11
aI71
aI237
aI553
aI829
aI3653
aI4991
aI1003
aI8353
aI52173
asS'd'
I4958
sa(dp9916
S'a'
I20611
sS's'
I16
sS'm'
(lp9917
I1
aI3
aI3
aI9
aI27
aI39
aI83
aI137
aI315
aI883
aI1155
aI3541
aI3815
aI10633
aI13277
aI20269
asS'd'
I4959
sa(dp9918
S'a'
I20614
sS's'
I16
sS'm'
(lp9919
I1
aI3
aI3
aI15
aI13
aI55
aI43
aI19
aI345
aI351
aI1117
aI1747
aI1949
aI3195
aI12241
aI52441
asS'd'
I4960
sa(dp9920
S'a'
I20626
sS's'
I16
sS'm'
(lp9921
I1
aI1
aI3
aI5
aI1
aI11
aI113
aI117
aI37
aI103
aI1681
aI3269
aI1659
aI14779
aI30479
aI31123
asS'd'
I4961
sa(dp9922
S'a'
I20628
sS's'
I16
sS'm'
(lp9923
I1
aI3
aI7
aI13
aI1
aI63
aI9
aI63
aI65
aI737
aI785
aI1713
aI8123
aI10053
aI29871
aI17647
asS'd'
I4962
sa(dp9924
S'a'
I20635
sS's'
I16
sS'm'
(lp9925
I1
aI1
aI3
aI5
aI17
aI45
aI71
aI37
aI283
aI145
aI1927
aI75
aI7355
aI4681
aI2777
aI31465
asS'd'
I4963
sa(dp9926
S'a'
I20642
sS's'
I16
sS'm'
(lp9927
I1
aI1
aI3
aI7
aI21
aI19
aI113
aI89
aI67
aI751
aI99
aI421
aI201
aI6345
aI9473
aI39849
asS'd'
I4964
sa(dp9928
S'a'
I20674
sS's'
I16
sS'm'
(lp9929
I1
aI1
aI5
aI11
aI31
aI57
aI75
aI79
aI393
aI745
aI1435
aI3039
aI1175
aI983
aI923
aI42867
asS'd'
I4965
sa(dp9930
S'a'
I20716
sS's'
I16
sS'm'
(lp9931
I1
aI1
aI5
aI9
aI31
aI47
aI31
aI61
aI85
aI651
aI1733
aI3973
aI1979
aI7223
aI13817
aI30593
asS'd'
I4966
sa(dp9932
S'a'
I20734
sS's'
I16
sS'm'
(lp9933
I1
aI1
aI1
aI11
aI31
aI21
aI23
aI177
aI401
aI55
aI537
aI3775
aI3241
aI15157
aI11849
aI15629
asS'd'
I4967
sa(dp9934
S'a'
I20765
sS's'
I16
sS'm'
(lp9935
I1
aI1
aI1
aI13
aI31
aI53
aI79
aI57
aI35
aI617
aI61
aI89
aI6917
aI6045
aI23879
aI45485
asS'd'
I4968
sa(dp9936
S'a'
I20801
sS's'
I16
sS'm'
(lp9937
I1
aI3
aI7
aI7
aI3
aI43
aI57
aI243
aI107
aI321
aI273
aI2171
aI2069
aI6171
aI29181
aI8281
asS'd'
I4969
sa(dp9938
S'a'
I20804
sS's'
I16
sS'm'
(lp9939
I1
aI1
aI1
aI11
aI3
aI27
aI51
aI57
aI81
aI795
aI1673
aI2601
aI7335
aI16243
aI863
aI20663
asS'd'
I4970
sa(dp9940
S'a'
I20808
sS's'
I16
sS'm'
(lp9941
I1
aI1
aI5
aI9
aI7
aI19
aI31
aI87
aI509
aI899
aI1133
aI1609
aI2163
aI7595
aI10523
aI43181
asS'd'
I4971
sa(dp9942
S'a'
I20831
sS's'
I16
sS'm'
(lp9943
I1
aI1
aI1
aI7
aI21
aI53
aI103
aI43
aI507
aI317
aI685
aI1329
aI7057
aI7275
aI2277
aI28271
asS'd'
I4972
sa(dp9944
S'a'
I20832
sS's'
I16
sS'm'
(lp9945
I1
aI1
aI7
aI7
aI3
aI35
aI81
aI81
aI261
aI587
aI1571
aI2771
aI4653
aI6517
aI25101
aI56753
asS'd'
I4973
sa(dp9946
S'a'
I20862
sS's'
I16
sS'm'
(lp9947
I1
aI3
aI1
aI11
aI17
aI61
aI29
aI137
aI7
aI929
aI393
aI2513
aI2423
aI5457
aI6285
aI12525
asS'd'
I4974
sa(dp9948
S'a'
I20877
sS's'
I16
sS'm'
(lp9949
I1
aI3
aI1
aI9
aI25
aI63
aI17
aI45
aI439
aI591
aI273
aI877
aI7741
aI8381
aI32277
aI24635
asS'd'
I4975
sa(dp9950
S'a'
I20880
sS's'
I16
sS'm'
(lp9951
I1
aI3
aI1
aI5
aI19
aI11
aI17
aI175
aI297
aI77
aI961
aI3331
aI5193
aI14347
aI12713
aI32067
asS'd'
I4976
sa(dp9952
S'a'
I20885
sS's'
I16
sS'm'
(lp9953
I1
aI1
aI5
aI11
aI3
aI17
aI13
aI21
aI219
aI653
aI1279
aI1657
aI7659
aI14459
aI27661
aI38093
asS'd'
I4977
sa(dp9954
S'a'
I20889
sS's'
I16
sS'm'
(lp9955
I1
aI3
aI7
aI7
aI29
aI17
aI67
aI35
aI451
aI91
aI919
aI3163
aI7459
aI14971
aI4317
aI42503
asS'd'
I4978
sa(dp9956
S'a'
I20905
sS's'
I16
sS'm'
(lp9957
I1
aI3
aI3
aI15
aI7
aI61
aI69
aI211
aI349
aI97
aI911
aI503
aI3903
aI12327
aI11049
aI29387
asS'd'
I4979
sa(dp9958
S'a'
I20914
sS's'
I16
sS'm'
(lp9959
I1
aI1
aI7
aI3
aI5
aI7
aI63
aI237
aI387
aI931
aI693
aI1143
aI6545
aI8529
aI25217
aI53087
asS'd'
I4980
sa(dp9960
S'a'
I20967
sS's'
I16
sS'm'
(lp9961
I1
aI1
aI5
aI7
aI1
aI13
aI21
aI169
aI259
aI289
aI437
aI649
aI4905
aI15261
aI29997
aI34043
asS'd'
I4981
sa(dp9962
S'a'
I21028
sS's'
I16
sS'm'
(lp9963
I1
aI1
aI1
aI9
aI25
aI13
aI19
aI229
aI29
aI213
aI1941
aI1501
aI3463
aI15761
aI15635
aI39687
asS'd'
I4982
sa(dp9964
S'a'
I21031
sS's'
I16
sS'm'
(lp9965
I1
aI1
aI5
aI7
aI13
aI29
aI101
aI57
aI483
aI913
aI1025
aI2139
aI4327
aI7847
aI12455
aI41797
asS'd'
I4983
sa(dp9966
S'a'
I21043
sS's'
I16
sS'm'
(lp9967
I1
aI1
aI3
aI11
aI17
aI27
aI97
aI79
aI411
aI9
aI1797
aI3721
aI5291
aI859
aI8889
aI6151
asS'd'
I4984
sa(dp9968
S'a'
I21052
sS's'
I16
sS'm'
(lp9969
I1
aI1
aI1
aI5
aI17
aI61
aI45
aI187
aI157
aI301
aI1017
aI1507
aI6031
aI4271
aI32593
aI23739
asS'd'
I4985
sa(dp9970
S'a'
I21058
sS's'
I16
sS'm'
(lp9971
I1
aI1
aI3
aI11
aI31
aI39
aI7
aI169
aI15
aI799
aI1585
aI2055
aI4683
aI13247
aI23743
aI50399
asS'd'
I4986
sa(dp9972
S'a'
I21087
sS's'
I16
sS'm'
(lp9973
I1
aI1
aI1
aI9
aI25
aI37
aI15
aI39
aI339
aI383
aI1153
aI1211
aI5745
aI15249
aI26021
aI39871
asS'd'
I4987
sa(dp9974
S'a'
I21088
sS's'
I16
sS'm'
(lp9975
I1
aI1
aI3
aI13
aI17
aI51
aI27
aI137
aI231
aI877
aI309
aI3633
aI2575
aI12259
aI2743
aI14781
asS'd'
I4988
sa(dp9976
S'a'
I21093
sS's'
I16
sS'm'
(lp9977
I1
aI1
aI5
aI7
aI5
aI33
aI95
aI19
aI37
aI829
aI1489
aI3525
aI3887
aI8277
aI21867
aI3581
asS'd'
I4989
sa(dp9978
S'a'
I21106
sS's'
I16
sS'm'
(lp9979
I1
aI1
aI1
aI15
aI11
aI33
aI99
aI213
aI365
aI549
aI1603
aI777
aI3787
aI12209
aI14999
aI50607
asS'd'
I4990
sa(dp9980
S'a'
I21108
sS's'
I16
sS'm'
(lp9981
I1
aI3
aI1
aI9
aI23
aI25
aI57
aI147
aI73
aI285
aI1229
aI1763
aI4579
aI14771
aI4003
aI38197
asS'd'
I4991
sa(dp9982
S'a'
I21118
sS's'
I16
sS'm'
(lp9983
I1
aI1
aI5
aI1
aI15
aI55
aI25
aI209
aI135
aI895
aI311
aI139
aI2445
aI6903
aI12129
aI27907
asS'd'
I4992
sa(dp9984
S'a'
I21122
sS's'
I16
sS'm'
(lp9985
I1
aI1
aI5
aI7
aI23
aI29
aI33
aI135
aI325
aI517
aI2021
aI1721
aI4235
aI2363
aI12905
aI18811
asS'd'
I4993
sa(dp9986
S'a'
I21131
sS's'
I16
sS'm'
(lp9987
I1
aI1
aI1
aI9
aI3
aI19
aI69
aI29
aI157
aI787
aI1969
aI3711
aI8179
aI5691
aI4059
aI25541
asS'd'
I4994
sa(dp9988
S'a'
I21139
sS's'
I16
sS'm'
(lp9989
I1
aI1
aI5
aI15
aI1
aI61
aI11
aI195
aI317
aI13
aI923
aI2149
aI4001
aI12843
aI27109
aI30625
asS'd'
I4995
sa(dp9990
S'a'
I21141
sS's'
I16
sS'm'
(lp9991
I1
aI3
aI1
aI7
aI3
aI13
aI45
aI187
aI445
aI859
aI53
aI3227
aI4381
aI8273
aI32699
aI48719
asS'd'
I4996
sa(dp9992
S'a'
I21146
sS's'
I16
sS'm'
(lp9993
I1
aI1
aI7
aI7
aI21
aI19
aI47
aI101
aI119
aI311
aI577
aI3309
aI4585
aI12109
aI15153
aI22915
asS'd'
I4997
sa(dp9994
S'a'
I21162
sS's'
I16
sS'm'
(lp9995
I1
aI3
aI5
aI15
aI21
aI39
aI15
aI211
aI349
aI237
aI1873
aI3333
aI7837
aI12811
aI14321
aI20227
asS'd'
I4998
sa(dp9996
S'a'
I21164
sS's'
I16
sS'm'
(lp9997
I1
aI1
aI5
aI5
aI19
aI47
aI15
aI239
aI23
aI763
aI537
aI1477
aI2231
aI10885
aI19487
aI47385
asS'd'
I4999
sa(dp9998
S'a'
I21184
sS's'
I16
sS'm'
(lp9999
I1
aI1
aI3
aI1
aI19
aI37
aI67
aI85
aI11
aI817
aI869
aI2249
aI4111
aI12411
aI10405
aI20055
asS'd'
I5000
sa(dp10000
S'a'
I21208
sS's'
I16
sS'm'
(lp10001
I1
aI1
aI3
aI3
aI1
aI41
aI85
aI137
aI91
aI369
aI1863
aI759
aI303
aI15859
aI8063
aI12811
asS'd'
I5001
sa(dp10002
S'a'
I21211
sS's'
I16
sS'm'
(lp10003
I1
aI3
aI1
aI11
aI23
aI1
aI11
aI219
aI201
aI573
aI1573
aI619
aI2959
aI6485
aI7483
aI46099
asS'd'
I5002
sa(dp10004
S'a'
I21213
sS's'
I16
sS'm'
(lp10005
I1
aI3
aI3
aI9
aI13
aI9
aI9
aI255
aI47
aI375
aI409
aI1435
aI1665
aI14967
aI3247
aI18193
asS'd'
I5003
sa(dp10006
S'a'
I21230
sS's'
I16
sS'm'
(lp10007
I1
aI1
aI1
aI5
aI9
aI61
aI121
aI173
aI51
aI415
aI1621
aI3947
aI1379
aI847
aI23187
aI39259
asS'd'
I5004
sa(dp10008
S'a'
I21270
sS's'
I16
sS'm'
(lp10009
I1
aI1
aI1
aI7
aI3
aI19
aI95
aI59
aI187
aI453
aI1533
aI445
aI2699
aI4817
aI25983
aI50925
asS'd'
I5005
sa(dp10010
S'a'
I21276
sS's'
I16
sS'm'
(lp10011
I1
aI3
aI5
aI13
aI25
aI25
aI33
aI5
aI497
aI1
aI535
aI1653
aI6541
aI10939
aI17721
aI43829
asS'd'
I5006
sa(dp10012
S'a'
I21285
sS's'
I16
sS'm'
(lp10013
I1
aI3
aI7
aI11
aI9
aI59
aI115
aI127
aI85
aI505
aI541
aI3243
aI5853
aI12673
aI25275
aI39577
asS'd'
I5007
sa(dp10014
S'a'
I21297
sS's'
I16
sS'm'
(lp10015
I1
aI3
aI7
aI1
aI17
aI25
aI83
aI127
aI225
aI295
aI1823
aI2051
aI847
aI4249
aI13763
aI5723
asS'd'
I5008
sa(dp10016
S'a'
I21304
sS's'
I16
sS'm'
(lp10017
I1
aI1
aI1
aI5
aI3
aI63
aI39
aI131
aI191
aI983
aI119
aI3287
aI3335
aI7969
aI31347
aI39439
asS'd'
I5009
sa(dp10018
S'a'
I21310
sS's'
I16
sS'm'
(lp10019
I1
aI3
aI7
aI9
aI19
aI31
aI19
aI91
aI35
aI677
aI1229
aI1371
aI6497
aI3315
aI15239
aI54235
asS'd'
I5010
sa(dp10020
S'a'
I21330
sS's'
I16
sS'm'
(lp10021
I1
aI1
aI1
aI15
aI3
aI49
aI113
aI199
aI135
aI35
aI709
aI385
aI7923
aI3711
aI18351
aI12711
asS'd'
I5011
sa(dp10022
S'a'
I21339
sS's'
I16
sS'm'
(lp10023
I1
aI1
aI3
aI15
aI31
aI13
aI41
aI1
aI183
aI95
aI1625
aI1675
aI7881
aI6607
aI4165
aI27151
asS'd'
I5012
sa(dp10024
S'a'
I21346
sS's'
I16
sS'm'
(lp10025
I1
aI3
aI3
aI15
aI21
aI57
aI81
aI49
aI5
aI297
aI131
aI883
aI1113
aI2497
aI32129
aI39139
asS'd'
I5013
sa(dp10026
S'a'
I21391
sS's'
I16
sS'm'
(lp10027
I1
aI3
aI5
aI7
aI29
aI47
aI101
aI173
aI299
aI879
aI143
aI3341
aI3929
aI6797
aI8753
aI47963
asS'd'
I5014
sa(dp10028
S'a'
I21427
sS's'
I16
sS'm'
(lp10029
I1
aI3
aI3
aI13
aI11
aI39
aI27
aI187
aI27
aI763
aI1515
aI1903
aI5897
aI10061
aI14595
aI12713
asS'd'
I5015
sa(dp10030
S'a'
I21451
sS's'
I16
sS'm'
(lp10031
I1
aI3
aI5
aI11
aI27
aI35
aI37
aI213
aI45
aI867
aI1591
aI3083
aI8123
aI5045
aI31703
aI61487
asS'd'
I5016
sa(dp10032
S'a'
I21465
sS's'
I16
sS'm'
(lp10033
I1
aI1
aI3
aI5
aI3
aI31
aI23
aI89
aI369
aI371
aI1165
aI3673
aI6821
aI333
aI10881
aI4153
asS'd'
I5017
sa(dp10034
S'a'
I21468
sS's'
I16
sS'm'
(lp10035
I1
aI1
aI7
aI13
aI1
aI33
aI49
aI195
aI223
aI197
aI1799
aI2427
aI6171
aI493
aI13503
aI23619
asS'd'
I5018
sa(dp10036
S'a'
I21471
sS's'
I16
sS'm'
(lp10037
I1
aI1
aI3
aI9
aI5
aI59
aI105
aI215
aI449
aI527
aI1661
aI2643
aI309
aI11239
aI11633
aI63459
asS'd'
I5019
sa(dp10038
S'a'
I21533
sS's'
I16
sS'm'
(lp10039
I1
aI1
aI3
aI11
aI13
aI11
aI15
aI99
aI409
aI807
aI1911
aI883
aI1323
aI9037
aI6401
aI545
asS'd'
I5020
sa(dp10040
S'a'
I21610
sS's'
I16
sS'm'
(lp10041
I1
aI1
aI5
aI7
aI7
aI7
aI1
aI167
aI353
aI923
aI1403
aI3109
aI4981
aI3877
aI25451
aI4667
asS'd'
I5021
sa(dp10042
S'a'
I21615
sS's'
I16
sS'm'
(lp10043
I1
aI1
aI5
aI11
aI11
aI25
aI121
aI153
aI111
aI785
aI1301
aI1497
aI6267
aI14919
aI24125
aI52029
asS'd'
I5022
sa(dp10044
S'a'
I21630
sS's'
I16
sS'm'
(lp10045
I1
aI3
aI3
aI5
aI29
aI55
aI63
aI177
aI305
aI41
aI111
aI1065
aI1127
aI113
aI2815
aI12979
asS'd'
I5023
sa(dp10046
S'a'
I21633
sS's'
I16
sS'm'
(lp10047
I1
aI3
aI5
aI7
aI23
aI39
aI17
aI179
aI267
aI917
aI511
aI3923
aI915
aI14173
aI10689
aI50749
asS'd'
I5024
sa(dp10048
S'a'
I21657
sS's'
I16
sS'm'
(lp10049
I1
aI1
aI5
aI3
aI9
aI45
aI15
aI157
aI495
aI625
aI407
aI2769
aI3267
aI7593
aI17957
aI54597
asS'd'
I5025
sa(dp10050
S'a'
I21658
sS's'
I16
sS'm'
(lp10051
I1
aI3
aI3
aI11
aI21
aI13
aI5
aI207
aI107
aI965
aI1803
aI1541
aI3487
aI3483
aI109
aI26923
asS'd'
I5026
sa(dp10052
S'a'
I21669
sS's'
I16
sS'm'
(lp10053
I1
aI1
aI5
aI11
aI25
aI49
aI99
aI135
aI109
aI371
aI1307
aI1815
aI1095
aI2329
aI27101
aI52269
asS'd'
I5027
sa(dp10054
S'a'
I21670
sS's'
I16
sS'm'
(lp10055
I1
aI3
aI5
aI5
aI13
aI9
aI109
aI79
aI151
aI47
aI311
aI3873
aI3645
aI3773
aI1083
aI31599
asS'd'
I5028
sa(dp10056
S'a'
I21673
sS's'
I16
sS'm'
(lp10057
I1
aI3
aI5
aI15
aI1
aI9
aI87
aI21
aI145
aI583
aI159
aI2435
aI587
aI10123
aI24803
aI19993
asS'd'
I5029
sa(dp10058
S'a'
I21701
sS's'
I16
sS'm'
(lp10059
I1
aI3
aI1
aI1
aI23
aI11
aI5
aI45
aI373
aI1011
aI1353
aI277
aI7051
aI3845
aI12391
aI25313
asS'd'
I5030
sa(dp10060
S'a'
I21719
sS's'
I16
sS'm'
(lp10061
I1
aI1
aI1
aI9
aI13
aI13
aI109
aI251
aI97
aI483
aI1723
aI2555
aI813
aI9345
aI11351
aI44705
asS'd'
I5031
sa(dp10062
S'a'
I21720
sS's'
I16
sS'm'
(lp10063
I1
aI3
aI5
aI7
aI21
aI49
aI63
aI13
aI119
aI813
aI1559
aI983
aI499
aI15159
aI24163
aI59903
asS'd'
I5032
sa(dp10064
S'a'
I21747
sS's'
I16
sS'm'
(lp10065
I1
aI1
aI3
aI5
aI27
aI33
aI27
aI165
aI207
aI693
aI1401
aI1357
aI7637
aI337
aI10163
aI43273
asS'd'
I5033
sa(dp10066
S'a'
I21819
sS's'
I16
sS'm'
(lp10067
I1
aI1
aI5
aI13
aI29
aI7
aI71
aI187
aI1
aI655
aI1829
aI2645
aI6969
aI5435
aI28415
aI33199
asS'd'
I5034
sa(dp10068
S'a'
I21839
sS's'
I16
sS'm'
(lp10069
I1
aI1
aI1
aI13
aI17
aI21
aI13
aI141
aI41
aI267
aI1165
aI1893
aI3455
aI10737
aI16693
aI33065
asS'd'
I5035
sa(dp10070
S'a'
I21854
sS's'
I16
sS'm'
(lp10071
I1
aI1
aI5
aI1
aI7
aI27
aI7
aI67
aI107
aI11
aI763
aI2529
aI3023
aI15745
aI17023
aI51911
asS'd'
I5036
sa(dp10072
S'a'
I21857
sS's'
I16
sS'm'
(lp10073
I1
aI3
aI3
aI3
aI7
aI57
aI123
aI249
aI309
aI511
aI1655
aI1379
aI725
aI7325
aI20261
aI65039
asS'd'
I5037
sa(dp10074
S'a'
I21864
sS's'
I16
sS'm'
(lp10075
I1
aI1
aI5
aI11
aI3
aI27
aI23
aI27
aI285
aI771
aI2017
aI1727
aI4847
aI2665
aI30655
aI47625
asS'd'
I5038
sa(dp10076
S'a'
I21882
sS's'
I16
sS'm'
(lp10077
I1
aI3
aI7
aI7
aI17
aI3
aI93
aI133
aI427
aI1021
aI1135
aI341
aI6711
aI11713
aI24157
aI1561
asS'd'
I5039
sa(dp10078
S'a'
I21900
sS's'
I16
sS'm'
(lp10079
I1
aI1
aI3
aI7
aI15
aI55
aI11
aI247
aI65
aI115
aI1967
aI535
aI841
aI15131
aI28381
aI31337
asS'd'
I5040
sa(dp10080
S'a'
I21903
sS's'
I16
sS'm'
(lp10081
I1
aI3
aI1
aI7
aI9
aI45
aI73
aI151
aI127
aI125
aI767
aI2003
aI6893
aI3875
aI451
aI30275
asS'd'
I5041
sa(dp10082
S'a'
I21928
sS's'
I16
sS'm'
(lp10083
I1
aI1
aI7
aI3
aI5
aI55
aI127
aI123
aI163
aI763
aI1165
aI1637
aI6267
aI7215
aI23403
aI20961
asS'd'
I5042
sa(dp10084
S'a'
I21931
sS's'
I16
sS'm'
(lp10085
I1
aI1
aI1
aI13
aI1
aI21
aI65
aI141
aI369
aI413
aI1675
aI27
aI7357
aI6929
aI18083
aI829
asS'd'
I5043
sa(dp10086
S'a'
I21946
sS's'
I16
sS'm'
(lp10087
I1
aI3
aI5
aI13
aI1
aI17
aI97
aI107
aI249
aI931
aI47
aI3537
aI2245
aI9827
aI13673
aI23217
asS'd'
I5044
sa(dp10088
S'a'
I21971
sS's'
I16
sS'm'
(lp10089
I1
aI1
aI1
aI11
aI13
aI19
aI43
aI31
aI51
aI727
aI1001
aI927
aI771
aI11853
aI5761
aI60537
asS'd'
I5045
sa(dp10090
S'a'
I21974
sS's'
I16
sS'm'
(lp10091
I1
aI3
aI1
aI7
aI23
aI27
aI115
aI5
aI201
aI431
aI1021
aI585
aI6069
aI12511
aI6129
aI2105
asS'd'
I5046
sa(dp10092
S'a'
I21978
sS's'
I16
sS'm'
(lp10093
I1
aI1
aI3
aI11
aI3
aI25
aI75
aI129
aI389
aI131
aI223
aI2263
aI5377
aI5561
aI15633
aI39527
asS'd'
I5047
sa(dp10094
S'a'
I21993
sS's'
I16
sS'm'
(lp10095
I1
aI3
aI3
aI1
aI27
aI43
aI101
aI55
aI319
aI549
aI1971
aI2255
aI353
aI93
aI25661
aI59077
asS'd'
I5048
sa(dp10096
S'a'
I21994
sS's'
I16
sS'm'
(lp10097
I1
aI1
aI5
aI11
aI29
aI57
aI27
aI135
aI341
aI913
aI1637
aI1781
aI457
aI11293
aI1013
aI53863
asS'd'
I5049
sa(dp10098
S'a'
I22030
sS's'
I16
sS'm'
(lp10099
I1
aI1
aI1
aI11
aI3
aI51
aI79
aI251
aI443
aI651
aI393
aI3635
aI7397
aI5443
aI4225
aI991
asS'd'
I5050
sa(dp10100
S'a'
I22035
sS's'
I16
sS'm'
(lp10101
I1
aI3
aI7
aI13
aI31
aI9
aI3
aI109
aI427
aI735
aI891
aI2789
aI2169
aI6459
aI355
aI43177
asS'd'
I5051
sa(dp10102
S'a'
I22063
sS's'
I16
sS'm'
(lp10103
I1
aI3
aI3
aI3
aI13
aI7
aI93
aI195
aI293
aI37
aI75
aI2467
aI933
aI8017
aI9925
aI61397
asS'd'
I5052
sa(dp10104
S'a'
I22068
sS's'
I16
sS'm'
(lp10105
I1
aI1
aI5
aI7
aI25
aI15
aI69
aI199
aI161
aI769
aI387
aI1491
aI4553
aI4173
aI25631
aI37089
asS'd'
I5053
sa(dp10106
S'a'
I22086
sS's'
I16
sS'm'
(lp10107
I1
aI3
aI1
aI3
aI7
aI59
aI53
aI93
aI103
aI413
aI531
aI887
aI6149
aI2901
aI22611
aI27135
asS'd'
I5054
sa(dp10108
S'a'
I22104
sS's'
I16
sS'm'
(lp10109
I1
aI1
aI1
aI13
aI31
aI39
aI71
aI215
aI385
aI821
aI1603
aI3043
aI4967
aI10953
aI11543
aI64433
asS'd'
I5055
sa(dp10110
S'a'
I22119
sS's'
I16
sS'm'
(lp10111
I1
aI3
aI1
aI7
aI7
aI63
aI5
aI143
aI1
aI339
aI1165
aI3809
aI4257
aI12879
aI21581
aI21307
asS'd'
I5056
sa(dp10112
S'a'
I22153
sS's'
I16
sS'm'
(lp10113
I1
aI1
aI1
aI15
aI1
aI35
aI67
aI227
aI277
aI879
aI513
aI3423
aI6153
aI11573
aI12809
aI34335
asS'd'
I5057
sa(dp10114
S'a'
I22168
sS's'
I16
sS'm'
(lp10115
I1
aI3
aI7
aI9
aI9
aI39
aI47
aI17
aI101
aI179
aI631
aI1307
aI481
aI871
aI16807
aI39811
asS'd'
I5058
sa(dp10116
S'a'
I22183
sS's'
I16
sS'm'
(lp10117
I1
aI3
aI1
aI1
aI13
aI25
aI53
aI179
aI285
aI267
aI407
aI3781
aI3267
aI3545
aI525
aI15733
asS'd'
I5059
sa(dp10118
S'a'
I22189
sS's'
I16
sS'm'
(lp10119
I1
aI1
aI1
aI13
aI11
aI35
aI45
aI181
aI319
aI767
aI283
aI3021
aI5405
aI403
aI3587
aI7291
asS'd'
I5060
sa(dp10120
S'a'
I22204
sS's'
I16
sS'm'
(lp10121
I1
aI1
aI7
aI3
aI5
aI9
aI67
aI129
aI101
aI117
aI267
aI1925
aI1209
aI13095
aI7123
aI62941
asS'd'
I5061
sa(dp10122
S'a'
I22230
sS's'
I16
sS'm'
(lp10123
I1
aI1
aI7
aI3
aI5
aI63
aI109
aI199
aI95
aI421
aI193
aI3519
aI6551
aI955
aI1679
aI16627
asS'd'
I5062
sa(dp10124
S'a'
I22240
sS's'
I16
sS'm'
(lp10125
I1
aI1
aI5
aI13
aI1
aI47
aI71
aI157
aI343
aI653
aI981
aI1335
aI3737
aI7185
aI28861
aI22883
asS'd'
I5063
sa(dp10126
S'a'
I22246
sS's'
I16
sS'm'
(lp10127
I1
aI1
aI3
aI15
aI7
aI63
aI7
aI81
aI481
aI5
aI1159
aI1361
aI4167
aI2575
aI7437
aI16917
asS'd'
I5064
sa(dp10128
S'a'
I22269
sS's'
I16
sS'm'
(lp10129
I1
aI3
aI7
aI1
aI27
aI17
aI61
aI193
aI317
aI841
aI1149
aI955
aI5161
aI4275
aI1955
aI15665
asS'd'
I5065
sa(dp10130
S'a'
I22282
sS's'
I16
sS'm'
(lp10131
I1
aI1
aI1
aI7
aI19
aI63
aI77
aI57
aI367
aI237
aI579
aI3701
aI5763
aI4951
aI24151
aI45215
asS'd'
I5066
sa(dp10132
S'a'
I22302
sS's'
I16
sS'm'
(lp10133
I1
aI1
aI5
aI11
aI29
aI7
aI119
aI155
aI431
aI999
aI757
aI2433
aI5811
aI3709
aI29573
aI23639
asS'd'
I5067
sa(dp10134
S'a'
I22330
sS's'
I16
sS'm'
(lp10135
I1
aI3
aI3
aI3
aI15
aI35
aI125
aI13
aI275
aI507
aI1719
aI1537
aI2349
aI12909
aI8107
aI9845
asS'd'
I5068
sa(dp10136
S'a'
I22347
sS's'
I16
sS'm'
(lp10137
I1
aI3
aI1
aI13
aI27
aI27
aI11
aI69
aI15
aI1017
aI207
aI625
aI809
aI2921
aI8939
aI30293
asS'd'
I5069
sa(dp10138
S'a'
I22349
sS's'
I16
sS'm'
(lp10139
I1
aI1
aI7
aI11
aI31
aI51
aI113
aI193
aI69
aI845
aI73
aI919
aI3523
aI3987
aI23665
aI36527
asS'd'
I5070
sa(dp10140
S'a'
I22383
sS's'
I16
sS'm'
(lp10141
I1
aI3
aI7
aI11
aI21
aI31
aI103
aI29
aI5
aI81
aI1081
aI3847
aI4697
aI8857
aI30769
aI40053
asS'd'
I5071
sa(dp10142
S'a'
I22386
sS's'
I16
sS'm'
(lp10143
I1
aI1
aI1
aI1
aI5
aI11
aI5
aI169
aI13
aI899
aI769
aI3823
aI5405
aI5991
aI3821
aI27767
asS'd'
I5072
sa(dp10144
S'a'
I22432
sS's'
I16
sS'm'
(lp10145
I1
aI1
aI3
aI15
aI1
aI35
aI9
aI83
aI23
aI701
aI1807
aI1681
aI4009
aI127
aI31751
aI38059
asS'd'
I5073
sa(dp10146
S'a'
I22450
sS's'
I16
sS'm'
(lp10147
I1
aI3
aI3
aI7
aI9
aI61
aI73
aI111
aI193
aI607
aI2001
aI413
aI3751
aI16337
aI16597
aI23959
asS'd'
I5074
sa(dp10148
S'a'
I22473
sS's'
I16
sS'm'
(lp10149
I1
aI1
aI7
aI7
aI21
aI29
aI53
aI253
aI187
aI507
aI1191
aI3521
aI463
aI2167
aI23785
aI19867
asS'd'
I5075
sa(dp10150
S'a'
I22487
sS's'
I16
sS'm'
(lp10151
I1
aI3
aI5
aI3
aI19
aI43
aI101
aI93
aI257
aI61
aI1589
aI3883
aI1975
aI7283
aI5253
aI23257
asS'd'
I5076
sa(dp10152
S'a'
I22527
sS's'
I16
sS'm'
(lp10153
I1
aI3
aI1
aI9
aI1
aI63
aI25
aI101
aI377
aI571
aI1701
aI3385
aI243
aI12051
aI32619
aI10459
asS'd'
I5077
sa(dp10154
S'a'
I22537
sS's'
I16
sS'm'
(lp10155
I1
aI1
aI1
aI5
aI17
aI11
aI37
aI197
aI205
aI879
aI625
aI959
aI7389
aI7857
aI20615
aI20405
asS'd'
I5078
sa(dp10156
S'a'
I22557
sS's'
I16
sS'm'
(lp10157
I1
aI3
aI5
aI5
aI27
aI41
aI9
aI109
aI197
aI623
aI1045
aI63
aI7977
aI11355
aI28613
aI5131
asS'd'
I5079
sa(dp10158
S'a'
I22561
sS's'
I16
sS'm'
(lp10159
I1
aI1
aI5
aI11
aI5
aI29
aI27
aI85
aI131
aI247
aI433
aI3981
aI2677
aI15415
aI869
aI6045
asS'd'
I5080
sa(dp10160
S'a'
I22568
sS's'
I16
sS'm'
(lp10161
I1
aI3
aI1
aI13
aI9
aI49
aI25
aI79
aI135
aI719
aI93
aI631
aI2149
aI5929
aI29833
aI38673
asS'd'
I5081
sa(dp10162
S'a'
I22573
sS's'
I16
sS'm'
(lp10163
I1
aI3
aI7
aI11
aI15
aI13
aI37
aI233
aI227
aI205
aI1579
aI65
aI1429
aI13499
aI26355
aI63821
asS'd'
I5082
sa(dp10164
S'a'
I22591
sS's'
I16
sS'm'
(lp10165
I1
aI1
aI5
aI11
aI21
aI19
aI7
aI183
aI409
aI275
aI899
aI3665
aI6207
aI849
aI24339
aI1617
asS'd'
I5083
sa(dp10166
S'a'
I22593
sS's'
I16
sS'm'
(lp10167
I1
aI3
aI1
aI3
aI21
aI61
aI23
aI125
aI463
aI489
aI1265
aI2975
aI3907
aI11881
aI7533
aI43331
asS'd'
I5084
sa(dp10168
S'a'
I22605
sS's'
I16
sS'm'
(lp10169
I1
aI3
aI1
aI15
aI15
aI51
aI83
aI31
aI175
aI47
aI1791
aI3651
aI6735
aI5013
aI723
aI24141
asS'd'
I5085
sa(dp10170
S'a'
I22620
sS's'
I16
sS'm'
(lp10171
I1
aI3
aI1
aI9
aI17
aI41
aI57
aI43
aI469
aI911
aI1251
aI2105
aI3133
aI3437
aI10097
aI26687
asS'd'
I5086
sa(dp10172
S'a'
I22627
sS's'
I16
sS'm'
(lp10173
I1
aI1
aI3
aI3
aI9
aI9
aI125
aI201
aI141
aI943
aI1509
aI1239
aI6575
aI8707
aI363
aI23309
asS'd'
I5087
sa(dp10174
S'a'
I22636
sS's'
I16
sS'm'
(lp10175
I1
aI1
aI5
aI3
aI19
aI37
aI43
aI141
aI413
aI149
aI1449
aI1003
aI4473
aI12395
aI4101
aI61201
asS'd'
I5088
sa(dp10176
S'a'
I22647
sS's'
I16
sS'm'
(lp10177
I1
aI1
aI5
aI11
aI17
aI37
aI41
aI33
aI57
aI627
aI325
aI1895
aI1773
aI1339
aI24859
aI587
asS'd'
I5089
sa(dp10178
S'a'
I22697
sS's'
I16
sS'm'
(lp10179
I1
aI1
aI1
aI3
aI5
aI49
aI127
aI109
aI361
aI853
aI1437
aI3451
aI4031
aI5379
aI27463
aI47425
asS'd'
I5090
sa(dp10180
S'a'
I22715
sS's'
I16
sS'm'
(lp10181
I1
aI3
aI5
aI7
aI9
aI57
aI71
aI219
aI347
aI791
aI797
aI73
aI6241
aI12717
aI24429
aI40977
asS'd'
I5091
sa(dp10182
S'a'
I22725
sS's'
I16
sS'm'
(lp10183
I1
aI1
aI5
aI9
aI27
aI43
aI43
aI227
aI433
aI413
aI1109
aI2589
aI4535
aI8947
aI8121
aI43479
asS'd'
I5092
sa(dp10184
S'a'
I22732
sS's'
I16
sS'm'
(lp10185
I1
aI3
aI7
aI1
aI9
aI21
aI81
aI23
aI157
aI313
aI197
aI2845
aI8059
aI15957
aI28657
aI28093
asS'd'
I5093
sa(dp10186
S'a'
I22749
sS's'
I16
sS'm'
(lp10187
I1
aI3
aI1
aI11
aI15
aI17
aI115
aI27
aI421
aI743
aI1885
aI2089
aI5011
aI7137
aI7395
aI36853
asS'd'
I5094
sa(dp10188
S'a'
I22766
sS's'
I16
sS'm'
(lp10189
I1
aI1
aI5
aI15
aI5
aI53
aI69
aI255
aI63
aI29
aI1011
aI3201
aI1467
aI15441
aI26255
aI62895
asS'd'
I5095
sa(dp10190
S'a'
I22768
sS's'
I16
sS'm'
(lp10191
I1
aI3
aI1
aI11
aI3
aI47
aI35
aI195
aI149
aI849
aI1317
aI439
aI3539
aI845
aI15295
aI42183
asS'd'
I5096
sa(dp10192
S'a'
I22771
sS's'
I16
sS'm'
(lp10193
I1
aI1
aI5
aI15
aI15
aI37
aI67
aI105
aI495
aI985
aI1777
aI3137
aI8039
aI11795
aI29771
aI35045
asS'd'
I5097
sa(dp10194
S'a'
I22788
sS's'
I16
sS'm'
(lp10195
I1
aI1
aI3
aI1
aI25
aI17
aI67
aI227
aI229
aI419
aI1319
aI3325
aI1293
aI8585
aI28425
aI34013
asS'd'
I5098
sa(dp10196
S'a'
I22797
sS's'
I16
sS'm'
(lp10197
I1
aI1
aI5
aI1
aI27
aI51
aI71
aI197
aI375
aI407
aI259
aI4005
aI3309
aI5475
aI13421
aI60065
asS'd'
I5099
sa(dp10198
S'a'
I22822
sS's'
I16
sS'm'
(lp10199
I1
aI3
aI1
aI5
aI11
aI17
aI89
aI45
aI311
aI425
aI1629
aI773
aI7267
aI8699
aI27547
aI37081
asS'd'
I5100
sa(dp10200
S'a'
I22828
sS's'
I16
sS'm'
(lp10201
I1
aI3
aI1
aI7
aI9
aI25
aI101
aI105
aI489
aI217
aI103
aI1959
aI4049
aI5793
aI31201
aI11947
asS'd'
I5101
sa(dp10202
S'a'
I22851
sS's'
I16
sS'm'
(lp10203
I1
aI1
aI5
aI5
aI19
aI3
aI63
aI55
aI431
aI49
aI273
aI3253
aI5357
aI13801
aI9735
aI21883
asS'd'
I5102
sa(dp10204
S'a'
I22888
sS's'
I16
sS'm'
(lp10205
I1
aI1
aI1
aI11
aI13
aI3
aI75
aI201
aI477
aI201
aI1875
aI657
aI6217
aI8651
aI2207
aI16421
asS'd'
I5103
sa(dp10206
S'a'
I22893
sS's'
I16
sS'm'
(lp10207
I1
aI1
aI5
aI13
aI5
aI31
aI75
aI113
aI25
aI171
aI1147
aI3089
aI7095
aI4405
aI26155
aI42323
asS'd'
I5104
sa(dp10208
S'a'
I22901
sS's'
I16
sS'm'
(lp10209
I1
aI3
aI5
aI5
aI5
aI49
aI99
aI171
aI445
aI1023
aI1793
aI3159
aI5809
aI12509
aI31723
aI60411
asS'd'
I5105
sa(dp10210
S'a'
I22902
sS's'
I16
sS'm'
(lp10211
I1
aI3
aI7
aI3
aI23
aI51
aI111
aI27
aI103
aI159
aI433
aI293
aI1741
aI3599
aI4067
aI40667
asS'd'
I5106
sa(dp10212
S'a'
I22921
sS's'
I16
sS'm'
(lp10213
I1
aI3
aI3
aI13
aI11
aI9
aI11
aI21
aI453
aI35
aI1649
aI1261
aI3539
aI14081
aI5581
aI57105
asS'd'
I5107
sa(dp10214
S'a'
I22929
sS's'
I16
sS'm'
(lp10215
I1
aI3
aI3
aI13
aI7
aI9
aI113
aI87
aI391
aI647
aI223
aI1345
aI4481
aI9855
aI20129
aI10807
asS'd'
I5108
sa(dp10216
S'a'
I22946
sS's'
I16
sS'm'
(lp10217
I1
aI3
aI7
aI15
aI3
aI61
aI15
aI117
aI97
aI495
aI985
aI819
aI181
aI1363
aI13111
aI35857
asS'd'
I5109
sa(dp10218
S'a'
I22948
sS's'
I16
sS'm'
(lp10219
I1
aI3
aI1
aI9
aI3
aI27
aI125
aI151
aI217
aI961
aI707
aI2647
aI5307
aI621
aI12581
aI13941
asS'd'
I5110
sa(dp10220
S'a'
I22951
sS's'
I16
sS'm'
(lp10221
I1
aI3
aI1
aI11
aI17
aI37
aI35
aI211
aI179
aI29
aI627
aI3623
aI6429
aI16237
aI24699
aI14385
asS'd'
I5111
sa(dp10222
S'a'
I22958
sS's'
I16
sS'm'
(lp10223
I1
aI3
aI3
aI9
aI3
aI57
aI35
aI3
aI85
aI1017
aI1739
aI2241
aI7297
aI15637
aI27085
aI41237
asS'd'
I5112
sa(dp10224
S'a'
I22975
sS's'
I16
sS'm'
(lp10225
I1
aI1
aI3
aI7
aI7
aI13
aI5
aI85
aI505
aI51
aI409
aI867
aI677
aI12451
aI13633
aI47883
asS'd'
I5113
sa(dp10226
S'a'
I22983
sS's'
I16
sS'm'
(lp10227
I1
aI3
aI5
aI13
aI5
aI51
aI37
aI79
aI237
aI133
aI1331
aI3263
aI349
aI4971
aI16067
aI62485
asS'd'
I5114
sa(dp10228
S'a'
I22990
sS's'
I16
sS'm'
(lp10229
I1
aI1
aI7
aI11
aI29
aI41
aI101
aI219
aI391
aI1023
aI1681
aI3163
aI4071
aI14665
aI11041
aI14523
asS'd'
I5115
sa(dp10230
S'a'
I23032
sS's'
I16
sS'm'
(lp10231
I1
aI1
aI3
aI3
aI13
aI55
aI37
aI119
aI471
aI665
aI1315
aI3071
aI5993
aI12005
aI13549
aI63425
asS'd'
I5116
sa(dp10232
S'a'
I23047
sS's'
I16
sS'm'
(lp10233
I1
aI3
aI5
aI7
aI5
aI25
aI59
aI71
aI77
aI841
aI91
aI1841
aI6997
aI1139
aI11843
aI52209
asS'd'
I5117
sa(dp10234
S'a'
I23053
sS's'
I16
sS'm'
(lp10235
I1
aI3
aI7
aI15
aI17
aI25
aI85
aI173
aI373
aI459
aI1713
aI1055
aI5337
aI9921
aI15213
aI44235
asS'd'
I5118
sa(dp10236
S'a'
I23054
sS's'
I16
sS'm'
(lp10237
I1
aI1
aI1
aI15
aI7
aI11
aI89
aI237
aI131
aI565
aI745
aI457
aI4447
aI5581
aI11053
aI43819
asS'd'
I5119
sa(dp10238
S'a'
I23082
sS's'
I16
sS'm'
(lp10239
I1
aI3
aI5
aI1
aI29
aI21
aI11
aI7
aI125
aI851
aI2023
aI3321
aI1885
aI67
aI8809
aI43291
asS'd'
I5120
sa(dp10240
S'a'
I23095
sS's'
I16
sS'm'
(lp10241
I1
aI3
aI5
aI11
aI11
aI43
aI41
aI97
aI353
aI813
aI85
aI2453
aI769
aI11709
aI4697
aI2849
asS'd'
I5121
sa(dp10242
S'a'
I23116
sS's'
I16
sS'm'
(lp10243
I1
aI1
aI5
aI5
aI21
aI29
aI87
aI179
aI157
aI981
aI129
aI2139
aI6841
aI5479
aI27111
aI20749
asS'd'
I5122
sa(dp10244
S'a'
I23197
sS's'
I16
sS'm'
(lp10245
I1
aI1
aI3
aI9
aI31
aI59
aI61
aI15
aI423
aI33
aI467
aI1817
aI6535
aI7341
aI31061
aI20937
asS'd'
I5123
sa(dp10246
S'a'
I23221
sS's'
I16
sS'm'
(lp10247
I1
aI1
aI7
aI3
aI1
aI21
aI127
aI135
aI321
aI699
aI727
aI3079
aI753
aI3971
aI5611
aI28345
asS'd'
I5124
sa(dp10248
S'a'
I23257
sS's'
I16
sS'm'
(lp10249
I1
aI3
aI7
aI11
aI27
aI3
aI39
aI63
aI389
aI433
aI43
aI1443
aI6241
aI10769
aI20485
aI58823
asS'd'
I5125
sa(dp10250
S'a'
I23260
sS's'
I16
sS'm'
(lp10251
I1
aI1
aI1
aI11
aI3
aI13
aI5
aI57
aI503
aI707
aI677
aI3355
aI6691
aI8841
aI20041
aI11867
asS'd'
I5126
sa(dp10252
S'a'
I23263
sS's'
I16
sS'm'
(lp10253
I1
aI1
aI3
aI11
aI31
aI39
aI107
aI221
aI81
aI125
aI1439
aI2429
aI2109
aI3931
aI31007
aI10915
asS'd'
I5127
sa(dp10254
S'a'
I23267
sS's'
I16
sS'm'
(lp10255
I1
aI3
aI3
aI9
aI17
aI53
aI13
aI121
aI127
aI15
aI1825
aI1909
aI5951
aI13503
aI31565
aI56163
asS'd'
I5128
sa(dp10256
S'a'
I23282
sS's'
I16
sS'm'
(lp10257
I1
aI1
aI1
aI1
aI19
aI55
aI3
aI153
aI385
aI277
aI1297
aI3655
aI6027
aI3057
aI11341
aI46989
asS'd'
I5129
sa(dp10258
S'a'
I23284
sS's'
I16
sS'm'
(lp10259
I1
aI1
aI5
aI9
aI3
aI55
aI37
aI223
aI353
aI141
aI1917
aI3827
aI2255
aI7617
aI10971
aI10641
asS'd'
I5130
sa(dp10260
S'a'
I23314
sS's'
I16
sS'm'
(lp10261
I1
aI3
aI7
aI9
aI29
aI41
aI71
aI19
aI137
aI243
aI2047
aI395
aI6981
aI15887
aI9479
aI60199
asS'd'
I5131
sa(dp10262
S'a'
I23326
sS's'
I16
sS'm'
(lp10263
I1
aI1
aI1
aI9
aI17
aI43
aI51
aI191
aI405
aI727
aI485
aI987
aI1855
aI15801
aI22529
aI10165
asS'd'
I5132
sa(dp10264
S'a'
I23335
sS's'
I16
sS'm'
(lp10265
I1
aI3
aI1
aI7
aI27
aI31
aI69
aI255
aI153
aI793
aI1353
aI1981
aI83
aI11387
aI6747
aI23379
asS'd'
I5133
sa(dp10266
S'a'
I23368
sS's'
I16
sS'm'
(lp10267
I1
aI1
aI5
aI5
aI31
aI49
aI83
aI157
aI63
aI647
aI1367
aI3995
aI5899
aI8429
aI18317
aI3471
asS'd'
I5134
sa(dp10268
S'a'
I23373
sS's'
I16
sS'm'
(lp10269
I1
aI3
aI5
aI13
aI19
aI15
aI99
aI13
aI143
aI887
aI529
aI2855
aI573
aI9799
aI13585
aI59263
asS'd'
I5135
sa(dp10270
S'a'
I23401
sS's'
I16
sS'm'
(lp10271
I1
aI3
aI5
aI3
aI13
aI47
aI103
aI87
aI11
aI381
aI397
aI899
aI71
aI15539
aI13005
aI38297
asS'd'
I5136
sa(dp10272
S'a'
I23415
sS's'
I16
sS'm'
(lp10273
I1
aI1
aI1
aI3
aI1
aI53
aI45
aI141
aI1
aI941
aI261
aI3291
aI5177
aI9843
aI6309
aI62799
asS'd'
I5137
sa(dp10274
S'a'
I23432
sS's'
I16
sS'm'
(lp10275
I1
aI1
aI5
aI9
aI29
aI31
aI107
aI57
aI135
aI229
aI1147
aI247
aI265
aI12757
aI21365
aI7219
asS'd'
I5138
sa(dp10276
S'a'
I23476
sS's'
I16
sS'm'
(lp10277
I1
aI1
aI1
aI3
aI1
aI49
aI55
aI247
aI495
aI449
aI141
aI1349
aI7393
aI2589
aI23587
aI1097
asS'd'
I5139
sa(dp10278
S'a'
I23479
sS's'
I16
sS'm'
(lp10279
I1
aI3
aI5
aI1
aI9
aI11
aI73
aI153
aI89
aI575
aI1805
aI137
aI435
aI3687
aI32169
aI24275
asS'd'
I5140
sa(dp10280
S'a'
I23497
sS's'
I16
sS'm'
(lp10281
I1
aI1
aI7
aI15
aI25
aI61
aI51
aI21
aI109
aI139
aI611
aI3907
aI6721
aI5081
aI6665
aI51463
asS'd'
I5141
sa(dp10282
S'a'
I23505
sS's'
I16
sS'm'
(lp10283
I1
aI1
aI1
aI9
aI13
aI23
aI59
aI203
aI33
aI1013
aI1533
aI291
aI6171
aI15463
aI8581
aI9497
asS'd'
I5142
sa(dp10284
S'a'
I23508
sS's'
I16
sS'm'
(lp10285
I1
aI3
aI3
aI9
aI7
aI25
aI51
aI189
aI49
aI265
aI1163
aI1141
aI2467
aI7839
aI7083
aI65527
asS'd'
I5143
sa(dp10286
S'a'
I23531
sS's'
I16
sS'm'
(lp10287
I1
aI1
aI3
aI9
aI19
aI33
aI77
aI9
aI181
aI919
aI623
aI1521
aI7853
aI2199
aI24115
aI60607
asS'd'
I5144
sa(dp10288
S'a'
I23536
sS's'
I16
sS'm'
(lp10289
I1
aI1
aI3
aI11
aI9
aI11
aI27
aI57
aI355
aI313
aI151
aI3391
aI4869
aI12541
aI30031
aI29455
asS'd'
I5145
sa(dp10290
S'a'
I23542
sS's'
I16
sS'm'
(lp10291
I1
aI3
aI5
aI9
aI17
aI13
aI91
aI123
aI235
aI841
aI1113
aI1451
aI501
aI3863
aI32483
aI9445
asS'd'
I5146
sa(dp10292
S'a'
I23565
sS's'
I16
sS'm'
(lp10293
I1
aI3
aI7
aI3
aI13
aI25
aI87
aI243
aI449
aI293
aI1279
aI3571
aI2693
aI13459
aI5895
aI38127
asS'd'
I5147
sa(dp10294
S'a'
I23573
sS's'
I16
sS'm'
(lp10295
I1
aI1
aI3
aI15
aI27
aI61
aI7
aI57
aI255
aI971
aI131
aI2585
aI2187
aI7191
aI17779
aI34587
asS'd'
I5148
sa(dp10296
S'a'
I23590
sS's'
I16
sS'm'
(lp10297
I1
aI3
aI3
aI7
aI23
aI29
aI55
aI41
aI463
aI873
aI1781
aI2851
aI4731
aI9819
aI25587
aI32199
asS'd'
I5149
sa(dp10298
S'a'
I23593
sS's'
I16
sS'm'
(lp10299
I1
aI1
aI1
aI9
aI29
aI39
aI25
aI143
aI171
aI141
aI223
aI467
aI4417
aI9575
aI23159
aI33819
asS'd'
I5150
sa(dp10300
S'a'
I23604
sS's'
I16
sS'm'
(lp10301
I1
aI1
aI5
aI13
aI19
aI61
aI19
aI75
aI25
aI361
aI585
aI1627
aI2231
aI1831
aI24885
aI37917
asS'd'
I5151
sa(dp10302
S'a'
I23621
sS's'
I16
sS'm'
(lp10303
I1
aI3
aI7
aI7
aI23
aI19
aI59
aI55
aI323
aI55
aI143
aI131
aI27
aI15363
aI26423
aI43963
asS'd'
I5152
sa(dp10304
S'a'
I23622
sS's'
I16
sS'm'
(lp10305
I1
aI1
aI5
aI5
aI25
aI9
aI33
aI17
aI427
aI481
aI315
aI3999
aI4757
aI4545
aI7695
aI56733
asS'd'
I5153
sa(dp10306
S'a'
I23636
sS's'
I16
sS'm'
(lp10307
I1
aI3
aI5
aI13
aI5
aI59
aI45
aI117
aI115
aI263
aI1441
aI3307
aI1085
aI3875
aI25869
aI19725
asS'd'
I5154
sa(dp10308
S'a'
I23662
sS's'
I16
sS'm'
(lp10309
I1
aI3
aI3
aI15
aI13
aI39
aI31
aI243
aI293
aI345
aI343
aI1911
aI6123
aI12577
aI32081
aI59993
asS'd'
I5155
sa(dp10310
S'a'
I23667
sS's'
I16
sS'm'
(lp10311
I1
aI3
aI5
aI13
aI17
aI37
aI105
aI201
aI205
aI929
aI435
aI1467
aI8063
aI6963
aI13709
aI53275
asS'd'
I5156
sa(dp10312
S'a'
I23703
sS's'
I16
sS'm'
(lp10313
I1
aI3
aI7
aI15
aI31
aI3
aI65
aI221
aI191
aI413
aI427
aI2579
aI377
aI2793
aI26943
aI61299
asS'd'
I5157
sa(dp10314
S'a'
I23725
sS's'
I16
sS'm'
(lp10315
I1
aI3
aI5
aI3
aI11
aI61
aI75
aI107
aI53
aI689
aI1845
aI859
aI333
aI889
aI27607
aI48795
asS'd'
I5158
sa(dp10316
S'a'
I23734
sS's'
I16
sS'm'
(lp10317
I1
aI1
aI5
aI7
aI1
aI11
aI37
aI181
aI323
aI187
aI1511
aI25
aI5517
aI11957
aI7093
aI429
asS'd'
I5159
sa(dp10318
S'a'
I23738
sS's'
I16
sS'm'
(lp10319
I1
aI3
aI3
aI1
aI25
aI31
aI125
aI139
aI433
aI583
aI683
aI587
aI5389
aI1225
aI26047
aI18717
asS'd'
I5160
sa(dp10320
S'a'
I23752
sS's'
I16
sS'm'
(lp10321
I1
aI3
aI1
aI15
aI23
aI33
aI23
aI147
aI279
aI513
aI157
aI4023
aI2669
aI7543
aI2111
aI9191
asS'd'
I5161
sa(dp10322
S'a'
I23800
sS's'
I16
sS'm'
(lp10323
I1
aI3
aI1
aI1
aI5
aI39
aI55
aI127
aI257
aI649
aI347
aI3001
aI2215
aI15579
aI29665
aI10337
asS'd'
I5162
sa(dp10324
S'a'
I23803
sS's'
I16
sS'm'
(lp10325
I1
aI1
aI1
aI15
aI19
aI55
aI105
aI183
aI505
aI1003
aI1311
aI2253
aI1861
aI3561
aI19581
aI46183
asS'd'
I5163
sa(dp10326
S'a'
I23818
sS's'
I16
sS'm'
(lp10327
I1
aI3
aI1
aI9
aI23
aI5
aI127
aI215
aI195
aI817
aI719
aI1285
aI919
aI8627
aI20427
aI2723
asS'd'
I5164
sa(dp10328
S'a'
I23832
sS's'
I16
sS'm'
(lp10329
I1
aI1
aI1
aI5
aI15
aI31
aI43
aI131
aI377
aI57
aI1531
aI915
aI2871
aI1805
aI19765
aI60529
asS'd'
I5165
sa(dp10330
S'a'
I23842
sS's'
I16
sS'm'
(lp10331
I1
aI3
aI3
aI3
aI15
aI1
aI93
aI55
aI477
aI221
aI643
aI1319
aI959
aI475
aI14015
aI48823
asS'd'
I5166
sa(dp10332
S'a'
I23851
sS's'
I16
sS'm'
(lp10333
I1
aI3
aI3
aI7
aI19
aI13
aI13
aI191
aI421
aI751
aI1103
aI2129
aI1973
aI14935
aI26485
aI41269
asS'd'
I5167
sa(dp10334
S'a'
I23873
sS's'
I16
sS'm'
(lp10335
I1
aI1
aI5
aI13
aI17
aI43
aI81
aI83
aI67
aI643
aI1799
aI1157
aI4365
aI2815
aI29871
aI5351
asS'd'
I5168
sa(dp10336
S'a'
I23883
sS's'
I16
sS'm'
(lp10337
I1
aI3
aI1
aI9
aI21
aI31
aI87
aI177
aI25
aI403
aI1357
aI4047
aI959
aI5133
aI7307
aI4333
asS'd'
I5169
sa(dp10338
S'a'
I23888
sS's'
I16
sS'm'
(lp10339
I1
aI1
aI7
aI7
aI29
aI39
aI27
aI139
aI159
aI529
aI1323
aI3823
aI4569
aI12711
aI30263
aI10961
asS'd'
I5170
sa(dp10340
S'a'
I23910
sS's'
I16
sS'm'
(lp10341
I1
aI3
aI1
aI13
aI27
aI15
aI59
aI1
aI107
aI723
aI497
aI43
aI143
aI16119
aI29177
aI5653
asS'd'
I5171
sa(dp10342
S'a'
I23934
sS's'
I16
sS'm'
(lp10343
I1
aI1
aI5
aI9
aI15
aI41
aI23
aI109
aI101
aI639
aI277
aI1687
aI1311
aI1995
aI5403
aI6867
asS'd'
I5172
sa(dp10344
S'a'
I23938
sS's'
I16
sS'm'
(lp10345
I1
aI1
aI5
aI3
aI13
aI1
aI95
aI177
aI379
aI545
aI789
aI3479
aI4135
aI445
aI5869
aI38923
asS'd'
I5173
sa(dp10346
S'a'
I23949
sS's'
I16
sS'm'
(lp10347
I1
aI1
aI3
aI9
aI21
aI31
aI23
aI65
aI209
aI383
aI271
aI367
aI6605
aI1169
aI27267
aI9331
asS'd'
I5174
sa(dp10348
S'a'
I23955
sS's'
I16
sS'm'
(lp10349
I1
aI1
aI1
aI1
aI27
aI39
aI121
aI29
aI155
aI465
aI947
aI2675
aI6753
aI11647
aI29781
aI30251
asS'd'
I5175
sa(dp10350
S'a'
I23967
sS's'
I16
sS'm'
(lp10351
I1
aI3
aI1
aI5
aI7
aI45
aI7
aI21
aI223
aI363
aI1021
aI751
aI2257
aI14423
aI19629
aI64867
asS'd'
I5176
sa(dp10352
S'a'
I23973
sS's'
I16
sS'm'
(lp10353
I1
aI3
aI1
aI9
aI31
aI53
aI13
aI99
aI49
aI389
aI867
aI327
aI4145
aI3265
aI15423
aI14985
asS'd'
I5177
sa(dp10354
S'a'
I23991
sS's'
I16
sS'm'
(lp10355
I1
aI1
aI1
aI15
aI11
aI11
aI27
aI41
aI333
aI161
aI1343
aI2023
aI3789
aI13563
aI16957
aI26849
asS'd'
I5178
sa(dp10356
S'a'
I24024
sS's'
I16
sS'm'
(lp10357
I1
aI3
aI7
aI1
aI7
aI51
aI7
aI163
aI239
aI393
aI231
aI3985
aI309
aI875
aI837
aI24791
asS'd'
I5179
sa(dp10358
S'a'
I24030
sS's'
I16
sS'm'
(lp10359
I1
aI1
aI1
aI7
aI1
aI43
aI105
aI7
aI351
aI755
aI1781
aI1925
aI4961
aI2961
aI13427
aI44317
asS'd'
I5180
sa(dp10360
S'a'
I24039
sS's'
I16
sS'm'
(lp10361
I1
aI3
aI1
aI9
aI17
aI39
aI81
aI75
aI201
aI931
aI1547
aI1857
aI7251
aI11687
aI14437
aI12603
asS'd'
I5181
sa(dp10362
S'a'
I24067
sS's'
I16
sS'm'
(lp10363
I1
aI3
aI3
aI15
aI15
aI23
aI95
aI7
aI193
aI9
aI1749
aI809
aI5083
aI14645
aI24893
aI35979
asS'd'
I5182
sa(dp10364
S'a'
I24069
sS's'
I16
sS'm'
(lp10365
I1
aI1
aI7
aI1
aI9
aI9
aI127
aI149
aI433
aI985
aI1347
aI3379
aI2881
aI681
aI20777
aI30703
asS'd'
I5183
sa(dp10366
S'a'
I24084
sS's'
I16
sS'm'
(lp10367
I1
aI3
aI1
aI11
aI1
aI27
aI121
aI111
aI251
aI45
aI1341
aI1709
aI3733
aI11297
aI29063
aI57119
asS'd'
I5184
sa(dp10368
S'a'
I24112
sS's'
I16
sS'm'
(lp10369
I1
aI3
aI3
aI1
aI19
aI3
aI77
aI127
aI187
aI831
aI1125
aI3119
aI4665
aI9857
aI5301
aI36755
asS'd'
I5185
sa(dp10370
S'a'
I24115
sS's'
I16
sS'm'
(lp10371
I1
aI3
aI3
aI3
aI29
aI29
aI61
aI225
aI257
aI635
aI665
aI1279
aI3019
aI2401
aI8253
aI40251
asS'd'
I5186
sa(dp10372
S'a'
I24184
sS's'
I16
sS'm'
(lp10373
I1
aI1
aI7
aI9
aI29
aI43
aI47
aI95
aI221
aI823
aI257
aI1485
aI5183
aI225
aI27675
aI60479
asS'd'
I5187
sa(dp10374
S'a'
I24187
sS's'
I16
sS'm'
(lp10375
I1
aI1
aI5
aI3
aI15
aI49
aI25
aI69
aI101
aI393
aI901
aI305
aI4917
aI13479
aI30209
aI9199
asS'd'
I5188
sa(dp10376
S'a'
I24262
sS's'
I16
sS'm'
(lp10377
I1
aI1
aI3
aI15
aI1
aI9
aI47
aI243
aI403
aI341
aI143
aI1365
aI1817
aI6017
aI3853
aI58625
asS'd'
I5189
sa(dp10378
S'a'
I24276
sS's'
I16
sS'm'
(lp10379
I1
aI3
aI3
aI11
aI9
aI49
aI93
aI149
aI39
aI177
aI1863
aI1027
aI659
aI9253
aI2131
aI26943
asS'd'
I5190
sa(dp10380
S'a'
I24279
sS's'
I16
sS'm'
(lp10381
I1
aI3
aI1
aI3
aI25
aI1
aI43
aI255
aI217
aI353
aI957
aI39
aI4607
aI15761
aI24291
aI33619
asS'd'
I5191
sa(dp10382
S'a'
I24313
sS's'
I16
sS'm'
(lp10383
I1
aI1
aI1
aI7
aI29
aI51
aI71
aI237
aI1
aI703
aI19
aI213
aI6429
aI11783
aI22049
aI30597
asS'd'
I5192
sa(dp10384
S'a'
I24331
sS's'
I16
sS'm'
(lp10385
I1
aI3
aI1
aI7
aI31
aI63
aI105
aI203
aI473
aI731
aI257
aI91
aI5749
aI4099
aI30125
aI40171
asS'd'
I5193
sa(dp10386
S'a'
I24336
sS's'
I16
sS'm'
(lp10387
I1
aI3
aI7
aI9
aI7
aI29
aI119
aI181
aI225
aI743
aI1031
aI55
aI3997
aI16221
aI11663
aI14847
asS'd'
I5194
sa(dp10388
S'a'
I24348
sS's'
I16
sS'm'
(lp10389
I1
aI3
aI3
aI11
aI5
aI21
aI43
aI17
aI473
aI981
aI125
aI2077
aI6141
aI4757
aI7585
aI29207
asS'd'
I5195
sa(dp10390
S'a'
I24367
sS's'
I16
sS'm'
(lp10391
I1
aI1
aI7
aI1
aI27
aI61
aI27
aI45
aI267
aI483
aI119
aI767
aI957
aI3411
aI2653
aI53967
asS'd'
I5196
sa(dp10392
S'a'
I24372
sS's'
I16
sS'm'
(lp10393
I1
aI1
aI1
aI3
aI9
aI41
aI43
aI17
aI485
aI253
aI825
aI3605
aI5919
aI9285
aI1815
aI6095
asS'd'
I5197
sa(dp10394
S'a'
I24402
sS's'
I16
sS'm'
(lp10395
I1
aI1
aI1
aI11
aI7
aI5
aI53
aI107
aI309
aI609
aI1389
aI2035
aI861
aI15565
aI9375
aI42363
asS'd'
I5198
sa(dp10396
S'a'
I24420
sS's'
I16
sS'm'
(lp10397
I1
aI3
aI5
aI3
aI27
aI57
aI7
aI177
aI183
aI227
aI865
aI183
aI2903
aI6325
aI4393
aI5257
asS'd'
I5199
sa(dp10398
S'a'
I24444
sS's'
I16
sS'm'
(lp10399
I1
aI3
aI1
aI5
aI21
aI29
aI79
aI107
aI365
aI427
aI813
aI3563
aI7713
aI3865
aI4289
aI28555
asS'd'
I5200
sa(dp10400
S'a'
I24465
sS's'
I16
sS'm'
(lp10401
I1
aI3
aI5
aI7
aI11
aI27
aI1
aI197
aI425
aI769
aI1737
aI3627
aI1273
aI4469
aI11967
aI823
asS'd'
I5201
sa(dp10402
S'a'
I24466
sS's'
I16
sS'm'
(lp10403
I1
aI1
aI1
aI13
aI3
aI31
aI127
aI159
aI471
aI367
aI2047
aI949
aI1591
aI12429
aI21497
aI27153
asS'd'
I5202
sa(dp10404
S'a'
I24471
sS's'
I16
sS'm'
(lp10405
I1
aI3
aI1
aI3
aI3
aI31
aI31
aI87
aI243
aI413
aI1777
aI1361
aI4575
aI1147
aI17451
aI33985
asS'd'
I5203
sa(dp10406
S'a'
I24475
sS's'
I16
sS'm'
(lp10407
I1
aI3
aI3
aI5
aI13
aI47
aI45
aI3
aI165
aI329
aI743
aI397
aI2479
aI4999
aI12921
aI26689
asS'd'
I5204
sa(dp10408
S'a'
I24481
sS's'
I16
sS'm'
(lp10409
I1
aI1
aI5
aI7
aI17
aI59
aI25
aI117
aI217
aI601
aI235
aI2691
aI5569
aI7853
aI31063
aI28281
asS'd'
I5205
sa(dp10410
S'a'
I24488
sS's'
I16
sS'm'
(lp10411
I1
aI3
aI1
aI11
aI11
aI39
aI71
aI77
aI481
aI39
aI363
aI1921
aI3263
aI12849
aI13325
aI41803
asS'd'
I5206
sa(dp10412
S'a'
I24501
sS's'
I16
sS'm'
(lp10413
I1
aI3
aI7
aI5
aI19
aI1
aI59
aI197
aI239
aI787
aI1657
aI1449
aI4245
aI1317
aI19609
aI53583
asS'd'
I5207
sa(dp10414
S'a'
I24514
sS's'
I16
sS'm'
(lp10415
I1
aI1
aI7
aI1
aI13
aI33
aI81
aI53
aI223
aI323
aI477
aI2421
aI4045
aI1855
aI5823
aI9661
asS'd'
I5208
sa(dp10416
S'a'
I24531
sS's'
I16
sS'm'
(lp10417
I1
aI3
aI1
aI7
aI1
aI3
aI121
aI19
aI329
aI569
aI481
aI3443
aI499
aI12407
aI19625
aI4627
asS'd'
I5209
sa(dp10418
S'a'
I24534
sS's'
I16
sS'm'
(lp10419
I1
aI1
aI1
aI7
aI3
aI33
aI91
aI241
aI69
aI581
aI1635
aI1025
aI4677
aI14651
aI5229
aI19555
asS'd'
I5210
sa(dp10420
S'a'
I24559
sS's'
I16
sS'm'
(lp10421
I1
aI3
aI1
aI15
aI9
aI11
aI99
aI47
aI489
aI755
aI601
aI1221
aI4251
aI4377
aI20567
aI33839
asS'd'
I5211
sa(dp10422
S'a'
I24602
sS's'
I16
sS'm'
(lp10423
I1
aI3
aI3
aI5
aI21
aI21
aI127
aI151
aI499
aI971
aI1627
aI609
aI2365
aI3183
aI7413
aI697
asS'd'
I5212
sa(dp10424
S'a'
I24614
sS's'
I16
sS'm'
(lp10425
I1
aI3
aI7
aI13
aI5
aI35
aI61
aI121
aI51
aI297
aI29
aI1825
aI495
aI1299
aI12741
aI3253
asS'd'
I5213
sa(dp10426
S'a'
I24637
sS's'
I16
sS'm'
(lp10427
I1
aI1
aI1
aI1
aI13
aI15
aI49
aI205
aI235
aI9
aI849
aI1101
aI4533
aI10221
aI32419
aI50151
asS'd'
I5214
sa(dp10428
S'a'
I24664
sS's'
I16
sS'm'
(lp10429
I1
aI1
aI3
aI13
aI29
aI31
aI57
aI77
aI217
aI195
aI1951
aI189
aI981
aI6169
aI22677
aI64415
asS'd'
I5215
sa(dp10430
S'a'
I24676
sS's'
I16
sS'm'
(lp10431
I1
aI3
aI1
aI5
aI15
aI37
aI25
aI233
aI43
aI843
aI1205
aI153
aI6339
aI3767
aI16725
aI32699
asS'd'
I5216
sa(dp10432
S'a'
I24679
sS's'
I16
sS'm'
(lp10433
I1
aI3
aI3
aI1
aI19
aI37
aI11
aI217
aI461
aI897
aI1181
aI2815
aI295
aI15153
aI25351
aI57211
asS'd'
I5217
sa(dp10434
S'a'
I24697
sS's'
I16
sS'm'
(lp10435
I1
aI3
aI5
aI11
aI15
aI9
aI5
aI179
aI353
aI769
aI1457
aI2919
aI1201
aI14871
aI29549
aI52265
asS'd'
I5218
sa(dp10436
S'a'
I24709
sS's'
I16
sS'm'
(lp10437
I1
aI1
aI3
aI3
aI5
aI51
aI127
aI221
aI329
aI543
aI1537
aI2701
aI2709
aI9311
aI2715
aI42669
asS'd'
I5219
sa(dp10438
S'a'
I24714
sS's'
I16
sS'm'
(lp10439
I1
aI3
aI5
aI15
aI5
aI41
aI79
aI233
aI445
aI265
aI2001
aI935
aI3133
aI3977
aI3601
aI21389
asS'd'
I5220
sa(dp10440
S'a'
I24716
sS's'
I16
sS'm'
(lp10441
I1
aI3
aI5
aI11
aI3
aI7
aI115
aI45
aI311
aI827
aI1897
aI3399
aI4251
aI5341
aI22621
aI25519
asS'd'
I5221
sa(dp10442
S'a'
I24731
sS's'
I16
sS'm'
(lp10443
I1
aI3
aI7
aI1
aI11
aI57
aI117
aI103
aI401
aI505
aI1683
aI2161
aI4363
aI11189
aI20263
aI13065
asS'd'
I5222
sa(dp10444
S'a'
I24744
sS's'
I16
sS'm'
(lp10445
I1
aI1
aI1
aI7
aI23
aI29
aI31
aI77
aI63
aI179
aI195
aI2747
aI579
aI11701
aI5101
aI11497
asS'd'
I5223
sa(dp10446
S'a'
I24762
sS's'
I16
sS'm'
(lp10447
I1
aI3
aI7
aI3
aI9
aI33
aI81
aI165
aI433
aI545
aI1893
aI3709
aI3813
aI6615
aI1485
aI6395
asS'd'
I5224
sa(dp10448
S'a'
I24764
sS's'
I16
sS'm'
(lp10449
I1
aI3
aI1
aI15
aI9
aI5
aI27
aI157
aI389
aI683
aI1919
aI509
aI455
aI3865
aI2303
aI6993
asS'd'
I5225
sa(dp10450
S'a'
I24769
sS's'
I16
sS'm'
(lp10451
I1
aI3
aI3
aI9
aI5
aI5
aI3
aI5
aI299
aI59
aI539
aI1199
aI2443
aI10821
aI3359
aI44345
asS'd'
I5226
sa(dp10452
S'a'
I24806
sS's'
I16
sS'm'
(lp10453
I1
aI3
aI5
aI9
aI21
aI37
aI87
aI35
aI501
aI943
aI1313
aI3929
aI351
aI9851
aI22971
aI35659
asS'd'
I5227
sa(dp10454
S'a'
I24812
sS's'
I16
sS'm'
(lp10455
I1
aI3
aI7
aI11
aI11
aI33
aI77
aI175
aI411
aI315
aI1797
aI2731
aI4611
aI1809
aI22027
aI50595
asS'd'
I5228
sa(dp10456
S'a'
I24872
sS's'
I16
sS'm'
(lp10457
I1
aI3
aI7
aI13
aI15
aI11
aI13
aI189
aI209
aI1015
aI1869
aI1593
aI6887
aI8571
aI7217
aI2641
asS'd'
I5229
sa(dp10458
S'a'
I24878
sS's'
I16
sS'm'
(lp10459
I1
aI1
aI3
aI13
aI29
aI15
aI119
aI127
aI329
aI275
aI865
aI1693
aI225
aI15735
aI11071
aI37127
asS'd'
I5230
sa(dp10460
S'a'
I24883
sS's'
I16
sS'm'
(lp10461
I1
aI3
aI7
aI13
aI9
aI31
aI85
aI25
aI281
aI401
aI1923
aI2391
aI3875
aI2079
aI2055
aI53275
asS'd'
I5231
sa(dp10462
S'a'
I24909
sS's'
I16
sS'm'
(lp10463
I1
aI3
aI1
aI5
aI23
aI57
aI79
aI127
aI209
aI901
aI315
aI1165
aI3393
aI15095
aI1981
aI34993
asS'd'
I5232
sa(dp10464
S'a'
I24943
sS's'
I16
sS'm'
(lp10465
I1
aI3
aI7
aI7
aI25
aI13
aI15
aI223
aI157
aI335
aI1061
aI395
aI6895
aI6283
aI18375
aI4887
asS'd'
I5233
sa(dp10466
S'a'
I24946
sS's'
I16
sS'm'
(lp10467
I1
aI3
aI7
aI13
aI9
aI15
aI99
aI201
aI105
aI643
aI117
aI3027
aI641
aI13353
aI4343
aI11875
asS'd'
I5234
sa(dp10468
S'a'
I24964
sS's'
I16
sS'm'
(lp10469
I1
aI1
aI3
aI11
aI5
aI51
aI5
aI77
aI281
aI207
aI1201
aI1187
aI8107
aI6625
aI7517
aI34377
asS'd'
I5235
sa(dp10470
S'a'
I24992
sS's'
I16
sS'm'
(lp10471
I1
aI1
aI1
aI5
aI29
aI61
aI3
aI181
aI297
aI151
aI565
aI2713
aI6611
aI8665
aI32425
aI50399
asS'd'
I5236
sa(dp10472
S'a'
I25016
sS's'
I16
sS'm'
(lp10473
I1
aI3
aI1
aI5
aI1
aI61
aI41
aI245
aI95
aI647
aI49
aI2195
aI5927
aI15531
aI28547
aI51075
asS'd'
I5237
sa(dp10474
S'a'
I25022
sS's'
I16
sS'm'
(lp10475
I1
aI3
aI3
aI15
aI15
aI63
aI123
aI63
aI77
aI813
aI423
aI715
aI91
aI3793
aI20901
aI54927
asS'd'
I5238
sa(dp10476
S'a'
I25024
sS's'
I16
sS'm'
(lp10477
I1
aI3
aI7
aI9
aI15
aI35
aI31
aI161
aI127
aI13
aI1667
aI1225
aI5279
aI15487
aI18769
aI24887
asS'd'
I5239
sa(dp10478
S'a'
I25039
sS's'
I16
sS'm'
(lp10479
I1
aI1
aI3
aI5
aI27
aI25
aI61
aI21
aI447
aI175
aI1419
aI2691
aI1993
aI13059
aI30809
aI38325
asS'd'
I5240
sa(dp10480
S'a'
I25051
sS's'
I16
sS'm'
(lp10481
I1
aI3
aI1
aI3
aI15
aI21
aI15
aI193
aI233
aI435
aI1993
aI4003
aI4581
aI13837
aI16535
aI43781
asS'd'
I5241
sa(dp10482
S'a'
I25060
sS's'
I16
sS'm'
(lp10483
I1
aI1
aI1
aI5
aI1
aI11
aI21
aI253
aI59
aI59
aI497
aI77
aI2165
aI8197
aI5933
aI25743
asS'd'
I5242
sa(dp10484
S'a'
I25072
sS's'
I16
sS'm'
(lp10485
I1
aI1
aI3
aI9
aI25
aI61
aI29
aI217
aI95
aI269
aI799
aI409
aI801
aI421
aI19065
aI53443
asS'd'
I5243
sa(dp10486
S'a'
I25097
sS's'
I16
sS'm'
(lp10487
I1
aI1
aI7
aI1
aI1
aI41
aI71
aI59
aI191
aI867
aI223
aI1467
aI6679
aI16031
aI4451
aI15313
asS'd'
I5244
sa(dp10488
S'a'
I25122
sS's'
I16
sS'm'
(lp10489
I1
aI1
aI1
aI11
aI13
aI27
aI7
aI241
aI167
aI969
aI1267
aI2953
aI3769
aI2415
aI30065
aI39483
asS'd'
I5245
sa(dp10490
S'a'
I25148
sS's'
I16
sS'm'
(lp10491
I1
aI1
aI1
aI3
aI25
aI61
aI103
aI23
aI53
aI799
aI989
aI3859
aI7299
aI13613
aI12007
aI37535
asS'd'
I5246
sa(dp10492
S'a'
I25194
sS's'
I16
sS'm'
(lp10493
I1
aI1
aI7
aI1
aI29
aI19
aI121
aI57
aI355
aI663
aI643
aI3723
aI1775
aI10363
aI1389
aI16039
asS'd'
I5247
sa(dp10494
S'a'
I25201
sS's'
I16
sS'm'
(lp10495
I1
aI3
aI7
aI3
aI21
aI55
aI51
aI67
aI363
aI843
aI1187
aI1983
aI7757
aI5413
aI26575
aI53329
asS'd'
I5248
sa(dp10496
S'a'
I25204
sS's'
I16
sS'm'
(lp10497
I1
aI3
aI1
aI3
aI31
aI55
aI73
aI55
aI75
aI533
aI197
aI1463
aI2933
aI6033
aI24397
aI41579
asS'd'
I5249
sa(dp10498
S'a'
I25238
sS's'
I16
sS'm'
(lp10499
I1
aI3
aI1
aI11
aI9
aI15
aI107
aI141
aI473
aI825
aI901
aI937
aI7433
aI13119
aI20047
aI6695
asS'd'
I5250
sa(dp10500
S'a'
I25241
sS's'
I16
sS'm'
(lp10501
I1
aI3
aI5
aI7
aI19
aI27
aI3
aI149
aI507
aI137
aI1025
aI1841
aI33
aI3113
aI15101
aI28187
asS'd'
I5251
sa(dp10502
S'a'
I25248
sS's'
I16
sS'm'
(lp10503
I1
aI3
aI5
aI7
aI31
aI27
aI53
aI45
aI177
aI129
aI1241
aI1525
aI991
aI4141
aI17681
aI39435
asS'd'
I5252
sa(dp10504
S'a'
I25257
sS's'
I16
sS'm'
(lp10505
I1
aI1
aI1
aI15
aI31
aI57
aI29
aI137
aI395
aI563
aI101
aI3367
aI1277
aI5431
aI1169
aI44321
asS'd'
I5253
sa(dp10506
S'a'
I25275
sS's'
I16
sS'm'
(lp10507
I1
aI3
aI5
aI7
aI21
aI15
aI123
aI181
aI113
aI313
aI1763
aI1429
aI2985
aI715
aI26129
aI549
asS'd'
I5254
sa(dp10508
S'a'
I25278
sS's'
I16
sS'm'
(lp10509
I1
aI3
aI3
aI1
aI15
aI27
aI27
aI139
aI507
aI79
aI1999
aI2505
aI485
aI7011
aI13369
aI7159
asS'd'
I5255
sa(dp10510
S'a'
I25304
sS's'
I16
sS'm'
(lp10511
I1
aI3
aI3
aI11
aI27
aI53
aI109
aI179
aI399
aI657
aI697
aI421
aI5467
aI4273
aI7837
aI11631
asS'd'
I5256
sa(dp10512
S'a'
I25307
sS's'
I16
sS'm'
(lp10513
I1
aI1
aI1
aI15
aI1
aI57
aI91
aI199
aI443
aI569
aI1945
aI2531
aI6349
aI4851
aI3931
aI20537
asS'd'
I5257
sa(dp10514
S'a'
I25320
sS's'
I16
sS'm'
(lp10515
I1
aI1
aI3
aI13
aI3
aI3
aI107
aI237
aI261
aI377
aI135
aI2809
aI7239
aI1613
aI24035
aI26053
asS'd'
I5258
sa(dp10516
S'a'
I25334
sS's'
I16
sS'm'
(lp10517
I1
aI3
aI3
aI5
aI3
aI59
aI65
aI197
aI411
aI363
aI1729
aI967
aI3963
aI4535
aI111
aI7273
asS'd'
I5259
sa(dp10518
S'a'
I25348
sS's'
I16
sS'm'
(lp10519
I1
aI1
aI7
aI3
aI13
aI39
aI105
aI235
aI203
aI1015
aI1031
aI3127
aI1209
aI10817
aI22177
aI44117
asS'd'
I5260
sa(dp10520
S'a'
I25357
sS's'
I16
sS'm'
(lp10521
I1
aI3
aI3
aI13
aI19
aI21
aI123
aI31
aI59
aI185
aI1007
aI1115
aI1965
aI13087
aI18489
aI34063
asS'd'
I5261
sa(dp10522
S'a'
I25372
sS's'
I16
sS'm'
(lp10523
I1
aI1
aI7
aI5
aI27
aI7
aI33
aI159
aI245
aI57
aI2003
aI3229
aI2095
aI4939
aI31355
aI23121
asS'd'
I5262
sa(dp10524
S'a'
I25394
sS's'
I16
sS'm'
(lp10525
I1
aI3
aI3
aI9
aI9
aI41
aI49
aI203
aI397
aI915
aI821
aI3685
aI2269
aI11159
aI25441
aI46377
asS'd'
I5263
sa(dp10526
S'a'
I25454
sS's'
I16
sS'm'
(lp10527
I1
aI3
aI7
aI5
aI29
aI33
aI29
aI227
aI361
aI961
aI1905
aI1149
aI2799
aI8115
aI28235
aI25685
asS'd'
I5264
sa(dp10528
S'a'
I25465
sS's'
I16
sS'm'
(lp10529
I1
aI3
aI1
aI1
aI19
aI13
aI73
aI103
aI11
aI183
aI853
aI2425
aI3809
aI2391
aI18615
aI32735
asS'd'
I5265
sa(dp10530
S'a'
I25472
sS's'
I16
sS'm'
(lp10531
I1
aI1
aI3
aI3
aI21
aI57
aI47
aI57
aI157
aI43
aI1085
aI3319
aI461
aI11499
aI6809
aI10435
asS'd'
I5266
sa(dp10532
S'a'
I25492
sS's'
I16
sS'm'
(lp10533
I1
aI1
aI5
aI5
aI17
aI21
aI55
aI17
aI421
aI865
aI159
aI1643
aI4523
aI1485
aI11937
aI8287
asS'd'
I5267
sa(dp10534
S'a'
I25505
sS's'
I16
sS'm'
(lp10535
I1
aI1
aI3
aI11
aI7
aI43
aI39
aI37
aI187
aI797
aI1273
aI1227
aI2683
aI1249
aI3375
aI44499
asS'd'
I5268
sa(dp10536
S'a'
I25517
sS's'
I16
sS'm'
(lp10537
I1
aI1
aI5
aI11
aI17
aI35
aI27
aI73
aI97
aI59
aI921
aI2171
aI7577
aI2847
aI93
aI35911
asS'd'
I5269
sa(dp10538
S'a'
I25530
sS's'
I16
sS'm'
(lp10539
I1
aI1
aI5
aI1
aI5
aI17
aI117
aI189
aI357
aI581
aI1945
aI2141
aI1679
aI12019
aI11249
aI6809
asS'd'
I5270
sa(dp10540
S'a'
I25558
sS's'
I16
sS'm'
(lp10541
I1
aI1
aI5
aI7
aI15
aI53
aI9
aI191
aI153
aI257
aI533
aI493
aI2389
aI4657
aI20757
aI57275
asS'd'
I5271
sa(dp10542
S'a'
I25562
sS's'
I16
sS'm'
(lp10543
I1
aI1
aI1
aI11
aI13
aI35
aI85
aI37
aI501
aI525
aI543
aI4057
aI2001
aI6183
aI949
aI18465
asS'd'
I5272
sa(dp10544
S'a'
I25598
sS's'
I16
sS'm'
(lp10545
I1
aI1
aI1
aI3
aI15
aI7
aI39
aI169
aI359
aI671
aI731
aI1523
aI211
aI1233
aI29515
aI57787
asS'd'
I5273
sa(dp10546
S'a'
I25609
sS's'
I16
sS'm'
(lp10547
I1
aI1
aI3
aI7
aI27
aI7
aI41
aI15
aI71
aI733
aI1919
aI401
aI1915
aI4815
aI10571
aI839
asS'd'
I5274
sa(dp10548
S'a'
I25612
sS's'
I16
sS'm'
(lp10549
I1
aI1
aI7
aI13
aI27
aI61
aI5
aI25
aI293
aI779
aI477
aI1537
aI6695
aI7435
aI1281
aI64369
asS'd'
I5275
sa(dp10550
S'a'
I25627
sS's'
I16
sS'm'
(lp10551
I1
aI1
aI7
aI7
aI19
aI11
aI101
aI45
aI449
aI883
aI1181
aI3521
aI6019
aI16305
aI23429
aI43789
asS'd'
I5276
sa(dp10552
S'a'
I25651
sS's'
I16
sS'm'
(lp10553
I1
aI1
aI7
aI5
aI1
aI49
aI121
aI89
aI275
aI367
aI461
aI1717
aI2733
aI4403
aI9123
aI35217
asS'd'
I5277
sa(dp10554
S'a'
I25653
sS's'
I16
sS'm'
(lp10555
I1
aI1
aI7
aI1
aI1
aI37
aI41
aI221
aI281
aI531
aI357
aI3783
aI3561
aI8135
aI18405
aI56045
asS'd'
I5278
sa(dp10556
S'a'
I25668
sS's'
I16
sS'm'
(lp10557
I1
aI3
aI5
aI7
aI29
aI9
aI127
aI37
aI13
aI519
aI1059
aI3973
aI7253
aI15159
aI19337
aI57103
asS'd'
I5279
sa(dp10558
S'a'
I25711
sS's'
I16
sS'm'
(lp10559
I1
aI3
aI3
aI15
aI3
aI41
aI91
aI7
aI63
aI747
aI1649
aI3367
aI5945
aI3603
aI28465
aI511
asS'd'
I5280
sa(dp10560
S'a'
I25765
sS's'
I16
sS'm'
(lp10561
I1
aI3
aI3
aI15
aI27
aI19
aI67
aI179
aI505
aI131
aI149
aI1753
aI3603
aI1135
aI15811
aI5305
asS'd'
I5281
sa(dp10562
S'a'
I25770
sS's'
I16
sS'm'
(lp10563
I1
aI1
aI1
aI5
aI5
aI63
aI71
aI235
aI151
aI651
aI1383
aI249
aI3223
aI14559
aI26809
aI4551
asS'd'
I5282
sa(dp10564
S'a'
I25775
sS's'
I16
sS'm'
(lp10565
I1
aI3
aI3
aI9
aI29
aI41
aI67
aI111
aI175
aI515
aI1123
aI1707
aI6653
aI8233
aI22775
aI61029
asS'd'
I5283
sa(dp10566
S'a'
I25777
sS's'
I16
sS'm'
(lp10567
I1
aI3
aI3
aI9
aI23
aI1
aI75
aI145
aI159
aI785
aI537
aI1995
aI2241
aI8641
aI30709
aI41373
asS'd'
I5284
sa(dp10568
S'a'
I25783
sS's'
I16
sS'm'
(lp10569
I1
aI1
aI5
aI9
aI21
aI1
aI87
aI233
aI401
aI643
aI197
aI3437
aI8163
aI6363
aI6537
aI17283
asS'd'
I5285
sa(dp10570
S'a'
I25801
sS's'
I16
sS'm'
(lp10571
I1
aI3
aI5
aI3
aI23
aI19
aI55
aI243
aI405
aI369
aI1221
aI1959
aI5455
aI11729
aI26117
aI9097
asS'd'
I5286
sa(dp10572
S'a'
I25802
sS's'
I16
sS'm'
(lp10573
I1
aI1
aI3
aI13
aI3
aI57
aI71
aI235
aI125
aI263
aI873
aI1079
aI2627
aI1343
aI1979
aI49519
asS'd'
I5287
sa(dp10574
S'a'
I25812
sS's'
I16
sS'm'
(lp10575
I1
aI3
aI1
aI11
aI21
aI15
aI27
aI7
aI425
aI935
aI305
aI2593
aI4437
aI9517
aI26207
aI4229
asS'd'
I5288
sa(dp10576
S'a'
I25821
sS's'
I16
sS'm'
(lp10577
I1
aI1
aI3
aI13
aI11
aI53
aI1
aI149
aI97
aI939
aI147
aI3365
aI5023
aI607
aI2083
aI8715
asS'd'
I5289
sa(dp10578
S'a'
I25897
sS's'
I16
sS'm'
(lp10579
I1
aI1
aI5
aI3
aI13
aI13
aI113
aI51
aI263
aI837
aI1145
aI3621
aI5697
aI2867
aI7975
aI22839
asS'd'
I5290
sa(dp10580
S'a'
I25906
sS's'
I16
sS'm'
(lp10581
I1
aI1
aI3
aI15
aI31
aI9
aI91
aI231
aI399
aI295
aI1935
aI4021
aI7669
aI3867
aI28015
aI9865
asS'd'
I5291
sa(dp10582
S'a'
I25929
sS's'
I16
sS'm'
(lp10583
I1
aI1
aI1
aI1
aI13
aI59
aI51
aI35
aI407
aI733
aI1629
aI2429
aI291
aI11923
aI32129
aI28847
asS'd'
I5292
sa(dp10584
S'a'
I25940
sS's'
I16
sS'm'
(lp10585
I1
aI3
aI3
aI11
aI25
aI21
aI13
aI117
aI31
aI547
aI327
aI2801
aI2247
aI4051
aI27523
aI4257
asS'd'
I5293
sa(dp10586
S'a'
I25950
sS's'
I16
sS'm'
(lp10587
I1
aI1
aI7
aI7
aI15
aI33
aI15
aI17
aI255
aI363
aI1013
aI1463
aI7537
aI14093
aI21883
aI35389
asS'd'
I5294
sa(dp10588
S'a'
I25968
sS's'
I16
sS'm'
(lp10589
I1
aI1
aI5
aI9
aI11
aI61
aI7
aI161
aI121
aI413
aI515
aI413
aI4439
aI15405
aI30265
aI23939
asS'd'
I5295
sa(dp10590
S'a'
I25971
sS's'
I16
sS'm'
(lp10591
I1
aI3
aI7
aI7
aI11
aI15
aI5
aI181
aI315
aI231
aI1567
aI2985
aI1653
aI12251
aI269
aI37317
asS'd'
I5296
sa(dp10592
S'a'
I25977
sS's'
I16
sS'm'
(lp10593
I1
aI3
aI1
aI11
aI3
aI15
aI91
aI45
aI489
aI571
aI11
aI1239
aI7705
aI4303
aI12535
aI21359
asS'd'
I5297
sa(dp10594
S'a'
I25994
sS's'
I16
sS'm'
(lp10595
I1
aI1
aI7
aI15
aI29
aI43
aI81
aI63
aI483
aI851
aI389
aI1719
aI6111
aI15293
aI2513
aI44397
asS'd'
I5298
sa(dp10596
S'a'
I25996
sS's'
I16
sS'm'
(lp10597
I1
aI1
aI5
aI15
aI25
aI33
aI97
aI131
aI183
aI269
aI1903
aI2733
aI7197
aI4507
aI24471
aI36871
asS'd'
I5299
sa(dp10598
S'a'
I25999
sS's'
I16
sS'm'
(lp10599
I1
aI3
aI3
aI13
aI17
aI33
aI73
aI83
aI247
aI207
aI79
aI1139
aI581
aI12147
aI3539
aI45423
asS'd'
I5300
sa(dp10600
S'a'
I26001
sS's'
I16
sS'm'
(lp10601
I1
aI1
aI1
aI15
aI29
aI49
aI79
aI151
aI267
aI393
aI1995
aI105
aI2873
aI3981
aI19147
aI53987
asS'd'
I5301
sa(dp10602
S'a'
I26030
sS's'
I16
sS'm'
(lp10603
I1
aI1
aI5
aI1
aI31
aI25
aI39
aI203
aI459
aI181
aI661
aI717
aI6235
aI13413
aI1197
aI40665
asS'd'
I5302
sa(dp10604
S'a'
I26069
sS's'
I16
sS'm'
(lp10605
I1
aI1
aI5
aI9
aI19
aI33
aI65
aI239
aI463
aI133
aI461
aI3601
aI7755
aI1771
aI20683
aI7417
asS'd'
I5303
sa(dp10606
S'a'
I26100
sS's'
I16
sS'm'
(lp10607
I1
aI1
aI1
aI1
aI25
aI19
aI25
aI155
aI431
aI33
aI341
aI959
aI5679
aI1205
aI2599
aI37499
asS'd'
I5304
sa(dp10608
S'a'
I26109
sS's'
I16
sS'm'
(lp10609
I1
aI1
aI3
aI5
aI25
aI5
aI83
aI87
aI91
aI991
aI1833
aI4063
aI147
aI14497
aI25725
aI4617
asS'd'
I5305
sa(dp10610
S'a'
I26131
sS's'
I16
sS'm'
(lp10611
I1
aI3
aI5
aI7
aI31
aI7
aI73
aI51
aI339
aI313
aI1593
aI2089
aI7387
aI15759
aI3249
aI7953
asS'd'
I5306
sa(dp10612
S'a'
I26144
sS's'
I16
sS'm'
(lp10613
I1
aI3
aI7
aI1
aI27
aI49
aI35
aI11
aI21
aI45
aI1689
aI1665
aI4591
aI3713
aI12781
aI4805
asS'd'
I5307
sa(dp10614
S'a'
I26149
sS's'
I16
sS'm'
(lp10615
I1
aI1
aI3
aI9
aI29
aI51
aI73
aI51
aI303
aI179
aI67
aI3917
aI7615
aI6131
aI26225
aI55991
asS'd'
I5308
sa(dp10616
S'a'
I26162
sS's'
I16
sS'm'
(lp10617
I1
aI3
aI7
aI11
aI9
aI63
aI29
aI47
aI153
aI883
aI1859
aI1913
aI3563
aI11451
aI6333
aI51367
asS'd'
I5309
sa(dp10618
S'a'
I26167
sS's'
I16
sS'm'
(lp10619
I1
aI3
aI1
aI3
aI5
aI25
aI69
aI87
aI389
aI613
aI989
aI3557
aI1339
aI12503
aI14505
aI63209
asS'd'
I5310
sa(dp10620
S'a'
I26173
sS's'
I16
sS'm'
(lp10621
I1
aI1
aI3
aI1
aI5
aI13
aI37
aI163
aI499
aI163
aI2025
aI1467
aI5059
aI8479
aI2889
aI62957
asS'd'
I5311
sa(dp10622
S'a'
I26179
sS's'
I16
sS'm'
(lp10623
I1
aI1
aI7
aI9
aI23
aI31
aI109
aI49
aI73
aI197
aI337
aI2763
aI4789
aI8983
aI9691
aI32155
asS'd'
I5312
sa(dp10624
S'a'
I26193
sS's'
I16
sS'm'
(lp10625
I1
aI3
aI1
aI3
aI31
aI25
aI121
aI91
aI371
aI339
aI833
aI2217
aI4997
aI9425
aI10685
aI60037
asS'd'
I5313
sa(dp10626
S'a'
I26230
sS's'
I16
sS'm'
(lp10627
I1
aI1
aI7
aI11
aI31
aI3
aI105
aI125
aI255
aI175
aI803
aI1787
aI6231
aI4441
aI5031
aI29737
asS'd'
I5314
sa(dp10628
S'a'
I26234
sS's'
I16
sS'm'
(lp10629
I1
aI1
aI1
aI11
aI21
aI63
aI75
aI209
aI393
aI437
aI495
aI2397
aI4759
aI15703
aI29869
aI62685
asS'd'
I5315
sa(dp10630
S'a'
I26246
sS's'
I16
sS'm'
(lp10631
I1
aI1
aI7
aI7
aI25
aI33
aI117
aI7
aI293
aI623
aI2001
aI2409
aI2487
aI14803
aI3329
aI38277
asS'd'
I5316
sa(dp10632
S'a'
I26267
sS's'
I16
sS'm'
(lp10633
I1
aI1
aI7
aI9
aI31
aI51
aI75
aI151
aI487
aI85
aI639
aI3529
aI4491
aI1957
aI22099
aI20263
asS'd'
I5317
sa(dp10634
S'a'
I26283
sS's'
I16
sS'm'
(lp10635
I1
aI1
aI7
aI5
aI3
aI29
aI11
aI1
aI255
aI555
aI1269
aI779
aI1525
aI7689
aI25847
aI39347
asS'd'
I5318
sa(dp10636
S'a'
I26300
sS's'
I16
sS'm'
(lp10637
I1
aI1
aI7
aI7
aI19
aI21
aI9
aI123
aI3
aI943
aI1627
aI2979
aI919
aI4565
aI31435
aI59789
asS'd'
I5319
sa(dp10638
S'a'
I26341
sS's'
I16
sS'm'
(lp10639
I1
aI3
aI7
aI5
aI29
aI13
aI57
aI221
aI9
aI893
aI1685
aI1879
aI4575
aI7369
aI26191
aI6067
asS'd'
I5320
sa(dp10640
S'a'
I26356
sS's'
I16
sS'm'
(lp10641
I1
aI3
aI7
aI9
aI13
aI11
aI9
aI27
aI313
aI751
aI1377
aI1121
aI3799
aI1673
aI16305
aI30665
asS'd'
I5321
sa(dp10642
S'a'
I26377
sS's'
I16
sS'm'
(lp10643
I1
aI3
aI3
aI13
aI23
aI17
aI59
aI47
aI499
aI525
aI681
aI3195
aI1611
aI7003
aI7325
aI53019
asS'd'
I5322
sa(dp10644
S'a'
I26397
sS's'
I16
sS'm'
(lp10645
I1
aI3
aI1
aI7
aI23
aI51
aI59
aI127
aI67
aI263
aI1305
aI2479
aI637
aI9441
aI6329
aI12857
asS'd'
I5323
sa(dp10646
S'a'
I26404
sS's'
I16
sS'm'
(lp10647
I1
aI1
aI7
aI7
aI9
aI3
aI51
aI193
aI205
aI503
aI19
aI2513
aI7489
aI9241
aI15371
aI20875
asS'd'
I5324
sa(dp10648
S'a'
I26411
sS's'
I16
sS'm'
(lp10649
I1
aI3
aI3
aI1
aI1
aI57
aI17
aI181
aI23
aI549
aI769
aI2325
aI3669
aI7017
aI25601
aI64479
asS'd'
I5325
sa(dp10650
S'a'
I26422
sS's'
I16
sS'm'
(lp10651
I1
aI3
aI1
aI15
aI5
aI55
aI53
aI13
aI327
aI447
aI1031
aI1599
aI3639
aI15305
aI1387
aI16035
asS'd'
I5326
sa(dp10652
S'a'
I26451
sS's'
I16
sS'm'
(lp10653
I1
aI3
aI7
aI15
aI9
aI41
aI53
aI113
aI97
aI99
aI1377
aI4047
aI3713
aI8891
aI5601
aI5853
asS'd'
I5327
sa(dp10654
S'a'
I26454
sS's'
I16
sS'm'
(lp10655
I1
aI1
aI7
aI9
aI9
aI7
aI29
aI249
aI411
aI315
aI181
aI2153
aI6135
aI6947
aI27595
aI15553
asS'd'
I5328
sa(dp10656
S'a'
I26463
sS's'
I16
sS'm'
(lp10657
I1
aI1
aI7
aI11
aI3
aI57
aI35
aI55
aI165
aI313
aI577
aI3049
aI4259
aI4231
aI7225
aI58973
asS'd'
I5329
sa(dp10658
S'a'
I26488
sS's'
I16
sS'm'
(lp10659
I1
aI1
aI1
aI1
aI15
aI43
aI53
aI143
aI157
aI843
aI465
aI3897
aI4797
aI12305
aI28807
aI46381
asS'd'
I5330
sa(dp10660
S'a'
I26498
sS's'
I16
sS'm'
(lp10661
I1
aI3
aI7
aI9
aI17
aI3
aI99
aI61
aI475
aI507
aI831
aI2207
aI367
aI27
aI23205
aI2303
asS'd'
I5331
sa(dp10662
S'a'
I26509
sS's'
I16
sS'm'
(lp10663
I1
aI1
aI3
aI11
aI27
aI29
aI99
aI237
aI43
aI955
aI361
aI3231
aI1863
aI7973
aI8969
aI58663
asS'd'
I5332
sa(dp10664
S'a'
I26512
sS's'
I16
sS'm'
(lp10665
I1
aI3
aI1
aI7
aI15
aI15
aI11
aI251
aI135
aI261
aI675
aI3723
aI7675
aI7993
aI1725
aI41149
asS'd'
I5333
sa(dp10666
S'a'
I26517
sS's'
I16
sS'm'
(lp10667
I1
aI3
aI3
aI9
aI29
aI11
aI105
aI37
aI151
aI215
aI1911
aI4051
aI5427
aI11019
aI9073
aI33129
asS'd'
I5334
sa(dp10668
S'a'
I26534
sS's'
I16
sS'm'
(lp10669
I1
aI3
aI3
aI1
aI19
aI7
aI103
aI81
aI371
aI253
aI1043
aI1231
aI6497
aI10377
aI2349
aI29047
asS'd'
I5335
sa(dp10670
S'a'
I26545
sS's'
I16
sS'm'
(lp10671
I1
aI3
aI7
aI9
aI15
aI11
aI85
aI61
aI507
aI629
aI811
aI3883
aI1435
aI13035
aI31913
aI2153
asS'd'
I5336
sa(dp10672
S'a'
I26546
sS's'
I16
sS'm'
(lp10673
I1
aI1
aI5
aI11
aI13
aI7
aI63
aI147
aI117
aI223
aI1217
aI3317
aI3275
aI6851
aI2917
aI55901
asS'd'
I5337
sa(dp10674
S'a'
I26636
sS's'
I16
sS'm'
(lp10675
I1
aI3
aI3
aI9
aI7
aI21
aI1
aI63
aI117
aI297
aI405
aI797
aI337
aI10173
aI8327
aI29157
asS'd'
I5338
sa(dp10676
S'a'
I26749
sS's'
I16
sS'm'
(lp10677
I1
aI1
aI7
aI11
aI31
aI63
aI97
aI191
aI259
aI421
aI1829
aI2117
aI4203
aI11919
aI18001
aI55791
asS'd'
I5339
sa(dp10678
S'a'
I26753
sS's'
I16
sS'm'
(lp10679
I1
aI3
aI7
aI9
aI21
aI13
aI125
aI247
aI133
aI611
aI463
aI227
aI7815
aI8877
aI17961
aI3641
asS'd'
I5340
sa(dp10680
S'a'
I26759
sS's'
I16
sS'm'
(lp10681
I1
aI1
aI7
aI9
aI27
aI21
aI97
aI165
aI371
aI715
aI491
aI3929
aI3067
aI12501
aI5511
aI18217
asS'd'
I5341
sa(dp10682
S'a'
I26774
sS's'
I16
sS'm'
(lp10683
I1
aI3
aI3
aI15
aI27
aI17
aI81
aI161
aI263
aI273
aI135
aI1159
aI7535
aI4581
aI16065
aI11493
asS'd'
I5342
sa(dp10684
S'a'
I26789
sS's'
I16
sS'm'
(lp10685
I1
aI3
aI3
aI7
aI11
aI59
aI111
aI47
aI381
aI413
aI243
aI2173
aI4957
aI2451
aI15669
aI22071
asS'd'
I5343
sa(dp10686
S'a'
I26808
sS's'
I16
sS'm'
(lp10687
I1
aI3
aI7
aI5
aI3
aI31
aI65
aI131
aI111
aI141
aI1891
aI2983
aI3331
aI769
aI24075
aI40865
asS'd'
I5344
sa(dp10688
S'a'
I26825
sS's'
I16
sS'm'
(lp10689
I1
aI3
aI7
aI11
aI11
aI63
aI7
aI213
aI333
aI111
aI1235
aI961
aI3749
aI9123
aI5067
aI9657
asS'd'
I5345
sa(dp10690
S'a'
I26831
sS's'
I16
sS'm'
(lp10691
I1
aI3
aI3
aI1
aI9
aI33
aI1
aI225
aI37
aI951
aI1995
aI3215
aI3117
aI3723
aI15013
aI64525
asS'd'
I5346
sa(dp10692
S'a'
I26859
sS's'
I16
sS'm'
(lp10693
I1
aI3
aI3
aI13
aI29
aI19
aI103
aI65
aI67
aI423
aI1679
aI3791
aI5551
aI11711
aI195
aI52291
asS'd'
I5347
sa(dp10694
S'a'
I26888
sS's'
I16
sS'm'
(lp10695
I1
aI3
aI1
aI15
aI31
aI7
aI65
aI249
aI489
aI287
aI1385
aI1075
aI1357
aI13593
aI20853
aI46221
asS'd'
I5348
sa(dp10696
S'a'
I26918
sS's'
I16
sS'm'
(lp10697
I1
aI1
aI1
aI13
aI23
aI45
aI29
aI175
aI147
aI101
aI1007
aI1867
aI5387
aI12683
aI29609
aI55861
asS'd'
I5349
sa(dp10698
S'a'
I26929
sS's'
I16
sS'm'
(lp10699
I1
aI3
aI5
aI13
aI21
aI31
aI85
aI187
aI183
aI179
aI1337
aI481
aI71
aI6117
aI2177
aI27915
asS'd'
I5350
sa(dp10700
S'a'
I26950
sS's'
I16
sS'm'
(lp10701
I1
aI3
aI5
aI1
aI15
aI5
aI11
aI141
aI205
aI177
aI891
aI3591
aI4371
aI889
aI12957
aI61083
asS'd'
I5351
sa(dp10702
S'a'
I26954
sS's'
I16
sS'm'
(lp10703
I1
aI3
aI7
aI7
aI11
aI39
aI81
aI241
aI13
aI757
aI521
aI3029
aI2345
aI12385
aI20683
aI64053
asS'd'
I5352
sa(dp10704
S'a'
I26973
sS's'
I16
sS'm'
(lp10705
I1
aI1
aI5
aI13
aI7
aI3
aI77
aI211
aI215
aI97
aI1409
aI1021
aI1267
aI4785
aI27231
aI2877
asS'd'
I5353
sa(dp10706
S'a'
I26997
sS's'
I16
sS'm'
(lp10707
I1
aI3
aI5
aI3
aI11
aI57
aI93
aI39
aI415
aI179
aI1033
aI3267
aI5383
aI10451
aI27117
aI10711
asS'd'
I5354
sa(dp10708
S'a'
I26998
sS's'
I16
sS'm'
(lp10709
I1
aI1
aI1
aI1
aI3
aI45
aI93
aI179
aI453
aI995
aI1423
aI3849
aI4381
aI15789
aI20789
aI18775
asS'd'
I5355
sa(dp10710
S'a'
I27008
sS's'
I16
sS'm'
(lp10711
I1
aI3
aI1
aI3
aI13
aI25
aI33
aI165
aI351
aI887
aI1109
aI195
aI8131
aI3061
aI16743
aI22997
asS'd'
I5356
sa(dp10712
S'a'
I27038
sS's'
I16
sS'm'
(lp10713
I1
aI3
aI1
aI5
aI23
aI35
aI93
aI155
aI333
aI603
aI1571
aI229
aI2979
aI6295
aI20793
aI40901
asS'd'
I5357
sa(dp10714
S'a'
I27109
sS's'
I16
sS'm'
(lp10715
I1
aI3
aI3
aI11
aI29
aI57
aI101
aI61
aI487
aI719
aI1009
aI1933
aI7815
aI15329
aI12489
aI26105
asS'd'
I5358
sa(dp10716
S'a'
I27127
sS's'
I16
sS'm'
(lp10717
I1
aI3
aI3
aI9
aI23
aI59
aI73
aI13
aI141
aI815
aI1819
aI3557
aI2555
aI5901
aI23039
aI62321
asS'd'
I5359
sa(dp10718
S'a'
I27150
sS's'
I16
sS'm'
(lp10719
I1
aI1
aI3
aI5
aI19
aI49
aI27
aI139
aI35
aI995
aI565
aI323
aI6439
aI15679
aI27017
aI30889
asS'd'
I5360
sa(dp10720
S'a'
I27168
sS's'
I16
sS'm'
(lp10721
I1
aI3
aI7
aI3
aI1
aI3
aI27
aI153
aI235
aI59
aI989
aI2763
aI4197
aI3931
aI31227
aI27129
asS'd'
I5361
sa(dp10722
S'a'
I27178
sS's'
I16
sS'm'
(lp10723
I1
aI3
aI1
aI15
aI23
aI45
aI71
aI205
aI465
aI451
aI347
aI1909
aI3287
aI8363
aI9081
aI35641
asS'd'
I5362
sa(dp10724
S'a'
I27212
sS's'
I16
sS'm'
(lp10725
I1
aI1
aI5
aI1
aI25
aI29
aI17
aI201
aI463
aI903
aI1729
aI3435
aI2483
aI10835
aI14315
aI52505
asS'd'
I5363
sa(dp10726
S'a'
I27224
sS's'
I16
sS'm'
(lp10727
I1
aI1
aI1
aI15
aI13
aI23
aI3
aI175
aI273
aI305
aI1945
aI3319
aI7777
aI9411
aI4209
aI4047
asS'd'
I5364
sa(dp10728
S'a'
I27229
sS's'
I16
sS'm'
(lp10729
I1
aI1
aI5
aI15
aI25
aI5
aI71
aI35
aI307
aI89
aI301
aI3465
aI1497
aI13467
aI21911
aI13611
asS'd'
I5365
sa(dp10730
S'a'
I27246
sS's'
I16
sS'm'
(lp10731
I1
aI3
aI1
aI7
aI11
aI7
aI33
aI241
aI349
aI751
aI633
aI3281
aI6733
aI13833
aI23605
aI34307
asS'd'
I5366
sa(dp10732
S'a'
I27251
sS's'
I16
sS'm'
(lp10733
I1
aI1
aI1
aI15
aI17
aI27
aI45
aI13
aI259
aI843
aI1207
aI1735
aI4063
aI6259
aI1751
aI45107
asS'd'
I5367
sa(dp10734
S'a'
I27257
sS's'
I16
sS'm'
(lp10735
I1
aI1
aI5
aI15
aI29
aI51
aI73
aI95
aI5
aI31
aI933
aI423
aI5505
aI2193
aI14919
aI2715
asS'd'
I5368
sa(dp10736
S'a'
I27258
sS's'
I16
sS'm'
(lp10737
I1
aI3
aI1
aI3
aI23
aI5
aI29
aI7
aI271
aI485
aI827
aI1159
aI77
aI16337
aI27933
aI18741
asS'd'
I5369
sa(dp10738
S'a'
I27260
sS's'
I16
sS'm'
(lp10739
I1
aI3
aI7
aI9
aI23
aI33
aI47
aI191
aI59
aI301
aI1277
aI3745
aI7837
aI799
aI2861
aI25853
asS'd'
I5370
sa(dp10740
S'a'
I27287
sS's'
I16
sS'm'
(lp10741
I1
aI3
aI7
aI13
aI13
aI39
aI33
aI91
aI279
aI855
aI1917
aI3601
aI3971
aI6193
aI16951
aI6115
asS'd'
I5371
sa(dp10742
S'a'
I27300
sS's'
I16
sS'm'
(lp10743
I1
aI1
aI3
aI13
aI15
aI59
aI89
aI239
aI313
aI545
aI431
aI3823
aI5741
aI14981
aI2647
aI42813
asS'd'
I5372
sa(dp10744
S'a'
I27315
sS's'
I16
sS'm'
(lp10745
I1
aI1
aI1
aI3
aI17
aI21
aI45
aI37
aI343
aI737
aI1795
aI2659
aI2897
aI7683
aI15191
aI1393
asS'd'
I5373
sa(dp10746
S'a'
I27329
sS's'
I16
sS'm'
(lp10747
I1
aI1
aI3
aI3
aI19
aI55
aI27
aI201
aI459
aI953
aI1531
aI671
aI5667
aI11695
aI149
aI14605
asS'd'
I5374
sa(dp10748
S'a'
I27332
sS's'
I16
sS'm'
(lp10749
I1
aI3
aI7
aI13
aI9
aI63
aI67
aI229
aI69
aI819
aI859
aI2035
aI5725
aI13403
aI24197
aI2599
asS'd'
I5375
sa(dp10750
S'a'
I27349
sS's'
I16
sS'm'
(lp10751
I1
aI1
aI7
aI7
aI1
aI59
aI45
aI105
aI327
aI779
aI59
aI791
aI7593
aI8189
aI28161
aI13339
asS'd'
I5376
sa(dp10752
S'a'
I27350
sS's'
I16
sS'm'
(lp10753
I1
aI3
aI3
aI15
aI25
aI55
aI125
aI189
aI327
aI733
aI115
aI3541
aI5227
aI12143
aI32719
aI15785
asS'd'
I5377
sa(dp10754
S'a'
I27354
sS's'
I16
sS'm'
(lp10755
I1
aI3
aI3
aI7
aI19
aI51
aI35
aI63
aI507
aI89
aI1441
aI2369
aI4927
aI7953
aI10193
aI8331
asS'd'
I5378
sa(dp10756
S'a'
I27359
sS's'
I16
sS'm'
(lp10757
I1
aI1
aI5
aI5
aI21
aI1
aI41
aI49
aI217
aI1001
aI1649
aI2789
aI5267
aI1525
aI3811
aI40785
asS'd'
I5379
sa(dp10758
S'a'
I27377
sS's'
I16
sS'm'
(lp10759
I1
aI1
aI7
aI15
aI31
aI21
aI33
aI183
aI425
aI393
aI367
aI3253
aI3047
aI465
aI28229
aI44743
asS'd'
I5380
sa(dp10760
S'a'
I27378
sS's'
I16
sS'm'
(lp10761
I1
aI3
aI7
aI5
aI1
aI23
aI11
aI71
aI269
aI707
aI5
aI2931
aI1959
aI15089
aI9299
aI43927
asS'd'
I5381
sa(dp10762
S'a'
I27383
sS's'
I16
sS'm'
(lp10763
I1
aI3
aI5
aI15
aI21
aI51
aI31
aI15
aI49
aI481
aI297
aI3871
aI751
aI10661
aI26401
aI62923
asS'd'
I5382
sa(dp10764
S'a'
I27384
sS's'
I16
sS'm'
(lp10765
I1
aI3
aI1
aI7
aI17
aI27
aI35
aI255
aI205
aI865
aI1659
aI1921
aI5457
aI11633
aI2825
aI13549
asS'd'
I5383
sa(dp10766
S'a'
I27387
sS's'
I16
sS'm'
(lp10767
I1
aI1
aI5
aI15
aI17
aI35
aI83
aI237
aI437
aI7
aI2001
aI2225
aI2601
aI10841
aI7953
aI20651
asS'd'
I5384
sa(dp10768
S'a'
I27392
sS's'
I16
sS'm'
(lp10769
I1
aI1
aI1
aI3
aI3
aI37
aI43
aI135
aI451
aI203
aI1319
aI261
aI3889
aI14489
aI9635
aI52545
asS'd'
I5385
sa(dp10770
S'a'
I27402
sS's'
I16
sS'm'
(lp10771
I1
aI3
aI3
aI13
aI15
aI41
aI95
aI207
aI43
aI997
aI207
aI3459
aI5995
aI5187
aI15851
aI28551
asS'd'
I5386
sa(dp10772
S'a'
I27438
sS's'
I16
sS'm'
(lp10773
I1
aI1
aI1
aI5
aI23
aI57
aI49
aI101
aI303
aI921
aI745
aI1407
aI7071
aI2765
aI18703
aI32671
asS'd'
I5387
sa(dp10774
S'a'
I27481
sS's'
I16
sS'm'
(lp10775
I1
aI1
aI7
aI13
aI9
aI59
aI65
aI157
aI209
aI295
aI107
aI3175
aI3401
aI1197
aI1875
aI9033
asS'd'
I5388
sa(dp10776
S'a'
I27482
sS's'
I16
sS'm'
(lp10777
I1
aI1
aI3
aI3
aI17
aI9
aI101
aI75
aI177
aI905
aI1013
aI397
aI3421
aI6475
aI15897
aI11269
asS'd'
I5389
sa(dp10778
S'a'
I27494
sS's'
I16
sS'm'
(lp10779
I1
aI3
aI1
aI5
aI29
aI53
aI105
aI83
aI383
aI137
aI1169
aI1245
aI6973
aI8701
aI317
aI10073
asS'd'
I5390
sa(dp10780
S'a'
I27531
sS's'
I16
sS'm'
(lp10781
I1
aI1
aI1
aI3
aI15
aI55
aI69
aI219
aI507
aI707
aI945
aI397
aI779
aI4157
aI10333
aI7869
asS'd'
I5391
sa(dp10782
S'a'
I27542
sS's'
I16
sS'm'
(lp10783
I1
aI3
aI1
aI3
aI9
aI21
aI125
aI153
aI107
aI969
aI1979
aI651
aI1199
aI11419
aI17043
aI32269
asS'd'
I5392
sa(dp10784
S'a'
I27546
sS's'
I16
sS'm'
(lp10785
I1
aI1
aI1
aI7
aI1
aI29
aI71
aI127
aI209
aI853
aI1515
aI1169
aI5055
aI9981
aI30291
aI29569
asS'd'
I5393
sa(dp10786
S'a'
I27564
sS's'
I16
sS'm'
(lp10787
I1
aI3
aI1
aI11
aI1
aI1
aI109
aI251
aI329
aI633
aI2021
aI1237
aI2147
aI11471
aI26025
aI19649
asS'd'
I5394
sa(dp10788
S'a'
I27567
sS's'
I16
sS'm'
(lp10789
I1
aI1
aI5
aI1
aI5
aI7
aI77
aI175
aI251
aI143
aI711
aI1241
aI2133
aI9993
aI9203
aI24949
asS'd'
I5395
sa(dp10790
S'a'
I27582
sS's'
I16
sS'm'
(lp10791
I1
aI3
aI5
aI11
aI19
aI11
aI101
aI83
aI91
aI595
aI753
aI2389
aI1887
aI11569
aI29759
aI55785
asS'd'
I5396
sa(dp10792
S'a'
I27608
sS's'
I16
sS'm'
(lp10793
I1
aI1
aI1
aI3
aI29
aI47
aI49
aI249
aI495
aI451
aI1887
aI2547
aI543
aI2755
aI17207
aI24379
asS'd'
I5397
sa(dp10794
S'a'
I27611
sS's'
I16
sS'm'
(lp10795
I1
aI3
aI7
aI7
aI19
aI15
aI95
aI143
aI109
aI221
aI2041
aI3593
aI4571
aI14547
aI14217
aI16711
asS'd'
I5398
sa(dp10796
S'a'
I27624
sS's'
I16
sS'm'
(lp10797
I1
aI3
aI5
aI13
aI27
aI55
aI31
aI209
aI39
aI989
aI1435
aI1665
aI7265
aI14127
aI13517
aI37647
asS'd'
I5399
sa(dp10798
S'a'
I27629
sS's'
I16
sS'm'
(lp10799
I1
aI1
aI3
aI7
aI1
aI49
aI63
aI71
aI249
aI371
aI435
aI3591
aI2677
aI143
aI28897
aI38981
asS'd'
I5400
sa(dp10800
S'a'
I27655
sS's'
I16
sS'm'
(lp10801
I1
aI1
aI7
aI7
aI21
aI9
aI53
aI221
aI23
aI515
aI1971
aI3759
aI3511
aI10207
aI12929
aI42021
asS'd'
I5401
sa(dp10802
S'a'
I27667
sS's'
I16
sS'm'
(lp10803
I1
aI3
aI1
aI13
aI25
aI21
aI3
aI85
aI421
aI19
aI663
aI3219
aI3541
aI13021
aI5765
aI39623
asS'd'
I5402
sa(dp10804
S'a'
I27676
sS's'
I16
sS'm'
(lp10805
I1
aI3
aI1
aI7
aI11
aI5
aI125
aI169
aI293
aI715
aI1111
aI2965
aI4815
aI6047
aI27207
aI23093
asS'd'
I5403
sa(dp10806
S'a'
I27710
sS's'
I16
sS'm'
(lp10807
I1
aI1
aI5
aI13
aI11
aI15
aI37
aI201
aI457
aI551
aI821
aI25
aI435
aI14307
aI25855
aI1811
asS'd'
I5404
sa(dp10808
S'a'
I27724
sS's'
I16
sS'm'
(lp10809
I1
aI3
aI3
aI9
aI27
aI15
aI1
aI253
aI217
aI549
aI1013
aI2277
aI4171
aI3813
aI19857
aI8641
asS'd'
I5405
sa(dp10810
S'a'
I27745
sS's'
I16
sS'm'
(lp10811
I1
aI3
aI5
aI5
aI29
aI37
aI71
aI49
aI163
aI949
aI425
aI2459
aI945
aI13125
aI13981
aI21669
asS'd'
I5406
sa(dp10812
S'a'
I27752
sS's'
I16
sS'm'
(lp10813
I1
aI3
aI3
aI9
aI17
aI23
aI53
aI235
aI83
aI887
aI439
aI1939
aI7601
aI15275
aI15739
aI17623
asS'd'
I5407
sa(dp10814
S'a'
I27770
sS's'
I16
sS'm'
(lp10815
I1
aI3
aI5
aI13
aI7
aI51
aI73
aI67
aI167
aI849
aI2021
aI2977
aI6591
aI3721
aI5827
aI40897
asS'd'
I5408
sa(dp10816
S'a'
I27779
sS's'
I16
sS'm'
(lp10817
I1
aI1
aI5
aI11
aI27
aI19
aI81
aI181
aI383
aI23
aI1061
aI3327
aI1671
aI7113
aI7435
aI17591
asS'd'
I5409
sa(dp10818
S'a'
I27781
sS's'
I16
sS'm'
(lp10819
I1
aI3
aI3
aI7
aI25
aI33
aI73
aI23
aI103
aI821
aI917
aI1425
aI4739
aI7227
aI12365
aI29509
asS'd'
I5410
sa(dp10820
S'a'
I27791
sS's'
I16
sS'm'
(lp10821
I1
aI1
aI1
aI7
aI3
aI37
aI81
aI231
aI135
aI663
aI1983
aI399
aI6881
aI14991
aI4957
aI20913
asS'd'
I5411
sa(dp10822
S'a'
I27809
sS's'
I16
sS'm'
(lp10823
I1
aI3
aI7
aI15
aI25
aI41
aI65
aI215
aI301
aI471
aI1669
aI65
aI227
aI9307
aI29867
aI9503
asS'd'
I5412
sa(dp10824
S'a'
I27810
sS's'
I16
sS'm'
(lp10825
I1
aI1
aI7
aI3
aI25
aI47
aI31
aI63
aI53
aI995
aI33
aI1297
aI3423
aI12301
aI16255
aI14467
asS'd'
I5413
sa(dp10826
S'a'
I27815
sS's'
I16
sS'm'
(lp10827
I1
aI3
aI1
aI1
aI31
aI25
aI79
aI131
aI353
aI169
aI1425
aI2257
aI2631
aI1559
aI793
aI48383
asS'd'
I5414
sa(dp10828
S'a'
I27827
sS's'
I16
sS'm'
(lp10829
I1
aI3
aI3
aI5
aI31
aI9
aI93
aI35
aI503
aI243
aI595
aI2939
aI771
aI7333
aI13429
aI59457
asS'd'
I5415
sa(dp10830
S'a'
I27834
sS's'
I16
sS'm'
(lp10831
I1
aI3
aI1
aI7
aI5
aI51
aI21
aI237
aI453
aI743
aI775
aI2207
aI453
aI12077
aI12283
aI9735
asS'd'
I5416
sa(dp10832
S'a'
I27865
sS's'
I16
sS'm'
(lp10833
I1
aI3
aI1
aI15
aI5
aI47
aI59
aI239
aI87
aI97
aI885
aI3191
aI2547
aI13227
aI7433
aI4989
asS'd'
I5417
sa(dp10834
S'a'
I27899
sS's'
I16
sS'm'
(lp10835
I1
aI3
aI5
aI15
aI21
aI33
aI41
aI155
aI509
aI317
aI517
aI1101
aI133
aI1897
aI8235
aI57673
asS'd'
I5418
sa(dp10836
S'a'
I27901
sS's'
I16
sS'm'
(lp10837
I1
aI1
aI5
aI15
aI7
aI9
aI59
aI155
aI415
aI831
aI1173
aI1263
aI5451
aI7181
aI7233
aI51483
asS'd'
I5419
sa(dp10838
S'a'
I27914
sS's'
I16
sS'm'
(lp10839
I1
aI1
aI7
aI3
aI31
aI43
aI71
aI39
aI155
aI529
aI895
aI827
aI3203
aI4625
aI32185
aI53507
asS'd'
I5420
sa(dp10840
S'a'
I27950
sS's'
I16
sS'm'
(lp10841
I1
aI3
aI1
aI11
aI15
aI17
aI85
aI141
aI277
aI439
aI1775
aI4015
aI4457
aI281
aI22455
aI47591
asS'd'
I5421
sa(dp10842
S'a'
I27994
sS's'
I16
sS'm'
(lp10843
I1
aI3
aI5
aI11
aI25
aI41
aI93
aI39
aI51
aI655
aI1347
aI3109
aI2479
aI9057
aI18939
aI26217
asS'd'
I5422
sa(dp10844
S'a'
I28005
sS's'
I16
sS'm'
(lp10845
I1
aI3
aI3
aI11
aI31
aI41
aI7
aI189
aI241
aI443
aI65
aI1723
aI4817
aI13405
aI9641
aI63965
asS'd'
I5423
sa(dp10846
S'a'
I28009
sS's'
I16
sS'm'
(lp10847
I1
aI1
aI5
aI3
aI19
aI29
aI111
aI11
aI399
aI277
aI425
aI1331
aI5365
aI14521
aI16449
aI29411
asS'd'
I5424
sa(dp10848
S'a'
I28033
sS's'
I16
sS'm'
(lp10849
I1
aI1
aI3
aI9
aI25
aI53
aI91
aI175
aI481
aI307
aI1025
aI71
aI7425
aI10667
aI4053
aI25605
asS'd'
I5425
sa(dp10850
S'a'
I28039
sS's'
I16
sS'm'
(lp10851
I1
aI3
aI7
aI7
aI3
aI13
aI75
aI175
aI467
aI363
aI1889
aI1759
aI1155
aI5511
aI13047
aI39637
asS'd'
I5426
sa(dp10852
S'a'
I28060
sS's'
I16
sS'm'
(lp10853
I1
aI3
aI7
aI9
aI5
aI21
aI65
aI43
aI223
aI97
aI835
aI2253
aI3313
aI9817
aI23015
aI55365
asS'd'
I5427
sa(dp10854
S'a'
I28067
sS's'
I16
sS'm'
(lp10855
I1
aI1
aI1
aI13
aI9
aI63
aI95
aI61
aI417
aI713
aI1469
aI1815
aI795
aI13609
aI1567
aI33535
asS'd'
I5428
sa(dp10856
S'a'
I28069
sS's'
I16
sS'm'
(lp10857
I1
aI3
aI7
aI1
aI25
aI45
aI41
aI27
aI53
aI407
aI1633
aI1317
aI6267
aI3293
aI8899
aI45235
asS'd'
I5429
sa(dp10858
S'a'
I28099
sS's'
I16
sS'm'
(lp10859
I1
aI3
aI5
aI11
aI23
aI47
aI91
aI211
aI41
aI775
aI1301
aI1407
aI7931
aI4491
aI7579
aI62321
asS'd'
I5430
sa(dp10860
S'a'
I28113
sS's'
I16
sS'm'
(lp10861
I1
aI1
aI1
aI7
aI23
aI15
aI57
aI31
aI437
aI293
aI1999
aI2589
aI5453
aI2519
aI15533
aI26949
asS'd'
I5431
sa(dp10862
S'a'
I28114
sS's'
I16
sS'm'
(lp10863
I1
aI3
aI1
aI9
aI1
aI27
aI41
aI165
aI129
aI297
aI1887
aI1171
aI201
aI5817
aI24503
aI38911
asS'd'
I5432
sa(dp10864
S'a'
I28139
sS's'
I16
sS'm'
(lp10865
I1
aI3
aI1
aI7
aI1
aI11
aI63
aI225
aI191
aI623
aI1281
aI3275
aI167
aI14697
aI4905
aI47289
asS'd'
I5433
sa(dp10866
S'a'
I28142
sS's'
I16
sS'm'
(lp10867
I1
aI3
aI7
aI7
aI15
aI47
aI87
aI177
aI303
aI391
aI355
aI3997
aI7557
aI6201
aI20531
aI22483
asS'd'
I5434
sa(dp10868
S'a'
I28153
sS's'
I16
sS'm'
(lp10869
I1
aI3
aI3
aI15
aI17
aI31
aI111
aI87
aI61
aI477
aI1581
aI3787
aI5919
aI10511
aI2607
aI62951
asS'd'
I5435
sa(dp10870
S'a'
I28166
sS's'
I16
sS'm'
(lp10871
I1
aI3
aI3
aI9
aI29
aI19
aI63
aI27
aI205
aI915
aI1131
aI3821
aI673
aI2875
aI12703
aI14367
asS'd'
I5436
sa(dp10872
S'a'
I28172
sS's'
I16
sS'm'
(lp10873
I1
aI3
aI7
aI1
aI21
aI19
aI25
aI97
aI281
aI635
aI629
aI181
aI5207
aI11133
aI3687
aI3489
asS'd'
I5437
sa(dp10874
S'a'
I28183
sS's'
I16
sS'm'
(lp10875
I1
aI3
aI3
aI9
aI5
aI63
aI3
aI21
aI385
aI713
aI1805
aI3583
aI2807
aI15455
aI13057
aI39771
asS'd'
I5438
sa(dp10876
S'a'
I28194
sS's'
I16
sS'm'
(lp10877
I1
aI3
aI5
aI9
aI3
aI59
aI1
aI253
aI123
aI405
aI575
aI3911
aI4609
aI11869
aI23593
aI947
asS'd'
I5439
sa(dp10878
S'a'
I28232
sS's'
I16
sS'm'
(lp10879
I1
aI1
aI7
aI5
aI21
aI27
aI101
aI221
aI413
aI153
aI1647
aI3637
aI803
aI5697
aI20761
aI61137
asS'd'
I5440
sa(dp10880
S'a'
I28245
sS's'
I16
sS'm'
(lp10881
I1
aI3
aI7
aI13
aI31
aI35
aI111
aI253
aI187
aI499
aI465
aI157
aI5551
aI10417
aI323
aI34913
asS'd'
I5441
sa(dp10882
S'a'
I28246
sS's'
I16
sS'm'
(lp10883
I1
aI1
aI1
aI7
aI11
aI41
aI29
aI65
aI393
aI69
aI1373
aI2291
aI7807
aI13159
aI13735
aI31001
asS'd'
I5442
sa(dp10884
S'a'
I28252
sS's'
I16
sS'm'
(lp10885
I1
aI3
aI7
aI13
aI31
aI49
aI1
aI35
aI377
aI11
aI427
aI2803
aI1725
aI9165
aI22633
aI63985
asS'd'
I5443
sa(dp10886
S'a'
I28265
sS's'
I16
sS'm'
(lp10887
I1
aI3
aI7
aI13
aI3
aI41
aI27
aI43
aI269
aI599
aI1035
aI3681
aI309
aI6011
aI1065
aI27901
asS'd'
I5444
sa(dp10888
S'a'
I28301
sS's'
I16
sS'm'
(lp10889
I1
aI3
aI5
aI13
aI1
aI19
aI105
aI143
aI425
aI883
aI1669
aI155
aI189
aI8573
aI10759
aI25507
asS'd'
I5445
sa(dp10890
S'a'
I28323
sS's'
I16
sS'm'
(lp10891
I1
aI3
aI5
aI1
aI15
aI37
aI115
aI9
aI149
aI79
aI1733
aI1045
aI1849
aI3289
aI13957
aI63171
asS'd'
I5446
sa(dp10892
S'a'
I28344
sS's'
I16
sS'm'
(lp10893
I1
aI1
aI3
aI9
aI17
aI27
aI49
aI201
aI155
aI901
aI47
aI1585
aI4419
aI8117
aI25425
aI14699
asS'd'
I5447
sa(dp10894
S'a'
I28362
sS's'
I16
sS'm'
(lp10895
I1
aI1
aI7
aI13
aI19
aI55
aI19
aI21
aI427
aI77
aI1295
aI1471
aI6271
aI7985
aI19337
aI12701
asS'd'
I5448
sa(dp10896
S'a'
I28400
sS's'
I16
sS'm'
(lp10897
I1
aI1
aI1
aI1
aI11
aI49
aI101
aI53
aI175
aI157
aI839
aI2713
aI6149
aI6391
aI8089
aI27739
asS'd'
I5449
sa(dp10898
S'a'
I28417
sS's'
I16
sS'm'
(lp10899
I1
aI3
aI1
aI1
aI5
aI21
aI121
aI199
aI107
aI221
aI993
aI1737
aI409
aI2545
aI9287
aI54605
asS'd'
I5450
sa(dp10900
S'a'
I28454
sS's'
I16
sS'm'
(lp10901
I1
aI1
aI1
aI3
aI25
aI25
aI51
aI121
aI371
aI861
aI967
aI3257
aI6221
aI11701
aI27897
aI42509
asS'd'
I5451
sa(dp10902
S'a'
I28466
sS's'
I16
sS'm'
(lp10903
I1
aI1
aI1
aI3
aI17
aI25
aI101
aI191
aI313
aI817
aI815
aI1855
aI7999
aI12649
aI23385
aI26081
asS'd'
I5452
sa(dp10904
S'a'
I28468
sS's'
I16
sS'm'
(lp10905
I1
aI1
aI5
aI1
aI25
aI55
aI51
aI237
aI63
aI943
aI455
aI619
aI2381
aI9773
aI14575
aI34205
asS'd'
I5453
sa(dp10906
S'a'
I28477
sS's'
I16
sS'm'
(lp10907
I1
aI3
aI3
aI3
aI13
aI49
aI101
aI37
aI457
aI727
aI1009
aI2389
aI4841
aI16303
aI9599
aI17773
asS'd'
I5454
sa(dp10908
S'a'
I28498
sS's'
I16
sS'm'
(lp10909
I1
aI1
aI7
aI9
aI19
aI59
aI111
aI205
aI19
aI229
aI1755
aI1169
aI7767
aI13335
aI19669
aI33269
asS'd'
I5455
sa(dp10910
S'a'
I28510
sS's'
I16
sS'm'
(lp10911
I1
aI3
aI1
aI15
aI29
aI1
aI51
aI167
aI7
aI415
aI1295
aI3165
aI1241
aI12859
aI5531
aI20083
asS'd'
I5456
sa(dp10912
S'a'
I28513
sS's'
I16
sS'm'
(lp10913
I1
aI1
aI3
aI7
aI7
aI51
aI31
aI221
aI57
aI643
aI1461
aI3951
aI6237
aI5757
aI1907
aI40471
asS'd'
I5457
sa(dp10914
S'a'
I28571
sS's'
I16
sS'm'
(lp10915
I1
aI3
aI3
aI5
aI23
aI39
aI49
aI177
aI183
aI117
aI1379
aI3803
aI771
aI12723
aI22291
aI32667
asS'd'
I5458
sa(dp10916
S'a'
I28573
sS's'
I16
sS'm'
(lp10917
I1
aI1
aI3
aI13
aI27
aI17
aI39
aI27
aI313
aI141
aI1421
aI2967
aI2213
aI1915
aI23219
aI15113
asS'd'
I5459
sa(dp10918
S'a'
I28578
sS's'
I16
sS'm'
(lp10919
I1
aI1
aI1
aI11
aI5
aI55
aI51
aI55
aI389
aI895
aI57
aI1447
aI1497
aI2799
aI19585
aI11587
asS'd'
I5460
sa(dp10920
S'a'
I28587
sS's'
I16
sS'm'
(lp10921
I1
aI1
aI5
aI13
aI11
aI55
aI91
aI77
aI69
aI131
aI93
aI1383
aI3321
aI10487
aI15087
aI8539
asS'd'
I5461
sa(dp10922
S'a'
I28601
sS's'
I16
sS'm'
(lp10923
I1
aI1
aI3
aI9
aI23
aI49
aI107
aI131
aI363
aI733
aI1189
aI3575
aI7815
aI10071
aI20291
aI7533
asS'd'
I5462
sa(dp10924
S'a'
I28646
sS's'
I16
sS'm'
(lp10925
I1
aI1
aI3
aI15
aI31
aI31
aI73
aI15
aI199
aI17
aI761
aI3271
aI1419
aI12985
aI32717
aI37317
asS'd'
I5463
sa(dp10926
S'a'
I28663
sS's'
I16
sS'm'
(lp10927
I1
aI3
aI1
aI13
aI23
aI9
aI3
aI59
aI109
aI729
aI1321
aI4023
aI7041
aI14445
aI22205
aI8993
asS'd'
I5464
sa(dp10928
S'a'
I28681
sS's'
I16
sS'm'
(lp10929
I1
aI1
aI3
aI15
aI19
aI43
aI99
aI59
aI491
aI479
aI715
aI2235
aI7493
aI889
aI31465
aI1375
asS'd'
I5465
sa(dp10930
S'a'
I28682
sS's'
I16
sS'm'
(lp10931
I1
aI3
aI3
aI15
aI9
aI47
aI35
aI115
aI227
aI615
aI605
aI1143
aI5923
aI10939
aI9497
aI24911
asS'd'
I5466
sa(dp10932
S'a'
I28699
sS's'
I16
sS'm'
(lp10933
I1
aI1
aI3
aI15
aI23
aI53
aI111
aI87
aI423
aI497
aI85
aI3525
aI7341
aI8885
aI21543
aI30083
asS'd'
I5467
sa(dp10934
S'a'
I28706
sS's'
I16
sS'm'
(lp10935
I1
aI1
aI5
aI3
aI21
aI5
aI117
aI157
aI407
aI743
aI715
aI1883
aI4425
aI10187
aI6395
aI43673
asS'd'
I5468
sa(dp10936
S'a'
I28708
sS's'
I16
sS'm'
(lp10937
I1
aI3
aI3
aI3
aI31
aI39
aI119
aI77
aI269
aI891
aI1391
aI3085
aI2881
aI10639
aI3391
aI44911
asS'd'
I5469
sa(dp10938
S'a'
I28717
sS's'
I16
sS'm'
(lp10939
I1
aI3
aI7
aI5
aI7
aI5
aI115
aI91
aI5
aI107
aI1401
aI1409
aI1811
aI737
aI5399
aI9119
asS'd'
I5470
sa(dp10940
S'a'
I28720
sS's'
I16
sS'm'
(lp10941
I1
aI1
aI5
aI9
aI17
aI45
aI107
aI15
aI397
aI299
aI1219
aI1675
aI963
aI10111
aI31679
aI13809
asS'd'
I5471
sa(dp10942
S'a'
I28735
sS's'
I16
sS'm'
(lp10943
I1
aI1
aI3
aI7
aI29
aI17
aI43
aI95
aI261
aI601
aI1091
aI3633
aI1357
aI13461
aI16583
aI12183
asS'd'
I5472
sa(dp10944
S'a'
I28761
sS's'
I16
sS'm'
(lp10945
I1
aI1
aI5
aI1
aI19
aI55
aI5
aI195
aI187
aI427
aI421
aI1717
aI4223
aI2943
aI23147
aI32985
asS'd'
I5473
sa(dp10946
S'a'
I28783
sS's'
I16
sS'm'
(lp10947
I1
aI3
aI1
aI3
aI3
aI23
aI69
aI95
aI347
aI273
aI1223
aI3061
aI1587
aI4707
aI32415
aI53991
asS'd'
I5474
sa(dp10948
S'a'
I28788
sS's'
I16
sS'm'
(lp10949
I1
aI1
aI7
aI13
aI15
aI13
aI29
aI151
aI325
aI949
aI2029
aI813
aI5339
aI14165
aI1159
aI56917
asS'd'
I5475
sa(dp10950
S'a'
I28811
sS's'
I16
sS'm'
(lp10951
I1
aI1
aI1
aI13
aI9
aI33
aI67
aI109
aI215
aI313
aI1407
aI3543
aI2403
aI5051
aI20367
aI13527
asS'd'
I5476
sa(dp10952
S'a'
I28825
sS's'
I16
sS'm'
(lp10953
I1
aI3
aI1
aI9
aI5
aI1
aI9
aI195
aI497
aI495
aI1107
aI745
aI1647
aI10637
aI1933
aI44965
asS'd'
I5477
sa(dp10954
S'a'
I28838
sS's'
I16
sS'm'
(lp10955
I1
aI1
aI3
aI9
aI13
aI19
aI49
aI183
aI497
aI519
aI1433
aI519
aI4317
aI2359
aI10349
aI63789
asS'd'
I5478
sa(dp10956
S'a'
I28850
sS's'
I16
sS'm'
(lp10957
I1
aI3
aI5
aI9
aI29
aI45
aI55
aI163
aI189
aI533
aI275
aI237
aI5453
aI8895
aI6377
aI14117
asS'd'
I5479
sa(dp10958
S'a'
I28891
sS's'
I16
sS'm'
(lp10959
I1
aI3
aI7
aI5
aI25
aI3
aI111
aI241
aI139
aI383
aI689
aI3481
aI2557
aI11163
aI5275
aI14671
asS'd'
I5480
sa(dp10960
S'a'
I28897
sS's'
I16
sS'm'
(lp10961
I1
aI3
aI3
aI9
aI3
aI5
aI5
aI141
aI507
aI331
aI645
aI1957
aI5857
aI2083
aI24717
aI11131
asS'd'
I5481
sa(dp10962
S'a'
I28932
sS's'
I16
sS'm'
(lp10963
I1
aI1
aI5
aI1
aI11
aI49
aI113
aI45
aI491
aI945
aI1467
aI3485
aI6369
aI15983
aI14489
aI12679
asS'd'
I5482
sa(dp10964
S'a'
I28975
sS's'
I16
sS'm'
(lp10965
I1
aI3
aI7
aI9
aI11
aI41
aI77
aI127
aI147
aI635
aI1757
aI587
aI7423
aI4233
aI14875
aI30531
asS'd'
I5483
sa(dp10966
S'a'
I28998
sS's'
I16
sS'm'
(lp10967
I1
aI3
aI3
aI9
aI17
aI29
aI21
aI249
aI155
aI441
aI1443
aI2093
aI1967
aI2117
aI5815
aI3857
asS'd'
I5484
sa(dp10968
S'a'
I29052
sS's'
I16
sS'm'
(lp10969
I1
aI3
aI5
aI3
aI11
aI55
aI75
aI157
aI105
aI507
aI309
aI3737
aI2645
aI7547
aI29373
aI62775
asS'd'
I5485
sa(dp10970
S'a'
I29090
sS's'
I16
sS'm'
(lp10971
I1
aI1
aI3
aI3
aI11
aI29
aI49
aI241
aI21
aI653
aI1273
aI715
aI8123
aI14241
aI25257
aI1681
asS'd'
I5486
sa(dp10972
S'a'
I29096
sS's'
I16
sS'm'
(lp10973
I1
aI1
aI7
aI5
aI11
aI31
aI33
aI215
aI243
aI369
aI247
aI3365
aI4065
aI9389
aI32457
aI58393
asS'd'
I5487
sa(dp10974
S'a'
I29104
sS's'
I16
sS'm'
(lp10975
I1
aI3
aI5
aI3
aI31
aI55
aI51
aI201
aI439
aI835
aI1805
aI25
aI7987
aI10611
aI26893
aI43663
asS'd'
I5488
sa(dp10976
S'a'
I29113
sS's'
I16
sS'm'
(lp10977
I1
aI1
aI5
aI9
aI27
aI51
aI29
aI31
aI17
aI163
aI71
aI603
aI3083
aI12439
aI11043
aI6471
asS'd'
I5489
sa(dp10978
S'a'
I29133
sS's'
I16
sS'm'
(lp10979
I1
aI1
aI5
aI7
aI13
aI1
aI91
aI109
aI213
aI721
aI1345
aI3037
aI3047
aI5209
aI15559
aI17467
asS'd'
I5490
sa(dp10980
S'a'
I29142
sS's'
I16
sS'm'
(lp10981
I1
aI1
aI3
aI9
aI19
aI37
aI93
aI185
aI107
aI859
aI501
aI3843
aI1631
aI4389
aI2215
aI52225
asS'd'
I5491
sa(dp10982
S'a'
I29170
sS's'
I16
sS'm'
(lp10983
I1
aI1
aI3
aI3
aI25
aI5
aI119
aI17
aI33
aI841
aI997
aI439
aI6135
aI7405
aI8445
aI40087
asS'd'
I5492
sa(dp10984
S'a'
I29192
sS's'
I16
sS'm'
(lp10985
I1
aI1
aI7
aI15
aI19
aI17
aI101
aI43
aI423
aI647
aI29
aI1143
aI3259
aI7807
aI15929
aI809
asS'd'
I5493
sa(dp10986
S'a'
I29221
sS's'
I16
sS'm'
(lp10987
I1
aI1
aI7
aI13
aI21
aI57
aI83
aI101
aI183
aI309
aI171
aI3173
aI7919
aI7263
aI29403
aI11055
asS'd'
I5494
sa(dp10988
S'a'
I29236
sS's'
I16
sS'm'
(lp10989
I1
aI1
aI1
aI13
aI3
aI1
aI57
aI15
aI435
aI713
aI459
aI847
aI3115
aI191
aI19809
aI43037
asS'd'
I5495
sa(dp10990
S'a'
I29246
sS's'
I16
sS'm'
(lp10991
I1
aI1
aI7
aI7
aI17
aI45
aI91
aI117
aI157
aI647
aI121
aI4091
aI3611
aI14169
aI19883
aI9069
asS'd'
I5496
sa(dp10992
S'a'
I29293
sS's'
I16
sS'm'
(lp10993
I1
aI1
aI7
aI7
aI1
aI47
aI21
aI253
aI419
aI157
aI549
aI2105
aI4475
aI3127
aI3939
aI5809
asS'd'
I5497
sa(dp10994
S'a'
I29305
sS's'
I16
sS'm'
(lp10995
I1
aI1
aI5
aI7
aI15
aI7
aI71
aI195
aI87
aI757
aI77
aI1391
aI151
aI12995
aI26403
aI17789
asS'd'
I5498
sa(dp10996
S'a'
I29312
sS's'
I16
sS'm'
(lp10997
I1
aI1
aI1
aI15
aI15
aI3
aI79
aI43
aI475
aI263
aI1195
aI2385
aI5955
aI7039
aI15625
aI19263
asS'd'
I5499
sa(dp10998
S'a'
I29339
sS's'
I16
sS'm'
(lp10999
I1
aI1
aI5
aI13
aI13
aI29
aI5
aI29
aI489
aI929
aI2027
aI2771
aI6899
aI14173
aI13747
aI1019
asS'd'
I5500
sa(dp11000
S'a'
I29365
sS's'
I16
sS'm'
(lp11001
I1
aI1
aI7
aI1
aI5
aI45
aI37
aI85
aI221
aI871
aI627
aI3445
aI4853
aI4243
aI21651
aI30201
asS'd'
I5501
sa(dp11002
S'a'
I29389
sS's'
I16
sS'm'
(lp11003
I1
aI3
aI7
aI11
aI9
aI49
aI73
aI245
aI161
aI321
aI579
aI2641
aI6653
aI5513
aI11555
aI53091
asS'd'
I5502
sa(dp11004
S'a'
I29402
sS's'
I16
sS'm'
(lp11005
I1
aI1
aI7
aI7
aI25
aI63
aI101
aI179
aI497
aI113
aI9
aI549
aI5341
aI6097
aI13305
aI52421
asS'd'
I5503
sa(dp11006
S'a'
I29423
sS's'
I16
sS'm'
(lp11007
I1
aI3
aI3
aI7
aI21
aI7
aI89
aI79
aI137
aI651
aI189
aI3025
aI1403
aI4559
aI32611
aI1857
asS'd'
I5504
sa(dp11008
S'a'
I29443
sS's'
I16
sS'm'
(lp11009
I1
aI3
aI1
aI13
aI27
aI55
aI61
aI135
aI81
aI195
aI799
aI3477
aI4873
aI2691
aI29769
aI59033
asS'd'
I5505
sa(dp11010
S'a'
I29445
sS's'
I16
sS'm'
(lp11011
I1
aI3
aI3
aI15
aI29
aI11
aI7
aI11
aI151
aI649
aI1511
aI2327
aI6921
aI12911
aI3571
aI35039
asS'd'
I5506
sa(dp11012
S'a'
I29463
sS's'
I16
sS'm'
(lp11013
I1
aI1
aI5
aI11
aI25
aI19
aI49
aI133
aI455
aI373
aI1827
aI3619
aI2127
aI3365
aI11327
aI52785
asS'd'
I5507
sa(dp11014
S'a'
I29473
sS's'
I16
sS'm'
(lp11015
I1
aI3
aI5
aI1
aI9
aI19
aI107
aI171
aI205
aI93
aI1557
aI2693
aI4297
aI4415
aI20407
aI19221
asS'd'
I5508
sa(dp11016
S'a'
I29493
sS's'
I16
sS'm'
(lp11017
I1
aI3
aI3
aI11
aI15
aI45
aI37
aI143
aI61
aI759
aI2047
aI2465
aI3923
aI9477
aI30831
aI46377
asS'd'
I5509
sa(dp11018
S'a'
I29506
sS's'
I16
sS'm'
(lp11019
I1
aI3
aI7
aI11
aI17
aI51
aI117
aI129
aI77
aI579
aI1167
aI1575
aI1967
aI10099
aI22137
aI31431
asS'd'
I5510
sa(dp11020
S'a'
I29518
sS's'
I16
sS'm'
(lp11021
I1
aI1
aI5
aI13
aI31
aI61
aI67
aI37
aI49
aI283
aI235
aI783
aI7353
aI5149
aI12245
aI18725
asS'd'
I5511
sa(dp11022
S'a'
I29532
sS's'
I16
sS'm'
(lp11023
I1
aI1
aI5
aI3
aI17
aI33
aI35
aI83
aI359
aI253
aI1911
aI913
aI6481
aI4635
aI24223
aI19693
asS'd'
I5512
sa(dp11024
S'a'
I29560
sS's'
I16
sS'm'
(lp11025
I1
aI1
aI1
aI13
aI19
aI15
aI81
aI131
aI417
aI969
aI1911
aI2829
aI3097
aI5333
aI11175
aI52269
asS'd'
I5513
sa(dp11026
S'a'
I29590
sS's'
I16
sS'm'
(lp11027
I1
aI3
aI7
aI15
aI5
aI39
aI19
aI205
aI329
aI83
aI1473
aI3259
aI6409
aI12297
aI30557
aI40917
asS'd'
I5514
sa(dp11028
S'a'
I29594
sS's'
I16
sS'm'
(lp11029
I1
aI3
aI1
aI15
aI17
aI33
aI123
aI185
aI501
aI299
aI621
aI929
aI5797
aI10539
aI12321
aI61043
asS'd'
I5515
sa(dp11030
S'a'
I29637
sS's'
I16
sS'm'
(lp11031
I1
aI3
aI3
aI1
aI7
aI51
aI119
aI19
aI17
aI203
aI373
aI2145
aI2367
aI9965
aI28071
aI50083
asS'd'
I5516
sa(dp11032
S'a'
I29647
sS's'
I16
sS'm'
(lp11033
I1
aI1
aI1
aI5
aI1
aI35
aI43
aI243
aI91
aI793
aI1299
aI2705
aI7987
aI1291
aI10147
aI17863
asS'd'
I5517
sa(dp11034
S'a'
I29650
sS's'
I16
sS'm'
(lp11035
I1
aI3
aI5
aI15
aI27
aI13
aI99
aI33
aI179
aI479
aI897
aI1113
aI1639
aI12321
aI23987
aI36219
asS'd'
I5518
sa(dp11036
S'a'
I29655
sS's'
I16
sS'm'
(lp11037
I1
aI1
aI5
aI9
aI29
aI41
aI85
aI9
aI389
aI583
aI293
aI1727
aI2575
aI13767
aI15443
aI40027
asS'd'
I5519
sa(dp11038
S'a'
I29661
sS's'
I16
sS'm'
(lp11039
I1
aI1
aI7
aI11
aI29
aI33
aI93
aI115
aI51
aI747
aI1569
aI3557
aI869
aI1991
aI29877
aI44131
asS'd'
I5520
sa(dp11040
S'a'
I29680
sS's'
I16
sS'm'
(lp11041
I1
aI3
aI7
aI7
aI29
aI11
aI33
aI137
aI411
aI689
aI1815
aI1789
aI6557
aI5973
aI19445
aI49449
asS'd'
I5521
sa(dp11042
S'a'
I29721
sS's'
I16
sS'm'
(lp11043
I1
aI1
aI5
aI1
aI17
aI3
aI77
aI55
aI351
aI325
aI983
aI3935
aI819
aI14127
aI18893
aI62553
asS'd'
I5522
sa(dp11044
S'a'
I29751
sS's'
I16
sS'm'
(lp11045
I1
aI3
aI3
aI1
aI15
aI33
aI25
aI159
aI135
aI385
aI837
aI3615
aI1649
aI1687
aI3421
aI47579
asS'd'
I5523
sa(dp11046
S'a'
I29755
sS's'
I16
sS'm'
(lp11047
I1
aI3
aI1
aI7
aI17
aI25
aI125
aI169
aI469
aI919
aI1789
aI863
aI2827
aI949
aI21347
aI10517
asS'd'
I5524
sa(dp11048
S'a'
I29760
sS's'
I16
sS'm'
(lp11049
I1
aI3
aI1
aI11
aI27
aI19
aI45
aI255
aI175
aI483
aI1073
aI3779
aI611
aI2809
aI179
aI19767
asS'd'
I5525
sa(dp11050
S'a'
I29772
sS's'
I16
sS'm'
(lp11051
I1
aI1
aI3
aI1
aI21
aI61
aI47
aI171
aI179
aI85
aI61
aI1209
aI4005
aI11439
aI8477
aI27229
asS'd'
I5526
sa(dp11052
S'a'
I29778
sS's'
I16
sS'm'
(lp11053
I1
aI1
aI1
aI1
aI3
aI1
aI43
aI159
aI261
aI411
aI1449
aI1621
aI3681
aI3465
aI24029
aI3493
asS'd'
I5527
sa(dp11054
S'a'
I29799
sS's'
I16
sS'm'
(lp11055
I1
aI3
aI1
aI11
aI5
aI13
aI9
aI23
aI369
aI769
aI363
aI3329
aI409
aI13151
aI30269
aI9621
asS'd'
I5528
sa(dp11056
S'a'
I29824
sS's'
I16
sS'm'
(lp11057
I1
aI1
aI5
aI1
aI13
aI39
aI121
aI39
aI295
aI981
aI1151
aI4039
aI8179
aI5007
aI25527
aI1249
asS'd'
I5529
sa(dp11058
S'a'
I29841
sS's'
I16
sS'm'
(lp11059
I1
aI3
aI7
aI5
aI17
aI21
aI47
aI233
aI211
aI349
aI643
aI109
aI7553
aI11453
aI30967
aI30959
asS'd'
I5530
sa(dp11060
S'a'
I29842
sS's'
I16
sS'm'
(lp11061
I1
aI1
aI5
aI3
aI31
aI39
aI105
aI137
aI487
aI855
aI107
aI1567
aI2385
aI2889
aI25777
aI33709
asS'd'
I5531
sa(dp11062
S'a'
I29853
sS's'
I16
sS'm'
(lp11063
I1
aI1
aI1
aI9
aI1
aI7
aI9
aI69
aI465
aI965
aI355
aI299
aI3327
aI14997
aI14599
aI2241
asS'd'
I5532
sa(dp11064
S'a'
I29867
sS's'
I16
sS'm'
(lp11065
I1
aI3
aI5
aI11
aI5
aI39
aI69
aI203
aI367
aI611
aI199
aI3931
aI5039
aI8683
aI8675
aI49151
asS'd'
I5533
sa(dp11066
S'a'
I29949
sS's'
I16
sS'm'
(lp11067
I1
aI1
aI7
aI13
aI31
aI35
aI101
aI213
aI273
aI827
aI203
aI2773
aI4131
aI1397
aI15311
aI62903
asS'd'
I5534
sa(dp11068
S'a'
I29950
sS's'
I16
sS'm'
(lp11069
I1
aI3
aI7
aI9
aI23
aI41
aI33
aI213
aI411
aI965
aI563
aI3035
aI247
aI15019
aI20429
aI61081
asS'd'
I5535
sa(dp11070
S'a'
I29964
sS's'
I16
sS'm'
(lp11071
I1
aI1
aI5
aI5
aI5
aI1
aI1
aI203
aI27
aI199
aI67
aI1301
aI7831
aI12839
aI2777
aI6325
asS'd'
I5536
sa(dp11072
S'a'
I29967
sS's'
I16
sS'm'
(lp11073
I1
aI3
aI1
aI11
aI27
aI3
aI11
aI173
aI9
aI121
aI1701
aI2741
aI29
aI16319
aI15849
aI11989
asS'd'
I5537
sa(dp11074
S'a'
I29985
sS's'
I16
sS'm'
(lp11075
I1
aI1
aI5
aI13
aI17
aI49
aI125
aI153
aI261
aI603
aI1617
aI3967
aI6083
aI7745
aI19683
aI49885
asS'd'
I5538
sa(dp11076
S'a'
I29992
sS's'
I16
sS'm'
(lp11077
I1
aI3
aI3
aI7
aI23
aI13
aI39
aI169
aI135
aI361
aI579
aI1443
aI7615
aI2389
aI5669
aI651
asS'd'
I5539
sa(dp11078
S'a'
I30000
sS's'
I16
sS'm'
(lp11079
I1
aI3
aI5
aI9
aI31
aI19
aI81
aI83
aI483
aI93
aI1895
aI2285
aI7771
aI8281
aI8353
aI39677
asS'd'
I5540
sa(dp11080
S'a'
I30020
sS's'
I16
sS'm'
(lp11081
I1
aI1
aI7
aI7
aI23
aI51
aI127
aI25
aI101
aI611
aI1095
aI3013
aI2685
aI8153
aI22629
aI53355
asS'd'
I5541
sa(dp11082
S'a'
I30024
sS's'
I16
sS'm'
(lp11083
I1
aI1
aI1
aI11
aI11
aI37
aI35
aI127
aI317
aI877
aI1591
aI401
aI4121
aI9945
aI12121
aI28257
asS'd'
I5542
sa(dp11084
S'a'
I30030
sS's'
I16
sS'm'
(lp11085
I1
aI3
aI5
aI11
aI23
aI9
aI43
aI135
aI37
aI405
aI2009
aI2903
aI3065
aI6591
aI8473
aI58231
asS'd'
I5543
sa(dp11086
S'a'
I30066
sS's'
I16
sS'm'
(lp11087
I1
aI1
aI3
aI11
aI21
aI45
aI21
aI205
aI425
aI891
aI357
aI2609
aI495
aI7541
aI2161
aI37853
asS'd'
I5544
sa(dp11088
S'a'
I30071
sS's'
I16
sS'm'
(lp11089
I1
aI3
aI1
aI1
aI25
aI9
aI113
aI243
aI317
aI491
aI997
aI2023
aI5869
aI13643
aI11483
aI6733
asS'd'
I5545
sa(dp11090
S'a'
I30078
sS's'
I16
sS'm'
(lp11091
I1
aI3
aI1
aI15
aI13
aI3
aI75
aI25
aI409
aI421
aI1817
aI857
aI4575
aI12559
aI1211
aI62177
asS'd'
I5546
sa(dp11092
S'a'
I30082
sS's'
I16
sS'm'
(lp11093
I1
aI1
aI3
aI7
aI17
aI35
aI115
aI195
aI217
aI223
aI1195
aI749
aI5619
aI7265
aI7369
aI46907
asS'd'
I5547
sa(dp11094
S'a'
I30096
sS's'
I16
sS'm'
(lp11095
I1
aI1
aI1
aI13
aI5
aI57
aI117
aI161
aI121
aI533
aI987
aI3959
aI5047
aI15213
aI15811
aI41841
asS'd'
I5548
sa(dp11096
S'a'
I30101
sS's'
I16
sS'm'
(lp11097
I1
aI3
aI7
aI1
aI19
aI55
aI97
aI191
aI217
aI75
aI1881
aI3351
aI3737
aI12179
aI22875
aI28767
asS'd'
I5549
sa(dp11098
S'a'
I30102
sS's'
I16
sS'm'
(lp11099
I1
aI3
aI1
aI9
aI15
aI41
aI9
aI97
aI491
aI31
aI1191
aI963
aI875
aI8259
aI2723
aI9503
asS'd'
I5550
sa(dp11100
S'a'
I30122
sS's'
I16
sS'm'
(lp11101
I1
aI3
aI7
aI9
aI3
aI17
aI21
aI71
aI1
aI523
aI2031
aI3469
aI3181
aI8707
aI6093
aI8837
asS'd'
I5551
sa(dp11102
S'a'
I30141
sS's'
I16
sS'm'
(lp11103
I1
aI3
aI5
aI3
aI5
aI1
aI11
aI91
aI33
aI37
aI643
aI85
aI4325
aI4293
aI8351
aI28311
asS'd'
I5552
sa(dp11104
S'a'
I30159
sS's'
I16
sS'm'
(lp11105
I1
aI3
aI7
aI5
aI15
aI45
aI47
aI183
aI391
aI113
aI493
aI3607
aI2541
aI13521
aI31613
aI36049
asS'd'
I5553
sa(dp11106
S'a'
I30168
sS's'
I16
sS'm'
(lp11107
I1
aI1
aI3
aI9
aI15
aI33
aI115
aI69
aI289
aI217
aI1875
aI1339
aI4995
aI9073
aI6909
aI15977
asS'd'
I5554
sa(dp11108
S'a'
I30177
sS's'
I16
sS'm'
(lp11109
I1
aI1
aI7
aI3
aI9
aI29
aI39
aI219
aI119
aI369
aI893
aI1293
aI4511
aI15703
aI11093
aI30259
asS'd'
I5555
sa(dp11110
S'a'
I30183
sS's'
I16
sS'm'
(lp11111
I1
aI1
aI5
aI13
aI19
aI9
aI17
aI75
aI149
aI415
aI35
aI97
aI563
aI1689
aI18311
aI54291
asS'd'
I5556
sa(dp11112
S'a'
I30197
sS's'
I16
sS'm'
(lp11113
I1
aI1
aI7
aI3
aI17
aI15
aI71
aI29
aI25
aI883
aI1801
aI1675
aI5585
aI9413
aI3813
aI26673
asS'd'
I5557
sa(dp11114
S'a'
I30213
sS's'
I16
sS'm'
(lp11115
I1
aI1
aI3
aI15
aI5
aI13
aI31
aI41
aI311
aI411
aI573
aI281
aI8075
aI7163
aI11817
aI29121
asS'd'
I5558
sa(dp11116
S'a'
I30231
sS's'
I16
sS'm'
(lp11117
I1
aI1
aI7
aI9
aI7
aI57
aI15
aI141
aI337
aI123
aI269
aI3737
aI6455
aI2539
aI13655
aI59809
asS'd'
I5559
sa(dp11118
S'a'
I30232
sS's'
I16
sS'm'
(lp11119
I1
aI3
aI1
aI15
aI15
aI23
aI111
aI51
aI429
aI483
aI1567
aI1317
aI8057
aI1609
aI30181
aI35687
asS'd'
I5560
sa(dp11120
S'a'
I30241
sS's'
I16
sS'm'
(lp11121
I1
aI3
aI7
aI9
aI25
aI43
aI67
aI13
aI319
aI587
aI1827
aI443
aI2031
aI8563
aI16173
aI58667
asS'd'
I5561
sa(dp11122
S'a'
I30253
sS's'
I16
sS'm'
(lp11123
I1
aI3
aI5
aI13
aI11
aI63
aI89
aI105
aI377
aI257
aI7
aI4077
aI5091
aI5125
aI25
aI39033
asS'd'
I5562
sa(dp11124
S'a'
I30259
sS's'
I16
sS'm'
(lp11125
I1
aI3
aI3
aI1
aI9
aI29
aI7
aI87
aI239
aI469
aI1851
aI1711
aI5797
aI7137
aI11405
aI20175
asS'd'
I5563
sa(dp11126
S'a'
I30262
sS's'
I16
sS'm'
(lp11127
I1
aI3
aI3
aI1
aI13
aI17
aI101
aI209
aI301
aI95
aI1181
aI3091
aI4451
aI1241
aI17057
aI335
asS'd'
I5564
sa(dp11128
S'a'
I30268
sS's'
I16
sS'm'
(lp11129
I1
aI1
aI1
aI9
aI31
aI7
aI81
aI161
aI391
aI677
aI115
aI141
aI5375
aI7279
aI1887
aI1645
asS'd'
I5565
sa(dp11130
S'a'
I30297
sS's'
I16
sS'm'
(lp11131
I1
aI1
aI1
aI11
aI27
aI61
aI3
aI195
aI189
aI409
aI1747
aI331
aI2931
aI9535
aI1369
aI47233
asS'd'
I5566
sa(dp11132
S'a'
I30316
sS's'
I16
sS'm'
(lp11133
I1
aI3
aI5
aI15
aI7
aI15
aI105
aI255
aI491
aI689
aI97
aI1131
aI3459
aI7147
aI27541
aI62307
asS'd'
I5567
sa(dp11134
S'a'
I30322
sS's'
I16
sS'm'
(lp11135
I1
aI3
aI5
aI9
aI5
aI23
aI1
aI209
aI233
aI717
aI1919
aI1835
aI5073
aI10403
aI28979
aI1945
asS'd'
I5568
sa(dp11136
S'a'
I30344
sS's'
I16
sS'm'
(lp11137
I1
aI1
aI3
aI9
aI3
aI35
aI107
aI209
aI255
aI447
aI227
aI273
aI443
aI9367
aI24105
aI34095
asS'd'
I5569
sa(dp11138
S'a'
I30350
sS's'
I16
sS'm'
(lp11139
I1
aI1
aI3
aI11
aI3
aI33
aI5
aI165
aI83
aI787
aI1555
aI31
aI4351
aI16301
aI27453
aI56739
asS'd'
I5570
sa(dp11140
S'a'
I30355
sS's'
I16
sS'm'
(lp11141
I1
aI1
aI5
aI5
aI29
aI9
aI127
aI253
aI281
aI487
aI441
aI1129
aI2811
aI9113
aI28855
aI57117
asS'd'
I5571
sa(dp11142
S'a'
I30429
sS's'
I16
sS'm'
(lp11143
I1
aI1
aI1
aI13
aI13
aI1
aI17
aI143
aI121
aI917
aI1571
aI3777
aI2297
aI3691
aI3001
aI42327
asS'd'
I5572
sa(dp11144
S'a'
I30445
sS's'
I16
sS'm'
(lp11145
I1
aI1
aI5
aI1
aI25
aI7
aI41
aI245
aI241
aI929
aI1203
aI3755
aI7113
aI9333
aI22549
aI12253
asS'd'
I5573
sa(dp11146
S'a'
I30477
sS's'
I16
sS'm'
(lp11147
I1
aI3
aI1
aI11
aI1
aI13
aI69
aI73
aI285
aI975
aI1331
aI3411
aI7777
aI3489
aI2763
aI44297
asS'd'
I5574
sa(dp11148
S'a'
I30513
sS's'
I16
sS'm'
(lp11149
I1
aI3
aI5
aI11
aI3
aI37
aI21
aI105
aI153
aI307
aI989
aI627
aI3127
aI6169
aI10573
aI22139
asS'd'
I5575
sa(dp11150
S'a'
I30520
sS's'
I16
sS'm'
(lp11151
I1
aI3
aI5
aI15
aI11
aI11
aI39
aI21
aI355
aI437
aI547
aI2283
aI6443
aI5561
aI6367
aI53899
asS'd'
I5576
sa(dp11152
S'a'
I30540
sS's'
I16
sS'm'
(lp11153
I1
aI1
aI1
aI9
aI25
aI51
aI97
aI175
aI131
aI207
aI1367
aI2561
aI7455
aI8289
aI5877
aI4383
asS'd'
I5577
sa(dp11154
S'a'
I30551
sS's'
I16
sS'm'
(lp11155
I1
aI3
aI7
aI1
aI29
aI17
aI7
aI1
aI43
aI831
aI591
aI2145
aI975
aI909
aI23107
aI43987
asS'd'
I5578
sa(dp11156
S'a'
I30573
sS's'
I16
sS'm'
(lp11157
I1
aI3
aI5
aI13
aI5
aI47
aI65
aI65
aI439
aI807
aI271
aI1615
aI1873
aI10905
aI30537
aI3337
asS'd'
I5579
sa(dp11158
S'a'
I30609
sS's'
I16
sS'm'
(lp11159
I1
aI1
aI1
aI13
aI29
aI1
aI53
aI5
aI307
aI347
aI1059
aI545
aI1129
aI11883
aI5969
aI50433
asS'd'
I5580
sa(dp11160
S'a'
I30622
sS's'
I16
sS'm'
(lp11161
I1
aI1
aI3
aI5
aI31
aI29
aI63
aI201
aI255
aI803
aI677
aI1499
aI1691
aI14037
aI18251
aI6881
asS'd'
I5581
sa(dp11162
S'a'
I30635
sS's'
I16
sS'm'
(lp11163
I1
aI3
aI1
aI5
aI5
aI13
aI13
aI121
aI505
aI855
aI467
aI2803
aI3297
aI4689
aI18443
aI60757
asS'd'
I5582
sa(dp11164
S'a'
I30658
sS's'
I16
sS'm'
(lp11165
I1
aI1
aI5
aI13
aI11
aI19
aI11
aI201
aI101
aI431
aI693
aI1267
aI6909
aI7759
aI2265
aI6125
asS'd'
I5583
sa(dp11166
S'a'
I30667
sS's'
I16
sS'm'
(lp11167
I1
aI1
aI7
aI13
aI9
aI3
aI37
aI209
aI269
aI27
aI1041
aI2587
aI4667
aI11077
aI27009
aI27967
asS'd'
I5584
sa(dp11168
S'a'
I30681
sS's'
I16
sS'm'
(lp11169
I1
aI1
aI5
aI5
aI1
aI5
aI127
aI179
aI463
aI949
aI1525
aI231
aI1201
aI3283
aI9929
aI46677
asS'd'
I5585
sa(dp11170
S'a'
I30684
sS's'
I16
sS'm'
(lp11171
I1
aI3
aI1
aI15
aI9
aI11
aI89
aI129
aI331
aI833
aI1415
aI229
aI2059
aI13145
aI30525
aI33773
asS'd'
I5586
sa(dp11172
S'a'
I30703
sS's'
I16
sS'm'
(lp11173
I1
aI1
aI7
aI15
aI7
aI43
aI95
aI177
aI313
aI989
aI483
aI825
aI1885
aI4535
aI8213
aI39835
asS'd'
I5587
sa(dp11174
S'a'
I30705
sS's'
I16
sS'm'
(lp11175
I1
aI1
aI7
aI3
aI19
aI27
aI45
aI163
aI17
aI523
aI1565
aI3753
aI7433
aI14117
aI8499
aI40177
asS'd'
I5588
sa(dp11176
S'a'
I30715
sS's'
I16
sS'm'
(lp11177
I1
aI3
aI3
aI15
aI23
aI45
aI95
aI31
aI55
aI469
aI383
aI237
aI6287
aI5561
aI20901
aI48259
asS'd'
I5589
sa(dp11178
S'a'
I30742
sS's'
I16
sS'm'
(lp11179
I1
aI1
aI3
aI1
aI23
aI61
aI101
aI185
aI35
aI553
aI463
aI1169
aI2875
aI12491
aI14327
aI47999
asS'd'
I5590
sa(dp11180
S'a'
I30748
sS's'
I16
sS'm'
(lp11181
I1
aI3
aI3
aI13
aI23
aI29
aI77
aI21
aI19
aI3
aI769
aI1943
aI2081
aI9135
aI29767
aI11367
asS'd'
I5591
sa(dp11182
S'a'
I30758
sS's'
I16
sS'm'
(lp11183
I1
aI1
aI5
aI15
aI5
aI11
aI59
aI163
aI355
aI993
aI375
aI3181
aI2675
aI8515
aI17007
aI38467
asS'd'
I5592
sa(dp11184
S'a'
I30767
sS's'
I16
sS'm'
(lp11185
I1
aI1
aI5
aI13
aI19
aI5
aI107
aI83
aI123
aI843
aI413
aI2137
aI7531
aI3833
aI6149
aI55925
asS'd'
I5593
sa(dp11186
S'a'
I30770
sS's'
I16
sS'm'
(lp11187
I1
aI3
aI1
aI13
aI23
aI9
aI41
aI145
aI265
aI591
aI1899
aI3145
aI5255
aI13653
aI12245
aI25367
asS'd'
I5594
sa(dp11188
S'a'
I30772
sS's'
I16
sS'm'
(lp11189
I1
aI1
aI3
aI15
aI1
aI45
aI119
aI79
aI121
aI137
aI1945
aI2041
aI2409
aI1377
aI29501
aI29885
asS'd'
I5595
sa(dp11190
S'a'
I30807
sS's'
I16
sS'm'
(lp11191
I1
aI1
aI7
aI11
aI27
aI57
aI75
aI183
aI341
aI237
aI1909
aI2785
aI5973
aI9965
aI21729
aI45089
asS'd'
I5596
sa(dp11192
S'a'
I30814
sS's'
I16
sS'm'
(lp11193
I1
aI3
aI5
aI7
aI21
aI1
aI41
aI189
aI131
aI1021
aI1375
aI1463
aI5985
aI12499
aI4115
aI9131
asS'd'
I5597
sa(dp11194
S'a'
I30824
sS's'
I16
sS'm'
(lp11195
I1
aI3
aI7
aI15
aI21
aI19
aI59
aI171
aI339
aI841
aI1725
aI2909
aI6437
aI2499
aI17191
aI43275
asS'd'
I5598
sa(dp11196
S'a'
I30857
sS's'
I16
sS'm'
(lp11197
I1
aI3
aI1
aI1
aI15
aI55
aI31
aI199
aI351
aI183
aI1819
aI1873
aI7877
aI12407
aI7881
aI1663
asS'd'
I5599
sa(dp11198
S'a'
I30875
sS's'
I16
sS'm'
(lp11199
I1
aI1
aI3
aI15
aI1
aI61
aI111
aI61
aI115
aI243
aI1281
aI3195
aI1229
aI10973
aI189
aI36049
asS'd'
I5600
sa(dp11200
S'a'
I30931
sS's'
I16
sS'm'
(lp11201
I1
aI1
aI3
aI15
aI13
aI13
aI3
aI49
aI61
aI839
aI1615
aI1853
aI3619
aI7805
aI25441
aI8789
asS'd'
I5601
sa(dp11202
S'a'
I30933
sS's'
I16
sS'm'
(lp11203
I1
aI3
aI1
aI9
aI27
aI43
aI7
aI193
aI397
aI869
aI1079
aI1785
aI6535
aI1801
aI29363
aI59269
asS'd'
I5602
sa(dp11204
S'a'
I30949
sS's'
I16
sS'm'
(lp11205
I1
aI3
aI5
aI5
aI31
aI57
aI37
aI53
aI41
aI871
aI809
aI1235
aI1011
aI12979
aI8749
aI52151
asS'd'
I5603
sa(dp11206
S'a'
I30953
sS's'
I16
sS'm'
(lp11207
I1
aI1
aI7
aI13
aI25
aI59
aI69
aI117
aI463
aI587
aI513
aI297
aI6991
aI5905
aI25737
aI37249
asS'd'
I5604
sa(dp11208
S'a'
I30982
sS's'
I16
sS'm'
(lp11209
I1
aI1
aI5
aI1
aI27
aI19
aI121
aI97
aI349
aI793
aI1971
aI3057
aI4781
aI15841
aI22625
aI58637
asS'd'
I5605
sa(dp11210
S'a'
I31010
sS's'
I16
sS'm'
(lp11211
I1
aI1
aI5
aI5
aI25
aI31
aI11
aI133
aI411
aI239
aI1071
aI3473
aI1733
aI7175
aI31841
aI46665
asS'd'
I5606
sa(dp11212
S'a'
I31012
sS's'
I16
sS'm'
(lp11213
I1
aI3
aI3
aI13
aI19
aI25
aI99
aI175
aI271
aI175
aI1755
aI3597
aI4615
aI15207
aI25573
aI16089
asS'd'
I5607
sa(dp11214
S'a'
I31039
sS's'
I16
sS'm'
(lp11215
I1
aI1
aI7
aI11
aI17
aI19
aI119
aI91
aI505
aI791
aI55
aI2979
aI7463
aI10147
aI23647
aI33283
asS'd'
I5608
sa(dp11216
S'a'
I31041
sS's'
I16
sS'm'
(lp11217
I1
aI3
aI1
aI1
aI21
aI11
aI43
aI173
aI239
aI839
aI1533
aI1559
aI549
aI15621
aI22133
aI46387
asS'd'
I5609
sa(dp11218
S'a'
I31061
sS's'
I16
sS'm'
(lp11219
I1
aI1
aI3
aI13
aI31
aI15
aI73
aI15
aI209
aI267
aI701
aI2899
aI1163
aI10093
aI7727
aI44211
asS'd'
I5610
sa(dp11220
S'a'
I31082
sS's'
I16
sS'm'
(lp11221
I1
aI3
aI1
aI11
aI29
aI21
aI5
aI39
aI421
aI375
aI411
aI3693
aI3901
aI8507
aI10883
aI16189
asS'd'
I5611
sa(dp11222
S'a'
I31087
sS's'
I16
sS'm'
(lp11223
I1
aI3
aI1
aI7
aI13
aI13
aI73
aI167
aI149
aI677
aI1435
aI621
aI2511
aI13813
aI13129
aI55327
asS'd'
I5612
sa(dp11224
S'a'
I31096
sS's'
I16
sS'm'
(lp11225
I1
aI3
aI5
aI15
aI7
aI59
aI83
aI221
aI77
aI357
aI281
aI2689
aI5629
aI5837
aI1701
aI30811
asS'd'
I5613
sa(dp11226
S'a'
I31115
sS's'
I16
sS'm'
(lp11227
I1
aI3
aI3
aI11
aI17
aI1
aI43
aI95
aI473
aI981
aI1487
aI1337
aI905
aI3307
aI22357
aI181
asS'd'
I5614
sa(dp11228
S'a'
I31130
sS's'
I16
sS'm'
(lp11229
I1
aI1
aI3
aI7
aI1
aI27
aI9
aI3
aI489
aI1
aI1265
aI2463
aI539
aI12769
aI825
aI6149
asS'd'
I5615
sa(dp11230
S'a'
I31168
sS's'
I16
sS'm'
(lp11231
I1
aI3
aI3
aI3
aI11
aI27
aI81
aI237
aI411
aI241
aI1613
aI931
aI6397
aI4325
aI29651
aI49003
asS'd'
I5616
sa(dp11232
S'a'
I31171
sS's'
I16
sS'm'
(lp11233
I1
aI3
aI3
aI13
aI1
aI19
aI55
aI73
aI47
aI203
aI1661
aI1245
aI6847
aI2457
aI25427
aI33069
asS'd'
I5617
sa(dp11234
S'a'
I31177
sS's'
I16
sS'm'
(lp11235
I1
aI3
aI7
aI3
aI7
aI47
aI11
aI165
aI391
aI457
aI301
aI1213
aI1913
aI14531
aI7847
aI14811
asS'd'
I5618
sa(dp11236
S'a'
I31180
sS's'
I16
sS'm'
(lp11237
I1
aI3
aI1
aI9
aI1
aI9
aI57
aI203
aI15
aI733
aI1131
aI2751
aI5869
aI3165
aI21497
aI28881
asS'd'
I5619
sa(dp11238
S'a'
I31191
sS's'
I16
sS'm'
(lp11239
I1
aI3
aI1
aI5
aI9
aI7
aI29
aI85
aI71
aI571
aI469
aI2395
aI2819
aI8443
aI2281
aI50489
asS'd'
I5620
sa(dp11240
S'a'
I31207
sS's'
I16
sS'm'
(lp11241
I1
aI3
aI5
aI11
aI13
aI63
aI47
aI47
aI349
aI21
aI861
aI2217
aI2945
aI6967
aI6605
aI16459
asS'd'
I5621
sa(dp11242
S'a'
I31247
sS's'
I16
sS'm'
(lp11243
I1
aI1
aI7
aI5
aI13
aI3
aI41
aI53
aI409
aI289
aI1225
aI2965
aI5283
aI1785
aI14443
aI51755
asS'd'
I5622
sa(dp11244
S'a'
I31249
sS's'
I16
sS'm'
(lp11245
I1
aI3
aI7
aI13
aI19
aI1
aI29
aI191
aI119
aI37
aI697
aI1909
aI481
aI14157
aI13425
aI60581
asS'd'
I5623
sa(dp11246
S'a'
I31285
sS's'
I16
sS'm'
(lp11247
I1
aI3
aI1
aI13
aI1
aI15
aI105
aI79
aI505
aI681
aI1741
aI3683
aI5775
aI7479
aI11387
aI1321
asS'd'
I5624
sa(dp11248
S'a'
I31303
sS's'
I16
sS'm'
(lp11249
I1
aI1
aI1
aI11
aI9
aI35
aI77
aI73
aI351
aI217
aI2029
aI2845
aI5143
aI5677
aI15465
aI33123
asS'd'
I5625
sa(dp11250
S'a'
I31310
sS's'
I16
sS'm'
(lp11251
I1
aI1
aI3
aI3
aI19
aI49
aI63
aI109
aI335
aI743
aI741
aI1673
aI3311
aI3139
aI25197
aI13793
asS'd'
I5626
sa(dp11252
S'a'
I31337
sS's'
I16
sS'm'
(lp11253
I1
aI3
aI1
aI3
aI29
aI63
aI79
aI1
aI493
aI13
aI1487
aI4015
aI6983
aI1433
aI26023
aI55591
asS'd'
I5627
sa(dp11254
S'a'
I31352
sS's'
I16
sS'm'
(lp11255
I1
aI3
aI3
aI11
aI1
aI25
aI57
aI207
aI309
aI201
aI1513
aI1749
aI3785
aI9217
aI11531
aI40597
asS'd'
I5628
sa(dp11256
S'a'
I31357
sS's'
I16
sS'm'
(lp11257
I1
aI3
aI7
aI13
aI3
aI23
aI69
aI253
aI311
aI773
aI807
aI1063
aI745
aI4843
aI25221
aI55885
asS'd'
I5629
sa(dp11258
S'a'
I31374
sS's'
I16
sS'm'
(lp11259
I1
aI1
aI3
aI11
aI29
aI47
aI67
aI183
aI11
aI259
aI5
aI1935
aI2295
aI8105
aI19139
aI11707
asS'd'
I5630
sa(dp11260
S'a'
I31379
sS's'
I16
sS'm'
(lp11261
I1
aI1
aI3
aI3
aI23
aI3
aI53
aI165
aI255
aI501
aI1547
aI3649
aI5631
aI13307
aI8869
aI5595
asS'd'
I5631
sa(dp11262
S'a'
I31388
sS's'
I16
sS'm'
(lp11263
I1
aI1
aI3
aI5
aI7
aI29
aI37
aI223
aI289
aI925
aI959
aI309
aI1479
aI3141
aI18661
aI52123
asS'd'
I5632
sa(dp11264
S'a'
I31410
sS's'
I16
sS'm'
(lp11265
I1
aI3
aI1
aI1
aI7
aI59
aI101
aI219
aI91
aI793
aI1103
aI1485
aI7547
aI12889
aI19097
aI15613
asS'd'
I5633
sa(dp11266
S'a'
I31416
sS's'
I16
sS'm'
(lp11267
I1
aI1
aI5
aI15
aI1
aI17
aI79
aI83
aI131
aI683
aI1611
aI1635
aI5405
aI9621
aI29489
aI4801
asS'd'
I5634
sa(dp11268
S'a'
I31467
sS's'
I16
sS'm'
(lp11269
I1
aI1
aI5
aI7
aI31
aI63
aI59
aI125
aI401
aI261
aI1445
aI33
aI187
aI12913
aI8639
aI48413
asS'd'
I5635
sa(dp11270
S'a'
I31495
sS's'
I16
sS'm'
(lp11271
I1
aI3
aI3
aI13
aI27
aI37
aI27
aI99
aI379
aI851
aI1311
aI4051
aI5483
aI13935
aI29679
aI30905
asS'd'
I5636
sa(dp11272
S'a'
I31504
sS's'
I16
sS'm'
(lp11273
I1
aI1
aI3
aI1
aI7
aI57
aI79
aI23
aI97
aI561
aI1083
aI2327
aI1545
aI5387
aI12119
aI29717
asS'd'
I5637
sa(dp11274
S'a'
I31507
sS's'
I16
sS'm'
(lp11275
I1
aI1
aI7
aI7
aI9
aI41
aI63
aI165
aI315
aI247
aI89
aI2055
aI7399
aI1399
aI2057
aI39851
asS'd'
I5638
sa(dp11276
S'a'
I31509
sS's'
I16
sS'm'
(lp11277
I1
aI1
aI1
aI15
aI9
aI23
aI7
aI15
aI457
aI669
aI661
aI3269
aI915
aI3475
aI15845
aI59769
asS'd'
I5639
sa(dp11278
S'a'
I31514
sS's'
I16
sS'm'
(lp11279
I1
aI3
aI7
aI15
aI17
aI53
aI83
aI5
aI457
aI103
aI1297
aI2413
aI1095
aI7711
aI27935
aI56357
asS'd'
I5640
sa(dp11280
S'a'
I31516
sS's'
I16
sS'm'
(lp11281
I1
aI1
aI3
aI5
aI17
aI3
aI81
aI23
aI165
aI341
aI781
aI3583
aI1751
aI6763
aI13937
aI35331
asS'd'
I5641
sa(dp11282
S'a'
I31530
sS's'
I16
sS'm'
(lp11283
I1
aI1
aI5
aI11
aI31
aI21
aI7
aI63
aI369
aI867
aI573
aI45
aI2781
aI4507
aI21553
aI51933
asS'd'
I5642
sa(dp11284
S'a'
I31555
sS's'
I16
sS'm'
(lp11285
I1
aI1
aI5
aI15
aI1
aI37
aI85
aI133
aI489
aI733
aI1471
aI2089
aI979
aI7723
aI7339
aI59595
asS'd'
I5643
sa(dp11286
S'a'
I31567
sS's'
I16
sS'm'
(lp11287
I1
aI1
aI1
aI1
aI7
aI3
aI3
aI77
aI137
aI1009
aI481
aI1343
aI397
aI15865
aI21701
aI37509
asS'd'
I5644
sa(dp11288
S'a'
I31570
sS's'
I16
sS'm'
(lp11289
I1
aI3
aI7
aI5
aI17
aI57
aI19
aI245
aI249
aI289
aI1847
aI3057
aI4905
aI5905
aI32459
aI41305
asS'd'
I5645
sa(dp11290
S'a'
I31586
sS's'
I16
sS'm'
(lp11291
I1
aI1
aI5
aI1
aI23
aI23
aI1
aI177
aI115
aI337
aI983
aI421
aI3135
aI6319
aI27109
aI59641
asS'd'
I5646
sa(dp11292
S'a'
I31598
sS's'
I16
sS'm'
(lp11293
I1
aI3
aI1
aI5
aI25
aI1
aI63
aI73
aI61
aI967
aI1567
aI2645
aI7347
aI11877
aI28777
aI38507
asS'd'
I5647
sa(dp11294
S'a'
I31605
sS's'
I16
sS'm'
(lp11295
I1
aI1
aI3
aI9
aI5
aI41
aI39
aI101
aI339
aI337
aI1079
aI3861
aI5049
aI5601
aI14377
aI34093
asS'd'
I5648
sa(dp11296
S'a'
I31609
sS's'
I16
sS'm'
(lp11297
I1
aI3
aI7
aI7
aI3
aI47
aI95
aI157
aI167
aI1011
aI1117
aI3669
aI7993
aI11735
aI8505
aI64713
asS'd'
I5649
sa(dp11298
S'a'
I31612
sS's'
I16
sS'm'
(lp11299
I1
aI3
aI1
aI9
aI3
aI33
aI11
aI33
aI65
aI329
aI401
aI2659
aI2851
aI3903
aI29791
aI41613
asS'd'
I5650
sa(dp11300
S'a'
I31626
sS's'
I16
sS'm'
(lp11301
I1
aI1
aI1
aI15
aI15
aI17
aI9
aI69
aI359
aI41
aI1475
aI1919
aI5829
aI2189
aI21295
aI33255
asS'd'
I5651
sa(dp11302
S'a'
I31634
sS's'
I16
sS'm'
(lp11303
I1
aI3
aI1
aI3
aI9
aI23
aI73
aI247
aI399
aI775
aI419
aI3033
aI865
aI12595
aI16345
aI15079
asS'd'
I5652
sa(dp11304
S'a'
I31655
sS's'
I16
sS'm'
(lp11305
I1
aI3
aI1
aI5
aI1
aI17
aI33
aI23
aI419
aI585
aI673
aI929
aI6955
aI10247
aI12647
aI29107
asS'd'
I5653
sa(dp11306
S'a'
I31681
sS's'
I16
sS'm'
(lp11307
I1
aI3
aI3
aI13
aI9
aI33
aI11
aI13
aI127
aI529
aI1219
aI2401
aI6459
aI14745
aI5123
aI53023
asS'd'
I5654
sa(dp11308
S'a'
I31705
sS's'
I16
sS'm'
(lp11309
I1
aI3
aI5
aI11
aI23
aI11
aI5
aI19
aI281
aI121
aI1671
aI2171
aI4545
aI10691
aI24875
aI28849
asS'd'
I5655
sa(dp11310
S'a'
I31706
sS's'
I16
sS'm'
(lp11311
I1
aI3
aI1
aI3
aI13
aI25
aI85
aI131
aI127
aI977
aI1599
aI3319
aI3107
aI3185
aI4879
aI3455
asS'd'
I5656
sa(dp11312
S'a'
I31718
sS's'
I16
sS'm'
(lp11313
I1
aI1
aI5
aI1
aI3
aI13
aI77
aI15
aI133
aI185
aI1319
aI727
aI2181
aI12175
aI28017
aI28023
asS'd'
I5657
sa(dp11314
S'a'
I31735
sS's'
I16
sS'm'
(lp11315
I1
aI3
aI7
aI5
aI29
aI51
aI113
aI203
aI331
aI847
aI1
aI3445
aI3669
aI7711
aI13647
aI58651
asS'd'
I5658
sa(dp11316
S'a'
I31741
sS's'
I16
sS'm'
(lp11317
I1
aI3
aI1
aI3
aI31
aI27
aI35
aI199
aI491
aI839
aI1275
aI3385
aI4743
aI821
aI26259
aI11345
asS'd'
I5659
sa(dp11318
S'a'
I31744
sS's'
I16
sS'm'
(lp11319
I1
aI1
aI7
aI9
aI21
aI47
aI9
aI67
aI119
aI985
aI127
aI1987
aI5451
aI6403
aI26183
aI8349
asS'd'
I5660
sa(dp11320
S'a'
I31762
sS's'
I16
sS'm'
(lp11321
I1
aI3
aI5
aI1
aI19
aI3
aI91
aI217
aI301
aI595
aI1789
aI735
aI4993
aI229
aI18033
aI59625
asS'd'
I5661
sa(dp11322
S'a'
I31774
sS's'
I16
sS'm'
(lp11323
I1
aI3
aI3
aI3
aI11
aI25
aI103
aI211
aI117
aI9
aI773
aI1521
aI2265
aI8277
aI23179
aI22433
asS'd'
I5662
sa(dp11324
S'a'
I31864
sS's'
I16
sS'm'
(lp11325
I1
aI1
aI7
aI9
aI3
aI27
aI63
aI255
aI175
aI699
aI293
aI2409
aI3155
aI285
aI8663
aI53503
asS'd'
I5663
sa(dp11326
S'a'
I31874
sS's'
I16
sS'm'
(lp11327
I1
aI1
aI5
aI7
aI27
aI23
aI63
aI213
aI323
aI697
aI1541
aI3497
aI2985
aI12389
aI11155
aI26217
asS'd'
I5664
sa(dp11328
S'a'
I31900
sS's'
I16
sS'm'
(lp11329
I1
aI3
aI1
aI3
aI31
aI7
aI47
aI207
aI185
aI873
aI1063
aI1055
aI205
aI12469
aI23505
aI56245
asS'd'
I5665
sa(dp11330
S'a'
I31910
sS's'
I16
sS'm'
(lp11331
I1
aI3
aI7
aI13
aI31
aI17
aI47
aI95
aI91
aI483
aI1997
aI3273
aI445
aI2601
aI15219
aI10997
asS'd'
I5666
sa(dp11332
S'a'
I31928
sS's'
I16
sS'm'
(lp11333
I1
aI3
aI3
aI5
aI29
aI45
aI29
aI83
aI457
aI823
aI1395
aI1411
aI1879
aI9409
aI11609
aI32001
asS'd'
I5667
sa(dp11334
S'a'
I31965
sS's'
I16
sS'm'
(lp11335
I1
aI3
aI5
aI11
aI21
aI11
aI43
aI73
aI159
aI137
aI29
aI1957
aI815
aI5077
aI16127
aI42199
asS'd'
I5668
sa(dp11336
S'a'
I31976
sS's'
I16
sS'm'
(lp11337
I1
aI3
aI5
aI13
aI9
aI59
aI47
aI215
aI293
aI807
aI309
aI1951
aI2285
aI9287
aI1019
aI49501
asS'd'
I5669
sa(dp11338
S'a'
I32016
sS's'
I16
sS'm'
(lp11339
I1
aI1
aI7
aI13
aI31
aI7
aI95
aI189
aI233
aI363
aI1039
aI1675
aI1715
aI9049
aI8537
aI31051
asS'd'
I5670
sa(dp11340
S'a'
I32032
sS's'
I16
sS'm'
(lp11341
I1
aI3
aI7
aI9
aI23
aI35
aI125
aI251
aI107
aI401
aI1113
aI3585
aI6331
aI2363
aI27889
aI28877
asS'd'
I5671
sa(dp11342
S'a'
I32037
sS's'
I16
sS'm'
(lp11343
I1
aI1
aI7
aI13
aI9
aI1
aI13
aI69
aI257
aI369
aI547
aI1595
aI1823
aI9553
aI25653
aI31181
asS'd'
I5672
sa(dp11344
S'a'
I32062
sS's'
I16
sS'm'
(lp11345
I1
aI1
aI7
aI11
aI9
aI43
aI3
aI93
aI69
aI1019
aI1935
aI3297
aI47
aI7101
aI1037
aI63473
asS'd'
I5673
sa(dp11346
S'a'
I32069
sS's'
I16
sS'm'
(lp11347
I1
aI1
aI7
aI5
aI21
aI9
aI97
aI105
aI405
aI893
aI1673
aI3783
aI2965
aI7329
aI4549
aI25433
asS'd'
I5674
sa(dp11348
S'a'
I32115
sS's'
I16
sS'm'
(lp11349
I1
aI1
aI5
aI13
aI5
aI17
aI31
aI123
aI415
aI173
aI1333
aI2245
aI1557
aI16011
aI28321
aI4039
asS'd'
I5675
sa(dp11350
S'a'
I32128
sS's'
I16
sS'm'
(lp11351
I1
aI1
aI5
aI9
aI15
aI3
aI27
aI79
aI511
aI39
aI945
aI49
aI3231
aI9199
aI21327
aI11183
asS'd'
I5676
sa(dp11352
S'a'
I32171
sS's'
I16
sS'm'
(lp11353
I1
aI3
aI3
aI9
aI3
aI15
aI115
aI141
aI387
aI341
aI953
aI399
aI6109
aI12037
aI21079
aI26745
asS'd'
I5677
sa(dp11354
S'a'
I32173
sS's'
I16
sS'm'
(lp11355
I1
aI3
aI3
aI1
aI5
aI5
aI31
aI195
aI477
aI755
aI687
aI3811
aI805
aI679
aI20687
aI46299
asS'd'
I5678
sa(dp11356
S'a'
I32182
sS's'
I16
sS'm'
(lp11357
I1
aI1
aI7
aI15
aI1
aI31
aI67
aI159
aI205
aI141
aI1667
aI3077
aI451
aI13161
aI16211
aI6887
asS'd'
I5679
sa(dp11358
S'a'
I32191
sS's'
I16
sS'm'
(lp11359
I1
aI3
aI3
aI1
aI7
aI43
aI87
aI5
aI49
aI205
aI231
aI3957
aI2947
aI13199
aI15743
aI4681
asS'd'
I5680
sa(dp11360
S'a'
I32193
sS's'
I16
sS'm'
(lp11361
I1
aI3
aI3
aI15
aI25
aI37
aI95
aI11
aI439
aI553
aI59
aI1241
aI7407
aI13467
aI22403
aI44441
asS'd'
I5681
sa(dp11362
S'a'
I32194
sS's'
I16
sS'm'
(lp11363
I1
aI1
aI1
aI3
aI21
aI3
aI127
aI239
aI491
aI139
aI1411
aI3417
aI4247
aI6247
aI13809
aI31609
asS'd'
I5682
sa(dp11364
S'a'
I32229
sS's'
I16
sS'm'
(lp11365
I1
aI1
aI5
aI1
aI9
aI13
aI5
aI155
aI109
aI593
aI119
aI4091
aI1911
aI8301
aI4239
aI50081
asS'd'
I5683
sa(dp11366
S'a'
I32230
sS's'
I16
sS'm'
(lp11367
I1
aI3
aI5
aI13
aI27
aI3
aI99
aI225
aI253
aI169
aI801
aI3741
aI1905
aI12073
aI31831
aI17997
asS'd'
I5684
sa(dp11368
S'a'
I32248
sS's'
I16
sS'm'
(lp11369
I1
aI3
aI7
aI15
aI9
aI23
aI93
aI171
aI453
aI983
aI1657
aI1133
aI6381
aI5229
aI32303
aI17439
asS'd'
I5685
sa(dp11370
S'a'
I32263
sS's'
I16
sS'm'
(lp11371
I1
aI1
aI7
aI11
aI7
aI5
aI125
aI141
aI63
aI763
aI1293
aI1007
aI4579
aI1479
aI11977
aI59261
asS'd'
I5686
sa(dp11372
S'a'
I32264
sS's'
I16
sS'm'
(lp11373
I1
aI3
aI1
aI7
aI1
aI15
aI49
aI41
aI367
aI639
aI1933
aI401
aI2335
aI2441
aI13653
aI55555
asS'd'
I5687
sa(dp11374
S'a'
I32269
sS's'
I16
sS'm'
(lp11375
I1
aI3
aI1
aI7
aI15
aI23
aI5
aI213
aI45
aI721
aI543
aI2133
aI4525
aI9719
aI28053
aI54075
asS'd'
I5688
sa(dp11376
S'a'
I32298
sS's'
I16
sS'm'
(lp11377
I1
aI3
aI7
aI3
aI11
aI7
aI23
aI35
aI169
aI829
aI1957
aI2423
aI3583
aI4951
aI28957
aI29753
asS'd'
I5689
sa(dp11378
S'a'
I32335
sS's'
I16
sS'm'
(lp11379
I1
aI1
aI3
aI3
aI1
aI5
aI19
aI235
aI175
aI969
aI229
aI2335
aI7215
aI10195
aI7487
aI64191
asS'd'
I5690
sa(dp11380
S'a'
I32340
sS's'
I16
sS'm'
(lp11381
I1
aI1
aI7
aI3
aI27
aI1
aI73
aI49
aI445
aI863
aI69
aI3555
aI993
aI9553
aI31941
aI29901
asS'd'
I5691
sa(dp11382
S'a'
I32356
sS's'
I16
sS'm'
(lp11383
I1
aI3
aI5
aI11
aI9
aI25
aI59
aI177
aI23
aI997
aI1041
aI1135
aI3879
aI767
aI2263
aI51267
asS'd'
I5692
sa(dp11384
S'a'
I32374
sS's'
I16
sS'm'
(lp11385
I1
aI1
aI7
aI3
aI1
aI63
aI49
aI51
aI237
aI569
aI1293
aI1143
aI3125
aI16315
aI17009
aI24821
asS'd'
I5693
sa(dp11386
S'a'
I32390
sS's'
I16
sS'm'
(lp11387
I1
aI3
aI3
aI15
aI11
aI17
aI121
aI25
aI349
aI833
aI557
aI1975
aI5405
aI15189
aI31243
aI53541
asS'd'
I5694
sa(dp11388
S'a'
I32401
sS's'
I16
sS'm'
(lp11389
I1
aI3
aI7
aI9
aI11
aI15
aI39
aI15
aI75
aI87
aI55
aI2069
aI3291
aI507
aI16925
aI57751
asS'd'
I5695
sa(dp11390
S'a'
I32414
sS's'
I16
sS'm'
(lp11391
I1
aI1
aI3
aI15
aI1
aI21
aI61
aI139
aI357
aI931
aI647
aI947
aI2291
aI15557
aI6739
aI5881
asS'd'
I5696
sa(dp11392
S'a'
I32417
sS's'
I16
sS'm'
(lp11393
I1
aI3
aI1
aI9
aI1
aI47
aI73
aI59
aI115
aI497
aI733
aI1777
aI905
aI16181
aI4351
aI7345
asS'd'
I5697
sa(dp11394
S'a'
I32442
sS's'
I16
sS'm'
(lp11395
I1
aI3
aI3
aI7
aI5
aI21
aI67
aI113
aI71
aI743
aI757
aI1851
aI7899
aI10315
aI15437
aI61803
asS'd'
I5698
sa(dp11396
S'a'
I32450
sS's'
I16
sS'm'
(lp11397
I1
aI3
aI7
aI1
aI9
aI23
aI77
aI131
aI395
aI767
aI1229
aI2629
aI5731
aI11907
aI32217
aI18473
asS'd'
I5699
sa(dp11398
S'a'
I32461
sS's'
I16
sS'm'
(lp11399
I1
aI3
aI5
aI15
aI1
aI23
aI123
aI207
aI291
aI565
aI1211
aI501
aI2111
aI11381
aI5171
aI54841
asS'd'
I5700
sa(dp11400
S'a'
I32473
sS's'
I16
sS'm'
(lp11401
I1
aI1
aI1
aI15
aI21
aI13
aI3
aI175
aI405
aI109
aI1353
aI2495
aI7619
aI14971
aI28179
aI34737
asS'd'
I5701
sa(dp11402
S'a'
I32479
sS's'
I16
sS'm'
(lp11403
I1
aI3
aI5
aI3
aI17
aI25
aI53
aI71
aI229
aI729
aI1953
aI3119
aI7747
aI1551
aI23417
aI35563
asS'd'
I5702
sa(dp11404
S'a'
I32530
sS's'
I16
sS'm'
(lp11405
I1
aI1
aI7
aI7
aI11
aI31
aI81
aI43
aI149
aI537
aI1253
aI2759
aI431
aI4813
aI8375
aI46329
asS'd'
I5703
sa(dp11406
S'a'
I32536
sS's'
I16
sS'm'
(lp11407
I1
aI1
aI1
aI5
aI11
aI27
aI61
aI199
aI239
aI889
aI723
aI2353
aI5663
aI7385
aI28165
aI14675
asS'd'
I5704
sa(dp11408
S'a'
I32548
sS's'
I16
sS'm'
(lp11409
I1
aI3
aI1
aI7
aI3
aI3
aI83
aI247
aI247
aI57
aI579
aI1163
aI2615
aI4051
aI2809
aI46413
asS'd'
I5705
sa(dp11410
S'a'
I32577
sS's'
I16
sS'm'
(lp11411
I1
aI1
aI3
aI11
aI13
aI47
aI11
aI235
aI475
aI35
aI843
aI2329
aI3519
aI8899
aI14533
aI24889
asS'd'
I5706
sa(dp11412
S'a'
I32628
sS's'
I16
sS'm'
(lp11413
I1
aI3
aI1
aI1
aI7
aI31
aI15
aI101
aI327
aI499
aI471
aI1001
aI339
aI11863
aI24787
aI47191
asS'd'
I5707
sa(dp11414
S'a'
I32642
sS's'
I16
sS'm'
(lp11415
I1
aI1
aI7
aI1
aI3
aI55
aI93
aI43
aI11
aI65
aI289
aI1249
aI5325
aI13867
aI29841
aI34333
asS'd'
I5708
sa(dp11416
S'a'
I32665
sS's'
I16
sS'm'
(lp11417
I1
aI3
aI3
aI1
aI25
aI61
aI87
aI113
aI115
aI265
aI1007
aI1129
aI7633
aI6109
aI5733
aI22649
asS'd'
I5709
sa(dp11418
S'a'
I32666
sS's'
I16
sS'm'
(lp11419
I1
aI3
aI1
aI11
aI31
aI59
aI127
aI83
aI33
aI419
aI1037
aI3777
aI6383
aI2711
aI2113
aI17233
asS'd'
I5710
sa(dp11420
S'a'
I32668
sS's'
I16
sS'm'
(lp11421
I1
aI1
aI5
aI13
aI11
aI17
aI73
aI41
aI257
aI223
aI359
aI3821
aI4617
aI1943
aI11331
aI40153
asS'd'
I5711
sa(dp11422
S'a'
I32696
sS's'
I16
sS'm'
(lp11423
I1
aI1
aI1
aI1
aI9
aI25
aI43
aI179
aI17
aI1021
aI1323
aI761
aI5861
aI11547
aI26017
aI5165
asS'd'
I5712
sa(dp11424
S'a'
I32722
sS's'
I16
sS'm'
(lp11425
I1
aI3
aI5
aI3
aI29
aI21
aI53
aI111
aI213
aI717
aI1101
aI3215
aI3021
aI16343
aI23387
aI33439
asS'd'
I5713
sa(dp11426
S'a'
I32757
sS's'
I16
sS'm'
(lp11427
I1
aI3
aI5
aI13
aI29
aI11
aI21
aI89
aI107
aI111
aI1121
aI2785
aI3493
aI9873
aI13
aI40863
asS'd'
I5714
sa(dp11428
S'a'
I32758
sS's'
I16
sS'm'
(lp11429
I1
aI1
aI5
aI13
aI15
aI15
aI111
aI219
aI59
aI43
aI333
aI3581
aI1311
aI2799
aI23987
aI21637
asS'd'
I5715
sa(dp11430
S'a'
I4
sS's'
I17
sS'm'
(lp11431
I1
aI3
aI1
aI11
aI21
aI57
aI115
aI247
aI499
aI525
aI1629
aI3679
aI2109
aI6607
aI27435
aI1745
aI71201
asS'd'
I5716
sa(dp11432
S'a'
I7
sS's'
I17
sS'm'
(lp11433
I1
aI3
aI3
aI3
aI31
aI17
aI113
aI165
aI189
aI361
aI103
aI1775
aI3001
aI3865
aI30591
aI2873
aI17129
asS'd'
I5717
sa(dp11434
S'a'
I16
sS's'
I17
sS'm'
(lp11435
I1
aI1
aI5
aI5
aI15
aI47
aI47
aI85
aI247
aI471
aI713
aI3571
aI2407
aI9811
aI8187
aI32133
aI8541
asS'd'
I5718
sa(dp11436
S'a'
I22
sS's'
I17
sS'm'
(lp11437
I1
aI3
aI3
aI1
aI15
aI1
aI59
aI151
aI469
aI351
aI671
aI2925
aI7207
aI5061
aI28691
aI4363
aI50767
asS'd'
I5719
sa(dp11438
S'a'
I25
sS's'
I17
sS'm'
(lp11439
I1
aI1
aI5
aI7
aI11
aI35
aI67
aI45
aI193
aI3
aI627
aI3333
aI6497
aI12307
aI28807
aI13997
aI108645
asS'd'
I5720
sa(dp11440
S'a'
I31
sS's'
I17
sS'm'
(lp11441
I1
aI3
aI1
aI1
aI17
aI63
aI125
aI185
aI485
aI759
aI717
aI1993
aI6707
aI3993
aI2181
aI8173
aI18057
asS'd'
I5721
sa(dp11442
S'a'
I32
sS's'
I17
sS'm'
(lp11443
I1
aI1
aI3
aI13
aI7
aI15
aI113
aI207
aI103
aI191
aI1895
aI2595
aI3873
aI12021
aI19259
aI12553
aI119119
asS'd'
I5722
sa(dp11444
S'a'
I42
sS's'
I17
sS'm'
(lp11445
I1
aI3
aI7
aI1
aI17
aI11
aI101
aI209
aI315
aI9
aI901
aI2303
aI7623
aI7459
aI26391
aI45143
aI5753
asS'd'
I5723
sa(dp11446
S'a'
I52
sS's'
I17
sS'm'
(lp11447
I1
aI1
aI5
aI15
aI1
aI5
aI71
aI155
aI167
aI89
aI145
aI3483
aI2385
aI15205
aI9193
aI20637
aI58473
asS'd'
I5724
sa(dp11448
S'a'
I61
sS's'
I17
sS'm'
(lp11449
I1
aI1
aI5
aI7
aI25
aI55
aI57
aI51
aI333
aI299
aI1721
aI1667
aI6513
aI10191
aI29405
aI21923
aI76593
asS'd'
I5725
sa(dp11450
S'a'
I70
sS's'
I17
sS'm'
(lp11451
I1
aI1
aI5
aI1
aI7
aI37
aI107
aI91
aI241
aI137
aI627
aI2749
aI5573
aI11243
aI26197
aI4545
aI105599
asS'd'
I5726
sa(dp11452
S'a'
I76
sS's'
I17
sS'm'
(lp11453
I1
aI3
aI1
aI5
aI25
aI37
aI73
aI61
aI57
aI249
aI1953
aI1385
aI6479
aI3701
aI10693
aI617
aI62535
asS'd'
I5727
sa(dp11454
S'a'
I81
sS's'
I17
sS'm'
(lp11455
I1
aI1
aI1
aI15
aI5
aI63
aI41
aI151
aI395
aI681
aI227
aI3027
aI8123
aI15091
aI15475
aI35671
aI21129
asS'd'
I5728
sa(dp11456
S'a'
I87
sS's'
I17
sS'm'
(lp11457
I1
aI3
aI5
aI11
aI29
aI21
aI15
aI233
aI103
aI463
aI1829
aI2257
aI1717
aI2249
aI9599
aI5097
aI55705
asS'd'
I5729
sa(dp11458
S'a'
I93
sS's'
I17
sS'm'
(lp11459
I1
aI3
aI5
aI1
aI29
aI3
aI35
aI151
aI193
aI105
aI1107
aI827
aI7169
aI1843
aI15225
aI29025
aI43165
asS'd'
I5730
sa(dp11460
S'a'
I98
sS's'
I17
sS'm'
(lp11461
I1
aI1
aI7
aI15
aI17
aI51
aI93
aI199
aI205
aI41
aI113
aI1081
aI1571
aI11471
aI11057
aI16149
aI66905
asS'd'
I5731
sa(dp11462
S'a'
I122
sS's'
I17
sS'm'
(lp11463
I1
aI1
aI3
aI11
aI5
aI25
aI107
aI195
aI51
aI675
aI1683
aI3739
aI1653
aI611
aI23249
aI53157
aI127785
asS'd'
I5732
sa(dp11464
S'a'
I133
sS's'
I17
sS'm'
(lp11465
I1
aI1
aI7
aI5
aI7
aI3
aI25
aI145
aI453
aI735
aI441
aI77
aI8171
aI9281
aI22749
aI36973
aI106237
asS'd'
I5733
sa(dp11466
S'a'
I134
sS's'
I17
sS'm'
(lp11467
I1
aI1
aI3
aI13
aI13
aI5
aI95
aI33
aI223
aI369
aI453
aI2031
aI3531
aI6931
aI8977
aI54109
aI115487
asS'd'
I5734
sa(dp11468
S'a'
I140
sS's'
I17
sS'm'
(lp11469
I1
aI1
aI7
aI7
aI1
aI61
aI33
aI183
aI245
aI623
aI529
aI1831
aI1867
aI2845
aI8311
aI10143
aI67897
asS'd'
I5735
sa(dp11470
S'a'
I146
sS's'
I17
sS'm'
(lp11471
I1
aI3
aI7
aI11
aI27
aI23
aI93
aI9
aI61
aI451
aI67
aI1695
aI4227
aI2415
aI19249
aI44765
aI24611
asS'd'
I5736
sa(dp11472
S'a'
I158
sS's'
I17
sS'm'
(lp11473
I1
aI3
aI3
aI11
aI29
aI57
aI65
aI117
aI349
aI149
aI363
aI1095
aI4989
aI3071
aI17519
aI18079
aI7277
asS'd'
I5737
sa(dp11474
S'a'
I171
sS's'
I17
sS'm'
(lp11475
I1
aI3
aI5
aI9
aI1
aI7
aI59
aI87
aI307
aI111
aI1291
aI789
aI7361
aI6477
aI11229
aI36785
aI33303
asS'd'
I5738
sa(dp11476
S'a'
I176
sS's'
I17
sS'm'
(lp11477
I1
aI3
aI5
aI1
aI19
aI47
aI53
aI81
aI127
aI849
aI1479
aI1459
aI1889
aI15087
aI22115
aI20587
aI121005
asS'd'
I5739
sa(dp11478
S'a'
I179
sS's'
I17
sS'm'
(lp11479
I1
aI1
aI7
aI15
aI31
aI31
aI71
aI55
aI253
aI927
aI277
aI2087
aI1313
aI3721
aI22729
aI34709
aI9821
asS'd'
I5740
sa(dp11480
S'a'
I182
sS's'
I17
sS'm'
(lp11481
I1
aI3
aI5
aI13
aI13
aI63
aI73
aI41
aI165
aI315
aI1907
aI2005
aI691
aI725
aI22685
aI8673
aI76011
asS'd'
I5741
sa(dp11482
S'a'
I191
sS's'
I17
sS'm'
(lp11483
I1
aI1
aI5
aI9
aI23
aI61
aI47
aI167
aI279
aI683
aI683
aI1261
aI4037
aI15251
aI9421
aI45359
aI38001
asS'd'
I5742
sa(dp11484
S'a'
I193
sS's'
I17
sS'm'
(lp11485
I1
aI1
aI7
aI3
aI17
aI33
aI69
aI139
aI235
aI709
aI1475
aI2483
aI7559
aI8581
aI23965
aI31153
aI5097
asS'd'
I5743
sa(dp11486
S'a'
I224
sS's'
I17
sS'm'
(lp11487
I1
aI1
aI7
aI15
aI23
aI61
aI43
aI5
aI433
aI531
aI761
aI2749
aI2881
aI5225
aI13491
aI16479
aI50203
asS'd'
I5744
sa(dp11488
S'a'
I227
sS's'
I17
sS'm'
(lp11489
I1
aI1
aI3
aI9
aI29
aI7
aI9
aI23
aI339
aI315
aI1723
aI779
aI2983
aI6571
aI16025
aI63055
aI111103
asS'd'
I5745
sa(dp11490
S'a'
I229
sS's'
I17
sS'm'
(lp11491
I1
aI1
aI7
aI13
aI23
aI55
aI71
aI121
aI297
aI193
aI41
aI3165
aI4419
aI5853
aI28127
aI56151
aI16597
asS'd'
I5746
sa(dp11492
S'a'
I236
sS's'
I17
sS'm'
(lp11493
I1
aI1
aI5
aI7
aI7
aI23
aI93
aI11
aI261
aI297
aI1769
aI1239
aI2579
aI531
aI4423
aI7891
aI21729
asS'd'
I5747
sa(dp11494
S'a'
I248
sS's'
I17
sS'm'
(lp11495
I1
aI3
aI5
aI1
aI13
aI35
aI83
aI85
aI125
aI887
aI161
aI3311
aI7261
aI9557
aI28975
aI28643
aI21479
asS'd'
I5748
sa(dp11496
S'a'
I262
sS's'
I17
sS'm'
(lp11497
I1
aI3
aI5
aI3
aI27
aI5
aI47
aI175
aI287
aI867
aI141
aI3079
aI7583
aI4997
aI18271
aI24097
aI96319
asS'd'
I5749
sa(dp11498
S'a'
I273
sS's'
I17
sS'm'
(lp11499
I1
aI3
aI5
aI1
aI21
aI51
aI47
aI67
aI211
aI281
aI1861
aI1169
aI6403
aI4229
aI3995
aI9921
aI41515
asS'd'
I5750
sa(dp11500
S'a'
I276
sS's'
I17
sS'm'
(lp11501
I1
aI3
aI3
aI11
aI23
aI23
aI81
aI55
aI441
aI211
aI169
aI3197
aI7213
aI7205
aI15
aI11771
aI129091
asS'd'
I5751
sa(dp11502
S'a'
I280
sS's'
I17
sS'm'
(lp11503
I1
aI3
aI7
aI3
aI23
aI39
aI23
aI163
aI253
aI1005
aI1775
aI3393
aI7659
aI8065
aI30021
aI61065
aI35171
asS'd'
I5752
sa(dp11504
S'a'
I283
sS's'
I17
sS'm'
(lp11505
I1
aI3
aI1
aI1
aI29
aI29
aI39
aI143
aI191
aI711
aI1077
aI13
aI4137
aI15425
aI11139
aI1269
aI71915
asS'd'
I5753
sa(dp11506
S'a'
I290
sS's'
I17
sS'm'
(lp11507
I1
aI3
aI3
aI5
aI11
aI41
aI101
aI127
aI301
aI335
aI45
aI2065
aI5835
aI7801
aI2639
aI5735
aI63445
asS'd'
I5754
sa(dp11508
S'a'
I309
sS's'
I17
sS'm'
(lp11509
I1
aI3
aI5
aI9
aI3
aI39
aI51
aI53
aI489
aI663
aI951
aI3931
aI3075
aI753
aI22179
aI20573
aI10775
asS'd'
I5755
sa(dp11510
S'a'
I316
sS's'
I17
sS'm'
(lp11511
I1
aI3
aI3
aI15
aI13
aI31
aI1
aI237
aI79
aI587
aI395
aI591
aI607
aI13105
aI21301
aI26829
aI112181
asS'd'
I5756
sa(dp11512
S'a'
I319
sS's'
I17
sS'm'
(lp11513
I1
aI1
aI7
aI7
aI5
aI55
aI31
aI117
aI247
aI229
aI247
aI307
aI3821
aI6483
aI31317
aI22975
aI40535
asS'd'
I5757
sa(dp11514
S'a'
I321
sS's'
I17
sS'm'
(lp11515
I1
aI3
aI7
aI15
aI15
aI59
aI101
aI17
aI437
aI373
aI1727
aI471
aI2783
aI7825
aI24555
aI58765
aI5097
asS'd'
I5758
sa(dp11516
S'a'
I328
sS's'
I17
sS'm'
(lp11517
I1
aI1
aI3
aI9
aI31
aI27
aI71
aI147
aI71
aI871
aI793
aI2363
aI3213
aI13383
aI29801
aI53187
aI70021
asS'd'
I5759
sa(dp11518
S'a'
I346
sS's'
I17
sS'm'
(lp11519
I1
aI3
aI1
aI1
aI19
aI47
aI121
aI61
aI303
aI565
aI1371
aI3703
aI2201
aI6835
aI26041
aI56039
aI80227
asS'd'
I5760
sa(dp11520
S'a'
I355
sS's'
I17
sS'm'
(lp11521
I1
aI1
aI5
aI5
aI3
aI45
aI91
aI61
aI257
aI947
aI1449
aI4031
aI4925
aI8627
aI11909
aI9529
aI3429
asS'd'
I5761
sa(dp11522
S'a'
I367
sS's'
I17
sS'm'
(lp11523
I1
aI1
aI1
aI7
aI9
aI63
aI69
aI233
aI141
aI361
aI1443
aI2157
aI2877
aI643
aI2779
aI8109
aI126911
asS'd'
I5762
sa(dp11524
S'a'
I369
sS's'
I17
sS'm'
(lp11525
I1
aI1
aI5
aI1
aI5
aI3
aI67
aI157
aI21
aI1
aI361
aI35
aI1475
aI12877
aI22169
aI6653
aI85005
asS'd'
I5763
sa(dp11526
S'a'
I372
sS's'
I17
sS'm'
(lp11527
I1
aI1
aI7
aI9
aI25
aI1
aI7
aI175
aI47
aI963
aI405
aI3955
aI3905
aI8429
aI8483
aI62037
aI11323
asS'd'
I5764
sa(dp11528
S'a'
I382
sS's'
I17
sS'm'
(lp11529
I1
aI1
aI5
aI11
aI29
aI23
aI77
aI211
aI319
aI745
aI1935
aI2429
aI1687
aI2173
aI1571
aI19457
aI117777
asS'd'
I5765
sa(dp11530
S'a'
I388
sS's'
I17
sS'm'
(lp11531
I1
aI1
aI7
aI5
aI15
aI57
aI121
aI189
aI303
aI79
aI527
aI1801
aI71
aI9857
aI14197
aI59007
aI75341
asS'd'
I5766
sa(dp11532
S'a'
I392
sS's'
I17
sS'm'
(lp11533
I1
aI3
aI3
aI5
aI25
aI3
aI19
aI141
aI155
aI157
aI287
aI769
aI5789
aI8443
aI31823
aI1019
aI79111
asS'd'
I5767
sa(dp11534
S'a'
I395
sS's'
I17
sS'm'
(lp11535
I1
aI1
aI5
aI11
aI27
aI27
aI117
aI141
aI355
aI1023
aI869
aI995
aI6311
aI6573
aI11721
aI1565
aI35517
asS'd'
I5768
sa(dp11536
S'a'
I397
sS's'
I17
sS'm'
(lp11537
I1
aI1
aI1
aI9
aI1
aI33
aI107
aI51
aI41
aI889
aI1191
aI1055
aI503
aI14779
aI6641
aI58117
aI74157
asS'd'
I5769
sa(dp11538
S'a'
I403
sS's'
I17
sS'm'
(lp11539
I1
aI1
aI7
aI5
aI13
aI39
aI39
aI33
aI293
aI75
aI963
aI3379
aI1847
aI12371
aI9005
aI38107
aI69753
asS'd'
I5770
sa(dp11540
S'a'
I409
sS's'
I17
sS'm'
(lp11541
I1
aI1
aI5
aI5
aI7
aI37
aI19
aI241
aI427
aI635
aI1711
aI3835
aI773
aI10525
aI17207
aI1675
aI127255
asS'd'
I5771
sa(dp11542
S'a'
I410
sS's'
I17
sS'm'
(lp11543
I1
aI1
aI3
aI7
aI17
aI19
aI11
aI113
aI191
aI947
aI1133
aI3173
aI213
aI10125
aI1373
aI56797
aI111011
asS'd'
I5772
sa(dp11544
S'a'
I425
sS's'
I17
sS'm'
(lp11545
I1
aI3
aI1
aI1
aI29
aI45
aI65
aI237
aI223
aI695
aI697
aI3197
aI6887
aI8079
aI22099
aI12079
aI54847
asS'd'
I5773
sa(dp11546
S'a'
I443
sS's'
I17
sS'm'
(lp11547
I1
aI3
aI3
aI7
aI5
aI47
aI19
aI215
aI341
aI863
aI1879
aI571
aI7113
aI2465
aI23407
aI52555
aI44375
asS'd'
I5774
sa(dp11548
S'a'
I472
sS's'
I17
sS'm'
(lp11549
I1
aI3
aI5
aI11
aI25
aI31
aI109
aI73
aI429
aI553
aI1905
aI1753
aI6733
aI4433
aI13785
aI32041
aI27115
asS'd'
I5775
sa(dp11550
S'a'
I475
sS's'
I17
sS'm'
(lp11551
I1
aI1
aI1
aI3
aI27
aI5
aI97
aI47
aI343
aI977
aI1241
aI721
aI3355
aI3559
aI28349
aI56389
aI63103
asS'd'
I5776
sa(dp11552
S'a'
I481
sS's'
I17
sS'm'
(lp11553
I1
aI3
aI3
aI9
aI21
aI53
aI57
aI211
aI73
aI155
aI1855
aI715
aI3179
aI5963
aI10061
aI35141
aI63131
asS'd'
I5777
sa(dp11554
S'a'
I488
sS's'
I17
sS'm'
(lp11555
I1
aI3
aI1
aI15
aI21
aI25
aI51
aI73
aI31
aI25
aI1385
aI637
aI6585
aI49
aI2105
aI6829
aI9353
asS'd'
I5778
sa(dp11556
S'a'
I493
sS's'
I17
sS'm'
(lp11557
I1
aI1
aI7
aI5
aI11
aI55
aI31
aI69
aI145
aI637
aI1131
aI2175
aI3547
aI13031
aI2131
aI12361
aI74737
asS'd'
I5779
sa(dp11558
S'a'
I501
sS's'
I17
sS'm'
(lp11559
I1
aI3
aI3
aI5
aI31
aI7
aI119
aI119
aI309
aI925
aI895
aI3813
aI1131
aI4765
aI17865
aI48707
aI113577
asS'd'
I5780
sa(dp11560
S'a'
I515
sS's'
I17
sS'm'
(lp11561
I1
aI3
aI3
aI9
aI13
aI33
aI127
aI177
aI323
aI727
aI1881
aI775
aI7329
aI11881
aI28309
aI987
aI116093
asS'd'
I5781
sa(dp11562
S'a'
I522
sS's'
I17
sS'm'
(lp11563
I1
aI1
aI3
aI5
aI31
aI55
aI39
aI41
aI511
aI157
aI1655
aI2991
aI3633
aI8521
aI27049
aI18771
aI54015
asS'd'
I5782
sa(dp11564
S'a'
I524
sS's'
I17
sS'm'
(lp11565
I1
aI3
aI5
aI13
aI11
aI45
aI113
aI185
aI375
aI661
aI1331
aI4013
aI5521
aI1037
aI23365
aI30239
aI76957
asS'd'
I5783
sa(dp11566
S'a'
I527
sS's'
I17
sS'm'
(lp11567
I1
aI3
aI3
aI7
aI19
aI7
aI23
aI17
aI435
aI913
aI1985
aI353
aI6049
aI7549
aI3371
aI60867
aI41099
asS'd'
I5784
sa(dp11568
S'a'
I535
sS's'
I17
sS'm'
(lp11569
I1
aI3
aI3
aI15
aI17
aI9
aI53
aI127
aI149
aI849
aI1181
aI2237
aI1345
aI539
aI19715
aI26277
aI125445
asS'd'
I5785
sa(dp11570
S'a'
I542
sS's'
I17
sS'm'
(lp11571
I1
aI1
aI1
aI3
aI1
aI9
aI67
aI79
aI79
aI795
aI1793
aI3167
aI5917
aI5323
aI22043
aI22007
aI3917
asS'd'
I5786
sa(dp11572
S'a'
I545
sS's'
I17
sS'm'
(lp11573
I1
aI3
aI5
aI9
aI15
aI19
aI59
aI37
aI141
aI145
aI413
aI1095
aI7709
aI669
aI27061
aI40171
aI101499
asS'd'
I5787
sa(dp11574
S'a'
I555
sS's'
I17
sS'm'
(lp11575
I1
aI3
aI1
aI1
aI9
aI49
aI109
aI7
aI119
aI861
aI875
aI1049
aI4125
aI6113
aI15699
aI6105
aI48799
asS'd'
I5788
sa(dp11576
S'a'
I558
sS's'
I17
sS'm'
(lp11577
I1
aI1
aI3
aI9
aI11
aI29
aI43
aI175
aI371
aI357
aI1181
aI3933
aI43
aI4559
aI10333
aI23603
aI83095
asS'd'
I5789
sa(dp11578
S'a'
I560
sS's'
I17
sS'm'
(lp11579
I1
aI3
aI3
aI9
aI9
aI7
aI57
aI61
aI409
aI143
aI591
aI761
aI4107
aI8117
aI1051
aI4471
aI91771
asS'd'
I5790
sa(dp11580
S'a'
I563
sS's'
I17
sS'm'
(lp11581
I1
aI1
aI3
aI11
aI3
aI53
aI119
aI21
aI213
aI219
aI51
aI3491
aI7143
aI937
aI24693
aI3211
aI99463
asS'd'
I5791
sa(dp11582
S'a'
I570
sS's'
I17
sS'm'
(lp11583
I1
aI1
aI3
aI3
aI1
aI47
aI53
aI153
aI211
aI523
aI1637
aI3351
aI3753
aI12489
aI31825
aI27613
aI96431
asS'd'
I5792
sa(dp11584
S'a'
I578
sS's'
I17
sS'm'
(lp11585
I1
aI1
aI5
aI15
aI23
aI57
aI81
aI231
aI147
aI9
aI1043
aI3157
aI1463
aI4835
aI22435
aI57407
aI59615
asS'd'
I5793
sa(dp11586
S'a'
I583
sS's'
I17
sS'm'
(lp11587
I1
aI3
aI3
aI13
aI15
aI63
aI111
aI5
aI449
aI957
aI1175
aI2887
aI7741
aI8975
aI28775
aI4067
aI69283
asS'd'
I5794
sa(dp11588
S'a'
I590
sS's'
I17
sS'm'
(lp11589
I1
aI3
aI1
aI1
aI5
aI61
aI109
aI211
aI349
aI179
aI951
aI153
aI3147
aI7555
aI27037
aI59829
aI16077
asS'd'
I5795
sa(dp11590
S'a'
I597
sS's'
I17
sS'm'
(lp11591
I1
aI3
aI3
aI7
aI15
aI33
aI53
aI61
aI309
aI991
aI227
aI3437
aI3983
aI14559
aI13065
aI46387
aI49105
asS'd'
I5796
sa(dp11592
S'a'
I604
sS's'
I17
sS'm'
(lp11593
I1
aI3
aI5
aI3
aI25
aI23
aI97
aI139
aI315
aI601
aI1179
aI1083
aI6799
aI1813
aI15511
aI60433
aI65641
asS'd'
I5797
sa(dp11594
S'a'
I608
sS's'
I17
sS'm'
(lp11595
I1
aI1
aI7
aI1
aI11
aI43
aI87
aI87
aI173
aI161
aI91
aI3011
aI1869
aI2313
aI13691
aI3509
aI39433
asS'd'
I5798
sa(dp11596
S'a'
I614
sS's'
I17
sS'm'
(lp11597
I1
aI3
aI5
aI7
aI15
aI5
aI39
aI251
aI269
aI819
aI815
aI2283
aI5635
aI6953
aI27017
aI65143
aI45281
asS'd'
I5799
sa(dp11598
S'a'
I635
sS's'
I17
sS'm'
(lp11599
I1
aI3
aI7
aI9
aI1
aI37
aI9
aI57
aI467
aI37
aI1743
aI4031
aI3751
aI8105
aI23789
aI46847
aI21911
asS'd'
I5800
sa(dp11600
S'a'
I637
sS's'
I17
sS'm'
(lp11601
I1
aI1
aI7
aI1
aI23
aI47
aI63
aI99
aI59
aI951
aI1837
aI2829
aI161
aI857
aI4045
aI9945
aI53487
asS'd'
I5801
sa(dp11602
S'a'
I653
sS's'
I17
sS'm'
(lp11603
I1
aI3
aI7
aI7
aI11
aI47
aI43
aI99
aI279
aI945
aI1189
aI2091
aI4597
aI183
aI15527
aI7151
aI112403
asS'd'
I5802
sa(dp11604
S'a'
I654
sS's'
I17
sS'm'
(lp11605
I1
aI3
aI3
aI15
aI9
aI53
aI63
aI135
aI119
aI95
aI131
aI2461
aI157
aI10631
aI20847
aI51699
aI58865
asS'd'
I5803
sa(dp11606
S'a'
I659
sS's'
I17
sS'm'
(lp11607
I1
aI1
aI3
aI1
aI25
aI3
aI115
aI29
aI303
aI361
aI1529
aI3993
aI5899
aI11501
aI4463
aI47121
aI75333
asS'd'
I5804
sa(dp11608
S'a'
I666
sS's'
I17
sS'm'
(lp11609
I1
aI3
aI1
aI15
aI9
aI39
aI31
aI199
aI305
aI279
aI15
aI611
aI561
aI6593
aI3189
aI1863
aI61875
asS'd'
I5805
sa(dp11610
S'a'
I671
sS's'
I17
sS'm'
(lp11611
I1
aI3
aI5
aI15
aI5
aI49
aI87
aI17
aI87
aI5
aI1179
aI1351
aI7647
aI7529
aI15901
aI30351
aI31959
asS'd'
I5806
sa(dp11612
S'a'
I689
sS's'
I17
sS'm'
(lp11613
I1
aI3
aI3
aI9
aI31
aI57
aI127
aI239
aI349
aI773
aI547
aI2649
aI1309
aI8071
aI10741
aI57645
aI14423
asS'd'
I5807
sa(dp11614
S'a'
I690
sS's'
I17
sS'm'
(lp11615
I1
aI1
aI5
aI9
aI5
aI15
aI59
aI185
aI315
aI411
aI1425
aI3905
aI853
aI12393
aI21
aI15195
aI114291
asS'd'
I5808
sa(dp11616
S'a'
I695
sS's'
I17
sS'm'
(lp11617
I1
aI3
aI1
aI5
aI29
aI47
aI19
aI203
aI319
aI673
aI1169
aI2413
aI5295
aI6251
aI19883
aI2725
aI28937
asS'd'
I5809
sa(dp11618
S'a'
I713
sS's'
I17
sS'm'
(lp11619
I1
aI3
aI1
aI5
aI21
aI55
aI19
aI185
aI103
aI827
aI117
aI341
aI3315
aI5625
aI345
aI63845
aI49081
asS'd'
I5810
sa(dp11620
S'a'
I722
sS's'
I17
sS'm'
(lp11621
I1
aI1
aI7
aI9
aI27
aI51
aI105
aI15
aI243
aI735
aI1221
aI1641
aI293
aI14423
aI5363
aI60873
aI66223
asS'd'
I5811
sa(dp11622
S'a'
I733
sS's'
I17
sS'm'
(lp11623
I1
aI1
aI5
aI1
aI19
aI5
aI109
aI131
aI131
aI67
aI231
aI2907
aI4389
aI5079
aI20503
aI59045
aI33625
asS'd'
I5812
sa(dp11624
S'a'
I758
sS's'
I17
sS'm'
(lp11625
I1
aI3
aI1
aI5
aI5
aI15
aI79
aI67
aI287
aI225
aI519
aI1543
aI2389
aI671
aI7767
aI62625
aI61639
asS'd'
I5813
sa(dp11626
S'a'
I770
sS's'
I17
sS'm'
(lp11627
I1
aI1
aI1
aI9
aI25
aI35
aI83
aI15
aI291
aI207
aI1757
aI3691
aI5669
aI11255
aI27939
aI57813
aI46251
asS'd'
I5814
sa(dp11628
S'a'
I782
sS's'
I17
sS'm'
(lp11629
I1
aI3
aI1
aI1
aI29
aI3
aI83
aI109
aI323
aI179
aI1855
aI3205
aI7665
aI16201
aI13863
aI16347
aI98977
asS'd'
I5815
sa(dp11630
S'a'
I784
sS's'
I17
sS'm'
(lp11631
I1
aI3
aI1
aI13
aI17
aI1
aI101
aI183
aI153
aI985
aI125
aI999
aI855
aI15897
aI19491
aI8953
aI23277
asS'd'
I5816
sa(dp11632
S'a'
I793
sS's'
I17
sS'm'
(lp11633
I1
aI1
aI7
aI11
aI9
aI33
aI45
aI229
aI411
aI155
aI537
aI3037
aI1785
aI11719
aI8589
aI16617
aI47339
asS'd'
I5817
sa(dp11634
S'a'
I803
sS's'
I17
sS'm'
(lp11635
I1
aI1
aI5
aI5
aI9
aI11
aI7
aI163
aI305
aI621
aI1647
aI2609
aI7901
aI14421
aI23447
aI1205
aI52681
asS'd'
I5818
sa(dp11636
S'a'
I805
sS's'
I17
sS'm'
(lp11637
I1
aI3
aI3
aI1
aI7
aI29
aI39
aI227
aI419
aI561
aI129
aI3299
aI3123
aI4243
aI18689
aI12335
aI71783
asS'd'
I5819
sa(dp11638
S'a'
I812
sS's'
I17
sS'm'
(lp11639
I1
aI3
aI1
aI9
aI11
aI61
aI65
aI207
aI123
aI763
aI485
aI1943
aI3617
aI415
aI22397
aI58597
aI128017
asS'd'
I5820
sa(dp11640
S'a'
I838
sS's'
I17
sS'm'
(lp11641
I1
aI1
aI5
aI13
aI25
aI43
aI115
aI73
aI269
aI137
aI1765
aI705
aI1705
aI16137
aI22751
aI60021
aI4333
asS'd'
I5821
sa(dp11642
S'a'
I849
sS's'
I17
sS'm'
(lp11643
I1
aI1
aI5
aI13
aI3
aI57
aI9
aI141
aI75
aI695
aI597
aI3435
aI1085
aI4905
aI19625
aI16061
aI12111
asS'd'
I5822
sa(dp11644
S'a'
I875
sS's'
I17
sS'm'
(lp11645
I1
aI1
aI5
aI9
aI29
aI13
aI119
aI251
aI353
aI421
aI1955
aI3503
aI2605
aI2587
aI12503
aI46419
aI128815
asS'd'
I5823
sa(dp11646
S'a'
I877
sS's'
I17
sS'm'
(lp11647
I1
aI3
aI5
aI7
aI7
aI29
aI67
aI25
aI37
aI327
aI1607
aI1899
aI1691
aI5801
aI17441
aI9755
aI24993
asS'd'
I5824
sa(dp11648
S'a'
I880
sS's'
I17
sS'm'
(lp11649
I1
aI1
aI3
aI11
aI17
aI29
aI121
aI201
aI371
aI597
aI213
aI2361
aI6615
aI169
aI24801
aI56175
aI129241
asS'd'
I5825
sa(dp11650
S'a'
I892
sS's'
I17
sS'm'
(lp11651
I1
aI3
aI5
aI1
aI31
aI63
aI85
aI77
aI151
aI599
aI103
aI677
aI4431
aI12897
aI6373
aI40349
aI100819
asS'd'
I5826
sa(dp11652
S'a'
I895
sS's'
I17
sS'm'
(lp11653
I1
aI3
aI5
aI9
aI25
aI9
aI119
aI219
aI379
aI939
aI1907
aI945
aI5819
aI7433
aI32519
aI56493
aI50441
asS'd'
I5827
sa(dp11654
S'a'
I899
sS's'
I17
sS'm'
(lp11655
I1
aI1
aI3
aI9
aI13
aI1
aI63
aI189
aI135
aI839
aI1821
aI2247
aI2547
aI965
aI6847
aI63335
aI32921
asS'd'
I5828
sa(dp11656
S'a'
I919
sS's'
I17
sS'm'
(lp11657
I1
aI3
aI5
aI13
aI21
aI25
aI111
aI37
aI319
aI469
aI1999
aI1637
aI8167
aI2641
aI24615
aI63713
aI115923
asS'd'
I5829
sa(dp11658
S'a'
I920
sS's'
I17
sS'm'
(lp11659
I1
aI3
aI5
aI9
aI9
aI27
aI1
aI63
aI275
aI223
aI1675
aI3833
aI7377
aI9755
aI6279
aI37161
aI108805
asS'd'
I5830
sa(dp11660
S'a'
I932
sS's'
I17
sS'm'
(lp11661
I1
aI3
aI3
aI13
aI29
aI23
aI21
aI73
aI401
aI863
aI701
aI2527
aI4557
aI5549
aI22493
aI6651
aI39167
asS'd'
I5831
sa(dp11662
S'a'
I935
sS's'
I17
sS'm'
(lp11663
I1
aI1
aI3
aI15
aI25
aI21
aI97
aI25
aI83
aI925
aI2029
aI3789
aI3241
aI7617
aI13699
aI31123
aI124619
asS'd'
I5832
sa(dp11664
S'a'
I936
sS's'
I17
sS'm'
(lp11665
I1
aI3
aI7
aI5
aI23
aI7
aI95
aI227
aI123
aI215
aI359
aI2099
aI4505
aI8477
aI32665
aI18211
aI99679
asS'd'
I5833
sa(dp11666
S'a'
I941
sS's'
I17
sS'm'
(lp11667
I1
aI3
aI1
aI9
aI11
aI57
aI75
aI17
aI105
aI175
aI831
aI1033
aI5425
aI8419
aI16163
aI23901
aI33889
asS'd'
I5834
sa(dp11668
S'a'
I950
sS's'
I17
sS'm'
(lp11669
I1
aI1
aI7
aI1
aI17
aI49
aI71
aI23
aI129
aI413
aI333
aI2547
aI4627
aI14961
aI16745
aI53649
aI73059
asS'd'
I5835
sa(dp11670
S'a'
I961
sS's'
I17
sS'm'
(lp11671
I1
aI3
aI5
aI3
aI13
aI33
aI121
aI147
aI443
aI187
aI1949
aI319
aI8141
aI14359
aI11203
aI53569
aI70415
asS'd'
I5836
sa(dp11672
S'a'
I962
sS's'
I17
sS'm'
(lp11673
I1
aI3
aI1
aI11
aI15
aI1
aI23
aI29
aI509
aI985
aI1217
aI3755
aI385
aI3697
aI24631
aI37619
aI62435
asS'd'
I5837
sa(dp11674
S'a'
I971
sS's'
I17
sS'm'
(lp11675
I1
aI3
aI3
aI3
aI17
aI11
aI107
aI37
aI227
aI913
aI259
aI2799
aI3249
aI2347
aI9703
aI52741
aI101187
asS'd'
I5838
sa(dp11676
S'a'
I982
sS's'
I17
sS'm'
(lp11677
I1
aI1
aI5
aI13
aI25
aI25
aI47
aI77
aI405
aI415
aI1947
aI1675
aI5079
aI1333
aI10059
aI32033
aI88975
asS'd'
I5839
sa(dp11678
S'a'
I986
sS's'
I17
sS'm'
(lp11679
I1
aI3
aI5
aI9
aI27
aI7
aI19
aI241
aI445
aI205
aI333
aI285
aI7997
aI6339
aI29643
aI10229
aI29965
asS'd'
I5840
sa(dp11680
S'a'
I1012
sS's'
I17
sS'm'
(lp11681
I1
aI3
aI5
aI11
aI17
aI9
aI91
aI223
aI173
aI1013
aI779
aI3967
aI781
aI5471
aI4309
aI24795
aI99203
asS'd'
I5841
sa(dp11682
S'a'
I1021
sS's'
I17
sS'm'
(lp11683
I1
aI1
aI1
aI3
aI19
aI53
aI7
aI159
aI351
aI515
aI223
aI3375
aI1
aI4985
aI16729
aI43333
aI85917
asS'd'
I5842
sa(dp11684
S'a'
I1024
sS's'
I17
sS'm'
(lp11685
I1
aI3
aI3
aI1
aI19
aI35
aI95
aI69
aI19
aI157
aI1177
aI579
aI7109
aI3499
aI3219
aI26641
aI49491
asS'd'
I5843
sa(dp11686
S'a'
I1029
sS's'
I17
sS'm'
(lp11687
I1
aI3
aI3
aI5
aI25
aI21
aI125
aI5
aI39
aI857
aI615
aI2925
aI2005
aI5503
aI25523
aI36711
aI30939
asS'd'
I5844
sa(dp11688
S'a'
I1030
sS's'
I17
sS'm'
(lp11689
I1
aI3
aI1
aI5
aI11
aI33
aI29
aI5
aI425
aI125
aI939
aI1641
aI321
aI1023
aI12551
aI4587
aI116617
asS'd'
I5845
sa(dp11690
S'a'
I1051
sS's'
I17
sS'm'
(lp11691
I1
aI3
aI3
aI13
aI9
aI59
aI93
aI137
aI103
aI517
aI1555
aI13
aI7965
aI13629
aI14339
aI37425
aI65891
asS'd'
I5846
sa(dp11692
S'a'
I1054
sS's'
I17
sS'm'
(lp11693
I1
aI3
aI7
aI1
aI31
aI31
aI87
aI237
aI365
aI951
aI267
aI2019
aI5085
aI6133
aI29371
aI50319
aI94313
asS'd'
I5847
sa(dp11694
S'a'
I1064
sS's'
I17
sS'm'
(lp11695
I1
aI3
aI5
aI7
aI17
aI19
aI23
aI225
aI501
aI189
aI1291
aI603
aI6873
aI8633
aI11425
aI30565
aI26355
asS'd'
I5848
sa(dp11696
S'a'
I1067
sS's'
I17
sS'm'
(lp11697
I1
aI3
aI7
aI11
aI23
aI17
aI91
aI111
aI415
aI225
aI1287
aI2081
aI4683
aI12069
aI3627
aI32281
aI17995
asS'd'
I5849
sa(dp11698
S'a'
I1082
sS's'
I17
sS'm'
(lp11699
I1
aI1
aI5
aI15
aI25
aI59
aI75
aI203
aI179
aI405
aI1711
aI3147
aI7483
aI5583
aI3729
aI11765
aI61019
asS'd'
I5850
sa(dp11700
S'a'
I1096
sS's'
I17
sS'm'
(lp11701
I1
aI3
aI3
aI9
aI3
aI43
aI65
aI7
aI269
aI33
aI829
aI1789
aI967
aI13119
aI26329
aI16937
aI18533
asS'd'
I5851
sa(dp11702
S'a'
I1116
sS's'
I17
sS'm'
(lp11703
I1
aI1
aI3
aI15
aI11
aI39
aI73
aI11
aI31
aI143
aI1913
aI1227
aI1363
aI11831
aI28687
aI50489
aI106373
asS'd'
I5852
sa(dp11704
S'a'
I1119
sS's'
I17
sS'm'
(lp11705
I1
aI1
aI3
aI3
aI25
aI19
aI15
aI11
aI349
aI1011
aI421
aI3193
aI3665
aI6149
aI20729
aI6997
aI51437
asS'd'
I5853
sa(dp11706
S'a'
I1129
sS's'
I17
sS'm'
(lp11707
I1
aI3
aI5
aI9
aI13
aI63
aI73
aI55
aI417
aI223
aI1753
aI2913
aI4809
aI3947
aI10769
aI5751
aI93867
asS'd'
I5854
sa(dp11708
S'a'
I1130
sS's'
I17
sS'm'
(lp11709
I1
aI3
aI7
aI13
aI31
aI39
aI39
aI133
aI483
aI839
aI1137
aI3303
aI7285
aI4309
aI24079
aI60529
aI103337
asS'd'
I5855
sa(dp11710
S'a'
I1132
sS's'
I17
sS'm'
(lp11711
I1
aI1
aI3
aI7
aI1
aI55
aI3
aI253
aI435
aI589
aI1949
aI1461
aI513
aI381
aI29455
aI4263
aI16831
asS'd'
I5856
sa(dp11712
S'a'
I1137
sS's'
I17
sS'm'
(lp11713
I1
aI1
aI1
aI15
aI25
aI19
aI77
aI101
aI299
aI187
aI1021
aI1533
aI8021
aI4165
aI2277
aI18927
aI110439
asS'd'
I5857
sa(dp11714
S'a'
I1147
sS's'
I17
sS'm'
(lp11715
I1
aI1
aI1
aI11
aI9
aI35
aI71
aI159
aI409
aI527
aI15
aI4073
aI5749
aI8563
aI2503
aI53015
aI111581
asS'd'
I5858
sa(dp11716
S'a'
I1150
sS's'
I17
sS'm'
(lp11717
I1
aI1
aI7
aI5
aI21
aI47
aI113
aI23
aI477
aI559
aI543
aI409
aI4701
aI11479
aI30761
aI8373
aI87777
asS'd'
I5859
sa(dp11718
S'a'
I1154
sS's'
I17
sS'm'
(lp11719
I1
aI3
aI5
aI13
aI9
aI27
aI25
aI137
aI81
aI37
aI799
aI857
aI3539
aI4471
aI15753
aI59015
aI48589
asS'd'
I5860
sa(dp11720
S'a'
I1165
sS's'
I17
sS'm'
(lp11721
I1
aI1
aI3
aI7
aI11
aI57
aI103
aI83
aI209
aI71
aI193
aI3251
aI4839
aI13959
aI32009
aI6471
aI23631
asS'd'
I5861
sa(dp11722
S'a'
I1166
sS's'
I17
sS'm'
(lp11723
I1
aI1
aI7
aI11
aI25
aI33
aI85
aI31
aI371
aI253
aI1667
aI1627
aI6159
aI10039
aI15177
aI52121
aI39475
asS'd'
I5862
sa(dp11724
S'a'
I1174
sS's'
I17
sS'm'
(lp11725
I1
aI1
aI5
aI9
aI13
aI55
aI37
aI13
aI95
aI113
aI1895
aI1525
aI1907
aI6361
aI5863
aI27767
aI108143
asS'd'
I5863
sa(dp11726
S'a'
I1177
sS's'
I17
sS'm'
(lp11727
I1
aI1
aI3
aI13
aI21
aI5
aI53
aI39
aI485
aI171
aI1355
aI2117
aI3127
aI6467
aI31697
aI45343
aI111477
asS'd'
I5864
sa(dp11728
S'a'
I1184
sS's'
I17
sS'm'
(lp11729
I1
aI1
aI7
aI15
aI13
aI57
aI11
aI231
aI329
aI703
aI1823
aI2983
aI215
aI2835
aI19719
aI56637
aI126169
asS'd'
I5865
sa(dp11730
S'a'
I1194
sS's'
I17
sS'm'
(lp11731
I1
aI3
aI5
aI15
aI13
aI51
aI13
aI173
aI301
aI867
aI127
aI2391
aI2795
aI4945
aI13293
aI49947
aI10765
asS'd'
I5866
sa(dp11732
S'a'
I1204
sS's'
I17
sS'm'
(lp11733
I1
aI3
aI3
aI9
aI23
aI5
aI29
aI165
aI467
aI599
aI1181
aI3213
aI4069
aI5473
aI8937
aI51495
aI42611
asS'd'
I5867
sa(dp11734
S'a'
I1208
sS's'
I17
sS'm'
(lp11735
I1
aI1
aI7
aI15
aI5
aI5
aI31
aI125
aI397
aI519
aI1465
aI115
aI7877
aI7025
aI14213
aI50343
aI85827
asS'd'
I5868
sa(dp11736
S'a'
I1213
sS's'
I17
sS'm'
(lp11737
I1
aI3
aI7
aI3
aI25
aI59
aI95
aI103
aI101
aI347
aI95
aI3
aI1251
aI15109
aI12615
aI7511
aI56789
asS'd'
I5869
sa(dp11738
S'a'
I1219
sS's'
I17
sS'm'
(lp11739
I1
aI3
aI5
aI9
aI13
aI59
aI71
aI19
aI107
aI73
aI345
aI3177
aI6519
aI2407
aI18033
aI31075
aI113185
asS'd'
I5870
sa(dp11740
S'a'
I1233
sS's'
I17
sS'm'
(lp11741
I1
aI1
aI1
aI3
aI27
aI37
aI5
aI219
aI169
aI149
aI355
aI549
aI1811
aI11351
aI22627
aI53931
aI88619
asS'd'
I5871
sa(dp11742
S'a'
I1264
sS's'
I17
sS'm'
(lp11743
I1
aI3
aI1
aI3
aI27
aI7
aI9
aI97
aI399
aI947
aI1393
aI3917
aI5439
aI15845
aI19465
aI30123
aI69099
asS'd'
I5872
sa(dp11744
S'a'
I1267
sS's'
I17
sS'm'
(lp11745
I1
aI1
aI7
aI9
aI13
aI25
aI107
aI45
aI111
aI409
aI967
aI3359
aI2499
aI1703
aI20763
aI45187
aI16265
asS'd'
I5873
sa(dp11746
S'a'
I1281
sS's'
I17
sS'm'
(lp11747
I1
aI1
aI1
aI13
aI5
aI49
aI43
aI249
aI49
aI947
aI597
aI1773
aI2387
aI2693
aI15297
aI57969
aI53385
asS'd'
I5874
sa(dp11748
S'a'
I1312
sS's'
I17
sS'm'
(lp11749
I1
aI1
aI7
aI15
aI27
aI25
aI27
aI121
aI421
aI781
aI143
aI817
aI7335
aI14211
aI139
aI55601
aI56671
asS'd'
I5875
sa(dp11750
S'a'
I1321
sS's'
I17
sS'm'
(lp11751
I1
aI3
aI1
aI5
aI29
aI47
aI77
aI23
aI413
aI931
aI785
aI1221
aI769
aI13131
aI26955
aI56441
aI85745
asS'd'
I5876
sa(dp11752
S'a'
I1330
sS's'
I17
sS'm'
(lp11753
I1
aI1
aI1
aI11
aI27
aI3
aI53
aI21
aI467
aI43
aI1533
aI1053
aI691
aI6369
aI8325
aI51087
aI71261
asS'd'
I5877
sa(dp11754
S'a'
I1332
sS's'
I17
sS'm'
(lp11755
I1
aI1
aI3
aI15
aI7
aI9
aI43
aI225
aI293
aI143
aI1049
aI3095
aI6119
aI3165
aI9913
aI26023
aI62657
asS'd'
I5878
sa(dp11756
S'a'
I1335
sS's'
I17
sS'm'
(lp11757
I1
aI3
aI7
aI9
aI11
aI39
aI99
aI193
aI217
aI941
aI259
aI3811
aI6757
aI281
aI10377
aI46961
aI48949
asS'd'
I5879
sa(dp11758
S'a'
I1341
sS's'
I17
sS'm'
(lp11759
I1
aI1
aI1
aI1
aI25
aI1
aI99
aI61
aI495
aI861
aI2013
aI487
aI2821
aI12921
aI30111
aI27213
aI97363
asS'd'
I5880
sa(dp11760
S'a'
I1356
sS's'
I17
sS'm'
(lp11761
I1
aI1
aI5
aI9
aI23
aI33
aI103
aI237
aI161
aI721
aI2021
aI159
aI995
aI475
aI20615
aI30961
aI31767
asS'd'
I5881
sa(dp11762
S'a'
I1371
sS's'
I17
sS'm'
(lp11763
I1
aI3
aI1
aI1
aI5
aI59
aI63
aI139
aI451
aI789
aI1285
aI655
aI5501
aI273
aI21061
aI35937
aI20811
asS'd'
I5882
sa(dp11764
S'a'
I1377
sS's'
I17
sS'm'
(lp11765
I1
aI3
aI3
aI9
aI9
aI15
aI121
aI233
aI287
aI929
aI1605
aI1243
aI417
aI1695
aI29903
aI28699
aI85981
asS'd'
I5883
sa(dp11766
S'a'
I1380
sS's'
I17
sS'm'
(lp11767
I1
aI3
aI3
aI5
aI7
aI25
aI27
aI253
aI469
aI255
aI285
aI2467
aI4897
aI4079
aI29759
aI50351
aI76451
asS'd'
I5884
sa(dp11768
S'a'
I1384
sS's'
I17
sS'm'
(lp11769
I1
aI1
aI3
aI3
aI5
aI33
aI29
aI209
aI291
aI967
aI1429
aI1953
aI5957
aI14065
aI8875
aI32675
aI4629
asS'd'
I5885
sa(dp11770
S'a'
I1395
sS's'
I17
sS'm'
(lp11771
I1
aI3
aI5
aI9
aI7
aI31
aI97
aI21
aI177
aI485
aI1115
aI4051
aI6683
aI7761
aI30181
aI37531
aI51789
asS'd'
I5886
sa(dp11772
S'a'
I1397
sS's'
I17
sS'm'
(lp11773
I1
aI1
aI7
aI3
aI25
aI51
aI23
aI183
aI57
aI699
aI1245
aI2519
aI2783
aI4457
aI6381
aI43199
aI40071
asS'd'
I5887
sa(dp11774
S'a'
I1411
sS's'
I17
sS'm'
(lp11775
I1
aI3
aI5
aI5
aI19
aI55
aI45
aI101
aI299
aI461
aI1009
aI319
aI7335
aI7769
aI5479
aI61113
aI7937
asS'd'
I5888
sa(dp11776
S'a'
I1414
sS's'
I17
sS'm'
(lp11777
I1
aI1
aI7
aI3
aI29
aI21
aI55
aI55
aI437
aI771
aI363
aI683
aI4299
aI15569
aI13813
aI40663
aI86285
asS'd'
I5889
sa(dp11778
S'a'
I1426
sS's'
I17
sS'm'
(lp11779
I1
aI1
aI1
aI13
aI31
aI35
aI93
aI175
aI451
aI387
aI1145
aI3367
aI3833
aI13495
aI11019
aI48925
aI85721
asS'd'
I5890
sa(dp11780
S'a'
I1432
sS's'
I17
sS'm'
(lp11781
I1
aI1
aI7
aI15
aI31
aI21
aI55
aI205
aI117
aI895
aI535
aI2627
aI1473
aI10779
aI24493
aI42999
aI130805
asS'd'
I5891
sa(dp11782
S'a'
I1435
sS's'
I17
sS'm'
(lp11783
I1
aI1
aI3
aI13
aI27
aI11
aI45
aI37
aI193
aI237
aI1505
aI1405
aI3613
aI9565
aI3037
aI53643
aI85211
asS'd'
I5892
sa(dp11784
S'a'
I1437
sS's'
I17
sS'm'
(lp11785
I1
aI1
aI3
aI13
aI9
aI17
aI19
aI27
aI117
aI503
aI65
aI1033
aI7891
aI4005
aI9229
aI20999
aI96601
asS'd'
I5893
sa(dp11786
S'a'
I1442
sS's'
I17
sS'm'
(lp11787
I1
aI3
aI3
aI5
aI17
aI3
aI71
aI79
aI145
aI985
aI935
aI3997
aI6239
aI12511
aI13895
aI65031
aI126383
asS'd'
I5894
sa(dp11788
S'a'
I1454
sS's'
I17
sS'm'
(lp11789
I1
aI1
aI5
aI1
aI23
aI55
aI3
aI105
aI71
aI243
aI1479
aI111
aI7103
aI10753
aI26193
aI35833
aI14583
asS'd'
I5895
sa(dp11790
S'a'
I1468
sS's'
I17
sS'm'
(lp11791
I1
aI3
aI3
aI3
aI15
aI3
aI73
aI125
aI267
aI29
aI1775
aI1437
aI8091
aI10891
aI25731
aI54381
aI12821
asS'd'
I5896
sa(dp11792
S'a'
I1473
sS's'
I17
sS'm'
(lp11793
I1
aI1
aI1
aI3
aI23
aI15
aI67
aI123
aI401
aI347
aI807
aI1097
aI31
aI11209
aI8727
aI58149
aI129099
asS'd'
I5897
sa(dp11794
S'a'
I1488
sS's'
I17
sS'm'
(lp11795
I1
aI3
aI3
aI7
aI7
aI61
aI49
aI129
aI423
aI535
aI135
aI3587
aI233
aI4509
aI23209
aI59203
aI41297
asS'd'
I5898
sa(dp11796
S'a'
I1491
sS's'
I17
sS'm'
(lp11797
I1
aI3
aI1
aI7
aI5
aI29
aI65
aI31
aI335
aI855
aI835
aI1421
aI3081
aI14219
aI16321
aI48269
aI41603
asS'd'
I5899
sa(dp11798
S'a'
I1509
sS's'
I17
sS'm'
(lp11799
I1
aI1
aI1
aI13
aI3
aI21
aI5
aI117
aI163
aI603
aI1519
aI3789
aI7873
aI10981
aI4615
aI9165
aI83929
asS'd'
I5900
sa(dp11800
S'a'
I1524
sS's'
I17
sS'm'
(lp11801
I1
aI3
aI5
aI11
aI15
aI21
aI75
aI151
aI193
aI757
aI647
aI1603
aI333
aI10515
aI22771
aI55459
aI3315
asS'd'
I5901
sa(dp11802
S'a'
I1533
sS's'
I17
sS'm'
(lp11803
I1
aI1
aI7
aI1
aI27
aI3
aI63
aI197
aI271
aI175
aI1599
aI2119
aI1031
aI8671
aI10893
aI35641
aI94535
asS'd'
I5902
sa(dp11804
S'a'
I1555
sS's'
I17
sS'm'
(lp11805
I1
aI1
aI1
aI15
aI1
aI59
aI93
aI17
aI5
aI213
aI1663
aI941
aI435
aI8107
aI1963
aI34951
aI106181
asS'd'
I5903
sa(dp11806
S'a'
I1567
sS's'
I17
sS'm'
(lp11807
I1
aI1
aI5
aI11
aI13
aI35
aI111
aI97
aI267
aI737
aI2023
aI1301
aI7407
aI11249
aI31785
aI31933
aI20673
asS'd'
I5904
sa(dp11808
S'a'
I1571
sS's'
I17
sS'm'
(lp11809
I1
aI3
aI3
aI15
aI5
aI15
aI29
aI63
aI189
aI687
aI27
aI2005
aI7129
aI11377
aI23175
aI42389
aI30933
asS'd'
I5905
sa(dp11810
S'a'
I1586
sS's'
I17
sS'm'
(lp11811
I1
aI1
aI1
aI9
aI13
aI63
aI7
aI155
aI67
aI291
aI1419
aI755
aI2623
aI4749
aI22971
aI7545
aI55711
asS'd'
I5906
sa(dp11812
S'a'
I1592
sS's'
I17
sS'm'
(lp11813
I1
aI3
aI7
aI7
aI23
aI29
aI83
aI151
aI213
aI201
aI157
aI3051
aI6553
aI6401
aI15931
aI47941
aI22869
asS'd'
I5907
sa(dp11814
S'a'
I1595
sS's'
I17
sS'm'
(lp11815
I1
aI3
aI5
aI5
aI7
aI45
aI33
aI155
aI225
aI25
aI49
aI2419
aI4241
aI6835
aI11401
aI50725
aI118343
asS'd'
I5908
sa(dp11816
S'a'
I1600
sS's'
I17
sS'm'
(lp11817
I1
aI1
aI3
aI13
aI31
aI27
aI37
aI41
aI19
aI375
aI1771
aI1789
aI2313
aI2577
aI12615
aI22715
aI22179
asS'd'
I5909
sa(dp11818
S'a'
I1606
sS's'
I17
sS'm'
(lp11819
I1
aI3
aI1
aI11
aI17
aI53
aI55
aI229
aI235
aI837
aI143
aI3583
aI2789
aI5471
aI6515
aI44565
aI8619
asS'd'
I5910
sa(dp11820
S'a'
I1627
sS's'
I17
sS'm'
(lp11821
I1
aI1
aI5
aI15
aI5
aI17
aI23
aI95
aI217
aI551
aI353
aI27
aI3973
aI2547
aI27903
aI50611
aI72277
asS'd'
I5911
sa(dp11822
S'a'
I1648
sS's'
I17
sS'm'
(lp11823
I1
aI1
aI3
aI7
aI5
aI13
aI41
aI111
aI157
aI215
aI1327
aI3073
aI1871
aI11875
aI24239
aI40527
aI97637
asS'd'
I5912
sa(dp11824
S'a'
I1651
sS's'
I17
sS'm'
(lp11825
I1
aI3
aI1
aI1
aI29
aI63
aI111
aI187
aI369
aI395
aI1197
aI3229
aI4353
aI14715
aI29671
aI50503
aI89321
asS'd'
I5913
sa(dp11826
S'a'
I1654
sS's'
I17
sS'm'
(lp11827
I1
aI3
aI1
aI1
aI5
aI63
aI11
aI39
aI171
aI209
aI463
aI3421
aI3451
aI4453
aI14397
aI2219
aI98261
asS'd'
I5914
sa(dp11828
S'a'
I1667
sS's'
I17
sS'm'
(lp11829
I1
aI3
aI3
aI5
aI1
aI1
aI13
aI101
aI67
aI815
aI1521
aI1543
aI7221
aI7337
aI10765
aI30029
aI47881
asS'd'
I5915
sa(dp11830
S'a'
I1669
sS's'
I17
sS'm'
(lp11831
I1
aI1
aI5
aI7
aI9
aI9
aI33
aI197
aI439
aI893
aI961
aI11
aI4319
aI14265
aI24839
aI33581
aI35531
asS'd'
I5916
sa(dp11832
S'a'
I1674
sS's'
I17
sS'm'
(lp11833
I1
aI3
aI3
aI15
aI29
aI35
aI43
aI229
aI313
aI369
aI955
aI1069
aI2939
aI12623
aI20373
aI1533
aI9105
asS'd'
I5917
sa(dp11834
S'a'
I1687
sS's'
I17
sS'm'
(lp11835
I1
aI3
aI1
aI7
aI21
aI7
aI127
aI243
aI103
aI353
aI859
aI3789
aI4369
aI12063
aI22369
aI14531
aI94289
asS'd'
I5918
sa(dp11836
S'a'
I1698
sS's'
I17
sS'm'
(lp11837
I1
aI3
aI5
aI15
aI1
aI27
aI65
aI127
aI229
aI99
aI627
aI2693
aI7173
aI7305
aI29971
aI7097
aI10113
asS'd'
I5919
sa(dp11838
S'a'
I1710
sS's'
I17
sS'm'
(lp11839
I1
aI1
aI5
aI15
aI3
aI47
aI61
aI29
aI155
aI725
aI1727
aI2667
aI7003
aI16277
aI21983
aI21365
aI129365
asS'd'
I5920
sa(dp11840
S'a'
I1717
sS's'
I17
sS'm'
(lp11841
I1
aI1
aI5
aI7
aI27
aI61
aI115
aI133
aI137
aI661
aI1201
aI2151
aI367
aI3567
aI12885
aI62143
aI53955
asS'd'
I5921
sa(dp11842
S'a'
I1722
sS's'
I17
sS'm'
(lp11843
I1
aI1
aI1
aI11
aI9
aI41
aI113
aI103
aI469
aI687
aI1541
aI3679
aI6833
aI10493
aI32747
aI39909
aI121445
asS'd'
I5922
sa(dp11844
S'a'
I1735
sS's'
I17
sS'm'
(lp11845
I1
aI1
aI7
aI5
aI5
aI5
aI91
aI91
aI5
aI405
aI529
aI3999
aI6783
aI2387
aI16621
aI12919
aI8659
asS'd'
I5923
sa(dp11846
S'a'
I1741
sS's'
I17
sS'm'
(lp11847
I1
aI1
aI7
aI13
aI21
aI47
aI125
aI155
aI83
aI913
aI1833
aI4027
aI6657
aI7031
aI31231
aI58201
aI88943
asS'd'
I5924
sa(dp11848
S'a'
I1749
sS's'
I17
sS'm'
(lp11849
I1
aI3
aI7
aI3
aI17
aI55
aI25
aI29
aI181
aI205
aI1173
aI1081
aI6475
aI5037
aI18461
aI22487
aI114131
asS'd'
I5925
sa(dp11850
S'a'
I1750
sS's'
I17
sS'm'
(lp11851
I1
aI1
aI7
aI7
aI25
aI63
aI101
aI103
aI171
aI191
aI1863
aI3441
aI2515
aI14179
aI30123
aI19145
aI31669
asS'd'
I5926
sa(dp11852
S'a'
I1769
sS's'
I17
sS'm'
(lp11853
I1
aI3
aI7
aI11
aI29
aI49
aI73
aI163
aI415
aI821
aI1809
aI723
aI7049
aI14565
aI4829
aI19395
aI61131
asS'd'
I5927
sa(dp11854
S'a'
I1775
sS's'
I17
sS'm'
(lp11855
I1
aI1
aI7
aI9
aI5
aI25
aI103
aI167
aI381
aI757
aI813
aI471
aI3021
aI6619
aI20929
aI38133
aI129505
asS'd'
I5928
sa(dp11856
S'a'
I1777
sS's'
I17
sS'm'
(lp11857
I1
aI1
aI5
aI13
aI25
aI61
aI59
aI199
aI257
aI999
aI169
aI3289
aI7181
aI2049
aI2185
aI39045
aI102703
asS'd'
I5929
sa(dp11858
S'a'
I1778
sS's'
I17
sS'm'
(lp11859
I1
aI1
aI3
aI1
aI21
aI1
aI111
aI125
aI289
aI33
aI701
aI3491
aI5569
aI8055
aI23149
aI26793
aI102563
asS'd'
I5930
sa(dp11860
S'a'
I1792
sS's'
I17
sS'm'
(lp11861
I1
aI1
aI7
aI3
aI25
aI15
aI105
aI235
aI307
aI201
aI1947
aI699
aI2519
aI10615
aI29345
aI17061
aI112949
asS'd'
I5931
sa(dp11862
S'a'
I1797
sS's'
I17
sS'm'
(lp11863
I1
aI3
aI3
aI15
aI19
aI1
aI93
aI173
aI399
aI13
aI269
aI1189
aI523
aI5145
aI32731
aI54087
aI94123
asS'd'
I5932
sa(dp11864
S'a'
I1802
sS's'
I17
sS'm'
(lp11865
I1
aI3
aI1
aI15
aI9
aI41
aI59
aI79
aI217
aI833
aI1993
aI2429
aI3599
aI6919
aI30911
aI12615
aI67947
asS'd'
I5933
sa(dp11866
S'a'
I1822
sS's'
I17
sS'm'
(lp11867
I1
aI3
aI3
aI13
aI31
aI9
aI95
aI37
aI343
aI955
aI1363
aI3851
aI4091
aI13165
aI15241
aI14853
aI35747
asS'd'
I5934
sa(dp11868
S'a'
I1825
sS's'
I17
sS'm'
(lp11869
I1
aI1
aI3
aI5
aI27
aI39
aI37
aI217
aI385
aI473
aI1997
aI2247
aI7353
aI1503
aI9003
aI15055
aI27289
asS'd'
I5935
sa(dp11870
S'a'
I1831
sS's'
I17
sS'm'
(lp11871
I1
aI3
aI7
aI11
aI1
aI13
aI21
aI243
aI375
aI91
aI1295
aI1661
aI203
aI15251
aI15355
aI16065
aI24183
asS'd'
I5936
sa(dp11872
S'a'
I1838
sS's'
I17
sS'm'
(lp11873
I1
aI3
aI1
aI13
aI11
aI45
aI85
aI5
aI275
aI741
aI1395
aI4011
aI7987
aI16087
aI24113
aI50555
aI128147
asS'd'
I5937
sa(dp11874
S'a'
I1852
sS's'
I17
sS'm'
(lp11875
I1
aI1
aI1
aI7
aI3
aI11
aI13
aI189
aI55
aI151
aI395
aI657
aI807
aI11973
aI26297
aI13043
aI109641
asS'd'
I5938
sa(dp11876
S'a'
I1855
sS's'
I17
sS'm'
(lp11877
I1
aI1
aI7
aI13
aI31
aI19
aI33
aI235
aI491
aI647
aI1115
aI2299
aI6381
aI7525
aI2237
aI36197
aI126457
asS'd'
I5939
sa(dp11878
S'a'
I1860
sS's'
I17
sS'm'
(lp11879
I1
aI3
aI5
aI1
aI21
aI15
aI53
aI231
aI77
aI347
aI969
aI141
aI4501
aI9429
aI1815
aI50887
aI74581
asS'd'
I5940
sa(dp11880
S'a'
I1867
sS's'
I17
sS'm'
(lp11881
I1
aI1
aI1
aI9
aI29
aI43
aI47
aI103
aI327
aI131
aI927
aI441
aI7517
aI7277
aI21065
aI409
aI50351
asS'd'
I5941
sa(dp11882
S'a'
I1869
sS's'
I17
sS'm'
(lp11883
I1
aI1
aI5
aI1
aI11
aI13
aI103
aI157
aI239
aI69
aI1347
aI477
aI5017
aI9723
aI28133
aI65135
aI12359
asS'd'
I5942
sa(dp11884
S'a'
I1875
sS's'
I17
sS'm'
(lp11885
I1
aI1
aI1
aI13
aI17
aI63
aI117
aI189
aI323
aI565
aI927
aI1727
aI5337
aI13243
aI5739
aI31241
aI14209
asS'd'
I5943
sa(dp11886
S'a'
I1882
sS's'
I17
sS'm'
(lp11887
I1
aI1
aI3
aI9
aI29
aI9
aI103
aI61
aI467
aI217
aI1367
aI2405
aI5355
aI5743
aI31469
aI30149
aI98775
asS'd'
I5944
sa(dp11888
S'a'
I1903
sS's'
I17
sS'm'
(lp11889
I1
aI1
aI1
aI15
aI23
aI23
aI17
aI229
aI103
aI583
aI179
aI115
aI7081
aI9437
aI32623
aI62639
aI72391
asS'd'
I5945
sa(dp11890
S'a'
I1908
sS's'
I17
sS'm'
(lp11891
I1
aI1
aI5
aI11
aI11
aI39
aI97
aI209
aI115
aI107
aI593
aI2347
aI1445
aI6179
aI32011
aI8435
aI65847
asS'd'
I5946
sa(dp11892
S'a'
I1917
sS's'
I17
sS'm'
(lp11893
I1
aI3
aI7
aI3
aI29
aI27
aI55
aI111
aI27
aI731
aI995
aI1871
aI5017
aI1485
aI11313
aI2559
aI6561
asS'd'
I5947
sa(dp11894
S'a'
I1927
sS's'
I17
sS'm'
(lp11895
I1
aI3
aI1
aI3
aI27
aI9
aI103
aI247
aI83
aI197
aI517
aI1629
aI2189
aI7255
aI183
aI35111
aI15077
asS'd'
I5948
sa(dp11896
S'a'
I1941
sS's'
I17
sS'm'
(lp11897
I1
aI3
aI7
aI5
aI31
aI37
aI87
aI223
aI343
aI331
aI1361
aI3371
aI2007
aI13235
aI10897
aI63839
aI109837
asS'd'
I5949
sa(dp11898
S'a'
I1945
sS's'
I17
sS'm'
(lp11899
I1
aI1
aI7
aI11
aI17
aI5
aI41
aI197
aI489
aI625
aI1595
aI2663
aI5941
aI14029
aI30999
aI16781
aI116001
asS'd'
I5950
sa(dp11900
S'a'
I1948
sS's'
I17
sS'm'
(lp11901
I1
aI3
aI3
aI7
aI19
aI19
aI61
aI175
aI125
aI609
aI1391
aI147
aI3001
aI4189
aI10133
aI24031
aI46219
asS'd'
I5951
sa(dp11902
S'a'
I1962
sS's'
I17
sS'm'
(lp11903
I1
aI1
aI3
aI13
aI13
aI57
aI117
aI181
aI299
aI939
aI583
aI3151
aI829
aI6561
aI30449
aI12211
aI107879
asS'd'
I5952
sa(dp11904
S'a'
I1975
sS's'
I17
sS'm'
(lp11905
I1
aI1
aI5
aI11
aI23
aI45
aI87
aI115
aI259
aI613
aI1001
aI171
aI57
aI13789
aI22173
aI56837
aI26263
asS'd'
I5953
sa(dp11906
S'a'
I1976
sS's'
I17
sS'm'
(lp11907
I1
aI1
aI3
aI3
aI7
aI43
aI45
aI131
aI87
aI251
aI1411
aI2737
aI2739
aI4595
aI12561
aI12043
aI82885
asS'd'
I5954
sa(dp11908
S'a'
I1987
sS's'
I17
sS'm'
(lp11909
I1
aI3
aI3
aI7
aI19
aI39
aI87
aI223
aI461
aI37
aI283
aI3937
aI6193
aI10887
aI11509
aI41131
aI38359
asS'd'
I5955
sa(dp11910
S'a'
I1993
sS's'
I17
sS'm'
(lp11911
I1
aI3
aI1
aI11
aI11
aI37
aI25
aI133
aI105
aI1013
aI925
aI3301
aI239
aI16295
aI4831
aI8649
aI125767
asS'd'
I5956
sa(dp11912
S'a'
I2004
sS's'
I17
sS'm'
(lp11913
I1
aI3
aI3
aI11
aI25
aI11
aI41
aI155
aI1
aI717
aI1587
aI635
aI279
aI1803
aI14817
aI28669
aI88835
asS'd'
I5957
sa(dp11914
S'a'
I2020
sS's'
I17
sS'm'
(lp11915
I1
aI3
aI3
aI11
aI29
aI17
aI39
aI51
aI13
aI871
aI1197
aI2561
aI6671
aI8465
aI22709
aI15933
aI15923
asS'd'
I5958
sa(dp11916
S'a'
I2029
sS's'
I17
sS'm'
(lp11917
I1
aI3
aI7
aI1
aI13
aI17
aI57
aI43
aI267
aI261
aI901
aI241
aI3767
aI15053
aI11017
aI36321
aI72497
asS'd'
I5959
sa(dp11918
S'a'
I2030
sS's'
I17
sS'm'
(lp11919
I1
aI3
aI1
aI15
aI23
aI13
aI17
aI63
aI171
aI919
aI1387
aI2673
aI7605
aI8523
aI14807
aI21187
aI56057
asS'd'
I5960
sa(dp11920
S'a'
I2038
sS's'
I17
sS'm'
(lp11921
I1
aI3
aI7
aI15
aI23
aI41
aI85
aI95
aI53
aI629
aI1877
aI3167
aI2411
aI9619
aI24621
aI31213
aI30069
asS'd'
I5961
sa(dp11922
S'a'
I2041
sS's'
I17
sS'm'
(lp11923
I1
aI1
aI5
aI3
aI3
aI25
aI99
aI39
aI321
aI549
aI599
aI1279
aI2401
aI2335
aI8227
aI59429
aI94549
asS'd'
I5962
sa(dp11924
S'a'
I2048
sS's'
I17
sS'm'
(lp11925
I1
aI3
aI3
aI11
aI9
aI21
aI29
aI55
aI477
aI19
aI1275
aI29
aI2253
aI11421
aI30401
aI57059
aI93219
asS'd'
I5963
sa(dp11926
S'a'
I2054
sS's'
I17
sS'm'
(lp11927
I1
aI1
aI7
aI1
aI27
aI13
aI117
aI249
aI463
aI769
aI281
aI515
aI7467
aI11507
aI1621
aI39765
aI31109
asS'd'
I5964
sa(dp11928
S'a'
I2057
sS's'
I17
sS'm'
(lp11929
I1
aI3
aI5
aI7
aI19
aI7
aI77
aI107
aI23
aI895
aI1013
aI2701
aI3805
aI7327
aI27247
aI6119
aI102395
asS'd'
I5965
sa(dp11930
S'a'
I2058
sS's'
I17
sS'm'
(lp11931
I1
aI1
aI3
aI13
aI21
aI49
aI99
aI15
aI163
aI641
aI1703
aI3061
aI163
aI4265
aI32571
aI13957
aI75005
asS'd'
I5966
sa(dp11932
S'a'
I2068
sS's'
I17
sS'm'
(lp11933
I1
aI1
aI5
aI11
aI27
aI17
aI87
aI169
aI427
aI959
aI361
aI1023
aI5727
aI16279
aI1099
aI39081
aI67215
asS'd'
I5967
sa(dp11934
S'a'
I2072
sS's'
I17
sS'm'
(lp11935
I1
aI3
aI3
aI9
aI23
aI13
aI1
aI91
aI173
aI325
aI1881
aI1385
aI8023
aI935
aI9221
aI19673
aI36949
asS'd'
I5968
sa(dp11936
S'a'
I2075
sS's'
I17
sS'm'
(lp11937
I1
aI3
aI1
aI7
aI7
aI25
aI119
aI189
aI107
aI249
aI811
aI973
aI6499
aI101
aI11281
aI55227
aI32361
asS'd'
I5969
sa(dp11938
S'a'
I2077
sS's'
I17
sS'm'
(lp11939
I1
aI1
aI5
aI13
aI19
aI37
aI117
aI95
aI463
aI587
aI1419
aI445
aI4019
aI7257
aI29757
aI50773
aI52247
asS'd'
I5970
sa(dp11940
S'a'
I2082
sS's'
I17
sS'm'
(lp11941
I1
aI1
aI1
aI1
aI17
aI57
aI81
aI57
aI43
aI789
aI1035
aI625
aI1707
aI9683
aI3681
aI12411
aI110623
asS'd'
I5971
sa(dp11942
S'a'
I2084
sS's'
I17
sS'm'
(lp11943
I1
aI1
aI7
aI5
aI7
aI57
aI49
aI91
aI459
aI513
aI1869
aI3377
aI139
aI10037
aI24091
aI54247
aI41279
asS'd'
I5972
sa(dp11944
S'a'
I2087
sS's'
I17
sS'm'
(lp11945
I1
aI3
aI3
aI9
aI9
aI33
aI29
aI51
aI355
aI415
aI1907
aI809
aI6543
aI349
aI18507
aI12919
aI41667
asS'd'
I5973
sa(dp11946
S'a'
I2101
sS's'
I17
sS'm'
(lp11947
I1
aI1
aI5
aI11
aI3
aI17
aI73
aI201
aI121
aI909
aI1623
aI799
aI3271
aI9051
aI5717
aI15169
aI127861
asS'd'
I5974
sa(dp11948
S'a'
I2111
sS's'
I17
sS'm'
(lp11949
I1
aI1
aI7
aI7
aI23
aI31
aI1
aI155
aI475
aI87
aI2001
aI2459
aI1285
aI5931
aI6803
aI56757
aI71671
asS'd'
I5975
sa(dp11950
S'a'
I2113
sS's'
I17
sS'm'
(lp11951
I1
aI1
aI5
aI13
aI5
aI1
aI21
aI109
aI263
aI841
aI723
aI1539
aI7529
aI433
aI23721
aI33195
aI57001
asS'd'
I5976
sa(dp11952
S'a'
I2119
sS's'
I17
sS'm'
(lp11953
I1
aI3
aI3
aI13
aI29
aI55
aI105
aI231
aI405
aI265
aI671
aI351
aI4693
aI9033
aI21963
aI52073
aI125131
asS'd'
I5977
sa(dp11954
S'a'
I2147
sS's'
I17
sS'm'
(lp11955
I1
aI3
aI1
aI13
aI25
aI51
aI55
aI227
aI245
aI983
aI251
aI2553
aI2017
aI1381
aI31461
aI3953
aI75775
asS'd'
I5978
sa(dp11956
S'a'
I2154
sS's'
I17
sS'm'
(lp11957
I1
aI1
aI1
aI11
aI31
aI11
aI91
aI91
aI287
aI749
aI1019
aI4055
aI3237
aI6965
aI14765
aI1663
aI82987
asS'd'
I5979
sa(dp11958
S'a'
I2161
sS's'
I17
sS'm'
(lp11959
I1
aI1
aI7
aI3
aI11
aI15
aI67
aI161
aI79
aI729
aI1115
aI3713
aI2715
aI9361
aI9365
aI26093
aI63409
asS'd'
I5980
sa(dp11960
S'a'
I2164
sS's'
I17
sS'm'
(lp11961
I1
aI3
aI1
aI7
aI1
aI51
aI125
aI15
aI457
aI433
aI405
aI2329
aI157
aI4817
aI25867
aI38177
aI45319
asS'd'
I5981
sa(dp11962
S'a'
I2177
sS's'
I17
sS'm'
(lp11963
I1
aI3
aI7
aI9
aI25
aI57
aI5
aI233
aI481
aI781
aI1313
aI3179
aI7219
aI8717
aI14825
aI16079
aI127149
asS'd'
I5982
sa(dp11964
S'a'
I2178
sS's'
I17
sS'm'
(lp11965
I1
aI1
aI7
aI15
aI27
aI51
aI5
aI65
aI77
aI313
aI1751
aI1489
aI4307
aI10541
aI11345
aI52577
aI18143
asS'd'
I5983
sa(dp11966
S'a'
I2184
sS's'
I17
sS'm'
(lp11967
I1
aI1
aI1
aI15
aI21
aI5
aI113
aI71
aI411
aI327
aI1681
aI1023
aI5661
aI15815
aI5387
aI10351
aI21121
asS'd'
I5984
sa(dp11968
S'a'
I2198
sS's'
I17
sS'm'
(lp11969
I1
aI1
aI5
aI5
aI29
aI55
aI25
aI255
aI69
aI879
aI501
aI1915
aI3731
aI633
aI12197
aI5249
aI31129
asS'd'
I5985
sa(dp11970
S'a'
I2201
sS's'
I17
sS'm'
(lp11971
I1
aI3
aI5
aI7
aI3
aI23
aI107
aI163
aI485
aI853
aI359
aI3069
aI4353
aI371
aI6027
aI53239
aI105541
asS'd'
I5986
sa(dp11972
S'a'
I2213
sS's'
I17
sS'm'
(lp11973
I1
aI3
aI5
aI15
aI7
aI41
aI9
aI47
aI33
aI327
aI621
aI147
aI577
aI29
aI14623
aI3403
aI9791
asS'd'
I5987
sa(dp11974
S'a'
I2217
sS's'
I17
sS'm'
(lp11975
I1
aI3
aI3
aI15
aI29
aI47
aI41
aI149
aI477
aI127
aI573
aI877
aI3101
aI5963
aI28457
aI14231
aI67425
asS'd'
I5988
sa(dp11976
S'a'
I2228
sS's'
I17
sS'm'
(lp11977
I1
aI1
aI1
aI15
aI31
aI7
aI55
aI191
aI101
aI259
aI1071
aI219
aI2233
aI3583
aI21969
aI32745
aI80529
asS'd'
I5989
sa(dp11978
S'a'
I2240
sS's'
I17
sS'm'
(lp11979
I1
aI3
aI7
aI13
aI17
aI53
aI115
aI69
aI241
aI71
aI1475
aI191
aI509
aI3721
aI15537
aI53773
aI18005
asS'd'
I5990
sa(dp11980
S'a'
I2245
sS's'
I17
sS'm'
(lp11981
I1
aI1
aI3
aI9
aI5
aI57
aI13
aI95
aI103
aI871
aI2043
aI2239
aI7833
aI10727
aI6513
aI55273
aI3781
asS'd'
I5991
sa(dp11982
S'a'
I2250
sS's'
I17
sS'm'
(lp11983
I1
aI1
aI5
aI5
aI9
aI11
aI55
aI151
aI239
aI537
aI135
aI2779
aI7393
aI15393
aI11097
aI58593
aI100745
asS'd'
I5992
sa(dp11984
S'a'
I2263
sS's'
I17
sS'm'
(lp11985
I1
aI1
aI1
aI9
aI15
aI39
aI29
aI105
aI441
aI181
aI1113
aI2125
aI8145
aI11045
aI6589
aI33603
aI83377
asS'd'
I5993
sa(dp11986
S'a'
I2267
sS's'
I17
sS'm'
(lp11987
I1
aI3
aI1
aI1
aI11
aI63
aI69
aI153
aI225
aI845
aI675
aI407
aI4691
aI13383
aI27359
aI38881
aI5509
asS'd'
I5994
sa(dp11988
S'a'
I2285
sS's'
I17
sS'm'
(lp11989
I1
aI3
aI7
aI11
aI23
aI31
aI69
aI3
aI41
aI57
aI683
aI887
aI6861
aI12161
aI14537
aI27293
aI113001
asS'd'
I5995
sa(dp11990
S'a'
I2286
sS's'
I17
sS'm'
(lp11991
I1
aI1
aI1
aI11
aI5
aI1
aI101
aI175
aI437
aI3
aI1477
aI1005
aI6607
aI7429
aI7213
aI4025
aI66479
asS'd'
I5996
sa(dp11992
S'a'
I2291
sS's'
I17
sS'm'
(lp11993
I1
aI1
aI7
aI5
aI19
aI7
aI99
aI131
aI273
aI977
aI1717
aI3831
aI175
aI5673
aI12577
aI36787
aI30945
asS'd'
I5997
sa(dp11994
S'a'
I2298
sS's'
I17
sS'm'
(lp11995
I1
aI3
aI1
aI1
aI15
aI37
aI105
aI195
aI61
aI869
aI255
aI2625
aI7401
aI9361
aI13217
aI52811
aI130811
asS'd'
I5998
sa(dp11996
S'a'
I2306
sS's'
I17
sS'm'
(lp11997
I1
aI3
aI5
aI3
aI29
aI27
aI105
aI23
aI511
aI813
aI1311
aI2859
aI1647
aI1949
aI1329
aI27589
aI125209
asS'd'
I5999
sa(dp11998
S'a'
I2325
sS's'
I17
sS'm'
(lp11999
I1
aI3
aI3
aI1
aI21
aI11
aI119
aI247
aI123
aI401
aI409
aI1845
aI2133
aI10793
aI221
aI43217
aI14069
asS'd'
I6000
sa(dp12000
S'a'
I2329
sS's'
I17
sS'm'
(lp12001
I1
aI1
aI5
aI1
aI29
aI21
aI51
aI73
aI501
aI861
aI725
aI249
aI4249
aI8029
aI15767
aI11985
aI18637
asS'd'
I6001
sa(dp12002
S'a'
I2332
sS's'
I17
sS'm'
(lp12003
I1
aI1
aI5
aI11
aI19
aI39
aI97
aI65
aI13
aI283
aI489
aI2307
aI5239
aI4161
aI18639
aI60035
aI22405
asS'd'
I6002
sa(dp12004
S'a'
I2335
sS's'
I17
sS'm'
(lp12005
I1
aI3
aI5
aI1
aI3
aI7
aI109
aI27
aI429
aI663
aI1569
aI3001
aI3453
aI8627
aI9719
aI23941
aI110451
asS'd'
I6003
sa(dp12006
S'a'
I2339
sS's'
I17
sS'm'
(lp12007
I1
aI3
aI7
aI5
aI17
aI13
aI125
aI209
aI347
aI95
aI1937
aI1419
aI5661
aI7171
aI20607
aI9777
aI68343
asS'd'
I6004
sa(dp12008
S'a'
I2346
sS's'
I17
sS'm'
(lp12009
I1
aI1
aI1
aI1
aI7
aI41
aI43
aI229
aI57
aI49
aI1863
aI2819
aI3735
aI915
aI1571
aI11603
aI116275
asS'd'
I6005
sa(dp12010
S'a'
I2351
sS's'
I17
sS'm'
(lp12011
I1
aI1
aI7
aI9
aI21
aI27
aI5
aI199
aI181
aI521
aI303
aI1097
aI5427
aI8899
aI30325
aI55457
aI16189
asS'd'
I6006
sa(dp12012
S'a'
I2353
sS's'
I17
sS'm'
(lp12013
I1
aI3
aI3
aI7
aI19
aI41
aI3
aI205
aI279
aI223
aI971
aI633
aI2617
aI13191
aI10193
aI23375
aI62563
asS'd'
I6007
sa(dp12014
S'a'
I2363
sS's'
I17
sS'm'
(lp12015
I1
aI3
aI3
aI13
aI23
aI59
aI85
aI25
aI253
aI405
aI65
aI1625
aI4401
aI4679
aI14381
aI57833
aI30001
asS'd'
I6008
sa(dp12016
S'a'
I2378
sS's'
I17
sS'm'
(lp12017
I1
aI3
aI3
aI3
aI13
aI35
aI11
aI157
aI123
aI397
aI119
aI2513
aI1919
aI14583
aI5469
aI11463
aI94711
asS'd'
I6009
sa(dp12018
S'a'
I2383
sS's'
I17
sS'm'
(lp12019
I1
aI1
aI1
aI7
aI17
aI37
aI83
aI211
aI451
aI939
aI449
aI13
aI6671
aI1457
aI19855
aI15053
aI52327
asS'd'
I6010
sa(dp12020
S'a'
I2391
sS's'
I17
sS'm'
(lp12021
I1
aI1
aI5
aI3
aI9
aI57
aI39
aI183
aI331
aI451
aI1391
aI1865
aI7801
aI14293
aI29069
aI705
aI109497
asS'd'
I6011
sa(dp12022
S'a'
I2401
sS's'
I17
sS'm'
(lp12023
I1
aI3
aI7
aI7
aI23
aI21
aI85
aI81
aI255
aI9
aI1685
aI2879
aI6327
aI12675
aI31657
aI38877
aI74131
asS'd'
I6012
sa(dp12024
S'a'
I2408
sS's'
I17
sS'm'
(lp12025
I1
aI1
aI5
aI9
aI25
aI19
aI41
aI195
aI31
aI555
aI927
aI1445
aI593
aI11067
aI10819
aI17205
aI82037
asS'd'
I6013
sa(dp12026
S'a'
I2414
sS's'
I17
sS'm'
(lp12027
I1
aI3
aI1
aI13
aI1
aI35
aI29
aI71
aI323
aI705
aI53
aI3885
aI6223
aI1319
aI30853
aI59935
aI35949
asS'd'
I6014
sa(dp12028
S'a'
I2419
sS's'
I17
sS'm'
(lp12029
I1
aI1
aI7
aI3
aI27
aI63
aI67
aI31
aI149
aI61
aI1611
aI77
aI4271
aI3161
aI12493
aI38341
aI53837
asS'd'
I6015
sa(dp12030
S'a'
I2428
sS's'
I17
sS'm'
(lp12031
I1
aI1
aI1
aI15
aI27
aI53
aI31
aI249
aI429
aI925
aI1485
aI1855
aI4421
aI5703
aI10097
aI14827
aI36685
asS'd'
I6016
sa(dp12032
S'a'
I2441
sS's'
I17
sS'm'
(lp12033
I1
aI3
aI7
aI13
aI7
aI63
aI53
aI9
aI317
aI485
aI1679
aI3631
aI3745
aI5643
aI21615
aI45129
aI48027
asS'd'
I6017
sa(dp12034
S'a'
I2444
sS's'
I17
sS'm'
(lp12035
I1
aI1
aI1
aI1
aI17
aI43
aI19
aI163
aI441
aI847
aI937
aI959
aI6649
aI13071
aI1065
aI55193
aI129509
asS'd'
I6018
sa(dp12036
S'a'
I2461
sS's'
I17
sS'm'
(lp12037
I1
aI1
aI1
aI11
aI29
aI47
aI9
aI215
aI397
aI637
aI961
aI3139
aI2007
aI12603
aI27657
aI22825
aI72873
asS'd'
I6019
sa(dp12038
S'a'
I2480
sS's'
I17
sS'm'
(lp12039
I1
aI3
aI3
aI15
aI7
aI45
aI55
aI163
aI259
aI899
aI951
aI3245
aI4191
aI15813
aI20195
aI8361
aI54025
asS'd'
I6020
sa(dp12040
S'a'
I2483
sS's'
I17
sS'm'
(lp12041
I1
aI1
aI5
aI11
aI3
aI17
aI13
aI223
aI289
aI255
aI875
aI2937
aI1593
aI9729
aI21569
aI63199
aI83875
asS'd'
I6021
sa(dp12042
S'a'
I2486
sS's'
I17
sS'm'
(lp12043
I1
aI1
aI1
aI15
aI19
aI31
aI17
aI129
aI267
aI9
aI2015
aI3233
aI6799
aI12891
aI18473
aI37865
aI19547
asS'd'
I6022
sa(dp12044
S'a'
I2489
sS's'
I17
sS'm'
(lp12045
I1
aI1
aI5
aI5
aI5
aI29
aI81
aI37
aI357
aI539
aI1525
aI2839
aI8041
aI5569
aI4423
aI8907
aI35461
asS'd'
I6023
sa(dp12046
S'a'
I2490
sS's'
I17
sS'm'
(lp12047
I1
aI1
aI5
aI5
aI29
aI11
aI85
aI61
aI333
aI521
aI1111
aI3627
aI325
aI9805
aI17889
aI25655
aI39537
asS'd'
I6024
sa(dp12048
S'a'
I2518
sS's'
I17
sS'm'
(lp12049
I1
aI3
aI5
aI11
aI11
aI53
aI81
aI25
aI79
aI253
aI1963
aI287
aI7487
aI15045
aI21431
aI35417
aI102391
asS'd'
I6025
sa(dp12050
S'a'
I2527
sS's'
I17
sS'm'
(lp12051
I1
aI1
aI1
aI5
aI11
aI33
aI45
aI45
aI425
aI773
aI1817
aI4077
aI1471
aI11655
aI683
aI7115
aI92651
asS'd'
I6026
sa(dp12052
S'a'
I2540
sS's'
I17
sS'm'
(lp12053
I1
aI1
aI3
aI3
aI21
aI13
aI101
aI215
aI311
aI853
aI41
aI1007
aI5511
aI2581
aI25565
aI13155
aI117225
asS'd'
I6027
sa(dp12054
S'a'
I2558
sS's'
I17
sS'm'
(lp12055
I1
aI1
aI3
aI11
aI19
aI9
aI125
aI59
aI273
aI691
aI499
aI1547
aI567
aI10259
aI21963
aI48725
aI3601
asS'd'
I6028
sa(dp12056
S'a'
I2567
sS's'
I17
sS'm'
(lp12057
I1
aI1
aI3
aI7
aI27
aI31
aI39
aI125
aI317
aI625
aI1329
aI3947
aI3943
aI6889
aI2811
aI34055
aI1449
asS'd'
I6029
sa(dp12058
S'a'
I2571
sS's'
I17
sS'm'
(lp12059
I1
aI1
aI1
aI3
aI29
aI45
aI73
aI239
aI319
aI611
aI647
aI1839
aI5277
aI7807
aI3107
aI14683
aI20203
asS'd'
I6030
sa(dp12060
S'a'
I2574
sS's'
I17
sS'm'
(lp12061
I1
aI3
aI3
aI3
aI5
aI5
aI107
aI139
aI103
aI809
aI1343
aI4041
aI3273
aI1789
aI16205
aI47873
aI27803
asS'd'
I6031
sa(dp12062
S'a'
I2579
sS's'
I17
sS'm'
(lp12063
I1
aI3
aI1
aI9
aI21
aI23
aI13
aI131
aI105
aI741
aI1773
aI981
aI5633
aI14609
aI12281
aI50343
aI14317
asS'd'
I6032
sa(dp12064
S'a'
I2585
sS's'
I17
sS'm'
(lp12065
I1
aI1
aI1
aI5
aI11
aI5
aI125
aI171
aI109
aI555
aI159
aI905
aI691
aI12401
aI22817
aI41411
aI70113
asS'd'
I6033
sa(dp12066
S'a'
I2615
sS's'
I17
sS'm'
(lp12067
I1
aI3
aI3
aI9
aI31
aI37
aI109
aI231
aI59
aI615
aI799
aI319
aI2459
aI4521
aI8525
aI4827
aI22969
asS'd'
I6034
sa(dp12068
S'a'
I2639
sS's'
I17
sS'm'
(lp12069
I1
aI3
aI1
aI5
aI11
aI7
aI49
aI237
aI345
aI473
aI981
aI2073
aI6525
aI8805
aI13403
aI3659
aI69897
asS'd'
I6035
sa(dp12070
S'a'
I2641
sS's'
I17
sS'm'
(lp12071
I1
aI3
aI1
aI5
aI9
aI37
aI13
aI203
aI141
aI573
aI745
aI2613
aI5589
aI607
aI24483
aI38427
aI95775
asS'd'
I6036
sa(dp12072
S'a'
I2644
sS's'
I17
sS'm'
(lp12073
I1
aI1
aI3
aI1
aI23
aI61
aI75
aI57
aI299
aI191
aI805
aI2993
aI5175
aI12037
aI13649
aI58831
aI48791
asS'd'
I6037
sa(dp12074
S'a'
I2663
sS's'
I17
sS'm'
(lp12075
I1
aI3
aI7
aI13
aI31
aI57
aI13
aI219
aI185
aI717
aI1607
aI3785
aI4719
aI11583
aI29285
aI48207
aI92021
asS'd'
I6038
sa(dp12076
S'a'
I2667
sS's'
I17
sS'm'
(lp12077
I1
aI3
aI7
aI15
aI23
aI35
aI23
aI69
aI411
aI773
aI1549
aI1087
aI1685
aI15703
aI27193
aI62675
aI43505
asS'd'
I6039
sa(dp12078
S'a'
I2669
sS's'
I17
sS'm'
(lp12079
I1
aI1
aI5
aI3
aI25
aI19
aI97
aI75
aI493
aI549
aI1655
aI2881
aI4989
aI2765
aI4797
aI43143
aI113955
asS'd'
I6040
sa(dp12080
S'a'
I2672
sS's'
I17
sS'm'
(lp12081
I1
aI1
aI5
aI7
aI21
aI5
aI65
aI37
aI383
aI133
aI1907
aI3747
aI1525
aI5803
aI19977
aI50551
aI23157
asS'd'
I6041
sa(dp12082
S'a'
I2687
sS's'
I17
sS'm'
(lp12083
I1
aI1
aI1
aI11
aI15
aI61
aI59
aI109
aI489
aI901
aI1787
aI1611
aI6101
aI10653
aI3071
aI35643
aI56227
asS'd'
I6042
sa(dp12084
S'a'
I2700
sS's'
I17
sS'm'
(lp12085
I1
aI3
aI1
aI5
aI15
aI25
aI121
aI111
aI25
aI251
aI1467
aI1795
aI1631
aI13753
aI32391
aI14831
aI90739
asS'd'
I6043
sa(dp12086
S'a'
I2705
sS's'
I17
sS'm'
(lp12087
I1
aI1
aI1
aI13
aI23
aI55
aI119
aI147
aI45
aI871
aI1389
aI1929
aI1023
aI16131
aI10041
aI40055
aI23337
asS'd'
I6044
sa(dp12088
S'a'
I2724
sS's'
I17
sS'm'
(lp12089
I1
aI3
aI1
aI15
aI27
aI33
aI23
aI41
aI463
aI603
aI1633
aI3445
aI2007
aI5999
aI11175
aI18343
aI13159
asS'd'
I6045
sa(dp12090
S'a'
I2728
sS's'
I17
sS'm'
(lp12091
I1
aI3
aI1
aI9
aI17
aI15
aI107
aI63
aI493
aI411
aI293
aI3669
aI6143
aI3057
aI8253
aI25491
aI58907
asS'd'
I6046
sa(dp12092
S'a'
I2733
sS's'
I17
sS'm'
(lp12093
I1
aI3
aI5
aI11
aI1
aI43
aI5
aI117
aI127
aI813
aI1881
aI3711
aI2567
aI7819
aI5809
aI64471
aI104221
asS'd'
I6047
sa(dp12094
S'a'
I2741
sS's'
I17
sS'm'
(lp12095
I1
aI3
aI5
aI9
aI25
aI27
aI49
aI93
aI77
aI705
aI1773
aI1745
aI4605
aI16137
aI14621
aI62893
aI81637
asS'd'
I6048
sa(dp12096
S'a'
I2748
sS's'
I17
sS'm'
(lp12097
I1
aI3
aI1
aI15
aI9
aI29
aI41
aI101
aI291
aI763
aI1475
aI3185
aI3661
aI10351
aI26645
aI50375
aI59373
asS'd'
I6049
sa(dp12098
S'a'
I2751
sS's'
I17
sS'm'
(lp12099
I1
aI1
aI5
aI15
aI9
aI31
aI107
aI159
aI125
aI471
aI1023
aI2361
aI4805
aI8073
aI21563
aI14903
aI77801
asS'd'
I6050
sa(dp12100
S'a'
I2756
sS's'
I17
sS'm'
(lp12101
I1
aI3
aI7
aI1
aI27
aI17
aI75
aI129
aI71
aI697
aI551
aI1969
aI6597
aI13821
aI2605
aI61783
aI74791
asS'd'
I6051
sa(dp12102
S'a'
I2771
sS's'
I17
sS'm'
(lp12103
I1
aI1
aI7
aI15
aI17
aI27
aI49
aI47
aI59
aI47
aI1671
aI2535
aI1299
aI2387
aI24349
aI23661
aI91123
asS'd'
I6052
sa(dp12104
S'a'
I2774
sS's'
I17
sS'm'
(lp12105
I1
aI1
aI5
aI15
aI21
aI61
aI45
aI37
aI415
aI189
aI143
aI351
aI1815
aI3479
aI2399
aI56753
aI123893
asS'd'
I6053
sa(dp12106
S'a'
I2793
sS's'
I17
sS'm'
(lp12107
I1
aI1
aI3
aI7
aI7
aI19
aI93
aI249
aI335
aI305
aI1437
aI1329
aI2693
aI13201
aI9589
aI61513
aI115995
asS'd'
I6054
sa(dp12108
S'a'
I2796
sS's'
I17
sS'm'
(lp12109
I1
aI1
aI1
aI11
aI21
aI57
aI33
aI205
aI235
aI253
aI751
aI259
aI6029
aI9811
aI10231
aI36899
aI78035
asS'd'
I6055
sa(dp12110
S'a'
I2804
sS's'
I17
sS'm'
(lp12111
I1
aI1
aI1
aI11
aI13
aI25
aI115
aI195
aI111
aI913
aI1851
aI3283
aI6083
aI11717
aI2773
aI40727
aI493
asS'd'
I6056
sa(dp12112
S'a'
I2814
sS's'
I17
sS'm'
(lp12113
I1
aI3
aI3
aI9
aI9
aI17
aI83
aI137
aI465
aI671
aI1277
aI325
aI2767
aI12413
aI21977
aI47525
aI23041
asS'd'
I6057
sa(dp12114
S'a'
I2822
sS's'
I17
sS'm'
(lp12115
I1
aI1
aI1
aI11
aI15
aI47
aI65
aI219
aI271
aI197
aI297
aI3195
aI1325
aI9991
aI26385
aI46055
aI43151
asS'd'
I6058
sa(dp12116
S'a'
I2845
sS's'
I17
sS'm'
(lp12117
I1
aI1
aI1
aI13
aI31
aI21
aI39
aI89
aI127
aI629
aI367
aI2935
aI6259
aI6627
aI15691
aI55781
aI97251
asS'd'
I6059
sa(dp12118
S'a'
I2846
sS's'
I17
sS'm'
(lp12119
I1
aI1
aI7
aI13
aI11
aI45
aI65
aI75
aI211
aI785
aI1221
aI2087
aI7751
aI15619
aI25489
aI28195
aI69007
asS'd'
I6060
sa(dp12120
S'a'
I2850
sS's'
I17
sS'm'
(lp12121
I1
aI3
aI5
aI15
aI27
aI37
aI75
aI111
aI487
aI219
aI233
aI583
aI6433
aI15105
aI355
aI28331
aI21105
asS'd'
I6061
sa(dp12122
S'a'
I2855
sS's'
I17
sS'm'
(lp12123
I1
aI3
aI3
aI15
aI31
aI53
aI33
aI95
aI27
aI197
aI1727
aI1467
aI7115
aI15479
aI26873
aI31075
aI12793
asS'd'
I6062
sa(dp12124
S'a'
I2856
sS's'
I17
sS'm'
(lp12125
I1
aI3
aI7
aI1
aI19
aI3
aI19
aI105
aI225
aI599
aI737
aI107
aI7951
aI10193
aI31699
aI59207
aI85619
asS'd'
I6063
sa(dp12126
S'a'
I2867
sS's'
I17
sS'm'
(lp12127
I1
aI3
aI1
aI3
aI7
aI17
aI73
aI191
aI247
aI421
aI537
aI1473
aI189
aI4219
aI29993
aI25491
aI21189
asS'd'
I6064
sa(dp12128
S'a'
I2891
sS's'
I17
sS'm'
(lp12129
I1
aI3
aI7
aI7
aI13
aI21
aI33
aI95
aI147
aI699
aI943
aI2275
aI4093
aI6067
aI9063
aI25503
aI111085
asS'd'
I6065
sa(dp12130
S'a'
I2894
sS's'
I17
sS'm'
(lp12131
I1
aI1
aI7
aI9
aI13
aI47
aI123
aI121
aI347
aI467
aI225
aI957
aI2329
aI14075
aI29843
aI61753
aI97179
asS'd'
I6066
sa(dp12132
S'a'
I2902
sS's'
I17
sS'm'
(lp12133
I1
aI3
aI3
aI7
aI17
aI55
aI37
aI167
aI215
aI819
aI163
aI1747
aI4485
aI15991
aI28011
aI36351
aI106495
asS'd'
I6067
sa(dp12134
S'a'
I2908
sS's'
I17
sS'm'
(lp12135
I1
aI1
aI3
aI9
aI25
aI5
aI83
aI199
aI209
aI395
aI1757
aI1967
aI5739
aI2573
aI13989
aI32145
aI4847
asS'd'
I6068
sa(dp12136
S'a'
I2951
sS's'
I17
sS'm'
(lp12137
I1
aI3
aI3
aI13
aI11
aI21
aI25
aI223
aI239
aI569
aI1877
aI299
aI8089
aI3697
aI801
aI64775
aI26827
asS'd'
I6069
sa(dp12138
S'a'
I2970
sS's'
I17
sS'm'
(lp12139
I1
aI3
aI5
aI7
aI17
aI9
aI127
aI9
aI65
aI919
aI1073
aI2661
aI1337
aI10065
aI30099
aI30929
aI90067
asS'd'
I6070
sa(dp12140
S'a'
I2972
sS's'
I17
sS'm'
(lp12141
I1
aI3
aI1
aI13
aI25
aI41
aI35
aI251
aI279
aI351
aI111
aI3917
aI2815
aI7989
aI9895
aI54859
aI126355
asS'd'
I6071
sa(dp12142
S'a'
I2975
sS's'
I17
sS'm'
(lp12143
I1
aI1
aI3
aI7
aI17
aI61
aI13
aI73
aI335
aI831
aI703
aI37
aI2765
aI13169
aI12513
aI56301
aI13907
asS'd'
I6072
sa(dp12144
S'a'
I2976
sS's'
I17
sS'm'
(lp12145
I1
aI1
aI5
aI13
aI11
aI15
aI33
aI45
aI505
aI127
aI1723
aI17
aI4927
aI11453
aI28859
aI9671
aI80041
asS'd'
I6073
sa(dp12146
S'a'
I2981
sS's'
I17
sS'm'
(lp12147
I1
aI3
aI1
aI5
aI9
aI1
aI25
aI147
aI281
aI601
aI243
aI2687
aI5533
aI6725
aI11075
aI34807
aI24619
asS'd'
I6074
sa(dp12148
S'a'
I2986
sS's'
I17
sS'm'
(lp12149
I1
aI1
aI3
aI1
aI7
aI21
aI71
aI31
aI485
aI561
aI1361
aI1237
aI8171
aI15885
aI7941
aI4583
aI32851
asS'd'
I6075
sa(dp12150
S'a'
I2999
sS's'
I17
sS'm'
(lp12151
I1
aI3
aI7
aI1
aI5
aI35
aI95
aI155
aI283
aI959
aI577
aI1343
aI4269
aI13481
aI30819
aI40273
aI8711
asS'd'
I6076
sa(dp12152
S'a'
I3000
sS's'
I17
sS'm'
(lp12153
I1
aI3
aI7
aI3
aI1
aI53
aI77
aI45
aI215
aI537
aI1045
aI77
aI2791
aI3553
aI13273
aI23819
aI62263
asS'd'
I6077
sa(dp12154
S'a'
I3006
sS's'
I17
sS'm'
(lp12155
I1
aI3
aI1
aI15
aI29
aI59
aI7
aI145
aI85
aI3
aI251
aI2691
aI7547
aI11241
aI32295
aI24645
aI75739
asS'd'
I6078
sa(dp12156
S'a'
I3014
sS's'
I17
sS'm'
(lp12157
I1
aI1
aI5
aI9
aI19
aI9
aI39
aI163
aI303
aI233
aI2039
aI2027
aI7169
aI2773
aI28649
aI38317
aI66761
asS'd'
I6079
sa(dp12158
S'a'
I3028
sS's'
I17
sS'm'
(lp12159
I1
aI3
aI7
aI5
aI21
aI27
aI93
aI227
aI131
aI1019
aI1619
aI1497
aI4043
aI1131
aI25761
aI20173
aI99957
asS'd'
I6080
sa(dp12160
S'a'
I3031
sS's'
I17
sS'm'
(lp12161
I1
aI3
aI7
aI5
aI19
aI33
aI15
aI173
aI435
aI399
aI531
aI2001
aI3221
aI12627
aI10153
aI24421
aI61805
asS'd'
I6081
sa(dp12162
S'a'
I3035
sS's'
I17
sS'm'
(lp12163
I1
aI3
aI1
aI9
aI11
aI3
aI69
aI105
aI289
aI183
aI1103
aI831
aI2297
aI1613
aI18801
aI54395
aI54243
asS'd'
I6082
sa(dp12164
S'a'
I3037
sS's'
I17
sS'm'
(lp12165
I1
aI3
aI3
aI9
aI3
aI53
aI113
aI183
aI79
aI355
aI1629
aI1061
aI3713
aI4563
aI14365
aI43529
aI56073
asS'd'
I6083
sa(dp12166
S'a'
I3053
sS's'
I17
sS'm'
(lp12167
I1
aI3
aI7
aI11
aI31
aI39
aI107
aI139
aI187
aI873
aI225
aI33
aI4943
aI15837
aI225
aI6407
aI85967
asS'd'
I6084
sa(dp12168
S'a'
I3059
sS's'
I17
sS'm'
(lp12169
I1
aI3
aI1
aI11
aI17
aI47
aI93
aI233
aI119
aI699
aI1429
aI2845
aI2061
aI8887
aI20665
aI45497
aI33107
asS'd'
I6085
sa(dp12170
S'a'
I3065
sS's'
I17
sS'm'
(lp12171
I1
aI3
aI5
aI1
aI25
aI11
aI55
aI75
aI91
aI1009
aI1887
aI3167
aI515
aI15929
aI11659
aI57953
aI63401
asS'd'
I6086
sa(dp12172
S'a'
I3080
sS's'
I17
sS'm'
(lp12173
I1
aI1
aI3
aI15
aI27
aI59
aI103
aI53
aI353
aI553
aI2021
aI1543
aI2785
aI9373
aI14609
aI21213
aI19911
asS'd'
I6087
sa(dp12174
S'a'
I3091
sS's'
I17
sS'm'
(lp12175
I1
aI3
aI7
aI9
aI3
aI1
aI101
aI133
aI437
aI773
aI1399
aI1067
aI7419
aI1793
aI16589
aI3483
aI42065
asS'd'
I6088
sa(dp12176
S'a'
I3094
sS's'
I17
sS'm'
(lp12177
I1
aI3
aI7
aI1
aI25
aI57
aI127
aI113
aI65
aI577
aI1865
aI1527
aI6485
aI11273
aI15803
aI39625
aI75219
asS'd'
I6089
sa(dp12178
S'a'
I3109
sS's'
I17
sS'm'
(lp12179
I1
aI3
aI5
aI9
aI7
aI63
aI29
aI89
aI155
aI45
aI1029
aI2407
aI6783
aI4749
aI4849
aI26639
aI54059
asS'd'
I6090
sa(dp12180
S'a'
I3110
sS's'
I17
sS'm'
(lp12181
I1
aI3
aI7
aI9
aI25
aI13
aI113
aI41
aI267
aI767
aI1071
aI1689
aI269
aI14437
aI21255
aI39473
aI65771
asS'd'
I6091
sa(dp12182
S'a'
I3113
sS's'
I17
sS'm'
(lp12183
I1
aI3
aI1
aI15
aI5
aI3
aI77
aI43
aI391
aI763
aI59
aI1027
aI6263
aI3715
aI31061
aI43311
aI130725
asS'd'
I6092
sa(dp12184
S'a'
I3116
sS's'
I17
sS'm'
(lp12185
I1
aI3
aI7
aI7
aI21
aI51
aI127
aI71
aI229
aI171
aI397
aI1099
aI871
aI2717
aI1643
aI17363
aI125979
asS'd'
I6093
sa(dp12186
S'a'
I3136
sS's'
I17
sS'm'
(lp12187
I1
aI1
aI5
aI15
aI25
aI11
aI11
aI113
aI203
aI795
aI1703
aI3901
aI1113
aI12819
aI25345
aI46691
aI112313
asS'd'
I6094
sa(dp12188
S'a'
I3139
sS's'
I17
sS'm'
(lp12189
I1
aI3
aI7
aI5
aI1
aI59
aI91
aI81
aI325
aI483
aI595
aI1491
aI7455
aI6699
aI199
aI35597
aI59851
asS'd'
I6095
sa(dp12190
S'a'
I3141
sS's'
I17
sS'm'
(lp12191
I1
aI3
aI5
aI1
aI3
aI33
aI43
aI195
aI201
aI575
aI1395
aI1305
aI7001
aI2023
aI22419
aI15233
aI120355
asS'd'
I6096
sa(dp12192
S'a'
I3154
sS's'
I17
sS'm'
(lp12193
I1
aI1
aI3
aI3
aI15
aI37
aI81
aI59
aI87
aI675
aI199
aI3231
aI4473
aI5023
aI16753
aI51475
aI102113
asS'd'
I6097
sa(dp12194
S'a'
I3160
sS's'
I17
sS'm'
(lp12195
I1
aI1
aI7
aI9
aI13
aI39
aI65
aI9
aI51
aI565
aI1171
aI119
aI7875
aI12149
aI6565
aI56849
aI123235
asS'd'
I6098
sa(dp12196
S'a'
I3169
sS's'
I17
sS'm'
(lp12197
I1
aI3
aI3
aI7
aI15
aI45
aI53
aI93
aI111
aI533
aI1849
aI643
aI2265
aI10241
aI24741
aI11559
aI74333
asS'd'
I6099
sa(dp12198
S'a'
I3182
sS's'
I17
sS'm'
(lp12199
I1
aI3
aI1
aI1
aI11
aI61
aI75
aI51
aI5
aI199
aI535
aI279
aI5821
aI6005
aI2907
aI32521
aI74121
asS'd'
I6100
sa(dp12200
S'a'
I3187
sS's'
I17
sS'm'
(lp12201
I1
aI1
aI3
aI15
aI3
aI21
aI29
aI193
aI71
aI993
aI1719
aI1865
aI6135
aI7683
aI12171
aI29275
aI14539
asS'd'
I6101
sa(dp12202
S'a'
I3189
sS's'
I17
sS'm'
(lp12203
I1
aI1
aI1
aI7
aI7
aI13
aI1
aI61
aI315
aI431
aI1145
aI2067
aI5745
aI1641
aI1047
aI55111
aI129477
asS'd'
I6102
sa(dp12204
S'a'
I3190
sS's'
I17
sS'm'
(lp12205
I1
aI1
aI5
aI1
aI21
aI43
aI115
aI193
aI153
aI573
aI1181
aI3947
aI7809
aI11317
aI30649
aI56891
aI47741
asS'd'
I6103
sa(dp12206
S'a'
I3203
sS's'
I17
sS'm'
(lp12207
I1
aI1
aI5
aI7
aI19
aI15
aI61
aI239
aI109
aI683
aI395
aI2869
aI3103
aI1531
aI12019
aI45159
aI37525
asS'd'
I6104
sa(dp12208
S'a'
I3217
sS's'
I17
sS'm'
(lp12209
I1
aI1
aI5
aI7
aI29
aI55
aI45
aI7
aI353
aI659
aI591
aI3371
aI5777
aI8475
aI2743
aI47483
aI11983
asS'd'
I6105
sa(dp12210
S'a'
I3229
sS's'
I17
sS'm'
(lp12211
I1
aI3
aI1
aI3
aI13
aI17
aI39
aI195
aI43
aI5
aI1749
aI2559
aI5843
aI8719
aI21421
aI58511
aI105637
asS'd'
I6106
sa(dp12212
S'a'
I3236
sS's'
I17
sS'm'
(lp12213
I1
aI3
aI5
aI5
aI5
aI21
aI29
aI63
aI387
aI301
aI567
aI3325
aI2109
aI403
aI23053
aI24851
aI14493
asS'd'
I6107
sa(dp12214
S'a'
I3248
sS's'
I17
sS'm'
(lp12215
I1
aI1
aI3
aI3
aI17
aI57
aI107
aI131
aI85
aI855
aI1101
aI3199
aI7159
aI14739
aI4197
aI27943
aI113009
asS'd'
I6108
sa(dp12216
S'a'
I3257
sS's'
I17
sS'm'
(lp12217
I1
aI1
aI3
aI11
aI1
aI61
aI31
aI79
aI33
aI123
aI1509
aI507
aI6679
aI2279
aI8465
aI37279
aI17553
asS'd'
I6109
sa(dp12218
S'a'
I3278
sS's'
I17
sS'm'
(lp12219
I1
aI3
aI1
aI15
aI7
aI33
aI11
aI71
aI217
aI609
aI1661
aI3437
aI5497
aI13365
aI6247
aI649
aI26407
asS'd'
I6110
sa(dp12220
S'a'
I3283
sS's'
I17
sS'm'
(lp12221
I1
aI1
aI3
aI1
aI19
aI45
aI49
aI125
aI5
aI455
aI1669
aI4083
aI253
aI10101
aI27327
aI16401
aI120399
asS'd'
I6111
sa(dp12222
S'a'
I3289
sS's'
I17
sS'm'
(lp12223
I1
aI3
aI1
aI1
aI27
aI19
aI117
aI137
aI261
aI341
aI1697
aI457
aI7553
aI12169
aI30049
aI49281
aI36937
asS'd'
I6112
sa(dp12224
S'a'
I3292
sS's'
I17
sS'm'
(lp12225
I1
aI1
aI1
aI3
aI9
aI49
aI33
aI13
aI461
aI545
aI1537
aI2623
aI883
aI10921
aI5583
aI58997
aI114183
asS'd'
I6113
sa(dp12226
S'a'
I3302
sS's'
I17
sS'm'
(lp12227
I1
aI1
aI7
aI9
aI29
aI53
aI29
aI165
aI205
aI989
aI1347
aI2343
aI7505
aI7609
aI18503
aI51677
aI105993
asS'd'
I6114
sa(dp12228
S'a'
I3316
sS's'
I17
sS'm'
(lp12229
I1
aI1
aI1
aI13
aI1
aI29
aI59
aI121
aI297
aI659
aI1965
aI1765
aI5255
aI10971
aI32613
aI18763
aI41983
asS'd'
I6115
sa(dp12230
S'a'
I3328
sS's'
I17
sS'm'
(lp12231
I1
aI3
aI7
aI11
aI21
aI41
aI19
aI47
aI125
aI485
aI475
aI2745
aI4075
aI8101
aI31227
aI4679
aI115473
asS'd'
I6116
sa(dp12232
S'a'
I3333
sS's'
I17
sS'm'
(lp12233
I1
aI3
aI3
aI7
aI21
aI23
aI55
aI65
aI223
aI1001
aI317
aI1459
aI183
aI5139
aI26553
aI41471
aI116373
asS'd'
I6117
sa(dp12234
S'a'
I3337
sS's'
I17
sS'm'
(lp12235
I1
aI1
aI7
aI3
aI1
aI9
aI29
aI139
aI343
aI913
aI1993
aI3139
aI3791
aI5869
aI6057
aI23863
aI35737
asS'd'
I6118
sa(dp12236
S'a'
I3340
sS's'
I17
sS'm'
(lp12237
I1
aI3
aI3
aI3
aI7
aI21
aI77
aI197
aI239
aI467
aI35
aI591
aI1061
aI3417
aI31811
aI38825
aI124981
asS'd'
I6119
sa(dp12238
S'a'
I3368
sS's'
I17
sS'm'
(lp12239
I1
aI3
aI3
aI1
aI21
aI29
aI5
aI213
aI417
aI111
aI1681
aI1409
aI2899
aI16233
aI1053
aI51235
aI87767
asS'd'
I6120
sa(dp12240
S'a'
I3371
sS's'
I17
sS'm'
(lp12241
I1
aI1
aI5
aI3
aI13
aI47
aI61
aI203
aI223
aI73
aI1947
aI3613
aI5885
aI13567
aI7593
aI34329
aI68597
asS'd'
I6121
sa(dp12242
S'a'
I3376
sS's'
I17
sS'm'
(lp12243
I1
aI3
aI1
aI1
aI17
aI9
aI11
aI187
aI361
aI973
aI781
aI1835
aI1539
aI12917
aI21725
aI48279
aI115037
asS'd'
I6122
sa(dp12244
S'a'
I3385
sS's'
I17
sS'm'
(lp12245
I1
aI3
aI1
aI1
aI9
aI25
aI117
aI157
aI433
aI395
aI403
aI2183
aI3327
aI5427
aI7505
aI2673
aI77137
asS'd'
I6123
sa(dp12246
S'a'
I3386
sS's'
I17
sS'm'
(lp12247
I1
aI1
aI7
aI15
aI31
aI15
aI27
aI155
aI441
aI837
aI1877
aI3829
aI5139
aI16331
aI31183
aI15803
aI95699
asS'd'
I6124
sa(dp12248
S'a'
I3393
sS's'
I17
sS'm'
(lp12249
I1
aI1
aI7
aI15
aI5
aI51
aI77
aI179
aI289
aI727
aI1763
aI2529
aI6715
aI3967
aI29267
aI27293
aI67953
asS'd'
I6125
sa(dp12250
S'a'
I3399
sS's'
I17
sS'm'
(lp12251
I1
aI3
aI7
aI13
aI7
aI3
aI3
aI17
aI311
aI547
aI1465
aI1413
aI3937
aI2725
aI24523
aI12321
aI109763
asS'd'
I6126
sa(dp12252
S'a'
I3405
sS's'
I17
sS'm'
(lp12253
I1
aI3
aI5
aI15
aI9
aI5
aI87
aI135
aI281
aI97
aI2021
aI1903
aI8007
aI10321
aI27989
aI18993
aI110407
asS'd'
I6127
sa(dp12254
S'a'
I3414
sS's'
I17
sS'm'
(lp12255
I1
aI1
aI1
aI13
aI25
aI61
aI89
aI107
aI233
aI823
aI1375
aI3531
aI1757
aI1577
aI29457
aI1461
aI17217
asS'd'
I6128
sa(dp12256
S'a'
I3433
sS's'
I17
sS'm'
(lp12257
I1
aI1
aI1
aI13
aI17
aI17
aI27
aI193
aI485
aI759
aI145
aI3943
aI4183
aI14119
aI11217
aI3793
aI1935
asS'd'
I6129
sa(dp12258
S'a'
I3436
sS's'
I17
sS'm'
(lp12259
I1
aI1
aI1
aI3
aI13
aI31
aI101
aI227
aI311
aI363
aI1925
aI1525
aI5275
aI2385
aI15093
aI48769
aI121189
asS'd'
I6130
sa(dp12260
S'a'
I3448
sS's'
I17
sS'm'
(lp12261
I1
aI1
aI5
aI13
aI11
aI61
aI89
aI141
aI117
aI229
aI417
aI3935
aI7249
aI13869
aI30591
aI62763
aI67521
asS'd'
I6131
sa(dp12262
S'a'
I3467
sS's'
I17
sS'm'
(lp12263
I1
aI1
aI3
aI15
aI7
aI59
aI105
aI239
aI453
aI221
aI1101
aI395
aI2031
aI8941
aI23155
aI7077
aI125593
asS'd'
I6132
sa(dp12264
S'a'
I3469
sS's'
I17
sS'm'
(lp12265
I1
aI1
aI1
aI11
aI7
aI55
aI99
aI31
aI305
aI371
aI1035
aI577
aI4473
aI577
aI371
aI46093
aI69157
asS'd'
I6133
sa(dp12266
S'a'
I3472
sS's'
I17
sS'm'
(lp12267
I1
aI3
aI1
aI9
aI9
aI33
aI35
aI245
aI95
aI47
aI1623
aI2965
aI6849
aI7269
aI5321
aI31641
aI73321
asS'd'
I6134
sa(dp12268
S'a'
I3477
sS's'
I17
sS'm'
(lp12269
I1
aI1
aI1
aI15
aI21
aI61
aI65
aI65
aI159
aI151
aI625
aI2281
aI2993
aI1311
aI29757
aI24703
aI71029
asS'd'
I6135
sa(dp12270
S'a'
I3484
sS's'
I17
sS'm'
(lp12271
I1
aI3
aI5
aI15
aI29
aI59
aI29
aI69
aI351
aI901
aI631
aI3501
aI7031
aI703
aI20805
aI36437
aI94931
asS'd'
I6136
sa(dp12272
S'a'
I3494
sS's'
I17
sS'm'
(lp12273
I1
aI3
aI7
aI1
aI21
aI11
aI19
aI125
aI237
aI807
aI1651
aI2389
aI7347
aI11759
aI27151
aI38669
aI965
asS'd'
I6137
sa(dp12274
S'a'
I3505
sS's'
I17
sS'm'
(lp12275
I1
aI1
aI5
aI1
aI15
aI41
aI1
aI105
aI89
aI127
aI895
aI29
aI2339
aI15951
aI18633
aI2781
aI67269
asS'd'
I6138
sa(dp12276
S'a'
I3515
sS's'
I17
sS'm'
(lp12277
I1
aI1
aI5
aI15
aI25
aI7
aI3
aI33
aI375
aI447
aI203
aI2579
aI6145
aI14015
aI9939
aI52777
aI123181
asS'd'
I6139
sa(dp12278
S'a'
I3523
sS's'
I17
sS'm'
(lp12279
I1
aI3
aI1
aI15
aI29
aI7
aI7
aI27
aI451
aI869
aI107
aI2457
aI5557
aI11601
aI28957
aI36181
aI41419
asS'd'
I6140
sa(dp12280
S'a'
I3530
sS's'
I17
sS'm'
(lp12281
I1
aI1
aI1
aI7
aI1
aI57
aI33
aI213
aI329
aI763
aI815
aI169
aI623
aI155
aI20529
aI20603
aI73311
asS'd'
I6141
sa(dp12282
S'a'
I3543
sS's'
I17
sS'm'
(lp12283
I1
aI3
aI5
aI7
aI25
aI21
aI7
aI217
aI159
aI89
aI1373
aI1735
aI705
aI4093
aI13083
aI3855
aI55875
asS'd'
I6142
sa(dp12284
S'a'
I3559
sS's'
I17
sS'm'
(lp12285
I1
aI3
aI1
aI1
aI29
aI33
aI105
aI127
aI95
aI543
aI235
aI67
aI691
aI5015
aI22139
aI18251
aI89945
asS'd'
I6143
sa(dp12286
S'a'
I3568
sS's'
I17
sS'm'
(lp12287
I1
aI1
aI3
aI11
aI27
aI53
aI105
aI83
aI337
aI331
aI1571
aI1145
aI745
aI1845
aI17881
aI17697
aI88139
asS'd'
I6144
sa(dp12288
S'a'
I3577
sS's'
I17
sS'm'
(lp12289
I1
aI3
aI7
aI15
aI19
aI37
aI119
aI35
aI35
aI463
aI1925
aI1665
aI673
aI12193
aI12137
aI62371
aI10957
asS'd'
I6145
sa(dp12290
S'a'
I3578
sS's'
I17
sS'm'
(lp12291
I1
aI3
aI3
aI3
aI19
aI21
aI113
aI29
aI459
aI467
aI623
aI2661
aI857
aI16265
aI27509
aI46555
aI18867
asS'd'
I6146
sa(dp12292
S'a'
I3594
sS's'
I17
sS'm'
(lp12293
I1
aI3
aI7
aI5
aI17
aI49
aI123
aI41
aI85
aI673
aI41
aI1871
aI7649
aI8687
aI28269
aI64423
aI93675
asS'd'
I6147
sa(dp12294
S'a'
I3601
sS's'
I17
sS'm'
(lp12295
I1
aI3
aI3
aI3
aI7
aI23
aI101
aI171
aI181
aI527
aI65
aI2387
aI6629
aI6089
aI17387
aI46551
aI36143
asS'd'
I6148
sa(dp12296
S'a'
I3607
sS's'
I17
sS'm'
(lp12297
I1
aI1
aI5
aI1
aI13
aI51
aI21
aI251
aI139
aI429
aI1993
aI3767
aI1089
aI5459
aI19407
aI41747
aI41033
asS'd'
I6149
sa(dp12298
S'a'
I3608
sS's'
I17
sS'm'
(lp12299
I1
aI1
aI1
aI11
aI15
aI9
aI81
aI91
aI73
aI969
aI1513
aI2067
aI7959
aI2605
aI26641
aI37631
aI124571
asS'd'
I6150
sa(dp12300
S'a'
I3620
sS's'
I17
sS'm'
(lp12301
I1
aI1
aI3
aI15
aI29
aI15
aI5
aI57
aI247
aI901
aI527
aI3325
aI5859
aI11299
aI9871
aI63947
aI125247
asS'd'
I6151
sa(dp12302
S'a'
I3629
sS's'
I17
sS'm'
(lp12303
I1
aI3
aI1
aI5
aI1
aI35
aI75
aI21
aI307
aI43
aI1111
aI3299
aI1647
aI3585
aI31045
aI18217
aI95169
asS'd'
I6152
sa(dp12304
S'a'
I3644
sS's'
I17
sS'm'
(lp12305
I1
aI3
aI1
aI7
aI23
aI35
aI11
aI103
aI3
aI461
aI1915
aI4019
aI453
aI13111
aI26941
aI43091
aI22917
asS'd'
I6153
sa(dp12306
S'a'
I3656
sS's'
I17
sS'm'
(lp12307
I1
aI1
aI5
aI5
aI1
aI61
aI121
aI167
aI475
aI5
aI1749
aI887
aI2237
aI5055
aI7077
aI29453
aI17691
asS'd'
I6154
sa(dp12308
S'a'
I3664
sS's'
I17
sS'm'
(lp12309
I1
aI3
aI3
aI15
aI15
aI15
aI9
aI15
aI171
aI787
aI1965
aI577
aI4507
aI7325
aI20901
aI8557
aI111909
asS'd'
I6155
sa(dp12310
S'a'
I3670
sS's'
I17
sS'm'
(lp12311
I1
aI3
aI5
aI1
aI27
aI15
aI123
aI141
aI63
aI55
aI599
aI4095
aI1245
aI13919
aI27485
aI49977
aI74551
asS'd'
I6156
sa(dp12312
S'a'
I3680
sS's'
I17
sS'm'
(lp12313
I1
aI3
aI5
aI9
aI21
aI61
aI79
aI119
aI7
aI573
aI1923
aI2775
aI3127
aI12689
aI12135
aI53429
aI130163
asS'd'
I6157
sa(dp12314
S'a'
I3685
sS's'
I17
sS'm'
(lp12315
I1
aI3
aI3
aI13
aI27
aI41
aI67
aI249
aI447
aI277
aI311
aI775
aI8187
aI10161
aI12953
aI22885
aI121247
asS'd'
I6158
sa(dp12316
S'a'
I3686
sS's'
I17
sS'm'
(lp12317
I1
aI3
aI5
aI9
aI21
aI55
aI115
aI65
aI45
aI395
aI481
aI2063
aI6493
aI4199
aI19219
aI27119
aI62255
asS'd'
I6159
sa(dp12318
S'a'
I3695
sS's'
I17
sS'm'
(lp12319
I1
aI1
aI3
aI13
aI7
aI41
aI3
aI127
aI383
aI923
aI1725
aI1033
aI7731
aI11971
aI3089
aI46459
aI98369
asS'd'
I6160
sa(dp12320
S'a'
I3698
sS's'
I17
sS'm'
(lp12321
I1
aI1
aI3
aI11
aI13
aI39
aI39
aI149
aI309
aI311
aI1491
aI807
aI2109
aI363
aI14637
aI65429
aI124731
asS'd'
I6161
sa(dp12322
S'a'
I3703
sS's'
I17
sS'm'
(lp12323
I1
aI1
aI7
aI13
aI13
aI35
aI67
aI81
aI493
aI859
aI1177
aI237
aI4605
aI15319
aI16669
aI16661
aI21385
asS'd'
I6162
sa(dp12324
S'a'
I3710
sS's'
I17
sS'm'
(lp12325
I1
aI1
aI3
aI7
aI7
aI39
aI57
aI103
aI239
aI753
aI221
aI1611
aI1557
aI13317
aI27453
aI10245
aI33839
asS'd'
I6163
sa(dp12326
S'a'
I3714
sS's'
I17
sS'm'
(lp12327
I1
aI1
aI5
aI13
aI27
aI53
aI97
aI41
aI123
aI253
aI535
aI1839
aI5827
aI7587
aI1261
aI20313
aI65961
asS'd'
I6164
sa(dp12328
S'a'
I3726
sS's'
I17
sS'm'
(lp12329
I1
aI1
aI7
aI1
aI11
aI47
aI93
aI135
aI223
aI591
aI1087
aI3329
aI3293
aI14207
aI6187
aI54789
aI23781
asS'd'
I6165
sa(dp12330
S'a'
I3731
sS's'
I17
sS'm'
(lp12331
I1
aI3
aI7
aI7
aI25
aI21
aI97
aI105
aI269
aI515
aI1805
aI3711
aI3295
aI7307
aI21065
aI65205
aI116969
asS'd'
I6166
sa(dp12332
S'a'
I3733
sS's'
I17
sS'm'
(lp12333
I1
aI3
aI1
aI11
aI25
aI37
aI21
aI89
aI109
aI581
aI1055
aI2393
aI1291
aI1115
aI25545
aI36383
aI93605
asS'd'
I6167
sa(dp12334
S'a'
I3737
sS's'
I17
sS'm'
(lp12335
I1
aI3
aI7
aI1
aI27
aI13
aI113
aI11
aI395
aI473
aI943
aI4045
aI5507
aI15051
aI25203
aI2971
aI31961
asS'd'
I6168
sa(dp12336
S'a'
I3756
sS's'
I17
sS'm'
(lp12337
I1
aI1
aI5
aI5
aI27
aI35
aI57
aI219
aI67
aI949
aI659
aI203
aI5235
aI6509
aI13731
aI61533
aI54963
asS'd'
I6169
sa(dp12338
S'a'
I3759
sS's'
I17
sS'm'
(lp12339
I1
aI3
aI1
aI1
aI15
aI39
aI85
aI13
aI347
aI99
aI25
aI3595
aI3081
aI13617
aI14373
aI58909
aI102181
asS'd'
I6170
sa(dp12340
S'a'
I3767
sS's'
I17
sS'm'
(lp12341
I1
aI1
aI7
aI13
aI3
aI25
aI97
aI91
aI287
aI389
aI665
aI2981
aI2301
aI12625
aI4495
aI57489
aI68677
asS'd'
I6171
sa(dp12342
S'a'
I3776
sS's'
I17
sS'm'
(lp12343
I1
aI1
aI5
aI1
aI15
aI57
aI77
aI55
aI299
aI713
aI1457
aI3699
aI2807
aI5549
aI467
aI47367
aI8163
asS'd'
I6172
sa(dp12344
S'a'
I3785
sS's'
I17
sS'm'
(lp12345
I1
aI1
aI7
aI3
aI23
aI45
aI91
aI251
aI501
aI193
aI1121
aI2359
aI4781
aI12797
aI13713
aI55171
aI927
asS'd'
I6173
sa(dp12346
S'a'
I3793
sS's'
I17
sS'm'
(lp12347
I1
aI3
aI3
aI7
aI7
aI31
aI87
aI163
aI249
aI163
aI937
aI1293
aI4827
aI10299
aI31935
aI58787
aI80589
asS'd'
I6174
sa(dp12348
S'a'
I3812
sS's'
I17
sS'm'
(lp12349
I1
aI3
aI1
aI9
aI7
aI1
aI73
aI65
aI475
aI791
aI1429
aI3319
aI7149
aI433
aI10373
aI44061
aI121195
asS'd'
I6175
sa(dp12350
S'a'
I3815
sS's'
I17
sS'm'
(lp12351
I1
aI1
aI5
aI9
aI9
aI61
aI27
aI249
aI435
aI437
aI1329
aI2163
aI5859
aI13663
aI623
aI55569
aI94283
asS'd'
I6176
sa(dp12352
S'a'
I3824
sS's'
I17
sS'm'
(lp12353
I1
aI3
aI7
aI11
aI1
aI29
aI117
aI195
aI399
aI999
aI1705
aI1325
aI6043
aI9823
aI27335
aI30377
aI16627
asS'd'
I6177
sa(dp12354
S'a'
I3844
sS's'
I17
sS'm'
(lp12355
I1
aI1
aI1
aI15
aI5
aI11
aI63
aI185
aI15
aI741
aI1061
aI2961
aI3455
aI5
aI26587
aI54081
aI18107
asS'd'
I6178
sa(dp12356
S'a'
I3859
sS's'
I17
sS'm'
(lp12357
I1
aI1
aI5
aI7
aI29
aI57
aI17
aI203
aI501
aI177
aI49
aI2773
aI8069
aI12513
aI14437
aI64489
aI58661
asS'd'
I6179
sa(dp12358
S'a'
I3866
sS's'
I17
sS'm'
(lp12359
I1
aI3
aI3
aI9
aI11
aI23
aI121
aI3
aI415
aI447
aI1773
aI135
aI5901
aI4951
aI2683
aI437
aI126251
asS'd'
I6180
sa(dp12360
S'a'
I3872
sS's'
I17
sS'm'
(lp12361
I1
aI3
aI3
aI1
aI7
aI23
aI17
aI23
aI115
aI591
aI1075
aI3133
aI49
aI15183
aI10615
aI37857
aI122609
asS'd'
I6181
sa(dp12362
S'a'
I3884
sS's'
I17
sS'm'
(lp12363
I1
aI1
aI3
aI3
aI13
aI49
aI63
aI37
aI275
aI763
aI1135
aI2913
aI1563
aI11037
aI6693
aI18799
aI32089
asS'd'
I6182
sa(dp12364
S'a'
I3889
sS's'
I17
sS'm'
(lp12365
I1
aI3
aI5
aI11
aI7
aI29
aI59
aI45
aI227
aI941
aI1947
aI2733
aI797
aI10485
aI7071
aI14741
aI11451
asS'd'
I6183
sa(dp12366
S'a'
I3899
sS's'
I17
sS'm'
(lp12367
I1
aI1
aI1
aI9
aI21
aI19
aI77
aI97
aI75
aI991
aI187
aI1003
aI5619
aI11013
aI3931
aI19907
aI79723
asS'd'
I6184
sa(dp12368
S'a'
I3902
sS's'
I17
sS'm'
(lp12369
I1
aI1
aI7
aI13
aI1
aI57
aI61
aI177
aI443
aI227
aI1347
aI2665
aI2011
aI12329
aI14137
aI37795
aI63331
asS'd'
I6185
sa(dp12370
S'a'
I3909
sS's'
I17
sS'm'
(lp12371
I1
aI3
aI3
aI9
aI31
aI59
aI87
aI93
aI485
aI635
aI901
aI1845
aI6153
aI10797
aI1289
aI8989
aI41717
asS'd'
I6186
sa(dp12372
S'a'
I3913
sS's'
I17
sS'm'
(lp12373
I1
aI1
aI1
aI1
aI3
aI7
aI85
aI17
aI67
aI309
aI1891
aI435
aI303
aI8011
aI32127
aI54309
aI21457
asS'd'
I6187
sa(dp12374
S'a'
I3933
sS's'
I17
sS'm'
(lp12375
I1
aI3
aI7
aI1
aI29
aI27
aI41
aI239
aI293
aI717
aI1331
aI917
aI6145
aI7131
aI28199
aI35093
aI103683
asS'd'
I6188
sa(dp12376
S'a'
I3938
sS's'
I17
sS'm'
(lp12377
I1
aI3
aI7
aI3
aI21
aI63
aI65
aI233
aI257
aI789
aI1095
aI505
aI4557
aI16259
aI7397
aI24815
aI89529
asS'd'
I6189
sa(dp12378
S'a'
I3949
sS's'
I17
sS'm'
(lp12379
I1
aI3
aI3
aI11
aI29
aI41
aI55
aI17
aI335
aI715
aI779
aI2121
aI6393
aI8887
aI32753
aI45647
aI82665
asS'd'
I6190
sa(dp12380
S'a'
I3952
sS's'
I17
sS'm'
(lp12381
I1
aI1
aI1
aI11
aI27
aI47
aI71
aI13
aI141
aI283
aI967
aI3359
aI4309
aI6661
aI20481
aI23175
aI50835
asS'd'
I6191
sa(dp12382
S'a'
I3980
sS's'
I17
sS'm'
(lp12383
I1
aI3
aI3
aI7
aI3
aI25
aI19
aI241
aI409
aI573
aI1565
aI3355
aI1307
aI12205
aI18017
aI8271
aI117007
asS'd'
I6192
sa(dp12384
S'a'
I3991
sS's'
I17
sS'm'
(lp12385
I1
aI3
aI3
aI9
aI21
aI39
aI21
aI253
aI439
aI963
aI341
aI3637
aI2275
aI1845
aI11015
aI481
aI83369
asS'd'
I6193
sa(dp12386
S'a'
I3992
sS's'
I17
sS'm'
(lp12387
I1
aI3
aI7
aI9
aI31
aI29
aI29
aI163
aI111
aI983
aI571
aI713
aI2621
aI11569
aI13341
aI28341
aI130381
asS'd'
I6194
sa(dp12388
S'a'
I4002
sS's'
I17
sS'm'
(lp12389
I1
aI3
aI7
aI7
aI11
aI35
aI89
aI49
aI81
aI115
aI113
aI1857
aI3527
aI14819
aI6909
aI14659
aI23557
asS'd'
I6195
sa(dp12390
S'a'
I4008
sS's'
I17
sS'm'
(lp12391
I1
aI3
aI3
aI15
aI29
aI41
aI85
aI241
aI317
aI737
aI213
aI1667
aI5789
aI16321
aI13991
aI36165
aI124151
asS'd'
I6196
sa(dp12392
S'a'
I4011
sS's'
I17
sS'm'
(lp12393
I1
aI3
aI1
aI3
aI31
aI1
aI75
aI99
aI495
aI241
aI1499
aI1535
aI2033
aI2135
aI6699
aI58893
aI37031
asS'd'
I6197
sa(dp12394
S'a'
I4016
sS's'
I17
sS'm'
(lp12395
I1
aI1
aI7
aI9
aI25
aI15
aI101
aI23
aI477
aI563
aI1691
aI2655
aI2321
aI2323
aI4255
aI22055
aI99661
asS'd'
I6198
sa(dp12396
S'a'
I4034
sS's'
I17
sS'm'
(lp12397
I1
aI3
aI7
aI5
aI7
aI7
aI49
aI221
aI51
aI83
aI279
aI2205
aI2939
aI2119
aI14073
aI32839
aI108075
asS'd'
I6199
sa(dp12398
S'a'
I4036
sS's'
I17
sS'm'
(lp12399
I1
aI3
aI5
aI11
aI17
aI39
aI3
aI127
aI87
aI501
aI799
aI401
aI4439
aI9895
aI13017
aI64975
aI67177
asS'd'
I6200
sa(dp12400
S'a'
I4063
sS's'
I17
sS'm'
(lp12401
I1
aI3
aI3
aI9
aI17
aI41
aI59
aI95
aI283
aI309
aI83
aI1293
aI6385
aI5783
aI30115
aI33997
aI12531
asS'd'
I6201
sa(dp12402
S'a'
I4067
sS's'
I17
sS'm'
(lp12403
I1
aI3
aI5
aI3
aI7
aI31
aI69
aI171
aI225
aI409
aI1237
aI3343
aI835
aI8039
aI16723
aI37203
aI129047
asS'd'
I6202
sa(dp12404
S'a'
I4073
sS's'
I17
sS'm'
(lp12405
I1
aI3
aI3
aI15
aI17
aI23
aI107
aI1
aI105
aI135
aI1245
aI993
aI4101
aI7325
aI7425
aI17379
aI98121
asS'd'
I6203
sa(dp12406
S'a'
I4082
sS's'
I17
sS'm'
(lp12407
I1
aI1
aI7
aI9
aI27
aI5
aI67
aI111
aI75
aI531
aI243
aI2239
aI2527
aI4513
aI27059
aI40533
aI88169
asS'd'
I6204
sa(dp12408
S'a'
I4091
sS's'
I17
sS'm'
(lp12409
I1
aI3
aI5
aI7
aI21
aI63
aI57
aI15
aI75
aI679
aI1729
aI1845
aI6259
aI8531
aI18691
aI49321
aI101599
asS'd'
I6205
sa(dp12410
S'a'
I4093
sS's'
I17
sS'm'
(lp12411
I1
aI1
aI5
aI9
aI3
aI35
aI7
aI201
aI351
aI885
aI669
aI2339
aI5009
aI279
aI26469
aI54597
aI67933
asS'd'
I6206
sa(dp12412
S'a'
I4101
sS's'
I17
sS'm'
(lp12413
I1
aI3
aI5
aI13
aI27
aI5
aI85
aI161
aI141
aI733
aI1017
aI2021
aI6951
aI15595
aI21817
aI17243
aI88607
asS'd'
I6207
sa(dp12414
S'a'
I4113
sS's'
I17
sS'm'
(lp12415
I1
aI3
aI5
aI1
aI11
aI31
aI117
aI97
aI175
aI629
aI995
aI1207
aI2941
aI5825
aI5319
aI48191
aI9505
asS'd'
I6208
sa(dp12416
S'a'
I4120
sS's'
I17
sS'm'
(lp12417
I1
aI3
aI3
aI7
aI25
aI39
aI45
aI79
aI21
aI607
aI1593
aI1749
aI7951
aI10425
aI17491
aI16617
aI56903
asS'd'
I6209
sa(dp12418
S'a'
I4125
sS's'
I17
sS'm'
(lp12419
I1
aI1
aI1
aI5
aI15
aI41
aI107
aI115
aI79
aI693
aI919
aI3513
aI6793
aI6541
aI5545
aI58583
aI27963
asS'd'
I6210
sa(dp12420
S'a'
I4126
sS's'
I17
sS'm'
(lp12421
I1
aI3
aI7
aI11
aI21
aI19
aI123
aI1
aI441
aI531
aI359
aI2117
aI2465
aI11389
aI13489
aI32755
aI4577
asS'd'
I6211
sa(dp12422
S'a'
I4139
sS's'
I17
sS'm'
(lp12423
I1
aI1
aI5
aI13
aI7
aI7
aI7
aI127
aI201
aI377
aI1423
aI269
aI2611
aI3339
aI19153
aI25659
aI33069
asS'd'
I6212
sa(dp12424
S'a'
I4142
sS's'
I17
sS'm'
(lp12425
I1
aI3
aI7
aI1
aI13
aI35
aI45
aI5
aI313
aI739
aI1779
aI2983
aI1815
aI8817
aI14239
aI3921
aI57975
asS'd'
I6213
sa(dp12426
S'a'
I4144
sS's'
I17
sS'm'
(lp12427
I1
aI3
aI1
aI11
aI9
aI39
aI33
aI111
aI39
aI255
aI159
aI2345
aI2193
aI11475
aI12841
aI47579
aI90309
asS'd'
I6214
sa(dp12428
S'a'
I4147
sS's'
I17
sS'm'
(lp12429
I1
aI1
aI1
aI3
aI27
aI49
aI85
aI157
aI243
aI247
aI1473
aI323
aI4631
aI1787
aI15193
aI5533
aI104999
asS'd'
I6215
sa(dp12430
S'a'
I4153
sS's'
I17
sS'm'
(lp12431
I1
aI1
aI7
aI9
aI11
aI29
aI23
aI219
aI57
aI339
aI1797
aI409
aI6025
aI10569
aI27409
aI15147
aI130281
asS'd'
I6216
sa(dp12432
S'a'
I4154
sS's'
I17
sS'm'
(lp12433
I1
aI1
aI7
aI1
aI31
aI31
aI113
aI229
aI63
aI877
aI319
aI2655
aI3335
aI7743
aI19593
aI10089
aI28215
asS'd'
I6217
sa(dp12434
S'a'
I4164
sS's'
I17
sS'm'
(lp12435
I1
aI1
aI3
aI11
aI23
aI3
aI71
aI235
aI329
aI751
aI159
aI2579
aI5363
aI12681
aI20233
aI53855
aI16407
asS'd'
I6218
sa(dp12436
S'a'
I4174
sS's'
I17
sS'm'
(lp12437
I1
aI1
aI5
aI1
aI7
aI61
aI21
aI235
aI379
aI849
aI61
aI2969
aI6399
aI2655
aI21635
aI16955
aI58675
asS'd'
I6219
sa(dp12438
S'a'
I4182
sS's'
I17
sS'm'
(lp12439
I1
aI3
aI7
aI7
aI29
aI15
aI5
aI11
aI143
aI699
aI1875
aI2115
aI6633
aI6195
aI5829
aI53633
aI111221
asS'd'
I6220
sa(dp12440
S'a'
I4185
sS's'
I17
sS'm'
(lp12441
I1
aI3
aI7
aI11
aI19
aI41
aI17
aI219
aI483
aI829
aI1233
aI3183
aI6283
aI2363
aI25245
aI63075
aI82733
asS'd'
I6221
sa(dp12442
S'a'
I4188
sS's'
I17
sS'm'
(lp12443
I1
aI3
aI7
aI13
aI21
aI17
aI1
aI207
aI443
aI575
aI521
aI2585
aI6875
aI14871
aI14739
aI10211
aI127435
asS'd'
I6222
sa(dp12444
S'a'
I4191
sS's'
I17
sS'm'
(lp12445
I1
aI3
aI7
aI7
aI15
aI39
aI99
aI197
aI219
aI259
aI1723
aI3737
aI6625
aI849
aI887
aI41293
aI53825
asS'd'
I6223
sa(dp12446
S'a'
I4195
sS's'
I17
sS'm'
(lp12447
I1
aI3
aI3
aI3
aI5
aI3
aI75
aI155
aI189
aI935
aI85
aI2273
aI1375
aI4217
aI10709
aI58047
aI81689
asS'd'
I6224
sa(dp12448
S'a'
I4219
sS's'
I17
sS'm'
(lp12449
I1
aI3
aI5
aI5
aI27
aI27
aI107
aI229
aI179
aI887
aI91
aI421
aI7313
aI6495
aI451
aI43859
aI40033
asS'd'
I6225
sa(dp12450
S'a'
I4225
sS's'
I17
sS'm'
(lp12451
I1
aI3
aI5
aI11
aI25
aI49
aI121
aI73
aI169
aI311
aI1387
aI1037
aI6519
aI9317
aI26975
aI50627
aI46805
asS'd'
I6226
sa(dp12452
S'a'
I4228
sS's'
I17
sS'm'
(lp12453
I1
aI1
aI5
aI11
aI17
aI21
aI19
aI125
aI387
aI697
aI1017
aI1759
aI7295
aI9869
aI28241
aI9367
aI119255
asS'd'
I6227
sa(dp12454
S'a'
I4232
sS's'
I17
sS'm'
(lp12455
I1
aI1
aI7
aI5
aI29
aI27
aI87
aI187
aI95
aI625
aI933
aI1751
aI5253
aI313
aI30841
aI16349
aI67347
asS'd'
I6228
sa(dp12456
S'a'
I4246
sS's'
I17
sS'm'
(lp12457
I1
aI1
aI3
aI3
aI15
aI51
aI23
aI101
aI183
aI267
aI243
aI711
aI983
aI12461
aI17801
aI1429
aI47273
asS'd'
I6229
sa(dp12458
S'a'
I4255
sS's'
I17
sS'm'
(lp12459
I1
aI1
aI1
aI3
aI17
aI3
aI73
aI67
aI49
aI449
aI879
aI2559
aI401
aI11983
aI13697
aI12023
aI78855
asS'd'
I6230
sa(dp12460
S'a'
I4274
sS's'
I17
sS'm'
(lp12461
I1
aI3
aI7
aI15
aI25
aI25
aI43
aI81
aI141
aI161
aI595
aI621
aI1165
aI10869
aI22875
aI6741
aI90017
asS'd'
I6231
sa(dp12462
S'a'
I4283
sS's'
I17
sS'm'
(lp12463
I1
aI3
aI5
aI11
aI13
aI57
aI53
aI219
aI145
aI937
aI769
aI1961
aI4725
aI3335
aI12623
aI8335
aI46305
asS'd'
I6232
sa(dp12464
S'a'
I4286
sS's'
I17
sS'm'
(lp12465
I1
aI1
aI3
aI5
aI7
aI39
aI19
aI101
aI313
aI583
aI483
aI2515
aI125
aI5211
aI2559
aI11937
aI126717
asS'd'
I6233
sa(dp12466
S'a'
I4306
sS's'
I17
sS'm'
(lp12467
I1
aI3
aI1
aI7
aI7
aI1
aI117
aI49
aI231
aI133
aI381
aI697
aI927
aI8263
aI26529
aI64881
aI25059
asS'd'
I6234
sa(dp12468
S'a'
I4311
sS's'
I17
sS'm'
(lp12469
I1
aI1
aI1
aI15
aI11
aI25
aI77
aI149
aI233
aI215
aI1239
aI3045
aI99
aI11183
aI30279
aI32271
aI100943
asS'd'
I6235
sa(dp12470
S'a'
I4317
sS's'
I17
sS'm'
(lp12471
I1
aI1
aI5
aI7
aI31
aI25
aI1
aI51
aI221
aI607
aI1733
aI2145
aI6765
aI7011
aI16927
aI29257
aI2445
asS'd'
I6236
sa(dp12472
S'a'
I4321
sS's'
I17
sS'm'
(lp12473
I1
aI3
aI5
aI1
aI19
aI23
aI123
aI93
aI381
aI295
aI765
aI2335
aI8025
aI14003
aI4801
aI54243
aI57297
asS'd'
I6237
sa(dp12474
S'a'
I4324
sS's'
I17
sS'm'
(lp12475
I1
aI1
aI7
aI9
aI9
aI31
aI63
aI191
aI495
aI527
aI251
aI2119
aI1663
aI209
aI7445
aI1441
aI4075
asS'd'
I6238
sa(dp12476
S'a'
I4331
sS's'
I17
sS'm'
(lp12477
I1
aI3
aI5
aI5
aI13
aI17
aI97
aI79
aI369
aI55
aI677
aI2031
aI7315
aI4769
aI31659
aI21975
aI22061
asS'd'
I6239
sa(dp12478
S'a'
I4333
sS's'
I17
sS'm'
(lp12479
I1
aI3
aI3
aI7
aI3
aI63
aI121
aI243
aI39
aI917
aI1917
aI297
aI7241
aI1565
aI31675
aI14443
aI67239
asS'd'
I6240
sa(dp12480
S'a'
I4359
sS's'
I17
sS'm'
(lp12481
I1
aI3
aI7
aI1
aI13
aI25
aI51
aI65
aI145
aI475
aI1853
aI4023
aI5121
aI14411
aI15993
aI42165
aI13615
asS'd'
I6241
sa(dp12482
S'a'
I4360
sS's'
I17
sS'm'
(lp12483
I1
aI3
aI3
aI1
aI3
aI51
aI75
aI29
aI169
aI311
aI1309
aI2929
aI7669
aI1507
aI14605
aI32667
aI103861
asS'd'
I6242
sa(dp12484
S'a'
I4368
sS's'
I17
sS'm'
(lp12485
I1
aI3
aI7
aI1
aI23
aI37
aI89
aI211
aI137
aI495
aI1469
aI3425
aI1167
aI12429
aI27301
aI46857
aI83007
asS'd'
I6243
sa(dp12486
S'a'
I4373
sS's'
I17
sS'm'
(lp12487
I1
aI3
aI7
aI7
aI27
aI37
aI33
aI129
aI73
aI23
aI761
aI119
aI6217
aI4749
aI20835
aI47477
aI33665
asS'd'
I6244
sa(dp12488
S'a'
I4389
sS's'
I17
sS'm'
(lp12489
I1
aI1
aI3
aI5
aI29
aI35
aI79
aI21
aI183
aI933
aI43
aI3149
aI5273
aI12159
aI20695
aI5387
aI23569
asS'd'
I6245
sa(dp12490
S'a'
I4394
sS's'
I17
sS'm'
(lp12491
I1
aI1
aI5
aI5
aI3
aI11
aI57
aI205
aI349
aI657
aI1509
aI3693
aI5495
aI11865
aI13861
aI62215
aI94141
asS'd'
I6246
sa(dp12492
S'a'
I4413
sS's'
I17
sS'm'
(lp12493
I1
aI3
aI1
aI7
aI17
aI43
aI117
aI119
aI75
aI849
aI1247
aI643
aI2691
aI2289
aI9759
aI18683
aI68649
asS'd'
I6247
sa(dp12494
S'a'
I4422
sS's'
I17
sS'm'
(lp12495
I1
aI1
aI1
aI15
aI5
aI55
aI89
aI177
aI427
aI701
aI735
aI2993
aI5293
aI15395
aI567
aI5501
aI102393
asS'd'
I6248
sa(dp12496
S'a'
I4431
sS's'
I17
sS'm'
(lp12497
I1
aI3
aI3
aI15
aI5
aI37
aI73
aI111
aI9
aI141
aI407
aI1579
aI6691
aI11843
aI6377
aI64181
aI97347
asS'd'
I6249
sa(dp12498
S'a'
I4436
sS's'
I17
sS'm'
(lp12499
I1
aI1
aI5
aI1
aI9
aI17
aI71
aI127
aI285
aI929
aI1243
aI2605
aI359
aI14589
aI32603
aI39879
aI115901
asS'd'
I6250
sa(dp12500
S'a'
I4440
sS's'
I17
sS'm'
(lp12501
I1
aI3
aI7
aI15
aI3
aI27
aI91
aI121
aI47
aI631
aI1589
aI385
aI5997
aI14077
aI21285
aI33895
aI36985
asS'd'
I6251
sa(dp12502
S'a'
I4445
sS's'
I17
sS'm'
(lp12503
I1
aI3
aI3
aI9
aI1
aI47
aI89
aI79
aI213
aI27
aI547
aI1703
aI4035
aI13205
aI4341
aI21895
aI34247
asS'd'
I6252
sa(dp12504
S'a'
I4452
sS's'
I17
sS'm'
(lp12505
I1
aI3
aI5
aI7
aI9
aI9
aI47
aI89
aI231
aI857
aI297
aI2949
aI2715
aI1275
aI14427
aI20227
aI21569
asS'd'
I6253
sa(dp12506
S'a'
I4462
sS's'
I17
sS'm'
(lp12507
I1
aI3
aI1
aI3
aI15
aI57
aI61
aI183
aI377
aI477
aI1135
aI1729
aI2863
aI8607
aI29241
aI34983
aI84443
asS'd'
I6254
sa(dp12508
S'a'
I4469
sS's'
I17
sS'm'
(lp12509
I1
aI1
aI7
aI7
aI5
aI53
aI91
aI149
aI71
aI41
aI1025
aI3945
aI3989
aI15853
aI20903
aI26943
aI99841
asS'd'
I6255
sa(dp12510
S'a'
I4470
sS's'
I17
sS'm'
(lp12511
I1
aI3
aI3
aI3
aI29
aI21
aI59
aI217
aI483
aI257
aI331
aI657
aI2935
aI945
aI9821
aI42501
aI98087
asS'd'
I6256
sa(dp12512
S'a'
I4473
sS's'
I17
sS'm'
(lp12513
I1
aI3
aI5
aI3
aI17
aI39
aI123
aI103
aI109
aI957
aI853
aI3821
aI555
aI10869
aI27673
aI38315
aI83105
asS'd'
I6257
sa(dp12514
S'a'
I4479
sS's'
I17
sS'm'
(lp12515
I1
aI3
aI1
aI3
aI27
aI7
aI97
aI57
aI429
aI53
aI1791
aI1405
aI4113
aI8435
aI12845
aI21567
aI91559
asS'd'
I6258
sa(dp12516
S'a'
I4480
sS's'
I17
sS'm'
(lp12517
I1
aI3
aI3
aI1
aI17
aI61
aI125
aI77
aI225
aI395
aI945
aI3213
aI1363
aI15947
aI27049
aI4389
aI64037
asS'd'
I6259
sa(dp12518
S'a'
I4483
sS's'
I17
sS'm'
(lp12519
I1
aI1
aI1
aI3
aI15
aI51
aI15
aI189
aI449
aI989
aI939
aI985
aI6929
aI13779
aI25011
aI22277
aI72543
asS'd'
I6260
sa(dp12520
S'a'
I4489
sS's'
I17
sS'm'
(lp12521
I1
aI3
aI3
aI1
aI25
aI53
aI5
aI219
aI195
aI703
aI163
aI1405
aI821
aI6797
aI14329
aI1675
aI96653
asS'd'
I6261
sa(dp12522
S'a'
I4503
sS's'
I17
sS'm'
(lp12523
I1
aI1
aI7
aI13
aI7
aI1
aI45
aI135
aI369
aI125
aI711
aI2509
aI131
aI13663
aI29769
aI19497
aI116779
asS'd'
I6262
sa(dp12524
S'a'
I4519
sS's'
I17
sS'm'
(lp12525
I1
aI1
aI7
aI15
aI23
aI25
aI7
aI225
aI435
aI835
aI1981
aI2537
aI5727
aI15961
aI30089
aI58905
aI100339
asS'd'
I6263
sa(dp12526
S'a'
I4520
sS's'
I17
sS'm'
(lp12527
I1
aI3
aI7
aI3
aI19
aI9
aI79
aI63
aI371
aI419
aI1357
aI3649
aI7987
aI14541
aI6631
aI50555
aI84217
asS'd'
I6264
sa(dp12528
S'a'
I4525
sS's'
I17
sS'm'
(lp12529
I1
aI3
aI3
aI9
aI7
aI61
aI11
aI157
aI99
aI95
aI945
aI2803
aI1703
aI117
aI12891
aI21817
aI84259
asS'd'
I6265
sa(dp12530
S'a'
I4526
sS's'
I17
sS'm'
(lp12531
I1
aI3
aI7
aI7
aI25
aI37
aI111
aI99
aI65
aI599
aI1313
aI2557
aI5489
aI3625
aI7429
aI19309
aI78111
asS'd'
I6266
sa(dp12532
S'a'
I4533
sS's'
I17
sS'm'
(lp12533
I1
aI3
aI1
aI1
aI19
aI15
aI85
aI253
aI347
aI315
aI1349
aI983
aI2507
aI4155
aI15311
aI43535
aI101409
asS'd'
I6267
sa(dp12534
S'a'
I4552
sS's'
I17
sS'm'
(lp12535
I1
aI3
aI3
aI3
aI1
aI55
aI3
aI57
aI375
aI107
aI177
aI1673
aI6871
aI7137
aI10297
aI65363
aI42293
asS'd'
I6268
sa(dp12536
S'a'
I4581
sS's'
I17
sS'm'
(lp12537
I1
aI1
aI1
aI3
aI9
aI5
aI83
aI45
aI139
aI893
aI63
aI2859
aI6333
aI15591
aI18491
aI26387
aI25573
asS'd'
I6269
sa(dp12538
S'a'
I4585
sS's'
I17
sS'm'
(lp12539
I1
aI1
aI7
aI15
aI1
aI39
aI113
aI127
aI503
aI617
aI1367
aI1855
aI185
aI4233
aI5787
aI8265
aI42097
asS'd'
I6270
sa(dp12540
S'a'
I4591
sS's'
I17
sS'm'
(lp12541
I1
aI1
aI3
aI11
aI11
aI41
aI119
aI165
aI331
aI625
aI81
aI2495
aI7247
aI9139
aI15269
aI31447
aI128425
asS'd'
I6271
sa(dp12542
S'a'
I4594
sS's'
I17
sS'm'
(lp12543
I1
aI1
aI5
aI5
aI17
aI35
aI39
aI1
aI91
aI563
aI1841
aI2975
aI1233
aI3837
aI22145
aI36719
aI104503
asS'd'
I6272
sa(dp12544
S'a'
I4596
sS's'
I17
sS'm'
(lp12545
I1
aI1
aI7
aI3
aI23
aI35
aI77
aI69
aI271
aI487
aI921
aI2597
aI8011
aI13037
aI6001
aI20519
aI32673
asS'd'
I6273
sa(dp12546
S'a'
I4599
sS's'
I17
sS'm'
(lp12547
I1
aI1
aI1
aI1
aI29
aI17
aI11
aI145
aI473
aI877
aI813
aI727
aI6805
aI3563
aI13371
aI22169
aI17239
asS'd'
I6274
sa(dp12548
S'a'
I4612
sS's'
I17
sS'm'
(lp12549
I1
aI1
aI1
aI13
aI17
aI13
aI1
aI125
aI313
aI423
aI1079
aI2401
aI2325
aI2219
aI24071
aI25613
aI34163
asS'd'
I6275
sa(dp12550
S'a'
I4621
sS's'
I17
sS'm'
(lp12551
I1
aI1
aI5
aI7
aI29
aI33
aI53
aI215
aI11
aI555
aI555
aI1965
aI3643
aI5433
aI12923
aI59655
aI25339
asS'd'
I6276
sa(dp12552
S'a'
I4630
sS's'
I17
sS'm'
(lp12553
I1
aI3
aI3
aI3
aI23
aI37
aI119
aI117
aI459
aI359
aI1849
aI1019
aI433
aI15391
aI5625
aI52649
aI81313
asS'd'
I6277
sa(dp12554
S'a'
I4636
sS's'
I17
sS'm'
(lp12555
I1
aI3
aI3
aI1
aI21
aI31
aI121
aI161
aI113
aI667
aI863
aI105
aI3805
aI14459
aI28235
aI24543
aI89755
asS'd'
I6278
sa(dp12556
S'a'
I4640
sS's'
I17
sS'm'
(lp12557
I1
aI1
aI5
aI15
aI17
aI37
aI15
aI111
aI511
aI477
aI611
aI955
aI2591
aI16137
aI14179
aI30995
aI129575
asS'd'
I6279
sa(dp12558
S'a'
I4649
sS's'
I17
sS'm'
(lp12559
I1
aI3
aI3
aI3
aI21
aI49
aI25
aI37
aI287
aI263
aI851
aI1015
aI8133
aI9429
aI10959
aI64483
aI82533
asS'd'
I6280
sa(dp12560
S'a'
I4650
sS's'
I17
sS'm'
(lp12561
I1
aI1
aI5
aI1
aI25
aI19
aI49
aI159
aI155
aI443
aI975
aI1413
aI321
aI7871
aI22935
aI57303
aI124027
asS'd'
I6281
sa(dp12562
S'a'
I4660
sS's'
I17
sS'm'
(lp12563
I1
aI3
aI1
aI1
aI19
aI45
aI47
aI89
aI409
aI509
aI1249
aI2445
aI2053
aI3781
aI7517
aI61869
aI125137
asS'd'
I6282
sa(dp12564
S'a'
I4677
sS's'
I17
sS'm'
(lp12565
I1
aI1
aI5
aI13
aI27
aI57
aI45
aI43
aI361
aI329
aI1321
aI771
aI4665
aI12245
aI18993
aI15121
aI127485
asS'd'
I6283
sa(dp12566
S'a'
I4687
sS's'
I17
sS'm'
(lp12567
I1
aI3
aI3
aI7
aI3
aI41
aI127
aI75
aI485
aI821
aI497
aI2649
aI6423
aI12419
aI31421
aI9441
aI63645
asS'd'
I6284
sa(dp12568
S'a'
I4696
sS's'
I17
sS'm'
(lp12569
I1
aI1
aI3
aI5
aI19
aI61
aI91
aI35
aI311
aI287
aI449
aI3955
aI5805
aI5631
aI25613
aI55409
aI104545
asS'd'
I6285
sa(dp12570
S'a'
I4701
sS's'
I17
sS'm'
(lp12571
I1
aI3
aI7
aI11
aI27
aI19
aI27
aI53
aI19
aI35
aI1687
aI3923
aI3379
aI10435
aI15053
aI12343
aI89077
asS'd'
I6286
sa(dp12572
S'a'
I4705
sS's'
I17
sS'm'
(lp12573
I1
aI3
aI5
aI13
aI31
aI41
aI15
aI239
aI349
aI533
aI1771
aI737
aI6503
aI14355
aI18781
aI27805
aI79049
asS'd'
I6287
sa(dp12574
S'a'
I4706
sS's'
I17
sS'm'
(lp12575
I1
aI3
aI1
aI3
aI13
aI11
aI69
aI227
aI169
aI873
aI533
aI2217
aI1047
aI12415
aI12271
aI22447
aI14163
asS'd'
I6288
sa(dp12576
S'a'
I4711
sS's'
I17
sS'm'
(lp12577
I1
aI1
aI3
aI9
aI7
aI31
aI23
aI155
aI133
aI305
aI1569
aI521
aI201
aI10339
aI16999
aI29163
aI32817
asS'd'
I6289
sa(dp12578
S'a'
I4720
sS's'
I17
sS'm'
(lp12579
I1
aI1
aI1
aI5
aI31
aI57
aI43
aI223
aI121
aI803
aI357
aI1855
aI4321
aI10245
aI25725
aI2543
aI47395
asS'd'
I6290
sa(dp12580
S'a'
I4723
sS's'
I17
sS'm'
(lp12581
I1
aI3
aI5
aI9
aI3
aI5
aI47
aI189
aI217
aI899
aI1455
aI691
aI1277
aI7861
aI3627
aI14895
aI41109
asS'd'
I6291
sa(dp12582
S'a'
I4732
sS's'
I17
sS'm'
(lp12583
I1
aI3
aI7
aI3
aI29
aI9
aI37
aI63
aI453
aI709
aI921
aI771
aI8069
aI239
aI22639
aI59937
aI10635
asS'd'
I6292
sa(dp12584
S'a'
I4736
sS's'
I17
sS'm'
(lp12585
I1
aI3
aI7
aI1
aI11
aI51
aI79
aI131
aI225
aI757
aI549
aI1605
aI3921
aI1849
aI16307
aI29809
aI120597
asS'd'
I6293
sa(dp12586
S'a'
I4742
sS's'
I17
sS'm'
(lp12587
I1
aI3
aI7
aI7
aI1
aI45
aI33
aI185
aI23
aI881
aI1941
aI4093
aI4741
aI11633
aI2059
aI32007
aI11103
asS'd'
I6294
sa(dp12588
S'a'
I4748
sS's'
I17
sS'm'
(lp12589
I1
aI3
aI5
aI11
aI17
aI21
aI43
aI205
aI363
aI559
aI697
aI4057
aI631
aI6697
aI883
aI61705
aI102791
asS'd'
I6295
sa(dp12590
S'a'
I4754
sS's'
I17
sS'm'
(lp12591
I1
aI1
aI7
aI9
aI29
aI35
aI109
aI85
aI373
aI321
aI415
aI2969
aI6163
aI6999
aI9999
aI36435
aI125267
asS'd'
I6296
sa(dp12592
S'a'
I4759
sS's'
I17
sS'm'
(lp12593
I1
aI1
aI7
aI11
aI25
aI9
aI113
aI91
aI337
aI889
aI947
aI2093
aI5289
aI1367
aI13297
aI36155
aI21825
asS'd'
I6297
sa(dp12594
S'a'
I4769
sS's'
I17
sS'm'
(lp12595
I1
aI1
aI3
aI9
aI17
aI25
aI35
aI79
aI275
aI687
aI335
aI1181
aI7327
aI3729
aI1561
aI27441
aI114355
asS'd'
I6298
sa(dp12596
S'a'
I4787
sS's'
I17
sS'm'
(lp12597
I1
aI3
aI3
aI11
aI25
aI41
aI27
aI89
aI115
aI361
aI871
aI1497
aI5735
aI6365
aI1737
aI14277
aI63847
asS'd'
I6299
sa(dp12598
S'a'
I4807
sS's'
I17
sS'm'
(lp12599
I1
aI3
aI7
aI7
aI1
aI63
aI31
aI73
aI289
aI67
aI277
aI1821
aI4883
aI10795
aI11755
aI15471
aI105871
asS'd'
I6300
sa(dp12600
S'a'
I4814
sS's'
I17
sS'm'
(lp12601
I1
aI3
aI7
aI9
aI23
aI17
aI37
aI179
aI409
aI957
aI373
aI2393
aI2363
aI6735
aI28737
aI41927
aI115735
asS'd'
I6301
sa(dp12602
S'a'
I4837
sS's'
I17
sS'm'
(lp12603
I1
aI1
aI3
aI9
aI15
aI43
aI111
aI61
aI455
aI181
aI1643
aI3063
aI4311
aI13705
aI29993
aI21731
aI25243
asS'd'
I6302
sa(dp12604
S'a'
I4867
sS's'
I17
sS'm'
(lp12605
I1
aI1
aI1
aI15
aI13
aI13
aI69
aI187
aI91
aI395
aI209
aI3477
aI4649
aI7727
aI30557
aI14719
aI1953
asS'd'
I6303
sa(dp12606
S'a'
I4873
sS's'
I17
sS'm'
(lp12607
I1
aI1
aI1
aI15
aI9
aI39
aI119
aI193
aI459
aI135
aI567
aI25
aI4583
aI8401
aI22161
aI14771
aI74165
asS'd'
I6304
sa(dp12608
S'a'
I4879
sS's'
I17
sS'm'
(lp12609
I1
aI1
aI3
aI7
aI5
aI39
aI77
aI149
aI293
aI585
aI1245
aI3615
aI357
aI11613
aI13865
aI40227
aI41023
asS'd'
I6305
sa(dp12610
S'a'
I4884
sS's'
I17
sS'm'
(lp12611
I1
aI1
aI7
aI9
aI9
aI37
aI5
aI177
aI121
aI181
aI771
aI733
aI7683
aI4855
aI13629
aI8349
aI46137
asS'd'
I6306
sa(dp12612
S'a'
I4898
sS's'
I17
sS'm'
(lp12613
I1
aI1
aI3
aI13
aI3
aI37
aI73
aI69
aI281
aI109
aI563
aI1427
aI5127
aI8957
aI16749
aI41489
aI49531
asS'd'
I6307
sa(dp12614
S'a'
I4907
sS's'
I17
sS'm'
(lp12615
I1
aI1
aI7
aI11
aI29
aI63
aI79
aI127
aI95
aI809
aI1175
aI1567
aI6353
aI7505
aI26551
aI5073
aI53733
asS'd'
I6308
sa(dp12616
S'a'
I4910
sS's'
I17
sS'm'
(lp12617
I1
aI1
aI1
aI5
aI25
aI41
aI59
aI103
aI59
aI365
aI1111
aI3909
aI3749
aI14889
aI3639
aI10435
aI45407
asS'd'
I6309
sa(dp12618
S'a'
I4918
sS's'
I17
sS'm'
(lp12619
I1
aI1
aI1
aI5
aI3
aI61
aI93
aI199
aI97
aI779
aI67
aI241
aI6197
aI6785
aI16869
aI7573
aI46745
asS'd'
I6310
sa(dp12620
S'a'
I4924
sS's'
I17
sS'm'
(lp12621
I1
aI1
aI5
aI9
aI27
aI29
aI21
aI69
aI165
aI661
aI1245
aI1265
aI2979
aI9685
aI17781
aI23329
aI48029
asS'd'
I6311
sa(dp12622
S'a'
I4953
sS's'
I17
sS'm'
(lp12623
I1
aI1
aI1
aI7
aI7
aI23
aI39
aI197
aI169
aI561
aI499
aI2197
aI4371
aI157
aI6837
aI44635
aI94861
asS'd'
I6312
sa(dp12624
S'a'
I4956
sS's'
I17
sS'm'
(lp12625
I1
aI1
aI5
aI13
aI7
aI5
aI9
aI207
aI321
aI243
aI899
aI2967
aI3553
aI15413
aI8961
aI55039
aI6459
asS'd'
I6313
sa(dp12626
S'a'
I4965
sS's'
I17
sS'm'
(lp12627
I1
aI3
aI5
aI3
aI13
aI25
aI33
aI145
aI45
aI979
aI33
aI2211
aI7003
aI11147
aI11327
aI55151
aI30697
asS'd'
I6314
sa(dp12628
S'a'
I4966
sS's'
I17
sS'm'
(lp12629
I1
aI1
aI3
aI13
aI7
aI51
aI25
aI229
aI231
aI115
aI1815
aI3867
aI1533
aI15259
aI8067
aI64803
aI87535
asS'd'
I6315
sa(dp12630
S'a'
I4970
sS's'
I17
sS'm'
(lp12631
I1
aI1
aI3
aI3
aI21
aI51
aI101
aI49
aI227
aI393
aI1659
aI955
aI545
aI7395
aI31563
aI5499
aI130541
asS'd'
I6316
sa(dp12632
S'a'
I4972
sS's'
I17
sS'm'
(lp12633
I1
aI3
aI1
aI1
aI21
aI41
aI57
aI161
aI269
aI35
aI893
aI1817
aI857
aI7027
aI973
aI12529
aI46659
asS'd'
I6317
sa(dp12634
S'a'
I4983
sS's'
I17
sS'm'
(lp12635
I1
aI1
aI3
aI7
aI17
aI35
aI23
aI29
aI335
aI725
aI453
aI1051
aI6019
aI7595
aI29451
aI1853
aI116615
asS'd'
I6318
sa(dp12636
S'a'
I4989
sS's'
I17
sS'm'
(lp12637
I1
aI3
aI3
aI1
aI3
aI55
aI73
aI187
aI213
aI329
aI997
aI703
aI5829
aI7903
aI1081
aI33359
aI119123
asS'd'
I6319
sa(dp12638
S'a'
I4994
sS's'
I17
sS'm'
(lp12639
I1
aI3
aI3
aI15
aI29
aI55
aI15
aI17
aI245
aI117
aI1735
aI767
aI4457
aI8803
aI17621
aI26925
aI72487
asS'd'
I6320
sa(dp12640
S'a'
I5000
sS's'
I17
sS'm'
(lp12641
I1
aI3
aI5
aI3
aI25
aI7
aI119
aI139
aI159
aI199
aI317
aI3875
aI8115
aI7581
aI29239
aI50225
aI48459
asS'd'
I6321
sa(dp12642
S'a'
I5005
sS's'
I17
sS'm'
(lp12643
I1
aI3
aI7
aI11
aI11
aI41
aI107
aI225
aI395
aI545
aI259
aI2379
aI6709
aI11669
aI14545
aI43663
aI69979
asS'd'
I6322
sa(dp12644
S'a'
I5014
sS's'
I17
sS'm'
(lp12645
I1
aI3
aI5
aI13
aI23
aI45
aI73
aI137
aI447
aI305
aI117
aI2659
aI7989
aI233
aI31991
aI60495
aI571
asS'd'
I6323
sa(dp12646
S'a'
I5018
sS's'
I17
sS'm'
(lp12647
I1
aI3
aI7
aI9
aI1
aI37
aI31
aI1
aI433
aI701
aI159
aI3811
aI4529
aI6697
aI7121
aI31107
aI61555
asS'd'
I6324
sa(dp12648
S'a'
I5023
sS's'
I17
sS'm'
(lp12649
I1
aI3
aI5
aI5
aI13
aI21
aI81
aI63
aI95
aI741
aI1189
aI1567
aI1223
aI12371
aI28435
aI10537
aI53785
asS'd'
I6325
sa(dp12650
S'a'
I5039
sS's'
I17
sS'm'
(lp12651
I1
aI1
aI1
aI11
aI17
aI31
aI67
aI121
aI281
aI593
aI561
aI1759
aI387
aI9639
aI28595
aI22473
aI4935
asS'd'
I6326
sa(dp12652
S'a'
I5053
sS's'
I17
sS'm'
(lp12653
I1
aI3
aI7
aI3
aI5
aI43
aI59
aI151
aI351
aI263
aI297
aI423
aI1681
aI3785
aI15171
aI7145
aI57531
asS'd'
I6327
sa(dp12654
S'a'
I5054
sS's'
I17
sS'm'
(lp12655
I1
aI3
aI7
aI15
aI9
aI35
aI105
aI189
aI261
aI175
aI1669
aI1289
aI5401
aI12801
aI19585
aI48169
aI93195
asS'd'
I6328
sa(dp12656
S'a'
I5061
sS's'
I17
sS'm'
(lp12657
I1
aI1
aI7
aI1
aI31
aI41
aI23
aI237
aI151
aI549
aI1079
aI2933
aI5509
aI15593
aI1791
aI15757
aI44607
asS'd'
I6329
sa(dp12658
S'a'
I5065
sS's'
I17
sS'm'
(lp12659
I1
aI1
aI1
aI3
aI29
aI1
aI59
aI115
aI13
aI999
aI1179
aI3561
aI2749
aI10059
aI12861
aI6797
aI11793
asS'd'
I6330
sa(dp12660
S'a'
I5080
sS's'
I17
sS'm'
(lp12661
I1
aI3
aI3
aI7
aI11
aI5
aI23
aI217
aI101
aI775
aI1497
aI4047
aI2427
aI5117
aI9683
aI28895
aI27557
asS'd'
I6331
sa(dp12662
S'a'
I5083
sS's'
I17
sS'm'
(lp12663
I1
aI3
aI7
aI5
aI31
aI55
aI99
aI65
aI55
aI587
aI1271
aI2277
aI7947
aI12995
aI13149
aI4463
aI37625
asS'd'
I6332
sa(dp12664
S'a'
I5107
sS's'
I17
sS'm'
(lp12665
I1
aI1
aI7
aI11
aI3
aI63
aI23
aI191
aI125
aI365
aI1153
aI2657
aI6763
aI4557
aI21643
aI26885
aI36753
asS'd'
I6333
sa(dp12666
S'a'
I5119
sS's'
I17
sS'm'
(lp12667
I1
aI1
aI1
aI15
aI25
aI15
aI111
aI135
aI507
aI745
aI1947
aI2545
aI4329
aI14325
aI8187
aI52021
aI63401
asS'd'
I6334
sa(dp12668
S'a'
I5146
sS's'
I17
sS'm'
(lp12669
I1
aI1
aI3
aI3
aI27
aI25
aI19
aI211
aI393
aI467
aI1015
aI2495
aI7135
aI495
aI10385
aI26961
aI49325
asS'd'
I6335
sa(dp12670
S'a'
I5151
sS's'
I17
sS'm'
(lp12671
I1
aI1
aI3
aI5
aI15
aI35
aI3
aI203
aI337
aI337
aI703
aI1989
aI6869
aI6055
aI21095
aI4749
aI125669
asS'd'
I6336
sa(dp12672
S'a'
I5152
sS's'
I17
sS'm'
(lp12673
I1
aI1
aI5
aI1
aI31
aI39
aI57
aI101
aI419
aI717
aI1489
aI199
aI5729
aI3003
aI2607
aI64593
aI11515
asS'd'
I6337
sa(dp12674
S'a'
I5155
sS's'
I17
sS'm'
(lp12675
I1
aI3
aI7
aI13
aI15
aI3
aI33
aI61
aI17
aI433
aI1097
aI957
aI5351
aI3043
aI3679
aI44881
aI126909
asS'd'
I6338
sa(dp12676
S'a'
I5169
sS's'
I17
sS'm'
(lp12677
I1
aI1
aI3
aI11
aI5
aI1
aI121
aI175
aI119
aI367
aI399
aI2527
aI2157
aI2667
aI31069
aI24797
aI119621
asS'd'
I6339
sa(dp12678
S'a'
I5170
sS's'
I17
sS'm'
(lp12679
I1
aI3
aI1
aI7
aI27
aI47
aI115
aI229
aI455
aI775
aI73
aI837
aI1181
aI3457
aI4057
aI33907
aI67151
asS'd'
I6340
sa(dp12680
S'a'
I5176
sS's'
I17
sS'm'
(lp12681
I1
aI3
aI3
aI1
aI7
aI51
aI71
aI177
aI463
aI921
aI393
aI3137
aI1225
aI5709
aI303
aI20597
aI77581
asS'd'
I6341
sa(dp12682
S'a'
I5179
sS's'
I17
sS'm'
(lp12683
I1
aI3
aI5
aI3
aI31
aI1
aI93
aI53
aI177
aI433
aI1471
aI2191
aI4471
aI9211
aI19397
aI57727
aI60367
asS'd'
I6342
sa(dp12684
S'a'
I5182
sS's'
I17
sS'm'
(lp12685
I1
aI1
aI3
aI11
aI29
aI55
aI121
aI89
aI67
aI869
aI1631
aI2657
aI7357
aI7159
aI22449
aI16357
aI20077
asS'd'
I6343
sa(dp12686
S'a'
I5189
sS's'
I17
sS'm'
(lp12687
I1
aI3
aI7
aI15
aI11
aI39
aI127
aI63
aI211
aI359
aI971
aI1221
aI1909
aI9963
aI7827
aI60923
aI98495
asS'd'
I6344
sa(dp12688
S'a'
I5193
sS's'
I17
sS'm'
(lp12689
I1
aI1
aI7
aI9
aI23
aI47
aI47
aI85
aI307
aI471
aI1287
aI3825
aI5451
aI15151
aI15647
aI63043
aI92443
asS'd'
I6345
sa(dp12690
S'a'
I5196
sS's'
I17
sS'm'
(lp12691
I1
aI3
aI7
aI5
aI19
aI11
aI11
aI27
aI307
aI695
aI99
aI1037
aI1997
aI13673
aI591
aI8183
aI82197
asS'd'
I6346
sa(dp12692
S'a'
I5204
sS's'
I17
sS'm'
(lp12693
I1
aI3
aI5
aI5
aI3
aI53
aI109
aI227
aI503
aI855
aI1269
aI3903
aI5049
aI10647
aI21751
aI58707
aI78311
asS'd'
I6347
sa(dp12694
S'a'
I5207
sS's'
I17
sS'm'
(lp12695
I1
aI1
aI3
aI11
aI31
aI3
aI51
aI211
aI285
aI919
aI487
aI3393
aI3463
aI2271
aI8053
aI56791
aI33763
asS'd'
I6348
sa(dp12696
S'a'
I5211
sS's'
I17
sS'm'
(lp12697
I1
aI3
aI3
aI5
aI21
aI15
aI5
aI5
aI327
aI809
aI915
aI1365
aI7323
aI4247
aI31603
aI26261
aI80389
asS'd'
I6349
sa(dp12698
S'a'
I5220
sS's'
I17
sS'm'
(lp12699
I1
aI3
aI7
aI7
aI15
aI33
aI31
aI221
aI291
aI815
aI1307
aI929
aI3249
aI14573
aI13613
aI59509
aI59741
asS'd'
I6350
sa(dp12700
S'a'
I5258
sS's'
I17
sS'm'
(lp12701
I1
aI3
aI7
aI15
aI19
aI41
aI61
aI27
aI353
aI965
aI1901
aI87
aI2669
aI12757
aI29723
aI47165
aI16521
asS'd'
I6351
sa(dp12702
S'a'
I5265
sS's'
I17
sS'm'
(lp12703
I1
aI3
aI5
aI3
aI11
aI43
aI97
aI215
aI361
aI901
aI1425
aI4063
aI5327
aI14119
aI457
aI43145
aI107401
asS'd'
I6352
sa(dp12704
S'a'
I5271
sS's'
I17
sS'm'
(lp12705
I1
aI1
aI3
aI15
aI19
aI37
aI101
aI69
aI131
aI927
aI897
aI477
aI7641
aI4299
aI21213
aI26017
aI123801
asS'd'
I6353
sa(dp12706
S'a'
I5277
sS's'
I17
sS'm'
(lp12707
I1
aI3
aI7
aI7
aI19
aI5
aI11
aI51
aI277
aI985
aI1071
aI3437
aI6595
aI9547
aI11855
aI64249
aI30957
asS'd'
I6354
sa(dp12708
S'a'
I5278
sS's'
I17
sS'm'
(lp12709
I1
aI1
aI7
aI9
aI21
aI41
aI89
aI113
aI61
aI235
aI685
aI1419
aI7619
aI9863
aI21221
aI28685
aI53409
asS'd'
I6355
sa(dp12710
S'a'
I5282
sS's'
I17
sS'm'
(lp12711
I1
aI1
aI1
aI1
aI27
aI1
aI19
aI3
aI473
aI827
aI269
aI1659
aI2621
aI12347
aI13359
aI64687
aI99293
asS'd'
I6356
sa(dp12712
S'a'
I5296
sS's'
I17
sS'm'
(lp12713
I1
aI3
aI7
aI7
aI29
aI37
aI61
aI49
aI215
aI883
aI625
aI2671
aI3743
aI4517
aI2075
aI64865
aI58611
asS'd'
I6357
sa(dp12714
S'a'
I5299
sS's'
I17
sS'm'
(lp12715
I1
aI3
aI3
aI7
aI15
aI11
aI35
aI37
aI255
aI781
aI613
aI3587
aI7643
aI13081
aI32467
aI14427
aI15235
asS'd'
I6358
sa(dp12716
S'a'
I5319
sS's'
I17
sS'm'
(lp12717
I1
aI1
aI1
aI11
aI31
aI47
aI107
aI65
aI489
aI377
aI425
aI3453
aI2901
aI9999
aI7687
aI13311
aI103947
asS'd'
I6359
sa(dp12718
S'a'
I5328
sS's'
I17
sS'm'
(lp12719
I1
aI3
aI3
aI7
aI9
aI17
aI7
aI107
aI33
aI545
aI407
aI3335
aI7563
aI14315
aI32725
aI8483
aI69093
asS'd'
I6360
sa(dp12720
S'a'
I5343
sS's'
I17
sS'm'
(lp12721
I1
aI1
aI1
aI5
aI17
aI9
aI87
aI229
aI417
aI769
aI423
aI569
aI7073
aI8705
aI24487
aI63743
aI69807
asS'd'
I6361
sa(dp12722
S'a'
I5353
sS's'
I17
sS'm'
(lp12723
I1
aI3
aI1
aI9
aI1
aI29
aI75
aI25
aI483
aI259
aI1941
aI1533
aI8147
aI14127
aI24087
aI37475
aI130961
asS'd'
I6362
sa(dp12724
S'a'
I5364
sS's'
I17
sS'm'
(lp12725
I1
aI3
aI3
aI11
aI15
aI15
aI51
aI45
aI215
aI283
aI1687
aI185
aI4521
aI12205
aI13041
aI33283
aI77007
asS'd'
I6363
sa(dp12726
S'a'
I5368
sS's'
I17
sS'm'
(lp12727
I1
aI1
aI3
aI3
aI5
aI47
aI107
aI67
aI325
aI87
aI1831
aI2845
aI1645
aI1741
aI10811
aI8983
aI58515
asS'd'
I6364
sa(dp12728
S'a'
I5379
sS's'
I17
sS'm'
(lp12729
I1
aI3
aI1
aI13
aI19
aI17
aI1
aI151
aI411
aI915
aI1739
aI3781
aI4939
aI15767
aI25897
aI7205
aI17285
asS'd'
I6365
sa(dp12730
S'a'
I5381
sS's'
I17
sS'm'
(lp12731
I1
aI3
aI5
aI15
aI19
aI1
aI125
aI33
aI321
aI325
aI639
aI4013
aI967
aI4347
aI19743
aI13445
aI61229
asS'd'
I6366
sa(dp12732
S'a'
I5399
sS's'
I17
sS'm'
(lp12733
I1
aI3
aI3
aI13
aI13
aI37
aI71
aI85
aI51
aI775
aI973
aI739
aI4341
aI15707
aI12221
aI24321
aI48073
asS'd'
I6367
sa(dp12734
S'a'
I5415
sS's'
I17
sS'm'
(lp12735
I1
aI1
aI7
aI13
aI15
aI13
aI9
aI211
aI331
aI429
aI1323
aI3027
aI1091
aI13311
aI289
aI57789
aI93261
asS'd'
I6368
sa(dp12736
S'a'
I5422
sS's'
I17
sS'm'
(lp12737
I1
aI1
aI1
aI1
aI27
aI7
aI13
aI27
aI67
aI573
aI455
aI2353
aI113
aI11831
aI9069
aI4503
aI89291
asS'd'
I6369
sa(dp12738
S'a'
I5441
sS's'
I17
sS'm'
(lp12739
I1
aI1
aI1
aI7
aI21
aI63
aI47
aI39
aI419
aI991
aI1623
aI11
aI3153
aI12633
aI9425
aI65087
aI44935
asS'd'
I6370
sa(dp12740
S'a'
I5451
sS's'
I17
sS'm'
(lp12741
I1
aI3
aI1
aI7
aI23
aI11
aI15
aI11
aI99
aI543
aI1739
aI3955
aI5883
aI12469
aI7529
aI14177
aI1945
asS'd'
I6371
sa(dp12742
S'a'
I5456
sS's'
I17
sS'm'
(lp12743
I1
aI3
aI1
aI3
aI5
aI17
aI31
aI251
aI387
aI311
aI725
aI3827
aI6835
aI5065
aI3141
aI43441
aI87955
asS'd'
I6372
sa(dp12744
S'a'
I5462
sS's'
I17
sS'm'
(lp12745
I1
aI1
aI1
aI11
aI25
aI7
aI75
aI135
aI67
aI589
aI889
aI3429
aI155
aI9081
aI28653
aI8059
aI57251
asS'd'
I6373
sa(dp12746
S'a'
I5490
sS's'
I17
sS'm'
(lp12747
I1
aI3
aI5
aI15
aI21
aI15
aI103
aI149
aI311
aI407
aI1391
aI717
aI1765
aI14887
aI14381
aI37483
aI29587
asS'd'
I6374
sa(dp12748
S'a'
I5495
sS's'
I17
sS'm'
(lp12749
I1
aI3
aI5
aI5
aI19
aI31
aI93
aI5
aI507
aI193
aI1735
aI3841
aI7895
aI9853
aI10317
aI14867
aI49529
asS'd'
I6375
sa(dp12750
S'a'
I5501
sS's'
I17
sS'm'
(lp12751
I1
aI3
aI7
aI7
aI19
aI3
aI99
aI201
aI479
aI313
aI693
aI3435
aI5453
aI1157
aI23127
aI49005
aI20167
asS'd'
I6376
sa(dp12752
S'a'
I5502
sS's'
I17
sS'm'
(lp12753
I1
aI3
aI7
aI9
aI15
aI21
aI123
aI41
aI19
aI281
aI1837
aI2589
aI1003
aI1993
aI18345
aI10039
aI89325
asS'd'
I6377
sa(dp12754
S'a'
I5505
sS's'
I17
sS'm'
(lp12755
I1
aI3
aI5
aI1
aI19
aI21
aI77
aI151
aI145
aI951
aI2017
aI609
aI5847
aI4475
aI12439
aI6357
aI108277
asS'd'
I6378
sa(dp12756
S'a'
I5512
sS's'
I17
sS'm'
(lp12757
I1
aI1
aI1
aI9
aI17
aI21
aI91
aI91
aI111
aI951
aI497
aI1759
aI503
aI12787
aI25117
aI24323
aI96447
asS'd'
I6379
sa(dp12758
S'a'
I5523
sS's'
I17
sS'm'
(lp12759
I1
aI1
aI3
aI11
aI13
aI9
aI73
aI205
aI329
aI243
aI1187
aI829
aI2821
aI5563
aI14391
aI771
aI116441
asS'd'
I6380
sa(dp12760
S'a'
I5529
sS's'
I17
sS'm'
(lp12761
I1
aI1
aI1
aI1
aI11
aI57
aI39
aI221
aI41
aI521
aI1541
aI3515
aI2367
aI4179
aI21039
aI52943
aI11627
asS'd'
I6381
sa(dp12762
S'a'
I5548
sS's'
I17
sS'm'
(lp12763
I1
aI3
aI3
aI3
aI23
aI13
aI103
aI125
aI67
aI217
aI863
aI3755
aI213
aI12657
aI31399
aI3771
aI54107
asS'd'
I6382
sa(dp12764
S'a'
I5551
sS's'
I17
sS'm'
(lp12765
I1
aI3
aI3
aI7
aI3
aI9
aI107
aI217
aI497
aI935
aI519
aI3041
aI323
aI14895
aI5695
aI28789
aI36085
asS'd'
I6383
sa(dp12766
S'a'
I5553
sS's'
I17
sS'm'
(lp12767
I1
aI1
aI5
aI11
aI23
aI33
aI81
aI23
aI167
aI3
aI1683
aI2279
aI5365
aI847
aI14717
aI9689
aI64481
asS'd'
I6384
sa(dp12768
S'a'
I5565
sS's'
I17
sS'm'
(lp12769
I1
aI3
aI1
aI7
aI1
aI15
aI107
aI93
aI429
aI363
aI1745
aI1459
aI5879
aI8351
aI17527
aI44001
aI70293
asS'd'
I6385
sa(dp12770
S'a'
I5568
sS's'
I17
sS'm'
(lp12771
I1
aI3
aI3
aI9
aI27
aI55
aI125
aI211
aI141
aI827
aI1239
aI663
aI4803
aI11067
aI32039
aI28091
aI56421
asS'd'
I6386
sa(dp12772
S'a'
I5577
sS's'
I17
sS'm'
(lp12773
I1
aI3
aI5
aI5
aI7
aI13
aI125
aI231
aI427
aI483
aI967
aI549
aI3105
aI13919
aI3017
aI39207
aI23253
asS'd'
I6387
sa(dp12774
S'a'
I5578
sS's'
I17
sS'm'
(lp12775
I1
aI3
aI7
aI3
aI21
aI29
aI79
aI67
aI39
aI451
aI157
aI337
aI3585
aI3621
aI9545
aI31205
aI63201
asS'd'
I6388
sa(dp12776
S'a'
I5583
sS's'
I17
sS'm'
(lp12777
I1
aI3
aI1
aI1
aI29
aI25
aI77
aI57
aI167
aI899
aI95
aI2487
aI3743
aI5381
aI3637
aI56289
aI39453
asS'd'
I6389
sa(dp12778
S'a'
I5585
sS's'
I17
sS'm'
(lp12779
I1
aI1
aI1
aI9
aI29
aI19
aI41
aI97
aI75
aI199
aI1709
aI483
aI4099
aI3113
aI10953
aI20659
aI109273
asS'd'
I6390
sa(dp12780
S'a'
I5588
sS's'
I17
sS'm'
(lp12781
I1
aI3
aI5
aI15
aI13
aI9
aI83
aI43
aI111
aI789
aI965
aI4061
aI1239
aI14577
aI10113
aI26359
aI52609
asS'd'
I6391
sa(dp12782
S'a'
I5613
sS's'
I17
sS'm'
(lp12783
I1
aI3
aI5
aI5
aI11
aI39
aI113
aI31
aI457
aI119
aI725
aI831
aI4143
aI5675
aI27431
aI12431
aI94977
asS'd'
I6392
sa(dp12784
S'a'
I5614
sS's'
I17
sS'm'
(lp12785
I1
aI1
aI3
aI3
aI25
aI17
aI93
aI253
aI307
aI625
aI143
aI1061
aI4415
aI3563
aI3313
aI53527
aI29537
asS'd'
I6393
sa(dp12786
S'a'
I5616
sS's'
I17
sS'm'
(lp12787
I1
aI3
aI5
aI3
aI29
aI41
aI43
aI109
aI147
aI919
aI1675
aI465
aI6101
aI12251
aI28915
aI15397
aI85233
asS'd'
I6394
sa(dp12788
S'a'
I5622
sS's'
I17
sS'm'
(lp12789
I1
aI1
aI1
aI1
aI31
aI25
aI59
aI187
aI439
aI561
aI559
aI413
aI1917
aI9319
aI27475
aI49715
aI32953
asS'd'
I6395
sa(dp12790
S'a'
I5631
sS's'
I17
sS'm'
(lp12791
I1
aI1
aI7
aI13
aI23
aI31
aI95
aI231
aI141
aI207
aI1373
aI2173
aI2905
aI169
aI23825
aI55071
aI6147
asS'd'
I6396
sa(dp12792
S'a'
I5637
sS's'
I17
sS'm'
(lp12793
I1
aI1
aI7
aI13
aI15
aI39
aI43
aI117
aI321
aI297
aI661
aI2941
aI7359
aI11675
aI15483
aI24093
aI7269
asS'd'
I6397
sa(dp12794
S'a'
I5638
sS's'
I17
sS'm'
(lp12795
I1
aI3
aI3
aI13
aI9
aI59
aI51
aI49
aI81
aI563
aI745
aI1843
aI295
aI4689
aI19847
aI42137
aI63197
asS'd'
I6398
sa(dp12796
S'a'
I5668
sS's'
I17
sS'm'
(lp12797
I1
aI3
aI1
aI9
aI5
aI33
aI21
aI199
aI509
aI927
aI1777
aI1349
aI3593
aI1065
aI24943
aI55667
aI73539
asS'd'
I6399
sa(dp12798
S'a'
I5675
sS's'
I17
sS'm'
(lp12799
I1
aI3
aI1
aI11
aI17
aI15
aI91
aI21
aI59
aI587
aI1207
aI543
aI6669
aI10861
aI24755
aI1789
aI91249
asS'd'
I6400
sa(dp12800
S'a'
I5683
sS's'
I17
sS'm'
(lp12801
I1
aI3
aI7
aI15
aI13
aI47
aI57
aI147
aI381
aI1021
aI921
aI1347
aI3847
aI5969
aI9075
aI39081
aI127241
asS'd'
I6401
sa(dp12802
S'a'
I5695
sS's'
I17
sS'm'
(lp12803
I1
aI3
aI3
aI15
aI19
aI15
aI1
aI97
aI203
aI409
aI1745
aI1217
aI2199
aI7945
aI24361
aI41771
aI123127
asS'd'
I6402
sa(dp12804
S'a'
I5703
sS's'
I17
sS'm'
(lp12805
I1
aI3
aI3
aI5
aI17
aI17
aI43
aI255
aI179
aI717
aI1993
aI645
aI6527
aI1533
aI32719
aI27481
aI122425
asS'd'
I6403
sa(dp12806
S'a'
I5710
sS's'
I17
sS'm'
(lp12807
I1
aI3
aI5
aI9
aI13
aI59
aI15
aI157
aI373
aI937
aI27
aI3325
aI2297
aI89
aI10861
aI48615
aI16083
asS'd'
I6404
sa(dp12808
S'a'
I5715
sS's'
I17
sS'm'
(lp12809
I1
aI3
aI1
aI3
aI19
aI27
aI109
aI243
aI189
aI17
aI99
aI1879
aI695
aI11329
aI12467
aI6053
aI41749
asS'd'
I6405
sa(dp12810
S'a'
I5727
sS's'
I17
sS'm'
(lp12811
I1
aI1
aI5
aI5
aI23
aI41
aI103
aI69
aI171
aI917
aI1303
aI2101
aI617
aI10017
aI26525
aI11009
aI66137
asS'd'
I6406
sa(dp12812
S'a'
I5738
sS's'
I17
sS'm'
(lp12813
I1
aI1
aI1
aI9
aI21
aI45
aI47
aI171
aI455
aI257
aI411
aI4021
aI6995
aI12881
aI4793
aI51193
aI60775
asS'd'
I6407
sa(dp12814
S'a'
I5752
sS's'
I17
sS'm'
(lp12815
I1
aI3
aI7
aI5
aI25
aI31
aI89
aI53
aI321
aI593
aI1795
aI2435
aI3833
aI2767
aI17241
aI63373
aI25457
asS'd'
I6408
sa(dp12816
S'a'
I5767
sS's'
I17
sS'm'
(lp12817
I1
aI3
aI1
aI1
aI3
aI45
aI19
aI255
aI179
aI991
aI1407
aI3683
aI1435
aI6803
aI12215
aI12835
aI2005
asS'd'
I6409
sa(dp12818
S'a'
I5773
sS's'
I17
sS'm'
(lp12819
I1
aI3
aI7
aI3
aI17
aI5
aI117
aI251
aI71
aI983
aI1391
aI3499
aI5119
aI7257
aI7325
aI16565
aI6321
asS'd'
I6410
sa(dp12820
S'a'
I5776
sS's'
I17
sS'm'
(lp12821
I1
aI3
aI5
aI7
aI5
aI49
aI47
aI201
aI297
aI485
aI1879
aI2205
aI4903
aI13619
aI22537
aI5479
aI121625
asS'd'
I6411
sa(dp12822
S'a'
I5781
sS's'
I17
sS'm'
(lp12823
I1
aI1
aI3
aI5
aI27
aI27
aI87
aI61
aI145
aI943
aI343
aI1639
aI6307
aI4549
aI20765
aI33479
aI113697
asS'd'
I6412
sa(dp12824
S'a'
I5791
sS's'
I17
sS'm'
(lp12825
I1
aI1
aI3
aI9
aI17
aI5
aI101
aI129
aI305
aI653
aI1901
aI3901
aI6361
aI2369
aI7449
aI55259
aI75215
asS'd'
I6413
sa(dp12826
S'a'
I5792
sS's'
I17
sS'm'
(lp12827
I1
aI1
aI7
aI5
aI31
aI45
aI117
aI55
aI335
aI827
aI1309
aI2603
aI2111
aI11005
aI14747
aI56999
aI97373
asS'd'
I6414
sa(dp12828
S'a'
I5795
sS's'
I17
sS'm'
(lp12829
I1
aI1
aI7
aI11
aI29
aI29
aI81
aI175
aI169
aI453
aI293
aI2589
aI1057
aI15795
aI32397
aI65433
aI79455
asS'd'
I6415
sa(dp12830
S'a'
I5798
sS's'
I17
sS'm'
(lp12831
I1
aI1
aI1
aI5
aI11
aI7
aI13
aI249
aI29
aI407
aI1289
aI2385
aI8113
aI15327
aI4029
aI32005
aI105901
asS'd'
I6416
sa(dp12832
S'a'
I5801
sS's'
I17
sS'm'
(lp12833
I1
aI1
aI5
aI5
aI7
aI61
aI103
aI141
aI109
aI391
aI631
aI821
aI1479
aI14771
aI25057
aI1415
aI8081
asS'd'
I6417
sa(dp12834
S'a'
I5810
sS's'
I17
sS'm'
(lp12835
I1
aI3
aI1
aI1
aI9
aI37
aI17
aI231
aI501
aI745
aI1695
aI45
aI7797
aI2945
aI5529
aI34747
aI39069
asS'd'
I6418
sa(dp12836
S'a'
I5812
sS's'
I17
sS'm'
(lp12837
I1
aI1
aI7
aI9
aI21
aI59
aI103
aI103
aI33
aI875
aI723
aI3477
aI4729
aI7311
aI29979
aI60901
aI72187
asS'd'
I6419
sa(dp12838
S'a'
I5836
sS's'
I17
sS'm'
(lp12839
I1
aI3
aI3
aI3
aI15
aI63
aI93
aI237
aI203
aI635
aI1189
aI2035
aI6499
aI9943
aI9133
aI62977
aI29657
asS'd'
I6420
sa(dp12840
S'a'
I5839
sS's'
I17
sS'm'
(lp12841
I1
aI1
aI1
aI9
aI3
aI11
aI63
aI207
aI95
aI563
aI775
aI3009
aI7125
aI13141
aI4489
aI16343
aI120951
asS'd'
I6421
sa(dp12842
S'a'
I5841
sS's'
I17
sS'm'
(lp12843
I1
aI1
aI3
aI1
aI21
aI57
aI15
aI217
aI185
aI305
aI463
aI1597
aI6529
aI4989
aI14011
aI11265
aI131031
asS'd'
I6422
sa(dp12844
S'a'
I5867
sS's'
I17
sS'm'
(lp12845
I1
aI3
aI5
aI15
aI17
aI61
aI35
aI127
aI411
aI579
aI1349
aI615
aI3293
aI8475
aI9773
aI30635
aI117639
asS'd'
I6423
sa(dp12846
S'a'
I5870
sS's'
I17
sS'm'
(lp12847
I1
aI1
aI7
aI9
aI11
aI3
aI55
aI105
aI305
aI223
aI1899
aI2217
aI1261
aI9831
aI23693
aI3013
aI30489
asS'd'
I6424
sa(dp12848
S'a'
I5877
sS's'
I17
sS'm'
(lp12849
I1
aI3
aI7
aI15
aI15
aI29
aI1
aI99
aI67
aI293
aI499
aI1941
aI5303
aI1329
aI24547
aI14065
aI7927
asS'd'
I6425
sa(dp12850
S'a'
I5881
sS's'
I17
sS'm'
(lp12851
I1
aI1
aI5
aI11
aI17
aI55
aI71
aI49
aI499
aI435
aI985
aI2803
aI6139
aI1503
aI24167
aI47181
aI102529
asS'd'
I6426
sa(dp12852
S'a'
I5899
sS's'
I17
sS'm'
(lp12853
I1
aI3
aI5
aI1
aI19
aI53
aI71
aI17
aI63
aI469
aI1871
aI2051
aI357
aI11661
aI5689
aI36373
aI13379
asS'd'
I6427
sa(dp12854
S'a'
I5914
sS's'
I17
sS'm'
(lp12855
I1
aI1
aI5
aI1
aI27
aI47
aI23
aI247
aI59
aI381
aI1895
aI2453
aI3665
aI5487
aI24081
aI50501
aI91659
asS'd'
I6428
sa(dp12856
S'a'
I5925
sS's'
I17
sS'm'
(lp12857
I1
aI1
aI5
aI7
aI29
aI19
aI3
aI33
aI83
aI301
aI133
aI3603
aI5133
aI16171
aI22905
aI36271
aI10405
asS'd'
I6429
sa(dp12858
S'a'
I5929
sS's'
I17
sS'm'
(lp12859
I1
aI3
aI7
aI9
aI11
aI23
aI57
aI87
aI9
aI731
aI631
aI3703
aI2593
aI12851
aI7115
aI8801
aI108919
asS'd'
I6430
sa(dp12860
S'a'
I5943
sS's'
I17
sS'm'
(lp12861
I1
aI3
aI3
aI3
aI23
aI35
aI33
aI99
aI343
aI837
aI231
aI3921
aI6975
aI15093
aI15049
aI64623
aI123523
asS'd'
I6431
sa(dp12862
S'a'
I5949
sS's'
I17
sS'm'
(lp12863
I1
aI1
aI7
aI11
aI15
aI61
aI113
aI103
aI501
aI57
aI1345
aI3155
aI2965
aI4433
aI10605
aI43765
aI42169
asS'd'
I6432
sa(dp12864
S'a'
I5962
sS's'
I17
sS'm'
(lp12865
I1
aI1
aI7
aI13
aI7
aI53
aI91
aI121
aI229
aI127
aI103
aI833
aI7829
aI1571
aI10847
aI20861
aI101155
asS'd'
I6433
sa(dp12866
S'a'
I5969
sS's'
I17
sS'm'
(lp12867
I1
aI3
aI7
aI1
aI9
aI25
aI71
aI103
aI37
aI473
aI1133
aI1129
aI1651
aI6965
aI6937
aI16597
aI20439
asS'd'
I6434
sa(dp12868
S'a'
I5976
sS's'
I17
sS'm'
(lp12869
I1
aI1
aI5
aI9
aI1
aI9
aI47
aI131
aI285
aI967
aI1869
aI1075
aI8127
aI135
aI15575
aI38569
aI123729
asS'd'
I6435
sa(dp12870
S'a'
I5988
sS's'
I17
sS'm'
(lp12871
I1
aI1
aI7
aI9
aI5
aI31
aI33
aI227
aI347
aI41
aI2025
aI3755
aI857
aI7805
aI13121
aI38307
aI125825
asS'd'
I6436
sa(dp12872
S'a'
I5997
sS's'
I17
sS'm'
(lp12873
I1
aI3
aI5
aI7
aI11
aI11
aI19
aI55
aI23
aI627
aI1477
aI3093
aI2779
aI7653
aI7165
aI23053
aI76123
asS'd'
I6437
sa(dp12874
S'a'
I6006
sS's'
I17
sS'm'
(lp12875
I1
aI1
aI3
aI1
aI3
aI47
aI83
aI89
aI177
aI381
aI1247
aI141
aI7051
aI6443
aI27369
aI34323
aI43063
asS'd'
I6438
sa(dp12876
S'a'
I6010
sS's'
I17
sS'm'
(lp12877
I1
aI1
aI7
aI7
aI13
aI15
aI55
aI223
aI351
aI525
aI1051
aI3009
aI5443
aI11499
aI8335
aI37949
aI69149
asS'd'
I6439
sa(dp12878
S'a'
I6016
sS's'
I17
sS'm'
(lp12879
I1
aI1
aI1
aI3
aI13
aI61
aI89
aI33
aI129
aI921
aI1905
aI201
aI3141
aI5531
aI135
aI34103
aI56883
asS'd'
I6440
sa(dp12880
S'a'
I6022
sS's'
I17
sS'm'
(lp12881
I1
aI1
aI5
aI13
aI17
aI27
aI13
aI163
aI169
aI471
aI1263
aI1421
aI7015
aI7927
aI21027
aI58001
aI26739
asS'd'
I6441
sa(dp12882
S'a'
I6026
sS's'
I17
sS'm'
(lp12883
I1
aI1
aI1
aI15
aI19
aI49
aI109
aI207
aI245
aI49
aI1271
aI3635
aI2561
aI5091
aI24415
aI59195
aI67701
asS'd'
I6442
sa(dp12884
S'a'
I6031
sS's'
I17
sS'm'
(lp12885
I1
aI3
aI5
aI7
aI27
aI57
aI99
aI155
aI461
aI595
aI1859
aI1727
aI857
aI4993
aI31733
aI42141
aI10035
asS'd'
I6443
sa(dp12886
S'a'
I6040
sS's'
I17
sS'm'
(lp12887
I1
aI1
aI1
aI15
aI11
aI11
aI85
aI9
aI251
aI375
aI155
aI379
aI7501
aI12559
aI32583
aI36317
aI4675
asS'd'
I6444
sa(dp12888
S'a'
I6043
sS's'
I17
sS'm'
(lp12889
I1
aI1
aI5
aI13
aI19
aI57
aI81
aI69
aI201
aI293
aI593
aI3169
aI4519
aI9057
aI16685
aI12847
aI123797
asS'd'
I6445
sa(dp12890
S'a'
I6050
sS's'
I17
sS'm'
(lp12891
I1
aI3
aI1
aI5
aI5
aI1
aI19
aI243
aI345
aI661
aI561
aI3549
aI2541
aI5887
aI25879
aI41467
aI72799
asS'd'
I6446
sa(dp12892
S'a'
I6059
sS's'
I17
sS'm'
(lp12893
I1
aI1
aI5
aI13
aI15
aI51
aI67
aI61
aI79
aI89
aI447
aI1471
aI4915
aI10637
aI10901
aI48157
aI103545
asS'd'
I6447
sa(dp12894
S'a'
I6079
sS's'
I17
sS'm'
(lp12895
I1
aI3
aI5
aI13
aI31
aI25
aI73
aI129
aI435
aI659
aI1851
aI3595
aI753
aI7717
aI10927
aI30115
aI109221
asS'd'
I6448
sa(dp12896
S'a'
I6099
sS's'
I17
sS'm'
(lp12897
I1
aI1
aI1
aI3
aI25
aI3
aI121
aI43
aI349
aI205
aI1209
aI2671
aI6445
aI8755
aI7171
aI58631
aI74319
asS'd'
I6449
sa(dp12898
S'a'
I6101
sS's'
I17
sS'm'
(lp12899
I1
aI1
aI3
aI1
aI11
aI15
aI83
aI37
aI483
aI65
aI759
aI1835
aI3883
aI1693
aI30051
aI61077
aI1187
asS'd'
I6450
sa(dp12900
S'a'
I6105
sS's'
I17
sS'm'
(lp12901
I1
aI3
aI7
aI15
aI29
aI23
aI85
aI77
aI139
aI903
aI1821
aI943
aI6453
aI1523
aI18539
aI49039
aI110787
asS'd'
I6451
sa(dp12902
S'a'
I6108
sS's'
I17
sS'm'
(lp12903
I1
aI1
aI7
aI15
aI15
aI17
aI69
aI253
aI507
aI921
aI523
aI79
aI747
aI4011
aI25795
aI42029
aI88309
asS'd'
I6452
sa(dp12904
S'a'
I6124
sS's'
I17
sS'm'
(lp12905
I1
aI1
aI7
aI3
aI25
aI47
aI119
aI83
aI313
aI45
aI985
aI145
aI205
aI3407
aI9013
aI64517
aI115811
asS'd'
I6453
sa(dp12906
S'a'
I6132
sS's'
I17
sS'm'
(lp12907
I1
aI1
aI7
aI1
aI29
aI21
aI9
aI123
aI97
aI545
aI1987
aI2979
aI6901
aI12667
aI23325
aI63635
aI70593
asS'd'
I6454
sa(dp12908
S'a'
I6145
sS's'
I17
sS'm'
(lp12909
I1
aI3
aI7
aI3
aI23
aI45
aI81
aI255
aI41
aI29
aI1493
aI4065
aI3201
aI10479
aI17193
aI39999
aI55493
asS'd'
I6455
sa(dp12910
S'a'
I6146
sS's'
I17
sS'm'
(lp12911
I1
aI3
aI1
aI3
aI9
aI43
aI43
aI135
aI235
aI603
aI481
aI3139
aI2729
aI14759
aI7269
aI7995
aI110351
asS'd'
I6456
sa(dp12912
S'a'
I6151
sS's'
I17
sS'm'
(lp12913
I1
aI3
aI1
aI11
aI17
aI35
aI113
aI93
aI417
aI967
aI755
aI659
aI3115
aI16163
aI22997
aI38205
aI126961
asS'd'
I6457
sa(dp12914
S'a'
I6152
sS's'
I17
sS'm'
(lp12915
I1
aI1
aI7
aI11
aI29
aI57
aI81
aI235
aI93
aI869
aI475
aI825
aI6269
aI15819
aI14977
aI53057
aI116021
asS'd'
I6458
sa(dp12916
S'a'
I6158
sS's'
I17
sS'm'
(lp12917
I1
aI1
aI7
aI13
aI5
aI61
aI5
aI241
aI245
aI673
aI1651
aI3367
aI2355
aI713
aI20107
aI30133
aI735
asS'd'
I6459
sa(dp12918
S'a'
I6160
sS's'
I17
sS'm'
(lp12919
I1
aI1
aI5
aI9
aI21
aI3
aI121
aI241
aI129
aI703
aI1435
aI1943
aI5087
aI13123
aI30023
aI58287
aI50377
asS'd'
I6460
sa(dp12920
S'a'
I6163
sS's'
I17
sS'm'
(lp12921
I1
aI1
aI1
aI15
aI23
aI27
aI67
aI197
aI123
aI629
aI169
aI3303
aI1679
aI11051
aI16875
aI28055
aI12379
asS'd'
I6461
sa(dp12922
S'a'
I6165
sS's'
I17
sS'm'
(lp12923
I1
aI1
aI3
aI3
aI7
aI63
aI97
aI43
aI89
aI739
aI779
aI2893
aI7763
aI6351
aI26135
aI44647
aI127987
asS'd'
I6462
sa(dp12924
S'a'
I6170
sS's'
I17
sS'm'
(lp12925
I1
aI3
aI3
aI9
aI31
aI59
aI95
aI131
aI131
aI321
aI1125
aI127
aI4865
aI145
aI26237
aI47871
aI114549
asS'd'
I6463
sa(dp12926
S'a'
I6182
sS's'
I17
sS'm'
(lp12927
I1
aI3
aI3
aI13
aI21
aI3
aI33
aI17
aI445
aI693
aI1599
aI2517
aI1679
aI2237
aI15053
aI30983
aI106755
asS'd'
I6464
sa(dp12928
S'a'
I6196
sS's'
I17
sS'm'
(lp12929
I1
aI1
aI5
aI13
aI31
aI37
aI49
aI67
aI403
aI27
aI575
aI1795
aI3385
aI1067
aI585
aI60277
aI123189
asS'd'
I6465
sa(dp12930
S'a'
I6199
sS's'
I17
sS'm'
(lp12931
I1
aI3
aI1
aI15
aI13
aI35
aI23
aI247
aI493
aI305
aI363
aI451
aI4011
aI3679
aI18281
aI31751
aI127933
asS'd'
I6466
sa(dp12932
S'a'
I6200
sS's'
I17
sS'm'
(lp12933
I1
aI1
aI7
aI5
aI21
aI45
aI123
aI253
aI469
aI267
aI985
aI2349
aI3427
aI7653
aI25685
aI13747
aI531
asS'd'
I6467
sa(dp12934
S'a'
I6205
sS's'
I17
sS'm'
(lp12935
I1
aI1
aI5
aI11
aI7
aI59
aI105
aI209
aI27
aI847
aI593
aI3775
aI6165
aI1655
aI29867
aI28465
aI92193
asS'd'
I6468
sa(dp12936
S'a'
I6226
sS's'
I17
sS'm'
(lp12937
I1
aI3
aI1
aI11
aI7
aI25
aI101
aI81
aI233
aI311
aI9
aI2735
aI3951
aI485
aI10105
aI24489
aI649
asS'd'
I6469
sa(dp12938
S'a'
I6228
sS's'
I17
sS'm'
(lp12939
I1
aI3
aI1
aI7
aI27
aI5
aI115
aI243
aI295
aI659
aI215
aI1787
aI5131
aI2513
aI29201
aI21195
aI103383
asS'd'
I6470
sa(dp12940
S'a'
I6237
sS's'
I17
sS'm'
(lp12941
I1
aI3
aI5
aI13
aI29
aI21
aI7
aI57
aI345
aI467
aI1297
aI207
aI5115
aI335
aI6153
aI32959
aI125697
asS'd'
I6471
sa(dp12942
S'a'
I6247
sS's'
I17
sS'm'
(lp12943
I1
aI1
aI1
aI9
aI3
aI63
aI63
aI5
aI373
aI123
aI1265
aI2365
aI1623
aI1561
aI14805
aI17487
aI104787
asS'd'
I6472
sa(dp12944
S'a'
I6251
sS's'
I17
sS'm'
(lp12945
I1
aI3
aI1
aI5
aI15
aI13
aI55
aI69
aI251
aI341
aI463
aI2611
aI4793
aI12157
aI4669
aI11613
aI128705
asS'd'
I6473
sa(dp12946
S'a'
I6253
sS's'
I17
sS'm'
(lp12947
I1
aI3
aI7
aI13
aI19
aI7
aI93
aI149
aI453
aI693
aI1731
aI861
aI6971
aI943
aI18891
aI56547
aI34411
asS'd'
I6474
sa(dp12948
S'a'
I6256
sS's'
I17
sS'm'
(lp12949
I1
aI1
aI7
aI1
aI27
aI49
aI27
aI9
aI281
aI121
aI581
aI393
aI2583
aI1159
aI26989
aI39955
aI100765
asS'd'
I6475
sa(dp12950
S'a'
I6268
sS's'
I17
sS'm'
(lp12951
I1
aI1
aI3
aI9
aI3
aI43
aI97
aI207
aI311
aI617
aI1987
aI2559
aI2101
aI15791
aI30085
aI40713
aI41909
asS'd'
I6476
sa(dp12952
S'a'
I6272
sS's'
I17
sS'm'
(lp12953
I1
aI3
aI1
aI3
aI15
aI19
aI53
aI183
aI375
aI867
aI397
aI3203
aI4207
aI5381
aI25065
aI60357
aI88739
asS'd'
I6477
sa(dp12954
S'a'
I6275
sS's'
I17
sS'm'
(lp12955
I1
aI3
aI3
aI3
aI27
aI51
aI85
aI231
aI19
aI559
aI567
aI4049
aI4875
aI14201
aI11623
aI39763
aI57339
asS'd'
I6478
sa(dp12956
S'a'
I6281
sS's'
I17
sS'm'
(lp12957
I1
aI1
aI5
aI1
aI19
aI7
aI81
aI249
aI41
aI789
aI985
aI3725
aI4053
aI4255
aI9861
aI1609
aI29511
asS'd'
I6479
sa(dp12958
S'a'
I6289
sS's'
I17
sS'm'
(lp12959
I1
aI3
aI5
aI5
aI21
aI13
aI49
aI41
aI367
aI283
aI1161
aI2753
aI4733
aI3691
aI27931
aI53055
aI83625
asS'd'
I6480
sa(dp12960
S'a'
I6335
sS's'
I17
sS'm'
(lp12961
I1
aI3
aI5
aI11
aI29
aI47
aI95
aI51
aI265
aI85
aI385
aI833
aI7957
aI14985
aI7017
aI41937
aI41377
asS'd'
I6481
sa(dp12962
S'a'
I6338
sS's'
I17
sS'm'
(lp12963
I1
aI1
aI7
aI5
aI1
aI23
aI17
aI191
aI185
aI323
aI515
aI3183
aI7685
aI7361
aI21143
aI5227
aI110297
asS'd'
I6482
sa(dp12964
S'a'
I6355
sS's'
I17
sS'm'
(lp12965
I1
aI3
aI3
aI7
aI11
aI39
aI31
aI97
aI237
aI497
aI1649
aI3529
aI6153
aI5055
aI29021
aI35125
aI121581
asS'd'
I6483
sa(dp12966
S'a'
I6362
sS's'
I17
sS'm'
(lp12967
I1
aI3
aI5
aI3
aI17
aI47
aI105
aI75
aI55
aI343
aI595
aI2447
aI5575
aI10673
aI32015
aI37541
aI127867
asS'd'
I6484
sa(dp12968
S'a'
I6373
sS's'
I17
sS'm'
(lp12969
I1
aI3
aI1
aI7
aI19
aI39
aI31
aI135
aI167
aI979
aI219
aI1353
aI489
aI9667
aI27107
aI55565
aI72291
asS'd'
I6485
sa(dp12970
S'a'
I6386
sS's'
I17
sS'm'
(lp12971
I1
aI1
aI3
aI13
aI31
aI49
aI87
aI93
aI235
aI577
aI1551
aI2663
aI387
aI1129
aI26683
aI31285
aI15913
asS'd'
I6486
sa(dp12972
S'a'
I6388
sS's'
I17
sS'm'
(lp12973
I1
aI3
aI3
aI7
aI15
aI29
aI61
aI33
aI115
aI511
aI1781
aI2029
aI4265
aI6745
aI1467
aI34415
aI40907
asS'd'
I6487
sa(dp12974
S'a'
I6391
sS's'
I17
sS'm'
(lp12975
I1
aI1
aI7
aI5
aI1
aI55
aI13
aI129
aI167
aI937
aI79
aI2047
aI3589
aI1979
aI4153
aI15229
aI85745
asS'd'
I6488
sa(dp12976
S'a'
I6397
sS's'
I17
sS'm'
(lp12977
I1
aI1
aI7
aI15
aI15
aI25
aI89
aI129
aI31
aI435
aI1359
aI49
aI2659
aI2829
aI8741
aI25215
aI4239
asS'd'
I6489
sa(dp12978
S'a'
I6405
sS's'
I17
sS'm'
(lp12979
I1
aI3
aI5
aI3
aI11
aI39
aI95
aI239
aI187
aI615
aI1481
aI3509
aI1133
aI13497
aI24833
aI59635
aI45695
asS'd'
I6490
sa(dp12980
S'a'
I6406
sS's'
I17
sS'm'
(lp12981
I1
aI1
aI5
aI3
aI19
aI17
aI17
aI235
aI315
aI943
aI883
aI1381
aI7129
aI15709
aI9847
aI41183
aI116071
asS'd'
I6491
sa(dp12982
S'a'
I6410
sS's'
I17
sS'm'
(lp12983
I1
aI1
aI1
aI3
aI9
aI63
aI109
aI209
aI309
aI1015
aI1391
aI2617
aI1481
aI6483
aI4151
aI28063
aI49887
asS'd'
I6492
sa(dp12984
S'a'
I6417
sS's'
I17
sS'm'
(lp12985
I1
aI1
aI5
aI13
aI23
aI37
aI31
aI89
aI501
aI461
aI41
aI931
aI7863
aI15499
aI25635
aI16995
aI41651
asS'd'
I6493
sa(dp12986
S'a'
I6443
sS's'
I17
sS'm'
(lp12987
I1
aI1
aI1
aI9
aI29
aI29
aI125
aI161
aI219
aI439
aI1465
aI1615
aI7483
aI7497
aI1121
aI49693
aI30269
asS'd'
I6494
sa(dp12988
S'a'
I6457
sS's'
I17
sS'm'
(lp12989
I1
aI3
aI1
aI5
aI7
aI43
aI27
aI161
aI431
aI375
aI419
aI2995
aI527
aI8207
aI747
aI18491
aI15351
asS'd'
I6495
sa(dp12990
S'a'
I6468
sS's'
I17
sS'm'
(lp12991
I1
aI1
aI3
aI13
aI25
aI21
aI67
aI177
aI9
aI453
aI1171
aI65
aI2845
aI16147
aI12699
aI30905
aI122255
asS'd'
I6496
sa(dp12992
S'a'
I6475
sS's'
I17
sS'm'
(lp12993
I1
aI3
aI1
aI5
aI29
aI47
aI77
aI251
aI473
aI385
aI947
aI3239
aI5375
aI13617
aI10639
aI36005
aI95821
asS'd'
I6497
sa(dp12994
S'a'
I6486
sS's'
I17
sS'm'
(lp12995
I1
aI3
aI1
aI15
aI13
aI1
aI75
aI223
aI509
aI19
aI175
aI1541
aI637
aI5711
aI1097
aI44901
aI35277
asS'd'
I6498
sa(dp12996
S'a'
I6489
sS's'
I17
sS'm'
(lp12997
I1
aI3
aI3
aI7
aI3
aI27
aI17
aI151
aI39
aI583
aI391
aI2739
aI7339
aI2051
aI17005
aI49573
aI85969
asS'd'
I6499
sa(dp12998
S'a'
I6495
sS's'
I17
sS'm'
(lp12999
I1
aI3
aI1
aI11
aI3
aI25
aI119
aI125
aI17
aI629
aI201
aI2347
aI2923
aI1273
aI14871
aI58299
aI97667
asS'd'
I6500
sa(dp13000
S'a'
I6499
sS's'
I17
sS'm'
(lp13001
I1
aI1
aI7
aI1
aI31
aI39
aI11
aI121
aI339
aI667
aI1863
aI3479
aI1895
aI11319
aI5683
aI64969
aI9261
asS'd'
I6501
sa(dp13002
S'a'
I6505
sS's'
I17
sS'm'
(lp13003
I1
aI1
aI5
aI9
aI27
aI61
aI101
aI221
aI221
aI583
aI287
aI707
aI5931
aI4225
aI29537
aI46097
aI114361
asS'd'
I6502
sa(dp13004
S'a'
I6511
sS's'
I17
sS'm'
(lp13005
I1
aI1
aI1
aI9
aI23
aI47
aI1
aI35
aI85
aI1021
aI151
aI3153
aI3867
aI971
aI31573
aI4745
aI107639
asS'd'
I6503
sa(dp13006
S'a'
I6520
sS's'
I17
sS'm'
(lp13007
I1
aI1
aI7
aI13
aI15
aI15
aI63
aI37
aI291
aI907
aI411
aI1571
aI6415
aI7443
aI26635
aI27945
aI130909
asS'd'
I6504
sa(dp13008
S'a'
I6529
sS's'
I17
sS'm'
(lp13009
I1
aI3
aI1
aI9
aI21
aI13
aI77
aI147
aI485
aI107
aI235
aI481
aI2389
aI957
aI11493
aI53033
aI46373
asS'd'
I6505
sa(dp13010
S'a'
I6542
sS's'
I17
sS'm'
(lp13011
I1
aI3
aI5
aI7
aI3
aI55
aI125
aI237
aI205
aI411
aI1911
aI4053
aI5983
aI15489
aI29333
aI44727
aI62167
asS'd'
I6506
sa(dp13012
S'a'
I6547
sS's'
I17
sS'm'
(lp13013
I1
aI1
aI3
aI3
aI17
aI3
aI59
aI239
aI209
aI495
aI447
aI3427
aI3425
aI2347
aI10057
aI26147
aI52243
asS'd'
I6507
sa(dp13014
S'a'
I6550
sS's'
I17
sS'm'
(lp13015
I1
aI1
aI3
aI1
aI11
aI31
aI3
aI139
aI441
aI997
aI295
aI1267
aI2181
aI6047
aI32419
aI62657
aI24921
asS'd'
I6508
sa(dp13016
S'a'
I6554
sS's'
I17
sS'm'
(lp13017
I1
aI3
aI7
aI15
aI5
aI3
aI11
aI9
aI211
aI701
aI1987
aI2611
aI6195
aI14379
aI22919
aI15785
aI52149
asS'd'
I6509
sa(dp13018
S'a'
I6556
sS's'
I17
sS'm'
(lp13019
I1
aI1
aI7
aI9
aI7
aI27
aI35
aI253
aI343
aI679
aI103
aI1217
aI3983
aI8677
aI17671
aI41347
aI89355
asS'd'
I6510
sa(dp13020
S'a'
I6560
sS's'
I17
sS'm'
(lp13021
I1
aI1
aI1
aI5
aI7
aI55
aI111
aI115
aI231
aI999
aI773
aI2111
aI3617
aI2469
aI16967
aI60735
aI24557
asS'd'
I6511
sa(dp13022
S'a'
I6569
sS's'
I17
sS'm'
(lp13023
I1
aI3
aI5
aI1
aI29
aI5
aI77
aI217
aI131
aI307
aI473
aI3595
aI2713
aI6503
aI18459
aI57245
aI91897
asS'd'
I6512
sa(dp13024
S'a'
I6572
sS's'
I17
sS'm'
(lp13025
I1
aI3
aI5
aI13
aI9
aI33
aI93
aI31
aI59
aI343
aI1337
aI1971
aI7593
aI15629
aI22693
aI19885
aI4139
asS'd'
I6513
sa(dp13026
S'a'
I6590
sS's'
I17
sS'm'
(lp13027
I1
aI3
aI3
aI3
aI21
aI33
aI115
aI205
aI373
aI587
aI739
aI669
aI8065
aI5339
aI16507
aI29455
aI15863
asS'd'
I6514
sa(dp13028
S'a'
I6592
sS's'
I17
sS'm'
(lp13029
I1
aI3
aI5
aI11
aI9
aI43
aI45
aI41
aI91
aI87
aI19
aI1523
aI5059
aI9403
aI6739
aI36893
aI6395
asS'd'
I6515
sa(dp13030
S'a'
I6601
sS's'
I17
sS'm'
(lp13031
I1
aI1
aI5
aI15
aI19
aI43
aI81
aI3
aI401
aI621
aI1839
aI1443
aI179
aI8085
aI27021
aI7757
aI95011
asS'd'
I6516
sa(dp13032
S'a'
I6610
sS's'
I17
sS'm'
(lp13033
I1
aI3
aI5
aI15
aI19
aI21
aI45
aI167
aI77
aI977
aI309
aI431
aI3437
aI8327
aI12895
aI50521
aI68473
asS'd'
I6517
sa(dp13034
S'a'
I6632
sS's'
I17
sS'm'
(lp13035
I1
aI3
aI3
aI15
aI7
aI21
aI49
aI169
aI327
aI271
aI7
aI785
aI1767
aI14747
aI7083
aI65223
aI24213
asS'd'
I6518
sa(dp13036
S'a'
I6635
sS's'
I17
sS'm'
(lp13037
I1
aI1
aI5
aI9
aI9
aI51
aI101
aI197
aI507
aI839
aI1413
aI3131
aI331
aI15725
aI32293
aI60433
aI86759
asS'd'
I6519
sa(dp13038
S'a'
I6640
sS's'
I17
sS'm'
(lp13039
I1
aI1
aI7
aI1
aI17
aI39
aI127
aI201
aI341
aI607
aI1565
aI1615
aI1367
aI16043
aI28265
aI29139
aI63813
asS'd'
I6520
sa(dp13040
S'a'
I6643
sS's'
I17
sS'm'
(lp13041
I1
aI3
aI5
aI7
aI9
aI1
aI107
aI73
aI121
aI649
aI1385
aI3203
aI2897
aI8479
aI28519
aI34041
aI1359
asS'd'
I6521
sa(dp13042
S'a'
I6649
sS's'
I17
sS'm'
(lp13043
I1
aI1
aI7
aI7
aI21
aI55
aI19
aI13
aI415
aI647
aI2015
aI107
aI4167
aI5033
aI16849
aI41407
aI94387
asS'd'
I6522
sa(dp13044
S'a'
I6659
sS's'
I17
sS'm'
(lp13045
I1
aI3
aI5
aI13
aI31
aI27
aI107
aI95
aI425
aI679
aI55
aI3521
aI6737
aI11459
aI19995
aI64189
aI44323
asS'd'
I6523
sa(dp13046
S'a'
I6662
sS's'
I17
sS'm'
(lp13047
I1
aI1
aI3
aI9
aI17
aI47
aI29
aI167
aI17
aI63
aI5
aI2505
aI6483
aI14089
aI7127
aI7907
aI68555
asS'd'
I6524
sa(dp13048
S'a'
I6666
sS's'
I17
sS'm'
(lp13049
I1
aI1
aI5
aI3
aI29
aI3
aI87
aI107
aI227
aI893
aI1821
aI341
aI5481
aI13317
aI10637
aI8611
aI28625
asS'd'
I6525
sa(dp13050
S'a'
I6690
sS's'
I17
sS'm'
(lp13051
I1
aI1
aI1
aI13
aI11
aI19
aI59
aI157
aI397
aI103
aI1821
aI3913
aI3083
aI6053
aI1015
aI25475
aI94813
asS'd'
I6526
sa(dp13052
S'a'
I6692
sS's'
I17
sS'm'
(lp13053
I1
aI3
aI1
aI3
aI15
aI45
aI1
aI209
aI335
aI1015
aI539
aI2959
aI1711
aI2567
aI30169
aI147
aI25383
asS'd'
I6527
sa(dp13054
S'a'
I6704
sS's'
I17
sS'm'
(lp13055
I1
aI3
aI7
aI1
aI17
aI5
aI99
aI121
aI91
aI531
aI865
aI1667
aI5615
aI4729
aI7473
aI21445
aI37925
asS'd'
I6528
sa(dp13056
S'a'
I6713
sS's'
I17
sS'm'
(lp13057
I1
aI1
aI7
aI13
aI3
aI51
aI27
aI115
aI439
aI761
aI1121
aI1503
aI3047
aI2127
aI29253
aI48147
aI10813
asS'd'
I6529
sa(dp13058
S'a'
I6728
sS's'
I17
sS'm'
(lp13059
I1
aI3
aI7
aI15
aI1
aI51
aI33
aI161
aI509
aI159
aI1705
aI3365
aI7953
aI14027
aI3873
aI29609
aI33101
asS'd'
I6530
sa(dp13060
S'a'
I6731
sS's'
I17
sS'm'
(lp13061
I1
aI1
aI5
aI15
aI15
aI53
aI119
aI115
aI433
aI75
aI497
aI1259
aI1681
aI7715
aI24767
aI34647
aI82007
asS'd'
I6531
sa(dp13062
S'a'
I6734
sS's'
I17
sS'm'
(lp13063
I1
aI1
aI5
aI3
aI27
aI63
aI41
aI181
aI393
aI439
aI95
aI2765
aI7617
aI817
aI1311
aI18595
aI16921
asS'd'
I6532
sa(dp13064
S'a'
I6746
sS's'
I17
sS'm'
(lp13065
I1
aI3
aI1
aI15
aI31
aI7
aI57
aI89
aI371
aI745
aI475
aI3211
aI6893
aI10681
aI18547
aI28373
aI127787
asS'd'
I6533
sa(dp13066
S'a'
I6755
sS's'
I17
sS'm'
(lp13067
I1
aI3
aI5
aI13
aI5
aI55
aI45
aI167
aI147
aI833
aI765
aI1153
aI4037
aI8503
aI10751
aI49541
aI77489
asS'd'
I6534
sa(dp13068
S'a'
I6757
sS's'
I17
sS'm'
(lp13069
I1
aI3
aI1
aI11
aI11
aI7
aI45
aI167
aI431
aI759
aI1035
aI1367
aI1649
aI11711
aI4915
aI58915
aI72479
asS'd'
I6535
sa(dp13070
S'a'
I6764
sS's'
I17
sS'm'
(lp13071
I1
aI1
aI5
aI1
aI11
aI3
aI15
aI135
aI427
aI637
aI879
aI1667
aI6139
aI14759
aI25665
aI13083
aI67961
asS'd'
I6536
sa(dp13072
S'a'
I6772
sS's'
I17
sS'm'
(lp13073
I1
aI3
aI3
aI9
aI1
aI3
aI73
aI167
aI269
aI51
aI1481
aI3659
aI7863
aI7187
aI3951
aI10711
aI5909
asS'd'
I6537
sa(dp13074
S'a'
I6792
sS's'
I17
sS'm'
(lp13075
I1
aI1
aI3
aI3
aI9
aI53
aI101
aI209
aI109
aI691
aI1641
aI919
aI1083
aI6247
aI23041
aI44681
aI130105
asS'd'
I6538
sa(dp13076
S'a'
I6797
sS's'
I17
sS'm'
(lp13077
I1
aI3
aI7
aI5
aI21
aI55
aI127
aI9
aI437
aI225
aI1599
aI2575
aI5407
aI8099
aI20009
aI40339
aI110581
asS'd'
I6539
sa(dp13078
S'a'
I6821
sS's'
I17
sS'm'
(lp13079
I1
aI3
aI3
aI13
aI7
aI41
aI15
aI137
aI363
aI337
aI995
aI1215
aI3651
aI11011
aI27209
aI53927
aI78065
asS'd'
I6540
sa(dp13080
S'a'
I6822
sS's'
I17
sS'm'
(lp13081
I1
aI1
aI1
aI7
aI11
aI17
aI27
aI9
aI481
aI79
aI905
aI1297
aI811
aI10221
aI463
aI12979
aI114731
asS'd'
I6541
sa(dp13082
S'a'
I6831
sS's'
I17
sS'm'
(lp13083
I1
aI1
aI3
aI13
aI7
aI59
aI105
aI79
aI253
aI699
aI139
aI3823
aI4939
aI12955
aI32069
aI7255
aI18159
asS'd'
I6542
sa(dp13084
S'a'
I6834
sS's'
I17
sS'm'
(lp13085
I1
aI3
aI5
aI7
aI29
aI7
aI79
aI79
aI147
aI921
aI425
aI1423
aI5967
aI6397
aI17393
aI30009
aI84075
asS'd'
I6543
sa(dp13086
S'a'
I6851
sS's'
I17
sS'm'
(lp13087
I1
aI3
aI7
aI13
aI23
aI45
aI51
aI141
aI237
aI443
aI1101
aI309
aI4533
aI7479
aI22415
aI31517
aI120407
asS'd'
I6544
sa(dp13088
S'a'
I6858
sS's'
I17
sS'm'
(lp13089
I1
aI1
aI5
aI13
aI3
aI19
aI97
aI185
aI59
aI179
aI1343
aI2537
aI3165
aI16295
aI25005
aI49769
aI78007
asS'd'
I6545
sa(dp13090
S'a'
I6860
sS's'
I17
sS'm'
(lp13091
I1
aI3
aI7
aI15
aI11
aI53
aI127
aI195
aI309
aI121
aI1741
aI1415
aI225
aI15645
aI16365
aI38729
aI70061
asS'd'
I6546
sa(dp13092
S'a'
I6871
sS's'
I17
sS'm'
(lp13093
I1
aI3
aI7
aI11
aI29
aI35
aI47
aI109
aI179
aI3
aI849
aI2305
aI3035
aI15289
aI31569
aI28851
aI90057
asS'd'
I6547
sa(dp13094
S'a'
I6875
sS's'
I17
sS'm'
(lp13095
I1
aI1
aI7
aI1
aI13
aI27
aI93
aI119
aI439
aI45
aI623
aI1263
aI6595
aI6669
aI12981
aI64721
aI130109
asS'd'
I6548
sa(dp13096
S'a'
I6884
sS's'
I17
sS'm'
(lp13097
I1
aI1
aI7
aI13
aI5
aI43
aI43
aI99
aI395
aI417
aI795
aI3991
aI5601
aI13115
aI12803
aI52247
aI39245
asS'd'
I6549
sa(dp13098
S'a'
I6888
sS's'
I17
sS'm'
(lp13099
I1
aI3
aI3
aI3
aI15
aI61
aI85
aI91
aI407
aI391
aI359
aI3885
aI1925
aI4873
aI169
aI41727
aI129471
asS'd'
I6550
sa(dp13100
S'a'
I6894
sS's'
I17
sS'm'
(lp13101
I1
aI3
aI3
aI9
aI11
aI47
aI3
aI33
aI355
aI853
aI1329
aI1347
aI1995
aI8197
aI10015
aI787
aI66773
asS'd'
I6551
sa(dp13102
S'a'
I6919
sS's'
I17
sS'm'
(lp13103
I1
aI3
aI3
aI13
aI31
aI31
aI49
aI195
aI55
aI185
aI1743
aI3523
aI1781
aI8469
aI7623
aI55933
aI74953
asS'd'
I6552
sa(dp13104
S'a'
I6940
sS's'
I17
sS'm'
(lp13105
I1
aI3
aI5
aI15
aI29
aI31
aI5
aI45
aI149
aI71
aI2033
aI3171
aI4601
aI9941
aI15005
aI55709
aI74403
asS'd'
I6553
sa(dp13106
S'a'
I6950
sS's'
I17
sS'm'
(lp13107
I1
aI3
aI5
aI3
aI1
aI27
aI105
aI7
aI139
aI805
aI1877
aI915
aI1843
aI11897
aI29485
aI19275
aI44711
asS'd'
I6554
sa(dp13108
S'a'
I6959
sS's'
I17
sS'm'
(lp13109
I1
aI1
aI5
aI7
aI25
aI57
aI111
aI57
aI401
aI935
aI1685
aI2985
aI2015
aI13501
aI14581
aI53579
aI117011
asS'd'
I6555
sa(dp13110
S'a'
I6968
sS's'
I17
sS'm'
(lp13111
I1
aI1
aI5
aI11
aI13
aI47
aI63
aI137
aI145
aI77
aI1727
aI2629
aI7377
aI6311
aI537
aI13703
aI129503
asS'd'
I6556
sa(dp13112
S'a'
I6981
sS's'
I17
sS'm'
(lp13113
I1
aI1
aI7
aI9
aI5
aI49
aI67
aI51
aI163
aI989
aI845
aI7
aI2141
aI14467
aI3197
aI57581
aI121087
asS'd'
I6557
sa(dp13114
S'a'
I6988
sS's'
I17
sS'm'
(lp13115
I1
aI1
aI5
aI3
aI31
aI49
aI57
aI103
aI171
aI491
aI1109
aI1255
aI4353
aI11927
aI29525
aI16685
aI48469
asS'd'
I6558
sa(dp13116
S'a'
I6996
sS's'
I17
sS'm'
(lp13117
I1
aI1
aI1
aI3
aI7
aI29
aI17
aI111
aI339
aI747
aI763
aI179
aI7747
aI2483
aI18415
aI45301
aI25155
asS'd'
I6559
sa(dp13118
S'a'
I6999
sS's'
I17
sS'm'
(lp13119
I1
aI1
aI7
aI7
aI1
aI41
aI71
aI109
aI401
aI815
aI1311
aI3933
aI1349
aI13327
aI20847
aI44391
aI49721
asS'd'
I6560
sa(dp13120
S'a'
I7015
sS's'
I17
sS'm'
(lp13121
I1
aI1
aI1
aI15
aI27
aI57
aI39
aI129
aI391
aI701
aI619
aI3925
aI701
aI403
aI11821
aI30517
aI22035
asS'd'
I6561
sa(dp13122
S'a'
I7019
sS's'
I17
sS'm'
(lp13123
I1
aI1
aI5
aI11
aI21
aI49
aI109
aI101
aI497
aI417
aI73
aI2727
aI2899
aI2777
aI22161
aI35561
aI70211
asS'd'
I6562
sa(dp13124
S'a'
I7022
sS's'
I17
sS'm'
(lp13125
I1
aI1
aI3
aI3
aI15
aI43
aI1
aI159
aI41
aI833
aI55
aI2415
aI5009
aI9663
aI31295
aI29397
aI3187
asS'd'
I6563
sa(dp13126
S'a'
I7040
sS's'
I17
sS'm'
(lp13127
I1
aI1
aI3
aI7
aI27
aI5
aI113
aI187
aI453
aI753
aI1649
aI1605
aI2405
aI11791
aI4239
aI20915
aI54033
asS'd'
I6564
sa(dp13128
S'a'
I7045
sS's'
I17
sS'm'
(lp13129
I1
aI3
aI1
aI11
aI1
aI57
aI49
aI229
aI283
aI113
aI345
aI785
aI8009
aI11977
aI30169
aI63787
aI32011
asS'd'
I6565
sa(dp13130
S'a'
I7049
sS's'
I17
sS'm'
(lp13131
I1
aI1
aI7
aI3
aI5
aI59
aI57
aI91
aI327
aI685
aI219
aI1949
aI3095
aI8389
aI2035
aI11903
aI73461
asS'd'
I6566
sa(dp13132
S'a'
I7055
sS's'
I17
sS'm'
(lp13133
I1
aI1
aI3
aI3
aI19
aI59
aI19
aI37
aI453
aI1
aI1811
aI3263
aI1807
aI16147
aI24861
aI14003
aI31747
asS'd'
I6567
sa(dp13134
S'a'
I7073
sS's'
I17
sS'm'
(lp13135
I1
aI1
aI3
aI11
aI1
aI53
aI93
aI203
aI429
aI629
aI1931
aI1487
aI3301
aI8805
aI4901
aI2459
aI98555
asS'd'
I6568
sa(dp13136
S'a'
I7076
sS's'
I17
sS'm'
(lp13137
I1
aI1
aI7
aI5
aI21
aI5
aI37
aI135
aI159
aI749
aI1589
aI2631
aI8145
aI7279
aI28397
aI47113
aI82309
asS'd'
I6569
sa(dp13138
S'a'
I7085
sS's'
I17
sS'm'
(lp13139
I1
aI1
aI5
aI15
aI25
aI61
aI19
aI51
aI217
aI495
aI109
aI1179
aI2743
aI12107
aI12509
aI13003
aI94375
asS'd'
I6570
sa(dp13140
S'a'
I7091
sS's'
I17
sS'm'
(lp13141
I1
aI3
aI3
aI15
aI11
aI7
aI67
aI165
aI57
aI925
aI427
aI2549
aI7189
aI5917
aI13113
aI30933
aI62703
asS'd'
I6571
sa(dp13142
S'a'
I7103
sS's'
I17
sS'm'
(lp13143
I1
aI1
aI5
aI5
aI9
aI5
aI43
aI5
aI485
aI159
aI757
aI3979
aI4963
aI3389
aI29731
aI48477
aI113429
asS'd'
I6572
sa(dp13144
S'a'
I7112
sS's'
I17
sS'm'
(lp13145
I1
aI3
aI5
aI1
aI5
aI5
aI81
aI163
aI493
aI357
aI2005
aI1093
aI5951
aI1045
aI10569
aI40321
aI56881
asS'd'
I6573
sa(dp13146
S'a'
I7117
sS's'
I17
sS'm'
(lp13147
I1
aI3
aI1
aI5
aI7
aI29
aI11
aI7
aI7
aI13
aI1641
aI1031
aI4025
aI16337
aI24333
aI9589
aI37779
asS'd'
I6574
sa(dp13148
S'a'
I7118
sS's'
I17
sS'm'
(lp13149
I1
aI3
aI5
aI11
aI15
aI3
aI69
aI19
aI141
aI79
aI749
aI391
aI4505
aI6939
aI3079
aI3647
aI22363
asS'd'
I6575
sa(dp13150
S'a'
I7123
sS's'
I17
sS'm'
(lp13151
I1
aI3
aI3
aI3
aI29
aI3
aI7
aI189
aI183
aI513
aI1225
aI239
aI4203
aI9197
aI23507
aI33089
aI124433
asS'd'
I6576
sa(dp13152
S'a'
I7126
sS's'
I17
sS'm'
(lp13153
I1
aI3
aI3
aI13
aI27
aI37
aI81
aI221
aI287
aI891
aI1197
aI3501
aI539
aI2053
aI20509
aI48635
aI50269
asS'd'
I6577
sa(dp13154
S'a'
I7154
sS's'
I17
sS'm'
(lp13155
I1
aI1
aI5
aI7
aI13
aI3
aI35
aI79
aI3
aI885
aI343
aI3527
aI1043
aI7197
aI6973
aI8515
aI39315
asS'd'
I6578
sa(dp13156
S'a'
I7180
sS's'
I17
sS'm'
(lp13157
I1
aI3
aI3
aI9
aI21
aI53
aI79
aI225
aI229
aI759
aI457
aI293
aI953
aI12857
aI20483
aI3677
aI93839
asS'd'
I6579
sa(dp13158
S'a'
I7192
sS's'
I17
sS'm'
(lp13159
I1
aI3
aI5
aI3
aI5
aI17
aI45
aI107
aI153
aI279
aI761
aI1923
aI7013
aI2989
aI10137
aI19107
aI126897
asS'd'
I6580
sa(dp13160
S'a'
I7195
sS's'
I17
sS'm'
(lp13161
I1
aI3
aI1
aI3
aI23
aI53
aI91
aI1
aI133
aI729
aI13
aI2017
aI6933
aI7405
aI1255
aI49509
aI105571
asS'd'
I6581
sa(dp13162
S'a'
I7207
sS's'
I17
sS'm'
(lp13163
I1
aI3
aI5
aI1
aI9
aI45
aI35
aI153
aI209
aI289
aI1779
aI2557
aI315
aI981
aI15347
aI30391
aI16027
asS'd'
I6582
sa(dp13164
S'a'
I7208
sS's'
I17
sS'm'
(lp13165
I1
aI3
aI3
aI5
aI17
aI3
aI51
aI105
aI263
aI959
aI1255
aI1177
aI8143
aI10541
aI7687
aI38731
aI93561
asS'd'
I6583
sa(dp13166
S'a'
I7214
sS's'
I17
sS'm'
(lp13167
I1
aI1
aI1
aI13
aI19
aI1
aI15
aI135
aI447
aI847
aI663
aI3893
aI3539
aI6833
aI13265
aI62923
aI8375
asS'd'
I6584
sa(dp13168
S'a'
I7222
sS's'
I17
sS'm'
(lp13169
I1
aI3
aI1
aI15
aI31
aI11
aI105
aI1
aI91
aI523
aI1583
aI3493
aI2665
aI117
aI10757
aI29845
aI127201
asS'd'
I6585
sa(dp13170
S'a'
I7234
sS's'
I17
sS'm'
(lp13171
I1
aI1
aI1
aI3
aI29
aI49
aI9
aI103
aI309
aI605
aI1751
aI1981
aI833
aI3653
aI14001
aI16545
aI58513
asS'd'
I6586
sa(dp13172
S'a'
I7254
sS's'
I17
sS'm'
(lp13173
I1
aI1
aI5
aI9
aI1
aI19
aI117
aI71
aI237
aI765
aI249
aI1983
aI2289
aI6019
aI26505
aI31427
aI64333
asS'd'
I6587
sa(dp13174
S'a'
I7258
sS's'
I17
sS'm'
(lp13175
I1
aI1
aI3
aI11
aI15
aI31
aI5
aI207
aI347
aI143
aI11
aI1987
aI3569
aI2051
aI31051
aI22193
aI93289
asS'd'
I6588
sa(dp13176
S'a'
I7264
sS's'
I17
sS'm'
(lp13177
I1
aI1
aI3
aI5
aI13
aI15
aI5
aI73
aI457
aI611
aI673
aI2675
aI8071
aI13245
aI19443
aI14399
aI99599
asS'd'
I6589
sa(dp13178
S'a'
I7279
sS's'
I17
sS'm'
(lp13179
I1
aI1
aI1
aI9
aI11
aI5
aI103
aI231
aI31
aI457
aI1031
aI2257
aI3159
aI8323
aI31585
aI26163
aI45159
asS'd'
I6590
sa(dp13180
S'a'
I7282
sS's'
I17
sS'm'
(lp13181
I1
aI3
aI1
aI11
aI29
aI51
aI29
aI7
aI89
aI331
aI783
aI951
aI6353
aI15421
aI12801
aI8337
aI119171
asS'd'
I6591
sa(dp13182
S'a'
I7293
sS's'
I17
sS'm'
(lp13183
I1
aI1
aI3
aI13
aI23
aI57
aI63
aI43
aI505
aI1
aI657
aI4005
aI6327
aI7545
aI15455
aI27097
aI53649
asS'd'
I6592
sa(dp13184
S'a'
I7297
sS's'
I17
sS'm'
(lp13185
I1
aI1
aI1
aI5
aI31
aI7
aI51
aI107
aI175
aI461
aI1893
aI305
aI157
aI4819
aI18549
aI33087
aI9499
asS'd'
I6593
sa(dp13186
S'a'
I7322
sS's'
I17
sS'm'
(lp13187
I1
aI3
aI1
aI3
aI19
aI45
aI37
aI9
aI459
aI143
aI1327
aI3611
aI1899
aI15109
aI30151
aI17911
aI13233
asS'd'
I6594
sa(dp13188
S'a'
I7324
sS's'
I17
sS'm'
(lp13189
I1
aI1
aI5
aI15
aI19
aI49
aI11
aI227
aI375
aI661
aI665
aI259
aI3659
aI13723
aI28239
aI48159
aI59209
asS'd'
I6595
sa(dp13190
S'a'
I7351
sS's'
I17
sS'm'
(lp13191
I1
aI3
aI7
aI7
aI17
aI49
aI77
aI161
aI505
aI713
aI1521
aI935
aI3629
aI5033
aI26717
aI47199
aI3693
asS'd'
I6596
sa(dp13192
S'a'
I7363
sS's'
I17
sS'm'
(lp13193
I1
aI3
aI5
aI9
aI17
aI61
aI1
aI201
aI259
aI179
aI1637
aI2485
aI4995
aI2813
aI19923
aI43739
aI32183
asS'd'
I6597
sa(dp13194
S'a'
I7380
sS's'
I17
sS'm'
(lp13195
I1
aI1
aI3
aI5
aI1
aI23
aI125
aI61
aI225
aI703
aI2011
aI1013
aI6651
aI14029
aI27375
aI23301
aI80269
asS'd'
I6598
sa(dp13196
S'a'
I7384
sS's'
I17
sS'm'
(lp13197
I1
aI1
aI3
aI9
aI11
aI57
aI37
aI49
aI321
aI443
aI1055
aI1989
aI4755
aI8467
aI22001
aI18647
aI112617
asS'd'
I6599
sa(dp13198
S'a'
I7389
sS's'
I17
sS'm'
(lp13199
I1
aI3
aI1
aI5
aI5
aI39
aI21
aI139
aI101
aI583
aI1881
aI2599
aI4185
aI15679
aI22215
aI19093
aI76737
asS'd'
I6600
sa(dp13200
S'a'
I7396
sS's'
I17
sS'm'
(lp13201
I1
aI3
aI1
aI11
aI31
aI51
aI85
aI91
aI159
aI421
aI2005
aI1075
aI7757
aI12653
aI25489
aI3545
aI62961
asS'd'
I6601
sa(dp13202
S'a'
I7413
sS's'
I17
sS'm'
(lp13203
I1
aI1
aI1
aI15
aI27
aI57
aI75
aI151
aI357
aI571
aI395
aI299
aI5607
aI12865
aI2149
aI21059
aI120753
asS'd'
I6602
sa(dp13204
S'a'
I7417
sS's'
I17
sS'm'
(lp13205
I1
aI1
aI1
aI3
aI15
aI57
aI63
aI171
aI265
aI709
aI1089
aI677
aI7243
aI10207
aI9789
aI38431
aI130415
asS'd'
I6603
sa(dp13206
S'a'
I7431
sS's'
I17
sS'm'
(lp13207
I1
aI3
aI7
aI5
aI21
aI9
aI73
aI149
aI197
aI773
aI773
aI3931
aI4135
aI5671
aI2733
aI57173
aI90693
asS'd'
I6604
sa(dp13208
S'a'
I7443
sS's'
I17
sS'm'
(lp13209
I1
aI1
aI5
aI1
aI23
aI1
aI47
aI201
aI33
aI167
aI1643
aI4009
aI2687
aI5725
aI28733
aI27859
aI55163
asS'd'
I6605
sa(dp13210
S'a'
I7445
sS's'
I17
sS'm'
(lp13211
I1
aI1
aI5
aI15
aI25
aI11
aI57
aI139
aI471
aI625
aI1067
aI3647
aI6213
aI15605
aI23537
aI5005
aI32593
asS'd'
I6606
sa(dp13212
S'a'
I7450
sS's'
I17
sS'm'
(lp13213
I1
aI3
aI1
aI11
aI17
aI11
aI25
aI163
aI199
aI21
aI1775
aI3721
aI2845
aI15769
aI2381
aI27643
aI19909
asS'd'
I6607
sa(dp13214
S'a'
I7456
sS's'
I17
sS'm'
(lp13215
I1
aI3
aI5
aI5
aI21
aI41
aI23
aI125
aI401
aI483
aI535
aI925
aI7065
aI1727
aI3761
aI8485
aI3519
asS'd'
I6608
sa(dp13216
S'a'
I7466
sS's'
I17
sS'm'
(lp13217
I1
aI1
aI3
aI15
aI27
aI31
aI11
aI7
aI93
aI237
aI611
aI3635
aI4747
aI9751
aI20607
aI20473
aI73935
asS'd'
I6609
sa(dp13218
S'a'
I7468
sS's'
I17
sS'm'
(lp13219
I1
aI1
aI7
aI3
aI15
aI19
aI69
aI169
aI387
aI291
aI1981
aI635
aI3387
aI15817
aI20357
aI47537
aI107311
asS'd'
I6610
sa(dp13220
S'a'
I7474
sS's'
I17
sS'm'
(lp13221
I1
aI3
aI7
aI15
aI13
aI59
aI31
aI235
aI399
aI343
aI1265
aI2975
aI6839
aI13335
aI5397
aI58915
aI31313
asS'd'
I6611
sa(dp13222
S'a'
I7479
sS's'
I17
sS'm'
(lp13223
I1
aI1
aI7
aI1
aI3
aI35
aI81
aI243
aI387
aI421
aI1533
aI799
aI2615
aI13219
aI9041
aI26967
aI22677
asS'd'
I6612
sa(dp13224
S'a'
I7486
sS's'
I17
sS'm'
(lp13225
I1
aI1
aI7
aI15
aI17
aI41
aI89
aI115
aI67
aI569
aI1647
aI1831
aI5533
aI4629
aI1413
aI20037
aI97343
asS'd'
I6613
sa(dp13226
S'a'
I7497
sS's'
I17
sS'm'
(lp13227
I1
aI1
aI5
aI1
aI23
aI41
aI11
aI149
aI319
aI377
aI439
aI1237
aI4819
aI14303
aI14657
aI61711
aI129235
asS'd'
I6614
sa(dp13228
S'a'
I7508
sS's'
I17
sS'm'
(lp13229
I1
aI3
aI3
aI7
aI9
aI11
aI79
aI219
aI249
aI607
aI1453
aI2931
aI3407
aI13725
aI28289
aI42869
aI96759
asS'd'
I6615
sa(dp13230
S'a'
I7515
sS's'
I17
sS'm'
(lp13231
I1
aI1
aI5
aI11
aI7
aI9
aI101
aI51
aI11
aI893
aI697
aI1221
aI4237
aI1873
aI11191
aI25517
aI119861
asS'd'
I6616
sa(dp13232
S'a'
I7533
sS's'
I17
sS'm'
(lp13233
I1
aI1
aI3
aI11
aI23
aI23
aI19
aI245
aI485
aI317
aI1945
aI2339
aI193
aI9389
aI30709
aI33927
aI95089
asS'd'
I6617
sa(dp13234
S'a'
I7542
sS's'
I17
sS'm'
(lp13235
I1
aI1
aI3
aI1
aI27
aI55
aI5
aI81
aI63
aI185
aI223
aI3639
aI6093
aI10053
aI1793
aI11885
aI29307
asS'd'
I6618
sa(dp13236
S'a'
I7546
sS's'
I17
sS'm'
(lp13237
I1
aI1
aI7
aI13
aI15
aI41
aI33
aI133
aI467
aI457
aI213
aI3687
aI1313
aI2555
aI19487
aI44257
aI108667
asS'd'
I6619
sa(dp13238
S'a'
I7551
sS's'
I17
sS'm'
(lp13239
I1
aI1
aI3
aI5
aI31
aI51
aI53
aI115
aI449
aI273
aI1043
aI2743
aI1759
aI2013
aI28171
aI57091
aI76837
asS'd'
I6620
sa(dp13240
S'a'
I7569
sS's'
I17
sS'm'
(lp13241
I1
aI1
aI5
aI15
aI21
aI43
aI11
aI215
aI151
aI253
aI913
aI1889
aI2799
aI13787
aI3869
aI54413
aI50991
asS'd'
I6621
sa(dp13242
S'a'
I7572
sS's'
I17
sS'm'
(lp13243
I1
aI1
aI3
aI13
aI29
aI19
aI81
aI123
aI461
aI203
aI81
aI555
aI6601
aI15689
aI12637
aI41467
aI105343
asS'd'
I6622
sa(dp13244
S'a'
I7595
sS's'
I17
sS'm'
(lp13245
I1
aI1
aI3
aI13
aI7
aI21
aI75
aI111
aI47
aI481
aI1519
aI3299
aI6199
aI3501
aI31323
aI29215
aI45607
asS'd'
I6623
sa(dp13246
S'a'
I7603
sS's'
I17
sS'm'
(lp13247
I1
aI3
aI1
aI3
aI17
aI51
aI45
aI223
aI321
aI233
aI267
aI3333
aI3803
aI3099
aI4601
aI29061
aI88441
asS'd'
I6624
sa(dp13248
S'a'
I7605
sS's'
I17
sS'm'
(lp13249
I1
aI1
aI5
aI13
aI23
aI27
aI7
aI57
aI273
aI893
aI773
aI239
aI6357
aI15875
aI5497
aI21775
aI108291
asS'd'
I6625
sa(dp13250
S'a'
I7629
sS's'
I17
sS'm'
(lp13251
I1
aI3
aI1
aI15
aI25
aI17
aI11
aI229
aI175
aI909
aI691
aI3507
aI5247
aI2933
aI1741
aI35059
aI62841
asS'd'
I6626
sa(dp13252
S'a'
I7632
sS's'
I17
sS'm'
(lp13253
I1
aI3
aI5
aI1
aI29
aI7
aI11
aI69
aI345
aI87
aI99
aI3243
aI5669
aI11053
aI1185
aI6979
aI117069
asS'd'
I6627
sa(dp13254
S'a'
I7638
sS's'
I17
sS'm'
(lp13255
I1
aI3
aI5
aI11
aI13
aI33
aI23
aI183
aI89
aI475
aI643
aI2773
aI7899
aI15219
aI133
aI5073
aI129355
asS'd'
I6628
sa(dp13256
S'a'
I7648
sS's'
I17
sS'm'
(lp13257
I1
aI3
aI7
aI9
aI23
aI17
aI31
aI53
aI455
aI193
aI1695
aI2557
aI1645
aI12675
aI27857
aI50447
aI121335
asS'd'
I6629
sa(dp13258
S'a'
I7654
sS's'
I17
sS'm'
(lp13259
I1
aI1
aI3
aI11
aI15
aI19
aI41
aI57
aI305
aI235
aI1131
aI1165
aI1857
aI13667
aI19285
aI29755
aI118885
asS'd'
I6630
sa(dp13260
S'a'
I7663
sS's'
I17
sS'm'
(lp13261
I1
aI3
aI7
aI3
aI9
aI43
aI107
aI3
aI275
aI673
aI677
aI3769
aI3097
aI5497
aI24911
aI4617
aI80505
asS'd'
I6631
sa(dp13262
S'a'
I7675
sS's'
I17
sS'm'
(lp13263
I1
aI1
aI7
aI9
aI21
aI39
aI107
aI155
aI471
aI753
aI579
aI2929
aI4951
aI4245
aI25035
aI41795
aI86955
asS'd'
I6632
sa(dp13264
S'a'
I7693
sS's'
I17
sS'm'
(lp13265
I1
aI3
aI1
aI7
aI31
aI51
aI27
aI165
aI147
aI751
aI709
aI399
aI45
aI947
aI9893
aI32721
aI122127
asS'd'
I6633
sa(dp13266
S'a'
I7705
sS's'
I17
sS'm'
(lp13267
I1
aI3
aI3
aI1
aI31
aI31
aI73
aI59
aI351
aI293
aI845
aI3139
aI4177
aI3537
aI9465
aI20689
aI65837
asS'd'
I6634
sa(dp13268
S'a'
I7717
sS's'
I17
sS'm'
(lp13269
I1
aI3
aI5
aI9
aI27
aI29
aI13
aI115
aI417
aI435
aI465
aI1291
aI5225
aI11687
aI29207
aI39895
aI55443
asS'd'
I6635
sa(dp13270
S'a'
I7724
sS's'
I17
sS'm'
(lp13271
I1
aI3
aI3
aI15
aI29
aI49
aI111
aI179
aI221
aI565
aI787
aI1811
aI4055
aI7863
aI27273
aI32975
aI26985
asS'd'
I6636
sa(dp13272
S'a'
I7727
sS's'
I17
sS'm'
(lp13273
I1
aI1
aI1
aI7
aI15
aI49
aI121
aI145
aI277
aI27
aI149
aI965
aI4903
aI3497
aI32333
aI37217
aI105073
asS'd'
I6637
sa(dp13274
S'a'
I7735
sS's'
I17
sS'm'
(lp13275
I1
aI1
aI7
aI1
aI23
aI29
aI31
aI77
aI353
aI349
aI755
aI2081
aI4291
aI567
aI641
aI41751
aI20397
asS'd'
I6638
sa(dp13276
S'a'
I7761
sS's'
I17
sS'm'
(lp13277
I1
aI1
aI5
aI3
aI25
aI31
aI97
aI3
aI405
aI607
aI965
aI2981
aI3217
aI14695
aI25977
aI22457
aI113539
asS'd'
I6639
sa(dp13278
S'a'
I7767
sS's'
I17
sS'm'
(lp13279
I1
aI3
aI3
aI15
aI25
aI3
aI91
aI125
aI269
aI825
aI1163
aI2181
aI4247
aI6813
aI4699
aI35091
aI87771
asS'd'
I6640
sa(dp13280
S'a'
I7783
sS's'
I17
sS'm'
(lp13281
I1
aI1
aI5
aI9
aI25
aI23
aI113
aI145
aI71
aI31
aI1115
aI3729
aI6793
aI11869
aI26509
aI18779
aI99499
asS'd'
I6641
sa(dp13282
S'a'
I7784
sS's'
I17
sS'm'
(lp13283
I1
aI1
aI1
aI9
aI31
aI51
aI77
aI217
aI247
aI599
aI1541
aI3217
aI1383
aI5203
aI27971
aI23647
aI71823
asS'd'
I6642
sa(dp13284
S'a'
I7798
sS's'
I17
sS'm'
(lp13285
I1
aI1
aI5
aI7
aI17
aI35
aI113
aI73
aI475
aI511
aI35
aI1961
aI5311
aI2257
aI1935
aI58963
aI94349
asS'd'
I6643
sa(dp13286
S'a'
I7802
sS's'
I17
sS'm'
(lp13287
I1
aI3
aI1
aI7
aI27
aI31
aI67
aI253
aI95
aI883
aI1213
aI855
aI3429
aI15049
aI26715
aI56099
aI101797
asS'd'
I6644
sa(dp13288
S'a'
I7811
sS's'
I17
sS'm'
(lp13289
I1
aI1
aI3
aI5
aI9
aI9
aI61
aI57
aI511
aI537
aI1803
aI949
aI1327
aI3921
aI11297
aI13807
aI64629
asS'd'
I6645
sa(dp13290
S'a'
I7817
sS's'
I17
sS'm'
(lp13291
I1
aI1
aI5
aI1
aI31
aI57
aI105
aI161
aI309
aI283
aI1291
aI2737
aI7141
aI7497
aI25893
aI14453
aI35375
asS'd'
I6646
sa(dp13292
S'a'
I7823
sS's'
I17
sS'm'
(lp13293
I1
aI1
aI3
aI1
aI21
aI3
aI77
aI37
aI13
aI211
aI1863
aI1895
aI8035
aI5801
aI25981
aI12317
aI48375
asS'd'
I6647
sa(dp13294
S'a'
I7832
sS's'
I17
sS'm'
(lp13295
I1
aI3
aI7
aI7
aI25
aI45
aI13
aI77
aI185
aI553
aI1501
aI1349
aI5951
aI15581
aI32657
aI18467
aI128363
asS'd'
I6648
sa(dp13296
S'a'
I7837
sS's'
I17
sS'm'
(lp13297
I1
aI3
aI5
aI9
aI23
aI63
aI105
aI239
aI213
aI935
aI1331
aI3653
aI2775
aI6591
aI6067
aI34597
aI19217
asS'd'
I6649
sa(dp13298
S'a'
I7842
sS's'
I17
sS'm'
(lp13299
I1
aI3
aI7
aI13
aI15
aI19
aI79
aI91
aI391
aI637
aI1685
aI2263
aI3507
aI2025
aI2111
aI15875
aI14831
asS'd'
I6650
sa(dp13300
S'a'
I7853
sS's'
I17
sS'm'
(lp13301
I1
aI3
aI3
aI5
aI7
aI29
aI81
aI69
aI511
aI399
aI343
aI737
aI2833
aI1021
aI10471
aI18689
aI36181
asS'd'
I6651
sa(dp13302
S'a'
I7854
sS's'
I17
sS'm'
(lp13303
I1
aI1
aI5
aI11
aI21
aI17
aI39
aI137
aI145
aI857
aI583
aI789
aI8057
aI15995
aI32113
aI64163
aI37153
asS'd'
I6652
sa(dp13304
S'a'
I7856
sS's'
I17
sS'm'
(lp13305
I1
aI3
aI3
aI11
aI9
aI61
aI87
aI131
aI487
aI667
aI1295
aI493
aI4629
aI7719
aI18157
aI49715
aI2051
asS'd'
I6653
sa(dp13306
S'a'
I7861
sS's'
I17
sS'm'
(lp13307
I1
aI3
aI5
aI9
aI19
aI5
aI85
aI3
aI491
aI353
aI571
aI2829
aI4411
aI343
aI24781
aI62325
aI123959
asS'd'
I6654
sa(dp13308
S'a'
I7862
sS's'
I17
sS'm'
(lp13309
I1
aI1
aI7
aI13
aI13
aI39
aI11
aI31
aI413
aI285
aI27
aI2433
aI3307
aI6165
aI26565
aI40065
aI102655
asS'd'
I6655
sa(dp13310
S'a'
I7873
sS's'
I17
sS'm'
(lp13311
I1
aI1
aI5
aI11
aI25
aI45
aI7
aI97
aI9
aI973
aI1833
aI2537
aI1457
aI7389
aI24087
aI38061
aI122805
asS'd'
I6656
sa(dp13312
S'a'
I7874
sS's'
I17
sS'm'
(lp13313
I1
aI3
aI5
aI3
aI21
aI63
aI77
aI21
aI249
aI525
aI1145
aI1421
aI8011
aI3357
aI15051
aI30293
aI127017
asS'd'
I6657
sa(dp13314
S'a'
I7886
sS's'
I17
sS'm'
(lp13315
I1
aI1
aI5
aI3
aI13
aI53
aI81
aI185
aI303
aI307
aI1579
aI841
aI2277
aI607
aI10899
aI34209
aI215
asS'd'
I6658
sa(dp13316
S'a'
I7914
sS's'
I17
sS'm'
(lp13317
I1
aI3
aI3
aI13
aI17
aI1
aI125
aI145
aI205
aI763
aI127
aI1865
aI4129
aI849
aI27247
aI29845
aI36515
asS'd'
I6659
sa(dp13318
S'a'
I7927
sS's'
I17
sS'm'
(lp13319
I1
aI3
aI7
aI13
aI5
aI59
aI19
aI71
aI227
aI111
aI365
aI1309
aI5857
aI6035
aI32379
aI11303
aI127329
asS'd'
I6660
sa(dp13320
S'a'
I7936
sS's'
I17
sS'm'
(lp13321
I1
aI1
aI1
aI1
aI19
aI61
aI79
aI253
aI459
aI23
aI725
aI3615
aI4583
aI429
aI13215
aI31879
aI4523
asS'd'
I6661
sa(dp13322
S'a'
I7951
sS's'
I17
sS'm'
(lp13323
I1
aI1
aI1
aI7
aI19
aI13
aI53
aI41
aI243
aI107
aI1767
aI983
aI3483
aI2249
aI2209
aI58895
aI14805
asS'd'
I6662
sa(dp13324
S'a'
I7963
sS's'
I17
sS'm'
(lp13325
I1
aI1
aI1
aI9
aI5
aI63
aI31
aI85
aI119
aI307
aI633
aI3295
aI841
aI3495
aI22965
aI57587
aI108271
asS'd'
I6663
sa(dp13326
S'a'
I7966
sS's'
I17
sS'm'
(lp13327
I1
aI3
aI5
aI9
aI17
aI13
aI57
aI49
aI97
aI613
aI405
aI2637
aI3229
aI14253
aI4663
aI61345
aI33415
asS'd'
I6664
sa(dp13328
S'a'
I7976
sS's'
I17
sS'm'
(lp13329
I1
aI3
aI1
aI1
aI17
aI37
aI63
aI3
aI5
aI375
aI1073
aI3971
aI665
aI4445
aI153
aI20437
aI38513
asS'd'
I6665
sa(dp13330
S'a'
I7993
sS's'
I17
sS'm'
(lp13331
I1
aI3
aI3
aI15
aI5
aI9
aI77
aI161
aI409
aI461
aI443
aI567
aI5581
aI8623
aI27735
aI9041
aI5517
asS'd'
I6666
sa(dp13332
S'a'
I8001
sS's'
I17
sS'm'
(lp13333
I1
aI3
aI5
aI13
aI13
aI5
aI19
aI53
aI263
aI155
aI557
aI3973
aI6841
aI4829
aI30751
aI30025
aI121973
asS'd'
I6667
sa(dp13334
S'a'
I8004
sS's'
I17
sS'm'
(lp13335
I1
aI3
aI7
aI9
aI27
aI37
aI49
aI243
aI107
aI1013
aI1743
aI1509
aI4465
aI15415
aI4741
aI41409
aI72695
asS'd'
I6668
sa(dp13336
S'a'
I8013
sS's'
I17
sS'm'
(lp13337
I1
aI1
aI3
aI5
aI11
aI49
aI39
aI45
aI21
aI463
aI875
aI3681
aI1901
aI15325
aI24553
aI51369
aI82227
asS'd'
I6669
sa(dp13338
S'a'
I8021
sS's'
I17
sS'm'
(lp13339
I1
aI1
aI3
aI15
aI11
aI35
aI21
aI91
aI383
aI149
aI1815
aI911
aI4633
aI12027
aI12413
aI22307
aI120049
asS'd'
I6670
sa(dp13340
S'a'
I8026
sS's'
I17
sS'm'
(lp13341
I1
aI3
aI5
aI7
aI7
aI3
aI15
aI83
aI477
aI687
aI145
aI1705
aI6893
aI5233
aI20171
aI43337
aI72603
asS'd'
I6671
sa(dp13342
S'a'
I8028
sS's'
I17
sS'm'
(lp13343
I1
aI1
aI3
aI9
aI25
aI35
aI19
aI173
aI67
aI5
aI561
aI2139
aI4557
aI4911
aI26273
aI38409
aI22801
asS'd'
I6672
sa(dp13344
S'a'
I8031
sS's'
I17
sS'm'
(lp13345
I1
aI1
aI3
aI13
aI15
aI39
aI85
aI91
aI91
aI187
aI1851
aI1181
aI4049
aI16353
aI26525
aI43703
aI19415
asS'd'
I6673
sa(dp13346
S'a'
I8035
sS's'
I17
sS'm'
(lp13347
I1
aI3
aI1
aI9
aI13
aI41
aI77
aI179
aI415
aI705
aI693
aI3017
aI5847
aI16191
aI11435
aI28979
aI51839
asS'd'
I6674
sa(dp13348
S'a'
I8042
sS's'
I17
sS'm'
(lp13349
I1
aI1
aI3
aI5
aI23
aI15
aI3
aI159
aI269
aI67
aI625
aI4043
aI4701
aI1599
aI6467
aI10949
aI80073
asS'd'
I6675
sa(dp13350
S'a'
I8071
sS's'
I17
sS'm'
(lp13351
I1
aI3
aI3
aI15
aI7
aI43
aI81
aI157
aI393
aI321
aI1875
aI2801
aI4359
aI11703
aI1063
aI64015
aI109997
asS'd'
I6676
sa(dp13352
S'a'
I8085
sS's'
I17
sS'm'
(lp13353
I1
aI1
aI7
aI3
aI25
aI21
aI37
aI123
aI133
aI691
aI973
aI3115
aI2291
aI10519
aI13339
aI22751
aI85445
asS'd'
I6677
sa(dp13354
S'a'
I8092
sS's'
I17
sS'm'
(lp13355
I1
aI3
aI1
aI1
aI21
aI21
aI9
aI23
aI431
aI679
aI1873
aI289
aI4503
aI3939
aI14417
aI36081
aI18709
asS'd'
I6678
sa(dp13356
S'a'
I8102
sS's'
I17
sS'm'
(lp13357
I1
aI3
aI5
aI5
aI1
aI53
aI109
aI133
aI33
aI279
aI727
aI2233
aI3065
aI8557
aI7487
aI25797
aI109177
asS'd'
I6679
sa(dp13358
S'a'
I8105
sS's'
I17
sS'm'
(lp13359
I1
aI1
aI7
aI7
aI1
aI9
aI47
aI127
aI179
aI757
aI1985
aI547
aI169
aI13393
aI22669
aI58795
aI92897
asS'd'
I6680
sa(dp13360
S'a'
I8114
sS's'
I17
sS'm'
(lp13361
I1
aI3
aI5
aI11
aI17
aI21
aI95
aI219
aI263
aI579
aI1493
aI3283
aI5461
aI1235
aI1749
aI33325
aI36033
asS'd'
I6681
sa(dp13362
S'a'
I8123
sS's'
I17
sS'm'
(lp13363
I1
aI1
aI3
aI11
aI21
aI49
aI45
aI143
aI511
aI983
aI1933
aI965
aI7905
aI1925
aI27857
aI40723
aI68251
asS'd'
I6682
sa(dp13364
S'a'
I8131
sS's'
I17
sS'm'
(lp13365
I1
aI3
aI7
aI3
aI27
aI9
aI73
aI7
aI441
aI877
aI107
aI1599
aI4795
aI7251
aI6819
aI7671
aI21137
asS'd'
I6683
sa(dp13366
S'a'
I8140
sS's'
I17
sS'm'
(lp13367
I1
aI1
aI3
aI3
aI21
aI25
aI49
aI43
aI247
aI949
aI627
aI2859
aI2507
aI4787
aI11269
aI53221
aI126387
asS'd'
I6684
sa(dp13368
S'a'
I8145
sS's'
I17
sS'm'
(lp13369
I1
aI1
aI5
aI3
aI5
aI53
aI127
aI65
aI353
aI521
aI1701
aI2981
aI3201
aI611
aI13475
aI58015
aI2605
asS'd'
I6685
sa(dp13370
S'a'
I8157
sS's'
I17
sS'm'
(lp13371
I1
aI1
aI5
aI13
aI9
aI39
aI55
aI103
aI53
aI281
aI705
aI2433
aI6179
aI3381
aI31973
aI30267
aI91307
asS'd'
I6686
sa(dp13372
S'a'
I8158
sS's'
I17
sS'm'
(lp13373
I1
aI1
aI7
aI13
aI31
aI23
aI29
aI161
aI347
aI449
aI123
aI3427
aI5731
aI12691
aI15175
aI20487
aI74695
asS'd'
I6687
sa(dp13374
S'a'
I8185
sS's'
I17
sS'm'
(lp13375
I1
aI3
aI3
aI15
aI13
aI19
aI83
aI137
aI437
aI317
aI921
aI913
aI7979
aI6665
aI5313
aI1435
aI60271
asS'd'
I6688
sa(dp13376
S'a'
I8186
sS's'
I17
sS'm'
(lp13377
I1
aI3
aI5
aI7
aI19
aI23
aI31
aI131
aI421
aI95
aI1999
aI897
aI4839
aI1815
aI12387
aI45009
aI2609
asS'd'
I6689
sa(dp13378
S'a'
I8188
sS's'
I17
sS'm'
(lp13379
I1
aI1
aI1
aI7
aI3
aI53
aI121
aI33
aI47
aI283
aI813
aI3841
aI4449
aI2543
aI15211
aI59285
aI42551
asS'd'
I6690
sa(dp13380
S'a'
I8192
sS's'
I17
sS'm'
(lp13381
I1
aI3
aI1
aI13
aI9
aI43
aI37
aI167
aI93
aI417
aI213
aI2721
aI3395
aI2089
aI13743
aI32925
aI91147
asS'd'
I6691
sa(dp13382
S'a'
I8212
sS's'
I17
sS'm'
(lp13383
I1
aI3
aI7
aI5
aI31
aI25
aI97
aI25
aI19
aI11
aI543
aI1889
aI455
aI5765
aI9517
aI56963
aI131069
asS'd'
I6692
sa(dp13384
S'a'
I8219
sS's'
I17
sS'm'
(lp13385
I1
aI3
aI1
aI7
aI3
aI7
aI87
aI61
aI209
aI39
aI1303
aI1637
aI6687
aI8001
aI5003
aI47911
aI110657
asS'd'
I6693
sa(dp13386
S'a'
I8221
sS's'
I17
sS'm'
(lp13387
I1
aI1
aI5
aI3
aI11
aI25
aI99
aI77
aI379
aI843
aI1423
aI2933
aI7535
aI4181
aI32223
aI49327
aI48041
asS'd'
I6694
sa(dp13388
S'a'
I8235
sS's'
I17
sS'm'
(lp13389
I1
aI3
aI3
aI13
aI9
aI7
aI85
aI59
aI47
aI777
aI401
aI2449
aI2795
aI11289
aI25023
aI7725
aI73887
asS'd'
I6695
sa(dp13390
S'a'
I8237
sS's'
I17
sS'm'
(lp13391
I1
aI1
aI3
aI5
aI11
aI51
aI93
aI57
aI369
aI871
aI1175
aI2705
aI1253
aI5169
aI24691
aI14243
aI119667
asS'd'
I6696
sa(dp13392
S'a'
I8249
sS's'
I17
sS'm'
(lp13393
I1
aI3
aI1
aI3
aI5
aI7
aI33
aI171
aI359
aI115
aI1909
aI2003
aI1413
aI13829
aI3471
aI36185
aI118399
asS'd'
I6697
sa(dp13394
S'a'
I8260
sS's'
I17
sS'm'
(lp13395
I1
aI1
aI1
aI11
aI5
aI49
aI97
aI145
aI415
aI731
aI671
aI2309
aI7211
aI11359
aI22757
aI15415
aI70951
asS'd'
I6698
sa(dp13396
S'a'
I8264
sS's'
I17
sS'm'
(lp13397
I1
aI1
aI3
aI5
aI7
aI51
aI61
aI101
aI375
aI575
aI1321
aI2835
aI7569
aI9599
aI4707
aI7655
aI53417
asS'd'
I6699
sa(dp13398
S'a'
I8270
sS's'
I17
sS'm'
(lp13399
I1
aI3
aI1
aI15
aI9
aI63
aI25
aI117
aI203
aI5
aI1345
aI2571
aI5273
aI2059
aI4689
aI27237
aI23199
asS'd'
I6700
sa(dp13400
S'a'
I8282
sS's'
I17
sS'm'
(lp13401
I1
aI1
aI3
aI15
aI15
aI23
aI69
aI49
aI349
aI995
aI5
aI1565
aI903
aI10165
aI9565
aI6343
aI16653
asS'd'
I6701
sa(dp13402
S'a'
I8291
sS's'
I17
sS'm'
(lp13403
I1
aI1
aI3
aI9
aI21
aI15
aI69
aI9
aI463
aI69
aI1447
aI2347
aI5125
aI7479
aI18257
aI14405
aI51321
asS'd'
I6702
sa(dp13404
S'a'
I8293
sS's'
I17
sS'm'
(lp13405
I1
aI1
aI7
aI11
aI23
aI57
aI57
aI179
aI17
aI129
aI999
aI777
aI6281
aI1693
aI31885
aI31085
aI29237
asS'd'
I6703
sa(dp13406
S'a'
I8297
sS's'
I17
sS'm'
(lp13407
I1
aI3
aI5
aI1
aI25
aI55
aI15
aI21
aI199
aI271
aI1645
aI1719
aI2023
aI10049
aI15215
aI11959
aI44875
asS'd'
I6704
sa(dp13408
S'a'
I8298
sS's'
I17
sS'm'
(lp13409
I1
aI3
aI1
aI3
aI29
aI43
aI83
aI11
aI281
aI27
aI429
aI685
aI7189
aI9151
aI8665
aI9553
aI115293
asS'd'
I6705
sa(dp13410
S'a'
I8305
sS's'
I17
sS'm'
(lp13411
I1
aI3
aI1
aI7
aI17
aI43
aI125
aI11
aI189
aI803
aI713
aI683
aI7285
aI4455
aI18195
aI45333
aI32281
asS'd'
I6706
sa(dp13412
S'a'
I8306
sS's'
I17
sS'm'
(lp13413
I1
aI3
aI3
aI3
aI11
aI55
aI21
aI59
aI173
aI283
aI709
aI1561
aI5391
aI5097
aI24725
aI19217
aI13769
asS'd'
I6707
sa(dp13414
S'a'
I8311
sS's'
I17
sS'm'
(lp13415
I1
aI3
aI5
aI13
aI7
aI29
aI117
aI207
aI415
aI525
aI567
aI1741
aI3553
aI6729
aI433
aI17619
aI45971
asS'd'
I6708
sa(dp13416
S'a'
I8318
sS's'
I17
sS'm'
(lp13417
I1
aI1
aI7
aI7
aI3
aI23
aI43
aI43
aI213
aI823
aI609
aI1037
aI3797
aI4733
aI30717
aI61067
aI89581
asS'd'
I6709
sa(dp13418
S'a'
I8327
sS's'
I17
sS'm'
(lp13419
I1
aI3
aI5
aI7
aI11
aI7
aI7
aI241
aI379
aI217
aI739
aI2815
aI2549
aI14297
aI10283
aI1509
aI80613
asS'd'
I6710
sa(dp13420
S'a'
I8345
sS's'
I17
sS'm'
(lp13421
I1
aI1
aI1
aI1
aI17
aI45
aI53
aI229
aI193
aI893
aI1881
aI227
aI6751
aI7135
aI20823
aI36939
aI27667
asS'd'
I6711
sa(dp13422
S'a'
I8379
sS's'
I17
sS'm'
(lp13423
I1
aI3
aI3
aI1
aI15
aI39
aI27
aI217
aI101
aI949
aI1963
aI2213
aI2357
aI4129
aI11925
aI841
aI59259
asS'd'
I6712
sa(dp13424
S'a'
I8390
sS's'
I17
sS'm'
(lp13425
I1
aI1
aI3
aI3
aI5
aI53
aI59
aI255
aI421
aI1009
aI683
aI2171
aI6691
aI12489
aI20865
aI29363
aI70611
asS'd'
I6713
sa(dp13426
S'a'
I8394
sS's'
I17
sS'm'
(lp13427
I1
aI1
aI7
aI15
aI7
aI31
aI105
aI141
aI153
aI401
aI549
aI3045
aI5443
aI11147
aI18159
aI24283
aI21859
asS'd'
I6714
sa(dp13428
S'a'
I8414
sS's'
I17
sS'm'
(lp13429
I1
aI3
aI7
aI1
aI11
aI17
aI17
aI231
aI175
aI603
aI1915
aI111
aI3203
aI10627
aI9687
aI47235
aI87057
asS'd'
I6715
sa(dp13430
S'a'
I8417
sS's'
I17
sS'm'
(lp13431
I1
aI1
aI1
aI11
aI19
aI21
aI115
aI41
aI45
aI727
aI1523
aI739
aI3025
aI10321
aI27353
aI63139
aI16051
asS'd'
I6716
sa(dp13432
S'a'
I8432
sS's'
I17
sS'm'
(lp13433
I1
aI3
aI7
aI11
aI13
aI9
aI33
aI121
aI237
aI565
aI2043
aI2131
aI3079
aI12575
aI2187
aI14427
aI85939
asS'd'
I6717
sa(dp13434
S'a'
I8437
sS's'
I17
sS'm'
(lp13435
I1
aI3
aI1
aI15
aI21
aI19
aI91
aI227
aI485
aI49
aI101
aI15
aI1903
aI4039
aI23819
aI40001
aI66405
asS'd'
I6718
sa(dp13436
S'a'
I8441
sS's'
I17
sS'm'
(lp13437
I1
aI3
aI1
aI5
aI15
aI25
aI65
aI25
aI393
aI287
aI1435
aI1851
aI6437
aI5983
aI13769
aI37847
aI120907
asS'd'
I6719
sa(dp13438
S'a'
I8449
sS's'
I17
sS'm'
(lp13439
I1
aI3
aI7
aI15
aI15
aI21
aI97
aI37
aI359
aI155
aI807
aI1421
aI517
aI13135
aI2955
aI56979
aI52299
asS'd'
I6720
sa(dp13440
S'a'
I8456
sS's'
I17
sS'm'
(lp13441
I1
aI1
aI5
aI1
aI27
aI53
aI79
aI27
aI467
aI605
aI267
aI1193
aI31
aI6177
aI12369
aI32621
aI38319
asS'd'
I6721
sa(dp13442
S'a'
I8473
sS's'
I17
sS'm'
(lp13443
I1
aI1
aI1
aI11
aI27
aI15
aI15
aI231
aI205
aI677
aI331
aI133
aI3313
aI7193
aI8059
aI36449
aI21671
asS'd'
I6722
sa(dp13444
S'a'
I8489
sS's'
I17
sS'm'
(lp13445
I1
aI3
aI3
aI11
aI19
aI57
aI113
aI83
aI399
aI801
aI1843
aI2119
aI2779
aI14061
aI30901
aI28745
aI120903
asS'd'
I6723
sa(dp13446
S'a'
I8495
sS's'
I17
sS'm'
(lp13447
I1
aI1
aI1
aI11
aI5
aI27
aI121
aI247
aI467
aI251
aI1487
aI251
aI897
aI3171
aI28383
aI22473
aI1709
asS'd'
I6724
sa(dp13448
S'a'
I8522
sS's'
I17
sS'm'
(lp13449
I1
aI1
aI1
aI15
aI7
aI59
aI123
aI165
aI123
aI373
aI167
aI1323
aI5239
aI9027
aI13791
aI55593
aI78785
asS'd'
I6725
sa(dp13450
S'a'
I8524
sS's'
I17
sS'm'
(lp13451
I1
aI3
aI1
aI11
aI31
aI11
aI81
aI229
aI123
aI183
aI461
aI1751
aI5713
aI2615
aI27795
aI1657
aI39253
asS'd'
I6726
sa(dp13452
S'a'
I8529
sS's'
I17
sS'm'
(lp13453
I1
aI1
aI7
aI1
aI21
aI45
aI107
aI3
aI283
aI149
aI549
aI3731
aI6435
aI3595
aI32753
aI16079
aI84257
asS'd'
I6727
sa(dp13454
S'a'
I8545
sS's'
I17
sS'm'
(lp13455
I1
aI3
aI3
aI15
aI19
aI9
aI81
aI37
aI51
aI341
aI909
aI985
aI1503
aI12787
aI16129
aI37789
aI113515
asS'd'
I6728
sa(dp13456
S'a'
I8557
sS's'
I17
sS'm'
(lp13457
I1
aI3
aI5
aI13
aI3
aI33
aI127
aI219
aI369
aI341
aI1191
aI1305
aI567
aI2339
aI31221
aI49435
aI114927
asS'd'
I6729
sa(dp13458
S'a'
I8565
sS's'
I17
sS'm'
(lp13459
I1
aI1
aI7
aI15
aI29
aI47
aI103
aI107
aI257
aI15
aI2029
aI2133
aI2129
aI11235
aI29553
aI49139
aI33809
asS'd'
I6730
sa(dp13460
S'a'
I8572
sS's'
I17
sS'm'
(lp13461
I1
aI3
aI3
aI13
aI23
aI33
aI105
aI43
aI155
aI815
aI1087
aI2261
aI2781
aI3461
aI7371
aI4479
aI123093
asS'd'
I6731
sa(dp13462
S'a'
I8576
sS's'
I17
sS'm'
(lp13463
I1
aI1
aI1
aI13
aI17
aI7
aI89
aI107
aI143
aI349
aI637
aI3651
aI4153
aI12131
aI28393
aI45781
aI84133
asS'd'
I6732
sa(dp13464
S'a'
I8582
sS's'
I17
sS'm'
(lp13465
I1
aI3
aI5
aI11
aI31
aI47
aI105
aI101
aI267
aI403
aI1853
aI3977
aI3277
aI1737
aI15503
aI47365
aI14361
asS'd'
I6733
sa(dp13466
S'a'
I8594
sS's'
I17
sS'm'
(lp13467
I1
aI1
aI1
aI13
aI1
aI63
aI125
aI107
aI123
aI183
aI1027
aI3491
aI3597
aI15949
aI5779
aI34665
aI81257
asS'd'
I6734
sa(dp13468
S'a'
I8629
sS's'
I17
sS'm'
(lp13469
I1
aI3
aI1
aI9
aI13
aI5
aI125
aI41
aI389
aI73
aI1487
aI1983
aI957
aI12645
aI13983
aI7675
aI72711
asS'd'
I6735
sa(dp13470
S'a'
I8636
sS's'
I17
sS'm'
(lp13471
I1
aI3
aI7
aI5
aI17
aI5
aI25
aI63
aI211
aI591
aI261
aI2345
aI3883
aI4403
aI773
aI43963
aI93509
asS'd'
I6736
sa(dp13472
S'a'
I8668
sS's'
I17
sS'm'
(lp13473
I1
aI3
aI3
aI1
aI11
aI35
aI15
aI251
aI225
aI643
aI537
aI3769
aI7593
aI6113
aI1377
aI52185
aI81459
asS'd'
I6737
sa(dp13474
S'a'
I8678
sS's'
I17
sS'm'
(lp13475
I1
aI3
aI5
aI15
aI27
aI27
aI51
aI35
aI389
aI853
aI1437
aI2803
aI5739
aI1887
aI15099
aI3299
aI111827
asS'd'
I6738
sa(dp13476
S'a'
I8701
sS's'
I17
sS'm'
(lp13477
I1
aI1
aI3
aI15
aI25
aI63
aI31
aI201
aI79
aI131
aI31
aI3929
aI4195
aI13045
aI8681
aI48121
aI110723
asS'd'
I6739
sa(dp13478
S'a'
I8702
sS's'
I17
sS'm'
(lp13479
I1
aI1
aI5
aI7
aI11
aI43
aI101
aI57
aI69
aI271
aI189
aI3087
aI4893
aI11365
aI6945
aI14285
aI41961
asS'd'
I6740
sa(dp13480
S'a'
I8708
sS's'
I17
sS'm'
(lp13481
I1
aI1
aI7
aI9
aI21
aI61
aI41
aI123
aI25
aI947
aI1619
aI2895
aI7879
aI12397
aI17405
aI48139
aI71519
asS'd'
I6741
sa(dp13482
S'a'
I8712
sS's'
I17
sS'm'
(lp13483
I1
aI3
aI1
aI15
aI1
aI27
aI113
aI225
aI441
aI855
aI541
aI357
aI3111
aI4867
aI20571
aI30627
aI70123
asS'd'
I6742
sa(dp13484
S'a'
I8745
sS's'
I17
sS'm'
(lp13485
I1
aI3
aI5
aI3
aI5
aI33
aI103
aI1
aI21
aI93
aI383
aI407
aI5145
aI7857
aI20289
aI51943
aI16223
asS'd'
I6743
sa(dp13486
S'a'
I8754
sS's'
I17
sS'm'
(lp13487
I1
aI1
aI7
aI15
aI1
aI13
aI41
aI215
aI463
aI417
aI513
aI3417
aI1755
aI16165
aI7271
aI3101
aI54353
asS'd'
I6744
sa(dp13488
S'a'
I8759
sS's'
I17
sS'm'
(lp13489
I1
aI3
aI3
aI13
aI19
aI29
aI5
aI205
aI245
aI927
aI1249
aI773
aI3653
aI9959
aI357
aI40863
aI37289
asS'd'
I6745
sa(dp13490
S'a'
I8763
sS's'
I17
sS'm'
(lp13491
I1
aI3
aI3
aI7
aI3
aI5
aI85
aI241
aI29
aI627
aI1963
aI3133
aI1369
aI503
aI11449
aI4699
aI2573
asS'd'
I6746
sa(dp13492
S'a'
I8766
sS's'
I17
sS'm'
(lp13493
I1
aI1
aI7
aI15
aI3
aI35
aI47
aI157
aI413
aI437
aI1627
aI3953
aI947
aI12721
aI22209
aI34303
aI81237
asS'd'
I6747
sa(dp13494
S'a'
I8780
sS's'
I17
sS'm'
(lp13495
I1
aI1
aI5
aI5
aI1
aI45
aI47
aI245
aI253
aI349
aI1853
aI3481
aI6105
aI7267
aI3159
aI38833
aI117889
asS'd'
I6748
sa(dp13496
S'a'
I8783
sS's'
I17
sS'm'
(lp13497
I1
aI3
aI7
aI15
aI23
aI43
aI25
aI181
aI121
aI681
aI479
aI1239
aI6155
aI3317
aI9419
aI28717
aI44643
asS'd'
I6749
sa(dp13498
S'a'
I8786
sS's'
I17
sS'm'
(lp13499
I1
aI3
aI3
aI15
aI31
aI43
aI111
aI99
aI405
aI991
aI301
aI1689
aI7107
aI16131
aI16703
aI24059
aI40345
asS'd'
I6750
sa(dp13500
S'a'
I8798
sS's'
I17
sS'm'
(lp13501
I1
aI1
aI3
aI9
aI25
aI5
aI107
aI91
aI117
aI351
aI1595
aI163
aI3007
aI13743
aI24535
aI38671
aI29745
asS'd'
I6751
sa(dp13502
S'a'
I8804
sS's'
I17
sS'm'
(lp13503
I1
aI3
aI3
aI5
aI27
aI47
aI15
aI195
aI119
aI919
aI665
aI1903
aI1981
aI7753
aI21709
aI33699
aI15963
asS'd'
I6752
sa(dp13504
S'a'
I8819
sS's'
I17
sS'm'
(lp13505
I1
aI3
aI1
aI11
aI23
aI23
aI75
aI115
aI477
aI105
aI541
aI1111
aI209
aI13939
aI17129
aI7565
aI75415
asS'd'
I6753
sa(dp13506
S'a'
I8826
sS's'
I17
sS'm'
(lp13507
I1
aI1
aI1
aI11
aI7
aI61
aI123
aI201
aI305
aI713
aI779
aI2059
aI4899
aI13733
aI20529
aI15617
aI39833
asS'd'
I6754
sa(dp13508
S'a'
I8835
sS's'
I17
sS'm'
(lp13509
I1
aI1
aI7
aI11
aI21
aI7
aI63
aI113
aI213
aI871
aI375
aI29
aI1925
aI15237
aI7091
aI12229
aI8457
asS'd'
I6755
sa(dp13510
S'a'
I8838
sS's'
I17
sS'm'
(lp13511
I1
aI1
aI1
aI7
aI19
aI57
aI83
aI91
aI297
aI255
aI1993
aI63
aI5337
aI4569
aI21243
aI40867
aI46969
asS'd'
I6756
sa(dp13512
S'a'
I8856
sS's'
I17
sS'm'
(lp13513
I1
aI1
aI3
aI7
aI13
aI63
aI91
aI191
aI281
aI259
aI1367
aI3505
aI5885
aI10557
aI12423
aI56303
aI14731
asS'd'
I6757
sa(dp13514
S'a'
I8862
sS's'
I17
sS'm'
(lp13515
I1
aI1
aI5
aI15
aI27
aI15
aI29
aI67
aI115
aI287
aI253
aI1497
aI3739
aI2183
aI14427
aI44931
aI11547
asS'd'
I6758
sa(dp13516
S'a'
I8871
sS's'
I17
sS'm'
(lp13517
I1
aI3
aI1
aI9
aI25
aI61
aI25
aI113
aI137
aI819
aI781
aI3741
aI2457
aI7817
aI31209
aI20707
aI93007
asS'd'
I6759
sa(dp13518
S'a'
I8875
sS's'
I17
sS'm'
(lp13519
I1
aI1
aI7
aI3
aI5
aI13
aI23
aI3
aI365
aI77
aI1117
aI3061
aI4707
aI3013
aI27899
aI10887
aI78677
asS'd'
I6760
sa(dp13520
S'a'
I8890
sS's'
I17
sS'm'
(lp13521
I1
aI3
aI1
aI15
aI1
aI39
aI85
aI107
aI483
aI83
aI603
aI3121
aI1995
aI5241
aI32319
aI9515
aI94551
asS'd'
I6761
sa(dp13522
S'a'
I8892
sS's'
I17
sS'm'
(lp13523
I1
aI1
aI7
aI3
aI27
aI13
aI105
aI41
aI285
aI237
aI1589
aI517
aI2009
aI10833
aI1459
aI26217
aI50759
asS'd'
I6762
sa(dp13524
S'a'
I8898
sS's'
I17
sS'm'
(lp13525
I1
aI1
aI3
aI11
aI27
aI1
aI127
aI83
aI355
aI107
aI1003
aI657
aI4997
aI4123
aI13151
aI56601
aI122307
asS'd'
I6763
sa(dp13526
S'a'
I8927
sS's'
I17
sS'm'
(lp13527
I1
aI1
aI1
aI7
aI13
aI17
aI93
aI75
aI481
aI473
aI131
aI1359
aI4859
aI1319
aI23919
aI50079
aI128849
asS'd'
I6764
sa(dp13528
S'a'
I8928
sS's'
I17
sS'm'
(lp13529
I1
aI1
aI3
aI7
aI9
aI33
aI111
aI229
aI11
aI283
aI1089
aI3049
aI1635
aI959
aI19109
aI62821
aI105391
asS'd'
I6765
sa(dp13530
S'a'
I8945
sS's'
I17
sS'm'
(lp13531
I1
aI3
aI1
aI3
aI9
aI47
aI49
aI169
aI343
aI929
aI1379
aI1985
aI5867
aI6053
aI12179
aI39727
aI116053
asS'd'
I6766
sa(dp13532
S'a'
I8952
sS's'
I17
sS'm'
(lp13533
I1
aI3
aI3
aI15
aI27
aI39
aI61
aI113
aI439
aI719
aI1313
aI3701
aI4793
aI10275
aI2943
aI32405
aI95457
asS'd'
I6767
sa(dp13534
S'a'
I8955
sS's'
I17
sS'm'
(lp13535
I1
aI1
aI1
aI1
aI27
aI49
aI121
aI171
aI319
aI365
aI1593
aI1655
aI63
aI6257
aI18097
aI35285
aI112245
asS'd'
I6768
sa(dp13536
S'a'
I8965
sS's'
I17
sS'm'
(lp13537
I1
aI3
aI1
aI1
aI19
aI33
aI89
aI235
aI281
aI519
aI1867
aI525
aI4475
aI12059
aI26611
aI14789
aI59541
asS'd'
I6769
sa(dp13538
S'a'
I8972
sS's'
I17
sS'm'
(lp13539
I1
aI3
aI1
aI15
aI1
aI51
aI65
aI71
aI131
aI599
aI117
aI2459
aI7421
aI7157
aI24393
aI48139
aI53701
asS'd'
I6770
sa(dp13540
S'a'
I8977
sS's'
I17
sS'm'
(lp13541
I1
aI1
aI7
aI7
aI1
aI41
aI57
aI191
aI207
aI329
aI43
aI1235
aI5671
aI12243
aI22549
aI40751
aI104513
asS'd'
I6771
sa(dp13542
S'a'
I8990
sS's'
I17
sS'm'
(lp13543
I1
aI3
aI5
aI13
aI15
aI21
aI55
aI187
aI283
aI209
aI1511
aI1329
aI6665
aI15953
aI4521
aI16879
aI57625
asS'd'
I6772
sa(dp13544
S'a'
I8996
sS's'
I17
sS'm'
(lp13545
I1
aI1
aI5
aI3
aI3
aI53
aI75
aI123
aI291
aI663
aI1893
aI3669
aI4903
aI8575
aI27971
aI46977
aI56357
asS'd'
I6773
sa(dp13546
S'a'
I9025
sS's'
I17
sS'm'
(lp13547
I1
aI3
aI1
aI5
aI27
aI41
aI19
aI199
aI489
aI197
aI439
aI3299
aI6315
aI6957
aI15809
aI35297
aI5559
asS'd'
I6774
sa(dp13548
S'a'
I9037
sS's'
I17
sS'm'
(lp13549
I1
aI3
aI5
aI1
aI3
aI25
aI109
aI191
aI33
aI543
aI125
aI2309
aI429
aI14059
aI3149
aI45747
aI47357
asS'd'
I6775
sa(dp13550
S'a'
I9040
sS's'
I17
sS'm'
(lp13551
I1
aI1
aI3
aI11
aI15
aI61
aI109
aI103
aI305
aI1
aI1479
aI2781
aI6521
aI8921
aI23681
aI9583
aI87257
asS'd'
I6776
sa(dp13552
S'a'
I9049
sS's'
I17
sS'm'
(lp13553
I1
aI1
aI7
aI15
aI5
aI19
aI121
aI139
aI177
aI967
aI1363
aI705
aI211
aI11877
aI22457
aI34563
aI7801
asS'd'
I6777
sa(dp13554
S'a'
I9062
sS's'
I17
sS'm'
(lp13555
I1
aI1
aI7
aI13
aI9
aI21
aI103
aI95
aI483
aI567
aI5
aI2095
aI4659
aI2447
aI23521
aI27273
aI85867
asS'd'
I6778
sa(dp13556
S'a'
I9068
sS's'
I17
sS'm'
(lp13557
I1
aI3
aI5
aI15
aI23
aI55
aI13
aI237
aI275
aI113
aI1431
aI2931
aI5165
aI5317
aI5625
aI51865
aI42177
asS'd'
I6779
sa(dp13558
S'a'
I9076
sS's'
I17
sS'm'
(lp13559
I1
aI3
aI3
aI7
aI1
aI23
aI15
aI171
aI303
aI43
aI1137
aI1255
aI3843
aI9049
aI1799
aI7075
aI2115
asS'd'
I6780
sa(dp13560
S'a'
I9079
sS's'
I17
sS'm'
(lp13561
I1
aI1
aI7
aI5
aI23
aI53
aI75
aI129
aI1
aI511
aI793
aI265
aI6535
aI9641
aI25173
aI9449
aI46949
asS'd'
I6781
sa(dp13562
S'a'
I9099
sS's'
I17
sS'm'
(lp13563
I1
aI3
aI3
aI1
aI19
aI39
aI51
aI173
aI5
aI281
aI2047
aI4065
aI3225
aI14587
aI16947
aI1459
aI87227
asS'd'
I6782
sa(dp13564
S'a'
I9107
sS's'
I17
sS'm'
(lp13565
I1
aI3
aI7
aI13
aI13
aI53
aI39
aI115
aI403
aI37
aI1533
aI2727
aI2229
aI8291
aI18687
aI59553
aI37629
asS'd'
I6783
sa(dp13566
S'a'
I9114
sS's'
I17
sS'm'
(lp13567
I1
aI3
aI1
aI9
aI3
aI55
aI63
aI191
aI147
aI321
aI1287
aI2419
aI6881
aI2249
aI11141
aI54839
aI50263
asS'd'
I6784
sa(dp13568
S'a'
I9123
sS's'
I17
sS'm'
(lp13569
I1
aI1
aI5
aI3
aI9
aI61
aI85
aI139
aI1
aI409
aI633
aI53
aI163
aI14677
aI13043
aI12253
aI106939
asS'd'
I6785
sa(dp13570
S'a'
I9126
sS's'
I17
sS'm'
(lp13571
I1
aI1
aI7
aI3
aI19
aI3
aI7
aI165
aI497
aI621
aI1563
aI1267
aI8113
aI2383
aI17205
aI13337
aI102547
asS'd'
I6786
sa(dp13572
S'a'
I9137
sS's'
I17
sS'm'
(lp13573
I1
aI3
aI3
aI13
aI15
aI29
aI23
aI31
aI481
aI535
aI471
aI2125
aI331
aI9421
aI29799
aI27097
aI5307
asS'd'
I6787
sa(dp13574
S'a'
I9149
sS's'
I17
sS'm'
(lp13575
I1
aI1
aI1
aI1
aI31
aI45
aI47
aI139
aI235
aI509
aI889
aI685
aI1855
aI13599
aI24431
aI62105
aI109509
asS'd'
I6788
sa(dp13576
S'a'
I9150
sS's'
I17
sS'm'
(lp13577
I1
aI3
aI1
aI7
aI3
aI13
aI25
aI197
aI111
aI45
aI1815
aI1031
aI4803
aI349
aI32369
aI40837
aI111529
asS'd'
I6789
sa(dp13578
S'a'
I9155
sS's'
I17
sS'm'
(lp13579
I1
aI1
aI7
aI1
aI27
aI9
aI3
aI73
aI403
aI321
aI967
aI2713
aI6953
aI16123
aI8611
aI48651
aI120635
asS'd'
I6790
sa(dp13580
S'a'
I9161
sS's'
I17
sS'm'
(lp13581
I1
aI3
aI5
aI3
aI3
aI25
aI69
aI231
aI249
aI393
aI1141
aI1721
aI7071
aI3711
aI15627
aI21815
aI104735
asS'd'
I6791
sa(dp13582
S'a'
I9162
sS's'
I17
sS'm'
(lp13583
I1
aI3
aI1
aI11
aI19
aI63
aI77
aI5
aI55
aI481
aI1021
aI119
aI3941
aI1227
aI10997
aI29513
aI18923
asS'd'
I6792
sa(dp13584
S'a'
I9167
sS's'
I17
sS'm'
(lp13585
I1
aI3
aI7
aI5
aI1
aI11
aI13
aI99
aI365
aI797
aI1993
aI699
aI3091
aI11401
aI3659
aI15339
aI90395
asS'd'
I6793
sa(dp13586
S'a'
I9172
sS's'
I17
sS'm'
(lp13587
I1
aI3
aI5
aI7
aI31
aI43
aI55
aI143
aI273
aI379
aI1189
aI1689
aI4811
aI5159
aI3281
aI63819
aI57065
asS'd'
I6794
sa(dp13588
S'a'
I9186
sS's'
I17
sS'm'
(lp13589
I1
aI1
aI1
aI13
aI9
aI25
aI9
aI3
aI461
aI281
aI959
aI2439
aI3187
aI4837
aI13857
aI20221
aI29733
asS'd'
I6795
sa(dp13590
S'a'
I9188
sS's'
I17
sS'm'
(lp13591
I1
aI1
aI7
aI11
aI31
aI17
aI13
aI101
aI81
aI921
aI1329
aI2421
aI2747
aI9435
aI23313
aI7093
aI7547
asS'd'
I6796
sa(dp13592
S'a'
I9191
sS's'
I17
sS'm'
(lp13593
I1
aI1
aI3
aI3
aI9
aI51
aI67
aI95
aI511
aI1011
aI1519
aI4089
aI5001
aI1351
aI15367
aI50665
aI92111
asS'd'
I6797
sa(dp13594
S'a'
I9198
sS's'
I17
sS'm'
(lp13595
I1
aI1
aI5
aI13
aI27
aI43
aI115
aI77
aI439
aI589
aI31
aI915
aI7027
aI697
aI25143
aI1443
aI59093
asS'd'
I6798
sa(dp13596
S'a'
I9200
sS's'
I17
sS'm'
(lp13597
I1
aI1
aI7
aI3
aI17
aI5
aI107
aI117
aI133
aI649
aI1309
aI2979
aI969
aI9789
aI12597
aI24507
aI106825
asS'd'
I6799
sa(dp13598
S'a'
I9205
sS's'
I17
sS'm'
(lp13599
I1
aI1
aI7
aI13
aI1
aI27
aI97
aI35
aI431
aI183
aI199
aI2619
aI515
aI89
aI20281
aI30291
aI97977
asS'd'
I6800
sa(dp13600
S'a'
I9206
sS's'
I17
sS'm'
(lp13601
I1
aI1
aI7
aI1
aI31
aI9
aI35
aI11
aI359
aI21
aI1875
aI3227
aI1307
aI15691
aI17343
aI21163
aI84671
asS'd'
I6801
sa(dp13602
S'a'
I9215
sS's'
I17
sS'm'
(lp13603
I1
aI3
aI1
aI11
aI29
aI21
aI47
aI137
aI441
aI841
aI1641
aI3283
aI1371
aI8835
aI16287
aI45009
aI13779
asS'd'
I6802
sa(dp13604
S'a'
I9227
sS's'
I17
sS'm'
(lp13605
I1
aI1
aI3
aI9
aI23
aI53
aI1
aI99
aI473
aI649
aI447
aI2589
aI5667
aI15579
aI6497
aI44321
aI46993
asS'd'
I6803
sa(dp13606
S'a'
I9232
sS's'
I17
sS'm'
(lp13607
I1
aI1
aI7
aI9
aI31
aI63
aI95
aI81
aI197
aI373
aI1027
aI3959
aI7189
aI13369
aI17287
aI53643
aI12673
asS'd'
I6804
sa(dp13608
S'a'
I9241
sS's'
I17
sS'm'
(lp13609
I1
aI3
aI1
aI5
aI25
aI61
aI79
aI183
aI489
aI725
aI1077
aI1147
aI113
aI7357
aI27505
aI529
aI61855
asS'd'
I6805
sa(dp13610
S'a'
I9244
sS's'
I17
sS'm'
(lp13611
I1
aI1
aI7
aI11
aI19
aI35
aI73
aI223
aI125
aI765
aI1303
aI2887
aI7861
aI14839
aI9537
aI27027
aI94327
asS'd'
I6806
sa(dp13612
S'a'
I9251
sS's'
I17
sS'm'
(lp13613
I1
aI3
aI1
aI3
aI17
aI35
aI63
aI233
aI317
aI133
aI1837
aI3339
aI4351
aI10071
aI5005
aI13245
aI34327
asS'd'
I6807
sa(dp13614
S'a'
I9254
sS's'
I17
sS'm'
(lp13615
I1
aI3
aI1
aI3
aI17
aI13
aI59
aI113
aI247
aI1015
aI1831
aI3391
aI6337
aI6853
aI7145
aI64309
aI40109
asS'd'
I6808
sa(dp13616
S'a'
I9275
sS's'
I17
sS'm'
(lp13617
I1
aI3
aI5
aI13
aI15
aI23
aI65
aI203
aI241
aI545
aI1521
aI1253
aI3171
aI7777
aI21145
aI565
aI87813
asS'd'
I6809
sa(dp13618
S'a'
I9283
sS's'
I17
sS'm'
(lp13619
I1
aI1
aI5
aI15
aI31
aI9
aI9
aI145
aI409
aI155
aI409
aI2935
aI5817
aI11427
aI32617
aI38167
aI69465
asS'd'
I6810
sa(dp13620
S'a'
I9285
sS's'
I17
sS'm'
(lp13621
I1
aI1
aI5
aI11
aI19
aI31
aI43
aI85
aI97
aI931
aI687
aI1501
aI3991
aI2215
aI11091
aI64735
aI56999
asS'd'
I6811
sa(dp13622
S'a'
I9303
sS's'
I17
sS'm'
(lp13623
I1
aI1
aI1
aI3
aI7
aI11
aI101
aI21
aI345
aI829
aI531
aI1475
aI6617
aI1187
aI26885
aI32135
aI9733
asS'd'
I6812
sa(dp13624
S'a'
I9304
sS's'
I17
sS'm'
(lp13625
I1
aI3
aI5
aI11
aI7
aI49
aI79
aI197
aI57
aI15
aI1845
aI1485
aI6167
aI10887
aI17083
aI59367
aI7411
asS'd'
I6813
sa(dp13626
S'a'
I9313
sS's'
I17
sS'm'
(lp13627
I1
aI3
aI7
aI5
aI9
aI33
aI7
aI91
aI311
aI847
aI1435
aI3573
aI3693
aI5369
aI26817
aI30105
aI115337
asS'd'
I6814
sa(dp13628
S'a'
I9314
sS's'
I17
sS'm'
(lp13629
I1
aI3
aI1
aI9
aI25
aI43
aI65
aI69
aI225
aI337
aI575
aI1979
aI5555
aI8499
aI8127
aI33035
aI52549
asS'd'
I6815
sa(dp13630
S'a'
I9320
sS's'
I17
sS'm'
(lp13631
I1
aI1
aI3
aI11
aI17
aI29
aI71
aI99
aI379
aI145
aI1067
aI2561
aI7635
aI5647
aI32491
aI56621
aI93603
asS'd'
I6816
sa(dp13632
S'a'
I9328
sS's'
I17
sS'm'
(lp13633
I1
aI1
aI5
aI13
aI25
aI43
aI75
aI237
aI407
aI393
aI1219
aI3651
aI7719
aI11685
aI26123
aI62767
aI1043
asS'd'
I6817
sa(dp13634
S'a'
I9333
sS's'
I17
sS'm'
(lp13635
I1
aI1
aI7
aI15
aI13
aI59
aI9
aI163
aI273
aI225
aI873
aI3201
aI633
aI6121
aI18777
aI58763
aI77731
asS'd'
I6818
sa(dp13636
S'a'
I9337
sS's'
I17
sS'm'
(lp13637
I1
aI3
aI7
aI7
aI3
aI7
aI99
aI155
aI279
aI991
aI799
aI753
aI7205
aI9567
aI23643
aI38263
aI19083
asS'd'
I6819
sa(dp13638
S'a'
I9338
sS's'
I17
sS'm'
(lp13639
I1
aI3
aI7
aI11
aI11
aI29
aI65
aI3
aI207
aI575
aI253
aI2407
aI7935
aI11323
aI23239
aI1923
aI47737
asS'd'
I6820
sa(dp13640
S'a'
I9340
sS's'
I17
sS'm'
(lp13641
I1
aI1
aI5
aI9
aI25
aI47
aI1
aI25
aI397
aI1009
aI193
aI4031
aI3023
aI2029
aI10561
aI32363
aI104405
asS'd'
I6821
sa(dp13642
S'a'
I9353
sS's'
I17
sS'm'
(lp13643
I1
aI3
aI7
aI9
aI19
aI55
aI63
aI179
aI385
aI97
aI461
aI3393
aI8137
aI8929
aI17621
aI9611
aI58925
asS'd'
I6822
sa(dp13644
S'a'
I9356
sS's'
I17
sS'm'
(lp13645
I1
aI1
aI1
aI7
aI1
aI17
aI127
aI45
aI157
aI529
aI809
aI3545
aI5173
aI5083
aI13325
aI52295
aI91261
asS'd'
I6823
sa(dp13646
S'a'
I9364
sS's'
I17
sS'm'
(lp13647
I1
aI1
aI7
aI9
aI25
aI49
aI99
aI79
aI157
aI535
aI1569
aI2195
aI1725
aI1187
aI18423
aI47957
aI10043
asS'd'
I6824
sa(dp13648
S'a'
I9373
sS's'
I17
sS'm'
(lp13649
I1
aI1
aI3
aI7
aI3
aI31
aI83
aI45
aI199
aI665
aI1261
aI3497
aI7885
aI5761
aI17187
aI12041
aI12867
asS'd'
I6825
sa(dp13650
S'a'
I9374
sS's'
I17
sS'm'
(lp13651
I1
aI3
aI1
aI7
aI3
aI55
aI73
aI215
aI41
aI1011
aI1883
aI1051
aI7293
aI1881
aI27435
aI29459
aI130933
asS'd'
I6826
sa(dp13652
S'a'
I9378
sS's'
I17
sS'm'
(lp13653
I1
aI1
aI3
aI9
aI21
aI31
aI113
aI209
aI35
aI771
aI365
aI3151
aI787
aI3845
aI26555
aI13823
aI36951
asS'd'
I6827
sa(dp13654
S'a'
I9380
sS's'
I17
sS'm'
(lp13655
I1
aI3
aI7
aI15
aI13
aI21
aI119
aI91
aI15
aI251
aI1337
aI2715
aI1665
aI3451
aI8309
aI11033
aI127159
asS'd'
I6828
sa(dp13656
S'a'
I9389
sS's'
I17
sS'm'
(lp13657
I1
aI3
aI1
aI9
aI9
aI63
aI5
aI145
aI357
aI9
aI859
aI1565
aI1141
aI14689
aI25121
aI41337
aI83357
asS'd'
I6829
sa(dp13658
S'a'
I9395
sS's'
I17
sS'm'
(lp13659
I1
aI1
aI7
aI11
aI13
aI63
aI57
aI151
aI33
aI595
aI2025
aI571
aI4713
aI11019
aI26771
aI16221
aI92439
asS'd'
I6830
sa(dp13660
S'a'
I9412
sS's'
I17
sS'm'
(lp13661
I1
aI3
aI3
aI15
aI29
aI49
aI93
aI131
aI167
aI835
aI33
aI263
aI93
aI8475
aI16139
aI61237
aI95081
asS'd'
I6831
sa(dp13662
S'a'
I9422
sS's'
I17
sS'm'
(lp13663
I1
aI1
aI7
aI13
aI1
aI57
aI43
aI91
aI485
aI841
aI1415
aI3083
aI2665
aI8521
aI9825
aI59955
aI21763
asS'd'
I6832
sa(dp13664
S'a'
I9439
sS's'
I17
sS'm'
(lp13665
I1
aI1
aI1
aI1
aI29
aI47
aI63
aI107
aI439
aI847
aI537
aI2011
aI7571
aI3699
aI23961
aI54887
aI92681
asS'd'
I6833
sa(dp13666
S'a'
I9450
sS's'
I17
sS'm'
(lp13667
I1
aI3
aI7
aI5
aI27
aI41
aI105
aI161
aI95
aI821
aI451
aI2627
aI4687
aI1899
aI18851
aI35167
aI6869
asS'd'
I6834
sa(dp13668
S'a'
I9452
sS's'
I17
sS'm'
(lp13669
I1
aI1
aI1
aI11
aI7
aI7
aI13
aI163
aI399
aI471
aI1587
aI2561
aI1241
aI5365
aI27189
aI49883
aI68101
asS'd'
I6835
sa(dp13670
S'a'
I9482
sS's'
I17
sS'm'
(lp13671
I1
aI3
aI7
aI9
aI19
aI5
aI119
aI251
aI151
aI359
aI235
aI2387
aI3919
aI7135
aI17591
aI1053
aI6265
asS'd'
I6836
sa(dp13672
S'a'
I9487
sS's'
I17
sS'm'
(lp13673
I1
aI1
aI5
aI9
aI13
aI25
aI43
aI23
aI453
aI693
aI517
aI1235
aI1045
aI4299
aI27877
aI3733
aI72269
asS'd'
I6837
sa(dp13674
S'a'
I9489
sS's'
I17
sS'm'
(lp13675
I1
aI1
aI7
aI1
aI27
aI43
aI103
aI249
aI487
aI67
aI855
aI3239
aI2157
aI8121
aI4701
aI37803
aI49971
asS'd'
I6838
sa(dp13676
S'a'
I9499
sS's'
I17
sS'm'
(lp13677
I1
aI1
aI3
aI13
aI1
aI37
aI125
aI115
aI365
aI57
aI1419
aI4085
aI7039
aI10079
aI14991
aI48861
aI61979
asS'd'
I6839
sa(dp13678
S'a'
I9501
sS's'
I17
sS'm'
(lp13679
I1
aI1
aI5
aI5
aI3
aI35
aI109
aI19
aI219
aI653
aI1219
aI1625
aI6847
aI11271
aI4525
aI56341
aI57801
asS'd'
I6840
sa(dp13680
S'a'
I9508
sS's'
I17
sS'm'
(lp13681
I1
aI3
aI7
aI5
aI31
aI19
aI37
aI73
aI185
aI13
aI1723
aI1139
aI5919
aI11717
aI27161
aI13635
aI51765
asS'd'
I6841
sa(dp13682
S'a'
I9515
sS's'
I17
sS'm'
(lp13683
I1
aI1
aI1
aI1
aI19
aI61
aI53
aI111
aI215
aI189
aI1199
aI591
aI943
aI2111
aI17171
aI15621
aI128459
asS'd'
I6842
sa(dp13684
S'a'
I9518
sS's'
I17
sS'm'
(lp13685
I1
aI1
aI7
aI9
aI17
aI61
aI101
aI159
aI85
aI537
aI15
aI1427
aI6139
aI4091
aI32639
aI28655
aI115385
asS'd'
I6843
sa(dp13686
S'a'
I9520
sS's'
I17
sS'm'
(lp13687
I1
aI1
aI7
aI5
aI23
aI31
aI125
aI7
aI151
aI967
aI1079
aI4059
aI3287
aI11673
aI19307
aI49469
aI65981
asS'd'
I6844
sa(dp13688
S'a'
I9526
sS's'
I17
sS'm'
(lp13689
I1
aI3
aI3
aI1
aI29
aI59
aI95
aI119
aI31
aI427
aI1653
aI721
aI5509
aI6385
aI17043
aI45133
aI74155
asS'd'
I6845
sa(dp13690
S'a'
I9537
sS's'
I17
sS'm'
(lp13691
I1
aI1
aI3
aI9
aI13
aI61
aI35
aI189
aI1
aI559
aI119
aI3719
aI4137
aI1369
aI19147
aI10923
aI43909
asS'd'
I6846
sa(dp13692
S'a'
I9552
sS's'
I17
sS'm'
(lp13693
I1
aI3
aI3
aI13
aI1
aI41
aI31
aI185
aI451
aI379
aI29
aI153
aI4121
aI13153
aI4171
aI36993
aI109241
asS'd'
I6847
sa(dp13694
S'a'
I9571
sS's'
I17
sS'm'
(lp13695
I1
aI1
aI1
aI9
aI15
aI41
aI99
aI17
aI21
aI93
aI649
aI2765
aI6955
aI10843
aI12547
aI64989
aI63713
asS'd'
I6848
sa(dp13696
S'a'
I9588
sS's'
I17
sS'm'
(lp13697
I1
aI1
aI7
aI5
aI5
aI5
aI73
aI187
aI473
aI235
aI1907
aI409
aI7335
aI4429
aI7493
aI20703
aI14505
asS'd'
I6849
sa(dp13698
S'a'
I9613
sS's'
I17
sS'm'
(lp13699
I1
aI1
aI3
aI11
aI27
aI59
aI17
aI103
aI337
aI117
aI1241
aI951
aI3701
aI10407
aI16741
aI46531
aI56485
asS'd'
I6850
sa(dp13700
S'a'
I9619
sS's'
I17
sS'm'
(lp13701
I1
aI1
aI3
aI15
aI11
aI51
aI111
aI189
aI137
aI939
aI97
aI1563
aI851
aI13949
aI1375
aI41463
aI61445
asS'd'
I6851
sa(dp13702
S'a'
I9622
sS's'
I17
sS'm'
(lp13703
I1
aI1
aI7
aI9
aI19
aI39
aI117
aI173
aI165
aI547
aI483
aI361
aI6819
aI15093
aI13631
aI29785
aI29593
asS'd'
I6852
sa(dp13704
S'a'
I9637
sS's'
I17
sS'm'
(lp13705
I1
aI3
aI3
aI5
aI15
aI39
aI41
aI249
aI455
aI79
aI233
aI3133
aI405
aI9487
aI23161
aI32751
aI117743
asS'd'
I6853
sa(dp13706
S'a'
I9652
sS's'
I17
sS'm'
(lp13707
I1
aI1
aI5
aI15
aI7
aI63
aI7
aI57
aI127
aI349
aI1913
aI1145
aI3371
aI3733
aI30971
aI35717
aI60935
asS'd'
I6854
sa(dp13708
S'a'
I9655
sS's'
I17
sS'm'
(lp13709
I1
aI1
aI7
aI11
aI7
aI57
aI49
aI63
aI51
aI233
aI855
aI2125
aI6961
aI15011
aI28503
aI40549
aI47175
asS'd'
I6855
sa(dp13710
S'a'
I9661
sS's'
I17
sS'm'
(lp13711
I1
aI3
aI7
aI1
aI25
aI49
aI35
aI39
aI237
aI545
aI1637
aI1401
aI3279
aI10499
aI14463
aI34973
aI29485
asS'd'
I6856
sa(dp13712
S'a'
I9664
sS's'
I17
sS'm'
(lp13713
I1
aI3
aI3
aI13
aI7
aI13
aI79
aI141
aI55
aI277
aI843
aI3087
aI2339
aI6855
aI10635
aI13021
aI11273
asS'd'
I6857
sa(dp13714
S'a'
I9669
sS's'
I17
sS'm'
(lp13715
I1
aI3
aI1
aI1
aI11
aI39
aI51
aI255
aI119
aI691
aI559
aI3287
aI5485
aI791
aI19283
aI51027
aI8061
asS'd'
I6858
sa(dp13716
S'a'
I9681
sS's'
I17
sS'm'
(lp13717
I1
aI3
aI7
aI7
aI3
aI59
aI119
aI241
aI185
aI81
aI1843
aI2313
aI7471
aI15689
aI2271
aI59781
aI107439
asS'd'
I6859
sa(dp13718
S'a'
I9682
sS's'
I17
sS'm'
(lp13719
I1
aI3
aI3
aI3
aI17
aI63
aI93
aI217
aI329
aI39
aI583
aI3031
aI4315
aI4623
aI12557
aI42063
aI11877
asS'd'
I6860
sa(dp13720
S'a'
I9688
sS's'
I17
sS'm'
(lp13721
I1
aI1
aI1
aI1
aI15
aI57
aI37
aI233
aI387
aI639
aI37
aI425
aI637
aI1577
aI16449
aI33665
aI80417
asS'd'
I6861
sa(dp13722
S'a'
I9697
sS's'
I17
sS'm'
(lp13723
I1
aI1
aI1
aI15
aI25
aI1
aI67
aI159
aI423
aI961
aI959
aI417
aI5657
aI8417
aI8127
aI29251
aI105893
asS'd'
I6862
sa(dp13724
S'a'
I9700
sS's'
I17
sS'm'
(lp13725
I1
aI3
aI5
aI15
aI31
aI9
aI87
aI217
aI259
aI771
aI1663
aI2899
aI1531
aI7849
aI1961
aI61487
aI55399
asS'd'
I6863
sa(dp13726
S'a'
I9715
sS's'
I17
sS'm'
(lp13727
I1
aI1
aI3
aI9
aI21
aI13
aI39
aI107
aI89
aI811
aI449
aI2569
aI4617
aI8977
aI1649
aI37721
aI48943
asS'd'
I6864
sa(dp13728
S'a'
I9722
sS's'
I17
sS'm'
(lp13729
I1
aI3
aI7
aI15
aI15
aI59
aI63
aI195
aI287
aI677
aI269
aI1715
aI3545
aI3269
aI5231
aI46433
aI25921
asS'd'
I6865
sa(dp13730
S'a'
I9727
sS's'
I17
sS'm'
(lp13731
I1
aI1
aI5
aI7
aI19
aI27
aI57
aI221
aI243
aI47
aI1791
aI2309
aI2751
aI4403
aI7083
aI34223
aI64905
asS'd'
I6866
sa(dp13732
S'a'
I9734
sS's'
I17
sS'm'
(lp13733
I1
aI1
aI1
aI15
aI1
aI63
aI119
aI155
aI383
aI649
aI429
aI3857
aI7309
aI9823
aI9539
aI8933
aI128573
asS'd'
I6867
sa(dp13734
S'a'
I9740
sS's'
I17
sS'm'
(lp13735
I1
aI3
aI7
aI11
aI17
aI19
aI99
aI19
aI321
aI415
aI1501
aI2123
aI6119
aI9705
aI11397
aI39521
aI34327
asS'd'
I6868
sa(dp13736
S'a'
I9743
sS's'
I17
sS'm'
(lp13737
I1
aI1
aI5
aI15
aI29
aI37
aI9
aI95
aI417
aI19
aI1637
aI2949
aI4961
aI10743
aI9619
aI16045
aI48083
asS'd'
I6869
sa(dp13738
S'a'
I9745
sS's'
I17
sS'm'
(lp13739
I1
aI1
aI1
aI11
aI21
aI17
aI57
aI23
aI247
aI201
aI1781
aI779
aI2207
aI2511
aI4829
aI13847
aI77593
asS'd'
I6870
sa(dp13740
S'a'
I9757
sS's'
I17
sS'm'
(lp13741
I1
aI3
aI1
aI13
aI7
aI1
aI95
aI87
aI223
aI73
aI1129
aI383
aI1355
aI4965
aI29645
aI63465
aI76281
asS'd'
I6871
sa(dp13742
S'a'
I9761
sS's'
I17
sS'm'
(lp13743
I1
aI3
aI3
aI13
aI3
aI47
aI33
aI123
aI155
aI621
aI1019
aI1817
aI4083
aI4723
aI24701
aI47503
aI18007
asS'd'
I6872
sa(dp13744
S'a'
I9762
sS's'
I17
sS'm'
(lp13745
I1
aI1
aI7
aI15
aI13
aI41
aI73
aI93
aI379
aI923
aI1183
aI2475
aI5901
aI10599
aI10053
aI9941
aI112107
asS'd'
I6873
sa(dp13746
S'a'
I9767
sS's'
I17
sS'm'
(lp13747
I1
aI1
aI3
aI3
aI13
aI35
aI59
aI231
aI45
aI1011
aI1101
aI2467
aI2703
aI10305
aI12575
aI7587
aI25737
asS'd'
I6874
sa(dp13748
S'a'
I9768
sS's'
I17
sS'm'
(lp13749
I1
aI3
aI7
aI1
aI21
aI31
aI9
aI55
aI373
aI779
aI397
aI1551
aI5139
aI16339
aI1769
aI10413
aI74059
asS'd'
I6875
sa(dp13750
S'a'
I9774
sS's'
I17
sS'm'
(lp13751
I1
aI1
aI7
aI15
aI7
aI3
aI67
aI179
aI411
aI217
aI1219
aI13
aI1577
aI13463
aI12263
aI41465
aI83001
asS'd'
I6876
sa(dp13752
S'a'
I9786
sS's'
I17
sS'm'
(lp13753
I1
aI3
aI7
aI1
aI21
aI53
aI7
aI187
aI395
aI777
aI391
aI737
aI47
aI12681
aI16749
aI26507
aI49415
asS'd'
I6877
sa(dp13754
S'a'
I9796
sS's'
I17
sS'm'
(lp13755
I1
aI1
aI5
aI7
aI5
aI57
aI93
aI53
aI419
aI731
aI825
aI487
aI45
aI9199
aI20947
aI56067
aI45343
asS'd'
I6878
sa(dp13756
S'a'
I9820
sS's'
I17
sS'm'
(lp13757
I1
aI3
aI3
aI9
aI31
aI41
aI35
aI133
aI63
aI293
aI1503
aI51
aI3111
aI15711
aI15051
aI1965
aI64951
asS'd'
I6879
sa(dp13758
S'a'
I9823
sS's'
I17
sS'm'
(lp13759
I1
aI1
aI5
aI9
aI9
aI47
aI53
aI229
aI405
aI621
aI1795
aI1923
aI6609
aI6983
aI1695
aI18021
aI71893
asS'd'
I6880
sa(dp13760
S'a'
I9839
sS's'
I17
sS'm'
(lp13761
I1
aI1
aI5
aI9
aI23
aI13
aI107
aI13
aI149
aI759
aI1113
aI1329
aI1747
aI14159
aI16705
aI61841
aI82955
asS'd'
I6881
sa(dp13762
S'a'
I9844
sS's'
I17
sS'm'
(lp13763
I1
aI3
aI3
aI9
aI25
aI49
aI31
aI145
aI481
aI609
aI1847
aI1485
aI6345
aI7859
aI21231
aI37303
aI69975
asS'd'
I6882
sa(dp13764
S'a'
I9851
sS's'
I17
sS'm'
(lp13765
I1
aI3
aI1
aI15
aI13
aI49
aI59
aI221
aI27
aI517
aI431
aI3961
aI6401
aI8483
aI10161
aI37453
aI128237
asS'd'
I6883
sa(dp13766
S'a'
I9853
sS's'
I17
sS'm'
(lp13767
I1
aI1
aI3
aI1
aI3
aI55
aI37
aI111
aI263
aI735
aI655
aI2831
aI2219
aI9449
aI8413
aI49585
aI91355
asS'd'
I6884
sa(dp13768
S'a'
I9863
sS's'
I17
sS'm'
(lp13769
I1
aI3
aI7
aI1
aI31
aI33
aI7
aI55
aI261
aI977
aI1215
aI1967
aI7297
aI14815
aI27009
aI35001
aI89671
asS'd'
I6885
sa(dp13770
S'a'
I9864
sS's'
I17
sS'm'
(lp13771
I1
aI1
aI7
aI11
aI13
aI21
aI33
aI151
aI195
aI373
aI181
aI1631
aI355
aI7857
aI12555
aI7531
aI50417
asS'd'
I6886
sa(dp13772
S'a'
I9877
sS's'
I17
sS'm'
(lp13773
I1
aI3
aI1
aI15
aI19
aI25
aI79
aI195
aI237
aI385
aI1531
aI2509
aI4371
aI16103
aI3575
aI62265
aI124251
asS'd'
I6887
sa(dp13774
S'a'
I9884
sS's'
I17
sS'm'
(lp13775
I1
aI3
aI1
aI11
aI5
aI61
aI21
aI159
aI51
aI37
aI845
aI3075
aI8039
aI14269
aI10505
aI36369
aI73793
asS'd'
I6888
sa(dp13776
S'a'
I9888
sS's'
I17
sS'm'
(lp13777
I1
aI3
aI5
aI9
aI11
aI43
aI67
aI57
aI271
aI451
aI989
aI3705
aI2481
aI10717
aI10861
aI63785
aI10183
asS'd'
I6889
sa(dp13778
S'a'
I9897
sS's'
I17
sS'm'
(lp13779
I1
aI3
aI3
aI5
aI13
aI29
aI119
aI171
aI439
aI459
aI479
aI3173
aI3781
aI11131
aI6827
aI53925
aI119939
asS'd'
I6890
sa(dp13780
S'a'
I9915
sS's'
I17
sS'm'
(lp13781
I1
aI3
aI7
aI3
aI27
aI21
aI1
aI167
aI79
aI305
aI1283
aI1903
aI5483
aI5727
aI17911
aI16075
aI97629
asS'd'
I6891
sa(dp13782
S'a'
I9925
sS's'
I17
sS'm'
(lp13783
I1
aI3
aI1
aI3
aI23
aI21
aI29
aI185
aI227
aI295
aI915
aI2033
aI6269
aI2089
aI20785
aI15207
aI115675
asS'd'
I6892
sa(dp13784
S'a'
I9949
sS's'
I17
sS'm'
(lp13785
I1
aI3
aI7
aI15
aI11
aI15
aI65
aI103
aI249
aI27
aI1805
aI2079
aI4797
aI2535
aI16865
aI61449
aI90923
asS'd'
I6893
sa(dp13786
S'a'
I9954
sS's'
I17
sS'm'
(lp13787
I1
aI3
aI7
aI9
aI27
aI41
aI77
aI181
aI457
aI677
aI633
aI1601
aI8085
aI2431
aI7957
aI55913
aI38677
asS'd'
I6894
sa(dp13788
S'a'
I9960
sS's'
I17
sS'm'
(lp13789
I1
aI1
aI5
aI7
aI11
aI37
aI3
aI221
aI79
aI895
aI1023
aI653
aI3925
aI12755
aI19729
aI18221
aI91123
asS'd'
I6895
sa(dp13790
S'a'
I9965
sS's'
I17
sS'm'
(lp13791
I1
aI3
aI1
aI5
aI23
aI61
aI119
aI191
aI425
aI41
aI853
aI3497
aI6915
aI1927
aI5513
aI55303
aI4895
asS'd'
I6896
sa(dp13792
S'a'
I9978
sS's'
I17
sS'm'
(lp13793
I1
aI3
aI5
aI3
aI7
aI35
aI47
aI243
aI167
aI821
aI267
aI2149
aI5797
aI6329
aI32495
aI51037
aI18313
asS'd'
I6897
sa(dp13794
S'a'
I9986
sS's'
I17
sS'm'
(lp13795
I1
aI1
aI7
aI9
aI23
aI29
aI79
aI205
aI115
aI839
aI1217
aI479
aI1601
aI9681
aI1
aI35293
aI28731
asS'd'
I6898
sa(dp13796
S'a'
I9992
sS's'
I17
sS'm'
(lp13797
I1
aI3
aI3
aI5
aI31
aI17
aI31
aI161
aI35
aI953
aI377
aI451
aI7985
aI11371
aI15115
aI60131
aI27033
asS'd'
I6899
sa(dp13798
S'a'
I9995
sS's'
I17
sS'm'
(lp13799
I1
aI1
aI3
aI9
aI15
aI19
aI43
aI215
aI327
aI429
aI145
aI1837
aI725
aI14775
aI10465
aI7367
aI21271
asS'd'
I6900
sa(dp13800
S'a'
I10005
sS's'
I17
sS'm'
(lp13801
I1
aI3
aI7
aI13
aI31
aI17
aI85
aI49
aI487
aI795
aI1679
aI599
aI3783
aI3195
aI2683
aI53475
aI38603
asS'd'
I6901
sa(dp13802
S'a'
I10026
sS's'
I17
sS'm'
(lp13803
I1
aI1
aI1
aI7
aI19
aI11
aI71
aI143
aI443
aI199
aI1117
aI3445
aI6429
aI12037
aI13751
aI43609
aI101563
asS'd'
I6902
sa(dp13804
S'a'
I10031
sS's'
I17
sS'm'
(lp13805
I1
aI3
aI5
aI7
aI29
aI63
aI65
aI87
aI305
aI721
aI851
aI2235
aI4987
aI3051
aI23015
aI1281
aI15755
asS'd'
I6903
sa(dp13806
S'a'
I10040
sS's'
I17
sS'm'
(lp13807
I1
aI1
aI3
aI9
aI17
aI3
aI57
aI47
aI223
aI305
aI1409
aI235
aI4379
aI5779
aI27695
aI22535
aI9387
asS'd'
I6904
sa(dp13808
S'a'
I10051
sS's'
I17
sS'm'
(lp13809
I1
aI1
aI3
aI11
aI25
aI33
aI75
aI141
aI155
aI699
aI85
aI1729
aI2551
aI7101
aI7739
aI18025
aI100819
asS'd'
I6905
sa(dp13810
S'a'
I10057
sS's'
I17
sS'm'
(lp13811
I1
aI3
aI3
aI13
aI5
aI45
aI63
aI83
aI141
aI383
aI1931
aI3343
aI7397
aI4823
aI28893
aI41279
aI67805
asS'd'
I6906
sa(dp13812
S'a'
I10072
sS's'
I17
sS'm'
(lp13813
I1
aI3
aI5
aI7
aI19
aI29
aI97
aI67
aI177
aI583
aI1783
aI4007
aI5087
aI805
aI30999
aI23197
aI117553
asS'd'
I6907
sa(dp13814
S'a'
I10096
sS's'
I17
sS'm'
(lp13815
I1
aI3
aI5
aI1
aI25
aI41
aI33
aI109
aI511
aI449
aI653
aI995
aI5881
aI2163
aI13689
aI54385
aI97419
asS'd'
I6908
sa(dp13816
S'a'
I10102
sS's'
I17
sS'm'
(lp13817
I1
aI3
aI3
aI13
aI25
aI17
aI49
aI77
aI497
aI659
aI783
aI3513
aI3735
aI3541
aI573
aI50237
aI99247
asS'd'
I6909
sa(dp13818
S'a'
I10105
sS's'
I17
sS'm'
(lp13819
I1
aI3
aI1
aI7
aI17
aI13
aI37
aI169
aI19
aI965
aI289
aI455
aI6855
aI11233
aI7553
aI7007
aI57389
asS'd'
I6910
sa(dp13820
S'a'
I10115
sS's'
I17
sS'm'
(lp13821
I1
aI1
aI7
aI11
aI5
aI15
aI11
aI177
aI75
aI243
aI453
aI3861
aI3091
aI4625
aI12489
aI11537
aI74199
asS'd'
I6911
sa(dp13822
S'a'
I10124
sS's'
I17
sS'm'
(lp13823
I1
aI1
aI5
aI13
aI17
aI21
aI23
aI57
aI343
aI985
aI1755
aI3947
aI3899
aI11847
aI19321
aI62295
aI51265
asS'd'
I6912
sa(dp13824
S'a'
I10139
sS's'
I17
sS'm'
(lp13825
I1
aI1
aI3
aI9
aI19
aI37
aI31
aI243
aI433
aI725
aI535
aI3733
aI33
aI7885
aI1425
aI41919
aI66507
asS'd'
I6913
sa(dp13826
S'a'
I10145
sS's'
I17
sS'm'
(lp13827
I1
aI3
aI5
aI11
aI15
aI11
aI25
aI255
aI93
aI33
aI71
aI2389
aI1855
aI317
aI12773
aI13311
aI81927
asS'd'
I6914
sa(dp13828
S'a'
I10148
sS's'
I17
sS'm'
(lp13829
I1
aI3
aI1
aI3
aI7
aI55
aI21
aI175
aI357
aI235
aI1679
aI931
aI2051
aI14213
aI20539
aI38049
aI122513
asS'd'
I6915
sa(dp13830
S'a'
I10157
sS's'
I17
sS'm'
(lp13831
I1
aI1
aI5
aI15
aI5
aI51
aI127
aI79
aI297
aI135
aI1423
aI2783
aI7229
aI14451
aI27619
aI7299
aI49189
asS'd'
I6916
sa(dp13832
S'a'
I10158
sS's'
I17
sS'm'
(lp13833
I1
aI1
aI1
aI3
aI5
aI13
aI9
aI209
aI455
aI483
aI1745
aI323
aI789
aI7645
aI26373
aI61659
aI23671
asS'd'
I6917
sa(dp13834
S'a'
I10163
sS's'
I17
sS'm'
(lp13835
I1
aI1
aI1
aI9
aI23
aI63
aI99
aI91
aI377
aI275
aI275
aI3005
aI1563
aI5945
aI23825
aI33211
aI52753
asS'd'
I6918
sa(dp13836
S'a'
I10180
sS's'
I17
sS'm'
(lp13837
I1
aI1
aI1
aI1
aI31
aI55
aI31
aI109
aI481
aI581
aI771
aI197
aI6155
aI3465
aI8451
aI25925
aI41159
asS'd'
I6919
sa(dp13838
S'a'
I10187
sS's'
I17
sS'm'
(lp13839
I1
aI3
aI7
aI13
aI5
aI33
aI113
aI161
aI265
aI493
aI1723
aI513
aI5111
aI10177
aI21755
aI5321
aI58831
asS'd'
I6920
sa(dp13840
S'a'
I10198
sS's'
I17
sS'm'
(lp13841
I1
aI1
aI7
aI1
aI21
aI33
aI117
aI183
aI89
aI689
aI1253
aI2215
aI6565
aI3079
aI16343
aI22427
aI96447
asS'd'
I6921
sa(dp13842
S'a'
I10208
sS's'
I17
sS'm'
(lp13843
I1
aI1
aI1
aI5
aI15
aI61
aI5
aI139
aI111
aI463
aI573
aI1907
aI4615
aI14975
aI5715
aI51017
aI69827
asS'd'
I6922
sa(dp13844
S'a'
I10214
sS's'
I17
sS'm'
(lp13845
I1
aI1
aI1
aI13
aI3
aI3
aI117
aI249
aI25
aI361
aI1177
aI2901
aI1601
aI11381
aI18981
aI44811
aI47117
asS'd'
I6923
sa(dp13846
S'a'
I10220
sS's'
I17
sS'm'
(lp13847
I1
aI1
aI5
aI3
aI29
aI5
aI49
aI221
aI247
aI57
aI553
aI1889
aI479
aI15581
aI7035
aI7293
aI53065
asS'd'
I6924
sa(dp13848
S'a'
I10237
sS's'
I17
sS'm'
(lp13849
I1
aI3
aI3
aI3
aI15
aI49
aI91
aI187
aI213
aI981
aI1417
aI211
aI3719
aI13693
aI17671
aI16691
aI57147
asS'd'
I6925
sa(dp13850
S'a'
I10238
sS's'
I17
sS'm'
(lp13851
I1
aI1
aI7
aI9
aI7
aI17
aI109
aI185
aI459
aI769
aI1783
aI899
aI885
aI2291
aI30023
aI26315
aI7337
asS'd'
I6926
sa(dp13852
S'a'
I10241
sS's'
I17
sS'm'
(lp13853
I1
aI1
aI5
aI11
aI11
aI31
aI73
aI191
aI95
aI25
aI1953
aI1387
aI1077
aI7547
aI9661
aI57739
aI76799
asS'd'
I6927
sa(dp13854
S'a'
I10244
sS's'
I17
sS'm'
(lp13855
I1
aI1
aI7
aI13
aI23
aI41
aI69
aI177
aI407
aI699
aI1055
aI3653
aI1239
aI8113
aI12823
aI1803
aI117815
asS'd'
I6928
sa(dp13856
S'a'
I10251
sS's'
I17
sS'm'
(lp13857
I1
aI1
aI1
aI15
aI1
aI55
aI71
aI133
aI401
aI593
aI605
aI2855
aI4569
aI3533
aI14141
aI65457
aI125655
asS'd'
I6929
sa(dp13858
S'a'
I10253
sS's'
I17
sS'm'
(lp13859
I1
aI1
aI7
aI9
aI31
aI55
aI53
aI11
aI65
aI17
aI561
aI925
aI1561
aI8929
aI19859
aI57111
aI12777
asS'd'
I6930
sa(dp13860
S'a'
I10256
sS's'
I17
sS'm'
(lp13861
I1
aI3
aI3
aI11
aI7
aI59
aI125
aI205
aI473
aI655
aI1429
aI337
aI6829
aI7551
aI27873
aI11667
aI39231
asS'd'
I6931
sa(dp13862
S'a'
I10259
sS's'
I17
sS'm'
(lp13863
I1
aI3
aI3
aI9
aI13
aI23
aI25
aI161
aI443
aI545
aI1967
aI1895
aI6929
aI5975
aI17801
aI41769
aI30429
asS'd'
I6932
sa(dp13864
S'a'
I10266
sS's'
I17
sS'm'
(lp13865
I1
aI3
aI7
aI13
aI15
aI1
aI99
aI43
aI45
aI451
aI21
aI639
aI7121
aI4781
aI2813
aI419
aI17761
asS'd'
I6933
sa(dp13866
S'a'
I10284
sS's'
I17
sS'm'
(lp13867
I1
aI1
aI5
aI13
aI11
aI9
aI53
aI83
aI443
aI441
aI1601
aI3177
aI1913
aI12211
aI25835
aI1733
aI4793
asS'd'
I6934
sa(dp13868
S'a'
I10290
sS's'
I17
sS'm'
(lp13869
I1
aI3
aI3
aI1
aI13
aI15
aI11
aI187
aI471
aI699
aI1751
aI3279
aI2305
aI15259
aI31541
aI21357
aI73763
asS'd'
I6935
sa(dp13870
S'a'
I10331
sS's'
I17
sS'm'
(lp13871
I1
aI3
aI5
aI9
aI23
aI11
aI125
aI57
aI261
aI479
aI879
aI719
aI3221
aI2943
aI10593
aI11521
aI83979
asS'd'
I6936
sa(dp13872
S'a'
I10334
sS's'
I17
sS'm'
(lp13873
I1
aI3
aI7
aI13
aI3
aI39
aI119
aI135
aI85
aI417
aI1675
aI971
aI7577
aI12709
aI20407
aI26105
aI97021
asS'd'
I6937
sa(dp13874
S'a'
I10350
sS's'
I17
sS'm'
(lp13875
I1
aI1
aI5
aI11
aI15
aI63
aI83
aI141
aI281
aI663
aI1745
aI2775
aI5605
aI9127
aI553
aI7177
aI115969
asS'd'
I6938
sa(dp13876
S'a'
I10355
sS's'
I17
sS'm'
(lp13877
I1
aI1
aI7
aI1
aI19
aI47
aI7
aI165
aI87
aI95
aI361
aI1879
aI6351
aI2861
aI9103
aI37489
aI24525
asS'd'
I6939
sa(dp13878
S'a'
I10357
sS's'
I17
sS'm'
(lp13879
I1
aI3
aI3
aI11
aI9
aI21
aI51
aI149
aI375
aI967
aI1583
aI1427
aI1223
aI11611
aI7481
aI36619
aI128429
asS'd'
I6940
sa(dp13880
S'a'
I10367
sS's'
I17
sS'm'
(lp13881
I1
aI1
aI5
aI1
aI3
aI31
aI7
aI217
aI453
aI565
aI1517
aI2847
aI6937
aI1197
aI24339
aI44311
aI66843
asS'd'
I6941
sa(dp13882
S'a'
I10368
sS's'
I17
sS'm'
(lp13883
I1
aI1
aI5
aI3
aI3
aI17
aI127
aI59
aI3
aI905
aI531
aI1179
aI3559
aI5175
aI24627
aI60941
aI129457
asS'd'
I6942
sa(dp13884
S'a'
I10377
sS's'
I17
sS'm'
(lp13885
I1
aI1
aI1
aI7
aI15
aI15
aI1
aI31
aI373
aI643
aI279
aI3831
aI4881
aI9763
aI17641
aI43219
aI83109
asS'd'
I6943
sa(dp13886
S'a'
I10388
sS's'
I17
sS'm'
(lp13887
I1
aI3
aI3
aI9
aI5
aI21
aI41
aI71
aI371
aI201
aI573
aI1481
aI3631
aI10783
aI6679
aI1089
aI117347
asS'd'
I6944
sa(dp13888
S'a'
I10407
sS's'
I17
sS'm'
(lp13889
I1
aI1
aI7
aI7
aI5
aI25
aI73
aI63
aI173
aI197
aI147
aI981
aI1491
aI1597
aI11733
aI14285
aI74021
asS'd'
I6945
sa(dp13890
S'a'
I10421
sS's'
I17
sS'm'
(lp13891
I1
aI1
aI5
aI11
aI17
aI15
aI3
aI175
aI391
aI503
aI1745
aI319
aI791
aI5607
aI18173
aI37319
aI92025
asS'd'
I6946
sa(dp13892
S'a'
I10434
sS's'
I17
sS'm'
(lp13893
I1
aI3
aI1
aI1
aI9
aI37
aI43
aI81
aI439
aI951
aI805
aI251
aI4625
aI15617
aI13715
aI62263
aI3827
asS'd'
I6947
sa(dp13894
S'a'
I10439
sS's'
I17
sS'm'
(lp13895
I1
aI3
aI1
aI1
aI25
aI21
aI67
aI191
aI499
aI205
aI1355
aI105
aI1637
aI563
aI22291
aI9045
aI6545
asS'd'
I6948
sa(dp13896
S'a'
I10440
sS's'
I17
sS'm'
(lp13897
I1
aI1
aI5
aI5
aI9
aI3
aI75
aI75
aI287
aI303
aI1767
aI1789
aI3437
aI4637
aI9605
aI2537
aI64935
asS'd'
I6949
sa(dp13898
S'a'
I10443
sS's'
I17
sS'm'
(lp13899
I1
aI1
aI3
aI3
aI1
aI51
aI27
aI155
aI375
aI149
aI885
aI187
aI1551
aI13109
aI27011
aI57301
aI41047
asS'd'
I6950
sa(dp13900
S'a'
I10446
sS's'
I17
sS'm'
(lp13901
I1
aI1
aI7
aI5
aI21
aI23
aI1
aI81
aI163
aI231
aI2039
aI1519
aI1279
aI15379
aI25549
aI6711
aI81499
asS'd'
I6951
sa(dp13902
S'a'
I10457
sS's'
I17
sS'm'
(lp13903
I1
aI1
aI3
aI5
aI3
aI37
aI71
aI243
aI165
aI365
aI379
aI351
aI4649
aI4287
aI13395
aI30329
aI78383
asS'd'
I6952
sa(dp13904
S'a'
I10469
sS's'
I17
sS'm'
(lp13905
I1
aI3
aI1
aI1
aI25
aI63
aI27
aI215
aI223
aI699
aI2029
aI3737
aI5947
aI7287
aI20813
aI4931
aI19345
asS'd'
I6953
sa(dp13906
S'a'
I10476
sS's'
I17
sS'm'
(lp13907
I1
aI1
aI3
aI15
aI21
aI7
aI25
aI187
aI219
aI53
aI1749
aI1797
aI3533
aI14307
aI53
aI11095
aI75469
asS'd'
I6954
sa(dp13908
S'a'
I10479
sS's'
I17
sS'm'
(lp13909
I1
aI1
aI3
aI13
aI27
aI31
aI91
aI121
aI481
aI291
aI915
aI535
aI4291
aI5271
aI12181
aI55921
aI125917
asS'd'
I6955
sa(dp13910
S'a'
I10481
sS's'
I17
sS'm'
(lp13911
I1
aI3
aI1
aI1
aI3
aI29
aI21
aI251
aI361
aI747
aI997
aI2989
aI1809
aI7235
aI17855
aI31027
aI100689
asS'd'
I6956
sa(dp13912
S'a'
I10494
sS's'
I17
sS'm'
(lp13913
I1
aI3
aI7
aI1
aI21
aI13
aI49
aI93
aI183
aI673
aI881
aI1931
aI7009
aI2565
aI26021
aI53815
aI19807
asS'd'
I6957
sa(dp13914
S'a'
I10501
sS's'
I17
sS'm'
(lp13915
I1
aI1
aI7
aI13
aI9
aI23
aI47
aI237
aI487
aI843
aI1357
aI919
aI1753
aI903
aI2911
aI31527
aI73027
asS'd'
I6958
sa(dp13916
S'a'
I10505
sS's'
I17
sS'm'
(lp13917
I1
aI1
aI1
aI1
aI25
aI33
aI97
aI241
aI421
aI375
aI73
aI2541
aI6231
aI14659
aI15335
aI5915
aI110791
asS'd'
I6959
sa(dp13918
S'a'
I10516
sS's'
I17
sS'm'
(lp13919
I1
aI3
aI3
aI7
aI21
aI17
aI97
aI125
aI7
aI271
aI167
aI475
aI4887
aI1847
aI30173
aI25913
aI36659
asS'd'
I6960
sa(dp13920
S'a'
I10532
sS's'
I17
sS'm'
(lp13921
I1
aI1
aI3
aI15
aI15
aI37
aI67
aI5
aI463
aI423
aI823
aI941
aI1551
aI14175
aI15377
aI6017
aI118297
asS'd'
I6961
sa(dp13922
S'a'
I10541
sS's'
I17
sS'm'
(lp13923
I1
aI1
aI1
aI7
aI31
aI51
aI71
aI127
aI73
aI517
aI881
aI3205
aI6219
aI11213
aI14783
aI64275
aI70033
asS'd'
I6962
sa(dp13924
S'a'
I10547
sS's'
I17
sS'm'
(lp13925
I1
aI3
aI1
aI5
aI17
aI17
aI57
aI107
aI359
aI999
aI1415
aI757
aI4743
aI7775
aI14111
aI20075
aI73269
asS'd'
I6963
sa(dp13926
S'a'
I10550
sS's'
I17
sS'm'
(lp13927
I1
aI3
aI5
aI3
aI21
aI57
aI87
aI43
aI307
aI777
aI717
aI3329
aI4159
aI12545
aI31355
aI31329
aI41377
asS'd'
I6964
sa(dp13928
S'a'
I10591
sS's'
I17
sS'm'
(lp13929
I1
aI3
aI7
aI15
aI25
aI43
aI19
aI147
aI487
aI517
aI977
aI3625
aI2311
aI14173
aI15167
aI56563
aI110417
asS'd'
I6965
sa(dp13930
S'a'
I10597
sS's'
I17
sS'm'
(lp13931
I1
aI3
aI3
aI11
aI23
aI1
aI67
aI157
aI461
aI169
aI231
aI1977
aI5657
aI865
aI711
aI24213
aI76895
asS'd'
I6966
sa(dp13932
S'a'
I10602
sS's'
I17
sS'm'
(lp13933
I1
aI1
aI7
aI13
aI5
aI37
aI51
aI165
aI331
aI97
aI431
aI3819
aI1379
aI12083
aI27521
aI19689
aI100119
asS'd'
I6967
sa(dp13934
S'a'
I10610
sS's'
I17
sS'm'
(lp13935
I1
aI1
aI7
aI15
aI29
aI21
aI59
aI193
aI397
aI467
aI951
aI3037
aI2955
aI13235
aI20981
aI63865
aI30069
asS'd'
I6968
sa(dp13936
S'a'
I10619
sS's'
I17
sS'm'
(lp13937
I1
aI3
aI3
aI5
aI7
aI49
aI41
aI143
aI319
aI71
aI353
aI2159
aI3043
aI15317
aI24095
aI12017
aI64393
asS'd'
I6969
sa(dp13938
S'a'
I10631
sS's'
I17
sS'm'
(lp13939
I1
aI1
aI5
aI13
aI25
aI45
aI57
aI153
aI311
aI805
aI953
aI1763
aI5655
aI3961
aI12085
aI58761
aI76533
asS'd'
I6970
sa(dp13940
S'a'
I10646
sS's'
I17
sS'm'
(lp13941
I1
aI1
aI3
aI15
aI29
aI19
aI71
aI107
aI203
aI221
aI1173
aI1597
aI1179
aI9649
aI21659
aI10463
aI8195
asS'd'
I6971
sa(dp13942
S'a'
I10655
sS's'
I17
sS'm'
(lp13943
I1
aI1
aI3
aI9
aI31
aI29
aI53
aI151
aI247
aI577
aI543
aI459
aI8141
aI5613
aI12029
aI24199
aI118603
asS'd'
I6972
sa(dp13944
S'a'
I10665
sS's'
I17
sS'm'
(lp13945
I1
aI3
aI1
aI5
aI1
aI55
aI103
aI23
aI405
aI5
aI181
aI3805
aI1103
aI13389
aI6725
aI48733
aI99639
asS'd'
I6973
sa(dp13946
S'a'
I10673
sS's'
I17
sS'm'
(lp13947
I1
aI1
aI5
aI9
aI1
aI47
aI115
aI231
aI151
aI885
aI427
aI2849
aI361
aI12969
aI705
aI41711
aI53587
asS'd'
I6974
sa(dp13948
S'a'
I10674
sS's'
I17
sS'm'
(lp13949
I1
aI1
aI3
aI11
aI9
aI3
aI11
aI231
aI77
aI775
aI657
aI2721
aI3431
aI11919
aI10425
aI29405
aI91561
asS'd'
I6975
sa(dp13950
S'a'
I10680
sS's'
I17
sS'm'
(lp13951
I1
aI1
aI1
aI5
aI5
aI7
aI79
aI41
aI181
aI333
aI963
aI3117
aI7703
aI2259
aI16671
aI51139
aI27997
asS'd'
I6976
sa(dp13952
S'a'
I10693
sS's'
I17
sS'm'
(lp13953
I1
aI3
aI7
aI7
aI13
aI55
aI59
aI157
aI377
aI711
aI1475
aI1509
aI1375
aI6825
aI13729
aI28613
aI109199
asS'd'
I6977
sa(dp13954
S'a'
I10700
sS's'
I17
sS'm'
(lp13955
I1
aI3
aI3
aI3
aI13
aI11
aI51
aI1
aI67
aI609
aI467
aI2161
aI7693
aI9019
aI1847
aI27969
aI74863
asS'd'
I6978
sa(dp13956
S'a'
I10721
sS's'
I17
sS'm'
(lp13957
I1
aI1
aI3
aI3
aI11
aI33
aI87
aI217
aI239
aI505
aI1451
aI2801
aI1417
aI695
aI29883
aI15877
aI99969
asS'd'
I6979
sa(dp13958
S'a'
I10727
sS's'
I17
sS'm'
(lp13959
I1
aI3
aI3
aI5
aI3
aI61
aI9
aI171
aI57
aI547
aI2003
aI2335
aI2259
aI3205
aI5639
aI21721
aI25893
asS'd'
I6980
sa(dp13960
S'a'
I10746
sS's'
I17
sS'm'
(lp13961
I1
aI3
aI1
aI3
aI19
aI15
aI83
aI69
aI47
aI897
aI627
aI2839
aI7123
aI8567
aI14707
aI13159
aI125139
asS'd'
I6981
sa(dp13962
S'a'
I10748
sS's'
I17
sS'm'
(lp13963
I1
aI3
aI7
aI11
aI1
aI59
aI53
aI33
aI135
aI1009
aI1829
aI3011
aI1245
aI421
aI28909
aI45517
aI55071
asS'd'
I6982
sa(dp13964
S'a'
I10757
sS's'
I17
sS'm'
(lp13965
I1
aI1
aI5
aI9
aI3
aI27
aI11
aI243
aI235
aI683
aI1329
aI3145
aI2141
aI14027
aI3707
aI5933
aI51965
asS'd'
I6983
sa(dp13966
S'a'
I10761
sS's'
I17
sS'm'
(lp13967
I1
aI1
aI5
aI7
aI13
aI63
aI79
aI105
aI27
aI195
aI1657
aI3107
aI1245
aI1681
aI29619
aI10589
aI78197
asS'd'
I6984
sa(dp13968
S'a'
I10770
sS's'
I17
sS'm'
(lp13969
I1
aI3
aI3
aI7
aI21
aI1
aI5
aI79
aI73
aI125
aI1587
aI3053
aI5977
aI10745
aI28343
aI39023
aI56201
asS'd'
I6985
sa(dp13970
S'a'
I10776
sS's'
I17
sS'm'
(lp13971
I1
aI1
aI3
aI15
aI23
aI21
aI39
aI41
aI173
aI913
aI1267
aI1323
aI2967
aI1979
aI16763
aI53753
aI21905
asS'd'
I6986
sa(dp13972
S'a'
I10782
sS's'
I17
sS'm'
(lp13973
I1
aI1
aI5
aI7
aI11
aI11
aI117
aI151
aI409
aI345
aI1461
aI1703
aI687
aI557
aI31651
aI35507
aI54909
asS'd'
I6987
sa(dp13974
S'a'
I10791
sS's'
I17
sS'm'
(lp13975
I1
aI1
aI1
aI15
aI15
aI49
aI55
aI223
aI289
aI765
aI1737
aI1117
aI3717
aI15465
aI31949
aI55061
aI97091
asS'd'
I6988
sa(dp13976
S'a'
I10792
sS's'
I17
sS'm'
(lp13977
I1
aI1
aI5
aI9
aI21
aI29
aI99
aI13
aI119
aI35
aI1461
aI61
aI5155
aI6785
aI15957
aI11295
aI52203
asS'd'
I6989
sa(dp13978
S'a'
I10805
sS's'
I17
sS'm'
(lp13979
I1
aI3
aI5
aI7
aI23
aI39
aI73
aI161
aI465
aI715
aI153
aI3529
aI2243
aI13773
aI16573
aI26233
aI130263
asS'd'
I6990
sa(dp13980
S'a'
I10810
sS's'
I17
sS'm'
(lp13981
I1
aI3
aI7
aI9
aI11
aI51
aI5
aI149
aI501
aI119
aI2047
aI3417
aI3955
aI15055
aI31633
aI473
aI127305
asS'd'
I6991
sa(dp13982
S'a'
I10832
sS's'
I17
sS'm'
(lp13983
I1
aI1
aI1
aI9
aI31
aI57
aI91
aI119
aI215
aI11
aI1013
aI3969
aI1285
aI11521
aI8039
aI36737
aI86365
asS'd'
I6992
sa(dp13984
S'a'
I10835
sS's'
I17
sS'm'
(lp13985
I1
aI1
aI5
aI3
aI7
aI17
aI9
aI27
aI59
aI883
aI541
aI3027
aI6219
aI1091
aI2453
aI38247
aI21323
asS'd'
I6993
sa(dp13986
S'a'
I10841
sS's'
I17
sS'm'
(lp13987
I1
aI1
aI1
aI1
aI25
aI39
aI55
aI249
aI61
aI313
aI467
aI1763
aI4067
aI8367
aI32431
aI44463
aI66439
asS'd'
I6994
sa(dp13988
S'a'
I10842
sS's'
I17
sS'm'
(lp13989
I1
aI3
aI3
aI1
aI13
aI3
aI37
aI209
aI21
aI653
aI1971
aI3649
aI6165
aI3789
aI12793
aI56327
aI60351
asS'd'
I6995
sa(dp13990
S'a'
I10847
sS's'
I17
sS'm'
(lp13991
I1
aI1
aI7
aI9
aI31
aI33
aI21
aI51
aI313
aI631
aI515
aI1761
aI4149
aI2601
aI12481
aI25975
aI94061
asS'd'
I6996
sa(dp13992
S'a'
I10853
sS's'
I17
sS'm'
(lp13993
I1
aI1
aI7
aI15
aI3
aI7
aI55
aI129
aI297
aI735
aI779
aI633
aI3265
aI11713
aI3893
aI61197
aI113991
asS'd'
I6997
sa(dp13994
S'a'
I10860
sS's'
I17
sS'm'
(lp13995
I1
aI3
aI5
aI13
aI1
aI15
aI27
aI253
aI435
aI595
aI1163
aI2753
aI7399
aI15225
aI26215
aI59753
aI74933
asS'd'
I6998
sa(dp13996
S'a'
I10871
sS's'
I17
sS'm'
(lp13997
I1
aI1
aI7
aI7
aI15
aI23
aI111
aI43
aI467
aI957
aI1687
aI2893
aI2315
aI2025
aI1475
aI9061
aI101611
asS'd'
I6999
sa(dp13998
S'a'
I10878
sS's'
I17
sS'm'
(lp13999
I1
aI1
aI3
aI3
aI29
aI41
aI53
aI169
aI125
aI415
aI361
aI869
aI3399
aI8821
aI18193
aI38575
aI73979
asS'd'
I7000
sa(dp14000
S'a'
I10881
sS's'
I17
sS'm'
(lp14001
I1
aI1
aI1
aI15
aI3
aI5
aI27
aI5
aI293
aI765
aI1809
aI1961
aI955
aI12441
aI10915
aI2363
aI49617
asS'd'
I7001
sa(dp14002
S'a'
I10888
sS's'
I17
sS'm'
(lp14003
I1
aI1
aI5
aI15
aI19
aI11
aI3
aI91
aI59
aI323
aI545
aI1177
aI7967
aI2729
aI14085
aI3283
aI79859
asS'd'
I7002
sa(dp14004
S'a'
I10894
sS's'
I17
sS'm'
(lp14005
I1
aI1
aI7
aI13
aI11
aI17
aI29
aI163
aI295
aI951
aI311
aI3471
aI1339
aI10719
aI701
aI32377
aI41685
asS'd'
I7003
sa(dp14006
S'a'
I10901
sS's'
I17
sS'm'
(lp14007
I1
aI3
aI5
aI7
aI21
aI19
aI81
aI247
aI495
aI767
aI251
aI3455
aI6383
aI7221
aI19943
aI64865
aI33193
asS'd'
I7004
sa(dp14008
S'a'
I10915
sS's'
I17
sS'm'
(lp14009
I1
aI1
aI7
aI15
aI23
aI41
aI63
aI195
aI311
aI619
aI211
aI743
aI889
aI7627
aI12527
aI15865
aI40103
asS'd'
I7005
sa(dp14010
S'a'
I10918
sS's'
I17
sS'm'
(lp14011
I1
aI1
aI3
aI1
aI23
aI23
aI57
aI221
aI153
aI27
aI939
aI3949
aI411
aI6357
aI31985
aI939
aI91001
asS'd'
I7006
sa(dp14012
S'a'
I10922
sS's'
I17
sS'm'
(lp14013
I1
aI3
aI5
aI15
aI7
aI5
aI35
aI135
aI245
aI921
aI307
aI823
aI775
aI4891
aI24575
aI53503
aI48147
asS'd'
I7007
sa(dp14014
S'a'
I10936
sS's'
I17
sS'm'
(lp14015
I1
aI1
aI5
aI7
aI9
aI31
aI23
aI139
aI477
aI495
aI287
aI807
aI1855
aI8321
aI13963
aI52197
aI78509
asS'd'
I7008
sa(dp14016
S'a'
I10954
sS's'
I17
sS'm'
(lp14017
I1
aI3
aI3
aI3
aI29
aI59
aI33
aI83
aI211
aI65
aI623
aI1269
aI1745
aI16383
aI10759
aI57199
aI14035
asS'd'
I7009
sa(dp14018
S'a'
I10968
sS's'
I17
sS'm'
(lp14019
I1
aI3
aI3
aI15
aI25
aI55
aI69
aI171
aI411
aI937
aI731
aI2275
aI2597
aI4133
aI5089
aI50507
aI39989
asS'd'
I7010
sa(dp14020
S'a'
I10971
sS's'
I17
sS'm'
(lp14021
I1
aI3
aI1
aI9
aI5
aI47
aI51
aI21
aI171
aI913
aI233
aI43
aI2673
aI471
aI27077
aI57039
aI32579
asS'd'
I7011
sa(dp14022
S'a'
I10973
sS's'
I17
sS'm'
(lp14023
I1
aI3
aI5
aI3
aI29
aI35
aI5
aI105
aI233
aI379
aI77
aI1775
aI2409
aI4597
aI19879
aI12691
aI49739
asS'd'
I7012
sa(dp14024
S'a'
I10978
sS's'
I17
sS'm'
(lp14025
I1
aI3
aI7
aI13
aI17
aI29
aI117
aI177
aI163
aI927
aI45
aI3227
aI7263
aI5551
aI9219
aI32101
aI122473
asS'd'
I7013
sa(dp14026
S'a'
I10998
sS's'
I17
sS'm'
(lp14027
I1
aI1
aI7
aI5
aI31
aI39
aI75
aI147
aI311
aI991
aI1431
aI3821
aI6891
aI9637
aI17887
aI661
aI23067
asS'd'
I7014
sa(dp14028
S'a'
I11009
sS's'
I17
sS'm'
(lp14029
I1
aI3
aI5
aI13
aI31
aI53
aI69
aI79
aI153
aI329
aI207
aI479
aI2395
aI6505
aI29553
aI52023
aI31531
asS'd'
I7015
sa(dp14030
S'a'
I11021
sS's'
I17
sS'm'
(lp14031
I1
aI3
aI1
aI7
aI15
aI7
aI87
aI233
aI25
aI275
aI981
aI1207
aI3083
aI16349
aI30185
aI60611
aI120607
asS'd'
I7016
sa(dp14032
S'a'
I11029
sS's'
I17
sS'm'
(lp14033
I1
aI1
aI5
aI3
aI21
aI7
aI47
aI173
aI291
aI965
aI65
aI545
aI7465
aI4471
aI2249
aI34281
aI107217
asS'd'
I7017
sa(dp14034
S'a'
I11030
sS's'
I17
sS'm'
(lp14035
I1
aI1
aI3
aI11
aI19
aI53
aI17
aI243
aI193
aI297
aI1937
aI1513
aI4979
aI14867
aI15497
aI10049
aI9135
asS'd'
I7018
sa(dp14036
S'a'
I11034
sS's'
I17
sS'm'
(lp14037
I1
aI3
aI1
aI3
aI25
aI39
aI29
aI63
aI231
aI145
aI247
aI1745
aI3439
aI8635
aI26687
aI18595
aI67123
asS'd'
I7019
sa(dp14038
S'a'
I11050
sS's'
I17
sS'm'
(lp14039
I1
aI1
aI7
aI9
aI29
aI33
aI89
aI175
aI429
aI675
aI891
aI1739
aI3567
aI5453
aI30427
aI33671
aI83395
asS'd'
I7020
sa(dp14040
S'a'
I11063
sS's'
I17
sS'm'
(lp14041
I1
aI3
aI1
aI5
aI31
aI25
aI69
aI237
aI235
aI307
aI1217
aI3805
aI153
aI13387
aI6209
aI14179
aI128725
asS'd'
I7021
sa(dp14042
S'a'
I11064
sS's'
I17
sS'm'
(lp14043
I1
aI1
aI3
aI3
aI19
aI45
aI117
aI135
aI67
aI601
aI369
aI3369
aI5505
aI2049
aI24099
aI22515
aI96575
asS'd'
I7022
sa(dp14044
S'a'
I11077
sS's'
I17
sS'm'
(lp14045
I1
aI1
aI1
aI3
aI3
aI45
aI29
aI255
aI327
aI77
aI1103
aI4067
aI2875
aI6487
aI5903
aI26625
aI19631
asS'd'
I7023
sa(dp14046
S'a'
I11078
sS's'
I17
sS'm'
(lp14047
I1
aI3
aI5
aI1
aI31
aI63
aI115
aI7
aI255
aI855
aI913
aI1779
aI7001
aI14387
aI26765
aI51987
aI3191
asS'd'
I7024
sa(dp14048
S'a'
I11105
sS's'
I17
sS'm'
(lp14049
I1
aI1
aI3
aI11
aI15
aI43
aI71
aI247
aI303
aI231
aI445
aI3963
aI3699
aI11851
aI18941
aI43465
aI63431
asS'd'
I7025
sa(dp14050
S'a'
I11106
sS's'
I17
sS'm'
(lp14051
I1
aI1
aI3
aI5
aI31
aI33
aI93
aI127
aI267
aI399
aI653
aI1997
aI5005
aI14535
aI4813
aI64065
aI2159
asS'd'
I7026
sa(dp14052
S'a'
I11126
sS's'
I17
sS'm'
(lp14053
I1
aI3
aI7
aI13
aI31
aI39
aI61
aI155
aI141
aI515
aI1217
aI161
aI4309
aI3697
aI22445
aI43599
aI43329
asS'd'
I7027
sa(dp14054
S'a'
I11129
sS's'
I17
sS'm'
(lp14055
I1
aI3
aI3
aI3
aI7
aI51
aI103
aI147
aI511
aI971
aI195
aI3731
aI6629
aI12125
aI12053
aI34951
aI60059
asS'd'
I7028
sa(dp14056
S'a'
I11135
sS's'
I17
sS'm'
(lp14057
I1
aI1
aI5
aI11
aI21
aI49
aI99
aI31
aI55
aI309
aI1805
aI2253
aI7095
aI15265
aI28445
aI54813
aI48615
asS'd'
I7029
sa(dp14058
S'a'
I11151
sS's'
I17
sS'm'
(lp14059
I1
aI3
aI1
aI15
aI9
aI41
aI61
aI125
aI65
aI143
aI1567
aI3259
aI6757
aI653
aI31601
aI63127
aI52179
asS'd'
I7030
sa(dp14060
S'a'
I11159
sS's'
I17
sS'm'
(lp14061
I1
aI1
aI5
aI3
aI29
aI5
aI19
aI197
aI153
aI447
aI7
aI1713
aI469
aI6043
aI1259
aI63641
aI29171
asS'd'
I7031
sa(dp14062
S'a'
I11165
sS's'
I17
sS'm'
(lp14063
I1
aI3
aI3
aI7
aI3
aI41
aI95
aI245
aI445
aI15
aI607
aI565
aI2361
aI2673
aI21077
aI20153
aI6199
asS'd'
I7032
sa(dp14064
S'a'
I11176
sS's'
I17
sS'm'
(lp14065
I1
aI1
aI5
aI1
aI5
aI59
aI93
aI127
aI485
aI663
aI683
aI635
aI1599
aI16377
aI31819
aI6539
aI27789
asS'd'
I7033
sa(dp14066
S'a'
I11179
sS's'
I17
sS'm'
(lp14067
I1
aI3
aI1
aI3
aI31
aI3
aI11
aI215
aI441
aI1005
aI1815
aI3945
aI5109
aI5539
aI23935
aI62671
aI90731
asS'd'
I7034
sa(dp14068
S'a'
I11182
sS's'
I17
sS'm'
(lp14069
I1
aI3
aI3
aI1
aI13
aI47
aI19
aI229
aI191
aI427
aI1141
aI2321
aI7105
aI1587
aI26347
aI63265
aI23377
asS'd'
I7035
sa(dp14070
S'a'
I11189
sS's'
I17
sS'm'
(lp14071
I1
aI1
aI5
aI15
aI31
aI55
aI61
aI93
aI89
aI945
aI1203
aI3631
aI4457
aI15097
aI32019
aI41747
aI46009
asS'd'
I7036
sa(dp14072
S'a'
I11204
sS's'
I17
sS'm'
(lp14073
I1
aI1
aI5
aI13
aI5
aI33
aI69
aI59
aI93
aI247
aI503
aI421
aI1923
aI9855
aI9825
aI14257
aI98663
asS'd'
I7037
sa(dp14074
S'a'
I11216
sS's'
I17
sS'm'
(lp14075
I1
aI3
aI1
aI13
aI27
aI21
aI91
aI39
aI131
aI571
aI1527
aI2715
aI2061
aI627
aI19705
aI47165
aI84345
asS'd'
I7038
sa(dp14076
S'a'
I11232
sS's'
I17
sS'm'
(lp14077
I1
aI1
aI1
aI7
aI3
aI3
aI7
aI251
aI225
aI959
aI1017
aI2423
aI6163
aI1549
aI7473
aI3193
aI104259
asS'd'
I7039
sa(dp14078
S'a'
I11235
sS's'
I17
sS'm'
(lp14079
I1
aI3
aI3
aI1
aI5
aI5
aI115
aI221
aI505
aI649
aI1525
aI2459
aI167
aI1899
aI23939
aI29253
aI122835
asS'd'
I7040
sa(dp14080
S'a'
I11242
sS's'
I17
sS'm'
(lp14081
I1
aI3
aI1
aI5
aI15
aI9
aI123
aI221
aI133
aI43
aI31
aI1211
aI4737
aI5001
aI20065
aI6369
aI93865
asS'd'
I7041
sa(dp14082
S'a'
I11250
sS's'
I17
sS'm'
(lp14083
I1
aI1
aI5
aI11
aI11
aI5
aI23
aI29
aI333
aI133
aI1469
aI1895
aI5879
aI15599
aI2131
aI25005
aI96271
asS'd'
I7042
sa(dp14084
S'a'
I11264
sS's'
I17
sS'm'
(lp14085
I1
aI1
aI3
aI11
aI25
aI11
aI19
aI57
aI397
aI645
aI1233
aI2433
aI6371
aI10577
aI15489
aI60709
aI3957
asS'd'
I7043
sa(dp14086
S'a'
I11273
sS's'
I17
sS'm'
(lp14087
I1
aI3
aI1
aI1
aI19
aI3
aI33
aI131
aI429
aI835
aI1363
aI2213
aI3185
aI14385
aI8831
aI43159
aI32975
asS'd'
I7044
sa(dp14088
S'a'
I11282
sS's'
I17
sS'm'
(lp14089
I1
aI1
aI5
aI5
aI23
aI11
aI127
aI139
aI213
aI259
aI897
aI1913
aI5737
aI1287
aI26617
aI4885
aI30193
asS'd'
I7045
sa(dp14090
S'a'
I11288
sS's'
I17
sS'm'
(lp14091
I1
aI3
aI5
aI13
aI3
aI27
aI99
aI31
aI11
aI27
aI1003
aI2473
aI7055
aI12923
aI4269
aI41433
aI90637
asS'd'
I7046
sa(dp14092
S'a'
I11291
sS's'
I17
sS'm'
(lp14093
I1
aI3
aI1
aI7
aI17
aI25
aI95
aI151
aI199
aI237
aI207
aI1879
aI2943
aI9845
aI3765
aI53533
aI111191
asS'd'
I7047
sa(dp14094
S'a'
I11293
sS's'
I17
sS'm'
(lp14095
I1
aI3
aI1
aI9
aI19
aI27
aI5
aI249
aI85
aI185
aI1883
aI1401
aI2041
aI12721
aI20593
aI30993
aI2601
asS'd'
I7048
sa(dp14096
S'a'
I11318
sS's'
I17
sS'm'
(lp14097
I1
aI3
aI3
aI9
aI23
aI1
aI15
aI133
aI387
aI779
aI707
aI2723
aI4485
aI989
aI27125
aI37295
aI125319
asS'd'
I7049
sa(dp14098
S'a'
I11321
sS's'
I17
sS'm'
(lp14099
I1
aI1
aI7
aI3
aI9
aI41
aI81
aI151
aI349
aI941
aI357
aI3817
aI7123
aI10079
aI27519
aI107
aI102281
asS'd'
I7050
sa(dp14100
S'a'
I11329
sS's'
I17
sS'm'
(lp14101
I1
aI1
aI1
aI1
aI13
aI5
aI111
aI167
aI73
aI85
aI1185
aI1213
aI333
aI153
aI13101
aI38087
aI39389
asS'd'
I7051
sa(dp14102
S'a'
I11336
sS's'
I17
sS'm'
(lp14103
I1
aI3
aI3
aI15
aI11
aI41
aI99
aI231
aI377
aI539
aI1335
aI1059
aI5373
aI9611
aI27927
aI29801
aI85749
asS'd'
I7052
sa(dp14104
S'a'
I11339
sS's'
I17
sS'm'
(lp14105
I1
aI3
aI1
aI9
aI19
aI37
aI125
aI27
aI15
aI699
aI1867
aI2711
aI1589
aI1675
aI32007
aI61339
aI96919
asS'd'
I7053
sa(dp14106
S'a'
I11342
sS's'
I17
sS'm'
(lp14107
I1
aI3
aI3
aI3
aI3
aI27
aI21
aI159
aI249
aI783
aI1517
aI2923
aI2609
aI1207
aI13705
aI57371
aI43603
asS'd'
I7054
sa(dp14108
S'a'
I11356
sS's'
I17
sS'm'
(lp14109
I1
aI1
aI5
aI15
aI17
aI55
aI77
aI1
aI401
aI897
aI987
aI345
aI5283
aI5827
aI17755
aI44371
aI13253
asS'd'
I7055
sa(dp14110
S'a'
I11383
sS's'
I17
sS'm'
(lp14111
I1
aI3
aI1
aI7
aI3
aI3
aI99
aI237
aI487
aI405
aI771
aI3503
aI1199
aI4779
aI26893
aI45821
aI46383
asS'd'
I7056
sa(dp14112
S'a'
I11389
sS's'
I17
sS'm'
(lp14113
I1
aI1
aI7
aI3
aI9
aI47
aI81
aI27
aI459
aI989
aI1891
aI3997
aI4081
aI4075
aI15079
aI65081
aI125185
asS'd'
I7057
sa(dp14114
S'a'
I11405
sS's'
I17
sS'm'
(lp14115
I1
aI3
aI5
aI9
aI25
aI23
aI71
aI251
aI251
aI197
aI353
aI3553
aI2165
aI2953
aI3733
aI52369
aI100641
asS'd'
I7058
sa(dp14116
S'a'
I11411
sS's'
I17
sS'm'
(lp14117
I1
aI1
aI1
aI5
aI25
aI43
aI63
aI187
aI495
aI345
aI1547
aI2293
aI7327
aI7797
aI14001
aI61865
aI40329
asS'd'
I7059
sa(dp14118
S'a'
I11423
sS's'
I17
sS'm'
(lp14119
I1
aI1
aI5
aI15
aI25
aI37
aI67
aI23
aI315
aI801
aI71
aI1235
aI7293
aI7207
aI30929
aI9417
aI94735
asS'd'
I7060
sa(dp14120
S'a'
I11424
sS's'
I17
sS'm'
(lp14121
I1
aI1
aI1
aI3
aI23
aI29
aI87
aI171
aI337
aI457
aI1597
aI3933
aI4151
aI1237
aI19563
aI56997
aI81497
asS'd'
I7061
sa(dp14122
S'a'
I11430
sS's'
I17
sS'm'
(lp14123
I1
aI3
aI3
aI11
aI3
aI33
aI79
aI239
aI277
aI611
aI205
aI2283
aI7459
aI425
aI21999
aI26491
aI58681
asS'd'
I7062
sa(dp14124
S'a'
I11444
sS's'
I17
sS'm'
(lp14125
I1
aI1
aI7
aI1
aI5
aI37
aI53
aI93
aI205
aI97
aI779
aI3623
aI7777
aI521
aI21915
aI46539
aI128811
asS'd'
I7063
sa(dp14126
S'a'
I11447
sS's'
I17
sS'm'
(lp14127
I1
aI1
aI5
aI7
aI19
aI7
aI39
aI183
aI299
aI193
aI1351
aI3867
aI5709
aI11655
aI1231
aI15555
aI128023
asS'd'
I7064
sa(dp14128
S'a'
I11459
sS's'
I17
sS'm'
(lp14129
I1
aI3
aI7
aI11
aI31
aI13
aI113
aI57
aI197
aI841
aI921
aI2087
aI2195
aI8279
aI8353
aI1955
aI22121
asS'd'
I7065
sa(dp14130
S'a'
I11465
sS's'
I17
sS'm'
(lp14131
I1
aI3
aI3
aI11
aI21
aI55
aI61
aI105
aI357
aI747
aI363
aI3511
aI2547
aI16283
aI25747
aI56041
aI33695
asS'd'
I7066
sa(dp14132
S'a'
I11473
sS's'
I17
sS'm'
(lp14133
I1
aI3
aI3
aI13
aI27
aI13
aI5
aI27
aI93
aI691
aI1869
aI2331
aI3131
aI14411
aI2355
aI37195
aI129273
asS'd'
I7067
sa(dp14134
S'a'
I11479
sS's'
I17
sS'm'
(lp14135
I1
aI3
aI3
aI7
aI27
aI9
aI11
aI165
aI435
aI811
aI215
aI1617
aI347
aI4289
aI29373
aI15749
aI91445
asS'd'
I7068
sa(dp14136
S'a'
I11501
sS's'
I17
sS'm'
(lp14137
I1
aI1
aI7
aI13
aI29
aI3
aI95
aI53
aI457
aI633
aI959
aI3705
aI7461
aI9307
aI21963
aI51599
aI6751
asS'd'
I7069
sa(dp14138
S'a'
I11510
sS's'
I17
sS'm'
(lp14139
I1
aI1
aI1
aI15
aI29
aI25
aI95
aI1
aI125
aI61
aI683
aI2067
aI6485
aI9095
aI5571
aI61281
aI70865
asS'd'
I7070
sa(dp14140
S'a'
I11514
sS's'
I17
sS'm'
(lp14141
I1
aI1
aI7
aI7
aI1
aI35
aI119
aI107
aI247
aI991
aI237
aI1865
aI3961
aI12583
aI11417
aI14913
aI90897
asS'd'
I7071
sa(dp14142
S'a'
I11522
sS's'
I17
sS'm'
(lp14143
I1
aI3
aI7
aI15
aI11
aI51
aI73
aI193
aI289
aI381
aI1767
aI3803
aI3197
aI3797
aI15059
aI19393
aI98947
asS'd'
I7072
sa(dp14144
S'a'
I11527
sS's'
I17
sS'm'
(lp14145
I1
aI1
aI5
aI3
aI13
aI7
aI91
aI223
aI347
aI59
aI1721
aI1501
aI6391
aI4141
aI14495
aI47283
aI47237
asS'd'
I7073
sa(dp14146
S'a'
I11533
sS's'
I17
sS'm'
(lp14147
I1
aI3
aI7
aI11
aI17
aI39
aI43
aI247
aI35
aI423
aI1859
aI3199
aI5343
aI7061
aI8609
aI6819
aI88575
asS'd'
I7074
sa(dp14148
S'a'
I11536
sS's'
I17
sS'm'
(lp14149
I1
aI1
aI5
aI13
aI31
aI27
aI57
aI19
aI499
aI1007
aI1965
aI795
aI1231
aI12755
aI24631
aI53343
aI82305
asS'd'
I7075
sa(dp14150
S'a'
I11548
sS's'
I17
sS'm'
(lp14151
I1
aI1
aI1
aI9
aI13
aI23
aI127
aI161
aI245
aI467
aI2025
aI2545
aI3085
aI13035
aI27087
aI14461
aI35971
asS'd'
I7076
sa(dp14152
S'a'
I11551
sS's'
I17
sS'm'
(lp14153
I1
aI3
aI5
aI1
aI7
aI3
aI99
aI159
aI75
aI341
aI1755
aI2337
aI5981
aI5055
aI19445
aI30043
aI61427
asS'd'
I7077
sa(dp14154
S'a'
I11552
sS's'
I17
sS'm'
(lp14155
I1
aI1
aI1
aI7
aI13
aI33
aI41
aI73
aI267
aI21
aI961
aI3509
aI6839
aI13215
aI8471
aI46735
aI93071
asS'd'
I7078
sa(dp14156
S'a'
I11555
sS's'
I17
sS'm'
(lp14157
I1
aI3
aI7
aI7
aI3
aI25
aI81
aI239
aI357
aI445
aI1483
aI389
aI3891
aI5131
aI21357
aI34757
aI111063
asS'd'
I7079
sa(dp14158
S'a'
I11572
sS's'
I17
sS'm'
(lp14159
I1
aI3
aI7
aI1
aI1
aI37
aI119
aI121
aI195
aI935
aI1711
aI2049
aI7001
aI7117
aI9957
aI7309
aI102293
asS'd'
I7080
sa(dp14160
S'a'
I11587
sS's'
I17
sS'm'
(lp14161
I1
aI1
aI7
aI11
aI1
aI49
aI107
aI95
aI149
aI329
aI289
aI1121
aI7217
aI15091
aI19071
aI13801
aI13
asS'd'
I7081
sa(dp14162
S'a'
I11601
sS's'
I17
sS'm'
(lp14163
I1
aI1
aI1
aI13
aI17
aI17
aI7
aI105
aI81
aI1017
aI1867
aI1567
aI5133
aI7325
aI19797
aI16301
aI40471
asS'd'
I7082
sa(dp14164
S'a'
I11613
sS's'
I17
sS'm'
(lp14165
I1
aI3
aI5
aI5
aI27
aI45
aI117
aI135
aI499
aI53
aI973
aI121
aI53
aI8771
aI11893
aI35827
aI57691
asS'd'
I7083
sa(dp14166
S'a'
I11614
sS's'
I17
sS'm'
(lp14167
I1
aI1
aI1
aI1
aI7
aI23
aI11
aI163
aI17
aI871
aI129
aI2959
aI5583
aI12253
aI1419
aI28367
aI32539
asS'd'
I7084
sa(dp14168
S'a'
I11618
sS's'
I17
sS'm'
(lp14169
I1
aI1
aI3
aI5
aI23
aI31
aI127
aI33
aI115
aI799
aI331
aI1873
aI1729
aI1383
aI23601
aI51145
aI72027
asS'd'
I7085
sa(dp14170
S'a'
I11624
sS's'
I17
sS'm'
(lp14171
I1
aI1
aI1
aI9
aI15
aI49
aI105
aI163
aI51
aI539
aI451
aI3983
aI6509
aI1073
aI30757
aI13971
aI51371
asS'd'
I7086
sa(dp14172
S'a'
I11630
sS's'
I17
sS'm'
(lp14173
I1
aI1
aI7
aI1
aI1
aI57
aI71
aI135
aI5
aI171
aI983
aI951
aI777
aI9257
aI3607
aI3239
aI76237
asS'd'
I7087
sa(dp14174
S'a'
I11663
sS's'
I17
sS'm'
(lp14175
I1
aI1
aI7
aI7
aI21
aI17
aI49
aI175
aI9
aI807
aI289
aI2777
aI7309
aI14911
aI28349
aI43871
aI96019
asS'd'
I7088
sa(dp14176
S'a'
I11682
sS's'
I17
sS'm'
(lp14177
I1
aI3
aI1
aI13
aI5
aI7
aI83
aI215
aI297
aI319
aI347
aI633
aI7285
aI8293
aI18811
aI31065
aI114077
asS'd'
I7089
sa(dp14178
S'a'
I11684
sS's'
I17
sS'm'
(lp14179
I1
aI3
aI1
aI11
aI3
aI29
aI91
aI231
aI161
aI601
aI355
aI2719
aI2941
aI6065
aI21849
aI58051
aI46515
asS'd'
I7090
sa(dp14180
S'a'
I11702
sS's'
I17
sS'm'
(lp14181
I1
aI1
aI3
aI9
aI25
aI41
aI111
aI135
aI71
aI755
aI29
aI131
aI1339
aI5053
aI15713
aI14557
aI106777
asS'd'
I7091
sa(dp14182
S'a'
I11705
sS's'
I17
sS'm'
(lp14183
I1
aI1
aI7
aI13
aI21
aI59
aI13
aI45
aI503
aI71
aI1611
aI4021
aI2359
aI11653
aI7261
aI14537
aI33031
asS'd'
I7092
sa(dp14184
S'a'
I11713
sS's'
I17
sS'm'
(lp14185
I1
aI1
aI1
aI11
aI5
aI31
aI1
aI181
aI37
aI527
aI1345
aI1979
aI4899
aI3289
aI25181
aI49959
aI44609
asS'd'
I7093
sa(dp14186
S'a'
I11731
sS's'
I17
sS'm'
(lp14187
I1
aI3
aI3
aI13
aI21
aI25
aI33
aI105
aI57
aI637
aI841
aI1595
aI3881
aI5053
aI9441
aI58717
aI127255
asS'd'
I7094
sa(dp14188
S'a'
I11734
sS's'
I17
sS'm'
(lp14189
I1
aI3
aI5
aI7
aI23
aI57
aI9
aI117
aI281
aI769
aI1573
aI2857
aI1139
aI6413
aI14001
aI21097
aI55215
asS'd'
I7095
sa(dp14190
S'a'
I11740
sS's'
I17
sS'm'
(lp14191
I1
aI1
aI7
aI7
aI3
aI5
aI75
aI149
aI269
aI353
aI437
aI61
aI2451
aI11987
aI17243
aI5649
aI105107
asS'd'
I7096
sa(dp14192
S'a'
I11747
sS's'
I17
sS'm'
(lp14193
I1
aI1
aI3
aI3
aI25
aI61
aI53
aI21
aI113
aI57
aI1415
aI2825
aI11
aI14977
aI6159
aI4181
aI96765
asS'd'
I7097
sa(dp14194
S'a'
I11754
sS's'
I17
sS'm'
(lp14195
I1
aI1
aI7
aI5
aI15
aI25
aI121
aI159
aI71
aI773
aI601
aI147
aI6507
aI16171
aI16607
aI32017
aI77845
asS'd'
I7098
sa(dp14196
S'a'
I11761
sS's'
I17
sS'm'
(lp14197
I1
aI3
aI1
aI1
aI27
aI19
aI59
aI109
aI347
aI991
aI165
aI683
aI6147
aI493
aI22017
aI19069
aI52857
asS'd'
I7099
sa(dp14198
S'a'
I11762
sS's'
I17
sS'm'
(lp14199
I1
aI1
aI5
aI5
aI21
aI1
aI93
aI115
aI407
aI15
aI421
aI1305
aI3495
aI14287
aI31831
aI65347
aI35339
asS'd'
I7100
sa(dp14200
S'a'
I11787
sS's'
I17
sS'm'
(lp14201
I1
aI3
aI5
aI11
aI29
aI35
aI87
aI27
aI453
aI769
aI1991
aI2757
aI2607
aI9225
aI293
aI49441
aI18185
asS'd'
I7101
sa(dp14202
S'a'
I11792
sS's'
I17
sS'm'
(lp14203
I1
aI1
aI5
aI3
aI23
aI41
aI67
aI195
aI499
aI903
aI197
aI1121
aI4691
aI9277
aI29225
aI34597
aI37395
asS'd'
I7102
sa(dp14204
S'a'
I11814
sS's'
I17
sS'm'
(lp14205
I1
aI1
aI7
aI7
aI21
aI7
aI65
aI245
aI241
aI909
aI1063
aI2271
aI1979
aI10287
aI1747
aI61523
aI72969
asS'd'
I7103
sa(dp14206
S'a'
I11823
sS's'
I17
sS'm'
(lp14207
I1
aI3
aI1
aI13
aI23
aI25
aI3
aI89
aI385
aI481
aI1463
aI3431
aI6907
aI1129
aI3519
aI35789
aI82585
asS'd'
I7104
sa(dp14208
S'a'
I11825
sS's'
I17
sS'm'
(lp14209
I1
aI3
aI5
aI3
aI31
aI17
aI11
aI209
aI77
aI991
aI885
aI3341
aI6895
aI3429
aI21611
aI38555
aI35475
asS'd'
I7105
sa(dp14210
S'a'
I11837
sS's'
I17
sS'm'
(lp14211
I1
aI1
aI3
aI1
aI9
aI61
aI27
aI219
aI433
aI787
aI281
aI1155
aI2915
aI4449
aI30881
aI34461
aI15357
asS'd'
I7106
sa(dp14212
S'a'
I11838
sS's'
I17
sS'm'
(lp14213
I1
aI1
aI3
aI15
aI27
aI55
aI51
aI101
aI117
aI799
aI1475
aI4013
aI5145
aI14991
aI27847
aI49537
aI57339
asS'd'
I7107
sa(dp14214
S'a'
I11846
sS's'
I17
sS'm'
(lp14215
I1
aI3
aI7
aI13
aI13
aI9
aI13
aI167
aI283
aI883
aI1501
aI2635
aI1463
aI3353
aI14961
aI30349
aI62043
asS'd'
I7108
sa(dp14216
S'a'
I11855
sS's'
I17
sS'm'
(lp14217
I1
aI1
aI7
aI3
aI3
aI47
aI119
aI37
aI389
aI655
aI701
aI2471
aI5749
aI6645
aI845
aI27065
aI82299
asS'd'
I7109
sa(dp14218
S'a'
I11864
sS's'
I17
sS'm'
(lp14219
I1
aI1
aI7
aI15
aI27
aI5
aI95
aI195
aI227
aI991
aI1137
aI3715
aI4901
aI9459
aI1917
aI43857
aI126505
asS'd'
I7110
sa(dp14220
S'a'
I11876
sS's'
I17
sS'm'
(lp14221
I1
aI3
aI7
aI5
aI29
aI35
aI45
aI165
aI361
aI257
aI641
aI1265
aI6533
aI11333
aI26081
aI12621
aI66909
asS'd'
I7111
sa(dp14222
S'a'
I11885
sS's'
I17
sS'm'
(lp14223
I1
aI1
aI1
aI11
aI19
aI55
aI73
aI137
aI29
aI355
aI725
aI1161
aI6717
aI2035
aI19769
aI43531
aI72577
asS'd'
I7112
sa(dp14224
S'a'
I11904
sS's'
I17
sS'm'
(lp14225
I1
aI3
aI7
aI5
aI19
aI3
aI99
aI17
aI387
aI621
aI137
aI117
aI6567
aI7667
aI14979
aI17981
aI68319
asS'd'
I7113
sa(dp14226
S'a'
I11909
sS's'
I17
sS'm'
(lp14227
I1
aI1
aI5
aI5
aI7
aI53
aI31
aI33
aI245
aI371
aI691
aI2763
aI95
aI16369
aI7853
aI29839
aI34957
asS'd'
I7114
sa(dp14228
S'a'
I11913
sS's'
I17
sS'm'
(lp14229
I1
aI1
aI3
aI1
aI9
aI1
aI83
aI177
aI17
aI141
aI1739
aI1791
aI3849
aI3093
aI22271
aI53755
aI7817
asS'd'
I7115
sa(dp14230
S'a'
I11916
sS's'
I17
sS'm'
(lp14231
I1
aI3
aI3
aI1
aI3
aI51
aI71
aI69
aI439
aI987
aI807
aI3353
aI4747
aI16031
aI29591
aI61091
aI95675
asS'd'
I7116
sa(dp14232
S'a'
I11940
sS's'
I17
sS'm'
(lp14233
I1
aI3
aI5
aI1
aI17
aI47
aI51
aI211
aI7
aI5
aI1751
aI1735
aI1647
aI13389
aI13861
aI49427
aI13577
asS'd'
I7117
sa(dp14234
S'a'
I11943
sS's'
I17
sS'm'
(lp14235
I1
aI3
aI7
aI5
aI11
aI23
aI17
aI55
aI11
aI61
aI809
aI927
aI6533
aI1509
aI29261
aI21555
aI55075
asS'd'
I7118
sa(dp14236
S'a'
I11972
sS's'
I17
sS'm'
(lp14237
I1
aI3
aI1
aI1
aI15
aI51
aI37
aI47
aI183
aI117
aI597
aI3225
aI1435
aI13359
aI19127
aI17339
aI17345
asS'd'
I7119
sa(dp14238
S'a'
I11981
sS's'
I17
sS'm'
(lp14239
I1
aI1
aI5
aI3
aI5
aI11
aI33
aI179
aI295
aI129
aI29
aI713
aI1561
aI27
aI21087
aI50305
aI39253
asS'd'
I7120
sa(dp14240
S'a'
I11990
sS's'
I17
sS'm'
(lp14241
I1
aI1
aI5
aI7
aI17
aI25
aI105
aI241
aI41
aI915
aI1223
aI2625
aI617
aI10983
aI10749
aI2137
aI101831
asS'd'
I7121
sa(dp14242
S'a'
I11993
sS's'
I17
sS'm'
(lp14243
I1
aI3
aI5
aI7
aI15
aI15
aI85
aI23
aI193
aI625
aI1803
aI2903
aI1935
aI523
aI8377
aI12165
aI105851
asS'd'
I7122
sa(dp14244
S'a'
I12000
sS's'
I17
sS'm'
(lp14245
I1
aI3
aI3
aI7
aI3
aI35
aI5
aI107
aI191
aI855
aI405
aI1659
aI5523
aI5011
aI6401
aI45187
aI31345
asS'd'
I7123
sa(dp14246
S'a'
I12005
sS's'
I17
sS'm'
(lp14247
I1
aI3
aI3
aI1
aI9
aI21
aI103
aI75
aI501
aI669
aI547
aI3685
aI411
aI2663
aI14743
aI13869
aI124389
asS'd'
I7124
sa(dp14248
S'a'
I12015
sS's'
I17
sS'm'
(lp14249
I1
aI3
aI5
aI13
aI15
aI37
aI39
aI79
aI19
aI165
aI1685
aI1367
aI5951
aI12303
aI13423
aI51083
aI119933
asS'd'
I7125
sa(dp14250
S'a'
I12020
sS's'
I17
sS'm'
(lp14251
I1
aI1
aI3
aI1
aI7
aI25
aI1
aI221
aI415
aI591
aI859
aI1457
aI1789
aI2269
aI15947
aI31913
aI86397
asS'd'
I7126
sa(dp14252
S'a'
I12038
sS's'
I17
sS'm'
(lp14253
I1
aI3
aI7
aI15
aI11
aI49
aI15
aI171
aI45
aI925
aI407
aI1719
aI4505
aI5695
aI17397
aI28849
aI77
asS'd'
I7127
sa(dp14254
S'a'
I12042
sS's'
I17
sS'm'
(lp14255
I1
aI1
aI3
aI11
aI21
aI33
aI91
aI115
aI263
aI141
aI753
aI3335
aI7695
aI1981
aI6029
aI22629
aI2467
asS'd'
I7128
sa(dp14256
S'a'
I12056
sS's'
I17
sS'm'
(lp14257
I1
aI3
aI5
aI3
aI25
aI5
aI21
aI67
aI429
aI323
aI223
aI2395
aI761
aI14817
aI12387
aI37905
aI19551
asS'd'
I7129
sa(dp14258
S'a'
I12065
sS's'
I17
sS'm'
(lp14259
I1
aI3
aI1
aI15
aI31
aI43
aI35
aI255
aI73
aI533
aI1093
aI965
aI557
aI607
aI6913
aI35283
aI12261
asS'd'
I7130
sa(dp14260
S'a'
I12066
sS's'
I17
sS'm'
(lp14261
I1
aI3
aI1
aI15
aI25
aI13
aI39
aI83
aI77
aI269
aI1205
aI1577
aI4095
aI6669
aI8643
aI48807
aI98227
asS'd'
I7131
sa(dp14262
S'a'
I12072
sS's'
I17
sS'm'
(lp14263
I1
aI3
aI3
aI7
aI31
aI57
aI25
aI177
aI441
aI973
aI1255
aI675
aI5579
aI4899
aI27925
aI52555
aI70845
asS'd'
I7132
sa(dp14264
S'a'
I12080
sS's'
I17
sS'm'
(lp14265
I1
aI3
aI1
aI5
aI13
aI47
aI15
aI75
aI387
aI461
aI1909
aI841
aI7
aI9567
aI913
aI41411
aI12565
asS'd'
I7133
sa(dp14266
S'a'
I12083
sS's'
I17
sS'm'
(lp14267
I1
aI1
aI5
aI7
aI5
aI21
aI17
aI189
aI319
aI645
aI403
aI2723
aI6747
aI15471
aI26533
aI12709
aI49417
asS'd'
I7134
sa(dp14268
S'a'
I12090
sS's'
I17
sS'm'
(lp14269
I1
aI1
aI5
aI7
aI7
aI41
aI99
aI179
aI137
aI435
aI1061
aI3987
aI4583
aI4101
aI23781
aI54263
aI36695
asS'd'
I7135
sa(dp14270
S'a'
I12092
sS's'
I17
sS'm'
(lp14271
I1
aI3
aI1
aI11
aI19
aI37
aI125
aI177
aI111
aI921
aI1003
aI1433
aI1399
aI3991
aI28193
aI40471
aI97041
asS'd'
I7136
sa(dp14272
S'a'
I12103
sS's'
I17
sS'm'
(lp14273
I1
aI1
aI7
aI1
aI5
aI33
aI7
aI139
aI127
aI413
aI1171
aI2237
aI265
aI10145
aI18793
aI28957
aI25037
asS'd'
I7137
sa(dp14274
S'a'
I12109
sS's'
I17
sS'm'
(lp14275
I1
aI3
aI1
aI1
aI25
aI37
aI13
aI17
aI471
aI195
aI1645
aI3165
aI5635
aI8433
aI28507
aI453
aI107709
asS'd'
I7138
sa(dp14276
S'a'
I12112
sS's'
I17
sS'm'
(lp14277
I1
aI3
aI3
aI11
aI1
aI55
aI119
aI97
aI243
aI371
aI95
aI97
aI7833
aI777
aI12177
aI1861
aI56323
asS'd'
I7139
sa(dp14278
S'a'
I12117
sS's'
I17
sS'm'
(lp14279
I1
aI1
aI7
aI5
aI7
aI29
aI59
aI219
aI405
aI411
aI275
aI111
aI4899
aI10367
aI24331
aI57295
aI47065
asS'd'
I7140
sa(dp14280
S'a'
I12121
sS's'
I17
sS'm'
(lp14281
I1
aI1
aI3
aI3
aI19
aI23
aI91
aI111
aI221
aI195
aI1013
aI3001
aI3227
aI6359
aI30383
aI49699
aI49157
asS'd'
I7141
sa(dp14282
S'a'
I12137
sS's'
I17
sS'm'
(lp14283
I1
aI1
aI5
aI7
aI1
aI21
aI125
aI23
aI177
aI291
aI249
aI861
aI1899
aI14101
aI5079
aI5211
aI14373
asS'd'
I7142
sa(dp14284
S'a'
I12143
sS's'
I17
sS'm'
(lp14285
I1
aI1
aI7
aI11
aI11
aI59
aI33
aI41
aI291
aI919
aI253
aI609
aI1657
aI14633
aI15189
aI22245
aI99815
asS'd'
I7143
sa(dp14286
S'a'
I12145
sS's'
I17
sS'm'
(lp14287
I1
aI3
aI5
aI3
aI23
aI49
aI71
aI137
aI393
aI343
aI1845
aI343
aI5853
aI6639
aI17435
aI62143
aI76041
asS'd'
I7144
sa(dp14288
S'a'
I12148
sS's'
I17
sS'm'
(lp14289
I1
aI1
aI5
aI3
aI9
aI27
aI55
aI193
aI25
aI965
aI1453
aI2739
aI3785
aI12497
aI29607
aI11111
aI25145
asS'd'
I7145
sa(dp14290
S'a'
I12168
sS's'
I17
sS'm'
(lp14291
I1
aI1
aI1
aI1
aI29
aI11
aI111
aI73
aI491
aI629
aI405
aI2779
aI5313
aI589
aI1459
aI47555
aI67945
asS'd'
I7146
sa(dp14292
S'a'
I12174
sS's'
I17
sS'm'
(lp14293
I1
aI3
aI1
aI7
aI13
aI21
aI99
aI75
aI79
aI963
aI207
aI1725
aI6875
aI8359
aI10573
aI45219
aI130463
asS'd'
I7147
sa(dp14294
S'a'
I12188
sS's'
I17
sS'm'
(lp14295
I1
aI3
aI7
aI13
aI1
aI17
aI105
aI227
aI487
aI891
aI1053
aI1333
aI7651
aI5415
aI18661
aI22085
aI82055
asS'd'
I7148
sa(dp14296
S'a'
I12191
sS's'
I17
sS'm'
(lp14297
I1
aI1
aI3
aI3
aI31
aI27
aI91
aI93
aI383
aI331
aI965
aI3035
aI4931
aI13265
aI9729
aI28985
aI118227
asS'd'
I7149
sa(dp14298
S'a'
I12192
sS's'
I17
sS'm'
(lp14299
I1
aI3
aI1
aI1
aI11
aI9
aI59
aI191
aI253
aI909
aI301
aI3811
aI255
aI14937
aI28627
aI54509
aI95993
asS'd'
I7150
sa(dp14300
S'a'
I12201
sS's'
I17
sS'm'
(lp14301
I1
aI3
aI3
aI5
aI11
aI5
aI105
aI77
aI323
aI713
aI637
aI1857
aI2697
aI12473
aI12261
aI2933
aI101287
asS'd'
I7151
sa(dp14302
S'a'
I12224
sS's'
I17
sS'm'
(lp14303
I1
aI3
aI3
aI11
aI9
aI63
aI19
aI19
aI213
aI859
aI1479
aI2849
aI1067
aI5749
aI13511
aI14933
aI11125
asS'd'
I7152
sa(dp14304
S'a'
I12230
sS's'
I17
sS'm'
(lp14305
I1
aI1
aI5
aI9
aI19
aI19
aI13
aI49
aI237
aI511
aI533
aI543
aI575
aI8095
aI27335
aI18847
aI18173
asS'd'
I7153
sa(dp14306
S'a'
I12239
sS's'
I17
sS'm'
(lp14307
I1
aI3
aI5
aI5
aI9
aI53
aI47
aI157
aI35
aI827
aI637
aI2327
aI787
aI5269
aI5145
aI10135
aI111273
asS'd'
I7154
sa(dp14308
S'a'
I12242
sS's'
I17
sS'm'
(lp14309
I1
aI3
aI3
aI7
aI27
aI41
aI69
aI173
aI53
aI655
aI809
aI481
aI6999
aI3101
aI20781
aI2481
aI94957
asS'd'
I7155
sa(dp14310
S'a'
I12251
sS's'
I17
sS'm'
(lp14311
I1
aI1
aI5
aI11
aI17
aI23
aI95
aI201
aI363
aI613
aI863
aI1365
aI1131
aI15417
aI20705
aI8283
aI55235
asS'd'
I7156
sa(dp14312
S'a'
I12258
sS's'
I17
sS'm'
(lp14313
I1
aI1
aI5
aI13
aI3
aI15
aI37
aI219
aI291
aI595
aI1665
aI1861
aI1953
aI15385
aI20569
aI46085
aI15163
asS'd'
I7157
sa(dp14314
S'a'
I12264
sS's'
I17
sS'm'
(lp14315
I1
aI3
aI3
aI11
aI23
aI43
aI125
aI133
aI85
aI45
aI819
aI243
aI7325
aI8723
aI1499
aI58139
aI120353
asS'd'
I7158
sa(dp14316
S'a'
I12310
sS's'
I17
sS'm'
(lp14317
I1
aI1
aI1
aI11
aI21
aI49
aI91
aI145
aI175
aI619
aI1817
aI3533
aI8155
aI7521
aI30361
aI45431
aI130175
asS'd'
I7159
sa(dp14318
S'a'
I12319
sS's'
I17
sS'm'
(lp14319
I1
aI1
aI3
aI1
aI11
aI59
aI57
aI51
aI37
aI903
aI1221
aI3813
aI8043
aI14165
aI31503
aI7905
aI61515
asS'd'
I7160
sa(dp14320
S'a'
I12323
sS's'
I17
sS'm'
(lp14321
I1
aI1
aI1
aI1
aI15
aI9
aI115
aI175
aI285
aI839
aI97
aI3119
aI719
aI15283
aI22947
aI25417
aI40665
asS'd'
I7161
sa(dp14322
S'a'
I12325
sS's'
I17
sS'm'
(lp14323
I1
aI3
aI1
aI7
aI5
aI49
aI127
aI111
aI373
aI747
aI393
aI2351
aI4577
aI15227
aI23149
aI16901
aI80253
asS'd'
I7162
sa(dp14324
S'a'
I12332
sS's'
I17
sS'm'
(lp14325
I1
aI1
aI5
aI3
aI15
aI5
aI95
aI197
aI251
aI275
aI831
aI1389
aI3907
aI12343
aI11599
aI24369
aI65361
asS'd'
I7163
sa(dp14326
S'a'
I12343
sS's'
I17
sS'm'
(lp14327
I1
aI3
aI7
aI5
aI25
aI37
aI11
aI75
aI417
aI789
aI745
aI811
aI2189
aI15381
aI4785
aI41657
aI2897
asS'd'
I7164
sa(dp14328
S'a'
I12344
sS's'
I17
sS'm'
(lp14329
I1
aI3
aI1
aI13
aI29
aI55
aI55
aI33
aI279
aI383
aI1645
aI975
aI4683
aI1357
aI1149
aI30271
aI90527
asS'd'
I7165
sa(dp14330
S'a'
I12352
sS's'
I17
sS'm'
(lp14331
I1
aI3
aI5
aI3
aI5
aI3
aI79
aI61
aI371
aI225
aI141
aI369
aI1037
aI12249
aI29431
aI37253
aI9899
asS'd'
I7166
sa(dp14332
S'a'
I12370
sS's'
I17
sS'm'
(lp14333
I1
aI1
aI3
aI13
aI13
aI7
aI127
aI147
aI507
aI119
aI1085
aI1949
aI6289
aI10179
aI10107
aI55989
aI74395
asS'd'
I7167
sa(dp14334
S'a'
I12388
sS's'
I17
sS'm'
(lp14335
I1
aI3
aI1
aI7
aI21
aI35
aI53
aI209
aI103
aI365
aI683
aI553
aI4977
aI14371
aI24037
aI11453
aI45369
asS'd'
I7168
sa(dp14336
S'a'
I12395
sS's'
I17
sS'm'
(lp14337
I1
aI1
aI5
aI11
aI27
aI39
aI41
aI145
aI437
aI55
aI893
aI2375
aI4977
aI5451
aI21225
aI46815
aI1423
asS'd'
I7169
sa(dp14338
S'a'
I12403
sS's'
I17
sS'm'
(lp14339
I1
aI3
aI5
aI1
aI23
aI53
aI113
aI75
aI209
aI323
aI1975
aI3809
aI1829
aI14625
aI3821
aI53773
aI129173
asS'd'
I7170
sa(dp14340
S'a'
I12409
sS's'
I17
sS'm'
(lp14341
I1
aI1
aI5
aI3
aI7
aI51
aI97
aI73
aI289
aI481
aI339
aI1375
aI3101
aI4395
aI13933
aI33267
aI68115
asS'd'
I7171
sa(dp14342
S'a'
I12410
sS's'
I17
sS'm'
(lp14343
I1
aI3
aI5
aI1
aI5
aI45
aI83
aI57
aI3
aI667
aI109
aI3979
aI6447
aI8603
aI20147
aI49291
aI18023
asS'd'
I7172
sa(dp14344
S'a'
I12415
sS's'
I17
sS'm'
(lp14345
I1
aI3
aI7
aI1
aI11
aI7
aI45
aI233
aI65
aI745
aI1009
aI2979
aI5965
aI10681
aI3499
aI23077
aI87479
asS'd'
I7173
sa(dp14346
S'a'
I12419
sS's'
I17
sS'm'
(lp14347
I1
aI3
aI3
aI3
aI13
aI25
aI25
aI189
aI197
aI83
aI1429
aI2857
aI2877
aI8577
aI24811
aI33049
aI46009
asS'd'
I7174
sa(dp14348
S'a'
I12426
sS's'
I17
sS'm'
(lp14349
I1
aI1
aI1
aI7
aI11
aI47
aI47
aI255
aI89
aI625
aI449
aI3747
aI2035
aI3509
aI4901
aI2961
aI14073
asS'd'
I7175
sa(dp14350
S'a'
I12439
sS's'
I17
sS'm'
(lp14351
I1
aI1
aI1
aI13
aI9
aI55
aI35
aI47
aI389
aI573
aI847
aI1037
aI1345
aI5487
aI7575
aI57435
aI77303
asS'd'
I7176
sa(dp14352
S'a'
I12445
sS's'
I17
sS'm'
(lp14353
I1
aI1
aI5
aI11
aI25
aI51
aI113
aI109
aI79
aI339
aI95
aI2049
aI5881
aI13209
aI20041
aI26419
aI110319
asS'd'
I7177
sa(dp14354
S'a'
I12459
sS's'
I17
sS'm'
(lp14355
I1
aI1
aI7
aI1
aI27
aI15
aI93
aI145
aI253
aI917
aI1211
aI2221
aI1087
aI14209
aI32097
aI20083
aI67841
asS'd'
I7178
sa(dp14356
S'a'
I12464
sS's'
I17
sS'm'
(lp14357
I1
aI1
aI3
aI15
aI13
aI19
aI67
aI107
aI75
aI919
aI2047
aI3675
aI6231
aI1243
aI14335
aI35939
aI17281
asS'd'
I7179
sa(dp14358
S'a'
I12474
sS's'
I17
sS'm'
(lp14359
I1
aI3
aI7
aI5
aI27
aI47
aI53
aI239
aI475
aI231
aI1645
aI825
aI4039
aI15985
aI10853
aI32951
aI34985
asS'd'
I7180
sa(dp14360
S'a'
I12484
sS's'
I17
sS'm'
(lp14361
I1
aI1
aI7
aI5
aI15
aI61
aI107
aI93
aI51
aI221
aI717
aI2859
aI7885
aI9571
aI11841
aI45143
aI33723
asS'd'
I7181
sa(dp14362
S'a'
I12491
sS's'
I17
sS'm'
(lp14363
I1
aI1
aI7
aI7
aI9
aI25
aI63
aI25
aI47
aI55
aI2041
aI3965
aI215
aI14857
aI31669
aI54775
aI42157
asS'd'
I7182
sa(dp14364
S'a'
I12501
sS's'
I17
sS'm'
(lp14365
I1
aI3
aI5
aI1
aI5
aI45
aI123
aI109
aI471
aI599
aI479
aI475
aI3499
aI11963
aI23709
aI18851
aI66861
asS'd'
I7183
sa(dp14366
S'a'
I12505
sS's'
I17
sS'm'
(lp14367
I1
aI3
aI3
aI3
aI5
aI29
aI71
aI81
aI315
aI329
aI1471
aI3995
aI623
aI5871
aI11171
aI15645
aI97251
asS'd'
I7184
sa(dp14368
S'a'
I12508
sS's'
I17
sS'm'
(lp14369
I1
aI1
aI7
aI11
aI15
aI15
aI101
aI173
aI445
aI871
aI765
aI1121
aI1937
aI13055
aI7309
aI54175
aI85559
asS'd'
I7185
sa(dp14370
S'a'
I12511
sS's'
I17
sS'm'
(lp14371
I1
aI3
aI5
aI7
aI7
aI13
aI43
aI237
aI361
aI981
aI19
aI3113
aI4681
aI3313
aI19147
aI35193
aI87281
asS'd'
I7186
sa(dp14372
S'a'
I12521
sS's'
I17
sS'm'
(lp14373
I1
aI3
aI5
aI3
aI27
aI13
aI37
aI51
aI233
aI573
aI1599
aI2807
aI7149
aI12083
aI28927
aI7797
aI130879
asS'd'
I7187
sa(dp14374
S'a'
I12522
sS's'
I17
sS'm'
(lp14375
I1
aI1
aI1
aI13
aI31
aI63
aI127
aI89
aI209
aI717
aI1075
aI3887
aI1427
aI87
aI18565
aI39973
aI55025
asS'd'
I7188
sa(dp14376
S'a'
I12530
sS's'
I17
sS'm'
(lp14377
I1
aI3
aI1
aI5
aI15
aI11
aI121
aI247
aI273
aI613
aI1857
aI2059
aI7399
aI13951
aI9025
aI39523
aI68121
asS'd'
I7189
sa(dp14378
S'a'
I12544
sS's'
I17
sS'm'
(lp14379
I1
aI3
aI7
aI13
aI31
aI9
aI61
aI143
aI375
aI433
aI471
aI1315
aI5299
aI1167
aI10099
aI11445
aI51693
asS'd'
I7190
sa(dp14380
S'a'
I12547
sS's'
I17
sS'm'
(lp14381
I1
aI1
aI7
aI9
aI25
aI31
aI125
aI5
aI13
aI595
aI621
aI3551
aI7959
aI10643
aI14345
aI37683
aI118377
asS'd'
I7191
sa(dp14382
S'a'
I12561
sS's'
I17
sS'm'
(lp14383
I1
aI1
aI5
aI11
aI1
aI33
aI45
aI31
aI447
aI229
aI893
aI3777
aI4101
aI2505
aI4855
aI14057
aI20133
asS'd'
I7192
sa(dp14384
S'a'
I12571
sS's'
I17
sS'm'
(lp14385
I1
aI1
aI1
aI1
aI7
aI23
aI89
aI53
aI483
aI873
aI521
aI2115
aI1461
aI11241
aI1003
aI28749
aI68227
asS'd'
I7193
sa(dp14386
S'a'
I12580
sS's'
I17
sS'm'
(lp14387
I1
aI3
aI5
aI5
aI3
aI17
aI23
aI219
aI281
aI975
aI895
aI4043
aI6505
aI5991
aI27401
aI38791
aI89239
asS'd'
I7194
sa(dp14388
S'a'
I12597
sS's'
I17
sS'm'
(lp14389
I1
aI1
aI1
aI7
aI29
aI41
aI63
aI151
aI195
aI495
aI469
aI305
aI7437
aI1107
aI31147
aI30755
aI116551
asS'd'
I7195
sa(dp14390
S'a'
I12607
sS's'
I17
sS'm'
(lp14391
I1
aI3
aI7
aI3
aI13
aI25
aI33
aI193
aI23
aI135
aI3
aI513
aI4169
aI15355
aI2255
aI32167
aI68691
asS'd'
I7196
sa(dp14392
S'a'
I12609
sS's'
I17
sS'm'
(lp14393
I1
aI3
aI3
aI11
aI29
aI19
aI125
aI177
aI83
aI361
aI393
aI663
aI1859
aI1333
aI17507
aI10661
aI72387
asS'd'
I7197
sa(dp14394
S'a'
I12610
sS's'
I17
sS'm'
(lp14395
I1
aI1
aI5
aI11
aI23
aI13
aI61
aI33
aI149
aI145
aI995
aI649
aI7587
aI6743
aI25225
aI54997
aI10193
asS'd'
I7198
sa(dp14396
S'a'
I12616
sS's'
I17
sS'm'
(lp14397
I1
aI1
aI7
aI13
aI29
aI51
aI107
aI79
aI467
aI881
aI1227
aI1083
aI3277
aI2559
aI26819
aI57311
aI48095
asS'd'
I7199
sa(dp14398
S'a'
I12621
sS's'
I17
sS'm'
(lp14399
I1
aI3
aI1
aI1
aI1
aI19
aI23
aI25
aI239
aI703
aI119
aI2525
aI8079
aI5433
aI8989
aI42517
aI116755
asS'd'
I7200
sa(dp14400
S'a'
I12624
sS's'
I17
sS'm'
(lp14401
I1
aI1
aI7
aI11
aI31
aI9
aI9
aI113
aI381
aI363
aI447
aI3751
aI7523
aI15995
aI3853
aI42069
aI81455
asS'd'
I7201
sa(dp14402
S'a'
I12639
sS's'
I17
sS'm'
(lp14403
I1
aI1
aI5
aI9
aI29
aI41
aI103
aI179
aI477
aI527
aI1593
aI3003
aI1095
aI6823
aI6911
aI44987
aI32445
asS'd'
I7202
sa(dp14404
S'a'
I12645
sS's'
I17
sS'm'
(lp14405
I1
aI1
aI7
aI15
aI5
aI31
aI55
aI181
aI149
aI127
aI1745
aI2753
aI801
aI285
aI20199
aI33707
aI118397
asS'd'
I7203
sa(dp14406
S'a'
I12652
sS's'
I17
sS'm'
(lp14407
I1
aI3
aI7
aI7
aI11
aI29
aI89
aI215
aI351
aI303
aI1519
aI2593
aI2045
aI14699
aI1657
aI40799
aI39641
asS'd'
I7204
sa(dp14408
S'a'
I12655
sS's'
I17
sS'm'
(lp14409
I1
aI1
aI7
aI13
aI3
aI35
aI73
aI111
aI15
aI803
aI1819
aI3453
aI3611
aI8337
aI14239
aI14875
aI83983
asS'd'
I7205
sa(dp14410
S'a'
I12660
sS's'
I17
sS'm'
(lp14411
I1
aI1
aI5
aI15
aI15
aI49
aI27
aI101
aI149
aI3
aI717
aI2229
aI7397
aI6579
aI10965
aI35997
aI28823
asS'd'
I7206
sa(dp14412
S'a'
I12667
sS's'
I17
sS'm'
(lp14413
I1
aI1
aI5
aI7
aI3
aI17
aI49
aI245
aI343
aI657
aI15
aI749
aI6413
aI10811
aI2909
aI47309
aI34613
asS'd'
I7207
sa(dp14414
S'a'
I12686
sS's'
I17
sS'm'
(lp14415
I1
aI3
aI5
aI15
aI13
aI35
aI67
aI99
aI481
aI379
aI2003
aI3367
aI3065
aI5845
aI7799
aI43931
aI15263
asS'd'
I7208
sa(dp14416
S'a'
I12688
sS's'
I17
sS'm'
(lp14417
I1
aI1
aI5
aI13
aI21
aI49
aI81
aI77
aI395
aI919
aI1931
aI661
aI123
aI9965
aI10487
aI55131
aI1567
asS'd'
I7209
sa(dp14418
S'a'
I12697
sS's'
I17
sS'm'
(lp14419
I1
aI3
aI5
aI11
aI23
aI39
aI41
aI121
aI159
aI473
aI191
aI1983
aI6411
aI10503
aI10523
aI40601
aI64153
asS'd'
I7210
sa(dp14420
S'a'
I12700
sS's'
I17
sS'm'
(lp14421
I1
aI1
aI5
aI7
aI9
aI37
aI73
aI207
aI497
aI789
aI1671
aI325
aI1697
aI11281
aI31185
aI4961
aI124431
asS'd'
I7211
sa(dp14422
S'a'
I12707
sS's'
I17
sS'm'
(lp14423
I1
aI1
aI5
aI15
aI7
aI51
aI71
aI91
aI449
aI707
aI621
aI2427
aI627
aI1747
aI12779
aI17569
aI98289
asS'd'
I7212
sa(dp14424
S'a'
I12710
sS's'
I17
sS'm'
(lp14425
I1
aI1
aI5
aI5
aI31
aI3
aI89
aI163
aI77
aI647
aI1747
aI2965
aI1669
aI3311
aI17651
aI8111
aI30739
asS'd'
I7213
sa(dp14426
S'a'
I12719
sS's'
I17
sS'm'
(lp14427
I1
aI1
aI3
aI11
aI15
aI31
aI77
aI173
aI405
aI913
aI459
aI2955
aI6153
aI13391
aI20439
aI64433
aI12371
asS'd'
I7214
sa(dp14428
S'a'
I12739
sS's'
I17
sS'm'
(lp14429
I1
aI1
aI3
aI11
aI13
aI55
aI29
aI37
aI379
aI689
aI407
aI1373
aI397
aI5027
aI15497
aI25687
aI48193
asS'd'
I7215
sa(dp14430
S'a'
I12742
sS's'
I17
sS'm'
(lp14431
I1
aI3
aI3
aI15
aI13
aI7
aI81
aI207
aI395
aI901
aI779
aI1683
aI2491
aI3807
aI31979
aI32403
aI81113
asS'd'
I7216
sa(dp14432
S'a'
I12746
sS's'
I17
sS'm'
(lp14433
I1
aI3
aI3
aI13
aI29
aI31
aI25
aI81
aI459
aI991
aI793
aI3285
aI2775
aI16199
aI11423
aI52597
aI86041
asS'd'
I7217
sa(dp14434
S'a'
I12754
sS's'
I17
sS'm'
(lp14435
I1
aI3
aI3
aI13
aI17
aI17
aI101
aI183
aI19
aI735
aI671
aI1097
aI2461
aI9863
aI25985
aI31915
aI73047
asS'd'
I7218
sa(dp14436
S'a'
I12765
sS's'
I17
sS'm'
(lp14437
I1
aI3
aI3
aI3
aI3
aI11
aI71
aI63
aI429
aI899
aI351
aI1275
aI3907
aI14241
aI19135
aI14875
aI43325
asS'd'
I7219
sa(dp14438
S'a'
I12772
sS's'
I17
sS'm'
(lp14439
I1
aI1
aI7
aI11
aI11
aI61
aI15
aI213
aI411
aI13
aI1409
aI1741
aI5257
aI8729
aI28351
aI6381
aI77501
asS'd'
I7220
sa(dp14440
S'a'
I12784
sS's'
I17
sS'm'
(lp14441
I1
aI1
aI7
aI7
aI29
aI27
aI51
aI217
aI411
aI261
aI599
aI3027
aI7871
aI9133
aI32423
aI44275
aI34701
asS'd'
I7221
sa(dp14442
S'a'
I12789
sS's'
I17
sS'm'
(lp14443
I1
aI3
aI7
aI7
aI1
aI1
aI127
aI209
aI151
aI845
aI1421
aI3115
aI7775
aI10133
aI6163
aI41165
aI91187
asS'd'
I7222
sa(dp14444
S'a'
I12800
sS's'
I17
sS'm'
(lp14445
I1
aI3
aI1
aI9
aI1
aI35
aI75
aI3
aI81
aI477
aI131
aI1383
aI1377
aI6857
aI3863
aI12583
aI7855
asS'd'
I7223
sa(dp14446
S'a'
I12805
sS's'
I17
sS'm'
(lp14447
I1
aI3
aI1
aI3
aI3
aI11
aI1
aI167
aI347
aI317
aI557
aI3763
aI7175
aI13341
aI759
aI23275
aI78039
asS'd'
I7224
sa(dp14448
S'a'
I12809
sS's'
I17
sS'm'
(lp14449
I1
aI3
aI5
aI11
aI19
aI53
aI85
aI139
aI67
aI757
aI487
aI919
aI6001
aI16031
aI24959
aI28013
aI65771
asS'd'
I7225
sa(dp14450
S'a'
I12815
sS's'
I17
sS'm'
(lp14451
I1
aI1
aI1
aI1
aI23
aI9
aI83
aI55
aI249
aI305
aI1305
aI109
aI5559
aI5129
aI30973
aI19889
aI6691
asS'd'
I7226
sa(dp14452
S'a'
I12827
sS's'
I17
sS'm'
(lp14453
I1
aI1
aI1
aI3
aI21
aI19
aI85
aI89
aI213
aI847
aI861
aI1651
aI6613
aI6001
aI8157
aI2555
aI98673
asS'd'
I7227
sa(dp14454
S'a'
I12830
sS's'
I17
sS'm'
(lp14455
I1
aI1
aI1
aI15
aI25
aI15
aI125
aI133
aI177
aI295
aI549
aI1763
aI2811
aI4381
aI1079
aI7813
aI87909
asS'd'
I7228
sa(dp14456
S'a'
I12833
sS's'
I17
sS'm'
(lp14457
I1
aI1
aI1
aI5
aI5
aI17
aI25
aI225
aI353
aI997
aI1565
aI2225
aI7265
aI16227
aI28209
aI9011
aI97193
asS'd'
I7229
sa(dp14458
S'a'
I12840
sS's'
I17
sS'm'
(lp14459
I1
aI3
aI7
aI15
aI13
aI13
aI35
aI239
aI331
aI965
aI1547
aI1627
aI6409
aI7745
aI30899
aI36915
aI59293
asS'd'
I7230
sa(dp14460
S'a'
I12851
sS's'
I17
sS'm'
(lp14461
I1
aI1
aI1
aI13
aI27
aI45
aI23
aI179
aI193
aI801
aI381
aI3783
aI3551
aI11855
aI11041
aI49911
aI62101
asS'd'
I7231
sa(dp14462
S'a'
I12868
sS's'
I17
sS'm'
(lp14463
I1
aI1
aI7
aI3
aI3
aI31
aI61
aI5
aI421
aI939
aI1637
aI217
aI389
aI1797
aI32141
aI28817
aI6997
asS'd'
I7232
sa(dp14464
S'a'
I12871
sS's'
I17
sS'm'
(lp14465
I1
aI3
aI3
aI5
aI21
aI31
aI83
aI65
aI421
aI577
aI1137
aI2561
aI2943
aI4171
aI2803
aI23325
aI92315
asS'd'
I7233
sa(dp14466
S'a'
I12886
sS's'
I17
sS'm'
(lp14467
I1
aI1
aI3
aI3
aI27
aI33
aI75
aI81
aI477
aI3
aI1903
aI773
aI5551
aI10069
aI7285
aI58103
aI98311
asS'd'
I7234
sa(dp14468
S'a'
I12899
sS's'
I17
sS'm'
(lp14469
I1
aI3
aI7
aI15
aI1
aI17
aI95
aI209
aI65
aI747
aI1633
aI581
aI7395
aI1393
aI21795
aI15735
aI129757
asS'd'
I7235
sa(dp14470
S'a'
I12923
sS's'
I17
sS'm'
(lp14471
I1
aI3
aI5
aI3
aI17
aI3
aI9
aI131
aI51
aI693
aI1571
aI1865
aI8137
aI915
aI13345
aI35137
aI59005
asS'd'
I7236
sa(dp14472
S'a'
I12926
sS's'
I17
sS'm'
(lp14473
I1
aI1
aI3
aI7
aI23
aI27
aI61
aI163
aI449
aI87
aI717
aI1075
aI4309
aI4887
aI11741
aI24549
aI96729
asS'd'
I7237
sa(dp14474
S'a'
I12932
sS's'
I17
sS'm'
(lp14475
I1
aI3
aI7
aI13
aI21
aI5
aI3
aI97
aI191
aI999
aI1193
aI1215
aI5907
aI10491
aI2281
aI6455
aI68625
asS'd'
I7238
sa(dp14476
S'a'
I12935
sS's'
I17
sS'm'
(lp14477
I1
aI1
aI5
aI7
aI9
aI9
aI101
aI5
aI375
aI137
aI1473
aI1265
aI5307
aI259
aI20699
aI25367
aI129393
asS'd'
I7239
sa(dp14478
S'a'
I12939
sS's'
I17
sS'm'
(lp14479
I1
aI3
aI7
aI7
aI21
aI1
aI77
aI65
aI23
aI139
aI945
aI491
aI1069
aI253
aI12335
aI26861
aI129467
asS'd'
I7240
sa(dp14480
S'a'
I12942
sS's'
I17
sS'm'
(lp14481
I1
aI1
aI3
aI9
aI15
aI33
aI85
aI225
aI45
aI311
aI281
aI1601
aI7325
aI12265
aI2591
aI51955
aI130681
asS'd'
I7241
sa(dp14482
S'a'
I12953
sS's'
I17
sS'm'
(lp14483
I1
aI1
aI1
aI3
aI27
aI33
aI17
aI89
aI495
aI91
aI527
aI3347
aI7883
aI9481
aI28731
aI54729
aI15265
asS'd'
I7242
sa(dp14484
S'a'
I12959
sS's'
I17
sS'm'
(lp14485
I1
aI1
aI3
aI3
aI9
aI47
aI115
aI161
aI299
aI493
aI1857
aI3597
aI7175
aI15603
aI11523
aI33837
aI57557
asS'd'
I7243
sa(dp14486
S'a'
I12960
sS's'
I17
sS'm'
(lp14487
I1
aI3
aI7
aI3
aI15
aI47
aI127
aI195
aI391
aI869
aI99
aI429
aI7125
aI10413
aI5063
aI61845
aI71843
asS'd'
I7244
sa(dp14488
S'a'
I13002
sS's'
I17
sS'm'
(lp14489
I1
aI3
aI3
aI1
aI7
aI31
aI27
aI69
aI7
aI83
aI315
aI2749
aI5693
aI13377
aI28091
aI13065
aI111029
asS'd'
I7245
sa(dp14490
S'a'
I13004
sS's'
I17
sS'm'
(lp14491
I1
aI1
aI3
aI15
aI15
aI45
aI125
aI229
aI459
aI611
aI1167
aI3375
aI3587
aI81
aI9275
aI45327
aI39749
asS'd'
I7246
sa(dp14492
S'a'
I13016
sS's'
I17
sS'm'
(lp14493
I1
aI3
aI7
aI11
aI9
aI3
aI43
aI161
aI221
aI209
aI51
aI1475
aI3577
aI13973
aI15285
aI35553
aI83935
asS'd'
I7247
sa(dp14494
S'a'
I13021
sS's'
I17
sS'm'
(lp14495
I1
aI1
aI7
aI9
aI15
aI55
aI25
aI119
aI39
aI537
aI317
aI1331
aI2161
aI1791
aI19221
aI63459
aI124595
asS'd'
I7248
sa(dp14496
S'a'
I13035
sS's'
I17
sS'm'
(lp14497
I1
aI1
aI1
aI11
aI9
aI7
aI113
aI187
aI295
aI67
aI1795
aI113
aI119
aI9127
aI32119
aI7719
aI67591
asS'd'
I7249
sa(dp14498
S'a'
I13038
sS's'
I17
sS'm'
(lp14499
I1
aI3
aI7
aI13
aI1
aI53
aI17
aI19
aI331
aI711
aI359
aI2945
aI5847
aI7237
aI23617
aI17411
aI2203
asS'd'
I7250
sa(dp14500
S'a'
I13052
sS's'
I17
sS'm'
(lp14501
I1
aI3
aI3
aI7
aI21
aI63
aI115
aI159
aI225
aI161
aI1255
aI2381
aI7411
aI95
aI1625
aI30493
aI56685
asS'd'
I7251
sa(dp14502
S'a'
I13058
sS's'
I17
sS'm'
(lp14503
I1
aI1
aI5
aI9
aI13
aI57
aI5
aI107
aI195
aI271
aI677
aI2081
aI6027
aI11091
aI14171
aI19007
aI102119
asS'd'
I7252
sa(dp14504
S'a'
I13069
sS's'
I17
sS'm'
(lp14505
I1
aI3
aI3
aI3
aI19
aI13
aI31
aI155
aI209
aI89
aI955
aI523
aI615
aI5319
aI16079
aI9289
aI49135
asS'd'
I7253
sa(dp14506
S'a'
I13082
sS's'
I17
sS'm'
(lp14507
I1
aI3
aI1
aI13
aI1
aI31
aI69
aI143
aI329
aI813
aI635
aI891
aI2967
aI5563
aI19643
aI35813
aI14345
asS'd'
I7254
sa(dp14508
S'a'
I13093
sS's'
I17
sS'm'
(lp14509
I1
aI1
aI5
aI5
aI17
aI47
aI97
aI49
aI123
aI997
aI15
aI3685
aI3925
aI4973
aI11195
aI17115
aI63709
asS'd'
I7255
sa(dp14510
S'a'
I13094
sS's'
I17
sS'm'
(lp14511
I1
aI1
aI7
aI13
aI13
aI17
aI99
aI149
aI309
aI281
aI329
aI905
aI6487
aI4495
aI31831
aI24413
aI26431
asS'd'
I7256
sa(dp14512
S'a'
I13100
sS's'
I17
sS'm'
(lp14513
I1
aI1
aI5
aI5
aI5
aI47
aI113
aI115
aI61
aI157
aI955
aI2323
aI4445
aI229
aI24049
aI14753
aI15189
asS'd'
I7257
sa(dp14514
S'a'
I13115
sS's'
I17
sS'm'
(lp14515
I1
aI3
aI7
aI15
aI25
aI21
aI13
aI137
aI377
aI45
aI629
aI1339
aI8037
aI5073
aI24741
aI48589
aI28953
asS'd'
I7258
sa(dp14516
S'a'
I13125
sS's'
I17
sS'm'
(lp14517
I1
aI3
aI3
aI9
aI3
aI41
aI7
aI101
aI333
aI59
aI1213
aI1871
aI3993
aI11261
aI4403
aI42785
aI58753
asS'd'
I7259
sa(dp14518
S'a'
I13132
sS's'
I17
sS'm'
(lp14519
I1
aI3
aI1
aI7
aI5
aI33
aI87
aI73
aI317
aI575
aI1459
aI905
aI1033
aI14179
aI19595
aI30269
aI103853
asS'd'
I7260
sa(dp14520
S'a'
I13143
sS's'
I17
sS'm'
(lp14521
I1
aI1
aI1
aI1
aI19
aI49
aI63
aI181
aI227
aI401
aI695
aI1811
aI2383
aI3835
aI14379
aI30685
aI114731
asS'd'
I7261
sa(dp14522
S'a'
I13144
sS's'
I17
sS'm'
(lp14523
I1
aI1
aI5
aI15
aI9
aI41
aI35
aI91
aI357
aI659
aI155
aI3725
aI6509
aI405
aI25449
aI37719
aI6013
asS'd'
I7262
sa(dp14524
S'a'
I13153
sS's'
I17
sS'm'
(lp14525
I1
aI1
aI7
aI3
aI11
aI59
aI33
aI151
aI291
aI393
aI741
aI3961
aI2787
aI993
aI10361
aI11737
aI42047
asS'd'
I7263
sa(dp14526
S'a'
I13160
sS's'
I17
sS'm'
(lp14527
I1
aI3
aI7
aI3
aI15
aI15
aI55
aI59
aI419
aI203
aI55
aI801
aI2719
aI15487
aI13213
aI58473
aI50315
asS'd'
I7264
sa(dp14528
S'a'
I13165
sS's'
I17
sS'm'
(lp14529
I1
aI3
aI7
aI13
aI17
aI21
aI113
aI111
aI159
aI163
aI711
aI1135
aI1133
aI15519
aI30515
aI55777
aI25025
asS'd'
I7265
sa(dp14530
S'a'
I13173
sS's'
I17
sS'm'
(lp14531
I1
aI1
aI3
aI5
aI13
aI25
aI23
aI3
aI93
aI873
aI559
aI1815
aI3381
aI5311
aI14365
aI34349
aI17333
asS'd'
I7266
sa(dp14532
S'a'
I13174
sS's'
I17
sS'm'
(lp14533
I1
aI3
aI5
aI7
aI15
aI43
aI85
aI33
aI23
aI903
aI1247
aI3279
aI1393
aI12059
aI19251
aI19389
aI5097
asS'd'
I7267
sa(dp14534
S'a'
I13187
sS's'
I17
sS'm'
(lp14535
I1
aI3
aI1
aI11
aI21
aI59
aI3
aI153
aI403
aI95
aI1939
aI2679
aI419
aI9035
aI31219
aI2897
aI15727
asS'd'
I7268
sa(dp14536
S'a'
I13190
sS's'
I17
sS'm'
(lp14537
I1
aI3
aI5
aI1
aI11
aI21
aI35
aI169
aI453
aI15
aI791
aI3931
aI1021
aI16321
aI6033
aI10639
aI16173
asS'd'
I7269
sa(dp14538
S'a'
I13204
sS's'
I17
sS'm'
(lp14539
I1
aI3
aI1
aI11
aI7
aI57
aI39
aI61
aI381
aI465
aI451
aI2863
aI575
aI5597
aI31041
aI8625
aI82373
asS'd'
I7270
sa(dp14540
S'a'
I13218
sS's'
I17
sS'm'
(lp14541
I1
aI1
aI3
aI5
aI13
aI5
aI63
aI1
aI75
aI245
aI1305
aI285
aI3367
aI10107
aI5853
aI35275
aI128255
asS'd'
I7271
sa(dp14542
S'a'
I13247
sS's'
I17
sS'm'
(lp14543
I1
aI1
aI1
aI3
aI1
aI57
aI21
aI91
aI139
aI669
aI765
aI1867
aI2153
aI10347
aI26119
aI35517
aI4725
asS'd'
I7272
sa(dp14544
S'a'
I13250
sS's'
I17
sS'm'
(lp14545
I1
aI3
aI5
aI1
aI21
aI41
aI59
aI247
aI473
aI1015
aI975
aI485
aI2161
aI11941
aI10341
aI35245
aI55587
asS'd'
I7273
sa(dp14546
S'a'
I13262
sS's'
I17
sS'm'
(lp14547
I1
aI1
aI5
aI9
aI7
aI59
aI33
aI149
aI97
aI619
aI393
aI3613
aI6037
aI10895
aI19461
aI15975
aI47919
asS'd'
I7274
sa(dp14548
S'a'
I13267
sS's'
I17
sS'm'
(lp14549
I1
aI3
aI3
aI15
aI7
aI17
aI95
aI13
aI147
aI361
aI915
aI2585
aI4483
aI3159
aI12255
aI44685
aI116163
asS'd'
I7275
sa(dp14550
S'a'
I13274
sS's'
I17
sS'm'
(lp14551
I1
aI3
aI1
aI15
aI27
aI31
aI75
aI31
aI423
aI233
aI1453
aI2815
aI3633
aI6531
aI25721
aI29649
aI80645
asS'd'
I7276
sa(dp14552
S'a'
I13304
sS's'
I17
sS'm'
(lp14553
I1
aI3
aI3
aI3
aI19
aI7
aI73
aI33
aI163
aI495
aI1483
aI2277
aI6455
aI6523
aI9331
aI21869
aI52175
asS'd'
I7277
sa(dp14554
S'a'
I13309
sS's'
I17
sS'm'
(lp14555
I1
aI3
aI5
aI13
aI5
aI1
aI63
aI35
aI335
aI189
aI713
aI2997
aI3277
aI10049
aI4681
aI16753
aI17107
asS'd'
I7278
sa(dp14556
S'a'
I13315
sS's'
I17
sS'm'
(lp14557
I1
aI3
aI5
aI9
aI3
aI55
aI29
aI171
aI395
aI585
aI671
aI1875
aI4449
aI12895
aI5455
aI11023
aI106189
asS'd'
I7279
sa(dp14558
S'a'
I13317
sS's'
I17
sS'm'
(lp14559
I1
aI3
aI5
aI3
aI25
aI53
aI33
aI169
aI109
aI285
aI787
aI861
aI5549
aI5171
aI15293
aI2977
aI14559
asS'd'
I7280
sa(dp14560
S'a'
I13324
sS's'
I17
sS'm'
(lp14561
I1
aI3
aI7
aI1
aI21
aI25
aI97
aI115
aI1
aI999
aI1033
aI3471
aI129
aI16093
aI495
aI16859
aI34615
asS'd'
I7281
sa(dp14562
S'a'
I13332
sS's'
I17
sS'm'
(lp14563
I1
aI3
aI5
aI9
aI13
aI5
aI109
aI41
aI57
aI957
aI231
aI3771
aI2917
aI15649
aI8869
aI14857
aI64943
asS'd'
I7282
sa(dp14564
S'a'
I13342
sS's'
I17
sS'm'
(lp14565
I1
aI1
aI7
aI7
aI21
aI41
aI101
aI59
aI167
aI441
aI997
aI2951
aI7891
aI16325
aI12669
aI53829
aI100705
asS'd'
I7283
sa(dp14566
S'a'
I13346
sS's'
I17
sS'm'
(lp14567
I1
aI1
aI7
aI9
aI19
aI59
aI23
aI141
aI193
aI237
aI1067
aI1823
aI3351
aI3239
aI3135
aI9275
aI37069
asS'd'
I7284
sa(dp14568
S'a'
I13355
sS's'
I17
sS'm'
(lp14569
I1
aI3
aI3
aI9
aI3
aI17
aI95
aI73
aI19
aI231
aI779
aI3065
aI2245
aI2967
aI24971
aI62589
aI16729
asS'd'
I7285
sa(dp14570
S'a'
I13358
sS's'
I17
sS'm'
(lp14571
I1
aI3
aI3
aI13
aI25
aI19
aI117
aI147
aI443
aI123
aI157
aI2037
aI327
aI14715
aI5693
aI54641
aI33325
asS'd'
I7286
sa(dp14572
S'a'
I13360
sS's'
I17
sS'm'
(lp14573
I1
aI3
aI1
aI9
aI21
aI21
aI21
aI125
aI49
aI787
aI767
aI2831
aI511
aI2461
aI31537
aI27155
aI44053
asS'd'
I7287
sa(dp14574
S'a'
I13369
sS's'
I17
sS'm'
(lp14575
I1
aI3
aI7
aI9
aI31
aI19
aI125
aI67
aI119
aI465
aI287
aI1869
aI3979
aI15723
aI21069
aI8581
aI66939
asS'd'
I7288
sa(dp14576
S'a'
I13372
sS's'
I17
sS'm'
(lp14577
I1
aI1
aI1
aI11
aI7
aI37
aI123
aI237
aI353
aI499
aI113
aI3829
aI217
aI4751
aI7385
aI20343
aI83699
asS'd'
I7289
sa(dp14578
S'a'
I13398
sS's'
I17
sS'm'
(lp14579
I1
aI3
aI7
aI13
aI9
aI3
aI53
aI27
aI487
aI87
aI35
aI2645
aI3481
aI14409
aI27875
aI31695
aI78489
asS'd'
I7290
sa(dp14580
S'a'
I13404
sS's'
I17
sS'm'
(lp14581
I1
aI3
aI3
aI13
aI9
aI43
aI67
aI51
aI153
aI83
aI591
aI1991
aI1787
aI11973
aI7273
aI34801
aI47199
asS'd'
I7291
sa(dp14582
S'a'
I13407
sS's'
I17
sS'm'
(lp14583
I1
aI3
aI7
aI1
aI15
aI53
aI71
aI11
aI205
aI853
aI2011
aI581
aI1281
aI7819
aI23083
aI33731
aI74951
asS'd'
I7292
sa(dp14584
S'a'
I13414
sS's'
I17
sS'm'
(lp14585
I1
aI1
aI5
aI5
aI17
aI63
aI109
aI219
aI225
aI997
aI1251
aI3287
aI1441
aI13489
aI22723
aI45191
aI50249
asS'd'
I7293
sa(dp14586
S'a'
I13426
sS's'
I17
sS'm'
(lp14587
I1
aI1
aI5
aI3
aI13
aI1
aI43
aI53
aI293
aI685
aI1369
aI1515
aI7479
aI3233
aI20007
aI65235
aI102467
asS'd'
I7294
sa(dp14588
S'a'
I13432
sS's'
I17
sS'm'
(lp14589
I1
aI3
aI5
aI7
aI29
aI45
aI63
aI45
aI219
aI445
aI2047
aI317
aI7553
aI325
aI1465
aI949
aI35163
asS'd'
I7295
sa(dp14590
S'a'
I13466
sS's'
I17
sS'm'
(lp14591
I1
aI1
aI3
aI9
aI7
aI31
aI73
aI211
aI501
aI233
aI1495
aI701
aI5857
aI10763
aI9743
aI10289
aI23801
asS'd'
I7296
sa(dp14592
S'a'
I13481
sS's'
I17
sS'm'
(lp14593
I1
aI3
aI7
aI3
aI23
aI47
aI99
aI61
aI179
aI833
aI1425
aI1275
aI4467
aI4367
aI5567
aI23513
aI68677
asS'd'
I7297
sa(dp14594
S'a'
I13490
sS's'
I17
sS'm'
(lp14595
I1
aI1
aI5
aI5
aI27
aI33
aI119
aI229
aI329
aI51
aI1025
aI3167
aI3405
aI4039
aI4135
aI6655
aI43771
asS'd'
I7298
sa(dp14596
S'a'
I13496
sS's'
I17
sS'm'
(lp14597
I1
aI3
aI7
aI15
aI5
aI49
aI91
aI55
aI425
aI15
aI2003
aI1571
aI3539
aI10375
aI29645
aI5889
aI51887
asS'd'
I7299
sa(dp14598
S'a'
I13504
sS's'
I17
sS'm'
(lp14599
I1
aI1
aI7
aI3
aI13
aI55
aI85
aI91
aI181
aI723
aI1941
aI75
aI4443
aI11507
aI7027
aI14189
aI50685
asS'd'
I7300
sa(dp14600
S'a'
I13516
sS's'
I17
sS'm'
(lp14601
I1
aI3
aI5
aI5
aI29
aI49
aI13
aI3
aI97
aI165
aI41
aI3039
aI3325
aI2161
aI775
aI38501
aI42381
asS'd'
I7301
sa(dp14602
S'a'
I13527
sS's'
I17
sS'm'
(lp14603
I1
aI1
aI5
aI11
aI9
aI57
aI47
aI109
aI9
aI585
aI375
aI1839
aI937
aI6877
aI29847
aI60163
aI103081
asS'd'
I7302
sa(dp14604
S'a'
I13533
sS's'
I17
sS'm'
(lp14605
I1
aI1
aI3
aI13
aI5
aI47
aI11
aI195
aI253
aI235
aI275
aI2313
aI163
aI14683
aI5681
aI13381
aI84553
asS'd'
I7303
sa(dp14606
S'a'
I13537
sS's'
I17
sS'm'
(lp14607
I1
aI3
aI3
aI15
aI15
aI1
aI93
aI157
aI437
aI557
aI307
aI1179
aI6857
aI3101
aI16723
aI50579
aI69603
asS'd'
I7304
sa(dp14608
S'a'
I13552
sS's'
I17
sS'm'
(lp14609
I1
aI1
aI5
aI9
aI11
aI29
aI23
aI219
aI337
aI689
aI1155
aI2007
aI6853
aI6749
aI20127
aI13199
aI48433
asS'd'
I7305
sa(dp14610
S'a'
I13561
sS's'
I17
sS'm'
(lp14611
I1
aI1
aI1
aI13
aI1
aI61
aI73
aI213
aI335
aI539
aI903
aI2719
aI775
aI2775
aI29109
aI33367
aI3281
asS'd'
I7306
sa(dp14612
S'a'
I13567
sS's'
I17
sS'm'
(lp14613
I1
aI3
aI1
aI5
aI15
aI31
aI65
aI231
aI439
aI623
aI1871
aI2299
aI5365
aI10333
aI9147
aI2781
aI63813
asS'd'
I7307
sa(dp14614
S'a'
I13582
sS's'
I17
sS'm'
(lp14615
I1
aI1
aI3
aI1
aI1
aI25
aI23
aI229
aI173
aI279
aI181
aI1299
aI2893
aI15475
aI12473
aI46097
aI123387
asS'd'
I7308
sa(dp14616
S'a'
I13587
sS's'
I17
sS'm'
(lp14617
I1
aI1
aI7
aI13
aI7
aI43
aI17
aI187
aI467
aI113
aI1293
aI2013
aI6091
aI14621
aI22195
aI24079
aI45379
asS'd'
I7309
sa(dp14618
S'a'
I13589
sS's'
I17
sS'm'
(lp14619
I1
aI3
aI1
aI7
aI1
aI7
aI119
aI159
aI377
aI11
aI705
aI2853
aI3767
aI13739
aI23375
aI25563
aI73987
asS'd'
I7310
sa(dp14620
S'a'
I13593
sS's'
I17
sS'm'
(lp14621
I1
aI1
aI3
aI15
aI21
aI13
aI111
aI119
aI401
aI1005
aI777
aI1699
aI2431
aI15139
aI27887
aI28415
aI71519
asS'd'
I7311
sa(dp14622
S'a'
I13596
sS's'
I17
sS'm'
(lp14623
I1
aI1
aI7
aI9
aI1
aI49
aI19
aI171
aI297
aI77
aI1343
aI1249
aI5769
aI13889
aI21401
aI24915
aI17641
asS'd'
I7312
sa(dp14624
S'a'
I13615
sS's'
I17
sS'm'
(lp14625
I1
aI1
aI7
aI11
aI31
aI45
aI51
aI231
aI123
aI817
aI13
aI791
aI6235
aI2787
aI475
aI1717
aI5071
asS'd'
I7313
sa(dp14626
S'a'
I13617
sS's'
I17
sS'm'
(lp14627
I1
aI3
aI3
aI13
aI5
aI9
aI21
aI129
aI253
aI731
aI785
aI2275
aI7343
aI7841
aI5477
aI8973
aI101033
asS'd'
I7314
sa(dp14628
S'a'
I13623
sS's'
I17
sS'm'
(lp14629
I1
aI1
aI7
aI13
aI23
aI1
aI119
aI221
aI293
aI709
aI2031
aI3019
aI1529
aI2007
aI10823
aI43193
aI82661
asS'd'
I7315
sa(dp14630
S'a'
I13641
sS's'
I17
sS'm'
(lp14631
I1
aI3
aI1
aI11
aI29
aI29
aI87
aI79
aI415
aI679
aI1899
aI3453
aI7355
aI8627
aI28225
aI41857
aI106645
asS'd'
I7316
sa(dp14632
S'a'
I13647
sS's'
I17
sS'm'
(lp14633
I1
aI1
aI5
aI15
aI9
aI13
aI21
aI241
aI491
aI927
aI999
aI2131
aI3501
aI11063
aI28595
aI54691
aI21297
asS'd'
I7317
sa(dp14634
S'a'
I13650
sS's'
I17
sS'm'
(lp14635
I1
aI1
aI1
aI3
aI5
aI41
aI85
aI89
aI483
aI309
aI791
aI825
aI3043
aI2715
aI16573
aI6551
aI77875
asS'd'
I7318
sa(dp14636
S'a'
I13659
sS's'
I17
sS'm'
(lp14637
I1
aI3
aI1
aI1
aI25
aI21
aI107
aI123
aI79
aI1019
aI821
aI1251
aI4943
aI1429
aI17843
aI37013
aI53285
asS'd'
I7319
sa(dp14638
S'a'
I13671
sS's'
I17
sS'm'
(lp14639
I1
aI1
aI5
aI3
aI7
aI5
aI35
aI123
aI445
aI315
aI627
aI2543
aI1261
aI13737
aI15991
aI36591
aI18309
asS'd'
I7320
sa(dp14640
S'a'
I13677
sS's'
I17
sS'm'
(lp14641
I1
aI3
aI3
aI5
aI25
aI43
aI65
aI249
aI309
aI1023
aI737
aI1933
aI4735
aI7725
aI12063
aI53023
aI126677
asS'd'
I7321
sa(dp14642
S'a'
I13678
sS's'
I17
sS'm'
(lp14643
I1
aI3
aI1
aI9
aI13
aI37
aI77
aI61
aI179
aI275
aI277
aI1431
aI2869
aI14563
aI665
aI60553
aI7661
asS'd'
I7322
sa(dp14644
S'a'
I13680
sS's'
I17
sS'm'
(lp14645
I1
aI3
aI5
aI3
aI29
aI1
aI127
aI73
aI363
aI311
aI1591
aI3863
aI6481
aI4725
aI8287
aI61311
aI39011
asS'd'
I7323
sa(dp14646
S'a'
I13685
sS's'
I17
sS'm'
(lp14647
I1
aI3
aI1
aI7
aI13
aI23
aI115
aI215
aI385
aI563
aI1033
aI2343
aI5023
aI11013
aI12131
aI26997
aI48645
asS'd'
I7324
sa(dp14648
S'a'
I13689
sS's'
I17
sS'm'
(lp14649
I1
aI1
aI5
aI13
aI3
aI59
aI41
aI155
aI263
aI507
aI1175
aI2967
aI7929
aI8237
aI11841
aI15365
aI51881
asS'd'
I7325
sa(dp14650
S'a'
I13701
sS's'
I17
sS'm'
(lp14651
I1
aI1
aI3
aI11
aI19
aI35
aI89
aI115
aI121
aI315
aI1697
aI2121
aI1867
aI6865
aI23639
aI26525
aI44687
asS'd'
I7326
sa(dp14652
S'a'
I13706
sS's'
I17
sS'm'
(lp14653
I1
aI3
aI5
aI15
aI9
aI5
aI125
aI183
aI149
aI447
aI309
aI1743
aI6089
aI369
aI16153
aI63799
aI57657
asS'd'
I7327
sa(dp14654
S'a'
I13720
sS's'
I17
sS'm'
(lp14655
I1
aI1
aI1
aI7
aI7
aI39
aI89
aI139
aI457
aI741
aI1613
aI2883
aI5057
aI12495
aI18669
aI55469
aI97941
asS'd'
I7328
sa(dp14656
S'a'
I13729
sS's'
I17
sS'm'
(lp14657
I1
aI1
aI5
aI7
aI29
aI39
aI97
aI9
aI481
aI667
aI1353
aI3387
aI2813
aI16205
aI8353
aI22121
aI92965
asS'd'
I7329
sa(dp14658
S'a'
I13735
sS's'
I17
sS'm'
(lp14659
I1
aI3
aI7
aI9
aI11
aI55
aI79
aI159
aI349
aI717
aI829
aI3157
aI1457
aI6199
aI5861
aI2553
aI20387
asS'd'
I7330
sa(dp14660
S'a'
I13736
sS's'
I17
sS'm'
(lp14661
I1
aI1
aI1
aI11
aI3
aI51
aI113
aI53
aI287
aI109
aI1717
aI2405
aI7207
aI4473
aI11145
aI2549
aI591
asS'd'
I7331
sa(dp14662
S'a'
I13756
sS's'
I17
sS'm'
(lp14663
I1
aI1
aI3
aI13
aI3
aI61
aI31
aI141
aI217
aI487
aI299
aI2755
aI3389
aI10053
aI1105
aI21129
aI74203
asS'd'
I7332
sa(dp14664
S'a'
I13759
sS's'
I17
sS'm'
(lp14665
I1
aI1
aI1
aI3
aI11
aI55
aI7
aI113
aI413
aI449
aI787
aI3279
aI5123
aI16025
aI15005
aI12175
aI6795
asS'd'
I7333
sa(dp14666
S'a'
I13761
sS's'
I17
sS'm'
(lp14667
I1
aI3
aI1
aI1
aI25
aI23
aI107
aI191
aI3
aI3
aI49
aI1083
aI3275
aI10385
aI7989
aI53739
aI25505
asS'd'
I7334
sa(dp14668
S'a'
I13771
sS's'
I17
sS'm'
(lp14669
I1
aI1
aI5
aI13
aI7
aI17
aI59
aI13
aI471
aI147
aI1627
aI2119
aI3555
aI15555
aI10333
aI49363
aI80959
asS'd'
I7335
sa(dp14670
S'a'
I13782
sS's'
I17
sS'm'
(lp14671
I1
aI1
aI1
aI15
aI23
aI33
aI61
aI191
aI207
aI939
aI45
aI2781
aI71
aI9661
aI28433
aI13089
aI76419
asS'd'
I7336
sa(dp14672
S'a'
I13786
sS's'
I17
sS'm'
(lp14673
I1
aI3
aI3
aI7
aI29
aI47
aI111
aI19
aI315
aI381
aI851
aI1303
aI2627
aI6255
aI30369
aI37723
aI12949
asS'd'
I7337
sa(dp14674
S'a'
I13798
sS's'
I17
sS'm'
(lp14675
I1
aI1
aI1
aI13
aI31
aI43
aI3
aI193
aI5
aI99
aI769
aI2523
aI1949
aI129
aI9693
aI60535
aI67059
asS'd'
I7338
sa(dp14676
S'a'
I13810
sS's'
I17
sS'm'
(lp14677
I1
aI1
aI3
aI15
aI5
aI33
aI73
aI149
aI253
aI985
aI863
aI1551
aI4369
aI5911
aI8269
aI35463
aI117055
asS'd'
I7339
sa(dp14678
S'a'
I13819
sS's'
I17
sS'm'
(lp14679
I1
aI3
aI1
aI5
aI27
aI57
aI3
aI105
aI253
aI731
aI119
aI3287
aI613
aI4627
aI22003
aI56027
aI123005
asS'd'
I7340
sa(dp14680
S'a'
I13826
sS's'
I17
sS'm'
(lp14681
I1
aI1
aI3
aI3
aI27
aI47
aI67
aI147
aI495
aI865
aI1233
aI3707
aI2511
aI2951
aI7367
aI15625
aI86417
asS'd'
I7341
sa(dp14682
S'a'
I13831
sS's'
I17
sS'm'
(lp14683
I1
aI1
aI7
aI1
aI7
aI7
aI13
aI255
aI457
aI529
aI953
aI1481
aI5565
aI12495
aI4723
aI41615
aI121829
asS'd'
I7342
sa(dp14684
S'a'
I13837
sS's'
I17
sS'm'
(lp14685
I1
aI3
aI5
aI11
aI1
aI51
aI91
aI153
aI323
aI609
aI1353
aI2995
aI4035
aI13835
aI28619
aI46217
aI4967
asS'd'
I7343
sa(dp14686
S'a'
I13846
sS's'
I17
sS'm'
(lp14687
I1
aI1
aI5
aI7
aI25
aI59
aI81
aI101
aI185
aI709
aI1249
aI2285
aI6579
aI8655
aI17563
aI9707
aI63845
asS'd'
I7344
sa(dp14688
S'a'
I13856
sS's'
I17
sS'm'
(lp14689
I1
aI3
aI3
aI9
aI31
aI25
aI17
aI19
aI111
aI627
aI1187
aI2621
aI6529
aI9457
aI25027
aI18069
aI47559
asS'd'
I7345
sa(dp14690
S'a'
I13862
sS's'
I17
sS'm'
(lp14691
I1
aI3
aI7
aI15
aI7
aI15
aI103
aI201
aI391
aI1023
aI817
aI535
aI2713
aI1317
aI13469
aI56043
aI70847
asS'd'
I7346
sa(dp14692
S'a'
I13866
sS's'
I17
sS'm'
(lp14693
I1
aI1
aI3
aI7
aI17
aI57
aI35
aI99
aI439
aI367
aI27
aI2695
aI3519
aI8337
aI14047
aI58489
aI69
asS'd'
I7347
sa(dp14694
S'a'
I13885
sS's'
I17
sS'm'
(lp14695
I1
aI1
aI1
aI3
aI17
aI23
aI71
aI189
aI57
aI39
aI715
aI1779
aI3081
aI14657
aI21895
aI59203
aI31005
asS'd'
I7348
sa(dp14696
S'a'
I13891
sS's'
I17
sS'm'
(lp14697
I1
aI1
aI7
aI13
aI1
aI47
aI69
aI159
aI353
aI517
aI271
aI973
aI5077
aI15707
aI11095
aI19671
aI3389
asS'd'
I7349
sa(dp14698
S'a'
I13893
sS's'
I17
sS'm'
(lp14699
I1
aI1
aI7
aI13
aI25
aI55
aI115
aI21
aI43
aI939
aI1697
aI101
aI4751
aI1993
aI2389
aI28353
aI45251
asS'd'
I7350
sa(dp14700
S'a'
I13905
sS's'
I17
sS'm'
(lp14701
I1
aI3
aI1
aI15
aI11
aI57
aI17
aI49
aI121
aI419
aI909
aI121
aI5047
aI4235
aI13051
aI21529
aI42097
asS'd'
I7351
sa(dp14702
S'a'
I13908
sS's'
I17
sS'm'
(lp14703
I1
aI1
aI1
aI3
aI19
aI37
aI31
aI233
aI251
aI175
aI929
aI1527
aI7527
aI3605
aI17075
aI61053
aI56235
asS'd'
I7352
sa(dp14704
S'a'
I13912
sS's'
I17
sS'm'
(lp14705
I1
aI1
aI1
aI3
aI9
aI5
aI117
aI131
aI251
aI475
aI1695
aI1381
aI2445
aI5921
aI14921
aI937
aI80791
asS'd'
I7353
sa(dp14706
S'a'
I13917
sS's'
I17
sS'm'
(lp14707
I1
aI1
aI3
aI9
aI11
aI5
aI31
aI215
aI37
aI567
aI1537
aI2183
aI3291
aI1601
aI14025
aI48807
aI7243
asS'd'
I7354
sa(dp14708
S'a'
I13918
sS's'
I17
sS'm'
(lp14709
I1
aI3
aI1
aI9
aI7
aI13
aI81
aI249
aI321
aI473
aI1419
aI3977
aI7037
aI14191
aI10865
aI56131
aI43225
asS'd'
I7355
sa(dp14710
S'a'
I13946
sS's'
I17
sS'm'
(lp14711
I1
aI1
aI1
aI13
aI15
aI23
aI31
aI69
aI449
aI491
aI1461
aI729
aI7955
aI4003
aI16817
aI37273
aI72025
asS'd'
I7356
sa(dp14712
S'a'
I13948
sS's'
I17
sS'm'
(lp14713
I1
aI1
aI5
aI13
aI7
aI41
aI93
aI169
aI347
aI1013
aI301
aI2813
aI1455
aI13187
aI10769
aI60807
aI46333
asS'd'
I7357
sa(dp14714
S'a'
I13964
sS's'
I17
sS'm'
(lp14715
I1
aI1
aI5
aI3
aI23
aI15
aI1
aI161
aI29
aI35
aI415
aI235
aI93
aI14543
aI29585
aI29657
aI36489
asS'd'
I7358
sa(dp14716
S'a'
I13970
sS's'
I17
sS'm'
(lp14717
I1
aI3
aI1
aI3
aI31
aI63
aI39
aI235
aI153
aI549
aI43
aI147
aI2317
aI3537
aI25561
aI58287
aI58725
asS'd'
I7359
sa(dp14718
S'a'
I13975
sS's'
I17
sS'm'
(lp14719
I1
aI1
aI3
aI5
aI5
aI11
aI59
aI97
aI349
aI307
aI501
aI1701
aI4243
aI13717
aI17419
aI23387
aI29533
asS'd'
I7360
sa(dp14720
S'a'
I13979
sS's'
I17
sS'm'
(lp14721
I1
aI3
aI1
aI7
aI7
aI19
aI33
aI243
aI67
aI353
aI2023
aI3111
aI7173
aI10979
aI28117
aI40175
aI45337
asS'd'
I7361
sa(dp14722
S'a'
I13997
sS's'
I17
sS'm'
(lp14723
I1
aI1
aI1
aI5
aI15
aI59
aI55
aI135
aI107
aI543
aI1743
aI2695
aI3293
aI111
aI32629
aI8249
aI52273
asS'd'
I7362
sa(dp14724
S'a'
I14000
sS's'
I17
sS'm'
(lp14725
I1
aI3
aI5
aI13
aI15
aI57
aI39
aI79
aI5
aI451
aI571
aI1445
aI1393
aI2125
aI31713
aI59655
aI20897
asS'd'
I7363
sa(dp14726
S'a'
I14006
sS's'
I17
sS'm'
(lp14727
I1
aI1
aI3
aI11
aI29
aI61
aI1
aI37
aI173
aI513
aI1779
aI2649
aI3289
aI4679
aI2039
aI47587
aI28973
asS'd'
I7364
sa(dp14728
S'a'
I14020
sS's'
I17
sS'm'
(lp14729
I1
aI1
aI5
aI5
aI15
aI19
aI17
aI143
aI387
aI359
aI275
aI625
aI7383
aI15537
aI10311
aI40005
aI20729
asS'd'
I7365
sa(dp14730
S'a'
I14023
sS's'
I17
sS'm'
(lp14731
I1
aI1
aI3
aI9
aI7
aI23
aI71
aI179
aI85
aI447
aI345
aI3459
aI2857
aI8331
aI5489
aI62207
aI64933
asS'd'
I7366
sa(dp14732
S'a'
I14024
sS's'
I17
sS'm'
(lp14733
I1
aI1
aI1
aI1
aI11
aI61
aI47
aI131
aI213
aI611
aI701
aI713
aI1269
aI9563
aI25223
aI50697
aI88679
asS'd'
I7367
sa(dp14734
S'a'
I14029
sS's'
I17
sS'm'
(lp14735
I1
aI1
aI5
aI15
aI21
aI5
aI77
aI59
aI455
aI243
aI459
aI2809
aI13
aI9325
aI32047
aI3939
aI48389
asS'd'
I7368
sa(dp14736
S'a'
I14035
sS's'
I17
sS'm'
(lp14737
I1
aI3
aI7
aI1
aI21
aI53
aI111
aI225
aI407
aI119
aI713
aI3635
aI1539
aI15321
aI29827
aI36069
aI74483
asS'd'
I7369
sa(dp14738
S'a'
I14044
sS's'
I17
sS'm'
(lp14739
I1
aI1
aI5
aI13
aI7
aI45
aI75
aI43
aI191
aI715
aI169
aI759
aI33
aI11329
aI1069
aI36103
aI28055
asS'd'
I7370
sa(dp14740
S'a'
I14047
sS's'
I17
sS'm'
(lp14741
I1
aI3
aI7
aI5
aI7
aI13
aI7
aI35
aI27
aI391
aI517
aI1439
aI5699
aI1067
aI23857
aI7293
aI66167
asS'd'
I7371
sa(dp14742
S'a'
I14058
sS's'
I17
sS'm'
(lp14743
I1
aI1
aI7
aI11
aI3
aI31
aI1
aI83
aI299
aI345
aI65
aI669
aI1529
aI7569
aI28959
aI50561
aI69493
asS'd'
I7372
sa(dp14744
S'a'
I14066
sS's'
I17
sS'm'
(lp14745
I1
aI3
aI1
aI5
aI25
aI25
aI43
aI149
aI83
aI225
aI1589
aI1691
aI7777
aI773
aI10421
aI49523
aI23533
asS'd'
I7373
sa(dp14746
S'a'
I14075
sS's'
I17
sS'm'
(lp14747
I1
aI1
aI5
aI11
aI25
aI29
aI81
aI11
aI497
aI43
aI951
aI2551
aI821
aI13805
aI12315
aI61299
aI81397
asS'd'
I7374
sa(dp14748
S'a'
I14080
sS's'
I17
sS'm'
(lp14749
I1
aI3
aI1
aI9
aI29
aI23
aI109
aI123
aI235
aI255
aI1519
aI3289
aI7761
aI14575
aI11851
aI1719
aI51655
asS'd'
I7375
sa(dp14750
S'a'
I14095
sS's'
I17
sS'm'
(lp14751
I1
aI3
aI5
aI15
aI21
aI49
aI13
aI43
aI87
aI517
aI687
aI1457
aI1501
aI15959
aI31907
aI13771
aI69379
asS'd'
I7376
sa(dp14752
S'a'
I14100
sS's'
I17
sS'm'
(lp14753
I1
aI1
aI5
aI3
aI21
aI11
aI87
aI9
aI343
aI317
aI845
aI1663
aI7933
aI14063
aI24915
aI31487
aI17445
asS'd'
I7377
sa(dp14754
S'a'
I14114
sS's'
I17
sS'm'
(lp14755
I1
aI1
aI3
aI13
aI21
aI31
aI87
aI99
aI185
aI333
aI993
aI3899
aI971
aI2851
aI23643
aI195
aI66957
asS'd'
I7378
sa(dp14756
S'a'
I14116
sS's'
I17
sS'm'
(lp14757
I1
aI1
aI1
aI15
aI19
aI47
aI23
aI1
aI67
aI57
aI165
aI3903
aI421
aI10561
aI11621
aI13815
aI10349
asS'd'
I7379
sa(dp14758
S'a'
I14123
sS's'
I17
sS'm'
(lp14759
I1
aI3
aI5
aI11
aI9
aI19
aI73
aI17
aI229
aI913
aI459
aI3809
aI2667
aI9775
aI3693
aI52945
aI90837
asS'd'
I7380
sa(dp14760
S'a'
I14134
sS's'
I17
sS'm'
(lp14761
I1
aI1
aI5
aI15
aI3
aI25
aI109
aI131
aI507
aI637
aI1615
aI859
aI6785
aI14891
aI24801
aI39095
aI79557
asS'd'
I7381
sa(dp14762
S'a'
I14143
sS's'
I17
sS'm'
(lp14763
I1
aI1
aI5
aI7
aI1
aI51
aI71
aI251
aI19
aI799
aI835
aI1119
aI2349
aI15083
aI16509
aI55621
aI123501
asS'd'
I7382
sa(dp14764
S'a'
I14151
sS's'
I17
sS'm'
(lp14765
I1
aI3
aI5
aI9
aI13
aI39
aI127
aI1
aI233
aI37
aI735
aI3307
aI5163
aI4529
aI5961
aI12893
aI103641
asS'd'
I7383
sa(dp14766
S'a'
I14160
sS's'
I17
sS'm'
(lp14767
I1
aI1
aI7
aI5
aI23
aI15
aI49
aI123
aI511
aI201
aI2025
aI289
aI3847
aI15755
aI24279
aI52543
aI42017
asS'd'
I7384
sa(dp14768
S'a'
I14163
sS's'
I17
sS'm'
(lp14769
I1
aI1
aI5
aI3
aI9
aI61
aI19
aI37
aI3
aI361
aI1065
aI2971
aI2517
aI1259
aI27359
aI3823
aI60181
asS'd'
I7385
sa(dp14770
S'a'
I14179
sS's'
I17
sS'm'
(lp14771
I1
aI3
aI1
aI7
aI15
aI17
aI57
aI249
aI57
aI979
aI147
aI2407
aI2579
aI3159
aI8467
aI8433
aI72873
asS'd'
I7386
sa(dp14772
S'a'
I14181
sS's'
I17
sS'm'
(lp14773
I1
aI1
aI3
aI1
aI25
aI7
aI47
aI117
aI449
aI321
aI143
aI3867
aI165
aI7961
aI27597
aI10033
aI2437
asS'd'
I7387
sa(dp14774
S'a'
I14193
sS's'
I17
sS'm'
(lp14775
I1
aI3
aI5
aI13
aI19
aI49
aI1
aI83
aI477
aI549
aI509
aI2911
aI1559
aI14017
aI10469
aI62171
aI82829
asS'd'
I7388
sa(dp14776
S'a'
I14209
sS's'
I17
sS'm'
(lp14777
I1
aI3
aI3
aI7
aI27
aI21
aI15
aI63
aI31
aI45
aI1223
aI3903
aI5469
aI11983
aI29627
aI27453
aI32019
asS'd'
I7389
sa(dp14778
S'a'
I14210
sS's'
I17
sS'm'
(lp14779
I1
aI1
aI7
aI7
aI9
aI9
aI7
aI77
aI349
aI467
aI61
aI3465
aI6921
aI15761
aI15179
aI38649
aI2469
asS'd'
I7390
sa(dp14780
S'a'
I14224
sS's'
I17
sS'm'
(lp14781
I1
aI3
aI1
aI13
aI9
aI59
aI55
aI67
aI271
aI617
aI643
aI4071
aI7963
aI8153
aI5121
aI43917
aI26219
asS'd'
I7391
sa(dp14782
S'a'
I14245
sS's'
I17
sS'm'
(lp14783
I1
aI1
aI3
aI7
aI29
aI21
aI63
aI103
aI327
aI623
aI931
aI1511
aI3125
aI229
aI28949
aI61315
aI72667
asS'd'
I7392
sa(dp14784
S'a'
I14249
sS's'
I17
sS'm'
(lp14785
I1
aI3
aI7
aI1
aI19
aI37
aI49
aI63
aI403
aI885
aI161
aI121
aI1447
aI9227
aI15019
aI50049
aI26939
asS'd'
I7393
sa(dp14786
S'a'
I14255
sS's'
I17
sS'm'
(lp14787
I1
aI3
aI3
aI3
aI23
aI57
aI95
aI79
aI485
aI173
aI93
aI835
aI7161
aI11247
aI3485
aI5759
aI36393
asS'd'
I7394
sa(dp14788
S'a'
I14267
sS's'
I17
sS'm'
(lp14789
I1
aI3
aI7
aI13
aI23
aI33
aI5
aI97
aI235
aI531
aI313
aI2925
aI2223
aI847
aI18591
aI15477
aI3129
asS'd'
I7395
sa(dp14790
S'a'
I14270
sS's'
I17
sS'm'
(lp14791
I1
aI1
aI3
aI13
aI25
aI25
aI101
aI183
aI477
aI947
aI1251
aI2631
aI7987
aI13417
aI23759
aI55305
aI123817
asS'd'
I7396
sa(dp14792
S'a'
I14277
sS's'
I17
sS'm'
(lp14793
I1
aI1
aI5
aI9
aI27
aI63
aI49
aI137
aI179
aI861
aI33
aI2375
aI3827
aI6485
aI19689
aI7867
aI124429
asS'd'
I7397
sa(dp14794
S'a'
I14305
sS's'
I17
sS'm'
(lp14795
I1
aI3
aI7
aI3
aI15
aI43
aI63
aI103
aI45
aI947
aI1837
aI833
aI7055
aI7487
aI19669
aI12045
aI78377
asS'd'
I7398
sa(dp14796
S'a'
I14308
sS's'
I17
sS'm'
(lp14797
I1
aI3
aI5
aI3
aI29
aI35
aI57
aI19
aI471
aI985
aI1147
aI741
aI5403
aI10057
aI25375
aI50889
aI82719
asS'd'
I7399
sa(dp14798
S'a'
I14312
sS's'
I17
sS'm'
(lp14799
I1
aI3
aI3
aI1
aI17
aI19
aI111
aI13
aI121
aI821
aI1831
aI4043
aI123
aI9529
aI1511
aI10917
aI105961
asS'd'
I7400
sa(dp14800
S'a'
I14325
sS's'
I17
sS'm'
(lp14801
I1
aI1
aI3
aI11
aI1
aI43
aI23
aI75
aI345
aI9
aI1379
aI2157
aI5887
aI1197
aI14849
aI17103
aI91925
asS'd'
I7401
sa(dp14802
S'a'
I14332
sS's'
I17
sS'm'
(lp14803
I1
aI1
aI3
aI3
aI19
aI11
aI1
aI179
aI343
aI1023
aI1801
aI915
aI255
aI519
aI5787
aI32913
aI43471
asS'd'
I7402
sa(dp14804
S'a'
I14345
sS's'
I17
sS'm'
(lp14805
I1
aI3
aI5
aI5
aI3
aI3
aI3
aI211
aI461
aI55
aI851
aI3165
aI2903
aI15077
aI8537
aI2037
aI109057
asS'd'
I7403
sa(dp14806
S'a'
I14354
sS's'
I17
sS'm'
(lp14807
I1
aI1
aI7
aI15
aI7
aI7
aI43
aI249
aI27
aI511
aI1369
aI735
aI6093
aI12575
aI26675
aI21745
aI117053
asS'd'
I7404
sa(dp14808
S'a'
I14372
sS's'
I17
sS'm'
(lp14809
I1
aI1
aI5
aI7
aI21
aI53
aI45
aI83
aI415
aI645
aI325
aI4027
aI5181
aI8485
aI1917
aI55623
aI45203
asS'd'
I7405
sa(dp14810
S'a'
I14387
sS's'
I17
sS'm'
(lp14811
I1
aI1
aI3
aI15
aI7
aI1
aI121
aI221
aI387
aI403
aI1877
aI1671
aI2113
aI2379
aI5667
aI39867
aI8079
asS'd'
I7406
sa(dp14812
S'a'
I14390
sS's'
I17
sS'm'
(lp14813
I1
aI1
aI1
aI7
aI5
aI29
aI35
aI77
aI197
aI661
aI1859
aI2539
aI4045
aI13497
aI305
aI44987
aI31215
asS'd'
I7407
sa(dp14814
S'a'
I14402
sS's'
I17
sS'm'
(lp14815
I1
aI1
aI5
aI5
aI13
aI37
aI13
aI85
aI287
aI347
aI579
aI2283
aI7911
aI5463
aI21141
aI9035
aI105067
asS'd'
I7408
sa(dp14816
S'a'
I14408
sS's'
I17
sS'm'
(lp14817
I1
aI1
aI1
aI9
aI17
aI17
aI63
aI97
aI57
aI629
aI1917
aI1133
aI779
aI12365
aI17127
aI52549
aI18755
asS'd'
I7409
sa(dp14818
S'a'
I14413
sS's'
I17
sS'm'
(lp14819
I1
aI1
aI7
aI11
aI7
aI17
aI65
aI137
aI485
aI841
aI653
aI2921
aI4935
aI16273
aI23333
aI7399
aI43129
asS'd'
I7410
sa(dp14820
S'a'
I14431
sS's'
I17
sS'm'
(lp14821
I1
aI3
aI1
aI11
aI31
aI55
aI93
aI225
aI319
aI35
aI947
aI1909
aI7733
aI8303
aI20739
aI55713
aI6633
asS'd'
I7411
sa(dp14822
S'a'
I14438
sS's'
I17
sS'm'
(lp14823
I1
aI1
aI1
aI3
aI11
aI25
aI1
aI165
aI305
aI275
aI607
aI3845
aI5203
aI1989
aI13803
aI597
aI39751
asS'd'
I7412
sa(dp14824
S'a'
I14447
sS's'
I17
sS'm'
(lp14825
I1
aI1
aI5
aI11
aI31
aI43
aI83
aI237
aI453
aI59
aI457
aI741
aI411
aI15895
aI18891
aI30133
aI66767
asS'd'
I7413
sa(dp14826
S'a'
I14455
sS's'
I17
sS'm'
(lp14827
I1
aI3
aI5
aI11
aI3
aI23
aI65
aI81
aI299
aI527
aI1057
aI2731
aI3839
aI6023
aI28887
aI64929
aI41405
asS'd'
I7414
sa(dp14828
S'a'
I14461
sS's'
I17
sS'm'
(lp14829
I1
aI3
aI1
aI1
aI3
aI5
aI11
aI169
aI123
aI957
aI1495
aI1717
aI4079
aI13239
aI28339
aI33677
aI30591
asS'd'
I7415
sa(dp14830
S'a'
I14466
sS's'
I17
sS'm'
(lp14831
I1
aI1
aI7
aI15
aI3
aI1
aI37
aI245
aI169
aI273
aI2039
aI415
aI6555
aI13131
aI11181
aI62179
aI36885
asS'd'
I7416
sa(dp14832
S'a'
I14480
sS's'
I17
sS'm'
(lp14833
I1
aI1
aI3
aI11
aI1
aI55
aI19
aI19
aI425
aI113
aI1367
aI2101
aI5581
aI985
aI2475
aI53983
aI68999
asS'd'
I7417
sa(dp14834
S'a'
I14490
sS's'
I17
sS'm'
(lp14835
I1
aI1
aI5
aI9
aI5
aI33
aI101
aI193
aI303
aI579
aI1265
aI2791
aI479
aI12083
aI17609
aI31801
aI113089
asS'd'
I7418
sa(dp14836
S'a'
I14492
sS's'
I17
sS'm'
(lp14837
I1
aI1
aI3
aI3
aI17
aI61
aI59
aI249
aI81
aI821
aI1
aI431
aI5327
aI8675
aI23469
aI15349
aI67711
asS'd'
I7419
sa(dp14838
S'a'
I14508
sS's'
I17
sS'm'
(lp14839
I1
aI1
aI7
aI9
aI31
aI51
aI89
aI19
aI469
aI843
aI561
aI559
aI4823
aI7803
aI31699
aI44537
aI56835
asS'd'
I7420
sa(dp14840
S'a'
I14513
sS's'
I17
sS'm'
(lp14841
I1
aI3
aI7
aI9
aI11
aI57
aI27
aI43
aI469
aI655
aI433
aI3081
aI6719
aI6651
aI30823
aI61503
aI110711
asS'd'
I7421
sa(dp14842
S'a'
I14516
sS's'
I17
sS'm'
(lp14843
I1
aI3
aI5
aI11
aI9
aI53
aI25
aI147
aI61
aI533
aI1369
aI879
aI7935
aI13829
aI26655
aI17327
aI52983
asS'd'
I7422
sa(dp14844
S'a'
I14519
sS's'
I17
sS'm'
(lp14845
I1
aI3
aI7
aI11
aI15
aI27
aI97
aI175
aI435
aI53
aI75
aI807
aI549
aI5277
aI1831
aI19421
aI55669
asS'd'
I7423
sa(dp14846
S'a'
I14525
sS's'
I17
sS'm'
(lp14847
I1
aI1
aI7
aI15
aI23
aI5
aI99
aI133
aI485
aI587
aI65
aI2585
aI7667
aI2783
aI19437
aI52769
aI1587
asS'd'
I7424
sa(dp14848
S'a'
I14534
sS's'
I17
sS'm'
(lp14849
I1
aI1
aI7
aI7
aI13
aI39
aI111
aI165
aI489
aI355
aI1963
aI333
aI2993
aI5233
aI9173
aI18951
aI93737
asS'd'
I7425
sa(dp14850
S'a'
I14537
sS's'
I17
sS'm'
(lp14851
I1
aI1
aI5
aI7
aI1
aI29
aI67
aI135
aI427
aI91
aI53
aI3109
aI3745
aI9529
aI17567
aI42361
aI84577
asS'd'
I7426
sa(dp14852
S'a'
I14543
sS's'
I17
sS'm'
(lp14853
I1
aI3
aI5
aI1
aI31
aI35
aI59
aI181
aI87
aI345
aI1975
aI781
aI603
aI16365
aI19453
aI9933
aI112739
asS'd'
I7427
sa(dp14854
S'a'
I14545
sS's'
I17
sS'm'
(lp14855
I1
aI3
aI3
aI1
aI31
aI41
aI127
aI35
aI263
aI403
aI1811
aI383
aI1523
aI8477
aI5973
aI41569
aI99309
asS'd'
I7428
sa(dp14856
S'a'
I14552
sS's'
I17
sS'm'
(lp14857
I1
aI3
aI7
aI7
aI5
aI25
aI11
aI201
aI231
aI679
aI519
aI2481
aI7415
aI12397
aI21265
aI49419
aI13903
asS'd'
I7429
sa(dp14858
S'a'
I14562
sS's'
I17
sS'm'
(lp14859
I1
aI1
aI7
aI5
aI1
aI11
aI63
aI221
aI327
aI509
aI419
aI871
aI7891
aI11835
aI11099
aI10669
aI43853
asS'd'
I7430
sa(dp14860
S'a'
I14571
sS's'
I17
sS'm'
(lp14861
I1
aI1
aI5
aI11
aI19
aI11
aI37
aI105
aI265
aI513
aI1013
aI707
aI6083
aI14571
aI17573
aI7645
aI5363
asS'd'
I7431
sa(dp14862
S'a'
I14574
sS's'
I17
sS'm'
(lp14863
I1
aI1
aI1
aI13
aI19
aI19
aI67
aI93
aI113
aI509
aI1013
aI4037
aI1939
aI7015
aI24487
aI57183
aI123463
asS'd'
I7432
sa(dp14864
S'a'
I14582
sS's'
I17
sS'm'
(lp14865
I1
aI1
aI1
aI1
aI21
aI17
aI95
aI25
aI261
aI1005
aI685
aI691
aI4467
aI14723
aI24043
aI32287
aI19651
asS'd'
I7433
sa(dp14866
S'a'
I14611
sS's'
I17
sS'm'
(lp14867
I1
aI3
aI1
aI15
aI15
aI15
aI57
aI191
aI27
aI719
aI229
aI1977
aI241
aI9021
aI21335
aI30967
aI81207
asS'd'
I7434
sa(dp14868
S'a'
I14614
sS's'
I17
sS'm'
(lp14869
I1
aI3
aI1
aI9
aI23
aI61
aI103
aI67
aI361
aI925
aI811
aI1007
aI5707
aI11479
aI5907
aI3897
aI65141
asS'd'
I7435
sa(dp14870
S'a'
I14620
sS's'
I17
sS'm'
(lp14871
I1
aI3
aI5
aI9
aI17
aI61
aI11
aI15
aI351
aI715
aI939
aI2141
aI4857
aI8397
aI9693
aI26845
aI120007
asS'd'
I7436
sa(dp14872
S'a'
I14633
sS's'
I17
sS'm'
(lp14873
I1
aI3
aI1
aI5
aI19
aI55
aI99
aI19
aI291
aI309
aI287
aI1969
aI4341
aI7579
aI30909
aI37277
aI54927
asS'd'
I7437
sa(dp14874
S'a'
I14641
sS's'
I17
sS'm'
(lp14875
I1
aI3
aI7
aI3
aI19
aI29
aI43
aI163
aI367
aI753
aI1733
aI1463
aI7927
aI10671
aI16817
aI41229
aI113887
asS'd'
I7438
sa(dp14876
S'a'
I14648
sS's'
I17
sS'm'
(lp14877
I1
aI3
aI7
aI1
aI11
aI51
aI39
aI207
aI283
aI73
aI1423
aI2473
aI7593
aI3581
aI30179
aI6369
aI112217
asS'd'
I7439
sa(dp14878
S'a'
I14671
sS's'
I17
sS'm'
(lp14879
I1
aI1
aI3
aI15
aI15
aI25
aI43
aI5
aI271
aI611
aI959
aI537
aI303
aI3659
aI18073
aI8147
aI81531
asS'd'
I7440
sa(dp14880
S'a'
I14674
sS's'
I17
sS'm'
(lp14881
I1
aI3
aI7
aI1
aI27
aI55
aI77
aI11
aI367
aI209
aI1967
aI3409
aI935
aI5309
aI18857
aI46225
aI8367
asS'd'
I7441
sa(dp14882
S'a'
I14689
sS's'
I17
sS'm'
(lp14883
I1
aI1
aI5
aI11
aI11
aI63
aI75
aI73
aI43
aI869
aI2021
aI3285
aI269
aI9113
aI32699
aI2091
aI17327
asS'd'
I7442
sa(dp14884
S'a'
I14690
sS's'
I17
sS'm'
(lp14885
I1
aI1
aI5
aI11
aI9
aI25
aI31
aI245
aI109
aI805
aI1645
aI3607
aI817
aI9571
aI12767
aI65441
aI129977
asS'd'
I7443
sa(dp14886
S'a'
I14692
sS's'
I17
sS'm'
(lp14887
I1
aI3
aI7
aI5
aI11
aI61
aI67
aI223
aI433
aI387
aI935
aI1615
aI7915
aI6133
aI24087
aI55323
aI100619
asS'd'
I7444
sa(dp14888
S'a'
I14699
sS's'
I17
sS'm'
(lp14889
I1
aI1
aI1
aI15
aI25
aI61
aI7
aI39
aI311
aI353
aI183
aI33
aI2591
aI4951
aI31377
aI9081
aI9707
asS'd'
I7445
sa(dp14890
S'a'
I14710
sS's'
I17
sS'm'
(lp14891
I1
aI1
aI3
aI3
aI1
aI9
aI65
aI229
aI185
aI47
aI1255
aI1365
aI2231
aI6843
aI26927
aI27195
aI60651
asS'd'
I7446
sa(dp14892
S'a'
I14719
sS's'
I17
sS'm'
(lp14893
I1
aI1
aI7
aI5
aI7
aI25
aI91
aI133
aI159
aI737
aI1767
aI3117
aI7321
aI6159
aI3361
aI27793
aI33473
asS'd'
I7447
sa(dp14894
S'a'
I14730
sS's'
I17
sS'm'
(lp14895
I1
aI3
aI7
aI3
aI11
aI7
aI5
aI125
aI369
aI951
aI1277
aI65
aI7703
aI1817
aI11773
aI25657
aI67045
asS'd'
I7448
sa(dp14896
S'a'
I14732
sS's'
I17
sS'm'
(lp14897
I1
aI1
aI3
aI9
aI21
aI27
aI21
aI41
aI131
aI605
aI1
aI119
aI1553
aI1361
aI31973
aI43135
aI119321
asS'd'
I7449
sa(dp14898
S'a'
I14743
sS's'
I17
sS'm'
(lp14899
I1
aI3
aI7
aI1
aI25
aI63
aI55
aI173
aI323
aI403
aI1401
aI1367
aI3455
aI15335
aI13045
aI20759
aI8309
asS'd'
I7450
sa(dp14900
S'a'
I14744
sS's'
I17
sS'm'
(lp14901
I1
aI1
aI3
aI5
aI3
aI61
aI59
aI7
aI39
aI439
aI721
aI2829
aI3035
aI2293
aI32015
aI28509
aI104831
asS'd'
I7451
sa(dp14902
S'a'
I14750
sS's'
I17
sS'm'
(lp14903
I1
aI3
aI5
aI1
aI29
aI35
aI71
aI87
aI351
aI917
aI1661
aI547
aI4501
aI7107
aI5493
aI17833
aI130729
asS'd'
I7452
sa(dp14904
S'a'
I14759
sS's'
I17
sS'm'
(lp14905
I1
aI1
aI5
aI5
aI7
aI5
aI69
aI57
aI319
aI595
aI1749
aI3789
aI1437
aI6327
aI24089
aI7387
aI125109
asS'd'
I7453
sa(dp14906
S'a'
I14763
sS's'
I17
sS'm'
(lp14907
I1
aI3
aI5
aI9
aI15
aI53
aI95
aI59
aI217
aI37
aI1561
aI401
aI5259
aI4361
aI1049
aI3437
aI30559
asS'd'
I7454
sa(dp14908
S'a'
I14768
sS's'
I17
sS'm'
(lp14909
I1
aI3
aI7
aI13
aI15
aI15
aI107
aI167
aI475
aI157
aI1565
aI2219
aI1891
aI1433
aI11829
aI43433
aI48111
asS'd'
I7455
sa(dp14910
S'a'
I14773
sS's'
I17
sS'm'
(lp14911
I1
aI1
aI1
aI3
aI11
aI41
aI25
aI211
aI243
aI355
aI1831
aI2093
aI2747
aI2523
aI9885
aI9503
aI120089
asS'd'
I7456
sa(dp14912
S'a'
I14777
sS's'
I17
sS'm'
(lp14913
I1
aI3
aI7
aI5
aI11
aI3
aI1
aI231
aI243
aI541
aI341
aI887
aI3567
aI14759
aI26763
aI35705
aI29417
asS'd'
I7457
sa(dp14914
S'a'
I14786
sS's'
I17
sS'm'
(lp14915
I1
aI1
aI7
aI13
aI17
aI35
aI117
aI177
aI81
aI361
aI1425
aI2437
aI6821
aI1061
aI15019
aI19135
aI106007
asS'd'
I7458
sa(dp14916
S'a'
I14795
sS's'
I17
sS'm'
(lp14917
I1
aI3
aI3
aI11
aI19
aI5
aI39
aI23
aI367
aI9
aI879
aI3583
aI2527
aI14375
aI28359
aI27393
aI55041
asS'd'
I7459
sa(dp14918
S'a'
I14800
sS's'
I17
sS'm'
(lp14919
I1
aI1
aI7
aI11
aI9
aI41
aI63
aI125
aI33
aI337
aI587
aI3939
aI2635
aI4559
aI1007
aI38991
aI35651
asS'd'
I7460
sa(dp14920
S'a'
I14812
sS's'
I17
sS'm'
(lp14921
I1
aI1
aI3
aI1
aI19
aI11
aI83
aI13
aI227
aI649
aI415
aI1661
aI3285
aI55
aI3683
aI22319
aI2127
asS'd'
I7461
sa(dp14922
S'a'
I14816
sS's'
I17
sS'm'
(lp14923
I1
aI3
aI7
aI13
aI19
aI49
aI113
aI129
aI83
aI5
aI19
aI1095
aI6561
aI11049
aI3805
aI11355
aI84265
asS'd'
I7462
sa(dp14924
S'a'
I14836
sS's'
I17
sS'm'
(lp14925
I1
aI3
aI1
aI9
aI19
aI41
aI111
aI193
aI429
aI319
aI67
aI1717
aI1819
aI12959
aI31449
aI21035
aI113161
asS'd'
I7463
sa(dp14926
S'a'
I14840
sS's'
I17
sS'm'
(lp14927
I1
aI1
aI5
aI11
aI19
aI19
aI115
aI237
aI145
aI681
aI1525
aI2215
aI7915
aI15529
aI7533
aI45981
aI85461
asS'd'
I7464
sa(dp14928
S'a'
I14856
sS's'
I17
sS'm'
(lp14929
I1
aI1
aI1
aI1
aI25
aI3
aI73
aI207
aI15
aI69
aI43
aI1643
aI7707
aI12505
aI27101
aI40735
aI6091
asS'd'
I7465
sa(dp14930
S'a'
I14859
sS's'
I17
sS'm'
(lp14931
I1
aI1
aI5
aI11
aI21
aI61
aI119
aI7
aI37
aI147
aI1379
aI3165
aI6555
aI3867
aI24027
aI45161
aI93015
asS'd'
I7466
sa(dp14932
S'a'
I14870
sS's'
I17
sS'm'
(lp14933
I1
aI1
aI3
aI9
aI9
aI25
aI51
aI125
aI511
aI209
aI75
aI2849
aI2299
aI2901
aI25157
aI13079
aI67733
asS'd'
I7467
sa(dp14934
S'a'
I14873
sS's'
I17
sS'm'
(lp14935
I1
aI3
aI7
aI9
aI31
aI49
aI99
aI21
aI89
aI1
aI1391
aI1741
aI2733
aI7283
aI12087
aI9287
aI39713
asS'd'
I7468
sa(dp14936
S'a'
I14879
sS's'
I17
sS'm'
(lp14937
I1
aI1
aI5
aI5
aI1
aI5
aI89
aI109
aI499
aI343
aI431
aI401
aI2023
aI5541
aI16615
aI40059
aI119195
asS'd'
I7469
sa(dp14938
S'a'
I14880
sS's'
I17
sS'm'
(lp14939
I1
aI3
aI5
aI15
aI27
aI27
aI9
aI159
aI395
aI31
aI865
aI2793
aI55
aI10961
aI23123
aI63731
aI54385
asS'd'
I7470
sa(dp14940
S'a'
I14889
sS's'
I17
sS'm'
(lp14941
I1
aI3
aI7
aI1
aI11
aI47
aI123
aI239
aI399
aI383
aI1497
aI4075
aI4659
aI2911
aI2101
aI8295
aI115717
asS'd'
I7471
sa(dp14942
S'a'
I14892
sS's'
I17
sS'm'
(lp14943
I1
aI1
aI3
aI1
aI11
aI63
aI125
aI171
aI65
aI15
aI349
aI753
aI2981
aI6713
aI6219
aI14093
aI78797
asS'd'
I7472
sa(dp14944
S'a'
I14895
sS's'
I17
sS'm'
(lp14945
I1
aI1
aI1
aI13
aI9
aI15
aI1
aI113
aI221
aI867
aI1907
aI103
aI1411
aI27
aI22743
aI377
aI116907
asS'd'
I7473
sa(dp14946
S'a'
I14900
sS's'
I17
sS'm'
(lp14947
I1
aI3
aI1
aI5
aI27
aI5
aI27
aI245
aI221
aI575
aI2009
aI1561
aI4263
aI11843
aI28331
aI12865
aI10483
asS'd'
I7474
sa(dp14948
S'a'
I14903
sS's'
I17
sS'm'
(lp14949
I1
aI3
aI7
aI9
aI1
aI51
aI119
aI241
aI439
aI913
aI1191
aI2343
aI2055
aI10247
aI18283
aI40175
aI63321
asS'd'
I7475
sa(dp14950
S'a'
I14910
sS's'
I17
sS'm'
(lp14951
I1
aI1
aI3
aI15
aI21
aI59
aI45
aI151
aI485
aI293
aI981
aI3523
aI7689
aI2789
aI5003
aI62383
aI126221
asS'd'
I7476
sa(dp14952
S'a'
I14912
sS's'
I17
sS'm'
(lp14953
I1
aI1
aI1
aI1
aI13
aI15
aI39
aI201
aI405
aI513
aI1721
aI2077
aI5995
aI2433
aI20421
aI12695
aI20393
asS'd'
I7477
sa(dp14954
S'a'
I14942
sS's'
I17
sS'm'
(lp14955
I1
aI3
aI5
aI15
aI11
aI35
aI113
aI133
aI187
aI583
aI577
aI291
aI7563
aI12959
aI9383
aI44255
aI81763
asS'd'
I7478
sa(dp14956
S'a'
I14948
sS's'
I17
sS'm'
(lp14957
I1
aI3
aI7
aI15
aI9
aI55
aI57
aI227
aI189
aI595
aI1311
aI1131
aI1323
aI11347
aI12777
aI50963
aI13827
asS'd'
I7479
sa(dp14958
S'a'
I14957
sS's'
I17
sS'm'
(lp14959
I1
aI3
aI5
aI3
aI11
aI49
aI77
aI157
aI107
aI959
aI761
aI1457
aI7121
aI3027
aI9269
aI26291
aI125261
asS'd'
I7480
sa(dp14960
S'a'
I14963
sS's'
I17
sS'm'
(lp14961
I1
aI1
aI5
aI9
aI23
aI53
aI125
aI211
aI303
aI433
aI1103
aI41
aI2643
aI5325
aI11885
aI23825
aI80415
asS'd'
I7481
sa(dp14962
S'a'
I14975
sS's'
I17
sS'm'
(lp14963
I1
aI1
aI7
aI1
aI29
aI25
aI51
aI107
aI209
aI165
aI707
aI1855
aI7429
aI1583
aI5941
aI47509
aI90105
asS'd'
I7482
sa(dp14964
S'a'
I14985
sS's'
I17
sS'm'
(lp14965
I1
aI1
aI3
aI3
aI1
aI15
aI121
aI165
aI181
aI259
aI1949
aI3049
aI3545
aI3093
aI5967
aI49207
aI37129
asS'd'
I7483
sa(dp14966
S'a'
I14993
sS's'
I17
sS'm'
(lp14967
I1
aI1
aI5
aI13
aI9
aI59
aI93
aI87
aI57
aI343
aI389
aI1995
aI4001
aI11495
aI12909
aI13491
aI61759
asS'd'
I7484
sa(dp14968
S'a'
I15003
sS's'
I17
sS'm'
(lp14969
I1
aI3
aI1
aI5
aI11
aI27
aI27
aI133
aI459
aI733
aI1845
aI1795
aI4613
aI3397
aI12313
aI52839
aI129583
asS'd'
I7485
sa(dp14970
S'a'
I15010
sS's'
I17
sS'm'
(lp14971
I1
aI3
aI5
aI3
aI19
aI1
aI7
aI145
aI255
aI337
aI1649
aI1473
aI4113
aI4425
aI12233
aI55477
aI69157
asS'd'
I7486
sa(dp14972
S'a'
I15022
sS's'
I17
sS'm'
(lp14973
I1
aI1
aI3
aI1
aI25
aI27
aI93
aI59
aI415
aI437
aI25
aI1565
aI319
aI8981
aI2453
aI53579
aI45033
asS'd'
I7487
sa(dp14974
S'a'
I15039
sS's'
I17
sS'm'
(lp14975
I1
aI3
aI7
aI1
aI27
aI49
aI47
aI233
aI341
aI101
aI2017
aI2827
aI8085
aI237
aI6363
aI61139
aI88903
asS'd'
I7488
sa(dp14976
S'a'
I15041
sS's'
I17
sS'm'
(lp14977
I1
aI1
aI1
aI5
aI23
aI47
aI65
aI251
aI423
aI957
aI1751
aI3541
aI5405
aI1335
aI22703
aI12587
aI60201
asS'd'
I7489
sa(dp14978
S'a'
I15047
sS's'
I17
sS'm'
(lp14979
I1
aI1
aI3
aI3
aI5
aI51
aI85
aI195
aI423
aI519
aI1797
aI3821
aI5915
aI12257
aI5377
aI62733
aI41197
asS'd'
I7490
sa(dp14980
S'a'
I15048
sS's'
I17
sS'm'
(lp14981
I1
aI3
aI7
aI15
aI3
aI47
aI97
aI1
aI5
aI175
aI1449
aI1609
aI6873
aI12017
aI5579
aI2665
aI58389
asS'd'
I7491
sa(dp14982
S'a'
I15056
sS's'
I17
sS'm'
(lp14983
I1
aI3
aI3
aI15
aI19
aI37
aI35
aI29
aI79
aI767
aI21
aI1279
aI1997
aI11611
aI14381
aI35607
aI127701
asS'd'
I7492
sa(dp14984
S'a'
I15066
sS's'
I17
sS'm'
(lp14985
I1
aI3
aI7
aI7
aI7
aI43
aI47
aI33
aI69
aI155
aI703
aI1373
aI1589
aI6997
aI8627
aI50647
aI16989
asS'd'
I7493
sa(dp14986
S'a'
I15068
sS's'
I17
sS'm'
(lp14987
I1
aI3
aI1
aI5
aI13
aI33
aI69
aI133
aI399
aI361
aI1633
aI321
aI2077
aI8857
aI13419
aI23227
aI40003
asS'd'
I7494
sa(dp14988
S'a'
I15075
sS's'
I17
sS'm'
(lp14989
I1
aI1
aI1
aI15
aI15
aI9
aI45
aI181
aI427
aI1005
aI341
aI1697
aI6423
aI5727
aI7163
aI10401
aI38957
asS'd'
I7495
sa(dp14990
S'a'
I15077
sS's'
I17
sS'm'
(lp14991
I1
aI1
aI7
aI1
aI17
aI5
aI17
aI95
aI279
aI171
aI825
aI2459
aI5243
aI10683
aI1849
aI32809
aI8995
asS'd'
I7496
sa(dp14992
S'a'
I15082
sS's'
I17
sS'm'
(lp14993
I1
aI3
aI5
aI15
aI27
aI47
aI103
aI69
aI69
aI255
aI961
aI2173
aI5297
aI5987
aI5863
aI14311
aI117569
asS'd'
I7497
sa(dp14994
S'a'
I15096
sS's'
I17
sS'm'
(lp14995
I1
aI1
aI1
aI11
aI21
aI27
aI61
aI239
aI183
aI1013
aI1955
aI3171
aI4183
aI965
aI14885
aI49605
aI87851
asS'd'
I7498
sa(dp14996
S'a'
I15102
sS's'
I17
sS'm'
(lp14997
I1
aI1
aI7
aI9
aI9
aI53
aI99
aI211
aI267
aI803
aI1545
aI4011
aI7613
aI13889
aI28277
aI6817
aI26515
asS'd'
I7499
sa(dp14998
S'a'
I15116
sS's'
I17
sS'm'
(lp14999
I1
aI1
aI3
aI9
aI1
aI19
aI33
aI227
aI461
aI679
aI499
aI1069
aI837
aI12129
aI20779
aI12937
aI104367
asS'd'
I7500
sa(dp15000
S'a'
I15122
sS's'
I17
sS'm'
(lp15001
I1
aI3
aI3
aI15
aI7
aI3
aI29
aI245
aI179
aI1015
aI1651
aI3753
aI4185
aI15357
aI17379
aI52835
aI51953
asS'd'
I7501
sa(dp15002
S'a'
I15127
sS's'
I17
sS'm'
(lp15003
I1
aI3
aI3
aI3
aI3
aI25
aI95
aI239
aI263
aI427
aI1749
aI183
aI5251
aI361
aI32549
aI24331
aI30789
asS'd'
I7502
sa(dp15004
S'a'
I15133
sS's'
I17
sS'm'
(lp15005
I1
aI1
aI7
aI1
aI5
aI3
aI79
aI9
aI403
aI195
aI1433
aI385
aI8105
aI7893
aI16415
aI23253
aI127837
asS'd'
I7503
sa(dp15006
S'a'
I15137
sS's'
I17
sS'm'
(lp15007
I1
aI3
aI7
aI3
aI23
aI45
aI115
aI27
aI473
aI241
aI361
aI1787
aI4247
aI13451
aI5627
aI32923
aI29375
asS'd'
I7504
sa(dp15008
S'a'
I15138
sS's'
I17
sS'm'
(lp15009
I1
aI3
aI7
aI1
aI5
aI55
aI43
aI37
aI481
aI899
aI51
aI2459
aI5005
aI12365
aI19261
aI32797
aI45843
asS'd'
I7505
sa(dp15010
S'a'
I15149
sS's'
I17
sS'm'
(lp15011
I1
aI3
aI7
aI5
aI9
aI41
aI83
aI163
aI241
aI899
aI567
aI231
aI4897
aI15175
aI10329
aI6625
aI95927
asS'd'
I7506
sa(dp15012
S'a'
I15152
sS's'
I17
sS'm'
(lp15013
I1
aI3
aI1
aI1
aI7
aI51
aI61
aI55
aI253
aI315
aI1893
aI2635
aI4061
aI257
aI14147
aI36639
aI24893
asS'd'
I7507
sa(dp15014
S'a'
I15155
sS's'
I17
sS'm'
(lp15015
I1
aI1
aI5
aI1
aI13
aI63
aI115
aI119
aI205
aI309
aI277
aI2191
aI341
aI4715
aI13111
aI58043
aI51241
asS'd'
I7508
sa(dp15016
S'a'
I15158
sS's'
I17
sS'm'
(lp15017
I1
aI3
aI1
aI15
aI17
aI23
aI89
aI121
aI205
aI15
aI295
aI667
aI421
aI14071
aI27719
aI1335
aI9887
asS'd'
I7509
sa(dp15018
S'a'
I15187
sS's'
I17
sS'm'
(lp15019
I1
aI3
aI5
aI5
aI17
aI49
aI5
aI93
aI251
aI613
aI1029
aI945
aI1547
aI10479
aI20183
aI26787
aI120441
asS'd'
I7510
sa(dp15020
S'a'
I15189
sS's'
I17
sS'm'
(lp15021
I1
aI3
aI3
aI15
aI17
aI11
aI63
aI97
aI499
aI313
aI881
aI2233
aI4287
aI5141
aI13841
aI40725
aI49285
asS'd'
I7511
sa(dp15022
S'a'
I15190
sS's'
I17
sS'm'
(lp15023
I1
aI3
aI3
aI11
aI19
aI33
aI105
aI203
aI325
aI337
aI353
aI1923
aI7157
aI8623
aI23881
aI4513
aI71495
asS'd'
I7512
sa(dp15024
S'a'
I15196
sS's'
I17
sS'm'
(lp15025
I1
aI1
aI5
aI1
aI3
aI15
aI119
aI43
aI85
aI869
aI1597
aI2433
aI845
aI5065
aI12813
aI64849
aI58491
asS'd'
I7513
sa(dp15026
S'a'
I15199
sS's'
I17
sS'm'
(lp15027
I1
aI3
aI7
aI7
aI25
aI63
aI119
aI93
aI303
aI665
aI571
aI1795
aI5853
aI13527
aI12715
aI36483
aI57723
asS'd'
I7514
sa(dp15028
S'a'
I15205
sS's'
I17
sS'm'
(lp15029
I1
aI3
aI7
aI13
aI19
aI43
aI55
aI85
aI189
aI627
aI1457
aI3185
aI3491
aI1913
aI13399
aI30681
aI69015
asS'd'
I7515
sa(dp15030
S'a'
I15212
sS's'
I17
sS'm'
(lp15031
I1
aI3
aI5
aI9
aI5
aI41
aI51
aI65
aI147
aI425
aI569
aI1317
aI1557
aI7631
aI17243
aI37847
aI51161
asS'd'
I7516
sa(dp15032
S'a'
I15236
sS's'
I17
sS'm'
(lp15033
I1
aI1
aI3
aI7
aI29
aI39
aI61
aI127
aI489
aI89
aI749
aI2073
aI195
aI14367
aI13533
aI27403
aI16365
asS'd'
I7517
sa(dp15034
S'a'
I15243
sS's'
I17
sS'm'
(lp15035
I1
aI3
aI7
aI15
aI13
aI35
aI45
aI157
aI373
aI415
aI725
aI779
aI3559
aI7489
aI11369
aI36501
aI60761
asS'd'
I7518
sa(dp15036
S'a'
I15246
sS's'
I17
sS'm'
(lp15037
I1
aI3
aI1
aI3
aI13
aI45
aI25
aI215
aI385
aI709
aI499
aI3861
aI761
aI15597
aI3335
aI37013
aI13173
asS'd'
I7519
sa(dp15038
S'a'
I15260
sS's'
I17
sS'm'
(lp15039
I1
aI1
aI7
aI1
aI13
aI49
aI89
aI135
aI175
aI1015
aI67
aI957
aI4893
aI9843
aI13027
aI14709
aI59721
asS'd'
I7520
sa(dp15040
S'a'
I15267
sS's'
I17
sS'm'
(lp15041
I1
aI3
aI3
aI11
aI19
aI37
aI109
aI143
aI135
aI535
aI1543
aI3991
aI189
aI6739
aI28087
aI18845
aI41819
asS'd'
I7521
sa(dp15042
S'a'
I15274
sS's'
I17
sS'm'
(lp15043
I1
aI1
aI7
aI5
aI1
aI7
aI11
aI5
aI211
aI251
aI1593
aI2527
aI3539
aI10471
aI25595
aI60119
aI89213
asS'd'
I7522
sa(dp15044
S'a'
I15279
sS's'
I17
sS'm'
(lp15045
I1
aI1
aI5
aI7
aI13
aI51
aI121
aI167
aI299
aI403
aI977
aI521
aI279
aI15521
aI15901
aI935
aI14065
asS'd'
I7523
sa(dp15046
S'a'
I15281
sS's'
I17
sS'm'
(lp15047
I1
aI1
aI7
aI13
aI7
aI21
aI27
aI205
aI377
aI801
aI1365
aI1567
aI6651
aI139
aI14229
aI30827
aI50429
asS'd'
I7524
sa(dp15048
S'a'
I15282
sS's'
I17
sS'm'
(lp15049
I1
aI1
aI1
aI1
aI17
aI11
aI75
aI87
aI217
aI413
aI1923
aI1765
aI2037
aI14061
aI12433
aI30671
aI24883
asS'd'
I7525
sa(dp15050
S'a'
I15284
sS's'
I17
sS'm'
(lp15051
I1
aI1
aI5
aI13
aI17
aI51
aI91
aI241
aI95
aI505
aI349
aI2689
aI1117
aI4435
aI1713
aI44501
aI125619
asS'd'
I7526
sa(dp15052
S'a'
I15291
sS's'
I17
sS'm'
(lp15053
I1
aI1
aI3
aI15
aI11
aI21
aI25
aI59
aI511
aI353
aI799
aI91
aI4517
aI16005
aI17061
aI21841
aI46311
asS'd'
I7527
sa(dp15054
S'a'
I15293
sS's'
I17
sS'm'
(lp15055
I1
aI1
aI1
aI5
aI19
aI53
aI109
aI177
aI213
aI373
aI761
aI453
aI5753
aI69
aI3503
aI49411
aI111105
asS'd'
I7528
sa(dp15056
S'a'
I15302
sS's'
I17
sS'm'
(lp15057
I1
aI3
aI1
aI5
aI21
aI27
aI103
aI167
aI109
aI55
aI1849
aI3999
aI7801
aI4185
aI9789
aI7515
aI124983
asS'd'
I7529
sa(dp15058
S'a'
I15319
sS's'
I17
sS'm'
(lp15059
I1
aI3
aI7
aI7
aI25
aI9
aI65
aI127
aI141
aI169
aI1079
aI3377
aI691
aI5119
aI6629
aI3517
aI28963
asS'd'
I7530
sa(dp15060
S'a'
I15329
sS's'
I17
sS'm'
(lp15061
I1
aI1
aI3
aI11
aI15
aI61
aI127
aI35
aI87
aI891
aI1459
aI483
aI6763
aI16173
aI5633
aI6939
aI63411
asS'd'
I7531
sa(dp15062
S'a'
I15336
sS's'
I17
sS'm'
(lp15063
I1
aI3
aI5
aI3
aI7
aI63
aI111
aI85
aI415
aI273
aI1705
aI4045
aI5551
aI2377
aI29025
aI16831
aI90203
asS'd'
I7532
sa(dp15064
S'a'
I15347
sS's'
I17
sS'm'
(lp15065
I1
aI3
aI5
aI13
aI7
aI23
aI103
aI227
aI477
aI985
aI1059
aI1489
aI7233
aI1917
aI10409
aI38759
aI86761
asS'd'
I7533
sa(dp15066
S'a'
I15353
sS's'
I17
sS'm'
(lp15067
I1
aI3
aI5
aI7
aI31
aI33
aI75
aI41
aI355
aI577
aI225
aI5
aI897
aI15653
aI27415
aI83
aI14911
asS'd'
I7534
sa(dp15068
S'a'
I15361
sS's'
I17
sS'm'
(lp15069
I1
aI3
aI5
aI5
aI23
aI13
aI5
aI43
aI165
aI53
aI149
aI2005
aI4545
aI477
aI17885
aI21343
aI35751
asS'd'
I7535
sa(dp15070
S'a'
I15371
sS's'
I17
sS'm'
(lp15071
I1
aI3
aI3
aI3
aI25
aI51
aI33
aI203
aI291
aI835
aI241
aI3255
aI3709
aI3573
aI9859
aI33027
aI122801
asS'd'
I7536
sa(dp15072
S'a'
I15397
sS's'
I17
sS'm'
(lp15073
I1
aI3
aI5
aI7
aI13
aI7
aI3
aI141
aI455
aI67
aI2003
aI3411
aI4717
aI157
aI29491
aI14429
aI44849
asS'd'
I7537
sa(dp15074
S'a'
I15404
sS's'
I17
sS'm'
(lp15075
I1
aI1
aI1
aI11
aI3
aI33
aI101
aI93
aI219
aI371
aI1191
aI1521
aI1663
aI8485
aI24815
aI38283
aI120867
asS'd'
I7538
sa(dp15076
S'a'
I15407
sS's'
I17
sS'm'
(lp15077
I1
aI1
aI3
aI9
aI25
aI61
aI71
aI173
aI69
aI181
aI1525
aI2129
aI2979
aI19
aI13489
aI627
aI72619
asS'd'
I7539
sa(dp15078
S'a'
I15421
sS's'
I17
sS'm'
(lp15079
I1
aI3
aI1
aI7
aI25
aI33
aI39
aI247
aI221
aI7
aI683
aI1837
aI8037
aI9125
aI4259
aI63049
aI63021
asS'd'
I7540
sa(dp15080
S'a'
I15433
sS's'
I17
sS'm'
(lp15081
I1
aI3
aI3
aI9
aI17
aI15
aI9
aI189
aI357
aI707
aI521
aI711
aI8189
aI12945
aI29675
aI11851
aI126813
asS'd'
I7541
sa(dp15082
S'a'
I15441
sS's'
I17
sS'm'
(lp15083
I1
aI3
aI1
aI1
aI23
aI3
aI57
aI133
aI245
aI301
aI957
aI239
aI3139
aI7949
aI27133
aI18229
aI93015
asS'd'
I7542
sa(dp15084
S'a'
I15442
sS's'
I17
sS'm'
(lp15085
I1
aI1
aI3
aI1
aI29
aI23
aI35
aI87
aI231
aI257
aI1997
aI271
aI3019
aI3409
aI10613
aI42245
aI111309
asS'd'
I7543
sa(dp15086
S'a'
I15463
sS's'
I17
sS'm'
(lp15087
I1
aI1
aI3
aI3
aI1
aI21
aI17
aI37
aI393
aI943
aI791
aI3101
aI6715
aI11907
aI25369
aI9061
aI75381
asS'd'
I7544
sa(dp15088
S'a'
I15472
sS's'
I17
sS'm'
(lp15089
I1
aI1
aI7
aI7
aI17
aI31
aI25
aI7
aI183
aI819
aI1265
aI3343
aI6845
aI2039
aI3779
aI41861
aI38309
asS'd'
I7545
sa(dp15090
S'a'
I15478
sS's'
I17
sS'm'
(lp15091
I1
aI1
aI1
aI5
aI17
aI25
aI1
aI41
aI173
aI995
aI863
aI3515
aI1779
aI2159
aI28223
aI64661
aI40697
asS'd'
I7546
sa(dp15092
S'a'
I15488
sS's'
I17
sS'm'
(lp15093
I1
aI1
aI3
aI15
aI29
aI49
aI81
aI241
aI511
aI817
aI1301
aI3593
aI6759
aI7483
aI8859
aI30339
aI106137
asS'd'
I7547
sa(dp15094
S'a'
I15493
sS's'
I17
sS'm'
(lp15095
I1
aI3
aI1
aI11
aI17
aI61
aI95
aI231
aI3
aI693
aI37
aI1091
aI3111
aI11941
aI17475
aI8073
aI62373
asS'd'
I7548
sa(dp15096
S'a'
I15498
sS's'
I17
sS'm'
(lp15097
I1
aI1
aI1
aI3
aI7
aI25
aI93
aI7
aI291
aI957
aI859
aI2519
aI241
aI10963
aI10403
aI933
aI50599
asS'd'
I7549
sa(dp15098
S'a'
I15511
sS's'
I17
sS'm'
(lp15099
I1
aI1
aI7
aI1
aI7
aI33
aI121
aI91
aI369
aI333
aI229
aI4073
aI6063
aI6491
aI31711
aI65061
aI107843
asS'd'
I7550
sa(dp15100
S'a'
I15521
sS's'
I17
sS'm'
(lp15101
I1
aI1
aI5
aI15
aI17
aI1
aI117
aI195
aI445
aI547
aI867
aI2893
aI4835
aI6513
aI29091
aI60367
aI33409
asS'd'
I7551
sa(dp15102
S'a'
I15534
sS's'
I17
sS'm'
(lp15103
I1
aI3
aI3
aI7
aI15
aI5
aI125
aI131
aI165
aI127
aI207
aI853
aI5927
aI3605
aI17083
aI44481
aI111333
asS'd'
I7552
sa(dp15104
S'a'
I15539
sS's'
I17
sS'm'
(lp15105
I1
aI1
aI1
aI9
aI3
aI43
aI75
aI191
aI319
aI889
aI1513
aI3301
aI1535
aI4693
aI10367
aI12491
aI43175
asS'd'
I7553
sa(dp15106
S'a'
I15541
sS's'
I17
sS'm'
(lp15107
I1
aI3
aI5
aI5
aI29
aI19
aI75
aI221
aI393
aI977
aI1373
aI1571
aI7377
aI1763
aI18073
aI11381
aI101241
asS'd'
I7554
sa(dp15108
S'a'
I15560
sS's'
I17
sS'm'
(lp15109
I1
aI3
aI1
aI15
aI3
aI15
aI73
aI91
aI165
aI213
aI1077
aI1267
aI2411
aI15807
aI3979
aI12731
aI86597
asS'd'
I7555
sa(dp15110
S'a'
I15563
sS's'
I17
sS'm'
(lp15111
I1
aI1
aI3
aI7
aI3
aI21
aI5
aI135
aI95
aI337
aI1853
aI1675
aI2449
aI12535
aI18505
aI60127
aI76949
asS'd'
I7556
sa(dp15112
S'a'
I15574
sS's'
I17
sS'm'
(lp15113
I1
aI1
aI7
aI3
aI15
aI11
aI63
aI127
aI329
aI169
aI1569
aI675
aI4801
aI5859
aI3243
aI25811
aI77841
asS'd'
I7557
sa(dp15114
S'a'
I15578
sS's'
I17
sS'm'
(lp15115
I1
aI1
aI1
aI9
aI19
aI13
aI73
aI119
aI105
aI537
aI951
aI1033
aI5303
aI5775
aI815
aI19277
aI57607
asS'd'
I7558
sa(dp15116
S'a'
I15584
sS's'
I17
sS'm'
(lp15117
I1
aI3
aI5
aI5
aI23
aI21
aI91
aI231
aI117
aI1007
aI1603
aI841
aI2595
aI11223
aI17171
aI25963
aI17049
asS'd'
I7559
sa(dp15118
S'a'
I15593
sS's'
I17
sS'm'
(lp15119
I1
aI1
aI5
aI11
aI15
aI43
aI7
aI229
aI55
aI129
aI599
aI993
aI563
aI15677
aI16703
aI36253
aI17847
asS'd'
I7560
sa(dp15120
S'a'
I15604
sS's'
I17
sS'm'
(lp15121
I1
aI3
aI1
aI9
aI25
aI3
aI109
aI21
aI87
aI721
aI1927
aI3219
aI3395
aI3267
aI9117
aI13591
aI89267
asS'd'
I7561
sa(dp15122
S'a'
I15614
sS's'
I17
sS'm'
(lp15123
I1
aI1
aI1
aI15
aI11
aI17
aI47
aI49
aI125
aI925
aI333
aI945
aI2411
aI10907
aI12021
aI47857
aI84303
asS'd'
I7562
sa(dp15124
S'a'
I15619
sS's'
I17
sS'm'
(lp15125
I1
aI1
aI1
aI1
aI23
aI11
aI99
aI215
aI105
aI417
aI823
aI1289
aI421
aI12285
aI17711
aI35389
aI1935
asS'd'
I7563
sa(dp15126
S'a'
I15622
sS's'
I17
sS'm'
(lp15127
I1
aI1
aI1
aI15
aI27
aI7
aI23
aI141
aI7
aI929
aI147
aI681
aI5473
aI4173
aI28645
aI42053
aI83573
asS'd'
I7564
sa(dp15128
S'a'
I15633
sS's'
I17
sS'm'
(lp15129
I1
aI1
aI1
aI11
aI5
aI61
aI71
aI65
aI287
aI697
aI1183
aI3257
aI7251
aI14011
aI21349
aI42445
aI4701
asS'd'
I7565
sa(dp15130
S'a'
I15639
sS's'
I17
sS'm'
(lp15131
I1
aI3
aI5
aI9
aI5
aI23
aI45
aI217
aI369
aI189
aI1495
aI107
aI425
aI10467
aI4909
aI64293
aI17885
asS'd'
I7566
sa(dp15132
S'a'
I15646
sS's'
I17
sS'm'
(lp15133
I1
aI1
aI3
aI11
aI21
aI45
aI75
aI65
aI57
aI893
aI783
aI3429
aI409
aI13617
aI483
aI62489
aI2919
asS'd'
I7567
sa(dp15134
S'a'
I15673
sS's'
I17
sS'm'
(lp15135
I1
aI1
aI7
aI3
aI5
aI61
aI51
aI255
aI501
aI839
aI367
aI1165
aI7055
aI8139
aI23891
aI18807
aI20739
asS'd'
I7568
sa(dp15136
S'a'
I15674
sS's'
I17
sS'm'
(lp15137
I1
aI1
aI3
aI7
aI23
aI15
aI97
aI139
aI323
aI463
aI921
aI1529
aI6655
aI8697
aI23577
aI56761
aI62023
asS'd'
I7569
sa(dp15138
S'a'
I15684
sS's'
I17
sS'm'
(lp15139
I1
aI1
aI5
aI11
aI13
aI11
aI57
aI225
aI277
aI713
aI1427
aI95
aI1135
aI7721
aI30731
aI32625
aI107891
asS'd'
I7570
sa(dp15140
S'a'
I15691
sS's'
I17
sS'm'
(lp15141
I1
aI1
aI5
aI7
aI23
aI35
aI39
aI91
aI291
aI609
aI919
aI3325
aI6843
aI7659
aI5603
aI37471
aI41495
asS'd'
I7571
sa(dp15142
S'a'
I15694
sS's'
I17
sS'm'
(lp15143
I1
aI1
aI1
aI1
aI25
aI11
aI117
aI15
aI389
aI589
aI1345
aI423
aI6531
aI9903
aI20243
aI9523
aI22991
asS'd'
I7572
sa(dp15144
S'a'
I15696
sS's'
I17
sS'm'
(lp15145
I1
aI3
aI5
aI15
aI29
aI7
aI57
aI113
aI387
aI883
aI1141
aI3295
aI2973
aI4129
aI16973
aI33429
aI109997
asS'd'
I7573
sa(dp15146
S'a'
I15701
sS's'
I17
sS'm'
(lp15147
I1
aI3
aI5
aI3
aI25
aI1
aI73
aI207
aI353
aI203
aI1479
aI985
aI6373
aI3079
aI28403
aI63675
aI21787
asS'd'
I7574
sa(dp15148
S'a'
I15705
sS's'
I17
sS'm'
(lp15149
I1
aI3
aI1
aI5
aI31
aI39
aI107
aI197
aI359
aI45
aI203
aI559
aI4721
aI6579
aI11305
aI12957
aI10061
asS'd'
I7575
sa(dp15150
S'a'
I15715
sS's'
I17
sS'm'
(lp15151
I1
aI3
aI7
aI15
aI9
aI3
aI55
aI153
aI373
aI981
aI575
aI827
aI4757
aI15743
aI14295
aI43875
aI17847
asS'd'
I7576
sa(dp15152
S'a'
I15729
sS's'
I17
sS'm'
(lp15153
I1
aI3
aI5
aI1
aI17
aI1
aI93
aI87
aI207
aI997
aI1695
aI3643
aI6973
aI9507
aI29309
aI58531
aI6849
asS'd'
I7577
sa(dp15154
S'a'
I15730
sS's'
I17
sS'm'
(lp15155
I1
aI3
aI1
aI11
aI3
aI39
aI17
aI241
aI83
aI931
aI39
aI3839
aI6437
aI5159
aI28869
aI61859
aI96873
asS'd'
I7578
sa(dp15156
S'a'
I15741
sS's'
I17
sS'm'
(lp15157
I1
aI1
aI5
aI13
aI29
aI43
aI71
aI159
aI261
aI563
aI695
aI1205
aI2273
aI8077
aI12569
aI17187
aI54369
asS'd'
I7579
sa(dp15158
S'a'
I15745
sS's'
I17
sS'm'
(lp15159
I1
aI3
aI7
aI5
aI11
aI57
aI17
aI31
aI311
aI1001
aI1419
aI3899
aI6679
aI15531
aI28877
aI28221
aI105413
asS'd'
I7580
sa(dp15160
S'a'
I15748
sS's'
I17
sS'm'
(lp15161
I1
aI3
aI3
aI13
aI23
aI29
aI127
aI19
aI345
aI1003
aI1571
aI2219
aI3199
aI9903
aI18701
aI31865
aI108879
asS'd'
I7581
sa(dp15162
S'a'
I15757
sS's'
I17
sS'm'
(lp15163
I1
aI3
aI7
aI13
aI23
aI51
aI95
aI43
aI35
aI439
aI25
aI323
aI2365
aI12407
aI27525
aI57795
aI74495
asS'd'
I7582
sa(dp15164
S'a'
I15765
sS's'
I17
sS'm'
(lp15165
I1
aI1
aI1
aI1
aI17
aI43
aI57
aI185
aI439
aI929
aI69
aI813
aI6205
aI3139
aI3853
aI56967
aI19073
asS'd'
I7583
sa(dp15166
S'a'
I15766
sS's'
I17
sS'm'
(lp15167
I1
aI3
aI7
aI1
aI27
aI5
aI43
aI211
aI395
aI113
aI1675
aI1505
aI6171
aI5169
aI9991
aI21641
aI27101
asS'd'
I7584
sa(dp15168
S'a'
I15775
sS's'
I17
sS'm'
(lp15169
I1
aI3
aI3
aI1
aI17
aI41
aI59
aI131
aI131
aI339
aI955
aI1145
aI5301
aI4585
aI20441
aI43227
aI23123
asS'd'
I7585
sa(dp15170
S'a'
I15776
sS's'
I17
sS'm'
(lp15171
I1
aI1
aI7
aI9
aI9
aI55
aI61
aI31
aI71
aI229
aI963
aI3247
aI4677
aI9595
aI21715
aI36391
aI86997
asS'd'
I7586
sa(dp15172
S'a'
I15779
sS's'
I17
sS'm'
(lp15173
I1
aI1
aI7
aI5
aI9
aI17
aI55
aI179
aI27
aI229
aI79
aI1335
aI5887
aI1003
aI22085
aI34377
aI51367
asS'd'
I7587
sa(dp15174
S'a'
I15786
sS's'
I17
sS'm'
(lp15175
I1
aI1
aI1
aI5
aI11
aI45
aI15
aI219
aI411
aI27
aI1003
aI1553
aI303
aI13571
aI13985
aI6801
aI52407
asS'd'
I7588
sa(dp15176
S'a'
I15788
sS's'
I17
sS'm'
(lp15177
I1
aI3
aI3
aI7
aI7
aI55
aI111
aI255
aI453
aI409
aI1863
aI1449
aI4103
aI8725
aI26923
aI5017
aI43657
asS'd'
I7589
sa(dp15178
S'a'
I15813
sS's'
I17
sS'm'
(lp15179
I1
aI1
aI1
aI15
aI23
aI3
aI95
aI57
aI29
aI727
aI1111
aI3309
aI1089
aI471
aI16099
aI11517
aI51563
asS'd'
I7590
sa(dp15180
S'a'
I15814
sS's'
I17
sS'm'
(lp15181
I1
aI3
aI1
aI15
aI17
aI57
aI83
aI163
aI251
aI987
aI1159
aI2079
aI3463
aI13109
aI7443
aI8665
aI123397
asS'd'
I7591
sa(dp15182
S'a'
I15842
sS's'
I17
sS'm'
(lp15183
I1
aI1
aI7
aI1
aI27
aI13
aI35
aI209
aI471
aI843
aI1029
aI1383
aI5413
aI2085
aI13431
aI26557
aI47033
asS'd'
I7592
sa(dp15184
S'a'
I15851
sS's'
I17
sS'm'
(lp15185
I1
aI3
aI1
aI1
aI21
aI21
aI83
aI135
aI303
aI27
aI1407
aI1751
aI331
aI9207
aI31891
aI59287
aI120687
asS'd'
I7593
sa(dp15186
S'a'
I15862
sS's'
I17
sS'm'
(lp15187
I1
aI1
aI1
aI9
aI11
aI35
aI103
aI157
aI1
aI855
aI175
aI3203
aI4381
aI3113
aI27589
aI4567
aI31897
asS'd'
I7594
sa(dp15188
S'a'
I15875
sS's'
I17
sS'm'
(lp15189
I1
aI1
aI3
aI5
aI21
aI5
aI123
aI161
aI301
aI101
aI909
aI947
aI6893
aI15459
aI29139
aI49377
aI94901
asS'd'
I7595
sa(dp15190
S'a'
I15878
sS's'
I17
sS'm'
(lp15191
I1
aI3
aI7
aI7
aI21
aI27
aI5
aI69
aI427
aI409
aI1389
aI3737
aI847
aI2775
aI603
aI1001
aI87651
asS'd'
I7596
sa(dp15192
S'a'
I15889
sS's'
I17
sS'm'
(lp15193
I1
aI1
aI3
aI5
aI1
aI57
aI109
aI89
aI99
aI593
aI581
aI3527
aI1557
aI4971
aI27523
aI26909
aI35787
asS'd'
I7597
sa(dp15194
S'a'
I15896
sS's'
I17
sS'm'
(lp15195
I1
aI1
aI7
aI3
aI31
aI19
aI83
aI65
aI239
aI919
aI15
aI2289
aI4117
aI9127
aI6033
aI49667
aI89343
asS'd'
I7598
sa(dp15196
S'a'
I15901
sS's'
I17
sS'm'
(lp15197
I1
aI3
aI7
aI7
aI9
aI31
aI87
aI117
aI195
aI681
aI1711
aI1753
aI2221
aI10053
aI1985
aI6273
aI21801
asS'd'
I7599
sa(dp15198
S'a'
I15908
sS's'
I17
sS'm'
(lp15199
I1
aI3
aI1
aI7
aI21
aI61
aI53
aI231
aI309
aI115
aI1729
aI3883
aI6085
aI4825
aI31455
aI50097
aI59779
asS'd'
I7600
sa(dp15200
S'a'
I15911
sS's'
I17
sS'm'
(lp15201
I1
aI1
aI1
aI9
aI29
aI25
aI45
aI91
aI145
aI927
aI147
aI371
aI2603
aI12537
aI17267
aI59895
aI128009
asS'd'
I7601
sa(dp15202
S'a'
I15915
sS's'
I17
sS'm'
(lp15203
I1
aI1
aI1
aI1
aI15
aI41
aI63
aI43
aI167
aI215
aI15
aI3387
aI1811
aI12391
aI25721
aI6961
aI13701
asS'd'
I7602
sa(dp15204
S'a'
I15938
sS's'
I17
sS'm'
(lp15205
I1
aI1
aI7
aI1
aI27
aI63
aI25
aI85
aI337
aI799
aI87
aI2237
aI4085
aI14529
aI11493
aI60149
aI86399
asS'd'
I7603
sa(dp15206
S'a'
I15944
sS's'
I17
sS'm'
(lp15207
I1
aI3
aI1
aI11
aI1
aI41
aI103
aI145
aI279
aI805
aI1201
aI823
aI5411
aI4227
aI25999
aI14373
aI36295
asS'd'
I7604
sa(dp15208
S'a'
I15950
sS's'
I17
sS'm'
(lp15209
I1
aI1
aI7
aI3
aI27
aI51
aI83
aI105
aI155
aI657
aI1879
aI3869
aI2559
aI2939
aI19785
aI47167
aI34503
asS'd'
I7605
sa(dp15210
S'a'
I15955
sS's'
I17
sS'm'
(lp15211
I1
aI3
aI1
aI5
aI3
aI31
aI47
aI241
aI257
aI15
aI983
aI4095
aI3745
aI3901
aI1639
aI5421
aI81585
asS'd'
I7606
sa(dp15212
S'a'
I15974
sS's'
I17
sS'm'
(lp15213
I1
aI3
aI3
aI5
aI31
aI13
aI127
aI125
aI175
aI577
aI1103
aI3573
aI6229
aI13969
aI6267
aI19067
aI3933
asS'd'
I7607
sa(dp15214
S'a'
I15978
sS's'
I17
sS'm'
(lp15215
I1
aI1
aI7
aI1
aI31
aI17
aI15
aI15
aI411
aI553
aI1929
aI3731
aI1955
aI11749
aI21991
aI39189
aI124427
asS'd'
I7608
sa(dp15216
S'a'
I15980
sS's'
I17
sS'm'
(lp15217
I1
aI3
aI5
aI5
aI19
aI63
aI93
aI201
aI491
aI599
aI1093
aI767
aI3411
aI13087
aI23569
aI42981
aI35757
asS'd'
I7609
sa(dp15218
S'a'
I15983
sS's'
I17
sS'm'
(lp15219
I1
aI1
aI1
aI15
aI27
aI7
aI51
aI101
aI429
aI939
aI111
aI781
aI2055
aI14227
aI17821
aI42097
aI32485
asS'd'
I7610
sa(dp15220
S'a'
I15991
sS's'
I17
sS'm'
(lp15221
I1
aI3
aI7
aI13
aI11
aI21
aI3
aI161
aI353
aI389
aI285
aI2633
aI6245
aI7089
aI21907
aI40765
aI88869
asS'd'
I7611
sa(dp15222
S'a'
I16004
sS's'
I17
sS'm'
(lp15223
I1
aI1
aI5
aI9
aI7
aI27
aI101
aI203
aI243
aI897
aI1375
aI1619
aI5275
aI12935
aI22103
aI38005
aI65603
asS'd'
I7612
sa(dp15224
S'a'
I16011
sS's'
I17
sS'm'
(lp15225
I1
aI1
aI5
aI9
aI13
aI25
aI15
aI21
aI447
aI7
aI947
aI1613
aI5055
aI129
aI18057
aI58551
aI6603
asS'd'
I7613
sa(dp15226
S'a'
I16016
sS's'
I17
sS'm'
(lp15227
I1
aI3
aI7
aI15
aI17
aI41
aI11
aI55
aI103
aI339
aI349
aI1813
aI7423
aI11837
aI20641
aI51951
aI61615
asS'd'
I7614
sa(dp15228
S'a'
I16019
sS's'
I17
sS'm'
(lp15229
I1
aI3
aI3
aI15
aI21
aI59
aI113
aI3
aI123
aI689
aI465
aI3039
aI4109
aI3241
aI30317
aI65053
aI117845
asS'd'
I7615
sa(dp15230
S'a'
I16025
sS's'
I17
sS'm'
(lp15231
I1
aI3
aI3
aI1
aI31
aI33
aI73
aI155
aI245
aI401
aI473
aI51
aI1387
aI489
aI10573
aI55401
aI106733
asS'd'
I7616
sa(dp15232
S'a'
I16041
sS's'
I17
sS'm'
(lp15233
I1
aI3
aI3
aI1
aI31
aI37
aI15
aI139
aI127
aI201
aI229
aI1753
aI7287
aI9045
aI18321
aI63485
aI26399
asS'd'
I7617
sa(dp15234
S'a'
I16064
sS's'
I17
sS'm'
(lp15235
I1
aI3
aI5
aI5
aI5
aI23
aI93
aI3
aI125
aI715
aI1827
aI419
aI1213
aI9031
aI25139
aI20771
aI41345
asS'd'
I7618
sa(dp15236
S'a'
I16067
sS's'
I17
sS'm'
(lp15237
I1
aI3
aI5
aI15
aI23
aI15
aI13
aI145
aI105
aI477
aI1131
aI2699
aI1929
aI10447
aI9655
aI26791
aI80101
asS'd'
I7619
sa(dp15238
S'a'
I16074
sS's'
I17
sS'm'
(lp15239
I1
aI1
aI1
aI13
aI1
aI35
aI75
aI73
aI269
aI851
aI737
aI1909
aI6805
aI11359
aI28991
aI52435
aI83767
asS'd'
I7620
sa(dp15240
S'a'
I16082
sS's'
I17
sS'm'
(lp15241
I1
aI1
aI7
aI5
aI11
aI31
aI31
aI91
aI111
aI161
aI1865
aI2545
aI133
aI12215
aI8957
aI20671
aI92975
asS'd'
I7621
sa(dp15242
S'a'
I16103
sS's'
I17
sS'm'
(lp15243
I1
aI1
aI7
aI5
aI25
aI53
aI55
aI121
aI53
aI457
aI831
aI2493
aI339
aI10955
aI30783
aI9095
aI97921
asS'd'
I7622
sa(dp15244
S'a'
I16109
sS's'
I17
sS'm'
(lp15245
I1
aI1
aI5
aI3
aI25
aI33
aI81
aI51
aI211
aI737
aI1865
aI4039
aI6931
aI8473
aI22459
aI24885
aI96355
asS'd'
I7623
sa(dp15246
S'a'
I16122
sS's'
I17
sS'm'
(lp15247
I1
aI3
aI7
aI13
aI23
aI5
aI101
aI171
aI65
aI793
aI443
aI411
aI7629
aI14791
aI28633
aI9055
aI123763
asS'd'
I7624
sa(dp15248
S'a'
I16149
sS's'
I17
sS'm'
(lp15249
I1
aI3
aI3
aI1
aI11
aI7
aI99
aI79
aI461
aI481
aI1689
aI3777
aI2125
aI4783
aI13061
aI19537
aI68109
asS'd'
I7625
sa(dp15250
S'a'
I16170
sS's'
I17
sS'm'
(lp15251
I1
aI1
aI3
aI11
aI31
aI53
aI109
aI7
aI49
aI925
aI1017
aI2371
aI1537
aI13557
aI75
aI40677
aI49181
asS'd'
I7626
sa(dp15252
S'a'
I16175
sS's'
I17
sS'm'
(lp15253
I1
aI3
aI3
aI3
aI9
aI1
aI95
aI113
aI189
aI389
aI377
aI393
aI6523
aI3183
aI6461
aI30201
aI66549
asS'd'
I7627
sa(dp15254
S'a'
I16178
sS's'
I17
sS'm'
(lp15255
I1
aI1
aI7
aI15
aI13
aI19
aI41
aI171
aI475
aI157
aI949
aI3245
aI5581
aI2783
aI25263
aI53023
aI11155
asS'd'
I7628
sa(dp15256
S'a'
I16189
sS's'
I17
sS'm'
(lp15257
I1
aI3
aI5
aI7
aI29
aI63
aI61
aI65
aI315
aI595
aI905
aI899
aI5059
aI4243
aI27287
aI14023
aI64213
asS'd'
I7629
sa(dp15258
S'a'
I16202
sS's'
I17
sS'm'
(lp15259
I1
aI3
aI1
aI3
aI15
aI37
aI109
aI161
aI9
aI867
aI1023
aI2513
aI4593
aI7747
aI1505
aI4801
aI127091
asS'd'
I7630
sa(dp15260
S'a'
I16204
sS's'
I17
sS'm'
(lp15261
I1
aI3
aI1
aI7
aI11
aI59
aI75
aI129
aI469
aI695
aI63
aI2757
aI6357
aI8675
aI6193
aI23439
aI66445
asS'd'
I7631
sa(dp15262
S'a'
I16222
sS's'
I17
sS'm'
(lp15263
I1
aI1
aI3
aI13
aI17
aI9
aI47
aI91
aI161
aI265
aI139
aI129
aI6707
aI9659
aI8917
aI54757
aI77835
asS'd'
I7632
sa(dp15264
S'a'
I16231
sS's'
I17
sS'm'
(lp15265
I1
aI1
aI3
aI13
aI19
aI37
aI113
aI255
aI99
aI913
aI1445
aI487
aI337
aI1001
aI16395
aI37141
aI66595
asS'd'
I7633
sa(dp15266
S'a'
I16238
sS's'
I17
sS'm'
(lp15267
I1
aI1
aI1
aI15
aI3
aI63
aI69
aI43
aI185
aI293
aI1137
aI2061
aI2377
aI8741
aI26817
aI5833
aI7807
asS'd'
I7634
sa(dp15268
S'a'
I16262
sS's'
I17
sS'm'
(lp15269
I1
aI1
aI1
aI5
aI3
aI29
aI39
aI33
aI263
aI355
aI597
aI539
aI5055
aI13075
aI8977
aI19829
aI88171
asS'd'
I7635
sa(dp15270
S'a'
I16265
sS's'
I17
sS'm'
(lp15271
I1
aI3
aI7
aI9
aI17
aI49
aI125
aI101
aI447
aI597
aI1337
aI559
aI2807
aI7925
aI12421
aI17427
aI34815
asS'd'
I7636
sa(dp15272
S'a'
I16276
sS's'
I17
sS'm'
(lp15273
I1
aI3
aI1
aI9
aI11
aI57
aI31
aI163
aI503
aI925
aI911
aI3721
aI2515
aI8429
aI25749
aI55209
aI90105
asS'd'
I7637
sa(dp15274
S'a'
I16285
sS's'
I17
sS'm'
(lp15275
I1
aI3
aI5
aI3
aI21
aI57
aI119
aI233
aI319
aI745
aI563
aI3057
aI2683
aI7063
aI11513
aI49157
aI64561
asS'd'
I7638
sa(dp15276
S'a'
I16313
sS's'
I17
sS'm'
(lp15277
I1
aI1
aI3
aI9
aI15
aI21
aI93
aI99
aI227
aI479
aI965
aI51
aI6941
aI9887
aI32409
aI23171
aI98387
asS'd'
I7639
sa(dp15278
S'a'
I16314
sS's'
I17
sS'm'
(lp15279
I1
aI3
aI5
aI5
aI19
aI1
aI47
aI49
aI233
aI931
aI971
aI2369
aI2827
aI1291
aI18653
aI725
aI19791
asS'd'
I7640
sa(dp15280
S'a'
I16321
sS's'
I17
sS'm'
(lp15281
I1
aI1
aI5
aI15
aI3
aI7
aI71
aI251
aI341
aI861
aI1203
aI793
aI7627
aI10929
aI10717
aI10677
aI49743
asS'd'
I7641
sa(dp15282
S'a'
I16327
sS's'
I17
sS'm'
(lp15283
I1
aI3
aI1
aI7
aI3
aI43
aI9
aI187
aI247
aI621
aI1069
aI2875
aI1525
aI4221
aI18813
aI35807
aI117609
asS'd'
I7642
sa(dp15284
S'a'
I16333
sS's'
I17
sS'm'
(lp15285
I1
aI3
aI3
aI3
aI29
aI39
aI83
aI201
aI205
aI337
aI231
aI547
aI2893
aI2483
aI6197
aI26869
aI18921
asS'd'
I7643
sa(dp15286
S'a'
I16334
sS's'
I17
sS'm'
(lp15287
I1
aI1
aI7
aI3
aI23
aI29
aI33
aI137
aI491
aI691
aI979
aI65
aI5711
aI11685
aI5137
aI37993
aI37075
asS'd'
I7644
sa(dp15288
S'a'
I16364
sS's'
I17
sS'm'
(lp15289
I1
aI3
aI3
aI1
aI11
aI3
aI99
aI119
aI203
aI901
aI1887
aI879
aI7547
aI4613
aI31233
aI13279
aI105089
asS'd'
I7645
sa(dp15290
S'a'
I16369
sS's'
I17
sS'm'
(lp15291
I1
aI1
aI1
aI13
aI25
aI23
aI111
aI167
aI313
aI141
aI127
aI1223
aI5711
aI4101
aI10977
aI34695
aI128303
asS'd'
I7646
sa(dp15292
S'a'
I16370
sS's'
I17
sS'm'
(lp15293
I1
aI1
aI7
aI15
aI5
aI3
aI89
aI151
aI289
aI769
aI539
aI2883
aI8121
aI15403
aI22345
aI63765
aI117015
asS'd'
I7647
sa(dp15294
S'a'
I16375
sS's'
I17
sS'm'
(lp15295
I1
aI1
aI1
aI13
aI15
aI9
aI71
aI95
aI37
aI705
aI1575
aI3735
aI7445
aI2027
aI27523
aI53321
aI106085
asS'd'
I7648
sa(dp15296
S'a'
I16376
sS's'
I17
sS'm'
(lp15297
I1
aI3
aI5
aI7
aI5
aI29
aI7
aI25
aI181
aI491
aI1173
aI1947
aI3321
aI9233
aI17265
aI26999
aI97783
asS'd'
I7649
sa(dp15298
S'a'
I16379
sS's'
I17
sS'm'
(lp15299
I1
aI1
aI3
aI15
aI1
aI63
aI111
aI113
aI279
aI123
aI345
aI1529
aI2725
aI8643
aI8551
aI30073
aI26689
asS'd'
I7650
sa(dp15300
S'a'
I16393
sS's'
I17
sS'm'
(lp15301
I1
aI3
aI7
aI7
aI5
aI55
aI117
aI211
aI293
aI851
aI1491
aI3265
aI4009
aI14949
aI10297
aI16219
aI69983
asS'd'
I7651
sa(dp15302
S'a'
I16402
sS's'
I17
sS'm'
(lp15303
I1
aI1
aI3
aI11
aI23
aI45
aI35
aI91
aI97
aI191
aI417
aI3545
aI1733
aI3955
aI10763
aI10229
aI75027
asS'd'
I7652
sa(dp15304
S'a'
I16408
sS's'
I17
sS'm'
(lp15305
I1
aI1
aI3
aI13
aI3
aI61
aI69
aI205
aI379
aI627
aI295
aI3979
aI85
aI11305
aI2493
aI35583
aI3133
asS'd'
I7653
sa(dp15306
S'a'
I16418
sS's'
I17
sS'm'
(lp15307
I1
aI3
aI5
aI9
aI5
aI63
aI67
aI201
aI351
aI367
aI1009
aI739
aI5409
aI8715
aI28939
aI31511
aI34599
asS'd'
I7654
sa(dp15308
S'a'
I16438
sS's'
I17
sS'm'
(lp15309
I1
aI1
aI1
aI5
aI3
aI25
aI21
aI25
aI477
aI301
aI623
aI157
aI563
aI9457
aI24515
aI30135
aI107165
asS'd'
I7655
sa(dp15310
S'a'
I16441
sS's'
I17
sS'm'
(lp15311
I1
aI1
aI3
aI15
aI5
aI41
aI49
aI171
aI469
aI427
aI857
aI2165
aI1437
aI2151
aI24061
aI63243
aI105331
asS'd'
I7656
sa(dp15312
S'a'
I16447
sS's'
I17
sS'm'
(lp15313
I1
aI3
aI5
aI11
aI21
aI25
aI59
aI167
aI29
aI653
aI1503
aI2223
aI3889
aI4605
aI28381
aI36075
aI74907
asS'd'
I7657
sa(dp15314
S'a'
I16450
sS's'
I17
sS'm'
(lp15315
I1
aI3
aI7
aI7
aI17
aI55
aI73
aI127
aI33
aI319
aI1565
aI2761
aI6473
aI2187
aI19939
aI56687
aI112137
asS'd'
I7658
sa(dp15316
S'a'
I16455
sS's'
I17
sS'm'
(lp15317
I1
aI1
aI1
aI9
aI7
aI53
aI105
aI3
aI299
aI15
aI1009
aI607
aI6885
aI12875
aI20719
aI16841
aI70471
asS'd'
I7659
sa(dp15318
S'a'
I16459
sS's'
I17
sS'm'
(lp15319
I1
aI3
aI5
aI9
aI7
aI33
aI23
aI163
aI279
aI739
aI1541
aI3017
aI2309
aI11827
aI3875
aI44337
aI82063
asS'd'
I7660
sa(dp15320
S'a'
I16483
sS's'
I17
sS'm'
(lp15321
I1
aI1
aI1
aI5
aI19
aI53
aI109
aI193
aI331
aI339
aI477
aI4093
aI5177
aI13527
aI25731
aI64137
aI81411
asS'd'
I7661
sa(dp15322
S'a'
I16490
sS's'
I17
sS'm'
(lp15323
I1
aI3
aI7
aI13
aI15
aI63
aI101
aI145
aI127
aI13
aI1431
aI3581
aI4993
aI14287
aI12125
aI60217
aI102563
asS'd'
I7662
sa(dp15324
S'a'
I16492
sS's'
I17
sS'm'
(lp15325
I1
aI3
aI1
aI7
aI17
aI27
aI127
aI81
aI223
aI763
aI761
aI2061
aI1031
aI12251
aI14141
aI23587
aI124813
asS'd'
I7663
sa(dp15326
S'a'
I16495
sS's'
I17
sS'm'
(lp15327
I1
aI3
aI5
aI13
aI27
aI21
aI9
aI249
aI285
aI875
aI65
aI4075
aI6749
aI13417
aI3079
aI29343
aI87075
asS'd'
I7664
sa(dp15328
S'a'
I16523
sS's'
I17
sS'm'
(lp15329
I1
aI3
aI5
aI13
aI1
aI31
aI61
aI21
aI169
aI145
aI1681
aI1229
aI5059
aI13555
aI21373
aI35597
aI70669
asS'd'
I7665
sa(dp15330
S'a'
I16528
sS's'
I17
sS'm'
(lp15331
I1
aI3
aI7
aI15
aI23
aI31
aI43
aI237
aI139
aI9
aI1905
aI3197
aI801
aI14205
aI13323
aI18717
aI88523
asS'd'
I7666
sa(dp15332
S'a'
I16543
sS's'
I17
sS'm'
(lp15333
I1
aI1
aI1
aI11
aI1
aI7
aI21
aI83
aI15
aI459
aI537
aI4029
aI6973
aI4019
aI1
aI35147
aI16329
asS'd'
I7667
sa(dp15334
S'a'
I16553
sS's'
I17
sS'm'
(lp15335
I1
aI3
aI7
aI15
aI23
aI11
aI17
aI101
aI235
aI683
aI913
aI3529
aI4363
aI13899
aI3603
aI27741
aI74143
asS'd'
I7668
sa(dp15336
S'a'
I16562
sS's'
I17
sS'm'
(lp15337
I1
aI1
aI7
aI7
aI3
aI3
aI91
aI107
aI499
aI723
aI315
aI2805
aI5909
aI11041
aI18281
aI54981
aI76041
asS'd'
I7669
sa(dp15338
S'a'
I16564
sS's'
I17
sS'm'
(lp15339
I1
aI3
aI7
aI9
aI15
aI7
aI93
aI171
aI275
aI647
aI655
aI3565
aI2199
aI14795
aI21945
aI9373
aI122299
asS'd'
I7670
sa(dp15340
S'a'
I16576
sS's'
I17
sS'm'
(lp15341
I1
aI1
aI1
aI5
aI27
aI53
aI73
aI27
aI431
aI707
aI53
aI1281
aI49
aI13199
aI1973
aI18935
aI114821
asS'd'
I7671
sa(dp15342
S'a'
I16588
sS's'
I17
sS'm'
(lp15343
I1
aI1
aI3
aI3
aI25
aI1
aI17
aI159
aI217
aI413
aI1393
aI2119
aI5611
aI7659
aI6003
aI19927
aI22287
asS'd'
I7672
sa(dp15344
S'a'
I16612
sS's'
I17
sS'm'
(lp15345
I1
aI1
aI7
aI15
aI29
aI59
aI77
aI9
aI205
aI795
aI627
aI2167
aI2477
aI6841
aI17663
aI34871
aI79823
asS'd'
I7673
sa(dp15346
S'a'
I16630
sS's'
I17
sS'm'
(lp15347
I1
aI3
aI5
aI9
aI13
aI35
aI79
aI237
aI11
aI335
aI789
aI2291
aI13
aI853
aI20373
aI39049
aI407
asS'd'
I7674
sa(dp15348
S'a'
I16654
sS's'
I17
sS'm'
(lp15349
I1
aI1
aI5
aI7
aI13
aI27
aI21
aI173
aI137
aI659
aI123
aI2677
aI2153
aI14879
aI26737
aI56291
aI47613
asS'd'
I7675
sa(dp15350
S'a'
I16656
sS's'
I17
sS'm'
(lp15351
I1
aI3
aI5
aI15
aI23
aI47
aI15
aI109
aI311
aI597
aI261
aI2407
aI8139
aI3215
aI28169
aI60731
aI79937
asS'd'
I7676
sa(dp15352
S'a'
I16668
sS's'
I17
sS'm'
(lp15353
I1
aI3
aI3
aI5
aI11
aI61
aI71
aI29
aI189
aI741
aI1171
aI397
aI2669
aI10627
aI20037
aI51703
aI6697
asS'd'
I7677
sa(dp15354
S'a'
I16672
sS's'
I17
sS'm'
(lp15355
I1
aI3
aI3
aI3
aI9
aI41
aI125
aI1
aI381
aI399
aI349
aI3265
aI6337
aI8113
aI14869
aI5305
aI83409
asS'd'
I7678
sa(dp15356
S'a'
I16675
sS's'
I17
sS'm'
(lp15357
I1
aI1
aI3
aI13
aI5
aI19
aI33
aI225
aI45
aI55
aI1809
aI1037
aI5443
aI15719
aI9963
aI363
aI15145
asS'd'
I7679
sa(dp15358
S'a'
I16678
sS's'
I17
sS'm'
(lp15359
I1
aI3
aI7
aI1
aI31
aI25
aI103
aI29
aI207
aI169
aI305
aI913
aI7501
aI15323
aI10575
aI13477
aI65245
asS'd'
I7680
sa(dp15360
S'a'
I16681
sS's'
I17
sS'm'
(lp15361
I1
aI3
aI3
aI15
aI13
aI23
aI69
aI255
aI333
aI157
aI279
aI1989
aI3439
aI12955
aI13649
aI52431
aI90009
asS'd'
I7681
sa(dp15362
S'a'
I16689
sS's'
I17
sS'm'
(lp15363
I1
aI3
aI7
aI5
aI23
aI61
aI111
aI121
aI79
aI469
aI89
aI1545
aI3405
aI12393
aI2035
aI15989
aI84855
asS'd'
I7682
sa(dp15364
S'a'
I16699
sS's'
I17
sS'm'
(lp15365
I1
aI1
aI7
aI5
aI17
aI21
aI127
aI151
aI283
aI521
aI5
aI3023
aI5365
aI11633
aI21177
aI42207
aI48925
asS'd'
I7683
sa(dp15366
S'a'
I16719
sS's'
I17
sS'm'
(lp15367
I1
aI3
aI7
aI5
aI21
aI21
aI61
aI17
aI415
aI879
aI1485
aI3727
aI935
aI9899
aI23241
aI651
aI103701
asS'd'
I7684
sa(dp15368
S'a'
I16734
sS's'
I17
sS'm'
(lp15369
I1
aI3
aI5
aI15
aI31
aI47
aI19
aI245
aI249
aI467
aI253
aI1575
aI337
aI863
aI19353
aI13153
aI125453
asS'd'
I7685
sa(dp15370
S'a'
I16737
sS's'
I17
sS'm'
(lp15371
I1
aI1
aI7
aI15
aI9
aI41
aI39
aI63
aI139
aI875
aI1011
aI1961
aI1627
aI7461
aI28961
aI47195
aI16239
asS'd'
I7686
sa(dp15372
S'a'
I16750
sS's'
I17
sS'm'
(lp15373
I1
aI3
aI3
aI7
aI27
aI55
aI51
aI245
aI231
aI619
aI43
aI91
aI2125
aI2685
aI23661
aI10189
aI43085
asS'd'
I7687
sa(dp15374
S'a'
I16752
sS's'
I17
sS'm'
(lp15375
I1
aI1
aI7
aI9
aI27
aI55
aI35
aI139
aI187
aI143
aI1545
aI2685
aI3173
aI12065
aI21607
aI42619
aI105279
asS'd'
I7688
sa(dp15376
S'a'
I16757
sS's'
I17
sS'm'
(lp15377
I1
aI1
aI5
aI3
aI29
aI63
aI15
aI197
aI49
aI995
aI389
aI1959
aI2441
aI11509
aI31753
aI40539
aI26989
asS'd'
I7689
sa(dp15378
S'a'
I16761
sS's'
I17
sS'm'
(lp15379
I1
aI3
aI7
aI15
aI19
aI37
aI17
aI37
aI305
aI469
aI945
aI2335
aI1493
aI13843
aI19905
aI49031
aI107893
asS'd'
I7690
sa(dp15380
S'a'
I16773
sS's'
I17
sS'm'
(lp15381
I1
aI3
aI1
aI11
aI3
aI35
aI113
aI181
aI223
aI27
aI485
aI2435
aI3423
aI11321
aI1687
aI45755
aI18017
asS'd'
I7691
sa(dp15382
S'a'
I16774
sS's'
I17
sS'm'
(lp15383
I1
aI3
aI3
aI13
aI17
aI47
aI109
aI145
aI287
aI769
aI1373
aI3423
aI1251
aI14357
aI3209
aI28363
aI97987
asS'd'
I7692
sa(dp15384
S'a'
I16801
sS's'
I17
sS'm'
(lp15385
I1
aI1
aI3
aI13
aI7
aI25
aI93
aI11
aI23
aI331
aI517
aI1705
aI1957
aI291
aI763
aI10411
aI120367
asS'd'
I7693
sa(dp15386
S'a'
I16802
sS's'
I17
sS'm'
(lp15387
I1
aI3
aI7
aI15
aI25
aI9
aI1
aI33
aI83
aI61
aI97
aI509
aI5387
aI8701
aI14243
aI31883
aI7375
asS'd'
I7694
sa(dp15388
S'a'
I16822
sS's'
I17
sS'm'
(lp15389
I1
aI3
aI1
aI5
aI19
aI11
aI59
aI95
aI265
aI205
aI533
aI1857
aI693
aI12469
aI24445
aI19449
aI130623
asS'd'
I7695
sa(dp15390
S'a'
I16831
sS's'
I17
sS'm'
(lp15391
I1
aI1
aI7
aI7
aI1
aI5
aI15
aI159
aI333
aI361
aI391
aI1889
aI2645
aI15115
aI30709
aI60515
aI13315
asS'd'
I7696
sa(dp15392
S'a'
I16840
sS's'
I17
sS'm'
(lp15393
I1
aI3
aI5
aI15
aI25
aI61
aI69
aI213
aI183
aI575
aI1573
aI3147
aI1753
aI2387
aI23063
aI12853
aI108507
asS'd'
I7697
sa(dp15394
S'a'
I16854
sS's'
I17
sS'm'
(lp15395
I1
aI1
aI1
aI15
aI17
aI31
aI11
aI177
aI411
aI23
aI469
aI3985
aI2159
aI2273
aI14175
aI20425
aI107741
asS'd'
I7698
sa(dp15396
S'a'
I16858
sS's'
I17
sS'm'
(lp15397
I1
aI1
aI3
aI9
aI5
aI35
aI55
aI225
aI263
aI641
aI1393
aI1277
aI595
aI2671
aI7039
aI64999
aI114387
asS'd'
I7699
sa(dp15398
S'a'
I16863
sS's'
I17
sS'm'
(lp15399
I1
aI1
aI3
aI3
aI11
aI23
aI1
aI161
aI77
aI755
aI1325
aI1773
aI4291
aI13119
aI29677
aI27295
aI81713
asS'd'
I7700
sa(dp15400
S'a'
I16867
sS's'
I17
sS'm'
(lp15401
I1
aI1
aI5
aI13
aI31
aI45
aI115
aI141
aI449
aI171
aI1413
aI2411
aI7937
aI10859
aI19453
aI64403
aI45169
asS'd'
I7701
sa(dp15402
S'a'
I16876
sS's'
I17
sS'm'
(lp15403
I1
aI3
aI5
aI7
aI1
aI27
aI117
aI157
aI99
aI119
aI1281
aI2633
aI5117
aI16009
aI19545
aI7421
aI30807
asS'd'
I7702
sa(dp15404
S'a'
I16891
sS's'
I17
sS'm'
(lp15405
I1
aI1
aI3
aI13
aI19
aI11
aI61
aI239
aI331
aI731
aI1723
aI1773
aI2623
aI15255
aI17197
aI63793
aI100433
asS'd'
I7703
sa(dp15406
S'a'
I16894
sS's'
I17
sS'm'
(lp15407
I1
aI3
aI7
aI11
aI11
aI7
aI119
aI33
aI195
aI521
aI811
aI2599
aI3113
aI5497
aI16751
aI2541
aI21813
asS'd'
I7704
sa(dp15408
S'a'
I16898
sS's'
I17
sS'm'
(lp15409
I1
aI1
aI1
aI15
aI23
aI47
aI25
aI73
aI429
aI213
aI557
aI1613
aI7055
aI7211
aI2225
aI1345
aI58033
asS'd'
I7705
sa(dp15410
S'a'
I16907
sS's'
I17
sS'm'
(lp15411
I1
aI1
aI1
aI13
aI15
aI39
aI69
aI71
aI11
aI543
aI267
aI2803
aI4853
aI9819
aI603
aI4629
aI78343
asS'd'
I7706
sa(dp15412
S'a'
I16915
sS's'
I17
sS'm'
(lp15413
I1
aI1
aI7
aI1
aI15
aI55
aI47
aI223
aI63
aI679
aI1135
aI3225
aI3845
aI12031
aI6761
aI20337
aI29021
asS'd'
I7707
sa(dp15414
S'a'
I16917
sS's'
I17
sS'm'
(lp15415
I1
aI1
aI3
aI3
aI3
aI51
aI127
aI103
aI43
aI379
aI169
aI2549
aI7775
aI2553
aI27415
aI30671
aI34043
asS'd'
I7708
sa(dp15416
S'a'
I16922
sS's'
I17
sS'm'
(lp15417
I1
aI1
aI3
aI11
aI1
aI31
aI89
aI113
aI475
aI857
aI499
aI3901
aI5343
aI8819
aI4503
aI58757
aI60513
asS'd'
I7709
sa(dp15418
S'a'
I16924
sS's'
I17
sS'm'
(lp15419
I1
aI3
aI5
aI11
aI27
aI49
aI97
aI217
aI91
aI971
aI1835
aI3447
aI2021
aI3747
aI20533
aI13659
aI84007
asS'd'
I7710
sa(dp15420
S'a'
I16933
sS's'
I17
sS'm'
(lp15421
I1
aI1
aI5
aI1
aI31
aI39
aI49
aI21
aI135
aI983
aI579
aI3509
aI3611
aI15101
aI29781
aI49941
aI14353
asS'd'
I7711
sa(dp15422
S'a'
I16938
sS's'
I17
sS'm'
(lp15423
I1
aI1
aI1
aI9
aI7
aI17
aI55
aI233
aI295
aI161
aI823
aI3823
aI4771
aI13531
aI24197
aI42629
aI60269
asS'd'
I7712
sa(dp15424
S'a'
I16952
sS's'
I17
sS'm'
(lp15425
I1
aI1
aI3
aI15
aI23
aI5
aI101
aI167
aI55
aI297
aI1733
aI3819
aI7041
aI9915
aI27803
aI60359
aI10249
asS'd'
I7713
sa(dp15426
S'a'
I16960
sS's'
I17
sS'm'
(lp15427
I1
aI1
aI7
aI9
aI25
aI47
aI67
aI253
aI303
aI313
aI1389
aI3785
aI2729
aI11471
aI27267
aI42783
aI111595
asS'd'
I7714
sa(dp15428
S'a'
I16963
sS's'
I17
sS'm'
(lp15429
I1
aI1
aI5
aI13
aI25
aI63
aI17
aI195
aI457
aI793
aI1553
aI1673
aI6799
aI12171
aI9003
aI22195
aI90229
asS'd'
I7715
sa(dp15430
S'a'
I16969
sS's'
I17
sS'm'
(lp15431
I1
aI1
aI3
aI15
aI11
aI43
aI43
aI221
aI423
aI985
aI873
aI599
aI1753
aI4875
aI7149
aI34625
aI8941
asS'd'
I7716
sa(dp15432
S'a'
I16978
sS's'
I17
sS'm'
(lp15433
I1
aI3
aI5
aI11
aI1
aI7
aI109
aI163
aI309
aI477
aI1291
aI3019
aI1933
aI14055
aI15005
aI1141
aI66867
asS'd'
I7717
sa(dp15434
S'a'
I17014
sS's'
I17
sS'm'
(lp15435
I1
aI3
aI3
aI15
aI21
aI35
aI95
aI131
aI413
aI1009
aI147
aI2165
aI6333
aI8313
aI20873
aI18377
aI23579
asS'd'
I7718
sa(dp15436
S'a'
I17020
sS's'
I17
sS'm'
(lp15437
I1
aI3
aI1
aI5
aI21
aI49
aI29
aI187
aI67
aI419
aI253
aI2345
aI3179
aI12331
aI23127
aI8799
aI102493
asS'd'
I7719
sa(dp15438
S'a'
I17034
sS's'
I17
sS'm'
(lp15439
I1
aI1
aI7
aI5
aI29
aI59
aI13
aI189
aI377
aI595
aI1893
aI527
aI7993
aI14867
aI24671
aI14585
aI38645
asS'd'
I7720
sa(dp15440
S'a'
I17036
sS's'
I17
sS'm'
(lp15441
I1
aI3
aI5
aI13
aI3
aI11
aI99
aI69
aI253
aI833
aI1961
aI2719
aI3953
aI8143
aI21277
aI16257
aI26929
asS'd'
I7721
sa(dp15442
S'a'
I17042
sS's'
I17
sS'm'
(lp15443
I1
aI3
aI7
aI3
aI3
aI19
aI19
aI57
aI393
aI187
aI945
aI2107
aI669
aI14785
aI13895
aI26907
aI92439
asS'd'
I7722
sa(dp15444
S'a'
I17047
sS's'
I17
sS'm'
(lp15445
I1
aI3
aI5
aI15
aI11
aI5
aI73
aI167
aI99
aI887
aI1213
aI2019
aI3781
aI14345
aI30249
aI16215
aI1893
asS'd'
I7723
sa(dp15446
S'a'
I17051
sS's'
I17
sS'm'
(lp15447
I1
aI1
aI5
aI1
aI17
aI11
aI69
aI145
aI97
aI393
aI1587
aI2513
aI1011
aI6933
aI7945
aI41387
aI34361
asS'd'
I7724
sa(dp15448
S'a'
I17054
sS's'
I17
sS'm'
(lp15449
I1
aI1
aI5
aI1
aI5
aI59
aI57
aI1
aI501
aI855
aI1485
aI977
aI4981
aI7631
aI31853
aI30737
aI103023
asS'd'
I7725
sa(dp15450
S'a'
I17063
sS's'
I17
sS'm'
(lp15451
I1
aI3
aI1
aI5
aI3
aI27
aI55
aI171
aI317
aI641
aI1875
aI2523
aI1631
aI4971
aI18743
aI25119
aI118913
asS'd'
I7726
sa(dp15452
S'a'
I17069
sS's'
I17
sS'm'
(lp15453
I1
aI1
aI3
aI15
aI7
aI39
aI73
aI209
aI125
aI29
aI1031
aI1569
aI1793
aI5461
aI985
aI59441
aI92997
asS'd'
I7727
sa(dp15454
S'a'
I17075
sS's'
I17
sS'm'
(lp15455
I1
aI3
aI5
aI11
aI27
aI23
aI57
aI13
aI65
aI555
aI1309
aI1149
aI5125
aI11573
aI3835
aI57913
aI78699
asS'd'
I7728
sa(dp15456
S'a'
I17077
sS's'
I17
sS'm'
(lp15457
I1
aI3
aI7
aI5
aI29
aI7
aI51
aI131
aI443
aI623
aI1491
aI1067
aI6647
aI6277
aI25799
aI54843
aI90869
asS'd'
I7729
sa(dp15458
S'a'
I17089
sS's'
I17
sS'm'
(lp15459
I1
aI1
aI1
aI11
aI7
aI33
aI67
aI113
aI319
aI665
aI11
aI1225
aI3137
aI16269
aI20101
aI40263
aI31091
asS'd'
I7730
sa(dp15460
S'a'
I17090
sS's'
I17
sS'm'
(lp15461
I1
aI3
aI5
aI15
aI7
aI5
aI101
aI153
aI165
aI173
aI97
aI1651
aI6633
aI6071
aI29079
aI35641
aI77305
asS'd'
I7731
sa(dp15462
S'a'
I17107
sS's'
I17
sS'm'
(lp15463
I1
aI3
aI7
aI13
aI9
aI45
aI103
aI55
aI121
aI1021
aI1841
aI315
aI8127
aI6547
aI1093
aI7181
aI39575
asS'd'
I7732
sa(dp15464
S'a'
I17126
sS's'
I17
sS'm'
(lp15465
I1
aI3
aI3
aI11
aI15
aI17
aI27
aI55
aI341
aI443
aI377
aI681
aI3635
aI1091
aI16719
aI49403
aI85507
asS'd'
I7733
sa(dp15466
S'a'
I17135
sS's'
I17
sS'm'
(lp15467
I1
aI3
aI5
aI5
aI29
aI53
aI51
aI213
aI273
aI475
aI981
aI549
aI539
aI14989
aI4037
aI23911
aI45997
asS'd'
I7734
sa(dp15468
S'a'
I17150
sS's'
I17
sS'm'
(lp15469
I1
aI3
aI5
aI3
aI27
aI37
aI73
aI115
aI331
aI911
aI991
aI4049
aI6299
aI3919
aI10231
aI31507
aI98651
asS'd'
I7735
sa(dp15470
S'a'
I17162
sS's'
I17
sS'm'
(lp15471
I1
aI1
aI5
aI13
aI21
aI13
aI1
aI175
aI137
aI837
aI1067
aI2845
aI307
aI4399
aI15671
aI1309
aI107409
asS'd'
I7736
sa(dp15472
S'a'
I17169
sS's'
I17
sS'm'
(lp15473
I1
aI1
aI3
aI1
aI5
aI47
aI111
aI75
aI193
aI389
aI157
aI3731
aI6237
aI5053
aI9933
aI28413
aI32939
asS'd'
I7737
sa(dp15474
S'a'
I17172
sS's'
I17
sS'm'
(lp15475
I1
aI1
aI7
aI5
aI29
aI1
aI51
aI85
aI267
aI935
aI1021
aI3135
aI3135
aI9263
aI32597
aI6779
aI71473
asS'd'
I7738
sa(dp15476
S'a'
I17175
sS's'
I17
sS'm'
(lp15477
I1
aI3
aI5
aI9
aI21
aI59
aI27
aI99
aI155
aI507
aI1911
aI3501
aI4307
aI6755
aI17127
aI29815
aI1577
asS'd'
I7739
sa(dp15478
S'a'
I17176
sS's'
I17
sS'm'
(lp15479
I1
aI1
aI5
aI1
aI15
aI63
aI45
aI105
aI125
aI299
aI689
aI3935
aI7229
aI5007
aI25003
aI30453
aI27819
asS'd'
I7740
sa(dp15480
S'a'
I17191
sS's'
I17
sS'm'
(lp15481
I1
aI1
aI7
aI15
aI19
aI9
aI67
aI151
aI45
aI985
aI2015
aI833
aI5435
aI15383
aI25881
aI46735
aI56717
asS'd'
I7741
sa(dp15482
S'a'
I17209
sS's'
I17
sS'm'
(lp15483
I1
aI1
aI5
aI15
aI27
aI59
aI119
aI163
aI293
aI63
aI1251
aI1309
aI485
aI4937
aI27207
aI47481
aI114357
asS'd'
I7742
sa(dp15484
S'a'
I17218
sS's'
I17
sS'm'
(lp15485
I1
aI3
aI5
aI13
aI23
aI11
aI111
aI87
aI329
aI467
aI1657
aI3309
aI3421
aI12013
aI23163
aI14105
aI88761
asS'd'
I7743
sa(dp15486
S'a'
I17220
sS's'
I17
sS'm'
(lp15487
I1
aI1
aI5
aI11
aI17
aI63
aI9
aI61
aI299
aI585
aI341
aI3375
aI3213
aI15953
aI11455
aI5333
aI66889
asS'd'
I7744
sa(dp15488
S'a'
I17227
sS's'
I17
sS'm'
(lp15489
I1
aI3
aI5
aI5
aI5
aI35
aI57
aI235
aI137
aI543
aI77
aI2811
aI857
aI12793
aI10791
aI55711
aI93353
asS'd'
I7745
sa(dp15490
S'a'
I17229
sS's'
I17
sS'm'
(lp15491
I1
aI3
aI7
aI3
aI23
aI37
aI19
aI81
aI321
aI23
aI1625
aI2359
aI3569
aI4685
aI7385
aI32677
aI18073
asS'd'
I7746
sa(dp15492
S'a'
I17238
sS's'
I17
sS'm'
(lp15493
I1
aI3
aI3
aI7
aI21
aI35
aI81
aI229
aI207
aI547
aI1397
aI2709
aI7159
aI1265
aI16823
aI9921
aI29159
asS'd'
I7747
sa(dp15494
S'a'
I17251
sS's'
I17
sS'm'
(lp15495
I1
aI3
aI7
aI13
aI27
aI13
aI107
aI241
aI395
aI317
aI307
aI3927
aI1153
aI15915
aI25179
aI25173
aI21503
asS'd'
I7748
sa(dp15496
S'a'
I17257
sS's'
I17
sS'm'
(lp15497
I1
aI3
aI1
aI5
aI1
aI51
aI25
aI135
aI381
aI229
aI1491
aI2009
aI3331
aI16165
aI8169
aI65161
aI9335
asS'd'
I7749
sa(dp15498
S'a'
I17258
sS's'
I17
sS'm'
(lp15499
I1
aI1
aI5
aI5
aI17
aI15
aI57
aI221
aI183
aI225
aI1649
aI3701
aI299
aI12349
aI4691
aI64479
aI82237
asS'd'
I7750
sa(dp15500
S'a'
I17272
sS's'
I17
sS'm'
(lp15501
I1
aI3
aI7
aI7
aI31
aI39
aI65
aI183
aI149
aI67
aI1697
aI3933
aI3709
aI15501
aI12583
aI60117
aI88691
asS'd'
I7751
sa(dp15502
S'a'
I17277
sS's'
I17
sS'm'
(lp15503
I1
aI1
aI5
aI15
aI17
aI49
aI117
aI233
aI161
aI891
aI789
aI1347
aI4887
aI10713
aI10613
aI4389
aI42619
asS'd'
I7752
sa(dp15504
S'a'
I17308
sS's'
I17
sS'm'
(lp15505
I1
aI3
aI5
aI9
aI13
aI3
aI83
aI69
aI381
aI777
aI743
aI2843
aI7233
aI3285
aI8931
aI48667
aI120777
asS'd'
I7753
sa(dp15506
S'a'
I17311
sS's'
I17
sS'm'
(lp15507
I1
aI3
aI1
aI3
aI11
aI7
aI55
aI107
aI165
aI533
aI1897
aI3385
aI1069
aI12805
aI30125
aI42729
aI123977
asS'd'
I7754
sa(dp15508
S'a'
I17321
sS's'
I17
sS'm'
(lp15509
I1
aI1
aI1
aI5
aI13
aI17
aI103
aI237
aI77
aI537
aI1843
aI2817
aI7467
aI13647
aI15259
aI3525
aI18313
asS'd'
I7755
sa(dp15510
S'a'
I17329
sS's'
I17
sS'm'
(lp15511
I1
aI1
aI7
aI7
aI13
aI59
aI29
aI197
aI309
aI917
aI1173
aI2605
aI4313
aI12007
aI25611
aI60409
aI104931
asS'd'
I7756
sa(dp15512
S'a'
I17342
sS's'
I17
sS'm'
(lp15513
I1
aI3
aI3
aI3
aI27
aI57
aI7
aI207
aI491
aI467
aI1973
aI3075
aI8043
aI3977
aI14517
aI13179
aI47111
asS'd'
I7757
sa(dp15514
S'a'
I17344
sS's'
I17
sS'm'
(lp15515
I1
aI1
aI7
aI5
aI31
aI33
aI125
aI235
aI79
aI847
aI1893
aI3875
aI7513
aI1435
aI24959
aI46813
aI82053
asS'd'
I7758
sa(dp15516
S'a'
I17350
sS's'
I17
sS'm'
(lp15517
I1
aI3
aI7
aI5
aI3
aI53
aI103
aI1
aI215
aI71
aI787
aI223
aI1399
aI6793
aI11281
aI39201
aI122119
asS'd'
I7759
sa(dp15518
S'a'
I17356
sS's'
I17
sS'm'
(lp15519
I1
aI3
aI3
aI3
aI3
aI57
aI7
aI151
aI319
aI463
aI685
aI2917
aI4037
aI14929
aI11971
aI41827
aI57449
asS'd'
I7760
sa(dp15520
S'a'
I17371
sS's'
I17
sS'm'
(lp15521
I1
aI1
aI7
aI3
aI5
aI11
aI15
aI139
aI379
aI563
aI135
aI65
aI5633
aI7535
aI1451
aI18289
aI62457
asS'd'
I7761
sa(dp15522
S'a'
I17374
sS's'
I17
sS'm'
(lp15523
I1
aI1
aI1
aI15
aI11
aI23
aI37
aI57
aI205
aI107
aI995
aI151
aI3279
aI2015
aI28927
aI40731
aI95551
asS'd'
I7762
sa(dp15524
S'a'
I17392
sS's'
I17
sS'm'
(lp15525
I1
aI3
aI5
aI9
aI15
aI43
aI95
aI217
aI203
aI215
aI203
aI2207
aI8189
aI465
aI2175
aI29285
aI25375
asS'd'
I7763
sa(dp15526
S'a'
I17402
sS's'
I17
sS'm'
(lp15527
I1
aI3
aI3
aI5
aI19
aI59
aI51
aI123
aI285
aI721
aI1335
aI1777
aI1645
aI15811
aI16539
aI14637
aI123323
asS'd'
I7764
sa(dp15528
S'a'
I17410
sS's'
I17
sS'm'
(lp15529
I1
aI3
aI5
aI5
aI11
aI35
aI23
aI23
aI259
aI789
aI567
aI1921
aI4743
aI6635
aI6965
aI43025
aI43175
asS'd'
I7765
sa(dp15530
S'a'
I17421
sS's'
I17
sS'm'
(lp15531
I1
aI3
aI7
aI3
aI7
aI27
aI77
aI121
aI285
aI65
aI647
aI591
aI2553
aI7163
aI12057
aI43675
aI24227
asS'd'
I7766
sa(dp15532
S'a'
I17424
sS's'
I17
sS'm'
(lp15533
I1
aI1
aI5
aI9
aI3
aI25
aI17
aI85
aI235
aI715
aI1913
aI2391
aI3719
aI11029
aI18359
aI6323
aI4703
asS'd'
I7767
sa(dp15534
S'a'
I17427
sS's'
I17
sS'm'
(lp15535
I1
aI1
aI1
aI3
aI25
aI31
aI37
aI31
aI89
aI311
aI1797
aI3409
aI6785
aI9627
aI29721
aI58591
aI111429
asS'd'
I7768
sa(dp15536
S'a'
I17434
sS's'
I17
sS'm'
(lp15537
I1
aI3
aI1
aI5
aI9
aI37
aI47
aI45
aI419
aI115
aI1009
aI1359
aI65
aI1161
aI15673
aI16075
aI63895
asS'd'
I7769
sa(dp15538
S'a'
I17449
sS's'
I17
sS'm'
(lp15539
I1
aI1
aI3
aI5
aI25
aI47
aI27
aI87
aI441
aI547
aI1801
aI3589
aI3773
aI12215
aI14509
aI12669
aI99983
asS'd'
I7770
sa(dp15540
S'a'
I17452
sS's'
I17
sS'm'
(lp15541
I1
aI1
aI1
aI3
aI19
aI33
aI51
aI91
aI447
aI577
aI491
aI539
aI3177
aI7033
aI21633
aI51737
aI47089
asS'd'
I7771
sa(dp15542
S'a'
I17463
sS's'
I17
sS'm'
(lp15543
I1
aI1
aI3
aI15
aI23
aI53
aI93
aI113
aI143
aI973
aI999
aI2355
aI1489
aI3451
aI29821
aI23769
aI74633
asS'd'
I7772
sa(dp15544
S'a'
I17470
sS's'
I17
sS'm'
(lp15545
I1
aI3
aI7
aI11
aI27
aI1
aI35
aI109
aI111
aI51
aI425
aI3203
aI2585
aI11255
aI20939
aI281
aI92133
asS'd'
I7773
sa(dp15546
S'a'
I17477
sS's'
I17
sS'm'
(lp15547
I1
aI1
aI1
aI11
aI13
aI37
aI13
aI149
aI421
aI655
aI79
aI3093
aI6429
aI1145
aI27663
aI52861
aI81431
asS'd'
I7774
sa(dp15548
S'a'
I17482
sS's'
I17
sS'm'
(lp15549
I1
aI1
aI5
aI13
aI19
aI23
aI105
aI39
aI97
aI239
aI469
aI1047
aI4727
aI12009
aI8055
aI45557
aI124219
asS'd'
I7775
sa(dp15550
S'a'
I17490
sS's'
I17
sS'm'
(lp15551
I1
aI1
aI1
aI7
aI7
aI7
aI5
aI53
aI269
aI391
aI1893
aI2263
aI2109
aI11531
aI12109
aI31437
aI20445
asS'd'
I7776
sa(dp15552
S'a'
I17496
sS's'
I17
sS'm'
(lp15553
I1
aI1
aI3
aI11
aI9
aI35
aI69
aI209
aI93
aI455
aI1117
aI3297
aI2597
aI15035
aI17943
aI19955
aI829
asS'd'
I7777
sa(dp15554
S'a'
I17508
sS's'
I17
sS'm'
(lp15555
I1
aI1
aI5
aI7
aI23
aI23
aI101
aI71
aI339
aI553
aI1653
aI2997
aI1191
aI3121
aI4575
aI49979
aI17353
asS'd'
I7778
sa(dp15556
S'a'
I17511
sS's'
I17
sS'm'
(lp15557
I1
aI3
aI3
aI13
aI13
aI9
aI51
aI181
aI33
aI185
aI111
aI589
aI3117
aI10105
aI28811
aI31529
aI79657
asS'd'
I7779
sa(dp15558
S'a'
I17536
sS's'
I17
sS'm'
(lp15559
I1
aI1
aI5
aI3
aI9
aI57
aI103
aI65
aI211
aI473
aI519
aI3815
aI4087
aI2767
aI10213
aI37829
aI9523
asS'd'
I7780
sa(dp15560
S'a'
I17563
sS's'
I17
sS'm'
(lp15561
I1
aI1
aI5
aI7
aI7
aI31
aI81
aI161
aI311
aI617
aI1689
aI3133
aI57
aI14595
aI22783
aI18475
aI85811
asS'd'
I7781
sa(dp15562
S'a'
I17570
sS's'
I17
sS'm'
(lp15563
I1
aI3
aI5
aI5
aI21
aI51
aI99
aI249
aI7
aI525
aI885
aI3981
aI2851
aI529
aI947
aI29885
aI122901
asS'd'
I7782
sa(dp15564
S'a'
I17581
sS's'
I17
sS'm'
(lp15565
I1
aI3
aI1
aI5
aI1
aI23
aI85
aI91
aI309
aI747
aI183
aI1347
aI2399
aI15777
aI16205
aI15687
aI117333
asS'd'
I7783
sa(dp15566
S'a'
I17590
sS's'
I17
sS'm'
(lp15567
I1
aI3
aI7
aI3
aI29
aI21
aI99
aI137
aI297
aI229
aI1063
aI2747
aI6415
aI7791
aI4775
aI62863
aI50849
asS'd'
I7784
sa(dp15568
S'a'
I17608
sS's'
I17
sS'm'
(lp15569
I1
aI3
aI1
aI3
aI11
aI3
aI53
aI153
aI103
aI911
aI1595
aI1899
aI1049
aI11643
aI21105
aI61587
aI92399
asS'd'
I7785
sa(dp15570
S'a'
I17616
sS's'
I17
sS'm'
(lp15571
I1
aI1
aI5
aI15
aI29
aI55
aI99
aI101
aI181
aI453
aI1917
aI2081
aI4687
aI4335
aI2817
aI11861
aI103167
asS'd'
I7786
sa(dp15572
S'a'
I17621
sS's'
I17
sS'm'
(lp15573
I1
aI3
aI7
aI15
aI11
aI7
aI9
aI3
aI477
aI281
aI1141
aI453
aI4993
aI7843
aI6169
aI49785
aI53827
asS'd'
I7787
sa(dp15574
S'a'
I17628
sS's'
I17
sS'm'
(lp15575
I1
aI3
aI7
aI11
aI25
aI61
aI77
aI159
aI83
aI95
aI1223
aI85
aI6309
aI16145
aI18729
aI133
aI14193
asS'd'
I7788
sa(dp15576
S'a'
I17644
sS's'
I17
sS'm'
(lp15577
I1
aI3
aI3
aI1
aI7
aI27
aI97
aI183
aI263
aI59
aI915
aI3857
aI3349
aI8123
aI11251
aI55163
aI125703
asS'd'
I7789
sa(dp15578
S'a'
I17649
sS's'
I17
sS'm'
(lp15579
I1
aI3
aI5
aI5
aI17
aI33
aI57
aI55
aI503
aI811
aI953
aI349
aI1949
aI9127
aI31015
aI14475
aI116769
asS'd'
I7790
sa(dp15580
S'a'
I17652
sS's'
I17
sS'm'
(lp15581
I1
aI3
aI1
aI1
aI3
aI53
aI59
aI131
aI421
aI971
aI89
aI3047
aI3513
aI13599
aI4569
aI54525
aI54779
asS'd'
I7791
sa(dp15582
S'a'
I17679
sS's'
I17
sS'm'
(lp15583
I1
aI1
aI3
aI11
aI9
aI45
aI95
aI123
aI197
aI257
aI1073
aI1461
aI5
aI12701
aI12559
aI34989
aI71631
asS'd'
I7792
sa(dp15584
S'a'
I17691
sS's'
I17
sS'm'
(lp15585
I1
aI3
aI3
aI7
aI1
aI27
aI55
aI191
aI447
aI561
aI1975
aI2665
aI1341
aI8969
aI18519
aI47389
aI70847
asS'd'
I7793
sa(dp15586
S'a'
I17693
sS's'
I17
sS'm'
(lp15587
I1
aI1
aI5
aI5
aI3
aI31
aI15
aI165
aI95
aI423
aI233
aI2309
aI7777
aI3503
aI20105
aI3085
aI92349
asS'd'
I7794
sa(dp15588
S'a'
I17697
sS's'
I17
sS'm'
(lp15589
I1
aI3
aI1
aI13
aI23
aI61
aI7
aI55
aI157
aI1
aI83
aI515
aI2169
aI14397
aI18149
aI56855
aI117265
asS'd'
I7795
sa(dp15590
S'a'
I17698
sS's'
I17
sS'm'
(lp15591
I1
aI3
aI3
aI3
aI3
aI59
aI69
aI95
aI127
aI241
aI65
aI3145
aI491
aI13809
aI17529
aI20223
aI96579
asS'd'
I7796
sa(dp15592
S'a'
I17700
sS's'
I17
sS'm'
(lp15593
I1
aI1
aI1
aI5
aI13
aI43
aI117
aI163
aI305
aI955
aI2007
aI3337
aI807
aI16019
aI5721
aI5479
aI90937
asS'd'
I7797
sa(dp15594
S'a'
I17704
sS's'
I17
sS'm'
(lp15595
I1
aI3
aI3
aI1
aI19
aI9
aI127
aI5
aI113
aI491
aI1873
aI2127
aI7949
aI5207
aI32531
aI775
aI131065
asS'd'
I7798
sa(dp15596
S'a'
I17707
sS's'
I17
sS'm'
(lp15597
I1
aI1
aI7
aI3
aI1
aI3
aI91
aI187
aI37
aI873
aI1039
aI4075
aI5645
aI243
aI15127
aI45615
aI3813
asS'd'
I7799
sa(dp15598
S'a'
I17715
sS's'
I17
sS'm'
(lp15599
I1
aI1
aI3
aI11
aI3
aI37
aI67
aI59
aI439
aI763
aI213
aI1099
aI1659
aI12783
aI30297
aI60713
aI43497
asS'd'
I7800
sa(dp15600
S'a'
I17718
sS's'
I17
sS'm'
(lp15601
I1
aI3
aI3
aI13
aI23
aI49
aI47
aI191
aI245
aI985
aI487
aI3165
aI7803
aI2437
aI19073
aI30605
aI119641
asS'd'
I7801
sa(dp15602
S'a'
I17721
sS's'
I17
sS'm'
(lp15603
I1
aI3
aI7
aI7
aI19
aI43
aI7
aI253
aI93
aI99
aI145
aI219
aI699
aI2433
aI3009
aI565
aI99671
asS'd'
I7802
sa(dp15604
S'a'
I17722
sS's'
I17
sS'm'
(lp15605
I1
aI1
aI3
aI13
aI7
aI5
aI9
aI23
aI219
aI155
aI925
aI3427
aI6631
aI16353
aI4115
aI20831
aI49525
asS'd'
I7803
sa(dp15606
S'a'
I17727
sS's'
I17
sS'm'
(lp15607
I1
aI1
aI7
aI11
aI15
aI45
aI41
aI35
aI301
aI273
aI241
aI4031
aI5441
aI8281
aI9341
aI8499
aI73841
asS'd'
I7804
sa(dp15608
S'a'
I17729
sS's'
I17
sS'm'
(lp15609
I1
aI3
aI7
aI13
aI9
aI19
aI79
aI3
aI163
aI197
aI509
aI2301
aI6971
aI11525
aI8805
aI33805
aI111595
asS'd'
I7805
sa(dp15610
S'a'
I17747
sS's'
I17
sS'm'
(lp15611
I1
aI3
aI3
aI1
aI15
aI45
aI69
aI253
aI155
aI639
aI1045
aI749
aI3619
aI14871
aI18063
aI49763
aI66687
asS'd'
I7806
sa(dp15612
S'a'
I17754
sS's'
I17
sS'm'
(lp15613
I1
aI3
aI3
aI3
aI29
aI5
aI41
aI171
aI265
aI677
aI1719
aI2623
aI1721
aI12243
aI18741
aI39595
aI92873
asS'd'
I7807
sa(dp15614
S'a'
I17756
sS's'
I17
sS'm'
(lp15615
I1
aI3
aI5
aI7
aI27
aI23
aI69
aI61
aI453
aI399
aI1857
aI3901
aI6565
aI15595
aI1083
aI15065
aI91249
asS'd'
I7808
sa(dp15616
S'a'
I17760
sS's'
I17
sS'm'
(lp15617
I1
aI1
aI5
aI7
aI1
aI27
aI9
aI145
aI95
aI983
aI685
aI2079
aI5117
aI5037
aI22695
aI53135
aI43569
asS'd'
I7809
sa(dp15618
S'a'
I17765
sS's'
I17
sS'm'
(lp15619
I1
aI1
aI3
aI5
aI5
aI7
aI69
aI59
aI331
aI409
aI179
aI333
aI1293
aI3863
aI9473
aI12537
aI55605
asS'd'
I7810
sa(dp15620
S'a'
I17778
sS's'
I17
sS'm'
(lp15621
I1
aI3
aI5
aI1
aI1
aI19
aI1
aI49
aI317
aI769
aI91
aI2073
aI1765
aI1197
aI15029
aI52553
aI57361
asS'd'
I7811
sa(dp15622
S'a'
I17784
sS's'
I17
sS'm'
(lp15623
I1
aI1
aI5
aI1
aI23
aI13
aI11
aI69
aI345
aI877
aI41
aI817
aI617
aI14415
aI8337
aI53973
aI50007
asS'd'
I7812
sa(dp15624
S'a'
I17794
sS's'
I17
sS'm'
(lp15625
I1
aI1
aI7
aI3
aI19
aI27
aI69
aI103
aI115
aI893
aI219
aI2891
aI2813
aI9933
aI26401
aI63323
aI30909
asS'd'
I7813
sa(dp15626
S'a'
I17805
sS's'
I17
sS'm'
(lp15627
I1
aI1
aI5
aI5
aI27
aI15
aI119
aI3
aI11
aI783
aI541
aI2431
aI2395
aI3921
aI15471
aI34657
aI100295
asS'd'
I7814
sa(dp15628
S'a'
I17806
sS's'
I17
sS'm'
(lp15629
I1
aI1
aI7
aI11
aI15
aI25
aI39
aI191
aI345
aI1001
aI1773
aI715
aI1627
aI15957
aI30085
aI34097
aI58747
asS'd'
I7815
sa(dp15630
S'a'
I17817
sS's'
I17
sS'm'
(lp15631
I1
aI1
aI1
aI5
aI17
aI43
aI65
aI81
aI487
aI387
aI1359
aI145
aI2231
aI6693
aI15857
aI59539
aI79615
asS'd'
I7816
sa(dp15632
S'a'
I17824
sS's'
I17
sS'm'
(lp15633
I1
aI1
aI3
aI5
aI15
aI19
aI17
aI233
aI247
aI611
aI587
aI2587
aI2321
aI2835
aI1477
aI41991
aI88143
asS'd'
I7817
sa(dp15634
S'a'
I17839
sS's'
I17
sS'm'
(lp15635
I1
aI3
aI3
aI15
aI27
aI15
aI53
aI61
aI359
aI989
aI283
aI3569
aI5551
aI11849
aI20995
aI34065
aI69407
asS'd'
I7818
sa(dp15636
S'a'
I17842
sS's'
I17
sS'm'
(lp15637
I1
aI3
aI3
aI11
aI13
aI31
aI41
aI87
aI379
aI47
aI1289
aI3143
aI4213
aI8643
aI17065
aI10707
aI62773
asS'd'
I7819
sa(dp15638
S'a'
I17844
sS's'
I17
sS'm'
(lp15639
I1
aI3
aI7
aI1
aI9
aI61
aI59
aI121
aI453
aI663
aI27
aI793
aI4501
aI7713
aI285
aI13279
aI11633
asS'd'
I7820
sa(dp15640
S'a'
I17851
sS's'
I17
sS'm'
(lp15641
I1
aI1
aI7
aI5
aI29
aI51
aI57
aI15
aI233
aI743
aI879
aI2317
aI3399
aI15741
aI605
aI57529
aI87427
asS'd'
I7821
sa(dp15642
S'a'
I17862
sS's'
I17
sS'm'
(lp15643
I1
aI1
aI1
aI1
aI19
aI59
aI51
aI119
aI273
aI403
aI1649
aI3877
aI3561
aI10931
aI21139
aI2599
aI77957
asS'd'
I7822
sa(dp15644
S'a'
I17868
sS's'
I17
sS'm'
(lp15645
I1
aI3
aI1
aI3
aI5
aI1
aI79
aI131
aI251
aI585
aI359
aI2073
aI7041
aI13611
aI22937
aI24645
aI72827
asS'd'
I7823
sa(dp15646
S'a'
I17871
sS's'
I17
sS'm'
(lp15647
I1
aI3
aI7
aI9
aI19
aI39
aI93
aI137
aI359
aI565
aI1123
aI1301
aI4111
aI13683
aI1361
aI25147
aI38315
asS'd'
I7824
sa(dp15648
S'a'
I17873
sS's'
I17
sS'm'
(lp15649
I1
aI1
aI5
aI13
aI27
aI31
aI11
aI243
aI111
aI243
aI1615
aI1649
aI2999
aI15873
aI19161
aI57485
aI35819
asS'd'
I7825
sa(dp15650
S'a'
I17896
sS's'
I17
sS'm'
(lp15651
I1
aI3
aI3
aI5
aI25
aI57
aI61
aI207
aI219
aI969
aI303
aI1165
aI6753
aI13473
aI10789
aI52883
aI45205
asS'd'
I7826
sa(dp15652
S'a'
I17904
sS's'
I17
sS'm'
(lp15653
I1
aI1
aI7
aI11
aI9
aI53
aI55
aI175
aI399
aI981
aI255
aI2499
aI373
aI13131
aI26803
aI48017
aI25599
asS'd'
I7827
sa(dp15654
S'a'
I17909
sS's'
I17
sS'm'
(lp15655
I1
aI1
aI3
aI3
aI11
aI23
aI73
aI25
aI83
aI39
aI1813
aI747
aI3287
aI795
aI11917
aI55509
aI105057
asS'd'
I7828
sa(dp15656
S'a'
I17920
sS's'
I17
sS'm'
(lp15657
I1
aI3
aI7
aI15
aI29
aI59
aI47
aI171
aI219
aI875
aI71
aI123
aI8131
aI15595
aI12471
aI62439
aI131
asS'd'
I7829
sa(dp15658
S'a'
I17923
sS's'
I17
sS'm'
(lp15659
I1
aI3
aI5
aI13
aI9
aI27
aI119
aI233
aI323
aI943
aI375
aI3647
aI185
aI1639
aI431
aI27225
aI130175
asS'd'
I7830
sa(dp15660
S'a'
I17947
sS's'
I17
sS'm'
(lp15661
I1
aI3
aI7
aI3
aI7
aI17
aI31
aI155
aI89
aI835
aI1015
aI2019
aI3973
aI7087
aI16899
aI29591
aI40797
asS'd'
I7831
sa(dp15662
S'a'
I17950
sS's'
I17
sS'm'
(lp15663
I1
aI3
aI3
aI1
aI3
aI11
aI83
aI231
aI209
aI537
aI1227
aI1519
aI1059
aI14027
aI18591
aI34031
aI125755
asS'd'
I7832
sa(dp15664
S'a'
I17956
sS's'
I17
sS'm'
(lp15665
I1
aI3
aI3
aI1
aI7
aI39
aI19
aI99
aI169
aI961
aI385
aI1621
aI7373
aI7459
aI8979
aI23643
aI17101
asS'd'
I7833
sa(dp15666
S'a'
I17959
sS's'
I17
sS'm'
(lp15667
I1
aI1
aI3
aI11
aI11
aI23
aI61
aI37
aI359
aI981
aI209
aI2555
aI2673
aI6501
aI12731
aI10735
aI97321
asS'd'
I7834
sa(dp15668
S'a'
I17966
sS's'
I17
sS'm'
(lp15669
I1
aI1
aI3
aI13
aI15
aI61
aI115
aI119
aI99
aI755
aI1933
aI345
aI3133
aI12071
aI26657
aI7133
aI18553
asS'd'
I7835
sa(dp15670
S'a'
I17971
sS's'
I17
sS'm'
(lp15671
I1
aI3
aI1
aI5
aI17
aI7
aI29
aI119
aI445
aI911
aI89
aI19
aI6137
aI8037
aI19527
aI22467
aI29253
asS'd'
I7836
sa(dp15672
S'a'
I17973
sS's'
I17
sS'm'
(lp15673
I1
aI1
aI3
aI11
aI31
aI21
aI119
aI21
aI249
aI371
aI343
aI4037
aI7539
aI14473
aI23829
aI46415
aI60911
asS'd'
I7837
sa(dp15674
S'a'
I17992
sS's'
I17
sS'm'
(lp15675
I1
aI1
aI7
aI9
aI21
aI53
aI29
aI149
aI467
aI893
aI479
aI359
aI1007
aI13955
aI9667
aI10245
aI74761
asS'd'
I7838
sa(dp15676
S'a'
I18006
sS's'
I17
sS'm'
(lp15677
I1
aI3
aI1
aI7
aI7
aI45
aI7
aI77
aI289
aI271
aI1329
aI261
aI5675
aI8275
aI7443
aI57945
aI117825
asS'd'
I7839
sa(dp15678
S'a'
I18009
sS's'
I17
sS'm'
(lp15679
I1
aI1
aI1
aI3
aI21
aI57
aI117
aI77
aI287
aI119
aI1073
aI915
aI2521
aI455
aI7433
aI56953
aI84433
asS'd'
I7840
sa(dp15680
S'a'
I18010
sS's'
I17
sS'm'
(lp15681
I1
aI1
aI1
aI9
aI27
aI47
aI1
aI189
aI303
aI375
aI215
aI3117
aI4541
aI12877
aI15523
aI32317
aI104213
asS'd'
I7841
sa(dp15682
S'a'
I18022
sS's'
I17
sS'm'
(lp15683
I1
aI1
aI3
aI1
aI13
aI39
aI37
aI249
aI371
aI159
aI1073
aI1351
aI4703
aI2715
aI17463
aI3945
aI51523
asS'd'
I7842
sa(dp15684
S'a'
I18039
sS's'
I17
sS'm'
(lp15685
I1
aI3
aI5
aI5
aI29
aI15
aI79
aI25
aI61
aI995
aI1081
aI3377
aI345
aI13773
aI4205
aI20589
aI83591
asS'd'
I7843
sa(dp15686
S'a'
I18046
sS's'
I17
sS'm'
(lp15687
I1
aI1
aI3
aI1
aI9
aI1
aI41
aI39
aI389
aI509
aI561
aI3273
aI1911
aI15271
aI22655
aI34713
aI2045
asS'd'
I7844
sa(dp15688
S'a'
I18069
sS's'
I17
sS'm'
(lp15689
I1
aI3
aI1
aI15
aI17
aI1
aI55
aI55
aI119
aI707
aI843
aI2657
aI3687
aI11557
aI18331
aI4935
aI110639
asS'd'
I7845
sa(dp15690
S'a'
I18074
sS's'
I17
sS'm'
(lp15691
I1
aI3
aI5
aI1
aI23
aI35
aI119
aI215
aI471
aI643
aI1581
aI1965
aI2627
aI2991
aI3361
aI6737
aI111657
asS'd'
I7846
sa(dp15692
S'a'
I18076
sS's'
I17
sS'm'
(lp15693
I1
aI3
aI5
aI7
aI9
aI19
aI33
aI197
aI33
aI993
aI1795
aI595
aI7113
aI14721
aI12647
aI41035
aI13669
asS'd'
I7847
sa(dp15694
S'a'
I18085
sS's'
I17
sS'm'
(lp15695
I1
aI1
aI7
aI15
aI31
aI39
aI51
aI157
aI373
aI473
aI665
aI3541
aI6695
aI11741
aI5617
aI17189
aI129851
asS'd'
I7848
sa(dp15696
S'a'
I18086
sS's'
I17
sS'm'
(lp15697
I1
aI3
aI3
aI7
aI15
aI37
aI33
aI175
aI391
aI159
aI717
aI593
aI113
aI9331
aI10685
aI59003
aI26975
asS'd'
I7849
sa(dp15698
S'a'
I18095
sS's'
I17
sS'm'
(lp15699
I1
aI1
aI3
aI5
aI13
aI41
aI11
aI109
aI9
aI899
aI1503
aI901
aI6237
aI7789
aI9963
aI14923
aI63665
asS'd'
I7850
sa(dp15700
S'a'
I18100
sS's'
I17
sS'm'
(lp15701
I1
aI3
aI7
aI7
aI25
aI61
aI3
aI231
aI235
aI29
aI1049
aI1997
aI5371
aI9047
aI29595
aI49239
aI108649
asS'd'
I7851
sa(dp15702
S'a'
I18109
sS's'
I17
sS'm'
(lp15703
I1
aI1
aI3
aI5
aI27
aI1
aI53
aI209
aI315
aI747
aI1803
aI11
aI1815
aI6539
aI8839
aI18385
aI88681
asS'd'
I7852
sa(dp15704
S'a'
I18121
sS's'
I17
sS'm'
(lp15705
I1
aI1
aI5
aI13
aI25
aI55
aI117
aI197
aI13
aI689
aI751
aI1203
aI2277
aI1763
aI23453
aI54459
aI118023
asS'd'
I7853
sa(dp15706
S'a'
I18127
sS's'
I17
sS'm'
(lp15707
I1
aI3
aI3
aI11
aI21
aI1
aI51
aI101
aI429
aI723
aI273
aI3021
aI1491
aI9923
aI6629
aI63741
aI98813
asS'd'
I7854
sa(dp15708
S'a'
I18129
sS's'
I17
sS'm'
(lp15709
I1
aI1
aI1
aI15
aI17
aI25
aI111
aI251
aI43
aI403
aI465
aI17
aI787
aI6045
aI32185
aI22921
aI115851
asS'd'
I7855
sa(dp15710
S'a'
I18132
sS's'
I17
sS'm'
(lp15711
I1
aI1
aI5
aI11
aI11
aI13
aI13
aI93
aI489
aI941
aI1391
aI383
aI7735
aI1921
aI16199
aI53099
aI25181
asS'd'
I7856
sa(dp15712
S'a'
I18141
sS's'
I17
sS'm'
(lp15713
I1
aI3
aI3
aI7
aI15
aI1
aI3
aI159
aI507
aI867
aI1589
aI2111
aI3839
aI8989
aI12589
aI37657
aI97055
asS'd'
I7857
sa(dp15714
S'a'
I18146
sS's'
I17
sS'm'
(lp15715
I1
aI3
aI3
aI13
aI25
aI23
aI7
aI95
aI489
aI1001
aI105
aI2737
aI5013
aI14465
aI25383
aI57551
aI77425
asS'd'
I7858
sa(dp15716
S'a'
I18151
sS's'
I17
sS'm'
(lp15717
I1
aI3
aI5
aI3
aI3
aI7
aI81
aI15
aI255
aI297
aI1183
aI655
aI741
aI3081
aI2141
aI34493
aI103707
asS'd'
I7859
sa(dp15718
S'a'
I18157
sS's'
I17
sS'm'
(lp15719
I1
aI1
aI7
aI9
aI27
aI57
aI49
aI121
aI21
aI239
aI829
aI2001
aI613
aI9569
aI4419
aI20007
aI59109
asS'd'
I7860
sa(dp15720
S'a'
I18160
sS's'
I17
sS'm'
(lp15721
I1
aI3
aI7
aI1
aI3
aI21
aI109
aI255
aI45
aI333
aI915
aI1245
aI5893
aI765
aI28289
aI53927
aI15335
asS'd'
I7861
sa(dp15722
S'a'
I18183
sS's'
I17
sS'm'
(lp15723
I1
aI3
aI3
aI7
aI5
aI35
aI33
aI79
aI111
aI509
aI347
aI3915
aI2017
aI6461
aI11847
aI17807
aI48601
asS'd'
I7862
sa(dp15724
S'a'
I18204
sS's'
I17
sS'm'
(lp15725
I1
aI3
aI5
aI1
aI13
aI63
aI87
aI65
aI507
aI277
aI339
aI3637
aI6289
aI719
aI9383
aI38887
aI55061
asS'd'
I7863
sa(dp15726
S'a'
I18218
sS's'
I17
sS'm'
(lp15727
I1
aI1
aI5
aI15
aI17
aI5
aI59
aI107
aI355
aI1021
aI1849
aI1807
aI7679
aI305
aI31533
aI1221
aI98165
asS'd'
I7864
sa(dp15728
S'a'
I18226
sS's'
I17
sS'm'
(lp15729
I1
aI1
aI1
aI13
aI19
aI7
aI37
aI63
aI267
aI399
aI1451
aI2149
aI1003
aI13635
aI18693
aI215
aI15887
asS'd'
I7865
sa(dp15730
S'a'
I18238
sS's'
I17
sS'm'
(lp15731
I1
aI1
aI3
aI7
aI11
aI63
aI81
aI251
aI253
aI963
aI635
aI1697
aI6393
aI9775
aI24189
aI9099
aI106277
asS'd'
I7866
sa(dp15732
S'a'
I18245
sS's'
I17
sS'm'
(lp15733
I1
aI3
aI3
aI13
aI17
aI47
aI63
aI47
aI279
aI879
aI271
aI1655
aI1897
aI10743
aI2607
aI16695
aI32779
asS'd'
I7867
sa(dp15734
S'a'
I18249
sS's'
I17
sS'm'
(lp15735
I1
aI3
aI5
aI15
aI3
aI1
aI121
aI181
aI303
aI973
aI19
aI3327
aI3827
aI2197
aI31857
aI22835
aI122611
asS'd'
I7868
sa(dp15736
S'a'
I18260
sS's'
I17
sS'm'
(lp15737
I1
aI1
aI5
aI13
aI25
aI41
aI105
aI197
aI195
aI85
aI1515
aI2735
aI7539
aI7557
aI24297
aI38721
aI46895
asS'd'
I7869
sa(dp15738
S'a'
I18267
sS's'
I17
sS'm'
(lp15739
I1
aI1
aI1
aI1
aI15
aI63
aI33
aI7
aI43
aI971
aI781
aI1461
aI4483
aI2113
aI32459
aI37653
aI68017
asS'd'
I7870
sa(dp15740
S'a'
I18270
sS's'
I17
sS'm'
(lp15741
I1
aI3
aI3
aI9
aI7
aI1
aI65
aI183
aI171
aI695
aI191
aI3675
aI6749
aI6823
aI3577
aI45031
aI81597
asS'd'
I7871
sa(dp15742
S'a'
I18273
sS's'
I17
sS'm'
(lp15743
I1
aI3
aI3
aI3
aI9
aI61
aI13
aI159
aI295
aI329
aI943
aI3293
aI79
aI14497
aI21461
aI4667
aI96435
asS'd'
I7872
sa(dp15744
S'a'
I18274
sS's'
I17
sS'm'
(lp15745
I1
aI1
aI7
aI9
aI29
aI37
aI117
aI215
aI295
aI591
aI1139
aI3093
aI7469
aI7995
aI13581
aI48075
aI5943
asS'd'
I7873
sa(dp15746
S'a'
I18276
sS's'
I17
sS'm'
(lp15747
I1
aI3
aI1
aI9
aI11
aI11
aI117
aI255
aI419
aI551
aI1445
aI1987
aI1169
aI14227
aI31095
aI36041
aI63739
asS'd'
I7874
sa(dp15748
S'a'
I18283
sS's'
I17
sS'm'
(lp15749
I1
aI1
aI7
aI15
aI17
aI25
aI81
aI27
aI87
aI463
aI1871
aI551
aI7449
aI12231
aI28645
aI32663
aI43037
asS'd'
I7875
sa(dp15750
S'a'
I18307
sS's'
I17
sS'm'
(lp15751
I1
aI3
aI5
aI11
aI3
aI49
aI109
aI123
aI397
aI113
aI1269
aI2433
aI4463
aI1257
aI2127
aI6677
aI96009
asS'd'
I7876
sa(dp15752
S'a'
I18314
sS's'
I17
sS'm'
(lp15753
I1
aI1
aI1
aI11
aI27
aI19
aI83
aI123
aI297
aI867
aI941
aI3929
aI3483
aI4641
aI32505
aI48999
aI66169
asS'd'
I7877
sa(dp15754
S'a'
I18321
sS's'
I17
sS'm'
(lp15755
I1
aI1
aI5
aI11
aI5
aI21
aI11
aI255
aI369
aI859
aI657
aI587
aI5245
aI12973
aI22403
aI47935
aI121375
asS'd'
I7878
sa(dp15756
S'a'
I18334
sS's'
I17
sS'm'
(lp15757
I1
aI3
aI1
aI13
aI17
aI43
aI83
aI51
aI339
aI967
aI499
aI1485
aI5203
aI10053
aI31707
aI31443
aI75033
asS'd'
I7879
sa(dp15758
S'a'
I18355
sS's'
I17
sS'm'
(lp15759
I1
aI1
aI5
aI13
aI11
aI5
aI121
aI121
aI73
aI101
aI1751
aI3805
aI1333
aI14043
aI26957
aI27557
aI110899
asS'd'
I7880
sa(dp15760
S'a'
I18364
sS's'
I17
sS'm'
(lp15761
I1
aI3
aI7
aI11
aI9
aI7
aI125
aI237
aI437
aI177
aI841
aI175
aI5509
aI9157
aI3129
aI54119
aI109315
asS'd'
I7881
sa(dp15762
S'a'
I18372
sS's'
I17
sS'm'
(lp15763
I1
aI3
aI7
aI15
aI1
aI59
aI87
aI121
aI43
aI475
aI1589
aI1267
aI1501
aI1585
aI31705
aI33959
aI27247
asS'd'
I7882
sa(dp15764
S'a'
I18390
sS's'
I17
sS'm'
(lp15765
I1
aI1
aI5
aI3
aI27
aI63
aI117
aI205
aI169
aI701
aI1081
aI2835
aI8049
aI11875
aI4143
aI17663
aI90043
asS'd'
I7883
sa(dp15766
S'a'
I18399
sS's'
I17
sS'm'
(lp15767
I1
aI3
aI1
aI9
aI23
aI27
aI31
aI141
aI411
aI145
aI1177
aI3681
aI3403
aI6943
aI10729
aI47219
aI102713
asS'd'
I7884
sa(dp15768
S'a'
I18415
sS's'
I17
sS'm'
(lp15769
I1
aI1
aI7
aI11
aI5
aI49
aI33
aI27
aI121
aI865
aI471
aI1871
aI6945
aI10051
aI4493
aI7121
aI65551
asS'd'
I7885
sa(dp15770
S'a'
I18420
sS's'
I17
sS'm'
(lp15771
I1
aI1
aI5
aI1
aI17
aI27
aI53
aI13
aI31
aI403
aI1319
aI1381
aI1371
aI11693
aI18805
aI54683
aI30137
asS'd'
I7886
sa(dp15772
S'a'
I18434
sS's'
I17
sS'm'
(lp15773
I1
aI1
aI7
aI11
aI9
aI21
aI71
aI155
aI79
aI145
aI943
aI3891
aI641
aI3163
aI28493
aI14729
aI83391
asS'd'
I7887
sa(dp15774
S'a'
I18443
sS's'
I17
sS'm'
(lp15775
I1
aI3
aI3
aI13
aI3
aI53
aI21
aI189
aI245
aI803
aI1625
aI4005
aI1163
aI16033
aI5549
aI14301
aI115859
asS'd'
I7888
sa(dp15776
S'a'
I18446
sS's'
I17
sS'm'
(lp15777
I1
aI3
aI1
aI5
aI17
aI59
aI61
aI31
aI293
aI677
aI1753
aI1803
aI1671
aI14619
aI22361
aI61453
aI78203
asS'd'
I7889
sa(dp15778
S'a'
I18460
sS's'
I17
sS'm'
(lp15779
I1
aI3
aI3
aI1
aI5
aI51
aI99
aI231
aI175
aI97
aI1335
aI689
aI1913
aI6157
aI22757
aI52395
aI118347
asS'd'
I7890
sa(dp15780
S'a'
I18467
sS's'
I17
sS'm'
(lp15781
I1
aI3
aI3
aI7
aI25
aI11
aI113
aI19
aI289
aI507
aI1143
aI3437
aI7965
aI12551
aI27603
aI8423
aI46713
asS'd'
I7891
sa(dp15782
S'a'
I18482
sS's'
I17
sS'm'
(lp15783
I1
aI1
aI3
aI9
aI13
aI1
aI73
aI9
aI425
aI407
aI363
aI2915
aI4269
aI2903
aI9251
aI17733
aI80321
asS'd'
I7892
sa(dp15784
S'a'
I18484
sS's'
I17
sS'm'
(lp15785
I1
aI1
aI3
aI11
aI31
aI47
aI37
aI211
aI433
aI237
aI1069
aI1891
aI6175
aI9305
aI30385
aI2497
aI94775
asS'd'
I7893
sa(dp15786
S'a'
I18501
sS's'
I17
sS'm'
(lp15787
I1
aI1
aI3
aI1
aI23
aI5
aI113
aI103
aI427
aI587
aI1863
aI1863
aI679
aI2575
aI13059
aI16163
aI42289
asS'd'
I7894
sa(dp15788
S'a'
I18506
sS's'
I17
sS'm'
(lp15789
I1
aI1
aI5
aI3
aI7
aI17
aI45
aI33
aI209
aI609
aI1897
aI95
aI5123
aI2239
aI5483
aI60715
aI126497
asS'd'
I7895
sa(dp15790
S'a'
I18516
sS's'
I17
sS'm'
(lp15791
I1
aI1
aI5
aI11
aI1
aI55
aI67
aI223
aI41
aI967
aI337
aI2511
aI7879
aI1157
aI17635
aI64081
aI421
asS'd'
I7896
sa(dp15792
S'a'
I18519
sS's'
I17
sS'm'
(lp15793
I1
aI3
aI3
aI9
aI27
aI33
aI87
aI97
aI231
aI895
aI1337
aI829
aI47
aI8481
aI14059
aI57209
aI109159
asS'd'
I7897
sa(dp15794
S'a'
I18547
sS's'
I17
sS'm'
(lp15795
I1
aI3
aI7
aI1
aI25
aI5
aI41
aI161
aI393
aI523
aI1623
aI3761
aI1933
aI8319
aI17309
aI46717
aI97299
asS'd'
I7898
sa(dp15796
S'a'
I18569
sS's'
I17
sS'm'
(lp15797
I1
aI1
aI1
aI11
aI5
aI55
aI19
aI191
aI41
aI791
aI1611
aI59
aI2633
aI13873
aI25859
aI42879
aI54807
asS'd'
I7899
sa(dp15798
S'a'
I18575
sS's'
I17
sS'm'
(lp15799
I1
aI3
aI1
aI11
aI11
aI33
aI5
aI13
aI199
aI411
aI895
aI759
aI2735
aI16225
aI31469
aI24081
aI31857
asS'd'
I7900
sa(dp15800
S'a'
I18589
sS's'
I17
sS'm'
(lp15801
I1
aI1
aI7
aI13
aI27
aI57
aI21
aI191
aI389
aI977
aI1013
aI3493
aI6401
aI15957
aI23181
aI52461
aI41853
asS'd'
I7901
sa(dp15802
S'a'
I18590
sS's'
I17
sS'm'
(lp15803
I1
aI3
aI7
aI5
aI23
aI19
aI117
aI117
aI427
aI923
aI1347
aI3099
aI247
aI8879
aI5309
aI53277
aI100625
asS'd'
I7902
sa(dp15804
S'a'
I18605
sS's'
I17
sS'm'
(lp15805
I1
aI1
aI5
aI13
aI11
aI23
aI69
aI37
aI119
aI329
aI1935
aI2851
aI5127
aI6907
aI24651
aI11135
aI118287
asS'd'
I7903
sa(dp15806
S'a'
I18611
sS's'
I17
sS'm'
(lp15807
I1
aI1
aI3
aI15
aI23
aI1
aI69
aI107
aI253
aI771
aI1697
aI4035
aI3219
aI15011
aI6995
aI19255
aI39909
asS'd'
I7904
sa(dp15808
S'a'
I18625
sS's'
I17
sS'm'
(lp15809
I1
aI3
aI1
aI1
aI5
aI21
aI35
aI173
aI407
aI603
aI27
aI3589
aI2879
aI2755
aI17679
aI6145
aI95989
asS'd'
I7905
sa(dp15810
S'a'
I18652
sS's'
I17
sS'm'
(lp15811
I1
aI1
aI5
aI13
aI31
aI23
aI61
aI139
aI341
aI593
aI1673
aI4031
aI4809
aI1107
aI22657
aI29073
aI53401
asS'd'
I7906
sa(dp15812
S'a'
I18665
sS's'
I17
sS'm'
(lp15813
I1
aI3
aI1
aI15
aI13
aI37
aI39
aI61
aI443
aI417
aI1125
aI1529
aI1943
aI7317
aI2985
aI50285
aI107069
asS'd'
I7907
sa(dp15814
S'a'
I18673
sS's'
I17
sS'm'
(lp15815
I1
aI1
aI3
aI9
aI5
aI51
aI87
aI91
aI31
aI491
aI1455
aI1685
aI2579
aI6023
aI10989
aI64635
aI130147
asS'd'
I7908
sa(dp15816
S'a'
I18674
sS's'
I17
sS'm'
(lp15817
I1
aI3
aI5
aI5
aI31
aI23
aI15
aI163
aI357
aI161
aI1597
aI1571
aI5039
aI13213
aI32221
aI4405
aI88079
asS'd'
I7909
sa(dp15818
S'a'
I18683
sS's'
I17
sS'm'
(lp15819
I1
aI1
aI1
aI15
aI13
aI43
aI7
aI109
aI243
aI389
aI683
aI2671
aI8003
aI4187
aI6507
aI11171
aI116727
asS'd'
I7910
sa(dp15820
S'a'
I18688
sS's'
I17
sS'm'
(lp15821
I1
aI3
aI7
aI1
aI17
aI31
aI53
aI5
aI227
aI755
aI1755
aI2939
aI1789
aI8951
aI16777
aI30203
aI79005
asS'd'
I7911
sa(dp15822
S'a'
I18691
sS's'
I17
sS'm'
(lp15823
I1
aI3
aI3
aI9
aI27
aI5
aI111
aI241
aI89
aI333
aI371
aI1035
aI5719
aI2433
aI29343
aI50829
aI25131
asS'd'
I7912
sa(dp15824
S'a'
I18698
sS's'
I17
sS'm'
(lp15825
I1
aI1
aI3
aI13
aI7
aI37
aI125
aI69
aI79
aI397
aI1595
aI123
aI255
aI2257
aI10881
aI27085
aI99411
asS'd'
I7913
sa(dp15826
S'a'
I18717
sS's'
I17
sS'm'
(lp15827
I1
aI1
aI3
aI15
aI1
aI35
aI61
aI73
aI507
aI775
aI1631
aI2005
aI4277
aI8421
aI5669
aI39221
aI19053
asS'd'
I7914
sa(dp15828
S'a'
I18733
sS's'
I17
sS'm'
(lp15829
I1
aI1
aI3
aI7
aI15
aI17
aI65
aI157
aI19
aI997
aI861
aI1249
aI4059
aI7975
aI955
aI5833
aI97733
asS'd'
I7915
sa(dp15830
S'a'
I18754
sS's'
I17
sS'm'
(lp15831
I1
aI1
aI5
aI5
aI21
aI43
aI1
aI181
aI1
aI17
aI1337
aI3333
aI3761
aI12283
aI20941
aI231
aI30457
asS'd'
I7916
sa(dp15832
S'a'
I18759
sS's'
I17
sS'm'
(lp15833
I1
aI3
aI3
aI7
aI7
aI23
aI105
aI101
aI101
aI757
aI1407
aI565
aI2187
aI1529
aI29385
aI22847
aI57675
asS'd'
I7917
sa(dp15834
S'a'
I18760
sS's'
I17
sS'm'
(lp15835
I1
aI3
aI3
aI1
aI11
aI3
aI65
aI93
aI201
aI773
aI1037
aI1325
aI673
aI6625
aI2909
aI63435
aI114311
asS'd'
I7918
sa(dp15836
S'a'
I18771
sS's'
I17
sS'm'
(lp15837
I1
aI3
aI5
aI1
aI21
aI43
aI87
aI37
aI491
aI323
aI1093
aI2493
aI4755
aI7225
aI12037
aI9483
aI70351
asS'd'
I7919
sa(dp15838
S'a'
I18777
sS's'
I17
sS'm'
(lp15839
I1
aI1
aI7
aI9
aI23
aI39
aI59
aI117
aI103
aI645
aI1975
aI1177
aI55
aI325
aI23781
aI64365
aI94915
asS'd'
I7920
sa(dp15840
S'a'
I18796
sS's'
I17
sS'm'
(lp15841
I1
aI3
aI7
aI15
aI21
aI29
aI109
aI35
aI307
aI847
aI777
aI3457
aI7899
aI17
aI24065
aI10517
aI35651
asS'd'
I7921
sa(dp15842
S'a'
I18799
sS's'
I17
sS'm'
(lp15843
I1
aI1
aI7
aI9
aI25
aI35
aI49
aI131
aI377
aI429
aI1773
aI2107
aI6305
aI15209
aI9567
aI17685
aI5599
asS'd'
I7922
sa(dp15844
S'a'
I18807
sS's'
I17
sS'm'
(lp15845
I1
aI1
aI3
aI11
aI13
aI27
aI47
aI125
aI483
aI229
aI921
aI2733
aI2217
aI2615
aI24135
aI28759
aI109411
asS'd'
I7923
sa(dp15846
S'a'
I18813
sS's'
I17
sS'm'
(lp15847
I1
aI3
aI1
aI7
aI19
aI45
aI23
aI195
aI445
aI955
aI853
aI2877
aI6889
aI9507
aI2009
aI18747
aI50545
asS'd'
I7924
sa(dp15848
S'a'
I18817
sS's'
I17
sS'm'
(lp15849
I1
aI1
aI1
aI5
aI15
aI35
aI75
aI177
aI145
aI683
aI309
aI893
aI4999
aI827
aI26563
aI30819
aI111115
asS'd'
I7925
sa(dp15850
S'a'
I18820
sS's'
I17
sS'm'
(lp15851
I1
aI3
aI3
aI11
aI5
aI45
aI49
aI39
aI93
aI653
aI1053
aI2553
aI863
aI12185
aI30261
aI16459
aI121061
asS'd'
I7926
sa(dp15852
S'a'
I18827
sS's'
I17
sS'm'
(lp15853
I1
aI3
aI7
aI5
aI11
aI29
aI57
aI43
aI409
aI71
aI67
aI3537
aI263
aI13237
aI8825
aI58411
aI44629
asS'd'
I7927
sa(dp15854
S'a'
I18829
sS's'
I17
sS'm'
(lp15855
I1
aI1
aI5
aI13
aI1
aI37
aI23
aI171
aI13
aI309
aI239
aI2023
aI6233
aI8751
aI11371
aI5825
aI77355
asS'd'
I7928
sa(dp15856
S'a'
I18838
sS's'
I17
sS'm'
(lp15857
I1
aI3
aI1
aI13
aI5
aI1
aI47
aI217
aI369
aI609
aI453
aI879
aI4337
aI4441
aI8785
aI51963
aI53819
asS'd'
I7929
sa(dp15858
S'a'
I18842
sS's'
I17
sS'm'
(lp15859
I1
aI3
aI5
aI5
aI23
aI1
aI67
aI147
aI27
aI121
aI1369
aI569
aI1519
aI11585
aI18193
aI30889
aI78055
asS'd'
I7930
sa(dp15860
S'a'
I18848
sS's'
I17
sS'm'
(lp15861
I1
aI1
aI1
aI13
aI11
aI53
aI33
aI37
aI419
aI111
aI1649
aI2495
aI6105
aI12385
aI30865
aI3683
aI63813
asS'd'
I7931
sa(dp15862
S'a'
I18853
sS's'
I17
sS'm'
(lp15863
I1
aI3
aI3
aI5
aI17
aI35
aI107
aI235
aI471
aI735
aI1093
aI1007
aI567
aI173
aI9623
aI39533
aI56455
asS'd'
I7932
sa(dp15864
S'a'
I18885
sS's'
I17
sS'm'
(lp15865
I1
aI1
aI7
aI15
aI27
aI13
aI123
aI211
aI407
aI857
aI801
aI3951
aI8153
aI4427
aI15333
aI13217
aI92675
asS'd'
I7933
sa(dp15866
S'a'
I18890
sS's'
I17
sS'm'
(lp15867
I1
aI1
aI1
aI7
aI11
aI61
aI99
aI131
aI121
aI119
aI1483
aI1485
aI3521
aI13163
aI24899
aI56849
aI111943
asS'd'
I7934
sa(dp15868
S'a'
I18898
sS's'
I17
sS'm'
(lp15869
I1
aI3
aI3
aI1
aI19
aI1
aI29
aI139
aI127
aI557
aI1913
aI1487
aI1381
aI185
aI11195
aI52499
aI45059
asS'd'
I7935
sa(dp15870
S'a'
I18903
sS's'
I17
sS'm'
(lp15871
I1
aI3
aI7
aI11
aI5
aI29
aI95
aI111
aI235
aI55
aI1101
aI2631
aI1219
aI9867
aI22209
aI3095
aI56793
asS'd'
I7936
sa(dp15872
S'a'
I18910
sS's'
I17
sS'm'
(lp15873
I1
aI3
aI7
aI1
aI31
aI61
aI37
aI125
aI241
aI985
aI1079
aI1439
aI433
aI2779
aI8463
aI59985
aI39667
asS'd'
I7937
sa(dp15874
S'a'
I18913
sS's'
I17
sS'm'
(lp15875
I1
aI3
aI7
aI15
aI5
aI7
aI71
aI7
aI435
aI727
aI1611
aI135
aI1421
aI8329
aI29995
aI64243
aI58285
asS'd'
I7938
sa(dp15876
S'a'
I18931
sS's'
I17
sS'm'
(lp15877
I1
aI3
aI3
aI15
aI27
aI11
aI121
aI27
aI281
aI499
aI267
aI2651
aI7575
aI9499
aI5051
aI49475
aI79573
asS'd'
I7939
sa(dp15878
S'a'
I18934
sS's'
I17
sS'm'
(lp15879
I1
aI3
aI3
aI15
aI29
aI47
aI11
aI183
aI235
aI537
aI79
aI2467
aI3751
aI831
aI6725
aI52173
aI108645
asS'd'
I7940
sa(dp15880
S'a'
I18956
sS's'
I17
sS'm'
(lp15881
I1
aI3
aI5
aI13
aI23
aI31
aI23
aI19
aI477
aI511
aI727
aI183
aI5955
aI7613
aI31979
aI8441
aI39835
asS'd'
I7941
sa(dp15882
S'a'
I18961
sS's'
I17
sS'm'
(lp15883
I1
aI1
aI5
aI7
aI17
aI31
aI53
aI133
aI387
aI787
aI1573
aI89
aI1975
aI1825
aI19963
aI27203
aI124923
asS'd'
I7942
sa(dp15884
S'a'
I18968
sS's'
I17
sS'm'
(lp15885
I1
aI1
aI1
aI9
aI3
aI15
aI125
aI135
aI89
aI37
aI517
aI3931
aI2013
aI2143
aI25413
aI18421
aI6097
asS'd'
I7943
sa(dp15886
S'a'
I18978
sS's'
I17
sS'm'
(lp15887
I1
aI1
aI3
aI11
aI23
aI29
aI89
aI45
aI53
aI135
aI223
aI3523
aI7921
aI3271
aI1819
aI40931
aI65471
asS'd'
I7944
sa(dp15888
S'a'
I18983
sS's'
I17
sS'm'
(lp15889
I1
aI1
aI1
aI13
aI17
aI3
aI121
aI25
aI509
aI61
aI1009
aI2009
aI7813
aI8499
aI5807
aI63171
aI75991
asS'd'
I7945
sa(dp15890
S'a'
I18987
sS's'
I17
sS'm'
(lp15891
I1
aI3
aI5
aI13
aI15
aI35
aI37
aI45
aI161
aI683
aI1665
aI59
aI6297
aI9595
aI10193
aI46745
aI105411
asS'd'
I7946
sa(dp15892
S'a'
I18992
sS's'
I17
sS'm'
(lp15893
I1
aI3
aI1
aI7
aI21
aI19
aI85
aI107
aI3
aI845
aI673
aI1271
aI7581
aI15971
aI27085
aI35375
aI29027
asS'd'
I7947
sa(dp15894
S'a'
I18997
sS's'
I17
sS'm'
(lp15895
I1
aI3
aI3
aI9
aI5
aI17
aI79
aI137
aI123
aI809
aI583
aI3507
aI7559
aI2857
aI13911
aI57083
aI8595
asS'd'
I7948
sa(dp15896
S'a'
I19002
sS's'
I17
sS'm'
(lp15897
I1
aI1
aI7
aI5
aI31
aI29
aI77
aI33
aI439
aI787
aI1223
aI2471
aI5851
aI15891
aI27925
aI51661
aI82645
asS'd'
I7949
sa(dp15898
S'a'
I19004
sS's'
I17
sS'm'
(lp15899
I1
aI1
aI7
aI15
aI19
aI35
aI35
aI37
aI197
aI245
aI799
aI3971
aI277
aI11289
aI20111
aI13857
aI104571
asS'd'
I7950
sa(dp15900
S'a'
I19010
sS's'
I17
sS'm'
(lp15901
I1
aI3
aI5
aI15
aI19
aI3
aI65
aI17
aI201
aI1007
aI1665
aI107
aI2409
aI1469
aI23265
aI24547
aI8189
asS'd'
I7951
sa(dp15902
S'a'
I19012
sS's'
I17
sS'm'
(lp15903
I1
aI3
aI1
aI13
aI29
aI45
aI109
aI243
aI43
aI383
aI631
aI265
aI6671
aI15333
aI21931
aI30675
aI103077
asS'd'
I7952
sa(dp15904
S'a'
I19030
sS's'
I17
sS'm'
(lp15905
I1
aI1
aI5
aI1
aI25
aI25
aI77
aI189
aI109
aI777
aI1485
aI2265
aI1403
aI2627
aI13843
aI27263
aI14263
asS'd'
I7953
sa(dp15906
S'a'
I19043
sS's'
I17
sS'm'
(lp15907
I1
aI3
aI5
aI1
aI13
aI49
aI73
aI107
aI225
aI243
aI1253
aI3503
aI735
aI2605
aI27165
aI45467
aI93001
asS'd'
I7954
sa(dp15908
S'a'
I19049
sS's'
I17
sS'm'
(lp15909
I1
aI1
aI5
aI9
aI15
aI17
aI1
aI33
aI69
aI321
aI1375
aI3635
aI8131
aI6579
aI1225
aI38699
aI17447
asS'd'
I7955
sa(dp15910
S'a'
I19079
sS's'
I17
sS'm'
(lp15911
I1
aI3
aI5
aI3
aI25
aI49
aI15
aI149
aI483
aI37
aI1929
aI1219
aI5841
aI11975
aI805
aI31339
aI130971
asS'd'
I7956
sa(dp15912
S'a'
I19086
sS's'
I17
sS'm'
(lp15913
I1
aI3
aI3
aI3
aI15
aI29
aI3
aI143
aI291
aI593
aI1769
aI3603
aI1693
aI151
aI27701
aI9015
aI25847
asS'd'
I7957
sa(dp15914
S'a'
I19100
sS's'
I17
sS'm'
(lp15915
I1
aI3
aI1
aI11
aI5
aI35
aI55
aI127
aI137
aI71
aI967
aI2501
aI1023
aI2061
aI31387
aI44011
aI130121
asS'd'
I7958
sa(dp15916
S'a'
I19103
sS's'
I17
sS'm'
(lp15917
I1
aI1
aI7
aI1
aI29
aI13
aI93
aI41
aI125
aI263
aI521
aI2633
aI4361
aI12153
aI30647
aI55883
aI65185
asS'd'
I7959
sa(dp15918
S'a'
I19107
sS's'
I17
sS'm'
(lp15919
I1
aI3
aI7
aI9
aI23
aI19
aI61
aI197
aI139
aI463
aI1867
aI3627
aI5185
aI8251
aI26977
aI48027
aI66301
asS'd'
I7960
sa(dp15920
S'a'
I19109
sS's'
I17
sS'm'
(lp15921
I1
aI3
aI3
aI7
aI27
aI53
aI25
aI137
aI175
aI155
aI1843
aI2253
aI4797
aI4989
aI32613
aI55779
aI91625
asS'd'
I7961
sa(dp15922
S'a'
I19113
sS's'
I17
sS'm'
(lp15923
I1
aI3
aI3
aI11
aI1
aI5
aI21
aI233
aI295
aI675
aI47
aI2995
aI8075
aI8201
aI3845
aI23925
aI82559
asS'd'
I7962
sa(dp15924
S'a'
I19116
sS's'
I17
sS'm'
(lp15925
I1
aI1
aI3
aI7
aI31
aI53
aI93
aI21
aI307
aI709
aI9
aI3061
aI3935
aI11009
aI13411
aI3657
aI30251
asS'd'
I7963
sa(dp15926
S'a'
I19127
sS's'
I17
sS'm'
(lp15927
I1
aI3
aI7
aI13
aI13
aI25
aI51
aI205
aI391
aI897
aI275
aI333
aI6831
aI9383
aI16101
aI14301
aI99101
asS'd'
I7964
sa(dp15928
S'a'
I19134
sS's'
I17
sS'm'
(lp15929
I1
aI1
aI5
aI15
aI17
aI47
aI119
aI7
aI189
aI765
aI753
aI2909
aI3373
aI2379
aI20331
aI61535
aI51345
asS'd'
I7965
sa(dp15930
S'a'
I19141
sS's'
I17
sS'm'
(lp15931
I1
aI1
aI3
aI1
aI27
aI43
aI9
aI185
aI9
aI197
aI1179
aI67
aI7689
aI9679
aI29683
aI29905
aI29393
asS'd'
I7966
sa(dp15932
S'a'
I19179
sS's'
I17
sS'm'
(lp15933
I1
aI1
aI5
aI5
aI31
aI55
aI69
aI9
aI477
aI91
aI1705
aI1877
aI5463
aI15401
aI13449
aI27541
aI125521
asS'd'
I7967
sa(dp15934
S'a'
I19193
sS's'
I17
sS'm'
(lp15935
I1
aI1
aI7
aI15
aI15
aI33
aI11
aI233
aI69
aI771
aI845
aI2715
aI5293
aI10351
aI19557
aI15319
aI36857
asS'd'
I7968
sa(dp15936
S'a'
I19194
sS's'
I17
sS'm'
(lp15937
I1
aI3
aI7
aI7
aI15
aI9
aI123
aI47
aI165
aI739
aI361
aI1675
aI2743
aI8021
aI10241
aI48275
aI51935
asS'd'
I7969
sa(dp15938
S'a'
I19201
sS's'
I17
sS'm'
(lp15939
I1
aI1
aI5
aI1
aI9
aI25
aI99
aI83
aI487
aI627
aI343
aI3233
aI6697
aI13197
aI19771
aI38337
aI89139
asS'd'
I7970
sa(dp15940
S'a'
I19208
sS's'
I17
sS'm'
(lp15941
I1
aI3
aI7
aI13
aI1
aI31
aI15
aI63
aI463
aI621
aI935
aI1093
aI6043
aI14051
aI13665
aI43413
aI104893
asS'd'
I7971
sa(dp15942
S'a'
I19214
sS's'
I17
sS'm'
(lp15943
I1
aI1
aI1
aI3
aI27
aI1
aI47
aI151
aI127
aI357
aI689
aI3263
aI141
aI4459
aI9847
aI205
aI88493
asS'd'
I7972
sa(dp15944
S'a'
I19225
sS's'
I17
sS'm'
(lp15945
I1
aI3
aI7
aI15
aI29
aI13
aI41
aI113
aI495
aI421
aI195
aI197
aI6857
aI10555
aI22861
aI30229
aI31707
asS'd'
I7973
sa(dp15946
S'a'
I19226
sS's'
I17
sS'm'
(lp15947
I1
aI3
aI5
aI11
aI11
aI1
aI89
aI227
aI425
aI623
aI1605
aI1901
aI7933
aI7211
aI16301
aI3403
aI59651
asS'd'
I7974
sa(dp15948
S'a'
I19235
sS's'
I17
sS'm'
(lp15949
I1
aI1
aI3
aI3
aI27
aI41
aI37
aI89
aI395
aI903
aI1641
aI2739
aI5599
aI11371
aI8683
aI61125
aI105231
asS'd'
I7975
sa(dp15950
S'a'
I19242
sS's'
I17
sS'm'
(lp15951
I1
aI3
aI7
aI9
aI1
aI17
aI51
aI233
aI507
aI783
aI459
aI1187
aI7281
aI15809
aI27637
aI6067
aI125877
asS'd'
I7976
sa(dp15952
S'a'
I19264
sS's'
I17
sS'm'
(lp15953
I1
aI3
aI1
aI3
aI13
aI57
aI5
aI199
aI261
aI357
aI1255
aI1849
aI8013
aI10313
aI9375
aI1271
aI64117
asS'd'
I7977
sa(dp15954
S'a'
I19267
sS's'
I17
sS'm'
(lp15955
I1
aI3
aI1
aI11
aI9
aI59
aI55
aI95
aI401
aI423
aI1657
aI513
aI3565
aI12957
aI19243
aI53027
aI11323
asS'd'
I7978
sa(dp15956
S'a'
I19293
sS's'
I17
sS'm'
(lp15957
I1
aI3
aI7
aI13
aI27
aI35
aI121
aI215
aI397
aI991
aI191
aI3443
aI3829
aI6107
aI5381
aI48497
aI107997
asS'd'
I7979
sa(dp15958
S'a'
I19309
sS's'
I17
sS'm'
(lp15959
I1
aI1
aI5
aI5
aI19
aI15
aI21
aI53
aI165
aI835
aI1599
aI3245
aI5609
aI5991
aI18141
aI28075
aI102809
asS'd'
I7980
sa(dp15960
S'a'
I19318
sS's'
I17
sS'm'
(lp15961
I1
aI3
aI5
aI9
aI25
aI21
aI71
aI15
aI453
aI475
aI915
aI3097
aI5869
aI699
aI13883
aI34919
aI127211
asS'd'
I7981
sa(dp15962
S'a'
I19324
sS's'
I17
sS'm'
(lp15963
I1
aI1
aI3
aI7
aI21
aI53
aI27
aI207
aI373
aI703
aI593
aI17
aI6991
aI15013
aI10125
aI34801
aI129245
asS'd'
I7982
sa(dp15964
S'a'
I19337
sS's'
I17
sS'm'
(lp15965
I1
aI3
aI3
aI13
aI17
aI9
aI89
aI95
aI291
aI681
aI1415
aI2323
aI2885
aI11381
aI7703
aI3691
aI51505
asS'd'
I7983
sa(dp15966
S'a'
I19340
sS's'
I17
sS'm'
(lp15967
I1
aI1
aI1
aI11
aI15
aI63
aI55
aI153
aI373
aI665
aI983
aI3987
aI5997
aI6873
aI27031
aI65449
aI22021
asS'd'
I7984
sa(dp15968
S'a'
I19345
sS's'
I17
sS'm'
(lp15969
I1
aI1
aI5
aI5
aI1
aI55
aI119
aI61
aI159
aI889
aI225
aI709
aI1879
aI2521
aI69
aI7815
aI18733
asS'd'
I7985
sa(dp15970
S'a'
I19346
sS's'
I17
sS'm'
(lp15971
I1
aI3
aI5
aI11
aI23
aI53
aI23
aI61
aI71
aI993
aI633
aI1829
aI3465
aI12465
aI30205
aI40723
aI74499
asS'd'
I7986
sa(dp15972
S'a'
I19352
sS's'
I17
sS'm'
(lp15973
I1
aI3
aI3
aI1
aI17
aI37
aI19
aI43
aI55
aI133
aI1171
aI3101
aI3963
aI5177
aI24791
aI7255
aI10263
asS'd'
I7987
sa(dp15974
S'a'
I19364
sS's'
I17
sS'm'
(lp15975
I1
aI3
aI7
aI1
aI21
aI13
aI21
aI87
aI237
aI629
aI1167
aI3699
aI597
aI6251
aI11545
aI34429
aI104393
asS'd'
I7988
sa(dp15976
S'a'
I19382
sS's'
I17
sS'm'
(lp15977
I1
aI1
aI7
aI1
aI11
aI53
aI105
aI111
aI463
aI869
aI549
aI2423
aI17
aI917
aI879
aI49367
aI72235
asS'd'
I7989
sa(dp15978
S'a'
I19391
sS's'
I17
sS'm'
(lp15979
I1
aI1
aI5
aI15
aI17
aI51
aI69
aI55
aI309
aI867
aI257
aI573
aI4821
aI5245
aI28033
aI61801
aI18253
asS'd'
I7990
sa(dp15980
S'a'
I19405
sS's'
I17
sS'm'
(lp15981
I1
aI1
aI5
aI3
aI1
aI23
aI103
aI241
aI13
aI267
aI21
aI1751
aI6637
aI12537
aI26741
aI40651
aI94493
asS'd'
I7991
sa(dp15982
S'a'
I19411
sS's'
I17
sS'm'
(lp15983
I1
aI3
aI3
aI13
aI25
aI35
aI21
aI83
aI337
aI363
aI1111
aI1865
aI7889
aI985
aI465
aI40287
aI64469
asS'd'
I7992
sa(dp15984
S'a'
I19439
sS's'
I17
sS'm'
(lp15985
I1
aI1
aI7
aI5
aI27
aI1
aI99
aI95
aI209
aI211
aI1445
aI1577
aI6097
aI13813
aI22463
aI64395
aI106383
asS'd'
I7993
sa(dp15986
S'a'
I19442
sS's'
I17
sS'm'
(lp15987
I1
aI3
aI1
aI15
aI1
aI45
aI77
aI247
aI273
aI1023
aI1377
aI1989
aI5787
aI15267
aI24363
aI42717
aI125617
asS'd'
I7994
sa(dp15988
S'a'
I19444
sS's'
I17
sS'm'
(lp15989
I1
aI1
aI1
aI3
aI9
aI49
aI79
aI171
aI427
aI439
aI1725
aI3179
aI6263
aI12437
aI31353
aI22077
aI94455
asS'd'
I7995
sa(dp15990
S'a'
I19451
sS's'
I17
sS'm'
(lp15991
I1
aI3
aI5
aI11
aI11
aI45
aI57
aI97
aI409
aI935
aI967
aI2509
aI3431
aI5707
aI19473
aI15853
aI129059
asS'd'
I7996
sa(dp15992
S'a'
I19465
sS's'
I17
sS'm'
(lp15993
I1
aI1
aI7
aI5
aI7
aI21
aI105
aI29
aI359
aI145
aI1691
aI131
aI6721
aI10971
aI16173
aI38193
aI37091
asS'd'
I7997
sa(dp15994
S'a'
I19471
sS's'
I17
sS'm'
(lp15995
I1
aI1
aI1
aI15
aI15
aI35
aI5
aI185
aI455
aI507
aI681
aI3355
aI2091
aI3437
aI27231
aI28527
aI5383
asS'd'
I7998
sa(dp15996
S'a'
I19474
sS's'
I17
sS'm'
(lp15997
I1
aI3
aI5
aI3
aI7
aI29
aI33
aI127
aI57
aI495
aI1069
aI3635
aI7461
aI9861
aI18757
aI39039
aI92421
asS'd'
I7999
sa(dp15998
S'a'
I19476
sS's'
I17
sS'm'
(lp15999
I1
aI3
aI5
aI5
aI11
aI31
aI51
aI59
aI413
aI417
aI1577
aI2837
aI5229
aI4501
aI18645
aI37613
aI31325
asS'd'
I8000
sa(dp16000
S'a'
I19479
sS's'
I17
sS'm'
(lp16001
I1
aI1
aI5
aI13
aI15
aI61
aI17
aI247
aI413
aI817
aI907
aI2249
aI3901
aI11275
aI7469
aI33403
aI30629
asS'd'
I8001
sa(dp16002
S'a'
I19485
sS's'
I17
sS'm'
(lp16003
I1
aI3
aI5
aI7
aI31
aI7
aI109
aI177
aI277
aI75
aI449
aI3029
aI7135
aI427
aI26641
aI43157
aI47671
asS'd'
I8002
sa(dp16004
S'a'
I19489
sS's'
I17
sS'm'
(lp16005
I1
aI3
aI7
aI13
aI29
aI63
aI21
aI187
aI471
aI289
aI835
aI3885
aI6111
aI8721
aI9841
aI24017
aI18673
asS'd'
I8003
sa(dp16006
S'a'
I19496
sS's'
I17
sS'm'
(lp16007
I1
aI1
aI5
aI13
aI25
aI37
aI15
aI35
aI227
aI623
aI47
aI189
aI3443
aI1911
aI8579
aI50911
aI10895
asS'd'
I8004
sa(dp16008
S'a'
I19507
sS's'
I17
sS'm'
(lp16009
I1
aI3
aI1
aI1
aI29
aI53
aI89
aI101
aI251
aI203
aI821
aI2485
aI633
aI7943
aI27563
aI58735
aI72057
asS'd'
I8005
sa(dp16010
S'a'
I19513
sS's'
I17
sS'm'
(lp16011
I1
aI3
aI7
aI9
aI23
aI61
aI121
aI199
aI19
aI165
aI131
aI1373
aI637
aI7307
aI7109
aI42475
aI126669
asS'd'
I8006
sa(dp16012
S'a'
I19514
sS's'
I17
sS'm'
(lp16013
I1
aI3
aI7
aI13
aI7
aI5
aI125
aI173
aI365
aI65
aI565
aI751
aI3343
aI13421
aI6177
aI39095
aI97375
asS'd'
I8007
sa(dp16014
S'a'
I19521
sS's'
I17
sS'm'
(lp16015
I1
aI1
aI7
aI3
aI1
aI59
aI65
aI39
aI307
aI793
aI887
aI3291
aI3405
aI4497
aI19351
aI1821
aI67861
asS'd'
I8008
sa(dp16016
S'a'
I19524
sS's'
I17
sS'm'
(lp16017
I1
aI1
aI1
aI3
aI19
aI9
aI101
aI183
aI163
aI819
aI769
aI49
aI5293
aI3715
aI4055
aI32403
aI90763
asS'd'
I8009
sa(dp16018
S'a'
I19546
sS's'
I17
sS'm'
(lp16019
I1
aI3
aI3
aI1
aI27
aI31
aI21
aI123
aI457
aI1021
aI1791
aI2217
aI6171
aI11445
aI15605
aI59945
aI19663
asS'd'
I8010
sa(dp16020
S'a'
I19552
sS's'
I17
sS'm'
(lp16021
I1
aI1
aI1
aI9
aI13
aI53
aI61
aI201
aI457
aI111
aI1217
aI377
aI5871
aI4591
aI16379
aI42817
aI129807
asS'd'
I8011
sa(dp16022
S'a'
I19555
sS's'
I17
sS'm'
(lp16023
I1
aI3
aI1
aI5
aI23
aI37
aI25
aI7
aI125
aI651
aI349
aI3727
aI1487
aI5103
aI31407
aI40215
aI16065
asS'd'
I8012
sa(dp16024
S'a'
I19557
sS's'
I17
sS'm'
(lp16025
I1
aI1
aI3
aI11
aI19
aI29
aI1
aI193
aI13
aI287
aI331
aI985
aI5391
aI7307
aI28075
aI9939
aI84999
asS'd'
I8013
sa(dp16026
S'a'
I19575
sS's'
I17
sS'm'
(lp16027
I1
aI1
aI1
aI11
aI21
aI37
aI117
aI241
aI229
aI957
aI2019
aI819
aI459
aI6185
aI21533
aI64725
aI24709
asS'd'
I8014
sa(dp16028
S'a'
I19579
sS's'
I17
sS'm'
(lp16029
I1
aI3
aI5
aI13
aI11
aI25
aI107
aI245
aI175
aI519
aI629
aI537
aI2227
aI15123
aI10619
aI32611
aI118697
asS'd'
I8015
sa(dp16030
S'a'
I19591
sS's'
I17
sS'm'
(lp16031
I1
aI3
aI1
aI11
aI5
aI53
aI119
aI253
aI489
aI181
aI1365
aI3465
aI1323
aI949
aI3657
aI2467
aI38545
asS'd'
I8016
sa(dp16032
S'a'
I19595
sS's'
I17
sS'm'
(lp16033
I1
aI1
aI3
aI9
aI27
aI17
aI109
aI19
aI297
aI541
aI89
aI3021
aI761
aI5577
aI907
aI21405
aI128029
asS'd'
I8017
sa(dp16034
S'a'
I19605
sS's'
I17
sS'm'
(lp16035
I1
aI1
aI3
aI9
aI31
aI9
aI61
aI149
aI267
aI707
aI671
aI2733
aI1247
aI14623
aI6395
aI42579
aI30845
asS'd'
I8018
sa(dp16036
S'a'
I19615
sS's'
I17
sS'm'
(lp16037
I1
aI1
aI7
aI7
aI25
aI29
aI63
aI41
aI309
aI275
aI2019
aI1373
aI1003
aI13891
aI16571
aI16209
aI30115
asS'd'
I8019
sa(dp16038
S'a'
I19616
sS's'
I17
sS'm'
(lp16039
I1
aI3
aI7
aI1
aI5
aI21
aI53
aI97
aI475
aI799
aI1963
aI1181
aI4187
aI8767
aI24779
aI10403
aI98349
asS'd'
I8020
sa(dp16040
S'a'
I19626
sS's'
I17
sS'm'
(lp16041
I1
aI3
aI3
aI13
aI19
aI9
aI125
aI227
aI347
aI535
aI353
aI3087
aI769
aI9049
aI20145
aI27433
aI23105
asS'd'
I8021
sa(dp16042
S'a'
I19631
sS's'
I17
sS'm'
(lp16043
I1
aI1
aI1
aI15
aI7
aI61
aI51
aI113
aI403
aI501
aI1767
aI2785
aI7151
aI14517
aI17533
aI24913
aI7121
asS'd'
I8022
sa(dp16044
S'a'
I19634
sS's'
I17
sS'm'
(lp16045
I1
aI1
aI1
aI9
aI7
aI21
aI27
aI169
aI425
aI567
aI31
aI35
aI7859
aI929
aI6545
aI33983
aI13227
asS'd'
I8023
sa(dp16046
S'a'
I19640
sS's'
I17
sS'm'
(lp16047
I1
aI1
aI5
aI15
aI11
aI15
aI69
aI33
aI127
aI1005
aI1947
aI989
aI6333
aI15587
aI18523
aI53547
aI115613
asS'd'
I8024
sa(dp16048
S'a'
I19645
sS's'
I17
sS'm'
(lp16049
I1
aI3
aI5
aI3
aI1
aI55
aI7
aI99
aI213
aI737
aI363
aI3167
aI3949
aI3723
aI15777
aI23207
aI22901
asS'd'
I8025
sa(dp16050
S'a'
I19678
sS's'
I17
sS'm'
(lp16051
I1
aI1
aI1
aI9
aI9
aI29
aI121
aI85
aI467
aI811
aI1
aI3543
aI6259
aI4477
aI31371
aI27633
aI22995
asS'd'
I8026
sa(dp16052
S'a'
I19681
sS's'
I17
sS'm'
(lp16053
I1
aI3
aI5
aI3
aI11
aI21
aI95
aI19
aI55
aI71
aI803
aI3655
aI3749
aI5113
aI13611
aI38097
aI20943
asS'd'
I8027
sa(dp16054
S'a'
I19682
sS's'
I17
sS'm'
(lp16055
I1
aI3
aI3
aI11
aI19
aI25
aI127
aI105
aI447
aI499
aI485
aI881
aI2649
aI10297
aI22283
aI18305
aI128919
asS'd'
I8028
sa(dp16056
S'a'
I19706
sS's'
I17
sS'm'
(lp16057
I1
aI3
aI7
aI1
aI11
aI45
aI21
aI87
aI481
aI645
aI815
aI793
aI5763
aI3945
aI14379
aI19623
aI103199
asS'd'
I8029
sa(dp16058
S'a'
I19708
sS's'
I17
sS'm'
(lp16059
I1
aI3
aI5
aI1
aI3
aI45
aI39
aI229
aI359
aI151
aI1079
aI3955
aI2107
aI9593
aI6701
aI2811
aI55215
asS'd'
I8030
sa(dp16060
S'a'
I19713
sS's'
I17
sS'm'
(lp16061
I1
aI3
aI7
aI7
aI27
aI59
aI69
aI211
aI373
aI643
aI977
aI3545
aI2647
aI10473
aI27919
aI10719
aI24823
asS'd'
I8031
sa(dp16062
S'a'
I19714
sS's'
I17
sS'm'
(lp16063
I1
aI1
aI3
aI11
aI7
aI27
aI117
aI21
aI59
aI679
aI969
aI3813
aI2701
aI7363
aI17525
aI31229
aI100665
asS'd'
I8032
sa(dp16064
S'a'
I19720
sS's'
I17
sS'm'
(lp16065
I1
aI3
aI3
aI5
aI29
aI53
aI113
aI141
aI197
aI991
aI81
aI2701
aI6831
aI7949
aI16569
aI44185
aI29631
asS'd'
I8033
sa(dp16066
S'a'
I19725
sS's'
I17
sS'm'
(lp16067
I1
aI1
aI1
aI3
aI1
aI31
aI9
aI101
aI347
aI585
aI577
aI2529
aI7461
aI11921
aI29475
aI34505
aI74911
asS'd'
I8034
sa(dp16068
S'a'
I19743
sS's'
I17
sS'm'
(lp16069
I1
aI1
aI1
aI15
aI25
aI19
aI95
aI37
aI93
aI755
aI1891
aI2309
aI623
aI13503
aI5811
aI45863
aI106501
asS'd'
I8035
sa(dp16070
S'a'
I19753
sS's'
I17
sS'm'
(lp16071
I1
aI1
aI5
aI3
aI15
aI23
aI51
aI225
aI87
aI679
aI1225
aI4015
aI3971
aI163
aI3185
aI12921
aI51267
asS'd'
I8036
sa(dp16072
S'a'
I19756
sS's'
I17
sS'm'
(lp16073
I1
aI1
aI5
aI1
aI1
aI37
aI105
aI181
aI379
aI657
aI571
aI2775
aI5905
aI8391
aI32069
aI18713
aI125833
asS'd'
I8037
sa(dp16074
S'a'
I19759
sS's'
I17
sS'm'
(lp16075
I1
aI1
aI7
aI11
aI11
aI19
aI109
aI125
aI371
aI321
aI629
aI2165
aI2861
aI7883
aI15503
aI37679
aI33057
asS'd'
I8038
sa(dp16076
S'a'
I19762
sS's'
I17
sS'm'
(lp16077
I1
aI1
aI3
aI5
aI7
aI5
aI21
aI5
aI169
aI321
aI1145
aI2243
aI6143
aI2537
aI4429
aI56493
aI39391
asS'd'
I8039
sa(dp16078
S'a'
I19776
sS's'
I17
sS'm'
(lp16079
I1
aI3
aI5
aI5
aI31
aI7
aI15
aI175
aI441
aI663
aI921
aI3113
aI2261
aI13033
aI19135
aI28657
aI92225
asS'd'
I8040
sa(dp16080
S'a'
I19786
sS's'
I17
sS'm'
(lp16081
I1
aI3
aI1
aI11
aI13
aI9
aI25
aI57
aI297
aI3
aI1561
aI825
aI2803
aI11327
aI2699
aI28631
aI57277
asS'd'
I8041
sa(dp16082
S'a'
I19799
sS's'
I17
sS'm'
(lp16083
I1
aI1
aI3
aI9
aI15
aI25
aI81
aI197
aI345
aI341
aI1557
aI1375
aI2509
aI11949
aI30201
aI6807
aI95199
asS'd'
I8042
sa(dp16084
S'a'
I19800
sS's'
I17
sS'm'
(lp16085
I1
aI3
aI5
aI3
aI15
aI23
aI91
aI147
aI277
aI59
aI495
aI1423
aI1775
aI12065
aI8401
aI22639
aI111199
asS'd'
I8043
sa(dp16086
S'a'
I19803
sS's'
I17
sS'm'
(lp16087
I1
aI1
aI5
aI1
aI1
aI59
aI35
aI255
aI229
aI293
aI187
aI2663
aI3967
aI6493
aI20103
aI36703
aI108939
asS'd'
I8044
sa(dp16088
S'a'
I19806
sS's'
I17
sS'm'
(lp16089
I1
aI3
aI7
aI11
aI15
aI1
aI23
aI39
aI27
aI281
aI11
aI3119
aI2791
aI1691
aI16521
aI39715
aI32145
asS'd'
I8045
sa(dp16090
S'a'
I19815
sS's'
I17
sS'm'
(lp16091
I1
aI3
aI5
aI5
aI9
aI53
aI43
aI49
aI107
aI1015
aI431
aI3017
aI3317
aI9655
aI19193
aI45621
aI56861
asS'd'
I8046
sa(dp16092
S'a'
I19816
sS's'
I17
sS'm'
(lp16093
I1
aI3
aI1
aI15
aI27
aI63
aI127
aI31
aI21
aI245
aI1503
aI3339
aI6375
aI5411
aI12369
aI35973
aI9473
asS'd'
I8047
sa(dp16094
S'a'
I19857
sS's'
I17
sS'm'
(lp16095
I1
aI1
aI3
aI13
aI31
aI61
aI19
aI99
aI25
aI593
aI539
aI1807
aI673
aI12339
aI23567
aI22005
aI130341
asS'd'
I8048
sa(dp16096
S'a'
I19860
sS's'
I17
sS'm'
(lp16097
I1
aI1
aI5
aI3
aI3
aI3
aI13
aI183
aI255
aI41
aI641
aI581
aI6509
aI1891
aI19195
aI28277
aI51725
asS'd'
I8049
sa(dp16098
S'a'
I19874
sS's'
I17
sS'm'
(lp16099
I1
aI1
aI3
aI5
aI3
aI59
aI17
aI227
aI9
aI345
aI1303
aI1535
aI3089
aI2653
aI20647
aI63159
aI124677
asS'd'
I8050
sa(dp16100
S'a'
I19883
sS's'
I17
sS'm'
(lp16101
I1
aI3
aI1
aI11
aI25
aI19
aI117
aI29
aI221
aI461
aI1285
aI1427
aI7183
aI3051
aI10839
aI47491
aI92613
asS'd'
I8051
sa(dp16102
S'a'
I19885
sS's'
I17
sS'm'
(lp16103
I1
aI1
aI3
aI5
aI27
aI19
aI1
aI235
aI51
aI215
aI783
aI2325
aI1191
aI4679
aI14365
aI35479
aI65083
asS'd'
I8052
sa(dp16104
S'a'
I19886
sS's'
I17
sS'm'
(lp16105
I1
aI3
aI3
aI5
aI27
aI17
aI79
aI185
aI259
aI369
aI1399
aI1029
aI2219
aI10975
aI30487
aI15247
aI39789
asS'd'
I8053
sa(dp16106
S'a'
I19893
sS's'
I17
sS'm'
(lp16107
I1
aI3
aI3
aI1
aI13
aI13
aI59
aI119
aI249
aI471
aI1433
aI1165
aI5345
aI4431
aI375
aI64999
aI85577
asS'd'
I8054
sa(dp16108
S'a'
I19932
sS's'
I17
sS'm'
(lp16109
I1
aI1
aI1
aI3
aI1
aI19
aI13
aI243
aI35
aI675
aI321
aI3625
aI7835
aI6403
aI651
aI48283
aI91819
asS'd'
I8055
sa(dp16110
S'a'
I19960
sS's'
I17
sS'm'
(lp16111
I1
aI3
aI3
aI1
aI27
aI13
aI73
aI159
aI145
aI59
aI287
aI2419
aI8115
aI7923
aI18933
aI36109
aI123879
asS'd'
I8056
sa(dp16112
S'a'
I19972
sS's'
I17
sS'm'
(lp16113
I1
aI3
aI1
aI7
aI21
aI1
aI83
aI245
aI477
aI623
aI391
aI129
aI6897
aI3447
aI11109
aI17017
aI68277
asS'd'
I8057
sa(dp16114
S'a'
I19975
sS's'
I17
sS'm'
(lp16115
I1
aI1
aI3
aI11
aI13
aI43
aI119
aI93
aI99
aI393
aI1219
aI995
aI1881
aI7929
aI4337
aI33579
aI103211
asS'd'
I8058
sa(dp16116
S'a'
I20005
sS's'
I17
sS'm'
(lp16117
I1
aI1
aI7
aI7
aI31
aI5
aI39
aI25
aI119
aI819
aI409
aI2395
aI151
aI12763
aI28265
aI28909
aI35685
asS'd'
I8059
sa(dp16118
S'a'
I20009
sS's'
I17
sS'm'
(lp16119
I1
aI1
aI1
aI1
aI3
aI13
aI59
aI205
aI19
aI843
aI1691
aI955
aI1859
aI1791
aI22083
aI37843
aI63615
asS'd'
I8060
sa(dp16120
S'a'
I20010
sS's'
I17
sS'm'
(lp16121
I1
aI1
aI1
aI3
aI11
aI63
aI41
aI243
aI103
aI875
aI1337
aI3731
aI6317
aI12951
aI31743
aI56935
aI55975
asS'd'
I8061
sa(dp16122
S'a'
I20012
sS's'
I17
sS'm'
(lp16123
I1
aI1
aI3
aI13
aI19
aI11
aI51
aI97
aI469
aI279
aI1621
aI3521
aI853
aI11849
aI3331
aI27907
aI119081
asS'd'
I8062
sa(dp16124
S'a'
I20023
sS's'
I17
sS'm'
(lp16125
I1
aI1
aI5
aI1
aI23
aI55
aI9
aI141
aI449
aI41
aI167
aI2441
aI6783
aI2785
aI3547
aI35379
aI74973
asS'd'
I8063
sa(dp16126
S'a'
I20024
sS's'
I17
sS'm'
(lp16127
I1
aI1
aI5
aI3
aI15
aI55
aI13
aI75
aI107
aI153
aI1841
aI3991
aI3229
aI6523
aI18541
aI21571
aI31539
asS'd'
I8064
sa(dp16128
S'a'
I20030
sS's'
I17
sS'm'
(lp16129
I1
aI3
aI1
aI1
aI27
aI37
aI35
aI201
aI401
aI867
aI1861
aI1783
aI6255
aI14001
aI29543
aI25843
aI39719
asS'd'
I8065
sa(dp16130
S'a'
I20038
sS's'
I17
sS'm'
(lp16131
I1
aI1
aI7
aI15
aI3
aI43
aI113
aI173
aI297
aI457
aI1369
aI4053
aI5033
aI5513
aI27387
aI14725
aI79937
asS'd'
I8066
sa(dp16132
S'a'
I20041
sS's'
I17
sS'm'
(lp16133
I1
aI1
aI5
aI13
aI5
aI27
aI109
aI93
aI187
aI833
aI1551
aI2899
aI1681
aI6979
aI1289
aI3507
aI66499
asS'd'
I8067
sa(dp16134
S'a'
I20055
sS's'
I17
sS'm'
(lp16135
I1
aI1
aI3
aI11
aI11
aI47
aI121
aI29
aI129
aI807
aI2037
aI1527
aI6083
aI14803
aI10669
aI46047
aI70241
asS'd'
I8068
sa(dp16136
S'a'
I20059
sS's'
I17
sS'm'
(lp16137
I1
aI3
aI1
aI9
aI29
aI3
aI19
aI191
aI461
aI527
aI1389
aI3359
aI81
aI6773
aI12185
aI49207
aI19091
asS'd'
I8069
sa(dp16138
S'a'
I20061
sS's'
I17
sS'm'
(lp16139
I1
aI3
aI7
aI7
aI5
aI47
aI33
aI27
aI445
aI1
aI149
aI343
aI4827
aI91
aI29233
aI37775
aI89321
asS'd'
I8070
sa(dp16140
S'a'
I20080
sS's'
I17
sS'm'
(lp16141
I1
aI3
aI5
aI1
aI13
aI55
aI107
aI99
aI259
aI591
aI983
aI3863
aI1231
aI3457
aI29645
aI10709
aI16543
asS'd'
I8071
sa(dp16142
S'a'
I20086
sS's'
I17
sS'm'
(lp16143
I1
aI3
aI7
aI9
aI29
aI5
aI9
aI165
aI337
aI187
aI219
aI97
aI6511
aI193
aI23947
aI36329
aI35317
asS'd'
I8072
sa(dp16144
S'a'
I20089
sS's'
I17
sS'm'
(lp16145
I1
aI3
aI7
aI1
aI31
aI25
aI5
aI175
aI409
aI1021
aI1873
aI289
aI7143
aI15341
aI31501
aI25707
aI106453
asS'd'
I8073
sa(dp16146
S'a'
I20095
sS's'
I17
sS'm'
(lp16147
I1
aI3
aI7
aI7
aI27
aI1
aI15
aI221
aI341
aI987
aI1739
aI1183
aI8139
aI11081
aI29721
aI42991
aI72805
asS'd'
I8074
sa(dp16148
S'a'
I20111
sS's'
I17
sS'm'
(lp16149
I1
aI1
aI1
aI9
aI11
aI1
aI13
aI17
aI501
aI543
aI1485
aI987
aI1861
aI8527
aI1621
aI30461
aI23057
asS'd'
I8075
sa(dp16150
S'a'
I20116
sS's'
I17
sS'm'
(lp16151
I1
aI3
aI7
aI7
aI9
aI41
aI1
aI253
aI71
aI1009
aI427
aI3347
aI6987
aI3303
aI30535
aI33345
aI126459
asS'd'
I8076
sa(dp16152
S'a'
I20130
sS's'
I17
sS'm'
(lp16153
I1
aI1
aI1
aI7
aI11
aI11
aI15
aI11
aI305
aI559
aI1805
aI1111
aI377
aI1495
aI13471
aI34831
aI123125
asS'd'
I8077
sa(dp16154
S'a'
I20136
sS's'
I17
sS'm'
(lp16155
I1
aI1
aI5
aI7
aI9
aI37
aI27
aI45
aI61
aI705
aI263
aI3181
aI7077
aI5227
aI28121
aI42865
aI3809
asS'd'
I8078
sa(dp16156
S'a'
I20147
sS's'
I17
sS'm'
(lp16157
I1
aI1
aI5
aI1
aI23
aI25
aI29
aI199
aI259
aI959
aI661
aI43
aI6157
aI1547
aI1497
aI24077
aI129939
asS'd'
I8079
sa(dp16158
S'a'
I20153
sS's'
I17
sS'm'
(lp16159
I1
aI3
aI5
aI3
aI13
aI49
aI33
aI19
aI367
aI891
aI1777
aI3119
aI5673
aI8383
aI14487
aI1763
aI63495
asS'd'
I8080
sa(dp16160
S'a'
I20156
sS's'
I17
sS'm'
(lp16161
I1
aI1
aI1
aI13
aI9
aI57
aI35
aI181
aI7
aI225
aI449
aI3843
aI6257
aI4983
aI31307
aI16559
aI27633
asS'd'
I8081
sa(dp16162
S'a'
I20167
sS's'
I17
sS'm'
(lp16163
I1
aI3
aI1
aI11
aI7
aI33
aI55
aI81
aI41
aI61
aI1711
aI3273
aI7629
aI11283
aI9103
aI24105
aI107547
asS'd'
I8082
sa(dp16164
S'a'
I20173
sS's'
I17
sS'm'
(lp16165
I1
aI3
aI5
aI5
aI13
aI17
aI13
aI51
aI235
aI869
aI1543
aI1249
aI7749
aI14773
aI21751
aI53497
aI108709
asS'd'
I8083
sa(dp16166
S'a'
I20182
sS's'
I17
sS'm'
(lp16167
I1
aI1
aI3
aI9
aI3
aI63
aI89
aI43
aI23
aI479
aI115
aI3917
aI7943
aI7323
aI5659
aI64507
aI46941
asS'd'
I8084
sa(dp16168
S'a'
I20185
sS's'
I17
sS'm'
(lp16169
I1
aI1
aI3
aI1
aI25
aI63
aI25
aI169
aI459
aI633
aI1785
aI1059
aI5113
aI4799
aI29281
aI24561
aI17017
asS'd'
I8085
sa(dp16170
S'a'
I20202
sS's'
I17
sS'm'
(lp16171
I1
aI1
aI3
aI3
aI15
aI3
aI11
aI173
aI493
aI5
aI1575
aI653
aI7391
aI7453
aI8297
aI28653
aI6213
asS'd'
I8086
sa(dp16172
S'a'
I20209
sS's'
I17
sS'm'
(lp16173
I1
aI1
aI7
aI5
aI29
aI5
aI1
aI57
aI75
aI479
aI787
aI3417
aI3349
aI111
aI17787
aI41141
aI97597
asS'd'
I8087
sa(dp16174
S'a'
I20229
sS's'
I17
sS'm'
(lp16175
I1
aI3
aI7
aI1
aI11
aI7
aI107
aI159
aI435
aI159
aI1401
aI803
aI7065
aI5923
aI4005
aI37271
aI113791
asS'd'
I8088
sa(dp16176
S'a'
I20233
sS's'
I17
sS'm'
(lp16177
I1
aI1
aI5
aI1
aI23
aI55
aI7
aI59
aI351
aI801
aI1279
aI3231
aI4561
aI2857
aI20563
aI46115
aI79489
asS'd'
I8089
sa(dp16178
S'a'
I20236
sS's'
I17
sS'm'
(lp16179
I1
aI3
aI3
aI15
aI19
aI13
aI113
aI33
aI149
aI175
aI1127
aI3815
aI4357
aI12645
aI4449
aI61355
aI83023
asS'd'
I8090
sa(dp16180
S'a'
I20241
sS's'
I17
sS'm'
(lp16181
I1
aI1
aI7
aI15
aI3
aI17
aI41
aI57
aI243
aI319
aI1631
aI2751
aI7853
aI5977
aI28367
aI20023
aI56049
asS'd'
I8091
sa(dp16182
S'a'
I20242
sS's'
I17
sS'm'
(lp16183
I1
aI3
aI1
aI7
aI27
aI59
aI7
aI13
aI497
aI241
aI1827
aI2861
aI1331
aI1759
aI6037
aI18967
aI42849
asS'd'
I8092
sa(dp16184
S'a'
I20248
sS's'
I17
sS'm'
(lp16185
I1
aI3
aI1
aI1
aI31
aI43
aI41
aI183
aI241
aI219
aI335
aI2331
aI755
aI10589
aI29431
aI29007
aI66667
asS'd'
I8093
sa(dp16186
S'a'
I20278
sS's'
I17
sS'm'
(lp16187
I1
aI1
aI3
aI1
aI27
aI37
aI61
aI117
aI471
aI39
aI139
aI3821
aI2945
aI7035
aI23673
aI20167
aI56169
asS'd'
I8094
sa(dp16188
S'a'
I20282
sS's'
I17
sS'm'
(lp16189
I1
aI3
aI1
aI1
aI9
aI29
aI123
aI61
aI171
aI1015
aI1029
aI1695
aI1039
aI11883
aI259
aI10879
aI97709
asS'd'
I8095
sa(dp16190
S'a'
I20299
sS's'
I17
sS'm'
(lp16191
I1
aI3
aI3
aI5
aI29
aI39
aI65
aI193
aI285
aI635
aI999
aI717
aI5465
aI1849
aI4293
aI19775
aI79121
asS'd'
I8096
sa(dp16192
S'a'
I20307
sS's'
I17
sS'm'
(lp16193
I1
aI1
aI7
aI1
aI3
aI3
aI103
aI15
aI451
aI307
aI1027
aI263
aI6585
aI11651
aI14457
aI62695
aI38407
asS'd'
I8097
sa(dp16194
S'a'
I20313
sS's'
I17
sS'm'
(lp16195
I1
aI3
aI7
aI11
aI13
aI13
aI29
aI83
aI267
aI561
aI2041
aI13
aI3167
aI3475
aI14735
aI34455
aI117533
asS'd'
I8098
sa(dp16196
S'a'
I20314
sS's'
I17
sS'm'
(lp16197
I1
aI3
aI1
aI15
aI5
aI1
aI35
aI225
aI151
aI637
aI1347
aI833
aI7077
aI13145
aI10285
aI46583
aI14351
asS'd'
I8099
sa(dp16198
S'a'
I20320
sS's'
I17
sS'm'
(lp16199
I1
aI1
aI3
aI15
aI27
aI63
aI119
aI159
aI209
aI421
aI1413
aI2727
aI1607
aI7175
aI6133
aI29051
aI97387
asS'd'
I8100
sa(dp16200
S'a'
I20326
sS's'
I17
sS'm'
(lp16201
I1
aI1
aI3
aI5
aI9
aI29
aI35
aI93
aI353
aI903
aI1037
aI469
aI5473
aI7193
aI21883
aI14709
aI89023
asS'd'
I8101
sa(dp16202
S'a'
I20332
sS's'
I17
sS'm'
(lp16203
I1
aI1
aI1
aI11
aI9
aI17
aI51
aI155
aI145
aI443
aI1985
aI423
aI4721
aI15721
aI9747
aI10303
aI21909
asS'd'
I8102
sa(dp16204
S'a'
I20350
sS's'
I17
sS'm'
(lp16205
I1
aI3
aI7
aI15
aI19
aI49
aI53
aI153
aI241
aI739
aI1585
aI3945
aI4869
aI11
aI15845
aI17937
aI69397
asS'd'
I8103
sa(dp16206
S'a'
I20360
sS's'
I17
sS'm'
(lp16207
I1
aI1
aI5
aI7
aI19
aI53
aI43
aI211
aI327
aI723
aI1513
aI1569
aI919
aI1771
aI11309
aI50787
aI7459
asS'd'
I8104
sa(dp16208
S'a'
I20371
sS's'
I17
sS'm'
(lp16209
I1
aI1
aI1
aI9
aI7
aI29
aI49
aI89
aI409
aI685
aI201
aI1327
aI2807
aI13101
aI2485
aI62909
aI102595
asS'd'
I8105
sa(dp16210
S'a'
I20373
sS's'
I17
sS'm'
(lp16211
I1
aI3
aI1
aI13
aI21
aI51
aI37
aI231
aI271
aI475
aI855
aI3703
aI4447
aI5161
aI17937
aI14471
aI47173
asS'd'
I8106
sa(dp16212
S'a'
I20377
sS's'
I17
sS'm'
(lp16213
I1
aI1
aI7
aI3
aI9
aI7
aI121
aI197
aI71
aI147
aI1669
aI1745
aI6589
aI15029
aI1529
aI12625
aI121925
asS'd'
I8107
sa(dp16214
S'a'
I20390
sS's'
I17
sS'm'
(lp16215
I1
aI1
aI1
aI3
aI7
aI47
aI63
aI61
aI187
aI341
aI919
aI307
aI389
aI14141
aI12941
aI17917
aI104289
asS'd'
I8108
sa(dp16216
S'a'
I20396
sS's'
I17
sS'm'
(lp16217
I1
aI3
aI5
aI13
aI19
aI43
aI57
aI11
aI383
aI311
aI1229
aI3527
aI3301
aI12473
aI24377
aI16279
aI92733
asS'd'
I8109
sa(dp16218
S'a'
I20404
sS's'
I17
sS'm'
(lp16219
I1
aI3
aI3
aI5
aI25
aI35
aI63
aI23
aI131
aI481
aI809
aI3627
aI5685
aI13695
aI14121
aI64751
aI66181
asS'd'
I8110
sa(dp16220
S'a'
I20413
sS's'
I17
sS'm'
(lp16221
I1
aI3
aI1
aI5
aI11
aI43
aI89
aI55
aI103
aI219
aI1861
aI3223
aI5111
aI5879
aI31399
aI1395
aI87419
asS'd'
I8111
sa(dp16222
S'a'
I20434
sS's'
I17
sS'm'
(lp16223
I1
aI3
aI1
aI11
aI21
aI27
aI123
aI205
aI47
aI923
aI7
aI1635
aI4019
aI8431
aI28313
aI24275
aI129617
asS'd'
I8112
sa(dp16224
S'a'
I20436
sS's'
I17
sS'm'
(lp16225
I1
aI1
aI1
aI3
aI1
aI11
aI91
aI215
aI393
aI999
aI1071
aI3225
aI4415
aI759
aI24499
aI16109
aI33791
asS'd'
I8113
sa(dp16226
S'a'
I20440
sS's'
I17
sS'm'
(lp16227
I1
aI1
aI3
aI13
aI19
aI45
aI77
aI103
aI105
aI395
aI529
aI3631
aI8179
aI4467
aI30263
aI39379
aI70507
asS'd'
I8114
sa(dp16228
S'a'
I20443
sS's'
I17
sS'm'
(lp16229
I1
aI3
aI3
aI9
aI17
aI45
aI101
aI219
aI433
aI971
aI471
aI1243
aI6955
aI5941
aI20641
aI16119
aI129383
asS'd'
I8115
sa(dp16230
S'a'
I20445
sS's'
I17
sS'm'
(lp16231
I1
aI1
aI7
aI7
aI9
aI9
aI91
aI95
aI503
aI171
aI129
aI1509
aI7179
aI5367
aI2219
aI50445
aI112459
asS'd'
I8116
sa(dp16232
S'a'
I20464
sS's'
I17
sS'm'
(lp16233
I1
aI3
aI7
aI1
aI17
aI17
aI19
aI173
aI229
aI519
aI147
aI1835
aI3797
aI8091
aI20335
aI33177
aI90197
asS'd'
I8117
sa(dp16234
S'a'
I20476
sS's'
I17
sS'm'
(lp16235
I1
aI3
aI3
aI5
aI23
aI29
aI107
aI205
aI43
aI969
aI799
aI1239
aI1353
aI5221
aI15175
aI42945
aI34043
asS'd'
I8118
sa(dp16236
S'a'
I20494
sS's'
I17
sS'm'
(lp16237
I1
aI1
aI5
aI7
aI31
aI63
aI67
aI87
aI189
aI301
aI1719
aI3937
aI965
aI2505
aI24781
aI25133
aI91675
asS'd'
I8119
sa(dp16238
S'a'
I20496
sS's'
I17
sS'm'
(lp16239
I1
aI3
aI1
aI7
aI15
aI25
aI11
aI39
aI281
aI35
aI1149
aI1445
aI6451
aI12069
aI20959
aI29895
aI60059
asS'd'
I8120
sa(dp16240
S'a'
I20501
sS's'
I17
sS'm'
(lp16241
I1
aI1
aI5
aI1
aI1
aI17
aI65
aI213
aI359
aI561
aI2015
aI1629
aI3521
aI6877
aI8099
aI62483
aI103903
asS'd'
I8121
sa(dp16242
S'a'
I20518
sS's'
I17
sS'm'
(lp16243
I1
aI1
aI7
aI9
aI7
aI49
aI1
aI227
aI49
aI823
aI1141
aI2419
aI2697
aI13293
aI14143
aI6323
aI16081
asS'd'
I8122
sa(dp16244
S'a'
I20527
sS's'
I17
sS'm'
(lp16245
I1
aI3
aI3
aI1
aI9
aI13
aI99
aI235
aI343
aI601
aI927
aI183
aI4545
aI14529
aI5521
aI55571
aI90675
asS'd'
I8123
sa(dp16246
S'a'
I20529
sS's'
I17
sS'm'
(lp16247
I1
aI1
aI5
aI1
aI13
aI49
aI95
aI153
aI131
aI955
aI283
aI2951
aI3651
aI7743
aI4285
aI42621
aI110577
asS'd'
I8124
sa(dp16248
S'a'
I20535
sS's'
I17
sS'm'
(lp16249
I1
aI1
aI1
aI9
aI19
aI59
aI97
aI181
aI67
aI357
aI497
aI287
aI3523
aI3729
aI28981
aI59687
aI39463
asS'd'
I8125
sa(dp16250
S'a'
I20544
sS's'
I17
sS'm'
(lp16251
I1
aI1
aI3
aI7
aI5
aI19
aI107
aI55
aI393
aI225
aI1953
aI669
aI8063
aI6537
aI15983
aI59891
aI95079
asS'd'
I8126
sa(dp16252
S'a'
I20568
sS's'
I17
sS'm'
(lp16253
I1
aI3
aI1
aI5
aI29
aI21
aI17
aI169
aI447
aI697
aI1613
aI3483
aI3139
aI11175
aI28865
aI12065
aI130241
asS'd'
I8127
sa(dp16254
S'a'
I20589
sS's'
I17
sS'm'
(lp16255
I1
aI3
aI5
aI7
aI5
aI49
aI35
aI181
aI85
aI505
aI1537
aI1345
aI773
aI3255
aI27405
aI3959
aI126377
asS'd'
I8128
sa(dp16256
S'a'
I20592
sS's'
I17
sS'm'
(lp16257
I1
aI1
aI7
aI15
aI9
aI9
aI17
aI99
aI409
aI319
aI807
aI1721
aI4023
aI2171
aI32657
aI51663
aI23253
asS'd'
I8129
sa(dp16258
S'a'
I20601
sS's'
I17
sS'm'
(lp16259
I1
aI3
aI5
aI1
aI5
aI3
aI37
aI219
aI89
aI263
aI397
aI573
aI6147
aI9525
aI2521
aI11153
aI94319
asS'd'
I8130
sa(dp16260
S'a'
I20617
sS's'
I17
sS'm'
(lp16261
I1
aI1
aI5
aI5
aI11
aI39
aI55
aI205
aI209
aI239
aI1443
aI2477
aI1941
aI8337
aI2883
aI54593
aI129859
asS'd'
I8131
sa(dp16262
S'a'
I20625
sS's'
I17
sS'm'
(lp16263
I1
aI1
aI1
aI7
aI11
aI13
aI127
aI65
aI127
aI413
aI1553
aI413
aI3395
aI9451
aI7517
aI34103
aI57029
asS'd'
I8132
sa(dp16264
S'a'
I20626
sS's'
I17
sS'm'
(lp16265
I1
aI1
aI1
aI15
aI5
aI25
aI109
aI181
aI399
aI1023
aI277
aI3365
aI6209
aI827
aI13933
aI27483
aI63835
asS'd'
I8133
sa(dp16266
S'a'
I20632
sS's'
I17
sS'm'
(lp16267
I1
aI1
aI3
aI3
aI21
aI57
aI95
aI127
aI481
aI365
aI197
aI3631
aI7443
aI4925
aI31277
aI35061
aI35875
asS'd'
I8134
sa(dp16268
S'a'
I20638
sS's'
I17
sS'm'
(lp16269
I1
aI1
aI7
aI13
aI3
aI31
aI59
aI127
aI441
aI967
aI1049
aI1281
aI3553
aI375
aI9683
aI45755
aI18889
asS'd'
I8135
sa(dp16270
S'a'
I20644
sS's'
I17
sS'm'
(lp16271
I1
aI1
aI1
aI13
aI11
aI39
aI49
aI43
aI383
aI607
aI157
aI1887
aI3623
aI13335
aI31949
aI49843
aI96781
asS'd'
I8136
sa(dp16272
S'a'
I20647
sS's'
I17
sS'm'
(lp16273
I1
aI3
aI7
aI13
aI19
aI35
aI21
aI9
aI299
aI425
aI1921
aI3481
aI6849
aI4149
aI5227
aI56737
aI27559
asS'd'
I8137
sa(dp16274
S'a'
I20662
sS's'
I17
sS'm'
(lp16275
I1
aI3
aI7
aI5
aI21
aI11
aI79
aI97
aI1
aI849
aI819
aI1133
aI3393
aI5429
aI10621
aI38787
aI120785
asS'd'
I8138
sa(dp16276
S'a'
I20671
sS's'
I17
sS'm'
(lp16277
I1
aI1
aI1
aI13
aI21
aI29
aI3
aI239
aI399
aI619
aI759
aI2655
aI3691
aI655
aI30979
aI15507
aI114791
asS'd'
I8139
sa(dp16278
S'a'
I20674
sS's'
I17
sS'm'
(lp16279
I1
aI3
aI5
aI3
aI1
aI61
aI79
aI43
aI273
aI965
aI1759
aI3901
aI2437
aI1703
aI20205
aI46291
aI23679
asS'd'
I8140
sa(dp16280
S'a'
I20683
sS's'
I17
sS'm'
(lp16281
I1
aI1
aI1
aI9
aI19
aI57
aI75
aI245
aI377
aI261
aI231
aI3683
aI6745
aI7797
aI28471
aI56269
aI109969
asS'd'
I8141
sa(dp16282
S'a'
I20704
sS's'
I17
sS'm'
(lp16283
I1
aI3
aI1
aI11
aI9
aI55
aI53
aI87
aI33
aI431
aI1805
aI2933
aI455
aI12479
aI16235
aI2667
aI70105
asS'd'
I8142
sa(dp16284
S'a'
I20724
sS's'
I17
sS'm'
(lp16285
I1
aI3
aI5
aI1
aI29
aI1
aI101
aI17
aI377
aI499
aI1365
aI209
aI4819
aI15099
aI8769
aI37003
aI53003
asS'd'
I8143
sa(dp16286
S'a'
I20742
sS's'
I17
sS'm'
(lp16287
I1
aI3
aI5
aI11
aI11
aI39
aI109
aI235
aI337
aI393
aI35
aI1071
aI7085
aI7165
aI25143
aI24223
aI71493
asS'd'
I8144
sa(dp16288
S'a'
I20748
sS's'
I17
sS'm'
(lp16289
I1
aI3
aI1
aI5
aI13
aI49
aI9
aI205
aI305
aI943
aI799
aI2405
aI319
aI10755
aI9785
aI32023
aI48015
asS'd'
I8145
sa(dp16290
S'a'
I20765
sS's'
I17
sS'm'
(lp16291
I1
aI3
aI1
aI9
aI29
aI35
aI123
aI101
aI73
aI747
aI1257
aI407
aI5871
aI4271
aI14837
aI52727
aI13339
asS'd'
I8146
sa(dp16292
S'a'
I20776
sS's'
I17
sS'm'
(lp16293
I1
aI3
aI3
aI9
aI31
aI7
aI113
aI27
aI89
aI123
aI1117
aI531
aI5531
aI7897
aI11209
aI35267
aI123457
asS'd'
I8147
sa(dp16294
S'a'
I20789
sS's'
I17
sS'm'
(lp16295
I1
aI1
aI1
aI1
aI29
aI19
aI93
aI11
aI61
aI743
aI267
aI13
aI6561
aI7667
aI20537
aI12675
aI10481
asS'd'
I8148
sa(dp16296
S'a'
I20796
sS's'
I17
sS'm'
(lp16297
I1
aI1
aI5
aI13
aI27
aI47
aI103
aI171
aI349
aI139
aI1709
aI961
aI783
aI7147
aI5569
aI53395
aI80797
asS'd'
I8149
sa(dp16298
S'a'
I20802
sS's'
I17
sS'm'
(lp16299
I1
aI3
aI1
aI9
aI21
aI49
aI41
aI175
aI507
aI861
aI1157
aI1033
aI6795
aI5795
aI603
aI12485
aI75263
asS'd'
I8150
sa(dp16300
S'a'
I20807
sS's'
I17
sS'm'
(lp16301
I1
aI1
aI5
aI7
aI23
aI3
aI21
aI165
aI123
aI951
aI785
aI2065
aI8001
aI7009
aI22981
aI10011
aI9807
asS'd'
I8151
sa(dp16302
S'a'
I20814
sS's'
I17
sS'm'
(lp16303
I1
aI3
aI7
aI15
aI1
aI53
aI49
aI197
aI231
aI351
aI141
aI3465
aI7907
aI10695
aI30913
aI26753
aI71079
asS'd'
I8152
sa(dp16304
S'a'
I20821
sS's'
I17
sS'm'
(lp16305
I1
aI3
aI5
aI3
aI29
aI45
aI23
aI131
aI65
aI507
aI75
aI275
aI7287
aI187
aI1093
aI52657
aI31533
asS'd'
I8153
sa(dp16306
S'a'
I20832
sS's'
I17
sS'm'
(lp16307
I1
aI3
aI5
aI9
aI9
aI7
aI113
aI125
aI441
aI75
aI663
aI4081
aI3147
aI1469
aI28375
aI35747
aI122965
asS'd'
I8154
sa(dp16308
S'a'
I20835
sS's'
I17
sS'm'
(lp16309
I1
aI1
aI7
aI3
aI3
aI57
aI5
aI17
aI183
aI237
aI965
aI3709
aI4161
aI9513
aI217
aI58835
aI78789
asS'd'
I8155
sa(dp16310
S'a'
I20847
sS's'
I17
sS'm'
(lp16311
I1
aI1
aI3
aI1
aI7
aI25
aI13
aI29
aI173
aI319
aI1723
aI57
aI2401
aI10405
aI15541
aI52915
aI93747
asS'd'
I8156
sa(dp16312
S'a'
I20859
sS's'
I17
sS'm'
(lp16313
I1
aI1
aI7
aI5
aI1
aI31
aI11
aI61
aI341
aI97
aI1199
aI2585
aI5909
aI3707
aI31507
aI51233
aI2389
asS'd'
I8157
sa(dp16314
S'a'
I20871
sS's'
I17
sS'm'
(lp16315
I1
aI1
aI5
aI15
aI15
aI21
aI127
aI155
aI229
aI203
aI1303
aI3215
aI1965
aI9481
aI31909
aI52307
aI112207
asS'd'
I8158
sa(dp16316
S'a'
I20883
sS's'
I17
sS'm'
(lp16317
I1
aI3
aI1
aI13
aI9
aI45
aI91
aI39
aI113
aI587
aI895
aI637
aI2475
aI1695
aI9347
aI53255
aI75797
asS'd'
I8159
sa(dp16318
S'a'
I20886
sS's'
I17
sS'm'
(lp16319
I1
aI3
aI5
aI13
aI17
aI11
aI35
aI83
aI69
aI255
aI741
aI3927
aI153
aI11001
aI29145
aI37107
aI51873
asS'd'
I8160
sa(dp16320
S'a'
I20892
sS's'
I17
sS'm'
(lp16321
I1
aI1
aI7
aI5
aI5
aI37
aI35
aI219
aI153
aI1005
aI973
aI2555
aI893
aI5931
aI23857
aI34631
aI74561
asS'd'
I8161
sa(dp16322
S'a'
I20906
sS's'
I17
sS'm'
(lp16323
I1
aI3
aI1
aI11
aI21
aI63
aI113
aI29
aI115
aI307
aI957
aI407
aI879
aI4819
aI2865
aI1773
aI116825
asS'd'
I8162
sa(dp16324
S'a'
I20908
sS's'
I17
sS'm'
(lp16325
I1
aI3
aI7
aI3
aI19
aI55
aI87
aI21
aI139
aI571
aI311
aI2295
aI2729
aI6371
aI11845
aI30223
aI19247
asS'd'
I8163
sa(dp16326
S'a'
I20957
sS's'
I17
sS'm'
(lp16327
I1
aI3
aI7
aI5
aI23
aI9
aI59
aI25
aI357
aI863
aI435
aI2509
aI5599
aI14039
aI25731
aI41645
aI255
asS'd'
I8164
sa(dp16328
S'a'
I20962
sS's'
I17
sS'm'
(lp16329
I1
aI3
aI7
aI13
aI9
aI3
aI63
aI115
aI503
aI489
aI1585
aI813
aI5419
aI691
aI23973
aI18677
aI59979
asS'd'
I8165
sa(dp16330
S'a'
I20968
sS's'
I17
sS'm'
(lp16331
I1
aI1
aI1
aI1
aI13
aI3
aI55
aI23
aI185
aI731
aI459
aI1497
aI433
aI16243
aI29995
aI20777
aI59513
asS'd'
I8166
sa(dp16332
S'a'
I20979
sS's'
I17
sS'm'
(lp16333
I1
aI1
aI7
aI3
aI27
aI55
aI77
aI57
aI349
aI5
aI617
aI385
aI6225
aI7025
aI23335
aI877
aI21973
asS'd'
I8167
sa(dp16334
S'a'
I20991
sS's'
I17
sS'm'
(lp16335
I1
aI3
aI3
aI5
aI3
aI37
aI105
aI197
aI153
aI639
aI1643
aI1093
aI801
aI4605
aI4551
aI46081
aI7739
asS'd'
I8168
sa(dp16336
S'a'
I20998
sS's'
I17
sS'm'
(lp16337
I1
aI1
aI5
aI11
aI29
aI23
aI5
aI91
aI39
aI489
aI2029
aI887
aI4451
aI11463
aI5641
aI56299
aI34027
asS'd'
I8169
sa(dp16338
S'a'
I21007
sS's'
I17
sS'm'
(lp16339
I1
aI1
aI7
aI3
aI17
aI11
aI25
aI161
aI317
aI701
aI155
aI1989
aI7549
aI11529
aI9945
aI18395
aI61251
asS'd'
I8170
sa(dp16340
S'a'
I21010
sS's'
I17
sS'm'
(lp16341
I1
aI1
aI7
aI13
aI23
aI55
aI113
aI91
aI17
aI149
aI1893
aI2793
aI8185
aI81
aI29487
aI47925
aI51837
asS'd'
I8171
sa(dp16342
S'a'
I21026
sS's'
I17
sS'm'
(lp16343
I1
aI3
aI7
aI7
aI9
aI29
aI103
aI161
aI215
aI129
aI113
aI1987
aI919
aI9639
aI20715
aI6541
aI15041
asS'd'
I8172
sa(dp16344
S'a'
I21037
sS's'
I17
sS'm'
(lp16345
I1
aI1
aI5
aI9
aI19
aI19
aI127
aI43
aI263
aI997
aI1687
aI3801
aI4249
aI6309
aI25889
aI1787
aI122771
asS'd'
I8173
sa(dp16346
S'a'
I21038
sS's'
I17
sS'm'
(lp16347
I1
aI3
aI5
aI13
aI17
aI3
aI91
aI183
aI349
aI467
aI333
aI3299
aI3085
aI12135
aI16801
aI31471
aI37227
asS'd'
I8174
sa(dp16348
S'a'
I21045
sS's'
I17
sS'm'
(lp16349
I1
aI1
aI5
aI3
aI7
aI53
aI11
aI221
aI407
aI545
aI237
aI3631
aI1791
aI3729
aI19737
aI921
aI57303
asS'd'
I8175
sa(dp16350
S'a'
I21057
sS's'
I17
sS'm'
(lp16351
I1
aI3
aI7
aI15
aI3
aI27
aI71
aI45
aI219
aI9
aI1135
aI2267
aI6841
aI8637
aI30317
aI9397
aI115425
asS'd'
I8176
sa(dp16352
S'a'
I21082
sS's'
I17
sS'm'
(lp16353
I1
aI1
aI3
aI5
aI29
aI59
aI121
aI225
aI419
aI813
aI1725
aI3969
aI3451
aI8457
aI31803
aI57659
aI33263
asS'd'
I8177
sa(dp16354
S'a'
I21093
sS's'
I17
sS'm'
(lp16355
I1
aI3
aI3
aI3
aI17
aI3
aI65
aI249
aI423
aI293
aI1333
aI3947
aI1477
aI4005
aI30445
aI28171
aI95823
asS'd'
I8178
sa(dp16356
S'a'
I21094
sS's'
I17
sS'm'
(lp16357
I1
aI3
aI3
aI11
aI29
aI45
aI67
aI89
aI75
aI95
aI555
aI1823
aI2795
aI11929
aI1995
aI30013
aI116241
asS'd'
I8179
sa(dp16358
S'a'
I21105
sS's'
I17
sS'm'
(lp16359
I1
aI3
aI3
aI3
aI23
aI35
aI87
aI221
aI385
aI275
aI803
aI387
aI7765
aI15637
aI27953
aI20913
aI25279
asS'd'
I8180
sa(dp16360
S'a'
I21117
sS's'
I17
sS'm'
(lp16361
I1
aI3
aI7
aI15
aI15
aI43
aI21
aI179
aI393
aI95
aI1913
aI1715
aI4467
aI15093
aI13613
aI50775
aI37133
asS'd'
I8181
sa(dp16362
S'a'
I21121
sS's'
I17
sS'm'
(lp16363
I1
aI1
aI7
aI7
aI31
aI27
aI49
aI71
aI323
aI123
aI597
aI2395
aI4449
aI7249
aI20197
aI19789
aI92685
asS'd'
I8182
sa(dp16364
S'a'
I21124
sS's'
I17
sS'm'
(lp16365
I1
aI1
aI5
aI13
aI31
aI37
aI89
aI225
aI357
aI201
aI1887
aI3915
aI2165
aI10809
aI21623
aI34375
aI110905
asS'd'
I8183
sa(dp16366
S'a'
I21136
sS's'
I17
sS'm'
(lp16367
I1
aI1
aI5
aI7
aI11
aI53
aI37
aI55
aI61
aI679
aI1465
aI1587
aI2215
aI16237
aI14985
aI50507
aI128637
asS'd'
I8184
sa(dp16368
S'a'
I21139
sS's'
I17
sS'm'
(lp16369
I1
aI1
aI7
aI1
aI15
aI53
aI115
aI21
aI279
aI555
aI43
aI2429
aI7251
aI2141
aI4813
aI47047
aI119551
asS'd'
I8185
sa(dp16370
S'a'
I21142
sS's'
I17
sS'm'
(lp16371
I1
aI1
aI5
aI13
aI11
aI41
aI59
aI245
aI337
aI117
aI1125
aI4007
aI947
aI10591
aI17795
aI48535
aI72171
asS'd'
I8186
sa(dp16372
S'a'
I21145
sS's'
I17
sS'm'
(lp16373
I1
aI1
aI5
aI15
aI27
aI41
aI71
aI43
aI505
aI539
aI975
aI1567
aI795
aI4433
aI11689
aI53051
aI98819
asS'd'
I8187
sa(dp16374
S'a'
I21167
sS's'
I17
sS'm'
(lp16375
I1
aI1
aI7
aI9
aI1
aI57
aI57
aI137
aI323
aI311
aI759
aI3027
aI3713
aI13
aI24133
aI21451
aI1153
asS'd'
I8188
sa(dp16376
S'a'
I21170
sS's'
I17
sS'm'
(lp16377
I1
aI1
aI5
aI15
aI31
aI49
aI23
aI123
aI271
aI549
aI1995
aI5
aI6065
aI3797
aI1085
aI50137
aI19741
asS'd'
I8189
sa(dp16378
S'a'
I21175
sS's'
I17
sS'm'
(lp16379
I1
aI3
aI3
aI13
aI5
aI15
aI21
aI117
aI487
aI43
aI1759
aI2899
aI4239
aI9729
aI16711
aI31559
aI34553
asS'd'
I8190
sa(dp16380
S'a'
I21179
sS's'
I17
sS'm'
(lp16381
I1
aI1
aI5
aI13
aI5
aI23
aI83
aI49
aI147
aI267
aI923
aI1377
aI1687
aI1793
aI30383
aI19537
aI66989
asS'd'
I8191
sa(dp16382
S'a'
I21182
sS's'
I17
sS'm'
(lp16383
I1
aI1
aI1
aI13
aI9
aI41
aI105
aI241
aI499
aI891
aI885
aI3349
aI4703
aI5609
aI11999
aI58025
aI69089
asS'd'
I8192
sa(dp16384
S'a'
I21193
sS's'
I17
sS'm'
(lp16385
I1
aI1
aI7
aI9
aI21
aI11
aI121
aI69
aI115
aI895
aI91
aI3745
aI41
aI12787
aI26181
aI31399
aI30463
asS'd'
I8193
sa(dp16386
S'a'
I21194
sS's'
I17
sS'm'
(lp16387
I1
aI1
aI7
aI13
aI11
aI3
aI23
aI173
aI5
aI907
aI45
aI3465
aI2807
aI3731
aI14347
aI27973
aI8567
asS'd'
I8194
sa(dp16388
S'a'
I21207
sS's'
I17
sS'm'
(lp16389
I1
aI3
aI7
aI5
aI27
aI47
aI43
aI25
aI499
aI57
aI649
aI705
aI6223
aI4213
aI4549
aI23213
aI13657
asS'd'
I8195
sa(dp16390
S'a'
I21217
sS's'
I17
sS'm'
(lp16391
I1
aI1
aI7
aI11
aI21
aI35
aI5
aI79
aI295
aI359
aI1993
aI99
aI7917
aI14917
aI2131
aI45527
aI31451
asS'd'
I8196
sa(dp16392
S'a'
I21224
sS's'
I17
sS'm'
(lp16393
I1
aI1
aI5
aI15
aI1
aI39
aI85
aI93
aI65
aI991
aI389
aI585
aI4835
aI11671
aI10913
aI41787
aI84953
asS'd'
I8197
sa(dp16394
S'a'
I21244
sS's'
I17
sS'm'
(lp16395
I1
aI1
aI1
aI5
aI27
aI5
aI1
aI15
aI11
aI83
aI1191
aI3945
aI4697
aI7703
aI6929
aI6057
aI88721
asS'd'
I8198
sa(dp16396
S'a'
I21247
sS's'
I17
sS'm'
(lp16397
I1
aI1
aI3
aI7
aI27
aI39
aI71
aI181
aI191
aI997
aI419
aI1671
aI7771
aI15305
aI18677
aI45033
aI64745
asS'd'
I8199
sa(dp16398
S'a'
I21252
sS's'
I17
sS'm'
(lp16399
I1
aI3
aI7
aI3
aI15
aI41
aI33
aI239
aI93
aI307
aI153
aI2701
aI1549
aI5011
aI6913
aI19257
aI55829
asS'd'
I8200
sa(dp16400
S'a'
I21264
sS's'
I17
sS'm'
(lp16401
I1
aI3
aI3
aI11
aI21
aI47
aI69
aI223
aI365
aI877
aI431
aI1629
aI4803
aI11591
aI13973
aI56359
aI11897
asS'd'
I8201
sa(dp16402
S'a'
I21273
sS's'
I17
sS'm'
(lp16403
I1
aI3
aI7
aI7
aI1
aI59
aI63
aI129
aI251
aI873
aI603
aI2707
aI2847
aI8739
aI31343
aI63291
aI5607
asS'd'
I8202
sa(dp16404
S'a'
I21289
sS's'
I17
sS'm'
(lp16405
I1
aI3
aI5
aI5
aI19
aI13
aI79
aI235
aI151
aI571
aI953
aI2191
aI5973
aI4751
aI11119
aI14439
aI97755
asS'd'
I8203
sa(dp16406
S'a'
I21290
sS's'
I17
sS'm'
(lp16407
I1
aI1
aI5
aI1
aI27
aI3
aI105
aI139
aI13
aI821
aI221
aI2025
aI7303
aI1891
aI28193
aI45537
aI92703
asS'd'
I8204
sa(dp16408
S'a'
I21295
sS's'
I17
sS'm'
(lp16409
I1
aI3
aI7
aI9
aI13
aI63
aI27
aI149
aI51
aI121
aI587
aI3695
aI4115
aI3955
aI22493
aI34903
aI51669
asS'd'
I8205
sa(dp16410
S'a'
I21297
sS's'
I17
sS'm'
(lp16411
I1
aI1
aI5
aI7
aI19
aI5
aI89
aI87
aI269
aI585
aI421
aI3827
aI315
aI14739
aI109
aI43009
aI94969
asS'd'
I8206
sa(dp16412
S'a'
I21309
sS's'
I17
sS'm'
(lp16413
I1
aI1
aI5
aI15
aI9
aI53
aI125
aI83
aI159
aI917
aI1583
aI585
aI6839
aI14957
aI20007
aI60467
aI96309
asS'd'
I8207
sa(dp16414
S'a'
I21318
sS's'
I17
sS'm'
(lp16415
I1
aI3
aI5
aI1
aI23
aI49
aI109
aI91
aI17
aI731
aI1083
aI3153
aI1825
aI14293
aI25639
aI44599
aI47541
asS'd'
I8208
sa(dp16416
S'a'
I21322
sS's'
I17
sS'm'
(lp16417
I1
aI1
aI3
aI7
aI21
aI51
aI45
aI25
aI367
aI925
aI799
aI1673
aI6977
aI7155
aI829
aI25899
aI104615
asS'd'
I8209
sa(dp16418
S'a'
I21324
sS's'
I17
sS'm'
(lp16419
I1
aI3
aI3
aI13
aI13
aI49
aI95
aI239
aI195
aI353
aI1967
aI1419
aI929
aI503
aI11717
aI9485
aI62885
asS'd'
I8210
sa(dp16420
S'a'
I21329
sS's'
I17
sS'm'
(lp16421
I1
aI1
aI1
aI15
aI3
aI41
aI109
aI91
aI327
aI789
aI1429
aI1159
aI2801
aI4845
aI19663
aI47737
aI11029
asS'd'
I8211
sa(dp16422
S'a'
I21332
sS's'
I17
sS'm'
(lp16423
I1
aI3
aI5
aI1
aI21
aI63
aI57
aI107
aI229
aI771
aI1911
aI647
aI6989
aI12615
aI23191
aI64941
aI97595
asS'd'
I8212
sa(dp16424
S'a'
I21336
sS's'
I17
sS'm'
(lp16425
I1
aI1
aI1
aI15
aI5
aI13
aI15
aI109
aI459
aI447
aI1625
aI1269
aI7629
aI7929
aI4405
aI12143
aI40481
asS'd'
I8213
sa(dp16426
S'a'
I21342
sS's'
I17
sS'm'
(lp16427
I1
aI3
aI3
aI1
aI23
aI3
aI95
aI229
aI363
aI379
aI1149
aI1615
aI5125
aI3645
aI27535
aI58791
aI38091
asS'd'
I8214
sa(dp16428
S'a'
I21351
sS's'
I17
sS'm'
(lp16429
I1
aI1
aI1
aI1
aI9
aI45
aI119
aI85
aI151
aI171
aI1123
aI41
aI6517
aI8067
aI17845
aI23301
aI95383
asS'd'
I8215
sa(dp16430
S'a'
I21355
sS's'
I17
sS'm'
(lp16431
I1
aI3
aI1
aI15
aI17
aI31
aI103
aI71
aI35
aI1019
aI1687
aI1175
aI6119
aI14075
aI26601
aI28873
aI36617
asS'd'
I8216
sa(dp16432
S'a'
I21363
sS's'
I17
sS'm'
(lp16433
I1
aI3
aI3
aI9
aI13
aI39
aI7
aI17
aI207
aI219
aI637
aI3025
aI1709
aI4031
aI563
aI14865
aI129389
asS'd'
I8217
sa(dp16434
S'a'
I21372
sS's'
I17
sS'm'
(lp16435
I1
aI3
aI7
aI1
aI21
aI11
aI121
aI85
aI111
aI641
aI1163
aI3173
aI5189
aI13261
aI19471
aI39635
aI76545
asS'd'
I8218
sa(dp16436
S'a'
I21382
sS's'
I17
sS'm'
(lp16437
I1
aI3
aI7
aI15
aI3
aI45
aI3
aI37
aI121
aI967
aI1861
aI3257
aI3699
aI6881
aI11905
aI8413
aI59397
asS'd'
I8219
sa(dp16438
S'a'
I21388
sS's'
I17
sS'm'
(lp16439
I1
aI3
aI3
aI13
aI25
aI53
aI85
aI181
aI1
aI979
aI2045
aI297
aI1739
aI8139
aI17897
aI35251
aI7193
asS'd'
I8220
sa(dp16440
S'a'
I21394
sS's'
I17
sS'm'
(lp16441
I1
aI1
aI1
aI3
aI5
aI49
aI77
aI115
aI377
aI209
aI1415
aI3747
aI485
aI803
aI2507
aI27729
aI52201
asS'd'
I8221
sa(dp16442
S'a'
I21403
sS's'
I17
sS'm'
(lp16443
I1
aI3
aI1
aI5
aI31
aI55
aI85
aI171
aI135
aI893
aI1423
aI3693
aI6155
aI5321
aI8297
aI39183
aI88377
asS'd'
I8222
sa(dp16444
S'a'
I21409
sS's'
I17
sS'm'
(lp16445
I1
aI3
aI3
aI15
aI1
aI35
aI73
aI239
aI181
aI101
aI509
aI2449
aI4955
aI13049
aI27631
aI16871
aI40151
asS'd'
I8223
sa(dp16446
S'a'
I21416
sS's'
I17
sS'm'
(lp16447
I1
aI1
aI7
aI13
aI27
aI7
aI109
aI71
aI437
aI835
aI563
aI1355
aI3681
aI7431
aI32743
aI59693
aI125055
asS'd'
I8224
sa(dp16448
S'a'
I21421
sS's'
I17
sS'm'
(lp16449
I1
aI1
aI7
aI5
aI19
aI23
aI29
aI147
aI291
aI507
aI1943
aI2069
aI3309
aI11569
aI1031
aI49345
aI86735
asS'd'
I8225
sa(dp16450
S'a'
I21424
sS's'
I17
sS'm'
(lp16451
I1
aI1
aI3
aI13
aI17
aI45
aI91
aI167
aI19
aI137
aI527
aI961
aI4435
aI2277
aI6863
aI57917
aI129407
asS'd'
I8226
sa(dp16452
S'a'
I21444
sS's'
I17
sS'm'
(lp16453
I1
aI3
aI5
aI7
aI11
aI31
aI79
aI207
aI43
aI871
aI1121
aI2929
aI6899
aI4099
aI29533
aI45333
aI33299
asS'd'
I8227
sa(dp16454
S'a'
I21453
sS's'
I17
sS'm'
(lp16455
I1
aI1
aI7
aI5
aI5
aI49
aI13
aI95
aI475
aI91
aI337
aI3531
aI3157
aI1331
aI32655
aI46169
aI3549
asS'd'
I8228
sa(dp16456
S'a'
I21466
sS's'
I17
sS'm'
(lp16457
I1
aI3
aI1
aI5
aI5
aI9
aI73
aI177
aI123
aI251
aI717
aI541
aI7083
aI6907
aI1417
aI31203
aI9755
asS'd'
I8229
sa(dp16458
S'a'
I21471
sS's'
I17
sS'm'
(lp16459
I1
aI3
aI1
aI11
aI21
aI11
aI91
aI155
aI447
aI165
aI1525
aI2073
aI5103
aI193
aI2677
aI43673
aI70579
asS'd'
I8230
sa(dp16460
S'a'
I21495
sS's'
I17
sS'm'
(lp16461
I1
aI3
aI7
aI1
aI7
aI27
aI115
aI247
aI211
aI779
aI1999
aI209
aI3215
aI111
aI25567
aI34641
aI130873
asS'd'
I8231
sa(dp16462
S'a'
I21499
sS's'
I17
sS'm'
(lp16463
I1
aI1
aI5
aI9
aI25
aI7
aI15
aI19
aI217
aI831
aI1577
aI2051
aI3533
aI2337
aI7675
aI2845
aI69135
asS'd'
I8232
sa(dp16464
S'a'
I21504
sS's'
I17
sS'm'
(lp16465
I1
aI3
aI5
aI15
aI29
aI11
aI91
aI59
aI221
aI383
aI1235
aI1261
aI2967
aI14989
aI11455
aI6459
aI58047
asS'd'
I8233
sa(dp16466
S'a'
I21507
sS's'
I17
sS'm'
(lp16467
I1
aI3
aI5
aI1
aI3
aI35
aI5
aI127
aI99
aI981
aI493
aI3001
aI5651
aI3125
aI27789
aI57389
aI115631
asS'd'
I8234
sa(dp16468
S'a'
I21521
sS's'
I17
sS'm'
(lp16469
I1
aI3
aI5
aI5
aI29
aI63
aI123
aI161
aI247
aI177
aI1653
aI2665
aI3903
aI11129
aI20961
aI49429
aI44075
asS'd'
I8235
sa(dp16470
S'a'
I21527
sS's'
I17
sS'm'
(lp16471
I1
aI3
aI1
aI1
aI13
aI9
aI57
aI167
aI291
aI765
aI1929
aI397
aI5503
aI5601
aI6957
aI62003
aI104631
asS'd'
I8236
sa(dp16472
S'a'
I21555
sS's'
I17
sS'm'
(lp16473
I1
aI1
aI7
aI15
aI9
aI43
aI57
aI85
aI157
aI361
aI1931
aI2183
aI8045
aI14939
aI2169
aI25733
aI29095
asS'd'
I8237
sa(dp16474
S'a'
I21558
sS's'
I17
sS'm'
(lp16475
I1
aI1
aI3
aI15
aI13
aI35
aI47
aI123
aI13
aI667
aI1373
aI4069
aI6259
aI13453
aI13439
aI25349
aI99437
asS'd'
I8238
sa(dp16476
S'a'
I21562
sS's'
I17
sS'm'
(lp16477
I1
aI3
aI7
aI1
aI31
aI15
aI69
aI45
aI355
aI919
aI415
aI793
aI3987
aI8785
aI4905
aI8177
aI123989
asS'd'
I8239
sa(dp16478
S'a'
I21570
sS's'
I17
sS'm'
(lp16479
I1
aI3
aI7
aI13
aI29
aI27
aI69
aI57
aI385
aI185
aI171
aI2499
aI3983
aI13437
aI23585
aI21501
aI88079
asS'd'
I8240
sa(dp16480
S'a'
I21576
sS's'
I17
sS'm'
(lp16481
I1
aI1
aI5
aI11
aI27
aI3
aI77
aI99
aI221
aI997
aI1653
aI1963
aI2251
aI15505
aI26347
aI51933
aI100679
asS'd'
I8241
sa(dp16482
S'a'
I21579
sS's'
I17
sS'm'
(lp16483
I1
aI1
aI1
aI7
aI19
aI39
aI49
aI69
aI193
aI235
aI959
aI2823
aI2573
aI8001
aI4389
aI13217
aI60975
asS'd'
I8242
sa(dp16484
S'a'
I21581
sS's'
I17
sS'm'
(lp16485
I1
aI1
aI7
aI5
aI1
aI3
aI3
aI189
aI199
aI293
aI1225
aI1913
aI7271
aI2255
aI661
aI23293
aI82069
asS'd'
I8243
sa(dp16486
S'a'
I21587
sS's'
I17
sS'm'
(lp16487
I1
aI1
aI5
aI5
aI21
aI31
aI35
aI113
aI47
aI479
aI325
aI1663
aI7409
aI8975
aI14151
aI56317
aI79663
asS'd'
I8244
sa(dp16488
S'a'
I21590
sS's'
I17
sS'm'
(lp16489
I1
aI1
aI5
aI9
aI15
aI63
aI99
aI135
aI277
aI715
aI667
aI387
aI6929
aI12873
aI12913
aI2599
aI84939
asS'd'
I8245
sa(dp16490
S'a'
I21599
sS's'
I17
sS'm'
(lp16491
I1
aI1
aI7
aI15
aI23
aI39
aI67
aI25
aI179
aI313
aI459
aI295
aI1103
aI1737
aI7529
aI29463
aI104693
asS'd'
I8246
sa(dp16492
S'a'
I21605
sS's'
I17
sS'm'
(lp16493
I1
aI1
aI3
aI13
aI23
aI11
aI111
aI67
aI105
aI191
aI1967
aI3497
aI7621
aI487
aI18545
aI59521
aI115315
asS'd'
I8247
sa(dp16494
S'a'
I21612
sS's'
I17
sS'm'
(lp16495
I1
aI1
aI1
aI7
aI25
aI45
aI83
aI61
aI231
aI569
aI155
aI2817
aI6985
aI5289
aI6731
aI3087
aI89749
asS'd'
I8248
sa(dp16496
S'a'
I21618
sS's'
I17
sS'm'
(lp16497
I1
aI3
aI7
aI1
aI1
aI61
aI103
aI49
aI135
aI411
aI659
aI1735
aI4461
aI8077
aI12885
aI62791
aI114769
asS'd'
I8249
sa(dp16498
S'a'
I21630
sS's'
I17
sS'm'
(lp16499
I1
aI1
aI7
aI13
aI3
aI53
aI21
aI81
aI433
aI563
aI857
aI891
aI195
aI11669
aI24769
aI56539
aI47601
asS'd'
I8250
sa(dp16500
S'a'
I21639
sS's'
I17
sS'm'
(lp16501
I1
aI3
aI1
aI13
aI3
aI41
aI59
aI101
aI67
aI803
aI1209
aI3267
aI1255
aI5763
aI5483
aI36339
aI38451
asS'd'
I8251
sa(dp16502
S'a'
I21640
sS's'
I17
sS'm'
(lp16503
I1
aI3
aI5
aI3
aI25
aI51
aI53
aI213
aI329
aI11
aI483
aI81
aI2151
aI7623
aI26309
aI15289
aI85103
asS'd'
I8252
sa(dp16504
S'a'
I21643
sS's'
I17
sS'm'
(lp16505
I1
aI3
aI3
aI13
aI23
aI17
aI9
aI161
aI417
aI207
aI39
aI3615
aI7567
aI15207
aI20383
aI58885
aI121765
asS'd'
I8253
sa(dp16506
S'a'
I21648
sS's'
I17
sS'm'
(lp16507
I1
aI3
aI1
aI7
aI15
aI59
aI9
aI195
aI187
aI583
aI341
aI2737
aI3891
aI331
aI18055
aI60455
aI113271
asS'd'
I8254
sa(dp16508
S'a'
I21669
sS's'
I17
sS'm'
(lp16509
I1
aI1
aI3
aI3
aI19
aI25
aI95
aI37
aI281
aI59
aI1613
aI2733
aI5715
aI4067
aI5509
aI5851
aI35189
asS'd'
I8255
sa(dp16510
S'a'
I21679
sS's'
I17
sS'm'
(lp16511
I1
aI3
aI1
aI3
aI31
aI43
aI125
aI107
aI341
aI109
aI1991
aI849
aI7795
aI13607
aI20421
aI3339
aI78979
asS'd'
I8256
sa(dp16512
S'a'
I21681
sS's'
I17
sS'm'
(lp16513
I1
aI3
aI7
aI13
aI15
aI57
aI87
aI151
aI479
aI99
aI479
aI447
aI7407
aI303
aI16397
aI15699
aI122273
asS'd'
I8257
sa(dp16514
S'a'
I21687
sS's'
I17
sS'm'
(lp16515
I1
aI1
aI3
aI1
aI27
aI61
aI79
aI195
aI5
aI839
aI1411
aI3451
aI4627
aI13715
aI18401
aI24325
aI44027
asS'd'
I8258
sa(dp16516
S'a'
I21688
sS's'
I17
sS'm'
(lp16517
I1
aI1
aI7
aI15
aI21
aI39
aI57
aI207
aI213
aI367
aI547
aI1203
aI6385
aI2555
aI31465
aI15675
aI7133
asS'd'
I8259
sa(dp16518
S'a'
I21706
sS's'
I17
sS'm'
(lp16519
I1
aI1
aI5
aI15
aI27
aI3
aI75
aI123
aI337
aI1019
aI1525
aI3065
aI1919
aI10779
aI27409
aI6291
aI86291
asS'd'
I8260
sa(dp16520
S'a'
I21713
sS's'
I17
sS'm'
(lp16521
I1
aI1
aI7
aI11
aI15
aI27
aI67
aI145
aI125
aI521
aI647
aI2957
aI6337
aI14973
aI24139
aI29107
aI27151
asS'd'
I8261
sa(dp16522
S'a'
I21714
sS's'
I17
sS'm'
(lp16523
I1
aI1
aI1
aI13
aI25
aI57
aI103
aI83
aI321
aI111
aI131
aI2051
aI5267
aI4723
aI1939
aI40389
aI4803
asS'd'
I8262
sa(dp16524
S'a'
I21716
sS's'
I17
sS'm'
(lp16525
I1
aI3
aI1
aI7
aI29
aI7
aI35
aI133
aI349
aI855
aI613
aI2563
aI2261
aI2119
aI13939
aI24727
aI26719
asS'd'
I8263
sa(dp16526
S'a'
I21730
sS's'
I17
sS'm'
(lp16527
I1
aI3
aI3
aI1
aI11
aI61
aI25
aI177
aI427
aI1005
aI2027
aI649
aI7871
aI7803
aI4717
aI59207
aI57945
asS'd'
I8264
sa(dp16528
S'a'
I21732
sS's'
I17
sS'm'
(lp16529
I1
aI1
aI7
aI1
aI15
aI45
aI75
aI133
aI193
aI745
aI485
aI197
aI6001
aI13837
aI615
aI43629
aI127321
asS'd'
I8265
sa(dp16530
S'a'
I21749
sS's'
I17
sS'm'
(lp16531
I1
aI3
aI3
aI13
aI5
aI7
aI101
aI183
aI211
aI283
aI1327
aI1395
aI8175
aI13359
aI18361
aI54243
aI3555
asS'd'
I8266
sa(dp16532
S'a'
I21756
sS's'
I17
sS'm'
(lp16533
I1
aI1
aI7
aI13
aI7
aI43
aI19
aI41
aI319
aI701
aI795
aI1407
aI7113
aI9149
aI31953
aI17075
aI53975
asS'd'
I8267
sa(dp16534
S'a'
I21774
sS's'
I17
sS'm'
(lp16535
I1
aI3
aI5
aI13
aI11
aI19
aI101
aI125
aI327
aI75
aI1471
aI3465
aI2247
aI5107
aI11519
aI45161
aI71373
asS'd'
I8268
sa(dp16536
S'a'
I21779
sS's'
I17
sS'm'
(lp16537
I1
aI3
aI7
aI13
aI23
aI59
aI53
aI57
aI137
aI575
aI59
aI3829
aI963
aI11259
aI25771
aI29223
aI79535
asS'd'
I8269
sa(dp16538
S'a'
I21781
sS's'
I17
sS'm'
(lp16539
I1
aI3
aI3
aI11
aI17
aI31
aI111
aI147
aI499
aI441
aI1385
aI769
aI6901
aI8349
aI1427
aI16561
aI7485
asS'd'
I8270
sa(dp16540
S'a'
I21786
sS's'
I17
sS'm'
(lp16541
I1
aI1
aI7
aI9
aI21
aI7
aI47
aI83
aI351
aI867
aI265
aI1329
aI7853
aI6959
aI11821
aI44947
aI42275
asS'd'
I8271
sa(dp16542
S'a'
I21792
sS's'
I17
sS'm'
(lp16543
I1
aI1
aI7
aI15
aI3
aI17
aI79
aI143
aI449
aI577
aI1007
aI1101
aI3229
aI6861
aI23921
aI37551
aI117309
asS'd'
I8272
sa(dp16544
S'a'
I21810
sS's'
I17
sS'm'
(lp16545
I1
aI3
aI5
aI11
aI27
aI63
aI107
aI15
aI289
aI821
aI1723
aI1945
aI1373
aI14469
aI30985
aI55987
aI21255
asS'd'
I8273
sa(dp16546
S'a'
I21819
sS's'
I17
sS'm'
(lp16547
I1
aI3
aI5
aI3
aI21
aI39
aI79
aI85
aI485
aI733
aI2031
aI1573
aI6873
aI12225
aI30471
aI54233
aI26967
asS'd'
I8274
sa(dp16548
S'a'
I21829
sS's'
I17
sS'm'
(lp16549
I1
aI3
aI5
aI7
aI17
aI29
aI93
aI251
aI437
aI583
aI825
aI551
aI6545
aI10905
aI27863
aI37037
aI52129
asS'd'
I8275
sa(dp16550
S'a'
I21830
sS's'
I17
sS'm'
(lp16551
I1
aI3
aI7
aI9
aI23
aI1
aI23
aI85
aI195
aI319
aI1759
aI3985
aI2413
aI16205
aI22197
aI48821
aI94907
asS'd'
I8276
sa(dp16552
S'a'
I21844
sS's'
I17
sS'm'
(lp16553
I1
aI1
aI3
aI7
aI17
aI47
aI3
aI195
aI167
aI925
aI11
aI3431
aI1767
aI5917
aI13915
aI54565
aI64625
asS'd'
I8277
sa(dp16554
S'a'
I21853
sS's'
I17
sS'm'
(lp16555
I1
aI3
aI1
aI13
aI23
aI43
aI97
aI93
aI313
aI773
aI591
aI127
aI6005
aI11497
aI32573
aI8173
aI92053
asS'd'
I8278
sa(dp16556
S'a'
I21867
sS's'
I17
sS'm'
(lp16557
I1
aI1
aI5
aI9
aI17
aI47
aI115
aI237
aI389
aI619
aI377
aI561
aI1333
aI6433
aI9743
aI32673
aI98039
asS'd'
I8279
sa(dp16558
S'a'
I21869
sS's'
I17
sS'm'
(lp16559
I1
aI3
aI7
aI15
aI23
aI17
aI99
aI225
aI145
aI191
aI2041
aI581
aI841
aI9377
aI18123
aI32773
aI66849
asS'd'
I8280
sa(dp16560
S'a'
I21882
sS's'
I17
sS'm'
(lp16561
I1
aI3
aI7
aI15
aI21
aI49
aI97
aI41
aI357
aI527
aI2019
aI2083
aI2611
aI12449
aI20037
aI52503
aI105991
asS'd'
I8281
sa(dp16562
S'a'
I21891
sS's'
I17
sS'm'
(lp16563
I1
aI1
aI5
aI13
aI17
aI53
aI41
aI75
aI355
aI207
aI1675
aI591
aI5797
aI9217
aI16443
aI3205
aI81905
asS'd'
I8282
sa(dp16564
S'a'
I21898
sS's'
I17
sS'm'
(lp16565
I1
aI3
aI7
aI11
aI1
aI61
aI29
aI207
aI449
aI103
aI1527
aI2327
aI7895
aI10137
aI25223
aI51607
aI60809
asS'd'
I8283
sa(dp16566
S'a'
I21917
sS's'
I17
sS'm'
(lp16567
I1
aI3
aI3
aI5
aI15
aI57
aI87
aI233
aI301
aI989
aI485
aI2143
aI7411
aI5475
aI23377
aI56005
aI59721
asS'd'
I8284
sa(dp16568
S'a'
I21934
sS's'
I17
sS'm'
(lp16569
I1
aI3
aI1
aI15
aI29
aI7
aI95
aI141
aI369
aI231
aI735
aI1103
aI1565
aI11575
aI571
aI3257
aI62961
asS'd'
I8285
sa(dp16570
S'a'
I21946
sS's'
I17
sS'm'
(lp16571
I1
aI1
aI5
aI15
aI27
aI19
aI25
aI35
aI303
aI555
aI95
aI1323
aI6139
aI5079
aI21763
aI59591
aI103537
asS'd'
I8286
sa(dp16572
S'a'
I21948
sS's'
I17
sS'm'
(lp16573
I1
aI1
aI1
aI13
aI25
aI23
aI85
aI151
aI135
aI349
aI1753
aI1061
aI7697
aI1723
aI5213
aI12581
aI103995
asS'd'
I8287
sa(dp16574
S'a'
I21963
sS's'
I17
sS'm'
(lp16575
I1
aI3
aI1
aI9
aI29
aI51
aI101
aI195
aI59
aI809
aI1527
aI2179
aI63
aI3681
aI29823
aI57537
aI121371
asS'd'
I8288
sa(dp16576
S'a'
I21966
sS's'
I17
sS'm'
(lp16577
I1
aI1
aI7
aI11
aI27
aI61
aI85
aI213
aI245
aI261
aI1649
aI2423
aI6127
aI5687
aI4247
aI56061
aI109793
asS'd'
I8289
sa(dp16578
S'a'
I21968
sS's'
I17
sS'm'
(lp16579
I1
aI3
aI5
aI15
aI11
aI33
aI127
aI31
aI269
aI857
aI2027
aI2611
aI1729
aI11783
aI16459
aI31083
aI30671
asS'd'
I8290
sa(dp16580
S'a'
I21973
sS's'
I17
sS'm'
(lp16581
I1
aI1
aI7
aI9
aI11
aI29
aI127
aI177
aI505
aI227
aI1499
aI1309
aI6855
aI9999
aI21815
aI32987
aI79109
asS'd'
I8291
sa(dp16582
S'a'
I21977
sS's'
I17
sS'm'
(lp16583
I1
aI3
aI7
aI11
aI7
aI21
aI107
aI1
aI493
aI459
aI867
aI3199
aI7985
aI12957
aI28197
aI41133
aI105985
asS'd'
I8292
sa(dp16584
S'a'
I21980
sS's'
I17
sS'm'
(lp16585
I1
aI1
aI3
aI15
aI1
aI57
aI113
aI97
aI213
aI547
aI1017
aI2961
aI461
aI16125
aI10621
aI4243
aI58277
asS'd'
I8293
sa(dp16586
S'a'
I21984
sS's'
I17
sS'm'
(lp16587
I1
aI1
aI3
aI5
aI11
aI57
aI61
aI47
aI209
aI961
aI333
aI795
aI4491
aI15115
aI25745
aI62633
aI66269
asS'd'
I8294
sa(dp16588
S'a'
I21994
sS's'
I17
sS'm'
(lp16589
I1
aI1
aI7
aI3
aI19
aI13
aI49
aI167
aI455
aI863
aI581
aI1407
aI4247
aI15023
aI2247
aI19981
aI125891
asS'd'
I8295
sa(dp16590
S'a'
I21999
sS's'
I17
sS'm'
(lp16591
I1
aI1
aI7
aI15
aI17
aI55
aI27
aI35
aI33
aI349
aI879
aI1781
aI1075
aI2475
aI30689
aI42043
aI29423
asS'd'
I8296
sa(dp16592
S'a'
I22018
sS's'
I17
sS'm'
(lp16593
I1
aI1
aI1
aI11
aI25
aI53
aI121
aI117
aI117
aI845
aI447
aI3927
aI1951
aI8643
aI24497
aI44833
aI99533
asS'd'
I8297
sa(dp16594
S'a'
I22020
sS's'
I17
sS'm'
(lp16595
I1
aI1
aI7
aI13
aI3
aI59
aI117
aI9
aI359
aI453
aI327
aI3419
aI5957
aI97
aI20541
aI49441
aI5673
asS'd'
I8298
sa(dp16596
S'a'
I22029
sS's'
I17
sS'm'
(lp16597
I1
aI3
aI5
aI5
aI31
aI35
aI95
aI107
aI435
aI733
aI827
aI1361
aI6627
aI8905
aI2681
aI25473
aI46093
asS'd'
I8299
sa(dp16598
S'a'
I22032
sS's'
I17
sS'm'
(lp16599
I1
aI3
aI3
aI5
aI7
aI23
aI75
aI137
aI231
aI915
aI637
aI2963
aI4409
aI12799
aI31587
aI65363
aI69539
asS'd'
I8300
sa(dp16600
S'a'
I22041
sS's'
I17
sS'm'
(lp16601
I1
aI1
aI1
aI7
aI15
aI35
aI19
aI233
aI189
aI837
aI243
aI2525
aI6185
aI565
aI8133
aI4265
aI3089
asS'd'
I8301
sa(dp16602
S'a'
I22047
sS's'
I17
sS'm'
(lp16603
I1
aI1
aI5
aI5
aI19
aI59
aI103
aI201
aI287
aI449
aI21
aI2331
aI341
aI13145
aI18607
aI46407
aI2767
asS'd'
I8302
sa(dp16604
S'a'
I22048
sS's'
I17
sS'm'
(lp16605
I1
aI3
aI3
aI15
aI19
aI41
aI49
aI179
aI109
aI367
aI1185
aI1045
aI1635
aI9647
aI16613
aI25357
aI34291
asS'd'
I8303
sa(dp16606
S'a'
I22071
sS's'
I17
sS'm'
(lp16607
I1
aI3
aI5
aI1
aI13
aI11
aI89
aI25
aI159
aI637
aI1979
aI549
aI3553
aI9163
aI227
aI50553
aI46307
asS'd'
I8304
sa(dp16608
S'a'
I22075
sS's'
I17
sS'm'
(lp16609
I1
aI1
aI3
aI1
aI17
aI33
aI73
aI239
aI261
aI751
aI1267
aI2643
aI2549
aI8331
aI25083
aI9715
aI67289
asS'd'
I8305
sa(dp16610
S'a'
I22078
sS's'
I17
sS'm'
(lp16611
I1
aI1
aI1
aI13
aI3
aI49
aI7
aI35
aI367
aI293
aI903
aI1045
aI569
aI6017
aI27635
aI51833
aI32963
asS'd'
I8306
sa(dp16612
S'a'
I22083
sS's'
I17
sS'm'
(lp16613
I1
aI3
aI5
aI3
aI31
aI3
aI69
aI137
aI57
aI87
aI719
aI3977
aI3031
aI7675
aI24605
aI8757
aI93173
asS'd'
I8307
sa(dp16614
S'a'
I22089
sS's'
I17
sS'm'
(lp16615
I1
aI3
aI3
aI1
aI7
aI45
aI97
aI35
aI233
aI69
aI1525
aI4047
aI2599
aI13679
aI4389
aI49079
aI121465
asS'd'
I8308
sa(dp16616
S'a'
I22097
sS's'
I17
sS'm'
(lp16617
I1
aI1
aI7
aI13
aI7
aI25
aI57
aI211
aI337
aI189
aI1825
aI2451
aI7651
aI11277
aI27763
aI40671
aI57223
asS'd'
I8309
sa(dp16618
S'a'
I22110
sS's'
I17
sS'm'
(lp16619
I1
aI1
aI1
aI1
aI15
aI59
aI55
aI169
aI461
aI907
aI407
aI803
aI3349
aI4727
aI20983
aI47717
aI51647
asS'd'
I8310
sa(dp16620
S'a'
I22113
sS's'
I17
sS'm'
(lp16621
I1
aI3
aI7
aI1
aI15
aI51
aI25
aI119
aI439
aI593
aI1289
aI3959
aI5489
aI13283
aI31837
aI8441
aI58373
asS'd'
I8311
sa(dp16622
S'a'
I22119
sS's'
I17
sS'm'
(lp16623
I1
aI3
aI1
aI9
aI5
aI1
aI81
aI45
aI13
aI537
aI1091
aI3861
aI6781
aI5679
aI2807
aI29757
aI40917
asS'd'
I8312
sa(dp16624
S'a'
I22120
sS's'
I17
sS'm'
(lp16625
I1
aI3
aI5
aI3
aI27
aI41
aI19
aI235
aI207
aI697
aI775
aI837
aI3431
aI3175
aI10807
aI42775
aI67987
asS'd'
I8313
sa(dp16626
S'a'
I22126
sS's'
I17
sS'm'
(lp16627
I1
aI3
aI7
aI3
aI29
aI33
aI35
aI119
aI271
aI609
aI1747
aI2839
aI3415
aI2275
aI30979
aI41293
aI99341
asS'd'
I8314
sa(dp16628
S'a'
I22133
sS's'
I17
sS'm'
(lp16629
I1
aI3
aI3
aI3
aI5
aI17
aI13
aI169
aI269
aI709
aI1449
aI3169
aI1545
aI16075
aI8937
aI39705
aI19609
asS'd'
I8315
sa(dp16630
S'a'
I22134
sS's'
I17
sS'm'
(lp16631
I1
aI3
aI5
aI15
aI29
aI13
aI1
aI199
aI65
aI385
aI977
aI797
aI1181
aI10979
aI241
aI40393
aI73663
asS'd'
I8316
sa(dp16632
S'a'
I22140
sS's'
I17
sS'm'
(lp16633
I1
aI1
aI3
aI7
aI17
aI35
aI47
aI63
aI193
aI451
aI151
aI3415
aI99
aI14557
aI26025
aI31361
aI112639
asS'd'
I8317
sa(dp16634
S'a'
I22147
sS's'
I17
sS'm'
(lp16635
I1
aI1
aI3
aI5
aI19
aI13
aI29
aI33
aI365
aI311
aI1241
aI217
aI6205
aI13067
aI18585
aI21693
aI97127
asS'd'
I8318
sa(dp16636
S'a'
I22161
sS's'
I17
sS'm'
(lp16637
I1
aI1
aI3
aI15
aI19
aI7
aI87
aI25
aI91
aI13
aI1839
aI1445
aI957
aI9779
aI25557
aI37027
aI38987
asS'd'
I8319
sa(dp16638
S'a'
I22164
sS's'
I17
sS'm'
(lp16639
I1
aI1
aI5
aI1
aI21
aI5
aI79
aI67
aI481
aI455
aI37
aI1321
aI7723
aI1413
aI7919
aI11035
aI5739
asS'd'
I8320
sa(dp16640
S'a'
I22173
sS's'
I17
sS'm'
(lp16641
I1
aI1
aI1
aI15
aI9
aI55
aI111
aI1
aI383
aI439
aI1037
aI4055
aI4243
aI10443
aI26737
aI21039
aI130847
asS'd'
I8321
sa(dp16642
S'a'
I22197
sS's'
I17
sS'm'
(lp16643
I1
aI1
aI7
aI9
aI13
aI25
aI71
aI137
aI307
aI717
aI1009
aI2477
aI3861
aI14145
aI14549
aI59589
aI93401
asS'd'
I8322
sa(dp16644
S'a'
I22207
sS's'
I17
sS'm'
(lp16645
I1
aI1
aI7
aI5
aI29
aI63
aI77
aI49
aI471
aI267
aI1457
aI1743
aI1915
aI14793
aI17899
aI28011
aI92183
asS'd'
I8323
sa(dp16646
S'a'
I22210
sS's'
I17
sS'm'
(lp16647
I1
aI3
aI7
aI7
aI7
aI41
aI47
aI251
aI75
aI749
aI1915
aI1015
aI5869
aI3211
aI24097
aI14349
aI130571
asS'd'
I8324
sa(dp16648
S'a'
I22216
sS's'
I17
sS'm'
(lp16649
I1
aI1
aI1
aI1
aI31
aI63
aI105
aI83
aI345
aI147
aI975
aI135
aI7299
aI15801
aI19311
aI26143
aI80293
asS'd'
I8325
sa(dp16650
S'a'
I22234
sS's'
I17
sS'm'
(lp16651
I1
aI1
aI3
aI1
aI7
aI1
aI47
aI45
aI251
aI635
aI583
aI3515
aI5233
aI6281
aI7797
aI37949
aI75877
asS'd'
I8326
sa(dp16652
S'a'
I22257
sS's'
I17
sS'm'
(lp16653
I1
aI1
aI3
aI3
aI5
aI53
aI99
aI175
aI155
aI327
aI1841
aI211
aI2811
aI16099
aI17255
aI34253
aI124141
asS'd'
I8327
sa(dp16654
S'a'
I22264
sS's'
I17
sS'm'
(lp16655
I1
aI3
aI1
aI3
aI13
aI27
aI81
aI217
aI115
aI245
aI101
aI1641
aI29
aI1441
aI4829
aI28399
aI102303
asS'd'
I8328
sa(dp16656
S'a'
I22278
sS's'
I17
sS'm'
(lp16657
I1
aI3
aI1
aI5
aI11
aI55
aI31
aI91
aI337
aI203
aI987
aI977
aI4929
aI14933
aI25149
aI20493
aI19783
asS'd'
I8329
sa(dp16658
S'a'
I22284
sS's'
I17
sS'm'
(lp16659
I1
aI1
aI5
aI9
aI9
aI37
aI103
aI211
aI349
aI165
aI1421
aI3015
aI5133
aI4615
aI28173
aI45787
aI10711
asS'd'
I8330
sa(dp16660
S'a'
I22287
sS's'
I17
sS'm'
(lp16661
I1
aI1
aI1
aI1
aI1
aI17
aI29
aI117
aI421
aI651
aI1617
aI1677
aI7841
aI16303
aI8843
aI1321
aI90701
asS'd'
I8331
sa(dp16662
S'a'
I22299
sS's'
I17
sS'm'
(lp16663
I1
aI1
aI1
aI15
aI27
aI23
aI49
aI195
aI139
aI319
aI1277
aI901
aI63
aI14677
aI21815
aI19497
aI24883
asS'd'
I8332
sa(dp16664
S'a'
I22301
sS's'
I17
sS'm'
(lp16665
I1
aI3
aI3
aI13
aI1
aI23
aI17
aI189
aI293
aI765
aI1503
aI1485
aI7427
aI11807
aI17629
aI61739
aI111365
asS'd'
I8333
sa(dp16666
S'a'
I22308
sS's'
I17
sS'm'
(lp16667
I1
aI1
aI5
aI5
aI15
aI41
aI25
aI53
aI221
aI449
aI1597
aI2763
aI4119
aI6319
aI17509
aI23493
aI104707
asS'd'
I8334
sa(dp16668
S'a'
I22337
sS's'
I17
sS'm'
(lp16669
I1
aI3
aI7
aI11
aI29
aI21
aI101
aI197
aI161
aI457
aI331
aI3817
aI5139
aI14307
aI23225
aI55567
aI62535
asS'd'
I8335
sa(dp16670
S'a'
I22349
sS's'
I17
sS'm'
(lp16671
I1
aI1
aI7
aI5
aI9
aI57
aI39
aI101
aI5
aI847
aI1311
aI313
aI2877
aI14811
aI21969
aI31741
aI8075
asS'd'
I8336
sa(dp16672
S'a'
I22350
sS's'
I17
sS'm'
(lp16673
I1
aI3
aI5
aI3
aI1
aI11
aI45
aI163
aI251
aI775
aI1031
aI1957
aI1631
aI1691
aI3191
aI6255
aI13037
asS'd'
I8337
sa(dp16674
S'a'
I22357
sS's'
I17
sS'm'
(lp16675
I1
aI1
aI3
aI13
aI7
aI11
aI95
aI97
aI409
aI835
aI707
aI1579
aI2409
aI9451
aI15069
aI62425
aI106499
asS'd'
I8338
sa(dp16676
S'a'
I22364
sS's'
I17
sS'm'
(lp16677
I1
aI3
aI3
aI11
aI5
aI25
aI23
aI207
aI429
aI299
aI537
aI1467
aI6309
aI891
aI15009
aI56733
aI60397
asS'd'
I8339
sa(dp16678
S'a'
I22373
sS's'
I17
sS'm'
(lp16679
I1
aI3
aI5
aI3
aI29
aI47
aI95
aI115
aI207
aI177
aI543
aI427
aI145
aI11169
aI7441
aI10911
aI87413
asS'd'
I8340
sa(dp16680
S'a'
I22377
sS's'
I17
sS'm'
(lp16681
I1
aI3
aI7
aI11
aI25
aI53
aI15
aI225
aI115
aI295
aI919
aI39
aI513
aI9989
aI11045
aI24015
aI102387
asS'd'
I8341
sa(dp16682
S'a'
I22380
sS's'
I17
sS'm'
(lp16683
I1
aI1
aI7
aI15
aI13
aI31
aI103
aI143
aI357
aI825
aI183
aI137
aI2671
aI9803
aI14777
aI48333
aI79483
asS'd'
I8342
sa(dp16684
S'a'
I22386
sS's'
I17
sS'm'
(lp16685
I1
aI3
aI5
aI1
aI25
aI13
aI65
aI9
aI461
aI307
aI1289
aI1035
aI7253
aI14223
aI16829
aI23361
aI84987
asS'd'
I8343
sa(dp16686
S'a'
I22391
sS's'
I17
sS'm'
(lp16687
I1
aI3
aI5
aI7
aI5
aI57
aI47
aI251
aI5
aI9
aI965
aI2883
aI3105
aI13931
aI807
aI31977
aI119035
asS'd'
I8344
sa(dp16688
S'a'
I22392
sS's'
I17
sS'm'
(lp16689
I1
aI1
aI3
aI5
aI3
aI7
aI55
aI165
aI3
aI787
aI1587
aI989
aI6049
aI14021
aI30789
aI15283
aI92851
asS'd'
I8345
sa(dp16690
S'a'
I22411
sS's'
I17
sS'm'
(lp16691
I1
aI1
aI5
aI5
aI3
aI17
aI11
aI167
aI487
aI885
aI193
aI3485
aI8179
aI9485
aI24913
aI40267
aI70625
asS'd'
I8346
sa(dp16692
S'a'
I22422
sS's'
I17
sS'm'
(lp16693
I1
aI1
aI7
aI1
aI27
aI31
aI9
aI139
aI73
aI137
aI783
aI321
aI691
aI6157
aI19905
aI45525
aI84877
asS'd'
I8347
sa(dp16694
S'a'
I22425
sS's'
I17
sS'm'
(lp16695
I1
aI3
aI1
aI9
aI17
aI39
aI127
aI177
aI301
aI579
aI1065
aI3899
aI281
aI9177
aI16295
aI51217
aI120293
asS'd'
I8348
sa(dp16696
S'a'
I22431
sS's'
I17
sS'm'
(lp16697
I1
aI1
aI7
aI9
aI31
aI59
aI17
aI93
aI247
aI779
aI847
aI1183
aI3453
aI1073
aI18597
aI2655
aI121633
asS'd'
I8349
sa(dp16698
S'a'
I22438
sS's'
I17
sS'm'
(lp16699
I1
aI1
aI7
aI1
aI25
aI43
aI47
aI253
aI23
aI999
aI973
aI1201
aI1061
aI5947
aI5619
aI36311
aI1545
asS'd'
I8350
sa(dp16700
S'a'
I22441
sS's'
I17
sS'm'
(lp16701
I1
aI3
aI5
aI7
aI11
aI5
aI103
aI119
aI229
aI657
aI1993
aI1991
aI1597
aI13165
aI19137
aI7161
aI83487
asS'd'
I8351
sa(dp16702
S'a'
I22442
sS's'
I17
sS'm'
(lp16703
I1
aI1
aI1
aI1
aI11
aI23
aI105
aI183
aI467
aI83
aI899
aI2447
aI4949
aI4171
aI28943
aI4829
aI13033
asS'd'
I8352
sa(dp16704
S'a'
I22449
sS's'
I17
sS'm'
(lp16705
I1
aI3
aI1
aI7
aI15
aI7
aI47
aI215
aI253
aI109
aI1975
aI3337
aI1553
aI13575
aI16835
aI61525
aI26423
asS'd'
I8353
sa(dp16706
S'a'
I22452
sS's'
I17
sS'm'
(lp16707
I1
aI1
aI7
aI1
aI21
aI17
aI53
aI79
aI175
aI267
aI999
aI249
aI6177
aI10453
aI12475
aI59801
aI47351
asS'd'
I8354
sa(dp16708
S'a'
I22461
sS's'
I17
sS'm'
(lp16709
I1
aI3
aI5
aI11
aI3
aI57
aI5
aI193
aI421
aI799
aI1833
aI2635
aI6537
aI4669
aI9597
aI40661
aI12113
asS'd'
I8355
sa(dp16710
S'a'
I22467
sS's'
I17
sS'm'
(lp16711
I1
aI1
aI7
aI11
aI9
aI11
aI69
aI103
aI139
aI167
aI159
aI2469
aI703
aI1519
aI21553
aI62875
aI60449
asS'd'
I8356
sa(dp16712
S'a'
I22479
sS's'
I17
sS'm'
(lp16713
I1
aI1
aI5
aI3
aI9
aI11
aI17
aI183
aI499
aI301
aI1275
aI605
aI7655
aI12141
aI7783
aI39413
aI116263
asS'd'
I8357
sa(dp16714
S'a'
I22484
sS's'
I17
sS'm'
(lp16715
I1
aI1
aI1
aI7
aI31
aI55
aI23
aI79
aI49
aI247
aI761
aI3573
aI8187
aI4879
aI27379
aI15725
aI81415
asS'd'
I8358
sa(dp16716
S'a'
I22487
sS's'
I17
sS'm'
(lp16717
I1
aI3
aI5
aI5
aI5
aI49
aI23
aI205
aI509
aI383
aI1165
aI3839
aI7663
aI1539
aI19967
aI55901
aI4351
asS'd'
I8359
sa(dp16718
S'a'
I22493
sS's'
I17
sS'm'
(lp16719
I1
aI1
aI1
aI11
aI7
aI15
aI3
aI159
aI235
aI735
aI391
aI2231
aI5043
aI9759
aI4569
aI35601
aI71989
asS'd'
I8360
sa(dp16720
S'a'
I22510
sS's'
I17
sS'm'
(lp16721
I1
aI3
aI3
aI15
aI23
aI3
aI49
aI97
aI99
aI517
aI1097
aI3517
aI1035
aI2319
aI27705
aI25547
aI101555
asS'd'
I8361
sa(dp16722
S'a'
I22521
sS's'
I17
sS'm'
(lp16723
I1
aI3
aI7
aI11
aI27
aI29
aI33
aI241
aI205
aI113
aI291
aI1993
aI3277
aI13155
aI1039
aI42367
aI130477
asS'd'
I8362
sa(dp16724
S'a'
I22533
sS's'
I17
sS'm'
(lp16725
I1
aI1
aI1
aI3
aI29
aI19
aI15
aI159
aI35
aI153
aI1177
aI3011
aI6271
aI8203
aI8971
aI19183
aI102871
asS'd'
I8363
sa(dp16726
S'a'
I22534
sS's'
I17
sS'm'
(lp16727
I1
aI1
aI1
aI5
aI5
aI51
aI19
aI175
aI209
aI895
aI229
aI2355
aI499
aI7877
aI4935
aI22737
aI35587
asS'd'
I8364
sa(dp16728
S'a'
I22543
sS's'
I17
sS'm'
(lp16729
I1
aI3
aI7
aI11
aI15
aI9
aI7
aI113
aI41
aI835
aI1593
aI3933
aI7165
aI10959
aI15487
aI30019
aI114505
asS'd'
I8365
sa(dp16730
S'a'
I22548
sS's'
I17
sS'm'
(lp16731
I1
aI1
aI7
aI5
aI31
aI21
aI27
aI11
aI421
aI165
aI1605
aI1859
aI29
aI13051
aI3273
aI3893
aI56089
asS'd'
I8366
sa(dp16732
S'a'
I22552
sS's'
I17
sS'm'
(lp16733
I1
aI3
aI5
aI5
aI17
aI51
aI55
aI187
aI401
aI977
aI95
aI2617
aI727
aI9609
aI5075
aI48989
aI120299
asS'd'
I8367
sa(dp16734
S'a'
I22558
sS's'
I17
sS'm'
(lp16735
I1
aI1
aI5
aI7
aI21
aI31
aI127
aI87
aI379
aI125
aI247
aI3607
aI2555
aI11873
aI32535
aI16677
aI122273
asS'd'
I8368
sa(dp16736
S'a'
I22561
sS's'
I17
sS'm'
(lp16737
I1
aI1
aI1
aI5
aI19
aI21
aI51
aI185
aI203
aI145
aI1073
aI167
aI235
aI12753
aI17315
aI14683
aI44101
asS'd'
I8369
sa(dp16738
S'a'
I22562
sS's'
I17
sS'm'
(lp16739
I1
aI3
aI3
aI1
aI5
aI61
aI71
aI17
aI63
aI151
aI823
aI17
aI5315
aI4861
aI17279
aI23049
aI84971
asS'd'
I8370
sa(dp16740
S'a'
I22568
sS's'
I17
sS'm'
(lp16741
I1
aI3
aI3
aI5
aI21
aI63
aI21
aI235
aI295
aI467
aI1661
aI2487
aI335
aI6107
aI28709
aI55875
aI129085
asS'd'
I8371
sa(dp16742
S'a'
I22571
sS's'
I17
sS'm'
(lp16743
I1
aI3
aI3
aI5
aI1
aI55
aI35
aI187
aI5
aI687
aI1633
aI2999
aI4513
aI10105
aI15249
aI22591
aI102857
asS'd'
I8372
sa(dp16744
S'a'
I22574
sS's'
I17
sS'm'
(lp16745
I1
aI3
aI1
aI5
aI19
aI1
aI113
aI27
aI261
aI623
aI831
aI3011
aI4091
aI11967
aI17191
aI17433
aI99925
asS'd'
I8373
sa(dp16746
S'a'
I22581
sS's'
I17
sS'm'
(lp16747
I1
aI3
aI5
aI5
aI25
aI59
aI81
aI249
aI463
aI523
aI183
aI3049
aI3675
aI2705
aI28379
aI1279
aI25579
asS'd'
I8374
sa(dp16748
S'a'
I22594
sS's'
I17
sS'm'
(lp16749
I1
aI1
aI3
aI9
aI19
aI19
aI71
aI127
aI189
aI613
aI647
aI1449
aI7755
aI1415
aI9067
aI30683
aI79703
asS'd'
I8375
sa(dp16750
S'a'
I22603
sS's'
I17
sS'm'
(lp16751
I1
aI1
aI7
aI1
aI27
aI33
aI61
aI135
aI233
aI633
aI1969
aI2245
aI5841
aI14069
aI6497
aI63617
aI101483
asS'd'
I8376
sa(dp16752
S'a'
I22605
sS's'
I17
sS'm'
(lp16753
I1
aI3
aI3
aI9
aI23
aI3
aI17
aI163
aI309
aI741
aI2023
aI2647
aI5847
aI7871
aI22311
aI38377
aI70663
asS'd'
I8377
sa(dp16754
S'a'
I22606
sS's'
I17
sS'm'
(lp16755
I1
aI3
aI5
aI15
aI31
aI33
aI51
aI243
aI209
aI273
aI1305
aI1599
aI6115
aI6249
aI8639
aI5903
aI17215
asS'd'
I8378
sa(dp16756
S'a'
I22623
sS's'
I17
sS'm'
(lp16757
I1
aI1
aI1
aI1
aI21
aI11
aI107
aI185
aI463
aI435
aI149
aI3789
aI6283
aI1327
aI20893
aI10417
aI78673
asS'd'
I8379
sa(dp16758
S'a'
I22630
sS's'
I17
sS'm'
(lp16759
I1
aI1
aI1
aI13
aI5
aI53
aI121
aI129
aI493
aI419
aI1711
aI2765
aI7673
aI8979
aI25845
aI62759
aI9669
asS'd'
I8380
sa(dp16760
S'a'
I22651
sS's'
I17
sS'm'
(lp16761
I1
aI3
aI5
aI5
aI1
aI39
aI123
aI47
aI449
aI639
aI625
aI2355
aI511
aI1685
aI1415
aI32417
aI76529
asS'd'
I8381
sa(dp16762
S'a'
I22657
sS's'
I17
sS'm'
(lp16763
I1
aI3
aI1
aI11
aI1
aI49
aI67
aI237
aI203
aI967
aI1401
aI2773
aI4951
aI13889
aI14147
aI41031
aI71897
asS'd'
I8382
sa(dp16764
S'a'
I22669
sS's'
I17
sS'm'
(lp16765
I1
aI3
aI5
aI11
aI13
aI49
aI17
aI113
aI315
aI207
aI1057
aI3395
aI6151
aI2767
aI16571
aI1811
aI66403
asS'd'
I8383
sa(dp16766
S'a'
I22670
sS's'
I17
sS'm'
(lp16767
I1
aI1
aI7
aI7
aI29
aI63
aI49
aI115
aI327
aI987
aI1853
aI3355
aI8139
aI2703
aI30039
aI51343
aI86999
asS'd'
I8384
sa(dp16768
S'a'
I22677
sS's'
I17
sS'm'
(lp16769
I1
aI1
aI3
aI9
aI1
aI3
aI45
aI35
aI509
aI483
aI159
aI1795
aI8023
aI6989
aI3755
aI20887
aI13587
asS'd'
I8385
sa(dp16770
S'a'
I22682
sS's'
I17
sS'm'
(lp16771
I1
aI1
aI3
aI7
aI1
aI27
aI39
aI159
aI283
aI843
aI317
aI3229
aI2297
aI15031
aI22039
aI21721
aI70583
asS'd'
I8386
sa(dp16772
S'a'
I22698
sS's'
I17
sS'm'
(lp16773
I1
aI1
aI3
aI11
aI9
aI23
aI1
aI35
aI79
aI77
aI1671
aI2583
aI647
aI12313
aI16271
aI2959
aI108389
asS'd'
I8387
sa(dp16774
S'a'
I22712
sS's'
I17
sS'm'
(lp16775
I1
aI1
aI1
aI7
aI5
aI55
aI1
aI233
aI429
aI231
aI833
aI1279
aI7815
aI1051
aI30627
aI4435
aI25997
asS'd'
I8388
sa(dp16776
S'a'
I22715
sS's'
I17
sS'm'
(lp16777
I1
aI3
aI1
aI15
aI19
aI53
aI9
aI165
aI307
aI437
aI551
aI2477
aI1841
aI11799
aI18477
aI5871
aI20065
asS'd'
I8389
sa(dp16778
S'a'
I22729
sS's'
I17
sS'm'
(lp16779
I1
aI1
aI1
aI1
aI21
aI5
aI65
aI41
aI77
aI909
aI93
aI751
aI2973
aI7341
aI30427
aI60075
aI71457
asS'd'
I8390
sa(dp16780
S'a'
I22732
sS's'
I17
sS'm'
(lp16781
I1
aI1
aI3
aI11
aI25
aI51
aI49
aI63
aI165
aI263
aI1915
aI747
aI8053
aI6361
aI4843
aI20189
aI110147
asS'd'
I8391
sa(dp16782
S'a'
I22735
sS's'
I17
sS'm'
(lp16783
I1
aI3
aI1
aI9
aI29
aI9
aI45
aI177
aI415
aI557
aI1555
aI2967
aI1239
aI8115
aI12853
aI19193
aI73681
asS'd'
I8392
sa(dp16784
S'a'
I22738
sS's'
I17
sS'm'
(lp16785
I1
aI1
aI5
aI5
aI11
aI5
aI51
aI157
aI325
aI517
aI1601
aI3911
aI1487
aI13631
aI7483
aI61515
aI48937
asS'd'
I8393
sa(dp16786
S'a'
I22740
sS's'
I17
sS'm'
(lp16787
I1
aI3
aI7
aI5
aI29
aI31
aI107
aI47
aI437
aI837
aI1791
aI477
aI1717
aI7
aI25855
aI48793
aI16385
asS'd'
I8394
sa(dp16788
S'a'
I22750
sS's'
I17
sS'm'
(lp16789
I1
aI1
aI1
aI3
aI29
aI49
aI31
aI255
aI233
aI935
aI1993
aI125
aI2255
aI12785
aI2807
aI54697
aI62591
asS'd'
I8395
sa(dp16790
S'a'
I22753
sS's'
I17
sS'm'
(lp16791
I1
aI3
aI1
aI7
aI15
aI13
aI9
aI245
aI79
aI289
aI841
aI253
aI5259
aI16123
aI29189
aI63837
aI127915
asS'd'
I8396
sa(dp16792
S'a'
I22760
sS's'
I17
sS'm'
(lp16793
I1
aI3
aI7
aI15
aI15
aI55
aI91
aI103
aI445
aI289
aI1471
aI423
aI3387
aI15609
aI19311
aI28993
aI23473
asS'd'
I8397
sa(dp16794
S'a'
I22765
sS's'
I17
sS'm'
(lp16795
I1
aI1
aI3
aI11
aI31
aI39
aI69
aI125
aI115
aI309
aI397
aI3417
aI5693
aI10301
aI1489
aI25955
aI2699
asS'd'
I8398
sa(dp16796
S'a'
I22768
sS's'
I17
sS'm'
(lp16797
I1
aI3
aI3
aI5
aI13
aI21
aI51
aI207
aI239
aI311
aI1601
aI2925
aI6285
aI9597
aI30579
aI62957
aI6153
asS'd'
I8399
sa(dp16798
S'a'
I22778
sS's'
I17
sS'm'
(lp16799
I1
aI1
aI7
aI1
aI27
aI21
aI63
aI143
aI399
aI971
aI1385
aI1875
aI5143
aI6423
aI6223
aI27009
aI14237
asS'd'
I8400
sa(dp16800
S'a'
I22785
sS's'
I17
sS'm'
(lp16801
I1
aI3
aI5
aI1
aI5
aI59
aI125
aI133
aI151
aI997
aI1315
aI3007
aI8173
aI16289
aI13409
aI839
aI103519
asS'd'
I8401
sa(dp16802
S'a'
I22809
sS's'
I17
sS'm'
(lp16803
I1
aI1
aI1
aI13
aI7
aI57
aI83
aI33
aI191
aI121
aI939
aI3927
aI6089
aI10083
aI5903
aI52229
aI78325
asS'd'
I8402
sa(dp16804
S'a'
I22810
sS's'
I17
sS'm'
(lp16805
I1
aI1
aI3
aI5
aI9
aI61
aI43
aI107
aI279
aI135
aI1109
aI3779
aI5305
aI15333
aI12217
aI41257
aI20265
asS'd'
I8403
sa(dp16806
S'a'
I22812
sS's'
I17
sS'm'
(lp16807
I1
aI3
aI7
aI1
aI31
aI59
aI83
aI43
aI219
aI119
aI511
aI2973
aI4587
aI10701
aI30959
aI21489
aI124077
asS'd'
I8404
sa(dp16808
S'a'
I22828
sS's'
I17
sS'm'
(lp16809
I1
aI1
aI7
aI9
aI17
aI3
aI59
aI151
aI281
aI209
aI1405
aI173
aI3589
aI7679
aI29803
aI53947
aI68291
asS'd'
I8405
sa(dp16810
S'a'
I22840
sS's'
I17
sS'm'
(lp16811
I1
aI1
aI7
aI7
aI5
aI19
aI53
aI91
aI1
aI513
aI1495
aI231
aI3627
aI1115
aI16121
aI12953
aI108343
asS'd'
I8406
sa(dp16812
S'a'
I22845
sS's'
I17
sS'm'
(lp16813
I1
aI3
aI1
aI13
aI17
aI3
aI35
aI35
aI211
aI481
aI2029
aI1035
aI3131
aI5653
aI18097
aI10735
aI102453
asS'd'
I8407
sa(dp16814
S'a'
I22848
sS's'
I17
sS'm'
(lp16815
I1
aI3
aI1
aI11
aI29
aI7
aI121
aI135
aI51
aI837
aI681
aI1497
aI7435
aI2215
aI26527
aI33029
aI93241
asS'd'
I8408
sa(dp16816
S'a'
I22857
sS's'
I17
sS'm'
(lp16817
I1
aI3
aI3
aI15
aI29
aI43
aI17
aI243
aI195
aI315
aI499
aI3801
aI5691
aI12119
aI4061
aI51769
aI80497
asS'd'
I8409
sa(dp16818
S'a'
I22877
sS's'
I17
sS'm'
(lp16819
I1
aI1
aI3
aI1
aI11
aI1
aI113
aI11
aI387
aI579
aI275
aI2995
aI895
aI11859
aI4017
aI1543
aI11853
asS'd'
I8410
sa(dp16820
S'a'
I22882
sS's'
I17
sS'm'
(lp16821
I1
aI1
aI7
aI9
aI31
aI27
aI63
aI217
aI97
aI275
aI435
aI1355
aI5205
aI6587
aI32589
aI46485
aI103587
asS'd'
I8411
sa(dp16822
S'a'
I22887
sS's'
I17
sS'm'
(lp16823
I1
aI3
aI7
aI3
aI7
aI19
aI51
aI41
aI81
aI261
aI1909
aI1475
aI425
aI3173
aI5679
aI34701
aI34977
asS'd'
I8412
sa(dp16824
S'a'
I22894
sS's'
I17
sS'm'
(lp16825
I1
aI1
aI7
aI3
aI27
aI15
aI35
aI49
aI387
aI471
aI1997
aI3643
aI2701
aI11853
aI21311
aI36027
aI104357
asS'd'
I8413
sa(dp16826
S'a'
I22912
sS's'
I17
sS'm'
(lp16827
I1
aI3
aI1
aI9
aI5
aI47
aI73
aI163
aI309
aI891
aI229
aI2433
aI6715
aI6721
aI25233
aI37043
aI29367
asS'd'
I8414
sa(dp16828
S'a'
I22930
sS's'
I17
sS'm'
(lp16829
I1
aI1
aI1
aI7
aI27
aI15
aI9
aI185
aI421
aI597
aI565
aI143
aI1531
aI15585
aI17057
aI54309
aI82915
asS'd'
I8415
sa(dp16830
S'a'
I22936
sS's'
I17
sS'm'
(lp16831
I1
aI1
aI7
aI1
aI5
aI43
aI87
aI61
aI121
aI341
aI25
aI3795
aI7161
aI11985
aI32197
aI789
aI69543
asS'd'
I8416
sa(dp16832
S'a'
I22939
sS's'
I17
sS'm'
(lp16833
I1
aI3
aI5
aI13
aI29
aI39
aI81
aI39
aI263
aI729
aI1833
aI365
aI1073
aI9869
aI1845
aI52621
aI5
asS'd'
I8417
sa(dp16834
S'a'
I22957
sS's'
I17
sS'm'
(lp16835
I1
aI1
aI7
aI7
aI5
aI33
aI117
aI11
aI371
aI161
aI1303
aI629
aI2285
aI5827
aI32355
aI43359
aI115595
asS'd'
I8418
sa(dp16836
S'a'
I22970
sS's'
I17
sS'm'
(lp16837
I1
aI3
aI5
aI5
aI13
aI57
aI63
aI9
aI243
aI533
aI173
aI2197
aI717
aI13441
aI22131
aI17783
aI3319
asS'd'
I8419
sa(dp16838
S'a'
I22980
sS's'
I17
sS'm'
(lp16839
I1
aI1
aI7
aI11
aI15
aI31
aI87
aI255
aI183
aI273
aI805
aI2347
aI5881
aI15401
aI273
aI17397
aI41827
asS'd'
I8420
sa(dp16840
S'a'
I22984
sS's'
I17
sS'm'
(lp16841
I1
aI3
aI1
aI13
aI7
aI17
aI121
aI49
aI47
aI121
aI333
aI3629
aI5337
aI4117
aI2735
aI36581
aI61345
asS'd'
I8421
sa(dp16842
S'a'
I22992
sS's'
I17
sS'm'
(lp16843
I1
aI3
aI3
aI11
aI9
aI7
aI25
aI223
aI379
aI119
aI385
aI1217
aI4803
aI2947
aI30665
aI7733
aI77893
asS'd'
I8422
sa(dp16844
S'a'
I22998
sS's'
I17
sS'm'
(lp16845
I1
aI3
aI3
aI7
aI31
aI35
aI127
aI97
aI5
aI373
aI7
aI3035
aI843
aI5991
aI9265
aI34289
aI42785
asS'd'
I8423
sa(dp16846
S'a'
I23001
sS's'
I17
sS'm'
(lp16847
I1
aI3
aI7
aI3
aI27
aI19
aI95
aI253
aI349
aI871
aI807
aI413
aI5847
aI10467
aI4277
aI12429
aI75773
asS'd'
I8424
sa(dp16848
S'a'
I23044
sS's'
I17
sS'm'
(lp16849
I1
aI3
aI3
aI7
aI21
aI1
aI79
aI89
aI219
aI505
aI41
aI505
aI5159
aI12839
aI3317
aI49873
aI73705
asS'd'
I8425
sa(dp16850
S'a'
I23061
sS's'
I17
sS'm'
(lp16851
I1
aI3
aI1
aI7
aI21
aI43
aI121
aI113
aI477
aI559
aI1831
aI3759
aI3315
aI6367
aI7149
aI16395
aI44703
asS'd'
I8426
sa(dp16852
S'a'
I23062
sS's'
I17
sS'm'
(lp16853
I1
aI1
aI1
aI7
aI13
aI53
aI35
aI53
aI489
aI975
aI631
aI863
aI3067
aI1905
aI21351
aI14705
aI80041
asS'd'
I8427
sa(dp16854
S'a'
I23071
sS's'
I17
sS'm'
(lp16855
I1
aI1
aI1
aI5
aI13
aI27
aI121
aI65
aI351
aI123
aI1731
aI367
aI8061
aI5229
aI8537
aI20897
aI130373
asS'd'
I8428
sa(dp16856
S'a'
I23075
sS's'
I17
sS'm'
(lp16857
I1
aI1
aI5
aI11
aI15
aI63
aI101
aI107
aI105
aI619
aI1771
aI3549
aI7191
aI9083
aI16827
aI29639
aI34219
asS'd'
I8429
sa(dp16858
S'a'
I23089
sS's'
I17
sS'm'
(lp16859
I1
aI3
aI1
aI9
aI15
aI13
aI87
aI157
aI379
aI433
aI217
aI815
aI5079
aI1797
aI26707
aI35165
aI92305
asS'd'
I8430
sa(dp16860
S'a'
I23090
sS's'
I17
sS'm'
(lp16861
I1
aI1
aI5
aI13
aI27
aI35
aI31
aI65
aI313
aI629
aI375
aI1391
aI5373
aI3497
aI7311
aI23105
aI45293
asS'd'
I8431
sa(dp16862
S'a'
I23096
sS's'
I17
sS'm'
(lp16863
I1
aI3
aI1
aI3
aI5
aI39
aI91
aI37
aI401
aI419
aI949
aI2431
aI3685
aI6671
aI20789
aI8597
aI44215
asS'd'
I8432
sa(dp16864
S'a'
I23101
sS's'
I17
sS'm'
(lp16865
I1
aI1
aI7
aI11
aI7
aI15
aI3
aI181
aI363
aI913
aI309
aI2009
aI3805
aI6651
aI27677
aI37711
aI40813
asS'd'
I8433
sa(dp16866
S'a'
I23114
sS's'
I17
sS'm'
(lp16867
I1
aI3
aI5
aI5
aI17
aI11
aI47
aI9
aI27
aI459
aI773
aI1403
aI7069
aI12651
aI8163
aI42425
aI126697
asS'd'
I8434
sa(dp16868
S'a'
I23121
sS's'
I17
sS'm'
(lp16869
I1
aI3
aI1
aI3
aI11
aI21
aI65
aI103
aI405
aI843
aI59
aI3653
aI1759
aI5265
aI401
aI58019
aI124999
asS'd'
I8435
sa(dp16870
S'a'
I23124
sS's'
I17
sS'm'
(lp16871
I1
aI1
aI3
aI7
aI11
aI25
aI61
aI211
aI199
aI849
aI1835
aI1181
aI5003
aI3873
aI23743
aI45451
aI54901
asS'd'
I8436
sa(dp16872
S'a'
I23127
sS's'
I17
sS'm'
(lp16873
I1
aI3
aI5
aI3
aI29
aI25
aI43
aI199
aI481
aI991
aI699
aI3937
aI7601
aI1253
aI24399
aI6625
aI93917
asS'd'
I8437
sa(dp16874
S'a'
I23128
sS's'
I17
sS'm'
(lp16875
I1
aI1
aI7
aI3
aI29
aI33
aI33
aI151
aI3
aI825
aI743
aI773
aI7825
aI8157
aI22121
aI50095
aI16435
asS'd'
I8438
sa(dp16876
S'a'
I23137
sS's'
I17
sS'm'
(lp16877
I1
aI3
aI1
aI1
aI27
aI15
aI81
aI151
aI271
aI167
aI1755
aI1289
aI7473
aI8525
aI12525
aI63139
aI48787
asS'd'
I8439
sa(dp16878
S'a'
I23138
sS's'
I17
sS'm'
(lp16879
I1
aI1
aI7
aI13
aI27
aI33
aI87
aI125
aI211
aI631
aI149
aI3451
aI643
aI6975
aI2659
aI12629
aI33187
asS'd'
I8440
sa(dp16880
S'a'
I23150
sS's'
I17
sS'm'
(lp16881
I1
aI1
aI3
aI3
aI5
aI49
aI99
aI99
aI85
aI647
aI351
aI2829
aI7005
aI7283
aI5857
aI46157
aI52061
asS'd'
I8441
sa(dp16882
S'a'
I23155
sS's'
I17
sS'm'
(lp16883
I1
aI1
aI3
aI5
aI11
aI37
aI43
aI129
aI21
aI639
aI187
aI2279
aI8189
aI12877
aI28707
aI7133
aI93639
asS'd'
I8442
sa(dp16884
S'a'
I23168
sS's'
I17
sS'm'
(lp16885
I1
aI1
aI3
aI7
aI19
aI13
aI35
aI51
aI77
aI811
aI1553
aI2769
aI763
aI4965
aI4643
aI37639
aI44229
asS'd'
I8443
sa(dp16886
S'a'
I23173
sS's'
I17
sS'm'
(lp16887
I1
aI3
aI5
aI15
aI11
aI29
aI103
aI203
aI435
aI1017
aI531
aI1453
aI1407
aI6569
aI619
aI52103
aI45213
asS'd'
I8444
sa(dp16888
S'a'
I23174
sS's'
I17
sS'm'
(lp16889
I1
aI1
aI7
aI5
aI25
aI25
aI47
aI229
aI201
aI843
aI473
aI2637
aI2265
aI4627
aI20013
aI41217
aI76095
asS'd'
I8445
sa(dp16890
S'a'
I23195
sS's'
I17
sS'm'
(lp16891
I1
aI3
aI3
aI15
aI23
aI61
aI109
aI31
aI57
aI595
aI1303
aI3915
aI67
aI8205
aI3553
aI9543
aI103385
asS'd'
I8446
sa(dp16892
S'a'
I23202
sS's'
I17
sS'm'
(lp16893
I1
aI1
aI3
aI3
aI21
aI19
aI21
aI41
aI137
aI905
aI2045
aI491
aI1783
aI151
aI20963
aI38009
aI735
asS'd'
I8447
sa(dp16894
S'a'
I23225
sS's'
I17
sS'm'
(lp16895
I1
aI1
aI7
aI11
aI13
aI33
aI95
aI251
aI179
aI211
aI1687
aI3189
aI6213
aI3905
aI2117
aI15153
aI4855
asS'd'
I8448
sa(dp16896
S'a'
I23226
sS's'
I17
sS'm'
(lp16897
I1
aI1
aI5
aI3
aI19
aI9
aI67
aI243
aI23
aI611
aI1007
aI1317
aI7303
aI11065
aI21157
aI56677
aI81683
asS'd'
I8449
sa(dp16898
S'a'
I23239
sS's'
I17
sS'm'
(lp16899
I1
aI1
aI3
aI5
aI19
aI41
aI63
aI129
aI233
aI15
aI37
aI1445
aI1095
aI11309
aI30181
aI49199
aI85113
asS'd'
I8450
sa(dp16900
S'a'
I23253
sS's'
I17
sS'm'
(lp16901
I1
aI3
aI7
aI1
aI21
aI53
aI83
aI79
aI155
aI379
aI773
aI1823
aI1003
aI2787
aI31107
aI36115
aI40987
asS'd'
I8451
sa(dp16902
S'a'
I23263
sS's'
I17
sS'm'
(lp16903
I1
aI3
aI3
aI5
aI3
aI19
aI7
aI247
aI417
aI573
aI407
aI3577
aI6079
aI10275
aI29791
aI35149
aI102565
asS'd'
I8452
sa(dp16904
S'a'
I23264
sS's'
I17
sS'm'
(lp16905
I1
aI3
aI3
aI9
aI21
aI49
aI57
aI223
aI125
aI671
aI655
aI2995
aI5849
aI5355
aI21171
aI54857
aI114841
asS'd'
I8453
sa(dp16906
S'a'
I23281
sS's'
I17
sS'm'
(lp16907
I1
aI3
aI7
aI3
aI27
aI23
aI125
aI103
aI485
aI955
aI963
aI1865
aI2321
aI2263
aI32497
aI47973
aI122111
asS'd'
I8454
sa(dp16908
S'a'
I23282
sS's'
I17
sS'm'
(lp16909
I1
aI1
aI3
aI15
aI3
aI1
aI37
aI19
aI287
aI165
aI1717
aI851
aI3619
aI13623
aI24295
aI48253
aI13143
asS'd'
I8455
sa(dp16910
S'a'
I23288
sS's'
I17
sS'm'
(lp16911
I1
aI1
aI7
aI9
aI13
aI59
aI69
aI97
aI113
aI163
aI871
aI1795
aI2719
aI13675
aI11767
aI23687
aI65841
asS'd'
I8456
sa(dp16912
S'a'
I23294
sS's'
I17
sS'm'
(lp16913
I1
aI1
aI5
aI3
aI21
aI31
aI41
aI115
aI469
aI177
aI137
aI2129
aI1385
aI10835
aI16471
aI59411
aI30795
asS'd'
I8457
sa(dp16914
S'a'
I23302
sS's'
I17
sS'm'
(lp16915
I1
aI1
aI7
aI7
aI13
aI45
aI73
aI119
aI457
aI673
aI1481
aI3735
aI2675
aI11413
aI9069
aI34741
aI8757
asS'd'
I8458
sa(dp16916
S'a'
I23311
sS's'
I17
sS'm'
(lp16917
I1
aI3
aI5
aI3
aI15
aI9
aI11
aI191
aI499
aI51
aI1963
aI3957
aI1341
aI7129
aI13491
aI65369
aI4339
asS'd'
I8459
sa(dp16918
S'a'
I23320
sS's'
I17
sS'm'
(lp16919
I1
aI3
aI7
aI1
aI5
aI45
aI103
aI209
aI183
aI205
aI525
aI2417
aI847
aI10801
aI10699
aI16723
aI36421
asS'd'
I8460
sa(dp16920
S'a'
I23325
sS's'
I17
sS'm'
(lp16921
I1
aI3
aI7
aI13
aI3
aI57
aI37
aI75
aI299
aI359
aI2017
aI125
aI6737
aI4859
aI18443
aI20765
aI40319
asS'd'
I8461
sa(dp16922
S'a'
I23356
sS's'
I17
sS'm'
(lp16923
I1
aI1
aI3
aI5
aI5
aI17
aI43
aI141
aI31
aI141
aI1019
aI1685
aI6831
aI9433
aI31245
aI29227
aI64083
asS'd'
I8462
sa(dp16924
S'a'
I23374
sS's'
I17
sS'm'
(lp16925
I1
aI3
aI1
aI13
aI25
aI47
aI107
aI69
aI459
aI595
aI1759
aI3391
aI1531
aI15197
aI25975
aI16971
aI70861
asS'd'
I8463
sa(dp16926
S'a'
I23388
sS's'
I17
sS'm'
(lp16927
I1
aI1
aI3
aI11
aI3
aI53
aI63
aI211
aI69
aI469
aI1407
aI1435
aI2763
aI917
aI19943
aI16591
aI97101
asS'd'
I8464
sa(dp16928
S'a'
I23402
sS's'
I17
sS'm'
(lp16929
I1
aI3
aI5
aI13
aI25
aI41
aI39
aI61
aI319
aI809
aI1109
aI169
aI3101
aI8801
aI21697
aI50759
aI130985
asS'd'
I8465
sa(dp16930
S'a'
I23415
sS's'
I17
sS'm'
(lp16931
I1
aI3
aI1
aI9
aI23
aI1
aI11
aI249
aI243
aI605
aI1419
aI269
aI1601
aI2063
aI5365
aI38077
aI106161
asS'd'
I8466
sa(dp16932
S'a'
I23421
sS's'
I17
sS'm'
(lp16933
I1
aI1
aI7
aI7
aI19
aI55
aI97
aI155
aI477
aI845
aI61
aI263
aI1337
aI8857
aI31611
aI44417
aI43111
asS'd'
I8467
sa(dp16934
S'a'
I23426
sS's'
I17
sS'm'
(lp16935
I1
aI3
aI3
aI15
aI7
aI63
aI45
aI239
aI291
aI279
aI1875
aI3769
aI1571
aI15857
aI13335
aI17209
aI34399
asS'd'
I8468
sa(dp16936
S'a'
I23443
sS's'
I17
sS'm'
(lp16937
I1
aI1
aI7
aI11
aI19
aI19
aI69
aI111
aI217
aI927
aI1643
aI1077
aI4763
aI15893
aI17491
aI39737
aI10705
asS'd'
I8469
sa(dp16938
S'a'
I23446
sS's'
I17
sS'm'
(lp16939
I1
aI1
aI5
aI11
aI3
aI3
aI31
aI199
aI109
aI403
aI973
aI3833
aI2729
aI7285
aI26743
aI53915
aI96203
asS'd'
I8470
sa(dp16940
S'a'
I23455
sS's'
I17
sS'm'
(lp16941
I1
aI3
aI3
aI11
aI9
aI7
aI19
aI145
aI495
aI805
aI381
aI919
aI1323
aI4343
aI15887
aI5163
aI68267
asS'd'
I8471
sa(dp16942
S'a'
I23461
sS's'
I17
sS'm'
(lp16943
I1
aI1
aI3
aI11
aI15
aI31
aI27
aI201
aI251
aI279
aI1377
aI1313
aI7143
aI9731
aI10451
aI63431
aI31307
asS'd'
I8472
sa(dp16944
S'a'
I23468
sS's'
I17
sS'm'
(lp16945
I1
aI1
aI7
aI1
aI1
aI55
aI35
aI249
aI133
aI645
aI425
aI279
aI6401
aI11687
aI751
aI947
aI21791
asS'd'
I8473
sa(dp16946
S'a'
I23471
sS's'
I17
sS'm'
(lp16947
I1
aI3
aI5
aI9
aI5
aI43
aI89
aI31
aI419
aI573
aI1087
aI2197
aI3451
aI2393
aI6569
aI4859
aI36607
asS'd'
I8474
sa(dp16948
S'a'
I23485
sS's'
I17
sS'm'
(lp16949
I1
aI3
aI5
aI15
aI25
aI51
aI11
aI149
aI483
aI789
aI661
aI967
aI3537
aI15511
aI26587
aI29861
aI120337
asS'd'
I8475
sa(dp16950
S'a'
I23486
sS's'
I17
sS'm'
(lp16951
I1
aI3
aI5
aI13
aI21
aI39
aI75
aI111
aI57
aI321
aI1591
aI381
aI7399
aI10807
aI26651
aI62489
aI78341
asS'd'
I8476
sa(dp16952
S'a'
I23488
sS's'
I17
sS'm'
(lp16953
I1
aI3
aI1
aI13
aI1
aI1
aI49
aI137
aI193
aI967
aI805
aI221
aI803
aI11381
aI27803
aI51013
aI10475
asS'd'
I8477
sa(dp16954
S'a'
I23498
sS's'
I17
sS'm'
(lp16955
I1
aI3
aI7
aI5
aI3
aI13
aI47
aI195
aI123
aI753
aI397
aI1203
aI981
aI12863
aI20845
aI36155
aI19055
asS'd'
I8478
sa(dp16956
S'a'
I23500
sS's'
I17
sS'm'
(lp16957
I1
aI1
aI1
aI9
aI9
aI11
aI53
aI203
aI3
aI163
aI1537
aI2061
aI941
aI12629
aI16053
aI34881
aI31489
asS'd'
I8479
sa(dp16958
S'a'
I23515
sS's'
I17
sS'm'
(lp16959
I1
aI1
aI1
aI15
aI5
aI23
aI51
aI197
aI459
aI21
aI1989
aI2529
aI4267
aI1505
aI8951
aI15777
aI20493
asS'd'
I8480
sa(dp16960
S'a'
I23521
sS's'
I17
sS'm'
(lp16961
I1
aI1
aI7
aI3
aI31
aI55
aI9
aI55
aI217
aI695
aI1563
aI4077
aI3207
aI7029
aI10881
aI39581
aI82511
asS'd'
I8481
sa(dp16962
S'a'
I23527
sS's'
I17
sS'm'
(lp16963
I1
aI3
aI1
aI5
aI1
aI11
aI81
aI1
aI505
aI631
aI1093
aI3655
aI2085
aI7349
aI5009
aI49381
aI30527
asS'd'
I8482
sa(dp16964
S'a'
I23534
sS's'
I17
sS'm'
(lp16965
I1
aI1
aI7
aI1
aI27
aI51
aI25
aI235
aI213
aI59
aI611
aI3883
aI2909
aI6411
aI19605
aI49001
aI114529
asS'd'
I8483
sa(dp16966
S'a'
I23546
sS's'
I17
sS'm'
(lp16967
I1
aI3
aI5
aI3
aI25
aI29
aI19
aI137
aI199
aI681
aI1625
aI2711
aI4873
aI14677
aI9767
aI30441
aI54673
asS'd'
I8484
sa(dp16968
S'a'
I23559
sS's'
I17
sS'm'
(lp16969
I1
aI1
aI1
aI9
aI27
aI43
aI109
aI161
aI139
aI675
aI741
aI2839
aI1425
aI5701
aI19897
aI12787
aI33069
asS'd'
I8485
sa(dp16970
S'a'
I23560
sS's'
I17
sS'm'
(lp16971
I1
aI3
aI5
aI11
aI21
aI19
aI77
aI107
aI197
aI591
aI1899
aI1311
aI3347
aI6369
aI26891
aI3771
aI32455
asS'd'
I8486
sa(dp16972
S'a'
I23566
sS's'
I17
sS'm'
(lp16973
I1
aI1
aI7
aI15
aI31
aI13
aI109
aI69
aI207
aI349
aI249
aI971
aI7891
aI10919
aI31579
aI38453
aI124601
asS'd'
I8487
sa(dp16974
S'a'
I23584
sS's'
I17
sS'm'
(lp16975
I1
aI3
aI5
aI5
aI27
aI61
aI67
aI193
aI53
aI259
aI1729
aI4033
aI2637
aI8217
aI22351
aI4001
aI118527
asS'd'
I8488
sa(dp16976
S'a'
I23587
sS's'
I17
sS'm'
(lp16977
I1
aI1
aI3
aI5
aI9
aI45
aI55
aI73
aI189
aI131
aI1947
aI1889
aI837
aI4085
aI10393
aI64359
aI1037
asS'd'
I8489
sa(dp16978
S'a'
I23594
sS's'
I17
sS'm'
(lp16979
I1
aI3
aI7
aI3
aI13
aI51
aI55
aI37
aI335
aI939
aI35
aI461
aI5057
aI2595
aI3305
aI58823
aI3941
asS'd'
I8490
sa(dp16980
S'a'
I23602
sS's'
I17
sS'm'
(lp16981
I1
aI1
aI7
aI11
aI7
aI3
aI121
aI139
aI241
aI477
aI615
aI2707
aI5391
aI7611
aI11563
aI41083
aI16719
asS'd'
I8491
sa(dp16982
S'a'
I23607
sS's'
I17
sS'm'
(lp16983
I1
aI3
aI3
aI15
aI27
aI55
aI13
aI221
aI195
aI543
aI215
aI4035
aI1647
aI8111
aI26425
aI43571
aI79893
asS'd'
I8492
sa(dp16984
S'a'
I23616
sS's'
I17
sS'm'
(lp16985
I1
aI1
aI1
aI5
aI31
aI5
aI35
aI145
aI481
aI339
aI1951
aI2155
aI1309
aI9851
aI31505
aI37371
aI21247
asS'd'
I8493
sa(dp16986
S'a'
I23621
sS's'
I17
sS'm'
(lp16987
I1
aI1
aI7
aI9
aI7
aI5
aI73
aI119
aI3
aI741
aI1351
aI2855
aI2207
aI1465
aI12047
aI13507
aI129173
asS'd'
I8494
sa(dp16988
S'a'
I23631
sS's'
I17
sS'm'
(lp16989
I1
aI1
aI7
aI13
aI5
aI57
aI119
aI63
aI367
aI327
aI1257
aI3191
aI6929
aI9593
aI16565
aI54397
aI100305
asS'd'
I8495
sa(dp16990
S'a'
I23634
sS's'
I17
sS'm'
(lp16991
I1
aI3
aI1
aI11
aI9
aI1
aI85
aI53
aI65
aI945
aI17
aI1963
aI4819
aI16173
aI11669
aI53579
aI33701
asS'd'
I8496
sa(dp16992
S'a'
I23636
sS's'
I17
sS'm'
(lp16993
I1
aI1
aI3
aI15
aI25
aI27
aI3
aI25
aI23
aI429
aI197
aI2717
aI6107
aI6719
aI12457
aI31793
aI78647
asS'd'
I8497
sa(dp16994
S'a'
I23649
sS's'
I17
sS'm'
(lp16995
I1
aI1
aI3
aI1
aI7
aI63
aI111
aI235
aI299
aI91
aI369
aI1423
aI7083
aI4229
aI18535
aI33793
aI19943
asS'd'
I8498
sa(dp16996
S'a'
I23652
sS's'
I17
sS'm'
(lp16997
I1
aI1
aI7
aI13
aI9
aI11
aI123
aI9
aI169
aI895
aI1989
aI1047
aI6139
aI11773
aI19381
aI9593
aI14809
asS'd'
I8499
sa(dp16998
S'a'
I23679
sS's'
I17
sS'm'
(lp16999
I1
aI3
aI1
aI3
aI29
aI31
aI63
aI91
aI59
aI391
aI1695
aI2459
aI3301
aI5615
aI3425
aI8029
aI16069
asS'd'
I8500
sa(dp17000
S'a'
I23686
sS's'
I17
sS'm'
(lp17001
I1
aI1
aI7
aI1
aI25
aI25
aI79
aI49
aI131
aI695
aI987
aI2911
aI1109
aI8237
aI18227
aI37287
aI22443
asS'd'
I8501
sa(dp17002
S'a'
I23697
sS's'
I17
sS'm'
(lp17003
I1
aI3
aI3
aI3
aI25
aI21
aI33
aI207
aI187
aI381
aI129
aI445
aI2967
aI5119
aI18777
aI14849
aI97115
asS'd'
I8502
sa(dp17004
S'a'
I23703
sS's'
I17
sS'm'
(lp17005
I1
aI1
aI7
aI13
aI19
aI9
aI93
aI185
aI391
aI579
aI1509
aI3245
aI3921
aI9473
aI4795
aI6685
aI49549
asS'd'
I8503
sa(dp17006
S'a'
I23714
sS's'
I17
sS'm'
(lp17007
I1
aI1
aI5
aI11
aI1
aI49
aI57
aI127
aI363
aI811
aI1383
aI2869
aI7625
aI15177
aI2581
aI64253
aI53677
asS'd'
I8504
sa(dp17008
S'a'
I23719
sS's'
I17
sS'm'
(lp17009
I1
aI1
aI7
aI3
aI7
aI27
aI73
aI187
aI31
aI1011
aI1013
aI3269
aI6625
aI5001
aI20805
aI13331
aI93725
asS'd'
I8505
sa(dp17010
S'a'
I23723
sS's'
I17
sS'm'
(lp17011
I1
aI3
aI7
aI1
aI23
aI61
aI123
aI9
aI141
aI113
aI1009
aI3713
aI4947
aI9929
aI24125
aI1101
aI104249
asS'd'
I8506
sa(dp17012
S'a'
I23726
sS's'
I17
sS'm'
(lp17013
I1
aI3
aI7
aI3
aI23
aI17
aI25
aI187
aI189
aI875
aI1435
aI163
aI4197
aI6619
aI29031
aI23117
aI45347
asS'd'
I8507
sa(dp17014
S'a'
I23728
sS's'
I17
sS'm'
(lp17015
I1
aI1
aI5
aI7
aI11
aI17
aI9
aI55
aI117
aI223
aI417
aI3993
aI1843
aI5817
aI20435
aI56705
aI98337
asS'd'
I8508
sa(dp17016
S'a'
I23733
sS's'
I17
sS'm'
(lp17017
I1
aI1
aI7
aI3
aI21
aI59
aI3
aI77
aI297
aI61
aI407
aI1603
aI3209
aI1611
aI30185
aI50275
aI56139
asS'd'
I8509
sa(dp17018
S'a'
I23740
sS's'
I17
sS'm'
(lp17019
I1
aI1
aI1
aI5
aI31
aI3
aI101
aI167
aI367
aI543
aI339
aI1885
aI7855
aI9989
aI30969
aI6735
aI108123
asS'd'
I8510
sa(dp17020
S'a'
I23751
sS's'
I17
sS'm'
(lp17021
I1
aI1
aI3
aI9
aI27
aI63
aI9
aI79
aI335
aI351
aI673
aI3107
aI3955
aI1799
aI16879
aI57631
aI109073
asS'd'
I8511
sa(dp17022
S'a'
I23755
sS's'
I17
sS'm'
(lp17023
I1
aI1
aI1
aI3
aI27
aI17
aI107
aI115
aI155
aI371
aI379
aI2837
aI6213
aI2663
aI1101
aI451
aI69517
asS'd'
I8512
sa(dp17024
S'a'
I23765
sS's'
I17
sS'm'
(lp17025
I1
aI1
aI3
aI9
aI13
aI3
aI55
aI9
aI449
aI43
aI1011
aI3281
aI5311
aI223
aI10715
aI6639
aI79949
asS'd'
I8513
sa(dp17026
S'a'
I23766
sS's'
I17
sS'm'
(lp17027
I1
aI3
aI3
aI11
aI23
aI9
aI43
aI185
aI271
aI1005
aI1041
aI2633
aI377
aI4247
aI10417
aI51903
aI19239
asS'd'
I8514
sa(dp17028
S'a'
I23769
sS's'
I17
sS'm'
(lp17029
I1
aI3
aI1
aI9
aI15
aI39
aI115
aI233
aI33
aI425
aI1979
aI583
aI1901
aI8943
aI1527
aI56065
aI50159
asS'd'
I8515
sa(dp17030
S'a'
I23779
sS's'
I17
sS'm'
(lp17031
I1
aI1
aI3
aI1
aI13
aI1
aI105
aI149
aI13
aI625
aI671
aI1811
aI3701
aI241
aI27357
aI25835
aI127265
asS'd'
I8516
sa(dp17032
S'a'
I23794
sS's'
I17
sS'm'
(lp17033
I1
aI3
aI1
aI9
aI11
aI23
aI107
aI197
aI21
aI589
aI1065
aI2591
aI1163
aI15013
aI8931
aI6355
aI87079
asS'd'
I8517
sa(dp17034
S'a'
I23796
sS's'
I17
sS'm'
(lp17035
I1
aI3
aI5
aI3
aI17
aI5
aI121
aI61
aI99
aI987
aI2033
aI2237
aI2299
aI14689
aI19785
aI9599
aI101035
asS'd'
I8518
sa(dp17036
S'a'
I23803
sS's'
I17
sS'm'
(lp17037
I1
aI1
aI1
aI1
aI17
aI25
aI5
aI97
aI55
aI75
aI1419
aI2793
aI7215
aI3185
aI7029
aI23023
aI89089
asS'd'
I8519
sa(dp17038
S'a'
I23813
sS's'
I17
sS'm'
(lp17039
I1
aI3
aI3
aI3
aI11
aI57
aI103
aI191
aI405
aI463
aI1421
aI253
aI6069
aI10905
aI18193
aI719
aI17337
asS'd'
I8520
sa(dp17040
S'a'
I23820
sS's'
I17
sS'm'
(lp17041
I1
aI3
aI5
aI11
aI23
aI37
aI39
aI169
aI295
aI527
aI1671
aI3913
aI6057
aI689
aI27719
aI47245
aI95895
asS'd'
I8521
sa(dp17042
S'a'
I23841
sS's'
I17
sS'm'
(lp17043
I1
aI3
aI7
aI5
aI13
aI9
aI43
aI189
aI411
aI155
aI559
aI3701
aI1623
aI2401
aI10359
aI22675
aI41897
asS'd'
I8522
sa(dp17044
S'a'
I23853
sS's'
I17
sS'm'
(lp17045
I1
aI1
aI1
aI11
aI17
aI55
aI47
aI101
aI357
aI669
aI857
aI2745
aI6425
aI11839
aI13095
aI10757
aI52383
asS'd'
I8523
sa(dp17046
S'a'
I23861
sS's'
I17
sS'm'
(lp17047
I1
aI1
aI7
aI5
aI11
aI13
aI53
aI151
aI93
aI455
aI133
aI3353
aI1417
aI7917
aI12913
aI2615
aI34281
asS'd'
I8524
sa(dp17048
S'a'
I23862
sS's'
I17
sS'm'
(lp17049
I1
aI1
aI3
aI5
aI29
aI57
aI43
aI35
aI203
aI423
aI311
aI3133
aI1757
aI1291
aI2019
aI3115
aI126939
asS'd'
I8525
sa(dp17050
S'a'
I23873
sS's'
I17
sS'm'
(lp17051
I1
aI1
aI3
aI11
aI9
aI43
aI119
aI95
aI135
aI351
aI1865
aI2821
aI717
aI6275
aI19713
aI42315
aI97935
asS'd'
I8526
sa(dp17052
S'a'
I23876
sS's'
I17
sS'm'
(lp17053
I1
aI3
aI7
aI7
aI31
aI51
aI7
aI29
aI405
aI31
aI1765
aI3231
aI1315
aI1307
aI26469
aI62033
aI35619
asS'd'
I8527
sa(dp17054
S'a'
I23897
sS's'
I17
sS'm'
(lp17055
I1
aI1
aI5
aI7
aI5
aI17
aI49
aI137
aI501
aI631
aI1401
aI2851
aI6971
aI14721
aI4329
aI26483
aI120007
asS'd'
I8528
sa(dp17056
S'a'
I23898
sS's'
I17
sS'm'
(lp17057
I1
aI1
aI5
aI13
aI21
aI19
aI95
aI93
aI125
aI331
aI1797
aI1653
aI1891
aI11081
aI30989
aI24671
aI95421
asS'd'
I8529
sa(dp17058
S'a'
I23903
sS's'
I17
sS'm'
(lp17059
I1
aI3
aI3
aI11
aI13
aI29
aI61
aI157
aI165
aI39
aI661
aI89
aI637
aI1397
aI12561
aI62399
aI129107
asS'd'
I8530
sa(dp17060
S'a'
I23904
sS's'
I17
sS'm'
(lp17061
I1
aI3
aI1
aI15
aI13
aI19
aI5
aI115
aI345
aI903
aI531
aI4069
aI6775
aI7433
aI569
aI21779
aI13271
asS'd'
I8531
sa(dp17062
S'a'
I23910
sS's'
I17
sS'm'
(lp17063
I1
aI3
aI3
aI9
aI5
aI53
aI17
aI115
aI67
aI939
aI1907
aI3979
aI4311
aI3573
aI857
aI34931
aI112397
asS'd'
I8532
sa(dp17064
S'a'
I23931
sS's'
I17
sS'm'
(lp17065
I1
aI3
aI7
aI11
aI9
aI47
aI83
aI85
aI277
aI219
aI1701
aI3013
aI3037
aI3473
aI3797
aI40713
aI118573
asS'd'
I8533
sa(dp17066
S'a'
I23933
sS's'
I17
sS'm'
(lp17067
I1
aI1
aI3
aI13
aI25
aI33
aI117
aI115
aI179
aI119
aI487
aI3213
aI2873
aI17
aI20865
aI20043
aI64381
asS'd'
I8534
sa(dp17068
S'a'
I23934
sS's'
I17
sS'm'
(lp17069
I1
aI1
aI1
aI3
aI1
aI45
aI73
aI103
aI75
aI579
aI981
aI2449
aI2141
aI8697
aI22995
aI59693
aI104461
asS'd'
I8535
sa(dp17070
S'a'
I23943
sS's'
I17
sS'm'
(lp17071
I1
aI3
aI1
aI1
aI29
aI9
aI9
aI201
aI55
aI389
aI1069
aI2057
aI4149
aI9217
aI10753
aI7889
aI95849
asS'd'
I8536
sa(dp17072
S'a'
I23952
sS's'
I17
sS'm'
(lp17073
I1
aI3
aI7
aI9
aI27
aI39
aI19
aI223
aI7
aI253
aI55
aI503
aI3339
aI6049
aI32603
aI34807
aI115403
asS'd'
I8537
sa(dp17074
S'a'
I23955
sS's'
I17
sS'm'
(lp17075
I1
aI1
aI5
aI3
aI13
aI21
aI67
aI87
aI205
aI309
aI1371
aI1579
aI281
aI16135
aI28403
aI25951
aI24109
asS'd'
I8538
sa(dp17076
S'a'
I23962
sS's'
I17
sS'm'
(lp17077
I1
aI3
aI1
aI3
aI17
aI21
aI49
aI77
aI393
aI943
aI1701
aI2661
aI5173
aI12875
aI2731
aI40531
aI19301
asS'd'
I8539
sa(dp17078
S'a'
I23971
sS's'
I17
sS'm'
(lp17079
I1
aI3
aI1
aI5
aI23
aI29
aI61
aI161
aI373
aI389
aI1699
aI359
aI2513
aI4717
aI30397
aI24395
aI20881
asS'd'
I8540
sa(dp17080
S'a'
I23978
sS's'
I17
sS'm'
(lp17081
I1
aI3
aI5
aI5
aI29
aI3
aI115
aI251
aI277
aI487
aI7
aI3301
aI7945
aI14233
aI20497
aI62035
aI21537
asS'd'
I8541
sa(dp17082
S'a'
I23998
sS's'
I17
sS'm'
(lp17083
I1
aI1
aI1
aI9
aI7
aI59
aI23
aI85
aI367
aI109
aI1761
aI4011
aI6535
aI8263
aI2081
aI63647
aI69807
asS'd'
I8542
sa(dp17084
S'a'
I24003
sS's'
I17
sS'm'
(lp17085
I1
aI1
aI7
aI11
aI21
aI41
aI29
aI219
aI271
aI617
aI929
aI407
aI2899
aI14299
aI7645
aI44815
aI58817
asS'd'
I8543
sa(dp17086
S'a'
I24009
sS's'
I17
sS'm'
(lp17087
I1
aI3
aI5
aI7
aI11
aI29
aI119
aI33
aI261
aI571
aI2013
aI3327
aI2181
aI12767
aI93
aI2437
aI76533
asS'd'
I8544
sa(dp17088
S'a'
I24017
sS's'
I17
sS'm'
(lp17089
I1
aI1
aI7
aI13
aI17
aI39
aI55
aI203
aI261
aI917
aI967
aI3651
aI7235
aI13751
aI14439
aI7591
aI96553
asS'd'
I8545
sa(dp17090
S'a'
I24045
sS's'
I17
sS'm'
(lp17091
I1
aI1
aI1
aI1
aI11
aI39
aI19
aI21
aI125
aI93
aI1773
aI1155
aI6213
aI7173
aI9057
aI6219
aI4643
asS'd'
I8546
sa(dp17092
S'a'
I24046
sS's'
I17
sS'm'
(lp17093
I1
aI3
aI1
aI5
aI1
aI31
aI55
aI143
aI425
aI539
aI61
aI3377
aI7647
aI257
aI15007
aI24511
aI8707
asS'd'
I8547
sa(dp17094
S'a'
I24060
sS's'
I17
sS'm'
(lp17095
I1
aI3
aI3
aI11
aI27
aI51
aI103
aI197
aI427
aI139
aI181
aI1169
aI3123
aI11803
aI5285
aI1321
aI62267
asS'd'
I8548
sa(dp17096
S'a'
I24064
sS's'
I17
sS'm'
(lp17097
I1
aI3
aI5
aI9
aI11
aI3
aI13
aI149
aI197
aI37
aI31
aI927
aI3313
aI16149
aI14209
aI60177
aI46525
asS'd'
I8549
sa(dp17098
S'a'
I24076
sS's'
I17
sS'm'
(lp17099
I1
aI1
aI5
aI13
aI15
aI29
aI103
aI49
aI355
aI797
aI1253
aI1833
aI621
aI3877
aI9981
aI49207
aI91035
asS'd'
I8550
sa(dp17100
S'a'
I24079
sS's'
I17
sS'm'
(lp17101
I1
aI1
aI3
aI3
aI13
aI19
aI27
aI51
aI151
aI275
aI35
aI3755
aI7511
aI14197
aI26141
aI43765
aI104327
asS'd'
I8551
sa(dp17102
S'a'
I24087
sS's'
I17
sS'm'
(lp17103
I1
aI3
aI5
aI15
aI23
aI47
aI101
aI213
aI97
aI957
aI831
aI1533
aI7913
aI15763
aI29717
aI60425
aI38559
asS'd'
I8552
sa(dp17104
S'a'
I24094
sS's'
I17
sS'm'
(lp17105
I1
aI1
aI7
aI9
aI29
aI31
aI49
aI245
aI361
aI299
aI151
aI2969
aI1487
aI1761
aI11697
aI4043
aI100909
asS'd'
I8553
sa(dp17106
S'a'
I24100
sS's'
I17
sS'm'
(lp17107
I1
aI1
aI1
aI3
aI17
aI49
aI99
aI159
aI3
aI525
aI1527
aI3435
aI5113
aI459
aI13341
aI54103
aI85813
asS'd'
I8554
sa(dp17108
S'a'
I24118
sS's'
I17
sS'm'
(lp17109
I1
aI3
aI1
aI1
aI5
aI59
aI35
aI75
aI107
aI91
aI1621
aI3261
aI619
aI3271
aI10813
aI29857
aI1547
asS'd'
I8555
sa(dp17110
S'a'
I24121
sS's'
I17
sS'm'
(lp17111
I1
aI1
aI5
aI9
aI9
aI33
aI85
aI245
aI39
aI879
aI1621
aI2587
aI3825
aI12939
aI30113
aI24611
aI68491
asS'd'
I8556
sa(dp17112
S'a'
I24132
sS's'
I17
sS'm'
(lp17113
I1
aI3
aI1
aI3
aI9
aI39
aI93
aI241
aI307
aI237
aI3
aI1763
aI7729
aI9257
aI31911
aI32591
aI77333
asS'd'
I8557
sa(dp17114
S'a'
I24147
sS's'
I17
sS'm'
(lp17115
I1
aI3
aI1
aI3
aI27
aI7
aI51
aI121
aI317
aI361
aI1027
aI95
aI7035
aI3097
aI21007
aI38311
aI88287
asS'd'
I8558
sa(dp17116
S'a'
I24165
sS's'
I17
sS'm'
(lp17117
I1
aI3
aI7
aI3
aI19
aI3
aI111
aI115
aI339
aI793
aI1571
aI3101
aI1911
aI14929
aI12841
aI45871
aI119905
asS'd'
I8559
sa(dp17118
S'a'
I24172
sS's'
I17
sS'm'
(lp17119
I1
aI1
aI5
aI7
aI31
aI61
aI37
aI143
aI279
aI941
aI1215
aI2411
aI7617
aI1657
aI10189
aI19139
aI6307
asS'd'
I8560
sa(dp17120
S'a'
I24177
sS's'
I17
sS'm'
(lp17121
I1
aI1
aI3
aI9
aI21
aI35
aI13
aI157
aI187
aI79
aI689
aI1085
aI37
aI4549
aI5901
aI15321
aI61627
asS'd'
I8561
sa(dp17122
S'a'
I24184
sS's'
I17
sS'm'
(lp17123
I1
aI3
aI1
aI13
aI15
aI39
aI21
aI231
aI39
aI327
aI801
aI2321
aI587
aI1877
aI3489
aI54467
aI95773
asS'd'
I8562
sa(dp17124
S'a'
I24187
sS's'
I17
sS'm'
(lp17125
I1
aI1
aI5
aI7
aI1
aI9
aI53
aI1
aI243
aI365
aI789
aI3833
aI317
aI10697
aI26567
aI65187
aI22507
asS'd'
I8563
sa(dp17126
S'a'
I24213
sS's'
I17
sS'm'
(lp17127
I1
aI3
aI3
aI7
aI9
aI41
aI31
aI135
aI425
aI939
aI15
aI2043
aI6593
aI7651
aI25467
aI62549
aI35847
asS'd'
I8564
sa(dp17128
S'a'
I24217
sS's'
I17
sS'm'
(lp17129
I1
aI1
aI1
aI7
aI15
aI23
aI19
aI57
aI421
aI25
aI1037
aI3055
aI6173
aI12451
aI485
aI54567
aI109561
asS'd'
I8565
sa(dp17130
S'a'
I24223
sS's'
I17
sS'm'
(lp17131
I1
aI3
aI5
aI1
aI3
aI29
aI67
aI233
aI157
aI677
aI1711
aI513
aI4673
aI2895
aI1983
aI31075
aI1861
asS'd'
I8566
sa(dp17132
S'a'
I24230
sS's'
I17
sS'm'
(lp17133
I1
aI1
aI1
aI1
aI7
aI39
aI115
aI251
aI275
aI791
aI15
aI1685
aI6835
aI14685
aI12607
aI28213
aI121475
asS'd'
I8567
sa(dp17134
S'a'
I24234
sS's'
I17
sS'm'
(lp17135
I1
aI1
aI5
aI5
aI13
aI11
aI103
aI93
aI489
aI709
aI1339
aI2407
aI1663
aI10195
aI3135
aI15531
aI88427
asS'd'
I8568
sa(dp17136
S'a'
I24241
sS's'
I17
sS'm'
(lp17137
I1
aI1
aI7
aI7
aI17
aI1
aI123
aI143
aI31
aI721
aI1739
aI2273
aI3785
aI10261
aI14741
aI52573
aI113677
asS'd'
I8569
sa(dp17138
S'a'
I24248
sS's'
I17
sS'm'
(lp17139
I1
aI3
aI7
aI3
aI9
aI21
aI77
aI13
aI241
aI429
aI165
aI3399
aI7543
aI2633
aI21129
aI13537
aI84473
asS'd'
I8570
sa(dp17140
S'a'
I24259
sS's'
I17
sS'm'
(lp17141
I1
aI1
aI1
aI11
aI21
aI33
aI125
aI123
aI189
aI337
aI163
aI3727
aI2101
aI14113
aI1719
aI46017
aI68601
asS'd'
I8571
sa(dp17142
S'a'
I24262
sS's'
I17
sS'm'
(lp17143
I1
aI1
aI7
aI9
aI9
aI53
aI101
aI111
aI125
aI605
aI1419
aI3901
aI1769
aI4585
aI20063
aI20857
aI21901
asS'd'
I8572
sa(dp17144
S'a'
I24271
sS's'
I17
sS'm'
(lp17145
I1
aI3
aI7
aI11
aI1
aI19
aI51
aI7
aI457
aI119
aI871
aI3847
aI57
aI11437
aI28763
aI58831
aI675
asS'd'
I8573
sa(dp17146
S'a'
I24279
sS's'
I17
sS'm'
(lp17147
I1
aI3
aI1
aI15
aI25
aI63
aI69
aI25
aI405
aI823
aI1701
aI2441
aI7561
aI8679
aI31643
aI29325
aI25563
asS'd'
I8574
sa(dp17148
S'a'
I24286
sS's'
I17
sS'm'
(lp17149
I1
aI1
aI3
aI9
aI15
aI5
aI89
aI13
aI73
aI951
aI959
aI2693
aI4565
aI13095
aI991
aI12419
aI8267
asS'd'
I8575
sa(dp17150
S'a'
I24289
sS's'
I17
sS'm'
(lp17151
I1
aI1
aI7
aI1
aI15
aI1
aI119
aI223
aI213
aI585
aI1047
aI2623
aI4141
aI2043
aI1583
aI59155
aI5133
asS'd'
I8576
sa(dp17152
S'a'
I24295
sS's'
I17
sS'm'
(lp17153
I1
aI3
aI3
aI3
aI17
aI37
aI81
aI233
aI87
aI843
aI1597
aI1251
aI4713
aI10813
aI24357
aI48499
aI84465
asS'd'
I8577
sa(dp17154
S'a'
I24296
sS's'
I17
sS'm'
(lp17155
I1
aI1
aI1
aI1
aI11
aI55
aI125
aI5
aI255
aI809
aI543
aI2351
aI7079
aI7801
aI29247
aI23937
aI97405
asS'd'
I8578
sa(dp17156
S'a'
I24299
sS's'
I17
sS'm'
(lp17157
I1
aI3
aI3
aI5
aI17
aI55
aI87
aI245
aI371
aI679
aI943
aI655
aI5857
aI261
aI28229
aI22519
aI35191
asS'd'
I8579
sa(dp17158
S'a'
I24314
sS's'
I17
sS'm'
(lp17159
I1
aI1
aI7
aI15
aI9
aI49
aI25
aI155
aI13
aI893
aI1303
aI2317
aI2903
aI15601
aI1433
aI20397
aI70125
asS'd'
I8580
sa(dp17160
S'a'
I24336
sS's'
I17
sS'm'
(lp17161
I1
aI3
aI5
aI3
aI11
aI47
aI99
aI63
aI253
aI95
aI1023
aI397
aI4307
aI4771
aI17027
aI19833
aI18269
asS'd'
I8581
sa(dp17162
S'a'
I24342
sS's'
I17
sS'm'
(lp17163
I1
aI3
aI3
aI7
aI25
aI17
aI69
aI119
aI475
aI575
aI1637
aI3785
aI649
aI11557
aI22457
aI38633
aI96153
asS'd'
I8582
sa(dp17164
S'a'
I24346
sS's'
I17
sS'm'
(lp17165
I1
aI1
aI1
aI5
aI31
aI55
aI85
aI83
aI307
aI201
aI1543
aI727
aI977
aI15779
aI21907
aI31025
aI38969
asS'd'
I8583
sa(dp17166
S'a'
I24357
sS's'
I17
sS'm'
(lp17167
I1
aI3
aI5
aI1
aI7
aI53
aI107
aI239
aI341
aI237
aI1567
aI2717
aI3197
aI12419
aI23733
aI42119
aI86619
asS'd'
I8584
sa(dp17168
S'a'
I24367
sS's'
I17
sS'm'
(lp17169
I1
aI1
aI5
aI13
aI3
aI7
aI105
aI95
aI201
aI953
aI781
aI2043
aI5263
aI13427
aI10303
aI60027
aI19297
asS'd'
I8585
sa(dp17170
S'a'
I24370
sS's'
I17
sS'm'
(lp17171
I1
aI1
aI5
aI15
aI25
aI51
aI5
aI77
aI165
aI297
aI1281
aI1635
aI4139
aI11569
aI32325
aI23135
aI27013
asS'd'
I8586
sa(dp17172
S'a'
I24372
sS's'
I17
sS'm'
(lp17173
I1
aI1
aI3
aI9
aI3
aI59
aI107
aI137
aI251
aI715
aI1477
aI511
aI5629
aI12205
aI7541
aI62559
aI4253
asS'd'
I8587
sa(dp17174
S'a'
I24387
sS's'
I17
sS'm'
(lp17175
I1
aI1
aI7
aI11
aI31
aI29
aI7
aI251
aI119
aI547
aI1179
aI3063
aI1625
aI8941
aI30515
aI13601
aI72741
asS'd'
I8588
sa(dp17176
S'a'
I24389
sS's'
I17
sS'm'
(lp17177
I1
aI3
aI7
aI13
aI27
aI43
aI31
aI43
aI465
aI355
aI1063
aI2305
aI1425
aI11963
aI27327
aI53335
aI127517
asS'd'
I8589
sa(dp17178
S'a'
I24402
sS's'
I17
sS'm'
(lp17179
I1
aI3
aI1
aI3
aI21
aI17
aI53
aI171
aI269
aI783
aI349
aI1879
aI575
aI13537
aI16931
aI61171
aI23499
asS'd'
I8590
sa(dp17180
S'a'
I24414
sS's'
I17
sS'm'
(lp17181
I1
aI3
aI5
aI3
aI11
aI5
aI121
aI227
aI237
aI841
aI431
aI3209
aI3241
aI6071
aI23465
aI39533
aI102391
asS'd'
I8591
sa(dp17182
S'a'
I24420
sS's'
I17
sS'm'
(lp17183
I1
aI3
aI5
aI11
aI9
aI1
aI59
aI143
aI181
aI869
aI1859
aI1543
aI6419
aI13305
aI29075
aI28051
aI105799
asS'd'
I8592
sa(dp17184
S'a'
I24435
sS's'
I17
sS'm'
(lp17185
I1
aI1
aI7
aI13
aI31
aI1
aI105
aI169
aI67
aI693
aI1667
aI2181
aI4127
aI4605
aI3701
aI36467
aI19631
asS'd'
I8593
sa(dp17186
S'a'
I24437
sS's'
I17
sS'm'
(lp17187
I1
aI1
aI7
aI5
aI31
aI15
aI119
aI161
aI55
aI549
aI1195
aI4051
aI1923
aI2497
aI8289
aI60393
aI96181
asS'd'
I8594
sa(dp17188
S'a'
I24442
sS's'
I17
sS'm'
(lp17189
I1
aI1
aI3
aI3
aI5
aI43
aI13
aI123
aI469
aI603
aI2047
aI2347
aI815
aI3457
aI7503
aI25261
aI71951
asS'd'
I8595
sa(dp17190
S'a'
I24444
sS's'
I17
sS'm'
(lp17191
I1
aI1
aI7
aI3
aI13
aI25
aI85
aI141
aI497
aI405
aI957
aI1407
aI2075
aI12445
aI6675
aI9993
aI40429
asS'd'
I8596
sa(dp17192
S'a'
I24447
sS's'
I17
sS'm'
(lp17193
I1
aI1
aI5
aI13
aI29
aI43
aI99
aI113
aI307
aI1003
aI859
aI723
aI7513
aI12249
aI12653
aI57685
aI89551
asS'd'
I8597
sa(dp17194
S'a'
I24468
sS's'
I17
sS'm'
(lp17195
I1
aI3
aI7
aI3
aI11
aI3
aI9
aI141
aI501
aI113
aI69
aI2285
aI4525
aI9049
aI24765
aI11585
aI53787
asS'd'
I8598
sa(dp17196
S'a'
I24475
sS's'
I17
sS'm'
(lp17197
I1
aI1
aI3
aI1
aI25
aI41
aI103
aI159
aI215
aI871
aI77
aI1849
aI609
aI15877
aI32515
aI22931
aI11933
asS'd'
I8599
sa(dp17198
S'a'
I24484
sS's'
I17
sS'm'
(lp17199
I1
aI1
aI5
aI11
aI3
aI27
aI27
aI111
aI479
aI861
aI1041
aI3777
aI4443
aI3095
aI30379
aI6293
aI30823
asS'd'
I8600
sa(dp17200
S'a'
I24493
sS's'
I17
sS'm'
(lp17201
I1
aI3
aI5
aI5
aI27
aI45
aI9
aI25
aI451
aI845
aI1153
aI897
aI325
aI15679
aI30151
aI37695
aI54593
asS'd'
I8601
sa(dp17202
S'a'
I24494
sS's'
I17
sS'm'
(lp17203
I1
aI3
aI7
aI1
aI15
aI47
aI87
aI135
aI87
aI567
aI221
aI3173
aI769
aI8173
aI2957
aI51287
aI20961
asS'd'
I8602
sa(dp17204
S'a'
I24496
sS's'
I17
sS'm'
(lp17205
I1
aI3
aI1
aI9
aI3
aI33
aI1
aI71
aI147
aI983
aI1485
aI3531
aI213
aI2353
aI28269
aI49353
aI88343
asS'd'
I8603
sa(dp17206
S'a'
I24514
sS's'
I17
sS'm'
(lp17207
I1
aI1
aI3
aI11
aI11
aI63
aI109
aI255
aI35
aI127
aI1777
aI791
aI1379
aI9539
aI4915
aI21593
aI98901
asS'd'
I8604
sa(dp17208
S'a'
I24519
sS's'
I17
sS'm'
(lp17209
I1
aI1
aI7
aI5
aI3
aI47
aI93
aI219
aI381
aI963
aI359
aI2461
aI7629
aI2803
aI17345
aI54311
aI79057
asS'd'
I8605
sa(dp17210
S'a'
I24533
sS's'
I17
sS'm'
(lp17211
I1
aI3
aI5
aI13
aI13
aI21
aI1
aI65
aI455
aI203
aI29
aI3717
aI4495
aI1285
aI25289
aI38597
aI42431
asS'd'
I8606
sa(dp17212
S'a'
I24538
sS's'
I17
sS'm'
(lp17213
I1
aI1
aI3
aI3
aI27
aI57
aI7
aI171
aI65
aI469
aI1921
aI3855
aI1637
aI5517
aI14907
aI48239
aI117573
asS'd'
I8607
sa(dp17214
S'a'
I24559
sS's'
I17
sS'm'
(lp17215
I1
aI3
aI5
aI1
aI11
aI35
aI105
aI251
aI19
aI219
aI1191
aI2177
aI7885
aI8399
aI30527
aI61415
aI122215
asS'd'
I8608
sa(dp17216
S'a'
I24568
sS's'
I17
sS'm'
(lp17217
I1
aI3
aI5
aI5
aI21
aI25
aI59
aI193
aI509
aI147
aI1805
aI561
aI3505
aI9639
aI14221
aI31
aI99261
asS'd'
I8609
sa(dp17218
S'a'
I24573
sS's'
I17
sS'm'
(lp17219
I1
aI1
aI5
aI13
aI31
aI23
aI35
aI143
aI367
aI385
aI1335
aI2497
aI3573
aI8113
aI16661
aI16147
aI8763
asS'd'
I8610
sa(dp17220
S'a'
I24577
sS's'
I17
sS'm'
(lp17221
I1
aI1
aI7
aI13
aI15
aI27
aI35
aI15
aI7
aI539
aI633
aI1145
aI2267
aI11527
aI20975
aI16689
aI58227
asS'd'
I8611
sa(dp17222
S'a'
I24587
sS's'
I17
sS'm'
(lp17223
I1
aI1
aI1
aI15
aI9
aI11
aI51
aI121
aI381
aI331
aI1445
aI187
aI519
aI15827
aI27611
aI32891
aI113671
asS'd'
I8612
sa(dp17224
S'a'
I24592
sS's'
I17
sS'm'
(lp17225
I1
aI3
aI1
aI5
aI19
aI3
aI77
aI67
aI107
aI105
aI1025
aI3229
aI6869
aI5717
aI4227
aI28489
aI59759
asS'd'
I8613
sa(dp17226
S'a'
I24601
sS's'
I17
sS'm'
(lp17227
I1
aI1
aI5
aI15
aI25
aI23
aI7
aI25
aI103
aI733
aI525
aI453
aI6467
aI2901
aI7197
aI33267
aI68177
asS'd'
I8614
sa(dp17228
S'a'
I24602
sS's'
I17
sS'm'
(lp17229
I1
aI1
aI5
aI7
aI27
aI27
aI41
aI93
aI449
aI733
aI571
aI411
aI1709
aI9557
aI549
aI5925
aI24123
asS'd'
I8615
sa(dp17230
S'a'
I24608
sS's'
I17
sS'm'
(lp17231
I1
aI1
aI7
aI5
aI31
aI57
aI119
aI227
aI105
aI533
aI717
aI3357
aI2495
aI6467
aI7211
aI38169
aI44603
asS'd'
I8616
sa(dp17232
S'a'
I24645
sS's'
I17
sS'm'
(lp17233
I1
aI1
aI5
aI7
aI29
aI9
aI125
aI241
aI471
aI571
aI1271
aI2911
aI8087
aI5067
aI31139
aI39681
aI28579
asS'd'
I8617
sa(dp17234
S'a'
I24650
sS's'
I17
sS'm'
(lp17235
I1
aI3
aI5
aI11
aI25
aI53
aI109
aI35
aI183
aI109
aI1961
aI1681
aI7773
aI6935
aI28049
aI37279
aI96829
asS'd'
I8618
sa(dp17236
S'a'
I24657
sS's'
I17
sS'm'
(lp17237
I1
aI1
aI1
aI11
aI1
aI17
aI47
aI245
aI231
aI747
aI1395
aI1635
aI5129
aI3165
aI627
aI34463
aI38967
asS'd'
I8619
sa(dp17238
S'a'
I24664
sS's'
I17
sS'm'
(lp17239
I1
aI3
aI5
aI1
aI9
aI41
aI25
aI215
aI251
aI525
aI1399
aI3405
aI7399
aI11041
aI5599
aI51167
aI38697
asS'd'
I8620
sa(dp17240
S'a'
I24670
sS's'
I17
sS'm'
(lp17241
I1
aI3
aI3
aI13
aI11
aI15
aI121
aI95
aI139
aI611
aI633
aI3941
aI2619
aI15123
aI28535
aI64823
aI17527
asS'd'
I8621
sa(dp17242
S'a'
I24673
sS's'
I17
sS'm'
(lp17243
I1
aI3
aI7
aI13
aI21
aI53
aI65
aI175
aI81
aI5
aI699
aI1525
aI7397
aI2465
aI4479
aI58225
aI26387
asS'd'
I8622
sa(dp17244
S'a'
I24676
sS's'
I17
sS'm'
(lp17245
I1
aI1
aI5
aI7
aI9
aI31
aI31
aI149
aI359
aI613
aI397
aI153
aI4861
aI8195
aI22969
aI26003
aI124423
asS'd'
I8623
sa(dp17246
S'a'
I24680
sS's'
I17
sS'm'
(lp17247
I1
aI3
aI1
aI13
aI27
aI17
aI107
aI27
aI19
aI13
aI1481
aI573
aI7701
aI6273
aI30255
aI16125
aI11809
asS'd'
I8624
sa(dp17248
S'a'
I24686
sS's'
I17
sS'm'
(lp17249
I1
aI3
aI1
aI9
aI15
aI1
aI45
aI105
aI287
aI901
aI667
aI3197
aI3493
aI12259
aI1511
aI63361
aI94257
asS'd'
I8625
sa(dp17250
S'a'
I24700
sS's'
I17
sS'm'
(lp17251
I1
aI3
aI1
aI3
aI25
aI53
aI19
aI87
aI365
aI585
aI1569
aI1731
aI3747
aI11985
aI22673
aI17767
aI113779
asS'd'
I8626
sa(dp17252
S'a'
I24704
sS's'
I17
sS'm'
(lp17253
I1
aI3
aI3
aI9
aI7
aI21
aI103
aI201
aI501
aI149
aI1939
aI3111
aI4739
aI8389
aI27127
aI55889
aI54487
asS'd'
I8627
sa(dp17254
S'a'
I24714
sS's'
I17
sS'm'
(lp17255
I1
aI3
aI5
aI7
aI25
aI53
aI75
aI57
aI19
aI505
aI849
aI2631
aI6999
aI11269
aI24541
aI17695
aI97671
asS'd'
I8628
sa(dp17256
S'a'
I24728
sS's'
I17
sS'm'
(lp17257
I1
aI1
aI7
aI15
aI5
aI51
aI123
aI93
aI445
aI379
aI1729
aI2747
aI5821
aI10779
aI29335
aI57419
aI109339
asS'd'
I8629
sa(dp17258
S'a'
I24731
sS's'
I17
sS'm'
(lp17259
I1
aI1
aI7
aI3
aI7
aI57
aI117
aI65
aI297
aI891
aI487
aI1535
aI2361
aI10457
aI30759
aI34571
aI129949
asS'd'
I8630
sa(dp17260
S'a'
I24733
sS's'
I17
sS'm'
(lp17261
I1
aI3
aI5
aI5
aI17
aI51
aI27
aI103
aI55
aI925
aI947
aI1237
aI1629
aI12687
aI14775
aI49627
aI100939
asS'd'
I8631
sa(dp17262
S'a'
I24747
sS's'
I17
sS'm'
(lp17263
I1
aI3
aI3
aI15
aI1
aI11
aI75
aI177
aI399
aI55
aI1705
aI1165
aI7525
aI8909
aI13071
aI60703
aI11561
asS'd'
I8632
sa(dp17264
S'a'
I24749
sS's'
I17
sS'm'
(lp17265
I1
aI1
aI1
aI7
aI13
aI29
aI23
aI65
aI279
aI853
aI637
aI3947
aI4099
aI6465
aI7061
aI50417
aI35015
asS'd'
I8633
sa(dp17266
S'a'
I24750
sS's'
I17
sS'm'
(lp17267
I1
aI1
aI3
aI3
aI15
aI11
aI111
aI169
aI135
aI279
aI1941
aI3035
aI3027
aI6813
aI13363
aI20387
aI3257
asS'd'
I8634
sa(dp17268
S'a'
I24764
sS's'
I17
sS'm'
(lp17269
I1
aI3
aI3
aI11
aI3
aI5
aI95
aI181
aI405
aI313
aI39
aI1503
aI2443
aI3221
aI17021
aI23485
aI43909
asS'd'
I8635
sa(dp17270
S'a'
I24782
sS's'
I17
sS'm'
(lp17271
I1
aI1
aI1
aI3
aI17
aI63
aI27
aI247
aI441
aI533
aI449
aI3845
aI4021
aI14269
aI31477
aI7013
aI37473
asS'd'
I8636
sa(dp17272
S'a'
I24784
sS's'
I17
sS'm'
(lp17273
I1
aI1
aI5
aI13
aI29
aI39
aI41
aI95
aI417
aI21
aI685
aI609
aI5787
aI13145
aI32677
aI6121
aI50919
asS'd'
I8637
sa(dp17274
S'a'
I24793
sS's'
I17
sS'm'
(lp17275
I1
aI1
aI5
aI3
aI17
aI5
aI93
aI143
aI171
aI681
aI1143
aI2875
aI805
aI15823
aI29649
aI63327
aI12041
asS'd'
I8638
sa(dp17276
S'a'
I24794
sS's'
I17
sS'm'
(lp17277
I1
aI1
aI1
aI11
aI3
aI53
aI123
aI105
aI59
aI485
aI1799
aI2939
aI657
aI2485
aI29563
aI36221
aI89095
asS'd'
I8639
sa(dp17278
S'a'
I24810
sS's'
I17
sS'm'
(lp17279
I1
aI1
aI5
aI5
aI15
aI13
aI127
aI87
aI211
aI579
aI175
aI793
aI6895
aI9051
aI17681
aI28831
aI31783
asS'd'
I8640
sa(dp17280
S'a'
I24817
sS's'
I17
sS'm'
(lp17281
I1
aI3
aI7
aI5
aI11
aI37
aI9
aI219
aI453
aI697
aI139
aI335
aI6411
aI8495
aI4203
aI29065
aI114837
asS'd'
I8641
sa(dp17282
S'a'
I24820
sS's'
I17
sS'm'
(lp17283
I1
aI1
aI3
aI5
aI31
aI25
aI89
aI215
aI249
aI271
aI1731
aI3133
aI3947
aI10227
aI9679
aI51303
aI82833
asS'd'
I8642
sa(dp17284
S'a'
I24823
sS's'
I17
sS'm'
(lp17285
I1
aI3
aI5
aI1
aI31
aI15
aI7
aI131
aI369
aI757
aI1963
aI3223
aI35
aI13967
aI31807
aI5093
aI113743
asS'd'
I8643
sa(dp17286
S'a'
I24832
sS's'
I17
sS'm'
(lp17287
I1
aI1
aI7
aI3
aI15
aI23
aI21
aI173
aI295
aI929
aI1137
aI3943
aI1985
aI13015
aI8523
aI59117
aI127
asS'd'
I8644
sa(dp17288
S'a'
I24855
sS's'
I17
sS'm'
(lp17289
I1
aI3
aI7
aI1
aI31
aI1
aI115
aI229
aI345
aI859
aI1757
aI1835
aI7491
aI4545
aI1483
aI40149
aI122321
asS'd'
I8645
sa(dp17290
S'a'
I24859
sS's'
I17
sS'm'
(lp17291
I1
aI1
aI1
aI3
aI13
aI5
aI3
aI133
aI177
aI47
aI1515
aI17
aI5663
aI3185
aI2775
aI31389
aI37409
asS'd'
I8646
sa(dp17292
S'a'
I24862
sS's'
I17
sS'm'
(lp17293
I1
aI1
aI3
aI3
aI31
aI3
aI43
aI137
aI185
aI803
aI709
aI391
aI3513
aI8117
aI32593
aI46593
aI61037
asS'd'
I8647
sa(dp17294
S'a'
I24877
sS's'
I17
sS'm'
(lp17295
I1
aI1
aI1
aI7
aI29
aI27
aI13
aI35
aI61
aI961
aI777
aI2725
aI7379
aI7053
aI21781
aI60285
aI49221
asS'd'
I8648
sa(dp17296
S'a'
I24890
sS's'
I17
sS'm'
(lp17297
I1
aI3
aI7
aI15
aI7
aI7
aI15
aI123
aI109
aI97
aI361
aI791
aI4773
aI8111
aI4319
aI13981
aI92505
asS'd'
I8649
sa(dp17298
S'a'
I24900
sS's'
I17
sS'm'
(lp17299
I1
aI1
aI3
aI11
aI21
aI33
aI113
aI221
aI453
aI981
aI341
aI4041
aI5129
aI5981
aI11051
aI17243
aI19023
asS'd'
I8650
sa(dp17300
S'a'
I24904
sS's'
I17
sS'm'
(lp17301
I1
aI3
aI1
aI1
aI19
aI7
aI75
aI213
aI467
aI221
aI1829
aI1275
aI5729
aI6843
aI23855
aI44805
aI89269
asS'd'
I8651
sa(dp17302
S'a'
I24909
sS's'
I17
sS'm'
(lp17303
I1
aI1
aI3
aI7
aI5
aI29
aI39
aI125
aI147
aI329
aI1485
aI2793
aI2329
aI14979
aI18395
aI37951
aI58699
asS'd'
I8652
sa(dp17304
S'a'
I24910
sS's'
I17
sS'm'
(lp17305
I1
aI3
aI1
aI3
aI11
aI37
aI117
aI189
aI103
aI381
aI39
aI31
aI5205
aI5601
aI17127
aI49073
aI121417
asS'd'
I8653
sa(dp17306
S'a'
I24915
sS's'
I17
sS'm'
(lp17307
I1
aI3
aI3
aI13
aI23
aI49
aI57
aI187
aI441
aI189
aI349
aI2559
aI3313
aI1321
aI7731
aI57309
aI80195
asS'd'
I8654
sa(dp17308
S'a'
I24922
sS's'
I17
sS'm'
(lp17309
I1
aI3
aI7
aI1
aI17
aI9
aI21
aI15
aI447
aI333
aI959
aI3471
aI5301
aI8573
aI9761
aI23183
aI57997
asS'd'
I8655
sa(dp17310
S'a'
I24937
sS's'
I17
sS'm'
(lp17311
I1
aI3
aI1
aI9
aI19
aI1
aI101
aI71
aI325
aI309
aI85
aI2097
aI8003
aI12249
aI1887
aI2097
aI68375
asS'd'
I8656
sa(dp17312
S'a'
I24945
sS's'
I17
sS'm'
(lp17313
I1
aI1
aI7
aI7
aI11
aI39
aI85
aI241
aI293
aI205
aI387
aI1539
aI6583
aI1395
aI8869
aI48843
aI49983
asS'd'
I8657
sa(dp17314
S'a'
I24962
sS's'
I17
sS'm'
(lp17315
I1
aI3
aI7
aI13
aI11
aI23
aI83
aI125
aI55
aI429
aI169
aI1893
aI4657
aI643
aI3405
aI9943
aI128753
asS'd'
I8658
sa(dp17316
S'a'
I24964
sS's'
I17
sS'm'
(lp17317
I1
aI1
aI3
aI11
aI19
aI43
aI13
aI171
aI495
aI117
aI437
aI3697
aI6723
aI6199
aI1859
aI39637
aI111499
asS'd'
I8659
sa(dp17318
S'a'
I24991
sS's'
I17
sS'm'
(lp17319
I1
aI1
aI1
aI5
aI1
aI31
aI83
aI199
aI129
aI941
aI1637
aI1997
aI5011
aI14957
aI32427
aI60797
aI4989
asS'd'
I8660
sa(dp17320
S'a'
I24992
sS's'
I17
sS'm'
(lp17321
I1
aI3
aI3
aI3
aI5
aI61
aI33
aI225
aI315
aI157
aI1709
aI807
aI7809
aI11063
aI319
aI20901
aI73599
asS'd'
I8661
sa(dp17322
S'a'
I24995
sS's'
I17
sS'm'
(lp17323
I1
aI1
aI7
aI3
aI27
aI3
aI1
aI173
aI125
aI769
aI1203
aI3357
aI4899
aI13115
aI7081
aI42459
aI18525
asS'd'
I8662
sa(dp17324
S'a'
I25001
sS's'
I17
sS'm'
(lp17325
I1
aI1
aI7
aI9
aI9
aI27
aI43
aI115
aI229
aI867
aI661
aI1825
aI2883
aI4285
aI22393
aI65141
aI24699
asS'd'
I8663
sa(dp17326
S'a'
I25009
sS's'
I17
sS'm'
(lp17327
I1
aI1
aI3
aI7
aI5
aI9
aI93
aI47
aI33
aI823
aI309
aI2977
aI5791
aI9177
aI27645
aI35683
aI57455
asS'd'
I8664
sa(dp17328
S'a'
I25019
sS's'
I17
sS'm'
(lp17329
I1
aI1
aI5
aI7
aI9
aI53
aI9
aI77
aI499
aI1023
aI917
aI209
aI7311
aI249
aI773
aI18303
aI41447
asS'd'
I8665
sa(dp17330
S'a'
I25021
sS's'
I17
sS'm'
(lp17331
I1
aI1
aI3
aI5
aI7
aI9
aI33
aI73
aI325
aI369
aI1657
aI2257
aI2893
aI13911
aI10797
aI21055
aI103511
asS'd'
I8666
sa(dp17332
S'a'
I25029
sS's'
I17
sS'm'
(lp17333
I1
aI3
aI1
aI3
aI21
aI31
aI125
aI29
aI149
aI513
aI979
aI2271
aI989
aI9541
aI4179
aI13215
aI71369
asS'd'
I8667
sa(dp17334
S'a'
I25034
sS's'
I17
sS'm'
(lp17335
I1
aI1
aI7
aI7
aI19
aI41
aI39
aI165
aI59
aI79
aI137
aI3479
aI3389
aI6635
aI21467
aI51073
aI20765
asS'd'
I8668
sa(dp17336
S'a'
I25036
sS's'
I17
sS'm'
(lp17337
I1
aI3
aI3
aI5
aI7
aI13
aI109
aI53
aI335
aI627
aI339
aI3825
aI287
aI6077
aI11319
aI2377
aI112693
asS'd'
I8669
sa(dp17338
S'a'
I25039
sS's'
I17
sS'm'
(lp17339
I1
aI3
aI3
aI1
aI3
aI57
aI9
aI47
aI437
aI717
aI563
aI1219
aI6191
aI9081
aI21533
aI2651
aI17275
asS'd'
I8670
sa(dp17340
S'a'
I25057
sS's'
I17
sS'm'
(lp17341
I1
aI1
aI1
aI5
aI21
aI9
aI109
aI109
aI339
aI947
aI1699
aI1487
aI6477
aI12601
aI12327
aI39427
aI80937
asS'd'
I8671
sa(dp17342
S'a'
I25063
sS's'
I17
sS'm'
(lp17343
I1
aI1
aI7
aI9
aI1
aI5
aI55
aI43
aI95
aI733
aI1151
aI3657
aI2119
aI11947
aI21279
aI21581
aI22053
asS'd'
I8672
sa(dp17344
S'a'
I25064
sS's'
I17
sS'm'
(lp17345
I1
aI3
aI5
aI11
aI7
aI9
aI97
aI149
aI55
aI523
aI1911
aI1389
aI5343
aI5533
aI15439
aI65313
aI73421
asS'd'
I8673
sa(dp17346
S'a'
I25075
sS's'
I17
sS'm'
(lp17347
I1
aI1
aI3
aI7
aI19
aI15
aI119
aI141
aI57
aI243
aI423
aI981
aI1407
aI12633
aI20455
aI53069
aI98593
asS'd'
I8674
sa(dp17348
S'a'
I25077
sS's'
I17
sS'm'
(lp17349
I1
aI1
aI3
aI3
aI15
aI57
aI71
aI203
aI15
aI133
aI601
aI805
aI2821
aI11623
aI147
aI4333
aI97681
asS'd'
I8675
sa(dp17350
S'a'
I25084
sS's'
I17
sS'm'
(lp17351
I1
aI1
aI5
aI7
aI17
aI61
aI15
aI251
aI53
aI699
aI105
aI1195
aI3979
aI41
aI9077
aI5145
aI80057
asS'd'
I8676
sa(dp17352
S'a'
I25088
sS's'
I17
sS'm'
(lp17353
I1
aI1
aI5
aI15
aI29
aI33
aI53
aI19
aI41
aI471
aI1143
aI65
aI5833
aI8417
aI17263
aI35859
aI45035
asS'd'
I8677
sa(dp17354
S'a'
I25091
sS's'
I17
sS'm'
(lp17355
I1
aI1
aI1
aI1
aI15
aI51
aI73
aI131
aI181
aI147
aI1863
aI3777
aI1749
aI10135
aI11591
aI12395
aI85163
asS'd'
I8678
sa(dp17356
S'a'
I25105
sS's'
I17
sS'm'
(lp17357
I1
aI3
aI1
aI9
aI23
aI63
aI83
aI199
aI87
aI499
aI2025
aI863
aI4665
aI3941
aI17647
aI52463
aI108615
asS'd'
I8679
sa(dp17358
S'a'
I25134
sS's'
I17
sS'm'
(lp17359
I1
aI3
aI5
aI7
aI11
aI39
aI65
aI161
aI367
aI593
aI699
aI1807
aI7217
aI5221
aI22093
aI44933
aI6201
asS'd'
I8680
sa(dp17360
S'a'
I25165
sS's'
I17
sS'm'
(lp17361
I1
aI1
aI7
aI13
aI9
aI41
aI35
aI77
aI353
aI291
aI1267
aI3923
aI5397
aI15401
aI30317
aI14945
aI8715
asS'd'
I8681
sa(dp17362
S'a'
I25183
sS's'
I17
sS'm'
(lp17363
I1
aI3
aI1
aI15
aI11
aI3
aI29
aI25
aI505
aI945
aI1425
aI2297
aI1133
aI4675
aI8069
aI55115
aI114177
asS'd'
I8682
sa(dp17364
S'a'
I25184
sS's'
I17
sS'm'
(lp17365
I1
aI3
aI1
aI5
aI27
aI63
aI25
aI7
aI5
aI399
aI473
aI1325
aI7391
aI5953
aI27755
aI65407
aI89435
asS'd'
I8683
sa(dp17366
S'a'
I25202
sS's'
I17
sS'm'
(lp17367
I1
aI3
aI3
aI13
aI21
aI61
aI5
aI119
aI23
aI999
aI849
aI1225
aI3077
aI821
aI12059
aI43223
aI45427
asS'd'
I8684
sa(dp17368
S'a'
I25204
sS's'
I17
sS'm'
(lp17369
I1
aI3
aI7
aI13
aI1
aI5
aI93
aI173
aI181
aI453
aI1449
aI3823
aI1713
aI14737
aI8891
aI43643
aI1983
asS'd'
I8685
sa(dp17370
S'a'
I25211
sS's'
I17
sS'm'
(lp17371
I1
aI1
aI3
aI3
aI29
aI53
aI31
aI163
aI321
aI539
aI1283
aI429
aI3449
aI15617
aI4761
aI21187
aI120725
asS'd'
I8686
sa(dp17372
S'a'
I25223
sS's'
I17
sS'm'
(lp17373
I1
aI1
aI1
aI1
aI13
aI27
aI49
aI37
aI33
aI631
aI375
aI425
aI2465
aI8773
aI2777
aI2115
aI35633
asS'd'
I8687
sa(dp17374
S'a'
I25224
sS's'
I17
sS'm'
(lp17375
I1
aI3
aI5
aI3
aI25
aI25
aI27
aI201
aI63
aI259
aI1571
aI1143
aI2325
aI6773
aI11941
aI28897
aI19719
asS'd'
I8688
sa(dp17376
S'a'
I25235
sS's'
I17
sS'm'
(lp17377
I1
aI1
aI3
aI5
aI11
aI39
aI59
aI203
aI37
aI899
aI559
aI2599
aI4397
aI12159
aI29579
aI51251
aI83213
asS'd'
I8689
sa(dp17378
S'a'
I25241
sS's'
I17
sS'm'
(lp17379
I1
aI1
aI1
aI7
aI9
aI19
aI63
aI169
aI257
aI921
aI381
aI3605
aI3479
aI1739
aI26599
aI20599
aI29617
asS'd'
I8690
sa(dp17380
S'a'
I25253
sS's'
I17
sS'm'
(lp17381
I1
aI1
aI1
aI9
aI7
aI29
aI123
aI35
aI419
aI963
aI855
aI1903
aI6199
aI2727
aI29811
aI49279
aI101673
asS'd'
I8691
sa(dp17382
S'a'
I25258
sS's'
I17
sS'm'
(lp17383
I1
aI3
aI5
aI11
aI29
aI23
aI73
aI13
aI467
aI935
aI181
aI3837
aI8117
aI11501
aI18361
aI26803
aI99471
asS'd'
I8692
sa(dp17384
S'a'
I25277
sS's'
I17
sS'm'
(lp17385
I1
aI1
aI1
aI5
aI31
aI41
aI109
aI45
aI115
aI113
aI1893
aI727
aI2453
aI13463
aI22339
aI13495
aI11473
asS'd'
I8693
sa(dp17386
S'a'
I25278
sS's'
I17
sS'm'
(lp17387
I1
aI1
aI5
aI9
aI5
aI31
aI109
aI145
aI511
aI243
aI57
aI2219
aI1601
aI1821
aI12787
aI48239
aI89645
asS'd'
I8694
sa(dp17388
S'a'
I25280
sS's'
I17
sS'm'
(lp17389
I1
aI3
aI1
aI7
aI19
aI41
aI25
aI57
aI45
aI489
aI1531
aI3959
aI2007
aI14247
aI13445
aI1991
aI114977
asS'd'
I8695
sa(dp17390
S'a'
I25290
sS's'
I17
sS'm'
(lp17391
I1
aI3
aI7
aI15
aI7
aI17
aI107
aI27
aI249
aI207
aI183
aI2483
aI5817
aI8927
aI10715
aI63631
aI51947
asS'd'
I8696
sa(dp17392
S'a'
I25295
sS's'
I17
sS'm'
(lp17393
I1
aI3
aI1
aI3
aI13
aI21
aI57
aI113
aI171
aI885
aI1335
aI783
aI7575
aI4443
aI19497
aI13827
aI130727
asS'd'
I8697
sa(dp17394
S'a'
I25300
sS's'
I17
sS'm'
(lp17395
I1
aI1
aI5
aI7
aI19
aI33
aI95
aI13
aI387
aI297
aI1597
aI767
aI7543
aI16063
aI10367
aI51683
aI119811
asS'd'
I8698
sa(dp17396
S'a'
I25307
sS's'
I17
sS'm'
(lp17397
I1
aI3
aI7
aI9
aI27
aI57
aI111
aI209
aI305
aI139
aI179
aI25
aI2295
aI2593
aI31361
aI23677
aI74501
asS'd'
I8699
sa(dp17398
S'a'
I25319
sS's'
I17
sS'm'
(lp17399
I1
aI3
aI7
aI3
aI21
aI63
aI97
aI189
aI3
aI693
aI209
aI2227
aI7169
aI9
aI32575
aI61521
aI115155
asS'd'
I8700
sa(dp17400
S'a'
I25323
sS's'
I17
sS'm'
(lp17401
I1
aI1
aI1
aI11
aI13
aI21
aI125
aI249
aI193
aI895
aI139
aI1207
aI5941
aI5821
aI6623
aI7753
aI80939
asS'd'
I8701
sa(dp17402
S'a'
I25326
sS's'
I17
sS'm'
(lp17403
I1
aI3
aI5
aI5
aI11
aI49
aI17
aI21
aI423
aI497
aI835
aI539
aI6195
aI12783
aI1271
aI20069
aI2657
asS'd'
I8702
sa(dp17404
S'a'
I25333
sS's'
I17
sS'm'
(lp17405
I1
aI1
aI7
aI15
aI13
aI39
aI83
aI191
aI77
aI95
aI661
aI3627
aI1853
aI1349
aI23109
aI43583
aI104121
asS'd'
I8703
sa(dp17406
S'a'
I25334
sS's'
I17
sS'm'
(lp17407
I1
aI3
aI1
aI15
aI31
aI15
aI71
aI255
aI489
aI91
aI351
aI367
aI309
aI6275
aI18325
aI51231
aI52159
asS'd'
I8704
sa(dp17408
S'a'
I25337
sS's'
I17
sS'm'
(lp17409
I1
aI1
aI7
aI13
aI21
aI49
aI37
aI135
aI355
aI421
aI507
aI2563
aI4955
aI4095
aI1933
aI29517
aI119699
asS'd'
I8705
sa(dp17410
S'a'
I25348
sS's'
I17
sS'm'
(lp17411
I1
aI1
aI1
aI1
aI27
aI41
aI15
aI161
aI475
aI635
aI863
aI3773
aI6015
aI6197
aI24261
aI26271
aI42375
asS'd'
I8706
sa(dp17412
S'a'
I25351
sS's'
I17
sS'm'
(lp17413
I1
aI1
aI7
aI13
aI25
aI7
aI23
aI185
aI129
aI597
aI1561
aI3003
aI2879
aI15187
aI4913
aI24589
aI12927
asS'd'
I8707
sa(dp17414
S'a'
I25372
sS's'
I17
sS'm'
(lp17415
I1
aI1
aI3
aI3
aI9
aI23
aI49
aI233
aI345
aI83
aI823
aI2627
aI5019
aI2365
aI23755
aI9855
aI48515
asS'd'
I8708
sa(dp17416
S'a'
I25381
sS's'
I17
sS'm'
(lp17417
I1
aI3
aI1
aI1
aI11
aI7
aI117
aI213
aI27
aI923
aI375
aI2597
aI8173
aI8935
aI16487
aI49283
aI104569
asS'd'
I8709
sa(dp17418
S'a'
I25403
sS's'
I17
sS'm'
(lp17419
I1
aI1
aI7
aI7
aI23
aI13
aI61
aI131
aI313
aI883
aI495
aI1105
aI6207
aI1473
aI21655
aI51883
aI403
asS'd'
I8710
sa(dp17420
S'a'
I25406
sS's'
I17
sS'm'
(lp17421
I1
aI3
aI3
aI1
aI25
aI5
aI5
aI159
aI243
aI929
aI1429
aI1151
aI5043
aI11551
aI21231
aI38767
aI105299
asS'd'
I8711
sa(dp17422
S'a'
I25431
sS's'
I17
sS'm'
(lp17423
I1
aI3
aI7
aI7
aI15
aI37
aI49
aI219
aI67
aI147
aI873
aI2391
aI455
aI9565
aI8977
aI64759
aI40347
asS'd'
I8712
sa(dp17424
S'a'
I25437
sS's'
I17
sS'm'
(lp17425
I1
aI1
aI1
aI13
aI21
aI13
aI13
aI243
aI303
aI333
aI187
aI3591
aI871
aI2501
aI30853
aI5247
aI48855
asS'd'
I8713
sa(dp17426
S'a'
I25453
sS's'
I17
sS'm'
(lp17427
I1
aI1
aI1
aI5
aI1
aI5
aI127
aI249
aI23
aI79
aI789
aI3507
aI8119
aI5025
aI26545
aI54009
aI100633
asS'd'
I8714
sa(dp17428
S'a'
I25459
sS's'
I17
sS'm'
(lp17429
I1
aI3
aI3
aI11
aI3
aI31
aI27
aI115
aI423
aI309
aI1805
aI169
aI789
aI4081
aI28139
aI35355
aI47991
asS'd'
I8715
sa(dp17430
S'a'
I25462
sS's'
I17
sS'm'
(lp17431
I1
aI3
aI1
aI5
aI19
aI13
aI43
aI165
aI165
aI241
aI309
aI1703
aI7631
aI5899
aI12041
aI21235
aI97045
asS'd'
I8716
sa(dp17432
S'a'
I25466
sS's'
I17
sS'm'
(lp17433
I1
aI3
aI1
aI13
aI15
aI49
aI29
aI199
aI93
aI611
aI77
aI2681
aI191
aI10215
aI8115
aI11895
aI108687
asS'd'
I8717
sa(dp17434
S'a'
I25477
sS's'
I17
sS'm'
(lp17435
I1
aI1
aI3
aI3
aI13
aI45
aI15
aI151
aI345
aI111
aI1829
aI1357
aI6317
aI5239
aI26193
aI46763
aI73101
asS'd'
I8718
sa(dp17436
S'a'
I25484
sS's'
I17
sS'm'
(lp17437
I1
aI3
aI7
aI3
aI1
aI19
aI119
aI63
aI23
aI759
aI173
aI307
aI967
aI2731
aI9353
aI14479
aI119
asS'd'
I8719
sa(dp17438
S'a'
I25495
sS's'
I17
sS'm'
(lp17439
I1
aI3
aI5
aI15
aI5
aI21
aI127
aI21
aI419
aI575
aI991
aI3465
aI7365
aI5711
aI30657
aI43513
aI22447
asS'd'
I8720
sa(dp17440
S'a'
I25501
sS's'
I17
sS'm'
(lp17441
I1
aI3
aI7
aI1
aI19
aI5
aI49
aI7
aI45
aI963
aI1755
aI3745
aI4061
aI4619
aI9089
aI59953
aI100265
asS'd'
I8721
sa(dp17442
S'a'
I25506
sS's'
I17
sS'm'
(lp17443
I1
aI1
aI1
aI3
aI25
aI53
aI97
aI97
aI347
aI749
aI823
aI1499
aI8151
aI9957
aI731
aI22317
aI121623
asS'd'
I8722
sa(dp17444
S'a'
I25511
sS's'
I17
sS'm'
(lp17445
I1
aI3
aI5
aI5
aI19
aI3
aI121
aI127
aI313
aI457
aI1737
aI4065
aI5295
aI7957
aI16373
aI62085
aI5711
asS'd'
I8723
sa(dp17446
S'a'
I25515
sS's'
I17
sS'm'
(lp17447
I1
aI1
aI7
aI13
aI7
aI37
aI97
aI43
aI179
aI837
aI161
aI477
aI5095
aI4985
aI111
aI58743
aI24049
asS'd'
I8724
sa(dp17448
S'a'
I25525
sS's'
I17
sS'm'
(lp17449
I1
aI3
aI1
aI13
aI27
aI13
aI91
aI241
aI339
aI235
aI111
aI369
aI3361
aI15105
aI11097
aI23955
aI53561
asS'd'
I8725
sa(dp17450
S'a'
I25529
sS's'
I17
sS'm'
(lp17451
I1
aI3
aI5
aI3
aI9
aI17
aI103
aI133
aI309
aI683
aI71
aI3329
aI7229
aI8763
aI4165
aI9649
aI8529
asS'd'
I8726
sa(dp17452
S'a'
I25532
sS's'
I17
sS'm'
(lp17453
I1
aI3
aI5
aI13
aI29
aI55
aI29
aI205
aI433
aI1007
aI1173
aI731
aI5653
aI89
aI18447
aI37911
aI65603
asS'd'
I8727
sa(dp17454
S'a'
I25538
sS's'
I17
sS'm'
(lp17455
I1
aI3
aI5
aI1
aI15
aI1
aI7
aI195
aI397
aI877
aI1433
aI3487
aI1581
aI1539
aI3361
aI7453
aI46451
asS'd'
I8728
sa(dp17456
S'a'
I25549
sS's'
I17
sS'm'
(lp17457
I1
aI1
aI5
aI13
aI23
aI1
aI47
aI245
aI19
aI859
aI681
aI2971
aI2531
aI11393
aI32765
aI4595
aI45213
asS'd'
I8729
sa(dp17458
S'a'
I25552
sS's'
I17
sS'm'
(lp17459
I1
aI3
aI1
aI3
aI1
aI11
aI85
aI185
aI467
aI413
aI25
aI3677
aI881
aI1791
aI14655
aI44811
aI50819
asS'd'
I8730
sa(dp17460
S'a'
I25564
sS's'
I17
sS'm'
(lp17461
I1
aI3
aI5
aI9
aI9
aI21
aI65
aI99
aI441
aI215
aI1453
aI2873
aI5883
aI485
aI20883
aI1303
aI11837
asS'd'
I8731
sa(dp17462
S'a'
I25567
sS's'
I17
sS'm'
(lp17463
I1
aI3
aI3
aI5
aI9
aI37
aI87
aI211
aI247
aI535
aI1163
aI1785
aI4219
aI12559
aI17419
aI48201
aI21725
asS'd'
I8732
sa(dp17464
S'a'
I25574
sS's'
I17
sS'm'
(lp17465
I1
aI1
aI1
aI11
aI29
aI11
aI9
aI215
aI375
aI601
aI627
aI2641
aI6961
aI6175
aI10995
aI49299
aI102891
asS'd'
I8733
sa(dp17466
S'a'
I25577
sS's'
I17
sS'm'
(lp17467
I1
aI3
aI1
aI3
aI7
aI7
aI23
aI139
aI89
aI1005
aI1815
aI947
aI1507
aI10349
aI35
aI43595
aI104697
asS'd'
I8734
sa(dp17468
S'a'
I25583
sS's'
I17
sS'm'
(lp17469
I1
aI1
aI5
aI13
aI29
aI47
aI77
aI255
aI341
aI333
aI1211
aI3473
aI1303
aI11237
aI28371
aI43283
aI77617
asS'd'
I8735
sa(dp17470
S'a'
I25588
sS's'
I17
sS'm'
(lp17471
I1
aI3
aI3
aI13
aI27
aI17
aI73
aI95
aI227
aI241
aI1369
aI833
aI6683
aI2193
aI309
aI64249
aI6731
asS'd'
I8736
sa(dp17472
S'a'
I25603
sS's'
I17
sS'm'
(lp17473
I1
aI3
aI3
aI3
aI15
aI29
aI45
aI209
aI401
aI725
aI1123
aI1659
aI6099
aI15941
aI5797
aI30563
aI119385
asS'd'
I8737
sa(dp17474
S'a'
I25610
sS's'
I17
sS'm'
(lp17475
I1
aI1
aI1
aI1
aI7
aI55
aI95
aI151
aI351
aI373
aI1131
aI2357
aI7535
aI3899
aI19047
aI17879
aI34623
asS'd'
I8738
sa(dp17476
S'a'
I25615
sS's'
I17
sS'm'
(lp17477
I1
aI3
aI1
aI5
aI31
aI5
aI33
aI97
aI477
aI399
aI1255
aI1073
aI1513
aI11651
aI2951
aI31351
aI102635
asS'd'
I8739
sa(dp17478
S'a'
I25624
sS's'
I17
sS'm'
(lp17479
I1
aI1
aI3
aI13
aI17
aI63
aI51
aI209
aI57
aI87
aI977
aI3663
aI6717
aI15441
aI10709
aI607
aI48297
asS'd'
I8740
sa(dp17480
S'a'
I25636
sS's'
I17
sS'm'
(lp17481
I1
aI1
aI5
aI1
aI9
aI29
aI1
aI105
aI343
aI19
aI977
aI3401
aI3873
aI4259
aI23057
aI13071
aI105771
asS'd'
I8741
sa(dp17482
S'a'
I25639
sS's'
I17
sS'm'
(lp17483
I1
aI1
aI1
aI5
aI1
aI33
aI59
aI17
aI115
aI225
aI853
aI3295
aI965
aI12547
aI26971
aI50097
aI54999
asS'd'
I8742
sa(dp17484
S'a'
I25643
sS's'
I17
sS'm'
(lp17485
I1
aI3
aI3
aI13
aI1
aI51
aI29
aI19
aI245
aI781
aI493
aI1121
aI2937
aI4177
aI3267
aI47463
aI101195
asS'd'
I8743
sa(dp17486
S'a'
I25645
sS's'
I17
sS'm'
(lp17487
I1
aI3
aI7
aI5
aI3
aI51
aI25
aI131
aI451
aI997
aI1809
aI1583
aI355
aI15383
aI28159
aI39141
aI109379
asS'd'
I8744
sa(dp17488
S'a'
I25648
sS's'
I17
sS'm'
(lp17489
I1
aI1
aI5
aI7
aI3
aI19
aI75
aI103
aI401
aI115
aI1627
aI423
aI2485
aI7281
aI6177
aI54677
aI31499
asS'd'
I8745
sa(dp17490
S'a'
I25671
sS's'
I17
sS'm'
(lp17491
I1
aI1
aI1
aI11
aI23
aI7
aI57
aI121
aI5
aI921
aI1191
aI1779
aI1979
aI3427
aI25617
aI19423
aI73835
asS'd'
I8746
sa(dp17492
S'a'
I25678
sS's'
I17
sS'm'
(lp17493
I1
aI3
aI3
aI11
aI17
aI51
aI15
aI163
aI265
aI665
aI1399
aI1977
aI3097
aI7109
aI14741
aI24291
aI79239
asS'd'
I8747
sa(dp17494
S'a'
I25689
sS's'
I17
sS'm'
(lp17495
I1
aI1
aI7
aI3
aI25
aI61
aI69
aI77
aI341
aI23
aI713
aI2879
aI8075
aI14855
aI9691
aI58241
aI113277
asS'd'
I8748
sa(dp17496
S'a'
I25708
sS's'
I17
sS'm'
(lp17497
I1
aI3
aI7
aI9
aI27
aI43
aI95
aI11
aI239
aI445
aI951
aI3869
aI1049
aI6493
aI9569
aI9285
aI29183
asS'd'
I8749
sa(dp17498
S'a'
I25716
sS's'
I17
sS'm'
(lp17499
I1
aI1
aI3
aI1
aI1
aI23
aI27
aI101
aI337
aI171
aI1977
aI3181
aI2693
aI8591
aI32309
aI24909
aI106535
asS'd'
I8750
sa(dp17500
S'a'
I25725
sS's'
I17
sS'm'
(lp17501
I1
aI1
aI1
aI7
aI23
aI59
aI79
aI115
aI49
aI351
aI871
aI1209
aI1045
aI5985
aI28427
aI23047
aI113571
asS'd'
I8751
sa(dp17502
S'a'
I25730
sS's'
I17
sS'm'
(lp17503
I1
aI1
aI7
aI13
aI27
aI3
aI35
aI7
aI319
aI503
aI977
aI3747
aI4859
aI16315
aI30375
aI25999
aI24341
asS'd'
I8752
sa(dp17504
S'a'
I25739
sS's'
I17
sS'm'
(lp17505
I1
aI3
aI3
aI7
aI23
aI43
aI67
aI21
aI399
aI349
aI1541
aI2991
aI5781
aI14501
aI5609
aI65093
aI12789
asS'd'
I8753
sa(dp17506
S'a'
I25747
sS's'
I17
sS'm'
(lp17507
I1
aI3
aI1
aI11
aI5
aI21
aI17
aI157
aI311
aI663
aI469
aI4033
aI1557
aI7569
aI31163
aI14079
aI127771
asS'd'
I8754
sa(dp17508
S'a'
I25753
sS's'
I17
sS'm'
(lp17509
I1
aI1
aI7
aI15
aI15
aI31
aI15
aI183
aI365
aI35
aI1433
aI2793
aI6685
aI10565
aI30409
aI46815
aI14173
asS'd'
I8755
sa(dp17510
S'a'
I25790
sS's'
I17
sS'm'
(lp17511
I1
aI1
aI7
aI7
aI7
aI45
aI61
aI163
aI99
aI353
aI1535
aI3185
aI4023
aI7999
aI26173
aI12675
aI98073
asS'd'
I8756
sa(dp17512
S'a'
I25797
sS's'
I17
sS'm'
(lp17513
I1
aI1
aI5
aI13
aI1
aI11
aI107
aI41
aI171
aI773
aI1513
aI883
aI2117
aI14449
aI32323
aI58271
aI97173
asS'd'
I8757
sa(dp17514
S'a'
I25804
sS's'
I17
sS'm'
(lp17515
I1
aI1
aI3
aI13
aI27
aI15
aI123
aI247
aI281
aI851
aI233
aI1173
aI6863
aI14805
aI12401
aI30729
aI104127
asS'd'
I8758
sa(dp17516
S'a'
I25809
sS's'
I17
sS'm'
(lp17517
I1
aI1
aI7
aI11
aI25
aI9
aI97
aI215
aI217
aI51
aI1865
aI3897
aI725
aI4779
aI21661
aI11853
aI72225
asS'd'
I8759
sa(dp17518
S'a'
I25816
sS's'
I17
sS'm'
(lp17519
I1
aI1
aI5
aI3
aI5
aI31
aI125
aI81
aI367
aI705
aI325
aI519
aI3879
aI5607
aI3247
aI7149
aI33177
asS'd'
I8760
sa(dp17520
S'a'
I25825
sS's'
I17
sS'm'
(lp17521
I1
aI3
aI3
aI7
aI17
aI17
aI19
aI113
aI331
aI277
aI317
aI1893
aI1287
aI8965
aI27523
aI61355
aI45331
asS'd'
I8761
sa(dp17522
S'a'
I25831
sS's'
I17
sS'm'
(lp17523
I1
aI3
aI7
aI9
aI27
aI15
aI87
aI21
aI343
aI479
aI11
aI2945
aI1235
aI1591
aI28195
aI40559
aI42773
asS'd'
I8762
sa(dp17524
S'a'
I25845
sS's'
I17
sS'm'
(lp17525
I1
aI3
aI3
aI13
aI1
aI45
aI115
aI41
aI263
aI569
aI71
aI4051
aI739
aI1031
aI19213
aI23961
aI110767
asS'd'
I8763
sa(dp17526
S'a'
I25846
sS's'
I17
sS'm'
(lp17527
I1
aI3
aI7
aI1
aI9
aI41
aI21
aI131
aI3
aI617
aI191
aI4051
aI2445
aI13451
aI11889
aI25075
aI82631
asS'd'
I8764
sa(dp17528
S'a'
I25867
sS's'
I17
sS'm'
(lp17529
I1
aI3
aI3
aI15
aI7
aI55
aI65
aI67
aI443
aI1023
aI1445
aI1467
aI3907
aI11449
aI2247
aI65085
aI102161
asS'd'
I8765
sa(dp17530
S'a'
I25870
sS's'
I17
sS'm'
(lp17531
I1
aI3
aI5
aI15
aI19
aI27
aI97
aI181
aI51
aI591
aI99
aI1443
aI4927
aI9809
aI29693
aI44293
aI29369
asS'd'
I8766
sa(dp17532
S'a'
I25897
sS's'
I17
sS'm'
(lp17533
I1
aI1
aI7
aI7
aI17
aI59
aI69
aI163
aI37
aI171
aI107
aI2581
aI3567
aI9455
aI19707
aI6329
aI27755
asS'd'
I8767
sa(dp17534
S'a'
I25908
sS's'
I17
sS'm'
(lp17535
I1
aI1
aI1
aI11
aI15
aI17
aI83
aI223
aI183
aI861
aI1047
aI3739
aI3509
aI5571
aI28259
aI42781
aI130657
asS'd'
I8768
sa(dp17536
S'a'
I25937
sS's'
I17
sS'm'
(lp17537
I1
aI3
aI3
aI7
aI15
aI11
aI33
aI115
aI297
aI841
aI1629
aI1559
aI2261
aI11763
aI22255
aI63819
aI55831
asS'd'
I8769
sa(dp17538
S'a'
I25940
sS's'
I17
sS'm'
(lp17539
I1
aI3
aI3
aI5
aI19
aI49
aI17
aI251
aI507
aI251
aI805
aI1041
aI3947
aI2219
aI19977
aI65449
aI25031
asS'd'
I8770
sa(dp17540
S'a'
I25944
sS's'
I17
sS'm'
(lp17541
I1
aI1
aI1
aI11
aI3
aI7
aI81
aI17
aI219
aI729
aI949
aI3257
aI6495
aI4701
aI2181
aI7009
aI106465
asS'd'
I8771
sa(dp17542
S'a'
I25949
sS's'
I17
sS'm'
(lp17543
I1
aI3
aI7
aI5
aI27
aI35
aI15
aI83
aI43
aI1013
aI1427
aI1943
aI7555
aI6613
aI26879
aI42685
aI22071
asS'd'
I8772
sa(dp17544
S'a'
I25954
sS's'
I17
sS'm'
(lp17545
I1
aI1
aI3
aI13
aI23
aI55
aI15
aI87
aI15
aI579
aI717
aI777
aI149
aI11431
aI26197
aI17711
aI7337
asS'd'
I8773
sa(dp17546
S'a'
I25960
sS's'
I17
sS'm'
(lp17547
I1
aI1
aI5
aI1
aI31
aI45
aI113
aI253
aI211
aI915
aI1855
aI4043
aI2159
aI1803
aI5061
aI40473
aI3657
asS'd'
I8774
sa(dp17548
S'a'
I25963
sS's'
I17
sS'm'
(lp17549
I1
aI1
aI3
aI7
aI25
aI15
aI37
aI73
aI467
aI969
aI1123
aI4053
aI4837
aI10091
aI25461
aI40803
aI91157
asS'd'
I8775
sa(dp17550
S'a'
I25966
sS's'
I17
sS'm'
(lp17551
I1
aI1
aI5
aI1
aI7
aI31
aI77
aI207
aI21
aI623
aI577
aI1195
aI5839
aI13013
aI11189
aI61691
aI33327
asS'd'
I8776
sa(dp17552
S'a'
I25978
sS's'
I17
sS'm'
(lp17553
I1
aI3
aI7
aI7
aI13
aI3
aI9
aI55
aI47
aI779
aI599
aI3747
aI1533
aI14705
aI23185
aI4011
aI36003
asS'd'
I8777
sa(dp17554
S'a'
I25996
sS's'
I17
sS'm'
(lp17555
I1
aI1
aI5
aI5
aI31
aI17
aI99
aI253
aI103
aI957
aI241
aI1893
aI7435
aI14907
aI9089
aI23205
aI70639
asS'd'
I8778
sa(dp17556
S'a'
I26007
sS's'
I17
sS'm'
(lp17557
I1
aI3
aI7
aI15
aI7
aI55
aI53
aI101
aI227
aI541
aI2017
aI275
aI577
aI15621
aI1799
aI50373
aI43197
asS'd'
I8779
sa(dp17558
S'a'
I26011
sS's'
I17
sS'm'
(lp17559
I1
aI3
aI1
aI15
aI29
aI23
aI69
aI193
aI429
aI359
aI1045
aI4091
aI6551
aI1673
aI29113
aI43247
aI80993
asS'd'
I8780
sa(dp17560
S'a'
I26027
sS's'
I17
sS'm'
(lp17561
I1
aI3
aI7
aI11
aI5
aI37
aI13
aI27
aI277
aI65
aI565
aI2631
aI6919
aI5593
aI8481
aI14703
aI9719
asS'd'
I8781
sa(dp17562
S'a'
I26032
sS's'
I17
sS'm'
(lp17563
I1
aI3
aI1
aI15
aI5
aI7
aI83
aI51
aI77
aI307
aI1299
aI1373
aI5281
aI15359
aI15569
aI50093
aI59661
asS'd'
I8782
sa(dp17564
S'a'
I26038
sS's'
I17
sS'm'
(lp17565
I1
aI3
aI5
aI11
aI13
aI31
aI99
aI123
aI263
aI319
aI2033
aI4055
aI2427
aI103
aI2009
aI27517
aI112467
asS'd'
I8783
sa(dp17566
S'a'
I26049
sS's'
I17
sS'm'
(lp17567
I1
aI1
aI7
aI3
aI13
aI1
aI51
aI131
aI17
aI861
aI459
aI3925
aI5511
aI5255
aI28553
aI36437
aI54591
asS'd'
I8784
sa(dp17568
S'a'
I26052
sS's'
I17
sS'm'
(lp17569
I1
aI3
aI7
aI5
aI9
aI57
aI49
aI119
aI291
aI727
aI1611
aI4035
aI4517
aI10979
aI28445
aI26905
aI57517
asS'd'
I8785
sa(dp17570
S'a'
I26055
sS's'
I17
sS'm'
(lp17571
I1
aI1
aI5
aI9
aI9
aI55
aI43
aI209
aI411
aI137
aI1619
aI3965
aI5253
aI8217
aI7569
aI42043
aI104163
asS'd'
I8786
sa(dp17572
S'a'
I26085
sS's'
I17
sS'm'
(lp17573
I1
aI3
aI3
aI7
aI21
aI3
aI107
aI255
aI353
aI735
aI71
aI1789
aI3351
aI755
aI22805
aI53537
aI126859
asS'd'
I8787
sa(dp17574
S'a'
I26089
sS's'
I17
sS'm'
(lp17575
I1
aI1
aI7
aI5
aI15
aI55
aI13
aI167
aI165
aI289
aI1231
aI2547
aI8135
aI5475
aI2361
aI49019
aI110579
asS'd'
I8788
sa(dp17576
S'a'
I26090
sS's'
I17
sS'm'
(lp17577
I1
aI3
aI1
aI11
aI17
aI21
aI59
aI37
aI177
aI517
aI499
aI4035
aI749
aI14297
aI22415
aI54975
aI29769
asS'd'
I8789
sa(dp17578
S'a'
I26098
sS's'
I17
sS'm'
(lp17579
I1
aI3
aI7
aI3
aI3
aI59
aI55
aI17
aI483
aI625
aI875
aI1465
aI7583
aI2969
aI2741
aI36965
aI80367
asS'd'
I8790
sa(dp17580
S'a'
I26104
sS's'
I17
sS'm'
(lp17581
I1
aI1
aI3
aI13
aI31
aI5
aI11
aI149
aI7
aI297
aI1485
aI735
aI4095
aI10089
aI5757
aI64997
aI56629
asS'd'
I8791
sa(dp17582
S'a'
I26110
sS's'
I17
sS'm'
(lp17583
I1
aI3
aI1
aI13
aI19
aI43
aI77
aI209
aI309
aI739
aI1765
aI3297
aI8167
aI6523
aI27987
aI25235
aI80555
asS'd'
I8792
sa(dp17584
S'a'
I26113
sS's'
I17
sS'm'
(lp17585
I1
aI1
aI3
aI9
aI31
aI57
aI125
aI75
aI3
aI633
aI85
aI3339
aI1691
aI9721
aI17465
aI36801
aI106147
asS'd'
I8793
sa(dp17586
S'a'
I26126
sS's'
I17
sS'm'
(lp17587
I1
aI3
aI5
aI15
aI27
aI7
aI111
aI7
aI475
aI523
aI1825
aI1367
aI1549
aI15533
aI13827
aI14471
aI100271
asS'd'
I8794
sa(dp17588
S'a'
I26133
sS's'
I17
sS'm'
(lp17589
I1
aI1
aI5
aI3
aI5
aI61
aI1
aI221
aI163
aI183
aI1701
aI3549
aI349
aI10057
aI26169
aI20725
aI55305
asS'd'
I8795
sa(dp17590
S'a'
I26138
sS's'
I17
sS'm'
(lp17591
I1
aI1
aI3
aI1
aI15
aI41
aI13
aI71
aI269
aI909
aI1253
aI2553
aI83
aI10055
aI1057
aI39841
aI20437
asS'd'
I8796
sa(dp17592
S'a'
I26140
sS's'
I17
sS'm'
(lp17593
I1
aI3
aI3
aI5
aI29
aI39
aI113
aI23
aI137
aI601
aI361
aI1779
aI279
aI15803
aI8993
aI2633
aI114847
asS'd'
I8797
sa(dp17594
S'a'
I26156
sS's'
I17
sS'm'
(lp17595
I1
aI1
aI3
aI7
aI29
aI45
aI35
aI27
aI71
aI253
aI231
aI3449
aI1955
aI9109
aI9043
aI50593
aI15023
asS'd'
I8798
sa(dp17596
S'a'
I26159
sS's'
I17
sS'm'
(lp17597
I1
aI3
aI1
aI11
aI17
aI45
aI85
aI255
aI341
aI957
aI769
aI3009
aI3997
aI6435
aI1161
aI34219
aI97077
asS'd'
I8799
sa(dp17598
S'a'
I26176
sS's'
I17
sS'm'
(lp17599
I1
aI1
aI5
aI1
aI1
aI19
aI9
aI51
aI477
aI129
aI1411
aI3223
aI5069
aI3237
aI15947
aI27215
aI70401
asS'd'
I8800
sa(dp17600
S'a'
I26186
sS's'
I17
sS'm'
(lp17601
I1
aI1
aI3
aI1
aI1
aI1
aI73
aI31
aI301
aI227
aI119
aI607
aI3379
aI3907
aI1263
aI2651
aI43769
asS'd'
I8801
sa(dp17602
S'a'
I26203
sS's'
I17
sS'm'
(lp17603
I1
aI1
aI1
aI15
aI21
aI59
aI109
aI155
aI473
aI361
aI1871
aI3085
aI513
aI12607
aI12747
aI41067
aI44977
asS'd'
I8802
sa(dp17604
S'a'
I26210
sS's'
I17
sS'm'
(lp17605
I1
aI1
aI7
aI3
aI27
aI21
aI89
aI71
aI437
aI671
aI1469
aI2191
aI4225
aI6343
aI1131
aI29141
aI25221
asS'd'
I8803
sa(dp17606
S'a'
I26222
sS's'
I17
sS'm'
(lp17607
I1
aI1
aI7
aI9
aI7
aI45
aI95
aI197
aI391
aI11
aI1913
aI3107
aI5247
aI959
aI30395
aI32809
aI20893
asS'd'
I8804
sa(dp17608
S'a'
I26224
sS's'
I17
sS'm'
(lp17609
I1
aI3
aI7
aI15
aI13
aI49
aI77
aI245
aI463
aI769
aI1807
aI1311
aI2715
aI14819
aI27595
aI57521
aI105221
asS'd'
I8805
sa(dp17610
S'a'
I26229
sS's'
I17
sS'm'
(lp17611
I1
aI1
aI1
aI5
aI23
aI45
aI119
aI77
aI325
aI405
aI1631
aI461
aI6357
aI7037
aI31699
aI46295
aI118577
asS'd'
I8806
sa(dp17612
S'a'
I26233
sS's'
I17
sS'm'
(lp17613
I1
aI3
aI5
aI1
aI21
aI3
aI31
aI13
aI7
aI571
aI633
aI425
aI6547
aI3423
aI19355
aI49481
aI76289
asS'd'
I8807
sa(dp17614
S'a'
I26236
sS's'
I17
sS'm'
(lp17615
I1
aI1
aI3
aI9
aI7
aI51
aI113
aI173
aI169
aI97
aI1821
aI979
aI2553
aI11505
aI20047
aI39277
aI122905
asS'd'
I8808
sa(dp17616
S'a'
I26239
sS's'
I17
sS'm'
(lp17617
I1
aI1
aI5
aI13
aI17
aI9
aI111
aI205
aI261
aI671
aI657
aI507
aI3903
aI10767
aI4387
aI3045
aI102617
asS'd'
I8809
sa(dp17618
S'a'
I26267
sS's'
I17
sS'm'
(lp17619
I1
aI1
aI3
aI3
aI5
aI11
aI19
aI61
aI401
aI605
aI455
aI2457
aI4471
aI7255
aI18435
aI49673
aI97289
asS'd'
I8810
sa(dp17620
S'a'
I26291
sS's'
I17
sS'm'
(lp17621
I1
aI3
aI1
aI1
aI31
aI25
aI77
aI35
aI65
aI127
aI929
aI2325
aI2315
aI13819
aI5509
aI12515
aI36991
asS'd'
I8811
sa(dp17622
S'a'
I26293
sS's'
I17
sS'm'
(lp17623
I1
aI1
aI7
aI5
aI21
aI49
aI33
aI119
aI181
aI645
aI1425
aI2411
aI245
aI13755
aI18775
aI50061
aI47119
asS'd'
I8812
sa(dp17624
S'a'
I26298
sS's'
I17
sS'm'
(lp17625
I1
aI3
aI7
aI5
aI27
aI43
aI81
aI191
aI233
aI435
aI829
aI3881
aI713
aI11153
aI4637
aI37721
aI115567
asS'd'
I8813
sa(dp17626
S'a'
I26303
sS's'
I17
sS'm'
(lp17627
I1
aI3
aI7
aI1
aI27
aI59
aI51
aI165
aI59
aI931
aI1921
aI3059
aI6843
aI813
aI22063
aI29445
aI114765
asS'd'
I8814
sa(dp17628
S'a'
I26308
sS's'
I17
sS'm'
(lp17629
I1
aI1
aI5
aI11
aI31
aI53
aI89
aI69
aI29
aI893
aI1241
aI7
aI1707
aI16167
aI8371
aI14021
aI103281
asS'd'
I8815
sa(dp17630
S'a'
I26312
sS's'
I17
sS'm'
(lp17631
I1
aI3
aI1
aI1
aI23
aI21
aI3
aI35
aI73
aI769
aI1417
aI4051
aI3223
aI813
aI1141
aI18823
aI31951
asS'd'
I8816
sa(dp17632
S'a'
I26315
sS's'
I17
sS'm'
(lp17633
I1
aI3
aI7
aI11
aI9
aI17
aI89
aI85
aI407
aI137
aI1865
aI2881
aI1495
aI3757
aI3711
aI36651
aI1797
asS'd'
I8817
sa(dp17634
S'a'
I26342
sS's'
I17
sS'm'
(lp17635
I1
aI3
aI5
aI1
aI25
aI29
aI29
aI217
aI15
aI173
aI479
aI2363
aI2341
aI6193
aI16403
aI64097
aI1173
asS'd'
I8818
sa(dp17636
S'a'
I26354
sS's'
I17
sS'm'
(lp17637
I1
aI3
aI3
aI3
aI11
aI29
aI113
aI167
aI333
aI573
aI1467
aI2223
aI5693
aI1063
aI20299
aI40993
aI69055
asS'd'
I8819
sa(dp17638
S'a'
I26363
sS's'
I17
sS'm'
(lp17639
I1
aI1
aI3
aI7
aI3
aI51
aI45
aI139
aI79
aI393
aI1251
aI1087
aI1423
aI1827
aI23445
aI41635
aI78511
asS'd'
I8820
sa(dp17640
S'a'
I26365
sS's'
I17
sS'm'
(lp17641
I1
aI3
aI3
aI13
aI29
aI45
aI85
aI229
aI33
aI373
aI113
aI1205
aI997
aI11777
aI7663
aI18513
aI5797
asS'd'
I8821
sa(dp17642
S'a'
I26383
sS's'
I17
sS'm'
(lp17643
I1
aI1
aI5
aI5
aI15
aI5
aI127
aI85
aI49
aI345
aI901
aI3215
aI2347
aI2329
aI19597
aI39555
aI25031
asS'd'
I8822
sa(dp17644
S'a'
I26391
sS's'
I17
sS'm'
(lp17645
I1
aI1
aI7
aI11
aI9
aI25
aI71
aI183
aI341
aI1011
aI439
aI3649
aI2859
aI10183
aI7635
aI45297
aI38581
asS'd'
I8823
sa(dp17646
S'a'
I26398
sS's'
I17
sS'm'
(lp17647
I1
aI1
aI1
aI11
aI23
aI13
aI1
aI69
aI461
aI77
aI1641
aI2851
aI1889
aI2413
aI1131
aI39009
aI33773
asS'd'
I8824
sa(dp17648
S'a'
I26416
sS's'
I17
sS'm'
(lp17649
I1
aI3
aI7
aI7
aI25
aI19
aI67
aI233
aI141
aI207
aI1501
aI453
aI4773
aI7411
aI22839
aI25365
aI53189
asS'd'
I8825
sa(dp17650
S'a'
I26421
sS's'
I17
sS'm'
(lp17651
I1
aI1
aI7
aI3
aI3
aI17
aI13
aI167
aI73
aI1005
aI887
aI2595
aI4351
aI3249
aI5653
aI36025
aI33733
asS'd'
I8826
sa(dp17652
S'a'
I26425
sS's'
I17
sS'm'
(lp17653
I1
aI1
aI7
aI15
aI11
aI1
aI105
aI215
aI329
aI601
aI1477
aI723
aI4597
aI3525
aI26235
aI63957
aI26677
asS'd'
I8827
sa(dp17654
S'a'
I26434
sS's'
I17
sS'm'
(lp17655
I1
aI1
aI1
aI11
aI27
aI15
aI111
aI133
aI327
aI567
aI845
aI2135
aI7905
aI7297
aI29255
aI14947
aI104885
asS'd'
I8828
sa(dp17656
S'a'
I26439
sS's'
I17
sS'm'
(lp17657
I1
aI1
aI7
aI9
aI21
aI11
aI67
aI179
aI459
aI423
aI295
aI3445
aI1681
aI12907
aI29281
aI7445
aI35539
asS'd'
I8829
sa(dp17658
S'a'
I26453
sS's'
I17
sS'm'
(lp17659
I1
aI1
aI3
aI11
aI23
aI61
aI111
aI123
aI81
aI439
aI299
aI3557
aI2821
aI705
aI15393
aI37175
aI11533
asS'd'
I8830
sa(dp17660
S'a'
I26454
sS's'
I17
sS'm'
(lp17661
I1
aI1
aI3
aI1
aI13
aI15
aI113
aI227
aI285
aI313
aI687
aI2085
aI6363
aI8003
aI32661
aI36461
aI68759
asS'd'
I8831
sa(dp17662
S'a'
I26470
sS's'
I17
sS'm'
(lp17663
I1
aI3
aI3
aI1
aI27
aI7
aI101
aI177
aI363
aI461
aI1519
aI2339
aI473
aI7469
aI4335
aI30951
aI130987
asS'd'
I8832
sa(dp17664
S'a'
I26491
sS's'
I17
sS'm'
(lp17665
I1
aI1
aI1
aI3
aI31
aI39
aI59
aI159
aI207
aI93
aI581
aI1973
aI945
aI11343
aI15901
aI22001
aI3515
asS'd'
I8833
sa(dp17666
S'a'
I26493
sS's'
I17
sS'm'
(lp17667
I1
aI3
aI7
aI5
aI11
aI53
aI125
aI57
aI389
aI985
aI1055
aI3803
aI3879
aI5537
aI28221
aI36805
aI16025
asS'd'
I8834
sa(dp17668
S'a'
I26500
sS's'
I17
sS'm'
(lp17669
I1
aI1
aI1
aI9
aI1
aI63
aI81
aI57
aI59
aI813
aI865
aI3491
aI3771
aI6121
aI6847
aI14765
aI68567
asS'd'
I8835
sa(dp17670
S'a'
I26509
sS's'
I17
sS'm'
(lp17671
I1
aI1
aI7
aI7
aI13
aI17
aI23
aI211
aI435
aI167
aI933
aI357
aI3567
aI3019
aI28439
aI17701
aI119937
asS'd'
I8836
sa(dp17672
S'a'
I26518
sS's'
I17
sS'm'
(lp17673
I1
aI1
aI1
aI9
aI7
aI53
aI103
aI155
aI211
aI719
aI413
aI3673
aI2795
aI15687
aI1737
aI50855
aI129133
asS'd'
I8837
sa(dp17674
S'a'
I26531
sS's'
I17
sS'm'
(lp17675
I1
aI1
aI3
aI13
aI11
aI23
aI53
aI121
aI497
aI383
aI1655
aI937
aI5563
aI2549
aI23183
aI46149
aI78875
asS'd'
I8838
sa(dp17676
S'a'
I26533
sS's'
I17
sS'm'
(lp17677
I1
aI3
aI5
aI11
aI25
aI1
aI45
aI139
aI437
aI729
aI2009
aI3245
aI4091
aI551
aI25993
aI31655
aI33641
asS'd'
I8839
sa(dp17678
S'a'
I26540
sS's'
I17
sS'm'
(lp17679
I1
aI3
aI1
aI13
aI7
aI23
aI87
aI111
aI471
aI501
aI1767
aI1051
aI7037
aI3199
aI19609
aI43227
aI53667
asS'd'
I8840
sa(dp17680
S'a'
I26552
sS's'
I17
sS'm'
(lp17681
I1
aI3
aI7
aI1
aI25
aI1
aI49
aI189
aI55
aI375
aI601
aI2065
aI2991
aI7697
aI25739
aI14951
aI43705
asS'd'
I8841
sa(dp17682
S'a'
I26566
sS's'
I17
sS'm'
(lp17683
I1
aI1
aI7
aI5
aI29
aI21
aI51
aI111
aI81
aI809
aI1963
aI2143
aI5529
aI15701
aI4719
aI11857
aI88207
asS'd'
I8842
sa(dp17684
S'a'
I26569
sS's'
I17
sS'm'
(lp17685
I1
aI1
aI5
aI11
aI27
aI27
aI7
aI145
aI281
aI939
aI537
aI1255
aI1475
aI11383
aI15081
aI9105
aI102775
asS'd'
I8843
sa(dp17686
S'a'
I26583
sS's'
I17
sS'm'
(lp17687
I1
aI1
aI5
aI7
aI9
aI45
aI67
aI23
aI65
aI139
aI1871
aI3975
aI6357
aI6515
aI25423
aI23915
aI76289
asS'd'
I8844
sa(dp17688
S'a'
I26590
sS's'
I17
sS'm'
(lp17689
I1
aI1
aI5
aI11
aI31
aI31
aI89
aI65
aI451
aI341
aI819
aI2439
aI6753
aI15113
aI9085
aI32687
aI5055
asS'd'
I8845
sa(dp17690
S'a'
I26596
sS's'
I17
sS'm'
(lp17691
I1
aI1
aI5
aI1
aI15
aI29
aI123
aI83
aI495
aI185
aI303
aI315
aI6015
aI5257
aI2467
aI4903
aI78269
asS'd'
I8846
sa(dp17692
S'a'
I26624
sS's'
I17
sS'm'
(lp17693
I1
aI1
aI3
aI5
aI31
aI51
aI49
aI199
aI501
aI407
aI733
aI1181
aI8023
aI7321
aI14765
aI17535
aI19893
asS'd'
I8847
sa(dp17694
S'a'
I26636
sS's'
I17
sS'm'
(lp17695
I1
aI1
aI5
aI5
aI19
aI15
aI103
aI183
aI13
aI969
aI1537
aI3053
aI3173
aI12983
aI21761
aI33733
aI67799
asS'd'
I8848
sa(dp17696
S'a'
I26667
sS's'
I17
sS'm'
(lp17697
I1
aI3
aI1
aI1
aI27
aI55
aI37
aI149
aI379
aI11
aI1655
aI2317
aI3135
aI6459
aI25941
aI12679
aI60245
asS'd'
I8849
sa(dp17698
S'a'
I26669
sS's'
I17
sS'm'
(lp17699
I1
aI1
aI3
aI9
aI9
aI13
aI33
aI243
aI337
aI741
aI1685
aI1147
aI5465
aI4865
aI559
aI23993
aI47273
asS'd'
I8850
sa(dp17700
S'a'
I26672
sS's'
I17
sS'm'
(lp17701
I1
aI3
aI5
aI13
aI21
aI11
aI39
aI169
aI135
aI209
aI1909
aI3655
aI3117
aI1075
aI8165
aI54633
aI28189
asS'd'
I8851
sa(dp17702
S'a'
I26675
sS's'
I17
sS'm'
(lp17703
I1
aI3
aI1
aI15
aI9
aI23
aI11
aI123
aI63
aI133
aI947
aI907
aI3853
aI10291
aI22905
aI4561
aI92497
asS'd'
I8852
sa(dp17704
S'a'
I26687
sS's'
I17
sS'm'
(lp17705
I1
aI1
aI3
aI13
aI17
aI9
aI5
aI209
aI429
aI3
aI2035
aI1497
aI6765
aI5991
aI24991
aI8155
aI103417
asS'd'
I8853
sa(dp17706
S'a'
I26689
sS's'
I17
sS'm'
(lp17707
I1
aI1
aI5
aI13
aI27
aI47
aI79
aI11
aI41
aI791
aI1939
aI3099
aI4069
aI4665
aI20801
aI18659
aI72163
asS'd'
I8854
sa(dp17708
S'a'
I26696
sS's'
I17
sS'm'
(lp17709
I1
aI3
aI1
aI13
aI25
aI37
aI79
aI131
aI233
aI647
aI291
aI1419
aI5157
aI4261
aI27715
aI611
aI83157
asS'd'
I8855
sa(dp17710
S'a'
I26702
sS's'
I17
sS'm'
(lp17711
I1
aI1
aI5
aI1
aI17
aI61
aI45
aI163
aI137
aI937
aI91
aI1695
aI1553
aI543
aI28615
aI6855
aI75201
asS'd'
I8856
sa(dp17712
S'a'
I26714
sS's'
I17
sS'm'
(lp17713
I1
aI1
aI3
aI13
aI7
aI63
aI109
aI13
aI351
aI159
aI1111
aI2791
aI4701
aI5805
aI9707
aI18361
aI98091
asS'd'
I8857
sa(dp17714
S'a'
I26716
sS's'
I17
sS'm'
(lp17715
I1
aI1
aI3
aI11
aI11
aI21
aI55
aI247
aI111
aI801
aI93
aI3091
aI1043
aI9761
aI23679
aI5555
aI195
asS'd'
I8858
sa(dp17716
S'a'
I26729
sS's'
I17
sS'm'
(lp17717
I1
aI1
aI1
aI1
aI13
aI43
aI123
aI113
aI265
aI561
aI659
aI3755
aI6605
aI10949
aI30511
aI29495
aI9075
asS'd'
I8859
sa(dp17718
S'a'
I26732
sS's'
I17
sS'm'
(lp17719
I1
aI3
aI1
aI7
aI23
aI63
aI19
aI73
aI233
aI1017
aI851
aI1971
aI3999
aI7407
aI25309
aI63991
aI92867
asS'd'
I8860
sa(dp17720
S'a'
I26737
sS's'
I17
sS'm'
(lp17721
I1
aI3
aI3
aI3
aI19
aI63
aI127
aI107
aI465
aI463
aI1507
aI1323
aI4729
aI14717
aI9129
aI24859
aI117565
asS'd'
I8861
sa(dp17722
S'a'
I26747
sS's'
I17
sS'm'
(lp17723
I1
aI1
aI7
aI1
aI19
aI11
aI13
aI85
aI339
aI939
aI895
aI887
aI765
aI15501
aI8783
aI65087
aI77899
asS'd'
I8862
sa(dp17724
S'a'
I26773
sS's'
I17
sS'm'
(lp17725
I1
aI3
aI7
aI3
aI7
aI15
aI43
aI153
aI365
aI223
aI1947
aI2295
aI787
aI5549
aI20089
aI29203
aI4807
asS'd'
I8863
sa(dp17726
S'a'
I26780
sS's'
I17
sS'm'
(lp17727
I1
aI3
aI7
aI3
aI31
aI27
aI51
aI217
aI483
aI623
aI633
aI2123
aI1211
aI9173
aI17949
aI54251
aI89161
asS'd'
I8864
sa(dp17728
S'a'
I26801
sS's'
I17
sS'm'
(lp17729
I1
aI3
aI3
aI11
aI3
aI11
aI111
aI73
aI283
aI23
aI1925
aI253
aI5141
aI12545
aI24913
aI16847
aI13067
asS'd'
I8865
sa(dp17730
S'a'
I26811
sS's'
I17
sS'm'
(lp17731
I1
aI3
aI5
aI5
aI31
aI39
aI35
aI235
aI135
aI85
aI191
aI999
aI6633
aI12527
aI21401
aI61339
aI81239
asS'd'
I8866
sa(dp17732
S'a'
I26813
sS's'
I17
sS'm'
(lp17733
I1
aI1
aI3
aI15
aI9
aI13
aI39
aI125
aI137
aI639
aI555
aI813
aI2821
aI1199
aI32075
aI15079
aI104609
asS'd'
I8867
sa(dp17734
S'a'
I26831
sS's'
I17
sS'm'
(lp17735
I1
aI3
aI7
aI7
aI15
aI43
aI99
aI51
aI245
aI25
aI147
aI89
aI6841
aI5523
aI28493
aI13967
aI113109
asS'd'
I8868
sa(dp17736
S'a'
I26836
sS's'
I17
sS'm'
(lp17737
I1
aI1
aI3
aI13
aI7
aI5
aI27
aI121
aI269
aI231
aI1011
aI1365
aI5055
aI11619
aI27393
aI48033
aI65725
asS'd'
I8869
sa(dp17738
S'a'
I26839
sS's'
I17
sS'm'
(lp17739
I1
aI1
aI7
aI5
aI9
aI23
aI41
aI71
aI327
aI339
aI1681
aI3303
aI4143
aI421
aI15213
aI48405
aI98067
asS'd'
I8870
sa(dp17740
S'a'
I26849
sS's'
I17
sS'm'
(lp17741
I1
aI3
aI7
aI15
aI15
aI33
aI73
aI33
aI351
aI131
aI1051
aI3909
aI7535
aI7659
aI9443
aI35015
aI329
asS'd'
I8871
sa(dp17742
S'a'
I26862
sS's'
I17
sS'm'
(lp17743
I1
aI1
aI3
aI9
aI19
aI55
aI97
aI223
aI265
aI877
aI235
aI867
aI4961
aI3137
aI19885
aI10955
aI7655
asS'd'
I8872
sa(dp17744
S'a'
I26869
sS's'
I17
sS'm'
(lp17745
I1
aI3
aI5
aI13
aI1
aI11
aI75
aI215
aI271
aI793
aI1691
aI1437
aI1317
aI10977
aI1311
aI64865
aI92361
asS'd'
I8873
sa(dp17746
S'a'
I26873
sS's'
I17
sS'm'
(lp17747
I1
aI1
aI1
aI5
aI23
aI23
aI35
aI53
aI187
aI345
aI115
aI929
aI3919
aI4523
aI31709
aI16771
aI33399
asS'd'
I8874
sa(dp17748
S'a'
I26874
sS's'
I17
sS'm'
(lp17749
I1
aI3
aI5
aI11
aI17
aI7
aI75
aI57
aI351
aI359
aI1737
aI2665
aI4259
aI13905
aI6999
aI45359
aI117891
asS'd'
I8875
sa(dp17750
S'a'
I26888
sS's'
I17
sS'm'
(lp17751
I1
aI1
aI5
aI3
aI23
aI29
aI49
aI209
aI417
aI843
aI531
aI1649
aI7829
aI6271
aI3759
aI39727
aI47415
asS'd'
I8876
sa(dp17752
S'a'
I26891
sS's'
I17
sS'm'
(lp17753
I1
aI1
aI7
aI1
aI21
aI45
aI101
aI105
aI385
aI797
aI985
aI1447
aI3757
aI3287
aI583
aI29283
aI96821
asS'd'
I8877
sa(dp17754
S'a'
I26896
sS's'
I17
sS'm'
(lp17755
I1
aI1
aI7
aI9
aI1
aI29
aI15
aI207
aI289
aI465
aI815
aI2289
aI449
aI9403
aI19197
aI13797
aI102651
asS'd'
I8878
sa(dp17756
S'a'
I26899
sS's'
I17
sS'm'
(lp17757
I1
aI3
aI7
aI15
aI5
aI45
aI81
aI187
aI21
aI433
aI679
aI2759
aI3375
aI6935
aI22595
aI50149
aI13557
asS'd'
I8879
sa(dp17758
S'a'
I26912
sS's'
I17
sS'm'
(lp17759
I1
aI3
aI3
aI5
aI11
aI55
aI69
aI219
aI95
aI21
aI645
aI1955
aI7527
aI6037
aI29427
aI36297
aI62013
asS'd'
I8880
sa(dp17760
S'a'
I26921
sS's'
I17
sS'm'
(lp17761
I1
aI1
aI7
aI15
aI13
aI35
aI25
aI67
aI383
aI13
aI539
aI2399
aI4611
aI8065
aI3815
aI27771
aI50411
asS'd'
I8881
sa(dp17762
S'a'
I26935
sS's'
I17
sS'm'
(lp17763
I1
aI3
aI1
aI3
aI27
aI47
aI65
aI33
aI393
aI895
aI1663
aI1289
aI1057
aI11887
aI1259
aI53611
aI36811
asS'd'
I8882
sa(dp17764
S'a'
I26941
sS's'
I17
sS'm'
(lp17765
I1
aI3
aI5
aI3
aI5
aI1
aI27
aI65
aI379
aI15
aI1643
aI1461
aI3009
aI8177
aI15589
aI5889
aI1103
asS'd'
I8883
sa(dp17766
S'a'
I26949
sS's'
I17
sS'm'
(lp17767
I1
aI1
aI5
aI15
aI27
aI53
aI43
aI173
aI377
aI665
aI581
aI1061
aI1363
aI15015
aI26709
aI29507
aI28075
asS'd'
I8884
sa(dp17768
S'a'
I26954
sS's'
I17
sS'm'
(lp17769
I1
aI1
aI7
aI3
aI9
aI11
aI45
aI71
aI23
aI995
aI1277
aI855
aI1001
aI12927
aI19753
aI46639
aI16949
asS'd'
I8885
sa(dp17770
S'a'
I26967
sS's'
I17
sS'm'
(lp17771
I1
aI1
aI7
aI5
aI15
aI33
aI27
aI27
aI437
aI415
aI1785
aI2091
aI279
aI8041
aI2209
aI15821
aI122363
asS'd'
I8886
sa(dp17772
S'a'
I26977
sS's'
I17
sS'm'
(lp17773
I1
aI1
aI7
aI1
aI21
aI1
aI47
aI215
aI463
aI959
aI849
aI1703
aI5175
aI10043
aI16991
aI11023
aI52201
asS'd'
I8887
sa(dp17774
S'a'
I27011
sS's'
I17
sS'm'
(lp17775
I1
aI1
aI1
aI11
aI21
aI7
aI121
aI31
aI95
aI631
aI1717
aI3017
aI2083
aI2047
aI12051
aI63117
aI25949
asS'd'
I8888
sa(dp17776
S'a'
I27018
sS's'
I17
sS'm'
(lp17777
I1
aI1
aI5
aI5
aI9
aI5
aI105
aI121
aI205
aI643
aI1721
aI2601
aI2991
aI2381
aI4873
aI12049
aI20043
asS'd'
I8889
sa(dp17778
S'a'
I27026
sS's'
I17
sS'm'
(lp17779
I1
aI3
aI5
aI11
aI7
aI11
aI97
aI187
aI253
aI571
aI101
aI3077
aI1479
aI9513
aI15451
aI37105
aI34445
asS'd'
I8890
sa(dp17780
S'a'
I27028
sS's'
I17
sS'm'
(lp17781
I1
aI1
aI7
aI11
aI19
aI19
aI39
aI115
aI221
aI13
aI217
aI369
aI6855
aI14529
aI143
aI13461
aI62927
asS'd'
I8891
sa(dp17782
S'a'
I27032
sS's'
I17
sS'm'
(lp17783
I1
aI3
aI3
aI7
aI29
aI27
aI9
aI171
aI419
aI571
aI837
aI3829
aI1871
aI12691
aI30693
aI4195
aI38905
asS'd'
I8892
sa(dp17784
S'a'
I27044
sS's'
I17
sS'm'
(lp17785
I1
aI1
aI1
aI11
aI5
aI55
aI17
aI41
aI241
aI419
aI337
aI897
aI4663
aI14469
aI18701
aI18009
aI44605
asS'd'
I8893
sa(dp17786
S'a'
I27053
sS's'
I17
sS'm'
(lp17787
I1
aI1
aI7
aI7
aI1
aI33
aI63
aI197
aI257
aI655
aI1287
aI2571
aI57
aI13275
aI29669
aI8501
aI61389
asS'd'
I8894
sa(dp17788
S'a'
I27062
sS's'
I17
sS'm'
(lp17789
I1
aI3
aI5
aI3
aI29
aI39
aI101
aI215
aI101
aI271
aI1373
aI2171
aI841
aI9865
aI28951
aI20369
aI42413
asS'd'
I8895
sa(dp17790
S'a'
I27065
sS's'
I17
sS'm'
(lp17791
I1
aI3
aI5
aI1
aI31
aI23
aI119
aI137
aI263
aI633
aI1239
aI281
aI4965
aI14913
aI30229
aI14147
aI37183
asS'd'
I8896
sa(dp17792
S'a'
I27068
sS's'
I17
sS'm'
(lp17793
I1
aI3
aI1
aI3
aI11
aI55
aI33
aI45
aI69
aI913
aI269
aI1021
aI4005
aI15191
aI11187
aI45917
aI76905
asS'd'
I8897
sa(dp17794
S'a'
I27080
sS's'
I17
sS'm'
(lp17795
I1
aI1
aI1
aI13
aI27
aI11
aI75
aI139
aI243
aI221
aI1289
aI2195
aI7041
aI10053
aI5731
aI35245
aI41317
asS'd'
I8898
sa(dp17796
S'a'
I27083
sS's'
I17
sS'm'
(lp17797
I1
aI1
aI7
aI9
aI25
aI11
aI81
aI243
aI233
aI137
aI831
aI2825
aI6007
aI7305
aI31733
aI64343
aI7047
asS'd'
I8899
sa(dp17798
S'a'
I27093
sS's'
I17
sS'm'
(lp17799
I1
aI3
aI5
aI9
aI17
aI61
aI25
aI245
aI285
aI969
aI1397
aI1331
aI5393
aI3653
aI15533
aI9121
aI111115
asS'd'
I8900
sa(dp17800
S'a'
I27098
sS's'
I17
sS'm'
(lp17801
I1
aI1
aI5
aI9
aI1
aI9
aI61
aI205
aI285
aI849
aI1071
aI1013
aI2655
aI10121
aI15165
aI25189
aI56207
asS'd'
I8901
sa(dp17802
S'a'
I27107
sS's'
I17
sS'm'
(lp17803
I1
aI3
aI5
aI7
aI19
aI45
aI121
aI19
aI237
aI711
aI1523
aI3251
aI693
aI13567
aI31993
aI11239
aI64127
asS'd'
I8902
sa(dp17804
S'a'
I27110
sS's'
I17
sS'm'
(lp17805
I1
aI1
aI1
aI11
aI23
aI25
aI33
aI211
aI321
aI1021
aI1855
aI291
aI2911
aI11841
aI21929
aI64147
aI63201
asS'd'
I8903
sa(dp17806
S'a'
I27114
sS's'
I17
sS'm'
(lp17807
I1
aI1
aI7
aI3
aI27
aI21
aI119
aI219
aI431
aI819
aI83
aI2487
aI7533
aI10697
aI3129
aI53301
aI104999
asS'd'
I8904
sa(dp17808
S'a'
I27121
sS's'
I17
sS'm'
(lp17809
I1
aI3
aI5
aI15
aI9
aI25
aI89
aI65
aI293
aI411
aI989
aI3103
aI5563
aI15703
aI8757
aI32595
aI43409
asS'd'
I8905
sa(dp17810
S'a'
I27127
sS's'
I17
sS'm'
(lp17811
I1
aI3
aI3
aI1
aI31
aI31
aI45
aI173
aI231
aI171
aI1185
aI1499
aI1713
aI9945
aI11575
aI37113
aI103989
asS'd'
I8906
sa(dp17812
S'a'
I27137
sS's'
I17
sS'm'
(lp17813
I1
aI1
aI5
aI13
aI27
aI3
aI93
aI253
aI23
aI71
aI1963
aI2571
aI6259
aI15757
aI9709
aI42835
aI42047
asS'd'
I8907
sa(dp17814
S'a'
I27144
sS's'
I17
sS'm'
(lp17815
I1
aI1
aI7
aI11
aI5
aI11
aI123
aI117
aI39
aI559
aI111
aI527
aI6253
aI781
aI9177
aI47189
aI94031
asS'd'
I8908
sa(dp17816
S'a'
I27155
sS's'
I17
sS'm'
(lp17817
I1
aI1
aI5
aI15
aI5
aI49
aI93
aI185
aI167
aI787
aI1553
aI3291
aI3723
aI1651
aI23225
aI5643
aI42967
asS'd'
I8909
sa(dp17818
S'a'
I27171
sS's'
I17
sS'm'
(lp17819
I1
aI3
aI1
aI13
aI15
aI35
aI19
aI193
aI505
aI127
aI661
aI1943
aI1249
aI5103
aI8233
aI64319
aI76955
asS'd'
I8910
sa(dp17820
S'a'
I27178
sS's'
I17
sS'm'
(lp17821
I1
aI3
aI7
aI7
aI17
aI13
aI97
aI185
aI415
aI331
aI283
aI3341
aI2903
aI2927
aI7729
aI4095
aI103083
asS'd'
I8911
sa(dp17822
S'a'
I27180
sS's'
I17
sS'm'
(lp17823
I1
aI1
aI3
aI15
aI15
aI25
aI65
aI45
aI413
aI521
aI747
aI2605
aI5845
aI12909
aI7651
aI45937
aI99043
asS'd'
I8912
sa(dp17824
S'a'
I27185
sS's'
I17
sS'm'
(lp17825
I1
aI1
aI5
aI9
aI9
aI21
aI3
aI75
aI335
aI745
aI1493
aI1721
aI1977
aI11105
aI22621
aI49281
aI107113
asS'd'
I8913
sa(dp17826
S'a'
I27191
sS's'
I17
sS'm'
(lp17827
I1
aI3
aI1
aI11
aI25
aI11
aI99
aI53
aI239
aI831
aI655
aI615
aI7565
aI14039
aI29115
aI26165
aI127159
asS'd'
I8914
sa(dp17828
S'a'
I27192
sS's'
I17
sS'm'
(lp17829
I1
aI1
aI7
aI5
aI31
aI35
aI75
aI157
aI441
aI815
aI119
aI565
aI2703
aI14059
aI7867
aI42487
aI93647
asS'd'
I8915
sa(dp17830
S'a'
I27195
sS's'
I17
sS'm'
(lp17831
I1
aI3
aI7
aI3
aI3
aI59
aI101
aI223
aI257
aI989
aI363
aI1059
aI5157
aI11129
aI1481
aI19287
aI117623
asS'd'
I8916
sa(dp17832
S'a'
I27212
sS's'
I17
sS'm'
(lp17833
I1
aI1
aI1
aI1
aI29
aI27
aI1
aI129
aI253
aI673
aI103
aI1881
aI7053
aI1261
aI32003
aI23345
aI102503
asS'd'
I8917
sa(dp17834
S'a'
I27215
sS's'
I17
sS'm'
(lp17835
I1
aI3
aI1
aI9
aI11
aI37
aI3
aI99
aI303
aI519
aI1175
aI3021
aI2275
aI9919
aI25011
aI45865
aI71351
asS'd'
I8918
sa(dp17836
S'a'
I27218
sS's'
I17
sS'm'
(lp17837
I1
aI1
aI7
aI15
aI27
aI9
aI107
aI61
aI385
aI21
aI861
aI2119
aI4643
aI8379
aI25455
aI22425
aI113387
asS'd'
I8919
sa(dp17838
S'a'
I27230
sS's'
I17
sS'm'
(lp17839
I1
aI1
aI1
aI7
aI27
aI15
aI73
aI211
aI497
aI527
aI873
aI329
aI2213
aI415
aI13987
aI56581
aI27829
asS'd'
I8920
sa(dp17840
S'a'
I27234
sS's'
I17
sS'm'
(lp17841
I1
aI3
aI5
aI1
aI31
aI43
aI107
aI149
aI209
aI211
aI2029
aI2793
aI2213
aI12389
aI27177
aI51375
aI51983
asS'd'
I8921
sa(dp17842
S'a'
I27245
sS's'
I17
sS'm'
(lp17843
I1
aI3
aI3
aI7
aI25
aI57
aI67
aI101
aI127
aI43
aI1775
aI801
aI3343
aI12203
aI8667
aI58387
aI10735
asS'd'
I8922
sa(dp17844
S'a'
I27248
sS's'
I17
sS'm'
(lp17845
I1
aI1
aI7
aI5
aI13
aI47
aI101
aI123
aI133
aI593
aI1409
aI3525
aI2545
aI13009
aI11873
aI38463
aI1075
asS'd'
I8923
sa(dp17846
S'a'
I27263
sS's'
I17
sS'm'
(lp17847
I1
aI3
aI3
aI1
aI3
aI19
aI75
aI221
aI157
aI67
aI397
aI1141
aI5073
aI10795
aI9857
aI35459
aI62701
asS'd'
I8924
sa(dp17848
S'a'
I27279
sS's'
I17
sS'm'
(lp17849
I1
aI1
aI7
aI7
aI23
aI17
aI41
aI179
aI83
aI543
aI1839
aI3709
aI131
aI15681
aI9147
aI18685
aI90389
asS'd'
I8925
sa(dp17850
S'a'
I27288
sS's'
I17
sS'm'
(lp17851
I1
aI1
aI5
aI7
aI17
aI15
aI31
aI217
aI79
aI687
aI1927
aI2889
aI6127
aI15095
aI28437
aI16403
aI123275
asS'd'
I8926
sa(dp17852
S'a'
I27294
sS's'
I17
sS'm'
(lp17853
I1
aI3
aI7
aI15
aI13
aI17
aI123
aI75
aI45
aI635
aI525
aI3897
aI6769
aI13855
aI16695
aI18039
aI37479
asS'd'
I8927
sa(dp17854
S'a'
I27303
sS's'
I17
sS'm'
(lp17855
I1
aI1
aI5
aI1
aI23
aI15
aI69
aI161
aI503
aI339
aI1061
aI839
aI9
aI10013
aI24493
aI32711
aI50147
asS'd'
I8928
sa(dp17856
S'a'
I27304
sS's'
I17
sS'm'
(lp17857
I1
aI3
aI3
aI11
aI5
aI45
aI9
aI233
aI131
aI629
aI1111
aI3311
aI2211
aI9079
aI19763
aI23793
aI85389
asS'd'
I8929
sa(dp17858
S'a'
I27312
sS's'
I17
sS'm'
(lp17859
I1
aI1
aI7
aI7
aI7
aI27
aI15
aI85
aI291
aI925
aI1545
aI3061
aI4867
aI1613
aI13467
aI53731
aI92811
asS'd'
I8930
sa(dp17860
S'a'
I27330
sS's'
I17
sS'm'
(lp17861
I1
aI3
aI5
aI1
aI21
aI21
aI71
aI33
aI141
aI675
aI1519
aI3275
aI1491
aI10717
aI28199
aI14983
aI18961
asS'd'
I8931
sa(dp17862
S'a'
I27341
sS's'
I17
sS'm'
(lp17863
I1
aI3
aI5
aI5
aI31
aI13
aI109
aI239
aI369
aI373
aI257
aI3765
aI2531
aI13281
aI11877
aI29439
aI106939
asS'd'
I8932
sa(dp17864
S'a'
I27342
sS's'
I17
sS'm'
(lp17865
I1
aI3
aI7
aI15
aI13
aI39
aI111
aI203
aI109
aI179
aI789
aI3849
aI433
aI5745
aI2343
aI15795
aI92603
asS'd'
I8933
sa(dp17866
S'a'
I27344
sS's'
I17
sS'm'
(lp17867
I1
aI3
aI5
aI5
aI11
aI57
aI3
aI245
aI289
aI249
aI713
aI315
aI2261
aI1249
aI6963
aI44507
aI100829
asS'd'
I8934
sa(dp17868
S'a'
I27353
sS's'
I17
sS'm'
(lp17869
I1
aI3
aI5
aI11
aI5
aI49
aI97
aI245
aI363
aI315
aI509
aI17
aI4485
aI15393
aI28491
aI17945
aI65663
asS'd'
I8935
sa(dp17870
S'a'
I27356
sS's'
I17
sS'm'
(lp17871
I1
aI3
aI5
aI1
aI5
aI13
aI15
aI17
aI141
aI119
aI1393
aI581
aI2439
aI2015
aI11527
aI8537
aI103005
asS'd'
I8936
sa(dp17872
S'a'
I27366
sS's'
I17
sS'm'
(lp17873
I1
aI3
aI5
aI1
aI25
aI9
aI117
aI25
aI99
aI777
aI985
aI1159
aI99
aI3013
aI21429
aI19027
aI61833
asS'd'
I8937
sa(dp17874
S'a'
I27369
sS's'
I17
sS'm'
(lp17875
I1
aI1
aI1
aI5
aI1
aI47
aI37
aI83
aI159
aI29
aI281
aI3789
aI2525
aI15999
aI8965
aI11145
aI14453
asS'd'
I8938
sa(dp17876
S'a'
I27377
sS's'
I17
sS'm'
(lp17877
I1
aI1
aI3
aI1
aI11
aI63
aI77
aI207
aI267
aI473
aI241
aI629
aI6969
aI9093
aI839
aI3875
aI18873
asS'd'
I8939
sa(dp17878
S'a'
I27387
sS's'
I17
sS'm'
(lp17879
I1
aI3
aI5
aI7
aI31
aI57
aI103
aI65
aI349
aI321
aI717
aI2403
aI105
aI643
aI27999
aI2509
aI123061
asS'd'
I8940
sa(dp17880
S'a'
I27409
sS's'
I17
sS'm'
(lp17881
I1
aI1
aI5
aI7
aI3
aI31
aI7
aI113
aI17
aI995
aI1211
aI1749
aI6757
aI3391
aI8011
aI47715
aI24301
asS'd'
I8941
sa(dp17882
S'a'
I27437
sS's'
I17
sS'm'
(lp17883
I1
aI3
aI7
aI11
aI7
aI55
aI29
aI155
aI373
aI81
aI1255
aI2205
aI3233
aI9537
aI22129
aI24505
aI79021
asS'd'
I8942
sa(dp17884
S'a'
I27443
sS's'
I17
sS'm'
(lp17885
I1
aI1
aI7
aI5
aI3
aI49
aI5
aI51
aI89
aI107
aI585
aI3933
aI745
aI11685
aI20663
aI12521
aI24357
asS'd'
I8943
sa(dp17886
S'a'
I27446
sS's'
I17
sS'm'
(lp17887
I1
aI1
aI7
aI1
aI17
aI17
aI83
aI215
aI357
aI581
aI2025
aI3411
aI7287
aI11925
aI2253
aI43513
aI98655
asS'd'
I8944
sa(dp17888
S'a'
I27450
sS's'
I17
sS'm'
(lp17889
I1
aI3
aI5
aI3
aI27
aI27
aI51
aI147
aI471
aI509
aI423
aI3807
aI677
aI8429
aI581
aI47999
aI35913
asS'd'
I8945
sa(dp17890
S'a'
I27452
sS's'
I17
sS'm'
(lp17891
I1
aI3
aI3
aI9
aI15
aI31
aI1
aI93
aI207
aI759
aI1991
aI473
aI2273
aI43
aI12547
aI58085
aI20953
asS'd'
I8946
sa(dp17892
S'a'
I27457
sS's'
I17
sS'm'
(lp17893
I1
aI1
aI3
aI3
aI1
aI27
aI39
aI219
aI381
aI187
aI159
aI2333
aI6141
aI3775
aI5693
aI14853
aI38765
asS'd'
I8947
sa(dp17894
S'a'
I27472
sS's'
I17
sS'm'
(lp17895
I1
aI3
aI1
aI5
aI19
aI1
aI19
aI237
aI231
aI975
aI1609
aI723
aI241
aI10105
aI18817
aI58373
aI118889
asS'd'
I8948
sa(dp17896
S'a'
I27475
sS's'
I17
sS'm'
(lp17897
I1
aI1
aI5
aI7
aI7
aI43
aI99
aI181
aI109
aI529
aI421
aI1493
aI1075
aI12219
aI24287
aI33479
aI29987
asS'd'
I8949
sa(dp17898
S'a'
I27487
sS's'
I17
sS'm'
(lp17899
I1
aI1
aI7
aI1
aI17
aI11
aI79
aI85
aI157
aI851
aI1429
aI3355
aI139
aI14327
aI30025
aI60303
aI109015
asS'd'
I8950
sa(dp17900
S'a'
I27488
sS's'
I17
sS'm'
(lp17901
I1
aI3
aI5
aI9
aI11
aI15
aI37
aI79
aI5
aI169
aI999
aI815
aI6255
aI11763
aI16299
aI49891
aI101917
asS'd'
I8951
sa(dp17902
S'a'
I27493
sS's'
I17
sS'm'
(lp17903
I1
aI3
aI5
aI9
aI29
aI45
aI51
aI211
aI159
aI771
aI1631
aI2871
aI4877
aI4941
aI18127
aI15669
aI57515
asS'd'
I8952
sa(dp17904
S'a'
I27506
sS's'
I17
sS'm'
(lp17905
I1
aI1
aI3
aI3
aI19
aI29
aI9
aI205
aI253
aI399
aI303
aI2441
aI3187
aI641
aI23341
aI52951
aI57559
asS'd'
I8953
sa(dp17906
S'a'
I27508
sS's'
I17
sS'm'
(lp17907
I1
aI3
aI3
aI15
aI11
aI29
aI121
aI227
aI69
aI935
aI365
aI217
aI4617
aI13193
aI27663
aI46903
aI107843
asS'd'
I8954
sa(dp17908
S'a'
I27521
sS's'
I17
sS'm'
(lp17909
I1
aI1
aI5
aI11
aI13
aI31
aI13
aI243
aI275
aI685
aI1613
aI1915
aI2775
aI11225
aI4729
aI45549
aI103875
asS'd'
I8955
sa(dp17910
S'a'
I27522
sS's'
I17
sS'm'
(lp17911
I1
aI1
aI5
aI1
aI9
aI61
aI35
aI143
aI165
aI441
aI517
aI1735
aI5281
aI10139
aI21107
aI11713
aI130483
asS'd'
I8956
sa(dp17912
S'a'
I27527
sS's'
I17
sS'm'
(lp17913
I1
aI3
aI5
aI5
aI13
aI21
aI7
aI219
aI97
aI887
aI1845
aI469
aI2523
aI1569
aI9959
aI4397
aI15823
asS'd'
I8957
sa(dp17914
S'a'
I27536
sS's'
I17
sS'm'
(lp17915
I1
aI3
aI7
aI11
aI15
aI27
aI73
aI223
aI365
aI939
aI331
aI145
aI683
aI6441
aI23421
aI59177
aI31763
asS'd'
I8958
sa(dp17916
S'a'
I27539
sS's'
I17
sS'm'
(lp17917
I1
aI3
aI1
aI5
aI9
aI59
aI85
aI113
aI343
aI831
aI121
aI3157
aI6059
aI14923
aI27401
aI19759
aI14223
asS'd'
I8959
sa(dp17918
S'a'
I27541
sS's'
I17
sS'm'
(lp17919
I1
aI3
aI1
aI7
aI17
aI25
aI3
aI39
aI471
aI759
aI2041
aI609
aI4293
aI7491
aI8041
aI50857
aI25601
asS'd'
I8960
sa(dp17920
S'a'
I27546
sS's'
I17
sS'm'
(lp17921
I1
aI1
aI5
aI15
aI19
aI45
aI21
aI5
aI269
aI197
aI527
aI1839
aI1719
aI15105
aI18671
aI42167
aI9617
asS'd'
I8961
sa(dp17922
S'a'
I27557
sS's'
I17
sS'm'
(lp17923
I1
aI3
aI1
aI3
aI5
aI35
aI3
aI105
aI395
aI113
aI1945
aI3945
aI3951
aI12207
aI32135
aI34121
aI10237
asS'd'
I8962
sa(dp17924
S'a'
I27564
sS's'
I17
sS'm'
(lp17925
I1
aI1
aI5
aI13
aI21
aI43
aI51
aI255
aI57
aI399
aI1937
aI1573
aI363
aI11589
aI26989
aI54345
aI94341
asS'd'
I8963
sa(dp17926
S'a'
I27572
sS's'
I17
sS'm'
(lp17927
I1
aI3
aI3
aI3
aI5
aI45
aI83
aI125
aI179
aI923
aI39
aI3617
aI7683
aI8191
aI31469
aI23633
aI79179
asS'd'
I8964
sa(dp17928
S'a'
I27576
sS's'
I17
sS'm'
(lp17929
I1
aI3
aI7
aI9
aI9
aI37
aI107
aI65
aI423
aI77
aI1779
aI1375
aI2085
aI11779
aI6535
aI2973
aI29425
asS'd'
I8965
sa(dp17930
S'a'
I27584
sS's'
I17
sS'm'
(lp17931
I1
aI1
aI7
aI3
aI11
aI39
aI101
aI137
aI407
aI855
aI1767
aI1717
aI2821
aI10447
aI31187
aI6329
aI124111
asS'd'
I8966
sa(dp17932
S'a'
I27608
sS's'
I17
sS'm'
(lp17933
I1
aI1
aI5
aI11
aI27
aI27
aI45
aI103
aI225
aI681
aI725
aI1791
aI2881
aI2923
aI14473
aI12269
aI58743
asS'd'
I8967
sa(dp17934
S'a'
I27611
sS's'
I17
sS'm'
(lp17935
I1
aI1
aI3
aI11
aI15
aI39
aI113
aI167
aI143
aI677
aI1189
aI1571
aI5339
aI6065
aI30187
aI19639
aI42227
asS'd'
I8968
sa(dp17936
S'a'
I27618
sS's'
I17
sS'm'
(lp17937
I1
aI1
aI1
aI3
aI13
aI7
aI113
aI75
aI129
aI619
aI1741
aI1495
aI4751
aI11085
aI22391
aI199
aI105463
asS'd'
I8969
sa(dp17938
S'a'
I27620
sS's'
I17
sS'm'
(lp17939
I1
aI1
aI3
aI3
aI19
aI47
aI77
aI195
aI209
aI453
aI495
aI1605
aI5255
aI15327
aI8941
aI18239
aI54511
asS'd'
I8970
sa(dp17940
S'a'
I27627
sS's'
I17
sS'm'
(lp17941
I1
aI1
aI7
aI7
aI21
aI29
aI95
aI175
aI3
aI689
aI611
aI2467
aI6919
aI12399
aI18869
aI16171
aI102753
asS'd'
I8971
sa(dp17942
S'a'
I27630
sS's'
I17
sS'm'
(lp17943
I1
aI1
aI5
aI1
aI27
aI43
aI61
aI133
aI37
aI603
aI315
aI1915
aI813
aI15769
aI27447
aI29589
aI122281
asS'd'
I8972
sa(dp17944
S'a'
I27637
sS's'
I17
sS'm'
(lp17945
I1
aI1
aI7
aI3
aI11
aI1
aI119
aI235
aI93
aI481
aI1811
aI1643
aI4853
aI11313
aI8991
aI6153
aI68985
asS'd'
I8973
sa(dp17946
S'a'
I27638
sS's'
I17
sS'm'
(lp17947
I1
aI1
aI1
aI7
aI1
aI13
aI99
aI83
aI221
aI775
aI1345
aI219
aI4445
aI11837
aI10405
aI43563
aI122111
asS'd'
I8974
sa(dp17948
S'a'
I27641
sS's'
I17
sS'm'
(lp17949
I1
aI1
aI5
aI13
aI21
aI33
aI105
aI19
aI343
aI571
aI703
aI429
aI2485
aI15531
aI9801
aI24101
aI88275
asS'd'
I8975
sa(dp17950
S'a'
I27647
sS's'
I17
sS'm'
(lp17951
I1
aI3
aI5
aI1
aI27
aI55
aI73
aI49
aI33
aI773
aI1411
aI1227
aI6827
aI1271
aI28897
aI24265
aI32383
asS'd'
I8976
sa(dp17952
S'a'
I27650
sS's'
I17
sS'm'
(lp17953
I1
aI3
aI7
aI3
aI9
aI45
aI59
aI5
aI157
aI669
aI261
aI2077
aI1425
aI8221
aI5849
aI40857
aI121029
asS'd'
I8977
sa(dp17954
S'a'
I27676
sS's'
I17
sS'm'
(lp17955
I1
aI3
aI7
aI11
aI23
aI5
aI87
aI113
aI279
aI611
aI1195
aI1775
aI5813
aI6737
aI18051
aI41341
aI93331
asS'd'
I8978
sa(dp17956
S'a'
I27683
sS's'
I17
sS'm'
(lp17957
I1
aI1
aI7
aI7
aI9
aI55
aI113
aI3
aI167
aI295
aI1579
aI2833
aI4003
aI7583
aI22833
aI44427
aI34781
asS'd'
I8979
sa(dp17958
S'a'
I27690
sS's'
I17
sS'm'
(lp17959
I1
aI1
aI5
aI13
aI21
aI33
aI127
aI175
aI153
aI961
aI819
aI143
aI3969
aI6159
aI29437
aI14123
aI65317
asS'd'
I8980
sa(dp17960
S'a'
I27695
sS's'
I17
sS'm'
(lp17961
I1
aI1
aI1
aI15
aI31
aI27
aI1
aI17
aI329
aI963
aI1907
aI421
aI535
aI2323
aI22749
aI44375
aI115531
asS'd'
I8981
sa(dp17962
S'a'
I27700
sS's'
I17
sS'm'
(lp17963
I1
aI1
aI5
aI15
aI15
aI23
aI57
aI171
aI253
aI401
aI1577
aI3855
aI197
aI7979
aI17577
aI25275
aI88831
asS'd'
I8982
sa(dp17964
S'a'
I27704
sS's'
I17
sS'm'
(lp17965
I1
aI1
aI7
aI9
aI27
aI9
aI7
aI13
aI381
aI847
aI533
aI357
aI6551
aI13441
aI5717
aI20209
aI64347
asS'd'
I8983
sa(dp17966
S'a'
I27709
sS's'
I17
sS'm'
(lp17967
I1
aI1
aI7
aI9
aI1
aI1
aI31
aI245
aI315
aI901
aI1857
aI497
aI4285
aI13227
aI3937
aI45025
aI59627
asS'd'
I8984
sa(dp17968
S'a'
I27715
sS's'
I17
sS'm'
(lp17969
I1
aI1
aI7
aI3
aI5
aI23
aI119
aI147
aI479
aI71
aI113
aI3379
aI863
aI8285
aI31259
aI15863
aI47267
asS'd'
I8985
sa(dp17970
S'a'
I27727
sS's'
I17
sS'm'
(lp17971
I1
aI3
aI5
aI1
aI5
aI7
aI77
aI163
aI421
aI353
aI1757
aI1335
aI4975
aI3011
aI11703
aI56075
aI12045
asS'd'
I8986
sa(dp17972
S'a'
I27745
sS's'
I17
sS'm'
(lp17973
I1
aI1
aI5
aI3
aI31
aI25
aI81
aI59
aI211
aI463
aI1693
aI609
aI3307
aI3641
aI19643
aI29361
aI8399
asS'd'
I8987
sa(dp17974
S'a'
I27763
sS's'
I17
sS'm'
(lp17975
I1
aI1
aI7
aI13
aI19
aI47
aI43
aI43
aI275
aI735
aI535
aI3689
aI3987
aI10695
aI17243
aI60565
aI72299
asS'd'
I8988
sa(dp17976
S'a'
I27769
sS's'
I17
sS'm'
(lp17977
I1
aI3
aI3
aI5
aI25
aI35
aI75
aI63
aI305
aI127
aI189
aI1785
aI731
aI12089
aI27811
aI43259
aI28191
asS'd'
I8989
sa(dp17978
S'a'
I27770
sS's'
I17
sS'm'
(lp17979
I1
aI3
aI7
aI11
aI17
aI17
aI59
aI107
aI139
aI355
aI1055
aI3181
aI4743
aI14785
aI26323
aI441
aI35613
asS'd'
I8990
sa(dp17980
S'a'
I27796
sS's'
I17
sS'm'
(lp17981
I1
aI3
aI1
aI1
aI17
aI17
aI39
aI203
aI373
aI601
aI449
aI1837
aI835
aI7061
aI14655
aI61765
aI80735
asS'd'
I8991
sa(dp17982
S'a'
I27810
sS's'
I17
sS'm'
(lp17983
I1
aI3
aI5
aI7
aI27
aI17
aI25
aI41
aI125
aI895
aI1843
aI3167
aI1527
aI4707
aI6477
aI33575
aI97247
asS'd'
I8992
sa(dp17984
S'a'
I27812
sS's'
I17
sS'm'
(lp17985
I1
aI3
aI3
aI3
aI13
aI39
aI25
aI15
aI279
aI347
aI1121
aI3603
aI3019
aI9577
aI16863
aI61483
aI15995
asS'd'
I8993
sa(dp17986
S'a'
I27834
sS's'
I17
sS'm'
(lp17987
I1
aI3
aI5
aI11
aI15
aI33
aI15
aI81
aI185
aI289
aI909
aI1237
aI3623
aI3983
aI24211
aI62719
aI79685
asS'd'
I8994
sa(dp17988
S'a'
I27847
sS's'
I17
sS'm'
(lp17989
I1
aI1
aI1
aI7
aI29
aI1
aI53
aI17
aI137
aI269
aI1209
aI3937
aI4167
aI14057
aI8061
aI38863
aI101477
asS'd'
I8995
sa(dp17990
S'a'
I27848
sS's'
I17
sS'm'
(lp17991
I1
aI1
aI1
aI9
aI5
aI45
aI95
aI127
aI507
aI159
aI1763
aI1527
aI5689
aI11007
aI549
aI22837
aI99207
asS'd'
I8996
sa(dp17992
S'a'
I27862
sS's'
I17
sS'm'
(lp17993
I1
aI3
aI3
aI1
aI15
aI57
aI127
aI39
aI73
aI397
aI67
aI3159
aI119
aI8929
aI29425
aI57687
aI68063
asS'd'
I8997
sa(dp17994
S'a'
I27868
sS's'
I17
sS'm'
(lp17995
I1
aI3
aI1
aI3
aI27
aI7
aI111
aI209
aI291
aI17
aI1381
aI1597
aI5389
aI4577
aI20463
aI28325
aI23743
asS'd'
I8998
sa(dp17996
S'a'
I27881
sS's'
I17
sS'm'
(lp17997
I1
aI3
aI3
aI7
aI23
aI41
aI83
aI81
aI213
aI537
aI1037
aI2371
aI1485
aI8391
aI12471
aI58541
aI27559
asS'd'
I8999
sa(dp17998
S'a'
I27884
sS's'
I17
sS'm'
(lp17999
I1
aI3
aI1
aI15
aI21
aI43
aI87
aI75
aI451
aI851
aI1917
aI2739
aI2167
aI12531
aI29931
aI8017
aI15163
asS'd'
I9000
sa(dp18000
S'a'
I27890
sS's'
I17
sS'm'
(lp18001
I1
aI1
aI3
aI9
aI27
aI19
aI41
aI145
aI401
aI759
aI527
aI3085
aI187
aI10615
aI4995
aI22421
aI69867
asS'd'
I9001
sa(dp18002
S'a'
I27895
sS's'
I17
sS'm'
(lp18003
I1
aI3
aI3
aI13
aI29
aI51
aI65
aI47
aI157
aI609
aI1061
aI1913
aI6195
aI12503
aI10375
aI55819
aI122091
asS'd'
I9002
sa(dp18004
S'a'
I27904
sS's'
I17
sS'm'
(lp18005
I1
aI1
aI3
aI7
aI1
aI19
aI3
aI149
aI453
aI279
aI569
aI3429
aI331
aI711
aI26773
aI21163
aI129339
asS'd'
I9003
sa(dp18006
S'a'
I27914
sS's'
I17
sS'm'
(lp18007
I1
aI1
aI5
aI3
aI7
aI47
aI39
aI181
aI115
aI771
aI2037
aI411
aI2697
aI7501
aI6393
aI19461
aI74967
asS'd'
I9004
sa(dp18008
S'a'
I27919
sS's'
I17
sS'm'
(lp18009
I1
aI3
aI3
aI5
aI5
aI17
aI89
aI161
aI409
aI49
aI1447
aI3977
aI4777
aI15553
aI3521
aI32553
aI126385
asS'd'
I9005
sa(dp18010
S'a'
I27921
sS's'
I17
sS'm'
(lp18011
I1
aI3
aI1
aI3
aI25
aI59
aI73
aI105
aI505
aI1009
aI1147
aI317
aI3457
aI13743
aI8337
aI38077
aI7709
asS'd'
I9006
sa(dp18012
S'a'
I27931
sS's'
I17
sS'm'
(lp18013
I1
aI3
aI3
aI15
aI23
aI37
aI25
aI123
aI413
aI911
aI637
aI2345
aI691
aI15199
aI22927
aI52467
aI126715
asS'd'
I9007
sa(dp18014
S'a'
I27933
sS's'
I17
sS'm'
(lp18015
I1
aI1
aI5
aI1
aI9
aI51
aI93
aI123
aI269
aI45
aI1947
aI179
aI5091
aI3743
aI31491
aI39771
aI119175
asS'd'
I9008
sa(dp18016
S'a'
I27938
sS's'
I17
sS'm'
(lp18017
I1
aI3
aI1
aI5
aI29
aI23
aI107
aI183
aI25
aI115
aI187
aI857
aI7337
aI469
aI8755
aI17281
aI45941
asS'd'
I9009
sa(dp18018
S'a'
I27943
sS's'
I17
sS'm'
(lp18019
I1
aI3
aI1
aI13
aI25
aI61
aI85
aI115
aI181
aI955
aI1365
aI837
aI5941
aI13209
aI27009
aI58865
aI115149
asS'd'
I9010
sa(dp18020
S'a'
I27949
sS's'
I17
sS'm'
(lp18021
I1
aI1
aI1
aI11
aI31
aI63
aI101
aI29
aI37
aI185
aI465
aI2651
aI6249
aI6887
aI25021
aI60539
aI50037
asS'd'
I9011
sa(dp18022
S'a'
I28005
sS's'
I17
sS'm'
(lp18023
I1
aI3
aI1
aI7
aI7
aI61
aI57
aI243
aI143
aI223
aI1759
aI4085
aI6765
aI13293
aI31929
aI29579
aI35053
asS'd'
I9012
sa(dp18024
S'a'
I28012
sS's'
I17
sS'm'
(lp18025
I1
aI3
aI1
aI3
aI29
aI9
aI121
aI3
aI285
aI199
aI1439
aI3151
aI5059
aI8535
aI17049
aI38917
aI46347
asS'd'
I9013
sa(dp18026
S'a'
I28017
sS's'
I17
sS'm'
(lp18027
I1
aI3
aI1
aI3
aI17
aI43
aI63
aI87
aI427
aI341
aI1251
aI3775
aI7729
aI3183
aI10579
aI917
aI49035
asS'd'
I9014
sa(dp18028
S'a'
I28024
sS's'
I17
sS'm'
(lp18029
I1
aI1
aI7
aI3
aI5
aI59
aI119
aI227
aI495
aI345
aI841
aI2021
aI2483
aI15987
aI24663
aI9819
aI33009
asS'd'
I9015
sa(dp18030
S'a'
I28030
sS's'
I17
sS'm'
(lp18031
I1
aI1
aI5
aI11
aI19
aI57
aI23
aI181
aI63
aI991
aI1
aI2927
aI4785
aI9645
aI17435
aI55627
aI1069
asS'd'
I9016
sa(dp18032
S'a'
I28034
sS's'
I17
sS'm'
(lp18033
I1
aI1
aI7
aI1
aI31
aI11
aI57
aI123
aI279
aI815
aI1407
aI3509
aI3963
aI8503
aI20345
aI7777
aI103701
asS'd'
I9017
sa(dp18034
S'a'
I28045
sS's'
I17
sS'm'
(lp18035
I1
aI1
aI5
aI5
aI19
aI51
aI37
aI15
aI363
aI939
aI1863
aI3485
aI7073
aI3035
aI31279
aI7289
aI39811
asS'd'
I9018
sa(dp18036
S'a'
I28048
sS's'
I17
sS'm'
(lp18037
I1
aI1
aI3
aI3
aI3
aI41
aI29
aI37
aI311
aI535
aI1993
aI3937
aI309
aI13055
aI22595
aI59641
aI95317
asS'd'
I9019
sa(dp18038
S'a'
I28054
sS's'
I17
sS'm'
(lp18039
I1
aI3
aI7
aI9
aI19
aI29
aI23
aI181
aI503
aI223
aI1655
aI997
aI7861
aI5867
aI16979
aI4559
aI7447
asS'd'
I9020
sa(dp18040
S'a'
I28063
sS's'
I17
sS'm'
(lp18041
I1
aI3
aI5
aI3
aI13
aI13
aI3
aI137
aI361
aI101
aI1005
aI2339
aI609
aI11891
aI15245
aI9653
aI63579
asS'd'
I9021
sa(dp18042
S'a'
I28069
sS's'
I17
sS'm'
(lp18043
I1
aI1
aI1
aI15
aI31
aI15
aI117
aI151
aI51
aI805
aI1403
aI3243
aI4007
aI11979
aI3945
aI61935
aI43225
asS'd'
I9022
sa(dp18044
S'a'
I28070
sS's'
I17
sS'm'
(lp18045
I1
aI1
aI7
aI3
aI1
aI43
aI93
aI105
aI9
aI555
aI731
aI655
aI2097
aI8015
aI27557
aI27985
aI11323
asS'd'
I9023
sa(dp18046
S'a'
I28073
sS's'
I17
sS'm'
(lp18047
I1
aI3
aI3
aI9
aI23
aI35
aI59
aI217
aI437
aI755
aI685
aI1431
aI2965
aI5269
aI25621
aI21735
aI1397
asS'd'
I9024
sa(dp18048
S'a'
I28084
sS's'
I17
sS'm'
(lp18049
I1
aI1
aI5
aI1
aI9
aI61
aI41
aI53
aI101
aI111
aI531
aI3385
aI4771
aI9535
aI15995
aI29687
aI99035
asS'd'
I9025
sa(dp18050
S'a'
I28096
sS's'
I17
sS'm'
(lp18051
I1
aI1
aI7
aI7
aI1
aI3
aI25
aI251
aI463
aI99
aI677
aI1889
aI3697
aI5579
aI11429
aI38301
aI57917
asS'd'
I9026
sa(dp18052
S'a'
I28114
sS's'
I17
sS'm'
(lp18053
I1
aI1
aI5
aI9
aI11
aI15
aI65
aI31
aI369
aI825
aI1229
aI1595
aI3891
aI5235
aI16973
aI25307
aI7805
asS'd'
I9027
sa(dp18054
S'a'
I28126
sS's'
I17
sS'm'
(lp18055
I1
aI3
aI7
aI15
aI27
aI37
aI35
aI103
aI393
aI781
aI1713
aI2009
aI1973
aI15461
aI6801
aI17557
aI105139
asS'd'
I9028
sa(dp18056
S'a'
I28141
sS's'
I17
sS'm'
(lp18057
I1
aI3
aI7
aI7
aI17
aI51
aI83
aI29
aI113
aI173
aI1733
aI2137
aI3041
aI11361
aI15999
aI25779
aI112493
asS'd'
I9029
sa(dp18058
S'a'
I28149
sS's'
I17
sS'm'
(lp18059
I1
aI3
aI5
aI11
aI19
aI3
aI89
aI103
aI449
aI375
aI437
aI4077
aI889
aI12163
aI29323
aI48845
aI7783
asS'd'
I9030
sa(dp18060
S'a'
I28165
sS's'
I17
sS'm'
(lp18061
I1
aI3
aI7
aI1
aI19
aI25
aI83
aI35
aI203
aI27
aI507
aI25
aI6629
aI13941
aI6187
aI17533
aI83349
asS'd'
I9031
sa(dp18062
S'a'
I28177
sS's'
I17
sS'm'
(lp18063
I1
aI3
aI5
aI9
aI15
aI59
aI3
aI87
aI473
aI733
aI1263
aI1733
aI4275
aI9283
aI32535
aI20807
aI59487
asS'd'
I9032
sa(dp18064
S'a'
I28189
sS's'
I17
sS'm'
(lp18065
I1
aI3
aI3
aI9
aI19
aI11
aI27
aI83
aI355
aI949
aI1339
aI171
aI921
aI14171
aI16271
aI41485
aI20405
asS'd'
I9033
sa(dp18066
S'a'
I28190
sS's'
I17
sS'm'
(lp18067
I1
aI1
aI3
aI11
aI25
aI51
aI9
aI241
aI367
aI519
aI1895
aI429
aI7689
aI9469
aI32709
aI46363
aI75767
asS'd'
I9034
sa(dp18068
S'a'
I28203
sS's'
I17
sS'm'
(lp18069
I1
aI1
aI7
aI7
aI27
aI59
aI85
aI87
aI467
aI273
aI1763
aI391
aI451
aI16165
aI7501
aI44779
aI68281
asS'd'
I9035
sa(dp18070
S'a'
I28206
sS's'
I17
sS'm'
(lp18071
I1
aI1
aI7
aI5
aI1
aI35
aI5
aI217
aI31
aI145
aI1151
aI2255
aI3543
aI401
aI17141
aI5981
aI25183
asS'd'
I9036
sa(dp18072
S'a'
I28208
sS's'
I17
sS'm'
(lp18073
I1
aI1
aI1
aI13
aI11
aI11
aI19
aI93
aI95
aI751
aI31
aI1091
aI2733
aI10517
aI2553
aI5247
aI42651
asS'd'
I9037
sa(dp18074
S'a'
I28220
sS's'
I17
sS'm'
(lp18075
I1
aI3
aI7
aI5
aI15
aI1
aI67
aI21
aI303
aI137
aI355
aI1989
aI5211
aI4985
aI645
aI6867
aI126931
asS'd'
I9038
sa(dp18076
S'a'
I28226
sS's'
I17
sS'm'
(lp18077
I1
aI1
aI3
aI15
aI21
aI23
aI59
aI209
aI121
aI623
aI575
aI2447
aI6149
aI10481
aI4959
aI22913
aI64963
asS'd'
I9039
sa(dp18078
S'a'
I28231
sS's'
I17
sS'm'
(lp18079
I1
aI3
aI1
aI1
aI25
aI55
aI47
aI95
aI215
aI609
aI639
aI1023
aI1579
aI5953
aI3063
aI13721
aI17607
asS'd'
I9040
sa(dp18080
S'a'
I28249
sS's'
I17
sS'm'
(lp18081
I1
aI1
aI1
aI11
aI7
aI61
aI127
aI173
aI307
aI623
aI453
aI3827
aI4847
aI16085
aI4407
aI4043
aI14881
asS'd'
I9041
sa(dp18082
S'a'
I28256
sS's'
I17
sS'm'
(lp18083
I1
aI1
aI7
aI11
aI5
aI15
aI51
aI125
aI439
aI795
aI203
aI91
aI3543
aI6925
aI32055
aI52277
aI26841
asS'd'
I9042
sa(dp18084
S'a'
I28259
sS's'
I17
sS'm'
(lp18085
I1
aI1
aI1
aI13
aI15
aI9
aI107
aI183
aI391
aI751
aI243
aI1105
aI8031
aI7443
aI137
aI45695
aI80163
asS'd'
I9043
sa(dp18086
S'a'
I28265
sS's'
I17
sS'm'
(lp18087
I1
aI3
aI5
aI9
aI5
aI61
aI117
aI113
aI121
aI291
aI225
aI1705
aI4017
aI13113
aI11035
aI28613
aI25927
asS'd'
I9044
sa(dp18088
S'a'
I28266
sS's'
I17
sS'm'
(lp18089
I1
aI1
aI3
aI11
aI27
aI9
aI45
aI85
aI309
aI991
aI1639
aI1183
aI8013
aI14587
aI7563
aI21111
aI48497
asS'd'
I9045
sa(dp18090
S'a'
I28271
sS's'
I17
sS'm'
(lp18091
I1
aI3
aI1
aI9
aI21
aI61
aI123
aI189
aI265
aI593
aI163
aI3681
aI2271
aI2795
aI753
aI48019
aI129507
asS'd'
I9046
sa(dp18092
S'a'
I28274
sS's'
I17
sS'm'
(lp18093
I1
aI1
aI1
aI5
aI31
aI51
aI127
aI79
aI333
aI177
aI1723
aI1365
aI2055
aI3063
aI10693
aI61223
aI60659
asS'd'
I9047
sa(dp18094
S'a'
I28280
sS's'
I17
sS'm'
(lp18095
I1
aI3
aI7
aI15
aI9
aI11
aI11
aI223
aI31
aI397
aI319
aI3283
aI3765
aI4729
aI4711
aI58323
aI114063
asS'd'
I9048
sa(dp18096
S'a'
I28296
sS's'
I17
sS'm'
(lp18097
I1
aI1
aI7
aI11
aI7
aI63
aI107
aI215
aI19
aI491
aI131
aI2491
aI6373
aI11081
aI2159
aI1311
aI109547
asS'd'
I9049
sa(dp18098
S'a'
I28316
sS's'
I17
sS'm'
(lp18099
I1
aI3
aI5
aI5
aI21
aI13
aI105
aI21
aI193
aI447
aI1331
aI2439
aI4165
aI15689
aI21273
aI4007
aI55161
asS'd'
I9050
sa(dp18100
S'a'
I28319
sS's'
I17
sS'm'
(lp18101
I1
aI3
aI5
aI11
aI19
aI47
aI25
aI211
aI335
aI437
aI1041
aI2093
aI7239
aI2869
aI18273
aI40505
aI13681
asS'd'
I9051
sa(dp18102
S'a'
I28343
sS's'
I17
sS'm'
(lp18103
I1
aI3
aI3
aI15
aI7
aI13
aI127
aI59
aI439
aI163
aI1841
aI1945
aI4915
aI16269
aI18315
aI15057
aI43197
asS'd'
I9052
sa(dp18104
S'a'
I28344
sS's'
I17
sS'm'
(lp18105
I1
aI1
aI3
aI3
aI15
aI33
aI101
aI241
aI131
aI353
aI1749
aI3965
aI1231
aI8167
aI9309
aI40337
aI4419
asS'd'
I9053
sa(dp18106
S'a'
I28357
sS's'
I17
sS'm'
(lp18107
I1
aI3
aI1
aI1
aI1
aI43
aI33
aI129
aI137
aI889
aI799
aI2937
aI3633
aI4769
aI2411
aI56059
aI585
asS'd'
I9054
sa(dp18108
S'a'
I28358
sS's'
I17
sS'm'
(lp18109
I1
aI1
aI3
aI9
aI25
aI31
aI45
aI199
aI229
aI175
aI1099
aI1143
aI1721
aI11811
aI22757
aI59843
aI117813
asS'd'
I9055
sa(dp18110
S'a'
I28364
sS's'
I17
sS'm'
(lp18111
I1
aI1
aI7
aI7
aI19
aI45
aI43
aI101
aI219
aI209
aI1169
aI599
aI5523
aI2463
aI15161
aI16675
aI85111
asS'd'
I9056
sa(dp18112
S'a'
I28370
sS's'
I17
sS'm'
(lp18113
I1
aI1
aI3
aI15
aI23
aI27
aI91
aI51
aI397
aI705
aI651
aI2345
aI3875
aI10005
aI29523
aI42805
aI76891
asS'd'
I9057
sa(dp18114
S'a'
I28372
sS's'
I17
sS'm'
(lp18115
I1
aI3
aI3
aI5
aI29
aI49
aI17
aI233
aI149
aI821
aI1953
aI1931
aI7127
aI957
aI6477
aI21259
aI126657
asS'd'
I9058
sa(dp18116
S'a'
I28379
sS's'
I17
sS'm'
(lp18117
I1
aI3
aI3
aI3
aI27
aI49
aI57
aI145
aI143
aI1
aI583
aI3987
aI651
aI12285
aI20139
aI51063
aI21449
asS'd'
I9059
sa(dp18118
S'a'
I28388
sS's'
I17
sS'm'
(lp18119
I1
aI1
aI3
aI5
aI29
aI61
aI41
aI93
aI277
aI111
aI395
aI2929
aI5325
aI15183
aI3981
aI23799
aI72781
asS'd'
I9060
sa(dp18120
S'a'
I28392
sS's'
I17
sS'm'
(lp18121
I1
aI1
aI7
aI5
aI25
aI43
aI85
aI137
aI401
aI261
aI1183
aI3959
aI1983
aI16209
aI30523
aI429
aI109181
asS'd'
I9061
sa(dp18122
S'a'
I28395
sS's'
I17
sS'm'
(lp18123
I1
aI1
aI1
aI5
aI7
aI19
aI79
aI237
aI373
aI929
aI907
aI1771
aI6991
aI211
aI5269
aI2135
aI32051
asS'd'
I9062
sa(dp18124
S'a'
I28406
sS's'
I17
sS'm'
(lp18125
I1
aI1
aI7
aI5
aI17
aI15
aI41
aI49
aI363
aI15
aI1483
aI1017
aI2439
aI11713
aI19983
aI26275
aI11945
asS'd'
I9063
sa(dp18126
S'a'
I28418
sS's'
I17
sS'm'
(lp18127
I1
aI1
aI7
aI7
aI19
aI5
aI55
aI15
aI15
aI573
aI1075
aI3225
aI6815
aI11893
aI18417
aI50833
aI71903
asS'd'
I9064
sa(dp18128
S'a'
I28423
sS's'
I17
sS'm'
(lp18129
I1
aI1
aI7
aI9
aI23
aI37
aI75
aI3
aI477
aI291
aI37
aI1861
aI2697
aI13369
aI24573
aI27285
aI96757
asS'd'
I9065
sa(dp18130
S'a'
I28435
sS's'
I17
sS'm'
(lp18131
I1
aI1
aI1
aI13
aI5
aI29
aI65
aI195
aI365
aI465
aI865
aI2705
aI5249
aI7051
aI3795
aI56611
aI72317
asS'd'
I9066
sa(dp18132
S'a'
I28444
sS's'
I17
sS'm'
(lp18133
I1
aI1
aI3
aI9
aI19
aI9
aI85
aI239
aI509
aI313
aI1137
aI2221
aI5475
aI71
aI11901
aI1877
aI68701
asS'd'
I9067
sa(dp18134
S'a'
I28454
sS's'
I17
sS'm'
(lp18135
I1
aI3
aI3
aI7
aI3
aI53
aI55
aI223
aI441
aI159
aI933
aI2573
aI3441
aI3295
aI25005
aI29021
aI97145
asS'd'
I9068
sa(dp18136
S'a'
I28468
sS's'
I17
sS'm'
(lp18137
I1
aI3
aI5
aI5
aI3
aI11
aI101
aI181
aI293
aI319
aI47
aI2971
aI387
aI4697
aI26613
aI8531
aI20461
asS'd'
I9069
sa(dp18138
S'a'
I28477
sS's'
I17
sS'm'
(lp18139
I1
aI3
aI3
aI13
aI17
aI11
aI41
aI233
aI455
aI353
aI1817
aI3065
aI4657
aI1717
aI3039
aI10937
aI107085
asS'd'
I9070
sa(dp18140
S'a'
I28478
sS's'
I17
sS'm'
(lp18141
I1
aI1
aI7
aI1
aI17
aI23
aI85
aI5
aI291
aI725
aI1791
aI3525
aI7705
aI6561
aI25311
aI44679
aI21419
asS'd'
I9071
sa(dp18142
S'a'
I28480
sS's'
I17
sS'm'
(lp18143
I1
aI1
aI7
aI1
aI23
aI41
aI97
aI117
aI435
aI261
aI2007
aI965
aI6913
aI12245
aI25723
aI8445
aI30871
asS'd'
I9072
sa(dp18144
S'a'
I28486
sS's'
I17
sS'm'
(lp18145
I1
aI1
aI1
aI15
aI29
aI39
aI101
aI33
aI55
aI1019
aI1431
aI2601
aI3837
aI14873
aI11785
aI12449
aI30815
asS'd'
I9073
sa(dp18146
S'a'
I28489
sS's'
I17
sS'm'
(lp18147
I1
aI1
aI5
aI3
aI15
aI35
aI101
aI7
aI479
aI535
aI1875
aI2435
aI1461
aI13967
aI2755
aI45879
aI93561
asS'd'
I9074
sa(dp18148
S'a'
I28500
sS's'
I17
sS'm'
(lp18149
I1
aI3
aI1
aI5
aI29
aI57
aI89
aI209
aI473
aI289
aI1843
aI2051
aI3997
aI1753
aI18179
aI41355
aI89301
asS'd'
I9075
sa(dp18150
S'a'
I28507
sS's'
I17
sS'm'
(lp18151
I1
aI3
aI1
aI11
aI17
aI45
aI47
aI57
aI109
aI309
aI1009
aI653
aI5175
aI15599
aI21617
aI35353
aI55253
asS'd'
I9076
sa(dp18152
S'a'
I28519
sS's'
I17
sS'm'
(lp18153
I1
aI3
aI5
aI3
aI1
aI11
aI57
aI83
aI385
aI765
aI1887
aI785
aI2115
aI8689
aI14783
aI14841
aI122221
asS'd'
I9077
sa(dp18154
S'a'
I28523
sS's'
I17
sS'm'
(lp18155
I1
aI3
aI5
aI11
aI11
aI5
aI77
aI115
aI189
aI371
aI887
aI3653
aI8159
aI15737
aI6763
aI52807
aI128841
asS'd'
I9078
sa(dp18156
S'a'
I28534
sS's'
I17
sS'm'
(lp18157
I1
aI1
aI1
aI11
aI11
aI33
aI9
aI145
aI439
aI565
aI171
aI3867
aI7149
aI4369
aI15073
aI3277
aI25873
asS'd'
I9079
sa(dp18158
S'a'
I28547
sS's'
I17
sS'm'
(lp18159
I1
aI1
aI3
aI1
aI11
aI9
aI17
aI255
aI129
aI835
aI1705
aI1551
aI877
aI4831
aI12717
aI2549
aI62723
asS'd'
I9080
sa(dp18160
S'a'
I28549
sS's'
I17
sS'm'
(lp18161
I1
aI1
aI7
aI11
aI17
aI33
aI21
aI195
aI143
aI153
aI1855
aI1323
aI1225
aI16359
aI16479
aI8883
aI76449
asS'd'
I9081
sa(dp18162
S'a'
I28562
sS's'
I17
sS'm'
(lp18163
I1
aI1
aI1
aI5
aI31
aI23
aI61
aI53
aI77
aI465
aI1983
aI4019
aI4865
aI14721
aI18601
aI48179
aI100453
asS'd'
I9082
sa(dp18164
S'a'
I28571
sS's'
I17
sS'm'
(lp18165
I1
aI3
aI1
aI13
aI19
aI53
aI83
aI63
aI165
aI393
aI469
aI1465
aI2669
aI10155
aI7029
aI26185
aI121223
asS'd'
I9083
sa(dp18166
S'a'
I28583
sS's'
I17
sS'm'
(lp18167
I1
aI1
aI1
aI3
aI3
aI3
aI123
aI23
aI45
aI359
aI1063
aI847
aI3943
aI6113
aI23749
aI30323
aI10583
asS'd'
I9084
sa(dp18168
S'a'
I28584
sS's'
I17
sS'm'
(lp18169
I1
aI3
aI5
aI15
aI1
aI55
aI65
aI149
aI139
aI217
aI141
aI2425
aI7019
aI14127
aI11725
aI50821
aI52643
asS'd'
I9085
sa(dp18170
S'a'
I28595
sS's'
I17
sS'm'
(lp18171
I1
aI3
aI5
aI15
aI13
aI13
aI15
aI93
aI457
aI869
aI117
aI585
aI7159
aI5957
aI15073
aI21861
aI118119
asS'd'
I9086
sa(dp18172
S'a'
I28597
sS's'
I17
sS'm'
(lp18173
I1
aI3
aI1
aI15
aI3
aI31
aI29
aI245
aI47
aI895
aI197
aI1169
aI945
aI11503
aI26615
aI14079
aI27175
asS'd'
I9087
sa(dp18174
S'a'
I28604
sS's'
I17
sS'm'
(lp18175
I1
aI3
aI5
aI5
aI31
aI1
aI107
aI109
aI253
aI999
aI1451
aI2243
aI6675
aI10779
aI26181
aI64597
aI16443
asS'd'
I9088
sa(dp18176
S'a'
I28607
sS's'
I17
sS'm'
(lp18177
I1
aI3
aI7
aI15
aI9
aI53
aI25
aI41
aI151
aI197
aI1955
aI2365
aI5305
aI2901
aI24825
aI9595
aI57377
asS'd'
I9089
sa(dp18178
S'a'
I28609
sS's'
I17
sS'm'
(lp18179
I1
aI3
aI1
aI3
aI25
aI37
aI37
aI193
aI417
aI373
aI1127
aI3239
aI4583
aI2861
aI14501
aI64163
aI30055
asS'd'
I9090
sa(dp18180
S'a'
I28612
sS's'
I17
sS'm'
(lp18181
I1
aI3
aI7
aI9
aI7
aI21
aI49
aI231
aI241
aI95
aI1757
aI2281
aI2679
aI1611
aI11115
aI31743
aI26851
asS'd'
I9091
sa(dp18182
S'a'
I28630
sS's'
I17
sS'm'
(lp18183
I1
aI3
aI5
aI5
aI1
aI1
aI23
aI173
aI195
aI593
aI1639
aI1449
aI4733
aI2451
aI12677
aI31959
aI128217
asS'd'
I9092
sa(dp18184
S'a'
I28640
sS's'
I17
sS'm'
(lp18185
I1
aI1
aI1
aI7
aI17
aI49
aI117
aI253
aI167
aI721
aI889
aI3027
aI7781
aI13521
aI15477
aI17981
aI95487
asS'd'
I9093
sa(dp18186
S'a'
I28669
sS's'
I17
sS'm'
(lp18187
I1
aI1
aI1
aI15
aI13
aI47
aI125
aI9
aI33
aI567
aI1733
aI1263
aI307
aI10285
aI6325
aI55827
aI39823
asS'd'
I9094
sa(dp18188
S'a'
I28677
sS's'
I17
sS'm'
(lp18189
I1
aI1
aI1
aI15
aI23
aI3
aI11
aI169
aI369
aI667
aI313
aI2287
aI6655
aI16067
aI5915
aI8605
aI92177
asS'd'
I9095
sa(dp18190
S'a'
I28682
sS's'
I17
sS'm'
(lp18191
I1
aI3
aI3
aI15
aI13
aI21
aI125
aI111
aI171
aI785
aI79
aI2281
aI1247
aI11321
aI30397
aI28555
aI84863
asS'd'
I9096
sa(dp18192
S'a'
I28690
sS's'
I17
sS'm'
(lp18193
I1
aI1
aI5
aI13
aI1
aI31
aI123
aI97
aI127
aI245
aI1213
aI2381
aI3545
aI13545
aI28657
aI54087
aI79039
asS'd'
I9097
sa(dp18194
S'a'
I28695
sS's'
I17
sS'm'
(lp18195
I1
aI1
aI7
aI9
aI9
aI21
aI87
aI111
aI27
aI33
aI843
aI199
aI1465
aI6555
aI8019
aI39521
aI98883
asS'd'
I9098
sa(dp18196
S'a'
I28696
sS's'
I17
sS'm'
(lp18197
I1
aI3
aI3
aI5
aI5
aI55
aI61
aI219
aI279
aI207
aI1811
aI667
aI2989
aI3133
aI25213
aI51979
aI87695
asS'd'
I9099
sa(dp18198
S'a'
I28717
sS's'
I17
sS'm'
(lp18199
I1
aI1
aI7
aI3
aI17
aI11
aI31
aI97
aI277
aI385
aI229
aI3045
aI557
aI13521
aI32733
aI36831
aI43003
asS'd'
I9100
sa(dp18200
S'a'
I28730
sS's'
I17
sS'm'
(lp18201
I1
aI3
aI5
aI13
aI27
aI57
aI31
aI207
aI147
aI405
aI1495
aI2471
aI4889
aI14861
aI4861
aI28185
aI62363
asS'd'
I9101
sa(dp18202
S'a'
I28737
sS's'
I17
sS'm'
(lp18203
I1
aI1
aI5
aI13
aI23
aI19
aI5
aI21
aI509
aI299
aI1077
aI1747
aI6229
aI2375
aI17903
aI58473
aI72637
asS'd'
I9102
sa(dp18204
S'a'
I28752
sS's'
I17
sS'm'
(lp18205
I1
aI3
aI7
aI11
aI15
aI61
aI63
aI165
aI27
aI461
aI1359
aI3375
aI3029
aI9907
aI17393
aI11097
aI43593
asS'd'
I9103
sa(dp18206
S'a'
I28761
sS's'
I17
sS'm'
(lp18207
I1
aI3
aI5
aI1
aI17
aI29
aI15
aI5
aI419
aI19
aI1981
aI3169
aI2389
aI9169
aI31697
aI26201
aI6997
asS'd'
I9104
sa(dp18208
S'a'
I28774
sS's'
I17
sS'm'
(lp18209
I1
aI3
aI3
aI1
aI29
aI31
aI89
aI79
aI347
aI707
aI505
aI2087
aI2163
aI5465
aI8677
aI11421
aI93217
asS'd'
I9105
sa(dp18210
S'a'
I28783
sS's'
I17
sS'm'
(lp18211
I1
aI3
aI7
aI9
aI3
aI23
aI75
aI215
aI7
aI971
aI925
aI3223
aI7825
aI12347
aI19763
aI10927
aI41245
asS'd'
I9106
sa(dp18212
S'a'
I28791
sS's'
I17
sS'm'
(lp18213
I1
aI3
aI5
aI5
aI3
aI43
aI119
aI79
aI373
aI761
aI709
aI1897
aI3953
aI13895
aI13421
aI16939
aI112449
asS'd'
I9107
sa(dp18214
S'a'
I28801
sS's'
I17
sS'm'
(lp18215
I1
aI1
aI3
aI15
aI11
aI25
aI65
aI101
aI315
aI1005
aI1319
aI1163
aI161
aI15331
aI4845
aI40375
aI121361
asS'd'
I9108
sa(dp18216
S'a'
I28802
sS's'
I17
sS'm'
(lp18217
I1
aI1
aI3
aI7
aI1
aI57
aI63
aI131
aI145
aI1007
aI549
aI2327
aI1513
aI3591
aI10839
aI56297
aI80613
asS'd'
I9109
sa(dp18218
S'a'
I28814
sS's'
I17
sS'm'
(lp18219
I1
aI1
aI3
aI5
aI1
aI19
aI79
aI81
aI505
aI945
aI65
aI319
aI6105
aI5491
aI13257
aI4651
aI48247
asS'd'
I9110
sa(dp18220
S'a'
I28837
sS's'
I17
sS'm'
(lp18221
I1
aI3
aI1
aI9
aI27
aI41
aI61
aI41
aI421
aI707
aI1279
aI3699
aI2403
aI4075
aI16947
aI53435
aI2917
asS'd'
I9111
sa(dp18222
S'a'
I28838
sS's'
I17
sS'm'
(lp18223
I1
aI1
aI5
aI13
aI11
aI29
aI35
aI141
aI313
aI769
aI749
aI4025
aI2597
aI12197
aI32265
aI32159
aI37003
asS'd'
I9112
sa(dp18224
S'a'
I28842
sS's'
I17
sS'm'
(lp18225
I1
aI3
aI7
aI11
aI25
aI63
aI121
aI15
aI273
aI877
aI637
aI409
aI5001
aI4723
aI27985
aI55501
aI43495
asS'd'
I9113
sa(dp18226
S'a'
I28844
sS's'
I17
sS'm'
(lp18227
I1
aI3
aI1
aI13
aI27
aI29
aI127
aI65
aI275
aI967
aI1723
aI4007
aI6147
aI13277
aI8361
aI12305
aI95433
asS'd'
I9114
sa(dp18228
S'a'
I28849
sS's'
I17
sS'm'
(lp18229
I1
aI3
aI3
aI13
aI11
aI45
aI7
aI101
aI169
aI361
aI517
aI2897
aI4283
aI7587
aI7495
aI31549
aI29113
asS'd'
I9115
sa(dp18230
S'a'
I28864
sS's'
I17
sS'm'
(lp18231
I1
aI3
aI3
aI1
aI9
aI27
aI65
aI15
aI279
aI127
aI1039
aI829
aI5739
aI1949
aI24789
aI30433
aI54503
asS'd'
I9116
sa(dp18232
S'a'
I28867
sS's'
I17
sS'm'
(lp18233
I1
aI3
aI7
aI5
aI13
aI19
aI95
aI133
aI25
aI271
aI1527
aI3571
aI101
aI15987
aI10985
aI55761
aI39833
asS'd'
I9117
sa(dp18234
S'a'
I28881
sS's'
I17
sS'm'
(lp18235
I1
aI3
aI5
aI9
aI27
aI5
aI37
aI219
aI249
aI947
aI1063
aI4081
aI1763
aI15003
aI23753
aI3975
aI109803
asS'd'
I9118
sa(dp18236
S'a'
I28882
sS's'
I17
sS'm'
(lp18237
I1
aI3
aI3
aI5
aI21
aI37
aI35
aI145
aI323
aI573
aI1939
aI885
aI4645
aI4515
aI16815
aI28783
aI76017
asS'd'
I9119
sa(dp18238
S'a'
I28893
sS's'
I17
sS'm'
(lp18239
I1
aI3
aI7
aI13
aI27
aI41
aI39
aI123
aI423
aI949
aI1487
aI2207
aI8069
aI15337
aI20671
aI20163
aI70667
asS'd'
I9120
sa(dp18240
S'a'
I28903
sS's'
I17
sS'm'
(lp18241
I1
aI1
aI3
aI3
aI15
aI29
aI69
aI15
aI151
aI729
aI35
aI2067
aI6715
aI12759
aI27611
aI54133
aI16561
asS'd'
I9121
sa(dp18242
S'a'
I28907
sS's'
I17
sS'm'
(lp18243
I1
aI3
aI7
aI13
aI21
aI13
aI7
aI161
aI327
aI339
aI535
aI2059
aI413
aI11161
aI18415
aI12415
aI63713
asS'd'
I9122
sa(dp18244
S'a'
I28909
sS's'
I17
sS'm'
(lp18245
I1
aI3
aI5
aI5
aI23
aI49
aI9
aI181
aI417
aI339
aI1013
aI1707
aI5097
aI13319
aI18951
aI56415
aI14397
asS'd'
I9123
sa(dp18246
S'a'
I28921
sS's'
I17
sS'm'
(lp18247
I1
aI1
aI5
aI7
aI29
aI23
aI117
aI159
aI287
aI695
aI71
aI2393
aI2655
aI6417
aI24349
aI20441
aI77987
asS'd'
I9124
sa(dp18248
S'a'
I28927
sS's'
I17
sS'm'
(lp18249
I1
aI1
aI5
aI7
aI31
aI23
aI81
aI125
aI145
aI141
aI1459
aI4095
aI713
aI1817
aI9263
aI21025
aI91983
asS'd'
I9125
sa(dp18250
S'a'
I28929
sS's'
I17
sS'm'
(lp18251
I1
aI1
aI1
aI9
aI17
aI23
aI91
aI39
aI459
aI299
aI1951
aI3229
aI6229
aI3267
aI15883
aI31719
aI96573
asS'd'
I9126
sa(dp18252
S'a'
I28935
sS's'
I17
sS'm'
(lp18253
I1
aI1
aI1
aI15
aI3
aI51
aI9
aI7
aI455
aI653
aI1447
aI153
aI8117
aI723
aI2177
aI9107
aI7757
asS'd'
I9127
sa(dp18254
S'a'
I28941
sS's'
I17
sS'm'
(lp18255
I1
aI3
aI1
aI15
aI27
aI47
aI49
aI245
aI499
aI807
aI175
aI1653
aI1693
aI3931
aI27479
aI30095
aI62353
asS'd'
I9128
sa(dp18256
S'a'
I28942
sS's'
I17
sS'm'
(lp18257
I1
aI1
aI5
aI5
aI23
aI7
aI15
aI193
aI499
aI193
aI201
aI2771
aI4153
aI11533
aI25883
aI23337
aI126685
asS'd'
I9129
sa(dp18258
S'a'
I28954
sS's'
I17
sS'm'
(lp18259
I1
aI1
aI1
aI7
aI9
aI43
aI125
aI181
aI425
aI557
aI1401
aI2593
aI1933
aI6803
aI20021
aI32687
aI126465
asS'd'
I9130
sa(dp18260
S'a'
I28959
sS's'
I17
sS'm'
(lp18261
I1
aI3
aI3
aI13
aI27
aI19
aI99
aI29
aI395
aI765
aI1137
aI2963
aI7397
aI9695
aI19259
aI27965
aI83157
asS'd'
I9131
sa(dp18262
S'a'
I28960
sS's'
I17
sS'm'
(lp18263
I1
aI3
aI7
aI7
aI17
aI29
aI7
aI241
aI5
aI281
aI1489
aI1599
aI5567
aI4579
aI7739
aI41413
aI110571
asS'd'
I9132
sa(dp18264
S'a'
I28990
sS's'
I17
sS'm'
(lp18265
I1
aI3
aI3
aI9
aI7
aI5
aI83
aI1
aI231
aI1003
aI631
aI2557
aI831
aI6495
aI30409
aI53519
aI79331
asS'd'
I9133
sa(dp18266
S'a'
I28992
sS's'
I17
sS'm'
(lp18267
I1
aI1
aI5
aI1
aI7
aI49
aI45
aI241
aI201
aI551
aI1645
aI2003
aI1455
aI3317
aI23639
aI7841
aI100765
asS'd'
I9134
sa(dp18268
S'a'
I29004
sS's'
I17
sS'm'
(lp18269
I1
aI3
aI5
aI13
aI25
aI47
aI103
aI37
aI81
aI263
aI1143
aI801
aI5863
aI6871
aI8615
aI57363
aI90161
asS'd'
I9135
sa(dp18270
S'a'
I29010
sS's'
I17
sS'm'
(lp18271
I1
aI3
aI7
aI11
aI27
aI23
aI119
aI211
aI473
aI207
aI605
aI637
aI3369
aI7287
aI27827
aI25003
aI65925
asS'd'
I9136
sa(dp18272
S'a'
I29012
sS's'
I17
sS'm'
(lp18273
I1
aI3
aI1
aI15
aI27
aI31
aI97
aI247
aI75
aI893
aI1099
aI3609
aI6807
aI4393
aI10253
aI62759
aI89971
asS'd'
I9137
sa(dp18274
S'a'
I29021
sS's'
I17
sS'm'
(lp18275
I1
aI1
aI7
aI15
aI27
aI19
aI43
aI59
aI419
aI263
aI387
aI3193
aI5589
aI4197
aI19143
aI64749
aI103093
asS'd'
I9138
sa(dp18276
S'a'
I29035
sS's'
I17
sS'm'
(lp18277
I1
aI1
aI7
aI11
aI21
aI51
aI97
aI227
aI251
aI869
aI1927
aI2331
aI7741
aI8207
aI12885
aI13267
aI17945
asS'd'
I9139
sa(dp18278
S'a'
I29040
sS's'
I17
sS'm'
(lp18279
I1
aI1
aI7
aI7
aI5
aI53
aI41
aI147
aI75
aI709
aI607
aI1073
aI2853
aI8081
aI12797
aI5279
aI86083
asS'd'
I9140
sa(dp18280
S'a'
I29059
sS's'
I17
sS'm'
(lp18281
I1
aI1
aI5
aI5
aI15
aI21
aI77
aI189
aI269
aI595
aI197
aI3117
aI5073
aI14277
aI26867
aI49505
aI75755
asS'd'
I9141
sa(dp18282
S'a'
I29068
sS's'
I17
sS'm'
(lp18283
I1
aI3
aI5
aI15
aI13
aI55
aI1
aI223
aI135
aI367
aI735
aI3139
aI4851
aI9773
aI11699
aI19081
aI26011
asS'd'
I9142
sa(dp18284
S'a'
I29080
sS's'
I17
sS'm'
(lp18285
I1
aI3
aI5
aI7
aI9
aI3
aI89
aI103
aI321
aI727
aI1809
aI3527
aI6881
aI2399
aI13593
aI27867
aI16219
asS'd'
I9143
sa(dp18286
S'a'
I29086
sS's'
I17
sS'm'
(lp18287
I1
aI3
aI3
aI1
aI23
aI5
aI53
aI51
aI403
aI753
aI2037
aI1261
aI7575
aI2725
aI11341
aI18533
aI98767
asS'd'
I9144
sa(dp18288
S'a'
I29089
sS's'
I17
sS'm'
(lp18289
I1
aI1
aI1
aI11
aI1
aI13
aI37
aI141
aI477
aI689
aI1789
aI1913
aI5753
aI6069
aI6965
aI55209
aI77329
asS'd'
I9145
sa(dp18290
S'a'
I29090
sS's'
I17
sS'm'
(lp18291
I1
aI3
aI7
aI3
aI17
aI59
aI79
aI249
aI381
aI163
aI1773
aI1645
aI7295
aI2359
aI21889
aI28429
aI117073
asS'd'
I9146
sa(dp18292
S'a'
I29095
sS's'
I17
sS'm'
(lp18293
I1
aI3
aI5
aI15
aI7
aI45
aI59
aI3
aI93
aI259
aI1257
aI2967
aI1175
aI10171
aI873
aI51423
aI67437
asS'd'
I9147
sa(dp18294
S'a'
I29113
sS's'
I17
sS'm'
(lp18295
I1
aI1
aI7
aI13
aI17
aI33
aI53
aI217
aI159
aI683
aI1169
aI3363
aI4591
aI3959
aI10089
aI35443
aI99677
asS'd'
I9148
sa(dp18296
S'a'
I29127
sS's'
I17
sS'm'
(lp18297
I1
aI3
aI7
aI9
aI3
aI1
aI5
aI171
aI17
aI635
aI369
aI1529
aI4861
aI4977
aI29303
aI42357
aI69309
asS'd'
I9149
sa(dp18298
S'a'
I29131
sS's'
I17
sS'm'
(lp18299
I1
aI3
aI7
aI9
aI21
aI17
aI77
aI127
aI105
aI427
aI525
aI1123
aI2365
aI7487
aI6315
aI64773
aI122747
asS'd'
I9150
sa(dp18300
S'a'
I29148
sS's'
I17
sS'm'
(lp18301
I1
aI1
aI1
aI15
aI19
aI63
aI65
aI83
aI219
aI987
aI169
aI2589
aI3809
aI8807
aI22473
aI6479
aI44617
asS'd'
I9151
sa(dp18302
S'a'
I29161
sS's'
I17
sS'm'
(lp18303
I1
aI3
aI1
aI7
aI11
aI51
aI107
aI19
aI379
aI191
aI1013
aI3145
aI1501
aI11871
aI14111
aI18269
aI98247
asS'd'
I9152
sa(dp18304
S'a'
I29172
sS's'
I17
sS'm'
(lp18305
I1
aI1
aI7
aI5
aI17
aI63
aI23
aI231
aI423
aI855
aI1955
aI907
aI4553
aI16173
aI7701
aI40329
aI42047
asS'd'
I9153
sa(dp18306
S'a'
I29179
sS's'
I17
sS'm'
(lp18307
I1
aI3
aI7
aI1
aI7
aI45
aI103
aI127
aI185
aI721
aI1035
aI839
aI691
aI6823
aI23819
aI50781
aI92767
asS'd'
I9154
sa(dp18308
S'a'
I29188
sS's'
I17
sS'm'
(lp18309
I1
aI1
aI1
aI3
aI9
aI21
aI57
aI253
aI285
aI85
aI1227
aI365
aI2347
aI7491
aI15183
aI8619
aI108819
asS'd'
I9155
sa(dp18310
S'a'
I29192
sS's'
I17
sS'm'
(lp18311
I1
aI1
aI3
aI15
aI27
aI13
aI5
aI85
aI45
aI1009
aI1315
aI1749
aI2797
aI3941
aI19367
aI50855
aI60693
asS'd'
I9156
sa(dp18312
S'a'
I29195
sS's'
I17
sS'm'
(lp18313
I1
aI3
aI5
aI15
aI29
aI63
aI115
aI197
aI317
aI601
aI711
aI377
aI7489
aI4247
aI4843
aI56549
aI108447
asS'd'
I9157
sa(dp18314
S'a'
I29206
sS's'
I17
sS'm'
(lp18315
I1
aI3
aI7
aI15
aI11
aI25
aI7
aI145
aI371
aI395
aI1743
aI267
aI2609
aI15707
aI13909
aI55031
aI71115
asS'd'
I9158
sa(dp18316
S'a'
I29210
sS's'
I17
sS'm'
(lp18317
I1
aI3
aI1
aI1
aI1
aI53
aI111
aI245
aI433
aI309
aI245
aI15
aI2091
aI9051
aI11095
aI31821
aI104535
asS'd'
I9159
sa(dp18318
S'a'
I29215
sS's'
I17
sS'm'
(lp18319
I1
aI3
aI1
aI15
aI25
aI31
aI99
aI89
aI259
aI595
aI1095
aI3681
aI5105
aI8671
aI23663
aI29717
aI126429
asS'd'
I9160
sa(dp18320
S'a'
I29221
sS's'
I17
sS'm'
(lp18321
I1
aI3
aI7
aI7
aI5
aI31
aI15
aI59
aI109
aI527
aI257
aI1785
aI6799
aI1283
aI11741
aI34589
aI102397
asS'd'
I9161
sa(dp18322
S'a'
I29222
sS's'
I17
sS'm'
(lp18323
I1
aI3
aI3
aI15
aI9
aI27
aI55
aI35
aI291
aI587
aI1281
aI779
aI4615
aI373
aI24037
aI64671
aI124019
asS'd'
I9162
sa(dp18324
S'a'
I29236
sS's'
I17
sS'm'
(lp18325
I1
aI1
aI5
aI5
aI13
aI51
aI49
aI19
aI37
aI857
aI539
aI1291
aI6021
aI8645
aI30351
aI33839
aI111515
asS'd'
I9163
sa(dp18326
S'a'
I29246
sS's'
I17
sS'm'
(lp18327
I1
aI1
aI5
aI13
aI3
aI47
aI9
aI197
aI19
aI337
aI2025
aI3003
aI7179
aI5755
aI31187
aI59317
aI69753
asS'd'
I9164
sa(dp18328
S'a'
I29248
sS's'
I17
sS'm'
(lp18329
I1
aI1
aI7
aI3
aI3
aI43
aI11
aI3
aI123
aI29
aI857
aI3349
aI791
aI11157
aI23967
aI33729
aI28445
asS'd'
I9165
sa(dp18330
S'a'
I29253
sS's'
I17
sS'm'
(lp18331
I1
aI1
aI5
aI1
aI1
aI11
aI73
aI231
aI173
aI925
aI331
aI161
aI3303
aI11015
aI17507
aI21271
aI56865
asS'd'
I9166
sa(dp18332
S'a'
I29272
sS's'
I17
sS'm'
(lp18333
I1
aI1
aI3
aI9
aI21
aI21
aI115
aI145
aI421
aI981
aI1789
aI3343
aI7591
aI12043
aI5795
aI17737
aI106501
asS'd'
I9167
sa(dp18334
S'a'
I29288
sS's'
I17
sS'm'
(lp18335
I1
aI1
aI7
aI13
aI7
aI15
aI51
aI75
aI497
aI637
aI1073
aI1505
aI5613
aI1415
aI30861
aI26159
aI79573
asS'd'
I9168
sa(dp18336
S'a'
I29293
sS's'
I17
sS'm'
(lp18337
I1
aI1
aI3
aI15
aI17
aI35
aI17
aI129
aI169
aI283
aI1383
aI149
aI211
aI1381
aI22205
aI28367
aI831
asS'd'
I9169
sa(dp18338
S'a'
I29294
sS's'
I17
sS'm'
(lp18339
I1
aI3
aI5
aI5
aI17
aI11
aI127
aI183
aI503
aI499
aI2011
aI2721
aI2717
aI3105
aI4731
aI60319
aI9361
asS'd'
I9170
sa(dp18340
S'a'
I29308
sS's'
I17
sS'm'
(lp18341
I1
aI1
aI1
aI7
aI27
aI55
aI77
aI203
aI255
aI761
aI1159
aI2915
aI4479
aI13671
aI19757
aI65497
aI4461
asS'd'
I9171
sa(dp18342
S'a'
I29318
sS's'
I17
sS'm'
(lp18343
I1
aI3
aI1
aI9
aI9
aI51
aI67
aI205
aI445
aI35
aI371
aI1837
aI3623
aI10365
aI19463
aI59005
aI3185
asS'd'
I9172
sa(dp18344
S'a'
I29321
sS's'
I17
sS'm'
(lp18345
I1
aI3
aI7
aI3
aI23
aI5
aI51
aI141
aI293
aI489
aI263
aI2187
aI1259
aI2729
aI1779
aI61027
aI53931
asS'd'
I9173
sa(dp18346
S'a'
I29322
sS's'
I17
sS'm'
(lp18347
I1
aI1
aI1
aI15
aI27
aI7
aI15
aI109
aI475
aI839
aI175
aI953
aI4531
aI437
aI22475
aI24167
aI19051
asS'd'
I9174
sa(dp18348
S'a'
I29339
sS's'
I17
sS'm'
(lp18349
I1
aI1
aI7
aI9
aI25
aI23
aI41
aI107
aI299
aI115
aI1713
aI1559
aI5701
aI5427
aI28813
aI39913
aI15941
asS'd'
I9175
sa(dp18350
S'a'
I29365
sS's'
I17
sS'm'
(lp18351
I1
aI3
aI5
aI7
aI5
aI25
aI99
aI9
aI307
aI591
aI1303
aI3501
aI1589
aI12095
aI26629
aI52127
aI60635
asS'd'
I9176
sa(dp18352
S'a'
I29369
sS's'
I17
sS'm'
(lp18353
I1
aI3
aI7
aI3
aI9
aI23
aI3
aI29
aI113
aI49
aI1601
aI541
aI1415
aI11601
aI19165
aI46595
aI111623
asS'd'
I9177
sa(dp18354
S'a'
I29372
sS's'
I17
sS'm'
(lp18355
I1
aI1
aI3
aI11
aI5
aI53
aI37
aI153
aI51
aI41
aI1267
aI545
aI2055
aI13137
aI7749
aI30721
aI119591
asS'd'
I9178
sa(dp18356
S'a'
I29377
sS's'
I17
sS'm'
(lp18357
I1
aI3
aI5
aI11
aI1
aI15
aI65
aI17
aI155
aI65
aI745
aI2547
aI6351
aI2347
aI13553
aI5785
aI129857
asS'd'
I9179
sa(dp18358
S'a'
I29392
sS's'
I17
sS'm'
(lp18359
I1
aI3
aI7
aI13
aI5
aI53
aI11
aI59
aI453
aI467
aI1275
aI3669
aI4481
aI5229
aI2953
aI23369
aI100161
asS'd'
I9180
sa(dp18360
S'a'
I29401
sS's'
I17
sS'm'
(lp18361
I1
aI3
aI1
aI1
aI13
aI41
aI91
aI179
aI331
aI547
aI1571
aI1787
aI6467
aI12375
aI4579
aI45023
aI119149
asS'd'
I9181
sa(dp18362
S'a'
I29402
sS's'
I17
sS'm'
(lp18363
I1
aI3
aI3
aI5
aI17
aI55
aI105
aI57
aI227
aI323
aI1517
aI1215
aI3149
aI13919
aI18595
aI5525
aI82445
asS'd'
I9182
sa(dp18364
S'a'
I29418
sS's'
I17
sS'm'
(lp18365
I1
aI3
aI1
aI13
aI19
aI23
aI17
aI239
aI81
aI481
aI1625
aI2003
aI7295
aI2185
aI7021
aI19357
aI37867
asS'd'
I9183
sa(dp18366
S'a'
I29437
sS's'
I17
sS'm'
(lp18367
I1
aI3
aI5
aI9
aI11
aI15
aI61
aI223
aI153
aI139
aI2023
aI2579
aI495
aI14319
aI2835
aI26541
aI113115
asS'd'
I9184
sa(dp18368
S'a'
I29445
sS's'
I17
sS'm'
(lp18369
I1
aI1
aI3
aI13
aI29
aI9
aI13
aI149
aI325
aI87
aI697
aI2345
aI2205
aI5069
aI9939
aI61351
aI127313
asS'd'
I9185
sa(dp18370
S'a'
I29446
sS's'
I17
sS'm'
(lp18371
I1
aI1
aI7
aI11
aI13
aI53
aI45
aI197
aI167
aI551
aI439
aI3715
aI4587
aI8549
aI28193
aI35827
aI96721
asS'd'
I9186
sa(dp18372
S'a'
I29460
sS's'
I17
sS'm'
(lp18373
I1
aI3
aI1
aI1
aI17
aI7
aI31
aI205
aI219
aI739
aI1165
aI243
aI3877
aI15943
aI15207
aI43857
aI120565
asS'd'
I9187
sa(dp18374
S'a'
I29467
sS's'
I17
sS'm'
(lp18375
I1
aI3
aI7
aI9
aI7
aI43
aI81
aI203
aI295
aI553
aI279
aI2717
aI9
aI751
aI24715
aI21591
aI11485
asS'd'
I9188
sa(dp18376
S'a'
I29473
sS's'
I17
sS'm'
(lp18377
I1
aI1
aI1
aI11
aI15
aI17
aI41
aI121
aI355
aI157
aI955
aI3875
aI7595
aI235
aI4937
aI20607
aI11401
asS'd'
I9189
sa(dp18378
S'a'
I29476
sS's'
I17
sS'm'
(lp18379
I1
aI1
aI7
aI7
aI13
aI49
aI33
aI161
aI65
aI251
aI1895
aI2665
aI3017
aI9725
aI10797
aI46313
aI43407
asS'd'
I9190
sa(dp18380
S'a'
I29488
sS's'
I17
sS'm'
(lp18381
I1
aI1
aI3
aI9
aI23
aI17
aI127
aI69
aI385
aI875
aI461
aI3305
aI3001
aI5875
aI13547
aI61239
aI113571
asS'd'
I9191
sa(dp18382
S'a'
I29511
sS's'
I17
sS'm'
(lp18383
I1
aI3
aI7
aI5
aI15
aI47
aI113
aI131
aI465
aI89
aI733
aI433
aI799
aI5689
aI723
aI63479
aI106945
asS'd'
I9192
sa(dp18384
S'a'
I29512
sS's'
I17
sS'm'
(lp18385
I1
aI3
aI3
aI15
aI29
aI1
aI51
aI115
aI317
aI1021
aI1219
aI1797
aI4005
aI10435
aI28935
aI49467
aI66833
asS'd'
I9193
sa(dp18386
S'a'
I29517
sS's'
I17
sS'm'
(lp18387
I1
aI1
aI3
aI11
aI15
aI9
aI51
aI209
aI477
aI479
aI1099
aI2781
aI5525
aI12715
aI9067
aI18317
aI121671
asS'd'
I9194
sa(dp18388
S'a'
I29523
sS's'
I17
sS'm'
(lp18389
I1
aI3
aI1
aI7
aI19
aI35
aI61
aI27
aI479
aI815
aI1639
aI2825
aI7449
aI13697
aI3079
aI49833
aI35119
asS'd'
I9195
sa(dp18390
S'a'
I29526
sS's'
I17
sS'm'
(lp18391
I1
aI3
aI7
aI3
aI17
aI53
aI95
aI155
aI505
aI185
aI717
aI3419
aI3857
aI2369
aI14525
aI22797
aI38553
asS'd'
I9196
sa(dp18392
S'a'
I29535
sS's'
I17
sS'm'
(lp18393
I1
aI3
aI1
aI13
aI27
aI5
aI11
aI21
aI507
aI65
aI39
aI2841
aI7887
aI2783
aI18767
aI34171
aI40527
asS'd'
I9197
sa(dp18394
S'a'
I29539
sS's'
I17
sS'm'
(lp18395
I1
aI3
aI1
aI7
aI9
aI47
aI69
aI217
aI251
aI775
aI631
aI1967
aI5541
aI10679
aI16439
aI33533
aI57817
asS'd'
I9198
sa(dp18396
S'a'
I29545
sS's'
I17
sS'm'
(lp18397
I1
aI1
aI5
aI11
aI27
aI57
aI103
aI255
aI359
aI745
aI63
aI3725
aI4113
aI10943
aI7833
aI46857
aI99239
asS'd'
I9199
sa(dp18398
S'a'
I29559
sS's'
I17
sS'm'
(lp18399
I1
aI3
aI1
aI5
aI31
aI41
aI69
aI245
aI401
aI451
aI959
aI351
aI6999
aI6187
aI21437
aI55067
aI53547
asS'd'
I9200
sa(dp18400
S'a'
I29560
sS's'
I17
sS'm'
(lp18401
I1
aI1
aI1
aI13
aI25
aI49
aI85
aI181
aI457
aI731
aI743
aI1901
aI7013
aI12027
aI14729
aI24193
aI89685
asS'd'
I9201
sa(dp18402
S'a'
I29589
sS's'
I17
sS'm'
(lp18403
I1
aI3
aI3
aI1
aI31
aI29
aI101
aI137
aI117
aI135
aI345
aI1419
aI7133
aI2695
aI3631
aI53049
aI45875
asS'd'
I9202
sa(dp18404
S'a'
I29593
sS's'
I17
sS'm'
(lp18405
I1
aI1
aI1
aI13
aI11
aI51
aI95
aI221
aI339
aI655
aI201
aI3007
aI8179
aI8093
aI22399
aI59123
aI127319
asS'd'
I9203
sa(dp18406
S'a'
I29599
sS's'
I17
sS'm'
(lp18407
I1
aI3
aI7
aI3
aI31
aI37
aI23
aI199
aI191
aI649
aI817
aI1959
aI893
aI2333
aI21477
aI29087
aI115667
asS'd'
I9204
sa(dp18408
S'a'
I29603
sS's'
I17
sS'm'
(lp18409
I1
aI3
aI3
aI5
aI9
aI55
aI123
aI67
aI39
aI533
aI797
aI2575
aI3955
aI14585
aI28587
aI13079
aI60053
asS'd'
I9205
sa(dp18410
S'a'
I29610
sS's'
I17
sS'm'
(lp18411
I1
aI3
aI1
aI1
aI17
aI19
aI15
aI219
aI185
aI21
aI967
aI667
aI3361
aI3883
aI8059
aI26199
aI80913
asS'd'
I9206
sa(dp18412
S'a'
I29629
sS's'
I17
sS'm'
(lp18413
I1
aI3
aI3
aI11
aI23
aI5
aI99
aI57
aI379
aI151
aI271
aI3735
aI7087
aI12731
aI2949
aI54831
aI37835
asS'd'
I9207
sa(dp18414
S'a'
I29644
sS's'
I17
sS'm'
(lp18415
I1
aI3
aI1
aI7
aI21
aI25
aI9
aI195
aI497
aI585
aI901
aI19
aI7675
aI13611
aI31155
aI14567
aI20545
asS'd'
I9208
sa(dp18416
S'a'
I29656
sS's'
I17
sS'm'
(lp18417
I1
aI1
aI3
aI3
aI27
aI45
aI51
aI169
aI397
aI531
aI673
aI2935
aI3779
aI7169
aI23479
aI16157
aI100237
asS'd'
I9209
sa(dp18418
S'a'
I29662
sS's'
I17
sS'm'
(lp18419
I1
aI1
aI1
aI1
aI19
aI49
aI3
aI75
aI455
aI805
aI591
aI1929
aI2883
aI2797
aI31379
aI12025
aI120929
asS'd'
I9210
sa(dp18420
S'a'
I29675
sS's'
I17
sS'm'
(lp18421
I1
aI3
aI5
aI3
aI17
aI39
aI115
aI93
aI341
aI329
aI1857
aI2753
aI4923
aI12539
aI25589
aI19437
aI29027
asS'd'
I9211
sa(dp18422
S'a'
I29692
sS's'
I17
sS'm'
(lp18423
I1
aI3
aI5
aI9
aI27
aI37
aI21
aI235
aI499
aI369
aI1341
aI3719
aI6819
aI3153
aI30619
aI50901
aI52999
asS'd'
I9212
sa(dp18424
S'a'
I29704
sS's'
I17
sS'm'
(lp18425
I1
aI3
aI1
aI11
aI3
aI55
aI43
aI219
aI83
aI771
aI783
aI3569
aI7879
aI2363
aI30605
aI5965
aI126855
asS'd'
I9213
sa(dp18426
S'a'
I29707
sS's'
I17
sS'm'
(lp18427
I1
aI3
aI7
aI13
aI7
aI25
aI111
aI63
aI355
aI317
aI1579
aI1523
aI2733
aI11963
aI25205
aI20545
aI67389
asS'd'
I9214
sa(dp18428
S'a'
I29715
sS's'
I17
sS'm'
(lp18429
I1
aI3
aI5
aI7
aI3
aI17
aI55
aI99
aI321
aI633
aI2013
aI1991
aI1019
aI9223
aI21117
aI23337
aI90589
asS'd'
I9215
sa(dp18430
S'a'
I29717
sS's'
I17
sS'm'
(lp18431
I1
aI3
aI1
aI1
aI17
aI25
aI79
aI171
aI303
aI403
aI2037
aI2595
aI3951
aI8021
aI8669
aI9363
aI20345
asS'd'
I9216
sa(dp18432
S'a'
I29722
sS's'
I17
sS'm'
(lp18433
I1
aI1
aI1
aI7
aI13
aI7
aI11
aI7
aI347
aI53
aI1763
aI3097
aI3353
aI3769
aI22947
aI20919
aI92247
asS'd'
I9217
sa(dp18434
S'a'
I29731
sS's'
I17
sS'm'
(lp18435
I1
aI3
aI5
aI11
aI19
aI29
aI91
aI191
aI511
aI705
aI1317
aI3367
aI7
aI4999
aI30251
aI18299
aI66983
asS'd'
I9218
sa(dp18436
S'a'
I29740
sS's'
I17
sS'm'
(lp18437
I1
aI1
aI3
aI7
aI19
aI17
aI71
aI77
aI285
aI189
aI853
aI2305
aI4205
aI15603
aI15501
aI48073
aI11411
asS'd'
I9219
sa(dp18438
S'a'
I29743
sS's'
I17
sS'm'
(lp18439
I1
aI3
aI7
aI5
aI21
aI15
aI47
aI13
aI277
aI969
aI1861
aI3493
aI6723
aI11521
aI22145
aI43779
aI44713
asS'd'
I9220
sa(dp18440
S'a'
I29752
sS's'
I17
sS'm'
(lp18441
I1
aI1
aI3
aI5
aI19
aI47
aI51
aI207
aI229
aI957
aI709
aI267
aI8081
aI611
aI26403
aI14871
aI111841
asS'd'
I9221
sa(dp18442
S'a'
I29760
sS's'
I17
sS'm'
(lp18443
I1
aI3
aI7
aI1
aI19
aI43
aI71
aI73
aI405
aI351
aI1131
aI3527
aI5949
aI14363
aI20041
aI48123
aI68123
asS'd'
I9222
sa(dp18444
S'a'
I29766
sS's'
I17
sS'm'
(lp18445
I1
aI3
aI3
aI7
aI23
aI51
aI81
aI13
aI161
aI453
aI365
aI1089
aI3505
aI12359
aI14277
aI56113
aI19717
asS'd'
I9223
sa(dp18446
S'a'
I29770
sS's'
I17
sS'm'
(lp18447
I1
aI1
aI1
aI7
aI29
aI35
aI103
aI137
aI317
aI417
aI1465
aI2787
aI6935
aI9885
aI12943
aI43937
aI28353
asS'd'
I9224
sa(dp18448
S'a'
I29794
sS's'
I17
sS'm'
(lp18449
I1
aI1
aI3
aI13
aI19
aI37
aI97
aI5
aI115
aI89
aI1005
aI3033
aI2011
aI2633
aI10615
aI6641
aI73385
asS'd'
I9225
sa(dp18450
S'a'
I29796
sS's'
I17
sS'm'
(lp18451
I1
aI3
aI7
aI7
aI31
aI39
aI107
aI165
aI61
aI1009
aI1159
aI865
aI3469
aI11093
aI10425
aI43959
aI37395
asS'd'
I9226
sa(dp18452
S'a'
I29814
sS's'
I17
sS'm'
(lp18453
I1
aI3
aI5
aI9
aI7
aI51
aI99
aI91
aI37
aI457
aI39
aI2455
aI7481
aI4929
aI29755
aI50603
aI48943
asS'd'
I9227
sa(dp18454
S'a'
I29817
sS's'
I17
sS'm'
(lp18455
I1
aI1
aI5
aI13
aI5
aI39
aI47
aI149
aI341
aI303
aI843
aI3619
aI7527
aI8739
aI5893
aI42623
aI99899
asS'd'
I9228
sa(dp18456
S'a'
I29829
sS's'
I17
sS'm'
(lp18457
I1
aI1
aI1
aI1
aI1
aI41
aI73
aI71
aI409
aI351
aI131
aI515
aI6657
aI337
aI23913
aI583
aI21665
asS'd'
I9229
sa(dp18458
S'a'
I29841
sS's'
I17
sS'm'
(lp18459
I1
aI1
aI3
aI3
aI11
aI45
aI39
aI113
aI315
aI965
aI1605
aI2779
aI501
aI7429
aI2567
aI7011
aI71445
asS'd'
I9230
sa(dp18460
S'a'
I29851
sS's'
I17
sS'm'
(lp18461
I1
aI3
aI7
aI13
aI21
aI13
aI45
aI105
aI385
aI281
aI1683
aI3997
aI6391
aI10943
aI22349
aI37261
aI57555
asS'd'
I9231
sa(dp18462
S'a'
I29860
sS's'
I17
sS'm'
(lp18463
I1
aI1
aI3
aI5
aI17
aI55
aI109
aI71
aI393
aI561
aI433
aI1091
aI1923
aI13861
aI12981
aI5523
aI15467
asS'd'
I9232
sa(dp18464
S'a'
I29864
sS's'
I17
sS'm'
(lp18465
I1
aI3
aI7
aI5
aI17
aI11
aI113
aI119
aI37
aI989
aI1609
aI2191
aI1511
aI11835
aI25423
aI793
aI15227
asS'd'
I9233
sa(dp18466
S'a'
I29869
sS's'
I17
sS'm'
(lp18467
I1
aI3
aI1
aI5
aI5
aI55
aI105
aI225
aI349
aI351
aI1259
aI1309
aI821
aI2733
aI1357
aI3665
aI38863
asS'd'
I9234
sa(dp18468
S'a'
I29870
sS's'
I17
sS'm'
(lp18469
I1
aI3
aI5
aI1
aI23
aI61
aI49
aI113
aI169
aI319
aI85
aI1581
aI97
aI5271
aI30625
aI37693
aI7297
asS'd'
I9235
sa(dp18470
S'a'
I29889
sS's'
I17
sS'm'
(lp18471
I1
aI3
aI5
aI1
aI1
aI25
aI31
aI125
aI307
aI731
aI1815
aI1047
aI7251
aI12481
aI20781
aI63275
aI51985
asS'd'
I9236
sa(dp18472
S'a'
I29896
sS's'
I17
sS'm'
(lp18473
I1
aI3
aI5
aI9
aI11
aI9
aI121
aI111
aI45
aI751
aI793
aI2593
aI6409
aI4355
aI30183
aI36959
aI105161
asS'd'
I9237
sa(dp18474
S'a'
I29901
sS's'
I17
sS'm'
(lp18475
I1
aI1
aI3
aI9
aI25
aI37
aI95
aI253
aI401
aI481
aI1521
aI3555
aI231
aI15459
aI1581
aI36661
aI121727
asS'd'
I9238
sa(dp18476
S'a'
I29913
sS's'
I17
sS'm'
(lp18477
I1
aI3
aI5
aI3
aI27
aI11
aI107
aI115
aI213
aI813
aI27
aI1789
aI603
aI383
aI1129
aI63365
aI51147
asS'd'
I9239
sa(dp18478
S'a'
I29919
sS's'
I17
sS'm'
(lp18479
I1
aI1
aI3
aI13
aI25
aI7
aI33
aI1
aI97
aI907
aI35
aI4069
aI5001
aI507
aI911
aI62037
aI22019
asS'd'
I9240
sa(dp18480
S'a'
I29920
sS's'
I17
sS'm'
(lp18481
I1
aI1
aI3
aI5
aI7
aI55
aI35
aI95
aI261
aI217
aI1565
aI3473
aI3475
aI12181
aI569
aI27389
aI81771
asS'd'
I9241
sa(dp18482
S'a'
I29947
sS's'
I17
sS'm'
(lp18483
I1
aI1
aI3
aI5
aI11
aI33
aI95
aI121
aI453
aI711
aI361
aI3927
aI5231
aI15685
aI31143
aI56915
aI23707
asS'd'
I9242
sa(dp18484
S'a'
I29958
sS's'
I17
sS'm'
(lp18485
I1
aI1
aI3
aI11
aI25
aI15
aI53
aI155
aI469
aI647
aI1547
aI335
aI3753
aI12873
aI13639
aI25129
aI79287
asS'd'
I9243
sa(dp18486
S'a'
I29975
sS's'
I17
sS'm'
(lp18487
I1
aI3
aI3
aI3
aI1
aI21
aI21
aI121
aI105
aI903
aI83
aI2287
aI4295
aI14369
aI29183
aI26841
aI38115
asS'd'
I9244
sa(dp18488
S'a'
I29981
sS's'
I17
sS'm'
(lp18489
I1
aI1
aI7
aI11
aI5
aI29
aI65
aI191
aI389
aI419
aI1315
aI739
aI3485
aI10587
aI2399
aI36377
aI28789
asS'd'
I9245
sa(dp18490
S'a'
I29985
sS's'
I17
sS'm'
(lp18491
I1
aI3
aI3
aI11
aI3
aI29
aI71
aI169
aI265
aI747
aI395
aI1211
aI3487
aI15705
aI25611
aI18183
aI85771
asS'd'
I9246
sa(dp18492
S'a'
I29998
sS's'
I17
sS'm'
(lp18493
I1
aI1
aI7
aI3
aI23
aI5
aI45
aI47
aI337
aI571
aI33
aI1221
aI5671
aI1233
aI28361
aI36945
aI911
asS'd'
I9247
sa(dp18494
S'a'
I30010
sS's'
I17
sS'm'
(lp18495
I1
aI1
aI5
aI11
aI17
aI15
aI57
aI97
aI185
aI999
aI1277
aI3371
aI2785
aI3341
aI13395
aI11925
aI86777
asS'd'
I9248
sa(dp18496
S'a'
I30012
sS's'
I17
sS'm'
(lp18497
I1
aI1
aI3
aI1
aI31
aI37
aI23
aI105
aI503
aI869
aI1309
aI3733
aI4629
aI8263
aI11763
aI30669
aI26179
asS'd'
I9249
sa(dp18498
S'a'
I30044
sS's'
I17
sS'm'
(lp18499
I1
aI1
aI3
aI15
aI25
aI61
aI37
aI27
aI325
aI413
aI809
aI2959
aI8137
aI3397
aI21185
aI27995
aI84297
asS'd'
I9250
sa(dp18500
S'a'
I30048
sS's'
I17
sS'm'
(lp18501
I1
aI3
aI5
aI5
aI1
aI55
aI95
aI41
aI493
aI469
aI331
aI1789
aI7037
aI7947
aI14239
aI16109
aI51795
asS'd'
I9251
sa(dp18502
S'a'
I30051
sS's'
I17
sS'm'
(lp18503
I1
aI1
aI1
aI1
aI19
aI33
aI111
aI237
aI261
aI331
aI871
aI3539
aI1731
aI6953
aI11345
aI37901
aI5623
asS'd'
I9252
sa(dp18504
S'a'
I30063
sS's'
I17
sS'm'
(lp18505
I1
aI1
aI5
aI7
aI21
aI41
aI49
aI179
aI49
aI797
aI231
aI1299
aI145
aI7743
aI725
aI60595
aI19581
asS'd'
I9253
sa(dp18506
S'a'
I30065
sS's'
I17
sS'm'
(lp18507
I1
aI3
aI7
aI15
aI7
aI43
aI67
aI219
aI133
aI641
aI1657
aI2301
aI1591
aI12309
aI6395
aI3999
aI92961
asS'd'
I9254
sa(dp18508
S'a'
I30072
sS's'
I17
sS'm'
(lp18509
I1
aI1
aI1
aI5
aI1
aI49
aI37
aI81
aI219
aI323
aI461
aI1379
aI1797
aI14825
aI21811
aI7347
aI35643
asS'd'
I9255
sa(dp18510
S'a'
I30077
sS's'
I17
sS'm'
(lp18511
I1
aI1
aI1
aI11
aI1
aI3
aI83
aI31
aI307
aI83
aI1169
aI3277
aI1875
aI13397
aI20265
aI46707
aI15205
asS'd'
I9256
sa(dp18512
S'a'
I30087
sS's'
I17
sS'm'
(lp18513
I1
aI3
aI7
aI11
aI29
aI41
aI69
aI33
aI405
aI991
aI1937
aI1217
aI2137
aI8657
aI4319
aI41119
aI43371
asS'd'
I9257
sa(dp18514
S'a'
I30088
sS's'
I17
sS'm'
(lp18515
I1
aI3
aI3
aI3
aI25
aI49
aI107
aI197
aI347
aI923
aI1585
aI3023
aI4087
aI13875
aI22015
aI35733
aI33755
asS'd'
I9258
sa(dp18516
S'a'
I30101
sS's'
I17
sS'm'
(lp18517
I1
aI3
aI3
aI5
aI15
aI61
aI89
aI249
aI141
aI853
aI1469
aI999
aI7425
aI10015
aI12341
aI51535
aI61619
asS'd'
I9259
sa(dp18518
S'a'
I30112
sS's'
I17
sS'm'
(lp18519
I1
aI1
aI7
aI13
aI31
aI61
aI89
aI113
aI117
aI429
aI1011
aI1589
aI1419
aI5083
aI4843
aI26759
aI47401
asS'd'
I9260
sa(dp18520
S'a'
I30121
sS's'
I17
sS'm'
(lp18521
I1
aI1
aI7
aI9
aI17
aI37
aI119
aI39
aI499
aI93
aI1155
aI3069
aI2033
aI12483
aI29849
aI40077
aI11103
asS'd'
I9261
sa(dp18522
S'a'
I30122
sS's'
I17
sS'm'
(lp18523
I1
aI3
aI1
aI15
aI11
aI5
aI23
aI121
aI283
aI717
aI1573
aI3911
aI2031
aI2617
aI20387
aI33157
aI301
asS'd'
I9262
sa(dp18524
S'a'
I30130
sS's'
I17
sS'm'
(lp18525
I1
aI3
aI5
aI7
aI17
aI61
aI109
aI3
aI205
aI617
aI1171
aI223
aI6609
aI15027
aI2629
aI15801
aI73749
asS'd'
I9263
sa(dp18526
S'a'
I30135
sS's'
I17
sS'm'
(lp18527
I1
aI1
aI3
aI7
aI5
aI49
aI9
aI73
aI333
aI401
aI675
aI2765
aI993
aI77
aI19237
aI60929
aI88703
asS'd'
I9264
sa(dp18528
S'a'
I30139
sS's'
I17
sS'm'
(lp18529
I1
aI1
aI1
aI9
aI21
aI25
aI53
aI249
aI241
aI43
aI1959
aI545
aI3729
aI11395
aI3027
aI12661
aI87729
asS'd'
I9265
sa(dp18530
S'a'
I30142
sS's'
I17
sS'm'
(lp18531
I1
aI3
aI3
aI15
aI15
aI61
aI33
aI59
aI155
aI773
aI1043
aI3111
aI6549
aI5397
aI29099
aI57851
aI107671
asS'd'
I9266
sa(dp18532
S'a'
I30164
sS's'
I17
sS'm'
(lp18533
I1
aI3
aI7
aI1
aI29
aI29
aI1
aI161
aI273
aI883
aI1913
aI2389
aI401
aI9425
aI17613
aI50443
aI84601
asS'd'
I9267
sa(dp18534
S'a'
I30167
sS's'
I17
sS'm'
(lp18535
I1
aI3
aI7
aI11
aI15
aI63
aI41
aI53
aI371
aI153
aI1491
aI3013
aI6635
aI4955
aI30145
aI20175
aI16541
asS'd'
I9268
sa(dp18536
S'a'
I30180
sS's'
I17
sS'm'
(lp18537
I1
aI3
aI3
aI11
aI31
aI27
aI127
aI207
aI11
aI313
aI1067
aI3445
aI3075
aI4071
aI28305
aI58911
aI85273
asS'd'
I9269
sa(dp18538
S'a'
I30202
sS's'
I17
sS'm'
(lp18539
I1
aI1
aI7
aI11
aI17
aI47
aI77
aI119
aI209
aI657
aI1181
aI459
aI5821
aI4267
aI5757
aI53703
aI35621
asS'd'
I9270
sa(dp18540
S'a'
I30211
sS's'
I17
sS'm'
(lp18541
I1
aI3
aI1
aI15
aI27
aI41
aI3
aI217
aI457
aI531
aI1749
aI2847
aI4715
aI11451
aI25071
aI53999
aI93503
asS'd'
I9271
sa(dp18542
S'a'
I30214
sS's'
I17
sS'm'
(lp18543
I1
aI3
aI5
aI3
aI19
aI29
aI9
aI177
aI355
aI265
aI883
aI1113
aI2397
aI1819
aI20757
aI50389
aI95551
asS'd'
I9272
sa(dp18544
S'a'
I30217
sS's'
I17
sS'm'
(lp18545
I1
aI1
aI3
aI15
aI3
aI45
aI85
aI211
aI377
aI293
aI1791
aI1193
aI1117
aI9383
aI28039
aI27155
aI129513
asS'd'
I9273
sa(dp18546
S'a'
I30223
sS's'
I17
sS'm'
(lp18547
I1
aI1
aI7
aI5
aI1
aI17
aI59
aI215
aI161
aI933
aI1653
aI2407
aI2693
aI3655
aI7515
aI2239
aI88985
asS'd'
I9274
sa(dp18548
S'a'
I30228
sS's'
I17
sS'm'
(lp18549
I1
aI1
aI3
aI11
aI31
aI3
aI1
aI77
aI459
aI817
aI651
aI603
aI1711
aI9391
aI20607
aI48195
aI127153
asS'd'
I9275
sa(dp18550
S'a'
I30235
sS's'
I17
sS'm'
(lp18551
I1
aI1
aI5
aI13
aI11
aI49
aI25
aI13
aI51
aI443
aI1877
aI1257
aI163
aI4673
aI30313
aI18841
aI24547
asS'd'
I9276
sa(dp18552
S'a'
I30256
sS's'
I17
sS'm'
(lp18553
I1
aI1
aI7
aI7
aI15
aI33
aI43
aI79
aI127
aI625
aI1991
aI1311
aI2095
aI14659
aI3477
aI56023
aI57955
asS'd'
I9277
sa(dp18554
S'a'
I30265
sS's'
I17
sS'm'
(lp18555
I1
aI3
aI7
aI15
aI29
aI7
aI119
aI183
aI123
aI323
aI1723
aI959
aI2733
aI2097
aI2927
aI57595
aI86067
asS'd'
I9278
sa(dp18556
S'a'
I30286
sS's'
I17
sS'm'
(lp18557
I1
aI3
aI5
aI5
aI29
aI57
aI93
aI139
aI495
aI739
aI1715
aI713
aI243
aI2027
aI11223
aI44143
aI119155
asS'd'
I9279
sa(dp18558
S'a'
I30293
sS's'
I17
sS'm'
(lp18559
I1
aI3
aI1
aI9
aI3
aI63
aI113
aI29
aI19
aI439
aI869
aI1101
aI4947
aI2191
aI14737
aI57049
aI93505
asS'd'
I9280
sa(dp18560
S'a'
I30298
sS's'
I17
sS'm'
(lp18561
I1
aI1
aI7
aI1
aI11
aI39
aI27
aI29
aI281
aI829
aI1979
aI2185
aI2207
aI14969
aI7447
aI55541
aI59593
asS'd'
I9281
sa(dp18562
S'a'
I30310
sS's'
I17
sS'm'
(lp18563
I1
aI1
aI7
aI13
aI11
aI15
aI15
aI143
aI383
aI469
aI1439
aI2823
aI7489
aI7675
aI15433
aI26203
aI80433
asS'd'
I9282
sa(dp18564
S'a'
I30314
sS's'
I17
sS'm'
(lp18565
I1
aI1
aI1
aI7
aI15
aI45
aI23
aI93
aI477
aI1
aI1431
aI3173
aI7879
aI12211
aI13051
aI56971
aI114289
asS'd'
I9283
sa(dp18566
S'a'
I30327
sS's'
I17
sS'm'
(lp18567
I1
aI3
aI1
aI1
aI27
aI55
aI61
aI185
aI323
aI569
aI1063
aI1357
aI7373
aI14947
aI15967
aI64517
aI104625
asS'd'
I9284
sa(dp18568
S'a'
I30333
sS's'
I17
sS'm'
(lp18569
I1
aI1
aI5
aI11
aI25
aI1
aI127
aI163
aI419
aI657
aI911
aI361
aI3675
aI10765
aI24565
aI2661
aI61979
asS'd'
I9285
sa(dp18570
S'a'
I30334
sS's'
I17
sS'm'
(lp18571
I1
aI3
aI7
aI15
aI29
aI53
aI29
aI149
aI465
aI535
aI1865
aI2243
aI4783
aI9327
aI24843
aI52313
aI15683
asS'd'
I9286
sa(dp18572
S'a'
I30337
sS's'
I17
sS'm'
(lp18573
I1
aI1
aI7
aI7
aI17
aI7
aI85
aI187
aI91
aI1013
aI1917
aI2959
aI3571
aI12047
aI25267
aI34095
aI9877
asS'd'
I9287
sa(dp18574
S'a'
I30338
sS's'
I17
sS'm'
(lp18575
I1
aI1
aI1
aI7
aI5
aI27
aI111
aI107
aI313
aI571
aI1081
aI3193
aI1025
aI2589
aI1523
aI40453
aI77065
asS'd'
I9288
sa(dp18576
S'a'
I30344
sS's'
I17
sS'm'
(lp18577
I1
aI3
aI7
aI1
aI19
aI27
aI1
aI103
aI479
aI405
aI583
aI1737
aI3495
aI9093
aI20397
aI16429
aI45609
asS'd'
I9289
sa(dp18578
S'a'
I30349
sS's'
I17
sS'm'
(lp18579
I1
aI1
aI3
aI11
aI17
aI1
aI125
aI97
aI261
aI651
aI901
aI1245
aI1181
aI14469
aI16229
aI31935
aI100227
asS'd'
I9290
sa(dp18580
S'a'
I30352
sS's'
I17
sS'm'
(lp18581
I1
aI1
aI7
aI11
aI15
aI1
aI19
aI151
aI453
aI833
aI1371
aI1109
aI5373
aI25
aI5619
aI58351
aI26349
asS'd'
I9291
sa(dp18582
S'a'
I30362
sS's'
I17
sS'm'
(lp18583
I1
aI1
aI7
aI15
aI17
aI55
aI51
aI67
aI123
aI13
aI1873
aI1035
aI5871
aI11943
aI11543
aI43261
aI62587
asS'd'
I9292
sa(dp18584
S'a'
I30374
sS's'
I17
sS'm'
(lp18585
I1
aI3
aI1
aI13
aI21
aI15
aI83
aI205
aI333
aI379
aI2021
aI1389
aI861
aI10395
aI20587
aI38207
aI49109
asS'd'
I9293
sa(dp18586
S'a'
I30380
sS's'
I17
sS'm'
(lp18587
I1
aI1
aI5
aI3
aI13
aI49
aI89
aI85
aI463
aI1005
aI1367
aI3487
aI581
aI12145
aI22445
aI35343
aI65745
asS'd'
I9294
sa(dp18588
S'a'
I30388
sS's'
I17
sS'm'
(lp18589
I1
aI3
aI3
aI3
aI29
aI27
aI99
aI195
aI89
aI793
aI1677
aI3989
aI4811
aI8303
aI9165
aI50349
aI96947
asS'd'
I9295
sa(dp18590
S'a'
I30391
sS's'
I17
sS'm'
(lp18591
I1
aI1
aI3
aI5
aI29
aI11
aI91
aI107
aI13
aI659
aI213
aI1613
aI2245
aI11567
aI28157
aI2937
aI53275
asS'd'
I9296
sa(dp18592
S'a'
I30395
sS's'
I17
sS'm'
(lp18593
I1
aI3
aI7
aI5
aI3
aI41
aI65
aI27
aI93
aI747
aI1143
aI505
aI3881
aI2123
aI2903
aI54137
aI96185
asS'd'
I9297
sa(dp18594
S'a'
I30403
sS's'
I17
sS'm'
(lp18595
I1
aI1
aI3
aI15
aI9
aI49
aI3
aI25
aI77
aI681
aI1709
aI915
aI2243
aI2127
aI18243
aI13915
aI67399
asS'd'
I9298
sa(dp18596
S'a'
I30409
sS's'
I17
sS'm'
(lp18597
I1
aI3
aI5
aI7
aI13
aI49
aI89
aI67
aI63
aI271
aI1651
aI897
aI4035
aI1067
aI13743
aI56791
aI44371
asS'd'
I9299
sa(dp18598
S'a'
I30424
sS's'
I17
sS'm'
(lp18599
I1
aI1
aI7
aI9
aI25
aI19
aI125
aI15
aI125
aI705
aI1359
aI817
aI1241
aI12447
aI19097
aI33209
aI89091
asS'd'
I9300
sa(dp18600
S'a'
I30427
sS's'
I17
sS'm'
(lp18601
I1
aI3
aI7
aI3
aI19
aI43
aI127
aI197
aI39
aI709
aI257
aI3547
aI3069
aI1187
aI21255
aI6453
aI40763
asS'd'
I9301
sa(dp18602
S'a'
I30430
sS's'
I17
sS'm'
(lp18603
I1
aI3
aI3
aI5
aI3
aI53
aI37
aI65
aI415
aI183
aI991
aI533
aI7805
aI9905
aI18925
aI52665
aI100987
asS'd'
I9302
sa(dp18604
S'a'
I30451
sS's'
I17
sS'm'
(lp18605
I1
aI3
aI3
aI11
aI17
aI41
aI17
aI137
aI143
aI277
aI945
aI1531
aI7427
aI7287
aI30869
aI27651
aI116507
asS'd'
I9303
sa(dp18606
S'a'
I30460
sS's'
I17
sS'm'
(lp18607
I1
aI1
aI1
aI3
aI13
aI3
aI9
aI163
aI113
aI373
aI1909
aI1051
aI97
aI10729
aI28615
aI40081
aI129297
asS'd'
I9304
sa(dp18608
S'a'
I30475
sS's'
I17
sS'm'
(lp18609
I1
aI3
aI3
aI9
aI11
aI47
aI113
aI27
aI307
aI339
aI1319
aI3083
aI7383
aI1551
aI26691
aI28769
aI114313
asS'd'
I9305
sa(dp18610
S'a'
I30480
sS's'
I17
sS'm'
(lp18611
I1
aI3
aI7
aI7
aI25
aI49
aI31
aI231
aI485
aI629
aI59
aI283
aI7463
aI6603
aI23055
aI63643
aI10121
asS'd'
I9306
sa(dp18612
S'a'
I30485
sS's'
I17
sS'm'
(lp18613
I1
aI3
aI7
aI9
aI5
aI55
aI53
aI127
aI147
aI103
aI1697
aI485
aI7051
aI14153
aI21631
aI35561
aI10393
asS'd'
I9307
sa(dp18614
S'a'
I30486
sS's'
I17
sS'm'
(lp18615
I1
aI3
aI3
aI7
aI23
aI7
aI83
aI17
aI135
aI487
aI315
aI719
aI7003
aI3919
aI13255
aI24031
aI110493
asS'd'
I9308
sa(dp18616
S'a'
I30489
sS's'
I17
sS'm'
(lp18617
I1
aI3
aI1
aI15
aI27
aI55
aI121
aI177
aI205
aI733
aI933
aI1535
aI2925
aI4259
aI22203
aI59059
aI89209
asS'd'
I9309
sa(dp18618
S'a'
I30492
sS's'
I17
sS'm'
(lp18619
I1
aI1
aI1
aI11
aI9
aI11
aI127
aI47
aI493
aI349
aI1415
aI3089
aI4739
aI14347
aI31579
aI20739
aI72997
asS'd'
I9310
sa(dp18620
S'a'
I30499
sS's'
I17
sS'm'
(lp18621
I1
aI3
aI1
aI7
aI9
aI31
aI121
aI111
aI163
aI733
aI1699
aI1507
aI5467
aI13499
aI25269
aI6303
aI70201
asS'd'
I9311
sa(dp18622
S'a'
I30511
sS's'
I17
sS'm'
(lp18623
I1
aI3
aI1
aI5
aI7
aI23
aI123
aI203
aI329
aI387
aI577
aI2331
aI2283
aI14029
aI19409
aI103
aI2477
asS'd'
I9312
sa(dp18624
S'a'
I30533
sS's'
I17
sS'm'
(lp18625
I1
aI1
aI7
aI15
aI11
aI29
aI29
aI153
aI289
aI333
aI1669
aI2065
aI5465
aI8835
aI28753
aI21209
aI34283
asS'd'
I9313
sa(dp18626
S'a'
I30534
sS's'
I17
sS'm'
(lp18627
I1
aI3
aI5
aI11
aI15
aI33
aI45
aI81
aI241
aI461
aI1167
aI1073
aI5511
aI795
aI30955
aI49121
aI42805
asS'd'
I9314
sa(dp18628
S'a'
I30540
sS's'
I17
sS'm'
(lp18629
I1
aI1
aI1
aI13
aI7
aI33
aI11
aI201
aI161
aI475
aI1359
aI2329
aI177
aI9883
aI8967
aI21399
aI73045
asS'd'
I9315
sa(dp18630
S'a'
I30545
sS's'
I17
sS'm'
(lp18631
I1
aI3
aI3
aI1
aI25
aI59
aI85
aI51
aI481
aI751
aI1213
aI3019
aI421
aI9903
aI30071
aI50661
aI94715
asS'd'
I9316
sa(dp18632
S'a'
I30552
sS's'
I17
sS'm'
(lp18633
I1
aI3
aI3
aI1
aI5
aI61
aI3
aI179
aI131
aI233
aI1627
aI3577
aI6323
aI14161
aI21595
aI44963
aI20215
asS'd'
I9317
sa(dp18634
S'a'
I30557
sS's'
I17
sS'm'
(lp18635
I1
aI1
aI1
aI5
aI9
aI53
aI45
aI105
aI275
aI769
aI105
aI2757
aI6769
aI14987
aI19955
aI18291
aI81707
asS'd'
I9318
sa(dp18636
S'a'
I30558
sS's'
I17
sS'm'
(lp18637
I1
aI1
aI7
aI1
aI1
aI59
aI33
aI19
aI385
aI775
aI423
aI1799
aI1325
aI13545
aI16027
aI58347
aI102397
asS'd'
I9319
sa(dp18638
S'a'
I30567
sS's'
I17
sS'm'
(lp18639
I1
aI1
aI3
aI11
aI15
aI61
aI63
aI59
aI355
aI659
aI1483
aI3781
aI7383
aI5563
aI32537
aI34175
aI40303
asS'd'
I9320
sa(dp18640
S'a'
I30568
sS's'
I17
sS'm'
(lp18641
I1
aI3
aI5
aI7
aI5
aI37
aI19
aI223
aI323
aI129
aI287
aI2655
aI3767
aI16201
aI4147
aI315
aI54885
asS'd'
I9321
sa(dp18642
S'a'
I30581
sS's'
I17
sS'm'
(lp18643
I1
aI1
aI7
aI13
aI13
aI23
aI43
aI129
aI405
aI205
aI1691
aI2189
aI3313
aI11789
aI10263
aI16367
aI65547
asS'd'
I9322
sa(dp18644
S'a'
I30597
sS's'
I17
sS'm'
(lp18645
I1
aI3
aI5
aI1
aI15
aI21
aI85
aI233
aI427
aI71
aI475
aI3731
aI3335
aI11483
aI28613
aI4335
aI35669
asS'd'
I9323
sa(dp18646
S'a'
I30607
sS's'
I17
sS'm'
(lp18647
I1
aI1
aI3
aI3
aI31
aI47
aI27
aI109
aI373
aI451
aI1459
aI3103
aI1941
aI10405
aI20233
aI30517
aI122787
asS'd'
I9324
sa(dp18648
S'a'
I30621
sS's'
I17
sS'm'
(lp18649
I1
aI3
aI1
aI7
aI3
aI11
aI113
aI49
aI355
aI465
aI1959
aI1355
aI6521
aI10863
aI11481
aI13385
aI31787
asS'd'
I9325
sa(dp18650
S'a'
I30632
sS's'
I17
sS'm'
(lp18651
I1
aI3
aI1
aI1
aI9
aI45
aI125
aI69
aI267
aI413
aI717
aI2767
aI833
aI317
aI23019
aI37753
aI3081
asS'd'
I9326
sa(dp18652
S'a'
I30650
sS's'
I17
sS'm'
(lp18653
I1
aI1
aI7
aI7
aI13
aI55
aI75
aI105
aI71
aI505
aI239
aI3739
aI4873
aI4257
aI18841
aI41711
aI24045
asS'd'
I9327
sa(dp18654
S'a'
I30655
sS's'
I17
sS'm'
(lp18655
I1
aI1
aI5
aI13
aI21
aI59
aI107
aI229
aI421
aI441
aI1079
aI3727
aI7941
aI8443
aI30433
aI56419
aI105751
asS'd'
I9328
sa(dp18656
S'a'
I30657
sS's'
I17
sS'm'
(lp18657
I1
aI1
aI1
aI9
aI15
aI59
aI29
aI49
aI117
aI1009
aI1971
aI115
aI2899
aI1069
aI29145
aI11855
aI35277
asS'd'
I9329
sa(dp18658
S'a'
I30660
sS's'
I17
sS'm'
(lp18659
I1
aI3
aI7
aI15
aI19
aI55
aI111
aI77
aI169
aI537
aI1695
aI2687
aI3871
aI14017
aI15119
aI27313
aI112807
asS'd'
I9330
sa(dp18660
S'a'
I30669
sS's'
I17
sS'm'
(lp18661
I1
aI1
aI3
aI7
aI29
aI5
aI41
aI201
aI211
aI127
aI1877
aI643
aI2441
aI8743
aI29393
aI6077
aI52597
asS'd'
I9331
sa(dp18662
S'a'
I30675
sS's'
I17
sS'm'
(lp18663
I1
aI3
aI1
aI11
aI7
aI1
aI95
aI15
aI229
aI339
aI475
aI3463
aI7827
aI9943
aI30903
aI65013
aI1145
asS'd'
I9332
sa(dp18664
S'a'
I30678
sS's'
I17
sS'm'
(lp18665
I1
aI1
aI1
aI5
aI23
aI19
aI81
aI23
aI475
aI169
aI373
aI1147
aI1805
aI12779
aI13173
aI8945
aI28175
asS'd'
I9333
sa(dp18666
S'a'
I30682
sS's'
I17
sS'm'
(lp18667
I1
aI3
aI5
aI9
aI3
aI53
aI127
aI33
aI237
aI803
aI121
aI307
aI4981
aI8765
aI12761
aI23601
aI92921
asS'd'
I9334
sa(dp18668
S'a'
I30687
sS's'
I17
sS'm'
(lp18669
I1
aI1
aI3
aI7
aI17
aI63
aI11
aI37
aI213
aI619
aI1095
aI1693
aI6747
aI7373
aI17355
aI5987
aI97923
asS'd'
I9335
sa(dp18670
S'a'
I30688
sS's'
I17
sS'm'
(lp18671
I1
aI1
aI3
aI15
aI11
aI37
aI109
aI175
aI503
aI339
aI591
aI2745
aI2387
aI7419
aI13915
aI4769
aI48229
asS'd'
I9336
sa(dp18672
S'a'
I30698
sS's'
I17
sS'm'
(lp18673
I1
aI3
aI5
aI7
aI19
aI17
aI5
aI81
aI471
aI989
aI249
aI437
aI7309
aI5747
aI25277
aI31911
aI87907
asS'd'
I9337
sa(dp18674
S'a'
I30711
sS's'
I17
sS'm'
(lp18675
I1
aI3
aI1
aI7
aI15
aI25
aI49
aI243
aI423
aI911
aI1957
aI911
aI6331
aI9831
aI26021
aI58877
aI99257
asS'd'
I9338
sa(dp18676
S'a'
I30728
sS's'
I17
sS'm'
(lp18677
I1
aI3
aI5
aI11
aI3
aI55
aI39
aI129
aI271
aI145
aI1973
aI3391
aI2905
aI9229
aI7989
aI15641
aI67933
asS'd'
I9339
sa(dp18678
S'a'
I30733
sS's'
I17
sS'm'
(lp18679
I1
aI3
aI5
aI9
aI13
aI13
aI43
aI135
aI183
aI319
aI1391
aI2953
aI2207
aI14205
aI31203
aI6329
aI98907
asS'd'
I9340
sa(dp18680
S'a'
I30741
sS's'
I17
sS'm'
(lp18681
I1
aI3
aI1
aI9
aI27
aI41
aI27
aI155
aI11
aI191
aI1747
aI975
aI7043
aI13139
aI30387
aI47099
aI120321
asS'd'
I9341
sa(dp18682
S'a'
I30746
sS's'
I17
sS'm'
(lp18683
I1
aI1
aI5
aI9
aI25
aI27
aI53
aI235
aI437
aI77
aI371
aI2413
aI4867
aI14245
aI27199
aI37387
aI88217
asS'd'
I9342
sa(dp18684
S'a'
I30752
sS's'
I17
sS'm'
(lp18685
I1
aI1
aI7
aI7
aI11
aI59
aI103
aI15
aI109
aI65
aI1987
aI3695
aI7737
aI7341
aI26963
aI16075
aI6301
asS'd'
I9343
sa(dp18686
S'a'
I30764
sS's'
I17
sS'm'
(lp18687
I1
aI3
aI7
aI5
aI7
aI59
aI109
aI159
aI121
aI377
aI1851
aI3983
aI5421
aI7633
aI7321
aI14869
aI54633
asS'd'
I9344
sa(dp18688
S'a'
I30769
sS's'
I17
sS'm'
(lp18689
I1
aI1
aI3
aI15
aI21
aI51
aI35
aI243
aI397
aI411
aI1107
aI3689
aI7913
aI14715
aI26349
aI23361
aI90665
asS'd'
I9345
sa(dp18690
S'a'
I30784
sS's'
I17
sS'm'
(lp18691
I1
aI1
aI1
aI3
aI5
aI11
aI77
aI205
aI187
aI981
aI1969
aI1749
aI6295
aI8267
aI16073
aI54451
aI103603
asS'd'
I9346
sa(dp18692
S'a'
I30796
sS's'
I17
sS'm'
(lp18693
I1
aI3
aI3
aI9
aI11
aI13
aI113
aI83
aI243
aI1021
aI2003
aI2277
aI6457
aI10535
aI13461
aI52741
aI9385
asS'd'
I9347
sa(dp18694
S'a'
I30799
sS's'
I17
sS'm'
(lp18695
I1
aI3
aI3
aI11
aI19
aI9
aI103
aI13
aI219
aI269
aI1805
aI2689
aI5219
aI11497
aI4909
aI57985
aI40141
asS'd'
I9348
sa(dp18696
S'a'
I30804
sS's'
I17
sS'm'
(lp18697
I1
aI1
aI1
aI1
aI29
aI25
aI15
aI217
aI69
aI567
aI839
aI1515
aI3627
aI9837
aI21433
aI37177
aI10487
asS'd'
I9349
sa(dp18698
S'a'
I30811
sS's'
I17
sS'm'
(lp18699
I1
aI1
aI7
aI15
aI15
aI23
aI119
aI217
aI277
aI447
aI551
aI825
aI7571
aI3083
aI16573
aI1189
aI64959
asS'd'
I9350
sa(dp18700
S'a'
I30814
sS's'
I17
sS'm'
(lp18701
I1
aI1
aI3
aI11
aI9
aI13
aI63
aI77
aI313
aI195
aI941
aI1621
aI165
aI8905
aI20265
aI53761
aI25091
asS'd'
I9351
sa(dp18702
S'a'
I30818
sS's'
I17
sS'm'
(lp18703
I1
aI3
aI3
aI9
aI17
aI5
aI9
aI183
aI175
aI1015
aI253
aI233
aI2883
aI15587
aI27175
aI38517
aI22707
asS'd'
I9352
sa(dp18704
S'a'
I30827
sS's'
I17
sS'm'
(lp18705
I1
aI3
aI3
aI11
aI23
aI63
aI83
aI17
aI49
aI671
aI749
aI3249
aI7821
aI7189
aI26735
aI28995
aI101737
asS'd'
I9353
sa(dp18706
S'a'
I30838
sS's'
I17
sS'm'
(lp18707
I1
aI1
aI7
aI5
aI25
aI15
aI97
aI247
aI161
aI585
aI1307
aI3803
aI1105
aI9093
aI23523
aI1383
aI65671
asS'd'
I9354
sa(dp18708
S'a'
I30842
sS's'
I17
sS'm'
(lp18709
I1
aI1
aI3
aI15
aI29
aI51
aI65
aI237
aI349
aI709
aI799
aI1425
aI6267
aI6283
aI4773
aI18123
aI74833
asS'd'
I9355
sa(dp18710
S'a'
I30854
sS's'
I17
sS'm'
(lp18711
I1
aI1
aI5
aI5
aI11
aI13
aI9
aI251
aI373
aI189
aI467
aI945
aI7279
aI11349
aI10917
aI6581
aI83967
asS'd'
I9356
sa(dp18712
S'a'
I30863
sS's'
I17
sS'm'
(lp18713
I1
aI1
aI7
aI15
aI23
aI27
aI1
aI197
aI41
aI325
aI757
aI1229
aI6295
aI345
aI26147
aI40135
aI123063
asS'd'
I9357
sa(dp18714
S'a'
I30865
sS's'
I17
sS'm'
(lp18715
I1
aI1
aI7
aI9
aI23
aI9
aI93
aI225
aI191
aI837
aI103
aI3367
aI5411
aI8289
aI7057
aI55391
aI10669
asS'd'
I9358
sa(dp18716
S'a'
I30877
sS's'
I17
sS'm'
(lp18717
I1
aI1
aI5
aI15
aI21
aI17
aI49
aI221
aI487
aI23
aI1943
aI1563
aI6157
aI4035
aI27769
aI14933
aI56913
asS'd'
I9359
sa(dp18718
S'a'
I30881
sS's'
I17
sS'm'
(lp18719
I1
aI1
aI5
aI13
aI13
aI43
aI67
aI245
aI457
aI365
aI1115
aI2205
aI6229
aI4173
aI25167
aI56333
aI55605
asS'd'
I9360
sa(dp18720
S'a'
I30887
sS's'
I17
sS'm'
(lp18721
I1
aI1
aI5
aI11
aI15
aI59
aI109
aI83
aI17
aI913
aI497
aI1299
aI5221
aI321
aI32139
aI13717
aI94311
asS'd'
I9361
sa(dp18722
S'a'
I30908
sS's'
I17
sS'm'
(lp18723
I1
aI3
aI3
aI3
aI31
aI11
aI5
aI203
aI3
aI843
aI2039
aI25
aI6211
aI14927
aI6015
aI46269
aI90369
asS'd'
I9362
sa(dp18724
S'a'
I30916
sS's'
I17
sS'm'
(lp18725
I1
aI1
aI3
aI9
aI21
aI51
aI91
aI203
aI149
aI147
aI197
aI1771
aI2921
aI6609
aI2343
aI35249
aI12963
asS'd'
I9363
sa(dp18726
S'a'
I30919
sS's'
I17
sS'm'
(lp18727
I1
aI3
aI1
aI7
aI17
aI43
aI91
aI229
aI107
aI521
aI737
aI2355
aI5855
aI6707
aI21217
aI47041
aI81833
asS'd'
I9364
sa(dp18728
S'a'
I30925
sS's'
I17
sS'm'
(lp18729
I1
aI3
aI3
aI7
aI7
aI31
aI97
aI135
aI503
aI665
aI1799
aI2937
aI6867
aI4125
aI7375
aI34401
aI18111
asS'd'
I9365
sa(dp18730
S'a'
I30928
sS's'
I17
sS'm'
(lp18731
I1
aI1
aI7
aI1
aI11
aI29
aI89
aI185
aI495
aI633
aI507
aI3727
aI5999
aI5871
aI5911
aI24877
aI10259
asS'd'
I9366
sa(dp18732
S'a'
I30944
sS's'
I17
sS'm'
(lp18733
I1
aI1
aI1
aI13
aI25
aI3
aI25
aI65
aI91
aI411
aI147
aI3699
aI7003
aI3017
aI25635
aI56619
aI101491
asS'd'
I9367
sa(dp18734
S'a'
I30947
sS's'
I17
sS'm'
(lp18735
I1
aI3
aI5
aI7
aI31
aI1
aI63
aI255
aI115
aI179
aI87
aI735
aI1649
aI4767
aI31093
aI60149
aI49829
asS'd'
I9368
sa(dp18736
S'a'
I30950
sS's'
I17
sS'm'
(lp18737
I1
aI1
aI3
aI1
aI21
aI63
aI67
aI85
aI399
aI279
aI1963
aI1759
aI4679
aI15423
aI11533
aI54387
aI36563
asS'd'
I9369
sa(dp18738
S'a'
I30968
sS's'
I17
sS'm'
(lp18739
I1
aI3
aI5
aI3
aI31
aI53
aI73
aI73
aI481
aI443
aI1393
aI2763
aI1199
aI5375
aI8977
aI5369
aI114603
asS'd'
I9370
sa(dp18740
S'a'
I30971
sS's'
I17
sS'm'
(lp18741
I1
aI1
aI1
aI3
aI29
aI47
aI73
aI205
aI469
aI187
aI815
aI2787
aI1431
aI4705
aI11455
aI53643
aI89269
asS'd'
I9371
sa(dp18742
S'a'
I30973
sS's'
I17
sS'm'
(lp18743
I1
aI3
aI3
aI9
aI27
aI57
aI93
aI55
aI287
aI539
aI1259
aI3279
aI1563
aI11399
aI22975
aI27077
aI41031
asS'd'
I9372
sa(dp18744
S'a'
I30976
sS's'
I17
sS'm'
(lp18745
I1
aI1
aI3
aI15
aI7
aI27
aI67
aI25
aI169
aI263
aI621
aI1921
aI509
aI11715
aI15363
aI27447
aI75535
asS'd'
I9373
sa(dp18746
S'a'
I30986
sS's'
I17
sS'm'
(lp18747
I1
aI1
aI1
aI9
aI29
aI63
aI31
aI99
aI321
aI361
aI1693
aI763
aI5593
aI10815
aI741
aI31257
aI70843
asS'd'
I9374
sa(dp18748
S'a'
I30993
sS's'
I17
sS'm'
(lp18749
I1
aI1
aI1
aI9
aI1
aI17
aI73
aI141
aI401
aI549
aI415
aI1289
aI1697
aI1903
aI8919
aI59563
aI60017
asS'd'
I9375
sa(dp18750
S'a'
I30994
sS's'
I17
sS'm'
(lp18751
I1
aI3
aI7
aI3
aI5
aI51
aI127
aI221
aI9
aI929
aI153
aI1045
aI6587
aI13653
aI5343
aI14043
aI116125
asS'd'
I9376
sa(dp18752
S'a'
I31012
sS's'
I17
sS'm'
(lp18753
I1
aI3
aI3
aI13
aI13
aI17
aI29
aI93
aI465
aI59
aI1207
aI3121
aI6331
aI8647
aI5047
aI41869
aI51969
asS'd'
I9377
sa(dp18754
S'a'
I31016
sS's'
I17
sS'm'
(lp18755
I1
aI1
aI1
aI15
aI23
aI29
aI47
aI149
aI119
aI855
aI367
aI1419
aI7739
aI1141
aI19787
aI38185
aI84403
asS'd'
I9378
sa(dp18756
S'a'
I31029
sS's'
I17
sS'm'
(lp18757
I1
aI3
aI1
aI9
aI29
aI63
aI5
aI63
aI435
aI401
aI1023
aI1981
aI6819
aI7547
aI30065
aI33833
aI7471
asS'd'
I9379
sa(dp18758
S'a'
I31039
sS's'
I17
sS'm'
(lp18759
I1
aI3
aI1
aI15
aI1
aI47
aI35
aI161
aI243
aI225
aI935
aI2179
aI7737
aI7841
aI28523
aI11505
aI103741
asS'd'
I9380
sa(dp18760
S'a'
I31041
sS's'
I17
sS'm'
(lp18761
I1
aI1
aI7
aI3
aI7
aI57
aI73
aI55
aI101
aI905
aI1705
aI4057
aI3781
aI8213
aI18997
aI17185
aI33265
asS'd'
I9381
sa(dp18762
S'a'
I31042
sS's'
I17
sS'm'
(lp18763
I1
aI1
aI5
aI1
aI7
aI57
aI31
aI77
aI323
aI395
aI927
aI1969
aI6973
aI9013
aI30789
aI757
aI84075
asS'd'
I9382
sa(dp18764
S'a'
I31044
sS's'
I17
sS'm'
(lp18765
I1
aI1
aI3
aI7
aI15
aI53
aI51
aI205
aI401
aI679
aI1295
aI1955
aI7739
aI11423
aI23207
aI55449
aI60419
asS'd'
I9383
sa(dp18766
S'a'
I31053
sS's'
I17
sS'm'
(lp18767
I1
aI3
aI5
aI11
aI23
aI21
aI67
aI141
aI157
aI767
aI219
aI3607
aI1847
aI11051
aI31499
aI8461
aI106353
asS'd'
I9384
sa(dp18768
S'a'
I31059
sS's'
I17
sS'm'
(lp18769
I1
aI1
aI3
aI9
aI17
aI19
aI123
aI169
aI1
aI31
aI1019
aI1823
aI6043
aI1895
aI17293
aI62079
aI16945
asS'd'
I9385
sa(dp18770
S'a'
I31062
sS's'
I17
sS'm'
(lp18771
I1
aI3
aI5
aI9
aI3
aI15
aI49
aI27
aI183
aI293
aI989
aI2161
aI1845
aI1103
aI20149
aI11121
aI31935
asS'd'
I9386
sa(dp18772
S'a'
I31077
sS's'
I17
sS'm'
(lp18773
I1
aI3
aI1
aI3
aI17
aI39
aI103
aI45
aI491
aI91
aI413
aI487
aI1381
aI5457
aI22503
aI40477
aI94297
asS'd'
I9387
sa(dp18774
S'a'
I31095
sS's'
I17
sS'm'
(lp18775
I1
aI1
aI3
aI7
aI29
aI11
aI87
aI79
aI349
aI437
aI945
aI3753
aI6691
aI4373
aI24875
aI54397
aI33697
asS'd'
I9388
sa(dp18776
S'a'
I31101
sS's'
I17
sS'm'
(lp18777
I1
aI1
aI7
aI1
aI9
aI31
aI105
aI121
aI97
aI947
aI129
aI1909
aI2371
aI5493
aI29523
aI52685
aI24325
asS'd'
I9389
sa(dp18778
S'a'
I31105
sS's'
I17
sS'm'
(lp18779
I1
aI1
aI5
aI9
aI19
aI21
aI63
aI115
aI511
aI525
aI49
aI1879
aI6075
aI8181
aI10135
aI56785
aI53309
asS'd'
I9390
sa(dp18780
S'a'
I31118
sS's'
I17
sS'm'
(lp18781
I1
aI1
aI5
aI15
aI3
aI55
aI75
aI135
aI451
aI697
aI1407
aI2765
aI2443
aI11589
aI24863
aI47187
aI98477
asS'd'
I9391
sa(dp18782
S'a'
I31129
sS's'
I17
sS'm'
(lp18783
I1
aI1
aI1
aI13
aI27
aI37
aI77
aI157
aI121
aI603
aI491
aI2201
aI619
aI9157
aI32511
aI19843
aI49919
asS'd'
I9392
sa(dp18784
S'a'
I31132
sS's'
I17
sS'm'
(lp18785
I1
aI1
aI3
aI1
aI23
aI17
aI23
aI115
aI119
aI349
aI987
aI2587
aI1847
aI12099
aI31955
aI31685
aI1989
asS'd'
I9393
sa(dp18786
S'a'
I31141
sS's'
I17
sS'm'
(lp18787
I1
aI3
aI7
aI7
aI5
aI47
aI63
aI209
aI69
aI921
aI1041
aI1391
aI7485
aI11121
aI30993
aI5691
aI74551
asS'd'
I9394
sa(dp18788
S'a'
I31159
sS's'
I17
sS'm'
(lp18789
I1
aI3
aI1
aI3
aI23
aI61
aI55
aI253
aI355
aI299
aI971
aI279
aI3543
aI10073
aI5199
aI50539
aI88303
asS'd'
I9395
sa(dp18790
S'a'
I31183
sS's'
I17
sS'm'
(lp18791
I1
aI1
aI1
aI11
aI13
aI19
aI7
aI255
aI189
aI267
aI2021
aI93
aI219
aI10537
aI28627
aI58141
aI53675
asS'd'
I9396
sa(dp18792
S'a'
I31185
sS's'
I17
sS'm'
(lp18793
I1
aI3
aI3
aI7
aI27
aI61
aI83
aI95
aI163
aI777
aI1533
aI2485
aI7211
aI6979
aI4013
aI20797
aI91411
asS'd'
I9397
sa(dp18794
S'a'
I31195
sS's'
I17
sS'm'
(lp18795
I1
aI1
aI7
aI13
aI15
aI37
aI5
aI109
aI133
aI225
aI59
aI3855
aI3351
aI659
aI24321
aI63531
aI15573
asS'd'
I9398
sa(dp18796
S'a'
I31202
sS's'
I17
sS'm'
(lp18797
I1
aI1
aI5
aI1
aI7
aI55
aI59
aI213
aI45
aI77
aI2003
aI2921
aI1105
aI11089
aI17197
aI45459
aI67681
asS'd'
I9399
sa(dp18798
S'a'
I31213
sS's'
I17
sS'm'
(lp18799
I1
aI1
aI1
aI5
aI13
aI21
aI107
aI245
aI505
aI409
aI1453
aI1201
aI6945
aI2103
aI7377
aI59413
aI8785
asS'd'
I9400
sa(dp18800
S'a'
I31238
sS's'
I17
sS'm'
(lp18801
I1
aI1
aI1
aI13
aI5
aI37
aI73
aI55
aI39
aI219
aI225
aI3877
aI6583
aI3105
aI25355
aI14839
aI23435
asS'd'
I9401
sa(dp18802
S'a'
I31241
sS's'
I17
sS'm'
(lp18803
I1
aI1
aI7
aI1
aI21
aI35
aI87
aI227
aI487
aI767
aI609
aI1685
aI2731
aI10135
aI381
aI24021
aI122137
asS'd'
I9402
sa(dp18804
S'a'
I31252
sS's'
I17
sS'm'
(lp18805
I1
aI1
aI1
aI3
aI29
aI13
aI19
aI255
aI355
aI505
aI1757
aI3537
aI3029
aI11403
aI22685
aI61169
aI397
asS'd'
I9403
sa(dp18806
S'a'
I31262
sS's'
I17
sS'm'
(lp18807
I1
aI1
aI7
aI1
aI29
aI43
aI11
aI207
aI83
aI97
aI435
aI1453
aI4709
aI4193
aI18517
aI47203
aI3255
asS'd'
I9404
sa(dp18808
S'a'
I31265
sS's'
I17
sS'm'
(lp18809
I1
aI1
aI1
aI1
aI21
aI49
aI39
aI163
aI459
aI849
aI561
aI1207
aI4109
aI1435
aI17519
aI14839
aI1331
asS'd'
I9405
sa(dp18810
S'a'
I31295
sS's'
I17
sS'm'
(lp18811
I1
aI1
aI3
aI11
aI27
aI35
aI65
aI219
aI135
aI559
aI1111
aI2959
aI7835
aI5165
aI26641
aI22765
aI121829
asS'd'
I9406
sa(dp18812
S'a'
I31300
sS's'
I17
sS'm'
(lp18813
I1
aI3
aI5
aI3
aI23
aI31
aI57
aI149
aI431
aI451
aI189
aI1771
aI5877
aI3503
aI7531
aI46485
aI129031
asS'd'
I9407
sa(dp18814
S'a'
I31303
sS's'
I17
sS'm'
(lp18815
I1
aI1
aI3
aI11
aI1
aI13
aI47
aI17
aI331
aI1003
aI215
aI2797
aI689
aI6289
aI12719
aI37139
aI35827
asS'd'
I9408
sa(dp18816
S'a'
I31317
sS's'
I17
sS'm'
(lp18817
I1
aI3
aI5
aI9
aI19
aI13
aI11
aI29
aI275
aI165
aI783
aI313
aI2153
aI6009
aI2247
aI5699
aI128753
asS'd'
I9409
sa(dp18818
S'a'
I31318
sS's'
I17
sS'm'
(lp18819
I1
aI1
aI7
aI13
aI5
aI43
aI51
aI75
aI411
aI743
aI335
aI217
aI559
aI15389
aI6567
aI41193
aI127443
asS'd'
I9410
sa(dp18820
S'a'
I31324
sS's'
I17
sS'm'
(lp18821
I1
aI3
aI5
aI15
aI5
aI63
aI7
aI145
aI445
aI835
aI825
aI35
aI5951
aI5121
aI16365
aI36789
aI2941
asS'd'
I9411
sa(dp18822
S'a'
I31338
sS's'
I17
sS'm'
(lp18823
I1
aI3
aI5
aI5
aI29
aI1
aI61
aI19
aI427
aI245
aI445
aI3505
aI3647
aI8817
aI8031
aI64577
aI60745
asS'd'
I9412
sa(dp18824
S'a'
I31340
sS's'
I17
sS'm'
(lp18825
I1
aI1
aI3
aI9
aI29
aI9
aI35
aI225
aI55
aI535
aI1537
aI831
aI6483
aI16123
aI26079
aI32809
aI62227
asS'd'
I9413
sa(dp18826
S'a'
I31345
sS's'
I17
sS'm'
(lp18827
I1
aI3
aI7
aI7
aI25
aI33
aI15
aI61
aI343
aI749
aI1963
aI2763
aI3171
aI6755
aI6529
aI49449
aI88903
asS'd'
I9414
sa(dp18828
S'a'
I31355
sS's'
I17
sS'm'
(lp18829
I1
aI1
aI7
aI13
aI17
aI35
aI91
aI119
aI87
aI1023
aI1101
aI1785
aI2005
aI15947
aI21679
aI63179
aI3389
asS'd'
I9415
sa(dp18830
S'a'
I31362
sS's'
I17
sS'm'
(lp18831
I1
aI3
aI1
aI1
aI1
aI1
aI123
aI195
aI315
aI681
aI153
aI1621
aI5097
aI3669
aI20505
aI39305
aI127065
asS'd'
I9416
sa(dp18832
S'a'
I31371
sS's'
I17
sS'm'
(lp18833
I1
aI1
aI5
aI11
aI1
aI17
aI73
aI251
aI185
aI59
aI1723
aI2321
aI2103
aI6331
aI29571
aI63811
aI66651
asS'd'
I9417
sa(dp18834
S'a'
I31373
sS's'
I17
sS'm'
(lp18835
I1
aI1
aI7
aI13
aI15
aI19
aI111
aI91
aI211
aI85
aI711
aI2197
aI3107
aI2717
aI16725
aI52995
aI65791
asS'd'
I9418
sa(dp18836
S'a'
I31381
sS's'
I17
sS'm'
(lp18837
I1
aI3
aI3
aI9
aI21
aI41
aI53
aI145
aI459
aI155
aI93
aI2833
aI6747
aI737
aI30625
aI40581
aI65825
asS'd'
I9419
sa(dp18838
S'a'
I31391
sS's'
I17
sS'm'
(lp18839
I1
aI3
aI3
aI3
aI1
aI45
aI119
aI81
aI185
aI431
aI1221
aI3043
aI7277
aI10537
aI12355
aI42261
aI126117
asS'd'
I9420
sa(dp18840
S'a'
I31409
sS's'
I17
sS'm'
(lp18841
I1
aI3
aI7
aI7
aI11
aI47
aI37
aI41
aI123
aI643
aI707
aI2963
aI6183
aI15527
aI10951
aI24031
aI38187
asS'd'
I9421
sa(dp18842
S'a'
I31410
sS's'
I17
sS'm'
(lp18843
I1
aI3
aI1
aI7
aI13
aI57
aI1
aI149
aI117
aI627
aI1999
aI2805
aI4857
aI12805
aI31453
aI25699
aI109447
asS'd'
I9422
sa(dp18844
S'a'
I31412
sS's'
I17
sS'm'
(lp18845
I1
aI3
aI5
aI3
aI9
aI37
aI83
aI221
aI77
aI573
aI661
aI465
aI1279
aI7355
aI24061
aI36151
aI96595
asS'd'
I9423
sa(dp18846
S'a'
I31434
sS's'
I17
sS'm'
(lp18847
I1
aI1
aI7
aI15
aI29
aI31
aI125
aI205
aI449
aI563
aI1263
aI3427
aI8013
aI14025
aI15235
aI11833
aI25601
asS'd'
I9424
sa(dp18848
S'a'
I31458
sS's'
I17
sS'm'
(lp18849
I1
aI3
aI7
aI11
aI31
aI35
aI99
aI193
aI163
aI527
aI1455
aI395
aI4853
aI2561
aI11909
aI57311
aI101007
asS'd'
I9425
sa(dp18850
S'a'
I31467
sS's'
I17
sS'm'
(lp18851
I1
aI1
aI5
aI3
aI17
aI39
aI99
aI173
aI497
aI245
aI1671
aI3457
aI83
aI11959
aI2963
aI3401
aI102259
asS'd'
I9426
sa(dp18852
S'a'
I31470
sS's'
I17
sS'm'
(lp18853
I1
aI1
aI1
aI5
aI25
aI41
aI119
aI81
aI301
aI797
aI661
aI2543
aI1195
aI2111
aI1785
aI41533
aI51947
asS'd'
I9427
sa(dp18854
S'a'
I31475
sS's'
I17
sS'm'
(lp18855
I1
aI3
aI3
aI13
aI5
aI59
aI61
aI153
aI213
aI541
aI1849
aI249
aI3897
aI3877
aI17095
aI6857
aI76781
asS'd'
I9428
sa(dp18856
S'a'
I31481
sS's'
I17
sS'm'
(lp18857
I1
aI3
aI7
aI1
aI19
aI13
aI57
aI47
aI359
aI165
aI1085
aI2263
aI3261
aI12825
aI17405
aI25853
aI20731
asS'd'
I9429
sa(dp18858
S'a'
I31482
sS's'
I17
sS'm'
(lp18859
I1
aI1
aI1
aI7
aI7
aI43
aI7
aI65
aI51
aI503
aI173
aI1023
aI283
aI14809
aI1183
aI33497
aI110683
asS'd'
I9430
sa(dp18860
S'a'
I31484
sS's'
I17
sS'm'
(lp18861
I1
aI3
aI5
aI11
aI19
aI51
aI29
aI157
aI159
aI191
aI1293
aI2951
aI6569
aI12433
aI14587
aI30631
aI30485
asS'd'
I9431
sa(dp18862
S'a'
I31492
sS's'
I17
sS'm'
(lp18863
I1
aI3
aI7
aI5
aI1
aI27
aI25
aI221
aI255
aI471
aI779
aI3991
aI6985
aI1803
aI28451
aI33403
aI5567
asS'd'
I9432
sa(dp18864
S'a'
I31507
sS's'
I17
sS'm'
(lp18865
I1
aI1
aI5
aI5
aI7
aI29
aI55
aI241
aI457
aI863
aI1715
aI3393
aI4127
aI13985
aI6313
aI13683
aI114837
asS'd'
I9433
sa(dp18866
S'a'
I31514
sS's'
I17
sS'm'
(lp18867
I1
aI3
aI5
aI5
aI11
aI27
aI55
aI109
aI247
aI199
aI1593
aI2881
aI307
aI97
aI24751
aI35921
aI121931
asS'd'
I9434
sa(dp18868
S'a'
I31538
sS's'
I17
sS'm'
(lp18869
I1
aI3
aI1
aI13
aI3
aI59
aI17
aI161
aI47
aI467
aI1019
aI3629
aI3017
aI15645
aI3983
aI32393
aI79213
asS'd'
I9435
sa(dp18870
S'a'
I31547
sS's'
I17
sS'm'
(lp18871
I1
aI1
aI3
aI11
aI19
aI57
aI67
aI199
aI319
aI107
aI2043
aI2045
aI4025
aI5733
aI29979
aI37721
aI117031
asS'd'
I9436
sa(dp18872
S'a'
I31549
sS's'
I17
sS'm'
(lp18873
I1
aI3
aI7
aI11
aI9
aI23
aI31
aI81
aI177
aI801
aI1177
aI3451
aI7777
aI15351
aI7579
aI39033
aI23847
asS'd'
I9437
sa(dp18874
S'a'
I31555
sS's'
I17
sS'm'
(lp18875
I1
aI1
aI3
aI5
aI17
aI61
aI63
aI7
aI371
aI905
aI1147
aI1383
aI4075
aI6721
aI17503
aI32015
aI112547
asS'd'
I9438
sa(dp18876
S'a'
I31557
sS's'
I17
sS'm'
(lp18877
I1
aI1
aI3
aI13
aI13
aI25
aI69
aI159
aI49
aI133
aI227
aI2155
aI1603
aI10077
aI3429
aI39131
aI18949
asS'd'
I9439
sa(dp18878
S'a'
I31597
sS's'
I17
sS'm'
(lp18879
I1
aI3
aI5
aI3
aI29
aI5
aI115
aI93
aI243
aI791
aI1113
aI2841
aI4733
aI3041
aI31733
aI28539
aI84567
asS'd'
I9440
sa(dp18880
S'a'
I31598
sS's'
I17
sS'm'
(lp18881
I1
aI3
aI3
aI7
aI21
aI9
aI5
aI95
aI489
aI517
aI1453
aI2697
aI7951
aI12369
aI19571
aI29811
aI51805
asS'd'
I9441
sa(dp18882
S'a'
I31610
sS's'
I17
sS'm'
(lp18883
I1
aI1
aI5
aI9
aI1
aI29
aI97
aI191
aI73
aI357
aI745
aI2787
aI7815
aI4565
aI19761
aI33729
aI86849
asS'd'
I9442
sa(dp18884
S'a'
I31625
sS's'
I17
sS'm'
(lp18885
I1
aI3
aI5
aI13
aI3
aI5
aI35
aI79
aI387
aI813
aI1673
aI3187
aI337
aI5539
aI6761
aI46903
aI122967
asS'd'
I9443
sa(dp18886
S'a'
I31634
sS's'
I17
sS'm'
(lp18887
I1
aI1
aI7
aI11
aI1
aI15
aI125
aI175
aI255
aI35
aI145
aI2391
aI887
aI10505
aI11587
aI53941
aI5089
asS'd'
I9444
sa(dp18888
S'a'
I31643
sS's'
I17
sS'm'
(lp18889
I1
aI1
aI7
aI13
aI9
aI13
aI15
aI215
aI361
aI227
aI1665
aI3345
aI3615
aI14031
aI16281
aI4457
aI52037
asS'd'
I9445
sa(dp18890
S'a'
I31645
sS's'
I17
sS'm'
(lp18891
I1
aI3
aI5
aI9
aI31
aI21
aI3
aI189
aI211
aI855
aI1781
aI2097
aI1345
aI6763
aI27651
aI54137
aI52689
asS'd'
I9446
sa(dp18892
S'a'
I31659
sS's'
I17
sS'm'
(lp18893
I1
aI3
aI1
aI5
aI29
aI9
aI99
aI183
aI183
aI205
aI149
aI53
aI7179
aI3387
aI9603
aI4281
aI47145
asS'd'
I9447
sa(dp18894
S'a'
I31669
sS's'
I17
sS'm'
(lp18895
I1
aI3
aI1
aI11
aI13
aI35
aI97
aI21
aI29
aI877
aI191
aI1621
aI2501
aI4283
aI1707
aI48957
aI129029
asS'd'
I9448
sa(dp18896
S'a'
I31670
sS's'
I17
sS'm'
(lp18897
I1
aI3
aI1
aI9
aI5
aI19
aI57
aI219
aI105
aI467
aI1179
aI3155
aI7743
aI4835
aI14845
aI35671
aI47655
asS'd'
I9449
sa(dp18898
S'a'
I31682
sS's'
I17
sS'm'
(lp18899
I1
aI3
aI1
aI7
aI27
aI41
aI27
aI185
aI271
aI611
aI1173
aI2875
aI529
aI11619
aI20231
aI18741
aI41799
asS'd'
I9450
sa(dp18900
S'a'
I31694
sS's'
I17
sS'm'
(lp18901
I1
aI3
aI7
aI13
aI9
aI3
aI35
aI71
aI467
aI689
aI1797
aI319
aI6657
aI13193
aI15861
aI7567
aI12891
asS'd'
I9451
sa(dp18902
S'a'
I31717
sS's'
I17
sS'm'
(lp18903
I1
aI1
aI7
aI13
aI19
aI57
aI25
aI141
aI195
aI995
aI859
aI811
aI4685
aI6711
aI8963
aI49657
aI54751
asS'd'
I9452
sa(dp18904
S'a'
I31718
sS's'
I17
sS'm'
(lp18905
I1
aI1
aI1
aI11
aI27
aI25
aI9
aI91
aI97
aI251
aI757
aI2783
aI5447
aI3617
aI26801
aI32501
aI55245
asS'd'
I9453
sa(dp18906
S'a'
I31729
sS's'
I17
sS'm'
(lp18907
I1
aI3
aI7
aI1
aI5
aI1
aI103
aI129
aI127
aI593
aI857
aI3957
aI3665
aI10279
aI26211
aI2095
aI15869
asS'd'
I9454
sa(dp18908
S'a'
I31736
sS's'
I17
sS'm'
(lp18909
I1
aI1
aI7
aI1
aI25
aI49
aI3
aI139
aI25
aI545
aI615
aI1353
aI4103
aI1099
aI21729
aI45383
aI110611
asS'd'
I9455
sa(dp18910
S'a'
I31742
sS's'
I17
sS'm'
(lp18911
I1
aI3
aI5
aI3
aI7
aI49
aI83
aI41
aI209
aI357
aI939
aI849
aI5851
aI3945
aI831
aI8131
aI105897
asS'd'
I9456
sa(dp18912
S'a'
I31749
sS's'
I17
sS'm'
(lp18913
I1
aI1
aI1
aI3
aI27
aI19
aI123
aI71
aI195
aI1019
aI1021
aI1287
aI5665
aI5277
aI8647
aI27033
aI89539
asS'd'
I9457
sa(dp18914
S'a'
I31773
sS's'
I17
sS'm'
(lp18915
I1
aI1
aI1
aI9
aI27
aI51
aI49
aI159
aI401
aI1013
aI763
aI653
aI1449
aI12441
aI21191
aI28871
aI106181
asS'd'
I9458
sa(dp18916
S'a'
I31777
sS's'
I17
sS'm'
(lp18917
I1
aI1
aI5
aI7
aI31
aI7
aI105
aI137
aI331
aI367
aI1305
aI2761
aI863
aI3915
aI12633
aI32251
aI82867
asS'd'
I9459
sa(dp18918
S'a'
I31778
sS's'
I17
sS'm'
(lp18919
I1
aI3
aI7
aI11
aI9
aI47
aI35
aI57
aI137
aI269
aI443
aI79
aI11
aI11817
aI28995
aI46681
aI104263
asS'd'
I9460
sa(dp18920
S'a'
I31784
sS's'
I17
sS'm'
(lp18921
I1
aI3
aI1
aI5
aI3
aI25
aI89
aI179
aI183
aI835
aI367
aI2215
aI295
aI5365
aI1899
aI10785
aI88979
asS'd'
I9461
sa(dp18922
S'a'
I31801
sS's'
I17
sS'm'
(lp18923
I1
aI3
aI7
aI13
aI3
aI5
aI93
aI43
aI409
aI363
aI267
aI2077
aI3745
aI445
aI25957
aI34103
aI29475
asS'd'
I9462
sa(dp18924
S'a'
I31812
sS's'
I17
sS'm'
(lp18925
I1
aI1
aI1
aI7
aI27
aI21
aI121
aI29
aI171
aI783
aI553
aI265
aI6835
aI3929
aI18127
aI33463
aI70999
asS'd'
I9463
sa(dp18926
S'a'
I31821
sS's'
I17
sS'm'
(lp18927
I1
aI3
aI3
aI15
aI15
aI55
aI13
aI1
aI297
aI935
aI1307
aI1779
aI2239
aI15471
aI32453
aI30649
aI45973
asS'd'
I9464
sa(dp18928
S'a'
I31822
sS's'
I17
sS'm'
(lp18929
I1
aI3
aI7
aI5
aI25
aI41
aI3
aI171
aI347
aI607
aI1873
aI1087
aI2433
aI8377
aI7959
aI19941
aI117319
asS'd'
I9465
sa(dp18930
S'a'
I31836
sS's'
I17
sS'm'
(lp18931
I1
aI1
aI1
aI3
aI5
aI47
aI107
aI69
aI431
aI63
aI325
aI1241
aI3487
aI11249
aI28559
aI30001
aI93789
asS'd'
I9466
sa(dp18932
S'a'
I31850
sS's'
I17
sS'm'
(lp18933
I1
aI1
aI1
aI5
aI15
aI17
aI9
aI145
aI335
aI169
aI1099
aI3637
aI5397
aI6711
aI16095
aI27053
aI124247
asS'd'
I9467
sa(dp18934
S'a'
I31855
sS's'
I17
sS'm'
(lp18935
I1
aI3
aI3
aI5
aI3
aI9
aI65
aI97
aI421
aI951
aI2003
aI2837
aI7095
aI15685
aI5147
aI56801
aI98679
asS'd'
I9468
sa(dp18936
S'a'
I31858
sS's'
I17
sS'm'
(lp18937
I1
aI3
aI7
aI15
aI1
aI33
aI115
aI45
aI215
aI253
aI361
aI555
aI787
aI15483
aI25531
aI53273
aI8933
asS'd'
I9469
sa(dp18938
S'a'
I31860
sS's'
I17
sS'm'
(lp18939
I1
aI3
aI1
aI9
aI3
aI63
aI47
aI205
aI457
aI977
aI991
aI3189
aI1369
aI14899
aI10937
aI56999
aI11525
asS'd'
I9470
sa(dp18940
S'a'
I31886
sS's'
I17
sS'm'
(lp18941
I1
aI1
aI7
aI5
aI11
aI61
aI53
aI55
aI231
aI357
aI1695
aI2489
aI2355
aI7583
aI14097
aI50039
aI96595
asS'd'
I9471
sa(dp18942
S'a'
I31891
sS's'
I17
sS'm'
(lp18943
I1
aI3
aI7
aI7
aI3
aI57
aI115
aI245
aI259
aI573
aI1275
aI2971
aI1793
aI13683
aI8683
aI51815
aI26807
asS'd'
I9472
sa(dp18944
S'a'
I31909
sS's'
I17
sS'm'
(lp18945
I1
aI1
aI5
aI3
aI17
aI59
aI55
aI237
aI491
aI757
aI1447
aI2941
aI2641
aI14175
aI4401
aI4367
aI36853
asS'd'
I9473
sa(dp18946
S'a'
I31928
sS's'
I17
sS'm'
(lp18947
I1
aI3
aI1
aI15
aI3
aI63
aI67
aI1
aI403
aI79
aI1161
aI2379
aI3337
aI14447
aI5877
aI40759
aI12573
asS'd'
I9474
sa(dp18948
S'a'
I31931
sS's'
I17
sS'm'
(lp18949
I1
aI1
aI7
aI15
aI17
aI1
aI91
aI5
aI173
aI215
aI1567
aI1851
aI3309
aI9813
aI21215
aI19151
aI96785
asS'd'
I9475
sa(dp18950
S'a'
I31934
sS's'
I17
sS'm'
(lp18951
I1
aI1
aI1
aI9
aI31
aI45
aI123
aI221
aI397
aI51
aI1489
aI3247
aI923
aI10423
aI10461
aI51231
aI92909
asS'd'
I9476
sa(dp18952
S'a'
I31941
sS's'
I17
sS'm'
(lp18953
I1
aI1
aI1
aI13
aI27
aI17
aI105
aI163
aI403
aI193
aI1487
aI2421
aI4415
aI14303
aI6419
aI24105
aI29997
asS'd'
I9477
sa(dp18954
S'a'
I31942
sS's'
I17
sS'm'
(lp18955
I1
aI1
aI5
aI13
aI31
aI55
aI17
aI125
aI341
aI219
aI401
aI1611
aI891
aI12909
aI13949
aI46245
aI26769
asS'd'
I9478
sa(dp18956
S'a'
I31945
sS's'
I17
sS'm'
(lp18957
I1
aI3
aI7
aI3
aI31
aI41
aI65
aI207
aI311
aI643
aI1617
aI271
aI3749
aI14635
aI26385
aI55251
aI50719
asS'd'
I9479
sa(dp18958
S'a'
I31951
sS's'
I17
sS'm'
(lp18959
I1
aI3
aI3
aI13
aI7
aI55
aI69
aI241
aI413
aI399
aI137
aI2255
aI5395
aI12625
aI26583
aI64603
aI22571
asS'd'
I9480
sa(dp18960
S'a'
I31959
sS's'
I17
sS'm'
(lp18961
I1
aI3
aI5
aI3
aI31
aI15
aI15
aI161
aI153
aI445
aI595
aI273
aI6631
aI12845
aI23331
aI16963
aI52099
asS'd'
I9481
sa(dp18962
S'a'
I31963
sS's'
I17
sS'm'
(lp18963
I1
aI3
aI3
aI1
aI27
aI39
aI71
aI41
aI455
aI841
aI831
aI1719
aI3531
aI5113
aI29183
aI1933
aI42227
asS'd'
I9482
sa(dp18964
S'a'
I31970
sS's'
I17
sS'm'
(lp18965
I1
aI3
aI7
aI3
aI1
aI15
aI31
aI183
aI429
aI557
aI1747
aI1059
aI2079
aI16361
aI29103
aI43207
aI921
asS'd'
I9483
sa(dp18966
S'a'
I31984
sS's'
I17
sS'm'
(lp18967
I1
aI3
aI1
aI1
aI31
aI39
aI97
aI73
aI339
aI405
aI1423
aI2215
aI5435
aI9205
aI1889
aI58249
aI61517
asS'd'
I9484
sa(dp18968
S'a'
I31987
sS's'
I17
sS'm'
(lp18969
I1
aI3
aI7
aI1
aI23
aI59
aI127
aI245
aI11
aI627
aI1555
aI2497
aI6427
aI7205
aI22675
aI62847
aI69691
asS'd'
I9485
sa(dp18970
S'a'
I31990
sS's'
I17
sS'm'
(lp18971
I1
aI1
aI3
aI5
aI1
aI13
aI95
aI9
aI167
aI481
aI947
aI3181
aI8057
aI5559
aI7537
aI33757
aI72419
asS'd'
I9486
sa(dp18972
S'a'
I32001
sS's'
I17
sS'm'
(lp18973
I1
aI1
aI7
aI3
aI15
aI9
aI105
aI205
aI287
aI375
aI115
aI1731
aI1063
aI11551
aI12077
aI41013
aI88853
asS'd'
I9487
sa(dp18974
S'a'
I32007
sS's'
I17
sS'm'
(lp18975
I1
aI3
aI3
aI9
aI5
aI63
aI127
aI33
aI409
aI279
aI1379
aI4069
aI4091
aI14703
aI27435
aI19525
aI71261
asS'd'
I9488
sa(dp18976
S'a'
I32008
sS's'
I17
sS'm'
(lp18977
I1
aI3
aI1
aI13
aI31
aI31
aI59
aI205
aI167
aI131
aI891
aI1259
aI6909
aI211
aI31517
aI8085
aI112065
asS'd'
I9489
sa(dp18978
S'a'
I32025
sS's'
I17
sS'm'
(lp18979
I1
aI1
aI5
aI11
aI17
aI25
aI119
aI77
aI449
aI569
aI381
aI825
aI2459
aI983
aI2959
aI51611
aI90721
asS'd'
I9490
sa(dp18980
S'a'
I32035
sS's'
I17
sS'm'
(lp18981
I1
aI3
aI1
aI7
aI17
aI55
aI91
aI231
aI133
aI541
aI499
aI3609
aI4237
aI11627
aI30007
aI58911
aI43443
asS'd'
I9491
sa(dp18982
S'a'
I32038
sS's'
I17
sS'm'
(lp18983
I1
aI3
aI7
aI7
aI29
aI5
aI47
aI187
aI71
aI695
aI1389
aI2855
aI5815
aI11605
aI3643
aI24961
aI25793
asS'd'
I9492
sa(dp18984
S'a'
I32047
sS's'
I17
sS'm'
(lp18985
I1
aI3
aI3
aI5
aI11
aI31
aI43
aI31
aI185
aI1021
aI795
aI3585
aI3981
aI8627
aI18117
aI42351
aI19513
asS'd'
I9493
sa(dp18986
S'a'
I32049
sS's'
I17
sS'm'
(lp18987
I1
aI1
aI5
aI13
aI9
aI3
aI115
aI45
aI39
aI577
aI1847
aI653
aI2625
aI9367
aI27923
aI35661
aI113613
asS'd'
I9494
sa(dp18988
S'a'
I32062
sS's'
I17
sS'm'
(lp18989
I1
aI3
aI7
aI7
aI17
aI9
aI69
aI233
aI367
aI673
aI11
aI2215
aI1177
aI4501
aI9693
aI62013
aI45647
asS'd'
I9495
sa(dp18990
S'a'
I32067
sS's'
I17
sS'm'
(lp18991
I1
aI3
aI5
aI7
aI7
aI53
aI11
aI227
aI465
aI843
aI2017
aI689
aI6767
aI10321
aI25163
aI56561
aI6865
asS'd'
I9496
sa(dp18992
S'a'
I32070
sS's'
I17
sS'm'
(lp18993
I1
aI3
aI3
aI5
aI13
aI43
aI119
aI9
aI185
aI893
aI133
aI863
aI7137
aI6653
aI7875
aI23167
aI13893
asS'd'
I9497
sa(dp18994
S'a'
I32073
sS's'
I17
sS'm'
(lp18995
I1
aI3
aI5
aI9
aI1
aI47
aI17
aI85
aI273
aI901
aI493
aI2411
aI983
aI15717
aI25151
aI21323
aI57939
asS'd'
I9498
sa(dp18996
S'a'
I32074
sS's'
I17
sS'm'
(lp18997
I1
aI1
aI7
aI5
aI19
aI17
aI49
aI37
aI425
aI443
aI781
aI2593
aI4929
aI12313
aI12727
aI42285
aI88451
asS'd'
I9499
sa(dp18998
S'a'
I32079
sS's'
I17
sS'm'
(lp18999
I1
aI3
aI3
aI11
aI9
aI53
aI17
aI67
aI237
aI463
aI1509
aI2153
aI3715
aI7909
aI21151
aI64517
aI87695
asS'd'
I9500
sa(dp19000
S'a'
I32081
sS's'
I17
sS'm'
(lp19001
I1
aI3
aI7
aI1
aI29
aI39
aI25
aI83
aI413
aI1005
aI2011
aI3933
aI2911
aI7041
aI10537
aI23135
aI22671
asS'd'
I9501
sa(dp19002
S'a'
I32082
sS's'
I17
sS'm'
(lp19003
I1
aI1
aI3
aI9
aI23
aI61
aI117
aI33
aI431
aI181
aI1819
aI683
aI1809
aI1723
aI27041
aI29113
aI99347
asS'd'
I9502
sa(dp19004
S'a'
I32107
sS's'
I17
sS'm'
(lp19005
I1
aI1
aI5
aI11
aI11
aI7
aI101
aI181
aI51
aI857
aI923
aI3495
aI7123
aI7775
aI30081
aI48513
aI116137
asS'd'
I9503
sa(dp19006
S'a'
I32127
sS's'
I17
sS'm'
(lp19007
I1
aI1
aI3
aI11
aI15
aI31
aI97
aI127
aI365
aI799
aI715
aI2101
aI6081
aI11607
aI1055
aI35027
aI62967
asS'd'
I9504
sa(dp19008
S'a'
I32145
sS's'
I17
sS'm'
(lp19009
I1
aI3
aI5
aI7
aI3
aI31
aI109
aI247
aI225
aI221
aI1093
aI2633
aI1847
aI7427
aI8767
aI16581
aI32145
asS'd'
I9505
sa(dp19010
S'a'
I32151
sS's'
I17
sS'm'
(lp19011
I1
aI3
aI1
aI7
aI15
aI23
aI43
aI109
aI327
aI417
aI1895
aI2333
aI6265
aI6599
aI6623
aI47375
aI92731
asS'd'
I9506
sa(dp19012
S'a'
I32152
sS's'
I17
sS'm'
(lp19013
I1
aI3
aI7
aI1
aI29
aI29
aI45
aI217
aI163
aI941
aI1327
aI3685
aI5481
aI15783
aI26281
aI60339
aI34277
asS'd'
I9507
sa(dp19014
S'a'
I32173
sS's'
I17
sS'm'
(lp19015
I1
aI1
aI7
aI11
aI1
aI7
aI119
aI201
aI29
aI193
aI1805
aI1395
aI267
aI2011
aI637
aI26765
aI48883
asS'd'
I9508
sa(dp19016
S'a'
I32174
sS's'
I17
sS'm'
(lp19017
I1
aI1
aI3
aI7
aI11
aI63
aI41
aI89
aI365
aI729
aI25
aI3185
aI2143
aI1737
aI29693
aI7443
aI78079
asS'd'
I9509
sa(dp19018
S'a'
I32186
sS's'
I17
sS'm'
(lp19019
I1
aI3
aI1
aI13
aI25
aI27
aI63
aI233
aI79
aI1007
aI1357
aI679
aI7581
aI8333
aI2469
aI31787
aI128531
asS'd'
I9510
sa(dp19020
S'a'
I32194
sS's'
I17
sS'm'
(lp19021
I1
aI3
aI1
aI3
aI23
aI39
aI53
aI99
aI219
aI475
aI931
aI507
aI3615
aI10613
aI14663
aI1151
aI123459
asS'd'
I9511
sa(dp19022
S'a'
I32196
sS's'
I17
sS'm'
(lp19023
I1
aI1
aI1
aI1
aI13
aI15
aI67
aI45
aI393
aI791
aI415
aI2731
aI1151
aI8935
aI28983
aI7239
aI106247
asS'd'
I9512
sa(dp19024
S'a'
I32200
sS's'
I17
sS'm'
(lp19025
I1
aI3
aI7
aI7
aI11
aI35
aI95
aI153
aI421
aI193
aI1997
aI2587
aI3183
aI9229
aI17663
aI28221
aI6759
asS'd'
I9513
sa(dp19026
S'a'
I32208
sS's'
I17
sS'm'
(lp19027
I1
aI3
aI1
aI7
aI5
aI5
aI123
aI55
aI509
aI973
aI261
aI463
aI2723
aI15225
aI1925
aI62283
aI86329
asS'd'
I9514
sa(dp19028
S'a'
I32218
sS's'
I17
sS'm'
(lp19029
I1
aI1
aI3
aI13
aI5
aI47
aI123
aI239
aI273
aI407
aI1725
aI717
aI1229
aI1387
aI11743
aI13739
aI104503
asS'd'
I9515
sa(dp19030
S'a'
I32236
sS's'
I17
sS'm'
(lp19031
I1
aI3
aI3
aI13
aI23
aI35
aI43
aI113
aI299
aI847
aI1903
aI3445
aI3395
aI641
aI11271
aI61517
aI40747
asS'd'
I9516
sa(dp19032
S'a'
I32260
sS's'
I17
sS'm'
(lp19033
I1
aI3
aI1
aI15
aI17
aI49
aI97
aI9
aI335
aI731
aI151
aI167
aI8129
aI11845
aI18285
aI20113
aI122397
asS'd'
I9517
sa(dp19034
S'a'
I32263
sS's'
I17
sS'm'
(lp19035
I1
aI1
aI5
aI11
aI11
aI63
aI3
aI153
aI345
aI511
aI1939
aI1815
aI7231
aI10555
aI14293
aI50753
aI14681
asS'd'
I9518
sa(dp19036
S'a'
I32288
sS's'
I17
sS'm'
(lp19037
I1
aI3
aI7
aI5
aI21
aI31
aI127
aI223
aI241
aI783
aI887
aI3519
aI4743
aI3541
aI4143
aI57461
aI27791
asS'd'
I9519
sa(dp19038
S'a'
I32298
sS's'
I17
sS'm'
(lp19039
I1
aI1
aI5
aI7
aI13
aI15
aI83
aI225
aI201
aI979
aI145
aI769
aI1491
aI12155
aI21307
aI64877
aI113277
asS'd'
I9520
sa(dp19040
S'a'
I32315
sS's'
I17
sS'm'
(lp19041
I1
aI1
aI7
aI1
aI27
aI25
aI105
aI69
aI239
aI323
aI1059
aI573
aI4913
aI14215
aI27007
aI42351
aI66315
asS'd'
I9521
sa(dp19042
S'a'
I32332
sS's'
I17
sS'm'
(lp19043
I1
aI1
aI3
aI11
aI21
aI33
aI93
aI23
aI363
aI633
aI935
aI637
aI6171
aI12695
aI14077
aI17505
aI69681
asS'd'
I9522
sa(dp19044
S'a'
I32340
sS's'
I17
sS'm'
(lp19045
I1
aI3
aI1
aI5
aI15
aI11
aI93
aI211
aI175
aI377
aI33
aI1403
aI5097
aI1503
aI8483
aI2881
aI85877
asS'd'
I9523
sa(dp19046
S'a'
I32354
sS's'
I17
sS'm'
(lp19047
I1
aI1
aI5
aI3
aI5
aI51
aI5
aI255
aI429
aI661
aI625
aI3015
aI4813
aI3573
aI22917
aI45967
aI70559
asS'd'
I9524
sa(dp19048
S'a'
I32359
sS's'
I17
sS'm'
(lp19049
I1
aI1
aI7
aI3
aI11
aI41
aI3
aI197
aI181
aI897
aI767
aI1385
aI7395
aI15543
aI4655
aI40309
aI73169
asS'd'
I9525
sa(dp19050
S'a'
I32366
sS's'
I17
sS'm'
(lp19051
I1
aI1
aI5
aI9
aI15
aI35
aI71
aI119
aI509
aI817
aI1169
aI75
aI1337
aI2959
aI611
aI38243
aI46987
asS'd'
I9526
sa(dp19052
S'a'
I32368
sS's'
I17
sS'm'
(lp19053
I1
aI1
aI1
aI9
aI1
aI7
aI43
aI65
aI479
aI625
aI1685
aI1309
aI5619
aI14163
aI13633
aI18169
aI8311
asS'd'
I9527
sa(dp19054
S'a'
I32377
sS's'
I17
sS'm'
(lp19055
I1
aI3
aI5
aI9
aI19
aI39
aI95
aI105
aI273
aI1023
aI79
aI229
aI6895
aI2931
aI5717
aI27911
aI22139
asS'd'
I9528
sa(dp19056
S'a'
I32384
sS's'
I17
sS'm'
(lp19057
I1
aI3
aI5
aI7
aI1
aI55
aI15
aI15
aI297
aI731
aI2029
aI2789
aI11
aI1333
aI26571
aI62595
aI15131
asS'd'
I9529
sa(dp19058
S'a'
I32399
sS's'
I17
sS'm'
(lp19059
I1
aI1
aI5
aI7
aI29
aI35
aI3
aI125
aI381
aI709
aI2047
aI2395
aI6315
aI2301
aI7175
aI19857
aI75085
asS'd'
I9530
sa(dp19060
S'a'
I32417
sS's'
I17
sS'm'
(lp19061
I1
aI1
aI5
aI15
aI23
aI45
aI95
aI117
aI49
aI635
aI1525
aI1105
aI7335
aI4653
aI18159
aI29729
aI62627
asS'd'
I9531
sa(dp19062
S'a'
I32424
sS's'
I17
sS'm'
(lp19063
I1
aI3
aI3
aI11
aI29
aI19
aI29
aI169
aI141
aI243
aI1765
aI1829
aI4555
aI16299
aI3053
aI58933
aI44605
asS'd'
I9532
sa(dp19064
S'a'
I32427
sS's'
I17
sS'm'
(lp19065
I1
aI1
aI3
aI15
aI5
aI45
aI35
aI213
aI385
aI993
aI1521
aI9
aI3561
aI10497
aI12601
aI38163
aI86501
asS'd'
I9533
sa(dp19066
S'a'
I32429
sS's'
I17
sS'm'
(lp19067
I1
aI3
aI3
aI13
aI9
aI23
aI109
aI95
aI491
aI1003
aI473
aI3325
aI6577
aI14617
aI17765
aI33391
aI82927
asS'd'
I9534
sa(dp19068
S'a'
I32438
sS's'
I17
sS'm'
(lp19069
I1
aI3
aI3
aI11
aI25
aI31
aI93
aI111
aI231
aI71
aI1233
aI3581
aI6789
aI4569
aI16741
aI61967
aI32249
asS'd'
I9535
sa(dp19070
S'a'
I32442
sS's'
I17
sS'm'
(lp19071
I1
aI3
aI3
aI15
aI15
aI63
aI39
aI247
aI79
aI923
aI327
aI2639
aI2013
aI12325
aI18133
aI60623
aI2215
asS'd'
I9536
sa(dp19072
S'a'
I32447
sS's'
I17
sS'm'
(lp19073
I1
aI3
aI5
aI1
aI5
aI49
aI121
aI53
aI283
aI529
aI37
aI3233
aI6285
aI12447
aI4355
aI9343
aI45631
asS'd'
I9537
sa(dp19074
S'a'
I32469
sS's'
I17
sS'm'
(lp19075
I1
aI1
aI7
aI11
aI11
aI11
aI111
aI139
aI429
aI279
aI1019
aI2139
aI2033
aI6809
aI8847
aI22535
aI107005
asS'd'
I9538
sa(dp19076
S'a'
I32479
sS's'
I17
sS'm'
(lp19077
I1
aI3
aI5
aI1
aI1
aI21
aI35
aI97
aI167
aI57
aI491
aI511
aI4065
aI11699
aI16851
aI6847
aI40929
asS'd'
I9539
sa(dp19078
S'a'
I32483
sS's'
I17
sS'm'
(lp19079
I1
aI3
aI1
aI15
aI3
aI55
aI113
aI33
aI255
aI537
aI835
aI1867
aI3927
aI839
aI955
aI29079
aI93727
asS'd'
I9540
sa(dp19080
S'a'
I32498
sS's'
I17
sS'm'
(lp19081
I1
aI1
aI7
aI3
aI5
aI7
aI35
aI111
aI165
aI885
aI115
aI3051
aI4541
aI1701
aI22827
aI361
aI91843
asS'd'
I9541
sa(dp19082
S'a'
I32503
sS's'
I17
sS'm'
(lp19083
I1
aI1
aI7
aI11
aI7
aI55
aI81
aI43
aI237
aI725
aI1761
aI1599
aI639
aI14189
aI31241
aI52827
aI107943
asS'd'
I9542
sa(dp19084
S'a'
I32507
sS's'
I17
sS'm'
(lp19085
I1
aI3
aI1
aI3
aI29
aI35
aI67
aI119
aI369
aI877
aI1861
aI123
aI8121
aI13861
aI31155
aI60245
aI79799
asS'd'
I9543
sa(dp19086
S'a'
I32521
sS's'
I17
sS'm'
(lp19087
I1
aI1
aI3
aI13
aI7
aI49
aI63
aI19
aI253
aI723
aI639
aI1677
aI291
aI13697
aI22231
aI46893
aI90069
asS'd'
I9544
sa(dp19088
S'a'
I32532
sS's'
I17
sS'm'
(lp19089
I1
aI3
aI5
aI1
aI7
aI57
aI29
aI233
aI35
aI715
aI515
aI3221
aI2715
aI13839
aI18321
aI4445
aI103843
asS'd'
I9545
sa(dp19090
S'a'
I32539
sS's'
I17
sS'm'
(lp19091
I1
aI3
aI1
aI7
aI1
aI63
aI33
aI7
aI481
aI461
aI1923
aI2679
aI2441
aI5449
aI13233
aI2245
aI48667
asS'd'
I9546
sa(dp19092
S'a'
I32551
sS's'
I17
sS'm'
(lp19093
I1
aI1
aI7
aI11
aI11
aI9
aI95
aI151
aI441
aI333
aI1871
aI1181
aI3027
aI12887
aI11923
aI63847
aI6953
asS'd'
I9547
sa(dp19094
S'a'
I32572
sS's'
I17
sS'm'
(lp19095
I1
aI3
aI5
aI5
aI15
aI33
aI53
aI47
aI351
aI387
aI55
aI393
aI5475
aI3027
aI18565
aI37997
aI120877
asS'd'
I9548
sa(dp19096
S'a'
I32577
sS's'
I17
sS'm'
(lp19097
I1
aI3
aI5
aI9
aI23
aI43
aI67
aI97
aI445
aI783
aI1499
aI1977
aI1441
aI10159
aI13479
aI149
aI4939
asS'd'
I9549
sa(dp19098
S'a'
I32578
sS's'
I17
sS'm'
(lp19099
I1
aI3
aI7
aI3
aI15
aI41
aI119
aI55
aI139
aI25
aI849
aI857
aI53
aI10421
aI2683
aI24839
aI107797
asS'd'
I9550
sa(dp19100
S'a'
I32587
sS's'
I17
sS'm'
(lp19101
I1
aI1
aI7
aI13
aI25
aI51
aI51
aI13
aI333
aI93
aI95
aI1755
aI3055
aI12585
aI3519
aI44857
aI11257
asS'd'
I9551
sa(dp19102
S'a'
I32592
sS's'
I17
sS'm'
(lp19103
I1
aI1
aI5
aI11
aI29
aI55
aI13
aI235
aI419
aI327
aI823
aI2675
aI8031
aI9303
aI8749
aI20215
aI12111
asS'd'
I9552
sa(dp19104
S'a'
I32602
sS's'
I17
sS'm'
(lp19105
I1
aI1
aI3
aI5
aI7
aI31
aI103
aI19
aI467
aI255
aI583
aI419
aI2845
aI12179
aI63
aI51693
aI9755
asS'd'
I9553
sa(dp19106
S'a'
I32604
sS's'
I17
sS'm'
(lp19107
I1
aI1
aI1
aI13
aI15
aI29
aI109
aI81
aI381
aI659
aI601
aI3867
aI7663
aI7307
aI16445
aI56327
aI48559
asS'd'
I9554
sa(dp19108
S'a'
I32613
sS's'
I17
sS'm'
(lp19109
I1
aI3
aI3
aI15
aI31
aI35
aI29
aI153
aI423
aI247
aI55
aI3259
aI6199
aI4199
aI13931
aI14433
aI52645
asS'd'
I9555
sa(dp19110
S'a'
I32625
sS's'
I17
sS'm'
(lp19111
I1
aI1
aI5
aI11
aI9
aI17
aI17
aI191
aI231
aI977
aI721
aI2817
aI2485
aI4965
aI32341
aI55131
aI4547
asS'd'
I9556
sa(dp19112
S'a'
I32631
sS's'
I17
sS'm'
(lp19113
I1
aI1
aI7
aI7
aI7
aI7
aI89
aI69
aI299
aI503
aI597
aI311
aI1321
aI2335
aI30193
aI45347
aI126631
asS'd'
I9557
sa(dp19114
S'a'
I32641
sS's'
I17
sS'm'
(lp19115
I1
aI1
aI7
aI11
aI13
aI43
aI105
aI153
aI89
aI229
aI1573
aI1549
aI3699
aI15981
aI28911
aI45011
aI83759
asS'd'
I9558
sa(dp19116
S'a'
I32642
sS's'
I17
sS'm'
(lp19117
I1
aI3
aI7
aI3
aI1
aI3
aI121
aI137
aI263
aI325
aI1449
aI3793
aI5795
aI7715
aI7449
aI26453
aI85081
asS'd'
I9559
sa(dp19118
S'a'
I32644
sS's'
I17
sS'm'
(lp19119
I1
aI3
aI1
aI7
aI23
aI15
aI39
aI217
aI99
aI873
aI1641
aI1411
aI4627
aI283
aI20707
aI41795
aI62239
asS'd'
I9560
sa(dp19120
S'a'
I32656
sS's'
I17
sS'm'
(lp19121
I1
aI3
aI5
aI9
aI15
aI15
aI35
aI255
aI501
aI945
aI79
aI799
aI2361
aI4495
aI27825
aI27699
aI129335
asS'd'
I9561
sa(dp19122
S'a'
I32678
sS's'
I17
sS'm'
(lp19123
I1
aI3
aI1
aI7
aI9
aI19
aI89
aI31
aI65
aI905
aI1475
aI1353
aI7253
aI12825
aI20723
aI47757
aI12007
asS'd'
I9562
sa(dp19124
S'a'
I32681
sS's'
I17
sS'm'
(lp19125
I1
aI1
aI3
aI3
aI15
aI35
aI83
aI239
aI463
aI835
aI1249
aI2521
aI3429
aI14073
aI13569
aI6161
aI71309
asS'd'
I9563
sa(dp19126
S'a'
I32701
sS's'
I17
sS'm'
(lp19127
I1
aI1
aI7
aI11
aI31
aI43
aI15
aI57
aI461
aI917
aI339
aI3787
aI2925
aI1879
aI7217
aI17091
aI108819
asS'd'
I9564
sa(dp19128
S'a'
I32713
sS's'
I17
sS'm'
(lp19129
I1
aI3
aI7
aI3
aI17
aI51
aI29
aI105
aI221
aI941
aI1291
aI835
aI1563
aI15623
aI2953
aI62985
aI63037
asS'd'
I9565
sa(dp19130
S'a'
I32721
sS's'
I17
sS'm'
(lp19131
I1
aI1
aI7
aI3
aI1
aI39
aI83
aI41
aI399
aI465
aI587
aI2011
aI137
aI6017
aI5067
aI52389
aI71053
asS'd'
I9566
sa(dp19132
S'a'
I32727
sS's'
I17
sS'm'
(lp19133
I1
aI1
aI7
aI11
aI17
aI55
aI103
aI239
aI173
aI181
aI1219
aI2671
aI5183
aI3799
aI19589
aI31247
aI68889
asS'd'
I9567
sa(dp19134
S'a'
I32731
sS's'
I17
sS'm'
(lp19135
I1
aI1
aI3
aI3
aI21
aI43
aI123
aI253
aI281
aI627
aI353
aI3077
aI1685
aI12143
aI19723
aI57775
aI70761
asS'd'
I9568
sa(dp19136
S'a'
I32734
sS's'
I17
sS'm'
(lp19137
I1
aI1
aI7
aI15
aI31
aI13
aI101
aI159
aI311
aI305
aI1783
aI3523
aI149
aI9269
aI7103
aI40315
aI30569
asS'd'
I9569
sa(dp19138
S'a'
I32740
sS's'
I17
sS'm'
(lp19139
I1
aI1
aI5
aI3
aI29
aI47
aI11
aI219
aI301
aI207
aI1361
aI563
aI7831
aI14469
aI18983
aI54535
aI64647
asS'd'
I9570
sa(dp19140
S'a'
I32773
sS's'
I17
sS'm'
(lp19141
I1
aI1
aI3
aI15
aI11
aI37
aI85
aI237
aI225
aI1009
aI1065
aI985
aI6849
aI5395
aI22853
aI43965
aI51363
asS'd'
I9571
sa(dp19142
S'a'
I32774
sS's'
I17
sS'm'
(lp19143
I1
aI3
aI3
aI1
aI11
aI61
aI45
aI131
aI201
aI609
aI757
aI2539
aI3817
aI9309
aI24759
aI26789
aI41437
asS'd'
I9572
sa(dp19144
S'a'
I32785
sS's'
I17
sS'm'
(lp19145
I1
aI1
aI7
aI3
aI21
aI5
aI19
aI137
aI75
aI573
aI583
aI2499
aI41
aI3429
aI24273
aI36711
aI110015
asS'd'
I9573
sa(dp19146
S'a'
I32788
sS's'
I17
sS'm'
(lp19147
I1
aI3
aI7
aI9
aI1
aI51
aI39
aI75
aI115
aI269
aI1983
aI2709
aI6989
aI6521
aI5551
aI43675
aI1019
asS'd'
I9574
sa(dp19148
S'a'
I32792
sS's'
I17
sS'm'
(lp19149
I1
aI1
aI3
aI9
aI27
aI1
aI125
aI7
aI67
aI821
aI275
aI1253
aI4635
aI3557
aI4155
aI13831
aI1523
asS'd'
I9575
sa(dp19150
S'a'
I32797
sS's'
I17
sS'm'
(lp19151
I1
aI1
aI5
aI15
aI23
aI15
aI79
aI43
aI275
aI791
aI1867
aI2495
aI2933
aI2167
aI22819
aI52913
aI88871
asS'd'
I9576
sa(dp19152
S'a'
I32801
sS's'
I17
sS'm'
(lp19153
I1
aI1
aI1
aI5
aI31
aI59
aI27
aI153
aI159
aI919
aI219
aI3373
aI3227
aI6321
aI27559
aI33905
aI126145
asS'd'
I9577
sa(dp19154
S'a'
I32811
sS's'
I17
sS'm'
(lp19155
I1
aI3
aI3
aI13
aI23
aI21
aI119
aI175
aI119
aI741
aI1745
aI3985
aI3847
aI5163
aI13699
aI32373
aI75201
asS'd'
I9578
sa(dp19156
S'a'
I32821
sS's'
I17
sS'm'
(lp19157
I1
aI3
aI7
aI15
aI1
aI47
aI101
aI89
aI425
aI269
aI713
aI3587
aI3373
aI13315
aI16481
aI40031
aI50353
asS'd'
I9579
sa(dp19158
S'a'
I32828
sS's'
I17
sS'm'
(lp19159
I1
aI3
aI7
aI3
aI19
aI29
aI5
aI69
aI385
aI979
aI1893
aI1849
aI8007
aI14415
aI18343
aI60555
aI109117
asS'd'
I9580
sa(dp19160
S'a'
I32839
sS's'
I17
sS'm'
(lp19161
I1
aI1
aI3
aI13
aI5
aI35
aI111
aI239
aI489
aI395
aI1565
aI1607
aI543
aI89
aI8971
aI22311
aI899
asS'd'
I9581
sa(dp19162
S'a'
I32854
sS's'
I17
sS'm'
(lp19163
I1
aI1
aI7
aI7
aI11
aI51
aI105
aI211
aI341
aI85
aI991
aI1275
aI3995
aI12611
aI2363
aI29501
aI44217
asS'd'
I9582
sa(dp19164
S'a'
I32867
sS's'
I17
sS'm'
(lp19165
I1
aI1
aI5
aI13
aI9
aI17
aI93
aI69
aI145
aI917
aI469
aI1109
aI7405
aI12903
aI8341
aI50383
aI20133
asS'd'
I9583
sa(dp19166
S'a'
I32870
sS's'
I17
sS'm'
(lp19167
I1
aI3
aI1
aI7
aI27
aI45
aI45
aI85
aI101
aI161
aI1117
aI2757
aI7847
aI359
aI17155
aI27073
aI123535
asS'd'
I9584
sa(dp19168
S'a'
I32873
sS's'
I17
sS'm'
(lp19169
I1
aI3
aI1
aI3
aI9
aI11
aI67
aI205
aI109
aI257
aI1635
aI141
aI3969
aI11571
aI211
aI48683
aI108671
asS'd'
I9585
sa(dp19170
S'a'
I32881
sS's'
I17
sS'm'
(lp19171
I1
aI1
aI3
aI7
aI13
aI9
aI29
aI251
aI113
aI851
aI1549
aI981
aI5553
aI6095
aI28885
aI32953
aI112563
asS'd'
I9586
sa(dp19172
S'a'
I32891
sS's'
I17
sS'm'
(lp19173
I1
aI1
aI5
aI7
aI11
aI5
aI13
aI83
aI343
aI499
aI587
aI3887
aI3859
aI11459
aI7361
aI25665
aI86151
asS'd'
I9587
sa(dp19174
S'a'
I32900
sS's'
I17
sS'm'
(lp19175
I1
aI1
aI5
aI1
aI13
aI43
aI3
aI37
aI273
aI749
aI1707
aI2069
aI3083
aI1095
aI3081
aI23919
aI21939
asS'd'
I9588
sa(dp19176
S'a'
I32903
sS's'
I17
sS'm'
(lp19177
I1
aI3
aI5
aI13
aI13
aI49
aI115
aI99
aI357
aI95
aI699
aI2615
aI1911
aI12675
aI8607
aI12535
aI118651
asS'd'
I9589
sa(dp19178
S'a'
I32910
sS's'
I17
sS'm'
(lp19179
I1
aI1
aI7
aI7
aI29
aI43
aI17
aI131
aI271
aI895
aI1427
aI3659
aI1843
aI8247
aI1175
aI48239
aI54435
asS'd'
I9590
sa(dp19180
S'a'
I32917
sS's'
I17
sS'm'
(lp19181
I1
aI1
aI1
aI9
aI1
aI27
aI85
aI163
aI353
aI669
aI745
aI317
aI2505
aI7685
aI14831
aI31131
aI106687
asS'd'
I9591
sa(dp19182
S'a'
I32922
sS's'
I17
sS'm'
(lp19183
I1
aI1
aI7
aI9
aI1
aI23
aI121
aI53
aI289
aI651
aI303
aI3049
aI6819
aI6733
aI17485
aI20023
aI110009
asS'd'
I9592
sa(dp19184
S'a'
I32928
sS's'
I17
sS'm'
(lp19185
I1
aI3
aI7
aI3
aI5
aI47
aI93
aI75
aI363
aI479
aI825
aI1801
aI6807
aI3341
aI6419
aI9889
aI5557
asS'd'
I9593
sa(dp19186
S'a'
I32945
sS's'
I17
sS'm'
(lp19187
I1
aI1
aI3
aI15
aI23
aI5
aI7
aI25
aI73
aI811
aI1597
aI2041
aI6707
aI6817
aI20427
aI50749
aI46255
asS'd'
I9594
sa(dp19188
S'a'
I32946
sS's'
I17
sS'm'
(lp19189
I1
aI3
aI7
aI9
aI1
aI11
aI61
aI63
aI435
aI977
aI1937
aI93
aI2685
aI643
aI20113
aI25873
aI63829
asS'd'
I9595
sa(dp19190
S'a'
I32951
sS's'
I17
sS'm'
(lp19191
I1
aI1
aI3
aI15
aI5
aI41
aI31
aI53
aI143
aI271
aI27
aI3899
aI5045
aI1063
aI17229
aI52715
aI67689
asS'd'
I9596
sa(dp19192
S'a'
I32958
sS's'
I17
sS'm'
(lp19193
I1
aI1
aI3
aI11
aI1
aI57
aI121
aI13
aI291
aI861
aI1547
aI3899
aI7949
aI15401
aI29807
aI52307
aI104359
asS'd'
I9597
sa(dp19194
S'a'
I32965
sS's'
I17
sS'm'
(lp19195
I1
aI3
aI5
aI15
aI23
aI3
aI95
aI43
aI377
aI437
aI1687
aI3075
aI5131
aI11791
aI3637
aI12621
aI105575
asS'd'
I9598
sa(dp19196
S'a'
I32978
sS's'
I17
sS'm'
(lp19197
I1
aI3
aI1
aI3
aI27
aI1
aI117
aI11
aI153
aI401
aI1971
aI2097
aI3227
aI14603
aI4757
aI56281
aI112263
asS'd'
I9599
sa(dp19198
S'a'
I32980
sS's'
I17
sS'm'
(lp19199
I1
aI3
aI3
aI5
aI13
aI25
aI51
aI209
aI367
aI327
aI1941
aI1943
aI1347
aI14393
aI31997
aI16001
aI129047
asS'd'
I9600
sa(dp19200
S'a'
I32983
sS's'
I17
sS'm'
(lp19201
I1
aI1
aI5
aI11
aI19
aI51
aI109
aI229
aI71
aI923
aI1741
aI1193
aI4657
aI6043
aI26703
aI17757
aI75009
asS'd'
I9601
sa(dp19202
S'a'
I32987
sS's'
I17
sS'm'
(lp19203
I1
aI1
aI7
aI3
aI23
aI3
aI125
aI165
aI137
aI999
aI1583
aI3493
aI859
aI15603
aI7143
aI28791
aI28201
asS'd'
I9602
sa(dp19204
S'a'
I33023
sS's'
I17
sS'm'
(lp19205
I1
aI1
aI5
aI11
aI29
aI57
aI65
aI41
aI295
aI729
aI635
aI1871
aI6347
aI3509
aI59
aI40765
aI42673
asS'd'
I9603
sa(dp19206
S'a'
I33031
sS's'
I17
sS'm'
(lp19207
I1
aI3
aI3
aI3
aI15
aI59
aI53
aI97
aI15
aI131
aI891
aI1105
aI841
aI6065
aI14427
aI4721
aI106433
asS'd'
I9604
sa(dp19208
S'a'
I33032
sS's'
I17
sS'm'
(lp19209
I1
aI1
aI1
aI7
aI19
aI37
aI101
aI121
aI141
aI613
aI1363
aI691
aI1731
aI12477
aI8339
aI55669
aI99379
asS'd'
I9605
sa(dp19210
S'a'
I33035
sS's'
I17
sS'm'
(lp19211
I1
aI3
aI5
aI13
aI17
aI49
aI75
aI25
aI447
aI113
aI1853
aI3465
aI5225
aI4531
aI14287
aI1039
aI17399
asS'd'
I9606
sa(dp19212
S'a'
I33038
sS's'
I17
sS'm'
(lp19213
I1
aI3
aI5
aI3
aI3
aI49
aI101
aI79
aI117
aI939
aI1161
aI1991
aI2343
aI7183
aI12599
aI52877
aI94337
asS'd'
I9607
sa(dp19214
S'a'
I33040
sS's'
I17
sS'm'
(lp19215
I1
aI3
aI1
aI1
aI19
aI47
aI73
aI195
aI475
aI435
aI1807
aI2723
aI7885
aI15469
aI26057
aI37325
aI57005
asS'd'
I9608
sa(dp19216
S'a'
I33043
sS's'
I17
sS'm'
(lp19217
I1
aI1
aI1
aI11
aI17
aI7
aI111
aI143
aI357
aI977
aI719
aI553
aI4559
aI7225
aI10405
aI26895
aI8385
asS'd'
I9609
sa(dp19218
S'a'
I33050
sS's'
I17
sS'm'
(lp19219
I1
aI3
aI3
aI9
aI17
aI5
aI1
aI73
aI125
aI913
aI1275
aI2387
aI5153
aI13611
aI20585
aI8465
aI27545
asS'd'
I9610
sa(dp19220
S'a'
I33059
sS's'
I17
sS'm'
(lp19221
I1
aI1
aI7
aI5
aI27
aI51
aI107
aI147
aI503
aI699
aI851
aI1729
aI2875
aI16331
aI28025
aI26451
aI92705
asS'd'
I9611
sa(dp19222
S'a'
I33080
sS's'
I17
sS'm'
(lp19223
I1
aI1
aI5
aI9
aI3
aI37
aI21
aI139
aI13
aI427
aI225
aI1345
aI2491
aI15495
aI25847
aI3095
aI128879
asS'd'
I9612
sa(dp19224
S'a'
I33098
sS's'
I17
sS'm'
(lp19225
I1
aI1
aI3
aI11
aI7
aI47
aI113
aI133
aI99
aI871
aI1151
aI1953
aI7931
aI6389
aI28715
aI36861
aI60017
asS'd'
I9613
sa(dp19226
S'a'
I33108
sS's'
I17
sS'm'
(lp19227
I1
aI1
aI7
aI1
aI21
aI47
aI35
aI83
aI137
aI945
aI2047
aI3491
aI3719
aI3001
aI20563
aI51243
aI14491
asS'd'
I9614
sa(dp19228
S'a'
I33115
sS's'
I17
sS'm'
(lp19229
I1
aI1
aI5
aI15
aI1
aI13
aI85
aI61
aI479
aI853
aI813
aI805
aI4931
aI12651
aI22757
aI29531
aI92861
asS'd'
I9615
sa(dp19230
S'a'
I33117
sS's'
I17
sS'm'
(lp19231
I1
aI3
aI7
aI7
aI27
aI63
aI31
aI169
aI43
aI185
aI637
aI729
aI7231
aI2381
aI23539
aI53885
aI90215
asS'd'
I9616
sa(dp19232
S'a'
I33133
sS's'
I17
sS'm'
(lp19233
I1
aI1
aI3
aI13
aI5
aI51
aI69
aI111
aI357
aI277
aI1889
aI3809
aI8031
aI13341
aI14261
aI34001
aI63317
asS'd'
I9617
sa(dp19234
S'a'
I33134
sS's'
I17
sS'm'
(lp19235
I1
aI1
aI7
aI3
aI11
aI59
aI1
aI43
aI227
aI503
aI1407
aI3917
aI7077
aI847
aI4513
aI53007
aI66721
asS'd'
I9618
sa(dp19236
S'a'
I33157
sS's'
I17
sS'm'
(lp19237
I1
aI1
aI5
aI11
aI15
aI25
aI109
aI169
aI25
aI391
aI597
aI2997
aI2377
aI9045
aI15239
aI25291
aI5451
asS'd'
I9619
sa(dp19238
S'a'
I33169
sS's'
I17
sS'm'
(lp19239
I1
aI3
aI3
aI11
aI15
aI11
aI1
aI59
aI347
aI707
aI239
aI2473
aI8057
aI4787
aI32247
aI17955
aI79151
asS'd'
I9620
sa(dp19240
S'a'
I33170
sS's'
I17
sS'm'
(lp19241
I1
aI3
aI7
aI11
aI9
aI59
aI9
aI117
aI137
aI713
aI451
aI1105
aI4485
aI14979
aI26271
aI46017
aI89211
asS'd'
I9621
sa(dp19242
S'a'
I33176
sS's'
I17
sS'm'
(lp19243
I1
aI3
aI3
aI3
aI3
aI19
aI95
aI131
aI413
aI291
aI1179
aI3265
aI7107
aI10419
aI13527
aI19905
aI8059
asS'd'
I9622
sa(dp19244
S'a'
I33182
sS's'
I17
sS'm'
(lp19245
I1
aI3
aI7
aI9
aI29
aI43
aI19
aI243
aI443
aI27
aI1401
aI3469
aI6925
aI2833
aI19715
aI39667
aI11983
asS'd'
I9623
sa(dp19246
S'a'
I33192
sS's'
I17
sS'm'
(lp19247
I1
aI3
aI3
aI7
aI23
aI33
aI115
aI59
aI29
aI61
aI1085
aI1115
aI4007
aI12673
aI26479
aI22397
aI95609
asS'd'
I9624
sa(dp19248
S'a'
I33205
sS's'
I17
sS'm'
(lp19249
I1
aI3
aI3
aI5
aI1
aI47
aI43
aI83
aI21
aI621
aI59
aI1
aI891
aI12285
aI31855
aI48641
aI52479
asS'd'
I9625
sa(dp19250
S'a'
I33212
sS's'
I17
sS'm'
(lp19251
I1
aI3
aI3
aI5
aI3
aI9
aI17
aI181
aI15
aI315
aI1705
aI2461
aI1853
aI14007
aI17665
aI40593
aI126179
asS'd'
I9626
sa(dp19252
S'a'
I33215
sS's'
I17
sS'm'
(lp19253
I1
aI3
aI5
aI3
aI3
aI23
aI83
aI163
aI29
aI293
aI1891
aI2631
aI2989
aI7295
aI2441
aI21689
aI8187
asS'd'
I9627
sa(dp19254
S'a'
I33217
sS's'
I17
sS'm'
(lp19255
I1
aI3
aI1
aI1
aI1
aI23
aI53
aI215
aI185
aI843
aI1083
aI2603
aI3857
aI4981
aI25079
aI20249
aI93717
asS'd'
I9628
sa(dp19256
S'a'
I33227
sS's'
I17
sS'm'
(lp19257
I1
aI3
aI5
aI11
aI7
aI61
aI127
aI13
aI449
aI395
aI1909
aI3967
aI2441
aI3073
aI8159
aI33979
aI26345
asS'd'
I9629
sa(dp19258
S'a'
I33229
sS's'
I17
sS'm'
(lp19259
I1
aI1
aI5
aI1
aI15
aI5
aI93
aI87
aI319
aI173
aI1729
aI1395
aI1019
aI5139
aI10819
aI29877
aI81025
asS'd'
I9630
sa(dp19260
S'a'
I33238
sS's'
I17
sS'm'
(lp19261
I1
aI3
aI3
aI7
aI17
aI55
aI61
aI227
aI299
aI245
aI849
aI211
aI895
aI2999
aI18215
aI37069
aI32821
asS'd'
I9631
sa(dp19262
S'a'
I33241
sS's'
I17
sS'm'
(lp19263
I1
aI1
aI5
aI3
aI17
aI49
aI115
aI55
aI447
aI533
aI1463
aI2983
aI3245
aI9345
aI11955
aI49145
aI128035
asS'd'
I9632
sa(dp19264
S'a'
I33260
sS's'
I17
sS'm'
(lp19265
I1
aI3
aI1
aI7
aI5
aI17
aI61
aI71
aI101
aI529
aI1761
aI827
aI7887
aI5713
aI31039
aI18087
aI82277
asS'd'
I9633
sa(dp19266
S'a'
I33271
sS's'
I17
sS'm'
(lp19267
I1
aI3
aI1
aI11
aI27
aI59
aI1
aI231
aI303
aI431
aI1279
aI3647
aI1333
aI3675
aI29401
aI55533
aI65997
asS'd'
I9634
sa(dp19268
S'a'
I33278
sS's'
I17
sS'm'
(lp19269
I1
aI1
aI5
aI9
aI7
aI9
aI111
aI245
aI269
aI919
aI1147
aI1601
aI6219
aI4931
aI3035
aI12231
aI4011
asS'd'
I9635
sa(dp19270
S'a'
I33293
sS's'
I17
sS'm'
(lp19271
I1
aI3
aI5
aI15
aI3
aI19
aI83
aI25
aI129
aI979
aI79
aI3027
aI3983
aI7703
aI16859
aI12085
aI83115
asS'd'
I9636
sa(dp19272
S'a'
I33294
sS's'
I17
sS'm'
(lp19273
I1
aI1
aI5
aI11
aI31
aI41
aI99
aI3
aI383
aI943
aI1579
aI2435
aI1209
aI161
aI31733
aI11755
aI95697
asS'd'
I9637
sa(dp19274
S'a'
I33296
sS's'
I17
sS'm'
(lp19275
I1
aI1
aI1
aI9
aI9
aI55
aI115
aI187
aI499
aI165
aI1081
aI813
aI2545
aI8065
aI10501
aI15475
aI85107
asS'd'
I9638
sa(dp19276
S'a'
I33302
sS's'
I17
sS'm'
(lp19277
I1
aI1
aI1
aI3
aI1
aI31
aI81
aI213
aI301
aI575
aI605
aI543
aI3347
aI12759
aI21645
aI37173
aI36127
asS'd'
I9639
sa(dp19278
S'a'
I33305
sS's'
I17
sS'm'
(lp19279
I1
aI3
aI3
aI9
aI21
aI29
aI51
aI91
aI307
aI617
aI1839
aI443
aI1013
aI4473
aI3885
aI57669
aI123271
asS'd'
I9640
sa(dp19280
S'a'
I33329
sS's'
I17
sS'm'
(lp19281
I1
aI3
aI1
aI15
aI31
aI43
aI83
aI187
aI51
aI513
aI1505
aI3895
aI3557
aI9527
aI27537
aI6173
aI99595
asS'd'
I9641
sa(dp19282
S'a'
I33330
sS's'
I17
sS'm'
(lp19283
I1
aI3
aI3
aI1
aI3
aI53
aI113
aI27
aI431
aI505
aI219
aI2143
aI6691
aI3219
aI9589
aI9885
aI24037
asS'd'
I9642
sa(dp19284
S'a'
I33332
sS's'
I17
sS'm'
(lp19285
I1
aI1
aI5
aI9
aI13
aI3
aI53
aI145
aI49
aI411
aI691
aI289
aI6443
aI4963
aI13815
aI23663
aI95497
asS'd'
I9643
sa(dp19286
S'a'
I33354
sS's'
I17
sS'm'
(lp19287
I1
aI3
aI5
aI9
aI19
aI7
aI53
aI101
aI199
aI69
aI1821
aI3233
aI3267
aI5947
aI4869
aI30095
aI21255
asS'd'
I9644
sa(dp19288
S'a'
I33383
sS's'
I17
sS'm'
(lp19289
I1
aI1
aI5
aI11
aI29
aI7
aI79
aI11
aI451
aI585
aI987
aI2333
aI1891
aI1853
aI14739
aI34399
aI62895
asS'd'
I9645
sa(dp19290
S'a'
I33387
sS's'
I17
sS'm'
(lp19291
I1
aI3
aI1
aI7
aI29
aI43
aI103
aI219
aI139
aI359
aI1663
aI3453
aI7469
aI1943
aI11457
aI19227
aI62211
asS'd'
I9646
sa(dp19292
S'a'
I33397
sS's'
I17
sS'm'
(lp19293
I1
aI3
aI3
aI11
aI9
aI47
aI17
aI237
aI87
aI881
aI583
aI3473
aI2579
aI975
aI1531
aI50997
aI76219
asS'd'
I9647
sa(dp19294
S'a'
I33408
sS's'
I17
sS'm'
(lp19295
I1
aI1
aI7
aI15
aI31
aI37
aI79
aI115
aI95
aI515
aI2003
aI2595
aI4077
aI4537
aI9171
aI31183
aI41219
asS'd'
I9648
sa(dp19296
S'a'
I33417
sS's'
I17
sS'm'
(lp19297
I1
aI1
aI1
aI9
aI21
aI41
aI93
aI33
aI211
aI341
aI233
aI2217
aI6657
aI12913
aI8329
aI3881
aI42563
asS'd'
I9649
sa(dp19298
S'a'
I33420
sS's'
I17
sS'm'
(lp19299
I1
aI3
aI3
aI11
aI25
aI3
aI23
aI197
aI49
aI339
aI877
aI1117
aI7817
aI14143
aI1575
aI50301
aI92367
asS'd'
I9650
sa(dp19300
S'a'
I33423
sS's'
I17
sS'm'
(lp19301
I1
aI3
aI5
aI5
aI19
aI45
aI69
aI179
aI447
aI861
aI1633
aI1941
aI5821
aI1843
aI4085
aI23501
aI109047
asS'd'
I9651
sa(dp19302
S'a'
I33431
sS's'
I17
sS'm'
(lp19303
I1
aI3
aI1
aI3
aI31
aI29
aI49
aI183
aI311
aI133
aI345
aI1541
aI111
aI5571
aI1943
aI11039
aI127673
asS'd'
I9652
sa(dp19304
S'a'
I33438
sS's'
I17
sS'm'
(lp19305
I1
aI3
aI1
aI5
aI3
aI13
aI63
aI5
aI59
aI789
aI71
aI3271
aI3871
aI9105
aI22525
aI31
aI117803
asS'd'
I9653
sa(dp19306
S'a'
I33442
sS's'
I17
sS'm'
(lp19307
I1
aI3
aI1
aI13
aI31
aI43
aI97
aI133
aI313
aI729
aI287
aI2971
aI5623
aI13183
aI15179
aI47271
aI28853
asS'd'
I9654
sa(dp19308
S'a'
I33444
sS's'
I17
sS'm'
(lp19309
I1
aI1
aI3
aI13
aI27
aI15
aI35
aI37
aI507
aI139
aI1933
aI2847
aI361
aI10261
aI21031
aI3889
aI56875
asS'd'
I9655
sa(dp19310
S'a'
I33448
sS's'
I17
sS'm'
(lp19311
I1
aI3
aI1
aI15
aI31
aI13
aI45
aI73
aI279
aI331
aI471
aI3881
aI3295
aI12035
aI28329
aI899
aI47397
asS'd'
I9656
sa(dp19312
S'a'
I33456
sS's'
I17
sS'm'
(lp19313
I1
aI1
aI3
aI13
aI1
aI7
aI81
aI255
aI315
aI595
aI43
aI3919
aI5229
aI7953
aI25711
aI19509
aI107181
asS'd'
I9657
sa(dp19314
S'a'
I33459
sS's'
I17
sS'm'
(lp19315
I1
aI1
aI3
aI15
aI7
aI33
aI117
aI169
aI71
aI577
aI629
aI3665
aI7761
aI13529
aI26375
aI17181
aI22125
asS'd'
I9658
sa(dp19316
S'a'
I33466
sS's'
I17
sS'm'
(lp19317
I1
aI3
aI5
aI7
aI5
aI7
aI1
aI93
aI489
aI289
aI329
aI2273
aI685
aI14835
aI11433
aI26041
aI112735
asS'd'
I9659
sa(dp19318
S'a'
I33473
sS's'
I17
sS'm'
(lp19319
I1
aI3
aI3
aI3
aI9
aI39
aI45
aI23
aI171
aI35
aI571
aI551
aI7815
aI6169
aI24283
aI61477
aI71877
asS'd'
I9660
sa(dp19320
S'a'
I33476
sS's'
I17
sS'm'
(lp19321
I1
aI1
aI5
aI7
aI23
aI15
aI81
aI215
aI297
aI269
aI655
aI2059
aI3643
aI12741
aI11955
aI41085
aI46047
asS'd'
I9661
sa(dp19322
S'a'
I33491
sS's'
I17
sS'm'
(lp19323
I1
aI1
aI7
aI5
aI3
aI35
aI125
aI141
aI419
aI137
aI1031
aI2053
aI7925
aI7267
aI6267
aI34323
aI77495
asS'd'
I9662
sa(dp19324
S'a'
I33494
sS's'
I17
sS'm'
(lp19325
I1
aI1
aI7
aI11
aI3
aI57
aI91
aI43
aI139
aI691
aI1569
aI1825
aI7855
aI1093
aI19263
aI31601
aI16019
asS'd'
I9663
sa(dp19326
S'a'
I33507
sS's'
I17
sS'm'
(lp19327
I1
aI3
aI1
aI5
aI21
aI7
aI11
aI225
aI105
aI757
aI1493
aI455
aI4757
aI12007
aI5139
aI3545
aI79717
asS'd'
I9664
sa(dp19328
S'a'
I33514
sS's'
I17
sS'm'
(lp19329
I1
aI3
aI1
aI13
aI17
aI29
aI125
aI249
aI475
aI79
aI1271
aI341
aI863
aI853
aI2105
aI32897
aI121261
asS'd'
I9665
sa(dp19330
S'a'
I33521
sS's'
I17
sS'm'
(lp19331
I1
aI3
aI1
aI11
aI17
aI59
aI3
aI29
aI61
aI399
aI1465
aI4029
aI2103
aI12481
aI28495
aI34363
aI63781
asS'd'
I9666
sa(dp19332
S'a'
I33528
sS's'
I17
sS'm'
(lp19333
I1
aI3
aI3
aI15
aI29
aI13
aI101
aI191
aI435
aI215
aI1355
aI2263
aI6059
aI4545
aI7535
aI15041
aI84091
asS'd'
I9667
sa(dp19334
S'a'
I33534
sS's'
I17
sS'm'
(lp19335
I1
aI1
aI3
aI9
aI29
aI23
aI99
aI55
aI91
aI145
aI235
aI2847
aI725
aI209
aI24565
aI16545
aI103669
asS'd'
I9668
sa(dp19336
S'a'
I33536
sS's'
I17
sS'm'
(lp19337
I1
aI1
aI1
aI1
aI31
aI15
aI93
aI197
aI207
aI357
aI667
aI3511
aI3865
aI5329
aI6491
aI9027
aI125979
asS'd'
I9669
sa(dp19338
S'a'
I33551
sS's'
I17
sS'm'
(lp19339
I1
aI3
aI3
aI13
aI17
aI35
aI99
aI187
aI153
aI589
aI1633
aI4053
aI1023
aI9541
aI9841
aI39585
aI24885
asS'd'
I9670
sa(dp19340
S'a'
I33554
sS's'
I17
sS'm'
(lp19341
I1
aI3
aI7
aI11
aI23
aI5
aI71
aI89
aI455
aI665
aI1221
aI1821
aI591
aI11459
aI503
aI56777
aI65691
asS'd'
I9671
sa(dp19342
S'a'
I33563
sS's'
I17
sS'm'
(lp19343
I1
aI3
aI1
aI1
aI9
aI33
aI51
aI203
aI223
aI709
aI1263
aI3535
aI7753
aI8279
aI8673
aI60259
aI2671
asS'd'
I9672
sa(dp19344
S'a'
I33575
sS's'
I17
sS'm'
(lp19345
I1
aI1
aI7
aI9
aI17
aI63
aI5
aI229
aI495
aI435
aI1711
aI3359
aI399
aI15901
aI28519
aI56627
aI8079
asS'd'
I9673
sa(dp19346
S'a'
I33579
sS's'
I17
sS'm'
(lp19347
I1
aI3
aI5
aI11
aI9
aI25
aI49
aI143
aI275
aI989
aI461
aI447
aI1917
aI9253
aI28421
aI1803
aI119725
asS'd'
I9674
sa(dp19348
S'a'
I33582
sS's'
I17
sS'm'
(lp19349
I1
aI3
aI3
aI7
aI25
aI3
aI39
aI171
aI303
aI905
aI1353
aI2561
aI7347
aI7339
aI15271
aI61945
aI26343
asS'd'
I9675
sa(dp19350
S'a'
I33601
sS's'
I17
sS'm'
(lp19351
I1
aI1
aI1
aI3
aI5
aI63
aI9
aI229
aI107
aI815
aI1705
aI3621
aI2345
aI3065
aI16315
aI17017
aI33667
asS'd'
I9676
sa(dp19352
S'a'
I33602
sS's'
I17
sS'm'
(lp19353
I1
aI3
aI5
aI13
aI29
aI13
aI91
aI111
aI475
aI561
aI443
aI3825
aI5331
aI11211
aI27639
aI28305
aI101831
asS'd'
I9677
sa(dp19354
S'a'
I33614
sS's'
I17
sS'm'
(lp19355
I1
aI3
aI1
aI9
aI15
aI33
aI17
aI47
aI249
aI89
aI429
aI3819
aI1959
aI14317
aI10737
aI28151
aI40395
asS'd'
I9678
sa(dp19356
S'a'
I33625
sS's'
I17
sS'm'
(lp19357
I1
aI3
aI7
aI13
aI19
aI29
aI83
aI81
aI511
aI783
aI823
aI2865
aI5823
aI9459
aI27413
aI63297
aI44181
asS'd'
I9679
sa(dp19358
S'a'
I33628
sS's'
I17
sS'm'
(lp19359
I1
aI3
aI1
aI1
aI19
aI53
aI45
aI227
aI193
aI631
aI289
aI1227
aI6241
aI6915
aI16051
aI31237
aI50201
asS'd'
I9680
sa(dp19360
S'a'
I33637
sS's'
I17
sS'm'
(lp19361
I1
aI3
aI7
aI7
aI15
aI49
aI77
aI147
aI421
aI515
aI927
aI1561
aI4391
aI12943
aI6807
aI36889
aI70249
asS'd'
I9681
sa(dp19362
S'a'
I33656
sS's'
I17
sS'm'
(lp19363
I1
aI3
aI7
aI7
aI17
aI15
aI63
aI123
aI101
aI283
aI59
aI977
aI5185
aI16161
aI5007
aI36255
aI11537
asS'd'
I9682
sa(dp19364
S'a'
I33665
sS's'
I17
sS'm'
(lp19365
I1
aI1
aI7
aI1
aI13
aI17
aI79
aI35
aI193
aI947
aI767
aI1365
aI2145
aI13267
aI30561
aI51949
aI37591
asS'd'
I9683
sa(dp19366
S'a'
I33683
sS's'
I17
sS'm'
(lp19367
I1
aI1
aI1
aI13
aI11
aI13
aI91
aI129
aI355
aI549
aI295
aI673
aI209
aI15953
aI14703
aI30857
aI47967
asS'd'
I9684
sa(dp19368
S'a'
I33695
sS's'
I17
sS'm'
(lp19369
I1
aI3
aI5
aI9
aI17
aI17
aI83
aI161
aI189
aI585
aI21
aI1019
aI4879
aI15943
aI17281
aI46013
aI94839
asS'd'
I9685
sa(dp19370
S'a'
I33696
sS's'
I17
sS'm'
(lp19371
I1
aI3
aI5
aI9
aI23
aI39
aI65
aI25
aI181
aI3
aI2005
aI635
aI201
aI9391
aI8755
aI38535
aI88697
asS'd'
I9686
sa(dp19372
S'a'
I33702
sS's'
I17
sS'm'
(lp19373
I1
aI3
aI1
aI15
aI13
aI35
aI47
aI125
aI429
aI901
aI895
aI3495
aI327
aI397
aI7847
aI62157
aI3489
asS'd'
I9687
sa(dp19374
S'a'
I33708
sS's'
I17
sS'm'
(lp19375
I1
aI3
aI5
aI3
aI19
aI21
aI81
aI39
aI85
aI169
aI1981
aI3323
aI113
aI2057
aI16617
aI58051
aI55059
asS'd'
I9688
sa(dp19376
S'a'
I33711
sS's'
I17
sS'm'
(lp19377
I1
aI3
aI1
aI13
aI9
aI1
aI101
aI81
aI129
aI717
aI1495
aI4077
aI5555
aI93
aI12957
aI14805
aI110219
asS'd'
I9689
sa(dp19378
S'a'
I33716
sS's'
I17
sS'm'
(lp19379
I1
aI3
aI5
aI5
aI5
aI47
aI107
aI111
aI387
aI987
aI2009
aI179
aI1111
aI3443
aI25579
aI12293
aI123035
asS'd'
I9690
sa(dp19380
S'a'
I33728
sS's'
I17
sS'm'
(lp19381
I1
aI1
aI7
aI13
aI21
aI25
aI33
aI211
aI9
aI783
aI1785
aI2691
aI6835
aI2867
aI22469
aI17853
aI90685
asS'd'
I9691
sa(dp19382
S'a'
I33737
sS's'
I17
sS'm'
(lp19383
I1
aI1
aI3
aI3
aI19
aI57
aI59
aI203
aI197
aI347
aI553
aI1361
aI7593
aI91
aI15303
aI30045
aI86605
asS'd'
I9692
sa(dp19384
S'a'
I33761
sS's'
I17
sS'm'
(lp19385
I1
aI3
aI5
aI7
aI29
aI23
aI1
aI235
aI159
aI277
aI1227
aI1727
aI1853
aI9717
aI2377
aI13597
aI18119
asS'd'
I9693
sa(dp19386
S'a'
I33774
sS's'
I17
sS'm'
(lp19387
I1
aI1
aI1
aI11
aI15
aI29
aI5
aI15
aI349
aI685
aI197
aI3127
aI1075
aI8847
aI27873
aI539
aI57149
asS'd'
I9694
sa(dp19388
S'a'
I33782
sS's'
I17
sS'm'
(lp19389
I1
aI1
aI7
aI9
aI23
aI25
aI121
aI239
aI219
aI747
aI1981
aI2683
aI5319
aI75
aI22569
aI29697
aI27627
asS'd'
I9695
sa(dp19390
S'a'
I33788
sS's'
I17
sS'm'
(lp19391
I1
aI3
aI7
aI5
aI31
aI43
aI95
aI131
aI423
aI547
aI1437
aI127
aI1953
aI861
aI839
aI54503
aI20465
asS'd'
I9696
sa(dp19392
S'a'
I33791
sS's'
I17
sS'm'
(lp19393
I1
aI1
aI5
aI3
aI29
aI29
aI71
aI237
aI275
aI493
aI513
aI4067
aI393
aI9415
aI20511
aI29257
aI86267
asS'd'
I9697
sa(dp19394
S'a'
I33793
sS's'
I17
sS'm'
(lp19395
I1
aI1
aI1
aI1
aI25
aI11
aI59
aI185
aI211
aI175
aI37
aI2999
aI4919
aI10225
aI16727
aI60447
aI59985
asS'd'
I9698
sa(dp19396
S'a'
I33811
sS's'
I17
sS'm'
(lp19397
I1
aI1
aI3
aI3
aI1
aI9
aI69
aI195
aI197
aI677
aI229
aI599
aI5613
aI4537
aI5495
aI58801
aI14297
asS'd'
I9699
sa(dp19398
S'a'
I33813
sS's'
I17
sS'm'
(lp19399
I1
aI3
aI1
aI15
aI17
aI23
aI5
aI101
aI331
aI943
aI1433
aI2199
aI313
aI469
aI3651
aI3281
aI100119
asS'd'
I9700
sa(dp19400
S'a'
I33818
sS's'
I17
sS'm'
(lp19401
I1
aI1
aI5
aI15
aI13
aI25
aI87
aI45
aI229
aI821
aI59
aI761
aI6259
aI15159
aI3197
aI39763
aI87301
asS'd'
I9701
sa(dp19402
S'a'
I33829
sS's'
I17
sS'm'
(lp19403
I1
aI3
aI5
aI7
aI19
aI21
aI89
aI15
aI19
aI623
aI603
aI4069
aI3531
aI13353
aI21267
aI6355
aI53821
asS'd'
I9702
sa(dp19404
S'a'
I33842
sS's'
I17
sS'm'
(lp19405
I1
aI1
aI5
aI9
aI13
aI13
aI111
aI77
aI439
aI599
aI1577
aI959
aI4567
aI3117
aI7127
aI49265
aI35667
asS'd'
I9703
sa(dp19406
S'a'
I33854
sS's'
I17
sS'm'
(lp19407
I1
aI3
aI7
aI9
aI27
aI61
aI1
aI19
aI43
aI475
aI221
aI655
aI4351
aI15827
aI30489
aI22245
aI41077
asS'd'
I9704
sa(dp19408
S'a'
I33856
sS's'
I17
sS'm'
(lp19409
I1
aI1
aI3
aI13
aI17
aI17
aI111
aI85
aI253
aI11
aI367
aI2349
aI4103
aI12517
aI27037
aI42481
aI84451
asS'd'
I9705
sa(dp19410
S'a'
I33868
sS's'
I17
sS'm'
(lp19411
I1
aI3
aI5
aI7
aI7
aI25
aI53
aI27
aI429
aI503
aI893
aI2923
aI2539
aI15849
aI30157
aI12111
aI108893
asS'd'
I9706
sa(dp19412
S'a'
I33879
sS's'
I17
sS'm'
(lp19413
I1
aI1
aI7
aI9
aI13
aI29
aI51
aI113
aI273
aI745
aI759
aI263
aI3031
aI705
aI23203
aI64245
aI127183
asS'd'
I9707
sa(dp19414
S'a'
I33885
sS's'
I17
sS'm'
(lp19415
I1
aI1
aI1
aI9
aI29
aI5
aI25
aI165
aI261
aI319
aI645
aI2199
aI3135
aI10263
aI10711
aI18713
aI63337
asS'd'
I9708
sa(dp19416
S'a'
I33886
sS's'
I17
sS'm'
(lp19417
I1
aI1
aI5
aI1
aI23
aI41
aI43
aI71
aI365
aI683
aI1107
aI1671
aI7079
aI8933
aI12815
aI8095
aI97955
asS'd'
I9709
sa(dp19418
S'a'
I33892
sS's'
I17
sS'm'
(lp19419
I1
aI3
aI1
aI15
aI9
aI43
aI105
aI217
aI131
aI299
aI1459
aI1087
aI3493
aI15297
aI11741
aI43383
aI35021
asS'd'
I9710
sa(dp19420
S'a'
I33907
sS's'
I17
sS'm'
(lp19421
I1
aI3
aI1
aI3
aI3
aI57
aI69
aI7
aI73
aI977
aI1163
aI3591
aI243
aI13129
aI23247
aI20609
aI22489
asS'd'
I9711
sa(dp19422
S'a'
I33913
sS's'
I17
sS'm'
(lp19423
I1
aI3
aI7
aI5
aI1
aI57
aI65
aI27
aI121
aI575
aI903
aI3527
aI5601
aI5597
aI1941
aI60079
aI88121
asS'd'
I9712
sa(dp19424
S'a'
I33923
sS's'
I17
sS'm'
(lp19425
I1
aI3
aI1
aI3
aI15
aI3
aI23
aI87
aI233
aI389
aI1671
aI1557
aI4825
aI1017
aI17697
aI26735
aI53421
asS'd'
I9713
sa(dp19426
S'a'
I33925
sS's'
I17
sS'm'
(lp19427
I1
aI3
aI5
aI3
aI5
aI43
aI61
aI249
aI273
aI251
aI1383
aI2415
aI1061
aI12363
aI3071
aI23785
aI127909
asS'd'
I9714
sa(dp19428
S'a'
I33935
sS's'
I17
sS'm'
(lp19429
I1
aI3
aI3
aI13
aI5
aI63
aI15
aI165
aI353
aI603
aI1627
aI2037
aI487
aI11603
aI719
aI54693
aI52645
asS'd'
I9715
sa(dp19430
S'a'
I33937
sS's'
I17
sS'm'
(lp19431
I1
aI3
aI5
aI11
aI31
aI41
aI41
aI83
aI481
aI251
aI1903
aI2655
aI5237
aI6073
aI20201
aI14069
aI91649
asS'd'
I9716
sa(dp19432
S'a'
I33954
sS's'
I17
sS'm'
(lp19433
I1
aI3
aI1
aI15
aI21
aI41
aI99
aI61
aI55
aI63
aI1595
aI1805
aI7625
aI12261
aI23275
aI43471
aI5147
asS'd'
I9717
sa(dp19434
S'a'
I33963
sS's'
I17
sS'm'
(lp19435
I1
aI3
aI1
aI5
aI23
aI21
aI71
aI169
aI197
aI51
aI1653
aI3053
aI4663
aI293
aI12751
aI15641
aI83993
asS'd'
I9718
sa(dp19436
S'a'
I33966
sS's'
I17
sS'm'
(lp19437
I1
aI3
aI5
aI15
aI29
aI45
aI55
aI199
aI275
aI103
aI1093
aI3569
aI5997
aI9445
aI2291
aI30973
aI68589
asS'd'
I9719
sa(dp19438
S'a'
I33977
sS's'
I17
sS'm'
(lp19439
I1
aI3
aI5
aI7
aI15
aI3
aI15
aI3
aI287
aI961
aI1759
aI1153
aI7613
aI9885
aI8981
aI5109
aI112865
asS'd'
I9720
sa(dp19440
S'a'
I33978
sS's'
I17
sS'm'
(lp19441
I1
aI1
aI1
aI9
aI1
aI37
aI111
aI61
aI309
aI581
aI875
aI2121
aI1035
aI4345
aI1351
aI59743
aI34955
asS'd'
I9721
sa(dp19442
S'a'
I33991
sS's'
I17
sS'm'
(lp19443
I1
aI3
aI7
aI7
aI11
aI23
aI51
aI235
aI23
aI697
aI991
aI1995
aI3615
aI6665
aI15885
aI18555
aI11711
asS'd'
I9722
sa(dp19444
S'a'
I33998
sS's'
I17
sS'm'
(lp19445
I1
aI3
aI7
aI13
aI3
aI59
aI87
aI129
aI405
aI689
aI1189
aI2071
aI877
aI12347
aI18381
aI28367
aI27247
asS'd'
I9723
sa(dp19446
S'a'
I34012
sS's'
I17
sS'm'
(lp19447
I1
aI1
aI1
aI9
aI23
aI29
aI113
aI71
aI479
aI421
aI215
aI1029
aI6125
aI13575
aI10823
aI45303
aI3153
asS'd'
I9724
sa(dp19448
S'a'
I34016
sS's'
I17
sS'm'
(lp19449
I1
aI1
aI3
aI11
aI13
aI5
aI31
aI29
aI279
aI597
aI791
aI319
aI1391
aI14487
aI3811
aI36913
aI11513
asS'd'
I9725
sa(dp19450
S'a'
I34025
sS's'
I17
sS'm'
(lp19451
I1
aI3
aI7
aI11
aI9
aI11
aI55
aI167
aI69
aI519
aI1887
aI145
aI6133
aI1307
aI14465
aI17419
aI18319
asS'd'
I9726
sa(dp19452
S'a'
I34033
sS's'
I17
sS'm'
(lp19453
I1
aI1
aI3
aI1
aI29
aI25
aI57
aI75
aI19
aI187
aI1591
aI421
aI959
aI7499
aI8377
aI42811
aI53423
asS'd'
I9727
sa(dp19454
S'a'
I34036
sS's'
I17
sS'm'
(lp19455
I1
aI3
aI1
aI3
aI7
aI9
aI73
aI217
aI383
aI755
aI1561
aI3923
aI3891
aI16129
aI13195
aI62097
aI67493
asS'd'
I9728
sa(dp19456
S'a'
I34045
sS's'
I17
sS'm'
(lp19457
I1
aI3
aI7
aI9
aI5
aI7
aI47
aI29
aI319
aI243
aI405
aI2867
aI5803
aI2273
aI4913
aI54777
aI88301
asS'd'
I9729
sa(dp19458
S'a'
I34065
sS's'
I17
sS'm'
(lp19459
I1
aI3
aI7
aI1
aI25
aI11
aI51
aI183
aI387
aI863
aI39
aI2119
aI2395
aI10175
aI20833
aI3235
aI108197
asS'd'
I9730
sa(dp19460
S'a'
I34078
sS's'
I17
sS'm'
(lp19461
I1
aI1
aI7
aI13
aI25
aI43
aI21
aI67
aI103
aI709
aI603
aI1045
aI7079
aI8867
aI29039
aI61499
aI39533
asS'd'
I9731
sa(dp19462
S'a'
I34093
sS's'
I17
sS'm'
(lp19463
I1
aI1
aI7
aI5
aI7
aI55
aI77
aI115
aI409
aI287
aI1149
aI1535
aI7459
aI5525
aI27129
aI43659
aI86953
asS'd'
I9732
sa(dp19464
S'a'
I34101
sS's'
I17
sS'm'
(lp19465
I1
aI3
aI5
aI3
aI21
aI41
aI47
aI147
aI267
aI473
aI1501
aI2663
aI5381
aI41
aI18265
aI53845
aI16039
asS'd'
I9733
sa(dp19466
S'a'
I34108
sS's'
I17
sS'm'
(lp19467
I1
aI1
aI7
aI15
aI27
aI63
aI95
aI103
aI169
aI1
aI133
aI3103
aI7539
aI5765
aI11453
aI4133
aI95133
asS'd'
I9734
sa(dp19468
S'a'
I34111
sS's'
I17
sS'm'
(lp19469
I1
aI3
aI3
aI15
aI3
aI53
aI121
aI135
aI385
aI475
aI889
aI2557
aI4937
aI11129
aI18461
aI16757
aI30339
asS'd'
I9735
sa(dp19470
S'a'
I34120
sS's'
I17
sS'm'
(lp19471
I1
aI3
aI1
aI13
aI11
aI39
aI111
aI13
aI475
aI201
aI1973
aI2151
aI6973
aI4083
aI12593
aI44093
aI108037
asS'd'
I9736
sa(dp19472
S'a'
I34123
sS's'
I17
sS'm'
(lp19473
I1
aI3
aI7
aI9
aI31
aI31
aI97
aI235
aI179
aI689
aI403
aI1995
aI7697
aI7511
aI29333
aI11005
aI50723
asS'd'
I9737
sa(dp19474
S'a'
I34125
sS's'
I17
sS'm'
(lp19475
I1
aI1
aI7
aI13
aI23
aI5
aI7
aI171
aI441
aI921
aI1455
aI3865
aI7089
aI5469
aI10423
aI53013
aI80559
asS'd'
I9738
sa(dp19476
S'a'
I34153
sS's'
I17
sS'm'
(lp19477
I1
aI3
aI5
aI3
aI25
aI43
aI105
aI157
aI507
aI143
aI297
aI1111
aI2761
aI14103
aI4965
aI36733
aI11741
asS'd'
I9739
sa(dp19478
S'a'
I34171
sS's'
I17
sS'm'
(lp19479
I1
aI3
aI7
aI9
aI29
aI61
aI49
aI239
aI271
aI697
aI211
aI1633
aI2991
aI14933
aI12347
aI44291
aI12219
asS'd'
I9740
sa(dp19480
S'a'
I34174
sS's'
I17
sS'm'
(lp19481
I1
aI1
aI7
aI7
aI17
aI61
aI29
aI43
aI87
aI633
aI937
aI1931
aI3541
aI12259
aI23045
aI5923
aI48479
asS'd'
I9741
sa(dp19482
S'a'
I34178
sS's'
I17
sS'm'
(lp19483
I1
aI3
aI3
aI3
aI15
aI25
aI105
aI17
aI159
aI863
aI1377
aI331
aI1475
aI10573
aI28947
aI8141
aI26671
asS'd'
I9742
sa(dp19484
S'a'
I34183
sS's'
I17
sS'm'
(lp19485
I1
aI1
aI7
aI7
aI31
aI59
aI81
aI23
aI467
aI241
aI1257
aI1337
aI7731
aI9071
aI3417
aI51191
aI78369
asS'd'
I9743
sa(dp19486
S'a'
I34190
sS's'
I17
sS'm'
(lp19487
I1
aI1
aI5
aI9
aI11
aI45
aI49
aI227
aI319
aI63
aI1339
aI885
aI4571
aI11649
aI5607
aI10509
aI55055
asS'd'
I9744
sa(dp19488
S'a'
I34201
sS's'
I17
sS'm'
(lp19489
I1
aI3
aI3
aI9
aI29
aI17
aI7
aI235
aI191
aI927
aI575
aI1115
aI4111
aI14179
aI2041
aI13331
aI29825
asS'd'
I9745
sa(dp19490
S'a'
I34211
sS's'
I17
sS'm'
(lp19491
I1
aI1
aI5
aI9
aI27
aI61
aI71
aI201
aI341
aI577
aI221
aI1371
aI1135
aI4347
aI24211
aI36171
aI23435
asS'd'
I9746
sa(dp19492
S'a'
I34220
sS's'
I17
sS'm'
(lp19493
I1
aI3
aI3
aI1
aI1
aI29
aI75
aI121
aI193
aI647
aI1429
aI275
aI5243
aI783
aI28533
aI13941
aI68035
asS'd'
I9747
sa(dp19494
S'a'
I34225
sS's'
I17
sS'm'
(lp19495
I1
aI3
aI5
aI15
aI21
aI27
aI117
aI183
aI251
aI991
aI935
aI3119
aI5133
aI2765
aI7423
aI28867
aI120565
asS'd'
I9748
sa(dp19496
S'a'
I34237
sS's'
I17
sS'm'
(lp19497
I1
aI3
aI5
aI5
aI13
aI23
aI29
aI101
aI299
aI699
aI1249
aI1225
aI1335
aI6079
aI17825
aI60467
aI87787
asS'd'
I9749
sa(dp19498
S'a'
I34249
sS's'
I17
sS'm'
(lp19499
I1
aI1
aI1
aI9
aI15
aI19
aI11
aI163
aI433
aI553
aI1487
aI813
aI3293
aI1195
aI895
aI28431
aI62905
asS'd'
I9750
sa(dp19500
S'a'
I34250
sS's'
I17
sS'm'
(lp19501
I1
aI1
aI1
aI13
aI25
aI37
aI111
aI129
aI391
aI813
aI1061
aI4065
aI7339
aI10731
aI23799
aI41463
aI99673
asS'd'
I9751
sa(dp19502
S'a'
I34264
sS's'
I17
sS'm'
(lp19503
I1
aI1
aI7
aI15
aI3
aI21
aI45
aI77
aI471
aI155
aI967
aI711
aI4947
aI13983
aI27827
aI28653
aI117839
asS'd'
I9752
sa(dp19504
S'a'
I34269
sS's'
I17
sS'm'
(lp19505
I1
aI1
aI5
aI9
aI13
aI39
aI107
aI237
aI233
aI881
aI297
aI2189
aI8085
aI1221
aI18659
aI299
aI90951
asS'd'
I9753
sa(dp19506
S'a'
I34276
sS's'
I17
sS'm'
(lp19507
I1
aI1
aI1
aI13
aI21
aI53
aI83
aI17
aI487
aI215
aI1203
aI3017
aI7887
aI3759
aI10521
aI31223
aI87917
asS'd'
I9754
sa(dp19508
S'a'
I34279
sS's'
I17
sS'm'
(lp19509
I1
aI1
aI7
aI1
aI13
aI31
aI123
aI219
aI127
aI743
aI1325
aI3907
aI129
aI8901
aI4855
aI22509
aI47331
asS'd'
I9755
sa(dp19510
S'a'
I34293
sS's'
I17
sS'm'
(lp19511
I1
aI1
aI7
aI11
aI29
aI37
aI11
aI157
aI401
aI35
aI2037
aI2873
aI7409
aI7837
aI1247
aI33911
aI3979
asS'd'
I9756
sa(dp19512
S'a'
I34303
sS's'
I17
sS'm'
(lp19513
I1
aI1
aI5
aI15
aI1
aI13
aI35
aI253
aI287
aI1007
aI1417
aI1613
aI6019
aI11617
aI6323
aI56263
aI45073
asS'd'
I9757
sa(dp19514
S'a'
I34310
sS's'
I17
sS'm'
(lp19515
I1
aI3
aI1
aI15
aI1
aI59
aI41
aI239
aI373
aI443
aI897
aI275
aI5783
aI8619
aI18559
aI16279
aI92063
asS'd'
I9758
sa(dp19516
S'a'
I34340
sS's'
I17
sS'm'
(lp19517
I1
aI3
aI1
aI9
aI23
aI33
aI83
aI43
aI231
aI819
aI1657
aI1031
aI5507
aI12621
aI8961
aI23059
aI63453
asS'd'
I9759
sa(dp19518
S'a'
I34349
sS's'
I17
sS'm'
(lp19519
I1
aI1
aI7
aI5
aI29
aI49
aI21
aI251
aI267
aI43
aI729
aI4013
aI1497
aI15489
aI16761
aI49689
aI122755
asS'd'
I9760
sa(dp19520
S'a'
I34352
sS's'
I17
sS'm'
(lp19521
I1
aI3
aI7
aI1
aI31
aI21
aI11
aI149
aI127
aI711
aI1249
aI49
aI5503
aI677
aI12313
aI61301
aI16279
asS'd'
I9761
sa(dp19522
S'a'
I34355
sS's'
I17
sS'm'
(lp19523
I1
aI1
aI5
aI11
aI9
aI15
aI41
aI61
aI81
aI991
aI1387
aI3567
aI221
aI15835
aI8609
aI28265
aI98199
asS'd'
I9762
sa(dp19524
S'a'
I34358
sS's'
I17
sS'm'
(lp19525
I1
aI3
aI1
aI7
aI21
aI35
aI13
aI59
aI173
aI637
aI107
aI393
aI4551
aI6523
aI27389
aI33129
aI45579
asS'd'
I9763
sa(dp19526
S'a'
I34362
sS's'
I17
sS'm'
(lp19527
I1
aI1
aI1
aI9
aI29
aI51
aI65
aI199
aI417
aI553
aI1321
aI2513
aI4749
aI8477
aI19721
aI24301
aI16301
asS'd'
I9764
sa(dp19528
S'a'
I34376
sS's'
I17
sS'm'
(lp19529
I1
aI3
aI5
aI1
aI25
aI13
aI7
aI55
aI163
aI581
aI1677
aI2313
aI6843
aI15697
aI3055
aI53171
aI59899
asS'd'
I9765
sa(dp19530
S'a'
I34381
sS's'
I17
sS'm'
(lp19531
I1
aI3
aI1
aI5
aI31
aI13
aI101
aI195
aI235
aI359
aI911
aI1017
aI2575
aI12801
aI997
aI7819
aI73243
asS'd'
I9766
sa(dp19532
S'a'
I34387
sS's'
I17
sS'm'
(lp19533
I1
aI1
aI7
aI1
aI9
aI39
aI59
aI83
aI57
aI885
aI317
aI2689
aI5741
aI11833
aI25563
aI62581
aI62239
asS'd'
I9767
sa(dp19534
S'a'
I34389
sS's'
I17
sS'm'
(lp19535
I1
aI1
aI5
aI15
aI25
aI25
aI55
aI207
aI223
aI907
aI913
aI387
aI5599
aI15567
aI8859
aI13703
aI66071
asS'd'
I9768
sa(dp19536
S'a'
I34394
sS's'
I17
sS'm'
(lp19537
I1
aI1
aI5
aI15
aI19
aI39
aI83
aI177
aI333
aI531
aI1257
aI2687
aI7793
aI15967
aI19175
aI1381
aI106629
asS'd'
I9769
sa(dp19538
S'a'
I34410
sS's'
I17
sS'm'
(lp19539
I1
aI3
aI5
aI13
aI29
aI29
aI77
aI1
aI273
aI483
aI725
aI3825
aI5115
aI4043
aI11571
aI8693
aI49761
asS'd'
I9770
sa(dp19540
S'a'
I34423
sS's'
I17
sS'm'
(lp19541
I1
aI1
aI7
aI3
aI5
aI45
aI37
aI65
aI267
aI191
aI301
aI2863
aI167
aI9303
aI14563
aI41553
aI119561
asS'd'
I9771
sa(dp19542
S'a'
I34434
sS's'
I17
sS'm'
(lp19543
I1
aI1
aI7
aI5
aI21
aI41
aI107
aI213
aI267
aI427
aI699
aI1485
aI2125
aI16011
aI29243
aI4691
aI50545
asS'd'
I9772
sa(dp19544
S'a'
I34436
sS's'
I17
sS'm'
(lp19545
I1
aI3
aI3
aI9
aI15
aI29
aI81
aI53
aI289
aI689
aI933
aI2667
aI5175
aI10409
aI28221
aI56375
aI49109
asS'd'
I9773
sa(dp19546
S'a'
I34448
sS's'
I17
sS'm'
(lp19547
I1
aI1
aI1
aI15
aI3
aI11
aI77
aI107
aI353
aI349
aI219
aI1961
aI7559
aI10081
aI25119
aI46041
aI103827
asS'd'
I9774
sa(dp19548
S'a'
I34453
sS's'
I17
sS'm'
(lp19549
I1
aI3
aI3
aI1
aI5
aI27
aI109
aI17
aI271
aI543
aI565
aI397
aI2649
aI12037
aI4525
aI37835
aI107071
asS'd'
I9775
sa(dp19550
S'a'
I34454
sS's'
I17
sS'm'
(lp19551
I1
aI1
aI5
aI15
aI3
aI37
aI123
aI157
aI389
aI619
aI1379
aI4093
aI6107
aI4419
aI21011
aI36189
aI21269
asS'd'
I9776
sa(dp19552
S'a'
I34460
sS's'
I17
sS'm'
(lp19553
I1
aI3
aI1
aI7
aI25
aI17
aI37
aI133
aI247
aI113
aI985
aI815
aI441
aI7869
aI25121
aI49459
aI429
asS'd'
I9777
sa(dp19554
S'a'
I34464
sS's'
I17
sS'm'
(lp19555
I1
aI3
aI3
aI11
aI7
aI23
aI59
aI51
aI403
aI685
aI2019
aI1167
aI7973
aI6915
aI10819
aI43807
aI127793
asS'd'
I9778
sa(dp19556
S'a'
I34479
sS's'
I17
sS'm'
(lp19557
I1
aI1
aI3
aI1
aI29
aI3
aI125
aI107
aI305
aI101
aI391
aI2733
aI6883
aI5867
aI5139
aI16025
aI112439
asS'd'
I9779
sa(dp19558
S'a'
I34491
sS's'
I17
sS'm'
(lp19559
I1
aI1
aI5
aI5
aI23
aI23
aI89
aI33
aI275
aI451
aI1033
aI649
aI3761
aI4735
aI26021
aI9627
aI102747
asS'd'
I9780
sa(dp19560
S'a'
I34501
sS's'
I17
sS'm'
(lp19561
I1
aI1
aI5
aI13
aI3
aI17
aI117
aI251
aI425
aI917
aI759
aI3047
aI8171
aI14421
aI27765
aI11085
aI64889
asS'd'
I9781
sa(dp19562
S'a'
I34508
sS's'
I17
sS'm'
(lp19563
I1
aI3
aI1
aI9
aI7
aI23
aI107
aI143
aI123
aI413
aI2045
aI655
aI6283
aI8783
aI20263
aI55463
aI33271
asS'd'
I9782
sa(dp19564
S'a'
I34516
sS's'
I17
sS'm'
(lp19565
I1
aI3
aI7
aI11
aI5
aI49
aI73
aI55
aI465
aI43
aI587
aI3943
aI521
aI12357
aI16273
aI26603
aI23219
asS'd'
I9783
sa(dp19566
S'a'
I34529
sS's'
I17
sS'm'
(lp19567
I1
aI3
aI5
aI13
aI9
aI3
aI127
aI171
aI271
aI227
aI993
aI1427
aI2235
aI6325
aI13501
aI1411
aI44393
asS'd'
I9784
sa(dp19568
S'a'
I34530
sS's'
I17
sS'm'
(lp19569
I1
aI1
aI1
aI3
aI13
aI27
aI19
aI37
aI175
aI423
aI5
aI3403
aI5427
aI16345
aI30297
aI11909
aI104647
asS'd'
I9785
sa(dp19570
S'a'
I34553
sS's'
I17
sS'm'
(lp19571
I1
aI3
aI1
aI3
aI3
aI39
aI111
aI179
aI487
aI923
aI1945
aI1609
aI4689
aI11807
aI13725
aI3081
aI48163
asS'd'
I9786
sa(dp19572
S'a'
I34564
sS's'
I17
sS'm'
(lp19573
I1
aI3
aI1
aI1
aI9
aI35
aI7
aI151
aI109
aI925
aI1249
aI3171
aI1207
aI2053
aI5135
aI34821
aI57291
asS'd'
I9787
sa(dp19574
S'a'
I34568
sS's'
I17
sS'm'
(lp19575
I1
aI1
aI5
aI13
aI31
aI35
aI101
aI199
aI499
aI725
aI1229
aI2857
aI6437
aI503
aI14437
aI35721
aI24971
asS'd'
I9788
sa(dp19576
S'a'
I34571
sS's'
I17
sS'm'
(lp19577
I1
aI1
aI1
aI15
aI3
aI49
aI75
aI101
aI373
aI119
aI875
aI245
aI15
aI12937
aI4731
aI13037
aI1555
asS'd'
I9789
sa(dp19578
S'a'
I34582
sS's'
I17
sS'm'
(lp19579
I1
aI1
aI1
aI7
aI15
aI5
aI53
aI5
aI423
aI69
aI73
aI2139
aI383
aI4035
aI6723
aI59941
aI124503
asS'd'
I9790
sa(dp19580
S'a'
I34586
sS's'
I17
sS'm'
(lp19581
I1
aI1
aI3
aI13
aI1
aI23
aI29
aI47
aI145
aI785
aI1013
aI1579
aI4579
aI107
aI17571
aI46311
aI27777
asS'd'
I9791
sa(dp19582
S'a'
I34598
sS's'
I17
sS'm'
(lp19583
I1
aI1
aI1
aI5
aI23
aI25
aI97
aI75
aI105
aI183
aI827
aI3871
aI2005
aI6453
aI28729
aI42583
aI62979
asS'd'
I9792
sa(dp19584
S'a'
I34604
sS's'
I17
sS'm'
(lp19585
I1
aI3
aI5
aI9
aI11
aI49
aI29
aI201
aI333
aI441
aI429
aI1955
aI5301
aI11775
aI22915
aI58693
aI111917
asS'd'
I9793
sa(dp19586
S'a'
I34610
sS's'
I17
sS'm'
(lp19587
I1
aI3
aI3
aI1
aI15
aI37
aI117
aI223
aI319
aI181
aI61
aI177
aI507
aI14871
aI16419
aI34261
aI106937
asS'd'
I9794
sa(dp19588
S'a'
I34619
sS's'
I17
sS'm'
(lp19589
I1
aI3
aI3
aI9
aI25
aI27
aI81
aI253
aI459
aI5
aI693
aI1271
aI485
aI16171
aI427
aI17917
aI4393
asS'd'
I9795
sa(dp19590
S'a'
I34621
sS's'
I17
sS'm'
(lp19591
I1
aI3
aI3
aI1
aI27
aI47
aI11
aI57
aI269
aI95
aI569
aI2733
aI3275
aI1599
aI15073
aI58071
aI86805
asS'd'
I9796
sa(dp19592
S'a'
I34633
sS's'
I17
sS'm'
(lp19593
I1
aI3
aI7
aI13
aI21
aI57
aI75
aI63
aI53
aI487
aI251
aI3193
aI4279
aI2311
aI6613
aI38319
aI93557
asS'd'
I9797
sa(dp19594
S'a'
I34634
sS's'
I17
sS'm'
(lp19595
I1
aI3
aI5
aI5
aI31
aI35
aI39
aI255
aI11
aI81
aI605
aI1457
aI6367
aI14121
aI8069
aI46653
aI79945
asS'd'
I9798
sa(dp19596
S'a'
I34657
sS's'
I17
sS'm'
(lp19597
I1
aI1
aI1
aI7
aI17
aI19
aI19
aI247
aI13
aI757
aI1069
aI2811
aI4969
aI10943
aI29399
aI4153
aI120817
asS'd'
I9799
sa(dp19598
S'a'
I34682
sS's'
I17
sS'm'
(lp19599
I1
aI1
aI1
aI15
aI31
aI13
aI1
aI247
aI157
aI785
aI1565
aI897
aI4825
aI8375
aI4933
aI60671
aI88403
asS'd'
I9800
sa(dp19600
S'a'
I34688
sS's'
I17
sS'm'
(lp19601
I1
aI3
aI3
aI7
aI31
aI53
aI117
aI207
aI243
aI603
aI625
aI1039
aI5725
aI5021
aI20227
aI28613
aI123759
asS'd'
I9801
sa(dp19602
S'a'
I34691
sS's'
I17
sS'm'
(lp19603
I1
aI1
aI5
aI1
aI7
aI29
aI65
aI153
aI393
aI821
aI295
aI2705
aI5999
aI15801
aI31301
aI15545
aI52917
asS'd'
I9802
sa(dp19604
S'a'
I34694
sS's'
I17
sS'm'
(lp19605
I1
aI1
aI1
aI1
aI11
aI51
aI97
aI143
aI27
aI279
aI1005
aI1235
aI5539
aI1523
aI26293
aI35015
aI47835
asS'd'
I9803
sa(dp19606
S'a'
I34706
sS's'
I17
sS'm'
(lp19607
I1
aI3
aI3
aI13
aI27
aI17
aI123
aI147
aI39
aI35
aI567
aI961
aI5431
aI5557
aI17849
aI46675
aI102181
asS'd'
I9804
sa(dp19608
S'a'
I34708
sS's'
I17
sS'm'
(lp19609
I1
aI1
aI7
aI11
aI7
aI25
aI73
aI223
aI459
aI207
aI1637
aI647
aI2057
aI685
aI24539
aI48809
aI26877
asS'd'
I9805
sa(dp19610
S'a'
I34724
sS's'
I17
sS'm'
(lp19611
I1
aI3
aI1
aI3
aI21
aI43
aI121
aI11
aI431
aI383
aI1703
aI1451
aI2349
aI11845
aI13801
aI20589
aI43125
asS'd'
I9806
sa(dp19612
S'a'
I34727
sS's'
I17
sS'm'
(lp19613
I1
aI1
aI5
aI1
aI27
aI29
aI89
aI233
aI437
aI303
aI853
aI3425
aI263
aI2073
aI14111
aI39129
aI59547
asS'd'
I9807
sa(dp19614
S'a'
I34751
sS's'
I17
sS'm'
(lp19615
I1
aI1
aI1
aI3
aI3
aI47
aI99
aI207
aI261
aI179
aI1761
aI2657
aI4339
aI6567
aI25455
aI18729
aI51431
asS'd'
I9808
sa(dp19616
S'a'
I34753
sS's'
I17
sS'm'
(lp19617
I1
aI3
aI3
aI13
aI5
aI5
aI109
aI125
aI123
aI233
aI1713
aI1539
aI4375
aI12187
aI18355
aI49597
aI109959
asS'd'
I9809
sa(dp19618
S'a'
I34759
sS's'
I17
sS'm'
(lp19619
I1
aI3
aI7
aI7
aI9
aI23
aI45
aI193
aI363
aI837
aI855
aI1413
aI7587
aI9091
aI27907
aI17809
aI63249
asS'd'
I9810
sa(dp19620
S'a'
I34763
sS's'
I17
sS'm'
(lp19621
I1
aI3
aI3
aI9
aI19
aI23
aI63
aI85
aI419
aI1007
aI1753
aI539
aI1471
aI2171
aI9239
aI36289
aI105503
asS'd'
I9811
sa(dp19622
S'a'
I34777
sS's'
I17
sS'm'
(lp19623
I1
aI3
aI1
aI11
aI23
aI5
aI105
aI79
aI473
aI879
aI1623
aI3155
aI5157
aI4699
aI697
aI41919
aI15441
asS'd'
I9812
sa(dp19624
S'a'
I34778
sS's'
I17
sS'm'
(lp19625
I1
aI1
aI7
aI11
aI5
aI21
aI43
aI207
aI491
aI355
aI857
aI2325
aI819
aI15849
aI24529
aI5789
aI110661
asS'd'
I9813
sa(dp19626
S'a'
I34780
sS's'
I17
sS'm'
(lp19627
I1
aI1
aI5
aI15
aI19
aI33
aI81
aI137
aI473
aI853
aI1681
aI3841
aI5617
aI13715
aI1987
aI52983
aI66327
asS'd'
I9814
sa(dp19628
S'a'
I34796
sS's'
I17
sS'm'
(lp19629
I1
aI3
aI5
aI7
aI11
aI31
aI69
aI85
aI33
aI197
aI1771
aI1957
aI1311
aI169
aI14159
aI7327
aI8577
asS'd'
I9815
sa(dp19630
S'a'
I34799
sS's'
I17
sS'm'
(lp19631
I1
aI1
aI3
aI9
aI11
aI23
aI19
aI143
aI9
aI579
aI111
aI2973
aI3567
aI8561
aI10447
aI55875
aI64305
asS'd'
I9816
sa(dp19632
S'a'
I34801
sS's'
I17
sS'm'
(lp19633
I1
aI1
aI5
aI7
aI1
aI17
aI93
aI11
aI423
aI1007
aI839
aI719
aI3965
aI14531
aI17301
aI29577
aI4083
asS'd'
I9817
sa(dp19634
S'a'
I34817
sS's'
I17
sS'm'
(lp19635
I1
aI3
aI5
aI13
aI19
aI17
aI123
aI61
aI59
aI115
aI1165
aI579
aI2545
aI633
aI5597
aI21865
aI109167
asS'd'
I9818
sa(dp19636
S'a'
I34824
sS's'
I17
sS'm'
(lp19637
I1
aI1
aI5
aI3
aI29
aI29
aI99
aI163
aI321
aI367
aI1523
aI3719
aI665
aI15843
aI28831
aI63823
aI113533
asS'd'
I9819
sa(dp19638
S'a'
I34827
sS's'
I17
sS'm'
(lp19639
I1
aI1
aI1
aI3
aI15
aI7
aI85
aI1
aI181
aI759
aI537
aI3315
aI7159
aI4363
aI4183
aI53775
aI8801
asS'd'
I9820
sa(dp19640
S'a'
I34837
sS's'
I17
sS'm'
(lp19641
I1
aI3
aI1
aI1
aI15
aI53
aI9
aI35
aI459
aI417
aI1169
aI2055
aI1175
aI10923
aI335
aI24269
aI93001
asS'd'
I9821
sa(dp19642
S'a'
I34841
sS's'
I17
sS'm'
(lp19643
I1
aI3
aI1
aI5
aI31
aI43
aI51
aI149
aI175
aI541
aI629
aI1147
aI7585
aI9725
aI18623
aI13345
aI65391
asS'd'
I9822
sa(dp19644
S'a'
I34853
sS's'
I17
sS'm'
(lp19645
I1
aI3
aI7
aI1
aI13
aI39
aI13
aI217
aI507
aI765
aI721
aI1491
aI5037
aI6267
aI2871
aI19181
aI123751
asS'd'
I9823
sa(dp19646
S'a'
I34858
sS's'
I17
sS'm'
(lp19647
I1
aI1
aI3
aI5
aI21
aI9
aI123
aI195
aI63
aI347
aI7
aI531
aI3015
aI9457
aI29543
aI51479
aI26607
asS'd'
I9824
sa(dp19648
S'a'
I34877
sS's'
I17
sS'm'
(lp19649
I1
aI1
aI1
aI1
aI21
aI15
aI81
aI127
aI429
aI15
aI901
aI1503
aI1919
aI6515
aI2477
aI53571
aI113447
asS'd'
I9825
sa(dp19650
S'a'
I34886
sS's'
I17
sS'm'
(lp19651
I1
aI3
aI1
aI13
aI9
aI33
aI79
aI169
aI499
aI767
aI441
aI2085
aI2429
aI10213
aI4125
aI2611
aI26137
asS'd'
I9826
sa(dp19652
S'a'
I34895
sS's'
I17
sS'm'
(lp19653
I1
aI1
aI3
aI1
aI19
aI23
aI83
aI179
aI447
aI513
aI913
aI1201
aI1861
aI11595
aI29037
aI7775
aI116417
asS'd'
I9827
sa(dp19654
S'a'
I34897
sS's'
I17
sS'm'
(lp19655
I1
aI3
aI3
aI7
aI3
aI57
aI47
aI183
aI413
aI319
aI1375
aI1401
aI2231
aI14331
aI28625
aI43839
aI102717
asS'd'
I9828
sa(dp19656
S'a'
I34898
sS's'
I17
sS'm'
(lp19657
I1
aI1
aI5
aI11
aI31
aI27
aI111
aI85
aI191
aI155
aI2025
aI1501
aI4991
aI4655
aI3451
aI10219
aI60391
asS'd'
I9829
sa(dp19658
S'a'
I34916
sS's'
I17
sS'm'
(lp19659
I1
aI3
aI3
aI7
aI17
aI19
aI113
aI37
aI423
aI479
aI709
aI3659
aI6567
aI1709
aI13483
aI61821
aI77101
asS'd'
I9830
sa(dp19660
S'a'
I34923
sS's'
I17
sS'm'
(lp19661
I1
aI3
aI1
aI13
aI3
aI17
aI73
aI61
aI275
aI359
aI1341
aI449
aI1373
aI12047
aI11207
aI52651
aI83305
asS'd'
I9831
sa(dp19662
S'a'
I34928
sS's'
I17
sS'm'
(lp19663
I1
aI1
aI7
aI9
aI9
aI45
aI15
aI121
aI15
aI51
aI509
aI2189
aI5057
aI6119
aI11669
aI14559
aI108323
asS'd'
I9832
sa(dp19664
S'a'
I34934
sS's'
I17
sS'm'
(lp19665
I1
aI1
aI7
aI7
aI25
aI13
aI13
aI141
aI157
aI249
aI823
aI821
aI1909
aI5925
aI3505
aI13187
aI19237
asS'd'
I9833
sa(dp19666
S'a'
I34940
sS's'
I17
sS'm'
(lp19667
I1
aI3
aI3
aI1
aI9
aI51
aI79
aI91
aI5
aI709
aI787
aI2427
aI4613
aI7307
aI20141
aI1675
aI49779
asS'd'
I9834
sa(dp19668
S'a'
I34944
sS's'
I17
sS'm'
(lp19669
I1
aI1
aI1
aI11
aI11
aI13
aI33
aI81
aI413
aI981
aI907
aI2709
aI4113
aI10607
aI2587
aI12845
aI11103
asS'd'
I9835
sa(dp19670
S'a'
I34947
sS's'
I17
sS'm'
(lp19671
I1
aI1
aI7
aI9
aI13
aI25
aI37
aI81
aI375
aI1013
aI2027
aI321
aI3947
aI2269
aI10687
aI7537
aI67495
asS'd'
I9836
sa(dp19672
S'a'
I34953
sS's'
I17
sS'm'
(lp19673
I1
aI3
aI5
aI11
aI9
aI43
aI53
aI111
aI339
aI841
aI503
aI3209
aI6437
aI10893
aI13627
aI51809
aI57229
asS'd'
I9837
sa(dp19674
S'a'
I34956
sS's'
I17
sS'm'
(lp19675
I1
aI3
aI1
aI1
aI21
aI15
aI71
aI93
aI453
aI405
aI1099
aI2979
aI7471
aI10173
aI17875
aI13179
aI48615
asS'd'
I9838
sa(dp19676
S'a'
I34967
sS's'
I17
sS'm'
(lp19677
I1
aI3
aI5
aI9
aI9
aI1
aI121
aI117
aI275
aI157
aI57
aI3459
aI4787
aI15005
aI24591
aI23963
aI45077
asS'd'
I9839
sa(dp19678
S'a'
I34968
sS's'
I17
sS'm'
(lp19679
I1
aI1
aI5
aI3
aI21
aI57
aI113
aI207
aI169
aI603
aI637
aI1455
aI6281
aI6527
aI17219
aI32307
aI18617
asS'd'
I9840
sa(dp19680
S'a'
I34971
sS's'
I17
sS'm'
(lp19681
I1
aI3
aI7
aI5
aI25
aI15
aI99
aI91
aI253
aI267
aI537
aI713
aI3929
aI895
aI7999
aI47989
aI118731
asS'd'
I9841
sa(dp19682
S'a'
I34974
sS's'
I17
sS'm'
(lp19683
I1
aI3
aI7
aI15
aI23
aI17
aI5
aI129
aI121
aI251
aI219
aI2547
aI7291
aI1079
aI14577
aI56229
aI35253
asS'd'
I9842
sa(dp19684
S'a'
I34977
sS's'
I17
sS'm'
(lp19685
I1
aI3
aI1
aI15
aI5
aI61
aI35
aI135
aI497
aI681
aI751
aI2303
aI6697
aI11225
aI30389
aI61673
aI87313
asS'd'
I9843
sa(dp19686
S'a'
I34980
sS's'
I17
sS'm'
(lp19687
I1
aI3
aI1
aI7
aI7
aI37
aI9
aI85
aI257
aI805
aI1325
aI3597
aI6065
aI727
aI18203
aI57077
aI437
asS'd'
I9844
sa(dp19688
S'a'
I34983
sS's'
I17
sS'm'
(lp19689
I1
aI3
aI5
aI7
aI5
aI43
aI29
aI179
aI73
aI173
aI1441
aI1233
aI1779
aI7893
aI10629
aI27547
aI7775
asS'd'
I9845
sa(dp19690
S'a'
I34998
sS's'
I17
sS'm'
(lp19691
I1
aI1
aI7
aI5
aI31
aI29
aI21
aI35
aI289
aI423
aI449
aI3331
aI2929
aI6827
aI15569
aI9873
aI76889
asS'd'
I9846
sa(dp19692
S'a'
I35004
sS's'
I17
sS'm'
(lp19693
I1
aI1
aI7
aI13
aI13
aI37
aI55
aI99
aI135
aI797
aI1263
aI2539
aI893
aI4225
aI16689
aI38259
aI50857
asS'd'
I9847
sa(dp19694
S'a'
I35010
sS's'
I17
sS'm'
(lp19695
I1
aI1
aI3
aI1
aI5
aI3
aI95
aI29
aI15
aI539
aI825
aI3931
aI4809
aI8299
aI29891
aI61357
aI97523
asS'd'
I9848
sa(dp19696
S'a'
I35012
sS's'
I17
sS'm'
(lp19697
I1
aI3
aI1
aI9
aI27
aI25
aI115
aI239
aI387
aI163
aI1153
aI31
aI2375
aI7943
aI31929
aI1121
aI33085
asS'd'
I9849
sa(dp19698
S'a'
I35030
sS's'
I17
sS'm'
(lp19699
I1
aI3
aI5
aI9
aI3
aI53
aI121
aI159
aI165
aI81
aI317
aI3051
aI1991
aI493
aI2029
aI43305
aI130209
asS'd'
I9850
sa(dp19700
S'a'
I35039
sS's'
I17
sS'm'
(lp19701
I1
aI1
aI1
aI5
aI9
aI57
aI39
aI247
aI73
aI613
aI1047
aI3289
aI2569
aI5363
aI18475
aI32749
aI39415
asS'd'
I9851
sa(dp19702
S'a'
I35058
sS's'
I17
sS'm'
(lp19703
I1
aI3
aI1
aI5
aI19
aI23
aI39
aI33
aI151
aI463
aI153
aI737
aI2501
aI7531
aI2769
aI35595
aI71799
asS'd'
I9852
sa(dp19704
S'a'
I35063
sS's'
I17
sS'm'
(lp19705
I1
aI3
aI5
aI5
aI29
aI49
aI105
aI81
aI67
aI441
aI1101
aI2241
aI6243
aI6177
aI7157
aI51635
aI81241
asS'd'
I9853
sa(dp19706
S'a'
I35082
sS's'
I17
sS'm'
(lp19707
I1
aI3
aI3
aI3
aI29
aI53
aI13
aI239
aI487
aI503
aI97
aI1323
aI1817
aI13021
aI12881
aI26943
aI21011
asS'd'
I9854
sa(dp19708
S'a'
I35095
sS's'
I17
sS'm'
(lp19709
I1
aI1
aI1
aI15
aI25
aI9
aI5
aI205
aI85
aI635
aI789
aI2495
aI5069
aI4987
aI847
aI26857
aI84225
asS'd'
I9855
sa(dp19710
S'a'
I35096
sS's'
I17
sS'm'
(lp19711
I1
aI1
aI3
aI15
aI9
aI51
aI79
aI13
aI377
aI637
aI159
aI3407
aI2057
aI13967
aI31781
aI40869
aI52987
asS'd'
I9856
sa(dp19712
S'a'
I35101
sS's'
I17
sS'm'
(lp19713
I1
aI3
aI1
aI13
aI11
aI27
aI103
aI207
aI383
aI887
aI749
aI1119
aI285
aI4269
aI31745
aI57539
aI5671
asS'd'
I9857
sa(dp19714
S'a'
I35102
sS's'
I17
sS'm'
(lp19715
I1
aI3
aI1
aI13
aI23
aI19
aI41
aI43
aI455
aI425
aI1653
aI4091
aI4855
aI16321
aI169
aI59289
aI82397
asS'd'
I9858
sa(dp19716
S'a'
I35105
sS's'
I17
sS'm'
(lp19717
I1
aI3
aI3
aI15
aI31
aI39
aI51
aI127
aI391
aI989
aI1831
aI3327
aI6487
aI6077
aI17277
aI52093
aI20389
asS'd'
I9859
sa(dp19718
S'a'
I35112
sS's'
I17
sS'm'
(lp19719
I1
aI3
aI5
aI15
aI19
aI1
aI21
aI241
aI15
aI543
aI1529
aI2355
aI1503
aI12795
aI17321
aI41219
aI61115
asS'd'
I9860
sa(dp19720
S'a'
I35118
sS's'
I17
sS'm'
(lp19721
I1
aI1
aI3
aI11
aI9
aI33
aI21
aI197
aI307
aI141
aI1663
aI371
aI1663
aI8307
aI3617
aI56941
aI62477
asS'd'
I9861
sa(dp19722
S'a'
I35120
sS's'
I17
sS'm'
(lp19723
I1
aI3
aI7
aI9
aI19
aI53
aI123
aI3
aI29
aI635
aI1795
aI2471
aI2491
aI15847
aI9169
aI2561
aI101515
asS'd'
I9862
sa(dp19724
S'a'
I35130
sS's'
I17
sS'm'
(lp19725
I1
aI1
aI5
aI3
aI19
aI11
aI117
aI231
aI475
aI837
aI1833
aI3499
aI4415
aI9961
aI28285
aI37821
aI81497
asS'd'
I9863
sa(dp19726
S'a'
I35143
sS's'
I17
sS'm'
(lp19727
I1
aI1
aI3
aI5
aI7
aI11
aI57
aI89
aI345
aI157
aI1519
aI3021
aI7157
aI2159
aI32557
aI31559
aI128907
asS'd'
I9864
sa(dp19728
S'a'
I35147
sS's'
I17
sS'm'
(lp19729
I1
aI1
aI7
aI3
aI27
aI1
aI15
aI177
aI489
aI405
aI811
aI3597
aI4939
aI15595
aI7279
aI58097
aI84703
asS'd'
I9865
sa(dp19730
S'a'
I35152
sS's'
I17
sS'm'
(lp19731
I1
aI3
aI1
aI9
aI25
aI61
aI119
aI219
aI111
aI339
aI1091
aI759
aI6087
aI16001
aI6757
aI15627
aI1691
asS'd'
I9866
sa(dp19732
S'a'
I35157
sS's'
I17
sS'm'
(lp19733
I1
aI3
aI7
aI9
aI1
aI39
aI107
aI139
aI143
aI917
aI421
aI1623
aI7135
aI4851
aI6687
aI6177
aI102425
asS'd'
I9867
sa(dp19734
S'a'
I35164
sS's'
I17
sS'm'
(lp19735
I1
aI1
aI7
aI13
aI23
aI17
aI19
aI167
aI317
aI331
aI743
aI3737
aI2195
aI545
aI2185
aI9125
aI30503
asS'd'
I9868
sa(dp19736
S'a'
I35178
sS's'
I17
sS'm'
(lp19737
I1
aI1
aI5
aI13
aI27
aI33
aI117
aI141
aI493
aI129
aI1553
aI2335
aI4161
aI14205
aI24177
aI35163
aI84869
asS'd'
I9869
sa(dp19738
S'a'
I35195
sS's'
I17
sS'm'
(lp19739
I1
aI3
aI7
aI1
aI11
aI9
aI75
aI133
aI113
aI507
aI2007
aI2473
aI4769
aI14655
aI17967
aI17709
aI90653
asS'd'
I9870
sa(dp19740
S'a'
I35197
sS's'
I17
sS'm'
(lp19741
I1
aI1
aI7
aI11
aI17
aI11
aI83
aI23
aI387
aI61
aI29
aI3905
aI4351
aI15173
aI28375
aI9129
aI111939
asS'd'
I9871
sa(dp19742
S'a'
I35201
sS's'
I17
sS'm'
(lp19743
I1
aI1
aI5
aI15
aI15
aI53
aI81
aI125
aI189
aI937
aI1607
aI2595
aI2847
aI7229
aI22241
aI26269
aI64781
asS'd'
I9872
sa(dp19744
S'a'
I35207
sS's'
I17
sS'm'
(lp19745
I1
aI3
aI1
aI7
aI5
aI11
aI61
aI111
aI13
aI423
aI885
aI2329
aI6003
aI16331
aI11207
aI25743
aI54619
asS'd'
I9873
sa(dp19746
S'a'
I35231
sS's'
I17
sS'm'
(lp19747
I1
aI3
aI5
aI9
aI1
aI13
aI95
aI241
aI237
aI629
aI263
aI1629
aI1063
aI12695
aI14501
aI5455
aI121483
asS'd'
I9874
sa(dp19748
S'a'
I35249
sS's'
I17
sS'm'
(lp19749
I1
aI1
aI7
aI15
aI5
aI17
aI45
aI255
aI143
aI79
aI87
aI1755
aI6215
aI5095
aI32411
aI8695
aI85511
asS'd'
I9875
sa(dp19750
S'a'
I35250
sS's'
I17
sS'm'
(lp19751
I1
aI3
aI7
aI7
aI21
aI11
aI117
aI135
aI333
aI73
aI1471
aI2749
aI5801
aI4209
aI9353
aI46171
aI90645
asS'd'
I9876
sa(dp19752
S'a'
I35256
sS's'
I17
sS'm'
(lp19753
I1
aI1
aI7
aI13
aI11
aI35
aI77
aI149
aI159
aI783
aI1527
aI2881
aI1409
aI3455
aI26991
aI3225
aI30693
asS'd'
I9877
sa(dp19754
S'a'
I35259
sS's'
I17
sS'm'
(lp19755
I1
aI1
aI3
aI15
aI19
aI55
aI21
aI245
aI207
aI103
aI775
aI2041
aI4637
aI7333
aI11267
aI60509
aI43099
asS'd'
I9878
sa(dp19756
S'a'
I35262
sS's'
I17
sS'm'
(lp19757
I1
aI3
aI3
aI15
aI17
aI63
aI23
aI81
aI183
aI923
aI75
aI391
aI615
aI13343
aI20839
aI56529
aI115747
asS'd'
I9879
sa(dp19758
S'a'
I35273
sS's'
I17
sS'm'
(lp19759
I1
aI3
aI1
aI13
aI5
aI5
aI15
aI27
aI263
aI497
aI1365
aI2733
aI5395
aI7461
aI2725
aI24735
aI89251
asS'd'
I9880
sa(dp19760
S'a'
I35282
sS's'
I17
sS'm'
(lp19761
I1
aI1
aI7
aI7
aI29
aI17
aI39
aI117
aI363
aI915
aI123
aI283
aI4575
aI3497
aI20995
aI37883
aI16645
asS'd'
I9881
sa(dp19762
S'a'
I35298
sS's'
I17
sS'm'
(lp19763
I1
aI3
aI3
aI9
aI1
aI25
aI79
aI181
aI331
aI617
aI393
aI1807
aI5145
aI8007
aI9173
aI45189
aI37945
asS'd'
I9882
sa(dp19764
S'a'
I35307
sS's'
I17
sS'm'
(lp19765
I1
aI3
aI1
aI5
aI1
aI9
aI127
aI137
aI379
aI371
aI367
aI3237
aI581
aI15295
aI18191
aI37689
aI103495
asS'd'
I9883
sa(dp19766
S'a'
I35328
sS's'
I17
sS'm'
(lp19767
I1
aI1
aI7
aI1
aI29
aI53
aI103
aI173
aI171
aI973
aI933
aI3847
aI3185
aI10107
aI31701
aI45021
aI106251
asS'd'
I9884
sa(dp19768
S'a'
I35334
sS's'
I17
sS'm'
(lp19769
I1
aI1
aI1
aI7
aI23
aI9
aI61
aI25
aI343
aI471
aI2041
aI2179
aI7647
aI1885
aI15353
aI50379
aI67681
asS'd'
I9885
sa(dp19770
S'a'
I35343
sS's'
I17
sS'm'
(lp19771
I1
aI1
aI5
aI11
aI31
aI13
aI51
aI185
aI83
aI917
aI85
aI1317
aI8185
aI14949
aI32455
aI57939
aI1217
asS'd'
I9886
sa(dp19772
S'a'
I35345
sS's'
I17
sS'm'
(lp19773
I1
aI1
aI7
aI5
aI23
aI45
aI101
aI227
aI497
aI941
aI985
aI167
aI6847
aI9611
aI20011
aI40069
aI83285
asS'd'
I9887
sa(dp19774
S'a'
I35355
sS's'
I17
sS'm'
(lp19775
I1
aI1
aI5
aI13
aI17
aI33
aI61
aI197
aI433
aI255
aI67
aI1479
aI5663
aI6501
aI30695
aI27235
aI80141
asS'd'
I9888
sa(dp19776
S'a'
I35388
sS's'
I17
sS'm'
(lp19777
I1
aI1
aI3
aI5
aI11
aI45
aI123
aI49
aI327
aI893
aI1963
aI2225
aI2611
aI8925
aI22811
aI2313
aI8411
asS'd'
I9889
sa(dp19778
S'a'
I35399
sS's'
I17
sS'm'
(lp19779
I1
aI3
aI7
aI7
aI15
aI39
aI75
aI235
aI13
aI847
aI575
aI3947
aI6947
aI2061
aI13467
aI103
aI86285
asS'd'
I9890
sa(dp19780
S'a'
I35403
sS's'
I17
sS'm'
(lp19781
I1
aI1
aI7
aI3
aI21
aI43
aI113
aI197
aI141
aI873
aI1139
aI2707
aI7235
aI10683
aI10831
aI33695
aI57063
asS'd'
I9891
sa(dp19782
S'a'
I35408
sS's'
I17
sS'm'
(lp19783
I1
aI3
aI5
aI1
aI3
aI27
aI45
aI43
aI119
aI979
aI1933
aI1851
aI6497
aI14937
aI4965
aI41285
aI120221
asS'd'
I9892
sa(dp19784
S'a'
I35413
sS's'
I17
sS'm'
(lp19785
I1
aI1
aI3
aI1
aI23
aI59
aI67
aI7
aI49
aI351
aI1053
aI1837
aI501
aI7671
aI26239
aI51951
aI95119
asS'd'
I9893
sa(dp19786
S'a'
I35418
sS's'
I17
sS'm'
(lp19787
I1
aI3
aI5
aI11
aI3
aI19
aI33
aI33
aI219
aI175
aI1439
aI197
aI1841
aI159
aI11229
aI20463
aI81797
asS'd'
I9894
sa(dp19788
S'a'
I35434
sS's'
I17
sS'm'
(lp19789
I1
aI1
aI7
aI1
aI13
aI11
aI79
aI75
aI53
aI525
aI91
aI233
aI5999
aI2921
aI21295
aI56831
aI116049
asS'd'
I9895
sa(dp19790
S'a'
I35436
sS's'
I17
sS'm'
(lp19791
I1
aI3
aI3
aI13
aI29
aI7
aI71
aI207
aI193
aI635
aI1393
aI3093
aI3775
aI12445
aI23281
aI29401
aI103225
asS'd'
I9896
sa(dp19792
S'a'
I35448
sS's'
I17
sS'm'
(lp19793
I1
aI1
aI7
aI3
aI29
aI57
aI111
aI163
aI63
aI593
aI881
aI1587
aI3027
aI12599
aI30977
aI38891
aI95495
asS'd'
I9897
sa(dp19794
S'a'
I35460
sS's'
I17
sS'm'
(lp19795
I1
aI1
aI5
aI15
aI17
aI57
aI111
aI169
aI149
aI767
aI377
aI765
aI7533
aI1539
aI22979
aI55489
aI29799
asS'd'
I9898
sa(dp19796
S'a'
I35475
sS's'
I17
sS'm'
(lp19797
I1
aI3
aI5
aI15
aI25
aI7
aI127
aI71
aI319
aI389
aI497
aI1513
aI1287
aI7359
aI12311
aI45457
aI45897
asS'd'
I9899
sa(dp19798
S'a'
I35494
sS's'
I17
sS'm'
(lp19799
I1
aI1
aI5
aI3
aI3
aI35
aI45
aI17
aI49
aI483
aI197
aI727
aI5355
aI7201
aI3035
aI14313
aI40933
asS'd'
I9900
sa(dp19800
S'a'
I35497
sS's'
I17
sS'm'
(lp19801
I1
aI1
aI7
aI15
aI1
aI9
aI27
aI59
aI455
aI653
aI1907
aI281
aI1435
aI14593
aI18909
aI37655
aI87603
asS'd'
I9901
sa(dp19802
S'a'
I35503
sS's'
I17
sS'm'
(lp19803
I1
aI1
aI7
aI11
aI29
aI9
aI67
aI17
aI353
aI709
aI859
aI3687
aI7741
aI4251
aI12263
aI41717
aI79393
asS'd'
I9902
sa(dp19804
S'a'
I35508
sS's'
I17
sS'm'
(lp19805
I1
aI3
aI3
aI3
aI1
aI15
aI113
aI187
aI255
aI851
aI503
aI4089
aI7923
aI1701
aI305
aI8353
aI16357
asS'd'
I9903
sa(dp19806
S'a'
I35511
sS's'
I17
sS'm'
(lp19807
I1
aI1
aI5
aI3
aI17
aI31
aI29
aI233
aI377
aI215
aI1889
aI3459
aI2443
aI3907
aI4193
aI16519
aI49089
asS'd'
I9904
sa(dp19808
S'a'
I35518
sS's'
I17
sS'm'
(lp19809
I1
aI1
aI3
aI1
aI17
aI39
aI11
aI255
aI247
aI305
aI669
aI1769
aI1355
aI12055
aI2275
aI51681
aI112337
asS'd'
I9905
sa(dp19810
S'a'
I35520
sS's'
I17
sS'm'
(lp19811
I1
aI3
aI1
aI1
aI17
aI17
aI75
aI95
aI409
aI21
aI1513
aI1443
aI4931
aI6491
aI1587
aI62979
aI90395
asS'd'
I9906
sa(dp19812
S'a'
I35530
sS's'
I17
sS'm'
(lp19813
I1
aI1
aI3
aI5
aI3
aI19
aI125
aI175
aI279
aI911
aI301
aI407
aI7773
aI949
aI32107
aI13571
aI58717
asS'd'
I9907
sa(dp19814
S'a'
I35537
sS's'
I17
sS'm'
(lp19815
I1
aI3
aI3
aI15
aI31
aI35
aI11
aI223
aI125
aI209
aI1719
aI1725
aI3387
aI14879
aI32243
aI7219
aI126791
asS'd'
I9908
sa(dp19816
S'a'
I35543
sS's'
I17
sS'm'
(lp19817
I1
aI1
aI3
aI1
aI31
aI29
aI67
aI79
aI93
aI193
aI1573
aI2285
aI3209
aI8397
aI17717
aI5657
aI61545
asS'd'
I9909
sa(dp19818
S'a'
I35560
sS's'
I17
sS'm'
(lp19819
I1
aI3
aI1
aI9
aI11
aI33
aI85
aI121
aI193
aI63
aI461
aI1835
aI889
aI10687
aI19831
aI49551
aI59087
asS'd'
I9910
sa(dp19820
S'a'
I35566
sS's'
I17
sS'm'
(lp19821
I1
aI3
aI3
aI7
aI11
aI3
aI9
aI87
aI91
aI487
aI289
aI1113
aI8135
aI7971
aI16693
aI31009
aI81197
asS'd'
I9911
sa(dp19822
S'a'
I35571
sS's'
I17
sS'm'
(lp19823
I1
aI3
aI3
aI1
aI23
aI23
aI61
aI209
aI409
aI845
aI547
aI1493
aI465
aI6399
aI17633
aI53647
aI52425
asS'd'
I9912
sa(dp19824
S'a'
I35598
sS's'
I17
sS'm'
(lp19825
I1
aI1
aI7
aI7
aI21
aI31
aI71
aI249
aI63
aI895
aI653
aI93
aI4429
aI8951
aI16873
aI48089
aI33947
asS'd'
I9913
sa(dp19826
S'a'
I35609
sS's'
I17
sS'm'
(lp19827
I1
aI3
aI5
aI11
aI3
aI35
aI49
aI15
aI379
aI645
aI855
aI3657
aI8019
aI2141
aI11233
aI60731
aI80455
asS'd'
I9914
sa(dp19828
S'a'
I35612
sS's'
I17
sS'm'
(lp19829
I1
aI3
aI1
aI3
aI1
aI53
aI101
aI157
aI255
aI765
aI1575
aI1615
aI7677
aI9699
aI13351
aI2207
aI90939
asS'd'
I9915
sa(dp19830
S'a'
I35615
sS's'
I17
sS'm'
(lp19831
I1
aI3
aI7
aI7
aI5
aI43
aI123
aI109
aI119
aI391
aI1889
aI1991
aI3151
aI1457
aI16321
aI65245
aI75891
asS'd'
I9916
sa(dp19832
S'a'
I35616
sS's'
I17
sS'm'
(lp19833
I1
aI3
aI1
aI15
aI9
aI1
aI113
aI249
aI1
aI675
aI501
aI487
aI2209
aI4411
aI6609
aI29243
aI100177
asS'd'
I9917
sa(dp19834
S'a'
I35622
sS's'
I17
sS'm'
(lp19835
I1
aI1
aI1
aI7
aI9
aI23
aI9
aI197
aI341
aI191
aI453
aI3733
aI5475
aI15515
aI28979
aI36077
aI17801
asS'd'
I9918
sa(dp19836
S'a'
I35626
sS's'
I17
sS'm'
(lp19837
I1
aI1
aI3
aI13
aI5
aI35
aI85
aI121
aI59
aI429
aI1251
aI3437
aI3121
aI12411
aI14713
aI28125
aI31921
asS'd'
I9919
sa(dp19838
S'a'
I35633
sS's'
I17
sS'm'
(lp19839
I1
aI3
aI5
aI3
aI27
aI17
aI61
aI255
aI485
aI709
aI83
aI3201
aI2191
aI3371
aI2941
aI10931
aI22141
asS'd'
I9920
sa(dp19840
S'a'
I35636
sS's'
I17
sS'm'
(lp19841
I1
aI1
aI1
aI1
aI19
aI19
aI25
aI177
aI397
aI579
aI529
aI1619
aI3887
aI4537
aI8123
aI52481
aI8305
asS'd'
I9921
sa(dp19842
S'a'
I35645
sS's'
I17
sS'm'
(lp19843
I1
aI1
aI3
aI15
aI3
aI15
aI77
aI51
aI31
aI881
aI203
aI2359
aI4947
aI6321
aI14705
aI16471
aI84395
asS'd'
I9922
sa(dp19844
S'a'
I35653
sS's'
I17
sS'm'
(lp19845
I1
aI3
aI7
aI9
aI13
aI53
aI67
aI41
aI289
aI721
aI1743
aI2725
aI435
aI1327
aI14953
aI14283
aI113211
asS'd'
I9923
sa(dp19846
S'a'
I35663
sS's'
I17
sS'm'
(lp19847
I1
aI3
aI1
aI5
aI19
aI23
aI73
aI181
aI187
aI675
aI125
aI1877
aI6167
aI7919
aI3955
aI25007
aI28299
asS'd'
I9924
sa(dp19848
S'a'
I35665
sS's'
I17
sS'm'
(lp19849
I1
aI1
aI3
aI1
aI5
aI11
aI123
aI189
aI173
aI123
aI499
aI2175
aI483
aI13017
aI14709
aI5797
aI36327
asS'd'
I9925
sa(dp19850
S'a'
I35682
sS's'
I17
sS'm'
(lp19851
I1
aI3
aI7
aI5
aI21
aI39
aI79
aI229
aI19
aI203
aI375
aI3901
aI1053
aI14209
aI13535
aI63155
aI99727
asS'd'
I9926
sa(dp19852
S'a'
I35687
sS's'
I17
sS'm'
(lp19853
I1
aI1
aI1
aI13
aI11
aI29
aI29
aI173
aI441
aI271
aI1147
aI2891
aI965
aI10777
aI16325
aI37135
aI101601
asS'd'
I9927
sa(dp19854
S'a'
I35688
sS's'
I17
sS'm'
(lp19855
I1
aI1
aI3
aI3
aI25
aI13
aI79
aI233
aI75
aI191
aI987
aI3231
aI3667
aI1525
aI14193
aI62027
aI77441
asS'd'
I9928
sa(dp19856
S'a'
I35691
sS's'
I17
sS'm'
(lp19857
I1
aI3
aI1
aI1
aI15
aI53
aI17
aI45
aI367
aI263
aI425
aI1565
aI6139
aI13833
aI12547
aI61103
aI75361
asS'd'
I9929
sa(dp19858
S'a'
I35696
sS's'
I17
sS'm'
(lp19859
I1
aI1
aI5
aI15
aI5
aI57
aI123
aI47
aI407
aI887
aI375
aI1181
aI5367
aI10283
aI24799
aI33121
aI76373
asS'd'
I9930
sa(dp19860
S'a'
I35727
sS's'
I17
sS'm'
(lp19861
I1
aI1
aI7
aI3
aI11
aI17
aI65
aI133
aI3
aI609
aI601
aI3391
aI7801
aI4137
aI32095
aI55983
aI23037
asS'd'
I9931
sa(dp19862
S'a'
I35741
sS's'
I17
sS'm'
(lp19863
I1
aI3
aI1
aI3
aI25
aI5
aI125
aI5
aI297
aI571
aI145
aI3601
aI1929
aI13457
aI16977
aI21049
aI92169
asS'd'
I9932
sa(dp19864
S'a'
I35742
sS's'
I17
sS'm'
(lp19865
I1
aI3
aI5
aI13
aI23
aI29
aI13
aI143
aI507
aI187
aI857
aI427
aI5125
aI1377
aI10947
aI58473
aI110541
asS'd'
I9933
sa(dp19866
S'a'
I35746
sS's'
I17
sS'm'
(lp19867
I1
aI3
aI3
aI15
aI15
aI49
aI39
aI103
aI193
aI507
aI639
aI2399
aI3829
aI12105
aI15993
aI52975
aI115935
asS'd'
I9934
sa(dp19868
S'a'
I35748
sS's'
I17
sS'm'
(lp19869
I1
aI3
aI7
aI3
aI7
aI41
aI95
aI127
aI193
aI923
aI1729
aI3039
aI7959
aI3345
aI7725
aI35293
aI34361
asS'd'
I9935
sa(dp19870
S'a'
I35752
sS's'
I17
sS'm'
(lp19871
I1
aI3
aI5
aI13
aI17
aI53
aI111
aI141
aI151
aI389
aI1955
aI3333
aI4523
aI6331
aI21239
aI57447
aI113325
asS'd'
I9936
sa(dp19872
S'a'
I35770
sS's'
I17
sS'm'
(lp19873
I1
aI3
aI7
aI15
aI31
aI7
aI11
aI35
aI105
aI607
aI1665
aI3281
aI487
aI9417
aI26205
aI26963
aI81537
asS'd'
I9937
sa(dp19874
S'a'
I35811
sS's'
I17
sS'm'
(lp19875
I1
aI3
aI1
aI1
aI17
aI15
aI3
aI55
aI451
aI691
aI1525
aI2009
aI6443
aI4629
aI15091
aI46961
aI83361
asS'd'
I9938
sa(dp19876
S'a'
I35817
sS's'
I17
sS'm'
(lp19877
I1
aI3
aI1
aI15
aI1
aI29
aI99
aI79
aI225
aI665
aI623
aI2389
aI3303
aI7221
aI20567
aI15917
aI24677
asS'd'
I9939
sa(dp19878
S'a'
I35832
sS's'
I17
sS'm'
(lp19879
I1
aI1
aI3
aI15
aI3
aI17
aI125
aI239
aI485
aI849
aI327
aI1459
aI3911
aI2145
aI14475
aI24337
aI19695
asS'd'
I9940
sa(dp19880
S'a'
I35838
sS's'
I17
sS'm'
(lp19881
I1
aI3
aI5
aI7
aI7
aI37
aI19
aI51
aI373
aI587
aI147
aI563
aI7623
aI7781
aI18289
aI37239
aI6803
asS'd'
I9941
sa(dp19882
S'a'
I35850
sS's'
I17
sS'm'
(lp19883
I1
aI3
aI5
aI1
aI9
aI63
aI5
aI87
aI171
aI5
aI1553
aI429
aI5001
aI7881
aI1493
aI20425
aI57727
asS'd'
I9942
sa(dp19884
S'a'
I35863
sS's'
I17
sS'm'
(lp19885
I1
aI3
aI5
aI9
aI25
aI43
aI17
aI71
aI87
aI869
aI1219
aI2661
aI4571
aI9689
aI18799
aI62467
aI128531
asS'd'
I9943
sa(dp19886
S'a'
I35870
sS's'
I17
sS'm'
(lp19887
I1
aI1
aI3
aI3
aI19
aI53
aI61
aI9
aI55
aI433
aI1555
aI2369
aI1423
aI9081
aI19185
aI8513
aI111079
asS'd'
I9944
sa(dp19888
S'a'
I35879
sS's'
I17
sS'm'
(lp19889
I1
aI3
aI5
aI15
aI11
aI61
aI1
aI147
aI17
aI71
aI1563
aI1113
aI4809
aI16229
aI23743
aI59757
aI64699
asS'd'
I9945
sa(dp19890
S'a'
I35880
sS's'
I17
sS'm'
(lp19891
I1
aI1
aI5
aI11
aI29
aI23
aI61
aI43
aI203
aI97
aI1119
aI237
aI6445
aI14507
aI9799
aI18447
aI14745
asS'd'
I9946
sa(dp19892
S'a'
I35891
sS's'
I17
sS'm'
(lp19893
I1
aI3
aI5
aI15
aI11
aI17
aI117
aI139
aI117
aI537
aI251
aI149
aI2731
aI15863
aI1381
aI25435
aI25501
asS'd'
I9947
sa(dp19894
S'a'
I35893
sS's'
I17
sS'm'
(lp19895
I1
aI3
aI3
aI15
aI31
aI57
aI53
aI43
aI95
aI445
aI1423
aI3833
aI2485
aI11789
aI16011
aI8101
aI39165
asS'd'
I9948
sa(dp19896
S'a'
I35903
sS's'
I17
sS'm'
(lp19897
I1
aI1
aI3
aI11
aI15
aI37
aI117
aI3
aI245
aI57
aI593
aI2771
aI7181
aI11397
aI5691
aI3217
aI44139
asS'd'
I9949
sa(dp19898
S'a'
I35905
sS's'
I17
sS'm'
(lp19899
I1
aI3
aI5
aI1
aI11
aI13
aI121
aI85
aI85
aI511
aI1837
aI611
aI237
aI4893
aI24025
aI28903
aI102025
asS'd'
I9950
sa(dp19900
S'a'
I35926
sS's'
I17
sS'm'
(lp19901
I1
aI3
aI1
aI11
aI5
aI45
aI43
aI45
aI393
aI741
aI1157
aI1511
aI1665
aI2359
aI19071
aI24537
aI122879
asS'd'
I9951
sa(dp19902
S'a'
I35930
sS's'
I17
sS'm'
(lp19903
I1
aI3
aI3
aI3
aI9
aI59
aI27
aI11
aI257
aI203
aI1535
aI2729
aI2313
aI3539
aI1689
aI31901
aI42949
asS'd'
I9952
sa(dp19904
S'a'
I35941
sS's'
I17
sS'm'
(lp19905
I1
aI1
aI1
aI11
aI17
aI7
aI21
aI35
aI479
aI697
aI107
aI1317
aI6585
aI705
aI3789
aI20439
aI33375
asS'd'
I9953
sa(dp19906
S'a'
I35956
sS's'
I17
sS'm'
(lp19907
I1
aI1
aI3
aI11
aI19
aI37
aI123
aI233
aI253
aI733
aI901
aI3047
aI3595
aI2357
aI24533
aI40519
aI109171
asS'd'
I9954
sa(dp19908
S'a'
I35963
sS's'
I17
sS'm'
(lp19909
I1
aI3
aI3
aI13
aI29
aI51
aI25
aI149
aI57
aI253
aI2001
aI351
aI7367
aI15361
aI4955
aI60951
aI19449
asS'd'
I9955
sa(dp19910
S'a'
I35970
sS's'
I17
sS'm'
(lp19911
I1
aI1
aI3
aI15
aI21
aI53
aI25
aI239
aI257
aI437
aI711
aI3599
aI5441
aI7405
aI15039
aI19207
aI63841
asS'd'
I9956
sa(dp19912
S'a'
I35984
sS's'
I17
sS'm'
(lp19913
I1
aI3
aI1
aI9
aI17
aI41
aI43
aI231
aI413
aI747
aI1447
aI1407
aI2615
aI14529
aI10781
aI20001
aI82713
asS'd'
I9957
sa(dp19914
S'a'
I35996
sS's'
I17
sS'm'
(lp19915
I1
aI3
aI7
aI7
aI9
aI29
aI25
aI55
aI53
aI423
aI1711
aI2871
aI2675
aI421
aI31703
aI57099
aI2955
asS'd'
I9958
sa(dp19916
S'a'
I36005
sS's'
I17
sS'm'
(lp19917
I1
aI3
aI1
aI7
aI31
aI17
aI113
aI83
aI387
aI611
aI1815
aI2137
aI3453
aI4409
aI20377
aI60263
aI81205
asS'd'
I9959
sa(dp19918
S'a'
I36012
sS's'
I17
sS'm'
(lp19919
I1
aI1
aI5
aI3
aI11
aI1
aI7
aI225
aI367
aI267
aI95
aI939
aI3801
aI2619
aI1207
aI62695
aI116407
asS'd'
I9960
sa(dp19920
S'a'
I36015
sS's'
I17
sS'm'
(lp19921
I1
aI3
aI3
aI9
aI5
aI39
aI85
aI45
aI247
aI483
aI491
aI865
aI3493
aI8243
aI8411
aI26449
aI50473
asS'd'
I9961
sa(dp19922
S'a'
I36030
sS's'
I17
sS'm'
(lp19923
I1
aI3
aI3
aI9
aI1
aI53
aI23
aI127
aI13
aI529
aI1925
aI2629
aI3451
aI15073
aI16075
aI29909
aI34101
asS'd'
I9962
sa(dp19924
S'a'
I36035
sS's'
I17
sS'm'
(lp19925
I1
aI3
aI1
aI11
aI1
aI9
aI125
aI57
aI79
aI633
aI979
aI3843
aI325
aI883
aI7769
aI40155
aI104057
asS'd'
I9963
sa(dp19926
S'a'
I36042
sS's'
I17
sS'm'
(lp19927
I1
aI1
aI7
aI13
aI23
aI53
aI27
aI157
aI493
aI901
aI1077
aI1079
aI1327
aI15903
aI20603
aI64377
aI103335
asS'd'
I9964
sa(dp19928
S'a'
I36047
sS's'
I17
sS'm'
(lp19929
I1
aI3
aI3
aI3
aI3
aI35
aI37
aI167
aI73
aI301
aI385
aI1045
aI6913
aI2269
aI22491
aI19735
aI70125
asS'd'
I9965
sa(dp19930
S'a'
I36049
sS's'
I17
sS'm'
(lp19931
I1
aI1
aI1
aI11
aI5
aI23
aI23
aI85
aI267
aI845
aI207
aI77
aI1245
aI16209
aI25579
aI12417
aI48723
asS'd'
I9966
sa(dp19932
S'a'
I36059
sS's'
I17
sS'm'
(lp19933
I1
aI1
aI5
aI15
aI11
aI17
aI43
aI83
aI373
aI1005
aI541
aI115
aI163
aI2165
aI8181
aI35839
aI44471
asS'd'
I9967
sa(dp19934
S'a'
I36071
sS's'
I17
sS'm'
(lp19935
I1
aI3
aI5
aI7
aI27
aI41
aI101
aI13
aI213
aI235
aI2037
aI2179
aI2121
aI4481
aI8127
aI20011
aI3981
asS'd'
I9968
sa(dp19936
S'a'
I36080
sS's'
I17
sS'm'
(lp19937
I1
aI1
aI5
aI11
aI7
aI43
aI59
aI129
aI127
aI387
aI489
aI1985
aI623
aI13307
aI19765
aI62155
aI93271
asS'd'
I9969
sa(dp19938
S'a'
I36085
sS's'
I17
sS'm'
(lp19939
I1
aI1
aI7
aI5
aI23
aI63
aI23
aI177
aI211
aI233
aI101
aI1809
aI7411
aI8003
aI25101
aI32601
aI75071
asS'd'
I9970
sa(dp19940
S'a'
I36097
sS's'
I17
sS'm'
(lp19941
I1
aI1
aI1
aI11
aI3
aI25
aI9
aI91
aI459
aI611
aI867
aI3639
aI5457
aI9101
aI15333
aI40069
aI67723
asS'd'
I9971
sa(dp19942
S'a'
I36110
sS's'
I17
sS'm'
(lp19943
I1
aI3
aI7
aI5
aI3
aI29
aI111
aI75
aI459
aI195
aI1405
aI2281
aI6085
aI4425
aI29061
aI57335
aI87449
asS'd'
I9972
sa(dp19944
S'a'
I36115
sS's'
I17
sS'm'
(lp19945
I1
aI3
aI7
aI11
aI21
aI45
aI53
aI81
aI77
aI863
aI1901
aI3355
aI5253
aI10897
aI26289
aI48399
aI26877
asS'd'
I9973
sa(dp19946
S'a'
I36118
sS's'
I17
sS'm'
(lp19947
I1
aI3
aI3
aI13
aI21
aI37
aI69
aI87
aI259
aI101
aI1203
aI167
aI6229
aI145
aI9355
aI15347
aI68047
asS'd'
I9974
sa(dp19948
S'a'
I36124
sS's'
I17
sS'm'
(lp19949
I1
aI1
aI3
aI1
aI31
aI1
aI15
aI229
aI429
aI915
aI929
aI381
aI1857
aI8441
aI22207
aI47071
aI127853
asS'd'
I9975
sa(dp19950
S'a'
I36137
sS's'
I17
sS'm'
(lp19951
I1
aI3
aI7
aI3
aI15
aI9
aI13
aI161
aI173
aI573
aI405
aI3253
aI7331
aI13965
aI3061
aI40687
aI130185
asS'd'
I9976
sa(dp19952
S'a'
I36138
sS's'
I17
sS'm'
(lp19953
I1
aI3
aI5
aI5
aI29
aI29
aI9
aI115
aI393
aI377
aI909
aI321
aI2861
aI9881
aI17863
aI52033
aI55133
asS'd'
I9977
sa(dp19954
S'a'
I36155
sS's'
I17
sS'm'
(lp19955
I1
aI1
aI7
aI7
aI27
aI53
aI101
aI213
aI199
aI301
aI1995
aI2549
aI5037
aI13639
aI18423
aI23547
aI79359
asS'd'
I9978
sa(dp19956
S'a'
I36160
sS's'
I17
sS'm'
(lp19957
I1
aI3
aI1
aI7
aI21
aI51
aI29
aI151
aI301
aI665
aI571
aI53
aI2637
aI7229
aI12517
aI33647
aI49413
asS'd'
I9979
sa(dp19958
S'a'
I36189
sS's'
I17
sS'm'
(lp19959
I1
aI3
aI3
aI13
aI13
aI49
aI49
aI131
aI325
aI273
aI1127
aI2981
aI2365
aI14287
aI23185
aI26915
aI81755
asS'd'
I9980
sa(dp19960
S'a'
I36190
sS's'
I17
sS'm'
(lp19961
I1
aI1
aI5
aI3
aI17
aI45
aI25
aI79
aI37
aI265
aI1205
aI1805
aI6707
aI11525
aI16473
aI39525
aI9571
asS'd'
I9981
sa(dp19962
S'a'
I36203
sS's'
I17
sS'm'
(lp19963
I1
aI3
aI3
aI15
aI9
aI43
aI55
aI101
aI469
aI939
aI365
aI3443
aI5759
aI4751
aI28893
aI46727
aI74569
asS'd'
I9982
sa(dp19964
S'a'
I36211
sS's'
I17
sS'm'
(lp19965
I1
aI3
aI7
aI9
aI5
aI33
aI11
aI201
aI263
aI227
aI1475
aI2795
aI1489
aI11129
aI18053
aI31009
aI73105
asS'd'
I9983
sa(dp19966
S'a'
I36218
sS's'
I17
sS'm'
(lp19967
I1
aI3
aI5
aI5
aI5
aI25
aI41
aI151
aI393
aI237
aI2017
aI3811
aI953
aI13835
aI28761
aI22439
aI76355
asS'd'
I9984
sa(dp19968
S'a'
I36230
sS's'
I17
sS'm'
(lp19969
I1
aI1
aI5
aI13
aI21
aI37
aI29
aI11
aI289
aI67
aI1317
aI511
aI685
aI15227
aI8731
aI15039
aI79491
asS'd'
I9985
sa(dp19970
S'a'
I36241
sS's'
I17
sS'm'
(lp19971
I1
aI3
aI1
aI9
aI31
aI59
aI123
aI169
aI473
aI139
aI575
aI1057
aI3213
aI8213
aI21845
aI28123
aI105335
asS'd'
I9986
sa(dp19972
S'a'
I36244
sS's'
I17
sS'm'
(lp19973
I1
aI1
aI1
aI5
aI21
aI47
aI23
aI121
aI403
aI5
aI1457
aI2137
aI569
aI9267
aI6367
aI6991
aI3113
asS'd'
I9987
sa(dp19974
S'a'
I36253
sS's'
I17
sS'm'
(lp19975
I1
aI3
aI3
aI7
aI13
aI7
aI25
aI215
aI81
aI1003
aI2041
aI1317
aI3913
aI14705
aI30551
aI50889
aI83441
asS'd'
I9988
sa(dp19976
S'a'
I36257
sS's'
I17
sS'm'
(lp19977
I1
aI3
aI3
aI3
aI13
aI17
aI63
aI229
aI83
aI901
aI953
aI2603
aI4685
aI6961
aI7519
aI52441
aI33223
asS'd'
I9989
sa(dp19978
S'a'
I36264
sS's'
I17
sS'm'
(lp19979
I1
aI3
aI7
aI5
aI7
aI57
aI65
aI73
aI243
aI531
aI261
aI2517
aI4083
aI5889
aI22913
aI49603
aI67135
asS'd'
I9990
sa(dp19980
S'a'
I36272
sS's'
I17
sS'm'
(lp19981
I1
aI3
aI5
aI11
aI15
aI47
aI81
aI83
aI35
aI1021
aI1313
aI1109
aI5103
aI5469
aI18149
aI15307
aI34939
asS'd'
I9991
sa(dp19982
S'a'
I36290
sS's'
I17
sS'm'
(lp19983
I1
aI3
aI7
aI5
aI21
aI13
aI105
aI157
aI435
aI23
aI931
aI3565
aI1
aI4987
aI8829
aI7327
aI51049
asS'd'
I9992
sa(dp19984
S'a'
I36292
sS's'
I17
sS'm'
(lp19985
I1
aI1
aI3
aI11
aI29
aI9
aI59
aI49
aI261
aI1009
aI1953
aI2683
aI8125
aI10937
aI16683
aI36013
aI5967
asS'd'
I9993
sa(dp19986
S'a'
I36301
sS's'
I17
sS'm'
(lp19987
I1
aI1
aI1
aI1
aI19
aI29
aI57
aI9
aI307
aI457
aI675
aI3023
aI495
aI15257
aI7945
aI10449
aI30155
asS'd'
I9994
sa(dp19988
S'a'
I36309
sS's'
I17
sS'm'
(lp19989
I1
aI1
aI7
aI13
aI25
aI9
aI51
aI135
aI491
aI205
aI1715
aI3253
aI1031
aI4137
aI14885
aI39925
aI6061
asS'd'
I9995
sa(dp19990
S'a'
I36313
sS's'
I17
sS'm'
(lp19991
I1
aI1
aI7
aI7
aI3
aI13
aI111
aI91
aI469
aI133
aI1221
aI1035
aI919
aI3697
aI26387
aI41675
aI487
asS'd'
I9996
sa(dp19992
S'a'
I36316
sS's'
I17
sS'm'
(lp19993
I1
aI1
aI3
aI1
aI19
aI53
aI11
aI113
aI245
aI747
aI189
aI4051
aI87
aI1767
aI3595
aI10259
aI100097
asS'd'
I9997
sa(dp19994
S'a'
I36319
sS's'
I17
sS'm'
(lp19995
I1
aI1
aI5
aI3
aI23
aI49
aI31
aI47
aI341
aI1019
aI723
aI2353
aI6191
aI3809
aI3297
aI39443
aI73529
asS'd'
I9998
sa(dp19996
S'a'
I36330
sS's'
I17
sS'm'
(lp19997
I1
aI3
aI3
aI9
aI25
aI27
aI123
aI49
aI51
aI85
aI1063
aI2633
aI6549
aI14493
aI7367
aI3557
aI60651
asS'd'
I9999
sa(dp19998
S'a'
I36335
sS's'
I17
sS'm'
(lp19999
I1
aI3
aI7
aI5
aI13
aI27
aI127
aI65
aI115
aI731
aI1147
aI283
aI91
aI14205
aI2457
aI57083
aI35815
asS'd'
I10000
sa(dp20000
S'a'
I36347
sS's'
I17
sS'm'
(lp20001
I1
aI3
aI3
aI3
aI25
aI63
aI99
aI249
aI25
aI951
aI733
aI3621
aI7139
aI14223
aI23641
aI20287
aI30743
asS'd'
I10001
sa(dp20002
S'a'
I36353
sS's'
I17
sS'm'
(lp20003
I1
aI3
aI3
aI7
aI21
aI23
aI83
aI207
aI235
aI467
aI1857
aI2661
aI1391
aI10097
aI12297
aI54825
aI5035
asS'd'
I10002
sa(dp20004
S'a'
I36356
sS's'
I17
sS'm'
(lp20005
I1
aI1
aI5
aI3
aI31
aI17
aI77
aI9
aI215
aI553
aI989
aI3643
aI729
aI2057
aI32053
aI50305
aI5499
asS'd'
I10003
sa(dp20006
S'a'
I36368
sS's'
I17
sS'm'
(lp20007
I1
aI1
aI7
aI1
aI23
aI5
aI111
aI195
aI431
aI947
aI403
aI1781
aI943
aI15073
aI67
aI52225
aI98987
asS'd'
I10004
sa(dp20008
S'a'
I36374
sS's'
I17
sS'm'
(lp20009
I1
aI1
aI5
aI11
aI23
aI1
aI41
aI33
aI457
aI767
aI275
aI801
aI5119
aI3781
aI14805
aI52789
aI41775
asS'd'
I10005
sa(dp20010
S'a'
I36377
sS's'
I17
sS'm'
(lp20011
I1
aI1
aI5
aI3
aI9
aI53
aI15
aI183
aI281
aI691
aI165
aI3277
aI7673
aI1509
aI16605
aI53799
aI100185
asS'd'
I10006
sa(dp20012
S'a'
I36384
sS's'
I17
sS'm'
(lp20013
I1
aI3
aI5
aI11
aI19
aI45
aI29
aI159
aI167
aI67
aI1259
aI879
aI7787
aI8855
aI24153
aI42667
aI102855
asS'd'
I10007
sa(dp20014
S'a'
I36407
sS's'
I17
sS'm'
(lp20015
I1
aI1
aI7
aI13
aI31
aI19
aI43
aI133
aI295
aI287
aI1985
aI2451
aI2297
aI3853
aI22401
aI27659
aI11149
asS'd'
I10008
sa(dp20016
S'a'
I36413
sS's'
I17
sS'm'
(lp20017
I1
aI1
aI7
aI13
aI31
aI39
aI125
aI21
aI173
aI103
aI1119
aI3739
aI6467
aI2113
aI4465
aI26537
aI129949
asS'd'
I10009
sa(dp20018
S'a'
I36419
sS's'
I17
sS'm'
(lp20019
I1
aI1
aI5
aI15
aI21
aI47
aI35
aI125
aI199
aI335
aI421
aI31
aI185
aI12769
aI30659
aI33427
aI106981
asS'd'
I10010
sa(dp20020
S'a'
I36425
sS's'
I17
sS'm'
(lp20021
I1
aI3
aI5
aI13
aI25
aI35
aI53
aI253
aI325
aI921
aI1705
aI2735
aI6437
aI2287
aI20479
aI61107
aI91453
asS'd'
I10011
sa(dp20022
S'a'
I36426
sS's'
I17
sS'm'
(lp20023
I1
aI3
aI7
aI13
aI25
aI63
aI83
aI183
aI5
aI401
aI329
aI525
aI3141
aI393
aI30469
aI16529
aI9605
asS'd'
I10012
sa(dp20024
S'a'
I36446
sS's'
I17
sS'm'
(lp20025
I1
aI3
aI3
aI13
aI19
aI23
aI15
aI85
aI323
aI545
aI149
aI3645
aI6269
aI15595
aI18453
aI39
aI128169
asS'd'
I10013
sa(dp20026
S'a'
I36461
sS's'
I17
sS'm'
(lp20027
I1
aI3
aI7
aI15
aI17
aI5
aI61
aI61
aI91
aI353
aI1039
aI2959
aI4147
aI13205
aI12599
aI53281
aI39509
asS'd'
I10014
sa(dp20028
S'a'
I36467
sS's'
I17
sS'm'
(lp20029
I1
aI1
aI3
aI7
aI21
aI9
aI97
aI111
aI249
aI775
aI845
aI1789
aI667
aI489
aI6689
aI29217
aI56527
asS'd'
I10015
sa(dp20030
S'a'
I36474
sS's'
I17
sS'm'
(lp20031
I1
aI3
aI5
aI7
aI11
aI5
aI59
aI219
aI29
aI803
aI923
aI3861
aI7953
aI8969
aI1819
aI43501
aI20513
asS'd'
I10016
sa(dp20032
S'a'
I36480
sS's'
I17
sS'm'
(lp20033
I1
aI1
aI5
aI11
aI7
aI53
aI63
aI231
aI193
aI293
aI1467
aI1409
aI6397
aI13237
aI15903
aI19271
aI66257
asS'd'
I10017
sa(dp20034
S'a'
I36486
sS's'
I17
sS'm'
(lp20035
I1
aI3
aI1
aI15
aI23
aI15
aI37
aI123
aI189
aI63
aI1121
aI751
aI6711
aI10095
aI6493
aI40709
aI47641
asS'd'
I10018
sa(dp20036
S'a'
I36489
sS's'
I17
sS'm'
(lp20037
I1
aI3
aI7
aI3
aI23
aI59
aI99
aI183
aI249
aI479
aI771
aI1087
aI7979
aI409
aI4819
aI4337
aI33345
asS'd'
I10019
sa(dp20038
S'a'
I36495
sS's'
I17
sS'm'
(lp20039
I1
aI1
aI5
aI1
aI17
aI7
aI15
aI167
aI305
aI411
aI1429
aI3127
aI23
aI9123
aI7185
aI44405
aI114841
asS'd'
I10020
sa(dp20040
S'a'
I36525
sS's'
I17
sS'm'
(lp20041
I1
aI1
aI5
aI11
aI3
aI29
aI29
aI31
aI399
aI777
aI251
aI1841
aI3607
aI211
aI23543
aI29111
aI54565
asS'd'
I10021
sa(dp20042
S'a'
I36526
sS's'
I17
sS'm'
(lp20043
I1
aI3
aI3
aI9
aI27
aI33
aI79
aI27
aI469
aI67
aI1327
aI183
aI5783
aI10039
aI13165
aI20443
aI4913
asS'd'
I10022
sa(dp20044
S'a'
I36533
sS's'
I17
sS'm'
(lp20045
I1
aI3
aI7
aI15
aI21
aI23
aI5
aI227
aI141
aI1021
aI69
aI3347
aI7221
aI13837
aI20921
aI20525
aI32567
asS'd'
I10023
sa(dp20046
S'a'
I36534
sS's'
I17
sS'm'
(lp20047
I1
aI1
aI5
aI5
aI25
aI53
aI73
aI111
aI319
aI311
aI1597
aI1809
aI5343
aI13963
aI6613
aI14471
aI53871
asS'd'
I10024
sa(dp20048
S'a'
I36540
sS's'
I17
sS'm'
(lp20049
I1
aI3
aI3
aI1
aI15
aI57
aI47
aI205
aI53
aI471
aI185
aI273
aI8077
aI5031
aI31195
aI30859
aI15979
asS'd'
I10025
sa(dp20050
S'a'
I36555
sS's'
I17
sS'm'
(lp20051
I1
aI1
aI3
aI5
aI23
aI15
aI87
aI211
aI83
aI265
aI1629
aI2979
aI69
aI12559
aI30455
aI36363
aI61461
asS'd'
I10026
sa(dp20052
S'a'
I36563
sS's'
I17
sS'm'
(lp20053
I1
aI1
aI7
aI7
aI1
aI47
aI5
aI199
aI95
aI17
aI57
aI1887
aI6847
aI9501
aI21361
aI57763
aI77069
asS'd'
I10027
sa(dp20054
S'a'
I36565
sS's'
I17
sS'm'
(lp20055
I1
aI1
aI3
aI5
aI9
aI15
aI15
aI149
aI141
aI605
aI639
aI2197
aI7237
aI5753
aI9415
aI4677
aI129947
asS'd'
I10028
sa(dp20056
S'a'
I36588
sS's'
I17
sS'm'
(lp20057
I1
aI3
aI7
aI1
aI7
aI9
aI29
aI249
aI275
aI461
aI1667
aI4093
aI5763
aI3205
aI24079
aI11883
aI86455
asS'd'
I10029
sa(dp20058
S'a'
I36593
sS's'
I17
sS'm'
(lp20059
I1
aI1
aI3
aI5
aI15
aI39
aI117
aI145
aI153
aI671
aI1819
aI111
aI3607
aI12279
aI4927
aI63759
aI42905
asS'd'
I10030
sa(dp20060
S'a'
I36596
sS's'
I17
sS'm'
(lp20061
I1
aI1
aI1
aI5
aI31
aI5
aI35
aI183
aI189
aI839
aI1811
aI1877
aI6545
aI11373
aI27947
aI27183
aI29857
asS'd'
I10031
sa(dp20062
S'a'
I36606
sS's'
I17
sS'm'
(lp20063
I1
aI3
aI5
aI7
aI29
aI47
aI3
aI183
aI511
aI145
aI1953
aI3419
aI6385
aI7745
aI12823
aI59783
aI69399
asS'd'
I10032
sa(dp20064
S'a'
I36614
sS's'
I17
sS'm'
(lp20065
I1
aI3
aI5
aI9
aI5
aI39
aI85
aI145
aI33
aI899
aI1009
aI2035
aI6145
aI3855
aI20583
aI4329
aI95231
asS'd'
I10033
sa(dp20066
S'a'
I36626
sS's'
I17
sS'm'
(lp20067
I1
aI1
aI3
aI3
aI15
aI61
aI85
aI181
aI247
aI705
aI413
aI1633
aI7489
aI1785
aI30397
aI42851
aI80197
asS'd'
I10034
sa(dp20068
S'a'
I36628
sS's'
I17
sS'm'
(lp20069
I1
aI3
aI3
aI13
aI23
aI11
aI3
aI97
aI307
aI183
aI113
aI3881
aI7455
aI8327
aI6749
aI23977
aI101629
asS'd'
I10035
sa(dp20070
S'a'
I36641
sS's'
I17
sS'm'
(lp20071
I1
aI1
aI7
aI13
aI1
aI23
aI59
aI219
aI125
aI789
aI1401
aI707
aI6915
aI6275
aI25813
aI46595
aI54119
asS'd'
I10036
sa(dp20072
S'a'
I36642
sS's'
I17
sS'm'
(lp20073
I1
aI3
aI7
aI9
aI5
aI7
aI37
aI33
aI165
aI181
aI833
aI1993
aI4541
aI5799
aI23323
aI39825
aI44575
asS'd'
I10037
sa(dp20074
S'a'
I36651
sS's'
I17
sS'm'
(lp20075
I1
aI3
aI1
aI13
aI13
aI43
aI69
aI219
aI437
aI521
aI503
aI2293
aI3607
aI6845
aI22583
aI291
aI65645
asS'd'
I10038
sa(dp20076
S'a'
I36653
sS's'
I17
sS'm'
(lp20077
I1
aI1
aI7
aI9
aI29
aI13
aI123
aI67
aI191
aI933
aI1875
aI1223
aI5525
aI13797
aI29771
aI58191
aI84469
asS'd'
I10039
sa(dp20078
S'a'
I36673
sS's'
I17
sS'm'
(lp20079
I1
aI1
aI7
aI7
aI3
aI57
aI101
aI69
aI23
aI239
aI1023
aI3289
aI1541
aI6245
aI23379
aI161
aI61155
asS'd'
I10040
sa(dp20080
S'a'
I36676
sS's'
I17
sS'm'
(lp20081
I1
aI3
aI7
aI13
aI25
aI33
aI49
aI145
aI487
aI681
aI451
aI1719
aI109
aI16273
aI20009
aI3003
aI115815
asS'd'
I10041
sa(dp20082
S'a'
I36679
sS's'
I17
sS'm'
(lp20083
I1
aI1
aI5
aI11
aI11
aI59
aI41
aI133
aI303
aI469
aI1975
aI847
aI5291
aI13947
aI8759
aI8533
aI25099
asS'd'
I10042
sa(dp20084
S'a'
I36694
sS's'
I17
sS'm'
(lp20085
I1
aI1
aI1
aI1
aI29
aI31
aI53
aI11
aI239
aI57
aI1627
aI1247
aI1577
aI3269
aI20751
aI4627
aI40499
asS'd'
I10043
sa(dp20086
S'a'
I36698
sS's'
I17
sS'm'
(lp20087
I1
aI3
aI7
aI15
aI1
aI1
aI51
aI39
aI383
aI203
aI1841
aI3867
aI4975
aI9937
aI1863
aI52611
aI83189
asS'd'
I10044
sa(dp20088
S'a'
I36704
sS's'
I17
sS'm'
(lp20089
I1
aI3
aI7
aI7
aI13
aI59
aI15
aI217
aI355
aI945
aI1317
aI815
aI2413
aI10985
aI30647
aI37745
aI126553
asS'd'
I10045
sa(dp20090
S'a'
I36714
sS's'
I17
sS'm'
(lp20091
I1
aI1
aI3
aI11
aI7
aI29
aI101
aI137
aI97
aI119
aI927
aI3269
aI6977
aI4253
aI10741
aI61907
aI122815
asS'd'
I10046
sa(dp20092
S'a'
I36721
sS's'
I17
sS'm'
(lp20093
I1
aI3
aI3
aI1
aI29
aI5
aI49
aI137
aI411
aI349
aI905
aI2481
aI4961
aI4513
aI29409
aI19503
aI77915
asS'd'
I10047
sa(dp20094
S'a'
I36722
sS's'
I17
sS'm'
(lp20095
I1
aI1
aI7
aI13
aI29
aI59
aI93
aI61
aI393
aI29
aI257
aI3601
aI6281
aI5105
aI17339
aI53827
aI83137
asS'd'
I10048
sa(dp20096
S'a'
I36727
sS's'
I17
sS'm'
(lp20097
I1
aI1
aI1
aI13
aI5
aI23
aI61
aI7
aI51
aI161
aI737
aI1549
aI6021
aI3385
aI5539
aI21261
aI69995
asS'd'
I10049
sa(dp20098
S'a'
I36749
sS's'
I17
sS'm'
(lp20099
I1
aI1
aI1
aI15
aI31
aI1
aI21
aI113
aI481
aI7
aI175
aI717
aI1593
aI5937
aI12347
aI51835
aI66649
asS'd'
I10050
sa(dp20100
S'a'
I36758
sS's'
I17
sS'm'
(lp20101
I1
aI1
aI3
aI7
aI9
aI51
aI9
aI199
aI39
aI607
aI1157
aI3913
aI7767
aI14195
aI28721
aI27655
aI34709
asS'd'
I10051
sa(dp20102
S'a'
I36761
sS's'
I17
sS'm'
(lp20103
I1
aI3
aI5
aI5
aI1
aI15
aI49
aI33
aI441
aI721
aI1749
aI1497
aI2023
aI8351
aI12641
aI11861
aI78545
asS'd'
I10052
sa(dp20104
S'a'
I36771
sS's'
I17
sS'm'
(lp20105
I1
aI3
aI1
aI7
aI7
aI17
aI103
aI113
aI243
aI25
aI889
aI1419
aI3163
aI12401
aI22459
aI39037
aI101719
asS'd'
I10053
sa(dp20106
S'a'
I36788
sS's'
I17
sS'm'
(lp20107
I1
aI1
aI7
aI11
aI17
aI45
aI121
aI215
aI3
aI409
aI1871
aI2149
aI4249
aI5071
aI14277
aI55869
aI91233
asS'd'
I10054
sa(dp20108
S'a'
I36797
sS's'
I17
sS'm'
(lp20109
I1
aI1
aI3
aI7
aI19
aI31
aI47
aI241
aI175
aI749
aI1709
aI355
aI6037
aI10555
aI24107
aI64683
aI42673
asS'd'
I10055
sa(dp20110
S'a'
I36805
sS's'
I17
sS'm'
(lp20111
I1
aI3
aI7
aI11
aI5
aI21
aI105
aI137
aI307
aI101
aI417
aI1903
aI1027
aI10257
aI27767
aI9755
aI92105
asS'd'
I10056
sa(dp20112
S'a'
I36830
sS's'
I17
sS'm'
(lp20113
I1
aI1
aI3
aI13
aI9
aI59
aI11
aI63
aI295
aI923
aI401
aI1471
aI3517
aI7761
aI28855
aI11525
aI72455
asS'd'
I10057
sa(dp20114
S'a'
I36833
sS's'
I17
sS'm'
(lp20115
I1
aI1
aI7
aI15
aI31
aI51
aI77
aI29
aI323
aI579
aI1313
aI3441
aI2903
aI1683
aI20605
aI8185
aI29753
asS'd'
I10058
sa(dp20116
S'a'
I36839
sS's'
I17
sS'm'
(lp20117
I1
aI1
aI5
aI15
aI11
aI59
aI119
aI109
aI233
aI1001
aI1527
aI2709
aI73
aI5311
aI18313
aI27155
aI85999
asS'd'
I10059
sa(dp20118
S'a'
I36843
sS's'
I17
sS'm'
(lp20119
I1
aI3
aI1
aI5
aI9
aI59
aI105
aI93
aI213
aI401
aI839
aI3225
aI3263
aI13501
aI2413
aI60367
aI121281
asS'd'
I10060
sa(dp20120
S'a'
I36860
sS's'
I17
sS'm'
(lp20121
I1
aI1
aI7
aI3
aI19
aI25
aI75
aI27
aI325
aI435
aI527
aI1465
aI3601
aI5785
aI6135
aI32841
aI60129
asS'd'
I10061
sa(dp20122
S'a'
I36866
sS's'
I17
sS'm'
(lp20123
I1
aI1
aI3
aI7
aI31
aI19
aI37
aI157
aI189
aI51
aI869
aI2963
aI5269
aI9151
aI14845
aI30441
aI89685
asS'd'
I10062
sa(dp20124
S'a'
I36871
sS's'
I17
sS'm'
(lp20125
I1
aI3
aI3
aI9
aI17
aI51
aI23
aI177
aI417
aI255
aI1739
aI3085
aI7811
aI15177
aI25433
aI38487
aI51021
asS'd'
I10063
sa(dp20126
S'a'
I36875
sS's'
I17
sS'm'
(lp20127
I1
aI1
aI3
aI7
aI27
aI1
aI45
aI235
aI59
aI491
aI1327
aI3967
aI7585
aI4313
aI29669
aI47193
aI89427
asS'd'
I10064
sa(dp20128
S'a'
I36877
sS's'
I17
sS'm'
(lp20129
I1
aI1
aI3
aI9
aI19
aI5
aI27
aI63
aI263
aI593
aI1599
aI1311
aI1029
aI603
aI25291
aI51391
aI98915
asS'd'
I10065
sa(dp20130
S'a'
I36880
sS's'
I17
sS'm'
(lp20131
I1
aI3
aI3
aI15
aI11
aI7
aI97
aI99
aI263
aI155
aI437
aI3849
aI2665
aI3371
aI8179
aI51883
aI3601
asS'd'
I10066
sa(dp20132
S'a'
I36892
sS's'
I17
sS'm'
(lp20133
I1
aI1
aI3
aI15
aI7
aI35
aI37
aI149
aI251
aI619
aI1423
aI553
aI4453
aI16365
aI22543
aI6951
aI34655
asS'd'
I10067
sa(dp20134
S'a'
I36911
sS's'
I17
sS'm'
(lp20135
I1
aI3
aI3
aI11
aI15
aI21
aI95
aI143
aI31
aI425
aI179
aI2383
aI4799
aI7655
aI26945
aI9273
aI103469
asS'd'
I10068
sa(dp20136
S'a'
I36914
sS's'
I17
sS'm'
(lp20137
I1
aI3
aI1
aI9
aI13
aI49
aI3
aI117
aI361
aI459
aI227
aI2067
aI4909
aI13461
aI22505
aI10259
aI59697
asS'd'
I10069
sa(dp20138
S'a'
I36916
sS's'
I17
sS'm'
(lp20139
I1
aI1
aI7
aI7
aI7
aI23
aI67
aI217
aI313
aI965
aI1747
aI995
aI579
aI6217
aI8915
aI49329
aI851
asS'd'
I10070
sa(dp20140
S'a'
I36923
sS's'
I17
sS'm'
(lp20141
I1
aI1
aI3
aI1
aI17
aI19
aI7
aI99
aI281
aI207
aI1685
aI2401
aI967
aI9399
aI28741
aI28839
aI6003
asS'd'
I10071
sa(dp20142
S'a'
I36940
sS's'
I17
sS'm'
(lp20143
I1
aI3
aI3
aI5
aI31
aI61
aI105
aI251
aI499
aI319
aI1167
aI2203
aI1195
aI2663
aI11797
aI12981
aI125523
asS'd'
I10072
sa(dp20144
S'a'
I36943
sS's'
I17
sS'm'
(lp20145
I1
aI3
aI1
aI5
aI23
aI19
aI99
aI101
aI85
aI837
aI501
aI2737
aI4051
aI2413
aI9275
aI38995
aI21633
asS'd'
I10073
sa(dp20146
S'a'
I36948
sS's'
I17
sS'm'
(lp20147
I1
aI3
aI7
aI13
aI17
aI17
aI119
aI75
aI281
aI527
aI1477
aI1515
aI7765
aI5573
aI10143
aI6219
aI57817
asS'd'
I10074
sa(dp20148
S'a'
I36957
sS's'
I17
sS'm'
(lp20149
I1
aI1
aI5
aI11
aI19
aI35
aI85
aI171
aI107
aI905
aI1395
aI1199
aI7345
aI15719
aI14021
aI47425
aI36081
asS'd'
I10075
sa(dp20150
S'a'
I36958
sS's'
I17
sS'm'
(lp20151
I1
aI1
aI3
aI9
aI9
aI63
aI109
aI15
aI323
aI73
aI1541
aI2227
aI5197
aI12617
aI23379
aI53415
aI105291
asS'd'
I10076
sa(dp20152
S'a'
I36967
sS's'
I17
sS'm'
(lp20153
I1
aI3
aI3
aI5
aI5
aI41
aI85
aI99
aI3
aI895
aI1383
aI3627
aI3897
aI1893
aI23673
aI56501
aI78411
asS'd'
I10077
sa(dp20154
S'a'
I36974
sS's'
I17
sS'm'
(lp20155
I1
aI1
aI7
aI1
aI25
aI27
aI45
aI185
aI475
aI577
aI1619
aI727
aI1407
aI2383
aI9215
aI55295
aI27349
asS'd'
I10078
sa(dp20156
S'a'
I36981
sS's'
I17
sS'm'
(lp20157
I1
aI3
aI7
aI11
aI3
aI51
aI53
aI53
aI399
aI711
aI1075
aI511
aI5369
aI10777
aI14419
aI63217
aI130181
asS'd'
I10079
sa(dp20158
S'a'
I37001
sS's'
I17
sS'm'
(lp20159
I1
aI1
aI7
aI13
aI25
aI19
aI107
aI71
aI151
aI73
aI735
aI3837
aI5307
aI10229
aI10529
aI9989
aI111925
asS'd'
I10080
sa(dp20160
S'a'
I37012
sS's'
I17
sS'm'
(lp20161
I1
aI1
aI1
aI15
aI19
aI59
aI65
aI77
aI465
aI957
aI1085
aI1359
aI3959
aI15823
aI6273
aI12565
aI126167
asS'd'
I10081
sa(dp20162
S'a'
I37015
sS's'
I17
sS'm'
(lp20163
I1
aI1
aI5
aI5
aI31
aI53
aI23
aI173
aI407
aI795
aI41
aI3275
aI1953
aI13673
aI26625
aI33477
aI14149
asS'd'
I10082
sa(dp20164
S'a'
I37019
sS's'
I17
sS'm'
(lp20165
I1
aI1
aI7
aI7
aI1
aI11
aI121
aI139
aI77
aI321
aI1939
aI2597
aI621
aI9579
aI11629
aI13119
aI30505
asS'd'
I10083
sa(dp20166
S'a'
I37035
sS's'
I17
sS'm'
(lp20167
I1
aI1
aI1
aI5
aI3
aI33
aI45
aI127
aI169
aI581
aI1521
aI1019
aI6489
aI1069
aI2469
aI40255
aI66619
asS'd'
I10084
sa(dp20168
S'a'
I37040
sS's'
I17
sS'm'
(lp20169
I1
aI3
aI7
aI5
aI29
aI47
aI7
aI245
aI459
aI417
aI1027
aI857
aI4905
aI11255
aI3267
aI9491
aI78013
asS'd'
I10085
sa(dp20170
S'a'
I37063
sS's'
I17
sS'm'
(lp20171
I1
aI3
aI5
aI9
aI25
aI49
aI61
aI215
aI19
aI731
aI303
aI1001
aI6031
aI3705
aI7797
aI31957
aI119383
asS'd'
I10086
sa(dp20172
S'a'
I37064
sS's'
I17
sS'm'
(lp20173
I1
aI3
aI5
aI5
aI1
aI9
aI37
aI187
aI235
aI453
aI963
aI2833
aI3501
aI605
aI2763
aI41215
aI93547
asS'd'
I10087
sa(dp20174
S'a'
I37069
sS's'
I17
sS'm'
(lp20175
I1
aI3
aI1
aI1
aI21
aI3
aI41
aI53
aI425
aI687
aI1051
aI2365
aI7835
aI3981
aI5557
aI61993
aI127417
asS'd'
I10088
sa(dp20176
S'a'
I37077
sS's'
I17
sS'm'
(lp20177
I1
aI3
aI3
aI7
aI13
aI61
aI41
aI189
aI261
aI163
aI1931
aI1803
aI2379
aI16379
aI25453
aI17911
aI123431
asS'd'
I10089
sa(dp20178
S'a'
I37093
sS's'
I17
sS'm'
(lp20179
I1
aI1
aI7
aI15
aI23
aI21
aI95
aI7
aI27
aI897
aI721
aI3917
aI7971
aI4643
aI5223
aI46583
aI32453
asS'd'
I10090
sa(dp20180
S'a'
I37097
sS's'
I17
sS'm'
(lp20181
I1
aI1
aI7
aI7
aI1
aI25
aI83
aI109
aI223
aI573
aI533
aI449
aI6477
aI10719
aI28705
aI8283
aI94963
asS'd'
I10091
sa(dp20182
S'a'
I37106
sS's'
I17
sS'm'
(lp20183
I1
aI1
aI5
aI13
aI21
aI45
aI63
aI31
aI21
aI223
aI31
aI1249
aI425
aI7199
aI11539
aI7731
aI44333
asS'd'
I10092
sa(dp20184
S'a'
I37115
sS's'
I17
sS'm'
(lp20185
I1
aI1
aI5
aI15
aI29
aI5
aI87
aI215
aI287
aI567
aI297
aI451
aI5867
aI15511
aI1005
aI57469
aI87257
asS'd'
I10093
sa(dp20186
S'a'
I37118
sS's'
I17
sS'm'
(lp20187
I1
aI3
aI5
aI11
aI13
aI51
aI117
aI139
aI377
aI1015
aI1237
aI2053
aI7625
aI1003
aI22673
aI64345
aI16203
asS'd'
I10094
sa(dp20188
S'a'
I37123
sS's'
I17
sS'm'
(lp20189
I1
aI1
aI3
aI15
aI19
aI39
aI73
aI205
aI185
aI331
aI869
aI857
aI5043
aI7247
aI25253
aI5799
aI64857
asS'd'
I10095
sa(dp20190
S'a'
I37129
sS's'
I17
sS'm'
(lp20191
I1
aI3
aI7
aI1
aI25
aI63
aI125
aI47
aI161
aI289
aI373
aI1603
aI1663
aI1123
aI28907
aI37855
aI47935
asS'd'
I10096
sa(dp20192
S'a'
I37130
sS's'
I17
sS'm'
(lp20193
I1
aI1
aI7
aI15
aI9
aI17
aI97
aI63
aI79
aI123
aI1357
aI3055
aI2323
aI16083
aI21861
aI38743
aI81291
asS'd'
I10097
sa(dp20194
S'a'
I37135
sS's'
I17
sS'm'
(lp20195
I1
aI1
aI3
aI15
aI5
aI23
aI7
aI159
aI127
aI511
aI55
aI2691
aI6823
aI16151
aI8059
aI43021
aI18911
asS'd'
I10098
sa(dp20196
S'a'
I37137
sS's'
I17
sS'm'
(lp20197
I1
aI1
aI3
aI9
aI27
aI19
aI41
aI75
aI375
aI921
aI1745
aI35
aI1189
aI5857
aI29869
aI43827
aI16899
asS'd'
I10099
sa(dp20198
S'a'
I37138
sS's'
I17
sS'm'
(lp20199
I1
aI1
aI1
aI5
aI3
aI21
aI13
aI235
aI51
aI529
aI291
aI2619
aI5419
aI12573
aI10907
aI8865
aI54987
asS'd'
I10100
sa(dp20200
S'a'
I37140
sS's'
I17
sS'm'
(lp20201
I1
aI3
aI1
aI13
aI7
aI9
aI85
aI131
aI159
aI743
aI1671
aI3001
aI4559
aI12343
aI27563
aI49941
aI68447
asS'd'
I10101
sa(dp20202
S'a'
I37144
sS's'
I17
sS'm'
(lp20203
I1
aI1
aI7
aI5
aI17
aI61
aI99
aI63
aI199
aI383
aI485
aI2569
aI5329
aI645
aI18805
aI20421
aI101229
asS'd'
I10102
sa(dp20204
S'a'
I37149
sS's'
I17
sS'm'
(lp20205
I1
aI1
aI1
aI15
aI3
aI59
aI41
aI247
aI213
aI843
aI2003
aI125
aI7755
aI4203
aI20277
aI47195
aI48249
asS'd'
I10103
sa(dp20206
S'a'
I37156
sS's'
I17
sS'm'
(lp20207
I1
aI1
aI5
aI15
aI15
aI17
aI113
aI101
aI27
aI811
aI1791
aI1777
aI749
aI14317
aI17267
aI54467
aI118369
asS'd'
I10104
sa(dp20208
S'a'
I37159
sS's'
I17
sS'm'
(lp20209
I1
aI3
aI3
aI3
aI19
aI37
aI23
aI117
aI275
aI733
aI1259
aI567
aI1769
aI12071
aI5413
aI49411
aI99259
asS'd'
I10105
sa(dp20210
S'a'
I37163
sS's'
I17
sS'm'
(lp20211
I1
aI3
aI1
aI11
aI3
aI27
aI103
aI113
aI251
aI731
aI481
aI2771
aI3205
aI14151
aI19403
aI30307
aI114691
asS'd'
I10106
sa(dp20212
S'a'
I37165
sS's'
I17
sS'm'
(lp20213
I1
aI1
aI5
aI15
aI19
aI15
aI103
aI25
aI357
aI197
aI1437
aI3621
aI4747
aI773
aI5769
aI33465
aI28307
asS'd'
I10107
sa(dp20214
S'a'
I37183
sS's'
I17
sS'm'
(lp20215
I1
aI1
aI5
aI15
aI5
aI17
aI89
aI87
aI423
aI611
aI549
aI2549
aI1275
aI14545
aI2931
aI3853
aI24577
asS'd'
I10108
sa(dp20216
S'a'
I37185
sS's'
I17
sS'm'
(lp20217
I1
aI3
aI5
aI1
aI15
aI13
aI29
aI49
aI279
aI495
aI697
aI1015
aI4899
aI15977
aI10765
aI47979
aI40237
asS'd'
I10109
sa(dp20218
S'a'
I37195
sS's'
I17
sS'm'
(lp20219
I1
aI3
aI3
aI9
aI31
aI51
aI21
aI5
aI279
aI947
aI1871
aI3075
aI5433
aI1631
aI30075
aI30517
aI99609
asS'd'
I10110
sa(dp20220
S'a'
I37198
sS's'
I17
sS'm'
(lp20221
I1
aI1
aI1
aI15
aI19
aI63
aI79
aI81
aI19
aI629
aI617
aI1887
aI4015
aI15501
aI10551
aI56419
aI108739
asS'd'
I10111
sa(dp20222
S'a'
I37203
sS's'
I17
sS'm'
(lp20223
I1
aI1
aI3
aI9
aI31
aI15
aI45
aI37
aI43
aI349
aI1357
aI189
aI4551
aI9363
aI15683
aI48445
aI89279
asS'd'
I10112
sa(dp20224
S'a'
I37212
sS's'
I17
sS'm'
(lp20225
I1
aI1
aI1
aI1
aI17
aI19
aI121
aI119
aI397
aI947
aI1797
aI613
aI1627
aI9591
aI15779
aI62295
aI118843
asS'd'
I10113
sa(dp20226
S'a'
I37233
sS's'
I17
sS'm'
(lp20227
I1
aI1
aI1
aI7
aI25
aI55
aI71
aI227
aI507
aI497
aI1209
aI2919
aI5733
aI15785
aI21437
aI40043
aI2325
asS'd'
I10114
sa(dp20228
S'a'
I37236
sS's'
I17
sS'm'
(lp20229
I1
aI1
aI1
aI15
aI11
aI1
aI59
aI93
aI69
aI859
aI67
aI1831
aI6345
aI5643
aI29515
aI20337
aI77281
asS'd'
I10115
sa(dp20230
S'a'
I37240
sS's'
I17
sS'm'
(lp20231
I1
aI3
aI5
aI9
aI19
aI53
aI59
aI63
aI161
aI853
aI697
aI1441
aI3457
aI951
aI29659
aI15337
aI38443
asS'd'
I10116
sa(dp20232
S'a'
I37256
sS's'
I17
sS'm'
(lp20233
I1
aI3
aI1
aI9
aI7
aI21
aI73
aI81
aI89
aI291
aI411
aI3793
aI4639
aI2829
aI6855
aI38113
aI32875
asS'd'
I10117
sa(dp20234
S'a'
I37264
sS's'
I17
sS'm'
(lp20235
I1
aI1
aI7
aI1
aI15
aI3
aI79
aI35
aI363
aI459
aI907
aI1157
aI5165
aI8021
aI10135
aI36367
aI111991
asS'd'
I10118
sa(dp20236
S'a'
I37267
sS's'
I17
sS'm'
(lp20237
I1
aI3
aI5
aI13
aI21
aI23
aI63
aI155
aI393
aI869
aI1553
aI3345
aI2711
aI8249
aI24907
aI28111
aI36667
asS'd'
I10119
sa(dp20238
S'a'
I37273
sS's'
I17
sS'm'
(lp20239
I1
aI1
aI7
aI11
aI15
aI25
aI29
aI93
aI45
aI637
aI1473
aI2053
aI313
aI8047
aI23411
aI8643
aI2925
asS'd'
I10120
sa(dp20240
S'a'
I37295
sS's'
I17
sS'm'
(lp20241
I1
aI3
aI7
aI9
aI11
aI5
aI73
aI69
aI311
aI949
aI2017
aI259
aI2861
aI10547
aI12017
aI34125
aI74101
asS'd'
I10121
sa(dp20242
S'a'
I37315
sS's'
I17
sS'm'
(lp20243
I1
aI3
aI1
aI13
aI19
aI61
aI115
aI59
aI447
aI787
aI1621
aI2221
aI7841
aI5329
aI18137
aI13857
aI51889
asS'd'
I10122
sa(dp20244
S'a'
I37330
sS's'
I17
sS'm'
(lp20245
I1
aI3
aI7
aI13
aI1
aI23
aI117
aI49
aI449
aI541
aI7
aI3269
aI1725
aI6677
aI15979
aI4319
aI40919
asS'd'
I10123
sa(dp20246
S'a'
I37336
sS's'
I17
sS'm'
(lp20247
I1
aI3
aI5
aI5
aI17
aI29
aI35
aI123
aI3
aI481
aI305
aI1589
aI4319
aI5183
aI31907
aI53019
aI49375
asS'd'
I10124
sa(dp20248
S'a'
I37339
sS's'
I17
sS'm'
(lp20249
I1
aI3
aI1
aI7
aI11
aI59
aI79
aI89
aI479
aI821
aI763
aI3597
aI7457
aI13775
aI11213
aI22777
aI80379
asS'd'
I10125
sa(dp20250
S'a'
I37342
sS's'
I17
sS'm'
(lp20251
I1
aI1
aI3
aI7
aI13
aI17
aI65
aI155
aI335
aI671
aI331
aI895
aI7459
aI1719
aI10675
aI60109
aI63143
asS'd'
I10126
sa(dp20252
S'a'
I37357
sS's'
I17
sS'm'
(lp20253
I1
aI3
aI5
aI1
aI29
aI33
aI105
aI249
aI61
aI469
aI1629
aI3777
aI4393
aI14457
aI11701
aI6065
aI2635
asS'd'
I10127
sa(dp20254
S'a'
I37365
sS's'
I17
sS'm'
(lp20255
I1
aI3
aI7
aI3
aI13
aI13
aI21
aI15
aI363
aI63
aI1263
aI1479
aI1459
aI6577
aI7481
aI30393
aI19831
asS'd'
I10128
sa(dp20256
S'a'
I37372
sS's'
I17
sS'm'
(lp20257
I1
aI1
aI3
aI7
aI29
aI25
aI71
aI247
aI501
aI815
aI1697
aI2457
aI4975
aI3821
aI25759
aI24901
aI120603
asS'd'
I10129
sa(dp20258
S'a'
I37381
sS's'
I17
sS'm'
(lp20259
I1
aI1
aI1
aI5
aI19
aI3
aI59
aI163
aI367
aI779
aI47
aI905
aI897
aI3293
aI13951
aI25497
aI99151
asS'd'
I10130
sa(dp20260
S'a'
I37399
sS's'
I17
sS'm'
(lp20261
I1
aI3
aI1
aI5
aI11
aI47
aI21
aI171
aI123
aI215
aI1797
aI3741
aI4921
aI7213
aI4847
aI3239
aI114839
asS'd'
I10131
sa(dp20262
S'a'
I37416
sS's'
I17
sS'm'
(lp20263
I1
aI3
aI3
aI5
aI23
aI63
aI57
aI31
aI409
aI431
aI1337
aI3301
aI4695
aI7401
aI9383
aI12639
aI34347
asS'd'
I10132
sa(dp20264
S'a'
I37429
sS's'
I17
sS'm'
(lp20265
I1
aI3
aI3
aI5
aI27
aI57
aI29
aI147
aI111
aI1015
aI815
aI1509
aI3967
aI7255
aI15109
aI26001
aI90775
asS'd'
I10133
sa(dp20266
S'a'
I37436
sS's'
I17
sS'm'
(lp20267
I1
aI1
aI7
aI13
aI31
aI45
aI21
aI99
aI377
aI399
aI255
aI459
aI6043
aI11055
aI5675
aI3333
aI32813
asS'd'
I10134
sa(dp20268
S'a'
I37442
sS's'
I17
sS'm'
(lp20269
I1
aI3
aI1
aI7
aI1
aI55
aI121
aI77
aI429
aI433
aI297
aI3181
aI3029
aI6777
aI22795
aI61515
aI58553
asS'd'
I10135
sa(dp20270
S'a'
I37451
sS's'
I17
sS'm'
(lp20271
I1
aI3
aI5
aI9
aI1
aI19
aI121
aI1
aI499
aI589
aI1597
aI2219
aI1029
aI4223
aI31613
aI45685
aI53517
asS'd'
I10136
sa(dp20272
S'a'
I37453
sS's'
I17
sS'm'
(lp20273
I1
aI3
aI1
aI9
aI29
aI39
aI83
aI193
aI43
aI41
aI467
aI1711
aI2761
aI10635
aI15503
aI38043
aI120615
asS'd'
I10137
sa(dp20274
S'a'
I37475
sS's'
I17
sS'm'
(lp20275
I1
aI1
aI7
aI13
aI27
aI61
aI1
aI181
aI163
aI613
aI221
aI63
aI6147
aI8215
aI15093
aI2417
aI71489
asS'd'
I10138
sa(dp20276
S'a'
I37482
sS's'
I17
sS'm'
(lp20277
I1
aI1
aI7
aI15
aI31
aI63
aI47
aI139
aI427
aI847
aI53
aI1275
aI1019
aI9455
aI12537
aI22467
aI129947
asS'd'
I10139
sa(dp20278
S'a'
I37489
sS's'
I17
sS'm'
(lp20279
I1
aI1
aI5
aI3
aI7
aI1
aI67
aI189
aI501
aI319
aI37
aI2849
aI2535
aI10917
aI11115
aI48083
aI67255
asS'd'
I10140
sa(dp20280
S'a'
I37490
sS's'
I17
sS'm'
(lp20281
I1
aI1
aI3
aI13
aI7
aI31
aI69
aI137
aI19
aI73
aI1553
aI3945
aI2381
aI8761
aI3977
aI24291
aI128189
asS'd'
I10141
sa(dp20282
S'a'
I37523
sS's'
I17
sS'm'
(lp20283
I1
aI3
aI5
aI11
aI1
aI59
aI43
aI229
aI301
aI771
aI559
aI195
aI1675
aI12605
aI22211
aI2915
aI90351
asS'd'
I10142
sa(dp20284
S'a'
I37525
sS's'
I17
sS'm'
(lp20285
I1
aI3
aI3
aI9
aI13
aI27
aI97
aI33
aI273
aI229
aI1537
aI1179
aI6985
aI11679
aI17889
aI44673
aI126641
asS'd'
I10143
sa(dp20286
S'a'
I37530
sS's'
I17
sS'm'
(lp20287
I1
aI1
aI7
aI3
aI31
aI29
aI41
aI123
aI491
aI639
aI269
aI45
aI2155
aI14103
aI6725
aI50781
aI42785
asS'd'
I10144
sa(dp20288
S'a'
I37535
sS's'
I17
sS'm'
(lp20289
I1
aI3
aI5
aI9
aI9
aI11
aI89
aI249
aI475
aI701
aI1029
aI985
aI8167
aI439
aI31897
aI24529
aI45759
asS'd'
I10145
sa(dp20290
S'a'
I37539
sS's'
I17
sS'm'
(lp20291
I1
aI1
aI5
aI11
aI9
aI39
aI127
aI179
aI15
aI135
aI1437
aI3331
aI5553
aI939
aI15319
aI64937
aI110783
asS'd'
I10146
sa(dp20292
S'a'
I37548
sS's'
I17
sS'm'
(lp20293
I1
aI3
aI1
aI5
aI7
aI61
aI1
aI219
aI111
aI801
aI85
aI3427
aI2533
aI12861
aI5395
aI28969
aI48091
asS'd'
I10147
sa(dp20294
S'a'
I37559
sS's'
I17
sS'm'
(lp20295
I1
aI1
aI1
aI9
aI23
aI57
aI77
aI41
aI61
aI635
aI457
aI231
aI8121
aI5349
aI27021
aI64807
aI87563
asS'd'
I10148
sa(dp20296
S'a'
I37560
sS's'
I17
sS'm'
(lp20297
I1
aI3
aI5
aI7
aI31
aI31
aI101
aI155
aI255
aI199
aI1973
aI903
aI7681
aI15379
aI12845
aI47943
aI60663
asS'd'
I10149
sa(dp20298
S'a'
I37566
sS's'
I17
sS'm'
(lp20299
I1
aI1
aI5
aI7
aI1
aI7
aI71
aI121
aI323
aI669
aI193
aI1209
aI267
aI9
aI21223
aI22037
aI121567
asS'd'
I10150
sa(dp20300
S'a'
I37597
sS's'
I17
sS'm'
(lp20301
I1
aI3
aI1
aI5
aI17
aI29
aI97
aI189
aI219
aI813
aI187
aI1763
aI5817
aI13185
aI467
aI40159
aI18037
asS'd'
I10151
sa(dp20302
S'a'
I37601
sS's'
I17
sS'm'
(lp20303
I1
aI1
aI7
aI9
aI7
aI59
aI3
aI189
aI379
aI843
aI631
aI3945
aI2909
aI4191
aI30343
aI11223
aI105629
asS'd'
I10152
sa(dp20304
S'a'
I37602
sS's'
I17
sS'm'
(lp20305
I1
aI3
aI1
aI3
aI15
aI17
aI23
aI73
aI439
aI699
aI657
aI451
aI6139
aI15869
aI4101
aI32327
aI55485
asS'd'
I10153
sa(dp20306
S'a'
I37604
sS's'
I17
sS'm'
(lp20307
I1
aI3
aI3
aI5
aI21
aI37
aI87
aI157
aI205
aI493
aI705
aI1539
aI2193
aI13539
aI2797
aI49063
aI55595
asS'd'
I10154
sa(dp20308
S'a'
I37616
sS's'
I17
sS'm'
(lp20309
I1
aI1
aI5
aI11
aI11
aI41
aI5
aI131
aI445
aI781
aI1153
aI1371
aI6763
aI3101
aI32449
aI16065
aI86579
asS'd'
I10155
sa(dp20310
S'a'
I37622
sS's'
I17
sS'm'
(lp20311
I1
aI3
aI5
aI1
aI23
aI51
aI97
aI87
aI161
aI261
aI269
aI2035
aI2139
aI3049
aI32217
aI25189
aI93571
asS'd'
I10156
sa(dp20312
S'a'
I37658
sS's'
I17
sS'm'
(lp20313
I1
aI3
aI1
aI11
aI23
aI1
aI111
aI45
aI19
aI19
aI1767
aI3571
aI6027
aI3593
aI17453
aI53821
aI28121
asS'd'
I10157
sa(dp20314
S'a'
I37660
sS's'
I17
sS'm'
(lp20315
I1
aI3
aI3
aI5
aI17
aI5
aI17
aI247
aI5
aI73
aI29
aI443
aI7713
aI15803
aI22311
aI56755
aI100119
asS'd'
I10158
sa(dp20316
S'a'
I37667
sS's'
I17
sS'm'
(lp20317
I1
aI3
aI1
aI13
aI7
aI1
aI41
aI139
aI317
aI977
aI1529
aI1217
aI529
aI3231
aI21491
aI28461
aI96699
asS'd'
I10159
sa(dp20318
S'a'
I37684
sS's'
I17
sS'm'
(lp20319
I1
aI3
aI1
aI13
aI11
aI41
aI103
aI99
aI81
aI849
aI231
aI1729
aI761
aI711
aI11499
aI25581
aI59433
asS'd'
I10160
sa(dp20320
S'a'
I37693
sS's'
I17
sS'm'
(lp20321
I1
aI1
aI5
aI5
aI13
aI33
aI79
aI175
aI89
aI29
aI295
aI2867
aI1197
aI6137
aI32063
aI23471
aI21721
asS'd'
I10161
sa(dp20322
S'a'
I37694
sS's'
I17
sS'm'
(lp20323
I1
aI1
aI5
aI15
aI17
aI29
aI123
aI249
aI273
aI437
aI443
aI2601
aI3957
aI11955
aI261
aI54863
aI85727
asS'd'
I10162
sa(dp20324
S'a'
I37696
sS's'
I17
sS'm'
(lp20325
I1
aI1
aI5
aI13
aI3
aI31
aI57
aI205
aI3
aI903
aI905
aI3851
aI757
aI13761
aI28615
aI48185
aI33227
asS'd'
I10163
sa(dp20326
S'a'
I37706
sS's'
I17
sS'm'
(lp20327
I1
aI3
aI7
aI1
aI1
aI1
aI107
aI15
aI307
aI405
aI45
aI297
aI4365
aI1569
aI9263
aI13685
aI36027
asS'd'
I10164
sa(dp20328
S'a'
I37711
sS's'
I17
sS'm'
(lp20329
I1
aI3
aI1
aI9
aI17
aI61
aI113
aI121
aI249
aI743
aI191
aI2523
aI6621
aI5395
aI23797
aI57975
aI51675
asS'd'
I10165
sa(dp20330
S'a'
I37730
sS's'
I17
sS'm'
(lp20331
I1
aI3
aI5
aI3
aI27
aI21
aI49
aI113
aI59
aI989
aI501
aI2651
aI3827
aI5121
aI29667
aI32903
aI84199
asS'd'
I10166
sa(dp20332
S'a'
I37735
sS's'
I17
sS'm'
(lp20333
I1
aI1
aI7
aI7
aI19
aI43
aI11
aI191
aI143
aI93
aI1167
aI2521
aI2569
aI12187
aI28575
aI13073
aI113545
asS'd'
I10167
sa(dp20334
S'a'
I37742
sS's'
I17
sS'm'
(lp20335
I1
aI1
aI7
aI3
aI27
aI39
aI11
aI85
aI61
aI979
aI49
aI2191
aI2607
aI13967
aI28123
aI48903
aI16327
asS'd'
I10168
sa(dp20336
S'a'
I37744
sS's'
I17
sS'm'
(lp20337
I1
aI1
aI5
aI3
aI17
aI17
aI1
aI149
aI189
aI1017
aI705
aI3119
aI6441
aI1595
aI30533
aI18795
aI34265
asS'd'
I10169
sa(dp20338
S'a'
I37759
sS's'
I17
sS'm'
(lp20339
I1
aI1
aI3
aI11
aI31
aI11
aI125
aI109
aI39
aI41
aI191
aI2615
aI1377
aI16089
aI8793
aI31425
aI90507
asS'd'
I10170
sa(dp20340
S'a'
I37777
sS's'
I17
sS'm'
(lp20341
I1
aI1
aI1
aI1
aI23
aI15
aI21
aI245
aI337
aI649
aI585
aI2893
aI927
aI883
aI15119
aI2595
aI127963
asS'd'
I10171
sa(dp20342
S'a'
I37783
sS's'
I17
sS'm'
(lp20343
I1
aI3
aI5
aI3
aI13
aI17
aI123
aI167
aI471
aI5
aI1671
aI2787
aI5081
aI12903
aI4257
aI19213
aI2503
asS'd'
I10172
sa(dp20344
S'a'
I37784
sS's'
I17
sS'm'
(lp20345
I1
aI1
aI5
aI7
aI21
aI57
aI75
aI171
aI509
aI591
aI85
aI407
aI1747
aI6375
aI19641
aI55683
aI111289
asS'd'
I10173
sa(dp20346
S'a'
I37793
sS's'
I17
sS'm'
(lp20347
I1
aI1
aI7
aI7
aI3
aI31
aI121
aI111
aI19
aI361
aI1033
aI4033
aI2359
aI13451
aI15095
aI61817
aI69683
asS'd'
I10174
sa(dp20348
S'a'
I37800
sS's'
I17
sS'm'
(lp20349
I1
aI1
aI5
aI9
aI21
aI33
aI83
aI179
aI387
aI69
aI1085
aI2147
aI2751
aI10899
aI16971
aI40623
aI110891
asS'd'
I10175
sa(dp20350
S'a'
I37818
sS's'
I17
sS'm'
(lp20351
I1
aI1
aI7
aI9
aI11
aI45
aI81
aI71
aI73
aI551
aI145
aI159
aI7519
aI3459
aI5197
aI48913
aI59045
asS'd'
I10176
sa(dp20352
S'a'
I37823
sS's'
I17
sS'm'
(lp20353
I1
aI3
aI1
aI3
aI7
aI35
aI17
aI249
aI207
aI767
aI1189
aI1451
aI4351
aI3673
aI28807
aI671
aI69271
asS'd'
I10177
sa(dp20354
S'a'
I37825
sS's'
I17
sS'm'
(lp20355
I1
aI3
aI1
aI15
aI21
aI27
aI81
aI243
aI55
aI191
aI1497
aI3205
aI1601
aI705
aI14891
aI14403
aI130729
asS'd'
I10178
sa(dp20356
S'a'
I37828
sS's'
I17
sS'm'
(lp20357
I1
aI3
aI1
aI7
aI17
aI43
aI41
aI123
aI507
aI201
aI1873
aI3547
aI5681
aI1819
aI4251
aI39661
aI57923
asS'd'
I10179
sa(dp20358
S'a'
I37832
sS's'
I17
sS'm'
(lp20359
I1
aI3
aI1
aI9
aI3
aI59
aI57
aI235
aI445
aI479
aI961
aI1937
aI2229
aI2511
aI15235
aI59707
aI72261
asS'd'
I10180
sa(dp20360
S'a'
I37849
sS's'
I17
sS'm'
(lp20361
I1
aI1
aI5
aI13
aI9
aI63
aI67
aI217
aI63
aI259
aI175
aI2469
aI3075
aI12365
aI7727
aI42215
aI12635
asS'd'
I10181
sa(dp20362
S'a'
I37856
sS's'
I17
sS'm'
(lp20363
I1
aI1
aI3
aI13
aI11
aI9
aI125
aI131
aI17
aI399
aI675
aI767
aI7349
aI10433
aI21615
aI46823
aI3955
asS'd'
I10182
sa(dp20364
S'a'
I37861
sS's'
I17
sS'm'
(lp20365
I1
aI1
aI3
aI15
aI19
aI53
aI73
aI171
aI125
aI531
aI1093
aI1449
aI2931
aI10897
aI12263
aI9799
aI98251
asS'd'
I10183
sa(dp20366
S'a'
I37862
sS's'
I17
sS'm'
(lp20367
I1
aI1
aI5
aI5
aI11
aI27
aI33
aI9
aI503
aI545
aI339
aI1099
aI1973
aI13261
aI26871
aI14569
aI22755
asS'd'
I10184
sa(dp20368
S'a'
I37886
sS's'
I17
sS'm'
(lp20369
I1
aI3
aI7
aI1
aI19
aI5
aI79
aI133
aI247
aI1021
aI1431
aI3707
aI4603
aI3285
aI5469
aI46963
aI98203
asS'd'
I10185
sa(dp20370
S'a'
I37893
sS's'
I17
sS'm'
(lp20371
I1
aI3
aI5
aI7
aI15
aI11
aI87
aI169
aI495
aI763
aI1295
aI2533
aI4213
aI8671
aI21683
aI12521
aI90071
asS'd'
I10186
sa(dp20372
S'a'
I37915
sS's'
I17
sS'm'
(lp20373
I1
aI1
aI3
aI1
aI17
aI55
aI7
aI165
aI313
aI659
aI49
aI377
aI6675
aI15255
aI9881
aI11751
aI87789
asS'd'
I10187
sa(dp20374
S'a'
I37922
sS's'
I17
sS'm'
(lp20375
I1
aI3
aI1
aI15
aI3
aI49
aI27
aI109
aI145
aI1011
aI1939
aI3201
aI6141
aI7229
aI20741
aI59285
aI129365
asS'd'
I10188
sa(dp20376
S'a'
I37936
sS's'
I17
sS'm'
(lp20377
I1
aI1
aI5
aI5
aI5
aI51
aI117
aI17
aI363
aI795
aI1343
aI2637
aI6209
aI1045
aI22515
aI10687
aI48487
asS'd'
I10189
sa(dp20378
S'a'
I37939
sS's'
I17
sS'm'
(lp20379
I1
aI1
aI5
aI15
aI1
aI37
aI91
aI241
aI245
aI703
aI505
aI3369
aI6163
aI10265
aI12497
aI46301
aI109523
asS'd'
I10190
sa(dp20380
S'a'
I37945
sS's'
I17
sS'm'
(lp20381
I1
aI1
aI7
aI11
aI3
aI37
aI67
aI35
aI229
aI823
aI193
aI913
aI3331
aI4475
aI9271
aI11859
aI52709
asS'd'
I10191
sa(dp20382
S'a'
I37956
sS's'
I17
sS'm'
(lp20383
I1
aI1
aI3
aI3
aI7
aI25
aI61
aI159
aI81
aI1011
aI1491
aI1439
aI1031
aI765
aI9839
aI61891
aI20969
asS'd'
I10192
sa(dp20384
S'a'
I37960
sS's'
I17
sS'm'
(lp20385
I1
aI3
aI3
aI7
aI25
aI39
aI73
aI59
aI101
aI101
aI225
aI1105
aI5943
aI5223
aI12585
aI16411
aI62699
asS'd'
I10193
sa(dp20386
S'a'
I37971
sS's'
I17
sS'm'
(lp20387
I1
aI1
aI7
aI5
aI25
aI19
aI27
aI113
aI465
aI319
aI2035
aI2127
aI1319
aI11793
aI26821
aI44805
aI28217
asS'd'
I10194
sa(dp20388
S'a'
I37980
sS's'
I17
sS'm'
(lp20389
I1
aI3
aI5
aI11
aI7
aI9
aI81
aI107
aI67
aI31
aI1503
aI3303
aI4451
aI11417
aI32681
aI26861
aI54845
asS'd'
I10195
sa(dp20390
S'a'
I38002
sS's'
I17
sS'm'
(lp20391
I1
aI3
aI1
aI1
aI3
aI51
aI93
aI235
aI93
aI247
aI2027
aI1517
aI6797
aI1703
aI10233
aI45313
aI60771
asS'd'
I10196
sa(dp20392
S'a'
I38007
sS's'
I17
sS'm'
(lp20393
I1
aI3
aI3
aI15
aI25
aI11
aI83
aI77
aI413
aI189
aI119
aI597
aI2199
aI12347
aI7935
aI40191
aI125569
asS'd'
I10197
sa(dp20394
S'a'
I38018
sS's'
I17
sS'm'
(lp20395
I1
aI3
aI7
aI9
aI11
aI3
aI77
aI31
aI89
aI163
aI1993
aI3017
aI3973
aI10943
aI22247
aI45565
aI7261
asS'd'
I10198
sa(dp20396
S'a'
I38020
sS's'
I17
sS'm'
(lp20397
I1
aI3
aI7
aI7
aI15
aI13
aI7
aI155
aI373
aI893
aI607
aI3521
aI7455
aI13809
aI6145
aI31743
aI86329
asS'd'
I10199
sa(dp20398
S'a'
I38038
sS's'
I17
sS'm'
(lp20399
I1
aI1
aI1
aI15
aI25
aI41
aI111
aI65
aI11
aI627
aI59
aI2725
aI995
aI3761
aI25361
aI45189
aI48355
asS'd'
I10200
sa(dp20400
S'a'
I38041
sS's'
I17
sS'm'
(lp20401
I1
aI1
aI7
aI5
aI29
aI43
aI91
aI139
aI323
aI503
aI679
aI4079
aI403
aI1899
aI1425
aI26989
aI117057
asS'd'
I10201
sa(dp20402
S'a'
I38044
sS's'
I17
sS'm'
(lp20403
I1
aI3
aI5
aI13
aI1
aI17
aI19
aI61
aI205
aI833
aI345
aI1031
aI7995
aI999
aI27469
aI15943
aI88011
asS'd'
I10202
sa(dp20404
S'a'
I38048
sS's'
I17
sS'm'
(lp20405
I1
aI3
aI1
aI7
aI23
aI49
aI123
aI9
aI11
aI761
aI1163
aI669
aI3837
aI15225
aI23393
aI19513
aI9457
asS'd'
I10203
sa(dp20406
S'a'
I38051
sS's'
I17
sS'm'
(lp20407
I1
aI1
aI5
aI9
aI9
aI33
aI29
aI123
aI277
aI433
aI1799
aI1583
aI3133
aI13461
aI26443
aI15807
aI80173
asS'd'
I10204
sa(dp20408
S'a'
I38054
sS's'
I17
sS'm'
(lp20409
I1
aI3
aI7
aI3
aI31
aI29
aI77
aI105
aI297
aI617
aI491
aI647
aI6541
aI5033
aI31841
aI48405
aI126985
asS'd'
I10205
sa(dp20410
S'a'
I38065
sS's'
I17
sS'm'
(lp20411
I1
aI1
aI5
aI5
aI31
aI39
aI17
aI25
aI3
aI279
aI89
aI3985
aI3333
aI5681
aI3701
aI36319
aI12585
asS'd'
I10206
sa(dp20412
S'a'
I38066
sS's'
I17
sS'm'
(lp20413
I1
aI1
aI7
aI13
aI13
aI19
aI93
aI129
aI51
aI875
aI1083
aI1739
aI5193
aI6217
aI10033
aI51839
aI66071
asS'd'
I10207
sa(dp20414
S'a'
I38077
sS's'
I17
sS'm'
(lp20415
I1
aI1
aI7
aI9
aI15
aI23
aI93
aI121
aI507
aI115
aI707
aI3181
aI1521
aI9609
aI4577
aI54389
aI19167
asS'd'
I10208
sa(dp20416
S'a'
I38090
sS's'
I17
sS'm'
(lp20417
I1
aI3
aI3
aI7
aI17
aI51
aI19
aI29
aI387
aI711
aI1105
aI1627
aI4421
aI15183
aI14149
aI26485
aI106425
asS'd'
I10209
sa(dp20418
S'a'
I38109
sS's'
I17
sS'm'
(lp20419
I1
aI3
aI3
aI15
aI25
aI59
aI11
aI45
aI259
aI1019
aI1997
aI3373
aI5083
aI5701
aI30217
aI44845
aI67559
asS'd'
I10210
sa(dp20420
S'a'
I38119
sS's'
I17
sS'm'
(lp20421
I1
aI3
aI3
aI9
aI17
aI47
aI5
aI103
aI477
aI785
aI235
aI1523
aI1505
aI8811
aI15255
aI53493
aI4383
asS'd'
I10211
sa(dp20422
S'a'
I38120
sS's'
I17
sS'm'
(lp20423
I1
aI3
aI7
aI3
aI7
aI37
aI73
aI247
aI397
aI409
aI1065
aI525
aI5665
aI8533
aI30627
aI19035
aI22937
asS'd'
I10212
sa(dp20424
S'a'
I38123
sS's'
I17
sS'm'
(lp20425
I1
aI3
aI3
aI15
aI15
aI47
aI123
aI215
aI413
aI249
aI55
aI2563
aI8033
aI8743
aI18659
aI7947
aI56057
asS'd'
I10213
sa(dp20426
S'a'
I38146
sS's'
I17
sS'm'
(lp20427
I1
aI1
aI7
aI3
aI11
aI61
aI103
aI109
aI313
aI293
aI149
aI999
aI901
aI13387
aI15351
aI52973
aI68385
asS'd'
I10214
sa(dp20428
S'a'
I38155
sS's'
I17
sS'm'
(lp20429
I1
aI1
aI1
aI5
aI31
aI13
aI57
aI43
aI263
aI141
aI335
aI2777
aI3435
aI4231
aI20623
aI2597
aI33481
asS'd'
I10215
sa(dp20430
S'a'
I38160
sS's'
I17
sS'm'
(lp20431
I1
aI1
aI7
aI13
aI21
aI53
aI101
aI75
aI237
aI275
aI1903
aI3501
aI8023
aI3651
aI19609
aI44417
aI60287
asS'd'
I10216
sa(dp20432
S'a'
I38181
sS's'
I17
sS'm'
(lp20433
I1
aI1
aI1
aI1
aI13
aI43
aI83
aI255
aI355
aI567
aI1781
aI2943
aI1061
aI2701
aI24861
aI58381
aI60255
asS'd'
I10217
sa(dp20434
S'a'
I38188
sS's'
I17
sS'm'
(lp20435
I1
aI1
aI3
aI9
aI25
aI5
aI81
aI85
aI445
aI857
aI517
aI3687
aI2641
aI6699
aI19273
aI4481
aI8715
asS'd'
I10218
sa(dp20436
S'a'
I38199
sS's'
I17
sS'm'
(lp20437
I1
aI1
aI3
aI7
aI17
aI39
aI33
aI31
aI29
aI269
aI379
aI3149
aI4731
aI10387
aI7941
aI49199
aI18423
asS'd'
I10219
sa(dp20438
S'a'
I38203
sS's'
I17
sS'm'
(lp20439
I1
aI1
aI7
aI15
aI19
aI37
aI105
aI157
aI185
aI1023
aI1865
aI53
aI6765
aI3
aI22897
aI17019
aI109521
asS'd'
I10220
sa(dp20440
S'a'
I38225
sS's'
I17
sS'm'
(lp20441
I1
aI3
aI7
aI5
aI5
aI7
aI117
aI211
aI19
aI149
aI1091
aI3721
aI201
aI4455
aI18965
aI51401
aI67225
asS'd'
I10221
sa(dp20442
S'a'
I38226
sS's'
I17
sS'm'
(lp20443
I1
aI3
aI5
aI11
aI1
aI55
aI101
aI41
aI469
aI271
aI1251
aI949
aI861
aI11903
aI14773
aI25675
aI114161
asS'd'
I10222
sa(dp20444
S'a'
I38231
sS's'
I17
sS'm'
(lp20445
I1
aI1
aI7
aI7
aI23
aI13
aI103
aI185
aI137
aI575
aI797
aI1195
aI5301
aI13307
aI12043
aI26003
aI31719
asS'd'
I10223
sa(dp20446
S'a'
I38244
sS's'
I17
sS'm'
(lp20447
I1
aI1
aI5
aI7
aI11
aI51
aI17
aI71
aI321
aI559
aI1461
aI3571
aI1033
aI15575
aI7097
aI14703
aI52359
asS'd'
I10224
sa(dp20448
S'a'
I38248
sS's'
I17
sS'm'
(lp20449
I1
aI1
aI1
aI5
aI21
aI9
aI123
aI211
aI233
aI81
aI111
aI1433
aI7825
aI11771
aI30743
aI23993
aI48717
asS'd'
I10225
sa(dp20450
S'a'
I38259
sS's'
I17
sS'm'
(lp20451
I1
aI3
aI5
aI15
aI7
aI3
aI109
aI33
aI99
aI135
aI393
aI3463
aI7271
aI14387
aI30723
aI19079
aI83073
asS'd'
I10226
sa(dp20452
S'a'
I38268
sS's'
I17
sS'm'
(lp20453
I1
aI3
aI3
aI15
aI3
aI51
aI77
aI219
aI409
aI11
aI67
aI3787
aI5155
aI9259
aI7185
aI21611
aI32577
asS'd'
I10227
sa(dp20454
S'a'
I38271
sS's'
I17
sS'm'
(lp20455
I1
aI3
aI7
aI1
aI5
aI49
aI125
aI85
aI151
aI301
aI887
aI1765
aI5
aI12849
aI11775
aI11319
aI29547
asS'd'
I10228
sa(dp20456
S'a'
I38272
sS's'
I17
sS'm'
(lp20457
I1
aI1
aI7
aI11
aI17
aI15
aI105
aI29
aI327
aI637
aI1493
aI3361
aI1823
aI14709
aI18355
aI741
aI57807
asS'd'
I10229
sa(dp20458
S'a'
I38278
sS's'
I17
sS'm'
(lp20459
I1
aI1
aI7
aI7
aI3
aI27
aI15
aI113
aI227
aI617
aI1543
aI1719
aI8065
aI13627
aI23525
aI20511
aI64759
asS'd'
I10230
sa(dp20460
S'a'
I38292
sS's'
I17
sS'm'
(lp20461
I1
aI1
aI3
aI3
aI21
aI47
aI89
aI177
aI381
aI711
aI1367
aI2405
aI887
aI2351
aI22957
aI49679
aI5963
asS'd'
I10231
sa(dp20462
S'a'
I38296
sS's'
I17
sS'm'
(lp20463
I1
aI3
aI7
aI9
aI7
aI1
aI39
aI71
aI9
aI275
aI875
aI1385
aI6215
aI10419
aI25921
aI63427
aI18031
asS'd'
I10232
sa(dp20464
S'a'
I38318
sS's'
I17
sS'm'
(lp20465
I1
aI1
aI7
aI5
aI23
aI57
aI27
aI7
aI445
aI111
aI953
aI37
aI2769
aI1967
aI8165
aI35417
aI36471
asS'd'
I10233
sa(dp20466
S'a'
I38326
sS's'
I17
sS'm'
(lp20467
I1
aI3
aI3
aI11
aI23
aI17
aI119
aI113
aI275
aI625
aI1957
aI2795
aI3815
aI7937
aI11049
aI31939
aI128053
asS'd'
I10234
sa(dp20468
S'a'
I38338
sS's'
I17
sS'm'
(lp20469
I1
aI3
aI3
aI5
aI7
aI35
aI45
aI41
aI251
aI491
aI1953
aI3201
aI751
aI5705
aI595
aI27003
aI77917
asS'd'
I10235
sa(dp20470
S'a'
I38347
sS's'
I17
sS'm'
(lp20471
I1
aI3
aI5
aI3
aI25
aI17
aI55
aI137
aI299
aI541
aI289
aI2225
aI4667
aI3569
aI13687
aI36193
aI75705
asS'd'
I10236
sa(dp20472
S'a'
I38373
sS's'
I17
sS'm'
(lp20473
I1
aI1
aI7
aI15
aI21
aI9
aI27
aI157
aI469
aI441
aI193
aI2097
aI4863
aI2147
aI31197
aI24283
aI102039
asS'd'
I10237
sa(dp20474
S'a'
I38378
sS's'
I17
sS'm'
(lp20475
I1
aI1
aI1
aI11
aI17
aI39
aI73
aI37
aI91
aI121
aI1283
aI367
aI1875
aI14365
aI28349
aI60993
aI10791
asS'd'
I10238
sa(dp20476
S'a'
I38380
sS's'
I17
sS'm'
(lp20477
I1
aI3
aI5
aI9
aI13
aI63
aI89
aI53
aI459
aI347
aI343
aI2321
aI5237
aI2497
aI279
aI63833
aI10709
asS'd'
I10239
sa(dp20478
S'a'
I38388
sS's'
I17
sS'm'
(lp20479
I1
aI1
aI1
aI15
aI11
aI23
aI41
aI79
aI45
aI567
aI1217
aI1669
aI1679
aI2561
aI16191
aI49041
aI4081
asS'd'
I10240
sa(dp20480
S'a'
I38392
sS's'
I17
sS'm'
(lp20481
I1
aI3
aI1
aI5
aI3
aI9
aI103
aI245
aI47
aI561
aI229
aI2945
aI6563
aI797
aI21571
aI25769
aI12995
asS'd'
I10241
sa(dp20482
S'a'
I38397
sS's'
I17
sS'm'
(lp20483
I1
aI3
aI5
aI7
aI5
aI47
aI99
aI55
aI49
aI951
aI765
aI2095
aI8021
aI4389
aI20501
aI26047
aI119967
asS'd'
I10242
sa(dp20484
S'a'
I38398
sS's'
I17
sS'm'
(lp20485
I1
aI1
aI5
aI3
aI15
aI47
aI81
aI121
aI379
aI527
aI419
aI1093
aI367
aI10939
aI17181
aI13905
aI49859
asS'd'
I10243
sa(dp20486
S'a'
I38401
sS's'
I17
sS'm'
(lp20487
I1
aI3
aI3
aI5
aI7
aI59
aI53
aI255
aI131
aI685
aI1677
aI3757
aI3601
aI89
aI6225
aI32705
aI28287
asS'd'
I10244
sa(dp20488
S'a'
I38404
sS's'
I17
sS'm'
(lp20489
I1
aI3
aI7
aI1
aI7
aI55
aI85
aI47
aI425
aI793
aI605
aI2313
aI791
aI4247
aI9693
aI10633
aI82915
asS'd'
I10245
sa(dp20490
S'a'
I38422
sS's'
I17
sS'm'
(lp20491
I1
aI3
aI5
aI13
aI13
aI49
aI127
aI213
aI27
aI657
aI419
aI55
aI6289
aI295
aI4211
aI8899
aI120237
asS'd'
I10246
sa(dp20492
S'a'
I38432
sS's'
I17
sS'm'
(lp20493
I1
aI1
aI7
aI11
aI11
aI7
aI75
aI35
aI315
aI125
aI517
aI3677
aI2323
aI6897
aI11535
aI36789
aI20179
asS'd'
I10247
sa(dp20494
S'a'
I38469
sS's'
I17
sS'm'
(lp20495
I1
aI1
aI5
aI9
aI11
aI3
aI77
aI43
aI323
aI647
aI383
aI485
aI3937
aI9081
aI27745
aI59149
aI103433
asS'd'
I10248
sa(dp20496
S'a'
I38482
sS's'
I17
sS'm'
(lp20497
I1
aI3
aI3
aI13
aI3
aI47
aI91
aI81
aI115
aI625
aI2003
aI3601
aI531
aI113
aI20719
aI47391
aI111039
asS'd'
I10249
sa(dp20498
S'a'
I38493
sS's'
I17
sS'm'
(lp20499
I1
aI1
aI3
aI13
aI5
aI49
aI123
aI189
aI109
aI325
aI497
aI923
aI3861
aI14029
aI22651
aI19857
aI104801
asS'd'
I10250
sa(dp20500
S'a'
I38507
sS's'
I17
sS'm'
(lp20501
I1
aI1
aI5
aI3
aI29
aI23
aI25
aI23
aI501
aI371
aI1983
aI1303
aI2261
aI11865
aI13281
aI2587
aI75741
asS'd'
I10251
sa(dp20502
S'a'
I38518
sS's'
I17
sS'm'
(lp20503
I1
aI3
aI5
aI13
aI27
aI61
aI45
aI11
aI157
aI615
aI897
aI1529
aI2213
aI757
aI30105
aI2011
aI27267
asS'd'
I10252
sa(dp20504
S'a'
I38521
sS's'
I17
sS'm'
(lp20505
I1
aI3
aI5
aI13
aI29
aI31
aI95
aI159
aI449
aI307
aI1575
aI1897
aI2301
aI14023
aI6921
aI30543
aI31843
asS'd'
I10253
sa(dp20506
S'a'
I38527
sS's'
I17
sS'm'
(lp20507
I1
aI3
aI3
aI5
aI1
aI1
aI79
aI147
aI437
aI623
aI1161
aI3407
aI3073
aI15425
aI4329
aI19651
aI90597
asS'd'
I10254
sa(dp20508
S'a'
I38533
sS's'
I17
sS'm'
(lp20509
I1
aI1
aI1
aI11
aI17
aI7
aI43
aI171
aI447
aI841
aI573
aI3775
aI5517
aI3629
aI18241
aI31907
aI51423
asS'd'
I10255
sa(dp20510
S'a'
I38534
sS's'
I17
sS'm'
(lp20511
I1
aI1
aI3
aI7
aI15
aI53
aI111
aI203
aI171
aI963
aI1983
aI2017
aI6067
aI12281
aI3417
aI7431
aI33803
asS'd'
I10256
sa(dp20512
S'a'
I38552
sS's'
I17
sS'm'
(lp20513
I1
aI3
aI1
aI1
aI31
aI49
aI125
aI65
aI7
aI579
aI1709
aI1815
aI2643
aI11537
aI30093
aI11813
aI65157
asS'd'
I10257
sa(dp20514
S'a'
I38561
sS's'
I17
sS'm'
(lp20515
I1
aI3
aI7
aI15
aI1
aI3
aI61
aI21
aI163
aI809
aI1263
aI2577
aI7811
aI12611
aI6921
aI18529
aI25709
asS'd'
I10258
sa(dp20516
S'a'
I38576
sS's'
I17
sS'm'
(lp20517
I1
aI1
aI3
aI5
aI17
aI43
aI13
aI81
aI29
aI905
aI1975
aI589
aI5875
aI15683
aI29447
aI46453
aI127911
asS'd'
I10259
sa(dp20518
S'a'
I38581
sS's'
I17
sS'm'
(lp20519
I1
aI1
aI5
aI3
aI19
aI29
aI11
aI67
aI375
aI771
aI755
aI3939
aI1465
aI3275
aI1119
aI24695
aI105105
asS'd'
I10260
sa(dp20520
S'a'
I38585
sS's'
I17
sS'm'
(lp20521
I1
aI1
aI3
aI11
aI23
aI37
aI33
aI135
aI329
aI733
aI245
aI2353
aI2547
aI7823
aI16265
aI5975
aI37877
asS'd'
I10261
sa(dp20522
S'a'
I38594
sS's'
I17
sS'm'
(lp20523
I1
aI3
aI7
aI13
aI15
aI9
aI47
aI181
aI239
aI723
aI1219
aI409
aI1685
aI5493
aI14189
aI18107
aI26231
asS'd'
I10262
sa(dp20524
S'a'
I38606
sS's'
I17
sS'm'
(lp20525
I1
aI1
aI5
aI1
aI9
aI1
aI65
aI125
aI439
aI591
aI1499
aI3797
aI5879
aI4231
aI18655
aI9643
aI42265
asS'd'
I10263
sa(dp20526
S'a'
I38613
sS's'
I17
sS'm'
(lp20527
I1
aI1
aI7
aI7
aI11
aI51
aI111
aI47
aI169
aI39
aI45
aI2211
aI6729
aI10987
aI22367
aI27257
aI112711
asS'd'
I10264
sa(dp20528
S'a'
I38617
sS's'
I17
sS'm'
(lp20529
I1
aI3
aI5
aI3
aI19
aI61
aI89
aI185
aI23
aI793
aI1457
aI1743
aI3743
aI15063
aI14053
aI50201
aI109175
asS'd'
I10265
sa(dp20530
S'a'
I38634
sS's'
I17
sS'm'
(lp20531
I1
aI1
aI5
aI13
aI31
aI51
aI69
aI135
aI427
aI527
aI1673
aI2393
aI5829
aI683
aI1509
aI56617
aI105735
asS'd'
I10266
sa(dp20532
S'a'
I38644
sS's'
I17
sS'm'
(lp20533
I1
aI1
aI1
aI15
aI31
aI51
aI3
aI105
aI125
aI593
aI1589
aI3217
aI7449
aI525
aI30599
aI11689
aI14781
asS'd'
I10267
sa(dp20534
S'a'
I38651
sS's'
I17
sS'm'
(lp20535
I1
aI1
aI1
aI11
aI9
aI37
aI113
aI45
aI487
aI961
aI87
aI1461
aI3521
aI8645
aI19771
aI43817
aI43277
asS'd'
I10268
sa(dp20536
S'a'
I38661
sS's'
I17
sS'm'
(lp20537
I1
aI1
aI7
aI3
aI7
aI11
aI45
aI97
aI11
aI593
aI319
aI2597
aI37
aI4157
aI6669
aI29929
aI17213
asS'd'
I10269
sa(dp20538
S'a'
I38665
sS's'
I17
sS'm'
(lp20539
I1
aI3
aI7
aI3
aI29
aI21
aI101
aI93
aI289
aI975
aI1937
aI3423
aI757
aI7075
aI12575
aI26801
aI90989
asS'd'
I10270
sa(dp20540
S'a'
I38668
sS's'
I17
sS'm'
(lp20541
I1
aI1
aI7
aI15
aI25
aI49
aI49
aI149
aI503
aI365
aI1359
aI2155
aI7977
aI14955
aI18439
aI44269
aI88995
asS'd'
I10271
sa(dp20542
S'a'
I38674
sS's'
I17
sS'm'
(lp20543
I1
aI3
aI7
aI13
aI25
aI27
aI15
aI67
aI157
aI873
aI339
aI2143
aI1405
aI12209
aI30939
aI36109
aI107699
asS'd'
I10272
sa(dp20544
S'a'
I38689
sS's'
I17
sS'm'
(lp20545
I1
aI3
aI5
aI5
aI21
aI33
aI121
aI95
aI61
aI159
aI1423
aI2899
aI3811
aI263
aI9139
aI54481
aI107375
asS'd'
I10273
sa(dp20546
S'a'
I38695
sS's'
I17
sS'm'
(lp20547
I1
aI1
aI7
aI7
aI13
aI49
aI83
aI53
aI267
aI131
aI673
aI3945
aI5255
aI2009
aI21017
aI41749
aI44817
asS'd'
I10274
sa(dp20548
S'a'
I38696
sS's'
I17
sS'm'
(lp20549
I1
aI3
aI1
aI13
aI25
aI57
aI125
aI5
aI323
aI653
aI221
aI2013
aI7225
aI8719
aI28049
aI41953
aI14725
asS'd'
I10275
sa(dp20550
S'a'
I38707
sS's'
I17
sS'm'
(lp20551
I1
aI3
aI7
aI7
aI5
aI13
aI35
aI161
aI221
aI951
aI769
aI717
aI267
aI2233
aI23387
aI47411
aI95739
asS'd'
I10276
sa(dp20552
S'a'
I38710
sS's'
I17
sS'm'
(lp20553
I1
aI3
aI7
aI11
aI23
aI47
aI37
aI67
aI501
aI159
aI763
aI4045
aI5125
aI5667
aI9895
aI33041
aI101171
asS'd'
I10277
sa(dp20554
S'a'
I38716
sS's'
I17
sS'm'
(lp20555
I1
aI1
aI7
aI1
aI31
aI31
aI111
aI183
aI33
aI895
aI1819
aI3593
aI1285
aI10145
aI3679
aI36615
aI82863
asS'd'
I10278
sa(dp20556
S'a'
I38733
sS's'
I17
sS'm'
(lp20557
I1
aI3
aI5
aI7
aI21
aI59
aI55
aI163
aI139
aI855
aI1903
aI3229
aI3745
aI10289
aI28831
aI46895
aI12647
asS'd'
I10279
sa(dp20558
S'a'
I38736
sS's'
I17
sS'm'
(lp20559
I1
aI3
aI7
aI9
aI25
aI31
aI113
aI177
aI459
aI201
aI565
aI2089
aI725
aI9273
aI26249
aI5987
aI49573
asS'd'
I10280
sa(dp20560
S'a'
I38742
sS's'
I17
sS'm'
(lp20561
I1
aI1
aI7
aI15
aI3
aI37
aI49
aI145
aI121
aI803
aI1197
aI2797
aI21
aI833
aI2277
aI28189
aI93171
asS'd'
I10281
sa(dp20562
S'a'
I38751
sS's'
I17
sS'm'
(lp20563
I1
aI1
aI7
aI7
aI13
aI31
aI93
aI153
aI345
aI363
aI1825
aI1481
aI3347
aI13277
aI26119
aI63153
aI118231
asS'd'
I10282
sa(dp20564
S'a'
I38752
sS's'
I17
sS'm'
(lp20565
I1
aI3
aI1
aI11
aI31
aI9
aI33
aI95
aI433
aI595
aI1131
aI465
aI1797
aI15453
aI32527
aI40789
aI37799
asS'd'
I10283
sa(dp20566
S'a'
I38775
sS's'
I17
sS'm'
(lp20567
I1
aI1
aI5
aI11
aI31
aI29
aI83
aI33
aI243
aI633
aI1325
aI3843
aI7717
aI851
aI29789
aI48827
aI89209
asS'd'
I10284
sa(dp20568
S'a'
I38779
sS's'
I17
sS'm'
(lp20569
I1
aI1
aI3
aI7
aI25
aI31
aI127
aI219
aI281
aI51
aI1843
aI3363
aI5985
aI1697
aI3083
aI18967
aI117421
asS'd'
I10285
sa(dp20570
S'a'
I38792
sS's'
I17
sS'm'
(lp20571
I1
aI3
aI7
aI9
aI1
aI19
aI125
aI199
aI41
aI117
aI903
aI1131
aI7731
aI14431
aI24753
aI62841
aI50251
asS'd'
I10286
sa(dp20572
S'a'
I38798
sS's'
I17
sS'm'
(lp20573
I1
aI3
aI5
aI5
aI11
aI37
aI19
aI249
aI97
aI59
aI1849
aI1151
aI2171
aI1217
aI31277
aI26547
aI86601
asS'd'
I10287
sa(dp20574
S'a'
I38819
sS's'
I17
sS'm'
(lp20575
I1
aI3
aI3
aI7
aI29
aI35
aI21
aI7
aI93
aI971
aI1155
aI1847
aI89
aI6759
aI29611
aI40793
aI88327
asS'd'
I10288
sa(dp20576
S'a'
I38822
sS's'
I17
sS'm'
(lp20577
I1
aI3
aI5
aI5
aI29
aI23
aI91
aI71
aI479
aI943
aI1839
aI3699
aI2491
aI9603
aI15061
aI43221
aI20435
asS'd'
I10289
sa(dp20578
S'a'
I38839
sS's'
I17
sS'm'
(lp20579
I1
aI3
aI3
aI7
aI29
aI11
aI15
aI83
aI21
aI585
aI501
aI161
aI4797
aI11243
aI14879
aI12519
aI19069
asS'd'
I10290
sa(dp20580
S'a'
I38851
sS's'
I17
sS'm'
(lp20581
I1
aI1
aI5
aI15
aI13
aI37
aI9
aI215
aI433
aI925
aI987
aI3253
aI8027
aI7013
aI20801
aI12891
aI36497
asS'd'
I10291
sa(dp20582
S'a'
I38854
sS's'
I17
sS'm'
(lp20583
I1
aI3
aI7
aI1
aI15
aI31
aI95
aI85
aI355
aI1013
aI1963
aI2943
aI1925
aI13691
aI15237
aI28943
aI63873
asS'd'
I10292
sa(dp20584
S'a'
I38863
sS's'
I17
sS'm'
(lp20585
I1
aI3
aI3
aI1
aI17
aI21
aI99
aI201
aI465
aI819
aI665
aI901
aI2671
aI2457
aI29603
aI35275
aI28339
asS'd'
I10293
sa(dp20586
S'a'
I38865
sS's'
I17
sS'm'
(lp20587
I1
aI1
aI1
aI9
aI5
aI23
aI111
aI107
aI27
aI433
aI1341
aI91
aI6879
aI1933
aI25433
aI37435
aI99061
asS'd'
I10294
sa(dp20588
S'a'
I38887
sS's'
I17
sS'm'
(lp20589
I1
aI3
aI5
aI13
aI11
aI55
aI27
aI151
aI397
aI591
aI89
aI1221
aI5581
aI2701
aI15033
aI41879
aI71415
asS'd'
I10295
sa(dp20590
S'a'
I38896
sS's'
I17
sS'm'
(lp20591
I1
aI3
aI3
aI5
aI17
aI35
aI15
aI119
aI487
aI903
aI875
aI3391
aI7731
aI12181
aI27823
aI32561
aI59133
asS'd'
I10296
sa(dp20592
S'a'
I38901
sS's'
I17
sS'm'
(lp20593
I1
aI1
aI5
aI7
aI15
aI33
aI67
aI53
aI307
aI947
aI857
aI2713
aI803
aI765
aI4175
aI15513
aI23985
asS'd'
I10297
sa(dp20594
S'a'
I38912
sS's'
I17
sS'm'
(lp20595
I1
aI3
aI1
aI15
aI23
aI11
aI15
aI101
aI467
aI429
aI153
aI3205
aI5627
aI7555
aI22515
aI12721
aI7905
asS'd'
I10298
sa(dp20596
S'a'
I38918
sS's'
I17
sS'm'
(lp20597
I1
aI1
aI3
aI7
aI19
aI61
aI55
aI187
aI413
aI49
aI1031
aI3415
aI3903
aI6933
aI20017
aI50429
aI116407
asS'd'
I10299
sa(dp20598
S'a'
I38929
sS's'
I17
sS'm'
(lp20599
I1
aI3
aI1
aI13
aI13
aI15
aI33
aI1
aI403
aI441
aI1969
aI775
aI2209
aI15061
aI15657
aI28819
aI62705
asS'd'
I10300
sa(dp20600
S'a'
I38941
sS's'
I17
sS'm'
(lp20601
I1
aI1
aI5
aI13
aI13
aI47
aI67
aI97
aI87
aI677
aI1639
aI3281
aI1395
aI8499
aI18449
aI49935
aI25775
asS'd'
I10301
sa(dp20602
S'a'
I38948
sS's'
I17
sS'm'
(lp20603
I1
aI3
aI3
aI13
aI7
aI13
aI77
aI45
aI405
aI881
aI293
aI2263
aI6517
aI15415
aI25809
aI5681
aI31327
asS'd'
I10302
sa(dp20604
S'a'
I38965
sS's'
I17
sS'm'
(lp20605
I1
aI3
aI7
aI5
aI5
aI51
aI63
aI171
aI401
aI671
aI1631
aI1735
aI7361
aI8741
aI31933
aI44761
aI12209
asS'd'
I10303
sa(dp20606
S'a'
I38969
sS's'
I17
sS'm'
(lp20607
I1
aI1
aI7
aI3
aI13
aI3
aI39
aI223
aI105
aI781
aI241
aI2895
aI5165
aI12135
aI5683
aI63009
aI58399
asS'd'
I10304
sa(dp20608
S'a'
I38977
sS's'
I17
sS'm'
(lp20609
I1
aI1
aI5
aI11
aI29
aI11
aI37
aI1
aI445
aI157
aI219
aI2269
aI3025
aI5721
aI24539
aI41879
aI128445
asS'd'
I10305
sa(dp20610
S'a'
I38987
sS's'
I17
sS'm'
(lp20611
I1
aI3
aI7
aI15
aI23
aI21
aI125
aI105
aI141
aI101
aI1981
aI3765
aI5349
aI13781
aI10055
aI7069
aI77721
asS'd'
I10306
sa(dp20612
S'a'
I38992
sS's'
I17
sS'm'
(lp20613
I1
aI1
aI7
aI9
aI3
aI37
aI111
aI95
aI33
aI53
aI1021
aI1629
aI6945
aI4657
aI19977
aI36715
aI101401
asS'd'
I10307
sa(dp20614
S'a'
I38995
sS's'
I17
sS'm'
(lp20615
I1
aI3
aI3
aI1
aI9
aI5
aI65
aI77
aI459
aI471
aI1045
aI2351
aI2787
aI13001
aI16211
aI22585
aI116205
asS'd'
I10308
sa(dp20616
S'a'
I39001
sS's'
I17
sS'm'
(lp20617
I1
aI1
aI7
aI9
aI25
aI41
aI21
aI187
aI471
aI997
aI583
aI2243
aI6537
aI11837
aI21089
aI51051
aI98517
asS'd'
I10309
sa(dp20618
S'a'
I39004
sS's'
I17
sS'm'
(lp20619
I1
aI1
aI7
aI5
aI13
aI15
aI81
aI39
aI223
aI935
aI951
aI5
aI4315
aI11789
aI18365
aI49647
aI92461
asS'd'
I10310
sa(dp20620
S'a'
I39011
sS's'
I17
sS'm'
(lp20621
I1
aI3
aI3
aI5
aI11
aI15
aI97
aI245
aI43
aI819
aI1415
aI3287
aI2051
aI15879
aI21727
aI54467
aI53875
asS'd'
I10311
sa(dp20622
S'a'
I39018
sS's'
I17
sS'm'
(lp20623
I1
aI3
aI1
aI11
aI7
aI47
aI125
aI155
aI301
aI469
aI805
aI3789
aI6967
aI12117
aI30369
aI55183
aI12671
asS'd'
I10312
sa(dp20624
S'a'
I39025
sS's'
I17
sS'm'
(lp20625
I1
aI1
aI3
aI13
aI17
aI25
aI45
aI199
aI69
aI1015
aI581
aI3891
aI7743
aI9273
aI7639
aI59055
aI93923
asS'd'
I10313
sa(dp20626
S'a'
I39031
sS's'
I17
sS'm'
(lp20627
I1
aI1
aI5
aI11
aI9
aI47
aI39
aI251
aI489
aI47
aI83
aI2147
aI943
aI1959
aI21361
aI5325
aI106079
asS'd'
I10314
sa(dp20628
S'a'
I39032
sS's'
I17
sS'm'
(lp20629
I1
aI3
aI5
aI13
aI27
aI59
aI35
aI1
aI155
aI367
aI165
aI2665
aI3021
aI1127
aI28585
aI45347
aI66763
asS'd'
I10315
sa(dp20630
S'a'
I39044
sS's'
I17
sS'm'
(lp20631
I1
aI3
aI1
aI5
aI31
aI31
aI15
aI125
aI485
aI581
aI1987
aI2293
aI4573
aI11431
aI20773
aI58661
aI79869
asS'd'
I10316
sa(dp20632
S'a'
I39056
sS's'
I17
sS'm'
(lp20633
I1
aI3
aI5
aI15
aI31
aI11
aI109
aI229
aI11
aI831
aI1545
aI919
aI6125
aI9337
aI4169
aI58041
aI67577
asS'd'
I10317
sa(dp20634
S'a'
I39065
sS's'
I17
sS'm'
(lp20635
I1
aI1
aI1
aI3
aI1
aI43
aI13
aI89
aI89
aI863
aI1607
aI1897
aI4831
aI5239
aI24503
aI51853
aI126983
asS'd'
I10318
sa(dp20636
S'a'
I39066
sS's'
I17
sS'm'
(lp20637
I1
aI1
aI5
aI11
aI11
aI37
aI11
aI253
aI495
aI83
aI941
aI3665
aI5187
aI13679
aI11811
aI29563
aI80571
asS'd'
I10319
sa(dp20638
S'a'
I39084
sS's'
I17
sS'm'
(lp20639
I1
aI3
aI7
aI15
aI3
aI45
aI45
aI157
aI477
aI321
aI1401
aI1133
aI271
aI12455
aI31543
aI18223
aI116701
asS'd'
I10320
sa(dp20640
S'a'
I39095
sS's'
I17
sS'm'
(lp20641
I1
aI1
aI5
aI3
aI7
aI5
aI17
aI127
aI195
aI583
aI715
aI3975
aI6865
aI7617
aI6749
aI15687
aI42375
asS'd'
I10321
sa(dp20642
S'a'
I39099
sS's'
I17
sS'm'
(lp20643
I1
aI1
aI1
aI7
aI5
aI7
aI21
aI163
aI303
aI45
aI1435
aI1345
aI2489
aI15333
aI18459
aI60837
aI104339
asS'd'
I10322
sa(dp20644
S'a'
I39101
sS's'
I17
sS'm'
(lp20645
I1
aI3
aI1
aI1
aI7
aI23
aI39
aI93
aI347
aI947
aI345
aI1713
aI6383
aI15411
aI10849
aI32559
aI126431
asS'd'
I10323
sa(dp20646
S'a'
I39102
sS's'
I17
sS'm'
(lp20647
I1
aI3
aI1
aI5
aI19
aI41
aI119
aI213
aI3
aI991
aI1745
aI3989
aI155
aI7761
aI28179
aI59805
aI106759
asS'd'
I10324
sa(dp20648
S'a'
I39104
sS's'
I17
sS'm'
(lp20649
I1
aI1
aI7
aI5
aI25
aI11
aI105
aI89
aI505
aI711
aI1213
aI2831
aI8087
aI8855
aI31171
aI49749
aI921
asS'd'
I10325
sa(dp20650
S'a'
I39109
sS's'
I17
sS'm'
(lp20651
I1
aI1
aI5
aI5
aI23
aI61
aI49
aI101
aI209
aI805
aI123
aI17
aI805
aI9033
aI25753
aI33261
aI33753
asS'd'
I10326
sa(dp20652
S'a'
I39114
sS's'
I17
sS'm'
(lp20653
I1
aI1
aI3
aI5
aI5
aI17
aI93
aI223
aI179
aI307
aI869
aI1851
aI4313
aI477
aI12925
aI21365
aI103999
asS'd'
I10327
sa(dp20654
S'a'
I39121
sS's'
I17
sS'm'
(lp20655
I1
aI1
aI7
aI13
aI21
aI23
aI105
aI53
aI393
aI939
aI291
aI2407
aI4815
aI4961
aI30305
aI8613
aI62599
asS'd'
I10328
sa(dp20656
S'a'
I39122
sS's'
I17
sS'm'
(lp20657
I1
aI1
aI1
aI11
aI9
aI55
aI55
aI135
aI411
aI225
aI205
aI3357
aI4553
aI10139
aI17649
aI51209
aI94037
asS'd'
I10329
sa(dp20658
S'a'
I39127
sS's'
I17
sS'm'
(lp20659
I1
aI3
aI5
aI15
aI17
aI17
aI119
aI15
aI121
aI581
aI169
aI2495
aI3673
aI7173
aI13099
aI7683
aI53397
asS'd'
I10330
sa(dp20660
S'a'
I39150
sS's'
I17
sS'm'
(lp20661
I1
aI3
aI1
aI11
aI29
aI29
aI119
aI255
aI447
aI85
aI845
aI1015
aI2793
aI2471
aI12639
aI32155
aI99193
asS'd'
I10331
sa(dp20662
S'a'
I39172
sS's'
I17
sS'm'
(lp20663
I1
aI3
aI3
aI3
aI9
aI33
aI77
aI251
aI425
aI1007
aI1003
aI2697
aI4989
aI7799
aI26581
aI15963
aI50443
asS'd'
I10332
sa(dp20664
S'a'
I39181
sS's'
I17
sS'm'
(lp20665
I1
aI3
aI1
aI5
aI13
aI47
aI13
aI203
aI473
aI529
aI147
aI2061
aI343
aI4029
aI14615
aI51355
aI27863
asS'd'
I10333
sa(dp20666
S'a'
I39184
sS's'
I17
sS'm'
(lp20667
I1
aI1
aI3
aI15
aI19
aI63
aI39
aI25
aI241
aI487
aI461
aI3021
aI3545
aI4537
aI8991
aI17689
aI77131
asS'd'
I10334
sa(dp20668
S'a'
I39193
sS's'
I17
sS'm'
(lp20669
I1
aI3
aI5
aI5
aI7
aI1
aI61
aI89
aI495
aI943
aI1061
aI405
aI449
aI12785
aI25151
aI24497
aI65709
asS'd'
I10335
sa(dp20670
S'a'
I39199
sS's'
I17
sS'm'
(lp20671
I1
aI1
aI5
aI3
aI7
aI43
aI51
aI55
aI193
aI615
aI37
aI1377
aI2541
aI3861
aI29447
aI32269
aI106335
asS'd'
I10336
sa(dp20672
S'a'
I39200
sS's'
I17
sS'm'
(lp20673
I1
aI1
aI5
aI11
aI21
aI55
aI103
aI43
aI421
aI673
aI175
aI979
aI5175
aI1301
aI8617
aI55875
aI111095
asS'd'
I10337
sa(dp20674
S'a'
I39206
sS's'
I17
sS'm'
(lp20675
I1
aI3
aI5
aI13
aI29
aI31
aI33
aI241
aI129
aI473
aI201
aI2015
aI447
aI99
aI23781
aI33517
aI107851
asS'd'
I10338
sa(dp20676
S'a'
I39209
sS's'
I17
sS'm'
(lp20677
I1
aI3
aI5
aI3
aI13
aI27
aI125
aI205
aI287
aI957
aI1609
aI2907
aI5481
aI14239
aI19719
aI22459
aI75365
asS'd'
I10339
sa(dp20678
S'a'
I39235
sS's'
I17
sS'm'
(lp20679
I1
aI3
aI3
aI5
aI27
aI23
aI53
aI39
aI329
aI381
aI745
aI517
aI7853
aI5333
aI2773
aI29119
aI7049
asS'd'
I10340
sa(dp20680
S'a'
I39238
sS's'
I17
sS'm'
(lp20681
I1
aI3
aI5
aI15
aI29
aI11
aI89
aI3
aI503
aI755
aI485
aI2669
aI6737
aI16241
aI7345
aI50991
aI107291
asS'd'
I10341
sa(dp20682
S'a'
I39242
sS's'
I17
sS'm'
(lp20683
I1
aI1
aI7
aI3
aI17
aI45
aI11
aI3
aI157
aI715
aI577
aI1309
aI3323
aI9401
aI10573
aI55135
aI100067
asS'd'
I10342
sa(dp20684
S'a'
I39244
sS's'
I17
sS'm'
(lp20685
I1
aI1
aI5
aI9
aI19
aI21
aI49
aI103
aI349
aI503
aI1447
aI675
aI4273
aI7673
aI27507
aI57697
aI80875
asS'd'
I10343
sa(dp20686
S'a'
I39280
sS's'
I17
sS'm'
(lp20687
I1
aI1
aI5
aI1
aI9
aI53
aI19
aI99
aI225
aI915
aI431
aI781
aI3291
aI4383
aI26505
aI50339
aI99799
asS'd'
I10344
sa(dp20688
S'a'
I39290
sS's'
I17
sS'm'
(lp20689
I1
aI3
aI1
aI5
aI7
aI37
aI87
aI201
aI481
aI991
aI1553
aI1867
aI7893
aI13147
aI18647
aI10373
aI51951
asS'd'
I10345
sa(dp20690
S'a'
I39292
sS's'
I17
sS'm'
(lp20691
I1
aI1
aI3
aI13
aI17
aI37
aI19
aI199
aI253
aI901
aI759
aI3545
aI3565
aI10461
aI11867
aI57605
aI75555
asS'd'
I10346
sa(dp20692
S'a'
I39296
sS's'
I17
sS'm'
(lp20693
I1
aI1
aI1
aI5
aI15
aI23
aI115
aI69
aI363
aI673
aI201
aI2451
aI3197
aI10059
aI1667
aI47145
aI89
asS'd'
I10347
sa(dp20694
S'a'
I39305
sS's'
I17
sS'm'
(lp20695
I1
aI1
aI7
aI13
aI19
aI31
aI63
aI35
aI93
aI939
aI1057
aI3221
aI951
aI3575
aI9659
aI7005
aI2087
asS'd'
I10348
sa(dp20696
S'a'
I39308
sS's'
I17
sS'm'
(lp20697
I1
aI1
aI7
aI15
aI15
aI21
aI79
aI7
aI23
aI595
aI1123
aI1909
aI6863
aI7383
aI28067
aI30113
aI107497
asS'd'
I10349
sa(dp20698
S'a'
I39311
sS's'
I17
sS'm'
(lp20699
I1
aI3
aI5
aI7
aI11
aI47
aI41
aI177
aI163
aI855
aI1853
aI3837
aI6995
aI9727
aI27285
aI62667
aI77531
asS'd'
I10350
sa(dp20700
S'a'
I39326
sS's'
I17
sS'm'
(lp20701
I1
aI3
aI7
aI3
aI3
aI29
aI99
aI163
aI95
aI893
aI1049
aI2001
aI2961
aI601
aI4613
aI59745
aI61203
asS'd'
I10351
sa(dp20702
S'a'
I39329
sS's'
I17
sS'm'
(lp20703
I1
aI3
aI5
aI1
aI27
aI5
aI5
aI47
aI119
aI631
aI1171
aI3467
aI2115
aI8581
aI24863
aI64193
aI52093
asS'd'
I10352
sa(dp20704
S'a'
I39330
sS's'
I17
sS'm'
(lp20705
I1
aI1
aI1
aI5
aI9
aI51
aI49
aI251
aI97
aI177
aI311
aI993
aI1103
aI7875
aI25273
aI51587
aI9081
asS'd'
I10353
sa(dp20706
S'a'
I39336
sS's'
I17
sS'm'
(lp20707
I1
aI3
aI7
aI5
aI31
aI21
aI43
aI137
aI143
aI779
aI691
aI2331
aI5073
aI5409
aI13335
aI999
aI127765
asS'd'
I10354
sa(dp20708
S'a'
I39339
sS's'
I17
sS'm'
(lp20709
I1
aI1
aI7
aI1
aI31
aI33
aI27
aI193
aI175
aI755
aI1559
aI659
aI5663
aI10491
aI29209
aI50979
aI116683
asS'd'
I10355
sa(dp20710
S'a'
I39353
sS's'
I17
sS'm'
(lp20711
I1
aI3
aI1
aI7
aI23
aI49
aI1
aI39
aI117
aI45
aI1767
aI3503
aI4901
aI5699
aI23613
aI44195
aI17867
asS'd'
I10356
sa(dp20712
S'a'
I39359
sS's'
I17
sS'm'
(lp20713
I1
aI3
aI7
aI13
aI3
aI5
aI105
aI89
aI343
aI627
aI1117
aI3419
aI2081
aI5747
aI7919
aI44329
aI125133
asS'd'
I10357
sa(dp20714
S'a'
I39374
sS's'
I17
sS'm'
(lp20715
I1
aI3
aI1
aI9
aI13
aI33
aI53
aI203
aI17
aI927
aI127
aI2195
aI415
aI11301
aI15115
aI54467
aI128777
asS'd'
I10358
sa(dp20716
S'a'
I39391
sS's'
I17
sS'm'
(lp20717
I1
aI3
aI7
aI1
aI9
aI41
aI15
aI89
aI403
aI333
aI57
aI1159
aI1325
aI2335
aI10609
aI20485
aI110317
asS'd'
I10359
sa(dp20718
S'a'
I39392
sS's'
I17
sS'm'
(lp20719
I1
aI3
aI3
aI5
aI3
aI61
aI25
aI155
aI477
aI907
aI359
aI359
aI5119
aI8157
aI29945
aI38955
aI106485
asS'd'
I10360
sa(dp20720
S'a'
I39402
sS's'
I17
sS'm'
(lp20721
I1
aI1
aI7
aI5
aI19
aI47
aI91
aI89
aI367
aI703
aI761
aI431
aI6813
aI2983
aI29739
aI52453
aI125935
asS'd'
I10361
sa(dp20722
S'a'
I39409
sS's'
I17
sS'm'
(lp20723
I1
aI1
aI1
aI7
aI7
aI61
aI127
aI239
aI277
aI649
aI1111
aI2319
aI1737
aI5071
aI13469
aI52119
aI48899
asS'd'
I10362
sa(dp20724
S'a'
I39410
sS's'
I17
sS'm'
(lp20725
I1
aI3
aI5
aI15
aI7
aI17
aI21
aI209
aI265
aI895
aI719
aI263
aI6871
aI5835
aI28483
aI49859
aI67619
asS'd'
I10363
sa(dp20726
S'a'
I39415
sS's'
I17
sS'm'
(lp20727
I1
aI3
aI3
aI15
aI3
aI7
aI113
aI109
aI333
aI545
aI597
aI1193
aI7593
aI3961
aI25169
aI64673
aI47839
asS'd'
I10364
sa(dp20728
S'a'
I39419
sS's'
I17
sS'm'
(lp20729
I1
aI1
aI1
aI3
aI15
aI45
aI55
aI41
aI317
aI719
aI1587
aI2953
aI2441
aI1127
aI9183
aI21637
aI69075
asS'd'
I10365
sa(dp20730
S'a'
I39431
sS's'
I17
sS'm'
(lp20731
I1
aI3
aI1
aI9
aI25
aI57
aI59
aI29
aI89
aI833
aI379
aI1085
aI763
aI14747
aI30797
aI24089
aI83367
asS'd'
I10366
sa(dp20732
S'a'
I39432
sS's'
I17
sS'm'
(lp20733
I1
aI1
aI5
aI13
aI29
aI3
aI117
aI239
aI453
aI595
aI243
aI3103
aI6047
aI631
aI22739
aI41669
aI11683
asS'd'
I10367
sa(dp20734
S'a'
I39438
sS's'
I17
sS'm'
(lp20735
I1
aI1
aI3
aI1
aI9
aI53
aI81
aI21
aI67
aI735
aI827
aI3519
aI7991
aI16249
aI4183
aI61295
aI4531
asS'd'
I10368
sa(dp20736
S'a'
I39443
sS's'
I17
sS'm'
(lp20737
I1
aI3
aI5
aI3
aI1
aI57
aI47
aI99
aI91
aI71
aI1421
aI2949
aI5951
aI15439
aI25239
aI26453
aI50199
asS'd'
I10369
sa(dp20738
S'a'
I39445
sS's'
I17
sS'm'
(lp20739
I1
aI1
aI3
aI15
aI21
aI3
aI93
aI21
aI41
aI809
aI855
aI727
aI7797
aI6957
aI15835
aI27175
aI60617
asS'd'
I10370
sa(dp20740
S'a'
I39449
sS's'
I17
sS'm'
(lp20741
I1
aI1
aI1
aI13
aI1
aI3
aI83
aI135
aI197
aI171
aI1459
aI2841
aI5021
aI6961
aI30675
aI38295
aI39555
asS'd'
I10371
sa(dp20742
S'a'
I39461
sS's'
I17
sS'm'
(lp20743
I1
aI1
aI7
aI5
aI7
aI49
aI83
aI83
aI117
aI73
aI639
aI2717
aI651
aI3253
aI31635
aI14427
aI116509
asS'd'
I10372
sa(dp20744
S'a'
I39462
sS's'
I17
sS'm'
(lp20745
I1
aI3
aI1
aI3
aI23
aI63
aI15
aI23
aI433
aI539
aI903
aI2655
aI1787
aI12901
aI12013
aI41315
aI128217
asS'd'
I10373
sa(dp20746
S'a'
I39485
sS's'
I17
sS'm'
(lp20747
I1
aI3
aI1
aI7
aI5
aI19
aI3
aI137
aI493
aI681
aI775
aI3725
aI4855
aI10817
aI25277
aI3631
aI60779
asS'd'
I10374
sa(dp20748
S'a'
I39508
sS's'
I17
sS'm'
(lp20749
I1
aI3
aI3
aI5
aI1
aI7
aI67
aI39
aI309
aI77
aI1679
aI2853
aI3803
aI2065
aI7461
aI1555
aI88219
asS'd'
I10375
sa(dp20750
S'a'
I39515
sS's'
I17
sS'm'
(lp20751
I1
aI1
aI3
aI5
aI3
aI47
aI17
aI193
aI429
aI789
aI1525
aI969
aI7905
aI6523
aI10149
aI64689
aI40037
asS'd'
I10376
sa(dp20752
S'a'
I39522
sS's'
I17
sS'm'
(lp20753
I1
aI3
aI5
aI7
aI21
aI17
aI65
aI61
aI255
aI517
aI1765
aI2603
aI4929
aI11073
aI7871
aI29313
aI84739
asS'd'
I10377
sa(dp20754
S'a'
I39533
sS's'
I17
sS'm'
(lp20755
I1
aI1
aI5
aI7
aI13
aI55
aI111
aI63
aI499
aI9
aI1715
aI957
aI6951
aI15839
aI13531
aI45483
aI17923
asS'd'
I10378
sa(dp20756
S'a'
I39541
sS's'
I17
sS'm'
(lp20757
I1
aI1
aI1
aI1
aI27
aI7
aI13
aI135
aI27
aI259
aI1735
aI3847
aI7931
aI14777
aI15249
aI62367
aI45773
asS'd'
I10379
sa(dp20758
S'a'
I39542
sS's'
I17
sS'm'
(lp20759
I1
aI1
aI5
aI3
aI5
aI7
aI99
aI171
aI491
aI1007
aI195
aI2223
aI2657
aI13557
aI6549
aI47125
aI25117
asS'd'
I10380
sa(dp20760
S'a'
I39557
sS's'
I17
sS'm'
(lp20761
I1
aI1
aI1
aI9
aI13
aI21
aI59
aI205
aI205
aI951
aI1707
aI3387
aI2901
aI5463
aI13403
aI1917
aI90591
asS'd'
I10381
sa(dp20762
S'a'
I39570
sS's'
I17
sS'm'
(lp20763
I1
aI3
aI3
aI5
aI21
aI37
aI71
aI91
aI297
aI885
aI1415
aI355
aI2877
aI9261
aI6485
aI45855
aI90081
asS'd'
I10382
sa(dp20764
S'a'
I39575
sS's'
I17
sS'm'
(lp20765
I1
aI1
aI5
aI9
aI23
aI51
aI107
aI75
aI93
aI1015
aI439
aI3589
aI3307
aI337
aI17247
aI42285
aI85417
asS'd'
I10383
sa(dp20766
S'a'
I39576
sS's'
I17
sS'm'
(lp20767
I1
aI1
aI7
aI15
aI29
aI33
aI51
aI23
aI269
aI35
aI1241
aI1137
aI729
aI14531
aI14603
aI47547
aI17151
asS'd'
I10384
sa(dp20768
S'a'
I39591
sS's'
I17
sS'm'
(lp20769
I1
aI3
aI3
aI1
aI25
aI21
aI51
aI229
aI55
aI561
aI653
aI3289
aI7629
aI15445
aI21115
aI35941
aI113669
asS'd'
I10385
sa(dp20770
S'a'
I39597
sS's'
I17
sS'm'
(lp20771
I1
aI1
aI5
aI15
aI1
aI33
aI119
aI171
aI75
aI621
aI2025
aI3235
aI1895
aI8279
aI13205
aI61085
aI105401
asS'd'
I10386
sa(dp20772
S'a'
I39603
sS's'
I17
sS'm'
(lp20773
I1
aI3
aI1
aI7
aI25
aI33
aI73
aI25
aI1
aI531
aI603
aI77
aI4939
aI5957
aI28065
aI59467
aI66659
asS'd'
I10387
sa(dp20774
S'a'
I39610
sS's'
I17
sS'm'
(lp20775
I1
aI1
aI7
aI3
aI17
aI61
aI103
aI47
aI289
aI39
aI917
aI2515
aI6607
aI1129
aI23361
aI46321
aI81929
asS'd'
I10388
sa(dp20776
S'a'
I39612
sS's'
I17
sS'm'
(lp20777
I1
aI1
aI7
aI5
aI29
aI53
aI5
aI191
aI151
aI19
aI895
aI1215
aI5401
aI9861
aI24751
aI15481
aI34179
asS'd'
I10389
sa(dp20778
S'a'
I39617
sS's'
I17
sS'm'
(lp20779
I1
aI3
aI7
aI9
aI5
aI3
aI29
aI205
aI173
aI547
aI727
aI947
aI5619
aI4181
aI30621
aI5553
aI37587
asS'd'
I10390
sa(dp20780
S'a'
I39620
sS's'
I17
sS'm'
(lp20781
I1
aI1
aI3
aI9
aI13
aI59
aI95
aI145
aI287
aI849
aI1483
aI3375
aI3531
aI6585
aI29565
aI4243
aI88333
asS'd'
I10391
sa(dp20782
S'a'
I39624
sS's'
I17
sS'm'
(lp20783
I1
aI1
aI7
aI11
aI21
aI23
aI59
aI223
aI71
aI743
aI443
aI697
aI7789
aI10371
aI28565
aI45127
aI37967
asS'd'
I10392
sa(dp20784
S'a'
I39627
sS's'
I17
sS'm'
(lp20785
I1
aI3
aI5
aI11
aI15
aI25
aI79
aI71
aI21
aI817
aI751
aI189
aI1769
aI3835
aI21465
aI17991
aI102043
asS'd'
I10393
sa(dp20786
S'a'
I39635
sS's'
I17
sS'm'
(lp20787
I1
aI1
aI7
aI11
aI3
aI19
aI25
aI5
aI261
aI181
aI49
aI261
aI7715
aI2195
aI19771
aI43463
aI36533
asS'd'
I10394
sa(dp20788
S'a'
I39641
sS's'
I17
sS'm'
(lp20789
I1
aI1
aI7
aI13
aI21
aI21
aI15
aI235
aI191
aI197
aI1305
aI1351
aI4511
aI625
aI6613
aI37053
aI59491
asS'd'
I10395
sa(dp20790
S'a'
I39666
sS's'
I17
sS'm'
(lp20791
I1
aI1
aI1
aI13
aI15
aI13
aI93
aI239
aI251
aI1009
aI527
aI1347
aI4173
aI14753
aI27389
aI20397
aI13101
asS'd'
I10396
sa(dp20792
S'a'
I39668
sS's'
I17
sS'm'
(lp20793
I1
aI1
aI3
aI1
aI15
aI11
aI127
aI141
aI277
aI775
aI1419
aI2353
aI6929
aI2265
aI7253
aI19807
aI74853
asS'd'
I10397
sa(dp20794
S'a'
I39686
sS's'
I17
sS'm'
(lp20795
I1
aI3
aI3
aI15
aI15
aI49
aI9
aI183
aI407
aI377
aI675
aI871
aI347
aI3417
aI4409
aI7805
aI40507
asS'd'
I10398
sa(dp20796
S'a'
I39692
sS's'
I17
sS'm'
(lp20797
I1
aI3
aI5
aI3
aI23
aI11
aI81
aI53
aI343
aI681
aI1777
aI3411
aI757
aI10875
aI3581
aI56105
aI79907
asS'd'
I10399
sa(dp20798
S'a'
I39700
sS's'
I17
sS'm'
(lp20799
I1
aI3
aI5
aI1
aI25
aI9
aI109
aI55
aI283
aI311
aI1607
aI2479
aI5675
aI8819
aI10853
aI38719
aI44471
asS'd'
I10400
sa(dp20800
S'a'
I39709
sS's'
I17
sS'm'
(lp20801
I1
aI1
aI7
aI7
aI9
aI53
aI33
aI195
aI503
aI167
aI993
aI3203
aI3885
aI1921
aI1039
aI25785
aI47411
asS'd'
I10401
sa(dp20802
S'a'
I39726
sS's'
I17
sS'm'
(lp20803
I1
aI3
aI3
aI3
aI31
aI3
aI11
aI85
aI475
aI743
aI1825
aI2649
aI2373
aI12177
aI21481
aI35579
aI85803
asS'd'
I10402
sa(dp20804
S'a'
I39728
sS's'
I17
sS'm'
(lp20805
I1
aI3
aI7
aI3
aI23
aI45
aI45
aI207
aI369
aI773
aI1579
aI783
aI2491
aI7441
aI21203
aI57091
aI107413
asS'd'
I10403
sa(dp20806
S'a'
I39737
sS's'
I17
sS'm'
(lp20807
I1
aI1
aI1
aI5
aI19
aI7
aI97
aI213
aI431
aI533
aI637
aI1767
aI4945
aI4693
aI977
aI64781
aI111811
asS'd'
I10404
sa(dp20808
S'a'
I39738
sS's'
I17
sS'm'
(lp20809
I1
aI3
aI7
aI7
aI1
aI55
aI101
aI251
aI153
aI95
aI1043
aI3219
aI3499
aI6297
aI11571
aI9131
aI61899
asS'd'
I10405
sa(dp20810
S'a'
I39755
sS's'
I17
sS'm'
(lp20811
I1
aI3
aI5
aI5
aI25
aI53
aI121
aI255
aI69
aI661
aI799
aI3559
aI2029
aI11701
aI14151
aI37897
aI18333
asS'd'
I10406
sa(dp20812
S'a'
I39757
sS's'
I17
sS'm'
(lp20813
I1
aI1
aI1
aI9
aI21
aI19
aI97
aI21
aI321
aI957
aI1115
aI251
aI5131
aI8465
aI24215
aI34423
aI12747
asS'd'
I10407
sa(dp20814
S'a'
I39758
sS's'
I17
sS'm'
(lp20815
I1
aI3
aI5
aI7
aI17
aI19
aI99
aI135
aI429
aI625
aI1401
aI1025
aI4193
aI2911
aI7349
aI34135
aI9341
asS'd'
I10408
sa(dp20816
S'a'
I39765
sS's'
I17
sS'm'
(lp20817
I1
aI3
aI5
aI1
aI5
aI63
aI97
aI121
aI307
aI547
aI1967
aI641
aI487
aI4627
aI30899
aI62049
aI94343
asS'd'
I10409
sa(dp20818
S'a'
I39766
sS's'
I17
sS'm'
(lp20819
I1
aI3
aI5
aI13
aI1
aI1
aI109
aI23
aI267
aI843
aI271
aI2277
aI855
aI4245
aI2177
aI33633
aI113835
asS'd'
I10410
sa(dp20820
S'a'
I39769
sS's'
I17
sS'm'
(lp20821
I1
aI1
aI3
aI7
aI3
aI27
aI91
aI79
aI27
aI855
aI2025
aI443
aI4797
aI9005
aI27533
aI20497
aI100431
asS'd'
I10411
sa(dp20822
S'a'
I39779
sS's'
I17
sS'm'
(lp20823
I1
aI3
aI3
aI5
aI23
aI7
aI73
aI35
aI395
aI649
aI881
aI2923
aI4065
aI853
aI10829
aI19461
aI82383
asS'd'
I10412
sa(dp20824
S'a'
I39785
sS's'
I17
sS'm'
(lp20825
I1
aI3
aI1
aI5
aI25
aI13
aI85
aI93
aI369
aI259
aI393
aI3233
aI799
aI12409
aI26631
aI64291
aI110133
asS'd'
I10413
sa(dp20826
S'a'
I39794
sS's'
I17
sS'm'
(lp20827
I1
aI3
aI5
aI5
aI31
aI35
aI25
aI239
aI455
aI893
aI573
aI1449
aI3359
aI12077
aI17149
aI12921
aI66931
asS'd'
I10414
sa(dp20828
S'a'
I39796
sS's'
I17
sS'm'
(lp20829
I1
aI1
aI7
aI3
aI25
aI39
aI67
aI87
aI215
aI325
aI627
aI3609
aI4417
aI10021
aI12047
aI64593
aI116525
asS'd'
I10415
sa(dp20830
S'a'
I39809
sS's'
I17
sS'm'
(lp20831
I1
aI1
aI5
aI5
aI23
aI51
aI125
aI247
aI83
aI419
aI655
aI635
aI7053
aI4907
aI12887
aI18083
aI49481
asS'd'
I10416
sa(dp20832
S'a'
I39815
sS's'
I17
sS'm'
(lp20833
I1
aI1
aI7
aI11
aI5
aI25
aI65
aI139
aI235
aI331
aI1885
aI1851
aI1061
aI13265
aI14371
aI23067
aI56757
asS'd'
I10417
sa(dp20834
S'a'
I39829
sS's'
I17
sS'm'
(lp20835
I1
aI1
aI7
aI9
aI11
aI15
aI29
aI255
aI509
aI869
aI561
aI2201
aI487
aI2989
aI14943
aI65373
aI35789
asS'd'
I10418
sa(dp20836
S'a'
I39834
sS's'
I17
sS'm'
(lp20837
I1
aI1
aI1
aI3
aI3
aI33
aI23
aI121
aI129
aI351
aI1481
aI65
aI321
aI15927
aI23849
aI2813
aI98547
asS'd'
I10419
sa(dp20838
S'a'
I39839
sS's'
I17
sS'm'
(lp20839
I1
aI1
aI1
aI3
aI23
aI55
aI121
aI35
aI339
aI87
aI1147
aI401
aI1477
aI10617
aI15943
aI20535
aI89321
asS'd'
I10420
sa(dp20840
S'a'
I39850
sS's'
I17
sS'm'
(lp20841
I1
aI3
aI5
aI15
aI25
aI59
aI111
aI185
aI305
aI47
aI523
aI2801
aI5485
aI625
aI30191
aI58153
aI9019
asS'd'
I10421
sa(dp20842
S'a'
I39857
sS's'
I17
sS'm'
(lp20843
I1
aI1
aI7
aI13
aI15
aI51
aI105
aI55
aI77
aI419
aI1011
aI1117
aI2705
aI15093
aI15629
aI51429
aI58487
asS'd'
I10422
sa(dp20844
S'a'
I39881
sS's'
I17
sS'm'
(lp20845
I1
aI3
aI7
aI5
aI15
aI27
aI19
aI7
aI401
aI295
aI1841
aI1167
aI2133
aI1967
aI6941
aI13571
aI29467
asS'd'
I10423
sa(dp20846
S'a'
I39905
sS's'
I17
sS'm'
(lp20847
I1
aI1
aI5
aI15
aI25
aI43
aI23
aI253
aI173
aI927
aI1299
aI2779
aI5489
aI16135
aI1503
aI51097
aI105751
asS'd'
I10424
sa(dp20848
S'a'
I39912
sS's'
I17
sS'm'
(lp20849
I1
aI3
aI3
aI5
aI9
aI13
aI5
aI13
aI411
aI639
aI1323
aI1495
aI2539
aI15087
aI21489
aI49653
aI76229
asS'd'
I10425
sa(dp20850
S'a'
I39938
sS's'
I17
sS'm'
(lp20851
I1
aI1
aI1
aI11
aI7
aI51
aI47
aI99
aI247
aI541
aI1355
aI2373
aI4121
aI13621
aI7715
aI16763
aI127985
asS'd'
I10426
sa(dp20852
S'a'
I39940
sS's'
I17
sS'm'
(lp20853
I1
aI1
aI3
aI9
aI7
aI1
aI85
aI45
aI269
aI769
aI581
aI2229
aI7143
aI5203
aI22483
aI18511
aI30997
asS'd'
I10427
sa(dp20854
S'a'
I39944
sS's'
I17
sS'm'
(lp20855
I1
aI3
aI5
aI7
aI21
aI41
aI97
aI225
aI109
aI195
aI1197
aI3417
aI7613
aI13225
aI29157
aI18969
aI82045
asS'd'
I10428
sa(dp20856
S'a'
I39955
sS's'
I17
sS'm'
(lp20857
I1
aI3
aI3
aI3
aI17
aI41
aI13
aI77
aI129
aI679
aI1659
aI1299
aI4809
aI8537
aI19081
aI1281
aI70793
asS'd'
I10429
sa(dp20858
S'a'
I39961
sS's'
I17
sS'm'
(lp20859
I1
aI1
aI5
aI5
aI5
aI49
aI5
aI15
aI313
aI941
aI775
aI259
aI6579
aI7745
aI20531
aI51669
aI35257
asS'd'
I10430
sa(dp20860
S'a'
I39977
sS's'
I17
sS'm'
(lp20861
I1
aI1
aI5
aI5
aI17
aI35
aI13
aI235
aI169
aI699
aI1365
aI3907
aI4231
aI10965
aI7737
aI6735
aI4253
asS'd'
I10431
sa(dp20862
S'a'
I39980
sS's'
I17
sS'm'
(lp20863
I1
aI1
aI5
aI3
aI29
aI3
aI1
aI197
aI133
aI935
aI571
aI3977
aI2467
aI2029
aI12803
aI64559
aI6427
asS'd'
I10432
sa(dp20864
S'a'
I39986
sS's'
I17
sS'm'
(lp20865
I1
aI3
aI5
aI5
aI27
aI5
aI69
aI57
aI439
aI925
aI1695
aI827
aI4685
aI10971
aI3011
aI56821
aI92187
asS'd'
I10433
sa(dp20866
S'a'
I39988
sS's'
I17
sS'm'
(lp20867
I1
aI1
aI1
aI3
aI9
aI45
aI77
aI179
aI173
aI1023
aI907
aI1999
aI3913
aI6973
aI26987
aI30237
aI62987
asS'd'
I10434
sa(dp20868
S'a'
I39991
sS's'
I17
sS'm'
(lp20869
I1
aI3
aI7
aI3
aI5
aI21
aI17
aI97
aI433
aI277
aI1515
aI2923
aI8025
aI14119
aI11243
aI3983
aI33943
asS'd'
I10435
sa(dp20870
S'a'
I39998
sS's'
I17
sS'm'
(lp20871
I1
aI1
aI5
aI7
aI15
aI13
aI119
aI169
aI21
aI927
aI439
aI361
aI2655
aI2237
aI19775
aI4157
aI84245
asS'd'
I10436
sa(dp20872
S'a'
I40003
sS's'
I17
sS'm'
(lp20873
I1
aI3
aI5
aI5
aI31
aI41
aI117
aI159
aI421
aI505
aI1617
aI3855
aI7835
aI8105
aI29525
aI56735
aI82335
asS'd'
I10437
sa(dp20874
S'a'
I40005
sS's'
I17
sS'm'
(lp20875
I1
aI1
aI5
aI5
aI1
aI33
aI51
aI3
aI79
aI933
aI389
aI493
aI5969
aI12493
aI26723
aI61159
aI116951
asS'd'
I10438
sa(dp20876
S'a'
I40023
sS's'
I17
sS'm'
(lp20877
I1
aI3
aI7
aI13
aI17
aI23
aI75
aI13
aI355
aI111
aI675
aI3191
aI3931
aI5651
aI17495
aI4595
aI49869
asS'd'
I10439
sa(dp20878
S'a'
I40024
sS's'
I17
sS'm'
(lp20879
I1
aI1
aI7
aI7
aI15
aI21
aI35
aI125
aI89
aI903
aI697
aI3493
aI4043
aI6631
aI4793
aI45655
aI86969
asS'd'
I10440
sa(dp20880
S'a'
I40045
sS's'
I17
sS'm'
(lp20881
I1
aI3
aI1
aI15
aI13
aI43
aI113
aI213
aI451
aI473
aI191
aI2913
aI6391
aI1321
aI29615
aI24791
aI26979
asS'd'
I10441
sa(dp20882
S'a'
I40046
sS's'
I17
sS'm'
(lp20883
I1
aI3
aI3
aI13
aI17
aI25
aI9
aI163
aI163
aI161
aI1647
aI3949
aI1343
aI12881
aI10931
aI31365
aI70013
asS'd'
I10442
sa(dp20884
S'a'
I40058
sS's'
I17
sS'm'
(lp20885
I1
aI3
aI7
aI3
aI3
aI19
aI1
aI121
aI387
aI543
aI1655
aI1797
aI6727
aI2951
aI21925
aI21595
aI73207
asS'd'
I10443
sa(dp20886
S'a'
I40088
sS's'
I17
sS'm'
(lp20887
I1
aI1
aI5
aI9
aI7
aI19
aI91
aI9
aI83
aI893
aI1393
aI163
aI2219
aI7763
aI32395
aI29569
aI98645
asS'd'
I10444
sa(dp20888
S'a'
I40091
sS's'
I17
sS'm'
(lp20889
I1
aI1
aI5
aI7
aI13
aI63
aI91
aI115
aI247
aI387
aI87
aI3239
aI7561
aI297
aI32615
aI48817
aI41761
asS'd'
I10445
sa(dp20890
S'a'
I40098
sS's'
I17
sS'm'
(lp20891
I1
aI3
aI5
aI3
aI21
aI23
aI27
aI141
aI257
aI377
aI1745
aI443
aI897
aI9033
aI1715
aI9225
aI110181
asS'd'
I10446
sa(dp20892
S'a'
I40109
sS's'
I17
sS'm'
(lp20893
I1
aI1
aI7
aI9
aI23
aI49
aI125
aI131
aI225
aI253
aI139
aI2057
aI3273
aI4049
aI6861
aI4463
aI11659
asS'd'
I10447
sa(dp20894
S'a'
I40112
sS's'
I17
sS'm'
(lp20895
I1
aI1
aI5
aI11
aI5
aI41
aI97
aI213
aI133
aI481
aI2009
aI2039
aI1533
aI10765
aI22427
aI23297
aI80661
asS'd'
I10448
sa(dp20896
S'a'
I40124
sS's'
I17
sS'm'
(lp20897
I1
aI1
aI5
aI15
aI9
aI11
aI77
aI129
aI421
aI219
aI1623
aI703
aI1611
aI13377
aI9859
aI42869
aI101943
asS'd'
I10449
sa(dp20898
S'a'
I40130
sS's'
I17
sS'm'
(lp20899
I1
aI3
aI3
aI3
aI17
aI63
aI55
aI29
aI317
aI973
aI1159
aI11
aI1733
aI14551
aI25911
aI39151
aI45861
asS'd'
I10450
sa(dp20900
S'a'
I40153
sS's'
I17
sS'm'
(lp20901
I1
aI3
aI7
aI11
aI29
aI63
aI107
aI193
aI263
aI799
aI1171
aI543
aI553
aI12591
aI21965
aI8165
aI64347
asS'd'
I10451
sa(dp20902
S'a'
I40166
sS's'
I17
sS'm'
(lp20903
I1
aI1
aI7
aI15
aI23
aI49
aI65
aI65
aI401
aI897
aI681
aI1113
aI6737
aI9157
aI1557
aI55891
aI129175
asS'd'
I10452
sa(dp20904
S'a'
I40175
sS's'
I17
sS'm'
(lp20905
I1
aI3
aI3
aI1
aI15
aI23
aI107
aI123
aI313
aI633
aI1009
aI2615
aI1155
aI11701
aI21945
aI7939
aI28111
asS'd'
I10453
sa(dp20906
S'a'
I40183
sS's'
I17
sS'm'
(lp20907
I1
aI3
aI1
aI11
aI15
aI11
aI47
aI137
aI299
aI393
aI877
aI1989
aI5903
aI6505
aI9599
aI4129
aI23073
asS'd'
I10454
sa(dp20908
S'a'
I40184
sS's'
I17
sS'm'
(lp20909
I1
aI1
aI7
aI15
aI9
aI49
aI67
aI15
aI79
aI125
aI505
aI17
aI8071
aI12957
aI13855
aI23611
aI66465
asS'd'
I10455
sa(dp20910
S'a'
I40207
sS's'
I17
sS'm'
(lp20911
I1
aI1
aI5
aI13
aI31
aI49
aI1
aI161
aI121
aI145
aI711
aI1347
aI5297
aI11309
aI9871
aI43075
aI95541
asS'd'
I10456
sa(dp20912
S'a'
I40215
sS's'
I17
sS'm'
(lp20913
I1
aI3
aI3
aI13
aI19
aI7
aI55
aI199
aI469
aI471
aI1269
aI3779
aI6251
aI3513
aI1775
aI19501
aI94055
asS'd'
I10457
sa(dp20914
S'a'
I40225
sS's'
I17
sS'm'
(lp20915
I1
aI3
aI3
aI13
aI9
aI41
aI109
aI211
aI197
aI227
aI1211
aI3327
aI1247
aI12253
aI4493
aI31507
aI38677
asS'd'
I10458
sa(dp20916
S'a'
I40235
sS's'
I17
sS'm'
(lp20917
I1
aI1
aI7
aI3
aI11
aI45
aI11
aI103
aI325
aI849
aI1817
aI3971
aI1059
aI9047
aI27237
aI32211
aI121165
asS'd'
I10459
sa(dp20918
S'a'
I40240
sS's'
I17
sS'm'
(lp20919
I1
aI3
aI3
aI3
aI13
aI43
aI7
aI35
aI293
aI3
aI679
aI1441
aI5189
aI7585
aI32009
aI6151
aI89803
asS'd'
I10460
sa(dp20920
S'a'
I40255
sS's'
I17
sS'm'
(lp20921
I1
aI1
aI7
aI9
aI29
aI41
aI127
aI255
aI363
aI913
aI2027
aI3891
aI5187
aI10233
aI8871
aI48085
aI125609
asS'd'
I10461
sa(dp20922
S'a'
I40263
sS's'
I17
sS'm'
(lp20923
I1
aI3
aI1
aI5
aI21
aI23
aI59
aI145
aI171
aI775
aI535
aI3803
aI6981
aI15901
aI20255
aI63199
aI92905
asS'd'
I10462
sa(dp20924
S'a'
I40270
sS's'
I17
sS'm'
(lp20925
I1
aI3
aI5
aI9
aI7
aI63
aI53
aI7
aI145
aI547
aI1753
aI3351
aI1273
aI8175
aI24103
aI42133
aI87459
asS'd'
I10463
sa(dp20926
S'a'
I40277
sS's'
I17
sS'm'
(lp20927
I1
aI3
aI7
aI7
aI25
aI33
aI5
aI217
aI469
aI473
aI1573
aI2525
aI7345
aI5261
aI7023
aI50893
aI124129
asS'd'
I10464
sa(dp20928
S'a'
I40282
sS's'
I17
sS'm'
(lp20929
I1
aI3
aI5
aI13
aI5
aI51
aI23
aI61
aI429
aI775
aI519
aI2671
aI1979
aI9005
aI21617
aI33611
aI120487
asS'd'
I10465
sa(dp20930
S'a'
I40297
sS's'
I17
sS'm'
(lp20931
I1
aI3
aI3
aI15
aI23
aI1
aI73
aI187
aI47
aI369
aI943
aI99
aI2529
aI5569
aI13649
aI51481
aI128949
asS'd'
I10466
sa(dp20932
S'a'
I40306
sS's'
I17
sS'm'
(lp20933
I1
aI3
aI1
aI5
aI25
aI55
aI35
aI191
aI327
aI845
aI1353
aI261
aI6297
aI6067
aI22241
aI32381
aI17749
asS'd'
I10467
sa(dp20934
S'a'
I40315
sS's'
I17
sS'm'
(lp20935
I1
aI1
aI5
aI15
aI31
aI5
aI29
aI129
aI15
aI47
aI739
aI755
aI7595
aI14743
aI14705
aI34347
aI11805
asS'd'
I10468
sa(dp20936
S'a'
I40333
sS's'
I17
sS'm'
(lp20937
I1
aI3
aI1
aI3
aI15
aI49
aI119
aI47
aI185
aI63
aI2003
aI2847
aI5393
aI855
aI7699
aI29521
aI67011
asS'd'
I10469
sa(dp20938
S'a'
I40334
sS's'
I17
sS'm'
(lp20939
I1
aI3
aI7
aI15
aI11
aI41
aI37
aI149
aI173
aI1015
aI29
aI1805
aI1269
aI16199
aI32337
aI11023
aI60065
asS'd'
I10470
sa(dp20940
S'a'
I40336
sS's'
I17
sS'm'
(lp20941
I1
aI1
aI1
aI7
aI31
aI19
aI65
aI81
aI255
aI875
aI1379
aI2347
aI1873
aI14427
aI29523
aI38413
aI65583
asS'd'
I10471
sa(dp20942
S'a'
I40342
sS's'
I17
sS'm'
(lp20943
I1
aI1
aI1
aI15
aI13
aI59
aI3
aI219
aI127
aI479
aI1029
aI3385
aI563
aI11825
aI10081
aI17423
aI26431
asS'd'
I10472
sa(dp20944
S'a'
I40345
sS's'
I17
sS'm'
(lp20945
I1
aI1
aI1
aI1
aI25
aI27
aI79
aI87
aI489
aI281
aI457
aI3527
aI5117
aI4705
aI21167
aI46211
aI90383
asS'd'
I10473
sa(dp20946
S'a'
I40348
sS's'
I17
sS'm'
(lp20947
I1
aI3
aI7
aI13
aI7
aI5
aI67
aI111
aI53
aI439
aI1483
aI3639
aI7781
aI9471
aI10957
aI60711
aI64957
asS'd'
I10474
sa(dp20948
S'a'
I40355
sS's'
I17
sS'm'
(lp20949
I1
aI3
aI7
aI9
aI7
aI7
aI41
aI137
aI159
aI245
aI551
aI4007
aI1277
aI4743
aI4863
aI48689
aI123289
asS'd'
I10475
sa(dp20950
S'a'
I40372
sS's'
I17
sS'm'
(lp20951
I1
aI3
aI7
aI9
aI15
aI49
aI55
aI77
aI41
aI475
aI1563
aI3569
aI5993
aI301
aI14831
aI44095
aI22641
asS'd'
I10476
sa(dp20952
S'a'
I40381
sS's'
I17
sS'm'
(lp20953
I1
aI1
aI1
aI1
aI15
aI33
aI39
aI135
aI81
aI533
aI869
aI305
aI1125
aI6399
aI14321
aI37217
aI121081
asS'd'
I10477
sa(dp20954
S'a'
I40390
sS's'
I17
sS'm'
(lp20955
I1
aI1
aI7
aI15
aI21
aI59
aI43
aI7
aI225
aI1
aI115
aI1531
aI2931
aI2593
aI15935
aI61973
aI106899
asS'd'
I10478
sa(dp20956
S'a'
I40407
sS's'
I17
sS'm'
(lp20957
I1
aI1
aI1
aI1
aI13
aI13
aI99
aI191
aI437
aI367
aI641
aI1933
aI5807
aI11677
aI13557
aI46475
aI34875
asS'd'
I10479
sa(dp20958
S'a'
I40435
sS's'
I17
sS'm'
(lp20959
I1
aI3
aI7
aI9
aI21
aI7
aI119
aI209
aI31
aI919
aI901
aI1229
aI5823
aI11439
aI18151
aI18991
aI114743
asS'd'
I10480
sa(dp20960
S'a'
I40437
sS's'
I17
sS'm'
(lp20961
I1
aI3
aI3
aI3
aI19
aI37
aI109
aI53
aI411
aI617
aI1841
aI2769
aI1271
aI5719
aI22359
aI1199
aI72405
asS'd'
I10481
sa(dp20962
S'a'
I40441
sS's'
I17
sS'm'
(lp20963
I1
aI1
aI1
aI5
aI29
aI3
aI51
aI59
aI141
aI897
aI1907
aI3799
aI1463
aI5661
aI181
aI50565
aI95085
asS'd'
I10482
sa(dp20964
S'a'
I40444
sS's'
I17
sS'm'
(lp20965
I1
aI1
aI1
aI7
aI1
aI35
aI77
aI225
aI341
aI587
aI137
aI35
aI2177
aI15177
aI12869
aI35013
aI39471
asS'd'
I10483
sa(dp20966
S'a'
I40458
sS's'
I17
sS'm'
(lp20967
I1
aI1
aI3
aI13
aI15
aI63
aI45
aI33
aI337
aI1
aI1133
aI263
aI4985
aI11591
aI1085
aI31197
aI67897
asS'd'
I10484
sa(dp20968
S'a'
I40460
sS's'
I17
sS'm'
(lp20969
I1
aI1
aI5
aI13
aI23
aI11
aI123
aI21
aI185
aI639
aI145
aI3865
aI2999
aI6261
aI23247
aI23055
aI32755
asS'd'
I10485
sa(dp20970
S'a'
I40481
sS's'
I17
sS'm'
(lp20971
I1
aI1
aI5
aI9
aI19
aI21
aI47
aI133
aI281
aI431
aI1661
aI3719
aI3637
aI973
aI9727
aI52627
aI60035
asS'd'
I10486
sa(dp20972
S'a'
I40484
sS's'
I17
sS'm'
(lp20973
I1
aI1
aI3
aI5
aI3
aI19
aI19
aI89
aI63
aI549
aI551
aI3357
aI5665
aI4781
aI22437
aI1149
aI10825
asS'd'
I10487
sa(dp20974
S'a'
I40487
sS's'
I17
sS'm'
(lp20975
I1
aI3
aI5
aI15
aI3
aI25
aI81
aI193
aI11
aI711
aI1481
aI1767
aI1159
aI4967
aI16915
aI3387
aI26245
asS'd'
I10488
sa(dp20976
S'a'
I40493
sS's'
I17
sS'm'
(lp20977
I1
aI1
aI1
aI3
aI29
aI39
aI23
aI131
aI473
aI107
aI765
aI2249
aI6087
aI9145
aI20751
aI21085
aI42989
asS'd'
I10489
sa(dp20978
S'a'
I40494
sS's'
I17
sS'm'
(lp20979
I1
aI3
aI1
aI9
aI7
aI39
aI13
aI199
aI475
aI333
aI269
aI1041
aI5927
aI14039
aI19081
aI9045
aI119645
asS'd'
I10490
sa(dp20980
S'a'
I40501
sS's'
I17
sS'm'
(lp20981
I1
aI1
aI5
aI13
aI11
aI61
aI99
aI71
aI151
aI175
aI1327
aI3397
aI5063
aI10683
aI7895
aI62255
aI85749
asS'd'
I10491
sa(dp20982
S'a'
I40502
sS's'
I17
sS'm'
(lp20983
I1
aI3
aI7
aI9
aI1
aI57
aI21
aI217
aI423
aI467
aI1435
aI2887
aI1567
aI8819
aI19961
aI36507
aI110309
asS'd'
I10492
sa(dp20984
S'a'
I40525
sS's'
I17
sS'm'
(lp20985
I1
aI3
aI3
aI11
aI11
aI35
aI77
aI127
aI153
aI357
aI865
aI1943
aI1947
aI10995
aI13617
aI44347
aI26851
asS'd'
I10493
sa(dp20986
S'a'
I40550
sS's'
I17
sS'm'
(lp20987
I1
aI3
aI1
aI11
aI9
aI43
aI31
aI81
aI123
aI813
aI995
aI169
aI6593
aI13621
aI32195
aI51125
aI53509
asS'd'
I10494
sa(dp20988
S'a'
I40553
sS's'
I17
sS'm'
(lp20989
I1
aI1
aI5
aI5
aI27
aI29
aI77
aI35
aI93
aI545
aI377
aI2345
aI6475
aI15729
aI15103
aI49591
aI101121
asS'd'
I10495
sa(dp20990
S'a'
I40559
sS's'
I17
sS'm'
(lp20991
I1
aI1
aI5
aI13
aI1
aI17
aI97
aI187
aI129
aI173
aI641
aI2937
aI3277
aI15087
aI28111
aI46905
aI112367
asS'd'
I10496
sa(dp20992
S'a'
I40562
sS's'
I17
sS'm'
(lp20993
I1
aI3
aI7
aI7
aI1
aI27
aI75
aI43
aI305
aI431
aI571
aI1327
aI7419
aI3093
aI2691
aI23417
aI11975
asS'd'
I10497
sa(dp20994
S'a'
I40573
sS's'
I17
sS'm'
(lp20995
I1
aI1
aI5
aI15
aI17
aI3
aI91
aI57
aI417
aI87
aI1891
aI1973
aI5765
aI5521
aI21931
aI60011
aI20883
asS'd'
I10498
sa(dp20996
S'a'
I40574
sS's'
I17
sS'm'
(lp20997
I1
aI3
aI1
aI3
aI27
aI13
aI105
aI153
aI495
aI371
aI453
aI1295
aI5675
aI6377
aI8971
aI40505
aI41149
asS'd'
I10499
sa(dp20998
S'a'
I40578
sS's'
I17
sS'm'
(lp20999
I1
aI1
aI1
aI15
aI1
aI17
aI105
aI177
aI41
aI455
aI611
aI3585
aI2307
aI2603
aI20985
aI5581
aI14033
asS'd'
I10500
sa(dp21000
S'a'
I40583
sS's'
I17
sS'm'
(lp21001
I1
aI3
aI3
aI9
aI7
aI41
aI33
aI145
aI307
aI293
aI1321
aI2151
aI3265
aI14845
aI15687
aI38715
aI8041
asS'd'
I10501
sa(dp21002
S'a'
I40584
sS's'
I17
sS'm'
(lp21003
I1
aI3
aI3
aI3
aI5
aI47
aI127
aI253
aI129
aI337
aI1467
aI5
aI2743
aI1921
aI26979
aI11737
aI41479
asS'd'
I10502
sa(dp21004
S'a'
I40587
sS's'
I17
sS'm'
(lp21005
I1
aI1
aI1
aI5
aI15
aI35
aI37
aI9
aI5
aI405
aI1041
aI1903
aI3655
aI14315
aI9441
aI20577
aI50715
asS'd'
I10503
sa(dp21006
S'a'
I40597
sS's'
I17
sS'm'
(lp21007
I1
aI1
aI5
aI15
aI7
aI5
aI53
aI61
aI409
aI353
aI87
aI1805
aI4523
aI11417
aI24105
aI21451
aI56387
asS'd'
I10504
sa(dp21008
S'a'
I40620
sS's'
I17
sS'm'
(lp21009
I1
aI3
aI3
aI5
aI5
aI9
aI25
aI249
aI511
aI795
aI559
aI2695
aI3071
aI3971
aI29421
aI46593
aI96563
asS'd'
I10505
sa(dp21010
S'a'
I40623
sS's'
I17
sS'm'
(lp21011
I1
aI1
aI3
aI1
aI3
aI39
aI61
aI85
aI399
aI105
aI1253
aI3787
aI3065
aI10553
aI8195
aI5637
aI129579
asS'd'
I10506
sa(dp21012
S'a'
I40631
sS's'
I17
sS'm'
(lp21013
I1
aI3
aI3
aI7
aI23
aI23
aI23
aI197
aI263
aI687
aI943
aI1977
aI5767
aI15373
aI17995
aI24509
aI81293
asS'd'
I10507
sa(dp21014
S'a'
I40643
sS's'
I17
sS'm'
(lp21015
I1
aI3
aI1
aI11
aI15
aI37
aI15
aI67
aI207
aI985
aI895
aI509
aI3435
aI11563
aI2055
aI19253
aI42649
asS'd'
I10508
sa(dp21016
S'a'
I40660
sS's'
I17
sS'm'
(lp21017
I1
aI1
aI7
aI3
aI1
aI51
aI59
aI133
aI241
aI569
aI1575
aI3633
aI2243
aI11939
aI5501
aI11249
aI86013
asS'd'
I10509
sa(dp21018
S'a'
I40667
sS's'
I17
sS'm'
(lp21019
I1
aI1
aI7
aI13
aI25
aI59
aI97
aI191
aI385
aI179
aI1195
aI1537
aI1837
aI11953
aI14231
aI37025
aI49803
asS'd'
I10510
sa(dp21020
S'a'
I40676
sS's'
I17
sS'm'
(lp21021
I1
aI3
aI5
aI5
aI13
aI49
aI19
aI171
aI503
aI433
aI1633
aI553
aI2759
aI4379
aI18313
aI62437
aI37453
asS'd'
I10511
sa(dp21022
S'a'
I40693
sS's'
I17
sS'm'
(lp21023
I1
aI3
aI3
aI15
aI29
aI49
aI107
aI239
aI21
aI913
aI1095
aI989
aI4749
aI10657
aI27169
aI15913
aI1573
asS'd'
I10512
sa(dp21024
S'a'
I40697
sS's'
I17
sS'm'
(lp21025
I1
aI1
aI1
aI1
aI3
aI3
aI53
aI241
aI287
aI149
aI557
aI2665
aI2027
aI449
aI29231
aI23025
aI102521
asS'd'
I10513
sa(dp21026
S'a'
I40708
sS's'
I17
sS'm'
(lp21027
I1
aI3
aI5
aI7
aI23
aI21
aI9
aI1
aI11
aI837
aI1337
aI2815
aI7883
aI16053
aI10031
aI43405
aI5037
asS'd'
I10514
sa(dp21028
S'a'
I40718
sS's'
I17
sS'm'
(lp21029
I1
aI3
aI7
aI1
aI23
aI53
aI113
aI125
aI337
aI491
aI1125
aI3083
aI4941
aI951
aI15805
aI1571
aI79779
asS'd'
I10515
sa(dp21030
S'a'
I40726
sS's'
I17
sS'm'
(lp21031
I1
aI3
aI7
aI13
aI1
aI3
aI95
aI105
aI431
aI723
aI1771
aI3773
aI177
aI2045
aI24719
aI57727
aI79005
asS'd'
I10516
sa(dp21032
S'a'
I40735
sS's'
I17
sS'm'
(lp21033
I1
aI3
aI1
aI1
aI7
aI17
aI107
aI171
aI213
aI437
aI409
aI2015
aI7543
aI12693
aI23597
aI44477
aI72543
asS'd'
I10517
sa(dp21034
S'a'
I40739
sS's'
I17
sS'm'
(lp21035
I1
aI3
aI5
aI9
aI7
aI21
aI27
aI167
aI473
aI901
aI1245
aI3737
aI3485
aI14593
aI7619
aI18753
aI14209
asS'd'
I10518
sa(dp21036
S'a'
I40748
sS's'
I17
sS'm'
(lp21037
I1
aI1
aI1
aI3
aI25
aI37
aI51
aI21
aI363
aI73
aI711
aI3749
aI5147
aI8495
aI30151
aI14275
aI128217
asS'd'
I10519
sa(dp21038
S'a'
I40760
sS's'
I17
sS'm'
(lp21039
I1
aI3
aI1
aI13
aI17
aI35
aI69
aI15
aI293
aI331
aI301
aI691
aI7315
aI6495
aI315
aI62909
aI105047
asS'd'
I10520
sa(dp21040
S'a'
I40763
sS's'
I17
sS'm'
(lp21041
I1
aI3
aI5
aI3
aI25
aI23
aI105
aI111
aI213
aI887
aI1701
aI2085
aI5931
aI9217
aI4009
aI2321
aI103631
asS'd'
I10521
sa(dp21042
S'a'
I40773
sS's'
I17
sS'm'
(lp21043
I1
aI1
aI7
aI15
aI17
aI57
aI59
aI249
aI267
aI941
aI777
aI2509
aI6587
aI12033
aI24969
aI31563
aI129049
asS'd'
I10522
sa(dp21044
S'a'
I40774
sS's'
I17
sS'm'
(lp21045
I1
aI1
aI1
aI5
aI31
aI23
aI31
aI217
aI509
aI973
aI659
aI673
aI7759
aI3865
aI21221
aI4319
aI117411
asS'd'
I10523
sa(dp21046
S'a'
I40786
sS's'
I17
sS'm'
(lp21047
I1
aI1
aI3
aI7
aI13
aI13
aI103
aI179
aI107
aI233
aI753
aI3121
aI835
aI13595
aI9271
aI31421
aI45275
asS'd'
I10524
sa(dp21048
S'a'
I40791
sS's'
I17
sS'm'
(lp21049
I1
aI3
aI5
aI13
aI23
aI61
aI125
aI189
aI283
aI83
aI1087
aI755
aI3697
aI14845
aI27901
aI16389
aI82993
asS'd'
I10525
sa(dp21050
S'a'
I40798
sS's'
I17
sS'm'
(lp21051
I1
aI3
aI1
aI3
aI1
aI55
aI25
aI139
aI435
aI681
aI1913
aI975
aI3109
aI6699
aI12943
aI50865
aI71811
asS'd'
I10526
sa(dp21052
S'a'
I40801
sS's'
I17
sS'm'
(lp21053
I1
aI3
aI1
aI5
aI15
aI61
aI17
aI219
aI29
aI805
aI1881
aI3761
aI3535
aI473
aI15629
aI26301
aI51085
asS'd'
I10527
sa(dp21054
S'a'
I40808
sS's'
I17
sS'm'
(lp21055
I1
aI3
aI1
aI1
aI7
aI43
aI87
aI93
aI355
aI247
aI641
aI2851
aI4565
aI9293
aI6025
aI1945
aI112549
asS'd'
I10528
sa(dp21056
S'a'
I40811
sS's'
I17
sS'm'
(lp21057
I1
aI3
aI7
aI5
aI19
aI55
aI69
aI227
aI107
aI443
aI1587
aI2457
aI2873
aI953
aI27529
aI57527
aI54145
asS'd'
I10529
sa(dp21058
S'a'
I40813
sS's'
I17
sS'm'
(lp21059
I1
aI1
aI5
aI9
aI1
aI33
aI31
aI241
aI339
aI791
aI399
aI3435
aI1711
aI10815
aI32657
aI59875
aI31291
asS'd'
I10530
sa(dp21060
S'a'
I40825
sS's'
I17
sS'm'
(lp21061
I1
aI1
aI1
aI7
aI25
aI59
aI87
aI115
aI435
aI47
aI1907
aI193
aI6069
aI10933
aI9877
aI46443
aI3451
asS'd'
I10531
sa(dp21062
S'a'
I40831
sS's'
I17
sS'm'
(lp21063
I1
aI3
aI3
aI15
aI25
aI33
aI19
aI121
aI133
aI253
aI1227
aI75
aI2839
aI3341
aI30727
aI52451
aI44883
asS'd'
I10532
sa(dp21064
S'a'
I40835
sS's'
I17
sS'm'
(lp21065
I1
aI1
aI7
aI11
aI5
aI47
aI97
aI255
aI235
aI565
aI1701
aI529
aI839
aI15473
aI24471
aI5749
aI73135
asS'd'
I10533
sa(dp21066
S'a'
I40856
sS's'
I17
sS'm'
(lp21067
I1
aI1
aI3
aI7
aI21
aI15
aI31
aI81
aI389
aI957
aI603
aI3879
aI2875
aI11987
aI24625
aI53667
aI77775
asS'd'
I10534
sa(dp21068
S'a'
I40861
sS's'
I17
sS'm'
(lp21069
I1
aI1
aI5
aI11
aI29
aI29
aI31
aI233
aI107
aI541
aI561
aI2533
aI1421
aI13587
aI6943
aI45635
aI71315
asS'd'
I10535
sa(dp21070
S'a'
I40880
sS's'
I17
sS'm'
(lp21071
I1
aI3
aI1
aI9
aI25
aI19
aI33
aI53
aI509
aI485
aI1637
aI2877
aI5927
aI16059
aI195
aI17279
aI127025
asS'd'
I10536
sa(dp21072
S'a'
I40889
sS's'
I17
sS'm'
(lp21073
I1
aI1
aI1
aI3
aI9
aI23
aI97
aI101
aI337
aI43
aI1979
aI1139
aI3693
aI2601
aI8225
aI53037
aI63709
asS'd'
I10537
sa(dp21074
S'a'
I40912
sS's'
I17
sS'm'
(lp21075
I1
aI1
aI7
aI7
aI17
aI25
aI121
aI253
aI63
aI105
aI527
aI1397
aI121
aI9665
aI29151
aI10795
aI79077
asS'd'
I10538
sa(dp21076
S'a'
I40918
sS's'
I17
sS'm'
(lp21077
I1
aI3
aI3
aI1
aI27
aI33
aI123
aI69
aI209
aI25
aI1677
aI1569
aI4441
aI7817
aI5165
aI29517
aI117165
asS'd'
I10539
sa(dp21078
S'a'
I40924
sS's'
I17
sS'm'
(lp21079
I1
aI1
aI5
aI15
aI3
aI59
aI13
aI25
aI359
aI71
aI179
aI3925
aI6899
aI6007
aI9121
aI36297
aI88541
asS'd'
I10540
sa(dp21080
S'a'
I40927
sS's'
I17
sS'm'
(lp21081
I1
aI1
aI3
aI11
aI23
aI17
aI55
aI133
aI27
aI277
aI1055
aI1057
aI807
aI1221
aI1665
aI64129
aI102395
asS'd'
I10541
sa(dp21082
S'a'
I40928
sS's'
I17
sS'm'
(lp21083
I1
aI3
aI1
aI15
aI13
aI15
aI105
aI141
aI329
aI73
aI609
aI1663
aI3277
aI1767
aI6371
aI34325
aI109563
asS'd'
I10542
sa(dp21084
S'a'
I40938
sS's'
I17
sS'm'
(lp21085
I1
aI1
aI5
aI1
aI17
aI21
aI37
aI81
aI187
aI403
aI291
aI1495
aI5071
aI14289
aI29075
aI44089
aI95001
asS'd'
I10543
sa(dp21086
S'a'
I40952
sS's'
I17
sS'm'
(lp21087
I1
aI3
aI3
aI3
aI15
aI33
aI49
aI155
aI41
aI853
aI15
aI3571
aI1433
aI8469
aI18711
aI59007
aI98703
asS'd'
I10544
sa(dp21088
S'a'
I40957
sS's'
I17
sS'm'
(lp21089
I1
aI3
aI1
aI13
aI17
aI47
aI61
aI151
aI127
aI87
aI207
aI3157
aI5141
aI14745
aI32567
aI18401
aI7497
asS'd'
I10545
sa(dp21090
S'a'
I40961
sS's'
I17
sS'm'
(lp21091
I1
aI3
aI5
aI1
aI19
aI25
aI49
aI147
aI137
aI603
aI1223
aI3195
aI5965
aI11335
aI20343
aI10109
aI63975
asS'd'
I10546
sa(dp21092
S'a'
I40968
sS's'
I17
sS'm'
(lp21093
I1
aI1
aI7
aI13
aI29
aI59
aI1
aI33
aI157
aI765
aI961
aI641
aI7303
aI3279
aI20287
aI37553
aI114573
asS'd'
I10547
sa(dp21094
S'a'
I40974
sS's'
I17
sS'm'
(lp21095
I1
aI3
aI5
aI1
aI11
aI63
aI63
aI41
aI15
aI717
aI1037
aI227
aI7875
aI8681
aI26943
aI11761
aI28005
asS'd'
I10548
sa(dp21096
S'a'
I40986
sS's'
I17
sS'm'
(lp21097
I1
aI3
aI1
aI3
aI19
aI5
aI67
aI169
aI209
aI293
aI343
aI2033
aI7669
aI1077
aI15513
aI54475
aI15459
asS'd'
I10549
sa(dp21098
S'a'
I40992
sS's'
I17
sS'm'
(lp21099
I1
aI1
aI3
aI3
aI17
aI47
aI49
aI187
aI341
aI767
aI1463
aI301
aI2083
aI9265
aI12313
aI14763
aI126627
asS'd'
I10550
sa(dp21100
S'a'
I41001
sS's'
I17
sS'm'
(lp21101
I1
aI3
aI5
aI13
aI11
aI15
aI45
aI237
aI445
aI55
aI319
aI2989
aI5043
aI1053
aI22809
aI23111
aI7617
asS'd'
I10551
sa(dp21102
S'a'
I41004
sS's'
I17
sS'm'
(lp21103
I1
aI1
aI7
aI9
aI7
aI15
aI41
aI185
aI511
aI701
aI1279
aI1995
aI7829
aI2947
aI3431
aI45799
aI1709
asS'd'
I10552
sa(dp21104
S'a'
I41022
sS's'
I17
sS'm'
(lp21105
I1
aI3
aI7
aI15
aI5
aI15
aI85
aI29
aI487
aI811
aI1653
aI483
aI1193
aI11331
aI21815
aI57215
aI8373
asS'd'
I10553
sa(dp21106
S'a'
I41033
sS's'
I17
sS'm'
(lp21107
I1
aI3
aI1
aI15
aI27
aI19
aI111
aI161
aI19
aI373
aI419
aI1547
aI2415
aI10705
aI17283
aI56663
aI73625
asS'd'
I10554
sa(dp21108
S'a'
I41036
sS's'
I17
sS'm'
(lp21109
I1
aI1
aI3
aI11
aI27
aI7
aI75
aI57
aI411
aI35
aI685
aI1249
aI5227
aI7313
aI3167
aI30537
aI40655
asS'd'
I10555
sa(dp21110
S'a'
I41039
sS's'
I17
sS'm'
(lp21111
I1
aI3
aI1
aI9
aI7
aI37
aI9
aI209
aI353
aI319
aI843
aI657
aI2069
aI6523
aI611
aI16291
aI107121
asS'd'
I10556
sa(dp21112
S'a'
I41044
sS's'
I17
sS'm'
(lp21113
I1
aI1
aI5
aI11
aI11
aI51
aI25
aI171
aI315
aI63
aI207
aI2279
aI2379
aI3583
aI31927
aI62451
aI109911
asS'd'
I10557
sa(dp21114
S'a'
I41064
sS's'
I17
sS'm'
(lp21115
I1
aI1
aI7
aI11
aI15
aI41
aI19
aI175
aI103
aI605
aI1889
aI3161
aI1217
aI3259
aI29655
aI11715
aI35551
asS'd'
I10558
sa(dp21116
S'a'
I41078
sS's'
I17
sS'm'
(lp21117
I1
aI3
aI5
aI13
aI23
aI11
aI121
aI147
aI179
aI397
aI659
aI3753
aI2355
aI1093
aI25863
aI39751
aI112381
asS'd'
I10559
sa(dp21118
S'a'
I41091
sS's'
I17
sS'm'
(lp21119
I1
aI3
aI5
aI7
aI1
aI23
aI37
aI117
aI7
aI361
aI991
aI661
aI4427
aI15333
aI5307
aI55171
aI96959
asS'd'
I10560
sa(dp21120
S'a'
I41103
sS's'
I17
sS'm'
(lp21121
I1
aI3
aI1
aI5
aI17
aI9
aI77
aI147
aI289
aI79
aI295
aI1271
aI7809
aI6387
aI31785
aI26489
aI9335
asS'd'
I10561
sa(dp21122
S'a'
I41108
sS's'
I17
sS'm'
(lp21123
I1
aI1
aI1
aI7
aI17
aI33
aI63
aI147
aI17
aI515
aI1349
aI1907
aI7703
aI5511
aI27773
aI54025
aI30019
asS'd'
I10562
sa(dp21124
S'a'
I41112
sS's'
I17
sS'm'
(lp21125
I1
aI3
aI5
aI3
aI27
aI57
aI75
aI129
aI219
aI533
aI207
aI3569
aI5799
aI6943
aI12271
aI53115
aI120389
asS'd'
I10563
sa(dp21126
S'a'
I41127
sS's'
I17
sS'm'
(lp21127
I1
aI1
aI1
aI13
aI11
aI25
aI101
aI251
aI289
aI215
aI1875
aI1821
aI703
aI15395
aI27167
aI43187
aI63401
asS'd'
I10564
sa(dp21128
S'a'
I41128
sS's'
I17
sS'm'
(lp21129
I1
aI1
aI7
aI15
aI7
aI39
aI125
aI41
aI57
aI513
aI17
aI965
aI3225
aI12833
aI21131
aI53243
aI60377
asS'd'
I10565
sa(dp21130
S'a'
I41136
sS's'
I17
sS'm'
(lp21131
I1
aI3
aI5
aI3
aI21
aI19
aI43
aI195
aI259
aI523
aI587
aI3393
aI6621
aI43
aI10951
aI51877
aI79967
asS'd'
I10566
sa(dp21132
S'a'
I41141
sS's'
I17
sS'm'
(lp21133
I1
aI3
aI3
aI7
aI7
aI19
aI11
aI89
aI321
aI821
aI99
aI2201
aI1297
aI949
aI11539
aI6295
aI19721
asS'd'
I10567
sa(dp21134
S'a'
I41146
sS's'
I17
sS'm'
(lp21135
I1
aI1
aI5
aI3
aI29
aI27
aI123
aI111
aI441
aI441
aI337
aI3849
aI1677
aI14403
aI17203
aI50661
aI92177
asS'd'
I10568
sa(dp21136
S'a'
I41156
sS's'
I17
sS'm'
(lp21137
I1
aI3
aI5
aI9
aI23
aI23
aI73
aI153
aI241
aI841
aI371
aI1503
aI5815
aI14117
aI4679
aI17997
aI112269
asS'd'
I10569
sa(dp21138
S'a'
I41159
sS's'
I17
sS'm'
(lp21139
I1
aI1
aI1
aI1
aI7
aI37
aI105
aI185
aI453
aI905
aI15
aI57
aI6963
aI9665
aI3371
aI2391
aI96023
asS'd'
I10570
sa(dp21140
S'a'
I41163
sS's'
I17
sS'm'
(lp21141
I1
aI3
aI7
aI1
aI1
aI21
aI35
aI43
aI449
aI111
aI191
aI2163
aI3249
aI15049
aI30215
aI43569
aI127973
asS'd'
I10571
sa(dp21142
S'a'
I41165
sS's'
I17
sS'm'
(lp21143
I1
aI3
aI3
aI3
aI17
aI13
aI77
aI123
aI471
aI929
aI1797
aI2061
aI355
aI4441
aI1101
aI24631
aI128711
asS'd'
I10572
sa(dp21144
S'a'
I41166
sS's'
I17
sS'm'
(lp21145
I1
aI3
aI7
aI7
aI17
aI51
aI1
aI69
aI23
aI1003
aI535
aI3751
aI765
aI5253
aI21027
aI52901
aI61951
asS'd'
I10573
sa(dp21146
S'a'
I41184
sS's'
I17
sS'm'
(lp21147
I1
aI1
aI7
aI9
aI25
aI13
aI33
aI13
aI423
aI787
aI223
aI729
aI4443
aI227
aI11487
aI14259
aI52951
asS'd'
I10574
sa(dp21148
S'a'
I41193
sS's'
I17
sS'm'
(lp21149
I1
aI3
aI5
aI5
aI25
aI27
aI113
aI93
aI13
aI679
aI1295
aI3773
aI7253
aI14629
aI8907
aI45885
aI85387
asS'd'
I10575
sa(dp21150
S'a'
I41202
sS's'
I17
sS'm'
(lp21151
I1
aI3
aI3
aI13
aI15
aI55
aI99
aI31
aI119
aI955
aI1477
aI3745
aI6777
aI973
aI4723
aI62133
aI65093
asS'd'
I10576
sa(dp21152
S'a'
I41211
sS's'
I17
sS'm'
(lp21153
I1
aI3
aI3
aI9
aI13
aI51
aI105
aI37
aI477
aI579
aI765
aI2573
aI6869
aI3891
aI30969
aI63413
aI56603
asS'd'
I10577
sa(dp21154
S'a'
I41216
sS's'
I17
sS'm'
(lp21155
I1
aI3
aI1
aI3
aI15
aI23
aI67
aI109
aI75
aI721
aI523
aI1433
aI3455
aI6377
aI23795
aI13711
aI121349
asS'd'
I10578
sa(dp21156
S'a'
I41239
sS's'
I17
sS'm'
(lp21157
I1
aI1
aI3
aI11
aI5
aI5
aI99
aI117
aI233
aI621
aI509
aI3235
aI7483
aI12325
aI13203
aI20075
aI27537
asS'd'
I10579
sa(dp21158
S'a'
I41243
sS's'
I17
sS'm'
(lp21159
I1
aI3
aI3
aI9
aI23
aI51
aI93
aI245
aI307
aI689
aI1993
aI3607
aI1985
aI11839
aI25553
aI54941
aI68741
asS'd'
I10580
sa(dp21160
S'a'
I41249
sS's'
I17
sS'm'
(lp21161
I1
aI1
aI3
aI5
aI19
aI21
aI33
aI71
aI447
aI539
aI351
aI2549
aI87
aI4317
aI1287
aI62289
aI121065
asS'd'
I10581
sa(dp21162
S'a'
I41262
sS's'
I17
sS'm'
(lp21163
I1
aI3
aI5
aI5
aI9
aI23
aI37
aI189
aI449
aI263
aI37
aI3127
aI1709
aI10793
aI7379
aI38565
aI8267
asS'd'
I10582
sa(dp21164
S'a'
I41267
sS's'
I17
sS'm'
(lp21165
I1
aI1
aI7
aI7
aI7
aI33
aI23
aI79
aI457
aI947
aI1275
aI2755
aI3747
aI9225
aI31385
aI8785
aI76945
asS'd'
I10583
sa(dp21166
S'a'
I41276
sS's'
I17
sS'm'
(lp21167
I1
aI3
aI1
aI9
aI17
aI33
aI29
aI59
aI505
aI649
aI1679
aI3609
aI1361
aI5987
aI26455
aI17295
aI98697
asS'd'
I10584
sa(dp21168
S'a'
I41279
sS's'
I17
sS'm'
(lp21169
I1
aI1
aI3
aI11
aI7
aI47
aI127
aI79
aI419
aI143
aI349
aI985
aI6397
aI10271
aI29427
aI19661
aI32629
asS'd'
I10585
sa(dp21170
S'a'
I41305
sS's'
I17
sS'm'
(lp21171
I1
aI1
aI5
aI13
aI15
aI5
aI79
aI171
aI491
aI223
aI1601
aI705
aI623
aI4405
aI10065
aI28057
aI105737
asS'd'
I10586
sa(dp21172
S'a'
I41306
sS's'
I17
sS'm'
(lp21173
I1
aI1
aI7
aI3
aI29
aI7
aI81
aI69
aI265
aI669
aI1763
aI2109
aI6275
aI7683
aI19561
aI26737
aI54449
asS'd'
I10587
sa(dp21174
S'a'
I41312
sS's'
I17
sS'm'
(lp21175
I1
aI1
aI1
aI7
aI1
aI1
aI5
aI9
aI65
aI487
aI1663
aI1021
aI1819
aI9971
aI22065
aI40407
aI4187
asS'd'
I10588
sa(dp21176
S'a'
I41317
sS's'
I17
sS'm'
(lp21177
I1
aI3
aI5
aI5
aI21
aI33
aI11
aI213
aI309
aI575
aI427
aI1421
aI6435
aI981
aI31533
aI16751
aI47813
asS'd'
I10589
sa(dp21178
S'a'
I41321
sS's'
I17
sS'm'
(lp21179
I1
aI3
aI3
aI13
aI7
aI59
aI65
aI65
aI401
aI195
aI211
aI421
aI1139
aI11729
aI19717
aI20699
aI111863
asS'd'
I10590
sa(dp21180
S'a'
I41332
sS's'
I17
sS'm'
(lp21181
I1
aI3
aI7
aI5
aI17
aI51
aI25
aI217
aI223
aI935
aI431
aI1703
aI4869
aI5635
aI199
aI5485
aI37311
asS'd'
I10591
sa(dp21182
S'a'
I41335
sS's'
I17
sS'm'
(lp21183
I1
aI1
aI3
aI11
aI23
aI25
aI15
aI37
aI187
aI1007
aI857
aI3327
aI5471
aI10089
aI13745
aI1741
aI37769
asS'd'
I10592
sa(dp21184
S'a'
I41345
sS's'
I17
sS'm'
(lp21185
I1
aI3
aI5
aI15
aI31
aI17
aI75
aI125
aI1
aI449
aI1293
aI3427
aI709
aI8285
aI31143
aI50655
aI130793
asS'd'
I10593
sa(dp21186
S'a'
I41346
sS's'
I17
sS'm'
(lp21187
I1
aI1
aI7
aI3
aI25
aI55
aI105
aI255
aI319
aI183
aI1571
aI2425
aI5429
aI7151
aI8569
aI37447
aI23055
asS'd'
I10594
sa(dp21188
S'a'
I41351
sS's'
I17
sS'm'
(lp21189
I1
aI3
aI1
aI1
aI23
aI37
aI17
aI61
aI161
aI559
aI1025
aI2651
aI5861
aI5231
aI1365
aI4853
aI127301
asS'd'
I10595
sa(dp21190
S'a'
I41365
sS's'
I17
sS'm'
(lp21191
I1
aI3
aI1
aI9
aI17
aI37
aI87
aI241
aI411
aI53
aI1555
aI3805
aI6867
aI125
aI9829
aI53581
aI117413
asS'd'
I10596
sa(dp21192
S'a'
I41388
sS's'
I17
sS'm'
(lp21193
I1
aI3
aI3
aI3
aI23
aI55
aI121
aI109
aI441
aI623
aI1345
aI3055
aI2591
aI11329
aI16891
aI61347
aI125643
asS'd'
I10597
sa(dp21194
S'a'
I41399
sS's'
I17
sS'm'
(lp21195
I1
aI3
aI1
aI1
aI5
aI29
aI53
aI97
aI15
aI275
aI1587
aI1245
aI379
aI16117
aI24369
aI26873
aI39547
asS'd'
I10598
sa(dp21196
S'a'
I41405
sS's'
I17
sS'm'
(lp21197
I1
aI3
aI1
aI5
aI3
aI63
aI85
aI167
aI301
aI45
aI1357
aI1185
aI3939
aI945
aI24961
aI59427
aI128129
asS'd'
I10599
sa(dp21198
S'a'
I41414
sS's'
I17
sS'm'
(lp21199
I1
aI3
aI1
aI7
aI23
aI25
aI109
aI253
aI37
aI151
aI17
aI1241
aI787
aI15895
aI7947
aI65071
aI14765
asS'd'
I10600
sa(dp21200
S'a'
I41432
sS's'
I17
sS'm'
(lp21201
I1
aI3
aI3
aI1
aI7
aI3
aI103
aI35
aI73
aI533
aI1055
aI823
aI7403
aI8117
aI28813
aI42457
aI56037
asS'd'
I10601
sa(dp21202
S'a'
I41454
sS's'
I17
sS'm'
(lp21203
I1
aI3
aI5
aI15
aI1
aI15
aI97
aI109
aI293
aI259
aI935
aI2977
aI5257
aI14563
aI28871
aI17647
aI34185
asS'd'
I10602
sa(dp21204
S'a'
I41461
sS's'
I17
sS'm'
(lp21205
I1
aI1
aI1
aI3
aI29
aI21
aI101
aI163
aI173
aI1019
aI1025
aI553
aI945
aI3781
aI1097
aI58025
aI124819
asS'd'
I10603
sa(dp21206
S'a'
I41462
sS's'
I17
sS'm'
(lp21207
I1
aI1
aI3
aI9
aI7
aI35
aI65
aI61
aI31
aI547
aI75
aI3515
aI6719
aI12809
aI23287
aI14609
aI30341
asS'd'
I10604
sa(dp21208
S'a'
I41471
sS's'
I17
sS'm'
(lp21209
I1
aI3
aI7
aI9
aI3
aI53
aI21
aI207
aI383
aI917
aI1383
aI2873
aI1663
aI15665
aI1787
aI50741
aI35145
asS'd'
I10605
sa(dp21210
S'a'
I41478
sS's'
I17
sS'm'
(lp21211
I1
aI3
aI7
aI5
aI3
aI35
aI113
aI191
aI171
aI635
aI1597
aI2943
aI2421
aI5555
aI6457
aI22087
aI104221
asS'd'
I10606
sa(dp21212
S'a'
I41490
sS's'
I17
sS'm'
(lp21213
I1
aI1
aI1
aI1
aI29
aI25
aI3
aI225
aI175
aI807
aI1325
aI215
aI6475
aI10729
aI18619
aI45401
aI20627
asS'd'
I10607
sa(dp21214
S'a'
I41506
sS's'
I17
sS'm'
(lp21215
I1
aI1
aI5
aI11
aI23
aI25
aI39
aI207
aI81
aI633
aI403
aI3369
aI1295
aI1289
aI20853
aI48899
aI16613
asS'd'
I10608
sa(dp21216
S'a'
I41508
sS's'
I17
sS'm'
(lp21217
I1
aI1
aI7
aI15
aI5
aI23
aI17
aI77
aI169
aI969
aI1459
aI3795
aI3121
aI5501
aI32323
aI46743
aI124175
asS'd'
I10609
sa(dp21218
S'a'
I41512
sS's'
I17
sS'm'
(lp21219
I1
aI1
aI7
aI13
aI3
aI25
aI77
aI153
aI105
aI1017
aI1599
aI237
aI4691
aI1993
aI6707
aI50265
aI13529
asS'd'
I10610
sa(dp21220
S'a'
I41517
sS's'
I17
sS'm'
(lp21221
I1
aI3
aI3
aI15
aI7
aI11
aI81
aI223
aI61
aI589
aI1263
aI3999
aI7643
aI12101
aI19853
aI49279
aI29999
asS'd'
I10611
sa(dp21222
S'a'
I41520
sS's'
I17
sS'm'
(lp21223
I1
aI3
aI1
aI13
aI3
aI31
aI61
aI59
aI41
aI313
aI115
aI561
aI3973
aI13513
aI6359
aI29395
aI34565
asS'd'
I10612
sa(dp21224
S'a'
I41529
sS's'
I17
sS'm'
(lp21225
I1
aI1
aI7
aI7
aI7
aI61
aI91
aI181
aI307
aI875
aI2045
aI1367
aI3743
aI6497
aI2443
aI12153
aI96431
asS'd'
I10613
sa(dp21226
S'a'
I41530
sS's'
I17
sS'm'
(lp21227
I1
aI1
aI3
aI7
aI19
aI63
aI97
aI211
aI157
aI945
aI891
aI3747
aI5483
aI3081
aI28939
aI11179
aI15935
asS'd'
I10614
sa(dp21228
S'a'
I41544
sS's'
I17
sS'm'
(lp21229
I1
aI3
aI7
aI3
aI23
aI39
aI51
aI137
aI91
aI179
aI1515
aI1397
aI2783
aI9343
aI11483
aI52407
aI111725
asS'd'
I10615
sa(dp21230
S'a'
I41550
sS's'
I17
sS'm'
(lp21231
I1
aI3
aI3
aI11
aI11
aI25
aI111
aI61
aI115
aI329
aI485
aI1713
aI565
aI8607
aI18869
aI6595
aI18605
asS'd'
I10616
sa(dp21232
S'a'
I41571
sS's'
I17
sS'm'
(lp21233
I1
aI1
aI5
aI1
aI13
aI59
aI67
aI231
aI443
aI695
aI1185
aI393
aI6393
aI12957
aI15817
aI37219
aI113127
asS'd'
I10617
sa(dp21234
S'a'
I41577
sS's'
I17
sS'm'
(lp21235
I1
aI3
aI5
aI3
aI15
aI57
aI25
aI97
aI321
aI627
aI15
aI2005
aI3813
aI10399
aI26779
aI24243
aI66463
asS'd'
I10618
sa(dp21236
S'a'
I41580
sS's'
I17
sS'm'
(lp21237
I1
aI3
aI7
aI7
aI17
aI43
aI117
aI179
aI447
aI1005
aI2007
aI1753
aI7685
aI13331
aI5187
aI49341
aI111927
asS'd'
I10619
sa(dp21238
S'a'
I41595
sS's'
I17
sS'm'
(lp21239
I1
aI1
aI3
aI3
aI5
aI53
aI35
aI185
aI93
aI847
aI1523
aI3039
aI25
aI3351
aI23195
aI41133
aI38547
asS'd'
I10620
sa(dp21240
S'a'
I41613
sS's'
I17
sS'm'
(lp21241
I1
aI1
aI7
aI5
aI27
aI59
aI95
aI137
aI55
aI129
aI331
aI127
aI7421
aI5633
aI557
aI18137
aI89055
asS'd'
I10621
sa(dp21242
S'a'
I41622
sS's'
I17
sS'm'
(lp21243
I1
aI3
aI3
aI11
aI5
aI53
aI93
aI137
aI175
aI191
aI1645
aI2047
aI2569
aI8177
aI22691
aI4037
aI31823
asS'd'
I10622
sa(dp21244
S'a'
I41635
sS's'
I17
sS'm'
(lp21245
I1
aI3
aI3
aI11
aI11
aI45
aI77
aI7
aI21
aI541
aI49
aI1689
aI171
aI829
aI28917
aI45095
aI1807
asS'd'
I10623
sa(dp21246
S'a'
I41642
sS's'
I17
sS'm'
(lp21247
I1
aI3
aI7
aI5
aI21
aI5
aI113
aI81
aI33
aI681
aI361
aI1107
aI1597
aI115
aI11503
aI27413
aI9199
asS'd'
I10624
sa(dp21248
S'a'
I41661
sS's'
I17
sS'm'
(lp21249
I1
aI1
aI3
aI11
aI29
aI57
aI15
aI249
aI105
aI683
aI833
aI2579
aI3517
aI16153
aI17373
aI32587
aI124333
asS'd'
I10625
sa(dp21250
S'a'
I41676
sS's'
I17
sS'm'
(lp21251
I1
aI3
aI7
aI13
aI3
aI35
aI55
aI23
aI293
aI5
aI2003
aI2741
aI4237
aI8117
aI20569
aI63967
aI106041
asS'd'
I10626
sa(dp21252
S'a'
I41681
sS's'
I17
sS'm'
(lp21253
I1
aI3
aI3
aI1
aI1
aI15
aI57
aI119
aI135
aI967
aI1495
aI801
aI4959
aI5037
aI10051
aI53915
aI116891
asS'd'
I10627
sa(dp21254
S'a'
I41684
sS's'
I17
sS'm'
(lp21255
I1
aI1
aI7
aI9
aI15
aI29
aI53
aI139
aI505
aI473
aI1179
aI3289
aI369
aI13147
aI15739
aI16739
aI54949
asS'd'
I10628
sa(dp21256
S'a'
I41687
sS's'
I17
sS'm'
(lp21257
I1
aI1
aI5
aI7
aI7
aI45
aI17
aI213
aI381
aI63
aI437
aI3099
aI3765
aI175
aI13521
aI11689
aI58675
asS'd'
I10629
sa(dp21258
S'a'
I41688
sS's'
I17
sS'm'
(lp21259
I1
aI1
aI7
aI1
aI15
aI35
aI55
aI43
aI147
aI873
aI1193
aI3801
aI2301
aI14569
aI31789
aI50443
aI62577
asS'd'
I10630
sa(dp21260
S'a'
I41694
sS's'
I17
sS'm'
(lp21261
I1
aI1
aI5
aI7
aI21
aI41
aI3
aI45
aI43
aI303
aI1465
aI1461
aI5295
aI13397
aI30439
aI7103
aI87505
asS'd'
I10631
sa(dp21262
S'a'
I41698
sS's'
I17
sS'm'
(lp21263
I1
aI1
aI1
aI15
aI19
aI27
aI81
aI141
aI307
aI259
aI521
aI1785
aI6917
aI15635
aI27781
aI64809
aI53297
asS'd'
I10632
sa(dp21264
S'a'
I41710
sS's'
I17
sS'm'
(lp21265
I1
aI1
aI1
aI7
aI27
aI15
aI53
aI99
aI377
aI935
aI1869
aI3835
aI741
aI8447
aI18947
aI10727
aI72179
asS'd'
I10633
sa(dp21266
S'a'
I41712
sS's'
I17
sS'm'
(lp21267
I1
aI1
aI3
aI5
aI15
aI51
aI91
aI207
aI7
aI997
aI935
aI591
aI7325
aI3025
aI11335
aI32087
aI109535
asS'd'
I10634
sa(dp21268
S'a'
I41721
sS's'
I17
sS'm'
(lp21269
I1
aI3
aI1
aI5
aI11
aI13
aI1
aI57
aI45
aI307
aI1839
aI1735
aI2247
aI13117
aI17471
aI16599
aI103063
asS'd'
I10635
sa(dp21270
S'a'
I41722
sS's'
I17
sS'm'
(lp21271
I1
aI3
aI5
aI11
aI19
aI7
aI121
aI3
aI325
aI731
aI1945
aI4025
aI7649
aI8939
aI11147
aI59065
aI49971
asS'd'
I10636
sa(dp21272
S'a'
I41729
sS's'
I17
sS'm'
(lp21273
I1
aI3
aI1
aI5
aI29
aI63
aI95
aI121
aI467
aI7
aI1857
aI2389
aI5213
aI3931
aI21187
aI43529
aI6767
asS'd'
I10637
sa(dp21274
S'a'
I41744
sS's'
I17
sS'm'
(lp21275
I1
aI1
aI7
aI7
aI9
aI53
aI31
aI227
aI95
aI827
aI927
aI3501
aI2003
aI12853
aI2595
aI33223
aI125799
asS'd'
I10638
sa(dp21276
S'a'
I41747
sS's'
I17
sS'm'
(lp21277
I1
aI3
aI3
aI3
aI27
aI25
aI105
aI143
aI233
aI887
aI1135
aI3449
aI5767
aI11447
aI10251
aI34621
aI102113
asS'd'
I10639
sa(dp21278
S'a'
I41753
sS's'
I17
sS'm'
(lp21279
I1
aI3
aI3
aI15
aI3
aI63
aI85
aI119
aI103
aI835
aI443
aI3861
aI4957
aI5389
aI6137
aI48851
aI51887
asS'd'
I10640
sa(dp21280
S'a'
I41766
sS's'
I17
sS'm'
(lp21281
I1
aI3
aI7
aI9
aI23
aI23
aI45
aI129
aI463
aI653
aI1309
aI3533
aI1303
aI2955
aI18023
aI37457
aI114765
asS'd'
I10641
sa(dp21282
S'a'
I41783
sS's'
I17
sS'm'
(lp21283
I1
aI3
aI7
aI1
aI23
aI17
aI31
aI151
aI71
aI515
aI781
aI1793
aI3507
aI6051
aI30279
aI29461
aI48271
asS'd'
I10642
sa(dp21284
S'a'
I41790
sS's'
I17
sS'm'
(lp21285
I1
aI3
aI5
aI15
aI1
aI31
aI9
aI187
aI131
aI571
aI1309
aI965
aI7561
aI16113
aI23209
aI54615
aI16969
asS'd'
I10643
sa(dp21286
S'a'
I41810
sS's'
I17
sS'm'
(lp21287
I1
aI3
aI5
aI11
aI11
aI9
aI109
aI161
aI9
aI697
aI1683
aI1245
aI2223
aI3571
aI18117
aI13085
aI99315
asS'd'
I10644
sa(dp21288
S'a'
I41819
sS's'
I17
sS'm'
(lp21289
I1
aI3
aI3
aI1
aI13
aI21
aI27
aI17
aI11
aI11
aI1095
aI1447
aI6941
aI3399
aI21245
aI36661
aI54283
asS'd'
I10645
sa(dp21290
S'a'
I41825
sS's'
I17
sS'm'
(lp21291
I1
aI3
aI1
aI3
aI21
aI51
aI21
aI197
aI161
aI689
aI1219
aI1337
aI6623
aI5765
aI11579
aI2679
aI23889
asS'd'
I10646
sa(dp21292
S'a'
I41828
sS's'
I17
sS'm'
(lp21293
I1
aI1
aI5
aI11
aI7
aI31
aI101
aI25
aI231
aI719
aI1677
aI1545
aI459
aI14735
aI25153
aI65079
aI15141
asS'd'
I10647
sa(dp21294
S'a'
I41843
sS's'
I17
sS'm'
(lp21295
I1
aI1
aI7
aI9
aI17
aI7
aI49
aI1
aI83
aI829
aI815
aI307
aI3405
aI15189
aI23699
aI50889
aI70391
asS'd'
I10648
sa(dp21296
S'a'
I41846
sS's'
I17
sS'm'
(lp21297
I1
aI1
aI3
aI15
aI21
aI57
aI97
aI191
aI415
aI899
aI197
aI2635
aI7507
aI14009
aI8633
aI48997
aI93925
asS'd'
I10649
sa(dp21298
S'a'
I41862
sS's'
I17
sS'm'
(lp21299
I1
aI3
aI5
aI15
aI23
aI13
aI67
aI127
aI33
aI551
aI911
aI3933
aI2027
aI10665
aI19509
aI18485
aI76111
asS'd'
I10650
sa(dp21300
S'a'
I41871
sS's'
I17
sS'm'
(lp21301
I1
aI1
aI5
aI7
aI23
aI63
aI19
aI149
aI139
aI155
aI1621
aI3391
aI2337
aI2809
aI21161
aI38565
aI401
asS'd'
I10651
sa(dp21302
S'a'
I41874
sS's'
I17
sS'm'
(lp21303
I1
aI1
aI1
aI7
aI19
aI23
aI81
aI49
aI339
aI879
aI1903
aI657
aI2677
aI2273
aI10853
aI3225
aI57933
asS'd'
I10652
sa(dp21304
S'a'
I41876
sS's'
I17
sS'm'
(lp21305
I1
aI3
aI5
aI5
aI13
aI31
aI19
aI203
aI269
aI1015
aI997
aI2151
aI4471
aI11331
aI5363
aI46519
aI51709
asS'd'
I10653
sa(dp21306
S'a'
I41892
sS's'
I17
sS'm'
(lp21307
I1
aI1
aI5
aI11
aI29
aI19
aI17
aI169
aI511
aI389
aI1429
aI2707
aI1341
aI10511
aI6779
aI43345
aI68693
asS'd'
I10654
sa(dp21308
S'a'
I41899
sS's'
I17
sS'm'
(lp21309
I1
aI1
aI5
aI11
aI19
aI25
aI29
aI37
aI423
aI345
aI953
aI2525
aI5937
aI6595
aI31389
aI39347
aI36343
asS'd'
I10655
sa(dp21310
S'a'
I41916
sS's'
I17
sS'm'
(lp21311
I1
aI3
aI1
aI3
aI15
aI25
aI45
aI95
aI111
aI207
aI19
aI1723
aI4113
aI421
aI3297
aI46771
aI8639
asS'd'
I10656
sa(dp21312
S'a'
I41928
sS's'
I17
sS'm'
(lp21313
I1
aI1
aI3
aI9
aI9
aI47
aI27
aI99
aI327
aI393
aI1547
aI1587
aI4463
aI719
aI14609
aI24347
aI68107
asS'd'
I10657
sa(dp21314
S'a'
I41957
sS's'
I17
sS'm'
(lp21315
I1
aI3
aI7
aI7
aI29
aI19
aI57
aI229
aI131
aI497
aI109
aI251
aI6599
aI8947
aI10255
aI12875
aI83831
asS'd'
I10658
sa(dp21316
S'a'
I41964
sS's'
I17
sS'm'
(lp21317
I1
aI3
aI3
aI7
aI17
aI5
aI17
aI45
aI423
aI393
aI1793
aI3
aI603
aI15221
aI13141
aI40585
aI37489
asS'd'
I10659
sa(dp21318
S'a'
I41969
sS's'
I17
sS'm'
(lp21319
I1
aI1
aI1
aI11
aI5
aI1
aI53
aI147
aI129
aI135
aI1473
aI17
aI7539
aI13513
aI16045
aI17375
aI41261
asS'd'
I10660
sa(dp21320
S'a'
I41981
sS's'
I17
sS'm'
(lp21321
I1
aI3
aI1
aI5
aI3
aI15
aI75
aI57
aI47
aI581
aI739
aI3529
aI4323
aI10225
aI27861
aI14431
aI106811
asS'd'
I10661
sa(dp21322
S'a'
I41996
sS's'
I17
sS'm'
(lp21323
I1
aI3
aI3
aI13
aI23
aI57
aI41
aI39
aI217
aI67
aI595
aI1381
aI6281
aI10125
aI30605
aI7935
aI124219
asS'd'
I10662
sa(dp21324
S'a'
I41999
sS's'
I17
sS'm'
(lp21325
I1
aI1
aI7
aI15
aI15
aI45
aI1
aI135
aI495
aI271
aI2023
aI3267
aI39
aI15025
aI32763
aI39023
aI20041
asS'd'
I10663
sa(dp21326
S'a'
I42001
sS's'
I17
sS'm'
(lp21327
I1
aI3
aI7
aI13
aI23
aI53
aI75
aI147
aI187
aI633
aI1989
aI1885
aI6581
aI12169
aI13639
aI19707
aI96429
asS'd'
I10664
sa(dp21328
S'a'
I42017
sS's'
I17
sS'm'
(lp21329
I1
aI1
aI5
aI9
aI13
aI55
aI13
aI41
aI305
aI105
aI1983
aI273
aI35
aI5185
aI22569
aI54203
aI31641
asS'd'
I10665
sa(dp21330
S'a'
I42023
sS's'
I17
sS'm'
(lp21331
I1
aI1
aI3
aI15
aI21
aI19
aI59
aI35
aI165
aI575
aI1961
aI1443
aI4803
aI2339
aI28329
aI47695
aI21505
asS'd'
I10666
sa(dp21332
S'a'
I42027
sS's'
I17
sS'm'
(lp21333
I1
aI3
aI1
aI3
aI23
aI45
aI95
aI85
aI55
aI457
aI1957
aI1243
aI4091
aI14669
aI13213
aI53901
aI122605
asS'd'
I10667
sa(dp21334
S'a'
I42032
sS's'
I17
sS'm'
(lp21335
I1
aI3
aI7
aI1
aI13
aI1
aI61
aI253
aI195
aI839
aI181
aI1153
aI1391
aI205
aI6725
aI1757
aI86817
asS'd'
I10668
sa(dp21336
S'a'
I42035
sS's'
I17
sS'm'
(lp21337
I1
aI1
aI3
aI9
aI7
aI13
aI115
aI137
aI169
aI851
aI299
aI509
aI6709
aI6331
aI51
aI31833
aI25217
asS'd'
I10669
sa(dp21338
S'a'
I42044
sS's'
I17
sS'm'
(lp21339
I1
aI1
aI5
aI15
aI29
aI23
aI119
aI15
aI41
aI585
aI1713
aI1203
aI1653
aI3287
aI25333
aI58873
aI71853
asS'd'
I10670
sa(dp21340
S'a'
I42050
sS's'
I17
sS'm'
(lp21341
I1
aI3
aI5
aI15
aI1
aI45
aI35
aI79
aI97
aI381
aI2027
aI3795
aI2127
aI4775
aI4579
aI63267
aI24719
asS'd'
I10671
sa(dp21342
S'a'
I42061
sS's'
I17
sS'm'
(lp21343
I1
aI1
aI5
aI7
aI17
aI21
aI123
aI75
aI3
aI887
aI1537
aI2017
aI1623
aI16315
aI12535
aI64281
aI54925
asS'd'
I10672
sa(dp21344
S'a'
I42062
sS's'
I17
sS'm'
(lp21345
I1
aI1
aI3
aI13
aI5
aI23
aI117
aI43
aI305
aI365
aI775
aI1599
aI5917
aI13995
aI6353
aI3113
aI106317
asS'd'
I10673
sa(dp21346
S'a'
I42073
sS's'
I17
sS'm'
(lp21347
I1
aI1
aI3
aI11
aI21
aI19
aI9
aI11
aI129
aI349
aI579
aI3523
aI5259
aI8083
aI24513
aI15077
aI115377
asS'd'
I10674
sa(dp21348
S'a'
I42098
sS's'
I17
sS'm'
(lp21349
I1
aI1
aI7
aI9
aI19
aI31
aI107
aI3
aI185
aI821
aI907
aI2389
aI7015
aI3161
aI13603
aI35063
aI60641
asS'd'
I10675
sa(dp21350
S'a'
I42104
sS's'
I17
sS'm'
(lp21351
I1
aI1
aI3
aI1
aI19
aI35
aI105
aI245
aI363
aI745
aI1287
aI4051
aI5201
aI7787
aI20919
aI26567
aI37357
asS'd'
I10676
sa(dp21352
S'a'
I42109
sS's'
I17
sS'm'
(lp21353
I1
aI3
aI1
aI1
aI23
aI31
aI1
aI149
aI61
aI489
aI371
aI987
aI3689
aI14275
aI8581
aI48221
aI44183
asS'd'
I10677
sa(dp21354
S'a'
I42120
sS's'
I17
sS'm'
(lp21355
I1
aI1
aI5
aI3
aI9
aI35
aI51
aI17
aI439
aI355
aI461
aI2129
aI1567
aI13261
aI22347
aI17013
aI53857
asS'd'
I10678
sa(dp21356
S'a'
I42125
sS's'
I17
sS'm'
(lp21357
I1
aI3
aI3
aI15
aI3
aI33
aI59
aI185
aI157
aI933
aI1489
aI647
aI4839
aI12139
aI3145
aI57819
aI11731
asS'd'
I10679
sa(dp21358
S'a'
I42131
sS's'
I17
sS'm'
(lp21359
I1
aI3
aI5
aI15
aI17
aI31
aI59
aI51
aI117
aI1001
aI1585
aI2861
aI2785
aI9579
aI28013
aI4481
aI126723
asS'd'
I10680
sa(dp21360
S'a'
I42143
sS's'
I17
sS'm'
(lp21361
I1
aI3
aI7
aI13
aI27
aI1
aI41
aI119
aI179
aI879
aI1617
aI4053
aI3537
aI15389
aI16381
aI40153
aI68019
asS'd'
I10681
sa(dp21362
S'a'
I42153
sS's'
I17
sS'm'
(lp21363
I1
aI1
aI3
aI13
aI13
aI35
aI45
aI203
aI333
aI337
aI1415
aI1889
aI2361
aI4207
aI10411
aI21013
aI44009
asS'd'
I10682
sa(dp21364
S'a'
I42176
sS's'
I17
sS'm'
(lp21365
I1
aI3
aI3
aI5
aI27
aI9
aI17
aI85
aI331
aI369
aI1219
aI247
aI1977
aI12267
aI1181
aI18811
aI54017
asS'd'
I10683
sa(dp21366
S'a'
I42182
sS's'
I17
sS'm'
(lp21367
I1
aI3
aI5
aI9
aI21
aI57
aI57
aI175
aI283
aI639
aI1155
aI1595
aI8187
aI9981
aI21451
aI7525
aI52751
asS'd'
I10684
sa(dp21368
S'a'
I42188
sS's'
I17
sS'm'
(lp21369
I1
aI3
aI1
aI5
aI27
aI61
aI95
aI25
aI271
aI81
aI1335
aI2821
aI7805
aI10167
aI13197
aI58341
aI62325
asS'd'
I10685
sa(dp21370
S'a'
I42203
sS's'
I17
sS'm'
(lp21371
I1
aI1
aI7
aI3
aI15
aI31
aI75
aI5
aI211
aI663
aI551
aI963
aI6015
aI11907
aI17045
aI22863
aI32389
asS'd'
I10686
sa(dp21372
S'a'
I42216
sS's'
I17
sS'm'
(lp21373
I1
aI1
aI7
aI5
aI21
aI53
aI67
aI71
aI251
aI135
aI1153
aI2247
aI2499
aI15431
aI21419
aI46737
aI2827
asS'd'
I10687
sa(dp21374
S'a'
I42219
sS's'
I17
sS'm'
(lp21375
I1
aI1
aI5
aI3
aI31
aI25
aI39
aI209
aI437
aI791
aI1595
aI637
aI1581
aI6575
aI26407
aI24043
aI11277
asS'd'
I10688
sa(dp21376
S'a'
I42227
sS's'
I17
sS'm'
(lp21377
I1
aI3
aI3
aI5
aI21
aI15
aI13
aI19
aI259
aI949
aI1237
aI239
aI5739
aI4661
aI3405
aI55775
aI58781
asS'd'
I10689
sa(dp21378
S'a'
I42234
sS's'
I17
sS'm'
(lp21379
I1
aI1
aI3
aI5
aI1
aI63
aI5
aI197
aI329
aI625
aI981
aI913
aI3957
aI2765
aI8801
aI56675
aI129511
asS'd'
I10690
sa(dp21380
S'a'
I42251
sS's'
I17
sS'm'
(lp21381
I1
aI3
aI3
aI3
aI29
aI53
aI65
aI145
aI435
aI937
aI787
aI2043
aI4945
aI14585
aI2789
aI15771
aI112335
asS'd'
I10691
sa(dp21382
S'a'
I42254
sS's'
I17
sS'm'
(lp21383
I1
aI3
aI7
aI13
aI3
aI23
aI33
aI141
aI131
aI375
aI739
aI711
aI897
aI469
aI3635
aI43335
aI3069
asS'd'
I10692
sa(dp21384
S'a'
I42256
sS's'
I17
sS'm'
(lp21385
I1
aI1
aI7
aI11
aI29
aI13
aI111
aI149
aI197
aI793
aI1541
aI1879
aI7683
aI9397
aI6873
aI43733
aI118507
asS'd'
I10693
sa(dp21386
S'a'
I42259
sS's'
I17
sS'm'
(lp21387
I1
aI3
aI7
aI7
aI29
aI21
aI97
aI113
aI139
aI573
aI1099
aI2615
aI5123
aI13021
aI9533
aI57673
aI79283
asS'd'
I10694
sa(dp21388
S'a'
I42282
sS's'
I17
sS'm'
(lp21389
I1
aI3
aI1
aI5
aI11
aI9
aI59
aI89
aI469
aI797
aI1119
aI1037
aI1667
aI5947
aI6051
aI65045
aI98275
asS'd'
I10695
sa(dp21390
S'a'
I42289
sS's'
I17
sS'm'
(lp21391
I1
aI3
aI3
aI9
aI11
aI7
aI51
aI191
aI321
aI677
aI1601
aI681
aI3579
aI14441
aI26579
aI18019
aI43065
asS'd'
I10696
sa(dp21392
S'a'
I42302
sS's'
I17
sS'm'
(lp21393
I1
aI3
aI5
aI11
aI7
aI11
aI79
aI21
aI335
aI537
aI801
aI3553
aI4311
aI375
aI7333
aI64839
aI88841
asS'd'
I10697
sa(dp21394
S'a'
I42307
sS's'
I17
sS'm'
(lp21395
I1
aI3
aI1
aI7
aI5
aI11
aI15
aI163
aI69
aI645
aI57
aI3685
aI5143
aI8275
aI12763
aI25035
aI68949
asS'd'
I10698
sa(dp21396
S'a'
I42310
sS's'
I17
sS'm'
(lp21397
I1
aI1
aI3
aI13
aI29
aI33
aI125
aI179
aI431
aI129
aI1367
aI951
aI5843
aI13419
aI13897
aI17315
aI58083
asS'd'
I10699
sa(dp21398
S'a'
I42322
sS's'
I17
sS'm'
(lp21399
I1
aI1
aI3
aI11
aI31
aI33
aI3
aI7
aI185
aI821
aI231
aI869
aI6147
aI15243
aI32029
aI20295
aI60871
asS'd'
I10700
sa(dp21400
S'a'
I42328
sS's'
I17
sS'm'
(lp21401
I1
aI1
aI1
aI1
aI31
aI43
aI21
aI103
aI275
aI573
aI805
aI225
aI2049
aI8375
aI32595
aI53201
aI126487
asS'd'
I10701
sa(dp21402
S'a'
I42338
sS's'
I17
sS'm'
(lp21403
I1
aI1
aI1
aI9
aI31
aI29
aI7
aI91
aI277
aI937
aI1223
aI2435
aI4335
aI7861
aI9647
aI13577
aI30059
asS'd'
I10702
sa(dp21404
S'a'
I42349
sS's'
I17
sS'm'
(lp21405
I1
aI1
aI1
aI1
aI23
aI25
aI69
aI175
aI293
aI905
aI765
aI1527
aI6655
aI15431
aI2511
aI3147
aI75431
asS'd'
I10703
sa(dp21406
S'a'
I42367
sS's'
I17
sS'm'
(lp21407
I1
aI3
aI3
aI3
aI15
aI53
aI109
aI195
aI87
aI557
aI1277
aI1471
aI7401
aI14127
aI11479
aI41505
aI769
asS'd'
I10704
sa(dp21408
S'a'
I42386
sS's'
I17
sS'm'
(lp21409
I1
aI1
aI5
aI11
aI23
aI37
aI121
aI181
aI199
aI359
aI1521
aI2561
aI3641
aI7621
aI14219
aI6959
aI77529
asS'd'
I10705
sa(dp21410
S'a'
I42398
sS's'
I17
sS'm'
(lp21411
I1
aI3
aI1
aI11
aI5
aI7
aI69
aI199
aI501
aI251
aI707
aI1485
aI8125
aI3209
aI30883
aI40259
aI85087
asS'd'
I10706
sa(dp21412
S'a'
I42404
sS's'
I17
sS'm'
(lp21413
I1
aI3
aI5
aI13
aI9
aI35
aI5
aI133
aI505
aI39
aI581
aI1605
aI6303
aI1211
aI27211
aI55591
aI31689
asS'd'
I10707
sa(dp21414
S'a'
I42413
sS's'
I17
sS'm'
(lp21415
I1
aI1
aI5
aI3
aI17
aI7
aI11
aI61
aI483
aI59
aI1569
aI2583
aI759
aI5759
aI3575
aI44547
aI89783
asS'd'
I10708
sa(dp21416
S'a'
I42419
sS's'
I17
sS'm'
(lp21417
I1
aI1
aI7
aI15
aI5
aI27
aI107
aI5
aI471
aI421
aI383
aI3591
aI3609
aI13817
aI633
aI22043
aI83119
asS'd'
I10709
sa(dp21418
S'a'
I42421
sS's'
I17
sS'm'
(lp21419
I1
aI1
aI3
aI7
aI27
aI55
aI61
aI249
aI37
aI241
aI1483
aI2839
aI1231
aI4765
aI1551
aI55801
aI129679
asS'd'
I10710
sa(dp21420
S'a'
I42422
sS's'
I17
sS'm'
(lp21421
I1
aI1
aI1
aI3
aI11
aI1
aI19
aI207
aI143
aI351
aI409
aI721
aI4597
aI13389
aI30297
aI43253
aI129923
asS'd'
I10711
sa(dp21422
S'a'
I42431
sS's'
I17
sS'm'
(lp21423
I1
aI3
aI3
aI7
aI7
aI53
aI83
aI27
aI167
aI163
aI537
aI3871
aI2459
aI12813
aI30019
aI41131
aI56109
asS'd'
I10712
sa(dp21424
S'a'
I42445
sS's'
I17
sS'm'
(lp21425
I1
aI1
aI5
aI1
aI23
aI37
aI11
aI67
aI161
aI751
aI123
aI307
aI3341
aI12983
aI21565
aI58529
aI94503
asS'd'
I10713
sa(dp21426
S'a'
I42448
sS's'
I17
sS'm'
(lp21427
I1
aI3
aI3
aI15
aI11
aI33
aI39
aI195
aI467
aI647
aI1479
aI1197
aI7949
aI6501
aI18375
aI15263
aI21121
asS'd'
I10714
sa(dp21428
S'a'
I42451
sS's'
I17
sS'm'
(lp21429
I1
aI3
aI5
aI13
aI3
aI35
aI9
aI253
aI299
aI679
aI69
aI165
aI2735
aI14725
aI4217
aI16391
aI107017
asS'd'
I10715
sa(dp21430
S'a'
I42454
sS's'
I17
sS'm'
(lp21431
I1
aI1
aI1
aI15
aI3
aI11
aI87
aI87
aI391
aI515
aI843
aI3957
aI1365
aI13201
aI15983
aI53647
aI35643
asS'd'
I10716
sa(dp21432
S'a'
I42458
sS's'
I17
sS'm'
(lp21433
I1
aI1
aI3
aI7
aI9
aI53
aI45
aI221
aI209
aI855
aI169
aI2729
aI1219
aI5229
aI14111
aI28877
aI114653
asS'd'
I10717
sa(dp21434
S'a'
I42470
sS's'
I17
sS'm'
(lp21435
I1
aI1
aI5
aI3
aI11
aI17
aI5
aI93
aI303
aI785
aI1895
aI2483
aI7399
aI14031
aI1007
aI2743
aI47307
asS'd'
I10718
sa(dp21436
S'a'
I42476
sS's'
I17
sS'm'
(lp21437
I1
aI1
aI7
aI11
aI9
aI13
aI115
aI31
aI223
aI1011
aI723
aI1291
aI5183
aI559
aI15881
aI43045
aI28131
asS'd'
I10719
sa(dp21438
S'a'
I42500
sS's'
I17
sS'm'
(lp21439
I1
aI3
aI7
aI11
aI7
aI59
aI85
aI111
aI79
aI227
aI691
aI1597
aI2453
aI10023
aI19255
aI47781
aI88351
asS'd'
I10720
sa(dp21440
S'a'
I42509
sS's'
I17
sS'm'
(lp21441
I1
aI3
aI3
aI7
aI21
aI33
aI39
aI35
aI253
aI743
aI563
aI2455
aI8015
aI13403
aI24883
aI47881
aI115559
asS'd'
I10721
sa(dp21442
S'a'
I42538
sS's'
I17
sS'm'
(lp21443
I1
aI3
aI1
aI1
aI5
aI33
aI69
aI37
aI225
aI157
aI1347
aI3241
aI4981
aI15985
aI9949
aI49189
aI21267
asS'd'
I10722
sa(dp21444
S'a'
I42543
sS's'
I17
sS'm'
(lp21445
I1
aI1
aI3
aI11
aI9
aI33
aI123
aI133
aI215
aI297
aI961
aI1571
aI1133
aI1
aI31871
aI25253
aI100097
asS'd'
I10723
sa(dp21446
S'a'
I42545
sS's'
I17
sS'm'
(lp21447
I1
aI1
aI1
aI7
aI13
aI29
aI101
aI127
aI113
aI785
aI1257
aI525
aI7397
aI13143
aI30315
aI5969
aI37829
asS'd'
I10724
sa(dp21448
S'a'
I42546
sS's'
I17
sS'm'
(lp21449
I1
aI1
aI1
aI7
aI29
aI33
aI17
aI95
aI439
aI577
aI1857
aI423
aI63
aI15365
aI4777
aI59073
aI7773
asS'd'
I10725
sa(dp21450
S'a'
I42563
sS's'
I17
sS'm'
(lp21451
I1
aI1
aI5
aI15
aI3
aI17
aI89
aI133
aI217
aI601
aI1979
aI391
aI105
aI13709
aI10081
aI37725
aI40957
asS'd'
I10726
sa(dp21452
S'a'
I42570
sS's'
I17
sS'm'
(lp21453
I1
aI1
aI1
aI15
aI25
aI7
aI85
aI197
aI155
aI367
aI1927
aI2007
aI2563
aI13147
aI2345
aI28735
aI88243
asS'd'
I10727
sa(dp21454
S'a'
I42580
sS's'
I17
sS'm'
(lp21455
I1
aI3
aI5
aI3
aI5
aI33
aI87
aI153
aI153
aI779
aI825
aI2163
aI385
aI11663
aI2005
aI51261
aI25893
asS'd'
I10728
sa(dp21456
S'a'
I42584
sS's'
I17
sS'm'
(lp21457
I1
aI3
aI5
aI5
aI23
aI15
aI19
aI99
aI71
aI723
aI523
aI3683
aI7773
aI191
aI17423
aI30497
aI129889
asS'd'
I10729
sa(dp21458
S'a'
I42589
sS's'
I17
sS'm'
(lp21459
I1
aI1
aI7
aI11
aI1
aI3
aI49
aI119
aI39
aI661
aI297
aI27
aI1575
aI12145
aI18519
aI57285
aI50059
asS'd'
I10730
sa(dp21460
S'a'
I42608
sS's'
I17
sS'm'
(lp21461
I1
aI3
aI7
aI5
aI7
aI37
aI75
aI235
aI403
aI743
aI603
aI1689
aI5031
aI8871
aI28241
aI16917
aI16947
asS'd'
I10731
sa(dp21462
S'a'
I42618
sS's'
I17
sS'm'
(lp21463
I1
aI1
aI5
aI13
aI17
aI41
aI67
aI219
aI237
aI365
aI833
aI3521
aI3211
aI1037
aI5657
aI34789
aI119739
asS'd'
I10732
sa(dp21464
S'a'
I42629
sS's'
I17
sS'm'
(lp21465
I1
aI1
aI5
aI7
aI3
aI61
aI89
aI107
aI335
aI825
aI803
aI2445
aI6861
aI5421
aI14585
aI44037
aI92711
asS'd'
I10733
sa(dp21466
S'a'
I42636
sS's'
I17
sS'm'
(lp21467
I1
aI3
aI7
aI3
aI19
aI25
aI81
aI51
aI101
aI477
aI1653
aI2841
aI6597
aI9261
aI30609
aI15681
aI48897
asS'd'
I10734
sa(dp21468
S'a'
I42639
sS's'
I17
sS'm'
(lp21469
I1
aI1
aI7
aI11
aI17
aI1
aI43
aI39
aI133
aI513
aI1839
aI553
aI6379
aI4865
aI28161
aI7249
aI80073
asS'd'
I10735
sa(dp21470
S'a'
I42644
sS's'
I17
sS'm'
(lp21471
I1
aI1
aI5
aI5
aI13
aI45
aI19
aI225
aI399
aI679
aI195
aI3613
aI413
aI2901
aI26749
aI39971
aI31435
asS'd'
I10736
sa(dp21472
S'a'
I42647
sS's'
I17
sS'm'
(lp21473
I1
aI3
aI7
aI3
aI23
aI55
aI57
aI77
aI447
aI721
aI677
aI271
aI6211
aI12631
aI5843
aI35991
aI82653
asS'd'
I10737
sa(dp21474
S'a'
I42651
sS's'
I17
sS'm'
(lp21475
I1
aI1
aI1
aI1
aI3
aI63
aI23
aI195
aI1
aI1019
aI723
aI3865
aI5913
aI5491
aI5495
aI27483
aI73637
asS'd'
I10738
sa(dp21476
S'a'
I42654
sS's'
I17
sS'm'
(lp21477
I1
aI3
aI1
aI11
aI17
aI31
aI27
aI211
aI411
aI789
aI1049
aI2487
aI2203
aI6457
aI7275
aI4833
aI14131
asS'd'
I10739
sa(dp21478
S'a'
I42658
sS's'
I17
sS'm'
(lp21479
I1
aI1
aI5
aI15
aI15
aI13
aI65
aI155
aI127
aI753
aI1605
aI1859
aI2873
aI9197
aI1763
aI11969
aI82971
asS'd'
I10740
sa(dp21480
S'a'
I42669
sS's'
I17
sS'm'
(lp21481
I1
aI1
aI3
aI11
aI11
aI63
aI13
aI29
aI31
aI851
aI251
aI3231
aI1227
aI5513
aI9785
aI34659
aI123811
asS'd'
I10741
sa(dp21482
S'a'
I42678
sS's'
I17
sS'm'
(lp21483
I1
aI3
aI5
aI1
aI19
aI57
aI41
aI205
aI91
aI39
aI989
aI1897
aI4789
aI16071
aI6507
aI29363
aI75773
asS'd'
I10742
sa(dp21484
S'a'
I42689
sS's'
I17
sS'm'
(lp21485
I1
aI1
aI1
aI1
aI5
aI29
aI113
aI203
aI53
aI599
aI1529
aI1417
aI7017
aI9609
aI4867
aI17659
aI80719
asS'd'
I10743
sa(dp21486
S'a'
I42695
sS's'
I17
sS'm'
(lp21487
I1
aI3
aI7
aI9
aI27
aI17
aI77
aI25
aI461
aI511
aI781
aI2977
aI7601
aI3551
aI23615
aI57669
aI119723
asS'd'
I10744
sa(dp21488
S'a'
I42696
sS's'
I17
sS'm'
(lp21489
I1
aI3
aI3
aI9
aI23
aI43
aI115
aI21
aI125
aI237
aI893
aI1431
aI7423
aI3717
aI4371
aI36193
aI30481
asS'd'
I10745
sa(dp21490
S'a'
I42710
sS's'
I17
sS'm'
(lp21491
I1
aI1
aI5
aI13
aI3
aI37
aI13
aI239
aI267
aI665
aI205
aI2745
aI3865
aI12167
aI26689
aI999
aI9355
asS'd'
I10746
sa(dp21492
S'a'
I42716
sS's'
I17
sS'm'
(lp21493
I1
aI1
aI1
aI1
aI31
aI35
aI55
aI115
aI387
aI217
aI657
aI2827
aI2963
aI3687
aI24271
aI41701
aI5911
asS'd'
I10747
sa(dp21494
S'a'
I42730
sS's'
I17
sS'm'
(lp21495
I1
aI1
aI3
aI3
aI27
aI57
aI41
aI183
aI351
aI841
aI1327
aI719
aI7043
aI12503
aI17953
aI60719
aI98223
asS'd'
I10748
sa(dp21496
S'a'
I42732
sS's'
I17
sS'm'
(lp21497
I1
aI3
aI1
aI1
aI27
aI1
aI119
aI85
aI197
aI673
aI1951
aI2949
aI4783
aI561
aI12807
aI43355
aI63397
asS'd'
I10749
sa(dp21498
S'a'
I42747
sS's'
I17
sS'm'
(lp21499
I1
aI1
aI7
aI7
aI17
aI63
aI109
aI87
aI303
aI439
aI529
aI685
aI111
aI8405
aI21249
aI33803
aI77927
asS'd'
I10750
sa(dp21500
S'a'
I42750
sS's'
I17
sS'm'
(lp21501
I1
aI1
aI7
aI9
aI11
aI63
aI27
aI185
aI445
aI25
aI1313
aI3979
aI4229
aI8797
aI10671
aI33995
aI84463
asS'd'
I10751
sa(dp21502
S'a'
I42752
sS's'
I17
sS'm'
(lp21503
I1
aI1
aI1
aI15
aI27
aI63
aI67
aI237
aI39
aI993
aI851
aI4075
aI3417
aI1077
aI11939
aI31737
aI93897
asS'd'
I10752
sa(dp21504
S'a'
I42761
sS's'
I17
sS'm'
(lp21505
I1
aI1
aI3
aI5
aI25
aI9
aI51
aI241
aI213
aI661
aI1135
aI213
aI7027
aI5933
aI24485
aI65029
aI8583
asS'd'
I10753
sa(dp21506
S'a'
I42772
sS's'
I17
sS'm'
(lp21507
I1
aI3
aI5
aI11
aI31
aI1
aI17
aI237
aI107
aI1021
aI279
aI181
aI1741
aI11099
aI7871
aI63231
aI64445
asS'd'
I10754
sa(dp21508
S'a'
I42776
sS's'
I17
sS'm'
(lp21509
I1
aI3
aI5
aI9
aI17
aI21
aI11
aI45
aI23
aI409
aI519
aI1703
aI5467
aI9591
aI13555
aI23739
aI73837
asS'd'
I10755
sa(dp21510
S'a'
I42779
sS's'
I17
sS'm'
(lp21511
I1
aI3
aI3
aI15
aI3
aI39
aI11
aI157
aI273
aI241
aI413
aI1723
aI3179
aI2125
aI16859
aI5231
aI122969
asS'd'
I10756
sa(dp21512
S'a'
I42797
sS's'
I17
sS'm'
(lp21513
I1
aI3
aI5
aI11
aI21
aI27
aI29
aI243
aI255
aI1011
aI1179
aI3545
aI3557
aI8091
aI31569
aI10217
aI108361
asS'd'
I10757
sa(dp21514
S'a'
I42815
sS's'
I17
sS'm'
(lp21515
I1
aI1
aI5
aI9
aI25
aI33
aI29
aI67
aI395
aI123
aI1405
aI3855
aI7481
aI5601
aI21231
aI17099
aI13399
asS'd'
I10758
sa(dp21516
S'a'
I42824
sS's'
I17
sS'm'
(lp21517
I1
aI1
aI5
aI5
aI13
aI17
aI111
aI47
aI77
aI827
aI577
aI1767
aI3367
aI11719
aI8801
aI22431
aI85451
asS'd'
I10759
sa(dp21518
S'a'
I42837
sS's'
I17
sS'm'
(lp21519
I1
aI3
aI7
aI11
aI11
aI31
aI17
aI141
aI149
aI293
aI55
aI3459
aI19
aI13709
aI29135
aI62765
aI66455
asS'd'
I10760
sa(dp21520
S'a'
I42844
sS's'
I17
sS'm'
(lp21521
I1
aI1
aI7
aI15
aI13
aI19
aI59
aI211
aI189
aI773
aI1791
aI2089
aI2857
aI1635
aI17777
aI46585
aI70115
asS'd'
I10761
sa(dp21522
S'a'
I42868
sS's'
I17
sS'm'
(lp21523
I1
aI1
aI5
aI11
aI29
aI29
aI15
aI7
aI93
aI733
aI1605
aI3731
aI2381
aI1063
aI15565
aI25081
aI46651
asS'd'
I10762
sa(dp21524
S'a'
I42877
sS's'
I17
sS'm'
(lp21525
I1
aI3
aI1
aI9
aI25
aI5
aI87
aI113
aI25
aI93
aI881
aI1137
aI3237
aI10983
aI14317
aI25945
aI121493
asS'd'
I10763
sa(dp21526
S'a'
I42888
sS's'
I17
sS'm'
(lp21527
I1
aI1
aI5
aI11
aI29
aI47
aI99
aI111
aI165
aI453
aI259
aI2001
aI7715
aI2609
aI15633
aI40273
aI2065
asS'd'
I10764
sa(dp21528
S'a'
I42891
sS's'
I17
sS'm'
(lp21529
I1
aI1
aI7
aI13
aI11
aI29
aI33
aI255
aI149
aI361
aI89
aI2837
aI49
aI3033
aI1917
aI9029
aI38123
asS'd'
I10765
sa(dp21530
S'a'
I42912
sS's'
I17
sS'm'
(lp21531
I1
aI1
aI1
aI7
aI27
aI31
aI105
aI61
aI469
aI497
aI1919
aI3005
aI3651
aI2143
aI24359
aI8053
aI103647
asS'd'
I10766
sa(dp21532
S'a'
I42918
sS's'
I17
sS'm'
(lp21533
I1
aI1
aI3
aI13
aI31
aI63
aI101
aI47
aI397
aI89
aI1915
aI2385
aI5399
aI8897
aI21001
aI42997
aI110333
asS'd'
I10767
sa(dp21534
S'a'
I42921
sS's'
I17
sS'm'
(lp21535
I1
aI3
aI7
aI5
aI29
aI1
aI5
aI119
aI493
aI349
aI153
aI1839
aI283
aI14343
aI12975
aI55597
aI89467
asS'd'
I10768
sa(dp21536
S'a'
I42927
sS's'
I17
sS'm'
(lp21537
I1
aI3
aI5
aI3
aI5
aI51
aI71
aI227
aI63
aI799
aI745
aI1387
aI2435
aI1003
aI27937
aI43421
aI12279
asS'd'
I10769
sa(dp21538
S'a'
I42949
sS's'
I17
sS'm'
(lp21539
I1
aI3
aI3
aI7
aI7
aI31
aI37
aI61
aI11
aI175
aI581
aI1583
aI4737
aI3087
aI10335
aI60683
aI57085
asS'd'
I10770
sa(dp21540
S'a'
I42953
sS's'
I17
sS'm'
(lp21541
I1
aI3
aI1
aI1
aI1
aI63
aI59
aI47
aI417
aI35
aI1673
aI3277
aI1873
aI14981
aI22463
aI26835
aI91115
asS'd'
I10771
sa(dp21542
S'a'
I42967
sS's'
I17
sS'm'
(lp21543
I1
aI1
aI7
aI5
aI15
aI23
aI115
aI13
aI253
aI583
aI219
aI1307
aI1189
aI9891
aI641
aI20841
aI87133
asS'd'
I10772
sa(dp21544
S'a'
I42974
sS's'
I17
sS'm'
(lp21545
I1
aI1
aI5
aI11
aI1
aI3
aI71
aI235
aI429
aI335
aI1649
aI1775
aI3077
aI13723
aI3209
aI19807
aI7283
asS'd'
I10773
sa(dp21546
S'a'
I42989
sS's'
I17
sS'm'
(lp21547
I1
aI1
aI7
aI1
aI31
aI49
aI39
aI141
aI127
aI63
aI1561
aI2559
aI7661
aI4825
aI9419
aI15327
aI87145
asS'd'
I10774
sa(dp21548
S'a'
I42995
sS's'
I17
sS'm'
(lp21549
I1
aI1
aI5
aI3
aI17
aI33
aI51
aI219
aI467
aI151
aI161
aI3301
aI7509
aI2235
aI30371
aI64031
aI62741
asS'd'
I10775
sa(dp21550
S'a'
I42997
sS's'
I17
sS'm'
(lp21551
I1
aI3
aI1
aI3
aI23
aI63
aI43
aI29
aI399
aI279
aI271
aI3537
aI1863
aI1811
aI14917
aI28247
aI34807
asS'd'
I10776
sa(dp21552
S'a'
I43007
sS's'
I17
sS'm'
(lp21553
I1
aI1
aI3
aI5
aI13
aI29
aI37
aI151
aI129
aI19
aI149
aI2145
aI5363
aI6835
aI19655
aI1207
aI74527
asS'd'
I10777
sa(dp21554
S'a'
I43018
sS's'
I17
sS'm'
(lp21555
I1
aI3
aI5
aI7
aI27
aI35
aI63
aI53
aI247
aI987
aI1767
aI483
aI3489
aI1711
aI10763
aI6981
aI78251
asS'd'
I10778
sa(dp21556
S'a'
I43025
sS's'
I17
sS'm'
(lp21557
I1
aI1
aI3
aI1
aI15
aI47
aI83
aI147
aI375
aI539
aI1623
aI29
aI4599
aI7981
aI23533
aI64659
aI48753
asS'd'
I10779
sa(dp21558
S'a'
I43031
sS's'
I17
sS'm'
(lp21559
I1
aI1
aI1
aI9
aI21
aI17
aI85
aI45
aI167
aI469
aI1319
aI2969
aI1605
aI1405
aI9961
aI28829
aI125757
asS'd'
I10780
sa(dp21560
S'a'
I43032
sS's'
I17
sS'm'
(lp21561
I1
aI3
aI1
aI11
aI3
aI45
aI43
aI159
aI301
aI579
aI1821
aI701
aI1149
aI457
aI16601
aI49377
aI99845
asS'd'
I10781
sa(dp21562
S'a'
I43038
sS's'
I17
sS'm'
(lp21563
I1
aI1
aI7
aI13
aI11
aI7
aI37
aI227
aI345
aI973
aI1167
aI1247
aI5109
aI10917
aI3029
aI60065
aI127347
asS'd'
I10782
sa(dp21564
S'a'
I43041
sS's'
I17
sS'm'
(lp21565
I1
aI1
aI3
aI5
aI3
aI63
aI95
aI233
aI495
aI225
aI1225
aI3451
aI7731
aI14677
aI10437
aI1417
aI33293
asS'd'
I10783
sa(dp21566
S'a'
I43054
sS's'
I17
sS'm'
(lp21567
I1
aI1
aI7
aI15
aI1
aI3
aI3
aI171
aI201
aI1009
aI1481
aI587
aI7661
aI10085
aI4961
aI46415
aI28573
asS'd'
I10784
sa(dp21568
S'a'
I43074
sS's'
I17
sS'm'
(lp21569
I1
aI1
aI5
aI1
aI3
aI45
aI67
aI79
aI463
aI733
aI2007
aI2811
aI2943
aI14857
aI23469
aI14479
aI97875
asS'd'
I10785
sa(dp21570
S'a'
I43085
sS's'
I17
sS'm'
(lp21571
I1
aI1
aI1
aI5
aI19
aI1
aI29
aI29
aI447
aI173
aI1081
aI153
aI5343
aI5707
aI1357
aI30169
aI122527
asS'd'
I10786
sa(dp21572
S'a'
I43097
sS's'
I17
sS'm'
(lp21573
I1
aI1
aI1
aI5
aI15
aI57
aI33
aI129
aI71
aI717
aI173
aI3271
aI4741
aI13211
aI28321
aI56793
aI119833
asS'd'
I10787
sa(dp21574
S'a'
I43098
sS's'
I17
sS'm'
(lp21575
I1
aI3
aI3
aI9
aI9
aI41
aI47
aI71
aI103
aI713
aI725
aI1335
aI5261
aI13835
aI17619
aI47429
aI69815
asS'd'
I10788
sa(dp21576
S'a'
I43110
sS's'
I17
sS'm'
(lp21577
I1
aI3
aI3
aI15
aI7
aI3
aI71
aI25
aI75
aI967
aI1037
aI3585
aI3407
aI9979
aI2195
aI51087
aI126535
asS'd'
I10789
sa(dp21578
S'a'
I43119
sS's'
I17
sS'm'
(lp21579
I1
aI3
aI3
aI11
aI25
aI7
aI25
aI249
aI473
aI339
aI1211
aI3503
aI4343
aI9707
aI26127
aI62061
aI52479
asS'd'
I10790
sa(dp21580
S'a'
I43131
sS's'
I17
sS'm'
(lp21581
I1
aI1
aI3
aI3
aI27
aI9
aI79
aI197
aI207
aI845
aI377
aI3231
aI5177
aI899
aI19497
aI41187
aI105897
asS'd'
I10791
sa(dp21582
S'a'
I43143
sS's'
I17
sS'm'
(lp21583
I1
aI3
aI5
aI15
aI5
aI27
aI65
aI151
aI207
aI677
aI713
aI2495
aI681
aI15341
aI5389
aI51965
aI43761
asS'd'
I10792
sa(dp21584
S'a'
I43144
sS's'
I17
sS'm'
(lp21585
I1
aI3
aI3
aI11
aI19
aI11
aI55
aI189
aI291
aI183
aI1345
aI2677
aI791
aI2391
aI25771
aI55147
aI24223
asS'd'
I10793
sa(dp21586
S'a'
I43152
sS's'
I17
sS'm'
(lp21587
I1
aI1
aI3
aI11
aI31
aI59
aI29
aI5
aI275
aI483
aI1361
aI1527
aI3019
aI245
aI17667
aI57905
aI41329
asS'd'
I10794
sa(dp21588
S'a'
I43157
sS's'
I17
sS'm'
(lp21589
I1
aI3
aI3
aI9
aI7
aI19
aI83
aI71
aI147
aI999
aI793
aI3535
aI1931
aI12817
aI2707
aI45735
aI31311
asS'd'
I10795
sa(dp21590
S'a'
I43178
sS's'
I17
sS'm'
(lp21591
I1
aI1
aI5
aI7
aI5
aI1
aI117
aI247
aI127
aI1011
aI1441
aI2449
aI4095
aI12239
aI4743
aI64781
aI32621
asS'd'
I10796
sa(dp21592
S'a'
I43180
sS's'
I17
sS'm'
(lp21593
I1
aI3
aI1
aI11
aI19
aI57
aI43
aI39
aI97
aI485
aI951
aI989
aI5975
aI5219
aI14421
aI43681
aI37305
asS'd'
I10797
sa(dp21594
S'a'
I43192
sS's'
I17
sS'm'
(lp21595
I1
aI1
aI5
aI15
aI7
aI49
aI113
aI161
aI199
aI545
aI1113
aI3821
aI2019
aI8747
aI4085
aI50823
aI31955
asS'd'
I10798
sa(dp21596
S'a'
I43197
sS's'
I17
sS'm'
(lp21597
I1
aI3
aI3
aI3
aI19
aI41
aI47
aI191
aI403
aI25
aI2043
aI3489
aI6263
aI4843
aI12961
aI63791
aI5027
asS'd'
I10799
sa(dp21598
S'a'
I43203
sS's'
I17
sS'm'
(lp21599
I1
aI1
aI7
aI1
aI25
aI55
aI5
aI51
aI121
aI273
aI973
aI3893
aI1771
aI9373
aI21927
aI29353
aI95935
asS'd'
I10800
sa(dp21600
S'a'
I43220
sS's'
I17
sS'm'
(lp21601
I1
aI3
aI3
aI3
aI27
aI1
aI97
aI63
aI445
aI179
aI481
aI2995
aI3123
aI4687
aI24359
aI35973
aI74535
asS'd'
I10801
sa(dp21602
S'a'
I43236
sS's'
I17
sS'm'
(lp21603
I1
aI1
aI5
aI1
aI29
aI23
aI117
aI183
aI197
aI819
aI695
aI641
aI4155
aI13593
aI30965
aI41407
aI42433
asS'd'
I10802
sa(dp21604
S'a'
I43245
sS's'
I17
sS'm'
(lp21605
I1
aI3
aI5
aI1
aI23
aI53
aI61
aI253
aI87
aI487
aI1995
aI1281
aI3367
aI15047
aI3493
aI41711
aI53407
asS'd'
I10803
sa(dp21606
S'a'
I43246
sS's'
I17
sS'm'
(lp21607
I1
aI1
aI1
aI9
aI27
aI49
aI83
aI21
aI63
aI181
aI1661
aI1649
aI281
aI12141
aI25771
aI35563
aI42643
asS'd'
I10804
sa(dp21608
S'a'
I43260
sS's'
I17
sS'm'
(lp21609
I1
aI3
aI5
aI13
aI15
aI59
aI121
aI113
aI379
aI487
aI1929
aI3725
aI2477
aI6527
aI8619
aI64869
aI57103
asS'd'
I10805
sa(dp21610
S'a'
I43265
sS's'
I17
sS'm'
(lp21611
I1
aI3
aI1
aI7
aI27
aI39
aI69
aI93
aI193
aI395
aI433
aI2091
aI151
aI6921
aI11599
aI36143
aI41179
asS'd'
I10806
sa(dp21612
S'a'
I43271
sS's'
I17
sS'm'
(lp21613
I1
aI1
aI7
aI1
aI31
aI33
aI73
aI199
aI57
aI37
aI1387
aI3505
aI7919
aI3507
aI2855
aI8239
aI84527
asS'd'
I10807
sa(dp21614
S'a'
I43285
sS's'
I17
sS'm'
(lp21615
I1
aI1
aI7
aI5
aI15
aI5
aI119
aI253
aI263
aI785
aI1409
aI1485
aI3675
aI5515
aI13057
aI30323
aI98015
asS'd'
I10808
sa(dp21616
S'a'
I43286
sS's'
I17
sS'm'
(lp21617
I1
aI3
aI1
aI1
aI11
aI5
aI57
aI83
aI365
aI703
aI1923
aI1397
aI1103
aI4015
aI13123
aI47093
aI113793
asS'd'
I10809
sa(dp21618
S'a'
I43290
sS's'
I17
sS'm'
(lp21619
I1
aI3
aI3
aI1
aI5
aI61
aI29
aI173
aI189
aI999
aI897
aI3389
aI6745
aI1487
aI2349
aI59105
aI107407
asS'd'
I10810
sa(dp21620
S'a'
I43299
sS's'
I17
sS'm'
(lp21621
I1
aI1
aI1
aI1
aI17
aI51
aI65
aI1
aI249
aI863
aI399
aI3819
aI2485
aI12215
aI12365
aI58909
aI25559
asS'd'
I10811
sa(dp21622
S'a'
I43314
sS's'
I17
sS'm'
(lp21623
I1
aI3
aI7
aI1
aI31
aI39
aI43
aI219
aI51
aI13
aI779
aI505
aI2259
aI14571
aI9049
aI21555
aI11869
asS'd'
I10812
sa(dp21624
S'a'
I43323
sS's'
I17
sS'm'
(lp21625
I1
aI1
aI7
aI7
aI13
aI5
aI97
aI85
aI111
aI511
aI587
aI63
aI2395
aI8099
aI26223
aI757
aI119821
asS'd'
I10813
sa(dp21626
S'a'
I43337
sS's'
I17
sS'm'
(lp21627
I1
aI3
aI3
aI5
aI5
aI19
aI113
aI35
aI101
aI41
aI499
aI1313
aI6489
aI6793
aI31435
aI45007
aI95691
asS'd'
I10814
sa(dp21628
S'a'
I43348
sS's'
I17
sS'm'
(lp21629
I1
aI3
aI5
aI15
aI19
aI37
aI103
aI187
aI347
aI667
aI1957
aI1825
aI7447
aI12359
aI21779
aI52749
aI18679
asS'd'
I10815
sa(dp21630
S'a'
I43355
sS's'
I17
sS'm'
(lp21631
I1
aI3
aI5
aI5
aI17
aI19
aI19
aI193
aI435
aI379
aI439
aI2093
aI725
aI2133
aI15659
aI54645
aI59567
asS'd'
I10816
sa(dp21632
S'a'
I43357
sS's'
I17
sS'm'
(lp21633
I1
aI3
aI7
aI3
aI23
aI35
aI33
aI13
aI23
aI349
aI231
aI1635
aI1625
aI5039
aI21299
aI36413
aI104681
asS'd'
I10817
sa(dp21634
S'a'
I43358
sS's'
I17
sS'm'
(lp21635
I1
aI1
aI3
aI13
aI23
aI49
aI15
aI253
aI509
aI9
aI411
aI2157
aI3737
aI11227
aI6021
aI42919
aI100375
asS'd'
I10818
sa(dp21636
S'a'
I43361
sS's'
I17
sS'm'
(lp21637
I1
aI1
aI7
aI1
aI17
aI11
aI33
aI167
aI219
aI63
aI137
aI741
aI4193
aI16149
aI9657
aI50223
aI85213
asS'd'
I10819
sa(dp21638
S'a'
I43362
sS's'
I17
sS'm'
(lp21639
I1
aI3
aI7
aI11
aI23
aI59
aI113
aI149
aI427
aI697
aI1723
aI255
aI201
aI10081
aI1079
aI323
aI109091
asS'd'
I10820
sa(dp21640
S'a'
I43364
sS's'
I17
sS'm'
(lp21641
I1
aI3
aI3
aI15
aI11
aI9
aI89
aI39
aI67
aI249
aI1939
aI1737
aI3719
aI10515
aI16517
aI22345
aI83959
asS'd'
I10821
sa(dp21642
S'a'
I43368
sS's'
I17
sS'm'
(lp21643
I1
aI3
aI3
aI13
aI5
aI33
aI127
aI9
aI329
aI429
aI563
aI1579
aI4427
aI8343
aI22083
aI5035
aI124915
asS'd'
I10822
sa(dp21644
S'a'
I43376
sS's'
I17
sS'm'
(lp21645
I1
aI1
aI1
aI5
aI15
aI57
aI121
aI171
aI315
aI983
aI743
aI2015
aI2421
aI12431
aI2561
aI13331
aI73163
asS'd'
I10823
sa(dp21646
S'a'
I43385
sS's'
I17
sS'm'
(lp21647
I1
aI1
aI3
aI9
aI1
aI39
aI85
aI159
aI23
aI979
aI1467
aI231
aI4231
aI3669
aI16747
aI24195
aI46745
asS'd'
I10824
sa(dp21648
S'a'
I43386
sS's'
I17
sS'm'
(lp21649
I1
aI1
aI3
aI7
aI3
aI11
aI65
aI67
aI85
aI455
aI365
aI2279
aI3471
aI12771
aI14443
aI42773
aI28723
asS'd'
I10825
sa(dp21650
S'a'
I43391
sS's'
I17
sS'm'
(lp21651
I1
aI3
aI5
aI1
aI13
aI9
aI105
aI237
aI103
aI59
aI1301
aI3125
aI509
aI12669
aI3893
aI9775
aI81303
asS'd'
I10826
sa(dp21652
S'a'
I43397
sS's'
I17
sS'm'
(lp21653
I1
aI1
aI3
aI11
aI19
aI9
aI125
aI23
aI191
aI979
aI533
aI429
aI3239
aI15013
aI13833
aI40689
aI102827
asS'd'
I10827
sa(dp21654
S'a'
I43431
sS's'
I17
sS'm'
(lp21655
I1
aI3
aI3
aI7
aI15
aI5
aI83
aI243
aI467
aI913
aI1279
aI3889
aI8049
aI8357
aI5957
aI39073
aI93521
asS'd'
I10828
sa(dp21656
S'a'
I43438
sS's'
I17
sS'm'
(lp21657
I1
aI3
aI3
aI3
aI19
aI5
aI123
aI77
aI289
aI57
aI2001
aI807
aI5257
aI1671
aI20273
aI10183
aI128439
asS'd'
I10829
sa(dp21658
S'a'
I43440
sS's'
I17
sS'm'
(lp21659
I1
aI1
aI7
aI13
aI19
aI45
aI25
aI47
aI135
aI929
aI1353
aI2731
aI3351
aI7637
aI27037
aI58835
aI50285
asS'd'
I10830
sa(dp21660
S'a'
I43452
sS's'
I17
sS'm'
(lp21661
I1
aI3
aI1
aI1
aI13
aI55
aI55
aI197
aI409
aI93
aI1351
aI161
aI1885
aI5913
aI27937
aI49793
aI84541
asS'd'
I10831
sa(dp21662
S'a'
I43463
sS's'
I17
sS'm'
(lp21663
I1
aI1
aI3
aI7
aI29
aI21
aI113
aI179
aI203
aI533
aI1471
aI2035
aI447
aI6781
aI28729
aI31099
aI23027
asS'd'
I10832
sa(dp21664
S'a'
I43470
sS's'
I17
sS'm'
(lp21665
I1
aI1
aI3
aI11
aI27
aI3
aI5
aI209
aI367
aI945
aI749
aI3637
aI2881
aI8139
aI27875
aI34223
aI97263
asS'd'
I10833
sa(dp21666
S'a'
I43478
sS's'
I17
sS'm'
(lp21667
I1
aI3
aI5
aI13
aI25
aI27
aI35
aI3
aI13
aI707
aI303
aI3663
aI6617
aI13501
aI25537
aI33077
aI71485
asS'd'
I10834
sa(dp21668
S'a'
I43481
sS's'
I17
sS'm'
(lp21669
I1
aI1
aI7
aI15
aI11
aI29
aI65
aI47
aI235
aI635
aI133
aI153
aI6175
aI2961
aI8171
aI28641
aI122589
asS'd'
I10835
sa(dp21670
S'a'
I43488
sS's'
I17
sS'm'
(lp21671
I1
aI1
aI5
aI15
aI17
aI41
aI85
aI147
aI323
aI673
aI1629
aI3477
aI3341
aI16373
aI13901
aI60961
aI39451
asS'd'
I10836
sa(dp21672
S'a'
I43491
sS's'
I17
sS'm'
(lp21673
I1
aI3
aI1
aI15
aI29
aI15
aI37
aI109
aI293
aI863
aI1835
aI1173
aI2263
aI13815
aI24995
aI6989
aI103417
asS'd'
I10837
sa(dp21674
S'a'
I43506
sS's'
I17
sS'm'
(lp21675
I1
aI3
aI3
aI15
aI3
aI31
aI23
aI47
aI15
aI717
aI1457
aI1067
aI6229
aI7051
aI21771
aI54815
aI115827
asS'd'
I10838
sa(dp21676
S'a'
I43512
sS's'
I17
sS'm'
(lp21677
I1
aI1
aI1
aI13
aI21
aI3
aI45
aI239
aI89
aI603
aI407
aI781
aI8095
aI7389
aI18035
aI32229
aI39867
asS'd'
I10839
sa(dp21678
S'a'
I43539
sS's'
I17
sS'm'
(lp21679
I1
aI1
aI3
aI7
aI7
aI59
aI79
aI51
aI411
aI917
aI803
aI2455
aI2623
aI12413
aI23957
aI44199
aI67903
asS'd'
I10840
sa(dp21680
S'a'
I43567
sS's'
I17
sS'm'
(lp21681
I1
aI3
aI1
aI9
aI17
aI37
aI117
aI47
aI101
aI733
aI1861
aI1111
aI6785
aI13743
aI24371
aI49427
aI54711
asS'd'
I10841
sa(dp21682
S'a'
I43579
sS's'
I17
sS'm'
(lp21683
I1
aI3
aI1
aI15
aI27
aI63
aI107
aI33
aI351
aI287
aI1765
aI1947
aI6209
aI8127
aI30007
aI18757
aI31453
asS'd'
I10842
sa(dp21684
S'a'
I43584
sS's'
I17
sS'm'
(lp21685
I1
aI3
aI5
aI13
aI11
aI13
aI29
aI247
aI7
aI609
aI1235
aI1767
aI5365
aI12673
aI10151
aI51579
aI106407
asS'd'
I10843
sa(dp21686
S'a'
I43601
sS's'
I17
sS'm'
(lp21687
I1
aI3
aI7
aI15
aI5
aI25
aI81
aI197
aI51
aI615
aI1695
aI259
aI7983
aI1403
aI7903
aI21441
aI73263
asS'd'
I10844
sa(dp21688
S'a'
I43614
sS's'
I17
sS'm'
(lp21689
I1
aI1
aI5
aI1
aI13
aI61
aI55
aI175
aI445
aI3
aI1957
aI1171
aI6823
aI4285
aI11847
aI12789
aI79787
asS'd'
I10845
sa(dp21690
S'a'
I43617
sS's'
I17
sS'm'
(lp21691
I1
aI1
aI5
aI15
aI17
aI51
aI111
aI201
aI45
aI97
aI45
aI2533
aI1125
aI3663
aI13685
aI45719
aI51497
asS'd'
I10846
sa(dp21692
S'a'
I43623
sS's'
I17
sS'm'
(lp21693
I1
aI3
aI3
aI13
aI29
aI59
aI111
aI97
aI381
aI477
aI1229
aI3709
aI5185
aI7055
aI32729
aI32881
aI25539
asS'd'
I10847
sa(dp21694
S'a'
I43630
sS's'
I17
sS'm'
(lp21695
I1
aI3
aI1
aI9
aI1
aI39
aI57
aI143
aI189
aI625
aI1717
aI1755
aI3129
aI807
aI27975
aI15511
aI66123
asS'd'
I10848
sa(dp21696
S'a'
I43647
sS's'
I17
sS'm'
(lp21697
I1
aI3
aI3
aI1
aI5
aI41
aI25
aI27
aI163
aI397
aI1595
aI2325
aI1803
aI12439
aI25743
aI24509
aI72613
asS'd'
I10849
sa(dp21698
S'a'
I43658
sS's'
I17
sS'm'
(lp21699
I1
aI1
aI5
aI13
aI29
aI41
aI125
aI113
aI367
aI709
aI1911
aI669
aI831
aI5375
aI31145
aI26197
aI33543
asS'd'
I10850
sa(dp21700
S'a'
I43663
sS's'
I17
sS'm'
(lp21701
I1
aI1
aI5
aI1
aI1
aI5
aI91
aI199
aI133
aI273
aI393
aI1179
aI717
aI12791
aI17693
aI6905
aI20433
asS'd'
I10851
sa(dp21702
S'a'
I43665
sS's'
I17
sS'm'
(lp21703
I1
aI1
aI3
aI15
aI29
aI35
aI9
aI127
aI383
aI673
aI1821
aI2765
aI2425
aI11789
aI19741
aI43189
aI99557
asS'd'
I10852
sa(dp21704
S'a'
I43691
sS's'
I17
sS'm'
(lp21705
I1
aI1
aI7
aI13
aI9
aI19
aI119
aI103
aI11
aI983
aI623
aI391
aI1609
aI2333
aI19843
aI28269
aI41237
asS'd'
I10853
sa(dp21706
S'a'
I43701
sS's'
I17
sS'm'
(lp21707
I1
aI3
aI7
aI5
aI29
aI3
aI13
aI213
aI387
aI361
aI749
aI669
aI1625
aI5687
aI11369
aI38119
aI38389
asS'd'
I10854
sa(dp21708
S'a'
I43705
sS's'
I17
sS'm'
(lp21709
I1
aI3
aI5
aI13
aI13
aI51
aI47
aI33
aI1
aI979
aI1817
aI2633
aI7181
aI47
aI3603
aI49211
aI4377
asS'd'
I10855
sa(dp21710
S'a'
I43708
sS's'
I17
sS'm'
(lp21711
I1
aI3
aI1
aI1
aI11
aI63
aI5
aI249
aI13
aI805
aI1097
aI1449
aI5235
aI16299
aI25855
aI30949
aI3013
asS'd'
I10856
sa(dp21712
S'a'
I43719
sS's'
I17
sS'm'
(lp21713
I1
aI3
aI7
aI9
aI29
aI35
aI89
aI135
aI475
aI945
aI999
aI771
aI6023
aI13317
aI32611
aI43971
aI10393
asS'd'
I10857
sa(dp21714
S'a'
I43731
sS's'
I17
sS'm'
(lp21715
I1
aI1
aI1
aI5
aI23
aI3
aI37
aI117
aI95
aI985
aI1599
aI2191
aI3617
aI5831
aI31113
aI10873
aI112219
asS'd'
I10858
sa(dp21716
S'a'
I43737
sS's'
I17
sS'm'
(lp21717
I1
aI3
aI5
aI7
aI11
aI15
aI55
aI65
aI239
aI365
aI1209
aI3509
aI8101
aI8619
aI24775
aI65291
aI50589
asS'd'
I10859
sa(dp21718
S'a'
I43740
sS's'
I17
sS'm'
(lp21719
I1
aI1
aI7
aI9
aI21
aI19
aI123
aI83
aI317
aI717
aI433
aI31
aI2597
aI14723
aI28839
aI7817
aI126123
asS'd'
I10860
sa(dp21720
S'a'
I43747
sS's'
I17
sS'm'
(lp21721
I1
aI1
aI7
aI11
aI3
aI33
aI99
aI39
aI227
aI279
aI353
aI1921
aI7883
aI16187
aI5157
aI41121
aI89425
asS'd'
I10861
sa(dp21722
S'a'
I43749
sS's'
I17
sS'm'
(lp21723
I1
aI3
aI5
aI9
aI25
aI7
aI29
aI165
aI129
aI77
aI159
aI923
aI1357
aI1159
aI23537
aI58087
aI56443
asS'd'
I10862
sa(dp21724
S'a'
I43750
sS's'
I17
sS'm'
(lp21725
I1
aI1
aI7
aI3
aI13
aI51
aI45
aI161
aI27
aI41
aI1295
aI2937
aI7223
aI5271
aI17927
aI23311
aI2543
asS'd'
I10863
sa(dp21726
S'a'
I43754
sS's'
I17
sS'm'
(lp21727
I1
aI1
aI1
aI1
aI11
aI53
aI119
aI165
aI409
aI785
aI1649
aI3587
aI259
aI10997
aI3171
aI31271
aI104631
asS'd'
I10864
sa(dp21728
S'a'
I43764
sS's'
I17
sS'm'
(lp21729
I1
aI1
aI5
aI7
aI5
aI7
aI49
aI201
aI373
aI825
aI1755
aI3751
aI8041
aI8133
aI21347
aI12039
aI3049
asS'd'
I10865
sa(dp21730
S'a'
I43767
sS's'
I17
sS'm'
(lp21731
I1
aI1
aI1
aI3
aI7
aI29
aI103
aI1
aI473
aI65
aI761
aI1611
aI5121
aI14345
aI32535
aI16679
aI11321
asS'd'
I10866
sa(dp21732
S'a'
I43768
sS's'
I17
sS'm'
(lp21733
I1
aI3
aI3
aI11
aI21
aI57
aI35
aI63
aI237
aI415
aI1943
aI483
aI5377
aI14647
aI23433
aI45459
aI32535
asS'd'
I10867
sa(dp21734
S'a'
I43773
sS's'
I17
sS'm'
(lp21735
I1
aI1
aI1
aI15
aI21
aI57
aI7
aI103
aI493
aI279
aI665
aI3699
aI169
aI7619
aI3571
aI11539
aI31983
asS'd'
I10868
sa(dp21736
S'a'
I43785
sS's'
I17
sS'm'
(lp21737
I1
aI1
aI1
aI1
aI9
aI5
aI81
aI159
aI105
aI927
aI379
aI1133
aI1805
aI14341
aI9833
aI63151
aI70877
asS'd'
I10869
sa(dp21738
S'a'
I43788
sS's'
I17
sS'm'
(lp21739
I1
aI1
aI7
aI5
aI19
aI5
aI63
aI127
aI129
aI43
aI757
aI2215
aI3899
aI643
aI19731
aI17345
aI102611
asS'd'
I10870
sa(dp21740
S'a'
I43810
sS's'
I17
sS'm'
(lp21741
I1
aI3
aI7
aI7
aI27
aI21
aI3
aI69
aI475
aI283
aI319
aI833
aI3683
aI11275
aI18191
aI44027
aI24901
asS'd'
I10871
sa(dp21742
S'a'
I43819
sS's'
I17
sS'm'
(lp21743
I1
aI1
aI5
aI5
aI31
aI25
aI63
aI33
aI505
aI765
aI257
aI1147
aI779
aI12505
aI19971
aI24695
aI65935
asS'd'
I10872
sa(dp21744
S'a'
I43834
sS's'
I17
sS'm'
(lp21745
I1
aI1
aI1
aI15
aI23
aI33
aI31
aI107
aI59
aI639
aI1307
aI3211
aI6171
aI15665
aI16775
aI61671
aI25569
asS'd'
I10873
sa(dp21746
S'a'
I43853
sS's'
I17
sS'm'
(lp21747
I1
aI3
aI3
aI9
aI31
aI3
aI113
aI199
aI425
aI895
aI1051
aI2125
aI1525
aI15199
aI14845
aI4213
aI18449
asS'd'
I10874
sa(dp21748
S'a'
I43866
sS's'
I17
sS'm'
(lp21749
I1
aI3
aI5
aI3
aI3
aI11
aI75
aI121
aI33
aI265
aI459
aI3879
aI909
aI6533
aI18451
aI32421
aI117427
asS'd'
I10875
sa(dp21750
S'a'
I43871
sS's'
I17
sS'm'
(lp21751
I1
aI1
aI1
aI9
aI11
aI9
aI125
aI175
aI309
aI847
aI959
aI2013
aI1557
aI9291
aI2963
aI43275
aI9917
asS'd'
I10876
sa(dp21752
S'a'
I43872
sS's'
I17
sS'm'
(lp21753
I1
aI1
aI5
aI3
aI15
aI39
aI67
aI35
aI373
aI601
aI463
aI1263
aI1615
aI15059
aI31011
aI36059
aI114493
asS'd'
I10877
sa(dp21754
S'a'
I43881
sS's'
I17
sS'm'
(lp21755
I1
aI1
aI5
aI15
aI5
aI43
aI49
aI239
aI461
aI171
aI1863
aI2249
aI2923
aI15897
aI22941
aI29925
aI21429
asS'd'
I10878
sa(dp21756
S'a'
I43889
sS's'
I17
sS'm'
(lp21757
I1
aI1
aI1
aI15
aI13
aI31
aI127
aI205
aI361
aI149
aI1641
aI1443
aI5959
aI13183
aI13861
aI9533
aI1011
asS'd'
I10879
sa(dp21758
S'a'
I43902
sS's'
I17
sS'm'
(lp21759
I1
aI1
aI3
aI13
aI9
aI49
aI39
aI67
aI165
aI695
aI611
aI2261
aI3425
aI6247
aI23575
aI51833
aI106167
asS'd'
I10880
sa(dp21760
S'a'
I43926
sS's'
I17
sS'm'
(lp21761
I1
aI1
aI7
aI9
aI29
aI21
aI75
aI251
aI87
aI263
aI2035
aI1007
aI3821
aI12719
aI8889
aI47901
aI39037
asS'd'
I10881
sa(dp21762
S'a'
I43935
sS's'
I17
sS'm'
(lp21763
I1
aI3
aI1
aI3
aI15
aI51
aI79
aI127
aI201
aI497
aI1881
aI3841
aI1821
aI14435
aI4933
aI6853
aI104305
asS'd'
I10882
sa(dp21764
S'a'
I43946
sS's'
I17
sS'm'
(lp21765
I1
aI1
aI5
aI11
aI23
aI47
aI33
aI109
aI481
aI585
aI333
aI2525
aI593
aI1625
aI5787
aI23839
aI30647
asS'd'
I10883
sa(dp21766
S'a'
I43951
sS's'
I17
sS'm'
(lp21767
I1
aI1
aI5
aI1
aI17
aI3
aI7
aI43
aI113
aI873
aI1433
aI3377
aI45
aI831
aI17015
aI21479
aI7257
asS'd'
I10884
sa(dp21768
S'a'
I43953
sS's'
I17
sS'm'
(lp21769
I1
aI1
aI1
aI1
aI13
aI21
aI59
aI159
aI279
aI871
aI53
aI3647
aI2599
aI12417
aI25807
aI6867
aI18251
asS'd'
I10885
sa(dp21770
S'a'
I43971
sS's'
I17
sS'm'
(lp21771
I1
aI1
aI5
aI9
aI29
aI61
aI7
aI81
aI353
aI761
aI269
aI4047
aI3051
aI8385
aI2919
aI18875
aI15239
asS'd'
I10886
sa(dp21772
S'a'
I44008
sS's'
I17
sS'm'
(lp21773
I1
aI1
aI7
aI13
aI31
aI17
aI71
aI103
aI107
aI655
aI1263
aI849
aI1809
aI349
aI3239
aI45381
aI117451
asS'd'
I10887
sa(dp21774
S'a'
I44011
sS's'
I17
sS'm'
(lp21775
I1
aI1
aI5
aI9
aI27
aI45
aI83
aI207
aI117
aI77
aI437
aI523
aI851
aI13595
aI12381
aI27271
aI59951
asS'd'
I10888
sa(dp21776
S'a'
I44026
sS's'
I17
sS'm'
(lp21777
I1
aI3
aI3
aI15
aI3
aI33
aI103
aI217
aI61
aI443
aI1077
aI2887
aI1751
aI11111
aI465
aI37051
aI89687
asS'd'
I10889
sa(dp21778
S'a'
I44033
sS's'
I17
sS'm'
(lp21779
I1
aI1
aI1
aI5
aI15
aI15
aI13
aI115
aI275
aI565
aI1257
aI1067
aI6561
aI8143
aI2149
aI53169
aI123637
asS'd'
I10890
sa(dp21780
S'a'
I44048
sS's'
I17
sS'm'
(lp21781
I1
aI3
aI3
aI15
aI27
aI63
aI25
aI191
aI143
aI103
aI1247
aI1053
aI2469
aI9823
aI4437
aI18195
aI91751
asS'd'
I10891
sa(dp21782
S'a'
I44057
sS's'
I17
sS'm'
(lp21783
I1
aI1
aI7
aI11
aI1
aI63
aI31
aI103
aI249
aI861
aI983
aI335
aI35
aI4291
aI16307
aI43669
aI68065
asS'd'
I10892
sa(dp21784
S'a'
I44058
sS's'
I17
sS'm'
(lp21785
I1
aI3
aI1
aI15
aI13
aI29
aI51
aI145
aI177
aI851
aI39
aI3531
aI4477
aI4243
aI3301
aI64293
aI15741
asS'd'
I10893
sa(dp21786
S'a'
I44067
sS's'
I17
sS'm'
(lp21787
I1
aI1
aI7
aI3
aI29
aI45
aI5
aI85
aI185
aI191
aI1007
aI3085
aI2177
aI14911
aI18319
aI265
aI25435
asS'd'
I10894
sa(dp21788
S'a'
I44081
sS's'
I17
sS'm'
(lp21789
I1
aI1
aI5
aI9
aI9
aI57
aI47
aI143
aI217
aI947
aI2021
aI1835
aI4773
aI15145
aI26519
aI46407
aI103667
asS'd'
I10895
sa(dp21790
S'a'
I44087
sS's'
I17
sS'm'
(lp21791
I1
aI3
aI1
aI11
aI1
aI7
aI51
aI75
aI207
aI757
aI89
aI1289
aI39
aI15641
aI9477
aI28503
aI47113
asS'd'
I10896
sa(dp21792
S'a'
I44099
sS's'
I17
sS'm'
(lp21793
I1
aI3
aI1
aI11
aI9
aI19
aI21
aI197
aI429
aI121
aI813
aI3447
aI6091
aI3167
aI5401
aI27791
aI26499
asS'd'
I10897
sa(dp21794
S'a'
I44105
sS's'
I17
sS'm'
(lp21795
I1
aI1
aI7
aI15
aI1
aI15
aI85
aI247
aI3
aI111
aI433
aI3103
aI5049
aI7929
aI22645
aI53247
aI53417
asS'd'
I10898
sa(dp21796
S'a'
I44106
sS's'
I17
sS'm'
(lp21797
I1
aI1
aI7
aI7
aI27
aI19
aI125
aI101
aI269
aI7
aI777
aI1289
aI1429
aI11561
aI18043
aI3601
aI125857
asS'd'
I10899
sa(dp21798
S'a'
I44114
sS's'
I17
sS'm'
(lp21799
I1
aI1
aI1
aI13
aI11
aI9
aI127
aI231
aI239
aI435
aI1291
aI4025
aI1049
aI15549
aI7577
aI51147
aI38121
asS'd'
I10900
sa(dp21800
S'a'
I44116
sS's'
I17
sS'm'
(lp21801
I1
aI1
aI7
aI3
aI9
aI55
aI57
aI137
aI387
aI565
aI873
aI1417
aI5993
aI4849
aI1731
aI51653
aI105697
asS'd'
I10901
sa(dp21802
S'a'
I44130
sS's'
I17
sS'm'
(lp21803
I1
aI1
aI7
aI9
aI7
aI47
aI115
aI119
aI325
aI881
aI1687
aI1009
aI7007
aI12541
aI6737
aI28471
aI7369
asS'd'
I10902
sa(dp21804
S'a'
I44139
sS's'
I17
sS'm'
(lp21805
I1
aI3
aI1
aI1
aI11
aI47
aI25
aI163
aI399
aI977
aI1777
aI727
aI5575
aI1311
aI23843
aI2199
aI93229
asS'd'
I10903
sa(dp21806
S'a'
I44141
sS's'
I17
sS'm'
(lp21807
I1
aI1
aI7
aI5
aI13
aI19
aI53
aI123
aI439
aI585
aI1977
aI3387
aI5305
aI1463
aI14307
aI9519
aI537
asS'd'
I10904
sa(dp21808
S'a'
I44153
sS's'
I17
sS'm'
(lp21809
I1
aI1
aI7
aI15
aI1
aI53
aI13
aI213
aI323
aI699
aI1585
aI3499
aI2441
aI3055
aI31263
aI63923
aI9779
asS'd'
I10905
sa(dp21810
S'a'
I44159
sS's'
I17
sS'm'
(lp21811
I1
aI1
aI5
aI5
aI21
aI43
aI123
aI43
aI475
aI521
aI1301
aI3185
aI5627
aI7443
aI1195
aI39485
aI113125
asS'd'
I10906
sa(dp21812
S'a'
I44160
sS's'
I17
sS'm'
(lp21813
I1
aI1
aI5
aI7
aI9
aI3
aI39
aI5
aI237
aI719
aI1743
aI1153
aI6401
aI14701
aI5503
aI38491
aI24123
asS'd'
I10907
sa(dp21814
S'a'
I44170
sS's'
I17
sS'm'
(lp21815
I1
aI3
aI5
aI9
aI17
aI33
aI117
aI23
aI409
aI63
aI1829
aI2587
aI3489
aI3209
aI4775
aI40069
aI4721
asS'd'
I10908
sa(dp21816
S'a'
I44172
sS's'
I17
sS'm'
(lp21817
I1
aI3
aI3
aI5
aI21
aI63
aI95
aI231
aI25
aI167
aI1181
aI813
aI4591
aI5227
aI21999
aI19633
aI37547
asS'd'
I10909
sa(dp21818
S'a'
I44187
sS's'
I17
sS'm'
(lp21819
I1
aI1
aI7
aI11
aI13
aI9
aI13
aI147
aI239
aI951
aI1247
aI1199
aI7907
aI12493
aI25371
aI1917
aI107499
asS'd'
I10910
sa(dp21820
S'a'
I44190
sS's'
I17
sS'm'
(lp21821
I1
aI1
aI5
aI15
aI3
aI49
aI31
aI103
aI189
aI561
aI1763
aI3941
aI3525
aI3165
aI7789
aI57729
aI92635
asS'd'
I10911
sa(dp21822
S'a'
I44193
sS's'
I17
sS'm'
(lp21823
I1
aI1
aI1
aI5
aI3
aI61
aI107
aI163
aI465
aI631
aI1519
aI169
aI4469
aI8153
aI11039
aI247
aI37657
asS'd'
I10912
sa(dp21824
S'a'
I44199
sS's'
I17
sS'm'
(lp21825
I1
aI3
aI1
aI5
aI9
aI37
aI51
aI195
aI465
aI975
aI169
aI1077
aI995
aI2669
aI7663
aI28997
aI25779
asS'd'
I10913
sa(dp21826
S'a'
I44213
sS's'
I17
sS'm'
(lp21827
I1
aI1
aI7
aI13
aI7
aI37
aI3
aI117
aI147
aI335
aI629
aI4077
aI5855
aI2893
aI5629
aI55075
aI83359
asS'd'
I10914
sa(dp21828
S'a'
I44218
sS's'
I17
sS'm'
(lp21829
I1
aI1
aI5
aI9
aI9
aI25
aI53
aI63
aI315
aI287
aI1833
aI1397
aI2395
aI5719
aI6719
aI18003
aI101073
asS'd'
I10915
sa(dp21830
S'a'
I44223
sS's'
I17
sS'm'
(lp21831
I1
aI1
aI7
aI1
aI13
aI19
aI13
aI81
aI497
aI399
aI413
aI2411
aI3915
aI14037
aI19735
aI4587
aI69655
asS'd'
I10916
sa(dp21832
S'a'
I44235
sS's'
I17
sS'm'
(lp21833
I1
aI3
aI1
aI7
aI5
aI61
aI101
aI209
aI299
aI729
aI1359
aI4013
aI2057
aI8439
aI8113
aI57417
aI8951
asS'd'
I10917
sa(dp21834
S'a'
I44243
sS's'
I17
sS'm'
(lp21835
I1
aI3
aI5
aI7
aI29
aI21
aI67
aI73
aI107
aI359
aI1655
aI3729
aI4403
aI10467
aI28103
aI10261
aI74651
asS'd'
I10918
sa(dp21836
S'a'
I44262
sS's'
I17
sS'm'
(lp21837
I1
aI1
aI1
aI9
aI3
aI39
aI25
aI91
aI287
aI497
aI1743
aI339
aI4739
aI1709
aI16351
aI45385
aI64693
asS'd'
I10919
sa(dp21838
S'a'
I44283
sS's'
I17
sS'm'
(lp21839
I1
aI3
aI1
aI1
aI7
aI13
aI41
aI93
aI49
aI285
aI997
aI891
aI4353
aI4249
aI11269
aI36935
aI71249
asS'd'
I10920
sa(dp21840
S'a'
I44291
sS's'
I17
sS'm'
(lp21841
I1
aI3
aI3
aI13
aI13
aI23
aI97
aI231
aI101
aI93
aI1183
aI201
aI6795
aI16287
aI30707
aI20845
aI105873
asS'd'
I10921
sa(dp21842
S'a'
I44293
sS's'
I17
sS'm'
(lp21843
I1
aI1
aI1
aI9
aI7
aI57
aI123
aI167
aI451
aI245
aI1887
aI1839
aI2967
aI2387
aI15075
aI11877
aI629
asS'd'
I10922
sa(dp21844
S'a'
I44308
sS's'
I17
sS'm'
(lp21845
I1
aI3
aI3
aI1
aI13
aI13
aI83
aI41
aI219
aI313
aI1743
aI1265
aI4435
aI11731
aI17625
aI64235
aI24865
asS'd'
I10923
sa(dp21846
S'a'
I44327
sS's'
I17
sS'm'
(lp21847
I1
aI3
aI1
aI9
aI13
aI17
aI109
aI235
aI387
aI581
aI887
aI1071
aI603
aI10955
aI5001
aI8419
aI20997
asS'd'
I10924
sa(dp21848
S'a'
I44341
sS's'
I17
sS'm'
(lp21849
I1
aI3
aI1
aI5
aI31
aI55
aI1
aI219
aI27
aI623
aI1425
aI1309
aI5409
aI9633
aI3231
aI15029
aI22989
asS'd'
I10925
sa(dp21850
S'a'
I44346
sS's'
I17
sS'm'
(lp21851
I1
aI3
aI3
aI13
aI25
aI47
aI23
aI223
aI283
aI189
aI1665
aI3743
aI387
aI1807
aI16919
aI8511
aI15933
asS'd'
I10926
sa(dp21852
S'a'
I44348
sS's'
I17
sS'm'
(lp21853
I1
aI1
aI1
aI1
aI13
aI11
aI81
aI59
aI423
aI1007
aI317
aI2761
aI2617
aI9715
aI24853
aI63585
aI77083
asS'd'
I10927
sa(dp21854
S'a'
I44354
sS's'
I17
sS'm'
(lp21855
I1
aI3
aI1
aI3
aI3
aI11
aI103
aI123
aI401
aI467
aI1159
aI2725
aI3275
aI15513
aI2281
aI21617
aI87211
asS'd'
I10928
sa(dp21856
S'a'
I44366
sS's'
I17
sS'm'
(lp21857
I1
aI1
aI5
aI7
aI23
aI17
aI25
aI83
aI11
aI901
aI809
aI3233
aI3929
aI8685
aI7609
aI50949
aI104841
asS'd'
I10929
sa(dp21858
S'a'
I44368
sS's'
I17
sS'm'
(lp21859
I1
aI3
aI7
aI1
aI15
aI33
aI37
aI245
aI275
aI453
aI729
aI721
aI1589
aI5417
aI29839
aI57315
aI67227
asS'd'
I10930
sa(dp21860
S'a'
I44373
sS's'
I17
sS'm'
(lp21861
I1
aI3
aI7
aI3
aI21
aI17
aI51
aI213
aI225
aI471
aI1201
aI931
aI1229
aI9503
aI5507
aI4057
aI7737
asS'd'
I10931
sa(dp21862
S'a'
I44384
sS's'
I17
sS'm'
(lp21863
I1
aI3
aI1
aI11
aI29
aI55
aI19
aI193
aI9
aI151
aI597
aI1377
aI827
aI8549
aI1293
aI10963
aI86183
asS'd'
I10932
sa(dp21864
S'a'
I44390
sS's'
I17
sS'm'
(lp21865
I1
aI3
aI3
aI15
aI17
aI23
aI89
aI47
aI195
aI333
aI2001
aI1001
aI6715
aI9797
aI21631
aI5723
aI88847
asS'd'
I10933
sa(dp21866
S'a'
I44393
sS's'
I17
sS'm'
(lp21867
I1
aI3
aI5
aI9
aI21
aI33
aI111
aI101
aI503
aI513
aI785
aI1947
aI1139
aI7921
aI13189
aI34831
aI80963
asS'd'
I10934
sa(dp21868
S'a'
I44394
sS's'
I17
sS'm'
(lp21869
I1
aI3
aI3
aI13
aI9
aI61
aI35
aI39
aI451
aI485
aI661
aI1993
aI4705
aI9477
aI32541
aI16553
aI33167
asS'd'
I10935
sa(dp21870
S'a'
I44399
sS's'
I17
sS'm'
(lp21871
I1
aI3
aI3
aI9
aI29
aI37
aI115
aI87
aI367
aI325
aI539
aI1975
aI6769
aI1453
aI31099
aI3335
aI16939
asS'd'
I10936
sa(dp21872
S'a'
I44401
sS's'
I17
sS'm'
(lp21873
I1
aI1
aI1
aI7
aI15
aI21
aI113
aI203
aI97
aI847
aI625
aI847
aI1819
aI1109
aI14503
aI25319
aI100259
asS'd'
I10937
sa(dp21874
S'a'
I44408
sS's'
I17
sS'm'
(lp21875
I1
aI1
aI5
aI11
aI9
aI13
aI65
aI21
aI429
aI865
aI513
aI2183
aI3785
aI11817
aI6283
aI23041
aI7969
asS'd'
I10938
sa(dp21876
S'a'
I44411
sS's'
I17
sS'm'
(lp21877
I1
aI1
aI5
aI13
aI1
aI41
aI109
aI43
aI91
aI211
aI1477
aI3543
aI5217
aI3133
aI12503
aI15523
aI12917
asS'd'
I10939
sa(dp21878
S'a'
I44417
sS's'
I17
sS'm'
(lp21879
I1
aI3
aI7
aI9
aI23
aI53
aI109
aI89
aI229
aI939
aI1211
aI2771
aI541
aI15915
aI5411
aI47273
aI54453
asS'd'
I10940
sa(dp21880
S'a'
I44420
sS's'
I17
sS'm'
(lp21881
I1
aI1
aI1
aI3
aI3
aI45
aI31
aI63
aI99
aI347
aI17
aI523
aI441
aI12325
aI15673
aI1887
aI15289
asS'd'
I10941
sa(dp21882
S'a'
I44424
sS's'
I17
sS'm'
(lp21883
I1
aI1
aI1
aI7
aI29
aI61
aI35
aI115
aI345
aI1011
aI5
aI595
aI465
aI3897
aI28147
aI791
aI98757
asS'd'
I10942
sa(dp21884
S'a'
I44444
sS's'
I17
sS'm'
(lp21885
I1
aI1
aI5
aI9
aI27
aI1
aI21
aI155
aI467
aI469
aI1565
aI1439
aI5809
aI851
aI32503
aI3025
aI97231
asS'd'
I10943
sa(dp21886
S'a'
I44451
sS's'
I17
sS'm'
(lp21887
I1
aI1
aI1
aI9
aI3
aI17
aI15
aI73
aI487
aI1011
aI63
aI2605
aI6647
aI9385
aI4527
aI21993
aI19783
asS'd'
I10944
sa(dp21888
S'a'
I44453
sS's'
I17
sS'm'
(lp21889
I1
aI1
aI3
aI9
aI17
aI17
aI65
aI75
aI175
aI897
aI1317
aI2593
aI1495
aI15835
aI12025
aI57457
aI29577
asS'd'
I10945
sa(dp21890
S'a'
I44466
sS's'
I17
sS'm'
(lp21891
I1
aI1
aI1
aI13
aI7
aI1
aI13
aI145
aI491
aI427
aI375
aI1235
aI3045
aI2991
aI26607
aI30581
aI43377
asS'd'
I10946
sa(dp21892
S'a'
I44472
sS's'
I17
sS'm'
(lp21893
I1
aI1
aI1
aI1
aI31
aI1
aI75
aI235
aI345
aI75
aI1505
aI1401
aI6921
aI6207
aI13729
aI21545
aI34703
asS'd'
I10947
sa(dp21894
S'a'
I44475
sS's'
I17
sS'm'
(lp21895
I1
aI3
aI7
aI9
aI31
aI35
aI53
aI233
aI85
aI385
aI2045
aI1401
aI5365
aI827
aI13093
aI41097
aI97381
asS'd'
I10948
sa(dp21896
S'a'
I44486
sS's'
I17
sS'm'
(lp21897
I1
aI3
aI7
aI15
aI5
aI9
aI19
aI125
aI49
aI29
aI1553
aI675
aI3947
aI4775
aI8161
aI12321
aI55191
asS'd'
I10949
sa(dp21898
S'a'
I44500
sS's'
I17
sS'm'
(lp21899
I1
aI3
aI3
aI7
aI17
aI17
aI27
aI237
aI87
aI927
aI275
aI1965
aI4993
aI1429
aI31613
aI38403
aI119319
asS'd'
I10950
sa(dp21900
S'a'
I44510
sS's'
I17
sS'm'
(lp21901
I1
aI3
aI7
aI13
aI25
aI61
aI87
aI133
aI37
aI725
aI697
aI371
aI7607
aI13861
aI8015
aI63997
aI25745
asS'd'
I10951
sa(dp21902
S'a'
I44531
sS's'
I17
sS'm'
(lp21903
I1
aI1
aI5
aI3
aI1
aI29
aI115
aI53
aI355
aI533
aI1711
aI3863
aI6983
aI4849
aI15787
aI38933
aI100299
asS'd'
I10952
sa(dp21904
S'a'
I44534
sS's'
I17
sS'm'
(lp21905
I1
aI1
aI3
aI5
aI7
aI11
aI95
aI21
aI363
aI1005
aI425
aI3497
aI841
aI8251
aI11933
aI47783
aI122699
asS'd'
I10953
sa(dp21906
S'a'
I44553
sS's'
I17
sS'm'
(lp21907
I1
aI1
aI1
aI11
aI15
aI41
aI23
aI159
aI191
aI433
aI919
aI3151
aI5311
aI2061
aI11277
aI4947
aI10549
asS'd'
I10954
sa(dp21908
S'a'
I44559
sS's'
I17
sS'm'
(lp21909
I1
aI1
aI5
aI1
aI29
aI57
aI23
aI239
aI179
aI821
aI1825
aI1745
aI4357
aI4041
aI27517
aI8557
aI86969
asS'd'
I10955
sa(dp21910
S'a'
I44564
sS's'
I17
sS'm'
(lp21911
I1
aI3
aI1
aI13
aI3
aI45
aI91
aI21
aI221
aI203
aI683
aI1787
aI375
aI4101
aI13555
aI43269
aI8063
asS'd'
I10956
sa(dp21912
S'a'
I44580
sS's'
I17
sS'm'
(lp21913
I1
aI1
aI5
aI15
aI17
aI61
aI95
aI95
aI285
aI597
aI1967
aI4061
aI389
aI3813
aI6061
aI50261
aI56035
asS'd'
I10957
sa(dp21914
S'a'
I44583
sS's'
I17
sS'm'
(lp21915
I1
aI1
aI7
aI9
aI9
aI35
aI103
aI255
aI239
aI77
aI145
aI4089
aI757
aI16151
aI29963
aI1229
aI31895
asS'd'
I10958
sa(dp21916
S'a'
I44589
sS's'
I17
sS'm'
(lp21917
I1
aI1
aI7
aI7
aI29
aI51
aI63
aI105
aI55
aI609
aI665
aI2101
aI4605
aI7085
aI18543
aI64221
aI102503
asS'd'
I10959
sa(dp21918
S'a'
I44592
sS's'
I17
sS'm'
(lp21919
I1
aI1
aI3
aI9
aI23
aI49
aI83
aI71
aI191
aI917
aI39
aI1013
aI4689
aI2407
aI1733
aI31113
aI31263
asS'd'
I10960
sa(dp21920
S'a'
I44609
sS's'
I17
sS'm'
(lp21921
I1
aI1
aI5
aI11
aI31
aI51
aI17
aI223
aI325
aI829
aI541
aI3561
aI5319
aI15397
aI12479
aI57199
aI38611
asS'd'
I10961
sa(dp21922
S'a'
I44627
sS's'
I17
sS'm'
(lp21923
I1
aI3
aI1
aI3
aI19
aI57
aI19
aI191
aI427
aI905
aI1111
aI695
aI5447
aI4061
aI25543
aI45699
aI113283
asS'd'
I10962
sa(dp21924
S'a'
I44633
sS's'
I17
sS'm'
(lp21925
I1
aI1
aI3
aI7
aI5
aI11
aI59
aI249
aI375
aI889
aI563
aI2757
aI5857
aI3595
aI23183
aI1785
aI105017
asS'd'
I10963
sa(dp21926
S'a'
I44643
sS's'
I17
sS'm'
(lp21927
I1
aI3
aI5
aI7
aI11
aI55
aI95
aI167
aI27
aI823
aI903
aI2403
aI1137
aI3209
aI6313
aI61871
aI129865
asS'd'
I10964
sa(dp21928
S'a'
I44646
sS's'
I17
sS'm'
(lp21929
I1
aI1
aI3
aI11
aI25
aI3
aI89
aI171
aI209
aI409
aI1357
aI3825
aI5261
aI10805
aI13493
aI3303
aI129987
asS'd'
I10965
sa(dp21930
S'a'
I44650
sS's'
I17
sS'm'
(lp21931
I1
aI1
aI5
aI1
aI23
aI21
aI3
aI207
aI471
aI375
aI1785
aI2555
aI1613
aI16235
aI1585
aI48221
aI10197
asS'd'
I10966
sa(dp21932
S'a'
I44674
sS's'
I17
sS'm'
(lp21933
I1
aI1
aI1
aI15
aI13
aI33
aI89
aI185
aI331
aI239
aI1401
aI789
aI2687
aI15193
aI20911
aI18935
aI28751
asS'd'
I10967
sa(dp21934
S'a'
I44676
sS's'
I17
sS'm'
(lp21935
I1
aI1
aI1
aI13
aI27
aI19
aI111
aI139
aI385
aI531
aI1069
aI2343
aI7405
aI10305
aI7049
aI48215
aI77591
asS'd'
I10968
sa(dp21936
S'a'
I44680
sS's'
I17
sS'm'
(lp21937
I1
aI3
aI7
aI13
aI23
aI9
aI113
aI107
aI441
aI265
aI1617
aI63
aI7629
aI5505
aI7059
aI47307
aI82527
asS'd'
I10969
sa(dp21938
S'a'
I44683
sS's'
I17
sS'm'
(lp21939
I1
aI3
aI1
aI9
aI27
aI27
aI35
aI233
aI189
aI517
aI1285
aI1843
aI1569
aI14921
aI6617
aI44337
aI46917
asS'd'
I10970
sa(dp21940
S'a'
I44703
sS's'
I17
sS'm'
(lp21941
I1
aI1
aI3
aI15
aI7
aI15
aI9
aI255
aI109
aI629
aI437
aI3601
aI6591
aI10873
aI1765
aI46459
aI110991
asS'd'
I10971
sa(dp21942
S'a'
I44704
sS's'
I17
sS'm'
(lp21943
I1
aI1
aI5
aI15
aI17
aI13
aI115
aI97
aI401
aI979
aI1139
aI2607
aI6537
aI5369
aI17775
aI7657
aI57175
asS'd'
I10972
sa(dp21944
S'a'
I44716
sS's'
I17
sS'm'
(lp21945
I1
aI1
aI5
aI15
aI27
aI15
aI43
aI95
aI271
aI945
aI1205
aI3505
aI7403
aI13203
aI27259
aI24821
aI62921
asS'd'
I10973
sa(dp21946
S'a'
I44733
sS's'
I17
sS'm'
(lp21947
I1
aI1
aI7
aI15
aI9
aI13
aI53
aI177
aI93
aI169
aI1933
aI1101
aI4847
aI15477
aI22107
aI13009
aI93675
asS'd'
I10974
sa(dp21948
S'a'
I44748
sS's'
I17
sS'm'
(lp21949
I1
aI3
aI1
aI3
aI13
aI57
aI121
aI229
aI353
aI449
aI769
aI1207
aI557
aI5673
aI13129
aI29383
aI35925
asS'd'
I10975
sa(dp21950
S'a'
I44759
sS's'
I17
sS'm'
(lp21951
I1
aI3
aI3
aI1
aI31
aI33
aI5
aI87
aI461
aI873
aI795
aI2715
aI1421
aI14723
aI17917
aI20681
aI46103
asS'd'
I10976
sa(dp21952
S'a'
I44763
sS's'
I17
sS'm'
(lp21953
I1
aI1
aI7
aI3
aI29
aI5
aI49
aI215
aI341
aI25
aI1473
aI177
aI1443
aI14181
aI26723
aI49143
aI73461
asS'd'
I10977
sa(dp21954
S'a'
I44781
sS's'
I17
sS'm'
(lp21955
I1
aI3
aI1
aI5
aI17
aI53
aI5
aI27
aI1
aI325
aI1335
aI2941
aI7195
aI8179
aI26971
aI63469
aI49357
asS'd'
I10978
sa(dp21956
S'a'
I44782
sS's'
I17
sS'm'
(lp21957
I1
aI3
aI5
aI3
aI3
aI5
aI29
aI241
aI119
aI415
aI1371
aI3201
aI2815
aI15567
aI32521
aI18635
aI2101
asS'd'
I10979
sa(dp21958
S'a'
I44789
sS's'
I17
sS'm'
(lp21959
I1
aI3
aI1
aI3
aI7
aI13
aI127
aI157
aI271
aI403
aI187
aI3663
aI4073
aI12613
aI1305
aI31061
aI48361
asS'd'
I10980
sa(dp21960
S'a'
I44794
sS's'
I17
sS'm'
(lp21961
I1
aI1
aI3
aI5
aI1
aI39
aI41
aI201
aI113
aI923
aI621
aI497
aI3823
aI12543
aI27273
aI58509
aI21613
asS'd'
I10981
sa(dp21962
S'a'
I44799
sS's'
I17
sS'm'
(lp21963
I1
aI1
aI1
aI11
aI5
aI51
aI93
aI39
aI345
aI175
aI679
aI617
aI3445
aI8591
aI4017
aI5147
aI88847
asS'd'
I10982
sa(dp21964
S'a'
I44804
sS's'
I17
sS'm'
(lp21965
I1
aI1
aI7
aI7
aI7
aI9
aI63
aI7
aI89
aI711
aI487
aI69
aI447
aI3355
aI31929
aI34719
aI93629
asS'd'
I10983
sa(dp21966
S'a'
I44813
sS's'
I17
sS'm'
(lp21967
I1
aI3
aI1
aI3
aI27
aI11
aI51
aI11
aI471
aI889
aI1935
aI2185
aI1277
aI3127
aI8853
aI17839
aI40279
asS'd'
I10984
sa(dp21968
S'a'
I44822
sS's'
I17
sS'm'
(lp21969
I1
aI3
aI3
aI15
aI25
aI35
aI71
aI213
aI121
aI935
aI1601
aI537
aI5753
aI8743
aI15243
aI59545
aI60399
asS'd'
I10985
sa(dp21970
S'a'
I44838
sS's'
I17
sS'm'
(lp21971
I1
aI1
aI3
aI15
aI31
aI41
aI51
aI205
aI123
aI215
aI305
aI3777
aI4103
aI7275
aI21603
aI56853
aI54575
asS'd'
I10986
sa(dp21972
S'a'
I44842
sS's'
I17
sS'm'
(lp21973
I1
aI3
aI7
aI9
aI17
aI19
aI37
aI59
aI193
aI303
aI1079
aI3627
aI6503
aI14649
aI10283
aI64469
aI83677
asS'd'
I10987
sa(dp21974
S'a'
I44849
sS's'
I17
sS'm'
(lp21975
I1
aI3
aI1
aI5
aI11
aI3
aI115
aI139
aI213
aI307
aI721
aI1611
aI5093
aI11817
aI32503
aI38559
aI38449
asS'd'
I10988
sa(dp21976
S'a'
I44856
sS's'
I17
sS'm'
(lp21977
I1
aI3
aI1
aI1
aI17
aI31
aI41
aI113
aI135
aI733
aI723
aI2021
aI7397
aI15917
aI15741
aI7295
aI69885
asS'd'
I10989
sa(dp21978
S'a'
I44870
sS's'
I17
sS'm'
(lp21979
I1
aI1
aI7
aI11
aI31
aI3
aI125
aI77
aI89
aI793
aI1441
aI1527
aI457
aI9457
aI13581
aI62979
aI125279
asS'd'
I10990
sa(dp21980
S'a'
I44887
sS's'
I17
sS'm'
(lp21981
I1
aI1
aI1
aI5
aI9
aI17
aI19
aI115
aI43
aI395
aI183
aI2091
aI7021
aI7555
aI20165
aI45165
aI58925
asS'd'
I10991
sa(dp21982
S'a'
I44904
sS's'
I17
sS'm'
(lp21983
I1
aI1
aI1
aI15
aI23
aI37
aI97
aI45
aI357
aI201
aI425
aI3605
aI5305
aI10079
aI16397
aI40635
aI15355
asS'd'
I10992
sa(dp21984
S'a'
I44915
sS's'
I17
sS'm'
(lp21985
I1
aI1
aI3
aI7
aI3
aI43
aI65
aI89
aI51
aI801
aI917
aI2835
aI5675
aI2347
aI16587
aI19701
aI68655
asS'd'
I10993
sa(dp21986
S'a'
I44917
sS's'
I17
sS'm'
(lp21987
I1
aI3
aI7
aI13
aI11
aI59
aI93
aI155
aI53
aI435
aI165
aI3231
aI429
aI12757
aI27033
aI14081
aI12625
asS'd'
I10994
sa(dp21988
S'a'
I44921
sS's'
I17
sS'm'
(lp21989
I1
aI3
aI1
aI15
aI15
aI33
aI121
aI157
aI271
aI295
aI901
aI1689
aI709
aI13395
aI17773
aI14397
aI37743
asS'd'
I10995
sa(dp21990
S'a'
I44928
sS's'
I17
sS'm'
(lp21991
I1
aI1
aI1
aI3
aI7
aI17
aI125
aI113
aI223
aI603
aI425
aI3213
aI2781
aI2921
aI15181
aI18649
aI93493
asS'd'
I10996
sa(dp21992
S'a'
I44933
sS's'
I17
sS'm'
(lp21993
I1
aI3
aI3
aI5
aI1
aI25
aI3
aI101
aI151
aI435
aI1339
aI1207
aI7687
aI12579
aI29331
aI4653
aI67353
asS'd'
I10997
sa(dp21994
S'a'
I44934
sS's'
I17
sS'm'
(lp21995
I1
aI1
aI7
aI1
aI29
aI53
aI101
aI61
aI31
aI633
aI1899
aI3919
aI1879
aI3143
aI25319
aI45809
aI77425
asS'd'
I10998
sa(dp21996
S'a'
I44937
sS's'
I17
sS'm'
(lp21997
I1
aI1
aI5
aI1
aI17
aI31
aI79
aI247
aI77
aI197
aI1693
aI313
aI2183
aI14343
aI4511
aI26009
aI44943
asS'd'
I10999
sa(dp21998
S'a'
I44940
sS's'
I17
sS'm'
(lp21999
I1
aI1
aI7
aI5
aI31
aI29
aI119
aI251
aI345
aI867
aI271
aI165
aI6425
aI8343
aI11251
aI28125
aI34849
asS'd'
I11000
sa(dp22000
S'a'
I44951
sS's'
I17
sS'm'
(lp22001
I1
aI3
aI1
aI1
aI13
aI35
aI9
aI103
aI365
aI675
aI1653
aI4095
aI3123
aI8245
aI4679
aI18951
aI88543
asS'd'
I11001
sa(dp22002
S'a'
I44961
sS's'
I17
sS'm'
(lp22003
I1
aI1
aI1
aI1
aI23
aI29
aI109
aI157
aI253
aI751
aI145
aI2077
aI4555
aI7523
aI30099
aI37709
aI97369
asS'd'
I11002
sa(dp22004
S'a'
I44962
sS's'
I17
sS'm'
(lp22005
I1
aI3
aI3
aI11
aI5
aI1
aI51
aI11
aI203
aI963
aI1961
aI351
aI6697
aI8137
aI25933
aI53505
aI28531
asS'd'
I11003
sa(dp22006
S'a'
I44971
sS's'
I17
sS'm'
(lp22007
I1
aI1
aI7
aI15
aI27
aI1
aI31
aI159
aI447
aI501
aI1873
aI2845
aI875
aI1671
aI5049
aI38901
aI32559
asS'd'
I11004
sa(dp22008
S'a'
I44982
sS's'
I17
sS'm'
(lp22009
I1
aI1
aI3
aI3
aI29
aI19
aI33
aI83
aI71
aI703
aI1861
aI3683
aI3589
aI15339
aI21075
aI40399
aI47853
asS'd'
I11005
sa(dp22010
S'a'
I44985
sS's'
I17
sS'm'
(lp22011
I1
aI3
aI3
aI7
aI5
aI41
aI61
aI181
aI319
aI77
aI777
aI2537
aI3887
aI2687
aI29227
aI55217
aI55813
asS'd'
I11006
sa(dp22012
S'a'
I44996
sS's'
I17
sS'm'
(lp22013
I1
aI3
aI3
aI1
aI25
aI41
aI23
aI31
aI31
aI775
aI693
aI891
aI861
aI7613
aI9557
aI43275
aI36311
asS'd'
I11007
sa(dp22014
S'a'
I44999
sS's'
I17
sS'm'
(lp22015
I1
aI1
aI7
aI13
aI11
aI5
aI99
aI217
aI81
aI441
aI765
aI3981
aI2921
aI9657
aI6905
aI30657
aI18395
asS'd'
I11008
sa(dp22016
S'a'
I45014
sS's'
I17
sS'm'
(lp22017
I1
aI3
aI1
aI11
aI21
aI55
aI25
aI209
aI13
aI1021
aI1373
aI785
aI3243
aI1541
aI12033
aI17309
aI116517
asS'd'
I11009
sa(dp22018
S'a'
I45029
sS's'
I17
sS'm'
(lp22019
I1
aI1
aI1
aI7
aI3
aI3
aI61
aI113
aI453
aI405
aI1321
aI2327
aI3529
aI12779
aI11707
aI55795
aI105137
asS'd'
I11010
sa(dp22020
S'a'
I45033
sS's'
I17
sS'm'
(lp22021
I1
aI3
aI1
aI13
aI15
aI53
aI17
aI189
aI197
aI459
aI1999
aI935
aI7835
aI9563
aI31231
aI47757
aI80807
asS'd'
I11011
sa(dp22022
S'a'
I45036
sS's'
I17
sS'm'
(lp22023
I1
aI3
aI5
aI13
aI11
aI15
aI91
aI115
aI427
aI723
aI1815
aI3527
aI5917
aI4931
aI28297
aI12257
aI5587
asS'd'
I11012
sa(dp22024
S'a'
I45047
sS's'
I17
sS'm'
(lp22025
I1
aI1
aI5
aI9
aI31
aI5
aI77
aI201
aI373
aI143
aI581
aI1199
aI6807
aI6059
aI3133
aI57069
aI4895
asS'd'
I11013
sa(dp22026
S'a'
I45065
sS's'
I17
sS'm'
(lp22027
I1
aI3
aI1
aI9
aI17
aI13
aI127
aI61
aI235
aI991
aI279
aI1545
aI2875
aI8453
aI13329
aI39763
aI66897
asS'd'
I11014
sa(dp22028
S'a'
I45076
sS's'
I17
sS'm'
(lp22029
I1
aI1
aI3
aI15
aI31
aI51
aI3
aI95
aI221
aI685
aI635
aI1747
aI177
aI9781
aI4859
aI45345
aI37607
asS'd'
I11015
sa(dp22030
S'a'
I45085
sS's'
I17
sS'm'
(lp22031
I1
aI3
aI5
aI1
aI3
aI55
aI63
aI51
aI63
aI707
aI883
aI2985
aI3699
aI3881
aI8159
aI41775
aI41411
asS'd'
I11016
sa(dp22032
S'a'
I45086
sS's'
I17
sS'm'
(lp22033
I1
aI1
aI1
aI11
aI3
aI41
aI69
aI181
aI413
aI33
aI525
aI1883
aI6063
aI13787
aI1259
aI19497
aI8119
asS'd'
I11017
sa(dp22034
S'a'
I45090
sS's'
I17
sS'm'
(lp22035
I1
aI1
aI5
aI15
aI13
aI27
aI65
aI63
aI117
aI831
aI855
aI369
aI1005
aI9069
aI16179
aI32027
aI6527
asS'd'
I11018
sa(dp22036
S'a'
I45107
sS's'
I17
sS'm'
(lp22037
I1
aI3
aI7
aI5
aI25
aI51
aI63
aI163
aI101
aI299
aI1637
aI641
aI2077
aI9195
aI11181
aI59783
aI109481
asS'd'
I11019
sa(dp22038
S'a'
I45119
sS's'
I17
sS'm'
(lp22039
I1
aI3
aI5
aI13
aI27
aI13
aI117
aI253
aI257
aI919
aI709
aI411
aI5525
aI1247
aI19951
aI51423
aI34605
asS'd'
I11020
sa(dp22040
S'a'
I45121
sS's'
I17
sS'm'
(lp22041
I1
aI1
aI5
aI5
aI1
aI37
aI49
aI125
aI87
aI291
aI339
aI3235
aI1477
aI9787
aI19637
aI22855
aI103013
asS'd'
I11021
sa(dp22042
S'a'
I45128
sS's'
I17
sS'm'
(lp22043
I1
aI3
aI7
aI15
aI25
aI17
aI77
aI23
aI303
aI739
aI1921
aI1425
aI6451
aI9521
aI6311
aI38551
aI123683
asS'd'
I11022
sa(dp22044
S'a'
I45139
sS's'
I17
sS'm'
(lp22045
I1
aI3
aI1
aI7
aI13
aI19
aI33
aI73
aI347
aI85
aI1693
aI3671
aI713
aI1191
aI3285
aI6815
aI61833
asS'd'
I11023
sa(dp22046
S'a'
I45151
sS's'
I17
sS'm'
(lp22047
I1
aI1
aI3
aI3
aI13
aI53
aI81
aI177
aI305
aI967
aI551
aI1177
aI2315
aI4899
aI5733
aI11147
aI128895
asS'd'
I11024
sa(dp22048
S'a'
I45157
sS's'
I17
sS'm'
(lp22049
I1
aI3
aI5
aI3
aI17
aI17
aI93
aI173
aI417
aI645
aI1631
aI1817
aI6127
aI3545
aI6127
aI22331
aI59751
asS'd'
I11025
sa(dp22050
S'a'
I45162
sS's'
I17
sS'm'
(lp22051
I1
aI1
aI5
aI11
aI7
aI53
aI61
aI117
aI133
aI141
aI283
aI3351
aI6745
aI599
aI7221
aI50583
aI9067
asS'd'
I11026
sa(dp22052
S'a'
I45164
sS's'
I17
sS'm'
(lp22053
I1
aI3
aI7
aI3
aI29
aI45
aI71
aI177
aI97
aI897
aI589
aI3319
aI1821
aI7207
aI25715
aI13043
aI96695
asS'd'
I11027
sa(dp22054
S'a'
I45176
sS's'
I17
sS'm'
(lp22055
I1
aI3
aI3
aI1
aI13
aI39
aI19
aI49
aI419
aI905
aI1063
aI4023
aI145
aI1479
aI22197
aI43883
aI45503
asS'd'
I11028
sa(dp22056
S'a'
I45179
sS's'
I17
sS'm'
(lp22057
I1
aI3
aI3
aI15
aI9
aI45
aI45
aI201
aI61
aI193
aI375
aI2439
aI2339
aI15981
aI5197
aI6285
aI109389
asS'd'
I11029
sa(dp22058
S'a'
I45198
sS's'
I17
sS'm'
(lp22059
I1
aI1
aI7
aI13
aI29
aI51
aI93
aI223
aI509
aI1003
aI1861
aI3715
aI2511
aI13843
aI25297
aI1241
aI12157
asS'd'
I11030
sa(dp22060
S'a'
I45209
sS's'
I17
sS'm'
(lp22061
I1
aI3
aI5
aI15
aI19
aI17
aI95
aI243
aI251
aI485
aI1837
aI1829
aI2081
aI15117
aI29635
aI63861
aI100397
asS'd'
I11031
sa(dp22062
S'a'
I45231
sS's'
I17
sS'm'
(lp22063
I1
aI1
aI7
aI3
aI1
aI37
aI31
aI53
aI483
aI849
aI1197
aI3069
aI2539
aI2529
aI12749
aI64331
aI45757
asS'd'
I11032
sa(dp22064
S'a'
I45234
sS's'
I17
sS'm'
(lp22065
I1
aI3
aI7
aI7
aI1
aI19
aI25
aI243
aI335
aI99
aI1507
aI2155
aI6085
aI2253
aI32439
aI16141
aI6781
asS'd'
I11033
sa(dp22066
S'a'
I45236
sS's'
I17
sS'm'
(lp22067
I1
aI3
aI7
aI15
aI9
aI13
aI35
aI63
aI371
aI373
aI1891
aI3913
aI4577
aI15553
aI13079
aI60251
aI71193
asS'd'
I11034
sa(dp22068
S'a'
I45251
sS's'
I17
sS'm'
(lp22069
I1
aI3
aI1
aI7
aI15
aI13
aI105
aI113
aI409
aI289
aI57
aI1095
aI791
aI15675
aI21471
aI42851
aI29203
asS'd'
I11035
sa(dp22070
S'a'
I45260
sS's'
I17
sS'm'
(lp22071
I1
aI1
aI1
aI13
aI1
aI57
aI65
aI7
aI153
aI929
aI1325
aI229
aI3841
aI8967
aI29889
aI49427
aI46853
asS'd'
I11036
sa(dp22072
S'a'
I45268
sS's'
I17
sS'm'
(lp22073
I1
aI1
aI3
aI11
aI29
aI1
aI79
aI111
aI479
aI931
aI1619
aI505
aI4503
aI4055
aI18849
aI3979
aI46091
asS'd'
I11037
sa(dp22074
S'a'
I45277
sS's'
I17
sS'm'
(lp22075
I1
aI1
aI7
aI3
aI31
aI27
aI127
aI63
aI219
aI43
aI883
aI1265
aI5733
aI9051
aI17059
aI61625
aI93843
asS'd'
I11038
sa(dp22076
S'a'
I45299
sS's'
I17
sS'm'
(lp22077
I1
aI1
aI7
aI7
aI23
aI21
aI35
aI211
aI243
aI399
aI1225
aI1415
aI5923
aI2143
aI25303
aI36171
aI126349
asS'd'
I11039
sa(dp22078
S'a'
I45301
sS's'
I17
sS'm'
(lp22079
I1
aI3
aI1
aI3
aI3
aI13
aI77
aI205
aI271
aI393
aI769
aI2101
aI4045
aI6159
aI3409
aI44065
aI102799
asS'd'
I11040
sa(dp22080
S'a'
I45338
sS's'
I17
sS'm'
(lp22081
I1
aI1
aI5
aI15
aI19
aI1
aI67
aI199
aI367
aI51
aI495
aI2051
aI3195
aI15239
aI10525
aI45319
aI50489
asS'd'
I11041
sa(dp22082
S'a'
I45344
sS's'
I17
sS'm'
(lp22083
I1
aI1
aI1
aI9
aI3
aI19
aI105
aI147
aI417
aI399
aI373
aI1025
aI2727
aI13779
aI30079
aI22723
aI41551
asS'd'
I11042
sa(dp22084
S'a'
I45349
sS's'
I17
sS'm'
(lp22085
I1
aI1
aI3
aI1
aI9
aI15
aI105
aI95
aI267
aI995
aI275
aI2627
aI3883
aI10785
aI8075
aI40591
aI54647
asS'd'
I11043
sa(dp22086
S'a'
I45364
sS's'
I17
sS'm'
(lp22087
I1
aI1
aI1
aI5
aI31
aI37
aI117
aI185
aI55
aI273
aI525
aI445
aI4221
aI2081
aI16017
aI19859
aI3297
asS'd'
I11044
sa(dp22088
S'a'
I45367
sS's'
I17
sS'm'
(lp22089
I1
aI3
aI5
aI13
aI21
aI13
aI105
aI231
aI461
aI831
aI393
aI3253
aI1213
aI2625
aI3393
aI36715
aI104889
asS'd'
I11045
sa(dp22090
S'a'
I45371
sS's'
I17
sS'm'
(lp22091
I1
aI3
aI5
aI15
aI1
aI17
aI103
aI129
aI257
aI1003
aI285
aI2927
aI3967
aI53
aI5197
aI39665
aI50751
asS'd'
I11046
sa(dp22092
S'a'
I45373
sS's'
I17
sS'm'
(lp22093
I1
aI1
aI1
aI13
aI1
aI61
aI47
aI255
aI137
aI849
aI213
aI301
aI681
aI9547
aI28209
aI32941
aI72109
asS'd'
I11047
sa(dp22094
S'a'
I45376
sS's'
I17
sS'm'
(lp22095
I1
aI1
aI7
aI11
aI31
aI15
aI81
aI117
aI327
aI289
aI1861
aI861
aI6189
aI13425
aI18279
aI7635
aI116969
asS'd'
I11048
sa(dp22096
S'a'
I45381
sS's'
I17
sS'm'
(lp22097
I1
aI3
aI3
aI3
aI9
aI11
aI13
aI181
aI183
aI621
aI329
aI2751
aI3989
aI6345
aI20319
aI52267
aI79695
asS'd'
I11049
sa(dp22098
S'a'
I45400
sS's'
I17
sS'm'
(lp22099
I1
aI1
aI7
aI13
aI9
aI1
aI5
aI125
aI1
aI735
aI691
aI13
aI3961
aI2273
aI18299
aI65221
aI20115
asS'd'
I11050
sa(dp22100
S'a'
I45406
sS's'
I17
sS'm'
(lp22101
I1
aI3
aI7
aI1
aI7
aI3
aI87
aI115
aI241
aI101
aI523
aI3019
aI7571
aI7721
aI27409
aI49751
aI97859
asS'd'
I11051
sa(dp22102
S'a'
I45416
sS's'
I17
sS'm'
(lp22103
I1
aI3
aI5
aI11
aI9
aI5
aI33
aI59
aI299
aI191
aI307
aI2115
aI2823
aI10187
aI10437
aI34137
aI93217
asS'd'
I11052
sa(dp22104
S'a'
I45422
sS's'
I17
sS'm'
(lp22105
I1
aI3
aI3
aI7
aI21
aI31
aI5
aI113
aI77
aI215
aI177
aI2029
aI7241
aI4465
aI31489
aI10165
aI19035
asS'd'
I11053
sa(dp22106
S'a'
I45427
sS's'
I17
sS'm'
(lp22107
I1
aI3
aI5
aI1
aI27
aI63
aI11
aI161
aI435
aI941
aI1593
aI1765
aI1519
aI9111
aI12787
aI35961
aI105263
asS'd'
I11054
sa(dp22108
S'a'
I45440
sS's'
I17
sS'm'
(lp22109
I1
aI1
aI1
aI9
aI11
aI57
aI41
aI229
aI387
aI617
aI1991
aI221
aI2857
aI4337
aI13851
aI23185
aI111031
asS'd'
I11055
sa(dp22110
S'a'
I45458
sS's'
I17
sS'm'
(lp22111
I1
aI1
aI3
aI5
aI21
aI27
aI125
aI83
aI129
aI919
aI65
aI403
aI2981
aI10111
aI17017
aI24829
aI12205
asS'd'
I11056
sa(dp22112
S'a'
I45467
sS's'
I17
sS'm'
(lp22113
I1
aI3
aI3
aI9
aI25
aI19
aI109
aI47
aI199
aI395
aI1909
aI2819
aI5361
aI6629
aI7067
aI18755
aI17921
asS'd'
I11057
sa(dp22114
S'a'
I45474
sS's'
I17
sS'm'
(lp22115
I1
aI1
aI3
aI15
aI25
aI37
aI111
aI129
aI409
aI291
aI1403
aI2785
aI3819
aI10245
aI24647
aI64799
aI64951
asS'd'
I11058
sa(dp22116
S'a'
I45476
sS's'
I17
sS'm'
(lp22117
I1
aI3
aI5
aI11
aI1
aI7
aI105
aI223
aI427
aI661
aI1817
aI1023
aI145
aI927
aI6507
aI13235
aI30147
asS'd'
I11059
sa(dp22118
S'a'
I45488
sS's'
I17
sS'm'
(lp22119
I1
aI3
aI5
aI13
aI7
aI15
aI65
aI125
aI121
aI113
aI923
aI2729
aI1397
aI14247
aI8487
aI54907
aI41921
asS'd'
I11060
sa(dp22120
S'a'
I45494
sS's'
I17
sS'm'
(lp22121
I1
aI1
aI5
aI1
aI13
aI15
aI47
aI111
aI453
aI375
aI1705
aI1539
aI4103
aI601
aI7499
aI33287
aI123689
asS'd'
I11061
sa(dp22122
S'a'
I45497
sS's'
I17
sS'm'
(lp22123
I1
aI1
aI5
aI3
aI21
aI11
aI87
aI115
aI483
aI617
aI1593
aI2817
aI6519
aI16203
aI361
aI34415
aI100829
asS'd'
I11062
sa(dp22124
S'a'
I45500
sS's'
I17
sS'm'
(lp22125
I1
aI3
aI7
aI15
aI23
aI25
aI41
aI193
aI473
aI517
aI1195
aI3627
aI1089
aI13391
aI3653
aI25637
aI5643
asS'd'
I11063
sa(dp22126
S'a'
I45512
sS's'
I17
sS'm'
(lp22127
I1
aI3
aI1
aI1
aI13
aI57
aI29
aI175
aI35
aI107
aI5
aI3641
aI1843
aI1507
aI7591
aI39967
aI66859
asS'd'
I11064
sa(dp22128
S'a'
I45515
sS's'
I17
sS'm'
(lp22129
I1
aI1
aI3
aI13
aI1
aI39
aI31
aI11
aI493
aI123
aI523
aI843
aI133
aI7971
aI14131
aI51927
aI97943
asS'd'
I11065
sa(dp22130
S'a'
I45523
sS's'
I17
sS'm'
(lp22131
I1
aI1
aI3
aI7
aI23
aI45
aI5
aI195
aI195
aI683
aI497
aI1215
aI5855
aI14569
aI20441
aI29541
aI30431
asS'd'
I11066
sa(dp22132
S'a'
I45542
sS's'
I17
sS'm'
(lp22133
I1
aI3
aI1
aI11
aI31
aI39
aI127
aI187
aI187
aI17
aI817
aI907
aI4657
aI8223
aI13305
aI36489
aI28909
asS'd'
I11067
sa(dp22134
S'a'
I45553
sS's'
I17
sS'm'
(lp22135
I1
aI1
aI7
aI13
aI9
aI1
aI59
aI27
aI449
aI887
aI39
aI191
aI803
aI2339
aI5213
aI2611
aI93175
asS'd'
I11068
sa(dp22136
S'a'
I45559
sS's'
I17
sS'm'
(lp22137
I1
aI1
aI1
aI1
aI29
aI17
aI105
aI13
aI175
aI401
aI1145
aI297
aI6873
aI889
aI10301
aI48993
aI49959
asS'd'
I11069
sa(dp22138
S'a'
I45589
sS's'
I17
sS'm'
(lp22139
I1
aI3
aI5
aI5
aI1
aI57
aI81
aI81
aI403
aI719
aI1887
aI2597
aI1069
aI5219
aI29767
aI46905
aI8025
asS'd'
I11070
sa(dp22140
S'a'
I45594
sS's'
I17
sS'm'
(lp22141
I1
aI1
aI5
aI11
aI13
aI37
aI41
aI3
aI487
aI895
aI343
aI1729
aI3777
aI8681
aI24737
aI34179
aI15015
asS'd'
I11071
sa(dp22142
S'a'
I45596
sS's'
I17
sS'm'
(lp22143
I1
aI1
aI1
aI15
aI9
aI43
aI67
aI203
aI71
aI399
aI23
aI529
aI2375
aI15373
aI21013
aI17389
aI93809
asS'd'
I11072
sa(dp22144
S'a'
I45603
sS's'
I17
sS'm'
(lp22145
I1
aI3
aI7
aI7
aI9
aI23
aI81
aI27
aI39
aI529
aI631
aI199
aI3555
aI953
aI4249
aI39297
aI88107
asS'd'
I11073
sa(dp22146
S'a'
I45605
sS's'
I17
sS'm'
(lp22147
I1
aI3
aI1
aI3
aI31
aI45
aI33
aI63
aI319
aI245
aI1567
aI3359
aI2051
aI11523
aI30177
aI20293
aI13245
asS'd'
I11074
sa(dp22148
S'a'
I45610
sS's'
I17
sS'm'
(lp22149
I1
aI1
aI1
aI13
aI9
aI61
aI39
aI127
aI453
aI1019
aI2037
aI3541
aI6983
aI10717
aI19587
aI8981
aI99637
asS'd'
I11075
sa(dp22150
S'a'
I45630
sS's'
I17
sS'm'
(lp22151
I1
aI3
aI5
aI9
aI15
aI7
aI55
aI79
aI93
aI303
aI1423
aI499
aI5499
aI795
aI14553
aI16945
aI46161
asS'd'
I11076
sa(dp22152
S'a'
I45638
sS's'
I17
sS'm'
(lp22153
I1
aI1
aI7
aI5
aI21
aI21
aI27
aI201
aI147
aI461
aI363
aI267
aI2963
aI3409
aI17835
aI40777
aI71879
asS'd'
I11077
sa(dp22154
S'a'
I45641
sS's'
I17
sS'm'
(lp22155
I1
aI1
aI7
aI9
aI23
aI63
aI115
aI243
aI103
aI119
aI2023
aI2223
aI7989
aI1365
aI26181
aI4631
aI88001
asS'd'
I11078
sa(dp22156
S'a'
I45647
sS's'
I17
sS'm'
(lp22157
I1
aI3
aI5
aI5
aI27
aI57
aI101
aI199
aI461
aI853
aI449
aI2733
aI2225
aI8609
aI19461
aI15265
aI54079
asS'd'
I11079
sa(dp22158
S'a'
I45655
sS's'
I17
sS'm'
(lp22159
I1
aI3
aI3
aI15
aI29
aI59
aI115
aI105
aI145
aI391
aI303
aI901
aI5481
aI1491
aI30441
aI22331
aI3841
asS'd'
I11080
sa(dp22160
S'a'
I45659
sS's'
I17
sS'm'
(lp22161
I1
aI1
aI3
aI1
aI27
aI45
aI11
aI167
aI73
aI181
aI253
aI1947
aI1731
aI15269
aI16971
aI12299
aI46439
asS'd'
I11081
sa(dp22162
S'a'
I45665
sS's'
I17
sS'm'
(lp22163
I1
aI1
aI7
aI13
aI11
aI21
aI83
aI157
aI75
aI705
aI1709
aI487
aI5029
aI9879
aI27589
aI21601
aI50575
asS'd'
I11082
sa(dp22164
S'a'
I45689
sS's'
I17
sS'm'
(lp22165
I1
aI1
aI5
aI3
aI27
aI37
aI101
aI163
aI115
aI903
aI1137
aI3807
aI2899
aI3407
aI27935
aI14203
aI31009
asS'd'
I11083
sa(dp22166
S'a'
I45695
sS's'
I17
sS'm'
(lp22167
I1
aI3
aI5
aI9
aI31
aI33
aI63
aI69
aI159
aI737
aI1973
aI3661
aI6159
aI1781
aI9239
aI12989
aI82947
asS'd'
I11084
sa(dp22168
S'a'
I45702
sS's'
I17
sS'm'
(lp22169
I1
aI3
aI5
aI9
aI15
aI33
aI41
aI89
aI183
aI933
aI1305
aI1013
aI7245
aI16225
aI10891
aI6641
aI61699
asS'd'
I11085
sa(dp22170
S'a'
I45708
sS's'
I17
sS'm'
(lp22171
I1
aI1
aI5
aI3
aI25
aI41
aI91
aI183
aI45
aI553
aI1817
aI3305
aI5169
aI9051
aI24917
aI52431
aI52505
asS'd'
I11086
sa(dp22172
S'a'
I45726
sS's'
I17
sS'm'
(lp22173
I1
aI3
aI3
aI9
aI3
aI9
aI127
aI59
aI117
aI1001
aI1255
aI3435
aI3797
aI8507
aI28593
aI24119
aI75569
asS'd'
I11087
sa(dp22174
S'a'
I45729
sS's'
I17
sS'm'
(lp22175
I1
aI3
aI1
aI5
aI17
aI43
aI45
aI21
aI461
aI339
aI1127
aI2213
aI7351
aI14585
aI2001
aI32619
aI33825
asS'd'
I11088
sa(dp22176
S'a'
I45739
sS's'
I17
sS'm'
(lp22177
I1
aI1
aI5
aI11
aI3
aI37
aI61
aI83
aI101
aI707
aI861
aI3037
aI1867
aI7747
aI16313
aI58745
aI14387
asS'd'
I11089
sa(dp22178
S'a'
I45744
sS's'
I17
sS'm'
(lp22179
I1
aI1
aI5
aI3
aI27
aI25
aI99
aI17
aI293
aI867
aI1655
aI2301
aI2007
aI7379
aI14487
aI18233
aI3625
asS'd'
I11090
sa(dp22180
S'a'
I45747
sS's'
I17
sS'm'
(lp22181
I1
aI1
aI7
aI13
aI25
aI29
aI21
aI133
aI207
aI119
aI423
aI1561
aI6587
aI1221
aI27295
aI48141
aI125473
asS'd'
I11091
sa(dp22182
S'a'
I45762
sS's'
I17
sS'm'
(lp22183
I1
aI3
aI3
aI1
aI19
aI45
aI39
aI85
aI127
aI249
aI157
aI1307
aI7343
aI6309
aI31073
aI16909
aI93223
asS'd'
I11092
sa(dp22184
S'a'
I45764
sS's'
I17
sS'm'
(lp22185
I1
aI1
aI5
aI13
aI19
aI43
aI111
aI109
aI385
aI847
aI1071
aI1009
aI2783
aI8471
aI5719
aI50459
aI110507
asS'd'
I11093
sa(dp22186
S'a'
I45773
sS's'
I17
sS'm'
(lp22187
I1
aI1
aI5
aI15
aI1
aI45
aI39
aI197
aI209
aI839
aI485
aI3943
aI5939
aI11835
aI18297
aI61217
aI85015
asS'd'
I11094
sa(dp22188
S'a'
I45774
sS's'
I17
sS'm'
(lp22189
I1
aI1
aI1
aI15
aI5
aI61
aI1
aI195
aI415
aI355
aI1593
aI151
aI8143
aI3527
aI11633
aI44337
aI99749
asS'd'
I11095
sa(dp22190
S'a'
I45781
sS's'
I17
sS'm'
(lp22191
I1
aI1
aI5
aI13
aI11
aI11
aI117
aI109
aI91
aI663
aI1351
aI2361
aI1409
aI9317
aI31133
aI17577
aI123919
asS'd'
I11096
sa(dp22192
S'a'
I45785
sS's'
I17
sS'm'
(lp22193
I1
aI3
aI3
aI9
aI3
aI5
aI115
aI173
aI459
aI937
aI1581
aI781
aI1069
aI573
aI24025
aI30721
aI116837
asS'd'
I11097
sa(dp22194
S'a'
I45792
sS's'
I17
sS'm'
(lp22195
I1
aI1
aI1
aI5
aI21
aI37
aI47
aI51
aI21
aI169
aI119
aI3285
aI2543
aI14023
aI29179
aI13407
aI130491
asS'd'
I11098
sa(dp22196
S'a'
I45801
sS's'
I17
sS'm'
(lp22197
I1
aI3
aI5
aI5
aI25
aI27
aI41
aI147
aI485
aI79
aI737
aI699
aI6763
aI16347
aI9265
aI52129
aI41431
asS'd'
I11099
sa(dp22198
S'a'
I45802
sS's'
I17
sS'm'
(lp22199
I1
aI1
aI1
aI3
aI5
aI33
aI115
aI187
aI311
aI717
aI1897
aI2215
aI2639
aI4167
aI1429
aI26359
aI52703
asS'd'
I11100
sa(dp22200
S'a'
I45812
sS's'
I17
sS'm'
(lp22201
I1
aI3
aI5
aI5
aI13
aI51
aI103
aI5
aI47
aI683
aI319
aI2969
aI7701
aI11031
aI9257
aI16725
aI80825
asS'd'
I11101
sa(dp22202
S'a'
I45816
sS's'
I17
sS'm'
(lp22203
I1
aI3
aI1
aI11
aI31
aI47
aI17
aI205
aI11
aI411
aI523
aI4053
aI6743
aI3095
aI3219
aI63163
aI84547
asS'd'
I11102
sa(dp22204
S'a'
I45829
sS's'
I17
sS'm'
(lp22205
I1
aI1
aI7
aI15
aI9
aI55
aI109
aI225
aI273
aI595
aI1697
aI2059
aI21
aI11319
aI23277
aI60613
aI4539
asS'd'
I11103
sa(dp22206
S'a'
I45833
sS's'
I17
sS'm'
(lp22207
I1
aI1
aI5
aI13
aI3
aI59
aI49
aI239
aI509
aI847
aI975
aI3361
aI5443
aI1941
aI29277
aI56379
aI38997
asS'd'
I11104
sa(dp22208
S'a'
I45847
sS's'
I17
sS'm'
(lp22209
I1
aI3
aI1
aI7
aI15
aI5
aI49
aI19
aI235
aI437
aI1309
aI827
aI4123
aI5839
aI22409
aI42535
aI98041
asS'd'
I11105
sa(dp22210
S'a'
I45851
sS's'
I17
sS'm'
(lp22211
I1
aI1
aI5
aI15
aI9
aI33
aI57
aI153
aI165
aI215
aI177
aI1271
aI1861
aI15489
aI4183
aI43701
aI114169
asS'd'
I11106
sa(dp22212
S'a'
I45854
sS's'
I17
sS'm'
(lp22213
I1
aI3
aI5
aI5
aI13
aI3
aI119
aI89
aI17
aI421
aI1205
aI835
aI4917
aI6113
aI28991
aI26839
aI114871
asS'd'
I11107
sa(dp22214
S'a'
I45863
sS's'
I17
sS'm'
(lp22215
I1
aI3
aI5
aI1
aI7
aI49
aI49
aI159
aI205
aI601
aI1939
aI4063
aI5975
aI11747
aI10329
aI21103
aI16779
asS'd'
I11108
sa(dp22216
S'a'
I45870
sS's'
I17
sS'm'
(lp22217
I1
aI1
aI5
aI15
aI13
aI33
aI89
aI21
aI113
aI639
aI891
aI989
aI829
aI1435
aI11475
aI42711
aI67049
asS'd'
I11109
sa(dp22218
S'a'
I45901
sS's'
I17
sS'm'
(lp22219
I1
aI1
aI5
aI5
aI9
aI59
aI57
aI105
aI385
aI733
aI1175
aI329
aI6809
aI7175
aI27267
aI9941
aI14203
asS'd'
I11110
sa(dp22220
S'a'
I45910
sS's'
I17
sS'm'
(lp22221
I1
aI3
aI1
aI13
aI21
aI53
aI83
aI139
aI287
aI659
aI1991
aI3225
aI4153
aI4325
aI16803
aI27719
aI86263
asS'd'
I11111
sa(dp22222
S'a'
I45920
sS's'
I17
sS'm'
(lp22223
I1
aI3
aI5
aI13
aI27
aI21
aI111
aI105
aI29
aI573
aI405
aI2781
aI1737
aI12057
aI25263
aI16903
aI45389
asS'd'
I11112
sa(dp22224
S'a'
I45932
sS's'
I17
sS'm'
(lp22225
I1
aI1
aI5
aI5
aI23
aI23
aI61
aI27
aI335
aI279
aI937
aI2509
aI4751
aI2993
aI28069
aI30187
aI3595
asS'd'
I11113
sa(dp22226
S'a'
I45938
sS's'
I17
sS'm'
(lp22227
I1
aI1
aI5
aI7
aI29
aI37
aI117
aI71
aI221
aI875
aI1987
aI2329
aI5953
aI15901
aI29813
aI17419
aI4745
asS'd'
I11114
sa(dp22228
S'a'
I45940
sS's'
I17
sS'm'
(lp22229
I1
aI3
aI3
aI13
aI21
aI51
aI77
aI85
aI53
aI573
aI1129
aI3415
aI2283
aI5221
aI29991
aI46091
aI65843
asS'd'
I11115
sa(dp22230
S'a'
I45943
sS's'
I17
sS'm'
(lp22231
I1
aI1
aI1
aI3
aI17
aI51
aI89
aI211
aI463
aI743
aI1189
aI4083
aI1437
aI5219
aI8373
aI15559
aI18557
asS'd'
I11116
sa(dp22232
S'a'
I45949
sS's'
I17
sS'm'
(lp22233
I1
aI1
aI5
aI3
aI29
aI27
aI1
aI207
aI285
aI739
aI505
aI1587
aI6565
aI14195
aI4995
aI39453
aI61023
asS'd'
I11117
sa(dp22234
S'a'
I45953
sS's'
I17
sS'm'
(lp22235
I1
aI3
aI3
aI15
aI7
aI57
aI19
aI45
aI39
aI881
aI1207
aI2829
aI3265
aI2637
aI7843
aI62889
aI53289
asS'd'
I11118
sa(dp22236
S'a'
I45963
sS's'
I17
sS'm'
(lp22237
I1
aI1
aI1
aI11
aI31
aI21
aI73
aI245
aI87
aI457
aI1523
aI2397
aI1157
aI8237
aI26195
aI23149
aI106523
asS'd'
I11119
sa(dp22238
S'a'
I45971
sS's'
I17
sS'm'
(lp22239
I1
aI3
aI5
aI13
aI3
aI55
aI3
aI179
aI107
aI85
aI639
aI2711
aI6359
aI1599
aI2325
aI59573
aI111941
asS'd'
I11120
sa(dp22240
S'a'
I45989
sS's'
I17
sS'm'
(lp22241
I1
aI1
aI1
aI13
aI17
aI61
aI45
aI253
aI45
aI149
aI1251
aI139
aI7113
aI6503
aI27675
aI37301
aI21713
asS'd'
I11121
sa(dp22242
S'a'
I45999
sS's'
I17
sS'm'
(lp22243
I1
aI3
aI5
aI9
aI31
aI31
aI67
aI79
aI355
aI225
aI1187
aI761
aI4927
aI5481
aI9139
aI13399
aI35653
asS'd'
I11122
sa(dp22244
S'a'
I46001
sS's'
I17
sS'm'
(lp22245
I1
aI1
aI5
aI3
aI7
aI3
aI95
aI119
aI161
aI529
aI1443
aI1099
aI609
aI3919
aI10935
aI37779
aI92993
asS'd'
I11123
sa(dp22246
S'a'
I46008
sS's'
I17
sS'm'
(lp22247
I1
aI1
aI7
aI9
aI13
aI21
aI13
aI7
aI165
aI173
aI989
aI2315
aI2305
aI13115
aI6933
aI56233
aI112113
asS'd'
I11124
sa(dp22248
S'a'
I46022
sS's'
I17
sS'm'
(lp22249
I1
aI1
aI7
aI3
aI9
aI11
aI25
aI45
aI493
aI119
aI839
aI3907
aI2273
aI14113
aI29453
aI55181
aI667
asS'd'
I11125
sa(dp22250
S'a'
I46026
sS's'
I17
sS'm'
(lp22251
I1
aI3
aI5
aI15
aI25
aI33
aI15
aI23
aI245
aI517
aI1883
aI2865
aI1483
aI7043
aI32615
aI12261
aI49297
asS'd'
I11126
sa(dp22252
S'a'
I46034
sS's'
I17
sS'm'
(lp22253
I1
aI3
aI7
aI9
aI31
aI35
aI89
aI103
aI245
aI441
aI1709
aI1321
aI3743
aI3767
aI23885
aI43587
aI18017
asS'd'
I11127
sa(dp22254
S'a'
I46040
sS's'
I17
sS'm'
(lp22255
I1
aI3
aI7
aI5
aI23
aI43
aI103
aI7
aI47
aI187
aI1257
aI3517
aI591
aI16263
aI12047
aI16699
aI81633
asS'd'
I11128
sa(dp22256
S'a'
I46043
sS's'
I17
sS'm'
(lp22257
I1
aI1
aI5
aI15
aI5
aI5
aI79
aI11
aI327
aI719
aI37
aI2913
aI6107
aI3463
aI25901
aI6125
aI100647
asS'd'
I11129
sa(dp22258
S'a'
I46045
sS's'
I17
sS'm'
(lp22259
I1
aI1
aI1
aI11
aI13
aI29
aI83
aI251
aI41
aI125
aI1137
aI2627
aI4643
aI29
aI24631
aI51435
aI98643
asS'd'
I11130
sa(dp22260
S'a'
I46061
sS's'
I17
sS'm'
(lp22261
I1
aI1
aI7
aI3
aI27
aI3
aI69
aI245
aI365
aI599
aI1575
aI2969
aI3441
aI12327
aI18951
aI56167
aI13861
asS'd'
I11131
sa(dp22262
S'a'
I46062
sS's'
I17
sS'm'
(lp22263
I1
aI3
aI3
aI11
aI5
aI47
aI103
aI233
aI351
aI821
aI867
aI3199
aI6133
aI4627
aI22663
aI14775
aI83205
asS'd'
I11132
sa(dp22264
S'a'
I46076
sS's'
I17
sS'm'
(lp22265
I1
aI3
aI1
aI13
aI9
aI35
aI27
aI251
aI281
aI727
aI873
aI3713
aI5247
aI8407
aI17739
aI57207
aI126201
asS'd'
I11133
sa(dp22266
S'a'
I46084
sS's'
I17
sS'm'
(lp22267
I1
aI1
aI7
aI11
aI11
aI35
aI53
aI115
aI93
aI663
aI625
aI565
aI3137
aI7869
aI18845
aI49155
aI83395
asS'd'
I11134
sa(dp22268
S'a'
I46094
sS's'
I17
sS'm'
(lp22269
I1
aI3
aI3
aI5
aI21
aI13
aI99
aI151
aI319
aI9
aI1363
aI1489
aI2545
aI1963
aI1271
aI24815
aI43355
asS'd'
I11135
sa(dp22270
S'a'
I46102
sS's'
I17
sS'm'
(lp22271
I1
aI3
aI5
aI9
aI15
aI51
aI109
aI85
aI67
aI131
aI1947
aI181
aI7331
aI15163
aI2255
aI33449
aI78107
asS'd'
I11136
sa(dp22272
S'a'
I46118
sS's'
I17
sS'm'
(lp22273
I1
aI1
aI7
aI9
aI27
aI61
aI1
aI163
aI309
aI739
aI453
aI1837
aI2093
aI16021
aI8485
aI19755
aI61335
asS'd'
I11137
sa(dp22274
S'a'
I46149
sS's'
I17
sS'm'
(lp22275
I1
aI3
aI5
aI7
aI3
aI13
aI11
aI195
aI91
aI143
aI203
aI2785
aI7319
aI7153
aI19265
aI11597
aI63365
asS'd'
I11138
sa(dp22276
S'a'
I46154
sS's'
I17
sS'm'
(lp22277
I1
aI3
aI1
aI9
aI29
aI1
aI123
aI247
aI253
aI757
aI191
aI1699
aI6625
aI1785
aI29199
aI29409
aI32577
asS'd'
I11139
sa(dp22278
S'a'
I46167
sS's'
I17
sS'm'
(lp22279
I1
aI3
aI5
aI11
aI23
aI21
aI31
aI35
aI383
aI587
aI65
aI1695
aI4045
aI12305
aI12437
aI5919
aI51465
asS'd'
I11140
sa(dp22280
S'a'
I46173
sS's'
I17
sS'm'
(lp22281
I1
aI1
aI5
aI9
aI11
aI13
aI123
aI171
aI499
aI877
aI1785
aI561
aI2547
aI1797
aI27679
aI56305
aI93223
asS'd'
I11141
sa(dp22282
S'a'
I46177
sS's'
I17
sS'm'
(lp22283
I1
aI3
aI1
aI3
aI25
aI41
aI63
aI243
aI219
aI533
aI753
aI1903
aI3257
aI11901
aI4777
aI28629
aI111141
asS'd'
I11142
sa(dp22284
S'a'
I46192
sS's'
I17
sS'm'
(lp22285
I1
aI3
aI3
aI5
aI31
aI47
aI1
aI253
aI283
aI995
aI1787
aI1767
aI6599
aI11913
aI21515
aI39259
aI117727
asS'd'
I11143
sa(dp22286
S'a'
I46197
sS's'
I17
sS'm'
(lp22287
I1
aI1
aI7
aI7
aI31
aI35
aI39
aI255
aI463
aI763
aI881
aI2583
aI347
aI14343
aI22761
aI45821
aI119155
asS'd'
I11144
sa(dp22288
S'a'
I46201
sS's'
I17
sS'm'
(lp22289
I1
aI3
aI5
aI9
aI5
aI37
aI43
aI55
aI423
aI525
aI157
aI3593
aI2831
aI11539
aI15675
aI11695
aI100609
asS'd'
I11145
sa(dp22290
S'a'
I46214
sS's'
I17
sS'm'
(lp22291
I1
aI3
aI3
aI5
aI11
aI9
aI27
aI57
aI409
aI201
aI1029
aI2461
aI5823
aI2593
aI32031
aI4203
aI55327
asS'd'
I11146
sa(dp22292
S'a'
I46217
sS's'
I17
sS'm'
(lp22293
I1
aI1
aI7
aI15
aI15
aI25
aI69
aI83
aI309
aI687
aI1607
aI819
aI7381
aI3697
aI5289
aI33153
aI48157
asS'd'
I11147
sa(dp22294
S'a'
I46223
sS's'
I17
sS'm'
(lp22295
I1
aI1
aI5
aI5
aI31
aI57
aI41
aI195
aI201
aI59
aI2045
aI2213
aI6695
aI3839
aI17331
aI4981
aI26803
asS'd'
I11148
sa(dp22296
S'a'
I46226
sS's'
I17
sS'm'
(lp22297
I1
aI1
aI1
aI11
aI7
aI53
aI109
aI169
aI387
aI181
aI391
aI19
aI4159
aI299
aI29059
aI27781
aI110193
asS'd'
I11149
sa(dp22298
S'a'
I46228
sS's'
I17
sS'm'
(lp22299
I1
aI3
aI7
aI15
aI5
aI31
aI95
aI155
aI47
aI601
aI1463
aI1799
aI8027
aI3003
aI18067
aI24589
aI108171
asS'd'
I11150
sa(dp22300
S'a'
I46237
sS's'
I17
sS'm'
(lp22301
I1
aI3
aI3
aI7
aI11
aI61
aI21
aI121
aI117
aI149
aI1037
aI3829
aI3581
aI15223
aI17051
aI34539
aI37263
asS'd'
I11151
sa(dp22302
S'a'
I46241
sS's'
I17
sS'm'
(lp22303
I1
aI3
aI3
aI3
aI15
aI15
aI115
aI91
aI443
aI309
aI1073
aI2053
aI789
aI7415
aI26253
aI62657
aI49729
asS'd'
I11152
sa(dp22304
S'a'
I46251
sS's'
I17
sS'm'
(lp22305
I1
aI3
aI5
aI5
aI31
aI19
aI23
aI221
aI19
aI105
aI1105
aI2025
aI4209
aI7531
aI30191
aI40777
aI46069
asS'd'
I11153
sa(dp22306
S'a'
I46259
sS's'
I17
sS'm'
(lp22307
I1
aI1
aI1
aI7
aI29
aI45
aI29
aI215
aI33
aI21
aI1147
aI1725
aI3711
aI2759
aI12731
aI57031
aI42361
asS'd'
I11154
sa(dp22308
S'a'
I46262
sS's'
I17
sS'm'
(lp22309
I1
aI1
aI5
aI15
aI13
aI59
aI111
aI169
aI317
aI841
aI1387
aI3513
aI3137
aI8265
aI31789
aI26963
aI126219
asS'd'
I11155
sa(dp22310
S'a'
I46266
sS's'
I17
sS'm'
(lp22311
I1
aI1
aI3
aI1
aI23
aI21
aI13
aI113
aI71
aI177
aI345
aI3149
aI1461
aI12945
aI3971
aI59759
aI61839
asS'd'
I11156
sa(dp22312
S'a'
I46294
sS's'
I17
sS'm'
(lp22313
I1
aI3
aI5
aI13
aI9
aI19
aI103
aI17
aI95
aI617
aI1477
aI263
aI4259
aI12899
aI24351
aI47431
aI11583
asS'd'
I11157
sa(dp22314
S'a'
I46300
sS's'
I17
sS'm'
(lp22315
I1
aI3
aI5
aI11
aI13
aI11
aI7
aI61
aI13
aI63
aI1687
aI631
aI381
aI5899
aI10225
aI19657
aI37087
asS'd'
I11158
sa(dp22316
S'a'
I46303
sS's'
I17
sS'm'
(lp22317
I1
aI3
aI5
aI3
aI7
aI53
aI11
aI193
aI103
aI427
aI1097
aI299
aI2905
aI5019
aI31803
aI28931
aI47495
asS'd'
I11159
sa(dp22318
S'a'
I46309
sS's'
I17
sS'm'
(lp22319
I1
aI3
aI3
aI11
aI21
aI31
aI125
aI249
aI233
aI941
aI975
aI2287
aI7837
aI6481
aI11021
aI52829
aI63023
asS'd'
I11160
sa(dp22320
S'a'
I46321
sS's'
I17
sS'm'
(lp22321
I1
aI1
aI7
aI15
aI13
aI17
aI33
aI85
aI503
aI11
aI689
aI637
aI4063
aI12223
aI1835
aI17161
aI35213
asS'd'
I11161
sa(dp22322
S'a'
I46331
sS's'
I17
sS'm'
(lp22323
I1
aI3
aI5
aI13
aI25
aI37
aI21
aI135
aI377
aI623
aI895
aI2547
aI2757
aI9055
aI17337
aI65457
aI24737
asS'd'
I11162
sa(dp22324
S'a'
I46339
sS's'
I17
sS'm'
(lp22325
I1
aI1
aI1
aI15
aI25
aI11
aI17
aI95
aI65
aI271
aI1791
aI841
aI1441
aI11177
aI10087
aI63963
aI71481
asS'd'
I11163
sa(dp22326
S'a'
I46342
sS's'
I17
sS'm'
(lp22327
I1
aI3
aI5
aI11
aI31
aI1
aI37
aI107
aI109
aI459
aI1185
aI2155
aI271
aI11775
aI23243
aI53517
aI103669
asS'd'
I11164
sa(dp22328
S'a'
I46354
sS's'
I17
sS'm'
(lp22329
I1
aI3
aI1
aI9
aI1
aI49
aI23
aI141
aI169
aI475
aI469
aI2271
aI1379
aI13139
aI1765
aI63625
aI14143
asS'd'
I11165
sa(dp22330
S'a'
I46370
sS's'
I17
sS'm'
(lp22331
I1
aI1
aI1
aI9
aI13
aI7
aI23
aI219
aI381
aI105
aI743
aI1745
aI2999
aI661
aI7245
aI39653
aI99913
asS'd'
I11166
sa(dp22332
S'a'
I46372
sS's'
I17
sS'm'
(lp22333
I1
aI3
aI1
aI11
aI7
aI35
aI9
aI215
aI41
aI537
aI1569
aI1803
aI3613
aI667
aI15089
aI39485
aI85457
asS'd'
I11167
sa(dp22334
S'a'
I46379
sS's'
I17
sS'm'
(lp22335
I1
aI1
aI1
aI15
aI9
aI49
aI75
aI235
aI119
aI97
aI273
aI209
aI2707
aI2071
aI21943
aI60249
aI57737
asS'd'
I11168
sa(dp22336
S'a'
I46382
sS's'
I17
sS'm'
(lp22337
I1
aI1
aI5
aI15
aI31
aI19
aI33
aI49
aI279
aI461
aI143
aI3001
aI3539
aI1015
aI27597
aI35389
aI36483
asS'd'
I11169
sa(dp22338
S'a'
I46393
sS's'
I17
sS'm'
(lp22339
I1
aI3
aI7
aI1
aI23
aI51
aI123
aI247
aI485
aI343
aI1365
aI593
aI6465
aI12305
aI29375
aI30641
aI43165
asS'd'
I11170
sa(dp22340
S'a'
I46402
sS's'
I17
sS'm'
(lp22341
I1
aI1
aI7
aI13
aI15
aI11
aI125
aI51
aI235
aI717
aI1427
aI3203
aI1711
aI12607
aI8805
aI5773
aI27467
asS'd'
I11171
sa(dp22342
S'a'
I46407
sS's'
I17
sS'm'
(lp22343
I1
aI3
aI5
aI5
aI5
aI13
aI51
aI181
aI133
aI977
aI469
aI2513
aI6819
aI12985
aI8917
aI47317
aI47557
asS'd'
I11172
sa(dp22344
S'a'
I46416
sS's'
I17
sS'm'
(lp22345
I1
aI1
aI5
aI7
aI3
aI7
aI71
aI17
aI345
aI921
aI1621
aI2801
aI5825
aI827
aI17711
aI33701
aI113503
asS'd'
I11173
sa(dp22346
S'a'
I46425
sS's'
I17
sS'm'
(lp22347
I1
aI3
aI1
aI1
aI17
aI7
aI99
aI83
aI73
aI349
aI567
aI713
aI5639
aI4969
aI11549
aI35317
aI28995
asS'd'
I11174
sa(dp22348
S'a'
I46437
sS's'
I17
sS'm'
(lp22349
I1
aI3
aI3
aI9
aI11
aI53
aI123
aI227
aI391
aI775
aI1013
aI3971
aI6183
aI14453
aI6403
aI57063
aI7123
asS'd'
I11175
sa(dp22350
S'a'
I46444
sS's'
I17
sS'm'
(lp22351
I1
aI1
aI1
aI15
aI27
aI13
aI51
aI147
aI151
aI535
aI2017
aI3019
aI6791
aI3931
aI12529
aI30855
aI33243
asS'd'
I11176
sa(dp22352
S'a'
I46447
sS's'
I17
sS'm'
(lp22353
I1
aI3
aI1
aI1
aI27
aI43
aI103
aI85
aI135
aI207
aI621
aI251
aI3723
aI10893
aI29533
aI31023
aI11043
asS'd'
I11177
sa(dp22354
S'a'
I46449
sS's'
I17
sS'm'
(lp22355
I1
aI3
aI7
aI5
aI7
aI39
aI55
aI133
aI141
aI63
aI237
aI3299
aI861
aI15123
aI11859
aI13271
aI32893
asS'd'
I11178
sa(dp22356
S'a'
I46456
sS's'
I17
sS'm'
(lp22357
I1
aI3
aI7
aI9
aI23
aI17
aI73
aI197
aI113
aI725
aI137
aI2835
aI2877
aI6913
aI22949
aI56071
aI67597
asS'd'
I11179
sa(dp22358
S'a'
I46462
sS's'
I17
sS'm'
(lp22359
I1
aI3
aI7
aI7
aI15
aI3
aI15
aI253
aI51
aI443
aI15
aI2549
aI7833
aI4713
aI29211
aI22339
aI6009
asS'd'
I11180
sa(dp22360
S'a'
I46472
sS's'
I17
sS'm'
(lp22361
I1
aI3
aI5
aI15
aI19
aI55
aI35
aI59
aI281
aI995
aI1113
aI605
aI2345
aI10009
aI14629
aI11757
aI53241
asS'd'
I11181
sa(dp22362
S'a'
I46475
sS's'
I17
sS'm'
(lp22363
I1
aI1
aI1
aI11
aI3
aI31
aI11
aI193
aI437
aI1003
aI873
aI909
aI6513
aI2045
aI10495
aI17387
aI25461
asS'd'
I11182
sa(dp22364
S'a'
I46486
sS's'
I17
sS'm'
(lp22365
I1
aI1
aI3
aI5
aI21
aI61
aI47
aI177
aI379
aI773
aI951
aI419
aI4455
aI10171
aI17403
aI19045
aI87327
asS'd'
I11183
sa(dp22366
S'a'
I46490
sS's'
I17
sS'm'
(lp22367
I1
aI3
aI3
aI11
aI31
aI41
aI69
aI229
aI207
aI299
aI1743
aI1417
aI4785
aI1327
aI26967
aI43077
aI124319
asS'd'
I11184
sa(dp22368
S'a'
I46495
sS's'
I17
sS'm'
(lp22369
I1
aI3
aI7
aI15
aI19
aI37
aI65
aI219
aI33
aI691
aI205
aI1577
aI4775
aI8427
aI28315
aI53559
aI100789
asS'd'
I11185
sa(dp22370
S'a'
I46501
sS's'
I17
sS'm'
(lp22371
I1
aI3
aI5
aI11
aI5
aI39
aI55
aI147
aI139
aI871
aI1563
aI3661
aI4791
aI423
aI30007
aI1589
aI20255
asS'd'
I11186
sa(dp22372
S'a'
I46514
sS's'
I17
sS'm'
(lp22373
I1
aI1
aI3
aI15
aI5
aI61
aI89
aI83
aI261
aI519
aI1367
aI2019
aI3799
aI8237
aI9011
aI28995
aI1587
asS'd'
I11187
sa(dp22374
S'a'
I46526
sS's'
I17
sS'm'
(lp22375
I1
aI1
aI7
aI15
aI17
aI55
aI49
aI41
aI353
aI507
aI1565
aI3365
aI7947
aI10391
aI1323
aI61591
aI126305
asS'd'
I11188
sa(dp22376
S'a'
I46528
sS's'
I17
sS'm'
(lp22377
I1
aI3
aI3
aI5
aI13
aI19
aI49
aI195
aI355
aI915
aI1867
aI3513
aI1239
aI4809
aI16925
aI22947
aI92641
asS'd'
I11189
sa(dp22378
S'a'
I46538
sS's'
I17
sS'm'
(lp22379
I1
aI1
aI7
aI3
aI31
aI51
aI45
aI241
aI55
aI195
aI1233
aI3675
aI8077
aI4981
aI17679
aI53025
aI77927
asS'd'
I11190
sa(dp22380
S'a'
I46551
sS's'
I17
sS'm'
(lp22381
I1
aI3
aI5
aI11
aI23
aI5
aI93
aI49
aI277
aI979
aI1093
aI3031
aI6131
aI8085
aI19121
aI45305
aI6705
asS'd'
I11191
sa(dp22382
S'a'
I46552
sS's'
I17
sS'm'
(lp22383
I1
aI1
aI7
aI13
aI29
aI25
aI83
aI105
aI469
aI729
aI1495
aI2607
aI2681
aI13959
aI101
aI1913
aI2671
asS'd'
I11192
sa(dp22384
S'a'
I46564
sS's'
I17
sS'm'
(lp22385
I1
aI1
aI1
aI15
aI7
aI19
aI63
aI105
aI253
aI807
aI1889
aI2433
aI1591
aI16267
aI11997
aI18939
aI113313
asS'd'
I11193
sa(dp22386
S'a'
I46579
sS's'
I17
sS'm'
(lp22387
I1
aI1
aI5
aI3
aI13
aI23
aI29
aI227
aI337
aI115
aI783
aI475
aI6949
aI9485
aI1797
aI18713
aI123981
asS'd'
I11194
sa(dp22388
S'a'
I46586
sS's'
I17
sS'm'
(lp22389
I1
aI3
aI3
aI9
aI21
aI43
aI115
aI225
aI147
aI753
aI919
aI1157
aI2901
aI14813
aI30035
aI52553
aI30225
asS'd'
I11195
sa(dp22390
S'a'
I46602
sS's'
I17
sS'm'
(lp22391
I1
aI1
aI3
aI15
aI23
aI19
aI65
aI83
aI457
aI965
aI579
aI2133
aI291
aI2033
aI7533
aI52995
aI92243
asS'd'
I11196
sa(dp22392
S'a'
I46615
sS's'
I17
sS'm'
(lp22393
I1
aI1
aI1
aI5
aI11
aI3
aI23
aI245
aI255
aI373
aI1119
aI3695
aI6449
aI13497
aI817
aI32215
aI103599
asS'd'
I11197
sa(dp22394
S'a'
I46616
sS's'
I17
sS'm'
(lp22395
I1
aI3
aI1
aI5
aI5
aI19
aI57
aI53
aI145
aI441
aI1253
aI929
aI1299
aI11491
aI29457
aI11245
aI55717
asS'd'
I11198
sa(dp22396
S'a'
I46621
sS's'
I17
sS'm'
(lp22397
I1
aI1
aI7
aI13
aI9
aI13
aI73
aI85
aI127
aI29
aI629
aI409
aI2487
aI13079
aI3767
aI27985
aI110139
asS'd'
I11199
sa(dp22398
S'a'
I46635
sS's'
I17
sS'm'
(lp22399
I1
aI1
aI5
aI11
aI29
aI39
aI27
aI9
aI487
aI623
aI757
aI2879
aI669
aI12521
aI23471
aI47933
aI41721
asS'd'
I11200
sa(dp22400
S'a'
I46643
sS's'
I17
sS'm'
(lp22401
I1
aI1
aI7
aI15
aI21
aI9
aI59
aI39
aI325
aI787
aI1347
aI3039
aI7333
aI9793
aI19337
aI41285
aI48339
asS'd'
I11201
sa(dp22402
S'a'
I46645
sS's'
I17
sS'm'
(lp22403
I1
aI3
aI3
aI11
aI11
aI21
aI127
aI45
aI173
aI981
aI483
aI3707
aI3651
aI10545
aI16865
aI62105
aI114847
asS'd'
I11202
sa(dp22404
S'a'
I46649
sS's'
I17
sS'm'
(lp22405
I1
aI3
aI5
aI11
aI11
aI49
aI89
aI179
aI393
aI455
aI1775
aI1903
aI8173
aI12589
aI17281
aI57687
aI56153
asS'd'
I11203
sa(dp22406
S'a'
I46667
sS's'
I17
sS'm'
(lp22407
I1
aI3
aI1
aI3
aI7
aI7
aI59
aI223
aI255
aI559
aI375
aI2427
aI6921
aI3709
aI24767
aI16213
aI60373
asS'd'
I11204
sa(dp22408
S'a'
I46670
sS's'
I17
sS'm'
(lp22409
I1
aI1
aI7
aI15
aI3
aI31
aI37
aI129
aI307
aI1023
aI1807
aI519
aI6779
aI8997
aI15383
aI4391
aI61161
asS'd'
I11205
sa(dp22410
S'a'
I46672
sS's'
I17
sS'm'
(lp22411
I1
aI1
aI1
aI1
aI9
aI25
aI53
aI27
aI263
aI459
aI1015
aI417
aI4195
aI10931
aI20507
aI19299
aI82371
asS'd'
I11206
sa(dp22412
S'a'
I46678
sS's'
I17
sS'm'
(lp22413
I1
aI1
aI3
aI1
aI7
aI7
aI49
aI221
aI47
aI7
aI1747
aI1533
aI3089
aI14369
aI32609
aI64157
aI78139
asS'd'
I11207
sa(dp22414
S'a'
I46681
sS's'
I17
sS'm'
(lp22415
I1
aI3
aI7
aI5
aI5
aI13
aI101
aI231
aI227
aI19
aI1359
aI3017
aI1405
aI3715
aI3541
aI933
aI1117
asS'd'
I11208
sa(dp22416
S'a'
I46684
sS's'
I17
sS'm'
(lp22417
I1
aI1
aI7
aI15
aI5
aI63
aI59
aI253
aI269
aI781
aI1639
aI2247
aI1041
aI667
aI7055
aI21221
aI84447
asS'd'
I11209
sa(dp22418
S'a'
I46687
sS's'
I17
sS'm'
(lp22419
I1
aI1
aI7
aI3
aI9
aI13
aI115
aI247
aI215
aI173
aI457
aI1125
aI5613
aI13171
aI17847
aI26323
aI68461
asS'd'
I11210
sa(dp22420
S'a'
I46705
sS's'
I17
sS'm'
(lp22421
I1
aI3
aI5
aI9
aI15
aI19
aI95
aI213
aI425
aI567
aI1625
aI1659
aI6961
aI10569
aI20985
aI17255
aI89919
asS'd'
I11211
sa(dp22422
S'a'
I46708
sS's'
I17
sS'm'
(lp22423
I1
aI3
aI5
aI5
aI3
aI11
aI107
aI123
aI265
aI743
aI499
aI1885
aI6079
aI7791
aI24953
aI30925
aI112517
asS'd'
I11212
sa(dp22424
S'a'
I46721
sS's'
I17
sS'm'
(lp22425
I1
aI1
aI5
aI15
aI29
aI27
aI103
aI195
aI119
aI873
aI1751
aI2091
aI6623
aI7583
aI20413
aI52367
aI16831
asS'd'
I11213
sa(dp22426
S'a'
I46724
sS's'
I17
sS'm'
(lp22427
I1
aI1
aI7
aI7
aI21
aI15
aI111
aI197
aI89
aI107
aI1317
aI2107
aI1951
aI189
aI31663
aI63007
aI21405
asS'd'
I11214
sa(dp22428
S'a'
I46739
sS's'
I17
sS'm'
(lp22429
I1
aI3
aI1
aI13
aI27
aI15
aI93
aI251
aI209
aI93
aI1419
aI3785
aI1899
aI3143
aI3205
aI16309
aI121455
asS'd'
I11215
sa(dp22430
S'a'
I46757
sS's'
I17
sS'm'
(lp22431
I1
aI1
aI7
aI7
aI17
aI43
aI23
aI251
aI425
aI591
aI1101
aI1317
aI6369
aI14209
aI10257
aI33813
aI59557
asS'd'
I11216
sa(dp22432
S'a'
I46762
sS's'
I17
sS'm'
(lp22433
I1
aI1
aI7
aI3
aI5
aI41
aI65
aI59
aI327
aI369
aI1867
aI1045
aI4953
aI3155
aI25679
aI8545
aI22753
asS'd'
I11217
sa(dp22434
S'a'
I46764
sS's'
I17
sS'm'
(lp22435
I1
aI1
aI5
aI3
aI5
aI25
aI117
aI97
aI369
aI721
aI1459
aI2501
aI4899
aI5299
aI3859
aI2509
aI127723
asS'd'
I11218
sa(dp22436
S'a'
I46776
sS's'
I17
sS'm'
(lp22437
I1
aI3
aI3
aI1
aI27
aI13
aI65
aI185
aI255
aI543
aI2013
aI2027
aI1131
aI4067
aI1327
aI44639
aI53275
asS'd'
I11219
sa(dp22438
S'a'
I46784
sS's'
I17
sS'm'
(lp22439
I1
aI1
aI7
aI3
aI11
aI15
aI53
aI67
aI265
aI477
aI971
aI3201
aI51
aI10599
aI23691
aI10493
aI130347
asS'd'
I11220
sa(dp22440
S'a'
I46790
sS's'
I17
sS'm'
(lp22441
I1
aI3
aI7
aI7
aI19
aI35
aI47
aI61
aI375
aI547
aI1867
aI1147
aI7775
aI12757
aI15101
aI63243
aI89817
asS'd'
I11221
sa(dp22442
S'a'
I46796
sS's'
I17
sS'm'
(lp22443
I1
aI1
aI5
aI3
aI5
aI21
aI57
aI59
aI145
aI901
aI835
aI1093
aI6487
aI12727
aI20585
aI6309
aI67803
asS'd'
I11222
sa(dp22444
S'a'
I46799
sS's'
I17
sS'm'
(lp22445
I1
aI1
aI1
aI11
aI13
aI31
aI75
aI171
aI189
aI741
aI1923
aI3503
aI4887
aI15423
aI2499
aI39125
aI4125
asS'd'
I11223
sa(dp22446
S'a'
I46802
sS's'
I17
sS'm'
(lp22447
I1
aI3
aI7
aI1
aI29
aI31
aI103
aI207
aI383
aI631
aI1017
aI1693
aI6251
aI9429
aI17491
aI60959
aI68131
asS'd'
I11224
sa(dp22448
S'a'
I46813
sS's'
I17
sS'm'
(lp22449
I1
aI1
aI1
aI1
aI19
aI45
aI127
aI105
aI451
aI287
aI657
aI3521
aI2021
aI15793
aI8993
aI34837
aI65441
asS'd'
I11225
sa(dp22450
S'a'
I46814
sS's'
I17
sS'm'
(lp22451
I1
aI1
aI7
aI9
aI17
aI15
aI13
aI189
aI255
aI753
aI1779
aI3047
aI1179
aI13201
aI28249
aI5909
aI35775
asS'd'
I11226
sa(dp22452
S'a'
I46824
sS's'
I17
sS'm'
(lp22453
I1
aI3
aI3
aI3
aI5
aI11
aI125
aI207
aI375
aI375
aI135
aI2939
aI1141
aI12211
aI727
aI16137
aI52253
asS'd'
I11227
sa(dp22454
S'a'
I46832
sS's'
I17
sS'm'
(lp22455
I1
aI3
aI7
aI9
aI29
aI3
aI83
aI221
aI281
aI299
aI667
aI3435
aI589
aI8039
aI7991
aI24289
aI13079
asS'd'
I11228
sa(dp22456
S'a'
I46841
sS's'
I17
sS'm'
(lp22457
I1
aI3
aI5
aI13
aI9
aI11
aI11
aI99
aI337
aI155
aI233
aI2497
aI3385
aI15045
aI5783
aI40915
aI19201
asS'd'
I11229
sa(dp22458
S'a'
I46856
sS's'
I17
sS'm'
(lp22459
I1
aI1
aI7
aI5
aI5
aI23
aI25
aI223
aI341
aI149
aI505
aI893
aI4933
aI14899
aI29899
aI207
aI125359
asS'd'
I11230
sa(dp22460
S'a'
I46862
sS's'
I17
sS'm'
(lp22461
I1
aI3
aI3
aI15
aI3
aI5
aI43
aI165
aI21
aI993
aI1091
aI3849
aI6005
aI1905
aI7199
aI13495
aI76915
asS'd'
I11231
sa(dp22462
S'a'
I46864
sS's'
I17
sS'm'
(lp22463
I1
aI1
aI1
aI15
aI1
aI19
aI55
aI217
aI179
aI431
aI935
aI2219
aI8135
aI15071
aI17437
aI43271
aI115963
asS'd'
I11232
sa(dp22464
S'a'
I46869
sS's'
I17
sS'm'
(lp22465
I1
aI1
aI7
aI15
aI7
aI61
aI45
aI157
aI441
aI107
aI1955
aI2877
aI5285
aI12157
aI21783
aI60999
aI102949
asS'd'
I11233
sa(dp22466
S'a'
I46874
sS's'
I17
sS'm'
(lp22467
I1
aI1
aI1
aI3
aI9
aI29
aI41
aI75
aI81
aI73
aI1859
aI2923
aI3009
aI10847
aI30257
aI44527
aI21933
asS'd'
I11234
sa(dp22468
S'a'
I46895
sS's'
I17
sS'm'
(lp22469
I1
aI1
aI3
aI3
aI29
aI29
aI103
aI3
aI401
aI197
aI237
aI3727
aI7919
aI13669
aI26869
aI64987
aI1581
asS'd'
I11235
sa(dp22470
S'a'
I46900
sS's'
I17
sS'm'
(lp22471
I1
aI1
aI5
aI13
aI13
aI47
aI99
aI209
aI45
aI745
aI1239
aI663
aI5535
aI3777
aI10479
aI15327
aI1441
asS'd'
I11236
sa(dp22472
S'a'
I46903
sS's'
I17
sS'm'
(lp22473
I1
aI3
aI1
aI9
aI1
aI49
aI31
aI231
aI15
aI1001
aI773
aI2113
aI1957
aI15271
aI25355
aI7461
aI33089
asS'd'
I11237
sa(dp22474
S'a'
I46927
sS's'
I17
sS'm'
(lp22475
I1
aI1
aI1
aI5
aI13
aI31
aI123
aI123
aI439
aI373
aI1817
aI2555
aI7905
aI3151
aI2311
aI62083
aI45535
asS'd'
I11238
sa(dp22476
S'a'
I46946
sS's'
I17
sS'm'
(lp22477
I1
aI3
aI1
aI3
aI31
aI37
aI83
aI117
aI177
aI483
aI1285
aI1725
aI821
aI2115
aI12893
aI54301
aI36491
asS'd'
I11239
sa(dp22478
S'a'
I46948
sS's'
I17
sS'm'
(lp22479
I1
aI3
aI3
aI3
aI9
aI13
aI87
aI7
aI467
aI287
aI1173
aI2739
aI3293
aI883
aI9123
aI30799
aI110221
asS'd'
I11240
sa(dp22480
S'a'
I46955
sS's'
I17
sS'm'
(lp22481
I1
aI3
aI1
aI15
aI3
aI5
aI119
aI235
aI393
aI789
aI67
aI1193
aI1613
aI8607
aI17371
aI16723
aI103747
asS'd'
I11241
sa(dp22482
S'a'
I46965
sS's'
I17
sS'm'
(lp22483
I1
aI3
aI1
aI13
aI13
aI47
aI61
aI51
aI447
aI1
aI1221
aI1619
aI3785
aI5851
aI10557
aI51181
aI6535
asS'd'
I11242
sa(dp22484
S'a'
I46972
sS's'
I17
sS'm'
(lp22485
I1
aI3
aI5
aI11
aI1
aI1
aI85
aI119
aI195
aI177
aI805
aI1161
aI4851
aI15765
aI24405
aI41757
aI110081
asS'd'
I11243
sa(dp22486
S'a'
I46976
sS's'
I17
sS'm'
(lp22487
I1
aI1
aI7
aI1
aI17
aI21
aI89
aI13
aI59
aI169
aI1847
aI2401
aI6243
aI2841
aI6153
aI16039
aI47407
asS'd'
I11244
sa(dp22488
S'a'
I46985
sS's'
I17
sS'm'
(lp22489
I1
aI1
aI3
aI15
aI21
aI53
aI103
aI187
aI143
aI897
aI65
aI3677
aI213
aI4027
aI22365
aI53703
aI82951
asS'd'
I11245
sa(dp22490
S'a'
I46999
sS's'
I17
sS'm'
(lp22491
I1
aI3
aI5
aI5
aI25
aI5
aI39
aI49
aI55
aI71
aI825
aI2123
aI2345
aI5683
aI18027
aI29897
aI53023
asS'd'
I11246
sa(dp22492
S'a'
I47003
sS's'
I17
sS'm'
(lp22493
I1
aI1
aI3
aI5
aI13
aI55
aI27
aI77
aI327
aI429
aI1219
aI2103
aI7095
aI13461
aI31027
aI15383
aI98485
asS'd'
I11247
sa(dp22494
S'a'
I47006
sS's'
I17
sS'm'
(lp22495
I1
aI3
aI1
aI3
aI9
aI41
aI33
aI241
aI487
aI229
aI1743
aI951
aI2319
aI15595
aI3213
aI5959
aI90721
asS'd'
I11248
sa(dp22496
S'a'
I47016
sS's'
I17
sS'm'
(lp22497
I1
aI1
aI7
aI1
aI29
aI45
aI45
aI163
aI123
aI227
aI305
aI1577
aI5465
aI5639
aI14507
aI65155
aI71425
asS'd'
I11249
sa(dp22498
S'a'
I47030
sS's'
I17
sS'm'
(lp22499
I1
aI1
aI5
aI1
aI3
aI5
aI33
aI15
aI203
aI141
aI465
aI3509
aI6653
aI14193
aI7073
aI22525
aI22951
asS'd'
I11250
sa(dp22500
S'a'
I47034
sS's'
I17
sS'm'
(lp22501
I1
aI1
aI7
aI3
aI27
aI39
aI3
aI27
aI75
aI821
aI1329
aI3655
aI4715
aI7659
aI31957
aI60219
aI79123
asS'd'
I11251
sa(dp22502
S'a'
I47053
sS's'
I17
sS'm'
(lp22503
I1
aI3
aI5
aI13
aI29
aI45
aI111
aI19
aI207
aI387
aI87
aI3731
aI7427
aI13351
aI9497
aI34285
aI25623
asS'd'
I11252
sa(dp22504
S'a'
I47056
sS's'
I17
sS'm'
(lp22505
I1
aI1
aI3
aI11
aI19
aI39
aI79
aI219
aI97
aI125
aI947
aI1397
aI3645
aI1021
aI9403
aI38695
aI54985
asS'd'
I11253
sa(dp22506
S'a'
I47059
sS's'
I17
sS'm'
(lp22507
I1
aI3
aI3
aI7
aI3
aI3
aI45
aI93
aI65
aI289
aI1843
aI1599
aI897
aI16159
aI23485
aI24699
aI43123
asS'd'
I11254
sa(dp22508
S'a'
I47072
sS's'
I17
sS'm'
(lp22509
I1
aI3
aI5
aI3
aI15
aI1
aI81
aI219
aI299
aI429
aI1115
aI1763
aI6381
aI869
aI7817
aI143
aI23583
asS'd'
I11255
sa(dp22510
S'a'
I47077
sS's'
I17
sS'm'
(lp22511
I1
aI1
aI3
aI5
aI1
aI35
aI95
aI147
aI425
aI1011
aI1039
aI2875
aI3089
aI3685
aI9995
aI13279
aI60923
asS'd'
I11256
sa(dp22512
S'a'
I47099
sS's'
I17
sS'm'
(lp22513
I1
aI1
aI5
aI7
aI5
aI59
aI105
aI241
aI151
aI307
aI735
aI1541
aI3115
aI12331
aI19535
aI56965
aI127015
asS'd'
I11257
sa(dp22514
S'a'
I47105
sS's'
I17
sS'm'
(lp22515
I1
aI3
aI5
aI7
aI5
aI33
aI83
aI179
aI65
aI397
aI787
aI3425
aI1305
aI10713
aI6973
aI9007
aI112081
asS'd'
I11258
sa(dp22516
S'a'
I47108
sS's'
I17
sS'm'
(lp22517
I1
aI3
aI7
aI13
aI29
aI11
aI37
aI31
aI271
aI501
aI897
aI1383
aI5333
aI13627
aI22091
aI38421
aI94575
asS'd'
I11259
sa(dp22518
S'a'
I47112
sS's'
I17
sS'm'
(lp22519
I1
aI3
aI1
aI1
aI9
aI61
aI87
aI115
aI13
aI79
aI391
aI2385
aI7157
aI3369
aI26035
aI883
aI34705
asS'd'
I11260
sa(dp22520
S'a'
I47115
sS's'
I17
sS'm'
(lp22521
I1
aI1
aI3
aI9
aI21
aI29
aI15
aI165
aI53
aI719
aI1509
aI1295
aI4437
aI8229
aI17961
aI55297
aI62635
asS'd'
I11261
sa(dp22522
S'a'
I47117
sS's'
I17
sS'm'
(lp22523
I1
aI1
aI7
aI5
aI27
aI21
aI23
aI141
aI341
aI423
aI9
aI2693
aI5555
aI5797
aI13179
aI1107
aI33489
asS'd'
I11262
sa(dp22524
S'a'
I47120
sS's'
I17
sS'm'
(lp22525
I1
aI3
aI1
aI13
aI27
aI33
aI101
aI29
aI379
aI119
aI1259
aI861
aI6843
aI69
aI3253
aI61977
aI80061
asS'd'
I11263
sa(dp22526
S'a'
I47129
sS's'
I17
sS'm'
(lp22527
I1
aI1
aI3
aI11
aI1
aI37
aI35
aI43
aI105
aI655
aI221
aI873
aI91
aI9095
aI8999
aI44033
aI24807
asS'd'
I11264
sa(dp22528
S'a'
I47135
sS's'
I17
sS'm'
(lp22529
I1
aI3
aI3
aI7
aI29
aI15
aI23
aI227
aI399
aI305
aI2007
aI747
aI2717
aI14767
aI6515
aI40617
aI7873
asS'd'
I11265
sa(dp22530
S'a'
I47151
sS's'
I17
sS'm'
(lp22531
I1
aI1
aI5
aI15
aI21
aI43
aI39
aI7
aI261
aI421
aI505
aI1433
aI1155
aI5621
aI2337
aI54027
aI54039
asS'd'
I11266
sa(dp22532
S'a'
I47154
sS's'
I17
sS'm'
(lp22533
I1
aI1
aI1
aI9
aI15
aI39
aI49
aI185
aI503
aI895
aI1321
aI375
aI4245
aI4929
aI9637
aI50561
aI65733
asS'd'
I11267
sa(dp22534
S'a'
I47156
sS's'
I17
sS'm'
(lp22535
I1
aI1
aI5
aI15
aI7
aI29
aI27
aI155
aI423
aI631
aI1295
aI973
aI4227
aI2637
aI8479
aI29527
aI70505
asS'd'
I11268
sa(dp22536
S'a'
I47166
sS's'
I17
sS'm'
(lp22537
I1
aI3
aI1
aI3
aI31
aI33
aI13
aI67
aI195
aI335
aI1577
aI3715
aI559
aI7251
aI7215
aI46443
aI125359
asS'd'
I11269
sa(dp22538
S'a'
I47171
sS's'
I17
sS'm'
(lp22539
I1
aI3
aI1
aI9
aI17
aI37
aI15
aI119
aI79
aI851
aI911
aI3549
aI99
aI9221
aI29897
aI63489
aI34937
asS'd'
I11270
sa(dp22540
S'a'
I47195
sS's'
I17
sS'm'
(lp22541
I1
aI3
aI5
aI9
aI31
aI11
aI125
aI1
aI265
aI467
aI835
aI2997
aI2401
aI9615
aI19397
aI50947
aI29963
asS'd'
I11271
sa(dp22542
S'a'
I47204
sS's'
I17
sS'm'
(lp22543
I1
aI1
aI5
aI3
aI11
aI21
aI63
aI15
aI471
aI763
aI1963
aI2815
aI4419
aI11457
aI7151
aI27009
aI124847
asS'd'
I11272
sa(dp22544
S'a'
I47213
sS's'
I17
sS'm'
(lp22545
I1
aI1
aI7
aI3
aI19
aI9
aI97
aI59
aI375
aI109
aI519
aI305
aI2787
aI3001
aI14199
aI27415
aI35403
asS'd'
I11273
sa(dp22546
S'a'
I47222
sS's'
I17
sS'm'
(lp22547
I1
aI1
aI1
aI3
aI13
aI47
aI3
aI93
aI307
aI979
aI419
aI2817
aI3741
aI305
aI1813
aI34549
aI116959
asS'd'
I11274
sa(dp22548
S'a'
I47228
sS's'
I17
sS'm'
(lp22549
I1
aI3
aI5
aI13
aI13
aI19
aI35
aI231
aI493
aI973
aI895
aI1583
aI1843
aI9057
aI27705
aI32333
aI130347
asS'd'
I11275
sa(dp22550
S'a'
I47241
sS's'
I17
sS'm'
(lp22551
I1
aI1
aI1
aI13
aI31
aI35
aI81
aI11
aI363
aI229
aI1865
aI2849
aI7805
aI877
aI3965
aI45337
aI33239
asS'd'
I11276
sa(dp22552
S'a'
I47242
sS's'
I17
sS'm'
(lp22553
I1
aI3
aI5
aI1
aI15
aI3
aI125
aI93
aI191
aI405
aI1359
aI929
aI3085
aI7907
aI7777
aI7815
aI103717
asS'd'
I11277
sa(dp22554
S'a'
I47247
sS's'
I17
sS'm'
(lp22555
I1
aI1
aI3
aI7
aI15
aI33
aI61
aI235
aI283
aI141
aI817
aI1611
aI665
aI13113
aI4197
aI45831
aI44505
asS'd'
I11278
sa(dp22556
S'a'
I47249
sS's'
I17
sS'm'
(lp22557
I1
aI1
aI5
aI5
aI25
aI3
aI63
aI159
aI223
aI531
aI1147
aI2323
aI2715
aI10319
aI32203
aI23943
aI95407
asS'd'
I11279
sa(dp22558
S'a'
I47250
sS's'
I17
sS'm'
(lp22559
I1
aI3
aI3
aI3
aI3
aI33
aI37
aI99
aI317
aI811
aI515
aI339
aI6527
aI11149
aI13071
aI7177
aI1549
asS'd'
I11280
sa(dp22560
S'a'
I47261
sS's'
I17
sS'm'
(lp22561
I1
aI3
aI1
aI11
aI31
aI53
aI9
aI17
aI297
aI259
aI1235
aI53
aI7065
aI1721
aI8191
aI21663
aI13393
asS'd'
I11281
sa(dp22562
S'a'
I47266
sS's'
I17
sS'm'
(lp22563
I1
aI3
aI7
aI7
aI29
aI41
aI127
aI179
aI113
aI191
aI783
aI861
aI6509
aI5199
aI3369
aI18327
aI30647
asS'd'
I11282
sa(dp22564
S'a'
I47268
sS's'
I17
sS'm'
(lp22565
I1
aI3
aI1
aI7
aI23
aI41
aI49
aI155
aI135
aI513
aI1127
aI1443
aI8081
aI2553
aI10389
aI35459
aI122513
asS'd'
I11283
sa(dp22566
S'a'
I47278
sS's'
I17
sS'm'
(lp22567
I1
aI1
aI7
aI3
aI25
aI3
aI117
aI125
aI283
aI165
aI1409
aI1587
aI7915
aI12899
aI12239
aI48161
aI7385
asS'd'
I11284
sa(dp22568
S'a'
I47286
sS's'
I17
sS'm'
(lp22569
I1
aI3
aI7
aI3
aI23
aI19
aI29
aI47
aI7
aI723
aI455
aI4013
aI2739
aI12303
aI29883
aI51485
aI1571
asS'd'
I11285
sa(dp22570
S'a'
I47312
sS's'
I17
sS'm'
(lp22571
I1
aI1
aI1
aI11
aI1
aI31
aI111
aI199
aI207
aI209
aI1163
aI2865
aI5335
aI2647
aI9125
aI6737
aI99881
asS'd'
I11286
sa(dp22572
S'a'
I47315
sS's'
I17
sS'm'
(lp22573
I1
aI3
aI1
aI11
aI31
aI19
aI59
aI153
aI65
aI133
aI1399
aI2709
aI905
aI3257
aI13603
aI46299
aI15139
asS'd'
I11287
sa(dp22574
S'a'
I47321
sS's'
I17
sS'm'
(lp22575
I1
aI3
aI1
aI1
aI23
aI19
aI123
aI115
aI59
aI667
aI333
aI2461
aI1843
aI16049
aI12353
aI17297
aI107779
asS'd'
I11288
sa(dp22576
S'a'
I47322
sS's'
I17
sS'm'
(lp22577
I1
aI1
aI5
aI1
aI5
aI29
aI91
aI241
aI97
aI557
aI1701
aI2441
aI2995
aI13103
aI9261
aI55833
aI843
asS'd'
I11289
sa(dp22578
S'a'
I47324
sS's'
I17
sS'm'
(lp22579
I1
aI1
aI1
aI5
aI15
aI19
aI23
aI189
aI69
aI91
aI427
aI3149
aI5199
aI13073
aI32273
aI41503
aI98749
asS'd'
I11290
sa(dp22580
S'a'
I47331
sS's'
I17
sS'm'
(lp22581
I1
aI3
aI5
aI11
aI19
aI3
aI71
aI125
aI307
aI241
aI861
aI681
aI5657
aI5189
aI7555
aI2037
aI72921
asS'd'
I11291
sa(dp22582
S'a'
I47338
sS's'
I17
sS'm'
(lp22583
I1
aI1
aI5
aI5
aI29
aI61
aI93
aI61
aI421
aI685
aI883
aI1559
aI5875
aI10561
aI11761
aI18879
aI31577
asS'd'
I11292
sa(dp22584
S'a'
I47366
sS's'
I17
sS'm'
(lp22585
I1
aI3
aI7
aI15
aI9
aI29
aI41
aI241
aI365
aI941
aI1087
aI3743
aI6781
aI9467
aI1409
aI20605
aI2361
asS'd'
I11293
sa(dp22586
S'a'
I47372
sS's'
I17
sS'm'
(lp22587
I1
aI1
aI5
aI13
aI25
aI29
aI7
aI21
aI41
aI621
aI1751
aI3731
aI2667
aI8613
aI20999
aI3851
aI39677
asS'd'
I11294
sa(dp22588
S'a'
I47375
sS's'
I17
sS'm'
(lp22589
I1
aI1
aI7
aI5
aI17
aI23
aI25
aI43
aI401
aI749
aI975
aI91
aI5277
aI2711
aI19847
aI41419
aI11851
asS'd'
I11295
sa(dp22590
S'a'
I47378
sS's'
I17
sS'm'
(lp22591
I1
aI1
aI7
aI9
aI13
aI37
aI113
aI89
aI435
aI749
aI1553
aI1853
aI7709
aI5449
aI25055
aI45207
aI2269
asS'd'
I11296
sa(dp22592
S'a'
I47387
sS's'
I17
sS'm'
(lp22593
I1
aI3
aI1
aI11
aI3
aI59
aI79
aI13
aI35
aI901
aI165
aI907
aI7579
aI12739
aI24679
aI54163
aI61059
asS'd'
I11297
sa(dp22594
S'a'
I47389
sS's'
I17
sS'm'
(lp22595
I1
aI1
aI5
aI7
aI9
aI47
aI101
aI61
aI25
aI461
aI1737
aI2825
aI4439
aI5363
aI28433
aI61979
aI120401
asS'd'
I11298
sa(dp22596
S'a'
I47390
sS's'
I17
sS'm'
(lp22597
I1
aI3
aI3
aI3
aI1
aI55
aI103
aI1
aI449
aI473
aI375
aI2609
aI4933
aI3411
aI19663
aI6067
aI61129
asS'd'
I11299
sa(dp22598
S'a'
I47396
sS's'
I17
sS'm'
(lp22599
I1
aI1
aI3
aI9
aI27
aI29
aI53
aI151
aI391
aI507
aI425
aI3469
aI6605
aI5783
aI31747
aI37677
aI116037
asS'd'
I11300
sa(dp22600
S'a'
I47400
sS's'
I17
sS'm'
(lp22601
I1
aI1
aI1
aI5
aI5
aI43
aI61
aI67
aI319
aI553
aI1163
aI3095
aI4447
aI7505
aI15617
aI26167
aI11145
asS'd'
I11301
sa(dp22602
S'a'
I47405
sS's'
I17
sS'm'
(lp22603
I1
aI1
aI5
aI7
aI3
aI3
aI9
aI161
aI155
aI869
aI337
aI3693
aI6847
aI8449
aI15077
aI54769
aI123335
asS'd'
I11302
sa(dp22604
S'a'
I47408
sS's'
I17
sS'm'
(lp22605
I1
aI3
aI3
aI7
aI11
aI19
aI17
aI71
aI105
aI649
aI323
aI3033
aI1695
aI15973
aI6361
aI3163
aI17669
asS'd'
I11303
sa(dp22606
S'a'
I47411
sS's'
I17
sS'm'
(lp22607
I1
aI3
aI7
aI13
aI13
aI9
aI3
aI251
aI149
aI513
aI637
aI2211
aI6397
aI1741
aI8547
aI3165
aI38241
asS'd'
I11304
sa(dp22608
S'a'
I47431
sS's'
I17
sS'm'
(lp22609
I1
aI1
aI5
aI5
aI1
aI57
aI73
aI35
aI287
aI347
aI221
aI3261
aI7693
aI5443
aI6175
aI18181
aI23733
asS'd'
I11305
sa(dp22610
S'a'
I47443
sS's'
I17
sS'm'
(lp22611
I1
aI3
aI7
aI5
aI5
aI19
aI1
aI167
aI5
aI825
aI815
aI1369
aI7657
aI6169
aI9583
aI34761
aI81003
asS'd'
I11306
sa(dp22612
S'a'
I47459
sS's'
I17
sS'm'
(lp22613
I1
aI3
aI3
aI5
aI15
aI17
aI23
aI157
aI475
aI297
aI1495
aI811
aI8135
aI11453
aI9683
aI55505
aI84361
asS'd'
I11307
sa(dp22614
S'a'
I47466
sS's'
I17
sS'm'
(lp22615
I1
aI1
aI1
aI11
aI1
aI63
aI59
aI189
aI205
aI1023
aI1065
aI1095
aI2293
aI14629
aI29399
aI2925
aI28327
asS'd'
I11308
sa(dp22616
S'a'
I47479
sS's'
I17
sS'm'
(lp22617
I1
aI3
aI1
aI3
aI31
aI33
aI85
aI183
aI383
aI731
aI1223
aI3353
aI3703
aI5655
aI31265
aI12249
aI22127
asS'd'
I11309
sa(dp22618
S'a'
I47480
sS's'
I17
sS'm'
(lp22619
I1
aI3
aI5
aI7
aI7
aI29
aI61
aI3
aI375
aI95
aI1815
aI4065
aI6287
aI3797
aI32397
aI50581
aI123371
asS'd'
I11310
sa(dp22620
S'a'
I47495
sS's'
I17
sS'm'
(lp22621
I1
aI3
aI1
aI9
aI13
aI37
aI57
aI139
aI43
aI561
aI425
aI3603
aI1167
aI10281
aI31825
aI32673
aI106169
asS'd'
I11311
sa(dp22622
S'a'
I47504
sS's'
I17
sS'm'
(lp22623
I1
aI3
aI5
aI3
aI15
aI41
aI29
aI223
aI79
aI851
aI1741
aI2241
aI5659
aI9773
aI18369
aI37239
aI14831
asS'd'
I11312
sa(dp22624
S'a'
I47519
sS's'
I17
sS'm'
(lp22625
I1
aI3
aI1
aI1
aI7
aI11
aI83
aI63
aI51
aI521
aI1911
aI475
aI5207
aI3219
aI10257
aI40461
aI9087
asS'd'
I11313
sa(dp22626
S'a'
I47520
sS's'
I17
sS'm'
(lp22627
I1
aI3
aI5
aI7
aI23
aI17
aI97
aI195
aI5
aI451
aI1971
aI1881
aI921
aI8729
aI3443
aI64529
aI67747
asS'd'
I11314
sa(dp22628
S'a'
I47530
sS's'
I17
sS'm'
(lp22629
I1
aI1
aI3
aI9
aI21
aI63
aI45
aI181
aI429
aI605
aI169
aI3493
aI2381
aI2887
aI19515
aI53151
aI60147
asS'd'
I11315
sa(dp22630
S'a'
I47537
sS's'
I17
sS'm'
(lp22631
I1
aI3
aI3
aI7
aI15
aI11
aI117
aI161
aI333
aI139
aI587
aI2331
aI3175
aI12093
aI12649
aI52381
aI107117
asS'd'
I11316
sa(dp22632
S'a'
I47555
sS's'
I17
sS'm'
(lp22633
I1
aI1
aI5
aI15
aI23
aI57
aI19
aI57
aI507
aI461
aI799
aI611
aI1589
aI10909
aI7649
aI17817
aI24677
asS'd'
I11317
sa(dp22634
S'a'
I47558
sS's'
I17
sS'm'
(lp22635
I1
aI1
aI3
aI15
aI1
aI61
aI9
aI201
aI401
aI387
aI527
aI2855
aI2339
aI3813
aI11825
aI14273
aI73745
asS'd'
I11318
sa(dp22636
S'a'
I47567
sS's'
I17
sS'm'
(lp22637
I1
aI3
aI1
aI5
aI1
aI37
aI27
aI105
aI475
aI335
aI1169
aI3233
aI3225
aI12861
aI10133
aI36673
aI55025
asS'd'
I11319
sa(dp22638
S'a'
I47569
sS's'
I17
sS'm'
(lp22639
I1
aI1
aI1
aI9
aI27
aI53
aI55
aI197
aI229
aI5
aI93
aI1157
aI7929
aI9745
aI5295
aI15359
aI75567
asS'd'
I11320
sa(dp22640
S'a'
I47572
sS's'
I17
sS'm'
(lp22641
I1
aI1
aI7
aI5
aI11
aI3
aI109
aI81
aI457
aI159
aI1899
aI557
aI1067
aI295
aI2661
aI1145
aI8745
asS'd'
I11321
sa(dp22642
S'a'
I47582
sS's'
I17
sS'm'
(lp22643
I1
aI1
aI1
aI15
aI31
aI29
aI77
aI171
aI411
aI425
aI1041
aI2791
aI2567
aI5357
aI21871
aI27689
aI103485
asS'd'
I11322
sa(dp22644
S'a'
I47591
sS's'
I17
sS'm'
(lp22645
I1
aI3
aI5
aI11
aI19
aI47
aI23
aI97
aI405
aI667
aI2045
aI2951
aI2063
aI7775
aI20629
aI34283
aI26925
asS'd'
I11323
sa(dp22646
S'a'
I47595
sS's'
I17
sS'm'
(lp22647
I1
aI3
aI5
aI7
aI5
aI51
aI125
aI67
aI165
aI145
aI733
aI1649
aI5787
aI4333
aI24355
aI33397
aI101001
asS'd'
I11324
sa(dp22648
S'a'
I47598
sS's'
I17
sS'm'
(lp22649
I1
aI3
aI7
aI11
aI13
aI7
aI117
aI147
aI449
aI201
aI953
aI553
aI1839
aI6903
aI10417
aI42751
aI36823
asS'd'
I11325
sa(dp22650
S'a'
I47609
sS's'
I17
sS'm'
(lp22651
I1
aI1
aI3
aI5
aI3
aI21
aI109
aI35
aI95
aI953
aI211
aI2849
aI5681
aI16287
aI16553
aI30345
aI69729
asS'd'
I11326
sa(dp22652
S'a'
I47612
sS's'
I17
sS'm'
(lp22653
I1
aI3
aI7
aI5
aI25
aI37
aI21
aI143
aI317
aI621
aI1417
aI283
aI3801
aI15375
aI3799
aI13345
aI59727
asS'd'
I11327
sa(dp22654
S'a'
I47616
sS's'
I17
sS'm'
(lp22655
I1
aI3
aI3
aI9
aI3
aI37
aI113
aI127
aI123
aI979
aI1225
aI2585
aI2055
aI2571
aI16727
aI38863
aI74347
asS'd'
I11328
sa(dp22656
S'a'
I47625
sS's'
I17
sS'm'
(lp22657
I1
aI3
aI3
aI3
aI23
aI13
aI49
aI111
aI277
aI143
aI1171
aI605
aI91
aI13693
aI1971
aI18209
aI114203
asS'd'
I11329
sa(dp22658
S'a'
I47640
sS's'
I17
sS'm'
(lp22659
I1
aI1
aI1
aI9
aI27
aI33
aI73
aI9
aI93
aI343
aI55
aI3045
aI2029
aI3665
aI28483
aI6601
aI72085
asS'd'
I11330
sa(dp22660
S'a'
I47649
sS's'
I17
sS'm'
(lp22661
I1
aI1
aI1
aI15
aI17
aI55
aI87
aI231
aI103
aI1005
aI1451
aI3617
aI7477
aI2045
aI10683
aI39053
aI1289
asS'd'
I11331
sa(dp22662
S'a'
I47655
sS's'
I17
sS'm'
(lp22663
I1
aI3
aI7
aI7
aI25
aI55
aI39
aI89
aI489
aI609
aI1969
aI159
aI7485
aI10713
aI28371
aI14935
aI95347
asS'd'
I11332
sa(dp22664
S'a'
I47661
sS's'
I17
sS'm'
(lp22665
I1
aI3
aI3
aI13
aI25
aI9
aI83
aI167
aI25
aI135
aI2017
aI3313
aI1493
aI7799
aI22479
aI49471
aI20149
asS'd'
I11333
sa(dp22666
S'a'
I47688
sS's'
I17
sS'm'
(lp22667
I1
aI1
aI1
aI15
aI25
aI13
aI63
aI117
aI97
aI449
aI1331
aI229
aI6027
aI1023
aI26705
aI15283
aI60385
asS'd'
I11334
sa(dp22668
S'a'
I47711
sS's'
I17
sS'm'
(lp22669
I1
aI1
aI7
aI5
aI5
aI37
aI3
aI79
aI337
aI861
aI1549
aI915
aI7303
aI1503
aI19245
aI60721
aI45313
asS'd'
I11335
sa(dp22670
S'a'
I47717
sS's'
I17
sS'm'
(lp22671
I1
aI3
aI7
aI15
aI19
aI1
aI11
aI39
aI505
aI757
aI1627
aI2137
aI3209
aI7651
aI31291
aI45913
aI26851
asS'd'
I11336
sa(dp22672
S'a'
I47721
sS's'
I17
sS'm'
(lp22673
I1
aI3
aI1
aI5
aI25
aI21
aI85
aI233
aI171
aI269
aI367
aI1651
aI3961
aI2487
aI1977
aI7027
aI2725
asS'd'
I11337
sa(dp22674
S'a'
I47724
sS's'
I17
sS'm'
(lp22675
I1
aI1
aI5
aI15
aI13
aI23
aI81
aI145
aI201
aI323
aI425
aI2785
aI1149
aI12617
aI11451
aI23205
aI117691
asS'd'
I11338
sa(dp22676
S'a'
I47727
sS's'
I17
sS'm'
(lp22677
I1
aI1
aI7
aI11
aI29
aI37
aI69
aI141
aI15
aI773
aI1299
aI2147
aI8129
aI12227
aI27811
aI58701
aI103637
asS'd'
I11339
sa(dp22678
S'a'
I47730
sS's'
I17
sS'm'
(lp22679
I1
aI3
aI5
aI13
aI3
aI33
aI29
aI227
aI261
aI221
aI823
aI1399
aI5107
aI2423
aI23809
aI42175
aI28207
asS'd'
I11340
sa(dp22680
S'a'
I47739
sS's'
I17
sS'm'
(lp22681
I1
aI3
aI1
aI15
aI21
aI1
aI121
aI255
aI259
aI441
aI45
aI1899
aI2489
aI4155
aI18317
aI52695
aI607
asS'd'
I11341
sa(dp22682
S'a'
I47745
sS's'
I17
sS'm'
(lp22683
I1
aI3
aI5
aI9
aI25
aI43
aI115
aI111
aI329
aI997
aI753
aI1513
aI6949
aI3197
aI28275
aI48855
aI25089
asS'd'
I11342
sa(dp22684
S'a'
I47752
sS's'
I17
sS'm'
(lp22685
I1
aI3
aI3
aI3
aI25
aI9
aI21
aI213
aI111
aI173
aI913
aI1465
aI4437
aI9725
aI1455
aI53517
aI81843
asS'd'
I11343
sa(dp22686
S'a'
I47758
sS's'
I17
sS'm'
(lp22687
I1
aI3
aI5
aI5
aI5
aI25
aI3
aI159
aI19
aI203
aI181
aI3447
aI6395
aI2145
aI11289
aI16797
aI59567
asS'd'
I11344
sa(dp22688
S'a'
I47765
sS's'
I17
sS'm'
(lp22689
I1
aI3
aI3
aI1
aI29
aI25
aI43
aI115
aI257
aI833
aI379
aI941
aI4389
aI5795
aI12593
aI2471
aI127149
asS'd'
I11345
sa(dp22690
S'a'
I47770
sS's'
I17
sS'm'
(lp22691
I1
aI3
aI1
aI9
aI3
aI15
aI81
aI141
aI155
aI515
aI1677
aI2569
aI1105
aI15653
aI24143
aI3439
aI17317
asS'd'
I11346
sa(dp22692
S'a'
I47782
sS's'
I17
sS'm'
(lp22693
I1
aI3
aI1
aI13
aI27
aI13
aI103
aI87
aI27
aI971
aI671
aI629
aI4943
aI13897
aI4003
aI21507
aI40193
asS'd'
I11347
sa(dp22694
S'a'
I47786
sS's'
I17
sS'm'
(lp22695
I1
aI3
aI5
aI9
aI1
aI49
aI11
aI15
aI511
aI837
aI1953
aI1585
aI1867
aI9095
aI543
aI16993
aI115187
asS'd'
I11348
sa(dp22696
S'a'
I47791
sS's'
I17
sS'm'
(lp22697
I1
aI1
aI5
aI11
aI21
aI49
aI37
aI61
aI9
aI629
aI1025
aI1635
aI4047
aI15491
aI28481
aI43235
aI53165
asS'd'
I11349
sa(dp22698
S'a'
I47793
sS's'
I17
sS'm'
(lp22699
I1
aI1
aI5
aI7
aI15
aI57
aI121
aI119
aI405
aI29
aI655
aI3085
aI7131
aI14761
aI2273
aI47113
aI8603
asS'd'
I11350
sa(dp22700
S'a'
I47796
sS's'
I17
sS'm'
(lp22701
I1
aI1
aI3
aI3
aI11
aI45
aI51
aI107
aI367
aI235
aI675
aI3777
aI6081
aI16319
aI19499
aI36893
aI25579
asS'd'
I11351
sa(dp22702
S'a'
I47814
sS's'
I17
sS'm'
(lp22703
I1
aI1
aI7
aI11
aI1
aI13
aI43
aI159
aI415
aI423
aI1223
aI2201
aI1089
aI10189
aI12457
aI26691
aI3603
asS'd'
I11352
sa(dp22704
S'a'
I47823
sS's'
I17
sS'm'
(lp22705
I1
aI1
aI3
aI3
aI3
aI37
aI109
aI67
aI487
aI785
aI637
aI3931
aI929
aI14153
aI25283
aI483
aI14371
asS'd'
I11353
sa(dp22706
S'a'
I47826
sS's'
I17
sS'm'
(lp22707
I1
aI1
aI7
aI9
aI31
aI59
aI9
aI245
aI479
aI113
aI1419
aI3265
aI8131
aI11123
aI32519
aI12141
aI82059
asS'd'
I11354
sa(dp22708
S'a'
I47837
sS's'
I17
sS'm'
(lp22709
I1
aI1
aI1
aI7
aI25
aI25
aI31
aI13
aI81
aI217
aI997
aI1161
aI1049
aI5487
aI8487
aI57807
aI126115
asS'd'
I11355
sa(dp22710
S'a'
I47854
sS's'
I17
sS'm'
(lp22711
I1
aI1
aI3
aI15
aI1
aI23
aI63
aI39
aI503
aI933
aI1915
aI687
aI547
aI779
aI7689
aI38607
aI125229
asS'd'
I11356
sa(dp22712
S'a'
I47856
sS's'
I17
sS'm'
(lp22713
I1
aI1
aI3
aI3
aI31
aI49
aI51
aI87
aI99
aI327
aI783
aI3487
aI7307
aI2759
aI22781
aI56343
aI126805
asS'd'
I11357
sa(dp22714
S'a'
I47871
sS's'
I17
sS'm'
(lp22715
I1
aI3
aI3
aI11
aI27
aI43
aI99
aI197
aI275
aI19
aI775
aI329
aI3815
aI14277
aI3363
aI26375
aI75427
asS'd'
I11358
sa(dp22716
S'a'
I47879
sS's'
I17
sS'm'
(lp22717
I1
aI1
aI5
aI15
aI15
aI19
aI65
aI109
aI103
aI411
aI1591
aI2569
aI1981
aI4773
aI7861
aI6303
aI127421
asS'd'
I11359
sa(dp22718
S'a'
I47880
sS's'
I17
sS'm'
(lp22719
I1
aI3
aI1
aI13
aI1
aI53
aI107
aI151
aI317
aI201
aI1053
aI2701
aI5039
aI2179
aI10085
aI31727
aI85579
asS'd'
I11360
sa(dp22720
S'a'
I47885
sS's'
I17
sS'm'
(lp22721
I1
aI3
aI1
aI9
aI5
aI55
aI71
aI143
aI187
aI29
aI1363
aI2403
aI2675
aI2187
aI12123
aI32825
aI56461
asS'd'
I11361
sa(dp22722
S'a'
I47886
sS's'
I17
sS'm'
(lp22723
I1
aI1
aI7
aI11
aI5
aI57
aI91
aI95
aI85
aI597
aI925
aI11
aI1915
aI9159
aI9997
aI29565
aI111655
asS'd'
I11362
sa(dp22724
S'a'
I47903
sS's'
I17
sS'm'
(lp22725
I1
aI1
aI5
aI9
aI25
aI57
aI79
aI189
aI205
aI687
aI471
aI601
aI5343
aI9031
aI7853
aI6079
aI40567
asS'd'
I11363
sa(dp22726
S'a'
I47907
sS's'
I17
sS'm'
(lp22727
I1
aI1
aI3
aI1
aI31
aI53
aI67
aI191
aI511
aI369
aI1273
aI3859
aI4253
aI14469
aI3427
aI17691
aI22599
asS'd'
I11364
sa(dp22728
S'a'
I47913
sS's'
I17
sS'm'
(lp22729
I1
aI1
aI3
aI9
aI17
aI53
aI101
aI219
aI41
aI663
aI635
aI3889
aI2197
aI8125
aI8313
aI14957
aI111445
asS'd'
I11365
sa(dp22730
S'a'
I47916
sS's'
I17
sS'm'
(lp22731
I1
aI1
aI5
aI5
aI7
aI49
aI113
aI9
aI399
aI257
aI617
aI63
aI2773
aI4411
aI1193
aI54449
aI89003
asS'd'
I11366
sa(dp22732
S'a'
I47928
sS's'
I17
sS'm'
(lp22733
I1
aI1
aI7
aI11
aI13
aI29
aI77
aI73
aI161
aI419
aI985
aI303
aI2237
aI15217
aI26621
aI20441
aI113955
asS'd'
I11367
sa(dp22734
S'a'
I47936
sS's'
I17
sS'm'
(lp22735
I1
aI3
aI3
aI9
aI17
aI39
aI37
aI139
aI289
aI421
aI1021
aI2635
aI2805
aI5815
aI9101
aI48077
aI114009
asS'd'
I11368
sa(dp22736
S'a'
I47945
sS's'
I17
sS'm'
(lp22737
I1
aI1
aI1
aI15
aI25
aI43
aI71
aI35
aI103
aI827
aI1301
aI3567
aI3425
aI3689
aI14453
aI10733
aI81257
asS'd'
I11369
sa(dp22738
S'a'
I47948
sS's'
I17
sS'm'
(lp22739
I1
aI3
aI5
aI5
aI15
aI31
aI41
aI161
aI481
aI231
aI1135
aI3045
aI439
aI11785
aI14863
aI39729
aI59539
asS'd'
I11370
sa(dp22740
S'a'
I47956
sS's'
I17
sS'm'
(lp22741
I1
aI1
aI5
aI7
aI27
aI9
aI81
aI141
aI395
aI105
aI41
aI3719
aI3105
aI13685
aI7451
aI31381
aI82907
asS'd'
I11371
sa(dp22742
S'a'
I47965
sS's'
I17
sS'm'
(lp22743
I1
aI3
aI1
aI13
aI15
aI25
aI127
aI231
aI433
aI837
aI1923
aI1301
aI2479
aI3243
aI21605
aI55789
aI11311
asS'd'
I11372
sa(dp22744
S'a'
I47970
sS's'
I17
sS'm'
(lp22745
I1
aI3
aI5
aI7
aI17
aI5
aI27
aI111
aI217
aI445
aI1245
aI1029
aI7663
aI10291
aI16483
aI37503
aI110205
asS'd'
I11373
sa(dp22746
S'a'
I47972
sS's'
I17
sS'm'
(lp22747
I1
aI1
aI1
aI7
aI31
aI21
aI51
aI235
aI487
aI457
aI1687
aI2947
aI5067
aI13779
aI7671
aI7257
aI119141
asS'd'
I11374
sa(dp22748
S'a'
I47981
sS's'
I17
sS'm'
(lp22749
I1
aI1
aI5
aI13
aI25
aI31
aI59
aI183
aI33
aI887
aI469
aI813
aI7939
aI11775
aI5795
aI26227
aI57703
asS'd'
I11375
sa(dp22750
S'a'
I47987
sS's'
I17
sS'm'
(lp22751
I1
aI3
aI3
aI15
aI17
aI7
aI19
aI87
aI365
aI177
aI1157
aI1023
aI3055
aI15439
aI27187
aI32593
aI112507
asS'd'
I11376
sa(dp22752
S'a'
I48010
sS's'
I17
sS'm'
(lp22753
I1
aI1
aI1
aI9
aI19
aI49
aI103
aI239
aI207
aI409
aI1613
aI2793
aI5477
aI11221
aI21611
aI19963
aI99333
asS'd'
I11377
sa(dp22754
S'a'
I48012
sS's'
I17
sS'm'
(lp22755
I1
aI3
aI3
aI11
aI17
aI1
aI103
aI77
aI93
aI197
aI1883
aI4053
aI4613
aI11571
aI1841
aI23189
aI4235
asS'd'
I11378
sa(dp22756
S'a'
I48023
sS's'
I17
sS'm'
(lp22757
I1
aI3
aI5
aI9
aI9
aI49
aI29
aI231
aI167
aI593
aI1909
aI2457
aI323
aI10549
aI6551
aI45597
aI96591
asS'd'
I11379
sa(dp22758
S'a'
I48033
sS's'
I17
sS'm'
(lp22759
I1
aI1
aI7
aI11
aI15
aI39
aI33
aI61
aI13
aI879
aI1589
aI2169
aI125
aI2427
aI13029
aI24919
aI14147
asS'd'
I11380
sa(dp22760
S'a'
I48051
sS's'
I17
sS'm'
(lp22761
I1
aI3
aI5
aI5
aI25
aI53
aI49
aI255
aI263
aI917
aI1997
aI171
aI2945
aI14243
aI12983
aI21821
aI119547
asS'd'
I11381
sa(dp22762
S'a'
I48071
sS's'
I17
sS'm'
(lp22763
I1
aI1
aI1
aI15
aI5
aI25
aI113
aI235
aI311
aI377
aI1059
aI1365
aI3457
aI8699
aI18617
aI25119
aI110659
asS'd'
I11382
sa(dp22764
S'a'
I48075
sS's'
I17
sS'm'
(lp22765
I1
aI1
aI5
aI1
aI9
aI61
aI1
aI137
aI301
aI85
aI1527
aI3831
aI923
aI13753
aI20909
aI14007
aI22939
asS'd'
I11383
sa(dp22766
S'a'
I48077
sS's'
I17
sS'm'
(lp22767
I1
aI1
aI7
aI9
aI23
aI17
aI31
aI107
aI55
aI293
aI425
aI3513
aI3503
aI10075
aI6299
aI40007
aI54355
asS'd'
I11384
sa(dp22768
S'a'
I48080
sS's'
I17
sS'm'
(lp22769
I1
aI3
aI5
aI5
aI27
aI39
aI95
aI187
aI239
aI949
aI531
aI3541
aI99
aI6339
aI32295
aI10377
aI111287
asS'd'
I11385
sa(dp22770
S'a'
I48099
sS's'
I17
sS'm'
(lp22771
I1
aI1
aI5
aI15
aI19
aI39
aI23
aI89
aI185
aI47
aI87
aI1721
aI2471
aI13221
aI13201
aI31
aI12897
asS'd'
I11386
sa(dp22772
S'a'
I48106
sS's'
I17
sS'm'
(lp22773
I1
aI3
aI1
aI9
aI7
aI7
aI109
aI241
aI39
aI687
aI175
aI139
aI583
aI1629
aI19775
aI6371
aI121879
asS'd'
I11387
sa(dp22774
S'a'
I48114
sS's'
I17
sS'm'
(lp22775
I1
aI1
aI1
aI13
aI13
aI9
aI73
aI153
aI367
aI425
aI217
aI3981
aI5203
aI1111
aI12333
aI59799
aI105259
asS'd'
I11388
sa(dp22776
S'a'
I48120
sS's'
I17
sS'm'
(lp22777
I1
aI1
aI3
aI5
aI7
aI33
aI27
aI123
aI335
aI727
aI1619
aI3999
aI7799
aI2807
aI2251
aI45835
aI113193
asS'd'
I11389
sa(dp22778
S'a'
I48128
sS's'
I17
sS'm'
(lp22779
I1
aI3
aI1
aI7
aI25
aI3
aI49
aI163
aI447
aI555
aI647
aI1461
aI5881
aI15755
aI32233
aI48915
aI96203
asS'd'
I11390
sa(dp22780
S'a'
I48145
sS's'
I17
sS'm'
(lp22781
I1
aI3
aI7
aI7
aI21
aI11
aI61
aI101
aI323
aI165
aI1489
aI2933
aI3363
aI8471
aI4311
aI65279
aI123813
asS'd'
I11391
sa(dp22782
S'a'
I48152
sS's'
I17
sS'm'
(lp22783
I1
aI3
aI1
aI1
aI15
aI1
aI101
aI225
aI473
aI479
aI1237
aI113
aI7591
aI2883
aI3891
aI53703
aI14607
asS'd'
I11392
sa(dp22784
S'a'
I48155
sS's'
I17
sS'm'
(lp22785
I1
aI3
aI5
aI3
aI17
aI51
aI23
aI31
aI487
aI957
aI1623
aI2329
aI2801
aI6213
aI7523
aI14131
aI23893
asS'd'
I11393
sa(dp22786
S'a'
I48182
sS's'
I17
sS'm'
(lp22787
I1
aI1
aI5
aI7
aI9
aI39
aI93
aI105
aI13
aI481
aI53
aI3785
aI5621
aI11889
aI993
aI23611
aI73651
asS'd'
I11394
sa(dp22788
S'a'
I48188
sS's'
I17
sS'm'
(lp22789
I1
aI3
aI7
aI15
aI21
aI57
aI97
aI69
aI339
aI289
aI1805
aI2661
aI1165
aI6079
aI1127
aI51285
aI54453
asS'd'
I11395
sa(dp22790
S'a'
I48203
sS's'
I17
sS'm'
(lp22791
I1
aI1
aI3
aI3
aI3
aI3
aI1
aI179
aI123
aI1011
aI1363
aI231
aI6983
aI9499
aI91
aI52573
aI32565
asS'd'
I11396
sa(dp22792
S'a'
I48217
sS's'
I17
sS'm'
(lp22793
I1
aI3
aI3
aI15
aI5
aI37
aI97
aI7
aI189
aI547
aI1965
aI3821
aI4907
aI8181
aI12857
aI7907
aI19361
asS'd'
I11397
sa(dp22794
S'a'
I48218
sS's'
I17
sS'm'
(lp22795
I1
aI1
aI5
aI3
aI29
aI33
aI45
aI99
aI359
aI337
aI1377
aI577
aI3117
aI9545
aI30093
aI26147
aI128509
asS'd'
I11398
sa(dp22796
S'a'
I48236
sS's'
I17
sS'm'
(lp22797
I1
aI3
aI5
aI11
aI7
aI49
aI99
aI249
aI37
aI755
aI383
aI2845
aI4153
aI695
aI11099
aI33653
aI105155
asS'd'
I11399
sa(dp22798
S'a'
I48241
sS's'
I17
sS'm'
(lp22799
I1
aI3
aI1
aI1
aI19
aI3
aI3
aI177
aI97
aI323
aI1367
aI213
aI4391
aI11223
aI26497
aI12289
aI39047
asS'd'
I11400
sa(dp22800
S'a'
I48244
sS's'
I17
sS'm'
(lp22801
I1
aI1
aI3
aI7
aI3
aI33
aI73
aI199
aI339
aI479
aI1797
aI3905
aI2849
aI5667
aI18015
aI36653
aI83491
asS'd'
I11401
sa(dp22802
S'a'
I48264
sS's'
I17
sS'm'
(lp22803
I1
aI1
aI5
aI3
aI27
aI59
aI27
aI241
aI49
aI161
aI451
aI3993
aI2489
aI10681
aI11895
aI60405
aI47021
asS'd'
I11402
sa(dp22804
S'a'
I48269
sS's'
I17
sS'm'
(lp22805
I1
aI1
aI7
aI5
aI9
aI23
aI115
aI131
aI383
aI895
aI1591
aI585
aI2571
aI7485
aI31535
aI12871
aI95717
asS'd'
I11403
sa(dp22806
S'a'
I48278
sS's'
I17
sS'm'
(lp22807
I1
aI1
aI1
aI15
aI13
aI49
aI3
aI157
aI311
aI159
aI1239
aI159
aI5643
aI6405
aI11763
aI34609
aI75259
asS'd'
I11404
sa(dp22808
S'a'
I48282
sS's'
I17
sS'm'
(lp22809
I1
aI1
aI1
aI13
aI11
aI41
aI119
aI3
aI165
aI943
aI2035
aI179
aI357
aI14591
aI20099
aI4787
aI12659
asS'd'
I11405
sa(dp22810
S'a'
I48288
sS's'
I17
sS'm'
(lp22811
I1
aI1
aI5
aI11
aI13
aI9
aI67
aI139
aI259
aI947
aI1559
aI283
aI1557
aI11297
aI29753
aI21953
aI100317
asS'd'
I11406
sa(dp22812
S'a'
I48311
sS's'
I17
sS'm'
(lp22813
I1
aI1
aI7
aI11
aI13
aI39
aI105
aI83
aI109
aI723
aI1643
aI3599
aI1471
aI13653
aI4583
aI18595
aI91935
asS'd'
I11407
sa(dp22814
S'a'
I48315
sS's'
I17
sS'm'
(lp22815
I1
aI1
aI5
aI9
aI25
aI5
aI69
aI189
aI209
aI779
aI1421
aI467
aI849
aI12887
aI10317
aI3005
aI100813
asS'd'
I11408
sa(dp22816
S'a'
I48317
sS's'
I17
sS'm'
(lp22817
I1
aI3
aI7
aI11
aI19
aI1
aI93
aI239
aI333
aI713
aI1525
aI813
aI5913
aI10811
aI7077
aI20573
aI86999
asS'd'
I11409
sa(dp22818
S'a'
I48329
sS's'
I17
sS'm'
(lp22819
I1
aI3
aI1
aI5
aI25
aI1
aI107
aI153
aI225
aI889
aI1319
aI497
aI2193
aI11511
aI17553
aI23733
aI83179
asS'd'
I11410
sa(dp22820
S'a'
I48330
sS's'
I17
sS'm'
(lp22821
I1
aI3
aI3
aI3
aI5
aI45
aI51
aI37
aI443
aI851
aI263
aI4067
aI2629
aI10887
aI29081
aI30489
aI1161
asS'd'
I11411
sa(dp22822
S'a'
I48337
sS's'
I17
sS'm'
(lp22823
I1
aI1
aI3
aI9
aI25
aI19
aI115
aI31
aI347
aI641
aI1103
aI1121
aI7051
aI14071
aI16663
aI48517
aI24655
asS'd'
I11412
sa(dp22824
S'a'
I48350
sS's'
I17
sS'm'
(lp22825
I1
aI1
aI3
aI15
aI11
aI33
aI123
aI181
aI225
aI437
aI875
aI1997
aI209
aI5487
aI1989
aI6745
aI38423
asS'd'
I11413
sa(dp22826
S'a'
I48353
sS's'
I17
sS'm'
(lp22827
I1
aI3
aI7
aI5
aI27
aI5
aI105
aI203
aI51
aI683
aI1523
aI347
aI6881
aI4353
aI4531
aI29589
aI108053
asS'd'
I11414
sa(dp22828
S'a'
I48360
sS's'
I17
sS'm'
(lp22829
I1
aI1
aI3
aI9
aI15
aI33
aI127
aI145
aI125
aI429
aI915
aI1307
aI1495
aI3553
aI26797
aI36819
aI121375
asS'd'
I11415
sa(dp22830
S'a'
I48374
sS's'
I17
sS'm'
(lp22831
I1
aI3
aI5
aI15
aI7
aI1
aI55
aI99
aI7
aI405
aI885
aI59
aI5359
aI13969
aI32037
aI53399
aI77293
asS'd'
I11416
sa(dp22832
S'a'
I48380
sS's'
I17
sS'm'
(lp22833
I1
aI1
aI7
aI1
aI29
aI31
aI89
aI69
aI405
aI837
aI1949
aI2337
aI1139
aI14129
aI16867
aI15167
aI21117
asS'd'
I11417
sa(dp22834
S'a'
I48405
sS's'
I17
sS'm'
(lp22835
I1
aI1
aI3
aI5
aI3
aI25
aI61
aI175
aI135
aI109
aI959
aI501
aI13
aI16057
aI25031
aI16321
aI27617
asS'd'
I11418
sa(dp22836
S'a'
I48406
sS's'
I17
sS'm'
(lp22837
I1
aI3
aI1
aI5
aI17
aI39
aI59
aI113
aI25
aI707
aI1641
aI3489
aI6193
aI10085
aI13837
aI57851
aI77475
asS'd'
I11419
sa(dp22838
S'a'
I48409
sS's'
I17
sS'm'
(lp22839
I1
aI1
aI3
aI5
aI17
aI35
aI111
aI163
aI253
aI641
aI715
aI1101
aI4411
aI10771
aI20241
aI46415
aI114719
asS'd'
I11420
sa(dp22840
S'a'
I48412
sS's'
I17
sS'm'
(lp22841
I1
aI3
aI1
aI1
aI7
aI11
aI77
aI59
aI71
aI869
aI1505
aI751
aI4367
aI9603
aI29735
aI7333
aI62487
asS'd'
I11421
sa(dp22842
S'a'
I48416
sS's'
I17
sS'm'
(lp22843
I1
aI3
aI7
aI9
aI3
aI51
aI93
aI195
aI497
aI659
aI1563
aI3801
aI6933
aI9089
aI10891
aI9853
aI93611
asS'd'
I11422
sa(dp22844
S'a'
I48419
sS's'
I17
sS'm'
(lp22845
I1
aI3
aI1
aI7
aI21
aI47
aI61
aI117
aI293
aI281
aI1547
aI3437
aI4947
aI12119
aI21425
aI5591
aI23951
asS'd'
I11423
sa(dp22846
S'a'
I48425
sS's'
I17
sS'm'
(lp22847
I1
aI1
aI7
aI13
aI19
aI59
aI121
aI249
aI147
aI759
aI1963
aI465
aI7785
aI7015
aI81
aI47869
aI123845
asS'd'
I11424
sa(dp22848
S'a'
I48434
sS's'
I17
sS'm'
(lp22849
I1
aI1
aI7
aI13
aI5
aI35
aI45
aI69
aI167
aI263
aI979
aI2855
aI1845
aI5531
aI17167
aI7363
aI89233
asS'd'
I11425
sa(dp22850
S'a'
I48443
sS's'
I17
sS'm'
(lp22851
I1
aI3
aI3
aI3
aI17
aI43
aI109
aI193
aI389
aI867
aI1403
aI1271
aI7127
aI13977
aI22547
aI53997
aI24475
asS'd'
I11426
sa(dp22852
S'a'
I48445
sS's'
I17
sS'm'
(lp22853
I1
aI1
aI1
aI5
aI25
aI43
aI47
aI163
aI417
aI967
aI819
aI2433
aI439
aI8499
aI29705
aI32697
aI109963
asS'd'
I11427
sa(dp22854
S'a'
I48453
sS's'
I17
sS'm'
(lp22855
I1
aI1
aI1
aI15
aI19
aI5
aI83
aI237
aI431
aI697
aI1383
aI2499
aI6907
aI9033
aI12147
aI23479
aI17649
asS'd'
I11428
sa(dp22856
S'a'
I48454
sS's'
I17
sS'm'
(lp22857
I1
aI3
aI5
aI13
aI27
aI35
aI103
aI95
aI239
aI903
aI537
aI601
aI417
aI15859
aI16533
aI16753
aI128229
asS'd'
I11429
sa(dp22858
S'a'
I48482
sS's'
I17
sS'm'
(lp22859
I1
aI1
aI3
aI1
aI11
aI55
aI57
aI9
aI229
aI135
aI805
aI2745
aI2023
aI12645
aI29135
aI39051
aI17065
asS'd'
I11430
sa(dp22860
S'a'
I48491
sS's'
I17
sS'm'
(lp22861
I1
aI3
aI5
aI3
aI13
aI25
aI77
aI29
aI449
aI31
aI1733
aI1451
aI3895
aI11551
aI1011
aI22817
aI35959
asS'd'
I11431
sa(dp22862
S'a'
I48501
sS's'
I17
sS'm'
(lp22863
I1
aI3
aI3
aI3
aI1
aI31
aI93
aI47
aI255
aI393
aI571
aI443
aI6079
aI6245
aI11773
aI42087
aI40651
asS'd'
I11432
sa(dp22864
S'a'
I48517
sS's'
I17
sS'm'
(lp22865
I1
aI3
aI1
aI5
aI17
aI49
aI119
aI219
aI375
aI337
aI29
aI3409
aI3187
aI15243
aI25853
aI44385
aI103675
asS'd'
I11433
sa(dp22866
S'a'
I48518
sS's'
I17
sS'm'
(lp22867
I1
aI3
aI7
aI15
aI15
aI21
aI79
aI109
aI81
aI119
aI603
aI1665
aI7813
aI3485
aI11801
aI48693
aI108307
asS'd'
I11434
sa(dp22868
S'a'
I48545
sS's'
I17
sS'm'
(lp22869
I1
aI1
aI1
aI15
aI9
aI5
aI103
aI141
aI181
aI841
aI7
aI1217
aI7713
aI4843
aI9089
aI53641
aI3029
asS'd'
I11435
sa(dp22870
S'a'
I48548
sS's'
I17
sS'm'
(lp22871
I1
aI3
aI3
aI1
aI29
aI47
aI13
aI179
aI439
aI387
aI1553
aI3141
aI947
aI4893
aI29119
aI30865
aI14207
asS'd'
I11436
sa(dp22872
S'a'
I48552
sS's'
I17
sS'm'
(lp22873
I1
aI1
aI1
aI15
aI29
aI1
aI41
aI135
aI43
aI673
aI1527
aI883
aI3211
aI5195
aI30219
aI47133
aI56819
asS'd'
I11437
sa(dp22874
S'a'
I48569
sS's'
I17
sS'm'
(lp22875
I1
aI3
aI5
aI13
aI13
aI47
aI97
aI219
aI277
aI397
aI1901
aI821
aI1961
aI705
aI7291
aI19435
aI123563
asS'd'
I11438
sa(dp22876
S'a'
I48575
sS's'
I17
sS'm'
(lp22877
I1
aI1
aI7
aI15
aI5
aI49
aI93
aI59
aI221
aI205
aI115
aI559
aI5633
aI5819
aI6923
aI18301
aI72639
asS'd'
I11439
sa(dp22878
S'a'
I48580
sS's'
I17
sS'm'
(lp22879
I1
aI3
aI5
aI13
aI29
aI49
aI25
aI203
aI125
aI385
aI487
aI1897
aI2177
aI6859
aI6105
aI763
aI36673
asS'd'
I11440
sa(dp22880
S'a'
I48584
sS's'
I17
sS'm'
(lp22881
I1
aI3
aI3
aI9
aI23
aI63
aI9
aI115
aI489
aI89
aI1113
aI1351
aI8181
aI2569
aI18263
aI32619
aI24795
asS'd'
I11441
sa(dp22882
S'a'
I48590
sS's'
I17
sS'm'
(lp22883
I1
aI1
aI5
aI11
aI25
aI51
aI97
aI155
aI15
aI139
aI1275
aI3479
aI5851
aI3099
aI7417
aI57155
aI90185
asS'd'
I11442
sa(dp22884
S'a'
I48595
sS's'
I17
sS'm'
(lp22885
I1
aI3
aI1
aI11
aI1
aI37
aI67
aI221
aI493
aI475
aI1881
aI2277
aI4365
aI9411
aI16629
aI65229
aI28803
asS'd'
I11443
sa(dp22886
S'a'
I48602
sS's'
I17
sS'm'
(lp22887
I1
aI3
aI1
aI1
aI15
aI53
aI39
aI141
aI453
aI151
aI335
aI795
aI1515
aI2719
aI24197
aI21153
aI129549
asS'd'
I11444
sa(dp22888
S'a'
I48604
sS's'
I17
sS'm'
(lp22889
I1
aI1
aI1
aI7
aI19
aI31
aI59
aI123
aI73
aI149
aI469
aI1199
aI3603
aI1539
aI29157
aI50031
aI22109
asS'd'
I11445
sa(dp22890
S'a'
I48618
sS's'
I17
sS'm'
(lp22891
I1
aI1
aI3
aI15
aI23
aI51
aI45
aI211
aI423
aI553
aI1289
aI1125
aI6347
aI6711
aI23761
aI14109
aI17261
asS'd'
I11446
sa(dp22892
S'a'
I48625
sS's'
I17
sS'm'
(lp22893
I1
aI3
aI5
aI13
aI5
aI7
aI35
aI185
aI263
aI791
aI161
aI325
aI4771
aI11913
aI31595
aI56675
aI68615
asS'd'
I11447
sa(dp22894
S'a'
I48626
sS's'
I17
sS'm'
(lp22895
I1
aI1
aI5
aI9
aI19
aI27
aI89
aI147
aI55
aI197
aI1695
aI99
aI755
aI15115
aI1933
aI41439
aI85063
asS'd'
I11448
sa(dp22896
S'a'
I48638
sS's'
I17
sS'm'
(lp22897
I1
aI1
aI7
aI7
aI17
aI11
aI103
aI55
aI281
aI707
aI1973
aI2055
aI5015
aI5713
aI3717
aI44149
aI8033
asS'd'
I11449
sa(dp22898
S'a'
I48668
sS's'
I17
sS'm'
(lp22899
I1
aI3
aI1
aI5
aI21
aI49
aI55
aI93
aI161
aI565
aI81
aI3241
aI3709
aI8185
aI16935
aI60369
aI118127
asS'd'
I11450
sa(dp22900
S'a'
I48677
sS's'
I17
sS'm'
(lp22901
I1
aI3
aI7
aI9
aI27
aI1
aI105
aI133
aI397
aI351
aI1021
aI739
aI161
aI9971
aI24733
aI29239
aI68853
asS'd'
I11451
sa(dp22902
S'a'
I48678
sS's'
I17
sS'm'
(lp22903
I1
aI1
aI7
aI1
aI1
aI19
aI97
aI243
aI73
aI969
aI313
aI399
aI2955
aI2467
aI18265
aI60637
aI35457
asS'd'
I11452
sa(dp22904
S'a'
I48687
sS's'
I17
sS'm'
(lp22905
I1
aI3
aI1
aI11
aI29
aI25
aI35
aI35
aI469
aI143
aI1195
aI911
aI1023
aI14685
aI10933
aI16449
aI102113
asS'd'
I11453
sa(dp22906
S'a'
I48692
sS's'
I17
sS'm'
(lp22907
I1
aI1
aI5
aI3
aI31
aI21
aI65
aI181
aI13
aI235
aI501
aI3621
aI3567
aI8771
aI32747
aI59231
aI91551
asS'd'
I11454
sa(dp22908
S'a'
I48696
sS's'
I17
sS'm'
(lp22909
I1
aI1
aI1
aI13
aI29
aI21
aI13
aI33
aI193
aI565
aI203
aI3927
aI4749
aI9897
aI26109
aI14221
aI27733
asS'd'
I11455
sa(dp22910
S'a'
I48704
sS's'
I17
sS'm'
(lp22911
I1
aI1
aI1
aI5
aI29
aI41
aI7
aI125
aI391
aI295
aI533
aI2135
aI3107
aI7711
aI27811
aI55767
aI78821
asS'd'
I11456
sa(dp22912
S'a'
I48707
sS's'
I17
sS'm'
(lp22913
I1
aI1
aI7
aI3
aI19
aI25
aI117
aI195
aI155
aI685
aI147
aI2049
aI3751
aI4585
aI24893
aI36895
aI80371
asS'd'
I11457
sa(dp22914
S'a'
I48714
sS's'
I17
sS'm'
(lp22915
I1
aI3
aI5
aI5
aI23
aI41
aI9
aI125
aI315
aI809
aI1019
aI1453
aI4605
aI13753
aI30641
aI50677
aI94781
asS'd'
I11458
sa(dp22916
S'a'
I48731
sS's'
I17
sS'm'
(lp22917
I1
aI1
aI1
aI7
aI27
aI45
aI103
aI199
aI37
aI291
aI651
aI185
aI6715
aI15387
aI30873
aI63051
aI123231
asS'd'
I11459
sa(dp22918
S'a'
I48740
sS's'
I17
sS'm'
(lp22919
I1
aI1
aI5
aI7
aI17
aI5
aI75
aI129
aI401
aI107
aI1681
aI2039
aI299
aI12399
aI30947
aI26327
aI91589
asS'd'
I11460
sa(dp22920
S'a'
I48750
sS's'
I17
sS'm'
(lp22921
I1
aI3
aI3
aI15
aI17
aI19
aI109
aI19
aI493
aI797
aI209
aI929
aI2821
aI395
aI22173
aI27803
aI87953
asS'd'
I11461
sa(dp22922
S'a'
I48761
sS's'
I17
sS'm'
(lp22923
I1
aI3
aI7
aI5
aI7
aI5
aI71
aI159
aI483
aI389
aI1817
aI4093
aI963
aI4253
aI31267
aI63919
aI62113
asS'd'
I11462
sa(dp22924
S'a'
I48785
sS's'
I17
sS'm'
(lp22925
I1
aI3
aI3
aI15
aI9
aI49
aI89
aI49
aI15
aI61
aI2041
aI2357
aI2173
aI3349
aI32565
aI23207
aI21177
asS'd'
I11463
sa(dp22926
S'a'
I48786
sS's'
I17
sS'm'
(lp22927
I1
aI1
aI1
aI13
aI15
aI21
aI31
aI143
aI387
aI371
aI567
aI1903
aI3793
aI9559
aI7055
aI31251
aI13663
asS'd'
I11464
sa(dp22928
S'a'
I48798
sS's'
I17
sS'm'
(lp22929
I1
aI1
aI3
aI13
aI3
aI7
aI49
aI31
aI255
aI581
aI627
aI1947
aI2965
aI2787
aI8275
aI59785
aI19979
asS'd'
I11465
sa(dp22930
S'a'
I48807
sS's'
I17
sS'm'
(lp22931
I1
aI3
aI3
aI13
aI29
aI47
aI53
aI133
aI301
aI253
aI1215
aI3409
aI5745
aI247
aI31585
aI5555
aI31011
asS'd'
I11466
sa(dp22932
S'a'
I48808
sS's'
I17
sS'm'
(lp22933
I1
aI3
aI7
aI15
aI23
aI49
aI1
aI89
aI141
aI423
aI707
aI645
aI7955
aI10485
aI27223
aI35867
aI45461
asS'd'
I11467
sa(dp22934
S'a'
I48814
sS's'
I17
sS'm'
(lp22935
I1
aI1
aI3
aI3
aI29
aI25
aI123
aI63
aI197
aI429
aI169
aI3229
aI3797
aI4029
aI29947
aI52781
aI16065
asS'd'
I11468
sa(dp22936
S'a'
I48826
sS's'
I17
sS'm'
(lp22937
I1
aI3
aI3
aI7
aI29
aI17
aI83
aI95
aI199
aI253
aI133
aI265
aI6723
aI6207
aI12863
aI61311
aI21937
asS'd'
I11469
sa(dp22938
S'a'
I48828
sS's'
I17
sS'm'
(lp22939
I1
aI3
aI7
aI13
aI19
aI61
aI89
aI151
aI249
aI597
aI1389
aI717
aI5111
aI3285
aI6251
aI50237
aI108703
asS'd'
I11470
sa(dp22940
S'a'
I48834
sS's'
I17
sS'm'
(lp22941
I1
aI3
aI5
aI11
aI1
aI23
aI35
aI61
aI143
aI45
aI625
aI1693
aI4943
aI2213
aI9317
aI7601
aI28359
asS'd'
I11471
sa(dp22942
S'a'
I48840
sS's'
I17
sS'm'
(lp22943
I1
aI3
aI3
aI15
aI29
aI43
aI115
aI23
aI167
aI355
aI977
aI439
aI4767
aI9967
aI22997
aI54725
aI125637
asS'd'
I11472
sa(dp22944
S'a'
I48845
sS's'
I17
sS'm'
(lp22945
I1
aI3
aI7
aI1
aI29
aI45
aI83
aI71
aI395
aI247
aI1
aI113
aI6393
aI12445
aI16137
aI35125
aI102053
asS'd'
I11473
sa(dp22946
S'a'
I48858
sS's'
I17
sS'm'
(lp22947
I1
aI1
aI1
aI7
aI17
aI55
aI17
aI159
aI33
aI237
aI207
aI1297
aI5611
aI6023
aI17709
aI60905
aI3533
asS'd'
I11474
sa(dp22948
S'a'
I48864
sS's'
I17
sS'm'
(lp22949
I1
aI3
aI1
aI13
aI27
aI57
aI27
aI235
aI141
aI917
aI1655
aI659
aI939
aI559
aI2651
aI705
aI80141
asS'd'
I11475
sa(dp22950
S'a'
I48869
sS's'
I17
sS'm'
(lp22951
I1
aI3
aI7
aI3
aI3
aI17
aI111
aI117
aI467
aI129
aI1105
aI3457
aI2093
aI8513
aI19941
aI22111
aI54597
asS'd'
I11476
sa(dp22952
S'a'
I48870
sS's'
I17
sS'm'
(lp22953
I1
aI3
aI7
aI3
aI5
aI17
aI59
aI195
aI23
aI547
aI1799
aI1427
aI391
aI4043
aI10407
aI31055
aI38495
asS'd'
I11477
sa(dp22954
S'a'
I48879
sS's'
I17
sS'm'
(lp22955
I1
aI1
aI1
aI7
aI21
aI9
aI71
aI33
aI209
aI773
aI1243
aI3239
aI3763
aI15229
aI9609
aI24395
aI56145
asS'd'
I11478
sa(dp22956
S'a'
I48887
sS's'
I17
sS'm'
(lp22957
I1
aI3
aI5
aI13
aI13
aI45
aI71
aI91
aI23
aI553
aI665
aI1753
aI5173
aI4355
aI14317
aI42517
aI32307
asS'd'
I11479
sa(dp22958
S'a'
I48899
sS's'
I17
sS'm'
(lp22959
I1
aI3
aI5
aI11
aI31
aI3
aI37
aI95
aI63
aI425
aI1611
aI2983
aI5075
aI1375
aI14305
aI11099
aI101203
asS'd'
I11480
sa(dp22960
S'a'
I48911
sS's'
I17
sS'm'
(lp22961
I1
aI1
aI3
aI1
aI21
aI7
aI15
aI141
aI389
aI871
aI617
aI271
aI1037
aI13569
aI13019
aI58899
aI54097
asS'd'
I11481
sa(dp22962
S'a'
I48920
sS's'
I17
sS'm'
(lp22963
I1
aI3
aI7
aI1
aI13
aI25
aI21
aI251
aI467
aI373
aI1539
aI4065
aI1871
aI791
aI26315
aI64187
aI119455
asS'd'
I11482
sa(dp22964
S'a'
I48926
sS's'
I17
sS'm'
(lp22965
I1
aI3
aI1
aI3
aI25
aI37
aI43
aI9
aI187
aI323
aI409
aI443
aI2861
aI14145
aI26185
aI24049
aI109613
asS'd'
I11483
sa(dp22966
S'a'
I48929
sS's'
I17
sS'm'
(lp22967
I1
aI1
aI3
aI7
aI21
aI61
aI3
aI81
aI445
aI673
aI1269
aI613
aI1279
aI8209
aI22911
aI48017
aI54181
asS'd'
I11484
sa(dp22968
S'a'
I48971
sS's'
I17
sS'm'
(lp22969
I1
aI1
aI1
aI15
aI25
aI63
aI71
aI147
aI217
aI491
aI183
aI977
aI4967
aI3471
aI8791
aI11843
aI68005
asS'd'
I11485
sa(dp22970
S'a'
I48974
sS's'
I17
sS'm'
(lp22971
I1
aI1
aI1
aI5
aI25
aI43
aI13
aI237
aI57
aI919
aI1641
aI399
aI4269
aI7357
aI3465
aI63901
aI61329
asS'd'
I11486
sa(dp22972
S'a'
I48981
sS's'
I17
sS'm'
(lp22973
I1
aI1
aI7
aI15
aI27
aI57
aI47
aI187
aI295
aI117
aI1223
aI2963
aI4995
aI13279
aI25107
aI56089
aI37293
asS'd'
I11487
sa(dp22974
S'a'
I48998
sS's'
I17
sS'm'
(lp22975
I1
aI1
aI5
aI13
aI29
aI19
aI83
aI121
aI129
aI931
aI1933
aI1141
aI3125
aI3321
aI22019
aI52729
aI119643
asS'd'
I11488
sa(dp22976
S'a'
I49007
sS's'
I17
sS'm'
(lp22977
I1
aI1
aI1
aI1
aI21
aI19
aI49
aI241
aI227
aI57
aI1919
aI113
aI6993
aI4687
aI1043
aI5247
aI15565
asS'd'
I11489
sa(dp22978
S'a'
I49028
sS's'
I17
sS'm'
(lp22979
I1
aI3
aI3
aI15
aI5
aI21
aI65
aI129
aI485
aI173
aI1663
aI2419
aI4279
aI4167
aI25827
aI28457
aI68219
asS'd'
I11490
sa(dp22980
S'a'
I49031
sS's'
I17
sS'm'
(lp22981
I1
aI1
aI3
aI1
aI17
aI9
aI65
aI21
aI487
aI875
aI1111
aI1679
aI6451
aI14825
aI23931
aI16053
aI79687
asS'd'
I11491
sa(dp22982
S'a'
I49056
sS's'
I17
sS'm'
(lp22983
I1
aI3
aI3
aI5
aI13
aI5
aI49
aI15
aI267
aI389
aI1111
aI1505
aI5815
aI6285
aI26075
aI167
aI70325
asS'd'
I11492
sa(dp22984
S'a'
I49065
sS's'
I17
sS'm'
(lp22985
I1
aI1
aI3
aI11
aI5
aI15
aI57
aI171
aI407
aI497
aI1979
aI2819
aI1267
aI6893
aI6601
aI30971
aI24477
asS'd'
I11493
sa(dp22986
S'a'
I49079
sS's'
I17
sS'm'
(lp22987
I1
aI1
aI1
aI15
aI7
aI5
aI69
aI181
aI195
aI847
aI1245
aI4019
aI2469
aI1407
aI17013
aI43437
aI16307
asS'd'
I11494
sa(dp22988
S'a'
I49083
sS's'
I17
sS'm'
(lp22989
I1
aI3
aI1
aI11
aI15
aI17
aI115
aI197
aI215
aI35
aI1489
aI659
aI4725
aI11339
aI30259
aI52539
aI13365
asS'd'
I11495
sa(dp22990
S'a'
I49088
sS's'
I17
sS'm'
(lp22991
I1
aI1
aI1
aI13
aI23
aI43
aI21
aI33
aI17
aI969
aI1321
aI2469
aI4371
aI7685
aI6817
aI20179
aI113483
asS'd'
I11496
sa(dp22992
S'a'
I49093
sS's'
I17
sS'm'
(lp22993
I1
aI3
aI7
aI3
aI31
aI11
aI79
aI55
aI123
aI263
aI1061
aI2087
aI183
aI11623
aI20703
aI60291
aI115261
asS'd'
I11497
sa(dp22994
S'a'
I49100
sS's'
I17
sS'm'
(lp22995
I1
aI1
aI5
aI13
aI5
aI53
aI21
aI71
aI399
aI165
aI1997
aI2667
aI7361
aI8863
aI27859
aI17235
aI77623
asS'd'
I11498
sa(dp22996
S'a'
I49103
sS's'
I17
sS'm'
(lp22997
I1
aI3
aI3
aI15
aI21
aI55
aI27
aI31
aI371
aI289
aI253
aI1453
aI105
aI7035
aI14787
aI2281
aI128359
asS'd'
I11499
sa(dp22998
S'a'
I49124
sS's'
I17
sS'm'
(lp22999
I1
aI1
aI3
aI15
aI23
aI29
aI3
aI143
aI47
aI255
aI115
aI2741
aI6773
aI16267
aI5975
aI3689
aI97497
asS'd'
I11500
sa(dp23000
S'a'
I49136
sS's'
I17
sS'm'
(lp23001
I1
aI1
aI1
aI15
aI9
aI57
aI109
aI43
aI359
aI365
aI1577
aI4091
aI7399
aI10521
aI7983
aI56119
aI65451
asS'd'
I11501
sa(dp23002
S'a'
I49139
sS's'
I17
sS'm'
(lp23003
I1
aI3
aI7
aI7
aI29
aI55
aI115
aI155
aI121
aI679
aI663
aI2345
aI3765
aI4493
aI9555
aI24043
aI41467
asS'd'
I11502
sa(dp23004
S'a'
I49142
sS's'
I17
sS'm'
(lp23005
I1
aI1
aI3
aI1
aI19
aI55
aI67
aI255
aI355
aI701
aI2027
aI3703
aI7839
aI1701
aI7657
aI36429
aI36623
asS'd'
I11503
sa(dp23006
S'a'
I49145
sS's'
I17
sS'm'
(lp23007
I1
aI3
aI5
aI5
aI29
aI7
aI31
aI123
aI21
aI901
aI1581
aI3993
aI5105
aI9715
aI18347
aI27415
aI19253
asS'd'
I11504
sa(dp23008
S'a'
I49146
sS's'
I17
sS'm'
(lp23009
I1
aI3
aI5
aI3
aI3
aI53
aI121
aI105
aI51
aI577
aI157
aI2151
aI5105
aI7855
aI8595
aI24457
aI55931
asS'd'
I11505
sa(dp23010
S'a'
I49171
sS's'
I17
sS'm'
(lp23011
I1
aI3
aI1
aI9
aI23
aI25
aI67
aI115
aI79
aI809
aI1215
aI943
aI1075
aI7103
aI729
aI18791
aI115977
asS'd'
I11506
sa(dp23012
S'a'
I49187
sS's'
I17
sS'm'
(lp23013
I1
aI3
aI1
aI13
aI3
aI57
aI105
aI161
aI277
aI393
aI1671
aI2765
aI5781
aI13429
aI24075
aI10931
aI50951
asS'd'
I11507
sa(dp23014
S'a'
I49190
sS's'
I17
sS'm'
(lp23015
I1
aI3
aI5
aI9
aI19
aI61
aI9
aI227
aI455
aI541
aI721
aI2855
aI949
aI10159
aI13801
aI48199
aI26747
asS'd'
I11508
sa(dp23016
S'a'
I49194
sS's'
I17
sS'm'
(lp23017
I1
aI3
aI1
aI3
aI25
aI61
aI53
aI177
aI441
aI697
aI1845
aI3573
aI6673
aI9691
aI911
aI46387
aI64727
asS'd'
I11509
sa(dp23018
S'a'
I49196
sS's'
I17
sS'm'
(lp23019
I1
aI1
aI1
aI11
aI7
aI13
aI95
aI221
aI455
aI967
aI869
aI883
aI6301
aI5261
aI18401
aI35745
aI114645
asS'd'
I11510
sa(dp23020
S'a'
I49207
sS's'
I17
sS'm'
(lp23021
I1
aI3
aI1
aI11
aI15
aI7
aI115
aI217
aI235
aI539
aI491
aI603
aI2201
aI241
aI25445
aI29773
aI122695
asS'd'
I11511
sa(dp23022
S'a'
I49216
sS's'
I17
sS'm'
(lp23023
I1
aI1
aI5
aI3
aI19
aI9
aI71
aI193
aI131
aI927
aI1931
aI3981
aI7537
aI10811
aI27285
aI45865
aI106171
asS'd'
I11512
sa(dp23024
S'a'
I49219
sS's'
I17
sS'm'
(lp23025
I1
aI1
aI7
aI5
aI5
aI21
aI107
aI77
aI363
aI733
aI1011
aI3259
aI5263
aI15043
aI19153
aI32117
aI129409
asS'd'
I11513
sa(dp23026
S'a'
I49222
sS's'
I17
sS'm'
(lp23027
I1
aI1
aI3
aI13
aI21
aI5
aI9
aI103
aI369
aI699
aI329
aI1065
aI895
aI13211
aI19017
aI5359
aI38335
asS'd'
I11514
sa(dp23028
S'a'
I49240
sS's'
I17
sS'm'
(lp23029
I1
aI1
aI7
aI15
aI21
aI11
aI73
aI153
aI371
aI753
aI805
aI3519
aI5839
aI1809
aI7201
aI8189
aI68361
asS'd'
I11515
sa(dp23030
S'a'
I49250
sS's'
I17
sS'm'
(lp23031
I1
aI3
aI7
aI9
aI27
aI45
aI25
aI175
aI317
aI381
aI961
aI619
aI4827
aI15161
aI19091
aI29369
aI21097
asS'd'
I11516
sa(dp23032
S'a'
I49262
sS's'
I17
sS'm'
(lp23033
I1
aI1
aI5
aI5
aI7
aI7
aI21
aI69
aI23
aI589
aI1413
aI653
aI911
aI13599
aI18349
aI47307
aI64047
asS'd'
I11517
sa(dp23034
S'a'
I49276
sS's'
I17
sS'm'
(lp23035
I1
aI1
aI1
aI13
aI27
aI23
aI87
aI249
aI135
aI727
aI375
aI3641
aI1489
aI13053
aI5151
aI62689
aI101347
asS'd'
I11518
sa(dp23036
S'a'
I49289
sS's'
I17
sS'm'
(lp23037
I1
aI1
aI5
aI11
aI7
aI39
aI1
aI109
aI203
aI961
aI973
aI1181
aI2357
aI5123
aI31481
aI58345
aI52705
asS'd'
I11519
sa(dp23038
S'a'
I49292
sS's'
I17
sS'm'
(lp23039
I1
aI1
aI1
aI15
aI25
aI15
aI85
aI49
aI77
aI235
aI1761
aI2731
aI4157
aI2057
aI27587
aI30299
aI52997
asS'd'
I11520
sa(dp23040
S'a'
I49298
sS's'
I17
sS'm'
(lp23041
I1
aI1
aI5
aI3
aI1
aI13
aI47
aI219
aI51
aI521
aI625
aI3243
aI7093
aI10823
aI9559
aI58191
aI95573
asS'd'
I11521
sa(dp23042
S'a'
I49319
sS's'
I17
sS'm'
(lp23043
I1
aI3
aI5
aI13
aI27
aI63
aI13
aI7
aI167
aI909
aI1559
aI2103
aI1807
aI943
aI28997
aI31015
aI85407
asS'd'
I11522
sa(dp23044
S'a'
I49320
sS's'
I17
sS'm'
(lp23045
I1
aI1
aI3
aI5
aI17
aI21
aI101
aI163
aI477
aI223
aI175
aI3435
aI7071
aI5121
aI28015
aI33365
aI121057
asS'd'
I11523
sa(dp23046
S'a'
I49325
sS's'
I17
sS'm'
(lp23047
I1
aI3
aI7
aI1
aI11
aI35
aI111
aI41
aI261
aI45
aI1009
aI2827
aI4019
aI5029
aI22289
aI20235
aI13481
asS'd'
I11524
sa(dp23048
S'a'
I49333
sS's'
I17
sS'm'
(lp23049
I1
aI3
aI5
aI5
aI15
aI41
aI109
aI7
aI329
aI447
aI65
aI1317
aI6169
aI15947
aI31191
aI47091
aI60643
asS'd'
I11525
sa(dp23050
S'a'
I49340
sS's'
I17
sS'm'
(lp23051
I1
aI3
aI7
aI7
aI21
aI13
aI29
aI113
aI511
aI407
aI1211
aI2065
aI455
aI10049
aI5745
aI48589
aI48669
asS'd'
I11526
sa(dp23052
S'a'
I49343
sS's'
I17
sS'm'
(lp23053
I1
aI1
aI5
aI5
aI21
aI45
aI89
aI19
aI279
aI165
aI1897
aI957
aI8045
aI565
aI4959
aI37173
aI100773
asS'd'
I11527
sa(dp23054
S'a'
I49352
sS's'
I17
sS'm'
(lp23055
I1
aI3
aI1
aI15
aI9
aI17
aI99
aI143
aI489
aI633
aI1721
aI1255
aI3655
aI10083
aI29079
aI17109
aI10035
asS'd'
I11528
sa(dp23056
S'a'
I49363
sS's'
I17
sS'm'
(lp23057
I1
aI1
aI3
aI3
aI3
aI23
aI47
aI9
aI255
aI169
aI1103
aI1799
aI7899
aI7673
aI19259
aI61919
aI112831
asS'd'
I11529
sa(dp23058
S'a'
I49375
sS's'
I17
sS'm'
(lp23059
I1
aI1
aI5
aI5
aI31
aI37
aI83
aI229
aI93
aI575
aI1589
aI2353
aI185
aI7783
aI9413
aI9617
aI123197
asS'd'
I11530
sa(dp23060
S'a'
I49391
sS's'
I17
sS'm'
(lp23061
I1
aI3
aI5
aI7
aI31
aI7
aI113
aI255
aI231
aI309
aI1215
aI737
aI3635
aI14631
aI28737
aI45127
aI111399
asS'd'
I11531
sa(dp23062
S'a'
I49405
sS's'
I17
sS'm'
(lp23063
I1
aI1
aI7
aI1
aI11
aI5
aI55
aI235
aI369
aI983
aI873
aI655
aI6277
aI11425
aI11191
aI38231
aI88941
asS'd'
I11532
sa(dp23064
S'a'
I49413
sS's'
I17
sS'm'
(lp23065
I1
aI3
aI1
aI9
aI3
aI1
aI119
aI93
aI245
aI167
aI853
aI2543
aI203
aI5313
aI14129
aI6283
aI107117
asS'd'
I11533
sa(dp23066
S'a'
I49420
sS's'
I17
sS'm'
(lp23067
I1
aI3
aI3
aI13
aI21
aI33
aI59
aI167
aI435
aI163
aI1873
aI3341
aI2895
aI13205
aI14147
aI19567
aI100127
asS'd'
I11534
sa(dp23068
S'a'
I49425
sS's'
I17
sS'm'
(lp23069
I1
aI1
aI5
aI9
aI7
aI15
aI39
aI81
aI475
aI511
aI1585
aI63
aI6861
aI10055
aI3577
aI48999
aI80979
asS'd'
I11535
sa(dp23070
S'a'
I49431
sS's'
I17
sS'm'
(lp23071
I1
aI1
aI5
aI13
aI21
aI29
aI17
aI3
aI499
aI739
aI1257
aI2925
aI8179
aI13367
aI18879
aI19917
aI109907
asS'd'
I11536
sa(dp23072
S'a'
I49432
sS's'
I17
sS'm'
(lp23073
I1
aI3
aI7
aI13
aI15
aI57
aI109
aI19
aI265
aI579
aI233
aI2507
aI5059
aI14713
aI9549
aI41915
aI56199
asS'd'
I11537
sa(dp23074
S'a'
I49441
sS's'
I17
sS'm'
(lp23075
I1
aI3
aI3
aI5
aI31
aI25
aI85
aI163
aI187
aI795
aI1597
aI1963
aI473
aI4673
aI4555
aI51365
aI73817
asS'd'
I11538
sa(dp23076
S'a'
I49456
sS's'
I17
sS'm'
(lp23077
I1
aI3
aI3
aI13
aI25
aI35
aI71
aI251
aI33
aI971
aI235
aI1919
aI6705
aI14657
aI23417
aI56377
aI21071
asS'd'
I11539
sa(dp23078
S'a'
I49476
sS's'
I17
sS'm'
(lp23079
I1
aI3
aI3
aI11
aI3
aI29
aI85
aI193
aI11
aI831
aI29
aI1233
aI6199
aI11991
aI9205
aI3323
aI23749
asS'd'
I11540
sa(dp23080
S'a'
I49479
sS's'
I17
sS'm'
(lp23081
I1
aI3
aI7
aI7
aI11
aI15
aI1
aI17
aI87
aI665
aI1593
aI2331
aI845
aI7821
aI89
aI7057
aI114975
asS'd'
I11541
sa(dp23082
S'a'
I49504
sS's'
I17
sS'm'
(lp23083
I1
aI1
aI5
aI11
aI9
aI37
aI39
aI79
aI455
aI397
aI1431
aI1541
aI7629
aI15133
aI21395
aI35619
aI123801
asS'd'
I11542
sa(dp23084
S'a'
I49514
sS's'
I17
sS'm'
(lp23085
I1
aI1
aI1
aI7
aI11
aI59
aI67
aI45
aI169
aI869
aI1547
aI2947
aI3025
aI12967
aI29927
aI22181
aI44783
asS'd'
I11543
sa(dp23086
S'a'
I49528
sS's'
I17
sS'm'
(lp23087
I1
aI1
aI5
aI7
aI3
aI57
aI123
aI253
aI369
aI537
aI83
aI1147
aI3469
aI9775
aI14137
aI38899
aI101143
asS'd'
I11544
sa(dp23088
S'a'
I49543
sS's'
I17
sS'm'
(lp23089
I1
aI3
aI5
aI3
aI19
aI35
aI11
aI215
aI343
aI677
aI1873
aI1211
aI3129
aI9017
aI29595
aI1291
aI39397
asS'd'
I11545
sa(dp23090
S'a'
I49544
sS's'
I17
sS'm'
(lp23091
I1
aI1
aI7
aI3
aI25
aI7
aI61
aI229
aI187
aI839
aI747
aI3347
aI4321
aI13201
aI19665
aI56951
aI85273
asS'd'
I11546
sa(dp23092
S'a'
I49550
sS's'
I17
sS'm'
(lp23093
I1
aI1
aI1
aI1
aI11
aI5
aI51
aI41
aI227
aI895
aI553
aI2673
aI6581
aI6583
aI15429
aI33211
aI100599
asS'd'
I11547
sa(dp23094
S'a'
I49558
sS's'
I17
sS'm'
(lp23095
I1
aI3
aI3
aI5
aI21
aI27
aI91
aI65
aI213
aI341
aI723
aI567
aI4761
aI11549
aI15041
aI23079
aI55245
asS'd'
I11548
sa(dp23096
S'a'
I49568
sS's'
I17
sS'm'
(lp23097
I1
aI1
aI3
aI1
aI15
aI43
aI83
aI57
aI473
aI453
aI1351
aI2055
aI5769
aI3887
aI29481
aI57915
aI14017
asS'd'
I11549
sa(dp23098
S'a'
I49578
sS's'
I17
sS'm'
(lp23099
I1
aI3
aI3
aI3
aI21
aI29
aI121
aI137
aI99
aI527
aI711
aI1169
aI7869
aI6245
aI25423
aI38989
aI87019
asS'd'
I11550
sa(dp23100
S'a'
I49600
sS's'
I17
sS'm'
(lp23101
I1
aI1
aI5
aI11
aI9
aI61
aI125
aI7
aI207
aI245
aI1019
aI635
aI7099
aI13133
aI11809
aI56705
aI18801
asS'd'
I11551
sa(dp23102
S'a'
I49612
sS's'
I17
sS'm'
(lp23103
I1
aI3
aI7
aI9
aI15
aI31
aI37
aI205
aI439
aI651
aI255
aI971
aI4311
aI7137
aI11821
aI45041
aI31081
asS'd'
I11552
sa(dp23104
S'a'
I49633
sS's'
I17
sS'm'
(lp23105
I1
aI1
aI5
aI9
aI7
aI55
aI51
aI147
aI371
aI881
aI359
aI3021
aI1141
aI14515
aI14605
aI64067
aI98231
asS'd'
I11553
sa(dp23106
S'a'
I49634
sS's'
I17
sS'm'
(lp23107
I1
aI3
aI1
aI3
aI29
aI9
aI109
aI21
aI437
aI321
aI753
aI3227
aI2929
aI14787
aI2451
aI54331
aI115921
asS'd'
I11554
sa(dp23108
S'a'
I49636
sS's'
I17
sS'm'
(lp23109
I1
aI3
aI5
aI7
aI1
aI29
aI13
aI167
aI89
aI185
aI409
aI209
aI6625
aI5237
aI22513
aI2095
aI26427
asS'd'
I11555
sa(dp23110
S'a'
I49643
sS's'
I17
sS'm'
(lp23111
I1
aI1
aI1
aI3
aI3
aI31
aI25
aI145
aI27
aI345
aI957
aI823
aI7873
aI9469
aI29115
aI12455
aI39447
asS'd'
I11556
sa(dp23112
S'a'
I49654
sS's'
I17
sS'm'
(lp23113
I1
aI1
aI3
aI3
aI31
aI15
aI99
aI181
aI247
aI165
aI441
aI59
aI1181
aI2851
aI1337
aI4929
aI31079
asS'd'
I11557
sa(dp23114
S'a'
I49657
sS's'
I17
sS'm'
(lp23115
I1
aI1
aI3
aI11
aI3
aI47
aI41
aI53
aI41
aI435
aI945
aI3839
aI8083
aI4927
aI26919
aI24689
aI61015
asS'd'
I11558
sa(dp23116
S'a'
I49663
sS's'
I17
sS'm'
(lp23117
I1
aI3
aI3
aI9
aI7
aI21
aI121
aI233
aI493
aI271
aI549
aI1627
aI5861
aI377
aI20751
aI52927
aI3649
asS'd'
I11559
sa(dp23118
S'a'
I49667
sS's'
I17
sS'm'
(lp23119
I1
aI1
aI1
aI7
aI15
aI49
aI29
aI149
aI57
aI513
aI873
aI93
aI337
aI12559
aI24287
aI27771
aI28207
asS'd'
I11560
sa(dp23120
S'a'
I49693
sS's'
I17
sS'm'
(lp23121
I1
aI1
aI5
aI7
aI7
aI7
aI75
aI161
aI419
aI601
aI55
aI2599
aI5325
aI12419
aI26755
aI5103
aI10231
asS'd'
I11561
sa(dp23122
S'a'
I49700
sS's'
I17
sS'm'
(lp23123
I1
aI3
aI5
aI15
aI3
aI55
aI67
aI183
aI7
aI371
aI141
aI4093
aI4567
aI13971
aI3327
aI20701
aI78819
asS'd'
I11562
sa(dp23124
S'a'
I49703
sS's'
I17
sS'm'
(lp23125
I1
aI1
aI1
aI13
aI15
aI41
aI45
aI29
aI375
aI235
aI1985
aI1051
aI5863
aI7043
aI11143
aI59381
aI55007
asS'd'
I11563
sa(dp23126
S'a'
I49712
sS's'
I17
sS'm'
(lp23127
I1
aI3
aI1
aI15
aI15
aI29
aI15
aI101
aI395
aI39
aI1839
aI1689
aI429
aI405
aI29337
aI28573
aI10599
asS'd'
I11564
sa(dp23128
S'a'
I49717
sS's'
I17
sS'm'
(lp23129
I1
aI3
aI3
aI11
aI7
aI5
aI11
aI153
aI235
aI227
aI561
aI1037
aI2283
aI6657
aI6729
aI17939
aI29809
asS'd'
I11565
sa(dp23130
S'a'
I49718
sS's'
I17
sS'm'
(lp23131
I1
aI1
aI1
aI5
aI15
aI59
aI33
aI69
aI275
aI447
aI661
aI2071
aI5811
aI10463
aI32707
aI64503
aI106313
asS'd'
I11566
sa(dp23132
S'a'
I49721
sS's'
I17
sS'm'
(lp23133
I1
aI1
aI7
aI1
aI21
aI53
aI21
aI235
aI497
aI309
aI1207
aI1613
aI7739
aI12785
aI7743
aI37671
aI29197
asS'd'
I11567
sa(dp23134
S'a'
I49727
sS's'
I17
sS'm'
(lp23135
I1
aI3
aI5
aI9
aI9
aI51
aI33
aI105
aI275
aI917
aI1911
aI3607
aI865
aI899
aI5405
aI59593
aI113965
asS'd'
I11568
sa(dp23136
S'a'
I49730
sS's'
I17
sS'm'
(lp23137
I1
aI3
aI1
aI7
aI7
aI63
aI51
aI83
aI481
aI347
aI1323
aI3549
aI2873
aI12527
aI16515
aI61077
aI63239
asS'd'
I11569
sa(dp23138
S'a'
I49735
sS's'
I17
sS'm'
(lp23139
I1
aI1
aI1
aI1
aI13
aI21
aI87
aI139
aI461
aI215
aI1173
aI1197
aI2091
aI11247
aI25647
aI23443
aI105761
asS'd'
I11570
sa(dp23140
S'a'
I49739
sS's'
I17
sS'm'
(lp23141
I1
aI1
aI1
aI1
aI27
aI33
aI21
aI27
aI365
aI75
aI351
aI2111
aI3897
aI13325
aI4821
aI41355
aI95681
asS'd'
I11571
sa(dp23142
S'a'
I49747
sS's'
I17
sS'm'
(lp23143
I1
aI3
aI5
aI5
aI15
aI41
aI15
aI93
aI397
aI461
aI1993
aI321
aI4375
aI1205
aI18417
aI37549
aI30181
asS'd'
I11572
sa(dp23144
S'a'
I49765
sS's'
I17
sS'm'
(lp23145
I1
aI3
aI5
aI15
aI1
aI49
aI101
aI129
aI215
aI773
aI1265
aI2245
aI2517
aI16261
aI32149
aI3545
aI27631
asS'd'
I11573
sa(dp23146
S'a'
I49770
sS's'
I17
sS'm'
(lp23147
I1
aI1
aI1
aI15
aI5
aI5
aI3
aI127
aI265
aI721
aI875
aI3117
aI2177
aI7843
aI15871
aI22687
aI89347
asS'd'
I11574
sa(dp23148
S'a'
I49772
sS's'
I17
sS'm'
(lp23149
I1
aI3
aI7
aI11
aI29
aI23
aI69
aI41
aI155
aI257
aI563
aI509
aI6105
aI9169
aI18341
aI25373
aI127397
asS'd'
I11575
sa(dp23150
S'a'
I49777
sS's'
I17
sS'm'
(lp23151
I1
aI1
aI5
aI13
aI31
aI23
aI65
aI131
aI131
aI61
aI1979
aI2737
aI3793
aI3617
aI14385
aI189
aI84567
asS'd'
I11576
sa(dp23152
S'a'
I49789
sS's'
I17
sS'm'
(lp23153
I1
aI1
aI1
aI13
aI21
aI33
aI43
aI97
aI83
aI903
aI1971
aI3209
aI5391
aI7703
aI13795
aI3895
aI78045
asS'd'
I11577
sa(dp23154
S'a'
I49817
sS's'
I17
sS'm'
(lp23155
I1
aI3
aI3
aI5
aI17
aI53
aI113
aI237
aI269
aI83
aI589
aI4029
aI3309
aI14531
aI11359
aI25803
aI25525
asS'd'
I11578
sa(dp23156
S'a'
I49829
sS's'
I17
sS'm'
(lp23157
I1
aI1
aI7
aI1
aI21
aI35
aI43
aI73
aI251
aI705
aI1737
aI3341
aI1581
aI9663
aI6251
aI16329
aI44491
asS'd'
I11579
sa(dp23158
S'a'
I49839
sS's'
I17
sS'm'
(lp23159
I1
aI3
aI3
aI7
aI17
aI5
aI65
aI19
aI203
aI717
aI807
aI1759
aI6907
aI15801
aI30369
aI2655
aI69565
asS'd'
I11580
sa(dp23160
S'a'
I49844
sS's'
I17
sS'm'
(lp23161
I1
aI1
aI1
aI1
aI31
aI21
aI75
aI221
aI115
aI395
aI1495
aI2739
aI1745
aI5981
aI28045
aI56581
aI130695
asS'd'
I11581
sa(dp23162
S'a'
I49851
sS's'
I17
sS'm'
(lp23163
I1
aI3
aI1
aI9
aI27
aI5
aI113
aI123
aI367
aI701
aI647
aI117
aI2389
aI12309
aI1747
aI23421
aI21583
asS'd'
I11582
sa(dp23164
S'a'
I49854
sS's'
I17
sS'm'
(lp23165
I1
aI1
aI1
aI15
aI27
aI57
aI95
aI81
aI347
aI765
aI1435
aI1727
aI153
aI6051
aI27085
aI62787
aI40903
asS'd'
I11583
sa(dp23166
S'a'
I49874
sS's'
I17
sS'm'
(lp23167
I1
aI1
aI3
aI11
aI23
aI29
aI23
aI29
aI169
aI653
aI835
aI357
aI5113
aI5293
aI11779
aI23567
aI64569
asS'd'
I11584
sa(dp23168
S'a'
I49883
sS's'
I17
sS'm'
(lp23169
I1
aI1
aI7
aI13
aI31
aI7
aI101
aI235
aI99
aI247
aI1267
aI509
aI3927
aI14317
aI3217
aI24389
aI34215
asS'd'
I11585
sa(dp23170
S'a'
I49886
sS's'
I17
sS'm'
(lp23171
I1
aI3
aI3
aI5
aI21
aI27
aI33
aI229
aI33
aI551
aI815
aI3551
aI4261
aI13325
aI31117
aI40689
aI66549
asS'd'
I11586
sa(dp23172
S'a'
I49892
sS's'
I17
sS'm'
(lp23173
I1
aI3
aI7
aI1
aI23
aI1
aI99
aI11
aI139
aI569
aI365
aI1233
aI3281
aI7817
aI8833
aI47699
aI97825
asS'd'
I11587
sa(dp23174
S'a'
I49895
sS's'
I17
sS'm'
(lp23175
I1
aI1
aI3
aI1
aI19
aI39
aI19
aI151
aI25
aI73
aI1271
aI1435
aI3109
aI2571
aI9191
aI48257
aI61001
asS'd'
I11588
sa(dp23176
S'a'
I49914
sS's'
I17
sS'm'
(lp23177
I1
aI3
aI7
aI1
aI23
aI63
aI1
aI115
aI159
aI943
aI1637
aI1443
aI809
aI10705
aI12563
aI63111
aI96343
asS'd'
I11589
sa(dp23178
S'a'
I49931
sS's'
I17
sS'm'
(lp23179
I1
aI1
aI7
aI15
aI5
aI17
aI65
aI157
aI45
aI199
aI371
aI2497
aI4367
aI9109
aI31955
aI64253
aI69279
asS'd'
I11590
sa(dp23180
S'a'
I49933
sS's'
I17
sS'm'
(lp23181
I1
aI3
aI3
aI15
aI29
aI45
aI103
aI183
aI87
aI543
aI97
aI1545
aI2719
aI5619
aI28871
aI4049
aI111825
asS'd'
I11591
sa(dp23182
S'a'
I49934
sS's'
I17
sS'm'
(lp23183
I1
aI1
aI5
aI7
aI27
aI63
aI65
aI241
aI103
aI483
aI579
aI3589
aI5673
aI13283
aI24193
aI31993
aI72713
asS'd'
I11592
sa(dp23184
S'a'
I49941
sS's'
I17
sS'm'
(lp23185
I1
aI3
aI7
aI9
aI21
aI35
aI59
aI183
aI459
aI211
aI753
aI3941
aI5389
aI10535
aI2895
aI44307
aI26577
asS'd'
I11593
sa(dp23186
S'a'
I49952
sS's'
I17
sS'm'
(lp23187
I1
aI3
aI1
aI3
aI9
aI11
aI15
aI141
aI159
aI853
aI1975
aI4027
aI8053
aI16129
aI32687
aI29117
aI67507
asS'd'
I11594
sa(dp23188
S'a'
I49967
sS's'
I17
sS'm'
(lp23189
I1
aI3
aI3
aI1
aI19
aI51
aI55
aI167
aI85
aI869
aI437
aI457
aI7879
aI2097
aI4403
aI2139
aI10589
asS'd'
I11595
sa(dp23190
S'a'
I49969
sS's'
I17
sS'm'
(lp23191
I1
aI3
aI3
aI15
aI19
aI33
aI63
aI229
aI197
aI269
aI1189
aI317
aI5087
aI3147
aI787
aI64317
aI43293
asS'd'
I11596
sa(dp23192
S'a'
I49972
sS's'
I17
sS'm'
(lp23193
I1
aI3
aI5
aI11
aI25
aI25
aI121
aI37
aI371
aI117
aI1683
aI2921
aI671
aI11353
aI32273
aI57597
aI56901
asS'd'
I11597
sa(dp23194
S'a'
I49976
sS's'
I17
sS'm'
(lp23195
I1
aI3
aI7
aI7
aI9
aI37
aI91
aI159
aI195
aI1
aI77
aI2085
aI985
aI9999
aI5639
aI25041
aI66393
asS'd'
I11598
sa(dp23196
S'a'
I49979
sS's'
I17
sS'm'
(lp23197
I1
aI1
aI7
aI5
aI11
aI17
aI67
aI21
aI301
aI971
aI591
aI3809
aI4795
aI12301
aI16977
aI27495
aI98539
asS'd'
I11599
sa(dp23198
S'a'
I49999
sS's'
I17
sS'm'
(lp23199
I1
aI3
aI1
aI9
aI13
aI53
aI83
aI205
aI111
aI609
aI631
aI23
aI1781
aI15401
aI1563
aI34367
aI40345
asS'd'
I11600
sa(dp23200
S'a'
I50008
sS's'
I17
sS'm'
(lp23201
I1
aI3
aI7
aI1
aI31
aI23
aI101
aI47
aI55
aI905
aI953
aI733
aI5173
aI5937
aI17703
aI31077
aI49707
asS'd'
I11601
sa(dp23202
S'a'
I50030
sS's'
I17
sS'm'
(lp23203
I1
aI3
aI5
aI3
aI5
aI3
aI127
aI171
aI511
aI289
aI685
aI1157
aI6521
aI3301
aI3017
aI58857
aI55289
asS'd'
I11602
sa(dp23204
S'a'
I50041
sS's'
I17
sS'm'
(lp23205
I1
aI1
aI5
aI1
aI1
aI29
aI59
aI7
aI423
aI83
aI797
aI2633
aI2015
aI1657
aI7575
aI39819
aI181
asS'd'
I11603
sa(dp23206
S'a'
I50042
sS's'
I17
sS'm'
(lp23207
I1
aI3
aI5
aI15
aI25
aI27
aI39
aI99
aI83
aI381
aI401
aI1033
aI867
aI15645
aI28643
aI34917
aI53215
asS'd'
I11604
sa(dp23208
S'a'
I50044
sS's'
I17
sS'm'
(lp23209
I1
aI3
aI5
aI15
aI17
aI1
aI67
aI63
aI355
aI841
aI681
aI2807
aI531
aI15295
aI7859
aI64031
aI121519
asS'd'
I11605
sa(dp23210
S'a'
I50057
sS's'
I17
sS'm'
(lp23211
I1
aI3
aI1
aI5
aI21
aI57
aI63
aI247
aI467
aI101
aI129
aI2627
aI4763
aI479
aI11145
aI8861
aI69803
asS'd'
I11606
sa(dp23212
S'a'
I50060
sS's'
I17
sS'm'
(lp23213
I1
aI3
aI1
aI15
aI13
aI13
aI77
aI39
aI297
aI401
aI1653
aI659
aI3909
aI13179
aI10477
aI45967
aI1665
asS'd'
I11607
sa(dp23214
S'a'
I50075
sS's'
I17
sS'm'
(lp23215
I1
aI3
aI7
aI5
aI29
aI17
aI35
aI157
aI309
aI747
aI1717
aI1279
aI6103
aI3509
aI17499
aI22989
aI43157
asS'd'
I11608
sa(dp23216
S'a'
I50077
sS's'
I17
sS'm'
(lp23217
I1
aI1
aI1
aI7
aI11
aI51
aI55
aI119
aI145
aI505
aI179
aI3979
aI1237
aI12801
aI15921
aI13561
aI69161
asS'd'
I11609
sa(dp23218
S'a'
I50091
sS's'
I17
sS'm'
(lp23219
I1
aI1
aI5
aI7
aI13
aI9
aI73
aI203
aI247
aI257
aI1607
aI1183
aI6237
aI12327
aI5059
aI51645
aI88781
asS'd'
I11610
sa(dp23220
S'a'
I50096
sS's'
I17
sS'm'
(lp23221
I1
aI3
aI5
aI9
aI1
aI9
aI27
aI59
aI235
aI81
aI689
aI2457
aI893
aI6107
aI27643
aI40145
aI2329
asS'd'
I11611
sa(dp23222
S'a'
I50099
sS's'
I17
sS'm'
(lp23223
I1
aI1
aI1
aI11
aI11
aI19
aI65
aI63
aI27
aI513
aI1473
aI2955
aI8037
aI4991
aI22035
aI41965
aI82589
asS'd'
I11612
sa(dp23224
S'a'
I50106
sS's'
I17
sS'm'
(lp23225
I1
aI3
aI1
aI1
aI9
aI53
aI97
aI95
aI247
aI379
aI259
aI2789
aI1433
aI15299
aI18309
aI51813
aI63271
asS'd'
I11613
sa(dp23226
S'a'
I50116
sS's'
I17
sS'm'
(lp23227
I1
aI1
aI5
aI1
aI11
aI41
aI69
aI193
aI391
aI27
aI1511
aI1575
aI1161
aI14741
aI25193
aI31149
aI79573
asS'd'
I11614
sa(dp23228
S'a'
I50133
sS's'
I17
sS'm'
(lp23229
I1
aI3
aI7
aI15
aI19
aI31
aI101
aI9
aI111
aI427
aI531
aI1335
aI767
aI15075
aI28373
aI54015
aI108021
asS'd'
I11615
sa(dp23230
S'a'
I50144
sS's'
I17
sS'm'
(lp23231
I1
aI3
aI1
aI3
aI29
aI17
aI83
aI163
aI179
aI703
aI2027
aI3027
aI5267
aI16111
aI23929
aI9653
aI38633
asS'd'
I11616
sa(dp23232
S'a'
I50153
sS's'
I17
sS'm'
(lp23233
I1
aI3
aI5
aI9
aI3
aI7
aI111
aI1
aI311
aI143
aI1563
aI2605
aI301
aI2447
aI5009
aI63767
aI37529
asS'd'
I11617
sa(dp23234
S'a'
I50154
sS's'
I17
sS'm'
(lp23235
I1
aI3
aI5
aI5
aI23
aI23
aI97
aI11
aI475
aI741
aI1385
aI2491
aI1717
aI14587
aI6289
aI27651
aI21873
asS'd'
I11618
sa(dp23236
S'a'
I50168
sS's'
I17
sS'm'
(lp23237
I1
aI1
aI5
aI1
aI31
aI31
aI119
aI93
aI209
aI861
aI1591
aI1233
aI3469
aI9799
aI5969
aI35965
aI110841
asS'd'
I11619
sa(dp23238
S'a'
I50179
sS's'
I17
sS'm'
(lp23239
I1
aI1
aI5
aI13
aI19
aI31
aI69
aI107
aI423
aI651
aI757
aI665
aI1297
aI9985
aI14983
aI3153
aI26425
asS'd'
I11620
sa(dp23240
S'a'
I50188
sS's'
I17
sS'm'
(lp23241
I1
aI1
aI5
aI3
aI23
aI5
aI89
aI77
aI461
aI799
aI683
aI2885
aI845
aI12847
aI26721
aI13145
aI88689
asS'd'
I11621
sa(dp23242
S'a'
I50194
sS's'
I17
sS'm'
(lp23243
I1
aI1
aI3
aI3
aI19
aI3
aI41
aI71
aI247
aI293
aI1047
aI2349
aI6815
aI2413
aI13683
aI51421
aI110737
asS'd'
I11622
sa(dp23244
S'a'
I50203
sS's'
I17
sS'm'
(lp23245
I1
aI3
aI3
aI9
aI19
aI51
aI91
aI193
aI447
aI305
aI751
aI1757
aI1547
aI12683
aI4645
aI39767
aI6433
asS'd'
I11623
sa(dp23246
S'a'
I50212
sS's'
I17
sS'm'
(lp23247
I1
aI1
aI7
aI1
aI11
aI51
aI83
aI175
aI461
aI259
aI1337
aI175
aI877
aI15895
aI22487
aI8079
aI71291
asS'd'
I11624
sa(dp23248
S'a'
I50216
sS's'
I17
sS'm'
(lp23249
I1
aI1
aI1
aI11
aI13
aI23
aI19
aI69
aI285
aI629
aI563
aI2433
aI815
aI4851
aI10617
aI59949
aI59119
asS'd'
I11625
sa(dp23250
S'a'
I50222
sS's'
I17
sS'm'
(lp23251
I1
aI3
aI5
aI5
aI19
aI49
aI47
aI27
aI343
aI579
aI197
aI35
aI7051
aI2441
aI30091
aI9645
aI101899
asS'd'
I11626
sa(dp23252
S'a'
I50229
sS's'
I17
sS'm'
(lp23253
I1
aI1
aI3
aI3
aI21
aI25
aI125
aI215
aI159
aI259
aI1915
aI2193
aI4213
aI16157
aI8665
aI10967
aI112793
asS'd'
I11627
sa(dp23254
S'a'
I50230
sS's'
I17
sS'm'
(lp23255
I1
aI1
aI5
aI11
aI29
aI11
aI53
aI45
aI121
aI533
aI257
aI1749
aI6311
aI7715
aI12037
aI12003
aI38987
asS'd'
I11628
sa(dp23256
S'a'
I50234
sS's'
I17
sS'm'
(lp23257
I1
aI3
aI1
aI11
aI25
aI31
aI93
aI191
aI231
aI801
aI361
aI1275
aI5031
aI7927
aI26333
aI39795
aI45875
asS'd'
I11629
sa(dp23258
S'a'
I50251
sS's'
I17
sS'm'
(lp23259
I1
aI1
aI5
aI1
aI7
aI37
aI117
aI35
aI257
aI225
aI1769
aI1805
aI1593
aI1507
aI27741
aI31561
aI52107
asS'd'
I11630
sa(dp23260
S'a'
I50253
sS's'
I17
sS'm'
(lp23261
I1
aI1
aI3
aI3
aI27
aI23
aI55
aI5
aI137
aI677
aI459
aI2821
aI1331
aI5845
aI17751
aI17557
aI60495
asS'd'
I11631
sa(dp23262
S'a'
I50262
sS's'
I17
sS'm'
(lp23263
I1
aI1
aI7
aI7
aI5
aI47
aI85
aI17
aI287
aI757
aI2013
aI2233
aI2975
aI13769
aI23199
aI32117
aI84429
asS'd'
I11632
sa(dp23264
S'a'
I50266
sS's'
I17
sS'm'
(lp23265
I1
aI3
aI3
aI11
aI1
aI15
aI39
aI133
aI79
aI915
aI147
aI1489
aI2319
aI13715
aI31317
aI46785
aI64147
asS'd'
I11633
sa(dp23266
S'a'
I50272
sS's'
I17
sS'm'
(lp23267
I1
aI1
aI3
aI11
aI1
aI25
aI83
aI69
aI395
aI537
aI895
aI565
aI2781
aI12685
aI7831
aI36369
aI27871
asS'd'
I11634
sa(dp23268
S'a'
I50287
sS's'
I17
sS'm'
(lp23269
I1
aI1
aI1
aI1
aI13
aI17
aI59
aI179
aI509
aI979
aI315
aI3495
aI1773
aI16375
aI27873
aI18065
aI20285
asS'd'
I11635
sa(dp23270
S'a'
I50295
sS's'
I17
sS'm'
(lp23271
I1
aI1
aI5
aI1
aI27
aI31
aI39
aI251
aI121
aI899
aI751
aI1603
aI7501
aI425
aI25791
aI35407
aI110405
asS'd'
I11636
sa(dp23272
S'a'
I50305
sS's'
I17
sS'm'
(lp23273
I1
aI1
aI7
aI11
aI11
aI3
aI107
aI247
aI79
aI349
aI405
aI3469
aI2201
aI8007
aI22891
aI7901
aI11413
asS'd'
I11637
sa(dp23274
S'a'
I50306
sS's'
I17
sS'm'
(lp23275
I1
aI1
aI1
aI11
aI15
aI55
aI121
aI151
aI127
aI239
aI115
aI611
aI6191
aI15435
aI19831
aI64745
aI110473
asS'd'
I11638
sa(dp23276
S'a'
I50311
sS's'
I17
sS'm'
(lp23277
I1
aI1
aI7
aI1
aI21
aI25
aI57
aI11
aI31
aI823
aI1855
aI2337
aI7655
aI10845
aI22167
aI36977
aI94265
asS'd'
I11639
sa(dp23278
S'a'
I50320
sS's'
I17
sS'm'
(lp23279
I1
aI3
aI5
aI1
aI31
aI11
aI117
aI97
aI341
aI953
aI1499
aI2487
aI559
aI8609
aI6321
aI20669
aI28945
asS'd'
I11640
sa(dp23280
S'a'
I50323
sS's'
I17
sS'm'
(lp23281
I1
aI3
aI7
aI1
aI5
aI27
aI15
aI161
aI83
aI139
aI75
aI3645
aI5227
aI16199
aI1833
aI21767
aI67579
asS'd'
I11641
sa(dp23282
S'a'
I50325
sS's'
I17
sS'm'
(lp23283
I1
aI1
aI7
aI5
aI31
aI35
aI75
aI115
aI451
aI773
aI1987
aI1069
aI651
aI961
aI16317
aI18391
aI56519
asS'd'
I11642
sa(dp23284
S'a'
I50339
sS's'
I17
sS'm'
(lp23285
I1
aI1
aI1
aI1
aI31
aI11
aI23
aI255
aI215
aI251
aI867
aI2381
aI2351
aI13189
aI17705
aI33569
aI102361
asS'd'
I11643
sa(dp23286
S'a'
I50356
sS's'
I17
sS'm'
(lp23287
I1
aI3
aI7
aI1
aI17
aI49
aI49
aI125
aI445
aI947
aI1985
aI2113
aI3025
aI6277
aI1981
aI33329
aI99413
asS'd'
I11644
sa(dp23288
S'a'
I50365
sS's'
I17
sS'm'
(lp23289
I1
aI3
aI3
aI3
aI27
aI7
aI59
aI109
aI37
aI777
aI1359
aI2157
aI3185
aI7317
aI30887
aI10499
aI126563
asS'd'
I11645
sa(dp23290
S'a'
I50366
sS's'
I17
sS'm'
(lp23291
I1
aI3
aI3
aI9
aI27
aI5
aI99
aI167
aI457
aI363
aI2031
aI1805
aI4661
aI8695
aI4667
aI61129
aI81143
asS'd'
I11646
sa(dp23292
S'a'
I50368
sS's'
I17
sS'm'
(lp23293
I1
aI1
aI5
aI13
aI5
aI47
aI95
aI249
aI289
aI631
aI1739
aI2947
aI7169
aI13019
aI10429
aI16197
aI11539
asS'd'
I11647
sa(dp23294
S'a'
I50377
sS's'
I17
sS'm'
(lp23295
I1
aI3
aI7
aI15
aI27
aI45
aI93
aI131
aI269
aI835
aI399
aI1133
aI6509
aI1279
aI3635
aI17977
aI38667
asS'd'
I11648
sa(dp23296
S'a'
I50386
sS's'
I17
sS'm'
(lp23297
I1
aI1
aI5
aI3
aI17
aI51
aI1
aI77
aI105
aI237
aI995
aI2643
aI6921
aI6707
aI30129
aI1543
aI94501
asS'd'
I11649
sa(dp23298
S'a'
I50397
sS's'
I17
sS'm'
(lp23299
I1
aI1
aI1
aI7
aI29
aI1
aI117
aI33
aI141
aI805
aI1553
aI3943
aI45
aI8911
aI24191
aI45191
aI36525
asS'd'
I11650
sa(dp23300
S'a'
I50411
sS's'
I17
sS'm'
(lp23301
I1
aI1
aI7
aI7
aI31
aI21
aI97
aI29
aI179
aI345
aI1059
aI701
aI7709
aI15831
aI22923
aI57233
aI58961
asS'd'
I11651
sa(dp23302
S'a'
I50414
sS's'
I17
sS'm'
(lp23303
I1
aI1
aI3
aI15
aI15
aI5
aI85
aI227
aI13
aI351
aI1167
aI3283
aI6833
aI565
aI21019
aI53249
aI4639
asS'd'
I11652
sa(dp23304
S'a'
I50425
sS's'
I17
sS'm'
(lp23305
I1
aI3
aI3
aI5
aI27
aI1
aI5
aI89
aI101
aI295
aI481
aI2397
aI1459
aI3729
aI3703
aI25109
aI69237
asS'd'
I11653
sa(dp23306
S'a'
I50451
sS's'
I17
sS'm'
(lp23307
I1
aI3
aI7
aI13
aI31
aI37
aI69
aI147
aI505
aI487
aI1701
aI1145
aI2061
aI10067
aI18269
aI13049
aI92091
asS'd'
I11654
sa(dp23308
S'a'
I50463
sS's'
I17
sS'm'
(lp23309
I1
aI3
aI3
aI1
aI29
aI39
aI33
aI199
aI499
aI377
aI1081
aI3787
aI4843
aI16287
aI23397
aI19083
aI91381
asS'd'
I11655
sa(dp23310
S'a'
I50467
sS's'
I17
sS'm'
(lp23311
I1
aI1
aI1
aI13
aI21
aI61
aI121
aI251
aI511
aI615
aI625
aI2245
aI5951
aI16165
aI2155
aI37301
aI68319
asS'd'
I11656
sa(dp23312
S'a'
I50473
sS's'
I17
sS'm'
(lp23313
I1
aI1
aI1
aI15
aI19
aI35
aI57
aI99
aI1
aI97
aI1177
aI3109
aI7213
aI5447
aI25251
aI24803
aI107449
asS'd'
I11657
sa(dp23314
S'a'
I50487
sS's'
I17
sS'm'
(lp23315
I1
aI3
aI1
aI11
aI11
aI59
aI95
aI135
aI329
aI931
aI843
aI2847
aI463
aI10725
aI3933
aI32325
aI44545
asS'd'
I11658
sa(dp23316
S'a'
I50501
sS's'
I17
sS'm'
(lp23317
I1
aI1
aI7
aI13
aI13
aI57
aI29
aI175
aI11
aI701
aI231
aI2907
aI5555
aI16159
aI1249
aI38049
aI40115
asS'd'
I11659
sa(dp23318
S'a'
I50505
sS's'
I17
sS'm'
(lp23319
I1
aI1
aI5
aI15
aI23
aI37
aI47
aI221
aI465
aI631
aI57
aI1189
aI2083
aI6561
aI10725
aI8015
aI21231
asS'd'
I11660
sa(dp23320
S'a'
I50514
sS's'
I17
sS'm'
(lp23321
I1
aI3
aI7
aI5
aI9
aI25
aI31
aI47
aI139
aI639
aI999
aI2909
aI39
aI16227
aI16967
aI30555
aI125541
asS'd'
I11661
sa(dp23322
S'a'
I50516
sS's'
I17
sS'm'
(lp23323
I1
aI3
aI3
aI9
aI5
aI3
aI9
aI9
aI43
aI999
aI159
aI3063
aI3661
aI7089
aI28929
aI55305
aI105521
asS'd'
I11662
sa(dp23324
S'a'
I50536
sS's'
I17
sS'm'
(lp23325
I1
aI3
aI1
aI11
aI17
aI7
aI3
aI57
aI457
aI753
aI135
aI3721
aI1111
aI7267
aI12603
aI50511
aI85433
asS'd'
I11663
sa(dp23326
S'a'
I50547
sS's'
I17
sS'm'
(lp23327
I1
aI3
aI5
aI1
aI27
aI3
aI107
aI187
aI247
aI891
aI1311
aI423
aI1767
aI14769
aI22119
aI36225
aI94267
asS'd'
I11664
sa(dp23328
S'a'
I50566
sS's'
I17
sS'm'
(lp23329
I1
aI3
aI5
aI15
aI1
aI13
aI65
aI35
aI435
aI557
aI1755
aI1343
aI2647
aI179
aI7781
aI62903
aI18741
asS'd'
I11665
sa(dp23330
S'a'
I50572
sS's'
I17
sS'm'
(lp23331
I1
aI1
aI7
aI15
aI29
aI57
aI63
aI227
aI407
aI163
aI1207
aI2717
aI2731
aI1737
aI6379
aI14937
aI46683
asS'd'
I11666
sa(dp23332
S'a'
I50583
sS's'
I17
sS'm'
(lp23333
I1
aI1
aI1
aI9
aI25
aI35
aI93
aI1
aI77
aI677
aI875
aI3787
aI3075
aI14033
aI23017
aI3487
aI14999
asS'd'
I11667
sa(dp23334
S'a'
I50593
sS's'
I17
sS'm'
(lp23335
I1
aI3
aI3
aI9
aI3
aI45
aI115
aI61
aI437
aI823
aI1401
aI459
aI301
aI5519
aI31003
aI64499
aI1757
asS'd'
I11668
sa(dp23336
S'a'
I50599
sS's'
I17
sS'm'
(lp23337
I1
aI3
aI1
aI11
aI23
aI37
aI69
aI215
aI197
aI961
aI1501
aI2953
aI3679
aI6775
aI24679
aI44215
aI52357
asS'd'
I11669
sa(dp23338
S'a'
I50613
sS's'
I17
sS'm'
(lp23339
I1
aI3
aI1
aI15
aI29
aI23
aI1
aI133
aI451
aI677
aI687
aI1269
aI5987
aI11975
aI11929
aI63691
aI48683
asS'd'
I11670
sa(dp23340
S'a'
I50617
sS's'
I17
sS'm'
(lp23341
I1
aI1
aI1
aI7
aI13
aI31
aI13
aI71
aI355
aI345
aI1193
aI3421
aI7601
aI7413
aI6719
aI28681
aI97605
asS'd'
I11671
sa(dp23342
S'a'
I50618
sS's'
I17
sS'm'
(lp23343
I1
aI1
aI5
aI13
aI23
aI3
aI15
aI253
aI109
aI17
aI341
aI471
aI1131
aI14901
aI31783
aI41369
aI64139
asS'd'
I11672
sa(dp23344
S'a'
I50620
sS's'
I17
sS'm'
(lp23345
I1
aI1
aI3
aI1
aI25
aI25
aI85
aI157
aI443
aI83
aI269
aI3789
aI7977
aI7783
aI28433
aI30563
aI82805
asS'd'
I11673
sa(dp23346
S'a'
I50628
sS's'
I17
sS'm'
(lp23347
I1
aI1
aI7
aI5
aI3
aI11
aI83
aI63
aI253
aI349
aI217
aI2733
aI4183
aI2759
aI7617
aI41749
aI14015
asS'd'
I11674
sa(dp23348
S'a'
I50638
sS's'
I17
sS'm'
(lp23349
I1
aI1
aI5
aI7
aI7
aI7
aI91
aI201
aI449
aI247
aI889
aI3829
aI3529
aI14253
aI24091
aI33521
aI6049
asS'd'
I11675
sa(dp23350
S'a'
I50652
sS's'
I17
sS'm'
(lp23351
I1
aI1
aI3
aI7
aI25
aI7
aI123
aI161
aI227
aI965
aI511
aI619
aI4359
aI11833
aI12859
aI26091
aI3867
asS'd'
I11676
sa(dp23352
S'a'
I50655
sS's'
I17
sS'm'
(lp23353
I1
aI1
aI7
aI9
aI5
aI41
aI71
aI111
aI95
aI261
aI1809
aI3835
aI7625
aI12085
aI28885
aI64829
aI48981
asS'd'
I11677
sa(dp23354
S'a'
I50662
sS's'
I17
sS'm'
(lp23355
I1
aI1
aI5
aI3
aI13
aI55
aI57
aI79
aI457
aI785
aI653
aI1429
aI3879
aI13559
aI3953
aI18205
aI5777
asS'd'
I11678
sa(dp23356
S'a'
I50665
sS's'
I17
sS'm'
(lp23357
I1
aI3
aI5
aI9
aI23
aI25
aI107
aI255
aI151
aI191
aI119
aI3367
aI1081
aI12691
aI3575
aI38171
aI42573
asS'd'
I11679
sa(dp23358
S'a'
I50673
sS's'
I17
sS'm'
(lp23359
I1
aI1
aI7
aI15
aI3
aI23
aI115
aI233
aI265
aI187
aI1961
aI1303
aI5101
aI1035
aI6803
aI14557
aI4527
asS'd'
I11680
sa(dp23360
S'a'
I50683
sS's'
I17
sS'm'
(lp23361
I1
aI1
aI3
aI11
aI19
aI37
aI45
aI167
aI17
aI793
aI1361
aI3571
aI5889
aI14421
aI20453
aI5093
aI59927
asS'd'
I11681
sa(dp23362
S'a'
I50689
sS's'
I17
sS'm'
(lp23363
I1
aI1
aI7
aI15
aI3
aI53
aI1
aI11
aI159
aI389
aI171
aI2351
aI7189
aI3109
aI1541
aI53595
aI24247
asS'd'
I11682
sa(dp23364
S'a'
I50690
sS's'
I17
sS'm'
(lp23365
I1
aI1
aI7
aI1
aI7
aI43
aI75
aI175
aI253
aI687
aI1811
aI3277
aI447
aI8711
aI14281
aI53265
aI7379
asS'd'
I11683
sa(dp23366
S'a'
I50692
sS's'
I17
sS'm'
(lp23367
I1
aI1
aI5
aI3
aI21
aI45
aI113
aI25
aI309
aI31
aI1765
aI305
aI1423
aI115
aI26421
aI50267
aI122115
asS'd'
I11684
sa(dp23368
S'a'
I50696
sS's'
I17
sS'm'
(lp23369
I1
aI3
aI5
aI13
aI15
aI47
aI17
aI17
aI445
aI775
aI243
aI3959
aI7263
aI9375
aI12017
aI57399
aI58203
asS'd'
I11685
sa(dp23370
S'a'
I50704
sS's'
I17
sS'm'
(lp23371
I1
aI3
aI3
aI1
aI31
aI37
aI37
aI213
aI125
aI929
aI243
aI1011
aI2841
aI4499
aI16961
aI12639
aI23381
asS'd'
I11686
sa(dp23372
S'a'
I50713
sS's'
I17
sS'm'
(lp23373
I1
aI3
aI1
aI3
aI7
aI53
aI31
aI165
aI311
aI239
aI731
aI1759
aI1769
aI203
aI23201
aI20267
aI33381
asS'd'
I11687
sa(dp23374
S'a'
I50719
sS's'
I17
sS'm'
(lp23375
I1
aI3
aI3
aI11
aI1
aI1
aI73
aI57
aI497
aI693
aI693
aI861
aI5587
aI16307
aI8559
aI28785
aI91147
asS'd'
I11688
sa(dp23376
S'a'
I50738
sS's'
I17
sS'm'
(lp23377
I1
aI3
aI1
aI3
aI11
aI61
aI11
aI241
aI473
aI479
aI1831
aI1771
aI25
aI10217
aI32683
aI40165
aI98433
asS'd'
I11689
sa(dp23378
S'a'
I50757
sS's'
I17
sS'm'
(lp23379
I1
aI1
aI5
aI3
aI17
aI51
aI39
aI27
aI189
aI631
aI689
aI2849
aI1849
aI9143
aI19263
aI32729
aI23031
asS'd'
I11690
sa(dp23380
S'a'
I50761
sS's'
I17
sS'm'
(lp23381
I1
aI1
aI5
aI9
aI15
aI39
aI103
aI83
aI485
aI689
aI1561
aI55
aI5219
aI12069
aI32225
aI7781
aI114299
asS'd'
I11691
sa(dp23382
S'a'
I50764
sS's'
I17
sS'm'
(lp23383
I1
aI1
aI1
aI1
aI31
aI49
aI71
aI145
aI485
aI907
aI1551
aI3931
aI4081
aI2159
aI24747
aI6953
aI15887
asS'd'
I11692
sa(dp23384
S'a'
I50770
sS's'
I17
sS'm'
(lp23385
I1
aI1
aI3
aI7
aI27
aI61
aI57
aI153
aI15
aI881
aI271
aI267
aI5827
aI7625
aI18179
aI3769
aI42211
asS'd'
I11693
sa(dp23386
S'a'
I50776
sS's'
I17
sS'm'
(lp23387
I1
aI3
aI1
aI3
aI21
aI49
aI65
aI177
aI341
aI851
aI1825
aI3347
aI113
aI8077
aI1117
aI9463
aI115821
asS'd'
I11694
sa(dp23388
S'a'
I50803
sS's'
I17
sS'm'
(lp23389
I1
aI3
aI5
aI11
aI27
aI17
aI75
aI35
aI475
aI389
aI313
aI2187
aI7005
aI911
aI21921
aI10979
aI13705
asS'd'
I11695
sa(dp23390
S'a'
I50805
sS's'
I17
sS'm'
(lp23391
I1
aI1
aI5
aI9
aI1
aI49
aI15
aI21
aI163
aI355
aI193
aI3473
aI4451
aI5325
aI28343
aI251
aI125963
asS'd'
I11696
sa(dp23392
S'a'
I50806
sS's'
I17
sS'm'
(lp23393
I1
aI3
aI7
aI1
aI9
aI63
aI9
aI129
aI453
aI887
aI1841
aI597
aI1989
aI14755
aI7847
aI7581
aI251
asS'd'
I11697
sa(dp23394
S'a'
I50816
sS's'
I17
sS'm'
(lp23395
I1
aI3
aI7
aI13
aI31
aI15
aI123
aI3
aI427
aI101
aI1039
aI1355
aI3653
aI2871
aI28937
aI31243
aI108827
asS'd'
I11698
sa(dp23396
S'a'
I50826
sS's'
I17
sS'm'
(lp23397
I1
aI3
aI7
aI13
aI3
aI17
aI71
aI243
aI145
aI747
aI1933
aI1105
aI455
aI6355
aI20321
aI60075
aI31575
asS'd'
I11699
sa(dp23398
S'a'
I50840
sS's'
I17
sS'm'
(lp23399
I1
aI3
aI5
aI5
aI11
aI25
aI95
aI85
aI461
aI459
aI313
aI173
aI1413
aI15761
aI31481
aI63793
aI29047
asS'd'
I11700
sa(dp23400
S'a'
I50845
sS's'
I17
sS'm'
(lp23401
I1
aI3
aI7
aI7
aI3
aI5
aI3
aI95
aI107
aI995
aI1203
aI2965
aI2419
aI5325
aI17667
aI40205
aI91059
asS'd'
I11701
sa(dp23402
S'a'
I50852
sS's'
I17
sS'm'
(lp23403
I1
aI1
aI3
aI9
aI25
aI3
aI113
aI79
aI359
aI69
aI93
aI1539
aI483
aI12701
aI19075
aI35021
aI17929
asS'd'
I11702
sa(dp23404
S'a'
I50855
sS's'
I17
sS'm'
(lp23405
I1
aI3
aI5
aI1
aI31
aI35
aI67
aI97
aI105
aI381
aI973
aI1355
aI3901
aI3847
aI12343
aI64309
aI29835
asS'd'
I11703
sa(dp23406
S'a'
I50862
sS's'
I17
sS'm'
(lp23407
I1
aI3
aI7
aI1
aI11
aI33
aI117
aI237
aI449
aI101
aI317
aI23
aI5157
aI8187
aI28839
aI29465
aI97485
asS'd'
I11704
sa(dp23408
S'a'
I50876
sS's'
I17
sS'm'
(lp23409
I1
aI3
aI5
aI5
aI1
aI49
aI93
aI71
aI89
aI607
aI1143
aI3271
aI5825
aI8529
aI18479
aI23859
aI40571
asS'd'
I11705
sa(dp23410
S'a'
I50879
sS's'
I17
sS'm'
(lp23411
I1
aI3
aI1
aI3
aI13
aI35
aI79
aI9
aI315
aI943
aI1199
aI1521
aI2023
aI11745
aI8273
aI27643
aI89545
asS'd'
I11706
sa(dp23412
S'a'
I50882
sS's'
I17
sS'm'
(lp23413
I1
aI3
aI5
aI1
aI21
aI3
aI15
aI111
aI19
aI895
aI1539
aI3331
aI6741
aI9087
aI5231
aI13435
aI60645
asS'd'
I11707
sa(dp23414
S'a'
I50894
sS's'
I17
sS'm'
(lp23415
I1
aI3
aI1
aI9
aI25
aI9
aI109
aI253
aI263
aI425
aI915
aI1955
aI659
aI6249
aI11803
aI34523
aI22645
asS'd'
I11708
sa(dp23416
S'a'
I50899
sS's'
I17
sS'm'
(lp23417
I1
aI1
aI1
aI9
aI15
aI23
aI13
aI79
aI369
aI689
aI565
aI743
aI3897
aI8837
aI13753
aI17213
aI86801
asS'd'
I11709
sa(dp23418
S'a'
I50905
sS's'
I17
sS'm'
(lp23419
I1
aI1
aI1
aI5
aI31
aI27
aI111
aI231
aI25
aI617
aI897
aI1325
aI4885
aI5731
aI2027
aI34639
aI67863
asS'd'
I11710
sa(dp23420
S'a'
I50924
sS's'
I17
sS'm'
(lp23421
I1
aI1
aI3
aI13
aI1
aI9
aI29
aI23
aI95
aI113
aI1035
aI2729
aI6555
aI335
aI24795
aI35387
aI31301
asS'd'
I11711
sa(dp23422
S'a'
I50936
sS's'
I17
sS'm'
(lp23423
I1
aI1
aI1
aI7
aI3
aI53
aI91
aI143
aI167
aI773
aI207
aI31
aI4993
aI7953
aI26997
aI40031
aI113939
asS'd'
I11712
sa(dp23424
S'a'
I50944
sS's'
I17
sS'm'
(lp23425
I1
aI1
aI3
aI5
aI17
aI43
aI121
aI231
aI411
aI575
aI1621
aI3079
aI535
aI3313
aI19443
aI58271
aI54207
asS'd'
I11713
sa(dp23426
S'a'
I50973
sS's'
I17
sS'm'
(lp23427
I1
aI1
aI1
aI7
aI11
aI41
aI61
aI81
aI97
aI91
aI1987
aI981
aI3745
aI819
aI23785
aI48331
aI63761
asS'd'
I11714
sa(dp23428
S'a'
I50974
sS's'
I17
sS'm'
(lp23429
I1
aI1
aI5
aI13
aI7
aI29
aI25
aI73
aI355
aI669
aI241
aI65
aI2249
aI5551
aI5217
aI58573
aI34049
asS'd'
I11715
sa(dp23430
S'a'
I51002
sS's'
I17
sS'm'
(lp23431
I1
aI1
aI7
aI11
aI1
aI45
aI125
aI107
aI127
aI639
aI1989
aI2727
aI2103
aI8985
aI30249
aI40037
aI40931
asS'd'
I11716
sa(dp23432
S'a'
I51007
sS's'
I17
sS'm'
(lp23433
I1
aI1
aI5
aI13
aI21
aI59
aI99
aI131
aI359
aI615
aI665
aI577
aI2559
aI3555
aI11355
aI26213
aI76427
asS'd'
I11717
sa(dp23434
S'a'
I51009
sS's'
I17
sS'm'
(lp23435
I1
aI3
aI3
aI1
aI19
aI9
aI85
aI111
aI381
aI661
aI561
aI3419
aI1355
aI8473
aI329
aI4989
aI9087
asS'd'
I11718
sa(dp23436
S'a'
I51029
sS's'
I17
sS'm'
(lp23437
I1
aI1
aI7
aI5
aI11
aI23
aI33
aI95
aI343
aI9
aI1579
aI2663
aI6245
aI267
aI7187
aI25381
aI103181
asS'd'
I11719
sa(dp23438
S'a'
I51036
sS's'
I17
sS'm'
(lp23439
I1
aI1
aI7
aI11
aI23
aI7
aI113
aI49
aI89
aI587
aI1221
aI409
aI873
aI15531
aI2721
aI44519
aI25349
asS'd'
I11720
sa(dp23440
S'a'
I51063
sS's'
I17
sS'm'
(lp23441
I1
aI3
aI3
aI1
aI17
aI17
aI45
aI239
aI453
aI639
aI1433
aI2829
aI6009
aI12447
aI9393
aI59979
aI93343
asS'd'
I11721
sa(dp23442
S'a'
I51067
sS's'
I17
sS'm'
(lp23443
I1
aI1
aI3
aI15
aI19
aI61
aI125
aI101
aI219
aI327
aI1551
aI1623
aI951
aI8379
aI21009
aI64089
aI21891
asS'd'
I11722
sa(dp23444
S'a'
I51070
sS's'
I17
sS'm'
(lp23445
I1
aI1
aI5
aI7
aI23
aI5
aI111
aI43
aI57
aI71
aI407
aI4027
aI4869
aI12033
aI19941
aI51969
aI120115
asS'd'
I11723
sa(dp23446
S'a'
I51074
sS's'
I17
sS'm'
(lp23447
I1
aI3
aI7
aI15
aI17
aI49
aI31
aI145
aI185
aI169
aI1559
aI4011
aI5293
aI7559
aI23487
aI12213
aI2757
asS'd'
I11724
sa(dp23448
S'a'
I51079
sS's'
I17
sS'm'
(lp23449
I1
aI3
aI7
aI3
aI3
aI59
aI119
aI3
aI509
aI539
aI1623
aI539
aI1405
aI3913
aI213
aI30293
aI68497
asS'd'
I11725
sa(dp23450
S'a'
I51083
sS's'
I17
sS'm'
(lp23451
I1
aI1
aI1
aI9
aI15
aI43
aI67
aI171
aI397
aI233
aI379
aI1681
aI6877
aI9169
aI19667
aI1971
aI115347
asS'd'
I11726
sa(dp23452
S'a'
I51093
sS's'
I17
sS'm'
(lp23453
I1
aI1
aI5
aI15
aI1
aI45
aI25
aI133
aI99
aI181
aI1825
aI3485
aI5633
aI4629
aI30181
aI15761
aI87161
asS'd'
I11727
sa(dp23454
S'a'
I51094
sS's'
I17
sS'm'
(lp23455
I1
aI1
aI5
aI3
aI5
aI55
aI97
aI117
aI303
aI591
aI733
aI3631
aI4305
aI169
aI5361
aI64491
aI71793
asS'd'
I11728
sa(dp23456
S'a'
I51124
sS's'
I17
sS'm'
(lp23457
I1
aI1
aI5
aI11
aI19
aI9
aI5
aI147
aI223
aI51
aI1763
aI3899
aI7393
aI8107
aI19619
aI60509
aI61427
asS'd'
I11729
sa(dp23458
S'a'
I51131
sS's'
I17
sS'm'
(lp23459
I1
aI1
aI1
aI15
aI15
aI3
aI103
aI15
aI423
aI649
aI613
aI1387
aI6229
aI4147
aI17517
aI225
aI35697
asS'd'
I11730
sa(dp23460
S'a'
I51151
sS's'
I17
sS'm'
(lp23461
I1
aI3
aI1
aI3
aI21
aI57
aI77
aI193
aI203
aI649
aI631
aI3753
aI4259
aI3983
aI27707
aI33623
aI83857
asS'd'
I11731
sa(dp23462
S'a'
I51153
sS's'
I17
sS'm'
(lp23463
I1
aI3
aI3
aI5
aI11
aI37
aI95
aI201
aI83
aI643
aI1639
aI153
aI7683
aI15249
aI23859
aI27021
aI43321
asS'd'
I11732
sa(dp23464
S'a'
I51156
sS's'
I17
sS'm'
(lp23465
I1
aI3
aI5
aI13
aI23
aI31
aI69
aI215
aI303
aI433
aI1325
aI1013
aI2903
aI12659
aI3813
aI34497
aI59651
asS'd'
I11733
sa(dp23466
S'a'
I51165
sS's'
I17
sS'm'
(lp23467
I1
aI3
aI1
aI9
aI15
aI39
aI113
aI253
aI173
aI393
aI19
aI2343
aI2939
aI8871
aI29741
aI2141
aI121675
asS'd'
I11734
sa(dp23468
S'a'
I51175
sS's'
I17
sS'm'
(lp23469
I1
aI1
aI7
aI9
aI7
aI9
aI91
aI211
aI169
aI299
aI507
aI2849
aI1321
aI15397
aI23949
aI32387
aI108691
asS'd'
I11735
sa(dp23470
S'a'
I51184
sS's'
I17
sS'm'
(lp23471
I1
aI1
aI7
aI13
aI1
aI21
aI119
aI127
aI229
aI253
aI39
aI323
aI1831
aI121
aI17385
aI45511
aI43743
asS'd'
I11736
sa(dp23472
S'a'
I51187
sS's'
I17
sS'm'
(lp23473
I1
aI1
aI1
aI15
aI25
aI25
aI97
aI209
aI375
aI945
aI1343
aI2205
aI1701
aI13085
aI25547
aI55555
aI129395
asS'd'
I11737
sa(dp23474
S'a'
I51203
sS's'
I17
sS'm'
(lp23475
I1
aI1
aI1
aI5
aI31
aI25
aI91
aI255
aI163
aI169
aI703
aI1607
aI4731
aI7413
aI10013
aI10925
aI109151
asS'd'
I11738
sa(dp23476
S'a'
I51220
sS's'
I17
sS'm'
(lp23477
I1
aI3
aI3
aI9
aI15
aI47
aI71
aI219
aI9
aI37
aI231
aI3227
aI3447
aI8129
aI23421
aI30113
aI120725
asS'd'
I11739
sa(dp23478
S'a'
I51224
sS's'
I17
sS'm'
(lp23479
I1
aI3
aI3
aI11
aI15
aI47
aI93
aI203
aI299
aI865
aI151
aI3999
aI1245
aI8487
aI13355
aI27373
aI93583
asS'd'
I11740
sa(dp23480
S'a'
I51233
sS's'
I17
sS'm'
(lp23481
I1
aI3
aI7
aI5
aI13
aI7
aI97
aI81
aI271
aI95
aI513
aI365
aI5039
aI403
aI5285
aI29475
aI129347
asS'd'
I11741
sa(dp23482
S'a'
I51234
sS's'
I17
sS'm'
(lp23483
I1
aI1
aI7
aI7
aI9
aI27
aI25
aI207
aI161
aI785
aI1453
aI3031
aI763
aI2313
aI29347
aI61457
aI52561
asS'd'
I11742
sa(dp23484
S'a'
I51239
sS's'
I17
sS'm'
(lp23485
I1
aI3
aI3
aI11
aI25
aI25
aI39
aI61
aI165
aI803
aI1435
aI3643
aI299
aI13751
aI24239
aI53955
aI94889
asS'd'
I11743
sa(dp23486
S'a'
I51246
sS's'
I17
sS'm'
(lp23487
I1
aI3
aI5
aI9
aI9
aI13
aI63
aI221
aI123
aI947
aI905
aI913
aI953
aI7429
aI25409
aI43017
aI2217
asS'd'
I11744
sa(dp23488
S'a'
I51248
sS's'
I17
sS'm'
(lp23489
I1
aI1
aI3
aI3
aI31
aI19
aI107
aI211
aI503
aI675
aI1921
aI3027
aI1273
aI5699
aI20683
aI55605
aI119803
asS'd'
I11745
sa(dp23490
S'a'
I51251
sS's'
I17
sS'm'
(lp23491
I1
aI1
aI3
aI3
aI9
aI17
aI115
aI183
aI325
aI259
aI2013
aI1505
aI6999
aI11573
aI5315
aI18731
aI9405
asS'd'
I11746
sa(dp23492
S'a'
I51257
sS's'
I17
sS'm'
(lp23493
I1
aI3
aI1
aI5
aI29
aI37
aI81
aI145
aI5
aI851
aI1803
aI2011
aI6655
aI3601
aI11325
aI17137
aI68501
asS'd'
I11747
sa(dp23494
S'a'
I51266
sS's'
I17
sS'm'
(lp23495
I1
aI3
aI5
aI1
aI25
aI51
aI111
aI19
aI75
aI765
aI1843
aI139
aI7253
aI12967
aI13387
aI48631
aI124881
asS'd'
I11748
sa(dp23496
S'a'
I51280
sS's'
I17
sS'm'
(lp23497
I1
aI3
aI3
aI7
aI15
aI29
aI7
aI231
aI13
aI901
aI1913
aI3817
aI3993
aI3049
aI32367
aI4201
aI90837
asS'd'
I11749
sa(dp23498
S'a'
I51285
sS's'
I17
sS'm'
(lp23499
I1
aI3
aI1
aI11
aI27
aI5
aI109
aI57
aI81
aI147
aI1141
aI2153
aI5255
aI6367
aI189
aI5959
aI88843
asS'd'
I11750
sa(dp23500
S'a'
I51286
sS's'
I17
sS'm'
(lp23501
I1
aI3
aI5
aI5
aI19
aI35
aI17
aI149
aI407
aI889
aI1583
aI1727
aI465
aI10785
aI6043
aI21785
aI80935
asS'd'
I11751
sa(dp23502
S'a'
I51289
sS's'
I17
sS'm'
(lp23503
I1
aI1
aI3
aI7
aI15
aI21
aI105
aI249
aI427
aI491
aI923
aI3189
aI8103
aI3875
aI18347
aI35799
aI36703
asS'd'
I11752
sa(dp23504
S'a'
I51295
sS's'
I17
sS'm'
(lp23505
I1
aI1
aI3
aI15
aI3
aI45
aI93
aI197
aI265
aI309
aI1909
aI1635
aI1743
aI9499
aI21897
aI36889
aI67449
asS'd'
I11753
sa(dp23506
S'a'
I51296
sS's'
I17
sS'm'
(lp23507
I1
aI3
aI1
aI11
aI15
aI57
aI31
aI231
aI363
aI879
aI1377
aI1941
aI6969
aI10721
aI21933
aI33419
aI102939
asS'd'
I11754
sa(dp23508
S'a'
I51311
sS's'
I17
sS'm'
(lp23509
I1
aI1
aI3
aI9
aI3
aI57
aI49
aI51
aI71
aI991
aI885
aI1367
aI2937
aI9301
aI29893
aI9867
aI113711
asS'd'
I11755
sa(dp23510
S'a'
I51329
sS's'
I17
sS'm'
(lp23511
I1
aI3
aI7
aI7
aI11
aI59
aI123
aI247
aI51
aI659
aI1323
aI3371
aI3417
aI12295
aI2021
aI62753
aI28059
asS'd'
I11756
sa(dp23512
S'a'
I51350
sS's'
I17
sS'm'
(lp23513
I1
aI1
aI7
aI11
aI3
aI57
aI53
aI1
aI203
aI287
aI219
aI3531
aI1365
aI6235
aI30187
aI4479
aI29567
asS'd'
I11757
sa(dp23514
S'a'
I51356
sS's'
I17
sS'm'
(lp23515
I1
aI1
aI7
aI9
aI5
aI41
aI41
aI39
aI137
aI495
aI149
aI2421
aI7365
aI11381
aI25403
aI16063
aI47491
asS'd'
I11758
sa(dp23516
S'a'
I51363
sS's'
I17
sS'm'
(lp23517
I1
aI1
aI5
aI13
aI25
aI25
aI47
aI25
aI213
aI661
aI1345
aI883
aI7573
aI3291
aI21303
aI8033
aI102639
asS'd'
I11759
sa(dp23518
S'a'
I51365
sS's'
I17
sS'm'
(lp23519
I1
aI1
aI5
aI3
aI9
aI49
aI75
aI221
aI455
aI139
aI1533
aI3155
aI1023
aI7249
aI10129
aI63165
aI1713
asS'd'
I11760
sa(dp23520
S'a'
I51370
sS's'
I17
sS'm'
(lp23521
I1
aI3
aI5
aI11
aI17
aI1
aI23
aI241
aI83
aI359
aI1243
aI2791
aI2975
aI6271
aI19035
aI55057
aI7625
asS'd'
I11761
sa(dp23522
S'a'
I51372
sS's'
I17
sS'm'
(lp23523
I1
aI3
aI1
aI9
aI17
aI61
aI109
aI255
aI447
aI939
aI899
aI551
aI7049
aI4247
aI17333
aI43369
aI30105
asS'd'
I11762
sa(dp23524
S'a'
I51377
sS's'
I17
sS'm'
(lp23525
I1
aI1
aI3
aI5
aI3
aI31
aI79
aI39
aI225
aI605
aI1893
aI3523
aI5391
aI6879
aI18619
aI2339
aI108295
asS'd'
I11763
sa(dp23526
S'a'
I51383
sS's'
I17
sS'm'
(lp23527
I1
aI1
aI5
aI1
aI29
aI15
aI123
aI39
aI239
aI57
aI843
aI2799
aI4755
aI4993
aI23383
aI45559
aI48359
asS'd'
I11764
sa(dp23528
S'a'
I51384
sS's'
I17
sS'm'
(lp23529
I1
aI1
aI5
aI5
aI7
aI5
aI99
aI1
aI351
aI213
aI1061
aI721
aI343
aI16071
aI29641
aI55607
aI76727
asS'd'
I11765
sa(dp23530
S'a'
I51397
sS's'
I17
sS'm'
(lp23531
I1
aI3
aI1
aI7
aI9
aI9
aI15
aI25
aI87
aI595
aI71
aI3769
aI2583
aI10105
aI28639
aI48899
aI49753
asS'd'
I11766
sa(dp23532
S'a'
I51407
sS's'
I17
sS'm'
(lp23533
I1
aI3
aI5
aI3
aI31
aI29
aI99
aI77
aI323
aI615
aI581
aI1725
aI2471
aI16263
aI4903
aI205
aI55441
asS'd'
I11767
sa(dp23534
S'a'
I51422
sS's'
I17
sS'm'
(lp23535
I1
aI1
aI5
aI11
aI17
aI53
aI47
aI53
aI125
aI717
aI867
aI1251
aI4009
aI13981
aI10165
aI4769
aI117431
asS'd'
I11768
sa(dp23536
S'a'
I51435
sS's'
I17
sS'm'
(lp23537
I1
aI1
aI3
aI7
aI7
aI19
aI119
aI27
aI163
aI11
aI693
aI3197
aI3981
aI12591
aI26017
aI62113
aI48391
asS'd'
I11769
sa(dp23538
S'a'
I51440
sS's'
I17
sS'm'
(lp23539
I1
aI1
aI3
aI15
aI13
aI17
aI13
aI81
aI19
aI821
aI677
aI233
aI5227
aI14855
aI18269
aI18895
aI90041
asS'd'
I11770
sa(dp23540
S'a'
I51446
sS's'
I17
sS'm'
(lp23541
I1
aI1
aI3
aI9
aI27
aI61
aI125
aI221
aI415
aI183
aI1137
aI1879
aI3451
aI3599
aI27317
aI53449
aI35499
asS'd'
I11771
sa(dp23542
S'a'
I51463
sS's'
I17
sS'm'
(lp23543
I1
aI3
aI3
aI15
aI3
aI27
aI53
aI93
aI17
aI405
aI373
aI2531
aI3121
aI2299
aI1593
aI34623
aI102389
asS'd'
I11772
sa(dp23544
S'a'
I51467
sS's'
I17
sS'm'
(lp23545
I1
aI1
aI1
aI11
aI23
aI19
aI85
aI87
aI209
aI17
aI1805
aI4067
aI7401
aI6097
aI5865
aI61383
aI42971
asS'd'
I11773
sa(dp23546
S'a'
I51491
sS's'
I17
sS'm'
(lp23547
I1
aI1
aI7
aI9
aI29
aI43
aI99
aI125
aI385
aI347
aI97
aI1121
aI1533
aI10545
aI17383
aI48649
aI78443
asS'd'
I11774
sa(dp23548
S'a'
I51493
sS's'
I17
sS'm'
(lp23549
I1
aI3
aI3
aI13
aI7
aI9
aI95
aI105
aI463
aI911
aI357
aI423
aI5701
aI2389
aI16307
aI17817
aI108775
asS'd'
I11775
sa(dp23550
S'a'
I51494
sS's'
I17
sS'm'
(lp23551
I1
aI3
aI5
aI11
aI21
aI21
aI79
aI53
aI511
aI995
aI1709
aI1715
aI6031
aI10507
aI10735
aI48817
aI28569
asS'd'
I11776
sa(dp23552
S'a'
I51508
sS's'
I17
sS'm'
(lp23553
I1
aI3
aI5
aI7
aI31
aI49
aI43
aI109
aI267
aI981
aI1529
aI3611
aI3379
aI1295
aI27489
aI46721
aI58423
asS'd'
I11777
sa(dp23554
S'a'
I51518
sS's'
I17
sS'm'
(lp23555
I1
aI3
aI3
aI11
aI3
aI31
aI21
aI3
aI79
aI31
aI1885
aI3029
aI6337
aI15457
aI8995
aI9955
aI95019
asS'd'
I11778
sa(dp23556
S'a'
I51520
sS's'
I17
sS'm'
(lp23557
I1
aI1
aI7
aI13
aI9
aI9
aI77
aI73
aI111
aI769
aI813
aI1599
aI5925
aI1063
aI12151
aI54125
aI67723
asS'd'
I11779
sa(dp23558
S'a'
I51526
sS's'
I17
sS'm'
(lp23559
I1
aI1
aI1
aI7
aI13
aI5
aI43
aI201
aI379
aI199
aI769
aI3227
aI3995
aI1543
aI21903
aI10651
aI122007
asS'd'
I11780
sa(dp23560
S'a'
I51529
sS's'
I17
sS'm'
(lp23561
I1
aI3
aI7
aI13
aI11
aI53
aI83
aI201
aI231
aI137
aI617
aI2395
aI2513
aI6659
aI9387
aI15653
aI96927
asS'd'
I11781
sa(dp23562
S'a'
I51530
sS's'
I17
sS'm'
(lp23563
I1
aI3
aI1
aI13
aI29
aI19
aI97
aI57
aI231
aI985
aI805
aI1169
aI831
aI15867
aI20195
aI53533
aI99735
asS'd'
I11782
sa(dp23564
S'a'
I51537
sS's'
I17
sS'm'
(lp23565
I1
aI1
aI3
aI13
aI19
aI15
aI19
aI39
aI99
aI31
aI421
aI755
aI7439
aI4927
aI19893
aI15449
aI47937
asS'd'
I11783
sa(dp23566
S'a'
I51547
sS's'
I17
sS'm'
(lp23567
I1
aI3
aI3
aI9
aI1
aI17
aI71
aI37
aI289
aI537
aI69
aI3687
aI6537
aI12295
aI28403
aI6627
aI72991
asS'd'
I11784
sa(dp23568
S'a'
I51559
sS's'
I17
sS'm'
(lp23569
I1
aI3
aI3
aI15
aI31
aI53
aI21
aI223
aI451
aI309
aI895
aI3923
aI3149
aI5167
aI1979
aI31425
aI53485
asS'd'
I11785
sa(dp23570
S'a'
I51565
sS's'
I17
sS'm'
(lp23571
I1
aI3
aI1
aI1
aI29
aI7
aI5
aI197
aI445
aI455
aI185
aI633
aI897
aI4561
aI28833
aI39477
aI46665
asS'd'
I11786
sa(dp23572
S'a'
I51568
sS's'
I17
sS'm'
(lp23573
I1
aI1
aI3
aI9
aI29
aI19
aI45
aI239
aI323
aI1005
aI101
aI2083
aI7403
aI10401
aI987
aI32301
aI58141
asS'd'
I11787
sa(dp23574
S'a'
I51580
sS's'
I17
sS'm'
(lp23575
I1
aI3
aI3
aI5
aI31
aI17
aI7
aI141
aI245
aI301
aI1607
aI3381
aI7517
aI6663
aI6327
aI15321
aI19963
asS'd'
I11788
sa(dp23576
S'a'
I51583
sS's'
I17
sS'm'
(lp23577
I1
aI1
aI7
aI5
aI5
aI37
aI109
aI31
aI285
aI767
aI1689
aI2961
aI5511
aI15415
aI32011
aI14889
aI7237
asS'd'
I11789
sa(dp23578
S'a'
I51584
sS's'
I17
sS'm'
(lp23579
I1
aI1
aI3
aI7
aI31
aI35
aI47
aI157
aI407
aI719
aI1213
aI1241
aI2475
aI10321
aI11547
aI52641
aI21603
asS'd'
I11790
sa(dp23580
S'a'
I51593
sS's'
I17
sS'm'
(lp23581
I1
aI1
aI1
aI7
aI5
aI45
aI35
aI137
aI403
aI321
aI1151
aI529
aI6297
aI3059
aI27791
aI18387
aI101431
asS'd'
I11791
sa(dp23582
S'a'
I51594
sS's'
I17
sS'm'
(lp23583
I1
aI1
aI3
aI11
aI21
aI19
aI97
aI121
aI377
aI325
aI741
aI1601
aI1115
aI6233
aI19089
aI40491
aI53259
asS'd'
I11792
sa(dp23584
S'a'
I51607
sS's'
I17
sS'm'
(lp23585
I1
aI1
aI1
aI9
aI3
aI13
aI83
aI243
aI443
aI91
aI1455
aI1875
aI3327
aI7245
aI12735
aI14943
aI44163
asS'd'
I11793
sa(dp23586
S'a'
I51608
sS's'
I17
sS'm'
(lp23587
I1
aI3
aI5
aI11
aI1
aI15
aI107
aI211
aI25
aI125
aI623
aI1319
aI6133
aI12177
aI1377
aI32547
aI110919
asS'd'
I11794
sa(dp23588
S'a'
I51620
sS's'
I17
sS'm'
(lp23589
I1
aI1
aI3
aI3
aI7
aI39
aI23
aI99
aI433
aI581
aI53
aI3421
aI733
aI12767
aI23595
aI22957
aI88821
asS'd'
I11795
sa(dp23590
S'a'
I51624
sS's'
I17
sS'm'
(lp23591
I1
aI1
aI1
aI13
aI5
aI53
aI103
aI127
aI409
aI379
aI1155
aI3097
aI6529
aI11685
aI22147
aI46003
aI59771
asS'd'
I11796
sa(dp23592
S'a'
I51642
sS's'
I17
sS'm'
(lp23593
I1
aI3
aI3
aI3
aI15
aI37
aI9
aI67
aI237
aI79
aI697
aI1943
aI1021
aI3217
aI16013
aI14727
aI105729
asS'd'
I11797
sa(dp23594
S'a'
I51649
sS's'
I17
sS'm'
(lp23595
I1
aI1
aI7
aI1
aI9
aI43
aI55
aI79
aI63
aI553
aI871
aI2881
aI6487
aI4667
aI24263
aI41995
aI60907
asS'd'
I11798
sa(dp23596
S'a'
I51652
sS's'
I17
sS'm'
(lp23597
I1
aI1
aI7
aI7
aI23
aI31
aI55
aI23
aI451
aI593
aI85
aI43
aI965
aI12491
aI15121
aI16129
aI90639
asS'd'
I11799
sa(dp23598
S'a'
I51659
sS's'
I17
sS'm'
(lp23599
I1
aI1
aI7
aI13
aI11
aI53
aI21
aI123
aI237
aI147
aI511
aI2105
aI5961
aI4465
aI4015
aI19069
aI89203
asS'd'
I11800
sa(dp23600
S'a'
I51661
sS's'
I17
sS'm'
(lp23601
I1
aI3
aI5
aI1
aI1
aI39
aI59
aI239
aI391
aI91
aI923
aI85
aI1047
aI1489
aI31119
aI58485
aI129171
asS'd'
I11801
sa(dp23602
S'a'
I51670
sS's'
I17
sS'm'
(lp23603
I1
aI3
aI5
aI13
aI3
aI21
aI35
aI205
aI219
aI795
aI901
aI2465
aI5887
aI275
aI22003
aI29659
aI50589
asS'd'
I11802
sa(dp23604
S'a'
I51674
sS's'
I17
sS'm'
(lp23605
I1
aI3
aI1
aI5
aI1
aI35
aI127
aI147
aI159
aI791
aI1643
aI1811
aI1199
aI3851
aI9681
aI19845
aI127075
asS'd'
I11803
sa(dp23606
S'a'
I51698
sS's'
I17
sS'm'
(lp23607
I1
aI3
aI3
aI7
aI17
aI19
aI13
aI223
aI395
aI971
aI125
aI181
aI4455
aI13305
aI30433
aI46161
aI122277
asS'd'
I11804
sa(dp23608
S'a'
I51707
sS's'
I17
sS'm'
(lp23609
I1
aI3
aI1
aI3
aI19
aI1
aI15
aI71
aI425
aI459
aI655
aI2251
aI5525
aI7611
aI5819
aI1255
aI43107
asS'd'
I11805
sa(dp23610
S'a'
I51725
sS's'
I17
sS'm'
(lp23611
I1
aI1
aI5
aI15
aI3
aI9
aI83
aI191
aI439
aI663
aI399
aI2263
aI1857
aI15421
aI2079
aI2381
aI59193
asS'd'
I11806
sa(dp23612
S'a'
I51737
sS's'
I17
sS'm'
(lp23613
I1
aI3
aI5
aI5
aI1
aI7
aI9
aI77
aI347
aI419
aI1329
aI3173
aI7295
aI3631
aI13435
aI3217
aI18053
asS'd'
I11807
sa(dp23614
S'a'
I51753
sS's'
I17
sS'm'
(lp23615
I1
aI3
aI3
aI7
aI29
aI39
aI35
aI71
aI119
aI745
aI603
aI2247
aI377
aI3297
aI30983
aI27857
aI105739
asS'd'
I11808
sa(dp23616
S'a'
I51754
sS's'
I17
sS'm'
(lp23617
I1
aI1
aI5
aI13
aI9
aI59
aI57
aI239
aI247
aI921
aI1383
aI2315
aI241
aI4435
aI24661
aI6129
aI122727
asS'd'
I11809
sa(dp23618
S'a'
I51779
sS's'
I17
sS'm'
(lp23619
I1
aI1
aI5
aI7
aI3
aI15
aI39
aI165
aI27
aI803
aI609
aI3081
aI6009
aI12665
aI24155
aI51647
aI3857
asS'd'
I11810
sa(dp23620
S'a'
I51791
sS's'
I17
sS'm'
(lp23621
I1
aI3
aI5
aI11
aI27
aI41
aI61
aI255
aI195
aI169
aI557
aI1739
aI4029
aI1791
aI471
aI16321
aI49801
asS'd'
I11811
sa(dp23622
S'a'
I51796
sS's'
I17
sS'm'
(lp23623
I1
aI3
aI7
aI13
aI17
aI45
aI35
aI177
aI109
aI113
aI551
aI219
aI4065
aI303
aI15489
aI25427
aI12349
asS'd'
I11812
sa(dp23624
S'a'
I51809
sS's'
I17
sS'm'
(lp23625
I1
aI3
aI5
aI5
aI25
aI15
aI79
aI165
aI231
aI867
aI483
aI3563
aI6611
aI11277
aI3193
aI37455
aI127373
asS'd'
I11813
sa(dp23626
S'a'
I51816
sS's'
I17
sS'm'
(lp23627
I1
aI3
aI3
aI11
aI25
aI21
aI47
aI255
aI27
aI543
aI485
aI2675
aI5893
aI3029
aI3857
aI50967
aI14681
asS'd'
I11814
sa(dp23628
S'a'
I51819
sS's'
I17
sS'm'
(lp23629
I1
aI3
aI7
aI11
aI3
aI23
aI81
aI135
aI77
aI227
aI417
aI1733
aI5175
aI15295
aI15985
aI50329
aI48641
asS'd'
I11815
sa(dp23630
S'a'
I51827
sS's'
I17
sS'm'
(lp23631
I1
aI3
aI5
aI1
aI3
aI47
aI33
aI67
aI201
aI235
aI1299
aI3703
aI1959
aI8091
aI11115
aI10869
aI9595
asS'd'
I11816
sa(dp23632
S'a'
I51829
sS's'
I17
sS'm'
(lp23633
I1
aI3
aI7
aI13
aI1
aI45
aI61
aI49
aI471
aI923
aI1827
aI2175
aI1433
aI3473
aI3781
aI7923
aI121525
asS'd'
I11817
sa(dp23634
S'a'
I51834
sS's'
I17
sS'm'
(lp23635
I1
aI3
aI5
aI7
aI25
aI59
aI113
aI19
aI415
aI839
aI167
aI3549
aI7435
aI6573
aI767
aI2751
aI18383
asS'd'
I11818
sa(dp23636
S'a'
I51836
sS's'
I17
sS'm'
(lp23637
I1
aI3
aI3
aI1
aI5
aI11
aI125
aI241
aI395
aI423
aI955
aI2551
aI963
aI8197
aI30253
aI10473
aI28505
asS'd'
I11819
sa(dp23638
S'a'
I51846
sS's'
I17
sS'm'
(lp23639
I1
aI1
aI1
aI3
aI5
aI31
aI69
aI103
aI153
aI505
aI1507
aI2827
aI165
aI4943
aI8343
aI54253
aI87593
asS'd'
I11820
sa(dp23640
S'a'
I51849
sS's'
I17
sS'm'
(lp23641
I1
aI3
aI3
aI1
aI7
aI37
aI27
aI75
aI251
aI623
aI965
aI1907
aI6063
aI761
aI765
aI10103
aI43479
asS'd'
I11821
sa(dp23642
S'a'
I51855
sS's'
I17
sS'm'
(lp23643
I1
aI1
aI1
aI13
aI7
aI21
aI53
aI219
aI267
aI57
aI959
aI579
aI2951
aI13797
aI3249
aI29895
aI47467
asS'd'
I11822
sa(dp23644
S'a'
I51858
sS's'
I17
sS'm'
(lp23645
I1
aI1
aI5
aI9
aI13
aI7
aI85
aI107
aI3
aI635
aI1235
aI1339
aI3263
aI3895
aI25911
aI7521
aI34149
asS'd'
I11823
sa(dp23646
S'a'
I51869
sS's'
I17
sS'm'
(lp23647
I1
aI1
aI5
aI15
aI29
aI37
aI73
aI43
aI413
aI993
aI499
aI719
aI1557
aI14397
aI11245
aI58197
aI127901
asS'd'
I11824
sa(dp23648
S'a'
I51870
sS's'
I17
sS'm'
(lp23649
I1
aI1
aI5
aI15
aI9
aI37
aI87
aI57
aI63
aI337
aI927
aI1887
aI6407
aI11237
aI23233
aI53567
aI120449
asS'd'
I11825
sa(dp23650
S'a'
I51874
sS's'
I17
sS'm'
(lp23651
I1
aI1
aI3
aI7
aI27
aI11
aI85
aI227
aI159
aI849
aI647
aI1977
aI4623
aI7343
aI8089
aI4251
aI26609
asS'd'
I11826
sa(dp23652
S'a'
I51918
sS's'
I17
sS'm'
(lp23653
I1
aI1
aI5
aI11
aI11
aI3
aI105
aI191
aI189
aI627
aI367
aI3935
aI6647
aI13069
aI26195
aI23137
aI56427
asS'd'
I11827
sa(dp23654
S'a'
I51926
sS's'
I17
sS'm'
(lp23655
I1
aI1
aI3
aI3
aI29
aI51
aI39
aI3
aI437
aI1011
aI1061
aI1747
aI3051
aI11165
aI8155
aI9723
aI41035
asS'd'
I11828
sa(dp23656
S'a'
I51932
sS's'
I17
sS'm'
(lp23657
I1
aI3
aI7
aI15
aI9
aI43
aI79
aI195
aI265
aI395
aI1349
aI337
aI911
aI15767
aI3593
aI42859
aI70181
asS'd'
I11829
sa(dp23658
S'a'
I51936
sS's'
I17
sS'm'
(lp23659
I1
aI3
aI7
aI7
aI11
aI47
aI11
aI85
aI489
aI801
aI1177
aI3861
aI3517
aI9209
aI27505
aI12291
aI35691
asS'd'
I11830
sa(dp23660
S'a'
I51948
sS's'
I17
sS'm'
(lp23661
I1
aI1
aI3
aI9
aI15
aI61
aI71
aI123
aI287
aI419
aI1079
aI3489
aI3519
aI12739
aI21341
aI24323
aI33961
asS'd'
I11831
sa(dp23662
S'a'
I51954
sS's'
I17
sS'm'
(lp23663
I1
aI3
aI7
aI3
aI9
aI17
aI119
aI137
aI389
aI391
aI601
aI1875
aI2197
aI5271
aI13289
aI43597
aI43279
asS'd'
I11832
sa(dp23664
S'a'
I51959
sS's'
I17
sS'm'
(lp23665
I1
aI3
aI7
aI15
aI29
aI35
aI41
aI171
aI183
aI701
aI1673
aI981
aI5479
aI21
aI11353
aI64953
aI88189
asS'd'
I11833
sa(dp23666
S'a'
I51971
sS's'
I17
sS'm'
(lp23667
I1
aI3
aI3
aI13
aI15
aI35
aI35
aI81
aI297
aI245
aI475
aI393
aI5401
aI12369
aI21129
aI65213
aI125013
asS'd'
I11834
sa(dp23668
S'a'
I51983
sS's'
I17
sS'm'
(lp23669
I1
aI1
aI7
aI13
aI15
aI57
aI25
aI143
aI389
aI111
aI1219
aI2195
aI769
aI9005
aI10367
aI39435
aI3631
asS'd'
I11835
sa(dp23670
S'a'
I51992
sS's'
I17
sS'm'
(lp23671
I1
aI3
aI3
aI13
aI9
aI29
aI9
aI47
aI127
aI377
aI1195
aI1221
aI5751
aI15481
aI10537
aI29909
aI112691
asS'd'
I11836
sa(dp23672
S'a'
I51997
sS's'
I17
sS'm'
(lp23673
I1
aI1
aI3
aI5
aI17
aI47
aI29
aI1
aI299
aI651
aI1023
aI1601
aI5917
aI3781
aI18421
aI42393
aI51789
asS'd'
I11837
sa(dp23674
S'a'
I52016
sS's'
I17
sS'm'
(lp23675
I1
aI3
aI3
aI15
aI31
aI51
aI101
aI147
aI367
aI159
aI359
aI705
aI3773
aI8649
aI31373
aI5733
aI58287
asS'd'
I11838
sa(dp23676
S'a'
I52021
sS's'
I17
sS'm'
(lp23677
I1
aI3
aI5
aI11
aI11
aI51
aI55
aI79
aI147
aI917
aI1945
aI1725
aI289
aI12777
aI30099
aI3013
aI91527
asS'd'
I11839
sa(dp23678
S'a'
I52031
sS's'
I17
sS'm'
(lp23679
I1
aI1
aI7
aI13
aI1
aI51
aI33
aI27
aI169
aI573
aI117
aI2479
aI761
aI1283
aI32723
aI13589
aI88821
asS'd'
I11840
sa(dp23680
S'a'
I52033
sS's'
I17
sS'm'
(lp23681
I1
aI1
aI1
aI3
aI23
aI13
aI33
aI207
aI137
aI391
aI1309
aI4093
aI6889
aI827
aI9331
aI57113
aI110193
asS'd'
I11841
sa(dp23682
S'a'
I52034
sS's'
I17
sS'm'
(lp23683
I1
aI1
aI1
aI13
aI15
aI57
aI115
aI53
aI59
aI443
aI1
aI3545
aI6923
aI6603
aI8631
aI41703
aI8519
asS'd'
I11842
sa(dp23684
S'a'
I52048
sS's'
I17
sS'm'
(lp23685
I1
aI1
aI5
aI5
aI25
aI29
aI53
aI153
aI107
aI423
aI1829
aI2469
aI1843
aI889
aI31727
aI20701
aI6343
asS'd'
I11843
sa(dp23686
S'a'
I52053
sS's'
I17
sS'm'
(lp23687
I1
aI1
aI7
aI13
aI11
aI41
aI7
aI219
aI77
aI663
aI329
aI2639
aI1111
aI1293
aI16771
aI20731
aI46973
asS'd'
I11844
sa(dp23688
S'a'
I52057
sS's'
I17
sS'm'
(lp23689
I1
aI3
aI3
aI15
aI23
aI19
aI45
aI107
aI111
aI155
aI1595
aI1821
aI471
aI6089
aI21587
aI49259
aI85393
asS'd'
I11845
sa(dp23690
S'a'
I52067
sS's'
I17
sS'm'
(lp23691
I1
aI3
aI1
aI3
aI27
aI21
aI39
aI227
aI359
aI885
aI449
aI2615
aI3519
aI5377
aI16017
aI57159
aI82399
asS'd'
I11846
sa(dp23692
S'a'
I52076
sS's'
I17
sS'm'
(lp23693
I1
aI3
aI3
aI15
aI31
aI33
aI77
aI33
aI87
aI821
aI701
aI2859
aI1777
aI3007
aI16757
aI5447
aI3557
asS'd'
I11847
sa(dp23694
S'a'
I52079
sS's'
I17
sS'm'
(lp23695
I1
aI1
aI1
aI15
aI11
aI31
aI127
aI79
aI363
aI341
aI169
aI3451
aI6351
aI6867
aI13511
aI833
aI103151
asS'd'
I11848
sa(dp23696
S'a'
I52081
sS's'
I17
sS'm'
(lp23697
I1
aI3
aI5
aI7
aI27
aI23
aI5
aI67
aI159
aI535
aI103
aI843
aI8187
aI6891
aI19169
aI22565
aI95255
asS'd'
I11849
sa(dp23698
S'a'
I52109
sS's'
I17
sS'm'
(lp23699
I1
aI3
aI5
aI5
aI15
aI27
aI53
aI49
aI343
aI815
aI1203
aI1031
aI6359
aI1337
aI1629
aI47783
aI103391
asS'd'
I11850
sa(dp23700
S'a'
I52127
sS's'
I17
sS'm'
(lp23701
I1
aI3
aI1
aI1
aI13
aI19
aI51
aI185
aI45
aI549
aI619
aI2247
aI2541
aI10421
aI31507
aI60785
aI87139
asS'd'
I11851
sa(dp23702
S'a'
I52128
sS's'
I17
sS'm'
(lp23703
I1
aI3
aI1
aI15
aI29
aI9
aI47
aI127
aI41
aI767
aI1375
aI2183
aI7169
aI12855
aI15021
aI377
aI69327
asS'd'
I11852
sa(dp23704
S'a'
I52133
sS's'
I17
sS'm'
(lp23705
I1
aI1
aI3
aI9
aI5
aI23
aI23
aI203
aI101
aI809
aI1155
aI2885
aI3901
aI3081
aI1827
aI65373
aI106133
asS'd'
I11853
sa(dp23706
S'a'
I52148
sS's'
I17
sS'm'
(lp23707
I1
aI3
aI3
aI13
aI3
aI21
aI73
aI135
aI353
aI905
aI1757
aI1361
aI3801
aI15541
aI2261
aI17159
aI18037
asS'd'
I11854
sa(dp23708
S'a'
I52155
sS's'
I17
sS'm'
(lp23709
I1
aI1
aI7
aI7
aI27
aI23
aI57
aI33
aI225
aI407
aI1709
aI1159
aI6353
aI13341
aI15883
aI17339
aI50423
asS'd'
I11855
sa(dp23710
S'a'
I52157
sS's'
I17
sS'm'
(lp23711
I1
aI1
aI3
aI15
aI13
aI21
aI33
aI91
aI183
aI975
aI1623
aI3187
aI5495
aI8947
aI7031
aI19687
aI104483
asS'd'
I11856
sa(dp23712
S'a'
I52172
sS's'
I17
sS'm'
(lp23713
I1
aI3
aI5
aI7
aI17
aI25
aI77
aI1
aI335
aI85
aI1783
aI2617
aI4463
aI3807
aI12883
aI24487
aI66205
asS'd'
I11857
sa(dp23714
S'a'
I52178
sS's'
I17
sS'm'
(lp23715
I1
aI1
aI3
aI15
aI23
aI37
aI83
aI93
aI211
aI757
aI903
aI2681
aI49
aI435
aI21057
aI63635
aI36489
asS'd'
I11858
sa(dp23716
S'a'
I52184
sS's'
I17
sS'm'
(lp23717
I1
aI3
aI1
aI13
aI3
aI45
aI63
aI57
aI65
aI21
aI627
aI1467
aI51
aI15887
aI27465
aI59227
aI108233
asS'd'
I11859
sa(dp23718
S'a'
I52199
sS's'
I17
sS'm'
(lp23719
I1
aI1
aI7
aI5
aI15
aI41
aI53
aI57
aI301
aI677
aI803
aI1675
aI6937
aI3159
aI14281
aI22355
aI37783
asS'd'
I11860
sa(dp23720
S'a'
I52200
sS's'
I17
sS'm'
(lp23721
I1
aI1
aI5
aI13
aI15
aI43
aI39
aI245
aI191
aI875
aI1505
aI2085
aI3903
aI185
aI24461
aI28939
aI98771
asS'd'
I11861
sa(dp23722
S'a'
I52205
sS's'
I17
sS'm'
(lp23723
I1
aI1
aI7
aI9
aI17
aI25
aI29
aI31
aI439
aI159
aI533
aI3177
aI4155
aI403
aI23735
aI61817
aI121909
asS'd'
I11862
sa(dp23724
S'a'
I52206
sS's'
I17
sS'm'
(lp23725
I1
aI1
aI3
aI15
aI29
aI43
aI111
aI47
aI483
aI513
aI63
aI2423
aI4979
aI5159
aI15499
aI33391
aI51575
asS'd'
I11863
sa(dp23726
S'a'
I52232
sS's'
I17
sS'm'
(lp23727
I1
aI1
aI5
aI15
aI15
aI43
aI13
aI41
aI445
aI929
aI1365
aI2023
aI6173
aI6067
aI30969
aI51457
aI51179
asS'd'
I11864
sa(dp23728
S'a'
I52237
sS's'
I17
sS'm'
(lp23729
I1
aI3
aI3
aI11
aI15
aI17
aI93
aI235
aI159
aI599
aI635
aI1113
aI1017
aI7413
aI7737
aI20051
aI79127
asS'd'
I11865
sa(dp23730
S'a'
I52243
sS's'
I17
sS'm'
(lp23731
I1
aI1
aI1
aI15
aI5
aI19
aI81
aI65
aI479
aI119
aI1831
aI2515
aI2929
aI15395
aI31607
aI29969
aI49935
asS'd'
I11866
sa(dp23732
S'a'
I52246
sS's'
I17
sS'm'
(lp23733
I1
aI3
aI1
aI13
aI23
aI47
aI45
aI151
aI51
aI217
aI803
aI3265
aI5907
aI14371
aI8287
aI25659
aI27655
asS'd'
I11867
sa(dp23734
S'a'
I52252
sS's'
I17
sS'm'
(lp23735
I1
aI1
aI1
aI13
aI13
aI53
aI11
aI63
aI501
aI487
aI1683
aI1147
aI4693
aI2761
aI19359
aI2215
aI112393
asS'd'
I11868
sa(dp23736
S'a'
I52261
sS's'
I17
sS'm'
(lp23737
I1
aI3
aI3
aI1
aI31
aI15
aI61
aI237
aI129
aI119
aI135
aI39
aI6509
aI3753
aI16997
aI3841
aI24521
asS'd'
I11869
sa(dp23738
S'a'
I52265
sS's'
I17
sS'm'
(lp23739
I1
aI3
aI5
aI7
aI5
aI27
aI113
aI251
aI217
aI923
aI229
aI2259
aI5241
aI6331
aI6773
aI41929
aI89605
asS'd'
I11870
sa(dp23740
S'a'
I52266
sS's'
I17
sS'm'
(lp23741
I1
aI1
aI5
aI9
aI17
aI41
aI21
aI185
aI95
aI137
aI1151
aI195
aI913
aI531
aI15731
aI22893
aI93521
asS'd'
I11871
sa(dp23742
S'a'
I52273
sS's'
I17
sS'm'
(lp23743
I1
aI1
aI5
aI1
aI29
aI57
aI123
aI11
aI345
aI581
aI227
aI1539
aI7527
aI8537
aI16429
aI8437
aI18953
asS'd'
I11872
sa(dp23744
S'a'
I52274
sS's'
I17
sS'm'
(lp23745
I1
aI1
aI3
aI7
aI7
aI21
aI103
aI239
aI115
aI513
aI1287
aI3717
aI331
aI5453
aI18943
aI17247
aI64975
asS'd'
I11873
sa(dp23746
S'a'
I52288
sS's'
I17
sS'm'
(lp23747
I1
aI3
aI7
aI11
aI21
aI37
aI79
aI83
aI93
aI155
aI1297
aI1371
aI1109
aI6569
aI21137
aI29237
aI24007
asS'd'
I11874
sa(dp23748
S'a'
I52300
sS's'
I17
sS'm'
(lp23749
I1
aI1
aI1
aI13
aI17
aI11
aI127
aI89
aI397
aI497
aI1017
aI721
aI2837
aI5623
aI31745
aI52243
aI107225
asS'd'
I11875
sa(dp23750
S'a'
I52303
sS's'
I17
sS'm'
(lp23751
I1
aI1
aI7
aI9
aI15
aI43
aI29
aI215
aI449
aI233
aI313
aI2587
aI2903
aI2735
aI4539
aI50481
aI85279
asS'd'
I11876
sa(dp23752
S'a'
I52321
sS's'
I17
sS'm'
(lp23753
I1
aI1
aI3
aI15
aI13
aI35
aI109
aI211
aI299
aI255
aI1595
aI533
aI1801
aI13965
aI25277
aI52347
aI13447
asS'd'
I11877
sa(dp23754
S'a'
I52322
sS's'
I17
sS'm'
(lp23755
I1
aI1
aI1
aI15
aI9
aI23
aI115
aI119
aI207
aI973
aI1339
aI1315
aI6465
aI9917
aI4593
aI65435
aI2515
asS'd'
I11878
sa(dp23756
S'a'
I52328
sS's'
I17
sS'm'
(lp23757
I1
aI3
aI3
aI3
aI7
aI25
aI115
aI115
aI213
aI845
aI1445
aI1217
aI1563
aI12491
aI5197
aI44409
aI79895
asS'd'
I11879
sa(dp23758
S'a'
I52333
sS's'
I17
sS'm'
(lp23759
I1
aI1
aI1
aI3
aI9
aI33
aI31
aI203
aI19
aI895
aI1145
aI2893
aI4807
aI7501
aI6999
aI54883
aI13797
asS'd'
I11880
sa(dp23760
S'a'
I52351
sS's'
I17
sS'm'
(lp23761
I1
aI1
aI3
aI1
aI19
aI51
aI73
aI29
aI451
aI533
aI83
aI2477
aI335
aI9703
aI9747
aI57427
aI69443
asS'd'
I11881
sa(dp23762
S'a'
I52357
sS's'
I17
sS'm'
(lp23763
I1
aI1
aI7
aI5
aI21
aI11
aI53
aI133
aI165
aI291
aI591
aI1419
aI3661
aI4697
aI21363
aI53467
aI84063
asS'd'
I11882
sa(dp23764
S'a'
I52372
sS's'
I17
sS'm'
(lp23765
I1
aI1
aI7
aI3
aI13
aI7
aI85
aI49
aI193
aI289
aI1531
aI709
aI2351
aI12085
aI28553
aI57145
aI129517
asS'd'
I11883
sa(dp23766
S'a'
I52381
sS's'
I17
sS'm'
(lp23767
I1
aI1
aI1
aI1
aI17
aI19
aI13
aI213
aI75
aI977
aI811
aI1813
aI7293
aI13795
aI28569
aI28133
aI11949
asS'd'
I11884
sa(dp23768
S'a'
I52386
sS's'
I17
sS'm'
(lp23769
I1
aI1
aI5
aI1
aI27
aI11
aI47
aI89
aI271
aI65
aI1651
aI2331
aI3289
aI6227
aI15027
aI19959
aI22945
asS'd'
I11885
sa(dp23770
S'a'
I52395
sS's'
I17
sS'm'
(lp23771
I1
aI3
aI7
aI9
aI17
aI37
aI17
aI245
aI249
aI501
aI405
aI951
aI3005
aI9757
aI10265
aI35575
aI70529
asS'd'
I11886
sa(dp23772
S'a'
I52403
sS's'
I17
sS'm'
(lp23773
I1
aI3
aI1
aI15
aI21
aI47
aI15
aI75
aI113
aI45
aI125
aI1393
aI3361
aI13477
aI24325
aI39743
aI67409
asS'd'
I11887
sa(dp23774
S'a'
I52423
sS's'
I17
sS'm'
(lp23775
I1
aI3
aI5
aI1
aI13
aI3
aI33
aI51
aI463
aI241
aI1421
aI1607
aI3937
aI3405
aI26653
aI33955
aI97915
asS'd'
I11888
sa(dp23776
S'a'
I52427
sS's'
I17
sS'm'
(lp23777
I1
aI3
aI7
aI13
aI29
aI17
aI41
aI193
aI461
aI787
aI459
aI4019
aI1887
aI13831
aI9387
aI25215
aI69801
asS'd'
I11889
sa(dp23778
S'a'
I52432
sS's'
I17
sS'm'
(lp23779
I1
aI3
aI1
aI11
aI31
aI55
aI13
aI235
aI85
aI953
aI109
aI233
aI1893
aI13225
aI26165
aI59237
aI15845
asS'd'
I11890
sa(dp23780
S'a'
I52438
sS's'
I17
sS'm'
(lp23781
I1
aI1
aI1
aI13
aI11
aI43
aI127
aI193
aI143
aI831
aI875
aI2471
aI7011
aI3063
aI21979
aI42113
aI80581
asS'd'
I11891
sa(dp23782
S'a'
I52444
sS's'
I17
sS'm'
(lp23783
I1
aI3
aI7
aI1
aI5
aI1
aI63
aI55
aI349
aI525
aI441
aI2695
aI3301
aI15737
aI13355
aI16727
aI25001
asS'd'
I11892
sa(dp23784
S'a'
I52457
sS's'
I17
sS'm'
(lp23785
I1
aI3
aI7
aI7
aI31
aI23
aI87
aI99
aI331
aI101
aI1341
aI3
aI1447
aI9507
aI18627
aI2503
aI57597
asS'd'
I11893
sa(dp23786
S'a'
I52468
sS's'
I17
sS'm'
(lp23787
I1
aI1
aI1
aI9
aI11
aI19
aI89
aI141
aI269
aI31
aI1933
aI429
aI7765
aI5905
aI15327
aI25913
aI17281
asS'd'
I11894
sa(dp23788
S'a'
I52480
sS's'
I17
sS'm'
(lp23789
I1
aI3
aI1
aI9
aI23
aI61
aI75
aI15
aI121
aI635
aI1409
aI615
aI7841
aI11993
aI1637
aI26073
aI70763
asS'd'
I11895
sa(dp23790
S'a'
I52498
sS's'
I17
sS'm'
(lp23791
I1
aI3
aI3
aI1
aI5
aI63
aI85
aI3
aI443
aI87
aI1201
aI275
aI3457
aI16187
aI26839
aI16593
aI36335
asS'd'
I11896
sa(dp23792
S'a'
I52516
sS's'
I17
sS'm'
(lp23793
I1
aI1
aI5
aI5
aI27
aI61
aI1
aI145
aI287
aI563
aI1135
aI2703
aI733
aI10209
aI3937
aI23807
aI56857
asS'd'
I11897
sa(dp23794
S'a'
I52520
sS's'
I17
sS'm'
(lp23795
I1
aI3
aI3
aI7
aI1
aI41
aI97
aI155
aI305
aI395
aI1607
aI1171
aI1061
aI12301
aI8041
aI12111
aI66831
asS'd'
I11898
sa(dp23796
S'a'
I52525
sS's'
I17
sS'm'
(lp23797
I1
aI3
aI7
aI15
aI9
aI61
aI127
aI225
aI125
aI231
aI87
aI2433
aI6951
aI2999
aI24859
aI61685
aI111943
asS'd'
I11899
sa(dp23798
S'a'
I52531
sS's'
I17
sS'm'
(lp23799
I1
aI3
aI5
aI15
aI13
aI39
aI87
aI57
aI305
aI469
aI1929
aI1559
aI1383
aI2779
aI3883
aI845
aI45181
asS'd'
I11900
sa(dp23800
S'a'
I52540
sS's'
I17
sS'm'
(lp23801
I1
aI1
aI1
aI15
aI19
aI1
aI23
aI41
aI207
aI731
aI501
aI1147
aI6543
aI5011
aI483
aI56889
aI48993
asS'd'
I11901
sa(dp23802
S'a'
I52545
sS's'
I17
sS'm'
(lp23803
I1
aI3
aI3
aI15
aI21
aI35
aI75
aI129
aI441
aI497
aI953
aI201
aI6849
aI2893
aI6351
aI62163
aI84127
asS'd'
I11902
sa(dp23804
S'a'
I52546
sS's'
I17
sS'm'
(lp23805
I1
aI1
aI1
aI9
aI21
aI31
aI91
aI79
aI345
aI649
aI1529
aI805
aI4931
aI12887
aI30167
aI52305
aI92561
asS'd'
I11903
sa(dp23806
S'a'
I52552
sS's'
I17
sS'm'
(lp23807
I1
aI1
aI5
aI1
aI3
aI21
aI121
aI223
aI67
aI185
aI801
aI889
aI7443
aI8419
aI19929
aI11451
aI11487
asS'd'
I11904
sa(dp23808
S'a'
I52557
sS's'
I17
sS'm'
(lp23809
I1
aI1
aI3
aI15
aI21
aI51
aI119
aI31
aI197
aI773
aI617
aI2055
aI799
aI9105
aI12353
aI33635
aI27589
asS'd'
I11905
sa(dp23810
S'a'
I52569
sS's'
I17
sS'm'
(lp23811
I1
aI3
aI5
aI3
aI27
aI11
aI41
aI105
aI289
aI877
aI1175
aI3111
aI3161
aI12537
aI18001
aI38061
aI59089
asS'd'
I11906
sa(dp23812
S'a'
I52575
sS's'
I17
sS'm'
(lp23813
I1
aI3
aI3
aI5
aI3
aI27
aI101
aI253
aI225
aI915
aI1757
aI1601
aI3391
aI10443
aI3983
aI58427
aI93391
asS'd'
I11907
sa(dp23814
S'a'
I52582
sS's'
I17
sS'm'
(lp23815
I1
aI3
aI1
aI7
aI9
aI43
aI85
aI115
aI169
aI285
aI1267
aI3791
aI2701
aI5599
aI10099
aI48105
aI45219
asS'd'
I11908
sa(dp23816
S'a'
I52594
sS's'
I17
sS'm'
(lp23817
I1
aI1
aI7
aI13
aI25
aI57
aI35
aI223
aI265
aI451
aI1913
aI2715
aI8017
aI3725
aI7079
aI34611
aI61159
asS'd'
I11909
sa(dp23818
S'a'
I52615
sS's'
I17
sS'm'
(lp23819
I1
aI1
aI3
aI7
aI23
aI27
aI93
aI195
aI449
aI845
aI865
aI655
aI4263
aI12743
aI7467
aI7929
aI7095
asS'd'
I11910
sa(dp23820
S'a'
I52619
sS's'
I17
sS'm'
(lp23821
I1
aI3
aI5
aI5
aI29
aI51
aI109
aI123
aI227
aI693
aI2033
aI3829
aI7187
aI4027
aI17861
aI45093
aI7355
asS'd'
I11911
sa(dp23822
S'a'
I52624
sS's'
I17
sS'm'
(lp23823
I1
aI1
aI1
aI11
aI27
aI31
aI127
aI75
aI443
aI479
aI865
aI1377
aI711
aI3791
aI27235
aI17405
aI25975
asS'd'
I11912
sa(dp23824
S'a'
I52645
sS's'
I17
sS'm'
(lp23825
I1
aI3
aI5
aI7
aI1
aI49
aI79
aI167
aI471
aI453
aI211
aI265
aI8163
aI6517
aI3413
aI17283
aI51961
asS'd'
I11913
sa(dp23826
S'a'
I52663
sS's'
I17
sS'm'
(lp23827
I1
aI3
aI1
aI5
aI17
aI29
aI15
aI239
aI385
aI239
aI425
aI2197
aI3553
aI14913
aI14889
aI31645
aI67477
asS'd'
I11914
sa(dp23828
S'a'
I52664
sS's'
I17
sS'm'
(lp23829
I1
aI3
aI3
aI5
aI1
aI11
aI25
aI105
aI367
aI253
aI1395
aI2077
aI2613
aI4535
aI18215
aI37657
aI48283
asS'd'
I11915
sa(dp23830
S'a'
I52675
sS's'
I17
sS'm'
(lp23831
I1
aI3
aI3
aI1
aI1
aI41
aI7
aI161
aI437
aI659
aI833
aI3175
aI2063
aI14497
aI6655
aI8817
aI127321
asS'd'
I11916
sa(dp23832
S'a'
I52687
sS's'
I17
sS'm'
(lp23833
I1
aI3
aI1
aI11
aI17
aI27
aI3
aI51
aI75
aI183
aI1889
aI287
aI5429
aI14007
aI14445
aI47395
aI94543
asS'd'
I11917
sa(dp23834
S'a'
I52696
sS's'
I17
sS'm'
(lp23835
I1
aI1
aI7
aI13
aI29
aI9
aI109
aI19
aI73
aI3
aI1529
aI457
aI6413
aI4113
aI14733
aI24455
aI44623
asS'd'
I11918
sa(dp23836
S'a'
I52701
sS's'
I17
sS'm'
(lp23837
I1
aI3
aI1
aI15
aI15
aI31
aI83
aI25
aI263
aI229
aI1801
aI377
aI1703
aI8571
aI10393
aI52021
aI100937
asS'd'
I11919
sa(dp23838
S'a'
I52706
sS's'
I17
sS'm'
(lp23839
I1
aI3
aI5
aI9
aI25
aI57
aI79
aI19
aI117
aI437
aI275
aI3439
aI6393
aI2111
aI8317
aI3521
aI96927
asS'd'
I11920
sa(dp23840
S'a'
I52708
sS's'
I17
sS'm'
(lp23841
I1
aI3
aI1
aI13
aI27
aI43
aI103
aI171
aI361
aI949
aI347
aI809
aI5819
aI2763
aI10447
aI35129
aI46985
asS'd'
I11921
sa(dp23842
S'a'
I52711
sS's'
I17
sS'm'
(lp23843
I1
aI3
aI5
aI11
aI17
aI1
aI27
aI37
aI473
aI851
aI1057
aI831
aI4373
aI5179
aI18193
aI48731
aI64317
asS'd'
I11922
sa(dp23844
S'a'
I52726
sS's'
I17
sS'm'
(lp23845
I1
aI3
aI1
aI7
aI17
aI5
aI19
aI217
aI439
aI549
aI1983
aI2473
aI6059
aI5271
aI10279
aI7793
aI114357
asS'd'
I11923
sa(dp23846
S'a'
I52748
sS's'
I17
sS'm'
(lp23847
I1
aI3
aI1
aI5
aI25
aI19
aI99
aI65
aI507
aI527
aI825
aI2517
aI2299
aI1725
aI9913
aI5779
aI12207
asS'd'
I11924
sa(dp23848
S'a'
I52754
sS's'
I17
sS'm'
(lp23849
I1
aI3
aI1
aI1
aI29
aI41
aI119
aI27
aI411
aI475
aI461
aI1885
aI2339
aI4945
aI24665
aI13621
aI78129
asS'd'
I11925
sa(dp23850
S'a'
I52756
sS's'
I17
sS'm'
(lp23851
I1
aI3
aI1
aI11
aI27
aI29
aI119
aI155
aI487
aI29
aI1545
aI675
aI1417
aI6119
aI12451
aI21345
aI39377
asS'd'
I11926
sa(dp23852
S'a'
I52759
sS's'
I17
sS'm'
(lp23853
I1
aI1
aI3
aI7
aI19
aI5
aI111
aI227
aI49
aI307
aI549
aI737
aI4793
aI13885
aI22971
aI18653
aI69573
asS'd'
I11927
sa(dp23854
S'a'
I52769
sS's'
I17
sS'm'
(lp23855
I1
aI3
aI3
aI1
aI27
aI59
aI87
aI7
aI379
aI497
aI903
aI591
aI6105
aI1957
aI25849
aI55957
aI120181
asS'd'
I11928
sa(dp23856
S'a'
I52784
sS's'
I17
sS'm'
(lp23857
I1
aI3
aI5
aI15
aI19
aI31
aI43
aI1
aI35
aI341
aI311
aI1343
aI3625
aI16181
aI31047
aI59679
aI89231
asS'd'
I11929
sa(dp23858
S'a'
I52790
sS's'
I17
sS'm'
(lp23859
I1
aI1
aI1
aI15
aI21
aI19
aI93
aI229
aI49
aI597
aI1465
aI2027
aI5935
aI12269
aI20239
aI17861
aI26311
asS'd'
I11930
sa(dp23860
S'a'
I52804
sS's'
I17
sS'm'
(lp23861
I1
aI1
aI1
aI1
aI3
aI31
aI115
aI87
aI129
aI419
aI871
aI2469
aI3807
aI4473
aI25025
aI36923
aI67959
asS'd'
I11931
sa(dp23862
S'a'
I52807
sS's'
I17
sS'm'
(lp23863
I1
aI1
aI1
aI3
aI23
aI31
aI41
aI247
aI295
aI369
aI1131
aI2183
aI8097
aI7609
aI4387
aI37565
aI50177
asS'd'
I11932
sa(dp23864
S'a'
I52808
sS's'
I17
sS'm'
(lp23865
I1
aI3
aI1
aI11
aI9
aI17
aI111
aI249
aI417
aI775
aI217
aI1435
aI6295
aI5065
aI2967
aI55361
aI91933
asS'd'
I11933
sa(dp23866
S'a'
I52819
sS's'
I17
sS'm'
(lp23867
I1
aI3
aI5
aI7
aI19
aI21
aI71
aI219
aI411
aI31
aI335
aI2915
aI3687
aI5691
aI12405
aI34659
aI76721
asS'd'
I11934
sa(dp23868
S'a'
I52826
sS's'
I17
sS'm'
(lp23869
I1
aI3
aI5
aI13
aI29
aI31
aI95
aI203
aI149
aI399
aI547
aI2529
aI2485
aI3371
aI20219
aI33647
aI34217
asS'd'
I11935
sa(dp23870
S'a'
I52828
sS's'
I17
sS'm'
(lp23871
I1
aI3
aI5
aI13
aI31
aI41
aI97
aI115
aI427
aI35
aI1319
aI2335
aI715
aI2541
aI4507
aI49395
aI33895
asS'd'
I11936
sa(dp23872
S'a'
I52832
sS's'
I17
sS'm'
(lp23873
I1
aI3
aI7
aI15
aI3
aI49
aI3
aI49
aI153
aI93
aI1343
aI1035
aI5685
aI15855
aI15751
aI27663
aI99553
asS'd'
I11937
sa(dp23874
S'a'
I52835
sS's'
I17
sS'm'
(lp23875
I1
aI1
aI7
aI5
aI27
aI7
aI53
aI135
aI453
aI981
aI1767
aI3503
aI1259
aI11973
aI23259
aI41051
aI96593
asS'd'
I11938
sa(dp23876
S'a'
I52849
sS's'
I17
sS'm'
(lp23877
I1
aI1
aI7
aI9
aI5
aI59
aI57
aI141
aI41
aI639
aI1681
aI145
aI7019
aI6621
aI14221
aI12051
aI71871
asS'd'
I11939
sa(dp23878
S'a'
I52859
sS's'
I17
sS'm'
(lp23879
I1
aI1
aI3
aI1
aI13
aI39
aI7
aI187
aI7
aI919
aI1555
aI2111
aI6507
aI2099
aI10643
aI22851
aI82033
asS'd'
I11940
sa(dp23880
S'a'
I52877
sS's'
I17
sS'm'
(lp23881
I1
aI3
aI7
aI9
aI25
aI59
aI27
aI225
aI239
aI715
aI1115
aI2309
aI7785
aI11849
aI8991
aI54305
aI107305
asS'd'
I11941
sa(dp23882
S'a'
I52880
sS's'
I17
sS'm'
(lp23883
I1
aI1
aI7
aI11
aI21
aI51
aI1
aI223
aI481
aI195
aI2005
aI2651
aI6797
aI12201
aI11013
aI1843
aI65167
asS'd'
I11942
sa(dp23884
S'a'
I52896
sS's'
I17
sS'm'
(lp23885
I1
aI3
aI3
aI11
aI27
aI3
aI117
aI5
aI255
aI595
aI399
aI1329
aI1437
aI12061
aI32679
aI16655
aI76235
asS'd'
I11943
sa(dp23886
S'a'
I52899
sS's'
I17
sS'm'
(lp23887
I1
aI1
aI7
aI13
aI21
aI1
aI35
aI159
aI329
aI37
aI1247
aI2663
aI3889
aI14603
aI25799
aI45363
aI87963
asS'd'
I11944
sa(dp23888
S'a'
I52905
sS's'
I17
sS'm'
(lp23889
I1
aI1
aI7
aI7
aI7
aI11
aI53
aI215
aI351
aI329
aI1039
aI969
aI4449
aI14785
aI28617
aI25953
aI78663
asS'd'
I11945
sa(dp23890
S'a'
I52913
sS's'
I17
sS'm'
(lp23891
I1
aI1
aI7
aI7
aI27
aI17
aI19
aI223
aI143
aI433
aI789
aI1941
aI5527
aI711
aI25097
aI4571
aI121975
asS'd'
I11946
sa(dp23892
S'a'
I52933
sS's'
I17
sS'm'
(lp23893
I1
aI3
aI1
aI13
aI11
aI47
aI31
aI249
aI325
aI1003
aI509
aI2741
aI3953
aI1691
aI12661
aI16097
aI80211
asS'd'
I11947
sa(dp23894
S'a'
I52934
sS's'
I17
sS'm'
(lp23895
I1
aI3
aI7
aI9
aI27
aI11
aI21
aI129
aI25
aI57
aI1523
aI3631
aI2639
aI2541
aI14249
aI34539
aI70551
asS'd'
I11948
sa(dp23896
S'a'
I52938
sS's'
I17
sS'm'
(lp23897
I1
aI1
aI5
aI3
aI31
aI47
aI47
aI73
aI71
aI783
aI1353
aI2157
aI2563
aI14015
aI997
aI31611
aI118649
asS'd'
I11949
sa(dp23898
S'a'
I52957
sS's'
I17
sS'm'
(lp23899
I1
aI1
aI5
aI5
aI25
aI35
aI25
aI207
aI349
aI503
aI121
aI3455
aI5783
aI10899
aI12745
aI35117
aI36679
asS'd'
I11950
sa(dp23900
S'a'
I52979
sS's'
I17
sS'm'
(lp23901
I1
aI3
aI1
aI3
aI11
aI39
aI123
aI177
aI19
aI441
aI1979
aI1257
aI1351
aI4253
aI15145
aI44559
aI59379
asS'd'
I11951
sa(dp23902
S'a'
I52981
sS's'
I17
sS'm'
(lp23903
I1
aI3
aI7
aI3
aI7
aI35
aI41
aI203
aI439
aI1013
aI1055
aI1165
aI1043
aI11183
aI1795
aI31253
aI113693
asS'd'
I11952
sa(dp23904
S'a'
I52986
sS's'
I17
sS'm'
(lp23905
I1
aI3
aI1
aI13
aI7
aI43
aI57
aI1
aI229
aI345
aI631
aI841
aI7923
aI5971
aI20489
aI47917
aI66833
asS'd'
I11953
sa(dp23906
S'a'
I53005
sS's'
I17
sS'm'
(lp23907
I1
aI1
aI1
aI15
aI27
aI5
aI31
aI117
aI153
aI755
aI1207
aI619
aI8185
aI4329
aI9979
aI57255
aI79045
asS'd'
I11954
sa(dp23908
S'a'
I53008
sS's'
I17
sS'm'
(lp23909
I1
aI3
aI3
aI7
aI23
aI1
aI7
aI227
aI337
aI417
aI1895
aI765
aI7799
aI13599
aI27253
aI4485
aI112391
asS'd'
I11955
sa(dp23910
S'a'
I53024
sS's'
I17
sS'm'
(lp23911
I1
aI3
aI5
aI13
aI27
aI63
aI5
aI87
aI101
aI351
aI953
aI2235
aI1587
aI5479
aI26529
aI34165
aI83303
asS'd'
I11956
sa(dp23912
S'a'
I53029
sS's'
I17
sS'm'
(lp23913
I1
aI1
aI5
aI15
aI1
aI43
aI63
aI193
aI143
aI711
aI1779
aI3531
aI1355
aI16253
aI14595
aI32343
aI131021
asS'd'
I11957
sa(dp23914
S'a'
I53054
sS's'
I17
sS'm'
(lp23915
I1
aI1
aI1
aI9
aI29
aI37
aI29
aI71
aI11
aI877
aI1301
aI2415
aI5593
aI1855
aI25223
aI6805
aI12901
asS'd'
I11958
sa(dp23916
S'a'
I53073
sS's'
I17
sS'm'
(lp23917
I1
aI1
aI7
aI9
aI31
aI5
aI49
aI63
aI185
aI373
aI129
aI1695
aI7841
aI4477
aI17809
aI42771
aI120221
asS'd'
I11959
sa(dp23918
S'a'
I53083
sS's'
I17
sS'm'
(lp23919
I1
aI1
aI5
aI3
aI15
aI43
aI49
aI45
aI47
aI775
aI699
aI2787
aI7831
aI4189
aI18317
aI63933
aI83669
asS'd'
I11960
sa(dp23920
S'a'
I53086
sS's'
I17
sS'm'
(lp23921
I1
aI3
aI5
aI3
aI23
aI33
aI85
aI255
aI119
aI685
aI1245
aI1647
aI1999
aI13063
aI9241
aI49017
aI32619
asS'd'
I11961
sa(dp23922
S'a'
I53095
sS's'
I17
sS'm'
(lp23923
I1
aI1
aI7
aI15
aI29
aI15
aI125
aI233
aI189
aI411
aI1251
aI3459
aI7213
aI10081
aI4403
aI56819
aI17103
asS'd'
I11962
sa(dp23924
S'a'
I53102
sS's'
I17
sS'm'
(lp23925
I1
aI3
aI3
aI11
aI21
aI13
aI93
aI125
aI213
aI793
aI1057
aI2363
aI661
aI12213
aI2259
aI3787
aI91451
asS'd'
I11963
sa(dp23926
S'a'
I53107
sS's'
I17
sS'm'
(lp23927
I1
aI3
aI5
aI5
aI19
aI35
aI5
aI153
aI507
aI691
aI1743
aI1777
aI7579
aI14229
aI10155
aI18529
aI35945
asS'd'
I11964
sa(dp23928
S'a'
I53126
sS's'
I17
sS'm'
(lp23929
I1
aI3
aI7
aI5
aI27
aI35
aI13
aI77
aI189
aI793
aI877
aI643
aI2787
aI5817
aI22589
aI58363
aI49059
asS'd'
I11965
sa(dp23930
S'a'
I53130
sS's'
I17
sS'm'
(lp23931
I1
aI3
aI7
aI9
aI9
aI37
aI21
aI251
aI119
aI895
aI1023
aI91
aI4317
aI10943
aI7355
aI36961
aI36903
asS'd'
I11966
sa(dp23932
S'a'
I53138
sS's'
I17
sS'm'
(lp23933
I1
aI3
aI3
aI13
aI19
aI49
aI15
aI105
aI399
aI29
aI1903
aI3503
aI3453
aI15429
aI31503
aI57815
aI34009
asS'd'
I11967
sa(dp23934
S'a'
I53144
sS's'
I17
sS'm'
(lp23935
I1
aI1
aI5
aI1
aI19
aI35
aI49
aI97
aI335
aI665
aI1871
aI887
aI4713
aI517
aI9571
aI41601
aI9673
asS'd'
I11968
sa(dp23936
S'a'
I53156
sS's'
I17
sS'm'
(lp23937
I1
aI3
aI5
aI13
aI29
aI45
aI111
aI233
aI251
aI407
aI1135
aI2791
aI6525
aI11633
aI22295
aI65381
aI117511
asS'd'
I11969
sa(dp23938
S'a'
I53163
sS's'
I17
sS'm'
(lp23939
I1
aI1
aI3
aI3
aI17
aI7
aI65
aI43
aI391
aI91
aI315
aI3559
aI479
aI7337
aI25629
aI785
aI19855
asS'd'
I11970
sa(dp23940
S'a'
I53165
sS's'
I17
sS'm'
(lp23941
I1
aI1
aI5
aI9
aI29
aI31
aI67
aI17
aI381
aI875
aI1001
aI415
aI2263
aI4415
aI11263
aI309
aI117623
asS'd'
I11971
sa(dp23942
S'a'
I53173
sS's'
I17
sS'm'
(lp23943
I1
aI1
aI7
aI11
aI25
aI1
aI59
aI61
aI247
aI649
aI687
aI907
aI1037
aI13935
aI7229
aI39769
aI92755
asS'd'
I11972
sa(dp23944
S'a'
I53177
sS's'
I17
sS'm'
(lp23945
I1
aI3
aI5
aI15
aI21
aI51
aI27
aI79
aI343
aI785
aI1567
aI1349
aI7991
aI8531
aI11243
aI61351
aI21297
asS'd'
I11973
sa(dp23946
S'a'
I53183
sS's'
I17
sS'm'
(lp23947
I1
aI1
aI1
aI3
aI31
aI41
aI67
aI169
aI83
aI959
aI813
aI1953
aI2467
aI12369
aI31431
aI50761
aI75731
asS'd'
I11974
sa(dp23948
S'a'
I53192
sS's'
I17
sS'm'
(lp23949
I1
aI1
aI5
aI11
aI25
aI37
aI83
aI163
aI3
aI161
aI1249
aI3009
aI7167
aI5473
aI10561
aI44899
aI130879
asS'd'
I11975
sa(dp23950
S'a'
I53195
sS's'
I17
sS'm'
(lp23951
I1
aI1
aI7
aI11
aI9
aI61
aI61
aI113
aI81
aI205
aI731
aI3887
aI5525
aI13415
aI25181
aI11557
aI59343
asS'd'
I11976
sa(dp23952
S'a'
I53200
sS's'
I17
sS'm'
(lp23953
I1
aI3
aI7
aI5
aI19
aI27
aI107
aI89
aI295
aI715
aI1439
aI1285
aI5813
aI8895
aI7233
aI32905
aI3273
asS'd'
I11977
sa(dp23954
S'a'
I53212
sS's'
I17
sS'm'
(lp23955
I1
aI1
aI5
aI1
aI29
aI11
aI125
aI253
aI445
aI295
aI1721
aI1271
aI2203
aI2215
aI7613
aI55655
aI112157
asS'd'
I11978
sa(dp23956
S'a'
I53219
sS's'
I17
sS'm'
(lp23957
I1
aI1
aI5
aI11
aI11
aI13
aI111
aI55
aI19
aI551
aI1365
aI477
aI2513
aI12311
aI22485
aI23291
aI92447
asS'd'
I11979
sa(dp23958
S'a'
I53221
sS's'
I17
sS'm'
(lp23959
I1
aI1
aI7
aI11
aI9
aI5
aI3
aI109
aI507
aI441
aI1767
aI1781
aI3077
aI219
aI29293
aI21237
aI71159
asS'd'
I11980
sa(dp23960
S'a'
I53245
sS's'
I17
sS'm'
(lp23961
I1
aI1
aI3
aI11
aI3
aI45
aI99
aI113
aI367
aI569
aI1907
aI1281
aI51
aI13693
aI14639
aI32999
aI77461
asS'd'
I11981
sa(dp23962
S'a'
I53254
sS's'
I17
sS'm'
(lp23963
I1
aI3
aI5
aI11
aI5
aI19
aI97
aI11
aI473
aI937
aI1623
aI1507
aI3245
aI9331
aI16005
aI37505
aI40085
asS'd'
I11982
sa(dp23964
S'a'
I53257
sS's'
I17
sS'm'
(lp23965
I1
aI1
aI7
aI13
aI21
aI61
aI103
aI111
aI35
aI141
aI61
aI1043
aI1989
aI1311
aI29021
aI2617
aI89915
asS'd'
I11983
sa(dp23966
S'a'
I53265
sS's'
I17
sS'm'
(lp23967
I1
aI3
aI7
aI15
aI19
aI31
aI39
aI175
aI371
aI459
aI1293
aI1645
aI1125
aI1199
aI4811
aI55721
aI76071
asS'd'
I11984
sa(dp23968
S'a'
I53266
sS's'
I17
sS'm'
(lp23969
I1
aI1
aI7
aI3
aI3
aI35
aI17
aI7
aI91
aI317
aI1615
aI3559
aI191
aI2579
aI15027
aI58711
aI36009
asS'd'
I11985
sa(dp23970
S'a'
I53268
sS's'
I17
sS'm'
(lp23971
I1
aI1
aI1
aI13
aI1
aI27
aI45
aI87
aI443
aI443
aI853
aI3917
aI1437
aI4053
aI14861
aI2897
aI109853
asS'd'
I11986
sa(dp23972
S'a'
I53275
sS's'
I17
sS'm'
(lp23973
I1
aI1
aI5
aI3
aI21
aI47
aI73
aI195
aI115
aI517
aI1781
aI2341
aI805
aI5679
aI12053
aI29113
aI100479
asS'd'
I11987
sa(dp23974
S'a'
I53277
sS's'
I17
sS'm'
(lp23975
I1
aI1
aI7
aI1
aI25
aI27
aI61
aI167
aI203
aI57
aI527
aI1071
aI7131
aI8403
aI9943
aI11503
aI33081
asS'd'
I11988
sa(dp23976
S'a'
I53284
sS's'
I17
sS'm'
(lp23977
I1
aI1
aI5
aI13
aI31
aI43
aI35
aI195
aI177
aI229
aI1401
aI4011
aI2363
aI15787
aI21125
aI32103
aI62337
asS'd'
I11989
sa(dp23978
S'a'
I53294
sS's'
I17
sS'm'
(lp23979
I1
aI1
aI5
aI11
aI19
aI13
aI3
aI249
aI119
aI35
aI747
aI1419
aI5451
aI13043
aI19813
aI54859
aI94825
asS'd'
I11990
sa(dp23980
S'a'
I53308
sS's'
I17
sS'm'
(lp23981
I1
aI3
aI1
aI9
aI17
aI13
aI51
aI125
aI391
aI157
aI1199
aI1805
aI1763
aI11831
aI20915
aI38547
aI14221
asS'd'
I11991
sa(dp23982
S'a'
I53314
sS's'
I17
sS'm'
(lp23983
I1
aI1
aI7
aI1
aI23
aI61
aI25
aI69
aI435
aI183
aI1379
aI1211
aI5529
aI9447
aI4575
aI14127
aI91867
asS'd'
I11992
sa(dp23984
S'a'
I53319
sS's'
I17
sS'm'
(lp23985
I1
aI3
aI3
aI15
aI11
aI15
aI101
aI135
aI419
aI685
aI1097
aI787
aI2045
aI3393
aI26221
aI23653
aI116917
asS'd'
I11993
sa(dp23986
S'a'
I53326
sS's'
I17
sS'm'
(lp23987
I1
aI3
aI1
aI11
aI29
aI23
aI13
aI153
aI27
aI683
aI1569
aI413
aI261
aI10291
aI23763
aI15579
aI39337
asS'd'
I11994
sa(dp23988
S'a'
I53328
sS's'
I17
sS'm'
(lp23989
I1
aI3
aI7
aI7
aI19
aI23
aI121
aI23
aI339
aI165
aI1137
aI2791
aI319
aI16111
aI14847
aI28171
aI79237
asS'd'
I11995
sa(dp23990
S'a'
I53340
sS's'
I17
sS'm'
(lp23991
I1
aI3
aI1
aI5
aI9
aI59
aI33
aI19
aI191
aI707
aI1883
aI1683
aI1161
aI12905
aI12299
aI22201
aI19811
asS'd'
I11996
sa(dp23992
S'a'
I53364
sS's'
I17
sS'm'
(lp23993
I1
aI3
aI1
aI3
aI27
aI11
aI69
aI199
aI415
aI251
aI1079
aI1709
aI4539
aI7867
aI21321
aI33617
aI53459
asS'd'
I11997
sa(dp23994
S'a'
I53367
sS's'
I17
sS'm'
(lp23995
I1
aI1
aI3
aI9
aI19
aI59
aI21
aI95
aI275
aI213
aI1819
aI721
aI6271
aI11845
aI9573
aI16105
aI12755
asS'd'
I11998
sa(dp23996
S'a'
I53377
sS's'
I17
sS'm'
(lp23997
I1
aI1
aI1
aI15
aI23
aI7
aI91
aI235
aI43
aI95
aI913
aI715
aI3229
aI12339
aI23089
aI30963
aI129525
asS'd'
I11999
sa(dp23998
S'a'
I53395
sS's'
I17
sS'm'
(lp23999
I1
aI1
aI7
aI9
aI7
aI41
aI43
aI131
aI485
aI621
aI1293
aI1955
aI5215
aI6545
aI29225
aI53587
aI46901
asS'd'
I12000
sa(dp24000
S'a'
I53398
sS's'
I17
sS'm'
(lp24001
I1
aI3
aI1
aI5
aI7
aI57
aI97
aI199
aI225
aI707
aI1223
aI1829
aI497
aI12587
aI24551
aI12907
aI82411
asS'd'
I12001
sa(dp24002
S'a'
I53407
sS's'
I17
sS'm'
(lp24003
I1
aI1
aI3
aI7
aI21
aI9
aI63
aI15
aI263
aI957
aI155
aI4021
aI4455
aI2025
aI16981
aI19743
aI88619
asS'd'
I12002
sa(dp24004
S'a'
I53413
sS's'
I17
sS'm'
(lp24005
I1
aI1
aI7
aI5
aI31
aI3
aI27
aI45
aI369
aI747
aI1559
aI1429
aI8049
aI15069
aI19897
aI50067
aI52861
asS'd'
I12003
sa(dp24006
S'a'
I53414
sS's'
I17
sS'm'
(lp24007
I1
aI1
aI5
aI13
aI23
aI35
aI91
aI139
aI73
aI275
aI207
aI2709
aI3801
aI12755
aI19155
aI61629
aI5513
asS'd'
I12004
sa(dp24008
S'a'
I53417
sS's'
I17
sS'm'
(lp24009
I1
aI1
aI5
aI7
aI5
aI25
aI33
aI45
aI325
aI847
aI81
aI891
aI3191
aI14115
aI25095
aI39867
aI3839
asS'd'
I12005
sa(dp24010
S'a'
I53443
sS's'
I17
sS'm'
(lp24011
I1
aI3
aI5
aI13
aI9
aI31
aI31
aI113
aI507
aI833
aI691
aI2041
aI4873
aI81
aI21365
aI35265
aI37627
asS'd'
I12006
sa(dp24012
S'a'
I53473
sS's'
I17
sS'm'
(lp24013
I1
aI1
aI5
aI13
aI9
aI51
aI127
aI131
aI285
aI683
aI593
aI3411
aI6685
aI3601
aI12255
aI8337
aI80597
asS'd'
I12007
sa(dp24014
S'a'
I53476
sS's'
I17
sS'm'
(lp24015
I1
aI1
aI5
aI15
aI29
aI13
aI79
aI199
aI157
aI421
aI1697
aI2063
aI2213
aI4141
aI21045
aI45785
aI124023
asS'd'
I12008
sa(dp24016
S'a'
I53480
sS's'
I17
sS'm'
(lp24017
I1
aI3
aI1
aI11
aI19
aI5
aI79
aI57
aI71
aI373
aI487
aI671
aI3671
aI9093
aI20989
aI48477
aI104951
asS'd'
I12009
sa(dp24018
S'a'
I53486
sS's'
I17
sS'm'
(lp24019
I1
aI3
aI5
aI15
aI13
aI7
aI39
aI243
aI507
aI739
aI1905
aI3431
aI4141
aI9345
aI27877
aI64735
aI112997
asS'd'
I12010
sa(dp24020
S'a'
I53506
sS's'
I17
sS'm'
(lp24021
I1
aI3
aI3
aI5
aI17
aI25
aI31
aI243
aI393
aI61
aI199
aI2825
aI6981
aI5887
aI22289
aI9201
aI77689
asS'd'
I12011
sa(dp24022
S'a'
I53542
sS's'
I17
sS'm'
(lp24023
I1
aI3
aI5
aI15
aI15
aI63
aI77
aI39
aI463
aI883
aI671
aI3285
aI6925
aI15085
aI1665
aI64005
aI130619
asS'd'
I12012
sa(dp24024
S'a'
I53546
sS's'
I17
sS'm'
(lp24025
I1
aI3
aI3
aI11
aI21
aI15
aI7
aI115
aI9
aI879
aI1097
aI3993
aI3929
aI9809
aI22105
aI9549
aI31819
asS'd'
I12013
sa(dp24026
S'a'
I53554
sS's'
I17
sS'm'
(lp24027
I1
aI1
aI7
aI15
aI3
aI9
aI19
aI97
aI327
aI105
aI1915
aI205
aI3873
aI1229
aI29915
aI57375
aI108217
asS'd'
I12014
sa(dp24028
S'a'
I53563
sS's'
I17
sS'm'
(lp24029
I1
aI1
aI3
aI11
aI29
aI41
aI77
aI11
aI183
aI73
aI1651
aI3739
aI3911
aI8695
aI15339
aI19293
aI1827
asS'd'
I12015
sa(dp24030
S'a'
I53580
sS's'
I17
sS'm'
(lp24031
I1
aI1
aI1
aI5
aI23
aI49
aI35
aI175
aI99
aI49
aI615
aI1733
aI6901
aI2351
aI18765
aI55553
aI99791
asS'd'
I12016
sa(dp24032
S'a'
I53591
sS's'
I17
sS'm'
(lp24033
I1
aI3
aI7
aI3
aI25
aI17
aI67
aI161
aI507
aI941
aI35
aI2619
aI339
aI791
aI6485
aI64277
aI123867
asS'd'
I12017
sa(dp24034
S'a'
I53598
sS's'
I17
sS'm'
(lp24035
I1
aI1
aI3
aI13
aI11
aI9
aI79
aI193
aI75
aI391
aI1753
aI3537
aI6971
aI6607
aI11933
aI4447
aI87793
asS'd'
I12018
sa(dp24036
S'a'
I53611
sS's'
I17
sS'm'
(lp24037
I1
aI1
aI1
aI5
aI19
aI9
aI63
aI203
aI51
aI395
aI1365
aI2393
aI7265
aI11709
aI13721
aI4519
aI118765
asS'd'
I12019
sa(dp24038
S'a'
I53621
sS's'
I17
sS'm'
(lp24039
I1
aI1
aI3
aI9
aI17
aI53
aI29
aI103
aI325
aI973
aI903
aI785
aI7535
aI9951
aI8121
aI12603
aI38679
asS'd'
I12020
sa(dp24040
S'a'
I53625
sS's'
I17
sS'm'
(lp24041
I1
aI3
aI1
aI7
aI7
aI63
aI1
aI123
aI439
aI181
aI1373
aI2705
aI995
aI10789
aI7495
aI54543
aI120109
asS'd'
I12021
sa(dp24042
S'a'
I53628
sS's'
I17
sS'm'
(lp24043
I1
aI1
aI7
aI3
aI17
aI13
aI79
aI179
aI165
aI965
aI1537
aI3753
aI3497
aI12127
aI6983
aI48605
aI113057
asS'd'
I12022
sa(dp24044
S'a'
I53632
sS's'
I17
sS'm'
(lp24045
I1
aI1
aI5
aI7
aI3
aI7
aI41
aI25
aI267
aI633
aI19
aI1317
aI3445
aI12377
aI27881
aI55249
aI40841
asS'd'
I12023
sa(dp24046
S'a'
I53650
sS's'
I17
sS'm'
(lp24047
I1
aI3
aI5
aI1
aI31
aI55
aI43
aI129
aI411
aI281
aI1
aI851
aI2419
aI7943
aI13721
aI39371
aI114557
asS'd'
I12024
sa(dp24048
S'a'
I53655
sS's'
I17
sS'm'
(lp24049
I1
aI1
aI7
aI7
aI23
aI19
aI83
aI37
aI9
aI161
aI125
aI3179
aI7973
aI9703
aI23199
aI32723
aI130915
asS'd'
I12025
sa(dp24050
S'a'
I53675
sS's'
I17
sS'm'
(lp24051
I1
aI1
aI7
aI5
aI27
aI21
aI11
aI219
aI403
aI239
aI1723
aI2957
aI3029
aI9911
aI10981
aI35421
aI74025
asS'd'
I12026
sa(dp24052
S'a'
I53677
sS's'
I17
sS'm'
(lp24053
I1
aI3
aI1
aI1
aI31
aI59
aI69
aI77
aI395
aI1
aI157
aI1259
aI4913
aI2089
aI17619
aI51033
aI130899
asS'd'
I12027
sa(dp24054
S'a'
I53680
sS's'
I17
sS'm'
(lp24055
I1
aI3
aI3
aI3
aI19
aI11
aI83
aI237
aI103
aI921
aI487
aI1833
aI8187
aI3811
aI18887
aI9389
aI80927
asS'd'
I12028
sa(dp24056
S'a'
I53683
sS's'
I17
sS'm'
(lp24057
I1
aI3
aI7
aI3
aI17
aI51
aI107
aI209
aI187
aI831
aI1501
aI1337
aI803
aI10361
aI11347
aI65291
aI42219
asS'd'
I12029
sa(dp24058
S'a'
I53700
sS's'
I17
sS'm'
(lp24059
I1
aI3
aI3
aI15
aI7
aI29
aI61
aI25
aI413
aI257
aI1185
aI4009
aI7463
aI1839
aI6645
aI28389
aI14449
asS'd'
I12030
sa(dp24060
S'a'
I53709
sS's'
I17
sS'm'
(lp24061
I1
aI3
aI1
aI9
aI5
aI31
aI83
aI55
aI375
aI399
aI945
aI997
aI7649
aI12631
aI7691
aI53325
aI50173
asS'd'
I12031
sa(dp24062
S'a'
I53724
sS's'
I17
sS'm'
(lp24063
I1
aI1
aI5
aI9
aI13
aI9
aI83
aI37
aI487
aI975
aI487
aI3587
aI7285
aI7505
aI10155
aI673
aI126505
asS'd'
I12032
sa(dp24064
S'a'
I53731
sS's'
I17
sS'm'
(lp24065
I1
aI3
aI5
aI7
aI21
aI3
aI35
aI21
aI367
aI323
aI1579
aI3351
aI5465
aI13719
aI17033
aI42573
aI55079
asS'd'
I12033
sa(dp24066
S'a'
I53733
sS's'
I17
sS'm'
(lp24067
I1
aI3
aI3
aI15
aI11
aI27
aI121
aI109
aI267
aI855
aI1417
aI3839
aI6535
aI1051
aI29355
aI23815
aI76031
asS'd'
I12034
sa(dp24068
S'a'
I53738
sS's'
I17
sS'm'
(lp24069
I1
aI1
aI7
aI9
aI5
aI31
aI35
aI53
aI369
aI137
aI1545
aI927
aI825
aI1333
aI13637
aI11003
aI96963
asS'd'
I12035
sa(dp24070
S'a'
I53762
sS's'
I17
sS'm'
(lp24071
I1
aI1
aI5
aI3
aI29
aI41
aI31
aI85
aI35
aI477
aI227
aI3325
aI1213
aI681
aI14591
aI31325
aI12199
asS'd'
I12036
sa(dp24072
S'a'
I53767
sS's'
I17
sS'm'
(lp24073
I1
aI3
aI7
aI11
aI11
aI11
aI33
aI255
aI335
aI747
aI855
aI31
aI6101
aI293
aI20423
aI47521
aI62573
asS'd'
I12037
sa(dp24074
S'a'
I53785
sS's'
I17
sS'm'
(lp24075
I1
aI1
aI1
aI15
aI31
aI15
aI33
aI175
aI321
aI441
aI1197
aI3579
aI4989
aI9275
aI30485
aI1077
aI122947
asS'd'
I12038
sa(dp24076
S'a'
I53786
sS's'
I17
sS'm'
(lp24077
I1
aI3
aI5
aI15
aI23
aI21
aI127
aI223
aI249
aI373
aI1309
aI1469
aI5701
aI9097
aI29897
aI26627
aI38489
asS'd'
I12039
sa(dp24078
S'a'
I53795
sS's'
I17
sS'm'
(lp24079
I1
aI3
aI7
aI3
aI3
aI35
aI83
aI149
aI259
aI315
aI1467
aI1953
aI6035
aI7961
aI10901
aI25171
aI130713
asS'd'
I12040
sa(dp24080
S'a'
I53802
sS's'
I17
sS'm'
(lp24081
I1
aI1
aI3
aI9
aI7
aI63
aI55
aI33
aI375
aI421
aI151
aI1721
aI1999
aI14937
aI17539
aI48323
aI97345
asS'd'
I12041
sa(dp24082
S'a'
I53812
sS's'
I17
sS'm'
(lp24083
I1
aI1
aI5
aI5
aI3
aI21
aI47
aI19
aI227
aI131
aI1591
aI3779
aI929
aI13879
aI13489
aI19805
aI20135
asS'd'
I12042
sa(dp24084
S'a'
I53821
sS's'
I17
sS'm'
(lp24085
I1
aI1
aI7
aI1
aI31
aI25
aI87
aI125
aI213
aI135
aI809
aI3067
aI5035
aI7407
aI2193
aI31423
aI123973
asS'd'
I12043
sa(dp24086
S'a'
I53827
sS's'
I17
sS'm'
(lp24087
I1
aI3
aI5
aI13
aI17
aI19
aI77
aI169
aI345
aI115
aI227
aI649
aI3609
aI15063
aI1895
aI17533
aI95859
asS'd'
I12044
sa(dp24088
S'a'
I53833
sS's'
I17
sS'm'
(lp24089
I1
aI3
aI5
aI15
aI17
aI29
aI17
aI11
aI145
aI601
aI1871
aI851
aI8161
aI14029
aI10039
aI4247
aI62393
asS'd'
I12045
sa(dp24090
S'a'
I53841
sS's'
I17
sS'm'
(lp24091
I1
aI1
aI7
aI13
aI25
aI5
aI49
aI231
aI261
aI71
aI335
aI4081
aI7915
aI11367
aI17087
aI26041
aI128737
asS'd'
I12046
sa(dp24092
S'a'
I53848
sS's'
I17
sS'm'
(lp24093
I1
aI1
aI1
aI13
aI13
aI21
aI77
aI113
aI373
aI1005
aI109
aI2877
aI3001
aI15011
aI2465
aI37015
aI69049
asS'd'
I12047
sa(dp24094
S'a'
I53869
sS's'
I17
sS'm'
(lp24095
I1
aI1
aI3
aI15
aI31
aI33
aI119
aI121
aI41
aI9
aI1567
aI577
aI1687
aI12117
aI17049
aI675
aI10729
asS'd'
I12048
sa(dp24096
S'a'
I53897
sS's'
I17
sS'm'
(lp24097
I1
aI3
aI5
aI11
aI31
aI7
aI47
aI41
aI127
aI81
aI273
aI1649
aI975
aI3953
aI17021
aI24163
aI12599
asS'd'
I12049
sa(dp24098
S'a'
I53905
sS's'
I17
sS'm'
(lp24099
I1
aI3
aI1
aI3
aI27
aI41
aI75
aI237
aI317
aI85
aI1995
aI2255
aI2191
aI6441
aI26629
aI25797
aI97681
asS'd'
I12050
sa(dp24100
S'a'
I53912
sS's'
I17
sS'm'
(lp24101
I1
aI1
aI1
aI3
aI11
aI5
aI31
aI109
aI227
aI977
aI59
aI793
aI3305
aI10905
aI16529
aI21345
aI2403
asS'd'
I12051
sa(dp24102
S'a'
I53921
sS's'
I17
sS'm'
(lp24103
I1
aI3
aI5
aI9
aI9
aI37
aI107
aI129
aI421
aI383
aI1415
aI885
aI3383
aI9547
aI7303
aI41745
aI59919
asS'd'
I12052
sa(dp24104
S'a'
I53928
sS's'
I17
sS'm'
(lp24105
I1
aI1
aI7
aI7
aI29
aI27
aI59
aI177
aI97
aI299
aI1019
aI1393
aI7763
aI5715
aI9253
aI58035
aI23431
asS'd'
I12053
sa(dp24106
S'a'
I53948
sS's'
I17
sS'm'
(lp24107
I1
aI3
aI3
aI3
aI23
aI13
aI51
aI101
aI75
aI857
aI1699
aI2687
aI3983
aI10427
aI19845
aI49503
aI96033
asS'd'
I12054
sa(dp24108
S'a'
I53954
sS's'
I17
sS'm'
(lp24109
I1
aI1
aI1
aI7
aI21
aI51
aI25
aI71
aI265
aI999
aI1259
aI625
aI775
aI11045
aI20769
aI42597
aI115521
asS'd'
I12055
sa(dp24110
S'a'
I53968
sS's'
I17
sS'm'
(lp24111
I1
aI3
aI1
aI13
aI25
aI47
aI21
aI245
aI201
aI667
aI1193
aI1087
aI407
aI6057
aI14929
aI35291
aI57615
asS'd'
I12056
sa(dp24112
S'a'
I53977
sS's'
I17
sS'm'
(lp24113
I1
aI1
aI1
aI7
aI27
aI25
aI93
aI85
aI321
aI1009
aI1045
aI1901
aI349
aI11393
aI10917
aI10413
aI94125
asS'd'
I12057
sa(dp24114
S'a'
I53983
sS's'
I17
sS'm'
(lp24115
I1
aI1
aI1
aI15
aI3
aI63
aI59
aI51
aI307
aI135
aI785
aI1921
aI6921
aI5689
aI8487
aI21061
aI69903
asS'd'
I12058
sa(dp24116
S'a'
I53984
sS's'
I17
sS'm'
(lp24117
I1
aI3
aI7
aI1
aI13
aI47
aI59
aI155
aI107
aI573
aI843
aI2849
aI6685
aI5927
aI31747
aI21541
aI94271
asS'd'
I12059
sa(dp24118
S'a'
I54002
sS's'
I17
sS'm'
(lp24119
I1
aI1
aI7
aI15
aI23
aI7
aI85
aI169
aI209
aI527
aI1027
aI3745
aI4773
aI14893
aI10789
aI1243
aI87133
asS'd'
I12060
sa(dp24120
S'a'
I54007
sS's'
I17
sS'm'
(lp24121
I1
aI1
aI5
aI9
aI1
aI1
aI53
aI57
aI245
aI899
aI1785
aI1951
aI7651
aI10909
aI30167
aI40789
aI66965
asS'd'
I12061
sa(dp24122
S'a'
I54013
sS's'
I17
sS'm'
(lp24123
I1
aI1
aI1
aI7
aI17
aI33
aI65
aI79
aI455
aI677
aI157
aI1313
aI1573
aI9697
aI4161
aI4609
aI42783
asS'd'
I12062
sa(dp24124
S'a'
I54014
sS's'
I17
sS'm'
(lp24125
I1
aI3
aI7
aI7
aI27
aI15
aI109
aI113
aI239
aI521
aI563
aI2493
aI1471
aI15817
aI14515
aI48465
aI66009
asS'd'
I12063
sa(dp24126
S'a'
I54021
sS's'
I17
sS'm'
(lp24127
I1
aI3
aI5
aI3
aI29
aI33
aI125
aI169
aI483
aI593
aI1665
aI657
aI3799
aI15829
aI29591
aI25813
aI40987
asS'd'
I12064
sa(dp24128
S'a'
I54026
sS's'
I17
sS'm'
(lp24129
I1
aI3
aI1
aI11
aI15
aI25
aI21
aI215
aI341
aI241
aI1599
aI3807
aI6633
aI15137
aI15377
aI56881
aI47499
asS'd'
I12065
sa(dp24130
S'a'
I54028
sS's'
I17
sS'm'
(lp24131
I1
aI3
aI3
aI3
aI15
aI49
aI89
aI117
aI191
aI641
aI675
aI2671
aI4243
aI1617
aI20261
aI42669
aI119173
asS'd'
I12066
sa(dp24132
S'a'
I54031
sS's'
I17
sS'm'
(lp24133
I1
aI1
aI1
aI13
aI13
aI43
aI73
aI103
aI183
aI239
aI555
aI2121
aI4889
aI1431
aI20601
aI21545
aI11809
asS'd'
I12067
sa(dp24134
S'a'
I54036
sS's'
I17
sS'm'
(lp24135
I1
aI3
aI1
aI9
aI9
aI9
aI121
aI51
aI77
aI455
aI1481
aI427
aI1961
aI12149
aI21273
aI16295
aI21909
asS'd'
I12068
sa(dp24136
S'a'
I54067
sS's'
I17
sS'm'
(lp24137
I1
aI1
aI5
aI11
aI19
aI55
aI37
aI63
aI493
aI663
aI945
aI2191
aI2491
aI11545
aI7407
aI36295
aI94293
asS'd'
I12069
sa(dp24138
S'a'
I54069
sS's'
I17
sS'm'
(lp24139
I1
aI3
aI5
aI15
aI25
aI35
aI103
aI33
aI171
aI425
aI409
aI5
aI2519
aI2239
aI30557
aI20007
aI69079
asS'd'
I12070
sa(dp24140
S'a'
I54074
sS's'
I17
sS'm'
(lp24141
I1
aI1
aI5
aI11
aI13
aI29
aI71
aI21
aI35
aI833
aI191
aI365
aI7013
aI12413
aI10227
aI56705
aI61705
asS'd'
I12071
sa(dp24142
S'a'
I54076
sS's'
I17
sS'm'
(lp24143
I1
aI1
aI1
aI1
aI21
aI13
aI87
aI113
aI63
aI537
aI283
aI925
aI2147
aI1683
aI31239
aI2775
aI131021
asS'd'
I12072
sa(dp24144
S'a'
I54105
sS's'
I17
sS'm'
(lp24145
I1
aI1
aI3
aI9
aI23
aI1
aI117
aI19
aI487
aI235
aI877
aI149
aI369
aI9615
aI12501
aI60175
aI35741
asS'd'
I12073
sa(dp24146
S'a'
I54111
sS's'
I17
sS'm'
(lp24147
I1
aI1
aI7
aI9
aI5
aI25
aI107
aI199
aI339
aI755
aI245
aI2861
aI1119
aI14683
aI2221
aI5227
aI81479
asS'd'
I12074
sa(dp24148
S'a'
I54118
sS's'
I17
sS'm'
(lp24149
I1
aI1
aI1
aI15
aI5
aI15
aI37
aI63
aI511
aI219
aI783
aI3245
aI5563
aI13231
aI22311
aI16803
aI10393
asS'd'
I12075
sa(dp24150
S'a'
I54129
sS's'
I17
sS'm'
(lp24151
I1
aI3
aI7
aI5
aI1
aI15
aI9
aI21
aI287
aI991
aI555
aI771
aI7683
aI1661
aI6553
aI43735
aI118713
asS'd'
I12076
sa(dp24152
S'a'
I54145
sS's'
I17
sS'm'
(lp24153
I1
aI3
aI1
aI3
aI3
aI29
aI119
aI157
aI13
aI599
aI537
aI2921
aI5207
aI11621
aI1515
aI6351
aI118429
asS'd'
I12077
sa(dp24154
S'a'
I54157
sS's'
I17
sS'm'
(lp24155
I1
aI1
aI5
aI1
aI27
aI39
aI111
aI117
aI481
aI25
aI549
aI913
aI6427
aI7703
aI23099
aI50501
aI7617
asS'd'
I12078
sa(dp24156
S'a'
I54158
sS's'
I17
sS'm'
(lp24157
I1
aI1
aI7
aI5
aI29
aI63
aI43
aI151
aI63
aI43
aI197
aI3165
aI3879
aI12435
aI461
aI64475
aI60597
asS'd'
I12079
sa(dp24158
S'a'
I54163
sS's'
I17
sS'm'
(lp24159
I1
aI3
aI1
aI11
aI31
aI35
aI59
aI207
aI387
aI441
aI1293
aI2117
aI3751
aI12653
aI2811
aI42585
aI33297
asS'd'
I12080
sa(dp24160
S'a'
I54166
sS's'
I17
sS'm'
(lp24161
I1
aI3
aI7
aI15
aI27
aI47
aI13
aI15
aI135
aI433
aI615
aI1
aI171
aI11503
aI27117
aI64635
aI122191
asS'd'
I12081
sa(dp24162
S'a'
I54172
sS's'
I17
sS'm'
(lp24163
I1
aI1
aI7
aI1
aI23
aI23
aI107
aI135
aI311
aI395
aI373
aI2771
aI81
aI12513
aI16739
aI6715
aI94999
asS'd'
I12082
sa(dp24164
S'a'
I54185
sS's'
I17
sS'm'
(lp24165
I1
aI3
aI5
aI7
aI19
aI9
aI21
aI139
aI307
aI231
aI65
aI59
aI7767
aI2897
aI3503
aI58163
aI48807
asS'd'
I12083
sa(dp24166
S'a'
I54186
sS's'
I17
sS'm'
(lp24167
I1
aI3
aI5
aI13
aI23
aI5
aI51
aI247
aI125
aI911
aI1395
aI1337
aI3215
aI15811
aI12729
aI21495
aI22597
asS'd'
I12084
sa(dp24168
S'a'
I54188
sS's'
I17
sS'm'
(lp24169
I1
aI3
aI5
aI5
aI1
aI19
aI123
aI125
aI197
aI533
aI1699
aI1397
aI3473
aI15201
aI24493
aI3395
aI98261
asS'd'
I12085
sa(dp24170
S'a'
I54208
sS's'
I17
sS'm'
(lp24171
I1
aI1
aI3
aI7
aI29
aI39
aI69
aI97
aI353
aI293
aI1103
aI543
aI5015
aI9913
aI6965
aI61921
aI122073
asS'd'
I12086
sa(dp24172
S'a'
I54223
sS's'
I17
sS'm'
(lp24173
I1
aI1
aI3
aI13
aI19
aI41
aI117
aI253
aI449
aI231
aI865
aI3055
aI4751
aI3277
aI22863
aI3249
aI38359
asS'd'
I12087
sa(dp24174
S'a'
I54237
sS's'
I17
sS'm'
(lp24175
I1
aI3
aI5
aI13
aI9
aI7
aI107
aI17
aI251
aI501
aI1925
aI3733
aI5035
aI13213
aI12535
aI13705
aI73047
asS'd'
I12088
sa(dp24176
S'a'
I54241
sS's'
I17
sS'm'
(lp24177
I1
aI3
aI7
aI5
aI23
aI5
aI83
aI45
aI457
aI667
aI913
aI1167
aI7063
aI10915
aI10911
aI20501
aI61823
asS'd'
I12089
sa(dp24178
S'a'
I54244
sS's'
I17
sS'm'
(lp24179
I1
aI3
aI3
aI13
aI7
aI15
aI29
aI223
aI503
aI713
aI667
aI3989
aI5927
aI5909
aI27633
aI17615
aI97931
asS'd'
I12090
sa(dp24180
S'a'
I54259
sS's'
I17
sS'm'
(lp24181
I1
aI3
aI7
aI13
aI19
aI53
aI25
aI41
aI311
aI327
aI1323
aI3361
aI1095
aI12701
aI1065
aI34155
aI34705
asS'd'
I12091
sa(dp24182
S'a'
I54273
sS's'
I17
sS'm'
(lp24183
I1
aI1
aI7
aI7
aI11
aI35
aI63
aI73
aI179
aI477
aI467
aI4043
aI3097
aI16089
aI12749
aI18233
aI50299
asS'd'
I12092
sa(dp24184
S'a'
I54276
sS's'
I17
sS'm'
(lp24185
I1
aI3
aI3
aI13
aI5
aI27
aI31
aI207
aI357
aI469
aI607
aI961
aI7393
aI6707
aI25833
aI22337
aI119083
asS'd'
I12093
sa(dp24186
S'a'
I54280
sS's'
I17
sS'm'
(lp24187
I1
aI1
aI3
aI3
aI7
aI53
aI47
aI55
aI267
aI107
aI1307
aI2151
aI793
aI15605
aI12153
aI13075
aI76529
asS'd'
I12094
sa(dp24188
S'a'
I54294
sS's'
I17
sS'm'
(lp24189
I1
aI3
aI5
aI1
aI13
aI35
aI63
aI191
aI375
aI221
aI1603
aI2049
aI5363
aI1481
aI32271
aI22635
aI118603
asS'd'
I12095
sa(dp24190
S'a'
I54298
sS's'
I17
sS'm'
(lp24191
I1
aI1
aI1
aI11
aI17
aI63
aI13
aI3
aI353
aI943
aI443
aI141
aI7441
aI12335
aI4499
aI15923
aI63677
asS'd'
I12096
sa(dp24192
S'a'
I54303
sS's'
I17
sS'm'
(lp24193
I1
aI3
aI7
aI13
aI21
aI51
aI125
aI61
aI203
aI1
aI707
aI3893
aI4627
aI3125
aI14629
aI62721
aI85101
asS'd'
I12097
sa(dp24194
S'a'
I54304
sS's'
I17
sS'm'
(lp24195
I1
aI1
aI3
aI5
aI31
aI23
aI63
aI241
aI41
aI721
aI599
aI1761
aI2593
aI1685
aI31247
aI7811
aI87561
asS'd'
I12098
sa(dp24196
S'a'
I54309
sS's'
I17
sS'm'
(lp24197
I1
aI1
aI7
aI9
aI7
aI53
aI51
aI9
aI303
aI675
aI1261
aI1591
aI4363
aI15
aI29723
aI54533
aI103869
asS'd'
I12099
sa(dp24198
S'a'
I54310
sS's'
I17
sS'm'
(lp24199
I1
aI3
aI5
aI7
aI27
aI21
aI103
aI113
aI463
aI379
aI635
aI2363
aI607
aI11445
aI22475
aI58433
aI93071
asS'd'
I12100
sa(dp24200
S'a'
I54316
sS's'
I17
sS'm'
(lp24201
I1
aI1
aI5
aI5
aI5
aI63
aI23
aI67
aI399
aI279
aI829
aI945
aI6545
aI14951
aI5135
aI22889
aI87625
asS'd'
I12101
sa(dp24202
S'a'
I54336
sS's'
I17
sS'm'
(lp24203
I1
aI1
aI7
aI15
aI1
aI59
aI69
aI123
aI169
aI821
aI1125
aI2051
aI3375
aI11691
aI1379
aI57461
aI124209
asS'd'
I12102
sa(dp24204
S'a'
I54354
sS's'
I17
sS'm'
(lp24205
I1
aI1
aI5
aI15
aI31
aI57
aI51
aI59
aI297
aI459
aI701
aI241
aI2801
aI11893
aI4007
aI13165
aI79403
asS'd'
I12103
sa(dp24206
S'a'
I54359
sS's'
I17
sS'm'
(lp24207
I1
aI1
aI5
aI9
aI11
aI41
aI79
aI47
aI19
aI529
aI21
aI1871
aI371
aI6269
aI7433
aI36183
aI96113
asS'd'
I12104
sa(dp24208
S'a'
I54375
sS's'
I17
sS'm'
(lp24209
I1
aI1
aI7
aI5
aI29
aI3
aI33
aI191
aI119
aI501
aI1637
aI2903
aI3347
aI4581
aI17407
aI18169
aI10595
asS'd'
I12105
sa(dp24210
S'a'
I54376
sS's'
I17
sS'm'
(lp24211
I1
aI3
aI5
aI11
aI9
aI35
aI95
aI193
aI413
aI727
aI1157
aI3331
aI5993
aI1781
aI22653
aI3975
aI110557
asS'd'
I12106
sa(dp24212
S'a'
I54381
sS's'
I17
sS'm'
(lp24213
I1
aI1
aI1
aI1
aI23
aI5
aI35
aI65
aI57
aI515
aI569
aI4031
aI7983
aI4603
aI29419
aI44847
aI63601
asS'd'
I12107
sa(dp24214
S'a'
I54390
sS's'
I17
sS'm'
(lp24215
I1
aI3
aI1
aI7
aI21
aI5
aI77
aI23
aI317
aI803
aI723
aI3229
aI7171
aI2883
aI10943
aI50323
aI108579
asS'd'
I12108
sa(dp24216
S'a'
I54396
sS's'
I17
sS'm'
(lp24217
I1
aI3
aI5
aI11
aI15
aI53
aI75
aI127
aI177
aI19
aI501
aI1201
aI5113
aI9069
aI8817
aI14725
aI104737
asS'd'
I12109
sa(dp24218
S'a'
I54399
sS's'
I17
sS'm'
(lp24219
I1
aI3
aI7
aI9
aI7
aI39
aI5
aI121
aI409
aI103
aI1075
aI451
aI7603
aI16023
aI32557
aI43159
aI94385
asS'd'
I12110
sa(dp24220
S'a'
I54409
sS's'
I17
sS'm'
(lp24221
I1
aI3
aI1
aI11
aI29
aI57
aI123
aI141
aI57
aI945
aI1777
aI2427
aI2359
aI12839
aI7325
aI7945
aI129811
asS'd'
I12111
sa(dp24222
S'a'
I54424
sS's'
I17
sS'm'
(lp24223
I1
aI3
aI5
aI15
aI5
aI3
aI17
aI55
aI467
aI61
aI131
aI2891
aI6331
aI5859
aI20437
aI49425
aI80731
asS'd'
I12112
sa(dp24224
S'a'
I54434
sS's'
I17
sS'm'
(lp24225
I1
aI1
aI1
aI15
aI29
aI13
aI127
aI181
aI361
aI1019
aI1675
aI2755
aI6533
aI8957
aI14691
aI4285
aI65459
asS'd'
I12113
sa(dp24226
S'a'
I54436
sS's'
I17
sS'm'
(lp24227
I1
aI1
aI7
aI11
aI23
aI43
aI111
aI183
aI103
aI269
aI229
aI3291
aI1873
aI11349
aI29319
aI64829
aI19639
asS'd'
I12114
sa(dp24228
S'a'
I54439
sS's'
I17
sS'm'
(lp24229
I1
aI3
aI7
aI11
aI15
aI63
aI1
aI253
aI489
aI863
aI1707
aI2769
aI3201
aI7901
aI18161
aI12515
aI130237
asS'd'
I12115
sa(dp24230
S'a'
I54445
sS's'
I17
sS'm'
(lp24231
I1
aI3
aI1
aI7
aI1
aI25
aI43
aI159
aI505
aI511
aI1745
aI1421
aI6779
aI11103
aI23535
aI61129
aI124571
asS'd'
I12116
sa(dp24232
S'a'
I54448
sS's'
I17
sS'm'
(lp24233
I1
aI1
aI3
aI13
aI19
aI33
aI17
aI243
aI481
aI617
aI1061
aI1891
aI7165
aI6821
aI18505
aI8965
aI70179
asS'd'
I12117
sa(dp24234
S'a'
I54457
sS's'
I17
sS'm'
(lp24235
I1
aI1
aI7
aI13
aI17
aI17
aI65
aI23
aI255
aI361
aI1873
aI1605
aI2041
aI11119
aI11419
aI63131
aI49207
asS'd'
I12118
sa(dp24236
S'a'
I54465
sS's'
I17
sS'm'
(lp24237
I1
aI1
aI5
aI13
aI15
aI57
aI27
aI223
aI199
aI529
aI1115
aI1513
aI8083
aI11713
aI21005
aI50741
aI122223
asS'd'
I12119
sa(dp24238
S'a'
I54471
sS's'
I17
sS'm'
(lp24239
I1
aI3
aI5
aI15
aI29
aI35
aI107
aI85
aI141
aI505
aI1553
aI1283
aI4581
aI5077
aI9461
aI59853
aI23219
asS'd'
I12120
sa(dp24240
S'a'
I54472
sS's'
I17
sS'm'
(lp24241
I1
aI3
aI5
aI11
aI11
aI45
aI53
aI195
aI199
aI773
aI1911
aI721
aI1563
aI3769
aI3267
aI30673
aI80313
asS'd'
I12121
sa(dp24242
S'a'
I54485
sS's'
I17
sS'm'
(lp24243
I1
aI3
aI7
aI7
aI21
aI37
aI9
aI129
aI431
aI79
aI1559
aI2125
aI7781
aI6441
aI23533
aI46919
aI25315
asS'd'
I12122
sa(dp24244
S'a'
I54508
sS's'
I17
sS'm'
(lp24245
I1
aI1
aI5
aI15
aI11
aI61
aI77
aI231
aI349
aI647
aI225
aI85
aI6789
aI12557
aI6505
aI21985
aI54965
asS'd'
I12123
sa(dp24246
S'a'
I54513
sS's'
I17
sS'm'
(lp24247
I1
aI1
aI5
aI1
aI19
aI21
aI33
aI211
aI347
aI491
aI1119
aI1619
aI3739
aI11255
aI26705
aI59691
aI35337
asS'd'
I12124
sa(dp24248
S'a'
I54528
sS's'
I17
sS'm'
(lp24249
I1
aI1
aI3
aI3
aI29
aI15
aI7
aI23
aI279
aI145
aI699
aI289
aI475
aI1681
aI3201
aI64477
aI24919
asS'd'
I12125
sa(dp24250
S'a'
I54534
sS's'
I17
sS'm'
(lp24251
I1
aI1
aI7
aI7
aI23
aI53
aI75
aI71
aI315
aI403
aI1521
aI1417
aI3749
aI11243
aI3951
aI61039
aI51143
asS'd'
I12126
sa(dp24252
S'a'
I54537
sS's'
I17
sS'm'
(lp24253
I1
aI3
aI7
aI3
aI15
aI21
aI81
aI219
aI249
aI387
aI1405
aI3495
aI7143
aI2599
aI25435
aI15259
aI66069
asS'd'
I12127
sa(dp24254
S'a'
I54540
sS's'
I17
sS'm'
(lp24255
I1
aI1
aI5
aI7
aI31
aI9
aI63
aI55
aI409
aI421
aI1851
aI847
aI1593
aI10447
aI2833
aI13209
aI47285
asS'd'
I12128
sa(dp24256
S'a'
I54551
sS's'
I17
sS'm'
(lp24257
I1
aI1
aI3
aI7
aI15
aI3
aI35
aI49
aI253
aI21
aI1705
aI357
aI2751
aI9671
aI12429
aI4549
aI118691
asS'd'
I12129
sa(dp24258
S'a'
I54561
sS's'
I17
sS'm'
(lp24259
I1
aI1
aI5
aI15
aI1
aI3
aI97
aI197
aI43
aI923
aI1273
aI663
aI4291
aI12357
aI28221
aI15291
aI60989
asS'd'
I12130
sa(dp24260
S'a'
I54573
sS's'
I17
sS'm'
(lp24261
I1
aI1
aI7
aI7
aI15
aI3
aI35
aI115
aI449
aI641
aI743
aI1855
aI359
aI10983
aI2831
aI43983
aI56465
asS'd'
I12131
sa(dp24262
S'a'
I54579
sS's'
I17
sS'm'
(lp24263
I1
aI3
aI7
aI11
aI1
aI51
aI69
aI27
aI29
aI187
aI1673
aI1273
aI7987
aI1223
aI8971
aI53805
aI4413
asS'd'
I12132
sa(dp24264
S'a'
I54586
sS's'
I17
sS'm'
(lp24265
I1
aI1
aI3
aI11
aI1
aI55
aI91
aI241
aI35
aI97
aI1027
aI3967
aI703
aI3535
aI21681
aI55825
aI50423
asS'd'
I12133
sa(dp24266
S'a'
I54591
sS's'
I17
sS'm'
(lp24267
I1
aI1
aI7
aI11
aI21
aI53
aI111
aI125
aI11
aI355
aI1585
aI3603
aI1705
aI16311
aI7045
aI15503
aI63625
asS'd'
I12134
sa(dp24268
S'a'
I54596
sS's'
I17
sS'm'
(lp24269
I1
aI1
aI5
aI15
aI25
aI47
aI31
aI29
aI333
aI361
aI1831
aI1545
aI7751
aI8679
aI32453
aI61755
aI94637
asS'd'
I12135
sa(dp24270
S'a'
I54599
sS's'
I17
sS'm'
(lp24271
I1
aI1
aI3
aI3
aI23
aI3
aI79
aI11
aI367
aI551
aI281
aI1273
aI5097
aI12527
aI473
aI33855
aI85783
asS'd'
I12136
sa(dp24272
S'a'
I54605
sS's'
I17
sS'm'
(lp24273
I1
aI1
aI1
aI15
aI27
aI21
aI107
aI121
aI187
aI495
aI1877
aI1957
aI3647
aI13263
aI30729
aI18131
aI33689
asS'd'
I12137
sa(dp24274
S'a'
I54613
sS's'
I17
sS'm'
(lp24275
I1
aI1
aI5
aI13
aI3
aI43
aI41
aI53
aI127
aI299
aI839
aI3327
aI7929
aI9921
aI29471
aI18075
aI34283
asS'd'
I12138
sa(dp24276
S'a'
I54623
sS's'
I17
sS'm'
(lp24277
I1
aI1
aI7
aI13
aI31
aI5
aI59
aI75
aI335
aI929
aI379
aI139
aI7121
aI9281
aI31161
aI3177
aI2615
asS'd'
I12139
sa(dp24278
S'a'
I54654
sS's'
I17
sS'm'
(lp24279
I1
aI3
aI7
aI1
aI11
aI19
aI81
aI199
aI425
aI639
aI497
aI693
aI1271
aI7363
aI10543
aI52513
aI130549
asS'd'
I12140
sa(dp24280
S'a'
I54667
sS's'
I17
sS'm'
(lp24281
I1
aI3
aI3
aI7
aI21
aI41
aI101
aI67
aI363
aI5
aI1455
aI1433
aI81
aI15609
aI16231
aI13285
aI38995
asS'd'
I12141
sa(dp24282
S'a'
I54681
sS's'
I17
sS'm'
(lp24283
I1
aI1
aI3
aI15
aI11
aI19
aI123
aI177
aI429
aI27
aI141
aI3095
aI5379
aI2241
aI29877
aI59383
aI25199
asS'd'
I12142
sa(dp24284
S'a'
I54684
sS's'
I17
sS'm'
(lp24285
I1
aI1
aI7
aI7
aI19
aI63
aI93
aI217
aI279
aI349
aI149
aI2479
aI2355
aI6475
aI29993
aI37941
aI58715
asS'd'
I12143
sa(dp24286
S'a'
I54687
sS's'
I17
sS'm'
(lp24287
I1
aI3
aI7
aI3
aI21
aI23
aI59
aI177
aI489
aI817
aI1209
aI1629
aI5805
aI3137
aI23767
aI62967
aI16609
asS'd'
I12144
sa(dp24288
S'a'
I54694
sS's'
I17
sS'm'
(lp24289
I1
aI3
aI3
aI7
aI9
aI55
aI59
aI31
aI191
aI891
aI833
aI1059
aI3007
aI2331
aI385
aI58247
aI110697
asS'd'
I12145
sa(dp24290
S'a'
I54706
sS's'
I17
sS'm'
(lp24291
I1
aI1
aI3
aI3
aI11
aI61
aI9
aI189
aI79
aI621
aI209
aI2785
aI2959
aI4133
aI20691
aI45605
aI117089
asS'd'
I12146
sa(dp24292
S'a'
I54712
sS's'
I17
sS'm'
(lp24293
I1
aI3
aI3
aI5
aI5
aI47
aI31
aI1
aI451
aI765
aI2027
aI2327
aI1725
aI14341
aI7997
aI6449
aI77893
asS'd'
I12147
sa(dp24294
S'a'
I54715
sS's'
I17
sS'm'
(lp24295
I1
aI1
aI7
aI1
aI7
aI27
aI27
aI129
aI227
aI505
aI1461
aI783
aI945
aI12653
aI17913
aI61631
aI41875
asS'd'
I12148
sa(dp24296
S'a'
I54723
sS's'
I17
sS'm'
(lp24297
I1
aI1
aI1
aI11
aI13
aI41
aI41
aI221
aI483
aI825
aI451
aI493
aI1717
aI10389
aI7805
aI37707
aI30733
asS'd'
I12149
sa(dp24298
S'a'
I54725
sS's'
I17
sS'm'
(lp24299
I1
aI1
aI5
aI3
aI31
aI31
aI75
aI3
aI323
aI83
aI563
aI919
aI7387
aI1673
aI6157
aI7415
aI14947
asS'd'
I12150
sa(dp24300
S'a'
I54726
sS's'
I17
sS'm'
(lp24301
I1
aI1
aI7
aI13
aI19
aI37
aI29
aI93
aI153
aI491
aI1033
aI1389
aI6361
aI11133
aI20049
aI24585
aI107325
asS'd'
I12151
sa(dp24302
S'a'
I54740
sS's'
I17
sS'm'
(lp24303
I1
aI1
aI1
aI15
aI3
aI35
aI79
aI251
aI383
aI665
aI2033
aI3165
aI3411
aI15965
aI28281
aI56521
aI56479
asS'd'
I12152
sa(dp24304
S'a'
I54750
sS's'
I17
sS'm'
(lp24305
I1
aI3
aI1
aI15
aI23
aI1
aI87
aI145
aI443
aI405
aI635
aI1597
aI1455
aI5983
aI12741
aI55133
aI2815
asS'd'
I12153
sa(dp24306
S'a'
I54760
sS's'
I17
sS'm'
(lp24307
I1
aI1
aI1
aI13
aI11
aI25
aI19
aI129
aI23
aI913
aI1121
aI223
aI1991
aI13449
aI30443
aI50573
aI108467
asS'd'
I12154
sa(dp24308
S'a'
I54768
sS's'
I17
sS'm'
(lp24309
I1
aI1
aI7
aI11
aI29
aI31
aI49
aI51
aI415
aI293
aI173
aI4091
aI159
aI2679
aI30643
aI58725
aI109287
asS'd'
I12155
sa(dp24310
S'a'
I54771
sS's'
I17
sS'm'
(lp24311
I1
aI1
aI1
aI7
aI15
aI53
aI69
aI231
aI387
aI693
aI1299
aI1383
aI7935
aI10313
aI22403
aI59341
aI3347
asS'd'
I12156
sa(dp24312
S'a'
I54773
sS's'
I17
sS'm'
(lp24313
I1
aI3
aI3
aI5
aI9
aI21
aI111
aI11
aI469
aI109
aI1565
aI3107
aI2975
aI12491
aI26773
aI33245
aI27589
asS'd'
I12157
sa(dp24314
S'a'
I54796
sS's'
I17
sS'm'
(lp24315
I1
aI1
aI3
aI5
aI3
aI9
aI103
aI127
aI345
aI301
aI857
aI2035
aI3269
aI13819
aI7555
aI5197
aI94557
asS'd'
I12158
sa(dp24316
S'a'
I54801
sS's'
I17
sS'm'
(lp24317
I1
aI3
aI7
aI3
aI31
aI3
aI61
aI253
aI221
aI359
aI1281
aI1405
aI4819
aI1329
aI17773
aI29539
aI127043
asS'd'
I12159
sa(dp24318
S'a'
I54808
sS's'
I17
sS'm'
(lp24319
I1
aI1
aI7
aI11
aI17
aI47
aI105
aI253
aI253
aI531
aI119
aI2009
aI6125
aI9387
aI13141
aI29079
aI28361
asS'd'
I12160
sa(dp24320
S'a'
I54814
sS's'
I17
sS'm'
(lp24321
I1
aI1
aI3
aI5
aI21
aI13
aI21
aI223
aI79
aI819
aI1425
aI1001
aI6517
aI8883
aI29997
aI30637
aI7717
asS'd'
I12161
sa(dp24322
S'a'
I54837
sS's'
I17
sS'm'
(lp24323
I1
aI3
aI1
aI3
aI1
aI23
aI113
aI69
aI235
aI95
aI1873
aI689
aI4611
aI13209
aI12681
aI16057
aI114071
asS'd'
I12162
sa(dp24324
S'a'
I54847
sS's'
I17
sS'm'
(lp24325
I1
aI3
aI3
aI13
aI25
aI21
aI93
aI55
aI253
aI373
aI1659
aI829
aI6539
aI7453
aI28195
aI33131
aI92559
asS'd'
I12163
sa(dp24326
S'a'
I54849
sS's'
I17
sS'm'
(lp24327
I1
aI1
aI3
aI11
aI25
aI29
aI81
aI235
aI429
aI811
aI1867
aI2923
aI5949
aI4423
aI93
aI64631
aI48357
asS'd'
I12164
sa(dp24328
S'a'
I54864
sS's'
I17
sS'm'
(lp24329
I1
aI3
aI5
aI13
aI29
aI27
aI35
aI15
aI105
aI849
aI247
aI3999
aI6441
aI12443
aI19817
aI49897
aI21515
asS'd'
I12165
sa(dp24330
S'a'
I54867
sS's'
I17
sS'm'
(lp24331
I1
aI1
aI5
aI15
aI13
aI59
aI3
aI199
aI267
aI463
aI655
aI3875
aI2895
aI13411
aI5081
aI22069
aI6053
asS'd'
I12166
sa(dp24332
S'a'
I54889
sS's'
I17
sS'm'
(lp24333
I1
aI1
aI5
aI9
aI5
aI13
aI111
aI83
aI281
aI543
aI629
aI1349
aI1863
aI9523
aI19201
aI39229
aI78265
asS'd'
I12167
sa(dp24334
S'a'
I54895
sS's'
I17
sS'm'
(lp24335
I1
aI3
aI7
aI1
aI29
aI23
aI109
aI75
aI347
aI643
aI97
aI1981
aI2797
aI11201
aI28355
aI54105
aI45551
asS'd'
I12168
sa(dp24336
S'a'
I54907
sS's'
I17
sS'm'
(lp24337
I1
aI3
aI1
aI7
aI9
aI5
aI77
aI17
aI179
aI957
aI621
aI779
aI7117
aI1491
aI11563
aI10131
aI98335
asS'd'
I12169
sa(dp24338
S'a'
I54919
sS's'
I17
sS'm'
(lp24339
I1
aI3
aI3
aI1
aI3
aI53
aI39
aI217
aI309
aI105
aI485
aI3123
aI3143
aI2359
aI4923
aI22307
aI120319
asS'd'
I12170
sa(dp24340
S'a'
I54923
sS's'
I17
sS'm'
(lp24341
I1
aI1
aI7
aI7
aI11
aI5
aI65
aI165
aI321
aI455
aI625
aI2417
aI999
aI14999
aI6777
aI13319
aI43399
asS'd'
I12171
sa(dp24342
S'a'
I54926
sS's'
I17
sS'm'
(lp24343
I1
aI1
aI7
aI3
aI11
aI55
aI43
aI119
aI135
aI129
aI581
aI3593
aI3475
aI14667
aI30509
aI5007
aI120135
asS'd'
I12172
sa(dp24344
S'a'
I54940
sS's'
I17
sS'm'
(lp24345
I1
aI1
aI1
aI11
aI7
aI17
aI95
aI169
aI401
aI87
aI1425
aI1821
aI7619
aI3605
aI10993
aI35837
aI87311
asS'd'
I12173
sa(dp24346
S'a'
I54950
sS's'
I17
sS'm'
(lp24347
I1
aI1
aI7
aI11
aI11
aI35
aI29
aI63
aI395
aI301
aI373
aI2457
aI6859
aI1915
aI11215
aI41075
aI78219
asS'd'
I12174
sa(dp24348
S'a'
I54954
sS's'
I17
sS'm'
(lp24349
I1
aI3
aI5
aI7
aI25
aI3
aI97
aI43
aI273
aI459
aI103
aI3441
aI71
aI10269
aI29893
aI46053
aI104801
asS'd'
I12175
sa(dp24350
S'a'
I54961
sS's'
I17
sS'm'
(lp24351
I1
aI3
aI3
aI7
aI31
aI3
aI121
aI255
aI73
aI783
aI977
aI513
aI6527
aI1189
aI8925
aI23245
aI22287
asS'd'
I12176
sa(dp24352
S'a'
I54973
sS's'
I17
sS'm'
(lp24353
I1
aI3
aI3
aI13
aI15
aI53
aI51
aI135
aI465
aI341
aI263
aI1687
aI4085
aI14257
aI18745
aI46945
aI115475
asS'd'
I12177
sa(dp24354
S'a'
I54974
sS's'
I17
sS'm'
(lp24355
I1
aI3
aI1
aI5
aI31
aI1
aI1
aI91
aI511
aI771
aI1501
aI2613
aI991
aI3859
aI28911
aI65417
aI201
asS'd'
I12178
sa(dp24356
S'a'
I54976
sS's'
I17
sS'm'
(lp24357
I1
aI1
aI7
aI1
aI27
aI21
aI107
aI153
aI163
aI949
aI811
aI3087
aI3443
aI5621
aI28795
aI58311
aI63763
asS'd'
I12179
sa(dp24358
S'a'
I54986
sS's'
I17
sS'm'
(lp24359
I1
aI1
aI7
aI5
aI29
aI29
aI57
aI175
aI29
aI821
aI1545
aI2643
aI725
aI16225
aI29111
aI19675
aI129995
asS'd'
I12180
sa(dp24360
S'a'
I54993
sS's'
I17
sS'm'
(lp24361
I1
aI1
aI7
aI3
aI31
aI31
aI61
aI155
aI265
aI323
aI1829
aI3891
aI6393
aI8573
aI10627
aI10839
aI78683
asS'd'
I12181
sa(dp24362
S'a'
I55000
sS's'
I17
sS'm'
(lp24363
I1
aI3
aI5
aI7
aI29
aI7
aI67
aI181
aI313
aI731
aI1761
aI1681
aI3673
aI8939
aI811
aI48931
aI82021
asS'd'
I12182
sa(dp24364
S'a'
I55010
sS's'
I17
sS'm'
(lp24365
I1
aI3
aI1
aI3
aI11
aI51
aI81
aI67
aI173
aI881
aI855
aI3627
aI1613
aI4825
aI7035
aI36261
aI64899
asS'd'
I12183
sa(dp24366
S'a'
I55019
sS's'
I17
sS'm'
(lp24367
I1
aI3
aI7
aI7
aI15
aI53
aI123
aI41
aI265
aI817
aI807
aI3875
aI7675
aI16225
aI13313
aI62217
aI47647
asS'd'
I12184
sa(dp24368
S'a'
I55021
sS's'
I17
sS'm'
(lp24369
I1
aI1
aI3
aI13
aI23
aI47
aI125
aI155
aI403
aI651
aI1693
aI2185
aI5565
aI9947
aI20893
aI11287
aI118943
asS'd'
I12185
sa(dp24370
S'a'
I55030
sS's'
I17
sS'm'
(lp24371
I1
aI3
aI7
aI3
aI19
aI47
aI69
aI5
aI209
aI259
aI367
aI3929
aI7579
aI12687
aI18109
aI51885
aI128281
asS'd'
I12186
sa(dp24372
S'a'
I55033
sS's'
I17
sS'm'
(lp24373
I1
aI1
aI7
aI5
aI27
aI41
aI45
aI41
aI205
aI1001
aI1509
aI2649
aI1141
aI5355
aI10265
aI34131
aI112111
asS'd'
I12187
sa(dp24374
S'a'
I55039
sS's'
I17
sS'm'
(lp24375
I1
aI3
aI7
aI5
aI19
aI41
aI103
aI63
aI49
aI25
aI271
aI793
aI3217
aI4741
aI2563
aI61333
aI113205
asS'd'
I12188
sa(dp24376
S'a'
I55047
sS's'
I17
sS'm'
(lp24377
I1
aI3
aI3
aI15
aI15
aI35
aI13
aI233
aI277
aI673
aI545
aI545
aI7419
aI6707
aI1867
aI58873
aI110027
asS'd'
I12189
sa(dp24378
S'a'
I55048
sS's'
I17
sS'm'
(lp24379
I1
aI1
aI7
aI3
aI9
aI23
aI67
aI55
aI3
aI1019
aI2001
aI2909
aI7311
aI9295
aI26953
aI43217
aI54597
asS'd'
I12190
sa(dp24380
S'a'
I55054
sS's'
I17
sS'm'
(lp24381
I1
aI3
aI5
aI7
aI13
aI33
aI67
aI27
aI75
aI569
aI1777
aI791
aI1223
aI1805
aI19167
aI60537
aI60039
asS'd'
I12191
sa(dp24382
S'a'
I55059
sS's'
I17
sS'm'
(lp24383
I1
aI1
aI5
aI13
aI15
aI61
aI49
aI59
aI289
aI907
aI1055
aI3579
aI8169
aI12119
aI25479
aI32867
aI65343
asS'd'
I12192
sa(dp24384
S'a'
I55068
sS's'
I17
sS'm'
(lp24385
I1
aI3
aI5
aI9
aI5
aI63
aI91
aI225
aI377
aI469
aI891
aI891
aI5115
aI11487
aI30151
aI44357
aI120077
asS'd'
I12193
sa(dp24386
S'a'
I55071
sS's'
I17
sS'm'
(lp24387
I1
aI1
aI1
aI15
aI29
aI59
aI19
aI51
aI295
aI585
aI149
aI497
aI5837
aI11629
aI7825
aI18129
aI113797
asS'd'
I12194
sa(dp24388
S'a'
I55075
sS's'
I17
sS'm'
(lp24389
I1
aI1
aI3
aI7
aI31
aI25
aI77
aI209
aI183
aI337
aI1753
aI2703
aI2559
aI11847
aI17349
aI27359
aI21771
asS'd'
I12195
sa(dp24390
S'a'
I55077
sS's'
I17
sS'm'
(lp24391
I1
aI1
aI7
aI7
aI13
aI23
aI69
aI61
aI353
aI339
aI833
aI1935
aI4333
aI10521
aI20331
aI62145
aI59245
asS'd'
I12196
sa(dp24392
S'a'
I55081
sS's'
I17
sS'm'
(lp24393
I1
aI1
aI5
aI13
aI19
aI57
aI35
aI59
aI203
aI99
aI487
aI2747
aI637
aI8213
aI27053
aI29
aI64335
asS'd'
I12197
sa(dp24394
S'a'
I55095
sS's'
I17
sS'm'
(lp24395
I1
aI3
aI5
aI7
aI27
aI5
aI71
aI147
aI339
aI313
aI913
aI2845
aI5713
aI4383
aI18969
aI54871
aI51931
asS'd'
I12198
sa(dp24396
S'a'
I55099
sS's'
I17
sS'm'
(lp24397
I1
aI1
aI1
aI5
aI7
aI23
aI19
aI11
aI111
aI543
aI311
aI1519
aI387
aI10175
aI18209
aI14115
aI123421
asS'd'
I12199
sa(dp24398
S'a'
I55110
sS's'
I17
sS'm'
(lp24399
I1
aI3
aI7
aI11
aI7
aI7
aI123
aI193
aI417
aI65
aI1317
aI3821
aI2315
aI14527
aI14113
aI25873
aI23977
asS'd'
I12200
sa(dp24400
S'a'
I55116
sS's'
I17
sS'm'
(lp24401
I1
aI1
aI3
aI15
aI21
aI11
aI3
aI37
aI115
aI395
aI877
aI1227
aI6997
aI4357
aI11397
aI52855
aI24899
asS'd'
I12201
sa(dp24402
S'a'
I55122
sS's'
I17
sS'm'
(lp24403
I1
aI1
aI7
aI5
aI15
aI45
aI45
aI17
aI441
aI605
aI429
aI739
aI4759
aI5249
aI11311
aI55049
aI56909
asS'd'
I12202
sa(dp24404
S'a'
I55134
sS's'
I17
sS'm'
(lp24405
I1
aI1
aI1
aI5
aI3
aI5
aI77
aI31
aI407
aI703
aI385
aI235
aI7751
aI617
aI16013
aI27269
aI66971
asS'd'
I12203
sa(dp24406
S'a'
I55144
sS's'
I17
sS'm'
(lp24407
I1
aI3
aI3
aI15
aI25
aI27
aI19
aI251
aI465
aI197
aI1039
aI3261
aI4557
aI4821
aI16083
aI43997
aI61371
asS'd'
I12204
sa(dp24408
S'a'
I55147
sS's'
I17
sS'm'
(lp24409
I1
aI3
aI3
aI15
aI21
aI45
aI13
aI139
aI213
aI797
aI619
aI2125
aI3805
aI4149
aI11427
aI59807
aI104587
asS'd'
I12205
sa(dp24410
S'a'
I55158
sS's'
I17
sS'm'
(lp24411
I1
aI3
aI1
aI1
aI29
aI27
aI25
aI7
aI371
aI535
aI1613
aI1083
aI4221
aI8913
aI10601
aI6447
aI17619
asS'd'
I12206
sa(dp24412
S'a'
I55162
sS's'
I17
sS'm'
(lp24413
I1
aI1
aI3
aI3
aI13
aI35
aI37
aI127
aI285
aI899
aI307
aI123
aI129
aI14035
aI26503
aI64103
aI27155
asS'd'
I12207
sa(dp24414
S'a'
I55171
sS's'
I17
sS'm'
(lp24415
I1
aI1
aI7
aI7
aI27
aI25
aI45
aI245
aI271
aI281
aI69
aI3505
aI7087
aI1529
aI7121
aI30327
aI89131
asS'd'
I12208
sa(dp24416
S'a'
I55185
sS's'
I17
sS'm'
(lp24417
I1
aI1
aI3
aI11
aI13
aI57
aI31
aI23
aI455
aI427
aI1683
aI3019
aI5827
aI8817
aI12943
aI321
aI39951
asS'd'
I12209
sa(dp24418
S'a'
I55192
sS's'
I17
sS'm'
(lp24419
I1
aI3
aI1
aI3
aI31
aI41
aI69
aI211
aI385
aI275
aI1569
aI2265
aI4017
aI11057
aI15
aI16619
aI126967
asS'd'
I12210
sa(dp24420
S'a'
I55197
sS's'
I17
sS'm'
(lp24421
I1
aI3
aI3
aI1
aI27
aI27
aI21
aI145
aI125
aI929
aI1371
aI1469
aI1591
aI5283
aI4651
aI1265
aI17161
asS'd'
I12211
sa(dp24422
S'a'
I55207
sS's'
I17
sS'm'
(lp24423
I1
aI3
aI7
aI5
aI29
aI31
aI41
aI141
aI49
aI967
aI1421
aI663
aI6089
aI3831
aI11353
aI38809
aI108605
asS'd'
I12212
sa(dp24424
S'a'
I55228
sS's'
I17
sS'm'
(lp24425
I1
aI1
aI7
aI15
aI11
aI23
aI91
aI31
aI9
aI717
aI265
aI1729
aI3563
aI8145
aI20441
aI22933
aI103683
asS'd'
I12213
sa(dp24426
S'a'
I55246
sS's'
I17
sS'm'
(lp24427
I1
aI1
aI7
aI3
aI17
aI13
aI47
aI13
aI241
aI1017
aI1803
aI2091
aI7379
aI2941
aI11783
aI36189
aI53513
asS'd'
I12214
sa(dp24428
S'a'
I55253
sS's'
I17
sS'm'
(lp24429
I1
aI3
aI1
aI11
aI31
aI63
aI107
aI79
aI427
aI385
aI1497
aI1265
aI5135
aI13597
aI27343
aI56733
aI100595
asS'd'
I12215
sa(dp24430
S'a'
I55254
sS's'
I17
sS'm'
(lp24431
I1
aI3
aI7
aI3
aI9
aI15
aI119
aI29
aI205
aI151
aI1453
aI3575
aI3627
aI7815
aI3553
aI31457
aI14267
asS'd'
I12216
sa(dp24432
S'a'
I55281
sS's'
I17
sS'm'
(lp24433
I1
aI3
aI1
aI7
aI15
aI21
aI73
aI47
aI417
aI29
aI1231
aI2477
aI161
aI15997
aI4457
aI3939
aI43929
asS'd'
I12217
sa(dp24434
S'a'
I55300
sS's'
I17
sS'm'
(lp24435
I1
aI1
aI5
aI5
aI19
aI49
aI103
aI251
aI359
aI69
aI669
aI299
aI8161
aI10579
aI13999
aI26859
aI92199
asS'd'
I12218
sa(dp24436
S'a'
I55303
sS's'
I17
sS'm'
(lp24437
I1
aI1
aI3
aI1
aI9
aI27
aI81
aI7
aI115
aI29
aI1067
aI1933
aI3061
aI2885
aI27883
aI65227
aI59365
asS'd'
I12219
sa(dp24438
S'a'
I55307
sS's'
I17
sS'm'
(lp24439
I1
aI1
aI1
aI5
aI23
aI17
aI1
aI113
aI495
aI155
aI1673
aI3945
aI8053
aI7935
aI3537
aI65141
aI11809
asS'd'
I12220
sa(dp24440
S'a'
I55312
sS's'
I17
sS'm'
(lp24441
I1
aI1
aI1
aI15
aI15
aI59
aI61
aI213
aI303
aI851
aI1893
aI615
aI6659
aI9351
aI16621
aI6097
aI114383
asS'd'
I12221
sa(dp24442
S'a'
I55328
sS's'
I17
sS'm'
(lp24443
I1
aI3
aI1
aI7
aI19
aI11
aI95
aI127
aI277
aI677
aI1631
aI2563
aI3295
aI7029
aI4059
aI44079
aI128857
asS'd'
I12222
sa(dp24444
S'a'
I55331
sS's'
I17
sS'm'
(lp24445
I1
aI3
aI7
aI11
aI27
aI49
aI99
aI43
aI279
aI771
aI123
aI2969
aI699
aI12915
aI22039
aI62257
aI79359
asS'd'
I12223
sa(dp24446
S'a'
I55337
sS's'
I17
sS'm'
(lp24447
I1
aI1
aI7
aI3
aI19
aI45
aI45
aI113
aI251
aI883
aI715
aI1541
aI1573
aI3345
aI23855
aI62681
aI57591
asS'd'
I12224
sa(dp24448
S'a'
I55348
sS's'
I17
sS'm'
(lp24449
I1
aI1
aI5
aI15
aI11
aI1
aI51
aI15
aI135
aI519
aI961
aI1447
aI4425
aI2139
aI3309
aI35111
aI74143
asS'd'
I12225
sa(dp24450
S'a'
I55352
sS's'
I17
sS'm'
(lp24451
I1
aI3
aI7
aI7
aI17
aI39
aI109
aI25
aI11
aI549
aI315
aI2175
aI685
aI11837
aI9151
aI6277
aI45011
asS'd'
I12226
sa(dp24452
S'a'
I55357
sS's'
I17
sS'm'
(lp24453
I1
aI1
aI1
aI9
aI27
aI7
aI95
aI1
aI385
aI167
aI453
aI1027
aI4105
aI16351
aI19
aI10375
aI62833
asS'd'
I12227
sa(dp24454
S'a'
I55372
sS's'
I17
sS'm'
(lp24455
I1
aI3
aI7
aI13
aI17
aI19
aI107
aI11
aI441
aI171
aI185
aI3567
aI1245
aI12161
aI30257
aI48105
aI87105
asS'd'
I12228
sa(dp24456
S'a'
I55375
sS's'
I17
sS'm'
(lp24457
I1
aI3
aI3
aI9
aI15
aI5
aI109
aI225
aI85
aI919
aI513
aI3559
aI5411
aI9009
aI27391
aI25115
aI84875
asS'd'
I12229
sa(dp24458
S'a'
I55377
sS's'
I17
sS'm'
(lp24459
I1
aI3
aI3
aI7
aI11
aI37
aI81
aI51
aI121
aI25
aI1897
aI2121
aI6425
aI16087
aI4259
aI29501
aI118067
asS'd'
I12230
sa(dp24460
S'a'
I55394
sS's'
I17
sS'm'
(lp24461
I1
aI1
aI7
aI3
aI5
aI53
aI73
aI127
aI137
aI739
aI543
aI1723
aI1163
aI2791
aI18519
aI1459
aI50869
asS'd'
I12231
sa(dp24462
S'a'
I55406
sS's'
I17
sS'm'
(lp24463
I1
aI1
aI3
aI11
aI29
aI51
aI101
aI189
aI193
aI839
aI25
aI3109
aI3035
aI3917
aI23929
aI38577
aI129705
asS'd'
I12232
sa(dp24464
S'a'
I55424
sS's'
I17
sS'm'
(lp24465
I1
aI3
aI7
aI15
aI9
aI29
aI93
aI101
aI271
aI791
aI1257
aI1843
aI2701
aI8205
aI15195
aI9109
aI120835
asS'd'
I12233
sa(dp24466
S'a'
I55429
sS's'
I17
sS'm'
(lp24467
I1
aI3
aI3
aI5
aI29
aI47
aI31
aI135
aI483
aI385
aI1395
aI2955
aI7291
aI12885
aI9491
aI14581
aI66293
asS'd'
I12234
sa(dp24468
S'a'
I55430
sS's'
I17
sS'm'
(lp24469
I1
aI1
aI7
aI1
aI5
aI37
aI105
aI149
aI63
aI617
aI1611
aI3025
aI3177
aI15463
aI3373
aI3503
aI95001
asS'd'
I12235
sa(dp24470
S'a'
I55436
sS's'
I17
sS'm'
(lp24471
I1
aI1
aI5
aI13
aI1
aI57
aI19
aI35
aI127
aI423
aI1221
aI1547
aI4083
aI347
aI17131
aI60087
aI27437
asS'd'
I12236
sa(dp24472
S'a'
I55439
sS's'
I17
sS'm'
(lp24473
I1
aI3
aI7
aI9
aI25
aI1
aI105
aI39
aI25
aI921
aI1897
aI1729
aI2207
aI7761
aI24197
aI457
aI64241
asS'd'
I12237
sa(dp24474
S'a'
I55447
sS's'
I17
sS'm'
(lp24475
I1
aI3
aI7
aI9
aI15
aI21
aI13
aI113
aI379
aI1021
aI489
aI1757
aI5869
aI4833
aI24717
aI52227
aI3209
asS'd'
I12238
sa(dp24476
S'a'
I55453
sS's'
I17
sS'm'
(lp24477
I1
aI3
aI1
aI5
aI9
aI61
aI25
aI41
aI183
aI473
aI383
aI2259
aI6939
aI3
aI32161
aI6319
aI93099
asS'd'
I12239
sa(dp24478
S'a'
I55458
sS's'
I17
sS'm'
(lp24479
I1
aI1
aI5
aI13
aI13
aI47
aI97
aI3
aI357
aI837
aI1655
aI485
aI4251
aI12153
aI9013
aI25121
aI51877
asS'd'
I12240
sa(dp24480
S'a'
I55460
sS's'
I17
sS'm'
(lp24481
I1
aI1
aI1
aI15
aI7
aI59
aI65
aI119
aI467
aI313
aI1333
aI2007
aI5165
aI13935
aI13679
aI3999
aI81811
asS'd'
I12241
sa(dp24482
S'a'
I55478
sS's'
I17
sS'm'
(lp24483
I1
aI1
aI5
aI13
aI13
aI1
aI63
aI117
aI449
aI13
aI1017
aI1583
aI7599
aI3669
aI32699
aI59455
aI32363
asS'd'
I12242
sa(dp24484
S'a'
I55489
sS's'
I17
sS'm'
(lp24485
I1
aI1
aI7
aI9
aI7
aI15
aI37
aI251
aI167
aI25
aI1085
aI2067
aI2771
aI5737
aI20661
aI19231
aI59547
asS'd'
I12243
sa(dp24486
S'a'
I55501
sS's'
I17
sS'm'
(lp24487
I1
aI1
aI5
aI13
aI29
aI11
aI63
aI37
aI281
aI657
aI1567
aI2879
aI7601
aI15617
aI16527
aI51695
aI5583
asS'd'
I12244
sa(dp24488
S'a'
I55502
sS's'
I17
sS'm'
(lp24489
I1
aI1
aI5
aI3
aI31
aI17
aI19
aI65
aI315
aI413
aI927
aI3617
aI4089
aI11899
aI3759
aI47991
aI1685
asS'd'
I12245
sa(dp24490
S'a'
I55519
sS's'
I17
sS'm'
(lp24491
I1
aI1
aI1
aI1
aI13
aI47
aI89
aI91
aI379
aI429
aI283
aI3765
aI2923
aI14955
aI26399
aI9579
aI39817
asS'd'
I12246
sa(dp24492
S'a'
I55525
sS's'
I17
sS'm'
(lp24493
I1
aI1
aI1
aI9
aI1
aI17
aI91
aI119
aI327
aI291
aI39
aI2883
aI6265
aI553
aI7559
aI60577
aI34393
asS'd'
I12247
sa(dp24494
S'a'
I55530
sS's'
I17
sS'm'
(lp24495
I1
aI3
aI5
aI5
aI9
aI33
aI123
aI219
aI103
aI529
aI181
aI1321
aI6815
aI2411
aI10555
aI43911
aI18889
asS'd'
I12248
sa(dp24496
S'a'
I55544
sS's'
I17
sS'm'
(lp24497
I1
aI1
aI5
aI7
aI9
aI13
aI7
aI45
aI427
aI523
aI1189
aI255
aI2103
aI7217
aI16249
aI14631
aI90409
asS'd'
I12249
sa(dp24498
S'a'
I55550
sS's'
I17
sS'm'
(lp24499
I1
aI1
aI1
aI9
aI11
aI35
aI55
aI71
aI89
aI637
aI1417
aI411
aI5305
aI10125
aI20715
aI62927
aI4993
asS'd'
I12250
sa(dp24500
S'a'
I55558
sS's'
I17
sS'm'
(lp24501
I1
aI3
aI1
aI9
aI1
aI59
aI27
aI221
aI267
aI797
aI1081
aI951
aI1369
aI2677
aI20763
aI63301
aI61963
asS'd'
I12251
sa(dp24502
S'a'
I55561
sS's'
I17
sS'm'
(lp24503
I1
aI3
aI5
aI5
aI17
aI9
aI67
aI177
aI89
aI953
aI1329
aI1649
aI989
aI7773
aI28747
aI26231
aI42331
asS'd'
I12252
sa(dp24504
S'a'
I55569
sS's'
I17
sS'm'
(lp24505
I1
aI3
aI3
aI9
aI23
aI35
aI17
aI145
aI53
aI519
aI1173
aI2079
aI2593
aI3633
aI32005
aI30573
aI55651
asS'd'
I12253
sa(dp24506
S'a'
I55586
sS's'
I17
sS'm'
(lp24507
I1
aI1
aI7
aI7
aI17
aI41
aI47
aI253
aI107
aI843
aI9
aI323
aI2391
aI3267
aI25813
aI1741
aI93493
asS'd'
I12254
sa(dp24508
S'a'
I55591
sS's'
I17
sS'm'
(lp24509
I1
aI1
aI1
aI9
aI31
aI43
aI47
aI91
aI235
aI569
aI2017
aI2385
aI5055
aI5747
aI26471
aI48819
aI47315
asS'd'
I12255
sa(dp24510
S'a'
I55598
sS's'
I17
sS'm'
(lp24511
I1
aI3
aI5
aI9
aI1
aI17
aI87
aI91
aI55
aI287
aI995
aI2577
aI1151
aI9119
aI22791
aI50899
aI16423
asS'd'
I12256
sa(dp24512
S'a'
I55605
sS's'
I17
sS'm'
(lp24513
I1
aI3
aI7
aI1
aI3
aI29
aI9
aI193
aI269
aI201
aI325
aI2209
aI1061
aI7957
aI23265
aI65083
aI27575
asS'd'
I12257
sa(dp24514
S'a'
I55609
sS's'
I17
sS'm'
(lp24515
I1
aI3
aI7
aI15
aI27
aI23
aI37
aI239
aI165
aI959
aI1965
aI2105
aI1581
aI6621
aI17315
aI49255
aI62487
asS'd'
I12258
sa(dp24516
S'a'
I55618
sS's'
I17
sS'm'
(lp24517
I1
aI3
aI7
aI7
aI11
aI31
aI73
aI145
aI429
aI421
aI571
aI3375
aI2797
aI15889
aI26523
aI12315
aI48061
asS'd'
I12259
sa(dp24518
S'a'
I55630
sS's'
I17
sS'm'
(lp24519
I1
aI3
aI3
aI15
aI23
aI27
aI105
aI75
aI497
aI137
aI475
aI1343
aI537
aI10499
aI27807
aI46623
aI32435
asS'd'
I12260
sa(dp24520
S'a'
I55642
sS's'
I17
sS'm'
(lp24521
I1
aI1
aI7
aI15
aI15
aI11
aI51
aI107
aI225
aI557
aI1461
aI3447
aI1243
aI13827
aI23675
aI26139
aI54603
asS'd'
I12261
sa(dp24522
S'a'
I55644
sS's'
I17
sS'm'
(lp24523
I1
aI3
aI5
aI5
aI7
aI25
aI51
aI3
aI85
aI371
aI1503
aI3217
aI1779
aI7141
aI29471
aI42247
aI107699
asS'd'
I12262
sa(dp24524
S'a'
I55653
sS's'
I17
sS'm'
(lp24525
I1
aI3
aI7
aI15
aI23
aI53
aI127
aI229
aI241
aI165
aI1985
aI1921
aI5917
aI15743
aI18349
aI23981
aI58241
asS'd'
I12263
sa(dp24526
S'a'
I55654
sS's'
I17
sS'm'
(lp24527
I1
aI1
aI3
aI13
aI9
aI63
aI49
aI67
aI21
aI57
aI377
aI1807
aI5603
aI13651
aI28039
aI3745
aI4903
asS'd'
I12264
sa(dp24528
S'a'
I55660
sS's'
I17
sS'm'
(lp24529
I1
aI3
aI7
aI11
aI1
aI43
aI17
aI95
aI79
aI343
aI1939
aI2349
aI5195
aI3047
aI4325
aI27829
aI53809
asS'd'
I12265
sa(dp24530
S'a'
I55671
sS's'
I17
sS'm'
(lp24531
I1
aI1
aI5
aI9
aI7
aI43
aI111
aI221
aI493
aI151
aI1635
aI3949
aI6661
aI4861
aI17661
aI61909
aI4975
asS'd'
I12266
sa(dp24532
S'a'
I55693
sS's'
I17
sS'm'
(lp24533
I1
aI3
aI1
aI3
aI31
aI47
aI63
aI45
aI401
aI153
aI1139
aI2125
aI6639
aI14093
aI31607
aI20645
aI52245
asS'd'
I12267
sa(dp24534
S'a'
I55706
sS's'
I17
sS'm'
(lp24535
I1
aI1
aI3
aI1
aI19
aI31
aI59
aI139
aI285
aI749
aI751
aI775
aI7795
aI14917
aI30295
aI61037
aI12315
asS'd'
I12268
sa(dp24536
S'a'
I55712
sS's'
I17
sS'm'
(lp24537
I1
aI3
aI3
aI9
aI15
aI55
aI79
aI183
aI373
aI663
aI497
aI2589
aI4955
aI5409
aI23527
aI2683
aI5487
asS'd'
I12269
sa(dp24538
S'a'
I55718
sS's'
I17
sS'm'
(lp24539
I1
aI3
aI7
aI11
aI25
aI47
aI53
aI225
aI197
aI109
aI1937
aI1375
aI7347
aI7353
aI2335
aI21775
aI14877
asS'd'
I12270
sa(dp24540
S'a'
I55739
sS's'
I17
sS'm'
(lp24541
I1
aI3
aI1
aI9
aI23
aI61
aI51
aI221
aI129
aI57
aI115
aI1031
aI6793
aI14773
aI3331
aI24951
aI94761
asS'd'
I12271
sa(dp24542
S'a'
I55747
sS's'
I17
sS'm'
(lp24543
I1
aI1
aI5
aI15
aI31
aI9
aI69
aI117
aI295
aI147
aI673
aI3627
aI7167
aI13835
aI20593
aI53163
aI83033
asS'd'
I12272
sa(dp24544
S'a'
I55754
sS's'
I17
sS'm'
(lp24545
I1
aI3
aI7
aI9
aI15
aI63
aI111
aI225
aI147
aI863
aI691
aI629
aI7485
aI483
aI21835
aI46251
aI94645
asS'd'
I12273
sa(dp24546
S'a'
I55771
sS's'
I17
sS'm'
(lp24547
I1
aI3
aI3
aI3
aI11
aI41
aI23
aI159
aI133
aI787
aI1617
aI629
aI5047
aI4465
aI29051
aI47499
aI7211
asS'd'
I12274
sa(dp24548
S'a'
I55774
sS's'
I17
sS'm'
(lp24549
I1
aI1
aI3
aI13
aI21
aI61
aI29
aI159
aI73
aI165
aI917
aI2577
aI7237
aI11807
aI3767
aI56861
aI51395
asS'd'
I12275
sa(dp24550
S'a'
I55780
sS's'
I17
sS'm'
(lp24551
I1
aI1
aI5
aI3
aI15
aI31
aI37
aI233
aI283
aI265
aI1645
aI3843
aI1971
aI4989
aI26823
aI15243
aI74931
asS'd'
I12276
sa(dp24552
S'a'
I55783
sS's'
I17
sS'm'
(lp24553
I1
aI3
aI7
aI11
aI9
aI51
aI7
aI119
aI237
aI905
aI1211
aI3041
aI7641
aI3387
aI8373
aI38961
aI68925
asS'd'
I12277
sa(dp24554
S'a'
I55784
sS's'
I17
sS'm'
(lp24555
I1
aI1
aI7
aI3
aI17
aI53
aI55
aI195
aI57
aI957
aI2027
aI3965
aI2993
aI411
aI13947
aI58349
aI32169
asS'd'
I12278
sa(dp24556
S'a'
I55789
sS's'
I17
sS'm'
(lp24557
I1
aI1
aI7
aI9
aI7
aI37
aI55
aI93
aI173
aI769
aI1381
aI3977
aI5293
aI5051
aI21455
aI45547
aI64653
asS'd'
I12279
sa(dp24558
S'a'
I55798
sS's'
I17
sS'm'
(lp24559
I1
aI3
aI5
aI7
aI1
aI41
aI89
aI161
aI315
aI361
aI1675
aI2993
aI3281
aI13043
aI19003
aI22129
aI130379
asS'd'
I12280
sa(dp24560
S'a'
I55802
sS's'
I17
sS'm'
(lp24561
I1
aI3
aI1
aI9
aI13
aI37
aI85
aI197
aI465
aI177
aI661
aI943
aI541
aI11117
aI9751
aI4193
aI98291
asS'd'
I12281
sa(dp24562
S'a'
I55823
sS's'
I17
sS'm'
(lp24563
I1
aI1
aI3
aI7
aI21
aI7
aI67
aI17
aI41
aI817
aI1159
aI1483
aI6937
aI10079
aI3639
aI27887
aI14541
asS'd'
I12282
sa(dp24564
S'a'
I55828
sS's'
I17
sS'm'
(lp24565
I1
aI3
aI3
aI15
aI17
aI63
aI69
aI215
aI437
aI883
aI1857
aI3319
aI3107
aI16279
aI10709
aI30433
aI52551
asS'd'
I12283
sa(dp24566
S'a'
I55837
sS's'
I17
sS'm'
(lp24567
I1
aI1
aI3
aI11
aI1
aI5
aI69
aI37
aI419
aI999
aI1711
aI875
aI3807
aI10811
aI16345
aI61155
aI116043
asS'd'
I12284
sa(dp24568
S'a'
I55838
sS's'
I17
sS'm'
(lp24569
I1
aI3
aI7
aI13
aI17
aI7
aI57
aI237
aI81
aI691
aI1143
aI4075
aI2481
aI643
aI8091
aI8243
aI80111
asS'd'
I12285
sa(dp24570
S'a'
I55847
sS's'
I17
sS'm'
(lp24571
I1
aI1
aI7
aI7
aI17
aI61
aI73
aI215
aI113
aI885
aI159
aI2243
aI1177
aI10981
aI10123
aI48995
aI123349
asS'd'
I12286
sa(dp24572
S'a'
I55848
sS's'
I17
sS'm'
(lp24573
I1
aI1
aI1
aI7
aI31
aI47
aI99
aI15
aI371
aI343
aI1483
aI1985
aI25
aI11125
aI8357
aI10677
aI130895
asS'd'
I12287
sa(dp24574
S'a'
I55885
sS's'
I17
sS'm'
(lp24575
I1
aI3
aI1
aI13
aI25
aI41
aI83
aI37
aI129
aI493
aI641
aI185
aI6607
aI7213
aI13285
aI10439
aI73227
asS'd'
I12288
sa(dp24576
S'a'
I55897
sS's'
I17
sS'm'
(lp24577
I1
aI1
aI5
aI7
aI17
aI15
aI93
aI53
aI281
aI91
aI115
aI3675
aI3081
aI9825
aI23653
aI40095
aI91803
asS'd'
I12289
sa(dp24578
S'a'
I55904
sS's'
I17
sS'm'
(lp24579
I1
aI1
aI1
aI5
aI3
aI25
aI39
aI207
aI419
aI361
aI953
aI2823
aI8105
aI15763
aI29199
aI61607
aI32633
asS'd'
I12290
sa(dp24580
S'a'
I55910
sS's'
I17
sS'm'
(lp24581
I1
aI1
aI3
aI1
aI31
aI51
aI55
aI3
aI277
aI639
aI191
aI1783
aI139
aI29
aI16659
aI30199
aI69109
asS'd'
I12291
sa(dp24582
S'a'
I55919
sS's'
I17
sS'm'
(lp24583
I1
aI1
aI1
aI7
aI31
aI59
aI25
aI13
aI239
aI617
aI115
aI1787
aI5757
aI9927
aI2417
aI37313
aI115135
asS'd'
I12292
sa(dp24584
S'a'
I55922
sS's'
I17
sS'm'
(lp24585
I1
aI1
aI3
aI5
aI19
aI35
aI5
aI187
aI483
aI823
aI1875
aI163
aI4235
aI853
aI23679
aI50899
aI94981
asS'd'
I12293
sa(dp24586
S'a'
I55931
sS's'
I17
sS'm'
(lp24587
I1
aI3
aI3
aI11
aI9
aI39
aI121
aI201
aI189
aI543
aI1493
aI1215
aI351
aI16063
aI1701
aI56559
aI108053
asS'd'
I12294
sa(dp24588
S'a'
I55933
sS's'
I17
sS'm'
(lp24589
I1
aI3
aI1
aI7
aI1
aI39
aI31
aI163
aI347
aI307
aI349
aI4081
aI1729
aI16265
aI363
aI28297
aI50631
asS'd'
I12295
sa(dp24590
S'a'
I55937
sS's'
I17
sS'm'
(lp24591
I1
aI3
aI7
aI13
aI21
aI55
aI127
aI161
aI75
aI9
aI1285
aI1839
aI5283
aI5667
aI10979
aI22185
aI7581
asS'd'
I12296
sa(dp24592
S'a'
I55943
sS's'
I17
sS'm'
(lp24593
I1
aI1
aI3
aI5
aI13
aI45
aI17
aI181
aI117
aI395
aI1685
aI663
aI3441
aI5359
aI7157
aI27759
aI102343
asS'd'
I12297
sa(dp24594
S'a'
I55955
sS's'
I17
sS'm'
(lp24595
I1
aI3
aI1
aI7
aI7
aI31
aI97
aI187
aI383
aI769
aI1469
aI4007
aI5521
aI13973
aI49
aI43823
aI75649
asS'd'
I12298
sa(dp24596
S'a'
I55957
sS's'
I17
sS'm'
(lp24597
I1
aI3
aI5
aI7
aI31
aI13
aI47
aI11
aI335
aI961
aI321
aI3367
aI1903
aI503
aI8409
aI1101
aI58215
asS'd'
I12299
sa(dp24598
S'a'
I55968
sS's'
I17
sS'm'
(lp24599
I1
aI1
aI7
aI7
aI25
aI49
aI39
aI1
aI453
aI419
aI333
aI1759
aI2287
aI6243
aI10723
aI13687
aI56853
asS'd'
I12300
sa(dp24600
S'a'
I55980
sS's'
I17
sS'm'
(lp24601
I1
aI3
aI3
aI7
aI11
aI55
aI125
aI197
aI19
aI591
aI1969
aI511
aI2501
aI8429
aI29467
aI27917
aI63457
asS'd'
I12301
sa(dp24602
S'a'
I55986
sS's'
I17
sS'm'
(lp24603
I1
aI1
aI3
aI11
aI11
aI43
aI35
aI213
aI231
aI119
aI379
aI3761
aI4891
aI5677
aI20317
aI5459
aI55487
asS'd'
I12302
sa(dp24604
S'a'
I55992
sS's'
I17
sS'm'
(lp24605
I1
aI1
aI5
aI7
aI21
aI9
aI127
aI59
aI97
aI963
aI847
aI2131
aI7907
aI11409
aI8785
aI48197
aI96907
asS'd'
I12303
sa(dp24606
S'a'
I56005
sS's'
I17
sS'm'
(lp24607
I1
aI3
aI5
aI3
aI23
aI7
aI45
aI95
aI179
aI691
aI1571
aI3091
aI6359
aI9105
aI26021
aI26925
aI43
asS'd'
I12304
sa(dp24608
S'a'
I56034
sS's'
I17
sS'm'
(lp24609
I1
aI3
aI3
aI7
aI21
aI7
aI11
aI219
aI439
aI465
aI1983
aI117
aI4639
aI8387
aI27637
aI15883
aI5567
asS'd'
I12305
sa(dp24610
S'a'
I56051
sS's'
I17
sS'm'
(lp24611
I1
aI1
aI7
aI3
aI1
aI3
aI51
aI205
aI425
aI133
aI563
aI1317
aI533
aI1227
aI8361
aI23407
aI39825
asS'd'
I12306
sa(dp24612
S'a'
I56054
sS's'
I17
sS'm'
(lp24613
I1
aI1
aI5
aI3
aI3
aI39
aI19
aI69
aI477
aI605
aI3
aI1887
aI2077
aI13673
aI2763
aI64415
aI104519
asS'd'
I12307
sa(dp24614
S'a'
I56065
sS's'
I17
sS'm'
(lp24615
I1
aI1
aI5
aI15
aI11
aI45
aI89
aI245
aI177
aI591
aI1313
aI587
aI4781
aI5103
aI26401
aI12643
aI38959
asS'd'
I12308
sa(dp24616
S'a'
I56080
sS's'
I17
sS'm'
(lp24617
I1
aI1
aI1
aI5
aI11
aI13
aI15
aI95
aI271
aI99
aI2001
aI2701
aI6065
aI3527
aI7423
aI37525
aI117161
asS'd'
I12309
sa(dp24618
S'a'
I56101
sS's'
I17
sS'm'
(lp24619
I1
aI3
aI7
aI7
aI21
aI17
aI111
aI149
aI373
aI591
aI1461
aI809
aI3877
aI8635
aI13209
aI31439
aI64285
asS'd'
I12310
sa(dp24620
S'a'
I56113
sS's'
I17
sS'm'
(lp24621
I1
aI3
aI1
aI15
aI25
aI51
aI55
aI161
aI357
aI181
aI41
aI2345
aI3553
aI9917
aI30123
aI40683
aI122497
asS'd'
I12311
sa(dp24622
S'a'
I56126
sS's'
I17
sS'm'
(lp24623
I1
aI3
aI7
aI15
aI5
aI55
aI119
aI239
aI291
aI665
aI1537
aI3309
aI2519
aI12397
aI25897
aI51529
aI28673
asS'd'
I12312
sa(dp24624
S'a'
I56131
sS's'
I17
sS'm'
(lp24625
I1
aI1
aI1
aI9
aI25
aI45
aI21
aI119
aI19
aI145
aI313
aI2509
aI1031
aI3319
aI14863
aI10759
aI22577
asS'd'
I12313
sa(dp24626
S'a'
I56134
sS's'
I17
sS'm'
(lp24627
I1
aI3
aI1
aI15
aI1
aI61
aI87
aI229
aI511
aI83
aI79
aI51
aI1407
aI16293
aI26217
aI25839
aI86207
asS'd'
I12314
sa(dp24628
S'a'
I56143
sS's'
I17
sS'm'
(lp24629
I1
aI3
aI7
aI7
aI23
aI43
aI89
aI11
aI43
aI801
aI569
aI3273
aI315
aI9537
aI681
aI34783
aI97101
asS'd'
I12315
sa(dp24630
S'a'
I56148
sS's'
I17
sS'm'
(lp24631
I1
aI1
aI5
aI1
aI13
aI31
aI77
aI115
aI501
aI669
aI27
aI3765
aI6789
aI9139
aI30587
aI45995
aI102433
asS'd'
I12316
sa(dp24632
S'a'
I56152
sS's'
I17
sS'm'
(lp24633
I1
aI1
aI7
aI1
aI15
aI21
aI57
aI197
aI243
aI353
aI71
aI341
aI7319
aI8467
aI9779
aI15755
aI4185
asS'd'
I12317
sa(dp24634
S'a'
I56174
sS's'
I17
sS'm'
(lp24635
I1
aI1
aI1
aI1
aI17
aI21
aI3
aI185
aI277
aI585
aI265
aI3189
aI3975
aI353
aI8541
aI23905
aI21881
asS'd'
I12318
sa(dp24636
S'a'
I56181
sS's'
I17
sS'm'
(lp24637
I1
aI3
aI5
aI3
aI5
aI23
aI113
aI253
aI343
aI73
aI1419
aI2529
aI4333
aI2007
aI14307
aI60591
aI55411
asS'd'
I12319
sa(dp24638
S'a'
I56182
sS's'
I17
sS'm'
(lp24639
I1
aI1
aI7
aI3
aI25
aI35
aI109
aI173
aI351
aI487
aI1551
aI3207
aI1189
aI5091
aI3581
aI4699
aI22085
asS'd'
I12320
sa(dp24640
S'a'
I56185
sS's'
I17
sS'm'
(lp24641
I1
aI1
aI3
aI11
aI9
aI9
aI71
aI173
aI17
aI595
aI2015
aI2543
aI4889
aI6025
aI15265
aI6459
aI3977
asS'd'
I12321
sa(dp24642
S'a'
I56192
sS's'
I17
sS'm'
(lp24643
I1
aI3
aI5
aI15
aI11
aI13
aI11
aI189
aI431
aI307
aI317
aI3131
aI1421
aI10863
aI5311
aI25273
aI43187
asS'd'
I12322
sa(dp24644
S'a'
I56201
sS's'
I17
sS'm'
(lp24645
I1
aI1
aI1
aI7
aI7
aI41
aI103
aI231
aI321
aI327
aI1849
aI2485
aI6461
aI10259
aI4577
aI52951
aI33053
asS'd'
I12323
sa(dp24646
S'a'
I56215
sS's'
I17
sS'm'
(lp24647
I1
aI3
aI3
aI15
aI11
aI33
aI73
aI155
aI453
aI597
aI575
aI2119
aI327
aI4227
aI32271
aI7429
aI102007
asS'd'
I12324
sa(dp24648
S'a'
I56219
sS's'
I17
sS'm'
(lp24649
I1
aI1
aI3
aI7
aI15
aI9
aI95
aI177
aI21
aI245
aI257
aI3637
aI821
aI16351
aI1733
aI10635
aI59885
asS'd'
I12325
sa(dp24650
S'a'
I56222
sS's'
I17
sS'm'
(lp24651
I1
aI1
aI3
aI7
aI23
aI41
aI107
aI147
aI57
aI877
aI1609
aI3275
aI339
aI12997
aI5989
aI62293
aI21549
asS'd'
I12326
sa(dp24652
S'a'
I56258
sS's'
I17
sS'm'
(lp24653
I1
aI1
aI7
aI13
aI19
aI39
aI111
aI229
aI321
aI487
aI873
aI3365
aI4915
aI251
aI30629
aI45775
aI73549
asS'd'
I12327
sa(dp24654
S'a'
I56264
sS's'
I17
sS'm'
(lp24655
I1
aI1
aI1
aI13
aI25
aI61
aI43
aI111
aI135
aI463
aI1921
aI1723
aI7505
aI13805
aI30633
aI51683
aI7353
asS'd'
I12328
sa(dp24656
S'a'
I56269
sS's'
I17
sS'm'
(lp24657
I1
aI1
aI1
aI1
aI31
aI41
aI35
aI205
aI375
aI189
aI635
aI3589
aI3507
aI8131
aI13437
aI22823
aI68451
asS'd'
I12329
sa(dp24658
S'a'
I56272
sS's'
I17
sS'm'
(lp24659
I1
aI3
aI7
aI5
aI31
aI47
aI13
aI229
aI105
aI195
aI685
aI529
aI39
aI2651
aI6821
aI11043
aI112123
asS'd'
I12330
sa(dp24660
S'a'
I56278
sS's'
I17
sS'm'
(lp24661
I1
aI3
aI7
aI13
aI3
aI23
aI21
aI203
aI89
aI957
aI1577
aI1711
aI585
aI3937
aI17681
aI55577
aI61075
asS'd'
I12331
sa(dp24662
S'a'
I56282
sS's'
I17
sS'm'
(lp24663
I1
aI3
aI3
aI3
aI17
aI37
aI49
aI7
aI287
aI183
aI1185
aI2979
aI2103
aI1217
aI22105
aI11677
aI19603
asS'd'
I12332
sa(dp24664
S'a'
I56306
sS's'
I17
sS'm'
(lp24665
I1
aI1
aI1
aI9
aI29
aI35
aI93
aI179
aI403
aI563
aI441
aI3485
aI6909
aI12647
aI3885
aI60089
aI29275
asS'd'
I12333
sa(dp24666
S'a'
I56308
sS's'
I17
sS'm'
(lp24667
I1
aI3
aI3
aI9
aI15
aI37
aI49
aI103
aI509
aI77
aI495
aI921
aI2599
aI14735
aI30951
aI22779
aI47747
asS'd'
I12334
sa(dp24668
S'a'
I56318
sS's'
I17
sS'm'
(lp24669
I1
aI3
aI3
aI11
aI17
aI45
aI43
aI235
aI379
aI51
aI925
aI89
aI2241
aI10273
aI27649
aI8101
aI93977
asS'd'
I12335
sa(dp24670
S'a'
I56329
sS's'
I17
sS'm'
(lp24671
I1
aI3
aI3
aI3
aI11
aI7
aI25
aI163
aI405
aI997
aI847
aI2743
aI4705
aI7041
aI10997
aI50189
aI10775
asS'd'
I12336
sa(dp24672
S'a'
I56356
sS's'
I17
sS'm'
(lp24673
I1
aI1
aI1
aI13
aI19
aI43
aI3
aI125
aI37
aI41
aI5
aI965
aI2681
aI3737
aI29057
aI37777
aI119537
asS'd'
I12337
sa(dp24674
S'a'
I56360
sS's'
I17
sS'm'
(lp24675
I1
aI1
aI1
aI1
aI11
aI27
aI101
aI67
aI73
aI199
aI1739
aI2835
aI5837
aI10595
aI9865
aI38493
aI99323
asS'd'
I12338
sa(dp24676
S'a'
I56383
sS's'
I17
sS'm'
(lp24677
I1
aI3
aI1
aI13
aI27
aI43
aI21
aI79
aI419
aI847
aI843
aI2563
aI8133
aI10295
aI10127
aI30839
aI104863
asS'd'
I12339
sa(dp24678
S'a'
I56388
sS's'
I17
sS'm'
(lp24679
I1
aI3
aI3
aI11
aI15
aI37
aI71
aI251
aI157
aI971
aI165
aI1647
aI2583
aI205
aI23555
aI55297
aI106893
asS'd'
I12340
sa(dp24680
S'a'
I56391
sS's'
I17
sS'm'
(lp24681
I1
aI1
aI5
aI15
aI1
aI21
aI113
aI107
aI287
aI727
aI71
aI2655
aI1435
aI11125
aI15257
aI18899
aI37737
asS'd'
I12341
sa(dp24682
S'a'
I56405
sS's'
I17
sS'm'
(lp24683
I1
aI1
aI3
aI3
aI27
aI49
aI17
aI85
aI57
aI237
aI349
aI4049
aI1103
aI2523
aI3919
aI36587
aI128595
asS'd'
I12342
sa(dp24684
S'a'
I56406
sS's'
I17
sS'm'
(lp24685
I1
aI3
aI5
aI9
aI13
aI1
aI65
aI13
aI361
aI409
aI413
aI2153
aI5953
aI10651
aI25383
aI49777
aI65399
asS'd'
I12343
sa(dp24686
S'a'
I56416
sS's'
I17
sS'm'
(lp24687
I1
aI1
aI7
aI9
aI19
aI47
aI69
aI127
aI121
aI925
aI57
aI2775
aI4981
aI3643
aI4077
aI3081
aI56093
asS'd'
I12344
sa(dp24688
S'a'
I56426
sS's'
I17
sS'm'
(lp24689
I1
aI3
aI3
aI13
aI1
aI53
aI45
aI13
aI489
aI445
aI623
aI3547
aI1659
aI1899
aI11971
aI3725
aI12445
asS'd'
I12345
sa(dp24690
S'a'
I56439
sS's'
I17
sS'm'
(lp24691
I1
aI3
aI7
aI11
aI13
aI9
aI59
aI157
aI125
aI975
aI1283
aI297
aI3609
aI3179
aI31341
aI54727
aI112515
asS'd'
I12346
sa(dp24692
S'a'
I56446
sS's'
I17
sS'm'
(lp24693
I1
aI1
aI1
aI13
aI23
aI63
aI69
aI249
aI159
aI593
aI47
aI3957
aI757
aI14693
aI26345
aI18839
aI111263
asS'd'
I12347
sa(dp24694
S'a'
I56450
sS's'
I17
sS'm'
(lp24695
I1
aI1
aI1
aI1
aI31
aI37
aI37
aI199
aI7
aI425
aI337
aI1475
aI271
aI16215
aI12089
aI16765
aI13519
asS'd'
I12348
sa(dp24696
S'a'
I56459
sS's'
I17
sS'm'
(lp24697
I1
aI1
aI7
aI11
aI31
aI59
aI121
aI139
aI413
aI807
aI737
aI1235
aI3505
aI5859
aI14205
aI31939
aI4713
asS'd'
I12349
sa(dp24698
S'a'
I56461
sS's'
I17
sS'm'
(lp24699
I1
aI1
aI1
aI9
aI21
aI51
aI113
aI159
aI345
aI807
aI635
aI523
aI5535
aI13307
aI4239
aI14847
aI23711
asS'd'
I12350
sa(dp24700
S'a'
I56464
sS's'
I17
sS'm'
(lp24701
I1
aI3
aI5
aI13
aI31
aI7
aI33
aI1
aI293
aI271
aI1829
aI2535
aI6333
aI12037
aI29401
aI35009
aI37789
asS'd'
I12351
sa(dp24702
S'a'
I56469
sS's'
I17
sS'm'
(lp24703
I1
aI1
aI7
aI3
aI7
aI57
aI31
aI45
aI177
aI475
aI843
aI1265
aI585
aI16099
aI29293
aI52407
aI56131
asS'd'
I12352
sa(dp24704
S'a'
I56489
sS's'
I17
sS'm'
(lp24705
I1
aI3
aI3
aI11
aI1
aI25
aI117
aI205
aI139
aI141
aI1229
aI903
aI1883
aI11269
aI30493
aI3979
aI4263
asS'd'
I12353
sa(dp24706
S'a'
I56490
sS's'
I17
sS'm'
(lp24707
I1
aI1
aI5
aI13
aI25
aI43
aI9
aI237
aI347
aI869
aI1765
aI1389
aI1931
aI13331
aI17325
aI45999
aI121201
asS'd'
I12354
sa(dp24708
S'a'
I56497
sS's'
I17
sS'm'
(lp24709
I1
aI3
aI3
aI11
aI3
aI13
aI95
aI49
aI389
aI53
aI491
aI1467
aI5105
aI16053
aI6305
aI15759
aI51991
asS'd'
I12355
sa(dp24710
S'a'
I56509
sS's'
I17
sS'm'
(lp24711
I1
aI3
aI5
aI15
aI1
aI53
aI85
aI69
aI75
aI409
aI1299
aI1245
aI7951
aI10709
aI9157
aI3509
aI103975
asS'd'
I12356
sa(dp24712
S'a'
I56515
sS's'
I17
sS'm'
(lp24713
I1
aI1
aI1
aI13
aI1
aI33
aI97
aI235
aI463
aI413
aI1759
aI1891
aI1781
aI5261
aI5759
aI201
aI69199
asS'd'
I12357
sa(dp24714
S'a'
I56518
sS's'
I17
sS'm'
(lp24715
I1
aI1
aI3
aI13
aI21
aI27
aI101
aI143
aI123
aI705
aI969
aI2461
aI6057
aI13091
aI6077
aI38311
aI30379
asS'd'
I12358
sa(dp24716
S'a'
I56524
sS's'
I17
sS'm'
(lp24717
I1
aI3
aI3
aI1
aI13
aI11
aI73
aI33
aI495
aI513
aI763
aI3089
aI421
aI13663
aI30169
aI56599
aI38847
asS'd'
I12359
sa(dp24718
S'a'
I56535
sS's'
I17
sS'm'
(lp24719
I1
aI1
aI7
aI13
aI13
aI31
aI91
aI63
aI233
aI137
aI859
aI2449
aI539
aI12461
aI13477
aI31605
aI58919
asS'd'
I12360
sa(dp24720
S'a'
I56542
sS's'
I17
sS'm'
(lp24721
I1
aI1
aI1
aI9
aI13
aI49
aI107
aI45
aI451
aI707
aI1735
aI1881
aI3451
aI9131
aI25481
aI10841
aI116067
asS'd'
I12361
sa(dp24722
S'a'
I56545
sS's'
I17
sS'm'
(lp24723
I1
aI3
aI7
aI1
aI27
aI21
aI51
aI117
aI63
aI53
aI575
aI3325
aI1099
aI11181
aI23609
aI47141
aI115421
asS'd'
I12362
sa(dp24724
S'a'
I56551
sS's'
I17
sS'm'
(lp24725
I1
aI3
aI1
aI13
aI25
aI29
aI53
aI135
aI165
aI319
aI1695
aI341
aI8157
aI10671
aI7095
aI60749
aI31513
asS'd'
I12363
sa(dp24726
S'a'
I56555
sS's'
I17
sS'm'
(lp24727
I1
aI1
aI7
aI5
aI3
aI53
aI123
aI137
aI449
aI87
aI951
aI693
aI6943
aI15331
aI1515
aI24019
aI56613
asS'd'
I12364
sa(dp24728
S'a'
I56560
sS's'
I17
sS'm'
(lp24729
I1
aI3
aI7
aI5
aI25
aI17
aI43
aI251
aI301
aI203
aI633
aI1271
aI6253
aI4475
aI10773
aI25003
aI67599
asS'd'
I12365
sa(dp24730
S'a'
I56569
sS's'
I17
sS'm'
(lp24731
I1
aI1
aI3
aI9
aI1
aI25
aI117
aI159
aI13
aI155
aI851
aI2497
aI6155
aI6549
aI27909
aI24423
aI82357
asS'd'
I12366
sa(dp24732
S'a'
I56580
sS's'
I17
sS'm'
(lp24733
I1
aI1
aI1
aI1
aI21
aI59
aI103
aI43
aI291
aI111
aI1355
aI401
aI5129
aI16017
aI25947
aI15391
aI46745
asS'd'
I12367
sa(dp24734
S'a'
I56584
sS's'
I17
sS'm'
(lp24735
I1
aI1
aI1
aI13
aI13
aI51
aI95
aI111
aI17
aI963
aI1535
aI3003
aI6163
aI11377
aI6787
aI57275
aI109559
asS'd'
I12368
sa(dp24736
S'a'
I56598
sS's'
I17
sS'm'
(lp24737
I1
aI1
aI5
aI11
aI11
aI3
aI85
aI207
aI489
aI117
aI269
aI747
aI5719
aI8501
aI7307
aI59223
aI18941
asS'd'
I12369
sa(dp24738
S'a'
I56601
sS's'
I17
sS'm'
(lp24739
I1
aI1
aI5
aI1
aI9
aI53
aI41
aI255
aI271
aI995
aI1939
aI2739
aI2221
aI14841
aI22617
aI10643
aI6427
asS'd'
I12370
sa(dp24740
S'a'
I56602
sS's'
I17
sS'm'
(lp24741
I1
aI3
aI3
aI9
aI7
aI55
aI109
aI143
aI427
aI45
aI579
aI115
aI2061
aI8447
aI29469
aI5523
aI129063
asS'd'
I12371
sa(dp24742
S'a'
I56623
sS's'
I17
sS'm'
(lp24743
I1
aI3
aI7
aI5
aI23
aI63
aI119
aI31
aI53
aI821
aI135
aI2677
aI807
aI4685
aI24391
aI55165
aI88079
asS'd'
I12372
sa(dp24744
S'a'
I56637
sS's'
I17
sS'm'
(lp24745
I1
aI3
aI5
aI9
aI15
aI11
aI73
aI177
aI243
aI375
aI115
aI1633
aI7983
aI15039
aI21169
aI25325
aI128479
asS'd'
I12373
sa(dp24746
S'a'
I56638
sS's'
I17
sS'm'
(lp24747
I1
aI3
aI3
aI5
aI21
aI51
aI13
aI51
aI75
aI993
aI77
aI209
aI2761
aI451
aI11987
aI40297
aI2383
asS'd'
I12374
sa(dp24748
S'a'
I56643
sS's'
I17
sS'm'
(lp24749
I1
aI3
aI7
aI1
aI19
aI9
aI11
aI161
aI19
aI851
aI1313
aI1169
aI4405
aI7493
aI23935
aI37323
aI107387
asS'd'
I12375
sa(dp24750
S'a'
I56650
sS's'
I17
sS'm'
(lp24751
I1
aI3
aI5
aI15
aI5
aI11
aI79
aI129
aI507
aI247
aI811
aI1145
aI3893
aI5205
aI11309
aI38205
aI2051
asS'd'
I12376
sa(dp24752
S'a'
I56673
sS's'
I17
sS'm'
(lp24753
I1
aI1
aI5
aI1
aI11
aI13
aI33
aI155
aI21
aI185
aI771
aI3261
aI981
aI743
aI12479
aI22611
aI25321
asS'd'
I12377
sa(dp24754
S'a'
I56676
sS's'
I17
sS'm'
(lp24755
I1
aI1
aI5
aI9
aI7
aI25
aI11
aI235
aI429
aI563
aI1647
aI1429
aI1385
aI14411
aI3831
aI19769
aI67599
asS'd'
I12378
sa(dp24756
S'a'
I56686
sS's'
I17
sS'm'
(lp24757
I1
aI3
aI3
aI5
aI5
aI7
aI109
aI117
aI251
aI823
aI669
aI2043
aI1843
aI11829
aI27051
aI35865
aI11461
asS'd'
I12379
sa(dp24758
S'a'
I56704
sS's'
I17
sS'm'
(lp24759
I1
aI3
aI5
aI5
aI7
aI3
aI45
aI63
aI305
aI99
aI393
aI1765
aI1711
aI15569
aI27295
aI16555
aI77631
asS'd'
I12380
sa(dp24760
S'a'
I56710
sS's'
I17
sS'm'
(lp24761
I1
aI3
aI5
aI11
aI23
aI33
aI49
aI125
aI85
aI677
aI1589
aI2667
aI5723
aI15619
aI30415
aI39561
aI122763
asS'd'
I12381
sa(dp24762
S'a'
I56721
sS's'
I17
sS'm'
(lp24763
I1
aI3
aI7
aI13
aI27
aI21
aI99
aI209
aI481
aI123
aI1285
aI115
aI6517
aI11753
aI11365
aI44959
aI89
asS'd'
I12382
sa(dp24764
S'a'
I56731
sS's'
I17
sS'm'
(lp24765
I1
aI1
aI1
aI7
aI15
aI15
aI45
aI151
aI489
aI169
aI933
aI2987
aI657
aI3095
aI6745
aI131
aI37767
asS'd'
I12383
sa(dp24766
S'a'
I56738
sS's'
I17
sS'm'
(lp24767
I1
aI3
aI1
aI15
aI1
aI37
aI99
aI137
aI151
aI891
aI715
aI383
aI1293
aI719
aI10957
aI5557
aI92841
asS'd'
I12384
sa(dp24768
S'a'
I56747
sS's'
I17
sS'm'
(lp24769
I1
aI1
aI1
aI3
aI27
aI59
aI93
aI49
aI473
aI313
aI431
aI1129
aI5995
aI13101
aI13185
aI7091
aI109677
asS'd'
I12385
sa(dp24770
S'a'
I56750
sS's'
I17
sS'm'
(lp24771
I1
aI3
aI3
aI7
aI1
aI41
aI55
aI135
aI271
aI527
aI1919
aI1093
aI2579
aI3725
aI22853
aI31613
aI4729
asS'd'
I12386
sa(dp24772
S'a'
I56752
sS's'
I17
sS'm'
(lp24773
I1
aI1
aI7
aI1
aI7
aI13
aI63
aI255
aI219
aI837
aI117
aI2323
aI4295
aI15697
aI8607
aI47047
aI117869
asS'd'
I12387
sa(dp24774
S'a'
I56757
sS's'
I17
sS'm'
(lp24775
I1
aI1
aI1
aI7
aI29
aI29
aI55
aI171
aI437
aI733
aI491
aI1037
aI7221
aI5705
aI31819
aI19583
aI103991
asS'd'
I12388
sa(dp24776
S'a'
I56787
sS's'
I17
sS'm'
(lp24777
I1
aI3
aI7
aI1
aI31
aI19
aI65
aI39
aI151
aI517
aI1985
aI2251
aI6147
aI12983
aI28263
aI35891
aI7545
asS'd'
I12389
sa(dp24778
S'a'
I56796
sS's'
I17
sS'm'
(lp24779
I1
aI1
aI7
aI7
aI21
aI41
aI97
aI253
aI427
aI391
aI849
aI611
aI4827
aI10807
aI6267
aI22513
aI62803
asS'd'
I12390
sa(dp24780
S'a'
I56803
sS's'
I17
sS'm'
(lp24781
I1
aI3
aI7
aI7
aI25
aI9
aI49
aI245
aI491
aI39
aI603
aI1853
aI5655
aI3517
aI10745
aI55069
aI121497
asS'd'
I12391
sa(dp24782
S'a'
I56812
sS's'
I17
sS'm'
(lp24783
I1
aI3
aI1
aI5
aI9
aI39
aI109
aI195
aI283
aI141
aI2007
aI3
aI1267
aI13053
aI8387
aI48665
aI48877
asS'd'
I12392
sa(dp24784
S'a'
I56815
sS's'
I17
sS'm'
(lp24785
I1
aI1
aI3
aI7
aI27
aI61
aI49
aI43
aI229
aI497
aI2015
aI1345
aI3195
aI7139
aI13453
aI56993
aI15099
asS'd'
I12393
sa(dp24786
S'a'
I56824
sS's'
I17
sS'm'
(lp24787
I1
aI3
aI5
aI11
aI31
aI53
aI87
aI97
aI385
aI387
aI1107
aI3287
aI2517
aI7421
aI1007
aI37421
aI124113
asS'd'
I12394
sa(dp24788
S'a'
I56829
sS's'
I17
sS'm'
(lp24789
I1
aI3
aI3
aI11
aI9
aI55
aI51
aI215
aI181
aI419
aI863
aI3149
aI5815
aI15579
aI28527
aI34715
aI61375
asS'd'
I12395
sa(dp24790
S'a'
I56833
sS's'
I17
sS'm'
(lp24791
I1
aI3
aI3
aI3
aI31
aI57
aI5
aI35
aI445
aI957
aI1897
aI105
aI2533
aI10255
aI19795
aI49127
aI38491
asS'd'
I12396
sa(dp24792
S'a'
I56839
sS's'
I17
sS'm'
(lp24793
I1
aI3
aI3
aI13
aI25
aI53
aI1
aI159
aI443
aI541
aI439
aI3377
aI5511
aI9667
aI26777
aI32599
aI36981
asS'd'
I12397
sa(dp24794
S'a'
I56840
sS's'
I17
sS'm'
(lp24795
I1
aI1
aI1
aI3
aI23
aI29
aI97
aI131
aI59
aI143
aI1601
aI2765
aI4569
aI11081
aI6027
aI38641
aI100745
asS'd'
I12398
sa(dp24796
S'a'
I56845
sS's'
I17
sS'm'
(lp24797
I1
aI3
aI3
aI15
aI9
aI15
aI19
aI35
aI321
aI935
aI465
aI2707
aI4799
aI7455
aI12743
aI31029
aI114149
asS'd'
I12399
sa(dp24798
S'a'
I56853
sS's'
I17
sS'm'
(lp24799
I1
aI1
aI1
aI1
aI11
aI51
aI23
aI79
aI387
aI701
aI107
aI623
aI231
aI12571
aI7719
aI3061
aI79605
asS'd'
I12400
sa(dp24800
S'a'
I56854
sS's'
I17
sS'm'
(lp24801
I1
aI3
aI5
aI15
aI17
aI49
aI109
aI83
aI185
aI295
aI853
aI219
aI3615
aI535
aI32001
aI6655
aI4185
asS'd'
I12401
sa(dp24802
S'a'
I56858
sS's'
I17
sS'm'
(lp24803
I1
aI3
aI3
aI15
aI15
aI7
aI35
aI151
aI305
aI705
aI1383
aI1595
aI5595
aI11995
aI15491
aI49119
aI83383
asS'd'
I12402
sa(dp24804
S'a'
I56863
sS's'
I17
sS'm'
(lp24805
I1
aI3
aI7
aI9
aI3
aI25
aI57
aI47
aI359
aI719
aI1937
aI1403
aI1399
aI10827
aI24181
aI29141
aI79017
asS'd'
I12403
sa(dp24806
S'a'
I56864
sS's'
I17
sS'm'
(lp24807
I1
aI3
aI5
aI1
aI1
aI21
aI21
aI245
aI361
aI485
aI1521
aI3935
aI1587
aI8653
aI25871
aI49449
aI103413
asS'd'
I12404
sa(dp24808
S'a'
I56881
sS's'
I17
sS'm'
(lp24809
I1
aI3
aI3
aI3
aI9
aI61
aI31
aI69
aI401
aI261
aI1217
aI3069
aI4045
aI12437
aI32017
aI15113
aI10769
asS'd'
I12405
sa(dp24810
S'a'
I56882
sS's'
I17
sS'm'
(lp24811
I1
aI3
aI1
aI1
aI21
aI33
aI123
aI87
aI481
aI793
aI625
aI4087
aI1361
aI11077
aI18835
aI13287
aI40107
asS'd'
I12406
sa(dp24812
S'a'
I56896
sS's'
I17
sS'm'
(lp24813
I1
aI3
aI7
aI3
aI9
aI49
aI101
aI213
aI467
aI77
aI1691
aI2621
aI4411
aI8025
aI30247
aI13691
aI20559
asS'd'
I12407
sa(dp24814
S'a'
I56901
sS's'
I17
sS'm'
(lp24815
I1
aI1
aI1
aI9
aI29
aI49
aI47
aI135
aI1
aI337
aI1649
aI389
aI3845
aI7213
aI19527
aI2619
aI78841
asS'd'
I12408
sa(dp24816
S'a'
I56902
sS's'
I17
sS'm'
(lp24817
I1
aI3
aI7
aI9
aI23
aI23
aI47
aI97
aI493
aI767
aI137
aI1467
aI7015
aI2883
aI12749
aI9267
aI12441
asS'd'
I12409
sa(dp24818
S'a'
I56908
sS's'
I17
sS'm'
(lp24819
I1
aI1
aI3
aI15
aI9
aI57
aI53
aI19
aI401
aI385
aI1159
aI1185
aI6977
aI14027
aI3183
aI59119
aI42065
asS'd'
I12410
sa(dp24820
S'a'
I56923
sS's'
I17
sS'm'
(lp24821
I1
aI1
aI1
aI15
aI21
aI33
aI83
aI251
aI147
aI395
aI321
aI443
aI6893
aI1877
aI6687
aI28863
aI86531
asS'd'
I12411
sa(dp24822
S'a'
I56925
sS's'
I17
sS'm'
(lp24823
I1
aI3
aI3
aI1
aI3
aI47
aI27
aI247
aI121
aI827
aI1399
aI4079
aI7545
aI11691
aI27915
aI28811
aI17099
asS'd'
I12412
sa(dp24824
S'a'
I56926
sS's'
I17
sS'm'
(lp24825
I1
aI3
aI5
aI7
aI5
aI59
aI73
aI69
aI117
aI897
aI905
aI3273
aI2935
aI11077
aI32443
aI60959
aI16081
asS'd'
I12413
sa(dp24826
S'a'
I56932
sS's'
I17
sS'm'
(lp24827
I1
aI1
aI7
aI5
aI27
aI49
aI107
aI169
aI75
aI435
aI1913
aI2089
aI5733
aI2361
aI5163
aI52239
aI87411
asS'd'
I12414
sa(dp24828
S'a'
I56939
sS's'
I17
sS'm'
(lp24829
I1
aI1
aI7
aI13
aI1
aI13
aI123
aI89
aI427
aI301
aI1217
aI1491
aI5361
aI10381
aI28971
aI57655
aI108607
asS'd'
I12415
sa(dp24830
S'a'
I56947
sS's'
I17
sS'm'
(lp24831
I1
aI1
aI7
aI11
aI13
aI5
aI23
aI151
aI117
aI369
aI623
aI2263
aI2609
aI109
aI32485
aI52133
aI69391
asS'd'
I12416
sa(dp24832
S'a'
I56954
sS's'
I17
sS'm'
(lp24833
I1
aI3
aI7
aI11
aI7
aI33
aI127
aI43
aI123
aI203
aI775
aI3215
aI5115
aI1805
aI14581
aI46791
aI128781
asS'd'
I12417
sa(dp24834
S'a'
I56970
sS's'
I17
sS'm'
(lp24835
I1
aI3
aI3
aI1
aI7
aI23
aI37
aI99
aI1
aI719
aI293
aI2727
aI6859
aI683
aI13241
aI17839
aI4215
asS'd'
I12418
sa(dp24836
S'a'
I56972
sS's'
I17
sS'm'
(lp24837
I1
aI1
aI1
aI13
aI27
aI41
aI93
aI25
aI59
aI947
aI971
aI1523
aI4443
aI1209
aI32317
aI58651
aI11121
asS'd'
I12419
sa(dp24838
S'a'
I56977
sS's'
I17
sS'm'
(lp24839
I1
aI1
aI1
aI3
aI31
aI33
aI23
aI87
aI349
aI265
aI445
aI3489
aI783
aI7833
aI5767
aI59295
aI45057
asS'd'
I12420
sa(dp24840
S'a'
I56980
sS's'
I17
sS'm'
(lp24841
I1
aI1
aI1
aI1
aI15
aI47
aI19
aI15
aI217
aI837
aI2043
aI2805
aI4701
aI5873
aI1517
aI46743
aI61655
asS'd'
I12421
sa(dp24842
S'a'
I56993
sS's'
I17
sS'm'
(lp24843
I1
aI3
aI7
aI1
aI27
aI27
aI9
aI107
aI25
aI897
aI955
aI3763
aI821
aI1535
aI14557
aI38537
aI128737
asS'd'
I12422
sa(dp24844
S'a'
I57013
sS's'
I17
sS'm'
(lp24845
I1
aI3
aI3
aI7
aI19
aI49
aI121
aI217
aI401
aI975
aI1189
aI715
aI3113
aI4219
aI4885
aI57861
aI6833
asS'd'
I12423
sa(dp24846
S'a'
I57017
sS's'
I17
sS'm'
(lp24847
I1
aI3
aI3
aI1
aI17
aI59
aI53
aI15
aI259
aI791
aI2035
aI499
aI7707
aI13685
aI14367
aI20155
aI91033
asS'd'
I12424
sa(dp24848
S'a'
I57018
sS's'
I17
sS'm'
(lp24849
I1
aI3
aI5
aI5
aI11
aI11
aI69
aI237
aI139
aI73
aI541
aI1135
aI2647
aI14109
aI18113
aI8051
aI31917
asS'd'
I12425
sa(dp24850
S'a'
I57031
sS's'
I17
sS'm'
(lp24851
I1
aI1
aI1
aI15
aI11
aI23
aI89
aI181
aI295
aI743
aI29
aI4009
aI4683
aI13989
aI4575
aI38865
aI30449
asS'd'
I12426
sa(dp24852
S'a'
I57059
sS's'
I17
sS'm'
(lp24853
I1
aI1
aI1
aI11
aI21
aI63
aI31
aI121
aI55
aI153
aI1143
aI4059
aI3247
aI11725
aI17659
aI48935
aI118369
asS'd'
I12427
sa(dp24854
S'a'
I57062
sS's'
I17
sS'm'
(lp24855
I1
aI1
aI7
aI9
aI17
aI29
aI27
aI167
aI69
aI957
aI2009
aI2795
aI3161
aI3493
aI16365
aI43637
aI102321
asS'd'
I12428
sa(dp24856
S'a'
I57065
sS's'
I17
sS'm'
(lp24857
I1
aI3
aI5
aI13
aI27
aI23
aI17
aI7
aI345
aI253
aI631
aI1389
aI2523
aI9993
aI32619
aI46731
aI4757
asS'd'
I12429
sa(dp24858
S'a'
I57071
sS's'
I17
sS'm'
(lp24859
I1
aI1
aI1
aI11
aI17
aI1
aI41
aI107
aI25
aI183
aI1361
aI1211
aI3607
aI12713
aI16011
aI42987
aI36415
asS'd'
I12430
sa(dp24860
S'a'
I57083
sS's'
I17
sS'm'
(lp24861
I1
aI1
aI7
aI5
aI13
aI29
aI33
aI69
aI261
aI213
aI73
aI3737
aI3867
aI503
aI28225
aI53735
aI91695
asS'd'
I12431
sa(dp24862
S'a'
I57093
sS's'
I17
sS'm'
(lp24863
I1
aI3
aI7
aI7
aI21
aI63
aI75
aI39
aI259
aI367
aI487
aI2087
aI5411
aI1925
aI29589
aI39019
aI73283
asS'd'
I12432
sa(dp24864
S'a'
I57097
sS's'
I17
sS'm'
(lp24865
I1
aI3
aI1
aI7
aI29
aI29
aI25
aI191
aI91
aI509
aI1485
aI853
aI7011
aI13321
aI27769
aI10249
aI87341
asS'd'
I12433
sa(dp24866
S'a'
I57106
sS's'
I17
sS'm'
(lp24867
I1
aI1
aI7
aI7
aI1
aI9
aI115
aI71
aI321
aI913
aI1679
aI2129
aI771
aI9217
aI4731
aI24353
aI35631
asS'd'
I12434
sa(dp24868
S'a'
I57118
sS's'
I17
sS'm'
(lp24869
I1
aI3
aI1
aI13
aI5
aI45
aI53
aI255
aI429
aI805
aI1983
aI1437
aI2677
aI6337
aI22221
aI55455
aI39855
asS'd'
I12435
sa(dp24870
S'a'
I57133
sS's'
I17
sS'm'
(lp24871
I1
aI1
aI1
aI1
aI1
aI5
aI111
aI231
aI321
aI961
aI371
aI3825
aI3623
aI3985
aI32151
aI6113
aI130687
asS'd'
I12436
sa(dp24872
S'a'
I57134
sS's'
I17
sS'm'
(lp24873
I1
aI3
aI7
aI3
aI15
aI29
aI103
aI181
aI261
aI149
aI1161
aI1745
aI1765
aI1677
aI20051
aI47033
aI84997
asS'd'
I12437
sa(dp24874
S'a'
I57141
sS's'
I17
sS'm'
(lp24875
I1
aI3
aI7
aI1
aI25
aI21
aI7
aI229
aI407
aI673
aI1525
aI1207
aI3099
aI14849
aI22103
aI45695
aI85951
asS'd'
I12438
sa(dp24876
S'a'
I57146
sS's'
I17
sS'm'
(lp24877
I1
aI3
aI7
aI1
aI9
aI5
aI105
aI149
aI181
aI81
aI1589
aI3477
aI5387
aI7943
aI29203
aI50355
aI39001
asS'd'
I12439
sa(dp24878
S'a'
I57156
sS's'
I17
sS'm'
(lp24879
I1
aI3
aI7
aI1
aI21
aI31
aI39
aI121
aI397
aI1023
aI711
aI623
aI6193
aI12315
aI11101
aI11911
aI50033
asS'd'
I12440
sa(dp24880
S'a'
I57174
sS's'
I17
sS'm'
(lp24881
I1
aI3
aI7
aI3
aI7
aI19
aI73
aI69
aI201
aI337
aI1037
aI3663
aI2679
aI5153
aI28171
aI24455
aI74685
asS'd'
I12441
sa(dp24882
S'a'
I57183
sS's'
I17
sS'm'
(lp24883
I1
aI3
aI1
aI1
aI21
aI35
aI121
aI111
aI217
aI809
aI507
aI1347
aI4439
aI4601
aI26371
aI23595
aI3583
asS'd'
I12442
sa(dp24884
S'a'
I57189
sS's'
I17
sS'm'
(lp24885
I1
aI3
aI3
aI7
aI1
aI51
aI83
aI231
aI419
aI597
aI305
aI3405
aI5831
aI11845
aI1861
aI48671
aI105315
asS'd'
I12443
sa(dp24886
S'a'
I57190
sS's'
I17
sS'm'
(lp24887
I1
aI3
aI1
aI15
aI25
aI37
aI103
aI141
aI495
aI727
aI1919
aI2821
aI4689
aI6727
aI27117
aI2259
aI54559
asS'd'
I12444
sa(dp24888
S'a'
I57193
sS's'
I17
sS'm'
(lp24889
I1
aI3
aI5
aI3
aI27
aI55
aI5
aI117
aI199
aI979
aI1745
aI401
aI7967
aI5345
aI29747
aI54085
aI124765
asS'd'
I12445
sa(dp24890
S'a'
I57196
sS's'
I17
sS'm'
(lp24891
I1
aI3
aI3
aI15
aI23
aI61
aI1
aI29
aI489
aI131
aI583
aI389
aI6033
aI8007
aI22933
aI44513
aI111845
asS'd'
I12446
sa(dp24892
S'a'
I57207
sS's'
I17
sS'm'
(lp24893
I1
aI1
aI7
aI13
aI3
aI55
aI119
aI147
aI181
aI485
aI793
aI3593
aI6971
aI2227
aI28507
aI62393
aI127303
asS'd'
I12447
sa(dp24894
S'a'
I57211
sS's'
I17
sS'm'
(lp24895
I1
aI1
aI5
aI3
aI29
aI53
aI37
aI59
aI213
aI283
aI1809
aI3685
aI2677
aI5761
aI19705
aI47079
aI3477
asS'd'
I12448
sa(dp24896
S'a'
I57214
sS's'
I17
sS'm'
(lp24897
I1
aI1
aI1
aI7
aI21
aI63
aI91
aI13
aI347
aI605
aI589
aI415
aI5737
aI10281
aI30941
aI25609
aI67973
asS'd'
I12449
sa(dp24898
S'a'
I57218
sS's'
I17
sS'm'
(lp24899
I1
aI1
aI3
aI1
aI27
aI49
aI87
aI161
aI507
aI693
aI59
aI1375
aI6737
aI1029
aI14731
aI32335
aI51961
asS'd'
I12450
sa(dp24900
S'a'
I57230
sS's'
I17
sS'm'
(lp24901
I1
aI3
aI3
aI9
aI11
aI15
aI121
aI121
aI151
aI335
aI221
aI3099
aI1999
aI1047
aI20891
aI23015
aI95809
asS'd'
I12451
sa(dp24902
S'a'
I57235
sS's'
I17
sS'm'
(lp24903
I1
aI1
aI7
aI7
aI1
aI49
aI63
aI227
aI113
aI161
aI863
aI801
aI2559
aI7737
aI27619
aI27419
aI128009
asS'd'
I12452
sa(dp24904
S'a'
I57241
sS's'
I17
sS'm'
(lp24905
I1
aI1
aI7
aI1
aI25
aI59
aI67
aI111
aI435
aI309
aI807
aI2107
aI8077
aI9671
aI6739
aI53757
aI41259
asS'd'
I12453
sa(dp24906
S'a'
I57251
sS's'
I17
sS'm'
(lp24907
I1
aI3
aI7
aI3
aI3
aI19
aI7
aI111
aI237
aI981
aI1717
aI3131
aI6631
aI467
aI13103
aI61435
aI126469
asS'd'
I12454
sa(dp24908
S'a'
I57260
sS's'
I17
sS'm'
(lp24909
I1
aI1
aI7
aI9
aI19
aI31
aI59
aI185
aI199
aI111
aI351
aI611
aI6355
aI1095
aI28549
aI32871
aI44537
asS'd'
I12455
sa(dp24910
S'a'
I57263
sS's'
I17
sS'm'
(lp24911
I1
aI1
aI3
aI11
aI23
aI25
aI31
aI1
aI461
aI83
aI1723
aI1711
aI3679
aI10963
aI14927
aI17377
aI911
asS'd'
I12456
sa(dp24912
S'a'
I57265
sS's'
I17
sS'm'
(lp24913
I1
aI1
aI5
aI13
aI3
aI37
aI127
aI159
aI199
aI223
aI1097
aI3033
aI5825
aI13777
aI22189
aI44305
aI20509
asS'd'
I12457
sa(dp24914
S'a'
I57271
sS's'
I17
sS'm'
(lp24915
I1
aI1
aI5
aI15
aI13
aI49
aI17
aI79
aI289
aI601
aI1023
aI657
aI1687
aI14477
aI15929
aI4279
aI68007
asS'd'
I12458
sa(dp24916
S'a'
I57272
sS's'
I17
sS'm'
(lp24917
I1
aI1
aI5
aI7
aI29
aI3
aI45
aI225
aI65
aI711
aI1039
aI3585
aI4957
aI9041
aI22761
aI26649
aI95627
asS'd'
I12459
sa(dp24918
S'a'
I57290
sS's'
I17
sS'm'
(lp24919
I1
aI3
aI5
aI5
aI13
aI29
aI33
aI211
aI461
aI799
aI1437
aI1057
aI485
aI9535
aI8133
aI57527
aI12873
asS'd'
I12460
sa(dp24920
S'a'
I57295
sS's'
I17
sS'm'
(lp24921
I1
aI1
aI3
aI3
aI15
aI43
aI53
aI15
aI395
aI561
aI1371
aI3543
aI7707
aI2399
aI13311
aI25641
aI58865
asS'd'
I12461
sa(dp24922
S'a'
I57298
sS's'
I17
sS'm'
(lp24923
I1
aI3
aI1
aI7
aI1
aI9
aI115
aI39
aI249
aI87
aI835
aI97
aI8137
aI6665
aI11951
aI21045
aI76387
asS'd'
I12462
sa(dp24924
S'a'
I57325
sS's'
I17
sS'm'
(lp24925
I1
aI3
aI3
aI9
aI5
aI63
aI115
aI163
aI331
aI1007
aI733
aI4027
aI2911
aI5329
aI6967
aI3535
aI107293
asS'd'
I12463
sa(dp24926
S'a'
I57328
sS's'
I17
sS'm'
(lp24927
I1
aI3
aI1
aI11
aI5
aI55
aI81
aI63
aI345
aI535
aI1093
aI207
aI4053
aI9129
aI10397
aI26641
aI95171
asS'd'
I12464
sa(dp24928
S'a'
I57340
sS's'
I17
sS'm'
(lp24929
I1
aI1
aI3
aI3
aI13
aI27
aI65
aI37
aI255
aI69
aI19
aI2565
aI4329
aI11223
aI18131
aI18701
aI31111
asS'd'
I12465
sa(dp24930
S'a'
I57344
sS's'
I17
sS'm'
(lp24931
I1
aI3
aI7
aI7
aI25
aI57
aI81
aI189
aI227
aI377
aI829
aI1583
aI1343
aI4643
aI23485
aI47463
aI83535
asS'd'
I12466
sa(dp24932
S'a'
I57353
sS's'
I17
sS'm'
(lp24933
I1
aI3
aI5
aI9
aI13
aI33
aI103
aI75
aI501
aI803
aI427
aI1171
aI1187
aI2655
aI24187
aI32907
aI120239
asS'd'
I12467
sa(dp24934
S'a'
I57368
sS's'
I17
sS'm'
(lp24935
I1
aI3
aI1
aI7
aI31
aI63
aI21
aI137
aI241
aI63
aI1925
aI2193
aI1135
aI11159
aI14685
aI28397
aI59
asS'd'
I12468
sa(dp24936
S'a'
I57371
sS's'
I17
sS'm'
(lp24937
I1
aI1
aI7
aI1
aI15
aI7
aI85
aI87
aI493
aI63
aI561
aI1069
aI5481
aI12253
aI25149
aI35283
aI16123
asS'd'
I12469
sa(dp24938
S'a'
I57383
sS's'
I17
sS'm'
(lp24939
I1
aI3
aI3
aI5
aI19
aI17
aI23
aI95
aI429
aI805
aI1343
aI2243
aI233
aI7219
aI6549
aI21477
aI83679
asS'd'
I12470
sa(dp24940
S'a'
I57402
sS's'
I17
sS'm'
(lp24941
I1
aI3
aI3
aI1
aI3
aI55
aI83
aI107
aI131
aI311
aI741
aI781
aI6227
aI10059
aI3903
aI235
aI45971
asS'd'
I12471
sa(dp24942
S'a'
I57407
sS's'
I17
sS'm'
(lp24943
I1
aI1
aI3
aI13
aI21
aI43
aI79
aI149
aI367
aI755
aI463
aI221
aI5117
aI7015
aI17599
aI64665
aI37443
asS'd'
I12472
sa(dp24944
S'a'
I57419
sS's'
I17
sS'm'
(lp24945
I1
aI1
aI7
aI15
aI3
aI59
aI107
aI13
aI213
aI287
aI1505
aI2131
aI6965
aI12873
aI23973
aI8449
aI24829
asS'd'
I12473
sa(dp24946
S'a'
I57424
sS's'
I17
sS'm'
(lp24947
I1
aI3
aI5
aI11
aI25
aI45
aI39
aI17
aI175
aI749
aI1179
aI3349
aI6723
aI12543
aI3557
aI34521
aI103197
asS'd'
I12474
sa(dp24948
S'a'
I57433
sS's'
I17
sS'm'
(lp24949
I1
aI1
aI3
aI13
aI23
aI41
aI21
aI25
aI91
aI941
aI879
aI4015
aI137
aI12949
aI17245
aI41903
aI39803
asS'd'
I12475
sa(dp24950
S'a'
I57452
sS's'
I17
sS'm'
(lp24951
I1
aI1
aI7
aI15
aI11
aI33
aI45
aI127
aI321
aI895
aI1543
aI4013
aI6179
aI14209
aI13317
aI46803
aI99891
asS'd'
I12476
sa(dp24952
S'a'
I57458
sS's'
I17
sS'm'
(lp24953
I1
aI3
aI3
aI13
aI13
aI11
aI101
aI11
aI177
aI869
aI509
aI2323
aI449
aI16379
aI31965
aI2899
aI59229
asS'd'
I12477
sa(dp24954
S'a'
I57467
sS's'
I17
sS'm'
(lp24955
I1
aI1
aI5
aI7
aI15
aI37
aI113
aI237
aI463
aI489
aI1145
aI1629
aI3101
aI10305
aI31705
aI29957
aI99665
asS'd'
I12478
sa(dp24956
S'a'
I57470
sS's'
I17
sS'm'
(lp24957
I1
aI3
aI5
aI7
aI21
aI29
aI45
aI133
aI367
aI657
aI1315
aI537
aI6069
aI8141
aI31479
aI32983
aI57
asS'd'
I12479
sa(dp24958
S'a'
I57473
sS's'
I17
sS'm'
(lp24959
I1
aI1
aI3
aI9
aI13
aI39
aI109
aI125
aI467
aI975
aI829
aI4007
aI773
aI6639
aI8793
aI4579
aI60547
asS'd'
I12480
sa(dp24960
S'a'
I57479
sS's'
I17
sS'm'
(lp24961
I1
aI1
aI1
aI15
aI21
aI5
aI113
aI51
aI159
aI501
aI1921
aI4095
aI5603
aI16055
aI16649
aI50229
aI49863
asS'd'
I12481
sa(dp24962
S'a'
I57480
sS's'
I17
sS'm'
(lp24963
I1
aI1
aI5
aI13
aI9
aI21
aI77
aI187
aI355
aI299
aI1017
aI491
aI6725
aI4177
aI16739
aI15909
aI84069
asS'd'
I12482
sa(dp24964
S'a'
I57486
sS's'
I17
sS'm'
(lp24965
I1
aI3
aI7
aI5
aI13
aI13
aI51
aI19
aI159
aI339
aI735
aI933
aI2523
aI11435
aI20793
aI21975
aI19007
asS'd'
I12483
sa(dp24966
S'a'
I57491
sS's'
I17
sS'm'
(lp24967
I1
aI1
aI7
aI5
aI1
aI11
aI61
aI111
aI129
aI643
aI1741
aI945
aI7349
aI11579
aI24793
aI1751
aI2367
asS'd'
I12484
sa(dp24968
S'a'
I57493
sS's'
I17
sS'm'
(lp24969
I1
aI1
aI5
aI5
aI29
aI45
aI63
aI177
aI507
aI277
aI1789
aI729
aI4277
aI10099
aI28985
aI43009
aI2319
asS'd'
I12485
sa(dp24970
S'a'
I57524
sS's'
I17
sS'm'
(lp24971
I1
aI3
aI3
aI13
aI19
aI53
aI73
aI227
aI487
aI131
aI1227
aI3735
aI3979
aI7383
aI6923
aI31979
aI6651
asS'd'
I12486
sa(dp24972
S'a'
I57528
sS's'
I17
sS'm'
(lp24973
I1
aI3
aI3
aI15
aI7
aI47
aI31
aI255
aI317
aI621
aI497
aI4069
aI5249
aI15093
aI18013
aI6891
aI81893
asS'd'
I12487
sa(dp24974
S'a'
I57534
sS's'
I17
sS'm'
(lp24975
I1
aI1
aI7
aI5
aI17
aI23
aI119
aI27
aI55
aI555
aI221
aI2693
aI1757
aI11117
aI23409
aI21135
aI122977
asS'd'
I12488
sa(dp24976
S'a'
I57542
sS's'
I17
sS'm'
(lp24977
I1
aI1
aI5
aI1
aI15
aI17
aI127
aI109
aI75
aI1017
aI1167
aI2975
aI3249
aI5399
aI12599
aI50779
aI78215
asS'd'
I12489
sa(dp24978
S'a'
I57560
sS's'
I17
sS'm'
(lp24979
I1
aI3
aI5
aI5
aI1
aI13
aI65
aI199
aI101
aI75
aI513
aI493
aI6931
aI9363
aI5607
aI16331
aI69219
asS'd'
I12490
sa(dp24980
S'a'
I57565
sS's'
I17
sS'm'
(lp24981
I1
aI1
aI1
aI15
aI19
aI37
aI63
aI177
aI397
aI645
aI905
aI1599
aI609
aI14291
aI14681
aI46719
aI117745
asS'd'
I12491
sa(dp24982
S'a'
I57570
sS's'
I17
sS'm'
(lp24983
I1
aI3
aI7
aI13
aI3
aI57
aI63
aI49
aI67
aI913
aI1659
aI1857
aI3595
aI9219
aI11341
aI39735
aI82275
asS'd'
I12492
sa(dp24984
S'a'
I57579
sS's'
I17
sS'm'
(lp24985
I1
aI1
aI5
aI11
aI19
aI1
aI5
aI181
aI55
aI763
aI469
aI2417
aI2349
aI12437
aI22589
aI17867
aI95701
asS'd'
I12493
sa(dp24986
S'a'
I57590
sS's'
I17
sS'm'
(lp24987
I1
aI3
aI7
aI13
aI1
aI9
aI11
aI159
aI103
aI737
aI1989
aI59
aI4711
aI8093
aI31703
aI45663
aI92913
asS'd'
I12494
sa(dp24988
S'a'
I57601
sS's'
I17
sS'm'
(lp24989
I1
aI3
aI3
aI11
aI15
aI9
aI35
aI217
aI299
aI479
aI1539
aI2605
aI2003
aI8841
aI27261
aI28853
aI98877
asS'd'
I12495
sa(dp24990
S'a'
I57604
sS's'
I17
sS'm'
(lp24991
I1
aI1
aI5
aI5
aI15
aI45
aI73
aI159
aI205
aI1017
aI159
aI3659
aI2377
aI3651
aI6489
aI19711
aI109959
asS'd'
I12496
sa(dp24992
S'a'
I57607
sS's'
I17
sS'm'
(lp24993
I1
aI1
aI3
aI15
aI7
aI41
aI7
aI47
aI357
aI433
aI211
aI111
aI2565
aI8637
aI15917
aI47887
aI128513
asS'd'
I12497
sa(dp24994
S'a'
I57621
sS's'
I17
sS'm'
(lp24995
I1
aI3
aI7
aI3
aI25
aI9
aI33
aI33
aI5
aI805
aI1541
aI3333
aI7257
aI9011
aI4813
aI53195
aI52469
asS'd'
I12498
sa(dp24996
S'a'
I57635
sS's'
I17
sS'm'
(lp24997
I1
aI1
aI7
aI15
aI27
aI3
aI11
aI57
aI415
aI371
aI563
aI2515
aI149
aI6555
aI31273
aI51465
aI2989
asS'd'
I12499
sa(dp24998
S'a'
I57642
sS's'
I17
sS'm'
(lp24999
I1
aI1
aI1
aI11
aI3
aI41
aI25
aI175
aI499
aI879
aI1145
aI1083
aI7857
aI10497
aI16991
aI23351
aI115353
asS'd'
I12500
sa(dp25000
S'a'
I57650
sS's'
I17
sS'm'
(lp25001
I1
aI1
aI3
aI15
aI15
aI9
aI39
aI239
aI285
aI413
aI989
aI2927
aI7205
aI79
aI20101
aI30115
aI113933
asS'd'
I12501
sa(dp25002
S'a'
I57655
sS's'
I17
sS'm'
(lp25003
I1
aI1
aI7
aI5
aI13
aI29
aI29
aI69
aI25
aI533
aI1731
aI1391
aI4065
aI12597
aI19167
aI51989
aI101273
asS'd'
I12502
sa(dp25004
S'a'
I57670
sS's'
I17
sS'm'
(lp25005
I1
aI1
aI5
aI3
aI9
aI41
aI83
aI127
aI197
aI627
aI173
aI273
aI593
aI15733
aI19285
aI26517
aI107877
asS'd'
I12503
sa(dp25006
S'a'
I57676
sS's'
I17
sS'm'
(lp25007
I1
aI1
aI1
aI5
aI29
aI17
aI89
aI163
aI125
aI473
aI1577
aI2435
aI3379
aI3057
aI1829
aI64325
aI111719
asS'd'
I12504
sa(dp25008
S'a'
I57682
sS's'
I17
sS'm'
(lp25009
I1
aI3
aI1
aI11
aI1
aI29
aI97
aI53
aI421
aI941
aI1737
aI3337
aI2715
aI1633
aI28485
aI30369
aI116047
asS'd'
I12505
sa(dp25010
S'a'
I57687
sS's'
I17
sS'm'
(lp25011
I1
aI1
aI5
aI13
aI25
aI43
aI39
aI85
aI385
aI129
aI819
aI1647
aI3527
aI12319
aI573
aI58703
aI29463
asS'd'
I12506
sa(dp25012
S'a'
I57691
sS's'
I17
sS'm'
(lp25013
I1
aI3
aI3
aI9
aI23
aI59
aI31
aI215
aI49
aI451
aI645
aI3687
aI4507
aI9359
aI28161
aI34609
aI123409
asS'd'
I12507
sa(dp25014
S'a'
I57700
sS's'
I17
sS'm'
(lp25015
I1
aI3
aI1
aI9
aI13
aI25
aI97
aI239
aI203
aI31
aI1465
aI1089
aI7665
aI3007
aI22465
aI28389
aI119869
asS'd'
I12508
sa(dp25016
S'a'
I57704
sS's'
I17
sS'm'
(lp25017
I1
aI3
aI3
aI1
aI29
aI51
aI19
aI83
aI443
aI193
aI647
aI3125
aI7269
aI3031
aI9967
aI40447
aI102179
asS'd'
I12509
sa(dp25018
S'a'
I57707
sS's'
I17
sS'm'
(lp25019
I1
aI3
aI5
aI15
aI5
aI63
aI125
aI55
aI295
aI787
aI559
aI3309
aI7491
aI9907
aI5775
aI15155
aI41793
asS'd'
I12510
sa(dp25020
S'a'
I57721
sS's'
I17
sS'm'
(lp25021
I1
aI3
aI7
aI7
aI21
aI61
aI125
aI207
aI43
aI159
aI539
aI435
aI1945
aI725
aI797
aI47489
aI43099
asS'd'
I12511
sa(dp25022
S'a'
I57731
sS's'
I17
sS'm'
(lp25023
I1
aI1
aI5
aI7
aI11
aI11
aI61
aI103
aI23
aI693
aI493
aI1045
aI4435
aI9009
aI22075
aI24839
aI125431
asS'd'
I12512
sa(dp25024
S'a'
I57733
sS's'
I17
sS'm'
(lp25025
I1
aI3
aI1
aI9
aI13
aI61
aI83
aI181
aI373
aI949
aI1397
aI247
aI5079
aI10933
aI31887
aI14147
aI55121
asS'd'
I12513
sa(dp25026
S'a'
I57738
sS's'
I17
sS'm'
(lp25027
I1
aI1
aI7
aI15
aI31
aI47
aI87
aI219
aI357
aI409
aI943
aI2993
aI7615
aI7071
aI14179
aI41489
aI104401
asS'd'
I12514
sa(dp25028
S'a'
I57745
sS's'
I17
sS'm'
(lp25029
I1
aI3
aI3
aI1
aI11
aI15
aI87
aI241
aI389
aI761
aI1523
aI3049
aI2111
aI11581
aI5493
aI11301
aI32017
asS'd'
I12515
sa(dp25030
S'a'
I57748
sS's'
I17
sS'm'
(lp25031
I1
aI1
aI7
aI11
aI29
aI39
aI37
aI63
aI411
aI671
aI1789
aI3541
aI5651
aI11721
aI10871
aI53867
aI112895
asS'd'
I12516
sa(dp25032
S'a'
I57751
sS's'
I17
sS'm'
(lp25033
I1
aI3
aI3
aI1
aI23
aI21
aI99
aI161
aI467
aI197
aI1263
aI451
aI5469
aI7667
aI22139
aI31599
aI8345
asS'd'
I12517
sa(dp25034
S'a'
I57757
sS's'
I17
sS'm'
(lp25035
I1
aI3
aI1
aI3
aI11
aI41
aI13
aI79
aI3
aI377
aI1865
aI2297
aI1383
aI14033
aI17141
aI24787
aI127911
asS'd'
I12518
sa(dp25036
S'a'
I57761
sS's'
I17
sS'm'
(lp25037
I1
aI1
aI1
aI1
aI5
aI11
aI5
aI181
aI323
aI853
aI831
aI1599
aI3939
aI6391
aI22817
aI37969
aI9997
asS'd'
I12519
sa(dp25038
S'a'
I57774
sS's'
I17
sS'm'
(lp25039
I1
aI1
aI3
aI11
aI29
aI61
aI11
aI171
aI181
aI631
aI757
aI3879
aI4779
aI16183
aI5969
aI7909
aI42855
asS'd'
I12520
sa(dp25040
S'a'
I57779
sS's'
I17
sS'm'
(lp25041
I1
aI3
aI3
aI15
aI19
aI17
aI107
aI187
aI19
aI743
aI909
aI1963
aI2131
aI2107
aI659
aI35829
aI57905
asS'd'
I12521
sa(dp25042
S'a'
I57786
sS's'
I17
sS'm'
(lp25043
I1
aI1
aI7
aI1
aI31
aI29
aI67
aI99
aI353
aI715
aI65
aI3907
aI1931
aI1289
aI9217
aI60635
aI32737
asS'd'
I12522
sa(dp25044
S'a'
I57808
sS's'
I17
sS'm'
(lp25045
I1
aI3
aI3
aI13
aI3
aI17
aI19
aI173
aI263
aI203
aI1233
aI1407
aI933
aI8905
aI26905
aI63343
aI64963
asS'd'
I12523
sa(dp25046
S'a'
I57814
sS's'
I17
sS'm'
(lp25047
I1
aI3
aI1
aI13
aI21
aI43
aI97
aI79
aI163
aI529
aI1571
aI1027
aI4339
aI16235
aI9189
aI29203
aI36789
asS'd'
I12524
sa(dp25048
S'a'
I57817
sS's'
I17
sS'm'
(lp25049
I1
aI1
aI7
aI1
aI31
aI1
aI117
aI247
aI193
aI333
aI1797
aI3515
aI285
aI2803
aI25345
aI9101
aI110569
asS'd'
I12525
sa(dp25050
S'a'
I57824
sS's'
I17
sS'm'
(lp25051
I1
aI1
aI5
aI11
aI17
aI63
aI45
aI167
aI35
aI925
aI569
aI3687
aI6739
aI12453
aI26171
aI28249
aI73827
asS'd'
I12526
sa(dp25052
S'a'
I57827
sS's'
I17
sS'm'
(lp25053
I1
aI1
aI5
aI1
aI19
aI49
aI107
aI183
aI435
aI305
aI1191
aI2711
aI891
aI15813
aI13449
aI23489
aI89749
asS'd'
I12527
sa(dp25054
S'a'
I57848
sS's'
I17
sS'm'
(lp25055
I1
aI1
aI1
aI11
aI7
aI53
aI105
aI17
aI433
aI351
aI1151
aI4077
aI5371
aI10183
aI18895
aI13229
aI101219
asS'd'
I12528
sa(dp25056
S'a'
I57853
sS's'
I17
sS'm'
(lp25057
I1
aI3
aI7
aI9
aI21
aI33
aI27
aI125
aI177
aI607
aI817
aI2689
aI2123
aI2037
aI29643
aI27219
aI44591
asS'd'
I12529
sa(dp25058
S'a'
I57857
sS's'
I17
sS'm'
(lp25059
I1
aI3
aI5
aI11
aI27
aI51
aI29
aI161
aI469
aI349
aI1445
aI3613
aI1487
aI961
aI29017
aI2235
aI45905
asS'd'
I12530
sa(dp25060
S'a'
I57858
sS's'
I17
sS'm'
(lp25061
I1
aI1
aI3
aI11
aI7
aI15
aI37
aI195
aI329
aI875
aI559
aI1361
aI7373
aI7143
aI15059
aI59205
aI37167
asS'd'
I12531
sa(dp25062
S'a'
I57867
sS's'
I17
sS'm'
(lp25063
I1
aI1
aI5
aI1
aI7
aI25
aI27
aI211
aI301
aI369
aI227
aI123
aI4415
aI15993
aI4829
aI43801
aI83639
asS'd'
I12532
sa(dp25064
S'a'
I57872
sS's'
I17
sS'm'
(lp25065
I1
aI1
aI3
aI5
aI25
aI37
aI31
aI205
aI69
aI275
aI855
aI1407
aI2989
aI11001
aI16963
aI31497
aI3505
asS'd'
I12533
sa(dp25066
S'a'
I57887
sS's'
I17
sS'm'
(lp25067
I1
aI1
aI7
aI7
aI25
aI33
aI81
aI181
aI197
aI717
aI207
aI535
aI8083
aI5765
aI2523
aI40347
aI27245
asS'd'
I12534
sa(dp25068
S'a'
I57894
sS's'
I17
sS'm'
(lp25069
I1
aI3
aI1
aI5
aI13
aI3
aI107
aI143
aI233
aI419
aI1831
aI2149
aI7277
aI13449
aI31609
aI48345
aI82621
asS'd'
I12535
sa(dp25070
S'a'
I57915
sS's'
I17
sS'm'
(lp25071
I1
aI3
aI5
aI9
aI31
aI35
aI39
aI113
aI415
aI803
aI1479
aI3169
aI8015
aI4659
aI2445
aI9159
aI91625
asS'd'
I12536
sa(dp25072
S'a'
I57920
sS's'
I17
sS'm'
(lp25073
I1
aI3
aI1
aI5
aI27
aI25
aI83
aI165
aI459
aI955
aI1535
aI377
aI5531
aI2945
aI18285
aI18097
aI21589
asS'd'
I12537
sa(dp25074
S'a'
I57923
sS's'
I17
sS'm'
(lp25075
I1
aI1
aI1
aI15
aI13
aI23
aI11
aI159
aI471
aI951
aI1971
aI677
aI2641
aI3227
aI14761
aI39421
aI29841
asS'd'
I12538
sa(dp25076
S'a'
I57932
sS's'
I17
sS'm'
(lp25077
I1
aI1
aI7
aI11
aI19
aI37
aI101
aI97
aI373
aI559
aI105
aI905
aI897
aI15309
aI14979
aI52029
aI38989
asS'd'
I12539
sa(dp25078
S'a'
I57937
sS's'
I17
sS'm'
(lp25079
I1
aI1
aI5
aI15
aI31
aI43
aI71
aI75
aI481
aI997
aI5
aI1005
aI4987
aI7907
aI16237
aI43075
aI94827
asS'd'
I12540
sa(dp25080
S'a'
I57943
sS's'
I17
sS'm'
(lp25081
I1
aI3
aI5
aI11
aI23
aI55
aI123
aI91
aI183
aI307
aI97
aI1999
aI3341
aI4717
aI19643
aI9455
aI26555
asS'd'
I12541
sa(dp25082
S'a'
I57950
sS's'
I17
sS'm'
(lp25083
I1
aI1
aI7
aI3
aI29
aI61
aI75
aI197
aI125
aI579
aI145
aI1333
aI85
aI15655
aI28177
aI6169
aI43289
asS'd'
I12542
sa(dp25084
S'a'
I57960
sS's'
I17
sS'm'
(lp25085
I1
aI3
aI7
aI13
aI21
aI15
aI1
aI221
aI293
aI991
aI35
aI2701
aI2909
aI7333
aI27319
aI32281
aI77861
asS'd'
I12543
sa(dp25086
S'a'
I57971
sS's'
I17
sS'm'
(lp25087
I1
aI1
aI3
aI1
aI17
aI11
aI81
aI209
aI499
aI125
aI105
aI2987
aI3891
aI9531
aI27963
aI39611
aI43633
asS'd'
I12544
sa(dp25088
S'a'
I57974
sS's'
I17
sS'm'
(lp25089
I1
aI3
aI3
aI3
aI29
aI45
aI51
aI63
aI209
aI75
aI1321
aI4055
aI7851
aI12329
aI8371
aI59513
aI49105
asS'd'
I12545
sa(dp25090
S'a'
I57983
sS's'
I17
sS'm'
(lp25091
I1
aI1
aI7
aI5
aI9
aI63
aI59
aI127
aI445
aI899
aI1857
aI3737
aI625
aI15021
aI25177
aI21007
aI25935
asS'd'
I12546
sa(dp25092
S'a'
I57987
sS's'
I17
sS'm'
(lp25093
I1
aI1
aI5
aI9
aI9
aI57
aI25
aI63
aI235
aI191
aI1527
aI1783
aI1401
aI1813
aI2553
aI1241
aI100029
asS'd'
I12547
sa(dp25094
S'a'
I57994
sS's'
I17
sS'm'
(lp25095
I1
aI3
aI1
aI11
aI27
aI5
aI49
aI39
aI505
aI653
aI1083
aI921
aI1045
aI11337
aI25499
aI36211
aI75851
asS'd'
I12548
sa(dp25096
S'a'
I58001
sS's'
I17
sS'm'
(lp25097
I1
aI3
aI3
aI1
aI29
aI19
aI71
aI93
aI223
aI153
aI561
aI1657
aI5821
aI14181
aI1275
aI24633
aI114787
asS'd'
I12549
sa(dp25098
S'a'
I58002
sS's'
I17
sS'm'
(lp25099
I1
aI1
aI1
aI9
aI7
aI13
aI17
aI239
aI457
aI731
aI1811
aI3015
aI4465
aI5865
aI2387
aI30455
aI17405
asS'd'
I12550
sa(dp25100
S'a'
I58013
sS's'
I17
sS'm'
(lp25101
I1
aI3
aI5
aI1
aI7
aI45
aI117
aI207
aI7
aI645
aI131
aI863
aI6443
aI14455
aI11885
aI39257
aI126431
asS'd'
I12551
sa(dp25102
S'a'
I58023
sS's'
I17
sS'm'
(lp25103
I1
aI1
aI3
aI5
aI17
aI23
aI3
aI27
aI475
aI731
aI431
aI1967
aI981
aI12727
aI1301
aI17647
aI62481
asS'd'
I12552
sa(dp25104
S'a'
I58027
sS's'
I17
sS'm'
(lp25105
I1
aI1
aI1
aI13
aI11
aI53
aI105
aI109
aI105
aI461
aI1787
aI2851
aI1299
aI9925
aI13055
aI58301
aI24483
asS'd'
I12553
sa(dp25106
S'a'
I58029
sS's'
I17
sS'm'
(lp25107
I1
aI1
aI1
aI11
aI7
aI35
aI11
aI173
aI247
aI289
aI1269
aI361
aI2059
aI5051
aI25731
aI58085
aI21387
asS'd'
I12554
sa(dp25108
S'a'
I58032
sS's'
I17
sS'm'
(lp25109
I1
aI3
aI1
aI5
aI9
aI31
aI59
aI151
aI435
aI519
aI1863
aI2255
aI2585
aI10033
aI29189
aI27189
aI25023
asS'd'
I12555
sa(dp25110
S'a'
I58041
sS's'
I17
sS'm'
(lp25111
I1
aI1
aI5
aI5
aI25
aI17
aI53
aI189
aI187
aI847
aI369
aI3287
aI6047
aI2385
aI26045
aI48821
aI23335
asS'd'
I12556
sa(dp25112
S'a'
I58047
sS's'
I17
sS'm'
(lp25113
I1
aI1
aI1
aI9
aI15
aI55
aI79
aI21
aI433
aI131
aI1677
aI3681
aI6173
aI9189
aI15053
aI12971
aI81135
asS'd'
I12557
sa(dp25114
S'a'
I58062
sS's'
I17
sS'm'
(lp25115
I1
aI1
aI3
aI5
aI29
aI5
aI11
aI213
aI151
aI77
aI1493
aI3959
aI7675
aI7689
aI10381
aI15871
aI123463
asS'd'
I12558
sa(dp25116
S'a'
I58086
sS's'
I17
sS'm'
(lp25117
I1
aI3
aI7
aI11
aI19
aI47
aI31
aI81
aI207
aI789
aI1945
aI2671
aI1987
aI6363
aI26401
aI9799
aI59531
asS'd'
I12559
sa(dp25118
S'a'
I58092
sS's'
I17
sS'm'
(lp25119
I1
aI1
aI1
aI7
aI17
aI7
aI83
aI181
aI339
aI423
aI267
aI2251
aI1847
aI14883
aI13423
aI27223
aI81799
asS'd'
I12560
sa(dp25120
S'a'
I58095
sS's'
I17
sS'm'
(lp25121
I1
aI3
aI5
aI15
aI9
aI9
aI73
aI95
aI197
aI267
aI743
aI2369
aI6417
aI3555
aI9885
aI6373
aI87651
asS'd'
I12561
sa(dp25122
S'a'
I58097
sS's'
I17
sS'm'
(lp25123
I1
aI1
aI3
aI13
aI21
aI29
aI57
aI59
aI395
aI631
aI993
aI17
aI2939
aI14117
aI27521
aI61387
aI74927
asS'd'
I12562
sa(dp25124
S'a'
I58100
sS's'
I17
sS'm'
(lp25125
I1
aI1
aI3
aI13
aI1
aI29
aI23
aI151
aI381
aI591
aI1217
aI2295
aI4403
aI15865
aI4325
aI31329
aI56989
asS'd'
I12563
sa(dp25126
S'a'
I58117
sS's'
I17
sS'm'
(lp25127
I1
aI3
aI5
aI1
aI27
aI7
aI111
aI79
aI287
aI945
aI1237
aI2857
aI1461
aI2477
aI10929
aI17117
aI98677
asS'd'
I12564
sa(dp25128
S'a'
I58121
sS's'
I17
sS'm'
(lp25129
I1
aI1
aI1
aI5
aI15
aI35
aI109
aI39
aI391
aI57
aI1783
aI3133
aI5563
aI7721
aI12651
aI3437
aI46697
asS'd'
I12565
sa(dp25130
S'a'
I58132
sS's'
I17
sS'm'
(lp25131
I1
aI3
aI5
aI5
aI13
aI47
aI123
aI51
aI383
aI91
aI641
aI1607
aI4461
aI7
aI6427
aI42755
aI130097
asS'd'
I12566
sa(dp25132
S'a'
I58142
sS's'
I17
sS'm'
(lp25133
I1
aI1
aI5
aI5
aI25
aI49
aI33
aI59
aI351
aI1003
aI301
aI2721
aI7705
aI5447
aI21367
aI63007
aI112465
asS'd'
I12567
sa(dp25134
S'a'
I58145
sS's'
I17
sS'm'
(lp25135
I1
aI1
aI5
aI13
aI31
aI11
aI73
aI1
aI359
aI201
aI1465
aI187
aI7385
aI12817
aI12911
aI17775
aI77937
asS'd'
I12568
sa(dp25136
S'a'
I58152
sS's'
I17
sS'm'
(lp25137
I1
aI3
aI1
aI15
aI29
aI5
aI117
aI69
aI79
aI93
aI947
aI133
aI1049
aI13907
aI3611
aI42123
aI8041
asS'd'
I12569
sa(dp25138
S'a'
I58155
sS's'
I17
sS'm'
(lp25139
I1
aI1
aI7
aI7
aI29
aI17
aI71
aI171
aI357
aI619
aI1199
aI3817
aI6889
aI9607
aI19075
aI15539
aI54939
asS'd'
I12570
sa(dp25140
S'a'
I58163
sS's'
I17
sS'm'
(lp25141
I1
aI1
aI1
aI7
aI23
aI29
aI91
aI227
aI119
aI865
aI943
aI381
aI5289
aI189
aI15917
aI13875
aI31095
asS'd'
I12571
sa(dp25142
S'a'
I58166
sS's'
I17
sS'm'
(lp25143
I1
aI1
aI7
aI1
aI9
aI49
aI93
aI175
aI105
aI61
aI1285
aI3659
aI2383
aI505
aI12337
aI43801
aI67035
asS'd'
I12572
sa(dp25144
S'a'
I58170
sS's'
I17
sS'm'
(lp25145
I1
aI3
aI5
aI1
aI31
aI47
aI75
aI17
aI455
aI231
aI1887
aI2295
aI2533
aI733
aI29001
aI22001
aI119423
asS'd'
I12573
sa(dp25146
S'a'
I58175
sS's'
I17
sS'm'
(lp25147
I1
aI3
aI7
aI3
aI5
aI5
aI69
aI205
aI459
aI805
aI1257
aI3283
aI3305
aI6845
aI1405
aI56051
aI63453
asS'd'
I12574
sa(dp25148
S'a'
I58184
sS's'
I17
sS'm'
(lp25149
I1
aI3
aI3
aI3
aI19
aI57
aI65
aI149
aI175
aI683
aI1719
aI637
aI4951
aI13645
aI30455
aI22379
aI105369
asS'd'
I12575
sa(dp25150
S'a'
I58197
sS's'
I17
sS'm'
(lp25151
I1
aI3
aI1
aI11
aI31
aI27
aI69
aI13
aI191
aI645
aI1563
aI3897
aI2387
aI2269
aI12999
aI42217
aI100853
asS'd'
I12576
sa(dp25152
S'a'
I58202
sS's'
I17
sS'm'
(lp25153
I1
aI3
aI1
aI13
aI31
aI7
aI11
aI227
aI267
aI373
aI1249
aI2591
aI4769
aI303
aI5865
aI59911
aI4991
asS'd'
I12577
sa(dp25154
S'a'
I58204
sS's'
I17
sS'm'
(lp25155
I1
aI3
aI7
aI11
aI7
aI31
aI87
aI205
aI45
aI785
aI1947
aI1965
aI5851
aI7501
aI89
aI38897
aI91939
asS'd'
I12578
sa(dp25156
S'a'
I58208
sS's'
I17
sS'm'
(lp25157
I1
aI3
aI5
aI9
aI21
aI37
aI79
aI181
aI163
aI355
aI1129
aI3439
aI5103
aI15903
aI2935
aI22505
aI97451
asS'd'
I12579
sa(dp25158
S'a'
I58213
sS's'
I17
sS'm'
(lp25159
I1
aI1
aI3
aI9
aI7
aI37
aI107
aI255
aI69
aI133
aI551
aI1357
aI31
aI7059
aI29195
aI30151
aI81785
asS'd'
I12580
sa(dp25160
S'a'
I58220
sS's'
I17
sS'm'
(lp25161
I1
aI3
aI3
aI5
aI3
aI51
aI3
aI143
aI305
aI697
aI1041
aI1267
aI7635
aI1483
aI649
aI28275
aI65059
asS'd'
I12581
sa(dp25162
S'a'
I58226
sS's'
I17
sS'm'
(lp25163
I1
aI3
aI1
aI7
aI11
aI13
aI113
aI81
aI159
aI993
aI797
aI3207
aI3787
aI13005
aI29393
aI51107
aI42709
asS'd'
I12582
sa(dp25164
S'a'
I58231
sS's'
I17
sS'm'
(lp25165
I1
aI1
aI3
aI13
aI7
aI11
aI113
aI195
aI349
aI377
aI2033
aI197
aI147
aI13839
aI5405
aI29577
aI64535
asS'd'
I12583
sa(dp25166
S'a'
I58235
sS's'
I17
sS'm'
(lp25167
I1
aI3
aI1
aI9
aI1
aI1
aI77
aI73
aI383
aI463
aI1871
aI1589
aI1805
aI2085
aI5195
aI17805
aI33159
asS'd'
I12584
sa(dp25168
S'a'
I58254
sS's'
I17
sS'm'
(lp25169
I1
aI1
aI1
aI9
aI25
aI37
aI123
aI119
aI507
aI515
aI1547
aI139
aI7501
aI4725
aI30195
aI18199
aI41145
asS'd'
I12585
sa(dp25170
S'a'
I58261
sS's'
I17
sS'm'
(lp25171
I1
aI3
aI3
aI7
aI17
aI33
aI99
aI11
aI487
aI343
aI403
aI337
aI3831
aI6031
aI20375
aI2071
aI39795
asS'd'
I12586
sa(dp25172
S'a'
I58271
sS's'
I17
sS'm'
(lp25173
I1
aI1
aI5
aI1
aI13
aI57
aI59
aI11
aI391
aI953
aI597
aI1411
aI3929
aI13963
aI15003
aI59385
aI12293
asS'd'
I12587
sa(dp25174
S'a'
I58278
sS's'
I17
sS'm'
(lp25175
I1
aI3
aI1
aI15
aI23
aI61
aI67
aI129
aI63
aI555
aI433
aI631
aI2725
aI317
aI10121
aI9217
aI124371
asS'd'
I12588
sa(dp25176
S'a'
I58287
sS's'
I17
sS'm'
(lp25177
I1
aI1
aI5
aI7
aI19
aI23
aI17
aI45
aI75
aI819
aI1879
aI2315
aI1439
aI2643
aI26561
aI19209
aI16081
asS'd'
I12589
sa(dp25178
S'a'
I58327
sS's'
I17
sS'm'
(lp25179
I1
aI3
aI5
aI7
aI5
aI35
aI121
aI233
aI91
aI89
aI1741
aI3015
aI4223
aI9605
aI22795
aI55845
aI47167
asS'd'
I12590
sa(dp25180
S'a'
I58333
sS's'
I17
sS'm'
(lp25181
I1
aI3
aI7
aI7
aI13
aI23
aI11
aI51
aI91
aI367
aI773
aI1303
aI2151
aI14423
aI22263
aI28413
aI107461
asS'd'
I12591
sa(dp25182
S'a'
I58337
sS's'
I17
sS'm'
(lp25183
I1
aI1
aI1
aI11
aI29
aI17
aI49
aI53
aI379
aI819
aI1551
aI2907
aI5805
aI2167
aI16123
aI9263
aI43903
asS'd'
I12592
sa(dp25184
S'a'
I58357
sS's'
I17
sS'm'
(lp25185
I1
aI3
aI1
aI13
aI19
aI27
aI33
aI127
aI131
aI697
aI489
aI3289
aI1895
aI3243
aI19497
aI32419
aI55741
asS'd'
I12593
sa(dp25186
S'a'
I58362
sS's'
I17
sS'm'
(lp25187
I1
aI3
aI1
aI15
aI7
aI51
aI91
aI45
aI251
aI829
aI2015
aI2659
aI151
aI3327
aI25281
aI49291
aI106343
asS'd'
I12594
sa(dp25188
S'a'
I58364
sS's'
I17
sS'm'
(lp25189
I1
aI1
aI5
aI5
aI27
aI11
aI85
aI45
aI203
aI413
aI293
aI3067
aI4109
aI959
aI22337
aI38767
aI75829
asS'd'
I12595
sa(dp25190
S'a'
I58375
sS's'
I17
sS'm'
(lp25191
I1
aI1
aI3
aI5
aI21
aI35
aI65
aI173
aI315
aI423
aI171
aI3837
aI817
aI10153
aI5517
aI18115
aI117437
asS'd'
I12596
sa(dp25192
S'a'
I58384
sS's'
I17
sS'm'
(lp25193
I1
aI1
aI1
aI7
aI9
aI29
aI101
aI55
aI253
aI37
aI1717
aI7
aI3181
aI1067
aI20637
aI54773
aI106777
asS'd'
I12597
sa(dp25194
S'a'
I58387
sS's'
I17
sS'm'
(lp25195
I1
aI3
aI7
aI11
aI13
aI11
aI87
aI83
aI83
aI969
aI589
aI3625
aI3373
aI3921
aI24487
aI34235
aI96289
asS'd'
I12598
sa(dp25196
S'a'
I58405
sS's'
I17
sS'm'
(lp25197
I1
aI3
aI1
aI11
aI25
aI35
aI65
aI131
aI491
aI167
aI449
aI2949
aI5807
aI5649
aI7569
aI37363
aI106819
asS'd'
I12599
sa(dp25198
S'a'
I58406
sS's'
I17
sS'm'
(lp25199
I1
aI1
aI5
aI1
aI9
aI37
aI33
aI237
aI79
aI163
aI1791
aI3499
aI4951
aI2009
aI16183
aI10121
aI40221
asS'd'
I12600
sa(dp25200
S'a'
I58417
sS's'
I17
sS'm'
(lp25201
I1
aI1
aI5
aI15
aI21
aI49
aI123
aI53
aI27
aI273
aI1655
aI3713
aI5699
aI1659
aI31985
aI6779
aI10195
asS'd'
I12601
sa(dp25202
S'a'
I58424
sS's'
I17
sS'm'
(lp25203
I1
aI3
aI5
aI11
aI27
aI37
aI93
aI17
aI263
aI41
aI1583
aI703
aI7689
aI6667
aI15497
aI15255
aI67153
asS'd'
I12602
sa(dp25204
S'a'
I58435
sS's'
I17
sS'm'
(lp25205
I1
aI1
aI5
aI9
aI19
aI13
aI121
aI39
aI321
aI177
aI711
aI3223
aI3859
aI14775
aI19061
aI40587
aI97199
asS'd'
I12603
sa(dp25206
S'a'
I58450
sS's'
I17
sS'm'
(lp25207
I1
aI3
aI7
aI9
aI23
aI11
aI93
aI243
aI3
aI811
aI679
aI1103
aI7579
aI14147
aI20255
aI22485
aI117179
asS'd'
I12604
sa(dp25208
S'a'
I58452
sS's'
I17
sS'm'
(lp25209
I1
aI3
aI3
aI5
aI23
aI9
aI21
aI243
aI27
aI501
aI1273
aI1501
aI5331
aI15663
aI19483
aI10637
aI90905
asS'd'
I12605
sa(dp25210
S'a'
I58459
sS's'
I17
sS'm'
(lp25211
I1
aI3
aI1
aI15
aI13
aI11
aI89
aI37
aI63
aI463
aI731
aI3615
aI3923
aI4677
aI21329
aI1069
aI57565
asS'd'
I12606
sa(dp25212
S'a'
I58461
sS's'
I17
sS'm'
(lp25213
I1
aI1
aI1
aI13
aI19
aI39
aI63
aI59
aI389
aI367
aI1285
aI2005
aI8103
aI4179
aI3805
aI55475
aI126589
asS'd'
I12607
sa(dp25214
S'a'
I58466
sS's'
I17
sS'm'
(lp25215
I1
aI3
aI5
aI11
aI27
aI57
aI27
aI63
aI365
aI171
aI665
aI2035
aI2831
aI15955
aI11551
aI26741
aI121059
asS'd'
I12608
sa(dp25216
S'a'
I58468
sS's'
I17
sS'm'
(lp25217
I1
aI3
aI3
aI3
aI23
aI19
aI13
aI7
aI331
aI613
aI1783
aI733
aI7465
aI1089
aI4683
aI15695
aI31125
asS'd'
I12609
sa(dp25218
S'a'
I58485
sS's'
I17
sS'm'
(lp25219
I1
aI1
aI5
aI3
aI3
aI23
aI95
aI157
aI303
aI987
aI307
aI2679
aI6173
aI6633
aI17727
aI30901
aI76109
asS'd'
I12610
sa(dp25220
S'a'
I58489
sS's'
I17
sS'm'
(lp25221
I1
aI1
aI1
aI1
aI11
aI21
aI111
aI81
aI31
aI727
aI1133
aI4083
aI5811
aI2707
aI31749
aI42939
aI92225
asS'd'
I12611
sa(dp25222
S'a'
I58492
sS's'
I17
sS'm'
(lp25223
I1
aI3
aI1
aI11
aI29
aI21
aI31
aI115
aI351
aI213
aI201
aI3511
aI3707
aI12821
aI8845
aI59789
aI59721
asS'd'
I12612
sa(dp25224
S'a'
I58495
sS's'
I17
sS'm'
(lp25225
I1
aI3
aI5
aI9
aI7
aI27
aI65
aI1
aI305
aI937
aI1201
aI2045
aI2431
aI12275
aI24431
aI3317
aI119671
asS'd'
I12613
sa(dp25226
S'a'
I58505
sS's'
I17
sS'm'
(lp25227
I1
aI1
aI7
aI11
aI19
aI43
aI107
aI107
aI261
aI933
aI1125
aI1875
aI1943
aI7477
aI20759
aI57853
aI29459
asS'd'
I12614
sa(dp25228
S'a'
I58523
sS's'
I17
sS'm'
(lp25229
I1
aI1
aI3
aI3
aI17
aI43
aI79
aI19
aI167
aI341
aI237
aI677
aI725
aI2353
aI12003
aI48921
aI20707
asS'd'
I12615
sa(dp25230
S'a'
I58535
sS's'
I17
sS'm'
(lp25231
I1
aI3
aI3
aI5
aI31
aI39
aI91
aI81
aI121
aI59
aI737
aI1767
aI4407
aI9159
aI10237
aI25361
aI38891
asS'd'
I12616
sa(dp25232
S'a'
I58536
sS's'
I17
sS'm'
(lp25233
I1
aI1
aI5
aI7
aI31
aI47
aI37
aI47
aI171
aI623
aI597
aI2807
aI4875
aI15139
aI18809
aI24279
aI36563
asS'd'
I12617
sa(dp25234
S'a'
I58549
sS's'
I17
sS'm'
(lp25235
I1
aI3
aI5
aI3
aI3
aI7
aI31
aI135
aI81
aI777
aI1055
aI337
aI4309
aI9575
aI11075
aI54429
aI30097
asS'd'
I12618
sa(dp25236
S'a'
I58550
sS's'
I17
sS'm'
(lp25237
I1
aI1
aI5
aI1
aI3
aI57
aI53
aI25
aI117
aI961
aI947
aI257
aI2645
aI6935
aI27511
aI54051
aI129095
asS'd'
I12619
sa(dp25238
S'a'
I58561
sS's'
I17
sS'm'
(lp25239
I1
aI1
aI1
aI15
aI23
aI45
aI11
aI173
aI407
aI373
aI1051
aI3519
aI1875
aI1291
aI4393
aI9047
aI102159
asS'd'
I12620
sa(dp25240
S'a'
I58562
sS's'
I17
sS'm'
(lp25241
I1
aI1
aI3
aI5
aI9
aI51
aI23
aI253
aI261
aI339
aI505
aI3601
aI265
aI8375
aI7241
aI40715
aI114439
asS'd'
I12621
sa(dp25242
S'a'
I58573
sS's'
I17
sS'm'
(lp25243
I1
aI1
aI7
aI7
aI3
aI63
aI5
aI233
aI259
aI947
aI1367
aI2699
aI6029
aI9371
aI6567
aI39961
aI31621
asS'd'
I12622
sa(dp25244
S'a'
I58576
sS's'
I17
sS'm'
(lp25245
I1
aI1
aI3
aI3
aI25
aI43
aI73
aI61
aI305
aI257
aI235
aI421
aI6621
aI1025
aI18831
aI44525
aI87665
asS'd'
I12623
sa(dp25246
S'a'
I58592
sS's'
I17
sS'm'
(lp25247
I1
aI3
aI5
aI7
aI19
aI33
aI77
aI163
aI191
aI115
aI1275
aI3551
aI3521
aI6767
aI30209
aI48895
aI114185
asS'd'
I12624
sa(dp25248
S'a'
I58601
sS's'
I17
sS'm'
(lp25249
I1
aI1
aI3
aI3
aI29
aI37
aI3
aI165
aI267
aI807
aI967
aI2893
aI3287
aI12249
aI21411
aI19291
aI91781
asS'd'
I12625
sa(dp25250
S'a'
I58622
sS's'
I17
sS'm'
(lp25251
I1
aI1
aI5
aI5
aI27
aI55
aI51
aI137
aI429
aI775
aI1525
aI3911
aI5367
aI5981
aI24373
aI16331
aI31887
asS'd'
I12626
sa(dp25252
S'a'
I58627
sS's'
I17
sS'm'
(lp25253
I1
aI1
aI5
aI11
aI17
aI49
aI37
aI243
aI509
aI563
aI1381
aI2013
aI7341
aI10509
aI10049
aI29135
aI32709
asS'd'
I12627
sa(dp25254
S'a'
I58629
sS's'
I17
sS'm'
(lp25255
I1
aI1
aI3
aI13
aI1
aI33
aI119
aI5
aI477
aI701
aI1329
aI1023
aI2091
aI12895
aI6443
aI1053
aI44741
asS'd'
I12628
sa(dp25256
S'a'
I58634
sS's'
I17
sS'm'
(lp25257
I1
aI3
aI5
aI7
aI13
aI19
aI1
aI113
aI3
aI481
aI1555
aI2857
aI3519
aI7903
aI16153
aI62909
aI78877
asS'd'
I12629
sa(dp25258
S'a'
I58639
sS's'
I17
sS'm'
(lp25259
I1
aI1
aI7
aI11
aI7
aI27
aI47
aI175
aI87
aI605
aI411
aI4065
aI8059
aI5023
aI27719
aI3111
aI62247
asS'd'
I12630
sa(dp25260
S'a'
I58641
sS's'
I17
sS'm'
(lp25261
I1
aI3
aI1
aI15
aI27
aI37
aI71
aI77
aI297
aI647
aI1651
aI2543
aI3925
aI4827
aI28587
aI49663
aI56581
asS'd'
I12631
sa(dp25262
S'a'
I58644
sS's'
I17
sS'm'
(lp25263
I1
aI3
aI5
aI1
aI19
aI49
aI63
aI9
aI43
aI243
aI931
aI3577
aI2371
aI10513
aI13691
aI27739
aI61011
asS'd'
I12632
sa(dp25264
S'a'
I58654
sS's'
I17
sS'm'
(lp25265
I1
aI1
aI3
aI1
aI23
aI27
aI53
aI149
aI157
aI735
aI1087
aI1529
aI6613
aI2493
aI4879
aI26771
aI123711
asS'd'
I12633
sa(dp25266
S'a'
I58663
sS's'
I17
sS'm'
(lp25267
I1
aI3
aI7
aI13
aI5
aI3
aI43
aI3
aI343
aI497
aI943
aI3443
aI4335
aI4779
aI4033
aI25871
aI10965
asS'd'
I12634
sa(dp25268
S'a'
I58667
sS's'
I17
sS'm'
(lp25269
I1
aI3
aI5
aI1
aI5
aI11
aI93
aI73
aI199
aI455
aI421
aI3495
aI4381
aI717
aI21033
aI41287
aI44743
asS'd'
I12635
sa(dp25270
S'a'
I58677
sS's'
I17
sS'm'
(lp25271
I1
aI1
aI3
aI9
aI13
aI33
aI77
aI143
aI81
aI57
aI1061
aI3205
aI2411
aI1347
aI23149
aI21913
aI119331
asS'd'
I12636
sa(dp25272
S'a'
I58682
sS's'
I17
sS'm'
(lp25273
I1
aI3
aI1
aI7
aI1
aI15
aI45
aI247
aI119
aI905
aI457
aI3939
aI4865
aI11191
aI27705
aI46315
aI68367
asS'd'
I12637
sa(dp25274
S'a'
I58690
sS's'
I17
sS'm'
(lp25275
I1
aI3
aI3
aI15
aI31
aI5
aI55
aI169
aI381
aI1009
aI1893
aI2751
aI411
aI8653
aI30367
aI54919
aI23541
asS'd'
I12638
sa(dp25276
S'a'
I58692
sS's'
I17
sS'm'
(lp25277
I1
aI3
aI7
aI1
aI29
aI27
aI77
aI239
aI109
aI125
aI355
aI2759
aI2229
aI3435
aI16241
aI53309
aI613
asS'd'
I12639
sa(dp25278
S'a'
I58695
sS's'
I17
sS'm'
(lp25279
I1
aI3
aI1
aI3
aI11
aI27
aI33
aI167
aI311
aI1003
aI1635
aI2479
aI1831
aI6225
aI17711
aI30185
aI87043
asS'd'
I12640
sa(dp25280
S'a'
I58710
sS's'
I17
sS'm'
(lp25281
I1
aI3
aI7
aI13
aI7
aI61
aI3
aI243
aI55
aI259
aI461
aI3357
aI121
aI9107
aI19393
aI3719
aI71403
asS'd'
I12641
sa(dp25282
S'a'
I58713
sS's'
I17
sS'm'
(lp25283
I1
aI1
aI7
aI15
aI23
aI25
aI55
aI89
aI241
aI297
aI113
aI59
aI4799
aI381
aI29127
aI811
aI91149
asS'd'
I12642
sa(dp25284
S'a'
I58723
sS's'
I17
sS'm'
(lp25285
I1
aI1
aI3
aI3
aI13
aI35
aI43
aI163
aI69
aI69
aI1949
aI2383
aI887
aI2349
aI12539
aI167
aI23697
asS'd'
I12643
sa(dp25286
S'a'
I58740
sS's'
I17
sS'm'
(lp25287
I1
aI3
aI3
aI11
aI3
aI45
aI23
aI149
aI55
aI963
aI1293
aI2715
aI1401
aI16081
aI12821
aI2129
aI26835
asS'd'
I12644
sa(dp25288
S'a'
I58749
sS's'
I17
sS'm'
(lp25289
I1
aI3
aI3
aI13
aI29
aI47
aI21
aI249
aI141
aI179
aI627
aI329
aI6377
aI12049
aI11715
aI1447
aI119359
asS'd'
I12645
sa(dp25290
S'a'
I58760
sS's'
I17
sS'm'
(lp25291
I1
aI1
aI5
aI13
aI7
aI59
aI103
aI189
aI401
aI169
aI455
aI1197
aI1881
aI12679
aI4533
aI25561
aI79281
asS'd'
I12646
sa(dp25292
S'a'
I58766
sS's'
I17
sS'm'
(lp25293
I1
aI1
aI7
aI15
aI21
aI59
aI97
aI77
aI397
aI487
aI647
aI2277
aI269
aI713
aI17741
aI37387
aI100143
asS'd'
I12647
sa(dp25294
S'a'
I58773
sS's'
I17
sS'm'
(lp25295
I1
aI3
aI5
aI1
aI1
aI43
aI31
aI201
aI21
aI805
aI1533
aI1407
aI4719
aI2673
aI22757
aI7605
aI72485
asS'd'
I12648
sa(dp25296
S'a'
I58780
sS's'
I17
sS'm'
(lp25297
I1
aI1
aI3
aI11
aI27
aI25
aI99
aI135
aI225
aI367
aI1311
aI683
aI7193
aI8209
aI17081
aI55709
aI3029
asS'd'
I12649
sa(dp25298
S'a'
I58787
sS's'
I17
sS'm'
(lp25299
I1
aI3
aI5
aI5
aI17
aI9
aI59
aI27
aI153
aI353
aI647
aI2919
aI1877
aI5359
aI19787
aI46237
aI7
asS'd'
I12650
sa(dp25300
S'a'
I58796
sS's'
I17
sS'm'
(lp25301
I1
aI3
aI3
aI15
aI11
aI39
aI75
aI223
aI489
aI57
aI1355
aI3941
aI6603
aI12883
aI20909
aI24065
aI53543
asS'd'
I12651
sa(dp25302
S'a'
I58816
sS's'
I17
sS'm'
(lp25303
I1
aI1
aI3
aI9
aI17
aI47
aI61
aI49
aI397
aI23
aI2019
aI37
aI23
aI8967
aI10357
aI54419
aI27279
asS'd'
I12652
sa(dp25304
S'a'
I58822
sS's'
I17
sS'm'
(lp25305
I1
aI1
aI1
aI7
aI31
aI43
aI87
aI111
aI293
aI969
aI1431
aI2275
aI3131
aI2915
aI24595
aI63049
aI71517
asS'd'
I12653
sa(dp25306
S'a'
I58831
sS's'
I17
sS'm'
(lp25307
I1
aI1
aI3
aI11
aI11
aI63
aI89
aI33
aI95
aI299
aI593
aI3353
aI7389
aI841
aI1895
aI64835
aI19915
asS'd'
I12654
sa(dp25308
S'a'
I58839
sS's'
I17
sS'm'
(lp25309
I1
aI1
aI1
aI3
aI13
aI63
aI51
aI3
aI327
aI401
aI1289
aI2699
aI5599
aI5101
aI3189
aI23415
aI53429
asS'd'
I12655
sa(dp25310
S'a'
I58849
sS's'
I17
sS'm'
(lp25311
I1
aI1
aI7
aI13
aI27
aI15
aI57
aI201
aI365
aI643
aI171
aI2417
aI1763
aI7567
aI5323
aI50911
aI24281
asS'd'
I12656
sa(dp25312
S'a'
I58859
sS's'
I17
sS'm'
(lp25313
I1
aI3
aI3
aI13
aI31
aI37
aI31
aI243
aI225
aI431
aI379
aI1565
aI1567
aI11477
aI4641
aI58823
aI88565
asS'd'
I12657
sa(dp25314
S'a'
I58870
sS's'
I17
sS'm'
(lp25315
I1
aI3
aI3
aI7
aI13
aI39
aI79
aI169
aI401
aI177
aI2023
aI1703
aI5563
aI15619
aI21445
aI59287
aI30141
asS'd'
I12658
sa(dp25316
S'a'
I58879
sS's'
I17
sS'm'
(lp25317
I1
aI3
aI3
aI15
aI17
aI47
aI13
aI163
aI361
aI353
aI1435
aI2083
aI4109
aI14105
aI28585
aI1721
aI119133
asS'd'
I12659
sa(dp25318
S'a'
I58889
sS's'
I17
sS'm'
(lp25319
I1
aI3
aI3
aI3
aI29
aI41
aI97
aI131
aI501
aI415
aI1735
aI2315
aI6499
aI11417
aI3879
aI3957
aI47117
asS'd'
I12660
sa(dp25320
S'a'
I58904
sS's'
I17
sS'm'
(lp25321
I1
aI1
aI3
aI13
aI5
aI25
aI29
aI65
aI235
aI949
aI1571
aI927
aI515
aI7519
aI23123
aI4127
aI71019
asS'd'
I12661
sa(dp25322
S'a'
I58914
sS's'
I17
sS'm'
(lp25323
I1
aI1
aI5
aI13
aI1
aI41
aI5
aI137
aI487
aI711
aI561
aI2495
aI5367
aI6955
aI9453
aI54499
aI118373
asS'd'
I12662
sa(dp25324
S'a'
I58923
sS's'
I17
sS'm'
(lp25325
I1
aI3
aI3
aI5
aI27
aI15
aI23
aI237
aI95
aI873
aI1949
aI1579
aI7089
aI8837
aI30463
aI9903
aI61919
asS'd'
I12663
sa(dp25326
S'a'
I58940
sS's'
I17
sS'm'
(lp25327
I1
aI3
aI1
aI13
aI5
aI5
aI15
aI5
aI429
aI617
aI383
aI2495
aI409
aI15541
aI11209
aI20625
aI58493
asS'd'
I12664
sa(dp25328
S'a'
I58945
sS's'
I17
sS'm'
(lp25329
I1
aI1
aI7
aI11
aI25
aI21
aI113
aI219
aI459
aI777
aI289
aI2435
aI6665
aI83
aI22997
aI6561
aI82923
asS'd'
I12665
sa(dp25330
S'a'
I58952
sS's'
I17
sS'm'
(lp25331
I1
aI1
aI1
aI7
aI19
aI53
aI37
aI249
aI273
aI639
aI2007
aI1463
aI7819
aI9013
aI19539
aI41235
aI58059
asS'd'
I12666
sa(dp25332
S'a'
I58960
sS's'
I17
sS'm'
(lp25333
I1
aI1
aI3
aI11
aI13
aI51
aI97
aI55
aI237
aI159
aI1305
aI3705
aI2527
aI13065
aI13873
aI10275
aI3769
asS'd'
I12667
sa(dp25334
S'a'
I58965
sS's'
I17
sS'm'
(lp25335
I1
aI1
aI5
aI7
aI13
aI49
aI41
aI13
aI301
aI461
aI1169
aI3805
aI7163
aI4133
aI17255
aI48283
aI87059
asS'd'
I12668
sa(dp25336
S'a'
I58986
sS's'
I17
sS'm'
(lp25337
I1
aI3
aI7
aI15
aI9
aI19
aI45
aI195
aI445
aI145
aI543
aI2839
aI315
aI10991
aI3317
aI46501
aI29209
asS'd'
I12669
sa(dp25338
S'a'
I59000
sS's'
I17
sS'm'
(lp25339
I1
aI1
aI3
aI5
aI23
aI51
aI53
aI119
aI217
aI747
aI1307
aI3039
aI5523
aI5891
aI21035
aI56471
aI80305
asS'd'
I12670
sa(dp25340
S'a'
I59006
sS's'
I17
sS'm'
(lp25341
I1
aI3
aI1
aI3
aI23
aI17
aI97
aI71
aI7
aI839
aI229
aI3407
aI3025
aI1989
aI16599
aI38755
aI11139
asS'd'
I12671
sa(dp25342
S'a'
I59027
sS's'
I17
sS'm'
(lp25343
I1
aI1
aI7
aI7
aI7
aI19
aI67
aI43
aI293
aI191
aI1549
aI1621
aI7083
aI11633
aI1899
aI4515
aI89753
asS'd'
I12672
sa(dp25344
S'a'
I59034
sS's'
I17
sS'm'
(lp25345
I1
aI3
aI7
aI11
aI1
aI23
aI17
aI63
aI265
aI709
aI453
aI1309
aI6861
aI7257
aI17705
aI28565
aI15231
asS'd'
I12673
sa(dp25346
S'a'
I59052
sS's'
I17
sS'm'
(lp25347
I1
aI3
aI5
aI13
aI1
aI53
aI99
aI79
aI241
aI771
aI497
aI987
aI2353
aI6931
aI227
aI57781
aI109583
asS'd'
I12674
sa(dp25348
S'a'
I59058
sS's'
I17
sS'm'
(lp25349
I1
aI3
aI3
aI5
aI31
aI47
aI17
aI97
aI93
aI119
aI2045
aI3597
aI7983
aI4993
aI30659
aI3419
aI32803
asS'd'
I12675
sa(dp25350
S'a'
I59081
sS's'
I17
sS'm'
(lp25351
I1
aI3
aI7
aI1
aI11
aI41
aI3
aI7
aI155
aI627
aI979
aI2405
aI7467
aI6763
aI28523
aI56493
aI1291
asS'd'
I12676
sa(dp25352
S'a'
I59082
sS's'
I17
sS'm'
(lp25353
I1
aI1
aI1
aI7
aI3
aI43
aI13
aI187
aI415
aI1001
aI667
aI1371
aI8075
aI12855
aI27215
aI9399
aI70657
asS'd'
I12677
sa(dp25354
S'a'
I59089
sS's'
I17
sS'm'
(lp25355
I1
aI3
aI1
aI9
aI29
aI53
aI117
aI17
aI443
aI381
aI507
aI643
aI5891
aI10725
aI20251
aI497
aI101313
asS'd'
I12678
sa(dp25356
S'a'
I59096
sS's'
I17
sS'm'
(lp25357
I1
aI3
aI3
aI1
aI31
aI47
aI59
aI75
aI173
aI347
aI1
aI117
aI3639
aI10231
aI863
aI56795
aI56107
asS'd'
I12679
sa(dp25358
S'a'
I59118
sS's'
I17
sS'm'
(lp25359
I1
aI3
aI1
aI15
aI11
aI41
aI97
aI251
aI177
aI275
aI1849
aI1281
aI3659
aI9709
aI16239
aI7445
aI47661
asS'd'
I12680
sa(dp25360
S'a'
I59123
sS's'
I17
sS'm'
(lp25361
I1
aI3
aI3
aI5
aI13
aI1
aI27
aI251
aI89
aI863
aI685
aI3915
aI2201
aI4313
aI32083
aI37171
aI120885
asS'd'
I12681
sa(dp25362
S'a'
I59126
sS's'
I17
sS'm'
(lp25363
I1
aI3
aI3
aI3
aI3
aI11
aI77
aI135
aI147
aI475
aI505
aI611
aI4763
aI9445
aI28639
aI51343
aI119093
asS'd'
I12682
sa(dp25364
S'a'
I59132
sS's'
I17
sS'm'
(lp25365
I1
aI3
aI7
aI1
aI11
aI9
aI67
aI73
aI375
aI977
aI1011
aI1621
aI6623
aI8077
aI26321
aI7461
aI130637
asS'd'
I12683
sa(dp25366
S'a'
I59135
sS's'
I17
sS'm'
(lp25367
I1
aI3
aI5
aI15
aI15
aI59
aI63
aI217
aI117
aI375
aI1151
aI3451
aI5843
aI13221
aI20673
aI10817
aI57711
asS'd'
I12684
sa(dp25368
S'a'
I59150
sS's'
I17
sS'm'
(lp25369
I1
aI1
aI3
aI5
aI19
aI5
aI9
aI101
aI321
aI815
aI451
aI719
aI523
aI1299
aI27823
aI48041
aI45939
asS'd'
I12685
sa(dp25370
S'a'
I59174
sS's'
I17
sS'm'
(lp25371
I1
aI1
aI1
aI9
aI5
aI49
aI59
aI59
aI409
aI5
aI1075
aI21
aI7837
aI4543
aI3085
aI55473
aI89309
asS'd'
I12686
sa(dp25372
S'a'
I59177
sS's'
I17
sS'm'
(lp25373
I1
aI3
aI3
aI13
aI9
aI31
aI93
aI181
aI153
aI795
aI713
aI933
aI6001
aI9075
aI29781
aI14029
aI46361
asS'd'
I12687
sa(dp25374
S'a'
I59185
sS's'
I17
sS'm'
(lp25375
I1
aI3
aI5
aI3
aI9
aI21
aI79
aI129
aI361
aI627
aI743
aI3041
aI5381
aI3627
aI12581
aI35183
aI83183
asS'd'
I12688
sa(dp25376
S'a'
I59195
sS's'
I17
sS'm'
(lp25377
I1
aI3
aI3
aI3
aI29
aI29
aI93
aI187
aI341
aI727
aI181
aI133
aI6525
aI5673
aI2739
aI23349
aI62505
asS'd'
I12689
sa(dp25378
S'a'
I59203
sS's'
I17
sS'm'
(lp25379
I1
aI3
aI5
aI13
aI27
aI59
aI127
aI149
aI321
aI613
aI1543
aI1537
aI2909
aI5139
aI3755
aI49285
aI35183
asS'd'
I12690
sa(dp25380
S'a'
I59206
sS's'
I17
sS'm'
(lp25381
I1
aI3
aI3
aI9
aI5
aI31
aI19
aI173
aI137
aI19
aI799
aI1847
aI3897
aI15775
aI22411
aI55405
aI95713
asS'd'
I12691
sa(dp25382
S'a'
I59234
sS's'
I17
sS'm'
(lp25383
I1
aI3
aI5
aI5
aI31
aI1
aI87
aI97
aI481
aI797
aI459
aI793
aI4339
aI5443
aI31717
aI6691
aI68841
asS'd'
I12692
sa(dp25384
S'a'
I59248
sS's'
I17
sS'm'
(lp25385
I1
aI3
aI3
aI5
aI3
aI9
aI123
aI21
aI393
aI683
aI1007
aI971
aI729
aI7113
aI2811
aI51183
aI37093
asS'd'
I12693
sa(dp25386
S'a'
I59258
sS's'
I17
sS'm'
(lp25387
I1
aI3
aI5
aI5
aI27
aI31
aI19
aI227
aI123
aI937
aI763
aI2117
aI3825
aI9151
aI95
aI54963
aI62919
asS'd'
I12694
sa(dp25388
S'a'
I59260
sS's'
I17
sS'm'
(lp25389
I1
aI3
aI1
aI3
aI23
aI21
aI21
aI7
aI7
aI165
aI99
aI3073
aI91
aI8725
aI17613
aI47119
aI41441
asS'd'
I12695
sa(dp25390
S'a'
I59270
sS's'
I17
sS'm'
(lp25391
I1
aI3
aI7
aI9
aI29
aI49
aI117
aI49
aI205
aI605
aI1567
aI3323
aI1559
aI10949
aI14349
aI34951
aI99901
asS'd'
I12696
sa(dp25392
S'a'
I59284
sS's'
I17
sS'm'
(lp25393
I1
aI1
aI1
aI1
aI15
aI41
aI31
aI133
aI439
aI305
aI719
aI147
aI6849
aI5947
aI31749
aI45171
aI15265
asS'd'
I12697
sa(dp25394
S'a'
I59291
sS's'
I17
sS'm'
(lp25395
I1
aI1
aI5
aI11
aI19
aI37
aI13
aI253
aI385
aI625
aI1801
aI1191
aI547
aI12025
aI17971
aI54127
aI97323
asS'd'
I12698
sa(dp25396
S'a'
I59293
sS's'
I17
sS'm'
(lp25397
I1
aI3
aI5
aI5
aI27
aI9
aI57
aI245
aI59
aI211
aI1195
aI763
aI6743
aI6309
aI25759
aI26633
aI27497
asS'd'
I12699
sa(dp25398
S'a'
I59297
sS's'
I17
sS'm'
(lp25399
I1
aI3
aI5
aI5
aI9
aI51
aI49
aI29
aI213
aI855
aI305
aI3961
aI3187
aI6815
aI19015
aI2539
aI81705
asS'd'
I12700
sa(dp25400
S'a'
I59312
sS's'
I17
sS'm'
(lp25401
I1
aI1
aI5
aI7
aI11
aI61
aI123
aI163
aI127
aI871
aI463
aI2279
aI5931
aI2913
aI23215
aI40215
aI91373
asS'd'
I12701
sa(dp25402
S'a'
I59315
sS's'
I17
sS'm'
(lp25403
I1
aI1
aI7
aI1
aI29
aI57
aI55
aI189
aI285
aI603
aI747
aI3753
aI3359
aI10277
aI25319
aI17569
aI80125
asS'd'
I12702
sa(dp25404
S'a'
I59327
sS's'
I17
sS'm'
(lp25405
I1
aI3
aI1
aI9
aI7
aI53
aI9
aI55
aI117
aI495
aI2045
aI2487
aI1625
aI775
aI17773
aI62159
aI79537
asS'd'
I12703
sa(dp25406
S'a'
I59332
sS's'
I17
sS'm'
(lp25407
I1
aI1
aI5
aI1
aI1
aI11
aI67
aI79
aI57
aI677
aI2045
aI3913
aI853
aI3581
aI10509
aI35097
aI24585
asS'd'
I12704
sa(dp25408
S'a'
I59347
sS's'
I17
sS'm'
(lp25409
I1
aI3
aI7
aI7
aI19
aI15
aI13
aI237
aI297
aI807
aI657
aI2229
aI2931
aI10283
aI49
aI56485
aI113889
asS'd'
I12705
sa(dp25410
S'a'
I59354
sS's'
I17
sS'm'
(lp25411
I1
aI3
aI5
aI13
aI11
aI23
aI37
aI229
aI253
aI411
aI39
aI1069
aI3683
aI12141
aI11087
aI64875
aI62991
asS'd'
I12706
sa(dp25412
S'a'
I59363
sS's'
I17
sS'm'
(lp25413
I1
aI1
aI1
aI1
aI13
aI11
aI21
aI227
aI305
aI1021
aI1039
aI3095
aI4621
aI5723
aI31989
aI30681
aI58487
asS'd'
I12707
sa(dp25414
S'a'
I59369
sS's'
I17
sS'm'
(lp25415
I1
aI3
aI3
aI5
aI17
aI47
aI127
aI65
aI85
aI961
aI277
aI549
aI2111
aI4183
aI771
aI11921
aI111489
asS'd'
I12708
sa(dp25416
S'a'
I59377
sS's'
I17
sS'm'
(lp25417
I1
aI3
aI3
aI5
aI31
aI39
aI89
aI51
aI277
aI705
aI375
aI1603
aI2793
aI7425
aI25065
aI44945
aI48391
asS'd'
I12709
sa(dp25418
S'a'
I59406
sS's'
I17
sS'm'
(lp25419
I1
aI1
aI5
aI7
aI9
aI45
aI21
aI217
aI183
aI65
aI625
aI1239
aI4241
aI8043
aI13615
aI5611
aI82501
asS'd'
I12710
sa(dp25420
S'a'
I59413
sS's'
I17
sS'm'
(lp25421
I1
aI1
aI7
aI13
aI15
aI51
aI29
aI35
aI73
aI601
aI233
aI3117
aI3031
aI4229
aI4299
aI62761
aI45291
asS'd'
I12711
sa(dp25422
S'a'
I59423
sS's'
I17
sS'm'
(lp25423
I1
aI3
aI5
aI5
aI27
aI55
aI7
aI1
aI333
aI501
aI913
aI1939
aI5131
aI5597
aI1271
aI30195
aI20947
asS'd'
I12712
sa(dp25424
S'a'
I59434
sS's'
I17
sS'm'
(lp25425
I1
aI3
aI5
aI15
aI19
aI27
aI43
aI111
aI435
aI615
aI811
aI2113
aI2503
aI7553
aI1619
aI24773
aI40881
asS'd'
I12713
sa(dp25426
S'a'
I59441
sS's'
I17
sS'm'
(lp25427
I1
aI3
aI5
aI11
aI11
aI31
aI21
aI147
aI151
aI141
aI527
aI1839
aI339
aI1225
aI29621
aI19691
aI22031
asS'd'
I12714
sa(dp25428
S'a'
I59442
sS's'
I17
sS'm'
(lp25429
I1
aI1
aI3
aI13
aI23
aI35
aI61
aI181
aI221
aI837
aI241
aI3033
aI1849
aI12563
aI11387
aI3027
aI33419
asS'd'
I12715
sa(dp25430
S'a'
I59456
sS's'
I17
sS'm'
(lp25431
I1
aI3
aI3
aI7
aI1
aI39
aI125
aI249
aI231
aI575
aI1847
aI197
aI3969
aI12945
aI10257
aI27227
aI94097
asS'd'
I12716
sa(dp25432
S'a'
I59465
sS's'
I17
sS'm'
(lp25433
I1
aI3
aI1
aI11
aI19
aI7
aI85
aI119
aI37
aI253
aI1575
aI447
aI6947
aI14399
aI32095
aI15385
aI62917
asS'd'
I12717
sa(dp25434
S'a'
I59479
sS's'
I17
sS'm'
(lp25435
I1
aI3
aI7
aI11
aI23
aI53
aI113
aI193
aI253
aI651
aI1717
aI447
aI4055
aI13675
aI18479
aI28375
aI66475
asS'd'
I12718
sa(dp25436
S'a'
I59501
sS's'
I17
sS'm'
(lp25437
I1
aI1
aI7
aI3
aI17
aI27
aI13
aI209
aI67
aI271
aI121
aI1565
aI2589
aI8033
aI3939
aI14181
aI111787
asS'd'
I12719
sa(dp25438
S'a'
I59504
sS's'
I17
sS'm'
(lp25439
I1
aI1
aI3
aI3
aI5
aI57
aI75
aI87
aI165
aI947
aI967
aI3353
aI2055
aI16195
aI3701
aI62637
aI58343
asS'd'
I12720
sa(dp25440
S'a'
I59507
sS's'
I17
sS'm'
(lp25441
I1
aI1
aI7
aI13
aI31
aI41
aI125
aI111
aI253
aI257
aI57
aI679
aI3333
aI8259
aI26331
aI15883
aI95023
asS'd'
I12721
sa(dp25442
S'a'
I59514
sS's'
I17
sS'm'
(lp25443
I1
aI3
aI7
aI11
aI1
aI47
aI5
aI249
aI353
aI271
aI1121
aI1935
aI4971
aI8447
aI9983
aI55959
aI66179
asS'd'
I12722
sa(dp25444
S'a'
I59523
sS's'
I17
sS'm'
(lp25445
I1
aI3
aI5
aI11
aI1
aI61
aI101
aI43
aI97
aI241
aI687
aI4027
aI4319
aI4905
aI12357
aI51099
aI97093
asS'd'
I12723
sa(dp25446
S'a'
I59525
sS's'
I17
sS'm'
(lp25447
I1
aI1
aI3
aI15
aI25
aI25
aI117
aI139
aI475
aI961
aI1585
aI2795
aI2411
aI11049
aI18209
aI15511
aI43193
asS'd'
I12724
sa(dp25448
S'a'
I59526
sS's'
I17
sS'm'
(lp25449
I1
aI3
aI3
aI9
aI9
aI11
aI107
aI15
aI189
aI27
aI289
aI3111
aI851
aI3401
aI31981
aI7181
aI47533
asS'd'
I12725
sa(dp25450
S'a'
I59537
sS's'
I17
sS'm'
(lp25451
I1
aI1
aI5
aI15
aI31
aI57
aI37
aI77
aI205
aI707
aI1505
aI1343
aI629
aI13335
aI6719
aI35405
aI38905
asS'd'
I12726
sa(dp25452
S'a'
I59543
sS's'
I17
sS'm'
(lp25453
I1
aI3
aI7
aI7
aI21
aI63
aI35
aI125
aI507
aI285
aI621
aI2257
aI3009
aI13703
aI9761
aI54927
aI16925
asS'd'
I12727
sa(dp25454
S'a'
I59560
sS's'
I17
sS'm'
(lp25455
I1
aI1
aI1
aI11
aI11
aI53
aI37
aI17
aI167
aI663
aI1349
aI1395
aI7721
aI9329
aI2161
aI37093
aI52425
asS'd'
I12728
sa(dp25456
S'a'
I59563
sS's'
I17
sS'm'
(lp25457
I1
aI1
aI7
aI9
aI1
aI15
aI113
aI53
aI87
aI133
aI131
aI847
aI609
aI14737
aI1639
aI15511
aI46987
asS'd'
I12729
sa(dp25458
S'a'
I59566
sS's'
I17
sS'm'
(lp25459
I1
aI1
aI7
aI7
aI25
aI21
aI125
aI149
aI201
aI791
aI1323
aI2817
aI1141
aI289
aI14349
aI64461
aI76575
asS'd'
I12730
sa(dp25460
S'a'
I59577
sS's'
I17
sS'm'
(lp25461
I1
aI3
aI3
aI7
aI15
aI1
aI123
aI227
aI83
aI967
aI2039
aI3205
aI715
aI10787
aI11235
aI50375
aI66911
asS'd'
I12731
sa(dp25462
S'a'
I59580
sS's'
I17
sS'm'
(lp25463
I1
aI3
aI3
aI11
aI9
aI45
aI41
aI57
aI327
aI903
aI1705
aI3947
aI3173
aI1035
aI25529
aI15217
aI80795
asS'd'
I12732
sa(dp25464
S'a'
I59586
sS's'
I17
sS'm'
(lp25465
I1
aI3
aI1
aI13
aI15
aI63
aI69
aI57
aI479
aI277
aI1641
aI3465
aI2629
aI9901
aI5781
aI55101
aI33049
asS'd'
I12733
sa(dp25466
S'a'
I59597
sS's'
I17
sS'm'
(lp25467
I1
aI1
aI3
aI3
aI31
aI25
aI57
aI189
aI491
aI335
aI373
aI3069
aI1623
aI10781
aI6559
aI27057
aI111491
asS'd'
I12734
sa(dp25468
S'a'
I59612
sS's'
I17
sS'm'
(lp25469
I1
aI3
aI3
aI15
aI19
aI57
aI67
aI225
aI343
aI93
aI315
aI1011
aI4437
aI10371
aI27927
aI51269
aI65653
asS'd'
I12735
sa(dp25470
S'a'
I59615
sS's'
I17
sS'm'
(lp25471
I1
aI3
aI5
aI5
aI21
aI43
aI105
aI153
aI65
aI167
aI369
aI3167
aI785
aI7509
aI3753
aI9035
aI29039
asS'd'
I12736
sa(dp25472
S'a'
I59622
sS's'
I17
sS'm'
(lp25473
I1
aI1
aI7
aI3
aI17
aI9
aI113
aI21
aI175
aI967
aI13
aI2701
aI5667
aI9761
aI20267
aI33497
aI88819
asS'd'
I12737
sa(dp25474
S'a'
I59631
sS's'
I17
sS'm'
(lp25475
I1
aI1
aI1
aI7
aI9
aI61
aI121
aI205
aI283
aI259
aI2027
aI2361
aI3995
aI6787
aI26867
aI61681
aI96149
asS'd'
I12738
sa(dp25476
S'a'
I59657
sS's'
I17
sS'm'
(lp25477
I1
aI3
aI3
aI13
aI31
aI37
aI125
aI151
aI317
aI947
aI423
aI2907
aI6905
aI13247
aI27997
aI4755
aI73173
asS'd'
I12739
sa(dp25478
S'a'
I59660
sS's'
I17
sS'm'
(lp25479
I1
aI1
aI7
aI13
aI1
aI63
aI85
aI75
aI33
aI483
aI85
aI1583
aI4783
aI14003
aI31147
aI12643
aI99447
asS'd'
I12740
sa(dp25480
S'a'
I59665
sS's'
I17
sS'm'
(lp25481
I1
aI1
aI1
aI7
aI31
aI5
aI93
aI179
aI213
aI857
aI3
aI1015
aI1481
aI2413
aI28759
aI13941
aI24851
asS'd'
I12741
sa(dp25482
S'a'
I59671
sS's'
I17
sS'm'
(lp25483
I1
aI1
aI5
aI5
aI5
aI35
aI73
aI57
aI469
aI885
aI1951
aI3599
aI5555
aI7259
aI22599
aI7627
aI109227
asS'd'
I12742
sa(dp25484
S'a'
I59675
sS's'
I17
sS'm'
(lp25485
I1
aI1
aI5
aI13
aI1
aI1
aI69
aI107
aI473
aI793
aI1851
aI887
aI2241
aI7851
aI21821
aI25431
aI118565
asS'd'
I12743
sa(dp25486
S'a'
I59688
sS's'
I17
sS'm'
(lp25487
I1
aI1
aI1
aI3
aI15
aI51
aI57
aI175
aI91
aI113
aI1671
aI925
aI2187
aI1097
aI28793
aI45819
aI41855
asS'd'
I12744
sa(dp25488
S'a'
I59706
sS's'
I17
sS'm'
(lp25489
I1
aI1
aI3
aI3
aI15
aI51
aI99
aI177
aI379
aI713
aI1273
aI3245
aI5515
aI14657
aI28981
aI1197
aI29283
asS'd'
I12745
sa(dp25490
S'a'
I59708
sS's'
I17
sS'm'
(lp25491
I1
aI3
aI3
aI15
aI13
aI27
aI29
aI7
aI17
aI373
aI779
aI1589
aI3077
aI13673
aI31029
aI43765
aI74339
asS'd'
I12746
sa(dp25492
S'a'
I59713
sS's'
I17
sS'm'
(lp25493
I1
aI1
aI3
aI1
aI25
aI17
aI15
aI223
aI219
aI569
aI155
aI1307
aI7143
aI7975
aI22581
aI53223
aI44271
asS'd'
I12747
sa(dp25494
S'a'
I59714
sS's'
I17
sS'm'
(lp25495
I1
aI1
aI1
aI7
aI11
aI23
aI67
aI177
aI189
aI267
aI1799
aI2453
aI2367
aI4193
aI1827
aI10191
aI12265
asS'd'
I12748
sa(dp25496
S'a'
I59720
sS's'
I17
sS'm'
(lp25497
I1
aI3
aI3
aI11
aI3
aI29
aI121
aI231
aI211
aI191
aI1803
aI1171
aI6801
aI4007
aI14111
aI42951
aI8311
asS'd'
I12749
sa(dp25498
S'a'
I59737
sS's'
I17
sS'm'
(lp25499
I1
aI3
aI3
aI11
aI9
aI23
aI11
aI231
aI349
aI47
aI1645
aI345
aI3681
aI12227
aI29955
aI32131
aI391
asS'd'
I12750
sa(dp25500
S'a'
I59743
sS's'
I17
sS'm'
(lp25501
I1
aI1
aI3
aI11
aI27
aI17
aI17
aI49
aI463
aI173
aI1993
aI2339
aI3763
aI1931
aI29969
aI55579
aI114805
asS'd'
I12751
sa(dp25502
S'a'
I59744
sS's'
I17
sS'm'
(lp25503
I1
aI3
aI5
aI15
aI5
aI21
aI105
aI69
aI173
aI771
aI815
aI3807
aI577
aI12589
aI32193
aI37601
aI23961
asS'd'
I12752
sa(dp25504
S'a'
I59749
sS's'
I17
sS'm'
(lp25505
I1
aI1
aI3
aI9
aI21
aI63
aI57
aI213
aI327
aI765
aI333
aI2065
aI719
aI6159
aI15769
aI49335
aI2289
asS'd'
I12753
sa(dp25506
S'a'
I59756
sS's'
I17
sS'm'
(lp25507
I1
aI3
aI3
aI1
aI27
aI51
aI13
aI87
aI465
aI729
aI507
aI2811
aI6721
aI14279
aI31733
aI56165
aI78169
asS'd'
I12754
sa(dp25508
S'a'
I59783
sS's'
I17
sS'm'
(lp25509
I1
aI1
aI1
aI9
aI25
aI1
aI47
aI37
aI407
aI623
aI1409
aI3703
aI4491
aI3037
aI8129
aI13547
aI50517
asS'd'
I12755
sa(dp25510
S'a'
I59784
sS's'
I17
sS'm'
(lp25511
I1
aI3
aI5
aI13
aI5
aI51
aI127
aI225
aI215
aI377
aI1013
aI2387
aI4237
aI1317
aI5245
aI17535
aI100707
asS'd'
I12756
sa(dp25512
S'a'
I59787
sS's'
I17
sS'm'
(lp25513
I1
aI3
aI5
aI13
aI31
aI25
aI123
aI243
aI317
aI505
aI483
aI1743
aI3097
aI4139
aI4525
aI63143
aI47665
asS'd'
I12757
sa(dp25514
S'a'
I59795
sS's'
I17
sS'm'
(lp25515
I1
aI3
aI3
aI13
aI27
aI19
aI21
aI187
aI211
aI471
aI1931
aI2877
aI3635
aI9233
aI12385
aI15741
aI50843
asS'd'
I12758
sa(dp25516
S'a'
I59804
sS's'
I17
sS'm'
(lp25517
I1
aI1
aI3
aI13
aI23
aI33
aI71
aI45
aI371
aI621
aI1057
aI1605
aI6329
aI3763
aI8613
aI2965
aI8141
asS'd'
I12759
sa(dp25518
S'a'
I59811
sS's'
I17
sS'm'
(lp25519
I1
aI1
aI5
aI7
aI7
aI17
aI33
aI209
aI293
aI35
aI1665
aI3721
aI6245
aI14567
aI15349
aI16195
aI59757
asS'd'
I12760
sa(dp25520
S'a'
I59814
sS's'
I17
sS'm'
(lp25521
I1
aI1
aI7
aI7
aI31
aI19
aI47
aI83
aI277
aI323
aI1225
aI969
aI2193
aI10175
aI27657
aI50265
aI9817
asS'd'
I12761
sa(dp25522
S'a'
I59823
sS's'
I17
sS'm'
(lp25523
I1
aI1
aI5
aI7
aI15
aI21
aI103
aI95
aI189
aI737
aI309
aI357
aI5953
aI9701
aI15757
aI20753
aI88647
asS'd'
I12762
sa(dp25524
S'a'
I59838
sS's'
I17
sS'm'
(lp25525
I1
aI1
aI5
aI13
aI5
aI13
aI61
aI235
aI333
aI889
aI1559
aI1653
aI1871
aI10631
aI18067
aI47037
aI9507
asS'd'
I12763
sa(dp25526
S'a'
I59840
sS's'
I17
sS'm'
(lp25527
I1
aI3
aI7
aI1
aI31
aI61
aI69
aI159
aI41
aI107
aI807
aI1517
aI3551
aI3435
aI6151
aI50025
aI126949
asS'd'
I12764
sa(dp25528
S'a'
I59846
sS's'
I17
sS'm'
(lp25529
I1
aI1
aI7
aI15
aI7
aI41
aI55
aI103
aI9
aI105
aI1397
aI3955
aI7723
aI3389
aI32435
aI36005
aI73733
asS'd'
I12765
sa(dp25530
S'a'
I59849
sS's'
I17
sS'm'
(lp25531
I1
aI1
aI1
aI13
aI15
aI59
aI43
aI151
aI321
aI215
aI411
aI3103
aI7455
aI14041
aI1673
aI56425
aI59085
asS'd'
I12766
sa(dp25532
S'a'
I59855
sS's'
I17
sS'm'
(lp25533
I1
aI1
aI1
aI1
aI5
aI39
aI81
aI183
aI509
aI455
aI753
aI3743
aI227
aI7807
aI23747
aI42289
aI122765
asS'd'
I12767
sa(dp25534
S'a'
I59860
sS's'
I17
sS'm'
(lp25535
I1
aI1
aI7
aI3
aI9
aI29
aI89
aI131
aI141
aI851
aI1221
aI67
aI3117
aI1329
aI13151
aI36827
aI34313
asS'd'
I12768
sa(dp25536
S'a'
I59867
sS's'
I17
sS'm'
(lp25537
I1
aI3
aI1
aI5
aI9
aI29
aI11
aI189
aI389
aI79
aI1903
aI835
aI6043
aI8953
aI18985
aI8305
aI51939
asS'd'
I12769
sa(dp25538
S'a'
I59869
sS's'
I17
sS'm'
(lp25539
I1
aI3
aI5
aI3
aI9
aI49
aI47
aI43
aI193
aI917
aI795
aI2719
aI5709
aI9993
aI30637
aI26841
aI93113
asS'd'
I12770
sa(dp25540
S'a'
I59885
sS's'
I17
sS'm'
(lp25541
I1
aI3
aI7
aI5
aI13
aI15
aI85
aI169
aI315
aI963
aI617
aI1191
aI6739
aI11535
aI3423
aI6695
aI2047
asS'd'
I12771
sa(dp25542
S'a'
I59888
sS's'
I17
sS'm'
(lp25543
I1
aI3
aI1
aI3
aI5
aI49
aI41
aI255
aI131
aI255
aI825
aI1485
aI7005
aI10107
aI3913
aI4731
aI33199
asS'd'
I12772
sa(dp25544
S'a'
I59893
sS's'
I17
sS'm'
(lp25545
I1
aI3
aI1
aI9
aI27
aI63
aI109
aI29
aI183
aI381
aI591
aI617
aI1187
aI4381
aI30543
aI9933
aI109785
asS'd'
I12773
sa(dp25546
S'a'
I59904
sS's'
I17
sS'm'
(lp25547
I1
aI3
aI7
aI7
aI13
aI29
aI125
aI105
aI353
aI677
aI623
aI1553
aI5435
aI10853
aI16663
aI42277
aI64333
asS'd'
I12774
sa(dp25548
S'a'
I59919
sS's'
I17
sS'm'
(lp25549
I1
aI3
aI3
aI11
aI3
aI47
aI49
aI41
aI249
aI497
aI963
aI237
aI625
aI11303
aI6871
aI60441
aI39559
asS'd'
I12775
sa(dp25550
S'a'
I59922
sS's'
I17
sS'm'
(lp25551
I1
aI3
aI3
aI1
aI1
aI27
aI107
aI253
aI59
aI445
aI1761
aI2865
aI7117
aI2363
aI4007
aI46047
aI10811
asS'd'
I12776
sa(dp25552
S'a'
I59927
sS's'
I17
sS'm'
(lp25553
I1
aI3
aI5
aI5
aI25
aI13
aI17
aI107
aI321
aI691
aI351
aI577
aI5001
aI9437
aI12451
aI44997
aI42727
asS'd'
I12777
sa(dp25554
S'a'
I59937
sS's'
I17
sS'm'
(lp25555
I1
aI3
aI1
aI9
aI1
aI31
aI87
aI117
aI111
aI379
aI1989
aI1155
aI4777
aI8563
aI14585
aI33375
aI66985
asS'd'
I12778
sa(dp25556
S'a'
I59940
sS's'
I17
sS'm'
(lp25557
I1
aI3
aI5
aI5
aI9
aI51
aI79
aI135
aI89
aI929
aI277
aI763
aI5569
aI15545
aI28393
aI56921
aI102093
asS'd'
I12779
sa(dp25558
S'a'
I59944
sS's'
I17
sS'm'
(lp25559
I1
aI1
aI3
aI7
aI7
aI55
aI15
aI37
aI3
aI439
aI577
aI2051
aI101
aI13655
aI11959
aI38127
aI6639
asS'd'
I12780
sa(dp25560
S'a'
I59947
sS's'
I17
sS'm'
(lp25561
I1
aI1
aI3
aI9
aI21
aI11
aI27
aI221
aI465
aI565
aI1313
aI1405
aI2421
aI10543
aI18369
aI33751
aI87785
asS'd'
I12781
sa(dp25562
S'a'
I59949
sS's'
I17
sS'm'
(lp25563
I1
aI3
aI5
aI11
aI21
aI39
aI105
aI231
aI469
aI711
aI997
aI427
aI5797
aI3249
aI15141
aI29413
aI66509
asS'd'
I12782
sa(dp25564
S'a'
I59961
sS's'
I17
sS'm'
(lp25565
I1
aI3
aI7
aI5
aI25
aI31
aI111
aI37
aI229
aI773
aI193
aI553
aI673
aI4693
aI24441
aI8713
aI121203
asS'd'
I12783
sa(dp25566
S'a'
I59975
sS's'
I17
sS'm'
(lp25567
I1
aI3
aI1
aI9
aI23
aI27
aI103
aI7
aI183
aI549
aI1433
aI2831
aI3383
aI13229
aI10005
aI14135
aI15967
asS'd'
I12784
sa(dp25568
S'a'
I59981
sS's'
I17
sS'm'
(lp25569
I1
aI3
aI3
aI11
aI11
aI59
aI11
aI251
aI373
aI399
aI255
aI1177
aI5493
aI13559
aI29037
aI23405
aI79495
asS'd'
I12785
sa(dp25570
S'a'
I59994
sS's'
I17
sS'm'
(lp25571
I1
aI3
aI1
aI11
aI25
aI45
aI69
aI115
aI153
aI259
aI527
aI9
aI5807
aI6015
aI3765
aI61621
aI8645
asS'd'
I12786
sa(dp25572
S'a'
I59999
sS's'
I17
sS'm'
(lp25573
I1
aI3
aI3
aI9
aI19
aI5
aI113
aI191
aI345
aI655
aI429
aI3975
aI4297
aI5723
aI5345
aI64457
aI79031
asS'd'
I12787
sa(dp25574
S'a'
I60018
sS's'
I17
sS'm'
(lp25575
I1
aI3
aI7
aI7
aI15
aI59
aI5
aI87
aI289
aI1005
aI931
aI2403
aI485
aI16043
aI15623
aI39253
aI61377
asS'd'
I12788
sa(dp25576
S'a'
I60023
sS's'
I17
sS'm'
(lp25577
I1
aI3
aI3
aI9
aI17
aI53
aI31
aI9
aI217
aI711
aI1007
aI1375
aI2733
aI13735
aI19825
aI59741
aI83827
asS'd'
I12789
sa(dp25578
S'a'
I60024
sS's'
I17
sS'm'
(lp25579
I1
aI1
aI3
aI15
aI11
aI11
aI41
aI195
aI79
aI1013
aI1353
aI1961
aI7365
aI7533
aI13315
aI8441
aI90705
asS'd'
I12790
sa(dp25580
S'a'
I60039
sS's'
I17
sS'm'
(lp25581
I1
aI1
aI5
aI7
aI23
aI17
aI93
aI165
aI371
aI495
aI865
aI2753
aI15
aI10729
aI16553
aI6039
aI96721
asS'd'
I12791
sa(dp25582
S'a'
I60043
sS's'
I17
sS'm'
(lp25583
I1
aI3
aI3
aI3
aI25
aI25
aI67
aI119
aI485
aI63
aI75
aI2365
aI4711
aI16129
aI5589
aI50621
aI1203
asS'd'
I12792
sa(dp25584
S'a'
I60048
sS's'
I17
sS'm'
(lp25585
I1
aI1
aI7
aI1
aI27
aI57
aI49
aI79
aI479
aI79
aI683
aI753
aI345
aI2007
aI18983
aI42729
aI56369
asS'd'
I12793
sa(dp25586
S'a'
I60054
sS's'
I17
sS'm'
(lp25587
I1
aI1
aI7
aI7
aI1
aI23
aI9
aI155
aI425
aI735
aI1625
aI2271
aI7875
aI11219
aI12147
aI52351
aI55845
asS'd'
I12794
sa(dp25588
S'a'
I60073
sS's'
I17
sS'm'
(lp25589
I1
aI1
aI1
aI9
aI13
aI11
aI67
aI139
aI259
aI59
aI1593
aI1207
aI237
aI11683
aI24719
aI27689
aI115617
asS'd'
I12795
sa(dp25590
S'a'
I60079
sS's'
I17
sS'm'
(lp25591
I1
aI1
aI5
aI7
aI7
aI15
aI9
aI171
aI35
aI131
aI133
aI3939
aI1401
aI6347
aI4051
aI64235
aI68581
asS'd'
I12796
sa(dp25592
S'a'
I60082
sS's'
I17
sS'm'
(lp25593
I1
aI3
aI3
aI1
aI27
aI21
aI29
aI119
aI201
aI527
aI763
aI203
aI1139
aI3951
aI3341
aI17023
aI13493
asS'd'
I12797
sa(dp25594
S'a'
I60102
sS's'
I17
sS'm'
(lp25595
I1
aI1
aI1
aI15
aI27
aI31
aI97
aI203
aI255
aI573
aI781
aI4095
aI3381
aI363
aI32733
aI34517
aI77973
asS'd'
I12798
sa(dp25596
S'a'
I60108
sS's'
I17
sS'm'
(lp25597
I1
aI1
aI3
aI9
aI29
aI9
aI115
aI207
aI267
aI513
aI1071
aI3943
aI5045
aI10071
aI6627
aI18043
aI44289
asS'd'
I12799
sa(dp25598
S'a'
I60111
sS's'
I17
sS'm'
(lp25599
I1
aI1
aI7
aI7
aI29
aI25
aI51
aI31
aI305
aI239
aI197
aI3825
aI2363
aI4903
aI16237
aI37571
aI66545
asS'd'
I12800
sa(dp25600
S'a'
I60120
sS's'
I17
sS'm'
(lp25601
I1
aI1
aI5
aI11
aI29
aI11
aI63
aI145
aI185
aI359
aI249
aI1179
aI105
aI1745
aI28819
aI40513
aI74525
asS'd'
I12801
sa(dp25602
S'a'
I60149
sS's'
I17
sS'm'
(lp25603
I1
aI1
aI7
aI1
aI15
aI35
aI39
aI5
aI139
aI119
aI2047
aI3369
aI2857
aI11037
aI30523
aI24813
aI89209
asS'd'
I12802
sa(dp25604
S'a'
I60153
sS's'
I17
sS'm'
(lp25605
I1
aI1
aI5
aI11
aI9
aI41
aI97
aI15
aI357
aI95
aI361
aI3
aI3227
aI8445
aI16541
aI30661
aI84215
asS'd'
I12803
sa(dp25606
S'a'
I60161
sS's'
I17
sS'm'
(lp25607
I1
aI1
aI7
aI5
aI11
aI55
aI77
aI19
aI331
aI621
aI893
aI577
aI1025
aI1561
aI32331
aI59469
aI67065
asS'd'
I12804
sa(dp25608
S'a'
I60168
sS's'
I17
sS'm'
(lp25609
I1
aI3
aI7
aI7
aI17
aI53
aI55
aI65
aI295
aI391
aI445
aI33
aI5361
aI669
aI6447
aI3833
aI129463
asS'd'
I12805
sa(dp25610
S'a'
I60176
sS's'
I17
sS'm'
(lp25611
I1
aI3
aI7
aI5
aI21
aI5
aI83
aI207
aI485
aI597
aI1429
aI581
aI6831
aI2885
aI24669
aI35211
aI69549
asS'd'
I12806
sa(dp25612
S'a'
I60182
sS's'
I17
sS'm'
(lp25613
I1
aI3
aI3
aI1
aI3
aI33
aI69
aI199
aI427
aI893
aI1823
aI1291
aI4533
aI11779
aI18515
aI17597
aI79159
asS'd'
I12807
sa(dp25614
S'a'
I60185
sS's'
I17
sS'm'
(lp25615
I1
aI3
aI1
aI1
aI17
aI41
aI91
aI21
aI509
aI875
aI777
aI639
aI4009
aI12103
aI12947
aI58395
aI36625
asS'd'
I12808
sa(dp25616
S'a'
I60204
sS's'
I17
sS'm'
(lp25617
I1
aI1
aI7
aI15
aI13
aI11
aI19
aI243
aI365
aI661
aI1193
aI279
aI6055
aI13921
aI5811
aI14337
aI105375
asS'd'
I12809
sa(dp25618
S'a'
I60212
sS's'
I17
sS'm'
(lp25619
I1
aI3
aI7
aI3
aI3
aI51
aI101
aI175
aI83
aI921
aI523
aI3909
aI8003
aI1295
aI4153
aI4757
aI107881
asS'd'
I12810
sa(dp25620
S'a'
I60215
sS's'
I17
sS'm'
(lp25621
I1
aI1
aI1
aI3
aI19
aI15
aI39
aI125
aI401
aI399
aI381
aI1123
aI2339
aI12231
aI13387
aI50829
aI79263
asS'd'
I12811
sa(dp25622
S'a'
I60219
sS's'
I17
sS'm'
(lp25623
I1
aI3
aI1
aI1
aI15
aI13
aI55
aI181
aI53
aI671
aI1929
aI1003
aI521
aI15279
aI1837
aI11877
aI79241
asS'd'
I12812
sa(dp25624
S'a'
I60222
sS's'
I17
sS'm'
(lp25625
I1
aI3
aI3
aI9
aI23
aI45
aI21
aI37
aI1
aI701
aI1253
aI2595
aI6261
aI4139
aI24443
aI6655
aI109755
asS'd'
I12813
sa(dp25626
S'a'
I60229
sS's'
I17
sS'm'
(lp25627
I1
aI3
aI7
aI1
aI1
aI13
aI57
aI41
aI95
aI985
aI1613
aI3487
aI7509
aI213
aI32139
aI27869
aI123589
asS'd'
I12814
sa(dp25628
S'a'
I60234
sS's'
I17
sS'm'
(lp25629
I1
aI3
aI7
aI11
aI15
aI27
aI27
aI97
aI167
aI755
aI1331
aI3961
aI3067
aI13827
aI8983
aI8755
aI77847
asS'd'
I12815
sa(dp25630
S'a'
I60241
sS's'
I17
sS'm'
(lp25631
I1
aI1
aI5
aI13
aI5
aI59
aI51
aI193
aI339
aI837
aI137
aI3807
aI2617
aI14449
aI11035
aI16827
aI24531
asS'd'
I12816
sa(dp25632
S'a'
I60248
sS's'
I17
sS'm'
(lp25633
I1
aI1
aI7
aI15
aI17
aI37
aI67
aI99
aI261
aI743
aI275
aI33
aI8107
aI4959
aI9683
aI19757
aI36471
asS'd'
I12817
sa(dp25634
S'a'
I60277
sS's'
I17
sS'm'
(lp25635
I1
aI3
aI5
aI5
aI29
aI7
aI107
aI95
aI235
aI761
aI1205
aI3125
aI4791
aI4645
aI25623
aI6463
aI121887
asS'd'
I12818
sa(dp25636
S'a'
I60282
sS's'
I17
sS'm'
(lp25637
I1
aI3
aI1
aI1
aI19
aI19
aI73
aI189
aI243
aI669
aI489
aI1927
aI1651
aI11391
aI30699
aI64719
aI60359
asS'd'
I12819
sa(dp25638
S'a'
I60287
sS's'
I17
sS'm'
(lp25639
I1
aI3
aI7
aI3
aI29
aI47
aI7
aI21
aI299
aI739
aI1605
aI749
aI5755
aI11579
aI793
aI36845
aI14695
asS'd'
I12820
sa(dp25640
S'a'
I60303
sS's'
I17
sS'm'
(lp25641
I1
aI1
aI5
aI15
aI17
aI53
aI107
aI49
aI103
aI505
aI1191
aI2881
aI7435
aI7515
aI24237
aI5397
aI47003
asS'd'
I12821
sa(dp25642
S'a'
I60306
sS's'
I17
sS'm'
(lp25643
I1
aI1
aI7
aI15
aI25
aI45
aI121
aI157
aI313
aI709
aI1519
aI2195
aI5487
aI1789
aI32401
aI4197
aI9329
asS'd'
I12822
sa(dp25644
S'a'
I60327
sS's'
I17
sS'm'
(lp25645
I1
aI3
aI3
aI11
aI21
aI17
aI77
aI153
aI275
aI765
aI1943
aI3395
aI5807
aI12809
aI29891
aI42579
aI75565
asS'd'
I12823
sa(dp25646
S'a'
I60333
sS's'
I17
sS'm'
(lp25647
I1
aI3
aI5
aI11
aI11
aI3
aI63
aI9
aI223
aI441
aI1047
aI441
aI867
aI3399
aI15813
aI13
aI25293
asS'd'
I12824
sa(dp25648
S'a'
I60334
sS's'
I17
sS'm'
(lp25649
I1
aI1
aI7
aI1
aI17
aI25
aI1
aI7
aI211
aI117
aI1417
aI1057
aI3369
aI13211
aI11437
aI20877
aI114867
asS'd'
I12825
sa(dp25650
S'a'
I60336
sS's'
I17
sS'm'
(lp25651
I1
aI1
aI1
aI15
aI5
aI41
aI89
aI165
aI343
aI447
aI55
aI1013
aI8179
aI12295
aI18615
aI23885
aI46149
asS'd'
I12826
sa(dp25652
S'a'
I60345
sS's'
I17
sS'm'
(lp25653
I1
aI1
aI5
aI1
aI15
aI37
aI109
aI141
aI323
aI151
aI1669
aI1365
aI3047
aI13145
aI30355
aI12377
aI102467
asS'd'
I12827
sa(dp25654
S'a'
I60359
sS's'
I17
sS'm'
(lp25655
I1
aI1
aI3
aI15
aI15
aI49
aI83
aI127
aI285
aI715
aI981
aI1153
aI3019
aI11071
aI24229
aI63807
aI16315
asS'd'
I12828
sa(dp25656
S'a'
I60368
sS's'
I17
sS'm'
(lp25657
I1
aI1
aI3
aI1
aI23
aI35
aI21
aI179
aI83
aI929
aI1033
aI643
aI3591
aI10363
aI7739
aI259
aI106879
asS'd'
I12829
sa(dp25658
S'a'
I60380
sS's'
I17
sS'm'
(lp25659
I1
aI1
aI5
aI7
aI19
aI9
aI63
aI241
aI387
aI851
aI1709
aI1161
aI7469
aI2093
aI6169
aI6085
aI29851
asS'd'
I12830
sa(dp25660
S'a'
I60389
sS's'
I17
sS'm'
(lp25661
I1
aI3
aI7
aI5
aI17
aI59
aI99
aI87
aI189
aI853
aI193
aI1191
aI5683
aI15865
aI27791
aI55575
aI13479
asS'd'
I12831
sa(dp25662
S'a'
I60390
sS's'
I17
sS'm'
(lp25663
I1
aI3
aI3
aI3
aI25
aI51
aI81
aI129
aI365
aI319
aI179
aI2863
aI511
aI14471
aI3689
aI59505
aI80105
asS'd'
I12832
sa(dp25664
S'a'
I60394
sS's'
I17
sS'm'
(lp25665
I1
aI1
aI1
aI15
aI5
aI33
aI69
aI35
aI429
aI715
aI1781
aI783
aI1089
aI8969
aI26987
aI23519
aI34227
asS'd'
I12833
sa(dp25666
S'a'
I60407
sS's'
I17
sS'm'
(lp25667
I1
aI3
aI5
aI9
aI1
aI51
aI9
aI121
aI325
aI945
aI2025
aI1985
aI7337
aI10837
aI21299
aI20591
aI76905
asS'd'
I12834
sa(dp25668
S'a'
I60414
sS's'
I17
sS'm'
(lp25669
I1
aI3
aI7
aI3
aI15
aI51
aI19
aI109
aI297
aI491
aI15
aI1905
aI1479
aI1997
aI18129
aI43861
aI84925
asS'd'
I12835
sa(dp25670
S'a'
I60419
sS's'
I17
sS'm'
(lp25671
I1
aI1
aI3
aI5
aI7
aI21
aI1
aI137
aI207
aI943
aI1171
aI2019
aI6687
aI10683
aI20937
aI34033
aI43907
asS'd'
I12836
sa(dp25672
S'a'
I60426
sS's'
I17
sS'm'
(lp25673
I1
aI1
aI1
aI11
aI25
aI21
aI47
aI227
aI247
aI933
aI471
aI955
aI4299
aI5605
aI18469
aI62357
aI98273
asS'd'
I12837
sa(dp25674
S'a'
I60428
sS's'
I17
sS'm'
(lp25675
I1
aI1
aI5
aI13
aI21
aI39
aI41
aI23
aI493
aI311
aI1401
aI537
aI2919
aI11519
aI12597
aI58321
aI41401
asS'd'
I12838
sa(dp25676
S'a'
I60431
sS's'
I17
sS'm'
(lp25677
I1
aI1
aI7
aI5
aI31
aI55
aI93
aI151
aI219
aI765
aI1247
aI2775
aI7167
aI13413
aI17071
aI57969
aI114069
asS'd'
I12839
sa(dp25678
S'a'
I60440
sS's'
I17
sS'm'
(lp25679
I1
aI3
aI1
aI15
aI31
aI41
aI85
aI161
aI379
aI337
aI1639
aI933
aI3511
aI925
aI3523
aI52379
aI18421
asS'd'
I12840
sa(dp25680
S'a'
I60450
sS's'
I17
sS'm'
(lp25681
I1
aI3
aI3
aI11
aI17
aI17
aI71
aI11
aI291
aI305
aI1295
aI1175
aI1803
aI6247
aI26523
aI46467
aI126999
asS'd'
I12841
sa(dp25682
S'a'
I60469
sS's'
I17
sS'm'
(lp25683
I1
aI1
aI1
aI13
aI9
aI43
aI113
aI7
aI1
aI443
aI719
aI3045
aI2527
aI5233
aI13969
aI50463
aI115629
asS'd'
I12842
sa(dp25684
S'a'
I60473
sS's'
I17
sS'm'
(lp25685
I1
aI1
aI1
aI13
aI9
aI31
aI87
aI119
aI351
aI53
aI985
aI2017
aI1187
aI10429
aI13719
aI41383
aI12749
asS'd'
I12843
sa(dp25686
S'a'
I60474
sS's'
I17
sS'm'
(lp25687
I1
aI1
aI1
aI15
aI23
aI17
aI5
aI215
aI383
aI299
aI305
aI3577
aI7707
aI6927
aI28591
aI44287
aI65697
asS'd'
I12844
sa(dp25688
S'a'
I60484
sS's'
I17
sS'm'
(lp25689
I1
aI1
aI5
aI7
aI23
aI61
aI89
aI235
aI97
aI463
aI237
aI2117
aI5503
aI13693
aI28231
aI7745
aI73631
asS'd'
I12845
sa(dp25690
S'a'
I60496
sS's'
I17
sS'm'
(lp25691
I1
aI3
aI7
aI3
aI17
aI43
aI73
aI245
aI145
aI267
aI855
aI187
aI6167
aI3999
aI9935
aI14347
aI57727
asS'd'
I12846
sa(dp25692
S'a'
I60502
sS's'
I17
sS'm'
(lp25693
I1
aI1
aI7
aI3
aI25
aI47
aI11
aI221
aI425
aI527
aI1341
aI3973
aI4635
aI16321
aI30021
aI48547
aI109409
asS'd'
I12847
sa(dp25694
S'a'
I60508
sS's'
I17
sS'm'
(lp25695
I1
aI3
aI1
aI13
aI5
aI11
aI41
aI191
aI121
aI219
aI1315
aI583
aI2997
aI5883
aI31689
aI64835
aI35351
asS'd'
I12848
sa(dp25696
S'a'
I60511
sS's'
I17
sS'm'
(lp25697
I1
aI1
aI3
aI3
aI15
aI47
aI49
aI115
aI107
aI757
aI329
aI1653
aI4633
aI14605
aI1579
aI62413
aI88685
asS'd'
I12849
sa(dp25698
S'a'
I60522
sS's'
I17
sS'm'
(lp25699
I1
aI1
aI1
aI5
aI19
aI47
aI63
aI131
aI427
aI335
aI269
aI3581
aI7613
aI15685
aI16957
aI30487
aI94965
asS'd'
I12850
sa(dp25700
S'a'
I60529
sS's'
I17
sS'm'
(lp25701
I1
aI1
aI3
aI15
aI31
aI11
aI77
aI115
aI467
aI249
aI247
aI651
aI3769
aI3701
aI21627
aI36219
aI77309
asS'd'
I12851
sa(dp25702
S'a'
I60530
sS's'
I17
sS'm'
(lp25703
I1
aI3
aI1
aI5
aI5
aI29
aI45
aI21
aI37
aI733
aI1773
aI2467
aI2747
aI9391
aI5449
aI23285
aI20089
asS'd'
I12852
sa(dp25704
S'a'
I60536
sS's'
I17
sS'm'
(lp25705
I1
aI3
aI5
aI13
aI29
aI31
aI51
aI199
aI77
aI711
aI1313
aI3303
aI2675
aI177
aI7915
aI37129
aI123641
asS'd'
I12853
sa(dp25706
S'a'
I60551
sS's'
I17
sS'm'
(lp25707
I1
aI3
aI7
aI15
aI17
aI17
aI99
aI43
aI9
aI699
aI491
aI669
aI1313
aI1377
aI30015
aI59261
aI97321
asS'd'
I12854
sa(dp25708
S'a'
I60557
sS's'
I17
sS'm'
(lp25709
I1
aI3
aI1
aI13
aI9
aI13
aI21
aI199
aI249
aI775
aI399
aI897
aI2205
aI15357
aI17281
aI3193
aI255
asS'd'
I12855
sa(dp25710
S'a'
I60566
sS's'
I17
sS'm'
(lp25711
I1
aI3
aI1
aI13
aI7
aI23
aI7
aI181
aI65
aI253
aI199
aI333
aI6507
aI4409
aI13319
aI30165
aI95191
asS'd'
I12856
sa(dp25712
S'a'
I60576
sS's'
I17
sS'm'
(lp25713
I1
aI3
aI3
aI1
aI31
aI9
aI31
aI71
aI301
aI867
aI1655
aI2065
aI597
aI15247
aI3019
aI31763
aI91889
asS'd'
I12857
sa(dp25714
S'a'
I60579
sS's'
I17
sS'm'
(lp25715
I1
aI3
aI5
aI5
aI3
aI35
aI113
aI133
aI39
aI1013
aI991
aI3521
aI5805
aI87
aI32393
aI28619
aI34325
asS'd'
I12858
sa(dp25716
S'a'
I60585
sS's'
I17
sS'm'
(lp25717
I1
aI3
aI1
aI9
aI15
aI27
aI45
aI85
aI61
aI99
aI1085
aI3251
aI7085
aI4137
aI27443
aI42581
aI94031
asS'd'
I12859
sa(dp25718
S'a'
I60588
sS's'
I17
sS'm'
(lp25719
I1
aI1
aI5
aI7
aI11
aI49
aI97
aI129
aI339
aI259
aI821
aI165
aI833
aI11383
aI21629
aI17473
aI2947
asS'd'
I12860
sa(dp25720
S'a'
I60600
sS's'
I17
sS'm'
(lp25721
I1
aI3
aI3
aI3
aI11
aI7
aI27
aI231
aI169
aI847
aI1767
aI1823
aI3855
aI14277
aI12457
aI55825
aI14377
asS'd'
I12861
sa(dp25722
S'a'
I60613
sS's'
I17
sS'm'
(lp25723
I1
aI1
aI7
aI3
aI27
aI5
aI47
aI193
aI207
aI747
aI271
aI3155
aI1097
aI2229
aI4919
aI22327
aI12659
asS'd'
I12862
sa(dp25724
S'a'
I60628
sS's'
I17
sS'm'
(lp25725
I1
aI1
aI5
aI3
aI27
aI29
aI105
aI199
aI31
aI73
aI1741
aI2173
aI4577
aI3917
aI31513
aI45983
aI118015
asS'd'
I12863
sa(dp25726
S'a'
I60631
sS's'
I17
sS'm'
(lp25727
I1
aI1
aI3
aI1
aI19
aI5
aI23
aI249
aI35
aI891
aI1105
aI1907
aI5453
aI1877
aI1965
aI3169
aI107091
asS'd'
I12864
sa(dp25728
S'a'
I60648
sS's'
I17
sS'm'
(lp25729
I1
aI1
aI7
aI9
aI11
aI47
aI57
aI171
aI255
aI661
aI1925
aI2223
aI525
aI4775
aI12327
aI8067
aI47083
asS'd'
I12865
sa(dp25730
S'a'
I60651
sS's'
I17
sS'm'
(lp25731
I1
aI3
aI3
aI11
aI29
aI43
aI37
aI33
aI459
aI117
aI7
aI1739
aI3585
aI7429
aI2217
aI9533
aI95299
asS'd'
I12866
sa(dp25732
S'a'
I60653
sS's'
I17
sS'm'
(lp25733
I1
aI1
aI5
aI11
aI23
aI3
aI33
aI13
aI45
aI201
aI467
aI597
aI4891
aI2673
aI32407
aI56935
aI121991
asS'd'
I12867
sa(dp25734
S'a'
I60671
sS's'
I17
sS'm'
(lp25735
I1
aI1
aI7
aI5
aI29
aI63
aI7
aI59
aI417
aI547
aI17
aI3701
aI5775
aI1079
aI26527
aI47187
aI14827
asS'd'
I12868
sa(dp25736
S'a'
I60673
sS's'
I17
sS'm'
(lp25737
I1
aI1
aI5
aI3
aI27
aI11
aI85
aI129
aI377
aI497
aI1659
aI1965
aI581
aI15075
aI31265
aI195
aI62307
asS'd'
I12869
sa(dp25738
S'a'
I60691
sS's'
I17
sS'm'
(lp25739
I1
aI1
aI3
aI9
aI3
aI57
aI33
aI57
aI279
aI955
aI741
aI955
aI6501
aI8069
aI27305
aI15363
aI34715
asS'd'
I12870
sa(dp25740
S'a'
I60697
sS's'
I17
sS'm'
(lp25741
I1
aI3
aI7
aI13
aI29
aI23
aI25
aI171
aI201
aI529
aI661
aI4089
aI5755
aI12459
aI31269
aI9763
aI53217
asS'd'
I12871
sa(dp25742
S'a'
I60700
sS's'
I17
sS'm'
(lp25743
I1
aI3
aI7
aI1
aI3
aI19
aI27
aI201
aI261
aI421
aI1487
aI2907
aI547
aI15791
aI7661
aI42871
aI116343
asS'd'
I12872
sa(dp25744
S'a'
I60714
sS's'
I17
sS'm'
(lp25745
I1
aI3
aI7
aI9
aI5
aI59
aI51
aI91
aI7
aI399
aI2001
aI661
aI6577
aI7473
aI5439
aI29073
aI3391
asS'd'
I12873
sa(dp25746
S'a'
I60724
sS's'
I17
sS'm'
(lp25747
I1
aI1
aI7
aI1
aI23
aI39
aI119
aI105
aI113
aI913
aI1097
aI2547
aI8143
aI11409
aI23197
aI59527
aI55677
asS'd'
I12874
sa(dp25748
S'a'
I60728
sS's'
I17
sS'm'
(lp25749
I1
aI3
aI3
aI15
aI31
aI35
aI103
aI207
aI247
aI801
aI1821
aI1995
aI4437
aI12891
aI659
aI15687
aI53
asS'd'
I12875
sa(dp25750
S'a'
I60733
sS's'
I17
sS'm'
(lp25751
I1
aI3
aI3
aI5
aI13
aI5
aI45
aI187
aI231
aI661
aI1553
aI2909
aI3715
aI4499
aI14773
aI5957
aI24171
asS'd'
I12876
sa(dp25752
S'a'
I60736
sS's'
I17
sS'm'
(lp25753
I1
aI1
aI1
aI11
aI3
aI53
aI93
aI29
aI379
aI713
aI299
aI445
aI2815
aI9825
aI30941
aI22413
aI91563
asS'd'
I12877
sa(dp25754
S'a'
I60745
sS's'
I17
sS'm'
(lp25755
I1
aI3
aI1
aI1
aI27
aI31
aI111
aI83
aI349
aI895
aI1007
aI2649
aI7139
aI5863
aI27739
aI53099
aI6837
asS'd'
I12878
sa(dp25756
S'a'
I60746
sS's'
I17
sS'm'
(lp25757
I1
aI3
aI1
aI5
aI23
aI57
aI121
aI229
aI487
aI405
aI2001
aI2761
aI3011
aI2219
aI10711
aI31139
aI83809
asS'd'
I12879
sa(dp25758
S'a'
I60753
sS's'
I17
sS'm'
(lp25759
I1
aI3
aI7
aI5
aI13
aI51
aI37
aI181
aI359
aI909
aI2001
aI793
aI1143
aI9219
aI5111
aI23021
aI126081
asS'd'
I12880
sa(dp25760
S'a'
I60754
sS's'
I17
sS'm'
(lp25761
I1
aI3
aI1
aI13
aI27
aI27
aI99
aI25
aI189
aI129
aI1831
aI1005
aI8119
aI2557
aI26985
aI63447
aI89693
asS'd'
I12881
sa(dp25762
S'a'
I60782
sS's'
I17
sS'm'
(lp25763
I1
aI1
aI7
aI5
aI1
aI21
aI79
aI33
aI99
aI7
aI433
aI1343
aI3121
aI3705
aI477
aI41191
aI13749
asS'd'
I12882
sa(dp25764
S'a'
I60784
sS's'
I17
sS'm'
(lp25765
I1
aI1
aI5
aI5
aI29
aI53
aI75
aI243
aI35
aI461
aI1399
aI129
aI177
aI6533
aI22209
aI23503
aI43819
asS'd'
I12883
sa(dp25766
S'a'
I60790
sS's'
I17
sS'm'
(lp25767
I1
aI3
aI7
aI7
aI31
aI37
aI109
aI9
aI255
aI263
aI35
aI3451
aI7629
aI9849
aI10387
aI3165
aI120623
asS'd'
I12884
sa(dp25768
S'a'
I60793
sS's'
I17
sS'm'
(lp25769
I1
aI3
aI5
aI3
aI27
aI53
aI93
aI111
aI239
aI723
aI293
aI1481
aI4427
aI13623
aI1989
aI47705
aI122069
asS'd'
I12885
sa(dp25770
S'a'
I60805
sS's'
I17
sS'm'
(lp25771
I1
aI3
aI7
aI7
aI31
aI37
aI37
aI213
aI191
aI627
aI1821
aI3621
aI2875
aI15759
aI17525
aI50969
aI35311
asS'd'
I12886
sa(dp25772
S'a'
I60830
sS's'
I17
sS'm'
(lp25773
I1
aI3
aI5
aI5
aI11
aI41
aI87
aI233
aI79
aI251
aI25
aI1385
aI3527
aI7853
aI5541
aI36519
aI42779
asS'd'
I12887
sa(dp25774
S'a'
I60836
sS's'
I17
sS'm'
(lp25775
I1
aI3
aI1
aI5
aI9
aI1
aI117
aI11
aI61
aI879
aI553
aI383
aI6237
aI15207
aI3057
aI28051
aI59149
asS'd'
I12888
sa(dp25776
S'a'
I60846
sS's'
I17
sS'm'
(lp25777
I1
aI1
aI1
aI15
aI15
aI7
aI37
aI133
aI81
aI815
aI893
aI2281
aI2459
aI15325
aI20107
aI2289
aI34535
asS'd'
I12889
sa(dp25778
S'a'
I60851
sS's'
I17
sS'm'
(lp25779
I1
aI1
aI5
aI11
aI17
aI3
aI45
aI159
aI409
aI643
aI969
aI1289
aI4353
aI10465
aI16233
aI55561
aI111667
asS'd'
I12890
sa(dp25780
S'a'
I60880
sS's'
I17
sS'm'
(lp25781
I1
aI1
aI5
aI13
aI23
aI1
aI79
aI127
aI485
aI1013
aI629
aI791
aI853
aI9247
aI26333
aI1123
aI17313
asS'd'
I12891
sa(dp25782
S'a'
I60896
sS's'
I17
sS'm'
(lp25783
I1
aI1
aI5
aI11
aI27
aI17
aI97
aI157
aI479
aI421
aI1739
aI3257
aI2419
aI6673
aI2759
aI19399
aI120305
asS'd'
I12892
sa(dp25784
S'a'
I60905
sS's'
I17
sS'm'
(lp25785
I1
aI3
aI5
aI1
aI3
aI43
aI71
aI55
aI111
aI949
aI1957
aI3777
aI3409
aI8229
aI26585
aI49221
aI33639
asS'd'
I12893
sa(dp25786
S'a'
I60923
sS's'
I17
sS'm'
(lp25787
I1
aI3
aI7
aI9
aI17
aI45
aI71
aI71
aI417
aI1007
aI213
aI1069
aI2693
aI5065
aI23489
aI33363
aI120459
asS'd'
I12894
sa(dp25788
S'a'
I60925
sS's'
I17
sS'm'
(lp25789
I1
aI1
aI1
aI1
aI25
aI47
aI81
aI251
aI341
aI101
aI1941
aI1133
aI3205
aI4141
aI26561
aI56095
aI37193
asS'd'
I12895
sa(dp25790
S'a'
I60932
sS's'
I17
sS'm'
(lp25791
I1
aI1
aI3
aI7
aI25
aI45
aI97
aI95
aI135
aI871
aI949
aI3489
aI7593
aI10717
aI26163
aI12711
aI76989
asS'd'
I12896
sa(dp25792
S'a'
I60939
sS's'
I17
sS'm'
(lp25793
I1
aI3
aI1
aI9
aI3
aI11
aI35
aI7
aI471
aI509
aI1335
aI385
aI1297
aI11201
aI28553
aI51609
aI45351
asS'd'
I12897
sa(dp25794
S'a'
I60942
sS's'
I17
sS'm'
(lp25795
I1
aI1
aI5
aI11
aI21
aI23
aI105
aI31
aI125
aI5
aI1721
aI1503
aI7807
aI13093
aI24873
aI18467
aI30183
asS'd'
I12898
sa(dp25796
S'a'
I60953
sS's'
I17
sS'm'
(lp25797
I1
aI3
aI7
aI15
aI15
aI57
aI61
aI213
aI79
aI655
aI517
aI1031
aI6719
aI4807
aI12805
aI2605
aI35407
asS'd'
I12899
sa(dp25798
S'a'
I60956
sS's'
I17
sS'm'
(lp25799
I1
aI1
aI3
aI15
aI21
aI5
aI93
aI61
aI103
aI945
aI935
aI115
aI1281
aI7735
aI20723
aI37211
aI50039
asS'd'
I12900
sa(dp25800
S'a'
I60959
sS's'
I17
sS'm'
(lp25801
I1
aI3
aI3
aI15
aI19
aI51
aI79
aI187
aI127
aI205
aI121
aI701
aI6065
aI15185
aI29343
aI58249
aI25331
asS'd'
I12901
sa(dp25802
S'a'
I60963
sS's'
I17
sS'm'
(lp25803
I1
aI3
aI3
aI15
aI17
aI49
aI23
aI163
aI201
aI809
aI1203
aI687
aI1777
aI695
aI18779
aI31571
aI118383
asS'd'
I12902
sa(dp25804
S'a'
I60969
sS's'
I17
sS'm'
(lp25805
I1
aI1
aI1
aI5
aI25
aI45
aI121
aI223
aI233
aI193
aI1459
aI1889
aI5537
aI4421
aI13659
aI4591
aI112563
asS'd'
I12903
sa(dp25806
S'a'
I60978
sS's'
I17
sS'm'
(lp25807
I1
aI3
aI5
aI3
aI31
aI37
aI109
aI15
aI381
aI373
aI993
aI3633
aI641
aI4411
aI32265
aI46481
aI49195
asS'd'
I12904
sa(dp25808
S'a'
I60980
sS's'
I17
sS'm'
(lp25809
I1
aI3
aI3
aI11
aI19
aI21
aI39
aI67
aI447
aI829
aI1163
aI55
aI2153
aI15045
aI6643
aI48235
aI59261
asS'd'
I12905
sa(dp25810
S'a'
I60987
sS's'
I17
sS'm'
(lp25811
I1
aI3
aI3
aI1
aI7
aI63
aI37
aI71
aI35
aI601
aI1719
aI447
aI8009
aI7235
aI13225
aI44103
aI82409
asS'd'
I12906
sa(dp25812
S'a'
I61007
sS's'
I17
sS'm'
(lp25813
I1
aI3
aI3
aI7
aI13
aI33
aI69
aI115
aI207
aI807
aI109
aI2989
aI3727
aI9017
aI29095
aI11377
aI122401
asS'd'
I12907
sa(dp25814
S'a'
I61012
sS's'
I17
sS'm'
(lp25815
I1
aI1
aI3
aI15
aI9
aI9
aI57
aI197
aI115
aI73
aI1277
aI1727
aI5275
aI11897
aI12157
aI34763
aI58273
asS'd'
I12908
sa(dp25816
S'a'
I61015
sS's'
I17
sS'm'
(lp25817
I1
aI1
aI3
aI15
aI19
aI19
aI127
aI105
aI289
aI663
aI877
aI1303
aI4901
aI8897
aI4803
aI36853
aI93361
asS'd'
I12909
sa(dp25818
S'a'
I61025
sS's'
I17
sS'm'
(lp25819
I1
aI1
aI3
aI7
aI23
aI29
aI121
aI29
aI439
aI795
aI1469
aI523
aI7767
aI12061
aI15613
aI57343
aI21291
asS'd'
I12910
sa(dp25820
S'a'
I61026
sS's'
I17
sS'm'
(lp25821
I1
aI1
aI1
aI9
aI25
aI29
aI15
aI165
aI383
aI233
aI91
aI2065
aI2741
aI7809
aI5325
aI48581
aI65551
asS'd'
I12911
sa(dp25822
S'a'
I61038
sS's'
I17
sS'm'
(lp25823
I1
aI1
aI5
aI15
aI29
aI19
aI103
aI143
aI283
aI597
aI1055
aI3525
aI6115
aI11083
aI22497
aI42893
aI86849
asS'd'
I12912
sa(dp25824
S'a'
I61040
sS's'
I17
sS'm'
(lp25825
I1
aI1
aI1
aI15
aI13
aI43
aI75
aI157
aI267
aI519
aI1231
aI929
aI1585
aI16137
aI1045
aI4353
aI63473
asS'd'
I12913
sa(dp25826
S'a'
I61052
sS's'
I17
sS'm'
(lp25827
I1
aI1
aI1
aI9
aI17
aI25
aI73
aI227
aI145
aI921
aI1845
aI2057
aI3099
aI15523
aI8993
aI14135
aI37811
asS'd'
I12914
sa(dp25828
S'a'
I61061
sS's'
I17
sS'm'
(lp25829
I1
aI3
aI3
aI15
aI17
aI57
aI107
aI215
aI271
aI841
aI1543
aI2803
aI625
aI15359
aI13341
aI36879
aI83191
asS'd'
I12915
sa(dp25830
S'a'
I61074
sS's'
I17
sS'm'
(lp25831
I1
aI3
aI5
aI13
aI3
aI17
aI127
aI81
aI193
aI253
aI71
aI3205
aI1157
aI1313
aI27341
aI49657
aI96539
asS'd'
I12916
sa(dp25832
S'a'
I61083
sS's'
I17
sS'm'
(lp25833
I1
aI3
aI1
aI5
aI27
aI43
aI1
aI111
aI23
aI963
aI1853
aI925
aI7401
aI13999
aI29797
aI47125
aI59955
asS'd'
I12917
sa(dp25834
S'a'
I61085
sS's'
I17
sS'm'
(lp25835
I1
aI3
aI3
aI1
aI31
aI55
aI107
aI121
aI37
aI159
aI61
aI577
aI5711
aI6745
aI20077
aI42333
aI37105
asS'd'
I12918
sa(dp25836
S'a'
I61086
sS's'
I17
sS'm'
(lp25837
I1
aI1
aI5
aI3
aI31
aI11
aI1
aI7
aI295
aI515
aI203
aI707
aI2919
aI9619
aI8877
aI45143
aI101861
asS'd'
I12919
sa(dp25838
S'a'
I61096
sS's'
I17
sS'm'
(lp25839
I1
aI3
aI7
aI11
aI5
aI23
aI71
aI9
aI99
aI947
aI1141
aI1651
aI1903
aI13607
aI15433
aI55005
aI127639
asS'd'
I12920
sa(dp25840
S'a'
I61119
sS's'
I17
sS'm'
(lp25841
I1
aI1
aI7
aI3
aI13
aI61
aI25
aI111
aI239
aI243
aI1069
aI3551
aI3339
aI743
aI29921
aI21313
aI54953
asS'd'
I12921
sa(dp25842
S'a'
I61124
sS's'
I17
sS'm'
(lp25843
I1
aI1
aI3
aI3
aI23
aI7
aI21
aI47
aI367
aI871
aI1647
aI2183
aI2615
aI2257
aI30447
aI25761
aI110221
asS'd'
I12922
sa(dp25844
S'a'
I61127
sS's'
I17
sS'm'
(lp25845
I1
aI3
aI7
aI15
aI13
aI51
aI115
aI19
aI277
aI463
aI475
aI3467
aI7313
aI2477
aI10841
aI13585
aI61449
asS'd'
I12923
sa(dp25846
S'a'
I61128
sS's'
I17
sS'm'
(lp25847
I1
aI1
aI7
aI9
aI1
aI27
aI111
aI209
aI391
aI621
aI1047
aI549
aI2013
aI549
aI24213
aI6369
aI68691
asS'd'
I12924
sa(dp25848
S'a'
I61133
sS's'
I17
sS'm'
(lp25849
I1
aI1
aI1
aI11
aI19
aI59
aI11
aI107
aI79
aI1013
aI357
aI1729
aI889
aI12823
aI6537
aI35717
aI42761
asS'd'
I12925
sa(dp25850
S'a'
I61134
sS's'
I17
sS'm'
(lp25851
I1
aI1
aI3
aI15
aI29
aI41
aI49
aI177
aI309
aI293
aI1035
aI1481
aI1395
aI2009
aI7917
aI365
aI28981
asS'd'
I12926
sa(dp25852
S'a'
I61146
sS's'
I17
sS'm'
(lp25853
I1
aI3
aI7
aI11
aI31
aI19
aI51
aI113
aI479
aI347
aI353
aI929
aI1089
aI3373
aI2807
aI55201
aI23137
asS'd'
I12927
sa(dp25854
S'a'
I61155
sS's'
I17
sS'm'
(lp25855
I1
aI3
aI1
aI15
aI7
aI55
aI79
aI191
aI267
aI701
aI1885
aI1241
aI7085
aI9835
aI24239
aI7609
aI13967
asS'd'
I12928
sa(dp25856
S'a'
I61161
sS's'
I17
sS'm'
(lp25857
I1
aI1
aI3
aI15
aI25
aI33
aI69
aI5
aI93
aI375
aI435
aI2401
aI1591
aI8173
aI17293
aI20281
aI63809
asS'd'
I12929
sa(dp25858
S'a'
I61162
sS's'
I17
sS'm'
(lp25859
I1
aI3
aI5
aI3
aI9
aI49
aI63
aI47
aI217
aI773
aI1241
aI1131
aI7521
aI15607
aI24341
aI20353
aI122801
asS'd'
I12930
sa(dp25860
S'a'
I61169
sS's'
I17
sS'm'
(lp25861
I1
aI3
aI7
aI15
aI21
aI1
aI57
aI159
aI279
aI987
aI1641
aI3883
aI1659
aI7875
aI24857
aI33273
aI88933
asS'd'
I12931
sa(dp25862
S'a'
I61179
sS's'
I17
sS'm'
(lp25863
I1
aI1
aI3
aI11
aI3
aI23
aI45
aI31
aI279
aI643
aI1285
aI471
aI137
aI15871
aI13927
aI52361
aI118901
asS'd'
I12932
sa(dp25864
S'a'
I61182
sS's'
I17
sS'm'
(lp25865
I1
aI3
aI1
aI15
aI27
aI51
aI83
aI19
aI299
aI213
aI1001
aI897
aI2751
aI13085
aI20841
aI24891
aI113173
asS'd'
I12933
sa(dp25866
S'a'
I61190
sS's'
I17
sS'm'
(lp25867
I1
aI1
aI7
aI5
aI5
aI27
aI77
aI77
aI165
aI489
aI359
aI1607
aI3903
aI16241
aI641
aI25999
aI29279
asS'd'
I12934
sa(dp25868
S'a'
I61201
sS's'
I17
sS'm'
(lp25869
I1
aI3
aI3
aI7
aI15
aI23
aI103
aI49
aI259
aI519
aI641
aI1577
aI3713
aI12181
aI287
aI29483
aI58505
asS'd'
I12935
sa(dp25870
S'a'
I61208
sS's'
I17
sS'm'
(lp25871
I1
aI1
aI7
aI13
aI11
aI29
aI125
aI45
aI45
aI167
aI261
aI2735
aI2421
aI15457
aI5965
aI44005
aI82141
asS'd'
I12936
sa(dp25872
S'a'
I61238
sS's'
I17
sS'm'
(lp25873
I1
aI1
aI3
aI9
aI25
aI21
aI9
aI3
aI57
aI1017
aI1359
aI79
aI6789
aI7805
aI20683
aI25695
aI38893
asS'd'
I12937
sa(dp25874
S'a'
I61241
sS's'
I17
sS'm'
(lp25875
I1
aI1
aI7
aI1
aI29
aI53
aI87
aI171
aI51
aI5
aI9
aI3033
aI787
aI10611
aI15913
aI35703
aI70459
asS'd'
I12938
sa(dp25876
S'a'
I61247
sS's'
I17
sS'm'
(lp25877
I1
aI1
aI3
aI5
aI1
aI33
aI111
aI139
aI477
aI33
aI1287
aI3615
aI5235
aI15491
aI2915
aI47821
aI55257
asS'd'
I12939
sa(dp25878
S'a'
I61259
sS's'
I17
sS'm'
(lp25879
I1
aI1
aI1
aI13
aI5
aI23
aI55
aI225
aI303
aI587
aI1595
aI307
aI3809
aI8093
aI13297
aI63213
aI104317
asS'd'
I12940
sa(dp25880
S'a'
I61267
sS's'
I17
sS'm'
(lp25881
I1
aI1
aI3
aI15
aI31
aI29
aI13
aI33
aI267
aI481
aI1039
aI3805
aI2549
aI861
aI12483
aI61829
aI127725
asS'd'
I12941
sa(dp25882
S'a'
I61269
sS's'
I17
sS'm'
(lp25883
I1
aI3
aI5
aI11
aI23
aI17
aI23
aI117
aI333
aI167
aI1965
aI1387
aI5453
aI15545
aI123
aI12991
aI36281
asS'd'
I12942
sa(dp25884
S'a'
I61295
sS's'
I17
sS'm'
(lp25885
I1
aI3
aI5
aI1
aI3
aI9
aI25
aI55
aI497
aI951
aI1377
aI993
aI6089
aI4801
aI32719
aI31579
aI126663
asS'd'
I12943
sa(dp25886
S'a'
I61304
sS's'
I17
sS'm'
(lp25887
I1
aI3
aI5
aI11
aI15
aI37
aI103
aI51
aI509
aI585
aI769
aI425
aI835
aI14027
aI30265
aI55735
aI36655
asS'd'
I12944
sa(dp25888
S'a'
I61309
sS's'
I17
sS'm'
(lp25889
I1
aI1
aI1
aI13
aI9
aI49
aI105
aI61
aI493
aI3
aI1663
aI815
aI8105
aI16361
aI32477
aI30437
aI61519
asS'd'
I12945
sa(dp25890
S'a'
I61310
sS's'
I17
sS'm'
(lp25891
I1
aI3
aI7
aI11
aI29
aI23
aI105
aI87
aI119
aI399
aI1405
aI1515
aI7017
aI12729
aI13769
aI29741
aI30921
asS'd'
I12946
sa(dp25892
S'a'
I61313
sS's'
I17
sS'm'
(lp25893
I1
aI3
aI7
aI13
aI13
aI7
aI93
aI227
aI489
aI843
aI923
aI3373
aI759
aI5105
aI9059
aI21079
aI101335
asS'd'
I12947
sa(dp25894
S'a'
I61320
sS's'
I17
sS'm'
(lp25895
I1
aI1
aI1
aI1
aI29
aI53
aI119
aI227
aI171
aI363
aI289
aI827
aI425
aI12827
aI25791
aI21587
aI109567
asS'd'
I12948
sa(dp25896
S'a'
I61325
sS's'
I17
sS'm'
(lp25897
I1
aI1
aI5
aI11
aI29
aI29
aI53
aI127
aI441
aI219
aI1049
aI275
aI525
aI5535
aI20907
aI9299
aI76319
asS'd'
I12949
sa(dp25898
S'a'
I61334
sS's'
I17
sS'm'
(lp25899
I1
aI3
aI7
aI15
aI3
aI53
aI109
aI61
aI275
aI391
aI1147
aI2953
aI1439
aI4417
aI679
aI10949
aI35101
asS'd'
I12950
sa(dp25900
S'a'
I61337
sS's'
I17
sS'm'
(lp25901
I1
aI1
aI5
aI13
aI9
aI1
aI109
aI137
aI249
aI835
aI721
aI129
aI2883
aI13043
aI31827
aI36741
aI107167
asS'd'
I12951
sa(dp25902
S'a'
I61386
sS's'
I17
sS'm'
(lp25903
I1
aI1
aI5
aI9
aI27
aI17
aI117
aI121
aI111
aI433
aI743
aI1987
aI6839
aI8439
aI2533
aI62135
aI54281
asS'd'
I12952
sa(dp25904
S'a'
I61399
sS's'
I17
sS'm'
(lp25905
I1
aI3
aI5
aI15
aI11
aI61
aI117
aI103
aI409
aI485
aI1047
aI469
aI2245
aI7193
aI2541
aI9399
aI88127
asS'd'
I12953
sa(dp25906
S'a'
I61422
sS's'
I17
sS'm'
(lp25907
I1
aI3
aI1
aI9
aI3
aI49
aI111
aI201
aI87
aI181
aI1243
aI3261
aI1827
aI10385
aI13045
aI58331
aI107729
asS'd'
I12954
sa(dp25908
S'a'
I61429
sS's'
I17
sS'm'
(lp25909
I1
aI3
aI5
aI15
aI13
aI29
aI61
aI223
aI227
aI733
aI1757
aI755
aI4231
aI13537
aI1509
aI54623
aI120221
asS'd'
I12955
sa(dp25910
S'a'
I61436
sS's'
I17
sS'm'
(lp25911
I1
aI1
aI1
aI15
aI13
aI9
aI7
aI233
aI391
aI689
aI355
aI1203
aI5811
aI7759
aI2647
aI54949
aI51891
asS'd'
I12956
sa(dp25912
S'a'
I61439
sS's'
I17
sS'm'
(lp25913
I1
aI1
aI3
aI1
aI3
aI27
aI95
aI51
aI497
aI315
aI915
aI1055
aI2917
aI167
aI1849
aI26591
aI102729
asS'd'
I12957
sa(dp25914
S'a'
I61466
sS's'
I17
sS'm'
(lp25915
I1
aI3
aI7
aI13
aI1
aI51
aI3
aI113
aI437
aI449
aI1889
aI2887
aI4735
aI5693
aI8191
aI12847
aI52651
asS'd'
I12958
sa(dp25916
S'a'
I61477
sS's'
I17
sS'm'
(lp25917
I1
aI1
aI7
aI13
aI1
aI45
aI41
aI221
aI403
aI185
aI1653
aI1809
aI6405
aI9193
aI1381
aI36677
aI43255
asS'd'
I12959
sa(dp25918
S'a'
I61478
sS's'
I17
sS'm'
(lp25919
I1
aI1
aI5
aI1
aI25
aI51
aI109
aI245
aI291
aI809
aI1381
aI3235
aI5933
aI10185
aI18663
aI15589
aI39539
asS'd'
I12960
sa(dp25920
S'a'
I61490
sS's'
I17
sS'm'
(lp25921
I1
aI3
aI3
aI1
aI27
aI29
aI3
aI227
aI275
aI705
aI489
aI681
aI323
aI7453
aI26005
aI13791
aI115219
asS'd'
I12961
sa(dp25922
S'a'
I61495
sS's'
I17
sS'm'
(lp25923
I1
aI3
aI5
aI1
aI3
aI51
aI101
aI75
aI157
aI529
aI45
aI3105
aI3617
aI13081
aI21665
aI50065
aI117823
asS'd'
I12962
sa(dp25924
S'a'
I61504
sS's'
I17
sS'm'
(lp25925
I1
aI3
aI5
aI15
aI9
aI43
aI41
aI169
aI391
aI455
aI1375
aI253
aI1257
aI14427
aI16325
aI11571
aI26361
asS'd'
I12963
sa(dp25926
S'a'
I61514
sS's'
I17
sS'm'
(lp25927
I1
aI1
aI5
aI7
aI5
aI41
aI81
aI173
aI275
aI225
aI301
aI335
aI5473
aI1509
aI20897
aI21951
aI103967
asS'd'
I12964
sa(dp25928
S'a'
I61516
sS's'
I17
sS'm'
(lp25929
I1
aI3
aI5
aI1
aI13
aI27
aI107
aI19
aI221
aI609
aI823
aI1193
aI665
aI4947
aI11967
aI57373
aI21665
asS'd'
I12965
sa(dp25930
S'a'
I61521
sS's'
I17
sS'm'
(lp25931
I1
aI3
aI7
aI13
aI7
aI11
aI109
aI59
aI193
aI103
aI943
aI595
aI5121
aI6159
aI2103
aI52863
aI57541
asS'd'
I12966
sa(dp25932
S'a'
I61527
sS's'
I17
sS'm'
(lp25933
I1
aI3
aI5
aI3
aI5
aI51
aI85
aI227
aI465
aI1013
aI601
aI1687
aI7615
aI5991
aI8635
aI64487
aI69967
asS'd'
I12967
sa(dp25934
S'a'
I61531
sS's'
I17
sS'm'
(lp25935
I1
aI3
aI1
aI11
aI29
aI49
aI79
aI25
aI447
aI119
aI569
aI383
aI5261
aI6209
aI21965
aI40863
aI96593
asS'd'
I12968
sa(dp25936
S'a'
I61540
sS's'
I17
sS'm'
(lp25937
I1
aI1
aI3
aI13
aI9
aI49
aI59
aI93
aI369
aI789
aI1373
aI425
aI3565
aI13357
aI17783
aI46435
aI129653
asS'd'
I12969
sa(dp25938
S'a'
I61550
sS's'
I17
sS'm'
(lp25939
I1
aI3
aI7
aI5
aI5
aI39
aI51
aI15
aI421
aI531
aI1855
aI2105
aI5335
aI8509
aI20841
aI44997
aI48235
asS'd'
I12970
sa(dp25940
S'a'
I61557
sS's'
I17
sS'm'
(lp25941
I1
aI3
aI7
aI3
aI27
aI47
aI113
aI1
aI453
aI565
aI1843
aI243
aI7663
aI10697
aI7725
aI24485
aI49435
asS'd'
I12971
sa(dp25942
S'a'
I61562
sS's'
I17
sS'm'
(lp25943
I1
aI1
aI1
aI11
aI25
aI25
aI47
aI1
aI475
aI831
aI1833
aI391
aI5173
aI14873
aI14263
aI36061
aI26781
asS'd'
I12972
sa(dp25944
S'a'
I61577
sS's'
I17
sS'm'
(lp25945
I1
aI1
aI7
aI15
aI21
aI13
aI5
aI169
aI241
aI235
aI547
aI1565
aI2053
aI6877
aI12811
aI22213
aI106907
asS'd'
I12973
sa(dp25946
S'a'
I61583
sS's'
I17
sS'm'
(lp25947
I1
aI3
aI7
aI1
aI21
aI11
aI101
aI115
aI243
aI985
aI1389
aI2189
aI563
aI12453
aI14951
aI58889
aI48079
asS'd'
I12974
sa(dp25948
S'a'
I61597
sS's'
I17
sS'm'
(lp25949
I1
aI1
aI5
aI7
aI9
aI37
aI33
aI241
aI337
aI453
aI1955
aI1731
aI4445
aI8887
aI27715
aI63975
aI95891
asS'd'
I12975
sa(dp25950
S'a'
I61602
sS's'
I17
sS'm'
(lp25951
I1
aI1
aI5
aI1
aI23
aI21
aI95
aI237
aI241
aI991
aI1159
aI2417
aI2279
aI8941
aI20987
aI39773
aI79327
asS'd'
I12976
sa(dp25952
S'a'
I61604
sS's'
I17
sS'm'
(lp25953
I1
aI3
aI3
aI1
aI19
aI39
aI73
aI253
aI137
aI1009
aI1793
aI4007
aI2017
aI7503
aI16689
aI29013
aI41571
asS'd'
I12977
sa(dp25954
S'a'
I61607
sS's'
I17
sS'm'
(lp25955
I1
aI3
aI7
aI15
aI3
aI63
aI77
aI11
aI293
aI495
aI339
aI3525
aI5747
aI1807
aI11705
aI55807
aI54163
asS'd'
I12978
sa(dp25956
S'a'
I61622
sS's'
I17
sS'm'
(lp25957
I1
aI3
aI7
aI3
aI25
aI41
aI127
aI23
aI113
aI807
aI387
aI3529
aI2173
aI11217
aI21257
aI61169
aI47749
asS'd'
I12979
sa(dp25958
S'a'
I61625
sS's'
I17
sS'm'
(lp25959
I1
aI3
aI3
aI5
aI27
aI5
aI43
aI55
aI207
aI995
aI811
aI1473
aI481
aI4317
aI2015
aI49161
aI94711
asS'd'
I12980
sa(dp25960
S'a'
I61633
sS's'
I17
sS'm'
(lp25961
I1
aI3
aI1
aI9
aI21
aI61
aI41
aI147
aI425
aI353
aI1943
aI2455
aI379
aI10729
aI3045
aI16013
aI44527
asS'd'
I12981
sa(dp25962
S'a'
I61640
sS's'
I17
sS'm'
(lp25963
I1
aI3
aI1
aI5
aI17
aI43
aI109
aI231
aI313
aI277
aI939
aI3491
aI5883
aI2297
aI4763
aI33403
aI62395
asS'd'
I12982
sa(dp25964
S'a'
I61643
sS's'
I17
sS'm'
(lp25965
I1
aI1
aI3
aI9
aI1
aI49
aI37
aI145
aI383
aI467
aI621
aI2873
aI873
aI6163
aI16475
aI49045
aI115599
asS'd'
I12983
sa(dp25966
S'a'
I61651
sS's'
I17
sS'm'
(lp25967
I1
aI1
aI1
aI9
aI5
aI15
aI125
aI157
aI459
aI727
aI807
aI2769
aI5531
aI11531
aI4277
aI42301
aI16969
asS'd'
I12984
sa(dp25968
S'a'
I61658
sS's'
I17
sS'm'
(lp25969
I1
aI1
aI3
aI1
aI15
aI23
aI39
aI121
aI163
aI537
aI409
aI1217
aI8007
aI5671
aI19681
aI25371
aI69227
asS'd'
I12985
sa(dp25970
S'a'
I61670
sS's'
I17
sS'm'
(lp25971
I1
aI3
aI7
aI9
aI23
aI1
aI93
aI41
aI267
aI995
aI1917
aI3441
aI6237
aI7233
aI30215
aI31945
aI33967
asS'd'
I12986
sa(dp25972
S'a'
I61674
sS's'
I17
sS'm'
(lp25973
I1
aI1
aI1
aI15
aI7
aI5
aI123
aI53
aI359
aI677
aI1061
aI1649
aI651
aI14079
aI30211
aI14827
aI123175
asS'd'
I12987
sa(dp25974
S'a'
I61676
sS's'
I17
sS'm'
(lp25975
I1
aI3
aI1
aI5
aI11
aI19
aI121
aI135
aI167
aI293
aI493
aI949
aI5459
aI11785
aI21445
aI57161
aI129737
asS'd'
I12988
sa(dp25976
S'a'
I61679
sS's'
I17
sS'm'
(lp25977
I1
aI1
aI3
aI13
aI19
aI39
aI43
aI55
aI149
aI549
aI479
aI925
aI341
aI1151
aI12007
aI23473
aI10671
asS'd'
I12989
sa(dp25978
S'a'
I61688
sS's'
I17
sS'm'
(lp25979
I1
aI3
aI5
aI9
aI7
aI41
aI37
aI103
aI381
aI373
aI1767
aI3959
aI3001
aI7903
aI24033
aI55123
aI93627
asS'd'
I12990
sa(dp25980
S'a'
I61693
sS's'
I17
sS'm'
(lp25981
I1
aI3
aI3
aI3
aI31
aI27
aI93
aI175
aI393
aI575
aI703
aI3715
aI6043
aI11763
aI7613
aI15907
aI56821
asS'd'
I12991
sa(dp25982
S'a'
I61701
sS's'
I17
sS'm'
(lp25983
I1
aI3
aI3
aI13
aI3
aI13
aI75
aI85
aI89
aI851
aI1171
aI3075
aI5265
aI10293
aI14745
aI32153
aI89877
asS'd'
I12992
sa(dp25984
S'a'
I61711
sS's'
I17
sS'm'
(lp25985
I1
aI1
aI7
aI11
aI1
aI25
aI101
aI149
aI187
aI197
aI1485
aI1555
aI1599
aI7413
aI23275
aI27093
aI73483
asS'd'
I12993
sa(dp25986
S'a'
I61714
sS's'
I17
sS'm'
(lp25987
I1
aI3
aI1
aI1
aI19
aI15
aI63
aI111
aI211
aI197
aI77
aI1683
aI3159
aI235
aI32601
aI35715
aI59537
asS'd'
I12994
sa(dp25988
S'a'
I61723
sS's'
I17
sS'm'
(lp25989
I1
aI3
aI5
aI11
aI23
aI61
aI91
aI135
aI403
aI669
aI267
aI2507
aI2931
aI7813
aI5047
aI40027
aI111705
asS'd'
I12995
sa(dp25990
S'a'
I61725
sS's'
I17
sS'm'
(lp25991
I1
aI1
aI5
aI13
aI7
aI5
aI65
aI37
aI87
aI405
aI335
aI1095
aI3717
aI1717
aI31551
aI28181
aI47407
asS'd'
I12996
sa(dp25992
S'a'
I61726
sS's'
I17
sS'm'
(lp25993
I1
aI1
aI5
aI13
aI3
aI43
aI67
aI99
aI507
aI861
aI1063
aI3003
aI6095
aI11079
aI6919
aI41597
aI97709
asS'd'
I12997
sa(dp25994
S'a'
I61729
sS's'
I17
sS'm'
(lp25995
I1
aI1
aI3
aI1
aI7
aI23
aI109
aI161
aI321
aI499
aI549
aI363
aI2061
aI6519
aI1531
aI1969
aI83845
asS'd'
I12998
sa(dp25996
S'a'
I61741
sS's'
I17
sS'm'
(lp25997
I1
aI3
aI7
aI5
aI17
aI39
aI55
aI59
aI455
aI433
aI601
aI365
aI7987
aI2207
aI3463
aI31755
aI94203
asS'd'
I12999
sa(dp25998
S'a'
I61761
sS's'
I17
sS'm'
(lp25999
I1
aI3
aI5
aI5
aI29
aI61
aI79
aI101
aI125
aI1011
aI867
aI2935
aI3269
aI13601
aI21935
aI50355
aI65883
asS'd'
I13000
sa(dp26000
S'a'
I61779
sS's'
I17
sS'm'
(lp26001
I1
aI1
aI1
aI5
aI3
aI41
aI101
aI107
aI299
aI125
aI81
aI2421
aI2937
aI787
aI19479
aI25803
aI74473
asS'd'
I13001
sa(dp26002
S'a'
I61781
sS's'
I17
sS'm'
(lp26003
I1
aI3
aI3
aI1
aI3
aI15
aI73
aI13
aI167
aI387
aI75
aI601
aI415
aI6927
aI32277
aI16709
aI12477
asS'd'
I13002
sa(dp26004
S'a'
I61782
sS's'
I17
sS'm'
(lp26005
I1
aI1
aI5
aI1
aI19
aI37
aI53
aI45
aI139
aI737
aI159
aI2299
aI6219
aI11613
aI22873
aI18303
aI56875
asS'd'
I13003
sa(dp26006
S'a'
I61797
sS's'
I17
sS'm'
(lp26007
I1
aI1
aI3
aI9
aI23
aI15
aI17
aI37
aI373
aI445
aI1369
aI2997
aI49
aI13901
aI13155
aI37375
aI29063
asS'd'
I13004
sa(dp26008
S'a'
I61809
sS's'
I17
sS'm'
(lp26009
I1
aI3
aI3
aI11
aI17
aI1
aI45
aI91
aI445
aI631
aI1297
aI1907
aI3765
aI13893
aI29379
aI17939
aI36573
asS'd'
I13005
sa(dp26010
S'a'
I61810
sS's'
I17
sS'm'
(lp26011
I1
aI1
aI7
aI13
aI11
aI31
aI101
aI165
aI413
aI305
aI361
aI4019
aI3183
aI2321
aI7819
aI49275
aI101041
asS'd'
I13006
sa(dp26012
S'a'
I61816
sS's'
I17
sS'm'
(lp26013
I1
aI1
aI7
aI1
aI13
aI43
aI125
aI165
aI357
aI293
aI1343
aI2219
aI4189
aI6095
aI28509
aI27763
aI53871
asS'd'
I13007
sa(dp26014
S'a'
I61822
sS's'
I17
sS'm'
(lp26015
I1
aI3
aI3
aI11
aI29
aI33
aI105
aI71
aI297
aI637
aI1493
aI3797
aI5525
aI15093
aI21647
aI57647
aI1467
asS'd'
I13008
sa(dp26016
S'a'
I61849
sS's'
I17
sS'm'
(lp26017
I1
aI1
aI1
aI13
aI19
aI7
aI5
aI141
aI71
aI221
aI923
aI1039
aI4777
aI9481
aI1267
aI55345
aI116061
asS'd'
I13009
sa(dp26018
S'a'
I61876
sS's'
I17
sS'm'
(lp26019
I1
aI1
aI7
aI11
aI19
aI43
aI57
aI243
aI21
aI217
aI1075
aI569
aI3735
aI10477
aI18595
aI34133
aI70391
asS'd'
I13010
sa(dp26020
S'a'
I61893
sS's'
I17
sS'm'
(lp26021
I1
aI3
aI3
aI1
aI21
aI61
aI7
aI135
aI401
aI101
aI321
aI2959
aI7371
aI3303
aI23023
aI28163
aI19833
asS'd'
I13011
sa(dp26022
S'a'
I61905
sS's'
I17
sS'm'
(lp26023
I1
aI1
aI3
aI9
aI31
aI43
aI27
aI243
aI297
aI145
aI663
aI3951
aI4283
aI10421
aI9355
aI30381
aI68317
asS'd'
I13012
sa(dp26024
S'a'
I61908
sS's'
I17
sS'm'
(lp26025
I1
aI3
aI7
aI13
aI29
aI53
aI101
aI253
aI49
aI129
aI831
aI513
aI5567
aI5063
aI157
aI6465
aI90983
asS'd'
I13013
sa(dp26026
S'a'
I61911
sS's'
I17
sS'm'
(lp26027
I1
aI1
aI5
aI15
aI27
aI29
aI3
aI231
aI503
aI173
aI913
aI3971
aI7685
aI9679
aI32243
aI967
aI73195
asS'd'
I13014
sa(dp26028
S'a'
I61912
sS's'
I17
sS'm'
(lp26029
I1
aI1
aI1
aI15
aI19
aI55
aI127
aI3
aI405
aI239
aI1063
aI3473
aI7543
aI4049
aI14509
aI22657
aI5611
asS'd'
I13015
sa(dp26030
S'a'
I61928
sS's'
I17
sS'm'
(lp26031
I1
aI3
aI7
aI1
aI21
aI39
aI61
aI249
aI421
aI401
aI1667
aI1981
aI5503
aI9191
aI24027
aI35049
aI12199
asS'd'
I13016
sa(dp26032
S'a'
I61934
sS's'
I17
sS'm'
(lp26033
I1
aI3
aI5
aI5
aI27
aI1
aI99
aI83
aI287
aI997
aI721
aI1345
aI2197
aI6335
aI4245
aI42575
aI102635
asS'd'
I13017
sa(dp26034
S'a'
I61957
sS's'
I17
sS'm'
(lp26035
I1
aI3
aI3
aI1
aI31
aI7
aI103
aI107
aI387
aI273
aI951
aI2475
aI1275
aI15607
aI2159
aI55083
aI86107
asS'd'
I13018
sa(dp26036
S'a'
I61961
sS's'
I17
sS'm'
(lp26037
I1
aI3
aI3
aI5
aI21
aI59
aI69
aI55
aI121
aI601
aI5
aI1871
aI7161
aI4583
aI23867
aI7933
aI3125
asS'd'
I13019
sa(dp26038
S'a'
I61969
sS's'
I17
sS'm'
(lp26039
I1
aI1
aI1
aI15
aI17
aI41
aI51
aI45
aI383
aI579
aI713
aI275
aI1395
aI11665
aI30521
aI11683
aI126493
asS'd'
I13020
sa(dp26040
S'a'
I61979
sS's'
I17
sS'm'
(lp26041
I1
aI1
aI1
aI15
aI17
aI47
aI15
aI139
aI175
aI283
aI1377
aI827
aI5753
aI8855
aI26437
aI59219
aI105601
asS'd'
I13021
sa(dp26042
S'a'
I61982
sS's'
I17
sS'm'
(lp26043
I1
aI1
aI7
aI11
aI13
aI3
aI27
aI141
aI137
aI851
aI767
aI2575
aI7685
aI11719
aI24401
aI52547
aI127299
asS'd'
I13022
sa(dp26044
S'a'
I62003
sS's'
I17
sS'm'
(lp26045
I1
aI3
aI3
aI9
aI11
aI41
aI75
aI69
aI167
aI897
aI1213
aI3723
aI6773
aI12141
aI28033
aI19695
aI128545
asS'd'
I13023
sa(dp26046
S'a'
I62006
sS's'
I17
sS'm'
(lp26047
I1
aI1
aI5
aI13
aI7
aI61
aI55
aI131
aI465
aI169
aI1669
aI711
aI5901
aI10769
aI11273
aI23809
aI63625
asS'd'
I13024
sa(dp26048
S'a'
I62010
sS's'
I17
sS'm'
(lp26049
I1
aI1
aI5
aI1
aI27
aI25
aI35
aI167
aI83
aI921
aI251
aI2571
aI6723
aI14767
aI26715
aI21699
aI60779
asS'd'
I13025
sa(dp26050
S'a'
I62015
sS's'
I17
sS'm'
(lp26051
I1
aI3
aI1
aI9
aI15
aI5
aI59
aI241
aI405
aI323
aI1917
aI1161
aI6079
aI13443
aI13079
aI58617
aI63381
asS'd'
I13026
sa(dp26052
S'a'
I62020
sS's'
I17
sS'm'
(lp26053
I1
aI3
aI1
aI5
aI9
aI5
aI79
aI123
aI87
aI395
aI667
aI2787
aI3711
aI3613
aI1803
aI17885
aI78975
asS'd'
I13027
sa(dp26054
S'a'
I62024
sS's'
I17
sS'm'
(lp26055
I1
aI3
aI3
aI5
aI17
aI45
aI61
aI107
aI485
aI163
aI33
aI1491
aI7131
aI59
aI27327
aI8043
aI14907
asS'd'
I13028
sa(dp26056
S'a'
I62029
sS's'
I17
sS'm'
(lp26057
I1
aI3
aI1
aI11
aI27
aI37
aI5
aI251
aI115
aI339
aI1621
aI2013
aI3517
aI2213
aI10145
aI47121
aI84485
asS'd'
I13029
sa(dp26058
S'a'
I62032
sS's'
I17
sS'm'
(lp26059
I1
aI3
aI3
aI1
aI9
aI11
aI71
aI25
aI363
aI867
aI1485
aI3897
aI3339
aI7599
aI20777
aI52009
aI127097
asS'd'
I13030
sa(dp26060
S'a'
I62035
sS's'
I17
sS'm'
(lp26061
I1
aI1
aI3
aI9
aI25
aI37
aI29
aI231
aI183
aI315
aI399
aI879
aI6169
aI6355
aI3729
aI9187
aI19405
asS'd'
I13031
sa(dp26062
S'a'
I62038
sS's'
I17
sS'm'
(lp26063
I1
aI3
aI5
aI3
aI31
aI37
aI127
aI71
aI171
aI687
aI1237
aI151
aI7391
aI2395
aI11979
aI23381
aI117879
asS'd'
I13032
sa(dp26064
S'a'
I62047
sS's'
I17
sS'm'
(lp26065
I1
aI1
aI1
aI13
aI13
aI43
aI71
aI235
aI131
aI79
aI1321
aI235
aI2221
aI1133
aI13509
aI12205
aI44771
asS'd'
I13033
sa(dp26066
S'a'
I62068
sS's'
I17
sS'm'
(lp26067
I1
aI3
aI7
aI5
aI29
aI51
aI125
aI191
aI153
aI35
aI1657
aI2141
aI3701
aI7177
aI31723
aI15189
aI55441
asS'd'
I13034
sa(dp26068
S'a'
I62071
sS's'
I17
sS'm'
(lp26069
I1
aI1
aI5
aI1
aI5
aI35
aI13
aI165
aI461
aI255
aI1825
aI1531
aI6195
aI7717
aI973
aI12367
aI71747
asS'd'
I13035
sa(dp26070
S'a'
I62082
sS's'
I17
sS'm'
(lp26071
I1
aI3
aI1
aI9
aI13
aI57
aI25
aI83
aI389
aI405
aI227
aI1037
aI3805
aI15653
aI25365
aI47991
aI54315
asS'd'
I13036
sa(dp26072
S'a'
I62088
sS's'
I17
sS'm'
(lp26073
I1
aI1
aI1
aI13
aI17
aI55
aI113
aI151
aI145
aI951
aI1849
aI2205
aI6513
aI7845
aI7947
aI59429
aI44911
asS'd'
I13037
sa(dp26074
S'a'
I62096
sS's'
I17
sS'm'
(lp26075
I1
aI3
aI5
aI3
aI25
aI9
aI99
aI159
aI183
aI445
aI153
aI3053
aI2537
aI1787
aI19029
aI60047
aI128203
asS'd'
I13038
sa(dp26076
S'a'
I62108
sS's'
I17
sS'm'
(lp26077
I1
aI1
aI5
aI5
aI31
aI37
aI13
aI11
aI271
aI491
aI1141
aI1827
aI3751
aI9471
aI7579
aI35969
aI95245
asS'd'
I13039
sa(dp26078
S'a'
I62124
sS's'
I17
sS'm'
(lp26079
I1
aI3
aI3
aI15
aI1
aI43
aI9
aI109
aI13
aI991
aI345
aI1577
aI947
aI3197
aI16747
aI8127
aI116937
asS'd'
I13040
sa(dp26080
S'a'
I62142
sS's'
I17
sS'm'
(lp26081
I1
aI3
aI3
aI15
aI11
aI17
aI103
aI89
aI33
aI741
aI1855
aI2879
aI3899
aI9535
aI15119
aI56165
aI86055
asS'd'
I13041
sa(dp26082
S'a'
I62147
sS's'
I17
sS'm'
(lp26083
I1
aI3
aI5
aI1
aI31
aI41
aI57
aI205
aI69
aI163
aI1383
aI2087
aI6483
aI6281
aI32079
aI40825
aI28709
asS'd'
I13042
sa(dp26084
S'a'
I62153
sS's'
I17
sS'm'
(lp26085
I1
aI1
aI1
aI13
aI23
aI57
aI103
aI247
aI421
aI773
aI1733
aI3249
aI6681
aI9675
aI11669
aI51673
aI86189
asS'd'
I13043
sa(dp26086
S'a'
I62159
sS's'
I17
sS'm'
(lp26087
I1
aI1
aI5
aI15
aI7
aI37
aI63
aI123
aI77
aI941
aI277
aI1061
aI803
aI2135
aI15745
aI47413
aI73843
asS'd'
I13044
sa(dp26088
S'a'
I62161
sS's'
I17
sS'm'
(lp26089
I1
aI1
aI3
aI5
aI19
aI15
aI35
aI59
aI321
aI527
aI1669
aI2929
aI513
aI4453
aI20521
aI19781
aI115501
asS'd'
I13045
sa(dp26090
S'a'
I62171
sS's'
I17
sS'm'
(lp26091
I1
aI1
aI7
aI3
aI23
aI1
aI99
aI251
aI129
aI271
aI1555
aI1191
aI2445
aI11533
aI11921
aI19131
aI80653
asS'd'
I13046
sa(dp26092
S'a'
I62189
sS's'
I17
sS'm'
(lp26093
I1
aI1
aI7
aI15
aI15
aI33
aI79
aI89
aI113
aI517
aI1655
aI43
aI6255
aI15415
aI19559
aI63309
aI16857
asS'd'
I13047
sa(dp26094
S'a'
I62207
sS's'
I17
sS'm'
(lp26095
I1
aI3
aI5
aI1
aI13
aI61
aI87
aI159
aI65
aI875
aI163
aI663
aI7651
aI8775
aI32505
aI39313
aI83331
asS'd'
I13048
sa(dp26096
S'a'
I62227
sS's'
I17
sS'm'
(lp26097
I1
aI1
aI7
aI15
aI27
aI19
aI63
aI117
aI427
aI233
aI1243
aI755
aI3201
aI5153
aI31959
aI64545
aI69219
asS'd'
I13049
sa(dp26098
S'a'
I62230
sS's'
I17
sS'm'
(lp26099
I1
aI1
aI5
aI3
aI3
aI27
aI15
aI11
aI427
aI431
aI107
aI2433
aI6923
aI7503
aI31347
aI64849
aI14541
asS'd'
I13050
sa(dp26100
S'a'
I62234
sS's'
I17
sS'm'
(lp26101
I1
aI1
aI3
aI11
aI7
aI23
aI53
aI253
aI483
aI63
aI1749
aI2989
aI5219
aI7361
aI6423
aI1503
aI95431
asS'd'
I13051
sa(dp26102
S'a'
I62236
sS's'
I17
sS'm'
(lp26103
I1
aI1
aI5
aI9
aI1
aI19
aI23
aI25
aI301
aI665
aI1457
aI2423
aI6623
aI9771
aI2755
aI8963
aI51037
asS'd'
I13052
sa(dp26104
S'a'
I62239
sS's'
I17
sS'm'
(lp26105
I1
aI3
aI3
aI7
aI21
aI1
aI3
aI131
aI377
aI897
aI15
aI437
aI4075
aI7669
aI31529
aI64123
aI101249
asS'd'
I13053
sa(dp26106
S'a'
I62257
sS's'
I17
sS'm'
(lp26107
I1
aI3
aI5
aI3
aI31
aI41
aI99
aI27
aI397
aI393
aI1895
aI2249
aI3925
aI6393
aI2839
aI375
aI56721
asS'd'
I13054
sa(dp26108
S'a'
I62270
sS's'
I17
sS'm'
(lp26109
I1
aI3
aI7
aI15
aI1
aI45
aI65
aI113
aI85
aI557
aI857
aI2281
aI1395
aI2055
aI2405
aI34541
aI63719
asS'd'
I13055
sa(dp26110
S'a'
I62278
sS's'
I17
sS'm'
(lp26111
I1
aI3
aI1
aI15
aI7
aI43
aI21
aI29
aI15
aI229
aI1287
aI3005
aI339
aI5833
aI21867
aI21643
aI37557
asS'd'
I13056
sa(dp26112
S'a'
I62287
sS's'
I17
sS'm'
(lp26113
I1
aI3
aI7
aI3
aI3
aI51
aI67
aI119
aI423
aI539
aI1995
aI4039
aI2999
aI2787
aI29327
aI62687
aI11893
asS'd'
I13057
sa(dp26114
S'a'
I62295
sS's'
I17
sS'm'
(lp26115
I1
aI3
aI7
aI3
aI25
aI23
aI85
aI11
aI105
aI523
aI889
aI2983
aI2031
aI2049
aI16119
aI41925
aI38345
asS'd'
I13058
sa(dp26116
S'a'
I62301
sS's'
I17
sS'm'
(lp26117
I1
aI3
aI7
aI3
aI13
aI63
aI59
aI65
aI183
aI695
aI293
aI3301
aI7895
aI13915
aI25847
aI22819
aI92189
asS'd'
I13059
sa(dp26118
S'a'
I62306
sS's'
I17
sS'm'
(lp26119
I1
aI1
aI3
aI3
aI7
aI27
aI101
aI229
aI435
aI227
aI1759
aI1275
aI5781
aI6079
aI25125
aI64833
aI69577
asS'd'
I13060
sa(dp26120
S'a'
I62312
sS's'
I17
sS'm'
(lp26121
I1
aI1
aI3
aI1
aI29
aI1
aI61
aI45
aI193
aI441
aI687
aI841
aI4491
aI10683
aI13989
aI60461
aI78071
asS'd'
I13061
sa(dp26122
S'a'
I62317
sS's'
I17
sS'm'
(lp26123
I1
aI3
aI1
aI9
aI5
aI33
aI99
aI229
aI181
aI675
aI1629
aI1855
aI4719
aI9585
aI8059
aI26363
aI31161
asS'd'
I13062
sa(dp26124
S'a'
I62330
sS's'
I17
sS'm'
(lp26125
I1
aI3
aI1
aI11
aI11
aI37
aI79
aI53
aI163
aI49
aI1173
aI1715
aI8087
aI6535
aI14985
aI24069
aI118597
asS'd'
I13063
sa(dp26126
S'a'
I62342
sS's'
I17
sS'm'
(lp26127
I1
aI1
aI7
aI15
aI9
aI59
aI123
aI79
aI237
aI131
aI1693
aI2525
aI6339
aI9843
aI24309
aI24969
aI37645
asS'd'
I13064
sa(dp26128
S'a'
I62359
sS's'
I17
sS'm'
(lp26129
I1
aI3
aI3
aI7
aI19
aI49
aI85
aI133
aI415
aI239
aI555
aI2581
aI6523
aI2733
aI19665
aI19989
aI105585
asS'd'
I13065
sa(dp26130
S'a'
I62365
sS's'
I17
sS'm'
(lp26131
I1
aI3
aI3
aI7
aI23
aI37
aI31
aI121
aI59
aI7
aI2031
aI2893
aI6429
aI10305
aI21221
aI20105
aI38879
asS'd'
I13066
sa(dp26132
S'a'
I62366
sS's'
I17
sS'm'
(lp26133
I1
aI3
aI1
aI13
aI23
aI21
aI93
aI93
aI343
aI641
aI411
aI971
aI1777
aI2135
aI22895
aI9055
aI114457
asS'd'
I13067
sa(dp26134
S'a'
I62370
sS's'
I17
sS'm'
(lp26135
I1
aI3
aI5
aI3
aI15
aI33
aI23
aI7
aI59
aI413
aI277
aI3551
aI7737
aI2285
aI7951
aI5013
aI94469
asS'd'
I13068
sa(dp26136
S'a'
I62375
sS's'
I17
sS'm'
(lp26137
I1
aI3
aI1
aI15
aI25
aI1
aI109
aI245
aI153
aI187
aI1099
aI1071
aI145
aI6735
aI10327
aI3921
aI62123
asS'd'
I13069
sa(dp26138
S'a'
I62376
sS's'
I17
sS'm'
(lp26139
I1
aI1
aI7
aI11
aI11
aI53
aI51
aI123
aI277
aI281
aI1763
aI3161
aI7639
aI14515
aI29725
aI3919
aI5525
asS'd'
I13070
sa(dp26140
S'a'
I62387
sS's'
I17
sS'm'
(lp26141
I1
aI3
aI3
aI15
aI27
aI47
aI109
aI121
aI317
aI221
aI187
aI617
aI1331
aI5401
aI861
aI62465
aI9227
asS'd'
I13071
sa(dp26142
S'a'
I62404
sS's'
I17
sS'm'
(lp26143
I1
aI3
aI3
aI15
aI27
aI25
aI101
aI111
aI469
aI85
aI1285
aI1621
aI5393
aI1367
aI17115
aI35141
aI126989
asS'd'
I13072
sa(dp26144
S'a'
I62411
sS's'
I17
sS'm'
(lp26145
I1
aI3
aI5
aI1
aI15
aI23
aI25
aI249
aI69
aI17
aI1103
aI2603
aI59
aI15637
aI22051
aI29243
aI53113
asS'd'
I13073
sa(dp26146
S'a'
I62435
sS's'
I17
sS'm'
(lp26147
I1
aI3
aI1
aI9
aI17
aI49
aI73
aI13
aI207
aI963
aI379
aI3561
aI6447
aI7895
aI18651
aI8109
aI3943
asS'd'
I13074
sa(dp26148
S'a'
I62441
sS's'
I17
sS'm'
(lp26149
I1
aI3
aI5
aI11
aI7
aI41
aI55
aI85
aI481
aI831
aI593
aI4093
aI1151
aI12353
aI24231
aI46091
aI80967
asS'd'
I13075
sa(dp26150
S'a'
I62442
sS's'
I17
sS'm'
(lp26151
I1
aI3
aI7
aI7
aI5
aI39
aI47
aI187
aI427
aI1007
aI813
aI3617
aI6063
aI12981
aI18573
aI34061
aI85741
asS'd'
I13076
sa(dp26152
S'a'
I62452
sS's'
I17
sS'm'
(lp26153
I1
aI3
aI3
aI11
aI9
aI17
aI29
aI141
aI341
aI485
aI1075
aI4067
aI7247
aI11295
aI31803
aI18347
aI54985
asS'd'
I13077
sa(dp26154
S'a'
I62459
sS's'
I17
sS'm'
(lp26155
I1
aI1
aI3
aI7
aI17
aI25
aI7
aI29
aI355
aI35
aI1753
aI669
aI4123
aI4293
aI22875
aI36677
aI61201
asS'd'
I13078
sa(dp26156
S'a'
I62461
sS's'
I17
sS'm'
(lp26157
I1
aI1
aI5
aI9
aI13
aI45
aI29
aI153
aI169
aI387
aI1275
aI3161
aI4937
aI5331
aI16203
aI43925
aI129231
asS'd'
I13079
sa(dp26158
S'a'
I62473
sS's'
I17
sS'm'
(lp26159
I1
aI3
aI3
aI9
aI19
aI27
aI109
aI95
aI499
aI929
aI1627
aI3215
aI6097
aI15837
aI5655
aI29877
aI122513
asS'd'
I13080
sa(dp26160
S'a'
I62474
sS's'
I17
sS'm'
(lp26161
I1
aI3
aI7
aI11
aI1
aI25
aI15
aI41
aI65
aI411
aI289
aI883
aI5069
aI8405
aI11159
aI57357
aI114253
asS'd'
I13081
sa(dp26162
S'a'
I62493
sS's'
I17
sS'm'
(lp26163
I1
aI1
aI3
aI11
aI31
aI57
aI39
aI89
aI77
aI321
aI1667
aI871
aI6429
aI1005
aI18905
aI13877
aI9305
asS'd'
I13082
sa(dp26164
S'a'
I62510
sS's'
I17
sS'm'
(lp26165
I1
aI1
aI7
aI7
aI27
aI57
aI23
aI37
aI281
aI625
aI1871
aI565
aI5979
aI13925
aI22591
aI2375
aI8577
asS'd'
I13083
sa(dp26166
S'a'
I62518
sS's'
I17
sS'm'
(lp26167
I1
aI1
aI1
aI7
aI31
aI35
aI91
aI221
aI495
aI811
aI1321
aI2235
aI4287
aI3009
aI5745
aI35013
aI93715
asS'd'
I13084
sa(dp26168
S'a'
I62524
sS's'
I17
sS'm'
(lp26169
I1
aI1
aI7
aI3
aI11
aI53
aI17
aI13
aI319
aI225
aI117
aI3365
aI537
aI5249
aI14219
aI23879
aI4321
asS'd'
I13085
sa(dp26170
S'a'
I62549
sS's'
I17
sS'm'
(lp26171
I1
aI3
aI5
aI1
aI31
aI57
aI35
aI95
aI257
aI933
aI471
aI627
aI6733
aI8707
aI25173
aI44291
aI105041
asS'd'
I13086
sa(dp26172
S'a'
I62556
sS's'
I17
sS'm'
(lp26173
I1
aI1
aI3
aI3
aI31
aI53
aI69
aI19
aI277
aI669
aI497
aI3957
aI2781
aI14107
aI27741
aI53519
aI41057
asS'd'
I13087
sa(dp26174
S'a'
I62565
sS's'
I17
sS'm'
(lp26175
I1
aI1
aI3
aI15
aI11
aI21
aI11
aI25
aI257
aI665
aI491
aI2119
aI3893
aI14401
aI29147
aI3369
aI116569
asS'd'
I13088
sa(dp26176
S'a'
I62566
sS's'
I17
sS'm'
(lp26177
I1
aI3
aI1
aI1
aI13
aI49
aI31
aI231
aI217
aI711
aI1987
aI1487
aI7073
aI473
aI28781
aI51283
aI86049
asS'd'
I13089
sa(dp26178
S'a'
I62580
sS's'
I17
sS'm'
(lp26179
I1
aI1
aI1
aI5
aI23
aI31
aI119
aI115
aI381
aI179
aI1725
aI2323
aI8013
aI15191
aI13255
aI57813
aI95437
asS'd'
I13090
sa(dp26180
S'a'
I62584
sS's'
I17
sS'm'
(lp26181
I1
aI1
aI3
aI15
aI15
aI37
aI83
aI223
aI259
aI605
aI2013
aI4089
aI395
aI2063
aI11735
aI51931
aI74677
asS'd'
I13091
sa(dp26182
S'a'
I62589
sS's'
I17
sS'm'
(lp26183
I1
aI1
aI7
aI3
aI1
aI61
aI107
aI169
aI213
aI789
aI425
aI2309
aI225
aI1305
aI20697
aI26281
aI60129
asS'd'
I13092
sa(dp26184
S'a'
I62596
sS's'
I17
sS'm'
(lp26185
I1
aI1
aI5
aI15
aI27
aI15
aI69
aI169
aI289
aI931
aI1491
aI3711
aI6875
aI7723
aI21103
aI31795
aI53955
asS'd'
I13093
sa(dp26186
S'a'
I62608
sS's'
I17
sS'm'
(lp26187
I1
aI1
aI1
aI3
aI3
aI43
aI49
aI205
aI247
aI817
aI2037
aI2305
aI7935
aI255
aI18835
aI49423
aI90727
asS'd'
I13094
sa(dp26188
S'a'
I62636
sS's'
I17
sS'm'
(lp26189
I1
aI3
aI7
aI9
aI17
aI3
aI95
aI239
aI431
aI665
aI1271
aI3559
aI5703
aI14607
aI9723
aI11807
aI122937
asS'd'
I13095
sa(dp26190
S'a'
I62642
sS's'
I17
sS'm'
(lp26191
I1
aI3
aI5
aI13
aI5
aI15
aI13
aI111
aI375
aI895
aI833
aI813
aI5451
aI13841
aI1321
aI25273
aI25443
asS'd'
I13096
sa(dp26192
S'a'
I62651
sS's'
I17
sS'm'
(lp26193
I1
aI1
aI3
aI1
aI11
aI49
aI3
aI97
aI467
aI631
aI51
aI3577
aI1777
aI15965
aI6837
aI38827
aI68627
asS'd'
I13097
sa(dp26194
S'a'
I62654
sS's'
I17
sS'm'
(lp26195
I1
aI1
aI7
aI1
aI3
aI11
aI73
aI155
aI77
aI623
aI811
aI337
aI6837
aI10925
aI2097
aI28325
aI97487
asS'd'
I13098
sa(dp26196
S'a'
I62659
sS's'
I17
sS'm'
(lp26197
I1
aI1
aI1
aI3
aI29
aI35
aI9
aI215
aI415
aI143
aI1837
aI3723
aI73
aI11305
aI23187
aI19995
aI52987
asS'd'
I13099
sa(dp26198
S'a'
I62666
sS's'
I17
sS'm'
(lp26199
I1
aI1
aI7
aI1
aI25
aI39
aI35
aI67
aI245
aI295
aI1143
aI2043
aI1049
aI629
aI14111
aI62893
aI86899
asS'd'
I13100
sa(dp26200
S'a'
I62680
sS's'
I17
sS'm'
(lp26201
I1
aI3
aI7
aI5
aI3
aI41
aI123
aI97
aI241
aI743
aI259
aI3163
aI2289
aI6363
aI24033
aI10789
aI44117
asS'd'
I13101
sa(dp26202
S'a'
I62692
sS's'
I17
sS'm'
(lp26203
I1
aI3
aI1
aI7
aI25
aI33
aI33
aI17
aI359
aI567
aI901
aI3595
aI179
aI8671
aI841
aI24787
aI4367
asS'd'
I13102
sa(dp26204
S'a'
I62701
sS's'
I17
sS'm'
(lp26205
I1
aI3
aI1
aI13
aI5
aI13
aI57
aI185
aI321
aI727
aI789
aI3081
aI5345
aI9721
aI32029
aI11663
aI55695
asS'd'
I13103
sa(dp26206
S'a'
I62702
sS's'
I17
sS'm'
(lp26207
I1
aI1
aI7
aI7
aI5
aI51
aI85
aI255
aI329
aI263
aI297
aI1687
aI6799
aI10973
aI8265
aI19615
aI115333
asS'd'
I13104
sa(dp26208
S'a'
I62714
sS's'
I17
sS'm'
(lp26209
I1
aI1
aI1
aI5
aI29
aI27
aI55
aI167
aI465
aI73
aI661
aI137
aI7831
aI2571
aI15373
aI64223
aI27335
asS'd'
I13105
sa(dp26210
S'a'
I62716
sS's'
I17
sS'm'
(lp26211
I1
aI3
aI7
aI13
aI5
aI23
aI77
aI15
aI345
aI21
aI1729
aI3231
aI967
aI12573
aI31415
aI24249
aI110525
asS'd'
I13106
sa(dp26212
S'a'
I62733
sS's'
I17
sS'm'
(lp26213
I1
aI1
aI7
aI9
aI31
aI37
aI41
aI119
aI169
aI891
aI1845
aI2139
aI1747
aI1147
aI21983
aI11617
aI25963
asS'd'
I13107
sa(dp26214
S'a'
I62762
sS's'
I17
sS'm'
(lp26215
I1
aI3
aI3
aI7
aI23
aI5
aI1
aI11
aI95
aI795
aI1371
aI2631
aI3241
aI6935
aI17353
aI25013
aI89765
asS'd'
I13108
sa(dp26216
S'a'
I62769
sS's'
I17
sS'm'
(lp26217
I1
aI3
aI1
aI7
aI19
aI3
aI121
aI19
aI389
aI117
aI1905
aI3135
aI7601
aI12541
aI20855
aI38613
aI15005
asS'd'
I13109
sa(dp26218
S'a'
I62770
sS's'
I17
sS'm'
(lp26219
I1
aI3
aI3
aI5
aI17
aI43
aI91
aI99
aI113
aI545
aI1955
aI37
aI3411
aI15173
aI24961
aI28761
aI15245
asS'd'
I13110
sa(dp26220
S'a'
I62787
sS's'
I17
sS'm'
(lp26221
I1
aI3
aI3
aI13
aI25
aI9
aI83
aI17
aI17
aI271
aI1133
aI3851
aI4607
aI11017
aI14867
aI20677
aI42881
asS'd'
I13111
sa(dp26222
S'a'
I62794
sS's'
I17
sS'm'
(lp26223
I1
aI1
aI5
aI15
aI5
aI9
aI99
aI179
aI263
aI623
aI441
aI2577
aI189
aI11595
aI21505
aI27215
aI54081
asS'd'
I13112
sa(dp26224
S'a'
I62801
sS's'
I17
sS'm'
(lp26225
I1
aI1
aI7
aI1
aI1
aI63
aI123
aI119
aI245
aI467
aI169
aI3075
aI909
aI3581
aI14571
aI33071
aI6261
asS'd'
I13113
sa(dp26226
S'a'
I62807
sS's'
I17
sS'm'
(lp26227
I1
aI1
aI1
aI13
aI9
aI35
aI47
aI161
aI47
aI893
aI57
aI703
aI3373
aI4167
aI26555
aI51265
aI1391
asS'd'
I13114
sa(dp26228
S'a'
I62808
sS's'
I17
sS'm'
(lp26229
I1
aI3
aI1
aI13
aI9
aI61
aI9
aI5
aI47
aI259
aI579
aI113
aI355
aI14539
aI25757
aI10119
aI96869
asS'd'
I13115
sa(dp26230
S'a'
I62813
sS's'
I17
sS'm'
(lp26231
I1
aI1
aI5
aI11
aI3
aI5
aI61
aI231
aI291
aI21
aI1711
aI2981
aI4727
aI14287
aI12149
aI40275
aI51809
asS'd'
I13116
sa(dp26232
S'a'
I62817
sS's'
I17
sS'm'
(lp26233
I1
aI3
aI5
aI3
aI21
aI5
aI87
aI251
aI373
aI679
aI949
aI1023
aI5183
aI14549
aI4135
aI54927
aI20369
asS'd'
I13117
sa(dp26234
S'a'
I62818
sS's'
I17
sS'm'
(lp26235
I1
aI3
aI3
aI11
aI7
aI43
aI127
aI97
aI469
aI81
aI1843
aI3955
aI125
aI8607
aI27185
aI50761
aI122753
asS'd'
I13118
sa(dp26236
S'a'
I62832
sS's'
I17
sS'm'
(lp26237
I1
aI3
aI5
aI5
aI25
aI61
aI1
aI55
aI333
aI949
aI1005
aI1051
aI6291
aI8343
aI9627
aI37739
aI116911
asS'd'
I13119
sa(dp26238
S'a'
I62841
sS's'
I17
sS'm'
(lp26239
I1
aI3
aI3
aI13
aI21
aI9
aI67
aI225
aI179
aI837
aI2009
aI3171
aI217
aI5629
aI23451
aI63171
aI53225
asS'd'
I13120
sa(dp26240
S'a'
I62857
sS's'
I17
sS'm'
(lp26241
I1
aI3
aI7
aI1
aI23
aI15
aI91
aI163
aI351
aI883
aI579
aI3923
aI2641
aI12735
aI24955
aI1131
aI65119
asS'd'
I13121
sa(dp26242
S'a'
I62860
sS's'
I17
sS'm'
(lp26243
I1
aI1
aI1
aI11
aI29
aI5
aI113
aI217
aI171
aI535
aI913
aI2419
aI3843
aI12365
aI8287
aI27367
aI57369
asS'd'
I13122
sa(dp26244
S'a'
I62871
sS's'
I17
sS'm'
(lp26245
I1
aI1
aI5
aI11
aI9
aI41
aI57
aI243
aI123
aI159
aI1517
aI2653
aI4307
aI4243
aI2801
aI43131
aI18435
asS'd'
I13123
sa(dp26246
S'a'
I62882
sS's'
I17
sS'm'
(lp26247
I1
aI1
aI7
aI9
aI23
aI59
aI83
aI159
aI57
aI723
aI1635
aI7
aI1463
aI121
aI541
aI7657
aI83917
asS'd'
I13124
sa(dp26248
S'a'
I62888
sS's'
I17
sS'm'
(lp26249
I1
aI3
aI7
aI7
aI23
aI27
aI125
aI103
aI247
aI1019
aI1063
aI3979
aI8085
aI6449
aI12443
aI63427
aI106235
asS'd'
I13125
sa(dp26250
S'a'
I62913
sS's'
I17
sS'm'
(lp26251
I1
aI1
aI7
aI5
aI9
aI31
aI23
aI83
aI503
aI605
aI1731
aI3341
aI7427
aI14571
aI5981
aI39043
aI42965
asS'd'
I13126
sa(dp26252
S'a'
I62919
sS's'
I17
sS'm'
(lp26253
I1
aI3
aI5
aI1
aI17
aI49
aI109
aI171
aI301
aI951
aI1879
aI1317
aI457
aI8085
aI6803
aI62521
aI67871
asS'd'
I13127
sa(dp26254
S'a'
I62920
sS's'
I17
sS'm'
(lp26255
I1
aI1
aI7
aI11
aI11
aI27
aI1
aI71
aI335
aI137
aI265
aI1267
aI6399
aI14823
aI925
aI49895
aI19731
asS'd'
I13128
sa(dp26256
S'a'
I62925
sS's'
I17
sS'm'
(lp26257
I1
aI3
aI1
aI13
aI3
aI35
aI75
aI253
aI211
aI483
aI1495
aI3695
aI3033
aI14643
aI1861
aI51269
aI32655
asS'd'
I13129
sa(dp26258
S'a'
I62933
sS's'
I17
sS'm'
(lp26259
I1
aI3
aI7
aI1
aI5
aI17
aI63
aI1
aI83
aI435
aI2007
aI2023
aI57
aI8639
aI27067
aI4039
aI1955
asS'd'
I13130
sa(dp26260
S'a'
I62938
sS's'
I17
sS'm'
(lp26261
I1
aI3
aI5
aI15
aI27
aI51
aI59
aI47
aI77
aI131
aI507
aI559
aI645
aI16067
aI20989
aI15565
aI39925
asS'd'
I13131
sa(dp26262
S'a'
I62940
sS's'
I17
sS'm'
(lp26263
I1
aI3
aI3
aI5
aI11
aI15
aI63
aI121
aI39
aI1019
aI1027
aI2821
aI3297
aI13769
aI18587
aI14199
aI82251
asS'd'
I13132
sa(dp26264
S'a'
I62949
sS's'
I17
sS'm'
(lp26265
I1
aI1
aI5
aI1
aI31
aI11
aI89
aI75
aI147
aI1007
aI917
aI3519
aI5097
aI5695
aI15185
aI14819
aI38597
asS'd'
I13133
sa(dp26266
S'a'
I62956
sS's'
I17
sS'm'
(lp26267
I1
aI3
aI3
aI3
aI15
aI7
aI127
aI55
aI83
aI887
aI1901
aI75
aI639
aI713
aI13631
aI27447
aI106707
asS'd'
I13134
sa(dp26268
S'a'
I62971
sS's'
I17
sS'm'
(lp26269
I1
aI3
aI3
aI15
aI27
aI25
aI85
aI163
aI187
aI959
aI815
aI1403
aI6129
aI6515
aI31597
aI28307
aI30077
asS'd'
I13135
sa(dp26270
S'a'
I62978
sS's'
I17
sS'm'
(lp26271
I1
aI3
aI1
aI13
aI5
aI19
aI117
aI89
aI11
aI489
aI845
aI2899
aI3695
aI3279
aI22355
aI38759
aI85849
asS'd'
I13136
sa(dp26272
S'a'
I62990
sS's'
I17
sS'm'
(lp26273
I1
aI3
aI1
aI7
aI25
aI59
aI109
aI185
aI87
aI591
aI825
aI1119
aI7439
aI5451
aI17959
aI51299
aI76693
asS'd'
I13137
sa(dp26274
S'a'
I62995
sS's'
I17
sS'm'
(lp26275
I1
aI1
aI7
aI5
aI25
aI29
aI115
aI249
aI185
aI529
aI593
aI103
aI1739
aI4769
aI25925
aI3317
aI102445
asS'd'
I13138
sa(dp26276
S'a'
I62997
sS's'
I17
sS'm'
(lp26277
I1
aI1
aI3
aI1
aI3
aI35
aI19
aI255
aI279
aI295
aI1075
aI2817
aI3513
aI7501
aI15885
aI21653
aI113447
asS'd'
I13139
sa(dp26278
S'a'
I63004
sS's'
I17
sS'm'
(lp26279
I1
aI3
aI1
aI5
aI27
aI1
aI21
aI137
aI303
aI981
aI631
aI2339
aI397
aI13075
aI28815
aI50925
aI44379
asS'd'
I13140
sa(dp26280
S'a'
I63011
sS's'
I17
sS'm'
(lp26281
I1
aI1
aI7
aI7
aI31
aI31
aI59
aI129
aI105
aI181
aI1041
aI2685
aI1061
aI1721
aI30365
aI6873
aI30011
asS'd'
I13141
sa(dp26282
S'a'
I63032
sS's'
I17
sS'm'
(lp26283
I1
aI1
aI3
aI1
aI19
aI31
aI125
aI39
aI9
aI631
aI1239
aI1061
aI6313
aI9639
aI5991
aI27293
aI84635
asS'd'
I13142
sa(dp26284
S'a'
I63038
sS's'
I17
sS'm'
(lp26285
I1
aI3
aI3
aI11
aI17
aI59
aI17
aI241
aI195
aI175
aI1845
aI251
aI3323
aI13399
aI20493
aI15241
aI69303
asS'd'
I13143
sa(dp26286
S'a'
I63067
sS's'
I17
sS'm'
(lp26287
I1
aI1
aI5
aI3
aI9
aI19
aI59
aI25
aI49
aI359
aI263
aI4045
aI1527
aI6703
aI555
aI26413
aI42757
asS'd'
I13144
sa(dp26288
S'a'
I63069
sS's'
I17
sS'm'
(lp26289
I1
aI1
aI5
aI9
aI23
aI5
aI7
aI223
aI247
aI407
aI1079
aI1069
aI3417
aI14795
aI5015
aI2965
aI99059
asS'd'
I13145
sa(dp26290
S'a'
I63076
sS's'
I17
sS'm'
(lp26291
I1
aI3
aI7
aI5
aI27
aI47
aI9
aI37
aI47
aI181
aI819
aI2049
aI2643
aI9231
aI8173
aI33495
aI91321
asS'd'
I13146
sa(dp26292
S'a'
I63083
sS's'
I17
sS'm'
(lp26293
I1
aI3
aI5
aI11
aI5
aI27
aI5
aI237
aI349
aI653
aI1443
aI137
aI7969
aI5961
aI24749
aI37523
aI88921
asS'd'
I13147
sa(dp26294
S'a'
I63088
sS's'
I17
sS'm'
(lp26295
I1
aI3
aI7
aI13
aI11
aI51
aI49
aI71
aI339
aI195
aI1239
aI3479
aI2771
aI15217
aI23729
aI7839
aI32633
asS'd'
I13148
sa(dp26296
S'a'
I63094
sS's'
I17
sS'm'
(lp26297
I1
aI1
aI5
aI5
aI5
aI13
aI103
aI185
aI13
aI273
aI1793
aI761
aI5327
aI8659
aI32599
aI38181
aI121115
asS'd'
I13149
sa(dp26298
S'a'
I63097
sS's'
I17
sS'm'
(lp26299
I1
aI3
aI7
aI15
aI17
aI55
aI69
aI151
aI231
aI421
aI1679
aI3657
aI8001
aI12599
aI13761
aI13517
aI130199
asS'd'
I13150
sa(dp26300
S'a'
I63100
sS's'
I17
sS'm'
(lp26301
I1
aI1
aI5
aI3
aI5
aI15
aI15
aI61
aI489
aI219
aI925
aI2329
aI3415
aI10779
aI31297
aI63805
aI13375
asS'd'
I13151
sa(dp26302
S'a'
I63104
sS's'
I17
sS'm'
(lp26303
I1
aI1
aI7
aI9
aI7
aI11
aI87
aI45
aI39
aI885
aI87
aI1877
aI8135
aI1247
aI25685
aI23631
aI65579
asS'd'
I13152
sa(dp26304
S'a'
I63114
sS's'
I17
sS'm'
(lp26305
I1
aI3
aI3
aI7
aI1
aI17
aI31
aI75
aI455
aI535
aI509
aI2151
aI1737
aI7579
aI12727
aI25139
aI32659
asS'd'
I13153
sa(dp26306
S'a'
I63116
sS's'
I17
sS'm'
(lp26307
I1
aI3
aI5
aI7
aI15
aI27
aI111
aI145
aI99
aI767
aI167
aI3391
aI2155
aI7895
aI3405
aI47451
aI65185
asS'd'
I13154
sa(dp26308
S'a'
I63122
sS's'
I17
sS'm'
(lp26309
I1
aI1
aI1
aI3
aI23
aI31
aI15
aI53
aI59
aI787
aI431
aI2691
aI71
aI2843
aI13469
aI54029
aI2233
asS'd'
I13155
sa(dp26310
S'a'
I63128
sS's'
I17
sS'm'
(lp26311
I1
aI1
aI5
aI1
aI5
aI39
aI57
aI31
aI75
aI507
aI811
aI2747
aI317
aI13545
aI7395
aI65161
aI87987
asS'd'
I13156
sa(dp26312
S'a'
I63149
sS's'
I17
sS'm'
(lp26313
I1
aI3
aI5
aI5
aI13
aI17
aI11
aI89
aI371
aI337
aI913
aI2775
aI27
aI4923
aI24013
aI62955
aI65185
asS'd'
I13157
sa(dp26314
S'a'
I63150
sS's'
I17
sS'm'
(lp26315
I1
aI3
aI3
aI7
aI9
aI27
aI91
aI187
aI17
aI443
aI807
aI853
aI6243
aI12351
aI8123
aI4203
aI61021
asS'd'
I13158
sa(dp26316
S'a'
I63157
sS's'
I17
sS'm'
(lp26317
I1
aI1
aI1
aI5
aI9
aI33
aI101
aI211
aI205
aI701
aI1289
aI1253
aI653
aI8591
aI13009
aI48525
aI77051
asS'd'
I13159
sa(dp26318
S'a'
I63167
sS's'
I17
sS'm'
(lp26319
I1
aI3
aI5
aI11
aI5
aI19
aI1
aI67
aI259
aI355
aI15
aI2169
aI6785
aI2019
aI8675
aI1019
aI85903
asS'd'
I13160
sa(dp26320
S'a'
I63187
sS's'
I17
sS'm'
(lp26321
I1
aI3
aI7
aI5
aI27
aI31
aI103
aI163
aI369
aI685
aI659
aI2009
aI5819
aI10437
aI17311
aI35083
aI122125
asS'd'
I13161
sa(dp26322
S'a'
I63200
sS's'
I17
sS'm'
(lp26323
I1
aI3
aI5
aI7
aI19
aI13
aI93
aI113
aI377
aI359
aI1697
aI4063
aI4379
aI9321
aI7335
aI25491
aI85939
asS'd'
I13162
sa(dp26324
S'a'
I63220
sS's'
I17
sS'm'
(lp26325
I1
aI3
aI3
aI5
aI7
aI25
aI41
aI225
aI355
aI257
aI743
aI2067
aI7627
aI14317
aI7385
aI25187
aI63103
asS'd'
I13163
sa(dp26326
S'a'
I63223
sS's'
I17
sS'm'
(lp26327
I1
aI1
aI7
aI7
aI17
aI43
aI75
aI1
aI95
aI547
aI1937
aI2263
aI1709
aI5067
aI22651
aI55733
aI44619
asS'd'
I13164
sa(dp26328
S'a'
I63229
sS's'
I17
sS'm'
(lp26329
I1
aI1
aI7
aI3
aI5
aI27
aI45
aI23
aI107
aI547
aI1733
aI1169
aI6709
aI861
aI4439
aI55381
aI96447
asS'd'
I13165
sa(dp26330
S'a'
I63235
sS's'
I17
sS'm'
(lp26331
I1
aI1
aI7
aI11
aI25
aI23
aI127
aI159
aI489
aI945
aI843
aI3715
aI5215
aI2131
aI9681
aI35515
aI108109
asS'd'
I13166
sa(dp26332
S'a'
I63247
sS's'
I17
sS'm'
(lp26333
I1
aI1
aI3
aI7
aI5
aI1
aI67
aI59
aI83
aI745
aI1337
aI855
aI6087
aI14319
aI13405
aI36261
aI49091
asS'd'
I13167
sa(dp26334
S'a'
I63252
sS's'
I17
sS'm'
(lp26335
I1
aI3
aI1
aI1
aI13
aI27
aI41
aI155
aI463
aI709
aI1111
aI2017
aI4701
aI8663
aI29703
aI45311
aI113347
asS'd'
I13168
sa(dp26336
S'a'
I63255
sS's'
I17
sS'm'
(lp26337
I1
aI1
aI7
aI5
aI1
aI11
aI83
aI101
aI283
aI505
aI893
aI705
aI2331
aI5127
aI21793
aI28095
aI59055
asS'd'
I13169
sa(dp26338
S'a'
I63289
sS's'
I17
sS'm'
(lp26339
I1
aI1
aI5
aI9
aI25
aI7
aI97
aI155
aI71
aI569
aI1481
aI897
aI6177
aI13367
aI12163
aI18171
aI24785
asS'd'
I13170
sa(dp26340
S'a'
I63298
sS's'
I17
sS'm'
(lp26341
I1
aI3
aI5
aI11
aI19
aI25
aI7
aI15
aI511
aI369
aI957
aI1247
aI6097
aI11181
aI17265
aI24777
aI87377
asS'd'
I13171
sa(dp26342
S'a'
I63303
sS's'
I17
sS'm'
(lp26343
I1
aI3
aI1
aI7
aI11
aI9
aI39
aI191
aI9
aI793
aI867
aI2779
aI3447
aI3805
aI21025
aI64719
aI15669
asS'd'
I13172
sa(dp26344
S'a'
I63327
sS's'
I17
sS'm'
(lp26345
I1
aI1
aI3
aI1
aI31
aI43
aI107
aI103
aI175
aI131
aI1525
aI993
aI635
aI14383
aI26835
aI15929
aI109977
asS'd'
I13173
sa(dp26346
S'a'
I63328
sS's'
I17
sS'm'
(lp26347
I1
aI1
aI1
aI3
aI19
aI17
aI99
aI249
aI47
aI467
aI853
aI2805
aI3155
aI1565
aI17291
aI18865
aI11039
asS'd'
I13174
sa(dp26348
S'a'
I63348
sS's'
I17
sS'm'
(lp26349
I1
aI1
aI5
aI13
aI25
aI61
aI91
aI67
aI361
aI947
aI1909
aI3403
aI945
aI3481
aI16703
aI50227
aI43963
asS'd'
I13175
sa(dp26350
S'a'
I63355
sS's'
I17
sS'm'
(lp26351
I1
aI3
aI1
aI3
aI19
aI27
aI31
aI219
aI185
aI579
aI1539
aI315
aI4421
aI9473
aI30289
aI48477
aI61365
asS'd'
I13176
sa(dp26352
S'a'
I63357
sS's'
I17
sS'm'
(lp26353
I1
aI1
aI3
aI1
aI23
aI11
aI101
aI1
aI133
aI305
aI1107
aI1145
aI1733
aI13275
aI221
aI5071
aI81987
asS'd'
I13177
sa(dp26354
S'a'
I63368
sS's'
I17
sS'm'
(lp26355
I1
aI1
aI1
aI13
aI7
aI61
aI47
aI5
aI137
aI979
aI1183
aI2049
aI5263
aI6515
aI5585
aI6093
aI119689
asS'd'
I13178
sa(dp26356
S'a'
I63391
sS's'
I17
sS'm'
(lp26357
I1
aI3
aI1
aI5
aI19
aI47
aI83
aI115
aI215
aI901
aI1685
aI755
aI2327
aI13297
aI6847
aI40329
aI19225
asS'd'
I13179
sa(dp26358
S'a'
I63402
sS's'
I17
sS'm'
(lp26359
I1
aI1
aI3
aI3
aI3
aI13
aI31
aI127
aI199
aI655
aI55
aI2183
aI5031
aI945
aI6073
aI54729
aI108281
asS'd'
I13180
sa(dp26360
S'a'
I63409
sS's'
I17
sS'm'
(lp26361
I1
aI1
aI1
aI1
aI11
aI51
aI37
aI19
aI73
aI205
aI1377
aI1881
aI3679
aI4487
aI14693
aI41735
aI27349
asS'd'
I13181
sa(dp26362
S'a'
I63416
sS's'
I17
sS'm'
(lp26363
I1
aI3
aI7
aI13
aI1
aI35
aI37
aI73
aI45
aI973
aI209
aI529
aI5283
aI9765
aI26367
aI12697
aI8685
asS'd'
I13182
sa(dp26364
S'a'
I63419
sS's'
I17
sS'm'
(lp26365
I1
aI3
aI3
aI9
aI3
aI45
aI115
aI35
aI475
aI663
aI487
aI3613
aI4151
aI15623
aI3057
aI31519
aI87545
asS'd'
I13183
sa(dp26366
S'a'
I63430
sS's'
I17
sS'm'
(lp26367
I1
aI3
aI7
aI5
aI23
aI13
aI25
aI255
aI355
aI433
aI1671
aI667
aI7463
aI14189
aI14107
aI1531
aI11695
asS'd'
I13184
sa(dp26368
S'a'
I63442
sS's'
I17
sS'm'
(lp26369
I1
aI1
aI7
aI3
aI15
aI25
aI37
aI127
aI265
aI493
aI1763
aI2721
aI4335
aI13753
aI5947
aI18375
aI29911
asS'd'
I13185
sa(dp26370
S'a'
I63457
sS's'
I17
sS'm'
(lp26371
I1
aI1
aI3
aI15
aI1
aI55
aI25
aI69
aI335
aI157
aI1923
aI1757
aI5689
aI6731
aI723
aI6471
aI57415
asS'd'
I13186
sa(dp26372
S'a'
I63458
sS's'
I17
sS'm'
(lp26373
I1
aI3
aI3
aI3
aI1
aI15
aI127
aI227
aI401
aI395
aI503
aI3783
aI1737
aI8785
aI16287
aI34949
aI91683
asS'd'
I13187
sa(dp26374
S'a'
I63482
sS's'
I17
sS'm'
(lp26375
I1
aI3
aI5
aI15
aI23
aI29
aI55
aI119
aI115
aI855
aI657
aI3729
aI5309
aI14773
aI5647
aI25953
aI67303
asS'd'
I13188
sa(dp26376
S'a'
I63487
sS's'
I17
sS'm'
(lp26377
I1
aI3
aI5
aI13
aI23
aI25
aI1
aI187
aI67
aI389
aI359
aI619
aI2523
aI11203
aI11049
aI60345
aI53931
asS'd'
I13189
sa(dp26378
S'a'
I63488
sS's'
I17
sS'm'
(lp26379
I1
aI3
aI1
aI7
aI7
aI45
aI99
aI123
aI495
aI797
aI939
aI3387
aI7563
aI16289
aI8309
aI14917
aI99867
asS'd'
I13190
sa(dp26380
S'a'
I63500
sS's'
I17
sS'm'
(lp26381
I1
aI3
aI5
aI11
aI29
aI49
aI89
aI205
aI447
aI541
aI1279
aI1153
aI7277
aI5393
aI8743
aI41057
aI100119
asS'd'
I13191
sa(dp26382
S'a'
I63511
sS's'
I17
sS'm'
(lp26383
I1
aI1
aI1
aI9
aI1
aI7
aI43
aI165
aI259
aI311
aI993
aI1381
aI3363
aI577
aI4023
aI443
aI101785
asS'd'
I13192
sa(dp26384
S'a'
I63517
sS's'
I17
sS'm'
(lp26385
I1
aI1
aI7
aI9
aI25
aI55
aI93
aI63
aI423
aI787
aI549
aI1039
aI383
aI15855
aI6013
aI51399
aI60007
asS'd'
I13193
sa(dp26386
S'a'
I63528
sS's'
I17
sS'm'
(lp26387
I1
aI3
aI3
aI1
aI5
aI17
aI103
aI91
aI309
aI85
aI1319
aI3869
aI559
aI3993
aI18111
aI17753
aI66681
asS'd'
I13194
sa(dp26388
S'a'
I63531
sS's'
I17
sS'm'
(lp26389
I1
aI3
aI7
aI9
aI1
aI11
aI87
aI151
aI311
aI597
aI811
aI3955
aI275
aI6555
aI17005
aI26763
aI31227
asS'd'
I13195
sa(dp26390
S'a'
I63559
sS's'
I17
sS'm'
(lp26391
I1
aI1
aI3
aI11
aI19
aI51
aI41
aI101
aI183
aI1003
aI1635
aI2061
aI3305
aI12925
aI7223
aI4859
aI24433
asS'd'
I13196
sa(dp26392
S'a'
I63566
sS's'
I17
sS'm'
(lp26393
I1
aI3
aI7
aI11
aI7
aI43
aI79
aI53
aI43
aI429
aI947
aI2533
aI7005
aI15147
aI13435
aI33997
aI21201
asS'd'
I13197
sa(dp26394
S'a'
I63578
sS's'
I17
sS'm'
(lp26395
I1
aI1
aI3
aI5
aI15
aI17
aI61
aI41
aI383
aI465
aI1439
aI3503
aI3981
aI14469
aI5075
aI25953
aI70461
asS'd'
I13198
sa(dp26396
S'a'
I63580
sS's'
I17
sS'm'
(lp26397
I1
aI1
aI1
aI13
aI21
aI53
aI25
aI59
aI59
aI195
aI665
aI3367
aI2777
aI9179
aI24207
aI56729
aI94241
asS'd'
I13199
sa(dp26398
S'a'
I63584
sS's'
I17
sS'm'
(lp26399
I1
aI3
aI7
aI15
aI27
aI13
aI41
aI147
aI415
aI351
aI961
aI3811
aI1605
aI14231
aI31789
aI41189
aI50265
asS'd'
I13200
sa(dp26400
S'a'
I63587
sS's'
I17
sS'm'
(lp26401
I1
aI3
aI3
aI7
aI31
aI39
aI85
aI219
aI323
aI657
aI423
aI1579
aI3623
aI7663
aI14631
aI47169
aI88795
asS'd'
I13201
sa(dp26402
S'a'
I63594
sS's'
I17
sS'm'
(lp26403
I1
aI3
aI1
aI3
aI1
aI3
aI125
aI65
aI259
aI3
aI1897
aI2203
aI347
aI4101
aI23841
aI20217
aI117407
asS'd'
I13202
sa(dp26404
S'a'
I63607
sS's'
I17
sS'm'
(lp26405
I1
aI1
aI3
aI7
aI29
aI1
aI75
aI255
aI413
aI237
aI1531
aI2103
aI6847
aI10395
aI9817
aI9383
aI60679
asS'd'
I13203
sa(dp26406
S'a'
I63611
sS's'
I17
sS'm'
(lp26407
I1
aI3
aI5
aI3
aI7
aI63
aI17
aI83
aI375
aI835
aI1707
aI3227
aI327
aI2205
aI25025
aI47471
aI39967
asS'd'
I13204
sa(dp26408
S'a'
I63630
sS's'
I17
sS'm'
(lp26409
I1
aI3
aI7
aI7
aI9
aI51
aI23
aI189
aI157
aI351
aI755
aI2695
aI3923
aI3481
aI12159
aI1041
aI94563
asS'd'
I13205
sa(dp26410
S'a'
I63632
sS's'
I17
sS'm'
(lp26411
I1
aI1
aI3
aI11
aI25
aI27
aI39
aI19
aI221
aI795
aI523
aI695
aI3257
aI8045
aI2643
aI43239
aI13291
asS'd'
I13206
sa(dp26412
S'a'
I63641
sS's'
I17
sS'm'
(lp26413
I1
aI3
aI3
aI5
aI29
aI1
aI33
aI117
aI477
aI147
aI1117
aI1943
aI7967
aI15999
aI10673
aI13349
aI89471
asS'd'
I13207
sa(dp26414
S'a'
I63647
sS's'
I17
sS'm'
(lp26415
I1
aI1
aI3
aI9
aI17
aI51
aI55
aI115
aI147
aI687
aI1751
aI3685
aI3099
aI15369
aI371
aI55673
aI67951
asS'd'
I13208
sa(dp26416
S'a'
I63651
sS's'
I17
sS'm'
(lp26417
I1
aI1
aI7
aI1
aI5
aI25
aI67
aI31
aI103
aI439
aI1581
aI705
aI3855
aI15985
aI7151
aI56511
aI23697
asS'd'
I13209
sa(dp26418
S'a'
I63666
sS's'
I17
sS'm'
(lp26419
I1
aI3
aI5
aI3
aI21
aI7
aI11
aI123
aI121
aI1009
aI277
aI623
aI7913
aI7525
aI4759
aI19245
aI16735
asS'd'
I13210
sa(dp26420
S'a'
I63668
sS's'
I17
sS'm'
(lp26421
I1
aI1
aI5
aI11
aI7
aI57
aI103
aI147
aI199
aI209
aI233
aI3665
aI4215
aI13511
aI16393
aI37873
aI120857
asS'd'
I13211
sa(dp26422
S'a'
I63695
sS's'
I17
sS'm'
(lp26423
I1
aI1
aI7
aI9
aI27
aI45
aI29
aI97
aI279
aI379
aI1683
aI1965
aI1183
aI11389
aI20445
aI38435
aI56893
asS'd'
I13212
sa(dp26424
S'a'
I63697
sS's'
I17
sS'm'
(lp26425
I1
aI3
aI5
aI5
aI27
aI23
aI89
aI169
aI329
aI659
aI393
aI903
aI6217
aI6459
aI27327
aI2843
aI44889
asS'd'
I13213
sa(dp26426
S'a'
I63709
sS's'
I17
sS'm'
(lp26427
I1
aI1
aI1
aI15
aI3
aI53
aI109
aI83
aI195
aI5
aI1865
aI729
aI3627
aI13307
aI20761
aI50375
aI60379
asS'd'
I13214
sa(dp26428
S'a'
I63723
sS's'
I17
sS'm'
(lp26429
I1
aI1
aI1
aI13
aI25
aI57
aI17
aI185
aI359
aI797
aI469
aI2637
aI973
aI2731
aI25299
aI15169
aI90187
asS'd'
I13215
sa(dp26430
S'a'
I63737
sS's'
I17
sS'm'
(lp26431
I1
aI3
aI5
aI1
aI19
aI39
aI87
aI161
aI117
aI565
aI1737
aI3995
aI6487
aI5067
aI18531
aI38803
aI45453
asS'd'
I13216
sa(dp26432
S'a'
I63746
sS's'
I17
sS'm'
(lp26433
I1
aI1
aI1
aI5
aI19
aI3
aI93
aI85
aI479
aI369
aI469
aI1407
aI475
aI7775
aI12273
aI34417
aI65611
asS'd'
I13217
sa(dp26434
S'a'
I63752
sS's'
I17
sS'm'
(lp26435
I1
aI1
aI3
aI15
aI31
aI11
aI105
aI19
aI281
aI711
aI713
aI3423
aI797
aI11215
aI31409
aI44891
aI110413
asS'd'
I13218
sa(dp26436
S'a'
I63755
sS's'
I17
sS'm'
(lp26437
I1
aI1
aI3
aI11
aI13
aI17
aI59
aI111
aI59
aI431
aI1517
aI2159
aI1697
aI12309
aI16293
aI2097
aI107273
asS'd'
I13219
sa(dp26438
S'a'
I63775
sS's'
I17
sS'm'
(lp26439
I1
aI3
aI5
aI15
aI25
aI19
aI97
aI107
aI97
aI563
aI247
aI3691
aI2775
aI10631
aI15113
aI25721
aI12995
asS'd'
I13220
sa(dp26440
S'a'
I63781
sS's'
I17
sS'm'
(lp26441
I1
aI1
aI5
aI3
aI31
aI25
aI47
aI201
aI231
aI123
aI1923
aI2287
aI1711
aI12271
aI1573
aI6605
aI72991
asS'd'
I13221
sa(dp26442
S'a'
I63794
sS's'
I17
sS'm'
(lp26443
I1
aI1
aI5
aI5
aI27
aI17
aI109
aI125
aI423
aI1
aI819
aI3041
aI685
aI8791
aI19697
aI13107
aI67681
asS'd'
I13222
sa(dp26444
S'a'
I63796
sS's'
I17
sS'm'
(lp26445
I1
aI1
aI5
aI3
aI5
aI63
aI115
aI95
aI117
aI715
aI1523
aI1231
aI8171
aI1615
aI9819
aI14361
aI87075
asS'd'
I13223
sa(dp26446
S'a'
I63808
sS's'
I17
sS'm'
(lp26447
I1
aI1
aI7
aI7
aI7
aI35
aI35
aI175
aI349
aI853
aI1665
aI3101
aI6051
aI10737
aI933
aI40591
aI9419
asS'd'
I13224
sa(dp26448
S'a'
I63817
sS's'
I17
sS'm'
(lp26449
I1
aI1
aI1
aI3
aI23
aI49
aI65
aI23
aI103
aI837
aI403
aI3799
aI6515
aI15363
aI28079
aI36381
aI59523
asS'd'
I13225
sa(dp26450
S'a'
I63820
sS's'
I17
sS'm'
(lp26451
I1
aI3
aI1
aI15
aI15
aI25
aI119
aI181
aI229
aI685
aI1047
aI2397
aI4855
aI15393
aI2371
aI42441
aI30151
asS'd'
I13226
sa(dp26452
S'a'
I63826
sS's'
I17
sS'm'
(lp26453
I1
aI3
aI7
aI11
aI15
aI5
aI13
aI93
aI219
aI203
aI475
aI523
aI5827
aI6579
aI26759
aI29795
aI108463
asS'd'
I13227
sa(dp26454
S'a'
I63838
sS's'
I17
sS'm'
(lp26455
I1
aI1
aI7
aI13
aI25
aI53
aI75
aI195
aI443
aI1003
aI501
aI2543
aI5453
aI3119
aI19225
aI59631
aI16413
asS'd'
I13228
sa(dp26456
S'a'
I63848
sS's'
I17
sS'm'
(lp26457
I1
aI1
aI7
aI13
aI13
aI25
aI93
aI211
aI191
aI1005
aI1567
aI3057
aI3001
aI1125
aI6237
aI35725
aI108257
asS'd'
I13229
sa(dp26458
S'a'
I63861
sS's'
I17
sS'm'
(lp26459
I1
aI1
aI3
aI7
aI21
aI11
aI57
aI205
aI487
aI263
aI1801
aI3235
aI1819
aI10875
aI6063
aI26211
aI54699
asS'd'
I13230
sa(dp26460
S'a'
I63862
sS's'
I17
sS'm'
(lp26461
I1
aI3
aI3
aI7
aI11
aI59
aI89
aI217
aI15
aI991
aI1343
aI1247
aI277
aI13377
aI18499
aI64987
aI26053
asS'd'
I13231
sa(dp26462
S'a'
I63866
sS's'
I17
sS'm'
(lp26463
I1
aI3
aI3
aI1
aI15
aI51
aI111
aI69
aI137
aI817
aI1207
aI1729
aI3877
aI9873
aI18449
aI50749
aI57457
asS'd'
I13232
sa(dp26464
S'a'
I63878
sS's'
I17
sS'm'
(lp26465
I1
aI3
aI3
aI5
aI3
aI39
aI97
aI147
aI327
aI257
aI1547
aI769
aI7077
aI5221
aI13679
aI44237
aI70053
asS'd'
I13233
sa(dp26466
S'a'
I63889
sS's'
I17
sS'm'
(lp26467
I1
aI1
aI5
aI11
aI19
aI15
aI79
aI187
aI335
aI645
aI1235
aI4041
aI4831
aI10847
aI28135
aI48353
aI64921
asS'd'
I13234
sa(dp26468
S'a'
I63892
sS's'
I17
sS'm'
(lp26469
I1
aI1
aI7
aI9
aI3
aI43
aI41
aI149
aI71
aI205
aI1513
aI2801
aI6785
aI3187
aI25401
aI55367
aI114491
asS'd'
I13235
sa(dp26470
S'a'
I63901
sS's'
I17
sS'm'
(lp26471
I1
aI1
aI7
aI1
aI25
aI11
aI37
aI205
aI365
aI435
aI147
aI1303
aI587
aI14563
aI32461
aI28983
aI86157
asS'd'
I13236
sa(dp26472
S'a'
I63915
sS's'
I17
sS'm'
(lp26473
I1
aI1
aI7
aI1
aI31
aI11
aI51
aI37
aI401
aI343
aI1677
aI991
aI501
aI11993
aI14781
aI37055
aI30161
asS'd'
I13237
sa(dp26474
S'a'
I63917
sS's'
I17
sS'm'
(lp26475
I1
aI3
aI5
aI9
aI9
aI21
aI95
aI45
aI447
aI957
aI943
aI3997
aI4033
aI8371
aI25007
aI52827
aI50207
asS'd'
I13238
sa(dp26476
S'a'
I63926
sS's'
I17
sS'm'
(lp26477
I1
aI1
aI7
aI1
aI9
aI45
aI3
aI255
aI297
aI341
aI215
aI3631
aI7049
aI7625
aI4145
aI50109
aI48615
asS'd'
I13239
sa(dp26478
S'a'
I63932
sS's'
I17
sS'm'
(lp26479
I1
aI3
aI3
aI9
aI27
aI49
aI41
aI143
aI291
aI343
aI719
aI311
aI3819
aI7699
aI17631
aI64785
aI49239
asS'd'
I13240
sa(dp26480
S'a'
I63937
sS's'
I17
sS'm'
(lp26481
I1
aI1
aI7
aI3
aI27
aI35
aI61
aI183
aI153
aI781
aI979
aI1465
aI3315
aI14893
aI29847
aI18461
aI74949
asS'd'
I13241
sa(dp26482
S'a'
I63938
sS's'
I17
sS'm'
(lp26483
I1
aI3
aI5
aI15
aI19
aI61
aI39
aI219
aI279
aI909
aI1295
aI1681
aI8021
aI957
aI7675
aI14001
aI77669
asS'd'
I13242
sa(dp26484
S'a'
I63943
sS's'
I17
sS'm'
(lp26485
I1
aI3
aI1
aI5
aI15
aI59
aI127
aI85
aI229
aI649
aI503
aI3267
aI2465
aI5637
aI2729
aI24831
aI44791
asS'd'
I13243
sa(dp26486
S'a'
I63944
sS's'
I17
sS'm'
(lp26487
I1
aI3
aI7
aI11
aI23
aI55
aI61
aI191
aI345
aI255
aI105
aI1361
aI3913
aI7655
aI8865
aI1825
aI80619
asS'd'
I13244
sa(dp26488
S'a'
I63950
sS's'
I17
sS'm'
(lp26489
I1
aI3
aI3
aI13
aI29
aI15
aI53
aI19
aI1
aI651
aI917
aI2043
aI2333
aI13695
aI28225
aI16457
aI11287
asS'd'
I13245
sa(dp26490
S'a'
I63952
sS's'
I17
sS'm'
(lp26491
I1
aI1
aI3
aI13
aI15
aI53
aI41
aI211
aI13
aI287
aI383
aI3923
aI665
aI10343
aI4803
aI22199
aI90521
asS'd'
I13246
sa(dp26492
S'a'
I63955
sS's'
I17
sS'm'
(lp26493
I1
aI3
aI7
aI11
aI23
aI27
aI127
aI241
aI11
aI451
aI495
aI2779
aI319
aI13119
aI5575
aI43043
aI11659
asS'd'
I13247
sa(dp26494
S'a'
I63957
sS's'
I17
sS'm'
(lp26495
I1
aI1
aI1
aI7
aI17
aI53
aI55
aI39
aI233
aI273
aI1873
aI843
aI7885
aI329
aI6809
aI33119
aI116017
asS'd'
I13248
sa(dp26496
S'a'
I63961
sS's'
I17
sS'm'
(lp26497
I1
aI1
aI1
aI7
aI21
aI41
aI23
aI113
aI283
aI265
aI1535
aI2371
aI3975
aI6293
aI22497
aI65349
aI48653
asS'd'
I13249
sa(dp26498
S'a'
I63962
sS's'
I17
sS'm'
(lp26499
I1
aI3
aI7
aI9
aI25
aI21
aI61
aI135
aI245
aI777
aI679
aI2603
aI565
aI3251
aI32469
aI12707
aI40297
asS'd'
I13250
sa(dp26500
S'a'
I63978
sS's'
I17
sS'm'
(lp26501
I1
aI1
aI1
aI5
aI31
aI49
aI35
aI215
aI445
aI669
aI779
aI2231
aI5399
aI5853
aI17941
aI33973
aI126141
asS'd'
I13251
sa(dp26502
S'a'
I63983
sS's'
I17
sS'm'
(lp26503
I1
aI3
aI5
aI5
aI3
aI31
aI45
aI235
aI51
aI65
aI295
aI3755
aI8101
aI821
aI28331
aI38837
aI55235
asS'd'
I13252
sa(dp26504
S'a'
I63988
sS's'
I17
sS'm'
(lp26505
I1
aI1
aI5
aI15
aI23
aI15
aI37
aI197
aI59
aI455
aI1875
aI1745
aI7565
aI8039
aI15901
aI63129
aI36095
asS'd'
I13253
sa(dp26506
S'a'
I64008
sS's'
I17
sS'm'
(lp26507
I1
aI1
aI5
aI11
aI7
aI1
aI77
aI235
aI309
aI245
aI1539
aI1421
aI3401
aI1477
aI12655
aI19851
aI86147
asS'd'
I13254
sa(dp26508
S'a'
I64013
sS's'
I17
sS'm'
(lp26509
I1
aI1
aI3
aI9
aI27
aI9
aI113
aI127
aI167
aI213
aI161
aI4065
aI1275
aI10699
aI26111
aI26213
aI129091
asS'd'
I13255
sa(dp26510
S'a'
I64019
sS's'
I17
sS'm'
(lp26511
I1
aI3
aI5
aI9
aI9
aI17
aI109
aI205
aI23
aI145
aI1261
aI51
aI5855
aI7411
aI20551
aI5801
aI47841
asS'd'
I13256
sa(dp26512
S'a'
I64026
sS's'
I17
sS'm'
(lp26513
I1
aI1
aI3
aI3
aI15
aI1
aI1
aI39
aI431
aI601
aI177
aI525
aI6951
aI6271
aI27031
aI37157
aI73979
asS'd'
I13257
sa(dp26514
S'a'
I64028
sS's'
I17
sS'm'
(lp26515
I1
aI3
aI1
aI3
aI19
aI61
aI11
aI131
aI31
aI223
aI959
aI3531
aI2433
aI15675
aI29201
aI49277
aI43977
asS'd'
I13258
sa(dp26516
S'a'
I64032
sS's'
I17
sS'm'
(lp26517
I1
aI1
aI5
aI9
aI5
aI27
aI57
aI3
aI503
aI755
aI1261
aI3659
aI6685
aI10041
aI24739
aI12201
aI19753
asS'd'
I13259
sa(dp26518
S'a'
I64042
sS's'
I17
sS'm'
(lp26519
I1
aI1
aI7
aI3
aI31
aI27
aI7
aI191
aI7
aI415
aI1665
aI1413
aI7493
aI2645
aI23577
aI46331
aI9481
asS'd'
I13260
sa(dp26520
S'a'
I64044
sS's'
I17
sS'm'
(lp26521
I1
aI1
aI5
aI1
aI29
aI59
aI99
aI231
aI33
aI613
aI1347
aI2671
aI1767
aI15685
aI26583
aI44699
aI73511
asS'd'
I13261
sa(dp26522
S'a'
I64055
sS's'
I17
sS'm'
(lp26523
I1
aI1
aI3
aI3
aI9
aI47
aI93
aI87
aI45
aI549
aI219
aI2141
aI233
aI10239
aI30325
aI14985
aI70325
asS'd'
I13262
sa(dp26524
S'a'
I64070
sS's'
I17
sS'm'
(lp26525
I1
aI1
aI3
aI3
aI21
aI39
aI81
aI179
aI319
aI853
aI93
aI2869
aI59
aI6675
aI22391
aI16089
aI33949
asS'd'
I13263
sa(dp26526
S'a'
I64079
sS's'
I17
sS'm'
(lp26527
I1
aI1
aI3
aI7
aI31
aI19
aI73
aI249
aI175
aI57
aI1717
aI3557
aI2307
aI4595
aI22045
aI33291
aI123003
asS'd'
I13264
sa(dp26528
S'a'
I64084
sS's'
I17
sS'm'
(lp26529
I1
aI1
aI1
aI3
aI7
aI23
aI81
aI229
aI387
aI1001
aI1371
aI17
aI667
aI3043
aI30507
aI44613
aI32239
asS'd'
I13265
sa(dp26530
S'a'
I64087
sS's'
I17
sS'm'
(lp26531
I1
aI1
aI7
aI15
aI15
aI59
aI83
aI99
aI101
aI863
aI333
aI845
aI7547
aI13345
aI7599
aI51
aI10963
asS'd'
I13266
sa(dp26532
S'a'
I64093
sS's'
I17
sS'm'
(lp26533
I1
aI1
aI1
aI3
aI15
aI55
aI73
aI37
aI429
aI711
aI1315
aI2911
aI5109
aI953
aI14721
aI25551
aI33527
asS'd'
I13267
sa(dp26534
S'a'
I64100
sS's'
I17
sS'm'
(lp26535
I1
aI1
aI5
aI9
aI11
aI57
aI75
aI107
aI449
aI293
aI1267
aI2633
aI5291
aI9939
aI12365
aI1975
aI75705
asS'd'
I13268
sa(dp26536
S'a'
I64104
sS's'
I17
sS'm'
(lp26537
I1
aI3
aI3
aI7
aI19
aI51
aI111
aI233
aI369
aI873
aI1419
aI425
aI6587
aI11371
aI29613
aI28041
aI77405
asS'd'
I13269
sa(dp26538
S'a'
I64109
sS's'
I17
sS'm'
(lp26539
I1
aI3
aI1
aI15
aI11
aI1
aI65
aI185
aI301
aI25
aI75
aI1353
aI6879
aI11519
aI24093
aI65223
aI130659
asS'd'
I13270
sa(dp26540
S'a'
I64140
sS's'
I17
sS'm'
(lp26541
I1
aI1
aI3
aI3
aI17
aI17
aI33
aI177
aI467
aI841
aI949
aI1119
aI7869
aI5835
aI22175
aI20439
aI98923
asS'd'
I13271
sa(dp26542
S'a'
I64148
sS's'
I17
sS'm'
(lp26543
I1
aI3
aI1
aI9
aI1
aI19
aI1
aI9
aI487
aI425
aI1095
aI1995
aI693
aI12661
aI27717
aI56167
aI34829
asS'd'
I13272
sa(dp26544
S'a'
I64151
sS's'
I17
sS'm'
(lp26545
I1
aI1
aI7
aI7
aI27
aI57
aI85
aI159
aI109
aI801
aI477
aI3953
aI3195
aI11079
aI26885
aI59833
aI4971
asS'd'
I13273
sa(dp26546
S'a'
I64152
sS's'
I17
sS'm'
(lp26547
I1
aI1
aI1
aI15
aI25
aI9
aI89
aI231
aI499
aI623
aI1385
aI3753
aI4781
aI15263
aI12721
aI17511
aI67327
asS'd'
I13274
sa(dp26548
S'a'
I64171
sS's'
I17
sS'm'
(lp26549
I1
aI1
aI3
aI7
aI9
aI11
aI103
aI65
aI319
aI681
aI1423
aI2355
aI6243
aI399
aI8483
aI23697
aI107995
asS'd'
I13275
sa(dp26550
S'a'
I64179
sS's'
I17
sS'm'
(lp26551
I1
aI1
aI1
aI1
aI5
aI7
aI63
aI117
aI151
aI905
aI163
aI3813
aI6931
aI13161
aI15131
aI63067
aI15649
asS'd'
I13276
sa(dp26552
S'a'
I64186
sS's'
I17
sS'm'
(lp26553
I1
aI3
aI3
aI3
aI13
aI57
aI69
aI199
aI283
aI153
aI617
aI123
aI3125
aI3057
aI8121
aI14483
aI28085
asS'd'
I13277
sa(dp26554
S'a'
I64203
sS's'
I17
sS'm'
(lp26555
I1
aI3
aI7
aI15
aI25
aI45
aI25
aI179
aI91
aI457
aI681
aI537
aI243
aI4369
aI11395
aI17565
aI47875
asS'd'
I13278
sa(dp26556
S'a'
I64206
sS's'
I17
sS'm'
(lp26557
I1
aI3
aI1
aI13
aI29
aI51
aI101
aI23
aI143
aI715
aI1725
aI791
aI6001
aI4283
aI10689
aI49237
aI5231
asS'd'
I13279
sa(dp26558
S'a'
I64213
sS's'
I17
sS'm'
(lp26559
I1
aI3
aI3
aI5
aI27
aI41
aI39
aI17
aI501
aI587
aI1067
aI1859
aI9
aI13449
aI31257
aI17675
aI99769
asS'd'
I13280
sa(dp26560
S'a'
I64214
sS's'
I17
sS'm'
(lp26561
I1
aI1
aI1
aI3
aI15
aI57
aI119
aI195
aI15
aI779
aI761
aI733
aI3505
aI4815
aI23167
aI411
aI52303
asS'd'
I13281
sa(dp26562
S'a'
I64220
sS's'
I17
sS'm'
(lp26563
I1
aI3
aI1
aI13
aI9
aI31
aI5
aI141
aI19
aI487
aI739
aI577
aI4383
aI1951
aI24293
aI45503
aI111923
asS'd'
I13282
sa(dp26564
S'a'
I64233
sS's'
I17
sS'm'
(lp26565
I1
aI3
aI1
aI11
aI25
aI37
aI107
aI245
aI89
aI107
aI1969
aI1569
aI7475
aI11795
aI6123
aI45311
aI52251
asS'd'
I13283
sa(dp26566
S'a'
I64239
sS's'
I17
sS'm'
(lp26567
I1
aI1
aI7
aI11
aI15
aI9
aI67
aI141
aI199
aI91
aI819
aI3721
aI6251
aI6107
aI9393
aI14941
aI98545
asS'd'
I13284
sa(dp26568
S'a'
I64248
sS's'
I17
sS'm'
(lp26569
I1
aI3
aI3
aI11
aI23
aI9
aI31
aI211
aI339
aI665
aI1507
aI2255
aI3589
aI11495
aI28393
aI2017
aI106735
asS'd'
I13285
sa(dp26570
S'a'
I64251
sS's'
I17
sS'm'
(lp26571
I1
aI3
aI5
aI11
aI27
aI13
aI105
aI217
aI173
aI337
aI1573
aI837
aI3771
aI8645
aI28749
aI27501
aI45045
asS'd'
I13286
sa(dp26572
S'a'
I64259
sS's'
I17
sS'm'
(lp26573
I1
aI1
aI5
aI1
aI11
aI43
aI99
aI217
aI131
aI545
aI1323
aI3089
aI5689
aI785
aI9043
aI29961
aI17855
asS'd'
I13287
sa(dp26574
S'a'
I64268
sS's'
I17
sS'm'
(lp26575
I1
aI1
aI3
aI9
aI31
aI41
aI61
aI239
aI271
aI123
aI1583
aI397
aI4243
aI12197
aI9847
aI12341
aI130533
asS'd'
I13288
sa(dp26576
S'a'
I64273
sS's'
I17
sS'm'
(lp26577
I1
aI3
aI5
aI3
aI27
aI11
aI33
aI31
aI77
aI403
aI823
aI2791
aI3475
aI4201
aI15967
aI39149
aI107137
asS'd'
I13289
sa(dp26578
S'a'
I64279
sS's'
I17
sS'm'
(lp26579
I1
aI3
aI1
aI11
aI9
aI5
aI103
aI145
aI85
aI341
aI1615
aI729
aI7209
aI10289
aI20807
aI54167
aI15613
asS'd'
I13290
sa(dp26580
S'a'
I64283
sS's'
I17
sS'm'
(lp26581
I1
aI3
aI7
aI1
aI29
aI33
aI91
aI219
aI171
aI367
aI907
aI3645
aI1059
aI9031
aI247
aI13231
aI14323
asS'd'
I13291
sa(dp26582
S'a'
I64292
sS's'
I17
sS'm'
(lp26583
I1
aI1
aI1
aI7
aI19
aI15
aI65
aI61
aI221
aI941
aI1005
aI1447
aI3513
aI8917
aI17399
aI52471
aI64245
asS'd'
I13292
sa(dp26584
S'a'
I64296
sS's'
I17
sS'm'
(lp26585
I1
aI1
aI5
aI7
aI5
aI35
aI15
aI253
aI325
aI313
aI2015
aI3239
aI1633
aI9745
aI11617
aI10575
aI35877
asS'd'
I13293
sa(dp26586
S'a'
I64301
sS's'
I17
sS'm'
(lp26587
I1
aI3
aI5
aI3
aI13
aI1
aI115
aI207
aI227
aI637
aI1119
aI781
aI2897
aI1573
aI16499
aI43167
aI20631
asS'd'
I13294
sa(dp26588
S'a'
I64302
sS's'
I17
sS'm'
(lp26589
I1
aI3
aI5
aI9
aI17
aI47
aI117
aI7
aI303
aI719
aI975
aI1167
aI2463
aI5255
aI28237
aI33495
aI57133
asS'd'
I13295
sa(dp26590
S'a'
I64324
sS's'
I17
sS'm'
(lp26591
I1
aI3
aI5
aI11
aI5
aI43
aI123
aI63
aI19
aI97
aI1423
aI695
aI5985
aI5923
aI5755
aI22721
aI5411
asS'd'
I13296
sa(dp26592
S'a'
I64331
sS's'
I17
sS'm'
(lp26593
I1
aI3
aI1
aI9
aI9
aI25
aI87
aI197
aI325
aI827
aI1679
aI1561
aI101
aI3951
aI17453
aI33537
aI121431
asS'd'
I13297
sa(dp26594
S'a'
I64346
sS's'
I17
sS'm'
(lp26595
I1
aI1
aI7
aI5
aI13
aI33
aI3
aI191
aI171
aI37
aI619
aI1917
aI7525
aI14103
aI25807
aI25455
aI57455
asS'd'
I13298
sa(dp26596
S'a'
I64364
sS's'
I17
sS'm'
(lp26597
I1
aI3
aI3
aI1
aI9
aI35
aI93
aI159
aI455
aI115
aI479
aI665
aI477
aI4483
aI29751
aI45047
aI41251
asS'd'
I13299
sa(dp26598
S'a'
I64382
sS's'
I17
sS'm'
(lp26599
I1
aI3
aI1
aI3
aI11
aI47
aI41
aI199
aI511
aI475
aI151
aI1163
aI239
aI6731
aI4461
aI39845
aI99555
asS'd'
I13300
sa(dp26600
S'a'
I64386
sS's'
I17
sS'm'
(lp26601
I1
aI1
aI5
aI7
aI9
aI5
aI49
aI221
aI503
aI637
aI1323
aI3303
aI4137
aI6675
aI17709
aI49233
aI38325
asS'd'
I13301
sa(dp26602
S'a'
I64400
sS's'
I17
sS'm'
(lp26603
I1
aI1
aI5
aI15
aI1
aI43
aI55
aI67
aI291
aI393
aI237
aI3555
aI4171
aI909
aI8655
aI46309
aI61799
asS'd'
I13302
sa(dp26604
S'a'
I64409
sS's'
I17
sS'm'
(lp26605
I1
aI3
aI5
aI3
aI3
aI37
aI125
aI249
aI509
aI611
aI983
aI4093
aI1633
aI10063
aI10811
aI60033
aI40999
asS'd'
I13303
sa(dp26606
S'a'
I64419
sS's'
I17
sS'm'
(lp26607
I1
aI3
aI5
aI11
aI1
aI37
aI75
aI255
aI279
aI545
aI1999
aI833
aI2789
aI14601
aI16707
aI64703
aI53545
asS'd'
I13304
sa(dp26608
S'a'
I64433
sS's'
I17
sS'm'
(lp26609
I1
aI1
aI5
aI7
aI3
aI15
aI59
aI11
aI17
aI711
aI721
aI765
aI3747
aI13549
aI28641
aI47437
aI42261
asS'd'
I13305
sa(dp26610
S'a'
I64454
sS's'
I17
sS'm'
(lp26611
I1
aI3
aI7
aI1
aI3
aI45
aI65
aI45
aI279
aI929
aI933
aI2215
aI7095
aI14593
aI6047
aI40747
aI109789
asS'd'
I13306
sa(dp26612
S'a'
I64458
sS's'
I17
sS'm'
(lp26613
I1
aI3
aI7
aI15
aI15
aI55
aI89
aI155
aI345
aI515
aI1005
aI2921
aI1761
aI1095
aI28463
aI20971
aI62451
asS'd'
I13307
sa(dp26614
S'a'
I64482
sS's'
I17
sS'm'
(lp26615
I1
aI3
aI3
aI1
aI1
aI41
aI35
aI149
aI481
aI171
aI305
aI1411
aI237
aI4515
aI32375
aI22645
aI741
asS'd'
I13308
sa(dp26616
S'a'
I64494
sS's'
I17
sS'm'
(lp26617
I1
aI1
aI1
aI15
aI17
aI1
aI123
aI235
aI221
aI495
aI1693
aI3109
aI6453
aI8827
aI23775
aI9303
aI30237
asS'd'
I13309
sa(dp26618
S'a'
I64496
sS's'
I17
sS'm'
(lp26619
I1
aI3
aI3
aI5
aI7
aI63
aI37
aI13
aI457
aI159
aI1683
aI2207
aI1731
aI3341
aI7415
aI21073
aI119417
asS'd'
I13310
sa(dp26620
S'a'
I64505
sS's'
I17
sS'm'
(lp26621
I1
aI1
aI7
aI15
aI21
aI27
aI5
aI67
aI267
aI919
aI203
aI1129
aI4029
aI3407
aI16767
aI35485
aI66903
asS'd'
I13311
sa(dp26622
S'a'
I64514
sS's'
I17
sS'm'
(lp26623
I1
aI1
aI1
aI5
aI15
aI29
aI99
aI5
aI219
aI677
aI443
aI3799
aI2461
aI747
aI20885
aI32661
aI44079
asS'd'
I13312
sa(dp26624
S'a'
I64519
sS's'
I17
sS'm'
(lp26625
I1
aI1
aI1
aI1
aI3
aI55
aI53
aI151
aI195
aI587
aI1155
aI2439
aI3817
aI8735
aI30849
aI54107
aI14113
asS'd'
I13313
sa(dp26626
S'a'
I64525
sS's'
I17
sS'm'
(lp26627
I1
aI1
aI1
aI9
aI29
aI15
aI89
aI175
aI373
aI925
aI301
aI3749
aI5439
aI2653
aI22819
aI41201
aI77043
asS'd'
I13314
sa(dp26628
S'a'
I64528
sS's'
I17
sS'm'
(lp26629
I1
aI1
aI1
aI1
aI25
aI49
aI29
aI129
aI331
aI539
aI1247
aI773
aI7891
aI5905
aI19571
aI17919
aI6815
asS'd'
I13315
sa(dp26630
S'a'
I64534
sS's'
I17
sS'm'
(lp26631
I1
aI1
aI3
aI15
aI5
aI63
aI123
aI133
aI141
aI383
aI1893
aI573
aI629
aI3939
aI9455
aI50433
aI111415
asS'd'
I13316
sa(dp26632
S'a'
I64561
sS's'
I17
sS'm'
(lp26633
I1
aI1
aI7
aI9
aI15
aI33
aI119
aI159
aI17
aI511
aI1841
aI427
aI3911
aI8609
aI4215
aI9799
aI84397
asS'd'
I13317
sa(dp26634
S'a'
I64571
sS's'
I17
sS'm'
(lp26635
I1
aI1
aI7
aI3
aI9
aI25
aI63
aI247
aI235
aI635
aI915
aI3423
aI5421
aI7021
aI9203
aI18121
aI3683
asS'd'
I13318
sa(dp26636
S'a'
I64579
sS's'
I17
sS'm'
(lp26637
I1
aI3
aI1
aI1
aI25
aI11
aI105
aI1
aI491
aI137
aI1923
aI103
aI3371
aI3543
aI5173
aI36777
aI23417
asS'd'
I13319
sa(dp26638
S'a'
I64591
sS's'
I17
sS'm'
(lp26639
I1
aI3
aI3
aI13
aI19
aI37
aI93
aI191
aI101
aI193
aI351
aI839
aI7147
aI5477
aI29225
aI45307
aI1455
asS'd'
I13320
sa(dp26640
S'a'
I64606
sS's'
I17
sS'm'
(lp26641
I1
aI3
aI1
aI5
aI11
aI17
aI95
aI239
aI105
aI407
aI395
aI919
aI3317
aI14825
aI23447
aI4897
aI128363
asS'd'
I13321
sa(dp26642
S'a'
I64616
sS's'
I17
sS'm'
(lp26643
I1
aI1
aI1
aI11
aI27
aI47
aI83
aI137
aI163
aI673
aI1291
aI3041
aI4559
aI7217
aI23613
aI19477
aI93805
asS'd'
I13322
sa(dp26644
S'a'
I64619
sS's'
I17
sS'm'
(lp26645
I1
aI1
aI1
aI15
aI25
aI51
aI37
aI9
aI23
aI757
aI1921
aI2649
aI5677
aI11421
aI10231
aI1775
aI124709
asS'd'
I13323
sa(dp26646
S'a'
I64640
sS's'
I17
sS'm'
(lp26647
I1
aI3
aI1
aI13
aI31
aI37
aI37
aI163
aI59
aI975
aI1203
aI1425
aI1255
aI3259
aI16681
aI38101
aI118165
asS'd'
I13324
sa(dp26648
S'a'
I64645
sS's'
I17
sS'm'
(lp26649
I1
aI1
aI3
aI11
aI17
aI17
aI31
aI23
aI169
aI305
aI3
aI1631
aI6853
aI7019
aI14539
aI57663
aI70377
asS'd'
I13325
sa(dp26650
S'a'
I64649
sS's'
I17
sS'm'
(lp26651
I1
aI1
aI7
aI3
aI15
aI61
aI113
aI31
aI497
aI935
aI473
aI819
aI1223
aI13907
aI5075
aI45177
aI20255
asS'd'
I13326
sa(dp26652
S'a'
I64652
sS's'
I17
sS'm'
(lp26653
I1
aI3
aI7
aI13
aI9
aI41
aI123
aI121
aI497
aI877
aI915
aI3323
aI4815
aI4175
aI25979
aI38751
aI107099
asS'd'
I13327
sa(dp26654
S'a'
I64670
sS's'
I17
sS'm'
(lp26655
I1
aI1
aI3
aI7
aI13
aI33
aI31
aI167
aI331
aI595
aI517
aI1237
aI1947
aI1905
aI28155
aI52431
aI93065
asS'd'
I13328
sa(dp26656
S'a'
I64673
sS's'
I17
sS'm'
(lp26657
I1
aI3
aI1
aI1
aI11
aI51
aI7
aI151
aI323
aI211
aI523
aI2929
aI233
aI3633
aI2785
aI6043
aI100101
asS'd'
I13329
sa(dp26658
S'a'
I64674
sS's'
I17
sS'm'
(lp26659
I1
aI1
aI7
aI13
aI29
aI3
aI125
aI247
aI121
aI567
aI857
aI3225
aI7461
aI15413
aI773
aI54939
aI67443
asS'd'
I13330
sa(dp26660
S'a'
I64683
sS's'
I17
sS'm'
(lp26661
I1
aI1
aI1
aI15
aI19
aI29
aI101
aI179
aI369
aI115
aI1777
aI3223
aI1499
aI12487
aI41
aI50607
aI111137
asS'd'
I13331
sa(dp26662
S'a'
I64697
sS's'
I17
sS'm'
(lp26663
I1
aI1
aI3
aI1
aI9
aI59
aI21
aI25
aI173
aI357
aI1143
aI1353
aI3907
aI10743
aI30325
aI39211
aI116671
asS'd'
I13332
sa(dp26664
S'a'
I64703
sS's'
I17
sS'm'
(lp26665
I1
aI1
aI7
aI15
aI9
aI63
aI67
aI229
aI7
aI399
aI2037
aI3531
aI6393
aI4273
aI9365
aI52009
aI118093
asS'd'
I13333
sa(dp26666
S'a'
I64711
sS's'
I17
sS'm'
(lp26667
I1
aI1
aI7
aI1
aI31
aI21
aI5
aI251
aI433
aI1
aI481
aI4041
aI6179
aI825
aI8671
aI20597
aI103257
asS'd'
I13334
sa(dp26668
S'a'
I64723
sS's'
I17
sS'm'
(lp26669
I1
aI1
aI7
aI1
aI15
aI41
aI69
aI93
aI47
aI17
aI1901
aI2671
aI4739
aI1883
aI30239
aI50763
aI108295
asS'd'
I13335
sa(dp26670
S'a'
I64736
sS's'
I17
sS'm'
(lp26671
I1
aI3
aI7
aI15
aI29
aI19
aI63
aI213
aI475
aI133
aI43
aI955
aI2001
aI555
aI10479
aI1333
aI52807
asS'd'
I13336
sa(dp26672
S'a'
I64739
sS's'
I17
sS'm'
(lp26673
I1
aI3
aI3
aI15
aI27
aI13
aI91
aI109
aI71
aI333
aI1971
aI3355
aI2175
aI11457
aI31101
aI30217
aI68263
asS'd'
I13337
sa(dp26674
S'a'
I64741
sS's'
I17
sS'm'
(lp26675
I1
aI1
aI5
aI1
aI21
aI33
aI51
aI169
aI365
aI475
aI1015
aI985
aI7217
aI15453
aI7727
aI49843
aI57733
asS'd'
I13338
sa(dp26676
S'a'
I64748
sS's'
I17
sS'm'
(lp26677
I1
aI1
aI7
aI1
aI11
aI37
aI67
aI135
aI429
aI403
aI1663
aI2037
aI7849
aI3757
aI6373
aI38703
aI46393
asS'd'
I13339
sa(dp26678
S'a'
I64759
sS's'
I17
sS'm'
(lp26679
I1
aI1
aI1
aI3
aI15
aI3
aI29
aI101
aI327
aI643
aI47
aI1805
aI6873
aI1659
aI31097
aI34847
aI46843
asS'd'
I13340
sa(dp26680
S'a'
I64768
sS's'
I17
sS'm'
(lp26681
I1
aI3
aI1
aI15
aI9
aI45
aI7
aI189
aI175
aI955
aI45
aI3545
aI3595
aI7443
aI2913
aI54501
aI63279
asS'd'
I13341
sa(dp26682
S'a'
I64771
sS's'
I17
sS'm'
(lp26683
I1
aI3
aI7
aI11
aI1
aI39
aI59
aI179
aI209
aI121
aI445
aI4077
aI4851
aI15161
aI29133
aI13543
aI106247
asS'd'
I13342
sa(dp26684
S'a'
I64778
sS's'
I17
sS'm'
(lp26685
I1
aI3
aI7
aI7
aI5
aI53
aI73
aI107
aI409
aI639
aI1731
aI1921
aI999
aI14445
aI17629
aI3667
aI74819
asS'd'
I13343
sa(dp26686
S'a'
I64792
sS's'
I17
sS'm'
(lp26687
I1
aI3
aI3
aI9
aI23
aI41
aI117
aI195
aI497
aI425
aI627
aI1599
aI7715
aI1401
aI7217
aI61113
aI67135
asS'd'
I13344
sa(dp26688
S'a'
I64821
sS's'
I17
sS'm'
(lp26689
I1
aI1
aI5
aI13
aI9
aI33
aI97
aI115
aI233
aI833
aI1041
aI1755
aI5317
aI12703
aI25709
aI62293
aI2569
asS'd'
I13345
sa(dp26690
S'a'
I64831
sS's'
I17
sS'm'
(lp26691
I1
aI1
aI1
aI11
aI1
aI7
aI27
aI151
aI325
aI905
aI1279
aI4093
aI7495
aI9803
aI17339
aI7977
aI24009
asS'd'
I13346
sa(dp26692
S'a'
I64839
sS's'
I17
sS'm'
(lp26693
I1
aI3
aI1
aI11
aI25
aI59
aI89
aI175
aI67
aI139
aI1507
aI411
aI7863
aI9585
aI14869
aI46655
aI126021
asS'd'
I13347
sa(dp26694
S'a'
I64848
sS's'
I17
sS'm'
(lp26695
I1
aI3
aI3
aI15
aI29
aI5
aI111
aI251
aI69
aI177
aI519
aI901
aI4331
aI5341
aI22031
aI3851
aI114369
asS'd'
I13348
sa(dp26696
S'a'
I64860
sS's'
I17
sS'm'
(lp26697
I1
aI3
aI5
aI3
aI19
aI9
aI83
aI69
aI411
aI673
aI1549
aI3429
aI3647
aI12601
aI17177
aI16161
aI114561
asS'd'
I13349
sa(dp26698
S'a'
I64867
sS's'
I17
sS'm'
(lp26699
I1
aI1
aI5
aI5
aI21
aI15
aI65
aI179
aI405
aI571
aI1245
aI3693
aI7471
aI12109
aI20177
aI28783
aI124339
asS'd'
I13350
sa(dp26700
S'a'
I64870
sS's'
I17
sS'm'
(lp26701
I1
aI1
aI5
aI5
aI9
aI61
aI69
aI99
aI9
aI829
aI1823
aI3803
aI1181
aI3073
aI10069
aI28689
aI21347
asS'd'
I13351
sa(dp26702
S'a'
I64874
sS's'
I17
sS'm'
(lp26703
I1
aI1
aI5
aI1
aI3
aI11
aI25
aI99
aI241
aI957
aI1137
aI7
aI3809
aI7073
aI21217
aI49447
aI41425
asS'd'
I13352
sa(dp26704
S'a'
I64879
sS's'
I17
sS'm'
(lp26705
I1
aI3
aI1
aI9
aI15
aI59
aI13
aI29
aI467
aI893
aI1667
aI31
aI3269
aI12599
aI28673
aI17101
aI81591
asS'd'
I13353
sa(dp26706
S'a'
I64887
sS's'
I17
sS'm'
(lp26707
I1
aI3
aI7
aI3
aI15
aI55
aI79
aI177
aI1
aI891
aI217
aI2725
aI6171
aI7779
aI16173
aI1003
aI37093
asS'd'
I13354
sa(dp26708
S'a'
I64894
sS's'
I17
sS'm'
(lp26709
I1
aI3
aI7
aI11
aI15
aI61
aI13
aI181
aI421
aI83
aI905
aI1089
aI4597
aI3291
aI23243
aI53123
aI21315
asS'd'
I13355
sa(dp26710
S'a'
I64897
sS's'
I17
sS'm'
(lp26711
I1
aI1
aI3
aI3
aI21
aI63
aI113
aI149
aI203
aI379
aI583
aI1955
aI8087
aI9155
aI23019
aI17757
aI1537
asS'd'
I13356
sa(dp26712
S'a'
I64898
sS's'
I17
sS'm'
(lp26713
I1
aI3
aI5
aI9
aI27
aI41
aI61
aI207
aI213
aI253
aI693
aI273
aI1835
aI14135
aI11519
aI40819
aI50999
asS'd'
I13357
sa(dp26714
S'a'
I64921
sS's'
I17
sS'm'
(lp26715
I1
aI3
aI5
aI3
aI1
aI51
aI71
aI237
aI355
aI327
aI1903
aI133
aI6075
aI4685
aI29689
aI48723
aI67791
asS'd'
I13358
sa(dp26716
S'a'
I64933
sS's'
I17
sS'm'
(lp26717
I1
aI3
aI5
aI9
aI21
aI13
aI101
aI23
aI95
aI369
aI1657
aI989
aI4081
aI1373
aI29005
aI7247
aI53923
asS'd'
I13359
sa(dp26718
S'a'
I64940
sS's'
I17
sS'm'
(lp26719
I1
aI1
aI7
aI13
aI15
aI3
aI71
aI189
aI345
aI771
aI251
aI937
aI1041
aI3017
aI27279
aI1635
aI32581
asS'd'
I13360
sa(dp26720
S'a'
I64957
sS's'
I17
sS'm'
(lp26721
I1
aI1
aI5
aI7
aI23
aI63
aI99
aI43
aI237
aI189
aI1549
aI25
aI63
aI14089
aI14387
aI51423
aI57193
asS'd'
I13361
sa(dp26722
S'a'
I64969
sS's'
I17
sS'm'
(lp26723
I1
aI1
aI7
aI15
aI13
aI55
aI89
aI87
aI95
aI241
aI827
aI501
aI2341
aI14357
aI831
aI27101
aI98285
asS'd'
I13362
sa(dp26724
S'a'
I64972
sS's'
I17
sS'm'
(lp26725
I1
aI1
aI1
aI9
aI29
aI29
aI125
aI81
aI73
aI123
aI329
aI2617
aI1259
aI4415
aI30007
aI19467
aI117847
asS'd'
I13363
sa(dp26726
S'a'
I64978
sS's'
I17
sS'm'
(lp26727
I1
aI3
aI1
aI11
aI15
aI63
aI85
aI121
aI409
aI885
aI1197
aI423
aI2673
aI12107
aI1127
aI14119
aI90541
asS'd'
I13364
sa(dp26728
S'a'
I64984
sS's'
I17
sS'm'
(lp26729
I1
aI3
aI1
aI3
aI1
aI35
aI117
aI149
aI213
aI925
aI923
aI1013
aI3547
aI6877
aI3467
aI47893
aI38645
asS'd'
I13365
sa(dp26730
S'a'
I64993
sS's'
I17
sS'm'
(lp26731
I1
aI1
aI3
aI15
aI3
aI21
aI87
aI199
aI197
aI851
aI1711
aI3449
aI1771
aI1727
aI11651
aI51903
aI99835
asS'd'
I13366
sa(dp26732
S'a'
I64999
sS's'
I17
sS'm'
(lp26733
I1
aI3
aI3
aI1
aI5
aI27
aI57
aI243
aI465
aI173
aI697
aI4011
aI6177
aI3019
aI31317
aI24699
aI53151
asS'd'
I13367
sa(dp26734
S'a'
I65000
sS's'
I17
sS'm'
(lp26735
I1
aI3
aI5
aI7
aI7
aI51
aI61
aI177
aI489
aI381
aI493
aI1975
aI3143
aI8003
aI7735
aI46363
aI110705
asS'd'
I13368
sa(dp26736
S'a'
I65006
sS's'
I17
sS'm'
(lp26737
I1
aI1
aI7
aI5
aI27
aI45
aI69
aI33
aI229
aI725
aI2033
aI3655
aI3027
aI11795
aI2941
aI7921
aI117605
asS'd'
I13369
sa(dp26738
S'a'
I65011
sS's'
I17
sS'm'
(lp26739
I1
aI3
aI1
aI7
aI3
aI37
aI91
aI255
aI13
aI651
aI49
aI309
aI7425
aI11641
aI3661
aI3929
aI94199
asS'd'
I13370
sa(dp26740
S'a'
I65014
sS's'
I17
sS'm'
(lp26741
I1
aI3
aI7
aI5
aI7
aI47
aI121
aI203
aI297
aI941
aI1585
aI3659
aI265
aI159
aI30729
aI31825
aI343
asS'd'
I13371
sa(dp26742
S'a'
I65036
sS's'
I17
sS'm'
(lp26743
I1
aI3
aI5
aI9
aI3
aI25
aI95
aI215
aI125
aI105
aI37
aI943
aI4095
aI8169
aI26763
aI20975
aI122307
asS'd'
I13372
sa(dp26744
S'a'
I65044
sS's'
I17
sS'm'
(lp26745
I1
aI1
aI3
aI15
aI9
aI13
aI81
aI25
aI51
aI15
aI599
aI835
aI6723
aI9487
aI25219
aI60401
aI48749
asS'd'
I13373
sa(dp26746
S'a'
I65063
sS's'
I17
sS'm'
(lp26747
I1
aI3
aI3
aI15
aI15
aI47
aI41
aI219
aI77
aI43
aI1705
aI2363
aI7005
aI7137
aI17687
aI665
aI116097
asS'd'
I13374
sa(dp26748
S'a'
I65067
sS's'
I17
sS'm'
(lp26749
I1
aI3
aI5
aI1
aI17
aI33
aI71
aI3
aI253
aI355
aI117
aI1995
aI3339
aI11789
aI13563
aI58889
aI18553
asS'd'
I13375
sa(dp26750
S'a'
I65075
sS's'
I17
sS'm'
(lp26751
I1
aI3
aI5
aI1
aI21
aI33
aI89
aI177
aI9
aI951
aI1593
aI1419
aI3295
aI9617
aI31661
aI7841
aI119939
asS'd'
I13376
sa(dp26752
S'a'
I65077
sS's'
I17
sS'm'
(lp26753
I1
aI3
aI3
aI1
aI31
aI35
aI25
aI9
aI379
aI271
aI923
aI2387
aI3351
aI5869
aI4501
aI6855
aI28273
asS'd'
I13377
sa(dp26754
S'a'
I65082
sS's'
I17
sS'm'
(lp26755
I1
aI1
aI5
aI9
aI11
aI15
aI127
aI79
aI405
aI579
aI395
aI2469
aI5847
aI7589
aI17577
aI61717
aI6493
asS'd'
I13378
sa(dp26756
S'a'
I65095
sS's'
I17
sS'm'
(lp26757
I1
aI3
aI7
aI13
aI29
aI13
aI99
aI209
aI79
aI469
aI5
aI2231
aI89
aI1557
aI5123
aI47169
aI46529
asS'd'
I13379
sa(dp26758
S'a'
I65101
sS's'
I17
sS'm'
(lp26759
I1
aI3
aI7
aI9
aI13
aI35
aI119
aI53
aI7
aI351
aI601
aI901
aI5407
aI13673
aI6929
aI38311
aI2659
asS'd'
I13380
sa(dp26760
S'a'
I65104
sS's'
I17
sS'm'
(lp26761
I1
aI3
aI7
aI9
aI13
aI23
aI61
aI255
aI113
aI331
aI367
aI2979
aI2741
aI6971
aI26447
aI6861
aI116267
asS'd'
I13381
sa(dp26762
S'a'
I65109
sS's'
I17
sS'm'
(lp26763
I1
aI1
aI3
aI3
aI25
aI57
aI93
aI5
aI387
aI87
aI1765
aI1277
aI8175
aI11185
aI4377
aI9779
aI95569
asS'd'
I13382
sa(dp26764
S'a'
I65110
sS's'
I17
sS'm'
(lp26765
I1
aI1
aI7
aI11
aI29
aI43
aI31
aI155
aI111
aI409
aI733
aI1919
aI2681
aI8435
aI5877
aI35439
aI15435
asS'd'
I13383
sa(dp26766
S'a'
I65116
sS's'
I17
sS'm'
(lp26767
I1
aI1
aI1
aI7
aI19
aI33
aI109
aI125
aI51
aI733
aI997
aI3467
aI5081
aI8371
aI263
aI31461
aI46117
asS'd'
I13384
sa(dp26768
S'a'
I65126
sS's'
I17
sS'm'
(lp26769
I1
aI3
aI3
aI7
aI27
aI61
aI57
aI75
aI317
aI247
aI1535
aI3757
aI4617
aI15627
aI11191
aI3581
aI64475
asS'd'
I13385
sa(dp26770
S'a'
I65129
sS's'
I17
sS'm'
(lp26771
I1
aI3
aI3
aI1
aI31
aI7
aI95
aI151
aI159
aI475
aI559
aI379
aI361
aI5953
aI5551
aI20313
aI64015
asS'd'
I13386
sa(dp26772
S'a'
I65138
sS's'
I17
sS'm'
(lp26773
I1
aI1
aI5
aI1
aI11
aI31
aI71
aI77
aI493
aI697
aI345
aI1809
aI611
aI14319
aI6591
aI23657
aI44071
asS'd'
I13387
sa(dp26774
S'a'
I65160
sS's'
I17
sS'm'
(lp26775
I1
aI1
aI1
aI13
aI5
aI1
aI9
aI233
aI229
aI397
aI1201
aI1817
aI7409
aI11521
aI3753
aI35611
aI123037
asS'd'
I13388
sa(dp26776
S'a'
I65171
sS's'
I17
sS'm'
(lp26777
I1
aI3
aI7
aI15
aI9
aI15
aI85
aI163
aI99
aI867
aI265
aI1021
aI129
aI11059
aI123
aI27185
aI68435
asS'd'
I13389
sa(dp26778
S'a'
I65173
sS's'
I17
sS'm'
(lp26779
I1
aI3
aI1
aI11
aI25
aI43
aI105
aI165
aI291
aI977
aI463
aI2699
aI5361
aI9951
aI29735
aI63501
aI86235
asS'd'
I13390
sa(dp26780
S'a'
I65180
sS's'
I17
sS'm'
(lp26781
I1
aI3
aI7
aI13
aI9
aI33
aI39
aI145
aI441
aI233
aI373
aI193
aI1451
aI7975
aI2871
aI64431
aI43339
asS'd'
I13391
sa(dp26782
S'a'
I65189
sS's'
I17
sS'm'
(lp26783
I1
aI1
aI7
aI13
aI15
aI25
aI45
aI27
aI319
aI719
aI1801
aI447
aI3027
aI769
aI271
aI37227
aI26447
asS'd'
I13392
sa(dp26784
S'a'
I65193
sS's'
I17
sS'm'
(lp26785
I1
aI1
aI5
aI1
aI29
aI1
aI59
aI59
aI121
aI251
aI387
aI55
aI5957
aI10527
aI24227
aI38841
aI29115
asS'd'
I13393
sa(dp26786
S'a'
I65208
sS's'
I17
sS'm'
(lp26787
I1
aI3
aI1
aI5
aI31
aI25
aI67
aI191
aI137
aI849
aI631
aI953
aI3103
aI9737
aI28993
aI49413
aI60709
asS'd'
I13394
sa(dp26788
S'a'
I65214
sS's'
I17
sS'm'
(lp26789
I1
aI3
aI3
aI15
aI7
aI5
aI37
aI179
aI357
aI961
aI1649
aI441
aI5287
aI4161
aI24013
aI39661
aI76233
asS'd'
I13395
sa(dp26790
S'a'
I65216
sS's'
I17
sS'm'
(lp26791
I1
aI1
aI5
aI1
aI9
aI1
aI47
aI209
aI219
aI1021
aI969
aI2343
aI5675
aI7137
aI14247
aI50305
aI72613
asS'd'
I13396
sa(dp26792
S'a'
I65225
sS's'
I17
sS'm'
(lp26793
I1
aI1
aI3
aI1
aI9
aI43
aI43
aI47
aI35
aI97
aI617
aI1033
aI2387
aI14155
aI17049
aI53333
aI108619
asS'd'
I13397
sa(dp26794
S'a'
I65236
sS's'
I17
sS'm'
(lp26795
I1
aI3
aI1
aI3
aI1
aI45
aI11
aI171
aI349
aI65
aI909
aI1801
aI1075
aI10905
aI7395
aI19997
aI128205
asS'd'
I13398
sa(dp26796
S'a'
I65239
sS's'
I17
sS'm'
(lp26797
I1
aI1
aI7
aI11
aI19
aI39
aI117
aI175
aI459
aI791
aI1383
aI3473
aI6937
aI8447
aI10077
aI13353
aI122063
asS'd'
I13399
sa(dp26798
S'a'
I65273
sS's'
I17
sS'm'
(lp26799
I1
aI3
aI3
aI9
aI3
aI27
aI115
aI29
aI135
aI305
aI1023
aI2517
aI1981
aI4969
aI18149
aI35565
aI120785
asS'd'
I13400
sa(dp26800
S'a'
I65274
sS's'
I17
sS'm'
(lp26801
I1
aI3
aI3
aI11
aI15
aI23
aI27
aI115
aI411
aI805
aI841
aI2205
aI5997
aI5141
aI10679
aI25235
aI81989
asS'd'
I13401
sa(dp26802
S'a'
I65281
sS's'
I17
sS'm'
(lp26803
I1
aI1
aI3
aI9
aI11
aI63
aI27
aI185
aI337
aI891
aI1447
aI1397
aI8009
aI4453
aI23077
aI37599
aI93389
asS'd'
I13402
sa(dp26804
S'a'
I65294
sS's'
I17
sS'm'
(lp26805
I1
aI1
aI5
aI13
aI27
aI11
aI77
aI11
aI447
aI81
aI1603
aI2317
aI6499
aI6631
aI27305
aI51049
aI40967
asS'd'
I13403
sa(dp26806
S'a'
I65324
sS's'
I17
sS'm'
(lp26807
I1
aI3
aI7
aI1
aI7
aI43
aI83
aI33
aI69
aI119
aI139
aI1391
aI4879
aI3759
aI31211
aI29203
aI110229
asS'd'
I13404
sa(dp26808
S'a'
I65335
sS's'
I17
sS'm'
(lp26809
I1
aI3
aI7
aI15
aI31
aI59
aI53
aI97
aI135
aI233
aI1421
aI587
aI2985
aI3627
aI7355
aI53829
aI51581
asS'd'
I13405
sa(dp26810
S'a'
I65354
sS's'
I17
sS'm'
(lp26811
I1
aI3
aI7
aI15
aI1
aI37
aI39
aI225
aI147
aI37
aI327
aI2819
aI6081
aI4337
aI22063
aI21177
aI91065
asS'd'
I13406
sa(dp26812
S'a'
I65359
sS's'
I17
sS'm'
(lp26813
I1
aI1
aI3
aI1
aI13
aI31
aI61
aI133
aI433
aI243
aI131
aI3625
aI6389
aI335
aI24029
aI33217
aI80833
asS'd'
I13407
sa(dp26814
S'a'
I65373
sS's'
I17
sS'm'
(lp26815
I1
aI1
aI7
aI11
aI21
aI39
aI95
aI181
aI35
aI499
aI677
aI3935
aI1379
aI6791
aI12633
aI13671
aI28317
asS'd'
I13408
sa(dp26816
S'a'
I65383
sS's'
I17
sS'm'
(lp26817
I1
aI3
aI1
aI5
aI7
aI57
aI5
aI229
aI389
aI197
aI1523
aI1221
aI609
aI10449
aI6389
aI9279
aI53871
asS'd'
I13409
sa(dp26818
S'a'
I65387
sS's'
I17
sS'm'
(lp26819
I1
aI3
aI5
aI7
aI1
aI39
aI69
aI131
aI387
aI839
aI1375
aI3841
aI81
aI7395
aI5837
aI32067
aI51183
asS'd'
I13410
sa(dp26820
S'a'
I65397
sS's'
I17
sS'm'
(lp26821
I1
aI1
aI5
aI3
aI5
aI27
aI107
aI171
aI53
aI923
aI345
aI445
aI1101
aI11201
aI20563
aI30889
aI72361
asS'd'
I13411
sa(dp26822
S'a'
I65411
sS's'
I17
sS'm'
(lp26823
I1
aI3
aI7
aI11
aI19
aI7
aI99
aI219
aI485
aI403
aI293
aI3967
aI7517
aI4765
aI11331
aI55
aI92641
asS'd'
I13412
sa(dp26824
S'a'
I65413
sS's'
I17
sS'm'
(lp26825
I1
aI1
aI7
aI13
aI19
aI9
aI73
aI31
aI405
aI513
aI941
aI3645
aI7075
aI8109
aI21431
aI52791
aI120927
asS'd'
I13413
sa(dp26826
S'a'
I65418
sS's'
I17
sS'm'
(lp26827
I1
aI1
aI1
aI15
aI29
aI33
aI75
aI65
aI479
aI47
aI35
aI4023
aI4853
aI2793
aI29895
aI2711
aI83779
asS'd'
I13414
sa(dp26828
S'a'
I65441
sS's'
I17
sS'm'
(lp26829
I1
aI1
aI3
aI15
aI11
aI1
aI9
aI149
aI503
aI845
aI647
aI1233
aI4355
aI3623
aI3197
aI36015
aI24839
asS'd'
I13415
sa(dp26830
S'a'
I65444
sS's'
I17
sS'm'
(lp26831
I1
aI1
aI7
aI1
aI31
aI35
aI59
aI25
aI393
aI503
aI227
aI3243
aI301
aI11121
aI32463
aI38185
aI69969
asS'd'
I13416
sa(dp26832
S'a'
I65447
sS's'
I17
sS'm'
(lp26833
I1
aI1
aI1
aI7
aI9
aI15
aI11
aI89
aI19
aI605
aI1657
aI3335
aI1967
aI29
aI28619
aI42301
aI79909
asS'd'
I13417
sa(dp26834
S'a'
I65448
sS's'
I17
sS'm'
(lp26835
I1
aI3
aI7
aI7
aI19
aI29
aI111
aI55
aI299
aI733
aI547
aI395
aI4831
aI1991
aI7357
aI25781
aI115129
asS'd'
I13418
sa(dp26836
S'a'
I65473
sS's'
I17
sS'm'
(lp26837
I1
aI1
aI3
aI7
aI5
aI33
aI31
aI101
aI163
aI389
aI1163
aI1843
aI4105
aI14209
aI29261
aI5821
aI17929
asS'd'
I13419
sa(dp26838
S'a'
I65476
sS's'
I17
sS'm'
(lp26839
I1
aI3
aI5
aI1
aI11
aI23
aI53
aI227
aI497
aI695
aI313
aI3305
aI6549
aI15401
aI9339
aI40283
aI60531
asS'd'
I13420
sa(dp26840
S'a'
I65480
sS's'
I17
sS'm'
(lp26841
I1
aI3
aI5
aI5
aI3
aI29
aI77
aI149
aI509
aI747
aI85
aI2561
aI4435
aI14475
aI22887
aI38177
aI24535
asS'd'
I13421
sa(dp26842
S'a'
I65494
sS's'
I17
sS'm'
(lp26843
I1
aI1
aI7
aI1
aI1
aI33
aI7
aI77
aI153
aI369
aI689
aI3325
aI1173
aI16203
aI1499
aI36627
aI66915
asS'd'
I13422
sa(dp26844
S'a'
I65509
sS's'
I17
sS'm'
(lp26845
I1
aI1
aI5
aI11
aI23
aI61
aI95
aI61
aI289
aI71
aI653
aI2817
aI365
aI7391
aI1613
aI48901
aI57471
asS'd'
I13423
sa(dp26846
S'a'
I65519
sS's'
I17
sS'm'
(lp26847
I1
aI3
aI3
aI7
aI15
aI29
aI65
aI133
aI15
aI921
aI1601
aI1941
aI6917
aI10945
aI20101
aI59809
aI9017
asS'd'
I13424
sa(dp26848
S'a'
I65527
sS's'
I17
sS'm'
(lp26849
I1
aI3
aI3
aI15
aI7
aI51
aI95
aI53
aI87
aI1017
aI1039
aI3405
aI1967
aI9855
aI4905
aI4651
aI83487
asS'd'
I13425
sa(dp26850
S'a'
I19
sS's'
I18
sS'm'
(lp26851
I1
aI3
aI5
aI13
aI23
aI27
aI31
aI179
aI121
aI597
aI829
aI4003
aI2487
aI3977
aI3087
aI26791
aI28305
aI138357
asS'd'
I13426
sa(dp26852
S'a'
I31
sS's'
I18
sS'm'
(lp26853
I1
aI1
aI5
aI5
aI5
aI39
aI95
aI117
aI461
aI117
aI109
aI2571
aI7651
aI12361
aI17921
aI555
aI33353
aI186427
asS'd'
I13427
sa(dp26854
S'a'
I38
sS's'
I18
sS'm'
(lp26855
I1
aI1
aI3
aI13
aI23
aI47
aI89
aI125
aI271
aI609
aI215
aI3861
aI6883
aI3217
aI2547
aI54943
aI60565
aI215939
asS'd'
I13428
sa(dp26856
S'a'
I61
sS's'
I18
sS'm'
(lp26857
I1
aI1
aI7
aI15
aI25
aI61
aI47
aI93
aI219
aI919
aI1551
aI1417
aI2753
aI4353
aI9201
aI46423
aI31227
aI150649
asS'd'
I13429
sa(dp26858
S'a'
I64
sS's'
I18
sS'm'
(lp26859
I1
aI1
aI1
aI3
aI29
aI39
aI11
aI61
aI137
aI809
aI147
aI2715
aI5455
aI9431
aI5725
aI46135
aI118193
aI54099
asS'd'
I13430
sa(dp26860
S'a'
I109
sS's'
I18
sS'm'
(lp26861
I1
aI1
aI3
aI1
aI25
aI37
aI83
aI211
aI423
aI779
aI1731
aI2827
aI883
aI10477
aI28771
aI21723
aI114333
aI56293
asS'd'
I13431
sa(dp26862
S'a'
I115
sS's'
I18
sS'm'
(lp26863
I1
aI1
aI5
aI5
aI27
aI17
aI21
aI125
aI495
aI655
aI1803
aI3555
aI1997
aI15593
aI29705
aI48537
aI53935
aI179773
asS'd'
I13432
sa(dp26864
S'a'
I118
sS's'
I18
sS'm'
(lp26865
I1
aI1
aI7
aI5
aI19
aI63
aI55
aI15
aI469
aI769
aI967
aI3047
aI1713
aI11655
aI15313
aI29965
aI78857
aI223391
asS'd'
I13433
sa(dp26866
S'a'
I131
sS's'
I18
sS'm'
(lp26867
I1
aI3
aI3
aI5
aI27
aI33
aI51
aI171
aI417
aI243
aI1203
aI3505
aI2533
aI2695
aI219
aI57423
aI5145
aI143165
asS'd'
I13434
sa(dp26868
S'a'
I167
sS's'
I18
sS'm'
(lp26869
I1
aI3
aI5
aI9
aI5
aI19
aI95
aI97
aI1
aI863
aI693
aI2977
aI4839
aI6649
aI22587
aI40745
aI113839
aI69131
asS'd'
I13435
sa(dp26870
S'a'
I200
sS's'
I18
sS'm'
(lp26871
I1
aI3
aI5
aI1
aI31
aI39
aI53
aI85
aI509
aI5
aI359
aI1947
aI3279
aI5433
aI21763
aI46713
aI37289
aI35911
asS'd'
I13436
sa(dp26872
S'a'
I241
sS's'
I18
sS'm'
(lp26873
I1
aI3
aI7
aI13
aI17
aI35
aI59
aI63
aI95
aI667
aI1775
aI2165
aI7861
aI15731
aI12159
aI36179
aI115457
aI184819
asS'd'
I13437
sa(dp26874
S'a'
I244
sS's'
I18
sS'm'
(lp26875
I1
aI3
aI3
aI15
aI19
aI51
aI7
aI83
aI367
aI573
aI503
aI535
aI333
aI13041
aI7187
aI14479
aI57473
aI242951
asS'd'
I13438
sa(dp26876
S'a'
I247
sS's'
I18
sS'm'
(lp26877
I1
aI3
aI5
aI1
aI7
aI27
aI65
aI201
aI365
aI445
aI985
aI1175
aI6391
aI7345
aI19935
aI29085
aI103001
aI231855
asS'd'
I13439
sa(dp26878
S'a'
I261
sS's'
I18
sS'm'
(lp26879
I1
aI3
aI5
aI13
aI15
aI61
aI95
aI125
aI135
aI217
aI1787
aI417
aI7641
aI11825
aI14531
aI48497
aI125087
aI73279
asS'd'
I13440
sa(dp26880
S'a'
I265
sS's'
I18
sS'm'
(lp26881
I1
aI1
aI5
aI13
aI7
aI25
aI77
aI99
aI341
aI447
aI1711
aI137
aI2749
aI3465
aI26255
aI719
aI102595
aI112825
asS'd'
I13441
sa(dp26882
S'a'
I304
sS's'
I18
sS'm'
(lp26883
I1
aI1
aI7
aI7
aI15
aI13
aI127
aI57
aI359
aI591
aI713
aI409
aI1293
aI4979
aI7035
aI11369
aI85255
aI207241
asS'd'
I13442
sa(dp26884
S'a'
I314
sS's'
I18
sS'm'
(lp26885
I1
aI3
aI3
aI5
aI1
aI45
aI123
aI183
aI297
aI375
aI1269
aI1197
aI2389
aI6269
aI24549
aI44643
aI75893
aI161509
asS'd'
I13443
sa(dp26886
S'a'
I341
sS's'
I18
sS'm'
(lp26887
I1
aI1
aI5
aI7
aI17
aI55
aI67
aI51
aI449
aI383
aI2037
aI871
aI1359
aI15317
aI22055
aI4655
aI18065
aI258271
asS'd'
I13444
sa(dp26888
S'a'
I376
sS's'
I18
sS'm'
(lp26889
I1
aI1
aI3
aI11
aI21
aI27
aI59
aI205
aI145
aI195
aI1747
aI1121
aI1061
aI8879
aI31455
aI56541
aI74765
aI183047
asS'd'
I13445
sa(dp26890
S'a'
I395
sS's'
I18
sS'm'
(lp26891
I1
aI1
aI5
aI15
aI1
aI11
aI69
aI157
aI13
aI185
aI1355
aI467
aI4383
aI13103
aI21679
aI35169
aI33427
aI32113
asS'd'
I13446
sa(dp26892
S'a'
I405
sS's'
I18
sS'm'
(lp26893
I1
aI3
aI1
aI3
aI29
aI41
aI15
aI209
aI313
aI61
aI1749
aI2457
aI1897
aI15595
aI24441
aI39913
aI40499
aI5179
asS'd'
I13447
sa(dp26894
S'a'
I406
sS's'
I18
sS'm'
(lp26895
I1
aI3
aI7
aI15
aI25
aI41
aI87
aI125
aI239
aI73
aI207
aI2043
aI1133
aI12845
aI8533
aI16339
aI117913
aI118677
asS'd'
I13448
sa(dp26896
S'a'
I443
sS's'
I18
sS'm'
(lp26897
I1
aI1
aI3
aI15
aI25
aI9
aI15
aI97
aI395
aI99
aI2017
aI1003
aI847
aI2535
aI11753
aI54769
aI54011
aI73541
asS'd'
I13449
sa(dp26898
S'a'
I451
sS's'
I18
sS'm'
(lp26899
I1
aI1
aI7
aI15
aI11
aI61
aI13
aI49
aI319
aI871
aI893
aI165
aI3957
aI8683
aI31197
aI39491
aI58705
aI213411
asS'd'
I13450
sa(dp26900
S'a'
I458
sS's'
I18
sS'm'
(lp26901
I1
aI3
aI7
aI3
aI17
aI43
aI29
aI81
aI461
aI595
aI541
aI243
aI5587
aI13083
aI29981
aI16187
aI124601
aI89543
asS'd'
I13451
sa(dp26902
S'a'
I460
sS's'
I18
sS'm'
(lp26903
I1
aI1
aI5
aI11
aI7
aI5
aI61
aI43
aI445
aI115
aI1705
aI419
aI4627
aI15063
aI16053
aI26249
aI112243
aI208711
asS'd'
I13452
sa(dp26904
S'a'
I468
sS's'
I18
sS'm'
(lp26905
I1
aI3
aI3
aI9
aI27
aI21
aI89
aI49
aI41
aI859
aI681
aI2043
aI7445
aI9591
aI13443
aI36981
aI66785
aI227899
asS'd'
I13453
sa(dp26906
S'a'
I472
sS's'
I18
sS'm'
(lp26907
I1
aI1
aI3
aI5
aI11
aI55
aI51
aI45
aI41
aI739
aI1199
aI191
aI4563
aI4035
aI3657
aI12189
aI52879
aI33961
asS'd'
I13454
sa(dp26908
S'a'
I482
sS's'
I18
sS'm'
(lp26909
I1
aI1
aI1
aI3
aI17
aI59
aI47
aI217
aI389
aI783
aI1501
aI517
aI6311
aI7903
aI1371
aI50617
aI41723
aI116473
asS'd'
I13455
sa(dp26910
S'a'
I491
sS's'
I18
sS'm'
(lp26911
I1
aI1
aI5
aI13
aI29
aI39
aI101
aI203
aI101
aI479
aI1337
aI2647
aI6447
aI563
aI2593
aI16533
aI122535
aI25587
asS'd'
I13456
sa(dp26912
S'a'
I496
sS's'
I18
sS'm'
(lp26913
I1
aI1
aI3
aI3
aI27
aI21
aI75
aI173
aI289
aI279
aI665
aI3177
aI559
aI8539
aI10903
aI16779
aI128219
aI125907
asS'd'
I13457
sa(dp26914
S'a'
I524
sS's'
I18
sS'm'
(lp26915
I1
aI1
aI1
aI11
aI27
aI1
aI61
aI247
aI113
aI585
aI331
aI3443
aI5939
aI5213
aI27289
aI57057
aI17349
aI62359
asS'd'
I13458
sa(dp26916
S'a'
I536
sS's'
I18
sS'm'
(lp26917
I1
aI3
aI5
aI15
aI21
aI41
aI67
aI47
aI121
aI11
aI545
aI3609
aI7745
aI3669
aI9045
aI8377
aI97655
aI99631
asS'd'
I13459
sa(dp26918
S'a'
I542
sS's'
I18
sS'm'
(lp26919
I1
aI3
aI5
aI9
aI11
aI15
aI111
aI61
aI67
aI775
aI579
aI3421
aI7827
aI13607
aI32373
aI43531
aI86149
aI238827
asS'd'
I13460
sa(dp26920
S'a'
I557
sS's'
I18
sS'm'
(lp26921
I1
aI1
aI1
aI1
aI9
aI45
aI79
aI153
aI331
aI399
aI1777
aI3515
aI3363
aI3499
aI13461
aI48651
aI21731
aI220611
asS'd'
I13461
sa(dp26922
S'a'
I572
sS's'
I18
sS'm'
(lp26923
I1
aI1
aI1
aI1
aI31
aI57
aI117
aI223
aI139
aI725
aI1115
aI3203
aI8185
aI11983
aI20245
aI55913
aI36803
aI68101
asS'd'
I13462
sa(dp26924
S'a'
I580
sS's'
I18
sS'm'
(lp26925
I1
aI1
aI1
aI3
aI31
aI57
aI53
aI79
aI225
aI307
aI1645
aI3311
aI643
aI6587
aI12037
aI12453
aI83461
aI195503
asS'd'
I13463
sa(dp26926
S'a'
I592
sS's'
I18
sS'm'
(lp26927
I1
aI3
aI1
aI7
aI23
aI25
aI65
aI233
aI273
aI97
aI37
aI1563
aI3635
aI9299
aI24367
aI42761
aI55
aI128675
asS'd'
I13464
sa(dp26928
S'a'
I656
sS's'
I18
sS'm'
(lp26929
I1
aI3
aI3
aI11
aI29
aI21
aI97
aI143
aI447
aI345
aI389
aI381
aI1403
aI685
aI309
aI11103
aI69769
aI194441
asS'd'
I13465
sa(dp26930
S'a'
I713
sS's'
I18
sS'm'
(lp26931
I1
aI3
aI3
aI11
aI23
aI55
aI119
aI71
aI23
aI291
aI1241
aI1723
aI5025
aI4499
aI26617
aI22875
aI62185
aI240321
asS'd'
I13466
sa(dp26932
S'a'
I719
sS's'
I18
sS'm'
(lp26933
I1
aI1
aI7
aI11
aI19
aI63
aI31
aI131
aI393
aI99
aI1061
aI3805
aI7477
aI15357
aI8269
aI26067
aI113349
aI239333
asS'd'
I13467
sa(dp26934
S'a'
I738
sS's'
I18
sS'm'
(lp26935
I1
aI3
aI5
aI1
aI5
aI37
aI77
aI83
aI37
aI759
aI1297
aI3067
aI5369
aI5977
aI7531
aI49079
aI94503
aI192765
asS'd'
I13468
sa(dp26936
S'a'
I749
sS's'
I18
sS'm'
(lp26937
I1
aI1
aI7
aI1
aI23
aI9
aI119
aI137
aI469
aI73
aI2001
aI2629
aI2681
aI2295
aI2055
aI44027
aI47627
aI45283
asS'd'
I13469
sa(dp26938
S'a'
I752
sS's'
I18
sS'm'
(lp26939
I1
aI3
aI1
aI7
aI31
aI17
aI61
aI137
aI241
aI325
aI1417
aI2383
aI4171
aI2495
aI215
aI59593
aI98495
aI74727
asS'd'
I13470
sa(dp26940
S'a'
I767
sS's'
I18
sS'm'
(lp26941
I1
aI1
aI7
aI13
aI7
aI5
aI59
aI189
aI131
aI865
aI1963
aI1811
aI5629
aI16189
aI16397
aI58069
aI72081
aI191457
asS'd'
I13471
sa(dp26942
S'a'
I772
sS's'
I18
sS'm'
(lp26943
I1
aI1
aI7
aI15
aI23
aI33
aI93
aI247
aI395
aI643
aI693
aI3587
aI4375
aI5519
aI9449
aI37515
aI11455
aI218337
asS'd'
I13472
sa(dp26944
S'a'
I782
sS's'
I18
sS'm'
(lp26945
I1
aI1
aI3
aI1
aI27
aI63
aI113
aI91
aI477
aI55
aI1461
aI1547
aI4743
aI699
aI21639
aI1815
aI169
aI34239
asS'd'
I13473
sa(dp26946
S'a'
I789
sS's'
I18
sS'm'
(lp26947
I1
aI1
aI5
aI15
aI29
aI37
aI19
aI19
aI247
aI771
aI695
aI319
aI1779
aI10553
aI16165
aI60507
aI87161
aI86967
asS'd'
I13474
sa(dp26948
S'a'
I830
sS's'
I18
sS'm'
(lp26949
I1
aI1
aI7
aI1
aI25
aI61
aI13
aI167
aI251
aI861
aI1717
aI1533
aI7323
aI3945
aI20879
aI37759
aI129689
aI35901
asS'd'
I13475
sa(dp26950
S'a'
I838
sS's'
I18
sS'm'
(lp26951
I1
aI3
aI3
aI7
aI7
aI61
aI11
aI25
aI187
aI949
aI1393
aI1743
aI745
aI16313
aI5293
aI16921
aI17619
aI237705
asS'd'
I13476
sa(dp26952
S'a'
I916
sS's'
I18
sS'm'
(lp26953
I1
aI3
aI5
aI11
aI7
aI27
aI11
aI107
aI299
aI711
aI149
aI1581
aI7747
aI14285
aI6411
aI52209
aI79043
aI61117
asS'd'
I13477
sa(dp26954
S'a'
I920
sS's'
I18
sS'm'
(lp26955
I1
aI1
aI5
aI1
aI17
aI19
aI91
aI185
aI53
aI699
aI1185
aI4007
aI1099
aI1965
aI20239
aI19547
aI120859
aI234149
asS'd'
I13478
sa(dp26956
S'a'
I936
sS's'
I18
sS'm'
(lp26957
I1
aI1
aI5
aI5
aI13
aI61
aI117
aI187
aI149
aI957
aI837
aI3549
aI6221
aI501
aI24755
aI47975
aI67007
aI12329
asS'd'
I13479
sa(dp26958
S'a'
I991
sS's'
I18
sS'm'
(lp26959
I1
aI1
aI3
aI15
aI21
aI41
aI55
aI81
aI397
aI403
aI1699
aI1057
aI6125
aI11987
aI3103
aI43361
aI21277
aI156577
asS'd'
I13480
sa(dp26960
S'a'
I998
sS's'
I18
sS'm'
(lp26961
I1
aI1
aI5
aI11
aI5
aI27
aI5
aI177
aI387
aI859
aI809
aI3919
aI4085
aI1535
aI6009
aI13265
aI3065
aI217945
asS'd'
I13481
sa(dp26962
S'a'
I1016
sS's'
I18
sS'm'
(lp26963
I1
aI3
aI1
aI13
aI15
aI57
aI107
aI81
aI437
aI305
aI879
aI1691
aI3685
aI11415
aI3749
aI46999
aI113933
aI10515
asS'd'
I13482
sa(dp26964
S'a'
I1024
sS's'
I18
sS'm'
(lp26965
I1
aI1
aI7
aI13
aI9
aI43
aI59
aI223
aI189
aI329
aI829
aI2033
aI1835
aI8255
aI8121
aI46463
aI61433
aI86453
asS'd'
I13483
sa(dp26966
S'a'
I1053
sS's'
I18
sS'm'
(lp26967
I1
aI3
aI1
aI9
aI11
aI49
aI63
aI125
aI11
aI987
aI2017
aI2623
aI4753
aI13889
aI57
aI24755
aI108489
aI175383
asS'd'
I13484
sa(dp26968
S'a'
I1081
sS's'
I18
sS'm'
(lp26969
I1
aI1
aI1
aI3
aI25
aI33
aI39
aI151
aI405
aI657
aI1755
aI957
aI5557
aI7611
aI25839
aI51385
aI92713
aI64009
asS'd'
I13485
sa(dp26970
S'a'
I1090
sS's'
I18
sS'm'
(lp26971
I1
aI3
aI7
aI9
aI17
aI17
aI115
aI89
aI225
aI715
aI1085
aI543
aI1047
aI15053
aI14359
aI43301
aI31455
aI156555
asS'd'
I13486
sa(dp26972
S'a'
I1125
sS's'
I18
sS'm'
(lp26973
I1
aI1
aI7
aI11
aI11
aI21
aI115
aI5
aI371
aI1003
aI1053
aI1713
aI5921
aI7277
aI799
aI62483
aI28079
aI222319
asS'd'
I13487
sa(dp26974
S'a'
I1135
sS's'
I18
sS'm'
(lp26975
I1
aI1
aI3
aI3
aI31
aI15
aI127
aI213
aI459
aI229
aI1477
aI1863
aI1021
aI14881
aI16299
aI5953
aI121455
aI49659
asS'd'
I13488
sa(dp26976
S'a'
I1143
sS's'
I18
sS'm'
(lp26977
I1
aI1
aI5
aI9
aI3
aI39
aI87
aI219
aI57
aI479
aI69
aI2777
aI8105
aI11975
aI14743
aI26205
aI93303
aI45311
asS'd'
I13489
sa(dp26978
S'a'
I1150
sS's'
I18
sS'm'
(lp26979
I1
aI1
aI5
aI13
aI3
aI43
aI55
aI139
aI19
aI715
aI2035
aI2993
aI2945
aI9075
aI6275
aI32233
aI103127
aI49523
asS'd'
I13490
sa(dp26980
S'a'
I1154
sS's'
I18
sS'm'
(lp26981
I1
aI3
aI1
aI13
aI19
aI31
aI109
aI211
aI261
aI231
aI697
aI383
aI2173
aI14617
aI11877
aI37009
aI5485
aI236549
asS'd'
I13491
sa(dp26982
S'a'
I1171
sS's'
I18
sS'm'
(lp26983
I1
aI1
aI5
aI3
aI5
aI23
aI91
aI115
aI369
aI11
aI1021
aI519
aI655
aI4461
aI23743
aI56981
aI51687
aI114845
asS'd'
I13492
sa(dp26984
S'a'
I1174
sS's'
I18
sS'm'
(lp26985
I1
aI3
aI7
aI9
aI29
aI23
aI19
aI127
aI17
aI369
aI1537
aI2705
aI4993
aI1869
aI15447
aI28127
aI73609
aI97683
asS'd'
I13493
sa(dp26986
S'a'
I1202
sS's'
I18
sS'm'
(lp26987
I1
aI1
aI3
aI9
aI17
aI61
aI97
aI187
aI213
aI861
aI725
aI3205
aI103
aI12729
aI2915
aI28389
aI83123
aI124065
asS'd'
I13494
sa(dp26988
S'a'
I1213
sS's'
I18
sS'm'
(lp26989
I1
aI3
aI1
aI1
aI5
aI61
aI47
aI187
aI471
aI137
aI1595
aI707
aI2449
aI14315
aI16409
aI41467
aI37533
aI1649
asS'd'
I13495
sa(dp26990
S'a'
I1225
sS's'
I18
sS'm'
(lp26991
I1
aI3
aI5
aI5
aI7
aI39
aI1
aI245
aI361
aI43
aI1259
aI3149
aI3449
aI15723
aI6225
aI27445
aI80529
aI215349
asS'd'
I13496
sa(dp26992
S'a'
I1233
sS's'
I18
sS'm'
(lp26993
I1
aI3
aI3
aI11
aI17
aI27
aI37
aI47
aI157
aI345
aI1437
aI3219
aI5663
aI7299
aI23925
aI34067
aI102379
aI42767
asS'd'
I13497
sa(dp26994
S'a'
I1234
sS's'
I18
sS'm'
(lp26995
I1
aI3
aI5
aI13
aI21
aI59
aI43
aI189
aI17
aI303
aI1949
aI3627
aI3495
aI7981
aI18115
aI34221
aI43511
aI255257
asS'd'
I13498
sa(dp26996
S'a'
I1252
sS's'
I18
sS'm'
(lp26997
I1
aI3
aI7
aI15
aI3
aI29
aI81
aI243
aI321
aI853
aI595
aI2451
aI1713
aI11859
aI27689
aI12849
aI24505
aI9547
asS'd'
I13499
sa(dp26998
S'a'
I1255
sS's'
I18
sS'm'
(lp26999
I1
aI3
aI1
aI3
aI7
aI7
aI89
aI183
aI51
aI901
aI253
aI2421
aI7453
aI15827
aI21451
aI58653
aI51933
aI239113
asS'd'
I13500
sa(dp27000
S'a'
I1294
sS's'
I18
sS'm'
(lp27001
I1
aI1
aI7
aI3
aI21
aI59
aI93
aI25
aI219
aI805
aI1699
aI3777
aI3683
aI5351
aI5481
aI44797
aI651
aI32161
asS'd'
I13501
sa(dp27002
S'a'
I1349
sS's'
I18
sS'm'
(lp27003
I1
aI3
aI7
aI5
aI31
aI15
aI15
aI167
aI305
aI545
aI331
aI3765
aI8191
aI5763
aI16965
aI7239
aI73735
aI1049
asS'd'
I13502
sa(dp27004
S'a'
I1354
sS's'
I18
sS'm'
(lp27005
I1
aI3
aI1
aI15
aI13
aI19
aI59
aI107
aI213
aI39
aI1547
aI3413
aI6175
aI16195
aI4635
aI8945
aI60301
aI196697
asS'd'
I13503
sa(dp27006
S'a'
I1378
sS's'
I18
sS'm'
(lp27007
I1
aI1
aI3
aI1
aI29
aI17
aI51
aI61
aI261
aI951
aI643
aI2329
aI2235
aI9171
aI11265
aI3523
aI89781
aI227125
asS'd'
I13504
sa(dp27008
S'a'
I1383
sS's'
I18
sS'm'
(lp27009
I1
aI3
aI5
aI1
aI1
aI51
aI75
aI199
aI479
aI899
aI1425
aI3697
aI2039
aI4503
aI11789
aI16853
aI94607
aI236887
asS'd'
I13505
sa(dp27010
S'a'
I1387
sS's'
I18
sS'm'
(lp27011
I1
aI3
aI1
aI9
aI19
aI43
aI111
aI41
aI385
aI677
aI1067
aI3391
aI7819
aI13663
aI17713
aI10155
aI124243
aI56005
asS'd'
I13506
sa(dp27012
S'a'
I1392
sS's'
I18
sS'm'
(lp27013
I1
aI3
aI5
aI3
aI15
aI3
aI105
aI23
aI307
aI955
aI843
aI1277
aI6697
aI11903
aI8901
aI36129
aI51685
aI251115
asS'd'
I13507
sa(dp27014
S'a'
I1402
sS's'
I18
sS'm'
(lp27015
I1
aI3
aI1
aI5
aI27
aI35
aI95
aI57
aI207
aI49
aI1559
aI171
aI4703
aI511
aI4169
aI23241
aI111447
aI173109
asS'd'
I13508
sa(dp27016
S'a'
I1420
sS's'
I18
sS'm'
(lp27017
I1
aI3
aI1
aI13
aI23
aI5
aI31
aI15
aI223
aI673
aI1333
aI2243
aI2479
aI7489
aI31891
aI33909
aI96803
aI227027
asS'd'
I13509
sa(dp27018
S'a'
I1428
sS's'
I18
sS'm'
(lp27019
I1
aI3
aI5
aI11
aI5
aI45
aI19
aI13
aI367
aI475
aI1719
aI3947
aI5295
aI2319
aI20697
aI181
aI16925
aI80239
asS'd'
I13510
sa(dp27020
S'a'
I1437
sS's'
I18
sS'm'
(lp27021
I1
aI1
aI5
aI13
aI15
aI47
aI89
aI15
aI153
aI73
aI523
aI3529
aI5401
aI15881
aI13779
aI32123
aI82347
aI58749
asS'd'
I13511
sa(dp27022
S'a'
I1448
sS's'
I18
sS'm'
(lp27023
I1
aI3
aI5
aI7
aI5
aI7
aI123
aI217
aI261
aI65
aI685
aI2175
aI3289
aI7473
aI17857
aI48335
aI94183
aI216857
asS'd'
I13512
sa(dp27024
S'a'
I1459
sS's'
I18
sS'm'
(lp27025
I1
aI3
aI7
aI13
aI7
aI23
aI85
aI25
aI231
aI19
aI1179
aI2705
aI6433
aI10827
aI1969
aI51521
aI76775
aI260291
asS'd'
I13513
sa(dp27026
S'a'
I1473
sS's'
I18
sS'm'
(lp27027
I1
aI3
aI3
aI13
aI9
aI39
aI5
aI141
aI475
aI777
aI1809
aI1975
aI2347
aI12611
aI28303
aI15239
aI45429
aI170015
asS'd'
I13514
sa(dp27028
S'a'
I1507
sS's'
I18
sS'm'
(lp27029
I1
aI1
aI7
aI7
aI31
aI31
aI39
aI19
aI317
aI897
aI739
aI275
aI2261
aI16013
aI1123
aI33181
aI96603
aI37563
asS'd'
I13515
sa(dp27030
S'a'
I1516
sS's'
I18
sS'm'
(lp27031
I1
aI1
aI7
aI13
aI31
aI55
aI87
aI239
aI193
aI435
aI625
aI2153
aI3979
aI15537
aI19937
aI50621
aI48273
aI31381
asS'd'
I13516
sa(dp27032
S'a'
I1528
sS's'
I18
sS'm'
(lp27033
I1
aI1
aI1
aI15
aI1
aI57
aI73
aI237
aI361
aI749
aI379
aI2511
aI501
aI10783
aI2787
aI36983
aI12393
aI14345
asS'd'
I13517
sa(dp27034
S'a'
I1573
sS's'
I18
sS'm'
(lp27035
I1
aI1
aI1
aI3
aI25
aI33
aI85
aI25
aI83
aI939
aI139
aI2601
aI6385
aI16041
aI28463
aI38977
aI28163
aI232165
asS'd'
I13518
sa(dp27036
S'a'
I1592
sS's'
I18
sS'm'
(lp27037
I1
aI3
aI5
aI3
aI9
aI19
aI119
aI171
aI499
aI19
aI569
aI353
aI1619
aI6235
aI24431
aI47401
aI48125
aI168819
asS'd'
I13519
sa(dp27038
S'a'
I1597
sS's'
I18
sS'm'
(lp27039
I1
aI1
aI1
aI3
aI9
aI27
aI121
aI137
aI411
aI391
aI1437
aI1339
aI7475
aI3889
aI15451
aI34809
aI69807
aI162851
asS'd'
I13520
sa(dp27040
S'a'
I1654
sS's'
I18
sS'm'
(lp27041
I1
aI3
aI5
aI11
aI31
aI39
aI41
aI3
aI171
aI35
aI81
aI2713
aI1077
aI10697
aI12343
aI52133
aI52825
aI152255
asS'd'
I13521
sa(dp27042
S'a'
I1663
sS's'
I18
sS'm'
(lp27043
I1
aI1
aI3
aI11
aI17
aI51
aI83
aI19
aI357
aI207
aI897
aI2167
aI1333
aI4111
aI29295
aI65371
aI73447
aI61765
asS'd'
I13522
sa(dp27044
S'a'
I1730
sS's'
I18
sS'm'
(lp27045
I1
aI1
aI3
aI7
aI9
aI59
aI17
aI135
aI365
aI931
aI1203
aI277
aI5531
aI4213
aI12969
aI2617
aI591
aI154539
asS'd'
I13523
sa(dp27046
S'a'
I1739
sS's'
I18
sS'm'
(lp27047
I1
aI3
aI7
aI11
aI1
aI53
aI31
aI49
aI135
aI603
aI227
aI911
aI7371
aI8559
aI27195
aI33065
aI71351
aI245255
asS'd'
I13524
sa(dp27048
S'a'
I1741
sS's'
I18
sS'm'
(lp27049
I1
aI1
aI7
aI1
aI15
aI5
aI31
aI135
aI197
aI791
aI1531
aI2567
aI2545
aI15515
aI25417
aI27431
aI15571
aI176829
asS'd'
I13525
sa(dp27050
S'a'
I1753
sS's'
I18
sS'm'
(lp27051
I1
aI3
aI1
aI1
aI15
aI7
aI89
aI217
aI505
aI859
aI1329
aI2285
aI7921
aI11839
aI7699
aI56867
aI112483
aI3895
asS'd'
I13526
sa(dp27052
S'a'
I1783
sS's'
I18
sS'm'
(lp27053
I1
aI3
aI1
aI3
aI27
aI57
aI37
aI117
aI491
aI815
aI275
aI381
aI7443
aI3297
aI1523
aI34211
aI97589
aI232261
asS'd'
I13527
sa(dp27054
S'a'
I1804
sS's'
I18
sS'm'
(lp27055
I1
aI1
aI3
aI3
aI29
aI63
aI69
aI153
aI297
aI423
aI1435
aI3927
aI7265
aI13223
aI17607
aI21201
aI57929
aI73037
asS'd'
I13528
sa(dp27056
S'a'
I1807
sS's'
I18
sS'm'
(lp27057
I1
aI3
aI1
aI3
aI23
aI41
aI1
aI167
aI121
aI217
aI973
aI2149
aI3807
aI9895
aI29635
aI1625
aI99829
aI218541
asS'd'
I13529
sa(dp27058
S'a'
I1832
sS's'
I18
sS'm'
(lp27059
I1
aI3
aI5
aI7
aI31
aI33
aI53
aI165
aI51
aI119
aI7
aI1655
aI6521
aI5481
aI9503
aI6833
aI80483
aI252111
asS'd'
I13530
sa(dp27060
S'a'
I1850
sS's'
I18
sS'm'
(lp27061
I1
aI1
aI7
aI1
aI5
aI63
aI25
aI219
aI165
aI893
aI1665
aI2789
aI1113
aI9277
aI3151
aI12625
aI82403
aI59749
asS'd'
I13531
sa(dp27062
S'a'
I1852
sS's'
I18
sS'm'
(lp27063
I1
aI3
aI7
aI3
aI21
aI13
aI127
aI127
aI145
aI993
aI715
aI1947
aI7501
aI4385
aI11759
aI2179
aI26039
aI28027
asS'd'
I13532
sa(dp27064
S'a'
I1881
sS's'
I18
sS'm'
(lp27065
I1
aI3
aI5
aI9
aI23
aI27
aI123
aI1
aI231
aI709
aI1615
aI1433
aI5991
aI1045
aI16269
aI123
aI110249
aI154819
asS'd'
I13533
sa(dp27066
S'a'
I1894
sS's'
I18
sS'm'
(lp27067
I1
aI1
aI1
aI5
aI17
aI11
aI123
aI151
aI387
aI905
aI991
aI1571
aI4463
aI6765
aI31905
aI59307
aI75175
aI204571
asS'd'
I13534
sa(dp27068
S'a'
I1927
sS's'
I18
sS'm'
(lp27069
I1
aI3
aI1
aI11
aI27
aI49
aI1
aI181
aI77
aI1023
aI807
aI3479
aI7965
aI4633
aI17495
aI5991
aI77081
aI249343
asS'd'
I13535
sa(dp27070
S'a'
I1952
sS's'
I18
sS'm'
(lp27071
I1
aI3
aI1
aI1
aI13
aI53
aI105
aI79
aI269
aI173
aI1319
aI1695
aI1215
aI3651
aI25063
aI34949
aI77243
aI214671
asS'd'
I13536
sa(dp27072
S'a'
I1969
sS's'
I18
sS'm'
(lp27073
I1
aI1
aI1
aI1
aI3
aI19
aI103
aI233
aI1
aI507
aI721
aI1797
aI5025
aI405
aI13027
aI23693
aI89963
aI25771
asS'd'
I13537
sa(dp27074
S'a'
I1999
sS's'
I18
sS'm'
(lp27075
I1
aI3
aI5
aI9
aI21
aI53
aI1
aI241
aI405
aI707
aI1807
aI3615
aI1199
aI11155
aI27741
aI53931
aI55091
aI248677
asS'd'
I13538
sa(dp27076
S'a'
I2018
sS's'
I18
sS'm'
(lp27077
I1
aI3
aI5
aI7
aI27
aI27
aI39
aI77
aI475
aI845
aI1393
aI3779
aI5261
aI13017
aI13517
aI18595
aI64485
aI180577
asS'd'
I13539
sa(dp27078
S'a'
I2047
sS's'
I18
sS'm'
(lp27079
I1
aI1
aI3
aI5
aI7
aI21
aI95
aI59
aI203
aI233
aI1167
aI3457
aI3965
aI4321
aI14885
aI6335
aI78353
aI39341
asS'd'
I13540
sa(dp27080
S'a'
I2066
sS's'
I18
sS'm'
(lp27081
I1
aI1
aI7
aI13
aI27
aI19
aI27
aI133
aI419
aI507
aI945
aI3595
aI131
aI7981
aI31451
aI62347
aI19151
aI256127
asS'd'
I13541
sa(dp27082
S'a'
I2068
sS's'
I18
sS'm'
(lp27083
I1
aI3
aI7
aI3
aI7
aI15
aI9
aI173
aI257
aI983
aI223
aI2881
aI6911
aI3681
aI26183
aI38943
aI112171
aI148627
asS'd'
I13542
sa(dp27084
S'a'
I2093
sS's'
I18
sS'm'
(lp27085
I1
aI3
aI3
aI15
aI5
aI49
aI91
aI205
aI303
aI183
aI775
aI3841
aI4943
aI14417
aI23013
aI59337
aI85835
aI181771
asS'd'
I13543
sa(dp27086
S'a'
I2105
sS's'
I18
sS'm'
(lp27087
I1
aI3
aI5
aI9
aI21
aI1
aI117
aI27
aI509
aI263
aI1215
aI893
aI6677
aI3275
aI20831
aI5045
aI127323
aI62589
asS'd'
I13544
sa(dp27088
S'a'
I2116
sS's'
I18
sS'm'
(lp27089
I1
aI1
aI1
aI3
aI17
aI61
aI77
aI239
aI379
aI649
aI1151
aI2359
aI2659
aI13853
aI30589
aI55873
aI50359
aI184125
asS'd'
I13545
sa(dp27090
S'a'
I2149
sS's'
I18
sS'm'
(lp27091
I1
aI1
aI7
aI5
aI17
aI33
aI95
aI111
aI245
aI873
aI1721
aI3079
aI7753
aI12889
aI27107
aI8267
aI119413
aI249045
asS'd'
I13546
sa(dp27092
S'a'
I2201
sS's'
I18
sS'm'
(lp27093
I1
aI1
aI1
aI15
aI13
aI23
aI59
aI169
aI449
aI283
aI913
aI2099
aI5337
aI4307
aI3701
aI16395
aI112987
aI14183
asS'd'
I13547
sa(dp27094
S'a'
I2228
sS's'
I18
sS'm'
(lp27095
I1
aI1
aI5
aI3
aI5
aI15
aI3
aI249
aI97
aI849
aI1551
aI3437
aI1247
aI10915
aI24073
aI53723
aI40345
aI37215
asS'd'
I13548
sa(dp27096
S'a'
I2245
sS's'
I18
sS'm'
(lp27097
I1
aI1
aI1
aI5
aI21
aI59
aI109
aI79
aI9
aI827
aI1329
aI405
aI3821
aI8415
aI11239
aI1003
aI78967
aI112627
asS'd'
I13549
sa(dp27098
S'a'
I2246
sS's'
I18
sS'm'
(lp27099
I1
aI1
aI1
aI11
aI21
aI7
aI21
aI45
aI327
aI365
aI865
aI1409
aI1273
aI15675
aI21425
aI45367
aI22279
aI240943
asS'd'
I13550
sa(dp27100
S'a'
I2283
sS's'
I18
sS'm'
(lp27101
I1
aI3
aI3
aI7
aI3
aI19
aI83
aI163
aI381
aI547
aI195
aI1537
aI7905
aI9057
aI1309
aI41135
aI118857
aI101725
asS'd'
I13551
sa(dp27102
S'a'
I2288
sS's'
I18
sS'm'
(lp27103
I1
aI1
aI5
aI9
aI11
aI19
aI107
aI247
aI309
aI343
aI1697
aI699
aI7137
aI12815
aI18405
aI42673
aI505
aI104801
asS'd'
I13552
sa(dp27104
S'a'
I2320
sS's'
I18
sS'm'
(lp27105
I1
aI3
aI5
aI3
aI13
aI43
aI55
aI15
aI441
aI843
aI1153
aI3739
aI67
aI11053
aI30985
aI55329
aI57301
aI190991
asS'd'
I13553
sa(dp27106
S'a'
I2326
sS's'
I18
sS'm'
(lp27107
I1
aI1
aI5
aI3
aI23
aI41
aI9
aI239
aI227
aI145
aI1895
aI2645
aI945
aI6421
aI2859
aI16173
aI97043
aI234649
asS'd'
I13554
sa(dp27108
S'a'
I2386
sS's'
I18
sS'm'
(lp27109
I1
aI3
aI1
aI3
aI23
aI47
aI57
aI207
aI441
aI279
aI1951
aI3041
aI2465
aI6143
aI27669
aI41171
aI89627
aI2489
asS'd'
I13555
sa(dp27110
S'a'
I2392
sS's'
I18
sS'm'
(lp27111
I1
aI3
aI1
aI11
aI7
aI9
aI19
aI51
aI345
aI187
aI1699
aI1483
aI15
aI10321
aI25277
aI34889
aI85225
aI259071
asS'd'
I13556
sa(dp27112
S'a'
I2395
sS's'
I18
sS'm'
(lp27113
I1
aI1
aI1
aI15
aI27
aI15
aI79
aI51
aI407
aI757
aI611
aI3955
aI1123
aI14659
aI11273
aI56639
aI64727
aI183077
asS'd'
I13557
sa(dp27114
S'a'
I2413
sS's'
I18
sS'm'
(lp27115
I1
aI3
aI7
aI1
aI13
aI61
aI89
aI157
aI29
aI561
aI791
aI995
aI4233
aI11351
aI16335
aI47041
aI108671
aI120115
asS'd'
I13558
sa(dp27116
S'a'
I2419
sS's'
I18
sS'm'
(lp27117
I1
aI3
aI3
aI15
aI17
aI35
aI15
aI223
aI57
aI7
aI961
aI3327
aI7287
aI5537
aI26231
aI3289
aI106555
aI109781
asS'd'
I13559
sa(dp27118
S'a'
I2441
sS's'
I18
sS'm'
(lp27119
I1
aI3
aI7
aI15
aI17
aI3
aI25
aI121
aI349
aI995
aI1353
aI2991
aI3071
aI3583
aI26173
aI42343
aI60495
aI44035
asS'd'
I13560
sa(dp27120
S'a'
I2466
sS's'
I18
sS'm'
(lp27121
I1
aI3
aI1
aI11
aI5
aI5
aI83
aI249
aI427
aI173
aI1733
aI45
aI3277
aI7911
aI18091
aI61305
aI130251
aI31849
asS'd'
I13561
sa(dp27122
S'a'
I2477
sS's'
I18
sS'm'
(lp27123
I1
aI1
aI1
aI9
aI3
aI23
aI23
aI127
aI371
aI1011
aI573
aI1769
aI1707
aI15351
aI30077
aI61139
aI122963
aI203481
asS'd'
I13562
sa(dp27124
S'a'
I2485
sS's'
I18
sS'm'
(lp27125
I1
aI1
aI1
aI13
aI27
aI41
aI97
aI29
aI461
aI207
aI1393
aI707
aI5633
aI7155
aI13455
aI7305
aI107539
aI136413
asS'd'
I13563
sa(dp27126
S'a'
I2492
sS's'
I18
sS'm'
(lp27127
I1
aI1
aI1
aI9
aI3
aI13
aI61
aI115
aI297
aI333
aI1679
aI127
aI8049
aI3129
aI31845
aI40039
aI77087
aI6831
asS'd'
I13564
sa(dp27128
S'a'
I2495
sS's'
I18
sS'm'
(lp27129
I1
aI3
aI3
aI11
aI27
aI25
aI49
aI29
aI423
aI193
aI1955
aI2927
aI5679
aI3537
aI16911
aI47065
aI126803
aI129957
asS'd'
I13565
sa(dp27130
S'a'
I2498
sS's'
I18
sS'm'
(lp27131
I1
aI1
aI1
aI3
aI21
aI31
aI25
aI187
aI301
aI883
aI1301
aI415
aI1515
aI14761
aI227
aI24377
aI54415
aI64553
asS'd'
I13566
sa(dp27132
S'a'
I2504
sS's'
I18
sS'm'
(lp27133
I1
aI3
aI7
aI7
aI3
aI5
aI69
aI221
aI357
aI587
aI1387
aI3719
aI5355
aI10569
aI14731
aI22515
aI107237
aI1673
asS'd'
I13567
sa(dp27134
S'a'
I2515
sS's'
I18
sS'm'
(lp27135
I1
aI1
aI3
aI15
aI27
aI7
aI89
aI23
aI213
aI655
aI779
aI1641
aI1793
aI1499
aI27279
aI59423
aI56715
aI90313
asS'd'
I13568
sa(dp27136
S'a'
I2521
sS's'
I18
sS'm'
(lp27137
I1
aI3
aI3
aI7
aI3
aI33
aI85
aI181
aI509
aI327
aI353
aI1625
aI4995
aI15627
aI17071
aI31885
aI122423
aI100337
asS'd'
I13569
sa(dp27138
S'a'
I2561
sS's'
I18
sS'm'
(lp27139
I1
aI3
aI5
aI9
aI7
aI39
aI45
aI157
aI279
aI211
aI1163
aI3283
aI4419
aI10187
aI22397
aI42119
aI25105
aI163925
asS'd'
I13570
sa(dp27140
S'a'
I2579
sS's'
I18
sS'm'
(lp27141
I1
aI3
aI3
aI15
aI17
aI37
aI75
aI65
aI501
aI765
aI1171
aI2451
aI309
aI551
aI15573
aI65497
aI106435
aI20817
asS'd'
I13571
sa(dp27142
S'a'
I2604
sS's'
I18
sS'm'
(lp27143
I1
aI1
aI3
aI1
aI13
aI1
aI79
aI117
aI5
aI285
aI953
aI2401
aI2479
aI15765
aI25677
aI63611
aI91807
aI78153
asS'd'
I13572
sa(dp27144
S'a'
I2657
sS's'
I18
sS'm'
(lp27145
I1
aI1
aI5
aI15
aI1
aI7
aI123
aI159
aI217
aI307
aI1779
aI2625
aI101
aI13887
aI31721
aI55769
aI94899
aI183427
asS'd'
I13573
sa(dp27146
S'a'
I2681
sS's'
I18
sS'm'
(lp27147
I1
aI3
aI5
aI5
aI11
aI13
aI59
aI205
aI221
aI871
aI753
aI823
aI547
aI11055
aI31621
aI54379
aI23631
aI137027
asS'd'
I13574
sa(dp27148
S'a'
I2691
sS's'
I18
sS'm'
(lp27149
I1
aI3
aI3
aI7
aI5
aI17
aI7
aI31
aI37
aI237
aI1633
aI969
aI4123
aI6643
aI28499
aI3277
aI130223
aI37465
asS'd'
I13575
sa(dp27150
S'a'
I2731
sS's'
I18
sS'm'
(lp27151
I1
aI3
aI7
aI5
aI29
aI41
aI65
aI159
aI487
aI61
aI1217
aI4093
aI487
aI15257
aI13379
aI46641
aI88043
aI107425
asS'd'
I13576
sa(dp27152
S'a'
I2739
sS's'
I18
sS'm'
(lp27153
I1
aI1
aI7
aI7
aI19
aI29
aI87
aI119
aI13
aI877
aI467
aI2661
aI7733
aI9303
aI20069
aI8445
aI126159
aI69421
asS'd'
I13577
sa(dp27154
S'a'
I2765
sS's'
I18
sS'm'
(lp27155
I1
aI3
aI1
aI13
aI1
aI57
aI77
aI241
aI185
aI479
aI859
aI2397
aI1167
aI6545
aI20715
aI50701
aI107781
aI149965
asS'd'
I13578
sa(dp27156
S'a'
I2790
sS's'
I18
sS'm'
(lp27157
I1
aI1
aI5
aI1
aI1
aI3
aI19
aI31
aI473
aI685
aI1455
aI1537
aI1843
aI4051
aI17475
aI56717
aI70257
aI112815
asS'd'
I13579
sa(dp27158
S'a'
I2802
sS's'
I18
sS'm'
(lp27159
I1
aI3
aI7
aI15
aI9
aI21
aI19
aI201
aI13
aI551
aI1053
aI1291
aI3793
aI7923
aI30425
aI55513
aI30033
aI70597
asS'd'
I13580
sa(dp27160
S'a'
I2819
sS's'
I18
sS'm'
(lp27161
I1
aI3
aI1
aI15
aI21
aI47
aI127
aI117
aI199
aI655
aI1979
aI1291
aI8017
aI11769
aI9071
aI12029
aI112369
aI2529
asS'd'
I13581
sa(dp27162
S'a'
I2891
sS's'
I18
sS'm'
(lp27163
I1
aI3
aI5
aI1
aI15
aI3
aI25
aI199
aI101
aI997
aI597
aI2485
aI6509
aI11913
aI19573
aI13985
aI56165
aI249
asS'd'
I13582
sa(dp27164
S'a'
I2905
sS's'
I18
sS'm'
(lp27165
I1
aI1
aI7
aI3
aI19
aI45
aI107
aI229
aI241
aI747
aI1219
aI3133
aI3675
aI4441
aI13933
aI64571
aI95445
aI250713
asS'd'
I13583
sa(dp27166
S'a'
I2911
sS's'
I18
sS'm'
(lp27167
I1
aI3
aI1
aI5
aI11
aI31
aI89
aI119
aI503
aI99
aI75
aI349
aI7479
aI15161
aI6365
aI62461
aI39443
aI188455
asS'd'
I13584
sa(dp27168
S'a'
I2912
sS's'
I18
sS'm'
(lp27169
I1
aI1
aI5
aI13
aI25
aI31
aI65
aI237
aI259
aI329
aI89
aI1283
aI6033
aI4401
aI7655
aI38837
aI62367
aI76555
asS'd'
I13585
sa(dp27170
S'a'
I2921
sS's'
I18
sS'm'
(lp27171
I1
aI1
aI1
aI7
aI19
aI61
aI109
aI41
aI361
aI89
aI171
aI2319
aI3625
aI8905
aI24461
aI36135
aI28515
aI101547
asS'd'
I13586
sa(dp27172
S'a'
I2924
sS's'
I18
sS'm'
(lp27173
I1
aI3
aI5
aI3
aI5
aI45
aI123
aI227
aI339
aI79
aI309
aI2619
aI1621
aI1295
aI6395
aI6717
aI119933
aI187231
asS'd'
I13587
sa(dp27174
S'a'
I2945
sS's'
I18
sS'm'
(lp27175
I1
aI1
aI1
aI3
aI3
aI45
aI91
aI225
aI269
aI475
aI1159
aI2599
aI5087
aI4141
aI28375
aI22413
aI56235
aI256559
asS'd'
I13588
sa(dp27176
S'a'
I2952
sS's'
I18
sS'm'
(lp27177
I1
aI1
aI1
aI13
aI7
aI51
aI27
aI65
aI65
aI381
aI169
aI1759
aI4653
aI9885
aI25839
aI19851
aI4965
aI249097
asS'd'
I13589
sa(dp27178
S'a'
I2972
sS's'
I18
sS'm'
(lp27179
I1
aI3
aI7
aI11
aI25
aI11
aI83
aI137
aI419
aI277
aI503
aI2823
aI2759
aI8173
aI9405
aI23731
aI116087
aI9735
asS'd'
I13590
sa(dp27180
S'a'
I2986
sS's'
I18
sS'm'
(lp27181
I1
aI1
aI5
aI5
aI27
aI17
aI123
aI145
aI41
aI85
aI1099
aI1087
aI1465
aI7063
aI8585
aI39427
aI15479
aI243967
asS'd'
I13591
sa(dp27182
S'a'
I3000
sS's'
I18
sS'm'
(lp27183
I1
aI1
aI7
aI3
aI21
aI53
aI105
aI185
aI101
aI763
aI593
aI2649
aI3273
aI5655
aI12233
aI11761
aI27093
aI121347
asS'd'
I13592
sa(dp27184
S'a'
I3008
sS's'
I18
sS'm'
(lp27185
I1
aI1
aI1
aI5
aI11
aI55
aI107
aI167
aI179
aI681
aI741
aI1821
aI4297
aI14677
aI9949
aI9647
aI60465
aI36999
asS'd'
I13593
sa(dp27186
S'a'
I3011
sS's'
I18
sS'm'
(lp27187
I1
aI1
aI1
aI7
aI25
aI43
aI95
aI71
aI161
aI517
aI1475
aI1989
aI6273
aI13295
aI19681
aI51773
aI93523
aI33441
asS'd'
I13594
sa(dp27188
S'a'
I3018
sS's'
I18
sS'm'
(lp27189
I1
aI3
aI1
aI13
aI23
aI59
aI95
aI177
aI73
aI707
aI37
aI421
aI3747
aI14207
aI17159
aI4957
aI20161
aI26185
asS'd'
I13595
sa(dp27190
S'a'
I3047
sS's'
I18
sS'm'
(lp27191
I1
aI1
aI7
aI13
aI13
aI1
aI19
aI153
aI445
aI429
aI1911
aI3515
aI639
aI16015
aI833
aI54347
aI87717
aI82175
asS'd'
I13596
sa(dp27192
S'a'
I3071
sS's'
I18
sS'm'
(lp27193
I1
aI3
aI5
aI9
aI1
aI9
aI115
aI87
aI341
aI651
aI1583
aI807
aI559
aI13579
aI9647
aI37277
aI125555
aI169655
asS'd'
I13597
sa(dp27194
S'a'
I3079
sS's'
I18
sS'm'
(lp27195
I1
aI1
aI3
aI5
aI13
aI23
aI117
aI229
aI205
aI803
aI1381
aI2773
aI7099
aI4031
aI597
aI37135
aI11643
aI92325
asS'd'
I13598
sa(dp27196
S'a'
I3083
sS's'
I18
sS'm'
(lp27197
I1
aI3
aI7
aI9
aI27
aI15
aI33
aI147
aI1
aI799
aI1511
aI2609
aI1419
aI5991
aI15571
aI56995
aI97695
aI223969
asS'd'
I13599
sa(dp27198
S'a'
I3086
sS's'
I18
sS'm'
(lp27199
I1
aI3
aI1
aI3
aI17
aI9
aI17
aI189
aI407
aI355
aI765
aI2545
aI1079
aI15253
aI4785
aI5187
aI80775
aI238775
asS'd'
I13600
sa(dp27200
S'a'
I3148
sS's'
I18
sS'm'
(lp27201
I1
aI1
aI3
aI1
aI31
aI29
aI3
aI159
aI263
aI325
aI125
aI2221
aI6369
aI5717
aI13985
aI33829
aI21375
aI134249
asS'd'
I13601
sa(dp27202
S'a'
I3156
sS's'
I18
sS'm'
(lp27203
I1
aI3
aI7
aI3
aI5
aI29
aI39
aI75
aI183
aI155
aI1017
aI637
aI921
aI9561
aI14893
aI59695
aI38325
aI15503
asS'd'
I13602
sa(dp27204
S'a'
I3194
sS's'
I18
sS'm'
(lp27205
I1
aI1
aI3
aI13
aI9
aI31
aI43
aI71
aI241
aI661
aI325
aI357
aI431
aI903
aI5039
aI24535
aI94241
aI228605
asS'd'
I13603
sa(dp27206
S'a'
I3230
sS's'
I18
sS'm'
(lp27207
I1
aI3
aI1
aI1
aI17
aI37
aI93
aI47
aI25
aI207
aI611
aI415
aI6473
aI15979
aI2025
aI19003
aI8941
aI248779
asS'd'
I13604
sa(dp27208
S'a'
I3233
sS's'
I18
sS'm'
(lp27209
I1
aI1
aI7
aI15
aI19
aI17
aI81
aI201
aI121
aI11
aI1975
aI1289
aI4405
aI7851
aI9707
aI20057
aI33749
aI187161
asS'd'
I13605
sa(dp27210
S'a'
I3254
sS's'
I18
sS'm'
(lp27211
I1
aI1
aI3
aI5
aI29
aI31
aI47
aI99
aI435
aI795
aI947
aI1299
aI4011
aI8315
aI12827
aI48071
aI86567
aI154655
asS'd'
I13606
sa(dp27212
S'a'
I3268
sS's'
I18
sS'm'
(lp27213
I1
aI1
aI5
aI3
aI9
aI59
aI115
aI191
aI177
aI65
aI1835
aI3989
aI1819
aI14325
aI8939
aI25337
aI16099
aI200577
asS'd'
I13607
sa(dp27214
S'a'
I3305
sS's'
I18
sS'm'
(lp27215
I1
aI3
aI7
aI9
aI15
aI47
aI7
aI195
aI413
aI1013
aI1607
aI3317
aI6979
aI13243
aI275
aI34125
aI66069
aI90201
asS'd'
I13608
sa(dp27216
S'a'
I3323
sS's'
I18
sS'm'
(lp27217
I1
aI1
aI3
aI3
aI29
aI3
aI51
aI137
aI341
aI393
aI897
aI351
aI1937
aI6793
aI12551
aI18873
aI110949
aI133925
asS'd'
I13609
sa(dp27218
S'a'
I3326
sS's'
I18
sS'm'
(lp27219
I1
aI3
aI5
aI9
aI29
aI41
aI79
aI169
aI113
aI123
aI1229
aI1885
aI6153
aI1549
aI31729
aI41949
aI74083
aI41387
asS'd'
I13610
sa(dp27220
S'a'
I3343
sS's'
I18
sS'm'
(lp27221
I1
aI3
aI1
aI15
aI31
aI49
aI7
aI233
aI305
aI435
aI1299
aI3037
aI2387
aI15431
aI817
aI11783
aI24067
aI116527
asS'd'
I13611
sa(dp27222
S'a'
I3345
sS's'
I18
sS'm'
(lp27223
I1
aI3
aI5
aI13
aI7
aI17
aI49
aI33
aI133
aI45
aI689
aI2381
aI2649
aI2433
aI27535
aI21755
aI88611
aI200585
asS'd'
I13612
sa(dp27224
S'a'
I3382
sS's'
I18
sS'm'
(lp27225
I1
aI1
aI5
aI11
aI1
aI61
aI87
aI97
aI91
aI433
aI313
aI2541
aI5289
aI5769
aI17963
aI5719
aI12165
aI146849
asS'd'
I13613
sa(dp27226
S'a'
I3413
sS's'
I18
sS'm'
(lp27227
I1
aI3
aI7
aI13
aI17
aI21
aI37
aI191
aI489
aI847
aI841
aI3567
aI7339
aI15233
aI23973
aI1209
aI99741
aI243303
asS'd'
I13614
sa(dp27228
S'a'
I3420
sS's'
I18
sS'm'
(lp27229
I1
aI3
aI1
aI1
aI5
aI21
aI11
aI39
aI69
aI751
aI1679
aI143
aI6187
aI2963
aI695
aI45763
aI126749
aI243841
asS'd'
I13615
sa(dp27230
S'a'
I3434
sS's'
I18
sS'm'
(lp27231
I1
aI3
aI3
aI9
aI21
aI55
aI43
aI73
aI133
aI417
aI495
aI2899
aI5681
aI13049
aI30241
aI44519
aI19095
aI30673
asS'd'
I13616
sa(dp27232
S'a'
I3453
sS's'
I18
sS'm'
(lp27233
I1
aI1
aI5
aI9
aI17
aI51
aI121
aI205
aI273
aI597
aI1325
aI3755
aI5113
aI12287
aI21323
aI17947
aI23807
aI20025
asS'd'
I13617
sa(dp27234
S'a'
I3472
sS's'
I18
sS'm'
(lp27235
I1
aI1
aI7
aI7
aI21
aI11
aI25
aI33
aI207
aI13
aI1639
aI1971
aI7401
aI11771
aI7879
aI59027
aI111981
aI65451
asS'd'
I13618
sa(dp27236
S'a'
I3488
sS's'
I18
sS'm'
(lp27237
I1
aI3
aI5
aI15
aI3
aI15
aI121
aI23
aI199
aI839
aI937
aI3659
aI5379
aI2139
aI31631
aI17215
aI65349
aI157413
asS'd'
I13619
sa(dp27238
S'a'
I3503
sS's'
I18
sS'm'
(lp27239
I1
aI1
aI1
aI7
aI3
aI7
aI81
aI49
aI17
aI693
aI1819
aI2737
aI7329
aI49
aI1655
aI42317
aI31385
aI11435
asS'd'
I13620
sa(dp27240
S'a'
I3506
sS's'
I18
sS'm'
(lp27241
I1
aI3
aI5
aI15
aI25
aI51
aI121
aI133
aI457
aI159
aI869
aI855
aI3529
aI2691
aI147
aI58621
aI78379
aI148519
asS'd'
I13621
sa(dp27242
S'a'
I3518
sS's'
I18
sS'm'
(lp27243
I1
aI1
aI3
aI7
aI1
aI53
aI109
aI81
aI37
aI553
aI1921
aI3081
aI2665
aI12665
aI13887
aI1035
aI16987
aI48883
asS'd'
I13622
sa(dp27244
S'a'
I3532
sS's'
I18
sS'm'
(lp27245
I1
aI1
aI5
aI1
aI19
aI1
aI121
aI97
aI143
aI871
aI1401
aI2879
aI5657
aI5479
aI14011
aI65131
aI56011
aI241055
asS'd'
I13623
sa(dp27246
S'a'
I3543
sS's'
I18
sS'm'
(lp27247
I1
aI1
aI7
aI5
aI9
aI21
aI9
aI43
aI331
aI183
aI1313
aI2495
aI6905
aI2763
aI29567
aI7579
aI95169
aI130937
asS'd'
I13624
sa(dp27248
S'a'
I3547
sS's'
I18
sS'm'
(lp27249
I1
aI3
aI7
aI3
aI3
aI37
aI65
aI195
aI339
aI527
aI1383
aI3063
aI7749
aI11109
aI8097
aI27257
aI107615
aI134241
asS'd'
I13625
sa(dp27250
S'a'
I3573
sS's'
I18
sS'm'
(lp27251
I1
aI1
aI1
aI5
aI25
aI25
aI63
aI179
aI135
aI65
aI169
aI2709
aI5435
aI12119
aI21549
aI59847
aI129639
aI220163
asS'd'
I13626
sa(dp27252
S'a'
I3574
sS's'
I18
sS'm'
(lp27253
I1
aI3
aI5
aI1
aI3
aI17
aI87
aI181
aI9
aI923
aI731
aI3397
aI7079
aI3281
aI10455
aI35471
aI20439
aI206209
asS'd'
I13627
sa(dp27254
S'a'
I3587
sS's'
I18
sS'm'
(lp27255
I1
aI1
aI1
aI5
aI31
aI25
aI15
aI89
aI381
aI675
aI1217
aI3175
aI707
aI585
aI1695
aI57771
aI92433
aI203523
asS'd'
I13628
sa(dp27256
S'a'
I3632
sS's'
I18
sS'm'
(lp27257
I1
aI3
aI5
aI15
aI5
aI7
aI9
aI87
aI461
aI1017
aI869
aI1541
aI7833
aI3117
aI24917
aI13917
aI104797
aI149045
asS'd'
I13629
sa(dp27258
S'a'
I3664
sS's'
I18
sS'm'
(lp27259
I1
aI1
aI7
aI15
aI13
aI49
aI9
aI89
aI165
aI827
aI657
aI1977
aI7471
aI15437
aI25785
aI1455
aI52803
aI198793
asS'd'
I13630
sa(dp27260
S'a'
I3713
sS's'
I18
sS'm'
(lp27261
I1
aI3
aI1
aI15
aI3
aI39
aI27
aI205
aI325
aI345
aI965
aI1439
aI4403
aI10717
aI9591
aI46845
aI123983
aI76181
asS'd'
I13631
sa(dp27262
S'a'
I3726
sS's'
I18
sS'm'
(lp27263
I1
aI3
aI1
aI1
aI25
aI23
aI97
aI135
aI367
aI179
aI1563
aI75
aI455
aI3517
aI21539
aI59565
aI43449
aI139495
asS'd'
I13632
sa(dp27264
S'a'
I3768
sS's'
I18
sS'm'
(lp27265
I1
aI1
aI5
aI15
aI13
aI27
aI55
aI21
aI1
aI505
aI1349
aI409
aI2491
aI5299
aI15771
aI59389
aI110377
aI209275
asS'd'
I13633
sa(dp27266
S'a'
I3771
sS's'
I18
sS'm'
(lp27267
I1
aI1
aI7
aI9
aI31
aI15
aI63
aI91
aI3
aI559
aI419
aI1237
aI1157
aI5811
aI24335
aI19215
aI12581
aI148813
asS'd'
I13634
sa(dp27268
S'a'
I3810
sS's'
I18
sS'm'
(lp27269
I1
aI1
aI7
aI13
aI23
aI3
aI81
aI127
aI33
aI931
aI867
aI2905
aI1011
aI16207
aI1543
aI54309
aI10611
aI152733
asS'd'
I13635
sa(dp27270
S'a'
I3848
sS's'
I18
sS'm'
(lp27271
I1
aI3
aI5
aI7
aI21
aI19
aI45
aI101
aI439
aI537
aI267
aI945
aI8007
aI9383
aI13211
aI21867
aI5731
aI150203
asS'd'
I13636
sa(dp27272
S'a'
I3868
sS's'
I18
sS'm'
(lp27273
I1
aI1
aI3
aI9
aI29
aI3
aI31
aI219
aI217
aI775
aI1011
aI445
aI2663
aI1691
aI9837
aI5727
aI116283
aI128627
asS'd'
I13637
sa(dp27274
S'a'
I3896
sS's'
I18
sS'm'
(lp27275
I1
aI3
aI3
aI3
aI21
aI1
aI97
aI239
aI457
aI925
aI1923
aI1693
aI1187
aI13437
aI8529
aI22081
aI633
aI76109
asS'd'
I13638
sa(dp27276
S'a'
I3910
sS's'
I18
sS'm'
(lp27277
I1
aI3
aI7
aI7
aI19
aI5
aI9
aI15
aI337
aI855
aI1563
aI3159
aI2799
aI4103
aI2013
aI47789
aI77027
aI22425
asS'd'
I13639
sa(dp27278
S'a'
I3921
sS's'
I18
sS'm'
(lp27279
I1
aI1
aI3
aI15
aI15
aI41
aI27
aI77
aI489
aI377
aI1953
aI305
aI5081
aI1895
aI5117
aI51455
aI71859
aI190289
asS'd'
I13640
sa(dp27280
S'a'
I3928
sS's'
I18
sS'm'
(lp27281
I1
aI1
aI5
aI7
aI7
aI7
aI13
aI25
aI115
aI657
aI223
aI3185
aI5327
aI2559
aI5147
aI22237
aI91933
aI195429
asS'd'
I13641
sa(dp27282
S'a'
I3940
sS's'
I18
sS'm'
(lp27283
I1
aI1
aI3
aI5
aI5
aI19
aI3
aI197
aI371
aI237
aI555
aI2873
aI3401
aI3329
aI29165
aI4593
aI111677
aI244025
asS'd'
I13642
sa(dp27284
S'a'
I3947
sS's'
I18
sS'm'
(lp27285
I1
aI3
aI5
aI15
aI15
aI55
aI29
aI75
aI329
aI623
aI279
aI2831
aI4489
aI7803
aI24119
aI12959
aI59783
aI135213
asS'd'
I13643
sa(dp27286
S'a'
I3949
sS's'
I18
sS'm'
(lp27287
I1
aI3
aI5
aI13
aI31
aI21
aI93
aI77
aI401
aI353
aI893
aI917
aI4813
aI8027
aI7847
aI55315
aI60213
aI102763
asS'd'
I13644
sa(dp27288
S'a'
I4001
sS's'
I18
sS'm'
(lp27289
I1
aI1
aI5
aI13
aI29
aI49
aI91
aI35
aI79
aI625
aI1539
aI509
aI823
aI2239
aI30867
aI21729
aI33195
aI38189
asS'd'
I13645
sa(dp27290
S'a'
I4004
sS's'
I18
sS'm'
(lp27291
I1
aI3
aI3
aI3
aI19
aI11
aI39
aI145
aI5
aI329
aI1653
aI3205
aI4431
aI9291
aI30369
aI63173
aI72317
aI236103
asS'd'
I13646
sa(dp27292
S'a'
I4022
sS's'
I18
sS'm'
(lp27293
I1
aI3
aI3
aI15
aI27
aI9
aI111
aI191
aI249
aI845
aI1845
aI2097
aI6529
aI9559
aI25757
aI29085
aI2615
aI175759
asS'd'
I13647
sa(dp27294
S'a'
I4026
sS's'
I18
sS'm'
(lp27295
I1
aI3
aI7
aI1
aI17
aI59
aI119
aI125
aI213
aI995
aI601
aI2517
aI1225
aI2301
aI13031
aI40881
aI31623
aI165799
asS'd'
I13648
sa(dp27296
S'a'
I4036
sS's'
I18
sS'm'
(lp27297
I1
aI3
aI3
aI13
aI25
aI61
aI97
aI157
aI347
aI931
aI1731
aI3697
aI5815
aI7309
aI30605
aI3853
aI72395
aI103609
asS'd'
I13649
sa(dp27298
S'a'
I4073
sS's'
I18
sS'm'
(lp27299
I1
aI1
aI7
aI5
aI23
aI13
aI51
aI117
aI495
aI683
aI777
aI1629
aI5683
aI801
aI4907
aI24935
aI9457
aI214131
asS'd'
I13650
sa(dp27300
S'a'
I4093
sS's'
I18
sS'm'
(lp27301
I1
aI1
aI5
aI9
aI1
aI29
aI107
aI253
aI195
aI921
aI345
aI1451
aI2253
aI12723
aI571
aI12009
aI34149
aI140659
asS'd'
I13651
sa(dp27302
S'a'
I4099
sS's'
I18
sS'm'
(lp27303
I1
aI1
aI5
aI9
aI31
aI17
aI93
aI5
aI455
aI205
aI1439
aI1199
aI7371
aI12973
aI16455
aI675
aI60561
aI99575
asS'd'
I13652
sa(dp27304
S'a'
I4120
sS's'
I18
sS'm'
(lp27305
I1
aI3
aI3
aI3
aI31
aI37
aI115
aI49
aI31
aI285
aI2029
aI1369
aI3443
aI2411
aI10367
aI44859
aI26737
aI195703
asS'd'
I13653
sa(dp27306
S'a'
I4136
sS's'
I18
sS'm'
(lp27307
I1
aI1
aI3
aI1
aI15
aI39
aI113
aI37
aI257
aI3
aI817
aI2901
aI4029
aI12595
aI30475
aI34883
aI109133
aI92159
asS'd'
I13654
sa(dp27308
S'a'
I4156
sS's'
I18
sS'm'
(lp27309
I1
aI1
aI7
aI5
aI9
aI1
aI9
aI101
aI317
aI167
aI1975
aI411
aI6875
aI6951
aI4401
aI59483
aI129813
aI78289
asS'd'
I13655
sa(dp27310
S'a'
I4176
sS's'
I18
sS'm'
(lp27311
I1
aI1
aI7
aI9
aI9
aI5
aI73
aI7
aI57
aI907
aI1887
aI2923
aI961
aI8521
aI873
aI33791
aI114485
aI43081
asS'd'
I13656
sa(dp27312
S'a'
I4182
sS's'
I18
sS'm'
(lp27313
I1
aI1
aI5
aI7
aI13
aI45
aI91
aI179
aI499
aI197
aI1337
aI1321
aI5307
aI15503
aI20449
aI60813
aI97393
aI255741
asS'd'
I13657
sa(dp27314
S'a'
I4191
sS's'
I18
sS'm'
(lp27315
I1
aI1
aI1
aI5
aI25
aI13
aI69
aI221
aI207
aI823
aI845
aI3845
aI6743
aI5123
aI27447
aI2079
aI100635
aI124157
asS'd'
I13658
sa(dp27316
S'a'
I4198
sS's'
I18
sS'm'
(lp27317
I1
aI3
aI5
aI11
aI13
aI39
aI121
aI209
aI137
aI63
aI1479
aI323
aI5347
aI9797
aI17785
aI55541
aI108713
aI243347
asS'd'
I13659
sa(dp27318
S'a'
I4252
sS's'
I18
sS'm'
(lp27319
I1
aI1
aI3
aI9
aI29
aI45
aI43
aI81
aI115
aI979
aI727
aI423
aI1133
aI8757
aI27833
aI39907
aI104663
aI33067
asS'd'
I13660
sa(dp27320
S'a'
I4259
sS's'
I18
sS'm'
(lp27321
I1
aI3
aI5
aI1
aI13
aI61
aI49
aI17
aI409
aI567
aI1035
aI2299
aI3711
aI15485
aI7767
aI27809
aI1275
aI96455
asS'd'
I13661
sa(dp27322
S'a'
I4261
sS's'
I18
sS'm'
(lp27323
I1
aI1
aI5
aI9
aI5
aI33
aI13
aI9
aI505
aI459
aI747
aI4079
aI4271
aI6925
aI13933
aI31349
aI5793
aI68381
asS'd'
I13662
sa(dp27324
S'a'
I4294
sS's'
I18
sS'm'
(lp27325
I1
aI3
aI3
aI11
aI15
aI47
aI15
aI187
aI349
aI847
aI817
aI3551
aI6059
aI6451
aI32615
aI1635
aI108889
aI48003
asS'd'
I13663
sa(dp27326
S'a'
I4341
sS's'
I18
sS'm'
(lp27327
I1
aI3
aI5
aI7
aI3
aI31
aI11
aI255
aI367
aI295
aI1079
aI2981
aI5583
aI10771
aI25359
aI16083
aI24163
aI111201
asS'd'
I13664
sa(dp27328
S'a'
I4348
sS's'
I18
sS'm'
(lp27329
I1
aI3
aI5
aI5
aI7
aI5
aI127
aI19
aI343
aI849
aI287
aI1471
aI7299
aI1209
aI31349
aI33473
aI4989
aI229181
asS'd'
I13665
sa(dp27330
S'a'
I4356
sS's'
I18
sS'm'
(lp27331
I1
aI1
aI3
aI9
aI25
aI61
aI7
aI65
aI77
aI745
aI1871
aI2427
aI3669
aI8965
aI11177
aI5531
aI115801
aI34327
asS'd'
I13666
sa(dp27332
S'a'
I4384
sS's'
I18
sS'm'
(lp27333
I1
aI3
aI3
aI15
aI1
aI57
aI125
aI167
aI173
aI875
aI347
aI2317
aI6687
aI4339
aI10573
aI7841
aI16241
aI192225
asS'd'
I13667
sa(dp27334
S'a'
I4389
sS's'
I18
sS'm'
(lp27335
I1
aI3
aI1
aI3
aI15
aI37
aI45
aI189
aI75
aI1017
aI1919
aI3401
aI329
aI2539
aI32697
aI60801
aI52017
aI192611
asS'd'
I13668
sa(dp27336
S'a'
I4401
sS's'
I18
sS'm'
(lp27337
I1
aI3
aI1
aI5
aI1
aI23
aI43
aI55
aI1
aI443
aI1769
aI1633
aI5225
aI6855
aI5419
aI65139
aI22237
aI17415
asS'd'
I13669
sa(dp27338
S'a'
I4428
sS's'
I18
sS'm'
(lp27339
I1
aI3
aI5
aI15
aI25
aI7
aI107
aI209
aI325
aI367
aI373
aI1855
aI1313
aI12899
aI30137
aI19007
aI9911
aI11791
asS'd'
I13670
sa(dp27340
S'a'
I4431
sS's'
I18
sS'm'
(lp27341
I1
aI3
aI7
aI13
aI3
aI57
aI123
aI93
aI279
aI469
aI1817
aI3409
aI565
aI3997
aI14119
aI58341
aI59691
aI163323
asS'd'
I13671
sa(dp27342
S'a'
I4445
sS's'
I18
sS'm'
(lp27343
I1
aI1
aI3
aI9
aI3
aI3
aI69
aI109
aI47
aI487
aI1895
aI2003
aI7309
aI9803
aI9527
aI52211
aI31213
aI41521
asS'd'
I13672
sa(dp27344
S'a'
I4470
sS's'
I18
sS'm'
(lp27345
I1
aI1
aI7
aI7
aI9
aI15
aI101
aI227
aI75
aI501
aI25
aI1481
aI4847
aI13279
aI28673
aI11069
aI61987
aI5365
asS'd'
I13673
sa(dp27346
S'a'
I4473
sS's'
I18
sS'm'
(lp27347
I1
aI3
aI1
aI1
aI25
aI5
aI47
aI125
aI97
aI969
aI1077
aI1185
aI6033
aI13927
aI18149
aI34255
aI14353
aI66323
asS'd'
I13674
sa(dp27348
S'a'
I4474
sS's'
I18
sS'm'
(lp27349
I1
aI1
aI3
aI1
aI25
aI41
aI19
aI69
aI385
aI585
aI1049
aI3497
aI3615
aI13211
aI18855
aI61303
aI115739
aI42639
asS'd'
I13675
sa(dp27350
S'a'
I4490
sS's'
I18
sS'm'
(lp27351
I1
aI3
aI7
aI7
aI13
aI15
aI13
aI133
aI497
aI265
aI1809
aI4073
aI5673
aI7543
aI30823
aI13505
aI76167
aI98683
asS'd'
I13676
sa(dp27352
S'a'
I4509
sS's'
I18
sS'm'
(lp27353
I1
aI1
aI5
aI5
aI3
aI59
aI47
aI191
aI419
aI505
aI2035
aI329
aI553
aI1561
aI27885
aI39767
aI102611
aI12689
asS'd'
I13677
sa(dp27354
S'a'
I4510
sS's'
I18
sS'm'
(lp27355
I1
aI3
aI7
aI3
aI27
aI49
aI27
aI133
aI305
aI537
aI385
aI335
aI2417
aI14891
aI31299
aI26201
aI124655
aI150545
asS'd'
I13678
sa(dp27356
S'a'
I4533
sS's'
I18
sS'm'
(lp27357
I1
aI1
aI1
aI9
aI7
aI1
aI27
aI105
aI347
aI481
aI2043
aI1645
aI4367
aI10335
aI16457
aI48713
aI64699
aI63595
asS'd'
I13679
sa(dp27358
S'a'
I4548
sS's'
I18
sS'm'
(lp27359
I1
aI1
aI3
aI3
aI7
aI57
aI125
aI209
aI299
aI525
aI591
aI1265
aI7557
aI15113
aI19319
aI56269
aI43919
aI215435
asS'd'
I13680
sa(dp27360
S'a'
I4558
sS's'
I18
sS'm'
(lp27361
I1
aI1
aI1
aI11
aI29
aI59
aI119
aI245
aI63
aI919
aI1913
aI3969
aI545
aI1033
aI20975
aI61327
aI36783
aI124303
asS'd'
I13681
sa(dp27362
S'a'
I4594
sS's'
I18
sS'm'
(lp27363
I1
aI1
aI7
aI7
aI11
aI63
aI45
aI135
aI405
aI931
aI753
aI2559
aI5475
aI2107
aI6437
aI6055
aI43497
aI133571
asS'd'
I13682
sa(dp27364
S'a'
I4596
sS's'
I18
sS'm'
(lp27365
I1
aI3
aI1
aI13
aI31
aI39
aI39
aI141
aI231
aI83
aI69
aI473
aI1095
aI13617
aI10909
aI49861
aI98029
aI235003
asS'd'
I13683
sa(dp27366
S'a'
I4603
sS's'
I18
sS'm'
(lp27367
I1
aI3
aI7
aI13
aI13
aI41
aI73
aI107
aI505
aI359
aI957
aI1599
aI7617
aI1843
aI25531
aI63755
aI96295
aI167955
asS'd'
I13684
sa(dp27368
S'a'
I4610
sS's'
I18
sS'm'
(lp27369
I1
aI3
aI3
aI11
aI13
aI41
aI61
aI65
aI165
aI507
aI1007
aI1695
aI91
aI8781
aI15017
aI12063
aI95331
aI179853
asS'd'
I13685
sa(dp27370
S'a'
I4619
sS's'
I18
sS'm'
(lp27371
I1
aI3
aI7
aI7
aI29
aI19
aI7
aI95
aI303
aI641
aI581
aI3539
aI4495
aI13549
aI20195
aI20845
aI16961
aI95053
asS'd'
I13686
sa(dp27372
S'a'
I4630
sS's'
I18
sS'm'
(lp27373
I1
aI1
aI7
aI5
aI15
aI27
aI13
aI155
aI345
aI341
aI1583
aI2207
aI2497
aI6509
aI24343
aI3109
aI71431
aI184871
asS'd'
I13687
sa(dp27374
S'a'
I4652
sS's'
I18
sS'm'
(lp27375
I1
aI1
aI3
aI15
aI31
aI35
aI37
aI249
aI71
aI1005
aI681
aI3457
aI3387
aI13797
aI8781
aI11789
aI16825
aI11133
asS'd'
I13688
sa(dp27376
S'a'
I4658
sS's'
I18
sS'm'
(lp27377
I1
aI3
aI7
aI11
aI5
aI29
aI121
aI139
aI77
aI859
aI163
aI2749
aI6401
aI16303
aI22659
aI11817
aI61667
aI119993
asS'd'
I13689
sa(dp27378
S'a'
I4682
sS's'
I18
sS'm'
(lp27379
I1
aI1
aI7
aI11
aI15
aI45
aI71
aI87
aI293
aI981
aI1581
aI2789
aI4117
aI12791
aI13611
aI489
aI74823
aI71263
asS'd'
I13690
sa(dp27380
S'a'
I4708
sS's'
I18
sS'm'
(lp27381
I1
aI1
aI3
aI9
aI15
aI21
aI59
aI167
aI469
aI723
aI1609
aI2111
aI6359
aI10781
aI1043
aI51039
aI24429
aI14605
asS'd'
I13691
sa(dp27382
S'a'
I4736
sS's'
I18
sS'm'
(lp27383
I1
aI3
aI3
aI9
aI13
aI25
aI1
aI43
aI61
aI869
aI1919
aI601
aI8003
aI15841
aI10141
aI33187
aI124991
aI94205
asS'd'
I13692
sa(dp27384
S'a'
I4753
sS's'
I18
sS'm'
(lp27385
I1
aI1
aI7
aI5
aI23
aI13
aI67
aI43
aI167
aI667
aI1743
aI2523
aI2245
aI9287
aI8115
aI64995
aI121371
aI188321
asS'd'
I13693
sa(dp27386
S'a'
I4760
sS's'
I18
sS'm'
(lp27387
I1
aI1
aI1
aI9
aI13
aI19
aI45
aI249
aI21
aI751
aI239
aI4035
aI4549
aI8905
aI9377
aI47535
aI78135
aI210429
asS'd'
I13694
sa(dp27388
S'a'
I4781
sS's'
I18
sS'm'
(lp27389
I1
aI1
aI3
aI7
aI5
aI43
aI13
aI227
aI75
aI785
aI631
aI205
aI3475
aI9735
aI17867
aI61407
aI75897
aI51151
asS'd'
I13695
sa(dp27390
S'a'
I4784
sS's'
I18
sS'm'
(lp27391
I1
aI3
aI3
aI9
aI31
aI21
aI11
aI53
aI247
aI717
aI1505
aI3903
aI3249
aI3185
aI29007
aI48795
aI43413
aI158653
asS'd'
I13696
sa(dp27392
S'a'
I4799
sS's'
I18
sS'm'
(lp27393
I1
aI1
aI5
aI11
aI19
aI9
aI37
aI159
aI183
aI521
aI743
aI2877
aI2291
aI10317
aI1211
aI17951
aI16335
aI66439
asS'd'
I13697
sa(dp27394
S'a'
I4807
sS's'
I18
sS'm'
(lp27395
I1
aI1
aI3
aI7
aI3
aI41
aI15
aI113
aI125
aI391
aI201
aI3841
aI255
aI15381
aI16801
aI47219
aI119691
aI51811
asS'd'
I13698
sa(dp27396
S'a'
I4808
sS's'
I18
sS'm'
(lp27397
I1
aI1
aI3
aI1
aI1
aI29
aI79
aI181
aI481
aI969
aI297
aI625
aI7449
aI5813
aI5915
aI20011
aI44853
aI231933
asS'd'
I13699
sa(dp27398
S'a'
I4842
sS's'
I18
sS'm'
(lp27399
I1
aI1
aI1
aI5
aI5
aI49
aI63
aI171
aI93
aI107
aI1083
aI1277
aI121
aI4421
aI18951
aI61155
aI66643
aI120049
asS'd'
I13700
sa(dp27400
S'a'
I4882
sS's'
I18
sS'm'
(lp27401
I1
aI3
aI7
aI5
aI25
aI59
aI111
aI197
aI459
aI217
aI1819
aI1603
aI5581
aI11361
aI17721
aI57475
aI11171
aI186577
asS'd'
I13701
sa(dp27402
S'a'
I4897
sS's'
I18
sS'm'
(lp27403
I1
aI3
aI3
aI1
aI25
aI19
aI29
aI157
aI25
aI595
aI501
aI2145
aI7513
aI10323
aI11107
aI13269
aI21763
aI9427
asS'd'
I13702
sa(dp27404
S'a'
I4900
sS's'
I18
sS'm'
(lp27405
I1
aI3
aI7
aI3
aI9
aI49
aI119
aI117
aI445
aI91
aI227
aI1203
aI6245
aI9575
aI30653
aI65429
aI64987
aI81249
asS'd'
I13703
sa(dp27406
S'a'
I4922
sS's'
I18
sS'm'
(lp27407
I1
aI1
aI5
aI5
aI5
aI5
aI77
aI77
aI425
aI789
aI467
aI3931
aI4815
aI11195
aI21939
aI59513
aI78547
aI238035
asS'd'
I13704
sa(dp27408
S'a'
I4936
sS's'
I18
sS'm'
(lp27409
I1
aI3
aI1
aI11
aI11
aI29
aI115
aI37
aI423
aI997
aI1231
aI3987
aI5057
aI14533
aI18005
aI51513
aI71851
aI258137
asS'd'
I13705
sa(dp27410
S'a'
I4960
sS's'
I18
sS'm'
(lp27411
I1
aI3
aI1
aI1
aI27
aI31
aI7
aI223
aI23
aI59
aI1465
aI2045
aI6677
aI15707
aI25101
aI22269
aI46995
aI89141
asS'd'
I13706
sa(dp27412
S'a'
I4970
sS's'
I18
sS'm'
(lp27413
I1
aI1
aI1
aI5
aI21
aI7
aI115
aI133
aI407
aI373
aI1495
aI2551
aI6947
aI3309
aI14903
aI5683
aI67345
aI139381
asS'd'
I13707
sa(dp27414
S'a'
I4978
sS's'
I18
sS'm'
(lp27415
I1
aI1
aI5
aI3
aI13
aI31
aI5
aI221
aI187
aI9
aI165
aI2295
aI1239
aI5665
aI14543
aI3963
aI4931
aI8269
asS'd'
I13708
sa(dp27416
S'a'
I4994
sS's'
I18
sS'm'
(lp27417
I1
aI3
aI5
aI13
aI15
aI5
aI37
aI171
aI419
aI665
aI765
aI1619
aI1561
aI1661
aI5873
aI25595
aI34827
aI215599
asS'd'
I13709
sa(dp27418
S'a'
I5003
sS's'
I18
sS'm'
(lp27419
I1
aI3
aI1
aI15
aI13
aI33
aI45
aI107
aI275
aI771
aI1105
aI2895
aI187
aI5173
aI21179
aI35047
aI50825
aI176775
asS'd'
I13710
sa(dp27420
S'a'
I5044
sS's'
I18
sS'm'
(lp27421
I1
aI3
aI7
aI5
aI17
aI5
aI59
aI195
aI441
aI625
aI1205
aI207
aI4703
aI10627
aI17123
aI61785
aI100779
aI258597
asS'd'
I13711
sa(dp27422
S'a'
I5061
sS's'
I18
sS'm'
(lp27423
I1
aI1
aI1
aI13
aI17
aI3
aI13
aI201
aI241
aI657
aI153
aI289
aI5213
aI2129
aI13447
aI28807
aI25405
aI33803
asS'd'
I13712
sa(dp27424
S'a'
I5107
sS's'
I18
sS'm'
(lp27425
I1
aI3
aI1
aI9
aI19
aI9
aI51
aI133
aI159
aI743
aI1023
aI291
aI7137
aI6949
aI30419
aI13449
aI111505
aI212393
asS'd'
I13713
sa(dp27426
S'a'
I5127
sS's'
I18
sS'm'
(lp27427
I1
aI3
aI5
aI7
aI31
aI29
aI79
aI211
aI425
aI93
aI1173
aI1957
aI6737
aI1725
aI30703
aI43237
aI119747
aI157395
asS'd'
I13714
sa(dp27428
S'a'
I5131
sS's'
I18
sS'm'
(lp27429
I1
aI1
aI7
aI3
aI17
aI21
aI39
aI19
aI485
aI663
aI19
aI761
aI1525
aI11059
aI12833
aI17567
aI61123
aI124801
asS'd'
I13715
sa(dp27430
S'a'
I5136
sS's'
I18
sS'm'
(lp27431
I1
aI1
aI7
aI3
aI7
aI25
aI17
aI199
aI413
aI821
aI1561
aI3855
aI1871
aI14041
aI7525
aI19383
aI51017
aI213357
asS'd'
I13716
sa(dp27432
S'a'
I5145
sS's'
I18
sS'm'
(lp27433
I1
aI3
aI3
aI9
aI25
aI11
aI63
aI83
aI217
aI587
aI47
aI3775
aI767
aI9191
aI5127
aI9133
aI97689
aI122949
asS'd'
I13717
sa(dp27434
S'a'
I5146
sS's'
I18
sS'm'
(lp27435
I1
aI1
aI1
aI9
aI19
aI7
aI89
aI125
aI23
aI813
aI1277
aI2965
aI1939
aI1453
aI6349
aI53127
aI109813
aI63767
asS'd'
I13718
sa(dp27436
S'a'
I5175
sS's'
I18
sS'm'
(lp27437
I1
aI1
aI5
aI15
aI29
aI63
aI117
aI37
aI185
aI69
aI1823
aI2791
aI4125
aI11757
aI14847
aI15567
aI126141
aI185951
asS'd'
I13719
sa(dp27438
S'a'
I5218
sS's'
I18
sS'm'
(lp27439
I1
aI3
aI7
aI1
aI23
aI11
aI15
aI113
aI209
aI785
aI229
aI3207
aI97
aI2489
aI4587
aI14253
aI30421
aI51027
asS'd'
I13720
sa(dp27440
S'a'
I5223
sS's'
I18
sS'm'
(lp27441
I1
aI3
aI1
aI7
aI7
aI33
aI57
aI51
aI219
aI233
aI89
aI3781
aI2055
aI4163
aI10935
aI51913
aI63507
aI18645
asS'd'
I13721
sa(dp27442
S'a'
I5248
sS's'
I18
sS'm'
(lp27443
I1
aI3
aI1
aI7
aI7
aI55
aI107
aI187
aI109
aI867
aI955
aI139
aI4979
aI8627
aI5835
aI28761
aI72061
aI99413
asS'd'
I13722
sa(dp27444
S'a'
I5294
sS's'
I18
sS'm'
(lp27445
I1
aI1
aI1
aI3
aI5
aI59
aI17
aI121
aI511
aI29
aI1009
aI2875
aI2459
aI1817
aI11741
aI13869
aI72543
aI70485
asS'd'
I13723
sa(dp27446
S'a'
I5299
sS's'
I18
sS'm'
(lp27447
I1
aI1
aI5
aI1
aI27
aI19
aI125
aI65
aI379
aI803
aI411
aI2403
aI719
aI10683
aI23351
aI18113
aI66773
aI252223
asS'd'
I13724
sa(dp27448
S'a'
I5301
sS's'
I18
sS'm'
(lp27449
I1
aI1
aI7
aI9
aI31
aI51
aI65
aI233
aI171
aI357
aI1465
aI1609
aI4263
aI15207
aI18825
aI48831
aI69459
aI211321
asS'd'
I13725
sa(dp27450
S'a'
I5311
sS's'
I18
sS'm'
(lp27451
I1
aI1
aI3
aI3
aI5
aI35
aI53
aI245
aI469
aI1011
aI759
aI455
aI4487
aI9835
aI10349
aI61755
aI73279
aI186049
asS'd'
I13726
sa(dp27452
S'a'
I5343
sS's'
I18
sS'm'
(lp27453
I1
aI1
aI7
aI3
aI27
aI19
aI105
aI193
aI403
aI907
aI295
aI1445
aI1867
aI8867
aI7821
aI45309
aI129069
aI83953
asS'd'
I13727
sa(dp27454
S'a'
I5344
sS's'
I18
sS'm'
(lp27455
I1
aI1
aI7
aI7
aI7
aI51
aI85
aI97
aI473
aI837
aI201
aI501
aI2929
aI9457
aI6473
aI3653
aI126991
aI218069
asS'd'
I13728
sa(dp27456
S'a'
I5396
sS's'
I18
sS'm'
(lp27457
I1
aI3
aI3
aI5
aI25
aI49
aI85
aI223
aI127
aI563
aI239
aI1975
aI119
aI6029
aI19349
aI59533
aI44173
aI142229
asS'd'
I13729
sa(dp27458
S'a'
I5406
sS's'
I18
sS'm'
(lp27459
I1
aI1
aI3
aI7
aI23
aI11
aI27
aI49
aI467
aI701
aI2037
aI2367
aI5829
aI12533
aI9641
aI38629
aI90505
aI132013
asS'd'
I13730
sa(dp27460
S'a'
I5410
sS's'
I18
sS'm'
(lp27461
I1
aI1
aI7
aI1
aI13
aI25
aI79
aI107
aI37
aI331
aI355
aI3639
aI4875
aI6635
aI21703
aI18289
aI36257
aI201857
asS'd'
I13731
sa(dp27462
S'a'
I5424
sS's'
I18
sS'm'
(lp27463
I1
aI1
aI5
aI1
aI3
aI39
aI25
aI101
aI199
aI401
aI1495
aI3683
aI5447
aI12313
aI19707
aI20853
aI66821
aI73959
asS'd'
I13732
sa(dp27464
S'a'
I5447
sS's'
I18
sS'm'
(lp27465
I1
aI1
aI1
aI9
aI15
aI9
aI3
aI231
aI479
aI97
aI221
aI973
aI839
aI1757
aI8759
aI45625
aI44691
aI139803
asS'd'
I13733
sa(dp27466
S'a'
I5461
sS's'
I18
sS'm'
(lp27467
I1
aI1
aI7
aI15
aI9
aI51
aI23
aI233
aI311
aI83
aI287
aI4035
aI2087
aI4245
aI25457
aI43105
aI104903
aI132811
asS'd'
I13734
sa(dp27468
S'a'
I5468
sS's'
I18
sS'm'
(lp27469
I1
aI1
aI1
aI9
aI5
aI13
aI33
aI167
aI363
aI67
aI601
aI2143
aI5495
aI1277
aI14615
aI32759
aI34935
aI158625
asS'd'
I13735
sa(dp27470
S'a'
I5475
sS's'
I18
sS'm'
(lp27471
I1
aI1
aI5
aI11
aI5
aI63
aI35
aI49
aI183
aI705
aI377
aI2607
aI2947
aI10119
aI15631
aI60247
aI99309
aI25747
asS'd'
I13736
sa(dp27472
S'a'
I5478
sS's'
I18
sS'm'
(lp27473
I1
aI3
aI7
aI5
aI7
aI3
aI127
aI109
aI165
aI767
aI1873
aI3825
aI441
aI11957
aI2581
aI38309
aI129623
aI77451
asS'd'
I13737
sa(dp27474
S'a'
I5523
sS's'
I18
sS'm'
(lp27475
I1
aI3
aI1
aI7
aI19
aI53
aI101
aI117
aI505
aI363
aI1399
aI1015
aI631
aI8309
aI17507
aI28941
aI42585
aI116283
asS'd'
I13738
sa(dp27476
S'a'
I5530
sS's'
I18
sS'm'
(lp27477
I1
aI1
aI7
aI7
aI9
aI27
aI127
aI195
aI499
aI225
aI153
aI517
aI3909
aI9801
aI3787
aI32829
aI6599
aI190807
asS'd'
I13739
sa(dp27478
S'a'
I5535
sS's'
I18
sS'm'
(lp27479
I1
aI1
aI5
aI1
aI3
aI49
aI125
aI235
aI255
aI329
aI909
aI1685
aI759
aI2287
aI3479
aI23491
aI71157
aI81457
asS'd'
I13740
sa(dp27480
S'a'
I5548
sS's'
I18
sS'm'
(lp27481
I1
aI1
aI3
aI9
aI19
aI21
aI93
aI37
aI259
aI69
aI219
aI1943
aI4747
aI13951
aI14945
aI46099
aI87189
aI222287
asS'd'
I13741
sa(dp27482
S'a'
I5559
sS's'
I18
sS'm'
(lp27483
I1
aI1
aI7
aI5
aI21
aI33
aI61
aI227
aI167
aI569
aI1355
aI2997
aI4917
aI10765
aI7015
aI54335
aI125543
aI112867
asS'd'
I13742
sa(dp27484
S'a'
I5580
sS's'
I18
sS'm'
(lp27485
I1
aI1
aI5
aI3
aI25
aI35
aI97
aI23
aI365
aI159
aI1211
aI1283
aI979
aI8993
aI21323
aI6863
aI46869
aI36169
asS'd'
I13743
sa(dp27486
S'a'
I5604
sS's'
I18
sS'm'
(lp27487
I1
aI1
aI7
aI13
aI15
aI49
aI45
aI209
aI397
aI785
aI47
aI2307
aI4749
aI2735
aI29525
aI54921
aI23321
aI216197
asS'd'
I13744
sa(dp27488
S'a'
I5622
sS's'
I18
sS'm'
(lp27489
I1
aI1
aI1
aI11
aI31
aI23
aI27
aI127
aI197
aI595
aI29
aI773
aI3291
aI6355
aI11891
aI6635
aI99871
aI177531
asS'd'
I13745
sa(dp27490
S'a'
I5641
sS's'
I18
sS'm'
(lp27491
I1
aI1
aI7
aI11
aI9
aI49
aI85
aI59
aI211
aI307
aI1821
aI3947
aI4175
aI11287
aI27889
aI107
aI46463
aI237129
asS'd'
I13746
sa(dp27492
S'a'
I5649
sS's'
I18
sS'm'
(lp27493
I1
aI3
aI3
aI7
aI31
aI31
aI9
aI49
aI365
aI189
aI1211
aI943
aI337
aI13809
aI16941
aI17053
aI70125
aI149865
asS'd'
I13747
sa(dp27494
S'a'
I5656
sS's'
I18
sS'm'
(lp27495
I1
aI3
aI1
aI13
aI9
aI21
aI67
aI1
aI365
aI77
aI1701
aI559
aI3461
aI8961
aI13801
aI16111
aI65239
aI157713
asS'd'
I13748
sa(dp27496
S'a'
I5671
sS's'
I18
sS'm'
(lp27497
I1
aI3
aI7
aI13
aI19
aI49
aI29
aI233
aI361
aI1011
aI1617
aI2989
aI2387
aI14027
aI4021
aI28791
aI33155
aI171449
asS'd'
I13749
sa(dp27498
S'a'
I5689
sS's'
I18
sS'm'
(lp27499
I1
aI1
aI5
aI13
aI31
aI19
aI77
aI69
aI49
aI513
aI1411
aI77
aI4993
aI907
aI23483
aI20129
aI29491
aI138187
asS'd'
I13750
sa(dp27500
S'a'
I5721
sS's'
I18
sS'm'
(lp27501
I1
aI3
aI3
aI11
aI23
aI33
aI19
aI55
aI307
aI455
aI1783
aI3997
aI6411
aI3355
aI8815
aI39883
aI124381
aI49667
asS'd'
I13751
sa(dp27502
S'a'
I5731
sS's'
I18
sS'm'
(lp27503
I1
aI3
aI3
aI15
aI27
aI7
aI25
aI243
aI275
aI27
aI23
aI3039
aI6497
aI15975
aI5877
aI58611
aI6317
aI209119
asS'd'
I13752
sa(dp27504
S'a'
I5734
sS's'
I18
sS'm'
(lp27505
I1
aI3
aI1
aI5
aI7
aI21
aI97
aI247
aI297
aI181
aI773
aI3095
aI2441
aI15683
aI29609
aI50431
aI92813
aI723
asS'd'
I13753
sa(dp27506
S'a'
I5738
sS's'
I18
sS'm'
(lp27507
I1
aI1
aI1
aI3
aI17
aI25
aI69
aI171
aI27
aI83
aI173
aI163
aI7915
aI13547
aI5915
aI20275
aI101613
aI225081
asS'd'
I13754
sa(dp27508
S'a'
I5740
sS's'
I18
sS'm'
(lp27509
I1
aI1
aI7
aI15
aI19
aI13
aI53
aI95
aI171
aI889
aI131
aI1979
aI2537
aI7749
aI77
aI49293
aI68875
aI159125
asS'd'
I13755
sa(dp27510
S'a'
I5748
sS's'
I18
sS'm'
(lp27511
I1
aI1
aI1
aI5
aI11
aI7
aI7
aI29
aI397
aI435
aI1495
aI2263
aI3677
aI11121
aI1269
aI5415
aI44427
aI249943
asS'd'
I13756
sa(dp27512
S'a'
I5791
sS's'
I18
sS'm'
(lp27513
I1
aI3
aI1
aI5
aI23
aI21
aI13
aI185
aI231
aI757
aI1647
aI663
aI1273
aI11641
aI25563
aI46793
aI54231
aI113143
asS'd'
I13757
sa(dp27514
S'a'
I5801
sS's'
I18
sS'm'
(lp27515
I1
aI3
aI3
aI7
aI11
aI21
aI83
aI109
aI409
aI923
aI1541
aI2805
aI1781
aI6903
aI9093
aI37327
aI60923
aI167271
asS'd'
I13758
sa(dp27516
S'a'
I5807
sS's'
I18
sS'm'
(lp27517
I1
aI3
aI7
aI5
aI29
aI37
aI87
aI85
aI93
aI749
aI875
aI2869
aI1023
aI13303
aI26865
aI30971
aI40863
aI237075
asS'd'
I13759
sa(dp27518
S'a'
I5822
sS's'
I18
sS'm'
(lp27519
I1
aI1
aI7
aI13
aI21
aI25
aI39
aI213
aI303
aI265
aI1251
aI2963
aI3819
aI8507
aI23239
aI52625
aI123375
aI58553
asS'd'
I13760
sa(dp27520
S'a'
I5829
sS's'
I18
sS'm'
(lp27521
I1
aI1
aI1
aI1
aI3
aI1
aI7
aI67
aI339
aI583
aI3
aI2489
aI5481
aI12241
aI21695
aI31351
aI39389
aI131925
asS'd'
I13761
sa(dp27522
S'a'
I5834
sS's'
I18
sS'm'
(lp27523
I1
aI1
aI5
aI11
aI11
aI3
aI95
aI23
aI133
aI415
aI77
aI1891
aI4083
aI7097
aI26455
aI28689
aI83047
aI49759
asS'd'
I13762
sa(dp27524
S'a'
I5847
sS's'
I18
sS'm'
(lp27525
I1
aI3
aI7
aI15
aI19
aI31
aI65
aI189
aI489
aI461
aI1255
aI1897
aI3361
aI12223
aI9721
aI45937
aI102695
aI113431
asS'd'
I13763
sa(dp27526
S'a'
I5854
sS's'
I18
sS'm'
(lp27527
I1
aI3
aI1
aI5
aI9
aI57
aI3
aI225
aI241
aI769
aI1003
aI2255
aI7655
aI4837
aI25267
aI35845
aI49545
aI24931
asS'd'
I13764
sa(dp27528
S'a'
I5878
sS's'
I18
sS'm'
(lp27529
I1
aI1
aI1
aI1
aI1
aI43
aI3
aI217
aI397
aI419
aI1189
aI2037
aI5941
aI4341
aI19851
aI13773
aI15225
aI167581
asS'd'
I13765
sa(dp27530
S'a'
I5882
sS's'
I18
sS'm'
(lp27531
I1
aI1
aI1
aI9
aI3
aI27
aI65
aI49
aI115
aI787
aI1637
aI1867
aI7265
aI8541
aI1587
aI58987
aI82161
aI19997
asS'd'
I13766
sa(dp27532
S'a'
I5919
sS's'
I18
sS'm'
(lp27533
I1
aI3
aI7
aI11
aI17
aI17
aI93
aI103
aI309
aI159
aI781
aI3179
aI5759
aI7661
aI5693
aI48531
aI127375
aI141449
asS'd'
I13767
sa(dp27534
S'a'
I5947
sS's'
I18
sS'm'
(lp27535
I1
aI3
aI7
aI5
aI13
aI39
aI79
aI241
aI7
aI137
aI219
aI523
aI541
aI4787
aI23327
aI41665
aI111017
aI118901
asS'd'
I13768
sa(dp27536
S'a'
I5981
sS's'
I18
sS'm'
(lp27537
I1
aI1
aI3
aI15
aI31
aI23
aI107
aI221
aI295
aI935
aI1165
aI2463
aI1635
aI10205
aI18057
aI28217
aI51755
aI85579
asS'd'
I13769
sa(dp27538
S'a'
I5982
sS's'
I18
sS'm'
(lp27539
I1
aI3
aI1
aI11
aI23
aI47
aI7
aI59
aI75
aI603
aI1237
aI2601
aI6873
aI12735
aI32181
aI46849
aI106363
aI171753
asS'd'
I13770
sa(dp27540
S'a'
I6025
sS's'
I18
sS'm'
(lp27541
I1
aI1
aI7
aI13
aI15
aI31
aI3
aI113
aI355
aI955
aI919
aI1807
aI7903
aI5485
aI1733
aI64759
aI15817
aI93829
asS'd'
I13771
sa(dp27542
S'a'
I6028
sS's'
I18
sS'm'
(lp27543
I1
aI1
aI3
aI15
aI11
aI33
aI95
aI67
aI511
aI971
aI343
aI41
aI2849
aI10695
aI24487
aI8971
aI129279
aI197635
asS'd'
I13772
sa(dp27544
S'a'
I6039
sS's'
I18
sS'm'
(lp27545
I1
aI3
aI1
aI13
aI13
aI47
aI77
aI127
aI193
aI191
aI1185
aI3321
aI1685
aI1421
aI28675
aI12593
aI86689
aI186763
asS'd'
I13773
sa(dp27546
S'a'
I6056
sS's'
I18
sS'm'
(lp27547
I1
aI3
aI7
aI13
aI1
aI11
aI123
aI91
aI287
aI751
aI11
aI2753
aI7153
aI5253
aI21817
aI10459
aI122225
aI105775
asS'd'
I13774
sa(dp27548
S'a'
I6064
sS's'
I18
sS'm'
(lp27549
I1
aI3
aI3
aI13
aI1
aI17
aI121
aI13
aI391
aI253
aI1323
aI1515
aI2067
aI8009
aI5173
aI59543
aI109511
aI156821
asS'd'
I13775
sa(dp27550
S'a'
I6073
sS's'
I18
sS'm'
(lp27551
I1
aI1
aI5
aI3
aI7
aI1
aI119
aI151
aI281
aI859
aI675
aI2923
aI6627
aI16071
aI24653
aI41325
aI118413
aI191981
asS'd'
I13776
sa(dp27552
S'a'
I6081
sS's'
I18
sS'm'
(lp27553
I1
aI3
aI5
aI3
aI31
aI17
aI57
aI255
aI473
aI455
aI203
aI173
aI345
aI1477
aI27939
aI39289
aI105081
aI136179
asS'd'
I13777
sa(dp27554
S'a'
I6088
sS's'
I18
sS'm'
(lp27555
I1
aI3
aI3
aI7
aI1
aI35
aI29
aI81
aI337
aI483
aI951
aI955
aI4343
aI14827
aI17427
aI59919
aI81883
aI114289
asS'd'
I13778
sa(dp27556
S'a'
I6101
sS's'
I18
sS'm'
(lp27557
I1
aI3
aI5
aI11
aI3
aI39
aI49
aI177
aI335
aI57
aI173
aI1827
aI5729
aI2689
aI12109
aI13247
aI117559
aI31735
asS'd'
I13779
sa(dp27558
S'a'
I6139
sS's'
I18
sS'm'
(lp27559
I1
aI3
aI1
aI3
aI9
aI9
aI41
aI97
aI37
aI897
aI545
aI2289
aI7917
aI5701
aI21953
aI1863
aI33727
aI28451
asS'd'
I13780
sa(dp27560
S'a'
I6142
sS's'
I18
sS'm'
(lp27561
I1
aI1
aI5
aI3
aI29
aI61
aI59
aI129
aI387
aI965
aI285
aI3503
aI1651
aI10423
aI24861
aI31853
aI38491
aI155187
asS'd'
I13781
sa(dp27562
S'a'
I6194
sS's'
I18
sS'm'
(lp27563
I1
aI3
aI1
aI13
aI23
aI33
aI13
aI161
aI133
aI29
aI1073
aI1491
aI3687
aI6821
aI24153
aI3675
aI33771
aI230087
asS'd'
I13782
sa(dp27564
S'a'
I6199
sS's'
I18
sS'm'
(lp27565
I1
aI1
aI7
aI7
aI29
aI23
aI23
aI55
aI189
aI203
aI641
aI3391
aI1217
aI3199
aI32531
aI43103
aI24007
aI85613
asS'd'
I13783
sa(dp27566
S'a'
I6200
sS's'
I18
sS'm'
(lp27567
I1
aI1
aI1
aI11
aI7
aI57
aI117
aI245
aI467
aI861
aI1265
aI2827
aI2761
aI2817
aI15679
aI53223
aI47245
aI139871
asS'd'
I13784
sa(dp27568
S'a'
I6203
sS's'
I18
sS'm'
(lp27569
I1
aI1
aI3
aI15
aI25
aI1
aI125
aI237
aI489
aI1003
aI515
aI1117
aI4427
aI4877
aI8685
aI46211
aI19889
aI82491
asS'd'
I13785
sa(dp27570
S'a'
I6214
sS's'
I18
sS'm'
(lp27571
I1
aI1
aI3
aI3
aI25
aI3
aI63
aI217
aI485
aI699
aI161
aI1459
aI2973
aI15949
aI30681
aI30991
aI13933
aI86505
asS'd'
I13786
sa(dp27572
S'a'
I6238
sS's'
I18
sS'm'
(lp27573
I1
aI1
aI7
aI9
aI27
aI57
aI23
aI217
aI401
aI613
aI277
aI2827
aI7111
aI2133
aI17489
aI62059
aI7273
aI170917
asS'd'
I13787
sa(dp27574
S'a'
I6241
sS's'
I18
sS'm'
(lp27575
I1
aI1
aI3
aI13
aI7
aI19
aI39
aI63
aI203
aI1001
aI279
aI879
aI4293
aI10121
aI969
aI11571
aI96427
aI218969
asS'd'
I13788
sa(dp27576
S'a'
I6244
sS's'
I18
sS'm'
(lp27577
I1
aI3
aI7
aI7
aI5
aI21
aI113
aI203
aI77
aI971
aI1351
aI1097
aI2581
aI7519
aI16049
aI10565
aI5055
aI241561
asS'd'
I13789
sa(dp27578
S'a'
I6248
sS's'
I18
sS'm'
(lp27579
I1
aI3
aI3
aI11
aI1
aI21
aI93
aI111
aI221
aI31
aI1245
aI1499
aI2289
aI2299
aI23457
aI49221
aI68879
aI125029
asS'd'
I13790
sa(dp27580
S'a'
I6256
sS's'
I18
sS'm'
(lp27581
I1
aI1
aI7
aI3
aI15
aI19
aI57
aI189
aI243
aI785
aI399
aI3147
aI6107
aI2327
aI6275
aI9993
aI53051
aI34053
asS'd'
I13791
sa(dp27582
S'a'
I6266
sS's'
I18
sS'm'
(lp27583
I1
aI3
aI7
aI15
aI5
aI63
aI7
aI193
aI115
aI579
aI1987
aI765
aI7871
aI14179
aI26383
aI61455
aI14241
aI123515
asS'd'
I13792
sa(dp27584
S'a'
I6271
sS's'
I18
sS'm'
(lp27585
I1
aI3
aI7
aI13
aI19
aI13
aI91
aI225
aI295
aI675
aI1995
aI1145
aI4929
aI5163
aI1101
aI60681
aI76777
aI146875
asS'd'
I13793
sa(dp27586
S'a'
I6275
sS's'
I18
sS'm'
(lp27587
I1
aI3
aI7
aI15
aI21
aI37
aI57
aI89
aI297
aI143
aI717
aI4021
aI3259
aI8869
aI21189
aI39333
aI125045
aI94469
asS'd'
I13794
sa(dp27588
S'a'
I6295
sS's'
I18
sS'm'
(lp27589
I1
aI1
aI5
aI9
aI27
aI35
aI69
aI121
aI433
aI39
aI889
aI915
aI4055
aI11479
aI24757
aI53455
aI17503
aI113295
asS'd'
I13795
sa(dp27590
S'a'
I6315
sS's'
I18
sS'm'
(lp27591
I1
aI1
aI3
aI3
aI23
aI19
aI81
aI191
aI33
aI865
aI59
aI603
aI2819
aI4919
aI22495
aI25089
aI73905
aI44971
asS'd'
I13796
sa(dp27592
S'a'
I6364
sS's'
I18
sS'm'
(lp27593
I1
aI3
aI3
aI11
aI19
aI45
aI125
aI229
aI143
aI167
aI867
aI671
aI2225
aI16099
aI14909
aI14937
aI78063
aI135143
asS'd'
I13797
sa(dp27594
S'a'
I6368
sS's'
I18
sS'm'
(lp27595
I1
aI1
aI7
aI11
aI21
aI55
aI73
aI247
aI211
aI895
aI1147
aI17
aI2119
aI3261
aI19815
aI28055
aI50139
aI178459
asS'd'
I13798
sa(dp27596
S'a'
I6377
sS's'
I18
sS'm'
(lp27597
I1
aI1
aI3
aI1
aI19
aI3
aI37
aI221
aI243
aI459
aI1539
aI3899
aI4597
aI5503
aI23015
aI57019
aI62637
aI177821
asS'd'
I13799
sa(dp27598
S'a'
I6400
sS's'
I18
sS'm'
(lp27599
I1
aI3
aI1
aI3
aI9
aI5
aI91
aI3
aI319
aI609
aI1241
aI3953
aI5569
aI8757
aI6453
aI8083
aI55285
aI38297
asS'd'
I13800
sa(dp27600
S'a'
I6430
sS's'
I18
sS'm'
(lp27601
I1
aI3
aI5
aI15
aI9
aI51
aI37
aI53
aI137
aI95
aI123
aI157
aI15
aI7421
aI22469
aI49787
aI96245
aI199309
asS'd'
I13801
sa(dp27602
S'a'
I6477
sS's'
I18
sS'm'
(lp27603
I1
aI3
aI1
aI1
aI29
aI61
aI85
aI211
aI437
aI1013
aI1251
aI61
aI157
aI4325
aI24247
aI1065
aI24875
aI31509
asS'd'
I13802
sa(dp27604
S'a'
I6480
sS's'
I18
sS'm'
(lp27605
I1
aI3
aI1
aI13
aI7
aI43
aI13
aI171
aI53
aI567
aI77
aI3781
aI5077
aI6691
aI32485
aI24253
aI83919
aI159371
asS'd'
I13803
sa(dp27606
S'a'
I6490
sS's'
I18
sS'm'
(lp27607
I1
aI1
aI1
aI13
aI15
aI15
aI19
aI53
aI325
aI309
aI53
aI1857
aI7361
aI8831
aI31751
aI44749
aI109265
aI227875
asS'd'
I13804
sa(dp27608
S'a'
I6526
sS's'
I18
sS'm'
(lp27609
I1
aI3
aI7
aI3
aI31
aI19
aI113
aI253
aI361
aI697
aI1137
aI2029
aI3673
aI10323
aI10455
aI24935
aI7325
aI43673
asS'd'
I13805
sa(dp27610
S'a'
I6539
sS's'
I18
sS'm'
(lp27611
I1
aI3
aI1
aI1
aI17
aI31
aI3
aI55
aI121
aI967
aI1701
aI2171
aI4393
aI11937
aI3987
aI5139
aI68913
aI134233
asS'd'
I13806
sa(dp27612
S'a'
I6556
sS's'
I18
sS'm'
(lp27613
I1
aI1
aI7
aI15
aI23
aI37
aI121
aI241
aI297
aI419
aI373
aI1219
aI739
aI4567
aI28593
aI61267
aI95711
aI201299
asS'd'
I13807
sa(dp27614
S'a'
I6584
sS's'
I18
sS'm'
(lp27615
I1
aI1
aI1
aI9
aI23
aI31
aI101
aI243
aI163
aI333
aI1707
aI2553
aI5285
aI12827
aI5051
aI14165
aI505
aI253585
asS'd'
I13808
sa(dp27616
S'a'
I6598
sS's'
I18
sS'm'
(lp27617
I1
aI1
aI1
aI9
aI11
aI29
aI81
aI45
aI101
aI235
aI1079
aI4091
aI1069
aI3439
aI23599
aI6699
aI71783
aI236943
asS'd'
I13809
sa(dp27618
S'a'
I6601
sS's'
I18
sS'm'
(lp27619
I1
aI3
aI7
aI11
aI29
aI49
aI99
aI59
aI1
aI267
aI887
aI2941
aI6717
aI7501
aI22549
aI53393
aI34569
aI34671
asS'd'
I13810
sa(dp27620
S'a'
I6609
sS's'
I18
sS'm'
(lp27621
I1
aI3
aI1
aI15
aI23
aI13
aI113
aI47
aI11
aI79
aI989
aI1025
aI35
aI10475
aI8079
aI33121
aI32477
aI178595
asS'd'
I13811
sa(dp27622
S'a'
I6612
sS's'
I18
sS'm'
(lp27623
I1
aI3
aI7
aI1
aI21
aI19
aI51
aI31
aI393
aI171
aI553
aI2221
aI7017
aI8567
aI21803
aI51803
aI83737
aI196409
asS'd'
I13812
sa(dp27624
S'a'
I6645
sS's'
I18
sS'm'
(lp27625
I1
aI1
aI3
aI3
aI1
aI27
aI117
aI207
aI37
aI733
aI2001
aI2575
aI4849
aI5609
aI743
aI35987
aI109993
aI227663
asS'd'
I13813
sa(dp27626
S'a'
I6655
sS's'
I18
sS'm'
(lp27627
I1
aI1
aI7
aI15
aI29
aI47
aI85
aI213
aI335
aI633
aI849
aI3269
aI7723
aI4651
aI355
aI54565
aI58829
aI22781
asS'd'
I13814
sa(dp27628
S'a'
I6679
sS's'
I18
sS'm'
(lp27629
I1
aI3
aI7
aI5
aI5
aI3
aI91
aI243
aI17
aI85
aI1983
aI3909
aI1839
aI10403
aI503
aI28451
aI3221
aI215397
asS'd'
I13815
sa(dp27630
S'a'
I6683
sS's'
I18
sS'm'
(lp27631
I1
aI1
aI1
aI15
aI13
aI3
aI9
aI25
aI249
aI113
aI1619
aI2313
aI6461
aI2323
aI14319
aI59635
aI9569
aI220583
asS'd'
I13816
sa(dp27632
S'a'
I6719
sS's'
I18
sS'm'
(lp27633
I1
aI3
aI5
aI7
aI31
aI59
aI41
aI43
aI43
aI921
aI647
aI2141
aI7011
aI2749
aI24711
aI19067
aI107895
aI107145
asS'd'
I13817
sa(dp27634
S'a'
I6727
sS's'
I18
sS'm'
(lp27635
I1
aI1
aI3
aI3
aI15
aI63
aI41
aI241
aI181
aI729
aI843
aI3569
aI2645
aI2727
aI25331
aI23067
aI115421
aI86025
asS'd'
I13818
sa(dp27636
S'a'
I6734
sS's'
I18
sS'm'
(lp27637
I1
aI3
aI1
aI15
aI11
aI47
aI9
aI183
aI341
aI775
aI1067
aI1317
aI6835
aI7873
aI2653
aI33517
aI103979
aI196761
asS'd'
I13819
sa(dp27638
S'a'
I6770
sS's'
I18
sS'm'
(lp27639
I1
aI1
aI3
aI13
aI29
aI11
aI105
aI9
aI49
aI823
aI1343
aI759
aI1263
aI12413
aI26047
aI54285
aI57319
aI215387
asS'd'
I13820
sa(dp27640
S'a'
I6776
sS's'
I18
sS'm'
(lp27641
I1
aI1
aI5
aI13
aI21
aI55
aI75
aI149
aI63
aI737
aI1305
aI929
aI4149
aI2793
aI24505
aI11541
aI74765
aI8207
asS'd'
I13821
sa(dp27642
S'a'
I6791
sS's'
I18
sS'm'
(lp27643
I1
aI3
aI5
aI9
aI25
aI1
aI43
aI157
aI303
aI395
aI301
aI1561
aI5963
aI3501
aI2259
aI59777
aI100953
aI16051
asS'd'
I13822
sa(dp27644
S'a'
I6798
sS's'
I18
sS'm'
(lp27645
I1
aI3
aI1
aI13
aI19
aI25
aI33
aI209
aI11
aI95
aI655
aI595
aI3081
aI10345
aI26615
aI45129
aI84023
aI158079
asS'd'
I13823
sa(dp27646
S'a'
I6800
sS's'
I18
sS'm'
(lp27647
I1
aI3
aI1
aI9
aI31
aI61
aI103
aI203
aI471
aI215
aI1103
aI759
aI1197
aI3333
aI15859
aI36103
aI31563
aI5987
asS'd'
I13824
sa(dp27648
S'a'
I6826
sS's'
I18
sS'm'
(lp27649
I1
aI1
aI1
aI7
aI1
aI49
aI121
aI227
aI153
aI793
aI1723
aI1033
aI6875
aI6683
aI2503
aI57213
aI97967
aI120383
asS'd'
I13825
sa(dp27650
S'a'
I6833
sS's'
I18
sS'm'
(lp27651
I1
aI3
aI7
aI11
aI3
aI15
aI35
aI181
aI19
aI249
aI755
aI1385
aI3297
aI4665
aI2761
aI22717
aI126199
aI85065
asS'd'
I13826
sa(dp27652
S'a'
I6836
sS's'
I18
sS'm'
(lp27653
I1
aI3
aI7
aI15
aI5
aI45
aI17
aI55
aI111
aI597
aI553
aI1203
aI7183
aI8465
aI28523
aI50073
aI90889
aI187205
asS'd'
I13827
sa(dp27654
S'a'
I6858
sS's'
I18
sS'm'
(lp27655
I1
aI3
aI3
aI9
aI31
aI57
aI13
aI139
aI291
aI881
aI501
aI2051
aI617
aI5151
aI28225
aI44777
aI31645
aI6805
asS'd'
I13828
sa(dp27656
S'a'
I6882
sS's'
I18
sS'm'
(lp27657
I1
aI3
aI5
aI13
aI1
aI23
aI107
aI1
aI201
aI35
aI1673
aI2281
aI7663
aI1115
aI25061
aI59615
aI127955
aI169685
asS'd'
I13829
sa(dp27658
S'a'
I6901
sS's'
I18
sS'm'
(lp27659
I1
aI1
aI3
aI15
aI15
aI5
aI17
aI75
aI307
aI591
aI1661
aI855
aI4239
aI13359
aI20027
aI51871
aI35241
aI32769
asS'd'
I13830
sa(dp27660
S'a'
I6923
sS's'
I18
sS'm'
(lp27661
I1
aI1
aI1
aI15
aI5
aI23
aI59
aI205
aI223
aI103
aI1889
aI141
aI6157
aI9187
aI23571
aI15267
aI1941
aI119173
asS'd'
I13831
sa(dp27662
S'a'
I6967
sS's'
I18
sS'm'
(lp27663
I1
aI3
aI5
aI9
aI27
aI63
aI59
aI11
aI279
aI493
aI209
aI4087
aI1055
aI9841
aI31753
aI37459
aI27757
aI213151
asS'd'
I13832
sa(dp27664
S'a'
I6979
sS's'
I18
sS'm'
(lp27665
I1
aI1
aI3
aI11
aI31
aI33
aI13
aI87
aI285
aI847
aI2005
aI3431
aI253
aI15157
aI31359
aI45303
aI114337
aI42541
asS'd'
I13833
sa(dp27666
S'a'
I7022
sS's'
I18
sS'm'
(lp27667
I1
aI1
aI1
aI7
aI7
aI3
aI17
aI203
aI133
aI321
aI241
aI1323
aI5639
aI10953
aI10069
aI4941
aI17077
aI54493
asS'd'
I13834
sa(dp27668
S'a'
I7060
sS's'
I18
sS'm'
(lp27669
I1
aI1
aI7
aI1
aI9
aI5
aI125
aI5
aI421
aI609
aI645
aI1927
aI3785
aI2295
aI1491
aI23019
aI85497
aI161231
asS'd'
I13835
sa(dp27670
S'a'
I7064
sS's'
I18
sS'm'
(lp27671
I1
aI3
aI3
aI11
aI19
aI35
aI115
aI95
aI353
aI773
aI2025
aI2621
aI2821
aI6361
aI29589
aI20989
aI99645
aI90387
asS'd'
I13836
sa(dp27672
S'a'
I7083
sS's'
I18
sS'm'
(lp27673
I1
aI3
aI1
aI15
aI23
aI3
aI71
aI253
aI467
aI307
aI1109
aI2695
aI7175
aI15087
aI1587
aI48229
aI104307
aI218905
asS'd'
I13837
sa(dp27674
S'a'
I7093
sS's'
I18
sS'm'
(lp27675
I1
aI3
aI3
aI5
aI17
aI29
aI107
aI153
aI371
aI205
aI7
aI597
aI7393
aI2345
aI20149
aI47417
aI37983
aI200683
asS'd'
I13838
sa(dp27676
S'a'
I7094
sS's'
I18
sS'm'
(lp27677
I1
aI1
aI7
aI1
aI3
aI55
aI113
aI117
aI241
aI923
aI1217
aI3825
aI2635
aI8007
aI12673
aI9533
aI7121
aI3825
asS'd'
I13839
sa(dp27678
S'a'
I7120
sS's'
I18
sS'm'
(lp27679
I1
aI1
aI1
aI3
aI21
aI63
aI25
aI111
aI31
aI151
aI67
aI3735
aI7833
aI749
aI28743
aI59291
aI4989
aI93329
asS'd'
I13840
sa(dp27680
S'a'
I7125
sS's'
I18
sS'm'
(lp27681
I1
aI3
aI5
aI13
aI31
aI5
aI91
aI153
aI235
aI1019
aI431
aI1951
aI7501
aI8483
aI19625
aI57789
aI13203
aI36693
asS'd'
I13841
sa(dp27682
S'a'
I7142
sS's'
I18
sS'm'
(lp27683
I1
aI1
aI7
aI3
aI15
aI51
aI99
aI29
aI403
aI343
aI1903
aI907
aI3255
aI4149
aI29551
aI18885
aI74391
aI96119
asS'd'
I13842
sa(dp27684
S'a'
I7151
sS's'
I18
sS'm'
(lp27685
I1
aI1
aI7
aI11
aI31
aI11
aI63
aI213
aI437
aI879
aI359
aI555
aI7549
aI14269
aI31489
aI51001
aI76857
aI237305
asS'd'
I13843
sa(dp27686
S'a'
I7165
sS's'
I18
sS'm'
(lp27687
I1
aI1
aI1
aI15
aI23
aI21
aI31
aI227
aI311
aI273
aI253
aI2439
aI7217
aI2191
aI31743
aI47669
aI62279
aI201305
asS'd'
I13844
sa(dp27688
S'a'
I7173
sS's'
I18
sS'm'
(lp27689
I1
aI3
aI5
aI11
aI3
aI43
aI97
aI121
aI363
aI91
aI201
aI1095
aI5267
aI633
aI19111
aI36099
aI23035
aI205655
asS'd'
I13845
sa(dp27690
S'a'
I7188
sS's'
I18
sS'm'
(lp27691
I1
aI3
aI1
aI9
aI21
aI31
aI39
aI11
aI227
aI699
aI473
aI2109
aI2757
aI13821
aI31181
aI40493
aI57279
aI260085
asS'd'
I13846
sa(dp27692
S'a'
I7191
sS's'
I18
sS'm'
(lp27693
I1
aI1
aI5
aI7
aI27
aI57
aI65
aI11
aI173
aI709
aI1139
aI3735
aI5291
aI16053
aI32579
aI25275
aI79865
aI196033
asS'd'
I13847
sa(dp27694
S'a'
I7202
sS's'
I18
sS'm'
(lp27695
I1
aI1
aI1
aI9
aI19
aI35
aI83
aI153
aI287
aI207
aI593
aI2177
aI3243
aI10433
aI24583
aI881
aI71865
aI250223
asS'd'
I13848
sa(dp27696
S'a'
I7216
sS's'
I18
sS'm'
(lp27697
I1
aI3
aI3
aI9
aI3
aI35
aI107
aI3
aI193
aI1011
aI463
aI1643
aI2733
aI2157
aI6329
aI24583
aI116901
aI226385
asS'd'
I13849
sa(dp27698
S'a'
I7219
sS's'
I18
sS'm'
(lp27699
I1
aI3
aI1
aI11
aI3
aI47
aI39
aI33
aI495
aI137
aI1591
aI1335
aI1347
aI4527
aI389
aI43341
aI80163
aI5219
asS'd'
I13850
sa(dp27700
S'a'
I7267
sS's'
I18
sS'm'
(lp27701
I1
aI3
aI3
aI5
aI19
aI51
aI121
aI135
aI93
aI891
aI13
aI1339
aI5187
aI5005
aI12823
aI14465
aI73845
aI119685
asS'd'
I13851
sa(dp27702
S'a'
I7358
sS's'
I18
sS'm'
(lp27703
I1
aI3
aI7
aI7
aI19
aI1
aI57
aI193
aI325
aI49
aI813
aI5
aI4431
aI1119
aI13625
aI43613
aI127989
aI42669
asS'd'
I13852
sa(dp27704
S'a'
I7383
sS's'
I18
sS'm'
(lp27705
I1
aI3
aI5
aI15
aI11
aI33
aI7
aI45
aI215
aI469
aI1059
aI4095
aI3549
aI11839
aI5463
aI21383
aI4831
aI188345
asS'd'
I13853
sa(dp27706
S'a'
I7387
sS's'
I18
sS'm'
(lp27707
I1
aI1
aI1
aI13
aI5
aI61
aI7
aI161
aI99
aI623
aI1589
aI1045
aI2385
aI8899
aI19327
aI41373
aI109241
aI111895
asS'd'
I13854
sa(dp27708
S'a'
I7396
sS's'
I18
sS'm'
(lp27709
I1
aI1
aI5
aI7
aI11
aI39
aI115
aI41
aI21
aI491
aI1221
aI2805
aI4311
aI7137
aI3151
aI1387
aI24633
aI94679
asS'd'
I13855
sa(dp27710
S'a'
I7408
sS's'
I18
sS'm'
(lp27711
I1
aI1
aI3
aI7
aI3
aI39
aI71
aI175
aI443
aI187
aI1727
aI2535
aI5099
aI1881
aI21639
aI5717
aI48589
aI95037
asS'd'
I13856
sa(dp27712
S'a'
I7432
sS's'
I18
sS'm'
(lp27713
I1
aI1
aI1
aI11
aI25
aI21
aI37
aI227
aI407
aI73
aI721
aI3515
aI381
aI981
aI21389
aI5205
aI31851
aI140457
asS'd'
I13857
sa(dp27714
S'a'
I7462
sS's'
I18
sS'm'
(lp27715
I1
aI3
aI3
aI3
aI27
aI35
aI13
aI129
aI457
aI315
aI253
aI2545
aI5469
aI6695
aI25223
aI20115
aI38039
aI133655
asS'd'
I13858
sa(dp27716
S'a'
I7480
sS's'
I18
sS'm'
(lp27717
I1
aI1
aI3
aI11
aI21
aI47
aI77
aI231
aI87
aI245
aI2039
aI2515
aI2873
aI1711
aI3361
aI62123
aI67117
aI239047
asS'd'
I13859
sa(dp27718
S'a'
I7488
sS's'
I18
sS'm'
(lp27719
I1
aI1
aI7
aI7
aI29
aI21
aI39
aI175
aI477
aI813
aI447
aI1109
aI7391
aI14631
aI4437
aI42539
aI13003
aI75403
asS'd'
I13860
sa(dp27720
S'a'
I7497
sS's'
I18
sS'm'
(lp27721
I1
aI3
aI1
aI11
aI11
aI9
aI119
aI19
aI99
aI483
aI61
aI1883
aI3415
aI2137
aI30415
aI34519
aI115191
aI24437
asS'd'
I13861
sa(dp27722
S'a'
I7531
sS's'
I18
sS'm'
(lp27723
I1
aI1
aI1
aI9
aI7
aI47
aI115
aI233
aI419
aI427
aI1605
aI3821
aI6243
aI10861
aI28495
aI48265
aI80811
aI147701
asS'd'
I13862
sa(dp27724
S'a'
I7561
sS's'
I18
sS'm'
(lp27725
I1
aI3
aI5
aI5
aI11
aI61
aI51
aI155
aI279
aI463
aI31
aI1559
aI2837
aI8795
aI4049
aI13651
aI109227
aI52131
asS'd'
I13863
sa(dp27726
S'a'
I7582
sS's'
I18
sS'm'
(lp27727
I1
aI1
aI7
aI9
aI25
aI33
aI97
aI79
aI477
aI83
aI923
aI3293
aI6381
aI3063
aI23293
aI35381
aI82867
aI233189
asS'd'
I13864
sa(dp27728
S'a'
I7591
sS's'
I18
sS'm'
(lp27729
I1
aI1
aI1
aI9
aI11
aI39
aI109
aI189
aI219
aI1021
aI137
aI2041
aI2719
aI1763
aI31787
aI29377
aI96287
aI179685
asS'd'
I13865
sa(dp27730
S'a'
I7612
sS's'
I18
sS'm'
(lp27731
I1
aI3
aI7
aI5
aI3
aI15
aI37
aI179
aI77
aI751
aI709
aI893
aI7705
aI1563
aI7843
aI29843
aI1107
aI35919
asS'd'
I13866
sa(dp27732
S'a'
I7677
sS's'
I18
sS'm'
(lp27733
I1
aI3
aI3
aI3
aI29
aI17
aI123
aI201
aI275
aI487
aI1979
aI1361
aI7523
aI13783
aI10129
aI16877
aI127049
aI163221
asS'd'
I13867
sa(dp27734
S'a'
I7687
sS's'
I18
sS'm'
(lp27735
I1
aI1
aI1
aI9
aI27
aI35
aI69
aI167
aI509
aI133
aI1073
aI3773
aI265
aI8455
aI12341
aI127
aI115075
aI94537
asS'd'
I13868
sa(dp27736
S'a'
I7701
sS's'
I18
sS'm'
(lp27737
I1
aI3
aI7
aI9
aI13
aI35
aI57
aI83
aI123
aI211
aI739
aI253
aI3907
aI5405
aI3229
aI46837
aI77483
aI5915
asS'd'
I13869
sa(dp27738
S'a'
I7705
sS's'
I18
sS'm'
(lp27739
I1
aI1
aI7
aI3
aI7
aI27
aI71
aI235
aI133
aI803
aI611
aI529
aI4449
aI16113
aI8151
aI36519
aI34561
aI36361
asS'd'
I13870
sa(dp27740
S'a'
I7712
sS's'
I18
sS'm'
(lp27741
I1
aI1
aI3
aI1
aI31
aI9
aI123
aI85
aI407
aI415
aI353
aI3239
aI673
aI4641
aI25883
aI61117
aI7669
aI240851
asS'd'
I13871
sa(dp27742
S'a'
I7721
sS's'
I18
sS'm'
(lp27743
I1
aI1
aI5
aI11
aI29
aI9
aI49
aI31
aI3
aI249
aI1769
aI3325
aI503
aI1397
aI30677
aI22515
aI81279
aI90309
asS'd'
I13872
sa(dp27744
S'a'
I7736
sS's'
I18
sS'm'
(lp27745
I1
aI3
aI5
aI3
aI15
aI63
aI121
aI253
aI421
aI279
aI497
aI3881
aI6977
aI11061
aI5883
aI38347
aI8351
aI118123
asS'd'
I13873
sa(dp27746
S'a'
I7756
sS's'
I18
sS'm'
(lp27747
I1
aI1
aI1
aI13
aI1
aI27
aI3
aI91
aI281
aI563
aI1283
aI1893
aI7593
aI12171
aI27041
aI7769
aI95691
aI13791
asS'd'
I13874
sa(dp27748
S'a'
I7777
sS's'
I18
sS'm'
(lp27749
I1
aI1
aI3
aI3
aI29
aI59
aI87
aI153
aI337
aI819
aI787
aI2631
aI1889
aI13869
aI29237
aI57097
aI91621
aI4011
asS'd'
I13875
sa(dp27750
S'a'
I7783
sS's'
I18
sS'm'
(lp27751
I1
aI1
aI1
aI15
aI7
aI49
aI1
aI83
aI299
aI353
aI131
aI1635
aI3723
aI16209
aI1061
aI50669
aI68083
aI133443
asS'd'
I13876
sa(dp27752
S'a'
I7792
sS's'
I18
sS'm'
(lp27753
I1
aI3
aI5
aI5
aI27
aI17
aI5
aI239
aI285
aI831
aI1487
aI721
aI4891
aI4265
aI23753
aI43921
aI116709
aI105027
asS'd'
I13877
sa(dp27754
S'a'
I7797
sS's'
I18
sS'm'
(lp27755
I1
aI1
aI7
aI7
aI5
aI35
aI63
aI97
aI215
aI447
aI353
aI495
aI8119
aI12537
aI9679
aI58641
aI65057
aI21999
asS'd'
I13878
sa(dp27756
S'a'
I7802
sS's'
I18
sS'm'
(lp27757
I1
aI3
aI7
aI5
aI23
aI43
aI69
aI115
aI59
aI603
aI493
aI1665
aI5003
aI13607
aI28491
aI4439
aI11855
aI228183
asS'd'
I13879
sa(dp27758
S'a'
I7808
sS's'
I18
sS'm'
(lp27759
I1
aI3
aI3
aI3
aI15
aI35
aI19
aI63
aI241
aI357
aI979
aI2891
aI3105
aI14085
aI10539
aI62335
aI130903
aI163153
asS'd'
I13880
sa(dp27760
S'a'
I7866
sS's'
I18
sS'm'
(lp27761
I1
aI3
aI5
aI3
aI23
aI51
aI23
aI193
aI129
aI171
aI1913
aI1025
aI6397
aI15657
aI19611
aI57455
aI87531
aI51039
asS'd'
I13881
sa(dp27762
S'a'
I7885
sS's'
I18
sS'm'
(lp27763
I1
aI1
aI3
aI9
aI1
aI9
aI7
aI239
aI87
aI527
aI1401
aI2703
aI4021
aI3845
aI29269
aI48217
aI61091
aI131949
asS'd'
I13882
sa(dp27764
S'a'
I7891
sS's'
I18
sS'm'
(lp27765
I1
aI3
aI1
aI5
aI17
aI45
aI59
aI223
aI287
aI295
aI1959
aI3985
aI3671
aI14605
aI18949
aI34147
aI51251
aI10271
asS'd'
I13883
sa(dp27766
S'a'
I7913
sS's'
I18
sS'm'
(lp27767
I1
aI3
aI5
aI9
aI5
aI49
aI63
aI105
aI43
aI157
aI1827
aI495
aI5823
aI6323
aI6601
aI51379
aI64411
aI204103
asS'd'
I13884
sa(dp27768
S'a'
I7936
sS's'
I18
sS'm'
(lp27769
I1
aI1
aI7
aI1
aI11
aI31
aI117
aI9
aI13
aI965
aI177
aI1247
aI2487
aI9849
aI20367
aI49287
aI2193
aI235689
asS'd'
I13885
sa(dp27770
S'a'
I7946
sS's'
I18
sS'm'
(lp27771
I1
aI1
aI3
aI11
aI19
aI31
aI23
aI215
aI489
aI657
aI801
aI3937
aI379
aI12083
aI14969
aI37857
aI39027
aI63985
asS'd'
I13886
sa(dp27772
S'a'
I7963
sS's'
I18
sS'm'
(lp27773
I1
aI1
aI5
aI1
aI9
aI59
aI53
aI187
aI341
aI65
aI1251
aI767
aI4897
aI13263
aI17439
aI26625
aI122107
aI163653
asS'd'
I13887
sa(dp27774
S'a'
I7996
sS's'
I18
sS'm'
(lp27775
I1
aI1
aI3
aI13
aI31
aI41
aI125
aI253
aI481
aI107
aI233
aI2305
aI3321
aI7303
aI28585
aI12787
aI83307
aI31497
asS'd'
I13888
sa(dp27776
S'a'
I7999
sS's'
I18
sS'm'
(lp27777
I1
aI1
aI3
aI13
aI25
aI41
aI55
aI83
aI101
aI115
aI549
aI531
aI3085
aI9497
aI27989
aI28257
aI121075
aI189671
asS'd'
I13889
sa(dp27778
S'a'
I8011
sS's'
I18
sS'm'
(lp27779
I1
aI3
aI5
aI11
aI13
aI53
aI121
aI85
aI355
aI275
aI1925
aI2117
aI1349
aI5903
aI2041
aI20963
aI60803
aI1121
asS'd'
I13890
sa(dp27780
S'a'
I8014
sS's'
I18
sS'm'
(lp27781
I1
aI1
aI5
aI5
aI13
aI7
aI125
aI63
aI311
aI187
aI1127
aI643
aI6137
aI845
aI23945
aI9403
aI451
aI53027
asS'd'
I13891
sa(dp27782
S'a'
I8028
sS's'
I18
sS'm'
(lp27783
I1
aI1
aI3
aI7
aI19
aI31
aI27
aI239
aI337
aI61
aI641
aI1693
aI7289
aI5675
aI30067
aI41091
aI124607
aI36971
asS'd'
I13892
sa(dp27784
S'a'
I8047
sS's'
I18
sS'm'
(lp27785
I1
aI3
aI1
aI13
aI25
aI61
aI11
aI81
aI165
aI129
aI241
aI711
aI5193
aI13017
aI30821
aI35239
aI110809
aI60909
asS'd'
I13893
sa(dp27786
S'a'
I8077
sS's'
I18
sS'm'
(lp27787
I1
aI3
aI3
aI13
aI15
aI13
aI71
aI19
aI87
aI499
aI1395
aI1191
aI1445
aI2687
aI4691
aI16773
aI114269
aI186237
asS'd'
I13894
sa(dp27788
S'a'
I8101
sS's'
I18
sS'm'
(lp27789
I1
aI3
aI3
aI15
aI1
aI49
aI33
aI109
aI241
aI5
aI431
aI461
aI3865
aI14029
aI9827
aI54455
aI52159
aI211585
asS'd'
I13895
sa(dp27790
S'a'
I8119
sS's'
I18
sS'm'
(lp27791
I1
aI3
aI5
aI7
aI31
aI27
aI115
aI113
aI367
aI591
aI873
aI1447
aI6819
aI7011
aI14095
aI55243
aI4039
aI226985
asS'd'
I13896
sa(dp27792
S'a'
I8125
sS's'
I18
sS'm'
(lp27793
I1
aI1
aI7
aI3
aI1
aI19
aI69
aI239
aI417
aI833
aI1867
aI3111
aI2617
aI12781
aI5531
aI17345
aI75717
aI139667
asS'd'
I13897
sa(dp27794
S'a'
I8140
sS's'
I18
sS'm'
(lp27795
I1
aI1
aI5
aI11
aI11
aI63
aI23
aI141
aI221
aI897
aI1269
aI2185
aI6057
aI8865
aI20449
aI58255
aI27073
aI158305
asS'd'
I13898
sa(dp27796
S'a'
I8152
sS's'
I18
sS'm'
(lp27797
I1
aI3
aI3
aI5
aI23
aI23
aI121
aI39
aI457
aI935
aI691
aI2329
aI7055
aI2821
aI12669
aI28713
aI82321
aI245783
asS'd'
I13899
sa(dp27798
S'a'
I8209
sS's'
I18
sS'm'
(lp27799
I1
aI1
aI1
aI9
aI27
aI9
aI35
aI23
aI139
aI823
aI703
aI917
aI1281
aI12155
aI11681
aI26083
aI119445
aI181489
asS'd'
I13900
sa(dp27800
S'a'
I8212
sS's'
I18
sS'm'
(lp27801
I1
aI1
aI3
aI7
aI27
aI21
aI35
aI243
aI17
aI633
aI1665
aI3419
aI6301
aI16099
aI17477
aI24983
aI128455
aI127501
asS'd'
I13901
sa(dp27802
S'a'
I8231
sS's'
I18
sS'm'
(lp27803
I1
aI3
aI1
aI1
aI17
aI19
aI59
aI165
aI487
aI985
aI597
aI689
aI7103
aI14475
aI6985
aI29755
aI115977
aI105943
asS'd'
I13902
sa(dp27804
S'a'
I8258
sS's'
I18
sS'm'
(lp27805
I1
aI1
aI5
aI5
aI23
aI41
aI67
aI175
aI3
aI571
aI1501
aI3315
aI6111
aI1847
aI28975
aI54117
aI66605
aI69997
asS'd'
I13903
sa(dp27806
S'a'
I8275
sS's'
I18
sS'm'
(lp27807
I1
aI3
aI5
aI5
aI13
aI37
aI113
aI75
aI383
aI297
aI1187
aI2055
aI3433
aI14651
aI30393
aI29647
aI126403
aI32265
asS'd'
I13904
sa(dp27808
S'a'
I8303
sS's'
I18
sS'm'
(lp27809
I1
aI1
aI3
aI5
aI31
aI29
aI25
aI169
aI465
aI219
aI81
aI2019
aI4255
aI6003
aI7425
aI53269
aI31105
aI211937
asS'd'
I13905
sa(dp27810
S'a'
I8308
sS's'
I18
sS'm'
(lp27811
I1
aI3
aI7
aI11
aI13
aI7
aI11
aI195
aI327
aI883
aI1295
aI3721
aI1197
aI7585
aI5693
aI993
aI125017
aI12007
asS'd'
I13906
sa(dp27812
S'a'
I8312
sS's'
I18
sS'm'
(lp27813
I1
aI3
aI3
aI7
aI5
aI37
aI71
aI37
aI63
aI651
aI669
aI3445
aI3959
aI249
aI10599
aI22329
aI107701
aI107729
asS'd'
I13907
sa(dp27814
S'a'
I8345
sS's'
I18
sS'm'
(lp27815
I1
aI1
aI1
aI9
aI7
aI47
aI21
aI181
aI395
aI345
aI757
aI481
aI2759
aI8157
aI19847
aI55743
aI63137
aI224765
asS'd'
I13908
sa(dp27816
S'a'
I8346
sS's'
I18
sS'm'
(lp27817
I1
aI3
aI5
aI9
aI29
aI3
aI61
aI35
aI271
aI157
aI549
aI843
aI2907
aI91
aI16325
aI4241
aI94495
aI78861
asS'd'
I13909
sa(dp27818
S'a'
I8370
sS's'
I18
sS'm'
(lp27819
I1
aI3
aI1
aI9
aI17
aI11
aI53
aI243
aI49
aI911
aI1193
aI793
aI901
aI3727
aI21849
aI33987
aI565
aI154171
asS'd'
I13910
sa(dp27820
S'a'
I8402
sS's'
I18
sS'm'
(lp27821
I1
aI1
aI5
aI1
aI9
aI5
aI89
aI81
aI65
aI111
aI781
aI3775
aI591
aI4987
aI29833
aI58159
aI7253
aI206447
asS'd'
I13911
sa(dp27822
S'a'
I8411
sS's'
I18
sS'm'
(lp27823
I1
aI3
aI1
aI7
aI3
aI59
aI77
aI83
aI173
aI545
aI103
aI2541
aI8095
aI10797
aI11111
aI62351
aI88827
aI55081
asS'd'
I13912
sa(dp27824
S'a'
I8414
sS's'
I18
sS'm'
(lp27825
I1
aI1
aI3
aI11
aI29
aI37
aI19
aI47
aI145
aI19
aI513
aI3269
aI2205
aI5317
aI19207
aI38051
aI5413
aI78089
asS'd'
I13913
sa(dp27826
S'a'
I8424
sS's'
I18
sS'm'
(lp27827
I1
aI1
aI5
aI9
aI21
aI57
aI75
aI249
aI21
aI879
aI1377
aI3407
aI6123
aI11917
aI12493
aI44873
aI113539
aI114717
asS'd'
I13914
sa(dp27828
S'a'
I8435
sS's'
I18
sS'm'
(lp27829
I1
aI3
aI3
aI9
aI7
aI55
aI121
aI57
aI491
aI39
aI1561
aI2625
aI639
aI13553
aI1159
aI43071
aI68869
aI248837
asS'd'
I13915
sa(dp27830
S'a'
I8452
sS's'
I18
sS'm'
(lp27831
I1
aI1
aI1
aI11
aI25
aI19
aI107
aI239
aI171
aI1001
aI69
aI4095
aI49
aI9569
aI22613
aI59865
aI54959
aI70031
asS'd'
I13916
sa(dp27832
S'a'
I8462
sS's'
I18
sS'm'
(lp27833
I1
aI1
aI3
aI13
aI27
aI15
aI105
aI205
aI205
aI581
aI1965
aI1535
aI6531
aI15935
aI7623
aI33695
aI9317
aI44257
asS'd'
I13917
sa(dp27834
S'a'
I8479
sS's'
I18
sS'm'
(lp27835
I1
aI1
aI1
aI3
aI3
aI51
aI115
aI185
aI315
aI763
aI211
aI339
aI7083
aI4895
aI23277
aI14165
aI101731
aI218903
asS'd'
I13918
sa(dp27836
S'a'
I8509
sS's'
I18
sS'm'
(lp27837
I1
aI1
aI3
aI13
aI29
aI1
aI69
aI55
aI423
aI781
aI183
aI1417
aI151
aI14507
aI5217
aI27757
aI52447
aI145913
asS'd'
I13919
sa(dp27838
S'a'
I8515
sS's'
I18
sS'm'
(lp27839
I1
aI3
aI1
aI11
aI29
aI39
aI29
aI151
aI85
aI387
aI885
aI507
aI133
aI9819
aI12627
aI30951
aI79839
aI206267
asS'd'
I13920
sa(dp27840
S'a'
I8522
sS's'
I18
sS'm'
(lp27841
I1
aI3
aI3
aI7
aI1
aI53
aI99
aI141
aI91
aI51
aI143
aI1751
aI3989
aI6811
aI7339
aI52141
aI43473
aI18615
asS'd'
I13921
sa(dp27842
S'a'
I8541
sS's'
I18
sS'm'
(lp27843
I1
aI3
aI3
aI15
aI27
aI11
aI29
aI37
aI387
aI655
aI2019
aI1135
aI3619
aI12995
aI12755
aI26063
aI109419
aI103875
asS'd'
I13922
sa(dp27844
S'a'
I8560
sS's'
I18
sS'm'
(lp27845
I1
aI3
aI3
aI13
aI31
aI15
aI93
aI231
aI195
aI261
aI1055
aI2363
aI1123
aI3927
aI6907
aI365
aI27043
aI157049
asS'd'
I13923
sa(dp27846
S'a'
I8563
sS's'
I18
sS'm'
(lp27847
I1
aI1
aI1
aI15
aI7
aI29
aI105
aI199
aI507
aI437
aI117
aI2963
aI7801
aI6291
aI19261
aI30377
aI92205
aI20723
asS'd'
I13924
sa(dp27848
S'a'
I8585
sS's'
I18
sS'm'
(lp27849
I1
aI1
aI1
aI9
aI29
aI19
aI75
aI189
aI3
aI387
aI1491
aI2291
aI7739
aI12993
aI11835
aI10873
aI54583
aI207963
asS'd'
I13925
sa(dp27850
S'a'
I8594
sS's'
I18
sS'm'
(lp27851
I1
aI3
aI7
aI3
aI23
aI11
aI25
aI105
aI57
aI713
aI1291
aI3293
aI4693
aI13859
aI27541
aI31529
aI65929
aI245143
asS'd'
I13926
sa(dp27852
S'a'
I8596
sS's'
I18
sS'm'
(lp27853
I1
aI1
aI7
aI7
aI19
aI13
aI19
aI189
aI253
aI337
aI351
aI1751
aI6173
aI12207
aI24483
aI31381
aI82035
aI157143
asS'd'
I13927
sa(dp27854
S'a'
I8603
sS's'
I18
sS'm'
(lp27855
I1
aI3
aI3
aI11
aI11
aI49
aI117
aI177
aI301
aI417
aI855
aI2433
aI5619
aI7339
aI30361
aI29251
aI20411
aI184981
asS'd'
I13928
sa(dp27856
S'a'
I8610
sS's'
I18
sS'm'
(lp27857
I1
aI1
aI1
aI1
aI11
aI55
aI77
aI99
aI209
aI781
aI1193
aI2841
aI783
aI1485
aI19413
aI52255
aI19529
aI253927
asS'd'
I13929
sa(dp27858
S'a'
I8647
sS's'
I18
sS'm'
(lp27859
I1
aI3
aI1
aI3
aI15
aI49
aI85
aI191
aI389
aI411
aI479
aI341
aI4985
aI6193
aI19099
aI11497
aI103285
aI162333
asS'd'
I13930
sa(dp27860
S'a'
I8661
sS's'
I18
sS'm'
(lp27861
I1
aI3
aI1
aI3
aI11
aI31
aI71
aI91
aI357
aI615
aI2007
aI3601
aI5393
aI8079
aI16811
aI54127
aI26049
aI116341
asS'd'
I13931
sa(dp27862
S'a'
I8662
sS's'
I18
sS'm'
(lp27863
I1
aI1
aI3
aI15
aI9
aI39
aI121
aI53
aI43
aI617
aI905
aI3629
aI6327
aI13453
aI1435
aI24113
aI7523
aI228523
asS'd'
I13932
sa(dp27864
S'a'
I8672
sS's'
I18
sS'm'
(lp27865
I1
aI3
aI5
aI11
aI21
aI51
aI11
aI125
aI33
aI935
aI1069
aI2807
aI4951
aI13261
aI17611
aI38779
aI62203
aI135759
asS'd'
I13933
sa(dp27866
S'a'
I8690
sS's'
I18
sS'm'
(lp27867
I1
aI1
aI1
aI13
aI29
aI59
aI53
aI245
aI219
aI423
aI809
aI1109
aI7255
aI14679
aI25247
aI43235
aI129565
aI72649
asS'd'
I13934
sa(dp27868
S'a'
I8696
sS's'
I18
sS'm'
(lp27869
I1
aI3
aI5
aI7
aI27
aI29
aI119
aI91
aI297
aI407
aI187
aI2829
aI5637
aI13851
aI14073
aI461
aI64081
aI33971
asS'd'
I13935
sa(dp27870
S'a'
I8739
sS's'
I18
sS'm'
(lp27871
I1
aI1
aI5
aI11
aI15
aI27
aI29
aI233
aI487
aI859
aI1021
aI3117
aI1439
aI16021
aI31315
aI35775
aI117363
aI131635
asS'd'
I13936
sa(dp27872
S'a'
I8783
sS's'
I18
sS'm'
(lp27873
I1
aI1
aI5
aI3
aI3
aI1
aI91
aI229
aI327
aI777
aI393
aI3853
aI3455
aI1785
aI13749
aI25173
aI51575
aI167237
asS'd'
I13937
sa(dp27874
S'a'
I8811
sS's'
I18
sS'm'
(lp27875
I1
aI3
aI7
aI9
aI27
aI7
aI15
aI71
aI283
aI71
aI1783
aI1357
aI5581
aI3143
aI26075
aI47751
aI71001
aI157107
asS'd'
I13938
sa(dp27876
S'a'
I8813
sS's'
I18
sS'm'
(lp27877
I1
aI3
aI3
aI7
aI23
aI9
aI69
aI21
aI333
aI223
aI1735
aI1057
aI8091
aI1927
aI8507
aI40901
aI40233
aI164115
asS'd'
I13939
sa(dp27878
S'a'
I8821
sS's'
I18
sS'm'
(lp27879
I1
aI3
aI7
aI3
aI11
aI49
aI29
aI81
aI215
aI289
aI1137
aI765
aI6385
aI5935
aI3435
aI11991
aI30867
aI60745
asS'd'
I13940
sa(dp27880
S'a'
I8856
sS's'
I18
sS'm'
(lp27881
I1
aI1
aI1
aI1
aI7
aI39
aI33
aI173
aI225
aI533
aI1927
aI3607
aI1059
aI8779
aI2649
aI6801
aI103963
aI167471
asS'd'
I13941
sa(dp27882
S'a'
I8877
sS's'
I18
sS'm'
(lp27883
I1
aI3
aI3
aI15
aI27
aI51
aI107
aI3
aI195
aI87
aI739
aI1425
aI747
aI1501
aI22245
aI59233
aI124867
aI79753
asS'd'
I13942
sa(dp27884
S'a'
I8885
sS's'
I18
sS'm'
(lp27885
I1
aI1
aI3
aI3
aI13
aI41
aI125
aI101
aI225
aI749
aI221
aI2735
aI6441
aI11353
aI3943
aI35329
aI53437
aI149063
asS'd'
I13943
sa(dp27886
S'a'
I8897
sS's'
I18
sS'm'
(lp27887
I1
aI3
aI3
aI15
aI3
aI53
aI75
aI77
aI1
aI907
aI573
aI1909
aI363
aI6913
aI559
aI58489
aI1053
aI25513
asS'd'
I13944
sa(dp27888
S'a'
I8898
sS's'
I18
sS'm'
(lp27889
I1
aI1
aI7
aI11
aI7
aI15
aI91
aI155
aI447
aI555
aI473
aI3625
aI7529
aI16307
aI32241
aI64077
aI46943
aI85717
asS'd'
I13945
sa(dp27890
S'a'
I8934
sS's'
I18
sS'm'
(lp27891
I1
aI1
aI1
aI13
aI9
aI61
aI91
aI41
aI101
aI107
aI1081
aI2511
aI2881
aI14095
aI3861
aI22771
aI32687
aI77287
asS'd'
I13946
sa(dp27892
S'a'
I8955
sS's'
I18
sS'm'
(lp27893
I1
aI3
aI3
aI7
aI21
aI3
aI51
aI177
aI203
aI861
aI1507
aI1177
aI2369
aI11735
aI1667
aI28607
aI97671
aI123263
asS'd'
I13947
sa(dp27894
S'a'
I8972
sS's'
I18
sS'm'
(lp27895
I1
aI1
aI1
aI3
aI5
aI57
aI13
aI127
aI353
aI65
aI663
aI3849
aI3579
aI5521
aI11765
aI63427
aI76349
aI102517
asS'd'
I13948
sa(dp27896
S'a'
I8977
sS's'
I18
sS'm'
(lp27897
I1
aI1
aI7
aI11
aI27
aI55
aI79
aI249
aI397
aI77
aI1543
aI3787
aI4889
aI11145
aI18691
aI62899
aI66425
aI116195
asS'd'
I13949
sa(dp27898
S'a'
I9038
sS's'
I18
sS'm'
(lp27899
I1
aI1
aI1
aI3
aI5
aI3
aI1
aI143
aI73
aI999
aI2013
aI2001
aI4001
aI6563
aI30811
aI61445
aI2645
aI203631
asS'd'
I13950
sa(dp27900
S'a'
I9045
sS's'
I18
sS'm'
(lp27901
I1
aI1
aI1
aI15
aI1
aI49
aI35
aI61
aI493
aI101
aI1407
aI2211
aI7467
aI12321
aI15901
aI15479
aI62939
aI14643
asS'd'
I13951
sa(dp27902
S'a'
I9071
sS's'
I18
sS'm'
(lp27903
I1
aI1
aI3
aI11
aI21
aI33
aI123
aI95
aI449
aI355
aI1501
aI1627
aI1411
aI6183
aI17457
aI2199
aI96313
aI25023
asS'd'
I13952
sa(dp27904
S'a'
I9085
sS's'
I18
sS'm'
(lp27905
I1
aI1
aI5
aI5
aI13
aI49
aI73
aI203
aI83
aI3
aI137
aI119
aI3001
aI10685
aI18231
aI60727
aI31785
aI158605
asS'd'
I13953
sa(dp27906
S'a'
I9110
sS's'
I18
sS'm'
(lp27907
I1
aI3
aI1
aI11
aI23
aI19
aI123
aI9
aI269
aI501
aI2005
aI3695
aI3327
aI5353
aI12619
aI12987
aI18213
aI29355
asS'd'
I13954
sa(dp27908
S'a'
I9120
sS's'
I18
sS'm'
(lp27909
I1
aI3
aI1
aI5
aI1
aI25
aI99
aI197
aI327
aI575
aI773
aI2009
aI6653
aI1807
aI20381
aI55725
aI124359
aI176893
asS'd'
I13955
sa(dp27910
S'a'
I9157
sS's'
I18
sS'm'
(lp27911
I1
aI1
aI7
aI15
aI27
aI9
aI81
aI175
aI73
aI727
aI1907
aI1237
aI4983
aI16123
aI16479
aI2283
aI57805
aI13593
asS'd'
I13956
sa(dp27912
S'a'
I9164
sS's'
I18
sS'm'
(lp27913
I1
aI1
aI3
aI13
aI7
aI13
aI13
aI139
aI283
aI721
aI487
aI1821
aI4257
aI5105
aI8057
aI27193
aI46857
aI169927
asS'd'
I13957
sa(dp27914
S'a'
I9185
sS's'
I18
sS'm'
(lp27915
I1
aI1
aI5
aI5
aI29
aI5
aI81
aI211
aI441
aI685
aI981
aI3097
aI6253
aI10673
aI12253
aI54943
aI69401
aI147769
asS'd'
I13958
sa(dp27916
S'a'
I9203
sS's'
I18
sS'm'
(lp27917
I1
aI3
aI3
aI1
aI13
aI35
aI73
aI145
aI139
aI781
aI37
aI803
aI3607
aI4327
aI1153
aI11325
aI131025
aI168729
asS'd'
I13959
sa(dp27918
S'a'
I9235
sS's'
I18
sS'm'
(lp27919
I1
aI3
aI1
aI13
aI17
aI41
aI19
aI59
aI23
aI561
aI315
aI719
aI3325
aI275
aI12715
aI59843
aI16597
aI81691
asS'd'
I13960
sa(dp27920
S'a'
I9278
sS's'
I18
sS'm'
(lp27921
I1
aI3
aI1
aI11
aI1
aI53
aI11
aI237
aI363
aI345
aI331
aI129
aI6885
aI3105
aI12487
aI53803
aI8897
aI193777
asS'd'
I13961
sa(dp27922
S'a'
I9290
sS's'
I18
sS'm'
(lp27923
I1
aI3
aI7
aI15
aI3
aI53
aI55
aI101
aI389
aI839
aI413
aI2851
aI3989
aI12857
aI25723
aI16595
aI94145
aI193049
asS'd'
I13962
sa(dp27924
S'a'
I9292
sS's'
I18
sS'm'
(lp27925
I1
aI3
aI1
aI7
aI15
aI31
aI3
aI115
aI197
aI753
aI1035
aI1369
aI4925
aI4497
aI1641
aI63743
aI127089
aI114097
asS'd'
I13963
sa(dp27926
S'a'
I9319
sS's'
I18
sS'm'
(lp27927
I1
aI3
aI5
aI5
aI23
aI1
aI35
aI99
aI277
aI769
aI895
aI581
aI6969
aI15339
aI10309
aI27101
aI22611
aI86179
asS'd'
I13964
sa(dp27928
S'a'
I9334
sS's'
I18
sS'm'
(lp27929
I1
aI1
aI1
aI11
aI19
aI17
aI45
aI35
aI257
aI313
aI815
aI1469
aI3651
aI15101
aI22775
aI51729
aI75401
aI123653
asS'd'
I13965
sa(dp27930
S'a'
I9362
sS's'
I18
sS'm'
(lp27931
I1
aI3
aI1
aI15
aI5
aI11
aI83
aI141
aI373
aI935
aI1123
aI1849
aI1267
aI15427
aI10615
aI63303
aI109771
aI188601
asS'd'
I13966
sa(dp27932
S'a'
I9387
sS's'
I18
sS'm'
(lp27933
I1
aI3
aI5
aI3
aI29
aI23
aI79
aI193
aI261
aI29
aI1857
aI789
aI4359
aI14211
aI22181
aI64901
aI129089
aI65587
asS'd'
I13967
sa(dp27934
S'a'
I9404
sS's'
I18
sS'm'
(lp27935
I1
aI3
aI1
aI3
aI29
aI15
aI19
aI239
aI497
aI771
aI239
aI2853
aI2391
aI8153
aI31899
aI53759
aI127219
aI78833
asS'd'
I13968
sa(dp27936
S'a'
I9407
sS's'
I18
sS'm'
(lp27937
I1
aI1
aI7
aI7
aI5
aI57
aI9
aI93
aI69
aI993
aI193
aI3629
aI5761
aI9339
aI28073
aI50035
aI81635
aI83119
asS'd'
I13969
sa(dp27938
S'a'
I9410
sS's'
I18
sS'm'
(lp27939
I1
aI1
aI5
aI13
aI7
aI35
aI79
aI247
aI43
aI1011
aI1189
aI2881
aI1963
aI8889
aI9929
aI50043
aI112581
aI224139
asS'd'
I13970
sa(dp27940
S'a'
I9422
sS's'
I18
sS'm'
(lp27941
I1
aI3
aI3
aI7
aI15
aI63
aI85
aI33
aI107
aI37
aI45
aI1271
aI4735
aI1151
aI19793
aI6589
aI50875
aI185061
asS'd'
I13971
sa(dp27942
S'a'
I9478
sS's'
I18
sS'm'
(lp27943
I1
aI3
aI1
aI15
aI1
aI63
aI1
aI201
aI207
aI179
aI67
aI3703
aI2629
aI10517
aI1
aI39645
aI119733
aI6449
asS'd'
I13972
sa(dp27944
S'a'
I9512
sS's'
I18
sS'm'
(lp27945
I1
aI3
aI5
aI1
aI3
aI7
aI97
aI101
aI233
aI71
aI255
aI3767
aI8127
aI8041
aI25001
aI7601
aI129595
aI131657
asS'd'
I13973
sa(dp27946
S'a'
I9535
sS's'
I18
sS'm'
(lp27947
I1
aI1
aI7
aI1
aI25
aI29
aI105
aI25
aI267
aI191
aI267
aI3141
aI4445
aI5043
aI25203
aI32055
aI11035
aI229031
asS'd'
I13974
sa(dp27948
S'a'
I9604
sS's'
I18
sS'm'
(lp27949
I1
aI1
aI1
aI13
aI3
aI1
aI1
aI147
aI63
aI259
aI1171
aI401
aI6289
aI13577
aI28129
aI1349
aI85027
aI178123
asS'd'
I13975
sa(dp27950
S'a'
I9616
sS's'
I18
sS'm'
(lp27951
I1
aI1
aI1
aI13
aI1
aI59
aI109
aI95
aI49
aI309
aI1141
aI1355
aI3415
aI11237
aI21619
aI12039
aI1795
aI57775
asS'd'
I13976
sa(dp27952
S'a'
I9622
sS's'
I18
sS'm'
(lp27953
I1
aI3
aI1
aI11
aI19
aI3
aI51
aI227
aI277
aI49
aI703
aI2701
aI515
aI8893
aI20163
aI65297
aI114781
aI225687
asS'd'
I13977
sa(dp27954
S'a'
I9631
sS's'
I18
sS'm'
(lp27955
I1
aI3
aI7
aI11
aI19
aI47
aI121
aI199
aI173
aI905
aI1903
aI1781
aI2425
aI13381
aI25843
aI23279
aI87701
aI10723
asS'd'
I13978
sa(dp27956
S'a'
I9656
sS's'
I18
sS'm'
(lp27957
I1
aI3
aI1
aI13
aI7
aI21
aI17
aI15
aI85
aI241
aI119
aI2361
aI7921
aI6077
aI955
aI34221
aI78179
aI35511
asS'd'
I13979
sa(dp27958
S'a'
I9710
sS's'
I18
sS'm'
(lp27959
I1
aI1
aI7
aI11
aI9
aI1
aI1
aI29
aI445
aI557
aI241
aI959
aI6077
aI3547
aI30987
aI48129
aI79699
aI236611
asS'd'
I13980
sa(dp27960
S'a'
I9721
sS's'
I18
sS'm'
(lp27961
I1
aI3
aI1
aI15
aI13
aI29
aI57
aI117
aI347
aI719
aI1435
aI307
aI5209
aI4009
aI10517
aI3373
aI67667
aI260101
asS'd'
I13981
sa(dp27962
S'a'
I9728
sS's'
I18
sS'm'
(lp27963
I1
aI1
aI7
aI13
aI11
aI41
aI17
aI143
aI467
aI993
aI779
aI3991
aI623
aI8915
aI21615
aI56477
aI59721
aI164241
asS'd'
I13982
sa(dp27964
S'a'
I9733
sS's'
I18
sS'm'
(lp27965
I1
aI1
aI3
aI7
aI15
aI37
aI53
aI33
aI395
aI547
aI1815
aI2517
aI6575
aI14035
aI1
aI10919
aI25467
aI117521
asS'd'
I13983
sa(dp27966
S'a'
I9738
sS's'
I18
sS'm'
(lp27967
I1
aI1
aI3
aI9
aI17
aI47
aI45
aI3
aI509
aI53
aI1245
aI883
aI7917
aI15445
aI4169
aI49637
aI90933
aI109469
asS'd'
I13984
sa(dp27968
S'a'
I9774
sS's'
I18
sS'm'
(lp27969
I1
aI3
aI1
aI3
aI27
aI37
aI3
aI95
aI31
aI665
aI701
aI1979
aI3735
aI3257
aI18943
aI41201
aI95721
aI69451
asS'd'
I13985
sa(dp27970
S'a'
I9791
sS's'
I18
sS'm'
(lp27971
I1
aI1
aI1
aI15
aI19
aI49
aI61
aI5
aI115
aI801
aI805
aI2723
aI1387
aI13165
aI20717
aI40767
aI88857
aI28207
asS'd'
I13986
sa(dp27972
S'a'
I9803
sS's'
I18
sS'm'
(lp27973
I1
aI1
aI5
aI9
aI21
aI25
aI23
aI179
aI59
aI29
aI547
aI1829
aI4411
aI6689
aI22363
aI43975
aI52259
aI187563
asS'd'
I13987
sa(dp27974
S'a'
I9805
sS's'
I18
sS'm'
(lp27975
I1
aI1
aI5
aI11
aI13
aI31
aI97
aI131
aI135
aI415
aI53
aI4015
aI3629
aI6613
aI25541
aI47221
aI66483
aI224545
asS'd'
I13988
sa(dp27976
S'a'
I9817
sS's'
I18
sS'm'
(lp27977
I1
aI3
aI1
aI11
aI19
aI13
aI65
aI95
aI381
aI759
aI1319
aI2997
aI6321
aI9203
aI24483
aI9925
aI10799
aI117119
asS'd'
I13989
sa(dp27978
S'a'
I9823
sS's'
I18
sS'm'
(lp27979
I1
aI3
aI5
aI13
aI27
aI17
aI39
aI225
aI199
aI125
aI1125
aI2673
aI6787
aI8861
aI13139
aI13849
aI65459
aI40183
asS'd'
I13990
sa(dp27980
S'a'
I9839
sS's'
I18
sS'm'
(lp27981
I1
aI1
aI5
aI3
aI17
aI55
aI23
aI75
aI457
aI959
aI1507
aI1267
aI6857
aI16141
aI1889
aI10779
aI41331
aI166075
asS'd'
I13991
sa(dp27982
S'a'
I9847
sS's'
I18
sS'm'
(lp27983
I1
aI3
aI1
aI15
aI7
aI55
aI109
aI59
aI241
aI431
aI1281
aI183
aI1029
aI14617
aI4003
aI41871
aI36007
aI129617
asS'd'
I13992
sa(dp27984
S'a'
I9854
sS's'
I18
sS'm'
(lp27985
I1
aI3
aI1
aI1
aI27
aI61
aI79
aI93
aI217
aI251
aI671
aI989
aI7031
aI10035
aI15455
aI13685
aI95471
aI997
asS'd'
I13993
sa(dp27986
S'a'
I9863
sS's'
I18
sS'm'
(lp27987
I1
aI1
aI3
aI13
aI1
aI5
aI125
aI179
aI357
aI537
aI1303
aI2653
aI7319
aI2075
aI3861
aI11743
aI89659
aI221705
asS'd'
I13994
sa(dp27988
S'a'
I9872
sS's'
I18
sS'm'
(lp27989
I1
aI1
aI1
aI7
aI3
aI55
aI5
aI201
aI153
aI639
aI835
aI1913
aI3331
aI10727
aI30365
aI15133
aI67911
aI17851
asS'd'
I13995
sa(dp27990
S'a'
I9884
sS's'
I18
sS'm'
(lp27991
I1
aI1
aI3
aI13
aI21
aI1
aI67
aI71
aI265
aI43
aI279
aI2009
aI873
aI4447
aI32001
aI50783
aI76613
aI63919
asS'd'
I13996
sa(dp27992
S'a'
I9935
sS's'
I18
sS'm'
(lp27993
I1
aI1
aI3
aI11
aI17
aI43
aI19
aI195
aI233
aI17
aI1855
aI1227
aI3435
aI4313
aI6417
aI51019
aI130091
aI124947
asS'd'
I13997
sa(dp27994
S'a'
I9937
sS's'
I18
sS'm'
(lp27995
I1
aI1
aI7
aI9
aI19
aI9
aI95
aI87
aI297
aI817
aI1217
aI3637
aI2371
aI7073
aI387
aI62121
aI43507
aI93927
asS'd'
I13998
sa(dp27996
S'a'
I9974
sS's'
I18
sS'm'
(lp27997
I1
aI3
aI5
aI13
aI1
aI15
aI29
aI123
aI137
aI425
aI531
aI2659
aI2077
aI1345
aI2803
aI49469
aI29031
aI170825
asS'd'
I13999
sa(dp27998
S'a'
I9980
sS's'
I18
sS'm'
(lp27999
I1
aI1
aI5
aI7
aI15
aI13
aI119
aI231
aI139
aI673
aI1105
aI2355
aI3023
aI4437
aI17491
aI47367
aI12751
aI183319
asS'd'
I14000
sa(dp28000
S'a'
I10003
sS's'
I18
sS'm'
(lp28001
I1
aI1
aI5
aI15
aI19
aI5
aI125
aI121
aI509
aI539
aI473
aI2087
aI4421
aI4205
aI23457
aI34481
aI111231
aI145035
asS'd'
I14001
sa(dp28002
S'a'
I10015
sS's'
I18
sS'm'
(lp28003
I1
aI3
aI7
aI5
aI23
aI21
aI85
aI23
aI415
aI715
aI1579
aI3447
aI2373
aI233
aI19401
aI54869
aI15977
aI138119
asS'd'
I14002
sa(dp28004
S'a'
I10016
sS's'
I18
sS'm'
(lp28005
I1
aI1
aI3
aI11
aI21
aI1
aI37
aI127
aI101
aI943
aI79
aI2119
aI5679
aI10749
aI16209
aI16715
aI29421
aI259735
asS'd'
I14003
sa(dp28006
S'a'
I10066
sS's'
I18
sS'm'
(lp28007
I1
aI3
aI7
aI7
aI23
aI25
aI1
aI73
aI505
aI979
aI535
aI87
aI4165
aI9353
aI20075
aI57597
aI74651
aI22133
asS'd'
I14004
sa(dp28008
S'a'
I10093
sS's'
I18
sS'm'
(lp28009
I1
aI1
aI7
aI3
aI11
aI19
aI75
aI213
aI293
aI15
aI1981
aI1259
aI5455
aI2897
aI18861
aI6317
aI10339
aI123967
asS'd'
I14005
sa(dp28010
S'a'
I10118
sS's'
I18
sS'm'
(lp28011
I1
aI3
aI1
aI3
aI29
aI5
aI93
aI169
aI51
aI519
aI1649
aI2789
aI1251
aI8359
aI11489
aI62443
aI91549
aI148357
asS'd'
I14006
sa(dp28012
S'a'
I10132
sS's'
I18
sS'm'
(lp28013
I1
aI3
aI3
aI13
aI5
aI47
aI39
aI163
aI341
aI755
aI737
aI2335
aI2389
aI8351
aI26193
aI58111
aI18425
aI129313
asS'd'
I14007
sa(dp28014
S'a'
I10135
sS's'
I18
sS'm'
(lp28015
I1
aI1
aI3
aI1
aI31
aI49
aI101
aI69
aI345
aI291
aI1257
aI1801
aI1613
aI1479
aI4403
aI21307
aI44947
aI68591
asS'd'
I14008
sa(dp28016
S'a'
I10151
sS's'
I18
sS'm'
(lp28017
I1
aI3
aI3
aI9
aI5
aI23
aI65
aI65
aI187
aI709
aI883
aI2199
aI1037
aI8679
aI31527
aI23561
aI92225
aI254215
asS'd'
I14009
sa(dp28018
S'a'
I10158
sS's'
I18
sS'm'
(lp28019
I1
aI3
aI7
aI7
aI23
aI13
aI87
aI209
aI163
aI705
aI1199
aI3007
aI5469
aI2453
aI2691
aI17841
aI97045
aI174149
asS'd'
I14010
sa(dp28020
S'a'
I10169
sS's'
I18
sS'm'
(lp28021
I1
aI1
aI1
aI9
aI5
aI35
aI21
aI91
aI145
aI559
aI131
aI3911
aI1777
aI8225
aI6077
aI58223
aI100827
aI3641
asS'd'
I14011
sa(dp28022
S'a'
I10172
sS's'
I18
sS'm'
(lp28023
I1
aI1
aI5
aI5
aI7
aI5
aI31
aI189
aI117
aI785
aI1493
aI3899
aI471
aI10971
aI4607
aI21063
aI67225
aI195367
asS'd'
I14012
sa(dp28024
S'a'
I10180
sS's'
I18
sS'm'
(lp28025
I1
aI1
aI7
aI5
aI31
aI61
aI63
aI163
aI417
aI655
aI2033
aI1255
aI1139
aI6867
aI28655
aI55295
aI100519
aI166629
asS'd'
I14013
sa(dp28026
S'a'
I10187
sS's'
I18
sS'm'
(lp28027
I1
aI3
aI3
aI3
aI7
aI35
aI83
aI55
aI7
aI607
aI253
aI915
aI6801
aI7505
aI15929
aI16829
aI78469
aI150947
asS'd'
I14014
sa(dp28028
S'a'
I10189
sS's'
I18
sS'm'
(lp28029
I1
aI3
aI3
aI9
aI29
aI3
aI127
aI235
aI347
aI3
aI193
aI1547
aI8073
aI14963
aI20351
aI28951
aI53855
aI261375
asS'd'
I14015
sa(dp28030
S'a'
I10192
sS's'
I18
sS'm'
(lp28031
I1
aI1
aI7
aI3
aI31
aI19
aI75
aI87
aI23
aI419
aI75
aI1677
aI2371
aI8875
aI31993
aI4465
aI76085
aI86499
asS'd'
I14016
sa(dp28032
S'a'
I10197
sS's'
I18
sS'm'
(lp28033
I1
aI3
aI5
aI7
aI1
aI51
aI47
aI161
aI415
aI521
aI1099
aI1295
aI2545
aI15167
aI13983
aI7347
aI60631
aI4089
asS'd'
I14017
sa(dp28034
S'a'
I10211
sS's'
I18
sS'm'
(lp28035
I1
aI1
aI1
aI9
aI7
aI59
aI71
aI187
aI441
aI273
aI769
aI2649
aI3261
aI12661
aI23045
aI32035
aI104573
aI120589
asS'd'
I14018
sa(dp28036
S'a'
I10214
sS's'
I18
sS'm'
(lp28037
I1
aI3
aI7
aI13
aI23
aI51
aI113
aI205
aI443
aI291
aI475
aI2961
aI7615
aI105
aI22099
aI6045
aI22667
aI65515
asS'd'
I14019
sa(dp28038
S'a'
I10256
sS's'
I18
sS'm'
(lp28039
I1
aI1
aI5
aI11
aI1
aI1
aI23
aI231
aI413
aI371
aI1285
aI2695
aI2751
aI4235
aI15779
aI1903
aI24469
aI259157
asS'd'
I14020
sa(dp28040
S'a'
I10278
sS's'
I18
sS'm'
(lp28041
I1
aI3
aI3
aI7
aI7
aI47
aI87
aI105
aI311
aI251
aI573
aI3221
aI5757
aI11107
aI11161
aI8809
aI14467
aI33153
asS'd'
I14021
sa(dp28042
S'a'
I10290
sS's'
I18
sS'm'
(lp28043
I1
aI1
aI7
aI1
aI31
aI49
aI51
aI31
aI305
aI315
aI547
aI1159
aI2741
aI3773
aI13299
aI40115
aI62523
aI108487
asS'd'
I14022
sa(dp28044
S'a'
I10301
sS's'
I18
sS'm'
(lp28045
I1
aI3
aI1
aI11
aI11
aI43
aI33
aI213
aI107
aI467
aI1509
aI4081
aI4723
aI2409
aI1447
aI42759
aI64717
aI161991
asS'd'
I14023
sa(dp28046
S'a'
I10310
sS's'
I18
sS'm'
(lp28047
I1
aI3
aI7
aI3
aI31
aI23
aI25
aI159
aI95
aI721
aI1981
aI3659
aI4819
aI10119
aI25451
aI10165
aI31281
aI238319
asS'd'
I14024
sa(dp28048
S'a'
I10344
sS's'
I18
sS'm'
(lp28049
I1
aI3
aI1
aI5
aI19
aI27
aI67
aI125
aI481
aI585
aI43
aI3697
aI4997
aI581
aI6439
aI33477
aI115023
aI51759
asS'd'
I14025
sa(dp28050
S'a'
I10473
sS's'
I18
sS'm'
(lp28051
I1
aI3
aI7
aI7
aI15
aI53
aI89
aI173
aI369
aI365
aI91
aI1583
aI4611
aI7189
aI30383
aI47397
aI73657
aI158695
asS'd'
I14026
sa(dp28052
S'a'
I10474
sS's'
I18
sS'm'
(lp28053
I1
aI1
aI1
aI15
aI1
aI21
aI125
aI13
aI243
aI729
aI1397
aI2451
aI3233
aI15593
aI2815
aI56215
aI22685
aI167343
asS'd'
I14027
sa(dp28054
S'a'
I10482
sS's'
I18
sS'm'
(lp28055
I1
aI3
aI1
aI13
aI13
aI29
aI119
aI223
aI51
aI695
aI273
aI2381
aI4431
aI4891
aI29875
aI49511
aI111003
aI174413
asS'd'
I14028
sa(dp28056
S'a'
I10499
sS's'
I18
sS'm'
(lp28057
I1
aI3
aI3
aI15
aI9
aI13
aI19
aI177
aI371
aI957
aI255
aI115
aI4701
aI6089
aI7237
aI17077
aI87949
aI3111
asS'd'
I14029
sa(dp28058
S'a'
I10525
sS's'
I18
sS'm'
(lp28059
I1
aI1
aI7
aI11
aI29
aI49
aI59
aI201
aI145
aI219
aI1159
aI3863
aI715
aI10489
aI25883
aI56445
aI122103
aI149877
asS'd'
I14030
sa(dp28060
S'a'
I10532
sS's'
I18
sS'm'
(lp28061
I1
aI3
aI1
aI13
aI15
aI51
aI91
aI109
aI433
aI45
aI2045
aI3121
aI1109
aI14713
aI2667
aI40463
aI52185
aI64743
asS'd'
I14031
sa(dp28062
S'a'
I10535
sS's'
I18
sS'm'
(lp28063
I1
aI3
aI3
aI7
aI21
aI31
aI7
aI155
aI347
aI305
aI1557
aI1311
aI3315
aI1363
aI403
aI62063
aI114195
aI44623
asS'd'
I14032
sa(dp28064
S'a'
I10556
sS's'
I18
sS'm'
(lp28065
I1
aI3
aI5
aI11
aI9
aI11
aI11
aI245
aI49
aI21
aI239
aI3043
aI3525
aI1055
aI21891
aI19153
aI123689
aI170195
asS'd'
I14033
sa(dp28066
S'a'
I10568
sS's'
I18
sS'm'
(lp28067
I1
aI1
aI1
aI11
aI17
aI51
aI83
aI249
aI483
aI489
aI1063
aI469
aI6153
aI15551
aI13783
aI27945
aI103775
aI68175
asS'd'
I14034
sa(dp28068
S'a'
I10579
sS's'
I18
sS'm'
(lp28069
I1
aI3
aI3
aI13
aI11
aI61
aI107
aI113
aI503
aI819
aI1593
aI2851
aI6711
aI14623
aI3709
aI10931
aI8743
aI62321
asS'd'
I14035
sa(dp28070
S'a'
I10586
sS's'
I18
sS'm'
(lp28071
I1
aI1
aI5
aI15
aI5
aI37
aI23
aI131
aI329
aI499
aI1765
aI1273
aI819
aI11573
aI3307
aI46933
aI22087
aI173459
asS'd'
I14036
sa(dp28072
S'a'
I10609
sS's'
I18
sS'm'
(lp28073
I1
aI1
aI1
aI3
aI31
aI5
aI49
aI57
aI239
aI981
aI1863
aI3233
aI2727
aI7389
aI7923
aI63259
aI62873
aI113607
asS'd'
I14037
sa(dp28074
S'a'
I10612
sS's'
I18
sS'm'
(lp28075
I1
aI1
aI7
aI7
aI11
aI27
aI119
aI137
aI115
aI211
aI1239
aI2153
aI2579
aI11501
aI747
aI31141
aI129793
aI151589
asS'd'
I14038
sa(dp28076
S'a'
I10650
sS's'
I18
sS'm'
(lp28077
I1
aI3
aI3
aI9
aI9
aI55
aI121
aI199
aI91
aI835
aI521
aI2433
aI8123
aI2045
aI32553
aI48993
aI9935
aI220537
asS'd'
I14039
sa(dp28078
S'a'
I10665
sS's'
I18
sS'm'
(lp28079
I1
aI3
aI7
aI11
aI15
aI57
aI53
aI145
aI299
aI623
aI691
aI1557
aI785
aI15851
aI27075
aI5983
aI18043
aI22241
asS'd'
I14040
sa(dp28080
S'a'
I10685
sS's'
I18
sS'm'
(lp28081
I1
aI1
aI3
aI1
aI1
aI57
aI57
aI195
aI381
aI913
aI167
aI333
aI5541
aI2323
aI17001
aI34817
aI30795
aI144051
asS'd'
I14041
sa(dp28082
S'a'
I10686
sS's'
I18
sS'm'
(lp28083
I1
aI3
aI3
aI3
aI31
aI1
aI83
aI31
aI91
aI855
aI195
aI3449
aI8057
aI11061
aI29089
aI1597
aI127581
aI189033
asS'd'
I14042
sa(dp28084
S'a'
I10688
sS's'
I18
sS'm'
(lp28085
I1
aI3
aI1
aI1
aI21
aI59
aI113
aI179
aI13
aI523
aI629
aI3693
aI7155
aI893
aI17449
aI46535
aI18051
aI9191
asS'd'
I14043
sa(dp28086
S'a'
I10711
sS's'
I18
sS'm'
(lp28087
I1
aI1
aI1
aI3
aI27
aI19
aI75
aI229
aI181
aI653
aI1849
aI501
aI5871
aI14769
aI27461
aI59193
aI115013
aI72227
asS'd'
I14044
sa(dp28088
S'a'
I10728
sS's'
I18
sS'm'
(lp28089
I1
aI3
aI3
aI13
aI17
aI41
aI111
aI107
aI453
aI299
aI1699
aI2871
aI2955
aI4215
aI13919
aI19785
aI30339
aI148445
asS'd'
I14045
sa(dp28090
S'a'
I10748
sS's'
I18
sS'm'
(lp28091
I1
aI3
aI3
aI11
aI5
aI19
aI21
aI87
aI173
aI439
aI1651
aI2393
aI4137
aI16285
aI16093
aI22953
aI105663
aI226575
asS'd'
I14046
sa(dp28092
S'a'
I10751
sS's'
I18
sS'm'
(lp28093
I1
aI3
aI5
aI7
aI19
aI61
aI101
aI251
aI295
aI89
aI1695
aI1359
aI5797
aI8587
aI18753
aI65223
aI51079
aI96169
asS'd'
I14047
sa(dp28094
S'a'
I10781
sS's'
I18
sS'm'
(lp28095
I1
aI1
aI1
aI5
aI3
aI1
aI79
aI63
aI221
aI601
aI1385
aI1963
aI4601
aI15217
aI4861
aI58295
aI61043
aI88523
asS'd'
I14048
sa(dp28096
S'a'
I10805
sS's'
I18
sS'm'
(lp28097
I1
aI3
aI3
aI7
aI31
aI63
aI73
aI177
aI455
aI487
aI1009
aI2103
aI4753
aI3143
aI10121
aI36509
aI24753
aI230869
asS'd'
I14049
sa(dp28098
S'a'
I10823
sS's'
I18
sS'm'
(lp28099
I1
aI1
aI5
aI1
aI17
aI27
aI103
aI63
aI475
aI665
aI1189
aI3513
aI89
aI2669
aI1227
aI20635
aI121549
aI248851
asS'd'
I14050
sa(dp28100
S'a'
I10838
sS's'
I18
sS'm'
(lp28101
I1
aI1
aI3
aI7
aI25
aI19
aI117
aI243
aI337
aI207
aI903
aI3751
aI3309
aI11955
aI12651
aI25359
aI83419
aI19701
asS'd'
I14051
sa(dp28102
S'a'
I10857
sS's'
I18
sS'm'
(lp28103
I1
aI1
aI7
aI15
aI19
aI21
aI3
aI235
aI289
aI185
aI1175
aI2291
aI4003
aI7753
aI4775
aI65321
aI48957
aI220261
asS'd'
I14052
sa(dp28104
S'a'
I10865
sS's'
I18
sS'm'
(lp28105
I1
aI3
aI5
aI11
aI23
aI21
aI107
aI65
aI117
aI329
aI1085
aI3555
aI1183
aI15241
aI32663
aI50985
aI66753
aI38023
asS'd'
I14053
sa(dp28106
S'a'
I10872
sS's'
I18
sS'm'
(lp28107
I1
aI1
aI7
aI9
aI11
aI49
aI65
aI17
aI291
aI435
aI1221
aI3829
aI5467
aI5181
aI19891
aI7091
aI80673
aI90495
asS'd'
I14054
sa(dp28108
S'a'
I10893
sS's'
I18
sS'm'
(lp28109
I1
aI1
aI1
aI15
aI17
aI47
aI119
aI173
aI297
aI477
aI859
aI3661
aI8081
aI8257
aI20841
aI55123
aI11231
aI193669
asS'd'
I14055
sa(dp28110
S'a'
I10899
sS's'
I18
sS'm'
(lp28111
I1
aI1
aI7
aI7
aI27
aI11
aI119
aI109
aI199
aI727
aI1569
aI3749
aI4067
aI11675
aI30213
aI58091
aI64303
aI92785
asS'd'
I14056
sa(dp28112
S'a'
I10917
sS's'
I18
sS'm'
(lp28113
I1
aI3
aI7
aI15
aI15
aI39
aI101
aI149
aI299
aI449
aI1017
aI723
aI7731
aI7929
aI22465
aI61583
aI69851
aI150507
asS'd'
I14057
sa(dp28114
S'a'
I10935
sS's'
I18
sS'm'
(lp28115
I1
aI3
aI5
aI15
aI5
aI13
aI97
aI127
aI21
aI673
aI353
aI3885
aI5761
aI11443
aI10089
aI23701
aI85879
aI42217
asS'd'
I14058
sa(dp28116
S'a'
I11001
sS's'
I18
sS'm'
(lp28117
I1
aI3
aI1
aI5
aI27
aI55
aI31
aI167
aI69
aI453
aI925
aI555
aI5135
aI2759
aI27077
aI14497
aI94333
aI108729
asS'd'
I14059
sa(dp28118
S'a'
I11072
sS's'
I18
sS'm'
(lp28119
I1
aI1
aI7
aI15
aI11
aI55
aI9
aI241
aI55
aI611
aI149
aI2605
aI653
aI1631
aI15059
aI6349
aI12321
aI124561
asS'd'
I14060
sa(dp28120
S'a'
I11089
sS's'
I18
sS'm'
(lp28121
I1
aI1
aI1
aI9
aI3
aI11
aI95
aI67
aI443
aI103
aI1687
aI2667
aI4567
aI4271
aI15601
aI27859
aI4757
aI53289
asS'd'
I14061
sa(dp28122
S'a'
I11090
sS's'
I18
sS'm'
(lp28123
I1
aI1
aI5
aI9
aI23
aI21
aI1
aI125
aI105
aI975
aI1879
aI1821
aI5273
aI7079
aI25009
aI10471
aI29119
aI73249
asS'd'
I14062
sa(dp28124
S'a'
I11108
sS's'
I18
sS'm'
(lp28125
I1
aI1
aI7
aI1
aI31
aI61
aI17
aI23
aI485
aI565
aI1325
aI1559
aI4131
aI751
aI2071
aI4719
aI15925
aI101207
asS'd'
I14063
sa(dp28126
S'a'
I11117
sS's'
I18
sS'm'
(lp28127
I1
aI1
aI5
aI5
aI13
aI53
aI13
aI93
aI149
aI139
aI1429
aI3605
aI3545
aI11193
aI14139
aI6093
aI115727
aI183105
asS'd'
I14064
sa(dp28128
S'a'
I11136
sS's'
I18
sS'm'
(lp28129
I1
aI1
aI1
aI7
aI15
aI37
aI51
aI77
aI177
aI967
aI405
aI563
aI3047
aI8499
aI26787
aI27609
aI23613
aI239679
asS'd'
I14065
sa(dp28130
S'a'
I11148
sS's'
I18
sS'm'
(lp28131
I1
aI1
aI1
aI5
aI27
aI37
aI1
aI129
aI197
aI133
aI1329
aI3673
aI3143
aI1059
aI19209
aI39027
aI43787
aI42821
asS'd'
I14066
sa(dp28132
S'a'
I11153
sS's'
I18
sS'm'
(lp28133
I1
aI3
aI7
aI5
aI5
aI47
aI105
aI121
aI219
aI777
aI1569
aI1359
aI1955
aI13207
aI14895
aI7829
aI40499
aI182911
asS'd'
I14067
sa(dp28134
S'a'
I11181
sS's'
I18
sS'm'
(lp28135
I1
aI3
aI5
aI7
aI11
aI41
aI41
aI155
aI245
aI383
aI405
aI2415
aI5809
aI5117
aI31523
aI16927
aI76785
aI113731
asS'd'
I14068
sa(dp28136
S'a'
I11190
sS's'
I18
sS'm'
(lp28137
I1
aI3
aI3
aI9
aI9
aI21
aI13
aI197
aI409
aI931
aI305
aI1129
aI865
aI12961
aI5239
aI35823
aI82565
aI226765
asS'd'
I14069
sa(dp28138
S'a'
I11201
sS's'
I18
sS'm'
(lp28139
I1
aI1
aI5
aI3
aI17
aI27
aI59
aI79
aI359
aI601
aI979
aI1355
aI1657
aI10479
aI4741
aI36391
aI111527
aI105139
asS'd'
I14070
sa(dp28140
S'a'
I11204
sS's'
I18
sS'm'
(lp28141
I1
aI3
aI5
aI9
aI13
aI43
aI31
aI1
aI309
aI723
aI1049
aI803
aI1653
aI2551
aI26317
aI49731
aI67799
aI129225
asS'd'
I14071
sa(dp28142
S'a'
I11216
sS's'
I18
sS'm'
(lp28143
I1
aI1
aI5
aI3
aI1
aI39
aI95
aI243
aI499
aI809
aI1515
aI981
aI585
aI7907
aI16801
aI43381
aI117537
aI99787
asS'd'
I14072
sa(dp28144
S'a'
I11237
sS's'
I18
sS'm'
(lp28145
I1
aI1
aI5
aI5
aI25
aI23
aI15
aI127
aI33
aI799
aI647
aI2923
aI7805
aI2681
aI14773
aI42751
aI106861
aI119657
asS'd'
I14073
sa(dp28146
S'a'
I11259
sS's'
I18
sS'm'
(lp28147
I1
aI3
aI1
aI1
aI1
aI47
aI11
aI179
aI179
aI659
aI1061
aI2511
aI3601
aI7107
aI27887
aI48427
aI40559
aI106043
asS'd'
I14074
sa(dp28148
S'a'
I11279
sS's'
I18
sS'm'
(lp28149
I1
aI1
aI7
aI11
aI5
aI33
aI115
aI195
aI431
aI383
aI1571
aI3485
aI5741
aI5775
aI14891
aI26389
aI71723
aI198861
asS'd'
I14075
sa(dp28150
S'a'
I11282
sS's'
I18
sS'm'
(lp28151
I1
aI1
aI5
aI5
aI11
aI55
aI37
aI57
aI381
aI607
aI2017
aI1981
aI6113
aI3771
aI8827
aI13335
aI88587
aI102791
asS'd'
I14076
sa(dp28152
S'a'
I11304
sS's'
I18
sS'm'
(lp28153
I1
aI1
aI1
aI11
aI29
aI23
aI73
aI149
aI405
aI581
aI721
aI281
aI5315
aI4675
aI13013
aI39003
aI20335
aI109855
asS'd'
I14077
sa(dp28154
S'a'
I11312
sS's'
I18
sS'm'
(lp28155
I1
aI1
aI7
aI15
aI17
aI57
aI39
aI51
aI403
aI979
aI1543
aI1235
aI797
aI5949
aI26647
aI15125
aI33255
aI152861
asS'd'
I14078
sa(dp28156
S'a'
I11332
sS's'
I18
sS'm'
(lp28157
I1
aI1
aI5
aI3
aI25
aI27
aI7
aI147
aI257
aI163
aI1297
aI2289
aI693
aI7771
aI6341
aI22323
aI1653
aI177669
asS'd'
I14079
sa(dp28158
S'a'
I11341
sS's'
I18
sS'm'
(lp28159
I1
aI1
aI3
aI9
aI1
aI39
aI47
aI231
aI15
aI705
aI897
aI3943
aI6281
aI6679
aI21695
aI29553
aI39509
aI83135
asS'd'
I14080
sa(dp28160
S'a'
I11354
sS's'
I18
sS'm'
(lp28161
I1
aI3
aI3
aI15
aI17
aI43
aI15
aI195
aI31
aI501
aI529
aI3117
aI6031
aI12101
aI30687
aI52465
aI66171
aI149591
asS'd'
I14081
sa(dp28162
S'a'
I11356
sS's'
I18
sS'm'
(lp28163
I1
aI1
aI3
aI5
aI13
aI17
aI63
aI41
aI303
aI671
aI1225
aI1761
aI6159
aI3203
aI23611
aI18309
aI115027
aI116325
asS'd'
I14082
sa(dp28164
S'a'
I11365
sS's'
I18
sS'm'
(lp28165
I1
aI1
aI5
aI13
aI17
aI5
aI97
aI155
aI479
aI525
aI1403
aI4063
aI8167
aI6443
aI20627
aI41399
aI26897
aI102841
asS'd'
I14083
sa(dp28166
S'a'
I11377
sS's'
I18
sS'm'
(lp28167
I1
aI3
aI5
aI9
aI9
aI27
aI59
aI177
aI453
aI659
aI765
aI431
aI4209
aI12679
aI10719
aI22473
aI81597
aI20057
asS'd'
I14084
sa(dp28168
S'a'
I11417
sS's'
I18
sS'm'
(lp28169
I1
aI3
aI3
aI1
aI3
aI37
aI91
aI97
aI159
aI845
aI519
aI2603
aI6979
aI6711
aI29781
aI53639
aI103357
aI111671
asS'd'
I14085
sa(dp28170
S'a'
I11427
sS's'
I18
sS'm'
(lp28171
I1
aI1
aI3
aI3
aI25
aI15
aI27
aI9
aI503
aI719
aI153
aI3071
aI281
aI5341
aI32595
aI13069
aI6461
aI160319
asS'd'
I14086
sa(dp28172
S'a'
I11433
sS's'
I18
sS'm'
(lp28173
I1
aI3
aI3
aI5
aI29
aI7
aI119
aI229
aI117
aI925
aI465
aI1703
aI7277
aI10631
aI9429
aI41011
aI45181
aI229239
asS'd'
I14087
sa(dp28174
S'a'
I11434
sS's'
I18
sS'm'
(lp28175
I1
aI1
aI7
aI7
aI31
aI63
aI67
aI55
aI445
aI39
aI1363
aI1369
aI1061
aI8555
aI29263
aI47341
aI49563
aI80445
asS'd'
I14088
sa(dp28176
S'a'
I11439
sS's'
I18
sS'm'
(lp28177
I1
aI1
aI7
aI1
aI23
aI23
aI49
aI205
aI371
aI101
aI1963
aI2763
aI3475
aI835
aI20371
aI51343
aI9771
aI69713
asS'd'
I14089
sa(dp28178
S'a'
I11444
sS's'
I18
sS'm'
(lp28179
I1
aI3
aI3
aI7
aI3
aI29
aI7
aI185
aI511
aI93
aI1077
aI3971
aI2981
aI16367
aI12703
aI36179
aI47755
aI42767
asS'd'
I14090
sa(dp28180
S'a'
I11448
sS's'
I18
sS'm'
(lp28181
I1
aI1
aI1
aI11
aI27
aI47
aI43
aI39
aI129
aI337
aI1249
aI3557
aI2871
aI13565
aI19525
aI46263
aI49203
aI148235
asS'd'
I14091
sa(dp28182
S'a'
I11461
sS's'
I18
sS'm'
(lp28183
I1
aI3
aI5
aI11
aI19
aI3
aI83
aI151
aI425
aI199
aI847
aI3751
aI1729
aI12457
aI21819
aI295
aI53627
aI17555
asS'd'
I14092
sa(dp28184
S'a'
I11466
sS's'
I18
sS'm'
(lp28185
I1
aI1
aI3
aI5
aI7
aI43
aI21
aI221
aI93
aI785
aI1851
aI3891
aI2103
aI5219
aI31845
aI58943
aI42461
aI160149
asS'd'
I14093
sa(dp28186
S'a'
I11468
sS's'
I18
sS'm'
(lp28187
I1
aI3
aI5
aI9
aI25
aI11
aI43
aI171
aI445
aI335
aI1907
aI3401
aI815
aI10341
aI17779
aI24895
aI7727
aI168143
asS'd'
I14094
sa(dp28188
S'a'
I11473
sS's'
I18
sS'm'
(lp28189
I1
aI3
aI1
aI5
aI27
aI25
aI41
aI13
aI239
aI233
aI1861
aI3409
aI4325
aI2227
aI30197
aI59329
aI48501
aI168799
asS'd'
I14095
sa(dp28190
S'a'
I11489
sS's'
I18
sS'm'
(lp28191
I1
aI3
aI5
aI13
aI9
aI55
aI83
aI185
aI287
aI83
aI1545
aI2803
aI2177
aI6195
aI14455
aI30541
aI75731
aI98915
asS'd'
I14096
sa(dp28192
S'a'
I11499
sS's'
I18
sS'm'
(lp28193
I1
aI3
aI3
aI5
aI25
aI19
aI5
aI203
aI303
aI703
aI1861
aI3867
aI2683
aI8223
aI11107
aI54785
aI106053
aI135543
asS'd'
I14097
sa(dp28194
S'a'
I11516
sS's'
I18
sS'm'
(lp28195
I1
aI1
aI1
aI13
aI19
aI7
aI11
aI197
aI303
aI541
aI977
aI2083
aI4739
aI7971
aI2245
aI11029
aI77333
aI16573
asS'd'
I14098
sa(dp28196
S'a'
I11531
sS's'
I18
sS'm'
(lp28197
I1
aI1
aI1
aI3
aI11
aI33
aI77
aI59
aI283
aI791
aI365
aI4027
aI487
aI10559
aI4543
aI58111
aI91861
aI102905
asS'd'
I14099
sa(dp28198
S'a'
I11533
sS's'
I18
sS'm'
(lp28199
I1
aI3
aI7
aI7
aI15
aI3
aI19
aI51
aI339
aI377
aI929
aI693
aI1617
aI14057
aI7107
aI27181
aI7411
aI202843
asS'd'
I14100
sa(dp28200
S'a'
I11557
sS's'
I18
sS'm'
(lp28201
I1
aI3
aI3
aI9
aI19
aI9
aI73
aI109
aI333
aI917
aI1227
aI2871
aI4893
aI11029
aI5619
aI27091
aI9381
aI213403
asS'd'
I14101
sa(dp28202
S'a'
I11572
sS's'
I18
sS'm'
(lp28203
I1
aI1
aI1
aI9
aI9
aI13
aI77
aI131
aI163
aI619
aI169
aI315
aI1277
aI13705
aI16853
aI1179
aI86433
aI135427
asS'd'
I14102
sa(dp28204
S'a'
I11608
sS's'
I18
sS'm'
(lp28205
I1
aI3
aI5
aI9
aI15
aI47
aI57
aI119
aI325
aI529
aI893
aI2395
aI5159
aI5481
aI18689
aI6457
aI114733
aI159999
asS'd'
I14103
sa(dp28206
S'a'
I11635
sS's'
I18
sS'm'
(lp28207
I1
aI3
aI7
aI5
aI15
aI9
aI113
aI235
aI475
aI93
aI495
aI2983
aI2769
aI5209
aI7481
aI49699
aI46961
aI246393
asS'd'
I14104
sa(dp28208
S'a'
I11638
sS's'
I18
sS'm'
(lp28209
I1
aI1
aI7
aI1
aI5
aI31
aI113
aI27
aI359
aI635
aI955
aI2795
aI6289
aI11621
aI11059
aI2259
aI57443
aI243143
asS'd'
I14105
sa(dp28210
S'a'
I11644
sS's'
I18
sS'm'
(lp28211
I1
aI3
aI3
aI13
aI19
aI33
aI53
aI141
aI437
aI415
aI919
aI1375
aI2703
aI13731
aI31559
aI14115
aI50101
aI85199
asS'd'
I14106
sa(dp28212
S'a'
I11684
sS's'
I18
sS'm'
(lp28213
I1
aI1
aI7
aI13
aI27
aI57
aI111
aI89
aI89
aI313
aI1107
aI4049
aI2485
aI269
aI10197
aI36995
aI71381
aI112795
asS'd'
I14107
sa(dp28214
S'a'
I11702
sS's'
I18
sS'm'
(lp28215
I1
aI1
aI3
aI7
aI17
aI23
aI119
aI123
aI145
aI213
aI1273
aI1707
aI4005
aI13815
aI23495
aI36359
aI14391
aI94287
asS'd'
I14108
sa(dp28216
S'a'
I11740
sS's'
I18
sS'm'
(lp28217
I1
aI1
aI7
aI1
aI5
aI49
aI81
aI193
aI105
aI1003
aI413
aI2975
aI1725
aI5647
aI25447
aI43501
aI4431
aI115489
asS'd'
I14109
sa(dp28218
S'a'
I11749
sS's'
I18
sS'm'
(lp28219
I1
aI3
aI1
aI5
aI29
aI13
aI47
aI37
aI441
aI955
aI611
aI853
aI7225
aI4959
aI8739
aI31703
aI48095
aI124085
asS'd'
I14110
sa(dp28220
S'a'
I11754
sS's'
I18
sS'm'
(lp28221
I1
aI1
aI5
aI15
aI31
aI9
aI125
aI53
aI229
aI631
aI1031
aI3923
aI4417
aI12637
aI22093
aI46985
aI103417
aI193443
asS'd'
I14111
sa(dp28222
S'a'
I11764
sS's'
I18
sS'm'
(lp28223
I1
aI1
aI7
aI1
aI7
aI9
aI77
aI11
aI451
aI615
aI1259
aI3097
aI1513
aI13641
aI26845
aI17399
aI63661
aI9231
asS'd'
I14112
sa(dp28224
S'a'
I11767
sS's'
I18
sS'm'
(lp28225
I1
aI1
aI7
aI13
aI25
aI47
aI125
aI1
aI333
aI599
aI1133
aI3527
aI7451
aI2849
aI27227
aI40015
aI118185
aI24737
asS'd'
I14113
sa(dp28226
S'a'
I11784
sS's'
I18
sS'm'
(lp28227
I1
aI1
aI5
aI5
aI31
aI15
aI85
aI37
aI121
aI677
aI593
aI2757
aI739
aI839
aI3939
aI36339
aI116663
aI955
asS'd'
I14114
sa(dp28228
S'a'
I11789
sS's'
I18
sS'm'
(lp28229
I1
aI3
aI1
aI11
aI19
aI13
aI87
aI109
aI149
aI215
aI1811
aI3813
aI7699
aI16189
aI12841
aI52081
aI104545
aI245819
asS'd'
I14115
sa(dp28230
S'a'
I11795
sS's'
I18
sS'm'
(lp28231
I1
aI1
aI7
aI3
aI31
aI17
aI99
aI23
aI377
aI131
aI821
aI1167
aI4437
aI15727
aI20753
aI8163
aI43719
aI7243
asS'd'
I14116
sa(dp28232
S'a'
I11798
sS's'
I18
sS'm'
(lp28233
I1
aI3
aI7
aI13
aI21
aI5
aI5
aI167
aI9
aI1009
aI1013
aI797
aI6145
aI2855
aI19969
aI59887
aI3419
aI238661
asS'd'
I14117
sa(dp28234
S'a'
I11804
sS's'
I18
sS'm'
(lp28235
I1
aI1
aI7
aI1
aI5
aI39
aI47
aI91
aI185
aI139
aI959
aI3149
aI3423
aI8909
aI2045
aI18187
aI71935
aI238605
asS'd'
I14118
sa(dp28236
S'a'
I11818
sS's'
I18
sS'm'
(lp28237
I1
aI3
aI5
aI11
aI29
aI63
aI105
aI43
aI27
aI221
aI879
aI181
aI1499
aI10343
aI27135
aI823
aI4893
aI101707
asS'd'
I14119
sa(dp28238
S'a'
I11820
sS's'
I18
sS'm'
(lp28239
I1
aI3
aI5
aI11
aI5
aI13
aI59
aI83
aI315
aI999
aI1205
aI939
aI3661
aI3081
aI15551
aI13791
aI49027
aI26843
asS'd'
I14120
sa(dp28240
S'a'
I11860
sS's'
I18
sS'm'
(lp28241
I1
aI1
aI1
aI5
aI3
aI57
aI105
aI169
aI123
aI463
aI1471
aI445
aI743
aI13353
aI17661
aI23437
aI35451
aI115919
asS'd'
I14121
sa(dp28242
S'a'
I11869
sS's'
I18
sS'm'
(lp28243
I1
aI1
aI5
aI11
aI9
aI3
aI41
aI63
aI501
aI861
aI153
aI1591
aI1379
aI5189
aI24483
aI8073
aI43319
aI248959
asS'd'
I14122
sa(dp28244
S'a'
I11874
sS's'
I18
sS'm'
(lp28245
I1
aI1
aI7
aI3
aI29
aI45
aI51
aI177
aI1
aI961
aI1493
aI2179
aI3723
aI1923
aI1517
aI44823
aI81613
aI194641
asS'd'
I14123
sa(dp28246
S'a'
I11903
sS's'
I18
sS'm'
(lp28247
I1
aI1
aI5
aI11
aI17
aI17
aI61
aI141
aI5
aI529
aI379
aI2509
aI1487
aI13141
aI10877
aI18603
aI40569
aI69639
asS'd'
I14124
sa(dp28248
S'a'
I11916
sS's'
I18
sS'm'
(lp28249
I1
aI1
aI5
aI15
aI1
aI15
aI33
aI219
aI269
aI557
aI7
aI3627
aI183
aI6975
aI4627
aI15235
aI51863
aI172393
asS'd'
I14125
sa(dp28250
S'a'
I11927
sS's'
I18
sS'm'
(lp28251
I1
aI3
aI7
aI9
aI1
aI37
aI13
aI75
aI151
aI153
aI1693
aI2835
aI3093
aI8847
aI6721
aI44135
aI128931
aI230745
asS'd'
I14126
sa(dp28252
S'a'
I11933
sS's'
I18
sS'm'
(lp28253
I1
aI1
aI3
aI13
aI29
aI63
aI33
aI153
aI503
aI137
aI401
aI2315
aI2223
aI10843
aI4235
aI37295
aI103249
aI183899
asS'd'
I14127
sa(dp28254
S'a'
I11962
sS's'
I18
sS'm'
(lp28255
I1
aI1
aI7
aI11
aI15
aI25
aI49
aI55
aI39
aI13
aI269
aI3119
aI3445
aI8265
aI16781
aI57239
aI97489
aI204841
asS'd'
I14128
sa(dp28256
S'a'
I12009
sS's'
I18
sS'm'
(lp28257
I1
aI1
aI1
aI1
aI25
aI57
aI117
aI199
aI41
aI351
aI477
aI1891
aI7913
aI14439
aI25305
aI64811
aI57731
aI184265
asS'd'
I14129
sa(dp28258
S'a'
I12020
sS's'
I18
sS'm'
(lp28259
I1
aI3
aI3
aI1
aI13
aI41
aI33
aI53
aI381
aI31
aI1861
aI2207
aI1497
aI15539
aI23589
aI53215
aI36887
aI134007
asS'd'
I14130
sa(dp28260
S'a'
I12035
sS's'
I18
sS'm'
(lp28261
I1
aI1
aI5
aI7
aI15
aI37
aI13
aI99
aI17
aI325
aI643
aI2943
aI7967
aI11531
aI21301
aI5125
aI63201
aI101203
asS'd'
I14131
sa(dp28262
S'a'
I12041
sS's'
I18
sS'm'
(lp28263
I1
aI1
aI7
aI11
aI23
aI21
aI119
aI151
aI457
aI929
aI1917
aI3123
aI1133
aI11861
aI27889
aI40421
aI90949
aI113237
asS'd'
I14132
sa(dp28264
S'a'
I12049
sS's'
I18
sS'm'
(lp28265
I1
aI3
aI5
aI9
aI13
aI35
aI111
aI83
aI371
aI589
aI1507
aI3559
aI773
aI5895
aI31453
aI40865
aI124103
aI250473
asS'd'
I14133
sa(dp28266
S'a'
I12065
sS's'
I18
sS'm'
(lp28267
I1
aI3
aI3
aI15
aI11
aI7
aI93
aI163
aI285
aI763
aI2023
aI1047
aI3349
aI13575
aI22571
aI21513
aI56081
aI204765
asS'd'
I14134
sa(dp28268
S'a'
I12072
sS's'
I18
sS'm'
(lp28269
I1
aI3
aI3
aI5
aI19
aI19
aI47
aI25
aI49
aI717
aI1155
aI3901
aI407
aI2699
aI30961
aI55647
aI96043
aI185559
asS'd'
I14135
sa(dp28270
S'a'
I12098
sS's'
I18
sS'm'
(lp28271
I1
aI1
aI1
aI7
aI29
aI1
aI49
aI87
aI311
aI435
aI1235
aI1041
aI6595
aI1639
aI32495
aI44245
aI6593
aI236331
asS'd'
I14136
sa(dp28272
S'a'
I12100
sS's'
I18
sS'm'
(lp28273
I1
aI3
aI7
aI9
aI27
aI13
aI1
aI41
aI75
aI953
aI1635
aI101
aI7231
aI13019
aI14773
aI17315
aI120993
aI111215
asS'd'
I14137
sa(dp28274
S'a'
I12107
sS's'
I18
sS'm'
(lp28275
I1
aI1
aI5
aI9
aI23
aI31
aI87
aI47
aI83
aI791
aI1239
aI1453
aI5459
aI4847
aI7285
aI32667
aI45991
aI103593
asS'd'
I14138
sa(dp28276
S'a'
I12167
sS's'
I18
sS'm'
(lp28277
I1
aI3
aI1
aI3
aI27
aI47
aI97
aI191
aI5
aI961
aI1191
aI3897
aI6821
aI4257
aI22047
aI31557
aI52603
aI251405
asS'd'
I14139
sa(dp28278
S'a'
I12202
sS's'
I18
sS'm'
(lp28279
I1
aI1
aI5
aI3
aI23
aI45
aI103
aI115
aI287
aI47
aI93
aI2761
aI6467
aI14031
aI21881
aI31631
aI47605
aI237635
asS'd'
I14140
sa(dp28280
S'a'
I12207
sS's'
I18
sS'm'
(lp28281
I1
aI3
aI7
aI15
aI23
aI41
aI63
aI239
aI115
aI655
aI1949
aI1969
aI3145
aI91
aI16735
aI49295
aI117995
aI40537
asS'd'
I14141
sa(dp28282
S'a'
I12233
sS's'
I18
sS'm'
(lp28283
I1
aI1
aI7
aI11
aI5
aI25
aI99
aI247
aI11
aI707
aI1497
aI785
aI6055
aI8521
aI12179
aI56363
aI110131
aI55449
asS'd'
I14142
sa(dp28284
S'a'
I12269
sS's'
I18
sS'm'
(lp28285
I1
aI1
aI1
aI7
aI5
aI31
aI99
aI7
aI285
aI281
aI1207
aI1173
aI7637
aI9595
aI31413
aI16597
aI96157
aI39059
asS'd'
I14143
sa(dp28286
S'a'
I12290
sS's'
I18
sS'm'
(lp28287
I1
aI3
aI7
aI7
aI3
aI49
aI79
aI69
aI57
aI523
aI65
aI2785
aI2907
aI11295
aI16199
aI25845
aI51801
aI67417
asS'd'
I14144
sa(dp28288
S'a'
I12304
sS's'
I18
sS'm'
(lp28289
I1
aI1
aI3
aI1
aI1
aI53
aI57
aI117
aI1
aI927
aI1787
aI3059
aI7441
aI14663
aI10881
aI2225
aI29375
aI93717
asS'd'
I14145
sa(dp28290
S'a'
I12314
sS's'
I18
sS'm'
(lp28291
I1
aI1
aI5
aI5
aI29
aI17
aI67
aI35
aI475
aI367
aI155
aI3463
aI2339
aI6239
aI31073
aI56169
aI130309
aI28981
asS'd'
I14146
sa(dp28292
S'a'
I12340
sS's'
I18
sS'm'
(lp28293
I1
aI1
aI1
aI5
aI7
aI53
aI61
aI105
aI355
aI817
aI869
aI2863
aI6041
aI11459
aI4151
aI61115
aI100689
aI32917
asS'd'
I14147
sa(dp28294
S'a'
I12416
sS's'
I18
sS'm'
(lp28295
I1
aI3
aI5
aI11
aI31
aI11
aI33
aI105
aI437
aI767
aI101
aI2979
aI3911
aI4859
aI15551
aI23545
aI10705
aI6271
asS'd'
I14148
sa(dp28296
S'a'
I12425
sS's'
I18
sS'm'
(lp28297
I1
aI1
aI7
aI5
aI1
aI17
aI109
aI205
aI409
aI893
aI889
aI2181
aI6167
aI14273
aI25389
aI50279
aI5497
aI191755
asS'd'
I14149
sa(dp28298
S'a'
I12428
sS's'
I18
sS'm'
(lp28299
I1
aI3
aI7
aI15
aI29
aI11
aI79
aI101
aI123
aI399
aI1159
aI1263
aI3513
aI13169
aI2199
aI41057
aI98639
aI227107
asS'd'
I14150
sa(dp28300
S'a'
I12446
sS's'
I18
sS'm'
(lp28301
I1
aI1
aI3
aI11
aI13
aI31
aI51
aI119
aI257
aI829
aI337
aI3267
aI7673
aI15459
aI26681
aI4041
aI89429
aI198607
asS'd'
I14151
sa(dp28302
S'a'
I12462
sS's'
I18
sS'm'
(lp28303
I1
aI3
aI7
aI11
aI23
aI29
aI49
aI159
aI327
aI415
aI857
aI2411
aI2429
aI11839
aI20263
aI61813
aI31811
aI225443
asS'd'
I14152
sa(dp28304
S'a'
I12476
sS's'
I18
sS'm'
(lp28305
I1
aI1
aI3
aI11
aI7
aI61
aI61
aI119
aI431
aI299
aI1815
aI2857
aI7605
aI7517
aI15137
aI13727
aI73021
aI199325
asS'd'
I14153
sa(dp28306
S'a'
I12496
sS's'
I18
sS'm'
(lp28307
I1
aI1
aI7
aI3
aI5
aI19
aI51
aI19
aI59
aI637
aI591
aI2999
aI5997
aI13487
aI807
aI4887
aI112189
aI226597
asS'd'
I14154
sa(dp28308
S'a'
I12511
sS's'
I18
sS'm'
(lp28309
I1
aI1
aI1
aI13
aI21
aI5
aI55
aI167
aI463
aI679
aI891
aI3597
aI785
aI7717
aI17495
aI51681
aI55957
aI48561
asS'd'
I14155
sa(dp28310
S'a'
I12518
sS's'
I18
sS'm'
(lp28311
I1
aI1
aI5
aI5
aI9
aI55
aI55
aI143
aI193
aI839
aI785
aI1713
aI7457
aI11591
aI15803
aI2479
aI124663
aI72631
asS'd'
I14156
sa(dp28312
S'a'
I12530
sS's'
I18
sS'm'
(lp28313
I1
aI3
aI5
aI9
aI21
aI27
aI109
aI91
aI483
aI905
aI1369
aI2573
aI7173
aI13977
aI20131
aI17637
aI127477
aI66457
asS'd'
I14157
sa(dp28314
S'a'
I12532
sS's'
I18
sS'm'
(lp28315
I1
aI1
aI5
aI7
aI31
aI17
aI43
aI153
aI505
aI413
aI867
aI769
aI6947
aI10815
aI18805
aI5957
aI27715
aI24529
asS'd'
I14158
sa(dp28316
S'a'
I12539
sS's'
I18
sS'm'
(lp28317
I1
aI3
aI5
aI9
aI13
aI41
aI107
aI199
aI69
aI1019
aI2037
aI3221
aI1081
aI15051
aI6713
aI46379
aI17223
aI85453
asS'd'
I14159
sa(dp28318
S'a'
I12567
sS's'
I18
sS'm'
(lp28319
I1
aI3
aI3
aI7
aI23
aI51
aI45
aI133
aI227
aI373
aI1815
aI3795
aI5567
aI7153
aI25003
aI64951
aI75377
aI174115
asS'd'
I14160
sa(dp28320
S'a'
I12597
sS's'
I18
sS'm'
(lp28321
I1
aI1
aI7
aI15
aI7
aI51
aI33
aI239
aI113
aI133
aI1213
aI327
aI4841
aI11297
aI1093
aI40013
aI60843
aI99845
asS'd'
I14161
sa(dp28322
S'a'
I12601
sS's'
I18
sS'm'
(lp28323
I1
aI1
aI1
aI11
aI3
aI37
aI33
aI107
aI275
aI747
aI1451
aI1787
aI5029
aI3101
aI11575
aI36555
aI46181
aI221643
asS'd'
I14162
sa(dp28324
S'a'
I12602
sS's'
I18
sS'm'
(lp28325
I1
aI1
aI3
aI3
aI29
aI53
aI57
aI67
aI209
aI153
aI345
aI2897
aI5657
aI8907
aI14159
aI9899
aI102487
aI237721
asS'd'
I14163
sa(dp28326
S'a'
I12622
sS's'
I18
sS'm'
(lp28327
I1
aI3
aI1
aI7
aI7
aI59
aI17
aI151
aI423
aI903
aI2035
aI861
aI1057
aI2399
aI28547
aI3659
aI29583
aI100651
asS'd'
I14164
sa(dp28328
S'a'
I12693
sS's'
I18
sS'm'
(lp28329
I1
aI1
aI5
aI15
aI27
aI53
aI101
aI17
aI405
aI869
aI1253
aI1923
aI999
aI7787
aI23621
aI4351
aI48611
aI47129
asS'd'
I14165
sa(dp28330
S'a'
I12707
sS's'
I18
sS'm'
(lp28331
I1
aI3
aI7
aI11
aI13
aI43
aI61
aI161
aI43
aI831
aI2021
aI579
aI5353
aI12451
aI32261
aI39885
aI90051
aI34407
asS'd'
I14166
sa(dp28332
S'a'
I12713
sS's'
I18
sS'm'
(lp28333
I1
aI1
aI1
aI5
aI25
aI19
aI37
aI33
aI37
aI59
aI1399
aI1587
aI1517
aI4261
aI31215
aI33777
aI50447
aI87049
asS'd'
I14167
sa(dp28334
S'a'
I12716
sS's'
I18
sS'm'
(lp28335
I1
aI1
aI3
aI1
aI19
aI17
aI113
aI31
aI385
aI135
aI143
aI3997
aI1365
aI2625
aI22591
aI8887
aI31353
aI240603
asS'd'
I14168
sa(dp28336
S'a'
I12734
sS's'
I18
sS'm'
(lp28337
I1
aI3
aI7
aI11
aI21
aI7
aI55
aI171
aI233
aI1007
aI1321
aI2903
aI2457
aI3941
aI19667
aI49115
aI99119
aI185989
asS'd'
I14169
sa(dp28338
S'a'
I12763
sS's'
I18
sS'm'
(lp28339
I1
aI3
aI7
aI9
aI3
aI31
aI83
aI99
aI303
aI443
aI99
aI2285
aI1491
aI15897
aI21735
aI1575
aI74449
aI59615
asS'd'
I14170
sa(dp28340
S'a'
I12765
sS's'
I18
sS'm'
(lp28341
I1
aI3
aI1
aI5
aI29
aI37
aI125
aI213
aI277
aI115
aI255
aI137
aI6071
aI13561
aI23871
aI48129
aI120211
aI168603
asS'd'
I14171
sa(dp28342
S'a'
I12799
sS's'
I18
sS'm'
(lp28343
I1
aI1
aI7
aI3
aI9
aI21
aI93
aI239
aI399
aI21
aI9
aI409
aI3403
aI9517
aI6421
aI17121
aI65697
aI251985
asS'd'
I14172
sa(dp28344
S'a'
I12843
sS's'
I18
sS'm'
(lp28345
I1
aI1
aI1
aI15
aI27
aI35
aI17
aI113
aI471
aI357
aI1703
aI871
aI1803
aI3495
aI27437
aI48343
aI86425
aI155245
asS'd'
I14173
sa(dp28346
S'a'
I12853
sS's'
I18
sS'm'
(lp28347
I1
aI3
aI7
aI11
aI19
aI45
aI51
aI195
aI345
aI77
aI1403
aI2527
aI3405
aI14057
aI31965
aI17375
aI35107
aI246545
asS'd'
I14174
sa(dp28348
S'a'
I12865
sS's'
I18
sS'm'
(lp28349
I1
aI3
aI5
aI7
aI7
aI5
aI115
aI39
aI51
aI261
aI1883
aI1793
aI3423
aI3613
aI20399
aI27267
aI99875
aI119719
asS'd'
I14175
sa(dp28350
S'a'
I12866
sS's'
I18
sS'm'
(lp28351
I1
aI3
aI3
aI9
aI21
aI23
aI65
aI69
aI261
aI79
aI151
aI3387
aI7789
aI13275
aI30135
aI52229
aI40787
aI181297
asS'd'
I14176
sa(dp28352
S'a'
I12919
sS's'
I18
sS'm'
(lp28353
I1
aI1
aI5
aI13
aI19
aI49
aI111
aI125
aI433
aI99
aI1673
aI2091
aI5447
aI9377
aI9085
aI4659
aI75121
aI105809
asS'd'
I14177
sa(dp28354
S'a'
I12926
sS's'
I18
sS'm'
(lp28355
I1
aI3
aI7
aI1
aI3
aI27
aI107
aI109
aI245
aI431
aI1727
aI3269
aI2099
aI10777
aI21843
aI63377
aI47343
aI126269
asS'd'
I14178
sa(dp28356
S'a'
I12936
sS's'
I18
sS'm'
(lp28357
I1
aI3
aI1
aI15
aI7
aI17
aI107
aI153
aI37
aI287
aI1873
aI573
aI5025
aI3735
aI32545
aI35693
aI38083
aI89569
asS'd'
I14179
sa(dp28358
S'a'
I12944
sS's'
I18
sS'm'
(lp28359
I1
aI3
aI1
aI3
aI9
aI51
aI99
aI247
aI45
aI703
aI1231
aI1895
aI6309
aI12137
aI14697
aI25441
aI129701
aI198811
asS'd'
I14180
sa(dp28360
S'a'
I12959
sS's'
I18
sS'm'
(lp28361
I1
aI3
aI3
aI11
aI5
aI43
aI1
aI31
aI359
aI563
aI1013
aI3475
aI3935
aI7855
aI10085
aI15279
aI25109
aI225643
asS'd'
I14181
sa(dp28362
S'a'
I12965
sS's'
I18
sS'm'
(lp28363
I1
aI3
aI3
aI3
aI9
aI47
aI49
aI193
aI223
aI23
aI391
aI847
aI7233
aI14955
aI10645
aI50535
aI5415
aI119791
asS'd'
I14182
sa(dp28364
S'a'
I12980
sS's'
I18
sS'm'
(lp28365
I1
aI3
aI1
aI7
aI3
aI7
aI57
aI189
aI219
aI287
aI401
aI1767
aI5585
aI13983
aI18485
aI56725
aI71905
aI33779
asS'd'
I14183
sa(dp28366
S'a'
I13004
sS's'
I18
sS'm'
(lp28367
I1
aI3
aI5
aI15
aI23
aI17
aI115
aI223
aI35
aI263
aI345
aI3459
aI857
aI1467
aI30255
aI50127
aI72985
aI62509
asS'd'
I14184
sa(dp28368
S'a'
I13016
sS's'
I18
sS'm'
(lp28369
I1
aI3
aI1
aI9
aI3
aI27
aI125
aI43
aI47
aI183
aI1421
aI319
aI4273
aI10701
aI21761
aI23947
aI22531
aI10855
asS'd'
I14185
sa(dp28370
S'a'
I13019
sS's'
I18
sS'm'
(lp28371
I1
aI1
aI3
aI15
aI13
aI55
aI77
aI1
aI295
aI841
aI1115
aI4093
aI7993
aI13309
aI24851
aI35411
aI105201
aI188543
asS'd'
I14186
sa(dp28372
S'a'
I13037
sS's'
I18
sS'm'
(lp28373
I1
aI3
aI3
aI5
aI19
aI39
aI101
aI19
aI225
aI997
aI1999
aI407
aI3147
aI15393
aI30379
aI26221
aI21685
aI114167
asS'd'
I14187
sa(dp28374
S'a'
I13055
sS's'
I18
sS'm'
(lp28375
I1
aI3
aI3
aI3
aI23
aI15
aI3
aI57
aI45
aI381
aI47
aI1839
aI5491
aI6775
aI18477
aI51443
aI757
aI183111
asS'd'
I14188
sa(dp28376
S'a'
I13063
sS's'
I18
sS'm'
(lp28377
I1
aI3
aI3
aI7
aI31
aI27
aI107
aI229
aI1
aI977
aI125
aI1137
aI4873
aI14381
aI8705
aI64641
aI38447
aI239887
asS'd'
I14189
sa(dp28378
S'a'
I13098
sS's'
I18
sS'm'
(lp28379
I1
aI1
aI3
aI13
aI9
aI35
aI49
aI187
aI459
aI407
aI1677
aI2007
aI1091
aI12385
aI8911
aI38221
aI108681
aI171641
asS'd'
I14190
sa(dp28380
S'a'
I13103
sS's'
I18
sS'm'
(lp28381
I1
aI1
aI3
aI5
aI7
aI61
aI115
aI155
aI437
aI829
aI1041
aI2191
aI1489
aI1269
aI27613
aI48713
aI40095
aI206057
asS'd'
I14191
sa(dp28382
S'a'
I13106
sS's'
I18
sS'm'
(lp28383
I1
aI3
aI7
aI1
aI23
aI1
aI17
aI215
aI363
aI119
aI845
aI987
aI5619
aI5857
aI11307
aI63171
aI76921
aI201767
asS'd'
I14192
sa(dp28384
S'a'
I13115
sS's'
I18
sS'm'
(lp28385
I1
aI1
aI7
aI9
aI29
aI23
aI63
aI247
aI37
aI199
aI103
aI1215
aI913
aI12865
aI24089
aI35101
aI117677
aI261393
asS'd'
I14193
sa(dp28386
S'a'
I13143
sS's'
I18
sS'm'
(lp28387
I1
aI1
aI1
aI1
aI3
aI19
aI7
aI159
aI183
aI275
aI467
aI3629
aI6575
aI3351
aI26955
aI29247
aI119007
aI67895
asS'd'
I14194
sa(dp28388
S'a'
I13144
sS's'
I18
sS'm'
(lp28389
I1
aI1
aI7
aI13
aI1
aI33
aI31
aI211
aI103
aI495
aI417
aI817
aI7129
aI10169
aI11445
aI41511
aI73101
aI185357
asS'd'
I14195
sa(dp28390
S'a'
I13153
sS's'
I18
sS'm'
(lp28391
I1
aI1
aI3
aI5
aI3
aI27
aI99
aI1
aI425
aI295
aI131
aI835
aI1833
aI4547
aI8777
aI29489
aI60303
aI191437
asS'd'
I14196
sa(dp28392
S'a'
I13171
sS's'
I18
sS'm'
(lp28393
I1
aI1
aI5
aI5
aI31
aI63
aI71
aI189
aI317
aI61
aI385
aI891
aI2257
aI8281
aI17325
aI12207
aI125847
aI28259
asS'd'
I14197
sa(dp28394
S'a'
I13184
sS's'
I18
sS'm'
(lp28395
I1
aI1
aI3
aI3
aI19
aI19
aI97
aI83
aI495
aI551
aI1339
aI1699
aI3047
aI13623
aI27731
aI28999
aI101225
aI146139
asS'd'
I14198
sa(dp28396
S'a'
I13193
sS's'
I18
sS'm'
(lp28397
I1
aI3
aI5
aI7
aI25
aI9
aI37
aI73
aI239
aI47
aI583
aI3337
aI1329
aI79
aI30109
aI12451
aI10163
aI62943
asS'd'
I14199
sa(dp28398
S'a'
I13199
sS's'
I18
sS'm'
(lp28399
I1
aI1
aI3
aI1
aI9
aI1
aI31
aI181
aI231
aI441
aI1241
aI233
aI6049
aI2401
aI9867
aI58911
aI20599
aI26321
asS'd'
I14200
sa(dp28400
S'a'
I13207
sS's'
I18
sS'm'
(lp28401
I1
aI3
aI1
aI15
aI21
aI43
aI21
aI43
aI273
aI865
aI79
aI2069
aI3375
aI16069
aI12355
aI56355
aI9735
aI243719
asS'd'
I14201
sa(dp28402
S'a'
I13217
sS's'
I18
sS'm'
(lp28403
I1
aI3
aI5
aI11
aI13
aI21
aI5
aI73
aI423
aI761
aI1947
aI2935
aI3931
aI5573
aI83
aI58251
aI113115
aI245767
asS'd'
I14202
sa(dp28404
S'a'
I13227
sS's'
I18
sS'm'
(lp28405
I1
aI3
aI7
aI7
aI15
aI49
aI73
aI211
aI309
aI635
aI1257
aI2185
aI7151
aI11959
aI26885
aI45955
aI103503
aI161709
asS'd'
I14203
sa(dp28406
S'a'
I13232
sS's'
I18
sS'm'
(lp28407
I1
aI3
aI1
aI15
aI31
aI53
aI29
aI35
aI343
aI87
aI1833
aI2483
aI1847
aI4709
aI6105
aI21961
aI106541
aI46741
asS'd'
I14204
sa(dp28408
S'a'
I13276
sS's'
I18
sS'm'
(lp28409
I1
aI1
aI7
aI7
aI21
aI43
aI41
aI51
aI29
aI521
aI713
aI3693
aI483
aI11777
aI10453
aI43691
aI97585
aI133193
asS'd'
I14205
sa(dp28410
S'a'
I13283
sS's'
I18
sS'm'
(lp28411
I1
aI3
aI3
aI11
aI5
aI45
aI3
aI179
aI183
aI255
aI1291
aI1795
aI5721
aI10911
aI18395
aI64349
aI23141
aI99481
asS'd'
I14206
sa(dp28412
S'a'
I13307
sS's'
I18
sS'm'
(lp28413
I1
aI1
aI1
aI5
aI5
aI25
aI61
aI169
aI475
aI953
aI617
aI559
aI5945
aI16377
aI30063
aI30079
aI83305
aI81745
asS'd'
I14207
sa(dp28414
S'a'
I13317
sS's'
I18
sS'm'
(lp28415
I1
aI1
aI5
aI11
aI29
aI59
aI113
aI37
aI153
aI807
aI135
aI2639
aI4535
aI7079
aI6387
aI63523
aI89669
aI198803
asS'd'
I14208
sa(dp28416
S'a'
I13330
sS's'
I18
sS'm'
(lp28417
I1
aI1
aI7
aI7
aI19
aI23
aI71
aI51
aI165
aI733
aI1427
aI2473
aI331
aI5027
aI9299
aI20617
aI126775
aI91619
asS'd'
I14209
sa(dp28418
S'a'
I13345
sS's'
I18
sS'm'
(lp28419
I1
aI3
aI5
aI7
aI31
aI35
aI117
aI235
aI29
aI677
aI1243
aI281
aI6287
aI535
aI4783
aI37781
aI130929
aI14193
asS'd'
I14210
sa(dp28420
S'a'
I13363
sS's'
I18
sS'm'
(lp28421
I1
aI1
aI5
aI3
aI25
aI1
aI29
aI109
aI289
aI631
aI41
aI361
aI5537
aI9657
aI7475
aI63749
aI50325
aI169791
asS'd'
I14211
sa(dp28422
S'a'
I13387
sS's'
I18
sS'm'
(lp28423
I1
aI3
aI3
aI15
aI21
aI53
aI85
aI43
aI341
aI907
aI475
aI3257
aI2541
aI10397
aI30847
aI63681
aI121427
aI192135
asS'd'
I14212
sa(dp28424
S'a'
I13426
sS's'
I18
sS'm'
(lp28425
I1
aI1
aI1
aI1
aI9
aI59
aI59
aI233
aI335
aI345
aI1749
aI4007
aI1833
aI7789
aI21015
aI48939
aI15967
aI201321
asS'd'
I14213
sa(dp28426
S'a'
I13428
sS's'
I18
sS'm'
(lp28427
I1
aI1
aI1
aI5
aI11
aI9
aI101
aI243
aI391
aI1003
aI1019
aI311
aI3707
aI10223
aI21627
aI8237
aI53861
aI159785
asS'd'
I14214
sa(dp28428
S'a'
I13431
sS's'
I18
sS'm'
(lp28429
I1
aI3
aI7
aI15
aI5
aI3
aI109
aI197
aI507
aI947
aI833
aI1161
aI4021
aI5575
aI8081
aI45381
aI112597
aI70407
asS'd'
I14215
sa(dp28430
S'a'
I13489
sS's'
I18
sS'm'
(lp28431
I1
aI1
aI3
aI1
aI7
aI63
aI63
aI53
aI481
aI1003
aI43
aI2503
aI2303
aI12593
aI21403
aI2965
aI5377
aI91491
asS'd'
I14216
sa(dp28432
S'a'
I13501
sS's'
I18
sS'm'
(lp28433
I1
aI1
aI1
aI1
aI29
aI45
aI49
aI73
aI253
aI197
aI1245
aI1509
aI4747
aI6207
aI28321
aI59193
aI112687
aI83719
asS'd'
I14217
sa(dp28434
S'a'
I13533
sS's'
I18
sS'm'
(lp28435
I1
aI1
aI7
aI3
aI11
aI51
aI23
aI83
aI185
aI643
aI1427
aI227
aI2261
aI12521
aI27033
aI5129
aI53111
aI34975
asS'd'
I14218
sa(dp28436
S'a'
I13589
sS's'
I18
sS'm'
(lp28437
I1
aI1
aI7
aI3
aI5
aI41
aI55
aI175
aI447
aI603
aI723
aI2141
aI1189
aI4921
aI16905
aI2463
aI83641
aI247241
asS'd'
I14219
sa(dp28438
S'a'
I13600
sS's'
I18
sS'm'
(lp28439
I1
aI3
aI3
aI13
aI5
aI11
aI95
aI59
aI391
aI319
aI1675
aI329
aI7559
aI11585
aI28905
aI27843
aI106667
aI15531
asS'd'
I14220
sa(dp28440
S'a'
I13603
sS's'
I18
sS'm'
(lp28441
I1
aI3
aI3
aI3
aI27
aI17
aI103
aI115
aI447
aI657
aI267
aI2541
aI665
aI7819
aI4155
aI32191
aI60999
aI48737
asS'd'
I14221
sa(dp28442
S'a'
I13623
sS's'
I18
sS'm'
(lp28443
I1
aI1
aI5
aI7
aI7
aI49
aI87
aI171
aI457
aI149
aI1699
aI4081
aI3913
aI7889
aI29517
aI3339
aI33139
aI8925
asS'd'
I14222
sa(dp28444
S'a'
I13665
sS's'
I18
sS'm'
(lp28445
I1
aI1
aI1
aI9
aI11
aI51
aI87
aI115
aI429
aI505
aI1665
aI2361
aI5811
aI1621
aI12727
aI33703
aI52255
aI93835
asS'd'
I14223
sa(dp28446
S'a'
I13672
sS's'
I18
sS'm'
(lp28447
I1
aI3
aI3
aI5
aI27
aI11
aI35
aI49
aI281
aI607
aI1791
aI4065
aI5103
aI5253
aI975
aI20353
aI38253
aI139363
asS'd'
I14224
sa(dp28448
S'a'
I13675
sS's'
I18
sS'm'
(lp28449
I1
aI1
aI5
aI9
aI29
aI15
aI37
aI141
aI445
aI751
aI1219
aI2217
aI7207
aI14981
aI21113
aI3313
aI107127
aI135567
asS'd'
I14225
sa(dp28450
S'a'
I13708
sS's'
I18
sS'm'
(lp28451
I1
aI3
aI3
aI15
aI1
aI41
aI23
aI27
aI167
aI609
aI913
aI631
aI923
aI6939
aI9793
aI57869
aI126577
aI145271
asS'd'
I14226
sa(dp28452
S'a'
I13713
sS's'
I18
sS'm'
(lp28453
I1
aI1
aI3
aI7
aI27
aI47
aI127
aI5
aI213
aI575
aI559
aI2541
aI3457
aI2903
aI19529
aI53395
aI105353
aI212607
asS'd'
I14227
sa(dp28454
S'a'
I13720
sS's'
I18
sS'm'
(lp28455
I1
aI3
aI1
aI13
aI27
aI41
aI31
aI111
aI371
aI1019
aI241
aI2075
aI2571
aI10739
aI28163
aI16093
aI41127
aI69783
asS'd'
I14228
sa(dp28456
S'a'
I13730
sS's'
I18
sS'm'
(lp28457
I1
aI1
aI5
aI1
aI21
aI9
aI15
aI141
aI287
aI675
aI1721
aI2291
aI6587
aI7503
aI6363
aI9109
aI33547
aI259627
asS'd'
I14229
sa(dp28458
S'a'
I13735
sS's'
I18
sS'm'
(lp28459
I1
aI3
aI7
aI3
aI3
aI53
aI7
aI153
aI183
aI761
aI191
aI3735
aI2619
aI11153
aI19601
aI33855
aI82345
aI72755
asS'd'
I14230
sa(dp28460
S'a'
I13767
sS's'
I18
sS'm'
(lp28461
I1
aI3
aI1
aI9
aI19
aI21
aI41
aI105
aI281
aI833
aI981
aI2733
aI7179
aI14691
aI18365
aI56283
aI53719
aI191601
asS'd'
I14231
sa(dp28462
S'a'
I13773
sS's'
I18
sS'm'
(lp28463
I1
aI1
aI7
aI11
aI23
aI1
aI55
aI213
aI105
aI517
aI809
aI4087
aI825
aI7011
aI15701
aI54047
aI123831
aI49833
asS'd'
I14232
sa(dp28464
S'a'
I13792
sS's'
I18
sS'm'
(lp28465
I1
aI1
aI7
aI13
aI27
aI9
aI111
aI57
aI357
aI95
aI1489
aI887
aI5273
aI2833
aI8757
aI9371
aI44637
aI94939
asS'd'
I14233
sa(dp28466
S'a'
I13843
sS's'
I18
sS'm'
(lp28467
I1
aI1
aI3
aI5
aI1
aI17
aI43
aI31
aI509
aI353
aI401
aI1077
aI7567
aI9657
aI15065
aI32017
aI8491
aI214477
asS'd'
I14234
sa(dp28468
S'a'
I13891
sS's'
I18
sS'm'
(lp28469
I1
aI1
aI1
aI15
aI7
aI59
aI41
aI99
aI101
aI845
aI1479
aI2153
aI4281
aI12839
aI237
aI54095
aI125873
aI57165
asS'd'
I14235
sa(dp28470
S'a'
I13946
sS's'
I18
sS'm'
(lp28471
I1
aI3
aI7
aI13
aI5
aI17
aI1
aI249
aI309
aI351
aI709
aI3943
aI7775
aI6449
aI26611
aI54019
aI121015
aI213535
asS'd'
I14236
sa(dp28472
S'a'
I13967
sS's'
I18
sS'm'
(lp28473
I1
aI1
aI1
aI5
aI7
aI25
aI33
aI149
aI291
aI777
aI161
aI2729
aI117
aI4999
aI16781
aI23383
aI85161
aI71689
asS'd'
I14237
sa(dp28474
S'a'
I13976
sS's'
I18
sS'm'
(lp28475
I1
aI1
aI3
aI11
aI5
aI63
aI119
aI165
aI45
aI135
aI1723
aI811
aI1259
aI11055
aI28625
aI37559
aI128401
aI100715
asS'd'
I14238
sa(dp28476
S'a'
I14017
sS's'
I18
sS'm'
(lp28477
I1
aI3
aI7
aI1
aI1
aI39
aI11
aI255
aI423
aI289
aI1359
aI2827
aI4637
aI4089
aI26659
aI58701
aI117403
aI133971
asS'd'
I14239
sa(dp28478
S'a'
I14030
sS's'
I18
sS'm'
(lp28479
I1
aI1
aI7
aI1
aI25
aI9
aI127
aI121
aI147
aI831
aI17
aI3521
aI1535
aI10931
aI17305
aI56671
aI22425
aI157341
asS'd'
I14240
sa(dp28480
S'a'
I14041
sS's'
I18
sS'm'
(lp28481
I1
aI3
aI1
aI15
aI5
aI55
aI95
aI95
aI169
aI497
aI739
aI2031
aI339
aI13461
aI20619
aI24553
aI81805
aI90789
asS'd'
I14241
sa(dp28482
S'a'
I14054
sS's'
I18
sS'm'
(lp28483
I1
aI1
aI7
aI7
aI7
aI19
aI15
aI203
aI287
aI673
aI1033
aI3857
aI2761
aI10835
aI11039
aI62329
aI37893
aI6119
asS'd'
I14242
sa(dp28484
S'a'
I14058
sS's'
I18
sS'm'
(lp28485
I1
aI1
aI3
aI11
aI19
aI35
aI55
aI9
aI399
aI443
aI583
aI89
aI2387
aI747
aI9551
aI9907
aI96871
aI175457
asS'd'
I14243
sa(dp28486
S'a'
I14060
sS's'
I18
sS'm'
(lp28487
I1
aI1
aI3
aI1
aI11
aI57
aI121
aI89
aI491
aI133
aI545
aI683
aI5751
aI839
aI25975
aI44725
aI59863
aI142671
asS'd'
I14244
sa(dp28488
S'a'
I14063
sS's'
I18
sS'm'
(lp28489
I1
aI1
aI3
aI1
aI23
aI1
aI111
aI177
aI1
aI103
aI1933
aI2783
aI6857
aI51
aI14547
aI5945
aI14757
aI39783
asS'd'
I14245
sa(dp28490
S'a'
I14078
sS's'
I18
sS'm'
(lp28491
I1
aI3
aI7
aI13
aI25
aI13
aI51
aI247
aI325
aI361
aI1225
aI15
aI1929
aI1729
aI25293
aI59495
aI82111
aI101809
asS'd'
I14246
sa(dp28492
S'a'
I14080
sS's'
I18
sS'm'
(lp28493
I1
aI3
aI3
aI1
aI19
aI37
aI67
aI85
aI105
aI589
aI1273
aI2995
aI8017
aI1613
aI22189
aI2549
aI22671
aI72813
asS'd'
I14247
sa(dp28494
S'a'
I14089
sS's'
I18
sS'm'
(lp28495
I1
aI3
aI7
aI9
aI15
aI41
aI25
aI43
aI411
aI663
aI387
aI2861
aI3627
aI5839
aI733
aI53479
aI76241
aI116763
asS'd'
I14248
sa(dp28496
S'a'
I14131
sS's'
I18
sS'm'
(lp28497
I1
aI3
aI3
aI3
aI17
aI5
aI73
aI153
aI133
aI247
aI881
aI2853
aI6059
aI2259
aI10181
aI63251
aI107089
aI22579
asS'd'
I14249
sa(dp28498
S'a'
I14134
sS's'
I18
sS'm'
(lp28499
I1
aI3
aI1
aI5
aI11
aI15
aI17
aI235
aI23
aI15
aI521
aI235
aI4137
aI12705
aI24775
aI18197
aI56295
aI28035
asS'd'
I14250
sa(dp28500
S'a'
I14137
sS's'
I18
sS'm'
(lp28501
I1
aI3
aI5
aI5
aI13
aI9
aI77
aI69
aI19
aI755
aI1663
aI1499
aI1049
aI12935
aI28835
aI55413
aI71511
aI221223
asS'd'
I14251
sa(dp28502
S'a'
I14188
sS's'
I18
sS'm'
(lp28503
I1
aI1
aI1
aI1
aI27
aI31
aI21
aI39
aI197
aI519
aI1621
aI3703
aI2541
aI8865
aI21947
aI53605
aI114551
aI205697
asS'd'
I14252
sa(dp28504
S'a'
I14215
sS's'
I18
sS'm'
(lp28505
I1
aI1
aI1
aI13
aI11
aI53
aI41
aI245
aI495
aI275
aI385
aI3071
aI1913
aI11135
aI8571
aI58551
aI39049
aI5459
asS'd'
I14253
sa(dp28506
S'a'
I14222
sS's'
I18
sS'm'
(lp28507
I1
aI3
aI5
aI5
aI11
aI63
aI25
aI173
aI57
aI441
aI1749
aI79
aI3191
aI7733
aI13111
aI23453
aI118399
aI101845
asS'd'
I14254
sa(dp28508
S'a'
I14249
sS's'
I18
sS'm'
(lp28509
I1
aI1
aI3
aI11
aI29
aI25
aI119
aI39
aI65
aI623
aI517
aI1325
aI5981
aI8381
aI32031
aI25585
aI105537
aI214241
asS'd'
I14255
sa(dp28510
S'a'
I14292
sS's'
I18
sS'm'
(lp28511
I1
aI3
aI5
aI9
aI21
aI43
aI13
aI69
aI109
aI311
aI1893
aI1941
aI2491
aI7815
aI4067
aI56749
aI33761
aI191145
asS'd'
I14256
sa(dp28512
S'a'
I14320
sS's'
I18
sS'm'
(lp28513
I1
aI1
aI1
aI7
aI9
aI5
aI123
aI149
aI65
aI729
aI1967
aI3089
aI3757
aI2333
aI24587
aI36047
aI118105
aI146277
asS'd'
I14257
sa(dp28514
S'a'
I14339
sS's'
I18
sS'm'
(lp28515
I1
aI3
aI7
aI13
aI9
aI35
aI39
aI219
aI161
aI93
aI275
aI3619
aI353
aI14595
aI24673
aI54753
aI117175
aI81891
asS'd'
I14258
sa(dp28516
S'a'
I14346
sS's'
I18
sS'm'
(lp28517
I1
aI3
aI3
aI13
aI15
aI61
aI95
aI135
aI271
aI595
aI1103
aI877
aI747
aI2535
aI7733
aI25509
aI65673
aI62089
asS'd'
I14259
sa(dp28518
S'a'
I14353
sS's'
I18
sS'm'
(lp28519
I1
aI3
aI3
aI5
aI21
aI21
aI67
aI5
aI373
aI377
aI61
aI2337
aI489
aI5801
aI23203
aI42377
aI7801
aI178095
asS'd'
I14260
sa(dp28520
S'a'
I14365
sS's'
I18
sS'm'
(lp28521
I1
aI1
aI7
aI5
aI25
aI17
aI61
aI133
aI181
aI261
aI1627
aI1615
aI6851
aI4763
aI30353
aI53349
aI7545
aI66733
asS'd'
I14261
sa(dp28522
S'a'
I14384
sS's'
I18
sS'm'
(lp28523
I1
aI3
aI5
aI7
aI29
aI3
aI85
aI231
aI121
aI669
aI1925
aI403
aI777
aI10605
aI24125
aI60819
aI8253
aI81209
asS'd'
I14262
sa(dp28524
S'a'
I14414
sS's'
I18
sS'm'
(lp28525
I1
aI1
aI5
aI3
aI5
aI5
aI1
aI53
aI9
aI445
aI1339
aI2643
aI5527
aI13757
aI9409
aI31993
aI80845
aI97863
asS'd'
I14263
sa(dp28526
S'a'
I14428
sS's'
I18
sS'm'
(lp28527
I1
aI1
aI7
aI13
aI29
aI49
aI97
aI89
aI319
aI349
aI1739
aI3615
aI1113
aI11791
aI17429
aI37195
aI1159
aI32211
asS'd'
I14264
sa(dp28528
S'a'
I14435
sS's'
I18
sS'm'
(lp28529
I1
aI3
aI5
aI11
aI9
aI61
aI109
aI167
aI119
aI499
aI1157
aI3615
aI5773
aI8839
aI27915
aI47837
aI14945
aI187225
asS'd'
I14265
sa(dp28530
S'a'
I14483
sS's'
I18
sS'm'
(lp28531
I1
aI1
aI5
aI1
aI9
aI3
aI7
aI179
aI323
aI279
aI43
aI1337
aI4813
aI9917
aI2033
aI34657
aI130769
aI208089
asS'd'
I14266
sa(dp28532
S'a'
I14486
sS's'
I18
sS'm'
(lp28533
I1
aI1
aI3
aI1
aI31
aI57
aI57
aI73
aI21
aI661
aI1861
aI1661
aI7619
aI12155
aI23123
aI49751
aI130697
aI74143
asS'd'
I14267
sa(dp28534
S'a'
I14506
sS's'
I18
sS'm'
(lp28535
I1
aI3
aI3
aI13
aI11
aI61
aI95
aI75
aI227
aI491
aI463
aI597
aI2721
aI12323
aI26195
aI53657
aI44413
aI68965
asS'd'
I14268
sa(dp28536
S'a'
I14513
sS's'
I18
sS'm'
(lp28537
I1
aI1
aI7
aI11
aI5
aI51
aI103
aI123
aI203
aI911
aI203
aI1641
aI7009
aI9479
aI303
aI37649
aI32751
aI172777
asS'd'
I14269
sa(dp28538
S'a'
I14520
sS's'
I18
sS'm'
(lp28539
I1
aI1
aI3
aI7
aI11
aI59
aI111
aI5
aI271
aI863
aI269
aI3457
aI489
aI10877
aI8645
aI62567
aI24893
aI201587
asS'd'
I14270
sa(dp28540
S'a'
I14526
sS's'
I18
sS'm'
(lp28541
I1
aI1
aI5
aI7
aI29
aI23
aI127
aI151
aI371
aI121
aI1103
aI3951
aI3107
aI15563
aI6243
aI1631
aI75065
aI107681
asS'd'
I14271
sa(dp28542
S'a'
I14528
sS's'
I18
sS'm'
(lp28543
I1
aI1
aI7
aI15
aI27
aI45
aI43
aI83
aI461
aI673
aI715
aI3245
aI313
aI13731
aI21981
aI58853
aI46569
aI165463
asS'd'
I14272
sa(dp28544
S'a'
I14533
sS's'
I18
sS'm'
(lp28545
I1
aI1
aI1
aI9
aI7
aI53
aI63
aI43
aI3
aI187
aI1325
aI447
aI5113
aI4993
aI21807
aI24329
aI4499
aI30067
asS'd'
I14273
sa(dp28546
S'a'
I14576
sS's'
I18
sS'm'
(lp28547
I1
aI3
aI1
aI9
aI21
aI45
aI111
aI231
aI407
aI213
aI1977
aI2269
aI2323
aI4595
aI30427
aI54753
aI95049
aI195409
asS'd'
I14274
sa(dp28548
S'a'
I14599
sS's'
I18
sS'm'
(lp28549
I1
aI1
aI1
aI9
aI29
aI43
aI89
aI201
aI499
aI329
aI847
aI3831
aI5403
aI13001
aI6037
aI14371
aI25805
aI169237
asS'd'
I14275
sa(dp28550
S'a'
I14613
sS's'
I18
sS'm'
(lp28551
I1
aI1
aI1
aI11
aI29
aI61
aI61
aI203
aI91
aI189
aI1603
aI939
aI6575
aI3195
aI4731
aI44923
aI33627
aI21683
asS'd'
I14276
sa(dp28552
S'a'
I14639
sS's'
I18
sS'm'
(lp28553
I1
aI3
aI3
aI11
aI7
aI57
aI93
aI181
aI479
aI99
aI681
aI2875
aI7649
aI5555
aI27087
aI6841
aI69859
aI153689
asS'd'
I14277
sa(dp28554
S'a'
I14644
sS's'
I18
sS'm'
(lp28555
I1
aI1
aI1
aI9
aI17
aI45
aI97
aI47
aI91
aI879
aI1463
aI3041
aI2917
aI769
aI13675
aI26489
aI88559
aI120991
asS'd'
I14278
sa(dp28556
S'a'
I14653
sS's'
I18
sS'm'
(lp28557
I1
aI3
aI7
aI1
aI11
aI13
aI43
aI9
aI483
aI399
aI793
aI3965
aI2375
aI4957
aI17747
aI50905
aI56987
aI231265
asS'd'
I14279
sa(dp28558
S'a'
I14662
sS's'
I18
sS'm'
(lp28559
I1
aI3
aI3
aI13
aI23
aI45
aI69
aI67
aI397
aI437
aI1993
aI2569
aI8035
aI8531
aI27623
aI53567
aI123189
aI242515
asS'd'
I14280
sa(dp28560
S'a'
I14695
sS's'
I18
sS'm'
(lp28561
I1
aI1
aI1
aI3
aI5
aI9
aI21
aI227
aI499
aI205
aI431
aI3711
aI5307
aI15773
aI11337
aI6349
aI123507
aI95941
asS'd'
I14281
sa(dp28562
S'a'
I14704
sS's'
I18
sS'm'
(lp28563
I1
aI3
aI7
aI1
aI13
aI15
aI101
aI91
aI209
aI611
aI537
aI1427
aI101
aI2619
aI10513
aI44323
aI92745
aI249127
asS'd'
I14282
sa(dp28564
S'a'
I14713
sS's'
I18
sS'm'
(lp28565
I1
aI3
aI3
aI1
aI21
aI7
aI79
aI241
aI273
aI567
aI605
aI2371
aI5427
aI15147
aI20139
aI40987
aI75551
aI236213
asS'd'
I14283
sa(dp28566
S'a'
I14725
sS's'
I18
sS'm'
(lp28567
I1
aI1
aI7
aI11
aI25
aI19
aI77
aI209
aI313
aI663
aI115
aI3697
aI3641
aI12461
aI9877
aI18331
aI70809
aI78923
asS'd'
I14284
sa(dp28568
S'a'
I14783
sS's'
I18
sS'm'
(lp28569
I1
aI1
aI5
aI5
aI29
aI45
aI7
aI207
aI1
aI357
aI1089
aI3861
aI4161
aI3209
aI27845
aI20947
aI68909
aI125179
asS'd'
I14285
sa(dp28570
S'a'
I14810
sS's'
I18
sS'm'
(lp28571
I1
aI3
aI1
aI13
aI17
aI11
aI27
aI165
aI179
aI489
aI1611
aI2801
aI2385
aI2971
aI6777
aI16149
aI59811
aI151043
asS'd'
I14286
sa(dp28572
S'a'
I14816
sS's'
I18
sS'm'
(lp28573
I1
aI1
aI3
aI3
aI17
aI53
aI121
aI227
aI7
aI71
aI1855
aI639
aI5135
aI6349
aI7163
aI22997
aI112551
aI44167
asS'd'
I14287
sa(dp28574
S'a'
I14822
sS's'
I18
sS'm'
(lp28575
I1
aI3
aI1
aI11
aI15
aI9
aI125
aI213
aI485
aI291
aI1781
aI3621
aI7529
aI13353
aI13903
aI24151
aI130253
aI187097
asS'd'
I14288
sa(dp28576
S'a'
I14828
sS's'
I18
sS'm'
(lp28577
I1
aI3
aI1
aI3
aI1
aI61
aI39
aI157
aI455
aI945
aI739
aI589
aI7259
aI7149
aI16455
aI12649
aI72003
aI152419
asS'd'
I14289
sa(dp28578
S'a'
I14883
sS's'
I18
sS'm'
(lp28579
I1
aI1
aI3
aI3
aI31
aI23
aI17
aI255
aI319
aI907
aI563
aI2571
aI2149
aI15323
aI20289
aI46061
aI32769
aI184353
asS'd'
I14290
sa(dp28580
S'a'
I14889
sS's'
I18
sS'm'
(lp28581
I1
aI3
aI7
aI9
aI21
aI51
aI27
aI13
aI495
aI909
aI2039
aI1435
aI4791
aI10037
aI30119
aI3405
aI22535
aI42247
asS'd'
I14291
sa(dp28582
S'a'
I14904
sS's'
I18
sS'm'
(lp28583
I1
aI1
aI3
aI15
aI11
aI25
aI123
aI149
aI185
aI635
aI473
aI527
aI433
aI10373
aI18205
aI853
aI94619
aI202507
asS'd'
I14292
sa(dp28584
S'a'
I14917
sS's'
I18
sS'm'
(lp28585
I1
aI1
aI7
aI15
aI7
aI39
aI7
aI69
aI157
aI533
aI369
aI4031
aI1335
aI4279
aI8049
aI28491
aI103753
aI257477
asS'd'
I14293
sa(dp28586
S'a'
I14927
sS's'
I18
sS'm'
(lp28587
I1
aI3
aI1
aI15
aI29
aI51
aI113
aI77
aI5
aI961
aI1863
aI1477
aI5009
aI9519
aI32029
aI2367
aI55705
aI149597
asS'd'
I14294
sa(dp28588
S'a'
I14941
sS's'
I18
sS'm'
(lp28589
I1
aI3
aI1
aI5
aI19
aI43
aI49
aI107
aI59
aI693
aI867
aI3011
aI2703
aI3633
aI24567
aI52621
aI35839
aI134823
asS'd'
I14295
sa(dp28590
S'a'
I14946
sS's'
I18
sS'm'
(lp28591
I1
aI3
aI3
aI7
aI19
aI15
aI81
aI105
aI23
aI375
aI451
aI3017
aI1263
aI7589
aI24453
aI21885
aI57651
aI52613
asS'd'
I14296
sa(dp28592
S'a'
I14952
sS's'
I18
sS'm'
(lp28593
I1
aI1
aI3
aI7
aI3
aI59
aI19
aI1
aI3
aI243
aI1891
aI2041
aI4707
aI15557
aI28885
aI5959
aI22517
aI237131
asS'd'
I14297
sa(dp28594
S'a'
I15010
sS's'
I18
sS'm'
(lp28595
I1
aI3
aI7
aI11
aI25
aI33
aI105
aI15
aI245
aI247
aI1357
aI1255
aI7463
aI4815
aI13727
aI41687
aI112425
aI58827
asS'd'
I14298
sa(dp28596
S'a'
I15012
sS's'
I18
sS'm'
(lp28597
I1
aI1
aI7
aI1
aI19
aI31
aI37
aI201
aI217
aI127
aI927
aI763
aI6359
aI9951
aI2581
aI49171
aI104305
aI215923
asS'd'
I14299
sa(dp28598
S'a'
I15033
sS's'
I18
sS'm'
(lp28599
I1
aI1
aI7
aI7
aI13
aI9
aI9
aI139
aI363
aI85
aI1703
aI3615
aI2545
aI15991
aI20677
aI12109
aI54951
aI2171
asS'd'
I14300
sa(dp28600
S'a'
I15041
sS's'
I18
sS'm'
(lp28601
I1
aI3
aI1
aI13
aI27
aI3
aI37
aI195
aI185
aI829
aI815
aI1621
aI2917
aI8643
aI29071
aI45523
aI38475
aI243505
asS'd'
I14301
sa(dp28602
S'a'
I15047
sS's'
I18
sS'm'
(lp28603
I1
aI1
aI7
aI3
aI19
aI29
aI91
aI85
aI331
aI231
aI1609
aI2583
aI1091
aI4191
aI29929
aI55377
aI105077
aI168425
asS'd'
I14302
sa(dp28604
S'a'
I15066
sS's'
I18
sS'm'
(lp28605
I1
aI1
aI5
aI3
aI29
aI35
aI3
aI61
aI389
aI339
aI705
aI473
aI2075
aI7373
aI9699
aI38809
aI60415
aI66423
asS'd'
I14303
sa(dp28606
S'a'
I15068
sS's'
I18
sS'm'
(lp28607
I1
aI3
aI1
aI5
aI17
aI25
aI17
aI37
aI335
aI787
aI1891
aI1861
aI4325
aI12721
aI9675
aI13671
aI18655
aI235443
asS'd'
I14304
sa(dp28608
S'a'
I15072
sS's'
I18
sS'm'
(lp28609
I1
aI3
aI3
aI5
aI23
aI13
aI83
aI191
aI263
aI325
aI1847
aI1717
aI7089
aI15709
aI26567
aI44489
aI66523
aI3107
asS'd'
I14305
sa(dp28610
S'a'
I15095
sS's'
I18
sS'm'
(lp28611
I1
aI3
aI7
aI7
aI25
aI29
aI63
aI55
aI9
aI481
aI899
aI669
aI5481
aI11227
aI1637
aI17017
aI124509
aI102775
asS'd'
I14306
sa(dp28612
S'a'
I15150
sS's'
I18
sS'm'
(lp28613
I1
aI1
aI1
aI5
aI21
aI41
aI101
aI93
aI129
aI1023
aI561
aI2969
aI1525
aI2929
aI32729
aI54513
aI4359
aI28745
asS'd'
I14307
sa(dp28614
S'a'
I15152
sS's'
I18
sS'm'
(lp28615
I1
aI1
aI5
aI13
aI15
aI13
aI79
aI9
aI257
aI535
aI861
aI2703
aI6161
aI6659
aI10369
aI7
aI117467
aI146651
asS'd'
I14308
sa(dp28616
S'a'
I15175
sS's'
I18
sS'm'
(lp28617
I1
aI1
aI3
aI13
aI31
aI11
aI43
aI95
aI441
aI921
aI1323
aI343
aI5339
aI13149
aI19643
aI24253
aI32055
aI180327
asS'd'
I14309
sa(dp28618
S'a'
I15176
sS's'
I18
sS'm'
(lp28619
I1
aI1
aI5
aI15
aI11
aI27
aI109
aI149
aI255
aI1021
aI249
aI1913
aI5213
aI301
aI9939
aI49779
aI26097
aI66007
asS'd'
I14310
sa(dp28620
S'a'
I15210
sS's'
I18
sS'm'
(lp28621
I1
aI3
aI1
aI11
aI15
aI33
aI53
aI159
aI433
aI167
aI1107
aI3577
aI6231
aI8309
aI28125
aI55381
aI127309
aI14459
asS'd'
I14311
sa(dp28622
S'a'
I15245
sS's'
I18
sS'm'
(lp28623
I1
aI3
aI5
aI9
aI7
aI3
aI45
aI139
aI133
aI359
aI537
aI805
aI3931
aI5181
aI915
aI63317
aI86227
aI231573
asS'd'
I14312
sa(dp28624
S'a'
I15258
sS's'
I18
sS'm'
(lp28625
I1
aI1
aI1
aI3
aI11
aI31
aI97
aI127
aI117
aI151
aI711
aI2457
aI2777
aI3855
aI21829
aI7913
aI30785
aI141449
asS'd'
I14313
sa(dp28626
S'a'
I15263
sS's'
I18
sS'm'
(lp28627
I1
aI3
aI7
aI13
aI11
aI17
aI65
aI63
aI289
aI851
aI1929
aI4021
aI105
aI5207
aI17085
aI64119
aI48659
aI31687
asS'd'
I14314
sa(dp28628
S'a'
I15264
sS's'
I18
sS'm'
(lp28629
I1
aI1
aI7
aI11
aI31
aI57
aI63
aI251
aI341
aI727
aI505
aI1851
aI783
aI16191
aI9335
aI39421
aI14793
aI238215
asS'd'
I14315
sa(dp28630
S'a'
I15279
sS's'
I18
sS'm'
(lp28631
I1
aI3
aI3
aI11
aI23
aI13
aI119
aI195
aI117
aI579
aI693
aI3059
aI2967
aI12791
aI26905
aI28527
aI13393
aI11869
asS'd'
I14316
sa(dp28632
S'a'
I15301
sS's'
I18
sS'm'
(lp28633
I1
aI3
aI7
aI11
aI23
aI7
aI61
aI143
aI409
aI309
aI651
aI3321
aI4027
aI1351
aI10339
aI18451
aI18447
aI235665
asS'd'
I14317
sa(dp28634
S'a'
I15306
sS's'
I18
sS'm'
(lp28635
I1
aI1
aI1
aI9
aI13
aI21
aI7
aI65
aI103
aI789
aI973
aI475
aI2831
aI13337
aI18989
aI40573
aI105375
aI2221
asS'd'
I14318
sa(dp28636
S'a'
I15313
sS's'
I18
sS'm'
(lp28637
I1
aI3
aI5
aI15
aI3
aI59
aI115
aI61
aI365
aI653
aI523
aI3927
aI8175
aI6751
aI32561
aI55919
aI64903
aI139005
asS'd'
I14319
sa(dp28638
S'a'
I15329
sS's'
I18
sS'm'
(lp28639
I1
aI3
aI3
aI13
aI1
aI7
aI51
aI63
aI179
aI525
aI1899
aI373
aI3797
aI6329
aI5539
aI32669
aI65903
aI154993
asS'd'
I14320
sa(dp28640
S'a'
I15341
sS's'
I18
sS'm'
(lp28641
I1
aI3
aI1
aI1
aI31
aI53
aI87
aI39
aI317
aI71
aI1899
aI925
aI4719
aI11645
aI27125
aI50391
aI116491
aI219271
asS'd'
I14321
sa(dp28642
S'a'
I15367
sS's'
I18
sS'm'
(lp28643
I1
aI3
aI7
aI13
aI7
aI23
aI1
aI57
aI333
aI277
aI893
aI3245
aI1417
aI13115
aI21835
aI25879
aI91305
aI54691
asS'd'
I14322
sa(dp28644
S'a'
I15368
sS's'
I18
sS'm'
(lp28645
I1
aI3
aI1
aI11
aI27
aI5
aI109
aI69
aI221
aI453
aI299
aI517
aI609
aI11959
aI27789
aI33107
aI46559
aI121673
asS'd'
I14323
sa(dp28646
S'a'
I15386
sS's'
I18
sS'm'
(lp28647
I1
aI3
aI5
aI7
aI27
aI7
aI119
aI169
aI129
aI643
aI173
aI2479
aI6163
aI11159
aI11897
aI57153
aI11347
aI135337
asS'd'
I14324
sa(dp28648
S'a'
I15404
sS's'
I18
sS'm'
(lp28649
I1
aI1
aI5
aI13
aI13
aI59
aI21
aI13
aI429
aI601
aI267
aI1635
aI2579
aI12053
aI31583
aI14847
aI78187
aI217099
asS'd'
I14325
sa(dp28650
S'a'
I15422
sS's'
I18
sS'm'
(lp28651
I1
aI3
aI5
aI9
aI5
aI3
aI125
aI159
aI411
aI15
aI479
aI933
aI6307
aI9707
aI23491
aI6501
aI70993
aI161365
asS'd'
I14326
sa(dp28652
S'a'
I15436
sS's'
I18
sS'm'
(lp28653
I1
aI1
aI7
aI7
aI3
aI33
aI87
aI177
aI283
aI825
aI1935
aI1545
aI7071
aI9975
aI1795
aI48277
aI115725
aI173439
asS'd'
I14327
sa(dp28654
S'a'
I15444
sS's'
I18
sS'm'
(lp28655
I1
aI3
aI3
aI9
aI19
aI63
aI17
aI119
aI13
aI337
aI2021
aI2221
aI3237
aI3253
aI18661
aI15479
aI59377
aI76095
asS'd'
I14328
sa(dp28656
S'a'
I15467
sS's'
I18
sS'm'
(lp28657
I1
aI1
aI1
aI11
aI17
aI15
aI93
aI249
aI333
aI171
aI575
aI3251
aI5413
aI3587
aI22807
aI29273
aI56461
aI97801
asS'd'
I14329
sa(dp28658
S'a'
I15498
sS's'
I18
sS'm'
(lp28659
I1
aI3
aI3
aI11
aI13
aI7
aI27
aI167
aI389
aI693
aI1473
aI555
aI1603
aI3167
aI3985
aI3841
aI100283
aI195253
asS'd'
I14330
sa(dp28660
S'a'
I15503
sS's'
I18
sS'm'
(lp28661
I1
aI3
aI7
aI1
aI23
aI7
aI89
aI231
aI85
aI797
aI1935
aI2557
aI4365
aI2221
aI21069
aI44055
aI77723
aI226547
asS'd'
I14331
sa(dp28662
S'a'
I15528
sS's'
I18
sS'm'
(lp28663
I1
aI3
aI1
aI5
aI5
aI49
aI47
aI187
aI71
aI903
aI1279
aI3219
aI8041
aI10915
aI5249
aI17755
aI80077
aI3479
asS'd'
I14332
sa(dp28664
S'a'
I15534
sS's'
I18
sS'm'
(lp28665
I1
aI1
aI3
aI13
aI5
aI53
aI35
aI25
aI183
aI791
aI1651
aI1041
aI1221
aI2171
aI26221
aI20649
aI126851
aI163047
asS'd'
I14333
sa(dp28666
S'a'
I15536
sS's'
I18
sS'm'
(lp28667
I1
aI1
aI5
aI9
aI29
aI3
aI75
aI31
aI385
aI293
aI171
aI3023
aI2075
aI14541
aI30879
aI13895
aI67637
aI87831
asS'd'
I14334
sa(dp28668
S'a'
I15559
sS's'
I18
sS'm'
(lp28669
I1
aI3
aI5
aI7
aI3
aI41
aI115
aI213
aI23
aI895
aI361
aI27
aI5839
aI12447
aI13829
aI29183
aI106539
aI134891
asS'd'
I14335
sa(dp28670
S'a'
I15565
sS's'
I18
sS'm'
(lp28671
I1
aI3
aI7
aI7
aI11
aI39
aI99
aI229
aI195
aI633
aI837
aI3697
aI1161
aI15119
aI20831
aI27371
aI92195
aI26993
asS'd'
I14336
sa(dp28672
S'a'
I15583
sS's'
I18
sS'm'
(lp28673
I1
aI1
aI5
aI9
aI25
aI17
aI5
aI169
aI475
aI73
aI1451
aI2057
aI3671
aI12801
aI9671
aI57427
aI25321
aI154969
asS'd'
I14337
sa(dp28674
S'a'
I15599
sS's'
I18
sS'm'
(lp28675
I1
aI3
aI5
aI11
aI25
aI23
aI9
aI145
aI341
aI339
aI1855
aI981
aI8041
aI569
aI19851
aI29521
aI21767
aI136505
asS'd'
I14338
sa(dp28676
S'a'
I15602
sS's'
I18
sS'm'
(lp28677
I1
aI1
aI1
aI1
aI3
aI9
aI101
aI253
aI475
aI529
aI387
aI1893
aI5509
aI5763
aI29555
aI13307
aI30001
aI105057
asS'd'
I14339
sa(dp28678
S'a'
I15616
sS's'
I18
sS'm'
(lp28679
I1
aI3
aI3
aI5
aI17
aI23
aI127
aI161
aI375
aI817
aI1229
aI1197
aI1097
aI3053
aI14351
aI21213
aI12501
aI137397
asS'd'
I14340
sa(dp28680
S'a'
I15626
sS's'
I18
sS'm'
(lp28681
I1
aI1
aI1
aI9
aI7
aI1
aI57
aI185
aI281
aI65
aI181
aI2483
aI4739
aI4353
aI29837
aI40613
aI32489
aI23317
asS'd'
I14341
sa(dp28682
S'a'
I15667
sS's'
I18
sS'm'
(lp28683
I1
aI1
aI3
aI9
aI5
aI35
aI43
aI191
aI409
aI95
aI537
aI2465
aI515
aI1633
aI20887
aI32535
aI43863
aI199885
asS'd'
I14342
sa(dp28684
S'a'
I15684
sS's'
I18
sS'm'
(lp28685
I1
aI3
aI3
aI13
aI19
aI49
aI41
aI51
aI3
aI979
aI1623
aI3323
aI7711
aI3707
aI29417
aI58011
aI114467
aI227499
asS'd'
I14343
sa(dp28686
S'a'
I15711
sS's'
I18
sS'm'
(lp28687
I1
aI1
aI1
aI11
aI21
aI7
aI23
aI51
aI39
aI25
aI1971
aI213
aI369
aI9149
aI12845
aI57631
aI16597
aI22031
asS'd'
I14344
sa(dp28688
S'a'
I15745
sS's'
I18
sS'm'
(lp28689
I1
aI3
aI3
aI11
aI27
aI59
aI71
aI37
aI461
aI353
aI2041
aI1961
aI4643
aI6953
aI18129
aI60337
aI82769
aI20819
asS'd'
I14345
sa(dp28690
S'a'
I15763
sS's'
I18
sS'm'
(lp28691
I1
aI1
aI3
aI11
aI25
aI19
aI17
aI5
aI503
aI227
aI2021
aI733
aI2867
aI201
aI25779
aI49811
aI81167
aI95437
asS'd'
I14346
sa(dp28692
S'a'
I15779
sS's'
I18
sS'm'
(lp28693
I1
aI3
aI3
aI15
aI7
aI53
aI35
aI143
aI27
aI937
aI215
aI3249
aI4151
aI1933
aI25267
aI18047
aI35131
aI25903
asS'd'
I14347
sa(dp28694
S'a'
I15811
sS's'
I18
sS'm'
(lp28695
I1
aI1
aI1
aI3
aI3
aI39
aI71
aI99
aI291
aI97
aI1389
aI3803
aI2881
aI9765
aI11277
aI20071
aI15133
aI37349
asS'd'
I14348
sa(dp28696
S'a'
I15814
sS's'
I18
sS'm'
(lp28697
I1
aI1
aI7
aI11
aI9
aI55
aI1
aI241
aI391
aI935
aI1555
aI3585
aI1807
aI10057
aI2633
aI14023
aI14409
aI199643
asS'd'
I14349
sa(dp28698
S'a'
I15817
sS's'
I18
sS'm'
(lp28699
I1
aI3
aI3
aI3
aI19
aI9
aI57
aI237
aI107
aI869
aI147
aI2673
aI5271
aI8999
aI20723
aI63017
aI75989
aI20131
asS'd'
I14350
sa(dp28700
S'a'
I15853
sS's'
I18
sS'm'
(lp28701
I1
aI3
aI3
aI3
aI25
aI11
aI61
aI77
aI119
aI657
aI2011
aI3489
aI7835
aI4473
aI2531
aI65231
aI104797
aI161443
asS'd'
I14351
sa(dp28702
S'a'
I15881
sS's'
I18
sS'm'
(lp28703
I1
aI1
aI5
aI5
aI11
aI63
aI25
aI93
aI181
aI797
aI367
aI3357
aI5291
aI5087
aI28661
aI34093
aI75195
aI165345
asS'd'
I14352
sa(dp28704
S'a'
I15890
sS's'
I18
sS'm'
(lp28705
I1
aI1
aI1
aI7
aI17
aI1
aI77
aI149
aI59
aI633
aI1551
aI1305
aI7677
aI8671
aI17457
aI64037
aI104451
aI112387
asS'd'
I14353
sa(dp28706
S'a'
I15899
sS's'
I18
sS'm'
(lp28707
I1
aI3
aI1
aI1
aI15
aI33
aI37
aI187
aI247
aI261
aI1101
aI3451
aI7747
aI12197
aI22465
aI30589
aI12573
aI204517
asS'd'
I14354
sa(dp28708
S'a'
I15905
sS's'
I18
sS'm'
(lp28709
I1
aI3
aI3
aI11
aI3
aI39
aI71
aI139
aI145
aI139
aI101
aI2815
aI3457
aI14033
aI4531
aI42133
aI54147
aI71259
asS'd'
I14355
sa(dp28710
S'a'
I15915
sS's'
I18
sS'm'
(lp28711
I1
aI3
aI1
aI1
aI23
aI37
aI19
aI113
aI443
aI57
aI439
aI2929
aI3835
aI5431
aI11189
aI4539
aI72531
aI124453
asS'd'
I14356
sa(dp28712
S'a'
I15937
sS's'
I18
sS'm'
(lp28713
I1
aI3
aI3
aI5
aI3
aI17
aI21
aI217
aI41
aI665
aI1565
aI3753
aI5289
aI9789
aI29205
aI16453
aI88979
aI171387
asS'd'
I14357
sa(dp28714
S'a'
I15950
sS's'
I18
sS'm'
(lp28715
I1
aI3
aI3
aI13
aI27
aI15
aI15
aI223
aI231
aI311
aI311
aI1143
aI8113
aI13863
aI3191
aI51103
aI109437
aI245557
asS'd'
I14358
sa(dp28716
S'a'
I16002
sS's'
I18
sS'm'
(lp28717
I1
aI1
aI3
aI13
aI11
aI59
aI7
aI191
aI477
aI683
aI353
aI2845
aI7623
aI9035
aI453
aI48429
aI40111
aI162859
asS'd'
I14359
sa(dp28718
S'a'
I16041
sS's'
I18
sS'm'
(lp28719
I1
aI3
aI7
aI5
aI29
aI37
aI55
aI59
aI259
aI851
aI861
aI1951
aI7847
aI8537
aI30107
aI2999
aI59137
aI155615
asS'd'
I14360
sa(dp28720
S'a'
I16042
sS's'
I18
sS'm'
(lp28721
I1
aI3
aI7
aI11
aI3
aI13
aI73
aI147
aI393
aI327
aI1289
aI37
aI795
aI1413
aI19215
aI28345
aI124301
aI23135
asS'd'
I14361
sa(dp28722
S'a'
I16052
sS's'
I18
sS'm'
(lp28723
I1
aI1
aI5
aI11
aI29
aI17
aI107
aI69
aI433
aI845
aI1351
aI2551
aI807
aI15315
aI15511
aI39475
aI84879
aI129405
asS'd'
I14362
sa(dp28724
S'a'
I16121
sS's'
I18
sS'm'
(lp28725
I1
aI3
aI3
aI9
aI15
aI3
aI23
aI5
aI211
aI871
aI689
aI2319
aI39
aI2215
aI25171
aI43169
aI113715
aI186049
asS'd'
I14363
sa(dp28726
S'a'
I16132
sS's'
I18
sS'm'
(lp28727
I1
aI1
aI3
aI7
aI3
aI37
aI23
aI9
aI453
aI649
aI373
aI1273
aI1539
aI6221
aI27469
aI44675
aI13513
aI131179
asS'd'
I14364
sa(dp28728
S'a'
I16136
sS's'
I18
sS'm'
(lp28729
I1
aI3
aI1
aI5
aI29
aI41
aI119
aI133
aI37
aI761
aI1193
aI2311
aI4945
aI7337
aI17027
aI12873
aI51489
aI160633
asS'd'
I14365
sa(dp28730
S'a'
I16150
sS's'
I18
sS'm'
(lp28731
I1
aI3
aI1
aI3
aI21
aI63
aI75
aI115
aI105
aI223
aI933
aI445
aI5789
aI4611
aI13609
aI2873
aI16679
aI222895
asS'd'
I14366
sa(dp28732
S'a'
I16153
sS's'
I18
sS'm'
(lp28733
I1
aI1
aI7
aI5
aI17
aI13
aI15
aI217
aI193
aI863
aI1319
aI2337
aI3055
aI14879
aI8669
aI5705
aI42965
aI166443
asS'd'
I14367
sa(dp28734
S'a'
I16180
sS's'
I18
sS'm'
(lp28735
I1
aI1
aI7
aI11
aI3
aI55
aI57
aI131
aI289
aI843
aI1693
aI881
aI6737
aI5557
aI18365
aI12393
aI38479
aI189177
asS'd'
I14368
sa(dp28736
S'a'
I16190
sS's'
I18
sS'm'
(lp28737
I1
aI1
aI3
aI5
aI3
aI59
aI13
aI123
aI397
aI479
aI79
aI569
aI535
aI2529
aI26225
aI43475
aI76925
aI187763
asS'd'
I14369
sa(dp28738
S'a'
I16192
sS's'
I18
sS'm'
(lp28739
I1
aI1
aI5
aI15
aI15
aI37
aI1
aI97
aI489
aI331
aI1499
aI1759
aI3621
aI5373
aI1425
aI6477
aI45805
aI235511
asS'd'
I14370
sa(dp28740
S'a'
I16195
sS's'
I18
sS'm'
(lp28741
I1
aI3
aI1
aI3
aI7
aI51
aI55
aI157
aI61
aI751
aI1881
aI4093
aI2557
aI11129
aI23239
aI16335
aI8949
aI205007
asS'd'
I14371
sa(dp28742
S'a'
I16210
sS's'
I18
sS'm'
(lp28743
I1
aI1
aI1
aI1
aI13
aI21
aI67
aI141
aI85
aI1023
aI223
aI747
aI1951
aI10279
aI6399
aI49887
aI100437
aI76757
asS'd'
I14372
sa(dp28744
S'a'
I16225
sS's'
I18
sS'm'
(lp28745
I1
aI1
aI3
aI11
aI1
aI51
aI29
aI33
aI173
aI769
aI879
aI2883
aI417
aI15031
aI13377
aI63919
aI118803
aI87969
asS'd'
I14373
sa(dp28746
S'a'
I16256
sS's'
I18
sS'm'
(lp28747
I1
aI1
aI1
aI5
aI1
aI1
aI17
aI153
aI81
aI691
aI961
aI3399
aI5005
aI10617
aI18467
aI13775
aI34905
aI241349
asS'd'
I14374
sa(dp28748
S'a'
I16266
sS's'
I18
sS'm'
(lp28749
I1
aI1
aI1
aI13
aI7
aI37
aI57
aI187
aI389
aI575
aI1827
aI2017
aI4541
aI10513
aI23409
aI30945
aI126855
aI239657
asS'd'
I14375
sa(dp28750
S'a'
I16274
sS's'
I18
sS'm'
(lp28751
I1
aI1
aI5
aI5
aI17
aI41
aI83
aI177
aI285
aI695
aI29
aI1653
aI953
aI6377
aI13571
aI58663
aI9265
aI100759
asS'd'
I14376
sa(dp28752
S'a'
I16302
sS's'
I18
sS'm'
(lp28753
I1
aI3
aI5
aI3
aI5
aI13
aI27
aI153
aI207
aI699
aI1805
aI947
aI979
aI2719
aI389
aI61953
aI16991
aI160073
asS'd'
I14377
sa(dp28754
S'a'
I16310
sS's'
I18
sS'm'
(lp28755
I1
aI1
aI7
aI13
aI17
aI37
aI113
aI185
aI239
aI455
aI1557
aI3201
aI1111
aI4875
aI23197
aI41883
aI70507
aI255047
asS'd'
I14378
sa(dp28756
S'a'
I16322
sS's'
I18
sS'm'
(lp28757
I1
aI3
aI5
aI11
aI9
aI51
aI47
aI137
aI413
aI1015
aI259
aI1829
aI6043
aI11757
aI22317
aI15155
aI107827
aI171003
asS'd'
I14379
sa(dp28758
S'a'
I16339
sS's'
I18
sS'm'
(lp28759
I1
aI1
aI1
aI13
aI27
aI7
aI49
aI91
aI285
aI13
aI2007
aI3469
aI1223
aI2483
aI16155
aI8413
aI10529
aI224195
asS'd'
I14380
sa(dp28760
S'a'
I16345
sS's'
I18
sS'm'
(lp28761
I1
aI1
aI7
aI3
aI9
aI49
aI119
aI81
aI331
aI187
aI1695
aI1729
aI533
aI6359
aI7053
aI34665
aI37541
aI100225
asS'd'
I14381
sa(dp28762
S'a'
I16348
sS's'
I18
sS'm'
(lp28763
I1
aI3
aI7
aI1
aI7
aI35
aI115
aI91
aI479
aI515
aI1249
aI121
aI2885
aI16383
aI1777
aI44205
aI86459
aI255885
asS'd'
I14382
sa(dp28764
S'a'
I16357
sS's'
I18
sS'm'
(lp28765
I1
aI1
aI7
aI13
aI13
aI27
aI11
aI49
aI221
aI829
aI1787
aI2889
aI3875
aI1679
aI25333
aI1323
aI9813
aI189373
asS'd'
I14383
sa(dp28766
S'a'
I16382
sS's'
I18
sS'm'
(lp28767
I1
aI3
aI7
aI5
aI31
aI5
aI117
aI77
aI209
aI619
aI191
aI2969
aI2221
aI15339
aI11461
aI64201
aI130461
aI204467
asS'd'
I14384
sa(dp28768
S'a'
I16402
sS's'
I18
sS'm'
(lp28769
I1
aI3
aI1
aI1
aI29
aI5
aI91
aI31
aI313
aI901
aI1501
aI2837
aI3615
aI7765
aI341
aI13873
aI21663
aI260637
asS'd'
I14385
sa(dp28770
S'a'
I16461
sS's'
I18
sS'm'
(lp28771
I1
aI1
aI1
aI9
aI1
aI41
aI97
aI15
aI141
aI901
aI1309
aI3341
aI4871
aI16033
aI12343
aI1555
aI94989
aI78295
asS'd'
I14386
sa(dp28772
S'a'
I16469
sS's'
I18
sS'm'
(lp28773
I1
aI1
aI7
aI3
aI3
aI15
aI1
aI29
aI445
aI59
aI475
aI3033
aI4227
aI3219
aI6093
aI58953
aI92179
aI49343
asS'd'
I14387
sa(dp28774
S'a'
I16474
sS's'
I18
sS'm'
(lp28775
I1
aI1
aI5
aI3
aI27
aI25
aI109
aI13
aI219
aI983
aI131
aI2517
aI1161
aI16063
aI32737
aI6077
aI91183
aI37457
asS'd'
I14388
sa(dp28776
S'a'
I16483
sS's'
I18
sS'm'
(lp28777
I1
aI3
aI3
aI1
aI3
aI1
aI85
aI147
aI17
aI543
aI1475
aI3873
aI3719
aI2737
aI30977
aI15953
aI66077
aI258979
asS'd'
I14389
sa(dp28778
S'a'
I16497
sS's'
I18
sS'm'
(lp28779
I1
aI3
aI5
aI1
aI29
aI9
aI21
aI51
aI5
aI985
aI1177
aI3287
aI2183
aI7301
aI13713
aI53403
aI38439
aI195863
asS'd'
I14390
sa(dp28780
S'a'
I16523
sS's'
I18
sS'm'
(lp28781
I1
aI1
aI7
aI15
aI31
aI53
aI47
aI173
aI477
aI439
aI751
aI1019
aI3371
aI9319
aI17995
aI29029
aI90657
aI209277
asS'd'
I14391
sa(dp28782
S'a'
I16534
sS's'
I18
sS'm'
(lp28783
I1
aI1
aI5
aI5
aI17
aI5
aI59
aI115
aI375
aI231
aI1891
aI1321
aI3639
aI16117
aI32639
aI28793
aI68213
aI41091
asS'd'
I14392
sa(dp28784
S'a'
I16550
sS's'
I18
sS'm'
(lp28785
I1
aI3
aI5
aI11
aI17
aI15
aI13
aI11
aI459
aI767
aI849
aI1407
aI6611
aI6409
aI21515
aI63175
aI127155
aI171959
asS'd'
I14393
sa(dp28786
S'a'
I16564
sS's'
I18
sS'm'
(lp28787
I1
aI1
aI5
aI1
aI17
aI49
aI61
aI161
aI399
aI137
aI845
aI2673
aI2431
aI15343
aI389
aI42337
aI23031
aI94811
asS'd'
I14394
sa(dp28788
S'a'
I16582
sS's'
I18
sS'm'
(lp28789
I1
aI1
aI1
aI9
aI21
aI23
aI75
aI177
aI351
aI197
aI1619
aI2443
aI6829
aI3773
aI16399
aI31949
aI44975
aI221363
asS'd'
I14395
sa(dp28790
S'a'
I16591
sS's'
I18
sS'm'
(lp28791
I1
aI1
aI3
aI1
aI11
aI19
aI103
aI61
aI135
aI863
aI1427
aI2657
aI4553
aI1277
aI20249
aI3973
aI25467
aI18847
asS'd'
I14396
sa(dp28792
S'a'
I16609
sS's'
I18
sS'm'
(lp28793
I1
aI3
aI3
aI13
aI17
aI31
aI19
aI163
aI323
aI195
aI603
aI4069
aI3181
aI12069
aI22117
aI44229
aI23585
aI202785
asS'd'
I14397
sa(dp28794
S'a'
I16642
sS's'
I18
sS'm'
(lp28795
I1
aI1
aI7
aI5
aI17
aI3
aI77
aI111
aI491
aI829
aI1375
aI2829
aI5599
aI14057
aI21387
aI52345
aI108281
aI211285
asS'd'
I14398
sa(dp28796
S'a'
I16648
sS's'
I18
sS'm'
(lp28797
I1
aI3
aI3
aI1
aI17
aI43
aI71
aI13
aI321
aI393
aI1803
aI727
aI5101
aI13485
aI8693
aI60505
aI13893
aI3467
asS'd'
I14399
sa(dp28798
S'a'
I16656
sS's'
I18
sS'm'
(lp28799
I1
aI1
aI5
aI1
aI23
aI31
aI121
aI15
aI215
aI215
aI1113
aI3335
aI7431
aI4863
aI31429
aI49903
aI59403
aI60797
asS'd'
I14400
sa(dp28800
S'a'
I16662
sS's'
I18
sS'm'
(lp28801
I1
aI3
aI3
aI9
aI21
aI43
aI61
aI171
aI361
aI323
aI1895
aI3647
aI729
aI8809
aI9351
aI14573
aI93593
aI17485
asS'd'
I14401
sa(dp28802
S'a'
I16665
sS's'
I18
sS'm'
(lp28803
I1
aI3
aI3
aI7
aI7
aI19
aI45
aI247
aI203
aI757
aI1941
aI3753
aI5317
aI13239
aI18945
aI26173
aI43929
aI66889
asS'd'
I14402
sa(dp28804
S'a'
I16678
sS's'
I18
sS'm'
(lp28805
I1
aI1
aI1
aI15
aI5
aI17
aI11
aI21
aI193
aI941
aI517
aI191
aI6067
aI8403
aI27339
aI31035
aI34767
aI28675
asS'd'
I14403
sa(dp28806
S'a'
I16701
sS's'
I18
sS'm'
(lp28807
I1
aI3
aI1
aI7
aI27
aI59
aI27
aI7
aI491
aI551
aI867
aI3693
aI391
aI9799
aI11051
aI28347
aI57555
aI23079
asS'd'
I14404
sa(dp28808
S'a'
I16713
sS's'
I18
sS'm'
(lp28809
I1
aI3
aI3
aI1
aI25
aI21
aI63
aI253
aI459
aI603
aI107
aI1229
aI1433
aI4263
aI24341
aI20493
aI40165
aI254725
asS'd'
I14405
sa(dp28810
S'a'
I16716
sS's'
I18
sS'm'
(lp28811
I1
aI3
aI5
aI3
aI9
aI7
aI63
aI195
aI19
aI973
aI47
aI811
aI2207
aI3613
aI8911
aI17495
aI62403
aI77951
asS'd'
I14406
sa(dp28812
S'a'
I16758
sS's'
I18
sS'm'
(lp28813
I1
aI3
aI5
aI11
aI5
aI13
aI83
aI125
aI467
aI111
aI1819
aI3807
aI4259
aI2885
aI29577
aI13539
aI69859
aI97379
asS'd'
I14407
sa(dp28814
S'a'
I16768
sS's'
I18
sS'm'
(lp28815
I1
aI1
aI5
aI15
aI5
aI33
aI109
aI203
aI129
aI587
aI9
aI3025
aI2839
aI11405
aI11257
aI7779
aI30311
aI14015
asS'd'
I14408
sa(dp28816
S'a'
I16797
sS's'
I18
sS'm'
(lp28817
I1
aI1
aI7
aI15
aI5
aI47
aI103
aI199
aI391
aI61
aI129
aI3511
aI1295
aI15067
aI23919
aI2941
aI120463
aI21665
asS'd'
I14409
sa(dp28818
S'a'
I16804
sS's'
I18
sS'm'
(lp28819
I1
aI1
aI5
aI5
aI7
aI7
aI125
aI153
aI365
aI815
aI1423
aI4053
aI875
aI2405
aI21291
aI26785
aI31371
aI211045
asS'd'
I14410
sa(dp28820
S'a'
I16811
sS's'
I18
sS'm'
(lp28821
I1
aI3
aI3
aI9
aI31
aI47
aI1
aI247
aI197
aI1019
aI985
aI2277
aI875
aI3969
aI15093
aI15561
aI110101
aI156547
asS'd'
I14411
sa(dp28822
S'a'
I16813
sS's'
I18
sS'm'
(lp28823
I1
aI1
aI3
aI1
aI9
aI47
aI71
aI125
aI17
aI501
aI1783
aI2337
aI483
aI12719
aI22453
aI16701
aI102639
aI152955
asS'd'
I14412
sa(dp28824
S'a'
I16881
sS's'
I18
sS'm'
(lp28825
I1
aI1
aI3
aI11
aI13
aI31
aI9
aI63
aI261
aI257
aI319
aI1443
aI5011
aI9799
aI18639
aI53081
aI56879
aI102335
asS'd'
I14413
sa(dp28826
S'a'
I16888
sS's'
I18
sS'm'
(lp28827
I1
aI3
aI3
aI1
aI5
aI59
aI127
aI163
aI323
aI997
aI1755
aI1445
aI2285
aI4935
aI22123
aI815
aI115131
aI1009
asS'd'
I14414
sa(dp28828
S'a'
I16893
sS's'
I18
sS'm'
(lp28829
I1
aI1
aI3
aI11
aI11
aI43
aI65
aI127
aI137
aI583
aI173
aI2601
aI5809
aI15773
aI16129
aI2543
aI68281
aI96107
asS'd'
I14415
sa(dp28830
S'a'
I16922
sS's'
I18
sS'm'
(lp28831
I1
aI3
aI5
aI9
aI25
aI25
aI95
aI73
aI313
aI893
aI1805
aI2301
aI5917
aI15159
aI8637
aI25505
aI66053
aI31627
asS'd'
I14416
sa(dp28832
S'a'
I16937
sS's'
I18
sS'm'
(lp28833
I1
aI1
aI3
aI7
aI25
aI59
aI55
aI13
aI297
aI849
aI187
aI359
aI3745
aI12655
aI29293
aI58581
aI89799
aI195867
asS'd'
I14417
sa(dp28834
S'a'
I16958
sS's'
I18
sS'm'
(lp28835
I1
aI1
aI5
aI13
aI19
aI5
aI51
aI85
aI259
aI59
aI1003
aI2991
aI6605
aI8405
aI5221
aI45607
aI130729
aI99641
asS'd'
I14418
sa(dp28836
S'a'
I16965
sS's'
I18
sS'm'
(lp28837
I1
aI3
aI5
aI5
aI25
aI61
aI51
aI211
aI143
aI233
aI1465
aI1165
aI1769
aI3021
aI9491
aI30335
aI34787
aI142605
asS'd'
I14419
sa(dp28838
S'a'
I16978
sS's'
I18
sS'm'
(lp28839
I1
aI3
aI5
aI13
aI23
aI9
aI89
aI249
aI71
aI179
aI841
aI3375
aI21
aI6757
aI27495
aI7531
aI123725
aI253855
asS'd'
I14420
sa(dp28840
S'a'
I16980
sS's'
I18
sS'm'
(lp28841
I1
aI3
aI7
aI13
aI9
aI33
aI109
aI103
aI475
aI781
aI493
aI2079
aI6529
aI13443
aI2181
aI26925
aI31345
aI142863
asS'd'
I14421
sa(dp28842
S'a'
I16989
sS's'
I18
sS'm'
(lp28843
I1
aI1
aI7
aI9
aI15
aI41
aI17
aI85
aI503
aI839
aI533
aI731
aI2735
aI12949
aI11395
aI22539
aI130147
aI40045
asS'd'
I14422
sa(dp28844
S'a'
I17011
sS's'
I18
sS'm'
(lp28845
I1
aI1
aI5
aI9
aI25
aI21
aI29
aI79
aI405
aI383
aI1271
aI385
aI7629
aI3889
aI5319
aI57739
aI51411
aI50895
asS'd'
I14423
sa(dp28846
S'a'
I17014
sS's'
I18
sS'm'
(lp28847
I1
aI3
aI7
aI5
aI13
aI35
aI17
aI97
aI261
aI437
aI951
aI1403
aI2407
aI11447
aI13565
aI10165
aI100001
aI253093
asS'd'
I14424
sa(dp28848
S'a'
I17023
sS's'
I18
sS'm'
(lp28849
I1
aI3
aI1
aI3
aI11
aI39
aI31
aI187
aI473
aI565
aI351
aI4007
aI2621
aI14463
aI9009
aI40679
aI81069
aI51131
asS'd'
I14425
sa(dp28850
S'a'
I17063
sS's'
I18
sS'm'
(lp28851
I1
aI3
aI5
aI5
aI9
aI17
aI11
aI151
aI59
aI249
aI281
aI203
aI6423
aI4977
aI18557
aI65383
aI88361
aI87437
asS'd'
I14426
sa(dp28852
S'a'
I17081
sS's'
I18
sS'm'
(lp28853
I1
aI3
aI5
aI7
aI15
aI25
aI3
aI157
aI179
aI439
aI1627
aI3493
aI6641
aI6403
aI2361
aI3613
aI33817
aI22585
asS'd'
I14427
sa(dp28854
S'a'
I17084
sS's'
I18
sS'm'
(lp28855
I1
aI3
aI5
aI1
aI13
aI63
aI77
aI195
aI233
aI175
aI631
aI1021
aI637
aI13231
aI26187
aI131
aI127379
aI256183
asS'd'
I14428
sa(dp28856
S'a'
I17099
sS's'
I18
sS'm'
(lp28857
I1
aI1
aI7
aI9
aI5
aI13
aI15
aI187
aI55
aI37
aI1113
aI2191
aI3439
aI1073
aI26239
aI3049
aI19807
aI250869
asS'd'
I14429
sa(dp28858
S'a'
I17126
sS's'
I18
sS'm'
(lp28859
I1
aI1
aI7
aI3
aI13
aI15
aI77
aI47
aI317
aI285
aI753
aI2419
aI7795
aI11423
aI6043
aI2913
aI42819
aI50603
asS'd'
I14430
sa(dp28860
S'a'
I17129
sS's'
I18
sS'm'
(lp28861
I1
aI3
aI7
aI15
aI21
aI17
aI63
aI71
aI97
aI535
aI1085
aI1531
aI5165
aI13717
aI1537
aI26797
aI111787
aI189403
asS'd'
I14431
sa(dp28862
S'a'
I17138
sS's'
I18
sS'm'
(lp28863
I1
aI1
aI1
aI7
aI9
aI3
aI43
aI209
aI385
aI851
aI1411
aI4039
aI3259
aI13387
aI24505
aI33325
aI83741
aI241255
asS'd'
I14432
sa(dp28864
S'a'
I17164
sS's'
I18
sS'm'
(lp28865
I1
aI1
aI1
aI1
aI3
aI41
aI13
aI43
aI303
aI445
aI1097
aI3517
aI7753
aI8459
aI3017
aI16385
aI13775
aI248655
asS'd'
I14433
sa(dp28866
S'a'
I17170
sS's'
I18
sS'm'
(lp28867
I1
aI3
aI5
aI15
aI5
aI61
aI31
aI57
aI269
aI931
aI1071
aI1137
aI6181
aI13005
aI18493
aI1345
aI105203
aI117309
asS'd'
I14434
sa(dp28868
S'a'
I17203
sS's'
I18
sS'm'
(lp28869
I1
aI1
aI3
aI13
aI21
aI29
aI3
aI179
aI367
aI155
aI993
aI117
aI5849
aI10181
aI1175
aI55769
aI16025
aI67669
asS'd'
I14435
sa(dp28870
S'a'
I17206
sS's'
I18
sS'm'
(lp28871
I1
aI1
aI3
aI9
aI11
aI9
aI33
aI131
aI181
aI1003
aI253
aI2759
aI1877
aI11851
aI22959
aI37823
aI82737
aI110329
asS'd'
I14436
sa(dp28872
S'a'
I17224
sS's'
I18
sS'm'
(lp28873
I1
aI1
aI1
aI5
aI7
aI5
aI107
aI191
aI385
aI129
aI567
aI2585
aI7295
aI3005
aI28185
aI7095
aI54851
aI257587
asS'd'
I14437
sa(dp28874
S'a'
I17305
sS's'
I18
sS'm'
(lp28875
I1
aI3
aI7
aI9
aI21
aI61
aI103
aI155
aI503
aI307
aI993
aI683
aI1491
aI14895
aI9213
aI34535
aI17765
aI12457
asS'd'
I14438
sa(dp28876
S'a'
I17322
sS's'
I18
sS'm'
(lp28877
I1
aI3
aI3
aI1
aI7
aI47
aI27
aI173
aI97
aI889
aI853
aI3995
aI4943
aI71
aI20479
aI16741
aI35479
aI35307
asS'd'
I14439
sa(dp28878
S'a'
I17327
sS's'
I18
sS'm'
(lp28879
I1
aI1
aI5
aI3
aI15
aI35
aI29
aI207
aI117
aI267
aI1835
aI2565
aI1199
aI3813
aI13999
aI10537
aI129915
aI210651
asS'd'
I14440
sa(dp28880
S'a'
I17332
sS's'
I18
sS'm'
(lp28881
I1
aI3
aI3
aI13
aI5
aI57
aI77
aI193
aI11
aI279
aI745
aI2511
aI5775
aI13527
aI26329
aI16303
aI111511
aI70025
asS'd'
I14441
sa(dp28882
S'a'
I17342
sS's'
I18
sS'm'
(lp28883
I1
aI3
aI7
aI9
aI17
aI1
aI73
aI1
aI125
aI939
aI863
aI2763
aI1951
aI3191
aI5567
aI59729
aI32149
aI149417
asS'd'
I14442
sa(dp28884
S'a'
I17387
sS's'
I18
sS'm'
(lp28885
I1
aI3
aI3
aI1
aI19
aI25
aI119
aI63
aI101
aI33
aI77
aI3587
aI6367
aI8275
aI24957
aI32087
aI7031
aI217291
asS'd'
I14443
sa(dp28886
S'a'
I17410
sS's'
I18
sS'm'
(lp28887
I1
aI1
aI1
aI13
aI7
aI25
aI75
aI161
aI143
aI353
aI973
aI2957
aI749
aI13519
aI11295
aI34287
aI60727
aI83731
asS'd'
I14444
sa(dp28888
S'a'
I17419
sS's'
I18
sS'm'
(lp28889
I1
aI3
aI7
aI7
aI17
aI9
aI59
aI45
aI97
aI619
aI895
aI1955
aI8143
aI2507
aI4673
aI39425
aI35679
aI152069
asS'd'
I14445
sa(dp28890
S'a'
I17429
sS's'
I18
sS'm'
(lp28891
I1
aI1
aI5
aI9
aI31
aI19
aI115
aI177
aI349
aI877
aI525
aI305
aI2187
aI12195
aI13529
aI61641
aI102293
aI69941
asS'd'
I14446
sa(dp28892
S'a'
I17439
sS's'
I18
sS'm'
(lp28893
I1
aI3
aI7
aI11
aI23
aI59
aI15
aI243
aI511
aI465
aI905
aI1979
aI2263
aI2105
aI9009
aI3691
aI22241
aI97765
asS'd'
I14447
sa(dp28894
S'a'
I17440
sS's'
I18
sS'm'
(lp28895
I1
aI1
aI7
aI9
aI29
aI13
aI3
aI207
aI51
aI405
aI1703
aI1923
aI1781
aI14723
aI8103
aI10707
aI64799
aI99349
asS'd'
I14448
sa(dp28896
S'a'
I17457
sS's'
I18
sS'm'
(lp28897
I1
aI1
aI7
aI5
aI23
aI29
aI51
aI63
aI489
aI273
aI1577
aI2807
aI5427
aI9949
aI1929
aI19791
aI109405
aI241465
asS'd'
I14449
sa(dp28898
S'a'
I17458
sS's'
I18
sS'm'
(lp28899
I1
aI3
aI3
aI7
aI29
aI61
aI103
aI55
aI29
aI17
aI1081
aI21
aI5791
aI9803
aI19385
aI45091
aI118069
aI61383
asS'd'
I14450
sa(dp28900
S'a'
I17469
sS's'
I18
sS'm'
(lp28901
I1
aI3
aI1
aI7
aI3
aI15
aI75
aI47
aI475
aI87
aI1541
aI3933
aI1081
aI12361
aI29213
aI64333
aI7229
aI226909
asS'd'
I14451
sa(dp28902
S'a'
I17477
sS's'
I18
sS'm'
(lp28903
I1
aI1
aI5
aI7
aI21
aI45
aI19
aI137
aI351
aI229
aI1773
aI1829
aI5025
aI12661
aI18745
aI54917
aI10419
aI176667
asS'd'
I14452
sa(dp28904
S'a'
I17512
sS's'
I18
sS'm'
(lp28905
I1
aI1
aI1
aI3
aI9
aI37
aI81
aI25
aI11
aI327
aI1653
aI2751
aI2823
aI12575
aI30287
aI46265
aI17299
aI93595
asS'd'
I14453
sa(dp28906
S'a'
I17518
sS's'
I18
sS'm'
(lp28907
I1
aI1
aI1
aI3
aI15
aI17
aI43
aI163
aI223
aI731
aI631
aI2813
aI1723
aI6089
aI14245
aI64339
aI114291
aI40331
asS'd'
I14454
sa(dp28908
S'a'
I17532
sS's'
I18
sS'm'
(lp28909
I1
aI3
aI7
aI5
aI21
aI45
aI41
aI17
aI495
aI61
aI1369
aI369
aI4493
aI12071
aI3813
aI41455
aI62561
aI174399
asS'd'
I14455
sa(dp28910
S'a'
I17536
sS's'
I18
sS'm'
(lp28911
I1
aI1
aI1
aI5
aI9
aI41
aI95
aI113
aI109
aI519
aI1683
aI2265
aI2875
aI12649
aI15575
aI53511
aI100707
aI224035
asS'd'
I14456
sa(dp28912
S'a'
I17559
sS's'
I18
sS'm'
(lp28913
I1
aI3
aI7
aI9
aI29
aI7
aI109
aI109
aI283
aI111
aI1167
aI3679
aI369
aI11597
aI19459
aI759
aI128667
aI172427
asS'd'
I14457
sa(dp28914
S'a'
I17569
sS's'
I18
sS'm'
(lp28915
I1
aI1
aI1
aI3
aI13
aI31
aI97
aI31
aI477
aI507
aI835
aI465
aI7501
aI2485
aI19485
aI51055
aI56363
aI229341
asS'd'
I14458
sa(dp28916
S'a'
I17641
sS's'
I18
sS'm'
(lp28917
I1
aI3
aI5
aI11
aI3
aI23
aI67
aI173
aI99
aI963
aI977
aI1949
aI1263
aI2427
aI15181
aI23571
aI23509
aI26481
asS'd'
I14459
sa(dp28918
S'a'
I17667
sS's'
I18
sS'm'
(lp28919
I1
aI3
aI3
aI1
aI29
aI3
aI35
aI191
aI197
aI277
aI397
aI205
aI5945
aI1069
aI31789
aI3551
aI101901
aI222609
asS'd'
I14460
sa(dp28920
S'a'
I17674
sS's'
I18
sS'm'
(lp28921
I1
aI1
aI5
aI9
aI11
aI23
aI109
aI81
aI295
aI7
aI755
aI2345
aI2823
aI11133
aI22623
aI14515
aI57059
aI231099
asS'd'
I14461
sa(dp28922
S'a'
I17693
sS's'
I18
sS'm'
(lp28923
I1
aI3
aI1
aI3
aI21
aI29
aI37
aI71
aI111
aI737
aI1881
aI871
aI5843
aI5889
aI14615
aI49909
aI7105
aI48335
asS'd'
I14462
sa(dp28924
S'a'
I17710
sS's'
I18
sS'm'
(lp28925
I1
aI3
aI1
aI15
aI23
aI31
aI87
aI181
aI483
aI225
aI2003
aI365
aI1569
aI11153
aI14673
aI30085
aI56497
aI203723
asS'd'
I14463
sa(dp28926
S'a'
I17729
sS's'
I18
sS'm'
(lp28927
I1
aI1
aI3
aI15
aI17
aI47
aI99
aI167
aI485
aI431
aI1481
aI2225
aI1537
aI8513
aI19407
aI34165
aI27289
aI84393
asS'd'
I14464
sa(dp28928
S'a'
I17754
sS's'
I18
sS'm'
(lp28929
I1
aI3
aI7
aI3
aI11
aI17
aI115
aI205
aI403
aI831
aI1869
aI3623
aI5215
aI15511
aI11297
aI25181
aI127491
aI155887
asS'd'
I14465
sa(dp28930
S'a'
I17763
sS's'
I18
sS'm'
(lp28931
I1
aI1
aI7
aI5
aI31
aI37
aI23
aI21
aI403
aI529
aI1185
aI3363
aI6319
aI2435
aI2687
aI39407
aI121891
aI133047
asS'd'
I14466
sa(dp28932
S'a'
I17780
sS's'
I18
sS'm'
(lp28933
I1
aI3
aI7
aI1
aI21
aI31
aI43
aI61
aI371
aI987
aI1783
aI3811
aI6227
aI13199
aI31799
aI28863
aI49329
aI73947
asS'd'
I14467
sa(dp28934
S'a'
I17784
sS's'
I18
sS'm'
(lp28935
I1
aI3
aI1
aI1
aI13
aI1
aI5
aI99
aI35
aI793
aI483
aI2573
aI2249
aI6345
aI12793
aI61917
aI49419
aI58011
asS'd'
I14468
sa(dp28936
S'a'
I17790
sS's'
I18
sS'm'
(lp28937
I1
aI3
aI1
aI15
aI3
aI45
aI35
aI189
aI67
aI447
aI1455
aI3575
aI8191
aI7907
aI21559
aI38211
aI26945
aI240679
asS'd'
I14469
sa(dp28938
S'a'
I17830
sS's'
I18
sS'm'
(lp28939
I1
aI1
aI3
aI11
aI27
aI49
aI9
aI109
aI93
aI473
aI1465
aI271
aI7389
aI47
aI8101
aI6219
aI17437
aI220461
asS'd'
I14470
sa(dp28940
S'a'
I17851
sS's'
I18
sS'm'
(lp28941
I1
aI1
aI7
aI3
aI23
aI31
aI75
aI61
aI375
aI901
aI1329
aI2603
aI3469
aI12957
aI23949
aI62183
aI126763
aI68965
asS'd'
I14471
sa(dp28942
S'a'
I17868
sS's'
I18
sS'm'
(lp28943
I1
aI1
aI7
aI9
aI13
aI59
aI75
aI233
aI339
aI29
aI1117
aI1693
aI593
aI15317
aI29753
aI3079
aI43583
aI79939
asS'd'
I14472
sa(dp28944
S'a'
I17879
sS's'
I18
sS'm'
(lp28945
I1
aI1
aI5
aI1
aI17
aI57
aI81
aI123
aI101
aI765
aI1941
aI3143
aI7403
aI9105
aI23197
aI28983
aI128059
aI5931
asS'd'
I14473
sa(dp28946
S'a'
I17896
sS's'
I18
sS'm'
(lp28947
I1
aI3
aI5
aI11
aI19
aI31
aI89
aI165
aI213
aI251
aI965
aI3203
aI1621
aI4323
aI26877
aI17109
aI18321
aI162413
asS'd'
I14474
sa(dp28948
S'a'
I17901
sS's'
I18
sS'm'
(lp28949
I1
aI3
aI3
aI9
aI11
aI59
aI123
aI213
aI335
aI267
aI1767
aI3317
aI5189
aI10149
aI27921
aI19331
aI71541
aI170501
asS'd'
I14475
sa(dp28950
S'a'
I17916
sS's'
I18
sS'm'
(lp28951
I1
aI3
aI3
aI15
aI3
aI3
aI115
aI235
aI305
aI219
aI265
aI1535
aI4925
aI5597
aI20857
aI32381
aI117237
aI197533
asS'd'
I14476
sa(dp28952
S'a'
I17935
sS's'
I18
sS'm'
(lp28953
I1
aI3
aI5
aI13
aI13
aI59
aI93
aI85
aI419
aI337
aI513
aI2131
aI5665
aI12229
aI1389
aI34355
aI65485
aI81141
asS'd'
I14477
sa(dp28954
S'a'
I17953
sS's'
I18
sS'm'
(lp28955
I1
aI3
aI7
aI7
aI11
aI59
aI111
aI219
aI293
aI289
aI325
aI623
aI3853
aI3775
aI14771
aI5945
aI119451
aI162861
asS'd'
I14478
sa(dp28956
S'a'
I17991
sS's'
I18
sS'm'
(lp28957
I1
aI3
aI1
aI11
aI19
aI33
aI119
aI239
aI431
aI803
aI1119
aI2445
aI3203
aI7219
aI31963
aI34519
aI104953
aI254491
asS'd'
I14479
sa(dp28958
S'a'
I18009
sS's'
I18
sS'm'
(lp28959
I1
aI3
aI7
aI9
aI21
aI53
aI21
aI115
aI365
aI419
aI11
aI3803
aI4283
aI417
aI8937
aI64533
aI56433
aI166025
asS'd'
I14480
sa(dp28960
S'a'
I18016
sS's'
I18
sS'm'
(lp28961
I1
aI1
aI7
aI3
aI17
aI5
aI99
aI143
aI485
aI309
aI1255
aI2641
aI3427
aI1681
aI3301
aI64531
aI38629
aI20945
asS'd'
I14481
sa(dp28962
S'a'
I18034
sS's'
I18
sS'm'
(lp28963
I1
aI1
aI5
aI1
aI31
aI3
aI115
aI217
aI451
aI5
aI1447
aI2317
aI1725
aI12931
aI25799
aI23569
aI51747
aI28821
asS'd'
I14482
sa(dp28964
S'a'
I18061
sS's'
I18
sS'm'
(lp28965
I1
aI3
aI7
aI3
aI31
aI55
aI109
aI107
aI211
aI381
aI1067
aI3973
aI5007
aI8939
aI8605
aI55221
aI124603
aI47115
asS'd'
I14483
sa(dp28966
S'a'
I18070
sS's'
I18
sS'm'
(lp28967
I1
aI1
aI1
aI11
aI19
aI13
aI99
aI241
aI103
aI711
aI1823
aI2671
aI653
aI10217
aI14195
aI39735
aI54807
aI105599
asS'd'
I14484
sa(dp28968
S'a'
I18079
sS's'
I18
sS'm'
(lp28969
I1
aI3
aI3
aI7
aI9
aI33
aI43
aI131
aI493
aI141
aI827
aI2909
aI2847
aI12879
aI7879
aI6263
aI25981
aI57323
asS'd'
I14485
sa(dp28970
S'a'
I18132
sS's'
I18
sS'm'
(lp28971
I1
aI1
aI1
aI1
aI1
aI41
aI55
aI175
aI479
aI725
aI157
aI3403
aI5809
aI10685
aI20433
aI21729
aI9493
aI205685
asS'd'
I14486
sa(dp28972
S'a'
I18145
sS's'
I18
sS'm'
(lp28973
I1
aI1
aI7
aI5
aI1
aI33
aI31
aI245
aI109
aI711
aI1047
aI941
aI449
aI1055
aI16249
aI45211
aI48311
aI171339
asS'd'
I14487
sa(dp28974
S'a'
I18169
sS's'
I18
sS'm'
(lp28975
I1
aI3
aI3
aI9
aI27
aI9
aI113
aI69
aI269
aI643
aI1371
aI3521
aI4969
aI5373
aI11133
aI63109
aI42725
aI126969
asS'd'
I14488
sa(dp28976
S'a'
I18170
sS's'
I18
sS'm'
(lp28977
I1
aI1
aI5
aI15
aI9
aI21
aI1
aI195
aI421
aI429
aI1103
aI2727
aI463
aI9801
aI8955
aI62841
aI94687
aI114509
asS'd'
I14489
sa(dp28978
S'a'
I18190
sS's'
I18
sS'm'
(lp28979
I1
aI1
aI5
aI5
aI5
aI47
aI9
aI221
aI59
aI115
aI359
aI1147
aI749
aI1009
aI23129
aI641
aI39471
aI23073
asS'd'
I14490
sa(dp28980
S'a'
I18192
sS's'
I18
sS'm'
(lp28981
I1
aI3
aI5
aI13
aI27
aI29
aI19
aI3
aI121
aI773
aI625
aI2757
aI6377
aI15867
aI14563
aI40391
aI4351
aI21153
asS'd'
I14491
sa(dp28982
S'a'
I18201
sS's'
I18
sS'm'
(lp28983
I1
aI1
aI5
aI9
aI11
aI25
aI51
aI101
aI273
aI541
aI1761
aI593
aI7111
aI4369
aI30095
aI34867
aI103989
aI19855
asS'd'
I14492
sa(dp28984
S'a'
I18228
sS's'
I18
sS'm'
(lp28985
I1
aI1
aI3
aI13
aI27
aI55
aI79
aI115
aI105
aI855
aI627
aI2227
aI2927
aI8757
aI8713
aI54607
aI43671
aI130153
asS'd'
I14493
sa(dp28986
S'a'
I18243
sS's'
I18
sS'm'
(lp28987
I1
aI1
aI3
aI5
aI7
aI45
aI21
aI71
aI157
aI773
aI1265
aI841
aI2463
aI2217
aI6087
aI28683
aI21251
aI72377
asS'd'
I14494
sa(dp28988
S'a'
I18260
sS's'
I18
sS'm'
(lp28989
I1
aI1
aI3
aI1
aI15
aI11
aI117
aI211
aI223
aI713
aI545
aI907
aI6907
aI41
aI17039
aI23079
aI86657
aI5765
asS'd'
I14495
sa(dp28990
S'a'
I18279
sS's'
I18
sS'm'
(lp28991
I1
aI1
aI5
aI3
aI27
aI33
aI77
aI137
aI401
aI585
aI911
aI1189
aI2749
aI3427
aI2701
aI2453
aI84857
aI176585
asS'd'
I14496
sa(dp28992
S'a'
I18285
sS's'
I18
sS'm'
(lp28993
I1
aI1
aI1
aI3
aI7
aI39
aI73
aI143
aI29
aI569
aI939
aI301
aI7827
aI7691
aI11513
aI64517
aI113679
aI234165
asS'd'
I14497
sa(dp28994
S'a'
I18297
sS's'
I18
sS'm'
(lp28995
I1
aI3
aI1
aI11
aI29
aI57
aI127
aI181
aI175
aI973
aI1537
aI761
aI5205
aI13641
aI32649
aI8621
aI77509
aI261235
asS'd'
I14498
sa(dp28996
S'a'
I18300
sS's'
I18
sS'm'
(lp28997
I1
aI3
aI5
aI5
aI13
aI19
aI117
aI225
aI477
aI297
aI1807
aI2357
aI5653
aI3791
aI6325
aI54877
aI120659
aI91013
asS'd'
I14499
sa(dp28998
S'a'
I18321
sS's'
I18
sS'm'
(lp28999
I1
aI3
aI1
aI1
aI3
aI55
aI19
aI99
aI321
aI877
aI541
aI511
aI141
aI15047
aI26377
aI9
aI2765
aI223533
asS'd'
I14500
sa(dp29000
S'a'
I18344
sS's'
I18
sS'm'
(lp29001
I1
aI1
aI1
aI11
aI13
aI59
aI121
aI147
aI215
aI117
aI1047
aI3055
aI2129
aI15191
aI14425
aI28327
aI108541
aI114275
asS'd'
I14501
sa(dp29002
S'a'
I18358
sS's'
I18
sS'm'
(lp29003
I1
aI1
aI1
aI15
aI3
aI21
aI105
aI61
aI501
aI899
aI195
aI2745
aI5989
aI4433
aI19525
aI35477
aI22997
aI241657
asS'd'
I14502
sa(dp29004
S'a'
I18364
sS's'
I18
sS'm'
(lp29005
I1
aI1
aI3
aI5
aI19
aI47
aI77
aI247
aI413
aI317
aI1255
aI2087
aI4493
aI2211
aI9003
aI22145
aI94001
aI50579
asS'd'
I14503
sa(dp29006
S'a'
I18376
sS's'
I18
sS'm'
(lp29007
I1
aI3
aI7
aI11
aI31
aI47
aI25
aI191
aI65
aI409
aI1349
aI2481
aI7619
aI223
aI18051
aI63609
aI77187
aI75483
asS'd'
I14504
sa(dp29008
S'a'
I18390
sS's'
I18
sS'm'
(lp29009
I1
aI1
aI1
aI9
aI21
aI59
aI115
aI251
aI401
aI91
aI627
aI3273
aI2393
aI2949
aI11475
aI23669
aI16171
aI77507
asS'd'
I14505
sa(dp29010
S'a'
I18399
sS's'
I18
sS'm'
(lp29011
I1
aI3
aI5
aI1
aI19
aI7
aI65
aI253
aI217
aI493
aI227
aI3269
aI4261
aI2295
aI32037
aI5773
aI12925
aI41821
asS'd'
I14506
sa(dp29012
S'a'
I18400
sS's'
I18
sS'm'
(lp29013
I1
aI1
aI5
aI11
aI5
aI31
aI71
aI205
aI285
aI37
aI1863
aI1873
aI191
aI16137
aI2955
aI51993
aI91401
aI206967
asS'd'
I14507
sa(dp29014
S'a'
I18427
sS's'
I18
sS'm'
(lp29015
I1
aI3
aI1
aI7
aI23
aI31
aI21
aI81
aI37
aI903
aI817
aI3447
aI8067
aI3087
aI25831
aI46247
aI77255
aI68365
asS'd'
I14508
sa(dp29016
S'a'
I18443
sS's'
I18
sS'm'
(lp29017
I1
aI1
aI7
aI11
aI7
aI43
aI21
aI243
aI431
aI633
aI2047
aI577
aI7297
aI8151
aI15951
aI30313
aI121569
aI241687
asS'd'
I14509
sa(dp29018
S'a'
I18487
sS's'
I18
sS'm'
(lp29019
I1
aI1
aI3
aI5
aI7
aI45
aI35
aI189
aI381
aI849
aI1869
aI1193
aI6815
aI9017
aI29053
aI63605
aI113623
aI249097
asS'd'
I14510
sa(dp29020
S'a'
I18493
sS's'
I18
sS'm'
(lp29021
I1
aI3
aI3
aI11
aI13
aI1
aI73
aI151
aI197
aI591
aI1101
aI2437
aI6695
aI8337
aI26539
aI40147
aI45673
aI57727
asS'd'
I14511
sa(dp29022
S'a'
I18508
sS's'
I18
sS'm'
(lp29023
I1
aI1
aI5
aI1
aI19
aI15
aI61
aI151
aI37
aI893
aI1819
aI2317
aI6299
aI13097
aI5109
aI32613
aI123685
aI128173
asS'd'
I14512
sa(dp29024
S'a'
I18532
sS's'
I18
sS'm'
(lp29025
I1
aI1
aI1
aI7
aI25
aI29
aI29
aI203
aI179
aI211
aI1483
aI3315
aI7125
aI6931
aI609
aI849
aI117571
aI26829
asS'd'
I14513
sa(dp29026
S'a'
I18535
sS's'
I18
sS'm'
(lp29027
I1
aI3
aI3
aI11
aI11
aI47
aI33
aI101
aI181
aI431
aI183
aI2777
aI5269
aI4177
aI15727
aI717
aI111243
aI34825
asS'd'
I14514
sa(dp29028
S'a'
I18580
sS's'
I18
sS'm'
(lp29029
I1
aI3
aI5
aI11
aI17
aI19
aI19
aI143
aI137
aI537
aI1249
aI2889
aI1911
aI3895
aI15433
aI60165
aI83815
aI205569
asS'd'
I14515
sa(dp29030
S'a'
I18587
sS's'
I18
sS'm'
(lp29031
I1
aI3
aI3
aI7
aI9
aI59
aI13
aI159
aI307
aI625
aI1
aI2887
aI3307
aI16371
aI4269
aI56253
aI71171
aI55543
asS'd'
I14516
sa(dp29032
S'a'
I18606
sS's'
I18
sS'm'
(lp29033
I1
aI1
aI1
aI11
aI7
aI63
aI15
aI53
aI409
aI7
aI1317
aI473
aI7481
aI10321
aI27941
aI4941
aI40003
aI194153
asS'd'
I14517
sa(dp29034
S'a'
I18659
sS's'
I18
sS'm'
(lp29035
I1
aI1
aI1
aI1
aI11
aI53
aI93
aI157
aI289
aI231
aI31
aI273
aI8131
aI7861
aI31041
aI55221
aI58305
aI203403
asS'd'
I14518
sa(dp29036
S'a'
I18662
sS's'
I18
sS'm'
(lp29037
I1
aI1
aI5
aI11
aI15
aI53
aI103
aI41
aI439
aI601
aI1949
aI1087
aI4275
aI4675
aI31879
aI40909
aI22365
aI124781
asS'd'
I14519
sa(dp29038
S'a'
I18679
sS's'
I18
sS'm'
(lp29039
I1
aI3
aI5
aI9
aI1
aI47
aI81
aI47
aI197
aI499
aI329
aI2387
aI5455
aI15571
aI2289
aI44121
aI12105
aI11883
asS'd'
I14520
sa(dp29040
S'a'
I18694
sS's'
I18
sS'm'
(lp29041
I1
aI1
aI3
aI3
aI7
aI47
aI93
aI33
aI265
aI149
aI845
aI723
aI7783
aI6651
aI22939
aI58027
aI66959
aI3991
asS'd'
I14521
sa(dp29042
S'a'
I18697
sS's'
I18
sS'm'
(lp29043
I1
aI1
aI5
aI11
aI23
aI35
aI123
aI143
aI35
aI981
aI1269
aI2853
aI4547
aI7877
aI16181
aI17155
aI57605
aI11589
asS'd'
I14522
sa(dp29044
S'a'
I18706
sS's'
I18
sS'm'
(lp29045
I1
aI1
aI7
aI11
aI9
aI57
aI87
aI151
aI333
aI743
aI1939
aI3273
aI1047
aI5033
aI16061
aI37237
aI12013
aI17669
asS'd'
I14523
sa(dp29046
S'a'
I18708
sS's'
I18
sS'm'
(lp29047
I1
aI3
aI7
aI5
aI29
aI15
aI109
aI185
aI51
aI159
aI1353
aI3041
aI7821
aI14053
aI13643
aI62045
aI78475
aI43603
asS'd'
I14524
sa(dp29048
S'a'
I18728
sS's'
I18
sS'm'
(lp29049
I1
aI3
aI1
aI9
aI29
aI25
aI121
aI49
aI415
aI561
aI325
aI1139
aI1993
aI6437
aI6025
aI25225
aI20761
aI250589
asS'd'
I14525
sa(dp29050
S'a'
I18731
sS's'
I18
sS'm'
(lp29051
I1
aI3
aI5
aI3
aI15
aI39
aI33
aI43
aI437
aI605
aI1081
aI2397
aI3821
aI10961
aI4853
aI19517
aI95817
aI142023
asS'd'
I14526
sa(dp29052
S'a'
I18733
sS's'
I18
sS'm'
(lp29053
I1
aI1
aI3
aI11
aI23
aI51
aI119
aI13
aI227
aI981
aI2017
aI3265
aI1215
aI8737
aI10719
aI48027
aI43239
aI19425
asS'd'
I14527
sa(dp29054
S'a'
I18751
sS's'
I18
sS'm'
(lp29055
I1
aI3
aI1
aI15
aI5
aI5
aI33
aI175
aI509
aI611
aI451
aI2653
aI1553
aI1941
aI25221
aI31259
aI6027
aI159847
asS'd'
I14528
sa(dp29056
S'a'
I18766
sS's'
I18
sS'm'
(lp29057
I1
aI3
aI7
aI11
aI7
aI25
aI71
aI61
aI89
aI775
aI609
aI2363
aI4261
aI10677
aI1243
aI44895
aI49113
aI209603
asS'd'
I14529
sa(dp29058
S'a'
I18773
sS's'
I18
sS'm'
(lp29059
I1
aI3
aI5
aI15
aI23
aI23
aI3
aI15
aI489
aI455
aI1303
aI745
aI5311
aI1639
aI18317
aI33729
aI119303
aI255359
asS'd'
I14530
sa(dp29060
S'a'
I18796
sS's'
I18
sS'm'
(lp29061
I1
aI3
aI7
aI5
aI13
aI53
aI29
aI127
aI159
aI67
aI469
aI1735
aI3497
aI6985
aI24735
aI32957
aI1225
aI24447
asS'd'
I14531
sa(dp29062
S'a'
I18811
sS's'
I18
sS'm'
(lp29063
I1
aI3
aI5
aI5
aI9
aI13
aI119
aI83
aI387
aI777
aI361
aI3183
aI6351
aI9071
aI13699
aI53873
aI54663
aI67453
asS'd'
I14532
sa(dp29064
S'a'
I18823
sS's'
I18
sS'm'
(lp29065
I1
aI1
aI5
aI9
aI17
aI33
aI9
aI159
aI143
aI193
aI1055
aI2903
aI2719
aI12521
aI5231
aI37639
aI94963
aI105673
asS'd'
I14533
sa(dp29066
S'a'
I18832
sS's'
I18
sS'm'
(lp29067
I1
aI3
aI3
aI1
aI27
aI53
aI87
aI49
aI465
aI517
aI1333
aI411
aI4089
aI9985
aI12989
aI59511
aI49939
aI223481
asS'd'
I14534
sa(dp29068
S'a'
I18854
sS's'
I18
sS'm'
(lp29069
I1
aI1
aI5
aI9
aI27
aI59
aI35
aI125
aI393
aI271
aI1565
aI2847
aI8139
aI15627
aI16059
aI55319
aI11131
aI35141
asS'd'
I14535
sa(dp29070
S'a'
I18858
sS's'
I18
sS'm'
(lp29071
I1
aI1
aI1
aI1
aI25
aI1
aI27
aI195
aI113
aI539
aI1281
aI2273
aI4793
aI695
aI25599
aI41145
aI107431
aI160137
asS'd'
I14536
sa(dp29072
S'a'
I18883
sS's'
I18
sS'm'
(lp29073
I1
aI3
aI3
aI7
aI13
aI7
aI35
aI137
aI83
aI995
aI1671
aI1701
aI3157
aI15583
aI7637
aI18947
aI59675
aI9421
asS'd'
I14537
sa(dp29074
S'a'
I18900
sS's'
I18
sS'm'
(lp29075
I1
aI1
aI7
aI15
aI23
aI37
aI109
aI93
aI377
aI885
aI1843
aI1867
aI2013
aI10535
aI5717
aI55463
aI18307
aI125537
asS'd'
I14538
sa(dp29076
S'a'
I18967
sS's'
I18
sS'm'
(lp29077
I1
aI3
aI7
aI11
aI25
aI33
aI91
aI213
aI109
aI599
aI131
aI1879
aI1375
aI2911
aI4649
aI8809
aI41199
aI61629
asS'd'
I14539
sa(dp29078
S'a'
I18974
sS's'
I18
sS'm'
(lp29079
I1
aI3
aI1
aI1
aI11
aI17
aI117
aI243
aI427
aI913
aI495
aI527
aI4277
aI8867
aI3131
aI14143
aI81677
aI177369
asS'd'
I14540
sa(dp29080
S'a'
I18990
sS's'
I18
sS'm'
(lp29081
I1
aI3
aI7
aI11
aI11
aI37
aI71
aI185
aI487
aI161
aI1773
aI837
aI243
aI14105
aI6881
aI2155
aI63679
aI220387
asS'd'
I14541
sa(dp29082
S'a'
I19009
sS's'
I18
sS'm'
(lp29083
I1
aI1
aI5
aI3
aI11
aI41
aI33
aI99
aI495
aI757
aI1083
aI1987
aI1997
aI11057
aI18445
aI61903
aI78163
aI121701
asS'd'
I14542
sa(dp29084
S'a'
I19055
sS's'
I18
sS'm'
(lp29085
I1
aI1
aI1
aI1
aI23
aI37
aI9
aI19
aI411
aI11
aI1487
aI1279
aI2129
aI7449
aI29631
aI34559
aI129753
aI112627
asS'd'
I14543
sa(dp29086
S'a'
I19058
sS's'
I18
sS'm'
(lp29087
I1
aI1
aI3
aI11
aI31
aI39
aI41
aI207
aI141
aI383
aI723
aI3053
aI743
aI4479
aI12395
aI56659
aI130303
aI152005
asS'd'
I14544
sa(dp29088
S'a'
I19074
sS's'
I18
sS'm'
(lp29089
I1
aI3
aI1
aI7
aI27
aI25
aI19
aI37
aI29
aI781
aI1115
aI2569
aI4113
aI14033
aI18653
aI1055
aI50639
aI70413
asS'd'
I14545
sa(dp29090
S'a'
I19079
sS's'
I18
sS'm'
(lp29091
I1
aI1
aI3
aI1
aI9
aI15
aI109
aI7
aI221
aI161
aI569
aI2915
aI2717
aI2439
aI4257
aI61851
aI113183
aI63139
asS'd'
I14546
sa(dp29092
S'a'
I19086
sS's'
I18
sS'm'
(lp29093
I1
aI3
aI5
aI1
aI17
aI45
aI3
aI147
aI207
aI769
aI321
aI11
aI2747
aI7189
aI8067
aI34951
aI50851
aI42625
asS'd'
I14547
sa(dp29094
S'a'
I19128
sS's'
I18
sS'm'
(lp29095
I1
aI1
aI7
aI11
aI15
aI53
aI117
aI161
aI219
aI937
aI1661
aI3767
aI959
aI10351
aI26685
aI40095
aI109821
aI140139
asS'd'
I14548
sa(dp29096
S'a'
I19145
sS's'
I18
sS'm'
(lp29097
I1
aI3
aI3
aI9
aI15
aI47
aI61
aI35
aI289
aI743
aI1723
aI2189
aI749
aI13499
aI22897
aI55385
aI114953
aI67191
asS'd'
I14549
sa(dp29098
S'a'
I19163
sS's'
I18
sS'm'
(lp29099
I1
aI3
aI7
aI3
aI23
aI19
aI123
aI217
aI393
aI889
aI1665
aI13
aI5663
aI8695
aI29767
aI13433
aI65133
aI226713
asS'd'
I14550
sa(dp29100
S'a'
I19175
sS's'
I18
sS'm'
(lp29101
I1
aI1
aI7
aI11
aI5
aI57
aI59
aI171
aI321
aI519
aI1333
aI1975
aI5331
aI2383
aI26863
aI8989
aI82167
aI6915
asS'd'
I14551
sa(dp29102
S'a'
I19189
sS's'
I18
sS'm'
(lp29103
I1
aI1
aI3
aI3
aI7
aI17
aI105
aI79
aI7
aI827
aI1277
aI3805
aI5943
aI3161
aI28953
aI15657
aI615
aI149131
asS'd'
I14552
sa(dp29104
S'a'
I19196
sS's'
I18
sS'm'
(lp29105
I1
aI1
aI5
aI1
aI5
aI7
aI99
aI65
aI295
aI933
aI365
aI1867
aI1959
aI10733
aI26947
aI29659
aI121889
aI200379
asS'd'
I14553
sa(dp29106
S'a'
I19204
sS's'
I18
sS'm'
(lp29107
I1
aI3
aI1
aI13
aI25
aI21
aI89
aI247
aI251
aI43
aI1539
aI1317
aI1875
aI9237
aI20693
aI58433
aI16757
aI25451
asS'd'
I14554
sa(dp29108
S'a'
I19213
sS's'
I18
sS'm'
(lp29109
I1
aI3
aI3
aI13
aI11
aI47
aI73
aI21
aI467
aI337
aI1881
aI2723
aI7023
aI2767
aI12553
aI65533
aI20517
aI203749
asS'd'
I14555
sa(dp29110
S'a'
I19237
sS's'
I18
sS'm'
(lp29111
I1
aI1
aI1
aI1
aI1
aI17
aI85
aI133
aI369
aI577
aI71
aI859
aI8151
aI919
aI10843
aI44017
aI10097
aI199893
asS'd'
I14556
sa(dp29112
S'a'
I19276
sS's'
I18
sS'm'
(lp29113
I1
aI3
aI5
aI5
aI23
aI19
aI21
aI233
aI475
aI123
aI621
aI687
aI6945
aI2373
aI6447
aI31243
aI3525
aI256545
asS'd'
I14557
sa(dp29114
S'a'
I19287
sS's'
I18
sS'm'
(lp29115
I1
aI1
aI7
aI5
aI9
aI5
aI35
aI21
aI33
aI353
aI1429
aI3249
aI6159
aI8757
aI6213
aI855
aI75863
aI74507
asS'd'
I14558
sa(dp29116
S'a'
I19291
sS's'
I18
sS'm'
(lp29117
I1
aI1
aI5
aI11
aI29
aI21
aI45
aI155
aI369
aI769
aI1041
aI3929
aI7377
aI1621
aI5285
aI55213
aI66143
aI110251
asS'd'
I14559
sa(dp29118
S'a'
I19293
sS's'
I18
sS'm'
(lp29119
I1
aI3
aI7
aI11
aI13
aI57
aI45
aI207
aI259
aI907
aI573
aI663
aI7727
aI12677
aI5949
aI57625
aI42183
aI217491
asS'd'
I14560
sa(dp29120
S'a'
I19304
sS's'
I18
sS'm'
(lp29121
I1
aI3
aI5
aI3
aI21
aI63
aI113
aI159
aI87
aI551
aI1405
aI2867
aI239
aI10941
aI27633
aI13947
aI69689
aI225771
asS'd'
I14561
sa(dp29122
S'a'
I19371
sS's'
I18
sS'm'
(lp29123
I1
aI1
aI3
aI1
aI1
aI59
aI5
aI41
aI125
aI707
aI1457
aI1
aI4263
aI5519
aI26101
aI46339
aI44949
aI63689
asS'd'
I14562
sa(dp29124
S'a'
I19379
sS's'
I18
sS'm'
(lp29125
I1
aI1
aI1
aI1
aI11
aI9
aI65
aI155
aI3
aI85
aI273
aI2287
aI6059
aI3289
aI19045
aI14705
aI112465
aI202019
asS'd'
I14563
sa(dp29126
S'a'
I19381
sS's'
I18
sS'm'
(lp29127
I1
aI3
aI3
aI3
aI21
aI49
aI95
aI75
aI479
aI519
aI1511
aI1609
aI2421
aI14435
aI11749
aI49627
aI16221
aI98351
asS'd'
I14564
sa(dp29128
S'a'
I19405
sS's'
I18
sS'm'
(lp29129
I1
aI1
aI3
aI5
aI25
aI57
aI1
aI39
aI377
aI523
aI529
aI701
aI6749
aI10109
aI15845
aI53301
aI70979
aI210997
asS'd'
I14565
sa(dp29130
S'a'
I19417
sS's'
I18
sS'm'
(lp29131
I1
aI3
aI1
aI3
aI27
aI29
aI101
aI87
aI361
aI1
aI229
aI2653
aI769
aI16121
aI18221
aI31937
aI12187
aI63801
asS'd'
I14566
sa(dp29132
S'a'
I19420
sS's'
I18
sS'm'
(lp29133
I1
aI3
aI3
aI1
aI25
aI27
aI49
aI235
aI309
aI23
aI1625
aI589
aI1251
aI10305
aI26943
aI38949
aI82539
aI135491
asS'd'
I14567
sa(dp29134
S'a'
I19424
sS's'
I18
sS'm'
(lp29135
I1
aI1
aI7
aI7
aI13
aI13
aI13
aI61
aI509
aI73
aI201
aI2309
aI1601
aI3145
aI19867
aI5623
aI117455
aI180681
asS'd'
I14568
sa(dp29136
S'a'
I19462
sS's'
I18
sS'm'
(lp29137
I1
aI3
aI3
aI13
aI13
aI47
aI71
aI9
aI123
aI719
aI701
aI353
aI1877
aI3103
aI20017
aI64731
aI72729
aI147631
asS'd'
I14569
sa(dp29138
S'a'
I19474
sS's'
I18
sS'm'
(lp29139
I1
aI3
aI1
aI7
aI29
aI29
aI53
aI97
aI409
aI67
aI1033
aI2403
aI2471
aI10869
aI2837
aI43459
aI117415
aI213371
asS'd'
I14570
sa(dp29140
S'a'
I19492
sS's'
I18
sS'm'
(lp29141
I1
aI1
aI7
aI1
aI7
aI23
aI103
aI157
aI315
aI335
aI375
aI3493
aI4095
aI5331
aI7773
aI64173
aI23167
aI21259
asS'd'
I14571
sa(dp29142
S'a'
I19501
sS's'
I18
sS'm'
(lp29143
I1
aI3
aI5
aI13
aI13
aI55
aI107
aI147
aI447
aI281
aI401
aI1897
aI7887
aI15005
aI21645
aI26007
aI19673
aI238931
asS'd'
I14572
sa(dp29144
S'a'
I19504
sS's'
I18
sS'm'
(lp29145
I1
aI3
aI1
aI7
aI17
aI39
aI109
aI113
aI143
aI59
aI1095
aI225
aI1455
aI5021
aI5011
aI2039
aI4381
aI219847
asS'd'
I14573
sa(dp29146
S'a'
I19516
sS's'
I18
sS'm'
(lp29147
I1
aI3
aI3
aI7
aI1
aI35
aI121
aI145
aI297
aI251
aI1153
aI1955
aI7881
aI15461
aI26961
aI915
aI30253
aI15289
asS'd'
I14574
sa(dp29148
S'a'
I19519
sS's'
I18
sS'm'
(lp29149
I1
aI3
aI5
aI15
aI5
aI57
aI43
aI157
aI49
aI17
aI993
aI4085
aI5639
aI9405
aI28661
aI30191
aI73291
aI76913
asS'd'
I14575
sa(dp29150
S'a'
I19534
sS's'
I18
sS'm'
(lp29151
I1
aI1
aI7
aI1
aI25
aI63
aI117
aI55
aI63
aI649
aI1635
aI2505
aI2765
aI2715
aI30241
aI62699
aI19567
aI65953
asS'd'
I14576
sa(dp29152
S'a'
I19555
sS's'
I18
sS'm'
(lp29153
I1
aI3
aI5
aI13
aI21
aI49
aI111
aI127
aI179
aI819
aI1737
aI2519
aI815
aI10541
aI15821
aI54203
aI71767
aI7091
asS'd'
I14577
sa(dp29154
S'a'
I19597
sS's'
I18
sS'm'
(lp29155
I1
aI1
aI1
aI3
aI27
aI41
aI101
aI139
aI39
aI995
aI819
aI319
aI1481
aI15265
aI20611
aI22445
aI53733
aI82871
asS'd'
I14578
sa(dp29156
S'a'
I19600
sS's'
I18
sS'm'
(lp29157
I1
aI3
aI3
aI13
aI7
aI61
aI103
aI203
aI353
aI205
aI1927
aI2665
aI757
aI12277
aI31217
aI22247
aI14527
aI26385
asS'd'
I14579
sa(dp29158
S'a'
I19615
sS's'
I18
sS'm'
(lp29159
I1
aI1
aI7
aI3
aI5
aI35
aI87
aI235
aI139
aI785
aI417
aI3975
aI6753
aI4267
aI15201
aI8747
aI12491
aI159979
asS'd'
I14580
sa(dp29160
S'a'
I19621
sS's'
I18
sS'm'
(lp29161
I1
aI3
aI1
aI9
aI9
aI11
aI117
aI231
aI503
aI933
aI1461
aI2657
aI7771
aI2161
aI26723
aI4853
aI23215
aI162315
asS'd'
I14581
sa(dp29162
S'a'
I19646
sS's'
I18
sS'm'
(lp29163
I1
aI1
aI7
aI15
aI27
aI25
aI115
aI9
aI257
aI89
aI571
aI41
aI2169
aI10619
aI2695
aI2107
aI64747
aI40489
asS'd'
I14582
sa(dp29164
S'a'
I19651
sS's'
I18
sS'm'
(lp29165
I1
aI3
aI7
aI9
aI29
aI61
aI91
aI117
aI279
aI721
aI233
aI177
aI5509
aI7599
aI2379
aI20297
aI75425
aI25051
asS'd'
I14583
sa(dp29166
S'a'
I19693
sS's'
I18
sS'm'
(lp29167
I1
aI1
aI1
aI15
aI31
aI41
aI3
aI57
aI59
aI47
aI963
aI2831
aI1885
aI1989
aI26803
aI48243
aI112065
aI27753
asS'd'
I14584
sa(dp29168
S'a'
I19702
sS's'
I18
sS'm'
(lp29169
I1
aI1
aI5
aI15
aI9
aI57
aI41
aI255
aI179
aI719
aI1463
aI2857
aI285
aI9623
aI13111
aI20415
aI28819
aI149441
asS'd'
I14585
sa(dp29170
S'a'
I19726
sS's'
I18
sS'm'
(lp29171
I1
aI3
aI1
aI9
aI17
aI63
aI21
aI79
aI473
aI525
aI1557
aI3205
aI7097
aI14379
aI28039
aI30731
aI62383
aI247429
asS'd'
I14586
sa(dp29172
S'a'
I19754
sS's'
I18
sS'm'
(lp29173
I1
aI3
aI3
aI13
aI25
aI45
aI97
aI213
aI11
aI801
aI1519
aI1085
aI6167
aI13701
aI6707
aI47223
aI69923
aI66239
asS'd'
I14587
sa(dp29174
S'a'
I19759
sS's'
I18
sS'm'
(lp29175
I1
aI1
aI1
aI13
aI1
aI1
aI9
aI21
aI363
aI729
aI1715
aI1249
aI5299
aI11357
aI20627
aI33559
aI84255
aI133743
asS'd'
I14588
sa(dp29176
S'a'
I19764
sS's'
I18
sS'm'
(lp29177
I1
aI3
aI3
aI13
aI17
aI33
aI23
aI255
aI309
aI605
aI1177
aI1305
aI2717
aI6561
aI29193
aI7971
aI117525
aI79139
asS'd'
I14589
sa(dp29178
S'a'
I19788
sS's'
I18
sS'm'
(lp29179
I1
aI3
aI1
aI1
aI7
aI15
aI73
aI171
aI11
aI791
aI241
aI2641
aI5397
aI10403
aI22207
aI64123
aI124507
aI63855
asS'd'
I14590
sa(dp29180
S'a'
I19793
sS's'
I18
sS'm'
(lp29181
I1
aI3
aI1
aI11
aI11
aI7
aI109
aI103
aI321
aI1009
aI1237
aI3347
aI287
aI2389
aI16529
aI7789
aI3347
aI97827
asS'd'
I14591
sa(dp29182
S'a'
I19836
sS's'
I18
sS'm'
(lp29183
I1
aI3
aI3
aI3
aI1
aI27
aI17
aI9
aI223
aI755
aI559
aI3811
aI2997
aI1543
aI23197
aI42371
aI5837
aI13809
asS'd'
I14592
sa(dp29184
S'a'
I19855
sS's'
I18
sS'm'
(lp29185
I1
aI3
aI1
aI3
aI7
aI57
aI31
aI23
aI35
aI329
aI1155
aI2525
aI3029
aI5495
aI12005
aI18045
aI4539
aI75789
asS'd'
I14593
sa(dp29186
S'a'
I19858
sS's'
I18
sS'm'
(lp29187
I1
aI1
aI5
aI13
aI3
aI31
aI121
aI161
aI325
aI869
aI715
aI851
aI1273
aI1871
aI22711
aI61499
aI36291
aI11663
asS'd'
I14594
sa(dp29188
S'a'
I19880
sS's'
I18
sS'm'
(lp29189
I1
aI1
aI5
aI11
aI7
aI39
aI23
aI139
aI197
aI47
aI513
aI373
aI6859
aI11217
aI17725
aI60949
aI19299
aI91425
asS'd'
I14595
sa(dp29190
S'a'
I19883
sS's'
I18
sS'm'
(lp29191
I1
aI1
aI3
aI7
aI15
aI63
aI123
aI11
aI109
aI829
aI231
aI2591
aI7997
aI9061
aI18647
aI3209
aI38509
aI211219
asS'd'
I14596
sa(dp29192
S'a'
I19917
sS's'
I18
sS'm'
(lp29193
I1
aI1
aI1
aI11
aI13
aI35
aI73
aI223
aI325
aI49
aI1317
aI4063
aI4127
aI2755
aI555
aI51057
aI44909
aI205723
asS'd'
I14597
sa(dp29194
S'a'
I19918
sS's'
I18
sS'm'
(lp29195
I1
aI3
aI3
aI13
aI17
aI41
aI115
aI141
aI503
aI525
aI63
aI2487
aI3225
aI959
aI10623
aI28577
aI89127
aI157269
asS'd'
I14598
sa(dp29196
S'a'
I19936
sS's'
I18
sS'm'
(lp29197
I1
aI3
aI1
aI9
aI25
aI9
aI43
aI43
aI279
aI111
aI1141
aI3033
aI7229
aI5725
aI8277
aI59141
aI116811
aI127945
asS'd'
I14599
sa(dp29198
S'a'
I19946
sS's'
I18
sS'm'
(lp29199
I1
aI3
aI7
aI11
aI27
aI27
aI93
aI243
aI135
aI333
aI1475
aI1259
aI1583
aI7191
aI6831
aI53485
aI128819
aI174211
asS'd'
I14600
sa(dp29200
S'a'
I19954
sS's'
I18
sS'm'
(lp29201
I1
aI3
aI3
aI3
aI17
aI17
aI43
aI251
aI433
aI1011
aI1817
aI2835
aI7721
aI2449
aI9463
aI23779
aI31427
aI88127
asS'd'
I14601
sa(dp29202
S'a'
I19979
sS's'
I18
sS'm'
(lp29203
I1
aI1
aI3
aI3
aI11
aI49
aI61
aI41
aI211
aI559
aI1761
aI1303
aI2119
aI5743
aI25515
aI60705
aI54405
aI241063
asS'd'
I14602
sa(dp29204
S'a'
I19993
sS's'
I18
sS'm'
(lp29205
I1
aI1
aI3
aI11
aI7
aI61
aI15
aI115
aI29
aI35
aI187
aI3137
aI6177
aI1449
aI32723
aI15917
aI107851
aI101077
asS'd'
I14603
sa(dp29206
S'a'
I19994
sS's'
I18
sS'm'
(lp29207
I1
aI3
aI5
aI13
aI21
aI7
aI11
aI231
aI417
aI73
aI1175
aI735
aI627
aI7393
aI7233
aI39883
aI129481
aI106733
asS'd'
I14604
sa(dp29208
S'a'
I20006
sS's'
I18
sS'm'
(lp29209
I1
aI3
aI1
aI15
aI27
aI61
aI63
aI201
aI27
aI431
aI1127
aI1555
aI1953
aI13051
aI18701
aI30097
aI95549
aI198465
asS'd'
I14605
sa(dp29210
S'a'
I20017
sS's'
I18
sS'm'
(lp29211
I1
aI1
aI3
aI1
aI23
aI25
aI43
aI85
aI291
aI85
aI1861
aI675
aI7451
aI14701
aI3929
aI10835
aI25569
aI154687
asS'd'
I14606
sa(dp29212
S'a'
I20032
sS's'
I18
sS'm'
(lp29213
I1
aI3
aI7
aI15
aI5
aI43
aI91
aI225
aI283
aI259
aI1311
aI3977
aI585
aI14803
aI14117
aI2121
aI106981
aI157577
asS'd'
I14607
sa(dp29214
S'a'
I20038
sS's'
I18
sS'm'
(lp29215
I1
aI1
aI1
aI11
aI7
aI51
aI49
aI115
aI477
aI861
aI1115
aI743
aI5109
aI959
aI7105
aI9245
aI66297
aI188751
asS'd'
I14608
sa(dp29216
S'a'
I20050
sS's'
I18
sS'm'
(lp29217
I1
aI3
aI3
aI11
aI23
aI1
aI11
aI111
aI163
aI643
aI1907
aI3613
aI2967
aI10071
aI6023
aI1307
aI62341
aI241025
asS'd'
I14609
sa(dp29218
S'a'
I20080
sS's'
I18
sS'm'
(lp29219
I1
aI3
aI5
aI5
aI15
aI29
aI31
aI43
aI445
aI219
aI1261
aI421
aI6035
aI6461
aI25583
aI817
aI100509
aI239637
asS'd'
I14610
sa(dp29220
S'a'
I20135
sS's'
I18
sS'm'
(lp29221
I1
aI1
aI7
aI3
aI27
aI51
aI121
aI93
aI349
aI125
aI2013
aI1671
aI8049
aI7807
aI7291
aI64413
aI93625
aI245611
asS'd'
I14611
sa(dp29222
S'a'
I20142
sS's'
I18
sS'm'
(lp29223
I1
aI3
aI3
aI9
aI11
aI1
aI91
aI137
aI501
aI617
aI1513
aI799
aI1705
aI15737
aI14989
aI53611
aI48781
aI64481
asS'd'
I14612
sa(dp29224
S'a'
I20159
sS's'
I18
sS'm'
(lp29225
I1
aI3
aI3
aI1
aI21
aI55
aI95
aI79
aI383
aI617
aI1589
aI2671
aI4057
aI13525
aI9269
aI23539
aI13317
aI87701
asS'd'
I14613
sa(dp29226
S'a'
I20173
sS's'
I18
sS'm'
(lp29227
I1
aI1
aI3
aI1
aI29
aI17
aI121
aI45
aI91
aI215
aI325
aI2853
aI1213
aI10221
aI7233
aI34063
aI21887
aI142943
asS'd'
I14614
sa(dp29228
S'a'
I20186
sS's'
I18
sS'm'
(lp29229
I1
aI3
aI3
aI11
aI27
aI53
aI55
aI149
aI107
aI379
aI441
aI585
aI5697
aI16353
aI5613
aI4323
aI55315
aI197603
asS'd'
I14615
sa(dp29230
S'a'
I20229
sS's'
I18
sS'm'
(lp29231
I1
aI3
aI7
aI3
aI31
aI9
aI71
aI175
aI485
aI35
aI675
aI2091
aI2351
aI7985
aI14207
aI52687
aI8559
aI1067
asS'd'
I14616
sa(dp29232
S'a'
I20234
sS's'
I18
sS'm'
(lp29233
I1
aI1
aI5
aI15
aI29
aI37
aI9
aI73
aI357
aI961
aI489
aI875
aI7465
aI3231
aI27821
aI42499
aI127837
aI117215
asS'd'
I14617
sa(dp29234
S'a'
I20263
sS's'
I18
sS'm'
(lp29235
I1
aI3
aI7
aI7
aI19
aI43
aI75
aI153
aI27
aI291
aI2039
aI2661
aI5513
aI13429
aI27307
aI5305
aI44771
aI200621
asS'd'
I14618
sa(dp29236
S'a'
I20270
sS's'
I18
sS'm'
(lp29237
I1
aI1
aI1
aI15
aI5
aI39
aI61
aI107
aI201
aI485
aI319
aI335
aI5537
aI14195
aI31861
aI63637
aI68497
aI45637
asS'd'
I14619
sa(dp29238
S'a'
I20299
sS's'
I18
sS'm'
(lp29239
I1
aI3
aI7
aI7
aI23
aI49
aI95
aI225
aI25
aI933
aI667
aI2993
aI2181
aI15659
aI31343
aI20249
aI57039
aI43399
asS'd'
I14620
sa(dp29240
S'a'
I20304
sS's'
I18
sS'm'
(lp29241
I1
aI3
aI3
aI7
aI17
aI25
aI29
aI243
aI511
aI91
aI1409
aI203
aI2749
aI7067
aI12471
aI41737
aI32761
aI7535
asS'd'
I14621
sa(dp29242
S'a'
I20319
sS's'
I18
sS'm'
(lp29243
I1
aI1
aI7
aI9
aI27
aI43
aI63
aI65
aI325
aI817
aI1127
aI2039
aI6171
aI5867
aI10593
aI17205
aI95913
aI207417
asS'd'
I14622
sa(dp29244
S'a'
I20329
sS's'
I18
sS'm'
(lp29245
I1
aI1
aI7
aI3
aI3
aI51
aI107
aI153
aI193
aI579
aI593
aI2915
aI7641
aI5157
aI1131
aI29793
aI66579
aI81903
asS'd'
I14623
sa(dp29246
S'a'
I20337
sS's'
I18
sS'm'
(lp29247
I1
aI1
aI5
aI15
aI19
aI61
aI125
aI107
aI235
aI513
aI1897
aI875
aI6341
aI1817
aI10631
aI63905
aI42993
aI150699
asS'd'
I14624
sa(dp29248
S'a'
I20353
sS's'
I18
sS'm'
(lp29249
I1
aI1
aI3
aI11
aI27
aI1
aI93
aI107
aI325
aI459
aI1733
aI2527
aI4557
aI2277
aI19345
aI8205
aI67337
aI242559
asS'd'
I14625
sa(dp29250
S'a'
I20401
sS's'
I18
sS'm'
(lp29251
I1
aI1
aI7
aI3
aI3
aI45
aI27
aI227
aI201
aI99
aI589
aI1665
aI4851
aI2655
aI9915
aI41321
aI59865
aI71501
asS'd'
I14626
sa(dp29252
S'a'
I20434
sS's'
I18
sS'm'
(lp29253
I1
aI3
aI1
aI9
aI3
aI25
aI117
aI199
aI125
aI849
aI135
aI1771
aI4743
aI13475
aI23711
aI17389
aI52711
aI200143
asS'd'
I14627
sa(dp29254
S'a'
I20436
sS's'
I18
sS'm'
(lp29255
I1
aI3
aI1
aI3
aI11
aI23
aI67
aI155
aI133
aI557
aI1933
aI3169
aI1707
aI16045
aI11039
aI13889
aI71045
aI245885
asS'd'
I14628
sa(dp29256
S'a'
I20473
sS's'
I18
sS'm'
(lp29257
I1
aI1
aI5
aI9
aI1
aI7
aI99
aI13
aI315
aI251
aI1289
aI225
aI2847
aI8451
aI3139
aI46829
aI124745
aI64825
asS'd'
I14629
sa(dp29258
S'a'
I20488
sS's'
I18
sS'm'
(lp29259
I1
aI3
aI7
aI13
aI19
aI45
aI87
aI161
aI271
aI401
aI1995
aI935
aI1803
aI4051
aI11709
aI26993
aI120139
aI147895
asS'd'
I14630
sa(dp29260
S'a'
I20512
sS's'
I18
sS'm'
(lp29261
I1
aI1
aI7
aI5
aI15
aI11
aI47
aI215
aI51
aI1019
aI2039
aI3767
aI929
aI3845
aI3939
aI64077
aI48115
aI61845
asS'd'
I14631
sa(dp29262
S'a'
I20515
sS's'
I18
sS'm'
(lp29263
I1
aI1
aI1
aI5
aI1
aI39
aI15
aI77
aI179
aI13
aI1099
aI203
aI3363
aI9071
aI12033
aI49159
aI71137
aI124177
asS'd'
I14632
sa(dp29264
S'a'
I20517
sS's'
I18
sS'm'
(lp29265
I1
aI3
aI5
aI5
aI1
aI31
aI83
aI219
aI387
aI347
aI1099
aI925
aI4423
aI5081
aI15981
aI35881
aI79131
aI248301
asS'd'
I14633
sa(dp29266
S'a'
I20530
sS's'
I18
sS'm'
(lp29267
I1
aI3
aI7
aI3
aI25
aI19
aI53
aI43
aI347
aI845
aI1735
aI3237
aI2795
aI2253
aI2997
aI43729
aI122833
aI124869
asS'd'
I14634
sa(dp29268
S'a'
I20571
sS's'
I18
sS'm'
(lp29269
I1
aI3
aI5
aI11
aI5
aI19
aI93
aI55
aI297
aI231
aI239
aI3335
aI253
aI13607
aI16769
aI48879
aI61439
aI54827
asS'd'
I14635
sa(dp29270
S'a'
I20574
sS's'
I18
sS'm'
(lp29271
I1
aI3
aI7
aI11
aI11
aI55
aI121
aI73
aI19
aI1017
aI727
aI579
aI8011
aI9559
aI15051
aI7895
aI17609
aI103061
asS'd'
I14636
sa(dp29272
S'a'
I20589
sS's'
I18
sS'm'
(lp29273
I1
aI1
aI7
aI5
aI19
aI47
aI85
aI195
aI75
aI1003
aI439
aI3069
aI2107
aI12751
aI26729
aI2329
aI1191
aI86547
asS'd'
I14637
sa(dp29274
S'a'
I20592
sS's'
I18
sS'm'
(lp29275
I1
aI3
aI3
aI9
aI5
aI31
aI63
aI227
aI481
aI793
aI1853
aI1491
aI2109
aI4199
aI32149
aI45229
aI54685
aI124819
asS'd'
I14638
sa(dp29276
S'a'
I20611
sS's'
I18
sS'm'
(lp29277
I1
aI1
aI1
aI3
aI15
aI15
aI41
aI45
aI153
aI429
aI1691
aI1897
aI7253
aI7239
aI26133
aI36527
aI90319
aI186097
asS'd'
I14639
sa(dp29278
S'a'
I20613
sS's'
I18
sS'm'
(lp29279
I1
aI3
aI1
aI13
aI15
aI33
aI103
aI113
aI121
aI387
aI177
aI1943
aI3181
aI5483
aI18515
aI38807
aI22655
aI59787
asS'd'
I14640
sa(dp29280
S'a'
I20628
sS's'
I18
sS'm'
(lp29281
I1
aI1
aI5
aI7
aI15
aI3
aI53
aI155
aI99
aI133
aI579
aI2129
aI6881
aI11091
aI26715
aI15485
aI108071
aI230881
asS'd'
I14641
sa(dp29282
S'a'
I20637
sS's'
I18
sS'm'
(lp29283
I1
aI1
aI3
aI13
aI25
aI61
aI91
aI81
aI9
aI1011
aI1993
aI2485
aI3707
aI11127
aI21279
aI15853
aI104081
aI203769
asS'd'
I14642
sa(dp29284
S'a'
I20638
sS's'
I18
sS'm'
(lp29285
I1
aI1
aI7
aI3
aI17
aI23
aI37
aI171
aI315
aI247
aI275
aI3215
aI7139
aI11739
aI25859
aI34803
aI124601
aI9169
asS'd'
I14643
sa(dp29286
S'a'
I20719
sS's'
I18
sS'm'
(lp29287
I1
aI3
aI7
aI9
aI21
aI29
aI97
aI213
aI309
aI865
aI597
aI1811
aI5547
aI3741
aI31927
aI53379
aI43293
aI23589
asS'd'
I14644
sa(dp29288
S'a'
I20724
sS's'
I18
sS'm'
(lp29289
I1
aI3
aI7
aI9
aI7
aI43
aI107
aI187
aI485
aI977
aI1329
aI3037
aI3701
aI9667
aI13581
aI6283
aI39221
aI63841
asS'd'
I14645
sa(dp29290
S'a'
I20772
sS's'
I18
sS'm'
(lp29291
I1
aI1
aI3
aI11
aI3
aI51
aI117
aI45
aI293
aI409
aI689
aI153
aI1163
aI10921
aI22709
aI30415
aI120475
aI120751
asS'd'
I14646
sa(dp29292
S'a'
I20796
sS's'
I18
sS'm'
(lp29293
I1
aI3
aI5
aI15
aI31
aI59
aI57
aI63
aI249
aI763
aI1627
aI3039
aI4309
aI14115
aI25489
aI35009
aI126609
aI146041
asS'd'
I14647
sa(dp29294
S'a'
I20799
sS's'
I18
sS'm'
(lp29295
I1
aI1
aI1
aI9
aI3
aI47
aI21
aI183
aI495
aI361
aI1439
aI407
aI5757
aI12645
aI11425
aI1923
aI94511
aI205127
asS'd'
I14648
sa(dp29296
S'a'
I20816
sS's'
I18
sS'm'
(lp29297
I1
aI3
aI1
aI9
aI15
aI5
aI101
aI107
aI385
aI175
aI791
aI901
aI4427
aI10415
aI8163
aI14417
aI62997
aI139309
asS'd'
I14649
sa(dp29298
S'a'
I20841
sS's'
I18
sS'm'
(lp29299
I1
aI3
aI5
aI3
aI13
aI57
aI9
aI99
aI77
aI123
aI1607
aI3643
aI3879
aI503
aI6021
aI60211
aI106471
aI221801
asS'd'
I14650
sa(dp29300
S'a'
I20844
sS's'
I18
sS'm'
(lp29301
I1
aI1
aI7
aI5
aI27
aI35
aI11
aI33
aI415
aI387
aI1461
aI741
aI55
aI15095
aI21177
aI5715
aI109893
aI204843
asS'd'
I14651
sa(dp29302
S'a'
I20862
sS's'
I18
sS'm'
(lp29303
I1
aI1
aI1
aI15
aI7
aI49
aI51
aI81
aI157
aI421
aI279
aI1951
aI6847
aI10259
aI31925
aI60761
aI12395
aI49511
asS'd'
I14652
sa(dp29304
S'a'
I20865
sS's'
I18
sS'm'
(lp29305
I1
aI3
aI7
aI11
aI5
aI33
aI27
aI135
aI247
aI813
aI1889
aI2547
aI2359
aI9535
aI4141
aI59713
aI88685
aI214641
asS'd'
I14653
sa(dp29306
S'a'
I20902
sS's'
I18
sS'm'
(lp29307
I1
aI1
aI5
aI15
aI17
aI61
aI99
aI103
aI39
aI151
aI1033
aI2743
aI6639
aI5271
aI22059
aI12681
aI22763
aI88255
asS'd'
I14654
sa(dp29308
S'a'
I20938
sS's'
I18
sS'm'
(lp29309
I1
aI1
aI7
aI13
aI5
aI11
aI39
aI139
aI353
aI989
aI1391
aI169
aI3709
aI735
aI22965
aI227
aI103623
aI153893
asS'd'
I14655
sa(dp29310
S'a'
I20957
sS's'
I18
sS'm'
(lp29311
I1
aI1
aI7
aI3
aI9
aI51
aI53
aI87
aI411
aI617
aI671
aI681
aI5057
aI6003
aI23137
aI30881
aI2289
aI187133
asS'd'
I14656
sa(dp29312
S'a'
I20971
sS's'
I18
sS'm'
(lp29313
I1
aI1
aI7
aI7
aI17
aI59
aI77
aI219
aI25
aI53
aI145
aI129
aI4289
aI14257
aI7159
aI44833
aI22131
aI53393
asS'd'
I14657
sa(dp29314
S'a'
I20973
sS's'
I18
sS'm'
(lp29315
I1
aI3
aI7
aI1
aI9
aI59
aI79
aI177
aI149
aI637
aI1641
aI3713
aI2709
aI12321
aI5691
aI18239
aI8617
aI225979
asS'd'
I14658
sa(dp29316
S'a'
I20981
sS's'
I18
sS'm'
(lp29317
I1
aI1
aI7
aI5
aI9
aI9
aI67
aI51
aI451
aI815
aI295
aI813
aI1257
aI179
aI28769
aI57241
aI51753
aI164873
asS'd'
I14659
sa(dp29318
S'a'
I20982
sS's'
I18
sS'm'
(lp29319
I1
aI3
aI1
aI9
aI19
aI61
aI53
aI65
aI29
aI503
aI715
aI1837
aI7487
aI16187
aI27303
aI54681
aI98753
aI100471
asS'd'
I14660
sa(dp29320
S'a'
I20985
sS's'
I18
sS'm'
(lp29321
I1
aI1
aI3
aI1
aI15
aI51
aI1
aI79
aI179
aI367
aI841
aI1313
aI797
aI4777
aI1369
aI13317
aI65059
aI204877
asS'd'
I14661
sa(dp29322
S'a'
I20991
sS's'
I18
sS'm'
(lp29323
I1
aI3
aI5
aI9
aI15
aI19
aI109
aI45
aI473
aI517
aI1139
aI15
aI1997
aI4245
aI11169
aI56417
aI75017
aI37957
asS'd'
I14662
sa(dp29324
S'a'
I21012
sS's'
I18
sS'm'
(lp29325
I1
aI1
aI7
aI1
aI3
aI41
aI75
aI95
aI59
aI503
aI1439
aI2633
aI3527
aI5363
aI24357
aI43659
aI10387
aI208319
asS'd'
I14663
sa(dp29326
S'a'
I21022
sS's'
I18
sS'm'
(lp29327
I1
aI1
aI5
aI1
aI31
aI7
aI71
aI231
aI505
aI241
aI1579
aI3517
aI3995
aI8269
aI6793
aI15883
aI102779
aI75589
asS'd'
I14664
sa(dp29328
S'a'
I21026
sS's'
I18
sS'm'
(lp29329
I1
aI3
aI5
aI1
aI13
aI61
aI87
aI213
aI501
aI307
aI1629
aI2715
aI7245
aI747
aI20601
aI28105
aI79249
aI76231
asS'd'
I14665
sa(dp29330
S'a'
I21028
sS's'
I18
sS'm'
(lp29331
I1
aI1
aI7
aI11
aI5
aI13
aI69
aI221
aI485
aI59
aI2027
aI483
aI6851
aI11719
aI16787
aI54111
aI47579
aI49959
asS'd'
I14666
sa(dp29332
S'a'
I21050
sS's'
I18
sS'm'
(lp29333
I1
aI1
aI3
aI15
aI3
aI33
aI57
aI75
aI375
aI45
aI851
aI1673
aI8167
aI867
aI32087
aI34157
aI96701
aI72893
asS'd'
I14667
sa(dp29334
S'a'
I21075
sS's'
I18
sS'm'
(lp29335
I1
aI1
aI3
aI1
aI21
aI31
aI65
aI85
aI181
aI453
aI815
aI3139
aI205
aI429
aI7451
aI50855
aI41085
aI137927
asS'd'
I14668
sa(dp29336
S'a'
I21077
sS's'
I18
sS'm'
(lp29337
I1
aI3
aI1
aI9
aI3
aI57
aI99
aI183
aI305
aI991
aI809
aI4021
aI3131
aI4459
aI5839
aI32493
aI116541
aI59329
asS'd'
I14669
sa(dp29338
S'a'
I21078
sS's'
I18
sS'm'
(lp29339
I1
aI3
aI7
aI1
aI5
aI19
aI3
aI91
aI297
aI715
aI1081
aI445
aI393
aI12685
aI4457
aI61437
aI103701
aI75917
asS'd'
I14670
sa(dp29340
S'a'
I21106
sS's'
I18
sS'm'
(lp29341
I1
aI1
aI7
aI15
aI17
aI39
aI19
aI255
aI247
aI391
aI1055
aI1241
aI4515
aI10217
aI23363
aI40301
aI115053
aI234349
asS'd'
I14671
sa(dp29342
S'a'
I21122
sS's'
I18
sS'm'
(lp29343
I1
aI3
aI5
aI1
aI21
aI9
aI33
aI243
aI501
aI793
aI219
aI3595
aI2585
aI5083
aI15377
aI35761
aI90609
aI93761
asS'd'
I14672
sa(dp29344
S'a'
I21127
sS's'
I18
sS'm'
(lp29345
I1
aI1
aI5
aI13
aI3
aI1
aI5
aI77
aI265
aI525
aI1107
aI1879
aI1119
aI2277
aI30557
aI43547
aI81947
aI134075
asS'd'
I14673
sa(dp29346
S'a'
I21155
sS's'
I18
sS'm'
(lp29347
I1
aI1
aI7
aI5
aI11
aI47
aI71
aI83
aI255
aI183
aI515
aI2591
aI3933
aI16025
aI16727
aI43421
aI18725
aI106675
asS'd'
I14674
sa(dp29348
S'a'
I21167
sS's'
I18
sS'm'
(lp29349
I1
aI3
aI1
aI1
aI5
aI17
aI57
aI209
aI509
aI421
aI1247
aI3153
aI1835
aI8777
aI13285
aI27699
aI34001
aI186553
asS'd'
I14675
sa(dp29350
S'a'
I21169
sS's'
I18
sS'm'
(lp29351
I1
aI1
aI3
aI11
aI27
aI19
aI73
aI65
aI179
aI115
aI845
aI2507
aI7673
aI14429
aI10553
aI4999
aI82323
aI247379
asS'd'
I14676
sa(dp29352
S'a'
I21204
sS's'
I18
sS'm'
(lp29353
I1
aI3
aI7
aI7
aI17
aI59
aI97
aI183
aI407
aI697
aI1423
aI123
aI6479
aI3997
aI729
aI31587
aI114383
aI61673
asS'd'
I14677
sa(dp29354
S'a'
I21230
sS's'
I18
sS'm'
(lp29355
I1
aI1
aI5
aI9
aI21
aI23
aI21
aI153
aI187
aI255
aI125
aI1469
aI2639
aI8099
aI29689
aI36415
aI103959
aI231621
asS'd'
I14678
sa(dp29356
S'a'
I21256
sS's'
I18
sS'm'
(lp29357
I1
aI1
aI3
aI11
aI19
aI59
aI115
aI205
aI123
aI133
aI1953
aI3471
aI2495
aI329
aI32385
aI21931
aI9691
aI51405
asS'd'
I14679
sa(dp29358
S'a'
I21285
sS's'
I18
sS'm'
(lp29359
I1
aI3
aI3
aI13
aI7
aI7
aI115
aI65
aI301
aI621
aI1091
aI2137
aI5729
aI5027
aI21331
aI24803
aI114789
aI142039
asS'd'
I14680
sa(dp29360
S'a'
I21312
sS's'
I18
sS'm'
(lp29361
I1
aI1
aI7
aI5
aI31
aI19
aI103
aI69
aI503
aI663
aI1497
aI2867
aI5295
aI893
aI15927
aI37513
aI94553
aI72369
asS'd'
I14681
sa(dp29362
S'a'
I21329
sS's'
I18
sS'm'
(lp29363
I1
aI3
aI3
aI15
aI17
aI33
aI99
aI249
aI277
aI259
aI9
aI99
aI3073
aI12017
aI14847
aI7685
aI102499
aI26489
asS'd'
I14682
sa(dp29364
S'a'
I21351
sS's'
I18
sS'm'
(lp29365
I1
aI1
aI1
aI1
aI5
aI23
aI31
aI45
aI29
aI483
aI1977
aI1129
aI6925
aI2273
aI16573
aI53039
aI90251
aI137191
asS'd'
I14683
sa(dp29366
S'a'
I21372
sS's'
I18
sS'm'
(lp29367
I1
aI3
aI1
aI13
aI27
aI47
aI29
aI51
aI473
aI895
aI671
aI3917
aI6905
aI15769
aI9019
aI28879
aI120591
aI220753
asS'd'
I14684
sa(dp29368
S'a'
I21376
sS's'
I18
sS'm'
(lp29369
I1
aI3
aI1
aI13
aI27
aI29
aI53
aI255
aI507
aI819
aI1251
aI2463
aI1717
aI14461
aI31997
aI30829
aI8803
aI115539
asS'd'
I14685
sa(dp29370
S'a'
I21424
sS's'
I18
sS'm'
(lp29371
I1
aI3
aI3
aI15
aI27
aI1
aI109
aI225
aI451
aI409
aI2025
aI2701
aI4121
aI9949
aI1551
aI13625
aI73577
aI211549
asS'd'
I14686
sa(dp29372
S'a'
I21448
sS's'
I18
sS'm'
(lp29373
I1
aI1
aI1
aI3
aI23
aI57
aI49
aI35
aI365
aI711
aI2001
aI997
aI1853
aI2913
aI15667
aI30255
aI19535
aI2171
asS'd'
I14687
sa(dp29374
S'a'
I21465
sS's'
I18
sS'm'
(lp29375
I1
aI1
aI7
aI1
aI21
aI37
aI127
aI3
aI117
aI449
aI1689
aI1391
aI1427
aI12641
aI15199
aI23769
aI66553
aI34669
asS'd'
I14688
sa(dp29376
S'a'
I21495
sS's'
I18
sS'm'
(lp29377
I1
aI3
aI7
aI9
aI31
aI45
aI51
aI137
aI181
aI469
aI573
aI89
aI7257
aI10991
aI30705
aI37827
aI75071
aI152885
asS'd'
I14689
sa(dp29378
S'a'
I21509
sS's'
I18
sS'm'
(lp29379
I1
aI3
aI1
aI1
aI19
aI13
aI55
aI223
aI261
aI353
aI1497
aI183
aI8173
aI14421
aI9977
aI24095
aI47215
aI155189
asS'd'
I14690
sa(dp29380
S'a'
I21550
sS's'
I18
sS'm'
(lp29381
I1
aI1
aI3
aI15
aI15
aI41
aI31
aI105
aI459
aI27
aI299
aI159
aI2167
aI14809
aI9983
aI2755
aI121715
aI35921
asS'd'
I14691
sa(dp29382
S'a'
I21562
sS's'
I18
sS'm'
(lp29383
I1
aI1
aI3
aI7
aI31
aI5
aI85
aI137
aI431
aI849
aI1479
aI2681
aI167
aI5727
aI3211
aI30765
aI63295
aI39509
asS'd'
I14692
sa(dp29384
S'a'
I21575
sS's'
I18
sS'm'
(lp29385
I1
aI1
aI5
aI7
aI5
aI51
aI21
aI103
aI175
aI927
aI1115
aI1507
aI505
aI8093
aI25831
aI54303
aI40397
aI61249
asS'd'
I14693
sa(dp29386
S'a'
I21579
sS's'
I18
sS'm'
(lp29387
I1
aI1
aI7
aI3
aI23
aI53
aI49
aI225
aI7
aI425
aI403
aI3949
aI1081
aI15335
aI21737
aI647
aI107875
aI236183
asS'd'
I14694
sa(dp29388
S'a'
I21582
sS's'
I18
sS'm'
(lp29389
I1
aI1
aI5
aI7
aI17
aI21
aI85
aI229
aI325
aI57
aI601
aI2785
aI6417
aI5135
aI17917
aI12861
aI97675
aI115457
asS'd'
I14695
sa(dp29390
S'a'
I21600
sS's'
I18
sS'm'
(lp29391
I1
aI1
aI3
aI3
aI13
aI23
aI73
aI111
aI385
aI47
aI605
aI1169
aI1729
aI2335
aI18739
aI61293
aI41915
aI237645
asS'd'
I14696
sa(dp29392
S'a'
I21615
sS's'
I18
sS'm'
(lp29393
I1
aI1
aI1
aI5
aI31
aI11
aI123
aI13
aI465
aI755
aI1073
aI1885
aI2105
aI5971
aI2347
aI10911
aI125823
aI156037
asS'd'
I14697
sa(dp29394
S'a'
I21617
sS's'
I18
sS'm'
(lp29395
I1
aI3
aI7
aI11
aI17
aI47
aI3
aI165
aI227
aI355
aI87
aI839
aI7741
aI12275
aI28579
aI25337
aI87671
aI224847
asS'd'
I14698
sa(dp29396
S'a'
I21624
sS's'
I18
sS'm'
(lp29397
I1
aI1
aI5
aI15
aI23
aI33
aI9
aI1
aI257
aI121
aI1049
aI1009
aI187
aI9935
aI26093
aI21921
aI130247
aI240291
asS'd'
I14699
sa(dp29398
S'a'
I21633
sS's'
I18
sS'm'
(lp29399
I1
aI1
aI3
aI13
aI13
aI27
aI87
aI221
aI27
aI117
aI551
aI2533
aI7611
aI5333
aI14635
aI9911
aI37555
aI250621
asS'd'
I14700
sa(dp29400
S'a'
I21636
sS's'
I18
sS'm'
(lp29401
I1
aI3
aI7
aI15
aI29
aI39
aI33
aI1
aI495
aI889
aI1397
aI3415
aI7193
aI11533
aI27379
aI36425
aI13739
aI146635
asS'd'
I14701
sa(dp29402
S'a'
I21645
sS's'
I18
sS'm'
(lp29403
I1
aI1
aI7
aI11
aI1
aI23
aI85
aI127
aI79
aI989
aI321
aI1913
aI7571
aI9889
aI11803
aI1307
aI120513
aI218077
asS'd'
I14702
sa(dp29404
S'a'
I21654
sS's'
I18
sS'm'
(lp29405
I1
aI1
aI7
aI5
aI5
aI15
aI35
aI9
aI351
aI973
aI1455
aI2043
aI5527
aI9431
aI16059
aI53915
aI105785
aI180579
asS'd'
I14703
sa(dp29406
S'a'
I21660
sS's'
I18
sS'm'
(lp29407
I1
aI1
aI1
aI13
aI13
aI45
aI15
aI41
aI131
aI463
aI1011
aI3559
aI6393
aI4737
aI6041
aI33073
aI60989
aI56761
asS'd'
I14704
sa(dp29408
S'a'
I21667
sS's'
I18
sS'm'
(lp29409
I1
aI1
aI3
aI9
aI31
aI35
aI23
aI133
aI33
aI233
aI543
aI957
aI4913
aI12441
aI10293
aI31611
aI83383
aI154551
asS'd'
I14705
sa(dp29410
S'a'
I21702
sS's'
I18
sS'm'
(lp29411
I1
aI3
aI3
aI1
aI29
aI37
aI117
aI247
aI345
aI197
aI1617
aI3333
aI7901
aI8343
aI55
aI16529
aI34627
aI172703
asS'd'
I14706
sa(dp29412
S'a'
I21714
sS's'
I18
sS'm'
(lp29413
I1
aI1
aI1
aI13
aI23
aI51
aI7
aI219
aI503
aI215
aI375
aI2275
aI5467
aI13953
aI13987
aI22735
aI67505
aI185977
asS'd'
I14707
sa(dp29414
S'a'
I21719
sS's'
I18
sS'm'
(lp29415
I1
aI3
aI5
aI5
aI29
aI53
aI85
aI147
aI167
aI409
aI853
aI667
aI4431
aI5227
aI15535
aI34375
aI107135
aI220637
asS'd'
I14708
sa(dp29416
S'a'
I21736
sS's'
I18
sS'm'
(lp29417
I1
aI3
aI7
aI7
aI19
aI3
aI73
aI123
aI455
aI539
aI1735
aI1423
aI5337
aI16311
aI15469
aI36071
aI126437
aI219249
asS'd'
I14709
sa(dp29418
S'a'
I21767
sS's'
I18
sS'm'
(lp29419
I1
aI1
aI1
aI3
aI19
aI49
aI17
aI133
aI101
aI1013
aI683
aI869
aI6267
aI409
aI31379
aI2535
aI8039
aI63205
asS'd'
I14710
sa(dp29420
S'a'
I21781
sS's'
I18
sS'm'
(lp29421
I1
aI1
aI1
aI13
aI13
aI53
aI25
aI31
aI501
aI629
aI645
aI1811
aI3675
aI13317
aI17009
aI7359
aI85475
aI249823
asS'd'
I14711
sa(dp29422
S'a'
I21795
sS's'
I18
sS'm'
(lp29423
I1
aI1
aI3
aI11
aI5
aI1
aI41
aI17
aI159
aI361
aI1439
aI2083
aI1425
aI7221
aI9117
aI59543
aI59285
aI188615
asS'd'
I14712
sa(dp29424
S'a'
I21841
sS's'
I18
sS'm'
(lp29425
I1
aI3
aI7
aI1
aI11
aI27
aI71
aI121
aI471
aI749
aI1983
aI3715
aI6463
aI5793
aI1063
aI18201
aI189
aI243751
asS'd'
I14713
sa(dp29426
S'a'
I21853
sS's'
I18
sS'm'
(lp29427
I1
aI3
aI3
aI11
aI19
aI17
aI15
aI175
aI379
aI683
aI1491
aI2385
aI6981
aI1183
aI16829
aI2103
aI9309
aI46119
asS'd'
I14714
sa(dp29428
S'a'
I21867
sS's'
I18
sS'm'
(lp29429
I1
aI1
aI7
aI5
aI17
aI39
aI109
aI9
aI279
aI309
aI1
aI1523
aI4551
aI3855
aI13277
aI36125
aI54191
aI45085
asS'd'
I14715
sa(dp29430
S'a'
I21878
sS's'
I18
sS'm'
(lp29431
I1
aI3
aI5
aI3
aI9
aI59
aI51
aI5
aI431
aI657
aI161
aI2725
aI2401
aI9743
aI12925
aI43501
aI51551
aI163737
asS'd'
I14716
sa(dp29432
S'a'
I21891
sS's'
I18
sS'm'
(lp29433
I1
aI1
aI5
aI3
aI9
aI13
aI7
aI177
aI121
aI795
aI1169
aI3169
aI3793
aI3995
aI29027
aI32967
aI82273
aI207939
asS'd'
I14717
sa(dp29434
S'a'
I21897
sS's'
I18
sS'm'
(lp29435
I1
aI3
aI7
aI13
aI1
aI1
aI31
aI91
aI245
aI775
aI1589
aI2263
aI6303
aI15787
aI3111
aI52553
aI52507
aI183971
asS'd'
I14718
sa(dp29436
S'a'
I21954
sS's'
I18
sS'm'
(lp29437
I1
aI3
aI5
aI11
aI15
aI49
aI73
aI191
aI67
aI449
aI1245
aI2445
aI5617
aI8625
aI27971
aI35939
aI76907
aI76207
asS'd'
I14719
sa(dp29438
S'a'
I21956
sS's'
I18
sS'm'
(lp29439
I1
aI3
aI1
aI11
aI15
aI47
aI29
aI91
aI437
aI895
aI1941
aI249
aI2739
aI15479
aI29699
aI7257
aI39897
aI65985
asS'd'
I14720
sa(dp29440
S'a'
I21965
sS's'
I18
sS'm'
(lp29441
I1
aI3
aI5
aI7
aI13
aI23
aI45
aI113
aI297
aI373
aI1505
aI2317
aI7509
aI12059
aI13737
aI29081
aI87337
aI221917
asS'd'
I14721
sa(dp29442
S'a'
I21977
sS's'
I18
sS'm'
(lp29443
I1
aI3
aI3
aI3
aI1
aI5
aI13
aI215
aI221
aI461
aI1337
aI3569
aI2257
aI12135
aI14685
aI39721
aI16723
aI234791
asS'd'
I14722
sa(dp29444
S'a'
I21978
sS's'
I18
sS'm'
(lp29445
I1
aI1
aI7
aI5
aI11
aI25
aI71
aI103
aI87
aI533
aI779
aI379
aI6695
aI13451
aI24801
aI49235
aI35109
aI100865
asS'd'
I14723
sa(dp29446
S'a'
I21983
sS's'
I18
sS'm'
(lp29447
I1
aI3
aI3
aI15
aI13
aI51
aI27
aI11
aI279
aI847
aI135
aI1119
aI2765
aI3805
aI20273
aI29089
aI83379
aI190353
asS'd'
I14724
sa(dp29448
S'a'
I22002
sS's'
I18
sS'm'
(lp29449
I1
aI3
aI7
aI5
aI17
aI29
aI111
aI35
aI189
aI273
aI503
aI541
aI6691
aI9051
aI10403
aI7559
aI54787
aI25403
asS'd'
I14725
sa(dp29450
S'a'
I22013
sS's'
I18
sS'm'
(lp29451
I1
aI3
aI5
aI9
aI21
aI29
aI85
aI235
aI223
aI677
aI71
aI1313
aI6587
aI10983
aI199
aI27721
aI78627
aI105505
asS'd'
I14726
sa(dp29452
S'a'
I22014
sS's'
I18
sS'm'
(lp29453
I1
aI3
aI1
aI3
aI13
aI3
aI123
aI115
aI173
aI907
aI1555
aI1489
aI2745
aI6451
aI25347
aI24105
aI66471
aI181009
asS'd'
I14727
sa(dp29454
S'a'
I22054
sS's'
I18
sS'm'
(lp29455
I1
aI1
aI7
aI1
aI31
aI15
aI13
aI97
aI511
aI827
aI1193
aI3081
aI1517
aI13511
aI24887
aI39239
aI85175
aI150213
asS'd'
I14728
sa(dp29456
S'a'
I22058
sS's'
I18
sS'm'
(lp29457
I1
aI1
aI1
aI5
aI17
aI39
aI121
aI67
aI207
aI877
aI1885
aI171
aI2687
aI13081
aI27267
aI58699
aI118575
aI213025
asS'd'
I14729
sa(dp29458
S'a'
I22066
sS's'
I18
sS'm'
(lp29459
I1
aI1
aI3
aI9
aI9
aI27
aI101
aI215
aI31
aI37
aI1629
aI3631
aI3225
aI9667
aI31547
aI41939
aI38683
aI150805
asS'd'
I14730
sa(dp29460
S'a'
I22085
sS's'
I18
sS'm'
(lp29461
I1
aI3
aI1
aI11
aI11
aI59
aI17
aI15
aI187
aI667
aI747
aI2193
aI6749
aI6019
aI31805
aI52433
aI4141
aI52613
asS'd'
I14731
sa(dp29462
S'a'
I22103
sS's'
I18
sS'm'
(lp29463
I1
aI3
aI3
aI13
aI9
aI1
aI51
aI101
aI213
aI881
aI899
aI2197
aI3017
aI1591
aI9271
aI44017
aI99893
aI192005
asS'd'
I14732
sa(dp29464
S'a'
I22138
sS's'
I18
sS'm'
(lp29465
I1
aI3
aI7
aI13
aI23
aI41
aI79
aI83
aI123
aI585
aI49
aI849
aI2133
aI12473
aI6907
aI15487
aI45783
aI46609
asS'd'
I14733
sa(dp29466
S'a'
I22140
sS's'
I18
sS'm'
(lp29467
I1
aI3
aI7
aI13
aI27
aI23
aI71
aI13
aI319
aI903
aI1123
aI933
aI2603
aI11631
aI19953
aI47001
aI127751
aI84547
asS'd'
I14734
sa(dp29468
S'a'
I22153
sS's'
I18
sS'm'
(lp29469
I1
aI1
aI1
aI15
aI3
aI61
aI79
aI231
aI43
aI217
aI801
aI997
aI6545
aI13657
aI25589
aI30435
aI49497
aI1037
asS'd'
I14735
sa(dp29470
S'a'
I22164
sS's'
I18
sS'm'
(lp29471
I1
aI3
aI3
aI3
aI21
aI29
aI121
aI35
aI129
aI239
aI1645
aI3147
aI7647
aI1201
aI19287
aI7075
aI67961
aI62481
asS'd'
I14736
sa(dp29472
S'a'
I22167
sS's'
I18
sS'm'
(lp29473
I1
aI3
aI7
aI7
aI3
aI23
aI45
aI177
aI469
aI897
aI359
aI2521
aI2079
aI985
aI14993
aI56813
aI20667
aI187341
asS'd'
I14737
sa(dp29474
S'a'
I22189
sS's'
I18
sS'm'
(lp29475
I1
aI3
aI5
aI7
aI23
aI53
aI15
aI45
aI297
aI93
aI247
aI1165
aI2683
aI5899
aI7113
aI14859
aI22733
aI173835
asS'd'
I14738
sa(dp29476
S'a'
I22202
sS's'
I18
sS'm'
(lp29477
I1
aI1
aI3
aI15
aI23
aI17
aI43
aI179
aI103
aI197
aI1857
aI323
aI267
aI12417
aI2343
aI41527
aI12243
aI112023
asS'd'
I14739
sa(dp29478
S'a'
I22212
sS's'
I18
sS'm'
(lp29479
I1
aI3
aI7
aI13
aI7
aI43
aI75
aI19
aI169
aI621
aI735
aI141
aI3087
aI765
aI5901
aI34029
aI117603
aI5137
asS'd'
I14740
sa(dp29480
S'a'
I22234
sS's'
I18
sS'm'
(lp29481
I1
aI1
aI3
aI5
aI15
aI17
aI67
aI177
aI371
aI249
aI99
aI1651
aI3701
aI343
aI435
aI50307
aI33915
aI115391
asS'd'
I14741
sa(dp29482
S'a'
I22239
sS's'
I18
sS'm'
(lp29483
I1
aI1
aI3
aI13
aI19
aI53
aI69
aI1
aI435
aI71
aI339
aI2289
aI1591
aI8783
aI8087
aI25855
aI115311
aI191115
asS'd'
I14742
sa(dp29484
S'a'
I22245
sS's'
I18
sS'm'
(lp29485
I1
aI3
aI5
aI11
aI1
aI55
aI59
aI7
aI101
aI655
aI353
aI483
aI5681
aI12721
aI15973
aI51377
aI94921
aI246365
asS'd'
I14743
sa(dp29486
S'a'
I22272
sS's'
I18
sS'm'
(lp29487
I1
aI3
aI5
aI3
aI25
aI23
aI99
aI145
aI277
aI741
aI595
aI2653
aI1393
aI2867
aI271
aI49131
aI111973
aI118869
asS'd'
I14744
sa(dp29488
S'a'
I22308
sS's'
I18
sS'm'
(lp29489
I1
aI1
aI7
aI13
aI11
aI51
aI127
aI27
aI305
aI265
aI1755
aI3189
aI4679
aI9721
aI24409
aI46941
aI94353
aI95643
asS'd'
I14745
sa(dp29490
S'a'
I22318
sS's'
I18
sS'm'
(lp29491
I1
aI1
aI5
aI11
aI1
aI63
aI53
aI149
aI459
aI155
aI1431
aI3969
aI3417
aI12121
aI14535
aI52089
aI110745
aI57
asS'd'
I14746
sa(dp29492
S'a'
I22320
sS's'
I18
sS'm'
(lp29493
I1
aI1
aI5
aI9
aI23
aI33
aI17
aI175
aI313
aI185
aI101
aI531
aI2941
aI14999
aI31413
aI12103
aI33709
aI260555
asS'd'
I14747
sa(dp29494
S'a'
I22338
sS's'
I18
sS'm'
(lp29495
I1
aI1
aI3
aI13
aI3
aI11
aI67
aI95
aI211
aI673
aI23
aI2379
aI6985
aI12101
aI13021
aI9255
aI116437
aI228877
asS'd'
I14748
sa(dp29496
S'a'
I22350
sS's'
I18
sS'm'
(lp29497
I1
aI1
aI3
aI15
aI7
aI51
aI25
aI109
aI45
aI691
aI869
aI485
aI111
aI11465
aI27953
aI54375
aI10805
aI221023
asS'd'
I14749
sa(dp29498
S'a'
I22374
sS's'
I18
sS'm'
(lp29499
I1
aI3
aI7
aI7
aI17
aI53
aI59
aI101
aI221
aI593
aI587
aI873
aI931
aI14617
aI12067
aI58655
aI102437
aI31675
asS'd'
I14750
sa(dp29500
S'a'
I22388
sS's'
I18
sS'm'
(lp29501
I1
aI1
aI3
aI15
aI25
aI57
aI35
aI231
aI491
aI671
aI933
aI3525
aI1237
aI10155
aI27501
aI50781
aI23183
aI108283
asS'd'
I14751
sa(dp29502
S'a'
I22391
sS's'
I18
sS'm'
(lp29503
I1
aI3
aI5
aI5
aI31
aI63
aI117
aI205
aI199
aI841
aI1455
aI3901
aI2127
aI13573
aI20667
aI49489
aI60217
aI197421
asS'd'
I14752
sa(dp29504
S'a'
I22422
sS's'
I18
sS'm'
(lp29505
I1
aI1
aI3
aI7
aI15
aI21
aI73
aI211
aI421
aI873
aI607
aI709
aI9
aI10985
aI28653
aI64579
aI118145
aI3095
asS'd'
I14753
sa(dp29506
S'a'
I22426
sS's'
I18
sS'm'
(lp29507
I1
aI1
aI1
aI13
aI17
aI53
aI27
aI105
aI201
aI399
aI737
aI3235
aI1287
aI13859
aI6049
aI62249
aI88259
aI52991
asS'd'
I14754
sa(dp29508
S'a'
I22441
sS's'
I18
sS'm'
(lp29509
I1
aI3
aI7
aI7
aI25
aI45
aI67
aI147
aI275
aI315
aI1675
aI2289
aI4611
aI6325
aI26617
aI38079
aI125219
aI23569
asS'd'
I14755
sa(dp29510
S'a'
I22481
sS's'
I18
sS'm'
(lp29511
I1
aI1
aI7
aI7
aI9
aI61
aI115
aI251
aI297
aI691
aI1881
aI1815
aI7229
aI10859
aI8257
aI38097
aI87927
aI162845
asS'd'
I14756
sa(dp29512
S'a'
I22488
sS's'
I18
sS'm'
(lp29513
I1
aI3
aI3
aI9
aI9
aI59
aI17
aI207
aI433
aI825
aI93
aI697
aI7263
aI15983
aI14829
aI47471
aI17579
aI151519
asS'd'
I14757
sa(dp29514
S'a'
I22500
sS's'
I18
sS'm'
(lp29515
I1
aI1
aI1
aI11
aI21
aI31
aI7
aI41
aI383
aI731
aI2033
aI3417
aI4187
aI5515
aI10093
aI15875
aI78551
aI2057
asS'd'
I14758
sa(dp29516
S'a'
I22517
sS's'
I18
sS'm'
(lp29517
I1
aI3
aI7
aI15
aI5
aI29
aI7
aI171
aI129
aI727
aI1815
aI1361
aI6137
aI10333
aI22203
aI361
aI92437
aI6545
asS'd'
I14759
sa(dp29518
S'a'
I22531
sS's'
I18
sS'm'
(lp29519
I1
aI1
aI3
aI13
aI25
aI45
aI111
aI69
aI333
aI365
aI765
aI2755
aI3485
aI2729
aI23467
aI64809
aI120755
aI169279
asS'd'
I14760
sa(dp29520
S'a'
I22552
sS's'
I18
sS'm'
(lp29521
I1
aI1
aI3
aI1
aI19
aI13
aI33
aI165
aI157
aI429
aI1175
aI3435
aI7523
aI5055
aI12295
aI34309
aI36933
aI164037
asS'd'
I14761
sa(dp29522
S'a'
I22574
sS's'
I18
sS'm'
(lp29523
I1
aI1
aI3
aI11
aI31
aI49
aI37
aI161
aI465
aI311
aI1839
aI689
aI6837
aI13473
aI29883
aI61587
aI86077
aI156921
asS'd'
I14762
sa(dp29524
S'a'
I22579
sS's'
I18
sS'm'
(lp29525
I1
aI3
aI3
aI1
aI3
aI23
aI69
aI159
aI501
aI303
aI1495
aI9
aI6055
aI545
aI12247
aI23413
aI67247
aI38137
asS'd'
I14763
sa(dp29526
S'a'
I22582
sS's'
I18
sS'm'
(lp29527
I1
aI1
aI5
aI15
aI5
aI39
aI107
aI121
aI295
aI167
aI1055
aI2703
aI147
aI7291
aI3981
aI51989
aI92953
aI225987
asS'd'
I14764
sa(dp29528
S'a'
I22586
sS's'
I18
sS'm'
(lp29529
I1
aI3
aI1
aI1
aI21
aI5
aI91
aI129
aI57
aI53
aI365
aI2497
aI5017
aI13535
aI19305
aI60447
aI115467
aI225317
asS'd'
I14765
sa(dp29530
S'a'
I22594
sS's'
I18
sS'm'
(lp29531
I1
aI3
aI1
aI7
aI25
aI17
aI51
aI15
aI119
aI1013
aI719
aI991
aI2655
aI12587
aI15749
aI11723
aI18461
aI155937
asS'd'
I14766
sa(dp29532
S'a'
I22611
sS's'
I18
sS'm'
(lp29533
I1
aI3
aI7
aI3
aI25
aI33
aI59
aI135
aI501
aI813
aI235
aI3775
aI2781
aI13137
aI32673
aI31643
aI78881
aI207651
asS'd'
I14767
sa(dp29534
S'a'
I22614
sS's'
I18
sS'm'
(lp29535
I1
aI1
aI7
aI13
aI27
aI51
aI99
aI189
aI187
aI577
aI941
aI1275
aI7297
aI14731
aI12599
aI49049
aI96439
aI35093
asS'd'
I14768
sa(dp29536
S'a'
I22700
sS's'
I18
sS'm'
(lp29537
I1
aI3
aI1
aI15
aI9
aI45
aI1
aI149
aI305
aI231
aI935
aI1377
aI6345
aI14795
aI20969
aI26263
aI5711
aI146949
asS'd'
I14769
sa(dp29538
S'a'
I22711
sS's'
I18
sS'm'
(lp29539
I1
aI1
aI5
aI9
aI5
aI9
aI47
aI127
aI105
aI517
aI671
aI67
aI4639
aI2477
aI23109
aI56707
aI72131
aI100709
asS'd'
I14770
sa(dp29540
S'a'
I22749
sS's'
I18
sS'm'
(lp29541
I1
aI3
aI5
aI3
aI21
aI23
aI7
aI193
aI491
aI197
aI319
aI3207
aI2183
aI2133
aI3127
aI34555
aI53707
aI170875
asS'd'
I14771
sa(dp29542
S'a'
I22759
sS's'
I18
sS'm'
(lp29543
I1
aI1
aI3
aI9
aI5
aI23
aI109
aI91
aI359
aI913
aI179
aI1031
aI3617
aI12497
aI23299
aI53293
aI114603
aI9931
asS'd'
I14772
sa(dp29544
S'a'
I22774
sS's'
I18
sS'm'
(lp29545
I1
aI3
aI1
aI5
aI1
aI47
aI73
aI103
aI333
aI483
aI1015
aI3085
aI5229
aI3171
aI16539
aI13493
aI68957
aI177645
asS'd'
I14773
sa(dp29546
S'a'
I22777
sS's'
I18
sS'm'
(lp29547
I1
aI1
aI1
aI9
aI27
aI15
aI25
aI255
aI383
aI501
aI831
aI2463
aI237
aI16065
aI6991
aI56503
aI117303
aI140573
asS'd'
I14774
sa(dp29548
S'a'
I22780
sS's'
I18
sS'm'
(lp29549
I1
aI1
aI3
aI5
aI9
aI25
aI15
aI179
aI415
aI729
aI1163
aI2649
aI2907
aI9591
aI29129
aI42775
aI80537
aI139897
asS'd'
I14775
sa(dp29550
S'a'
I22783
sS's'
I18
sS'm'
(lp29551
I1
aI1
aI3
aI7
aI31
aI15
aI113
aI1
aI263
aI685
aI1953
aI1479
aI5143
aI8585
aI9057
aI61479
aI122065
aI191541
asS'd'
I14776
sa(dp29552
S'a'
I22785
sS's'
I18
sS'm'
(lp29553
I1
aI3
aI1
aI11
aI25
aI47
aI25
aI229
aI463
aI197
aI1123
aI2665
aI2345
aI11701
aI10435
aI15205
aI35437
aI137619
asS'd'
I14777
sa(dp29554
S'a'
I22840
sS's'
I18
sS'm'
(lp29555
I1
aI3
aI3
aI5
aI19
aI57
aI89
aI101
aI373
aI283
aI57
aI1701
aI5025
aI6677
aI20321
aI58459
aI9319
aI161501
asS'd'
I14778
sa(dp29556
S'a'
I22878
sS's'
I18
sS'm'
(lp29557
I1
aI1
aI1
aI13
aI3
aI51
aI111
aI23
aI325
aI813
aI441
aI2371
aI1993
aI6839
aI359
aI9873
aI33719
aI208163
asS'd'
I14779
sa(dp29558
S'a'
I22884
sS's'
I18
sS'm'
(lp29559
I1
aI3
aI1
aI11
aI23
aI53
aI35
aI89
aI91
aI601
aI433
aI1671
aI1919
aI2115
aI6355
aI10639
aI87305
aI194185
asS'd'
I14780
sa(dp29560
S'a'
I22888
sS's'
I18
sS'm'
(lp29561
I1
aI1
aI5
aI9
aI29
aI31
aI43
aI153
aI209
aI835
aI865
aI2431
aI1085
aI9771
aI14483
aI19551
aI98673
aI146881
asS'd'
I14781
sa(dp29562
S'a'
I22927
sS's'
I18
sS'm'
(lp29563
I1
aI1
aI7
aI3
aI7
aI33
aI49
aI111
aI111
aI843
aI479
aI2113
aI4575
aI14911
aI5161
aI7153
aI37525
aI217887
asS'd'
I14782
sa(dp29564
S'a'
I22941
sS's'
I18
sS'm'
(lp29565
I1
aI1
aI7
aI9
aI27
aI5
aI23
aI217
aI11
aI79
aI1637
aI2047
aI6697
aI5601
aI2877
aI63497
aI100127
aI157833
asS'd'
I14783
sa(dp29566
S'a'
I22951
sS's'
I18
sS'm'
(lp29567
I1
aI1
aI7
aI11
aI31
aI41
aI91
aI39
aI207
aI185
aI1163
aI2115
aI2963
aI7605
aI12597
aI54175
aI7221
aI117129
asS'd'
I14784
sa(dp29568
S'a'
I22958
sS's'
I18
sS'm'
(lp29569
I1
aI3
aI7
aI13
aI9
aI15
aI3
aI47
aI281
aI451
aI1111
aI3585
aI4505
aI9465
aI8047
aI45893
aI27179
aI124373
asS'd'
I14785
sa(dp29570
S'a'
I22980
sS's'
I18
sS'm'
(lp29571
I1
aI3
aI5
aI11
aI27
aI29
aI11
aI221
aI483
aI29
aI17
aI1067
aI6761
aI39
aI13419
aI7263
aI127547
aI178951
asS'd'
I14786
sa(dp29572
S'a'
I23007
sS's'
I18
sS'm'
(lp29573
I1
aI3
aI5
aI5
aI19
aI3
aI51
aI155
aI41
aI251
aI851
aI1191
aI4445
aI8337
aI25339
aI32931
aI4743
aI31883
asS'd'
I14787
sa(dp29574
S'a'
I23032
sS's'
I18
sS'm'
(lp29575
I1
aI3
aI7
aI15
aI9
aI3
aI113
aI151
aI239
aI611
aI381
aI1141
aI2865
aI3071
aI7293
aI61997
aI2891
aI14533
asS'd'
I14788
sa(dp29576
S'a'
I23041
sS's'
I18
sS'm'
(lp29577
I1
aI3
aI5
aI3
aI15
aI59
aI3
aI37
aI385
aI587
aI837
aI2483
aI5493
aI10571
aI26129
aI44835
aI63425
aI246953
asS'd'
I14789
sa(dp29578
S'a'
I23044
sS's'
I18
sS'm'
(lp29579
I1
aI3
aI5
aI13
aI9
aI9
aI93
aI11
aI139
aI619
aI581
aI2859
aI5481
aI11941
aI20661
aI37463
aI95369
aI177009
asS'd'
I14790
sa(dp29580
S'a'
I23059
sS's'
I18
sS'm'
(lp29581
I1
aI1
aI7
aI11
aI7
aI17
aI89
aI7
aI479
aI377
aI1631
aI509
aI7429
aI13733
aI24011
aI24191
aI98409
aI180761
asS'd'
I14791
sa(dp29582
S'a'
I23065
sS's'
I18
sS'm'
(lp29583
I1
aI3
aI7
aI1
aI5
aI17
aI51
aI113
aI181
aI75
aI1787
aI2221
aI6181
aI16069
aI3031
aI32531
aI107833
aI239907
asS'd'
I14792
sa(dp29584
S'a'
I23072
sS's'
I18
sS'm'
(lp29585
I1
aI1
aI5
aI11
aI3
aI25
aI13
aI35
aI311
aI865
aI873
aI1811
aI3101
aI4445
aI18155
aI18647
aI55693
aI144963
asS'd'
I14793
sa(dp29586
S'a'
I23137
sS's'
I18
sS'm'
(lp29587
I1
aI3
aI7
aI13
aI1
aI9
aI73
aI189
aI255
aI301
aI1579
aI597
aI6027
aI15621
aI27287
aI14615
aI76051
aI143445
asS'd'
I14794
sa(dp29588
S'a'
I23144
sS's'
I18
sS'm'
(lp29589
I1
aI3
aI1
aI13
aI19
aI59
aI11
aI97
aI501
aI857
aI1071
aI3633
aI8059
aI2469
aI16803
aI49395
aI73631
aI114297
asS'd'
I14795
sa(dp29590
S'a'
I23155
sS's'
I18
sS'm'
(lp29591
I1
aI1
aI1
aI5
aI19
aI3
aI59
aI179
aI343
aI745
aI497
aI2965
aI3841
aI3119
aI17707
aI31577
aI39801
aI108819
asS'd'
I14796
sa(dp29592
S'a'
I23162
sS's'
I18
sS'm'
(lp29593
I1
aI3
aI3
aI9
aI11
aI17
aI19
aI199
aI283
aI229
aI493
aI631
aI8133
aI1531
aI25271
aI11353
aI114759
aI70655
asS'd'
I14797
sa(dp29594
S'a'
I23167
sS's'
I18
sS'm'
(lp29595
I1
aI1
aI7
aI7
aI3
aI11
aI1
aI95
aI167
aI863
aI1009
aI1695
aI2773
aI11667
aI23515
aI12927
aI87883
aI28773
asS'd'
I14798
sa(dp29596
S'a'
I23183
sS's'
I18
sS'm'
(lp29597
I1
aI3
aI1
aI15
aI9
aI1
aI31
aI243
aI57
aI349
aI483
aI659
aI1971
aI7971
aI23797
aI4403
aI83837
aI239261
asS'd'
I14799
sa(dp29598
S'a'
I23192
sS's'
I18
sS'm'
(lp29599
I1
aI3
aI7
aI5
aI11
aI17
aI55
aI5
aI209
aI233
aI1969
aI925
aI695
aI1321
aI11965
aI29849
aI120519
aI195105
asS'd'
I14800
sa(dp29600
S'a'
I23197
sS's'
I18
sS'm'
(lp29601
I1
aI3
aI7
aI11
aI9
aI45
aI27
aI9
aI57
aI649
aI1801
aI2653
aI1535
aI45
aI8901
aI28755
aI26475
aI112341
asS'd'
I14801
sa(dp29602
S'a'
I23216
sS's'
I18
sS'm'
(lp29603
I1
aI1
aI3
aI13
aI11
aI57
aI103
aI213
aI193
aI779
aI541
aI3685
aI4191
aI6105
aI7199
aI63659
aI49673
aI208361
asS'd'
I14802
sa(dp29604
S'a'
I23221
sS's'
I18
sS'm'
(lp29605
I1
aI3
aI3
aI7
aI15
aI15
aI9
aI207
aI387
aI429
aI1213
aI1703
aI5753
aI10261
aI8705
aI62783
aI9643
aI248591
asS'd'
I14803
sa(dp29606
S'a'
I23228
sS's'
I18
sS'm'
(lp29607
I1
aI3
aI3
aI15
aI23
aI17
aI5
aI83
aI295
aI685
aI2003
aI1723
aI2799
aI14699
aI25171
aI20275
aI45597
aI214107
asS'd'
I14804
sa(dp29608
S'a'
I23233
sS's'
I18
sS'm'
(lp29609
I1
aI1
aI1
aI15
aI13
aI33
aI111
aI69
aI329
aI273
aI1303
aI3377
aI4151
aI12547
aI20411
aI54845
aI7839
aI173939
asS'd'
I14805
sa(dp29610
S'a'
I23234
sS's'
I18
sS'm'
(lp29611
I1
aI1
aI5
aI15
aI25
aI31
aI11
aI75
aI69
aI501
aI1485
aI3659
aI3889
aI9715
aI9633
aI45313
aI112377
aI27799
asS'd'
I14806
sa(dp29612
S'a'
I23251
sS's'
I18
sS'm'
(lp29613
I1
aI1
aI3
aI11
aI31
aI27
aI7
aI25
aI315
aI593
aI315
aI275
aI1453
aI9429
aI10023
aI17939
aI37651
aI217435
asS'd'
I14807
sa(dp29614
S'a'
I23257
sS's'
I18
sS'm'
(lp29615
I1
aI3
aI7
aI7
aI27
aI41
aI69
aI95
aI19
aI763
aI1733
aI2097
aI6723
aI7051
aI15209
aI53047
aI56117
aI87639
asS'd'
I14808
sa(dp29616
S'a'
I23258
sS's'
I18
sS'm'
(lp29617
I1
aI3
aI7
aI5
aI15
aI61
aI31
aI19
aI361
aI571
aI727
aI405
aI835
aI4847
aI26777
aI50311
aI104125
aI127197
asS'd'
I14809
sa(dp29618
S'a'
I23288
sS's'
I18
sS'm'
(lp29619
I1
aI1
aI1
aI11
aI11
aI61
aI59
aI63
aI409
aI219
aI1135
aI3385
aI5583
aI16143
aI22709
aI31247
aI19871
aI68557
asS'd'
I14810
sa(dp29620
S'a'
I23302
sS's'
I18
sS'm'
(lp29621
I1
aI3
aI7
aI1
aI11
aI3
aI121
aI41
aI135
aI427
aI1267
aI2169
aI507
aI757
aI12411
aI50655
aI75625
aI1199
asS'd'
I14811
sa(dp29622
S'a'
I23341
sS's'
I18
sS'm'
(lp29623
I1
aI1
aI7
aI5
aI17
aI21
aI89
aI119
aI55
aI395
aI979
aI909
aI1711
aI3289
aI8433
aI9
aI12743
aI109027
asS'd'
I14812
sa(dp29624
S'a'
I23373
sS's'
I18
sS'm'
(lp29625
I1
aI1
aI1
aI13
aI5
aI11
aI93
aI35
aI437
aI173
aI1157
aI2749
aI6855
aI8307
aI26145
aI22593
aI125415
aI65509
asS'd'
I14813
sa(dp29626
S'a'
I23402
sS's'
I18
sS'm'
(lp29627
I1
aI1
aI5
aI3
aI25
aI27
aI1
aI173
aI113
aI373
aI1769
aI2941
aI1895
aI3399
aI27665
aI50613
aI20747
aI31903
asS'd'
I14814
sa(dp29628
S'a'
I23409
sS's'
I18
sS'm'
(lp29629
I1
aI1
aI1
aI1
aI9
aI7
aI53
aI73
aI465
aI725
aI1537
aI579
aI83
aI925
aI15507
aI13595
aI16927
aI205087
asS'd'
I14815
sa(dp29630
S'a'
I23428
sS's'
I18
sS'm'
(lp29631
I1
aI1
aI3
aI7
aI7
aI23
aI31
aI127
aI27
aI727
aI1305
aI3879
aI817
aI15995
aI28607
aI22695
aI6367
aI161587
asS'd'
I14816
sa(dp29632
S'a'
I23435
sS's'
I18
sS'm'
(lp29633
I1
aI3
aI1
aI7
aI29
aI23
aI27
aI117
aI279
aI917
aI1105
aI2061
aI7719
aI13633
aI16501
aI33739
aI71939
aI143115
asS'd'
I14817
sa(dp29634
S'a'
I23440
sS's'
I18
sS'm'
(lp29635
I1
aI3
aI7
aI11
aI7
aI27
aI65
aI133
aI411
aI441
aI925
aI1485
aI2035
aI3067
aI14511
aI58511
aI120773
aI228731
asS'd'
I14818
sa(dp29636
S'a'
I23449
sS's'
I18
sS'm'
(lp29637
I1
aI1
aI3
aI9
aI21
aI55
aI27
aI73
aI175
aI395
aI1201
aI2599
aI3839
aI11163
aI5057
aI3385
aI43265
aI105211
asS'd'
I14819
sa(dp29638
S'a'
I23459
sS's'
I18
sS'm'
(lp29639
I1
aI1
aI1
aI3
aI7
aI63
aI91
aI197
aI417
aI763
aI1391
aI3729
aI2791
aI1975
aI23655
aI50611
aI110315
aI86879
asS'd'
I14820
sa(dp29640
S'a'
I23473
sS's'
I18
sS'm'
(lp29641
I1
aI3
aI3
aI5
aI31
aI35
aI67
aI67
aI89
aI933
aI1005
aI1837
aI5947
aI2559
aI27731
aI25151
aI102959
aI81557
asS'd'
I14821
sa(dp29642
S'a'
I23500
sS's'
I18
sS'm'
(lp29643
I1
aI3
aI3
aI1
aI3
aI39
aI57
aI199
aI87
aI91
aI1641
aI3407
aI2823
aI10441
aI26357
aI56677
aI17647
aI86831
asS'd'
I14822
sa(dp29644
S'a'
I23511
sS's'
I18
sS'm'
(lp29645
I1
aI3
aI5
aI7
aI15
aI5
aI49
aI227
aI395
aI837
aI1707
aI1677
aI1907
aI13101
aI1929
aI61701
aI1479
aI80671
asS'd'
I14823
sa(dp29646
S'a'
I23548
sS's'
I18
sS'm'
(lp29647
I1
aI1
aI1
aI11
aI17
aI57
aI37
aI151
aI61
aI709
aI2027
aI2239
aI3283
aI5467
aI17221
aI40759
aI91637
aI258167
asS'd'
I14824
sa(dp29648
S'a'
I23560
sS's'
I18
sS'm'
(lp29649
I1
aI3
aI5
aI11
aI27
aI29
aI121
aI181
aI503
aI705
aI225
aI1111
aI7183
aI3219
aI3233
aI2085
aI113619
aI32959
asS'd'
I14825
sa(dp29650
S'a'
I23563
sS's'
I18
sS'm'
(lp29651
I1
aI1
aI7
aI5
aI29
aI31
aI93
aI113
aI457
aI161
aI337
aI2003
aI1865
aI13357
aI19961
aI51485
aI62751
aI111285
asS'd'
I14826
sa(dp29652
S'a'
I23594
sS's'
I18
sS'm'
(lp29653
I1
aI3
aI1
aI1
aI23
aI25
aI65
aI99
aI11
aI835
aI661
aI3291
aI2655
aI1135
aI19957
aI5029
aI110483
aI2499
asS'd'
I14827
sa(dp29654
S'a'
I23601
sS's'
I18
sS'm'
(lp29655
I1
aI1
aI1
aI1
aI25
aI21
aI59
aI203
aI471
aI697
aI455
aI1561
aI3215
aI609
aI5097
aI8715
aI115705
aI21441
asS'd'
I14828
sa(dp29656
S'a'
I23625
sS's'
I18
sS'm'
(lp29657
I1
aI1
aI5
aI13
aI27
aI37
aI15
aI175
aI191
aI975
aI977
aI401
aI7053
aI14291
aI14621
aI48989
aI113033
aI172569
asS'd'
I14829
sa(dp29658
S'a'
I23640
sS's'
I18
sS'm'
(lp29659
I1
aI3
aI1
aI1
aI19
aI11
aI125
aI53
aI307
aI421
aI93
aI2487
aI5907
aI2195
aI30569
aI21009
aI20759
aI246937
asS'd'
I14830
sa(dp29660
S'a'
I23650
sS's'
I18
sS'm'
(lp29661
I1
aI3
aI7
aI3
aI23
aI21
aI103
aI115
aI453
aI537
aI473
aI1069
aI3007
aI15111
aI3477
aI5635
aI46423
aI68633
asS'd'
I14831
sa(dp29662
S'a'
I23662
sS's'
I18
sS'm'
(lp29663
I1
aI3
aI3
aI1
aI21
aI1
aI49
aI197
aI173
aI775
aI1877
aI1309
aI729
aI3555
aI5981
aI32539
aI22765
aI171077
asS'd'
I14832
sa(dp29664
S'a'
I23692
sS's'
I18
sS'm'
(lp29665
I1
aI1
aI3
aI13
aI19
aI5
aI75
aI149
aI441
aI665
aI1567
aI2433
aI8173
aI12639
aI27479
aI47221
aI66203
aI89017
asS'd'
I14833
sa(dp29666
S'a'
I23726
sS's'
I18
sS'm'
(lp29667
I1
aI1
aI1
aI11
aI1
aI55
aI99
aI119
aI491
aI621
aI619
aI2521
aI905
aI11601
aI26481
aI2023
aI127413
aI220387
asS'd'
I14834
sa(dp29668
S'a'
I23738
sS's'
I18
sS'm'
(lp29669
I1
aI1
aI7
aI11
aI9
aI21
aI57
aI93
aI243
aI229
aI1445
aI997
aI1317
aI2327
aI14141
aI45787
aI82295
aI72823
asS'd'
I14835
sa(dp29670
S'a'
I23743
sS's'
I18
sS'm'
(lp29671
I1
aI3
aI7
aI3
aI11
aI7
aI115
aI143
aI349
aI507
aI1047
aI2573
aI2491
aI13351
aI19019
aI4857
aI62781
aI261261
asS'd'
I14836
sa(dp29672
S'a'
I23755
sS's'
I18
sS'm'
(lp29673
I1
aI1
aI3
aI1
aI1
aI13
aI45
aI227
aI41
aI947
aI693
aI2853
aI7459
aI1485
aI22087
aI61195
aI111771
aI136389
asS'd'
I14837
sa(dp29674
S'a'
I23760
sS's'
I18
sS'm'
(lp29675
I1
aI1
aI3
aI11
aI13
aI53
aI49
aI15
aI425
aI29
aI681
aI1493
aI1385
aI9555
aI13291
aI36735
aI12351
aI29293
asS'd'
I14838
sa(dp29676
S'a'
I23799
sS's'
I18
sS'm'
(lp29677
I1
aI1
aI3
aI1
aI5
aI19
aI37
aI45
aI69
aI209
aI365
aI3949
aI6163
aI5207
aI9297
aI21147
aI71437
aI40487
asS'd'
I14839
sa(dp29678
S'a'
I23848
sS's'
I18
sS'm'
(lp29679
I1
aI3
aI3
aI13
aI31
aI21
aI9
aI177
aI95
aI285
aI1953
aI1969
aI7367
aI7401
aI12017
aI9939
aI11895
aI213133
asS'd'
I14840
sa(dp29680
S'a'
I23859
sS's'
I18
sS'm'
(lp29681
I1
aI1
aI7
aI1
aI1
aI63
aI103
aI141
aI39
aI679
aI123
aI2941
aI4335
aI199
aI12237
aI6599
aI48641
aI140063
asS'd'
I14841
sa(dp29682
S'a'
I23876
sS's'
I18
sS'm'
(lp29683
I1
aI3
aI7
aI3
aI31
aI17
aI21
aI77
aI65
aI979
aI109
aI3325
aI1781
aI6983
aI31477
aI23149
aI33943
aI96137
asS'd'
I14842
sa(dp29684
S'a'
I23897
sS's'
I18
sS'm'
(lp29685
I1
aI3
aI3
aI5
aI21
aI5
aI125
aI117
aI427
aI381
aI511
aI2643
aI409
aI4945
aI3167
aI45879
aI1469
aI56077
asS'd'
I14843
sa(dp29686
S'a'
I23907
sS's'
I18
sS'm'
(lp29687
I1
aI3
aI1
aI5
aI27
aI43
aI83
aI31
aI65
aI645
aI1205
aI1387
aI723
aI15359
aI13517
aI23601
aI61717
aI47079
asS'd'
I14844
sa(dp29688
S'a'
I23919
sS's'
I18
sS'm'
(lp29689
I1
aI3
aI3
aI13
aI15
aI37
aI101
aI175
aI225
aI513
aI483
aI1291
aI669
aI5335
aI16023
aI287
aI51819
aI239803
asS'd'
I14845
sa(dp29690
S'a'
I23921
sS's'
I18
sS'm'
(lp29691
I1
aI3
aI3
aI3
aI3
aI1
aI75
aI175
aI185
aI949
aI673
aI2239
aI4355
aI10687
aI27093
aI37409
aI23193
aI211819
asS'd'
I14846
sa(dp29692
S'a'
I23931
sS's'
I18
sS'm'
(lp29693
I1
aI1
aI3
aI13
aI21
aI3
aI41
aI55
aI243
aI501
aI285
aI7
aI6291
aI7725
aI17051
aI45753
aI115117
aI14323
asS'd'
I14847
sa(dp29694
S'a'
I23933
sS's'
I18
sS'm'
(lp29695
I1
aI1
aI1
aI5
aI13
aI11
aI51
aI175
aI435
aI673
aI67
aI1525
aI323
aI5739
aI19977
aI62317
aI97511
aI130883
asS'd'
I14848
sa(dp29696
S'a'
I23943
sS's'
I18
sS'm'
(lp29697
I1
aI3
aI5
aI11
aI7
aI11
aI97
aI59
aI295
aI409
aI453
aI2439
aI5217
aI10315
aI469
aI31187
aI17325
aI158079
asS'd'
I14849
sa(dp29698
S'a'
I23957
sS's'
I18
sS'm'
(lp29699
I1
aI1
aI3
aI5
aI31
aI9
aI15
aI63
aI411
aI427
aI277
aI2687
aI5021
aI1507
aI22453
aI35559
aI122081
aI121669
asS'd'
I14850
sa(dp29700
S'a'
I23986
sS's'
I18
sS'm'
(lp29701
I1
aI3
aI5
aI13
aI3
aI21
aI69
aI51
aI27
aI571
aI1981
aI2729
aI5733
aI1225
aI26821
aI43763
aI57355
aI169279
asS'd'
I14851
sa(dp29702
S'a'
I24020
sS's'
I18
sS'm'
(lp29703
I1
aI1
aI1
aI13
aI31
aI37
aI33
aI19
aI313
aI341
aI1141
aI1689
aI4511
aI789
aI15317
aI61263
aI79371
aI65157
asS'd'
I14852
sa(dp29704
S'a'
I24043
sS's'
I18
sS'm'
(lp29705
I1
aI3
aI3
aI15
aI27
aI41
aI107
aI23
aI499
aI339
aI273
aI1937
aI2743
aI10879
aI27127
aI64817
aI1217
aI45863
asS'd'
I14853
sa(dp29706
S'a'
I24064
sS's'
I18
sS'm'
(lp29707
I1
aI1
aI5
aI9
aI19
aI43
aI125
aI223
aI473
aI489
aI1999
aI1513
aI6479
aI9511
aI12503
aI29419
aI22559
aI209499
asS'd'
I14854
sa(dp29708
S'a'
I24082
sS's'
I18
sS'm'
(lp29709
I1
aI3
aI1
aI13
aI25
aI55
aI53
aI61
aI303
aI337
aI1325
aI2525
aI6503
aI1155
aI6841
aI58167
aI8175
aI183949
asS'd'
I14855
sa(dp29710
S'a'
I24084
sS's'
I18
sS'm'
(lp29711
I1
aI3
aI3
aI11
aI3
aI15
aI55
aI105
aI497
aI527
aI1007
aI3545
aI4187
aI8723
aI12761
aI20751
aI101583
aI225373
asS'd'
I14856
sa(dp29712
S'a'
I24088
sS's'
I18
sS'm'
(lp29713
I1
aI3
aI3
aI9
aI19
aI59
aI57
aI215
aI313
aI871
aI407
aI2475
aI879
aI15147
aI31945
aI23939
aI104073
aI217619
asS'd'
I14857
sa(dp29714
S'a'
I24117
sS's'
I18
sS'm'
(lp29715
I1
aI1
aI3
aI1
aI27
aI23
aI3
aI43
aI471
aI757
aI1525
aI3003
aI2779
aI6731
aI12423
aI59621
aI72935
aI192283
asS'd'
I14858
sa(dp29716
S'a'
I24149
sS's'
I18
sS'm'
(lp29717
I1
aI3
aI1
aI7
aI13
aI23
aI13
aI91
aI95
aI745
aI639
aI2627
aI4595
aI11735
aI4143
aI23573
aI98647
aI171201
asS'd'
I14859
sa(dp29718
S'a'
I24165
sS's'
I18
sS'm'
(lp29719
I1
aI1
aI1
aI15
aI3
aI61
aI33
aI181
aI351
aI777
aI1365
aI1691
aI2465
aI5289
aI24567
aI8059
aI95301
aI75855
asS'd'
I14860
sa(dp29720
S'a'
I24223
sS's'
I18
sS'm'
(lp29721
I1
aI3
aI5
aI13
aI1
aI57
aI57
aI187
aI1
aI601
aI563
aI1703
aI1307
aI14673
aI7793
aI44589
aI7629
aI254071
asS'd'
I14861
sa(dp29722
S'a'
I24247
sS's'
I18
sS'm'
(lp29723
I1
aI1
aI5
aI13
aI29
aI17
aI61
aI233
aI371
aI909
aI529
aI185
aI127
aI15773
aI19529
aI49271
aI26749
aI70869
asS'd'
I14862
sa(dp29724
S'a'
I24259
sS's'
I18
sS'm'
(lp29725
I1
aI1
aI3
aI9
aI21
aI41
aI37
aI71
aI505
aI969
aI301
aI1667
aI5879
aI13187
aI2461
aI17301
aI103673
aI235133
asS'd'
I14863
sa(dp29726
S'a'
I24283
sS's'
I18
sS'm'
(lp29727
I1
aI3
aI1
aI3
aI9
aI13
aI75
aI63
aI313
aI273
aI1061
aI3821
aI539
aI9887
aI19775
aI17259
aI93133
aI217245
asS'd'
I14864
sa(dp29728
S'a'
I24296
sS's'
I18
sS'm'
(lp29729
I1
aI3
aI5
aI5
aI21
aI27
aI9
aI11
aI461
aI575
aI507
aI577
aI4559
aI9995
aI13953
aI61023
aI121941
aI195419
asS'd'
I14865
sa(dp29730
S'a'
I24336
sS's'
I18
sS'm'
(lp29731
I1
aI3
aI3
aI7
aI17
aI17
aI45
aI193
aI271
aI571
aI1337
aI2107
aI1923
aI4791
aI23773
aI60923
aI58085
aI81219
asS'd'
I14866
sa(dp29732
S'a'
I24342
sS's'
I18
sS'm'
(lp29733
I1
aI3
aI1
aI7
aI11
aI7
aI85
aI33
aI231
aI307
aI993
aI1509
aI1427
aI9545
aI7919
aI39775
aI81145
aI79139
asS'd'
I14867
sa(dp29734
S'a'
I24389
sS's'
I18
sS'm'
(lp29735
I1
aI3
aI5
aI3
aI9
aI57
aI117
aI187
aI57
aI719
aI1635
aI2499
aI6747
aI6649
aI22643
aI16429
aI83233
aI122057
asS'd'
I14868
sa(dp29736
S'a'
I24394
sS's'
I18
sS'm'
(lp29737
I1
aI1
aI3
aI11
aI7
aI39
aI103
aI221
aI167
aI181
aI1355
aI989
aI3399
aI9471
aI10493
aI57267
aI106551
aI158599
asS'd'
I14869
sa(dp29738
S'a'
I24414
sS's'
I18
sS'm'
(lp29739
I1
aI3
aI1
aI15
aI23
aI19
aI29
aI11
aI355
aI923
aI1401
aI509
aI3647
aI5663
aI2353
aI53217
aI70687
aI145613
asS'd'
I14870
sa(dp29740
S'a'
I24424
sS's'
I18
sS'm'
(lp29741
I1
aI3
aI3
aI11
aI11
aI5
aI21
aI107
aI177
aI429
aI119
aI1029
aI5931
aI7543
aI15455
aI62797
aI118095
aI35387
asS'd'
I14871
sa(dp29742
S'a'
I24441
sS's'
I18
sS'm'
(lp29743
I1
aI3
aI5
aI15
aI19
aI53
aI17
aI215
aI279
aI497
aI1157
aI2235
aI5541
aI5899
aI20711
aI20843
aI113821
aI164231
asS'd'
I14872
sa(dp29744
S'a'
I24491
sS's'
I18
sS'm'
(lp29745
I1
aI1
aI1
aI15
aI21
aI33
aI67
aI247
aI55
aI573
aI1863
aI2703
aI5267
aI4071
aI18235
aI44659
aI102379
aI171529
asS'd'
I14873
sa(dp29746
S'a'
I24502
sS's'
I18
sS'm'
(lp29747
I1
aI3
aI1
aI15
aI5
aI59
aI69
aI189
aI313
aI243
aI339
aI3097
aI4999
aI5909
aI1903
aI56143
aI76209
aI83073
asS'd'
I14874
sa(dp29748
S'a'
I24534
sS's'
I18
sS'm'
(lp29749
I1
aI3
aI5
aI15
aI11
aI41
aI65
aI207
aI95
aI115
aI1203
aI3731
aI6845
aI11173
aI8281
aI40623
aI97119
aI218455
asS'd'
I14875
sa(dp29750
S'a'
I24540
sS's'
I18
sS'm'
(lp29751
I1
aI3
aI7
aI13
aI29
aI57
aI5
aI31
aI255
aI539
aI107
aI953
aI3707
aI9233
aI20295
aI17459
aI2005
aI56193
asS'd'
I14876
sa(dp29752
S'a'
I24543
sS's'
I18
sS'm'
(lp29753
I1
aI1
aI1
aI15
aI5
aI31
aI83
aI165
aI211
aI433
aI1411
aI2949
aI4817
aI1645
aI1693
aI9877
aI118493
aI142923
asS'd'
I14877
sa(dp29754
S'a'
I24549
sS's'
I18
sS'm'
(lp29755
I1
aI3
aI1
aI11
aI19
aI61
aI35
aI21
aI159
aI159
aI1717
aI3227
aI3351
aI8641
aI20575
aI13721
aI114649
aI129201
asS'd'
I14878
sa(dp29756
S'a'
I24567
sS's'
I18
sS'm'
(lp29757
I1
aI3
aI1
aI13
aI9
aI41
aI17
aI7
aI209
aI501
aI445
aI23
aI7911
aI5867
aI30129
aI643
aI36363
aI52037
asS'd'
I14879
sa(dp29758
S'a'
I24583
sS's'
I18
sS'm'
(lp29759
I1
aI3
aI5
aI11
aI31
aI55
aI27
aI81
aI413
aI167
aI599
aI2231
aI7055
aI4013
aI26729
aI63927
aI12075
aI208123
asS'd'
I14880
sa(dp29760
S'a'
I24693
sS's'
I18
sS'm'
(lp29761
I1
aI1
aI7
aI9
aI11
aI39
aI99
aI187
aI169
aI999
aI609
aI3647
aI2497
aI8969
aI30919
aI29145
aI67699
aI51601
asS'd'
I14881
sa(dp29762
S'a'
I24704
sS's'
I18
sS'm'
(lp29763
I1
aI3
aI1
aI1
aI11
aI11
aI69
aI29
aI197
aI979
aI1135
aI869
aI5435
aI5151
aI26349
aI55911
aI68051
aI131849
asS'd'
I14882
sa(dp29764
S'a'
I24719
sS's'
I18
sS'm'
(lp29765
I1
aI1
aI1
aI5
aI27
aI1
aI85
aI145
aI439
aI585
aI1713
aI677
aI1833
aI14139
aI5547
aI31265
aI82223
aI47605
asS'd'
I14883
sa(dp29766
S'a'
I24749
sS's'
I18
sS'm'
(lp29767
I1
aI3
aI5
aI9
aI17
aI23
aI31
aI199
aI447
aI551
aI683
aI2977
aI7839
aI8681
aI15923
aI61057
aI89875
aI52945
asS'd'
I14884
sa(dp29768
S'a'
I24799
sS's'
I18
sS'm'
(lp29769
I1
aI1
aI3
aI11
aI9
aI17
aI29
aI125
aI195
aI123
aI1259
aI2729
aI3099
aI2229
aI9683
aI13121
aI105399
aI111833
asS'd'
I14885
sa(dp29770
S'a'
I24830
sS's'
I18
sS'm'
(lp29771
I1
aI3
aI1
aI3
aI1
aI47
aI93
aI117
aI461
aI633
aI1641
aI933
aI7927
aI13569
aI483
aI28159
aI121561
aI164325
asS'd'
I14886
sa(dp29772
S'a'
I24835
sS's'
I18
sS'm'
(lp29773
I1
aI3
aI1
aI5
aI21
aI19
aI79
aI183
aI395
aI23
aI767
aI519
aI4857
aI10385
aI12425
aI26207
aI114623
aI37125
asS'd'
I14887
sa(dp29774
S'a'
I24907
sS's'
I18
sS'm'
(lp29775
I1
aI1
aI5
aI5
aI9
aI47
aI67
aI217
aI499
aI843
aI1539
aI301
aI1485
aI3157
aI22375
aI47199
aI26215
aI182785
asS'd'
I14888
sa(dp29776
S'a'
I24917
sS's'
I18
sS'm'
(lp29777
I1
aI1
aI1
aI13
aI9
aI37
aI87
aI49
aI445
aI681
aI1097
aI1049
aI4093
aI13167
aI18447
aI58243
aI41797
aI217929
asS'd'
I14889
sa(dp29778
S'a'
I24940
sS's'
I18
sS'm'
(lp29779
I1
aI1
aI5
aI13
aI13
aI49
aI21
aI149
aI79
aI113
aI1217
aI921
aI6321
aI9345
aI27987
aI21723
aI49249
aI18813
asS'd'
I14890
sa(dp29780
S'a'
I24943
sS's'
I18
sS'm'
(lp29781
I1
aI3
aI1
aI3
aI15
aI27
aI67
aI69
aI131
aI713
aI1741
aI1955
aI5665
aI8749
aI11971
aI11257
aI13999
aI124535
asS'd'
I14891
sa(dp29782
S'a'
I24958
sS's'
I18
sS'm'
(lp29783
I1
aI3
aI1
aI3
aI3
aI11
aI21
aI167
aI441
aI557
aI593
aI3261
aI3099
aI2801
aI21725
aI23247
aI106891
aI129187
asS'd'
I14892
sa(dp29784
S'a'
I24962
sS's'
I18
sS'm'
(lp29785
I1
aI1
aI1
aI11
aI5
aI55
aI33
aI71
aI505
aI85
aI1609
aI521
aI5459
aI12777
aI13007
aI255
aI67537
aI2877
asS'd'
I14893
sa(dp29786
S'a'
I24964
sS's'
I18
sS'm'
(lp29787
I1
aI1
aI1
aI3
aI31
aI47
aI49
aI119
aI351
aI797
aI1407
aI4089
aI2381
aI12409
aI12849
aI23489
aI53631
aI119387
asS'd'
I14894
sa(dp29788
S'a'
I24968
sS's'
I18
sS'm'
(lp29789
I1
aI3
aI5
aI11
aI25
aI11
aI25
aI185
aI85
aI849
aI141
aI385
aI3663
aI13133
aI8451
aI61463
aI35129
aI149933
asS'd'
I14895
sa(dp29790
S'a'
I25015
sS's'
I18
sS'm'
(lp29791
I1
aI3
aI7
aI9
aI23
aI17
aI21
aI197
aI15
aI893
aI939
aI707
aI5491
aI7249
aI14009
aI18973
aI111545
aI36809
asS'd'
I14896
sa(dp29792
S'a'
I25024
sS's'
I18
sS'm'
(lp29793
I1
aI3
aI3
aI13
aI23
aI15
aI19
aI193
aI223
aI627
aI1529
aI1963
aI1003
aI7199
aI15361
aI25233
aI110281
aI221761
asS'd'
I14897
sa(dp29794
S'a'
I25027
sS's'
I18
sS'm'
(lp29795
I1
aI3
aI3
aI1
aI17
aI51
aI61
aI215
aI311
aI919
aI349
aI59
aI2897
aI12137
aI5931
aI37611
aI124387
aI83503
asS'd'
I14898
sa(dp29796
S'a'
I25033
sS's'
I18
sS'm'
(lp29797
I1
aI3
aI7
aI9
aI13
aI47
aI53
aI139
aI481
aI733
aI389
aI1209
aI3281
aI593
aI29103
aI61521
aI41445
aI11015
asS'd'
I14899
sa(dp29798
S'a'
I25041
sS's'
I18
sS'm'
(lp29799
I1
aI1
aI1
aI13
aI3
aI31
aI19
aI47
aI151
aI883
aI1707
aI827
aI2129
aI4333
aI871
aI42967
aI79701
aI192211
asS'd'
I14900
sa(dp29800
S'a'
I25047
sS's'
I18
sS'm'
(lp29801
I1
aI3
aI3
aI11
aI11
aI51
aI121
aI241
aI199
aI881
aI1493
aI2381
aI5161
aI13287
aI8155
aI52481
aI120307
aI206203
asS'd'
I14901
sa(dp29802
S'a'
I25057
sS's'
I18
sS'm'
(lp29803
I1
aI1
aI1
aI13
aI15
aI37
aI27
aI151
aI17
aI851
aI1343
aI1447
aI43
aI10267
aI18267
aI21347
aI129277
aI83987
asS'd'
I14902
sa(dp29804
S'a'
I25082
sS's'
I18
sS'm'
(lp29805
I1
aI1
aI3
aI3
aI13
aI17
aI53
aI217
aI253
aI853
aI1461
aI1953
aI617
aI4209
aI9925
aI377
aI42789
aI150415
asS'd'
I14903
sa(dp29806
S'a'
I25093
sS's'
I18
sS'm'
(lp29807
I1
aI3
aI7
aI11
aI13
aI23
aI83
aI235
aI39
aI701
aI1091
aI25
aI1807
aI15431
aI2169
aI5339
aI123679
aI117053
asS'd'
I14904
sa(dp29808
S'a'
I25105
sS's'
I18
sS'm'
(lp29809
I1
aI3
aI7
aI13
aI3
aI29
aI43
aI149
aI33
aI873
aI1177
aI1961
aI7943
aI11317
aI30725
aI55765
aI50929
aI12335
asS'd'
I14905
sa(dp29810
S'a'
I25112
sS's'
I18
sS'm'
(lp29811
I1
aI1
aI3
aI5
aI25
aI1
aI91
aI121
aI295
aI25
aI1743
aI2125
aI2643
aI11175
aI15089
aI44979
aI28355
aI543
asS'd'
I14906
sa(dp29812
S'a'
I25117
sS's'
I18
sS'm'
(lp29813
I1
aI1
aI1
aI13
aI27
aI27
aI43
aI195
aI377
aI821
aI437
aI3445
aI2673
aI15221
aI15101
aI25143
aI22347
aI218549
asS'd'
I14907
sa(dp29814
S'a'
I25131
sS's'
I18
sS'm'
(lp29815
I1
aI1
aI3
aI7
aI9
aI51
aI121
aI231
aI91
aI913
aI1325
aI167
aI8067
aI8119
aI9307
aI33551
aI58069
aI170567
asS'd'
I14908
sa(dp29816
S'a'
I25139
sS's'
I18
sS'm'
(lp29817
I1
aI3
aI7
aI7
aI15
aI7
aI85
aI51
aI11
aI353
aI1117
aI2479
aI3091
aI2377
aI23589
aI38537
aI113047
aI261285
asS'd'
I14909
sa(dp29818
S'a'
I25194
sS's'
I18
sS'm'
(lp29819
I1
aI1
aI3
aI5
aI15
aI33
aI61
aI145
aI147
aI815
aI767
aI9
aI2059
aI11463
aI1883
aI8565
aI101043
aI117565
asS'd'
I14910
sa(dp29820
S'a'
I25204
sS's'
I18
sS'm'
(lp29821
I1
aI3
aI5
aI3
aI5
aI49
aI5
aI33
aI15
aI13
aI895
aI3973
aI7963
aI3831
aI26817
aI10799
aI111409
aI90679
asS'd'
I14911
sa(dp29822
S'a'
I25244
sS's'
I18
sS'm'
(lp29823
I1
aI1
aI5
aI7
aI5
aI51
aI35
aI237
aI217
aI531
aI719
aI2711
aI1937
aI16071
aI23233
aI22799
aI66023
aI145739
asS'd'
I14912
sa(dp29824
S'a'
I25247
sS's'
I18
sS'm'
(lp29825
I1
aI3
aI5
aI1
aI5
aI63
aI1
aI163
aI9
aI697
aI1379
aI2989
aI7113
aI9821
aI15941
aI6495
aI7825
aI29715
asS'd'
I14913
sa(dp29826
S'a'
I25253
sS's'
I18
sS'm'
(lp29827
I1
aI3
aI5
aI7
aI9
aI41
aI113
aI173
aI151
aI963
aI2019
aI3531
aI1133
aI4287
aI16917
aI16929
aI12345
aI31201
asS'd'
I14914
sa(dp29828
S'a'
I25258
sS's'
I18
sS'm'
(lp29829
I1
aI3
aI7
aI1
aI25
aI9
aI5
aI195
aI175
aI297
aI717
aI3725
aI33
aI5155
aI4405
aI56171
aI105597
aI132407
asS'd'
I14915
sa(dp29830
S'a'
I25275
sS's'
I18
sS'm'
(lp29831
I1
aI1
aI7
aI7
aI3
aI59
aI115
aI95
aI227
aI951
aI843
aI619
aI7791
aI10981
aI11773
aI57651
aI108391
aI179561
asS'd'
I14916
sa(dp29832
S'a'
I25320
sS's'
I18
sS'm'
(lp29833
I1
aI3
aI3
aI1
aI9
aI3
aI59
aI161
aI417
aI413
aI1933
aI1027
aI4575
aI10427
aI15643
aI16049
aI120089
aI176607
asS'd'
I14917
sa(dp29834
S'a'
I25338
sS's'
I18
sS'm'
(lp29835
I1
aI3
aI3
aI3
aI15
aI1
aI83
aI195
aI59
aI859
aI1669
aI1063
aI2069
aI15875
aI16459
aI53741
aI114521
aI37641
asS'd'
I14918
sa(dp29836
S'a'
I25343
sS's'
I18
sS'm'
(lp29837
I1
aI1
aI1
aI11
aI11
aI45
aI47
aI143
aI11
aI239
aI1329
aI865
aI2693
aI899
aI26265
aI43255
aI125679
aI130099
asS'd'
I14919
sa(dp29838
S'a'
I25351
sS's'
I18
sS'm'
(lp29839
I1
aI3
aI5
aI3
aI31
aI51
aI95
aI127
aI79
aI167
aI117
aI3177
aI5875
aI14039
aI20341
aI47815
aI118799
aI211871
asS'd'
I14920
sa(dp29840
S'a'
I25358
sS's'
I18
sS'm'
(lp29841
I1
aI1
aI1
aI1
aI3
aI21
aI65
aI203
aI11
aI565
aI537
aI1307
aI8189
aI11423
aI7745
aI56117
aI110959
aI95361
asS'd'
I14921
sa(dp29842
S'a'
I25372
sS's'
I18
sS'm'
(lp29843
I1
aI1
aI7
aI7
aI21
aI63
aI63
aI231
aI441
aI127
aI1943
aI13
aI4813
aI10607
aI23867
aI43891
aI15801
aI173245
asS'd'
I14922
sa(dp29844
S'a'
I25399
sS's'
I18
sS'm'
(lp29845
I1
aI3
aI3
aI5
aI25
aI23
aI123
aI133
aI129
aI303
aI1993
aI1453
aI1109
aI4649
aI30315
aI62399
aI121575
aI60069
asS'd'
I14923
sa(dp29846
S'a'
I25408
sS's'
I18
sS'm'
(lp29847
I1
aI3
aI5
aI3
aI29
aI23
aI69
aI141
aI137
aI1017
aI1915
aI35
aI3817
aI6249
aI22427
aI7281
aI88473
aI230167
asS'd'
I14924
sa(dp29848
S'a'
I25432
sS's'
I18
sS'm'
(lp29849
I1
aI3
aI7
aI7
aI19
aI37
aI93
aI217
aI287
aI731
aI583
aI3377
aI2879
aI4873
aI5549
aI52949
aI127285
aI211173
asS'd'
I14925
sa(dp29850
S'a'
I25442
sS's'
I18
sS'm'
(lp29851
I1
aI1
aI5
aI7
aI23
aI41
aI49
aI145
aI277
aI571
aI1225
aI455
aI2133
aI1229
aI25421
aI20179
aI70919
aI242825
asS'd'
I14926
sa(dp29852
S'a'
I25462
sS's'
I18
sS'm'
(lp29853
I1
aI3
aI5
aI5
aI29
aI3
aI1
aI89
aI413
aI901
aI1343
aI3963
aI6969
aI14649
aI18331
aI4573
aI82077
aI100693
asS'd'
I14927
sa(dp29854
S'a'
I25465
sS's'
I18
sS'm'
(lp29855
I1
aI1
aI1
aI13
aI31
aI53
aI107
aI95
aI151
aI539
aI1593
aI3763
aI1007
aI8959
aI25235
aI16461
aI121819
aI106143
asS'd'
I14928
sa(dp29856
S'a'
I25468
sS's'
I18
sS'm'
(lp29857
I1
aI1
aI1
aI3
aI11
aI15
aI5
aI157
aI347
aI81
aI2013
aI2025
aI6541
aI12287
aI1315
aI23285
aI23539
aI75027
asS'd'
I14929
sa(dp29858
S'a'
I25482
sS's'
I18
sS'm'
(lp29859
I1
aI1
aI5
aI7
aI17
aI11
aI65
aI157
aI93
aI607
aI1445
aI4089
aI3139
aI4699
aI1225
aI58935
aI93673
aI146467
asS'd'
I14930
sa(dp29860
S'a'
I25489
sS's'
I18
sS'm'
(lp29861
I1
aI3
aI3
aI13
aI29
aI59
aI69
aI141
aI257
aI463
aI93
aI649
aI8179
aI15205
aI6943
aI45317
aI31269
aI70825
asS'd'
I14931
sa(dp29862
S'a'
I25490
sS's'
I18
sS'm'
(lp29863
I1
aI3
aI5
aI9
aI23
aI39
aI113
aI61
aI315
aI463
aI1739
aI149
aI6007
aI2789
aI12021
aI969
aI18551
aI153669
asS'd'
I14932
sa(dp29864
S'a'
I25505
sS's'
I18
sS'm'
(lp29865
I1
aI1
aI3
aI15
aI19
aI9
aI23
aI211
aI265
aI877
aI325
aI2635
aI8131
aI4957
aI24371
aI60975
aI3887
aI198927
asS'd'
I14933
sa(dp29866
S'a'
I25511
sS's'
I18
sS'm'
(lp29867
I1
aI1
aI3
aI11
aI29
aI31
aI5
aI105
aI157
aI573
aI2009
aI1701
aI1549
aI1641
aI17429
aI13587
aI48421
aI8675
asS'd'
I14934
sa(dp29868
S'a'
I25512
sS's'
I18
sS'm'
(lp29869
I1
aI1
aI5
aI13
aI3
aI13
aI17
aI55
aI101
aI369
aI705
aI3635
aI5195
aI10439
aI12881
aI21565
aI1671
aI75489
asS'd'
I14935
sa(dp29870
S'a'
I25552
sS's'
I18
sS'm'
(lp29871
I1
aI3
aI3
aI1
aI1
aI23
aI85
aI189
aI347
aI205
aI5
aI3465
aI3269
aI3347
aI10163
aI26921
aI86555
aI9387
asS'd'
I14936
sa(dp29872
S'a'
I25555
sS's'
I18
sS'm'
(lp29873
I1
aI3
aI7
aI15
aI27
aI21
aI79
aI151
aI279
aI627
aI1093
aI1929
aI5549
aI12141
aI5245
aI55747
aI65939
aI193759
asS'd'
I14937
sa(dp29874
S'a'
I25595
sS's'
I18
sS'm'
(lp29875
I1
aI1
aI1
aI5
aI5
aI23
aI57
aI235
aI143
aI129
aI795
aI35
aI4375
aI12577
aI871
aI20879
aI82811
aI52279
asS'd'
I14938
sa(dp29876
S'a'
I25636
sS's'
I18
sS'm'
(lp29877
I1
aI3
aI1
aI5
aI11
aI1
aI125
aI99
aI89
aI629
aI857
aI2631
aI393
aI15075
aI27473
aI42695
aI61505
aI239651
asS'd'
I14939
sa(dp29878
S'a'
I25672
sS's'
I18
sS'm'
(lp29879
I1
aI1
aI5
aI1
aI9
aI11
aI55
aI203
aI453
aI677
aI259
aI1979
aI4101
aI16067
aI26783
aI17907
aI75349
aI62797
asS'd'
I14940
sa(dp29880
S'a'
I25685
sS's'
I18
sS'm'
(lp29881
I1
aI3
aI7
aI5
aI5
aI19
aI85
aI165
aI341
aI405
aI1779
aI87
aI889
aI265
aI9851
aI36175
aI69697
aI123769
asS'd'
I14941
sa(dp29882
S'a'
I25696
sS's'
I18
sS'm'
(lp29883
I1
aI3
aI7
aI1
aI7
aI49
aI93
aI57
aI85
aI597
aI183
aI3253
aI6301
aI9307
aI8753
aI38133
aI58743
aI19621
asS'd'
I14942
sa(dp29884
S'a'
I25720
sS's'
I18
sS'm'
(lp29885
I1
aI1
aI3
aI9
aI1
aI15
aI125
aI215
aI391
aI141
aI87
aI37
aI4333
aI5033
aI30549
aI64281
aI18577
aI156093
asS'd'
I14943
sa(dp29886
S'a'
I25730
sS's'
I18
sS'm'
(lp29887
I1
aI3
aI3
aI15
aI17
aI25
aI29
aI81
aI339
aI865
aI1619
aI773
aI901
aI8163
aI22275
aI57159
aI119951
aI137451
asS'd'
I14944
sa(dp29888
S'a'
I25735
sS's'
I18
sS'm'
(lp29889
I1
aI3
aI5
aI15
aI7
aI57
aI113
aI221
aI13
aI49
aI1653
aI3695
aI4423
aI4383
aI28669
aI64175
aI130355
aI202543
asS'd'
I14945
sa(dp29890
S'a'
I25753
sS's'
I18
sS'm'
(lp29891
I1
aI1
aI3
aI9
aI29
aI21
aI19
aI119
aI409
aI717
aI1853
aI3981
aI4489
aI3985
aI31205
aI10423
aI13223
aI131973
asS'd'
I14946
sa(dp29892
S'a'
I25756
sS's'
I18
sS'm'
(lp29893
I1
aI1
aI3
aI15
aI29
aI7
aI1
aI85
aI133
aI345
aI317
aI2363
aI7803
aI4975
aI19441
aI10497
aI42059
aI131531
asS'd'
I14947
sa(dp29894
S'a'
I25769
sS's'
I18
sS'm'
(lp29895
I1
aI1
aI7
aI13
aI3
aI53
aI49
aI27
aI487
aI901
aI801
aI335
aI6317
aI14205
aI26655
aI52747
aI102659
aI231359
asS'd'
I14948
sa(dp29896
S'a'
I25821
sS's'
I18
sS'm'
(lp29897
I1
aI3
aI3
aI3
aI13
aI33
aI113
aI107
aI407
aI499
aI903
aI3059
aI1343
aI11859
aI6315
aI23071
aI73627
aI44239
asS'd'
I14949
sa(dp29898
S'a'
I25825
sS's'
I18
sS'm'
(lp29899
I1
aI3
aI5
aI15
aI1
aI53
aI59
aI247
aI213
aI981
aI443
aI3
aI615
aI12067
aI3881
aI61759
aI101219
aI110407
asS'd'
I14950
sa(dp29900
S'a'
I25840
sS's'
I18
sS'm'
(lp29901
I1
aI3
aI5
aI13
aI7
aI31
aI87
aI161
aI61
aI1023
aI147
aI2075
aI7245
aI9025
aI7229
aI60935
aI104481
aI169561
asS'd'
I14951
sa(dp29902
S'a'
I25860
sS's'
I18
sS'm'
(lp29903
I1
aI3
aI1
aI15
aI25
aI17
aI53
aI107
aI311
aI621
aI1493
aI2443
aI4635
aI12163
aI12543
aI43031
aI90843
aI139645
asS'd'
I14952
sa(dp29904
S'a'
I25869
sS's'
I18
sS'm'
(lp29905
I1
aI3
aI5
aI11
aI7
aI49
aI91
aI165
aI91
aI329
aI493
aI3533
aI7429
aI7047
aI14767
aI31641
aI62005
aI77267
asS'd'
I14953
sa(dp29906
S'a'
I25872
sS's'
I18
sS'm'
(lp29907
I1
aI1
aI3
aI9
aI31
aI21
aI19
aI167
aI185
aI199
aI1989
aI1093
aI4213
aI4769
aI21659
aI19685
aI122123
aI215233
asS'd'
I14954
sa(dp29908
S'a'
I25956
sS's'
I18
sS'm'
(lp29909
I1
aI3
aI3
aI7
aI27
aI23
aI99
aI205
aI365
aI689
aI1281
aI419
aI4207
aI5355
aI20245
aI25029
aI123029
aI61499
asS'd'
I14955
sa(dp29910
S'a'
I25973
sS's'
I18
sS'm'
(lp29911
I1
aI3
aI3
aI7
aI19
aI15
aI29
aI185
aI165
aI203
aI1859
aI2895
aI6361
aI6331
aI13641
aI42577
aI33757
aI41305
asS'd'
I14956
sa(dp29912
S'a'
I25994
sS's'
I18
sS'm'
(lp29913
I1
aI3
aI5
aI11
aI21
aI43
aI15
aI11
aI425
aI125
aI1597
aI1109
aI3335
aI7009
aI20799
aI41261
aI127813
aI181261
asS'd'
I14957
sa(dp29914
S'a'
I26002
sS's'
I18
sS'm'
(lp29915
I1
aI1
aI5
aI5
aI27
aI35
aI35
aI159
aI111
aI1011
aI1487
aI813
aI4985
aI2555
aI23741
aI44675
aI97159
aI250477
asS'd'
I14958
sa(dp29916
S'a'
I26020
sS's'
I18
sS'm'
(lp29917
I1
aI3
aI3
aI7
aI17
aI41
aI81
aI187
aI367
aI767
aI1345
aI205
aI5797
aI9129
aI21973
aI39911
aI130131
aI96891
asS'd'
I14959
sa(dp29918
S'a'
I26023
sS's'
I18
sS'm'
(lp29919
I1
aI3
aI5
aI1
aI1
aI47
aI57
aI177
aI127
aI791
aI1427
aI1895
aI5995
aI12569
aI32711
aI58599
aI55641
aI80405
asS'd'
I14960
sa(dp29920
S'a'
I26041
sS's'
I18
sS'm'
(lp29921
I1
aI3
aI3
aI1
aI23
aI13
aI115
aI81
aI511
aI677
aI775
aI3143
aI4963
aI7093
aI15963
aI59893
aI22609
aI137601
asS'd'
I14961
sa(dp29922
S'a'
I26044
sS's'
I18
sS'm'
(lp29923
I1
aI1
aI5
aI9
aI17
aI17
aI115
aI127
aI397
aI139
aI1171
aI207
aI3485
aI15869
aI465
aI26267
aI29957
aI205459
asS'd'
I14962
sa(dp29924
S'a'
I26069
sS's'
I18
sS'm'
(lp29925
I1
aI3
aI7
aI3
aI9
aI29
aI23
aI189
aI447
aI481
aI753
aI2415
aI2669
aI6007
aI15201
aI7317
aI18861
aI173759
asS'd'
I14963
sa(dp29926
S'a'
I26097
sS's'
I18
sS'm'
(lp29927
I1
aI3
aI5
aI7
aI13
aI43
aI13
aI163
aI363
aI683
aI1869
aI1237
aI2523
aI3661
aI13887
aI5593
aI91513
aI220177
asS'd'
I14964
sa(dp29928
S'a'
I26113
sS's'
I18
sS'm'
(lp29929
I1
aI1
aI1
aI11
aI7
aI11
aI43
aI39
aI319
aI793
aI375
aI3159
aI7621
aI8965
aI25743
aI351
aI31873
aI18115
asS'd'
I14965
sa(dp29930
S'a'
I26138
sS's'
I18
sS'm'
(lp29931
I1
aI1
aI5
aI5
aI11
aI57
aI79
aI147
aI55
aI553
aI417
aI1365
aI3979
aI9789
aI22677
aI58645
aI104549
aI9019
asS'd'
I14966
sa(dp29932
S'a'
I26167
sS's'
I18
sS'm'
(lp29933
I1
aI1
aI3
aI11
aI17
aI37
aI127
aI5
aI165
aI867
aI79
aI2259
aI197
aI4789
aI28109
aI46721
aI3431
aI118939
asS'd'
I14967
sa(dp29934
S'a'
I26216
sS's'
I18
sS'm'
(lp29935
I1
aI1
aI5
aI13
aI19
aI45
aI113
aI11
aI125
aI351
aI1753
aI3201
aI1697
aI2567
aI9717
aI22247
aI84309
aI248583
asS'd'
I14968
sa(dp29936
S'a'
I26230
sS's'
I18
sS'm'
(lp29937
I1
aI1
aI5
aI9
aI5
aI37
aI65
aI47
aI261
aI855
aI1573
aI2267
aI7977
aI13029
aI32527
aI59805
aI103591
aI180041
asS'd'
I14969
sa(dp29938
S'a'
I26243
sS's'
I18
sS'm'
(lp29939
I1
aI3
aI3
aI9
aI31
aI5
aI115
aI159
aI111
aI899
aI1907
aI2671
aI1575
aI7021
aI10281
aI34905
aI641
aI63549
asS'd'
I14970
sa(dp29940
S'a'
I26245
sS's'
I18
sS'm'
(lp29941
I1
aI3
aI5
aI1
aI17
aI61
aI45
aI9
aI375
aI761
aI117
aI1767
aI4657
aI12217
aI12067
aI42807
aI118587
aI72715
asS'd'
I14971
sa(dp29942
S'a'
I26263
sS's'
I18
sS'm'
(lp29943
I1
aI1
aI1
aI13
aI3
aI3
aI93
aI3
aI351
aI97
aI119
aI1743
aI81
aI12761
aI22529
aI47191
aI111315
aI256501
asS'd'
I14972
sa(dp29944
S'a'
I26264
sS's'
I18
sS'm'
(lp29945
I1
aI1
aI5
aI7
aI29
aI23
aI41
aI9
aI231
aI567
aI1565
aI3539
aI7241
aI11535
aI7375
aI10391
aI127045
aI9371
asS'd'
I14973
sa(dp29946
S'a'
I26269
sS's'
I18
sS'm'
(lp29947
I1
aI1
aI1
aI15
aI23
aI47
aI39
aI57
aI73
aI809
aI513
aI3233
aI8071
aI8595
aI13817
aI821
aI89091
aI107173
asS'd'
I14974
sa(dp29948
S'a'
I26311
sS's'
I18
sS'm'
(lp29949
I1
aI1
aI1
aI13
aI13
aI43
aI75
aI239
aI487
aI175
aI1561
aI3925
aI3743
aI14247
aI15713
aI55005
aI116135
aI199827
asS'd'
I14975
sa(dp29950
S'a'
I26346
sS's'
I18
sS'm'
(lp29951
I1
aI1
aI3
aI7
aI13
aI15
aI67
aI147
aI77
aI241
aI1763
aI651
aI1107
aI4943
aI15651
aI23259
aI45931
aI34717
asS'd'
I14976
sa(dp29952
S'a'
I26354
sS's'
I18
sS'm'
(lp29953
I1
aI1
aI3
aI5
aI15
aI15
aI67
aI153
aI163
aI179
aI1567
aI685
aI3245
aI2205
aI8373
aI56567
aI32091
aI23313
asS'd'
I14977
sa(dp29954
S'a'
I26365
sS's'
I18
sS'm'
(lp29955
I1
aI1
aI7
aI9
aI1
aI27
aI79
aI209
aI263
aI517
aI635
aI3
aI103
aI2173
aI22659
aI11319
aI103757
aI144449
asS'd'
I14978
sa(dp29956
S'a'
I26368
sS's'
I18
sS'm'
(lp29957
I1
aI1
aI1
aI5
aI11
aI63
aI21
aI89
aI443
aI775
aI327
aI1559
aI1421
aI2309
aI18597
aI46385
aI16547
aI186813
asS'd'
I14979
sa(dp29958
S'a'
I26398
sS's'
I18
sS'm'
(lp29959
I1
aI1
aI7
aI9
aI27
aI37
aI43
aI7
aI305
aI117
aI1103
aI1801
aI3349
aI12225
aI28215
aI8857
aI118677
aI88909
asS'd'
I14980
sa(dp29960
S'a'
I26401
sS's'
I18
sS'm'
(lp29961
I1
aI3
aI3
aI7
aI15
aI59
aI21
aI19
aI371
aI81
aI755
aI1565
aI4823
aI16363
aI20301
aI33571
aI74423
aI177205
asS'd'
I14981
sa(dp29962
S'a'
I26426
sS's'
I18
sS'm'
(lp29963
I1
aI1
aI3
aI9
aI15
aI33
aI23
aI31
aI171
aI713
aI271
aI2437
aI3609
aI4271
aI24355
aI46283
aI121767
aI188501
asS'd'
I14982
sa(dp29964
S'a'
I26436
sS's'
I18
sS'm'
(lp29965
I1
aI1
aI1
aI1
aI21
aI3
aI3
aI241
aI339
aI211
aI443
aI1577
aI343
aI2625
aI1077
aI29933
aI106401
aI51439
asS'd'
I14983
sa(dp29966
S'a'
I26443
sS's'
I18
sS'm'
(lp29967
I1
aI3
aI5
aI15
aI13
aI31
aI11
aI167
aI15
aI101
aI373
aI2095
aI3017
aI1347
aI15029
aI6579
aI21233
aI87589
asS'd'
I14984
sa(dp29968
S'a'
I26448
sS's'
I18
sS'm'
(lp29969
I1
aI3
aI1
aI3
aI15
aI1
aI61
aI239
aI13
aI867
aI1621
aI2275
aI5757
aI8275
aI7923
aI44469
aI113513
aI84927
asS'd'
I14985
sa(dp29970
S'a'
I26458
sS's'
I18
sS'm'
(lp29971
I1
aI1
aI1
aI13
aI23
aI27
aI101
aI25
aI459
aI517
aI127
aI1131
aI669
aI13209
aI23671
aI3379
aI66091
aI72919
asS'd'
I14986
sa(dp29972
S'a'
I26464
sS's'
I18
sS'm'
(lp29973
I1
aI3
aI5
aI7
aI29
aI53
aI25
aI185
aI101
aI707
aI281
aI183
aI2823
aI7241
aI3127
aI48093
aI20195
aI208349
asS'd'
I14987
sa(dp29974
S'a'
I26476
sS's'
I18
sS'm'
(lp29975
I1
aI3
aI3
aI15
aI31
aI11
aI3
aI165
aI453
aI609
aI1053
aI3937
aI1989
aI13887
aI13415
aI8005
aI103537
aI17853
asS'd'
I14988
sa(dp29976
S'a'
I26503
sS's'
I18
sS'm'
(lp29977
I1
aI1
aI3
aI7
aI3
aI41
aI13
aI67
aI227
aI265
aI767
aI391
aI1835
aI8827
aI13131
aI42605
aI117089
aI12475
asS'd'
I14989
sa(dp29978
S'a'
I26504
sS's'
I18
sS'm'
(lp29979
I1
aI1
aI7
aI3
aI15
aI31
aI45
aI157
aI261
aI207
aI1109
aI1587
aI5389
aI13239
aI31697
aI35969
aI79839
aI209633
asS'd'
I14990
sa(dp29980
S'a'
I26515
sS's'
I18
sS'm'
(lp29981
I1
aI1
aI5
aI9
aI31
aI3
aI35
aI187
aI5
aI945
aI633
aI2645
aI171
aI2221
aI18369
aI41765
aI82373
aI8007
asS'd'
I14991
sa(dp29982
S'a'
I26531
sS's'
I18
sS'm'
(lp29983
I1
aI1
aI7
aI7
aI3
aI57
aI73
aI103
aI245
aI811
aI1637
aI101
aI6335
aI9911
aI663
aI21779
aI31681
aI119141
asS'd'
I14992
sa(dp29984
S'a'
I26548
sS's'
I18
sS'm'
(lp29985
I1
aI1
aI5
aI13
aI27
aI25
aI5
aI203
aI183
aI251
aI1803
aI665
aI6295
aI965
aI5269
aI379
aI78455
aI7097
asS'd'
I14993
sa(dp29986
S'a'
I26570
sS's'
I18
sS'm'
(lp29987
I1
aI1
aI5
aI3
aI19
aI55
aI45
aI161
aI481
aI737
aI1903
aI1093
aI3313
aI4427
aI7959
aI6231
aI94769
aI123827
asS'd'
I14994
sa(dp29988
S'a'
I26589
sS's'
I18
sS'm'
(lp29989
I1
aI1
aI7
aI11
aI29
aI41
aI77
aI165
aI49
aI875
aI137
aI2003
aI8093
aI1941
aI25979
aI10765
aI99241
aI71275
asS'd'
I14995
sa(dp29990
S'a'
I26594
sS's'
I18
sS'm'
(lp29991
I1
aI3
aI5
aI3
aI23
aI1
aI89
aI163
aI293
aI701
aI29
aI2543
aI4487
aI14873
aI28123
aI48643
aI31633
aI74179
asS'd'
I14996
sa(dp29992
S'a'
I26608
sS's'
I18
sS'm'
(lp29993
I1
aI1
aI7
aI1
aI13
aI33
aI33
aI173
aI111
aI959
aI205
aI1633
aI3127
aI3963
aI6455
aI41809
aI60655
aI247121
asS'd'
I14997
sa(dp29994
S'a'
I26623
sS's'
I18
sS'm'
(lp29995
I1
aI3
aI5
aI13
aI1
aI49
aI87
aI217
aI381
aI125
aI823
aI837
aI3967
aI8157
aI11097
aI35721
aI93591
aI3939
asS'd'
I14998
sa(dp29996
S'a'
I26678
sS's'
I18
sS'm'
(lp29997
I1
aI3
aI7
aI1
aI7
aI27
aI29
aI21
aI295
aI127
aI823
aI2409
aI1873
aI2417
aI27961
aI39211
aI14785
aI71557
asS'd'
I14999
sa(dp29998
S'a'
I26690
sS's'
I18
sS'm'
(lp29999
I1
aI3
aI1
aI5
aI31
aI59
aI43
aI121
aI217
aI417
aI2029
aI3983
aI5629
aI10447
aI1073
aI57515
aI58849
aI178927
asS'd'
I15000
sa(dp30000
S'a'
I26709
sS's'
I18
sS'm'
(lp30001
I1
aI3
aI3
aI11
aI5
aI59
aI45
aI39
aI269
aI483
aI757
aI3245
aI4383
aI11127
aI26535
aI17395
aI60953
aI259333
asS'd'
I15001
sa(dp30002
S'a'
I26725
sS's'
I18
sS'm'
(lp30003
I1
aI1
aI1
aI5
aI5
aI49
aI81
aI241
aI371
aI353
aI1293
aI3375
aI6725
aI11891
aI5973
aI13901
aI37999
aI17751
asS'd'
I15002
sa(dp30004
S'a'
I26744
sS's'
I18
sS'm'
(lp30005
I1
aI3
aI3
aI7
aI31
aI21
aI45
aI107
aI33
aI911
aI1869
aI2643
aI2655
aI3979
aI5509
aI33065
aI128463
aI246937
asS'd'
I15003
sa(dp30006
S'a'
I26763
sS's'
I18
sS'm'
(lp30007
I1
aI1
aI1
aI1
aI31
aI33
aI99
aI29
aI485
aI11
aI1423
aI1775
aI2045
aI741
aI30691
aI53957
aI13891
aI57303
asS'd'
I15004
sa(dp30008
S'a'
I26765
sS's'
I18
sS'm'
(lp30009
I1
aI1
aI7
aI13
aI7
aI37
aI117
aI121
aI51
aI743
aI887
aI1769
aI1049
aI12859
aI1663
aI27763
aI90969
aI38935
asS'd'
I15005
sa(dp30010
S'a'
I26768
sS's'
I18
sS'm'
(lp30011
I1
aI3
aI5
aI7
aI3
aI41
aI121
aI89
aI461
aI979
aI457
aI2973
aI8109
aI13819
aI30237
aI54671
aI66967
aI135233
asS'd'
I15006
sa(dp30012
S'a'
I26784
sS's'
I18
sS'm'
(lp30013
I1
aI1
aI7
aI13
aI13
aI47
aI51
aI121
aI295
aI847
aI681
aI1163
aI8123
aI14179
aI26561
aI54057
aI74043
aI146155
asS'd'
I15007
sa(dp30014
S'a'
I26802
sS's'
I18
sS'm'
(lp30015
I1
aI3
aI3
aI11
aI21
aI15
aI9
aI85
aI445
aI11
aI1525
aI3165
aI5929
aI12481
aI10769
aI31885
aI51487
aI248933
asS'd'
I15008
sa(dp30016
S'a'
I26814
sS's'
I18
sS'm'
(lp30017
I1
aI3
aI1
aI9
aI25
aI41
aI59
aI139
aI293
aI1021
aI2043
aI967
aI3949
aI7309
aI6545
aI62761
aI37761
aI22395
asS'd'
I15009
sa(dp30018
S'a'
I26856
sS's'
I18
sS'm'
(lp30019
I1
aI3
aI7
aI15
aI25
aI45
aI29
aI75
aI283
aI845
aI687
aI3285
aI7263
aI10237
aI5343
aI58635
aI85137
aI2387
asS'd'
I15010
sa(dp30020
S'a'
I26861
sS's'
I18
sS'm'
(lp30021
I1
aI1
aI3
aI15
aI29
aI33
aI111
aI175
aI251
aI181
aI709
aI1373
aI1661
aI1155
aI30479
aI57823
aI28809
aI74117
asS'd'
I15011
sa(dp30022
S'a'
I26862
sS's'
I18
sS'm'
(lp30023
I1
aI3
aI3
aI5
aI11
aI37
aI55
aI155
aI439
aI173
aI1861
aI1713
aI1675
aI12119
aI12531
aI50995
aI124657
aI58321
asS'd'
I15012
sa(dp30024
S'a'
I26894
sS's'
I18
sS'm'
(lp30025
I1
aI1
aI5
aI13
aI1
aI23
aI27
aI141
aI3
aI985
aI1439
aI781
aI7381
aI2223
aI26673
aI46607
aI54953
aI24547
asS'd'
I15013
sa(dp30026
S'a'
I26908
sS's'
I18
sS'm'
(lp30027
I1
aI1
aI5
aI7
aI5
aI3
aI41
aI115
aI503
aI731
aI633
aI3631
aI3455
aI15937
aI22639
aI41163
aI65243
aI233749
asS'd'
I15014
sa(dp30028
S'a'
I26915
sS's'
I18
sS'm'
(lp30029
I1
aI3
aI1
aI1
aI5
aI57
aI89
aI35
aI53
aI653
aI1763
aI1247
aI6999
aI1811
aI28191
aI52327
aI129421
aI191093
asS'd'
I15015
sa(dp30030
S'a'
I26929
sS's'
I18
sS'm'
(lp30031
I1
aI3
aI3
aI9
aI3
aI27
aI107
aI13
aI475
aI409
aI1623
aI483
aI3127
aI12841
aI4777
aI36157
aI24967
aI89795
asS'd'
I15016
sa(dp30032
S'a'
I26930
sS's'
I18
sS'm'
(lp30033
I1
aI1
aI3
aI5
aI7
aI29
aI15
aI225
aI257
aI923
aI251
aI21
aI4559
aI3571
aI9351
aI1739
aI37393
aI170789
asS'd'
I15017
sa(dp30034
S'a'
I26939
sS's'
I18
sS'm'
(lp30035
I1
aI3
aI7
aI5
aI11
aI7
aI107
aI237
aI343
aI665
aI767
aI2293
aI4781
aI4811
aI11227
aI25045
aI3951
aI44307
asS'd'
I15018
sa(dp30036
S'a'
I26941
sS's'
I18
sS'm'
(lp30037
I1
aI3
aI7
aI7
aI13
aI31
aI47
aI9
aI121
aI441
aI1011
aI2015
aI8053
aI355
aI13441
aI23213
aI60675
aI259761
asS'd'
I15019
sa(dp30038
S'a'
I26987
sS's'
I18
sS'm'
(lp30039
I1
aI1
aI5
aI9
aI21
aI61
aI85
aI141
aI271
aI577
aI17
aI243
aI3049
aI2479
aI28947
aI53351
aI67379
aI211133
asS'd'
I15020
sa(dp30040
S'a'
I27023
sS's'
I18
sS'm'
(lp30041
I1
aI1
aI1
aI15
aI23
aI1
aI119
aI141
aI311
aI543
aI1463
aI3633
aI8111
aI9439
aI4147
aI64913
aI28261
aI197217
asS'd'
I15021
sa(dp30042
S'a'
I27066
sS's'
I18
sS'm'
(lp30043
I1
aI3
aI1
aI15
aI15
aI33
aI125
aI231
aI225
aI797
aI605
aI259
aI3673
aI10423
aI7541
aI26289
aI61681
aI136463
asS'd'
I15022
sa(dp30044
S'a'
I27079
sS's'
I18
sS'm'
(lp30045
I1
aI3
aI1
aI13
aI23
aI15
aI15
aI151
aI289
aI657
aI1883
aI2923
aI6861
aI1411
aI17159
aI9353
aI79463
aI135813
asS'd'
I15023
sa(dp30046
S'a'
I27086
sS's'
I18
sS'm'
(lp30047
I1
aI3
aI7
aI11
aI5
aI59
aI101
aI167
aI291
aI63
aI753
aI1105
aI8173
aI2389
aI22097
aI7207
aI110377
aI15797
asS'd'
I15024
sa(dp30048
S'a'
I27091
sS's'
I18
sS'm'
(lp30049
I1
aI3
aI7
aI15
aI7
aI9
aI7
aI135
aI303
aI675
aI1803
aI2827
aI1711
aI9543
aI16567
aI24075
aI17065
aI22193
asS'd'
I15025
sa(dp30050
S'a'
I27113
sS's'
I18
sS'm'
(lp30051
I1
aI3
aI5
aI7
aI17
aI7
aI125
aI57
aI423
aI733
aI1813
aI4031
aI713
aI10687
aI27315
aI37599
aI78807
aI103429
asS'd'
I15026
sa(dp30052
S'a'
I27157
sS's'
I18
sS'm'
(lp30053
I1
aI1
aI5
aI15
aI21
aI11
aI87
aI21
aI415
aI571
aI1169
aI2561
aI7071
aI12499
aI195
aI20111
aI116757
aI157731
asS'd'
I15027
sa(dp30054
S'a'
I27174
sS's'
I18
sS'm'
(lp30055
I1
aI3
aI3
aI15
aI7
aI33
aI11
aI241
aI23
aI189
aI599
aI2891
aI2829
aI13327
aI21777
aI57733
aI38583
aI162161
asS'd'
I15028
sa(dp30056
S'a'
I27188
sS's'
I18
sS'm'
(lp30057
I1
aI1
aI1
aI15
aI3
aI49
aI7
aI143
aI291
aI301
aI1439
aI793
aI3447
aI1167
aI2815
aI24875
aI117437
aI112561
asS'd'
I15029
sa(dp30058
S'a'
I27191
sS's'
I18
sS'm'
(lp30059
I1
aI3
aI5
aI13
aI29
aI11
aI51
aI255
aI365
aI741
aI1919
aI2091
aI2865
aI12721
aI4329
aI37281
aI128703
aI739
asS'd'
I15030
sa(dp30060
S'a'
I27197
sS's'
I18
sS'm'
(lp30061
I1
aI1
aI1
aI13
aI19
aI31
aI39
aI141
aI81
aI133
aI297
aI3837
aI7537
aI16043
aI11755
aI10289
aI74399
aI95371
asS'd'
I15031
sa(dp30062
S'a'
I27206
sS's'
I18
sS'm'
(lp30063
I1
aI1
aI7
aI5
aI21
aI35
aI125
aI109
aI241
aI217
aI1219
aI2617
aI1925
aI9573
aI19305
aI44689
aI89365
aI248869
asS'd'
I15032
sa(dp30064
S'a'
I27210
sS's'
I18
sS'm'
(lp30065
I1
aI3
aI5
aI5
aI13
aI33
aI43
aI221
aI325
aI267
aI837
aI809
aI6025
aI9847
aI9267
aI13465
aI45937
aI204339
asS'd'
I15033
sa(dp30066
S'a'
I27245
sS's'
I18
sS'm'
(lp30067
I1
aI3
aI1
aI3
aI25
aI53
aI85
aI249
aI105
aI619
aI917
aI1213
aI5365
aI6197
aI22929
aI27529
aI71011
aI141651
asS'd'
I15034
sa(dp30068
S'a'
I27254
sS's'
I18
sS'm'
(lp30069
I1
aI1
aI5
aI1
aI29
aI9
aI27
aI161
aI269
aI775
aI1043
aI303
aI4503
aI5059
aI479
aI17237
aI51383
aI152495
asS'd'
I15035
sa(dp30070
S'a'
I27383
sS's'
I18
sS'm'
(lp30071
I1
aI3
aI1
aI5
aI19
aI5
aI127
aI139
aI1
aI461
aI943
aI593
aI7457
aI3357
aI1909
aI64633
aI91811
aI92703
asS'd'
I15036
sa(dp30072
S'a'
I27387
sS's'
I18
sS'm'
(lp30073
I1
aI1
aI5
aI7
aI1
aI21
aI83
aI29
aI123
aI83
aI1085
aI2727
aI651
aI15801
aI20561
aI34821
aI17671
aI162227
asS'd'
I15037
sa(dp30074
S'a'
I27416
sS's'
I18
sS'm'
(lp30075
I1
aI3
aI7
aI1
aI19
aI59
aI33
aI195
aI81
aI69
aI51
aI1473
aI3873
aI4247
aI3587
aI4293
aI30831
aI245345
asS'd'
I15038
sa(dp30076
S'a'
I27422
sS's'
I18
sS'm'
(lp30077
I1
aI3
aI1
aI3
aI23
aI27
aI19
aI115
aI275
aI293
aI705
aI131
aI1001
aI8881
aI30165
aI25149
aI38679
aI175167
asS'd'
I15039
sa(dp30078
S'a'
I27472
sS's'
I18
sS'm'
(lp30079
I1
aI1
aI3
aI15
aI11
aI11
aI79
aI55
aI323
aI217
aI859
aI897
aI6567
aI12529
aI3161
aI13009
aI100787
aI3501
asS'd'
I15040
sa(dp30080
S'a'
I27478
sS's'
I18
sS'm'
(lp30081
I1
aI3
aI1
aI15
aI17
aI63
aI51
aI71
aI55
aI207
aI1095
aI2527
aI611
aI9281
aI7375
aI14553
aI16021
aI88537
asS'd'
I15041
sa(dp30082
S'a'
I27548
sS's'
I18
sS'm'
(lp30083
I1
aI1
aI7
aI9
aI11
aI23
aI95
aI25
aI327
aI873
aI575
aI3583
aI6587
aI137
aI23737
aI59341
aI83281
aI93541
asS'd'
I15042
sa(dp30084
S'a'
I27579
sS's'
I18
sS'm'
(lp30085
I1
aI1
aI5
aI7
aI15
aI37
aI89
aI105
aI471
aI757
aI103
aI3747
aI3565
aI4957
aI23537
aI16193
aI84843
aI256757
asS'd'
I15043
sa(dp30086
S'a'
I27582
sS's'
I18
sS'm'
(lp30087
I1
aI3
aI5
aI1
aI15
aI17
aI119
aI231
aI387
aI715
aI797
aI3807
aI4985
aI8335
aI4885
aI45541
aI69597
aI238599
asS'd'
I15044
sa(dp30088
S'a'
I27589
sS's'
I18
sS'm'
(lp30089
I1
aI1
aI3
aI1
aI7
aI21
aI87
aI205
aI39
aI503
aI433
aI3643
aI4719
aI2051
aI10049
aI28997
aI75981
aI253647
asS'd'
I15045
sa(dp30090
S'a'
I27593
sS's'
I18
sS'm'
(lp30091
I1
aI3
aI1
aI13
aI9
aI21
aI103
aI63
aI27
aI267
aI185
aI3507
aI3009
aI5183
aI2261
aI40249
aI33733
aI70493
asS'd'
I15046
sa(dp30092
S'a'
I27613
sS's'
I18
sS'm'
(lp30093
I1
aI1
aI3
aI13
aI7
aI7
aI79
aI13
aI141
aI327
aI1035
aI1699
aI6273
aI5621
aI13877
aI57607
aI50207
aI184643
asS'd'
I15047
sa(dp30094
S'a'
I27649
sS's'
I18
sS'm'
(lp30095
I1
aI1
aI3
aI1
aI9
aI19
aI75
aI99
aI115
aI469
aI1025
aI1999
aI1985
aI9975
aI11069
aI59113
aI80877
aI124153
asS'd'
I15048
sa(dp30096
S'a'
I27717
sS's'
I18
sS'm'
(lp30097
I1
aI3
aI5
aI7
aI19
aI27
aI47
aI3
aI313
aI575
aI107
aI991
aI2575
aI11001
aI12323
aI21443
aI126245
aI219649
asS'd'
I15049
sa(dp30098
S'a'
I27729
sS's'
I18
sS'm'
(lp30099
I1
aI1
aI1
aI5
aI19
aI33
aI13
aI1
aI357
aI225
aI1355
aI1827
aI7127
aI6387
aI19299
aI24935
aI26847
aI251433
asS'd'
I15050
sa(dp30100
S'a'
I27751
sS's'
I18
sS'm'
(lp30101
I1
aI1
aI1
aI5
aI3
aI3
aI113
aI19
aI425
aI209
aI159
aI347
aI1349
aI6771
aI13125
aI8393
aI21435
aI186369
asS'd'
I15051
sa(dp30102
S'a'
I27757
sS's'
I18
sS'm'
(lp30103
I1
aI1
aI5
aI11
aI5
aI39
aI95
aI3
aI193
aI741
aI1755
aI3361
aI1927
aI12909
aI5413
aI29111
aI123429
aI109191
asS'd'
I15052
sa(dp30104
S'a'
I27766
sS's'
I18
sS'm'
(lp30105
I1
aI1
aI1
aI13
aI31
aI23
aI43
aI163
aI421
aI719
aI457
aI3149
aI7741
aI1465
aI15719
aI42831
aI99051
aI164675
asS'd'
I15053
sa(dp30106
S'a'
I27791
sS's'
I18
sS'm'
(lp30107
I1
aI1
aI1
aI7
aI15
aI1
aI29
aI15
aI9
aI577
aI269
aI31
aI4361
aI5081
aI32185
aI54869
aI115105
aI151233
asS'd'
I15054
sa(dp30108
S'a'
I27794
sS's'
I18
sS'm'
(lp30109
I1
aI1
aI1
aI11
aI19
aI3
aI67
aI3
aI377
aI487
aI1287
aI3463
aI6523
aI15237
aI3171
aI38673
aI7359
aI29739
asS'd'
I15055
sa(dp30110
S'a'
I27854
sS's'
I18
sS'm'
(lp30111
I1
aI3
aI5
aI11
aI9
aI13
aI47
aI191
aI97
aI641
aI807
aI1085
aI1537
aI2855
aI24615
aI32383
aI66425
aI53713
asS'd'
I15056
sa(dp30112
S'a'
I27856
sS's'
I18
sS'm'
(lp30113
I1
aI1
aI5
aI7
aI19
aI25
aI93
aI1
aI21
aI853
aI813
aI2535
aI4291
aI9051
aI3385
aI507
aI73889
aI85397
asS'd'
I15057
sa(dp30114
S'a'
I27866
sS's'
I18
sS'm'
(lp30115
I1
aI1
aI3
aI13
aI7
aI15
aI103
aI199
aI83
aI585
aI1859
aI2089
aI839
aI8923
aI14615
aI3399
aI7703
aI229937
asS'd'
I15058
sa(dp30116
S'a'
I27875
sS's'
I18
sS'm'
(lp30117
I1
aI1
aI3
aI3
aI11
aI23
aI125
aI135
aI475
aI613
aI327
aI339
aI3081
aI13221
aI4889
aI41233
aI36547
aI195357
asS'd'
I15059
sa(dp30118
S'a'
I27902
sS's'
I18
sS'm'
(lp30119
I1
aI3
aI7
aI7
aI19
aI23
aI85
aI217
aI501
aI447
aI1873
aI2175
aI6753
aI2825
aI5171
aI47561
aI13321
aI59583
asS'd'
I15060
sa(dp30120
S'a'
I27916
sS's'
I18
sS'm'
(lp30121
I1
aI1
aI5
aI13
aI23
aI59
aI109
aI195
aI487
aI785
aI21
aI1595
aI5641
aI10103
aI8115
aI60647
aI78425
aI237379
asS'd'
I15061
sa(dp30122
S'a'
I27924
sS's'
I18
sS'm'
(lp30123
I1
aI1
aI3
aI15
aI21
aI17
aI85
aI51
aI369
aI475
aI1021
aI1129
aI7233
aI6593
aI12467
aI55399
aI128157
aI80539
asS'd'
I15062
sa(dp30124
S'a'
I27933
sS's'
I18
sS'm'
(lp30125
I1
aI3
aI3
aI9
aI31
aI27
aI69
aI145
aI489
aI251
aI1997
aI1157
aI2027
aI16109
aI4085
aI24859
aI63561
aI79591
asS'd'
I15063
sa(dp30126
S'a'
I27940
sS's'
I18
sS'm'
(lp30127
I1
aI3
aI3
aI5
aI29
aI49
aI41
aI185
aI405
aI471
aI431
aI3539
aI6593
aI1185
aI24383
aI17009
aI111215
aI79839
asS'd'
I15064
sa(dp30128
S'a'
I27947
sS's'
I18
sS'm'
(lp30129
I1
aI1
aI7
aI3
aI3
aI15
aI97
aI157
aI301
aI227
aI717
aI3291
aI2471
aI11515
aI30657
aI30745
aI72147
aI98653
asS'd'
I15065
sa(dp30130
S'a'
I27969
sS's'
I18
sS'm'
(lp30131
I1
aI3
aI5
aI1
aI23
aI21
aI67
aI223
aI185
aI385
aI137
aI2897
aI2423
aI6119
aI28471
aI52269
aI95725
aI9105
asS'd'
I15066
sa(dp30132
S'a'
I27981
sS's'
I18
sS'm'
(lp30133
I1
aI3
aI3
aI11
aI19
aI1
aI111
aI131
aI293
aI495
aI1043
aI631
aI1375
aI15347
aI22029
aI29163
aI120025
aI81631
asS'd'
I15067
sa(dp30134
S'a'
I27996
sS's'
I18
sS'm'
(lp30135
I1
aI1
aI7
aI5
aI17
aI55
aI47
aI183
aI367
aI81
aI555
aI2857
aI4787
aI5605
aI32053
aI11815
aI81771
aI234993
asS'd'
I15068
sa(dp30136
S'a'
I28005
sS's'
I18
sS'm'
(lp30137
I1
aI3
aI7
aI15
aI15
aI49
aI45
aI221
aI49
aI299
aI887
aI3991
aI2097
aI10819
aI23297
aI1823
aI11319
aI205273
asS'd'
I15069
sa(dp30138
S'a'
I28030
sS's'
I18
sS'm'
(lp30139
I1
aI1
aI1
aI1
aI1
aI15
aI91
aI253
aI213
aI849
aI501
aI1073
aI5503
aI1379
aI28887
aI51811
aI109763
aI226149
asS'd'
I15070
sa(dp30140
S'a'
I28064
sS's'
I18
sS'm'
(lp30141
I1
aI1
aI1
aI7
aI27
aI39
aI17
aI29
aI359
aI655
aI1695
aI1781
aI1203
aI1125
aI8983
aI3477
aI13925
aI218399
asS'd'
I15071
sa(dp30142
S'a'
I28102
sS's'
I18
sS'm'
(lp30143
I1
aI3
aI7
aI13
aI1
aI25
aI33
aI185
aI87
aI19
aI151
aI371
aI1221
aI4859
aI20103
aI11435
aI104263
aI218515
asS'd'
I15072
sa(dp30144
S'a'
I28108
sS's'
I18
sS'm'
(lp30145
I1
aI1
aI3
aI5
aI17
aI43
aI29
aI149
aI207
aI39
aI1539
aI2933
aI6825
aI12391
aI18163
aI24543
aI35305
aI196295
asS'd'
I15073
sa(dp30146
S'a'
I28120
sS's'
I18
sS'm'
(lp30147
I1
aI3
aI7
aI9
aI21
aI61
aI69
aI231
aI401
aI95
aI1757
aI839
aI3395
aI7573
aI6583
aI34621
aI119303
aI20767
asS'd'
I15074
sa(dp30148
S'a'
I28125
sS's'
I18
sS'm'
(lp30149
I1
aI1
aI7
aI1
aI25
aI53
aI63
aI105
aI241
aI591
aI23
aI3219
aI2387
aI13945
aI3047
aI30939
aI63243
aI60941
asS'd'
I15075
sa(dp30150
S'a'
I28130
sS's'
I18
sS'm'
(lp30151
I1
aI1
aI5
aI7
aI25
aI47
aI7
aI227
aI57
aI279
aI81
aI4017
aI3117
aI6229
aI20029
aI30031
aI25049
aI102291
asS'd'
I15076
sa(dp30152
S'a'
I28142
sS's'
I18
sS'm'
(lp30153
I1
aI3
aI3
aI1
aI21
aI15
aI69
aI57
aI311
aI9
aI853
aI3377
aI2949
aI4781
aI15419
aI54741
aI11603
aI136821
asS'd'
I15077
sa(dp30154
S'a'
I28149
sS's'
I18
sS'm'
(lp30155
I1
aI3
aI3
aI13
aI13
aI5
aI103
aI253
aI27
aI449
aI821
aI3241
aI41
aI6643
aI15217
aI61691
aI58463
aI46117
asS'd'
I15078
sa(dp30156
S'a'
I28165
sS's'
I18
sS'm'
(lp30157
I1
aI1
aI7
aI7
aI27
aI51
aI1
aI239
aI71
aI955
aI145
aI1059
aI5645
aI7025
aI4839
aI11459
aI3051
aI235989
asS'd'
I15079
sa(dp30158
S'a'
I28169
sS's'
I18
sS'm'
(lp30159
I1
aI1
aI5
aI15
aI15
aI13
aI33
aI21
aI209
aI681
aI1163
aI3109
aI1441
aI6895
aI20829
aI48769
aI35373
aI195171
asS'd'
I15080
sa(dp30160
S'a'
I28177
sS's'
I18
sS'm'
(lp30161
I1
aI1
aI7
aI3
aI11
aI7
aI25
aI27
aI463
aI77
aI1293
aI1977
aI4931
aI8089
aI11079
aI14793
aI123049
aI32259
asS'd'
I15081
sa(dp30162
S'a'
I28211
sS's'
I18
sS'm'
(lp30163
I1
aI3
aI1
aI15
aI11
aI49
aI7
aI103
aI79
aI773
aI235
aI1653
aI6477
aI8835
aI26627
aI61101
aI40633
aI98155
asS'd'
I15082
sa(dp30164
S'a'
I28213
sS's'
I18
sS'm'
(lp30165
I1
aI3
aI5
aI7
aI3
aI53
aI77
aI197
aI49
aI57
aI1533
aI3485
aI6603
aI1131
aI9073
aI37023
aI85293
aI170883
asS'd'
I15083
sa(dp30166
S'a'
I28220
sS's'
I18
sS'm'
(lp30167
I1
aI3
aI7
aI15
aI23
aI5
aI125
aI75
aI413
aI521
aI1897
aI1099
aI35
aI2013
aI687
aI51511
aI21359
aI19995
asS'd'
I15084
sa(dp30168
S'a'
I28261
sS's'
I18
sS'm'
(lp30169
I1
aI3
aI1
aI11
aI19
aI13
aI91
aI181
aI39
aI613
aI1917
aI3149
aI669
aI9927
aI20967
aI38313
aI13537
aI181873
asS'd'
I15085
sa(dp30170
S'a'
I28271
sS's'
I18
sS'm'
(lp30171
I1
aI3
aI5
aI7
aI5
aI23
aI25
aI145
aI189
aI679
aI483
aI2689
aI2855
aI9631
aI8863
aI34841
aI83311
aI211507
asS'd'
I15086
sa(dp30172
S'a'
I28299
sS's'
I18
sS'm'
(lp30173
I1
aI3
aI5
aI15
aI15
aI15
aI87
aI53
aI309
aI807
aI1405
aI259
aI3181
aI12187
aI31529
aI8861
aI70557
aI247787
asS'd'
I15087
sa(dp30174
S'a'
I28326
sS's'
I18
sS'm'
(lp30175
I1
aI1
aI7
aI13
aI7
aI15
aI1
aI205
aI91
aI325
aI1371
aI531
aI4917
aI10291
aI30827
aI32491
aI34497
aI250301
asS'd'
I15088
sa(dp30176
S'a'
I28330
sS's'
I18
sS'm'
(lp30177
I1
aI3
aI1
aI11
aI29
aI17
aI97
aI37
aI259
aI1021
aI1705
aI4001
aI4385
aI7047
aI14593
aI63443
aI3283
aI18195
asS'd'
I15089
sa(dp30178
S'a'
I28350
sS's'
I18
sS'm'
(lp30179
I1
aI1
aI7
aI9
aI9
aI55
aI11
aI113
aI351
aI513
aI197
aI873
aI1595
aI11331
aI27711
aI419
aI73097
aI144357
asS'd'
I15090
sa(dp30180
S'a'
I28355
sS's'
I18
sS'm'
(lp30181
I1
aI3
aI7
aI15
aI29
aI31
aI37
aI15
aI233
aI573
aI1457
aI293
aI5437
aI15909
aI3087
aI24535
aI6507
aI173555
asS'd'
I15091
sa(dp30182
S'a'
I28367
sS's'
I18
sS'm'
(lp30183
I1
aI1
aI1
aI7
aI7
aI47
aI81
aI241
aI257
aI389
aI233
aI3275
aI919
aI14911
aI14473
aI58457
aI78195
aI121421
asS'd'
I15092
sa(dp30184
S'a'
I28376
sS's'
I18
sS'm'
(lp30185
I1
aI3
aI7
aI13
aI1
aI63
aI9
aI233
aI231
aI771
aI1851
aI3829
aI7089
aI4573
aI13297
aI58963
aI2065
aI8365
asS'd'
I15093
sa(dp30186
S'a'
I28379
sS's'
I18
sS'm'
(lp30187
I1
aI1
aI5
aI1
aI31
aI45
aI45
aI209
aI77
aI977
aI159
aI1521
aI969
aI10115
aI32387
aI52821
aI124209
aI51841
asS'd'
I15094
sa(dp30188
S'a'
I28381
sS's'
I18
sS'm'
(lp30189
I1
aI3
aI5
aI5
aI13
aI27
aI53
aI171
aI91
aI743
aI217
aI3805
aI7721
aI15127
aI20679
aI57459
aI53649
aI16381
asS'd'
I15095
sa(dp30190
S'a'
I28447
sS's'
I18
sS'm'
(lp30191
I1
aI3
aI1
aI15
aI23
aI43
aI105
aI169
aI143
aI759
aI463
aI1237
aI3311
aI2919
aI16675
aI53049
aI12403
aI153651
asS'd'
I15096
sa(dp30192
S'a'
I28465
sS's'
I18
sS'm'
(lp30193
I1
aI3
aI3
aI11
aI5
aI27
aI1
aI135
aI17
aI683
aI679
aI2591
aI2929
aI12417
aI18379
aI61903
aI81991
aI25231
asS'd'
I15097
sa(dp30194
S'a'
I28471
sS's'
I18
sS'm'
(lp30195
I1
aI1
aI1
aI5
aI13
aI59
aI73
aI119
aI369
aI445
aI553
aI243
aI7523
aI1105
aI20349
aI8417
aI87535
aI148857
asS'd'
I15098
sa(dp30196
S'a'
I28472
sS's'
I18
sS'm'
(lp30197
I1
aI3
aI7
aI15
aI29
aI3
aI49
aI49
aI7
aI753
aI1597
aI1427
aI7485
aI9119
aI17427
aI24961
aI114897
aI62841
asS'd'
I15099
sa(dp30198
S'a'
I28478
sS's'
I18
sS'm'
(lp30199
I1
aI1
aI7
aI7
aI17
aI35
aI49
aI225
aI267
aI801
aI1359
aI2131
aI6093
aI3859
aI11305
aI6287
aI106459
aI31093
asS'd'
I15100
sa(dp30200
S'a'
I28480
sS's'
I18
sS'm'
(lp30201
I1
aI1
aI5
aI3
aI1
aI45
aI19
aI89
aI145
aI23
aI1071
aI3053
aI3463
aI6781
aI8635
aI1961
aI54403
aI183401
asS'd'
I15101
sa(dp30202
S'a'
I28489
sS's'
I18
sS'm'
(lp30203
I1
aI3
aI1
aI13
aI17
aI35
aI105
aI155
aI145
aI597
aI1169
aI3731
aI725
aI2185
aI23365
aI31849
aI113717
aI185413
asS'd'
I15102
sa(dp30204
S'a'
I28514
sS's'
I18
sS'm'
(lp30205
I1
aI3
aI3
aI5
aI3
aI5
aI13
aI119
aI39
aI383
aI1595
aI63
aI7003
aI6465
aI19847
aI37213
aI42921
aI254479
asS'd'
I15103
sa(dp30206
S'a'
I28525
sS's'
I18
sS'm'
(lp30207
I1
aI3
aI3
aI1
aI3
aI33
aI43
aI255
aI227
aI151
aI1911
aI2657
aI6529
aI3855
aI24411
aI8485
aI30385
aI193265
asS'd'
I15104
sa(dp30208
S'a'
I28537
sS's'
I18
sS'm'
(lp30209
I1
aI1
aI7
aI7
aI21
aI53
aI101
aI37
aI193
aI107
aI1095
aI369
aI6423
aI3491
aI1219
aI53385
aI31041
aI122587
asS'd'
I15105
sa(dp30210
S'a'
I28550
sS's'
I18
sS'm'
(lp30211
I1
aI1
aI7
aI3
aI13
aI39
aI101
aI109
aI113
aI201
aI619
aI2489
aI4545
aI5017
aI25519
aI44281
aI128605
aI128595
asS'd'
I15106
sa(dp30212
S'a'
I28559
sS's'
I18
sS'm'
(lp30213
I1
aI1
aI3
aI13
aI25
aI7
aI99
aI141
aI465
aI625
aI667
aI1633
aI6719
aI16195
aI365
aI34355
aI65025
aI128025
asS'd'
I15107
sa(dp30214
S'a'
I28595
sS's'
I18
sS'm'
(lp30215
I1
aI1
aI1
aI7
aI15
aI51
aI43
aI159
aI223
aI493
aI411
aI65
aI5753
aI10219
aI21885
aI33267
aI116643
aI76777
asS'd'
I15108
sa(dp30216
S'a'
I28601
sS's'
I18
sS'm'
(lp30217
I1
aI3
aI5
aI13
aI25
aI39
aI97
aI31
aI245
aI367
aI685
aI103
aI4093
aI10449
aI3849
aI52659
aI63355
aI262059
asS'd'
I15109
sa(dp30218
S'a'
I28629
sS's'
I18
sS'm'
(lp30219
I1
aI3
aI3
aI1
aI9
aI49
aI25
aI157
aI107
aI821
aI265
aI2939
aI6365
aI7539
aI17935
aI50147
aI88907
aI214317
asS'd'
I15110
sa(dp30220
S'a'
I28657
sS's'
I18
sS'm'
(lp30221
I1
aI1
aI1
aI13
aI17
aI5
aI55
aI217
aI137
aI915
aI121
aI3187
aI3111
aI7145
aI30477
aI20023
aI71969
aI179417
asS'd'
I15111
sa(dp30222
S'a'
I28667
sS's'
I18
sS'm'
(lp30223
I1
aI1
aI1
aI5
aI7
aI15
aI47
aI71
aI197
aI725
aI523
aI2207
aI5729
aI741
aI8595
aI39125
aI25431
aI101093
asS'd'
I15112
sa(dp30224
S'a'
I28711
sS's'
I18
sS'm'
(lp30225
I1
aI3
aI1
aI7
aI19
aI37
aI117
aI235
aI353
aI459
aI207
aI953
aI4955
aI14979
aI22897
aI53911
aI7783
aI203667
asS'd'
I15113
sa(dp30226
S'a'
I28718
sS's'
I18
sS'm'
(lp30227
I1
aI1
aI5
aI13
aI9
aI17
aI21
aI95
aI37
aI751
aI1463
aI2491
aI791
aI1569
aI32055
aI61415
aI113885
aI259285
asS'd'
I15114
sa(dp30228
S'a'
I28730
sS's'
I18
sS'm'
(lp30229
I1
aI1
aI3
aI11
aI1
aI23
aI5
aI73
aI61
aI719
aI215
aI469
aI3267
aI12003
aI16535
aI46913
aI58321
aI2407
asS'd'
I15115
sa(dp30230
S'a'
I28732
sS's'
I18
sS'm'
(lp30231
I1
aI1
aI5
aI11
aI5
aI9
aI81
aI1
aI275
aI877
aI791
aI1591
aI2109
aI9983
aI29085
aI15069
aI44757
aI17887
asS'd'
I15116
sa(dp30232
S'a'
I28777
sS's'
I18
sS'm'
(lp30233
I1
aI3
aI7
aI11
aI23
aI47
aI121
aI53
aI55
aI677
aI1239
aI2591
aI579
aI11321
aI14231
aI53701
aI71947
aI56793
asS'd'
I15117
sa(dp30234
S'a'
I28801
sS's'
I18
sS'm'
(lp30235
I1
aI1
aI5
aI1
aI7
aI31
aI39
aI205
aI231
aI843
aI159
aI2301
aI7765
aI3317
aI8935
aI60647
aI110545
aI142543
asS'd'
I15118
sa(dp30236
S'a'
I28822
sS's'
I18
sS'm'
(lp30237
I1
aI3
aI3
aI11
aI25
aI39
aI9
aI131
aI145
aI373
aI41
aI1687
aI417
aI9427
aI8657
aI18315
aI18505
aI144055
asS'd'
I15119
sa(dp30238
S'a'
I28855
sS's'
I18
sS'm'
(lp30239
I1
aI3
aI7
aI3
aI1
aI51
aI61
aI223
aI409
aI607
aI1281
aI1767
aI4719
aI10741
aI21537
aI17307
aI5473
aI76127
asS'd'
I15120
sa(dp30240
S'a'
I28859
sS's'
I18
sS'm'
(lp30241
I1
aI1
aI1
aI1
aI1
aI43
aI35
aI157
aI183
aI835
aI1141
aI3235
aI1383
aI11381
aI4503
aI20435
aI73125
aI196955
asS'd'
I15121
sa(dp30242
S'a'
I28869
sS's'
I18
sS'm'
(lp30243
I1
aI1
aI5
aI7
aI13
aI47
aI9
aI191
aI349
aI587
aI1887
aI3667
aI619
aI9443
aI28781
aI7759
aI6023
aI81595
asS'd'
I15122
sa(dp30244
S'a'
I28893
sS's'
I18
sS'm'
(lp30245
I1
aI3
aI5
aI9
aI31
aI27
aI77
aI47
aI375
aI229
aI989
aI1241
aI4937
aI5881
aI18797
aI21743
aI49947
aI246165
asS'd'
I15123
sa(dp30246
S'a'
I28904
sS's'
I18
sS'm'
(lp30247
I1
aI1
aI5
aI1
aI29
aI23
aI43
aI237
aI293
aI391
aI243
aI3471
aI5205
aI9951
aI29329
aI19873
aI114325
aI19239
asS'd'
I15124
sa(dp30248
S'a'
I28922
sS's'
I18
sS'm'
(lp30249
I1
aI3
aI3
aI5
aI19
aI49
aI23
aI149
aI419
aI23
aI21
aI515
aI3321
aI3157
aI28559
aI8521
aI11119
aI192881
asS'd'
I15125
sa(dp30250
S'a'
I28930
sS's'
I18
sS'm'
(lp30251
I1
aI3
aI1
aI9
aI29
aI17
aI15
aI13
aI171
aI57
aI1849
aI3815
aI7361
aI7723
aI23657
aI60883
aI54953
aI159861
asS'd'
I15126
sa(dp30252
S'a'
I28949
sS's'
I18
sS'm'
(lp30253
I1
aI1
aI3
aI5
aI13
aI57
aI35
aI227
aI143
aI725
aI2023
aI2583
aI2277
aI4721
aI4395
aI61479
aI112487
aI211861
asS'd'
I15127
sa(dp30254
S'a'
I28972
sS's'
I18
sS'm'
(lp30255
I1
aI1
aI5
aI7
aI7
aI25
aI83
aI95
aI281
aI931
aI415
aI1661
aI1543
aI5313
aI13317
aI21203
aI23965
aI60891
asS'd'
I15128
sa(dp30256
S'a'
I29002
sS's'
I18
sS'm'
(lp30257
I1
aI3
aI3
aI1
aI25
aI51
aI65
aI147
aI7
aI521
aI235
aI2165
aI6219
aI14247
aI30621
aI43245
aI8133
aI49481
asS'd'
I15129
sa(dp30258
S'a'
I29009
sS's'
I18
sS'm'
(lp30259
I1
aI1
aI5
aI11
aI13
aI27
aI39
aI51
aI213
aI811
aI151
aI1157
aI7821
aI6481
aI32097
aI12319
aI52005
aI33291
asS'd'
I15130
sa(dp30260
S'a'
I29061
sS's'
I18
sS'm'
(lp30261
I1
aI1
aI1
aI13
aI19
aI15
aI39
aI205
aI481
aI253
aI1643
aI2969
aI3181
aI13995
aI29877
aI1307
aI101721
aI119111
asS'd'
I15131
sa(dp30262
S'a'
I29068
sS's'
I18
sS'm'
(lp30263
I1
aI1
aI1
aI11
aI5
aI63
aI57
aI53
aI187
aI315
aI1521
aI847
aI5955
aI3179
aI21459
aI25937
aI83215
aI181599
asS'd'
I15132
sa(dp30264
S'a'
I29107
sS's'
I18
sS'm'
(lp30265
I1
aI1
aI7
aI13
aI17
aI35
aI113
aI73
aI105
aI497
aI1183
aI3397
aI4443
aI14697
aI29201
aI40737
aI40943
aI3529
asS'd'
I15133
sa(dp30266
S'a'
I29119
sS's'
I18
sS'm'
(lp30267
I1
aI3
aI5
aI5
aI3
aI53
aI101
aI125
aI173
aI137
aI333
aI381
aI1143
aI1165
aI789
aI50013
aI23595
aI50235
asS'd'
I15134
sa(dp30268
S'a'
I29142
sS's'
I18
sS'm'
(lp30269
I1
aI1
aI3
aI1
aI23
aI3
aI21
aI143
aI475
aI337
aI1693
aI2341
aI6509
aI4167
aI21031
aI13887
aI83191
aI85187
asS'd'
I15135
sa(dp30270
S'a'
I29146
sS's'
I18
sS'm'
(lp30271
I1
aI1
aI3
aI1
aI29
aI59
aI39
aI217
aI77
aI943
aI1531
aI383
aI6535
aI2593
aI8601
aI61865
aI26629
aI57313
asS'd'
I15136
sa(dp30272
S'a'
I29152
sS's'
I18
sS'm'
(lp30273
I1
aI3
aI5
aI15
aI17
aI15
aI19
aI31
aI273
aI507
aI1193
aI2501
aI4677
aI13355
aI5287
aI1155
aI102959
aI185219
asS'd'
I15137
sa(dp30274
S'a'
I29200
sS's'
I18
sS'm'
(lp30275
I1
aI1
aI1
aI9
aI3
aI3
aI5
aI111
aI159
aI913
aI331
aI303
aI3673
aI12227
aI5245
aI63749
aI107123
aI26315
asS'd'
I15138
sa(dp30276
S'a'
I29206
sS's'
I18
sS'm'
(lp30277
I1
aI3
aI5
aI11
aI13
aI13
aI115
aI237
aI481
aI793
aI1783
aI1107
aI4811
aI3965
aI29571
aI63237
aI15013
aI176925
asS'd'
I15139
sa(dp30278
S'a'
I29245
sS's'
I18
sS'm'
(lp30279
I1
aI1
aI7
aI1
aI13
aI57
aI85
aI15
aI19
aI889
aI1637
aI1721
aI6299
aI6659
aI5541
aI24365
aI38363
aI67749
asS'd'
I15140
sa(dp30280
S'a'
I29254
sS's'
I18
sS'm'
(lp30281
I1
aI1
aI3
aI1
aI9
aI39
aI15
aI183
aI133
aI821
aI1361
aI2617
aI7197
aI8251
aI12599
aI60549
aI42947
aI72519
asS'd'
I15141
sa(dp30282
S'a'
I29258
sS's'
I18
sS'm'
(lp30283
I1
aI3
aI1
aI5
aI19
aI17
aI69
aI189
aI309
aI33
aI2003
aI569
aI6189
aI7845
aI22351
aI14623
aI35287
aI160511
asS'd'
I15142
sa(dp30284
S'a'
I29260
sS's'
I18
sS'm'
(lp30285
I1
aI1
aI7
aI13
aI25
aI3
aI1
aI203
aI163
aI661
aI513
aI3513
aI741
aI16259
aI29817
aI6059
aI23823
aI51869
asS'd'
I15143
sa(dp30286
S'a'
I29268
sS's'
I18
sS'm'
(lp30287
I1
aI3
aI3
aI1
aI9
aI43
aI59
aI77
aI465
aI223
aI2007
aI2187
aI1499
aI9373
aI10535
aI22207
aI111689
aI108485
asS'd'
I15144
sa(dp30288
S'a'
I29281
sS's'
I18
sS'm'
(lp30289
I1
aI1
aI5
aI15
aI1
aI21
aI87
aI163
aI177
aI751
aI115
aI3981
aI4257
aI5345
aI31211
aI44075
aI16983
aI69783
asS'd'
I15145
sa(dp30290
S'a'
I29306
sS's'
I18
sS'm'
(lp30291
I1
aI1
aI1
aI3
aI29
aI31
aI7
aI41
aI181
aI979
aI1661
aI1403
aI2577
aI983
aI545
aI6205
aI20183
aI44735
asS'd'
I15146
sa(dp30292
S'a'
I29317
sS's'
I18
sS'm'
(lp30293
I1
aI3
aI3
aI15
aI5
aI1
aI85
aI243
aI59
aI161
aI1053
aI803
aI1813
aI13583
aI2559
aI62761
aI105337
aI83209
asS'd'
I15147
sa(dp30294
S'a'
I29324
sS's'
I18
sS'm'
(lp30295
I1
aI3
aI3
aI3
aI5
aI21
aI101
aI61
aI379
aI369
aI1865
aI3289
aI2643
aI951
aI26493
aI17915
aI8185
aI42387
asS'd'
I15148
sa(dp30296
S'a'
I29342
sS's'
I18
sS'm'
(lp30297
I1
aI3
aI5
aI15
aI15
aI13
aI119
aI103
aI141
aI735
aI1317
aI3345
aI2885
aI4145
aI30719
aI965
aI10819
aI90295
asS'd'
I15149
sa(dp30298
S'a'
I29375
sS's'
I18
sS'm'
(lp30299
I1
aI3
aI7
aI13
aI15
aI11
aI19
aI163
aI495
aI369
aI1285
aI609
aI1559
aI9965
aI31123
aI55101
aI76743
aI104435
asS'd'
I15150
sa(dp30300
S'a'
I29384
sS's'
I18
sS'm'
(lp30301
I1
aI1
aI5
aI1
aI25
aI5
aI5
aI139
aI441
aI447
aI353
aI1369
aI959
aI14593
aI30991
aI20651
aI126945
aI2219
asS'd'
I15151
sa(dp30302
S'a'
I29389
sS's'
I18
sS'm'
(lp30303
I1
aI3
aI5
aI9
aI21
aI9
aI113
aI83
aI115
aI15
aI161
aI1559
aI3095
aI1447
aI10253
aI51481
aI114541
aI248375
asS'd'
I15152
sa(dp30304
S'a'
I29411
sS's'
I18
sS'm'
(lp30305
I1
aI3
aI3
aI1
aI31
aI61
aI111
aI69
aI495
aI195
aI1153
aI2605
aI5061
aI15509
aI8253
aI41909
aI126033
aI51173
asS'd'
I15153
sa(dp30306
S'a'
I29413
sS's'
I18
sS'm'
(lp30307
I1
aI1
aI7
aI15
aI15
aI41
aI121
aI75
aI471
aI539
aI341
aI441
aI5357
aI11509
aI32525
aI65477
aI101251
aI164835
asS'd'
I15154
sa(dp30308
S'a'
I29446
sS's'
I18
sS'm'
(lp30309
I1
aI1
aI5
aI7
aI3
aI63
aI13
aI123
aI285
aI499
aI1023
aI3533
aI483
aI13747
aI26515
aI52381
aI9073
aI256319
asS'd'
I15155
sa(dp30310
S'a'
I29457
sS's'
I18
sS'm'
(lp30311
I1
aI3
aI1
aI13
aI29
aI41
aI75
aI43
aI229
aI557
aI1775
aI1933
aI5567
aI11439
aI22045
aI10571
aI96761
aI98559
asS'd'
I15156
sa(dp30312
S'a'
I29458
sS's'
I18
sS'm'
(lp30313
I1
aI1
aI3
aI11
aI19
aI39
aI3
aI93
aI435
aI433
aI2005
aI1561
aI385
aI15865
aI19763
aI44105
aI48107
aI163063
asS'd'
I15157
sa(dp30314
S'a'
I29483
sS's'
I18
sS'm'
(lp30315
I1
aI1
aI7
aI5
aI29
aI37
aI53
aI19
aI335
aI325
aI133
aI2055
aI3029
aI14573
aI30395
aI38599
aI97637
aI203443
asS'd'
I15158
sa(dp30316
S'a'
I29488
sS's'
I18
sS'm'
(lp30317
I1
aI1
aI7
aI15
aI29
aI51
aI7
aI145
aI21
aI955
aI1013
aI579
aI4971
aI4849
aI11691
aI23725
aI71079
aI102641
asS'd'
I15159
sa(dp30318
S'a'
I29494
sS's'
I18
sS'm'
(lp30319
I1
aI1
aI7
aI3
aI9
aI49
aI79
aI187
aI237
aI823
aI1951
aI2947
aI3633
aI9119
aI14393
aI52969
aI44703
aI222389
asS'd'
I15160
sa(dp30320
S'a'
I29506
sS's'
I18
sS'm'
(lp30321
I1
aI3
aI5
aI11
aI13
aI9
aI13
aI245
aI499
aI661
aI1899
aI1313
aI6907
aI12259
aI4577
aI38547
aI79687
aI17555
asS'd'
I15161
sa(dp30322
S'a'
I29553
sS's'
I18
sS'm'
(lp30323
I1
aI3
aI3
aI1
aI5
aI59
aI123
aI197
aI293
aI247
aI687
aI695
aI7493
aI3115
aI28535
aI44335
aI31905
aI81607
asS'd'
I15162
sa(dp30324
S'a'
I29569
sS's'
I18
sS'm'
(lp30325
I1
aI3
aI1
aI3
aI19
aI5
aI45
aI101
aI457
aI395
aI565
aI3155
aI8081
aI4863
aI1199
aI32133
aI73087
aI27025
asS'd'
I15163
sa(dp30326
S'a'
I29587
sS's'
I18
sS'm'
(lp30327
I1
aI3
aI7
aI1
aI31
aI35
aI111
aI95
aI379
aI663
aI731
aI1813
aI4551
aI13105
aI18275
aI19729
aI121971
aI139959
asS'd'
I15164
sa(dp30328
S'a'
I29590
sS's'
I18
sS'm'
(lp30329
I1
aI1
aI7
aI3
aI23
aI47
aI11
aI117
aI323
aI943
aI183
aI2169
aI4625
aI12931
aI1305
aI23345
aI119521
aI67911
asS'd'
I15165
sa(dp30330
S'a'
I29600
sS's'
I18
sS'm'
(lp30331
I1
aI1
aI1
aI13
aI19
aI45
aI37
aI77
aI301
aI741
aI747
aI241
aI5865
aI11141
aI7961
aI10609
aI97833
aI256555
asS'd'
I15166
sa(dp30332
S'a'
I29612
sS's'
I18
sS'm'
(lp30333
I1
aI1
aI1
aI11
aI27
aI21
aI119
aI103
aI277
aI761
aI201
aI2063
aI1043
aI13303
aI6535
aI15553
aI57695
aI124187
asS'd'
I15167
sa(dp30334
S'a'
I29665
sS's'
I18
sS'm'
(lp30335
I1
aI3
aI7
aI3
aI1
aI11
aI79
aI143
aI345
aI237
aI1421
aI193
aI1889
aI2515
aI11729
aI559
aI35227
aI9393
asS'd'
I15168
sa(dp30336
S'a'
I29722
sS's'
I18
sS'm'
(lp30337
I1
aI3
aI1
aI1
aI27
aI27
aI117
aI159
aI183
aI871
aI47
aI989
aI999
aI303
aI30833
aI8229
aI116301
aI199439
asS'd'
I15169
sa(dp30338
S'a'
I29745
sS's'
I18
sS'm'
(lp30339
I1
aI1
aI1
aI1
aI31
aI27
aI41
aI83
aI435
aI409
aI999
aI2275
aI4489
aI1985
aI21455
aI23275
aI125039
aI192979
asS'd'
I15170
sa(dp30340
S'a'
I29746
sS's'
I18
sS'm'
(lp30341
I1
aI3
aI7
aI3
aI19
aI49
aI27
aI185
aI9
aI385
aI191
aI735
aI3439
aI16307
aI21181
aI58749
aI128393
aI140383
asS'd'
I15171
sa(dp30342
S'a'
I29752
sS's'
I18
sS'm'
(lp30343
I1
aI3
aI7
aI3
aI15
aI5
aI65
aI89
aI11
aI915
aI673
aI947
aI3847
aI6833
aI10095
aI34261
aI101645
aI42131
asS'd'
I15172
sa(dp30344
S'a'
I29775
sS's'
I18
sS'm'
(lp30345
I1
aI3
aI5
aI11
aI11
aI25
aI79
aI225
aI495
aI951
aI1033
aI5
aI699
aI9621
aI1791
aI48221
aI59275
aI49875
asS'd'
I15173
sa(dp30346
S'a'
I29778
sS's'
I18
sS'm'
(lp30347
I1
aI3
aI7
aI13
aI29
aI59
aI105
aI101
aI233
aI901
aI863
aI413
aI2087
aI16209
aI445
aI27463
aI61465
aI121795
asS'd'
I15174
sa(dp30348
S'a'
I29800
sS's'
I18
sS'm'
(lp30349
I1
aI1
aI1
aI11
aI5
aI19
aI11
aI51
aI503
aI313
aI195
aI3
aI7249
aI4919
aI11931
aI15569
aI118297
aI115989
asS'd'
I15175
sa(dp30350
S'a'
I29829
sS's'
I18
sS'm'
(lp30351
I1
aI3
aI3
aI13
aI13
aI61
aI63
aI57
aI429
aI487
aI2033
aI847
aI7539
aI1469
aI3197
aI1307
aI124557
aI211999
asS'd'
I15176
sa(dp30352
S'a'
I29830
sS's'
I18
sS'm'
(lp30353
I1
aI3
aI7
aI13
aI25
aI27
aI39
aI103
aI165
aI39
aI1587
aI3103
aI1745
aI12593
aI10779
aI37105
aI29059
aI256739
asS'd'
I15177
sa(dp30354
S'a'
I29877
sS's'
I18
sS'm'
(lp30355
I1
aI3
aI3
aI9
aI25
aI51
aI105
aI109
aI99
aI267
aI623
aI1351
aI3837
aI793
aI28609
aI52199
aI15621
aI77873
asS'd'
I15178
sa(dp30356
S'a'
I29881
sS's'
I18
sS'm'
(lp30357
I1
aI1
aI3
aI7
aI29
aI61
aI45
aI237
aI431
aI791
aI91
aI1259
aI8071
aI11103
aI27257
aI10153
aI18639
aI248949
asS'd'
I15179
sa(dp30358
S'a'
I29884
sS's'
I18
sS'm'
(lp30359
I1
aI1
aI1
aI9
aI15
aI47
aI113
aI189
aI291
aI837
aI1317
aI2263
aI7183
aI6669
aI17241
aI35275
aI9087
aI241577
asS'd'
I15180
sa(dp30360
S'a'
I29895
sS's'
I18
sS'm'
(lp30361
I1
aI1
aI3
aI1
aI15
aI59
aI85
aI21
aI69
aI569
aI1473
aI2735
aI713
aI3817
aI14677
aI26897
aI75291
aI251255
asS'd'
I15181
sa(dp30362
S'a'
I29899
sS's'
I18
sS'm'
(lp30363
I1
aI1
aI1
aI7
aI17
aI21
aI105
aI77
aI367
aI905
aI513
aI1807
aI5571
aI14627
aI10349
aI47821
aI34395
aI51143
asS'd'
I15182
sa(dp30364
S'a'
I29916
sS's'
I18
sS'm'
(lp30365
I1
aI3
aI7
aI13
aI27
aI19
aI123
aI145
aI371
aI857
aI1699
aI2231
aI373
aI781
aI28713
aI21441
aI64059
aI10689
asS'd'
I15183
sa(dp30366
S'a'
I29923
sS's'
I18
sS'm'
(lp30367
I1
aI1
aI1
aI7
aI19
aI57
aI81
aI223
aI87
aI315
aI1253
aI421
aI1371
aI1547
aI1411
aI6809
aI23889
aI213385
asS'd'
I15184
sa(dp30368
S'a'
I29935
sS's'
I18
sS'm'
(lp30369
I1
aI3
aI1
aI5
aI23
aI57
aI89
aI15
aI227
aI965
aI1247
aI3861
aI7723
aI15621
aI7151
aI53961
aI47167
aI73679
asS'd'
I15185
sa(dp30370
S'a'
I29955
sS's'
I18
sS'm'
(lp30371
I1
aI3
aI7
aI15
aI31
aI21
aI9
aI79
aI87
aI561
aI1001
aI3395
aI2095
aI15381
aI30725
aI48111
aI68031
aI121687
asS'd'
I15186
sa(dp30372
S'a'
I29962
sS's'
I18
sS'm'
(lp30373
I1
aI3
aI1
aI9
aI15
aI29
aI83
aI87
aI377
aI331
aI2035
aI93
aI2319
aI3637
aI4809
aI40091
aI93141
aI39881
asS'd'
I15187
sa(dp30374
S'a'
I29969
sS's'
I18
sS'm'
(lp30375
I1
aI3
aI3
aI11
aI21
aI39
aI27
aI159
aI161
aI439
aI1417
aI595
aI4873
aI15703
aI32743
aI56603
aI35881
aI160727
asS'd'
I15188
sa(dp30376
S'a'
I29997
sS's'
I18
sS'm'
(lp30377
I1
aI1
aI3
aI5
aI21
aI37
aI55
aI159
aI497
aI425
aI469
aI1185
aI5015
aI7045
aI7179
aI65325
aI97167
aI75723
asS'd'
I15189
sa(dp30378
S'a'
I30032
sS's'
I18
sS'm'
(lp30379
I1
aI3
aI1
aI7
aI31
aI21
aI125
aI223
aI479
aI765
aI1115
aI33
aI2765
aI12781
aI22923
aI36051
aI103749
aI33703
asS'd'
I15190
sa(dp30380
S'a'
I30041
sS's'
I18
sS'm'
(lp30381
I1
aI1
aI7
aI1
aI9
aI3
aI29
aI125
aI337
aI973
aI253
aI3179
aI3269
aI8801
aI19369
aI20693
aI17331
aI190295
asS'd'
I15191
sa(dp30382
S'a'
I30048
sS's'
I18
sS'm'
(lp30383
I1
aI1
aI3
aI3
aI13
aI5
aI115
aI31
aI481
aI45
aI855
aI81
aI3663
aI10443
aI13853
aI29847
aI99471
aI249943
asS'd'
I15192
sa(dp30384
S'a'
I30060
sS's'
I18
sS'm'
(lp30385
I1
aI1
aI3
aI7
aI3
aI1
aI47
aI31
aI169
aI625
aI201
aI2257
aI4617
aI1633
aI26681
aI53793
aI78257
aI8955
asS'd'
I15193
sa(dp30386
S'a'
I30063
sS's'
I18
sS'm'
(lp30387
I1
aI1
aI5
aI1
aI23
aI3
aI95
aI89
aI429
aI559
aI119
aI2619
aI1235
aI7609
aI21905
aI45495
aI19461
aI189091
asS'd'
I15194
sa(dp30388
S'a'
I30105
sS's'
I18
sS'm'
(lp30389
I1
aI3
aI5
aI5
aI11
aI33
aI123
aI45
aI89
aI899
aI1607
aI3717
aI6745
aI15199
aI22955
aI15891
aI50411
aI148201
asS'd'
I15195
sa(dp30390
S'a'
I30122
sS's'
I18
sS'm'
(lp30391
I1
aI3
aI5
aI9
aI19
aI23
aI87
aI21
aI39
aI117
aI603
aI823
aI3015
aI14853
aI4341
aI49029
aI97183
aI218713
asS'd'
I15196
sa(dp30392
S'a'
I30141
sS's'
I18
sS'm'
(lp30393
I1
aI1
aI3
aI7
aI11
aI55
aI31
aI255
aI399
aI861
aI745
aI1013
aI4583
aI15871
aI32453
aI25357
aI90645
aI100835
asS'd'
I15197
sa(dp30394
S'a'
I30142
sS's'
I18
sS'm'
(lp30395
I1
aI1
aI1
aI11
aI13
aI27
aI57
aI233
aI45
aI339
aI305
aI3689
aI5273
aI11801
aI29109
aI44139
aI83171
aI250559
asS'd'
I15198
sa(dp30396
S'a'
I30153
sS's'
I18
sS'm'
(lp30397
I1
aI3
aI7
aI13
aI19
aI1
aI29
aI113
aI207
aI313
aI1465
aI3563
aI2535
aI3307
aI14921
aI1923
aI31429
aI59815
asS'd'
I15199
sa(dp30398
S'a'
I30171
sS's'
I18
sS'm'
(lp30399
I1
aI3
aI1
aI5
aI29
aI25
aI59
aI95
aI177
aI795
aI353
aI3973
aI8029
aI1687
aI5045
aI16157
aI30361
aI218479
asS'd'
I15200
sa(dp30400
S'a'
I30192
sS's'
I18
sS'm'
(lp30401
I1
aI3
aI5
aI3
aI5
aI27
aI109
aI239
aI121
aI347
aI93
aI1645
aI3293
aI13181
aI23793
aI42935
aI98659
aI85385
asS'd'
I15201
sa(dp30402
S'a'
I30197
sS's'
I18
sS'm'
(lp30403
I1
aI1
aI7
aI9
aI9
aI37
aI127
aI211
aI77
aI557
aI177
aI2465
aI7895
aI5523
aI26665
aI23463
aI71715
aI126673
asS'd'
I15202
sa(dp30404
S'a'
I30223
sS's'
I18
sS'm'
(lp30405
I1
aI3
aI7
aI1
aI27
aI55
aI125
aI85
aI47
aI739
aI1513
aI3763
aI5335
aI3135
aI11913
aI22405
aI90785
aI88845
asS'd'
I15203
sa(dp30406
S'a'
I30235
sS's'
I18
sS'm'
(lp30407
I1
aI3
aI3
aI9
aI29
aI21
aI123
aI211
aI491
aI83
aI697
aI929
aI3507
aI7139
aI30569
aI16365
aI122657
aI77523
asS'd'
I15204
sa(dp30408
S'a'
I30241
sS's'
I18
sS'm'
(lp30409
I1
aI3
aI5
aI5
aI27
aI21
aI41
aI129
aI67
aI503
aI877
aI1893
aI6055
aI12709
aI24613
aI43831
aI124035
aI62631
asS'd'
I15205
sa(dp30410
S'a'
I30256
sS's'
I18
sS'm'
(lp30411
I1
aI3
aI3
aI1
aI21
aI15
aI59
aI185
aI405
aI487
aI627
aI3737
aI345
aI14751
aI2947
aI15815
aI55047
aI207137
asS'd'
I15206
sa(dp30412
S'a'
I30291
sS's'
I18
sS'm'
(lp30413
I1
aI1
aI1
aI9
aI15
aI29
aI19
aI155
aI101
aI405
aI597
aI329
aI2977
aI4333
aI5465
aI43863
aI6009
aI229481
asS'd'
I15207
sa(dp30414
S'a'
I30293
sS's'
I18
sS'm'
(lp30415
I1
aI1
aI1
aI11
aI7
aI27
aI93
aI207
aI43
aI599
aI1899
aI3979
aI4219
aI10199
aI2901
aI34261
aI19435
aI58317
asS'd'
I15208
sa(dp30416
S'a'
I30300
sS's'
I18
sS'm'
(lp30417
I1
aI1
aI7
aI15
aI9
aI47
aI33
aI209
aI235
aI655
aI253
aI3507
aI3355
aI1685
aI7045
aI58907
aI41791
aI175835
asS'd'
I15209
sa(dp30418
S'a'
I30331
sS's'
I18
sS'm'
(lp30419
I1
aI3
aI3
aI11
aI11
aI43
aI21
aI255
aI45
aI831
aI1051
aI1271
aI7945
aI9793
aI11125
aI17709
aI118267
aI169981
asS'd'
I15210
sa(dp30420
S'a'
I30349
sS's'
I18
sS'm'
(lp30421
I1
aI3
aI7
aI13
aI27
aI37
aI45
aI221
aI243
aI37
aI1493
aI2773
aI6655
aI7451
aI22609
aI56559
aI12063
aI221143
asS'd'
I15211
sa(dp30422
S'a'
I30361
sS's'
I18
sS'm'
(lp30423
I1
aI3
aI1
aI1
aI9
aI15
aI97
aI61
aI241
aI825
aI735
aI3953
aI5331
aI16373
aI19403
aI28933
aI31881
aI111545
asS'd'
I15212
sa(dp30424
S'a'
I30367
sS's'
I18
sS'm'
(lp30425
I1
aI1
aI1
aI15
aI29
aI1
aI127
aI111
aI329
aI741
aI1589
aI1653
aI5949
aI8703
aI27617
aI65285
aI122167
aI11895
asS'd'
I15213
sa(dp30426
S'a'
I30392
sS's'
I18
sS'm'
(lp30427
I1
aI1
aI3
aI3
aI17
aI31
aI91
aI197
aI421
aI865
aI1901
aI897
aI6917
aI15943
aI12823
aI15325
aI17011
aI110783
asS'd'
I15214
sa(dp30428
S'a'
I30405
sS's'
I18
sS'm'
(lp30429
I1
aI1
aI3
aI15
aI21
aI57
aI29
aI179
aI503
aI929
aI1513
aI205
aI6083
aI4015
aI32517
aI26921
aI54229
aI147789
asS'd'
I15215
sa(dp30430
S'a'
I30410
sS's'
I18
sS'm'
(lp30431
I1
aI1
aI7
aI15
aI27
aI21
aI95
aI59
aI193
aI97
aI1235
aI819
aI4435
aI371
aI627
aI24673
aI1775
aI261041
asS'd'
I15216
sa(dp30432
S'a'
I30415
sS's'
I18
sS'm'
(lp30433
I1
aI1
aI7
aI15
aI3
aI37
aI99
aI85
aI505
aI1011
aI19
aI1241
aI5299
aI15661
aI27323
aI44625
aI12683
aI225687
asS'd'
I15217
sa(dp30434
S'a'
I30439
sS's'
I18
sS'm'
(lp30435
I1
aI1
aI7
aI3
aI25
aI37
aI111
aI121
aI217
aI659
aI365
aI2627
aI5499
aI12911
aI951
aI54317
aI51927
aI235327
asS'd'
I15218
sa(dp30436
S'a'
I30446
sS's'
I18
sS'm'
(lp30437
I1
aI1
aI3
aI9
aI29
aI31
aI99
aI195
aI427
aI735
aI1817
aI3675
aI4269
aI1579
aI12593
aI39285
aI74909
aI230613
asS'd'
I15219
sa(dp30438
S'a'
I30501
sS's'
I18
sS'm'
(lp30439
I1
aI3
aI1
aI13
aI7
aI21
aI37
aI101
aI111
aI931
aI1581
aI465
aI4753
aI15607
aI14515
aI29769
aI107711
aI32703
asS'd'
I15220
sa(dp30440
S'a'
I30526
sS's'
I18
sS'm'
(lp30441
I1
aI1
aI7
aI15
aI25
aI57
aI117
aI119
aI309
aI345
aI491
aI3647
aI2933
aI5409
aI15431
aI43731
aI25537
aI17269
asS'd'
I15221
sa(dp30442
S'a'
I30531
sS's'
I18
sS'm'
(lp30443
I1
aI1
aI5
aI3
aI15
aI9
aI83
aI221
aI501
aI675
aI1441
aI129
aI213
aI5587
aI22361
aI16739
aI118845
aI192835
asS'd'
I15222
sa(dp30444
S'a'
I30537
sS's'
I18
sS'm'
(lp30445
I1
aI3
aI7
aI13
aI1
aI31
aI75
aI23
aI13
aI447
aI687
aI2711
aI7577
aI8275
aI5051
aI61835
aI22159
aI56477
asS'd'
I15223
sa(dp30446
S'a'
I30546
sS's'
I18
sS'm'
(lp30447
I1
aI1
aI3
aI1
aI23
aI19
aI111
aI45
aI395
aI841
aI1665
aI21
aI7435
aI12457
aI11065
aI20007
aI48785
aI15115
asS'd'
I15224
sa(dp30448
S'a'
I30568
sS's'
I18
sS'm'
(lp30449
I1
aI3
aI7
aI7
aI21
aI3
aI117
aI35
aI117
aI433
aI561
aI3045
aI2169
aI3255
aI18015
aI41093
aI99699
aI3479
asS'd'
I15225
sa(dp30450
S'a'
I30581
sS's'
I18
sS'm'
(lp30451
I1
aI1
aI3
aI7
aI9
aI59
aI65
aI143
aI315
aI63
aI29
aI3817
aI1259
aI7119
aI20847
aI44407
aI80015
aI37851
asS'd'
I15226
sa(dp30452
S'a'
I30588
sS's'
I18
sS'm'
(lp30453
I1
aI3
aI7
aI13
aI13
aI33
aI85
aI75
aI39
aI163
aI1759
aI1197
aI5971
aI8815
aI8745
aI45625
aI121873
aI246197
asS'd'
I15227
sa(dp30454
S'a'
I30598
sS's'
I18
sS'm'
(lp30455
I1
aI1
aI1
aI13
aI31
aI41
aI61
aI7
aI145
aI113
aI943
aI3757
aI2141
aI3523
aI22351
aI14143
aI107683
aI105579
asS'd'
I15228
sa(dp30456
S'a'
I30607
sS's'
I18
sS'm'
(lp30457
I1
aI1
aI5
aI5
aI23
aI27
aI75
aI77
aI25
aI901
aI1295
aI3091
aI981
aI10109
aI12649
aI15123
aI102433
aI145557
asS'd'
I15229
sa(dp30458
S'a'
I30609
sS's'
I18
sS'm'
(lp30459
I1
aI1
aI7
aI5
aI9
aI11
aI19
aI229
aI301
aI835
aI1577
aI3365
aI425
aI2271
aI15647
aI11685
aI57315
aI131043
asS'd'
I15230
sa(dp30460
S'a'
I30616
sS's'
I18
sS'm'
(lp30461
I1
aI3
aI1
aI13
aI31
aI3
aI113
aI7
aI473
aI395
aI1979
aI61
aI4205
aI2031
aI28007
aI34789
aI54463
aI94741
asS'd'
I15231
sa(dp30462
S'a'
I30626
sS's'
I18
sS'm'
(lp30463
I1
aI1
aI3
aI5
aI7
aI13
aI35
aI151
aI461
aI621
aI185
aI2645
aI907
aI9151
aI25953
aI26363
aI105531
aI235555
asS'd'
I15232
sa(dp30464
S'a'
I30628
sS's'
I18
sS'm'
(lp30465
I1
aI3
aI3
aI13
aI21
aI5
aI43
aI183
aI149
aI607
aI509
aI777
aI4089
aI16365
aI32201
aI60431
aI58773
aI92719
asS'd'
I15233
sa(dp30466
S'a'
I30638
sS's'
I18
sS'm'
(lp30467
I1
aI3
aI3
aI11
aI25
aI53
aI103
aI203
aI269
aI1017
aI77
aI3537
aI4839
aI15991
aI29223
aI57397
aI122735
aI67299
asS'd'
I15234
sa(dp30468
S'a'
I30658
sS's'
I18
sS'm'
(lp30469
I1
aI3
aI3
aI1
aI29
aI45
aI85
aI171
aI307
aI455
aI1399
aI2367
aI5991
aI5751
aI27957
aI36649
aI9251
aI38581
asS'd'
I15235
sa(dp30470
S'a'
I30663
sS's'
I18
sS'm'
(lp30471
I1
aI1
aI5
aI13
aI29
aI15
aI127
aI1
aI175
aI921
aI671
aI2469
aI3137
aI3679
aI32521
aI5321
aI92505
aI45901
asS'd'
I15236
sa(dp30472
S'a'
I30675
sS's'
I18
sS'm'
(lp30473
I1
aI3
aI3
aI11
aI23
aI23
aI113
aI255
aI443
aI609
aI1085
aI133
aI5369
aI885
aI17043
aI20961
aI36137
aI260457
asS'd'
I15237
sa(dp30474
S'a'
I30698
sS's'
I18
sS'm'
(lp30475
I1
aI3
aI3
aI9
aI13
aI55
aI117
aI19
aI111
aI323
aI275
aI495
aI6679
aI2217
aI12015
aI3053
aI32745
aI189413
asS'd'
I15238
sa(dp30476
S'a'
I30708
sS's'
I18
sS'm'
(lp30477
I1
aI3
aI5
aI13
aI31
aI43
aI37
aI225
aI67
aI755
aI1427
aI3967
aI6497
aI9987
aI28145
aI50583
aI59457
aI213217
asS'd'
I15239
sa(dp30478
S'a'
I30712
sS's'
I18
sS'm'
(lp30479
I1
aI3
aI1
aI15
aI1
aI5
aI121
aI249
aI293
aI695
aI1697
aI313
aI61
aI4037
aI11757
aI53739
aI5693
aI106225
asS'd'
I15240
sa(dp30480
S'a'
I30727
sS's'
I18
sS'm'
(lp30481
I1
aI1
aI7
aI1
aI23
aI9
aI111
aI211
aI303
aI147
aI1291
aI3807
aI1969
aI4115
aI7473
aI50077
aI60745
aI41605
asS'd'
I15241
sa(dp30482
S'a'
I30736
sS's'
I18
sS'm'
(lp30483
I1
aI3
aI3
aI15
aI21
aI51
aI63
aI171
aI197
aI403
aI1327
aI1851
aI6991
aI9069
aI19221
aI45765
aI55489
aI34853
asS'd'
I15242
sa(dp30484
S'a'
I30748
sS's'
I18
sS'm'
(lp30485
I1
aI1
aI5
aI3
aI5
aI53
aI87
aI241
aI255
aI309
aI1319
aI3727
aI3189
aI10887
aI13401
aI32371
aI24479
aI170571
asS'd'
I15243
sa(dp30486
S'a'
I30758
sS's'
I18
sS'm'
(lp30487
I1
aI1
aI7
aI11
aI5
aI5
aI59
aI177
aI317
aI835
aI527
aI165
aI2137
aI9597
aI30181
aI1779
aI75407
aI185827
asS'd'
I15244
sa(dp30488
S'a'
I30775
sS's'
I18
sS'm'
(lp30489
I1
aI3
aI3
aI7
aI25
aI15
aI15
aI183
aI235
aI955
aI27
aI2223
aI5587
aI11301
aI17653
aI56697
aI70787
aI198347
asS'd'
I15245
sa(dp30490
S'a'
I30793
sS's'
I18
sS'm'
(lp30491
I1
aI3
aI5
aI5
aI11
aI63
aI123
aI41
aI169
aI975
aI1709
aI2965
aI7491
aI4183
aI15217
aI41343
aI36139
aI9649
asS'd'
I15246
sa(dp30492
S'a'
I30802
sS's'
I18
sS'm'
(lp30493
I1
aI1
aI3
aI7
aI5
aI37
aI87
aI247
aI379
aI603
aI781
aI463
aI8063
aI13681
aI32005
aI43485
aI107401
aI42303
asS'd'
I15247
sa(dp30494
S'a'
I30820
sS's'
I18
sS'm'
(lp30495
I1
aI3
aI5
aI9
aI19
aI53
aI61
aI219
aI217
aI361
aI769
aI1687
aI5643
aI3145
aI12885
aI40303
aI86377
aI255051
asS'd'
I15248
sa(dp30496
S'a'
I30823
sS's'
I18
sS'm'
(lp30497
I1
aI1
aI7
aI11
aI15
aI49
aI127
aI99
aI127
aI515
aI647
aI1725
aI1605
aI2357
aI2069
aI31803
aI14179
aI180367
asS'd'
I15249
sa(dp30498
S'a'
I30827
sS's'
I18
sS'm'
(lp30499
I1
aI3
aI5
aI15
aI9
aI7
aI41
aI3
aI49
aI485
aI1471
aI207
aI6477
aI4463
aI12255
aI53481
aI110785
aI909
asS'd'
I15250
sa(dp30500
S'a'
I30848
sS's'
I18
sS'm'
(lp30501
I1
aI1
aI5
aI9
aI23
aI51
aI107
aI227
aI205
aI987
aI1525
aI2739
aI6761
aI12343
aI32311
aI12523
aI93351
aI29663
asS'd'
I15251
sa(dp30502
S'a'
I30893
sS's'
I18
sS'm'
(lp30503
I1
aI3
aI5
aI11
aI21
aI19
aI53
aI169
aI197
aI21
aI825
aI4029
aI6733
aI8943
aI13475
aI60677
aI31001
aI242291
asS'd'
I15252
sa(dp30504
S'a'
I30906
sS's'
I18
sS'm'
(lp30505
I1
aI3
aI7
aI13
aI9
aI27
aI87
aI37
aI265
aI877
aI735
aI2249
aI4801
aI2365
aI16923
aI40451
aI29693
aI222483
asS'd'
I15253
sa(dp30506
S'a'
I30950
sS's'
I18
sS'm'
(lp30507
I1
aI1
aI3
aI9
aI31
aI61
aI71
aI103
aI215
aI421
aI193
aI3451
aI6181
aI4271
aI30875
aI59573
aI80773
aI100369
asS'd'
I15254
sa(dp30508
S'a'
I30956
sS's'
I18
sS'm'
(lp30509
I1
aI1
aI3
aI13
aI17
aI55
aI73
aI191
aI233
aI821
aI961
aI1637
aI2393
aI3453
aI25959
aI5069
aI114585
aI186001
asS'd'
I15255
sa(dp30510
S'a'
I30996
sS's'
I18
sS'm'
(lp30511
I1
aI1
aI7
aI5
aI15
aI27
aI39
aI23
aI69
aI811
aI709
aI2349
aI6011
aI803
aI12497
aI7387
aI62023
aI247875
asS'd'
I15256
sa(dp30512
S'a'
I31015
sS's'
I18
sS'm'
(lp30513
I1
aI1
aI5
aI13
aI29
aI55
aI51
aI41
aI121
aI599
aI1633
aI3813
aI1913
aI3803
aI26097
aI53799
aI30997
aI261965
asS'd'
I15257
sa(dp30514
S'a'
I31016
sS's'
I18
sS'm'
(lp30515
I1
aI1
aI1
aI11
aI31
aI25
aI19
aI195
aI87
aI657
aI1005
aI3853
aI61
aI6585
aI24665
aI38283
aI5495
aI257201
asS'd'
I15258
sa(dp30516
S'a'
I31034
sS's'
I18
sS'm'
(lp30517
I1
aI1
aI1
aI11
aI23
aI43
aI91
aI121
aI49
aI491
aI1443
aI1873
aI7689
aI15957
aI30463
aI64079
aI100003
aI325
asS'd'
I15259
sa(dp30518
S'a'
I31099
sS's'
I18
sS'm'
(lp30519
I1
aI1
aI7
aI1
aI25
aI51
aI27
aI105
aI119
aI233
aI513
aI3403
aI2647
aI2847
aI12687
aI15619
aI71225
aI243759
asS'd'
I15260
sa(dp30520
S'a'
I31123
sS's'
I18
sS'm'
(lp30521
I1
aI1
aI7
aI3
aI17
aI49
aI19
aI123
aI307
aI463
aI1619
aI1853
aI7019
aI2605
aI17351
aI7971
aI20675
aI235929
asS'd'
I15261
sa(dp30522
S'a'
I31145
sS's'
I18
sS'm'
(lp30523
I1
aI1
aI5
aI3
aI17
aI29
aI71
aI215
aI365
aI955
aI1631
aI3549
aI1379
aI13881
aI25409
aI55703
aI29863
aI135401
asS'd'
I15262
sa(dp30524
S'a'
I31166
sS's'
I18
sS'm'
(lp30525
I1
aI3
aI7
aI15
aI3
aI27
aI107
aI149
aI65
aI681
aI505
aI3957
aI6697
aI11203
aI9321
aI63323
aI98587
aI199241
asS'd'
I15263
sa(dp30526
S'a'
I31168
sS's'
I18
sS'm'
(lp30527
I1
aI1
aI1
aI13
aI27
aI41
aI59
aI223
aI431
aI339
aI1805
aI899
aI639
aI6559
aI13233
aI4773
aI37415
aI110477
asS'd'
I15264
sa(dp30528
S'a'
I31174
sS's'
I18
sS'm'
(lp30529
I1
aI3
aI7
aI9
aI1
aI59
aI79
aI161
aI311
aI905
aI1755
aI3915
aI6259
aI8955
aI14187
aI11331
aI86185
aI209805
asS'd'
I15265
sa(dp30530
S'a'
I31216
sS's'
I18
sS'm'
(lp30531
I1
aI1
aI7
aI9
aI5
aI27
aI75
aI93
aI285
aI89
aI891
aI3341
aI1157
aI11219
aI8883
aI8093
aI68949
aI189643
asS'd'
I15266
sa(dp30532
S'a'
I31219
sS's'
I18
sS'm'
(lp30533
I1
aI3
aI1
aI5
aI1
aI43
aI97
aI71
aI67
aI605
aI739
aI1641
aI4415
aI4487
aI13437
aI12755
aI121595
aI55761
asS'd'
I15267
sa(dp30534
S'a'
I31238
sS's'
I18
sS'm'
(lp30535
I1
aI1
aI1
aI3
aI1
aI13
aI61
aI77
aI297
aI507
aI1527
aI3585
aI4515
aI13913
aI7679
aI28461
aI61807
aI196517
asS'd'
I15268
sa(dp30536
S'a'
I31256
sS's'
I18
sS'm'
(lp30537
I1
aI1
aI1
aI3
aI21
aI23
aI101
aI135
aI59
aI485
aI1601
aI3713
aI7409
aI349
aI16543
aI18897
aI97253
aI149055
asS'd'
I15269
sa(dp30538
S'a'
I31304
sS's'
I18
sS'm'
(lp30539
I1
aI1
aI1
aI13
aI5
aI37
aI15
aI37
aI109
aI1005
aI363
aI1925
aI2701
aI13169
aI17027
aI58453
aI27667
aI234027
asS'd'
I15270
sa(dp30540
S'a'
I31315
sS's'
I18
sS'm'
(lp30541
I1
aI1
aI7
aI1
aI1
aI41
aI67
aI231
aI143
aI951
aI2023
aI3465
aI1717
aI645
aI17353
aI9037
aI129127
aI199467
asS'd'
I15271
sa(dp30542
S'a'
I31324
sS's'
I18
sS'm'
(lp30543
I1
aI1
aI1
aI11
aI27
aI29
aI65
aI139
aI425
aI947
aI141
aI2601
aI7339
aI4451
aI25065
aI62691
aI62355
aI91819
asS'd'
I15272
sa(dp30544
S'a'
I31364
sS's'
I18
sS'm'
(lp30545
I1
aI3
aI7
aI15
aI29
aI29
aI93
aI25
aI139
aI267
aI1319
aI3839
aI7295
aI11855
aI17775
aI30199
aI44127
aI150875
asS'd'
I15273
sa(dp30546
S'a'
I31379
sS's'
I18
sS'm'
(lp30547
I1
aI3
aI5
aI1
aI3
aI55
aI23
aI191
aI199
aI583
aI1167
aI1357
aI6477
aI11827
aI15581
aI56541
aI16603
aI120139
asS'd'
I15274
sa(dp30548
S'a'
I31382
sS's'
I18
sS'm'
(lp30549
I1
aI1
aI1
aI3
aI5
aI47
aI103
aI211
aI443
aI491
aI1043
aI4001
aI121
aI1637
aI5685
aI42675
aI13009
aI22979
asS'd'
I15275
sa(dp30550
S'a'
I31412
sS's'
I18
sS'm'
(lp30551
I1
aI1
aI1
aI9
aI21
aI7
aI77
aI17
aI303
aI955
aI51
aI2389
aI3573
aI8817
aI28053
aI40269
aI35457
aI211023
asS'd'
I15276
sa(dp30552
S'a'
I31451
sS's'
I18
sS'm'
(lp30553
I1
aI3
aI5
aI15
aI3
aI39
aI17
aI75
aI223
aI37
aI1231
aI2127
aI3575
aI9085
aI10715
aI41871
aI103703
aI154181
asS'd'
I15277
sa(dp30554
S'a'
I31487
sS's'
I18
sS'm'
(lp30555
I1
aI3
aI3
aI15
aI25
aI31
aI31
aI223
aI473
aI267
aI1519
aI3205
aI7029
aI10753
aI24757
aI28187
aI117921
aI26783
asS'd'
I15278
sa(dp30556
S'a'
I31509
sS's'
I18
sS'm'
(lp30557
I1
aI3
aI7
aI1
aI5
aI21
aI105
aI191
aI55
aI115
aI1813
aI3701
aI1673
aI4199
aI2441
aI19737
aI46913
aI208725
asS'd'
I15279
sa(dp30558
S'a'
I31510
sS's'
I18
sS'm'
(lp30559
I1
aI1
aI5
aI5
aI19
aI63
aI89
aI141
aI143
aI783
aI545
aI883
aI2979
aI9219
aI24983
aI41793
aI88441
aI207095
asS'd'
I15280
sa(dp30560
S'a'
I31520
sS's'
I18
sS'm'
(lp30561
I1
aI1
aI1
aI9
aI21
aI21
aI93
aI161
aI93
aI733
aI417
aI3133
aI8155
aI12415
aI16343
aI11727
aI82877
aI94469
asS'd'
I15281
sa(dp30562
S'a'
I31535
sS's'
I18
sS'm'
(lp30563
I1
aI1
aI5
aI15
aI5
aI39
aI33
aI203
aI213
aI731
aI1849
aI1675
aI6029
aI2743
aI1529
aI16345
aI13955
aI54929
asS'd'
I15282
sa(dp30564
S'a'
I31564
sS's'
I18
sS'm'
(lp30565
I1
aI1
aI7
aI7
aI23
aI47
aI121
aI211
aI271
aI737
aI1015
aI1021
aI5641
aI12659
aI27545
aI52363
aI104761
aI150143
asS'd'
I15283
sa(dp30566
S'a'
I31585
sS's'
I18
sS'm'
(lp30567
I1
aI3
aI1
aI11
aI11
aI51
aI79
aI141
aI255
aI1007
aI481
aI3221
aI7741
aI6861
aI24943
aI63091
aI46741
aI33359
asS'd'
I15284
sa(dp30568
S'a'
I31586
sS's'
I18
sS'm'
(lp30569
I1
aI1
aI3
aI3
aI27
aI47
aI85
aI27
aI423
aI811
aI75
aI3911
aI1951
aI10821
aI7487
aI18971
aI83355
aI197479
asS'd'
I15285
sa(dp30570
S'a'
I31600
sS's'
I18
sS'm'
(lp30571
I1
aI1
aI5
aI5
aI1
aI63
aI125
aI251
aI457
aI795
aI557
aI217
aI2335
aI5659
aI18375
aI52183
aI9789
aI31417
asS'd'
I15286
sa(dp30572
S'a'
I31643
sS's'
I18
sS'm'
(lp30573
I1
aI3
aI3
aI3
aI19
aI61
aI41
aI129
aI345
aI361
aI187
aI3881
aI43
aI7197
aI7673
aI25011
aI115155
aI16375
asS'd'
I15287
sa(dp30574
S'a'
I31646
sS's'
I18
sS'm'
(lp30575
I1
aI3
aI3
aI7
aI13
aI7
aI55
aI91
aI153
aI341
aI2003
aI2013
aI6891
aI2411
aI14825
aI39555
aI50267
aI61405
asS'd'
I15288
sa(dp30576
S'a'
I31649
sS's'
I18
sS'm'
(lp30577
I1
aI3
aI5
aI7
aI13
aI57
aI21
aI91
aI331
aI615
aI1297
aI2881
aI2011
aI1907
aI16489
aI43061
aI75731
aI76675
asS'd'
I15289
sa(dp30578
S'a'
I31650
sS's'
I18
sS'm'
(lp30579
I1
aI1
aI5
aI5
aI25
aI15
aI77
aI175
aI101
aI885
aI835
aI529
aI2787
aI547
aI20191
aI50457
aI58557
aI61807
asS'd'
I15290
sa(dp30580
S'a'
I31674
sS's'
I18
sS'm'
(lp30581
I1
aI3
aI7
aI1
aI13
aI47
aI101
aI117
aI179
aI245
aI861
aI611
aI4377
aI5257
aI12807
aI26667
aI19889
aI112485
asS'd'
I15291
sa(dp30582
S'a'
I31679
sS's'
I18
sS'm'
(lp30583
I1
aI3
aI7
aI1
aI27
aI3
aI23
aI109
aI197
aI187
aI963
aI1827
aI5741
aI11921
aI6359
aI3989
aI108939
aI5761
asS'd'
I15292
sa(dp30584
S'a'
I31684
sS's'
I18
sS'm'
(lp30585
I1
aI1
aI7
aI5
aI27
aI7
aI119
aI159
aI53
aI969
aI557
aI597
aI7821
aI7121
aI17341
aI11233
aI10811
aI23969
asS'd'
I15293
sa(dp30586
S'a'
I31688
sS's'
I18
sS'm'
(lp30587
I1
aI1
aI3
aI13
aI23
aI55
aI75
aI131
aI339
aI917
aI317
aI2645
aI5973
aI9939
aI26375
aI29261
aI80883
aI229897
asS'd'
I15294
sa(dp30588
S'a'
I31706
sS's'
I18
sS'm'
(lp30589
I1
aI1
aI5
aI13
aI3
aI63
aI41
aI191
aI315
aI191
aI649
aI2119
aI317
aI14699
aI4097
aI19557
aI97015
aI124557
asS'd'
I15295
sa(dp30590
S'a'
I31735
sS's'
I18
sS'm'
(lp30591
I1
aI3
aI7
aI15
aI13
aI29
aI29
aI43
aI47
aI37
aI729
aI185
aI4477
aI15091
aI13991
aI18701
aI56785
aI218823
asS'd'
I15296
sa(dp30592
S'a'
I31754
sS's'
I18
sS'm'
(lp30593
I1
aI3
aI7
aI15
aI31
aI39
aI17
aI133
aI469
aI509
aI995
aI1683
aI391
aI1775
aI15431
aI63489
aI7405
aI122125
asS'd'
I15297
sa(dp30594
S'a'
I31761
sS's'
I18
sS'm'
(lp30595
I1
aI3
aI7
aI5
aI15
aI63
aI5
aI235
aI193
aI411
aI1493
aI3967
aI3279
aI6421
aI13359
aI20949
aI68097
aI69469
asS'd'
I15298
sa(dp30596
S'a'
I31774
sS's'
I18
sS'm'
(lp30597
I1
aI3
aI3
aI13
aI7
aI37
aI7
aI207
aI399
aI553
aI1629
aI1903
aI329
aI7577
aI5813
aI17151
aI40889
aI174811
asS'd'
I15299
sa(dp30598
S'a'
I31807
sS's'
I18
sS'm'
(lp30599
I1
aI3
aI1
aI13
aI7
aI39
aI119
aI109
aI323
aI61
aI749
aI1377
aI911
aI8195
aI19753
aI6265
aI60783
aI182339
asS'd'
I15300
sa(dp30600
S'a'
I31812
sS's'
I18
sS'm'
(lp30601
I1
aI1
aI5
aI11
aI3
aI21
aI89
aI61
aI243
aI273
aI1317
aI3443
aI117
aI6205
aI13907
aI12903
aI95485
aI103355
asS'd'
I15301
sa(dp30602
S'a'
I31821
sS's'
I18
sS'm'
(lp30603
I1
aI3
aI7
aI13
aI19
aI27
aI45
aI251
aI405
aI289
aI121
aI1501
aI2599
aI8111
aI5163
aI17437
aI75095
aI165847
asS'd'
I15302
sa(dp30604
S'a'
I31855
sS's'
I18
sS'm'
(lp30605
I1
aI3
aI3
aI5
aI23
aI13
aI29
aI145
aI333
aI573
aI1939
aI2133
aI5797
aI5263
aI18835
aI11945
aI42161
aI103123
asS'd'
I15303
sa(dp30606
S'a'
I31919
sS's'
I18
sS'm'
(lp30607
I1
aI3
aI5
aI13
aI9
aI31
aI45
aI17
aI181
aI111
aI219
aI3451
aI1591
aI5823
aI20307
aI41207
aI77047
aI173401
asS'd'
I15304
sa(dp30608
S'a'
I31934
sS's'
I18
sS'm'
(lp30609
I1
aI3
aI1
aI9
aI19
aI63
aI73
aI111
aI377
aI875
aI1749
aI2887
aI7035
aI14209
aI1805
aI20527
aI93839
aI225185
asS'd'
I15305
sa(dp30610
S'a'
I31956
sS's'
I18
sS'm'
(lp30611
I1
aI3
aI1
aI9
aI15
aI45
aI97
aI235
aI11
aI803
aI899
aI427
aI3353
aI7363
aI26687
aI1307
aI5451
aI176233
asS'd'
I15306
sa(dp30612
S'a'
I31965
sS's'
I18
sS'm'
(lp30613
I1
aI3
aI7
aI5
aI7
aI51
aI59
aI53
aI3
aI263
aI159
aI1005
aI6079
aI7237
aI17419
aI56653
aI61091
aI182209
asS'd'
I15307
sa(dp30614
S'a'
I31975
sS's'
I18
sS'm'
(lp30615
I1
aI1
aI3
aI15
aI19
aI47
aI17
aI185
aI167
aI219
aI233
aI2921
aI5755
aI1277
aI27281
aI33671
aI3191
aI169477
asS'd'
I15308
sa(dp30616
S'a'
I31981
sS's'
I18
sS'm'
(lp30617
I1
aI3
aI7
aI1
aI25
aI27
aI99
aI81
aI57
aI969
aI821
aI2397
aI2947
aI5913
aI15247
aI47651
aI449
aI183295
asS'd'
I15309
sa(dp30618
S'a'
I31999
sS's'
I18
sS'm'
(lp30619
I1
aI3
aI1
aI15
aI3
aI63
aI83
aI75
aI41
aI959
aI1005
aI153
aI97
aI15381
aI6901
aI55141
aI90215
aI161321
asS'd'
I15310
sa(dp30620
S'a'
I32014
sS's'
I18
sS'm'
(lp30621
I1
aI3
aI1
aI9
aI1
aI1
aI29
aI191
aI43
aI241
aI607
aI667
aI1189
aI4389
aI31335
aI14133
aI104049
aI100925
asS'd'
I15311
sa(dp30622
S'a'
I32022
sS's'
I18
sS'm'
(lp30623
I1
aI1
aI3
aI1
aI1
aI61
aI109
aI23
aI325
aI27
aI1601
aI3957
aI7181
aI8375
aI9823
aI50533
aI114895
aI73825
asS'd'
I15312
sa(dp30624
S'a'
I32049
sS's'
I18
sS'm'
(lp30625
I1
aI1
aI7
aI7
aI1
aI25
aI51
aI19
aI383
aI955
aI1717
aI2953
aI5431
aI7883
aI14451
aI18619
aI9601
aI153151
asS'd'
I15313
sa(dp30626
S'a'
I32055
sS's'
I18
sS'm'
(lp30627
I1
aI3
aI1
aI5
aI1
aI1
aI35
aI3
aI141
aI37
aI1531
aI1855
aI7905
aI6509
aI6223
aI45911
aI54267
aI172275
asS'd'
I15314
sa(dp30628
S'a'
I32082
sS's'
I18
sS'm'
(lp30629
I1
aI1
aI3
aI15
aI23
aI39
aI109
aI145
aI215
aI147
aI1191
aI2425
aI301
aI5543
aI997
aI31071
aI101697
aI169677
asS'd'
I15315
sa(dp30630
S'a'
I32087
sS's'
I18
sS'm'
(lp30631
I1
aI1
aI3
aI5
aI17
aI23
aI41
aI191
aI367
aI967
aI1625
aI3669
aI769
aI7599
aI111
aI4399
aI64121
aI232275
asS'd'
I15316
sa(dp30632
S'a'
I32100
sS's'
I18
sS'm'
(lp30633
I1
aI3
aI5
aI1
aI25
aI61
aI47
aI247
aI413
aI605
aI399
aI1233
aI2789
aI9775
aI7111
aI42853
aI2305
aI87423
asS'd'
I15317
sa(dp30634
S'a'
I32104
sS's'
I18
sS'm'
(lp30635
I1
aI3
aI7
aI3
aI1
aI25
aI73
aI247
aI221
aI235
aI169
aI889
aI5635
aI4325
aI27015
aI39549
aI107545
aI80885
asS'd'
I15318
sa(dp30636
S'a'
I32109
sS's'
I18
sS'm'
(lp30637
I1
aI1
aI7
aI11
aI15
aI55
aI61
aI103
aI179
aI157
aI481
aI3089
aI4539
aI375
aI25425
aI14995
aI60119
aI31031
asS'd'
I15319
sa(dp30638
S'a'
I32138
sS's'
I18
sS'm'
(lp30639
I1
aI3
aI3
aI1
aI31
aI17
aI87
aI123
aI27
aI309
aI1693
aI3871
aI7319
aI15615
aI20113
aI18239
aI86407
aI172381
asS'd'
I15320
sa(dp30640
S'a'
I32140
sS's'
I18
sS'm'
(lp30641
I1
aI1
aI1
aI13
aI17
aI31
aI83
aI149
aI451
aI305
aI847
aI223
aI5705
aI9697
aI4967
aI34273
aI4041
aI252891
asS'd'
I15321
sa(dp30642
S'a'
I32148
sS's'
I18
sS'm'
(lp30643
I1
aI3
aI3
aI15
aI7
aI27
aI105
aI207
aI443
aI825
aI701
aI1159
aI5267
aI14085
aI28295
aI41757
aI47799
aI14835
asS'd'
I15322
sa(dp30644
S'a'
I32161
sS's'
I18
sS'm'
(lp30645
I1
aI3
aI7
aI3
aI1
aI31
aI77
aI219
aI139
aI497
aI1575
aI905
aI4341
aI4611
aI27861
aI59871
aI45525
aI21735
asS'd'
I15323
sa(dp30646
S'a'
I32162
sS's'
I18
sS'm'
(lp30647
I1
aI3
aI3
aI3
aI7
aI17
aI65
aI183
aI231
aI955
aI1111
aI1899
aI1677
aI13685
aI29395
aI10449
aI62505
aI125643
asS'd'
I15324
sa(dp30648
S'a'
I32174
sS's'
I18
sS'm'
(lp30649
I1
aI3
aI5
aI7
aI27
aI57
aI81
aI165
aI279
aI989
aI1569
aI573
aI7593
aI10067
aI1343
aI12039
aI117175
aI225125
asS'd'
I15325
sa(dp30650
S'a'
I32196
sS's'
I18
sS'm'
(lp30651
I1
aI3
aI3
aI15
aI19
aI37
aI47
aI175
aI87
aI153
aI1137
aI1985
aI2473
aI14767
aI19587
aI41751
aI98937
aI66667
asS'd'
I15326
sa(dp30652
S'a'
I32220
sS's'
I18
sS'm'
(lp30653
I1
aI3
aI5
aI3
aI19
aI51
aI25
aI155
aI37
aI597
aI719
aI1039
aI165
aI1871
aI15677
aI59891
aI29899
aI231979
asS'd'
I15327
sa(dp30654
S'a'
I32223
sS's'
I18
sS'm'
(lp30655
I1
aI3
aI5
aI13
aI15
aI39
aI17
aI21
aI73
aI695
aI1813
aI2463
aI3549
aI3081
aI14037
aI13077
aI40417
aI258995
asS'd'
I15328
sa(dp30656
S'a'
I32267
sS's'
I18
sS'm'
(lp30657
I1
aI1
aI5
aI9
aI29
aI5
aI105
aI75
aI97
aI937
aI1767
aI975
aI637
aI9419
aI30673
aI33537
aI979
aI45381
asS'd'
I15329
sa(dp30658
S'a'
I32272
sS's'
I18
sS'm'
(lp30659
I1
aI1
aI5
aI9
aI3
aI31
aI91
aI193
aI171
aI163
aI925
aI3519
aI3621
aI4943
aI14093
aI22881
aI18459
aI110155
asS'd'
I15330
sa(dp30660
S'a'
I32308
sS's'
I18
sS'm'
(lp30661
I1
aI1
aI3
aI7
aI29
aI1
aI39
aI107
aI359
aI805
aI91
aI2911
aI4741
aI3099
aI16967
aI45849
aI95255
aI63225
asS'd'
I15331
sa(dp30662
S'a'
I32315
sS's'
I18
sS'm'
(lp30663
I1
aI1
aI5
aI13
aI25
aI41
aI49
aI145
aI345
aI823
aI1571
aI341
aI6323
aI9679
aI14855
aI19965
aI108367
aI99833
asS'd'
I15332
sa(dp30664
S'a'
I32317
sS's'
I18
sS'm'
(lp30665
I1
aI3
aI5
aI3
aI27
aI35
aI87
aI83
aI373
aI425
aI281
aI1313
aI5153
aI6301
aI2745
aI12677
aI34603
aI181835
asS'd'
I15333
sa(dp30666
S'a'
I32347
sS's'
I18
sS'm'
(lp30667
I1
aI1
aI7
aI7
aI13
aI17
aI83
aI101
aI141
aI789
aI1403
aI2777
aI2749
aI1551
aI9009
aI9909
aI48443
aI176679
asS'd'
I15334
sa(dp30668
S'a'
I32360
sS's'
I18
sS'm'
(lp30669
I1
aI3
aI5
aI9
aI7
aI25
aI125
aI109
aI155
aI357
aI1111
aI3057
aI771
aI11253
aI25811
aI60333
aI68505
aI146987
asS'd'
I15335
sa(dp30670
S'a'
I32394
sS's'
I18
sS'm'
(lp30671
I1
aI3
aI7
aI7
aI29
aI19
aI69
aI115
aI411
aI793
aI51
aI715
aI3035
aI11577
aI14237
aI23963
aI13915
aI196771
asS'd'
I15336
sa(dp30672
S'a'
I32411
sS's'
I18
sS'm'
(lp30673
I1
aI1
aI3
aI3
aI27
aI37
aI61
aI163
aI131
aI749
aI37
aI1333
aI47
aI2519
aI25473
aI40851
aI55861
aI113961
asS'd'
I15337
sa(dp30674
S'a'
I32420
sS's'
I18
sS'm'
(lp30675
I1
aI3
aI5
aI1
aI3
aI49
aI19
aI251
aI125
aI387
aI1887
aI3571
aI1465
aI4831
aI3859
aI43357
aI20859
aI225835
asS'd'
I15338
sa(dp30676
S'a'
I32423
sS's'
I18
sS'm'
(lp30677
I1
aI3
aI5
aI9
aI27
aI53
aI53
aI143
aI383
aI781
aI819
aI2921
aI3499
aI11551
aI18761
aI14453
aI58209
aI181763
asS'd'
I15339
sa(dp30678
S'a'
I32455
sS's'
I18
sS'm'
(lp30679
I1
aI3
aI7
aI9
aI17
aI17
aI79
aI61
aI145
aI413
aI541
aI895
aI2077
aI6957
aI28681
aI44821
aI30609
aI131705
asS'd'
I15340
sa(dp30680
S'a'
I32456
sS's'
I18
sS'm'
(lp30681
I1
aI1
aI3
aI15
aI1
aI57
aI17
aI125
aI11
aI43
aI1079
aI1023
aI5391
aI67
aI31701
aI5737
aI429
aI75411
asS'd'
I15341
sa(dp30682
S'a'
I32510
sS's'
I18
sS'm'
(lp30683
I1
aI1
aI7
aI7
aI21
aI45
aI65
aI127
aI447
aI793
aI161
aI333
aI637
aI7403
aI12861
aI30173
aI125121
aI254687
asS'd'
I15342
sa(dp30684
S'a'
I32517
sS's'
I18
sS'm'
(lp30685
I1
aI1
aI3
aI9
aI27
aI53
aI85
aI223
aI297
aI455
aI919
aI2371
aI7049
aI7167
aI18075
aI22815
aI10265
aI14765
asS'd'
I15343
sa(dp30686
S'a'
I32541
sS's'
I18
sS'm'
(lp30687
I1
aI1
aI3
aI7
aI1
aI53
aI91
aI181
aI471
aI101
aI771
aI4043
aI3039
aI1215
aI19289
aI15941
aI55187
aI147355
asS'd'
I15344
sa(dp30688
S'a'
I32552
sS's'
I18
sS'm'
(lp30689
I1
aI3
aI1
aI1
aI13
aI19
aI13
aI47
aI159
aI965
aI1383
aI297
aI4299
aI7181
aI1271
aI17057
aI114847
aI180883
asS'd'
I15345
sa(dp30690
S'a'
I32555
sS's'
I18
sS'm'
(lp30691
I1
aI3
aI1
aI15
aI29
aI55
aI113
aI243
aI215
aI665
aI641
aI1975
aI5907
aI2617
aI17077
aI43697
aI61101
aI70007
asS'd'
I15346
sa(dp30692
S'a'
I32580
sS's'
I18
sS'm'
(lp30693
I1
aI3
aI1
aI11
aI5
aI55
aI109
aI145
aI307
aI663
aI1327
aI1247
aI8033
aI15425
aI18539
aI57027
aI72161
aI181655
asS'd'
I15347
sa(dp30694
S'a'
I32598
sS's'
I18
sS'm'
(lp30695
I1
aI3
aI5
aI9
aI25
aI17
aI3
aI57
aI7
aI449
aI1049
aI3423
aI5769
aI12713
aI29849
aI1017
aI92579
aI131255
asS'd'
I15348
sa(dp30696
S'a'
I32654
sS's'
I18
sS'm'
(lp30697
I1
aI1
aI7
aI13
aI25
aI13
aI55
aI217
aI461
aI595
aI931
aI1883
aI2645
aI9625
aI20467
aI45825
aI72027
aI163767
asS'd'
I15349
sa(dp30698
S'a'
I32662
sS's'
I18
sS'm'
(lp30699
I1
aI3
aI5
aI15
aI29
aI49
aI23
aI47
aI45
aI645
aI973
aI3837
aI621
aI7373
aI3585
aI16083
aI93823
aI184593
asS'd'
I15350
sa(dp30700
S'a'
I32665
sS's'
I18
sS'm'
(lp30701
I1
aI3
aI7
aI11
aI27
aI19
aI59
aI125
aI447
aI33
aI541
aI1067
aI6983
aI3779
aI27275
aI34269
aI106937
aI65085
asS'd'
I15351
sa(dp30702
S'a'
I32678
sS's'
I18
sS'm'
(lp30703
I1
aI3
aI1
aI9
aI23
aI23
aI51
aI43
aI475
aI861
aI1759
aI2559
aI3059
aI1175
aI31555
aI27671
aI117653
aI162735
asS'd'
I15352
sa(dp30704
S'a'
I32682
sS's'
I18
sS'm'
(lp30705
I1
aI1
aI7
aI13
aI7
aI61
aI33
aI49
aI23
aI737
aI949
aI1785
aI2921
aI873
aI26631
aI61941
aI14467
aI76225
asS'd'
I15353
sa(dp30706
S'a'
I32687
sS's'
I18
sS'm'
(lp30707
I1
aI3
aI7
aI5
aI31
aI11
aI49
aI149
aI7
aI85
aI1929
aI1001
aI4185
aI221
aI23719
aI52265
aI52973
aI67967
asS'd'
I15354
sa(dp30708
S'a'
I32692
sS's'
I18
sS'm'
(lp30709
I1
aI3
aI1
aI13
aI17
aI31
aI35
aI191
aI65
aI527
aI51
aI1093
aI3673
aI11167
aI29985
aI59739
aI43567
aI109817
asS'd'
I15355
sa(dp30710
S'a'
I32710
sS's'
I18
sS'm'
(lp30711
I1
aI1
aI1
aI11
aI23
aI39
aI95
aI121
aI501
aI355
aI1043
aI993
aI7533
aI15763
aI18399
aI31601
aI49373
aI243209
asS'd'
I15356
sa(dp30712
S'a'
I32791
sS's'
I18
sS'm'
(lp30713
I1
aI1
aI3
aI3
aI15
aI37
aI57
aI183
aI27
aI981
aI153
aI1481
aI549
aI7847
aI2689
aI57401
aI46359
aI175401
asS'd'
I15357
sa(dp30714
S'a'
I32792
sS's'
I18
sS'm'
(lp30715
I1
aI3
aI5
aI9
aI31
aI19
aI83
aI79
aI413
aI597
aI1957
aI3027
aI4751
aI1437
aI11255
aI39513
aI56927
aI166841
asS'd'
I15358
sa(dp30716
S'a'
I32813
sS's'
I18
sS'm'
(lp30717
I1
aI3
aI7
aI5
aI27
aI61
aI69
aI65
aI143
aI321
aI1129
aI2521
aI4467
aI4369
aI11727
aI35643
aI80155
aI184241
asS'd'
I15359
sa(dp30718
S'a'
I32826
sS's'
I18
sS'm'
(lp30719
I1
aI3
aI7
aI1
aI25
aI27
aI107
aI219
aI481
aI457
aI2027
aI1057
aI6409
aI5641
aI19711
aI11009
aI44295
aI28171
asS'd'
I15360
sa(dp30720
S'a'
I32831
sS's'
I18
sS'm'
(lp30721
I1
aI3
aI3
aI9
aI3
aI17
aI85
aI5
aI341
aI107
aI2037
aI93
aI741
aI5279
aI20093
aI28637
aI80823
aI210279
asS'd'
I15361
sa(dp30722
S'a'
I32836
sS's'
I18
sS'm'
(lp30723
I1
aI1
aI7
aI15
aI5
aI9
aI33
aI167
aI451
aI463
aI1951
aI2395
aI3821
aI2915
aI15195
aI34517
aI113545
aI22173
asS'd'
I15362
sa(dp30724
S'a'
I32843
sS's'
I18
sS'm'
(lp30725
I1
aI1
aI3
aI1
aI23
aI55
aI113
aI159
aI139
aI795
aI69
aI2021
aI6769
aI10807
aI18605
aI54161
aI39501
aI233797
asS'd'
I15363
sa(dp30726
S'a'
I32854
sS's'
I18
sS'm'
(lp30727
I1
aI1
aI7
aI11
aI17
aI53
aI103
aI131
aI385
aI909
aI629
aI3127
aI6117
aI11457
aI31115
aI8255
aI33227
aI170877
asS'd'
I15364
sa(dp30728
S'a'
I32893
sS's'
I18
sS'm'
(lp30729
I1
aI1
aI3
aI9
aI23
aI19
aI99
aI221
aI141
aI731
aI311
aI2617
aI2763
aI375
aI26763
aI56473
aI6613
aI60519
asS'd'
I15365
sa(dp30730
S'a'
I32897
sS's'
I18
sS'm'
(lp30731
I1
aI3
aI1
aI9
aI7
aI29
aI15
aI105
aI243
aI159
aI1755
aI3999
aI5861
aI12009
aI30111
aI48269
aI70427
aI187173
asS'd'
I15366
sa(dp30732
S'a'
I32912
sS's'
I18
sS'm'
(lp30733
I1
aI1
aI3
aI13
aI1
aI35
aI39
aI121
aI409
aI571
aI1835
aI1535
aI4671
aI12671
aI4503
aI4783
aI48009
aI216837
asS'd'
I15367
sa(dp30734
S'a'
I32951
sS's'
I18
sS'm'
(lp30735
I1
aI1
aI1
aI9
aI25
aI61
aI87
aI109
aI489
aI107
aI1741
aI859
aI237
aI7161
aI27117
aI58587
aI55445
aI155763
asS'd'
I15368
sa(dp30736
S'a'
I32963
sS's'
I18
sS'm'
(lp30737
I1
aI1
aI7
aI15
aI9
aI21
aI61
aI159
aI301
aI863
aI1823
aI11
aI419
aI6717
aI28199
aI24129
aI58419
aI22147
asS'd'
I15369
sa(dp30738
S'a'
I32970
sS's'
I18
sS'm'
(lp30739
I1
aI3
aI1
aI5
aI9
aI7
aI13
aI205
aI185
aI759
aI777
aI2877
aI5991
aI14555
aI18793
aI51485
aI6373
aI232105
asS'd'
I15370
sa(dp30740
S'a'
I32975
sS's'
I18
sS'm'
(lp30741
I1
aI1
aI3
aI5
aI7
aI25
aI51
aI79
aI227
aI447
aI867
aI2709
aI2677
aI15249
aI22957
aI45577
aI39011
aI16839
asS'd'
I15371
sa(dp30742
S'a'
I33028
sS's'
I18
sS'm'
(lp30743
I1
aI1
aI1
aI3
aI5
aI51
aI69
aI135
aI395
aI339
aI685
aI3657
aI3789
aI16345
aI2911
aI51737
aI97471
aI126605
asS'd'
I15372
sa(dp30744
S'a'
I33052
sS's'
I18
sS'm'
(lp30745
I1
aI3
aI7
aI15
aI15
aI5
aI85
aI153
aI507
aI347
aI1457
aI527
aI1055
aI7773
aI4161
aI10487
aI92373
aI256535
asS'd'
I15373
sa(dp30746
S'a'
I33061
sS's'
I18
sS'm'
(lp30747
I1
aI3
aI3
aI3
aI1
aI9
aI25
aI1
aI155
aI321
aI1739
aI555
aI7719
aI10501
aI19075
aI12529
aI75975
aI229905
asS'd'
I15374
sa(dp30748
S'a'
I33083
sS's'
I18
sS'm'
(lp30749
I1
aI1
aI1
aI3
aI29
aI31
aI89
aI23
aI283
aI875
aI1653
aI855
aI7613
aI15277
aI23845
aI47443
aI287
aI217441
asS'd'
I15375
sa(dp30750
S'a'
I33103
sS's'
I18
sS'm'
(lp30751
I1
aI3
aI3
aI1
aI25
aI7
aI75
aI119
aI493
aI131
aI231
aI3063
aI7171
aI5437
aI16385
aI50347
aI87427
aI53603
asS'd'
I15376
sa(dp30752
S'a'
I33117
sS's'
I18
sS'm'
(lp30753
I1
aI3
aI7
aI7
aI27
aI55
aI103
aI223
aI219
aI103
aI733
aI1233
aI1931
aI2119
aI19333
aI59839
aI100421
aI256811
asS'd'
I15377
sa(dp30754
S'a'
I33139
sS's'
I18
sS'm'
(lp30755
I1
aI3
aI7
aI15
aI23
aI63
aI77
aI151
aI285
aI701
aI1403
aI1267
aI6975
aI11421
aI24943
aI51647
aI75651
aI191675
asS'd'
I15378
sa(dp30756
S'a'
I33151
sS's'
I18
sS'm'
(lp30757
I1
aI1
aI5
aI7
aI25
aI23
aI25
aI7
aI503
aI759
aI997
aI1711
aI1439
aI12483
aI30117
aI55205
aI8813
aI221589
asS'd'
I15379
sa(dp30758
S'a'
I33162
sS's'
I18
sS'm'
(lp30759
I1
aI3
aI7
aI5
aI21
aI3
aI117
aI65
aI461
aI851
aI915
aI575
aI3061
aI1055
aI11999
aI8375
aI128677
aI98005
asS'd'
I15380
sa(dp30760
S'a'
I33209
sS's'
I18
sS'm'
(lp30761
I1
aI3
aI7
aI7
aI19
aI13
aI123
aI23
aI41
aI293
aI79
aI1435
aI1175
aI7399
aI14907
aI4671
aI88029
aI220627
asS'd'
I15381
sa(dp30762
S'a'
I33210
sS's'
I18
sS'm'
(lp30763
I1
aI3
aI3
aI13
aI11
aI17
aI65
aI21
aI143
aI257
aI1001
aI2423
aI5659
aI11681
aI23605
aI40649
aI49797
aI55497
asS'd'
I15382
sa(dp30764
S'a'
I33223
sS's'
I18
sS'm'
(lp30765
I1
aI3
aI5
aI5
aI29
aI15
aI125
aI83
aI139
aI381
aI1435
aI2129
aI1699
aI10725
aI531
aI767
aI112477
aI134223
asS'd'
I15383
sa(dp30766
S'a'
I33227
sS's'
I18
sS'm'
(lp30767
I1
aI3
aI7
aI9
aI9
aI23
aI35
aI1
aI127
aI143
aI707
aI1475
aI4705
aI8921
aI14919
aI50909
aI64425
aI33381
asS'd'
I15384
sa(dp30768
S'a'
I33241
sS's'
I18
sS'm'
(lp30769
I1
aI1
aI1
aI15
aI11
aI63
aI87
aI101
aI243
aI833
aI707
aI4095
aI201
aI4877
aI10219
aI39019
aI129779
aI229857
asS'd'
I15385
sa(dp30770
S'a'
I33263
sS's'
I18
sS'm'
(lp30771
I1
aI1
aI1
aI11
aI5
aI9
aI35
aI177
aI303
aI545
aI917
aI1037
aI1789
aI12147
aI29095
aI27391
aI112833
aI104713
asS'd'
I15386
sa(dp30772
S'a'
I33341
sS's'
I18
sS'm'
(lp30773
I1
aI1
aI1
aI5
aI27
aI23
aI111
aI219
aI439
aI445
aI967
aI3527
aI6203
aI1829
aI19657
aI48965
aI85213
aI58719
asS'd'
I15387
sa(dp30774
S'a'
I33344
sS's'
I18
sS'm'
(lp30775
I1
aI3
aI7
aI11
aI15
aI7
aI95
aI113
aI317
aI225
aI1229
aI3033
aI5777
aI4075
aI24093
aI3539
aI19333
aI212757
asS'd'
I15388
sa(dp30776
S'a'
I33356
sS's'
I18
sS'm'
(lp30777
I1
aI3
aI7
aI7
aI1
aI35
aI35
aI67
aI459
aI769
aI1675
aI3509
aI7393
aI10433
aI12003
aI7385
aI4425
aI188989
asS'd'
I15389
sa(dp30778
S'a'
I33390
sS's'
I18
sS'm'
(lp30779
I1
aI1
aI7
aI9
aI5
aI45
aI35
aI27
aI443
aI301
aI533
aI2803
aI99
aI17
aI20749
aI58353
aI93067
aI118763
asS'd'
I15390
sa(dp30780
S'a'
I33398
sS's'
I18
sS'm'
(lp30781
I1
aI1
aI3
aI3
aI13
aI51
aI61
aI181
aI81
aI859
aI1461
aI3455
aI2277
aI13769
aI1251
aI4313
aI119973
aI30693
asS'd'
I15391
sa(dp30782
S'a'
I33423
sS's'
I18
sS'm'
(lp30783
I1
aI1
aI3
aI7
aI15
aI29
aI23
aI207
aI239
aI65
aI1889
aI151
aI7793
aI2657
aI13673
aI29033
aI74477
aI215027
asS'd'
I15392
sa(dp30784
S'a'
I33428
sS's'
I18
sS'm'
(lp30785
I1
aI1
aI5
aI1
aI19
aI35
aI19
aI71
aI187
aI783
aI543
aI505
aI347
aI3191
aI1087
aI49735
aI54109
aI27979
asS'd'
I15393
sa(dp30786
S'a'
I33444
sS's'
I18
sS'm'
(lp30787
I1
aI3
aI3
aI3
aI3
aI61
aI67
aI59
aI207
aI957
aI1709
aI1567
aI7973
aI5921
aI29841
aI8311
aI81165
aI91965
asS'd'
I15394
sa(dp30788
S'a'
I33454
sS's'
I18
sS'm'
(lp30789
I1
aI1
aI1
aI7
aI23
aI47
aI109
aI189
aI447
aI861
aI1615
aI3283
aI3059
aI749
aI28729
aI8713
aI38743
aI211019
asS'd'
I15395
sa(dp30790
S'a'
I33479
sS's'
I18
sS'm'
(lp30791
I1
aI3
aI5
aI1
aI17
aI51
aI127
aI181
aI355
aI515
aI603
aI465
aI1825
aI9281
aI31971
aI42793
aI22467
aI175777
asS'd'
I15396
sa(dp30792
S'a'
I33510
sS's'
I18
sS'm'
(lp30793
I1
aI3
aI1
aI5
aI3
aI29
aI111
aI91
aI99
aI15
aI535
aI3047
aI1083
aI7181
aI28003
aI60719
aI71825
aI12293
asS'd'
I15397
sa(dp30794
S'a'
I33514
sS's'
I18
sS'm'
(lp30795
I1
aI3
aI5
aI7
aI1
aI43
aI83
aI117
aI221
aI353
aI139
aI647
aI6017
aI4655
aI31823
aI22097
aI118537
aI71803
asS'd'
I15398
sa(dp30796
S'a'
I33528
sS's'
I18
sS'm'
(lp30797
I1
aI1
aI3
aI3
aI27
aI11
aI35
aI169
aI215
aI883
aI1195
aI2983
aI4651
aI15893
aI24051
aI313
aI103947
aI5227
asS'd'
I15399
sa(dp30798
S'a'
I33548
sS's'
I18
sS'm'
(lp30799
I1
aI3
aI5
aI13
aI3
aI51
aI33
aI159
aI499
aI763
aI845
aI1541
aI3837
aI9397
aI855
aI4187
aI112167
aI243817
asS'd'
I15400
sa(dp30800
S'a'
I33566
sS's'
I18
sS'm'
(lp30801
I1
aI1
aI7
aI9
aI1
aI15
aI19
aI239
aI227
aI561
aI1685
aI2841
aI53
aI9753
aI15105
aI34277
aI128859
aI100085
asS'd'
I15401
sa(dp30802
S'a'
I33570
sS's'
I18
sS'm'
(lp30803
I1
aI3
aI5
aI13
aI15
aI7
aI57
aI9
aI415
aI1005
aI583
aI1347
aI4349
aI3603
aI9125
aI15019
aI77735
aI237011
asS'd'
I15402
sa(dp30804
S'a'
I33593
sS's'
I18
sS'm'
(lp30805
I1
aI1
aI1
aI13
aI27
aI21
aI105
aI17
aI235
aI605
aI1417
aI2053
aI3233
aI11617
aI28651
aI43161
aI71663
aI98373
asS'd'
I15403
sa(dp30806
S'a'
I33599
sS's'
I18
sS'm'
(lp30807
I1
aI3
aI5
aI13
aI29
aI17
aI123
aI105
aI477
aI359
aI613
aI1699
aI2581
aI3007
aI8507
aI14391
aI95487
aI633
asS'd'
I15404
sa(dp30808
S'a'
I33619
sS's'
I18
sS'm'
(lp30809
I1
aI1
aI7
aI7
aI13
aI55
aI107
aI211
aI71
aI339
aI1169
aI2629
aI165
aI16207
aI25523
aI7101
aI47553
aI261131
asS'd'
I15405
sa(dp30810
S'a'
I33635
sS's'
I18
sS'm'
(lp30811
I1
aI3
aI5
aI15
aI11
aI7
aI45
aI207
aI39
aI41
aI781
aI3347
aI2529
aI4475
aI9665
aI31499
aI119837
aI128755
asS'd'
I15406
sa(dp30812
S'a'
I33659
sS's'
I18
sS'm'
(lp30813
I1
aI3
aI1
aI15
aI23
aI59
aI59
aI17
aI89
aI823
aI59
aI3991
aI305
aI14893
aI1411
aI8015
aI92193
aI66935
asS'd'
I15407
sa(dp30814
S'a'
I33680
sS's'
I18
sS'm'
(lp30815
I1
aI1
aI7
aI15
aI19
aI29
aI11
aI207
aI429
aI851
aI1661
aI2903
aI4413
aI447
aI29447
aI39243
aI70435
aI160451
asS'd'
I15408
sa(dp30816
S'a'
I33699
sS's'
I18
sS'm'
(lp30817
I1
aI1
aI3
aI13
aI17
aI5
aI93
aI89
aI455
aI67
aI33
aI65
aI7957
aI14383
aI28525
aI56983
aI71899
aI4881
asS'd'
I15409
sa(dp30818
S'a'
I33713
sS's'
I18
sS'm'
(lp30819
I1
aI3
aI3
aI5
aI11
aI7
aI47
aI233
aI433
aI791
aI47
aI2561
aI6539
aI1151
aI2083
aI12309
aI62353
aI69507
asS'd'
I15410
sa(dp30820
S'a'
I33758
sS's'
I18
sS'm'
(lp30821
I1
aI3
aI5
aI11
aI15
aI55
aI101
aI251
aI41
aI39
aI383
aI3481
aI1817
aI3447
aI6205
aI38169
aI98267
aI157091
asS'd'
I15411
sa(dp30822
S'a'
I33771
sS's'
I18
sS'm'
(lp30823
I1
aI3
aI5
aI7
aI9
aI47
aI113
aI55
aI421
aI703
aI849
aI2251
aI129
aI9257
aI28097
aI33475
aI98933
aI32041
asS'd'
I15412
sa(dp30824
S'a'
I33779
sS's'
I18
sS'm'
(lp30825
I1
aI3
aI3
aI11
aI9
aI59
aI1
aI211
aI277
aI969
aI977
aI3079
aI4707
aI3341
aI17679
aI9469
aI52859
aI125883
asS'd'
I15413
sa(dp30826
S'a'
I33800
sS's'
I18
sS'm'
(lp30827
I1
aI1
aI7
aI13
aI29
aI19
aI49
aI149
aI259
aI573
aI1137
aI2571
aI2661
aI12865
aI24993
aI10721
aI96921
aI85931
asS'd'
I15414
sa(dp30828
S'a'
I33829
sS's'
I18
sS'm'
(lp30829
I1
aI3
aI1
aI9
aI15
aI27
aI91
aI127
aI305
aI159
aI523
aI2539
aI1969
aI4325
aI595
aI37077
aI79919
aI26889
asS'd'
I15415
sa(dp30830
S'a'
I33836
sS's'
I18
sS'm'
(lp30831
I1
aI3
aI1
aI11
aI1
aI5
aI65
aI75
aI317
aI909
aI1601
aI2713
aI6891
aI4927
aI28427
aI5791
aI82285
aI35209
asS'd'
I15416
sa(dp30832
S'a'
I33844
sS's'
I18
sS'm'
(lp30833
I1
aI3
aI7
aI7
aI9
aI5
aI111
aI167
aI477
aI437
aI1133
aI2715
aI6189
aI5051
aI4765
aI26267
aI99819
aI70419
asS'd'
I15417
sa(dp30834
S'a'
I33856
sS's'
I18
sS'm'
(lp30835
I1
aI1
aI3
aI9
aI27
aI57
aI23
aI233
aI423
aI191
aI1159
aI1539
aI6397
aI16041
aI8803
aI19787
aI114447
aI17029
asS'd'
I15418
sa(dp30836
S'a'
I33859
sS's'
I18
sS'm'
(lp30837
I1
aI1
aI5
aI1
aI11
aI3
aI111
aI125
aI287
aI487
aI1663
aI1953
aI3771
aI2011
aI18167
aI47471
aI94041
aI87569
asS'd'
I15419
sa(dp30838
S'a'
I33873
sS's'
I18
sS'm'
(lp30839
I1
aI3
aI1
aI11
aI5
aI9
aI75
aI153
aI37
aI803
aI971
aI1667
aI4631
aI9183
aI20179
aI6905
aI80949
aI54443
asS'd'
I15420
sa(dp30840
S'a'
I33874
sS's'
I18
sS'm'
(lp30841
I1
aI1
aI1
aI5
aI13
aI29
aI91
aI201
aI49
aI739
aI1569
aI2725
aI257
aI5505
aI5289
aI40731
aI27843
aI16565
asS'd'
I15421
sa(dp30842
S'a'
I33929
sS's'
I18
sS'm'
(lp30843
I1
aI1
aI7
aI13
aI27
aI41
aI81
aI199
aI99
aI43
aI1331
aI3237
aI6493
aI3839
aI19329
aI44371
aI19715
aI60553
asS'd'
I15422
sa(dp30844
S'a'
I33944
sS's'
I18
sS'm'
(lp30845
I1
aI1
aI5
aI7
aI29
aI41
aI67
aI163
aI467
aI93
aI1977
aI2625
aI6967
aI6655
aI19835
aI39517
aI10259
aI200487
asS'd'
I15423
sa(dp30846
S'a'
I33965
sS's'
I18
sS'm'
(lp30847
I1
aI1
aI3
aI15
aI23
aI35
aI31
aI171
aI175
aI883
aI593
aI245
aI6209
aI7381
aI5555
aI54507
aI66159
aI40771
asS'd'
I15424
sa(dp30848
S'a'
I33978
sS's'
I18
sS'm'
(lp30849
I1
aI3
aI3
aI11
aI3
aI63
aI75
aI177
aI239
aI77
aI1543
aI875
aI7951
aI7571
aI961
aI9909
aI101781
aI160399
asS'd'
I15425
sa(dp30850
S'a'
I33986
sS's'
I18
sS'm'
(lp30851
I1
aI3
aI1
aI3
aI3
aI37
aI71
aI231
aI373
aI443
aI835
aI1321
aI2107
aI2929
aI7527
aI47969
aI15329
aI94537
asS'd'
I15426
sa(dp30852
S'a'
I34006
sS's'
I18
sS'm'
(lp30853
I1
aI3
aI3
aI15
aI7
aI5
aI127
aI121
aI159
aI25
aI399
aI3009
aI4401
aI9649
aI4311
aI18045
aI22557
aI135177
asS'd'
I15427
sa(dp30854
S'a'
I34025
sS's'
I18
sS'm'
(lp30855
I1
aI1
aI1
aI3
aI29
aI57
aI75
aI73
aI261
aI493
aI1417
aI1351
aI3407
aI8553
aI4893
aI10325
aI123149
aI161435
asS'd'
I15428
sa(dp30856
S'a'
I34106
sS's'
I18
sS'm'
(lp30857
I1
aI3
aI3
aI11
aI15
aI5
aI87
aI115
aI337
aI213
aI949
aI1925
aI5057
aI5831
aI6837
aI51167
aI25291
aI182197
asS'd'
I15429
sa(dp30858
S'a'
I34126
sS's'
I18
sS'm'
(lp30859
I1
aI3
aI1
aI11
aI25
aI49
aI27
aI101
aI403
aI989
aI1129
aI3933
aI1147
aI13091
aI11965
aI38075
aI68251
aI103293
asS'd'
I15430
sa(dp30860
S'a'
I34171
sS's'
I18
sS'm'
(lp30861
I1
aI1
aI7
aI5
aI7
aI49
aI1
aI189
aI275
aI63
aI149
aI3255
aI1175
aI7811
aI24845
aI20755
aI99391
aI140673
asS'd'
I15431
sa(dp30862
S'a'
I34189
sS's'
I18
sS'm'
(lp30863
I1
aI3
aI1
aI13
aI17
aI35
aI37
aI37
aI415
aI125
aI643
aI443
aI6215
aI299
aI31237
aI45687
aI78535
aI102123
asS'd'
I15432
sa(dp30864
S'a'
I34204
sS's'
I18
sS'm'
(lp30865
I1
aI3
aI5
aI3
aI27
aI41
aI85
aI215
aI47
aI21
aI725
aI1967
aI2317
aI121
aI7827
aI48229
aI82027
aI60271
asS'd'
I15433
sa(dp30866
S'a'
I34260
sS's'
I18
sS'm'
(lp30867
I1
aI3
aI5
aI1
aI1
aI55
aI37
aI183
aI117
aI421
aI383
aI3883
aI5519
aI6161
aI6823
aI18423
aI77747
aI215969
asS'd'
I15434
sa(dp30868
S'a'
I34270
sS's'
I18
sS'm'
(lp30869
I1
aI3
aI5
aI13
aI31
aI3
aI117
aI59
aI375
aI797
aI1129
aI1283
aI3245
aI12775
aI30353
aI3837
aI130273
aI228899
asS'd'
I15435
sa(dp30870
S'a'
I34280
sS's'
I18
sS'm'
(lp30871
I1
aI3
aI5
aI1
aI7
aI33
aI17
aI153
aI179
aI255
aI537
aI2911
aI1223
aI367
aI18131
aI25903
aI33509
aI220031
asS'd'
I15436
sa(dp30872
S'a'
I34298
sS's'
I18
sS'm'
(lp30873
I1
aI3
aI7
aI5
aI5
aI7
aI103
aI233
aI309
aI947
aI1835
aI3509
aI4267
aI15619
aI5895
aI30707
aI81841
aI191899
asS'd'
I15437
sa(dp30874
S'a'
I34313
sS's'
I18
sS'm'
(lp30875
I1
aI1
aI5
aI11
aI3
aI15
aI91
aI83
aI319
aI765
aI895
aI2565
aI6833
aI1719
aI2971
aI37483
aI21709
aI23193
asS'd'
I15438
sa(dp30876
S'a'
I34321
sS's'
I18
sS'm'
(lp30877
I1
aI1
aI3
aI11
aI31
aI57
aI83
aI233
aI439
aI161
aI1503
aI749
aI6347
aI15379
aI2317
aI24671
aI93399
aI239585
asS'd'
I15439
sa(dp30878
S'a'
I34333
sS's'
I18
sS'm'
(lp30879
I1
aI1
aI3
aI9
aI29
aI45
aI19
aI107
aI295
aI153
aI189
aI2521
aI5465
aI7321
aI6143
aI229
aI100553
aI258911
asS'd'
I15440
sa(dp30880
S'a'
I34370
sS's'
I18
sS'm'
(lp30881
I1
aI3
aI7
aI11
aI31
aI27
aI95
aI21
aI249
aI981
aI1725
aI1481
aI1025
aI9301
aI11809
aI53109
aI29007
aI127683
asS'd'
I15441
sa(dp30882
S'a'
I34376
sS's'
I18
sS'm'
(lp30883
I1
aI3
aI1
aI3
aI1
aI23
aI97
aI137
aI5
aI471
aI1887
aI1035
aI2681
aI5143
aI3145
aI1517
aI88107
aI245245
asS'd'
I15442
sa(dp30884
S'a'
I34387
sS's'
I18
sS'm'
(lp30885
I1
aI1
aI1
aI1
aI15
aI11
aI13
aI9
aI405
aI607
aI403
aI1693
aI4363
aI9365
aI6425
aI48121
aI78969
aI87341
asS'd'
I15443
sa(dp30886
S'a'
I34389
sS's'
I18
sS'm'
(lp30887
I1
aI1
aI7
aI15
aI3
aI17
aI7
aI51
aI111
aI1023
aI9
aI465
aI1909
aI16283
aI4763
aI50939
aI119029
aI198257
asS'd'
I15444
sa(dp30888
S'a'
I34417
sS's'
I18
sS'm'
(lp30889
I1
aI3
aI3
aI1
aI31
aI11
aI113
aI13
aI499
aI433
aI1941
aI991
aI5439
aI3123
aI24591
aI16171
aI55099
aI206015
asS'd'
I15445
sa(dp30890
S'a'
I34429
sS's'
I18
sS'm'
(lp30891
I1
aI3
aI5
aI15
aI25
aI49
aI125
aI101
aI251
aI619
aI1895
aI4063
aI3065
aI14965
aI20081
aI11233
aI58253
aI69633
asS'd'
I15446
sa(dp30892
S'a'
I34440
sS's'
I18
sS'm'
(lp30893
I1
aI1
aI1
aI5
aI21
aI35
aI29
aI241
aI359
aI553
aI1001
aI1865
aI5183
aI5233
aI16371
aI55277
aI102091
aI143275
asS'd'
I15447
sa(dp30894
S'a'
I34443
sS's'
I18
sS'm'
(lp30895
I1
aI1
aI1
aI3
aI29
aI37
aI3
aI191
aI239
aI961
aI2031
aI1337
aI1169
aI5229
aI22861
aI38257
aI55027
aI153703
asS'd'
I15448
sa(dp30896
S'a'
I34451
sS's'
I18
sS'm'
(lp30897
I1
aI3
aI5
aI7
aI7
aI35
aI49
aI139
aI509
aI381
aI1267
aI2641
aI747
aI16239
aI23133
aI32111
aI70471
aI128427
asS'd'
I15449
sa(dp30898
S'a'
I34470
sS's'
I18
sS'm'
(lp30899
I1
aI1
aI1
aI9
aI23
aI25
aI117
aI125
aI369
aI891
aI103
aI2215
aI3571
aI1291
aI9001
aI35671
aI67119
aI198847
asS'd'
I15450
sa(dp30900
S'a'
I34484
sS's'
I18
sS'm'
(lp30901
I1
aI1
aI5
aI9
aI17
aI19
aI27
aI7
aI207
aI55
aI1099
aI2117
aI7511
aI14999
aI7761
aI32215
aI103401
aI68599
asS'd'
I15451
sa(dp30902
S'a'
I34493
sS's'
I18
sS'm'
(lp30903
I1
aI1
aI7
aI9
aI1
aI59
aI41
aI91
aI9
aI225
aI457
aI3241
aI4713
aI2923
aI11973
aI2867
aI130583
aI202007
asS'd'
I15452
sa(dp30904
S'a'
I34496
sS's'
I18
sS'm'
(lp30905
I1
aI3
aI1
aI9
aI31
aI47
aI63
aI49
aI457
aI757
aI885
aI937
aI2973
aI12147
aI2607
aI52907
aI126047
aI83275
asS'd'
I15453
sa(dp30906
S'a'
I34514
sS's'
I18
sS'm'
(lp30907
I1
aI3
aI5
aI11
aI17
aI1
aI79
aI123
aI505
aI203
aI1779
aI71
aI4357
aI2285
aI31625
aI15225
aI86519
aI2021
asS'd'
I15454
sa(dp30908
S'a'
I34526
sS's'
I18
sS'm'
(lp30909
I1
aI3
aI1
aI11
aI21
aI17
aI41
aI169
aI125
aI995
aI351
aI1235
aI25
aI7463
aI2099
aI18917
aI71355
aI26983
asS'd'
I15455
sa(dp30910
S'a'
I34535
sS's'
I18
sS'm'
(lp30911
I1
aI1
aI7
aI1
aI21
aI23
aI41
aI5
aI415
aI405
aI1235
aI1245
aI151
aI11283
aI25293
aI45147
aI12597
aI39501
asS'd'
I15456
sa(dp30912
S'a'
I34585
sS's'
I18
sS'm'
(lp30913
I1
aI1
aI1
aI5
aI29
aI29
aI15
aI165
aI259
aI179
aI1479
aI3535
aI779
aI6583
aI885
aI34331
aI71193
aI154417
asS'd'
I15457
sa(dp30914
S'a'
I34591
sS's'
I18
sS'm'
(lp30915
I1
aI1
aI1
aI5
aI3
aI1
aI13
aI181
aI507
aI339
aI333
aI4059
aI7963
aI15649
aI15507
aI16913
aI34741
aI202039
asS'd'
I15458
sa(dp30916
S'a'
I34592
sS's'
I18
sS'm'
(lp30917
I1
aI3
aI7
aI1
aI1
aI9
aI17
aI119
aI77
aI583
aI259
aI883
aI4011
aI4275
aI9599
aI58663
aI73237
aI202783
asS'd'
I15459
sa(dp30918
S'a'
I34602
sS's'
I18
sS'm'
(lp30919
I1
aI1
aI5
aI11
aI23
aI27
aI19
aI171
aI373
aI779
aI661
aI1701
aI3363
aI13095
aI897
aI51233
aI1319
aI41093
asS'd'
I15460
sa(dp30920
S'a'
I34607
sS's'
I18
sS'm'
(lp30921
I1
aI3
aI3
aI3
aI29
aI21
aI105
aI29
aI429
aI657
aI1735
aI1279
aI809
aI14963
aI9735
aI23251
aI44879
aI159371
asS'd'
I15461
sa(dp30922
S'a'
I34644
sS's'
I18
sS'm'
(lp30923
I1
aI1
aI5
aI11
aI17
aI27
aI117
aI65
aI193
aI539
aI1095
aI439
aI1687
aI11277
aI513
aI30611
aI88885
aI69145
asS'd'
I15462
sa(dp30924
S'a'
I34657
sS's'
I18
sS'm'
(lp30925
I1
aI3
aI1
aI7
aI11
aI1
aI27
aI41
aI63
aI501
aI917
aI2397
aI6839
aI10835
aI26437
aI56169
aI46631
aI64095
asS'd'
I15463
sa(dp30926
S'a'
I34675
sS's'
I18
sS'm'
(lp30927
I1
aI1
aI7
aI15
aI29
aI5
aI17
aI217
aI333
aI389
aI403
aI3167
aI3599
aI12055
aI30669
aI44821
aI109811
aI237393
asS'd'
I15464
sa(dp30928
S'a'
I34681
sS's'
I18
sS'm'
(lp30929
I1
aI3
aI5
aI13
aI1
aI39
aI51
aI233
aI159
aI135
aI763
aI2499
aI7741
aI13099
aI8639
aI8043
aI39827
aI5989
asS'd'
I15465
sa(dp30930
S'a'
I34693
sS's'
I18
sS'm'
(lp30931
I1
aI1
aI7
aI3
aI1
aI61
aI41
aI37
aI37
aI67
aI867
aI2631
aI6265
aI5551
aI161
aI56643
aI126087
aI126829
asS'd'
I15466
sa(dp30932
S'a'
I34706
sS's'
I18
sS'm'
(lp30933
I1
aI1
aI7
aI1
aI21
aI39
aI101
aI225
aI489
aI123
aI661
aI2489
aI1865
aI6809
aI21663
aI59405
aI45579
aI51257
asS'd'
I15467
sa(dp30934
S'a'
I34745
sS's'
I18
sS'm'
(lp30935
I1
aI3
aI7
aI9
aI27
aI53
aI11
aI97
aI369
aI389
aI1933
aI3321
aI543
aI12331
aI11571
aI10685
aI49049
aI244027
asS'd'
I15468
sa(dp30936
S'a'
I34759
sS's'
I18
sS'm'
(lp30937
I1
aI3
aI5
aI3
aI7
aI15
aI21
aI165
aI181
aI877
aI1161
aI1815
aI2097
aI449
aI32411
aI22843
aI12467
aI55397
asS'd'
I15469
sa(dp30938
S'a'
I34765
sS's'
I18
sS'm'
(lp30939
I1
aI3
aI1
aI3
aI11
aI45
aI23
aI193
aI287
aI137
aI333
aI1831
aI457
aI583
aI23081
aI4525
aI4781
aI249509
asS'd'
I15470
sa(dp30940
S'a'
I34774
sS's'
I18
sS'm'
(lp30941
I1
aI3
aI5
aI5
aI15
aI13
aI27
aI199
aI267
aI297
aI923
aI3861
aI4949
aI7945
aI25291
aI45407
aI47529
aI127287
asS'd'
I15471
sa(dp30942
S'a'
I34780
sS's'
I18
sS'm'
(lp30943
I1
aI1
aI5
aI7
aI19
aI29
aI113
aI51
aI503
aI487
aI699
aI2097
aI2957
aI6519
aI19487
aI46873
aI38871
aI89997
asS'd'
I15472
sa(dp30944
S'a'
I34794
sS's'
I18
sS'm'
(lp30945
I1
aI1
aI5
aI13
aI17
aI31
aI57
aI127
aI335
aI223
aI1545
aI3749
aI1539
aI3293
aI21159
aI13019
aI48343
aI190895
asS'd'
I15473
sa(dp30946
S'a'
I34807
sS's'
I18
sS'm'
(lp30947
I1
aI1
aI7
aI9
aI25
aI19
aI75
aI41
aI511
aI269
aI819
aI3313
aI6805
aI15051
aI4349
aI1809
aI34841
aI190641
asS'd'
I15474
sa(dp30948
S'a'
I34808
sS's'
I18
sS'm'
(lp30949
I1
aI3
aI5
aI9
aI27
aI7
aI91
aI187
aI123
aI519
aI477
aI2719
aI211
aI1225
aI22689
aI37043
aI66291
aI205441
asS'd'
I15475
sa(dp30950
S'a'
I34835
sS's'
I18
sS'm'
(lp30951
I1
aI1
aI5
aI13
aI5
aI41
aI95
aI49
aI187
aI239
aI1213
aI2363
aI8075
aI12423
aI6361
aI42471
aI70047
aI188063
asS'd'
I15476
sa(dp30952
S'a'
I34842
sS's'
I18
sS'm'
(lp30953
I1
aI3
aI7
aI3
aI27
aI23
aI21
aI217
aI65
aI143
aI1171
aI1441
aI1603
aI2235
aI20923
aI32611
aI111903
aI132771
asS'd'
I15477
sa(dp30954
S'a'
I34865
sS's'
I18
sS'm'
(lp30955
I1
aI1
aI7
aI9
aI3
aI29
aI33
aI203
aI497
aI179
aI1253
aI2083
aI7407
aI12551
aI8371
aI62167
aI93875
aI193017
asS'd'
I15478
sa(dp30956
S'a'
I34907
sS's'
I18
sS'm'
(lp30957
I1
aI1
aI1
aI13
aI7
aI61
aI43
aI107
aI417
aI757
aI1701
aI3187
aI5489
aI11359
aI20469
aI20249
aI93581
aI207969
asS'd'
I15479
sa(dp30958
S'a'
I34928
sS's'
I18
sS'm'
(lp30959
I1
aI1
aI7
aI3
aI31
aI51
aI51
aI183
aI483
aI885
aI1627
aI3605
aI6687
aI1271
aI27013
aI40409
aI103807
aI189805
asS'd'
I15480
sa(dp30960
S'a'
I34949
sS's'
I18
sS'm'
(lp30961
I1
aI1
aI3
aI1
aI21
aI21
aI107
aI185
aI267
aI981
aI147
aI1873
aI1085
aI15829
aI10315
aI21673
aI7713
aI120087
asS'd'
I15481
sa(dp30962
S'a'
I34961
sS's'
I18
sS'm'
(lp30963
I1
aI1
aI5
aI3
aI7
aI27
aI73
aI131
aI287
aI657
aI1351
aI547
aI3655
aI2433
aI6753
aI2465
aI110299
aI194587
asS'd'
I15482
sa(dp30964
S'a'
I34964
sS's'
I18
sS'm'
(lp30965
I1
aI1
aI7
aI7
aI17
aI55
aI29
aI223
aI411
aI775
aI745
aI3515
aI4573
aI4289
aI7411
aI55999
aI22021
aI110567
asS'd'
I15483
sa(dp30966
S'a'
I34987
sS's'
I18
sS'm'
(lp30967
I1
aI3
aI1
aI3
aI29
aI55
aI7
aI183
aI507
aI773
aI1299
aI3653
aI7693
aI3773
aI29549
aI4171
aI123039
aI137495
asS'd'
I15484
sa(dp30968
S'a'
I34990
sS's'
I18
sS'm'
(lp30969
I1
aI1
aI1
aI7
aI5
aI25
aI53
aI85
aI41
aI837
aI587
aI2997
aI7281
aI6821
aI15609
aI47855
aI49017
aI108557
asS'd'
I15485
sa(dp30970
S'a'
I35019
sS's'
I18
sS'm'
(lp30971
I1
aI1
aI5
aI5
aI1
aI17
aI109
aI231
aI295
aI825
aI1909
aI683
aI2197
aI1895
aI8641
aI37917
aI36347
aI38683
asS'd'
I15486
sa(dp30972
S'a'
I35069
sS's'
I18
sS'm'
(lp30973
I1
aI3
aI7
aI3
aI23
aI39
aI91
aI121
aI223
aI505
aI127
aI3439
aI7779
aI12917
aI17351
aI33063
aI84019
aI40077
asS'd'
I15487
sa(dp30974
S'a'
I35077
sS's'
I18
sS'm'
(lp30975
I1
aI3
aI3
aI1
aI19
aI1
aI125
aI99
aI143
aI549
aI709
aI3605
aI2377
aI761
aI14369
aI52191
aI80811
aI214877
asS'd'
I15488
sa(dp30976
S'a'
I35090
sS's'
I18
sS'm'
(lp30977
I1
aI1
aI7
aI9
aI13
aI57
aI39
aI91
aI505
aI299
aI1241
aI1697
aI5821
aI5327
aI22439
aI42321
aI120941
aI40009
asS'd'
I15489
sa(dp30978
S'a'
I35152
sS's'
I18
sS'm'
(lp30979
I1
aI1
aI3
aI13
aI15
aI59
aI15
aI129
aI265
aI841
aI1255
aI1915
aI4645
aI5991
aI26801
aI7839
aI66961
aI59045
asS'd'
I15490
sa(dp30980
S'a'
I35173
sS's'
I18
sS'm'
(lp30981
I1
aI3
aI7
aI15
aI17
aI57
aI61
aI173
aI391
aI1001
aI1815
aI2565
aI1445
aI13237
aI2273
aI61683
aI62327
aI180255
asS'd'
I15491
sa(dp30982
S'a'
I35174
sS's'
I18
sS'm'
(lp30983
I1
aI1
aI3
aI3
aI23
aI29
aI115
aI185
aI333
aI103
aI1807
aI3271
aI4803
aI9743
aI3031
aI25263
aI30761
aI1899
asS'd'
I15492
sa(dp30984
S'a'
I35202
sS's'
I18
sS'm'
(lp30985
I1
aI1
aI1
aI7
aI1
aI13
aI63
aI113
aI467
aI17
aI1803
aI3141
aI7069
aI8895
aI25823
aI40347
aI11211
aI88769
asS'd'
I15493
sa(dp30986
S'a'
I35214
sS's'
I18
sS'm'
(lp30987
I1
aI3
aI5
aI1
aI3
aI3
aI29
aI101
aI315
aI915
aI341
aI287
aI4167
aI7579
aI19797
aI18287
aI19079
aI52805
asS'd'
I15494
sa(dp30988
S'a'
I35219
sS's'
I18
sS'm'
(lp30989
I1
aI3
aI7
aI15
aI31
aI61
aI27
aI153
aI387
aI273
aI343
aI881
aI2273
aI6621
aI19391
aI41735
aI123899
aI117851
asS'd'
I15495
sa(dp30990
S'a'
I35226
sS's'
I18
sS'm'
(lp30991
I1
aI1
aI7
aI11
aI5
aI49
aI83
aI223
aI87
aI341
aI1023
aI2785
aI3635
aI2651
aI5179
aI25907
aI115249
aI74001
asS'd'
I15496
sa(dp30992
S'a'
I35235
sS's'
I18
sS'm'
(lp30993
I1
aI3
aI3
aI7
aI31
aI37
aI123
aI79
aI365
aI455
aI639
aI691
aI2659
aI12215
aI26785
aI48785
aI120175
aI155501
asS'd'
I15497
sa(dp30994
S'a'
I35255
sS's'
I18
sS'm'
(lp30995
I1
aI3
aI3
aI11
aI19
aI49
aI81
aI97
aI429
aI317
aI257
aI663
aI5009
aI2855
aI22721
aI51553
aI32511
aI188977
asS'd'
I15498
sa(dp30996
S'a'
I35310
sS's'
I18
sS'm'
(lp30997
I1
aI1
aI7
aI11
aI5
aI37
aI1
aI123
aI477
aI747
aI839
aI3975
aI6347
aI489
aI31387
aI56037
aI62935
aI177587
asS'd'
I15499
sa(dp30998
S'a'
I35318
sS's'
I18
sS'm'
(lp30999
I1
aI1
aI1
aI1
aI29
aI7
aI119
aI233
aI255
aI25
aI127
aI1377
aI991
aI12151
aI31259
aI64863
aI34733
aI86101
asS'd'
I15500
sa(dp31000
S'a'
I35321
sS's'
I18
sS'm'
(lp31001
I1
aI3
aI7
aI5
aI19
aI57
aI67
aI1
aI81
aI719
aI891
aI2485
aI3817
aI1055
aI437
aI9779
aI23823
aI173219
asS'd'
I15501
sa(dp31002
S'a'
I35333
sS's'
I18
sS'm'
(lp31003
I1
aI1
aI1
aI5
aI1
aI63
aI87
aI163
aI135
aI809
aI637
aI1233
aI5245
aI481
aI11011
aI23477
aI114963
aI96051
asS'd'
I15502
sa(dp31004
S'a'
I35337
sS's'
I18
sS'm'
(lp31005
I1
aI1
aI7
aI5
aI25
aI39
aI57
aI129
aI311
aI525
aI1555
aI179
aI639
aI4949
aI8809
aI31215
aI95975
aI79407
asS'd'
I15503
sa(dp31006
S'a'
I35346
sS's'
I18
sS'm'
(lp31007
I1
aI3
aI1
aI1
aI15
aI59
aI77
aI87
aI479
aI889
aI1619
aI331
aI4781
aI10597
aI935
aI28105
aI83861
aI134273
asS'd'
I15504
sa(dp31008
S'a'
I35373
sS's'
I18
sS'm'
(lp31009
I1
aI3
aI5
aI15
aI21
aI55
aI61
aI105
aI373
aI185
aI1579
aI3487
aI2621
aI8993
aI6443
aI31709
aI57329
aI128165
asS'd'
I15505
sa(dp31010
S'a'
I35414
sS's'
I18
sS'm'
(lp31011
I1
aI3
aI3
aI3
aI7
aI21
aI117
aI159
aI177
aI927
aI1873
aI1865
aI3219
aI1693
aI1173
aI34365
aI107053
aI113949
asS'd'
I15506
sa(dp31012
S'a'
I35478
sS's'
I18
sS'm'
(lp31013
I1
aI3
aI7
aI1
aI17
aI37
aI35
aI101
aI305
aI141
aI1681
aI1949
aI47
aI11351
aI989
aI13887
aI127429
aI13059
asS'd'
I15507
sa(dp31014
S'a'
I35497
sS's'
I18
sS'm'
(lp31015
I1
aI1
aI3
aI7
aI13
aI41
aI125
aI115
aI65
aI621
aI1401
aI631
aI5875
aI8589
aI17185
aI22757
aI83625
aI92907
asS'd'
I15508
sa(dp31016
S'a'
I35503
sS's'
I18
sS'm'
(lp31017
I1
aI3
aI1
aI5
aI25
aI37
aI73
aI39
aI495
aI645
aI265
aI2685
aI5875
aI5919
aI23223
aI44593
aI26207
aI49921
asS'd'
I15509
sa(dp31018
S'a'
I35512
sS's'
I18
sS'm'
(lp31019
I1
aI3
aI1
aI13
aI25
aI31
aI39
aI15
aI179
aI791
aI1817
aI3617
aI2139
aI1827
aI21215
aI21767
aI15009
aI239443
asS'd'
I15510
sa(dp31020
S'a'
I35515
sS's'
I18
sS'm'
(lp31021
I1
aI1
aI3
aI7
aI17
aI9
aI33
aI121
aI235
aI535
aI1537
aI3307
aI2881
aI4351
aI4721
aI34131
aI129619
aI137993
asS'd'
I15511
sa(dp31022
S'a'
I35526
sS's'
I18
sS'm'
(lp31023
I1
aI1
aI3
aI1
aI3
aI51
aI79
aI213
aI205
aI323
aI1749
aI2563
aI2013
aI6519
aI18923
aI25937
aI74445
aI252283
asS'd'
I15512
sa(dp31024
S'a'
I35577
sS's'
I18
sS'm'
(lp31025
I1
aI3
aI5
aI11
aI3
aI53
aI17
aI195
aI305
aI543
aI1937
aI2997
aI4593
aI7801
aI15307
aI46359
aI39365
aI59537
asS'd'
I15513
sa(dp31026
S'a'
I35585
sS's'
I18
sS'm'
(lp31027
I1
aI1
aI1
aI13
aI31
aI53
aI111
aI163
aI139
aI163
aI999
aI83
aI5125
aI10047
aI11143
aI51407
aI13627
aI3621
asS'd'
I15514
sa(dp31028
S'a'
I35592
sS's'
I18
sS'm'
(lp31029
I1
aI1
aI3
aI9
aI5
aI1
aI125
aI95
aI281
aI939
aI913
aI1441
aI1209
aI12983
aI27759
aI22393
aI75985
aI178997
asS'd'
I15515
sa(dp31030
S'a'
I35615
sS's'
I18
sS'm'
(lp31031
I1
aI3
aI5
aI3
aI5
aI27
aI91
aI41
aI51
aI447
aI491
aI3405
aI497
aI2873
aI17865
aI30651
aI104197
aI71751
asS'd'
I15516
sa(dp31032
S'a'
I35616
sS's'
I18
sS'm'
(lp31033
I1
aI3
aI7
aI1
aI29
aI61
aI73
aI31
aI423
aI933
aI1327
aI809
aI1461
aI269
aI15121
aI18649
aI36095
aI139429
asS'd'
I15517
sa(dp31034
S'a'
I35622
sS's'
I18
sS'm'
(lp31035
I1
aI1
aI7
aI7
aI19
aI49
aI51
aI173
aI297
aI411
aI1255
aI1093
aI2821
aI6743
aI1927
aI27563
aI68459
aI261411
asS'd'
I15518
sa(dp31036
S'a'
I35634
sS's'
I18
sS'm'
(lp31037
I1
aI3
aI5
aI1
aI5
aI33
aI27
aI119
aI103
aI615
aI149
aI2299
aI6801
aI15615
aI7361
aI31045
aI87719
aI9557
asS'd'
I15519
sa(dp31038
S'a'
I35657
sS's'
I18
sS'm'
(lp31039
I1
aI1
aI3
aI9
aI17
aI23
aI89
aI35
aI151
aI385
aI319
aI2065
aI1897
aI1987
aI15159
aI34855
aI5395
aI110751
asS'd'
I15520
sa(dp31040
S'a'
I35672
sS's'
I18
sS'm'
(lp31041
I1
aI3
aI1
aI13
aI7
aI47
aI19
aI185
aI207
aI787
aI1179
aI1073
aI1463
aI6277
aI6129
aI25031
aI91969
aI123235
asS'd'
I15521
sa(dp31042
S'a'
I35708
sS's'
I18
sS'm'
(lp31043
I1
aI1
aI7
aI3
aI19
aI63
aI97
aI1
aI381
aI71
aI1169
aI339
aI6585
aI3629
aI31357
aI59451
aI102735
aI253667
asS'd'
I15522
sa(dp31044
S'a'
I35772
sS's'
I18
sS'm'
(lp31045
I1
aI3
aI3
aI13
aI9
aI27
aI69
aI17
aI509
aI599
aI1247
aI2267
aI3309
aI1905
aI17995
aI41263
aI5947
aI51607
asS'd'
I15523
sa(dp31046
S'a'
I35790
sS's'
I18
sS'm'
(lp31047
I1
aI3
aI1
aI9
aI31
aI45
aI69
aI243
aI171
aI555
aI61
aI1135
aI1993
aI8615
aI18363
aI19545
aI64015
aI81391
asS'd'
I15524
sa(dp31048
S'a'
I35804
sS's'
I18
sS'm'
(lp31049
I1
aI1
aI1
aI1
aI19
aI49
aI31
aI65
aI53
aI123
aI271
aI3007
aI4509
aI9465
aI3855
aI12673
aI19457
aI14677
asS'd'
I15525
sa(dp31050
S'a'
I35811
sS's'
I18
sS'm'
(lp31051
I1
aI3
aI5
aI13
aI29
aI53
aI91
aI145
aI115
aI53
aI839
aI1911
aI2887
aI6053
aI18437
aI42273
aI63093
aI70937
asS'd'
I15526
sa(dp31052
S'a'
I35814
sS's'
I18
sS'm'
(lp31053
I1
aI1
aI5
aI5
aI27
aI13
aI87
aI175
aI485
aI699
aI463
aI811
aI4991
aI15303
aI23007
aI10021
aI59125
aI39997
asS'd'
I15527
sa(dp31054
S'a'
I35837
sS's'
I18
sS'm'
(lp31055
I1
aI1
aI5
aI5
aI27
aI21
aI89
aI61
aI109
aI555
aI953
aI2811
aI3015
aI3249
aI16085
aI64413
aI84605
aI177333
asS'd'
I15528
sa(dp31056
S'a'
I35846
sS's'
I18
sS'm'
(lp31057
I1
aI1
aI7
aI3
aI1
aI29
aI83
aI143
aI67
aI577
aI971
aI2339
aI6521
aI6341
aI27141
aI37149
aI99813
aI37579
asS'd'
I15529
sa(dp31058
S'a'
I35873
sS's'
I18
sS'm'
(lp31059
I1
aI3
aI3
aI7
aI23
aI29
aI117
aI5
aI287
aI559
aI667
aI2349
aI7481
aI679
aI9633
aI40857
aI89841
aI98125
asS'd'
I15530
sa(dp31060
S'a'
I35883
sS's'
I18
sS'm'
(lp31061
I1
aI1
aI1
aI3
aI31
aI31
aI83
aI117
aI213
aI213
aI23
aI3803
aI5967
aI7759
aI19521
aI13229
aI62231
aI150687
asS'd'
I15531
sa(dp31062
S'a'
I35918
sS's'
I18
sS'm'
(lp31063
I1
aI3
aI7
aI9
aI1
aI15
aI37
aI191
aI19
aI107
aI1723
aI3517
aI3477
aI3777
aI4359
aI45815
aI58661
aI33217
asS'd'
I15532
sa(dp31064
S'a'
I35920
sS's'
I18
sS'm'
(lp31065
I1
aI1
aI5
aI1
aI17
aI3
aI3
aI255
aI501
aI1021
aI1731
aI481
aI6145
aI3475
aI3417
aI11847
aI92203
aI75109
asS'd'
I15533
sa(dp31066
S'a'
I35925
sS's'
I18
sS'm'
(lp31067
I1
aI1
aI5
aI1
aI1
aI61
aI89
aI107
aI503
aI627
aI931
aI1355
aI2067
aI12487
aI20665
aI61543
aI15501
aI103843
asS'd'
I15534
sa(dp31068
S'a'
I35926
sS's'
I18
sS'm'
(lp31069
I1
aI1
aI5
aI9
aI25
aI17
aI7
aI255
aI251
aI939
aI851
aI2241
aI939
aI15331
aI29357
aI2485
aI80791
aI152601
asS'd'
I15535
sa(dp31070
S'a'
I35945
sS's'
I18
sS'm'
(lp31071
I1
aI1
aI5
aI3
aI13
aI25
aI35
aI113
aI83
aI765
aI1317
aI1409
aI369
aI2215
aI5659
aI3581
aI13925
aI108673
asS'd'
I15536
sa(dp31072
S'a'
I35956
sS's'
I18
sS'm'
(lp31073
I1
aI1
aI1
aI1
aI13
aI13
aI83
aI27
aI5
aI563
aI723
aI2733
aI3155
aI6567
aI24595
aI45569
aI37587
aI144401
asS'd'
I15537
sa(dp31074
S'a'
I36003
sS's'
I18
sS'm'
(lp31075
I1
aI1
aI1
aI9
aI31
aI51
aI73
aI105
aI299
aI857
aI669
aI963
aI4115
aI14939
aI11669
aI46215
aI92707
aI149249
asS'd'
I15538
sa(dp31076
S'a'
I36023
sS's'
I18
sS'm'
(lp31077
I1
aI1
aI3
aI5
aI7
aI41
aI105
aI213
aI3
aI999
aI93
aI1497
aI6783
aI1559
aI20047
aI40761
aI88219
aI64769
asS'd'
I15539
sa(dp31078
S'a'
I36024
sS's'
I18
sS'm'
(lp31079
I1
aI3
aI1
aI5
aI13
aI17
aI79
aI29
aI453
aI75
aI1095
aI623
aI7401
aI4225
aI30467
aI60795
aI130045
aI154767
asS'd'
I15540
sa(dp31080
S'a'
I36059
sS's'
I18
sS'm'
(lp31081
I1
aI1
aI5
aI3
aI31
aI59
aI33
aI129
aI505
aI277
aI1623
aI3531
aI6841
aI12903
aI7231
aI5801
aI92405
aI260195
asS'd'
I15541
sa(dp31082
S'a'
I36061
sS's'
I18
sS'm'
(lp31083
I1
aI1
aI3
aI5
aI27
aI23
aI63
aI219
aI225
aI547
aI1163
aI1899
aI4191
aI9725
aI30077
aI30157
aI73395
aI38195
asS'd'
I15542
sa(dp31084
S'a'
I36072
sS's'
I18
sS'm'
(lp31085
I1
aI1
aI1
aI11
aI17
aI27
aI63
aI127
aI95
aI205
aI1753
aI2023
aI6803
aI4355
aI28169
aI16691
aI25105
aI8969
asS'd'
I15543
sa(dp31086
S'a'
I36075
sS's'
I18
sS'm'
(lp31087
I1
aI1
aI5
aI3
aI23
aI23
aI89
aI115
aI231
aI647
aI513
aI3161
aI3175
aI5061
aI5797
aI35387
aI109827
aI19511
asS'd'
I15544
sa(dp31088
S'a'
I36103
sS's'
I18
sS'm'
(lp31089
I1
aI3
aI5
aI5
aI11
aI9
aI39
aI251
aI367
aI253
aI2031
aI3909
aI1057
aI12545
aI25397
aI51571
aI91229
aI83721
asS'd'
I15545
sa(dp31090
S'a'
I36110
sS's'
I18
sS'm'
(lp31091
I1
aI3
aI5
aI7
aI5
aI35
aI57
aI153
aI111
aI789
aI177
aI2237
aI1333
aI13185
aI993
aI22099
aI62113
aI211815
asS'd'
I15546
sa(dp31092
S'a'
I36131
sS's'
I18
sS'm'
(lp31093
I1
aI1
aI5
aI15
aI19
aI37
aI123
aI221
aI375
aI605
aI791
aI1663
aI7537
aI7193
aI20149
aI58077
aI113129
aI185493
asS'd'
I15547
sa(dp31094
S'a'
I36151
sS's'
I18
sS'm'
(lp31095
I1
aI1
aI1
aI1
aI1
aI53
aI117
aI227
aI441
aI851
aI1171
aI4031
aI2313
aI2847
aI25533
aI31767
aI18197
aI153101
asS'd'
I15548
sa(dp31096
S'a'
I36160
sS's'
I18
sS'm'
(lp31097
I1
aI1
aI3
aI3
aI13
aI9
aI65
aI225
aI71
aI763
aI1507
aI3795
aI4321
aI399
aI12515
aI4527
aI89193
aI236161
asS'd'
I15549
sa(dp31098
S'a'
I36199
sS's'
I18
sS'm'
(lp31099
I1
aI1
aI3
aI11
aI21
aI63
aI73
aI125
aI369
aI309
aI953
aI3525
aI3925
aI13609
aI26061
aI21739
aI112867
aI112985
asS'd'
I15550
sa(dp31100
S'a'
I36223
sS's'
I18
sS'm'
(lp31101
I1
aI1
aI7
aI1
aI27
aI25
aI3
aI129
aI321
aI193
aI1871
aI233
aI837
aI11163
aI14861
aI42721
aI72849
aI206739
asS'd'
I15551
sa(dp31102
S'a'
I36227
sS's'
I18
sS'm'
(lp31103
I1
aI3
aI7
aI3
aI5
aI51
aI43
aI177
aI167
aI11
aI1297
aI1805
aI515
aI6485
aI8253
aI271
aI47435
aI252291
asS'd'
I15552
sa(dp31104
S'a'
I36234
sS's'
I18
sS'm'
(lp31105
I1
aI3
aI3
aI3
aI19
aI47
aI47
aI47
aI299
aI101
aI1535
aI3593
aI4669
aI10367
aI19219
aI16579
aI85269
aI36115
asS'd'
I15553
sa(dp31106
S'a'
I36236
sS's'
I18
sS'm'
(lp31107
I1
aI1
aI7
aI15
aI7
aI51
aI53
aI181
aI379
aI267
aI985
aI3401
aI2189
aI10197
aI14183
aI413
aI76797
aI24751
asS'd'
I15554
sa(dp31108
S'a'
I36284
sS's'
I18
sS'm'
(lp31109
I1
aI1
aI5
aI7
aI13
aI27
aI65
aI119
aI235
aI131
aI1921
aI3411
aI1511
aI11221
aI30315
aI11799
aI127563
aI203533
asS'd'
I15555
sa(dp31110
S'a'
I36319
sS's'
I18
sS'm'
(lp31111
I1
aI3
aI1
aI3
aI13
aI55
aI101
aI189
aI483
aI261
aI467
aI645
aI417
aI6203
aI9221
aI19671
aI102331
aI259335
asS'd'
I15556
sa(dp31112
S'a'
I36332
sS's'
I18
sS'm'
(lp31113
I1
aI1
aI5
aI15
aI19
aI7
aI81
aI1
aI371
aI119
aI1433
aI1211
aI303
aI14393
aI27107
aI45295
aI109211
aI224661
asS'd'
I15557
sa(dp31114
S'a'
I36343
sS's'
I18
sS'm'
(lp31115
I1
aI3
aI7
aI9
aI19
aI53
aI31
aI55
aI103
aI351
aI1511
aI377
aI981
aI6709
aI19553
aI53579
aI55043
aI170489
asS'd'
I15558
sa(dp31116
S'a'
I36373
sS's'
I18
sS'm'
(lp31117
I1
aI3
aI3
aI15
aI31
aI49
aI1
aI251
aI187
aI73
aI119
aI3041
aI5455
aI5355
aI22245
aI7735
aI14661
aI258447
asS'd'
I15559
sa(dp31118
S'a'
I36401
sS's'
I18
sS'm'
(lp31119
I1
aI3
aI7
aI11
aI13
aI1
aI61
aI97
aI179
aI975
aI1653
aI3301
aI4303
aI2271
aI30171
aI63287
aI51271
aI21909
asS'd'
I15560
sa(dp31120
S'a'
I36413
sS's'
I18
sS'm'
(lp31121
I1
aI1
aI5
aI11
aI21
aI45
aI101
aI131
aI121
aI881
aI1205
aI1849
aI4337
aI5687
aI31967
aI22559
aI98017
aI202557
asS'd'
I15561
sa(dp31122
S'a'
I36433
sS's'
I18
sS'm'
(lp31123
I1
aI3
aI3
aI7
aI1
aI49
aI11
aI35
aI141
aI309
aI651
aI3319
aI4313
aI3675
aI27699
aI49429
aI109805
aI167089
asS'd'
I15562
sa(dp31124
S'a'
I36459
sS's'
I18
sS'm'
(lp31125
I1
aI1
aI3
aI13
aI13
aI15
aI61
aI251
aI335
aI365
aI677
aI2183
aI6291
aI8857
aI15231
aI551
aI63149
aI76729
asS'd'
I15563
sa(dp31126
S'a'
I36480
sS's'
I18
sS'm'
(lp31127
I1
aI3
aI3
aI13
aI1
aI59
aI85
aI127
aI409
aI1007
aI1947
aI3495
aI6227
aI11447
aI14329
aI3769
aI109619
aI59063
asS'd'
I15564
sa(dp31128
S'a'
I36485
sS's'
I18
sS'm'
(lp31129
I1
aI3
aI5
aI11
aI11
aI59
aI67
aI209
aI491
aI757
aI1137
aI1977
aI3155
aI9339
aI11219
aI20303
aI66417
aI187911
asS'd'
I15565
sa(dp31130
S'a'
I36510
sS's'
I18
sS'm'
(lp31131
I1
aI1
aI5
aI9
aI27
aI51
aI87
aI249
aI327
aI867
aI29
aI3811
aI4769
aI12353
aI24803
aI35747
aI84101
aI210975
asS'd'
I15566
sa(dp31132
S'a'
I36513
sS's'
I18
sS'm'
(lp31133
I1
aI3
aI7
aI7
aI23
aI37
aI23
aI55
aI237
aI543
aI779
aI1305
aI1535
aI13333
aI1403
aI10903
aI113135
aI195799
asS'd'
I15567
sa(dp31134
S'a'
I36523
sS's'
I18
sS'm'
(lp31135
I1
aI1
aI3
aI11
aI1
aI1
aI3
aI153
aI401
aI35
aI981
aI79
aI4227
aI9203
aI8179
aI29325
aI104809
aI140613
asS'd'
I15568
sa(dp31136
S'a'
I36528
sS's'
I18
sS'm'
(lp31137
I1
aI3
aI5
aI9
aI13
aI39
aI101
aI181
aI507
aI307
aI1411
aI1443
aI6855
aI8747
aI22709
aI37869
aI102303
aI577
asS'd'
I15569
sa(dp31138
S'a'
I36537
sS's'
I18
sS'm'
(lp31139
I1
aI3
aI5
aI1
aI25
aI41
aI3
aI239
aI195
aI1
aI1277
aI2085
aI4253
aI14683
aI24171
aI56733
aI82795
aI213291
asS'd'
I15570
sa(dp31140
S'a'
I36558
sS's'
I18
sS'm'
(lp31141
I1
aI1
aI3
aI5
aI25
aI55
aI31
aI55
aI215
aI149
aI1813
aI3775
aI779
aI6137
aI10561
aI41671
aI96883
aI177435
asS'd'
I15571
sa(dp31142
S'a'
I36563
sS's'
I18
sS'm'
(lp31143
I1
aI1
aI5
aI11
aI15
aI5
aI1
aI237
aI131
aI13
aI229
aI3203
aI2431
aI1829
aI31983
aI59535
aI31381
aI175455
asS'd'
I15572
sa(dp31144
S'a'
I36576
sS's'
I18
sS'm'
(lp31145
I1
aI3
aI3
aI7
aI5
aI19
aI61
aI253
aI223
aI609
aI1395
aI2495
aI5501
aI6571
aI12989
aI889
aI49435
aI200251
asS'd'
I15573
sa(dp31146
S'a'
I36608
sS's'
I18
sS'm'
(lp31147
I1
aI1
aI7
aI13
aI25
aI49
aI33
aI157
aI457
aI259
aI1935
aI2249
aI7419
aI12685
aI30509
aI32187
aI108839
aI178963
asS'd'
I15574
sa(dp31148
S'a'
I36611
sS's'
I18
sS'm'
(lp31149
I1
aI3
aI3
aI15
aI19
aI27
aI91
aI133
aI369
aI931
aI359
aI759
aI2647
aI13643
aI14877
aI14031
aI115367
aI201745
asS'd'
I15575
sa(dp31150
S'a'
I36617
sS's'
I18
sS'm'
(lp31151
I1
aI1
aI5
aI3
aI9
aI23
aI87
aI27
aI203
aI995
aI1759
aI999
aI949
aI2733
aI29053
aI46581
aI129003
aI42585
asS'd'
I15576
sa(dp31152
S'a'
I36653
sS's'
I18
sS'm'
(lp31153
I1
aI1
aI3
aI1
aI1
aI21
aI63
aI243
aI257
aI741
aI681
aI2471
aI2455
aI15145
aI31739
aI8751
aI15963
aI165405
asS'd'
I15577
sa(dp31154
S'a'
I36716
sS's'
I18
sS'm'
(lp31155
I1
aI3
aI3
aI1
aI25
aI21
aI69
aI213
aI219
aI9
aI199
aI487
aI4103
aI141
aI18177
aI58797
aI60415
aI6313
asS'd'
I15578
sa(dp31156
S'a'
I36721
sS's'
I18
sS'm'
(lp31157
I1
aI3
aI1
aI5
aI23
aI43
aI61
aI121
aI123
aI89
aI283
aI1313
aI2707
aI10199
aI7699
aI17437
aI130995
aI140327
asS'd'
I15579
sa(dp31158
S'a'
I36733
sS's'
I18
sS'm'
(lp31159
I1
aI3
aI5
aI13
aI31
aI41
aI111
aI39
aI403
aI5
aI1125
aI2867
aI3143
aI7051
aI9891
aI43349
aI20751
aI88465
asS'd'
I15580
sa(dp31160
S'a'
I36761
sS's'
I18
sS'm'
(lp31161
I1
aI1
aI3
aI1
aI19
aI53
aI83
aI207
aI285
aI789
aI1515
aI3455
aI4057
aI15777
aI27879
aI46971
aI122661
aI143407
asS'd'
I15581
sa(dp31162
S'a'
I36783
sS's'
I18
sS'm'
(lp31163
I1
aI3
aI3
aI11
aI25
aI21
aI127
aI191
aI313
aI357
aI1625
aI1323
aI1151
aI12509
aI22275
aI23517
aI12221
aI258709
asS'd'
I15582
sa(dp31164
S'a'
I36786
sS's'
I18
sS'm'
(lp31165
I1
aI1
aI5
aI7
aI1
aI47
aI1
aI107
aI387
aI965
aI1319
aI2911
aI2121
aI8595
aI9
aI21587
aI81187
aI2803
asS'd'
I15583
sa(dp31166
S'a'
I36795
sS's'
I18
sS'm'
(lp31167
I1
aI3
aI3
aI5
aI19
aI55
aI37
aI213
aI23
aI767
aI1493
aI635
aI4289
aI2503
aI16835
aI47851
aI77335
aI60565
asS'd'
I15584
sa(dp31168
S'a'
I36800
sS's'
I18
sS'm'
(lp31169
I1
aI1
aI1
aI7
aI23
aI9
aI101
aI145
aI457
aI691
aI1895
aI2145
aI7527
aI7687
aI20781
aI10957
aI24859
aI79137
asS'd'
I15585
sa(dp31170
S'a'
I36810
sS's'
I18
sS'm'
(lp31171
I1
aI3
aI7
aI15
aI9
aI9
aI15
aI195
aI493
aI859
aI687
aI1445
aI429
aI8599
aI24591
aI40709
aI118361
aI148163
asS'd'
I15586
sa(dp31172
S'a'
I36812
sS's'
I18
sS'm'
(lp31173
I1
aI1
aI1
aI3
aI7
aI51
aI45
aI143
aI339
aI475
aI1177
aI2829
aI785
aI10141
aI4923
aI29135
aI22603
aI119973
asS'd'
I15587
sa(dp31174
S'a'
I36817
sS's'
I18
sS'm'
(lp31175
I1
aI3
aI5
aI15
aI25
aI37
aI1
aI13
aI351
aI127
aI143
aI2637
aI1215
aI14577
aI12465
aI10575
aI67997
aI21877
asS'd'
I15588
sa(dp31176
S'a'
I36818
sS's'
I18
sS'm'
(lp31177
I1
aI3
aI7
aI3
aI27
aI19
aI59
aI241
aI327
aI307
aI731
aI3471
aI6123
aI13607
aI8793
aI14825
aI110681
aI83259
asS'd'
I15589
sa(dp31178
S'a'
I36851
sS's'
I18
sS'm'
(lp31179
I1
aI1
aI1
aI11
aI25
aI5
aI59
aI85
aI335
aI189
aI499
aI1305
aI5801
aI7259
aI2397
aI14045
aI55585
aI258579
asS'd'
I15590
sa(dp31180
S'a'
I36853
sS's'
I18
sS'm'
(lp31181
I1
aI1
aI3
aI5
aI21
aI49
aI49
aI63
aI261
aI657
aI1453
aI55
aI1325
aI15513
aI14891
aI60689
aI15381
aI252641
asS'd'
I15591
sa(dp31182
S'a'
I36868
sS's'
I18
sS'm'
(lp31183
I1
aI1
aI7
aI15
aI25
aI3
aI85
aI33
aI495
aI867
aI903
aI1813
aI2871
aI365
aI17399
aI45695
aI102851
aI225873
asS'd'
I15592
sa(dp31184
S'a'
I36889
sS's'
I18
sS'm'
(lp31185
I1
aI1
aI1
aI13
aI13
aI63
aI29
aI35
aI325
aI893
aI1313
aI133
aI8169
aI7791
aI9271
aI36759
aI92275
aI169717
asS'd'
I15593
sa(dp31186
S'a'
I36890
sS's'
I18
sS'm'
(lp31187
I1
aI1
aI7
aI3
aI21
aI45
aI7
aI151
aI387
aI891
aI1921
aI1701
aI307
aI5323
aI16321
aI51229
aI79369
aI21513
asS'd'
I15594
sa(dp31188
S'a'
I36896
sS's'
I18
sS'm'
(lp31189
I1
aI1
aI1
aI11
aI7
aI27
aI17
aI75
aI161
aI649
aI337
aI1731
aI2905
aI4589
aI17387
aI10455
aI70673
aI228373
asS'd'
I15595
sa(dp31190
S'a'
I36905
sS's'
I18
sS'm'
(lp31191
I1
aI1
aI3
aI15
aI17
aI35
aI45
aI131
aI469
aI629
aI1771
aI1965
aI5065
aI6249
aI29041
aI52791
aI55619
aI154531
asS'd'
I15596
sa(dp31192
S'a'
I36919
sS's'
I18
sS'm'
(lp31193
I1
aI1
aI7
aI3
aI25
aI53
aI85
aI233
aI161
aI163
aI1155
aI3159
aI1551
aI13099
aI25647
aI26777
aI91647
aI162755
asS'd'
I15597
sa(dp31194
S'a'
I36938
sS's'
I18
sS'm'
(lp31195
I1
aI1
aI3
aI9
aI17
aI11
aI39
aI63
aI503
aI927
aI1621
aI3425
aI4835
aI7083
aI16449
aI47913
aI127905
aI165567
asS'd'
I15598
sa(dp31196
S'a'
I36946
sS's'
I18
sS'm'
(lp31197
I1
aI1
aI7
aI1
aI29
aI9
aI1
aI111
aI285
aI1009
aI1427
aI3071
aI205
aI12269
aI31337
aI14501
aI10923
aI14277
asS'd'
I15599
sa(dp31198
S'a'
I36951
sS's'
I18
sS'm'
(lp31199
I1
aI1
aI5
aI5
aI1
aI3
aI51
aI205
aI477
aI661
aI1555
aI2113
aI6487
aI4755
aI13633
aI16391
aI35775
aI52623
asS'd'
I15600
sa(dp31200
S'a'
I36952
sS's'
I18
sS'm'
(lp31201
I1
aI3
aI3
aI3
aI27
aI23
aI109
aI49
aI71
aI19
aI733
aI1361
aI4369
aI14527
aI20443
aI10507
aI120183
aI246115
asS'd'
I15601
sa(dp31202
S'a'
I36964
sS's'
I18
sS'm'
(lp31203
I1
aI3
aI3
aI5
aI7
aI47
aI51
aI197
aI97
aI471
aI1631
aI3317
aI5857
aI9405
aI30359
aI7741
aI45079
aI175929
asS'd'
I15602
sa(dp31204
S'a'
I36968
sS's'
I18
sS'm'
(lp31205
I1
aI1
aI3
aI3
aI13
aI63
aI39
aI173
aI511
aI525
aI1687
aI1735
aI6877
aI7383
aI27971
aI26503
aI6189
aI232251
asS'd'
I15603
sa(dp31206
S'a'
I36979
sS's'
I18
sS'm'
(lp31207
I1
aI1
aI5
aI3
aI5
aI31
aI101
aI99
aI51
aI987
aI1627
aI3899
aI6321
aI9441
aI4983
aI64001
aI30923
aI199495
asS'd'
I15604
sa(dp31208
S'a'
I36981
sS's'
I18
sS'm'
(lp31209
I1
aI3
aI1
aI1
aI11
aI39
aI119
aI123
aI33
aI1017
aI1477
aI283
aI4939
aI453
aI16445
aI25599
aI106857
aI257811
asS'd'
I15605
sa(dp31210
S'a'
I37021
sS's'
I18
sS'm'
(lp31211
I1
aI3
aI1
aI11
aI13
aI1
aI3
aI101
aI275
aI75
aI1795
aI1449
aI2503
aI11765
aI19299
aI14237
aI157
aI244825
asS'd'
I15606
sa(dp31212
S'a'
I37026
sS's'
I18
sS'm'
(lp31213
I1
aI3
aI7
aI15
aI23
aI1
aI85
aI65
aI55
aI103
aI1523
aI1443
aI1021
aI5733
aI3297
aI10889
aI22487
aI82503
asS'd'
I15607
sa(dp31214
S'a'
I37075
sS's'
I18
sS'm'
(lp31215
I1
aI1
aI7
aI3
aI17
aI59
aI109
aI113
aI17
aI173
aI1709
aI273
aI5327
aI3243
aI10751
aI58361
aI42303
aI78391
asS'd'
I15608
sa(dp31216
S'a'
I37077
sS's'
I18
sS'm'
(lp31217
I1
aI1
aI5
aI15
aI25
aI11
aI101
aI133
aI193
aI131
aI1671
aI3045
aI7111
aI14331
aI15665
aI56407
aI31561
aI154555
asS'd'
I15609
sa(dp31218
S'a'
I37108
sS's'
I18
sS'm'
(lp31219
I1
aI3
aI3
aI5
aI15
aI41
aI105
aI65
aI81
aI293
aI389
aI2653
aI1883
aI14741
aI23553
aI33349
aI39665
aI154233
asS'd'
I15610
sa(dp31220
S'a'
I37112
sS's'
I18
sS'm'
(lp31221
I1
aI1
aI5
aI15
aI31
aI19
aI121
aI41
aI261
aI511
aI1679
aI957
aI1647
aI12647
aI12285
aI30291
aI122483
aI187911
asS'd'
I15611
sa(dp31222
S'a'
I37150
sS's'
I18
sS'm'
(lp31223
I1
aI3
aI1
aI5
aI27
aI25
aI17
aI45
aI303
aI947
aI1123
aI2729
aI281
aI12389
aI27987
aI42667
aI16089
aI17129
asS'd'
I15612
sa(dp31224
S'a'
I37154
sS's'
I18
sS'm'
(lp31225
I1
aI3
aI7
aI15
aI13
aI17
aI25
aI223
aI125
aI837
aI159
aI253
aI2599
aI11381
aI639
aI32545
aI50633
aI139025
asS'd'
I15613
sa(dp31226
S'a'
I37156
sS's'
I18
sS'm'
(lp31227
I1
aI3
aI1
aI13
aI23
aI43
aI25
aI83
aI507
aI47
aI99
aI697
aI4453
aI2139
aI17151
aI50709
aI37099
aI212957
asS'd'
I15614
sa(dp31228
S'a'
I37163
sS's'
I18
sS'm'
(lp31229
I1
aI1
aI7
aI7
aI29
aI7
aI63
aI141
aI475
aI853
aI1073
aI143
aI6979
aI5663
aI29691
aI59489
aI89689
aI223047
asS'd'
I15615
sa(dp31230
S'a'
I37178
sS's'
I18
sS'm'
(lp31231
I1
aI1
aI1
aI1
aI13
aI27
aI101
aI61
aI27
aI735
aI207
aI2065
aI5811
aI5461
aI21493
aI15481
aI103727
aI80017
asS'd'
I15616
sa(dp31232
S'a'
I37183
sS's'
I18
sS'm'
(lp31233
I1
aI3
aI1
aI11
aI9
aI9
aI35
aI251
aI147
aI841
aI1891
aI1909
aI5053
aI5103
aI11751
aI16209
aI110475
aI114875
asS'd'
I15617
sa(dp31234
S'a'
I37185
sS's'
I18
sS'm'
(lp31235
I1
aI3
aI3
aI11
aI13
aI55
aI117
aI205
aI71
aI159
aI1797
aI989
aI2221
aI16087
aI18287
aI8355
aI96403
aI146613
asS'd'
I15618
sa(dp31236
S'a'
I37191
sS's'
I18
sS'm'
(lp31237
I1
aI1
aI5
aI5
aI29
aI25
aI73
aI63
aI299
aI839
aI1225
aI3583
aI5641
aI1341
aI29431
aI7035
aI99107
aI13493
asS'd'
I15619
sa(dp31238
S'a'
I37198
sS's'
I18
sS'm'
(lp31239
I1
aI1
aI3
aI5
aI27
aI53
aI9
aI51
aI79
aI701
aI667
aI1469
aI4455
aI13761
aI18607
aI39429
aI7687
aI201115
asS'd'
I15620
sa(dp31240
S'a'
I37203
sS's'
I18
sS'm'
(lp31241
I1
aI3
aI7
aI11
aI23
aI35
aI101
aI129
aI491
aI369
aI565
aI2557
aI2529
aI1003
aI16003
aI33873
aI52155
aI861
asS'd'
I15621
sa(dp31242
S'a'
I37225
sS's'
I18
sS'm'
(lp31243
I1
aI1
aI1
aI15
aI27
aI63
aI1
aI55
aI331
aI853
aI899
aI1027
aI7389
aI8935
aI12559
aI27315
aI101753
aI255331
asS'd'
I15622
sa(dp31244
S'a'
I37243
sS's'
I18
sS'm'
(lp31245
I1
aI3
aI3
aI15
aI5
aI41
aI93
aI39
aI473
aI887
aI1667
aI847
aI7619
aI8407
aI6539
aI31989
aI63807
aI21861
asS'd'
I15623
sa(dp31246
S'a'
I37252
sS's'
I18
sS'm'
(lp31247
I1
aI1
aI5
aI11
aI27
aI57
aI73
aI249
aI331
aI653
aI21
aI2937
aI4403
aI16195
aI18785
aI30375
aI22939
aI235291
asS'd'
I15624
sa(dp31248
S'a'
I37262
sS's'
I18
sS'm'
(lp31249
I1
aI1
aI7
aI1
aI7
aI41
aI59
aI161
aI295
aI503
aI595
aI3021
aI455
aI3991
aI8617
aI65361
aI107239
aI83205
asS'd'
I15625
sa(dp31250
S'a'
I37264
sS's'
I18
sS'm'
(lp31251
I1
aI3
aI3
aI15
aI17
aI41
aI61
aI229
aI273
aI687
aI657
aI1969
aI2817
aI2367
aI29183
aI41199
aI24123
aI184081
asS'd'
I15626
sa(dp31252
S'a'
I37276
sS's'
I18
sS'm'
(lp31253
I1
aI3
aI7
aI5
aI25
aI63
aI43
aI65
aI443
aI423
aI549
aI2031
aI3353
aI7041
aI6563
aI18819
aI46047
aI239823
asS'd'
I15627
sa(dp31254
S'a'
I37327
sS's'
I18
sS'm'
(lp31255
I1
aI3
aI3
aI3
aI3
aI17
aI13
aI115
aI377
aI623
aI1959
aI127
aI5125
aI13209
aI24731
aI23151
aI21303
aI7213
asS'd'
I15628
sa(dp31256
S'a'
I37355
sS's'
I18
sS'm'
(lp31257
I1
aI1
aI7
aI1
aI21
aI11
aI21
aI41
aI491
aI37
aI1759
aI2771
aI1301
aI12995
aI17621
aI30907
aI75511
aI82321
asS'd'
I15629
sa(dp31258
S'a'
I37403
sS's'
I18
sS'm'
(lp31259
I1
aI3
aI3
aI13
aI13
aI23
aI77
aI211
aI215
aI711
aI427
aI2213
aI8041
aI1595
aI26105
aI39051
aI105407
aI242141
asS'd'
I15630
sa(dp31260
S'a'
I37415
sS's'
I18
sS'm'
(lp31261
I1
aI3
aI3
aI9
aI13
aI35
aI117
aI207
aI75
aI395
aI723
aI3321
aI6643
aI2429
aI10043
aI10585
aI3529
aI64067
asS'd'
I15631
sa(dp31262
S'a'
I37422
sS's'
I18
sS'm'
(lp31263
I1
aI1
aI1
aI7
aI3
aI1
aI83
aI93
aI311
aI157
aI891
aI1717
aI7669
aI16067
aI11775
aI27693
aI11757
aI94471
asS'd'
I15632
sa(dp31264
S'a'
I37448
sS's'
I18
sS'm'
(lp31265
I1
aI3
aI3
aI5
aI17
aI63
aI23
aI177
aI289
aI921
aI315
aI3083
aI5903
aI8697
aI22425
aI37845
aI31171
aI49237
asS'd'
I15633
sa(dp31266
S'a'
I37451
sS's'
I18
sS'm'
(lp31267
I1
aI1
aI7
aI9
aI21
aI63
aI29
aI227
aI427
aI271
aI525
aI2071
aI7103
aI8389
aI29185
aI51601
aI110737
aI16949
asS'd'
I15634
sa(dp31268
S'a'
I37478
sS's'
I18
sS'm'
(lp31269
I1
aI3
aI3
aI3
aI3
aI49
aI25
aI173
aI79
aI343
aI509
aI1939
aI6389
aI15501
aI20135
aI54365
aI69931
aI135269
asS'd'
I15635
sa(dp31270
S'a'
I37484
sS's'
I18
sS'm'
(lp31271
I1
aI1
aI3
aI3
aI21
aI23
aI41
aI71
aI169
aI947
aI1027
aI2345
aI3397
aI12181
aI15409
aI31725
aI41223
aI58837
asS'd'
I15636
sa(dp31272
S'a'
I37525
sS's'
I18
sS'm'
(lp31273
I1
aI3
aI7
aI1
aI19
aI23
aI57
aI201
aI27
aI449
aI1479
aI921
aI4703
aI10949
aI14369
aI27929
aI45399
aI46055
asS'd'
I15637
sa(dp31274
S'a'
I37553
sS's'
I18
sS'm'
(lp31275
I1
aI1
aI3
aI9
aI13
aI17
aI125
aI17
aI393
aI295
aI497
aI3089
aI6589
aI4003
aI8687
aI48145
aI2683
aI175521
asS'd'
I15638
sa(dp31276
S'a'
I37640
sS's'
I18
sS'm'
(lp31277
I1
aI3
aI3
aI15
aI15
aI13
aI3
aI31
aI51
aI101
aI973
aI101
aI3709
aI13437
aI51
aI14293
aI21561
aI136497
asS'd'
I15639
sa(dp31278
S'a'
I37645
sS's'
I18
sS'm'
(lp31279
I1
aI1
aI5
aI11
aI17
aI27
aI51
aI45
aI77
aI539
aI225
aI2029
aI533
aI153
aI26085
aI33611
aI28153
aI75671
asS'd'
I15640
sa(dp31280
S'a'
I37658
sS's'
I18
sS'm'
(lp31281
I1
aI1
aI1
aI15
aI3
aI59
aI59
aI123
aI475
aI225
aI1613
aI3121
aI2865
aI4647
aI14553
aI35449
aI121657
aI37457
asS'd'
I15641
sa(dp31282
S'a'
I37667
sS's'
I18
sS'm'
(lp31283
I1
aI1
aI5
aI1
aI27
aI33
aI121
aI225
aI57
aI619
aI1293
aI3813
aI2121
aI3525
aI21995
aI47253
aI33095
aI257451
asS'd'
I15642
sa(dp31284
S'a'
I37708
sS's'
I18
sS'm'
(lp31285
I1
aI3
aI1
aI11
aI11
aI43
aI115
aI233
aI335
aI185
aI989
aI3567
aI4135
aI2357
aI20559
aI43325
aI43015
aI51695
asS'd'
I15643
sa(dp31286
S'a'
I37723
sS's'
I18
sS'm'
(lp31287
I1
aI1
aI5
aI9
aI11
aI49
aI45
aI187
aI93
aI967
aI1609
aI2511
aI1549
aI4045
aI21309
aI16341
aI13495
aI214827
asS'd'
I15644
sa(dp31288
S'a'
I37732
sS's'
I18
sS'm'
(lp31289
I1
aI1
aI1
aI13
aI21
aI23
aI81
aI7
aI259
aI483
aI1059
aI773
aI5297
aI10123
aI9857
aI61187
aI47355
aI76307
asS'd'
I15645
sa(dp31290
S'a'
I37747
sS's'
I18
sS'm'
(lp31291
I1
aI3
aI7
aI9
aI29
aI51
aI113
aI255
aI223
aI853
aI1173
aI1019
aI1587
aI9629
aI22373
aI32731
aI125179
aI193271
asS'd'
I15646
sa(dp31292
S'a'
I37753
sS's'
I18
sS'm'
(lp31293
I1
aI1
aI5
aI11
aI3
aI55
aI25
aI145
aI347
aI451
aI1447
aI3399
aI5873
aI11579
aI11107
aI64707
aI10161
aI142003
asS'd'
I15647
sa(dp31294
S'a'
I37772
sS's'
I18
sS'm'
(lp31295
I1
aI1
aI1
aI7
aI15
aI49
aI109
aI93
aI267
aI919
aI177
aI2247
aI8129
aI8039
aI15629
aI63767
aI98153
aI143255
asS'd'
I15648
sa(dp31296
S'a'
I37789
sS's'
I18
sS'm'
(lp31297
I1
aI1
aI5
aI3
aI3
aI27
aI47
aI151
aI231
aI35
aI155
aI2745
aI7349
aI6543
aI14117
aI19549
aI54927
aI10819
asS'd'
I15649
sa(dp31298
S'a'
I37817
sS's'
I18
sS'm'
(lp31299
I1
aI3
aI7
aI15
aI31
aI29
aI17
aI203
aI249
aI169
aI1071
aI3069
aI6269
aI3455
aI27177
aI33761
aI111003
aI4527
asS'd'
I15650
sa(dp31300
S'a'
I37826
sS's'
I18
sS'm'
(lp31301
I1
aI3
aI1
aI5
aI31
aI15
aI65
aI189
aI3
aI917
aI857
aI1221
aI3553
aI2883
aI3631
aI32971
aI68057
aI109081
asS'd'
I15651
sa(dp31302
S'a'
I37831
sS's'
I18
sS'm'
(lp31303
I1
aI3
aI1
aI9
aI3
aI55
aI127
aI57
aI125
aI463
aI199
aI317
aI3373
aI967
aI5569
aI55997
aI17167
aI33585
asS'd'
I15652
sa(dp31304
S'a'
I37845
sS's'
I18
sS'm'
(lp31305
I1
aI3
aI5
aI1
aI9
aI57
aI15
aI89
aI335
aI119
aI1445
aI1931
aI4177
aI2495
aI27507
aI8209
aI60003
aI29657
asS'd'
I15653
sa(dp31306
S'a'
I37855
sS's'
I18
sS'm'
(lp31307
I1
aI3
aI5
aI7
aI15
aI43
aI83
aI117
aI283
aI131
aI653
aI57
aI6789
aI7633
aI30525
aI64131
aI101981
aI122017
asS'd'
I15654
sa(dp31308
S'a'
I37859
sS's'
I18
sS'm'
(lp31309
I1
aI3
aI7
aI11
aI3
aI17
aI115
aI217
aI391
aI825
aI1633
aI885
aI7787
aI5595
aI12235
aI30233
aI53587
aI62985
asS'd'
I15655
sa(dp31310
S'a'
I37866
sS's'
I18
sS'm'
(lp31311
I1
aI1
aI5
aI3
aI5
aI13
aI99
aI1
aI75
aI329
aI1247
aI107
aI2337
aI4201
aI6217
aI12273
aI41585
aI46563
asS'd'
I15656
sa(dp31312
S'a'
I37880
sS's'
I18
sS'm'
(lp31313
I1
aI3
aI1
aI15
aI25
aI53
aI33
aI125
aI311
aI955
aI161
aI3631
aI581
aI11915
aI4223
aI63207
aI16517
aI201665
asS'd'
I15657
sa(dp31314
S'a'
I37885
sS's'
I18
sS'm'
(lp31315
I1
aI1
aI5
aI1
aI27
aI23
aI93
aI211
aI483
aI691
aI949
aI1825
aI391
aI12111
aI13639
aI61009
aI88503
aI104823
asS'd'
I15658
sa(dp31316
S'a'
I37897
sS's'
I18
sS'm'
(lp31317
I1
aI3
aI1
aI13
aI3
aI9
aI51
aI133
aI259
aI977
aI697
aI661
aI7661
aI3987
aI8327
aI50155
aI112235
aI236135
asS'd'
I15659
sa(dp31318
S'a'
I37906
sS's'
I18
sS'm'
(lp31319
I1
aI1
aI3
aI13
aI7
aI39
aI121
aI37
aI151
aI973
aI1275
aI2699
aI3345
aI7167
aI19245
aI55535
aI12305
aI33567
asS'd'
I15660
sa(dp31320
S'a'
I37948
sS's'
I18
sS'm'
(lp31321
I1
aI1
aI1
aI1
aI27
aI5
aI91
aI63
aI409
aI579
aI459
aI2335
aI4721
aI3305
aI11293
aI15405
aI74513
aI157863
asS'd'
I15661
sa(dp31322
S'a'
I37954
sS's'
I18
sS'm'
(lp31323
I1
aI1
aI5
aI1
aI21
aI45
aI55
aI111
aI475
aI381
aI659
aI1131
aI3575
aI5165
aI27221
aI46757
aI53587
aI90741
asS'd'
I15662
sa(dp31324
S'a'
I37978
sS's'
I18
sS'm'
(lp31325
I1
aI1
aI5
aI15
aI11
aI31
aI121
aI209
aI69
aI389
aI779
aI2833
aI4519
aI1801
aI4363
aI24723
aI105849
aI54475
asS'd'
I15663
sa(dp31326
S'a'
I37980
sS's'
I18
sS'm'
(lp31327
I1
aI1
aI3
aI9
aI7
aI19
aI11
aI75
aI275
aI77
aI1997
aI1661
aI6139
aI13165
aI30653
aI49469
aI67053
aI3811
asS'd'
I15664
sa(dp31328
S'a'
I37990
sS's'
I18
sS'm'
(lp31329
I1
aI1
aI3
aI9
aI5
aI11
aI5
aI151
aI395
aI715
aI1381
aI3011
aI5939
aI1805
aI8063
aI62877
aI99749
aI112951
asS'd'
I15665
sa(dp31330
S'a'
I38001
sS's'
I18
sS'm'
(lp31331
I1
aI1
aI5
aI13
aI19
aI15
aI113
aI47
aI455
aI173
aI1897
aI701
aI6093
aI2089
aI3977
aI20599
aI60947
aI23671
asS'd'
I15666
sa(dp31332
S'a'
I38020
sS's'
I18
sS'm'
(lp31333
I1
aI1
aI1
aI13
aI23
aI19
aI13
aI117
aI275
aI313
aI1683
aI2975
aI179
aI3949
aI4361
aI60211
aI91999
aI211219
asS'd'
I15667
sa(dp31334
S'a'
I38029
sS's'
I18
sS'm'
(lp31335
I1
aI3
aI5
aI15
aI13
aI53
aI83
aI161
aI491
aI1001
aI1773
aI1227
aI1965
aI14479
aI17677
aI24399
aI86431
aI203303
asS'd'
I15668
sa(dp31336
S'a'
I38047
sS's'
I18
sS'm'
(lp31337
I1
aI1
aI7
aI15
aI5
aI51
aI103
aI131
aI351
aI747
aI1227
aI2859
aI6693
aI10615
aI29485
aI6619
aI106239
aI148739
asS'd'
I15669
sa(dp31338
S'a'
I38063
sS's'
I18
sS'm'
(lp31339
I1
aI3
aI5
aI1
aI9
aI43
aI91
aI173
aI223
aI393
aI1181
aI3785
aI6589
aI1299
aI10217
aI20891
aI64125
aI63409
asS'd'
I15670
sa(dp31340
S'a'
I38077
sS's'
I18
sS'm'
(lp31341
I1
aI1
aI5
aI7
aI11
aI23
aI45
aI57
aI397
aI771
aI511
aI1849
aI343
aI14139
aI26271
aI56241
aI52581
aI163187
asS'd'
I15671
sa(dp31342
S'a'
I38110
sS's'
I18
sS'm'
(lp31343
I1
aI3
aI7
aI5
aI15
aI59
aI89
aI151
aI255
aI247
aI291
aI219
aI995
aI10821
aI1445
aI35581
aI88767
aI16871
asS'd'
I15672
sa(dp31344
S'a'
I38116
sS's'
I18
sS'm'
(lp31345
I1
aI1
aI7
aI11
aI7
aI25
aI3
aI175
aI253
aI193
aI1641
aI1669
aI7095
aI11871
aI10801
aI42567
aI120663
aI109347
asS'd'
I15673
sa(dp31346
S'a'
I38119
sS's'
I18
sS'm'
(lp31347
I1
aI3
aI5
aI7
aI31
aI41
aI119
aI39
aI149
aI653
aI153
aI2783
aI1377
aI5223
aI17915
aI3127
aI41869
aI193477
asS'd'
I15674
sa(dp31348
S'a'
I38176
sS's'
I18
sS'm'
(lp31349
I1
aI3
aI3
aI13
aI23
aI19
aI47
aI85
aI487
aI103
aI237
aI2363
aI4451
aI5077
aI23749
aI17011
aI73561
aI169165
asS'd'
I15675
sa(dp31350
S'a'
I38186
sS's'
I18
sS'm'
(lp31351
I1
aI1
aI3
aI9
aI21
aI25
aI77
aI235
aI53
aI681
aI1463
aI2093
aI1525
aI12797
aI5469
aI54277
aI15587
aI68395
asS'd'
I15676
sa(dp31352
S'a'
I38194
sS's'
I18
sS'm'
(lp31353
I1
aI1
aI1
aI15
aI23
aI63
aI63
aI225
aI239
aI143
aI1073
aI199
aI3231
aI1371
aI11215
aI5999
aI100705
aI174681
asS'd'
I15677
sa(dp31354
S'a'
I38218
sS's'
I18
sS'm'
(lp31355
I1
aI1
aI1
aI3
aI17
aI25
aI69
aI179
aI445
aI695
aI281
aI379
aI8115
aI9443
aI13221
aI50669
aI37369
aI62151
asS'd'
I15678
sa(dp31356
S'a'
I38241
sS's'
I18
sS'm'
(lp31357
I1
aI3
aI3
aI9
aI11
aI29
aI21
aI89
aI441
aI353
aI401
aI1139
aI5003
aI8087
aI24457
aI50237
aI110993
aI117233
asS'd'
I15679
sa(dp31358
S'a'
I38247
sS's'
I18
sS'm'
(lp31359
I1
aI3
aI3
aI1
aI13
aI45
aI31
aI249
aI295
aI149
aI591
aI2071
aI3611
aI931
aI16261
aI8239
aI82767
aI195665
asS'd'
I15680
sa(dp31360
S'a'
I38261
sS's'
I18
sS'm'
(lp31361
I1
aI3
aI3
aI9
aI19
aI47
aI69
aI177
aI493
aI231
aI431
aI1359
aI6867
aI7641
aI15661
aI25285
aI65477
aI212643
asS'd'
I15681
sa(dp31362
S'a'
I38268
sS's'
I18
sS'm'
(lp31363
I1
aI3
aI3
aI13
aI19
aI63
aI83
aI153
aI367
aI407
aI547
aI661
aI7743
aI5473
aI2993
aI62937
aI33811
aI101313
asS'd'
I15682
sa(dp31364
S'a'
I38277
sS's'
I18
sS'm'
(lp31365
I1
aI3
aI5
aI3
aI29
aI17
aI19
aI203
aI79
aI279
aI1333
aI1851
aI51
aI9793
aI12955
aI17383
aI73437
aI121743
asS'd'
I15683
sa(dp31366
S'a'
I38287
sS's'
I18
sS'm'
(lp31367
I1
aI1
aI1
aI11
aI11
aI43
aI31
aI187
aI463
aI827
aI1511
aI225
aI845
aI8963
aI1553
aI61269
aI122033
aI245633
asS'd'
I15684
sa(dp31368
S'a'
I38295
sS's'
I18
sS'm'
(lp31369
I1
aI1
aI3
aI3
aI31
aI23
aI9
aI241
aI377
aI317
aI655
aI2197
aI2461
aI13239
aI15649
aI7879
aI55085
aI129855
asS'd'
I15685
sa(dp31370
S'a'
I38299
sS's'
I18
sS'm'
(lp31371
I1
aI3
aI5
aI7
aI9
aI37
aI1
aI191
aI185
aI145
aI1567
aI3423
aI1127
aI1991
aI10741
aI38407
aI22915
aI222845
asS'd'
I15686
sa(dp31372
S'a'
I38301
sS's'
I18
sS'm'
(lp31373
I1
aI1
aI5
aI3
aI27
aI31
aI11
aI227
aI307
aI973
aI745
aI1079
aI7479
aI10065
aI31389
aI19195
aI114775
aI246615
asS'd'
I15687
sa(dp31374
S'a'
I38305
sS's'
I18
sS'm'
(lp31375
I1
aI3
aI1
aI11
aI29
aI27
aI11
aI83
aI205
aI399
aI1489
aI739
aI715
aI7955
aI16473
aI21127
aI62379
aI260399
asS'd'
I15688
sa(dp31376
S'a'
I38312
sS's'
I18
sS'm'
(lp31377
I1
aI3
aI3
aI3
aI25
aI25
aI123
aI163
aI399
aI841
aI963
aI2089
aI4949
aI13085
aI19831
aI15345
aI60377
aI164235
asS'd'
I15689
sa(dp31378
S'a'
I38315
sS's'
I18
sS'm'
(lp31379
I1
aI1
aI1
aI9
aI3
aI21
aI101
aI105
aI397
aI23
aI1505
aI3201
aI547
aI295
aI23247
aI18823
aI115243
aI151073
asS'd'
I15690
sa(dp31380
S'a'
I38317
sS's'
I18
sS'm'
(lp31381
I1
aI3
aI7
aI1
aI31
aI27
aI111
aI23
aI205
aI709
aI1625
aI3921
aI6225
aI11039
aI29549
aI51239
aI119003
aI133663
asS'd'
I15691
sa(dp31382
S'a'
I38343
sS's'
I18
sS'm'
(lp31383
I1
aI3
aI3
aI11
aI21
aI49
aI111
aI195
aI25
aI833
aI1991
aI563
aI7031
aI15429
aI5707
aI12351
aI32221
aI16599
asS'd'
I15692
sa(dp31384
S'a'
I38344
sS's'
I18
sS'm'
(lp31385
I1
aI1
aI5
aI7
aI5
aI7
aI39
aI171
aI39
aI921
aI385
aI2343
aI625
aI15355
aI4923
aI36597
aI56901
aI148377
asS'd'
I15693
sa(dp31386
S'a'
I38350
sS's'
I18
sS'm'
(lp31387
I1
aI1
aI3
aI15
aI7
aI43
aI89
aI217
aI67
aI271
aI853
aI147
aI6767
aI3183
aI341
aI40769
aI116767
aI22351
asS'd'
I15694
sa(dp31388
S'a'
I38358
sS's'
I18
sS'm'
(lp31389
I1
aI3
aI5
aI7
aI7
aI3
aI105
aI27
aI183
aI59
aI953
aI4027
aI1277
aI10323
aI29437
aI56085
aI32677
aI198067
asS'd'
I15695
sa(dp31390
S'a'
I38364
sS's'
I18
sS'm'
(lp31391
I1
aI1
aI1
aI15
aI23
aI29
aI51
aI209
aI13
aI177
aI1103
aI1723
aI2877
aI9199
aI25601
aI15537
aI8599
aI230819
asS'd'
I15696
sa(dp31392
S'a'
I38371
sS's'
I18
sS'm'
(lp31393
I1
aI1
aI7
aI1
aI29
aI39
aI41
aI217
aI467
aI423
aI431
aI2707
aI2017
aI11865
aI11989
aI12045
aI71349
aI6311
asS'd'
I15697
sa(dp31394
S'a'
I38373
sS's'
I18
sS'm'
(lp31395
I1
aI1
aI1
aI13
aI15
aI25
aI3
aI55
aI403
aI833
aI1843
aI1159
aI1703
aI2221
aI15379
aI65027
aI18327
aI108881
asS'd'
I15698
sa(dp31396
S'a'
I38377
sS's'
I18
sS'm'
(lp31397
I1
aI3
aI7
aI13
aI3
aI27
aI13
aI227
aI215
aI873
aI1073
aI1117
aI7941
aI13607
aI7571
aI6957
aI44991
aI239761
asS'd'
I15699
sa(dp31398
S'a'
I38392
sS's'
I18
sS'm'
(lp31399
I1
aI3
aI7
aI11
aI23
aI1
aI95
aI235
aI283
aI977
aI1443
aI161
aI5937
aI4351
aI30835
aI35569
aI57509
aI1835
asS'd'
I15700
sa(dp31400
S'a'
I38407
sS's'
I18
sS'm'
(lp31401
I1
aI3
aI3
aI13
aI11
aI1
aI85
aI75
aI261
aI543
aI9
aI899
aI5821
aI5465
aI9771
aI53707
aI101003
aI219215
asS'd'
I15701
sa(dp31402
S'a'
I38408
sS's'
I18
sS'm'
(lp31403
I1
aI3
aI1
aI7
aI21
aI49
aI35
aI19
aI35
aI759
aI1467
aI1423
aI6355
aI8415
aI563
aI24157
aI121029
aI87309
asS'd'
I15702
sa(dp31404
S'a'
I38421
sS's'
I18
sS'm'
(lp31405
I1
aI1
aI7
aI1
aI9
aI13
aI65
aI85
aI209
aI583
aI387
aI1743
aI2665
aI12021
aI7525
aI27665
aI45885
aI135039
asS'd'
I15703
sa(dp31406
S'a'
I38438
sS's'
I18
sS'm'
(lp31407
I1
aI1
aI7
aI11
aI29
aI41
aI91
aI17
aI291
aI211
aI1801
aI493
aI899
aI14491
aI1741
aI28971
aI35205
aI131417
asS'd'
I15704
sa(dp31408
S'a'
I38442
sS's'
I18
sS'm'
(lp31409
I1
aI1
aI5
aI13
aI23
aI55
aI119
aI165
aI61
aI653
aI1375
aI3575
aI5081
aI7767
aI19963
aI61583
aI107149
aI240639
asS'd'
I15705
sa(dp31410
S'a'
I38464
sS's'
I18
sS'm'
(lp31411
I1
aI3
aI5
aI15
aI25
aI3
aI51
aI27
aI127
aI259
aI55
aI2221
aI3951
aI6243
aI15825
aI42881
aI37009
aI254401
asS'd'
I15706
sa(dp31412
S'a'
I38473
sS's'
I18
sS'm'
(lp31413
I1
aI3
aI5
aI11
aI25
aI63
aI13
aI105
aI111
aI677
aI1545
aI2399
aI4419
aI10853
aI7213
aI17183
aI103411
aI67459
asS'd'
I15707
sa(dp31414
S'a'
I38484
sS's'
I18
sS'm'
(lp31415
I1
aI1
aI1
aI11
aI11
aI31
aI73
aI125
aI155
aI545
aI1857
aI2749
aI6389
aI4083
aI16239
aI23651
aI68881
aI43455
asS'd'
I15708
sa(dp31416
S'a'
I38491
sS's'
I18
sS'm'
(lp31417
I1
aI3
aI7
aI7
aI21
aI3
aI117
aI237
aI431
aI17
aI687
aI2613
aI7483
aI3253
aI30511
aI53833
aI33077
aI157055
asS'd'
I15709
sa(dp31418
S'a'
I38510
sS's'
I18
sS'm'
(lp31419
I1
aI1
aI1
aI1
aI1
aI57
aI65
aI97
aI415
aI477
aI1003
aI1415
aI1493
aI12993
aI30965
aI24809
aI1467
aI213021
asS'd'
I15710
sa(dp31420
S'a'
I38518
sS's'
I18
sS'm'
(lp31421
I1
aI1
aI3
aI7
aI25
aI33
aI45
aI25
aI511
aI733
aI1077
aI2483
aI5899
aI14295
aI11631
aI50609
aI128989
aI45177
asS'd'
I15711
sa(dp31422
S'a'
I38531
sS's'
I18
sS'm'
(lp31423
I1
aI1
aI3
aI3
aI25
aI17
aI115
aI31
aI115
aI191
aI293
aI3991
aI3039
aI6751
aI16217
aI16517
aI21121
aI193641
asS'd'
I15712
sa(dp31424
S'a'
I38537
sS's'
I18
sS'm'
(lp31425
I1
aI3
aI3
aI13
aI25
aI23
aI7
aI51
aI363
aI641
aI333
aI2533
aI605
aI1105
aI12941
aI4195
aI129571
aI13253
asS'd'
I15713
sa(dp31426
S'a'
I38538
sS's'
I18
sS'm'
(lp31427
I1
aI3
aI1
aI11
aI17
aI21
aI7
aI205
aI293
aI159
aI9
aI441
aI3287
aI10247
aI2115
aI54099
aI128109
aI8137
asS'd'
I15714
sa(dp31428
S'a'
I38567
sS's'
I18
sS'm'
(lp31429
I1
aI1
aI7
aI5
aI21
aI17
aI43
aI87
aI117
aI737
aI149
aI3175
aI343
aI8509
aI12147
aI22041
aI80037
aI23277
asS'd'
I15715
sa(dp31430
S'a'
I38594
sS's'
I18
sS'm'
(lp31431
I1
aI3
aI3
aI1
aI3
aI7
aI101
aI245
aI11
aI1003
aI175
aI2323
aI7807
aI15611
aI5161
aI10277
aI37009
aI83231
asS'd'
I15716
sa(dp31432
S'a'
I38647
sS's'
I18
sS'm'
(lp31433
I1
aI3
aI1
aI5
aI13
aI17
aI113
aI75
aI315
aI237
aI77
aI587
aI5409
aI2053
aI22551
aI15205
aI82545
aI18531
asS'd'
I15717
sa(dp31434
S'a'
I38651
sS's'
I18
sS'm'
(lp31435
I1
aI3
aI5
aI11
aI9
aI57
aI61
aI117
aI281
aI111
aI369
aI2411
aI1691
aI3391
aI5379
aI8237
aI87329
aI4253
asS'd'
I15718
sa(dp31436
S'a'
I38654
sS's'
I18
sS'm'
(lp31437
I1
aI3
aI3
aI3
aI19
aI25
aI101
aI1
aI495
aI25
aI1317
aI2333
aI6183
aI12215
aI27879
aI56403
aI37169
aI71635
asS'd'
I15719
sa(dp31438
S'a'
I38686
sS's'
I18
sS'm'
(lp31439
I1
aI1
aI3
aI5
aI17
aI63
aI49
aI127
aI171
aI405
aI1763
aI3697
aI405
aI2233
aI4137
aI28787
aI108319
aI53133
asS'd'
I15720
sa(dp31440
S'a'
I38702
sS's'
I18
sS'm'
(lp31441
I1
aI1
aI5
aI7
aI23
aI43
aI87
aI189
aI97
aI239
aI1459
aI2115
aI7517
aI7799
aI28957
aI37105
aI71835
aI199195
asS'd'
I15721
sa(dp31442
S'a'
I38751
sS's'
I18
sS'm'
(lp31443
I1
aI3
aI1
aI3
aI25
aI25
aI23
aI61
aI369
aI717
aI1711
aI1103
aI7535
aI9871
aI25
aI26849
aI55955
aI113389
asS'd'
I15722
sa(dp31444
S'a'
I38821
sS's'
I18
sS'm'
(lp31445
I1
aI1
aI1
aI11
aI25
aI57
aI33
aI175
aI127
aI541
aI1031
aI2847
aI2069
aI4033
aI25593
aI10615
aI50097
aI122955
asS'd'
I15723
sa(dp31446
S'a'
I38825
sS's'
I18
sS'm'
(lp31447
I1
aI3
aI3
aI13
aI11
aI27
aI97
aI171
aI245
aI33
aI213
aI4069
aI753
aI3535
aI11727
aI34941
aI100543
aI220789
asS'd'
I15724
sa(dp31448
S'a'
I38853
sS's'
I18
sS'm'
(lp31449
I1
aI3
aI3
aI9
aI3
aI17
aI13
aI237
aI477
aI507
aI1751
aI3191
aI3385
aI13977
aI23355
aI57355
aI64341
aI37683
asS'd'
I15725
sa(dp31450
S'a'
I38863
sS's'
I18
sS'm'
(lp31451
I1
aI1
aI7
aI13
aI13
aI43
aI7
aI153
aI209
aI7
aI63
aI585
aI1715
aI13313
aI25355
aI46759
aI71893
aI29755
asS'd'
I15726
sa(dp31452
S'a'
I38882
sS's'
I18
sS'm'
(lp31453
I1
aI3
aI3
aI3
aI11
aI23
aI11
aI147
aI135
aI1011
aI1105
aI3931
aI3861
aI13589
aI32183
aI30727
aI37685
aI67123
asS'd'
I15727
sa(dp31454
S'a'
I38884
sS's'
I18
sS'm'
(lp31455
I1
aI3
aI7
aI1
aI11
aI13
aI25
aI229
aI147
aI843
aI329
aI3337
aI7559
aI13193
aI3011
aI31549
aI102461
aI46195
asS'd'
I15728
sa(dp31456
S'a'
I38932
sS's'
I18
sS'm'
(lp31457
I1
aI1
aI5
aI11
aI5
aI47
aI127
aI89
aI53
aI663
aI261
aI541
aI7743
aI13037
aI9013
aI23079
aI81225
aI239875
asS'd'
I15729
sa(dp31458
S'a'
I38941
sS's'
I18
sS'm'
(lp31459
I1
aI1
aI7
aI3
aI5
aI39
aI15
aI177
aI357
aI357
aI1959
aI1721
aI6703
aI11829
aI1195
aI42113
aI88699
aI244347
asS'd'
I15730
sa(dp31460
S'a'
I38952
sS's'
I18
sS'm'
(lp31461
I1
aI1
aI5
aI15
aI25
aI19
aI7
aI3
aI225
aI773
aI1535
aI99
aI6555
aI4105
aI19137
aI56155
aI109141
aI161015
asS'd'
I15731
sa(dp31462
S'a'
I38960
sS's'
I18
sS'm'
(lp31463
I1
aI1
aI5
aI15
aI5
aI59
aI41
aI53
aI203
aI459
aI1063
aI3015
aI5397
aI1559
aI20835
aI57773
aI67687
aI206189
asS'd'
I15732
sa(dp31464
S'a'
I38980
sS's'
I18
sS'm'
(lp31465
I1
aI3
aI1
aI11
aI17
aI25
aI61
aI221
aI37
aI809
aI1461
aI1961
aI7697
aI1777
aI23179
aI54761
aI7787
aI177737
asS'd'
I15733
sa(dp31466
S'a'
I38995
sS's'
I18
sS'm'
(lp31467
I1
aI3
aI7
aI15
aI27
aI21
aI49
aI107
aI353
aI677
aI461
aI239
aI5871
aI1059
aI3011
aI32397
aI13149
aI103973
asS'd'
I15734
sa(dp31468
S'a'
I39004
sS's'
I18
sS'm'
(lp31469
I1
aI1
aI5
aI3
aI11
aI53
aI61
aI239
aI479
aI913
aI479
aI3435
aI4979
aI7931
aI16841
aI60077
aI26667
aI212601
asS'd'
I15735
sa(dp31470
S'a'
I39013
sS's'
I18
sS'm'
(lp31471
I1
aI1
aI3
aI5
aI3
aI19
aI43
aI143
aI353
aI507
aI871
aI2547
aI7321
aI6163
aI9425
aI62911
aI86153
aI239011
asS'd'
I15736
sa(dp31472
S'a'
I39017
sS's'
I18
sS'm'
(lp31473
I1
aI1
aI1
aI3
aI15
aI7
aI115
aI43
aI69
aI299
aI1235
aI1511
aI3111
aI7465
aI769
aI46981
aI127707
aI195839
asS'd'
I15737
sa(dp31474
S'a'
I39026
sS's'
I18
sS'm'
(lp31475
I1
aI1
aI1
aI5
aI23
aI27
aI19
aI21
aI273
aI291
aI953
aI3577
aI3147
aI3863
aI18625
aI53505
aI33699
aI123305
asS'd'
I15738
sa(dp31476
S'a'
I39056
sS's'
I18
sS'm'
(lp31477
I1
aI3
aI5
aI9
aI3
aI11
aI89
aI27
aI447
aI119
aI493
aI2605
aI8175
aI8837
aI27555
aI2319
aI99101
aI79121
asS'd'
I15739
sa(dp31478
S'a'
I39121
sS's'
I18
sS'm'
(lp31479
I1
aI1
aI7
aI11
aI1
aI11
aI77
aI129
aI97
aI261
aI1241
aI3117
aI1627
aI5397
aI6495
aI52339
aI52711
aI206237
asS'd'
I15740
sa(dp31480
S'a'
I39124
sS's'
I18
sS'm'
(lp31481
I1
aI3
aI7
aI9
aI27
aI57
aI77
aI147
aI35
aI845
aI1417
aI1615
aI6097
aI12559
aI10765
aI19027
aI91693
aI204339
asS'd'
I15741
sa(dp31482
S'a'
I39133
sS's'
I18
sS'm'
(lp31483
I1
aI1
aI3
aI5
aI25
aI47
aI17
aI145
aI7
aI969
aI1981
aI733
aI4303
aI7785
aI4241
aI39733
aI82569
aI78061
asS'd'
I15742
sa(dp31484
S'a'
I39157
sS's'
I18
sS'm'
(lp31485
I1
aI1
aI5
aI1
aI5
aI47
aI45
aI111
aI405
aI943
aI1911
aI1613
aI3817
aI10483
aI17729
aI7201
aI88033
aI261701
asS'd'
I15743
sa(dp31486
S'a'
I39181
sS's'
I18
sS'm'
(lp31487
I1
aI3
aI3
aI13
aI9
aI3
aI87
aI39
aI277
aI769
aI57
aI2503
aI7803
aI11041
aI20945
aI19623
aI32617
aI110027
asS'd'
I15744
sa(dp31488
S'a'
I39187
sS's'
I18
sS'm'
(lp31489
I1
aI1
aI3
aI3
aI27
aI57
aI1
aI103
aI427
aI935
aI1617
aI665
aI837
aI8001
aI13543
aI44771
aI64033
aI65239
asS'd'
I15745
sa(dp31490
S'a'
I39212
sS's'
I18
sS'm'
(lp31491
I1
aI1
aI5
aI3
aI13
aI21
aI31
aI59
aI225
aI945
aI1825
aI1511
aI3273
aI3171
aI30347
aI21993
aI40203
aI143297
asS'd'
I15746
sa(dp31492
S'a'
I39223
sS's'
I18
sS'm'
(lp31493
I1
aI3
aI5
aI7
aI11
aI3
aI3
aI217
aI167
aI885
aI975
aI3249
aI7909
aI13621
aI18697
aI61021
aI31497
aI198033
asS'd'
I15747
sa(dp31494
S'a'
I39250
sS's'
I18
sS'm'
(lp31495
I1
aI3
aI5
aI9
aI11
aI5
aI87
aI33
aI117
aI471
aI267
aI529
aI5879
aI13969
aI5731
aI52613
aI106411
aI74341
asS'd'
I15748
sa(dp31496
S'a'
I39265
sS's'
I18
sS'm'
(lp31497
I1
aI1
aI7
aI11
aI5
aI31
aI25
aI55
aI135
aI779
aI717
aI1953
aI7929
aI11011
aI6133
aI14099
aI106975
aI178337
asS'd'
I15749
sa(dp31498
S'a'
I39320
sS's'
I18
sS'm'
(lp31499
I1
aI3
aI1
aI1
aI29
aI17
aI125
aI7
aI445
aI299
aI1897
aI3235
aI8189
aI14339
aI14725
aI63185
aI126751
aI88747
asS'd'
I15750
sa(dp31500
S'a'
I39330
sS's'
I18
sS'm'
(lp31501
I1
aI1
aI5
aI3
aI1
aI11
aI65
aI161
aI243
aI605
aI1945
aI3141
aI6443
aI9703
aI13331
aI2239
aI6315
aI247595
asS'd'
I15751
sa(dp31502
S'a'
I39335
sS's'
I18
sS'm'
(lp31503
I1
aI3
aI1
aI7
aI15
aI23
aI83
aI215
aI331
aI631
aI453
aI879
aI4109
aI4897
aI16535
aI55749
aI90799
aI147287
asS'd'
I15752
sa(dp31504
S'a'
I39353
sS's'
I18
sS'm'
(lp31505
I1
aI3
aI5
aI13
aI25
aI1
aI109
aI205
aI49
aI471
aI1735
aI973
aI1279
aI9917
aI18225
aI44921
aI98519
aI211541
asS'd'
I15753
sa(dp31506
S'a'
I39371
sS's'
I18
sS'm'
(lp31507
I1
aI1
aI5
aI11
aI29
aI41
aI113
aI187
aI75
aI479
aI1633
aI841
aI6259
aI8919
aI27751
aI25179
aI115369
aI166567
asS'd'
I15754
sa(dp31508
S'a'
I39385
sS's'
I18
sS'm'
(lp31509
I1
aI3
aI5
aI11
aI17
aI31
aI107
aI41
aI435
aI647
aI811
aI2937
aI5819
aI3483
aI3835
aI57033
aI53543
aI61973
asS'd'
I15755
sa(dp31510
S'a'
I39421
sS's'
I18
sS'm'
(lp31511
I1
aI3
aI3
aI11
aI15
aI45
aI33
aI103
aI505
aI67
aI463
aI1275
aI2449
aI15261
aI22867
aI25215
aI38793
aI189309
asS'd'
I15756
sa(dp31512
S'a'
I39432
sS's'
I18
sS'm'
(lp31513
I1
aI1
aI1
aI5
aI19
aI45
aI35
aI173
aI365
aI39
aI1599
aI3623
aI2231
aI12141
aI19437
aI27053
aI15869
aI104719
asS'd'
I15757
sa(dp31514
S'a'
I39443
sS's'
I18
sS'm'
(lp31515
I1
aI3
aI7
aI7
aI9
aI17
aI87
aI151
aI249
aI81
aI1109
aI1951
aI7475
aI1699
aI17847
aI64149
aI50285
aI242793
asS'd'
I15758
sa(dp31516
S'a'
I39450
sS's'
I18
sS'm'
(lp31517
I1
aI3
aI5
aI13
aI15
aI51
aI35
aI105
aI479
aI763
aI1945
aI2349
aI2987
aI621
aI283
aI20411
aI65799
aI86517
asS'd'
I15759
sa(dp31518
S'a'
I39455
sS's'
I18
sS'm'
(lp31519
I1
aI3
aI3
aI1
aI19
aI31
aI49
aI229
aI249
aI689
aI737
aI4027
aI5405
aI15211
aI26785
aI39143
aI93163
aI190421
asS'd'
I15760
sa(dp31520
S'a'
I39468
sS's'
I18
sS'm'
(lp31521
I1
aI3
aI7
aI5
aI1
aI21
aI63
aI97
aI347
aI73
aI745
aI3455
aI2347
aI3821
aI31385
aI6545
aI91803
aI72895
asS'd'
I15761
sa(dp31522
S'a'
I39473
sS's'
I18
sS'm'
(lp31523
I1
aI1
aI1
aI15
aI23
aI11
aI107
aI47
aI183
aI235
aI1985
aI3277
aI933
aI8491
aI14423
aI24293
aI6783
aI162199
asS'd'
I15762
sa(dp31524
S'a'
I39488
sS's'
I18
sS'm'
(lp31525
I1
aI3
aI3
aI9
aI17
aI3
aI123
aI59
aI277
aI773
aI1617
aI2979
aI1555
aI9753
aI10947
aI24745
aI89043
aI45185
asS'd'
I15763
sa(dp31526
S'a'
I39497
sS's'
I18
sS'm'
(lp31527
I1
aI3
aI7
aI3
aI25
aI17
aI79
aI43
aI311
aI415
aI1045
aI1289
aI7451
aI11413
aI11319
aI37177
aI101327
aI147453
asS'd'
I15764
sa(dp31528
S'a'
I39521
sS's'
I18
sS'm'
(lp31529
I1
aI3
aI7
aI7
aI17
aI45
aI49
aI33
aI313
aI613
aI1773
aI773
aI161
aI13579
aI1207
aI5681
aI120597
aI178639
asS'd'
I15765
sa(dp31530
S'a'
I39531
sS's'
I18
sS'm'
(lp31531
I1
aI3
aI5
aI5
aI17
aI43
aI65
aI243
aI287
aI223
aI253
aI687
aI887
aI14887
aI1077
aI53337
aI62381
aI43531
asS'd'
I15766
sa(dp31532
S'a'
I39542
sS's'
I18
sS'm'
(lp31533
I1
aI3
aI5
aI1
aI21
aI3
aI39
aI149
aI497
aI939
aI1537
aI437
aI5345
aI10321
aI25151
aI48785
aI49879
aI90945
asS'd'
I15767
sa(dp31534
S'a'
I39585
sS's'
I18
sS'm'
(lp31535
I1
aI1
aI7
aI11
aI1
aI61
aI113
aI63
aI285
aI615
aI343
aI2897
aI1939
aI7911
aI16387
aI10781
aI92769
aI27995
asS'd'
I15768
sa(dp31536
S'a'
I39605
sS's'
I18
sS'm'
(lp31537
I1
aI1
aI3
aI3
aI19
aI29
aI85
aI227
aI355
aI857
aI883
aI1853
aI5065
aI13795
aI5749
aI59107
aI57947
aI35775
asS'd'
I15769
sa(dp31538
S'a'
I39658
sS's'
I18
sS'm'
(lp31539
I1
aI3
aI5
aI9
aI23
aI37
aI119
aI161
aI23
aI511
aI81
aI973
aI4769
aI10821
aI32607
aI61731
aI64907
aI99055
asS'd'
I15770
sa(dp31540
S'a'
I39677
sS's'
I18
sS'm'
(lp31541
I1
aI3
aI1
aI3
aI11
aI17
aI109
aI241
aI349
aI887
aI1651
aI3865
aI2045
aI15893
aI4597
aI11557
aI53313
aI48489
asS'd'
I15771
sa(dp31542
S'a'
I39703
sS's'
I18
sS'm'
(lp31543
I1
aI1
aI5
aI9
aI31
aI43
aI49
aI193
aI171
aI477
aI363
aI735
aI1379
aI8977
aI9759
aI56477
aI99495
aI170433
asS'd'
I15772
sa(dp31544
S'a'
I39726
sS's'
I18
sS'm'
(lp31545
I1
aI1
aI3
aI7
aI25
aI25
aI77
aI31
aI21
aI1001
aI1527
aI1725
aI6479
aI8927
aI11249
aI63969
aI86291
aI74391
asS'd'
I15773
sa(dp31546
S'a'
I39731
sS's'
I18
sS'm'
(lp31547
I1
aI3
aI5
aI13
aI1
aI43
aI27
aI7
aI507
aI569
aI251
aI2199
aI3895
aI7845
aI13641
aI1655
aI112949
aI119493
asS'd'
I15774
sa(dp31548
S'a'
I39745
sS's'
I18
sS'm'
(lp31549
I1
aI3
aI7
aI15
aI7
aI1
aI123
aI27
aI121
aI261
aI201
aI1469
aI4229
aI2933
aI25157
aI1919
aI127937
aI21607
asS'd'
I15775
sa(dp31550
S'a'
I39776
sS's'
I18
sS'm'
(lp31551
I1
aI1
aI3
aI9
aI29
aI59
aI47
aI81
aI293
aI191
aI401
aI849
aI4355
aI1643
aI23533
aI8469
aI389
aI97891
asS'd'
I15776
sa(dp31552
S'a'
I39796
sS's'
I18
sS'm'
(lp31553
I1
aI3
aI5
aI5
aI9
aI55
aI37
aI175
aI203
aI179
aI901
aI3473
aI1489
aI1009
aI24623
aI54895
aI8711
aI190271
asS'd'
I15777
sa(dp31554
S'a'
I39809
sS's'
I18
sS'm'
(lp31555
I1
aI1
aI7
aI1
aI13
aI39
aI49
aI105
aI385
aI189
aI1079
aI2799
aI5901
aI2511
aI23199
aI58925
aI111727
aI131193
asS'd'
I15778
sa(dp31556
S'a'
I39833
sS's'
I18
sS'm'
(lp31557
I1
aI1
aI1
aI7
aI31
aI63
aI37
aI181
aI493
aI745
aI1131
aI223
aI8055
aI9507
aI26667
aI22163
aI26495
aI200945
asS'd'
I15779
sa(dp31558
S'a'
I39850
sS's'
I18
sS'm'
(lp31559
I1
aI3
aI7
aI1
aI25
aI15
aI127
aI71
aI445
aI935
aI1439
aI1167
aI3751
aI799
aI27253
aI46209
aI33413
aI38553
asS'd'
I15780
sa(dp31560
S'a'
I39869
sS's'
I18
sS'm'
(lp31561
I1
aI3
aI3
aI11
aI29
aI35
aI125
aI77
aI129
aI851
aI731
aI3259
aI4651
aI4137
aI20921
aI19779
aI119261
aI141507
asS'd'
I15781
sa(dp31562
S'a'
I39882
sS's'
I18
sS'm'
(lp31563
I1
aI3
aI1
aI9
aI11
aI13
aI31
aI211
aI87
aI377
aI547
aI113
aI1071
aI7167
aI28377
aI52943
aI50669
aI156915
asS'd'
I15782
sa(dp31564
S'a'
I39906
sS's'
I18
sS'm'
(lp31565
I1
aI1
aI3
aI7
aI29
aI55
aI89
aI215
aI425
aI513
aI175
aI1145
aI6995
aI1929
aI14253
aI20563
aI118543
aI104403
asS'd'
I15783
sa(dp31566
S'a'
I39918
sS's'
I18
sS'm'
(lp31567
I1
aI1
aI1
aI5
aI23
aI25
aI1
aI23
aI175
aI571
aI1597
aI3801
aI1411
aI1783
aI13045
aI37499
aI86831
aI139101
asS'd'
I15784
sa(dp31568
S'a'
I39929
sS's'
I18
sS'm'
(lp31569
I1
aI3
aI3
aI13
aI15
aI19
aI35
aI139
aI483
aI17
aI1555
aI3431
aI3417
aI13695
aI15985
aI65243
aI96659
aI76027
asS'd'
I15785
sa(dp31570
S'a'
I39947
sS's'
I18
sS'm'
(lp31571
I1
aI3
aI7
aI9
aI23
aI7
aI17
aI89
aI33
aI353
aI1999
aI2561
aI331
aI15661
aI25757
aI63389
aI112913
aI131757
asS'd'
I15786
sa(dp31572
S'a'
I39985
sS's'
I18
sS'm'
(lp31573
I1
aI3
aI1
aI11
aI27
aI59
aI37
aI75
aI121
aI429
aI1833
aI3243
aI2029
aI2601
aI5013
aI29433
aI123565
aI234803
asS'd'
I15787
sa(dp31574
S'a'
I39986
sS's'
I18
sS'm'
(lp31575
I1
aI3
aI1
aI7
aI31
aI13
aI33
aI25
aI459
aI803
aI267
aI1573
aI5579
aI4575
aI8125
aI7491
aI72681
aI239409
asS'd'
I15788
sa(dp31576
S'a'
I40023
sS's'
I18
sS'm'
(lp31577
I1
aI1
aI3
aI7
aI31
aI43
aI93
aI191
aI237
aI75
aI1809
aI3257
aI4131
aI1983
aI29153
aI23205
aI38393
aI197859
asS'd'
I15789
sa(dp31578
S'a'
I40024
sS's'
I18
sS'm'
(lp31579
I1
aI3
aI5
aI5
aI17
aI47
aI17
aI153
aI499
aI529
aI1515
aI1587
aI2951
aI12431
aI12787
aI13245
aI54117
aI82663
asS'd'
I15790
sa(dp31580
S'a'
I40039
sS's'
I18
sS'm'
(lp31581
I1
aI1
aI1
aI13
aI7
aI23
aI23
aI7
aI441
aI991
aI641
aI2713
aI151
aI1863
aI6065
aI47381
aI60493
aI136325
asS'd'
I15791
sa(dp31582
S'a'
I40053
sS's'
I18
sS'm'
(lp31583
I1
aI3
aI3
aI11
aI11
aI15
aI31
aI137
aI285
aI439
aI835
aI3033
aI6083
aI7883
aI3405
aI35803
aI65059
aI150143
asS'd'
I15792
sa(dp31584
S'a'
I40079
sS's'
I18
sS'm'
(lp31585
I1
aI1
aI7
aI3
aI19
aI47
aI61
aI163
aI313
aI813
aI1315
aI2995
aI2805
aI14397
aI6589
aI62123
aI46229
aI206697
asS'd'
I15793
sa(dp31586
S'a'
I40084
sS's'
I18
sS'm'
(lp31587
I1
aI3
aI5
aI5
aI27
aI51
aI25
aI99
aI241
aI571
aI1411
aI1191
aI7095
aI8639
aI29195
aI53733
aI53219
aI42851
asS'd'
I15794
sa(dp31588
S'a'
I40087
sS's'
I18
sS'm'
(lp31589
I1
aI3
aI1
aI5
aI11
aI29
aI1
aI49
aI61
aI149
aI1931
aI29
aI7163
aI3717
aI525
aI42375
aI71451
aI8345
asS'd'
I15795
sa(dp31590
S'a'
I40094
sS's'
I18
sS'm'
(lp31591
I1
aI3
aI3
aI3
aI13
aI19
aI97
aI249
aI265
aI509
aI1347
aI3081
aI6535
aI7941
aI31565
aI59897
aI91909
aI171789
asS'd'
I15796
sa(dp31592
S'a'
I40100
sS's'
I18
sS'm'
(lp31593
I1
aI3
aI3
aI1
aI25
aI17
aI75
aI169
aI251
aI607
aI73
aI549
aI1397
aI10661
aI1743
aI9615
aI41327
aI243207
asS'd'
I15797
sa(dp31594
S'a'
I40112
sS's'
I18
sS'm'
(lp31595
I1
aI3
aI1
aI11
aI7
aI7
aI15
aI181
aI385
aI883
aI651
aI2939
aI5457
aI15309
aI9807
aI22221
aI72893
aI146331
asS'd'
I15798
sa(dp31596
S'a'
I40129
sS's'
I18
sS'm'
(lp31597
I1
aI3
aI3
aI5
aI7
aI7
aI53
aI75
aI139
aI459
aI1861
aI917
aI4101
aI10379
aI18555
aI12633
aI70023
aI254761
asS'd'
I15799
sa(dp31598
S'a'
I40132
sS's'
I18
sS'm'
(lp31599
I1
aI3
aI3
aI1
aI17
aI51
aI5
aI109
aI471
aI3
aI1555
aI3731
aI6711
aI9791
aI63
aI61931
aI75269
aI138697
asS'd'
I15800
sa(dp31600
S'a'
I40142
sS's'
I18
sS'm'
(lp31601
I1
aI1
aI7
aI15
aI11
aI1
aI53
aI141
aI423
aI567
aI1937
aI849
aI5657
aI7437
aI32657
aI16253
aI115219
aI106027
asS'd'
I15801
sa(dp31602
S'a'
I40154
sS's'
I18
sS'm'
(lp31603
I1
aI3
aI5
aI1
aI17
aI29
aI65
aI213
aI443
aI541
aI697
aI3859
aI1463
aI16081
aI23299
aI7645
aI19475
aI77857
asS'd'
I15802
sa(dp31604
S'a'
I40177
sS's'
I18
sS'm'
(lp31605
I1
aI1
aI7
aI3
aI21
aI43
aI99
aI101
aI329
aI755
aI1123
aI1277
aI1381
aI7017
aI21763
aI28243
aI109995
aI178377
asS'd'
I15803
sa(dp31606
S'a'
I40178
sS's'
I18
sS'm'
(lp31607
I1
aI3
aI5
aI7
aI9
aI31
aI103
aI123
aI43
aI895
aI1925
aI3383
aI3539
aI13669
aI873
aI57361
aI45281
aI256517
asS'd'
I15804
sa(dp31608
S'a'
I40202
sS's'
I18
sS'm'
(lp31609
I1
aI1
aI1
aI13
aI1
aI37
aI115
aI55
aI415
aI703
aI1217
aI939
aI1145
aI4015
aI7233
aI44799
aI79711
aI164725
asS'd'
I15805
sa(dp31610
S'a'
I40219
sS's'
I18
sS'm'
(lp31611
I1
aI1
aI7
aI1
aI29
aI17
aI101
aI15
aI205
aI281
aI1059
aI301
aI753
aI11953
aI10533
aI31881
aI67741
aI12683
asS'd'
I15806
sa(dp31612
S'a'
I40303
sS's'
I18
sS'm'
(lp31613
I1
aI1
aI7
aI13
aI9
aI23
aI31
aI237
aI181
aI813
aI1765
aI2237
aI4897
aI9955
aI2139
aI13113
aI123423
aI227629
asS'd'
I15807
sa(dp31614
S'a'
I40305
sS's'
I18
sS'm'
(lp31615
I1
aI1
aI7
aI15
aI27
aI57
aI37
aI75
aI405
aI185
aI1671
aI2245
aI7151
aI10531
aI13161
aI15695
aI107547
aI47689
asS'd'
I15808
sa(dp31616
S'a'
I40311
sS's'
I18
sS'm'
(lp31617
I1
aI1
aI5
aI15
aI17
aI53
aI75
aI251
aI277
aI1001
aI179
aI589
aI1401
aI4937
aI11601
aI47113
aI36677
aI39263
asS'd'
I15809
sa(dp31618
S'a'
I40336
sS's'
I18
sS'm'
(lp31619
I1
aI3
aI5
aI15
aI23
aI47
aI53
aI81
aI115
aI547
aI1363
aI2457
aI4407
aI10861
aI25649
aI64013
aI44747
aI97949
asS'd'
I15810
sa(dp31620
S'a'
I40342
sS's'
I18
sS'm'
(lp31621
I1
aI1
aI3
aI1
aI25
aI29
aI121
aI43
aI205
aI591
aI211
aI1899
aI5835
aI739
aI19627
aI60387
aI127369
aI11255
asS'd'
I15811
sa(dp31622
S'a'
I40358
sS's'
I18
sS'm'
(lp31623
I1
aI1
aI3
aI15
aI31
aI11
aI93
aI227
aI501
aI731
aI1355
aI3963
aI347
aI83
aI12595
aI56431
aI80049
aI42103
asS'd'
I15812
sa(dp31624
S'a'
I40381
sS's'
I18
sS'm'
(lp31625
I1
aI3
aI1
aI11
aI13
aI17
aI51
aI165
aI311
aI67
aI1873
aI1493
aI3815
aI13209
aI11637
aI5809
aI94219
aI118077
asS'd'
I15813
sa(dp31626
S'a'
I40417
sS's'
I18
sS'm'
(lp31627
I1
aI1
aI7
aI15
aI19
aI17
aI13
aI73
aI365
aI413
aI1215
aI2265
aI2173
aI8725
aI4725
aI1373
aI76733
aI95379
asS'd'
I15814
sa(dp31628
S'a'
I40438
sS's'
I18
sS'm'
(lp31629
I1
aI3
aI1
aI7
aI7
aI61
aI13
aI145
aI205
aI113
aI1579
aI3851
aI7515
aI10659
aI28665
aI5277
aI65925
aI10141
asS'd'
I15815
sa(dp31630
S'a'
I40463
sS's'
I18
sS'm'
(lp31631
I1
aI3
aI7
aI1
aI9
aI63
aI11
aI83
aI197
aI797
aI1065
aI1521
aI1751
aI7423
aI7473
aI4371
aI29533
aI225167
asS'd'
I15816
sa(dp31632
S'a'
I40471
sS's'
I18
sS'm'
(lp31633
I1
aI3
aI7
aI7
aI3
aI63
aI71
aI177
aI53
aI279
aI1837
aI2609
aI7819
aI7285
aI11059
aI65247
aI102869
aI24429
asS'd'
I15817
sa(dp31634
S'a'
I40472
sS's'
I18
sS'm'
(lp31635
I1
aI3
aI3
aI13
aI9
aI21
aI123
aI125
aI367
aI85
aI85
aI1009
aI1009
aI7779
aI3375
aI30999
aI5035
aI215107
asS'd'
I15818
sa(dp31636
S'a'
I40488
sS's'
I18
sS'm'
(lp31637
I1
aI3
aI3
aI9
aI31
aI53
aI5
aI43
aI483
aI483
aI1359
aI2605
aI377
aI4243
aI13291
aI50211
aI118603
aI259865
asS'd'
I15819
sa(dp31638
S'a'
I40491
sS's'
I18
sS'm'
(lp31639
I1
aI1
aI1
aI5
aI19
aI37
aI109
aI139
aI373
aI79
aI1951
aI3379
aI5679
aI6445
aI29127
aI56229
aI97369
aI232561
asS'd'
I15820
sa(dp31640
S'a'
I40525
sS's'
I18
sS'm'
(lp31641
I1
aI1
aI3
aI5
aI19
aI37
aI61
aI225
aI321
aI573
aI1831
aI971
aI6507
aI10005
aI6837
aI16433
aI70913
aI170873
asS'd'
I15821
sa(dp31642
S'a'
I40526
sS's'
I18
sS'm'
(lp31643
I1
aI1
aI5
aI11
aI31
aI17
aI21
aI29
aI329
aI679
aI869
aI389
aI5121
aI1819
aI3539
aI43793
aI31617
aI204983
asS'd'
I15822
sa(dp31644
S'a'
I40550
sS's'
I18
sS'm'
(lp31645
I1
aI1
aI7
aI7
aI21
aI11
aI83
aI97
aI297
aI275
aI1559
aI1899
aI4957
aI11463
aI25647
aI21095
aI70121
aI113395
asS'd'
I15823
sa(dp31646
S'a'
I40553
sS's'
I18
sS'm'
(lp31647
I1
aI3
aI5
aI11
aI19
aI57
aI39
aI37
aI441
aI715
aI383
aI4083
aI1937
aI12263
aI6989
aI36159
aI118135
aI238475
asS'd'
I15824
sa(dp31648
S'a'
I40562
sS's'
I18
sS'm'
(lp31649
I1
aI1
aI1
aI3
aI9
aI53
aI85
aI201
aI357
aI807
aI865
aI1621
aI1993
aI7623
aI3165
aI1005
aI93343
aI227765
asS'd'
I15825
sa(dp31650
S'a'
I40568
sS's'
I18
sS'm'
(lp31651
I1
aI1
aI7
aI7
aI21
aI29
aI123
aI175
aI319
aI621
aI303
aI117
aI5589
aI12511
aI26053
aI41603
aI78439
aI71819
asS'd'
I15826
sa(dp31652
S'a'
I40580
sS's'
I18
sS'm'
(lp31653
I1
aI1
aI7
aI15
aI31
aI47
aI75
aI225
aI295
aI67
aI1349
aI1749
aI1363
aI8763
aI9153
aI4059
aI72015
aI3155
asS'd'
I15827
sa(dp31654
S'a'
I40608
sS's'
I18
sS'm'
(lp31655
I1
aI3
aI5
aI13
aI19
aI23
aI79
aI25
aI319
aI475
aI1517
aI2757
aI4009
aI12663
aI535
aI51617
aI55695
aI64399
asS'd'
I15828
sa(dp31656
S'a'
I40613
sS's'
I18
sS'm'
(lp31657
I1
aI3
aI7
aI13
aI19
aI39
aI61
aI235
aI369
aI951
aI111
aI2169
aI353
aI15371
aI8611
aI42477
aI130981
aI97419
asS'd'
I15829
sa(dp31658
S'a'
I40652
sS's'
I18
sS'm'
(lp31659
I1
aI1
aI3
aI13
aI27
aI31
aI115
aI201
aI3
aI291
aI793
aI237
aI3593
aI2307
aI12745
aI54603
aI96451
aI80853
asS'd'
I15830
sa(dp31660
S'a'
I40703
sS's'
I18
sS'm'
(lp31661
I1
aI3
aI3
aI11
aI11
aI35
aI43
aI1
aI35
aI415
aI1307
aI2303
aI5407
aI6883
aI29023
aI31271
aI119721
aI90599
asS'd'
I15831
sa(dp31662
S'a'
I40712
sS's'
I18
sS'm'
(lp31663
I1
aI1
aI5
aI9
aI21
aI23
aI3
aI1
aI333
aI463
aI1277
aI1165
aI6521
aI4887
aI16029
aI32537
aI43681
aI21633
asS'd'
I15832
sa(dp31664
S'a'
I40720
sS's'
I18
sS'm'
(lp31665
I1
aI1
aI1
aI13
aI1
aI35
aI45
aI57
aI293
aI435
aI1113
aI2477
aI6641
aI14083
aI28489
aI26189
aI44695
aI220481
asS'd'
I15833
sa(dp31666
S'a'
I40723
sS's'
I18
sS'm'
(lp31667
I1
aI3
aI5
aI5
aI5
aI31
aI75
aI149
aI309
aI921
aI941
aI1063
aI7041
aI12651
aI29533
aI46955
aI88133
aI89989
asS'd'
I15834
sa(dp31668
S'a'
I40746
sS's'
I18
sS'm'
(lp31669
I1
aI1
aI3
aI5
aI15
aI23
aI127
aI143
aI193
aI739
aI281
aI991
aI3731
aI16243
aI25483
aI24979
aI102317
aI186657
asS'd'
I15835
sa(dp31670
S'a'
I40759
sS's'
I18
sS'm'
(lp31671
I1
aI1
aI3
aI13
aI3
aI63
aI23
aI69
aI181
aI163
aI1733
aI893
aI5513
aI1525
aI31483
aI15033
aI108021
aI167875
asS'd'
I15836
sa(dp31672
S'a'
I40765
sS's'
I18
sS'm'
(lp31673
I1
aI1
aI5
aI15
aI15
aI51
aI79
aI59
aI55
aI243
aI565
aI159
aI7925
aI8393
aI20059
aI35011
aI53779
aI166241
asS'd'
I15837
sa(dp31674
S'a'
I40771
sS's'
I18
sS'm'
(lp31675
I1
aI1
aI3
aI5
aI11
aI57
aI85
aI175
aI495
aI999
aI1577
aI2377
aI715
aI2473
aI16979
aI5949
aI87691
aI195607
asS'd'
I15838
sa(dp31676
S'a'
I40778
sS's'
I18
sS'm'
(lp31677
I1
aI1
aI1
aI13
aI17
aI21
aI53
aI73
aI187
aI63
aI335
aI3251
aI4439
aI5701
aI13469
aI23567
aI70125
aI68931
asS'd'
I15839
sa(dp31678
S'a'
I40788
sS's'
I18
sS'm'
(lp31679
I1
aI1
aI1
aI13
aI23
aI11
aI55
aI75
aI37
aI845
aI745
aI2193
aI7113
aI5657
aI29449
aI41153
aI115547
aI87261
asS'd'
I15840
sa(dp31680
S'a'
I40804
sS's'
I18
sS'm'
(lp31681
I1
aI3
aI3
aI9
aI21
aI39
aI47
aI145
aI301
aI883
aI625
aI2479
aI1089
aI3393
aI23265
aI49577
aI81027
aI186485
asS'd'
I15841
sa(dp31682
S'a'
I40835
sS's'
I18
sS'm'
(lp31683
I1
aI1
aI3
aI7
aI3
aI11
aI37
aI117
aI79
aI905
aI493
aI265
aI1819
aI12179
aI12361
aI27457
aI14459
aI231837
asS'd'
I15842
sa(dp31684
S'a'
I40841
sS's'
I18
sS'm'
(lp31685
I1
aI3
aI5
aI11
aI19
aI45
aI99
aI5
aI455
aI497
aI1851
aI2349
aI5213
aI3671
aI5871
aI43187
aI59011
aI211167
asS'd'
I15843
sa(dp31686
S'a'
I40907
sS's'
I18
sS'm'
(lp31687
I1
aI3
aI7
aI5
aI23
aI61
aI63
aI97
aI413
aI575
aI1073
aI2587
aI573
aI1805
aI32307
aI58463
aI84927
aI15065
asS'd'
I15844
sa(dp31688
S'a'
I40934
sS's'
I18
sS'm'
(lp31689
I1
aI3
aI1
aI13
aI23
aI9
aI39
aI1
aI53
aI383
aI1277
aI3599
aI7719
aI16175
aI4443
aI53143
aI24345
aI111899
asS'd'
I15845
sa(dp31690
S'a'
I40955
sS's'
I18
sS'm'
(lp31691
I1
aI1
aI7
aI1
aI19
aI37
aI103
aI245
aI253
aI5
aI1367
aI3127
aI4689
aI5089
aI30697
aI45513
aI111291
aI26599
asS'd'
I15846
sa(dp31692
S'a'
I40986
sS's'
I18
sS'm'
(lp31693
I1
aI3
aI3
aI7
aI31
aI31
aI107
aI163
aI1
aI855
aI163
aI875
aI7481
aI5325
aI30107
aI19377
aI3167
aI5613
asS'd'
I15847
sa(dp31694
S'a'
I41007
sS's'
I18
sS'm'
(lp31695
I1
aI1
aI5
aI13
aI21
aI17
aI115
aI203
aI233
aI333
aI441
aI3185
aI3197
aI3397
aI8515
aI61879
aI11163
aI233277
asS'd'
I15848
sa(dp31696
S'a'
I41016
sS's'
I18
sS'm'
(lp31697
I1
aI3
aI5
aI3
aI17
aI53
aI93
aI233
aI465
aI573
aI1075
aI1905
aI1141
aI4965
aI13469
aI24901
aI23653
aI225233
asS'd'
I15849
sa(dp31698
S'a'
I41034
sS's'
I18
sS'm'
(lp31699
I1
aI3
aI7
aI11
aI11
aI1
aI95
aI47
aI85
aI65
aI9
aI2413
aI7347
aI2127
aI305
aI4673
aI79281
aI188081
asS'd'
I15850
sa(dp31700
S'a'
I41041
sS's'
I18
sS'm'
(lp31701
I1
aI1
aI7
aI5
aI9
aI5
aI27
aI23
aI393
aI201
aI467
aI3677
aI2641
aI4671
aI24627
aI18927
aI45137
aI74167
asS'd'
I15851
sa(dp31702
S'a'
I41063
sS's'
I18
sS'm'
(lp31703
I1
aI3
aI5
aI11
aI11
aI9
aI19
aI247
aI423
aI693
aI1885
aI1129
aI7459
aI8411
aI2573
aI54111
aI98919
aI160075
asS'd'
I15852
sa(dp31704
S'a'
I41084
sS's'
I18
sS'm'
(lp31705
I1
aI3
aI3
aI3
aI1
aI3
aI67
aI131
aI317
aI915
aI151
aI3609
aI4083
aI6395
aI12877
aI44017
aI28877
aI244165
asS'd'
I15853
sa(dp31706
S'a'
I41093
sS's'
I18
sS'm'
(lp31707
I1
aI3
aI1
aI3
aI1
aI33
aI29
aI23
aI19
aI323
aI873
aI115
aI2439
aI4699
aI5449
aI51637
aI68889
aI105197
asS'd'
I15854
sa(dp31708
S'a'
I41097
sS's'
I18
sS'm'
(lp31709
I1
aI1
aI7
aI1
aI19
aI55
aI37
aI241
aI53
aI695
aI729
aI1565
aI19
aI12875
aI26993
aI18511
aI35615
aI169281
asS'd'
I15855
sa(dp31710
S'a'
I41111
sS's'
I18
sS'm'
(lp31711
I1
aI3
aI3
aI1
aI1
aI7
aI121
aI49
aI345
aI883
aI1001
aI657
aI2647
aI15387
aI30633
aI18107
aI13745
aI217735
asS'd'
I15856
sa(dp31712
S'a'
I41118
sS's'
I18
sS'm'
(lp31713
I1
aI3
aI5
aI15
aI11
aI45
aI73
aI77
aI307
aI373
aI1723
aI335
aI473
aI5735
aI11747
aI39257
aI87199
aI47663
asS'd'
I15857
sa(dp31714
S'a'
I41121
sS's'
I18
sS'm'
(lp31715
I1
aI1
aI7
aI7
aI27
aI21
aI121
aI169
aI427
aI605
aI1593
aI3147
aI1001
aI3773
aI31505
aI22823
aI21543
aI82931
asS'd'
I15858
sa(dp31716
S'a'
I41141
sS's'
I18
sS'm'
(lp31717
I1
aI1
aI1
aI9
aI11
aI59
aI91
aI165
aI249
aI859
aI483
aI3133
aI5729
aI12675
aI7761
aI6475
aI116823
aI224187
asS'd'
I15859
sa(dp31718
S'a'
I41160
sS's'
I18
sS'm'
(lp31719
I1
aI1
aI3
aI3
aI27
aI31
aI51
aI1
aI429
aI517
aI1439
aI3959
aI2343
aI6709
aI5287
aI24039
aI52409
aI20749
asS'd'
I15860
sa(dp31720
S'a'
I41207
sS's'
I18
sS'm'
(lp31721
I1
aI3
aI7
aI7
aI11
aI31
aI83
aI111
aI391
aI729
aI721
aI1675
aI5679
aI14637
aI22065
aI49903
aI113759
aI40881
asS'd'
I15861
sa(dp31722
S'a'
I41214
sS's'
I18
sS'm'
(lp31723
I1
aI3
aI7
aI1
aI25
aI15
aI91
aI59
aI87
aI313
aI155
aI1439
aI2419
aI2099
aI22709
aI10289
aI40655
aI17351
asS'd'
I15862
sa(dp31724
S'a'
I41255
sS's'
I18
sS'm'
(lp31725
I1
aI3
aI5
aI1
aI15
aI5
aI11
aI21
aI261
aI227
aI1563
aI1177
aI4731
aI3487
aI1607
aI46599
aI105599
aI193425
asS'd'
I15863
sa(dp31726
S'a'
I41284
sS's'
I18
sS'm'
(lp31727
I1
aI1
aI5
aI5
aI5
aI45
aI77
aI181
aI431
aI27
aI1985
aI881
aI2555
aI7589
aI16199
aI31041
aI66683
aI52499
asS'd'
I15864
sa(dp31728
S'a'
I41287
sS's'
I18
sS'm'
(lp31729
I1
aI1
aI1
aI15
aI5
aI29
aI111
aI209
aI335
aI747
aI93
aI3551
aI5951
aI14995
aI18451
aI33329
aI9117
aI167455
asS'd'
I15865
sa(dp31730
S'a'
I41308
sS's'
I18
sS'm'
(lp31731
I1
aI3
aI1
aI7
aI25
aI9
aI113
aI123
aI387
aI87
aI267
aI2251
aI3509
aI10829
aI32733
aI48025
aI58267
aI143553
asS'd'
I15866
sa(dp31732
S'a'
I41330
sS's'
I18
sS'm'
(lp31733
I1
aI3
aI5
aI15
aI17
aI17
aI65
aI107
aI175
aI427
aI733
aI797
aI3837
aI12773
aI27865
aI29481
aI4557
aI196163
asS'd'
I15867
sa(dp31734
S'a'
I41336
sS's'
I18
sS'm'
(lp31735
I1
aI3
aI1
aI3
aI1
aI53
aI93
aI175
aI509
aI351
aI1093
aI1039
aI6931
aI2691
aI14957
aI44395
aI84383
aI58915
asS'd'
I15868
sa(dp31736
S'a'
I41437
sS's'
I18
sS'm'
(lp31737
I1
aI1
aI1
aI11
aI1
aI43
aI61
aI123
aI377
aI813
aI1335
aI1597
aI147
aI13663
aI30781
aI47635
aI24111
aI64307
asS'd'
I15869
sa(dp31738
S'a'
I41444
sS's'
I18
sS'm'
(lp31739
I1
aI1
aI3
aI11
aI25
aI27
aI15
aI215
aI125
aI679
aI1491
aI3203
aI5403
aI4531
aI11839
aI44227
aI49239
aI110439
asS'd'
I15870
sa(dp31740
S'a'
I41456
sS's'
I18
sS'm'
(lp31741
I1
aI1
aI3
aI13
aI9
aI35
aI71
aI127
aI127
aI629
aI1363
aI585
aI6713
aI10637
aI6803
aI20963
aI47157
aI157781
asS'd'
I15871
sa(dp31742
S'a'
I41481
sS's'
I18
sS'm'
(lp31743
I1
aI1
aI3
aI11
aI13
aI21
aI117
aI241
aI365
aI175
aI1397
aI1279
aI4117
aI5427
aI24007
aI50711
aI465
aI225003
asS'd'
I15872
sa(dp31744
S'a'
I41489
sS's'
I18
sS'm'
(lp31745
I1
aI1
aI1
aI9
aI13
aI63
aI49
aI189
aI113
aI61
aI353
aI2221
aI7541
aI4075
aI5283
aI5505
aI51035
aI35191
asS'd'
I15873
sa(dp31746
S'a'
I41499
sS's'
I18
sS'm'
(lp31747
I1
aI1
aI7
aI9
aI11
aI37
aI123
aI63
aI331
aI691
aI1299
aI1661
aI3769
aI7827
aI32127
aI32149
aI7271
aI150363
asS'd'
I15874
sa(dp31748
S'a'
I41555
sS's'
I18
sS'm'
(lp31749
I1
aI1
aI1
aI5
aI25
aI5
aI127
aI195
aI413
aI657
aI479
aI879
aI6743
aI8959
aI791
aI22425
aI77119
aI180721
asS'd'
I15875
sa(dp31750
S'a'
I41561
sS's'
I18
sS'm'
(lp31751
I1
aI1
aI7
aI3
aI13
aI57
aI123
aI125
aI135
aI69
aI455
aI3363
aI1783
aI1557
aI20401
aI26707
aI130345
aI195881
asS'd'
I15876
sa(dp31752
S'a'
I41562
sS's'
I18
sS'm'
(lp31753
I1
aI3
aI3
aI7
aI27
aI19
aI125
aI71
aI201
aI1017
aI1285
aI3955
aI5255
aI14375
aI18163
aI28537
aI76157
aI247193
asS'd'
I15877
sa(dp31754
S'a'
I41571
sS's'
I18
sS'm'
(lp31755
I1
aI1
aI3
aI1
aI27
aI33
aI123
aI137
aI189
aI655
aI1891
aI2419
aI5195
aI97
aI32565
aI38581
aI62715
aI164697
asS'd'
I15878
sa(dp31756
S'a'
I41592
sS's'
I18
sS'm'
(lp31757
I1
aI1
aI1
aI15
aI9
aI47
aI23
aI147
aI197
aI503
aI1803
aI2953
aI2961
aI13787
aI10545
aI35465
aI53997
aI198655
asS'd'
I15879
sa(dp31758
S'a'
I41631
sS's'
I18
sS'm'
(lp31759
I1
aI3
aI1
aI9
aI9
aI43
aI65
aI237
aI119
aI621
aI1517
aI3479
aI4165
aI12797
aI14731
aI53131
aI105501
aI112845
asS'd'
I15880
sa(dp31760
S'a'
I41661
sS's'
I18
sS'm'
(lp31761
I1
aI3
aI3
aI15
aI3
aI23
aI47
aI163
aI469
aI363
aI1813
aI3107
aI6167
aI8987
aI26829
aI33099
aI6821
aI203017
asS'd'
I15881
sa(dp31762
S'a'
I41682
sS's'
I18
sS'm'
(lp31763
I1
aI1
aI1
aI9
aI1
aI11
aI85
aI11
aI251
aI907
aI395
aI3935
aI3403
aI229
aI16825
aI48337
aI103647
aI91425
asS'd'
I15882
sa(dp31764
S'a'
I41710
sS's'
I18
sS'm'
(lp31765
I1
aI3
aI1
aI5
aI17
aI57
aI21
aI181
aI31
aI27
aI235
aI4041
aI4927
aI8319
aI29765
aI61603
aI19081
aI75879
asS'd'
I15883
sa(dp31766
S'a'
I41724
sS's'
I18
sS'm'
(lp31767
I1
aI1
aI5
aI15
aI11
aI3
aI7
aI225
aI247
aI221
aI251
aI1979
aI1151
aI10829
aI26491
aI39705
aI41587
aI99063
asS'd'
I15884
sa(dp31768
S'a'
I41727
sS's'
I18
sS'm'
(lp31769
I1
aI3
aI5
aI5
aI19
aI23
aI57
aI127
aI467
aI409
aI43
aI829
aI3883
aI10767
aI24351
aI31365
aI115943
aI209231
asS'd'
I15885
sa(dp31770
S'a'
I41730
sS's'
I18
sS'm'
(lp31771
I1
aI1
aI7
aI15
aI3
aI51
aI17
aI251
aI219
aI33
aI1511
aI2027
aI4995
aI12277
aI7639
aI59895
aI85267
aI87735
asS'd'
I15886
sa(dp31772
S'a'
I41744
sS's'
I18
sS'm'
(lp31773
I1
aI3
aI7
aI3
aI7
aI29
aI93
aI57
aI427
aI235
aI1591
aI3475
aI1159
aI2387
aI851
aI43307
aI87081
aI151543
asS'd'
I15887
sa(dp31774
S'a'
I41769
sS's'
I18
sS'm'
(lp31775
I1
aI3
aI7
aI7
aI1
aI9
aI21
aI167
aI73
aI439
aI2035
aI2091
aI4563
aI4819
aI5591
aI57123
aI78739
aI63235
asS'd'
I15888
sa(dp31776
S'a'
I41795
sS's'
I18
sS'm'
(lp31777
I1
aI1
aI3
aI13
aI19
aI35
aI63
aI17
aI425
aI277
aI1669
aI931
aI597
aI5621
aI22367
aI1155
aI109099
aI169965
asS'd'
I15889
sa(dp31778
S'a'
I41798
sS's'
I18
sS'm'
(lp31779
I1
aI3
aI7
aI5
aI11
aI11
aI41
aI71
aI35
aI183
aI555
aI2631
aI5199
aI16381
aI16319
aI1851
aI121551
aI141711
asS'd'
I15890
sa(dp31780
S'a'
I41804
sS's'
I18
sS'm'
(lp31781
I1
aI3
aI5
aI5
aI21
aI17
aI3
aI95
aI47
aI1011
aI1757
aI3295
aI1111
aI16043
aI6377
aI16257
aI37941
aI206637
asS'd'
I15891
sa(dp31782
S'a'
I41826
sS's'
I18
sS'm'
(lp31783
I1
aI3
aI1
aI5
aI25
aI19
aI19
aI69
aI395
aI589
aI1311
aI1075
aI5763
aI12475
aI3633
aI40647
aI54487
aI97459
asS'd'
I15892
sa(dp31784
S'a'
I41852
sS's'
I18
sS'm'
(lp31785
I1
aI3
aI5
aI7
aI13
aI23
aI83
aI91
aI419
aI415
aI685
aI1685
aI2893
aI12953
aI30641
aI43565
aI11851
aI187837
asS'd'
I15893
sa(dp31786
S'a'
I41868
sS's'
I18
sS'm'
(lp31787
I1
aI1
aI5
aI11
aI27
aI1
aI61
aI155
aI279
aI737
aI215
aI2909
aI969
aI57
aI17979
aI34537
aI41861
aI243717
asS'd'
I15894
sa(dp31788
S'a'
I41910
sS's'
I18
sS'm'
(lp31789
I1
aI1
aI5
aI1
aI17
aI61
aI57
aI199
aI127
aI569
aI1109
aI3057
aI7301
aI16097
aI17579
aI25753
aI82653
aI237273
asS'd'
I15895
sa(dp31790
S'a'
I41914
sS's'
I18
sS'm'
(lp31791
I1
aI1
aI7
aI15
aI17
aI45
aI19
aI53
aI153
aI785
aI51
aI291
aI5261
aI1317
aI21163
aI44393
aI108131
aI254373
asS'd'
I15896
sa(dp31792
S'a'
I41942
sS's'
I18
sS'm'
(lp31793
I1
aI1
aI3
aI11
aI5
aI19
aI61
aI125
aI127
aI961
aI2019
aI1725
aI855
aI677
aI20853
aI38845
aI3239
aI95697
asS'd'
I15897
sa(dp31794
S'a'
I41969
sS's'
I18
sS'm'
(lp31795
I1
aI1
aI3
aI15
aI3
aI3
aI117
aI17
aI61
aI201
aI241
aI1759
aI4465
aI3985
aI6985
aI47703
aI58657
aI52633
asS'd'
I15898
sa(dp31796
S'a'
I41975
sS's'
I18
sS'm'
(lp31797
I1
aI3
aI3
aI11
aI31
aI39
aI107
aI171
aI19
aI825
aI1309
aI807
aI7787
aI10761
aI20215
aI9287
aI21553
aI179599
asS'd'
I15899
sa(dp31798
S'a'
I41976
sS's'
I18
sS'm'
(lp31799
I1
aI3
aI7
aI9
aI5
aI7
aI121
aI15
aI3
aI199
aI97
aI3177
aI5461
aI15713
aI27609
aI54349
aI24963
aI186279
asS'd'
I15900
sa(dp31800
S'a'
I41994
sS's'
I18
sS'm'
(lp31801
I1
aI3
aI1
aI15
aI11
aI9
aI123
aI187
aI363
aI5
aI837
aI451
aI1601
aI6597
aI10857
aI46893
aI83729
aI20409
asS'd'
I15901
sa(dp31802
S'a'
I42002
sS's'
I18
sS'm'
(lp31803
I1
aI3
aI1
aI9
aI1
aI53
aI71
aI191
aI217
aI165
aI1709
aI1827
aI1977
aI10073
aI11373
aI35311
aI26637
aI134519
asS'd'
I15902
sa(dp31804
S'a'
I42037
sS's'
I18
sS'm'
(lp31805
I1
aI3
aI5
aI11
aI31
aI55
aI101
aI189
aI277
aI347
aI629
aI223
aI785
aI5739
aI25505
aI55601
aI55017
aI212837
asS'd'
I15903
sa(dp31806
S'a'
I42056
sS's'
I18
sS'm'
(lp31807
I1
aI1
aI5
aI11
aI3
aI13
aI45
aI235
aI65
aI459
aI621
aI587
aI7105
aI6181
aI13193
aI5683
aI42935
aI198585
asS'd'
I15904
sa(dp31808
S'a'
I42079
sS's'
I18
sS'm'
(lp31809
I1
aI1
aI3
aI1
aI3
aI17
aI27
aI109
aI261
aI979
aI903
aI1499
aI4799
aI11759
aI10591
aI65429
aI74587
aI16629
asS'd'
I15905
sa(dp31810
S'a'
I42085
sS's'
I18
sS'm'
(lp31811
I1
aI1
aI5
aI1
aI13
aI63
aI29
aI11
aI441
aI151
aI611
aI4073
aI3933
aI6793
aI28845
aI39223
aI120823
aI49397
asS'd'
I15906
sa(dp31812
S'a'
I42089
sS's'
I18
sS'm'
(lp31813
I1
aI3
aI3
aI5
aI21
aI13
aI23
aI53
aI357
aI197
aI1327
aI1773
aI2961
aI11509
aI16585
aI10201
aI28451
aI45109
asS'd'
I15907
sa(dp31814
S'a'
I42123
sS's'
I18
sS'm'
(lp31815
I1
aI3
aI5
aI7
aI19
aI17
aI97
aI63
aI295
aI111
aI85
aI2981
aI6719
aI9193
aI15197
aI12117
aI2553
aI59909
asS'd'
I15908
sa(dp31816
S'a'
I42128
sS's'
I18
sS'm'
(lp31817
I1
aI1
aI5
aI5
aI15
aI3
aI9
aI85
aI333
aI379
aI1409
aI1445
aI4173
aI3953
aI833
aI48089
aI120249
aI122703
asS'd'
I15909
sa(dp31818
S'a'
I42140
sS's'
I18
sS'm'
(lp31819
I1
aI3
aI7
aI7
aI29
aI13
aI57
aI19
aI141
aI979
aI1991
aI4011
aI8125
aI3915
aI15753
aI1371
aI113771
aI117273
asS'd'
I15910
sa(dp31820
S'a'
I42149
sS's'
I18
sS'm'
(lp31821
I1
aI3
aI3
aI13
aI21
aI53
aI115
aI187
aI279
aI29
aI1355
aI1843
aI253
aI3531
aI8193
aI54731
aI6213
aI123467
asS'd'
I15911
sa(dp31822
S'a'
I42156
sS's'
I18
sS'm'
(lp31823
I1
aI3
aI3
aI1
aI25
aI29
aI109
aI19
aI37
aI385
aI901
aI3737
aI6247
aI9941
aI13185
aI2895
aI88819
aI53029
asS'd'
I15912
sa(dp31824
S'a'
I42186
sS's'
I18
sS'm'
(lp31825
I1
aI1
aI1
aI1
aI3
aI31
aI89
aI157
aI483
aI657
aI1833
aI2975
aI3187
aI631
aI28685
aI7277
aI4915
aI115955
asS'd'
I15913
sa(dp31826
S'a'
I42191
sS's'
I18
sS'm'
(lp31827
I1
aI3
aI3
aI1
aI25
aI41
aI13
aI99
aI385
aI303
aI297
aI419
aI7919
aI12411
aI757
aI9227
aI47867
aI120175
asS'd'
I15914
sa(dp31828
S'a'
I42205
sS's'
I18
sS'm'
(lp31829
I1
aI3
aI5
aI11
aI23
aI15
aI21
aI157
aI177
aI301
aI789
aI2791
aI5769
aI7809
aI2369
aI26123
aI116465
aI22595
asS'd'
I15915
sa(dp31830
S'a'
I42221
sS's'
I18
sS'm'
(lp31831
I1
aI1
aI5
aI9
aI17
aI63
aI45
aI239
aI465
aI811
aI1157
aI1443
aI8137
aI12587
aI26209
aI62057
aI59299
aI167171
asS'd'
I15916
sa(dp31832
S'a'
I42222
sS's'
I18
sS'm'
(lp31833
I1
aI3
aI3
aI7
aI3
aI45
aI41
aI17
aI341
aI461
aI571
aI541
aI989
aI4069
aI17531
aI46729
aI107915
aI47871
asS'd'
I15917
sa(dp31834
S'a'
I42230
sS's'
I18
sS'm'
(lp31835
I1
aI1
aI7
aI1
aI1
aI23
aI45
aI213
aI125
aI799
aI5
aI3443
aI2535
aI12983
aI2133
aI63411
aI93027
aI89831
asS'd'
I15918
sa(dp31836
S'a'
I42233
sS's'
I18
sS'm'
(lp31837
I1
aI1
aI1
aI15
aI7
aI31
aI49
aI181
aI213
aI923
aI281
aI2059
aI861
aI6951
aI25659
aI32209
aI66423
aI225885
asS'd'
I15919
sa(dp31838
S'a'
I42241
sS's'
I18
sS'm'
(lp31839
I1
aI3
aI3
aI7
aI9
aI39
aI107
aI197
aI383
aI179
aI27
aI1395
aI6397
aI16139
aI32049
aI33567
aI43977
aI203939
asS'd'
I15920
sa(dp31840
S'a'
I42247
sS's'
I18
sS'm'
(lp31841
I1
aI1
aI3
aI5
aI13
aI37
aI13
aI31
aI339
aI527
aI641
aI181
aI1413
aI8145
aI341
aI57605
aI108031
aI109311
asS'd'
I15921
sa(dp31842
S'a'
I42248
sS's'
I18
sS'm'
(lp31843
I1
aI1
aI1
aI1
aI21
aI35
aI87
aI15
aI279
aI967
aI1003
aI813
aI5075
aI10595
aI5609
aI33901
aI86443
aI150007
asS'd'
I15922
sa(dp31844
S'a'
I42254
sS's'
I18
sS'm'
(lp31845
I1
aI3
aI1
aI5
aI27
aI9
aI75
aI199
aI377
aI889
aI545
aI1987
aI6277
aI361
aI12563
aI35699
aI27105
aI187995
asS'd'
I15923
sa(dp31846
S'a'
I42259
sS's'
I18
sS'm'
(lp31847
I1
aI3
aI7
aI1
aI11
aI3
aI41
aI215
aI273
aI61
aI821
aI1207
aI2809
aI8731
aI26409
aI50323
aI22355
aI16521
asS'd'
I15924
sa(dp31848
S'a'
I42268
sS's'
I18
sS'm'
(lp31849
I1
aI1
aI7
aI5
aI29
aI13
aI99
aI133
aI15
aI673
aI215
aI445
aI4051
aI2187
aI9395
aI62491
aI58685
aI224707
asS'd'
I15925
sa(dp31850
S'a'
I42295
sS's'
I18
sS'm'
(lp31851
I1
aI3
aI3
aI13
aI21
aI59
aI17
aI1
aI271
aI613
aI939
aI1881
aI2379
aI16325
aI3275
aI63707
aI59961
aI106937
asS'd'
I15926
sa(dp31852
S'a'
I42304
sS's'
I18
sS'm'
(lp31853
I1
aI3
aI3
aI1
aI23
aI53
aI51
aI181
aI391
aI375
aI767
aI239
aI373
aI4593
aI25211
aI37173
aI70409
aI252345
asS'd'
I15927
sa(dp31854
S'a'
I42322
sS's'
I18
sS'm'
(lp31855
I1
aI3
aI3
aI11
aI15
aI17
aI37
aI145
aI41
aI107
aI151
aI1351
aI3457
aI14727
aI755
aI36321
aI99397
aI73359
asS'd'
I15928
sa(dp31856
S'a'
I42334
sS's'
I18
sS'm'
(lp31857
I1
aI3
aI5
aI7
aI25
aI31
aI39
aI185
aI341
aI721
aI1799
aI1803
aI5985
aI10587
aI11605
aI9937
aI23905
aI56485
asS'd'
I15929
sa(dp31858
S'a'
I42355
sS's'
I18
sS'm'
(lp31859
I1
aI3
aI1
aI3
aI19
aI35
aI25
aI71
aI109
aI209
aI1675
aI4043
aI2053
aI6285
aI25317
aI50171
aI68293
aI124385
asS'd'
I15930
sa(dp31860
S'a'
I42361
sS's'
I18
sS'm'
(lp31861
I1
aI1
aI1
aI1
aI21
aI11
aI63
aI137
aI361
aI157
aI1985
aI2161
aI7239
aI1795
aI10459
aI38511
aI36817
aI253347
asS'd'
I15931
sa(dp31862
S'a'
I42373
sS's'
I18
sS'm'
(lp31863
I1
aI1
aI1
aI5
aI31
aI3
aI13
aI83
aI185
aI175
aI567
aI295
aI459
aI11453
aI3765
aI9841
aI30333
aI201377
asS'd'
I15932
sa(dp31864
S'a'
I42374
sS's'
I18
sS'm'
(lp31865
I1
aI1
aI3
aI13
aI5
aI57
aI99
aI125
aI371
aI903
aI2001
aI285
aI2005
aI8475
aI31617
aI58265
aI641
aI115507
asS'd'
I15933
sa(dp31866
S'a'
I42398
sS's'
I18
sS'm'
(lp31867
I1
aI3
aI5
aI9
aI11
aI17
aI123
aI67
aI77
aI803
aI815
aI3173
aI4795
aI11917
aI1187
aI32389
aI102289
aI248359
asS'd'
I15934
sa(dp31868
S'a'
I42411
sS's'
I18
sS'm'
(lp31869
I1
aI1
aI5
aI5
aI27
aI19
aI59
aI145
aI361
aI821
aI33
aI1465
aI7643
aI11101
aI145
aI21705
aI55105
aI181245
asS'd'
I15935
sa(dp31870
S'a'
I42443
sS's'
I18
sS'm'
(lp31871
I1
aI3
aI7
aI5
aI15
aI55
aI127
aI133
aI157
aI989
aI1211
aI3573
aI4021
aI2967
aI2941
aI38657
aI97681
aI114505
asS'd'
I15936
sa(dp31872
S'a'
I42446
sS's'
I18
sS'm'
(lp31873
I1
aI1
aI7
aI7
aI23
aI17
aI39
aI209
aI117
aI195
aI931
aI5
aI7509
aI9187
aI6011
aI10297
aI13727
aI258007
asS'd'
I15937
sa(dp31874
S'a'
I42463
sS's'
I18
sS'm'
(lp31875
I1
aI3
aI1
aI13
aI3
aI17
aI105
aI69
aI9
aI163
aI1615
aI241
aI5207
aI13173
aI28521
aI51417
aI130645
aI106787
asS'd'
I15938
sa(dp31876
S'a'
I42479
sS's'
I18
sS'm'
(lp31877
I1
aI1
aI7
aI13
aI15
aI3
aI5
aI187
aI5
aI239
aI1799
aI3083
aI7801
aI12781
aI24817
aI59341
aI73867
aI175273
asS'd'
I15939
sa(dp31878
S'a'
I42482
sS's'
I18
sS'm'
(lp31879
I1
aI1
aI3
aI11
aI11
aI1
aI89
aI17
aI497
aI217
aI1565
aI2933
aI6449
aI7687
aI6561
aI57903
aI92751
aI261371
asS'd'
I15940
sa(dp31880
S'a'
I42484
sS's'
I18
sS'm'
(lp31881
I1
aI1
aI1
aI1
aI15
aI39
aI47
aI249
aI39
aI765
aI249
aI2785
aI4401
aI16033
aI11463
aI127
aI120109
aI83133
asS'd'
I15941
sa(dp31882
S'a'
I42498
sS's'
I18
sS'm'
(lp31883
I1
aI1
aI3
aI3
aI19
aI59
aI47
aI153
aI505
aI1009
aI413
aI1177
aI5999
aI6841
aI12013
aI40295
aI115641
aI189241
asS'd'
I15942
sa(dp31884
S'a'
I42534
sS's'
I18
sS'm'
(lp31885
I1
aI3
aI3
aI11
aI17
aI51
aI45
aI101
aI459
aI971
aI1133
aI3467
aI4945
aI12445
aI1267
aI41783
aI66825
aI130167
asS'd'
I15943
sa(dp31886
S'a'
I42537
sS's'
I18
sS'm'
(lp31887
I1
aI3
aI1
aI7
aI17
aI45
aI77
aI211
aI387
aI23
aI1903
aI2309
aI2681
aI6897
aI6959
aI30981
aI113537
aI207415
asS'd'
I15944
sa(dp31888
S'a'
I42557
sS's'
I18
sS'm'
(lp31889
I1
aI1
aI3
aI1
aI19
aI9
aI103
aI89
aI461
aI881
aI71
aI2019
aI6475
aI13563
aI18835
aI2375
aI34807
aI1373
asS'd'
I15945
sa(dp31890
S'a'
I42570
sS's'
I18
sS'm'
(lp31891
I1
aI1
aI7
aI11
aI3
aI31
aI125
aI121
aI43
aI737
aI1995
aI3043
aI811
aI8883
aI8169
aI22131
aI29295
aI194963
asS'd'
I15946
sa(dp31892
S'a'
I42599
sS's'
I18
sS'm'
(lp31893
I1
aI3
aI7
aI11
aI7
aI45
aI125
aI149
aI427
aI187
aI1361
aI2405
aI2815
aI8877
aI15255
aI36867
aI95517
aI261969
asS'd'
I15947
sa(dp31894
S'a'
I42600
sS's'
I18
sS'm'
(lp31895
I1
aI1
aI7
aI11
aI9
aI29
aI89
aI175
aI467
aI997
aI937
aI3869
aI7843
aI12629
aI8701
aI60717
aI30443
aI193427
asS'd'
I15948
sa(dp31896
S'a'
I42633
sS's'
I18
sS'm'
(lp31897
I1
aI1
aI5
aI7
aI3
aI3
aI57
aI199
aI315
aI477
aI189
aI2029
aI2059
aI3473
aI27411
aI30419
aI26465
aI187807
asS'd'
I15949
sa(dp31898
S'a'
I42639
sS's'
I18
sS'm'
(lp31899
I1
aI3
aI1
aI15
aI3
aI3
aI11
aI55
aI233
aI603
aI1749
aI1737
aI5709
aI4559
aI13427
aI39137
aI44885
aI61611
asS'd'
I15950
sa(dp31900
S'a'
I42672
sS's'
I18
sS'm'
(lp31901
I1
aI1
aI7
aI1
aI25
aI15
aI61
aI199
aI107
aI737
aI909
aI3229
aI1799
aI5129
aI27655
aI45937
aI919
aI36161
asS'd'
I15951
sa(dp31902
S'a'
I42681
sS's'
I18
sS'm'
(lp31903
I1
aI1
aI3
aI5
aI31
aI27
aI99
aI247
aI425
aI689
aI1335
aI1661
aI625
aI15817
aI22601
aI33293
aI113927
aI261931
asS'd'
I15952
sa(dp31904
S'a'
I42687
sS's'
I18
sS'm'
(lp31905
I1
aI1
aI1
aI9
aI5
aI57
aI117
aI121
aI465
aI859
aI335
aI879
aI665
aI12787
aI21313
aI46387
aI16437
aI53769
asS'd'
I15953
sa(dp31906
S'a'
I42704
sS's'
I18
sS'm'
(lp31907
I1
aI1
aI3
aI1
aI11
aI11
aI75
aI145
aI307
aI621
aI833
aI235
aI3907
aI11331
aI6633
aI51905
aI72581
aI129613
asS'd'
I15954
sa(dp31908
S'a'
I42713
sS's'
I18
sS'm'
(lp31909
I1
aI1
aI3
aI13
aI17
aI13
aI81
aI125
aI377
aI499
aI1597
aI3437
aI4323
aI789
aI23825
aI46609
aI105997
aI159709
asS'd'
I15955
sa(dp31910
S'a'
I42714
sS's'
I18
sS'm'
(lp31911
I1
aI3
aI3
aI9
aI5
aI29
aI113
aI51
aI23
aI957
aI1981
aI3205
aI2549
aI9771
aI2555
aI44289
aI103893
aI170241
asS'd'
I15956
sa(dp31912
S'a'
I42716
sS's'
I18
sS'm'
(lp31913
I1
aI1
aI5
aI15
aI13
aI17
aI101
aI67
aI71
aI7
aI185
aI3775
aI5399
aI5213
aI13095
aI26045
aI59467
aI95547
asS'd'
I15957
sa(dp31914
S'a'
I42726
sS's'
I18
sS'm'
(lp31915
I1
aI1
aI5
aI5
aI31
aI11
aI77
aI169
aI3
aI1007
aI1853
aI2245
aI509
aI13489
aI2807
aI55227
aI67541
aI242835
asS'd'
I15958
sa(dp31916
S'a'
I42730
sS's'
I18
sS'm'
(lp31917
I1
aI3
aI7
aI15
aI11
aI63
aI39
aI97
aI1
aI219
aI1827
aI2343
aI6009
aI4909
aI4327
aI21853
aI46079
aI87719
asS'd'
I15959
sa(dp31918
S'a'
I42737
sS's'
I18
sS'm'
(lp31919
I1
aI3
aI1
aI9
aI17
aI51
aI119
aI93
aI179
aI607
aI1051
aI2381
aI619
aI11215
aI10839
aI44771
aI20555
aI12721
asS'd'
I15960
sa(dp31920
S'a'
I42752
sS's'
I18
sS'm'
(lp31921
I1
aI3
aI3
aI5
aI21
aI47
aI35
aI133
aI61
aI937
aI1561
aI1655
aI6527
aI5085
aI4141
aI60811
aI59971
aI6309
asS'd'
I15961
sa(dp31922
S'a'
I42757
sS's'
I18
sS'm'
(lp31923
I1
aI3
aI1
aI7
aI5
aI63
aI7
aI235
aI489
aI675
aI945
aI943
aI7107
aI6005
aI32695
aI27655
aI113219
aI132963
asS'd'
I15962
sa(dp31924
S'a'
I42772
sS's'
I18
sS'm'
(lp31925
I1
aI1
aI1
aI7
aI7
aI5
aI81
aI237
aI365
aI1
aI811
aI3075
aI1771
aI5223
aI7337
aI24601
aI68383
aI156975
asS'd'
I15963
sa(dp31926
S'a'
I42809
sS's'
I18
sS'm'
(lp31927
I1
aI3
aI1
aI15
aI5
aI35
aI39
aI91
aI301
aI387
aI805
aI3537
aI737
aI7453
aI4655
aI16349
aI108261
aI123697
asS'd'
I15964
sa(dp31928
S'a'
I42853
sS's'
I18
sS'm'
(lp31929
I1
aI3
aI5
aI3
aI9
aI59
aI95
aI187
aI155
aI183
aI589
aI2107
aI967
aI1095
aI4875
aI46131
aI100699
aI212797
asS'd'
I15965
sa(dp31930
S'a'
I42857
sS's'
I18
sS'm'
(lp31931
I1
aI3
aI1
aI7
aI27
aI29
aI77
aI133
aI397
aI445
aI933
aI1483
aI5027
aI12569
aI22163
aI58989
aI16657
aI195347
asS'd'
I15966
sa(dp31932
S'a'
I42902
sS's'
I18
sS'm'
(lp31933
I1
aI3
aI5
aI15
aI21
aI51
aI97
aI135
aI165
aI311
aI525
aI171
aI4785
aI7947
aI14649
aI40837
aI58875
aI222303
asS'd'
I15967
sa(dp31934
S'a'
I42912
sS's'
I18
sS'm'
(lp31935
I1
aI3
aI3
aI5
aI23
aI47
aI41
aI23
aI321
aI709
aI1555
aI1139
aI3775
aI11617
aI13001
aI18235
aI51803
aI197863
asS'd'
I15968
sa(dp31936
S'a'
I42930
sS's'
I18
sS'm'
(lp31937
I1
aI3
aI5
aI3
aI3
aI41
aI91
aI157
aI29
aI1005
aI945
aI3471
aI2563
aI8493
aI24961
aI44759
aI103079
aI50841
asS'd'
I15969
sa(dp31938
S'a'
I42936
sS's'
I18
sS'm'
(lp31939
I1
aI1
aI5
aI11
aI23
aI19
aI91
aI115
aI165
aI291
aI1653
aI1061
aI1067
aI6171
aI18441
aI26067
aI3569
aI117329
asS'd'
I15970
sa(dp31940
S'a'
I42956
sS's'
I18
sS'm'
(lp31941
I1
aI3
aI7
aI15
aI17
aI19
aI89
aI23
aI103
aI389
aI623
aI1071
aI203
aI9545
aI21259
aI36155
aI55395
aI141741
asS'd'
I15971
sa(dp31942
S'a'
I42959
sS's'
I18
sS'm'
(lp31943
I1
aI3
aI5
aI11
aI27
aI43
aI31
aI193
aI55
aI489
aI353
aI1615
aI7461
aI13977
aI31901
aI64051
aI82667
aI258825
asS'd'
I15972
sa(dp31944
S'a'
I42978
sS's'
I18
sS'm'
(lp31945
I1
aI1
aI3
aI7
aI5
aI39
aI7
aI245
aI241
aI843
aI1545
aI3499
aI8117
aI15057
aI14153
aI2665
aI107401
aI66059
asS'd'
I15973
sa(dp31946
S'a'
I43001
sS's'
I18
sS'm'
(lp31947
I1
aI3
aI5
aI11
aI9
aI11
aI41
aI171
aI255
aI153
aI1973
aI759
aI51
aI15601
aI327
aI25889
aI110861
aI20555
asS'd'
I15974
sa(dp31948
S'a'
I43011
sS's'
I18
sS'm'
(lp31949
I1
aI3
aI5
aI7
aI5
aI15
aI41
aI77
aI87
aI143
aI1141
aI3975
aI2675
aI7131
aI5549
aI52397
aI42073
aI27585
asS'd'
I15975
sa(dp31950
S'a'
I43047
sS's'
I18
sS'm'
(lp31951
I1
aI1
aI3
aI7
aI29
aI5
aI79
aI243
aI359
aI817
aI1053
aI3509
aI3347
aI6207
aI5147
aI31063
aI116851
aI132627
asS'd'
I15976
sa(dp31952
S'a'
I43059
sS's'
I18
sS'm'
(lp31953
I1
aI3
aI7
aI3
aI27
aI51
aI39
aI7
aI95
aI239
aI263
aI3497
aI867
aI1869
aI16773
aI46947
aI59193
aI37523
asS'd'
I15977
sa(dp31954
S'a'
I43094
sS's'
I18
sS'm'
(lp31955
I1
aI3
aI1
aI1
aI25
aI31
aI113
aI127
aI187
aI77
aI675
aI3307
aI999
aI12255
aI26441
aI30933
aI122761
aI116783
asS'd'
I15978
sa(dp31956
S'a'
I43124
sS's'
I18
sS'm'
(lp31957
I1
aI1
aI5
aI7
aI7
aI11
aI87
aI59
aI437
aI485
aI685
aI3159
aI7259
aI16271
aI24899
aI17919
aI130271
aI52953
asS'd'
I15979
sa(dp31958
S'a'
I43218
sS's'
I18
sS'm'
(lp31959
I1
aI1
aI3
aI11
aI7
aI13
aI61
aI157
aI149
aI1001
aI285
aI3631
aI1923
aI16013
aI19507
aI64447
aI8073
aI261171
asS'd'
I15980
sa(dp31960
S'a'
I43220
sS's'
I18
sS'm'
(lp31961
I1
aI3
aI7
aI9
aI7
aI13
aI45
aI7
aI225
aI671
aI287
aI1059
aI5223
aI2077
aI7551
aI58385
aI92955
aI162725
asS'd'
I15981
sa(dp31962
S'a'
I43240
sS's'
I18
sS'm'
(lp31963
I1
aI1
aI5
aI11
aI11
aI21
aI17
aI145
aI97
aI633
aI475
aI2639
aI2069
aI9663
aI23633
aI50949
aI109941
aI119865
asS'd'
I15982
sa(dp31964
S'a'
I43258
sS's'
I18
sS'm'
(lp31965
I1
aI1
aI5
aI13
aI9
aI49
aI127
aI171
aI199
aI413
aI1315
aI645
aI305
aI12123
aI9559
aI38319
aI99945
aI103313
asS'd'
I15983
sa(dp31966
S'a'
I43263
sS's'
I18
sS'm'
(lp31967
I1
aI3
aI5
aI3
aI19
aI21
aI119
aI17
aI301
aI611
aI1785
aI2081
aI2245
aI8761
aI4755
aI9507
aI23133
aI144575
asS'd'
I15984
sa(dp31968
S'a'
I43265
sS's'
I18
sS'm'
(lp31969
I1
aI3
aI7
aI11
aI17
aI25
aI101
aI59
aI397
aI249
aI687
aI715
aI1151
aI15941
aI20525
aI5171
aI24073
aI46257
asS'd'
I15985
sa(dp31970
S'a'
I43290
sS's'
I18
sS'm'
(lp31971
I1
aI1
aI3
aI15
aI11
aI45
aI9
aI201
aI421
aI867
aI389
aI3615
aI5965
aI10561
aI18309
aI17143
aI52771
aI230743
asS'd'
I15986
sa(dp31972
S'a'
I43346
sS's'
I18
sS'm'
(lp31973
I1
aI1
aI3
aI5
aI9
aI1
aI109
aI233
aI431
aI849
aI421
aI475
aI1331
aI9903
aI20649
aI34759
aI118611
aI38541
asS'd'
I15987
sa(dp31974
S'a'
I43364
sS's'
I18
sS'm'
(lp31975
I1
aI3
aI5
aI15
aI9
aI55
aI113
aI217
aI83
aI409
aI1449
aI2325
aI4825
aI11311
aI14565
aI65075
aI124399
aI3591
asS'd'
I15988
sa(dp31976
S'a'
I43373
sS's'
I18
sS'm'
(lp31977
I1
aI3
aI3
aI9
aI27
aI13
aI101
aI181
aI255
aI313
aI693
aI951
aI1153
aI13941
aI14097
aI8325
aI4589
aI102883
asS'd'
I15989
sa(dp31978
S'a'
I43401
sS's'
I18
sS'm'
(lp31979
I1
aI3
aI7
aI3
aI3
aI1
aI43
aI65
aI321
aI623
aI1389
aI57
aI3461
aI9965
aI6743
aI34843
aI91673
aI249573
asS'd'
I15990
sa(dp31980
S'a'
I43416
sS's'
I18
sS'm'
(lp31981
I1
aI3
aI5
aI5
aI11
aI29
aI101
aI79
aI275
aI685
aI569
aI59
aI6921
aI16065
aI30625
aI53339
aI32283
aI93401
asS'd'
I15991
sa(dp31982
S'a'
I43443
sS's'
I18
sS'm'
(lp31983
I1
aI3
aI1
aI11
aI29
aI33
aI103
aI145
aI431
aI289
aI1845
aI1915
aI23
aI699
aI5475
aI18413
aI127185
aI162637
asS'd'
I15992
sa(dp31984
S'a'
I43458
sS's'
I18
sS'm'
(lp31985
I1
aI3
aI1
aI11
aI27
aI23
aI71
aI63
aI45
aI579
aI1187
aI1189
aI1781
aI5595
aI9043
aI10155
aI33321
aI36487
asS'd'
I15993
sa(dp31986
S'a'
I43497
sS's'
I18
sS'm'
(lp31987
I1
aI3
aI3
aI15
aI17
aI19
aI85
aI233
aI293
aI833
aI1711
aI857
aI7573
aI393
aI23141
aI58713
aI21399
aI228381
asS'd'
I15994
sa(dp31988
S'a'
I43506
sS's'
I18
sS'm'
(lp31989
I1
aI3
aI3
aI15
aI3
aI35
aI33
aI35
aI403
aI123
aI575
aI3509
aI6303
aI13203
aI17997
aI15649
aI64331
aI124101
asS'd'
I15995
sa(dp31990
S'a'
I43545
sS's'
I18
sS'm'
(lp31991
I1
aI1
aI3
aI9
aI23
aI25
aI39
aI53
aI119
aI879
aI573
aI3225
aI5069
aI15489
aI21887
aI11773
aI37783
aI169523
asS'd'
I15996
sa(dp31992
S'a'
I43558
sS's'
I18
sS'm'
(lp31993
I1
aI3
aI5
aI5
aI5
aI25
aI29
aI189
aI145
aI863
aI661
aI3939
aI1059
aI11993
aI1487
aI3157
aI118287
aI69835
asS'd'
I15997
sa(dp31994
S'a'
I43576
sS's'
I18
sS'm'
(lp31995
I1
aI3
aI7
aI3
aI15
aI9
aI85
aI107
aI17
aI965
aI1097
aI2087
aI1947
aI14649
aI4099
aI50941
aI64511
aI209153
asS'd'
I15998
sa(dp31996
S'a'
I43579
sS's'
I18
sS'm'
(lp31997
I1
aI3
aI5
aI15
aI19
aI21
aI127
aI71
aI429
aI97
aI1989
aI835
aI743
aI11973
aI14635
aI45371
aI114657
aI208085
asS'd'
I15999
sa(dp31998
S'a'
I43624
sS's'
I18
sS'm'
(lp31999
I1
aI3
aI5
aI7
aI23
aI61
aI99
aI133
aI235
aI237
aI435
aI2681
aI331
aI7859
aI20859
aI3573
aI102901
aI775
asS'd'
I16000
sa(dp32000
S'a'
I43630
sS's'
I18
sS'm'
(lp32001
I1
aI3
aI3
aI9
aI7
aI57
aI29
aI15
aI53
aI569
aI871
aI1703
aI2573
aI10791
aI719
aI3487
aI105709
aI89573
asS'd'
I16001
sa(dp32002
S'a'
I43637
sS's'
I18
sS'm'
(lp32003
I1
aI1
aI1
aI9
aI7
aI21
aI21
aI207
aI459
aI621
aI737
aI635
aI5101
aI4343
aI4961
aI32067
aI64017
aI73675
asS'd'
I16002
sa(dp32004
S'a'
I43654
sS's'
I18
sS'm'
(lp32005
I1
aI1
aI3
aI9
aI25
aI27
aI77
aI69
aI251
aI327
aI921
aI3759
aI1715
aI14537
aI21399
aI10937
aI9085
aI241329
asS'd'
I16003
sa(dp32006
S'a'
I43660
sS's'
I18
sS'm'
(lp32007
I1
aI1
aI5
aI1
aI23
aI13
aI63
aI117
aI81
aI427
aI1063
aI1987
aI2433
aI14837
aI13473
aI28623
aI44799
aI202223
asS'd'
I16004
sa(dp32008
S'a'
I43696
sS's'
I18
sS'm'
(lp32009
I1
aI3
aI1
aI11
aI17
aI55
aI89
aI89
aI455
aI255
aI1009
aI1891
aI2197
aI9045
aI23543
aI48783
aI22871
aI58613
asS'd'
I16005
sa(dp32010
S'a'
I43706
sS's'
I18
sS'm'
(lp32011
I1
aI3
aI7
aI15
aI25
aI5
aI11
aI71
aI399
aI23
aI239
aI93
aI6681
aI11311
aI23403
aI58131
aI59549
aI38917
asS'd'
I16006
sa(dp32012
S'a'
I43725
sS's'
I18
sS'm'
(lp32013
I1
aI1
aI1
aI1
aI19
aI63
aI73
aI31
aI69
aI145
aI921
aI2475
aI3505
aI4797
aI23805
aI28621
aI101153
aI98895
asS'd'
I16007
sa(dp32014
S'a'
I43733
sS's'
I18
sS'm'
(lp32015
I1
aI3
aI3
aI3
aI9
aI1
aI113
aI111
aI275
aI851
aI519
aI1607
aI635
aI13287
aI6191
aI24545
aI112039
aI114383
asS'd'
I16008
sa(dp32016
S'a'
I43753
sS's'
I18
sS'm'
(lp32017
I1
aI3
aI7
aI9
aI13
aI13
aI107
aI97
aI37
aI133
aI21
aI1059
aI4201
aI6777
aI2843
aI43503
aI23761
aI13247
asS'd'
I16009
sa(dp32018
S'a'
I43756
sS's'
I18
sS'm'
(lp32019
I1
aI1
aI7
aI9
aI25
aI23
aI101
aI135
aI471
aI901
aI539
aI4083
aI1765
aI15553
aI5879
aI10787
aI67543
aI104543
asS'd'
I16010
sa(dp32020
S'a'
I43761
sS's'
I18
sS'm'
(lp32021
I1
aI3
aI3
aI13
aI25
aI63
aI1
aI49
aI287
aI357
aI1701
aI3689
aI3895
aI16313
aI22619
aI20203
aI8195
aI93127
asS'd'
I16011
sa(dp32022
S'a'
I43768
sS's'
I18
sS'm'
(lp32023
I1
aI1
aI1
aI11
aI15
aI37
aI75
aI43
aI311
aI443
aI1639
aI549
aI4707
aI3099
aI677
aI59115
aI11709
aI71013
asS'd'
I16012
sa(dp32024
S'a'
I43812
sS's'
I18
sS'm'
(lp32025
I1
aI1
aI5
aI11
aI17
aI43
aI23
aI21
aI421
aI613
aI199
aI1029
aI5255
aI271
aI12911
aI63431
aI108889
aI253379
asS'd'
I16013
sa(dp32026
S'a'
I43822
sS's'
I18
sS'm'
(lp32027
I1
aI3
aI7
aI11
aI27
aI45
aI9
aI89
aI65
aI25
aI1183
aI3497
aI8123
aI2389
aI215
aI16819
aI63777
aI163423
asS'd'
I16014
sa(dp32028
S'a'
I43841
sS's'
I18
sS'm'
(lp32029
I1
aI1
aI5
aI5
aI29
aI43
aI65
aI201
aI391
aI861
aI1133
aI3985
aI983
aI13039
aI15545
aI13695
aI91467
aI963
asS'd'
I16015
sa(dp32030
S'a'
I43861
sS's'
I18
sS'm'
(lp32031
I1
aI3
aI7
aI3
aI9
aI55
aI73
aI205
aI287
aI765
aI941
aI353
aI7379
aI2511
aI555
aI64399
aI77605
aI121959
asS'd'
I16016
sa(dp32032
S'a'
I43868
sS's'
I18
sS'm'
(lp32033
I1
aI3
aI1
aI11
aI25
aI15
aI109
aI35
aI351
aI675
aI1219
aI3791
aI233
aI15133
aI30733
aI24477
aI86077
aI85857
asS'd'
I16017
sa(dp32034
S'a'
I43877
sS's'
I18
sS'm'
(lp32035
I1
aI1
aI1
aI7
aI13
aI33
aI63
aI45
aI503
aI525
aI781
aI517
aI2187
aI1587
aI17297
aI26147
aI35421
aI61269
asS'd'
I16018
sa(dp32036
S'a'
I43878
sS's'
I18
sS'm'
(lp32037
I1
aI3
aI1
aI9
aI9
aI35
aI69
aI79
aI447
aI675
aI803
aI2793
aI5793
aI4433
aI29227
aI5437
aI103347
aI37713
asS'd'
I16019
sa(dp32038
S'a'
I43908
sS's'
I18
sS'm'
(lp32039
I1
aI3
aI1
aI9
aI27
aI31
aI65
aI167
aI327
aI231
aI1959
aI657
aI3141
aI8659
aI11055
aI23923
aI73597
aI187139
asS'd'
I16020
sa(dp32040
S'a'
I43917
sS's'
I18
sS'm'
(lp32041
I1
aI1
aI5
aI9
aI29
aI5
aI37
aI3
aI281
aI693
aI133
aI2139
aI5867
aI12073
aI23669
aI11427
aI80627
aI249003
asS'd'
I16021
sa(dp32042
S'a'
I43936
sS's'
I18
sS'm'
(lp32043
I1
aI1
aI1
aI13
aI9
aI35
aI89
aI203
aI381
aI281
aI535
aI1061
aI7417
aI13373
aI12149
aI37943
aI113133
aI110797
asS'd'
I16022
sa(dp32044
S'a'
I43954
sS's'
I18
sS'm'
(lp32045
I1
aI1
aI7
aI5
aI7
aI17
aI71
aI65
aI385
aI851
aI1357
aI3435
aI4441
aI6999
aI27065
aI32753
aI129531
aI52447
asS'd'
I16023
sa(dp32046
S'a'
I43963
sS's'
I18
sS'm'
(lp32047
I1
aI3
aI7
aI1
aI3
aI21
aI25
aI113
aI361
aI219
aI1345
aI2193
aI5711
aI3331
aI14343
aI23075
aI39955
aI71223
asS'd'
I16024
sa(dp32048
S'a'
I43977
sS's'
I18
sS'm'
(lp32049
I1
aI1
aI7
aI9
aI15
aI9
aI101
aI183
aI59
aI861
aI931
aI3385
aI6517
aI12067
aI7753
aI37997
aI128361
aI4591
asS'd'
I16025
sa(dp32050
S'a'
I44008
sS's'
I18
sS'm'
(lp32051
I1
aI3
aI5
aI3
aI23
aI5
aI37
aI217
aI203
aI749
aI993
aI2537
aI2425
aI13949
aI17235
aI15461
aI21275
aI141815
asS'd'
I16026
sa(dp32052
S'a'
I44013
sS's'
I18
sS'm'
(lp32053
I1
aI3
aI5
aI1
aI19
aI59
aI63
aI241
aI211
aI285
aI2033
aI1111
aI2859
aI14801
aI9491
aI14557
aI12973
aI223089
asS'd'
I16027
sa(dp32054
S'a'
I44026
sS's'
I18
sS'm'
(lp32055
I1
aI1
aI1
aI11
aI15
aI41
aI57
aI59
aI233
aI897
aI1193
aI381
aI2237
aI5309
aI19237
aI57607
aI97941
aI116573
asS'd'
I16028
sa(dp32056
S'a'
I44028
sS's'
I18
sS'm'
(lp32057
I1
aI1
aI3
aI9
aI15
aI17
aI75
aI15
aI385
aI129
aI495
aI887
aI4933
aI15113
aI7449
aI56213
aI15841
aI248029
asS'd'
I16029
sa(dp32058
S'a'
I44045
sS's'
I18
sS'm'
(lp32059
I1
aI3
aI1
aI1
aI29
aI37
aI19
aI221
aI47
aI185
aI1105
aI3297
aI5235
aI6917
aI12041
aI10815
aI54747
aI132329
asS'd'
I16030
sa(dp32060
S'a'
I44046
sS's'
I18
sS'm'
(lp32061
I1
aI3
aI1
aI5
aI3
aI17
aI33
aI35
aI395
aI491
aI1157
aI2563
aI6257
aI1025
aI18295
aI5293
aI77851
aI140157
asS'd'
I16031
sa(dp32062
S'a'
I44053
sS's'
I18
sS'm'
(lp32063
I1
aI3
aI1
aI13
aI7
aI31
aI95
aI21
aI347
aI409
aI1485
aI925
aI327
aI11497
aI7305
aI6503
aI111175
aI70989
asS'd'
I16032
sa(dp32064
S'a'
I44091
sS's'
I18
sS'm'
(lp32065
I1
aI3
aI3
aI1
aI11
aI41
aI127
aI93
aI367
aI649
aI1585
aI3379
aI7269
aI5537
aI26077
aI28541
aI55379
aI22989
asS'd'
I16033
sa(dp32066
S'a'
I44101
sS's'
I18
sS'm'
(lp32067
I1
aI1
aI1
aI5
aI27
aI11
aI35
aI49
aI7
aI113
aI1477
aI3615
aI7567
aI505
aI13915
aI51023
aI50783
aI45031
asS'd'
I16034
sa(dp32068
S'a'
I44129
sS's'
I18
sS'm'
(lp32069
I1
aI1
aI7
aI5
aI25
aI1
aI35
aI59
aI269
aI427
aI791
aI3179
aI1423
aI9801
aI17717
aI23631
aI97947
aI126861
asS'd'
I16035
sa(dp32070
S'a'
I44139
sS's'
I18
sS'm'
(lp32071
I1
aI3
aI5
aI7
aI1
aI51
aI77
aI97
aI19
aI287
aI303
aI791
aI307
aI4939
aI13615
aI61225
aI98127
aI114693
asS'd'
I16036
sa(dp32072
S'a'
I44172
sS's'
I18
sS'm'
(lp32073
I1
aI3
aI3
aI15
aI5
aI5
aI93
aI119
aI429
aI977
aI1763
aI3727
aI1761
aI3597
aI16489
aI71
aI44103
aI257929
asS'd'
I16037
sa(dp32074
S'a'
I44203
sS's'
I18
sS'm'
(lp32075
I1
aI1
aI5
aI1
aI15
aI29
aI79
aI33
aI335
aI381
aI1233
aI47
aI6741
aI4953
aI29689
aI11223
aI129185
aI182487
asS'd'
I16038
sa(dp32076
S'a'
I44211
sS's'
I18
sS'm'
(lp32077
I1
aI1
aI1
aI11
aI29
aI27
aI27
aI189
aI507
aI523
aI1949
aI2567
aI4105
aI1227
aI16631
aI34187
aI28521
aI265
asS'd'
I16039
sa(dp32078
S'a'
I44240
sS's'
I18
sS'm'
(lp32079
I1
aI3
aI3
aI7
aI3
aI7
aI39
aI155
aI315
aI439
aI1953
aI1227
aI6135
aI16291
aI453
aI50929
aI67507
aI166981
asS'd'
I16040
sa(dp32080
S'a'
I44259
sS's'
I18
sS'm'
(lp32081
I1
aI3
aI1
aI5
aI7
aI55
aI121
aI119
aI87
aI869
aI1049
aI575
aI3675
aI13505
aI15661
aI43899
aI106877
aI141691
asS'd'
I16041
sa(dp32082
S'a'
I44305
sS's'
I18
sS'm'
(lp32083
I1
aI3
aI1
aI11
aI23
aI9
aI117
aI243
aI329
aI767
aI335
aI2951
aI2887
aI13441
aI27579
aI15437
aI31699
aI165177
asS'd'
I16042
sa(dp32084
S'a'
I44383
sS's'
I18
sS'm'
(lp32085
I1
aI3
aI7
aI11
aI5
aI49
aI47
aI125
aI431
aI511
aI299
aI3215
aI3287
aI7029
aI9737
aI28317
aI34355
aI232365
asS'd'
I16043
sa(dp32086
S'a'
I44384
sS's'
I18
sS'm'
(lp32087
I1
aI3
aI7
aI3
aI17
aI9
aI29
aI255
aI509
aI393
aI1583
aI1979
aI6735
aI941
aI4393
aI35741
aI82019
aI109633
asS'd'
I16044
sa(dp32088
S'a'
I44389
sS's'
I18
sS'm'
(lp32089
I1
aI1
aI5
aI1
aI13
aI31
aI95
aI133
aI117
aI257
aI993
aI1513
aI4669
aI12239
aI7841
aI751
aI79567
aI23289
asS'd'
I16045
sa(dp32090
S'a'
I44390
sS's'
I18
sS'm'
(lp32091
I1
aI1
aI7
aI9
aI25
aI33
aI127
aI181
aI61
aI333
aI1087
aI1661
aI2715
aI2569
aI30041
aI4937
aI36053
aI3435
asS'd'
I16046
sa(dp32092
S'a'
I44432
sS's'
I18
sS'm'
(lp32093
I1
aI3
aI1
aI13
aI15
aI27
aI123
aI73
aI377
aI85
aI435
aI3435
aI2079
aI9271
aI28685
aI30089
aI38799
aI210247
asS'd'
I16047
sa(dp32094
S'a'
I44435
sS's'
I18
sS'm'
(lp32095
I1
aI1
aI5
aI3
aI5
aI17
aI93
aI181
aI313
aI837
aI1115
aI3099
aI3603
aI3483
aI23185
aI9933
aI53127
aI123245
asS'd'
I16048
sa(dp32096
S'a'
I44468
sS's'
I18
sS'm'
(lp32097
I1
aI1
aI1
aI13
aI3
aI57
aI11
aI67
aI41
aI273
aI1005
aI2313
aI505
aI6593
aI27287
aI47359
aI104597
aI45475
asS'd'
I16049
sa(dp32098
S'a'
I44472
sS's'
I18
sS'm'
(lp32099
I1
aI1
aI1
aI5
aI13
aI41
aI97
aI251
aI317
aI483
aI163
aI1493
aI6629
aI11995
aI31293
aI4715
aI98569
aI178419
asS'd'
I16050
sa(dp32100
S'a'
I44483
sS's'
I18
sS'm'
(lp32101
I1
aI3
aI5
aI5
aI21
aI7
aI69
aI169
aI223
aI953
aI1573
aI1137
aI8075
aI7733
aI23031
aI14589
aI6453
aI228883
asS'd'
I16051
sa(dp32102
S'a'
I44497
sS's'
I18
sS'm'
(lp32103
I1
aI1
aI7
aI9
aI27
aI37
aI3
aI183
aI41
aI99
aI111
aI1713
aI1291
aI8263
aI6373
aI39549
aI3099
aI156793
asS'd'
I16052
sa(dp32104
S'a'
I44550
sS's'
I18
sS'm'
(lp32105
I1
aI1
aI3
aI1
aI23
aI43
aI15
aI27
aI223
aI173
aI1601
aI2159
aI3595
aI15143
aI31065
aI35799
aI77339
aI141397
asS'd'
I16053
sa(dp32106
S'a'
I44577
sS's'
I18
sS'm'
(lp32107
I1
aI1
aI5
aI13
aI9
aI35
aI27
aI93
aI23
aI999
aI593
aI563
aI5333
aI8825
aI27277
aI46381
aI3171
aI6311
asS'd'
I16054
sa(dp32108
S'a'
I44584
sS's'
I18
sS'm'
(lp32109
I1
aI3
aI7
aI15
aI25
aI17
aI7
aI237
aI379
aI649
aI1879
aI2643
aI5951
aI4227
aI4937
aI24549
aI43577
aI116327
asS'd'
I16055
sa(dp32110
S'a'
I44601
sS's'
I18
sS'm'
(lp32111
I1
aI1
aI3
aI9
aI29
aI35
aI101
aI105
aI19
aI151
aI1135
aI897
aI2427
aI15779
aI31851
aI29183
aI44471
aI187817
asS'd'
I16056
sa(dp32112
S'a'
I44674
sS's'
I18
sS'm'
(lp32113
I1
aI3
aI7
aI11
aI9
aI55
aI19
aI115
aI395
aI559
aI1883
aI409
aI2459
aI201
aI18975
aI339
aI108251
aI19429
asS'd'
I16057
sa(dp32114
S'a'
I44683
sS's'
I18
sS'm'
(lp32115
I1
aI3
aI1
aI13
aI1
aI11
aI67
aI45
aI407
aI169
aI1401
aI381
aI3913
aI6491
aI28133
aI63857
aI52095
aI115749
asS'd'
I16058
sa(dp32116
S'a'
I44688
sS's'
I18
sS'm'
(lp32117
I1
aI1
aI1
aI11
aI21
aI45
aI65
aI253
aI511
aI51
aI893
aI3533
aI2101
aI4779
aI23941
aI22449
aI82457
aI44447
asS'd'
I16059
sa(dp32118
S'a'
I44694
sS's'
I18
sS'm'
(lp32119
I1
aI1
aI7
aI11
aI19
aI33
aI101
aI241
aI493
aI111
aI1967
aI469
aI6575
aI10445
aI733
aI56467
aI27403
aI25863
asS'd'
I16060
sa(dp32120
S'a'
I44703
sS's'
I18
sS'm'
(lp32121
I1
aI3
aI1
aI1
aI5
aI49
aI21
aI79
aI53
aI621
aI43
aI1183
aI1385
aI9129
aI21727
aI35559
aI35269
aI211383
asS'd'
I16061
sa(dp32122
S'a'
I44713
sS's'
I18
sS'm'
(lp32123
I1
aI1
aI1
aI13
aI23
aI43
aI53
aI145
aI149
aI611
aI745
aI899
aI7095
aI3243
aI1993
aI35807
aI110783
aI246199
asS'd'
I16062
sa(dp32124
S'a'
I44721
sS's'
I18
sS'm'
(lp32125
I1
aI1
aI7
aI9
aI25
aI37
aI71
aI233
aI407
aI1
aI1749
aI759
aI7689
aI15573
aI7351
aI33505
aI22631
aI49125
asS'd'
I16063
sa(dp32126
S'a'
I44722
sS's'
I18
sS'm'
(lp32127
I1
aI3
aI1
aI13
aI31
aI49
aI105
aI13
aI257
aI843
aI1171
aI2819
aI621
aI12567
aI24339
aI6689
aI127413
aI249671
asS'd'
I16064
sa(dp32128
S'a'
I44733
sS's'
I18
sS'm'
(lp32129
I1
aI1
aI5
aI11
aI5
aI1
aI93
aI21
aI317
aI789
aI571
aI2493
aI7255
aI6459
aI14737
aI13989
aI47003
aI246561
asS'd'
I16065
sa(dp32130
S'a'
I44746
sS's'
I18
sS'm'
(lp32131
I1
aI3
aI5
aI7
aI9
aI11
aI69
aI143
aI175
aI581
aI825
aI2219
aI1165
aI9061
aI6073
aI57169
aI18135
aI94943
asS'd'
I16066
sa(dp32132
S'a'
I44748
sS's'
I18
sS'm'
(lp32133
I1
aI1
aI5
aI7
aI7
aI55
aI121
aI107
aI395
aI939
aI1291
aI2497
aI3757
aI1361
aI31823
aI19375
aI22551
aI224653
asS'd'
I16067
sa(dp32134
S'a'
I44760
sS's'
I18
sS'm'
(lp32135
I1
aI1
aI7
aI15
aI15
aI1
aI47
aI225
aI223
aI589
aI1539
aI173
aI6257
aI5461
aI11197
aI9801
aI80687
aI84279
asS'd'
I16068
sa(dp32136
S'a'
I44765
sS's'
I18
sS'm'
(lp32137
I1
aI3
aI7
aI11
aI23
aI47
aI83
aI119
aI367
aI895
aI431
aI1949
aI565
aI1397
aI24911
aI29661
aI67861
aI74621
asS'd'
I16069
sa(dp32138
S'a'
I44766
sS's'
I18
sS'm'
(lp32139
I1
aI3
aI3
aI9
aI17
aI29
aI15
aI65
aI157
aI255
aI303
aI1467
aI3279
aI8873
aI31279
aI431
aI8497
aI7209
asS'd'
I16070
sa(dp32140
S'a'
I44775
sS's'
I18
sS'm'
(lp32141
I1
aI1
aI5
aI13
aI9
aI39
aI85
aI133
aI427
aI897
aI1665
aI2109
aI2865
aI15573
aI27729
aI59905
aI2241
aI83099
asS'd'
I16071
sa(dp32142
S'a'
I44794
sS's'
I18
sS'm'
(lp32143
I1
aI3
aI1
aI13
aI1
aI45
aI65
aI65
aI249
aI79
aI1515
aI503
aI953
aI9859
aI13307
aI27419
aI102209
aI107529
asS'd'
I16072
sa(dp32144
S'a'
I44822
sS's'
I18
sS'm'
(lp32145
I1
aI3
aI7
aI9
aI17
aI13
aI79
aI93
aI231
aI5
aI1811
aI557
aI1837
aI8077
aI8109
aI3497
aI79985
aI134375
asS'd'
I16073
sa(dp32146
S'a'
I44841
sS's'
I18
sS'm'
(lp32147
I1
aI1
aI5
aI9
aI15
aI3
aI23
aI27
aI491
aI9
aI1657
aI3877
aI3783
aI12645
aI2599
aI10673
aI40035
aI197681
asS'd'
I16074
sa(dp32148
S'a'
I44859
sS's'
I18
sS'm'
(lp32149
I1
aI1
aI5
aI3
aI5
aI13
aI121
aI145
aI291
aI621
aI1731
aI145
aI2033
aI341
aI3667
aI4139
aI52035
aI115865
asS'd'
I16075
sa(dp32150
S'a'
I44894
sS's'
I18
sS'm'
(lp32151
I1
aI3
aI5
aI5
aI3
aI29
aI115
aI227
aI339
aI23
aI1659
aI2367
aI1079
aI10757
aI30709
aI41473
aI55847
aI228761
asS'd'
I16076
sa(dp32152
S'a'
I44900
sS's'
I18
sS'm'
(lp32153
I1
aI3
aI5
aI7
aI7
aI61
aI103
aI57
aI197
aI31
aI237
aI2507
aI5247
aI10529
aI13823
aI47845
aI129031
aI47029
asS'd'
I16077
sa(dp32154
S'a'
I44904
sS's'
I18
sS'm'
(lp32155
I1
aI3
aI7
aI5
aI3
aI17
aI91
aI31
aI277
aI977
aI1905
aI3991
aI3657
aI12197
aI14535
aI43263
aI109629
aI31665
asS'd'
I16078
sa(dp32156
S'a'
I44946
sS's'
I18
sS'm'
(lp32157
I1
aI1
aI5
aI15
aI29
aI43
aI121
aI113
aI159
aI929
aI669
aI2067
aI4999
aI6847
aI12369
aI61795
aI98525
aI78051
asS'd'
I16079
sa(dp32158
S'a'
I44967
sS's'
I18
sS'm'
(lp32159
I1
aI1
aI3
aI7
aI25
aI61
aI65
aI79
aI373
aI683
aI113
aI1495
aI5447
aI15507
aI16731
aI53959
aI62905
aI252173
asS'd'
I16080
sa(dp32160
S'a'
I44994
sS's'
I18
sS'm'
(lp32161
I1
aI3
aI1
aI11
aI1
aI15
aI101
aI55
aI477
aI357
aI333
aI3243
aI3325
aI5885
aI11385
aI685
aI90575
aI23015
asS'd'
I16081
sa(dp32162
S'a'
I44999
sS's'
I18
sS'm'
(lp32163
I1
aI3
aI1
aI7
aI15
aI13
aI91
aI141
aI209
aI865
aI2035
aI2791
aI367
aI9953
aI29547
aI36731
aI13649
aI192911
asS'd'
I16082
sa(dp32164
S'a'
I45013
sS's'
I18
sS'm'
(lp32165
I1
aI1
aI5
aI15
aI9
aI59
aI37
aI131
aI3
aI299
aI897
aI119
aI7515
aI5271
aI2207
aI18187
aI62613
aI210345
asS'd'
I16083
sa(dp32166
S'a'
I45017
sS's'
I18
sS'm'
(lp32167
I1
aI3
aI1
aI7
aI29
aI53
aI95
aI185
aI327
aI473
aI1525
aI3751
aI915
aI5883
aI4137
aI46343
aI104917
aI182895
asS'd'
I16084
sa(dp32168
S'a'
I45020
sS's'
I18
sS'm'
(lp32169
I1
aI1
aI7
aI3
aI31
aI57
aI53
aI169
aI215
aI607
aI541
aI1081
aI5265
aI2509
aI28379
aI41767
aI19435
aI143693
asS'd'
I16085
sa(dp32170
S'a'
I45027
sS's'
I18
sS'm'
(lp32171
I1
aI1
aI7
aI7
aI31
aI57
aI19
aI203
aI467
aI57
aI1305
aI1513
aI6069
aI15595
aI31717
aI44687
aI65335
aI159315
asS'd'
I16086
sa(dp32172
S'a'
I45029
sS's'
I18
sS'm'
(lp32173
I1
aI1
aI1
aI1
aI13
aI1
aI41
aI39
aI489
aI465
aI1645
aI2847
aI6193
aI11025
aI11297
aI55945
aI92085
aI243061
asS'd'
I16087
sa(dp32174
S'a'
I45066
sS's'
I18
sS'm'
(lp32175
I1
aI3
aI3
aI13
aI15
aI47
aI87
aI77
aI321
aI903
aI181
aI2093
aI1673
aI5375
aI17969
aI48467
aI83441
aI120867
asS'd'
I16088
sa(dp32176
S'a'
I45074
sS's'
I18
sS'm'
(lp32177
I1
aI3
aI3
aI13
aI13
aI17
aI65
aI181
aI283
aI471
aI745
aI1091
aI7255
aI5987
aI29423
aI27579
aI96201
aI218157
asS'd'
I16089
sa(dp32178
S'a'
I45080
sS's'
I18
sS'm'
(lp32179
I1
aI1
aI1
aI7
aI13
aI5
aI101
aI131
aI309
aI401
aI99
aI2353
aI55
aI2377
aI6059
aI64777
aI33401
aI225605
asS'd'
I16090
sa(dp32180
S'a'
I45090
sS's'
I18
sS'm'
(lp32181
I1
aI3
aI1
aI15
aI9
aI35
aI37
aI197
aI65
aI593
aI411
aI2233
aI1485
aI9599
aI9581
aI31935
aI115145
aI76867
asS'd'
I16091
sa(dp32182
S'a'
I45101
sS's'
I18
sS'm'
(lp32183
I1
aI3
aI7
aI9
aI3
aI41
aI103
aI237
aI453
aI335
aI1831
aI3947
aI7573
aI3859
aI12495
aI60617
aI20715
aI163119
asS'd'
I16092
sa(dp32184
S'a'
I45110
sS's'
I18
sS'm'
(lp32185
I1
aI3
aI5
aI7
aI9
aI9
aI25
aI79
aI421
aI267
aI585
aI1093
aI2237
aI8881
aI7311
aI39417
aI110901
aI8969
asS'd'
I16093
sa(dp32186
S'a'
I45113
sS's'
I18
sS'm'
(lp32187
I1
aI3
aI3
aI15
aI1
aI29
aI75
aI159
aI89
aI965
aI457
aI1645
aI1485
aI729
aI30547
aI2275
aI79633
aI126089
asS'd'
I16094
sa(dp32188
S'a'
I45134
sS's'
I18
sS'm'
(lp32189
I1
aI3
aI7
aI7
aI19
aI47
aI55
aI115
aI287
aI477
aI719
aI3311
aI2455
aI12033
aI13811
aI34011
aI45153
aI126991
asS'd'
I16095
sa(dp32190
S'a'
I45136
sS's'
I18
sS'm'
(lp32191
I1
aI3
aI7
aI5
aI9
aI33
aI27
aI215
aI7
aI113
aI1027
aI415
aI1057
aI13011
aI1547
aI7955
aI21347
aI79427
asS'd'
I16096
sa(dp32192
S'a'
I45148
sS's'
I18
sS'm'
(lp32193
I1
aI3
aI1
aI11
aI1
aI41
aI27
aI29
aI255
aI763
aI219
aI897
aI915
aI453
aI9417
aI22845
aI31655
aI228869
asS'd'
I16097
sa(dp32194
S'a'
I45158
sS's'
I18
sS'm'
(lp32195
I1
aI1
aI7
aI13
aI7
aI49
aI105
aI111
aI53
aI219
aI171
aI841
aI5027
aI5311
aI28171
aI58719
aI32241
aI170921
asS'd'
I16098
sa(dp32196
S'a'
I45176
sS's'
I18
sS'm'
(lp32197
I1
aI3
aI5
aI7
aI21
aI53
aI11
aI169
aI45
aI429
aI1727
aI1953
aI8119
aI14955
aI19997
aI62665
aI31345
aI135715
asS'd'
I16099
sa(dp32198
S'a'
I45181
sS's'
I18
sS'm'
(lp32199
I1
aI1
aI5
aI13
aI17
aI13
aI125
aI247
aI411
aI663
aI1725
aI2515
aI7995
aI8963
aI5797
aI32871
aI66603
aI137997
asS'd'
I16100
sa(dp32200
S'a'
I45185
sS's'
I18
sS'm'
(lp32201
I1
aI1
aI5
aI5
aI23
aI37
aI15
aI63
aI35
aI817
aI463
aI1413
aI1203
aI8031
aI4169
aI45755
aI93511
aI134751
asS'd'
I16101
sa(dp32202
S'a'
I45188
sS's'
I18
sS'm'
(lp32203
I1
aI1
aI1
aI1
aI31
aI17
aI15
aI125
aI199
aI57
aI1499
aI1567
aI6113
aI6503
aI515
aI57841
aI49885
aI213881
asS'd'
I16102
sa(dp32204
S'a'
I45203
sS's'
I18
sS'm'
(lp32205
I1
aI1
aI7
aI7
aI15
aI1
aI67
aI213
aI197
aI471
aI577
aI27
aI1533
aI13009
aI31861
aI62435
aI113139
aI77057
asS'd'
I16103
sa(dp32206
S'a'
I45209
sS's'
I18
sS'm'
(lp32207
I1
aI3
aI7
aI1
aI29
aI21
aI99
aI173
aI137
aI19
aI1727
aI1157
aI5215
aI4367
aI28803
aI26031
aI120195
aI60111
asS'd'
I16104
sa(dp32208
S'a'
I45231
sS's'
I18
sS'm'
(lp32209
I1
aI1
aI5
aI9
aI25
aI55
aI27
aI45
aI231
aI693
aI765
aI429
aI2897
aI6045
aI19705
aI61903
aI5385
aI172967
asS'd'
I16105
sa(dp32210
S'a'
I45234
sS's'
I18
sS'm'
(lp32211
I1
aI3
aI1
aI5
aI9
aI17
aI35
aI151
aI91
aI41
aI89
aI3751
aI27
aI5721
aI26117
aI6105
aI31609
aI79569
asS'd'
I16106
sa(dp32212
S'a'
I45239
sS's'
I18
sS'm'
(lp32213
I1
aI3
aI5
aI11
aI15
aI3
aI25
aI73
aI383
aI765
aI729
aI51
aI4245
aI6215
aI9435
aI45223
aI68071
aI68453
asS'd'
I16107
sa(dp32214
S'a'
I45246
sS's'
I18
sS'm'
(lp32215
I1
aI1
aI7
aI5
aI31
aI61
aI19
aI153
aI331
aI233
aI11
aI3047
aI3939
aI11959
aI897
aI4437
aI2941
aI174493
asS'd'
I16108
sa(dp32216
S'a'
I45260
sS's'
I18
sS'm'
(lp32217
I1
aI1
aI7
aI11
aI27
aI11
aI83
aI33
aI7
aI761
aI805
aI2327
aI2997
aI12269
aI18707
aI10615
aI114357
aI54951
asS'd'
I16109
sa(dp32218
S'a'
I45284
sS's'
I18
sS'm'
(lp32219
I1
aI3
aI5
aI9
aI3
aI45
aI65
aI193
aI347
aI771
aI663
aI2901
aI6467
aI14109
aI5169
aI38021
aI39605
aI216877
asS'd'
I16110
sa(dp32220
S'a'
I45301
sS's'
I18
sS'm'
(lp32221
I1
aI1
aI7
aI1
aI1
aI13
aI109
aI179
aI499
aI325
aI1763
aI1923
aI7429
aI259
aI20589
aI48473
aI49605
aI124709
asS'd'
I16111
sa(dp32222
S'a'
I45333
sS's'
I18
sS'm'
(lp32223
I1
aI3
aI1
aI3
aI5
aI27
aI25
aI127
aI313
aI541
aI589
aI751
aI5959
aI5801
aI32467
aI40597
aI75625
aI24827
asS'd'
I16112
sa(dp32224
S'a'
I45350
sS's'
I18
sS'm'
(lp32225
I1
aI1
aI1
aI3
aI13
aI35
aI71
aI223
aI281
aI767
aI447
aI1253
aI2227
aI7305
aI23125
aI62847
aI16783
aI76145
asS'd'
I16113
sa(dp32226
S'a'
I45354
sS's'
I18
sS'm'
(lp32227
I1
aI1
aI5
aI5
aI11
aI1
aI95
aI215
aI351
aI915
aI1471
aI143
aI4011
aI3373
aI19019
aI31797
aI85891
aI33871
asS'd'
I16114
sa(dp32228
S'a'
I45446
sS's'
I18
sS'm'
(lp32229
I1
aI1
aI3
aI9
aI15
aI39
aI125
aI29
aI85
aI625
aI1155
aI753
aI6679
aI14239
aI14597
aI32715
aI97313
aI162291
asS'd'
I16115
sa(dp32230
S'a'
I45458
sS's'
I18
sS'm'
(lp32231
I1
aI1
aI1
aI7
aI13
aI47
aI127
aI69
aI199
aI145
aI123
aI3207
aI7673
aI9991
aI27501
aI29189
aI93027
aI136881
asS'd'
I16116
sa(dp32232
S'a'
I45518
sS's'
I18
sS'm'
(lp32233
I1
aI3
aI1
aI15
aI9
aI21
aI121
aI23
aI457
aI315
aI437
aI1919
aI5531
aI13817
aI8883
aI4421
aI19487
aI88591
asS'd'
I16117
sa(dp32234
S'a'
I45566
sS's'
I18
sS'm'
(lp32235
I1
aI3
aI1
aI9
aI15
aI49
aI101
aI27
aI11
aI283
aI1277
aI971
aI7697
aI5915
aI12709
aI38251
aI88165
aI261609
asS'd'
I16118
sa(dp32236
S'a'
I45569
sS's'
I18
sS'm'
(lp32237
I1
aI1
aI3
aI9
aI17
aI19
aI53
aI131
aI327
aI917
aI603
aI2451
aI3597
aI14731
aI9223
aI29719
aI113507
aI69875
asS'd'
I16119
sa(dp32238
S'a'
I45575
sS's'
I18
sS'm'
(lp32239
I1
aI1
aI5
aI7
aI1
aI3
aI77
aI71
aI149
aI901
aI283
aI1599
aI1053
aI16305
aI9937
aI20907
aI4133
aI29623
asS'd'
I16120
sa(dp32240
S'a'
I45576
sS's'
I18
sS'm'
(lp32241
I1
aI1
aI3
aI9
aI25
aI63
aI101
aI255
aI397
aI233
aI1111
aI1583
aI3431
aI5245
aI30209
aI33201
aI63859
aI16551
asS'd'
I16121
sa(dp32242
S'a'
I45587
sS's'
I18
sS'm'
(lp32243
I1
aI1
aI7
aI5
aI19
aI25
aI9
aI137
aI105
aI363
aI867
aI811
aI5829
aI12595
aI18867
aI61021
aI19425
aI99399
asS'd'
I16122
sa(dp32244
S'a'
I45615
sS's'
I18
sS'm'
(lp32245
I1
aI1
aI3
aI7
aI11
aI1
aI119
aI95
aI367
aI239
aI1677
aI67
aI283
aI5701
aI3635
aI26917
aI112895
aI224049
asS'd'
I16123
sa(dp32246
S'a'
I45661
sS's'
I18
sS'm'
(lp32247
I1
aI3
aI5
aI11
aI13
aI33
aI11
aI191
aI305
aI551
aI159
aI1953
aI4231
aI811
aI8711
aI41291
aI110917
aI176177
asS'd'
I16124
sa(dp32248
S'a'
I45798
sS's'
I18
sS'm'
(lp32249
I1
aI3
aI5
aI9
aI15
aI3
aI27
aI29
aI77
aI881
aI849
aI1113
aI7151
aI3433
aI11199
aI38409
aI98173
aI21373
asS'd'
I16125
sa(dp32250
S'a'
I45829
sS's'
I18
sS'm'
(lp32251
I1
aI1
aI5
aI13
aI25
aI57
aI81
aI169
aI215
aI379
aI1707
aI493
aI1071
aI4869
aI19149
aI24585
aI61803
aI149305
asS'd'
I16126
sa(dp32252
S'a'
I45847
sS's'
I18
sS'm'
(lp32253
I1
aI1
aI5
aI11
aI13
aI11
aI9
aI45
aI207
aI347
aI1203
aI185
aI3919
aI3119
aI27879
aI50793
aI18499
aI109239
asS'd'
I16127
sa(dp32254
S'a'
I45858
sS's'
I18
sS'm'
(lp32255
I1
aI1
aI1
aI5
aI3
aI23
aI123
aI175
aI445
aI439
aI215
aI1883
aI6857
aI14837
aI29411
aI33633
aI96241
aI68361
asS'd'
I16128
sa(dp32256
S'a'
I45864
sS's'
I18
sS'm'
(lp32257
I1
aI1
aI5
aI7
aI31
aI53
aI19
aI41
aI477
aI203
aI1133
aI1471
aI5067
aI3875
aI30655
aI9207
aI24835
aI22019
asS'd'
I16129
sa(dp32258
S'a'
I45899
sS's'
I18
sS'm'
(lp32259
I1
aI3
aI1
aI11
aI19
aI9
aI91
aI217
aI135
aI169
aI489
aI727
aI5471
aI2125
aI20867
aI19689
aI78859
aI222433
asS'd'
I16130
sa(dp32260
S'a'
I45904
sS's'
I18
sS'm'
(lp32261
I1
aI1
aI1
aI13
aI5
aI37
aI37
aI39
aI397
aI527
aI789
aI1447
aI7123
aI6099
aI16849
aI48895
aI95543
aI56135
asS'd'
I16131
sa(dp32262
S'a'
I45919
sS's'
I18
sS'm'
(lp32263
I1
aI3
aI7
aI7
aI15
aI31
aI81
aI165
aI31
aI867
aI1879
aI1161
aI3651
aI5167
aI12855
aI60195
aI85611
aI191791
asS'd'
I16132
sa(dp32264
S'a'
I45974
sS's'
I18
sS'm'
(lp32265
I1
aI1
aI1
aI9
aI31
aI35
aI19
aI123
aI281
aI445
aI63
aI2683
aI7073
aI1489
aI9791
aI40125
aI43723
aI103765
asS'd'
I16133
sa(dp32266
S'a'
I45987
sS's'
I18
sS'm'
(lp32267
I1
aI3
aI3
aI9
aI23
aI31
aI113
aI241
aI149
aI463
aI1047
aI3737
aI8105
aI16295
aI4565
aI8095
aI16617
aI87455
asS'd'
I16134
sa(dp32268
S'a'
I45989
sS's'
I18
sS'm'
(lp32269
I1
aI3
aI5
aI15
aI1
aI47
aI29
aI77
aI27
aI1013
aI1091
aI3657
aI835
aI563
aI7139
aI58839
aI92697
aI114523
asS'd'
I16135
sa(dp32270
S'a'
I46008
sS's'
I18
sS'm'
(lp32271
I1
aI3
aI1
aI1
aI5
aI49
aI87
aI251
aI473
aI583
aI2033
aI809
aI5341
aI15835
aI691
aI57133
aI75751
aI127717
asS'd'
I16136
sa(dp32272
S'a'
I46036
sS's'
I18
sS'm'
(lp32273
I1
aI3
aI1
aI11
aI23
aI51
aI95
aI167
aI73
aI739
aI573
aI1113
aI7585
aI5457
aI25767
aI5583
aI29583
aI48263
asS'd'
I16137
sa(dp32274
S'a'
I46039
sS's'
I18
sS'm'
(lp32275
I1
aI1
aI3
aI5
aI9
aI59
aI1
aI235
aI17
aI973
aI1987
aI3629
aI1739
aI6419
aI29943
aI44963
aI63943
aI18571
asS'd'
I16138
sa(dp32276
S'a'
I46043
sS's'
I18
sS'm'
(lp32277
I1
aI3
aI7
aI11
aI9
aI37
aI25
aI253
aI307
aI411
aI891
aI1977
aI6979
aI11287
aI12619
aI50327
aI25831
aI135673
asS'd'
I16139
sa(dp32278
S'a'
I46045
sS's'
I18
sS'm'
(lp32279
I1
aI1
aI5
aI15
aI23
aI27
aI83
aI157
aI253
aI559
aI1989
aI2431
aI4821
aI6617
aI4295
aI16143
aI249
aI109855
asS'd'
I16140
sa(dp32280
S'a'
I46079
sS's'
I18
sS'm'
(lp32281
I1
aI1
aI3
aI9
aI31
aI59
aI117
aI139
aI379
aI703
aI49
aI3013
aI6383
aI4019
aI7289
aI4567
aI34931
aI224967
asS'd'
I16141
sa(dp32282
S'a'
I46084
sS's'
I18
sS'm'
(lp32283
I1
aI1
aI1
aI3
aI3
aI33
aI49
aI127
aI177
aI417
aI723
aI3259
aI1547
aI3297
aI19733
aI59465
aI122179
aI13209
asS'd'
I16142
sa(dp32284
S'a'
I46091
sS's'
I18
sS'm'
(lp32285
I1
aI3
aI3
aI1
aI3
aI61
aI23
aI177
aI77
aI579
aI1739
aI2707
aI5319
aI13291
aI10223
aI45395
aI62797
aI124675
asS'd'
I16143
sa(dp32286
S'a'
I46105
sS's'
I18
sS'm'
(lp32287
I1
aI1
aI3
aI15
aI25
aI47
aI79
aI111
aI245
aI853
aI1103
aI3741
aI5783
aI2075
aI26371
aI28801
aI117831
aI111735
asS'd'
I16144
sa(dp32288
S'a'
I46117
sS's'
I18
sS'm'
(lp32289
I1
aI1
aI5
aI11
aI3
aI47
aI43
aI61
aI75
aI963
aI1507
aI3491
aI6031
aI14171
aI32557
aI23779
aI44815
aI168409
asS'd'
I16145
sa(dp32290
S'a'
I46141
sS's'
I18
sS'm'
(lp32291
I1
aI3
aI3
aI7
aI21
aI39
aI61
aI109
aI13
aI833
aI373
aI4021
aI4035
aI7987
aI18957
aI10423
aI82823
aI8763
asS'd'
I16146
sa(dp32292
S'a'
I46147
sS's'
I18
sS'm'
(lp32293
I1
aI1
aI7
aI3
aI3
aI33
aI113
aI77
aI447
aI127
aI213
aI1605
aI3873
aI12345
aI7847
aI63903
aI80665
aI6647
asS'd'
I16147
sa(dp32294
S'a'
I46162
sS's'
I18
sS'm'
(lp32295
I1
aI3
aI1
aI5
aI25
aI11
aI49
aI233
aI377
aI791
aI43
aI195
aI393
aI1403
aI27567
aI29673
aI11327
aI190513
asS'd'
I16148
sa(dp32296
S'a'
I46184
sS's'
I18
sS'm'
(lp32297
I1
aI3
aI1
aI15
aI21
aI3
aI21
aI169
aI221
aI495
aI1045
aI3715
aI4923
aI2437
aI23203
aI59905
aI70285
aI258455
asS'd'
I16149
sa(dp32298
S'a'
I46197
sS's'
I18
sS'm'
(lp32299
I1
aI1
aI7
aI13
aI29
aI33
aI91
aI153
aI31
aI359
aI25
aI3947
aI1699
aI4081
aI20907
aI24953
aI64977
aI88115
asS'd'
I16150
sa(dp32300
S'a'
I46207
sS's'
I18
sS'm'
(lp32301
I1
aI1
aI5
aI11
aI17
aI7
aI17
aI129
aI91
aI835
aI139
aI3823
aI1827
aI6787
aI27367
aI26801
aI61513
aI189677
asS'd'
I16151
sa(dp32302
S'a'
I46214
sS's'
I18
sS'm'
(lp32303
I1
aI3
aI3
aI9
aI23
aI21
aI97
aI103
aI481
aI859
aI413
aI893
aI4021
aI8111
aI23703
aI18791
aI102735
aI82735
asS'd'
I16152
sa(dp32304
S'a'
I46218
sS's'
I18
sS'm'
(lp32305
I1
aI1
aI1
aI7
aI11
aI53
aI61
aI3
aI347
aI633
aI191
aI3605
aI41
aI12605
aI14381
aI60403
aI126223
aI186157
asS'd'
I16153
sa(dp32306
S'a'
I46237
sS's'
I18
sS'm'
(lp32307
I1
aI3
aI5
aI1
aI11
aI33
aI61
aI235
aI245
aI623
aI2019
aI3289
aI5761
aI15557
aI1685
aI2173
aI104825
aI245139
asS'd'
I16154
sa(dp32308
S'a'
I46247
sS's'
I18
sS'm'
(lp32309
I1
aI1
aI3
aI7
aI19
aI55
aI75
aI219
aI11
aI75
aI1765
aI1833
aI263
aI6605
aI26297
aI24969
aI17721
aI109495
asS'd'
I16155
sa(dp32310
S'a'
I46251
sS's'
I18
sS'm'
(lp32311
I1
aI1
aI7
aI5
aI7
aI33
aI23
aI109
aI207
aI137
aI385
aI3233
aI6765
aI3517
aI31389
aI57049
aI25645
aI176257
asS'd'
I16156
sa(dp32312
S'a'
I46327
sS's'
I18
sS'm'
(lp32313
I1
aI1
aI5
aI5
aI27
aI63
aI65
aI23
aI93
aI721
aI1665
aI3805
aI3611
aI7543
aI21119
aI38565
aI99921
aI72773
asS'd'
I16157
sa(dp32314
S'a'
I46333
sS's'
I18
sS'm'
(lp32315
I1
aI1
aI1
aI13
aI13
aI19
aI23
aI5
aI425
aI727
aI1469
aI1261
aI6597
aI725
aI27129
aI36953
aI25781
aI191581
asS'd'
I16158
sa(dp32316
S'a'
I46342
sS's'
I18
sS'm'
(lp32317
I1
aI3
aI3
aI1
aI3
aI29
aI25
aI109
aI237
aI135
aI409
aI2239
aI5033
aI11007
aI31861
aI55171
aI76313
aI216271
asS'd'
I16159
sa(dp32318
S'a'
I46366
sS's'
I18
sS'm'
(lp32319
I1
aI3
aI7
aI13
aI31
aI33
aI11
aI31
aI179
aI67
aI755
aI2513
aI37
aI12863
aI24053
aI12315
aI45009
aI166643
asS'd'
I16160
sa(dp32320
S'a'
I46376
sS's'
I18
sS'm'
(lp32321
I1
aI1
aI5
aI13
aI13
aI37
aI19
aI223
aI409
aI387
aI139
aI3283
aI243
aI15573
aI24173
aI63271
aI91561
aI168665
asS'd'
I16161
sa(dp32322
S'a'
I46384
sS's'
I18
sS'm'
(lp32323
I1
aI3
aI5
aI7
aI17
aI25
aI63
aI231
aI23
aI965
aI1873
aI1021
aI7927
aI11127
aI19553
aI9883
aI83009
aI258991
asS'd'
I16162
sa(dp32324
S'a'
I46404
sS's'
I18
sS'm'
(lp32325
I1
aI3
aI1
aI7
aI17
aI21
aI71
aI65
aI165
aI845
aI1739
aI2395
aI5959
aI14803
aI17333
aI59003
aI36477
aI202511
asS'd'
I16163
sa(dp32326
S'a'
I46478
sS's'
I18
sS'm'
(lp32327
I1
aI3
aI5
aI13
aI15
aI57
aI59
aI205
aI507
aI41
aI703
aI195
aI4373
aI4023
aI13399
aI62061
aI43645
aI204899
asS'd'
I16164
sa(dp32328
S'a'
I46511
sS's'
I18
sS'm'
(lp32329
I1
aI3
aI5
aI11
aI13
aI13
aI43
aI19
aI497
aI599
aI1345
aI2001
aI407
aI2731
aI16283
aI55161
aI130887
aI189201
asS'd'
I16165
sa(dp32330
S'a'
I46540
sS's'
I18
sS'm'
(lp32331
I1
aI1
aI7
aI5
aI13
aI3
aI105
aI83
aI235
aI363
aI869
aI1715
aI4031
aI3419
aI15149
aI10627
aI47787
aI226107
asS'd'
I16166
sa(dp32332
S'a'
I46638
sS's'
I18
sS'm'
(lp32333
I1
aI3
aI3
aI3
aI9
aI49
aI99
aI161
aI413
aI739
aI195
aI2815
aI1157
aI9069
aI15591
aI4509
aI30765
aI184013
asS'd'
I16167
sa(dp32334
S'a'
I46700
sS's'
I18
sS'm'
(lp32335
I1
aI3
aI7
aI5
aI25
aI63
aI23
aI91
aI507
aI647
aI1249
aI2035
aI4341
aI8811
aI8345
aI49463
aI69023
aI195775
asS'd'
I16168
sa(dp32336
S'a'
I46706
sS's'
I18
sS'm'
(lp32337
I1
aI3
aI5
aI5
aI5
aI21
aI117
aI1
aI321
aI495
aI251
aI1961
aI8043
aI6593
aI11017
aI6167
aI56607
aI144009
asS'd'
I16169
sa(dp32338
S'a'
I46745
sS's'
I18
sS'm'
(lp32339
I1
aI1
aI3
aI3
aI15
aI63
aI41
aI177
aI403
aI375
aI1771
aI3
aI7481
aI4887
aI799
aI59283
aI106319
aI32155
asS'd'
I16170
sa(dp32340
S'a'
I46757
sS's'
I18
sS'm'
(lp32341
I1
aI1
aI1
aI7
aI19
aI57
aI97
aI91
aI489
aI561
aI335
aI3809
aI3167
aI8879
aI1789
aI22329
aI58851
aI84461
asS'd'
I16171
sa(dp32342
S'a'
I46772
sS's'
I18
sS'm'
(lp32343
I1
aI3
aI3
aI11
aI3
aI23
aI79
aI25
aI187
aI497
aI1343
aI543
aI4495
aI14599
aI29365
aI14795
aI26341
aI26923
asS'd'
I16172
sa(dp32344
S'a'
I46776
sS's'
I18
sS'm'
(lp32345
I1
aI3
aI1
aI13
aI21
aI35
aI9
aI227
aI423
aI761
aI439
aI3099
aI5167
aI12955
aI12877
aI5591
aI123511
aI74227
asS'd'
I16173
sa(dp32346
S'a'
I46790
sS's'
I18
sS'm'
(lp32347
I1
aI1
aI3
aI7
aI1
aI49
aI115
aI95
aI481
aI659
aI183
aI2337
aI39
aI235
aI30869
aI10223
aI59673
aI65293
asS'd'
I16174
sa(dp32348
S'a'
I46794
sS's'
I18
sS'm'
(lp32349
I1
aI1
aI1
aI13
aI11
aI29
aI113
aI37
aI153
aI993
aI1195
aI1695
aI4741
aI13671
aI29097
aI65023
aI78281
aI156235
asS'd'
I16175
sa(dp32350
S'a'
I46804
sS's'
I18
sS'm'
(lp32351
I1
aI3
aI5
aI11
aI19
aI17
aI47
aI243
aI273
aI679
aI393
aI4059
aI705
aI12473
aI1867
aI13783
aI86821
aI240545
asS'd'
I16176
sa(dp32352
S'a'
I46820
sS's'
I18
sS'm'
(lp32353
I1
aI3
aI1
aI11
aI23
aI35
aI15
aI31
aI275
aI261
aI427
aI909
aI7925
aI4737
aI22825
aI34859
aI28593
aI20071
asS'd'
I16177
sa(dp32354
S'a'
I46849
sS's'
I18
sS'm'
(lp32355
I1
aI3
aI1
aI1
aI13
aI1
aI115
aI115
aI103
aI63
aI1023
aI815
aI7007
aI6063
aI13329
aI28051
aI29807
aI109861
asS'd'
I16178
sa(dp32356
S'a'
I46870
sS's'
I18
sS'm'
(lp32357
I1
aI1
aI1
aI11
aI19
aI57
aI49
aI169
aI33
aI59
aI579
aI2409
aI89
aI9655
aI18091
aI57771
aI34509
aI175991
asS'd'
I16179
sa(dp32358
S'a'
I46876
sS's'
I18
sS'm'
(lp32359
I1
aI1
aI5
aI9
aI27
aI19
aI41
aI217
aI313
aI359
aI1745
aI3887
aI589
aI3103
aI10087
aI30615
aI56793
aI102515
asS'd'
I16180
sa(dp32360
S'a'
I46903
sS's'
I18
sS'm'
(lp32361
I1
aI3
aI7
aI11
aI25
aI43
aI101
aI245
aI67
aI1003
aI1379
aI2141
aI8025
aI15231
aI20705
aI45513
aI82251
aI147295
asS'd'
I16181
sa(dp32362
S'a'
I46904
sS's'
I18
sS'm'
(lp32363
I1
aI3
aI5
aI5
aI5
aI47
aI19
aI33
aI107
aI773
aI627
aI4077
aI5829
aI6483
aI25791
aI35079
aI103073
aI201657
asS'd'
I16182
sa(dp32364
S'a'
I46907
sS's'
I18
sS'm'
(lp32365
I1
aI1
aI5
aI15
aI1
aI37
aI115
aI15
aI61
aI987
aI1029
aI2125
aI5357
aI10233
aI14907
aI12077
aI92143
aI207301
asS'd'
I16183
sa(dp32366
S'a'
I46932
sS's'
I18
sS'm'
(lp32367
I1
aI1
aI7
aI7
aI3
aI15
aI95
aI3
aI393
aI535
aI819
aI743
aI3613
aI11459
aI2269
aI17083
aI65547
aI74813
asS'd'
I16184
sa(dp32368
S'a'
I46955
sS's'
I18
sS'm'
(lp32369
I1
aI1
aI7
aI5
aI11
aI27
aI117
aI65
aI55
aI87
aI105
aI3219
aI1587
aI383
aI25349
aI54561
aI11935
aI101203
asS'd'
I16185
sa(dp32370
S'a'
I46963
sS's'
I18
sS'm'
(lp32371
I1
aI1
aI5
aI15
aI23
aI53
aI37
aI149
aI191
aI963
aI1407
aI4091
aI1647
aI9537
aI30247
aI23501
aI123745
aI76301
asS'd'
I16186
sa(dp32372
S'a'
I46991
sS's'
I18
sS'm'
(lp32373
I1
aI3
aI1
aI1
aI21
aI31
aI11
aI107
aI55
aI823
aI805
aI141
aI7177
aI15883
aI3307
aI52245
aI115171
aI260589
asS'd'
I16187
sa(dp32374
S'a'
I47029
sS's'
I18
sS'm'
(lp32375
I1
aI3
aI7
aI13
aI25
aI15
aI5
aI221
aI347
aI83
aI51
aI1227
aI4591
aI851
aI10173
aI37777
aI82441
aI175219
asS'd'
I16188
sa(dp32376
S'a'
I47036
sS's'
I18
sS'm'
(lp32377
I1
aI1
aI7
aI15
aI17
aI41
aI121
aI215
aI111
aI999
aI367
aI1961
aI3207
aI10145
aI10395
aI24381
aI95937
aI12693
asS'd'
I16189
sa(dp32378
S'a'
I47054
sS's'
I18
sS'm'
(lp32379
I1
aI1
aI1
aI11
aI23
aI27
aI117
aI255
aI87
aI519
aI599
aI3471
aI3983
aI2797
aI13477
aI56479
aI27321
aI101585
asS'd'
I16190
sa(dp32380
S'a'
I47099
sS's'
I18
sS'm'
(lp32381
I1
aI1
aI5
aI9
aI1
aI3
aI35
aI15
aI457
aI209
aI141
aI1295
aI1631
aI9745
aI30247
aI44865
aI78113
aI13207
asS'd'
I16191
sa(dp32382
S'a'
I47106
sS's'
I18
sS'm'
(lp32383
I1
aI3
aI3
aI1
aI27
aI63
aI29
aI31
aI277
aI173
aI1321
aI3847
aI4127
aI8713
aI10507
aI8697
aI97025
aI105995
asS'd'
I16192
sa(dp32384
S'a'
I47132
sS's'
I18
sS'm'
(lp32385
I1
aI1
aI5
aI13
aI13
aI47
aI33
aI69
aI113
aI369
aI539
aI4075
aI1013
aI9733
aI9291
aI33377
aI130567
aI238331
asS'd'
I16193
sa(dp32386
S'a'
I47145
sS's'
I18
sS'm'
(lp32387
I1
aI3
aI7
aI13
aI15
aI15
aI125
aI219
aI205
aI763
aI1233
aI1911
aI7733
aI7623
aI27305
aI6067
aI16169
aI238805
asS'd'
I16194
sa(dp32388
S'a'
I47151
sS's'
I18
sS'm'
(lp32389
I1
aI3
aI1
aI9
aI23
aI17
aI35
aI175
aI157
aI627
aI1045
aI1791
aI1675
aI11699
aI2151
aI28293
aI14529
aI30523
asS'd'
I16195
sa(dp32390
S'a'
I47156
sS's'
I18
sS'm'
(lp32391
I1
aI1
aI5
aI9
aI5
aI1
aI23
aI151
aI295
aI949
aI371
aI3317
aI2557
aI5815
aI9699
aI48379
aI104561
aI103747
asS'd'
I16196
sa(dp32392
S'a'
I47171
sS's'
I18
sS'm'
(lp32393
I1
aI3
aI5
aI5
aI11
aI57
aI125
aI247
aI29
aI257
aI979
aI2437
aI4391
aI8229
aI11231
aI30145
aI111165
aI92347
asS'd'
I16197
sa(dp32394
S'a'
I47188
sS's'
I18
sS'm'
(lp32395
I1
aI1
aI3
aI3
aI27
aI17
aI71
aI71
aI357
aI367
aI1213
aI2549
aI6049
aI15299
aI2891
aI21839
aI109889
aI34643
asS'd'
I16198
sa(dp32396
S'a'
I47201
sS's'
I18
sS'm'
(lp32397
I1
aI1
aI1
aI11
aI7
aI21
aI41
aI77
aI249
aI567
aI1947
aI2989
aI875
aI12975
aI23599
aI49313
aI67213
aI98415
asS'd'
I16199
sa(dp32398
S'a'
I47219
sS's'
I18
sS'm'
(lp32399
I1
aI1
aI7
aI7
aI27
aI51
aI103
aI221
aI295
aI247
aI1579
aI2435
aI67
aI3087
aI9421
aI59573
aI111143
aI42363
asS'd'
I16200
sa(dp32400
S'a'
I47226
sS's'
I18
sS'm'
(lp32401
I1
aI1
aI1
aI13
aI27
aI27
aI75
aI33
aI81
aI841
aI1295
aI2823
aI997
aI16329
aI6117
aI43361
aI63727
aI113347
asS'd'
I16201
sa(dp32402
S'a'
I47241
sS's'
I18
sS'm'
(lp32403
I1
aI1
aI1
aI3
aI19
aI7
aI43
aI93
aI397
aI293
aI803
aI3021
aI3915
aI1417
aI22255
aI38529
aI53737
aI133705
asS'd'
I16202
sa(dp32404
S'a'
I47252
sS's'
I18
sS'm'
(lp32405
I1
aI3
aI3
aI13
aI1
aI33
aI1
aI235
aI255
aI345
aI1621
aI315
aI2685
aI6451
aI7133
aI239
aI103075
aI175033
asS'd'
I16203
sa(dp32406
S'a'
I47261
sS's'
I18
sS'm'
(lp32407
I1
aI3
aI1
aI11
aI29
aI47
aI61
aI115
aI395
aI633
aI1913
aI2983
aI4581
aI3729
aI22511
aI16479
aI80607
aI232859
asS'd'
I16204
sa(dp32408
S'a'
I47295
sS's'
I18
sS'm'
(lp32409
I1
aI3
aI7
aI7
aI27
aI25
aI29
aI121
aI511
aI533
aI1791
aI3349
aI4915
aI8213
aI13913
aI6595
aI2353
aI207495
asS'd'
I16205
sa(dp32410
S'a'
I47297
sS's'
I18
sS'm'
(lp32411
I1
aI3
aI7
aI5
aI21
aI25
aI77
aI189
aI473
aI1015
aI1455
aI1923
aI3281
aI15163
aI2329
aI58529
aI55369
aI195007
asS'd'
I16206
sa(dp32412
S'a'
I47333
sS's'
I18
sS'm'
(lp32413
I1
aI3
aI1
aI1
aI25
aI37
aI117
aI41
aI207
aI413
aI143
aI1707
aI7463
aI15399
aI3013
aI4141
aI37669
aI70953
asS'd'
I16207
sa(dp32414
S'a'
I47337
sS's'
I18
sS'm'
(lp32415
I1
aI3
aI5
aI3
aI31
aI61
aI41
aI157
aI141
aI387
aI1705
aI1661
aI3607
aI6905
aI3305
aI63235
aI7977
aI253707
asS'd'
I16208
sa(dp32416
S'a'
I47360
sS's'
I18
sS'm'
(lp32417
I1
aI1
aI5
aI11
aI29
aI15
aI61
aI127
aI417
aI795
aI171
aI415
aI7935
aI4553
aI29979
aI21153
aI108811
aI219959
asS'd'
I16209
sa(dp32418
S'a'
I47377
sS's'
I18
sS'm'
(lp32419
I1
aI1
aI5
aI11
aI17
aI61
aI3
aI9
aI297
aI53
aI933
aI341
aI507
aI2683
aI15313
aI24113
aI78617
aI191127
asS'd'
I16210
sa(dp32420
S'a'
I47378
sS's'
I18
sS'm'
(lp32421
I1
aI1
aI7
aI15
aI5
aI57
aI65
aI187
aI15
aI541
aI1843
aI731
aI7331
aI2479
aI26259
aI32685
aI125259
aI108693
asS'd'
I16211
sa(dp32422
S'a'
I47384
sS's'
I18
sS'm'
(lp32423
I1
aI1
aI7
aI7
aI11
aI15
aI33
aI183
aI225
aI609
aI1755
aI3531
aI2767
aI6267
aI30823
aI36797
aI59699
aI136769
asS'd'
I16212
sa(dp32424
S'a'
I47389
sS's'
I18
sS'm'
(lp32425
I1
aI3
aI5
aI13
aI15
aI19
aI77
aI239
aI307
aI691
aI1165
aI1327
aI7901
aI9299
aI7777
aI39151
aI96261
aI79791
asS'd'
I16213
sa(dp32426
S'a'
I47417
sS's'
I18
sS'm'
(lp32427
I1
aI1
aI3
aI5
aI29
aI31
aI85
aI109
aI381
aI243
aI955
aI193
aI3461
aI5163
aI18607
aI51143
aI74457
aI84685
asS'd'
I16214
sa(dp32428
S'a'
I47420
sS's'
I18
sS'm'
(lp32429
I1
aI1
aI5
aI9
aI11
aI23
aI77
aI247
aI149
aI759
aI1153
aI1781
aI4107
aI16315
aI15513
aI48545
aI55607
aI163947
asS'd'
I16215
sa(dp32430
S'a'
I47426
sS's'
I18
sS'm'
(lp32431
I1
aI3
aI3
aI5
aI21
aI27
aI57
aI61
aI75
aI943
aI97
aI1507
aI4091
aI671
aI23023
aI49095
aI61649
aI222401
asS'd'
I16216
sa(dp32432
S'a'
I47443
sS's'
I18
sS'm'
(lp32433
I1
aI1
aI3
aI13
aI1
aI3
aI15
aI105
aI285
aI255
aI577
aI1347
aI2917
aI10257
aI21607
aI63041
aI79823
aI6447
asS'd'
I16217
sa(dp32434
S'a'
I47504
sS's'
I18
sS'm'
(lp32435
I1
aI1
aI3
aI1
aI17
aI25
aI109
aI47
aI445
aI225
aI1729
aI2835
aI4569
aI8755
aI21847
aI25839
aI43503
aI173599
asS'd'
I16218
sa(dp32436
S'a'
I47507
sS's'
I18
sS'm'
(lp32437
I1
aI1
aI5
aI15
aI19
aI1
aI121
aI119
aI33
aI77
aI1147
aI359
aI5747
aI2785
aI15567
aI5409
aI125979
aI199183
asS'd'
I16219
sa(dp32438
S'a'
I47510
sS's'
I18
sS'm'
(lp32439
I1
aI3
aI1
aI9
aI17
aI45
aI85
aI83
aI427
aI223
aI531
aI1681
aI2343
aI14959
aI27297
aI54607
aI70889
aI45529
asS'd'
I16220
sa(dp32440
S'a'
I47526
sS's'
I18
sS'm'
(lp32441
I1
aI1
aI5
aI11
aI31
aI61
aI109
aI195
aI505
aI197
aI159
aI2799
aI4517
aI11549
aI10297
aI17415
aI58277
aI206577
asS'd'
I16221
sa(dp32442
S'a'
I47550
sS's'
I18
sS'm'
(lp32443
I1
aI1
aI3
aI5
aI17
aI43
aI107
aI207
aI453
aI161
aI1775
aI1287
aI5775
aI12417
aI14201
aI48187
aI75073
aI121099
asS'd'
I16222
sa(dp32444
S'a'
I47552
sS's'
I18
sS'm'
(lp32445
I1
aI1
aI5
aI7
aI31
aI27
aI67
aI251
aI127
aI443
aI263
aI895
aI8081
aI14053
aI32023
aI54743
aI14723
aI221285
asS'd'
I16223
sa(dp32446
S'a'
I47557
sS's'
I18
sS'm'
(lp32447
I1
aI3
aI3
aI13
aI27
aI45
aI51
aI47
aI243
aI15
aI1283
aI2291
aI3613
aI14733
aI8777
aI3959
aI36769
aI104789
asS'd'
I16224
sa(dp32448
S'a'
I47564
sS's'
I18
sS'm'
(lp32449
I1
aI3
aI5
aI1
aI5
aI31
aI1
aI139
aI411
aI79
aI959
aI1431
aI2329
aI3595
aI19231
aI55747
aI18923
aI223709
asS'd'
I16225
sa(dp32450
S'a'
I47591
sS's'
I18
sS'm'
(lp32451
I1
aI3
aI1
aI1
aI7
aI9
aI69
aI201
aI305
aI411
aI459
aI3201
aI2525
aI6977
aI16249
aI17777
aI114321
aI243831
asS'd'
I16226
sa(dp32452
S'a'
I47610
sS's'
I18
sS'm'
(lp32453
I1
aI3
aI3
aI5
aI13
aI7
aI3
aI27
aI201
aI279
aI1253
aI1725
aI1481
aI8885
aI1233
aI40699
aI7267
aI189095
asS'd'
I16227
sa(dp32454
S'a'
I47621
sS's'
I18
sS'm'
(lp32455
I1
aI3
aI5
aI5
aI31
aI61
aI53
aI231
aI93
aI597
aI2027
aI2179
aI4517
aI565
aI27807
aI5447
aI130341
aI10411
asS'd'
I16228
sa(dp32456
S'a'
I47639
sS's'
I18
sS'm'
(lp32457
I1
aI1
aI7
aI1
aI9
aI17
aI63
aI245
aI405
aI23
aI1647
aI285
aI6625
aI8935
aI959
aI29095
aI657
aI185511
asS'd'
I16229
sa(dp32458
S'a'
I47646
sS's'
I18
sS'm'
(lp32459
I1
aI1
aI7
aI7
aI5
aI15
aI49
aI69
aI479
aI585
aI437
aI1097
aI5933
aI1709
aI26169
aI36895
aI16981
aI147033
asS'd'
I16230
sa(dp32460
S'a'
I47676
sS's'
I18
sS'm'
(lp32461
I1
aI3
aI7
aI13
aI29
aI19
aI89
aI249
aI195
aI687
aI379
aI1291
aI4791
aI9039
aI6381
aI12965
aI99995
aI105107
asS'd'
I16231
sa(dp32462
S'a'
I47682
sS's'
I18
sS'm'
(lp32463
I1
aI1
aI5
aI7
aI13
aI49
aI101
aI217
aI205
aI635
aI577
aI3301
aI911
aI1793
aI22285
aI20163
aI22593
aI45701
asS'd'
I16232
sa(dp32464
S'a'
I47732
sS's'
I18
sS'm'
(lp32465
I1
aI3
aI7
aI9
aI3
aI21
aI55
aI123
aI205
aI309
aI59
aI3739
aI1625
aI839
aI26733
aI27443
aI6699
aI244287
asS'd'
I16233
sa(dp32466
S'a'
I47741
sS's'
I18
sS'm'
(lp32467
I1
aI1
aI1
aI11
aI29
aI3
aI33
aI57
aI481
aI691
aI1401
aI3595
aI5435
aI571
aI6945
aI10911
aI94721
aI89751
asS'd'
I16234
sa(dp32468
S'a'
I47751
sS's'
I18
sS'm'
(lp32469
I1
aI1
aI3
aI5
aI21
aI19
aI23
aI169
aI263
aI137
aI771
aI1995
aI2211
aI6287
aI18691
aI14219
aI65647
aI89885
asS'd'
I16235
sa(dp32470
S'a'
I47757
sS's'
I18
sS'm'
(lp32471
I1
aI1
aI3
aI7
aI7
aI53
aI9
aI155
aI327
aI325
aI301
aI3703
aI1069
aI15573
aI14873
aI15665
aI71617
aI5079
asS'd'
I16236
sa(dp32472
S'a'
I47786
sS's'
I18
sS'm'
(lp32473
I1
aI3
aI1
aI3
aI27
aI15
aI1
aI203
aI465
aI629
aI71
aI1093
aI2071
aI7743
aI22441
aI42997
aI35249
aI113329
asS'd'
I16237
sa(dp32474
S'a'
I47803
sS's'
I18
sS'm'
(lp32475
I1
aI3
aI5
aI3
aI21
aI35
aI107
aI73
aI247
aI575
aI719
aI3215
aI3181
aI5861
aI25169
aI6503
aI12347
aI196371
asS'd'
I16238
sa(dp32476
S'a'
I47828
sS's'
I18
sS'm'
(lp32477
I1
aI3
aI1
aI3
aI7
aI29
aI117
aI115
aI221
aI345
aI165
aI1367
aI1491
aI15791
aI12647
aI34679
aI1043
aI219311
asS'd'
I16239
sa(dp32478
S'a'
I47848
sS's'
I18
sS'm'
(lp32479
I1
aI3
aI5
aI7
aI19
aI47
aI65
aI65
aI101
aI323
aI1209
aI3185
aI3803
aI1077
aI18933
aI17081
aI7475
aI165133
asS'd'
I16240
sa(dp32480
S'a'
I47853
sS's'
I18
sS'm'
(lp32481
I1
aI1
aI7
aI1
aI25
aI9
aI61
aI245
aI175
aI201
aI1837
aI2119
aI943
aI14043
aI14343
aI46299
aI81151
aI5587
asS'd'
I16241
sa(dp32482
S'a'
I47856
sS's'
I18
sS'm'
(lp32483
I1
aI1
aI1
aI1
aI31
aI27
aI49
aI89
aI387
aI975
aI1203
aI2995
aI2969
aI1465
aI925
aI39611
aI38101
aI126043
asS'd'
I16242
sa(dp32484
S'a'
I47862
sS's'
I18
sS'm'
(lp32485
I1
aI3
aI7
aI13
aI21
aI17
aI45
aI139
aI359
aI11
aI335
aI79
aI6629
aI6137
aI7879
aI62735
aI99493
aI138943
asS'd'
I16243
sa(dp32486
S'a'
I47874
sS's'
I18
sS'm'
(lp32487
I1
aI3
aI5
aI5
aI9
aI23
aI91
aI195
aI89
aI195
aI1931
aI3855
aI387
aI3491
aI29643
aI59939
aI32347
aI171539
asS'd'
I16244
sa(dp32488
S'a'
I47883
sS's'
I18
sS'm'
(lp32489
I1
aI3
aI3
aI15
aI21
aI55
aI5
aI13
aI139
aI125
aI1731
aI3131
aI1927
aI16051
aI8351
aI18625
aI32465
aI255813
asS'd'
I16245
sa(dp32490
S'a'
I47916
sS's'
I18
sS'm'
(lp32491
I1
aI3
aI7
aI11
aI21
aI45
aI53
aI225
aI33
aI733
aI115
aI639
aI4801
aI5529
aI11041
aI3557
aI83521
aI37525
asS'd'
I16246
sa(dp32492
S'a'
I47934
sS's'
I18
sS'm'
(lp32493
I1
aI3
aI7
aI15
aI3
aI37
aI111
aI127
aI463
aI565
aI543
aI2593
aI2203
aI5719
aI11667
aI63735
aI16481
aI155613
asS'd'
I16247
sa(dp32494
S'a'
I47951
sS's'
I18
sS'm'
(lp32495
I1
aI1
aI3
aI13
aI31
aI1
aI17
aI53
aI479
aI629
aI1517
aI89
aI3377
aI4831
aI9213
aI55029
aI69547
aI52363
asS'd'
I16248
sa(dp32496
S'a'
I47979
sS's'
I18
sS'm'
(lp32497
I1
aI1
aI7
aI1
aI5
aI47
aI115
aI73
aI59
aI407
aI1227
aI2955
aI5249
aI7921
aI4713
aI28699
aI41455
aI1161
asS'd'
I16249
sa(dp32498
S'a'
I48010
sS's'
I18
sS'm'
(lp32499
I1
aI1
aI7
aI7
aI5
aI7
aI39
aI7
aI97
aI867
aI247
aI639
aI3125
aI14961
aI19737
aI52589
aI59821
aI96095
asS'd'
I16250
sa(dp32500
S'a'
I48027
sS's'
I18
sS'm'
(lp32501
I1
aI1
aI7
aI5
aI11
aI51
aI63
aI245
aI385
aI1003
aI189
aI4039
aI6137
aI3621
aI13241
aI55753
aI14855
aI50221
asS'd'
I16251
sa(dp32502
S'a'
I48046
sS's'
I18
sS'm'
(lp32503
I1
aI3
aI5
aI3
aI9
aI61
aI83
aI209
aI301
aI917
aI259
aI187
aI6253
aI1579
aI28285
aI16571
aI100945
aI158067
asS'd'
I16252
sa(dp32504
S'a'
I48060
sS's'
I18
sS'm'
(lp32505
I1
aI3
aI5
aI11
aI5
aI47
aI123
aI173
aI253
aI183
aI1823
aI459
aI4719
aI13639
aI8455
aI12217
aI88779
aI134863
asS'd'
I16253
sa(dp32506
S'a'
I48065
sS's'
I18
sS'm'
(lp32507
I1
aI1
aI1
aI5
aI1
aI45
aI43
aI163
aI371
aI427
aI1791
aI1073
aI1615
aI14473
aI15895
aI4971
aI105269
aI109201
asS'd'
I16254
sa(dp32508
S'a'
I48075
sS's'
I18
sS'm'
(lp32509
I1
aI3
aI1
aI15
aI11
aI9
aI99
aI99
aI25
aI21
aI1499
aI83
aI6967
aI12923
aI13623
aI27423
aI4707
aI477
asS'd'
I16255
sa(dp32510
S'a'
I48096
sS's'
I18
sS'm'
(lp32511
I1
aI1
aI3
aI5
aI15
aI49
aI45
aI27
aI51
aI391
aI1849
aI347
aI6841
aI2831
aI4425
aI40701
aI61135
aI116945
asS'd'
I16256
sa(dp32512
S'a'
I48119
sS's'
I18
sS'm'
(lp32513
I1
aI1
aI3
aI7
aI21
aI23
aI15
aI223
aI403
aI395
aI1997
aI3247
aI6345
aI11739
aI6511
aI44323
aI86667
aI213287
asS'd'
I16257
sa(dp32514
S'a'
I48173
sS's'
I18
sS'm'
(lp32515
I1
aI3
aI7
aI11
aI31
aI21
aI75
aI129
aI427
aI777
aI1787
aI4031
aI1493
aI2279
aI10681
aI36675
aI25527
aI123533
asS'd'
I16258
sa(dp32516
S'a'
I48176
sS's'
I18
sS'm'
(lp32517
I1
aI3
aI5
aI15
aI9
aI5
aI117
aI147
aI259
aI265
aI1817
aI583
aI5341
aI12115
aI2369
aI4023
aI123479
aI218877
asS'd'
I16259
sa(dp32518
S'a'
I48186
sS's'
I18
sS'm'
(lp32519
I1
aI1
aI1
aI15
aI5
aI21
aI87
aI57
aI487
aI529
aI1129
aI2989
aI39
aI5995
aI28779
aI35813
aI97425
aI5227
asS'd'
I16260
sa(dp32520
S'a'
I48213
sS's'
I18
sS'm'
(lp32521
I1
aI1
aI7
aI3
aI7
aI1
aI41
aI231
aI195
aI205
aI1663
aI3149
aI4439
aI8241
aI3085
aI43965
aI58833
aI216779
asS'd'
I16261
sa(dp32522
S'a'
I48223
sS's'
I18
sS'm'
(lp32523
I1
aI1
aI3
aI1
aI23
aI45
aI65
aI3
aI55
aI653
aI131
aI2425
aI5653
aI9105
aI16921
aI55221
aI29241
aI220927
asS'd'
I16262
sa(dp32524
S'a'
I48224
sS's'
I18
sS'm'
(lp32525
I1
aI3
aI1
aI3
aI5
aI47
aI107
aI49
aI175
aI957
aI1287
aI1299
aI5215
aI4141
aI31697
aI9371
aI43339
aI133933
asS'd'
I16263
sa(dp32526
S'a'
I48229
sS's'
I18
sS'm'
(lp32527
I1
aI3
aI1
aI7
aI13
aI55
aI33
aI163
aI361
aI793
aI101
aI2159
aI3457
aI12893
aI11627
aI27115
aI95201
aI2945
asS'd'
I16264
sa(dp32528
S'a'
I48241
sS's'
I18
sS'm'
(lp32529
I1
aI1
aI7
aI15
aI11
aI39
aI79
aI113
aI385
aI41
aI1715
aI3887
aI1347
aI8141
aI6121
aI18653
aI50867
aI55745
asS'd'
I16265
sa(dp32530
S'a'
I48281
sS's'
I18
sS'm'
(lp32531
I1
aI1
aI7
aI13
aI3
aI5
aI23
aI59
aI223
aI665
aI1823
aI2989
aI1069
aI15161
aI8917
aI5539
aI47437
aI240933
asS'd'
I16266
sa(dp32532
S'a'
I48284
sS's'
I18
sS'm'
(lp32533
I1
aI3
aI7
aI13
aI13
aI19
aI73
aI25
aI413
aI211
aI439
aI339
aI1159
aI16063
aI9589
aI33451
aI28789
aI118883
asS'd'
I16267
sa(dp32534
S'a'
I48291
sS's'
I18
sS'm'
(lp32535
I1
aI1
aI5
aI11
aI13
aI61
aI81
aI63
aI197
aI569
aI961
aI253
aI2065
aI8969
aI24045
aI52811
aI26929
aI111329
asS'd'
I16268
sa(dp32536
S'a'
I48308
sS's'
I18
sS'm'
(lp32537
I1
aI3
aI7
aI5
aI19
aI23
aI27
aI163
aI37
aI103
aI1737
aI175
aI1853
aI597
aI14147
aI46159
aI26385
aI69427
asS'd'
I16269
sa(dp32538
S'a'
I48340
sS's'
I18
sS'm'
(lp32539
I1
aI3
aI7
aI9
aI27
aI43
aI45
aI209
aI61
aI115
aI645
aI1149
aI1473
aI10557
aI8541
aI51703
aI29207
aI92059
asS'd'
I16270
sa(dp32540
S'a'
I48344
sS's'
I18
sS'm'
(lp32541
I1
aI1
aI5
aI5
aI25
aI53
aI123
aI243
aI493
aI403
aI485
aI1505
aI6879
aI1921
aI13569
aI28271
aI24407
aI73057
asS'd'
I16271
sa(dp32542
S'a'
I48368
sS's'
I18
sS'm'
(lp32543
I1
aI1
aI3
aI11
aI15
aI31
aI99
aI29
aI503
aI819
aI55
aI773
aI2993
aI15341
aI7625
aI12835
aI28555
aI200609
asS'd'
I16272
sa(dp32544
S'a'
I48400
sS's'
I18
sS'm'
(lp32545
I1
aI1
aI7
aI9
aI21
aI49
aI47
aI207
aI13
aI571
aI57
aI1727
aI7441
aI1703
aI4253
aI64851
aI113
aI180273
asS'd'
I16273
sa(dp32546
S'a'
I48406
sS's'
I18
sS'm'
(lp32547
I1
aI3
aI3
aI9
aI23
aI1
aI1
aI53
aI195
aI569
aI925
aI1969
aI3031
aI14371
aI17673
aI39163
aI11819
aI117573
asS'd'
I16274
sa(dp32548
S'a'
I48409
sS's'
I18
sS'm'
(lp32549
I1
aI1
aI1
aI11
aI5
aI3
aI103
aI133
aI447
aI1015
aI1049
aI3283
aI7507
aI233
aI13721
aI50721
aI75511
aI227561
asS'd'
I16275
sa(dp32550
S'a'
I48421
sS's'
I18
sS'm'
(lp32551
I1
aI3
aI1
aI1
aI17
aI59
aI79
aI105
aI503
aI173
aI1575
aI1563
aI5807
aI12841
aI7983
aI28209
aI27527
aI229919
asS'd'
I16276
sa(dp32552
S'a'
I48448
sS's'
I18
sS'm'
(lp32553
I1
aI1
aI3
aI9
aI3
aI41
aI125
aI213
aI361
aI939
aI161
aI2235
aI2575
aI13519
aI7957
aI21527
aI40693
aI51055
asS'd'
I16277
sa(dp32554
S'a'
I48453
sS's'
I18
sS'm'
(lp32555
I1
aI3
aI7
aI11
aI11
aI7
aI117
aI21
aI7
aI745
aI1109
aI3393
aI953
aI5163
aI19909
aI49121
aI85061
aI209555
asS'd'
I16278
sa(dp32556
S'a'
I48457
sS's'
I18
sS'm'
(lp32557
I1
aI1
aI7
aI5
aI27
aI59
aI127
aI199
aI119
aI461
aI653
aI497
aI5867
aI767
aI16373
aI21201
aI87589
aI171491
asS'd'
I16279
sa(dp32558
S'a'
I48478
sS's'
I18
sS'm'
(lp32559
I1
aI1
aI7
aI11
aI7
aI21
aI121
aI97
aI17
aI827
aI1191
aI113
aI6527
aI14977
aI32567
aI15191
aI104541
aI140359
asS'd'
I16280
sa(dp32560
S'a'
I48493
sS's'
I18
sS'm'
(lp32561
I1
aI1
aI7
aI13
aI19
aI29
aI127
aI207
aI233
aI511
aI43
aI3177
aI7963
aI5559
aI24185
aI13373
aI37853
aI150537
asS'd'
I16281
sa(dp32562
S'a'
I48527
sS's'
I18
sS'm'
(lp32563
I1
aI3
aI3
aI9
aI15
aI17
aI119
aI201
aI443
aI51
aI17
aI605
aI4191
aI5251
aI28903
aI4861
aI92571
aI143499
asS'd'
I16282
sa(dp32564
S'a'
I48583
sS's'
I18
sS'm'
(lp32565
I1
aI1
aI5
aI11
aI21
aI45
aI111
aI245
aI115
aI705
aI1267
aI3013
aI7907
aI14973
aI30499
aI44117
aI118229
aI136571
asS'd'
I16283
sa(dp32566
S'a'
I48597
sS's'
I18
sS'm'
(lp32567
I1
aI3
aI7
aI7
aI31
aI35
aI69
aI143
aI269
aI27
aI1365
aI775
aI6067
aI11477
aI28475
aI54573
aI92827
aI226459
asS'd'
I16284
sa(dp32568
S'a'
I48641
sS's'
I18
sS'm'
(lp32569
I1
aI1
aI5
aI1
aI1
aI29
aI41
aI189
aI423
aI245
aI1031
aI1667
aI3465
aI1491
aI26787
aI53851
aI26189
aI215443
asS'd'
I16285
sa(dp32570
S'a'
I48668
sS's'
I18
sS'm'
(lp32571
I1
aI3
aI3
aI5
aI23
aI29
aI97
aI141
aI135
aI249
aI767
aI3627
aI7867
aI9311
aI25411
aI38325
aI118643
aI128453
asS'd'
I16286
sa(dp32572
S'a'
I48671
sS's'
I18
sS'm'
(lp32573
I1
aI3
aI7
aI13
aI3
aI25
aI61
aI121
aI285
aI401
aI1099
aI2327
aI5509
aI16127
aI2363
aI52395
aI114233
aI216013
asS'd'
I16287
sa(dp32574
S'a'
I48682
sS's'
I18
sS'm'
(lp32575
I1
aI1
aI1
aI13
aI7
aI37
aI83
aI151
aI447
aI605
aI289
aI2941
aI6273
aI5945
aI7493
aI29805
aI34935
aI101177
asS'd'
I16288
sa(dp32576
S'a'
I48709
sS's'
I18
sS'm'
(lp32577
I1
aI1
aI7
aI13
aI31
aI51
aI125
aI9
aI321
aI293
aI489
aI4023
aI2425
aI7645
aI2927
aI20973
aI111223
aI71255
asS'd'
I16289
sa(dp32578
S'a'
I48719
sS's'
I18
sS'm'
(lp32579
I1
aI1
aI7
aI3
aI27
aI9
aI59
aI143
aI307
aI543
aI1995
aI111
aI5807
aI4641
aI3581
aI2421
aI64213
aI187567
asS'd'
I16290
sa(dp32580
S'a'
I48731
sS's'
I18
sS'm'
(lp32581
I1
aI3
aI3
aI9
aI23
aI33
aI47
aI81
aI441
aI995
aI213
aI3501
aI1003
aI9885
aI24101
aI58767
aI49507
aI30525
asS'd'
I16291
sa(dp32582
S'a'
I48780
sS's'
I18
sS'm'
(lp32583
I1
aI3
aI1
aI3
aI19
aI51
aI117
aI53
aI49
aI167
aI1799
aI1421
aI2473
aI3183
aI14421
aI58621
aI130703
aI48095
asS'd'
I16292
sa(dp32584
S'a'
I48791
sS's'
I18
sS'm'
(lp32585
I1
aI1
aI7
aI7
aI5
aI25
aI33
aI237
aI447
aI377
aI517
aI2047
aI4357
aI2747
aI16491
aI23935
aI21655
aI144151
asS'd'
I16293
sa(dp32586
S'a'
I48797
sS's'
I18
sS'm'
(lp32587
I1
aI1
aI7
aI7
aI17
aI5
aI57
aI67
aI89
aI897
aI1163
aI3517
aI1651
aI6745
aI23449
aI2853
aI43829
aI50707
asS'd'
I16294
sa(dp32588
S'a'
I48807
sS's'
I18
sS'm'
(lp32589
I1
aI1
aI1
aI13
aI5
aI45
aI87
aI139
aI339
aI1021
aI649
aI2957
aI7887
aI11957
aI11235
aI11063
aI77329
aI93121
asS'd'
I16295
sa(dp32590
S'a'
I48811
sS's'
I18
sS'm'
(lp32591
I1
aI1
aI1
aI7
aI11
aI35
aI73
aI241
aI11
aI287
aI551
aI207
aI4701
aI2031
aI27191
aI44337
aI35577
aI226387
asS'd'
I16296
sa(dp32592
S'a'
I48814
sS's'
I18
sS'm'
(lp32593
I1
aI3
aI5
aI11
aI19
aI3
aI41
aI211
aI29
aI913
aI1455
aI2525
aI3935
aI5581
aI8043
aI12033
aI97479
aI73521
asS'd'
I16297
sa(dp32594
S'a'
I48821
sS's'
I18
sS'm'
(lp32595
I1
aI3
aI5
aI7
aI15
aI49
aI53
aI185
aI229
aI161
aI121
aI3407
aI903
aI7257
aI19535
aI42197
aI3983
aI222077
asS'd'
I16298
sa(dp32596
S'a'
I48836
sS's'
I18
sS'm'
(lp32597
I1
aI1
aI5
aI7
aI17
aI33
aI61
aI121
aI343
aI111
aI151
aI1147
aI3743
aI2423
aI24151
aI42307
aI50711
aI140317
asS'd'
I16299
sa(dp32598
S'a'
I48846
sS's'
I18
sS'm'
(lp32599
I1
aI3
aI3
aI7
aI27
aI1
aI119
aI209
aI243
aI185
aI1253
aI2307
aI7659
aI7839
aI9697
aI16799
aI5189
aI130005
asS'd'
I16300
sa(dp32600
S'a'
I48848
sS's'
I18
sS'm'
(lp32601
I1
aI3
aI3
aI1
aI27
aI21
aI115
aI175
aI157
aI735
aI1233
aI1133
aI1763
aI1125
aI4667
aI42569
aI125185
aI218417
asS'd'
I16301
sa(dp32602
S'a'
I48854
sS's'
I18
sS'm'
(lp32603
I1
aI3
aI5
aI9
aI5
aI37
aI97
aI165
aI87
aI447
aI1491
aI3141
aI597
aI10651
aI9727
aI65163
aI2469
aI141859
asS'd'
I16302
sa(dp32604
S'a'
I48874
sS's'
I18
sS'm'
(lp32605
I1
aI1
aI5
aI13
aI11
aI53
aI111
aI123
aI423
aI413
aI1519
aI3715
aI6623
aI15415
aI24085
aI20925
aI1529
aI85183
asS'd'
I16303
sa(dp32606
S'a'
I48908
sS's'
I18
sS'm'
(lp32607
I1
aI3
aI7
aI1
aI31
aI45
aI93
aI45
aI363
aI779
aI443
aI3687
aI6051
aI11683
aI31733
aI57251
aI55087
aI240877
asS'd'
I16304
sa(dp32608
S'a'
I48929
sS's'
I18
sS'm'
(lp32609
I1
aI1
aI1
aI15
aI1
aI17
aI53
aI237
aI131
aI673
aI1919
aI1531
aI7455
aI5043
aI1709
aI23375
aI75657
aI194393
asS'd'
I16305
sa(dp32610
S'a'
I48941
sS's'
I18
sS'm'
(lp32611
I1
aI3
aI5
aI9
aI21
aI3
aI101
aI141
aI387
aI887
aI1329
aI2627
aI2865
aI2685
aI20273
aI18901
aI43805
aI181049
asS'd'
I16306
sa(dp32612
S'a'
I48954
sS's'
I18
sS'm'
(lp32613
I1
aI3
aI5
aI5
aI7
aI49
aI55
aI243
aI311
aI653
aI959
aI3157
aI4891
aI8777
aI14381
aI28105
aI119323
aI181129
asS'd'
I16307
sa(dp32614
S'a'
I48961
sS's'
I18
sS'm'
(lp32615
I1
aI1
aI1
aI9
aI17
aI33
aI97
aI123
aI71
aI5
aI793
aI2829
aI4385
aI8577
aI9927
aI26213
aI53287
aI203555
asS'd'
I16308
sa(dp32616
S'a'
I48962
sS's'
I18
sS'm'
(lp32617
I1
aI1
aI1
aI9
aI15
aI61
aI51
aI31
aI195
aI1003
aI1915
aI2349
aI5319
aI13411
aI15265
aI36321
aI76157
aI200437
asS'd'
I16309
sa(dp32618
S'a'
I48985
sS's'
I18
sS'm'
(lp32619
I1
aI3
aI1
aI13
aI19
aI41
aI11
aI227
aI447
aI395
aI1885
aI2953
aI3537
aI3855
aI21611
aI14547
aI106573
aI233205
asS'd'
I16310
sa(dp32620
S'a'
I49001
sS's'
I18
sS'm'
(lp32621
I1
aI3
aI1
aI11
aI1
aI1
aI7
aI61
aI69
aI631
aI1687
aI1131
aI6901
aI5801
aI14431
aI17807
aI35777
aI253941
asS'd'
I16311
sa(dp32622
S'a'
I49007
sS's'
I18
sS'm'
(lp32623
I1
aI3
aI5
aI9
aI15
aI47
aI71
aI145
aI419
aI401
aI1781
aI2031
aI3157
aI14483
aI2393
aI4061
aI122053
aI81701
asS'd'
I16312
sa(dp32624
S'a'
I49015
sS's'
I18
sS'm'
(lp32625
I1
aI3
aI7
aI13
aI23
aI21
aI67
aI147
aI137
aI277
aI259
aI3119
aI4785
aI5349
aI3193
aI21805
aI108265
aI231111
asS'd'
I16313
sa(dp32626
S'a'
I49035
sS's'
I18
sS'm'
(lp32627
I1
aI1
aI3
aI9
aI23
aI63
aI77
aI153
aI75
aI643
aI1341
aI3607
aI6001
aI3387
aI17485
aI17893
aI124699
aI99515
asS'd'
I16314
sa(dp32628
S'a'
I49062
sS's'
I18
sS'm'
(lp32629
I1
aI1
aI7
aI11
aI9
aI51
aI79
aI255
aI255
aI723
aI1797
aI2805
aI2505
aI3437
aI4835
aI30731
aI73741
aI8051
asS'd'
I16315
sa(dp32630
S'a'
I49066
sS's'
I18
sS'm'
(lp32631
I1
aI1
aI5
aI7
aI5
aI19
aI73
aI75
aI19
aI889
aI1443
aI2263
aI5773
aI9997
aI329
aI46679
aI82313
aI130897
asS'd'
I16316
sa(dp32632
S'a'
I49085
sS's'
I18
sS'm'
(lp32633
I1
aI3
aI3
aI9
aI7
aI53
aI69
aI11
aI287
aI861
aI1367
aI1433
aI2693
aI16255
aI6785
aI24705
aI77463
aI231247
asS'd'
I16317
sa(dp32634
S'a'
I49091
sS's'
I18
sS'm'
(lp32635
I1
aI1
aI3
aI7
aI21
aI15
aI115
aI131
aI169
aI447
aI1731
aI2873
aI119
aI643
aI31719
aI22193
aI17959
aI150567
asS'd'
I16318
sa(dp32636
S'a'
I49094
sS's'
I18
sS'm'
(lp32637
I1
aI3
aI3
aI9
aI23
aI21
aI7
aI39
aI185
aI837
aI379
aI3687
aI3751
aI14801
aI15231
aI7239
aI77521
aI80135
asS'd'
I16319
sa(dp32638
S'a'
I49127
sS's'
I18
sS'm'
(lp32639
I1
aI1
aI1
aI1
aI7
aI33
aI93
aI247
aI487
aI873
aI1951
aI3273
aI551
aI3735
aI29477
aI62387
aI30361
aI145613
asS'd'
I16320
sa(dp32640
S'a'
I49134
sS's'
I18
sS'm'
(lp32641
I1
aI1
aI3
aI7
aI5
aI47
aI17
aI31
aI93
aI517
aI1175
aI1033
aI7421
aI75
aI541
aI1967
aI13059
aI104751
asS'd'
I16321
sa(dp32642
S'a'
I49183
sS's'
I18
sS'm'
(lp32643
I1
aI3
aI5
aI3
aI19
aI23
aI123
aI115
aI99
aI747
aI655
aI2885
aI2991
aI7497
aI5989
aI33419
aI54363
aI224947
asS'd'
I16322
sa(dp32644
S'a'
I49221
sS's'
I18
sS'm'
(lp32645
I1
aI3
aI3
aI1
aI15
aI9
aI51
aI3
aI307
aI879
aI1065
aI1835
aI2267
aI11299
aI21995
aI2997
aI102207
aI39
asS'd'
I16323
sa(dp32646
S'a'
I49231
sS's'
I18
sS'm'
(lp32647
I1
aI3
aI5
aI3
aI9
aI19
aI19
aI79
aI499
aI121
aI951
aI1151
aI379
aI5299
aI13727
aI49061
aI32605
aI208683
asS'd'
I16324
sa(dp32648
S'a'
I49250
sS's'
I18
sS'm'
(lp32649
I1
aI3
aI7
aI7
aI23
aI41
aI71
aI23
aI493
aI969
aI271
aI3729
aI6199
aI14693
aI17625
aI52509
aI121257
aI151175
asS'd'
I16325
sa(dp32650
S'a'
I49252
sS's'
I18
sS'm'
(lp32651
I1
aI1
aI7
aI15
aI9
aI51
aI61
aI9
aI147
aI527
aI803
aI3255
aI723
aI3241
aI23961
aI17497
aI122569
aI50863
asS'd'
I16326
sa(dp32652
S'a'
I49270
sS's'
I18
sS'm'
(lp32653
I1
aI3
aI5
aI1
aI19
aI19
aI31
aI31
aI95
aI667
aI377
aI1193
aI6759
aI8583
aI29801
aI7989
aI113899
aI43481
asS'd'
I16327
sa(dp32654
S'a'
I49297
sS's'
I18
sS'm'
(lp32655
I1
aI3
aI7
aI5
aI17
aI47
aI1
aI7
aI305
aI489
aI1035
aI1335
aI5901
aI9635
aI10433
aI4235
aI108577
aI52661
asS'd'
I16328
sa(dp32656
S'a'
I49313
sS's'
I18
sS'm'
(lp32657
I1
aI1
aI7
aI3
aI15
aI7
aI105
aI189
aI429
aI287
aI679
aI111
aI2087
aI8479
aI3053
aI21763
aI65655
aI16207
asS'd'
I16329
sa(dp32658
S'a'
I49340
sS's'
I18
sS'm'
(lp32659
I1
aI3
aI5
aI13
aI19
aI41
aI45
aI115
aI311
aI653
aI1301
aI3797
aI6183
aI7203
aI9829
aI13263
aI71649
aI255611
asS'd'
I16330
sa(dp32660
S'a'
I49365
sS's'
I18
sS'm'
(lp32661
I1
aI3
aI1
aI9
aI21
aI33
aI19
aI233
aI83
aI423
aI1701
aI347
aI3999
aI11885
aI19699
aI61931
aI63895
aI59615
asS'd'
I16331
sa(dp32662
S'a'
I49379
sS's'
I18
sS'm'
(lp32663
I1
aI1
aI7
aI7
aI15
aI19
aI61
aI43
aI183
aI223
aI1135
aI3099
aI3137
aI16181
aI31019
aI20151
aI30169
aI13125
asS'd'
I16332
sa(dp32664
S'a'
I49386
sS's'
I18
sS'm'
(lp32665
I1
aI1
aI1
aI11
aI9
aI11
aI3
aI157
aI495
aI207
aI603
aI1919
aI3565
aI7639
aI4281
aI49179
aI2465
aI224033
asS'd'
I16333
sa(dp32666
S'a'
I49391
sS's'
I18
sS'm'
(lp32667
I1
aI3
aI1
aI13
aI13
aI19
aI127
aI81
aI475
aI659
aI1053
aI3153
aI5223
aI12567
aI30945
aI47109
aI98253
aI41571
asS'd'
I16334
sa(dp32668
S'a'
I49417
sS's'
I18
sS'm'
(lp32669
I1
aI1
aI5
aI11
aI19
aI1
aI9
aI67
aI501
aI149
aI1695
aI1265
aI6467
aI5285
aI29865
aI46777
aI97581
aI209395
asS'd'
I16335
sa(dp32670
S'a'
I49438
sS's'
I18
sS'm'
(lp32671
I1
aI1
aI1
aI11
aI21
aI23
aI123
aI103
aI419
aI787
aI111
aI3327
aI673
aI3425
aI8097
aI2411
aI107353
aI121379
asS'd'
I16336
sa(dp32672
S'a'
I49444
sS's'
I18
sS'm'
(lp32673
I1
aI1
aI3
aI3
aI17
aI7
aI87
aI35
aI501
aI973
aI351
aI751
aI4813
aI12245
aI3053
aI38633
aI44995
aI37151
asS'd'
I16337
sa(dp32674
S'a'
I49453
sS's'
I18
sS'm'
(lp32675
I1
aI3
aI5
aI15
aI7
aI57
aI99
aI53
aI3
aI355
aI153
aI3897
aI6923
aI2075
aI10821
aI41819
aI75665
aI21237
asS'd'
I16338
sa(dp32676
S'a'
I49454
sS's'
I18
sS'm'
(lp32677
I1
aI3
aI5
aI9
aI3
aI29
aI85
aI115
aI191
aI977
aI1067
aI1539
aI5287
aI3589
aI18017
aI35571
aI41633
aI148565
asS'd'
I16339
sa(dp32678
S'a'
I49480
sS's'
I18
sS'm'
(lp32679
I1
aI1
aI3
aI9
aI9
aI57
aI1
aI63
aI333
aI107
aI1173
aI3377
aI3599
aI277
aI20643
aI19295
aI91169
aI168891
asS'd'
I16340
sa(dp32680
S'a'
I49485
sS's'
I18
sS'm'
(lp32681
I1
aI1
aI5
aI13
aI21
aI7
aI55
aI15
aI79
aI995
aI1553
aI949
aI6357
aI8137
aI9539
aI861
aI95635
aI63021
asS'd'
I16341
sa(dp32682
S'a'
I49498
sS's'
I18
sS'm'
(lp32683
I1
aI1
aI5
aI13
aI31
aI59
aI11
aI55
aI159
aI449
aI307
aI2725
aI2305
aI8259
aI13823
aI51225
aI44775
aI54253
asS'd'
I16342
sa(dp32684
S'a'
I49513
sS's'
I18
sS'm'
(lp32685
I1
aI3
aI3
aI15
aI19
aI61
aI105
aI19
aI211
aI693
aI835
aI3607
aI3703
aI14007
aI24597
aI47109
aI49855
aI166115
asS'd'
I16343
sa(dp32686
S'a'
I49527
sS's'
I18
sS'm'
(lp32687
I1
aI3
aI1
aI9
aI27
aI33
aI81
aI229
aI445
aI171
aI247
aI2019
aI5227
aI8759
aI30155
aI1851
aI26909
aI47145
asS'd'
I16344
sa(dp32688
S'a'
I49531
sS's'
I18
sS'm'
(lp32689
I1
aI1
aI3
aI5
aI23
aI1
aI57
aI49
aI177
aI451
aI1283
aI3859
aI4719
aI13507
aI29439
aI9155
aI129585
aI94713
asS'd'
I16345
sa(dp32690
S'a'
I49564
sS's'
I18
sS'm'
(lp32691
I1
aI3
aI3
aI5
aI1
aI9
aI37
aI23
aI423
aI611
aI113
aI119
aI4307
aI11747
aI93
aI51213
aI5479
aI172061
asS'd'
I16346
sa(dp32692
S'a'
I49588
sS's'
I18
sS'm'
(lp32693
I1
aI3
aI3
aI15
aI5
aI35
aI9
aI43
aI423
aI877
aI1563
aI3411
aI4233
aI6961
aI22421
aI7347
aI39553
aI199745
asS'd'
I16347
sa(dp32694
S'a'
I49609
sS's'
I18
sS'm'
(lp32695
I1
aI3
aI1
aI9
aI15
aI35
aI85
aI57
aI151
aI323
aI1627
aI27
aI3603
aI11475
aI6561
aI6091
aI72099
aI185321
asS'd'
I16348
sa(dp32696
S'a'
I49617
sS's'
I18
sS'm'
(lp32697
I1
aI1
aI3
aI15
aI17
aI21
aI3
aI119
aI455
aI147
aI1077
aI3955
aI1355
aI3083
aI28217
aI50745
aI114445
aI46563
asS'd'
I16349
sa(dp32698
S'a'
I49645
sS's'
I18
sS'm'
(lp32699
I1
aI1
aI5
aI3
aI13
aI11
aI35
aI185
aI357
aI447
aI649
aI139
aI2847
aI12011
aI7753
aI671
aI113693
aI236281
asS'd'
I16350
sa(dp32700
S'a'
I49646
sS's'
I18
sS'm'
(lp32701
I1
aI3
aI5
aI1
aI19
aI27
aI117
aI195
aI67
aI259
aI1073
aI81
aI5527
aI6829
aI26675
aI60029
aI8159
aI13553
asS'd'
I16351
sa(dp32702
S'a'
I49654
sS's'
I18
sS'm'
(lp32703
I1
aI3
aI3
aI7
aI27
aI63
aI115
aI191
aI431
aI191
aI1547
aI2261
aI6527
aI13459
aI12773
aI37485
aI114847
aI61709
asS'd'
I16352
sa(dp32704
S'a'
I49697
sS's'
I18
sS'm'
(lp32705
I1
aI1
aI3
aI1
aI29
aI63
aI47
aI77
aI415
aI191
aI325
aI2487
aI7457
aI13721
aI1573
aI15037
aI31941
aI226651
asS'd'
I16353
sa(dp32706
S'a'
I49718
sS's'
I18
sS'm'
(lp32707
I1
aI1
aI5
aI11
aI29
aI59
aI49
aI209
aI195
aI901
aI691
aI3167
aI7471
aI11609
aI30135
aI22067
aI71395
aI248151
asS'd'
I16354
sa(dp32708
S'a'
I49741
sS's'
I18
sS'm'
(lp32709
I1
aI3
aI7
aI11
aI7
aI43
aI55
aI69
aI289
aI735
aI1479
aI139
aI1395
aI6463
aI19827
aI52151
aI18963
aI103367
asS'd'
I16355
sa(dp32710
S'a'
I49754
sS's'
I18
sS'm'
(lp32711
I1
aI3
aI5
aI5
aI27
aI45
aI77
aI79
aI89
aI79
aI913
aI1437
aI3337
aI8861
aI5477
aI46195
aI105869
aI242599
asS'd'
I16356
sa(dp32712
S'a'
I49756
sS's'
I18
sS'm'
(lp32713
I1
aI3
aI5
aI3
aI3
aI27
aI103
aI207
aI233
aI995
aI1173
aI1143
aI3517
aI3207
aI8373
aI45145
aI79687
aI150107
asS'd'
I16357
sa(dp32714
S'a'
I49833
sS's'
I18
sS'm'
(lp32715
I1
aI3
aI7
aI3
aI7
aI37
aI1
aI137
aI45
aI479
aI861
aI3863
aI4249
aI3075
aI23639
aI12531
aI118473
aI255805
asS'd'
I16358
sa(dp32716
S'a'
I49834
sS's'
I18
sS'm'
(lp32717
I1
aI1
aI1
aI9
aI7
aI23
aI125
aI57
aI347
aI215
aI1749
aI3029
aI97
aI14715
aI16213
aI29291
aI104725
aI92043
asS'd'
I16359
sa(dp32718
S'a'
I49879
sS's'
I18
sS'm'
(lp32719
I1
aI3
aI3
aI15
aI31
aI39
aI111
aI97
aI369
aI675
aI1471
aI3889
aI6437
aI1499
aI13325
aI46141
aI121087
aI202493
asS'd'
I16360
sa(dp32720
S'a'
I49942
sS's'
I18
sS'm'
(lp32721
I1
aI1
aI1
aI15
aI15
aI21
aI3
aI167
aI251
aI793
aI1291
aI427
aI6407
aI15521
aI15441
aI20071
aI54513
aI200485
asS'd'
I16361
sa(dp32722
S'a'
I49946
sS's'
I18
sS'm'
(lp32723
I1
aI1
aI7
aI11
aI27
aI47
aI123
aI195
aI241
aI385
aI609
aI2323
aI1221
aI815
aI279
aI28553
aI17663
aI180977
asS'd'
I16362
sa(dp32724
S'a'
I49982
sS's'
I18
sS'm'
(lp32725
I1
aI1
aI3
aI15
aI23
aI53
aI21
aI19
aI329
aI903
aI561
aI1539
aI2829
aI13037
aI29867
aI27825
aI84005
aI14853
asS'd'
I16363
sa(dp32726
S'a'
I49994
sS's'
I18
sS'm'
(lp32727
I1
aI1
aI5
aI1
aI21
aI27
aI103
aI99
aI31
aI829
aI515
aI3949
aI3635
aI935
aI2127
aI56659
aI123477
aI47003
asS'd'
I16364
sa(dp32728
S'a'
I50001
sS's'
I18
sS'm'
(lp32729
I1
aI3
aI1
aI11
aI17
aI63
aI51
aI15
aI319
aI539
aI55
aI2561
aI1109
aI16115
aI23387
aI20051
aI120867
aI4019
asS'd'
I16365
sa(dp32730
S'a'
I50023
sS's'
I18
sS'm'
(lp32731
I1
aI3
aI3
aI7
aI31
aI11
aI45
aI251
aI209
aI21
aI553
aI887
aI393
aI15183
aI14735
aI13163
aI123681
aI15013
asS'd'
I16366
sa(dp32732
S'a'
I50047
sS's'
I18
sS'm'
(lp32733
I1
aI3
aI1
aI7
aI21
aI41
aI105
aI39
aI213
aI477
aI985
aI2375
aI937
aI7099
aI8867
aI36519
aI60503
aI143225
asS'd'
I16367
sa(dp32734
S'a'
I50120
sS's'
I18
sS'm'
(lp32735
I1
aI3
aI1
aI1
aI19
aI55
aI73
aI171
aI379
aI271
aI791
aI2477
aI5381
aI1703
aI3805
aI37227
aI55553
aI34549
asS'd'
I16368
sa(dp32736
S'a'
I50133
sS's'
I18
sS'm'
(lp32737
I1
aI1
aI5
aI5
aI29
aI51
aI127
aI111
aI209
aI663
aI975
aI2293
aI8155
aI11263
aI7891
aI15463
aI74889
aI227403
asS'd'
I16369
sa(dp32738
S'a'
I50137
sS's'
I18
sS'm'
(lp32739
I1
aI3
aI5
aI1
aI15
aI1
aI115
aI221
aI23
aI247
aI1597
aI87
aI7513
aI3329
aI14491
aI8961
aI78147
aI89499
asS'd'
I16370
sa(dp32740
S'a'
I50140
sS's'
I18
sS'm'
(lp32741
I1
aI3
aI5
aI15
aI25
aI23
aI99
aI85
aI189
aI487
aI1559
aI427
aI6179
aI5291
aI27279
aI65507
aI77347
aI177073
asS'd'
I16371
sa(dp32742
S'a'
I50156
sS's'
I18
sS'm'
(lp32743
I1
aI3
aI7
aI1
aI23
aI35
aI7
aI89
aI455
aI659
aI2039
aI3115
aI4057
aI139
aI1269
aI62319
aI13629
aI201571
asS'd'
I16372
sa(dp32744
S'a'
I50199
sS's'
I18
sS'm'
(lp32745
I1
aI1
aI1
aI5
aI9
aI47
aI55
aI171
aI213
aI813
aI135
aI1545
aI1421
aI4055
aI26697
aI18889
aI57653
aI201369
asS'd'
I16373
sa(dp32746
S'a'
I50205
sS's'
I18
sS'm'
(lp32747
I1
aI3
aI1
aI11
aI9
aI45
aI5
aI69
aI261
aI907
aI1229
aI1371
aI1867
aI1771
aI17309
aI41759
aI119129
aI43521
asS'd'
I16374
sa(dp32748
S'a'
I50215
sS's'
I18
sS'm'
(lp32749
I1
aI1
aI7
aI5
aI11
aI27
aI9
aI237
aI405
aI53
aI1359
aI3665
aI169
aI3547
aI23331
aI14353
aI106627
aI219711
asS'd'
I16375
sa(dp32750
S'a'
I50216
sS's'
I18
sS'm'
(lp32751
I1
aI3
aI3
aI5
aI17
aI23
aI9
aI41
aI81
aI601
aI1447
aI3225
aI6721
aI161
aI16109
aI54331
aI111273
aI81061
asS'd'
I16376
sa(dp32752
S'a'
I50224
sS's'
I18
sS'm'
(lp32753
I1
aI1
aI5
aI7
aI7
aI37
aI71
aI13
aI505
aI671
aI425
aI2771
aI1131
aI1259
aI13715
aI45779
aI114839
aI217499
asS'd'
I16377
sa(dp32754
S'a'
I50229
sS's'
I18
sS'm'
(lp32755
I1
aI1
aI1
aI5
aI9
aI51
aI85
aI223
aI117
aI403
aI1117
aI723
aI2465
aI11947
aI7495
aI62991
aI17825
aI169147
asS'd'
I16378
sa(dp32756
S'a'
I50233
sS's'
I18
sS'm'
(lp32757
I1
aI1
aI3
aI13
aI5
aI57
aI47
aI35
aI313
aI901
aI365
aI1265
aI137
aI6335
aI31419
aI38497
aI93285
aI177187
asS'd'
I16379
sa(dp32758
S'a'
I50251
sS's'
I18
sS'm'
(lp32759
I1
aI3
aI5
aI5
aI7
aI9
aI21
aI109
aI47
aI557
aI1449
aI3459
aI5175
aI5135
aI727
aI55425
aI60593
aI86571
asS'd'
I16380
sa(dp32760
S'a'
I50281
sS's'
I18
sS'm'
(lp32761
I1
aI3
aI7
aI5
aI1
aI17
aI119
aI239
aI347
aI75
aI1345
aI2765
aI1491
aI16297
aI25233
aI60401
aI85433
aI59287
asS'd'
I16381
sa(dp32762
S'a'
I50326
sS's'
I18
sS'm'
(lp32763
I1
aI1
aI5
aI5
aI19
aI57
aI41
aI53
aI161
aI475
aI1791
aI3617
aI1651
aI15227
aI30357
aI63547
aI69937
aI246473
asS'd'
I16382
sa(dp32764
S'a'
I50332
sS's'
I18
sS'm'
(lp32765
I1
aI3
aI7
aI11
aI11
aI43
aI79
aI79
aI195
aI437
aI39
aI1259
aI6041
aI14989
aI6615
aI58747
aI43583
aI246979
asS'd'
I16383
sa(dp32766
S'a'
I50348
sS's'
I18
sS'm'
(lp32767
I1
aI1
aI7
aI9
aI11
aI13
aI49
aI133
aI365
aI931
aI1089
aI713
aI5997
aI8759
aI5789
aI61329
aI22639
aI149845
asS'd'
I16384
sa(dp32768
S'a'
I50359
sS's'
I18
sS'm'
(lp32769
I1
aI3
aI7
aI13
aI9
aI3
aI11
aI253
aI339
aI883
aI1933
aI443
aI4265
aI14165
aI15845
aI12625
aI1453
aI70961
asS'd'
I16385
sa(dp32770
S'a'
I50374
sS's'
I18
sS'm'
(lp32771
I1
aI1
aI7
aI13
aI21
aI35
aI113
aI83
aI473
aI719
aI1601
aI1727
aI3715
aI631
aI28075
aI17725
aI11393
aI116883
asS'd'
I16386
sa(dp32772
S'a'
I50404
sS's'
I18
sS'm'
(lp32773
I1
aI3
aI7
aI7
aI7
aI13
aI117
aI83
aI365
aI529
aI1297
aI3903
aI2633
aI9617
aI15819
aI38137
aI49065
aI189445
asS'd'
I16387
sa(dp32774
S'a'
I50443
sS's'
I18
sS'm'
(lp32775
I1
aI1
aI1
aI7
aI11
aI41
aI107
aI33
aI381
aI395
aI1993
aI2819
aI3301
aI7543
aI6787
aI27437
aI113681
aI132403
asS'd'
I16388
sa(dp32776
S'a'
I50488
sS's'
I18
sS'm'
(lp32777
I1
aI1
aI1
aI1
aI3
aI59
aI91
aI235
aI67
aI987
aI1587
aI1119
aI5851
aI13201
aI1125
aI49709
aI381
aI183295
asS'd'
I16389
sa(dp32778
S'a'
I50493
sS's'
I18
sS'm'
(lp32779
I1
aI3
aI5
aI9
aI3
aI33
aI45
aI187
aI455
aI151
aI823
aI565
aI5725
aI1927
aI25387
aI61785
aI130271
aI134083
asS'd'
I16390
sa(dp32780
S'a'
I50505
sS's'
I18
sS'm'
(lp32781
I1
aI1
aI7
aI5
aI15
aI25
aI63
aI231
aI133
aI401
aI307
aI2961
aI4249
aI2639
aI10207
aI8349
aI8203
aI72783
asS'd'
I16391
sa(dp32782
S'a'
I50511
sS's'
I18
sS'm'
(lp32783
I1
aI1
aI5
aI13
aI19
aI27
aI43
aI153
aI165
aI815
aI1385
aI3041
aI853
aI7683
aI1035
aI13255
aI69779
aI128765
asS'd'
I16392
sa(dp32784
S'a'
I50525
sS's'
I18
sS'm'
(lp32785
I1
aI1
aI1
aI3
aI31
aI61
aI97
aI15
aI327
aI717
aI841
aI643
aI4781
aI11609
aI14181
aI14625
aI75369
aI251015
asS'd'
I16393
sa(dp32786
S'a'
I50542
sS's'
I18
sS'm'
(lp32787
I1
aI1
aI5
aI11
aI17
aI33
aI9
aI45
aI111
aI91
aI1923
aI967
aI5649
aI13647
aI30497
aI2925
aI18395
aI255089
asS'd'
I16394
sa(dp32788
S'a'
I50550
sS's'
I18
sS'm'
(lp32789
I1
aI3
aI3
aI1
aI3
aI55
aI33
aI29
aI101
aI211
aI1731
aI2351
aI7389
aI14935
aI29703
aI60031
aI122305
aI174323
asS'd'
I16395
sa(dp32790
S'a'
I50560
sS's'
I18
sS'm'
(lp32791
I1
aI1
aI1
aI3
aI13
aI23
aI79
aI117
aI367
aI267
aI143
aI1537
aI3159
aI2303
aI9653
aI12029
aI2841
aI226971
asS'd'
I16396
sa(dp32792
S'a'
I50570
sS's'
I18
sS'm'
(lp32793
I1
aI1
aI1
aI13
aI19
aI57
aI103
aI159
aI315
aI895
aI1879
aI2153
aI1901
aI7635
aI14145
aI56495
aI6203
aI151203
asS'd'
I16397
sa(dp32794
S'a'
I50587
sS's'
I18
sS'm'
(lp32795
I1
aI3
aI3
aI11
aI31
aI11
aI75
aI5
aI419
aI963
aI1809
aI3751
aI7291
aI12697
aI2841
aI17965
aI91707
aI80361
asS'd'
I16398
sa(dp32796
S'a'
I50593
sS's'
I18
sS'm'
(lp32797
I1
aI1
aI7
aI7
aI5
aI23
aI29
aI199
aI413
aI501
aI19
aI697
aI6523
aI3997
aI4945
aI59817
aI127613
aI220399
asS'd'
I16399
sa(dp32798
S'a'
I50611
sS's'
I18
sS'm'
(lp32799
I1
aI1
aI7
aI5
aI1
aI17
aI51
aI9
aI183
aI689
aI325
aI2391
aI2095
aI1907
aI10325
aI51045
aI20097
aI33719
asS'd'
I16400
sa(dp32800
S'a'
I50649
sS's'
I18
sS'm'
(lp32801
I1
aI3
aI5
aI11
aI3
aI33
aI45
aI221
aI325
aI7
aI253
aI1323
aI4959
aI14067
aI10035
aI39463
aI123171
aI194581
asS'd'
I16401
sa(dp32802
S'a'
I50656
sS's'
I18
sS'm'
(lp32803
I1
aI1
aI3
aI3
aI15
aI45
aI25
aI57
aI357
aI907
aI1249
aI3279
aI2631
aI9209
aI7857
aI58233
aI29049
aI173859
asS'd'
I16402
sa(dp32804
S'a'
I50685
sS's'
I18
sS'm'
(lp32805
I1
aI3
aI3
aI9
aI29
aI39
aI105
aI55
aI295
aI583
aI825
aI1777
aI2403
aI9489
aI9079
aI24855
aI18155
aI252733
asS'd'
I16403
sa(dp32806
S'a'
I50686
sS's'
I18
sS'm'
(lp32807
I1
aI1
aI5
aI15
aI5
aI63
aI77
aI215
aI287
aI743
aI1937
aI2103
aI2673
aI15487
aI27855
aI46683
aI120215
aI89721
asS'd'
I16404
sa(dp32808
S'a'
I50692
sS's'
I18
sS'm'
(lp32809
I1
aI3
aI3
aI5
aI27
aI1
aI45
aI19
aI309
aI679
aI1405
aI415
aI6107
aI13567
aI5803
aI61941
aI130285
aI51847
asS'd'
I16405
sa(dp32810
S'a'
I50702
sS's'
I18
sS'm'
(lp32811
I1
aI3
aI1
aI15
aI23
aI49
aI47
aI113
aI401
aI825
aI1299
aI2711
aI6509
aI12225
aI16147
aI20615
aI121305
aI121937
asS'd'
I16406
sa(dp32812
S'a'
I50704
sS's'
I18
sS'm'
(lp32813
I1
aI3
aI5
aI5
aI31
aI41
aI31
aI57
aI385
aI919
aI593
aI1371
aI6773
aI12099
aI23767
aI17663
aI128321
aI188921
asS'd'
I16407
sa(dp32814
S'a'
I50709
sS's'
I18
sS'm'
(lp32815
I1
aI1
aI3
aI15
aI5
aI41
aI17
aI7
aI479
aI143
aI549
aI1827
aI8107
aI14775
aI28817
aI12297
aI119893
aI191297
asS'd'
I16408
sa(dp32816
S'a'
I50720
sS's'
I18
sS'm'
(lp32817
I1
aI3
aI7
aI5
aI29
aI61
aI27
aI123
aI269
aI223
aI121
aI1745
aI3513
aI1989
aI9759
aI8129
aI78933
aI40085
asS'd'
I16409
sa(dp32818
S'a'
I50726
sS's'
I18
sS'm'
(lp32819
I1
aI1
aI5
aI13
aI23
aI49
aI125
aI225
aI479
aI123
aI41
aI2359
aI4443
aI4729
aI31717
aI3139
aI3869
aI118803
asS'd'
I16410
sa(dp32820
S'a'
I50729
sS's'
I18
sS'm'
(lp32821
I1
aI3
aI3
aI3
aI15
aI7
aI7
aI87
aI415
aI95
aI1799
aI2009
aI4711
aI15635
aI21997
aI47201
aI16815
aI99815
asS'd'
I16411
sa(dp32822
S'a'
I50737
sS's'
I18
sS'm'
(lp32823
I1
aI3
aI3
aI1
aI31
aI57
aI73
aI31
aI423
aI363
aI1469
aI2411
aI6449
aI15275
aI14189
aI51079
aI130201
aI130181
asS'd'
I16412
sa(dp32824
S'a'
I50758
sS's'
I18
sS'm'
(lp32825
I1
aI1
aI5
aI3
aI3
aI31
aI107
aI11
aI355
aI647
aI1463
aI3019
aI3263
aI13727
aI10461
aI26577
aI4439
aI132737
asS'd'
I16413
sa(dp32826
S'a'
I50772
sS's'
I18
sS'm'
(lp32827
I1
aI1
aI1
aI7
aI13
aI55
aI31
aI227
aI71
aI563
aI1467
aI3733
aI725
aI3443
aI19279
aI4111
aI35749
aI62275
asS'd'
I16414
sa(dp32828
S'a'
I50849
sS's'
I18
sS'm'
(lp32829
I1
aI3
aI3
aI15
aI9
aI17
aI61
aI95
aI43
aI583
aI1381
aI1285
aI2655
aI9213
aI27551
aI16237
aI29569
aI216879
asS'd'
I16415
sa(dp32830
S'a'
I50850
sS's'
I18
sS'm'
(lp32831
I1
aI1
aI1
aI3
aI29
aI3
aI87
aI193
aI53
aI599
aI1581
aI907
aI4381
aI8697
aI27299
aI40259
aI122653
aI43559
asS'd'
I16416
sa(dp32832
S'a'
I50881
sS's'
I18
sS'm'
(lp32833
I1
aI3
aI5
aI3
aI15
aI49
aI21
aI27
aI35
aI353
aI1281
aI3253
aI7339
aI9333
aI25253
aI35035
aI30379
aI87387
asS'd'
I16417
sa(dp32834
S'a'
I50901
sS's'
I18
sS'm'
(lp32835
I1
aI1
aI1
aI15
aI7
aI27
aI69
aI41
aI45
aI729
aI1005
aI3495
aI1445
aI7421
aI27443
aI29609
aI70105
aI93883
asS'd'
I16418
sa(dp32836
S'a'
I50906
sS's'
I18
sS'm'
(lp32837
I1
aI1
aI7
aI7
aI1
aI41
aI15
aI149
aI107
aI121
aI639
aI3703
aI3397
aI1727
aI14165
aI2845
aI78531
aI175767
asS'd'
I16419
sa(dp32838
S'a'
I50915
sS's'
I18
sS'm'
(lp32839
I1
aI3
aI7
aI15
aI27
aI17
aI121
aI175
aI399
aI551
aI1889
aI2283
aI4343
aI3633
aI653
aI3267
aI101901
aI162157
asS'd'
I16420
sa(dp32840
S'a'
I50961
sS's'
I18
sS'm'
(lp32841
I1
aI3
aI3
aI15
aI23
aI41
aI97
aI251
aI435
aI955
aI69
aI509
aI8001
aI11783
aI7397
aI7587
aI127089
aI15391
asS'd'
I16421
sa(dp32842
S'a'
I50992
sS's'
I18
sS'm'
(lp32843
I1
aI3
aI5
aI5
aI25
aI25
aI63
aI25
aI203
aI655
aI2039
aI2545
aI5405
aI1377
aI30543
aI65531
aI122825
aI6853
asS'd'
I16422
sa(dp32844
S'a'
I50997
sS's'
I18
sS'm'
(lp32845
I1
aI1
aI7
aI9
aI9
aI43
aI73
aI195
aI465
aI497
aI1085
aI3821
aI7115
aI7513
aI21913
aI32499
aI96467
aI181905
asS'd'
I16423
sa(dp32846
S'a'
I51001
sS's'
I18
sS'm'
(lp32847
I1
aI1
aI3
aI7
aI29
aI29
aI63
aI131
aI409
aI423
aI687
aI2549
aI7367
aI6867
aI2685
aI29137
aI61845
aI194409
asS'd'
I16424
sa(dp32848
S'a'
I51004
sS's'
I18
sS'm'
(lp32849
I1
aI3
aI7
aI7
aI3
aI23
aI47
aI31
aI121
aI629
aI1771
aI2387
aI861
aI2269
aI29565
aI19599
aI18051
aI121531
asS'd'
I16425
sa(dp32850
S'a'
I51033
sS's'
I18
sS'm'
(lp32851
I1
aI1
aI1
aI3
aI21
aI13
aI109
aI105
aI163
aI433
aI521
aI3467
aI6225
aI3705
aI30605
aI3265
aI119313
aI2535
asS'd'
I16426
sa(dp32852
S'a'
I51045
sS's'
I18
sS'm'
(lp32853
I1
aI3
aI5
aI3
aI25
aI25
aI89
aI177
aI415
aI67
aI361
aI3317
aI6411
aI4857
aI23249
aI41959
aI59931
aI35797
asS'd'
I16427
sa(dp32854
S'a'
I51052
sS's'
I18
sS'm'
(lp32855
I1
aI3
aI7
aI15
aI5
aI37
aI37
aI65
aI71
aI13
aI1621
aI2217
aI3723
aI2113
aI23755
aI46521
aI48091
aI44307
asS'd'
I16428
sa(dp32856
S'a'
I51064
sS's'
I18
sS'm'
(lp32857
I1
aI1
aI5
aI13
aI5
aI5
aI63
aI75
aI61
aI443
aI1663
aI3239
aI7717
aI2623
aI5723
aI42673
aI8519
aI58773
asS'd'
I16429
sa(dp32858
S'a'
I51085
sS's'
I18
sS'm'
(lp32859
I1
aI1
aI3
aI1
aI29
aI13
aI35
aI61
aI391
aI517
aI1007
aI17
aI2519
aI7121
aI20095
aI33449
aI21397
aI103787
asS'd'
I16430
sa(dp32860
S'a'
I51127
sS's'
I18
sS'm'
(lp32861
I1
aI1
aI5
aI1
aI9
aI45
aI49
aI151
aI39
aI347
aI1821
aI2687
aI1551
aI12117
aI29429
aI40963
aI77795
aI20941
asS'd'
I16431
sa(dp32862
S'a'
I51139
sS's'
I18
sS'm'
(lp32863
I1
aI1
aI5
aI3
aI13
aI49
aI107
aI105
aI201
aI601
aI1335
aI2389
aI6837
aI11123
aI22985
aI62705
aI59057
aI128333
asS'd'
I16432
sa(dp32864
S'a'
I51153
sS's'
I18
sS'm'
(lp32865
I1
aI3
aI3
aI13
aI5
aI3
aI21
aI103
aI481
aI621
aI2037
aI2963
aI425
aI4685
aI27475
aI24363
aI116419
aI171743
asS'd'
I16433
sa(dp32866
S'a'
I51181
sS's'
I18
sS'm'
(lp32867
I1
aI3
aI3
aI5
aI27
aI27
aI5
aI91
aI245
aI421
aI795
aI1869
aI6455
aI4463
aI23467
aI24039
aI69681
aI262127
asS'd'
I16434
sa(dp32868
S'a'
I51182
sS's'
I18
sS'm'
(lp32869
I1
aI3
aI1
aI11
aI11
aI5
aI31
aI253
aI469
aI593
aI877
aI2041
aI4615
aI1541
aI11823
aI58525
aI128689
aI95985
asS'd'
I16435
sa(dp32870
S'a'
I51194
sS's'
I18
sS'm'
(lp32871
I1
aI3
aI5
aI13
aI31
aI35
aI115
aI73
aI97
aI213
aI1499
aI1371
aI4239
aI7897
aI3987
aI4833
aI115043
aI222743
asS'd'
I16436
sa(dp32872
S'a'
I51196
sS's'
I18
sS'm'
(lp32873
I1
aI3
aI3
aI5
aI31
aI15
aI17
aI233
aI137
aI521
aI1721
aI1913
aI1881
aI13457
aI10733
aI61527
aI19825
aI192601
asS'd'
I16437
sa(dp32874
S'a'
I51200
sS's'
I18
sS'm'
(lp32875
I1
aI3
aI1
aI3
aI11
aI21
aI81
aI77
aI377
aI915
aI321
aI3925
aI867
aI5491
aI4707
aI37307
aI52141
aI29155
asS'd'
I16438
sa(dp32876
S'a'
I51239
sS's'
I18
sS'm'
(lp32877
I1
aI1
aI5
aI1
aI13
aI51
aI97
aI161
aI295
aI159
aI1717
aI3817
aI4687
aI1907
aI2655
aI60577
aI73867
aI161851
asS'd'
I16439
sa(dp32878
S'a'
I51240
sS's'
I18
sS'm'
(lp32879
I1
aI1
aI3
aI13
aI21
aI31
aI33
aI145
aI375
aI377
aI1429
aI1981
aI4851
aI9047
aI2685
aI49037
aI67399
aI124243
asS'd'
I16440
sa(dp32880
S'a'
I51254
sS's'
I18
sS'm'
(lp32881
I1
aI1
aI7
aI13
aI1
aI55
aI65
aI147
aI471
aI277
aI1585
aI3949
aI1885
aI1635
aI15687
aI46367
aI120931
aI192097
asS'd'
I16441
sa(dp32882
S'a'
I51260
sS's'
I18
sS'm'
(lp32883
I1
aI1
aI7
aI9
aI13
aI27
aI33
aI41
aI377
aI863
aI1297
aI181
aI2685
aI11285
aI3961
aI63201
aI59757
aI70231
asS'd'
I16442
sa(dp32884
S'a'
I51342
sS's'
I18
sS'm'
(lp32885
I1
aI1
aI3
aI15
aI29
aI3
aI79
aI25
aI147
aI683
aI1563
aI805
aI5891
aI3355
aI20113
aI48261
aI38195
aI14589
asS'd'
I16443
sa(dp32886
S'a'
I51370
sS's'
I18
sS'm'
(lp32887
I1
aI1
aI7
aI11
aI9
aI15
aI21
aI33
aI47
aI923
aI1291
aI4001
aI203
aI305
aI21575
aI41915
aI74769
aI114921
asS'd'
I16444
sa(dp32888
S'a'
I51372
sS's'
I18
sS'm'
(lp32889
I1
aI3
aI7
aI11
aI15
aI1
aI75
aI173
aI473
aI493
aI291
aI811
aI931
aI10731
aI9855
aI57891
aI81575
aI250565
asS'd'
I16445
sa(dp32890
S'a'
I51377
sS's'
I18
sS'm'
(lp32891
I1
aI3
aI3
aI3
aI31
aI31
aI71
aI141
aI389
aI661
aI71
aI3245
aI6827
aI9219
aI26607
aI50511
aI94783
aI130785
asS'd'
I16446
sa(dp32892
S'a'
I51398
sS's'
I18
sS'm'
(lp32893
I1
aI1
aI7
aI3
aI9
aI15
aI67
aI141
aI293
aI779
aI3
aI3311
aI7063
aI13709
aI29715
aI55227
aI11043
aI150343
asS'd'
I16447
sa(dp32894
S'a'
I51402
sS's'
I18
sS'm'
(lp32895
I1
aI3
aI3
aI3
aI21
aI9
aI45
aI111
aI207
aI715
aI345
aI1345
aI3079
aI3851
aI23709
aI33919
aI108213
aI25353
asS'd'
I16448
sa(dp32896
S'a'
I51407
sS's'
I18
sS'm'
(lp32897
I1
aI3
aI5
aI13
aI7
aI3
aI35
aI95
aI397
aI397
aI1159
aI2759
aI5233
aI16237
aI12469
aI29543
aI39133
aI64429
asS'd'
I16449
sa(dp32898
S'a'
I51410
sS's'
I18
sS'm'
(lp32899
I1
aI1
aI5
aI13
aI19
aI35
aI9
aI7
aI153
aI843
aI2025
aI1379
aI3361
aI15889
aI7411
aI26399
aI106295
aI19851
asS'd'
I16450
sa(dp32900
S'a'
I51412
sS's'
I18
sS'm'
(lp32901
I1
aI3
aI1
aI1
aI21
aI15
aI71
aI143
aI279
aI431
aI487
aI967
aI4445
aI11969
aI16671
aI48891
aI59605
aI230607
asS'd'
I16451
sa(dp32902
S'a'
I51419
sS's'
I18
sS'm'
(lp32903
I1
aI3
aI5
aI1
aI25
aI63
aI23
aI143
aI221
aI805
aI377
aI1441
aI1971
aI1985
aI10055
aI35991
aI115873
aI223455
asS'd'
I16452
sa(dp32904
S'a'
I51428
sS's'
I18
sS'm'
(lp32905
I1
aI3
aI1
aI7
aI21
aI57
aI1
aI185
aI117
aI75
aI1623
aI3805
aI2385
aI10245
aI29009
aI59149
aI130219
aI114763
asS'd'
I16453
sa(dp32906
S'a'
I51431
sS's'
I18
sS'm'
(lp32907
I1
aI3
aI1
aI13
aI19
aI13
aI105
aI241
aI47
aI597
aI1725
aI2579
aI3785
aI1667
aI12427
aI62623
aI60883
aI189977
asS'd'
I16454
sa(dp32908
S'a'
I51460
sS's'
I18
sS'm'
(lp32909
I1
aI3
aI5
aI13
aI1
aI55
aI59
aI133
aI263
aI415
aI1013
aI139
aI3037
aI13661
aI15303
aI27279
aI84095
aI184807
asS'd'
I16455
sa(dp32910
S'a'
I51475
sS's'
I18
sS'm'
(lp32911
I1
aI3
aI1
aI11
aI7
aI45
aI3
aI179
aI315
aI639
aI875
aI3299
aI5221
aI8463
aI17507
aI59673
aI73865
aI98867
asS'd'
I16456
sa(dp32912
S'a'
I51477
sS's'
I18
sS'm'
(lp32913
I1
aI3
aI7
aI7
aI3
aI3
aI53
aI233
aI219
aI519
aI585
aI3029
aI3623
aI9559
aI2251
aI43735
aI121513
aI208007
asS'd'
I16457
sa(dp32914
S'a'
I51478
sS's'
I18
sS'm'
(lp32915
I1
aI3
aI1
aI3
aI17
aI47
aI47
aI145
aI445
aI541
aI163
aI2653
aI165
aI7213
aI3311
aI57335
aI43967
aI191841
asS'd'
I16458
sa(dp32916
S'a'
I51482
sS's'
I18
sS'm'
(lp32917
I1
aI3
aI5
aI5
aI7
aI47
aI39
aI119
aI13
aI727
aI887
aI3743
aI3807
aI15267
aI3977
aI52833
aI14851
aI61851
asS'd'
I16459
sa(dp32918
S'a'
I51511
sS's'
I18
sS'm'
(lp32919
I1
aI1
aI3
aI5
aI9
aI9
aI107
aI119
aI501
aI723
aI1965
aI3093
aI6947
aI1783
aI11287
aI24243
aI14005
aI106677
asS'd'
I16460
sa(dp32920
S'a'
I51518
sS's'
I18
sS'm'
(lp32921
I1
aI3
aI3
aI11
aI11
aI29
aI85
aI243
aI359
aI195
aI221
aI1767
aI6969
aI15275
aI20853
aI26921
aI40903
aI29849
asS'd'
I16461
sa(dp32922
S'a'
I51568
sS's'
I18
sS'm'
(lp32923
I1
aI3
aI1
aI5
aI31
aI3
aI11
aI247
aI371
aI339
aI1263
aI119
aI791
aI7425
aI18879
aI11333
aI34359
aI178929
asS'd'
I16462
sa(dp32924
S'a'
I51607
sS's'
I18
sS'm'
(lp32925
I1
aI1
aI3
aI7
aI23
aI15
aI121
aI203
aI441
aI499
aI779
aI1971
aI339
aI8737
aI6859
aI32417
aI97073
aI256143
asS'd'
I16463
sa(dp32926
S'a'
I51618
sS's'
I18
sS'm'
(lp32927
I1
aI1
aI5
aI11
aI3
aI23
aI25
aI51
aI407
aI677
aI97
aI281
aI427
aI3671
aI7939
aI54485
aI3967
aI210199
asS'd'
I16464
sa(dp32928
S'a'
I51630
sS's'
I18
sS'm'
(lp32929
I1
aI1
aI5
aI9
aI5
aI45
aI23
aI171
aI255
aI967
aI313
aI3881
aI6039
aI10545
aI3591
aI51985
aI37145
aI99291
asS'd'
I16465
sa(dp32930
S'a'
I51655
sS's'
I18
sS'm'
(lp32931
I1
aI1
aI5
aI7
aI3
aI13
aI55
aI147
aI83
aI369
aI1707
aI1919
aI491
aI11507
aI29559
aI29169
aI65897
aI80587
asS'd'
I16466
sa(dp32932
S'a'
I51673
sS's'
I18
sS'm'
(lp32933
I1
aI1
aI3
aI13
aI31
aI41
aI41
aI237
aI245
aI109
aI969
aI1797
aI8169
aI6351
aI3657
aI9655
aI109201
aI245117
asS'd'
I16467
sa(dp32934
S'a'
I51683
sS's'
I18
sS'm'
(lp32935
I1
aI3
aI7
aI5
aI17
aI23
aI17
aI219
aI473
aI1
aI865
aI1949
aI7589
aI10107
aI3035
aI19903
aI79579
aI138695
asS'd'
I16468
sa(dp32936
S'a'
I51709
sS's'
I18
sS'm'
(lp32937
I1
aI3
aI1
aI15
aI13
aI57
aI109
aI117
aI277
aI773
aI31
aI3807
aI7615
aI2453
aI22655
aI51513
aI108367
aI248473
asS'd'
I16469
sa(dp32938
S'a'
I51716
sS's'
I18
sS'm'
(lp32939
I1
aI1
aI7
aI11
aI17
aI59
aI27
aI167
aI63
aI931
aI13
aI3721
aI1789
aI7621
aI31093
aI27541
aI37283
aI35193
asS'd'
I16470
sa(dp32940
S'a'
I51774
sS's'
I18
sS'm'
(lp32941
I1
aI3
aI3
aI11
aI13
aI9
aI7
aI85
aI45
aI151
aI1865
aI4049
aI4433
aI9517
aI231
aI30733
aI93701
aI126585
asS'd'
I16471
sa(dp32942
S'a'
I51829
sS's'
I18
sS'm'
(lp32943
I1
aI3
aI1
aI13
aI21
aI33
aI19
aI87
aI77
aI425
aI351
aI1163
aI7453
aI12567
aI24615
aI35563
aI127643
aI28625
asS'd'
I16472
sa(dp32944
S'a'
I51830
sS's'
I18
sS'm'
(lp32945
I1
aI1
aI3
aI11
aI7
aI35
aI27
aI47
aI465
aI595
aI985
aI573
aI2541
aI7841
aI14749
aI43761
aI26699
aI135895
asS'd'
I16473
sa(dp32946
S'a'
I51839
sS's'
I18
sS'm'
(lp32947
I1
aI1
aI1
aI7
aI25
aI51
aI93
aI237
aI355
aI575
aI1
aI443
aI5697
aI1997
aI28801
aI11621
aI62531
aI88449
asS'd'
I16474
sa(dp32948
S'a'
I51849
sS's'
I18
sS'm'
(lp32949
I1
aI1
aI1
aI9
aI23
aI35
aI23
aI91
aI161
aI601
aI1401
aI2187
aI6283
aI10711
aI21277
aI47771
aI12589
aI176625
asS'd'
I16475
sa(dp32950
S'a'
I51869
sS's'
I18
sS'm'
(lp32951
I1
aI3
aI3
aI5
aI13
aI23
aI33
aI65
aI213
aI835
aI539
aI2487
aI273
aI6113
aI18327
aI52493
aI17571
aI160909
asS'd'
I16476
sa(dp32952
S'a'
I51886
sS's'
I18
sS'm'
(lp32953
I1
aI3
aI1
aI5
aI25
aI9
aI117
aI201
aI457
aI331
aI1455
aI439
aI4891
aI5515
aI21701
aI9343
aI29085
aI120299
asS'd'
I16477
sa(dp32954
S'a'
I51905
sS's'
I18
sS'm'
(lp32955
I1
aI1
aI1
aI5
aI29
aI7
aI43
aI125
aI155
aI43
aI1949
aI2901
aI7293
aI13683
aI18289
aI16873
aI27899
aI168631
asS'd'
I16478
sa(dp32956
S'a'
I51906
sS's'
I18
sS'm'
(lp32957
I1
aI1
aI7
aI3
aI29
aI33
aI53
aI137
aI301
aI27
aI1101
aI837
aI5843
aI13167
aI6073
aI49083
aI120031
aI45065
asS'd'
I16479
sa(dp32958
S'a'
I51915
sS's'
I18
sS'm'
(lp32959
I1
aI3
aI5
aI1
aI31
aI47
aI127
aI185
aI279
aI603
aI1699
aI1693
aI3263
aI9055
aI3525
aI46065
aI79305
aI19949
asS'd'
I16480
sa(dp32960
S'a'
I51948
sS's'
I18
sS'm'
(lp32961
I1
aI3
aI7
aI9
aI17
aI47
aI25
aI191
aI369
aI877
aI1477
aI3041
aI5123
aI1393
aI5061
aI1755
aI61051
aI55299
asS'd'
I16481
sa(dp32962
S'a'
I51988
sS's'
I18
sS'm'
(lp32963
I1
aI1
aI5
aI13
aI13
aI7
aI89
aI141
aI251
aI321
aI1515
aI2677
aI5103
aI12901
aI29875
aI165
aI15073
aI47795
asS'd'
I16482
sa(dp32964
S'a'
I51998
sS's'
I18
sS'm'
(lp32965
I1
aI3
aI1
aI7
aI1
aI37
aI55
aI173
aI191
aI981
aI685
aI4003
aI6319
aI3037
aI30637
aI11955
aI81015
aI89239
asS'd'
I16483
sa(dp32966
S'a'
I52004
sS's'
I18
sS'm'
(lp32967
I1
aI3
aI5
aI5
aI1
aI5
aI55
aI251
aI229
aI153
aI425
aI2793
aI6779
aI15797
aI10087
aI7573
aI121789
aI115479
asS'd'
I16484
sa(dp32968
S'a'
I52026
sS's'
I18
sS'm'
(lp32969
I1
aI3
aI7
aI13
aI23
aI39
aI23
aI21
aI55
aI543
aI1539
aI3055
aI7825
aI2865
aI8967
aI56719
aI117219
aI142137
asS'd'
I16485
sa(dp32970
S'a'
I52036
sS's'
I18
sS'm'
(lp32971
I1
aI1
aI5
aI13
aI1
aI29
aI81
aI11
aI509
aI61
aI1681
aI1911
aI829
aI10583
aI7105
aI42047
aI128579
aI48891
asS'd'
I16486
sa(dp32972
S'a'
I52082
sS's'
I18
sS'm'
(lp32973
I1
aI1
aI3
aI13
aI7
aI27
aI81
aI89
aI129
aI239
aI309
aI1353
aI5265
aI12255
aI29391
aI1659
aI114857
aI43551
asS'd'
I16487
sa(dp32974
S'a'
I52127
sS's'
I18
sS'm'
(lp32975
I1
aI1
aI7
aI13
aI23
aI15
aI9
aI93
aI35
aI839
aI549
aI1793
aI4693
aI13295
aI10603
aI18179
aI33141
aI239555
asS'd'
I16488
sa(dp32976
S'a'
I52145
sS's'
I18
sS'm'
(lp32977
I1
aI3
aI1
aI7
aI17
aI5
aI11
aI193
aI143
aI579
aI1199
aI1239
aI4503
aI15855
aI23345
aI34789
aI59427
aI235319
asS'd'
I16489
sa(dp32978
S'a'
I52157
sS's'
I18
sS'm'
(lp32979
I1
aI3
aI5
aI3
aI25
aI63
aI11
aI203
aI415
aI135
aI261
aI1843
aI3409
aI4605
aI15213
aI28537
aI75787
aI100007
asS'd'
I16490
sa(dp32980
S'a'
I52165
sS's'
I18
sS'm'
(lp32981
I1
aI1
aI1
aI5
aI3
aI37
aI29
aI157
aI213
aI235
aI959
aI1087
aI2843
aI10265
aI28233
aI14281
aI25867
aI204031
asS'd'
I16491
sa(dp32982
S'a'
I52193
sS's'
I18
sS'm'
(lp32983
I1
aI3
aI3
aI7
aI19
aI49
aI55
aI111
aI253
aI823
aI533
aI941
aI2509
aI5595
aI9267
aI28457
aI84301
aI165385
asS'd'
I16492
sa(dp32984
S'a'
I52211
sS's'
I18
sS'm'
(lp32985
I1
aI3
aI5
aI9
aI5
aI59
aI13
aI85
aI339
aI889
aI597
aI3517
aI7001
aI5525
aI25451
aI13855
aI19033
aI182677
asS'd'
I16493
sa(dp32986
S'a'
I52213
sS's'
I18
sS'm'
(lp32987
I1
aI3
aI1
aI15
aI23
aI29
aI31
aI105
aI353
aI825
aI1977
aI2013
aI131
aI1969
aI427
aI16465
aI90817
aI257931
asS'd'
I16494
sa(dp32988
S'a'
I52218
sS's'
I18
sS'm'
(lp32989
I1
aI3
aI1
aI3
aI29
aI9
aI109
aI243
aI321
aI15
aI1479
aI787
aI4667
aI13925
aI7347
aI7977
aI105585
aI143959
asS'd'
I16495
sa(dp32990
S'a'
I52220
sS's'
I18
sS'm'
(lp32991
I1
aI1
aI7
aI15
aI5
aI45
aI11
aI95
aI215
aI719
aI827
aI77
aI7263
aI5705
aI26971
aI26845
aI82127
aI2849
asS'd'
I16496
sa(dp32992
S'a'
I52238
sS's'
I18
sS'm'
(lp32993
I1
aI1
aI7
aI9
aI5
aI43
aI103
aI133
aI203
aI127
aI2021
aI3609
aI6971
aI13447
aI27089
aI62287
aI104391
aI147901
asS'd'
I16497
sa(dp32994
S'a'
I52240
sS's'
I18
sS'm'
(lp32995
I1
aI3
aI1
aI3
aI3
aI41
aI61
aI101
aI381
aI985
aI1795
aI2465
aI2899
aI13517
aI23953
aI38831
aI43569
aI128643
asS'd'
I16498
sa(dp32996
S'a'
I52243
sS's'
I18
sS'm'
(lp32997
I1
aI3
aI5
aI5
aI9
aI49
aI13
aI7
aI215
aI85
aI1203
aI647
aI6627
aI1861
aI17901
aI40203
aI13007
aI84975
asS'd'
I16499
sa(dp32998
S'a'
I52250
sS's'
I18
sS'm'
(lp32999
I1
aI3
aI5
aI15
aI31
aI13
aI55
aI89
aI397
aI641
aI1599
aI3379
aI3401
aI4173
aI5757
aI42945
aI59269
aI106891
asS'd'
I16500
sa(dp33000
S'a'
I52279
sS's'
I18
sS'm'
(lp33001
I1
aI1
aI5
aI13
aI25
aI17
aI45
aI27
aI151
aI725
aI819
aI581
aI3675
aI3983
aI9499
aI47511
aI128039
aI56825
asS'd'
I16501
sa(dp33002
S'a'
I52285
sS's'
I18
sS'm'
(lp33003
I1
aI1
aI5
aI1
aI7
aI11
aI65
aI63
aI301
aI927
aI409
aI3729
aI7227
aI12849
aI17855
aI36527
aI2907
aI66819
asS'd'
I16502
sa(dp33004
S'a'
I52306
sS's'
I18
sS'm'
(lp33005
I1
aI3
aI3
aI3
aI29
aI35
aI39
aI1
aI349
aI429
aI805
aI3639
aI3909
aI4211
aI10393
aI36223
aI72543
aI136375
asS'd'
I16503
sa(dp33006
S'a'
I52370
sS's'
I18
sS'm'
(lp33007
I1
aI1
aI1
aI9
aI23
aI27
aI25
aI213
aI195
aI455
aI883
aI3357
aI7277
aI9061
aI14103
aI6005
aI35031
aI72703
asS'd'
I16504
sa(dp33008
S'a'
I52432
sS's'
I18
sS'm'
(lp33009
I1
aI1
aI3
aI3
aI11
aI17
aI19
aI181
aI25
aI775
aI897
aI3809
aI2031
aI13017
aI7505
aI37469
aI107335
aI174309
asS'd'
I16505
sa(dp33010
S'a'
I52457
sS's'
I18
sS'm'
(lp33011
I1
aI1
aI7
aI13
aI7
aI1
aI57
aI27
aI159
aI465
aI533
aI3409
aI3863
aI14001
aI8011
aI25873
aI14971
aI67243
asS'd'
I16506
sa(dp33012
S'a'
I52466
sS's'
I18
sS'm'
(lp33013
I1
aI1
aI3
aI11
aI19
aI5
aI11
aI19
aI75
aI489
aI1879
aI1539
aI6563
aI4729
aI15605
aI35203
aI47993
aI110139
asS'd'
I16507
sa(dp33014
S'a'
I52514
sS's'
I18
sS'm'
(lp33015
I1
aI3
aI5
aI9
aI23
aI17
aI67
aI89
aI379
aI849
aI1667
aI955
aI1537
aI11781
aI9791
aI46959
aI79481
aI237335
asS'd'
I16508
sa(dp33016
S'a'
I52557
sS's'
I18
sS'm'
(lp33017
I1
aI1
aI7
aI9
aI3
aI31
aI127
aI145
aI29
aI35
aI463
aI4009
aI4427
aI16215
aI12093
aI50085
aI51259
aI45091
asS'd'
I16509
sa(dp33018
S'a'
I52560
sS's'
I18
sS'm'
(lp33019
I1
aI1
aI5
aI9
aI25
aI3
aI1
aI131
aI221
aI951
aI117
aI3227
aI797
aI7617
aI13243
aI50139
aI26737
aI105875
asS'd'
I16510
sa(dp33020
S'a'
I52585
sS's'
I18
sS'm'
(lp33021
I1
aI1
aI5
aI3
aI11
aI5
aI7
aI155
aI211
aI865
aI27
aI3943
aI7923
aI9973
aI23233
aI37399
aI89951
aI106555
asS'd'
I16511
sa(dp33022
S'a'
I52586
sS's'
I18
sS'm'
(lp33023
I1
aI1
aI3
aI13
aI21
aI61
aI77
aI121
aI227
aI527
aI1641
aI3535
aI3801
aI7221
aI6423
aI9179
aI114935
aI33373
asS'd'
I16512
sa(dp33024
S'a'
I52603
sS's'
I18
sS'm'
(lp33025
I1
aI3
aI7
aI5
aI31
aI7
aI123
aI159
aI367
aI369
aI1905
aI1689
aI6773
aI675
aI30289
aI54149
aI71469
aI232835
asS'd'
I16513
sa(dp33026
S'a'
I52650
sS's'
I18
sS'm'
(lp33027
I1
aI1
aI1
aI13
aI9
aI23
aI83
aI237
aI251
aI941
aI781
aI1489
aI6037
aI6001
aI15909
aI50527
aI60143
aI238499
asS'd'
I16514
sa(dp33028
S'a'
I52672
sS's'
I18
sS'm'
(lp33029
I1
aI3
aI1
aI11
aI5
aI7
aI103
aI43
aI413
aI247
aI535
aI2107
aI1801
aI16381
aI32529
aI2355
aI39143
aI71281
asS'd'
I16515
sa(dp33030
S'a'
I52681
sS's'
I18
sS'm'
(lp33031
I1
aI1
aI7
aI5
aI5
aI31
aI19
aI11
aI191
aI643
aI923
aI1661
aI2215
aI11817
aI23937
aI62907
aI128301
aI21459
asS'd'
I16516
sa(dp33032
S'a'
I52692
sS's'
I18
sS'm'
(lp33033
I1
aI3
aI7
aI15
aI1
aI51
aI123
aI61
aI99
aI751
aI1819
aI1191
aI3865
aI8765
aI7131
aI33965
aI55697
aI87059
asS'd'
I16517
sa(dp33034
S'a'
I52735
sS's'
I18
sS'm'
(lp33035
I1
aI1
aI5
aI13
aI5
aI57
aI41
aI103
aI135
aI207
aI517
aI3995
aI2537
aI15705
aI9123
aI26193
aI30653
aI190549
asS'd'
I16518
sa(dp33036
S'a'
I52760
sS's'
I18
sS'm'
(lp33037
I1
aI1
aI3
aI1
aI23
aI5
aI109
aI209
aI407
aI143
aI943
aI2325
aI8087
aI559
aI23675
aI31815
aI43805
aI67497
asS'd'
I16519
sa(dp33038
S'a'
I52772
sS's'
I18
sS'm'
(lp33039
I1
aI1
aI5
aI9
aI25
aI33
aI13
aI21
aI93
aI357
aI1551
aI739
aI5595
aI16285
aI30761
aI54075
aI75505
aI177333
asS'd'
I16520
sa(dp33040
S'a'
I52781
sS's'
I18
sS'm'
(lp33041
I1
aI3
aI5
aI11
aI17
aI59
aI31
aI249
aI95
aI561
aI1849
aI4061
aI5577
aI2607
aI30083
aI59033
aI56697
aI89761
asS'd'
I16521
sa(dp33042
S'a'
I52793
sS's'
I18
sS'm'
(lp33043
I1
aI1
aI3
aI13
aI27
aI57
aI9
aI17
aI323
aI813
aI1197
aI2775
aI3443
aI9523
aI24509
aI12129
aI89697
aI169043
asS'd'
I16522
sa(dp33044
S'a'
I52796
sS's'
I18
sS'm'
(lp33045
I1
aI3
aI5
aI13
aI17
aI51
aI91
aI1
aI105
aI345
aI829
aI1365
aI2755
aI7197
aI26655
aI1303
aI52223
aI133893
asS'd'
I16523
sa(dp33046
S'a'
I52807
sS's'
I18
sS'm'
(lp33047
I1
aI1
aI7
aI1
aI31
aI3
aI51
aI21
aI327
aI851
aI153
aI3329
aI3393
aI8489
aI5879
aI25035
aI124403
aI172657
asS'd'
I16524
sa(dp33048
S'a'
I52821
sS's'
I18
sS'm'
(lp33049
I1
aI1
aI1
aI5
aI3
aI21
aI61
aI29
aI99
aI343
aI621
aI3163
aI3763
aI9347
aI7691
aI34667
aI24555
aI125819
asS'd'
I16525
sa(dp33050
S'a'
I52849
sS's'
I18
sS'm'
(lp33051
I1
aI1
aI1
aI1
aI3
aI17
aI83
aI191
aI83
aI315
aI1091
aI589
aI5081
aI4611
aI15521
aI25791
aI9103
aI13741
asS'd'
I16526
sa(dp33052
S'a'
I52866
sS's'
I18
sS'm'
(lp33053
I1
aI3
aI5
aI3
aI7
aI53
aI9
aI227
aI399
aI857
aI673
aI3027
aI5045
aI5573
aI7467
aI4813
aI99659
aI142845
asS'd'
I16527
sa(dp33054
S'a'
I52868
sS's'
I18
sS'm'
(lp33055
I1
aI1
aI5
aI11
aI23
aI37
aI71
aI151
aI279
aI879
aI601
aI2391
aI7091
aI12669
aI10203
aI11747
aI9613
aI248261
asS'd'
I16528
sa(dp33056
S'a'
I52872
sS's'
I18
sS'm'
(lp33057
I1
aI1
aI3
aI3
aI29
aI25
aI33
aI25
aI1
aI683
aI1475
aI457
aI3641
aI14219
aI20105
aI21449
aI6903
aI43819
asS'd'
I16529
sa(dp33058
S'a'
I52875
sS's'
I18
sS'm'
(lp33059
I1
aI1
aI7
aI7
aI5
aI1
aI73
aI79
aI357
aI971
aI699
aI1105
aI1683
aI1687
aI32677
aI62467
aI47671
aI88149
asS'd'
I16530
sa(dp33060
S'a'
I52886
sS's'
I18
sS'm'
(lp33061
I1
aI3
aI1
aI15
aI23
aI13
aI93
aI75
aI307
aI81
aI1607
aI1333
aI6969
aI7747
aI27135
aI58941
aI26355
aI5565
asS'd'
I16531
sa(dp33062
S'a'
I52890
sS's'
I18
sS'm'
(lp33063
I1
aI1
aI7
aI3
aI27
aI35
aI85
aI195
aI421
aI999
aI1721
aI2029
aI283
aI13995
aI21649
aI7519
aI73357
aI193171
asS'd'
I16532
sa(dp33064
S'a'
I52896
sS's'
I18
sS'm'
(lp33065
I1
aI3
aI7
aI15
aI17
aI63
aI21
aI187
aI475
aI671
aI1681
aI2731
aI8169
aI3327
aI19789
aI53295
aI43219
aI6949
asS'd'
I16533
sa(dp33066
S'a'
I52925
sS's'
I18
sS'm'
(lp33067
I1
aI1
aI7
aI13
aI9
aI33
aI47
aI115
aI295
aI123
aI1293
aI1627
aI4261
aI4503
aI15925
aI16521
aI81759
aI247089
asS'd'
I16534
sa(dp33068
S'a'
I52931
sS's'
I18
sS'm'
(lp33069
I1
aI3
aI1
aI13
aI11
aI15
aI83
aI129
aI409
aI465
aI873
aI1333
aI7939
aI973
aI2753
aI33727
aI128975
aI43333
asS'd'
I16535
sa(dp33070
S'a'
I52937
sS's'
I18
sS'm'
(lp33071
I1
aI3
aI3
aI3
aI27
aI59
aI1
aI137
aI145
aI29
aI1189
aI2615
aI3249
aI11197
aI17165
aI32313
aI14065
aI44199
asS'd'
I16536
sa(dp33072
S'a'
I52938
sS's'
I18
sS'm'
(lp33073
I1
aI3
aI3
aI11
aI17
aI49
aI107
aI111
aI45
aI963
aI1129
aI1775
aI7671
aI1495
aI14531
aI49743
aI63321
aI159841
asS'd'
I16537
sa(dp33074
S'a'
I52952
sS's'
I18
sS'm'
(lp33075
I1
aI1
aI5
aI9
aI5
aI11
aI79
aI99
aI155
aI347
aI1777
aI3793
aI1765
aI2319
aI3135
aI30237
aI100845
aI52689
asS'd'
I16538
sa(dp33076
S'a'
I52979
sS's'
I18
sS'm'
(lp33077
I1
aI3
aI7
aI13
aI21
aI57
aI71
aI207
aI149
aI161
aI265
aI991
aI6967
aI8905
aI21581
aI13921
aI79201
aI95667
asS'd'
I16539
sa(dp33078
S'a'
I53013
sS's'
I18
sS'm'
(lp33079
I1
aI3
aI5
aI15
aI13
aI53
aI95
aI81
aI443
aI161
aI1071
aI2749
aI6637
aI837
aI15015
aI62397
aI33295
aI112005
asS'd'
I16540
sa(dp33080
S'a'
I53023
sS's'
I18
sS'm'
(lp33081
I1
aI3
aI1
aI1
aI31
aI25
aI37
aI111
aI79
aI293
aI249
aI1523
aI1509
aI1993
aI17167
aI62939
aI118281
aI62699
asS'd'
I16541
sa(dp33082
S'a'
I53027
sS's'
I18
sS'm'
(lp33083
I1
aI3
aI7
aI7
aI5
aI33
aI61
aI179
aI265
aI405
aI287
aI1899
aI437
aI1609
aI19617
aI41093
aI36341
aI176593
asS'd'
I16542
sa(dp33084
S'a'
I53033
sS's'
I18
sS'm'
(lp33085
I1
aI1
aI5
aI3
aI9
aI33
aI97
aI45
aI23
aI807
aI1575
aI627
aI7465
aI4805
aI11191
aI35439
aI69433
aI47275
asS'd'
I16543
sa(dp33086
S'a'
I53039
sS's'
I18
sS'm'
(lp33087
I1
aI1
aI1
aI5
aI1
aI51
aI51
aI247
aI453
aI449
aI1487
aI141
aI2501
aI8039
aI14749
aI63733
aI91963
aI232951
asS'd'
I16544
sa(dp33088
S'a'
I53041
sS's'
I18
sS'm'
(lp33089
I1
aI3
aI3
aI13
aI15
aI7
aI81
aI211
aI445
aI15
aI899
aI835
aI5163
aI3403
aI7367
aI29963
aI80413
aI87209
asS'd'
I16545
sa(dp33090
S'a'
I53048
sS's'
I18
sS'm'
(lp33091
I1
aI3
aI1
aI11
aI25
aI43
aI113
aI139
aI381
aI391
aI485
aI1503
aI4195
aI10109
aI13771
aI35865
aI50909
aI224887
asS'd'
I16546
sa(dp33092
S'a'
I53073
sS's'
I18
sS'm'
(lp33093
I1
aI1
aI7
aI9
aI9
aI51
aI21
aI85
aI137
aI765
aI951
aI2453
aI227
aI9177
aI1457
aI47937
aI84203
aI118987
asS'd'
I16547
sa(dp33094
S'a'
I53095
sS's'
I18
sS'm'
(lp33095
I1
aI3
aI5
aI9
aI1
aI1
aI21
aI41
aI133
aI317
aI519
aI2249
aI3453
aI2957
aI2029
aI54737
aI42515
aI176017
asS'd'
I16548
sa(dp33096
S'a'
I53099
sS's'
I18
sS'm'
(lp33097
I1
aI3
aI1
aI7
aI9
aI27
aI79
aI193
aI209
aI281
aI267
aI1267
aI7013
aI13667
aI13331
aI32863
aI5289
aI15077
asS'd'
I16549
sa(dp33098
S'a'
I53120
sS's'
I18
sS'm'
(lp33099
I1
aI1
aI5
aI11
aI29
aI3
aI3
aI177
aI75
aI485
aI1735
aI3955
aI4349
aI7893
aI13075
aI58735
aI8629
aI78143
asS'd'
I16550
sa(dp33100
S'a'
I53154
sS's'
I18
sS'm'
(lp33101
I1
aI1
aI7
aI13
aI3
aI15
aI15
aI77
aI7
aI843
aI1561
aI461
aI6817
aI4363
aI7861
aI45697
aI115663
aI93789
asS'd'
I16551
sa(dp33102
S'a'
I53168
sS's'
I18
sS'm'
(lp33103
I1
aI1
aI1
aI7
aI5
aI35
aI83
aI213
aI229
aI383
aI747
aI337
aI2589
aI683
aI18575
aI42415
aI74889
aI227331
asS'd'
I16552
sa(dp33104
S'a'
I53192
sS's'
I18
sS'm'
(lp33105
I1
aI3
aI5
aI3
aI17
aI35
aI57
aI213
aI247
aI273
aI689
aI1889
aI5667
aI357
aI4267
aI11611
aI20621
aI159039
asS'd'
I16553
sa(dp33106
S'a'
I53205
sS's'
I18
sS'm'
(lp33107
I1
aI3
aI3
aI15
aI7
aI55
aI25
aI83
aI293
aI939
aI1169
aI2507
aI3939
aI7537
aI2959
aI40231
aI124511
aI181091
asS'd'
I16554
sa(dp33108
S'a'
I53206
sS's'
I18
sS'm'
(lp33109
I1
aI1
aI5
aI9
aI31
aI51
aI67
aI149
aI509
aI695
aI449
aI2761
aI6597
aI4741
aI4205
aI49177
aI45599
aI167807
asS'd'
I16555
sa(dp33110
S'a'
I53234
sS's'
I18
sS'm'
(lp33111
I1
aI1
aI7
aI15
aI3
aI7
aI113
aI71
aI279
aI885
aI251
aI2831
aI855
aI6673
aI6511
aI63861
aI41109
aI177119
asS'd'
I16556
sa(dp33112
S'a'
I53246
sS's'
I18
sS'm'
(lp33113
I1
aI1
aI5
aI3
aI7
aI23
aI125
aI11
aI217
aI1023
aI549
aI529
aI3891
aI10595
aI13751
aI37729
aI113469
aI110549
asS'd'
I16557
sa(dp33114
S'a'
I53253
sS's'
I18
sS'm'
(lp33115
I1
aI1
aI3
aI15
aI21
aI9
aI13
aI63
aI93
aI635
aI659
aI2837
aI6303
aI10083
aI10107
aI27859
aI33891
aI181229
asS'd'
I16558
sa(dp33116
S'a'
I53263
sS's'
I18
sS'm'
(lp33117
I1
aI1
aI7
aI3
aI3
aI43
aI5
aI149
aI353
aI353
aI565
aI2441
aI7113
aI6493
aI30355
aI17887
aI110787
aI187199
asS'd'
I16559
sa(dp33118
S'a'
I53275
sS's'
I18
sS'm'
(lp33119
I1
aI1
aI3
aI7
aI19
aI43
aI99
aI63
aI169
aI743
aI185
aI3817
aI6677
aI5549
aI1609
aI24701
aI98669
aI233701
asS'd'
I16560
sa(dp33120
S'a'
I53313
sS's'
I18
sS'm'
(lp33121
I1
aI1
aI1
aI1
aI21
aI49
aI73
aI169
aI223
aI551
aI553
aI917
aI4705
aI14951
aI14031
aI19753
aI96205
aI131755
asS'd'
I16561
sa(dp33122
S'a'
I53331
sS's'
I18
sS'm'
(lp33123
I1
aI1
aI3
aI3
aI17
aI49
aI51
aI249
aI293
aI921
aI435
aI2915
aI3125
aI3487
aI11417
aI35043
aI29543
aI35933
asS'd'
I16562
sa(dp33124
S'a'
I53349
sS's'
I18
sS'm'
(lp33125
I1
aI1
aI3
aI9
aI23
aI43
aI73
aI151
aI379
aI911
aI671
aI151
aI955
aI11885
aI28795
aI23967
aI117135
aI137985
asS'd'
I16563
sa(dp33126
S'a'
I53367
sS's'
I18
sS'm'
(lp33127
I1
aI1
aI3
aI7
aI29
aI3
aI51
aI231
aI59
aI227
aI443
aI3533
aI7785
aI12535
aI25725
aI7451
aI9391
aI239281
asS'd'
I16564
sa(dp33128
S'a'
I53383
sS's'
I18
sS'm'
(lp33129
I1
aI3
aI3
aI9
aI17
aI37
aI91
aI195
aI5
aI843
aI313
aI1417
aI1207
aI3225
aI15949
aI34023
aI1275
aI221057
asS'd'
I16565
sa(dp33130
S'a'
I53392
sS's'
I18
sS'm'
(lp33131
I1
aI1
aI3
aI15
aI3
aI51
aI111
aI135
aI63
aI495
aI1967
aI2151
aI197
aI10913
aI20705
aI1021
aI68431
aI67119
asS'd'
I16566
sa(dp33132
S'a'
I53404
sS's'
I18
sS'm'
(lp33133
I1
aI1
aI5
aI3
aI7
aI29
aI87
aI219
aI273
aI267
aI1317
aI797
aI6723
aI947
aI29867
aI32571
aI12337
aI234715
asS'd'
I16567
sa(dp33134
S'a'
I53407
sS's'
I18
sS'm'
(lp33135
I1
aI3
aI5
aI15
aI1
aI9
aI91
aI63
aI97
aI107
aI451
aI4025
aI233
aI7803
aI17031
aI7669
aI49417
aI256611
asS'd'
I16568
sa(dp33136
S'a'
I53411
sS's'
I18
sS'm'
(lp33137
I1
aI1
aI7
aI3
aI17
aI17
aI45
aI197
aI227
aI133
aI799
aI411
aI6739
aI8037
aI19553
aI53009
aI25201
aI107625
asS'd'
I16569
sa(dp33138
S'a'
I53417
sS's'
I18
sS'm'
(lp33139
I1
aI3
aI5
aI7
aI3
aI39
aI25
aI95
aI197
aI127
aI45
aI173
aI3305
aI6575
aI19633
aI54919
aI35373
aI59509
asS'd'
I16570
sa(dp33140
S'a'
I53425
sS's'
I18
sS'm'
(lp33141
I1
aI3
aI3
aI3
aI9
aI1
aI107
aI211
aI217
aI715
aI311
aI3641
aI8055
aI1
aI9017
aI29329
aI128467
aI46911
asS'd'
I16571
sa(dp33142
S'a'
I53443
sS's'
I18
sS'm'
(lp33143
I1
aI1
aI1
aI7
aI1
aI13
aI13
aI79
aI259
aI533
aI1761
aI449
aI3363
aI3061
aI26227
aI50407
aI122951
aI261425
asS'd'
I16572
sa(dp33144
S'a'
I53469
sS's'
I18
sS'm'
(lp33145
I1
aI3
aI5
aI5
aI29
aI19
aI41
aI7
aI25
aI203
aI587
aI3321
aI655
aI15877
aI10423
aI41481
aI70325
aI165527
asS'd'
I16573
sa(dp33146
S'a'
I53480
sS's'
I18
sS'm'
(lp33147
I1
aI3
aI5
aI9
aI11
aI45
aI91
aI253
aI7
aI137
aI795
aI2379
aI4527
aI1677
aI5081
aI6523
aI97245
aI3691
asS'd'
I16574
sa(dp33148
S'a'
I53508
sS's'
I18
sS'm'
(lp33149
I1
aI3
aI7
aI9
aI25
aI43
aI125
aI243
aI391
aI785
aI651
aI3245
aI7979
aI14689
aI15443
aI40501
aI5519
aI96551
asS'd'
I16575
sa(dp33150
S'a'
I53535
sS's'
I18
sS'm'
(lp33151
I1
aI1
aI1
aI3
aI25
aI53
aI45
aI159
aI47
aI701
aI1655
aI2019
aI2355
aI11213
aI12403
aI42291
aI44925
aI72689
asS'd'
I16576
sa(dp33152
S'a'
I53536
sS's'
I18
sS'm'
(lp33153
I1
aI1
aI1
aI5
aI21
aI19
aI77
aI31
aI3
aI161
aI149
aI3759
aI6331
aI12311
aI7021
aI1117
aI12459
aI134821
asS'd'
I16577
sa(dp33154
S'a'
I53542
sS's'
I18
sS'm'
(lp33155
I1
aI1
aI3
aI9
aI9
aI59
aI23
aI255
aI437
aI625
aI719
aI3727
aI7157
aI1889
aI31523
aI59127
aI114143
aI174179
asS'd'
I16578
sa(dp33156
S'a'
I53563
sS's'
I18
sS'm'
(lp33157
I1
aI1
aI1
aI15
aI23
aI7
aI47
aI137
aI77
aI519
aI1681
aI1159
aI6121
aI14789
aI21343
aI43101
aI44709
aI179863
asS'd'
I16579
sa(dp33158
S'a'
I53577
sS's'
I18
sS'm'
(lp33159
I1
aI3
aI1
aI3
aI17
aI27
aI103
aI11
aI327
aI735
aI1949
aI3719
aI811
aI2267
aI13187
aI29747
aI98433
aI242021
asS'd'
I16580
sa(dp33160
S'a'
I53591
sS's'
I18
sS'm'
(lp33161
I1
aI1
aI7
aI15
aI15
aI63
aI25
aI203
aI109
aI585
aI409
aI4093
aI6669
aI2381
aI30721
aI58975
aI17235
aI257959
asS'd'
I16581
sa(dp33162
S'a'
I53607
sS's'
I18
sS'm'
(lp33163
I1
aI3
aI3
aI5
aI5
aI19
aI27
aI69
aI69
aI193
aI693
aI1169
aI6321
aI3425
aI9285
aI28019
aI128343
aI185165
asS'd'
I16582
sa(dp33164
S'a'
I53626
sS's'
I18
sS'm'
(lp33165
I1
aI1
aI3
aI9
aI7
aI51
aI113
aI93
aI81
aI385
aI1811
aI2601
aI2065
aI1029
aI24515
aI7199
aI26425
aI174283
asS'd'
I16583
sa(dp33166
S'a'
I53628
sS's'
I18
sS'm'
(lp33167
I1
aI1
aI1
aI5
aI29
aI55
aI93
aI219
aI281
aI887
aI891
aI2763
aI6083
aI9627
aI18559
aI21329
aI73887
aI83699
asS'd'
I16584
sa(dp33168
S'a'
I53638
sS's'
I18
sS'm'
(lp33169
I1
aI1
aI1
aI3
aI21
aI31
aI49
aI173
aI15
aI177
aI1001
aI3453
aI5623
aI14107
aI8837
aI10163
aI26817
aI41947
asS'd'
I16585
sa(dp33170
S'a'
I53675
sS's'
I18
sS'm'
(lp33171
I1
aI1
aI3
aI5
aI27
aI63
aI117
aI49
aI405
aI281
aI981
aI2363
aI1525
aI9685
aI29089
aI16739
aI66509
aI125823
asS'd'
I16586
sa(dp33172
S'a'
I53685
sS's'
I18
sS'm'
(lp33173
I1
aI3
aI7
aI13
aI27
aI29
aI57
aI189
aI345
aI135
aI753
aI463
aI731
aI4823
aI14335
aI33299
aI105229
aI54705
asS'd'
I16587
sa(dp33174
S'a'
I53692
sS's'
I18
sS'm'
(lp33175
I1
aI3
aI1
aI1
aI9
aI43
aI51
aI25
aI371
aI261
aI1409
aI3493
aI2811
aI12915
aI16075
aI62159
aI125945
aI108453
asS'd'
I16588
sa(dp33176
S'a'
I53703
sS's'
I18
sS'm'
(lp33177
I1
aI3
aI7
aI15
aI13
aI33
aI47
aI53
aI263
aI669
aI1383
aI3059
aI4043
aI4777
aI14679
aI2077
aI11019
aI5803
asS'd'
I16589
sa(dp33178
S'a'
I53722
sS's'
I18
sS'm'
(lp33179
I1
aI1
aI3
aI11
aI21
aI7
aI39
aI71
aI215
aI79
aI1849
aI1261
aI45
aI1273
aI27591
aI4653
aI25119
aI30445
asS'd'
I16590
sa(dp33180
S'a'
I53738
sS's'
I18
sS'm'
(lp33181
I1
aI3
aI5
aI9
aI21
aI3
aI17
aI207
aI417
aI777
aI37
aI3349
aI2761
aI4469
aI3457
aI15593
aI87251
aI38601
asS'd'
I16591
sa(dp33182
S'a'
I53748
sS's'
I18
sS'm'
(lp33183
I1
aI1
aI7
aI13
aI29
aI29
aI101
aI103
aI487
aI87
aI1129
aI2497
aI5501
aI4813
aI8623
aI25077
aI50487
aI94053
asS'd'
I16592
sa(dp33184
S'a'
I53752
sS's'
I18
sS'm'
(lp33185
I1
aI3
aI3
aI11
aI7
aI23
aI95
aI245
aI127
aI55
aI431
aI2707
aI4955
aI15871
aI29589
aI60023
aI1921
aI227623
asS'd'
I16593
sa(dp33186
S'a'
I53764
sS's'
I18
sS'm'
(lp33187
I1
aI1
aI3
aI11
aI17
aI61
aI103
aI59
aI477
aI99
aI1203
aI157
aI203
aI557
aI22921
aI47363
aI12853
aI144067
asS'd'
I16594
sa(dp33188
S'a'
I53781
sS's'
I18
sS'm'
(lp33189
I1
aI3
aI1
aI13
aI15
aI23
aI51
aI109
aI499
aI841
aI1779
aI2515
aI2519
aI4945
aI20061
aI12395
aI9223
aI157901
asS'd'
I16595
sa(dp33190
S'a'
I53788
sS's'
I18
sS'm'
(lp33191
I1
aI3
aI7
aI9
aI9
aI15
aI57
aI223
aI223
aI463
aI427
aI2145
aI1219
aI12639
aI28361
aI46019
aI128101
aI198479
asS'd'
I16596
sa(dp33192
S'a'
I53791
sS's'
I18
sS'm'
(lp33193
I1
aI3
aI7
aI7
aI1
aI1
aI99
aI101
aI135
aI169
aI1885
aI3979
aI3051
aI13649
aI26607
aI45067
aI4503
aI74087
asS'd'
I16597
sa(dp33194
S'a'
I53816
sS's'
I18
sS'm'
(lp33195
I1
aI3
aI7
aI5
aI17
aI63
aI27
aI45
aI447
aI759
aI1099
aI3407
aI489
aI2719
aI31577
aI10355
aI126835
aI203439
asS'd'
I16598
sa(dp33196
S'a'
I53824
sS's'
I18
sS'm'
(lp33197
I1
aI1
aI5
aI1
aI21
aI19
aI1
aI239
aI433
aI531
aI1181
aI2021
aI423
aI3235
aI8457
aI44459
aI117401
aI63545
asS'd'
I16599
sa(dp33198
S'a'
I53830
sS's'
I18
sS'm'
(lp33199
I1
aI3
aI7
aI13
aI21
aI39
aI25
aI65
aI405
aI785
aI137
aI2899
aI3255
aI11165
aI7827
aI46425
aI89063
aI102787
asS'd'
I16600
sa(dp33200
S'a'
I53839
sS's'
I18
sS'm'
(lp33201
I1
aI1
aI1
aI11
aI25
aI3
aI39
aI61
aI395
aI35
aI2001
aI3201
aI2233
aI9625
aI26489
aI54167
aI88495
aI127441
asS'd'
I16601
sa(dp33202
S'a'
I53844
sS's'
I18
sS'm'
(lp33203
I1
aI1
aI7
aI7
aI3
aI27
aI11
aI49
aI117
aI711
aI1881
aI1457
aI6759
aI10517
aI12733
aI47435
aI103111
aI237237
asS'd'
I16602
sa(dp33204
S'a'
I53875
sS's'
I18
sS'm'
(lp33205
I1
aI1
aI5
aI5
aI1
aI61
aI121
aI155
aI223
aI733
aI1349
aI2825
aI2093
aI4481
aI21389
aI40227
aI20453
aI116907
asS'd'
I16603
sa(dp33206
S'a'
I53921
sS's'
I18
sS'm'
(lp33207
I1
aI3
aI7
aI7
aI5
aI11
aI85
aI131
aI345
aI723
aI853
aI3679
aI7859
aI11923
aI16619
aI63169
aI127261
aI155665
asS'd'
I16604
sa(dp33208
S'a'
I53922
sS's'
I18
sS'm'
(lp33209
I1
aI1
aI5
aI1
aI3
aI51
aI93
aI225
aI197
aI893
aI555
aI2611
aI6225
aI7819
aI31655
aI12235
aI24919
aI31451
asS'd'
I16605
sa(dp33210
S'a'
I53927
sS's'
I18
sS'm'
(lp33211
I1
aI1
aI3
aI3
aI11
aI23
aI95
aI205
aI85
aI705
aI545
aI155
aI5533
aI14837
aI8341
aI42473
aI96891
aI70695
asS'd'
I16606
sa(dp33212
S'a'
I53948
sS's'
I18
sS'm'
(lp33213
I1
aI3
aI3
aI5
aI17
aI31
aI99
aI187
aI219
aI275
aI685
aI2933
aI4535
aI13495
aI20351
aI60667
aI95211
aI129233
asS'd'
I16607
sa(dp33214
S'a'
I53956
sS's'
I18
sS'm'
(lp33215
I1
aI1
aI1
aI11
aI9
aI11
aI123
aI231
aI127
aI199
aI733
aI2495
aI2601
aI10565
aI3155
aI45251
aI128319
aI187457
asS'd'
I16608
sa(dp33216
S'a'
I53990
sS's'
I18
sS'm'
(lp33217
I1
aI3
aI1
aI9
aI3
aI33
aI41
aI109
aI279
aI851
aI1115
aI3773
aI2383
aI1885
aI6993
aI59693
aI69863
aI88081
asS'd'
I16609
sa(dp33218
S'a'
I53994
sS's'
I18
sS'm'
(lp33219
I1
aI1
aI7
aI13
aI3
aI27
aI27
aI203
aI337
aI965
aI959
aI1125
aI2897
aI8653
aI15827
aI51187
aI12121
aI4665
asS'd'
I16610
sa(dp33220
S'a'
I54001
sS's'
I18
sS'm'
(lp33221
I1
aI3
aI1
aI9
aI19
aI7
aI23
aI113
aI257
aI671
aI571
aI1061
aI4353
aI217
aI13603
aI27961
aI68431
aI147187
asS'd'
I16611
sa(dp33222
S'a'
I54016
sS's'
I18
sS'm'
(lp33223
I1
aI3
aI5
aI9
aI25
aI61
aI7
aI139
aI237
aI859
aI761
aI2005
aI5981
aI153
aI6553
aI53005
aI72653
aI33571
asS'd'
I16612
sa(dp33224
S'a'
I54019
sS's'
I18
sS'm'
(lp33225
I1
aI3
aI5
aI5
aI9
aI35
aI63
aI149
aI427
aI65
aI635
aI1955
aI1845
aI13781
aI9761
aI36147
aI91479
aI141305
asS'd'
I16613
sa(dp33226
S'a'
I54070
sS's'
I18
sS'm'
(lp33227
I1
aI1
aI3
aI5
aI13
aI39
aI53
aI113
aI481
aI933
aI239
aI3713
aI7453
aI12363
aI14763
aI46955
aI108545
aI74349
asS'd'
I16614
sa(dp33228
S'a'
I54074
sS's'
I18
sS'm'
(lp33229
I1
aI3
aI1
aI7
aI13
aI41
aI57
aI225
aI213
aI617
aI1947
aI2855
aI4885
aI8553
aI20259
aI57125
aI59369
aI178553
asS'd'
I16615
sa(dp33230
S'a'
I54088
sS's'
I18
sS'm'
(lp33231
I1
aI3
aI5
aI15
aI31
aI31
aI19
aI87
aI461
aI403
aI1193
aI2123
aI4991
aI14551
aI17153
aI14171
aI17157
aI194879
asS'd'
I16616
sa(dp33232
S'a'
I54102
sS's'
I18
sS'm'
(lp33233
I1
aI1
aI5
aI11
aI5
aI27
aI119
aI65
aI111
aI455
aI1949
aI3441
aI6951
aI6819
aI12839
aI6913
aI57695
aI145925
asS'd'
I16617
sa(dp33234
S'a'
I54111
sS's'
I18
sS'm'
(lp33235
I1
aI1
aI1
aI15
aI19
aI41
aI55
aI45
aI33
aI559
aI589
aI3773
aI745
aI8515
aI32389
aI47797
aI145
aI105503
asS'd'
I16618
sa(dp33236
S'a'
I54130
sS's'
I18
sS'm'
(lp33237
I1
aI1
aI3
aI15
aI13
aI53
aI35
aI223
aI247
aI893
aI149
aI553
aI829
aI5129
aI26417
aI15769
aI95411
aI6595
asS'd'
I16619
sa(dp33238
S'a'
I54152
sS's'
I18
sS'm'
(lp33239
I1
aI1
aI5
aI3
aI3
aI59
aI35
aI187
aI387
aI3
aI847
aI3579
aI7869
aI481
aI23955
aI22191
aI21041
aI230449
asS'd'
I16620
sa(dp33240
S'a'
I54170
sS's'
I18
sS'm'
(lp33241
I1
aI1
aI3
aI15
aI23
aI11
aI97
aI199
aI11
aI647
aI803
aI2391
aI5791
aI2223
aI22187
aI49675
aI87775
aI196871
asS'd'
I16621
sa(dp33242
S'a'
I54172
sS's'
I18
sS'm'
(lp33243
I1
aI3
aI7
aI5
aI25
aI63
aI107
aI227
aI133
aI337
aI1767
aI2459
aI2987
aI10463
aI25001
aI17047
aI79901
aI222877
asS'd'
I16622
sa(dp33244
S'a'
I54211
sS's'
I18
sS'm'
(lp33245
I1
aI3
aI1
aI13
aI25
aI5
aI47
aI109
aI473
aI389
aI1743
aI3951
aI4231
aI827
aI4189
aI29903
aI106909
aI152835
asS'd'
I16623
sa(dp33246
S'a'
I54218
sS's'
I18
sS'm'
(lp33247
I1
aI1
aI5
aI3
aI7
aI61
aI121
aI189
aI303
aI21
aI957
aI545
aI7893
aI3217
aI25847
aI29371
aI100569
aI132393
asS'd'
I16624
sa(dp33248
S'a'
I54228
sS's'
I18
sS'm'
(lp33249
I1
aI1
aI1
aI15
aI29
aI17
aI59
aI37
aI449
aI149
aI845
aI555
aI7603
aI11911
aI18477
aI23279
aI107167
aI160339
asS'd'
I16625
sa(dp33250
S'a'
I54251
sS's'
I18
sS'm'
(lp33251
I1
aI1
aI7
aI5
aI13
aI27
aI43
aI167
aI443
aI445
aI2011
aI2179
aI2785
aI13663
aI21957
aI2455
aI18217
aI19303
asS'd'
I16626
sa(dp33252
S'a'
I54253
sS's'
I18
sS'm'
(lp33253
I1
aI1
aI1
aI5
aI19
aI45
aI71
aI39
aI21
aI791
aI1467
aI855
aI3101
aI8267
aI15529
aI919
aI105313
aI75627
asS'd'
I16627
sa(dp33254
S'a'
I54268
sS's'
I18
sS'm'
(lp33255
I1
aI3
aI1
aI11
aI31
aI25
aI57
aI177
aI211
aI327
aI679
aI771
aI7725
aI6123
aI23931
aI48223
aI9063
aI133319
asS'd'
I16628
sa(dp33256
S'a'
I54271
sS's'
I18
sS'm'
(lp33257
I1
aI3
aI5
aI3
aI1
aI11
aI19
aI153
aI177
aI563
aI1919
aI117
aI5583
aI1519
aI16623
aI10871
aI15511
aI66113
asS'd'
I16629
sa(dp33258
S'a'
I54274
sS's'
I18
sS'm'
(lp33259
I1
aI1
aI7
aI7
aI9
aI45
aI63
aI253
aI415
aI347
aI81
aI2991
aI2691
aI2383
aI15573
aI33783
aI12445
aI224107
asS'd'
I16630
sa(dp33260
S'a'
I54288
sS's'
I18
sS'm'
(lp33261
I1
aI3
aI5
aI5
aI7
aI17
aI99
aI231
aI439
aI1009
aI623
aI833
aI685
aI6419
aI30313
aI56197
aI73239
aI260007
asS'd'
I16631
sa(dp33262
S'a'
I54314
sS's'
I18
sS'm'
(lp33263
I1
aI1
aI5
aI5
aI5
aI51
aI97
aI239
aI267
aI629
aI1211
aI2175
aI5681
aI3107
aI11381
aI57047
aI120175
aI131285
asS'd'
I16632
sa(dp33264
S'a'
I54319
sS's'
I18
sS'm'
(lp33265
I1
aI1
aI7
aI7
aI29
aI11
aI21
aI49
aI481
aI279
aI1795
aI1295
aI453
aI15985
aI19941
aI51853
aI15115
aI107271
asS'd'
I16633
sa(dp33266
S'a'
I54321
sS's'
I18
sS'm'
(lp33267
I1
aI1
aI5
aI1
aI23
aI61
aI33
aI21
aI409
aI57
aI903
aI557
aI1673
aI2759
aI23705
aI4109
aI58865
aI233081
asS'd'
I16634
sa(dp33268
S'a'
I54324
sS's'
I18
sS'm'
(lp33269
I1
aI1
aI1
aI5
aI11
aI37
aI79
aI15
aI213
aI485
aI1477
aI3925
aI3205
aI9267
aI22043
aI54197
aI57101
aI66185
asS'd'
I16635
sa(dp33270
S'a'
I54341
sS's'
I18
sS'm'
(lp33271
I1
aI1
aI7
aI13
aI31
aI27
aI95
aI141
aI131
aI43
aI2039
aI2257
aI17
aI14427
aI5699
aI22263
aI86851
aI226283
asS'd'
I16636
sa(dp33272
S'a'
I54353
sS's'
I18
sS'm'
(lp33273
I1
aI3
aI5
aI5
aI11
aI5
aI5
aI217
aI363
aI163
aI1241
aI3683
aI1409
aI1731
aI20973
aI63849
aI35041
aI94859
asS'd'
I16637
sa(dp33274
S'a'
I54366
sS's'
I18
sS'm'
(lp33275
I1
aI1
aI7
aI1
aI25
aI61
aI67
aI239
aI369
aI319
aI1157
aI2435
aI2147
aI12057
aI4451
aI3005
aI31787
aI199653
asS'd'
I16638
sa(dp33276
S'a'
I54370
sS's'
I18
sS'm'
(lp33277
I1
aI3
aI5
aI1
aI11
aI57
aI1
aI163
aI433
aI11
aI1299
aI1711
aI1601
aI4677
aI16481
aI25175
aI63893
aI41853
asS'd'
I16639
sa(dp33278
S'a'
I54420
sS's'
I18
sS'm'
(lp33279
I1
aI1
aI3
aI1
aI29
aI49
aI91
aI15
aI313
aI533
aI115
aI4005
aI3157
aI10615
aI27915
aI52613
aI64447
aI93091
asS'd'
I16640
sa(dp33280
S'a'
I54423
sS's'
I18
sS'm'
(lp33281
I1
aI3
aI7
aI3
aI7
aI17
aI103
aI67
aI237
aI595
aI1571
aI3421
aI3971
aI11123
aI145
aI52087
aI59273
aI21417
asS'd'
I16641
sa(dp33282
S'a'
I54434
sS's'
I18
sS'm'
(lp33283
I1
aI3
aI5
aI11
aI31
aI37
aI105
aI205
aI377
aI243
aI841
aI3153
aI6847
aI14171
aI19947
aI61561
aI35
aI261753
asS'd'
I16642
sa(dp33284
S'a'
I54440
sS's'
I18
sS'm'
(lp33285
I1
aI3
aI5
aI9
aI29
aI21
aI103
aI219
aI107
aI427
aI1841
aI2015
aI2919
aI5721
aI8631
aI48841
aI33281
aI35835
asS'd'
I16643
sa(dp33286
S'a'
I54454
sS's'
I18
sS'm'
(lp33287
I1
aI1
aI3
aI5
aI25
aI27
aI67
aI65
aI305
aI677
aI1975
aI1049
aI7277
aI15279
aI30181
aI48941
aI119087
aI130265
asS'd'
I16644
sa(dp33288
S'a'
I54495
sS's'
I18
sS'm'
(lp33289
I1
aI1
aI3
aI9
aI29
aI27
aI109
aI167
aI351
aI463
aI663
aI3155
aI919
aI10627
aI30163
aI62233
aI32927
aI210775
asS'd'
I16645
sa(dp33290
S'a'
I54501
sS's'
I18
sS'm'
(lp33291
I1
aI3
aI5
aI5
aI19
aI9
aI17
aI93
aI33
aI999
aI1537
aI3045
aI3735
aI4625
aI31363
aI46075
aI80985
aI108331
asS'd'
I16646
sa(dp33292
S'a'
I54526
sS's'
I18
sS'm'
(lp33293
I1
aI3
aI7
aI7
aI11
aI63
aI83
aI157
aI205
aI505
aI657
aI1901
aI1405
aI8349
aI16473
aI29397
aI130379
aI167963
asS'd'
I16647
sa(dp33294
S'a'
I54639
sS's'
I18
sS'm'
(lp33295
I1
aI3
aI1
aI15
aI25
aI49
aI65
aI189
aI461
aI923
aI1839
aI2597
aI2471
aI14103
aI2915
aI48429
aI74387
aI243465
asS'd'
I16648
sa(dp33296
S'a'
I54653
sS's'
I18
sS'm'
(lp33297
I1
aI3
aI7
aI11
aI31
aI47
aI109
aI21
aI205
aI343
aI1999
aI315
aI8119
aI15937
aI8761
aI55257
aI99983
aI131641
asS'd'
I16649
sa(dp33298
S'a'
I54667
sS's'
I18
sS'm'
(lp33299
I1
aI1
aI3
aI11
aI3
aI23
aI73
aI125
aI17
aI967
aI1811
aI1413
aI4783
aI8303
aI25301
aI26859
aI90583
aI140721
asS'd'
I16650
sa(dp33300
S'a'
I54678
sS's'
I18
sS'm'
(lp33301
I1
aI1
aI5
aI5
aI25
aI41
aI49
aI127
aI391
aI381
aI1575
aI1697
aI5205
aI12805
aI24365
aI20275
aI58819
aI167845
asS'd'
I16651
sa(dp33302
S'a'
I54700
sS's'
I18
sS'm'
(lp33303
I1
aI1
aI7
aI11
aI17
aI53
aI51
aI35
aI383
aI931
aI359
aI2863
aI119
aI6683
aI26247
aI14281
aI49205
aI256869
asS'd'
I16652
sa(dp33304
S'a'
I54717
sS's'
I18
sS'm'
(lp33305
I1
aI1
aI7
aI9
aI23
aI5
aI69
aI97
aI407
aI15
aI579
aI2905
aI47
aI6227
aI23997
aI42459
aI26569
aI225467
asS'd'
I16653
sa(dp33306
S'a'
I54780
sS's'
I18
sS'm'
(lp33307
I1
aI3
aI3
aI11
aI7
aI3
aI125
aI87
aI347
aI79
aI1571
aI1513
aI285
aI12101
aI1731
aI27887
aI42009
aI173243
asS'd'
I16654
sa(dp33308
S'a'
I54801
sS's'
I18
sS'm'
(lp33309
I1
aI3
aI7
aI13
aI3
aI5
aI99
aI29
aI77
aI873
aI1111
aI1451
aI5493
aI10669
aI22597
aI54437
aI55521
aI101617
asS'd'
I16655
sa(dp33310
S'a'
I54813
sS's'
I18
sS'm'
(lp33311
I1
aI3
aI1
aI3
aI3
aI17
aI41
aI215
aI207
aI71
aI683
aI1979
aI4849
aI2437
aI5717
aI28999
aI55005
aI233929
asS'd'
I16656
sa(dp33312
S'a'
I54814
sS's'
I18
sS'm'
(lp33313
I1
aI1
aI1
aI1
aI23
aI21
aI105
aI223
aI5
aI235
aI1533
aI3715
aI2689
aI13937
aI12125
aI63879
aI111537
aI39817
asS'd'
I16657
sa(dp33314
S'a'
I54850
sS's'
I18
sS'm'
(lp33315
I1
aI3
aI3
aI15
aI25
aI47
aI71
aI229
aI21
aI563
aI1851
aI2423
aI131
aI4431
aI31567
aI8883
aI1311
aI227893
asS'd'
I16658
sa(dp33316
S'a'
I54883
sS's'
I18
sS'm'
(lp33317
I1
aI3
aI5
aI11
aI7
aI23
aI19
aI59
aI397
aI315
aI1149
aI3595
aI5973
aI11027
aI5233
aI55237
aI102777
aI137421
asS'd'
I16659
sa(dp33318
S'a'
I54907
sS's'
I18
sS'm'
(lp33319
I1
aI1
aI7
aI9
aI17
aI61
aI45
aI235
aI387
aI171
aI1079
aI3119
aI6933
aI3591
aI751
aI35495
aI49969
aI204611
asS'd'
I16660
sa(dp33320
S'a'
I54919
sS's'
I18
sS'm'
(lp33321
I1
aI1
aI7
aI7
aI21
aI7
aI105
aI79
aI81
aI245
aI1229
aI409
aI5159
aI9815
aI6713
aI4687
aI120541
aI246133
asS'd'
I16661
sa(dp33322
S'a'
I54931
sS's'
I18
sS'm'
(lp33323
I1
aI3
aI7
aI13
aI23
aI31
aI85
aI97
aI481
aI497
aI581
aI1179
aI243
aI1767
aI11855
aI11599
aI3141
aI104741
asS'd'
I16662
sa(dp33324
S'a'
I54933
sS's'
I18
sS'm'
(lp33325
I1
aI3
aI7
aI3
aI3
aI45
aI15
aI29
aI413
aI631
aI273
aI1007
aI2979
aI11307
aI24535
aI9305
aI83591
aI77121
asS'd'
I16663
sa(dp33326
S'a'
I54991
sS's'
I18
sS'm'
(lp33327
I1
aI3
aI7
aI15
aI21
aI55
aI11
aI169
aI417
aI631
aI141
aI1489
aI3371
aI16073
aI11215
aI15479
aI125341
aI131731
asS'd'
I16664
sa(dp33328
S'a'
I55003
sS's'
I18
sS'm'
(lp33329
I1
aI1
aI7
aI5
aI13
aI33
aI7
aI121
aI295
aI191
aI497
aI2233
aI997
aI13833
aI14503
aI38357
aI79007
aI53985
asS'd'
I16665
sa(dp33330
S'a'
I55009
sS's'
I18
sS'm'
(lp33331
I1
aI3
aI3
aI3
aI29
aI63
aI97
aI27
aI449
aI643
aI317
aI1989
aI1481
aI2873
aI21247
aI35989
aI61295
aI101829
asS'd'
I16666
sa(dp33332
S'a'
I55030
sS's'
I18
sS'm'
(lp33333
I1
aI3
aI1
aI7
aI13
aI49
aI27
aI227
aI21
aI983
aI179
aI2761
aI2859
aI2851
aI26447
aI33295
aI126963
aI41441
asS'd'
I16667
sa(dp33334
S'a'
I55034
sS's'
I18
sS'm'
(lp33335
I1
aI1
aI5
aI13
aI27
aI1
aI61
aI115
aI185
aI867
aI2017
aI2257
aI5035
aI7855
aI25849
aI48189
aI28287
aI133261
asS'd'
I16668
sa(dp33336
S'a'
I55039
sS's'
I18
sS'm'
(lp33337
I1
aI1
aI7
aI13
aI27
aI17
aI13
aI205
aI379
aI717
aI247
aI3341
aI2841
aI10845
aI26979
aI5589
aI1935
aI48371
asS'd'
I16669
sa(dp33338
S'a'
I55048
sS's'
I18
sS'm'
(lp33339
I1
aI3
aI1
aI11
aI9
aI51
aI25
aI185
aI65
aI643
aI1867
aI3825
aI3395
aI8883
aI29239
aI20019
aI19071
aI3377
asS'd'
I16670
sa(dp33340
S'a'
I55059
sS's'
I18
sS'm'
(lp33341
I1
aI1
aI1
aI1
aI11
aI57
aI61
aI113
aI419
aI249
aI153
aI2883
aI87
aI7919
aI11941
aI46725
aI38701
aI73715
asS'd'
I16671
sa(dp33342
S'a'
I55061
sS's'
I18
sS'm'
(lp33343
I1
aI3
aI3
aI11
aI3
aI15
aI19
aI87
aI27
aI839
aI463
aI1757
aI3137
aI10821
aI2857
aI58101
aI91983
aI137045
asS'd'
I16672
sa(dp33344
S'a'
I55068
sS's'
I18
sS'm'
(lp33345
I1
aI3
aI3
aI1
aI25
aI25
aI15
aI93
aI359
aI5
aI53
aI647
aI6245
aI1957
aI4651
aI14697
aI12193
aI231303
asS'd'
I16673
sa(dp33346
S'a'
I55077
sS's'
I18
sS'm'
(lp33347
I1
aI1
aI5
aI9
aI31
aI49
aI69
aI223
aI133
aI595
aI777
aI1281
aI727
aI6671
aI21453
aI14193
aI51769
aI258301
asS'd'
I16674
sa(dp33348
S'a'
I55122
sS's'
I18
sS'm'
(lp33349
I1
aI3
aI5
aI11
aI29
aI37
aI75
aI17
aI229
aI121
aI313
aI2873
aI5233
aI13231
aI7589
aI40075
aI42101
aI137697
asS'd'
I16675
sa(dp33350
S'a'
I55149
sS's'
I18
sS'm'
(lp33351
I1
aI1
aI7
aI1
aI31
aI9
aI15
aI63
aI149
aI5
aI1785
aI21
aI2619
aI15071
aI3243
aI58023
aI20697
aI205181
asS'd'
I16676
sa(dp33352
S'a'
I55157
sS's'
I18
sS'm'
(lp33353
I1
aI3
aI7
aI7
aI25
aI61
aI59
aI157
aI251
aI303
aI1905
aI2389
aI1681
aI319
aI14155
aI49089
aI45381
aI124447
asS'd'
I16677
sa(dp33354
S'a'
I55158
sS's'
I18
sS'm'
(lp33355
I1
aI3
aI5
aI5
aI25
aI27
aI41
aI125
aI105
aI867
aI365
aI117
aI7215
aI2887
aI28499
aI9597
aI105999
aI150189
asS'd'
I16678
sa(dp33356
S'a'
I55178
sS's'
I18
sS'm'
(lp33357
I1
aI1
aI5
aI13
aI5
aI33
aI47
aI221
aI207
aI641
aI525
aI3215
aI5293
aI16343
aI16169
aI44393
aI26305
aI194411
asS'd'
I16679
sa(dp33358
S'a'
I55222
sS's'
I18
sS'm'
(lp33359
I1
aI3
aI5
aI13
aI29
aI17
aI31
aI77
aI511
aI465
aI1141
aI597
aI5111
aI6629
aI14557
aI13057
aI11643
aI250925
asS'd'
I16680
sa(dp33360
S'a'
I55234
sS's'
I18
sS'm'
(lp33361
I1
aI1
aI7
aI11
aI1
aI5
aI65
aI139
aI471
aI265
aI1145
aI965
aI47
aI10971
aI15615
aI62031
aI58523
aI175593
asS'd'
I16681
sa(dp33362
S'a'
I55236
sS's'
I18
sS'm'
(lp33363
I1
aI1
aI5
aI1
aI23
aI61
aI57
aI139
aI377
aI843
aI79
aI2873
aI1823
aI7551
aI26741
aI63031
aI124879
aI115295
asS'd'
I16682
sa(dp33364
S'a'
I55251
sS's'
I18
sS'm'
(lp33365
I1
aI1
aI5
aI13
aI9
aI19
aI1
aI61
aI331
aI1015
aI1035
aI1691
aI4057
aI6071
aI24929
aI39569
aI95695
aI39307
asS'd'
I16683
sa(dp33366
S'a'
I55269
sS's'
I18
sS'm'
(lp33367
I1
aI3
aI3
aI5
aI23
aI17
aI13
aI65
aI381
aI893
aI1879
aI3735
aI1547
aI6735
aI30251
aI11471
aI102997
aI126429
asS'd'
I16684
sa(dp33368
S'a'
I55270
sS's'
I18
sS'm'
(lp33369
I1
aI1
aI5
aI13
aI1
aI43
aI15
aI1
aI155
aI221
aI1463
aI3793
aI6467
aI7221
aI28027
aI55357
aI69397
aI87565
asS'd'
I16685
sa(dp33370
S'a'
I55284
sS's'
I18
sS'm'
(lp33371
I1
aI1
aI7
aI3
aI17
aI9
aI71
aI75
aI77
aI639
aI1251
aI701
aI473
aI12337
aI1893
aI6349
aI10837
aI27797
asS'd'
I16686
sa(dp33372
S'a'
I55309
sS's'
I18
sS'm'
(lp33373
I1
aI3
aI5
aI11
aI11
aI11
aI125
aI23
aI161
aI937
aI707
aI2487
aI695
aI8495
aI16219
aI33671
aI109463
aI248305
asS'd'
I16687
sa(dp33374
S'a'
I55322
sS's'
I18
sS'm'
(lp33375
I1
aI1
aI1
aI11
aI5
aI49
aI15
aI47
aI393
aI407
aI39
aI1867
aI7727
aI12701
aI7805
aI119
aI77401
aI186421
asS'd'
I16688
sa(dp33376
S'a'
I55334
sS's'
I18
sS'm'
(lp33377
I1
aI1
aI5
aI5
aI19
aI21
aI77
aI187
aI387
aI51
aI1497
aI1225
aI3101
aI791
aI529
aI4321
aI118435
aI112889
asS'd'
I16689
sa(dp33378
S'a'
I55340
sS's'
I18
sS'm'
(lp33379
I1
aI3
aI1
aI13
aI27
aI17
aI11
aI63
aI201
aI909
aI1549
aI3243
aI1803
aI9461
aI20985
aI24637
aI100993
aI200473
asS'd'
I16690
sa(dp33380
S'a'
I55348
sS's'
I18
sS'm'
(lp33381
I1
aI3
aI7
aI13
aI11
aI35
aI97
aI213
aI415
aI467
aI2013
aI2159
aI7017
aI7895
aI18235
aI50659
aI113169
aI141887
asS'd'
I16691
sa(dp33382
S'a'
I55377
sS's'
I18
sS'm'
(lp33383
I1
aI1
aI3
aI7
aI13
aI21
aI119
aI109
aI471
aI323
aI277
aI1685
aI2399
aI14777
aI2643
aI5879
aI113043
aI45223
asS'd'
I16692
sa(dp33384
S'a'
I55430
sS's'
I18
sS'm'
(lp33385
I1
aI3
aI1
aI13
aI19
aI5
aI1
aI75
aI499
aI297
aI1897
aI591
aI3223
aI12939
aI30593
aI4053
aI122207
aI215171
asS'd'
I16693
sa(dp33386
S'a'
I55433
sS's'
I18
sS'm'
(lp33387
I1
aI3
aI3
aI9
aI21
aI11
aI29
aI205
aI13
aI381
aI569
aI599
aI7089
aI8145
aI18531
aI34477
aI101057
aI64269
asS'd'
I16694
sa(dp33388
S'a'
I55441
sS's'
I18
sS'm'
(lp33389
I1
aI1
aI5
aI15
aI1
aI19
aI37
aI131
aI325
aI441
aI3
aI4001
aI6937
aI9207
aI27543
aI30321
aI37083
aI241019
asS'd'
I16695
sa(dp33390
S'a'
I55470
sS's'
I18
sS'm'
(lp33391
I1
aI3
aI7
aI13
aI7
aI15
aI9
aI159
aI97
aI905
aI557
aI1913
aI7325
aI4057
aI19461
aI14277
aI36873
aI25619
asS'd'
I16696
sa(dp33392
S'a'
I55535
sS's'
I18
sS'm'
(lp33393
I1
aI3
aI5
aI7
aI3
aI51
aI99
aI9
aI185
aI227
aI2041
aI331
aI3925
aI12481
aI17485
aI37137
aI3753
aI125269
asS'd'
I16697
sa(dp33394
S'a'
I55561
sS's'
I18
sS'm'
(lp33395
I1
aI1
aI7
aI11
aI31
aI49
aI89
aI37
aI49
aI863
aI833
aI3263
aI351
aI6277
aI23055
aI49727
aI25005
aI161585
asS'd'
I16698
sa(dp33396
S'a'
I55567
sS's'
I18
sS'm'
(lp33397
I1
aI3
aI5
aI1
aI9
aI35
aI89
aI101
aI117
aI365
aI1015
aI1159
aI4623
aI4541
aI6831
aI28091
aI10647
aI221415
asS'd'
I16699
sa(dp33398
S'a'
I55597
sS's'
I18
sS'm'
(lp33399
I1
aI1
aI5
aI5
aI13
aI47
aI125
aI209
aI199
aI885
aI927
aI1411
aI795
aI8835
aI28589
aI48753
aI27191
aI53455
asS'd'
I16700
sa(dp33400
S'a'
I55630
sS's'
I18
sS'm'
(lp33401
I1
aI1
aI5
aI9
aI7
aI19
aI3
aI87
aI157
aI121
aI1433
aI1463
aI3241
aI5969
aI203
aI36723
aI14779
aI63949
asS'd'
I16701
sa(dp33402
S'a'
I55648
sS's'
I18
sS'm'
(lp33403
I1
aI1
aI3
aI9
aI1
aI47
aI71
aI113
aI405
aI561
aI1149
aI3599
aI4173
aI6819
aI5493
aI45987
aI41521
aI221503
asS'd'
I16702
sa(dp33404
S'a'
I55653
sS's'
I18
sS'm'
(lp33405
I1
aI3
aI3
aI1
aI3
aI55
aI101
aI103
aI161
aI549
aI457
aI2529
aI2043
aI8843
aI5677
aI7449
aI45185
aI178289
asS'd'
I16703
sa(dp33406
S'a'
I55657
sS's'
I18
sS'm'
(lp33407
I1
aI1
aI1
aI3
aI31
aI25
aI1
aI161
aI7
aI503
aI641
aI2221
aI749
aI1521
aI6151
aI19245
aI55913
aI80141
asS'd'
I16704
sa(dp33408
S'a'
I55665
sS's'
I18
sS'm'
(lp33409
I1
aI1
aI1
aI9
aI3
aI45
aI73
aI217
aI249
aI929
aI163
aI2139
aI3921
aI11223
aI11161
aI52697
aI89633
aI14243
asS'd'
I16705
sa(dp33410
S'a'
I55678
sS's'
I18
sS'm'
(lp33411
I1
aI1
aI7
aI15
aI17
aI41
aI5
aI119
aI211
aI53
aI985
aI2679
aI679
aI9349
aI25577
aI26947
aI35141
aI93999
asS'd'
I16706
sa(dp33412
S'a'
I55684
sS's'
I18
sS'm'
(lp33413
I1
aI3
aI1
aI15
aI17
aI43
aI51
aI15
aI363
aI615
aI889
aI195
aI6279
aI15477
aI31545
aI50941
aI119711
aI66535
asS'd'
I16707
sa(dp33414
S'a'
I55691
sS's'
I18
sS'm'
(lp33415
I1
aI1
aI1
aI13
aI7
aI11
aI17
aI127
aI131
aI759
aI739
aI161
aI5937
aI13611
aI31757
aI10681
aI101357
aI82873
asS'd'
I16708
sa(dp33416
S'a'
I55693
sS's'
I18
sS'm'
(lp33417
I1
aI3
aI5
aI7
aI21
aI63
aI75
aI33
aI233
aI981
aI589
aI3409
aI3523
aI1871
aI8919
aI38513
aI32825
aI56935
asS'd'
I16709
sa(dp33418
S'a'
I55702
sS's'
I18
sS'm'
(lp33419
I1
aI3
aI5
aI3
aI9
aI9
aI85
aI221
aI203
aI727
aI1035
aI1069
aI2409
aI2687
aI235
aI23395
aI64163
aI193235
asS'd'
I16710
sa(dp33420
S'a'
I55708
sS's'
I18
sS'm'
(lp33421
I1
aI3
aI3
aI7
aI1
aI35
aI119
aI175
aI203
aI819
aI207
aI2283
aI4175
aI3581
aI11647
aI43073
aI104573
aI86607
asS'd'
I16711
sa(dp33422
S'a'
I55715
sS's'
I18
sS'm'
(lp33423
I1
aI3
aI3
aI15
aI11
aI63
aI59
aI153
aI279
aI779
aI261
aI3317
aI7671
aI11727
aI19381
aI33227
aI79331
aI187227
asS'd'
I16712
sa(dp33424
S'a'
I55739
sS's'
I18
sS'm'
(lp33425
I1
aI1
aI3
aI1
aI7
aI1
aI115
aI15
aI235
aI9
aI1877
aI1911
aI1089
aI9939
aI9537
aI39563
aI95327
aI70323
asS'd'
I16713
sa(dp33426
S'a'
I55761
sS's'
I18
sS'm'
(lp33427
I1
aI1
aI5
aI7
aI25
aI61
aI63
aI145
aI425
aI617
aI1813
aI3255
aI6797
aI16019
aI18849
aI44191
aI69877
aI179933
asS'd'
I16714
sa(dp33428
S'a'
I55767
sS's'
I18
sS'm'
(lp33429
I1
aI1
aI3
aI13
aI17
aI45
aI69
aI247
aI27
aI367
aI871
aI1185
aI895
aI7991
aI8145
aI22869
aI97609
aI14673
asS'd'
I16715
sa(dp33430
S'a'
I55768
sS's'
I18
sS'm'
(lp33431
I1
aI3
aI3
aI11
aI19
aI41
aI99
aI213
aI159
aI803
aI121
aI1197
aI2849
aI15191
aI15603
aI52445
aI105077
aI128231
asS'd'
I16716
sa(dp33432
S'a'
I55774
sS's'
I18
sS'm'
(lp33433
I1
aI3
aI1
aI11
aI21
aI61
aI117
aI167
aI437
aI447
aI419
aI1673
aI755
aI15331
aI29819
aI16099
aI130773
aI177547
asS'd'
I16717
sa(dp33434
S'a'
I55787
sS's'
I18
sS'm'
(lp33435
I1
aI3
aI7
aI7
aI1
aI15
aI79
aI109
aI351
aI71
aI985
aI89
aI7517
aI4175
aI30533
aI52125
aI100863
aI186477
asS'd'
I16718
sa(dp33436
S'a'
I55811
sS's'
I18
sS'm'
(lp33437
I1
aI1
aI3
aI1
aI15
aI1
aI103
aI65
aI511
aI241
aI1279
aI3233
aI7141
aI255
aI10925
aI28271
aI56151
aI252121
asS'd'
I16719
sa(dp33438
S'a'
I55835
sS's'
I18
sS'm'
(lp33439
I1
aI1
aI1
aI13
aI17
aI49
aI59
aI93
aI19
aI343
aI979
aI865
aI3447
aI4595
aI3067
aI26807
aI98915
aI126237
asS'd'
I16720
sa(dp33440
S'a'
I55894
sS's'
I18
sS'm'
(lp33441
I1
aI3
aI3
aI5
aI17
aI5
aI91
aI199
aI191
aI775
aI233
aI919
aI277
aI3485
aI9231
aI37025
aI23493
aI186745
asS'd'
I16721
sa(dp33442
S'a'
I55897
sS's'
I18
sS'm'
(lp33443
I1
aI3
aI1
aI1
aI11
aI5
aI103
aI187
aI85
aI47
aI1111
aI883
aI6155
aI15315
aI9041
aI58275
aI75037
aI7773
asS'd'
I16722
sa(dp33444
S'a'
I55904
sS's'
I18
sS'm'
(lp33445
I1
aI3
aI1
aI3
aI19
aI5
aI7
aI211
aI481
aI713
aI383
aI1203
aI6089
aI15817
aI31577
aI7283
aI25457
aI101455
asS'd'
I16723
sa(dp33446
S'a'
I55931
sS's'
I18
sS'm'
(lp33447
I1
aI3
aI5
aI7
aI21
aI9
aI59
aI127
aI375
aI477
aI721
aI3931
aI7089
aI9079
aI5015
aI62019
aI113747
aI36055
asS'd'
I16724
sa(dp33448
S'a'
I55950
sS's'
I18
sS'm'
(lp33449
I1
aI3
aI7
aI13
aI3
aI17
aI47
aI177
aI103
aI535
aI1787
aI509
aI5253
aI2857
aI13421
aI19875
aI37397
aI251353
asS'd'
I16725
sa(dp33450
S'a'
I55961
sS's'
I18
sS'm'
(lp33451
I1
aI1
aI5
aI7
aI19
aI31
aI41
aI93
aI301
aI45
aI251
aI2691
aI4657
aI2627
aI17321
aI24627
aI80221
aI117191
asS'd'
I16726
sa(dp33452
S'a'
I55973
sS's'
I18
sS'm'
(lp33453
I1
aI3
aI5
aI7
aI5
aI31
aI27
aI3
aI463
aI549
aI1669
aI499
aI815
aI4091
aI7049
aI60957
aI102849
aI235617
asS'd'
I16727
sa(dp33454
S'a'
I56078
sS's'
I18
sS'm'
(lp33455
I1
aI3
aI5
aI1
aI21
aI31
aI57
aI201
aI503
aI977
aI893
aI3927
aI1605
aI8265
aI5137
aI51009
aI89375
aI237909
asS'd'
I16728
sa(dp33456
S'a'
I56099
sS's'
I18
sS'm'
(lp33457
I1
aI3
aI3
aI1
aI27
aI5
aI11
aI81
aI445
aI229
aI5
aI543
aI3397
aI12961
aI31911
aI36945
aI59485
aI305
asS'd'
I16729
sa(dp33458
S'a'
I56105
sS's'
I18
sS'm'
(lp33459
I1
aI1
aI5
aI13
aI31
aI63
aI39
aI171
aI243
aI39
aI1147
aI459
aI7215
aI14603
aI20625
aI47369
aI121495
aI237741
asS'd'
I16730
sa(dp33460
S'a'
I56119
sS's'
I18
sS'm'
(lp33461
I1
aI3
aI3
aI13
aI15
aI63
aI39
aI23
aI305
aI685
aI1885
aI571
aI2657
aI16031
aI24759
aI10639
aI25619
aI246137
asS'd'
I16731
sa(dp33462
S'a'
I56133
sS's'
I18
sS'm'
(lp33463
I1
aI1
aI1
aI5
aI19
aI33
aI5
aI187
aI167
aI725
aI1405
aI511
aI701
aI13283
aI3513
aI16495
aI8755
aI221751
asS'd'
I16732
sa(dp33464
S'a'
I56168
sS's'
I18
sS'm'
(lp33465
I1
aI1
aI7
aI11
aI3
aI27
aI27
aI237
aI495
aI637
aI479
aI3247
aI3825
aI2567
aI12853
aI52881
aI34807
aI161483
asS'd'
I16733
sa(dp33466
S'a'
I56191
sS's'
I18
sS'm'
(lp33467
I1
aI3
aI3
aI9
aI23
aI43
aI101
aI175
aI19
aI443
aI787
aI1053
aI4113
aI12777
aI4615
aI53115
aI2873
aI117383
asS'd'
I16734
sa(dp33468
S'a'
I56202
sS's'
I18
sS'm'
(lp33469
I1
aI3
aI1
aI13
aI3
aI23
aI33
aI93
aI145
aI937
aI957
aI2463
aI827
aI383
aI16749
aI61567
aI10029
aI188159
asS'd'
I16735
sa(dp33470
S'a'
I56209
sS's'
I18
sS'm'
(lp33471
I1
aI1
aI7
aI15
aI21
aI23
aI3
aI71
aI323
aI995
aI645
aI1189
aI1029
aI519
aI3479
aI13587
aI95641
aI215337
asS'd'
I16736
sa(dp33472
S'a'
I56215
sS's'
I18
sS'm'
(lp33473
I1
aI3
aI7
aI11
aI9
aI17
aI101
aI59
aI421
aI417
aI797
aI3089
aI773
aI15959
aI18127
aI13681
aI104667
aI217433
asS'd'
I16737
sa(dp33474
S'a'
I56232
sS's'
I18
sS'm'
(lp33475
I1
aI3
aI5
aI7
aI31
aI21
aI9
aI7
aI377
aI589
aI1497
aI939
aI5389
aI10997
aI22291
aI19639
aI72187
aI66193
asS'd'
I16738
sa(dp33476
S'a'
I56240
sS's'
I18
sS'm'
(lp33477
I1
aI1
aI1
aI13
aI19
aI1
aI127
aI185
aI251
aI167
aI1289
aI2715
aI5885
aI12715
aI18261
aI36861
aI102721
aI246917
asS'd'
I16739
sa(dp33478
S'a'
I56260
sS's'
I18
sS'm'
(lp33479
I1
aI1
aI7
aI1
aI23
aI41
aI19
aI151
aI125
aI465
aI813
aI1711
aI7933
aI13561
aI29737
aI59207
aI62533
aI124149
asS'd'
I16740
sa(dp33480
S'a'
I56270
sS's'
I18
sS'm'
(lp33481
I1
aI3
aI5
aI9
aI7
aI13
aI17
aI119
aI425
aI877
aI1207
aI2211
aI2943
aI13921
aI28251
aI44143
aI112149
aI152341
asS'd'
I16741
sa(dp33482
S'a'
I56278
sS's'
I18
sS'm'
(lp33483
I1
aI3
aI5
aI9
aI15
aI21
aI87
aI83
aI77
aI731
aI91
aI3091
aI5687
aI9647
aI2037
aI39031
aI106583
aI66533
asS'd'
I16742
sa(dp33484
S'a'
I56281
sS's'
I18
sS'm'
(lp33485
I1
aI1
aI7
aI9
aI31
aI49
aI7
aI119
aI147
aI599
aI1191
aI297
aI1597
aI10723
aI16893
aI47387
aI106995
aI165409
asS'd'
I16743
sa(dp33486
S'a'
I56288
sS's'
I18
sS'm'
(lp33487
I1
aI3
aI3
aI3
aI3
aI63
aI11
aI193
aI241
aI63
aI1671
aI2139
aI5689
aI13967
aI9239
aI7535
aI34237
aI140283
asS'd'
I16744
sa(dp33488
S'a'
I56303
sS's'
I18
sS'm'
(lp33489
I1
aI3
aI5
aI13
aI9
aI23
aI65
aI247
aI473
aI825
aI109
aI1897
aI245
aI10517
aI8147
aI25989
aI96447
aI118689
asS'd'
I16745
sa(dp33490
S'a'
I56308
sS's'
I18
sS'm'
(lp33491
I1
aI1
aI3
aI5
aI27
aI35
aI65
aI23
aI159
aI729
aI189
aI2661
aI4245
aI14377
aI21043
aI15551
aI2717
aI146949
asS'd'
I16746
sa(dp33492
S'a'
I56312
sS's'
I18
sS'm'
(lp33493
I1
aI1
aI3
aI13
aI23
aI5
aI35
aI63
aI293
aI347
aI883
aI149
aI5145
aI10821
aI5813
aI24183
aI94711
aI64787
asS'd'
I16747
sa(dp33494
S'a'
I56320
sS's'
I18
sS'm'
(lp33495
I1
aI1
aI5
aI3
aI27
aI3
aI127
aI141
aI237
aI535
aI1509
aI2755
aI5843
aI2379
aI19413
aI52345
aI100247
aI42571
asS'd'
I16748
sa(dp33496
S'a'
I56326
sS's'
I18
sS'm'
(lp33497
I1
aI3
aI3
aI9
aI1
aI55
aI61
aI105
aI29
aI1021
aI1215
aI2157
aI7453
aI4643
aI26793
aI33553
aI2959
aI51485
asS'd'
I16749
sa(dp33498
S'a'
I56392
sS's'
I18
sS'm'
(lp33499
I1
aI1
aI3
aI7
aI31
aI51
aI59
aI49
aI321
aI207
aI415
aI2115
aI219
aI5045
aI31133
aI17961
aI130779
aI28255
asS'd'
I16750
sa(dp33500
S'a'
I56395
sS's'
I18
sS'm'
(lp33501
I1
aI3
aI7
aI15
aI9
aI29
aI31
aI185
aI111
aI959
aI7
aI827
aI7891
aI5449
aI22221
aI49933
aI2091
aI194683
asS'd'
I16751
sa(dp33502
S'a'
I56403
sS's'
I18
sS'm'
(lp33503
I1
aI3
aI7
aI1
aI11
aI59
aI75
aI255
aI387
aI913
aI423
aI2915
aI5079
aI6363
aI5175
aI57977
aI5559
aI13257
asS'd'
I16752
sa(dp33504
S'a'
I56419
sS's'
I18
sS'm'
(lp33505
I1
aI1
aI7
aI1
aI21
aI3
aI21
aI13
aI157
aI3
aI715
aI3525
aI7769
aI5333
aI25345
aI53473
aI44323
aI203167
asS'd'
I16753
sa(dp33506
S'a'
I56428
sS's'
I18
sS'm'
(lp33507
I1
aI1
aI7
aI11
aI31
aI25
aI55
aI5
aI169
aI695
aI1599
aI2357
aI1427
aI14469
aI15223
aI34275
aI42605
aI23005
asS'd'
I16754
sa(dp33508
S'a'
I56450
sS's'
I18
sS'm'
(lp33509
I1
aI1
aI1
aI15
aI19
aI51
aI117
aI135
aI297
aI831
aI329
aI3793
aI4673
aI3795
aI24185
aI52971
aI30423
aI68771
asS'd'
I16755
sa(dp33510
S'a'
I56452
sS's'
I18
sS'm'
(lp33511
I1
aI1
aI7
aI5
aI19
aI33
aI79
aI77
aI315
aI29
aI307
aI1709
aI3489
aI14515
aI12477
aI58939
aI53753
aI165031
asS'd'
I16756
sa(dp33512
S'a'
I56485
sS's'
I18
sS'm'
(lp33513
I1
aI1
aI7
aI1
aI27
aI57
aI119
aI207
aI355
aI279
aI1371
aI3917
aI2821
aI5285
aI12673
aI28973
aI54957
aI94001
asS'd'
I16757
sa(dp33514
S'a'
I56486
sS's'
I18
sS'm'
(lp33515
I1
aI3
aI7
aI3
aI19
aI57
aI53
aI199
aI485
aI805
aI301
aI1337
aI5993
aI2187
aI30573
aI12045
aI101205
aI129841
asS'd'
I16758
sa(dp33516
S'a'
I56492
sS's'
I18
sS'm'
(lp33517
I1
aI1
aI3
aI9
aI15
aI45
aI71
aI119
aI445
aI759
aI1361
aI1299
aI2927
aI2343
aI22085
aI53733
aI21241
aI1553
asS'd'
I16759
sa(dp33518
S'a'
I56498
sS's'
I18
sS'm'
(lp33519
I1
aI3
aI5
aI3
aI27
aI11
aI1
aI239
aI497
aI343
aI1989
aI1463
aI2473
aI5191
aI6271
aI14129
aI124453
aI96817
asS'd'
I16760
sa(dp33520
S'a'
I56510
sS's'
I18
sS'm'
(lp33521
I1
aI3
aI5
aI7
aI27
aI19
aI123
aI27
aI483
aI557
aI1545
aI1871
aI1297
aI587
aI1067
aI51259
aI119231
aI173659
asS'd'
I16761
sa(dp33522
S'a'
I56512
sS's'
I18
sS'm'
(lp33523
I1
aI3
aI1
aI1
aI27
aI45
aI41
aI113
aI453
aI553
aI2019
aI2039
aI1709
aI13017
aI5497
aI34459
aI60295
aI229405
asS'd'
I16762
sa(dp33524
S'a'
I56524
sS's'
I18
sS'm'
(lp33525
I1
aI3
aI1
aI11
aI1
aI57
aI51
aI125
aI261
aI915
aI1673
aI25
aI529
aI653
aI17247
aI64225
aI98991
aI248143
asS'd'
I16763
sa(dp33526
S'a'
I56530
sS's'
I18
sS'm'
(lp33527
I1
aI3
aI5
aI15
aI25
aI27
aI31
aI1
aI463
aI249
aI113
aI1955
aI2223
aI5463
aI12281
aI20843
aI26495
aI256759
asS'd'
I16764
sa(dp33528
S'a'
I56545
sS's'
I18
sS'm'
(lp33529
I1
aI1
aI3
aI11
aI27
aI33
aI57
aI205
aI89
aI435
aI1983
aI1165
aI3843
aI127
aI30179
aI63971
aI10211
aI105403
asS'd'
I16765
sa(dp33530
S'a'
I56551
sS's'
I18
sS'm'
(lp33531
I1
aI3
aI3
aI5
aI21
aI49
aI35
aI161
aI273
aI205
aI41
aI1881
aI2013
aI12549
aI24859
aI55711
aI98235
aI237281
asS'd'
I16766
sa(dp33532
S'a'
I56565
sS's'
I18
sS'm'
(lp33533
I1
aI3
aI3
aI1
aI15
aI35
aI95
aI1
aI221
aI675
aI385
aI2257
aI2531
aI2129
aI12895
aI11565
aI125977
aI51973
asS'd'
I16767
sa(dp33534
S'a'
I56580
sS's'
I18
sS'm'
(lp33535
I1
aI1
aI1
aI15
aI19
aI61
aI35
aI55
aI9
aI721
aI499
aI2577
aI3001
aI14861
aI22293
aI56195
aI72855
aI166703
asS'd'
I16768
sa(dp33536
S'a'
I56587
sS's'
I18
sS'm'
(lp33537
I1
aI1
aI1
aI7
aI5
aI25
aI59
aI175
aI81
aI989
aI935
aI2579
aI8183
aI1109
aI4645
aI53753
aI115795
aI105091
asS'd'
I16769
sa(dp33538
S'a'
I56589
sS's'
I18
sS'm'
(lp33539
I1
aI3
aI3
aI13
aI7
aI55
aI7
aI113
aI197
aI763
aI1747
aI3291
aI1109
aI4391
aI18257
aI28563
aI97413
aI5847
asS'd'
I16770
sa(dp33540
S'a'
I56592
sS's'
I18
sS'm'
(lp33541
I1
aI1
aI1
aI7
aI23
aI55
aI91
aI83
aI479
aI305
aI843
aI2055
aI3405
aI15243
aI31551
aI5275
aI8651
aI66915
asS'd'
I16771
sa(dp33542
S'a'
I56611
sS's'
I18
sS'm'
(lp33543
I1
aI3
aI7
aI9
aI3
aI19
aI83
aI229
aI235
aI903
aI1495
aI1033
aI2729
aI14927
aI11847
aI22979
aI13905
aI84413
asS'd'
I16772
sa(dp33544
S'a'
I56623
sS's'
I18
sS'm'
(lp33545
I1
aI3
aI3
aI13
aI27
aI37
aI83
aI193
aI475
aI439
aI745
aI757
aI7359
aI6683
aI5839
aI50765
aI6933
aI117411
asS'd'
I16773
sa(dp33546
S'a'
I56635
sS's'
I18
sS'm'
(lp33547
I1
aI3
aI5
aI11
aI31
aI25
aI33
aI77
aI113
aI815
aI123
aI2721
aI2133
aI8995
aI15237
aI54565
aI5155
aI51235
asS'd'
I16774
sa(dp33548
S'a'
I56646
sS's'
I18
sS'm'
(lp33549
I1
aI3
aI3
aI7
aI15
aI31
aI73
aI91
aI379
aI39
aI913
aI53
aI41
aI1059
aI25883
aI11769
aI63015
aI48125
asS'd'
I16775
sa(dp33550
S'a'
I56660
sS's'
I18
sS'm'
(lp33551
I1
aI1
aI5
aI5
aI5
aI13
aI81
aI169
aI71
aI529
aI1429
aI2101
aI4069
aI5509
aI30283
aI40625
aI103673
aI183243
asS'd'
I16776
sa(dp33552
S'a'
I56680
sS's'
I18
sS'm'
(lp33553
I1
aI3
aI3
aI5
aI23
aI39
aI39
aI237
aI445
aI567
aI343
aI2521
aI2287
aI1851
aI2315
aI59979
aI5015
aI243349
asS'd'
I16777
sa(dp33554
S'a'
I56686
sS's'
I18
sS'm'
(lp33555
I1
aI1
aI7
aI1
aI1
aI51
aI89
aI229
aI187
aI207
aI245
aI3521
aI2987
aI4347
aI6997
aI62565
aI54397
aI140473
asS'd'
I16778
sa(dp33556
S'a'
I56716
sS's'
I18
sS'm'
(lp33557
I1
aI3
aI1
aI5
aI7
aI59
aI45
aI161
aI457
aI655
aI1591
aI215
aI2213
aI15101
aI14791
aI40397
aI95811
aI126291
asS'd'
I16779
sa(dp33558
S'a'
I56749
sS's'
I18
sS'm'
(lp33559
I1
aI1
aI3
aI1
aI5
aI23
aI7
aI199
aI143
aI561
aI1669
aI17
aI8109
aI11003
aI4535
aI8593
aI112021
aI223153
asS'd'
I16780
sa(dp33560
S'a'
I56790
sS's'
I18
sS'm'
(lp33561
I1
aI3
aI5
aI9
aI3
aI37
aI111
aI15
aI235
aI697
aI385
aI2197
aI909
aI1247
aI26199
aI50661
aI100643
aI122577
asS'd'
I16781
sa(dp33562
S'a'
I56809
sS's'
I18
sS'm'
(lp33563
I1
aI3
aI5
aI11
aI23
aI53
aI95
aI75
aI463
aI137
aI1511
aI3373
aI3071
aI547
aI22399
aI51891
aI9123
aI240925
asS'd'
I16782
sa(dp33564
S'a'
I56869
sS's'
I18
sS'm'
(lp33565
I1
aI3
aI7
aI3
aI21
aI35
aI69
aI197
aI371
aI15
aI185
aI3539
aI29
aI15071
aI17069
aI34669
aI37023
aI189385
asS'd'
I16783
sa(dp33566
S'a'
I56884
sS's'
I18
sS'm'
(lp33567
I1
aI1
aI1
aI15
aI5
aI21
aI7
aI5
aI201
aI881
aI841
aI827
aI503
aI3545
aI17771
aI64481
aI65105
aI209947
asS'd'
I16784
sa(dp33568
S'a'
I56887
sS's'
I18
sS'm'
(lp33569
I1
aI1
aI1
aI5
aI3
aI31
aI83
aI201
aI455
aI169
aI1797
aI1769
aI1999
aI8629
aI14313
aI16851
aI64955
aI180631
asS'd'
I16785
sa(dp33570
S'a'
I56893
sS's'
I18
sS'm'
(lp33571
I1
aI1
aI5
aI5
aI1
aI35
aI49
aI61
aI499
aI619
aI1509
aI3015
aI237
aI8979
aI3471
aI11513
aI80193
aI24135
asS'd'
I16786
sa(dp33572
S'a'
I56906
sS's'
I18
sS'm'
(lp33573
I1
aI3
aI3
aI9
aI25
aI29
aI111
aI19
aI339
aI739
aI1751
aI2671
aI5399
aI5965
aI3943
aI45577
aI70605
aI203117
asS'd'
I16787
sa(dp33574
S'a'
I56932
sS's'
I18
sS'm'
(lp33575
I1
aI3
aI3
aI7
aI3
aI9
aI15
aI147
aI177
aI545
aI161
aI2211
aI4653
aI15891
aI15939
aI19153
aI77827
aI245787
asS'd'
I16788
sa(dp33576
S'a'
I56959
sS's'
I18
sS'm'
(lp33577
I1
aI1
aI1
aI1
aI25
aI47
aI37
aI159
aI273
aI825
aI1037
aI2047
aI7149
aI5517
aI699
aI49687
aI110115
aI159475
asS'd'
I16789
sa(dp33578
S'a'
I56965
sS's'
I18
sS'm'
(lp33579
I1
aI3
aI1
aI7
aI7
aI55
aI77
aI231
aI197
aI381
aI2013
aI2421
aI7551
aI9955
aI21031
aI11365
aI48271
aI190147
asS'd'
I16790
sa(dp33580
S'a'
I56983
sS's'
I18
sS'm'
(lp33581
I1
aI1
aI5
aI9
aI25
aI1
aI81
aI145
aI215
aI427
aI905
aI2307
aI6149
aI12777
aI131
aI57091
aI106137
aI24625
asS'd'
I16791
sa(dp33582
S'a'
I57018
sS's'
I18
sS'm'
(lp33583
I1
aI3
aI1
aI13
aI13
aI63
aI103
aI245
aI275
aI745
aI841
aI2993
aI2083
aI8903
aI4499
aI55979
aI22323
aI244447
asS'd'
I16792
sa(dp33584
S'a'
I57023
sS's'
I18
sS'm'
(lp33585
I1
aI1
aI5
aI5
aI15
aI11
aI59
aI181
aI191
aI219
aI599
aI59
aI1079
aI4445
aI16537
aI31127
aI103257
aI233855
asS'd'
I16793
sa(dp33586
S'a'
I57025
sS's'
I18
sS'm'
(lp33587
I1
aI3
aI7
aI9
aI9
aI37
aI109
aI41
aI145
aI1001
aI609
aI551
aI6843
aI13791
aI15103
aI27851
aI7693
aI145207
asS'd'
I16794
sa(dp33588
S'a'
I57032
sS's'
I18
sS'm'
(lp33589
I1
aI3
aI1
aI9
aI3
aI35
aI63
aI219
aI49
aI567
aI1537
aI1327
aI6487
aI16039
aI26019
aI13851
aI116929
aI175121
asS'd'
I16795
sa(dp33590
S'a'
I57040
sS's'
I18
sS'm'
(lp33591
I1
aI3
aI7
aI15
aI17
aI31
aI27
aI91
aI241
aI229
aI485
aI2601
aI3859
aI12609
aI19847
aI31939
aI50815
aI235529
asS'd'
I16796
sa(dp33592
S'a'
I57046
sS's'
I18
sS'm'
(lp33593
I1
aI1
aI5
aI15
aI27
aI31
aI3
aI47
aI69
aI427
aI95
aI1445
aI1223
aI2953
aI32343
aI6841
aI67851
aI79561
asS'd'
I16797
sa(dp33594
S'a'
I57071
sS's'
I18
sS'm'
(lp33595
I1
aI3
aI5
aI13
aI13
aI37
aI19
aI127
aI259
aI139
aI1597
aI651
aI4845
aI6413
aI18205
aI56005
aI32107
aI140783
asS'd'
I16798
sa(dp33596
S'a'
I57091
sS's'
I18
sS'm'
(lp33597
I1
aI1
aI7
aI5
aI15
aI23
aI81
aI195
aI127
aI113
aI499
aI733
aI5907
aI12107
aI18105
aI28113
aI16111
aI152327
asS'd'
I16799
sa(dp33598
S'a'
I57094
sS's'
I18
sS'm'
(lp33599
I1
aI1
aI5
aI15
aI9
aI49
aI109
aI181
aI187
aI591
aI1625
aI3641
aI313
aI1225
aI11725
aI9047
aI30351
aI124301
asS'd'
I16800
sa(dp33600
S'a'
I57108
sS's'
I18
sS'm'
(lp33601
I1
aI3
aI1
aI1
aI9
aI45
aI103
aI219
aI155
aI805
aI1775
aI759
aI1687
aI11415
aI21623
aI37831
aI18995
aI21667
asS'd'
I16801
sa(dp33602
S'a'
I57122
sS's'
I18
sS'm'
(lp33603
I1
aI3
aI3
aI5
aI25
aI13
aI11
aI37
aI489
aI935
aI373
aI811
aI5045
aI3615
aI2111
aI22909
aI117155
aI69483
asS'd'
I16802
sa(dp33604
S'a'
I57127
sS's'
I18
sS'm'
(lp33605
I1
aI3
aI3
aI5
aI9
aI45
aI71
aI87
aI265
aI93
aI161
aI2983
aI1023
aI3633
aI5965
aI9499
aI35653
aI219257
asS'd'
I16803
sa(dp33606
S'a'
I57168
sS's'
I18
sS'm'
(lp33607
I1
aI1
aI5
aI13
aI21
aI27
aI101
aI231
aI85
aI469
aI1023
aI3735
aI5093
aI253
aI22585
aI61975
aI81041
aI4175
asS'd'
I16804
sa(dp33608
S'a'
I57183
sS's'
I18
sS'm'
(lp33609
I1
aI3
aI1
aI7
aI5
aI41
aI105
aI153
aI391
aI5
aI1917
aI331
aI7679
aI14359
aI13177
aI40755
aI78669
aI133527
asS'd'
I16805
sa(dp33610
S'a'
I57184
sS's'
I18
sS'm'
(lp33611
I1
aI3
aI3
aI15
aI21
aI61
aI87
aI63
aI227
aI195
aI1095
aI1629
aI7787
aI5887
aI20855
aI30203
aI61973
aI30627
asS'd'
I16806
sa(dp33612
S'a'
I57193
sS's'
I18
sS'm'
(lp33613
I1
aI3
aI1
aI15
aI31
aI41
aI125
aI223
aI201
aI717
aI1309
aI595
aI5333
aI10585
aI32525
aI8597
aI92637
aI111073
asS'd'
I16807
sa(dp33614
S'a'
I57202
sS's'
I18
sS'm'
(lp33615
I1
aI3
aI5
aI3
aI21
aI29
aI39
aI105
aI275
aI515
aI503
aI79
aI6715
aI14203
aI14035
aI20871
aI122417
aI243167
asS'd'
I16808
sa(dp33616
S'a'
I57235
sS's'
I18
sS'm'
(lp33617
I1
aI3
aI7
aI5
aI29
aI41
aI3
aI89
aI165
aI879
aI773
aI3989
aI3945
aI4771
aI2809
aI59105
aI37177
aI193887
asS'd'
I16809
sa(dp33618
S'a'
I57237
sS's'
I18
sS'm'
(lp33619
I1
aI3
aI3
aI3
aI27
aI1
aI91
aI191
aI135
aI257
aI527
aI2971
aI7117
aI6013
aI8735
aI52363
aI110617
aI96959
asS'd'
I16810
sa(dp33620
S'a'
I57251
sS's'
I18
sS'm'
(lp33621
I1
aI3
aI7
aI9
aI3
aI63
aI67
aI67
aI231
aI23
aI1539
aI771
aI1485
aI4331
aI19231
aI50539
aI15081
aI75945
asS'd'
I16811
sa(dp33622
S'a'
I57289
sS's'
I18
sS'm'
(lp33623
I1
aI3
aI3
aI11
aI29
aI11
aI77
aI67
aI497
aI861
aI21
aI2939
aI2463
aI14435
aI27399
aI19733
aI118207
aI60909
asS'd'
I16812
sa(dp33624
S'a'
I57349
sS's'
I18
sS'm'
(lp33625
I1
aI1
aI5
aI5
aI1
aI11
aI117
aI55
aI485
aI877
aI1213
aI2231
aI2613
aI14027
aI18491
aI45431
aI113303
aI28457
asS'd'
I16813
sa(dp33626
S'a'
I57359
sS's'
I18
sS'm'
(lp33627
I1
aI3
aI7
aI1
aI13
aI49
aI77
aI59
aI455
aI251
aI1033
aI3451
aI7641
aI389
aI3987
aI62361
aI90125
aI94569
asS'd'
I16814
sa(dp33628
S'a'
I57374
sS's'
I18
sS'm'
(lp33629
I1
aI1
aI7
aI15
aI3
aI5
aI45
aI173
aI343
aI445
aI1871
aI2505
aI1385
aI2641
aI21299
aI35139
aI61781
aI101195
asS'd'
I16815
sa(dp33630
S'a'
I57377
sS's'
I18
sS'm'
(lp33631
I1
aI3
aI1
aI9
aI25
aI27
aI89
aI123
aI473
aI901
aI1513
aI2585
aI5641
aI13123
aI22653
aI32985
aI15763
aI9161
asS'd'
I16816
sa(dp33632
S'a'
I57387
sS's'
I18
sS'm'
(lp33633
I1
aI3
aI3
aI9
aI29
aI41
aI5
aI127
aI489
aI715
aI1981
aI3953
aI3557
aI10081
aI31913
aI52191
aI118727
aI4443
asS'd'
I16817
sa(dp33634
S'a'
I57415
sS's'
I18
sS'm'
(lp33635
I1
aI1
aI5
aI1
aI19
aI57
aI125
aI33
aI253
aI297
aI265
aI2249
aI6859
aI14971
aI3519
aI24783
aI127491
aI210441
asS'd'
I16818
sa(dp33636
S'a'
I57440
sS's'
I18
sS'm'
(lp33637
I1
aI1
aI7
aI7
aI31
aI1
aI47
aI175
aI305
aI933
aI679
aI317
aI7511
aI13219
aI9509
aI61183
aI58907
aI72905
asS'd'
I16819
sa(dp33638
S'a'
I57446
sS's'
I18
sS'm'
(lp33639
I1
aI1
aI1
aI7
aI13
aI49
aI75
aI85
aI341
aI911
aI1217
aI3631
aI1849
aI9715
aI23193
aI947
aI106647
aI180455
asS'd'
I16820
sa(dp33640
S'a'
I57450
sS's'
I18
sS'm'
(lp33641
I1
aI1
aI7
aI7
aI1
aI49
aI91
aI195
aI329
aI771
aI607
aI1707
aI2723
aI291
aI21393
aI6549
aI31645
aI151431
asS'd'
I16821
sa(dp33642
S'a'
I57469
sS's'
I18
sS'm'
(lp33643
I1
aI3
aI7
aI5
aI17
aI57
aI7
aI231
aI247
aI217
aI1729
aI3231
aI7515
aI15341
aI18681
aI21733
aI28723
aI228187
asS'd'
I16822
sa(dp33644
S'a'
I57491
sS's'
I18
sS'm'
(lp33645
I1
aI1
aI5
aI9
aI5
aI19
aI121
aI251
aI43
aI951
aI957
aI173
aI4863
aI5027
aI6781
aI29421
aI4877
aI47749
asS'd'
I16823
sa(dp33646
S'a'
I57503
sS's'
I18
sS'm'
(lp33647
I1
aI3
aI7
aI7
aI11
aI33
aI107
aI233
aI329
aI589
aI869
aI913
aI7687
aI13223
aI27577
aI24379
aI13037
aI214713
asS'd'
I16824
sa(dp33648
S'a'
I57507
sS's'
I18
sS'm'
(lp33649
I1
aI3
aI7
aI13
aI1
aI13
aI121
aI103
aI387
aI193
aI543
aI3085
aI4323
aI9885
aI24499
aI34985
aI45763
aI13107
asS'd'
I16825
sa(dp33650
S'a'
I57542
sS's'
I18
sS'm'
(lp33651
I1
aI1
aI3
aI15
aI25
aI63
aI85
aI41
aI457
aI779
aI1199
aI2235
aI309
aI2549
aI3341
aI36265
aI17873
aI32361
asS'd'
I16826
sa(dp33652
S'a'
I57569
sS's'
I18
sS'm'
(lp33653
I1
aI3
aI3
aI3
aI15
aI31
aI11
aI57
aI499
aI415
aI1625
aI1195
aI6863
aI6073
aI25083
aI57705
aI76203
aI130993
asS'd'
I16827
sa(dp33654
S'a'
I57599
sS's'
I18
sS'm'
(lp33655
I1
aI3
aI5
aI5
aI21
aI13
aI43
aI161
aI255
aI31
aI1901
aI3325
aI3209
aI9809
aI8227
aI9005
aI57263
aI95095
asS'd'
I16828
sa(dp33656
S'a'
I57601
sS's'
I18
sS'm'
(lp33657
I1
aI1
aI3
aI15
aI13
aI33
aI5
aI123
aI291
aI579
aI1747
aI3319
aI7351
aI1679
aI11365
aI26909
aI74445
aI139017
asS'd'
I16829
sa(dp33658
S'a'
I57607
sS's'
I18
sS'm'
(lp33659
I1
aI1
aI3
aI13
aI17
aI39
aI1
aI253
aI487
aI935
aI1711
aI1397
aI503
aI7817
aI28509
aI20665
aI78551
aI204319
asS'd'
I16830
sa(dp33660
S'a'
I57608
sS's'
I18
sS'm'
(lp33661
I1
aI1
aI3
aI1
aI5
aI39
aI123
aI105
aI305
aI77
aI63
aI3285
aI7463
aI11199
aI647
aI37757
aI91083
aI108325
asS'd'
I16831
sa(dp33662
S'a'
I57625
sS's'
I18
sS'm'
(lp33663
I1
aI1
aI7
aI5
aI9
aI49
aI121
aI155
aI389
aI119
aI1327
aI3583
aI7715
aI2705
aI20047
aI19151
aI101455
aI205263
asS'd'
I16832
sa(dp33664
S'a'
I57644
sS's'
I18
sS'm'
(lp33665
I1
aI1
aI5
aI7
aI31
aI23
aI13
aI109
aI103
aI41
aI433
aI3609
aI4973
aI11481
aI8381
aI4725
aI113633
aI134651
asS'd'
I16833
sa(dp33666
S'a'
I57662
sS's'
I18
sS'm'
(lp33667
I1
aI3
aI7
aI7
aI25
aI25
aI107
aI189
aI89
aI625
aI187
aI2185
aI713
aI10107
aI11139
aI63681
aI97005
aI79329
asS'd'
I16834
sa(dp33668
S'a'
I57664
sS's'
I18
sS'm'
(lp33669
I1
aI3
aI1
aI11
aI3
aI41
aI43
aI161
aI337
aI955
aI1035
aI451
aI5989
aI3593
aI18087
aI22667
aI110213
aI128545
asS'd'
I16835
sa(dp33670
S'a'
I57674
sS's'
I18
sS'm'
(lp33671
I1
aI1
aI5
aI1
aI25
aI31
aI95
aI113
aI205
aI565
aI557
aI3885
aI7163
aI10703
aI27159
aI11395
aI117459
aI52439
asS'd'
I16836
sa(dp33672
S'a'
I57698
sS's'
I18
sS'm'
(lp33673
I1
aI1
aI1
aI5
aI27
aI31
aI39
aI61
aI323
aI983
aI1361
aI2387
aI5401
aI8287
aI17855
aI49783
aI65327
aI202861
asS'd'
I16837
sa(dp33674
S'a'
I57700
sS's'
I18
sS'm'
(lp33675
I1
aI3
aI5
aI3
aI31
aI39
aI105
aI113
aI183
aI311
aI667
aI945
aI3677
aI14623
aI27907
aI16673
aI77899
aI182863
asS'd'
I16838
sa(dp33676
S'a'
I57709
sS's'
I18
sS'm'
(lp33677
I1
aI1
aI5
aI3
aI17
aI27
aI99
aI93
aI81
aI805
aI1799
aI2855
aI6859
aI3917
aI26177
aI22307
aI59213
aI210123
asS'd'
I16839
sa(dp33678
S'a'
I57724
sS's'
I18
sS'm'
(lp33679
I1
aI3
aI5
aI1
aI19
aI37
aI51
aI65
aI495
aI229
aI229
aI1283
aI2967
aI5329
aI24339
aI58739
aI23145
aI7033
asS'd'
I16840
sa(dp33680
S'a'
I57728
sS's'
I18
sS'm'
(lp33681
I1
aI3
aI3
aI15
aI11
aI51
aI121
aI41
aI75
aI845
aI1771
aI3625
aI6137
aI3463
aI11767
aI45181
aI70907
aI42771
asS'd'
I16841
sa(dp33682
S'a'
I57740
sS's'
I18
sS'm'
(lp33683
I1
aI3
aI7
aI9
aI15
aI25
aI55
aI219
aI265
aI655
aI167
aI1247
aI5409
aI5623
aI21045
aI12333
aI25799
aI218601
asS'd'
I16842
sa(dp33684
S'a'
I57745
sS's'
I18
sS'm'
(lp33685
I1
aI3
aI3
aI13
aI31
aI39
aI77
aI155
aI471
aI969
aI755
aI2745
aI3057
aI3621
aI32423
aI48687
aI9409
aI90997
asS'd'
I16843
sa(dp33686
S'a'
I57751
sS's'
I18
sS'm'
(lp33687
I1
aI1
aI3
aI15
aI27
aI1
aI77
aI231
aI147
aI235
aI2027
aI4045
aI7431
aI14655
aI6361
aI43155
aI9839
aI161713
asS'd'
I16844
sa(dp33688
S'a'
I57774
sS's'
I18
sS'm'
(lp33689
I1
aI3
aI7
aI5
aI25
aI19
aI25
aI75
aI415
aI931
aI457
aI3691
aI687
aI4849
aI15469
aI42871
aI37949
aI74163
asS'd'
I16845
sa(dp33690
S'a'
I57782
sS's'
I18
sS'm'
(lp33691
I1
aI3
aI5
aI9
aI17
aI19
aI29
aI117
aI387
aI1021
aI1159
aI2467
aI2585
aI2563
aI9155
aI44763
aI93319
aI6321
asS'd'
I16846
sa(dp33692
S'a'
I57796
sS's'
I18
sS'm'
(lp33693
I1
aI3
aI5
aI7
aI25
aI33
aI127
aI175
aI143
aI705
aI539
aI2563
aI945
aI11369
aI19971
aI19019
aI116195
aI84121
asS'd'
I16847
sa(dp33694
S'a'
I57803
sS's'
I18
sS'm'
(lp33695
I1
aI3
aI7
aI7
aI5
aI55
aI29
aI1
aI419
aI715
aI1275
aI2983
aI7853
aI12245
aI32109
aI27371
aI123547
aI82723
asS'd'
I16848
sa(dp33696
S'a'
I57823
sS's'
I18
sS'm'
(lp33697
I1
aI1
aI1
aI13
aI3
aI29
aI31
aI213
aI195
aI609
aI1465
aI1711
aI6747
aI13309
aI1131
aI3151
aI48779
aI91571
asS'd'
I16849
sa(dp33698
S'a'
I57863
sS's'
I18
sS'm'
(lp33699
I1
aI1
aI5
aI3
aI17
aI7
aI103
aI7
aI217
aI87
aI1641
aI833
aI4551
aI14205
aI15119
aI6711
aI111273
aI200545
asS'd'
I16850
sa(dp33700
S'a'
I57894
sS's'
I18
sS'm'
(lp33701
I1
aI3
aI1
aI5
aI3
aI39
aI99
aI15
aI433
aI895
aI165
aI4049
aI3183
aI4385
aI24695
aI40009
aI67151
aI156643
asS'd'
I16851
sa(dp33702
S'a'
I57925
sS's'
I18
sS'm'
(lp33703
I1
aI1
aI7
aI3
aI29
aI9
aI15
aI27
aI109
aI1019
aI327
aI2837
aI5297
aI12455
aI2355
aI37703
aI122995
aI177871
asS'd'
I16852
sa(dp33704
S'a'
I57971
sS's'
I18
sS'm'
(lp33705
I1
aI1
aI5
aI15
aI29
aI5
aI121
aI117
aI31
aI155
aI1027
aI1105
aI8057
aI8677
aI9523
aI3019
aI98801
aI15539
asS'd'
I16853
sa(dp33706
S'a'
I58013
sS's'
I18
sS'm'
(lp33707
I1
aI3
aI7
aI3
aI1
aI1
aI37
aI67
aI471
aI317
aI1571
aI2801
aI7383
aI4339
aI8095
aI45685
aI95885
aI39577
asS'd'
I16854
sa(dp33708
S'a'
I58020
sS's'
I18
sS'm'
(lp33709
I1
aI3
aI7
aI13
aI17
aI13
aI91
aI79
aI49
aI321
aI1235
aI311
aI129
aI6537
aI6643
aI25813
aI48251
aI138823
asS'd'
I16855
sa(dp33710
S'a'
I58032
sS's'
I18
sS'm'
(lp33711
I1
aI1
aI5
aI3
aI21
aI19
aI67
aI61
aI153
aI611
aI1819
aI3755
aI5959
aI3419
aI6117
aI1159
aI68925
aI146199
asS'd'
I16856
sa(dp33712
S'a'
I58038
sS's'
I18
sS'm'
(lp33713
I1
aI1
aI7
aI9
aI23
aI3
aI7
aI13
aI429
aI463
aI653
aI3461
aI6337
aI4511
aI18097
aI44837
aI99845
aI37101
asS'd'
I16857
sa(dp33714
S'a'
I58061
sS's'
I18
sS'm'
(lp33715
I1
aI3
aI5
aI13
aI9
aI5
aI123
aI199
aI83
aI409
aI1391
aI1567
aI7327
aI8173
aI30971
aI18241
aI7755
aI185375
asS'd'
I16858
sa(dp33716
S'a'
I58069
sS's'
I18
sS'm'
(lp33717
I1
aI3
aI1
aI7
aI19
aI51
aI51
aI23
aI85
aI923
aI1969
aI2329
aI7343
aI12489
aI16135
aI64783
aI117063
aI141071
asS'd'
I16859
sa(dp33718
S'a'
I58080
sS's'
I18
sS'm'
(lp33719
I1
aI3
aI1
aI3
aI23
aI29
aI5
aI77
aI207
aI351
aI367
aI2097
aI2639
aI9255
aI21971
aI64167
aI98069
aI81153
asS'd'
I16860
sa(dp33720
S'a'
I58089
sS's'
I18
sS'm'
(lp33721
I1
aI1
aI7
aI15
aI27
aI1
aI83
aI255
aI47
aI935
aI567
aI3573
aI3629
aI5833
aI483
aI1001
aI9337
aI119847
asS'd'
I16861
sa(dp33722
S'a'
I58107
sS's'
I18
sS'm'
(lp33723
I1
aI3
aI7
aI11
aI31
aI53
aI25
aI35
aI463
aI51
aI401
aI3279
aI7709
aI11265
aI17905
aI40423
aI26277
aI43355
asS'd'
I16862
sa(dp33724
S'a'
I58121
sS's'
I18
sS'm'
(lp33725
I1
aI1
aI7
aI5
aI9
aI15
aI73
aI217
aI239
aI405
aI1651
aI2131
aI6791
aI11241
aI21717
aI7393
aI77251
aI28131
asS'd'
I16863
sa(dp33726
S'a'
I58130
sS's'
I18
sS'm'
(lp33727
I1
aI3
aI5
aI3
aI13
aI43
aI115
aI159
aI215
aI811
aI1349
aI2941
aI2073
aI1821
aI6891
aI17285
aI72027
aI137849
asS'd'
I16864
sa(dp33728
S'a'
I58146
sS's'
I18
sS'm'
(lp33729
I1
aI3
aI3
aI3
aI15
aI11
aI29
aI53
aI307
aI409
aI1069
aI3713
aI3205
aI6185
aI2565
aI14973
aI46149
aI162527
asS'd'
I16865
sa(dp33730
S'a'
I58190
sS's'
I18
sS'm'
(lp33731
I1
aI1
aI3
aI15
aI21
aI39
aI61
aI209
aI211
aI123
aI697
aI2285
aI859
aI2501
aI5847
aI56449
aI106575
aI261069
asS'd'
I16866
sa(dp33732
S'a'
I58195
sS's'
I18
sS'm'
(lp33733
I1
aI3
aI3
aI5
aI25
aI21
aI39
aI131
aI189
aI747
aI1499
aI1865
aI3369
aI9161
aI12543
aI63155
aI70083
aI69441
asS'd'
I16867
sa(dp33734
S'a'
I58202
sS's'
I18
sS'm'
(lp33735
I1
aI3
aI1
aI15
aI31
aI43
aI127
aI57
aI169
aI109
aI979
aI1399
aI3065
aI5865
aI16891
aI56003
aI14319
aI94109
asS'd'
I16868
sa(dp33736
S'a'
I58237
sS's'
I18
sS'm'
(lp33737
I1
aI1
aI1
aI13
aI23
aI57
aI13
aI239
aI139
aI41
aI1959
aI429
aI209
aI543
aI21297
aI15343
aI16521
aI52305
asS'd'
I16869
sa(dp33738
S'a'
I58253
sS's'
I18
sS'm'
(lp33739
I1
aI1
aI7
aI1
aI17
aI1
aI115
aI139
aI93
aI123
aI867
aI3257
aI8135
aI12089
aI1503
aI33287
aI79283
aI151419
asS'd'
I16870
sa(dp33740
S'a'
I58299
sS's'
I18
sS'm'
(lp33741
I1
aI3
aI7
aI7
aI27
aI17
aI15
aI253
aI89
aI959
aI597
aI2193
aI3505
aI13865
aI2179
aI58711
aI114615
aI15227
asS'd'
I16871
sa(dp33742
S'a'
I58302
sS's'
I18
sS'm'
(lp33743
I1
aI3
aI7
aI5
aI1
aI5
aI105
aI241
aI361
aI229
aI1069
aI3815
aI1409
aI4909
aI31785
aI46555
aI123523
aI53259
asS'd'
I16872
sa(dp33744
S'a'
I58327
sS's'
I18
sS'm'
(lp33745
I1
aI1
aI5
aI5
aI15
aI49
aI13
aI195
aI467
aI285
aI1405
aI3011
aI2069
aI8331
aI13953
aI31107
aI46581
aI154615
asS'd'
I16873
sa(dp33746
S'a'
I58328
sS's'
I18
sS'm'
(lp33747
I1
aI3
aI5
aI7
aI21
aI23
aI17
aI17
aI345
aI369
aI1521
aI3755
aI2165
aI15387
aI2851
aI11115
aI60483
aI236049
asS'd'
I16874
sa(dp33748
S'a'
I58364
sS's'
I18
sS'm'
(lp33749
I1
aI3
aI3
aI1
aI5
aI41
aI53
aI239
aI127
aI237
aI609
aI927
aI3787
aI5059
aI1865
aI52991
aI56229
aI102093
asS'd'
I16875
sa(dp33750
S'a'
I58367
sS's'
I18
sS'm'
(lp33751
I1
aI1
aI7
aI5
aI23
aI7
aI15
aI199
aI325
aI695
aI1525
aI3435
aI3997
aI11577
aI22985
aI57713
aI94309
aI218433
asS'd'
I16876
sa(dp33752
S'a'
I58375
sS's'
I18
sS'm'
(lp33753
I1
aI3
aI3
aI5
aI25
aI25
aI61
aI99
aI237
aI447
aI1905
aI783
aI5239
aI11415
aI16833
aI27815
aI115539
aI161111
asS'd'
I16877
sa(dp33754
S'a'
I58394
sS's'
I18
sS'm'
(lp33755
I1
aI1
aI5
aI9
aI31
aI49
aI55
aI199
aI159
aI751
aI849
aI1045
aI5485
aI8883
aI8549
aI11735
aI35983
aI161067
asS'd'
I16878
sa(dp33756
S'a'
I58405
sS's'
I18
sS'm'
(lp33757
I1
aI1
aI5
aI9
aI23
aI51
aI79
aI171
aI87
aI493
aI1911
aI3867
aI3435
aI493
aI16639
aI64085
aI97797
aI244959
asS'd'
I16879
sa(dp33758
S'a'
I58417
sS's'
I18
sS'm'
(lp33759
I1
aI1
aI5
aI11
aI29
aI33
aI15
aI107
aI283
aI545
aI1995
aI995
aI7181
aI3581
aI8621
aI42391
aI117997
aI397
asS'd'
I16880
sa(dp33760
S'a'
I58424
sS's'
I18
sS'm'
(lp33761
I1
aI3
aI7
aI13
aI31
aI25
aI91
aI75
aI123
aI451
aI1023
aI375
aI4505
aI13235
aI8913
aI34389
aI77385
aI168659
asS'd'
I16881
sa(dp33762
S'a'
I58430
sS's'
I18
sS'm'
(lp33763
I1
aI1
aI3
aI3
aI3
aI3
aI85
aI143
aI173
aI709
aI1313
aI593
aI6931
aI14609
aI13803
aI30305
aI109089
aI11473
asS'd'
I16882
sa(dp33764
S'a'
I58452
sS's'
I18
sS'm'
(lp33765
I1
aI3
aI7
aI5
aI25
aI45
aI25
aI223
aI407
aI597
aI83
aI2543
aI3823
aI13959
aI9089
aI28325
aI29237
aI57147
asS'd'
I16883
sa(dp33766
S'a'
I58466
sS's'
I18
sS'm'
(lp33767
I1
aI1
aI1
aI3
aI25
aI53
aI57
aI255
aI231
aI361
aI109
aI113
aI6091
aI13043
aI28399
aI29111
aI57987
aI137709
asS'd'
I16884
sa(dp33768
S'a'
I58468
sS's'
I18
sS'm'
(lp33769
I1
aI1
aI1
aI5
aI11
aI25
aI53
aI141
aI275
aI237
aI1427
aI1691
aI6043
aI8951
aI10683
aI17477
aI117645
aI89007
asS'd'
I16885
sa(dp33770
S'a'
I58495
sS's'
I18
sS'm'
(lp33771
I1
aI3
aI3
aI13
aI7
aI23
aI73
aI213
aI285
aI667
aI1765
aI1545
aI1401
aI12483
aI6349
aI47205
aI25791
aI16749
asS'd'
I16886
sa(dp33772
S'a'
I58501
sS's'
I18
sS'm'
(lp33773
I1
aI1
aI1
aI15
aI31
aI45
aI105
aI249
aI385
aI607
aI723
aI745
aI7037
aI15735
aI3637
aI29013
aI127315
aI165507
asS'd'
I16887
sa(dp33774
S'a'
I58544
sS's'
I18
sS'm'
(lp33775
I1
aI1
aI7
aI5
aI21
aI63
aI95
aI247
aI161
aI839
aI939
aI931
aI4277
aI7363
aI8289
aI55183
aI122413
aI152997
asS'd'
I16888
sa(dp33776
S'a'
I58571
sS's'
I18
sS'm'
(lp33777
I1
aI3
aI7
aI11
aI15
aI59
aI91
aI5
aI209
aI31
aI1581
aI979
aI6289
aI11443
aI26641
aI20183
aI106907
aI128647
asS'd'
I16889
sa(dp33778
S'a'
I58609
sS's'
I18
sS'm'
(lp33779
I1
aI1
aI3
aI15
aI21
aI33
aI117
aI89
aI457
aI405
aI1971
aI2211
aI4379
aI16189
aI7933
aI39351
aI79813
aI56373
asS'd'
I16890
sa(dp33780
S'a'
I58610
sS's'
I18
sS'm'
(lp33781
I1
aI3
aI3
aI9
aI5
aI9
aI93
aI75
aI55
aI271
aI321
aI3143
aI3893
aI2601
aI26169
aI35179
aI43063
aI156635
asS'd'
I16891
sa(dp33782
S'a'
I58616
sS's'
I18
sS'm'
(lp33783
I1
aI3
aI3
aI11
aI29
aI37
aI95
aI249
aI221
aI965
aI423
aI1637
aI4663
aI14839
aI16757
aI4261
aI128453
aI165593
asS'd'
I16892
sa(dp33784
S'a'
I58619
sS's'
I18
sS'm'
(lp33785
I1
aI3
aI3
aI7
aI1
aI55
aI31
aI235
aI447
aI839
aI721
aI1125
aI6503
aI4019
aI23351
aI37057
aI96103
aI143805
asS'd'
I16893
sa(dp33786
S'a'
I58641
sS's'
I18
sS'm'
(lp33787
I1
aI3
aI7
aI5
aI31
aI39
aI7
aI157
aI469
aI719
aI1613
aI395
aI8133
aI9753
aI17323
aI13849
aI45409
aI7601
asS'd'
I16894
sa(dp33788
S'a'
I58642
sS's'
I18
sS'm'
(lp33789
I1
aI3
aI7
aI7
aI31
aI37
aI89
aI215
aI453
aI659
aI605
aI3325
aI987
aI4611
aI29667
aI23229
aI4201
aI229675
asS'd'
I16895
sa(dp33790
S'a'
I58648
sS's'
I18
sS'm'
(lp33791
I1
aI1
aI5
aI5
aI3
aI3
aI21
aI249
aI377
aI343
aI1751
aI891
aI5275
aI14853
aI32703
aI51001
aI6759
aI162991
asS'd'
I16896
sa(dp33792
S'a'
I58660
sS's'
I18
sS'm'
(lp33793
I1
aI3
aI1
aI13
aI11
aI21
aI55
aI17
aI495
aI481
aI1817
aI919
aI2495
aI16367
aI3343
aI16997
aI83437
aI127791
asS'd'
I16897
sa(dp33794
S'a'
I58675
sS's'
I18
sS'm'
(lp33795
I1
aI3
aI1
aI1
aI5
aI57
aI65
aI223
aI33
aI491
aI1953
aI1521
aI4903
aI5007
aI14583
aI17321
aI82231
aI206299
asS'd'
I16898
sa(dp33796
S'a'
I58678
sS's'
I18
sS'm'
(lp33797
I1
aI3
aI7
aI11
aI21
aI45
aI55
aI141
aI185
aI379
aI851
aI885
aI3385
aI10311
aI701
aI2983
aI71045
aI171525
asS'd'
I16899
sa(dp33798
S'a'
I58690
sS's'
I18
sS'm'
(lp33799
I1
aI3
aI7
aI3
aI29
aI1
aI53
aI139
aI7
aI985
aI291
aI3949
aI1163
aI14637
aI363
aI59679
aI121571
aI121081
asS'd'
I16900
sa(dp33800
S'a'
I58735
sS's'
I18
sS'm'
(lp33801
I1
aI3
aI7
aI1
aI31
aI1
aI111
aI19
aI421
aI917
aI1529
aI1361
aI4461
aI12457
aI9791
aI19985
aI77283
aI117059
asS'd'
I16901
sa(dp33802
S'a'
I58760
sS's'
I18
sS'm'
(lp33803
I1
aI3
aI1
aI5
aI7
aI55
aI93
aI243
aI477
aI193
aI1983
aI489
aI3735
aI1391
aI24035
aI36395
aI49101
aI175861
asS'd'
I16902
sa(dp33804
S'a'
I58766
sS's'
I18
sS'm'
(lp33805
I1
aI1
aI1
aI11
aI3
aI25
aI69
aI167
aI351
aI193
aI1299
aI617
aI7455
aI2545
aI18359
aI9951
aI119513
aI128139
asS'd'
I16903
sa(dp33806
S'a'
I58799
sS's'
I18
sS'm'
(lp33807
I1
aI3
aI3
aI7
aI5
aI23
aI101
aI47
aI385
aI591
aI345
aI3501
aI531
aI3277
aI28945
aI18695
aI58587
aI87221
asS'd'
I16904
sa(dp33808
S'a'
I58825
sS's'
I18
sS'm'
(lp33809
I1
aI1
aI3
aI3
aI29
aI47
aI5
aI91
aI365
aI1
aI2015
aI323
aI1601
aI10615
aI28975
aI60263
aI4813
aI143351
asS'd'
I16905
sa(dp33810
S'a'
I58836
sS's'
I18
sS'm'
(lp33811
I1
aI3
aI1
aI7
aI25
aI43
aI65
aI211
aI91
aI759
aI985
aI3675
aI5701
aI4373
aI27781
aI51949
aI40667
aI102665
asS'd'
I16906
sa(dp33812
S'a'
I58855
sS's'
I18
sS'm'
(lp33813
I1
aI3
aI3
aI5
aI3
aI43
aI91
aI33
aI247
aI593
aI849
aI1955
aI7769
aI2307
aI2877
aI26037
aI28907
aI211021
asS'd'
I16907
sa(dp33814
S'a'
I58864
sS's'
I18
sS'm'
(lp33815
I1
aI3
aI5
aI15
aI29
aI29
aI85
aI97
aI99
aI979
aI2033
aI1415
aI2955
aI15733
aI5567
aI6241
aI100195
aI89077
asS'd'
I16908
sa(dp33816
S'a'
I58910
sS's'
I18
sS'm'
(lp33817
I1
aI3
aI7
aI13
aI13
aI19
aI121
aI211
aI381
aI73
aI1131
aI1881
aI1693
aI7873
aI27557
aI201
aI24997
aI202471
asS'd'
I16909
sa(dp33818
S'a'
I58937
sS's'
I18
sS'm'
(lp33819
I1
aI3
aI1
aI15
aI15
aI33
aI11
aI99
aI479
aI271
aI1873
aI1117
aI3559
aI6605
aI15995
aI44805
aI12465
aI71933
asS'd'
I16910
sa(dp33820
S'a'
I58943
sS's'
I18
sS'm'
(lp33821
I1
aI3
aI5
aI3
aI19
aI61
aI15
aI55
aI423
aI431
aI1321
aI3345
aI1633
aI4587
aI24909
aI54985
aI31831
aI181083
asS'd'
I16911
sa(dp33822
S'a'
I58952
sS's'
I18
sS'm'
(lp33823
I1
aI1
aI3
aI5
aI29
aI43
aI49
aI205
aI415
aI907
aI1651
aI57
aI3043
aI10763
aI16255
aI9567
aI59453
aI135637
asS'd'
I16912
sa(dp33824
S'a'
I58965
sS's'
I18
sS'm'
(lp33825
I1
aI3
aI3
aI1
aI17
aI11
aI29
aI33
aI293
aI203
aI1687
aI1565
aI6131
aI5435
aI29023
aI28425
aI102151
aI251913
asS'd'
I16913
sa(dp33826
S'a'
I58988
sS's'
I18
sS'm'
(lp33827
I1
aI1
aI5
aI7
aI9
aI9
aI43
aI191
aI269
aI681
aI607
aI3045
aI2799
aI14919
aI8083
aI57781
aI19345
aI49365
asS'd'
I16914
sa(dp33828
S'a'
I58994
sS's'
I18
sS'm'
(lp33829
I1
aI1
aI5
aI13
aI11
aI53
aI67
aI127
aI117
aI395
aI575
aI1651
aI2601
aI15019
aI21413
aI34433
aI66847
aI84159
asS'd'
I16915
sa(dp33830
S'a'
I58999
sS's'
I18
sS'm'
(lp33831
I1
aI3
aI5
aI5
aI15
aI59
aI33
aI41
aI301
aI699
aI1479
aI2285
aI1813
aI2459
aI4775
aI53213
aI26039
aI223155
asS'd'
I16916
sa(dp33832
S'a'
I59006
sS's'
I18
sS'm'
(lp33833
I1
aI3
aI1
aI15
aI17
aI57
aI5
aI211
aI357
aI175
aI945
aI3625
aI3943
aI12871
aI26805
aI29305
aI8839
aI107837
asS'd'
I16917
sa(dp33834
S'a'
I59029
sS's'
I18
sS'm'
(lp33835
I1
aI3
aI5
aI15
aI21
aI41
aI105
aI229
aI265
aI777
aI2047
aI767
aI2901
aI8873
aI7631
aI18545
aI86697
aI252965
asS'd'
I16918
sa(dp33836
S'a'
I59033
sS's'
I18
sS'm'
(lp33837
I1
aI1
aI5
aI11
aI31
aI63
aI115
aI119
aI271
aI921
aI1221
aI3341
aI6083
aI4293
aI28581
aI57323
aI33889
aI112577
asS'd'
I16919
sa(dp33838
S'a'
I59069
sS's'
I18
sS'm'
(lp33839
I1
aI1
aI5
aI5
aI31
aI21
aI119
aI93
aI287
aI139
aI451
aI2535
aI3925
aI10671
aI21279
aI55071
aI76127
aI248203
asS'd'
I16920
sa(dp33840
S'a'
I59096
sS's'
I18
sS'm'
(lp33841
I1
aI3
aI7
aI11
aI19
aI61
aI61
aI53
aI203
aI181
aI963
aI3581
aI519
aI14679
aI7717
aI31981
aI128709
aI197269
asS'd'
I16921
sa(dp33842
S'a'
I59106
sS's'
I18
sS'm'
(lp33843
I1
aI1
aI1
aI13
aI25
aI23
aI89
aI95
aI221
aI803
aI1433
aI3617
aI3217
aI2033
aI7859
aI14279
aI107239
aI5139
asS'd'
I16922
sa(dp33844
S'a'
I59123
sS's'
I18
sS'm'
(lp33845
I1
aI3
aI7
aI3
aI29
aI41
aI87
aI21
aI71
aI959
aI1149
aI2961
aI7471
aI11665
aI16037
aI5791
aI110155
aI35365
asS'd'
I16923
sa(dp33846
S'a'
I59130
sS's'
I18
sS'm'
(lp33847
I1
aI1
aI1
aI11
aI21
aI49
aI101
aI45
aI311
aI529
aI1301
aI1377
aI983
aI3937
aI6967
aI8413
aI33511
aI9617
asS'd'
I16924
sa(dp33848
S'a'
I59152
sS's'
I18
sS'm'
(lp33849
I1
aI3
aI3
aI5
aI15
aI41
aI107
aI49
aI409
aI537
aI289
aI3351
aI5307
aI16221
aI907
aI39847
aI61579
aI161487
asS'd'
I16925
sa(dp33850
S'a'
I59162
sS's'
I18
sS'm'
(lp33851
I1
aI1
aI3
aI11
aI5
aI49
aI71
aI107
aI431
aI469
aI453
aI1367
aI7811
aI10485
aI3861
aI62797
aI82025
aI253785
asS'd'
I16926
sa(dp33852
S'a'
I59180
sS's'
I18
sS'm'
(lp33853
I1
aI1
aI3
aI3
aI27
aI19
aI89
aI13
aI445
aI915
aI1259
aI1423
aI3987
aI3661
aI18183
aI18521
aI18831
aI191447
asS'd'
I16927
sa(dp33854
S'a'
I59183
sS's'
I18
sS'm'
(lp33855
I1
aI1
aI7
aI5
aI13
aI15
aI9
aI89
aI129
aI949
aI1733
aI245
aI6815
aI8477
aI1273
aI34737
aI33027
aI191415
asS'd'
I16928
sa(dp33856
S'a'
I59195
sS's'
I18
sS'm'
(lp33857
I1
aI1
aI7
aI15
aI25
aI63
aI83
aI195
aI319
aI987
aI1395
aI3559
aI6287
aI5139
aI25967
aI48711
aI58467
aI110983
asS'd'
I16929
sa(dp33858
S'a'
I59205
sS's'
I18
sS'm'
(lp33859
I1
aI1
aI3
aI9
aI5
aI3
aI35
aI171
aI15
aI883
aI915
aI2451
aI871
aI11741
aI32715
aI33475
aI81711
aI259157
asS'd'
I16930
sa(dp33860
S'a'
I59210
sS's'
I18
sS'm'
(lp33861
I1
aI1
aI7
aI13
aI23
aI63
aI33
aI11
aI117
aI351
aI1701
aI671
aI6753
aI5
aI9477
aI54701
aI65507
aI242621
asS'd'
I16931
sa(dp33862
S'a'
I59217
sS's'
I18
sS'm'
(lp33863
I1
aI3
aI7
aI11
aI21
aI37
aI127
aI143
aI369
aI819
aI1369
aI93
aI7009
aI3773
aI30153
aI30181
aI120783
aI137857
asS'd'
I16932
sa(dp33864
S'a'
I59218
sS's'
I18
sS'm'
(lp33865
I1
aI3
aI3
aI7
aI27
aI61
aI15
aI141
aI67
aI815
aI1449
aI1129
aI4703
aI3811
aI3067
aI61697
aI8881
aI110957
asS'd'
I16933
sa(dp33866
S'a'
I59236
sS's'
I18
sS'm'
(lp33867
I1
aI1
aI7
aI13
aI31
aI21
aI59
aI75
aI335
aI851
aI503
aI251
aI4869
aI11789
aI30871
aI14641
aI19319
aI156843
asS'd'
I16934
sa(dp33868
S'a'
I59267
sS's'
I18
sS'm'
(lp33869
I1
aI3
aI5
aI5
aI9
aI41
aI11
aI67
aI231
aI945
aI37
aI2925
aI5723
aI9053
aI13477
aI59735
aI75181
aI60335
asS'd'
I16935
sa(dp33870
S'a'
I59298
sS's'
I18
sS'm'
(lp33871
I1
aI1
aI5
aI1
aI13
aI39
aI81
aI43
aI363
aI611
aI1661
aI3833
aI7387
aI10531
aI21319
aI55579
aI102705
aI103009
asS'd'
I16936
sa(dp33872
S'a'
I59318
sS's'
I18
sS'm'
(lp33873
I1
aI3
aI1
aI7
aI23
aI25
aI67
aI179
aI327
aI401
aI1693
aI1453
aI4773
aI6363
aI27169
aI49747
aI29055
aI49145
asS'd'
I16937
sa(dp33874
S'a'
I59321
sS's'
I18
sS'm'
(lp33875
I1
aI3
aI5
aI7
aI13
aI47
aI5
aI175
aI369
aI921
aI507
aI113
aI6069
aI10919
aI11099
aI19795
aI95819
aI52419
asS'd'
I16938
sa(dp33876
S'a'
I59327
sS's'
I18
sS'm'
(lp33877
I1
aI1
aI1
aI5
aI5
aI53
aI93
aI47
aI75
aI837
aI109
aI3691
aI6961
aI10715
aI14269
aI63791
aI1941
aI136899
asS'd'
I16939
sa(dp33878
S'a'
I59354
sS's'
I18
sS'm'
(lp33879
I1
aI1
aI1
aI3
aI1
aI63
aI57
aI117
aI157
aI327
aI879
aI2411
aI3987
aI15393
aI8503
aI29829
aI77795
aI121307
asS'd'
I16940
sa(dp33880
S'a'
I59363
sS's'
I18
sS'm'
(lp33881
I1
aI3
aI5
aI1
aI25
aI5
aI47
aI45
aI433
aI121
aI607
aI1233
aI6433
aI3031
aI16369
aI58589
aI79357
aI151353
asS'd'
I16941
sa(dp33882
S'a'
I59377
sS's'
I18
sS'm'
(lp33883
I1
aI1
aI1
aI1
aI9
aI15
aI77
aI163
aI225
aI445
aI1479
aI1267
aI2571
aI2661
aI21489
aI5433
aI123969
aI191967
asS'd'
I16942
sa(dp33884
S'a'
I59389
sS's'
I18
sS'm'
(lp33885
I1
aI1
aI7
aI1
aI9
aI49
aI17
aI19
aI449
aI113
aI1289
aI2335
aI3309
aI2595
aI17819
aI18481
aI86605
aI125911
asS'd'
I16943
sa(dp33886
S'a'
I59403
sS's'
I18
sS'm'
(lp33887
I1
aI1
aI5
aI11
aI11
aI23
aI65
aI147
aI257
aI625
aI1901
aI913
aI5711
aI8159
aI16237
aI25133
aI100059
aI11395
asS'd'
I16944
sa(dp33888
S'a'
I59420
sS's'
I18
sS'm'
(lp33889
I1
aI3
aI7
aI13
aI5
aI33
aI89
aI189
aI171
aI185
aI751
aI2915
aI5025
aI15981
aI14853
aI12229
aI52829
aI59953
asS'd'
I16945
sa(dp33890
S'a'
I59444
sS's'
I18
sS'm'
(lp33891
I1
aI3
aI1
aI13
aI3
aI37
aI15
aI87
aI463
aI655
aI1927
aI2705
aI1885
aI14801
aI3491
aI52835
aI81761
aI90273
asS'd'
I16946
sa(dp33892
S'a'
I59471
sS's'
I18
sS'm'
(lp33893
I1
aI1
aI7
aI3
aI5
aI15
aI29
aI255
aI199
aI225
aI647
aI3215
aI6795
aI3821
aI31763
aI31059
aI65495
aI89981
asS'd'
I16947
sa(dp33894
S'a'
I59476
sS's'
I18
sS'm'
(lp33895
I1
aI1
aI7
aI7
aI9
aI25
aI11
aI85
aI111
aI283
aI507
aI2077
aI2993
aI5415
aI31785
aI16495
aI82361
aI122105
asS'd'
I16948
sa(dp33896
S'a'
I59483
sS's'
I18
sS'm'
(lp33897
I1
aI1
aI3
aI11
aI27
aI21
aI127
aI175
aI397
aI419
aI1115
aI2285
aI223
aI3881
aI4187
aI53759
aI115035
aI181647
asS'd'
I16949
sa(dp33898
S'a'
I59502
sS's'
I18
sS'm'
(lp33899
I1
aI3
aI7
aI11
aI27
aI31
aI29
aI233
aI137
aI827
aI1009
aI3879
aI7595
aI12989
aI27655
aI8517
aI28083
aI214985
asS'd'
I16950
sa(dp33900
S'a'
I59509
sS's'
I18
sS'm'
(lp33901
I1
aI3
aI5
aI9
aI25
aI23
aI85
aI191
aI475
aI445
aI621
aI1341
aI4045
aI4299
aI24933
aI32765
aI20219
aI86949
asS'd'
I16951
sa(dp33902
S'a'
I59538
sS's'
I18
sS'm'
(lp33903
I1
aI3
aI1
aI5
aI25
aI35
aI121
aI33
aI199
aI405
aI163
aI3487
aI1087
aI743
aI21989
aI47273
aI49221
aI124831
asS'd'
I16952
sa(dp33904
S'a'
I59556
sS's'
I18
sS'm'
(lp33905
I1
aI1
aI5
aI1
aI7
aI3
aI91
aI15
aI335
aI351
aI1311
aI777
aI4303
aI7203
aI19465
aI9135
aI32251
aI69805
asS'd'
I16953
sa(dp33906
S'a'
I59571
sS's'
I18
sS'm'
(lp33907
I1
aI3
aI3
aI7
aI23
aI55
aI73
aI77
aI189
aI801
aI1877
aI1901
aI2675
aI1015
aI3041
aI35925
aI125903
aI126227
asS'd'
I16954
sa(dp33908
S'a'
I59592
sS's'
I18
sS'm'
(lp33909
I1
aI3
aI3
aI1
aI1
aI23
aI105
aI75
aI435
aI743
aI651
aI1045
aI579
aI13637
aI14821
aI62683
aI95229
aI156475
asS'd'
I16955
sa(dp33910
S'a'
I59610
sS's'
I18
sS'm'
(lp33911
I1
aI3
aI3
aI5
aI1
aI53
aI89
aI239
aI439
aI195
aI189
aI731
aI1805
aI15123
aI23315
aI47737
aI29167
aI112081
asS'd'
I16956
sa(dp33912
S'a'
I59654
sS's'
I18
sS'm'
(lp33913
I1
aI1
aI5
aI7
aI31
aI11
aI119
aI191
aI155
aI61
aI247
aI915
aI5813
aI995
aI20093
aI23379
aI118969
aI65001
asS'd'
I16957
sa(dp33914
S'a'
I59677
sS's'
I18
sS'm'
(lp33915
I1
aI3
aI7
aI3
aI1
aI61
aI45
aI85
aI295
aI269
aI539
aI1787
aI6639
aI11093
aI11303
aI18509
aI77637
aI200743
asS'd'
I16958
sa(dp33916
S'a'
I59699
sS's'
I18
sS'm'
(lp33917
I1
aI3
aI1
aI1
aI13
aI17
aI75
aI51
aI199
aI151
aI1529
aI1443
aI4983
aI6723
aI6071
aI34711
aI39159
aI5441
asS'd'
I16959
sa(dp33918
S'a'
I59716
sS's'
I18
sS'm'
(lp33919
I1
aI3
aI3
aI3
aI31
aI15
aI91
aI125
aI261
aI683
aI1769
aI1697
aI2761
aI11373
aI13607
aI24933
aI19079
aI55497
asS'd'
I16960
sa(dp33920
S'a'
I59719
sS's'
I18
sS'm'
(lp33921
I1
aI1
aI1
aI15
aI21
aI49
aI117
aI99
aI29
aI969
aI463
aI3869
aI1251
aI8815
aI16443
aI46861
aI82839
aI233325
asS'd'
I16961
sa(dp33922
S'a'
I59737
sS's'
I18
sS'm'
(lp33923
I1
aI3
aI5
aI3
aI27
aI39
aI89
aI225
aI161
aI63
aI61
aI2875
aI4037
aI10413
aI5067
aI27893
aI78825
aI250207
asS'd'
I16962
sa(dp33924
S'a'
I59747
sS's'
I18
sS'm'
(lp33925
I1
aI1
aI1
aI9
aI13
aI49
aI93
aI11
aI23
aI25
aI2003
aI57
aI3065
aI11241
aI13935
aI2969
aI44235
aI39287
asS'd'
I16963
sa(dp33926
S'a'
I59789
sS's'
I18
sS'm'
(lp33927
I1
aI3
aI3
aI9
aI21
aI5
aI55
aI247
aI193
aI523
aI575
aI1235
aI3277
aI5253
aI5293
aI7919
aI7573
aI168809
asS'd'
I16964
sa(dp33928
S'a'
I59807
sS's'
I18
sS'm'
(lp33929
I1
aI3
aI1
aI13
aI29
aI39
aI43
aI21
aI511
aI205
aI303
aI703
aI3861
aI2467
aI3909
aI31597
aI51081
aI9863
asS'd'
I16965
sa(dp33930
S'a'
I59811
sS's'
I18
sS'm'
(lp33931
I1
aI3
aI7
aI1
aI25
aI55
aI11
aI131
aI5
aI49
aI371
aI1683
aI1907
aI5661
aI1015
aI15171
aI101477
aI11221
asS'd'
I16966
sa(dp33932
S'a'
I59818
sS's'
I18
sS'm'
(lp33933
I1
aI3
aI5
aI5
aI9
aI15
aI93
aI245
aI357
aI703
aI701
aI3675
aI4527
aI9225
aI16137
aI55433
aI81887
aI99153
asS'd'
I16967
sa(dp33934
S'a'
I59825
sS's'
I18
sS'm'
(lp33935
I1
aI1
aI3
aI3
aI11
aI1
aI39
aI251
aI291
aI599
aI643
aI231
aI4031
aI7055
aI99
aI14039
aI81811
aI184251
asS'd'
I16968
sa(dp33936
S'a'
I59826
sS's'
I18
sS'm'
(lp33937
I1
aI3
aI5
aI13
aI29
aI55
aI11
aI117
aI325
aI401
aI2013
aI3235
aI995
aI9255
aI2741
aI8211
aI71451
aI180619
asS'd'
I16969
sa(dp33938
S'a'
I59832
sS's'
I18
sS'm'
(lp33939
I1
aI1
aI1
aI5
aI31
aI41
aI41
aI175
aI247
aI3
aI739
aI1391
aI3311
aI5975
aI16921
aI4291
aI75065
aI161745
asS'd'
I16970
sa(dp33940
S'a'
I59858
sS's'
I18
sS'm'
(lp33941
I1
aI1
aI7
aI13
aI23
aI19
aI13
aI149
aI203
aI351
aI2033
aI1867
aI3871
aI14437
aI3793
aI17399
aI99577
aI171605
asS'd'
I16971
sa(dp33942
S'a'
I59860
sS's'
I18
sS'm'
(lp33943
I1
aI1
aI5
aI11
aI7
aI9
aI1
aI195
aI261
aI977
aI315
aI3771
aI1179
aI16281
aI20747
aI56309
aI108609
aI209205
asS'd'
I16972
sa(dp33944
S'a'
I59873
sS's'
I18
sS'm'
(lp33945
I1
aI3
aI3
aI5
aI19
aI15
aI123
aI153
aI325
aI601
aI393
aI753
aI93
aI4803
aI24343
aI42645
aI128209
aI45773
asS'd'
I16973
sa(dp33946
S'a'
I59876
sS's'
I18
sS'm'
(lp33947
I1
aI3
aI1
aI13
aI3
aI29
aI97
aI95
aI115
aI539
aI155
aI2789
aI1277
aI13127
aI20383
aI52807
aI97295
aI54589
asS'd'
I16974
sa(dp33948
S'a'
I59907
sS's'
I18
sS'm'
(lp33949
I1
aI1
aI1
aI3
aI25
aI59
aI27
aI149
aI365
aI317
aI773
aI3379
aI5931
aI14637
aI19881
aI37283
aI118027
aI21557
asS'd'
I16975
sa(dp33950
S'a'
I59928
sS's'
I18
sS'm'
(lp33951
I1
aI3
aI5
aI3
aI25
aI11
aI101
aI221
aI199
aI689
aI515
aI2255
aI6107
aI6259
aI2853
aI19039
aI117089
aI107181
asS'd'
I16976
sa(dp33952
S'a'
I59933
sS's'
I18
sS'm'
(lp33953
I1
aI3
aI5
aI7
aI29
aI63
aI19
aI113
aI249
aI147
aI737
aI3959
aI209
aI7001
aI24263
aI20443
aI99923
aI145709
asS'd'
I16977
sa(dp33954
S'a'
I59938
sS's'
I18
sS'm'
(lp33955
I1
aI1
aI5
aI9
aI25
aI37
aI69
aI41
aI87
aI369
aI1913
aI2255
aI7581
aI5301
aI25751
aI24981
aI1183
aI171969
asS'd'
I16978
sa(dp33956
S'a'
I59940
sS's'
I18
sS'm'
(lp33957
I1
aI3
aI3
aI1
aI9
aI25
aI55
aI5
aI267
aI295
aI43
aI819
aI4569
aI7065
aI31527
aI57811
aI48721
aI107707
asS'd'
I16979
sa(dp33958
S'a'
I59958
sS's'
I18
sS'm'
(lp33959
I1
aI1
aI7
aI9
aI19
aI19
aI1
aI199
aI371
aI1003
aI597
aI2097
aI4071
aI6185
aI879
aI13545
aI30033
aI120313
asS'd'
I16980
sa(dp33960
S'a'
I59984
sS's'
I18
sS'm'
(lp33961
I1
aI1
aI1
aI7
aI9
aI11
aI51
aI155
aI309
aI493
aI899
aI3121
aI2085
aI10541
aI21979
aI4725
aI70381
aI69643
asS'd'
I16981
sa(dp33962
S'a'
I60020
sS's'
I18
sS'm'
(lp33963
I1
aI1
aI1
aI1
aI13
aI45
aI123
aI119
aI459
aI295
aI1005
aI4093
aI393
aI11063
aI27235
aI28209
aI1671
aI215619
asS'd'
I16982
sa(dp33964
S'a'
I60024
sS's'
I18
sS'm'
(lp33965
I1
aI1
aI7
aI13
aI19
aI25
aI125
aI255
aI509
aI529
aI1577
aI3221
aI4051
aI7697
aI2065
aI42597
aI86295
aI131719
asS'd'
I16983
sa(dp33966
S'a'
I60033
sS's'
I18
sS'm'
(lp33967
I1
aI3
aI3
aI9
aI19
aI13
aI21
aI199
aI97
aI949
aI1297
aI379
aI1801
aI13247
aI22563
aI49517
aI22757
aI87371
asS'd'
I16984
sa(dp33968
S'a'
I60034
sS's'
I18
sS'm'
(lp33969
I1
aI3
aI3
aI1
aI17
aI63
aI109
aI175
aI301
aI565
aI1181
aI465
aI3457
aI7175
aI21225
aI33149
aI122169
aI148043
asS'd'
I16985
sa(dp33970
S'a'
I60063
sS's'
I18
sS'm'
(lp33971
I1
aI1
aI1
aI1
aI5
aI7
aI21
aI251
aI53
aI369
aI955
aI583
aI4703
aI9729
aI15853
aI55701
aI29317
aI27
asS'd'
I16986
sa(dp33972
S'a'
I60070
sS's'
I18
sS'm'
(lp33973
I1
aI3
aI3
aI1
aI31
aI3
aI53
aI57
aI231
aI441
aI109
aI149
aI8107
aI2303
aI29729
aI42279
aI46909
aI209877
asS'd'
I16987
sa(dp33974
S'a'
I60087
sS's'
I18
sS'm'
(lp33975
I1
aI1
aI1
aI11
aI23
aI57
aI63
aI189
aI259
aI657
aI1653
aI1155
aI2885
aI3317
aI22559
aI3145
aI19151
aI172507
asS'd'
I16988
sa(dp33976
S'a'
I60091
sS's'
I18
sS'm'
(lp33977
I1
aI3
aI7
aI5
aI31
aI63
aI103
aI147
aI287
aI685
aI1197
aI99
aI4907
aI12335
aI12001
aI20303
aI75503
aI231259
asS'd'
I16989
sa(dp33978
S'a'
I60105
sS's'
I18
sS'm'
(lp33979
I1
aI3
aI3
aI13
aI15
aI33
aI63
aI11
aI99
aI299
aI97
aI2669
aI3635
aI9969
aI1525
aI36555
aI85215
aI86915
asS'd'
I16990
sa(dp33980
S'a'
I60126
sS's'
I18
sS'm'
(lp33981
I1
aI3
aI7
aI5
aI25
aI47
aI25
aI61
aI227
aI939
aI1719
aI245
aI2389
aI14663
aI30671
aI22667
aI38873
aI245509
asS'd'
I16991
sa(dp33982
S'a'
I60132
sS's'
I18
sS'm'
(lp33983
I1
aI1
aI3
aI5
aI25
aI15
aI105
aI203
aI57
aI961
aI1941
aI1241
aI3163
aI6203
aI19631
aI10383
aI19235
aI57569
asS'd'
I16992
sa(dp33984
S'a'
I60154
sS's'
I18
sS'm'
(lp33985
I1
aI1
aI3
aI9
aI1
aI35
aI41
aI3
aI449
aI87
aI641
aI269
aI1529
aI14559
aI16571
aI4863
aI21625
aI921
asS'd'
I16993
sa(dp33986
S'a'
I60174
sS's'
I18
sS'm'
(lp33987
I1
aI1
aI7
aI11
aI25
aI53
aI85
aI209
aI181
aI417
aI1657
aI2117
aI4581
aI7069
aI15533
aI64475
aI82381
aI146943
asS'd'
I16994
sa(dp33988
S'a'
I60181
sS's'
I18
sS'm'
(lp33989
I1
aI3
aI1
aI7
aI17
aI53
aI5
aI199
aI347
aI887
aI1041
aI595
aI1843
aI10931
aI30559
aI42849
aI73723
aI220473
asS'd'
I16995
sa(dp33990
S'a'
I60198
sS's'
I18
sS'm'
(lp33991
I1
aI3
aI7
aI7
aI21
aI53
aI105
aI21
aI141
aI575
aI1965
aI2187
aI7293
aI13675
aI2471
aI1259
aI42485
aI62911
asS'd'
I16996
sa(dp33992
S'a'
I60212
sS's'
I18
sS'm'
(lp33993
I1
aI1
aI3
aI9
aI5
aI27
aI21
aI101
aI101
aI71
aI1215
aI3235
aI2451
aI14835
aI27817
aI30079
aI124301
aI253691
asS'd'
I16997
sa(dp33994
S'a'
I60247
sS's'
I18
sS'm'
(lp33995
I1
aI1
aI7
aI1
aI11
aI37
aI105
aI127
aI115
aI157
aI279
aI2425
aI2139
aI131
aI22717
aI40803
aI74867
aI86021
asS'd'
I16998
sa(dp33996
S'a'
I60254
sS's'
I18
sS'm'
(lp33997
I1
aI1
aI1
aI7
aI11
aI59
aI95
aI61
aI255
aI523
aI501
aI2895
aI7531
aI8151
aI18393
aI42069
aI120809
aI236537
asS'd'
I16999
sa(dp33998
S'a'
I60275
sS's'
I18
sS'm'
(lp33999
I1
aI1
aI1
aI7
aI17
aI23
aI31
aI59
aI377
aI187
aI873
aI1565
aI3459
aI2975
aI11633
aI13247
aI13095
aI193803
asS'd'
I17000
sa(dp34000
S'a'
I60284
sS's'
I18
sS'm'
(lp34001
I1
aI3
aI7
aI1
aI25
aI3
aI85
aI5
aI485
aI451
aI1385
aI1663
aI4825
aI14019
aI29437
aI33717
aI105343
aI161335
asS'd'
I17001
sa(dp34002
S'a'
I60317
sS's'
I18
sS'm'
(lp34003
I1
aI1
aI1
aI3
aI27
aI43
aI71
aI167
aI425
aI579
aI1739
aI3557
aI7403
aI2023
aI6533
aI61177
aI119273
aI85229
asS'd'
I17002
sa(dp34004
S'a'
I60318
sS's'
I18
sS'm'
(lp34005
I1
aI1
aI7
aI3
aI31
aI37
aI19
aI213
aI373
aI505
aI97
aI3669
aI7005
aI2205
aI26519
aI61999
aI18395
aI25967
asS'd'
I17003
sa(dp34006
S'a'
I60346
sS's'
I18
sS'm'
(lp34007
I1
aI3
aI3
aI7
aI29
aI17
aI9
aI137
aI265
aI875
aI887
aI3029
aI3295
aI11619
aI8357
aI46241
aI23543
aI43191
asS'd'
I17004
sa(dp34008
S'a'
I60360
sS's'
I18
sS'm'
(lp34009
I1
aI1
aI5
aI1
aI25
aI43
aI33
aI133
aI65
aI7
aI1581
aI3577
aI5997
aI6129
aI30649
aI18923
aI56459
aI227869
asS'd'
I17005
sa(dp34010
S'a'
I60380
sS's'
I18
sS'm'
(lp34011
I1
aI1
aI5
aI13
aI27
aI45
aI27
aI111
aI429
aI565
aI1449
aI1475
aI6613
aI4469
aI16083
aI42349
aI66843
aI214875
asS'd'
I17006
sa(dp34012
S'a'
I60389
sS's'
I18
sS'm'
(lp34013
I1
aI3
aI1
aI1
aI21
aI21
aI107
aI7
aI15
aI675
aI233
aI4021
aI1097
aI1393
aI6445
aI3323
aI102435
aI249355
asS'd'
I17007
sa(dp34014
S'a'
I60393
sS's'
I18
sS'm'
(lp34015
I1
aI1
aI5
aI15
aI17
aI51
aI99
aI249
aI437
aI667
aI1921
aI2371
aI3813
aI10543
aI19
aI39079
aI116825
aI242821
asS'd'
I17008
sa(dp34016
S'a'
I60401
sS's'
I18
sS'm'
(lp34017
I1
aI1
aI1
aI1
aI7
aI15
aI27
aI29
aI161
aI37
aI1847
aI287
aI4379
aI1399
aI24547
aI60361
aI68131
aI232883
asS'd'
I17009
sa(dp34018
S'a'
I60407
sS's'
I18
sS'm'
(lp34019
I1
aI1
aI3
aI9
aI17
aI21
aI41
aI169
aI61
aI771
aI241
aI1435
aI4151
aI1789
aI12195
aI27239
aI62371
aI165145
asS'd'
I17010
sa(dp34020
S'a'
I60408
sS's'
I18
sS'm'
(lp34021
I1
aI3
aI5
aI15
aI31
aI19
aI127
aI181
aI463
aI183
aI749
aI253
aI2403
aI1363
aI3965
aI7953
aI124025
aI226691
asS'd'
I17011
sa(dp34022
S'a'
I60462
sS's'
I18
sS'm'
(lp34023
I1
aI1
aI3
aI7
aI17
aI57
aI85
aI89
aI17
aI33
aI819
aI2191
aI1525
aI15651
aI23483
aI26027
aI86379
aI40191
asS'd'
I17012
sa(dp34024
S'a'
I60484
sS's'
I18
sS'm'
(lp34025
I1
aI1
aI5
aI1
aI9
aI45
aI65
aI65
aI359
aI5
aI531
aI2581
aI6313
aI13219
aI6005
aI36215
aI16275
aI208253
asS'd'
I17013
sa(dp34026
S'a'
I60505
sS's'
I18
sS'm'
(lp34027
I1
aI3
aI1
aI1
aI23
aI15
aI51
aI43
aI85
aI461
aI773
aI219
aI2681
aI3377
aI9797
aI54469
aI112871
aI231533
asS'd'
I17014
sa(dp34028
S'a'
I60521
sS's'
I18
sS'm'
(lp34029
I1
aI3
aI7
aI7
aI15
aI9
aI97
aI115
aI301
aI493
aI1085
aI2021
aI2305
aI15003
aI11381
aI9339
aI63015
aI179115
asS'd'
I17015
sa(dp34030
S'a'
I60527
sS's'
I18
sS'm'
(lp34031
I1
aI3
aI5
aI3
aI3
aI61
aI111
aI103
aI283
aI7
aI143
aI353
aI7815
aI7901
aI25795
aI7577
aI92991
aI228315
asS'd'
I17016
sa(dp34032
S'a'
I60532
sS's'
I18
sS'm'
(lp34033
I1
aI1
aI7
aI13
aI29
aI3
aI53
aI105
aI83
aI531
aI497
aI729
aI1375
aI7063
aI18655
aI35219
aI9671
aI102913
asS'd'
I17017
sa(dp34034
S'a'
I60545
sS's'
I18
sS'm'
(lp34035
I1
aI1
aI5
aI15
aI11
aI7
aI65
aI31
aI15
aI921
aI743
aI1469
aI5669
aI5437
aI20019
aI28123
aI5717
aI6181
asS'd'
I17018
sa(dp34036
S'a'
I60546
sS's'
I18
sS'm'
(lp34037
I1
aI3
aI3
aI15
aI19
aI1
aI3
aI183
aI315
aI595
aI1033
aI3259
aI7815
aI8281
aI32103
aI8699
aI59149
aI56657
asS'd'
I17019
sa(dp34038
S'a'
I60555
sS's'
I18
sS'm'
(lp34039
I1
aI1
aI1
aI7
aI9
aI1
aI87
aI81
aI267
aI637
aI1617
aI2113
aI487
aI23
aI11213
aI29211
aI92715
aI177767
asS'd'
I17020
sa(dp34040
S'a'
I60563
sS's'
I18
sS'm'
(lp34041
I1
aI3
aI5
aI13
aI29
aI37
aI31
aI55
aI343
aI759
aI813
aI2945
aI7189
aI4821
aI30661
aI38373
aI2793
aI98683
asS'd'
I17021
sa(dp34042
S'a'
I60586
sS's'
I18
sS'm'
(lp34043
I1
aI3
aI5
aI7
aI9
aI43
aI113
aI145
aI103
aI303
aI1065
aI3781
aI3527
aI9449
aI17355
aI38301
aI74859
aI30735
asS'd'
I17022
sa(dp34044
S'a'
I60626
sS's'
I18
sS'm'
(lp34045
I1
aI1
aI1
aI1
aI3
aI53
aI53
aI27
aI119
aI701
aI1777
aI3959
aI5911
aI8473
aI24997
aI17557
aI11593
aI201381
asS'd'
I17023
sa(dp34046
S'a'
I60637
sS's'
I18
sS'm'
(lp34047
I1
aI3
aI3
aI9
aI3
aI3
aI107
aI115
aI423
aI531
aI735
aI931
aI8053
aI4661
aI1919
aI29551
aI62515
aI210255
asS'd'
I17024
sa(dp34048
S'a'
I60665
sS's'
I18
sS'm'
(lp34049
I1
aI1
aI3
aI9
aI21
aI21
aI117
aI67
aI301
aI49
aI2025
aI781
aI7951
aI15719
aI27287
aI34551
aI115241
aI243981
asS'd'
I17025
sa(dp34050
S'a'
I60673
sS's'
I18
sS'm'
(lp34051
I1
aI1
aI7
aI9
aI9
aI25
aI87
aI229
aI375
aI353
aI445
aI3169
aI1865
aI7305
aI11175
aI47081
aI28609
aI107301
asS'd'
I17026
sa(dp34052
S'a'
I60685
sS's'
I18
sS'm'
(lp34053
I1
aI3
aI3
aI11
aI7
aI31
aI19
aI177
aI17
aI535
aI1353
aI2587
aI7723
aI8039
aI13607
aI5017
aI104937
aI207761
asS'd'
I17027
sa(dp34054
S'a'
I60703
sS's'
I18
sS'm'
(lp34055
I1
aI1
aI5
aI3
aI11
aI27
aI29
aI193
aI235
aI435
aI1451
aI3487
aI5749
aI4825
aI9487
aI53933
aI92061
aI223305
asS'd'
I17028
sa(dp34056
S'a'
I60734
sS's'
I18
sS'm'
(lp34057
I1
aI3
aI3
aI3
aI5
aI5
aI99
aI237
aI91
aI945
aI1373
aI3303
aI3079
aI5345
aI6843
aI34131
aI62851
aI259561
asS'd'
I17029
sa(dp34058
S'a'
I60742
sS's'
I18
sS'm'
(lp34059
I1
aI1
aI3
aI7
aI7
aI25
aI11
aI27
aI329
aI37
aI307
aI771
aI659
aI13045
aI25767
aI18887
aI54407
aI251313
asS'd'
I17030
sa(dp34060
S'a'
I60790
sS's'
I18
sS'm'
(lp34061
I1
aI1
aI1
aI11
aI29
aI59
aI37
aI121
aI281
aI55
aI495
aI159
aI3925
aI4447
aI14825
aI24831
aI103147
aI211951
asS'd'
I17031
sa(dp34062
S'a'
I60830
sS's'
I18
sS'm'
(lp34063
I1
aI3
aI3
aI7
aI23
aI31
aI59
aI67
aI303
aI383
aI1179
aI2347
aI4001
aI14797
aI14579
aI55365
aI112239
aI65309
asS'd'
I17032
sa(dp34064
S'a'
I60848
sS's'
I18
sS'm'
(lp34065
I1
aI3
aI3
aI9
aI15
aI17
aI61
aI123
aI339
aI319
aI765
aI1517
aI1269
aI69
aI9065
aI32347
aI21377
aI38449
asS'd'
I17033
sa(dp34066
S'a'
I60857
sS's'
I18
sS'm'
(lp34067
I1
aI3
aI5
aI3
aI7
aI35
aI71
aI63
aI251
aI457
aI351
aI385
aI4041
aI11489
aI14511
aI11875
aI45307
aI205041
asS'd'
I17034
sa(dp34068
S'a'
I60871
sS's'
I18
sS'm'
(lp34069
I1
aI1
aI1
aI7
aI1
aI25
aI115
aI195
aI41
aI1001
aI835
aI767
aI7991
aI7475
aI22397
aI36899
aI77255
aI194827
asS'd'
I17035
sa(dp34070
S'a'
I60875
sS's'
I18
sS'm'
(lp34071
I1
aI3
aI3
aI11
aI7
aI49
aI45
aI13
aI373
aI167
aI741
aI2569
aI3781
aI1131
aI2909
aI40387
aI77877
aI201859
asS'd'
I17036
sa(dp34072
S'a'
I60913
sS's'
I18
sS'm'
(lp34073
I1
aI3
aI5
aI3
aI17
aI11
aI123
aI137
aI65
aI835
aI1385
aI1157
aI7387
aI12301
aI5759
aI13137
aI30595
aI50923
asS'd'
I17037
sa(dp34074
S'a'
I60923
sS's'
I18
sS'm'
(lp34075
I1
aI1
aI5
aI7
aI1
aI55
aI57
aI97
aI377
aI223
aI115
aI2515
aI2565
aI14965
aI10485
aI23957
aI108239
aI160707
asS'd'
I17038
sa(dp34076
S'a'
I60972
sS's'
I18
sS'm'
(lp34077
I1
aI3
aI1
aI9
aI15
aI17
aI81
aI65
aI387
aI275
aI997
aI1485
aI4129
aI999
aI4915
aI55867
aI103799
aI191829
asS'd'
I17039
sa(dp34078
S'a'
I60989
sS's'
I18
sS'm'
(lp34079
I1
aI1
aI1
aI5
aI15
aI5
aI35
aI167
aI249
aI419
aI267
aI503
aI469
aI3163
aI19939
aI65501
aI88573
aI11621
asS'd'
I17040
sa(dp34080
S'a'
I61001
sS's'
I18
sS'm'
(lp34081
I1
aI3
aI3
aI13
aI7
aI9
aI101
aI125
aI371
aI97
aI1855
aI1755
aI4103
aI12283
aI18655
aI5965
aI17743
aI254779
asS'd'
I17041
sa(dp34082
S'a'
I61002
sS's'
I18
sS'm'
(lp34083
I1
aI1
aI3
aI7
aI13
aI15
aI119
aI227
aI451
aI863
aI1005
aI491
aI6515
aI717
aI12783
aI14161
aI106249
aI185297
asS'd'
I17042
sa(dp34084
S'a'
I61019
sS's'
I18
sS'm'
(lp34085
I1
aI1
aI7
aI13
aI17
aI23
aI95
aI143
aI133
aI219
aI897
aI2291
aI7469
aI923
aI22323
aI60583
aI2457
aI197231
asS'd'
I17043
sa(dp34086
S'a'
I61022
sS's'
I18
sS'm'
(lp34087
I1
aI1
aI1
aI11
aI3
aI25
aI115
aI187
aI319
aI999
aI867
aI1725
aI6969
aI239
aI2527
aI55283
aI91099
aI252153
asS'd'
I17044
sa(dp34088
S'a'
I61059
sS's'
I18
sS'm'
(lp34089
I1
aI1
aI7
aI9
aI3
aI37
aI107
aI25
aI425
aI95
aI631
aI2831
aI1265
aI11509
aI18865
aI39791
aI22281
aI220517
asS'd'
I17045
sa(dp34090
S'a'
I61065
sS's'
I18
sS'm'
(lp34091
I1
aI3
aI5
aI1
aI1
aI47
aI121
aI173
aI489
aI241
aI3
aI3707
aI7081
aI5341
aI23143
aI7321
aI30605
aI191665
asS'd'
I17046
sa(dp34092
S'a'
I61066
sS's'
I18
sS'm'
(lp34093
I1
aI1
aI7
aI7
aI7
aI27
aI23
aI43
aI145
aI11
aI1155
aI691
aI6993
aI9509
aI5991
aI40705
aI58215
aI202915
asS'd'
I17047
sa(dp34094
S'a'
I61071
sS's'
I18
sS'm'
(lp34095
I1
aI3
aI5
aI1
aI31
aI1
aI7
aI189
aI379
aI431
aI417
aI3843
aI3885
aI3263
aI16333
aI58123
aI68307
aI33795
asS'd'
I17048
sa(dp34096
S'a'
I61076
sS's'
I18
sS'm'
(lp34097
I1
aI1
aI5
aI5
aI19
aI27
aI19
aI217
aI509
aI535
aI287
aI1637
aI4829
aI2665
aI15393
aI35185
aI125335
aI10909
asS'd'
I17049
sa(dp34098
S'a'
I61141
sS's'
I18
sS'm'
(lp34099
I1
aI3
aI5
aI7
aI25
aI13
aI67
aI243
aI255
aI1021
aI1203
aI821
aI7811
aI149
aI26731
aI12913
aI18171
aI101385
asS'd'
I17050
sa(dp34100
S'a'
I61148
sS's'
I18
sS'm'
(lp34101
I1
aI1
aI1
aI3
aI25
aI61
aI59
aI207
aI449
aI789
aI1831
aI1731
aI513
aI10099
aI291
aI1963
aI100233
aI21847
asS'd'
I17051
sa(dp34102
S'a'
I61167
sS's'
I18
sS'm'
(lp34103
I1
aI1
aI7
aI1
aI27
aI19
aI45
aI81
aI479
aI31
aI707
aI2669
aI3589
aI15411
aI12089
aI38235
aI60897
aI135451
asS'd'
I17052
sa(dp34104
S'a'
I61190
sS's'
I18
sS'm'
(lp34105
I1
aI3
aI5
aI15
aI11
aI3
aI113
aI169
aI171
aI21
aI1291
aI2031
aI2023
aI5783
aI6137
aI54637
aI50247
aI233753
asS'd'
I17053
sa(dp34106
S'a'
I61218
sS's'
I18
sS'm'
(lp34107
I1
aI1
aI5
aI1
aI11
aI13
aI73
aI97
aI269
aI801
aI1015
aI1329
aI1779
aI15225
aI24251
aI35191
aI8619
aI130993
asS'd'
I17054
sa(dp34108
S'a'
I61247
sS's'
I18
sS'm'
(lp34109
I1
aI1
aI7
aI9
aI9
aI19
aI35
aI255
aI505
aI513
aI547
aI405
aI3065
aI4965
aI30877
aI50091
aI81319
aI29273
asS'd'
I17055
sa(dp34110
S'a'
I61256
sS's'
I18
sS'm'
(lp34111
I1
aI1
aI1
aI9
aI7
aI61
aI45
aI75
aI343
aI911
aI1683
aI453
aI1225
aI10939
aI19901
aI63685
aI123507
aI252027
asS'd'
I17056
sa(dp34112
S'a'
I61289
sS's'
I18
sS'm'
(lp34113
I1
aI1
aI5
aI11
aI21
aI55
aI37
aI161
aI143
aI463
aI1937
aI3349
aI2953
aI14827
aI7893
aI26581
aI128459
aI72325
asS'd'
I17057
sa(dp34114
S'a'
I61292
sS's'
I18
sS'm'
(lp34115
I1
aI3
aI5
aI9
aI31
aI57
aI115
aI77
aI225
aI859
aI621
aI731
aI5677
aI759
aI20773
aI52285
aI65555
aI4303
asS'd'
I17058
sa(dp34116
S'a'
I61307
sS's'
I18
sS'm'
(lp34117
I1
aI1
aI3
aI15
aI23
aI15
aI49
aI171
aI137
aI449
aI855
aI565
aI5579
aI5957
aI13643
aI8979
aI90327
aI116349
asS'd'
I17059
sa(dp34118
S'a'
I61319
sS's'
I18
sS'm'
(lp34119
I1
aI1
aI3
aI5
aI9
aI13
aI27
aI43
aI391
aI595
aI731
aI101
aI7121
aI13555
aI29181
aI38273
aI42309
aI175297
asS'd'
I17060
sa(dp34120
S'a'
I61353
sS's'
I18
sS'm'
(lp34121
I1
aI1
aI3
aI11
aI3
aI59
aI17
aI143
aI251
aI47
aI1391
aI1297
aI23
aI15871
aI13153
aI44081
aI65423
aI54875
asS'd'
I17061
sa(dp34122
S'a'
I61362
sS's'
I18
sS'm'
(lp34123
I1
aI3
aI7
aI11
aI25
aI43
aI3
aI163
aI273
aI277
aI755
aI2743
aI5909
aI10841
aI31331
aI64131
aI13945
aI91557
asS'd'
I17062
sa(dp34124
S'a'
I61379
sS's'
I18
sS'm'
(lp34125
I1
aI3
aI1
aI13
aI31
aI35
aI5
aI165
aI417
aI623
aI1083
aI1221
aI1051
aI8917
aI6725
aI11385
aI76315
aI119837
asS'd'
I17063
sa(dp34126
S'a'
I61382
sS's'
I18
sS'm'
(lp34127
I1
aI1
aI5
aI3
aI21
aI53
aI47
aI247
aI471
aI877
aI709
aI2425
aI3
aI1963
aI24331
aI52151
aI98859
aI119033
asS'd'
I17064
sa(dp34128
S'a'
I61433
sS's'
I18
sS'm'
(lp34129
I1
aI1
aI7
aI3
aI29
aI29
aI43
aI59
aI503
aI891
aI763
aI2927
aI1613
aI9091
aI10393
aI36003
aI61147
aI3437
asS'd'
I17065
sa(dp34130
S'a'
I61434
sS's'
I18
sS'm'
(lp34131
I1
aI1
aI5
aI15
aI27
aI59
aI73
aI163
aI425
aI855
aI349
aI3451
aI5779
aI10523
aI9103
aI46477
aI129873
aI39091
asS'd'
I17066
sa(dp34132
S'a'
I61454
sS's'
I18
sS'm'
(lp34133
I1
aI3
aI7
aI15
aI25
aI45
aI77
aI171
aI467
aI1017
aI1553
aI1877
aI5507
aI3909
aI12157
aI60441
aI98261
aI37781
asS'd'
I17067
sa(dp34134
S'a'
I61461
sS's'
I18
sS'm'
(lp34135
I1
aI1
aI7
aI13
aI13
aI39
aI99
aI51
aI197
aI327
aI1101
aI2679
aI8025
aI11853
aI7763
aI62537
aI96999
aI88673
asS'd'
I17068
sa(dp34136
S'a'
I61475
sS's'
I18
sS'm'
(lp34137
I1
aI3
aI1
aI11
aI5
aI61
aI29
aI219
aI471
aI387
aI319
aI433
aI5383
aI3933
aI27603
aI61171
aI104711
aI233295
asS'd'
I17069
sa(dp34138
S'a'
I61481
sS's'
I18
sS'm'
(lp34139
I1
aI3
aI5
aI11
aI15
aI23
aI91
aI119
aI207
aI717
aI1333
aI783
aI437
aI13073
aI10923
aI27049
aI87233
aI174899
asS'd'
I17070
sa(dp34140
S'a'
I61534
sS's'
I18
sS'm'
(lp34141
I1
aI1
aI1
aI1
aI13
aI19
aI109
aI139
aI183
aI299
aI1023
aI3265
aI5153
aI6307
aI27879
aI55311
aI95201
aI19481
asS'd'
I17071
sa(dp34142
S'a'
I61547
sS's'
I18
sS'm'
(lp34143
I1
aI1
aI5
aI11
aI13
aI61
aI81
aI115
aI53
aI483
aI693
aI3527
aI5033
aI8527
aI31345
aI46155
aI12403
aI126815
asS'd'
I17072
sa(dp34144
S'a'
I61564
sS's'
I18
sS'm'
(lp34145
I1
aI3
aI7
aI3
aI27
aI7
aI73
aI227
aI269
aI683
aI719
aI763
aI5417
aI9523
aI13625
aI6945
aI116225
aI223093
asS'd'
I17073
sa(dp34146
S'a'
I61568
sS's'
I18
sS'm'
(lp34147
I1
aI3
aI5
aI9
aI21
aI51
aI111
aI157
aI451
aI247
aI1375
aI1631
aI2783
aI3371
aI22713
aI34153
aI41949
aI141351
asS'd'
I17074
sa(dp34148
S'a'
I61588
sS's'
I18
sS'm'
(lp34149
I1
aI1
aI5
aI1
aI21
aI19
aI45
aI69
aI41
aI453
aI523
aI3163
aI7351
aI4467
aI18865
aI35371
aI129577
aI78039
asS'd'
I17075
sa(dp34150
S'a'
I61604
sS's'
I18
sS'm'
(lp34151
I1
aI3
aI5
aI1
aI29
aI33
aI13
aI19
aI341
aI321
aI117
aI1187
aI7021
aI5785
aI5553
aI58055
aI113557
aI46957
asS'd'
I17076
sa(dp34152
S'a'
I61626
sS's'
I18
sS'm'
(lp34153
I1
aI1
aI5
aI5
aI13
aI59
aI47
aI59
aI69
aI125
aI1491
aI2813
aI5005
aI5973
aI3145
aI27579
aI7763
aI129949
asS'd'
I17077
sa(dp34154
S'a'
I61645
sS's'
I18
sS'm'
(lp34155
I1
aI1
aI7
aI11
aI11
aI7
aI117
aI235
aI407
aI749
aI1925
aI1735
aI4499
aI13027
aI19355
aI1981
aI105657
aI242853
asS'd'
I17078
sa(dp34156
S'a'
I61646
sS's'
I18
sS'm'
(lp34157
I1
aI1
aI7
aI1
aI15
aI19
aI5
aI247
aI203
aI707
aI809
aI2085
aI5801
aI9947
aI569
aI9883
aI109861
aI156751
asS'd'
I17079
sa(dp34158
S'a'
I61654
sS's'
I18
sS'm'
(lp34159
I1
aI3
aI5
aI3
aI13
aI59
aI67
aI181
aI261
aI873
aI1589
aI2249
aI7213
aI14625
aI28403
aI41101
aI73439
aI46873
asS'd'
I17080
sa(dp34160
S'a'
I61663
sS's'
I18
sS'm'
(lp34161
I1
aI1
aI5
aI7
aI3
aI63
aI79
aI115
aI123
aI485
aI1373
aI3781
aI4315
aI4627
aI29003
aI64101
aI67521
aI184053
asS'd'
I17081
sa(dp34162
S'a'
I61669
sS's'
I18
sS'm'
(lp34163
I1
aI1
aI3
aI9
aI11
aI57
aI93
aI243
aI505
aI189
aI449
aI643
aI5267
aI7447
aI32265
aI44095
aI63015
aI36905
asS'd'
I17082
sa(dp34164
S'a'
I61702
sS's'
I18
sS'm'
(lp34165
I1
aI3
aI7
aI13
aI25
aI59
aI31
aI93
aI401
aI41
aI183
aI759
aI2473
aI8705
aI8211
aI13543
aI59749
aI235217
asS'd'
I17083
sa(dp34166
S'a'
I61705
sS's'
I18
sS'm'
(lp34167
I1
aI1
aI5
aI5
aI29
aI3
aI65
aI133
aI325
aI239
aI649
aI3225
aI4095
aI11691
aI4479
aI15419
aI100551
aI261981
asS'd'
I17084
sa(dp34168
S'a'
I61714
sS's'
I18
sS'm'
(lp34169
I1
aI3
aI7
aI1
aI17
aI11
aI63
aI97
aI431
aI161
aI1437
aI3679
aI1643
aI10583
aI20731
aI45919
aI94093
aI147067
asS'd'
I17085
sa(dp34170
S'a'
I61739
sS's'
I18
sS'm'
(lp34171
I1
aI1
aI7
aI1
aI25
aI13
aI63
aI155
aI221
aI345
aI189
aI1199
aI5465
aI14767
aI26263
aI54093
aI23697
aI71231
asS'd'
I17086
sa(dp34172
S'a'
I61744
sS's'
I18
sS'm'
(lp34173
I1
aI1
aI7
aI3
aI3
aI19
aI23
aI75
aI381
aI339
aI1989
aI1137
aI6449
aI1437
aI32279
aI17195
aI117423
aI259311
asS'd'
I17087
sa(dp34174
S'a'
I61749
sS's'
I18
sS'm'
(lp34175
I1
aI3
aI5
aI3
aI27
aI45
aI117
aI113
aI129
aI585
aI2019
aI807
aI5573
aI7407
aI9957
aI8741
aI52333
aI115607
asS'd'
I17088
sa(dp34176
S'a'
I61776
sS's'
I18
sS'm'
(lp34177
I1
aI1
aI3
aI7
aI9
aI5
aI77
aI9
aI417
aI725
aI429
aI1657
aI5445
aI1901
aI28745
aI26807
aI111743
aI169739
asS'd'
I17089
sa(dp34178
S'a'
I61786
sS's'
I18
sS'm'
(lp34179
I1
aI3
aI1
aI5
aI7
aI63
aI51
aI183
aI117
aI383
aI435
aI755
aI7849
aI5997
aI32697
aI5789
aI5189
aI80645
asS'd'
I17090
sa(dp34180
S'a'
I61822
sS's'
I18
sS'm'
(lp34181
I1
aI1
aI3
aI15
aI5
aI47
aI105
aI175
aI41
aI275
aI1441
aI3183
aI3651
aI9561
aI5749
aI20431
aI45969
aI59473
asS'd'
I17091
sa(dp34182
S'a'
I61826
sS's'
I18
sS'm'
(lp34183
I1
aI1
aI1
aI11
aI13
aI35
aI19
aI129
aI125
aI35
aI339
aI3099
aI5337
aI15605
aI10213
aI1171
aI61869
aI216681
asS'd'
I17092
sa(dp34184
S'a'
I61862
sS's'
I18
sS'm'
(lp34185
I1
aI3
aI7
aI1
aI7
aI25
aI23
aI9
aI431
aI73
aI1803
aI3969
aI7853
aI12845
aI8075
aI14553
aI124825
aI50561
asS'd'
I17093
sa(dp34186
S'a'
I61865
sS's'
I18
sS'm'
(lp34187
I1
aI1
aI1
aI7
aI29
aI21
aI79
aI247
aI313
aI143
aI59
aI2689
aI5643
aI827
aI26597
aI56423
aI107903
aI180809
asS'd'
I17094
sa(dp34188
S'a'
I61868
sS's'
I18
sS'm'
(lp34189
I1
aI3
aI7
aI1
aI3
aI3
aI25
aI39
aI269
aI529
aI67
aI3703
aI2163
aI12417
aI6307
aI29883
aI40303
aI171831
asS'd'
I17095
sa(dp34190
S'a'
I61879
sS's'
I18
sS'm'
(lp34191
I1
aI3
aI7
aI13
aI1
aI19
aI71
aI245
aI267
aI105
aI749
aI1203
aI7953
aI1881
aI9273
aI4629
aI71793
aI195393
asS'd'
I17096
sa(dp34192
S'a'
I61885
sS's'
I18
sS'm'
(lp34193
I1
aI1
aI1
aI15
aI3
aI49
aI53
aI145
aI47
aI959
aI1107
aI1361
aI4517
aI16055
aI32119
aI58433
aI110123
aI81487
asS'd'
I17097
sa(dp34194
S'a'
I61900
sS's'
I18
sS'm'
(lp34195
I1
aI3
aI1
aI1
aI5
aI3
aI99
aI93
aI257
aI659
aI19
aI3789
aI203
aI6183
aI11571
aI54845
aI80591
aI243303
asS'd'
I17098
sa(dp34196
S'a'
I61905
sS's'
I18
sS'm'
(lp34197
I1
aI3
aI7
aI1
aI1
aI7
aI27
aI11
aI255
aI261
aI769
aI2877
aI6013
aI8431
aI25669
aI43591
aI122501
aI208947
asS'd'
I17099
sa(dp34198
S'a'
I61906
sS's'
I18
sS'm'
(lp34199
I1
aI1
aI7
aI7
aI3
aI25
aI117
aI19
aI15
aI843
aI401
aI613
aI801
aI10579
aI129
aI12249
aI107465
aI95953
asS'd'
I17100
sa(dp34200
S'a'
I61962
sS's'
I18
sS'm'
(lp34201
I1
aI1
aI3
aI5
aI1
aI35
aI95
aI93
aI243
aI937
aI1543
aI3443
aI175
aI2199
aI12521
aI2521
aI87225
aI38631
asS'd'
I17101
sa(dp34202
S'a'
I61967
sS's'
I18
sS'm'
(lp34203
I1
aI3
aI7
aI13
aI21
aI29
aI81
aI139
aI247
aI937
aI1835
aI3887
aI6917
aI15709
aI20947
aI3341
aI125521
aI247195
asS'd'
I17102
sa(dp34204
S'a'
I61972
sS's'
I18
sS'm'
(lp34205
I1
aI1
aI5
aI11
aI31
aI19
aI111
aI215
aI191
aI347
aI1215
aI1757
aI6751
aI3099
aI755
aI43753
aI2813
aI159123
asS'd'
I17103
sa(dp34206
S'a'
I61976
sS's'
I18
sS'm'
(lp34207
I1
aI3
aI3
aI3
aI31
aI5
aI35
aI87
aI293
aI581
aI1501
aI3255
aI7041
aI5233
aI2053
aI63403
aI37943
aI12115
asS'd'
I17104
sa(dp34208
S'a'
I62027
sS's'
I18
sS'm'
(lp34209
I1
aI1
aI7
aI15
aI11
aI31
aI5
aI123
aI225
aI703
aI733
aI635
aI2193
aI3059
aI30933
aI43149
aI79409
aI106995
asS'd'
I17105
sa(dp34210
S'a'
I62048
sS's'
I18
sS'm'
(lp34211
I1
aI3
aI1
aI7
aI11
aI21
aI45
aI135
aI99
aI883
aI85
aI3861
aI6617
aI7169
aI29887
aI329
aI42487
aI129001
asS'd'
I17106
sa(dp34212
S'a'
I62051
sS's'
I18
sS'm'
(lp34213
I1
aI1
aI1
aI3
aI11
aI53
aI31
aI245
aI141
aI667
aI1615
aI3311
aI1475
aI12785
aI3509
aI47153
aI105747
aI141275
asS'd'
I17107
sa(dp34214
S'a'
I62066
sS's'
I18
sS'm'
(lp34215
I1
aI1
aI3
aI15
aI7
aI15
aI55
aI13
aI465
aI707
aI1299
aI1393
aI399
aI9229
aI4897
aI50313
aI1275
aI131811
asS'd'
I17108
sa(dp34216
S'a'
I62081
sS's'
I18
sS'm'
(lp34217
I1
aI1
aI3
aI15
aI5
aI57
aI43
aI19
aI335
aI929
aI459
aI327
aI5715
aI7173
aI27643
aI535
aI46221
aI144619
asS'd'
I17109
sa(dp34218
S'a'
I62108
sS's'
I18
sS'm'
(lp34219
I1
aI3
aI5
aI1
aI9
aI1
aI63
aI187
aI71
aI899
aI969
aI1349
aI1553
aI15593
aI22783
aI211
aI41643
aI163981
asS'd'
I17110
sa(dp34220
S'a'
I62178
sS's'
I18
sS'm'
(lp34221
I1
aI1
aI1
aI13
aI3
aI63
aI35
aI37
aI311
aI253
aI1393
aI629
aI5299
aI14837
aI15053
aI28041
aI81541
aI149037
asS'd'
I17111
sa(dp34222
S'a'
I62189
sS's'
I18
sS'm'
(lp34223
I1
aI3
aI3
aI11
aI13
aI45
aI17
aI165
aI497
aI751
aI635
aI2939
aI6891
aI14877
aI32763
aI20671
aI106845
aI258033
asS'd'
I17112
sa(dp34224
S'a'
I62224
sS's'
I18
sS'm'
(lp34225
I1
aI1
aI3
aI11
aI21
aI7
aI3
aI247
aI243
aI219
aI1651
aI929
aI2737
aI9507
aI31819
aI61389
aI14593
aI137207
asS'd'
I17113
sa(dp34226
S'a'
I62229
sS's'
I18
sS'm'
(lp34227
I1
aI1
aI7
aI5
aI15
aI33
aI31
aI29
aI467
aI75
aI523
aI1067
aI7313
aI11715
aI26581
aI47037
aI106385
aI199859
asS'd'
I17114
sa(dp34228
S'a'
I62282
sS's'
I18
sS'm'
(lp34229
I1
aI3
aI1
aI7
aI19
aI59
aI35
aI35
aI3
aI899
aI799
aI1379
aI5113
aI7653
aI17977
aI42197
aI52397
aI179705
asS'd'
I17115
sa(dp34230
S'a'
I62318
sS's'
I18
sS'm'
(lp34231
I1
aI1
aI7
aI5
aI13
aI13
aI67
aI157
aI181
aI633
aI21
aI3107
aI6301
aI7523
aI23981
aI9079
aI88875
aI195869
asS'd'
I17116
sa(dp34232
S'a'
I62320
sS's'
I18
sS'm'
(lp34233
I1
aI3
aI7
aI9
aI7
aI9
aI115
aI49
aI293
aI691
aI1729
aI4087
aI6353
aI963
aI12433
aI22135
aI96383
aI127745
asS'd'
I17117
sa(dp34234
S'a'
I62329
sS's'
I18
sS'm'
(lp34235
I1
aI3
aI7
aI7
aI21
aI5
aI43
aI247
aI89
aI275
aI1219
aI311
aI5677
aI7161
aI13853
aI38613
aI84935
aI223563
asS'd'
I17118
sa(dp34236
S'a'
I62341
sS's'
I18
sS'm'
(lp34237
I1
aI3
aI1
aI5
aI29
aI61
aI17
aI235
aI127
aI979
aI973
aI1463
aI371
aI5567
aI6949
aI34165
aI3075
aI169347
asS'd'
I17119
sa(dp34238
S'a'
I62353
sS's'
I18
sS'm'
(lp34239
I1
aI3
aI3
aI15
aI25
aI51
aI43
aI73
aI7
aI123
aI1761
aI1461
aI5291
aI14271
aI19335
aI45379
aI123469
aI190439
asS'd'
I17120
sa(dp34240
S'a'
I62359
sS's'
I18
sS'm'
(lp34241
I1
aI3
aI3
aI13
aI19
aI57
aI25
aI161
aI351
aI703
aI819
aI753
aI3101
aI9043
aI19179
aI22665
aI118533
aI45817
asS'd'
I17121
sa(dp34242
S'a'
I62382
sS's'
I18
sS'm'
(lp34243
I1
aI3
aI5
aI15
aI3
aI33
aI15
aI63
aI251
aI87
aI611
aI1187
aI2639
aI6001
aI16135
aI27505
aI71077
aI34101
asS'd'
I17122
sa(dp34244
S'a'
I62389
sS's'
I18
sS'm'
(lp34245
I1
aI1
aI5
aI3
aI31
aI5
aI13
aI239
aI119
aI803
aI1881
aI3479
aI1933
aI6421
aI21411
aI62923
aI76851
aI211029
asS'd'
I17123
sa(dp34246
S'a'
I62396
sS's'
I18
sS'm'
(lp34247
I1
aI3
aI1
aI11
aI13
aI59
aI13
aI77
aI87
aI343
aI1733
aI3493
aI5937
aI15733
aI7763
aI12839
aI68639
aI70965
asS'd'
I17124
sa(dp34248
S'a'
I62425
sS's'
I18
sS'm'
(lp34249
I1
aI1
aI1
aI3
aI5
aI19
aI73
aI109
aI197
aI1007
aI1369
aI623
aI3249
aI9263
aI12463
aI37105
aI40599
aI115323
asS'd'
I17125
sa(dp34250
S'a'
I62455
sS's'
I18
sS'm'
(lp34251
I1
aI1
aI7
aI1
aI21
aI23
aI27
aI221
aI117
aI27
aI1811
aI837
aI7355
aI8083
aI12657
aI34137
aI102025
aI6511
asS'd'
I17126
sa(dp34252
S'a'
I62484
sS's'
I18
sS'm'
(lp34253
I1
aI3
aI7
aI13
aI13
aI29
aI7
aI103
aI511
aI449
aI1443
aI775
aI3503
aI1057
aI8809
aI48583
aI27649
aI206219
asS'd'
I17127
sa(dp34254
S'a'
I62494
sS's'
I18
sS'm'
(lp34255
I1
aI1
aI7
aI1
aI15
aI37
aI53
aI205
aI393
aI691
aI989
aI3493
aI7813
aI12371
aI18125
aI62569
aI57075
aI100625
asS'd'
I17128
sa(dp34256
S'a'
I62507
sS's'
I18
sS'm'
(lp34257
I1
aI3
aI5
aI7
aI11
aI11
aI55
aI7
aI487
aI861
aI1589
aI1003
aI607
aI10031
aI22481
aI41905
aI67791
aI168167
asS'd'
I17129
sa(dp34258
S'a'
I62512
sS's'
I18
sS'm'
(lp34259
I1
aI3
aI1
aI9
aI21
aI31
aI25
aI187
aI315
aI379
aI961
aI2721
aI3395
aI12321
aI21693
aI56977
aI73197
aI160023
asS'd'
I17130
sa(dp34260
S'a'
I62550
sS's'
I18
sS'm'
(lp34261
I1
aI3
aI3
aI7
aI9
aI25
aI103
aI1
aI13
aI1021
aI1777
aI1015
aI2269
aI2131
aI191
aI2561
aI74755
aI27131
asS'd'
I17131
sa(dp34262
S'a'
I62577
sS's'
I18
sS'm'
(lp34263
I1
aI1
aI3
aI13
aI21
aI29
aI97
aI153
aI499
aI207
aI719
aI585
aI8155
aI2873
aI22073
aI45933
aI92875
aI19205
asS'd'
I17132
sa(dp34264
S'a'
I62590
sS's'
I18
sS'm'
(lp34265
I1
aI3
aI7
aI15
aI13
aI31
aI43
aI223
aI405
aI839
aI1241
aI2219
aI6911
aI9469
aI24477
aI63157
aI95503
aI128431
asS'd'
I17133
sa(dp34266
S'a'
I62599
sS's'
I18
sS'm'
(lp34267
I1
aI3
aI5
aI15
aI5
aI11
aI79
aI129
aI235
aI171
aI289
aI1791
aI6061
aI9107
aI13859
aI55923
aI30197
aI111025
asS'd'
I17134
sa(dp34268
S'a'
I62617
sS's'
I18
sS'm'
(lp34269
I1
aI1
aI7
aI11
aI13
aI23
aI51
aI139
aI219
aI467
aI1923
aI2847
aI1977
aI1503
aI1939
aI55579
aI65357
aI50047
asS'd'
I17135
sa(dp34270
S'a'
I62679
sS's'
I18
sS'm'
(lp34271
I1
aI3
aI7
aI11
aI27
aI25
aI91
aI95
aI73
aI189
aI1537
aI273
aI725
aI1215
aI15255
aI18847
aI67419
aI162153
asS'd'
I17136
sa(dp34272
S'a'
I62702
sS's'
I18
sS'm'
(lp34273
I1
aI3
aI3
aI11
aI3
aI63
aI49
aI131
aI219
aI285
aI819
aI2801
aI2645
aI2943
aI15055
aI15659
aI130641
aI82913
asS'd'
I17137
sa(dp34274
S'a'
I62745
sS's'
I18
sS'm'
(lp34275
I1
aI1
aI3
aI7
aI17
aI19
aI37
aI59
aI391
aI1009
aI1569
aI2569
aI2519
aI33
aI18827
aI23277
aI94797
aI103673
asS'd'
I17138
sa(dp34276
S'a'
I62772
sS's'
I18
sS'm'
(lp34277
I1
aI3
aI5
aI9
aI27
aI57
aI69
aI185
aI49
aI829
aI29
aI1247
aI6129
aI14935
aI8005
aI48343
aI55789
aI170099
asS'd'
I17139
sa(dp34278
S'a'
I62794
sS's'
I18
sS'm'
(lp34279
I1
aI3
aI3
aI7
aI19
aI55
aI77
aI231
aI79
aI787
aI1597
aI2701
aI4999
aI4247
aI31849
aI7797
aI118993
aI77871
asS'd'
I17140
sa(dp34280
S'a'
I62835
sS's'
I18
sS'm'
(lp34281
I1
aI1
aI7
aI13
aI5
aI45
aI105
aI137
aI239
aI923
aI593
aI3227
aI3603
aI15463
aI15533
aI55285
aI95295
aI141951
asS'd'
I17141
sa(dp34282
S'a'
I62842
sS's'
I18
sS'm'
(lp34283
I1
aI3
aI1
aI5
aI29
aI3
aI113
aI241
aI255
aI181
aI1933
aI2579
aI1865
aI11083
aI8023
aI34271
aI78603
aI240781
asS'd'
I17142
sa(dp34284
S'a'
I62847
sS's'
I18
sS'm'
(lp34285
I1
aI1
aI3
aI7
aI17
aI21
aI123
aI75
aI305
aI485
aI9
aI3037
aI677
aI8001
aI16803
aI25851
aI121773
aI77729
asS'd'
I17143
sa(dp34286
S'a'
I62857
sS's'
I18
sS'm'
(lp34287
I1
aI3
aI1
aI5
aI23
aI7
aI39
aI25
aI381
aI1003
aI361
aI995
aI1751
aI9599
aI6399
aI9627
aI19303
aI249899
asS'd'
I17144
sa(dp34288
S'a'
I62894
sS's'
I18
sS'm'
(lp34289
I1
aI3
aI5
aI5
aI13
aI39
aI65
aI145
aI351
aI135
aI981
aI3657
aI4711
aI13649
aI17253
aI46443
aI99187
aI176683
asS'd'
I17145
sa(dp34290
S'a'
I62911
sS's'
I18
sS'm'
(lp34291
I1
aI3
aI1
aI9
aI9
aI41
aI79
aI237
aI445
aI507
aI1947
aI2905
aI8161
aI715
aI24499
aI62397
aI26393
aI197221
asS'd'
I17146
sa(dp34292
S'a'
I62928
sS's'
I18
sS'm'
(lp34293
I1
aI3
aI5
aI3
aI23
aI9
aI107
aI121
aI59
aI265
aI177
aI3495
aI391
aI4537
aI32099
aI45217
aI128285
aI259285
asS'd'
I17147
sa(dp34294
S'a'
I62954
sS's'
I18
sS'm'
(lp34295
I1
aI3
aI3
aI15
aI5
aI61
aI87
aI209
aI139
aI461
aI485
aI3261
aI7425
aI6193
aI22221
aI22145
aI93989
aI101459
asS'd'
I17148
sa(dp34296
S'a'
I62964
sS's'
I18
sS'm'
(lp34297
I1
aI3
aI1
aI1
aI15
aI51
aI29
aI145
aI385
aI695
aI375
aI3743
aI1387
aI15385
aI7995
aI22993
aI64115
aI239897
asS'd'
I17149
sa(dp34298
S'a'
I62977
sS's'
I18
sS'm'
(lp34299
I1
aI1
aI5
aI15
aI9
aI11
aI73
aI219
aI293
aI941
aI477
aI3935
aI2717
aI9559
aI20537
aI6935
aI39711
aI13623
asS'd'
I17150
sa(dp34300
S'a'
I62984
sS's'
I18
sS'm'
(lp34301
I1
aI3
aI3
aI11
aI3
aI23
aI127
aI21
aI61
aI59
aI1685
aI507
aI3883
aI6587
aI6355
aI65407
aI54311
aI228555
asS'd'
I17151
sa(dp34302
S'a'
I63007
sS's'
I18
sS'm'
(lp34303
I1
aI3
aI1
aI1
aI25
aI47
aI51
aI111
aI77
aI871
aI1045
aI4017
aI7683
aI7729
aI24155
aI3481
aI31749
aI245155
asS'd'
I17152
sa(dp34304
S'a'
I63017
sS's'
I18
sS'm'
(lp34305
I1
aI3
aI5
aI1
aI25
aI29
aI119
aI131
aI475
aI763
aI1639
aI1937
aI7387
aI2307
aI24081
aI34797
aI91785
aI52055
asS'd'
I17153
sa(dp34306
S'a'
I63058
sS's'
I18
sS'm'
(lp34307
I1
aI1
aI5
aI1
aI29
aI19
aI119
aI111
aI119
aI751
aI1079
aI1911
aI4085
aI8909
aI4351
aI30037
aI37691
aI57175
asS'd'
I17154
sa(dp34308
S'a'
I63067
sS's'
I18
sS'm'
(lp34309
I1
aI1
aI7
aI7
aI27
aI33
aI71
aI189
aI105
aI821
aI1543
aI2939
aI3829
aI6485
aI22235
aI7097
aI76987
aI207121
asS'd'
I17155
sa(dp34310
S'a'
I63085
sS's'
I18
sS'm'
(lp34311
I1
aI1
aI3
aI3
aI7
aI7
aI65
aI121
aI355
aI405
aI1019
aI1779
aI7301
aI10609
aI25927
aI16501
aI37287
aI133383
asS'd'
I17156
sa(dp34312
S'a'
I63150
sS's'
I18
sS'm'
(lp34313
I1
aI1
aI3
aI11
aI31
aI57
aI109
aI197
aI165
aI711
aI271
aI653
aI5835
aI14905
aI26065
aI52287
aI106215
aI225075
asS'd'
I17157
sa(dp34314
S'a'
I63170
sS's'
I18
sS'm'
(lp34315
I1
aI1
aI3
aI3
aI1
aI41
aI5
aI169
aI15
aI49
aI1311
aI2715
aI579
aI1693
aI28001
aI17935
aI18585
aI123531
asS'd'
I17158
sa(dp34316
S'a'
I63184
sS's'
I18
sS'm'
(lp34317
I1
aI1
aI7
aI7
aI1
aI49
aI59
aI75
aI173
aI361
aI1947
aI2707
aI1835
aI12025
aI24051
aI24359
aI121841
aI215797
asS'd'
I17159
sa(dp34318
S'a'
I63210
sS's'
I18
sS'm'
(lp34319
I1
aI1
aI5
aI13
aI7
aI49
aI15
aI181
aI409
aI1005
aI383
aI3449
aI2987
aI13051
aI7097
aI34571
aI55495
aI65251
asS'd'
I17160
sa(dp34320
S'a'
I63215
sS's'
I18
sS'm'
(lp34321
I1
aI1
aI3
aI11
aI5
aI9
aI67
aI41
aI9
aI79
aI401
aI379
aI4107
aI5231
aI519
aI47877
aI17273
aI137479
asS'd'
I17161
sa(dp34322
S'a'
I63217
sS's'
I18
sS'm'
(lp34323
I1
aI1
aI3
aI13
aI25
aI7
aI9
aI165
aI103
aI37
aI1369
aI933
aI1119
aI1025
aI19767
aI25765
aI55487
aI249709
asS'd'
I17162
sa(dp34324
S'a'
I63229
sS's'
I18
sS'm'
(lp34325
I1
aI3
aI5
aI5
aI19
aI53
aI105
aI135
aI245
aI957
aI185
aI2901
aI1741
aI10429
aI747
aI23365
aI49363
aI84095
asS'd'
I17163
sa(dp34326
S'a'
I63237
sS's'
I18
sS'm'
(lp34327
I1
aI1
aI1
aI15
aI29
aI17
aI107
aI193
aI17
aI447
aI1261
aI1935
aI5749
aI2303
aI23287
aI59883
aI28655
aI188055
asS'd'
I17164
sa(dp34328
S'a'
I63259
sS's'
I18
sS'm'
(lp34329
I1
aI1
aI5
aI9
aI13
aI27
aI99
aI253
aI299
aI481
aI89
aI3041
aI1549
aI15417
aI30495
aI2063
aI53649
aI219883
asS'd'
I17165
sa(dp34330
S'a'
I63265
sS's'
I18
sS'm'
(lp34331
I1
aI3
aI3
aI15
aI19
aI19
aI7
aI149
aI67
aI349
aI789
aI129
aI2783
aI2887
aI28631
aI26001
aI62407
aI151767
asS'd'
I17166
sa(dp34332
S'a'
I63271
sS's'
I18
sS'm'
(lp34333
I1
aI3
aI3
aI13
aI7
aI29
aI65
aI25
aI93
aI627
aI301
aI721
aI7249
aI13295
aI19995
aI33715
aI36441
aI157625
asS'd'
I17167
sa(dp34334
S'a'
I63286
sS's'
I18
sS'm'
(lp34335
I1
aI1
aI1
aI3
aI29
aI63
aI85
aI27
aI507
aI543
aI1887
aI3169
aI4239
aI4455
aI22047
aI15369
aI48913
aI192071
asS'd'
I17168
sa(dp34336
S'a'
I63298
sS's'
I18
sS'm'
(lp34337
I1
aI3
aI7
aI5
aI9
aI33
aI125
aI41
aI7
aI723
aI1091
aI3311
aI8173
aI3861
aI31507
aI42669
aI68853
aI60043
asS'd'
I17169
sa(dp34338
S'a'
I63343
sS's'
I18
sS'm'
(lp34339
I1
aI3
aI1
aI13
aI13
aI7
aI121
aI41
aI181
aI913
aI371
aI163
aI7061
aI8779
aI18345
aI41915
aI1785
aI107113
asS'd'
I17170
sa(dp34340
S'a'
I63355
sS's'
I18
sS'm'
(lp34341
I1
aI1
aI1
aI9
aI13
aI41
aI23
aI35
aI157
aI247
aI1243
aI1101
aI5193
aI4027
aI29917
aI44099
aI46211
aI162059
asS'd'
I17171
sa(dp34342
S'a'
I63361
sS's'
I18
sS'm'
(lp34343
I1
aI3
aI1
aI5
aI15
aI51
aI3
aI241
aI131
aI741
aI1885
aI2397
aI5673
aI9097
aI9319
aI15381
aI55655
aI207569
asS'd'
I17172
sa(dp34344
S'a'
I63362
sS's'
I18
sS'm'
(lp34345
I1
aI1
aI3
aI15
aI25
aI15
aI69
aI55
aI435
aI727
aI1007
aI375
aI7871
aI10437
aI11011
aI36711
aI11269
aI105159
asS'd'
I17173
sa(dp34346
S'a'
I63379
sS's'
I18
sS'm'
(lp34347
I1
aI3
aI3
aI13
aI17
aI1
aI101
aI189
aI295
aI185
aI1715
aI2609
aI6767
aI11751
aI11469
aI3951
aI80743
aI114439
asS'd'
I17174
sa(dp34348
S'a'
I63397
sS's'
I18
sS'm'
(lp34349
I1
aI3
aI1
aI7
aI21
aI41
aI93
aI39
aI433
aI917
aI279
aI161
aI267
aI10201
aI26583
aI30363
aI110187
aI46501
asS'd'
I17175
sa(dp34350
S'a'
I63415
sS's'
I18
sS'm'
(lp34351
I1
aI1
aI7
aI7
aI13
aI15
aI89
aI167
aI365
aI925
aI107
aI3537
aI6815
aI15251
aI23149
aI61821
aI66569
aI135353
asS'd'
I17176
sa(dp34352
S'a'
I63421
sS's'
I18
sS'm'
(lp34353
I1
aI1
aI1
aI3
aI13
aI59
aI21
aI255
aI111
aI603
aI547
aI465
aI3001
aI16055
aI26389
aI64301
aI112751
aI219279
asS'd'
I17177
sa(dp34354
S'a'
I63463
sS's'
I18
sS'm'
(lp34355
I1
aI1
aI1
aI15
aI17
aI3
aI21
aI49
aI327
aI349
aI489
aI957
aI807
aI11685
aI23975
aI34729
aI100773
aI223551
asS'd'
I17178
sa(dp34356
S'a'
I63491
sS's'
I18
sS'm'
(lp34357
I1
aI3
aI5
aI15
aI19
aI59
aI63
aI71
aI233
aI767
aI1789
aI3609
aI5911
aI3405
aI7519
aI3611
aI92015
aI126669
asS'd'
I17179
sa(dp34358
S'a'
I63527
sS's'
I18
sS'm'
(lp34359
I1
aI1
aI1
aI11
aI7
aI31
aI79
aI57
aI115
aI763
aI1643
aI3329
aI7209
aI1385
aI15565
aI64353
aI60637
aI59445
asS'd'
I17180
sa(dp34360
S'a'
I63545
sS's'
I18
sS'm'
(lp34361
I1
aI1
aI3
aI1
aI13
aI3
aI47
aI89
aI507
aI523
aI1
aI1391
aI6973
aI7267
aI32527
aI52631
aI20775
aI234503
asS'd'
I17181
sa(dp34362
S'a'
I63553
sS's'
I18
sS'm'
(lp34363
I1
aI3
aI1
aI9
aI23
aI23
aI95
aI57
aI295
aI857
aI213
aI1211
aI3503
aI3043
aI24843
aI16149
aI118719
aI171585
asS'd'
I17182
sa(dp34364
S'a'
I63560
sS's'
I18
sS'm'
(lp34365
I1
aI3
aI1
aI5
aI1
aI13
aI63
aI167
aI305
aI711
aI759
aI2521
aI5051
aI9125
aI22917
aI24647
aI100777
aI261137
asS'd'
I17183
sa(dp34366
S'a'
I63563
sS's'
I18
sS'm'
(lp34367
I1
aI1
aI5
aI9
aI25
aI19
aI5
aI225
aI511
aI543
aI685
aI733
aI7249
aI10447
aI11115
aI25927
aI104327
aI92861
asS'd'
I17184
sa(dp34368
S'a'
I63566
sS's'
I18
sS'm'
(lp34369
I1
aI3
aI1
aI7
aI15
aI7
aI15
aI83
aI379
aI461
aI943
aI317
aI7735
aI12655
aI7549
aI6371
aI20901
aI170331
asS'd'
I17185
sa(dp34370
S'a'
I63589
sS's'
I18
sS'm'
(lp34371
I1
aI1
aI1
aI13
aI7
aI17
aI41
aI51
aI47
aI15
aI477
aI1203
aI819
aI1615
aI13805
aI40147
aI3967
aI192647
asS'd'
I17186
sa(dp34372
S'a'
I63599
sS's'
I18
sS'm'
(lp34373
I1
aI3
aI7
aI11
aI9
aI11
aI111
aI75
aI171
aI833
aI1503
aI2325
aI7279
aI2687
aI16499
aI11547
aI99409
aI186429
asS'd'
I17187
sa(dp34374
S'a'
I63601
sS's'
I18
sS'm'
(lp34375
I1
aI1
aI5
aI3
aI13
aI21
aI75
aI17
aI447
aI647
aI1309
aI2297
aI7911
aI12093
aI16237
aI50831
aI96123
aI134479
asS'd'
I17188
sa(dp34376
S'a'
I63608
sS's'
I18
sS'm'
(lp34377
I1
aI1
aI5
aI15
aI19
aI29
aI35
aI255
aI291
aI437
aI85
aI2143
aI3281
aI3629
aI29339
aI28169
aI46561
aI236595
asS'd'
I17189
sa(dp34378
S'a'
I63620
sS's'
I18
sS'm'
(lp34379
I1
aI3
aI7
aI1
aI31
aI57
aI125
aI109
aI317
aI461
aI681
aI1379
aI6387
aI14971
aI8451
aI17655
aI87619
aI51721
asS'd'
I17190
sa(dp34380
S'a'
I63654
sS's'
I18
sS'm'
(lp34381
I1
aI3
aI7
aI1
aI23
aI33
aI45
aI149
aI43
aI465
aI997
aI601
aI693
aI6273
aI12867
aI25885
aI81353
aI60437
asS'd'
I17191
sa(dp34382
S'a'
I63703
sS's'
I18
sS'm'
(lp34383
I1
aI1
aI7
aI7
aI31
aI25
aI113
aI205
aI481
aI141
aI1757
aI587
aI2981
aI7637
aI3869
aI4151
aI69541
aI68587
asS'd'
I17192
sa(dp34384
S'a'
I63719
sS's'
I18
sS'm'
(lp34385
I1
aI1
aI7
aI13
aI17
aI31
aI69
aI247
aI137
aI79
aI1221
aI1693
aI3747
aI10711
aI1671
aI31587
aI12139
aI248585
asS'd'
I17193
sa(dp34386
S'a'
I63758
sS's'
I18
sS'm'
(lp34387
I1
aI1
aI1
aI3
aI1
aI61
aI39
aI139
aI37
aI79
aI125
aI1145
aI7505
aI10129
aI29209
aI52045
aI99159
aI195553
asS'd'
I17194
sa(dp34388
S'a'
I63769
sS's'
I18
sS'm'
(lp34389
I1
aI1
aI5
aI3
aI13
aI41
aI13
aI11
aI167
aI953
aI1961
aI3557
aI871
aI1687
aI28479
aI10621
aI27533
aI243519
asS'd'
I17195
sa(dp34390
S'a'
I63782
sS's'
I18
sS'm'
(lp34391
I1
aI1
aI7
aI7
aI1
aI35
aI107
aI227
aI375
aI225
aI483
aI1239
aI7591
aI8549
aI7351
aI62001
aI70245
aI102795
asS'd'
I17196
sa(dp34392
S'a'
I63803
sS's'
I18
sS'm'
(lp34393
I1
aI1
aI1
aI9
aI29
aI35
aI15
aI3
aI337
aI1017
aI1065
aI2107
aI2457
aI9455
aI7069
aI55081
aI57887
aI149679
asS'd'
I17197
sa(dp34394
S'a'
I63811
sS's'
I18
sS'm'
(lp34395
I1
aI1
aI7
aI1
aI1
aI1
aI13
aI63
aI287
aI895
aI593
aI1253
aI4717
aI10313
aI10275
aI22143
aI59149
aI38865
asS'd'
I17198
sa(dp34396
S'a'
I63851
sS's'
I18
sS'm'
(lp34397
I1
aI1
aI3
aI15
aI27
aI39
aI73
aI11
aI509
aI391
aI1901
aI503
aI5523
aI6777
aI30849
aI41301
aI35067
aI68443
asS'd'
I17199
sa(dp34398
S'a'
I63854
sS's'
I18
sS'm'
(lp34399
I1
aI3
aI3
aI15
aI17
aI57
aI39
aI229
aI273
aI917
aI577
aI3627
aI3285
aI4495
aI28581
aI34011
aI38537
aI194999
asS'd'
I17200
sa(dp34400
S'a'
I63872
sS's'
I18
sS'm'
(lp34401
I1
aI1
aI7
aI3
aI17
aI51
aI91
aI203
aI161
aI757
aI581
aI1625
aI477
aI8839
aI16515
aI43101
aI121497
aI23603
asS'd'
I17201
sa(dp34402
S'a'
I63878
sS's'
I18
sS'm'
(lp34403
I1
aI3
aI3
aI5
aI19
aI29
aI55
aI127
aI283
aI999
aI1227
aI1937
aI4471
aI11305
aI8813
aI40509
aI78521
aI175573
asS'd'
I17202
sa(dp34404
S'a'
I63895
sS's'
I18
sS'm'
(lp34405
I1
aI3
aI3
aI3
aI5
aI29
aI33
aI249
aI25
aI213
aI1315
aI393
aI6967
aI12751
aI7485
aI39561
aI14801
aI191921
asS'd'
I17203
sa(dp34406
S'a'
I63917
sS's'
I18
sS'm'
(lp34407
I1
aI3
aI7
aI9
aI23
aI15
aI93
aI69
aI23
aI239
aI1993
aI3375
aI539
aI14141
aI10123
aI33561
aI127565
aI181527
asS'd'
I17204
sa(dp34408
S'a'
I63930
sS's'
I18
sS'm'
(lp34409
I1
aI3
aI3
aI15
aI13
aI15
aI65
aI241
aI83
aI351
aI1943
aI1305
aI7181
aI11803
aI31907
aI63623
aI5439
aI150661
asS'd'
I17205
sa(dp34410
S'a'
I63935
sS's'
I18
sS'm'
(lp34411
I1
aI3
aI7
aI11
aI13
aI17
aI17
aI37
aI409
aI577
aI973
aI797
aI1761
aI5333
aI13803
aI22991
aI29743
aI53051
asS'd'
I17206
sa(dp34412
S'a'
I63955
sS's'
I18
sS'm'
(lp34413
I1
aI3
aI5
aI5
aI27
aI25
aI91
aI225
aI411
aI23
aI877
aI2487
aI8061
aI12337
aI11471
aI8857
aI10791
aI112699
asS'd'
I17207
sa(dp34414
S'a'
I63964
sS's'
I18
sS'm'
(lp34415
I1
aI1
aI5
aI3
aI15
aI1
aI87
aI249
aI205
aI1011
aI2045
aI1879
aI4137
aI5877
aI12709
aI5231
aI74283
aI124315
asS'd'
I17208
sa(dp34416
S'a'
I63967
sS's'
I18
sS'm'
(lp34417
I1
aI1
aI7
aI7
aI31
aI37
aI117
aI71
aI139
aI391
aI1085
aI4033
aI3087
aI3063
aI19991
aI8787
aI96899
aI17279
asS'd'
I17209
sa(dp34418
S'a'
I63980
sS's'
I18
sS'm'
(lp34419
I1
aI1
aI7
aI15
aI19
aI47
aI45
aI181
aI303
aI151
aI337
aI2557
aI6131
aI3161
aI13097
aI52777
aI77783
aI259817
asS'd'
I17210
sa(dp34420
S'a'
I63985
sS's'
I18
sS'm'
(lp34421
I1
aI1
aI3
aI3
aI1
aI55
aI115
aI227
aI83
aI591
aI967
aI4067
aI3441
aI243
aI13443
aI4043
aI129365
aI161459
asS'd'
I17211
sa(dp34422
S'a'
I63992
sS's'
I18
sS'm'
(lp34423
I1
aI1
aI3
aI15
aI5
aI23
aI71
aI31
aI271
aI585
aI931
aI909
aI3375
aI15063
aI12111
aI35811
aI124047
aI68225
asS'd'
I17212
sa(dp34424
S'a'
I64021
sS's'
I18
sS'm'
(lp34425
I1
aI1
aI5
aI11
aI1
aI59
aI19
aI193
aI323
aI489
aI837
aI3709
aI1807
aI11617
aI30931
aI33561
aI2805
aI100979
asS'd'
I17213
sa(dp34426
S'a'
I64026
sS's'
I18
sS'm'
(lp34427
I1
aI3
aI5
aI7
aI27
aI7
aI71
aI67
aI167
aI521
aI1237
aI2911
aI3531
aI2885
aI4669
aI25703
aI87647
aI36381
asS'd'
I17214
sa(dp34428
S'a'
I64044
sS's'
I18
sS'm'
(lp34429
I1
aI3
aI3
aI13
aI13
aI21
aI97
aI225
aI477
aI1023
aI2029
aI877
aI3849
aI4675
aI17665
aI19257
aI9697
aI168577
asS'd'
I17215
sa(dp34430
S'a'
I64055
sS's'
I18
sS'm'
(lp34431
I1
aI1
aI7
aI15
aI25
aI31
aI19
aI255
aI45
aI539
aI1831
aI2655
aI7471
aI12011
aI12455
aI3681
aI123881
aI234471
asS'd'
I17216
sa(dp34432
S'a'
I64056
sS's'
I18
sS'm'
(lp34433
I1
aI1
aI3
aI9
aI17
aI39
aI105
aI73
aI271
aI555
aI987
aI873
aI5371
aI12381
aI13469
aI54961
aI125701
aI194063
asS'd'
I17217
sa(dp34434
S'a'
I64061
sS's'
I18
sS'm'
(lp34435
I1
aI1
aI5
aI5
aI7
aI27
aI15
aI195
aI121
aI175
aI991
aI955
aI5007
aI11423
aI1539
aI21381
aI79891
aI162149
asS'd'
I17218
sa(dp34436
S'a'
I64067
sS's'
I18
sS'm'
(lp34437
I1
aI3
aI1
aI7
aI25
aI23
aI69
aI69
aI177
aI545
aI481
aI3503
aI3721
aI1077
aI8763
aI6919
aI64743
aI172311
asS'd'
I17219
sa(dp34438
S'a'
I64081
sS's'
I18
sS'm'
(lp34439
I1
aI1
aI3
aI15
aI31
aI5
aI33
aI45
aI81
aI795
aI435
aI399
aI4591
aI3741
aI26493
aI14791
aI59529
aI89989
asS'd'
I17220
sa(dp34440
S'a'
I64100
sS's'
I18
sS'm'
(lp34441
I1
aI1
aI7
aI13
aI21
aI29
aI95
aI75
aI213
aI59
aI1635
aI479
aI441
aI14667
aI16389
aI9139
aI30955
aI169895
asS'd'
I17221
sa(dp34442
S'a'
I64109
sS's'
I18
sS'm'
(lp34443
I1
aI3
aI3
aI3
aI17
aI61
aI103
aI85
aI233
aI287
aI447
aI2687
aI4755
aI9489
aI1669
aI10405
aI58489
aI170429
asS'd'
I17222
sa(dp34444
S'a'
I64112
sS's'
I18
sS'm'
(lp34445
I1
aI1
aI5
aI5
aI13
aI9
aI63
aI129
aI321
aI531
aI393
aI3353
aI5309
aI16375
aI20473
aI12595
aI52239
aI183647
asS'd'
I17223
sa(dp34446
S'a'
I64118
sS's'
I18
sS'm'
(lp34447
I1
aI1
aI3
aI7
aI7
aI31
aI101
aI253
aI119
aI325
aI351
aI2321
aI1899
aI14073
aI8985
aI13609
aI32043
aI33225
asS'd'
I17224
sa(dp34448
S'a'
I64124
sS's'
I18
sS'm'
(lp34449
I1
aI1
aI3
aI13
aI7
aI25
aI73
aI191
aI399
aI591
aI819
aI2859
aI6053
aI815
aI30417
aI5709
aI18277
aI121991
asS'd'
I17225
sa(dp34450
S'a'
I64145
sS's'
I18
sS'm'
(lp34451
I1
aI3
aI3
aI1
aI7
aI47
aI7
aI81
aI451
aI463
aI699
aI1857
aI8169
aI15649
aI22693
aI28673
aI9717
aI227583
asS'd'
I17226
sa(dp34452
S'a'
I64151
sS's'
I18
sS'm'
(lp34453
I1
aI3
aI3
aI3
aI31
aI45
aI123
aI205
aI23
aI901
aI1003
aI1149
aI7481
aI6925
aI23845
aI18573
aI97047
aI248957
asS'd'
I17227
sa(dp34454
S'a'
I64203
sS's'
I18
sS'm'
(lp34455
I1
aI1
aI3
aI3
aI21
aI9
aI53
aI241
aI125
aI583
aI1055
aI3981
aI8113
aI12477
aI8455
aI6289
aI112253
aI17321
asS'd'
I17228
sa(dp34456
S'a'
I64208
sS's'
I18
sS'm'
(lp34457
I1
aI1
aI7
aI3
aI19
aI5
aI51
aI111
aI443
aI283
aI117
aI2127
aI4273
aI2335
aI20373
aI2885
aI57439
aI56839
asS'd'
I17229
sa(dp34458
S'a'
I64236
sS's'
I18
sS'm'
(lp34459
I1
aI1
aI5
aI15
aI7
aI5
aI65
aI163
aI27
aI691
aI1667
aI69
aI2459
aI7477
aI21349
aI52417
aI42299
aI75965
asS'd'
I17230
sa(dp34460
S'a'
I64254
sS's'
I18
sS'm'
(lp34461
I1
aI1
aI1
aI13
aI13
aI19
aI87
aI223
aI475
aI205
aI1113
aI887
aI2213
aI5533
aI15875
aI36173
aI53933
aI200173
asS'd'
I17231
sa(dp34462
S'a'
I64261
sS's'
I18
sS'm'
(lp34463
I1
aI3
aI3
aI13
aI23
aI17
aI93
aI37
aI391
aI127
aI873
aI1445
aI3007
aI10863
aI21245
aI55025
aI99275
aI255329
asS'd'
I17232
sa(dp34464
S'a'
I64280
sS's'
I18
sS'm'
(lp34465
I1
aI3
aI1
aI15
aI1
aI47
aI57
aI5
aI207
aI825
aI161
aI539
aI6151
aI12829
aI14121
aI51217
aI25547
aI234303
asS'd'
I17233
sa(dp34466
S'a'
I64296
sS's'
I18
sS'm'
(lp34467
I1
aI1
aI5
aI1
aI21
aI63
aI15
aI83
aI19
aI817
aI591
aI3131
aI889
aI12451
aI14363
aI27295
aI83877
aI124701
asS'd'
I17234
sa(dp34468
S'a'
I64313
sS's'
I18
sS'm'
(lp34469
I1
aI1
aI7
aI7
aI11
aI21
aI87
aI85
aI13
aI555
aI163
aI9
aI5973
aI14749
aI19585
aI57287
aI43421
aI66301
asS'd'
I17235
sa(dp34470
S'a'
I64314
sS's'
I18
sS'm'
(lp34471
I1
aI3
aI5
aI5
aI7
aI33
aI19
aI7
aI9
aI819
aI533
aI2105
aI4275
aI10611
aI30517
aI35863
aI84687
aI245157
asS'd'
I17236
sa(dp34472
S'a'
I64316
sS's'
I18
sS'm'
(lp34473
I1
aI3
aI5
aI13
aI3
aI55
aI111
aI157
aI235
aI405
aI39
aI2191
aI905
aI3099
aI245
aI37371
aI365
aI257385
asS'd'
I17237
sa(dp34474
S'a'
I64348
sS's'
I18
sS'm'
(lp34475
I1
aI3
aI3
aI13
aI29
aI39
aI125
aI235
aI213
aI879
aI497
aI1659
aI6689
aI12165
aI18621
aI14657
aI37079
aI167867
asS'd'
I17238
sa(dp34476
S'a'
I64352
sS's'
I18
sS'm'
(lp34477
I1
aI1
aI5
aI15
aI5
aI5
aI27
aI197
aI77
aI477
aI1115
aI3369
aI2253
aI5757
aI20855
aI4473
aI112501
aI76881
asS'd'
I17239
sa(dp34478
S'a'
I64355
sS's'
I18
sS'm'
(lp34479
I1
aI3
aI3
aI13
aI13
aI61
aI37
aI97
aI229
aI743
aI1381
aI3979
aI307
aI319
aI16765
aI56295
aI109303
aI21361
asS'd'
I17240
sa(dp34480
S'a'
I64361
sS's'
I18
sS'm'
(lp34481
I1
aI1
aI7
aI11
aI19
aI7
aI63
aI145
aI129
aI899
aI93
aI1851
aI7901
aI8767
aI15553
aI13913
aI4897
aI129483
asS'd'
I17241
sa(dp34482
S'a'
I64388
sS's'
I18
sS'm'
(lp34483
I1
aI3
aI5
aI11
aI7
aI23
aI19
aI5
aI465
aI365
aI883
aI3563
aI4395
aI2759
aI4273
aI623
aI75047
aI249519
asS'd'
I17242
sa(dp34484
S'a'
I64398
sS's'
I18
sS'm'
(lp34485
I1
aI3
aI1
aI5
aI29
aI43
aI75
aI7
aI509
aI373
aI359
aI2041
aI5957
aI1251
aI32431
aI37803
aI120915
aI45137
asS'd'
I17243
sa(dp34486
S'a'
I64403
sS's'
I18
sS'm'
(lp34487
I1
aI3
aI1
aI5
aI21
aI9
aI43
aI1
aI337
aI743
aI1359
aI1629
aI5117
aI2499
aI16129
aI22831
aI38795
aI32137
asS'd'
I17244
sa(dp34488
S'a'
I64419
sS's'
I18
sS'm'
(lp34489
I1
aI3
aI3
aI7
aI23
aI57
aI9
aI31
aI351
aI559
aI1729
aI1461
aI3037
aI12685
aI8899
aI14859
aI108851
aI170195
asS'd'
I17245
sa(dp34490
S'a'
I64421
sS's'
I18
sS'm'
(lp34491
I1
aI1
aI1
aI15
aI23
aI57
aI39
aI23
aI283
aI487
aI1055
aI1265
aI6781
aI7955
aI195
aI37745
aI66115
aI56413
asS'd'
I17246
sa(dp34492
S'a'
I64428
sS's'
I18
sS'm'
(lp34493
I1
aI1
aI3
aI7
aI27
aI35
aI57
aI17
aI137
aI17
aI905
aI4033
aI5775
aI5305
aI22975
aI17547
aI106297
aI146287
asS'd'
I17247
sa(dp34494
S'a'
I64453
sS's'
I18
sS'm'
(lp34495
I1
aI1
aI1
aI3
aI5
aI39
aI73
aI151
aI469
aI523
aI119
aI539
aI2817
aI7783
aI22957
aI59937
aI21331
aI172437
asS'd'
I17248
sa(dp34496
S'a'
I64463
sS's'
I18
sS'm'
(lp34497
I1
aI1
aI3
aI13
aI21
aI1
aI23
aI109
aI113
aI257
aI817
aI1671
aI6729
aI1571
aI15009
aI48539
aI94025
aI160379
asS'd'
I17249
sa(dp34498
S'a'
I64475
sS's'
I18
sS'm'
(lp34499
I1
aI1
aI1
aI1
aI31
aI23
aI83
aI107
aI225
aI715
aI949
aI69
aI2163
aI4777
aI7715
aI25901
aI82935
aI81455
asS'd'
I17250
sa(dp34500
S'a'
I64511
sS's'
I18
sS'm'
(lp34501
I1
aI3
aI5
aI13
aI29
aI11
aI61
aI169
aI241
aI973
aI315
aI3991
aI1389
aI3293
aI31123
aI59419
aI7359
aI170929
asS'd'
I17251
sa(dp34502
S'a'
I64537
sS's'
I18
sS'm'
(lp34503
I1
aI1
aI7
aI3
aI21
aI15
aI111
aI41
aI329
aI513
aI1175
aI4037
aI2747
aI11465
aI17253
aI54055
aI29409
aI230925
asS'd'
I17252
sa(dp34504
S'a'
I64576
sS's'
I18
sS'm'
(lp34505
I1
aI3
aI5
aI15
aI31
aI17
aI105
aI45
aI61
aI339
aI1387
aI1021
aI4499
aI13671
aI25521
aI52081
aI49153
aI31587
asS'd'
I17253
sa(dp34506
S'a'
I64579
sS's'
I18
sS'm'
(lp34507
I1
aI3
aI3
aI3
aI29
aI17
aI51
aI103
aI429
aI849
aI1759
aI1267
aI6255
aI4631
aI32643
aI44977
aI40875
aI239457
asS'd'
I17254
sa(dp34508
S'a'
I64581
sS's'
I18
sS'm'
(lp34509
I1
aI1
aI7
aI15
aI3
aI11
aI123
aI157
aI73
aI151
aI777
aI3855
aI1913
aI969
aI11821
aI16889
aI63503
aI197305
asS'd'
I17255
sa(dp34510
S'a'
I64606
sS's'
I18
sS'm'
(lp34511
I1
aI3
aI1
aI5
aI13
aI49
aI61
aI209
aI105
aI523
aI851
aI3667
aI7525
aI5537
aI12851
aI42867
aI50535
aI131403
asS'd'
I17256
sa(dp34512
S'a'
I64615
sS's'
I18
sS'm'
(lp34513
I1
aI3
aI1
aI7
aI13
aI19
aI107
aI71
aI479
aI895
aI405
aI89
aI1345
aI5543
aI12709
aI6093
aI97581
aI20483
asS'd'
I17257
sa(dp34514
S'a'
I64649
sS's'
I18
sS'm'
(lp34515
I1
aI3
aI3
aI11
aI3
aI47
aI117
aI175
aI175
aI321
aI1257
aI365
aI1193
aI12813
aI2713
aI26941
aI43605
aI223323
asS'd'
I17258
sa(dp34516
S'a'
I64655
sS's'
I18
sS'm'
(lp34517
I1
aI1
aI7
aI1
aI19
aI35
aI45
aI143
aI395
aI255
aI1599
aI575
aI2637
aI1287
aI27673
aI48329
aI57975
aI44173
asS'd'
I17259
sa(dp34518
S'a'
I64664
sS's'
I18
sS'm'
(lp34519
I1
aI3
aI5
aI1
aI1
aI19
aI107
aI233
aI465
aI661
aI91
aI4007
aI6409
aI3399
aI8175
aI54171
aI111417
aI124955
asS'd'
I17260
sa(dp34520
S'a'
I64667
sS's'
I18
sS'm'
(lp34521
I1
aI1
aI5
aI13
aI9
aI27
aI121
aI225
aI55
aI761
aI779
aI3015
aI6333
aI10779
aI26531
aI57103
aI33463
aI90219
asS'd'
I17261
sa(dp34522
S'a'
I64712
sS's'
I18
sS'm'
(lp34523
I1
aI3
aI1
aI9
aI17
aI3
aI85
aI147
aI111
aI133
aI869
aI1833
aI2401
aI5811
aI24415
aI27095
aI65529
aI164121
asS'd'
I17262
sa(dp34524
S'a'
I64715
sS's'
I18
sS'm'
(lp34525
I1
aI1
aI5
aI13
aI11
aI37
aI13
aI83
aI391
aI909
aI2013
aI1327
aI6697
aI1711
aI29265
aI10607
aI20127
aI57873
asS'd'
I17263
sa(dp34526
S'a'
I64718
sS's'
I18
sS'm'
(lp34527
I1
aI3
aI1
aI11
aI17
aI17
aI33
aI101
aI383
aI837
aI1769
aI1711
aI3735
aI14777
aI27101
aI56853
aI110643
aI101917
asS'd'
I17264
sa(dp34528
S'a'
I64741
sS's'
I18
sS'm'
(lp34529
I1
aI3
aI5
aI13
aI25
aI7
aI37
aI99
aI473
aI211
aI1469
aI1827
aI6307
aI8835
aI15853
aI22027
aI43095
aI15817
asS'd'
I17265
sa(dp34530
S'a'
I64766
sS's'
I18
sS'm'
(lp34531
I1
aI3
aI1
aI5
aI19
aI13
aI61
aI193
aI57
aI359
aI1277
aI749
aI5499
aI11239
aI20681
aI48477
aI7225
aI259259
asS'd'
I17266
sa(dp34532
S'a'
I64780
sS's'
I18
sS'm'
(lp34533
I1
aI1
aI1
aI7
aI27
aI17
aI79
aI213
aI307
aI761
aI429
aI1519
aI7483
aI6007
aI11251
aI13263
aI24851
aI7919
asS'd'
I17267
sa(dp34534
S'a'
I64788
sS's'
I18
sS'm'
(lp34535
I1
aI1
aI3
aI5
aI9
aI37
aI101
aI149
aI405
aI413
aI1213
aI157
aI3811
aI4485
aI13099
aI32697
aI75677
aI127815
asS'd'
I17268
sa(dp34536
S'a'
I64791
sS's'
I18
sS'm'
(lp34537
I1
aI3
aI3
aI7
aI29
aI29
aI13
aI113
aI45
aI885
aI1471
aI3433
aI2289
aI4375
aI815
aI16741
aI20933
aI9763
asS'd'
I17269
sa(dp34538
S'a'
I64795
sS's'
I18
sS'm'
(lp34539
I1
aI3
aI5
aI15
aI5
aI3
aI7
aI37
aI347
aI41
aI1977
aI395
aI6363
aI3591
aI21457
aI31455
aI60547
aI108153
asS'd'
I17270
sa(dp34540
S'a'
I64798
sS's'
I18
sS'm'
(lp34541
I1
aI3
aI5
aI7
aI19
aI9
aI113
aI1
aI241
aI439
aI731
aI1591
aI3347
aI1295
aI6635
aI25267
aI13239
aI214669
asS'd'
I17271
sa(dp34542
S'a'
I64816
sS's'
I18
sS'm'
(lp34543
I1
aI1
aI3
aI5
aI7
aI7
aI69
aI77
aI281
aI851
aI1533
aI1
aI7351
aI3429
aI29237
aI54597
aI11171
aI66613
asS'd'
I17272
sa(dp34544
S'a'
I64819
sS's'
I18
sS'm'
(lp34545
I1
aI1
aI7
aI15
aI29
aI7
aI59
aI9
aI105
aI129
aI1397
aI3841
aI3945
aI4755
aI19877
aI11109
aI17497
aI225473
asS'd'
I17273
sa(dp34546
S'a'
I64822
sS's'
I18
sS'm'
(lp34547
I1
aI3
aI7
aI7
aI15
aI61
aI3
aI207
aI97
aI229
aI1251
aI101
aI3157
aI5729
aI15579
aI14849
aI119119
aI91891
asS'd'
I17274
sa(dp34548
S'a'
I64833
sS's'
I18
sS'm'
(lp34549
I1
aI3
aI3
aI9
aI27
aI15
aI85
aI221
aI231
aI577
aI1787
aI3489
aI2393
aI7593
aI13175
aI25561
aI108505
aI97267
asS'd'
I17275
sa(dp34550
S'a'
I64836
sS's'
I18
sS'm'
(lp34551
I1
aI3
aI7
aI13
aI23
aI3
aI7
aI85
aI307
aI899
aI371
aI3539
aI3467
aI7955
aI9539
aI53583
aI125587
aI30969
asS'd'
I17276
sa(dp34552
S'a'
I64876
sS's'
I18
sS'm'
(lp34553
I1
aI1
aI5
aI7
aI7
aI31
aI115
aI245
aI375
aI803
aI1121
aI3775
aI3565
aI15283
aI25981
aI24681
aI34469
aI172003
asS'd'
I17277
sa(dp34554
S'a'
I64891
sS's'
I18
sS'm'
(lp34555
I1
aI1
aI5
aI1
aI31
aI5
aI5
aI161
aI153
aI235
aI1703
aI2163
aI1089
aI16233
aI6183
aI25167
aI102925
aI36673
asS'd'
I17278
sa(dp34556
S'a'
I64907
sS's'
I18
sS'm'
(lp34557
I1
aI3
aI3
aI5
aI1
aI57
aI59
aI5
aI87
aI497
aI151
aI1731
aI2727
aI4583
aI28165
aI63053
aI76003
aI29259
asS'd'
I17279
sa(dp34558
S'a'
I64943
sS's'
I18
sS'm'
(lp34559
I1
aI1
aI5
aI7
aI5
aI21
aI79
aI111
aI347
aI879
aI827
aI3947
aI4421
aI9589
aI23971
aI11681
aI104555
aI226535
asS'd'
I17280
sa(dp34560
S'a'
I64965
sS's'
I18
sS'm'
(lp34561
I1
aI1
aI5
aI1
aI17
aI35
aI105
aI159
aI391
aI495
aI1709
aI3731
aI261
aI2359
aI1413
aI37105
aI8979
aI189381
asS'd'
I17281
sa(dp34562
S'a'
I64977
sS's'
I18
sS'm'
(lp34563
I1
aI3
aI3
aI1
aI11
aI23
aI21
aI213
aI261
aI755
aI1503
aI2369
aI1765
aI14531
aI2605
aI15609
aI48691
aI113059
asS'd'
I17282
sa(dp34564
S'a'
I64999
sS's'
I18
sS'm'
(lp34565
I1
aI3
aI5
aI1
aI1
aI55
aI87
aI197
aI89
aI391
aI1157
aI3523
aI385
aI5871
aI13681
aI29097
aI101903
aI184553
asS'd'
I17283
sa(dp34566
S'a'
I65034
sS's'
I18
sS'm'
(lp34567
I1
aI3
aI7
aI5
aI17
aI51
aI87
aI191
aI495
aI761
aI1943
aI1845
aI2963
aI13133
aI22439
aI20101
aI96759
aI215215
asS'd'
I17284
sa(dp34568
S'a'
I65036
sS's'
I18
sS'm'
(lp34569
I1
aI3
aI5
aI11
aI23
aI9
aI53
aI41
aI229
aI233
aI2025
aI2835
aI2359
aI4755
aI3015
aI48267
aI20721
aI61001
asS'd'
I17285
sa(dp34570
S'a'
I65089
sS's'
I18
sS'm'
(lp34571
I1
aI1
aI7
aI9
aI27
aI35
aI45
aI201
aI137
aI291
aI151
aI733
aI6199
aI3127
aI3073
aI14491
aI95051
aI12469
asS'd'
I17286
sa(dp34572
S'a'
I65107
sS's'
I18
sS'm'
(lp34573
I1
aI1
aI1
aI9
aI7
aI49
aI73
aI233
aI239
aI881
aI1991
aI695
aI5947
aI9377
aI12027
aI41137
aI80217
aI122961
asS'd'
I17287
sa(dp34574
S'a'
I65110
sS's'
I18
sS'm'
(lp34575
I1
aI1
aI5
aI11
aI3
aI15
aI85
aI203
aI305
aI945
aI1007
aI1831
aI3999
aI373
aI21141
aI63829
aI91779
aI122495
asS'd'
I17288
sa(dp34576
S'a'
I65120
sS's'
I18
sS'm'
(lp34577
I1
aI1
aI1
aI11
aI23
aI51
aI127
aI215
aI441
aI467
aI229
aI3071
aI2731
aI8813
aI30155
aI60289
aI54531
aI196187
asS'd'
I17289
sa(dp34578
S'a'
I65132
sS's'
I18
sS'm'
(lp34579
I1
aI3
aI5
aI15
aI29
aI31
aI11
aI129
aI443
aI649
aI773
aI3035
aI7915
aI13831
aI31979
aI5577
aI42869
aI153591
asS'd'
I17290
sa(dp34580
S'a'
I65143
sS's'
I18
sS'm'
(lp34581
I1
aI1
aI3
aI11
aI21
aI37
aI23
aI79
aI153
aI7
aI1801
aI441
aI8189
aI7235
aI6311
aI965
aI71993
aI81755
asS'd'
I17291
sa(dp34582
S'a'
I65150
sS's'
I18
sS'm'
(lp34583
I1
aI3
aI5
aI5
aI23
aI13
aI93
aI39
aI247
aI367
aI811
aI1381
aI6809
aI16219
aI8755
aI41923
aI79873
aI105781
asS'd'
I17292
sa(dp34584
S'a'
I65177
sS's'
I18
sS'm'
(lp34585
I1
aI3
aI5
aI9
aI19
aI43
aI21
aI229
aI251
aI187
aI1047
aI2295
aI5529
aI2965
aI1507
aI16185
aI121183
aI30551
asS'd'
I17293
sa(dp34586
S'a'
I65201
sS's'
I18
sS'm'
(lp34587
I1
aI3
aI7
aI9
aI19
aI11
aI33
aI213
aI39
aI811
aI231
aI1527
aI6093
aI1507
aI3541
aI37585
aI78785
aI215419
asS'd'
I17294
sa(dp34588
S'a'
I65214
sS's'
I18
sS'm'
(lp34589
I1
aI3
aI7
aI7
aI11
aI13
aI109
aI119
aI175
aI311
aI719
aI3127
aI6351
aI1909
aI5441
aI5411
aI58751
aI80875
asS'd'
I17295
sa(dp34590
S'a'
I65234
sS's'
I18
sS'm'
(lp34591
I1
aI1
aI1
aI7
aI17
aI35
aI57
aI139
aI289
aI137
aI1919
aI2131
aI6145
aI3953
aI24887
aI64737
aI4677
aI23833
asS'd'
I17296
sa(dp34592
S'a'
I65240
sS's'
I18
sS'm'
(lp34593
I1
aI1
aI3
aI13
aI1
aI21
aI83
aI243
aI27
aI69
aI501
aI3925
aI3339
aI13313
aI27021
aI38319
aI76441
aI146397
asS'd'
I17297
sa(dp34594
S'a'
I65243
sS's'
I18
sS'm'
(lp34595
I1
aI3
aI7
aI9
aI15
aI17
aI97
aI117
aI505
aI673
aI1333
aI3891
aI7775
aI6323
aI12967
aI17387
aI19501
aI68347
asS'd'
I17298
sa(dp34596
S'a'
I65259
sS's'
I18
sS'm'
(lp34597
I1
aI1
aI1
aI5
aI27
aI55
aI43
aI47
aI399
aI147
aI1539
aI2663
aI5555
aI11993
aI8759
aI33783
aI8361
aI78633
asS'd'
I17299
sa(dp34598
S'a'
I65276
sS's'
I18
sS'm'
(lp34599
I1
aI1
aI1
aI15
aI1
aI17
aI21
aI85
aI129
aI117
aI339
aI1319
aI1119
aI6869
aI12913
aI56873
aI30795
aI76849
asS'd'
I17300
sa(dp34600
S'a'
I65287
sS's'
I18
sS'm'
(lp34601
I1
aI1
aI5
aI5
aI11
aI1
aI11
aI175
aI355
aI737
aI1367
aI3089
aI5993
aI4377
aI10325
aI3817
aI61735
aI187689
asS'd'
I17301
sa(dp34602
S'a'
I65306
sS's'
I18
sS'm'
(lp34603
I1
aI3
aI5
aI5
aI21
aI41
aI85
aI219
aI425
aI611
aI1219
aI1849
aI349
aI925
aI26185
aI31591
aI23855
aI35549
asS'd'
I17302
sa(dp34604
S'a'
I65311
sS's'
I18
sS'm'
(lp34605
I1
aI1
aI3
aI5
aI21
aI3
aI77
aI25
aI265
aI949
aI1979
aI1561
aI4243
aI12437
aI5215
aI23445
aI33295
aI130385
asS'd'
I17303
sa(dp34606
S'a'
I65321
sS's'
I18
sS'm'
(lp34607
I1
aI1
aI3
aI1
aI13
aI7
aI3
aI81
aI143
aI735
aI31
aI1781
aI1537
aI10789
aI11923
aI61589
aI75761
aI178837
asS'd'
I17304
sa(dp34608
S'a'
I65350
sS's'
I18
sS'm'
(lp34609
I1
aI1
aI7
aI7
aI11
aI47
aI55
aI37
aI39
aI533
aI1773
aI3121
aI183
aI7193
aI19403
aI45757
aI20457
aI158437
asS'd'
I17305
sa(dp34610
S'a'
I65364
sS's'
I18
sS'm'
(lp34611
I1
aI1
aI5
aI5
aI3
aI15
aI53
aI41
aI139
aI529
aI601
aI2967
aI4683
aI3869
aI13449
aI30155
aI85833
aI190053
asS'd'
I17306
sa(dp34612
S'a'
I65451
sS's'
I18
sS'm'
(lp34613
I1
aI1
aI5
aI1
aI11
aI39
aI85
aI131
aI349
aI175
aI267
aI779
aI923
aI5905
aI32727
aI22055
aI63087
aI247607
asS'd'
I17307
sa(dp34614
S'a'
I65488
sS's'
I18
sS'm'
(lp34615
I1
aI3
aI5
aI7
aI7
aI59
aI11
aI49
aI465
aI617
aI557
aI251
aI1303
aI10369
aI29207
aI13457
aI113591
aI43717
asS'd'
I17308
sa(dp34616
S'a'
I65510
sS's'
I18
sS'm'
(lp34617
I1
aI3
aI5
aI3
aI27
aI39
aI21
aI157
aI39
aI891
aI1833
aI2887
aI7395
aI7965
aI21771
aI42675
aI71705
aI177323
asS'd'
I17309
sa(dp34618
S'a'
I65528
sS's'
I18
sS'm'
(lp34619
I1
aI3
aI7
aI7
aI21
aI51
aI53
aI83
aI433
aI889
aI1033
aI1701
aI6285
aI14335
aI1683
aI3637
aI110241
aI110355
asS'd'
I17310
sa(dp34620
S'a'
I65533
sS's'
I18
sS'm'
(lp34621
I1
aI1
aI7
aI7
aI11
aI23
aI35
aI63
aI71
aI867
aI79
aI2551
aI1837
aI773
aI21093
aI60433
aI67305
aI70731
asS'd'
I17311
sa(dp34622
S'a'
I65553
sS's'
I18
sS'm'
(lp34623
I1
aI3
aI5
aI11
aI9
aI25
aI67
aI23
aI137
aI75
aI707
aI2229
aI6237
aI9871
aI29063
aI30433
aI112897
aI68037
asS'd'
I17312
sa(dp34624
S'a'
I65563
sS's'
I18
sS'm'
(lp34625
I1
aI3
aI7
aI13
aI1
aI45
aI119
aI149
aI487
aI667
aI1177
aI2927
aI1875
aI11963
aI20771
aI1177
aI2331
aI244039
asS'd'
I17313
sa(dp34626
S'a'
I65579
sS's'
I18
sS'm'
(lp34627
I1
aI3
aI3
aI7
aI19
aI61
aI89
aI163
aI91
aI409
aI1109
aI1947
aI1017
aI12385
aI13487
aI45645
aI64175
aI184221
asS'd'
I17314
sa(dp34628
S'a'
I65621
sS's'
I18
sS'm'
(lp34629
I1
aI1
aI1
aI15
aI13
aI47
aI21
aI203
aI341
aI845
aI443
aI1891
aI2591
aI2721
aI7515
aI52161
aI70359
aI173139
asS'd'
I17315
sa(dp34630
S'a'
I65625
sS's'
I18
sS'm'
(lp34631
I1
aI3
aI3
aI11
aI5
aI3
aI119
aI179
aI509
aI33
aI1909
aI2531
aI6713
aI12447
aI30157
aI61019
aI45857
aI165557
asS'd'
I17316
sa(dp34632
S'a'
I65635
sS's'
I18
sS'm'
(lp34633
I1
aI3
aI7
aI5
aI3
aI47
aI79
aI55
aI321
aI71
aI1917
aI4053
aI6603
aI3079
aI28133
aI15611
aI99161
aI118279
asS'd'
I17317
sa(dp34634
S'a'
I65644
sS's'
I18
sS'm'
(lp34635
I1
aI1
aI1
aI7
aI19
aI13
aI3
aI31
aI213
aI705
aI435
aI2381
aI991
aI4719
aI24473
aI8907
aI122013
aI228081
asS'd'
I17318
sa(dp34636
S'a'
I65690
sS's'
I18
sS'm'
(lp34637
I1
aI1
aI7
aI11
aI27
aI15
aI5
aI123
aI169
aI197
aI361
aI3803
aI2001
aI14547
aI22967
aI27575
aI118325
aI130651
asS'd'
I17319
sa(dp34638
S'a'
I65696
sS's'
I18
sS'm'
(lp34639
I1
aI3
aI3
aI5
aI29
aI43
aI77
aI15
aI463
aI753
aI695
aI3489
aI2023
aI9913
aI13029
aI26621
aI129393
aI209439
asS'd'
I17320
sa(dp34640
S'a'
I65705
sS's'
I18
sS'm'
(lp34641
I1
aI1
aI5
aI15
aI1
aI39
aI55
aI129
aI247
aI729
aI1537
aI2529
aI3981
aI13153
aI1505
aI12743
aI104173
aI218423
asS'd'
I17321
sa(dp34642
S'a'
I65716
sS's'
I18
sS'm'
(lp34643
I1
aI3
aI7
aI3
aI21
aI3
aI49
aI173
aI445
aI821
aI3
aI2671
aI1865
aI1377
aI7589
aI65485
aI96485
aI80193
asS'd'
I17322
sa(dp34644
S'a'
I65725
sS's'
I18
sS'm'
(lp34645
I1
aI3
aI3
aI11
aI7
aI21
aI99
aI143
aI333
aI869
aI1469
aI1579
aI1749
aI2203
aI18773
aI47377
aI103211
aI238357
asS'd'
I17323
sa(dp34646
S'a'
I65733
sS's'
I18
sS'm'
(lp34647
I1
aI3
aI5
aI11
aI11
aI19
aI25
aI253
aI229
aI755
aI101
aI269
aI6703
aI5603
aI23201
aI57163
aI28431
aI159653
asS'd'
I17324
sa(dp34648
S'a'
I65740
sS's'
I18
sS'm'
(lp34649
I1
aI3
aI5
aI3
aI3
aI15
aI45
aI225
aI325
aI997
aI1061
aI883
aI3885
aI7633
aI461
aI44411
aI52129
aI84535
asS'd'
I17325
sa(dp34650
S'a'
I65758
sS's'
I18
sS'm'
(lp34651
I1
aI1
aI7
aI1
aI27
aI29
aI51
aI23
aI473
aI443
aI117
aI3021
aI55
aI7413
aI7911
aI3063
aI47533
aI234941
asS'd'
I17326
sa(dp34652
S'a'
I65786
sS's'
I18
sS'm'
(lp34653
I1
aI3
aI3
aI15
aI19
aI43
aI37
aI95
aI249
aI805
aI603
aI865
aI2115
aI6999
aI9739
aI59029
aI12181
aI211159
asS'd'
I17327
sa(dp34654
S'a'
I65806
sS's'
I18
sS'm'
(lp34655
I1
aI3
aI7
aI3
aI3
aI61
aI105
aI113
aI11
aI169
aI1007
aI689
aI2553
aI14561
aI17473
aI38249
aI41225
aI80021
asS'd'
I17328
sa(dp34656
S'a'
I65834
sS's'
I18
sS'm'
(lp34657
I1
aI3
aI7
aI11
aI5
aI47
aI69
aI49
aI457
aI931
aI435
aI1423
aI411
aI15163
aI3171
aI29143
aI101153
aI240869
asS'd'
I17329
sa(dp34658
S'a'
I65844
sS's'
I18
sS'm'
(lp34659
I1
aI3
aI7
aI1
aI17
aI1
aI13
aI45
aI155
aI551
aI1783
aI3583
aI2767
aI2761
aI18019
aI61635
aI104527
aI123817
asS'd'
I17330
sa(dp34660
S'a'
I65866
sS's'
I18
sS'm'
(lp34661
I1
aI1
aI5
aI11
aI9
aI43
aI101
aI205
aI233
aI689
aI1247
aI2903
aI3117
aI12261
aI11827
aI50403
aI103727
aI35533
asS'd'
I17331
sa(dp34662
S'a'
I65895
sS's'
I18
sS'm'
(lp34663
I1
aI1
aI5
aI13
aI23
aI37
aI121
aI195
aI133
aI265
aI1517
aI823
aI5933
aI13917
aI6363
aI8533
aI58443
aI178549
asS'd'
I17332
sa(dp34664
S'a'
I65902
sS's'
I18
sS'm'
(lp34665
I1
aI1
aI7
aI9
aI29
aI1
aI3
aI195
aI221
aI877
aI71
aI473
aI1173
aI15285
aI6057
aI60005
aI92401
aI65357
asS'd'
I17333
sa(dp34666
S'a'
I65953
sS's'
I18
sS'm'
(lp34667
I1
aI3
aI7
aI1
aI5
aI25
aI15
aI207
aI455
aI447
aI1125
aI3731
aI1289
aI867
aI22111
aI38893
aI70779
aI88277
asS'd'
I17334
sa(dp34668
S'a'
I65954
sS's'
I18
sS'm'
(lp34669
I1
aI1
aI1
aI13
aI31
aI19
aI15
aI179
aI183
aI351
aI1197
aI1929
aI3569
aI12251
aI17641
aI4097
aI24141
aI186857
asS'd'
I17335
sa(dp34670
S'a'
I65983
sS's'
I18
sS'm'
(lp34671
I1
aI1
aI1
aI9
aI19
aI9
aI125
aI23
aI431
aI225
aI943
aI479
aI2615
aI443
aI30977
aI10889
aI17107
aI116819
asS'd'
I17336
sa(dp34672
S'a'
I65985
sS's'
I18
sS'm'
(lp34673
I1
aI3
aI1
aI9
aI9
aI13
aI85
aI123
aI85
aI857
aI125
aI3149
aI1105
aI3687
aI2313
aI38749
aI52131
aI259511
asS'd'
I17337
sa(dp34674
S'a'
I66003
sS's'
I18
sS'm'
(lp34675
I1
aI3
aI5
aI7
aI27
aI33
aI57
aI105
aI511
aI871
aI1089
aI2311
aI3291
aI2245
aI3365
aI30211
aI62549
aI56207
asS'd'
I17338
sa(dp34676
S'a'
I66010
sS's'
I18
sS'm'
(lp34677
I1
aI1
aI7
aI11
aI1
aI19
aI75
aI37
aI139
aI173
aI391
aI317
aI2575
aI11887
aI4289
aI32275
aI43487
aI487
asS'd'
I17339
sa(dp34678
S'a'
I66025
sS's'
I18
sS'm'
(lp34679
I1
aI1
aI7
aI1
aI11
aI3
aI9
aI217
aI343
aI35
aI59
aI93
aI1343
aI5043
aI14869
aI63717
aI40983
aI235373
asS'd'
I17340
sa(dp34680
S'a'
I66050
sS's'
I18
sS'm'
(lp34681
I1
aI1
aI3
aI5
aI15
aI13
aI93
aI247
aI417
aI179
aI307
aI3299
aI4383
aI5491
aI21271
aI37155
aI32289
aI75737
asS'd'
I17341
sa(dp34682
S'a'
I66074
sS's'
I18
sS'm'
(lp34683
I1
aI3
aI3
aI9
aI3
aI39
aI63
aI243
aI305
aI729
aI9
aI3317
aI3301
aI13165
aI20437
aI36505
aI32977
aI2761
asS'd'
I17342
sa(dp34684
S'a'
I66076
sS's'
I18
sS'm'
(lp34685
I1
aI3
aI7
aI5
aI3
aI37
aI61
aI109
aI351
aI641
aI1699
aI2517
aI2637
aI4995
aI27365
aI56971
aI53609
aI14373
asS'd'
I17343
sa(dp34686
S'a'
I66097
sS's'
I18
sS'm'
(lp34687
I1
aI3
aI1
aI15
aI31
aI53
aI127
aI123
aI219
aI1003
aI1425
aI1201
aI5303
aI10369
aI21481
aI26987
aI42541
aI37855
asS'd'
I17344
sa(dp34688
S'a'
I66132
sS's'
I18
sS'm'
(lp34689
I1
aI3
aI3
aI13
aI9
aI29
aI35
aI111
aI395
aI791
aI1619
aI2647
aI713
aI15955
aI19145
aI33883
aI65215
aI166267
asS'd'
I17345
sa(dp34690
S'a'
I66158
sS's'
I18
sS'm'
(lp34691
I1
aI1
aI7
aI11
aI17
aI5
aI45
aI249
aI421
aI273
aI411
aI2885
aI7027
aI11933
aI24847
aI36969
aI124701
aI214931
asS'd'
I17346
sa(dp34692
S'a'
I66175
sS's'
I18
sS'm'
(lp34693
I1
aI1
aI3
aI1
aI27
aI41
aI125
aI83
aI327
aI643
aI223
aI151
aI6709
aI15949
aI125
aI13275
aI90405
aI15759
asS'd'
I17347
sa(dp34694
S'a'
I66185
sS's'
I18
sS'm'
(lp34695
I1
aI1
aI5
aI15
aI19
aI45
aI55
aI109
aI497
aI1011
aI1363
aI1937
aI3697
aI7475
aI10533
aI65325
aI29681
aI76275
asS'd'
I17348
sa(dp34696
S'a'
I66199
sS's'
I18
sS'm'
(lp34697
I1
aI1
aI1
aI3
aI23
aI17
aI59
aI209
aI229
aI151
aI1199
aI279
aI191
aI8993
aI25939
aI13885
aI113477
aI166961
asS'd'
I17349
sa(dp34698
S'a'
I66206
sS's'
I18
sS'm'
(lp34699
I1
aI3
aI5
aI1
aI19
aI61
aI27
aI129
aI103
aI721
aI1451
aI2803
aI5879
aI3523
aI15443
aI4047
aI95927
aI50339
asS'd'
I17350
sa(dp34700
S'a'
I66222
sS's'
I18
sS'm'
(lp34701
I1
aI1
aI1
aI11
aI27
aI9
aI53
aI47
aI331
aI185
aI1337
aI3429
aI807
aI3341
aI14871
aI11035
aI50651
aI243843
asS'd'
I17351
sa(dp34702
S'a'
I66227
sS's'
I18
sS'm'
(lp34703
I1
aI3
aI5
aI11
aI7
aI57
aI125
aI15
aI271
aI811
aI1873
aI3093
aI7841
aI5761
aI19955
aI571
aI123319
aI149465
asS'd'
I17352
sa(dp34704
S'a'
I66241
sS's'
I18
sS'm'
(lp34705
I1
aI3
aI1
aI3
aI31
aI61
aI71
aI47
aI477
aI273
aI167
aI1069
aI3513
aI1463
aI2667
aI22097
aI60367
aI246045
asS'd'
I17353
sa(dp34706
S'a'
I66272
sS's'
I18
sS'm'
(lp34707
I1
aI1
aI1
aI11
aI1
aI55
aI35
aI233
aI37
aI659
aI1517
aI411
aI2981
aI10339
aI21857
aI33701
aI44393
aI6861
asS'd'
I17354
sa(dp34708
S'a'
I66324
sS's'
I18
sS'm'
(lp34709
I1
aI1
aI5
aI7
aI11
aI43
aI109
aI205
aI103
aI315
aI1925
aI2109
aI6307
aI7915
aI19793
aI61167
aI27963
aI251913
asS'd'
I17355
sa(dp34710
S'a'
I66424
sS's'
I18
sS'm'
(lp34711
I1
aI1
aI5
aI11
aI5
aI63
aI107
aI219
aI53
aI251
aI1053
aI2035
aI77
aI15885
aI22011
aI3945
aI91
aI204899
asS'd'
I17356
sa(dp34712
S'a'
I66460
sS's'
I18
sS'm'
(lp34713
I1
aI3
aI7
aI13
aI21
aI45
aI51
aI53
aI99
aI831
aI1421
aI3171
aI4241
aI14105
aI26161
aI45071
aI2813
aI54339
asS'd'
I17357
sa(dp34714
S'a'
I66464
sS's'
I18
sS'm'
(lp34715
I1
aI3
aI5
aI1
aI3
aI61
aI43
aI141
aI355
aI699
aI11
aI2203
aI8055
aI14815
aI24597
aI65201
aI32689
aI70167
asS'd'
I17358
sa(dp34716
S'a'
I66481
sS's'
I18
sS'm'
(lp34717
I1
aI3
aI1
aI5
aI11
aI27
aI109
aI239
aI199
aI23
aI375
aI1477
aI3197
aI4401
aI29901
aI46623
aI79593
aI133143
asS'd'
I17359
sa(dp34718
S'a'
I66482
sS's'
I18
sS'm'
(lp34719
I1
aI1
aI1
aI3
aI23
aI9
aI63
aI103
aI41
aI177
aI1365
aI1971
aI5937
aI13055
aI27713
aI13535
aI47371
aI57841
asS'd'
I17360
sa(dp34720
S'a'
I66523
sS's'
I18
sS'm'
(lp34721
I1
aI1
aI5
aI5
aI19
aI15
aI5
aI21
aI307
aI65
aI215
aI3801
aI4149
aI6565
aI10249
aI63541
aI30867
aI12129
asS'd'
I17361
sa(dp34722
S'a'
I66526
sS's'
I18
sS'm'
(lp34723
I1
aI3
aI3
aI9
aI11
aI1
aI107
aI99
aI235
aI331
aI1479
aI1365
aI2557
aI9545
aI25767
aI12461
aI6471
aI184643
asS'd'
I17362
sa(dp34724
S'a'
I66532
sS's'
I18
sS'm'
(lp34725
I1
aI3
aI3
aI5
aI17
aI13
aI103
aI223
aI95
aI955
aI1479
aI1825
aI705
aI5311
aI28531
aI22787
aI118899
aI181829
asS'd'
I17363
sa(dp34726
S'a'
I66539
sS's'
I18
sS'm'
(lp34727
I1
aI3
aI7
aI1
aI5
aI59
aI65
aI11
aI251
aI419
aI659
aI2559
aI5445
aI4221
aI5871
aI51845
aI33925
aI167037
asS'd'
I17364
sa(dp34728
S'a'
I66541
sS's'
I18
sS'm'
(lp34729
I1
aI3
aI5
aI13
aI15
aI45
aI35
aI181
aI325
aI293
aI1897
aI3321
aI6081
aI9919
aI27641
aI9407
aI35263
aI231009
asS'd'
I17365
sa(dp34730
S'a'
I66573
sS's'
I18
sS'm'
(lp34731
I1
aI3
aI3
aI3
aI3
aI35
aI85
aI33
aI293
aI777
aI1945
aI3771
aI6967
aI12353
aI2737
aI12501
aI127359
aI163591
asS'd'
I17366
sa(dp34732
S'a'
I66582
sS's'
I18
sS'm'
(lp34733
I1
aI3
aI7
aI9
aI13
aI7
aI119
aI107
aI309
aI811
aI1113
aI2465
aI4867
aI4295
aI565
aI59159
aI94587
aI119761
asS'd'
I17367
sa(dp34734
S'a'
I66591
sS's'
I18
sS'm'
(lp34735
I1
aI3
aI3
aI1
aI1
aI31
aI61
aI49
aI461
aI635
aI233
aI175
aI6237
aI10463
aI17847
aI54925
aI115675
aI260575
asS'd'
I17368
sa(dp34736
S'a'
I66615
sS's'
I18
sS'm'
(lp34737
I1
aI3
aI5
aI15
aI25
aI17
aI61
aI155
aI235
aI483
aI1771
aI2903
aI3163
aI2525
aI17153
aI54701
aI49521
aI11911
asS'd'
I17369
sa(dp34738
S'a'
I66622
sS's'
I18
sS'm'
(lp34739
I1
aI3
aI3
aI5
aI1
aI35
aI51
aI23
aI187
aI107
aI177
aI1381
aI165
aI6149
aI10841
aI3619
aI107811
aI188811
asS'd'
I17370
sa(dp34740
S'a'
I66629
sS's'
I18
sS'm'
(lp34741
I1
aI1
aI5
aI9
aI7
aI35
aI5
aI233
aI43
aI913
aI939
aI2195
aI1369
aI5355
aI7941
aI26075
aI66813
aI227623
asS'd'
I17371
sa(dp34742
S'a'
I66641
sS's'
I18
sS'm'
(lp34743
I1
aI3
aI7
aI11
aI5
aI43
aI97
aI211
aI427
aI875
aI1179
aI3631
aI7989
aI2419
aI17209
aI15789
aI128209
aI224117
asS'd'
I17372
sa(dp34744
S'a'
I66712
sS's'
I18
sS'm'
(lp34745
I1
aI1
aI1
aI7
aI17
aI7
aI109
aI255
aI111
aI883
aI371
aI3481
aI6031
aI14665
aI5905
aI28735
aI113003
aI327
asS'd'
I17373
sa(dp34746
S'a'
I66724
sS's'
I18
sS'm'
(lp34747
I1
aI3
aI5
aI3
aI5
aI61
aI7
aI155
aI87
aI861
aI39
aI3163
aI179
aI15493
aI16403
aI18755
aI116157
aI233185
asS'd'
I17374
sa(dp34748
S'a'
I66731
sS's'
I18
sS'm'
(lp34749
I1
aI3
aI5
aI5
aI23
aI45
aI67
aI205
aI395
aI417
aI1235
aI669
aI5097
aI6823
aI31483
aI61395
aI36073
aI24183
asS'd'
I17375
sa(dp34750
S'a'
I66745
sS's'
I18
sS'm'
(lp34751
I1
aI1
aI3
aI1
aI11
aI35
aI123
aI171
aI125
aI759
aI197
aI907
aI2273
aI3623
aI31861
aI60071
aI91857
aI158011
asS'd'
I17376
sa(dp34752
S'a'
I66759
sS's'
I18
sS'm'
(lp34753
I1
aI3
aI7
aI11
aI19
aI19
aI25
aI25
aI167
aI429
aI1565
aI3179
aI5453
aI15731
aI30727
aI32111
aI63685
aI113309
asS'd'
I17377
sa(dp34754
S'a'
I66808
sS's'
I18
sS'm'
(lp34755
I1
aI3
aI7
aI9
aI15
aI33
aI67
aI225
aI495
aI19
aI1881
aI1357
aI4311
aI9547
aI18717
aI20749
aI8819
aI209979
asS'd'
I17378
sa(dp34756
S'a'
I66814
sS's'
I18
sS'm'
(lp34757
I1
aI3
aI5
aI3
aI13
aI47
aI107
aI153
aI461
aI815
aI1521
aI2361
aI7721
aI10631
aI2799
aI62321
aI59755
aI170803
asS'd'
I17379
sa(dp34758
S'a'
I66821
sS's'
I18
sS'm'
(lp34759
I1
aI3
aI7
aI5
aI25
aI61
aI5
aI235
aI71
aI349
aI1555
aI3419
aI1159
aI2027
aI17391
aI29849
aI47145
aI122057
asS'd'
I17380
sa(dp34760
S'a'
I66831
sS's'
I18
sS'm'
(lp34761
I1
aI1
aI5
aI11
aI19
aI19
aI101
aI45
aI333
aI553
aI1431
aI4077
aI2629
aI15997
aI19793
aI65521
aI124287
aI174675
asS'd'
I17381
sa(dp34762
S'a'
I66834
sS's'
I18
sS'm'
(lp34763
I1
aI3
aI7
aI11
aI25
aI39
aI103
aI219
aI375
aI27
aI227
aI1061
aI445
aI14803
aI18883
aI49191
aI33303
aI114467
asS'd'
I17382
sa(dp34764
S'a'
I66879
sS's'
I18
sS'm'
(lp34765
I1
aI1
aI5
aI7
aI3
aI13
aI117
aI29
aI387
aI891
aI371
aI2199
aI7023
aI13671
aI26291
aI61563
aI2733
aI16093
asS'd'
I17383
sa(dp34766
S'a'
I66884
sS's'
I18
sS'm'
(lp34767
I1
aI3
aI5
aI11
aI29
aI5
aI17
aI249
aI149
aI777
aI1817
aI319
aI19
aI12321
aI15241
aI29069
aI58381
aI157467
asS'd'
I17384
sa(dp34768
S'a'
I66899
sS's'
I18
sS'm'
(lp34769
I1
aI3
aI7
aI9
aI29
aI17
aI81
aI141
aI201
aI383
aI429
aI3675
aI69
aI8155
aI22821
aI60707
aI127015
aI248279
asS'd'
I17385
sa(dp34770
S'a'
I66901
sS's'
I18
sS'm'
(lp34771
I1
aI3
aI7
aI9
aI25
aI5
aI11
aI27
aI423
aI987
aI99
aI3599
aI4849
aI4513
aI32119
aI34301
aI6327
aI249457
asS'd'
I17386
sa(dp34772
S'a'
I66924
sS's'
I18
sS'm'
(lp34773
I1
aI3
aI3
aI7
aI13
aI25
aI71
aI227
aI307
aI985
aI665
aI3097
aI6713
aI3823
aI6357
aI58199
aI84057
aI28055
asS'd'
I17387
sa(dp34774
S'a'
I66929
sS's'
I18
sS'm'
(lp34775
I1
aI3
aI5
aI7
aI19
aI21
aI93
aI45
aI159
aI527
aI493
aI59
aI1111
aI1415
aI1949
aI28525
aI50343
aI11039
asS'd'
I17388
sa(dp34776
S'a'
I66942
sS's'
I18
sS'm'
(lp34777
I1
aI3
aI5
aI13
aI17
aI35
aI79
aI229
aI449
aI533
aI235
aI3445
aI8153
aI15473
aI12975
aI53909
aI24589
aI237049
asS'd'
I17389
sa(dp34778
S'a'
I66985
sS's'
I18
sS'm'
(lp34779
I1
aI1
aI7
aI5
aI3
aI53
aI93
aI33
aI339
aI423
aI497
aI2691
aI6125
aI3931
aI25357
aI27509
aI92509
aI227209
asS'd'
I17390
sa(dp34780
S'a'
I66986
sS's'
I18
sS'm'
(lp34781
I1
aI1
aI7
aI3
aI13
aI49
aI111
aI179
aI449
aI279
aI827
aI1481
aI2477
aI6867
aI18079
aI6261
aI30885
aI205675
asS'd'
I17391
sa(dp34782
S'a'
I66994
sS's'
I18
sS'm'
(lp34783
I1
aI3
aI7
aI7
aI11
aI9
aI13
aI105
aI367
aI639
aI1307
aI1617
aI4759
aI8387
aI8909
aI13715
aI56599
aI113259
asS'd'
I17392
sa(dp34784
S'a'
I67066
sS's'
I18
sS'm'
(lp34785
I1
aI1
aI3
aI3
aI15
aI17
aI103
aI125
aI205
aI67
aI999
aI3965
aI907
aI13235
aI15275
aI58457
aI66889
aI227279
asS'd'
I17393
sa(dp34786
S'a'
I67075
sS's'
I18
sS'm'
(lp34787
I1
aI3
aI3
aI3
aI11
aI35
aI99
aI81
aI421
aI75
aI1757
aI2413
aI5655
aI1227
aI4019
aI14503
aI20719
aI224807
asS'd'
I17394
sa(dp34788
S'a'
I67099
sS's'
I18
sS'm'
(lp34789
I1
aI1
aI7
aI11
aI17
aI17
aI109
aI203
aI331
aI813
aI987
aI2925
aI1601
aI13617
aI29
aI8235
aI95129
aI117987
asS'd'
I17395
sa(dp34790
S'a'
I67130
sS's'
I18
sS'm'
(lp34791
I1
aI3
aI7
aI5
aI5
aI33
aI105
aI191
aI183
aI899
aI1949
aI2923
aI2473
aI3435
aI8097
aI35615
aI10109
aI62563
asS'd'
I17396
sa(dp34792
S'a'
I67173
sS's'
I18
sS'm'
(lp34793
I1
aI1
aI7
aI13
aI5
aI25
aI21
aI159
aI487
aI415
aI1507
aI2161
aI649
aI14425
aI2605
aI8357
aI92441
aI87323
asS'd'
I17397
sa(dp34794
S'a'
I67183
sS's'
I18
sS'm'
(lp34795
I1
aI3
aI7
aI15
aI3
aI23
aI87
aI209
aI407
aI765
aI975
aI3859
aI675
aI6351
aI18703
aI44919
aI57155
aI134961
asS'd'
I17398
sa(dp34796
S'a'
I67186
sS's'
I18
sS'm'
(lp34797
I1
aI3
aI5
aI15
aI7
aI59
aI77
aI37
aI235
aI565
aI1707
aI3531
aI6733
aI2223
aI12621
aI59523
aI83547
aI172355
asS'd'
I17399
sa(dp34798
S'a'
I67197
sS's'
I18
sS'm'
(lp34799
I1
aI3
aI7
aI5
aI23
aI15
aI57
aI217
aI151
aI333
aI1033
aI2549
aI303
aI1455
aI5329
aI20187
aI55415
aI166093
asS'd'
I17400
sa(dp34800
S'a'
I67201
sS's'
I18
sS'm'
(lp34801
I1
aI3
aI1
aI13
aI21
aI27
aI1
aI85
aI335
aI201
aI135
aI2603
aI291
aI10573
aI28411
aI1059
aI129871
aI98303
asS'd'
I17401
sa(dp34802
S'a'
I67208
sS's'
I18
sS'm'
(lp34803
I1
aI1
aI3
aI5
aI23
aI61
aI123
aI169
aI503
aI629
aI711
aI2795
aI2291
aI13273
aI32703
aI63377
aI72809
aI214927
asS'd'
I17402
sa(dp34804
S'a'
I67221
sS's'
I18
sS'm'
(lp34805
I1
aI3
aI5
aI13
aI31
aI11
aI115
aI133
aI443
aI709
aI263
aI3739
aI2777
aI11545
aI19137
aI61285
aI64065
aI214477
asS'd'
I17403
sa(dp34806
S'a'
I67238
sS's'
I18
sS'm'
(lp34807
I1
aI3
aI1
aI11
aI29
aI25
aI3
aI5
aI385
aI613
aI1277
aI1445
aI1643
aI15137
aI28041
aI47713
aI122051
aI62915
asS'd'
I17404
sa(dp34808
S'a'
I67252
sS's'
I18
sS'm'
(lp34809
I1
aI3
aI5
aI11
aI3
aI11
aI1
aI137
aI457
aI577
aI783
aI1745
aI5
aI5817
aI26569
aI50751
aI14075
aI246219
asS'd'
I17405
sa(dp34810
S'a'
I67256
sS's'
I18
sS'm'
(lp34811
I1
aI1
aI1
aI9
aI7
aI9
aI105
aI27
aI167
aI939
aI799
aI2773
aI6427
aI15579
aI1975
aI10695
aI111429
aI227105
asS'd'
I17406
sa(dp34812
S'a'
I67270
sS's'
I18
sS'm'
(lp34813
I1
aI3
aI1
aI11
aI23
aI23
aI73
aI103
aI103
aI61
aI1743
aI3061
aI8127
aI15893
aI21223
aI43549
aI103659
aI89129
asS'd'
I17407
sa(dp34814
S'a'
I67288
sS's'
I18
sS'm'
(lp34815
I1
aI1
aI7
aI13
aI5
aI43
aI59
aI235
aI139
aI961
aI839
aI3843
aI1317
aI4903
aI21043
aI15479
aI115065
aI112531
asS'd'
I17408
sa(dp34816
S'a'
I67318
sS's'
I18
sS'm'
(lp34817
I1
aI3
aI1
aI3
aI19
aI15
aI57
aI145
aI193
aI321
aI1919
aI385
aI125
aI15517
aI14243
aI62845
aI38995
aI120045
asS'd'
I17409
sa(dp34818
S'a'
I67329
sS's'
I18
sS'm'
(lp34819
I1
aI3
aI3
aI7
aI17
aI61
aI77
aI75
aI267
aI203
aI1911
aI2599
aI1797
aI761
aI28101
aI58603
aI107755
aI158689
asS'd'
I17410
sa(dp34820
S'a'
I67349
sS's'
I18
sS'm'
(lp34821
I1
aI1
aI5
aI11
aI15
aI21
aI71
aI227
aI377
aI361
aI2013
aI129
aI6271
aI1421
aI6009
aI52261
aI113389
aI74915
asS'd'
I17411
sa(dp34822
S'a'
I67369
sS's'
I18
sS'm'
(lp34823
I1
aI1
aI5
aI7
aI3
aI39
aI27
aI49
aI97
aI885
aI651
aI1633
aI3445
aI3415
aI20167
aI26667
aI52997
aI221391
asS'd'
I17412
sa(dp34824
S'a'
I67392
sS's'
I18
sS'm'
(lp34825
I1
aI1
aI7
aI9
aI7
aI59
aI95
aI127
aI479
aI871
aI845
aI2951
aI673
aI6385
aI10057
aI2605
aI78529
aI230771
asS'd'
I17413
sa(dp34826
S'a'
I67435
sS's'
I18
sS'm'
(lp34827
I1
aI3
aI5
aI7
aI31
aI15
aI53
aI125
aI223
aI711
aI875
aI429
aI7237
aI4005
aI2153
aI26865
aI63205
aI144125
asS'd'
I17414
sa(dp34828
S'a'
I67473
sS's'
I18
sS'm'
(lp34829
I1
aI1
aI7
aI9
aI1
aI57
aI19
aI189
aI67
aI423
aI1937
aI37
aI4925
aI15503
aI25969
aI20419
aI59921
aI58119
asS'd'
I17415
sa(dp34830
S'a'
I67496
sS's'
I18
sS'm'
(lp34831
I1
aI1
aI7
aI7
aI31
aI51
aI99
aI189
aI21
aI1011
aI1551
aI3529
aI7617
aI15805
aI11365
aI43123
aI84785
aI203703
asS'd'
I17416
sa(dp34832
S'a'
I67507
sS's'
I18
sS'm'
(lp34833
I1
aI1
aI3
aI15
aI3
aI63
aI9
aI67
aI399
aI151
aI253
aI1839
aI1365
aI16295
aI13145
aI29211
aI48681
aI177643
asS'd'
I17417
sa(dp34834
S'a'
I67514
sS's'
I18
sS'm'
(lp34835
I1
aI1
aI1
aI5
aI13
aI37
aI1
aI21
aI435
aI483
aI939
aI535
aI1505
aI10879
aI7027
aI5599
aI63261
aI158573
asS'd'
I17418
sa(dp34836
S'a'
I67536
sS's'
I18
sS'm'
(lp34837
I1
aI1
aI7
aI15
aI13
aI39
aI113
aI209
aI213
aI1017
aI1197
aI285
aI4221
aI6831
aI13383
aI2265
aI34313
aI160879
asS'd'
I17419
sa(dp34838
S'a'
I67555
sS's'
I18
sS'm'
(lp34839
I1
aI3
aI7
aI15
aI25
aI23
aI95
aI217
aI141
aI681
aI451
aI1275
aI4957
aI10197
aI21375
aI50905
aI11087
aI96135
asS'd'
I17420
sa(dp34840
S'a'
I67598
sS's'
I18
sS'm'
(lp34841
I1
aI3
aI1
aI11
aI17
aI35
aI87
aI15
aI57
aI777
aI1429
aI615
aI681
aI8437
aI23981
aI51781
aI112169
aI198471
asS'd'
I17421
sa(dp34842
S'a'
I67615
sS's'
I18
sS'm'
(lp34843
I1
aI3
aI1
aI5
aI21
aI37
aI113
aI105
aI123
aI327
aI549
aI1641
aI7697
aI2127
aI5709
aI8351
aI56787
aI260157
asS'd'
I17422
sa(dp34844
S'a'
I67625
sS's'
I18
sS'm'
(lp34845
I1
aI3
aI7
aI15
aI19
aI17
aI51
aI15
aI367
aI89
aI1635
aI353
aI4855
aI1551
aI7197
aI27403
aI11259
aI176029
asS'd'
I17423
sa(dp34846
S'a'
I67645
sS's'
I18
sS'm'
(lp34847
I1
aI3
aI1
aI1
aI27
aI9
aI3
aI35
aI213
aI31
aI885
aI797
aI7077
aI15641
aI22509
aI35193
aI112411
aI157335
asS'd'
I17424
sa(dp34848
S'a'
I67651
sS's'
I18
sS'm'
(lp34849
I1
aI1
aI5
aI1
aI5
aI9
aI101
aI149
aI169
aI581
aI1927
aI197
aI5935
aI6361
aI3915
aI15541
aI69575
aI102451
asS'd'
I17425
sa(dp34850
S'a'
I67663
sS's'
I18
sS'm'
(lp34851
I1
aI3
aI3
aI13
aI23
aI63
aI45
aI205
aI271
aI17
aI707
aI937
aI2547
aI12019
aI8559
aI26163
aI58117
aI138625
asS'd'
I17426
sa(dp34852
S'a'
I67672
sS's'
I18
sS'm'
(lp34853
I1
aI1
aI3
aI1
aI19
aI63
aI125
aI175
aI253
aI629
aI1121
aI3701
aI7755
aI61
aI13037
aI39417
aI6179
aI261923
asS'd'
I17427
sa(dp34854
S'a'
I67682
sS's'
I18
sS'm'
(lp34855
I1
aI1
aI5
aI1
aI25
aI63
aI27
aI245
aI371
aI657
aI157
aI3821
aI3279
aI8977
aI9065
aI35611
aI27325
aI205737
asS'd'
I17428
sa(dp34856
S'a'
I67699
sS's'
I18
sS'm'
(lp34857
I1
aI3
aI7
aI15
aI7
aI57
aI19
aI191
aI1
aI927
aI1379
aI2579
aI4335
aI7163
aI4877
aI51435
aI17309
aI100173
asS'd'
I17429
sa(dp34858
S'a'
I67718
sS's'
I18
sS'm'
(lp34859
I1
aI1
aI1
aI5
aI21
aI59
aI107
aI195
aI317
aI691
aI541
aI69
aI7235
aI2175
aI25191
aI23913
aI126369
aI9263
asS'd'
I17430
sa(dp34860
S'a'
I67780
sS's'
I18
sS'm'
(lp34861
I1
aI3
aI5
aI15
aI17
aI7
aI67
aI27
aI263
aI855
aI1065
aI973
aI6705
aI10729
aI8719
aI32741
aI59207
aI249107
asS'd'
I17431
sa(dp34862
S'a'
I67795
sS's'
I18
sS'm'
(lp34863
I1
aI3
aI1
aI1
aI21
aI23
aI115
aI119
aI351
aI207
aI1691
aI1105
aI7479
aI3877
aI24439
aI29017
aI34171
aI133797
asS'd'
I17432
sa(dp34864
S'a'
I67804
sS's'
I18
sS'm'
(lp34865
I1
aI3
aI3
aI3
aI11
aI23
aI39
aI165
aI99
aI1023
aI309
aI3933
aI4235
aI3891
aI27237
aI30887
aI34363
aI175017
asS'd'
I17433
sa(dp34866
S'a'
I67818
sS's'
I18
sS'm'
(lp34867
I1
aI3
aI1
aI5
aI17
aI33
aI79
aI105
aI253
aI515
aI823
aI1783
aI1523
aI2095
aI10355
aI8929
aI51001
aI112815
asS'd'
I17434
sa(dp34868
S'a'
I67832
sS's'
I18
sS'm'
(lp34869
I1
aI3
aI7
aI15
aI11
aI27
aI123
aI161
aI279
aI541
aI1343
aI1009
aI6015
aI8565
aI27031
aI233
aI2153
aI179243
asS'd'
I17435
sa(dp34870
S'a'
I67941
sS's'
I18
sS'm'
(lp34871
I1
aI1
aI7
aI7
aI1
aI53
aI1
aI33
aI75
aI57
aI723
aI3855
aI3301
aI14941
aI6637
aI25181
aI103441
aI208339
asS'd'
I17436
sa(dp34872
S'a'
I67984
sS's'
I18
sS'm'
(lp34873
I1
aI3
aI7
aI13
aI3
aI23
aI11
aI75
aI55
aI19
aI1181
aI3451
aI4901
aI2621
aI18323
aI42395
aI95701
aI237753
asS'd'
I17437
sa(dp34874
S'a'
I68044
sS's'
I18
sS'm'
(lp34875
I1
aI1
aI1
aI5
aI1
aI1
aI123
aI203
aI367
aI71
aI1897
aI295
aI6719
aI2647
aI7135
aI34511
aI79853
aI58351
asS'd'
I17438
sa(dp34876
S'a'
I68059
sS's'
I18
sS'm'
(lp34877
I1
aI3
aI3
aI7
aI7
aI63
aI87
aI91
aI223
aI265
aI927
aI3843
aI1703
aI11633
aI8077
aI26425
aI46573
aI181201
asS'd'
I17439
sa(dp34878
S'a'
I68062
sS's'
I18
sS'm'
(lp34879
I1
aI3
aI3
aI3
aI1
aI51
aI3
aI99
aI37
aI251
aI1739
aI2813
aI3955
aI8745
aI4425
aI42419
aI124177
aI173997
asS'd'
I17440
sa(dp34880
S'a'
I68065
sS's'
I18
sS'm'
(lp34881
I1
aI3
aI5
aI15
aI9
aI19
aI11
aI207
aI123
aI287
aI1703
aI2155
aI2729
aI4541
aI4925
aI4735
aI77159
aI97911
asS'd'
I17441
sa(dp34882
S'a'
I68071
sS's'
I18
sS'm'
(lp34883
I1
aI3
aI3
aI15
aI23
aI31
aI35
aI33
aI507
aI315
aI1071
aI3001
aI7569
aI11749
aI3183
aI6989
aI68637
aI177803
asS'd'
I17442
sa(dp34884
S'a'
I68072
sS's'
I18
sS'm'
(lp34885
I1
aI3
aI5
aI1
aI31
aI1
aI113
aI39
aI295
aI263
aI1113
aI619
aI5523
aI15385
aI24115
aI24233
aI91943
aI129299
asS'd'
I17443
sa(dp34886
S'a'
I68085
sS's'
I18
sS'm'
(lp34887
I1
aI1
aI5
aI11
aI9
aI57
aI89
aI49
aI67
aI601
aI1277
aI2275
aI6349
aI4141
aI28397
aI47061
aI28143
aI126291
asS'd'
I17444
sa(dp34888
S'a'
I68090
sS's'
I18
sS'm'
(lp34889
I1
aI3
aI7
aI15
aI11
aI63
aI123
aI187
aI305
aI1009
aI1509
aI2569
aI2235
aI8233
aI27351
aI53437
aI34353
aI105799
asS'd'
I17445
sa(dp34890
S'a'
I68101
sS's'
I18
sS'm'
(lp34891
I1
aI1
aI3
aI1
aI13
aI9
aI11
aI169
aI427
aI171
aI1031
aI633
aI4275
aI1173
aI11233
aI57997
aI107753
aI257337
asS'd'
I17446
sa(dp34892
S'a'
I68102
sS's'
I18
sS'm'
(lp34893
I1
aI3
aI5
aI5
aI3
aI39
aI49
aI233
aI309
aI999
aI1275
aI85
aI1663
aI16275
aI9145
aI18439
aI59055
aI249657
asS'd'
I17447
sa(dp34894
S'a'
I68106
sS's'
I18
sS'm'
(lp34895
I1
aI1
aI3
aI7
aI11
aI55
aI73
aI75
aI115
aI397
aI945
aI3657
aI6847
aI7341
aI21305
aI30119
aI65675
aI169281
asS'd'
I17448
sa(dp34896
S'a'
I68164
sS's'
I18
sS'm'
(lp34897
I1
aI1
aI7
aI7
aI3
aI5
aI31
aI179
aI183
aI479
aI329
aI217
aI1077
aI6893
aI23425
aI21903
aI34293
aI184819
asS'd'
I17449
sa(dp34898
S'a'
I68191
sS's'
I18
sS'm'
(lp34899
I1
aI1
aI3
aI3
aI7
aI63
aI97
aI255
aI289
aI917
aI1881
aI3829
aI2433
aI3473
aI11045
aI37739
aI73349
aI171899
asS'd'
I17450
sa(dp34900
S'a'
I68192
sS's'
I18
sS'm'
(lp34901
I1
aI3
aI5
aI5
aI27
aI23
aI61
aI151
aI353
aI667
aI1889
aI2323
aI3261
aI15999
aI24225
aI35265
aI97301
aI75743
asS'd'
I17451
sa(dp34902
S'a'
I68195
sS's'
I18
sS'm'
(lp34903
I1
aI3
aI7
aI1
aI19
aI41
aI81
aI61
aI461
aI275
aI131
aI2665
aI5615
aI1719
aI21047
aI42025
aI97725
aI196587
asS'd'
I17452
sa(dp34904
S'a'
I68209
sS's'
I18
sS'm'
(lp34905
I1
aI1
aI7
aI11
aI3
aI5
aI47
aI107
aI397
aI237
aI1795
aI3049
aI5317
aI14147
aI15299
aI50469
aI83855
aI75685
asS'd'
I17453
sa(dp34906
S'a'
I68245
sS's'
I18
sS'm'
(lp34907
I1
aI1
aI7
aI7
aI19
aI31
aI39
aI153
aI225
aI591
aI1547
aI3755
aI3219
aI15823
aI4015
aI30977
aI63999
aI198023
asS'd'
I17454
sa(dp34908
S'a'
I68246
sS's'
I18
sS'm'
(lp34909
I1
aI3
aI1
aI1
aI17
aI57
aI91
aI3
aI425
aI465
aI735
aI719
aI2955
aI3003
aI6669
aI14335
aI32137
aI82265
asS'd'
I17455
sa(dp34910
S'a'
I68274
sS's'
I18
sS'm'
(lp34911
I1
aI3
aI1
aI9
aI13
aI25
aI93
aI187
aI119
aI735
aI447
aI3387
aI5111
aI6525
aI28241
aI37643
aI99023
aI58551
asS'd'
I17456
sa(dp34912
S'a'
I68293
sS's'
I18
sS'm'
(lp34913
I1
aI1
aI7
aI1
aI23
aI15
aI105
aI89
aI109
aI743
aI2007
aI3131
aI4839
aI3285
aI14681
aI47097
aI69531
aI104647
asS'd'
I17457
sa(dp34914
S'a'
I68294
sS's'
I18
sS'm'
(lp34915
I1
aI1
aI5
aI7
aI27
aI33
aI85
aI109
aI165
aI569
aI511
aI3223
aI2201
aI2869
aI30457
aI42585
aI125187
aI83115
asS'd'
I17458
sa(dp34916
S'a'
I68322
sS's'
I18
sS'm'
(lp34917
I1
aI1
aI7
aI13
aI1
aI39
aI55
aI69
aI279
aI757
aI425
aI1317
aI2403
aI10711
aI3341
aI33491
aI5607
aI214161
asS'd'
I17459
sa(dp34918
S'a'
I68353
sS's'
I18
sS'm'
(lp34919
I1
aI3
aI3
aI9
aI23
aI57
aI109
aI9
aI473
aI323
aI1371
aI657
aI5039
aI1947
aI12787
aI29099
aI81887
aI44039
asS'd'
I17460
sa(dp34920
S'a'
I68359
sS's'
I18
sS'm'
(lp34921
I1
aI3
aI1
aI9
aI21
aI49
aI39
aI71
aI493
aI611
aI1465
aI3965
aI7509
aI5315
aI4095
aI21865
aI123533
aI148467
asS'd'
I17461
sa(dp34922
S'a'
I68377
sS's'
I18
sS'm'
(lp34923
I1
aI1
aI7
aI5
aI15
aI45
aI67
aI109
aI143
aI423
aI205
aI849
aI1291
aI245
aI26275
aI62873
aI69177
aI173705
asS'd'
I17462
sa(dp34924
S'a'
I68401
sS's'
I18
sS'm'
(lp34925
I1
aI3
aI7
aI13
aI3
aI13
aI103
aI45
aI171
aI919
aI903
aI2171
aI5025
aI14855
aI895
aI36937
aI37643
aI30311
asS'd'
I17463
sa(dp34926
S'a'
I68422
sS's'
I18
sS'm'
(lp34927
I1
aI3
aI1
aI5
aI31
aI35
aI67
aI111
aI201
aI183
aI375
aI905
aI5705
aI8839
aI31551
aI22525
aI53013
aI34189
asS'd'
I17464
sa(dp34928
S'a'
I68434
sS's'
I18
sS'm'
(lp34929
I1
aI1
aI3
aI7
aI25
aI11
aI85
aI231
aI285
aI957
aI1731
aI1267
aI8179
aI14195
aI18405
aI8489
aI32503
aI86257
asS'd'
I17465
sa(dp34930
S'a'
I68443
sS's'
I18
sS'm'
(lp34931
I1
aI3
aI7
aI5
aI29
aI7
aI73
aI15
aI501
aI909
aI1325
aI793
aI4479
aI12137
aI30871
aI36243
aI109781
aI7235
asS'd'
I17466
sa(dp34932
S'a'
I68445
sS's'
I18
sS'm'
(lp34933
I1
aI1
aI3
aI15
aI13
aI13
aI75
aI221
aI497
aI921
aI1939
aI2791
aI5277
aI6257
aI11129
aI109
aI27549
aI44901
asS'd'
I17467
sa(dp34934
S'a'
I68476
sS's'
I18
sS'm'
(lp34935
I1
aI1
aI7
aI3
aI21
aI9
aI35
aI113
aI101
aI15
aI545
aI2429
aI5869
aI11379
aI14427
aI28605
aI108313
aI220523
asS'd'
I17468
sa(dp34936
S'a'
I68492
sS's'
I18
sS'm'
(lp34937
I1
aI3
aI5
aI5
aI21
aI31
aI79
aI101
aI11
aI687
aI609
aI3741
aI1259
aI1529
aI10185
aI49863
aI86529
aI5147
asS'd'
I17469
sa(dp34938
S'a'
I68504
sS's'
I18
sS'm'
(lp34939
I1
aI3
aI1
aI5
aI5
aI49
aI105
aI213
aI435
aI201
aI511
aI525
aI5219
aI9503
aI32023
aI25407
aI2493
aI51165
asS'd'
I17470
sa(dp34940
S'a'
I68525
sS's'
I18
sS'm'
(lp34941
I1
aI1
aI5
aI9
aI9
aI61
aI67
aI107
aI351
aI519
aI1373
aI1261
aI1069
aI4325
aI9579
aI37117
aI71759
aI17601
asS'd'
I17471
sa(dp34942
S'a'
I68540
sS's'
I18
sS'm'
(lp34943
I1
aI1
aI7
aI7
aI27
aI63
aI63
aI229
aI239
aI291
aI1813
aI3831
aI8091
aI2553
aI18445
aI60707
aI88855
aI224325
asS'd'
I17472
sa(dp34944
S'a'
I68543
sS's'
I18
sS'm'
(lp34945
I1
aI1
aI5
aI7
aI15
aI23
aI109
aI7
aI409
aI447
aI185
aI3535
aI4643
aI13431
aI11107
aI48771
aI95843
aI155889
asS'd'
I17473
sa(dp34946
S'a'
I68555
sS's'
I18
sS'm'
(lp34947
I1
aI3
aI3
aI11
aI9
aI47
aI13
aI31
aI83
aI837
aI1661
aI2283
aI299
aI13161
aI25305
aI6079
aI107237
aI58477
asS'd'
I17474
sa(dp34948
S'a'
I68563
sS's'
I18
sS'm'
(lp34949
I1
aI3
aI7
aI15
aI3
aI31
aI21
aI245
aI105
aI141
aI703
aI71
aI1887
aI9345
aI15719
aI37737
aI58431
aI195997
asS'd'
I17475
sa(dp34950
S'a'
I68576
sS's'
I18
sS'm'
(lp34951
I1
aI1
aI1
aI13
aI25
aI21
aI23
aI67
aI349
aI581
aI1585
aI809
aI3955
aI4621
aI25989
aI25633
aI107229
aI193271
asS'd'
I17476
sa(dp34952
S'a'
I68581
sS's'
I18
sS'm'
(lp34953
I1
aI1
aI5
aI3
aI15
aI57
aI111
aI147
aI243
aI575
aI851
aI3461
aI5171
aI4203
aI21855
aI59579
aI90509
aI16897
asS'd'
I17477
sa(dp34954
S'a'
I68617
sS's'
I18
sS'm'
(lp34955
I1
aI1
aI5
aI13
aI5
aI45
aI83
aI59
aI253
aI261
aI1277
aI3179
aI6397
aI4277
aI6629
aI10979
aI55759
aI3033
asS'd'
I17478
sa(dp34956
S'a'
I68631
sS's'
I18
sS'm'
(lp34957
I1
aI1
aI1
aI9
aI25
aI1
aI127
aI159
aI273
aI357
aI1343
aI3209
aI649
aI6631
aI1365
aI40813
aI98955
aI181679
asS'd'
I17479
sa(dp34958
S'a'
I68656
sS's'
I18
sS'm'
(lp34959
I1
aI3
aI1
aI11
aI1
aI9
aI67
aI5
aI41
aI661
aI863
aI3769
aI2737
aI7261
aI26829
aI43093
aI113025
aI127975
asS'd'
I17480
sa(dp34960
S'a'
I68666
sS's'
I18
sS'm'
(lp34961
I1
aI1
aI3
aI3
aI7
aI5
aI77
aI207
aI125
aI625
aI437
aI1059
aI2635
aI1099
aI25567
aI63759
aI97575
aI231313
asS'd'
I17481
sa(dp34962
S'a'
I68714
sS's'
I18
sS'm'
(lp34963
I1
aI3
aI7
aI5
aI21
aI41
aI11
aI177
aI489
aI405
aI1831
aI1373
aI6267
aI11275
aI23613
aI55565
aI120353
aI98771
asS'd'
I17482
sa(dp34964
S'a'
I68743
sS's'
I18
sS'm'
(lp34965
I1
aI1
aI3
aI11
aI15
aI55
aI103
aI185
aI493
aI755
aI1235
aI3143
aI4355
aI4887
aI11245
aI60103
aI4023
aI184729
asS'd'
I17483
sa(dp34966
S'a'
I68761
sS's'
I18
sS'm'
(lp34967
I1
aI3
aI1
aI1
aI23
aI5
aI103
aI117
aI269
aI101
aI2013
aI1781
aI6445
aI8753
aI15041
aI13993
aI28753
aI47133
asS'd'
I17484
sa(dp34968
S'a'
I68764
sS's'
I18
sS'm'
(lp34969
I1
aI3
aI7
aI5
aI9
aI23
aI1
aI203
aI19
aI535
aI1445
aI1713
aI5503
aI11555
aI6195
aI35797
aI55663
aI10187
asS'd'
I17485
sa(dp34970
S'a'
I68767
sS's'
I18
sS'm'
(lp34971
I1
aI3
aI5
aI1
aI15
aI3
aI125
aI225
aI447
aI269
aI1663
aI1823
aI4309
aI12243
aI16689
aI3889
aI41111
aI123355
asS'd'
I17486
sa(dp34972
S'a'
I68768
sS's'
I18
sS'm'
(lp34973
I1
aI1
aI1
aI5
aI21
aI33
aI83
aI147
aI243
aI101
aI1085
aI121
aI4939
aI6081
aI22621
aI52995
aI103047
aI82531
asS'd'
I17487
sa(dp34974
S'a'
I68786
sS's'
I18
sS'm'
(lp34975
I1
aI3
aI5
aI3
aI21
aI3
aI5
aI81
aI431
aI191
aI1973
aI3675
aI6691
aI8687
aI31619
aI51669
aI95541
aI110447
asS'd'
I17488
sa(dp34976
S'a'
I68788
sS's'
I18
sS'm'
(lp34977
I1
aI1
aI1
aI15
aI1
aI11
aI83
aI145
aI75
aI133
aI967
aI2837
aI5111
aI6099
aI9119
aI53661
aI128647
aI40557
asS'd'
I17489
sa(dp34978
S'a'
I68848
sS's'
I18
sS'm'
(lp34979
I1
aI1
aI1
aI7
aI3
aI57
aI101
aI231
aI255
aI117
aI1903
aI2133
aI3867
aI11299
aI647
aI58853
aI22153
aI135959
asS'd'
I17490
sa(dp34980
S'a'
I68851
sS's'
I18
sS'm'
(lp34981
I1
aI3
aI5
aI5
aI9
aI19
aI89
aI229
aI313
aI421
aI201
aI953
aI2487
aI6283
aI1305
aI33421
aI20933
aI164841
asS'd'
I17491
sa(dp34982
S'a'
I68914
sS's'
I18
sS'm'
(lp34983
I1
aI1
aI3
aI3
aI1
aI45
aI93
aI165
aI343
aI577
aI1329
aI3019
aI2727
aI14397
aI7123
aI63347
aI45525
aI35133
asS'd'
I17492
sa(dp34984
S'a'
I68946
sS's'
I18
sS'm'
(lp34985
I1
aI3
aI5
aI5
aI19
aI1
aI33
aI55
aI49
aI1003
aI1567
aI2539
aI7461
aI14641
aI7655
aI37499
aI65525
aI84961
asS'd'
I17493
sa(dp34986
S'a'
I68952
sS's'
I18
sS'm'
(lp34987
I1
aI3
aI5
aI5
aI29
aI3
aI77
aI39
aI251
aI791
aI215
aI3779
aI1589
aI3577
aI22299
aI24133
aI105449
aI257157
asS'd'
I17494
sa(dp34988
S'a'
I68973
sS's'
I18
sS'm'
(lp34989
I1
aI3
aI1
aI5
aI7
aI23
aI109
aI209
aI35
aI571
aI1047
aI3453
aI3657
aI11713
aI19379
aI57101
aI29943
aI60909
asS'd'
I17495
sa(dp34990
S'a'
I69001
sS's'
I18
sS'm'
(lp34991
I1
aI1
aI7
aI11
aI17
aI63
aI15
aI175
aI333
aI831
aI1447
aI1991
aI3339
aI2519
aI30127
aI51481
aI71935
aI144995
asS'd'
I17496
sa(dp34992
S'a'
I69016
sS's'
I18
sS'm'
(lp34993
I1
aI3
aI1
aI11
aI13
aI51
aI17
aI67
aI43
aI209
aI789
aI1285
aI5655
aI5841
aI10203
aI32053
aI15721
aI211725
asS'd'
I17497
sa(dp34994
S'a'
I69035
sS's'
I18
sS'm'
(lp34995
I1
aI1
aI5
aI7
aI31
aI49
aI69
aI255
aI325
aI819
aI1769
aI1961
aI7403
aI1241
aI2241
aI40425
aI14839
aI178969
asS'd'
I17498
sa(dp34996
S'a'
I69037
sS's'
I18
sS'm'
(lp34997
I1
aI1
aI3
aI9
aI1
aI5
aI9
aI35
aI167
aI865
aI337
aI1079
aI6195
aI10139
aI19215
aI57607
aI122437
aI197147
asS'd'
I17499
sa(dp34998
S'a'
I69049
sS's'
I18
sS'm'
(lp34999
I1
aI1
aI1
aI15
aI23
aI7
aI47
aI155
aI345
aI547
aI333
aI3747
aI961
aI1397
aI17067
aI33385
aI48253
aI138611
asS'd'
I17500
sa(dp35000
S'a'
I69052
sS's'
I18
sS'm'
(lp35001
I1
aI1
aI7
aI9
aI29
aI27
aI81
aI183
aI153
aI171
aI1125
aI1929
aI1047
aI12463
aI1543
aI42981
aI126163
aI203259
asS'd'
I17501
sa(dp35002
S'a'
I69058
sS's'
I18
sS'm'
(lp35003
I1
aI1
aI5
aI3
aI1
aI41
aI123
aI213
aI7
aI179
aI1
aI3527
aI1437
aI3545
aI2025
aI5325
aI27097
aI187823
asS'd'
I17502
sa(dp35004
S'a'
I69063
sS's'
I18
sS'm'
(lp35005
I1
aI1
aI7
aI1
aI19
aI5
aI111
aI251
aI431
aI91
aI1437
aI1155
aI335
aI9587
aI18287
aI23937
aI123331
aI3939
asS'd'
I17503
sa(dp35006
S'a'
I69075
sS's'
I18
sS'm'
(lp35007
I1
aI1
aI1
aI9
aI13
aI59
aI75
aI219
aI225
aI313
aI525
aI2003
aI7829
aI7063
aI22123
aI4263
aI95491
aI9375
asS'd'
I17504
sa(dp35008
S'a'
I69112
sS's'
I18
sS'm'
(lp35009
I1
aI3
aI1
aI1
aI27
aI29
aI71
aI189
aI169
aI301
aI165
aI2967
aI5147
aI7127
aI2191
aI34259
aI66605
aI149603
asS'd'
I17505
sa(dp35010
S'a'
I69117
sS's'
I18
sS'm'
(lp35011
I1
aI3
aI5
aI13
aI29
aI57
aI105
aI31
aI495
aI311
aI1161
aI2109
aI1471
aI1275
aI12761
aI58379
aI46365
aI229935
asS'd'
I17506
sa(dp35012
S'a'
I69124
sS's'
I18
sS'm'
(lp35013
I1
aI1
aI7
aI3
aI7
aI49
aI125
aI189
aI309
aI399
aI1361
aI3969
aI2177
aI8171
aI26623
aI41855
aI2809
aI5825
asS'd'
I17507
sa(dp35014
S'a'
I69127
sS's'
I18
sS'm'
(lp35015
I1
aI3
aI5
aI7
aI17
aI21
aI77
aI101
aI37
aI661
aI1743
aI2243
aI823
aI12431
aI26931
aI7163
aI108093
aI191305
asS'd'
I17508
sa(dp35016
S'a'
I69134
sS's'
I18
sS'm'
(lp35017
I1
aI1
aI7
aI5
aI27
aI55
aI109
aI119
aI13
aI727
aI421
aI3469
aI1137
aI6125
aI5107
aI52733
aI102891
aI147425
asS'd'
I17509
sa(dp35018
S'a'
I69139
sS's'
I18
sS'm'
(lp35019
I1
aI1
aI3
aI5
aI17
aI45
aI17
aI211
aI137
aI21
aI689
aI1487
aI233
aI9845
aI6499
aI52617
aI73081
aI198137
asS'd'
I17510
sa(dp35020
S'a'
I69161
sS's'
I18
sS'm'
(lp35021
I1
aI3
aI7
aI1
aI25
aI27
aI67
aI7
aI161
aI633
aI729
aI807
aI7371
aI7301
aI29499
aI45939
aI110565
aI219491
asS'd'
I17511
sa(dp35022
S'a'
I69196
sS's'
I18
sS'm'
(lp35023
I1
aI3
aI3
aI13
aI17
aI7
aI55
aI211
aI103
aI981
aI1809
aI1913
aI5705
aI14011
aI7405
aI13893
aI92053
aI17997
asS'd'
I17512
sa(dp35024
S'a'
I69224
sS's'
I18
sS'm'
(lp35025
I1
aI1
aI1
aI1
aI15
aI9
aI75
aI37
aI5
aI443
aI157
aI2749
aI5587
aI16087
aI14953
aI26793
aI21229
aI226879
asS'd'
I17513
sa(dp35026
S'a'
I69238
sS's'
I18
sS'm'
(lp35027
I1
aI3
aI3
aI3
aI9
aI13
aI113
aI7
aI255
aI647
aI235
aI1713
aI525
aI8579
aI20873
aI49565
aI43869
aI145823
asS'd'
I17514
sa(dp35028
S'a'
I69301
sS's'
I18
sS'm'
(lp35029
I1
aI1
aI5
aI15
aI9
aI1
aI119
aI189
aI73
aI321
aI1045
aI467
aI1565
aI14381
aI22683
aI7939
aI44337
aI231901
asS'd'
I17515
sa(dp35030
S'a'
I69316
sS's'
I18
sS'm'
(lp35031
I1
aI1
aI3
aI13
aI21
aI61
aI35
aI105
aI425
aI395
aI381
aI1205
aI3631
aI8099
aI23723
aI29435
aI94683
aI180367
asS'd'
I17516
sa(dp35032
S'a'
I69325
sS's'
I18
sS'm'
(lp35033
I1
aI3
aI3
aI13
aI19
aI15
aI59
aI111
aI355
aI165
aI857
aI3131
aI5037
aI2527
aI17533
aI53563
aI621
aI89837
asS'd'
I17517
sa(dp35034
S'a'
I69334
sS's'
I18
sS'm'
(lp35035
I1
aI3
aI3
aI11
aI11
aI41
aI3
aI75
aI179
aI325
aI897
aI3141
aI75
aI1735
aI493
aI1123
aI126763
aI68645
asS'd'
I17518
sa(dp35036
S'a'
I69347
sS's'
I18
sS'm'
(lp35037
I1
aI3
aI5
aI7
aI19
aI61
aI9
aI99
aI101
aI583
aI1967
aI621
aI1869
aI10693
aI2025
aI62797
aI85727
aI212309
asS'd'
I17519
sa(dp35038
S'a'
I69359
sS's'
I18
sS'm'
(lp35039
I1
aI3
aI1
aI13
aI23
aI47
aI15
aI29
aI199
aI889
aI423
aI3995
aI1655
aI10753
aI25301
aI55551
aI94829
aI205833
asS'd'
I17520
sa(dp35040
S'a'
I69381
sS's'
I18
sS'm'
(lp35041
I1
aI1
aI1
aI1
aI21
aI1
aI91
aI237
aI195
aI721
aI881
aI1155
aI4109
aI10367
aI1873
aI6851
aI13295
aI182363
asS'd'
I17521
sa(dp35042
S'a'
I69388
sS's'
I18
sS'm'
(lp35043
I1
aI1
aI5
aI15
aI19
aI35
aI37
aI197
aI137
aI255
aI93
aI681
aI949
aI15183
aI24785
aI39357
aI65547
aI149013
asS'd'
I17522
sa(dp35044
S'a'
I69393
sS's'
I18
sS'm'
(lp35045
I1
aI3
aI3
aI3
aI27
aI27
aI95
aI239
aI171
aI513
aI655
aI1629
aI4577
aI3005
aI1681
aI2581
aI59995
aI83981
asS'd'
I17523
sa(dp35046
S'a'
I69409
sS's'
I18
sS'm'
(lp35047
I1
aI3
aI3
aI7
aI29
aI33
aI111
aI85
aI437
aI297
aI1563
aI2411
aI6171
aI2043
aI17625
aI59093
aI995
aI211599
asS'd'
I17524
sa(dp35048
S'a'
I69422
sS's'
I18
sS'm'
(lp35049
I1
aI3
aI1
aI13
aI19
aI35
aI33
aI9
aI57
aI153
aI819
aI2017
aI5879
aI13559
aI23135
aI25981
aI41091
aI50975
asS'd'
I17525
sa(dp35050
S'a'
I69453
sS's'
I18
sS'm'
(lp35051
I1
aI3
aI5
aI11
aI23
aI53
aI11
aI123
aI119
aI57
aI1775
aI3457
aI7939
aI4999
aI10771
aI23571
aI30099
aI17361
asS'd'
I17526
sa(dp35052
S'a'
I69459
sS's'
I18
sS'm'
(lp35053
I1
aI1
aI7
aI11
aI27
aI13
aI7
aI215
aI7
aI1009
aI1967
aI1845
aI6679
aI13781
aI21797
aI18755
aI47131
aI245907
asS'd'
I17527
sa(dp35054
S'a'
I69481
sS's'
I18
sS'm'
(lp35055
I1
aI1
aI1
aI3
aI19
aI47
aI35
aI13
aI287
aI349
aI439
aI3125
aI2387
aI12483
aI3833
aI29399
aI27037
aI30235
asS'd'
I17528
sa(dp35056
S'a'
I69487
sS's'
I18
sS'm'
(lp35057
I1
aI3
aI1
aI15
aI17
aI41
aI15
aI21
aI499
aI87
aI1899
aI2835
aI1919
aI925
aI4525
aI12935
aI25021
aI106657
asS'd'
I17529
sa(dp35058
S'a'
I69490
sS's'
I18
sS'm'
(lp35059
I1
aI1
aI1
aI13
aI17
aI59
aI73
aI75
aI443
aI199
aI1871
aI3447
aI4517
aI8395
aI16661
aI30655
aI17871
aI231337
asS'd'
I17530
sa(dp35060
S'a'
I69495
sS's'
I18
sS'm'
(lp35061
I1
aI1
aI7
aI1
aI1
aI53
aI17
aI49
aI259
aI77
aI917
aI631
aI6061
aI12291
aI17715
aI49761
aI70699
aI68313
asS'd'
I17531
sa(dp35062
S'a'
I69501
sS's'
I18
sS'm'
(lp35063
I1
aI1
aI3
aI9
aI13
aI27
aI67
aI149
aI229
aI347
aI1397
aI3457
aI6047
aI13117
aI11
aI18121
aI70323
aI36441
asS'd'
I17532
sa(dp35064
S'a'
I69511
sS's'
I18
sS'm'
(lp35065
I1
aI3
aI5
aI7
aI27
aI13
aI69
aI177
aI451
aI87
aI647
aI3797
aI5433
aI3137
aI20213
aI9809
aI126877
aI55243
asS'd'
I17533
sa(dp35066
S'a'
I69512
sS's'
I18
sS'm'
(lp35067
I1
aI3
aI7
aI13
aI21
aI57
aI73
aI157
aI173
aI631
aI1527
aI337
aI5605
aI8041
aI2181
aI19567
aI19829
aI63353
asS'd'
I17534
sa(dp35068
S'a'
I69532
sS's'
I18
sS'm'
(lp35069
I1
aI3
aI3
aI7
aI11
aI5
aI111
aI161
aI247
aI553
aI435
aI3883
aI5639
aI10889
aI8953
aI58297
aI15197
aI99711
asS'd'
I17535
sa(dp35070
S'a'
I69542
sS's'
I18
sS'm'
(lp35071
I1
aI1
aI7
aI13
aI11
aI29
aI71
aI251
aI387
aI1003
aI1275
aI763
aI67
aI10597
aI5995
aI53677
aI4683
aI2157
asS'd'
I17536
sa(dp35072
S'a'
I69560
sS's'
I18
sS'm'
(lp35073
I1
aI1
aI3
aI9
aI23
aI27
aI93
aI209
aI325
aI517
aI297
aI3215
aI4359
aI395
aI10377
aI36967
aI69803
aI190037
asS'd'
I17537
sa(dp35074
S'a'
I69565
sS's'
I18
sS'm'
(lp35075
I1
aI1
aI1
aI3
aI27
aI61
aI21
aI229
aI469
aI3
aI387
aI523
aI4753
aI2267
aI9879
aI32113
aI60837
aI76205
asS'd'
I17538
sa(dp35076
S'a'
I69571
sS's'
I18
sS'm'
(lp35077
I1
aI3
aI1
aI7
aI31
aI31
aI67
aI15
aI161
aI699
aI713
aI2973
aI2007
aI693
aI21823
aI57549
aI28989
aI157879
asS'd'
I17539
sa(dp35078
S'a'
I69611
sS's'
I18
sS'm'
(lp35079
I1
aI3
aI1
aI1
aI3
aI63
aI111
aI61
aI311
aI685
aI1029
aI345
aI6763
aI16217
aI14505
aI9777
aI3513
aI160985
asS'd'
I17540
sa(dp35080
S'a'
I69633
sS's'
I18
sS'm'
(lp35081
I1
aI1
aI1
aI11
aI5
aI25
aI13
aI79
aI337
aI3
aI1997
aI3489
aI7621
aI12115
aI9221
aI7953
aI19067
aI52697
asS'd'
I17541
sa(dp35082
S'a'
I69640
sS's'
I18
sS'm'
(lp35083
I1
aI1
aI1
aI3
aI19
aI3
aI85
aI127
aI475
aI391
aI293
aI2249
aI1211
aI1185
aI17133
aI6753
aI65517
aI98157
asS'd'
I17542
sa(dp35084
S'a'
I69667
sS's'
I18
sS'm'
(lp35085
I1
aI1
aI5
aI9
aI11
aI31
aI57
aI107
aI315
aI983
aI1117
aI2189
aI4813
aI9925
aI26635
aI30589
aI32989
aI44195
asS'd'
I17543
sa(dp35086
S'a'
I69669
sS's'
I18
sS'm'
(lp35087
I1
aI1
aI7
aI7
aI21
aI1
aI1
aI221
aI421
aI199
aI539
aI3981
aI4627
aI15655
aI12621
aI20427
aI11619
aI187185
asS'd'
I17544
sa(dp35088
S'a'
I69679
sS's'
I18
sS'm'
(lp35089
I1
aI1
aI7
aI5
aI13
aI19
aI49
aI31
aI55
aI35
aI1847
aI3173
aI475
aI15245
aI30907
aI50075
aI130837
aI87283
asS'd'
I17545
sa(dp35090
S'a'
I69688
sS's'
I18
sS'm'
(lp35091
I1
aI1
aI7
aI15
aI13
aI47
aI13
aI17
aI169
aI185
aI1411
aI1689
aI2339
aI2159
aI10591
aI52283
aI26785
aI255707
asS'd'
I17546
sa(dp35092
S'a'
I69699
sS's'
I18
sS'm'
(lp35093
I1
aI3
aI5
aI7
aI3
aI29
aI7
aI83
aI329
aI747
aI1755
aI1067
aI2565
aI2437
aI12309
aI15043
aI97589
aI69409
asS'd'
I17547
sa(dp35094
S'a'
I69756
sS's'
I18
sS'm'
(lp35095
I1
aI3
aI3
aI9
aI19
aI49
aI9
aI231
aI427
aI131
aI485
aI1637
aI1129
aI14723
aI19071
aI47997
aI74613
aI171539
asS'd'
I17548
sa(dp35096
S'a'
I69759
sS's'
I18
sS'm'
(lp35097
I1
aI1
aI1
aI1
aI5
aI17
aI105
aI39
aI313
aI407
aI1321
aI3013
aI8035
aI4395
aI15917
aI21105
aI53599
aI21341
asS'd'
I17549
sa(dp35098
S'a'
I69770
sS's'
I18
sS'm'
(lp35099
I1
aI3
aI1
aI15
aI7
aI35
aI5
aI153
aI485
aI1019
aI713
aI1891
aI5023
aI13885
aI15911
aI48215
aI81719
aI228189
asS'd'
I17550
sa(dp35100
S'a'
I69772
sS's'
I18
sS'm'
(lp35101
I1
aI1
aI3
aI3
aI19
aI3
aI103
aI55
aI221
aI847
aI27
aI1653
aI4887
aI3617
aI30235
aI42353
aI67007
aI21443
asS'd'
I17551
sa(dp35102
S'a'
I69826
sS's'
I18
sS'm'
(lp35103
I1
aI1
aI7
aI15
aI15
aI39
aI65
aI189
aI251
aI411
aI1953
aI1187
aI141
aI14919
aI7763
aI50879
aI2569
aI63467
asS'd'
I17552
sa(dp35104
S'a'
I69840
sS's'
I18
sS'm'
(lp35105
I1
aI1
aI7
aI1
aI15
aI3
aI37
aI133
aI11
aI745
aI697
aI3755
aI1233
aI2009
aI25597
aI40661
aI40743
aI198117
asS'd'
I17553
sa(dp35106
S'a'
I69846
sS's'
I18
sS'm'
(lp35107
I1
aI1
aI7
aI15
aI5
aI17
aI13
aI253
aI197
aI491
aI1499
aI2141
aI6803
aI13833
aI27297
aI385
aI54341
aI64305
asS'd'
I17554
sa(dp35108
S'a'
I69868
sS's'
I18
sS'm'
(lp35109
I1
aI3
aI5
aI7
aI3
aI11
aI19
aI193
aI441
aI575
aI1649
aI1821
aI2621
aI15803
aI7343
aI37361
aI16467
aI60629
asS'd'
I17555
sa(dp35110
S'a'
I69885
sS's'
I18
sS'm'
(lp35111
I1
aI3
aI3
aI11
aI11
aI29
aI109
aI1
aI83
aI475
aI1913
aI1975
aI1289
aI5221
aI24221
aI7479
aI26683
aI203435
asS'd'
I17556
sa(dp35112
S'a'
I69946
sS's'
I18
sS'm'
(lp35113
I1
aI1
aI3
aI13
aI3
aI35
aI119
aI131
aI323
aI413
aI147
aI4009
aI3167
aI11161
aI30523
aI65223
aI109859
aI239317
asS'd'
I17557
sa(dp35114
S'a'
I69966
sS's'
I18
sS'm'
(lp35115
I1
aI3
aI1
aI11
aI25
aI17
aI103
aI165
aI437
aI163
aI1141
aI105
aI3655
aI8105
aI20859
aI50727
aI27915
aI19309
asS'd'
I17558
sa(dp35116
S'a'
I69973
sS's'
I18
sS'm'
(lp35117
I1
aI3
aI3
aI11
aI9
aI59
aI17
aI135
aI131
aI781
aI675
aI2865
aI7287
aI11431
aI3717
aI56691
aI54971
aI83433
asS'd'
I17559
sa(dp35118
S'a'
I69977
sS's'
I18
sS'm'
(lp35119
I1
aI1
aI1
aI11
aI3
aI1
aI59
aI35
aI299
aI927
aI1761
aI823
aI287
aI13271
aI30321
aI32895
aI45961
aI23151
asS'd'
I17560
sa(dp35120
S'a'
I69980
sS's'
I18
sS'm'
(lp35121
I1
aI3
aI3
aI7
aI11
aI3
aI11
aI115
aI241
aI497
aI1359
aI1789
aI6677
aI2683
aI21145
aI58185
aI46131
aI17591
asS'd'
I17561
sa(dp35122
S'a'
I69984
sS's'
I18
sS'm'
(lp35123
I1
aI3
aI1
aI3
aI17
aI5
aI65
aI169
aI247
aI1001
aI1183
aI1801
aI759
aI2797
aI28721
aI7549
aI112463
aI127451
asS'd'
I17562
sa(dp35124
S'a'
I69994
sS's'
I18
sS'm'
(lp35125
I1
aI1
aI1
aI11
aI1
aI49
aI5
aI227
aI333
aI793
aI759
aI2845
aI6261
aI6325
aI6581
aI35853
aI39737
aI21457
asS'd'
I17563
sa(dp35126
S'a'
I70024
sS's'
I18
sS'm'
(lp35127
I1
aI1
aI1
aI3
aI7
aI17
aI81
aI105
aI453
aI207
aI1113
aI301
aI4933
aI14715
aI18815
aI29165
aI85251
aI209171
asS'd'
I17564
sa(dp35128
S'a'
I70044
sS's'
I18
sS'm'
(lp35129
I1
aI3
aI1
aI13
aI3
aI25
aI7
aI109
aI249
aI649
aI1009
aI937
aI659
aI14605
aI13325
aI26003
aI45507
aI166837
asS'd'
I17565
sa(dp35130
S'a'
I70053
sS's'
I18
sS'm'
(lp35131
I1
aI3
aI7
aI11
aI19
aI57
aI55
aI213
aI261
aI325
aI761
aI3167
aI6823
aI15039
aI13329
aI30195
aI52103
aI27877
asS'd'
I17566
sa(dp35132
S'a'
I70086
sS's'
I18
sS'm'
(lp35133
I1
aI3
aI3
aI11
aI31
aI45
aI3
aI185
aI225
aI143
aI651
aI327
aI4263
aI6005
aI31577
aI57779
aI90485
aI48393
asS'd'
I17567
sa(dp35134
S'a'
I70113
sS's'
I18
sS'm'
(lp35135
I1
aI1
aI3
aI13
aI9
aI21
aI97
aI63
aI285
aI531
aI1275
aI175
aI693
aI3735
aI15137
aI62193
aI80533
aI196545
asS'd'
I17568
sa(dp35136
S'a'
I70120
sS's'
I18
sS'm'
(lp35137
I1
aI1
aI1
aI11
aI5
aI25
aI101
aI111
aI101
aI17
aI1999
aI3709
aI19
aI5087
aI20151
aI4781
aI88417
aI186293
asS'd'
I17569
sa(dp35138
S'a'
I70131
sS's'
I18
sS'm'
(lp35139
I1
aI1
aI5
aI7
aI31
aI37
aI39
aI85
aI451
aI189
aI1521
aI619
aI5021
aI2601
aI32447
aI43513
aI8317
aI170611
asS'd'
I17570
sa(dp35140
S'a'
I70150
sS's'
I18
sS'm'
(lp35141
I1
aI1
aI7
aI1
aI25
aI45
aI33
aI111
aI443
aI719
aI1869
aI3619
aI5751
aI2649
aI27823
aI55465
aI113203
aI23875
asS'd'
I17571
sa(dp35142
S'a'
I70178
sS's'
I18
sS'm'
(lp35143
I1
aI3
aI5
aI15
aI19
aI47
aI49
aI241
aI75
aI395
aI307
aI1001
aI137
aI7029
aI21661
aI39159
aI94129
aI106693
asS'd'
I17572
sa(dp35144
S'a'
I70198
sS's'
I18
sS'm'
(lp35145
I1
aI3
aI7
aI1
aI7
aI35
aI85
aI27
aI285
aI975
aI565
aI2119
aI5861
aI9229
aI15877
aI25017
aI10551
aI155357
asS'd'
I17573
sa(dp35146
S'a'
I70227
sS's'
I18
sS'm'
(lp35147
I1
aI3
aI3
aI5
aI29
aI41
aI17
aI159
aI211
aI571
aI907
aI1745
aI6541
aI11643
aI4441
aI54599
aI83359
aI57227
asS'd'
I17574
sa(dp35148
S'a'
I70285
sS's'
I18
sS'm'
(lp35149
I1
aI3
aI7
aI5
aI19
aI11
aI37
aI191
aI75
aI443
aI1833
aI1715
aI6949
aI2477
aI31161
aI15647
aI84305
aI82887
asS'd'
I17575
sa(dp35150
S'a'
I70288
sS's'
I18
sS'm'
(lp35151
I1
aI1
aI3
aI9
aI29
aI35
aI87
aI11
aI147
aI443
aI1659
aI2457
aI1615
aI16135
aI10729
aI31583
aI111583
aI52607
asS'd'
I17576
sa(dp35152
S'a'
I70291
sS's'
I18
sS'm'
(lp35153
I1
aI3
aI7
aI1
aI7
aI47
aI55
aI133
aI53
aI23
aI225
aI2689
aI3075
aI12435
aI8337
aI37065
aI58631
aI247415
asS'd'
I17577
sa(dp35154
S'a'
I70309
sS's'
I18
sS'm'
(lp35155
I1
aI3
aI3
aI9
aI23
aI39
aI5
aI17
aI353
aI443
aI627
aI1609
aI5277
aI3899
aI31111
aI5935
aI25445
aI161043
asS'd'
I17578
sa(dp35156
S'a'
I70334
sS's'
I18
sS'm'
(lp35157
I1
aI3
aI3
aI11
aI31
aI11
aI97
aI99
aI37
aI169
aI1361
aI689
aI5481
aI5935
aI11957
aI36761
aI105641
aI250905
asS'd'
I17579
sa(dp35158
S'a'
I70339
sS's'
I18
sS'm'
(lp35159
I1
aI1
aI7
aI15
aI31
aI33
aI3
aI201
aI125
aI649
aI315
aI497
aI7715
aI2331
aI9081
aI16073
aI88459
aI70475
asS'd'
I17580
sa(dp35160
S'a'
I70381
sS's'
I18
sS'm'
(lp35161
I1
aI1
aI7
aI13
aI25
aI25
aI39
aI193
aI185
aI253
aI495
aI1143
aI3745
aI3459
aI10935
aI22029
aI70213
aI245827
asS'd'
I17581
sa(dp35162
S'a'
I70413
sS's'
I18
sS'm'
(lp35163
I1
aI3
aI3
aI11
aI1
aI47
aI93
aI27
aI117
aI755
aI1837
aI4045
aI4839
aI3413
aI21395
aI41905
aI6505
aI158029
asS'd'
I17582
sa(dp35164
S'a'
I70414
sS's'
I18
sS'm'
(lp35165
I1
aI3
aI5
aI5
aI3
aI41
aI23
aI207
aI3
aI409
aI1635
aI3511
aI899
aI747
aI10623
aI44933
aI62439
aI75577
asS'd'
I17583
sa(dp35166
S'a'
I70481
sS's'
I18
sS'm'
(lp35167
I1
aI3
aI7
aI15
aI1
aI15
aI113
aI175
aI43
aI513
aI515
aI1295
aI1903
aI9961
aI20995
aI57319
aI40649
aI22799
asS'd'
I17584
sa(dp35168
S'a'
I70488
sS's'
I18
sS'm'
(lp35169
I1
aI1
aI5
aI9
aI7
aI25
aI99
aI167
aI117
aI547
aI777
aI3819
aI4409
aI13465
aI3963
aI53355
aI67895
aI58007
asS'd'
I17585
sa(dp35170
S'a'
I70493
sS's'
I18
sS'm'
(lp35171
I1
aI1
aI5
aI9
aI25
aI9
aI11
aI113
aI455
aI563
aI143
aI1507
aI4055
aI6805
aI25027
aI37645
aI475
aI193193
asS'd'
I17586
sa(dp35172
S'a'
I70515
sS's'
I18
sS'm'
(lp35173
I1
aI1
aI1
aI11
aI15
aI27
aI123
aI199
aI229
aI27
aI1285
aI4013
aI6541
aI11203
aI23705
aI56821
aI59665
aI151109
asS'd'
I17587
sa(dp35174
S'a'
I70540
sS's'
I18
sS'm'
(lp35175
I1
aI1
aI3
aI1
aI31
aI19
aI27
aI129
aI235
aI407
aI865
aI2723
aI5387
aI7727
aI2309
aI45787
aI118107
aI199907
asS'd'
I17588
sa(dp35176
S'a'
I70543
sS's'
I18
sS'm'
(lp35177
I1
aI1
aI5
aI15
aI17
aI1
aI21
aI167
aI165
aI203
aI745
aI825
aI7993
aI15191
aI13731
aI13417
aI543
aI201511
asS'd'
I17589
sa(dp35178
S'a'
I70558
sS's'
I18
sS'm'
(lp35179
I1
aI3
aI1
aI1
aI29
aI49
aI45
aI81
aI321
aI755
aI1319
aI633
aI4889
aI7809
aI6305
aI58233
aI20213
aI144915
asS'd'
I17590
sa(dp35180
S'a'
I70568
sS's'
I18
sS'm'
(lp35181
I1
aI1
aI3
aI13
aI1
aI31
aI73
aI173
aI111
aI961
aI1995
aI3827
aI879
aI5567
aI31103
aI13227
aI126611
aI204507
asS'd'
I17591
sa(dp35182
S'a'
I70582
sS's'
I18
sS'm'
(lp35183
I1
aI1
aI3
aI7
aI21
aI3
aI75
aI137
aI125
aI981
aI1991
aI1167
aI1249
aI3821
aI19503
aI52855
aI122329
aI68717
asS'd'
I17592
sa(dp35184
S'a'
I70593
sS's'
I18
sS'm'
(lp35185
I1
aI1
aI1
aI13
aI17
aI17
aI69
aI167
aI327
aI635
aI427
aI2125
aI7499
aI9715
aI24097
aI39361
aI64301
aI63411
asS'd'
I17593
sa(dp35186
S'a'
I70596
sS's'
I18
sS'm'
(lp35187
I1
aI1
aI1
aI1
aI31
aI57
aI55
aI31
aI289
aI251
aI823
aI2301
aI5965
aI3381
aI479
aI39545
aI93051
aI68683
asS'd'
I17594
sa(dp35188
S'a'
I70613
sS's'
I18
sS'm'
(lp35189
I1
aI1
aI3
aI3
aI15
aI27
aI117
aI37
aI29
aI851
aI1891
aI3507
aI6279
aI323
aI11451
aI57961
aI41487
aI188359
asS'd'
I17595
sa(dp35190
S'a'
I70620
sS's'
I18
sS'm'
(lp35191
I1
aI1
aI5
aI1
aI25
aI55
aI125
aI207
aI129
aI849
aI589
aI1381
aI3395
aI645
aI1157
aI29285
aI105423
aI104429
asS'd'
I17596
sa(dp35192
S'a'
I70647
sS's'
I18
sS'm'
(lp35193
I1
aI1
aI7
aI7
aI9
aI47
aI41
aI103
aI473
aI395
aI883
aI1087
aI2827
aI9685
aI6313
aI15461
aI39803
aI254865
asS'd'
I17597
sa(dp35194
S'a'
I70666
sS's'
I18
sS'm'
(lp35195
I1
aI1
aI7
aI3
aI17
aI55
aI71
aI119
aI159
aI185
aI1415
aI3033
aI3045
aI1403
aI18349
aI2727
aI123995
aI45953
asS'd'
I17598
sa(dp35196
S'a'
I70710
sS's'
I18
sS'm'
(lp35197
I1
aI1
aI3
aI15
aI17
aI11
aI19
aI25
aI483
aI29
aI1329
aI1779
aI2885
aI6655
aI28327
aI42255
aI87555
aI211051
asS'd'
I17599
sa(dp35198
S'a'
I70719
sS's'
I18
sS'm'
(lp35199
I1
aI3
aI5
aI11
aI29
aI19
aI43
aI141
aI157
aI87
aI1091
aI3505
aI3139
aI11919
aI12123
aI31581
aI116229
aI167875
asS'd'
I17600
sa(dp35200
S'a'
I70721
sS's'
I18
sS'm'
(lp35201
I1
aI3
aI1
aI5
aI25
aI55
aI113
aI219
aI491
aI607
aI1641
aI3833
aI3153
aI1881
aI16027
aI39923
aI38551
aI204819
asS'd'
I17601
sa(dp35202
S'a'
I70722
sS's'
I18
sS'm'
(lp35203
I1
aI1
aI5
aI3
aI7
aI9
aI73
aI181
aI305
aI211
aI1699
aI983
aI3051
aI11643
aI12445
aI44827
aI74613
aI199699
asS'd'
I17602
sa(dp35204
S'a'
I70742
sS's'
I18
sS'm'
(lp35205
I1
aI1
aI3
aI5
aI23
aI21
aI115
aI49
aI311
aI205
aI963
aI1357
aI4013
aI8357
aI7065
aI47757
aI7937
aI249935
asS'd'
I17603
sa(dp35206
S'a'
I70757
sS's'
I18
sS'm'
(lp35207
I1
aI1
aI1
aI9
aI23
aI61
aI21
aI165
aI9
aI829
aI457
aI3975
aI5831
aI10901
aI15871
aI36769
aI45899
aI162083
asS'd'
I17604
sa(dp35208
S'a'
I70764
sS's'
I18
sS'm'
(lp35209
I1
aI1
aI3
aI3
aI25
aI41
aI91
aI45
aI37
aI939
aI299
aI3815
aI6433
aI3121
aI10585
aI62125
aI51333
aI171615
asS'd'
I17605
sa(dp35210
S'a'
I70781
sS's'
I18
sS'm'
(lp35211
I1
aI1
aI5
aI11
aI1
aI1
aI39
aI45
aI141
aI803
aI1493
aI1151
aI6243
aI8683
aI30223
aI53661
aI7949
aI197291
asS'd'
I17606
sa(dp35212
S'a'
I70795
sS's'
I18
sS'm'
(lp35213
I1
aI1
aI3
aI1
aI17
aI35
aI29
aI253
aI395
aI933
aI1015
aI3431
aI139
aI9095
aI30745
aI39747
aI58837
aI28517
asS'd'
I17607
sa(dp35214
S'a'
I70803
sS's'
I18
sS'm'
(lp35215
I1
aI1
aI5
aI3
aI21
aI17
aI105
aI21
aI249
aI387
aI1985
aI951
aI6323
aI8221
aI24601
aI57367
aI18751
aI240661
asS'd'
I17608
sa(dp35216
S'a'
I70809
sS's'
I18
sS'm'
(lp35217
I1
aI1
aI7
aI9
aI5
aI21
aI23
aI149
aI243
aI501
aI935
aI855
aI1821
aI15885
aI2239
aI39091
aI93615
aI31411
asS'd'
I17609
sa(dp35218
S'a'
I70821
sS's'
I18
sS'm'
(lp35219
I1
aI1
aI1
aI3
aI23
aI11
aI43
aI5
aI65
aI193
aI1723
aI3253
aI7533
aI12987
aI571
aI56073
aI125061
aI97117
asS'd'
I17610
sa(dp35220
S'a'
I70846
sS's'
I18
sS'm'
(lp35221
I1
aI1
aI1
aI13
aI13
aI21
aI113
aI79
aI115
aI867
aI777
aI2199
aI501
aI2913
aI18697
aI14959
aI18369
aI41631
asS'd'
I17611
sa(dp35222
S'a'
I70882
sS's'
I18
sS'm'
(lp35223
I1
aI1
aI7
aI13
aI13
aI53
aI101
aI165
aI447
aI995
aI587
aI201
aI1701
aI6429
aI8647
aI59265
aI27321
aI110841
asS'd'
I17612
sa(dp35224
S'a'
I70887
sS's'
I18
sS'm'
(lp35225
I1
aI3
aI1
aI3
aI25
aI35
aI67
aI95
aI173
aI877
aI1133
aI3027
aI2321
aI12517
aI4313
aI24469
aI40313
aI253095
asS'd'
I17613
sa(dp35226
S'a'
I70940
sS's'
I18
sS'm'
(lp35227
I1
aI1
aI1
aI9
aI17
aI33
aI103
aI141
aI259
aI963
aI1975
aI2979
aI5017
aI15689
aI30659
aI55145
aI73737
aI43539
asS'd'
I17614
sa(dp35228
S'a'
I70961
sS's'
I18
sS'm'
(lp35229
I1
aI3
aI1
aI15
aI7
aI7
aI7
aI1
aI267
aI415
aI1591
aI17
aI2451
aI13415
aI6993
aI16631
aI90019
aI237161
asS'd'
I17615
sa(dp35230
S'a'
I70996
sS's'
I18
sS'm'
(lp35231
I1
aI1
aI7
aI11
aI11
aI37
aI107
aI143
aI263
aI49
aI1391
aI3269
aI6139
aI1413
aI26557
aI16369
aI86789
aI89151
asS'd'
I17616
sa(dp35232
S'a'
I71012
sS's'
I18
sS'm'
(lp35233
I1
aI1
aI7
aI13
aI27
aI41
aI3
aI169
aI453
aI547
aI157
aI3219
aI4711
aI9805
aI10657
aI8121
aI40229
aI247825
asS'd'
I17617
sa(dp35234
S'a'
I71046
sS's'
I18
sS'm'
(lp35235
I1
aI3
aI3
aI3
aI25
aI25
aI109
aI253
aI67
aI901
aI259
aI1159
aI6161
aI6763
aI19669
aI42775
aI74089
aI69821
asS'd'
I17618
sa(dp35236
S'a'
I71057
sS's'
I18
sS'm'
(lp35237
I1
aI3
aI7
aI15
aI11
aI25
aI91
aI137
aI247
aI851
aI511
aI1847
aI1179
aI411
aI9545
aI31275
aI46201
aI169677
asS'd'
I17619
sa(dp35238
S'a'
I71060
sS's'
I18
sS'm'
(lp35239
I1
aI1
aI5
aI3
aI3
aI61
aI19
aI167
aI491
aI765
aI1997
aI3267
aI883
aI15439
aI27581
aI24865
aI128245
aI130055
asS'd'
I17620
sa(dp35240
S'a'
I71063
sS's'
I18
sS'm'
(lp35241
I1
aI1
aI3
aI9
aI17
aI61
aI7
aI109
aI325
aI347
aI1109
aI889
aI2995
aI4763
aI21551
aI60137
aI91833
aI126989
asS'd'
I17621
sa(dp35242
S'a'
I71079
sS's'
I18
sS'm'
(lp35243
I1
aI3
aI3
aI7
aI5
aI17
aI61
aI107
aI209
aI577
aI885
aI2611
aI1471
aI7549
aI16199
aI12319
aI48865
aI242229
asS'd'
I17622
sa(dp35244
S'a'
I71080
sS's'
I18
sS'm'
(lp35245
I1
aI3
aI5
aI1
aI5
aI49
aI85
aI177
aI213
aI583
aI857
aI179
aI1805
aI4297
aI5835
aI61923
aI22741
aI261983
asS'd'
I17623
sa(dp35246
S'a'
I71111
sS's'
I18
sS'm'
(lp35247
I1
aI3
aI1
aI13
aI1
aI1
aI83
aI227
aI457
aI375
aI567
aI1563
aI2085
aI8153
aI12563
aI44561
aI115487
aI188351
asS'd'
I17624
sa(dp35248
S'a'
I71118
sS's'
I18
sS'm'
(lp35249
I1
aI3
aI1
aI9
aI15
aI39
aI127
aI135
aI181
aI967
aI1495
aI3187
aI7463
aI9651
aI26261
aI57435
aI42069
aI48549
asS'd'
I17625
sa(dp35250
S'a'
I71129
sS's'
I18
sS'm'
(lp35251
I1
aI3
aI1
aI7
aI5
aI31
aI111
aI19
aI19
aI855
aI273
aI2089
aI6001
aI2799
aI26013
aI6625
aI75623
aI150185
asS'd'
I17626
sa(dp35252
S'a'
I71136
sS's'
I18
sS'm'
(lp35253
I1
aI1
aI7
aI1
aI31
aI19
aI15
aI159
aI35
aI791
aI1005
aI3947
aI7031
aI41
aI28807
aI45299
aI37761
aI101191
asS'd'
I17627
sa(dp35254
S'a'
I71142
sS's'
I18
sS'm'
(lp35255
I1
aI3
aI5
aI3
aI15
aI7
aI7
aI67
aI329
aI367
aI843
aI2309
aI3023
aI5369
aI21561
aI18881
aI14395
aI193369
asS'd'
I17628
sa(dp35256
S'a'
I71145
sS's'
I18
sS'm'
(lp35257
I1
aI3
aI3
aI11
aI21
aI53
aI3
aI251
aI87
aI131
aI563
aI847
aI8049
aI1639
aI30103
aI30461
aI108427
aI125197
asS'd'
I17629
sa(dp35258
S'a'
I71151
sS's'
I18
sS'm'
(lp35259
I1
aI1
aI1
aI5
aI21
aI45
aI79
aI229
aI29
aI133
aI1873
aI261
aI4221
aI3091
aI25569
aI11219
aI70693
aI227025
asS'd'
I17630
sa(dp35260
S'a'
I71163
sS's'
I18
sS'm'
(lp35261
I1
aI3
aI1
aI5
aI17
aI9
aI75
aI101
aI155
aI311
aI789
aI821
aI7361
aI3791
aI18511
aI57607
aI97647
aI42107
asS'd'
I17631
sa(dp35262
S'a'
I71218
sS's'
I18
sS'm'
(lp35263
I1
aI3
aI1
aI11
aI21
aI39
aI33
aI179
aI7
aI775
aI55
aI3779
aI6163
aI3575
aI27535
aI32363
aI9169
aI57133
asS'd'
I17632
sa(dp35264
S'a'
I71223
sS's'
I18
sS'm'
(lp35265
I1
aI3
aI7
aI3
aI19
aI33
aI19
aI11
aI173
aI175
aI219
aI3585
aI1115
aI15693
aI23481
aI45669
aI94149
aI19531
asS'd'
I17633
sa(dp35266
S'a'
I71232
sS's'
I18
sS'm'
(lp35267
I1
aI1
aI5
aI11
aI11
aI49
aI29
aI217
aI229
aI757
aI1031
aI3833
aI4235
aI13535
aI8765
aI20707
aI52851
aI9037
asS'd'
I17634
sa(dp35268
S'a'
I71237
sS's'
I18
sS'm'
(lp35269
I1
aI3
aI1
aI13
aI25
aI61
aI65
aI111
aI95
aI533
aI1235
aI2947
aI3239
aI9513
aI11395
aI9321
aI117535
aI228289
asS'd'
I17635
sa(dp35270
S'a'
I71272
sS's'
I18
sS'm'
(lp35271
I1
aI1
aI1
aI3
aI19
aI33
aI13
aI233
aI331
aI811
aI1931
aI1109
aI7705
aI3129
aI19757
aI44325
aI97903
aI165311
asS'd'
I17636
sa(dp35272
S'a'
I71323
sS's'
I18
sS'm'
(lp35273
I1
aI1
aI3
aI15
aI13
aI55
aI57
aI81
aI257
aI613
aI1305
aI653
aI6059
aI4935
aI15707
aI4717
aI1859
aI109265
asS'd'
I17637
sa(dp35274
S'a'
I71339
sS's'
I18
sS'm'
(lp35275
I1
aI1
aI7
aI15
aI15
aI19
aI19
aI91
aI213
aI311
aI1651
aI2215
aI6985
aI2989
aI11961
aI28647
aI111163
aI217187
asS'd'
I17638
sa(dp35276
S'a'
I71341
sS's'
I18
sS'm'
(lp35277
I1
aI3
aI3
aI5
aI15
aI31
aI45
aI193
aI119
aI11
aI511
aI3155
aI5989
aI813
aI32655
aI41531
aI121007
aI24733
asS'd'
I17639
sa(dp35278
S'a'
I71349
sS's'
I18
sS'm'
(lp35279
I1
aI3
aI7
aI1
aI19
aI63
aI61
aI11
aI225
aI677
aI1323
aI1655
aI7607
aI15691
aI27083
aI56743
aI116167
aI250413
asS'd'
I17640
sa(dp35280
S'a'
I71368
sS's'
I18
sS'm'
(lp35281
I1
aI1
aI1
aI15
aI7
aI25
aI27
aI213
aI171
aI1011
aI1483
aI119
aI6849
aI12527
aI20601
aI35701
aI68377
aI245669
asS'd'
I17641
sa(dp35282
S'a'
I71379
sS's'
I18
sS'm'
(lp35283
I1
aI3
aI7
aI5
aI27
aI5
aI7
aI117
aI127
aI871
aI631
aI3395
aI1501
aI4839
aI1857
aI45769
aI107597
aI90385
asS'd'
I17642
sa(dp35284
S'a'
I71395
sS's'
I18
sS'm'
(lp35285
I1
aI1
aI3
aI13
aI1
aI15
aI49
aI69
aI479
aI919
aI881
aI3069
aI5609
aI12795
aI30225
aI14411
aI122847
aI75569
asS'd'
I17643
sa(dp35286
S'a'
I71407
sS's'
I18
sS'm'
(lp35287
I1
aI1
aI5
aI9
aI1
aI15
aI91
aI207
aI235
aI667
aI321
aI2047
aI841
aI16049
aI12499
aI8799
aI8245
aI42199
asS'd'
I17644
sa(dp35288
S'a'
I71421
sS's'
I18
sS'm'
(lp35289
I1
aI3
aI7
aI15
aI11
aI19
aI99
aI163
aI331
aI953
aI791
aI3443
aI3215
aI8025
aI1999
aI43685
aI72595
aI153185
asS'd'
I17645
sa(dp35290
S'a'
I71430
sS's'
I18
sS'm'
(lp35291
I1
aI3
aI1
aI13
aI25
aI23
aI17
aI133
aI59
aI233
aI151
aI1971
aI3611
aI3951
aI16979
aI991
aI73325
aI158475
asS'd'
I17646
sa(dp35292
S'a'
I71436
sS's'
I18
sS'm'
(lp35293
I1
aI3
aI5
aI11
aI1
aI53
aI123
aI81
aI285
aI457
aI1183
aI489
aI939
aI3069
aI15845
aI24799
aI81301
aI105187
asS'd'
I17647
sa(dp35294
S'a'
I71454
sS's'
I18
sS'm'
(lp35295
I1
aI3
aI5
aI1
aI11
aI5
aI61
aI151
aI5
aI813
aI1347
aI1107
aI4915
aI4035
aI18709
aI20909
aI60569
aI55007
asS'd'
I17648
sa(dp35296
S'a'
I71467
sS's'
I18
sS'm'
(lp35297
I1
aI3
aI3
aI7
aI27
aI41
aI79
aI193
aI471
aI415
aI937
aI2561
aI1669
aI9213
aI21145
aI44917
aI64763
aI33195
asS'd'
I17649
sa(dp35298
S'a'
I71472
sS's'
I18
sS'm'
(lp35299
I1
aI3
aI7
aI13
aI31
aI5
aI71
aI237
aI419
aI957
aI1741
aI2829
aI5879
aI8143
aI8717
aI48995
aI114465
aI110295
asS'd'
I17650
sa(dp35300
S'a'
I71478
sS's'
I18
sS'm'
(lp35301
I1
aI1
aI7
aI3
aI7
aI23
aI83
aI161
aI381
aI313
aI383
aI2813
aI333
aI4647
aI18321
aI10437
aI111645
aI55509
asS'd'
I17651
sa(dp35302
S'a'
I71481
sS's'
I18
sS'm'
(lp35303
I1
aI1
aI5
aI9
aI23
aI1
aI83
aI121
aI245
aI37
aI1097
aI1437
aI3891
aI2727
aI30775
aI27649
aI95571
aI216245
asS'd'
I17652
sa(dp35304
S'a'
I71501
sS's'
I18
sS'm'
(lp35305
I1
aI1
aI5
aI7
aI1
aI43
aI59
aI253
aI329
aI421
aI791
aI3945
aI2599
aI2243
aI11121
aI37761
aI27223
aI176867
asS'd'
I17653
sa(dp35306
S'a'
I71519
sS's'
I18
sS'm'
(lp35307
I1
aI3
aI5
aI5
aI25
aI59
aI85
aI155
aI367
aI291
aI1025
aI1415
aI7871
aI14191
aI23249
aI32233
aI93253
aI177869
asS'd'
I17654
sa(dp35308
S'a'
I71530
sS's'
I18
sS'm'
(lp35309
I1
aI1
aI1
aI9
aI21
aI41
aI111
aI241
aI177
aI999
aI779
aI2827
aI1683
aI6405
aI16133
aI26523
aI102567
aI190313
asS'd'
I17655
sa(dp35310
S'a'
I71550
sS's'
I18
sS'm'
(lp35311
I1
aI3
aI3
aI15
aI13
aI59
aI69
aI239
aI231
aI511
aI1675
aI147
aI4041
aI3723
aI29191
aI24913
aI15601
aI198141
asS'd'
I17656
sa(dp35312
S'a'
I71573
sS's'
I18
sS'm'
(lp35313
I1
aI3
aI3
aI9
aI17
aI29
aI1
aI107
aI243
aI509
aI1949
aI205
aI1693
aI6339
aI31591
aI61527
aI128043
aI222497
asS'd'
I17657
sa(dp35314
S'a'
I71593
sS's'
I18
sS'm'
(lp35315
I1
aI3
aI7
aI3
aI21
aI7
aI87
aI57
aI9
aI209
aI1831
aI2189
aI5523
aI8509
aI23687
aI46221
aI87469
aI146815
asS'd'
I17658
sa(dp35316
S'a'
I71608
sS's'
I18
sS'm'
(lp35317
I1
aI3
aI7
aI9
aI5
aI45
aI51
aI207
aI401
aI681
aI469
aI1951
aI793
aI16379
aI32143
aI55457
aI91787
aI178569
asS'd'
I17659
sa(dp35318
S'a'
I71619
sS's'
I18
sS'm'
(lp35319
I1
aI3
aI3
aI9
aI19
aI15
aI121
aI143
aI243
aI795
aI1839
aI2411
aI7175
aI11535
aI31995
aI4157
aI20111
aI92653
asS'd'
I17660
sa(dp35320
S'a'
I71622
sS's'
I18
sS'm'
(lp35321
I1
aI3
aI1
aI1
aI1
aI5
aI47
aI107
aI455
aI429
aI1411
aI2375
aI2823
aI14657
aI16297
aI21893
aI115257
aI50343
asS'd'
I17661
sa(dp35322
S'a'
I71636
sS's'
I18
sS'm'
(lp35323
I1
aI3
aI5
aI9
aI19
aI63
aI55
aI197
aI281
aI797
aI1539
aI2601
aI4497
aI1631
aI26583
aI23819
aI104553
aI27285
asS'd'
I17662
sa(dp35324
S'a'
I71650
sS's'
I18
sS'm'
(lp35325
I1
aI1
aI3
aI9
aI23
aI41
aI123
aI37
aI327
aI789
aI1711
aI1299
aI6735
aI7243
aI30635
aI21251
aI56081
aI65623
asS'd'
I17663
sa(dp35326
S'a'
I71655
sS's'
I18
sS'm'
(lp35327
I1
aI3
aI7
aI15
aI25
aI49
aI89
aI231
aI133
aI1003
aI351
aI765
aI7115
aI16239
aI1141
aI44063
aI31519
aI233719
asS'd'
I17664
sa(dp35328
S'a'
I71685
sS's'
I18
sS'm'
(lp35329
I1
aI3
aI7
aI3
aI13
aI37
aI93
aI83
aI59
aI539
aI1185
aI525
aI705
aI993
aI1113
aI1871
aI60817
aI254075
asS'd'
I17665
sa(dp35330
S'a'
I71703
sS's'
I18
sS'm'
(lp35331
I1
aI3
aI7
aI11
aI25
aI33
aI23
aI21
aI141
aI451
aI25
aI1321
aI5139
aI8947
aI10305
aI30175
aI43123
aI113049
asS'd'
I17666
sa(dp35332
S'a'
I71714
sS's'
I18
sS'm'
(lp35333
I1
aI3
aI5
aI5
aI17
aI1
aI125
aI211
aI143
aI637
aI1175
aI1149
aI6775
aI11091
aI12503
aI5537
aI35379
aI30045
asS'd'
I17667
sa(dp35334
S'a'
I71748
sS's'
I18
sS'm'
(lp35335
I1
aI1
aI1
aI15
aI17
aI31
aI123
aI33
aI279
aI831
aI1247
aI2305
aI1033
aI3201
aI231
aI23173
aI34453
aI66617
asS'd'
I17668
sa(dp35336
S'a'
I71758
sS's'
I18
sS'm'
(lp35337
I1
aI3
aI3
aI1
aI3
aI1
aI23
aI115
aI421
aI553
aI273
aI4091
aI5965
aI7521
aI18393
aI31229
aI78533
aI243921
asS'd'
I17669
sa(dp35338
S'a'
I71766
sS's'
I18
sS'm'
(lp35339
I1
aI1
aI7
aI15
aI15
aI29
aI73
aI165
aI391
aI215
aI1801
aI45
aI7451
aI6969
aI27897
aI36599
aI103647
aI145165
asS'd'
I17670
sa(dp35340
S'a'
I71772
sS's'
I18
sS'm'
(lp35341
I1
aI3
aI3
aI13
aI5
aI7
aI67
aI81
aI477
aI301
aI1397
aI921
aI3777
aI12431
aI14753
aI50555
aI24497
aI52995
asS'd'
I17671
sa(dp35342
S'a'
I71785
sS's'
I18
sS'm'
(lp35343
I1
aI1
aI5
aI5
aI27
aI21
aI91
aI155
aI405
aI347
aI1135
aI3701
aI2471
aI577
aI3927
aI52605
aI1725
aI25803
asS'd'
I17672
sa(dp35344
S'a'
I71794
sS's'
I18
sS'm'
(lp35345
I1
aI3
aI1
aI3
aI15
aI57
aI117
aI175
aI437
aI13
aI1821
aI649
aI899
aI1295
aI2753
aI2183
aI47923
aI163407
asS'd'
I17673
sa(dp35346
S'a'
I71796
sS's'
I18
sS'm'
(lp35347
I1
aI3
aI5
aI9
aI23
aI15
aI103
aI179
aI233
aI787
aI715
aI3751
aI3321
aI2069
aI8299
aI43417
aI96549
aI180737
asS'd'
I17674
sa(dp35348
S'a'
I71803
sS's'
I18
sS'm'
(lp35349
I1
aI3
aI7
aI9
aI1
aI41
aI81
aI205
aI141
aI707
aI397
aI763
aI4797
aI8843
aI3311
aI37425
aI43873
aI131491
asS'd'
I17675
sa(dp35350
S'a'
I71809
sS's'
I18
sS'm'
(lp35351
I1
aI1
aI5
aI3
aI9
aI29
aI123
aI163
aI15
aI871
aI159
aI2615
aI6987
aI471
aI25653
aI11295
aI94481
aI195409
asS'd'
I17676
sa(dp35352
S'a'
I71858
sS's'
I18
sS'm'
(lp35353
I1
aI3
aI7
aI3
aI13
aI7
aI9
aI59
aI59
aI381
aI2027
aI2639
aI59
aI7977
aI14505
aI34327
aI99113
aI157439
asS'd'
I17677
sa(dp35354
S'a'
I71863
sS's'
I18
sS'm'
(lp35355
I1
aI1
aI5
aI9
aI27
aI13
aI87
aI85
aI443
aI531
aI1069
aI3479
aI6547
aI13943
aI13711
aI11007
aI37395
aI190293
asS'd'
I17678
sa(dp35356
S'a'
I71878
sS's'
I18
sS'm'
(lp35357
I1
aI3
aI7
aI9
aI19
aI47
aI95
aI229
aI395
aI979
aI359
aI1799
aI7389
aI14377
aI19371
aI56785
aI6699
aI215433
asS'd'
I17679
sa(dp35358
S'a'
I71895
sS's'
I18
sS'm'
(lp35359
I1
aI1
aI1
aI5
aI21
aI15
aI59
aI17
aI281
aI585
aI293
aI3029
aI2539
aI16089
aI19
aI34757
aI115811
aI235565
asS'd'
I17680
sa(dp35360
S'a'
I71899
sS's'
I18
sS'm'
(lp35361
I1
aI1
aI7
aI11
aI27
aI19
aI55
aI217
aI475
aI119
aI1291
aI1761
aI6879
aI4355
aI30019
aI17573
aI14987
aI204623
asS'd'
I17681
sa(dp35362
S'a'
I71906
sS's'
I18
sS'm'
(lp35363
I1
aI1
aI7
aI11
aI29
aI47
aI5
aI201
aI165
aI845
aI385
aI2903
aI7735
aI10855
aI14171
aI17881
aI45001
aI100725
asS'd'
I17682
sa(dp35364
S'a'
I71908
sS's'
I18
sS'm'
(lp35365
I1
aI1
aI5
aI5
aI31
aI57
aI99
aI35
aI315
aI363
aI1135
aI897
aI1041
aI729
aI26987
aI15299
aI29563
aI67293
asS'd'
I17683
sa(dp35366
S'a'
I71917
sS's'
I18
sS'm'
(lp35367
I1
aI1
aI5
aI7
aI27
aI13
aI99
aI81
aI491
aI887
aI1309
aI3343
aI7241
aI1289
aI12021
aI52533
aI101799
aI238721
asS'd'
I17684
sa(dp35368
S'a'
I71952
sS's'
I18
sS'm'
(lp35369
I1
aI1
aI1
aI11
aI27
aI49
aI79
aI235
aI49
aI215
aI2003
aI2771
aI5943
aI1183
aI31931
aI33885
aI56971
aI52665
asS'd'
I17685
sa(dp35370
S'a'
I71957
sS's'
I18
sS'm'
(lp35371
I1
aI1
aI1
aI13
aI13
aI1
aI123
aI1
aI191
aI747
aI859
aI2287
aI5113
aI3715
aI2217
aI61483
aI195
aI237163
asS'd'
I17686
sa(dp35372
S'a'
I71961
sS's'
I18
sS'm'
(lp35373
I1
aI3
aI7
aI15
aI13
aI13
aI17
aI207
aI141
aI821
aI231
aI1373
aI5355
aI6503
aI2403
aI18183
aI83717
aI170047
asS'd'
I17687
sa(dp35374
S'a'
I71973
sS's'
I18
sS'm'
(lp35375
I1
aI3
aI1
aI3
aI13
aI11
aI41
aI51
aI443
aI201
aI1349
aI2331
aI1009
aI16169
aI5247
aI50315
aI15589
aI150497
asS'd'
I17688
sa(dp35376
S'a'
I71980
sS's'
I18
sS'm'
(lp35377
I1
aI3
aI5
aI13
aI29
aI21
aI93
aI55
aI27
aI17
aI1615
aI3473
aI3641
aI10999
aI31955
aI4699
aI23585
aI141243
asS'd'
I17689
sa(dp35378
S'a'
I71997
sS's'
I18
sS'm'
(lp35379
I1
aI3
aI3
aI11
aI11
aI27
aI125
aI139
aI53
aI637
aI241
aI2651
aI4999
aI5923
aI16203
aI13645
aI95965
aI94459
asS'd'
I17690
sa(dp35380
S'a'
I72017
sS's'
I18
sS'm'
(lp35381
I1
aI3
aI7
aI3
aI9
aI53
aI87
aI171
aI489
aI691
aI303
aI3599
aI6093
aI841
aI3527
aI12953
aI22907
aI69823
asS'd'
I17691
sa(dp35382
S'a'
I72060
sS's'
I18
sS'm'
(lp35383
I1
aI1
aI5
aI13
aI31
aI11
aI33
aI207
aI437
aI683
aI703
aI1757
aI1443
aI14269
aI12677
aI20877
aI46791
aI176135
asS'd'
I17692
sa(dp35384
S'a'
I72091
sS's'
I18
sS'm'
(lp35385
I1
aI1
aI7
aI1
aI13
aI53
aI123
aI199
aI173
aI585
aI1099
aI3653
aI2253
aI13741
aI15675
aI38755
aI74545
aI139053
asS'd'
I17693
sa(dp35386
S'a'
I72097
sS's'
I18
sS'm'
(lp35387
I1
aI3
aI3
aI3
aI17
aI11
aI1
aI161
aI383
aI409
aI605
aI889
aI827
aI263
aI9677
aI42857
aI127691
aI99621
asS'd'
I17694
sa(dp35388
S'a'
I72166
sS's'
I18
sS'm'
(lp35389
I1
aI3
aI5
aI9
aI21
aI21
aI11
aI151
aI199
aI695
aI493
aI569
aI881
aI10533
aI11255
aI61997
aI124921
aI211139
asS'd'
I17695
sa(dp35390
S'a'
I72172
sS's'
I18
sS'm'
(lp35391
I1
aI3
aI5
aI5
aI9
aI47
aI109
aI7
aI195
aI287
aI97
aI3691
aI6929
aI6985
aI3063
aI16185
aI6313
aI228147
asS'd'
I17696
sa(dp35392
S'a'
I72183
sS's'
I18
sS'm'
(lp35393
I1
aI3
aI1
aI3
aI3
aI9
aI107
aI243
aI391
aI893
aI1207
aI2229
aI5295
aI723
aI14753
aI10921
aI104147
aI214941
asS'd'
I17697
sa(dp35394
S'a'
I72203
sS's'
I18
sS'm'
(lp35395
I1
aI3
aI1
aI9
aI5
aI63
aI63
aI247
aI413
aI805
aI285
aI4001
aI6735
aI3531
aI25949
aI44845
aI66959
aI194429
asS'd'
I17698
sa(dp35396
S'a'
I72205
sS's'
I18
sS'm'
(lp35397
I1
aI1
aI3
aI1
aI15
aI15
aI43
aI69
aI419
aI739
aI1739
aI1091
aI1043
aI3217
aI1139
aI44749
aI74131
aI165145
asS'd'
I17699
sa(dp35398
S'a'
I72213
sS's'
I18
sS'm'
(lp35399
I1
aI1
aI3
aI9
aI17
aI53
aI119
aI25
aI427
aI791
aI1873
aI481
aI6793
aI4767
aI30449
aI18079
aI52105
aI260371
asS'd'
I17700
sa(dp35400
S'a'
I72214
sS's'
I18
sS'm'
(lp35401
I1
aI1
aI3
aI15
aI15
aI41
aI15
aI53
aI395
aI571
aI1727
aI3081
aI4531
aI4215
aI22359
aI18165
aI91843
aI157273
asS'd'
I17701
sa(dp35402
S'a'
I72230
sS's'
I18
sS'm'
(lp35403
I1
aI3
aI3
aI9
aI19
aI15
aI55
aI185
aI321
aI285
aI695
aI1067
aI2551
aI1401
aI20023
aI22671
aI21365
aI89053
asS'd'
I17702
sa(dp35404
S'a'
I72239
sS's'
I18
sS'm'
(lp35405
I1
aI1
aI7
aI15
aI23
aI57
aI71
aI141
aI57
aI479
aI543
aI3783
aI3635
aI14011
aI23603
aI40877
aI21837
aI81079
asS'd'
I17703
sa(dp35406
S'a'
I72248
sS's'
I18
sS'm'
(lp35407
I1
aI3
aI7
aI11
aI31
aI5
aI83
aI177
aI105
aI981
aI331
aI2901
aI1781
aI8407
aI30199
aI19287
aI116219
aI78471
asS'd'
I17704
sa(dp35408
S'a'
I72251
sS's'
I18
sS'm'
(lp35409
I1
aI1
aI1
aI9
aI7
aI13
aI61
aI21
aI299
aI15
aI1045
aI475
aI7141
aI4827
aI5921
aI17323
aI42909
aI203623
asS'd'
I17705
sa(dp35410
S'a'
I72280
sS's'
I18
sS'm'
(lp35411
I1
aI3
aI7
aI1
aI17
aI59
aI99
aI221
aI77
aI37
aI1263
aI2137
aI1567
aI12473
aI20029
aI9231
aI32739
aI17021
asS'd'
I17706
sa(dp35412
S'a'
I72313
sS's'
I18
sS'm'
(lp35413
I1
aI1
aI1
aI5
aI23
aI61
aI39
aI13
aI97
aI191
aI1479
aI19
aI1913
aI3185
aI32393
aI59067
aI5483
aI158895
asS'd'
I17707
sa(dp35414
S'a'
I72338
sS's'
I18
sS'm'
(lp35415
I1
aI1
aI1
aI7
aI5
aI51
aI81
aI223
aI435
aI939
aI781
aI1153
aI6409
aI6369
aI30559
aI19007
aI50121
aI26525
asS'd'
I17708
sa(dp35416
S'a'
I72340
sS's'
I18
sS'm'
(lp35417
I1
aI3
aI3
aI5
aI29
aI1
aI57
aI127
aI153
aI897
aI161
aI683
aI295
aI11207
aI245
aI1819
aI3061
aI242609
asS'd'
I17709
sa(dp35418
S'a'
I72343
sS's'
I18
sS'm'
(lp35419
I1
aI1
aI1
aI7
aI5
aI19
aI105
aI57
aI263
aI433
aI1339
aI1479
aI6671
aI9917
aI26299
aI4573
aI68725
aI195
asS'd'
I17710
sa(dp35420
S'a'
I72368
sS's'
I18
sS'm'
(lp35421
I1
aI3
aI5
aI15
aI1
aI19
aI45
aI95
aI155
aI117
aI367
aI2051
aI1053
aI8847
aI6399
aI23641
aI95355
aI98415
asS'd'
I17711
sa(dp35422
S'a'
I72371
sS's'
I18
sS'm'
(lp35423
I1
aI3
aI3
aI15
aI3
aI1
aI5
aI115
aI349
aI747
aI1865
aI1669
aI659
aI7097
aI7871
aI3685
aI11013
aI59837
asS'd'
I17712
sa(dp35424
S'a'
I72426
sS's'
I18
sS'm'
(lp35425
I1
aI3
aI3
aI5
aI9
aI47
aI51
aI131
aI327
aI903
aI975
aI2481
aI3509
aI12481
aI4049
aI38053
aI4629
aI254415
asS'd'
I17713
sa(dp35426
S'a'
I72439
sS's'
I18
sS'm'
(lp35427
I1
aI3
aI5
aI9
aI31
aI47
aI1
aI29
aI37
aI683
aI1363
aI2527
aI4019
aI4965
aI14077
aI14191
aI101
aI1945
asS'd'
I17714
sa(dp35428
S'a'
I72446
sS's'
I18
sS'm'
(lp35429
I1
aI1
aI7
aI1
aI9
aI35
aI41
aI187
aI509
aI33
aI385
aI3907
aI1461
aI6827
aI6931
aI44723
aI109495
aI184641
asS'd'
I17715
sa(dp35430
S'a'
I72494
sS's'
I18
sS'm'
(lp35431
I1
aI3
aI5
aI9
aI5
aI49
aI21
aI171
aI353
aI927
aI409
aI913
aI5199
aI11747
aI8777
aI19891
aI63189
aI118839
asS'd'
I17716
sa(dp35432
S'a'
I72519
sS's'
I18
sS'm'
(lp35433
I1
aI3
aI3
aI1
aI27
aI49
aI43
aI157
aI75
aI469
aI787
aI3957
aI4147
aI13919
aI17489
aI57103
aI62091
aI135589
asS'd'
I17717
sa(dp35434
S'a'
I72573
sS's'
I18
sS'm'
(lp35435
I1
aI1
aI3
aI11
aI7
aI45
aI29
aI177
aI185
aI185
aI1537
aI127
aI121
aI817
aI31269
aI1677
aI20245
aI3835
asS'd'
I17718
sa(dp35436
S'a'
I72577
sS's'
I18
sS'm'
(lp35437
I1
aI1
aI5
aI3
aI21
aI9
aI67
aI79
aI391
aI971
aI1711
aI2607
aI5705
aI12863
aI12415
aI41255
aI26447
aI1643
asS'd'
I17719
sa(dp35438
S'a'
I72578
sS's'
I18
sS'm'
(lp35439
I1
aI1
aI5
aI5
aI9
aI63
aI67
aI245
aI31
aI225
aI309
aI1753
aI1507
aI817
aI4275
aI51843
aI22331
aI196875
asS'd'
I17720
sa(dp35440
S'a'
I72614
sS's'
I18
sS'm'
(lp35441
I1
aI3
aI1
aI13
aI15
aI39
aI3
aI245
aI147
aI485
aI241
aI2507
aI1859
aI7299
aI15037
aI41139
aI82757
aI224031
asS'd'
I17721
sa(dp35442
S'a'
I72628
sS's'
I18
sS'm'
(lp35443
I1
aI3
aI3
aI15
aI7
aI51
aI9
aI103
aI37
aI643
aI25
aI2067
aI7619
aI11991
aI12885
aI46809
aI109107
aI22393
asS'd'
I17722
sa(dp35444
S'a'
I72640
sS's'
I18
sS'm'
(lp35445
I1
aI1
aI1
aI9
aI1
aI55
aI119
aI85
aI115
aI827
aI187
aI2241
aI2553
aI577
aI12115
aI2391
aI69705
aI232101
asS'd'
I17723
sa(dp35446
S'a'
I72676
sS's'
I18
sS'm'
(lp35447
I1
aI1
aI5
aI7
aI13
aI61
aI125
aI129
aI475
aI703
aI1723
aI3233
aI5713
aI1941
aI21375
aI42119
aI75199
aI73163
asS'd'
I17724
sa(dp35448
S'a'
I72708
sS's'
I18
sS'm'
(lp35449
I1
aI1
aI5
aI5
aI5
aI21
aI73
aI155
aI493
aI81
aI1627
aI827
aI5925
aI7391
aI1587
aI39425
aI11807
aI64385
asS'd'
I17725
sa(dp35450
S'a'
I72717
sS's'
I18
sS'm'
(lp35451
I1
aI3
aI1
aI15
aI27
aI35
aI111
aI183
aI283
aI335
aI1387
aI669
aI6041
aI11637
aI26255
aI21113
aI121183
aI219703
asS'd'
I17726
sa(dp35452
S'a'
I72766
sS's'
I18
sS'm'
(lp35453
I1
aI3
aI5
aI15
aI15
aI23
aI119
aI91
aI197
aI809
aI975
aI3275
aI6171
aI11769
aI8385
aI5461
aI4561
aI29159
asS'd'
I17727
sa(dp35454
S'a'
I72786
sS's'
I18
sS'm'
(lp35455
I1
aI1
aI5
aI9
aI29
aI9
aI107
aI233
aI417
aI1005
aI799
aI1437
aI2679
aI15643
aI32341
aI54055
aI27861
aI115483
asS'd'
I17728
sa(dp35456
S'a'
I72788
sS's'
I18
sS'm'
(lp35457
I1
aI1
aI5
aI7
aI27
aI19
aI95
aI153
aI175
aI407
aI215
aI303
aI8165
aI14791
aI2099
aI61797
aI129411
aI10461
asS'd'
I17729
sa(dp35458
S'a'
I72795
sS's'
I18
sS'm'
(lp35459
I1
aI1
aI1
aI13
aI25
aI51
aI11
aI77
aI97
aI495
aI971
aI449
aI2833
aI7121
aI24105
aI34527
aI123135
aI129305
asS'd'
I17730
sa(dp35460
S'a'
I72797
sS's'
I18
sS'm'
(lp35461
I1
aI1
aI7
aI11
aI23
aI9
aI111
aI101
aI169
aI233
aI267
aI953
aI6379
aI15887
aI22921
aI33665
aI95195
aI159707
asS'd'
I17731
sa(dp35462
S'a'
I72798
sS's'
I18
sS'm'
(lp35463
I1
aI1
aI5
aI15
aI21
aI3
aI21
aI57
aI173
aI513
aI2027
aI1235
aI5031
aI5375
aI2717
aI23361
aI71817
aI232101
asS'd'
I17732
sa(dp35464
S'a'
I72808
sS's'
I18
sS'm'
(lp35465
I1
aI3
aI3
aI11
aI7
aI25
aI43
aI65
aI19
aI135
aI1611
aI85
aI7673
aI6459
aI27813
aI55557
aI100989
aI25205
asS'd'
I17733
sa(dp35466
S'a'
I72826
sS's'
I18
sS'm'
(lp35467
I1
aI3
aI7
aI1
aI15
aI37
aI55
aI141
aI239
aI205
aI647
aI3715
aI1617
aI13507
aI9847
aI64681
aI108711
aI231329
asS'd'
I17734
sa(dp35468
S'a'
I72835
sS's'
I18
sS'm'
(lp35469
I1
aI3
aI5
aI9
aI21
aI27
aI79
aI153
aI335
aI299
aI493
aI887
aI1457
aI16011
aI13795
aI50205
aI43319
aI130963
asS'd'
I17735
sa(dp35470
S'a'
I72837
sS's'
I18
sS'm'
(lp35471
I1
aI1
aI3
aI1
aI23
aI59
aI121
aI83
aI463
aI151
aI323
aI2977
aI4769
aI6011
aI20135
aI59541
aI23179
aI203487
asS'd'
I17736
sa(dp35472
S'a'
I72865
sS's'
I18
sS'm'
(lp35473
I1
aI1
aI5
aI7
aI9
aI17
aI63
aI149
aI59
aI281
aI763
aI619
aI2551
aI8179
aI2963
aI61283
aI107727
aI119817
asS'd'
I17737
sa(dp35474
S'a'
I72871
sS's'
I18
sS'm'
(lp35475
I1
aI3
aI7
aI3
aI15
aI39
aI11
aI145
aI141
aI965
aI505
aI2625
aI4335
aI7619
aI11007
aI43321
aI33199
aI212661
asS'd'
I17738
sa(dp35476
S'a'
I72872
sS's'
I18
sS'm'
(lp35477
I1
aI3
aI5
aI9
aI9
aI61
aI27
aI223
aI5
aI941
aI513
aI1437
aI481
aI9651
aI6567
aI57945
aI52547
aI21283
asS'd'
I17739
sa(dp35478
S'a'
I72898
sS's'
I18
sS'm'
(lp35479
I1
aI3
aI1
aI1
aI25
aI1
aI87
aI25
aI121
aI757
aI529
aI3857
aI1321
aI13479
aI5357
aI49341
aI5797
aI235895
asS'd'
I17740
sa(dp35480
S'a'
I72907
sS's'
I18
sS'm'
(lp35481
I1
aI3
aI5
aI1
aI21
aI35
aI37
aI215
aI509
aI165
aI1423
aI3067
aI4779
aI4693
aI12523
aI48099
aI69283
aI255111
asS'd'
I17741
sa(dp35482
S'a'
I72909
sS's'
I18
sS'm'
(lp35483
I1
aI1
aI3
aI1
aI31
aI15
aI45
aI127
aI339
aI331
aI1249
aI1075
aI6169
aI2941
aI30471
aI46789
aI118039
aI224651
asS'd'
I17742
sa(dp35484
S'a'
I72917
sS's'
I18
sS'm'
(lp35485
I1
aI1
aI5
aI13
aI21
aI37
aI39
aI61
aI191
aI17
aI177
aI3719
aI2177
aI11039
aI20047
aI14489
aI20475
aI171235
asS'd'
I17743
sa(dp35486
S'a'
I72934
sS's'
I18
sS'm'
(lp35487
I1
aI1
aI3
aI9
aI19
aI11
aI65
aI111
aI121
aI901
aI99
aI1861
aI3687
aI765
aI24861
aI46315
aI63433
aI171679
asS'd'
I17744
sa(dp35488
S'a'
I72987
sS's'
I18
sS'm'
(lp35489
I1
aI3
aI1
aI7
aI1
aI51
aI87
aI199
aI241
aI909
aI353
aI2471
aI7163
aI9547
aI16351
aI41129
aI12217
aI194099
asS'd'
I17745
sa(dp35490
S'a'
I73000
sS's'
I18
sS'm'
(lp35491
I1
aI1
aI7
aI7
aI17
aI17
aI127
aI67
aI51
aI217
aI1189
aI19
aI2099
aI10281
aI9071
aI21185
aI122821
aI110211
asS'd'
I17746
sa(dp35492
S'a'
I73005
sS's'
I18
sS'm'
(lp35493
I1
aI3
aI3
aI7
aI15
aI3
aI53
aI45
aI77
aI665
aI701
aI3175
aI6151
aI2639
aI19819
aI1063
aI25079
aI203343
asS'd'
I17747
sa(dp35494
S'a'
I73014
sS's'
I18
sS'm'
(lp35495
I1
aI1
aI5
aI11
aI11
aI5
aI103
aI11
aI481
aI999
aI713
aI499
aI5069
aI921
aI20619
aI25623
aI69601
aI82941
asS'd'
I17748
sa(dp35496
S'a'
I73023
sS's'
I18
sS'm'
(lp35497
I1
aI3
aI5
aI9
aI15
aI61
aI67
aI79
aI371
aI993
aI475
aI617
aI1611
aI12513
aI14907
aI55313
aI39207
aI112653
asS'd'
I17749
sa(dp35498
S'a'
I73025
sS's'
I18
sS'm'
(lp35499
I1
aI1
aI5
aI15
aI17
aI45
aI91
aI187
aI175
aI465
aI907
aI3371
aI3743
aI15657
aI30511
aI58191
aI105683
aI216759
asS'd'
I17750
sa(dp35500
S'a'
I73074
sS's'
I18
sS'm'
(lp35501
I1
aI1
aI5
aI9
aI23
aI1
aI17
aI79
aI73
aI717
aI1785
aI677
aI7377
aI4511
aI21927
aI34341
aI47119
aI193977
asS'd'
I17751
sa(dp35502
S'a'
I73076
sS's'
I18
sS'm'
(lp35503
I1
aI3
aI7
aI13
aI21
aI1
aI59
aI179
aI121
aI641
aI175
aI563
aI961
aI10549
aI15779
aI49875
aI8109
aI1039
asS'd'
I17752
sa(dp35504
S'a'
I73113
sS's'
I18
sS'm'
(lp35505
I1
aI1
aI5
aI7
aI17
aI9
aI37
aI171
aI335
aI135
aI1403
aI2541
aI3845
aI15311
aI1905
aI40853
aI11013
aI255669
asS'd'
I17753
sa(dp35506
S'a'
I73116
sS's'
I18
sS'm'
(lp35507
I1
aI1
aI7
aI1
aI5
aI23
aI113
aI111
aI337
aI755
aI2037
aI3067
aI2821
aI10549
aI28467
aI22615
aI71585
aI61871
asS'd'
I17754
sa(dp35508
S'a'
I73140
sS's'
I18
sS'm'
(lp35509
I1
aI1
aI3
aI9
aI7
aI3
aI49
aI229
aI111
aI871
aI1711
aI1793
aI3089
aI12571
aI30883
aI44773
aI80827
aI151709
asS'd'
I17755
sa(dp35510
S'a'
I73149
sS's'
I18
sS'm'
(lp35511
I1
aI1
aI5
aI1
aI13
aI41
aI3
aI253
aI399
aI881
aI1107
aI4081
aI1849
aI115
aI31557
aI2515
aI126751
aI195663
asS'd'
I17756
sa(dp35512
S'a'
I73186
sS's'
I18
sS'm'
(lp35513
I1
aI1
aI3
aI1
aI13
aI31
aI113
aI85
aI57
aI549
aI1653
aI2927
aI5433
aI11879
aI22709
aI41675
aI13395
aI46931
asS'd'
I17757
sa(dp35514
S'a'
I73188
sS's'
I18
sS'm'
(lp35515
I1
aI1
aI3
aI9
aI21
aI1
aI109
aI65
aI377
aI63
aI861
aI1031
aI2709
aI7265
aI9861
aI64109
aI34577
aI9743
asS'd'
I17758
sa(dp35516
S'a'
I73191
sS's'
I18
sS'm'
(lp35517
I1
aI1
aI7
aI7
aI17
aI31
aI5
aI177
aI253
aI387
aI1271
aI2805
aI2211
aI1813
aI11649
aI3217
aI123793
aI197753
asS'd'
I17759
sa(dp35518
S'a'
I73205
sS's'
I18
sS'm'
(lp35519
I1
aI1
aI3
aI9
aI15
aI63
aI89
aI59
aI455
aI783
aI1181
aI7
aI2309
aI15961
aI11231
aI37389
aI101221
aI119331
asS'd'
I17760
sa(dp35520
S'a'
I73215
sS's'
I18
sS'm'
(lp35521
I1
aI3
aI3
aI13
aI7
aI3
aI15
aI251
aI431
aI951
aI639
aI1585
aI1247
aI15927
aI9695
aI37469
aI34945
aI219723
asS'd'
I17761
sa(dp35522
S'a'
I73216
sS's'
I18
sS'm'
(lp35523
I1
aI1
aI5
aI3
aI21
aI29
aI83
aI151
aI383
aI227
aI215
aI2329
aI1297
aI13709
aI15653
aI3119
aI111319
aI222877
asS'd'
I17762
sa(dp35524
S'a'
I73239
sS's'
I18
sS'm'
(lp35525
I1
aI3
aI1
aI13
aI1
aI43
aI127
aI125
aI243
aI955
aI583
aI3497
aI6605
aI3821
aI4657
aI10599
aI90927
aI82725
asS'd'
I17763
sa(dp35526
S'a'
I73294
sS's'
I18
sS'm'
(lp35527
I1
aI1
aI1
aI1
aI15
aI51
aI61
aI167
aI489
aI603
aI873
aI907
aI575
aI6957
aI24409
aI63587
aI50205
aI159291
asS'd'
I17764
sa(dp35528
S'a'
I73311
sS's'
I18
sS'm'
(lp35529
I1
aI3
aI3
aI1
aI19
aI23
aI7
aI23
aI239
aI961
aI1001
aI1541
aI2211
aI4637
aI19931
aI39153
aI102769
aI242005
asS'd'
I17765
sa(dp35530
S'a'
I73330
sS's'
I18
sS'm'
(lp35531
I1
aI1
aI1
aI15
aI19
aI31
aI73
aI121
aI119
aI199
aI979
aI4061
aI3903
aI12055
aI27957
aI15999
aI5709
aI210329
asS'd'
I17766
sa(dp35532
S'a'
I73366
sS's'
I18
sS'm'
(lp35533
I1
aI1
aI7
aI13
aI25
aI35
aI21
aI25
aI241
aI937
aI13
aI947
aI943
aI3727
aI15321
aI46665
aI99437
aI233919
asS'd'
I17767
sa(dp35534
S'a'
I73379
sS's'
I18
sS'm'
(lp35535
I1
aI1
aI1
aI9
aI3
aI51
aI127
aI113
aI105
aI335
aI685
aI2173
aI4329
aI7569
aI5617
aI32407
aI21649
aI30609
asS'd'
I17768
sa(dp35536
S'a'
I73386
sS's'
I18
sS'm'
(lp35537
I1
aI3
aI7
aI9
aI9
aI13
aI69
aI191
aI95
aI727
aI1649
aI1201
aI2093
aI10053
aI29381
aI6207
aI70755
aI118505
asS'd'
I17769
sa(dp35538
S'a'
I73391
sS's'
I18
sS'm'
(lp35539
I1
aI1
aI1
aI11
aI21
aI57
aI77
aI139
aI271
aI21
aI1747
aI2337
aI7761
aI7753
aI6847
aI5219
aI87033
aI229105
asS'd'
I17770
sa(dp35540
S'a'
I73418
sS's'
I18
sS'm'
(lp35541
I1
aI3
aI3
aI1
aI5
aI51
aI15
aI235
aI87
aI567
aI391
aI3039
aI2253
aI11177
aI11899
aI25305
aI14815
aI51051
asS'd'
I17771
sa(dp35542
S'a'
I73438
sS's'
I18
sS'm'
(lp35543
I1
aI3
aI5
aI7
aI27
aI51
aI69
aI93
aI261
aI947
aI31
aI2751
aI6685
aI3655
aI24125
aI22161
aI108421
aI230865
asS'd'
I17772
sa(dp35544
S'a'
I73476
sS's'
I18
sS'm'
(lp35545
I1
aI1
aI3
aI3
aI27
aI17
aI35
aI97
aI285
aI855
aI1767
aI2545
aI825
aI11519
aI11231
aI50951
aI32883
aI78573
asS'd'
I17773
sa(dp35546
S'a'
I73503
sS's'
I18
sS'm'
(lp35547
I1
aI3
aI7
aI13
aI1
aI57
aI99
aI193
aI371
aI839
aI1319
aI2295
aI5897
aI7893
aI14339
aI64217
aI16951
aI234953
asS'd'
I17774
sa(dp35548
S'a'
I73509
sS's'
I18
sS'm'
(lp35549
I1
aI3
aI1
aI3
aI11
aI37
aI13
aI21
aI299
aI379
aI63
aI1209
aI7879
aI5001
aI10181
aI40173
aI1753
aI104821
asS'd'
I17775
sa(dp35550
S'a'
I73534
sS's'
I18
sS'm'
(lp35551
I1
aI3
aI5
aI9
aI23
aI51
aI75
aI103
aI249
aI533
aI621
aI15
aI1883
aI2109
aI20859
aI4635
aI120615
aI135515
asS'd'
I17776
sa(dp35552
S'a'
I73545
sS's'
I18
sS'm'
(lp35553
I1
aI1
aI3
aI3
aI21
aI45
aI113
aI57
aI495
aI457
aI685
aI3625
aI243
aI14831
aI12351
aI63001
aI118191
aI153875
asS'd'
I17777
sa(dp35554
S'a'
I73548
sS's'
I18
sS'm'
(lp35555
I1
aI3
aI5
aI5
aI21
aI15
aI65
aI251
aI183
aI241
aI1513
aI2711
aI4527
aI12675
aI26747
aI5181
aI4237
aI246479
asS'd'
I17778
sa(dp35556
S'a'
I73563
sS's'
I18
sS'm'
(lp35557
I1
aI1
aI5
aI3
aI27
aI29
aI91
aI93
aI345
aI893
aI195
aI3109
aI2611
aI12657
aI10401
aI15063
aI95807
aI244587
asS'd'
I17779
sa(dp35558
S'a'
I73575
sS's'
I18
sS'm'
(lp35559
I1
aI3
aI5
aI13
aI19
aI49
aI85
aI155
aI159
aI939
aI1139
aI1569
aI1129
aI8641
aI18391
aI55201
aI108491
aI77863
asS'd'
I17780
sa(dp35560
S'a'
I73582
sS's'
I18
sS'm'
(lp35561
I1
aI3
aI5
aI9
aI25
aI53
aI75
aI77
aI85
aI903
aI1399
aI2379
aI2219
aI14725
aI21877
aI24271
aI40955
aI61849
asS'd'
I17781
sa(dp35562
S'a'
I73603
sS's'
I18
sS'm'
(lp35563
I1
aI1
aI5
aI9
aI13
aI49
aI53
aI177
aI163
aI331
aI533
aI1469
aI1397
aI8187
aI12379
aI10185
aI125541
aI260271
asS'd'
I17782
sa(dp35564
S'a'
I73609
sS's'
I18
sS'm'
(lp35565
I1
aI1
aI7
aI9
aI19
aI31
aI81
aI89
aI281
aI397
aI1917
aI145
aI2723
aI15019
aI18841
aI13887
aI11859
aI171749
asS'd'
I17783
sa(dp35566
S'a'
I73610
sS's'
I18
sS'm'
(lp35567
I1
aI3
aI7
aI13
aI25
aI17
aI87
aI189
aI29
aI283
aI913
aI3855
aI5707
aI15881
aI12787
aI42357
aI84579
aI78531
asS'd'
I17784
sa(dp35568
S'a'
I73645
sS's'
I18
sS'm'
(lp35569
I1
aI1
aI5
aI15
aI15
aI31
aI37
aI249
aI445
aI119
aI431
aI4069
aI5699
aI10119
aI31661
aI9555
aI6869
aI1145
asS'd'
I17785
sa(dp35570
S'a'
I73646
sS's'
I18
sS'm'
(lp35571
I1
aI3
aI7
aI1
aI31
aI29
aI57
aI177
aI341
aI411
aI1019
aI1889
aI383
aI1461
aI26695
aI61777
aI18367
aI137233
asS'd'
I17786
sa(dp35572
S'a'
I73678
sS's'
I18
sS'm'
(lp35573
I1
aI3
aI5
aI1
aI7
aI53
aI29
aI53
aI387
aI675
aI435
aI461
aI6247
aI7519
aI14003
aI9037
aI116599
aI54471
asS'd'
I17787
sa(dp35574
S'a'
I73685
sS's'
I18
sS'm'
(lp35575
I1
aI3
aI3
aI7
aI17
aI59
aI91
aI77
aI169
aI591
aI95
aI113
aI6135
aI10479
aI17153
aI52953
aI16183
aI90775
asS'd'
I17788
sa(dp35576
S'a'
I73708
sS's'
I18
sS'm'
(lp35577
I1
aI3
aI3
aI15
aI11
aI11
aI117
aI141
aI493
aI163
aI65
aI1305
aI7477
aI7383
aI22651
aI64271
aI80983
aI154845
asS'd'
I17789
sa(dp35578
S'a'
I73747
sS's'
I18
sS'm'
(lp35579
I1
aI1
aI1
aI3
aI9
aI15
aI83
aI11
aI113
aI77
aI1115
aI1417
aI511
aI6825
aI21013
aI37241
aI104695
aI31335
asS'd'
I17790
sa(dp35580
S'a'
I73765
sS's'
I18
sS'm'
(lp35581
I1
aI1
aI7
aI3
aI13
aI33
aI115
aI121
aI245
aI673
aI1991
aI2157
aI479
aI9843
aI5963
aI4637
aI8925
aI27751
asS'd'
I17791
sa(dp35582
S'a'
I73797
sS's'
I18
sS'm'
(lp35583
I1
aI3
aI3
aI7
aI19
aI11
aI67
aI125
aI339
aI27
aI1545
aI2319
aI5977
aI11603
aI23219
aI48273
aI119265
aI20151
asS'd'
I17792
sa(dp35584
S'a'
I73810
sS's'
I18
sS'm'
(lp35585
I1
aI1
aI5
aI7
aI27
aI29
aI31
aI227
aI279
aI405
aI1133
aI689
aI1133
aI8957
aI29629
aI48849
aI109995
aI259749
asS'd'
I17793
sa(dp35586
S'a'
I73819
sS's'
I18
sS'm'
(lp35587
I1
aI3
aI5
aI7
aI7
aI61
aI95
aI243
aI91
aI741
aI1591
aI3169
aI2287
aI11015
aI15601
aI43043
aI65319
aI50671
asS'd'
I17794
sa(dp35588
S'a'
I73825
sS's'
I18
sS'm'
(lp35589
I1
aI1
aI5
aI5
aI13
aI47
aI31
aI95
aI425
aI715
aI1603
aI3485
aI673
aI12869
aI32561
aI42329
aI112809
aI181971
asS'd'
I17795
sa(dp35590
S'a'
I73845
sS's'
I18
sS'm'
(lp35591
I1
aI3
aI3
aI3
aI17
aI51
aI109
aI45
aI397
aI457
aI1379
aI3845
aI4215
aI14185
aI16597
aI27711
aI74283
aI98151
asS'd'
I17796
sa(dp35592
S'a'
I73855
sS's'
I18
sS'm'
(lp35593
I1
aI1
aI1
aI15
aI7
aI25
aI13
aI49
aI441
aI513
aI1769
aI707
aI6037
aI9689
aI18915
aI35647
aI110823
aI196633
asS'd'
I17797
sa(dp35594
S'a'
I73866
sS's'
I18
sS'm'
(lp35595
I1
aI1
aI1
aI1
aI23
aI53
aI93
aI61
aI277
aI125
aI55
aI2453
aI3331
aI14037
aI10809
aI33205
aI43785
aI248743
asS'd'
I17798
sa(dp35596
S'a'
I73890
sS's'
I18
sS'm'
(lp35597
I1
aI1
aI3
aI7
aI23
aI15
aI93
aI77
aI333
aI801
aI1969
aI31
aI51
aI5239
aI24241
aI5077
aI113503
aI132211
asS'd'
I17799
sa(dp35598
S'a'
I73919
sS's'
I18
sS'm'
(lp35599
I1
aI3
aI3
aI1
aI7
aI55
aI53
aI5
aI311
aI657
aI1507
aI3413
aI565
aI15745
aI6129
aI40285
aI91811
aI90527
asS'd'
I17800
sa(dp35600
S'a'
I73922
sS's'
I18
sS'm'
(lp35601
I1
aI3
aI3
aI13
aI19
aI23
aI45
aI25
aI509
aI313
aI915
aI2199
aI5549
aI8469
aI32735
aI37877
aI11607
aI37993
asS'd'
I17801
sa(dp35602
S'a'
I73936
sS's'
I18
sS'm'
(lp35603
I1
aI1
aI5
aI11
aI25
aI33
aI55
aI31
aI311
aI851
aI159
aI2103
aI2641
aI8957
aI9375
aI37179
aI33667
aI100513
asS'd'
I17802
sa(dp35604
S'a'
I73939
sS's'
I18
sS'm'
(lp35605
I1
aI3
aI1
aI1
aI7
aI7
aI17
aI75
aI217
aI171
aI359
aI1169
aI4105
aI929
aI6427
aI56349
aI77985
aI41941
asS'd'
I17803
sa(dp35606
S'a'
I73964
sS's'
I18
sS'm'
(lp35607
I1
aI1
aI1
aI15
aI5
aI51
aI73
aI63
aI259
aI351
aI1797
aI1001
aI5025
aI11203
aI30221
aI54345
aI11331
aI158415
asS'd'
I17804
sa(dp35608
S'a'
I73976
sS's'
I18
sS'm'
(lp35609
I1
aI3
aI1
aI5
aI3
aI21
aI13
aI3
aI63
aI779
aI871
aI2517
aI6345
aI4103
aI16321
aI30211
aI120815
aI83751
asS'd'
I17805
sa(dp35610
S'a'
I73984
sS's'
I18
sS'm'
(lp35611
I1
aI1
aI3
aI13
aI11
aI7
aI41
aI255
aI215
aI37
aI279
aI2485
aI6511
aI12855
aI22857
aI55695
aI122717
aI238151
asS'd'
I17806
sa(dp35612
S'a'
I73990
sS's'
I18
sS'm'
(lp35613
I1
aI3
aI3
aI3
aI1
aI55
aI27
aI193
aI509
aI677
aI1861
aI573
aI5341
aI5285
aI6909
aI51781
aI91203
aI139791
asS'd'
I17807
sa(dp35614
S'a'
I74013
sS's'
I18
sS'm'
(lp35615
I1
aI3
aI7
aI11
aI5
aI63
aI87
aI215
aI305
aI235
aI1049
aI1339
aI5301
aI9639
aI29861
aI58415
aI68303
aI76907
asS'd'
I17808
sa(dp35616
S'a'
I74014
sS's'
I18
sS'm'
(lp35617
I1
aI1
aI5
aI11
aI13
aI13
aI67
aI139
aI19
aI577
aI165
aI3067
aI8023
aI10905
aI3159
aI41289
aI118231
aI119673
asS'd'
I17809
sa(dp35618
S'a'
I74020
sS's'
I18
sS'm'
(lp35619
I1
aI1
aI3
aI1
aI15
aI5
aI69
aI119
aI363
aI703
aI461
aI2293
aI3801
aI14217
aI10709
aI9553
aI100651
aI186115
asS'd'
I17810
sa(dp35620
S'a'
I74035
sS's'
I18
sS'm'
(lp35621
I1
aI3
aI7
aI15
aI19
aI21
aI59
aI237
aI227
aI193
aI827
aI619
aI3447
aI13815
aI3467
aI38911
aI41403
aI99627
asS'd'
I17811
sa(dp35622
S'a'
I74038
sS's'
I18
sS'm'
(lp35623
I1
aI1
aI3
aI3
aI27
aI23
aI41
aI199
aI161
aI555
aI1629
aI3187
aI1355
aI5947
aI1157
aI25877
aI110989
aI231285
asS'd'
I17812
sa(dp35624
S'a'
I74069
sS's'
I18
sS'm'
(lp35625
I1
aI1
aI3
aI15
aI3
aI43
aI103
aI29
aI179
aI223
aI375
aI2877
aI1917
aI9367
aI15337
aI15381
aI62833
aI139003
asS'd'
I17813
sa(dp35626
S'a'
I74119
sS's'
I18
sS'm'
(lp35627
I1
aI3
aI3
aI11
aI23
aI33
aI107
aI189
aI511
aI209
aI1519
aI2809
aI6185
aI5921
aI20939
aI63879
aI113687
aI79149
asS'd'
I17814
sa(dp35628
S'a'
I74123
sS's'
I18
sS'm'
(lp35629
I1
aI3
aI5
aI15
aI1
aI63
aI101
aI227
aI419
aI803
aI59
aI2261
aI6905
aI10679
aI5393
aI54447
aI58521
aI59855
asS'd'
I17815
sa(dp35630
S'a'
I74126
sS's'
I18
sS'm'
(lp35631
I1
aI1
aI7
aI13
aI13
aI21
aI121
aI123
aI181
aI371
aI1485
aI3191
aI2627
aI6197
aI11169
aI44927
aI34739
aI10687
asS'd'
I17816
sa(dp35632
S'a'
I74153
sS's'
I18
sS'm'
(lp35633
I1
aI1
aI3
aI3
aI25
aI47
aI13
aI115
aI187
aI967
aI1439
aI1021
aI4413
aI3343
aI31463
aI3729
aI13511
aI162125
asS'd'
I17817
sa(dp35634
S'a'
I74179
sS's'
I18
sS'm'
(lp35635
I1
aI3
aI3
aI15
aI15
aI59
aI29
aI129
aI469
aI171
aI2045
aI2859
aI1097
aI11199
aI12147
aI37465
aI14179
aI197923
asS'd'
I17818
sa(dp35636
S'a'
I74200
sS's'
I18
sS'm'
(lp35637
I1
aI1
aI7
aI5
aI13
aI35
aI41
aI167
aI207
aI123
aI1077
aI3145
aI2803
aI15729
aI767
aI7321
aI84375
aI190855
asS'd'
I17819
sa(dp35638
S'a'
I74210
sS's'
I18
sS'm'
(lp35639
I1
aI3
aI5
aI11
aI3
aI5
aI65
aI3
aI111
aI725
aI143
aI2945
aI4755
aI3407
aI31801
aI15329
aI70311
aI119197
asS'd'
I17820
sa(dp35640
S'a'
I74216
sS's'
I18
sS'm'
(lp35641
I1
aI1
aI5
aI3
aI17
aI19
aI61
aI73
aI261
aI663
aI821
aI3389
aI3883
aI9961
aI17727
aI33113
aI98371
aI247097
asS'd'
I17821
sa(dp35642
S'a'
I74250
sS's'
I18
sS'm'
(lp35643
I1
aI3
aI1
aI9
aI29
aI17
aI43
aI43
aI481
aI1015
aI1249
aI607
aI3495
aI13259
aI29001
aI23083
aI51487
aI81723
asS'd'
I17822
sa(dp35644
S'a'
I74263
sS's'
I18
sS'm'
(lp35645
I1
aI3
aI1
aI3
aI25
aI57
aI5
aI33
aI313
aI21
aI1731
aI3417
aI7033
aI6609
aI31631
aI63231
aI61107
aI10941
asS'd'
I17823
sa(dp35646
S'a'
I74264
sS's'
I18
sS'm'
(lp35647
I1
aI3
aI3
aI11
aI29
aI29
aI117
aI131
aI417
aI789
aI1545
aI1677
aI3213
aI13869
aI5319
aI41387
aI13895
aI252387
asS'd'
I17824
sa(dp35648
S'a'
I74279
sS's'
I18
sS'm'
(lp35649
I1
aI3
aI3
aI13
aI25
aI47
aI55
aI213
aI83
aI345
aI1453
aI159
aI1521
aI14777
aI24177
aI7631
aI81259
aI135411
asS'd'
I17825
sa(dp35650
S'a'
I74285
sS's'
I18
sS'm'
(lp35651
I1
aI1
aI3
aI15
aI11
aI63
aI49
aI69
aI273
aI843
aI1661
aI1157
aI1285
aI12751
aI5
aI54909
aI114375
aI6395
asS'd'
I17826
sa(dp35652
S'a'
I74294
sS's'
I18
sS'm'
(lp35653
I1
aI3
aI1
aI11
aI25
aI1
aI83
aI55
aI161
aI125
aI1547
aI401
aI7639
aI4289
aI7075
aI9971
aI33825
aI135071
asS'd'
I17827
sa(dp35654
S'a'
I74298
sS's'
I18
sS'm'
(lp35655
I1
aI3
aI7
aI9
aI17
aI5
aI79
aI131
aI373
aI1023
aI573
aI2219
aI1789
aI5789
aI5347
aI26455
aI58661
aI206417
asS'd'
I17828
sa(dp35656
S'a'
I74303
sS's'
I18
sS'm'
(lp35657
I1
aI1
aI7
aI5
aI1
aI43
aI83
aI107
aI125
aI289
aI793
aI1731
aI5167
aI8943
aI28397
aI26877
aI53781
aI95899
asS'd'
I17829
sa(dp35658
S'a'
I74306
sS's'
I18
sS'm'
(lp35659
I1
aI3
aI3
aI13
aI21
aI53
aI123
aI103
aI385
aI753
aI1917
aI2075
aI4385
aI5757
aI9221
aI35797
aI86743
aI69069
asS'd'
I17830
sa(dp35660
S'a'
I74366
sS's'
I18
sS'm'
(lp35661
I1
aI3
aI3
aI3
aI5
aI19
aI119
aI221
aI457
aI907
aI359
aI3493
aI2331
aI5685
aI11133
aI29293
aI27051
aI213927
asS'd'
I17831
sa(dp35662
S'a'
I74369
sS's'
I18
sS'm'
(lp35663
I1
aI1
aI3
aI11
aI11
aI31
aI29
aI129
aI429
aI601
aI1217
aI3653
aI5935
aI14823
aI21161
aI33423
aI98391
aI214703
asS'd'
I17832
sa(dp35664
S'a'
I74403
sS's'
I18
sS'm'
(lp35665
I1
aI1
aI3
aI15
aI31
aI49
aI109
aI139
aI349
aI13
aI205
aI2483
aI8083
aI8391
aI4789
aI30355
aI12165
aI195263
asS'd'
I17833
sa(dp35666
S'a'
I74406
sS's'
I18
sS'm'
(lp35667
I1
aI1
aI1
aI9
aI19
aI63
aI15
aI167
aI45
aI185
aI811
aI529
aI6811
aI13441
aI27195
aI59047
aI106675
aI167125
asS'd'
I17834
sa(dp35668
S'a'
I74442
sS's'
I18
sS'm'
(lp35669
I1
aI1
aI7
aI1
aI19
aI19
aI39
aI1
aI349
aI467
aI551
aI4081
aI4743
aI11627
aI607
aI22005
aI60893
aI49101
asS'd'
I17835
sa(dp35670
S'a'
I74450
sS's'
I18
sS'm'
(lp35671
I1
aI3
aI7
aI9
aI9
aI25
aI105
aI119
aI113
aI825
aI1429
aI2019
aI5209
aI8491
aI6017
aI47783
aI88455
aI119083
asS'd'
I17836
sa(dp35672
S'a'
I74452
sS's'
I18
sS'm'
(lp35673
I1
aI3
aI3
aI13
aI21
aI39
aI107
aI119
aI321
aI251
aI563
aI311
aI4441
aI6491
aI3157
aI65479
aI107349
aI211621
asS'd'
I17837
sa(dp35674
S'a'
I74459
sS's'
I18
sS'm'
(lp35675
I1
aI1
aI7
aI13
aI9
aI3
aI17
aI117
aI327
aI459
aI489
aI7
aI6883
aI10047
aI31935
aI28069
aI37903
aI188281
asS'd'
I17838
sa(dp35676
S'a'
I74461
sS's'
I18
sS'm'
(lp35677
I1
aI1
aI7
aI1
aI15
aI53
aI47
aI127
aI483
aI595
aI811
aI1143
aI4543
aI10043
aI30349
aI24409
aI91947
aI240165
asS'd'
I17839
sa(dp35678
S'a'
I74477
sS's'
I18
sS'm'
(lp35679
I1
aI1
aI1
aI13
aI25
aI43
aI109
aI161
aI147
aI1009
aI1071
aI1533
aI2781
aI13439
aI20507
aI41387
aI26943
aI84675
asS'd'
I17840
sa(dp35680
S'a'
I74478
sS's'
I18
sS'm'
(lp35681
I1
aI3
aI5
aI13
aI19
aI21
aI23
aI167
aI135
aI257
aI587
aI2691
aI5877
aI3047
aI11745
aI24895
aI114799
aI48003
asS'd'
I17841
sa(dp35682
S'a'
I74489
sS's'
I18
sS'm'
(lp35683
I1
aI3
aI1
aI1
aI9
aI57
aI33
aI51
aI137
aI109
aI137
aI195
aI1233
aI11139
aI16833
aI27545
aI35877
aI126627
asS'd'
I17842
sa(dp35684
S'a'
I74504
sS's'
I18
sS'm'
(lp35685
I1
aI1
aI1
aI15
aI5
aI37
aI7
aI19
aI363
aI411
aI1193
aI767
aI6209
aI9115
aI14699
aI55515
aI46023
aI90693
asS'd'
I17843
sa(dp35686
S'a'
I74528
sS's'
I18
sS'm'
(lp35687
I1
aI3
aI5
aI3
aI19
aI55
aI85
aI117
aI391
aI757
aI861
aI3537
aI6507
aI6993
aI19589
aI6843
aI33557
aI64683
asS'd'
I17844
sa(dp35688
S'a'
I74533
sS's'
I18
sS'm'
(lp35689
I1
aI1
aI7
aI5
aI5
aI61
aI99
aI9
aI311
aI595
aI807
aI429
aI63
aI12359
aI28289
aI709
aI129911
aI143745
asS'd'
I17845
sa(dp35690
S'a'
I74538
sS's'
I18
sS'm'
(lp35691
I1
aI1
aI1
aI7
aI17
aI9
aI9
aI117
aI19
aI985
aI657
aI2803
aI2699
aI829
aI31069
aI13277
aI106769
aI109231
asS'd'
I17846
sa(dp35692
S'a'
I74551
sS's'
I18
sS'm'
(lp35693
I1
aI1
aI5
aI13
aI25
aI19
aI63
aI217
aI419
aI221
aI1921
aI215
aI2631
aI4659
aI29855
aI46549
aI62257
aI260113
asS'd'
I17847
sa(dp35694
S'a'
I74580
sS's'
I18
sS'm'
(lp35695
I1
aI3
aI3
aI9
aI13
aI45
aI63
aI129
aI147
aI489
aI879
aI3025
aI6777
aI1119
aI20963
aI30553
aI20863
aI169837
asS'd'
I17848
sa(dp35696
S'a'
I74589
sS's'
I18
sS'm'
(lp35697
I1
aI3
aI5
aI9
aI25
aI61
aI79
aI51
aI495
aI583
aI1519
aI1501
aI123
aI13871
aI32239
aI957
aI31921
aI255561
asS'd'
I17849
sa(dp35698
S'a'
I74590
sS's'
I18
sS'm'
(lp35699
I1
aI3
aI7
aI3
aI31
aI61
aI89
aI65
aI305
aI429
aI785
aI3871
aI7711
aI2745
aI24131
aI43055
aI51167
aI87743
asS'd'
I17850
sa(dp35700
S'a'
I74593
sS's'
I18
sS'm'
(lp35701
I1
aI3
aI1
aI11
aI19
aI23
aI89
aI117
aI185
aI121
aI109
aI1327
aI6553
aI14367
aI16069
aI28657
aI81751
aI10185
asS'd'
I17851
sa(dp35702
S'a'
I74608
sS's'
I18
sS'm'
(lp35703
I1
aI3
aI5
aI5
aI17
aI33
aI115
aI61
aI101
aI367
aI1465
aI3899
aI6601
aI4483
aI2447
aI49575
aI129987
aI11703
asS'd'
I17852
sa(dp35704
S'a'
I74611
sS's'
I18
sS'm'
(lp35705
I1
aI1
aI3
aI13
aI13
aI59
aI79
aI83
aI253
aI171
aI53
aI2467
aI5005
aI1045
aI943
aI62419
aI98563
aI78935
asS'd'
I17853
sa(dp35706
S'a'
I74617
sS's'
I18
sS'm'
(lp35707
I1
aI1
aI7
aI15
aI31
aI17
aI77
aI73
aI249
aI247
aI119
aI1655
aI7079
aI14593
aI105
aI55767
aI130401
aI74189
asS'd'
I17854
sa(dp35708
S'a'
I74623
sS's'
I18
sS'm'
(lp35709
I1
aI3
aI5
aI11
aI19
aI29
aI79
aI251
aI75
aI949
aI527
aI2779
aI5839
aI11451
aI24125
aI45991
aI127437
aI86541
asS'd'
I17855
sa(dp35710
S'a'
I74648
sS's'
I18
sS'm'
(lp35711
I1
aI1
aI5
aI7
aI1
aI29
aI127
aI27
aI477
aI807
aI829
aI1569
aI205
aI13319
aI16149
aI26003
aI38985
aI188587
asS'd'
I17856
sa(dp35712
S'a'
I74704
sS's'
I18
sS'm'
(lp35713
I1
aI1
aI1
aI3
aI19
aI29
aI39
aI71
aI17
aI51
aI1169
aI467
aI7505
aI1867
aI4469
aI32161
aI43031
aI31675
asS'd'
I17857
sa(dp35714
S'a'
I74735
sS's'
I18
sS'm'
(lp35715
I1
aI1
aI1
aI15
aI27
aI45
aI67
aI107
aI127
aI469
aI1955
aI1933
aI2379
aI8513
aI32071
aI35043
aI126537
aI23303
asS'd'
I17858
sa(dp35716
S'a'
I74737
sS's'
I18
sS'm'
(lp35717
I1
aI3
aI7
aI9
aI23
aI41
aI23
aI197
aI97
aI213
aI17
aI1751
aI5467
aI6179
aI29291
aI33397
aI42131
aI151093
asS'd'
I17859
sa(dp35718
S'a'
I74738
sS's'
I18
sS'm'
(lp35719
I1
aI3
aI7
aI15
aI27
aI3
aI39
aI87
aI365
aI77
aI487
aI293
aI6405
aI2239
aI30455
aI44723
aI12399
aI100013
asS'd'
I17860
sa(dp35720
S'a'
I74779
sS's'
I18
sS'm'
(lp35721
I1
aI3
aI5
aI9
aI29
aI21
aI55
aI183
aI343
aI263
aI1643
aI2027
aI2255
aI6259
aI18277
aI64661
aI39391
aI255839
asS'd'
I17861
sa(dp35722
S'a'
I74781
sS's'
I18
sS'm'
(lp35723
I1
aI1
aI7
aI11
aI21
aI55
aI11
aI139
aI261
aI11
aI1721
aI3779
aI85
aI8203
aI12089
aI50579
aI128341
aI119043
asS'd'
I17862
sa(dp35724
S'a'
I74809
sS's'
I18
sS'm'
(lp35725
I1
aI3
aI5
aI13
aI27
aI53
aI109
aI1
aI313
aI661
aI431
aI1543
aI1571
aI7337
aI18857
aI49951
aI7881
aI228161
asS'd'
I17863
sa(dp35726
S'a'
I74812
sS's'
I18
sS'm'
(lp35727
I1
aI1
aI7
aI5
aI5
aI31
aI27
aI149
aI239
aI199
aI1011
aI1979
aI5297
aI14609
aI26971
aI65531
aI64215
aI115109
asS'd'
I17864
sa(dp35728
S'a'
I74827
sS's'
I18
sS'm'
(lp35729
I1
aI1
aI5
aI13
aI15
aI11
aI63
aI225
aI165
aI405
aI1367
aI2291
aI5171
aI12419
aI19561
aI37719
aI621
aI137607
asS'd'
I17865
sa(dp35730
S'a'
I74868
sS's'
I18
sS'm'
(lp35731
I1
aI3
aI1
aI15
aI25
aI1
aI125
aI243
aI97
aI455
aI1977
aI3333
aI801
aI1343
aI993
aI18453
aI19285
aI71547
asS'd'
I17866
sa(dp35732
S'a'
I74871
sS's'
I18
sS'm'
(lp35733
I1
aI1
aI5
aI11
aI9
aI51
aI109
aI135
aI73
aI147
aI649
aI4071
aI7425
aI3093
aI26417
aI51139
aI1523
aI142225
asS'd'
I17867
sa(dp35734
S'a'
I74905
sS's'
I18
sS'm'
(lp35735
I1
aI3
aI7
aI11
aI1
aI39
aI109
aI119
aI337
aI715
aI1087
aI4005
aI1393
aI6397
aI31135
aI38935
aI106255
aI60723
asS'd'
I17868
sa(dp35736
S'a'
I74930
sS's'
I18
sS'm'
(lp35737
I1
aI1
aI1
aI9
aI9
aI11
aI45
aI11
aI171
aI671
aI965
aI109
aI2261
aI13775
aI8539
aI63669
aI10507
aI249113
asS'd'
I17869
sa(dp35738
S'a'
I74962
sS's'
I18
sS'm'
(lp35739
I1
aI3
aI1
aI15
aI21
aI33
aI127
aI173
aI419
aI31
aI299
aI857
aI4915
aI11331
aI29385
aI47375
aI111891
aI14505
asS'd'
I17870
sa(dp35740
S'a'
I74990
sS's'
I18
sS'm'
(lp35741
I1
aI3
aI5
aI13
aI7
aI43
aI85
aI183
aI377
aI275
aI803
aI1755
aI8005
aI15327
aI31043
aI18851
aI122581
aI229731
asS'd'
I17871
sa(dp35742
S'a'
I74995
sS's'
I18
sS'm'
(lp35743
I1
aI3
aI5
aI1
aI27
aI9
aI41
aI73
aI283
aI475
aI671
aI747
aI1419
aI15209
aI25465
aI60061
aI91417
aI103203
asS'd'
I17872
sa(dp35744
S'a'
I75007
sS's'
I18
sS'm'
(lp35745
I1
aI3
aI1
aI15
aI15
aI43
aI13
aI45
aI217
aI519
aI363
aI3265
aI6213
aI13045
aI3709
aI22119
aI79733
aI224195
asS'd'
I17873
sa(dp35746
S'a'
I75012
sS's'
I18
sS'm'
(lp35747
I1
aI3
aI1
aI15
aI15
aI59
aI95
aI71
aI171
aI769
aI1395
aI2673
aI4523
aI749
aI13411
aI60431
aI124651
aI11475
asS'd'
I17874
sa(dp35748
S'a'
I75030
sS's'
I18
sS'm'
(lp35749
I1
aI1
aI7
aI3
aI1
aI35
aI13
aI239
aI101
aI355
aI1201
aI3665
aI5403
aI11413
aI11983
aI52469
aI63621
aI155819
asS'd'
I17875
sa(dp35750
S'a'
I75033
sS's'
I18
sS'm'
(lp35751
I1
aI1
aI1
aI7
aI31
aI59
aI87
aI25
aI511
aI483
aI569
aI3337
aI4027
aI8347
aI3031
aI24351
aI57963
aI79425
asS'd'
I17876
sa(dp35752
S'a'
I75039
sS's'
I18
sS'm'
(lp35753
I1
aI3
aI1
aI3
aI11
aI17
aI29
aI249
aI61
aI923
aI585
aI2107
aI2727
aI8589
aI22809
aI3
aI17937
aI163267
asS'd'
I17877
sa(dp35754
S'a'
I75055
sS's'
I18
sS'm'
(lp35755
I1
aI3
aI5
aI11
aI27
aI3
aI73
aI187
aI19
aI975
aI257
aI2361
aI935
aI9071
aI29991
aI13619
aI92169
aI101031
asS'd'
I17878
sa(dp35756
S'a'
I75078
sS's'
I18
sS'm'
(lp35757
I1
aI1
aI5
aI13
aI17
aI53
aI105
aI157
aI343
aI673
aI237
aI3231
aI7311
aI1593
aI18521
aI57889
aI79805
aI97847
asS'd'
I17879
sa(dp35758
S'a'
I75120
sS's'
I18
sS'm'
(lp35759
I1
aI3
aI1
aI3
aI31
aI55
aI63
aI167
aI489
aI167
aI121
aI3333
aI2475
aI1545
aI13291
aI921
aI101757
aI62147
asS'd'
I17880
sa(dp35760
S'a'
I75130
sS's'
I18
sS'm'
(lp35761
I1
aI3
aI3
aI15
aI13
aI17
aI9
aI209
aI339
aI567
aI2011
aI1737
aI1455
aI9289
aI6105
aI49733
aI74237
aI93195
asS'd'
I17881
sa(dp35762
S'a'
I75159
sS's'
I18
sS'm'
(lp35763
I1
aI1
aI7
aI11
aI3
aI13
aI77
aI115
aI305
aI327
aI1005
aI3381
aI4269
aI4835
aI27221
aI16301
aI75173
aI244603
asS'd'
I17882
sa(dp35764
S'a'
I75163
sS's'
I18
sS'm'
(lp35765
I1
aI3
aI5
aI7
aI7
aI31
aI47
aI75
aI499
aI41
aI281
aI167
aI3525
aI8649
aI23623
aI4987
aI2057
aI204083
asS'd'
I17883
sa(dp35766
S'a'
I75170
sS's'
I18
sS'm'
(lp35767
I1
aI3
aI3
aI5
aI9
aI5
aI35
aI53
aI269
aI437
aI1035
aI1675
aI4567
aI13291
aI19787
aI28937
aI108915
aI62545
asS'd'
I17884
sa(dp35768
S'a'
I75193
sS's'
I18
sS'm'
(lp35769
I1
aI3
aI1
aI5
aI15
aI59
aI57
aI181
aI321
aI1
aI791
aI2149
aI591
aI6691
aI8759
aI62861
aI10815
aI257331
asS'd'
I17885
sa(dp35770
S'a'
I75228
sS's'
I18
sS'm'
(lp35771
I1
aI1
aI7
aI1
aI21
aI25
aI93
aI39
aI429
aI455
aI669
aI1725
aI7087
aI11805
aI22405
aI13083
aI88411
aI225967
asS'd'
I17886
sa(dp35772
S'a'
I75244
sS's'
I18
sS'm'
(lp35773
I1
aI1
aI5
aI15
aI5
aI45
aI15
aI1
aI55
aI281
aI2027
aI97
aI2639
aI57
aI23717
aI21669
aI92181
aI32731
asS'd'
I17887
sa(dp35774
S'a'
I75249
sS's'
I18
sS'm'
(lp35775
I1
aI3
aI7
aI3
aI7
aI3
aI67
aI201
aI445
aI577
aI1011
aI793
aI7763
aI10823
aI30309
aI41565
aI37263
aI218909
asS'd'
I17888
sa(dp35776
S'a'
I75328
sS's'
I18
sS'm'
(lp35777
I1
aI1
aI7
aI3
aI17
aI53
aI51
aI119
aI399
aI903
aI1785
aI1053
aI4315
aI2967
aI17579
aI64185
aI55005
aI12969
asS'd'
I17889
sa(dp35778
S'a'
I75357
sS's'
I18
sS'm'
(lp35779
I1
aI3
aI3
aI13
aI21
aI63
aI13
aI1
aI427
aI39
aI71
aI1811
aI1237
aI1623
aI11401
aI14371
aI44355
aI93089
asS'd'
I17890
sa(dp35780
S'a'
I75373
sS's'
I18
sS'm'
(lp35781
I1
aI1
aI5
aI1
aI17
aI39
aI39
aI105
aI187
aI691
aI251
aI3957
aI931
aI12149
aI18299
aI48819
aI23061
aI49179
asS'd'
I17891
sa(dp35782
S'a'
I75392
sS's'
I18
sS'm'
(lp35783
I1
aI1
aI5
aI11
aI31
aI3
aI23
aI211
aI101
aI763
aI237
aI3635
aI417
aI4935
aI14997
aI3859
aI22343
aI153541
asS'd'
I17892
sa(dp35784
S'a'
I75428
sS's'
I18
sS'm'
(lp35785
I1
aI3
aI7
aI5
aI21
aI37
aI59
aI137
aI13
aI179
aI527
aI895
aI3451
aI1743
aI3149
aI10665
aI119427
aI259343
asS'd'
I17893
sa(dp35786
S'a'
I75438
sS's'
I18
sS'm'
(lp35787
I1
aI3
aI5
aI3
aI7
aI37
aI103
aI173
aI453
aI327
aI131
aI2453
aI7795
aI12585
aI13947
aI59161
aI41845
aI29527
asS'd'
I17894
sa(dp35788
S'a'
I75469
sS's'
I18
sS'm'
(lp35789
I1
aI3
aI3
aI1
aI17
aI49
aI57
aI251
aI295
aI279
aI1545
aI3963
aI589
aI9211
aI32371
aI14963
aI116927
aI197321
asS'd'
I17895
sa(dp35790
S'a'
I75487
sS's'
I18
sS'm'
(lp35791
I1
aI1
aI3
aI7
aI17
aI59
aI37
aI115
aI315
aI591
aI481
aI767
aI4611
aI14741
aI6949
aI19507
aI6567
aI143371
asS'd'
I17896
sa(dp35792
S'a'
I75494
sS's'
I18
sS'm'
(lp35793
I1
aI1
aI5
aI3
aI19
aI53
aI121
aI229
aI355
aI909
aI339
aI1645
aI2747
aI7045
aI9085
aI5799
aI50997
aI17981
asS'd'
I17897
sa(dp35794
S'a'
I75515
sS's'
I18
sS'm'
(lp35795
I1
aI3
aI3
aI9
aI3
aI1
aI109
aI7
aI15
aI177
aI789
aI3911
aI6427
aI8453
aI22583
aI12039
aI124587
aI123887
asS'd'
I17898
sa(dp35796
S'a'
I75526
sS's'
I18
sS'm'
(lp35797
I1
aI1
aI5
aI1
aI7
aI23
aI15
aI193
aI109
aI685
aI1147
aI3921
aI2329
aI15153
aI25045
aI28389
aI34759
aI256611
asS'd'
I17899
sa(dp35798
S'a'
I75532
sS's'
I18
sS'm'
(lp35799
I1
aI1
aI3
aI3
aI23
aI27
aI55
aI43
aI485
aI541
aI1617
aI3761
aI1051
aI7525
aI19941
aI52699
aI35421
aI162939
asS'd'
I17900
sa(dp35800
S'a'
I75543
sS's'
I18
sS'm'
(lp35801
I1
aI3
aI7
aI1
aI25
aI9
aI113
aI251
aI477
aI1005
aI9
aI3321
aI5817
aI965
aI18523
aI29407
aI53353
aI205575
asS'd'
I17901
sa(dp35802
S'a'
I75549
sS's'
I18
sS'm'
(lp35803
I1
aI1
aI7
aI11
aI9
aI31
aI111
aI175
aI227
aI33
aI1745
aI1141
aI1547
aI2113
aI8785
aI40273
aI100301
aI190749
asS'd'
I17902
sa(dp35804
S'a'
I75559
sS's'
I18
sS'm'
(lp35805
I1
aI1
aI5
aI11
aI19
aI49
aI45
aI197
aI457
aI223
aI91
aI2769
aI6331
aI1161
aI6609
aI61905
aI42257
aI152117
asS'd'
I17903
sa(dp35806
S'a'
I75563
sS's'
I18
sS'm'
(lp35807
I1
aI1
aI7
aI11
aI1
aI17
aI37
aI153
aI431
aI933
aI269
aI1529
aI1297
aI15567
aI149
aI41701
aI59867
aI93631
asS'd'
I17904
sa(dp35808
S'a'
I75568
sS's'
I18
sS'm'
(lp35809
I1
aI1
aI5
aI9
aI25
aI33
aI83
aI127
aI305
aI667
aI343
aI185
aI3527
aI13079
aI10567
aI35753
aI72191
aI214091
asS'd'
I17905
sa(dp35810
S'a'
I75598
sS's'
I18
sS'm'
(lp35811
I1
aI3
aI7
aI1
aI1
aI7
aI75
aI241
aI185
aI81
aI2043
aI3081
aI3563
aI385
aI3055
aI59421
aI27081
aI32521
asS'd'
I17906
sa(dp35812
S'a'
I75612
sS's'
I18
sS'm'
(lp35813
I1
aI1
aI3
aI5
aI31
aI1
aI101
aI21
aI69
aI979
aI917
aI695
aI5601
aI12251
aI15031
aI18715
aI116985
aI53071
asS'd'
I17907
sa(dp35814
S'a'
I75622
sS's'
I18
sS'm'
(lp35815
I1
aI1
aI3
aI9
aI23
aI57
aI91
aI127
aI327
aI979
aI721
aI3855
aI1131
aI997
aI32227
aI33843
aI128299
aI15239
asS'd'
I17908
sa(dp35816
S'a'
I75640
sS's'
I18
sS'm'
(lp35817
I1
aI3
aI7
aI13
aI23
aI1
aI87
aI105
aI259
aI939
aI1935
aI1983
aI6619
aI1611
aI31901
aI14745
aI96641
aI211945
asS'd'
I17909
sa(dp35818
S'a'
I75683
sS's'
I18
sS'm'
(lp35819
I1
aI1
aI3
aI5
aI25
aI17
aI39
aI95
aI137
aI971
aI377
aI2493
aI981
aI329
aI25845
aI44513
aI100561
aI57985
asS'd'
I17910
sa(dp35820
S'a'
I75689
sS's'
I18
sS'm'
(lp35821
I1
aI1
aI3
aI9
aI27
aI37
aI69
aI103
aI167
aI131
aI487
aI2935
aI7099
aI15375
aI4825
aI12209
aI117165
aI84909
asS'd'
I17911
sa(dp35822
S'a'
I75690
sS's'
I18
sS'm'
(lp35823
I1
aI3
aI1
aI15
aI27
aI19
aI115
aI239
aI247
aI243
aI83
aI1535
aI8095
aI3953
aI25721
aI62983
aI89045
aI16783
asS'd'
I17912
sa(dp35824
S'a'
I75692
sS's'
I18
sS'm'
(lp35825
I1
aI1
aI5
aI1
aI19
aI13
aI125
aI39
aI439
aI411
aI171
aI155
aI5117
aI15137
aI19851
aI251
aI37921
aI97209
asS'd'
I17913
sa(dp35826
S'a'
I75752
sS's'
I18
sS'm'
(lp35827
I1
aI1
aI1
aI5
aI13
aI41
aI17
aI97
aI215
aI323
aI1333
aI775
aI1155
aI15269
aI19943
aI48489
aI71741
aI202501
asS'd'
I17914
sa(dp35828
S'a'
I75766
sS's'
I18
sS'm'
(lp35829
I1
aI3
aI5
aI3
aI1
aI31
aI55
aI125
aI69
aI437
aI1649
aI2791
aI8027
aI15509
aI31575
aI8491
aI106953
aI155215
asS'd'
I17915
sa(dp35830
S'a'
I75769
sS's'
I18
sS'm'
(lp35831
I1
aI3
aI3
aI13
aI19
aI45
aI39
aI63
aI227
aI67
aI2021
aI1243
aI6525
aI7211
aI6275
aI39719
aI74513
aI6713
asS'd'
I17916
sa(dp35832
S'a'
I75791
sS's'
I18
sS'm'
(lp35833
I1
aI1
aI5
aI11
aI13
aI45
aI15
aI101
aI171
aI613
aI1561
aI2939
aI3849
aI2917
aI29765
aI2027
aI53617
aI59939
asS'd'
I17917
sa(dp35834
S'a'
I75794
sS's'
I18
sS'm'
(lp35835
I1
aI1
aI5
aI9
aI27
aI21
aI119
aI19
aI441
aI759
aI703
aI2985
aI3007
aI2087
aI5207
aI64403
aI20273
aI66181
asS'd'
I17918
sa(dp35836
S'a'
I75806
sS's'
I18
sS'm'
(lp35837
I1
aI1
aI7
aI13
aI21
aI3
aI49
aI3
aI485
aI883
aI1863
aI1925
aI877
aI10009
aI24191
aI58639
aI107755
aI106539
asS'd'
I17919
sa(dp35838
S'a'
I75809
sS's'
I18
sS'm'
(lp35839
I1
aI1
aI1
aI5
aI27
aI37
aI23
aI185
aI281
aI533
aI437
aI555
aI8151
aI6489
aI22343
aI4573
aI91577
aI167919
asS'd'
I17920
sa(dp35840
S'a'
I75819
sS's'
I18
sS'm'
(lp35841
I1
aI1
aI7
aI15
aI3
aI61
aI103
aI221
aI223
aI703
aI133
aI2923
aI1027
aI14643
aI26413
aI16523
aI107223
aI97185
asS'd'
I17921
sa(dp35842
S'a'
I75834
sS's'
I18
sS'm'
(lp35843
I1
aI3
aI1
aI11
aI13
aI15
aI1
aI203
aI363
aI675
aI511
aI3225
aI1163
aI741
aI16063
aI8097
aI95905
aI148465
asS'd'
I17922
sa(dp35844
S'a'
I75895
sS's'
I18
sS'm'
(lp35845
I1
aI3
aI1
aI7
aI11
aI11
aI11
aI243
aI371
aI129
aI209
aI3533
aI1279
aI12181
aI31973
aI29165
aI122089
aI115117
asS'd'
I17923
sa(dp35846
S'a'
I75908
sS's'
I18
sS'm'
(lp35847
I1
aI3
aI1
aI5
aI25
aI17
aI31
aI45
aI215
aI809
aI1443
aI3245
aI1005
aI2903
aI20783
aI23041
aI96577
aI192063
asS'd'
I17924
sa(dp35848
S'a'
I75951
sS's'
I18
sS'm'
(lp35849
I1
aI3
aI3
aI3
aI19
aI17
aI101
aI219
aI91
aI805
aI189
aI761
aI4771
aI11629
aI7285
aI21631
aI21691
aI47421
asS'd'
I17925
sa(dp35850
S'a'
I75960
sS's'
I18
sS'm'
(lp35851
I1
aI3
aI7
aI1
aI31
aI11
aI71
aI149
aI303
aI793
aI35
aI3109
aI2769
aI11593
aI31839
aI2053
aI4541
aI202997
asS'd'
I17926
sa(dp35852
S'a'
I75965
sS's'
I18
sS'm'
(lp35853
I1
aI1
aI1
aI3
aI11
aI19
aI113
aI249
aI141
aI659
aI1117
aI2145
aI2617
aI1075
aI25347
aI12913
aI27457
aI222095
asS'd'
I17927
sa(dp35854
S'a'
I75974
sS's'
I18
sS'm'
(lp35855
I1
aI3
aI1
aI3
aI5
aI23
aI41
aI57
aI193
aI815
aI1293
aI1109
aI7597
aI999
aI10773
aI41065
aI18555
aI35617
asS'd'
I17928
sa(dp35856
S'a'
I75978
sS's'
I18
sS'm'
(lp35857
I1
aI1
aI3
aI1
aI31
aI11
aI127
aI99
aI163
aI293
aI299
aI3415
aI3761
aI8781
aI5327
aI47631
aI56411
aI242787
asS'd'
I17929
sa(dp35858
S'a'
I75998
sS's'
I18
sS'm'
(lp35859
I1
aI3
aI7
aI13
aI3
aI41
aI23
aI169
aI419
aI725
aI1419
aI2643
aI5265
aI77
aI24077
aI18639
aI78665
aI205303
asS'd'
I17930
sa(dp35860
S'a'
I76001
sS's'
I18
sS'm'
(lp35861
I1
aI1
aI5
aI1
aI31
aI39
aI39
aI205
aI413
aI393
aI1713
aI309
aI707
aI4153
aI10461
aI16053
aI26963
aI253993
asS'd'
I17931
sa(dp35862
S'a'
I76007
sS's'
I18
sS'm'
(lp35863
I1
aI3
aI7
aI5
aI23
aI37
aI125
aI87
aI199
aI631
aI1935
aI551
aI7047
aI4585
aI21257
aI42345
aI39365
aI249393
asS'd'
I17932
sa(dp35864
S'a'
I76016
sS's'
I18
sS'm'
(lp35865
I1
aI3
aI5
aI13
aI17
aI55
aI29
aI209
aI151
aI465
aI155
aI363
aI3097
aI4093
aI9869
aI23297
aI33973
aI115543
asS'd'
I17933
sa(dp35866
S'a'
I76091
sS's'
I18
sS'm'
(lp35867
I1
aI1
aI1
aI13
aI23
aI59
aI83
aI71
aI145
aI717
aI127
aI1299
aI1701
aI10885
aI5343
aI40793
aI87819
aI66621
asS'd'
I17934
sa(dp35868
S'a'
I76119
sS's'
I18
sS'm'
(lp35869
I1
aI3
aI3
aI9
aI19
aI37
aI23
aI11
aI269
aI603
aI871
aI851
aI837
aI15303
aI7595
aI56481
aI57819
aI185065
asS'd'
I17935
sa(dp35870
S'a'
I76135
sS's'
I18
sS'm'
(lp35871
I1
aI3
aI1
aI1
aI3
aI15
aI11
aI249
aI413
aI723
aI1403
aI3233
aI2747
aI10335
aI7127
aI63285
aI29237
aI191953
asS'd'
I17936
sa(dp35872
S'a'
I76170
sS's'
I18
sS'm'
(lp35873
I1
aI3
aI1
aI1
aI11
aI31
aI67
aI139
aI51
aI413
aI521
aI969
aI171
aI5943
aI31613
aI16477
aI85771
aI202139
asS'd'
I17937
sa(dp35874
S'a'
I76189
sS's'
I18
sS'm'
(lp35875
I1
aI3
aI7
aI5
aI5
aI21
aI109
aI25
aI463
aI873
aI493
aI2673
aI6409
aI11199
aI17195
aI40623
aI76821
aI72509
asS'd'
I17938
sa(dp35876
S'a'
I76214
sS's'
I18
sS'm'
(lp35877
I1
aI1
aI7
aI7
aI11
aI1
aI95
aI43
aI243
aI67
aI1289
aI3219
aI2255
aI4957
aI17561
aI40499
aI48537
aI108809
asS'd'
I17939
sa(dp35878
S'a'
I76225
sS's'
I18
sS'm'
(lp35879
I1
aI1
aI7
aI15
aI3
aI39
aI45
aI75
aI43
aI821
aI533
aI4043
aI1503
aI83
aI26937
aI56327
aI114149
aI156845
asS'd'
I17940
sa(dp35880
S'a'
I76235
sS's'
I18
sS'm'
(lp35881
I1
aI3
aI1
aI1
aI31
aI21
aI59
aI1
aI77
aI147
aI137
aI1827
aI4123
aI2791
aI27859
aI57921
aI40569
aI134753
asS'd'
I17941
sa(dp35882
S'a'
I76259
sS's'
I18
sS'm'
(lp35883
I1
aI1
aI5
aI13
aI31
aI41
aI111
aI11
aI181
aI963
aI459
aI2771
aI6123
aI4035
aI1627
aI2047
aI109537
aI33653
asS'd'
I17942
sa(dp35884
S'a'
I76261
sS's'
I18
sS'm'
(lp35885
I1
aI3
aI5
aI7
aI31
aI57
aI17
aI21
aI5
aI761
aI1833
aI1279
aI1239
aI10089
aI22531
aI32547
aI82699
aI28389
asS'd'
I17943
sa(dp35886
S'a'
I76262
sS's'
I18
sS'm'
(lp35887
I1
aI3
aI5
aI13
aI11
aI39
aI11
aI61
aI299
aI753
aI1067
aI1347
aI5189
aI12859
aI681
aI46309
aI31873
aI90333
asS'd'
I17944
sa(dp35888
S'a'
I76266
sS's'
I18
sS'm'
(lp35889
I1
aI3
aI1
aI5
aI13
aI27
aI119
aI205
aI377
aI457
aI817
aI3017
aI279
aI1859
aI30241
aI52089
aI61445
aI176203
asS'd'
I17945
sa(dp35890
S'a'
I76286
sS's'
I18
sS'm'
(lp35891
I1
aI1
aI3
aI5
aI11
aI35
aI17
aI163
aI27
aI1001
aI417
aI2899
aI1959
aI5513
aI1441
aI19743
aI67147
aI236591
asS'd'
I17946
sa(dp35892
S'a'
I76289
sS's'
I18
sS'm'
(lp35893
I1
aI3
aI5
aI13
aI15
aI39
aI53
aI179
aI447
aI675
aI933
aI1261
aI4415
aI9845
aI28459
aI33497
aI107375
aI156855
asS'd'
I17947
sa(dp35894
S'a'
I76310
sS's'
I18
sS'm'
(lp35895
I1
aI3
aI3
aI13
aI27
aI31
aI11
aI191
aI413
aI1011
aI2035
aI3965
aI2071
aI5429
aI16247
aI7439
aI15079
aI225041
asS'd'
I17948
sa(dp35896
S'a'
I76319
sS's'
I18
sS'm'
(lp35897
I1
aI3
aI3
aI11
aI7
aI23
aI87
aI215
aI241
aI687
aI1351
aI2399
aI4677
aI12967
aI22957
aI10443
aI116701
aI155477
asS'd'
I17949
sa(dp35898
S'a'
I76364
sS's'
I18
sS'm'
(lp35899
I1
aI3
aI7
aI5
aI25
aI55
aI5
aI197
aI359
aI879
aI619
aI1969
aI1513
aI12743
aI10953
aI28343
aI63685
aI39115
asS'd'
I17950
sa(dp35900
S'a'
I76375
sS's'
I18
sS'm'
(lp35901
I1
aI1
aI3
aI1
aI19
aI15
aI63
aI7
aI305
aI343
aI1333
aI3845
aI377
aI14031
aI28383
aI4271
aI60063
aI11827
asS'd'
I17951
sa(dp35902
S'a'
I76392
sS's'
I18
sS'm'
(lp35903
I1
aI3
aI5
aI15
aI25
aI21
aI115
aI101
aI171
aI735
aI787
aI3143
aI593
aI8793
aI4121
aI15471
aI53491
aI20617
asS'd'
I17952
sa(dp35904
S'a'
I76422
sS's'
I18
sS'm'
(lp35905
I1
aI1
aI5
aI15
aI15
aI51
aI103
aI17
aI433
aI611
aI1351
aI1729
aI6147
aI11623
aI3
aI6319
aI6133
aI19029
asS'd'
I17953
sa(dp35906
S'a'
I76450
sS's'
I18
sS'm'
(lp35907
I1
aI3
aI1
aI13
aI29
aI15
aI115
aI97
aI505
aI985
aI745
aI745
aI1459
aI7193
aI1247
aI58901
aI114255
aI212849
asS'd'
I17954
sa(dp35908
S'a'
I76496
sS's'
I18
sS'm'
(lp35909
I1
aI1
aI7
aI1
aI1
aI53
aI99
aI35
aI377
aI723
aI1751
aI2625
aI5113
aI13295
aI20133
aI26831
aI41657
aI51717
asS'd'
I17955
sa(dp35910
S'a'
I76521
sS's'
I18
sS'm'
(lp35911
I1
aI3
aI5
aI15
aI15
aI39
aI17
aI227
aI351
aI435
aI49
aI203
aI6959
aI11673
aI15755
aI29733
aI51445
aI64619
asS'd'
I17956
sa(dp35912
S'a'
I76522
sS's'
I18
sS'm'
(lp35913
I1
aI1
aI1
aI3
aI25
aI51
aI57
aI137
aI415
aI49
aI355
aI2149
aI7607
aI10781
aI30363
aI43889
aI55543
aI36637
asS'd'
I17957
sa(dp35914
S'a'
I76527
sS's'
I18
sS'm'
(lp35915
I1
aI1
aI7
aI5
aI5
aI15
aI73
aI189
aI153
aI949
aI527
aI587
aI513
aI12891
aI16765
aI41477
aI75569
aI80747
asS'd'
I17958
sa(dp35916
S'a'
I76536
sS's'
I18
sS'm'
(lp35917
I1
aI1
aI5
aI5
aI9
aI5
aI3
aI225
aI115
aI125
aI821
aI3551
aI4833
aI927
aI24331
aI63669
aI26549
aI220159
asS'd'
I17959
sa(dp35918
S'a'
I76541
sS's'
I18
sS'm'
(lp35919
I1
aI1
aI3
aI9
aI21
aI7
aI9
aI183
aI391
aI783
aI493
aI2785
aI3879
aI8311
aI9935
aI60629
aI119329
aI5791
asS'd'
I17960
sa(dp35920
S'a'
I76553
sS's'
I18
sS'm'
(lp35921
I1
aI1
aI5
aI15
aI15
aI41
aI61
aI97
aI33
aI29
aI199
aI3335
aI1531
aI6107
aI757
aI33797
aI3001
aI224507
asS'd'
I17961
sa(dp35922
S'a'
I76568
sS's'
I18
sS'm'
(lp35923
I1
aI1
aI3
aI9
aI13
aI39
aI25
aI247
aI407
aI1
aI1129
aI1453
aI7091
aI5557
aI8657
aI33961
aI100763
aI25099
asS'd'
I17962
sa(dp35924
S'a'
I76574
sS's'
I18
sS'm'
(lp35925
I1
aI3
aI3
aI13
aI31
aI21
aI69
aI73
aI431
aI827
aI861
aI235
aI2369
aI4283
aI27183
aI29095
aI99957
aI97577
asS'd'
I17963
sa(dp35926
S'a'
I76584
sS's'
I18
sS'm'
(lp35927
I1
aI1
aI7
aI13
aI1
aI25
aI21
aI173
aI365
aI921
aI21
aI3527
aI2481
aI8795
aI25621
aI41755
aI127249
aI221385
asS'd'
I17964
sa(dp35928
S'a'
I76604
sS's'
I18
sS'm'
(lp35929
I1
aI1
aI1
aI5
aI25
aI25
aI65
aI203
aI305
aI373
aI527
aI4033
aI3483
aI9403
aI28669
aI32083
aI52273
aI77037
asS'd'
I17965
sa(dp35930
S'a'
I76607
sS's'
I18
sS'm'
(lp35931
I1
aI3
aI1
aI15
aI9
aI23
aI19
aI7
aI29
aI83
aI1163
aI1147
aI5315
aI2381
aI21203
aI33915
aI109511
aI40669
asS'd'
I17966
sa(dp35932
S'a'
I76609
sS's'
I18
sS'm'
(lp35933
I1
aI3
aI7
aI15
aI23
aI19
aI69
aI127
aI113
aI937
aI935
aI1067
aI2431
aI7677
aI21327
aI44095
aI82799
aI5715
asS'd'
I17967
sa(dp35934
S'a'
I76669
sS's'
I18
sS'm'
(lp35935
I1
aI1
aI5
aI5
aI7
aI37
aI107
aI223
aI433
aI515
aI393
aI1721
aI1977
aI6383
aI18835
aI54841
aI103263
aI196997
asS'd'
I17968
sa(dp35936
S'a'
I76683
sS's'
I18
sS'm'
(lp35937
I1
aI1
aI1
aI3
aI1
aI11
aI85
aI173
aI259
aI685
aI595
aI1635
aI6979
aI4483
aI8097
aI42249
aI56259
aI105925
asS'd'
I17969
sa(dp35938
S'a'
I76745
sS's'
I18
sS'm'
(lp35939
I1
aI1
aI7
aI7
aI1
aI23
aI11
aI253
aI187
aI665
aI313
aI3745
aI2423
aI15835
aI32085
aI48643
aI75625
aI47511
asS'd'
I17970
sa(dp35940
S'a'
I76753
sS's'
I18
sS'm'
(lp35941
I1
aI1
aI3
aI5
aI1
aI59
aI127
aI83
aI501
aI387
aI977
aI3515
aI7921
aI12329
aI14757
aI20287
aI49699
aI91237
asS'd'
I17971
sa(dp35942
S'a'
I76754
sS's'
I18
sS'm'
(lp35943
I1
aI3
aI5
aI11
aI31
aI45
aI51
aI109
aI319
aI621
aI1013
aI3519
aI4023
aI12099
aI28829
aI26691
aI83131
aI261497
asS'd'
I17972
sa(dp35944
S'a'
I76756
sS's'
I18
sS'm'
(lp35945
I1
aI1
aI3
aI3
aI5
aI35
aI51
aI253
aI253
aI569
aI1017
aI2299
aI8159
aI13783
aI22123
aI55213
aI111527
aI110699
asS'd'
I17973
sa(dp35946
S'a'
I76760
sS's'
I18
sS'm'
(lp35947
I1
aI3
aI7
aI3
aI9
aI5
aI59
aI129
aI41
aI845
aI723
aI1607
aI3047
aI14323
aI19277
aI39447
aI12465
aI45925
asS'd'
I17974
sa(dp35948
S'a'
I76782
sS's'
I18
sS'm'
(lp35949
I1
aI1
aI3
aI1
aI17
aI35
aI51
aI79
aI115
aI361
aI739
aI2037
aI6167
aI14699
aI28187
aI65271
aI67285
aI48489
asS'd'
I17975
sa(dp35950
S'a'
I76821
sS's'
I18
sS'm'
(lp35951
I1
aI3
aI1
aI11
aI1
aI29
aI95
aI181
aI419
aI235
aI745
aI621
aI3889
aI2933
aI743
aI23801
aI32057
aI54103
asS'd'
I17976
sa(dp35952
S'a'
I76828
sS's'
I18
sS'm'
(lp35953
I1
aI3
aI1
aI11
aI17
aI47
aI43
aI55
aI7
aI695
aI1653
aI3983
aI961
aI3037
aI8669
aI10039
aI86571
aI6981
asS'd'
I17977
sa(dp35954
S'a'
I76849
sS's'
I18
sS'm'
(lp35955
I1
aI1
aI5
aI15
aI13
aI19
aI67
aI141
aI291
aI511
aI1913
aI397
aI7423
aI6541
aI21845
aI49821
aI126047
aI218587
asS'd'
I17978
sa(dp35956
S'a'
I76850
sS's'
I18
sS'm'
(lp35957
I1
aI1
aI5
aI3
aI11
aI13
aI103
aI213
aI189
aI115
aI1495
aI2695
aI2127
aI11979
aI13609
aI46615
aI64775
aI206417
asS'd'
I17979
sa(dp35958
S'a'
I76870
sS's'
I18
sS'm'
(lp35959
I1
aI3
aI1
aI5
aI5
aI9
aI57
aI207
aI253
aI251
aI1155
aI1319
aI6699
aI6613
aI21757
aI49703
aI124879
aI89987
asS'd'
I17980
sa(dp35960
S'a'
I76874
sS's'
I18
sS'm'
(lp35961
I1
aI3
aI5
aI7
aI23
aI25
aI35
aI81
aI165
aI789
aI771
aI415
aI5557
aI8431
aI12043
aI44359
aI9447
aI229481
asS'd'
I17981
sa(dp35962
S'a'
I76898
sS's'
I18
sS'm'
(lp35963
I1
aI1
aI1
aI13
aI17
aI21
aI63
aI251
aI387
aI767
aI85
aI3901
aI3227
aI10329
aI5049
aI56173
aI58065
aI78595
asS'd'
I17982
sa(dp35964
S'a'
I76903
sS's'
I18
sS'm'
(lp35965
I1
aI3
aI1
aI15
aI23
aI5
aI45
aI7
aI123
aI389
aI1041
aI1223
aI5865
aI5365
aI2915
aI24861
aI106893
aI170769
asS'd'
I17983
sa(dp35966
S'a'
I76910
sS's'
I18
sS'm'
(lp35967
I1
aI3
aI7
aI15
aI27
aI61
aI27
aI59
aI309
aI103
aI279
aI1829
aI1501
aI11277
aI4461
aI34817
aI60973
aI99805
asS'd'
I17984
sa(dp35968
S'a'
I76921
sS's'
I18
sS'm'
(lp35969
I1
aI1
aI3
aI9
aI1
aI25
aI57
aI85
aI411
aI699
aI911
aI1643
aI2687
aI13539
aI10187
aI21597
aI18883
aI212975
asS'd'
I17985
sa(dp35970
S'a'
I76968
sS's'
I18
sS'm'
(lp35971
I1
aI1
aI3
aI7
aI5
aI7
aI81
aI209
aI225
aI321
aI1867
aI2189
aI6315
aI5393
aI8859
aI47471
aI41677
aI222455
asS'd'
I17986
sa(dp35972
S'a'
I77013
sS's'
I18
sS'm'
(lp35973
I1
aI1
aI5
aI11
aI27
aI33
aI119
aI159
aI273
aI659
aI883
aI3773
aI6519
aI15449
aI17219
aI23923
aI33749
aI225489
asS'd'
I17987
sa(dp35974
S'a'
I77027
sS's'
I18
sS'm'
(lp35975
I1
aI1
aI7
aI13
aI29
aI39
aI1
aI161
aI165
aI531
aI1019
aI2369
aI2093
aI4341
aI24945
aI28537
aI49467
aI258065
asS'd'
I17988
sa(dp35976
S'a'
I77042
sS's'
I18
sS'm'
(lp35977
I1
aI3
aI3
aI3
aI15
aI59
aI23
aI143
aI377
aI943
aI1329
aI977
aI7025
aI2167
aI17973
aI65087
aI115757
aI75959
asS'd'
I17989
sa(dp35978
S'a'
I77059
sS's'
I18
sS'm'
(lp35979
I1
aI1
aI1
aI15
aI17
aI55
aI33
aI167
aI43
aI719
aI51
aI3873
aI3317
aI10763
aI639
aI58195
aI20023
aI100725
asS'd'
I17990
sa(dp35980
S'a'
I77062
sS's'
I18
sS'm'
(lp35981
I1
aI1
aI7
aI5
aI17
aI23
aI71
aI249
aI23
aI929
aI467
aI3073
aI3355
aI1343
aI18755
aI12247
aI49737
aI184103
asS'd'
I17991
sa(dp35982
S'a'
I77074
sS's'
I18
sS'm'
(lp35983
I1
aI3
aI3
aI15
aI1
aI9
aI17
aI193
aI157
aI265
aI983
aI1825
aI4805
aI2131
aI22117
aI32937
aI57
aI261867
asS'd'
I17992
sa(dp35984
S'a'
I77076
sS's'
I18
sS'm'
(lp35985
I1
aI3
aI5
aI9
aI5
aI1
aI101
aI141
aI511
aI489
aI73
aI1789
aI1303
aI2633
aI709
aI11891
aI44897
aI191229
asS'd'
I17993
sa(dp35986
S'a'
I77110
sS's'
I18
sS'm'
(lp35987
I1
aI1
aI7
aI15
aI1
aI27
aI121
aI7
aI129
aI421
aI725
aI1421
aI3883
aI13335
aI7247
aI8393
aI85029
aI127691
asS'd'
I17994
sa(dp35988
S'a'
I77113
sS's'
I18
sS'm'
(lp35989
I1
aI1
aI3
aI9
aI19
aI53
aI121
aI115
aI85
aI909
aI1535
aI3261
aI7063
aI16381
aI1719
aI19847
aI19041
aI215433
asS'd'
I17995
sa(dp35990
S'a'
I77131
sS's'
I18
sS'm'
(lp35991
I1
aI1
aI7
aI3
aI17
aI45
aI91
aI187
aI181
aI829
aI609
aI931
aI5727
aI3971
aI14567
aI15871
aI9825
aI184165
asS'd'
I17996
sa(dp35992
S'a'
I77157
sS's'
I18
sS'm'
(lp35993
I1
aI1
aI1
aI3
aI5
aI29
aI7
aI249
aI361
aI815
aI1101
aI1485
aI6879
aI5379
aI7179
aI27467
aI101427
aI196089
asS'd'
I17997
sa(dp35994
S'a'
I77158
sS's'
I18
sS'm'
(lp35995
I1
aI1
aI5
aI7
aI23
aI11
aI27
aI175
aI237
aI747
aI1911
aI3107
aI961
aI6649
aI29887
aI11003
aI27561
aI233841
asS'd'
I17998
sa(dp35996
S'a'
I77216
sS's'
I18
sS'm'
(lp35997
I1
aI1
aI1
aI5
aI11
aI5
aI125
aI227
aI303
aI315
aI1879
aI817
aI7445
aI1447
aI9333
aI54825
aI118865
aI216397
asS'd'
I17999
sa(dp35998
S'a'
I77246
sS's'
I18
sS'm'
(lp35999
I1
aI1
aI3
aI13
aI17
aI33
aI27
aI95
aI245
aI25
aI1741
aI2633
aI1869
aI14111
aI24507
aI61287
aI46397
aI220803
asS'd'
I18000
sa(dp36000
S'a'
I77258
sS's'
I18
sS'm'
(lp36001
I1
aI1
aI7
aI7
aI25
aI5
aI41
aI101
aI171
aI333
aI497
aI3417
aI4921
aI4553
aI25487
aI51529
aI72873
aI43525
asS'd'
I18001
sa(dp36002
S'a'
I77281
sS's'
I18
sS'm'
(lp36003
I1
aI1
aI7
aI7
aI9
aI19
aI25
aI161
aI235
aI929
aI1663
aI3237
aI323
aI3889
aI31423
aI2345
aI63113
aI212659
asS'd'
I18002
sa(dp36004
S'a'
I77284
sS's'
I18
sS'm'
(lp36005
I1
aI1
aI5
aI13
aI29
aI59
aI39
aI25
aI393
aI519
aI429
aI1461
aI5867
aI113
aI28091
aI36813
aI47827
aI163407
asS'd'
I18003
sa(dp36006
S'a'
I77294
sS's'
I18
sS'm'
(lp36007
I1
aI3
aI5
aI13
aI15
aI49
aI85
aI161
aI83
aI389
aI765
aI3349
aI4659
aI11007
aI24749
aI51121
aI93511
aI229885
asS'd'
I18004
sa(dp36008
S'a'
I77302
sS's'
I18
sS'm'
(lp36009
I1
aI3
aI5
aI11
aI5
aI27
aI107
aI233
aI221
aI425
aI941
aI1181
aI5403
aI4373
aI32625
aI41991
aI2019
aI245967
asS'd'
I18005
sa(dp36010
S'a'
I77306
sS's'
I18
sS'm'
(lp36011
I1
aI3
aI5
aI9
aI9
aI53
aI97
aI27
aI221
aI731
aI1301
aI3517
aI4407
aI11369
aI4251
aI31121
aI4813
aI42029
asS'd'
I18006
sa(dp36012
S'a'
I77315
sS's'
I18
sS'm'
(lp36013
I1
aI1
aI1
aI9
aI17
aI59
aI107
aI247
aI231
aI123
aI1177
aI3299
aI6163
aI4855
aI14547
aI63171
aI45201
aI27711
asS'd'
I18007
sa(dp36014
S'a'
I77352
sS's'
I18
sS'm'
(lp36015
I1
aI3
aI7
aI3
aI25
aI31
aI63
aI37
aI123
aI457
aI1531
aI3723
aI4807
aI14665
aI17973
aI42547
aI5417
aI170323
asS'd'
I18008
sa(dp36016
S'a'
I77360
sS's'
I18
sS'm'
(lp36017
I1
aI3
aI7
aI3
aI17
aI19
aI57
aI7
aI359
aI741
aI385
aI3127
aI855
aI10803
aI30093
aI24501
aI53629
aI40447
asS'd'
I18009
sa(dp36018
S'a'
I77365
sS's'
I18
sS'm'
(lp36019
I1
aI3
aI7
aI15
aI11
aI45
aI49
aI125
aI445
aI795
aI113
aI2425
aI7085
aI7337
aI16297
aI26447
aI94369
aI12371
asS'd'
I18010
sa(dp36020
S'a'
I77370
sS's'
I18
sS'm'
(lp36021
I1
aI3
aI7
aI9
aI29
aI59
aI35
aI191
aI123
aI619
aI415
aI1081
aI2469
aI4125
aI25587
aI7853
aI119781
aI9447
asS'd'
I18011
sa(dp36022
S'a'
I77378
sS's'
I18
sS'm'
(lp36023
I1
aI3
aI1
aI15
aI13
aI13
aI111
aI89
aI381
aI757
aI389
aI253
aI6929
aI33
aI8263
aI17385
aI122129
aI146679
asS'd'
I18012
sa(dp36024
S'a'
I77380
sS's'
I18
sS'm'
(lp36025
I1
aI1
aI3
aI3
aI15
aI35
aI101
aI95
aI479
aI577
aI1645
aI3781
aI7533
aI4665
aI6561
aI49897
aI72413
aI151383
asS'd'
I18013
sa(dp36026
S'a'
I77407
sS's'
I18
sS'm'
(lp36027
I1
aI1
aI1
aI1
aI7
aI49
aI23
aI223
aI189
aI763
aI227
aI2805
aI8093
aI389
aI11525
aI30915
aI91341
aI210231
asS'd'
I18014
sa(dp36028
S'a'
I77408
sS's'
I18
sS'm'
(lp36029
I1
aI1
aI7
aI3
aI19
aI23
aI3
aI137
aI79
aI569
aI1833
aI2091
aI4235
aI10739
aI22855
aI33845
aI120141
aI220267
asS'd'
I18015
sa(dp36030
S'a'
I77414
sS's'
I18
sS'm'
(lp36031
I1
aI3
aI7
aI3
aI11
aI43
aI85
aI63
aI419
aI681
aI365
aI3017
aI3603
aI6413
aI13515
aI16003
aI107949
aI241261
asS'd'
I18016
sa(dp36032
S'a'
I77444
sS's'
I18
sS'm'
(lp36033
I1
aI1
aI5
aI13
aI3
aI35
aI41
aI193
aI189
aI999
aI1395
aI2431
aI2227
aI7245
aI23929
aI16137
aI14591
aI54999
asS'd'
I18017
sa(dp36034
S'a'
I77454
sS's'
I18
sS'm'
(lp36035
I1
aI1
aI7
aI15
aI23
aI51
aI47
aI77
aI31
aI25
aI589
aI611
aI371
aI13329
aI5873
aI2133
aI40351
aI145293
asS'd'
I18018
sa(dp36036
S'a'
I77481
sS's'
I18
sS'm'
(lp36037
I1
aI1
aI7
aI15
aI17
aI19
aI53
aI155
aI309
aI573
aI1059
aI3557
aI2445
aI12205
aI4497
aI32061
aI130293
aI73859
asS'd'
I18019
sa(dp36038
S'a'
I77482
sS's'
I18
sS'm'
(lp36039
I1
aI1
aI7
aI13
aI3
aI25
aI71
aI157
aI237
aI185
aI1035
aI1759
aI1331
aI13533
aI25635
aI811
aI54391
aI91109
asS'd'
I18020
sa(dp36040
S'a'
I77531
sS's'
I18
sS'm'
(lp36041
I1
aI1
aI1
aI11
aI5
aI21
aI99
aI31
aI259
aI413
aI2033
aI2187
aI755
aI4591
aI28641
aI64031
aI88499
aI160789
asS'd'
I18021
sa(dp36042
S'a'
I77544
sS's'
I18
sS'm'
(lp36043
I1
aI3
aI7
aI13
aI29
aI33
aI13
aI157
aI97
aI981
aI329
aI81
aI6351
aI4171
aI10925
aI22733
aI72521
aI105477
asS'd'
I18022
sa(dp36044
S'a'
I77552
sS's'
I18
sS'm'
(lp36045
I1
aI3
aI7
aI9
aI11
aI31
aI97
aI35
aI337
aI309
aI847
aI3429
aI2697
aI3141
aI19481
aI43679
aI11129
aI205757
asS'd'
I18023
sa(dp36046
S'a'
I77569
sS's'
I18
sS'm'
(lp36047
I1
aI3
aI1
aI7
aI27
aI45
aI123
aI193
aI439
aI639
aI633
aI1375
aI7307
aI1599
aI23379
aI56811
aI100877
aI228687
asS'd'
I18024
sa(dp36048
S'a'
I77579
sS's'
I18
sS'm'
(lp36049
I1
aI3
aI7
aI1
aI29
aI43
aI103
aI131
aI103
aI933
aI143
aI2431
aI2221
aI4565
aI20841
aI58611
aI49163
aI13673
asS'd'
I18025
sa(dp36050
S'a'
I77610
sS's'
I18
sS'm'
(lp36051
I1
aI3
aI5
aI13
aI25
aI17
aI121
aI17
aI455
aI941
aI1577
aI509
aI5401
aI797
aI29573
aI38373
aI50527
aI17951
asS'd'
I18026
sa(dp36052
S'a'
I77612
sS's'
I18
sS'm'
(lp36053
I1
aI1
aI1
aI13
aI25
aI21
aI77
aI253
aI199
aI871
aI935
aI3919
aI1687
aI6653
aI20345
aI56969
aI77989
aI244767
asS'd'
I18027
sa(dp36054
S'a'
I77623
sS's'
I18
sS'm'
(lp36055
I1
aI1
aI3
aI1
aI17
aI5
aI5
aI191
aI279
aI33
aI579
aI651
aI969
aI6091
aI11659
aI1643
aI17935
aI85145
asS'd'
I18028
sa(dp36056
S'a'
I77649
sS's'
I18
sS'm'
(lp36057
I1
aI3
aI3
aI9
aI29
aI1
aI103
aI39
aI83
aI295
aI1237
aI207
aI4837
aI7899
aI27879
aI23195
aI29549
aI206885
asS'd'
I18029
sa(dp36058
S'a'
I77650
sS's'
I18
sS'm'
(lp36059
I1
aI3
aI5
aI1
aI9
aI55
aI115
aI37
aI225
aI447
aI943
aI1133
aI6203
aI949
aI9973
aI4309
aI43969
aI166795
asS'd'
I18030
sa(dp36060
S'a'
I77661
sS's'
I18
sS'm'
(lp36061
I1
aI1
aI7
aI7
aI17
aI43
aI75
aI251
aI35
aI489
aI1011
aI355
aI4113
aI2377
aI13775
aI34935
aI84905
aI252973
asS'd'
I18031
sa(dp36062
S'a'
I77675
sS's'
I18
sS'm'
(lp36063
I1
aI3
aI3
aI3
aI11
aI45
aI3
aI1
aI135
aI499
aI81
aI3265
aI6657
aI3875
aI27565
aI60931
aI13117
aI87931
asS'd'
I18032
sa(dp36064
S'a'
I77686
sS's'
I18
sS'm'
(lp36065
I1
aI1
aI3
aI1
aI9
aI1
aI77
aI69
aI137
aI241
aI1613
aI2607
aI3307
aI171
aI13551
aI54529
aI45937
aI180411
asS'd'
I18033
sa(dp36066
S'a'
I77742
sS's'
I18
sS'm'
(lp36067
I1
aI1
aI1
aI1
aI19
aI29
aI77
aI255
aI95
aI461
aI567
aI1103
aI2753
aI10627
aI19479
aI43411
aI128565
aI29869
asS'd'
I18034
sa(dp36068
S'a'
I77791
sS's'
I18
sS'm'
(lp36069
I1
aI1
aI3
aI5
aI5
aI63
aI123
aI159
aI165
aI733
aI1107
aI1711
aI5039
aI9221
aI15541
aI5527
aI27629
aI206505
asS'd'
I18035
sa(dp36070
S'a'
I77792
sS's'
I18
sS'm'
(lp36071
I1
aI3
aI1
aI3
aI7
aI45
aI73
aI63
aI413
aI693
aI433
aI2281
aI3981
aI7719
aI31473
aI56939
aI70391
aI67467
asS'd'
I18036
sa(dp36072
S'a'
I77807
sS's'
I18
sS'm'
(lp36073
I1
aI1
aI1
aI11
aI19
aI33
aI113
aI151
aI427
aI603
aI1653
aI2451
aI5367
aI12171
aI14373
aI33175
aI62013
aI209273
asS'd'
I18037
sa(dp36074
S'a'
I77815
sS's'
I18
sS'm'
(lp36075
I1
aI3
aI5
aI5
aI17
aI37
aI109
aI5
aI187
aI293
aI617
aI2663
aI7381
aI14217
aI23561
aI48999
aI108717
aI248289
asS'd'
I18038
sa(dp36076
S'a'
I77842
sS's'
I18
sS'm'
(lp36077
I1
aI1
aI5
aI1
aI9
aI27
aI35
aI127
aI355
aI479
aI281
aI2081
aI7303
aI259
aI8893
aI59141
aI20927
aI61611
asS'd'
I18039
sa(dp36078
S'a'
I77847
sS's'
I18
sS'm'
(lp36079
I1
aI3
aI3
aI15
aI31
aI33
aI71
aI209
aI315
aI363
aI593
aI1035
aI8029
aI12501
aI2859
aI54745
aI39391
aI153259
asS'd'
I18040
sa(dp36080
S'a'
I77899
sS's'
I18
sS'm'
(lp36081
I1
aI3
aI3
aI11
aI21
aI39
aI35
aI173
aI171
aI15
aI987
aI3737
aI7415
aI1827
aI973
aI6831
aI108643
aI241333
asS'd'
I18041
sa(dp36082
S'a'
I77901
sS's'
I18
sS'm'
(lp36083
I1
aI3
aI7
aI9
aI17
aI37
aI127
aI243
aI153
aI195
aI113
aI309
aI5301
aI13619
aI7927
aI35385
aI9501
aI99241
asS'd'
I18042
sa(dp36084
S'a'
I77904
sS's'
I18
sS'm'
(lp36085
I1
aI3
aI5
aI13
aI23
aI9
aI81
aI235
aI139
aI635
aI443
aI2235
aI2613
aI2389
aI18431
aI8409
aI2885
aI254811
asS'd'
I18043
sa(dp36086
S'a'
I77914
sS's'
I18
sS'm'
(lp36087
I1
aI3
aI7
aI9
aI1
aI5
aI15
aI109
aI141
aI173
aI1059
aI1961
aI7945
aI10381
aI17337
aI19591
aI42173
aI119831
asS'd'
I18044
sa(dp36088
S'a'
I77925
sS's'
I18
sS'm'
(lp36089
I1
aI3
aI1
aI13
aI19
aI7
aI111
aI111
aI345
aI327
aI1147
aI2293
aI49
aI16213
aI25309
aI60537
aI50421
aI108467
asS'd'
I18045
sa(dp36090
S'a'
I77950
sS's'
I18
sS'm'
(lp36091
I1
aI1
aI5
aI1
aI3
aI23
aI63
aI219
aI69
aI879
aI1397
aI3857
aI1859
aI1939
aI4851
aI26549
aI86019
aI7927
asS'd'
I18046
sa(dp36092
S'a'
I77959
sS's'
I18
sS'm'
(lp36093
I1
aI3
aI1
aI13
aI23
aI61
aI25
aI31
aI301
aI189
aI1031
aI2817
aI829
aI8777
aI26869
aI54405
aI43535
aI234687
asS'd'
I18047
sa(dp36094
S'a'
I77994
sS's'
I18
sS'm'
(lp36095
I1
aI1
aI1
aI9
aI11
aI31
aI13
aI139
aI77
aI567
aI949
aI3415
aI6955
aI14973
aI9565
aI37911
aI18395
aI94167
asS'd'
I18048
sa(dp36096
S'a'
I78004
sS's'
I18
sS'm'
(lp36097
I1
aI3
aI5
aI13
aI17
aI17
aI21
aI213
aI171
aI993
aI1001
aI979
aI5085
aI3909
aI11797
aI48669
aI73541
aI48979
asS'd'
I18049
sa(dp36098
S'a'
I78043
sS's'
I18
sS'm'
(lp36099
I1
aI1
aI7
aI9
aI7
aI37
aI35
aI107
aI347
aI239
aI585
aI2883
aI3235
aI1053
aI14871
aI25799
aI4861
aI56335
asS'd'
I18050
sa(dp36100
S'a'
I78052
sS's'
I18
sS'm'
(lp36101
I1
aI1
aI3
aI5
aI19
aI7
aI91
aI139
aI325
aI921
aI863
aI209
aI845
aI15943
aI8281
aI55103
aI110193
aI216091
asS'd'
I18051
sa(dp36102
S'a'
I78061
sS's'
I18
sS'm'
(lp36103
I1
aI3
aI1
aI13
aI31
aI33
aI65
aI155
aI177
aI103
aI1991
aI343
aI6299
aI3587
aI30215
aI64335
aI114301
aI220403
asS'd'
I18052
sa(dp36104
S'a'
I78064
sS's'
I18
sS'm'
(lp36105
I1
aI3
aI5
aI3
aI31
aI37
aI121
aI157
aI443
aI349
aI1097
aI3683
aI503
aI14061
aI14685
aI29755
aI61543
aI232983
asS'd'
I18053
sa(dp36106
S'a'
I78084
sS's'
I18
sS'm'
(lp36107
I1
aI3
aI3
aI11
aI17
aI59
aI29
aI161
aI381
aI791
aI1647
aI1077
aI6369
aI1095
aI17279
aI43141
aI65003
aI144609
asS'd'
I18054
sa(dp36108
S'a'
I78094
sS's'
I18
sS'm'
(lp36109
I1
aI1
aI5
aI1
aI1
aI15
aI67
aI77
aI3
aI585
aI1909
aI1485
aI3003
aI591
aI4711
aI10279
aI75901
aI226417
asS'd'
I18055
sa(dp36110
S'a'
I78099
sS's'
I18
sS'm'
(lp36111
I1
aI3
aI7
aI5
aI1
aI5
aI5
aI193
aI469
aI631
aI1065
aI607
aI2751
aI8163
aI13633
aI40563
aI1417
aI118169
asS'd'
I18056
sa(dp36112
S'a'
I78129
sS's'
I18
sS'm'
(lp36113
I1
aI1
aI7
aI9
aI25
aI25
aI109
aI27
aI157
aI495
aI225
aI1385
aI4315
aI995
aI10591
aI1629
aI129939
aI56765
asS'd'
I18057
sa(dp36114
S'a'
I78142
sS's'
I18
sS'm'
(lp36115
I1
aI3
aI1
aI7
aI9
aI23
aI61
aI63
aI35
aI145
aI1537
aI1029
aI4225
aI1467
aI10519
aI32861
aI519
aI53983
asS'd'
I18058
sa(dp36116
S'a'
I78149
sS's'
I18
sS'm'
(lp36117
I1
aI3
aI3
aI11
aI7
aI59
aI25
aI199
aI403
aI967
aI1089
aI1121
aI1063
aI6701
aI16827
aI55479
aI72983
aI36873
asS'd'
I18059
sa(dp36118
S'a'
I78153
sS's'
I18
sS'm'
(lp36119
I1
aI1
aI1
aI13
aI9
aI27
aI19
aI23
aI395
aI229
aI1837
aI1231
aI1737
aI10475
aI16743
aI42369
aI130331
aI47255
asS'd'
I18060
sa(dp36120
S'a'
I78171
sS's'
I18
sS'm'
(lp36121
I1
aI1
aI5
aI7
aI29
aI15
aI95
aI155
aI339
aI65
aI751
aI2399
aI5615
aI2987
aI16769
aI57381
aI113021
aI41417
asS'd'
I18061
sa(dp36122
S'a'
I78173
sS's'
I18
sS'm'
(lp36123
I1
aI3
aI1
aI9
aI15
aI17
aI1
aI111
aI197
aI7
aI417
aI3999
aI7261
aI5939
aI16773
aI29275
aI105559
aI84685
asS'd'
I18062
sa(dp36124
S'a'
I78174
sS's'
I18
sS'm'
(lp36125
I1
aI3
aI3
aI13
aI19
aI31
aI103
aI1
aI37
aI269
aI1257
aI1397
aI4293
aI3019
aI6503
aI7727
aI93943
aI237313
asS'd'
I18063
sa(dp36126
S'a'
I78195
sS's'
I18
sS'm'
(lp36127
I1
aI1
aI3
aI9
aI13
aI37
aI67
aI129
aI43
aI669
aI1331
aI1787
aI8185
aI323
aI18749
aI13737
aI86123
aI154131
asS'd'
I18064
sa(dp36128
S'a'
I78201
sS's'
I18
sS'm'
(lp36129
I1
aI3
aI1
aI11
aI3
aI51
aI13
aI35
aI197
aI867
aI559
aI1381
aI1057
aI13293
aI20603
aI18633
aI50503
aI169685
asS'd'
I18065
sa(dp36130
S'a'
I78202
sS's'
I18
sS'm'
(lp36131
I1
aI3
aI1
aI11
aI9
aI35
aI7
aI51
aI499
aI885
aI353
aI4095
aI6491
aI5917
aI15053
aI18363
aI99593
aI213089
asS'd'
I18066
sa(dp36132
S'a'
I78241
sS's'
I18
sS'm'
(lp36133
I1
aI1
aI3
aI9
aI19
aI23
aI107
aI147
aI339
aI331
aI1349
aI2855
aI3721
aI13317
aI26457
aI783
aI93949
aI196051
asS'd'
I18067
sa(dp36134
S'a'
I78247
sS's'
I18
sS'm'
(lp36135
I1
aI1
aI5
aI1
aI9
aI61
aI89
aI217
aI315
aI385
aI1729
aI2641
aI5753
aI6269
aI547
aI33737
aI20103
aI31533
asS'd'
I18068
sa(dp36136
S'a'
I78274
sS's'
I18
sS'm'
(lp36137
I1
aI3
aI5
aI13
aI13
aI61
aI3
aI191
aI57
aI683
aI1227
aI1255
aI3651
aI10687
aI9049
aI6529
aI60783
aI28639
asS'd'
I18069
sa(dp36138
S'a'
I78276
sS's'
I18
sS'm'
(lp36139
I1
aI1
aI7
aI11
aI25
aI41
aI79
aI19
aI383
aI363
aI1731
aI1597
aI1651
aI15037
aI22191
aI51883
aI41927
aI82419
asS'd'
I18070
sa(dp36140
S'a'
I78303
sS's'
I18
sS'm'
(lp36141
I1
aI3
aI5
aI9
aI15
aI61
aI39
aI149
aI49
aI633
aI709
aI1743
aI621
aI14659
aI3309
aI64129
aI91897
aI74235
asS'd'
I18071
sa(dp36142
S'a'
I78307
sS's'
I18
sS'm'
(lp36143
I1
aI3
aI7
aI15
aI5
aI59
aI7
aI197
aI111
aI885
aI1737
aI855
aI2807
aI3817
aI13759
aI29989
aI45105
aI171689
asS'd'
I18072
sa(dp36144
S'a'
I78328
sS's'
I18
sS'm'
(lp36145
I1
aI1
aI3
aI9
aI21
aI25
aI55
aI67
aI483
aI437
aI303
aI703
aI6993
aI1971
aI4565
aI56117
aI6105
aI254517
asS'd'
I18073
sa(dp36146
S'a'
I78344
sS's'
I18
sS'm'
(lp36147
I1
aI3
aI3
aI13
aI15
aI13
aI19
aI3
aI487
aI751
aI1185
aI2985
aI1619
aI7139
aI26087
aI21105
aI9049
aI236153
asS'd'
I18074
sa(dp36148
S'a'
I78362
sS's'
I18
sS'm'
(lp36149
I1
aI1
aI5
aI7
aI15
aI55
aI51
aI231
aI85
aI953
aI713
aI659
aI2021
aI4271
aI15961
aI26873
aI31141
aI76635
asS'd'
I18075
sa(dp36150
S'a'
I78367
sS's'
I18
sS'm'
(lp36151
I1
aI3
aI5
aI1
aI11
aI39
aI3
aI223
aI367
aI903
aI799
aI415
aI7247
aI9539
aI14479
aI37195
aI59951
aI181935
asS'd'
I18076
sa(dp36152
S'a'
I78368
sS's'
I18
sS'm'
(lp36153
I1
aI1
aI5
aI3
aI13
aI47
aI17
aI159
aI439
aI859
aI1067
aI3111
aI5277
aI13973
aI21999
aI28381
aI115685
aI231483
asS'd'
I18077
sa(dp36154
S'a'
I78409
sS's'
I18
sS'm'
(lp36155
I1
aI1
aI7
aI15
aI17
aI21
aI69
aI131
aI193
aI479
aI1075
aI3271
aI2057
aI1295
aI31235
aI35027
aI94145
aI65419
asS'd'
I18078
sa(dp36156
S'a'
I78412
sS's'
I18
sS'm'
(lp36157
I1
aI3
aI3
aI5
aI5
aI21
aI5
aI81
aI113
aI259
aI837
aI831
aI5985
aI6717
aI12041
aI40355
aI50957
aI111185
asS'd'
I18079
sa(dp36158
S'a'
I78417
sS's'
I18
sS'm'
(lp36159
I1
aI1
aI1
aI9
aI15
aI47
aI103
aI195
aI465
aI739
aI1415
aI225
aI3121
aI12623
aI7539
aI17555
aI36703
aI217641
asS'd'
I18080
sa(dp36160
S'a'
I78430
sS's'
I18
sS'm'
(lp36161
I1
aI3
aI1
aI3
aI31
aI17
aI91
aI153
aI221
aI217
aI525
aI981
aI281
aI9869
aI9713
aI10669
aI12049
aI97615
asS'd'
I18081
sa(dp36162
S'a'
I78433
sS's'
I18
sS'm'
(lp36163
I1
aI1
aI5
aI7
aI29
aI1
aI1
aI199
aI415
aI843
aI301
aI941
aI4589
aI13301
aI5833
aI41311
aI74019
aI78537
asS'd'
I18082
sa(dp36164
S'a'
I78440
sS's'
I18
sS'm'
(lp36165
I1
aI1
aI5
aI11
aI13
aI5
aI41
aI127
aI213
aI917
aI1297
aI2281
aI3193
aI3877
aI9517
aI40685
aI14657
aI185139
asS'd'
I18083
sa(dp36166
S'a'
I78451
sS's'
I18
sS'm'
(lp36167
I1
aI1
aI1
aI7
aI21
aI45
aI87
aI33
aI425
aI487
aI643
aI271
aI7087
aI5979
aI14795
aI27575
aI34541
aI173251
asS'd'
I18084
sa(dp36168
S'a'
I78453
sS's'
I18
sS'm'
(lp36169
I1
aI3
aI7
aI5
aI21
aI11
aI7
aI169
aI325
aI905
aI973
aI2853
aI7929
aI8801
aI1005
aI60641
aI45973
aI81859
asS'd'
I18085
sa(dp36170
S'a'
I78458
sS's'
I18
sS'm'
(lp36171
I1
aI3
aI3
aI1
aI1
aI35
aI39
aI81
aI93
aI463
aI697
aI2309
aI7769
aI5169
aI17595
aI41447
aI28837
aI52613
asS'd'
I18086
sa(dp36172
S'a'
I78467
sS's'
I18
sS'm'
(lp36173
I1
aI1
aI7
aI1
aI1
aI23
aI37
aI17
aI137
aI873
aI1657
aI681
aI503
aI7887
aI24463
aI32453
aI112727
aI133347
asS'd'
I18087
sa(dp36174
S'a'
I78479
sS's'
I18
sS'm'
(lp36175
I1
aI1
aI5
aI9
aI19
aI35
aI37
aI85
aI11
aI245
aI11
aI3
aI6475
aI5953
aI247
aI49447
aI32813
aI243841
asS'd'
I18088
sa(dp36176
S'a'
I78507
sS's'
I18
sS'm'
(lp36177
I1
aI3
aI5
aI3
aI19
aI53
aI37
aI45
aI431
aI259
aI1831
aI1443
aI2237
aI7651
aI20701
aI22857
aI50041
aI119667
asS'd'
I18089
sa(dp36178
S'a'
I78518
sS's'
I18
sS'm'
(lp36179
I1
aI1
aI7
aI1
aI5
aI37
aI113
aI69
aI389
aI369
aI1251
aI1989
aI7613
aI10669
aI4233
aI33379
aI72465
aI256861
asS'd'
I18090
sa(dp36180
S'a'
I78535
sS's'
I18
sS'm'
(lp36181
I1
aI1
aI7
aI5
aI27
aI55
aI17
aI75
aI373
aI325
aI1981
aI1743
aI7341
aI319
aI28169
aI3587
aI66057
aI169723
asS'd'
I18091
sa(dp36182
S'a'
I78542
sS's'
I18
sS'm'
(lp36183
I1
aI3
aI3
aI15
aI27
aI31
aI47
aI91
aI367
aI245
aI2045
aI979
aI2169
aI10935
aI29523
aI64871
aI119447
aI92131
asS'd'
I18092
sa(dp36184
S'a'
I78549
sS's'
I18
sS'm'
(lp36185
I1
aI3
aI7
aI15
aI9
aI11
aI93
aI61
aI249
aI107
aI1883
aI2547
aI375
aI4195
aI6451
aI14533
aI62529
aI93557
asS'd'
I18093
sa(dp36186
S'a'
I78554
sS's'
I18
sS'm'
(lp36187
I1
aI1
aI1
aI3
aI29
aI61
aI65
aI155
aI301
aI1017
aI131
aI1567
aI3649
aI3447
aI27943
aI52111
aI9133
aI88147
asS'd'
I18094
sa(dp36188
S'a'
I78556
sS's'
I18
sS'm'
(lp36189
I1
aI1
aI1
aI1
aI21
aI59
aI107
aI151
aI265
aI707
aI767
aI2325
aI8095
aI14027
aI15355
aI15465
aI83143
aI116199
asS'd'
I18095
sa(dp36190
S'a'
I78583
sS's'
I18
sS'm'
(lp36191
I1
aI3
aI1
aI15
aI23
aI51
aI31
aI25
aI439
aI357
aI1563
aI1091
aI2135
aI1327
aI18427
aI60965
aI29215
aI157351
asS'd'
I18096
sa(dp36192
S'a'
I78590
sS's'
I18
sS'm'
(lp36193
I1
aI3
aI3
aI13
aI29
aI37
aI25
aI215
aI149
aI487
aI703
aI1787
aI3641
aI8301
aI8795
aI13845
aI95245
aI169793
asS'd'
I18097
sa(dp36194
S'a'
I78615
sS's'
I18
sS'm'
(lp36195
I1
aI3
aI3
aI11
aI27
aI3
aI49
aI87
aI69
aI687
aI1181
aI3405
aI589
aI12901
aI14199
aI48607
aI74027
aI181379
asS'd'
I18098
sa(dp36196
S'a'
I78635
sS's'
I18
sS'm'
(lp36197
I1
aI3
aI5
aI13
aI9
aI15
aI33
aI229
aI135
aI769
aI1005
aI2435
aI4831
aI5493
aI16745
aI64379
aI20253
aI52661
asS'd'
I18099
sa(dp36198
S'a'
I78649
sS's'
I18
sS'm'
(lp36199
I1
aI1
aI1
aI13
aI9
aI61
aI33
aI127
aI339
aI15
aI945
aI219
aI4291
aI6995
aI29127
aI61853
aI40741
aI170541
asS'd'
I18100
sa(dp36200
S'a'
I78684
sS's'
I18
sS'm'
(lp36201
I1
aI3
aI3
aI15
aI9
aI33
aI75
aI39
aI327
aI133
aI733
aI1125
aI2747
aI15031
aI24575
aI65013
aI41997
aI158679
asS'd'
I18101
sa(dp36202
S'a'
I78691
sS's'
I18
sS'm'
(lp36203
I1
aI1
aI3
aI9
aI3
aI9
aI63
aI83
aI493
aI175
aI249
aI1977
aI8177
aI4067
aI2131
aI12467
aI86185
aI73417
asS'd'
I18102
sa(dp36204
S'a'
I78705
sS's'
I18
sS'm'
(lp36205
I1
aI1
aI3
aI13
aI29
aI55
aI91
aI109
aI73
aI913
aI1343
aI2147
aI105
aI8763
aI7613
aI55749
aI4339
aI61253
asS'd'
I18103
sa(dp36206
S'a'
I78724
sS's'
I18
sS'm'
(lp36207
I1
aI1
aI5
aI5
aI17
aI19
aI45
aI57
aI345
aI835
aI341
aI1365
aI5187
aI7485
aI22685
aI32321
aI67279
aI141119
asS'd'
I18104
sa(dp36208
S'a'
I78755
sS's'
I18
sS'm'
(lp36209
I1
aI1
aI3
aI11
aI9
aI47
aI11
aI231
aI241
aI681
aI255
aI3663
aI5547
aI997
aI2445
aI64413
aI55349
aI61785
asS'd'
I18105
sa(dp36210
S'a'
I78770
sS's'
I18
sS'm'
(lp36211
I1
aI3
aI5
aI5
aI23
aI29
aI23
aI249
aI149
aI1011
aI173
aI271
aI485
aI1239
aI81
aI59277
aI96669
aI210859
asS'd'
I18106
sa(dp36212
S'a'
I78772
sS's'
I18
sS'm'
(lp36213
I1
aI3
aI3
aI1
aI17
aI9
aI41
aI39
aI309
aI131
aI1431
aI1497
aI1669
aI14191
aI22795
aI48951
aI101731
aI70847
asS'd'
I18107
sa(dp36214
S'a'
I78818
sS's'
I18
sS'm'
(lp36215
I1
aI1
aI3
aI15
aI1
aI11
aI37
aI79
aI23
aI1023
aI585
aI127
aI7817
aI15009
aI3897
aI44601
aI83039
aI240457
asS'd'
I18108
sa(dp36216
S'a'
I78837
sS's'
I18
sS'm'
(lp36217
I1
aI3
aI5
aI9
aI21
aI33
aI55
aI31
aI193
aI745
aI1741
aI3637
aI7265
aI8969
aI11797
aI33239
aI29123
aI126077
asS'd'
I18109
sa(dp36218
S'a'
I78844
sS's'
I18
sS'm'
(lp36219
I1
aI3
aI3
aI13
aI31
aI5
aI87
aI215
aI271
aI573
aI1423
aI2611
aI947
aI14669
aI23785
aI60579
aI127099
aI55877
asS'd'
I18110
sa(dp36220
S'a'
I78849
sS's'
I18
sS'm'
(lp36221
I1
aI3
aI1
aI13
aI5
aI53
aI103
aI85
aI237
aI457
aI739
aI1201
aI133
aI8589
aI13471
aI6707
aI42257
aI141989
asS'd'
I18111
sa(dp36222
S'a'
I78909
sS's'
I18
sS'm'
(lp36223
I1
aI1
aI1
aI5
aI23
aI3
aI65
aI159
aI445
aI823
aI341
aI1723
aI6263
aI9421
aI16023
aI19145
aI52337
aI229397
asS'd'
I18112
sa(dp36224
S'a'
I78915
sS's'
I18
sS'm'
(lp36225
I1
aI3
aI5
aI3
aI15
aI3
aI15
aI251
aI407
aI137
aI951
aI1319
aI1035
aI7713
aI29579
aI19591
aI77841
aI84949
asS'd'
I18113
sa(dp36226
S'a'
I78941
sS's'
I18
sS'm'
(lp36227
I1
aI1
aI7
aI15
aI19
aI25
aI63
aI141
aI511
aI11
aI1027
aI1209
aI6627
aI8127
aI14879
aI12965
aI109973
aI144501
asS'd'
I18114
sa(dp36228
S'a'
I78958
sS's'
I18
sS'm'
(lp36229
I1
aI1
aI1
aI3
aI11
aI57
aI65
aI169
aI453
aI197
aI1249
aI2933
aI3743
aI1971
aI19373
aI32109
aI73265
aI46185
asS'd'
I18115
sa(dp36230
S'a'
I78975
sS's'
I18
sS'm'
(lp36231
I1
aI1
aI3
aI1
aI3
aI1
aI21
aI47
aI471
aI565
aI1795
aI1771
aI3187
aI7189
aI18627
aI22993
aI112319
aI158693
asS'd'
I18116
sa(dp36232
S'a'
I78979
sS's'
I18
sS'm'
(lp36233
I1
aI1
aI5
aI7
aI5
aI25
aI127
aI113
aI31
aI609
aI1273
aI2799
aI5713
aI16091
aI22239
aI43617
aI126003
aI218991
asS'd'
I18117
sa(dp36234
S'a'
I78986
sS's'
I18
sS'm'
(lp36235
I1
aI3
aI3
aI7
aI19
aI59
aI19
aI185
aI483
aI431
aI335
aI565
aI819
aI2555
aI18653
aI36573
aI50085
aI31007
asS'd'
I18118
sa(dp36236
S'a'
I79029
sS's'
I18
sS'm'
(lp36237
I1
aI1
aI3
aI13
aI17
aI61
aI5
aI219
aI297
aI755
aI2005
aI391
aI4927
aI1517
aI11341
aI9527
aI51739
aI182599
asS'd'
I18119
sa(dp36238
S'a'
I79030
sS's'
I18
sS'm'
(lp36239
I1
aI3
aI7
aI9
aI9
aI3
aI39
aI211
aI475
aI717
aI189
aI819
aI529
aI469
aI28559
aI7321
aI60213
aI79505
asS'd'
I18120
sa(dp36240
S'a'
I79044
sS's'
I18
sS'm'
(lp36241
I1
aI3
aI1
aI9
aI17
aI39
aI53
aI65
aI247
aI145
aI9
aI1669
aI7221
aI8359
aI11021
aI29775
aI24693
aI208655
asS'd'
I18121
sa(dp36242
S'a'
I79048
sS's'
I18
sS'm'
(lp36243
I1
aI1
aI5
aI13
aI7
aI7
aI31
aI135
aI375
aI439
aI1419
aI3579
aI4313
aI14057
aI31505
aI55249
aI5345
aI69537
asS'd'
I18122
sa(dp36244
S'a'
I79056
sS's'
I18
sS'm'
(lp36245
I1
aI3
aI5
aI9
aI21
aI3
aI125
aI223
aI9
aI73
aI1693
aI281
aI3941
aI10377
aI29365
aI19807
aI73973
aI169113
asS'd'
I18123
sa(dp36246
S'a'
I79095
sS's'
I18
sS'm'
(lp36247
I1
aI3
aI7
aI15
aI29
aI41
aI119
aI75
aI241
aI79
aI1969
aI1091
aI6241
aI10685
aI11579
aI3791
aI124443
aI5051
asS'd'
I18124
sa(dp36248
S'a'
I79099
sS's'
I18
sS'm'
(lp36249
I1
aI3
aI7
aI15
aI23
aI53
aI13
aI255
aI205
aI547
aI255
aI1589
aI7261
aI15735
aI14521
aI29679
aI109373
aI236433
asS'd'
I18125
sa(dp36250
S'a'
I79121
sS's'
I18
sS'm'
(lp36251
I1
aI3
aI7
aI3
aI17
aI37
aI71
aI163
aI95
aI265
aI1
aI3239
aI1779
aI9047
aI31387
aI32291
aI86741
aI55317
asS'd'
I18126
sa(dp36252
S'a'
I79150
sS's'
I18
sS'm'
(lp36253
I1
aI3
aI1
aI9
aI31
aI55
aI117
aI247
aI317
aI673
aI749
aI1155
aI7743
aI6427
aI25273
aI49701
aI62345
aI20913
asS'd'
I18127
sa(dp36254
S'a'
I79196
sS's'
I18
sS'm'
(lp36255
I1
aI3
aI3
aI7
aI27
aI55
aI35
aI111
aI69
aI799
aI213
aI3011
aI4359
aI14763
aI7387
aI13281
aI58397
aI38415
asS'd'
I18128
sa(dp36256
S'a'
I79199
sS's'
I18
sS'm'
(lp36257
I1
aI1
aI5
aI9
aI5
aI61
aI49
aI219
aI419
aI297
aI1019
aI2181
aI6069
aI12957
aI24637
aI23317
aI6389
aI240893
asS'd'
I18129
sa(dp36258
S'a'
I79220
sS's'
I18
sS'm'
(lp36259
I1
aI1
aI5
aI15
aI13
aI57
aI59
aI43
aI373
aI647
aI1407
aI3955
aI5583
aI15229
aI20935
aI38007
aI65971
aI95987
asS'd'
I18130
sa(dp36260
S'a'
I79229
sS's'
I18
sS'm'
(lp36261
I1
aI1
aI7
aI7
aI23
aI17
aI77
aI91
aI449
aI75
aI1059
aI3337
aI2041
aI261
aI25077
aI28161
aI44537
aI189443
asS'd'
I18131
sa(dp36262
S'a'
I79263
sS's'
I18
sS'm'
(lp36263
I1
aI1
aI7
aI11
aI9
aI7
aI117
aI225
aI457
aI941
aI161
aI1825
aI1101
aI193
aI32619
aI37245
aI102633
aI86707
asS'd'
I18132
sa(dp36264
S'a'
I79264
sS's'
I18
sS'm'
(lp36265
I1
aI1
aI1
aI7
aI13
aI43
aI33
aI137
aI275
aI691
aI1387
aI1265
aI759
aI1457
aI4877
aI41813
aI4159
aI234397
asS'd'
I18133
sa(dp36266
S'a'
I79296
sS's'
I18
sS'm'
(lp36267
I1
aI3
aI3
aI1
aI9
aI23
aI71
aI39
aI205
aI175
aI953
aI2965
aI3283
aI6025
aI5905
aI34691
aI120987
aI71841
asS'd'
I18134
sa(dp36268
S'a'
I79302
sS's'
I18
sS'm'
(lp36269
I1
aI3
aI1
aI13
aI31
aI63
aI49
aI73
aI299
aI169
aI1265
aI2205
aI1299
aI10045
aI6919
aI26067
aI56909
aI42549
asS'd'
I18135
sa(dp36270
S'a'
I79354
sS's'
I18
sS'm'
(lp36271
I1
aI3
aI3
aI1
aI31
aI41
aI75
aI219
aI457
aI407
aI5
aI1901
aI6823
aI531
aI3155
aI64375
aI38523
aI68217
asS'd'
I18136
sa(dp36272
S'a'
I79387
sS's'
I18
sS'm'
(lp36273
I1
aI1
aI7
aI9
aI7
aI35
aI123
aI193
aI145
aI1021
aI757
aI3775
aI2313
aI11885
aI11649
aI61071
aI129363
aI120467
asS'd'
I18137
sa(dp36274
S'a'
I79399
sS's'
I18
sS'm'
(lp36275
I1
aI3
aI7
aI3
aI29
aI21
aI127
aI93
aI415
aI641
aI453
aI923
aI7713
aI9569
aI5961
aI25969
aI31095
aI93317
asS'd'
I18138
sa(dp36276
S'a'
I79435
sS's'
I18
sS'm'
(lp36277
I1
aI3
aI3
aI1
aI5
aI15
aI21
aI235
aI211
aI663
aI385
aI2429
aI319
aI11571
aI17539
aI42975
aI43179
aI100105
asS'd'
I18139
sa(dp36278
S'a'
I79525
sS's'
I18
sS'm'
(lp36279
I1
aI3
aI7
aI7
aI25
aI57
aI51
aI215
aI393
aI167
aI1569
aI3235
aI5555
aI3391
aI2389
aI36485
aI21919
aI164479
asS'd'
I18140
sa(dp36280
S'a'
I79540
sS's'
I18
sS'm'
(lp36281
I1
aI3
aI3
aI3
aI29
aI21
aI81
aI59
aI239
aI671
aI605
aI583
aI2341
aI2321
aI28593
aI19035
aI10209
aI36433
asS'd'
I18141
sa(dp36282
S'a'
I79552
sS's'
I18
sS'm'
(lp36283
I1
aI3
aI3
aI11
aI31
aI33
aI1
aI147
aI111
aI523
aI427
aI3545
aI111
aI8009
aI29101
aI34549
aI122745
aI82117
asS'd'
I18142
sa(dp36284
S'a'
I79562
sS's'
I18
sS'm'
(lp36285
I1
aI3
aI5
aI15
aI19
aI37
aI97
aI141
aI387
aI523
aI467
aI1657
aI4161
aI5505
aI18091
aI39597
aI124423
aI74827
asS'd'
I18143
sa(dp36286
S'a'
I79576
sS's'
I18
sS'm'
(lp36287
I1
aI1
aI1
aI11
aI21
aI63
aI61
aI13
aI169
aI851
aI1863
aI3307
aI7189
aI10791
aI22619
aI24431
aI127781
aI14717
asS'd'
I18144
sa(dp36288
S'a'
I79579
sS's'
I18
sS'm'
(lp36289
I1
aI1
aI3
aI13
aI27
aI41
aI69
aI127
aI497
aI565
aI1489
aI277
aI2551
aI15409
aI9885
aI187
aI101319
aI194121
asS'd'
I18145
sa(dp36290
S'a'
I79605
sS's'
I18
sS'm'
(lp36291
I1
aI1
aI7
aI7
aI17
aI45
aI1
aI139
aI347
aI503
aI1189
aI1459
aI6117
aI14319
aI22153
aI2915
aI91991
aI246679
asS'd'
I18146
sa(dp36292
S'a'
I79618
sS's'
I18
sS'm'
(lp36293
I1
aI3
aI3
aI3
aI9
aI41
aI25
aI199
aI327
aI295
aI945
aI2765
aI563
aI11605
aI24267
aI37729
aI80057
aI169479
asS'd'
I18147
sa(dp36294
S'a'
I79648
sS's'
I18
sS'm'
(lp36295
I1
aI1
aI7
aI3
aI23
aI19
aI13
aI219
aI235
aI837
aI1015
aI2071
aI2727
aI3989
aI32539
aI26713
aI112391
aI163943
asS'd'
I18148
sa(dp36296
S'a'
I79654
sS's'
I18
sS'm'
(lp36297
I1
aI3
aI3
aI9
aI21
aI27
aI17
aI187
aI315
aI753
aI817
aI3053
aI5961
aI973
aI23973
aI37621
aI105637
aI247711
asS'd'
I18149
sa(dp36298
S'a'
I79666
sS's'
I18
sS'm'
(lp36299
I1
aI1
aI7
aI1
aI11
aI15
aI45
aI25
aI421
aI213
aI663
aI3829
aI469
aI15889
aI28773
aI14323
aI107705
aI111729
asS'd'
I18150
sa(dp36300
S'a'
I79686
sS's'
I18
sS'm'
(lp36301
I1
aI1
aI1
aI7
aI7
aI7
aI51
aI189
aI457
aI95
aI1903
aI639
aI1933
aI7409
aI22327
aI18959
aI42679
aI158987
asS'd'
I18151
sa(dp36302
S'a'
I79697
sS's'
I18
sS'm'
(lp36303
I1
aI1
aI5
aI9
aI13
aI13
aI49
aI159
aI387
aI365
aI1799
aI2399
aI6375
aI14965
aI32495
aI5383
aI73479
aI5653
asS'd'
I18152
sa(dp36304
S'a'
I79700
sS's'
I18
sS'm'
(lp36305
I1
aI1
aI3
aI1
aI29
aI23
aI81
aI73
aI183
aI563
aI435
aI133
aI5731
aI6663
aI21219
aI60007
aI101215
aI68775
asS'd'
I18153
sa(dp36306
S'a'
I79723
sS's'
I18
sS'm'
(lp36307
I1
aI1
aI7
aI11
aI31
aI47
aI43
aI159
aI221
aI745
aI1317
aI2405
aI4563
aI4073
aI27675
aI14225
aI114231
aI222553
asS'd'
I18154
sa(dp36308
S'a'
I79749
sS's'
I18
sS'm'
(lp36309
I1
aI1
aI1
aI5
aI11
aI63
aI105
aI99
aI413
aI81
aI771
aI547
aI1633
aI8097
aI30431
aI31417
aI101379
aI163575
asS'd'
I18155
sa(dp36310
S'a'
I79750
sS's'
I18
sS'm'
(lp36311
I1
aI1
aI3
aI9
aI23
aI29
aI123
aI149
aI241
aI267
aI1925
aI467
aI7743
aI4473
aI12223
aI10521
aI86265
aI89949
asS'd'
I18156
sa(dp36312
S'a'
I79764
sS's'
I18
sS'm'
(lp36313
I1
aI3
aI5
aI1
aI31
aI29
aI111
aI67
aI311
aI851
aI1919
aI2563
aI3725
aI4035
aI7241
aI13859
aI105207
aI200599
asS'd'
I18157
sa(dp36314
S'a'
I79771
sS's'
I18
sS'm'
(lp36315
I1
aI3
aI7
aI3
aI19
aI53
aI113
aI107
aI133
aI243
aI2021
aI2669
aI4633
aI14393
aI24827
aI1233
aI81471
aI20105
asS'd'
I18158
sa(dp36316
S'a'
I79774
sS's'
I18
sS'm'
(lp36317
I1
aI1
aI1
aI5
aI3
aI23
aI43
aI149
aI157
aI875
aI1175
aI963
aI6189
aI7343
aI13913
aI41375
aI112857
aI236047
asS'd'
I18159
sa(dp36318
S'a'
I79780
sS's'
I18
sS'm'
(lp36319
I1
aI3
aI5
aI15
aI11
aI31
aI43
aI225
aI469
aI229
aI703
aI3033
aI2341
aI10309
aI12057
aI13325
aI109019
aI130789
asS'd'
I18160
sa(dp36320
S'a'
I79789
sS's'
I18
sS'm'
(lp36321
I1
aI1
aI1
aI7
aI27
aI47
aI45
aI49
aI371
aI971
aI1121
aI2179
aI1267
aI9499
aI10771
aI28781
aI77059
aI90765
asS'd'
I18161
sa(dp36322
S'a'
I79798
sS's'
I18
sS'm'
(lp36323
I1
aI1
aI7
aI1
aI17
aI27
aI59
aI169
aI269
aI217
aI983
aI1365
aI1985
aI12287
aI5385
aI46407
aI24827
aI155761
asS'd'
I18162
sa(dp36324
S'a'
I79821
sS's'
I18
sS'm'
(lp36325
I1
aI1
aI7
aI11
aI9
aI5
aI19
aI205
aI159
aI937
aI763
aI3823
aI3625
aI14209
aI32031
aI58879
aI118449
aI50723
asS'd'
I18163
sa(dp36326
S'a'
I79850
sS's'
I18
sS'm'
(lp36327
I1
aI1
aI5
aI3
aI25
aI55
aI27
aI35
aI125
aI999
aI1541
aI3883
aI539
aI5691
aI18071
aI63199
aI112089
aI194825
asS'd'
I18164
sa(dp36328
S'a'
I79864
sS's'
I18
sS'm'
(lp36329
I1
aI3
aI3
aI1
aI27
aI43
aI57
aI225
aI173
aI673
aI1339
aI3433
aI5743
aI1375
aI32429
aI35071
aI98035
aI229973
asS'd'
I18165
sa(dp36330
S'a'
I79898
sS's'
I18
sS'm'
(lp36331
I1
aI3
aI3
aI9
aI3
aI51
aI5
aI203
aI439
aI41
aI529
aI863
aI6735
aI13211
aI7075
aI55637
aI24481
aI46673
asS'd'
I18166
sa(dp36332
S'a'
I79904
sS's'
I18
sS'm'
(lp36333
I1
aI1
aI5
aI11
aI15
aI23
aI93
aI7
aI181
aI843
aI777
aI1299
aI1941
aI7147
aI26253
aI10967
aI5387
aI84611
asS'd'
I18167
sa(dp36334
S'a'
I79934
sS's'
I18
sS'm'
(lp36335
I1
aI3
aI1
aI7
aI9
aI57
aI127
aI155
aI257
aI423
aI1421
aI261
aI4477
aI11169
aI22997
aI12371
aI8705
aI135883
asS'd'
I18168
sa(dp36336
S'a'
I79936
sS's'
I18
sS'm'
(lp36337
I1
aI3
aI1
aI9
aI17
aI9
aI15
aI209
aI427
aI889
aI1939
aI3623
aI2587
aI4037
aI32233
aI40391
aI32529
aI63851
asS'd'
I18169
sa(dp36338
S'a'
I79942
sS's'
I18
sS'm'
(lp36339
I1
aI3
aI3
aI13
aI3
aI19
aI49
aI155
aI213
aI239
aI817
aI1787
aI2999
aI9955
aI20155
aI44711
aI41367
aI59623
asS'd'
I18170
sa(dp36340
S'a'
I79945
sS's'
I18
sS'm'
(lp36341
I1
aI3
aI7
aI5
aI5
aI39
aI103
aI181
aI405
aI85
aI1997
aI3639
aI1259
aI10737
aI189
aI44377
aI23589
aI89371
asS'd'
I18171
sa(dp36342
S'a'
I79963
sS's'
I18
sS'm'
(lp36343
I1
aI3
aI5
aI11
aI15
aI13
aI57
aI81
aI203
aI773
aI1571
aI3235
aI6625
aI13803
aI2091
aI64265
aI131013
aI189705
asS'd'
I18172
sa(dp36344
S'a'
I79987
sS's'
I18
sS'm'
(lp36345
I1
aI1
aI1
aI13
aI15
aI3
aI113
aI159
aI149
aI55
aI355
aI2345
aI5043
aI4067
aI23277
aI32647
aI43755
aI5445
asS'd'
I18173
sa(dp36346
S'a'
I80057
sS's'
I18
sS'm'
(lp36347
I1
aI3
aI3
aI9
aI31
aI7
aI67
aI177
aI423
aI269
aI1731
aI3957
aI4383
aI13483
aI14653
aI8243
aI57689
aI37375
asS'd'
I18174
sa(dp36348
S'a'
I80077
sS's'
I18
sS'm'
(lp36349
I1
aI1
aI3
aI5
aI25
aI5
aI77
aI199
aI161
aI859
aI497
aI1679
aI6809
aI4877
aI1107
aI16443
aI15505
aI138155
asS'd'
I18175
sa(dp36350
S'a'
I80078
sS's'
I18
sS'm'
(lp36351
I1
aI3
aI5
aI1
aI11
aI57
aI7
aI49
aI145
aI569
aI571
aI2679
aI7531
aI14517
aI12425
aI6285
aI116961
aI116397
asS'd'
I18176
sa(dp36352
S'a'
I80080
sS's'
I18
sS'm'
(lp36353
I1
aI1
aI7
aI11
aI1
aI37
aI65
aI43
aI151
aI419
aI801
aI3231
aI5321
aI10725
aI12885
aI62771
aI16507
aI179009
asS'd'
I18177
sa(dp36354
S'a'
I80102
sS's'
I18
sS'm'
(lp36355
I1
aI1
aI5
aI11
aI29
aI55
aI89
aI81
aI325
aI47
aI1037
aI3235
aI2017
aI10875
aI8919
aI25115
aI118035
aI178227
asS'd'
I18178
sa(dp36356
S'a'
I80106
sS's'
I18
sS'm'
(lp36357
I1
aI3
aI1
aI7
aI1
aI43
aI101
aI25
aI449
aI617
aI381
aI3437
aI6655
aI1291
aI18693
aI53939
aI99143
aI195695
asS'd'
I18179
sa(dp36358
S'a'
I80111
sS's'
I18
sS'm'
(lp36359
I1
aI3
aI1
aI5
aI23
aI7
aI47
aI159
aI295
aI939
aI173
aI3087
aI1497
aI6353
aI13893
aI13465
aI118973
aI193737
asS'd'
I18180
sa(dp36360
S'a'
I80152
sS's'
I18
sS'm'
(lp36361
I1
aI1
aI3
aI9
aI3
aI41
aI65
aI79
aI449
aI345
aI2039
aI1193
aI5915
aI13689
aI1257
aI23273
aI48515
aI256793
asS'd'
I18181
sa(dp36362
S'a'
I80155
sS's'
I18
sS'm'
(lp36363
I1
aI3
aI5
aI11
aI11
aI55
aI13
aI117
aI343
aI899
aI1853
aI373
aI6885
aI12863
aI1209
aI34433
aI48215
aI218187
asS'd'
I18182
sa(dp36364
S'a'
I80173
sS's'
I18
sS'm'
(lp36365
I1
aI3
aI7
aI7
aI3
aI45
aI103
aI145
aI55
aI507
aI743
aI4027
aI2075
aI15707
aI4473
aI50077
aI64551
aI204305
asS'd'
I18183
sa(dp36366
S'a'
I80186
sS's'
I18
sS'm'
(lp36367
I1
aI1
aI3
aI5
aI31
aI45
aI123
aI233
aI363
aI1003
aI411
aI1459
aI6455
aI985
aI29451
aI17625
aI44153
aI137097
asS'd'
I18184
sa(dp36368
S'a'
I80203
sS's'
I18
sS'm'
(lp36369
I1
aI3
aI3
aI1
aI27
aI11
aI53
aI251
aI41
aI43
aI495
aI107
aI6145
aI8785
aI28997
aI7181
aI92903
aI105785
asS'd'
I18185
sa(dp36370
S'a'
I80223
sS's'
I18
sS'm'
(lp36371
I1
aI1
aI3
aI11
aI13
aI5
aI117
aI141
aI463
aI639
aI1857
aI2873
aI3627
aI6081
aI18207
aI29451
aI80909
aI73557
asS'd'
I18186
sa(dp36372
S'a'
I80233
sS's'
I18
sS'm'
(lp36373
I1
aI1
aI1
aI3
aI29
aI51
aI15
aI81
aI85
aI487
aI307
aI2481
aI2769
aI14901
aI9407
aI58321
aI52813
aI230393
asS'd'
I18187
sa(dp36374
S'a'
I80258
sS's'
I18
sS'm'
(lp36375
I1
aI3
aI3
aI11
aI31
aI7
aI107
aI43
aI205
aI811
aI1121
aI2757
aI2447
aI6843
aI21347
aI9143
aI41003
aI80507
asS'd'
I18188
sa(dp36376
S'a'
I80281
sS's'
I18
sS'm'
(lp36377
I1
aI1
aI1
aI1
aI29
aI19
aI13
aI203
aI47
aI689
aI2003
aI1477
aI7857
aI5031
aI21781
aI5745
aI3649
aI160389
asS'd'
I18189
sa(dp36378
S'a'
I80318
sS's'
I18
sS'm'
(lp36379
I1
aI3
aI3
aI7
aI21
aI21
aI65
aI3
aI351
aI157
aI167
aI3425
aI2395
aI9165
aI26143
aI57221
aI127171
aI54461
asS'd'
I18190
sa(dp36380
S'a'
I80326
sS's'
I18
sS'm'
(lp36381
I1
aI3
aI1
aI15
aI13
aI13
aI65
aI53
aI305
aI719
aI181
aI709
aI5485
aI13385
aI30287
aI52669
aI82647
aI83851
asS'd'
I18191
sa(dp36382
S'a'
I80330
sS's'
I18
sS'm'
(lp36383
I1
aI1
aI3
aI11
aI11
aI23
aI31
aI109
aI205
aI123
aI509
aI3831
aI7771
aI7341
aI31613
aI28035
aI38061
aI49375
asS'd'
I18192
sa(dp36384
S'a'
I80337
sS's'
I18
sS'm'
(lp36385
I1
aI3
aI3
aI3
aI15
aI33
aI47
aI159
aI321
aI589
aI393
aI3253
aI3743
aI6161
aI445
aI33129
aI8181
aI27793
asS'd'
I18193
sa(dp36386
S'a'
I80344
sS's'
I18
sS'm'
(lp36387
I1
aI1
aI1
aI13
aI9
aI57
aI111
aI253
aI203
aI539
aI673
aI855
aI1937
aI2699
aI25795
aI6889
aI13531
aI63561
asS'd'
I18194
sa(dp36388
S'a'
I80365
sS's'
I18
sS'm'
(lp36389
I1
aI1
aI1
aI13
aI31
aI45
aI13
aI101
aI113
aI903
aI1699
aI2423
aI7967
aI7957
aI20303
aI64395
aI124447
aI33947
asS'd'
I18195
sa(dp36390
S'a'
I80383
sS's'
I18
sS'm'
(lp36391
I1
aI3
aI5
aI11
aI17
aI39
aI59
aI181
aI421
aI535
aI1445
aI3927
aI5433
aI12885
aI12497
aI47231
aI39819
aI46371
asS'd'
I18196
sa(dp36392
S'a'
I80389
sS's'
I18
sS'm'
(lp36393
I1
aI3
aI5
aI7
aI27
aI3
aI75
aI49
aI461
aI781
aI433
aI1767
aI6903
aI11907
aI2063
aI55199
aI82823
aI229405
asS'd'
I18197
sa(dp36394
S'a'
I80413
sS's'
I18
sS'm'
(lp36395
I1
aI3
aI3
aI15
aI17
aI61
aI17
aI23
aI247
aI683
aI33
aI4027
aI341
aI8069
aI2529
aI9757
aI95653
aI12927
asS'd'
I18198
sa(dp36396
S'a'
I80441
sS's'
I18
sS'm'
(lp36397
I1
aI3
aI3
aI11
aI17
aI7
aI29
aI205
aI353
aI917
aI219
aI3509
aI7803
aI5939
aI25111
aI45357
aI9259
aI1549
asS'd'
I18199
sa(dp36398
S'a'
I80449
sS's'
I18
sS'm'
(lp36399
I1
aI3
aI3
aI15
aI21
aI7
aI23
aI25
aI459
aI291
aI31
aI2091
aI1177
aI9311
aI12231
aI16617
aI33575
aI252643
asS'd'
I18200
sa(dp36400
S'a'
I80461
sS's'
I18
sS'm'
(lp36401
I1
aI3
aI5
aI5
aI3
aI51
aI113
aI123
aI453
aI503
aI1575
aI2785
aI5011
aI1789
aI819
aI30857
aI12955
aI172421
asS'd'
I18201
sa(dp36402
S'a'
I80467
sS's'
I18
sS'm'
(lp36403
I1
aI1
aI5
aI3
aI15
aI15
aI125
aI65
aI113
aI281
aI53
aI3417
aI5279
aI6351
aI25931
aI54835
aI124077
aI204241
asS'd'
I18202
sa(dp36404
S'a'
I80476
sS's'
I18
sS'm'
(lp36405
I1
aI1
aI3
aI9
aI29
aI31
aI19
aI179
aI275
aI933
aI711
aI3351
aI6221
aI1711
aI9375
aI11645
aI118911
aI249395
asS'd'
I18203
sa(dp36406
S'a'
I80507
sS's'
I18
sS'm'
(lp36407
I1
aI3
aI7
aI13
aI23
aI59
aI43
aI61
aI85
aI267
aI691
aI3949
aI2135
aI3203
aI21455
aI61895
aI71157
aI136739
asS'd'
I18204
sa(dp36408
S'a'
I80516
sS's'
I18
sS'm'
(lp36409
I1
aI1
aI7
aI5
aI19
aI27
aI69
aI141
aI9
aI633
aI95
aI3789
aI7823
aI12635
aI27661
aI30285
aI129469
aI67163
asS'd'
I18205
sa(dp36410
S'a'
I80519
sS's'
I18
sS'm'
(lp36411
I1
aI3
aI3
aI9
aI11
aI25
aI103
aI47
aI425
aI809
aI1279
aI411
aI219
aI6703
aI24145
aI17303
aI56835
aI84879
asS'd'
I18206
sa(dp36412
S'a'
I80568
sS's'
I18
sS'm'
(lp36413
I1
aI1
aI5
aI13
aI29
aI41
aI47
aI133
aI197
aI615
aI169
aI2157
aI1795
aI4945
aI31693
aI57763
aI39369
aI83353
asS'd'
I18207
sa(dp36414
S'a'
I80571
sS's'
I18
sS'm'
(lp36415
I1
aI3
aI1
aI3
aI27
aI23
aI23
aI213
aI387
aI239
aI977
aI221
aI383
aI11005
aI7221
aI8795
aI100963
aI163777
asS'd'
I18208
sa(dp36416
S'a'
I80579
sS's'
I18
sS'm'
(lp36417
I1
aI1
aI1
aI1
aI31
aI29
aI87
aI93
aI239
aI399
aI801
aI3143
aI6973
aI16331
aI16865
aI1823
aI1127
aI41983
asS'd'
I18209
sa(dp36418
S'a'
I80586
sS's'
I18
sS'm'
(lp36419
I1
aI1
aI3
aI13
aI7
aI39
aI25
aI251
aI277
aI417
aI119
aI3033
aI6785
aI9783
aI1641
aI60169
aI25047
aI182263
asS'd'
I18210
sa(dp36420
S'a'
I80599
sS's'
I18
sS'm'
(lp36421
I1
aI1
aI5
aI5
aI7
aI35
aI17
aI47
aI295
aI861
aI1671
aI1971
aI4583
aI3925
aI31013
aI50039
aI125191
aI143019
asS'd'
I18211
sa(dp36422
S'a'
I80610
sS's'
I18
sS'm'
(lp36423
I1
aI1
aI5
aI1
aI3
aI57
aI11
aI23
aI273
aI209
aI617
aI1499
aI665
aI1193
aI7539
aI1625
aI48065
aI82843
asS'd'
I18212
sa(dp36424
S'a'
I80629
sS's'
I18
sS'm'
(lp36425
I1
aI1
aI3
aI15
aI15
aI17
aI39
aI145
aI193
aI503
aI1305
aI2071
aI93
aI11529
aI14267
aI14779
aI49327
aI51347
asS'd'
I18213
sa(dp36426
S'a'
I80642
sS's'
I18
sS'm'
(lp36427
I1
aI3
aI5
aI3
aI7
aI39
aI63
aI171
aI263
aI493
aI383
aI3209
aI4277
aI6259
aI1345
aI48013
aI110571
aI127865
asS'd'
I18214
sa(dp36428
S'a'
I80690
sS's'
I18
sS'm'
(lp36429
I1
aI3
aI1
aI7
aI15
aI29
aI93
aI75
aI37
aI235
aI1095
aI153
aI745
aI9785
aI28831
aI58899
aI67091
aI34743
asS'd'
I18215
sa(dp36430
S'a'
I80713
sS's'
I18
sS'm'
(lp36431
I1
aI3
aI7
aI9
aI27
aI23
aI67
aI85
aI491
aI447
aI1899
aI709
aI555
aI13979
aI12529
aI38383
aI16091
aI117301
asS'd'
I18216
sa(dp36432
S'a'
I80716
sS's'
I18
sS'm'
(lp36433
I1
aI1
aI5
aI3
aI9
aI55
aI109
aI173
aI29
aI19
aI1265
aI2391
aI7761
aI1953
aI5643
aI24079
aI14187
aI127017
asS'd'
I18217
sa(dp36434
S'a'
I80737
sS's'
I18
sS'm'
(lp36435
I1
aI3
aI7
aI13
aI21
aI57
aI105
aI145
aI73
aI421
aI403
aI5
aI3523
aI7005
aI1109
aI63357
aI111671
aI191857
asS'd'
I18218
sa(dp36436
S'a'
I80743
sS's'
I18
sS'm'
(lp36437
I1
aI1
aI7
aI13
aI5
aI27
aI21
aI5
aI199
aI515
aI917
aI365
aI2775
aI12453
aI26989
aI60593
aI98977
aI161759
asS'd'
I18219
sa(dp36438
S'a'
I80750
sS's'
I18
sS'm'
(lp36439
I1
aI3
aI1
aI13
aI15
aI37
aI71
aI65
aI27
aI533
aI1311
aI2981
aI1945
aI7183
aI5337
aI20659
aI67355
aI185633
asS'd'
I18220
sa(dp36440
S'a'
I80786
sS's'
I18
sS'm'
(lp36441
I1
aI1
aI5
aI7
aI21
aI39
aI21
aI195
aI443
aI979
aI1033
aI1823
aI3045
aI3023
aI31783
aI61803
aI1023
aI119291
asS'd'
I18221
sa(dp36442
S'a'
I80811
sS's'
I18
sS'm'
(lp36443
I1
aI1
aI3
aI11
aI5
aI15
aI107
aI155
aI465
aI249
aI1845
aI357
aI2769
aI3313
aI12335
aI16615
aI20809
aI103469
asS'd'
I18222
sa(dp36444
S'a'
I80834
sS's'
I18
sS'm'
(lp36445
I1
aI1
aI3
aI9
aI13
aI21
aI11
aI227
aI173
aI949
aI1255
aI3257
aI601
aI10865
aI12779
aI9173
aI87255
aI12867
asS'd'
I18223
sa(dp36446
S'a'
I80839
sS's'
I18
sS'm'
(lp36447
I1
aI3
aI3
aI15
aI3
aI41
aI97
aI141
aI385
aI23
aI1253
aI2905
aI1523
aI7647
aI7069
aI61143
aI101245
aI59747
asS'd'
I18224
sa(dp36448
S'a'
I80840
sS's'
I18
sS'm'
(lp36449
I1
aI3
aI1
aI3
aI7
aI35
aI117
aI93
aI357
aI741
aI1673
aI3295
aI6809
aI547
aI22949
aI42151
aI91241
aI16189
asS'd'
I18225
sa(dp36450
S'a'
I80846
sS's'
I18
sS'm'
(lp36451
I1
aI3
aI5
aI9
aI25
aI31
aI27
aI221
aI55
aI595
aI1513
aI3963
aI3143
aI1189
aI19843
aI6361
aI19575
aI231765
asS'd'
I18226
sa(dp36452
S'a'
I80848
sS's'
I18
sS'm'
(lp36453
I1
aI1
aI5
aI1
aI3
aI35
aI91
aI217
aI385
aI717
aI57
aI1471
aI3529
aI859
aI15259
aI4411
aI54491
aI79841
asS'd'
I18227
sa(dp36454
S'a'
I80876
sS's'
I18
sS'm'
(lp36455
I1
aI1
aI5
aI9
aI29
aI47
aI111
aI89
aI469
aI975
aI513
aI1339
aI1747
aI8839
aI30375
aI46217
aI128191
aI95831
asS'd'
I18228
sa(dp36456
S'a'
I80911
sS's'
I18
sS'm'
(lp36457
I1
aI1
aI5
aI13
aI9
aI45
aI3
aI221
aI223
aI461
aI1353
aI3953
aI5505
aI3139
aI3407
aI12953
aI74487
aI209401
asS'd'
I18229
sa(dp36458
S'a'
I80947
sS's'
I18
sS'm'
(lp36459
I1
aI1
aI7
aI7
aI7
aI43
aI33
aI143
aI427
aI183
aI573
aI2881
aI7355
aI10693
aI12841
aI14267
aI61847
aI47689
asS'd'
I18230
sa(dp36460
S'a'
I81001
sS's'
I18
sS'm'
(lp36461
I1
aI1
aI3
aI5
aI23
aI45
aI53
aI173
aI347
aI715
aI173
aI3385
aI429
aI8143
aI2831
aI57883
aI77245
aI37613
asS'd'
I18231
sa(dp36462
S'a'
I81031
sS's'
I18
sS'm'
(lp36463
I1
aI1
aI1
aI13
aI21
aI47
aI33
aI157
aI171
aI47
aI1981
aI2003
aI7401
aI7687
aI10553
aI38083
aI111901
aI30251
asS'd'
I18232
sa(dp36464
S'a'
I81045
sS's'
I18
sS'm'
(lp36465
I1
aI1
aI5
aI9
aI23
aI35
aI121
aI251
aI7
aI835
aI1561
aI1605
aI7023
aI15645
aI14313
aI6361
aI107973
aI211667
asS'd'
I18233
sa(dp36466
S'a'
I81056
sS's'
I18
sS'm'
(lp36467
I1
aI3
aI1
aI13
aI25
aI39
aI81
aI31
aI145
aI483
aI1587
aI3457
aI5293
aI927
aI3529
aI22457
aI69689
aI190371
asS'd'
I18234
sa(dp36468
S'a'
I81073
sS's'
I18
sS'm'
(lp36469
I1
aI1
aI1
aI3
aI25
aI61
aI87
aI111
aI441
aI829
aI313
aI2271
aI205
aI10187
aI3003
aI47237
aI99899
aI200553
asS'd'
I18235
sa(dp36470
S'a'
I81083
sS's'
I18
sS'm'
(lp36471
I1
aI1
aI7
aI13
aI31
aI51
aI9
aI243
aI219
aI139
aI1703
aI2001
aI959
aI11265
aI27897
aI9081
aI4473
aI107737
asS'd'
I18236
sa(dp36472
S'a'
I81145
sS's'
I18
sS'm'
(lp36473
I1
aI3
aI5
aI1
aI25
aI37
aI61
aI131
aI487
aI35
aI1293
aI833
aI3847
aI11315
aI11811
aI2763
aI2199
aI81127
asS'd'
I18237
sa(dp36474
S'a'
I81168
sS's'
I18
sS'm'
(lp36475
I1
aI3
aI7
aI7
aI31
aI33
aI87
aI111
aI429
aI809
aI173
aI1093
aI7719
aI14307
aI5735
aI61019
aI21223
aI26361
asS'd'
I18238
sa(dp36476
S'a'
I81177
sS's'
I18
sS'm'
(lp36477
I1
aI1
aI3
aI11
aI17
aI33
aI31
aI17
aI49
aI885
aI1279
aI2243
aI3693
aI61
aI30909
aI35807
aI14027
aI159225
asS'd'
I18239
sa(dp36478
S'a'
I81207
sS's'
I18
sS'm'
(lp36479
I1
aI3
aI7
aI11
aI9
aI35
aI61
aI75
aI171
aI117
aI1285
aI935
aI7271
aI3509
aI14119
aI31065
aI58181
aI136623
asS'd'
I18240
sa(dp36480
S'a'
I81208
sS's'
I18
sS'm'
(lp36481
I1
aI1
aI3
aI15
aI3
aI43
aI93
aI221
aI239
aI783
aI37
aI4007
aI3637
aI10461
aI18425
aI59629
aI93781
aI252689
asS'd'
I18241
sa(dp36482
S'a'
I81226
sS's'
I18
sS'm'
(lp36483
I1
aI3
aI5
aI7
aI5
aI61
aI19
aI107
aI123
aI417
aI1655
aI2307
aI8177
aI13617
aI17195
aI31597
aI66241
aI107199
asS'd'
I18242
sa(dp36484
S'a'
I81245
sS's'
I18
sS'm'
(lp36485
I1
aI3
aI7
aI7
aI5
aI5
aI25
aI69
aI383
aI217
aI993
aI2719
aI3425
aI8395
aI1125
aI10763
aI80111
aI70421
asS'd'
I18243
sa(dp36486
S'a'
I81269
sS's'
I18
sS'm'
(lp36487
I1
aI1
aI3
aI9
aI29
aI45
aI123
aI45
aI89
aI1015
aI1703
aI4049
aI4969
aI3801
aI23657
aI41031
aI66415
aI34063
asS'd'
I18244
sa(dp36488
S'a'
I81285
sS's'
I18
sS'm'
(lp36489
I1
aI1
aI3
aI3
aI7
aI53
aI125
aI63
aI67
aI335
aI1937
aI1793
aI4641
aI7115
aI10951
aI45503
aI54723
aI177433
asS'd'
I18245
sa(dp36490
S'a'
I81289
sS's'
I18
sS'm'
(lp36491
I1
aI3
aI5
aI1
aI21
aI55
aI83
aI199
aI509
aI331
aI695
aI2133
aI1881
aI14369
aI21687
aI2343
aI85895
aI99255
asS'd'
I18246
sa(dp36492
S'a'
I81292
sS's'
I18
sS'm'
(lp36493
I1
aI1
aI5
aI9
aI11
aI5
aI111
aI97
aI433
aI851
aI1537
aI411
aI6629
aI5185
aI30749
aI50017
aI46177
aI213347
asS'd'
I18247
sa(dp36494
S'a'
I81298
sS's'
I18
sS'm'
(lp36495
I1
aI3
aI3
aI1
aI7
aI21
aI95
aI229
aI311
aI605
aI1277
aI2435
aI5053
aI3051
aI15447
aI35479
aI2835
aI204149
asS'd'
I18248
sa(dp36496
S'a'
I81310
sS's'
I18
sS'm'
(lp36497
I1
aI3
aI5
aI9
aI31
aI27
aI79
aI201
aI329
aI735
aI1933
aI27
aI6201
aI9375
aI24801
aI34045
aI16227
aI61013
asS'd'
I18249
sa(dp36498
S'a'
I81346
sS's'
I18
sS'm'
(lp36499
I1
aI1
aI5
aI5
aI31
aI7
aI73
aI197
aI455
aI835
aI1845
aI2733
aI3371
aI513
aI10495
aI43659
aI4621
aI68969
asS'd'
I18250
sa(dp36500
S'a'
I81348
sS's'
I18
sS'm'
(lp36501
I1
aI1
aI1
aI15
aI21
aI55
aI15
aI83
aI419
aI471
aI1427
aI919
aI7125
aI7635
aI25579
aI19493
aI37381
aI191563
asS'd'
I18251
sa(dp36502
S'a'
I81355
sS's'
I18
sS'm'
(lp36503
I1
aI1
aI7
aI3
aI15
aI35
aI25
aI73
aI295
aI507
aI719
aI3307
aI4253
aI945
aI21005
aI24903
aI80287
aI48885
asS'd'
I18252
sa(dp36504
S'a'
I81388
sS's'
I18
sS'm'
(lp36505
I1
aI3
aI7
aI15
aI27
aI13
aI71
aI79
aI189
aI491
aI1185
aI3007
aI4285
aI13005
aI18973
aI33759
aI15327
aI45595
asS'd'
I18253
sa(dp36506
S'a'
I81396
sS's'
I18
sS'm'
(lp36507
I1
aI3
aI3
aI9
aI9
aI33
aI115
aI103
aI31
aI949
aI1817
aI2865
aI1215
aI9611
aI16019
aI7925
aI72945
aI208301
asS'd'
I18254
sa(dp36508
S'a'
I81415
sS's'
I18
sS'm'
(lp36509
I1
aI1
aI1
aI5
aI19
aI35
aI89
aI181
aI409
aI641
aI1277
aI2201
aI2825
aI5707
aI13463
aI34741
aI39303
aI217803
asS'd'
I18255
sa(dp36510
S'a'
I81460
sS's'
I18
sS'm'
(lp36511
I1
aI1
aI3
aI11
aI13
aI31
aI65
aI191
aI11
aI179
aI509
aI2513
aI3861
aI13323
aI11817
aI24901
aI53815
aI44343
asS'd'
I18256
sa(dp36512
S'a'
I81482
sS's'
I18
sS'm'
(lp36513
I1
aI3
aI1
aI5
aI5
aI57
aI97
aI25
aI83
aI177
aI1963
aI2367
aI6703
aI13361
aI8749
aI45533
aI87883
aI2977
asS'd'
I18257
sa(dp36514
S'a'
I81518
sS's'
I18
sS'm'
(lp36515
I1
aI3
aI5
aI3
aI15
aI41
aI113
aI145
aI39
aI509
aI81
aI1387
aI2881
aI1441
aI75
aI28409
aI61417
aI79393
asS'd'
I18258
sa(dp36516
S'a'
I81523
sS's'
I18
sS'm'
(lp36517
I1
aI3
aI3
aI3
aI17
aI1
aI41
aI19
aI173
aI133
aI2033
aI3637
aI7415
aI1841
aI19497
aI42643
aI122885
aI195301
asS'd'
I18259
sa(dp36518
S'a'
I81529
sS's'
I18
sS'm'
(lp36519
I1
aI3
aI3
aI9
aI15
aI37
aI11
aI87
aI291
aI881
aI1471
aI2469
aI6877
aI6813
aI8273
aI1455
aI30957
aI181887
asS'd'
I18260
sa(dp36520
S'a'
I81545
sS's'
I18
sS'm'
(lp36521
I1
aI3
aI5
aI9
aI25
aI41
aI7
aI71
aI451
aI831
aI495
aI3991
aI4173
aI4307
aI31249
aI7253
aI57141
aI35495
asS'd'
I18261
sa(dp36522
S'a'
I81570
sS's'
I18
sS'm'
(lp36523
I1
aI1
aI7
aI9
aI15
aI39
aI29
aI193
aI327
aI837
aI991
aI3503
aI1175
aI14965
aI18151
aI22479
aI51127
aI159019
asS'd'
I18262
sa(dp36524
S'a'
I81576
sS's'
I18
sS'm'
(lp36525
I1
aI3
aI1
aI9
aI23
aI41
aI89
aI211
aI179
aI507
aI1005
aI613
aI8083
aI15655
aI1927
aI23401
aI51025
aI21589
asS'd'
I18263
sa(dp36526
S'a'
I81604
sS's'
I18
sS'm'
(lp36527
I1
aI1
aI5
aI15
aI5
aI63
aI105
aI229
aI239
aI399
aI591
aI2233
aI391
aI2871
aI29829
aI49961
aI62045
aI190437
asS'd'
I18264
sa(dp36528
S'a'
I81613
sS's'
I18
sS'm'
(lp36529
I1
aI3
aI5
aI9
aI7
aI23
aI85
aI219
aI163
aI37
aI1881
aI589
aI4239
aI12845
aI19993
aI57267
aI29519
aI207597
asS'd'
I18265
sa(dp36530
S'a'
I81631
sS's'
I18
sS'm'
(lp36531
I1
aI3
aI7
aI15
aI19
aI19
aI115
aI141
aI41
aI405
aI657
aI2517
aI4231
aI10247
aI21383
aI11479
aI52955
aI121545
asS'd'
I18266
sa(dp36532
S'a'
I81656
sS's'
I18
sS'm'
(lp36533
I1
aI3
aI1
aI7
aI23
aI33
aI65
aI229
aI287
aI739
aI1265
aI1105
aI487
aI3801
aI5211
aI44731
aI5359
aI103685
asS'd'
I18267
sa(dp36534
S'a'
I81679
sS's'
I18
sS'm'
(lp36535
I1
aI3
aI1
aI13
aI23
aI29
aI101
aI153
aI395
aI335
aI899
aI303
aI2073
aI15767
aI1303
aI15539
aI12889
aI35517
asS'd'
I18268
sa(dp36536
S'a'
I81684
sS's'
I18
sS'm'
(lp36537
I1
aI1
aI5
aI11
aI5
aI63
aI41
aI53
aI99
aI339
aI563
aI2921
aI4959
aI13941
aI13655
aI10115
aI56867
aI42919
asS'd'
I18269
sa(dp36538
S'a'
I81698
sS's'
I18
sS'm'
(lp36539
I1
aI3
aI5
aI5
aI5
aI35
aI127
aI225
aI497
aI27
aI139
aI3269
aI3929
aI3369
aI22697
aI19421
aI2921
aI171927
asS'd'
I18270
sa(dp36540
S'a'
I81736
sS's'
I18
sS'm'
(lp36541
I1
aI1
aI1
aI15
aI15
aI21
aI35
aI251
aI67
aI447
aI1045
aI1173
aI2951
aI6589
aI27261
aI36597
aI98721
aI7205
asS'd'
I18271
sa(dp36542
S'a'
I81747
sS's'
I18
sS'm'
(lp36543
I1
aI3
aI3
aI9
aI11
aI63
aI83
aI19
aI163
aI381
aI87
aI1211
aI3007
aI4971
aI27105
aI2341
aI21389
aI32995
asS'd'
I18272
sa(dp36544
S'a'
I81765
sS's'
I18
sS'm'
(lp36545
I1
aI1
aI3
aI3
aI21
aI19
aI63
aI65
aI505
aI987
aI1821
aI2419
aI3195
aI2573
aI1481
aI35279
aI45135
aI597
asS'd'
I18273
sa(dp36546
S'a'
I81775
sS's'
I18
sS'm'
(lp36547
I1
aI3
aI1
aI15
aI29
aI5
aI77
aI65
aI121
aI223
aI2009
aI593
aI7929
aI10353
aI22301
aI25137
aI40289
aI95847
asS'd'
I18274
sa(dp36548
S'a'
I81805
sS's'
I18
sS'm'
(lp36549
I1
aI1
aI3
aI1
aI17
aI49
aI9
aI167
aI69
aI729
aI1189
aI1191
aI1
aI12603
aI8281
aI45193
aI1427
aI15887
asS'd'
I18275
sa(dp36550
S'a'
I81842
sS's'
I18
sS'm'
(lp36551
I1
aI1
aI3
aI7
aI17
aI5
aI11
aI217
aI505
aI317
aI505
aI1201
aI8025
aI13255
aI12591
aI16207
aI32387
aI242425
asS'd'
I18276
sa(dp36552
S'a'
I81859
sS's'
I18
sS'm'
(lp36553
I1
aI3
aI7
aI9
aI25
aI9
aI97
aI23
aI91
aI765
aI653
aI2689
aI2787
aI11719
aI8455
aI24665
aI26907
aI78525
asS'd'
I18277
sa(dp36554
S'a'
I81865
sS's'
I18
sS'm'
(lp36555
I1
aI3
aI3
aI15
aI27
aI19
aI79
aI157
aI117
aI715
aI1921
aI2453
aI499
aI13593
aI14173
aI1993
aI110087
aI151427
asS'd'
I18278
sa(dp36556
S'a'
I81866
sS's'
I18
sS'm'
(lp36557
I1
aI3
aI1
aI13
aI5
aI43
aI59
aI21
aI451
aI863
aI533
aI1723
aI2059
aI1611
aI10403
aI36479
aI36999
aI109553
asS'd'
I18279
sa(dp36558
S'a'
I81892
sS's'
I18
sS'm'
(lp36559
I1
aI3
aI7
aI7
aI29
aI63
aI51
aI5
aI475
aI549
aI123
aI1949
aI5279
aI8581
aI20053
aI52287
aI125223
aI152299
asS'd'
I18280
sa(dp36560
S'a'
I81902
sS's'
I18
sS'm'
(lp36561
I1
aI3
aI1
aI1
aI7
aI19
aI1
aI215
aI273
aI157
aI1557
aI425
aI7549
aI12337
aI1735
aI30917
aI116487
aI177335
asS'd'
I18281
sa(dp36562
S'a'
I81933
sS's'
I18
sS'm'
(lp36563
I1
aI1
aI1
aI1
aI7
aI47
aI61
aI191
aI73
aI551
aI1435
aI2283
aI3191
aI8545
aI11875
aI41389
aI17607
aI26869
asS'd'
I18282
sa(dp36564
S'a'
I81934
sS's'
I18
sS'm'
(lp36565
I1
aI1
aI7
aI9
aI13
aI61
aI109
aI121
aI365
aI223
aI1729
aI3311
aI7249
aI10765
aI12419
aI4235
aI64127
aI132257
asS'd'
I18283
sa(dp36566
S'a'
I81942
sS's'
I18
sS'm'
(lp36567
I1
aI1
aI3
aI13
aI17
aI25
aI65
aI49
aI417
aI311
aI141
aI1127
aI53
aI945
aI28277
aI33347
aI96399
aI166049
asS'd'
I18284
sa(dp36568
S'a'
I81969
sS's'
I18
sS'm'
(lp36569
I1
aI3
aI7
aI9
aI5
aI21
aI93
aI203
aI467
aI805
aI115
aI1757
aI4535
aI8687
aI10423
aI8065
aI2955
aI20403
asS'd'
I18285
sa(dp36570
S'a'
I81981
sS's'
I18
sS'm'
(lp36571
I1
aI3
aI3
aI15
aI7
aI63
aI103
aI137
aI227
aI111
aI735
aI2139
aI4293
aI5347
aI4131
aI63405
aI42599
aI173299
asS'd'
I18286
sa(dp36572
S'a'
I81999
sS's'
I18
sS'm'
(lp36573
I1
aI3
aI7
aI7
aI17
aI53
aI127
aI251
aI57
aI625
aI843
aI3045
aI1319
aI10085
aI18591
aI36115
aI104193
aI183891
asS'd'
I18287
sa(dp36574
S'a'
I82004
sS's'
I18
sS'm'
(lp36575
I1
aI3
aI7
aI1
aI31
aI57
aI107
aI253
aI207
aI739
aI1703
aI1377
aI3807
aI10289
aI22969
aI13087
aI2805
aI261279
asS'd'
I18288
sa(dp36576
S'a'
I82008
sS's'
I18
sS'm'
(lp36577
I1
aI1
aI5
aI5
aI5
aI59
aI59
aI63
aI77
aI663
aI1109
aI2159
aI3725
aI12355
aI4805
aI22433
aI81851
aI9419
asS'd'
I18289
sa(dp36578
S'a'
I82032
sS's'
I18
sS'm'
(lp36579
I1
aI1
aI7
aI15
aI1
aI1
aI101
aI101
aI295
aI311
aI447
aI3931
aI933
aI15713
aI8919
aI7185
aI38577
aI254203
asS'd'
I18290
sa(dp36580
S'a'
I82035
sS's'
I18
sS'm'
(lp36581
I1
aI1
aI5
aI15
aI7
aI35
aI35
aI141
aI283
aI665
aI1685
aI3875
aI495
aI1655
aI8269
aI23493
aI1523
aI248783
asS'd'
I18291
sa(dp36582
S'a'
I82060
sS's'
I18
sS'm'
(lp36583
I1
aI1
aI5
aI9
aI27
aI35
aI25
aI57
aI285
aI469
aI1491
aI1479
aI3705
aI11357
aI5319
aI11575
aI116207
aI215961
asS'd'
I18292
sa(dp36584
S'a'
I82063
sS's'
I18
sS'm'
(lp36585
I1
aI3
aI5
aI11
aI7
aI41
aI67
aI161
aI73
aI777
aI247
aI823
aI6677
aI1631
aI3431
aI2821
aI25291
aI17633
asS'd'
I18293
sa(dp36586
S'a'
I82071
sS's'
I18
sS'm'
(lp36587
I1
aI1
aI3
aI9
aI19
aI17
aI45
aI181
aI139
aI85
aI857
aI1231
aI7167
aI2951
aI26847
aI39113
aI51705
aI104617
asS'd'
I18294
sa(dp36588
S'a'
I82081
sS's'
I18
sS'm'
(lp36589
I1
aI3
aI5
aI1
aI5
aI55
aI101
aI209
aI1
aI47
aI1059
aI2175
aI1549
aI8007
aI11267
aI21863
aI125567
aI102775
asS'd'
I18295
sa(dp36590
S'a'
I82082
sS's'
I18
sS'm'
(lp36591
I1
aI3
aI3
aI15
aI15
aI21
aI79
aI85
aI427
aI963
aI1335
aI2129
aI6831
aI6613
aI13319
aI15781
aI3781
aI222547
asS'd'
I18296
sa(dp36592
S'a'
I82105
sS's'
I18
sS'm'
(lp36593
I1
aI3
aI3
aI13
aI19
aI63
aI25
aI123
aI1
aI215
aI139
aI1345
aI5035
aI3107
aI14381
aI6239
aI18481
aI202581
asS'd'
I18297
sa(dp36594
S'a'
I82106
sS's'
I18
sS'm'
(lp36595
I1
aI1
aI5
aI1
aI11
aI11
aI11
aI53
aI109
aI533
aI1113
aI177
aI609
aI15391
aI22735
aI62229
aI103591
aI89143
asS'd'
I18298
sa(dp36596
S'a'
I82120
sS's'
I18
sS'm'
(lp36597
I1
aI1
aI5
aI15
aI3
aI21
aI115
aI223
aI167
aI441
aI277
aI2971
aI933
aI2841
aI26893
aI48513
aI74553
aI250413
asS'd'
I18299
sa(dp36598
S'a'
I82125
sS's'
I18
sS'm'
(lp36599
I1
aI1
aI1
aI7
aI19
aI17
aI43
aI181
aI483
aI897
aI819
aI1657
aI5539
aI8847
aI23483
aI57605
aI104703
aI242559
asS'd'
I18300
sa(dp36600
S'a'
I82156
sS's'
I18
sS'm'
(lp36601
I1
aI3
aI5
aI11
aI3
aI63
aI3
aI129
aI45
aI981
aI45
aI845
aI1481
aI14735
aI30451
aI16937
aI13789
aI27107
asS'd'
I18301
sa(dp36602
S'a'
I82176
sS's'
I18
sS'm'
(lp36603
I1
aI3
aI3
aI15
aI25
aI11
aI33
aI49
aI155
aI947
aI521
aI3417
aI3299
aI1123
aI9517
aI32127
aI117795
aI223167
asS'd'
I18302
sa(dp36604
S'a'
I82203
sS's'
I18
sS'm'
(lp36605
I1
aI3
aI5
aI15
aI3
aI35
aI27
aI37
aI287
aI541
aI727
aI2779
aI7033
aI5189
aI21579
aI36895
aI109645
aI123353
asS'd'
I18303
sa(dp36606
S'a'
I82210
sS's'
I18
sS'm'
(lp36607
I1
aI3
aI7
aI9
aI15
aI53
aI123
aI125
aI405
aI841
aI119
aI63
aI853
aI8693
aI1537
aI25509
aI49345
aI54301
asS'd'
I18304
sa(dp36608
S'a'
I82241
sS's'
I18
sS'm'
(lp36609
I1
aI3
aI7
aI9
aI11
aI63
aI65
aI145
aI283
aI529
aI1553
aI883
aI3319
aI8601
aI29379
aI26991
aI127343
aI98701
asS'd'
I18305
sa(dp36610
S'a'
I82287
sS's'
I18
sS'm'
(lp36611
I1
aI1
aI1
aI7
aI23
aI59
aI11
aI89
aI407
aI869
aI445
aI659
aI3029
aI5465
aI5063
aI36775
aI69089
aI205367
asS'd'
I18306
sa(dp36612
S'a'
I82323
sS's'
I18
sS'm'
(lp36613
I1
aI3
aI7
aI5
aI19
aI35
aI99
aI49
aI257
aI287
aI1113
aI2825
aI2797
aI7283
aI31757
aI47015
aI106987
aI82589
asS'd'
I18307
sa(dp36614
S'a'
I82330
sS's'
I18
sS'm'
(lp36615
I1
aI3
aI7
aI11
aI15
aI37
aI41
aI101
aI493
aI725
aI1091
aI503
aI2611
aI13025
aI11071
aI39311
aI5193
aI92127
asS'd'
I18308
sa(dp36616
S'a'
I82366
sS's'
I18
sS'm'
(lp36617
I1
aI1
aI3
aI7
aI9
aI59
aI69
aI113
aI381
aI341
aI1495
aI3169
aI5099
aI69
aI7911
aI9721
aI84609
aI254171
asS'd'
I18309
sa(dp36618
S'a'
I82378
sS's'
I18
sS'm'
(lp36619
I1
aI3
aI5
aI7
aI21
aI19
aI75
aI71
aI7
aI617
aI1185
aI2787
aI4147
aI16045
aI18859
aI52347
aI66551
aI161563
asS'd'
I18310
sa(dp36620
S'a'
I82395
sS's'
I18
sS'm'
(lp36621
I1
aI3
aI5
aI3
aI27
aI39
aI17
aI205
aI425
aI3
aI1443
aI1947
aI7645
aI10125
aI24577
aI45373
aI38015
aI30407
asS'd'
I18311
sa(dp36622
S'a'
I82398
sS's'
I18
sS'm'
(lp36623
I1
aI3
aI3
aI11
aI1
aI57
aI105
aI251
aI65
aI389
aI1993
aI3933
aI3093
aI1425
aI9483
aI5953
aI13147
aI234121
asS'd'
I18312
sa(dp36624
S'a'
I82401
sS's'
I18
sS'm'
(lp36625
I1
aI3
aI3
aI3
aI1
aI27
aI105
aI45
aI435
aI393
aI609
aI291
aI545
aI4905
aI22621
aI62115
aI78955
aI84355
asS'd'
I18313
sa(dp36626
S'a'
I82404
sS's'
I18
sS'm'
(lp36627
I1
aI3
aI7
aI9
aI1
aI15
aI91
aI183
aI301
aI223
aI1183
aI1877
aI2141
aI5549
aI371
aI44147
aI6771
aI136777
asS'd'
I18314
sa(dp36628
S'a'
I82419
sS's'
I18
sS'm'
(lp36629
I1
aI1
aI5
aI15
aI5
aI49
aI127
aI161
aI121
aI979
aI1247
aI3681
aI3805
aI3363
aI11643
aI25735
aI21193
aI111657
asS'd'
I18315
sa(dp36630
S'a'
I82421
sS's'
I18
sS'm'
(lp36631
I1
aI3
aI5
aI15
aI15
aI33
aI47
aI91
aI137
aI323
aI1577
aI3723
aI3609
aI11533
aI4415
aI26467
aI120947
aI200919
asS'd'
I18316
sa(dp36632
S'a'
I82428
sS's'
I18
sS'm'
(lp36633
I1
aI3
aI3
aI3
aI3
aI33
aI121
aI161
aI453
aI205
aI1815
aI65
aI5893
aI4669
aI14377
aI10905
aI9559
aI56359
asS'd'
I18317
sa(dp36634
S'a'
I82442
sS's'
I18
sS'm'
(lp36635
I1
aI1
aI1
aI7
aI1
aI55
aI21
aI143
aI411
aI65
aI1009
aI2989
aI133
aI7059
aI30981
aI15417
aI2651
aI110345
asS'd'
I18318
sa(dp36636
S'a'
I82452
sS's'
I18
sS'm'
(lp36637
I1
aI1
aI3
aI7
aI19
aI25
aI91
aI241
aI193
aI903
aI661
aI665
aI7681
aI14111
aI29197
aI51299
aI109519
aI155827
asS'd'
I18319
sa(dp36638
S'a'
I82455
sS's'
I18
sS'm'
(lp36639
I1
aI1
aI1
aI15
aI25
aI3
aI79
aI57
aI417
aI73
aI705
aI7
aI4415
aI7699
aI28185
aI53005
aI88547
aI7281
asS'd'
I18320
sa(dp36640
S'a'
I82466
sS's'
I18
sS'm'
(lp36641
I1
aI1
aI7
aI13
aI27
aI21
aI35
aI197
aI65
aI171
aI1773
aI393
aI3759
aI8335
aI5987
aI20611
aI91373
aI80715
asS'd'
I18321
sa(dp36642
S'a'
I82510
sS's'
I18
sS'm'
(lp36643
I1
aI3
aI7
aI3
aI17
aI51
aI85
aI229
aI131
aI733
aI281
aI3157
aI1283
aI10751
aI20203
aI49955
aI23861
aI128517
asS'd'
I18322
sa(dp36644
S'a'
I82524
sS's'
I18
sS'm'
(lp36645
I1
aI1
aI5
aI15
aI3
aI27
aI35
aI87
aI391
aI509
aI1627
aI769
aI701
aI4933
aI24597
aI9695
aI111441
aI198493
asS'd'
I18323
sa(dp36646
S'a'
I82545
sS's'
I18
sS'm'
(lp36647
I1
aI1
aI1
aI15
aI3
aI31
aI73
aI235
aI341
aI263
aI883
aI2369
aI4887
aI4659
aI9493
aI6763
aI130625
aI15031
asS'd'
I18324
sa(dp36648
S'a'
I82555
sS's'
I18
sS'm'
(lp36649
I1
aI3
aI1
aI15
aI1
aI11
aI63
aI79
aI389
aI355
aI619
aI1361
aI313
aI1199
aI555
aI42213
aI81089
aI170863
asS'd'
I18325
sa(dp36650
S'a'
I82581
sS's'
I18
sS'm'
(lp36651
I1
aI3
aI1
aI15
aI21
aI27
aI1
aI179
aI19
aI241
aI1655
aI1803
aI5413
aI5353
aI65
aI31211
aI3501
aI27205
asS'd'
I18326
sa(dp36652
S'a'
I82588
sS's'
I18
sS'm'
(lp36653
I1
aI3
aI3
aI9
aI9
aI19
aI63
aI191
aI217
aI271
aI1453
aI2777
aI2915
aI13291
aI31391
aI37489
aI86435
aI22857
asS'd'
I18327
sa(dp36654
S'a'
I82591
sS's'
I18
sS'm'
(lp36655
I1
aI3
aI5
aI9
aI13
aI41
aI85
aI11
aI333
aI479
aI363
aI2591
aI697
aI8587
aI3647
aI5741
aI21627
aI244573
asS'd'
I18328
sa(dp36656
S'a'
I82636
sS's'
I18
sS'm'
(lp36657
I1
aI1
aI3
aI1
aI5
aI61
aI83
aI229
aI193
aI977
aI677
aI2585
aI3273
aI12035
aI2621
aI12943
aI49293
aI37985
asS'd'
I18329
sa(dp36658
S'a'
I82658
sS's'
I18
sS'm'
(lp36659
I1
aI1
aI5
aI7
aI27
aI9
aI69
aI189
aI489
aI747
aI519
aI719
aI1493
aI13337
aI14933
aI44359
aI11471
aI57245
asS'd'
I18330
sa(dp36660
S'a'
I82675
sS's'
I18
sS'm'
(lp36661
I1
aI3
aI5
aI1
aI5
aI17
aI75
aI89
aI417
aI367
aI57
aI1641
aI1573
aI1819
aI31237
aI5213
aI78821
aI149853
asS'd'
I18331
sa(dp36662
S'a'
I82678
sS's'
I18
sS'm'
(lp36663
I1
aI1
aI7
aI3
aI7
aI17
aI121
aI91
aI211
aI101
aI1145
aI3753
aI2997
aI67
aI10755
aI11261
aI122489
aI61679
asS'd'
I18332
sa(dp36664
S'a'
I82716
sS's'
I18
sS'm'
(lp36665
I1
aI3
aI7
aI7
aI15
aI17
aI73
aI133
aI429
aI285
aI201
aI1917
aI5677
aI1793
aI21653
aI49729
aI68965
aI5347
asS'd'
I18333
sa(dp36666
S'a'
I82725
sS's'
I18
sS'm'
(lp36667
I1
aI3
aI3
aI13
aI23
aI17
aI49
aI249
aI71
aI169
aI619
aI843
aI2163
aI585
aI23309
aI39509
aI68087
aI232233
asS'd'
I18334
sa(dp36668
S'a'
I82743
sS's'
I18
sS'm'
(lp36669
I1
aI1
aI7
aI15
aI23
aI15
aI19
aI227
aI89
aI719
aI1247
aI2521
aI1509
aI7553
aI12225
aI12865
aI100107
aI261847
asS'd'
I18335
sa(dp36670
S'a'
I82744
sS's'
I18
sS'm'
(lp36671
I1
aI1
aI5
aI3
aI23
aI17
aI117
aI5
aI401
aI57
aI1945
aI1081
aI1269
aI5921
aI31815
aI42341
aI112099
aI130047
asS'd'
I18336
sa(dp36672
S'a'
I82762
sS's'
I18
sS'm'
(lp36673
I1
aI1
aI1
aI1
aI9
aI5
aI87
aI203
aI211
aI1009
aI403
aI1617
aI3969
aI2541
aI7261
aI6989
aI16579
aI206159
asS'd'
I18337
sa(dp36674
S'a'
I82770
sS's'
I18
sS'm'
(lp36675
I1
aI3
aI7
aI9
aI5
aI13
aI93
aI191
aI79
aI631
aI1019
aI3639
aI7137
aI13859
aI19603
aI63263
aI82947
aI181023
asS'd'
I18338
sa(dp36676
S'a'
I82809
sS's'
I18
sS'm'
(lp36677
I1
aI1
aI5
aI11
aI25
aI17
aI85
aI51
aI61
aI311
aI517
aI2001
aI6325
aI6831
aI10835
aI20101
aI115241
aI15815
asS'd'
I18339
sa(dp36678
S'a'
I82846
sS's'
I18
sS'm'
(lp36679
I1
aI3
aI7
aI13
aI29
aI19
aI33
aI115
aI473
aI477
aI471
aI773
aI4097
aI11697
aI30781
aI20843
aI27089
aI181927
asS'd'
I18340
sa(dp36680
S'a'
I82855
sS's'
I18
sS'm'
(lp36681
I1
aI3
aI5
aI5
aI21
aI27
aI3
aI239
aI45
aI335
aI505
aI149
aI3005
aI3511
aI18037
aI31291
aI6145
aI2913
asS'd'
I18341
sa(dp36682
S'a'
I82882
sS's'
I18
sS'm'
(lp36683
I1
aI1
aI3
aI1
aI25
aI49
aI21
aI225
aI27
aI395
aI415
aI1813
aI5727
aI7211
aI9887
aI63533
aI99185
aI119269
asS'd'
I18342
sa(dp36684
S'a'
I82899
sS's'
I18
sS'm'
(lp36685
I1
aI3
aI3
aI5
aI15
aI53
aI127
aI195
aI81
aI895
aI587
aI561
aI5951
aI9901
aI18117
aI37855
aI19393
aI259031
asS'd'
I18343
sa(dp36686
S'a'
I82905
sS's'
I18
sS'm'
(lp36687
I1
aI1
aI7
aI13
aI9
aI49
aI109
aI127
aI53
aI735
aI391
aI1523
aI3759
aI10363
aI11299
aI3203
aI89121
aI122643
asS'd'
I18344
sa(dp36688
S'a'
I82950
sS's'
I18
sS'm'
(lp36689
I1
aI1
aI5
aI3
aI13
aI3
aI21
aI247
aI259
aI557
aI977
aI1465
aI6889
aI3879
aI4627
aI1439
aI122809
aI248941
asS'd'
I18345
sa(dp36690
S'a'
I82953
sS's'
I18
sS'm'
(lp36691
I1
aI3
aI7
aI15
aI7
aI19
aI113
aI251
aI245
aI63
aI267
aI1873
aI6601
aI16253
aI24643
aI7433
aI130051
aI233047
asS'd'
I18346
sa(dp36692
S'a'
I82967
sS's'
I18
sS'm'
(lp36693
I1
aI1
aI3
aI9
aI29
aI39
aI47
aI31
aI493
aI817
aI1697
aI2139
aI1059
aI11365
aI31653
aI56477
aI119191
aI45509
asS'd'
I18347
sa(dp36694
S'a'
I82971
sS's'
I18
sS'm'
(lp36695
I1
aI1
aI1
aI15
aI9
aI29
aI99
aI61
aI109
aI341
aI1009
aI1551
aI897
aI13075
aI10603
aI25153
aI65911
aI228213
asS'd'
I18348
sa(dp36696
S'a'
I82987
sS's'
I18
sS'm'
(lp36697
I1
aI3
aI7
aI7
aI29
aI47
aI57
aI85
aI263
aI767
aI1633
aI2473
aI199
aI49
aI22287
aI33345
aI118877
aI248435
asS'd'
I18349
sa(dp36698
S'a'
I83007
sS's'
I18
sS'm'
(lp36699
I1
aI3
aI1
aI9
aI5
aI45
aI5
aI179
aI9
aI129
aI1231
aI4075
aI7497
aI2159
aI18101
aI31039
aI95213
aI171913
asS'd'
I18350
sa(dp36700
S'a'
I83009
sS's'
I18
sS'm'
(lp36701
I1
aI1
aI1
aI13
aI23
aI1
aI89
aI63
aI21
aI983
aI481
aI773
aI5957
aI4823
aI4483
aI50405
aI42979
aI243567
asS'd'
I18351
sa(dp36702
S'a'
I83012
sS's'
I18
sS'm'
(lp36703
I1
aI1
aI5
aI13
aI15
aI21
aI65
aI133
aI347
aI511
aI1887
aI743
aI7825
aI1681
aI4857
aI49247
aI21277
aI212995
asS'd'
I18352
sa(dp36704
S'a'
I83030
sS's'
I18
sS'm'
(lp36705
I1
aI3
aI7
aI9
aI7
aI51
aI3
aI233
aI287
aI727
aI815
aI3609
aI397
aI5721
aI16473
aI7549
aI100455
aI136233
asS'd'
I18353
sa(dp36706
S'a'
I83045
sS's'
I18
sS'm'
(lp36707
I1
aI3
aI1
aI1
aI31
aI51
aI59
aI37
aI79
aI623
aI1219
aI2655
aI4619
aI11967
aI11377
aI28985
aI16069
aI188773
asS'd'
I18354
sa(dp36708
S'a'
I83050
sS's'
I18
sS'm'
(lp36709
I1
aI3
aI3
aI13
aI13
aI59
aI93
aI159
aI197
aI339
aI1633
aI1601
aI255
aI1631
aI4989
aI12019
aI23921
aI261273
asS'd'
I18355
sa(dp36710
S'a'
I83052
sS's'
I18
sS'm'
(lp36711
I1
aI1
aI3
aI13
aI27
aI25
aI55
aI43
aI147
aI981
aI65
aI725
aI5753
aI115
aI26125
aI25501
aI89099
aI233419
asS'd'
I18356
sa(dp36712
S'a'
I83070
sS's'
I18
sS'm'
(lp36713
I1
aI3
aI7
aI5
aI25
aI3
aI95
aI135
aI191
aI417
aI929
aI3855
aI5829
aI3827
aI13979
aI65367
aI63683
aI85911
asS'd'
I18357
sa(dp36714
S'a'
I83076
sS's'
I18
sS'm'
(lp36715
I1
aI3
aI5
aI1
aI7
aI63
aI45
aI187
aI355
aI735
aI1325
aI1461
aI3869
aI2127
aI18231
aI45891
aI24027
aI202997
asS'd'
I18358
sa(dp36716
S'a'
I83086
sS's'
I18
sS'm'
(lp36717
I1
aI3
aI7
aI13
aI19
aI47
aI89
aI229
aI253
aI659
aI355
aI3323
aI4081
aI8243
aI32553
aI46579
aI46431
aI53291
asS'd'
I18359
sa(dp36718
S'a'
I83128
sS's'
I18
sS'm'
(lp36719
I1
aI1
aI5
aI11
aI3
aI61
aI33
aI65
aI239
aI779
aI665
aI1337
aI6427
aI12787
aI1495
aI27105
aI71455
aI89715
asS'd'
I18360
sa(dp36720
S'a'
I83141
sS's'
I18
sS'm'
(lp36721
I1
aI3
aI1
aI11
aI31
aI33
aI115
aI69
aI511
aI187
aI99
aI1055
aI1065
aI9531
aI29897
aI23897
aI80581
aI166957
asS'd'
I18361
sa(dp36722
S'a'
I83153
sS's'
I18
sS'm'
(lp36723
I1
aI1
aI7
aI13
aI19
aI1
aI13
aI241
aI89
aI761
aI425
aI3865
aI961
aI14999
aI24175
aI19103
aI39095
aI38899
asS'd'
I18362
sa(dp36724
S'a'
I83156
sS's'
I18
sS'm'
(lp36725
I1
aI3
aI7
aI7
aI11
aI17
aI25
aI217
aI113
aI615
aI1455
aI1409
aI5679
aI2321
aI28687
aI8089
aI74031
aI230559
asS'd'
I18363
sa(dp36726
S'a'
I83194
sS's'
I18
sS'm'
(lp36727
I1
aI3
aI7
aI1
aI25
aI15
aI77
aI111
aI405
aI523
aI961
aI647
aI3857
aI14355
aI27063
aI48829
aI87913
aI254965
asS'd'
I18364
sa(dp36728
S'a'
I83225
sS's'
I18
sS'm'
(lp36729
I1
aI1
aI3
aI3
aI5
aI13
aI67
aI155
aI393
aI943
aI1875
aI1209
aI3765
aI8627
aI15123
aI43405
aI78473
aI146127
asS'd'
I18365
sa(dp36730
S'a'
I83261
sS's'
I18
sS'm'
(lp36731
I1
aI1
aI3
aI7
aI29
aI43
aI23
aI33
aI35
aI883
aI1859
aI1559
aI4163
aI13277
aI16971
aI15289
aI60305
aI56743
asS'd'
I18366
sa(dp36732
S'a'
I83264
sS's'
I18
sS'm'
(lp36733
I1
aI1
aI1
aI5
aI19
aI55
aI37
aI53
aI123
aI35
aI1477
aI1035
aI4683
aI259
aI20079
aI37041
aI48081
aI198685
asS'd'
I18367
sa(dp36734
S'a'
I83322
sS's'
I18
sS'm'
(lp36735
I1
aI1
aI3
aI3
aI1
aI19
aI27
aI129
aI427
aI685
aI959
aI2501
aI2761
aI9495
aI23649
aI18789
aI54521
aI219547
asS'd'
I18368
sa(dp36736
S'a'
I83382
sS's'
I18
sS'm'
(lp36737
I1
aI1
aI7
aI1
aI31
aI11
aI65
aI171
aI229
aI11
aI1825
aI1641
aI2731
aI11085
aI2567
aI30831
aI20365
aI242731
asS'd'
I18369
sa(dp36738
S'a'
I83391
sS's'
I18
sS'm'
(lp36739
I1
aI1
aI1
aI13
aI13
aI5
aI21
aI175
aI265
aI271
aI133
aI407
aI3415
aI5943
aI15385
aI12817
aI106159
aI41859
asS'd'
I18370
sa(dp36740
S'a'
I83399
sS's'
I18
sS'm'
(lp36741
I1
aI3
aI7
aI1
aI11
aI45
aI105
aI229
aI395
aI877
aI1495
aI2113
aI1733
aI10117
aI1125
aI9989
aI109637
aI124517
asS'd'
I18371
sa(dp36742
S'a'
I83406
sS's'
I18
sS'm'
(lp36743
I1
aI1
aI7
aI3
aI27
aI43
aI57
aI77
aI63
aI907
aI1137
aI3333
aI189
aI15285
aI13895
aI23773
aI73523
aI47811
asS'd'
I18372
sa(dp36744
S'a'
I83427
sS's'
I18
sS'm'
(lp36745
I1
aI3
aI3
aI11
aI17
aI19
aI81
aI197
aI73
aI897
aI515
aI3801
aI5105
aI6987
aI10125
aI7239
aI32339
aI124411
asS'd'
I18373
sa(dp36746
S'a'
I83439
sS's'
I18
sS'm'
(lp36747
I1
aI3
aI1
aI9
aI11
aI15
aI99
aI109
aI307
aI133
aI249
aI1463
aI5479
aI8565
aI19489
aI13773
aI11443
aI149799
asS'd'
I18374
sa(dp36748
S'a'
I83444
sS's'
I18
sS'm'
(lp36749
I1
aI3
aI7
aI7
aI19
aI53
aI61
aI75
aI83
aI545
aI1449
aI683
aI5845
aI8325
aI18111
aI35941
aI51843
aI97907
asS'd'
I18375
sa(dp36750
S'a'
I83453
sS's'
I18
sS'm'
(lp36751
I1
aI1
aI1
aI3
aI31
aI23
aI37
aI187
aI207
aI51
aI439
aI3095
aI2217
aI6393
aI9117
aI2779
aI47331
aI118275
asS'd'
I18376
sa(dp36752
S'a'
I83457
sS's'
I18
sS'm'
(lp36753
I1
aI1
aI3
aI9
aI23
aI17
aI41
aI37
aI59
aI281
aI319
aI1333
aI6207
aI2265
aI4445
aI50831
aI115893
aI120491
asS'd'
I18377
sa(dp36754
S'a'
I83475
sS's'
I18
sS'm'
(lp36755
I1
aI3
aI1
aI13
aI27
aI23
aI25
aI23
aI187
aI51
aI1257
aI379
aI921
aI3801
aI24537
aI59547
aI34191
aI184625
asS'd'
I18378
sa(dp36756
S'a'
I83484
sS's'
I18
sS'm'
(lp36757
I1
aI1
aI5
aI3
aI23
aI21
aI23
aI159
aI163
aI537
aI1589
aI2797
aI8007
aI6767
aI31331
aI20741
aI119969
aI174135
asS'd'
I18379
sa(dp36758
S'a'
I83488
sS's'
I18
sS'm'
(lp36759
I1
aI3
aI1
aI1
aI13
aI35
aI73
aI147
aI491
aI317
aI69
aI1069
aI5413
aI13973
aI19741
aI44717
aI63263
aI77145
asS'd'
I18380
sa(dp36760
S'a'
I83493
sS's'
I18
sS'm'
(lp36761
I1
aI3
aI1
aI15
aI31
aI41
aI23
aI79
aI55
aI863
aI129
aI2229
aI3395
aI1621
aI6273
aI44521
aI100047
aI42337
asS'd'
I18381
sa(dp36762
S'a'
I83512
sS's'
I18
sS'm'
(lp36763
I1
aI3
aI7
aI3
aI7
aI5
aI79
aI1
aI191
aI227
aI1039
aI2909
aI1085
aI3173
aI29311
aI13861
aI124785
aI212453
asS'd'
I18382
sa(dp36764
S'a'
I83525
sS's'
I18
sS'm'
(lp36765
I1
aI3
aI1
aI5
aI13
aI9
aI99
aI213
aI61
aI201
aI889
aI1171
aI3981
aI2091
aI31679
aI26643
aI5611
aI154339
asS'd'
I18383
sa(dp36766
S'a'
I83535
sS's'
I18
sS'm'
(lp36767
I1
aI3
aI3
aI7
aI27
aI49
aI53
aI77
aI285
aI441
aI1669
aI2157
aI223
aI1899
aI2725
aI36547
aI39273
aI206653
asS'd'
I18384
sa(dp36768
S'a'
I83560
sS's'
I18
sS'm'
(lp36769
I1
aI3
aI3
aI9
aI29
aI5
aI91
aI1
aI13
aI409
aI1275
aI891
aI6557
aI5157
aI6481
aI57381
aI87683
aI117277
asS'd'
I18385
sa(dp36770
S'a'
I83617
sS's'
I18
sS'm'
(lp36771
I1
aI3
aI3
aI1
aI11
aI7
aI13
aI69
aI9
aI1015
aI907
aI2685
aI6665
aI16307
aI24567
aI13191
aI9567
aI55073
asS'd'
I18386
sa(dp36772
S'a'
I83664
sS's'
I18
sS'm'
(lp36773
I1
aI3
aI1
aI5
aI19
aI15
aI65
aI13
aI503
aI427
aI1947
aI1869
aI5857
aI823
aI20533
aI25337
aI83551
aI128505
asS'd'
I18387
sa(dp36774
S'a'
I83676
sS's'
I18
sS'm'
(lp36775
I1
aI1
aI5
aI11
aI25
aI53
aI83
aI175
aI445
aI5
aI841
aI2773
aI4381
aI2829
aI1927
aI63689
aI63643
aI246629
asS'd'
I18388
sa(dp36776
S'a'
I83757
sS's'
I18
sS'm'
(lp36777
I1
aI3
aI1
aI11
aI27
aI63
aI43
aI95
aI453
aI235
aI673
aI117
aI6617
aI7589
aI5767
aI16465
aI36961
aI39395
asS'd'
I18389
sa(dp36778
S'a'
I83770
sS's'
I18
sS'm'
(lp36779
I1
aI1
aI3
aI13
aI3
aI27
aI119
aI87
aI209
aI167
aI721
aI1499
aI1955
aI9151
aI11649
aI29009
aI25249
aI26125
asS'd'
I18390
sa(dp36780
S'a'
I83775
sS's'
I18
sS'm'
(lp36781
I1
aI1
aI5
aI1
aI9
aI59
aI47
aI57
aI81
aI243
aI485
aI559
aI7311
aI15119
aI9827
aI47219
aI5941
aI16909
asS'd'
I18391
sa(dp36782
S'a'
I83835
sS's'
I18
sS'm'
(lp36783
I1
aI3
aI7
aI11
aI29
aI13
aI97
aI63
aI289
aI653
aI1811
aI835
aI801
aI13103
aI9333
aI7785
aI111587
aI10021
asS'd'
I18392
sa(dp36784
S'a'
I83838
sS's'
I18
sS'm'
(lp36785
I1
aI3
aI3
aI1
aI27
aI61
aI73
aI165
aI279
aI239
aI865
aI517
aI7763
aI1917
aI9839
aI20725
aI50721
aI171351
asS'd'
I18393
sa(dp36786
S'a'
I83848
sS's'
I18
sS'm'
(lp36787
I1
aI3
aI7
aI1
aI27
aI29
aI43
aI137
aI353
aI927
aI889
aI2511
aI709
aI3309
aI967
aI18119
aI48099
aI98139
asS'd'
I18394
sa(dp36788
S'a'
I83856
sS's'
I18
sS'm'
(lp36789
I1
aI3
aI1
aI1
aI17
aI5
aI79
aI23
aI367
aI231
aI605
aI3809
aI7557
aI14283
aI18417
aI15775
aI107421
aI9587
asS'd'
I18395
sa(dp36790
S'a'
I83878
sS's'
I18
sS'm'
(lp36791
I1
aI1
aI1
aI7
aI25
aI9
aI93
aI41
aI165
aI509
aI661
aI2165
aI3595
aI2555
aI11399
aI2403
aI76179
aI176003
asS'd'
I18396
sa(dp36792
S'a'
I83884
sS's'
I18
sS'm'
(lp36793
I1
aI3
aI3
aI7
aI27
aI19
aI55
aI213
aI83
aI601
aI377
aI2381
aI6831
aI5609
aI31321
aI26897
aI105321
aI144705
asS'd'
I18397
sa(dp36794
S'a'
I83896
sS's'
I18
sS'm'
(lp36795
I1
aI3
aI7
aI7
aI21
aI11
aI45
aI55
aI379
aI133
aI653
aI3593
aI7481
aI15789
aI12723
aI9697
aI20073
aI58211
asS'd'
I18398
sa(dp36796
S'a'
I83902
sS's'
I18
sS'm'
(lp36797
I1
aI3
aI5
aI1
aI1
aI57
aI89
aI159
aI461
aI719
aI1251
aI3899
aI1063
aI10753
aI6509
aI28391
aI129377
aI195279
asS'd'
I18399
sa(dp36798
S'a'
I83922
sS's'
I18
sS'm'
(lp36799
I1
aI3
aI7
aI7
aI5
aI39
aI59
aI81
aI27
aI169
aI1541
aI2213
aI3631
aI11601
aI13153
aI43221
aI14587
aI29719
asS'd'
I18400
sa(dp36800
S'a'
I83977
sS's'
I18
sS'm'
(lp36801
I1
aI1
aI1
aI1
aI3
aI29
aI103
aI125
aI35
aI455
aI255
aI3855
aI567
aI12013
aI13285
aI44753
aI117415
aI226285
asS'd'
I18401
sa(dp36802
S'a'
I83985
sS's'
I18
sS'm'
(lp36803
I1
aI1
aI5
aI1
aI21
aI41
aI59
aI69
aI83
aI813
aI1041
aI2559
aI1947
aI7343
aI5291
aI39281
aI56141
aI54487
asS'd'
I18402
sa(dp36804
S'a'
I83988
sS's'
I18
sS'm'
(lp36805
I1
aI1
aI7
aI15
aI25
aI17
aI83
aI115
aI321
aI659
aI1625
aI3253
aI281
aI6673
aI26301
aI45647
aI92151
aI150707
asS'd'
I18403
sa(dp36806
S'a'
I84034
sS's'
I18
sS'm'
(lp36807
I1
aI1
aI5
aI15
aI9
aI19
aI83
aI167
aI325
aI869
aI501
aI483
aI2155
aI14697
aI12755
aI54687
aI100637
aI6791
asS'd'
I18404
sa(dp36808
S'a'
I84046
sS's'
I18
sS'm'
(lp36809
I1
aI1
aI7
aI3
aI3
aI47
aI91
aI79
aI347
aI215
aI847
aI2957
aI5881
aI5371
aI20099
aI45603
aI29349
aI175357
asS'd'
I18405
sa(dp36810
S'a'
I84058
sS's'
I18
sS'm'
(lp36811
I1
aI3
aI1
aI3
aI13
aI43
aI101
aI235
aI505
aI289
aI691
aI673
aI5579
aI8721
aI9639
aI18569
aI44797
aI250887
asS'd'
I18406
sa(dp36812
S'a'
I84069
sS's'
I18
sS'm'
(lp36813
I1
aI1
aI3
aI11
aI23
aI27
aI85
aI223
aI365
aI767
aI577
aI2781
aI4179
aI12963
aI25235
aI51021
aI84989
aI149521
asS'd'
I18407
sa(dp36814
S'a'
I84107
sS's'
I18
sS'm'
(lp36815
I1
aI1
aI1
aI1
aI9
aI51
aI13
aI129
aI393
aI725
aI1301
aI1391
aI4693
aI4979
aI16801
aI21361
aI122157
aI56675
asS'd'
I18408
sa(dp36816
S'a'
I84155
sS's'
I18
sS'm'
(lp36817
I1
aI3
aI5
aI11
aI7
aI21
aI97
aI97
aI17
aI915
aI255
aI155
aI3961
aI7999
aI7493
aI52683
aI49377
aI131663
asS'd'
I18409
sa(dp36818
S'a'
I84157
sS's'
I18
sS'm'
(lp36819
I1
aI3
aI5
aI15
aI31
aI23
aI41
aI187
aI89
aI933
aI309
aI2519
aI6595
aI13785
aI14339
aI44393
aI64439
aI142105
asS'd'
I18410
sa(dp36820
S'a'
I84160
sS's'
I18
sS'm'
(lp36821
I1
aI1
aI1
aI11
aI13
aI57
aI29
aI249
aI467
aI863
aI77
aI3185
aI6221
aI13109
aI32397
aI13859
aI27331
aI35295
asS'd'
I18411
sa(dp36822
S'a'
I84165
sS's'
I18
sS'm'
(lp36823
I1
aI3
aI3
aI3
aI23
aI31
aI29
aI189
aI405
aI855
aI1597
aI3167
aI4171
aI13801
aI12297
aI38019
aI130141
aI135517
asS'd'
I18412
sa(dp36824
S'a'
I84187
sS's'
I18
sS'm'
(lp36825
I1
aI1
aI5
aI7
aI13
aI37
aI87
aI41
aI503
aI281
aI103
aI1997
aI3603
aI4185
aI25331
aI55123
aI74263
aI248695
asS'd'
I18413
sa(dp36826
S'a'
I84214
sS's'
I18
sS'm'
(lp36827
I1
aI3
aI1
aI15
aI13
aI57
aI67
aI135
aI429
aI489
aI829
aI2069
aI7657
aI15713
aI3907
aI5819
aI114005
aI187859
asS'd'
I18414
sa(dp36828
S'a'
I84226
sS's'
I18
sS'm'
(lp36829
I1
aI1
aI3
aI1
aI23
aI43
aI93
aI63
aI5
aI435
aI1649
aI1429
aI2923
aI9035
aI28667
aI13991
aI74491
aI236225
asS'd'
I18415
sa(dp36830
S'a'
I84235
sS's'
I18
sS'm'
(lp36831
I1
aI3
aI3
aI7
aI19
aI29
aI37
aI143
aI443
aI955
aI1431
aI3193
aI6023
aI2421
aI28955
aI29171
aI126785
aI124709
asS'd'
I18416
sa(dp36832
S'a'
I84240
sS's'
I18
sS'm'
(lp36833
I1
aI3
aI7
aI7
aI23
aI45
aI59
aI101
aI25
aI711
aI1685
aI851
aI3101
aI12273
aI10775
aI57633
aI52739
aI244681
asS'd'
I18417
sa(dp36834
S'a'
I84255
sS's'
I18
sS'm'
(lp36835
I1
aI1
aI1
aI5
aI3
aI13
aI97
aI143
aI367
aI139
aI1535
aI873
aI8005
aI2795
aI11103
aI3837
aI125833
aI194903
asS'd'
I18418
sa(dp36836
S'a'
I84273
sS's'
I18
sS'm'
(lp36837
I1
aI3
aI5
aI7
aI23
aI61
aI61
aI203
aI443
aI543
aI573
aI2835
aI941
aI12315
aI18453
aI34367
aI94359
aI132437
asS'd'
I18419
sa(dp36838
S'a'
I84291
sS's'
I18
sS'm'
(lp36839
I1
aI1
aI7
aI5
aI11
aI21
aI87
aI27
aI495
aI67
aI1267
aI2029
aI5041
aI4133
aI18821
aI50249
aI52397
aI101431
asS'd'
I18420
sa(dp36840
S'a'
I84311
sS's'
I18
sS'm'
(lp36841
I1
aI3
aI3
aI3
aI13
aI51
aI89
aI183
aI61
aI919
aI1841
aI373
aI7091
aI9413
aI1227
aI44515
aI72869
aI198769
asS'd'
I18421
sa(dp36842
S'a'
I84317
sS's'
I18
sS'm'
(lp36843
I1
aI3
aI3
aI11
aI13
aI63
aI13
aI253
aI203
aI571
aI91
aI3477
aI123
aI15353
aI7803
aI62729
aI14337
aI252725
asS'd'
I18422
sa(dp36844
S'a'
I84328
sS's'
I18
sS'm'
(lp36845
I1
aI1
aI7
aI1
aI11
aI57
aI45
aI251
aI351
aI895
aI1813
aI3857
aI7545
aI9739
aI32029
aI24915
aI46261
aI8149
asS'd'
I18423
sa(dp36846
S'a'
I84346
sS's'
I18
sS'm'
(lp36847
I1
aI3
aI1
aI15
aI25
aI41
aI71
aI47
aI265
aI567
aI307
aI4079
aI1943
aI10407
aI2999
aI6605
aI97621
aI194711
asS'd'
I18424
sa(dp36848
S'a'
I84357
sS's'
I18
sS'm'
(lp36849
I1
aI1
aI3
aI5
aI17
aI29
aI97
aI249
aI449
aI761
aI1727
aI1533
aI7417
aI16167
aI421
aI39075
aI1029
aI180923
asS'd'
I18425
sa(dp36850
S'a'
I84361
sS's'
I18
sS'm'
(lp36851
I1
aI1
aI1
aI3
aI11
aI27
aI67
aI227
aI131
aI453
aI951
aI3897
aI515
aI4513
aI17361
aI50049
aI4533
aI35953
asS'd'
I18426
sa(dp36852
S'a'
I84372
sS's'
I18
sS'm'
(lp36853
I1
aI1
aI1
aI13
aI7
aI53
aI25
aI163
aI453
aI195
aI1115
aI1019
aI3799
aI7489
aI12419
aI15141
aI112001
aI106459
asS'd'
I18427
sa(dp36854
S'a'
I84388
sS's'
I18
sS'm'
(lp36855
I1
aI1
aI3
aI9
aI27
aI15
aI63
aI109
aI293
aI867
aI645
aI1821
aI2867
aI9653
aI32617
aI39617
aI125589
aI249169
asS'd'
I18428
sa(dp36856
S'a'
I84415
sS's'
I18
sS'm'
(lp36857
I1
aI1
aI7
aI5
aI19
aI17
aI15
aI105
aI65
aI143
aI961
aI493
aI7301
aI11299
aI4549
aI49873
aI82447
aI107
asS'd'
I18429
sa(dp36858
S'a'
I84438
sS's'
I18
sS'm'
(lp36859
I1
aI1
aI7
aI1
aI31
aI21
aI19
aI61
aI255
aI815
aI421
aI3097
aI4993
aI9709
aI11529
aI53839
aI32653
aI137861
asS'd'
I18430
sa(dp36860
S'a'
I84444
sS's'
I18
sS'm'
(lp36861
I1
aI3
aI1
aI9
aI5
aI5
aI59
aI179
aI115
aI101
aI407
aI1143
aI309
aI843
aI31143
aI60639
aI126659
aI111695
asS'd'
I18431
sa(dp36862
S'a'
I84447
sS's'
I18
sS'm'
(lp36863
I1
aI3
aI7
aI13
aI7
aI47
aI65
aI127
aI159
aI817
aI1029
aI2983
aI5443
aI11087
aI10595
aI47143
aI128353
aI195189
asS'd'
I18432
sa(dp36864
S'a'
I84451
sS's'
I18
sS'm'
(lp36865
I1
aI3
aI7
aI1
aI27
aI21
aI61
aI235
aI433
aI929
aI581
aI1925
aI8185
aI6037
aI28859
aI16843
aI43499
aI217091
asS'd'
I18433
sa(dp36866
S'a'
I84465
sS's'
I18
sS'm'
(lp36867
I1
aI3
aI7
aI11
aI1
aI11
aI81
aI187
aI227
aI967
aI25
aI2285
aI1251
aI10743
aI2321
aI29029
aI89739
aI188023
asS'd'
I18434
sa(dp36868
S'a'
I84471
sS's'
I18
sS'm'
(lp36869
I1
aI3
aI5
aI11
aI21
aI25
aI57
aI201
aI89
aI965
aI1593
aI2879
aI2469
aI13675
aI28789
aI11407
aI13109
aI52749
asS'd'
I18435
sa(dp36870
S'a'
I84496
sS's'
I18
sS'm'
(lp36871
I1
aI3
aI3
aI3
aI19
aI25
aI87
aI3
aI127
aI881
aI645
aI207
aI1129
aI4235
aI1533
aI52503
aI128733
aI238679
asS'd'
I18436
sa(dp36872
S'a'
I84554
sS's'
I18
sS'm'
(lp36873
I1
aI3
aI3
aI1
aI5
aI5
aI63
aI181
aI493
aI457
aI1529
aI1795
aI219
aI10807
aI26713
aI49673
aI47167
aI103595
asS'd'
I18437
sa(dp36874
S'a'
I84561
sS's'
I18
sS'm'
(lp36875
I1
aI1
aI3
aI3
aI1
aI31
aI65
aI79
aI473
aI257
aI1477
aI387
aI2843
aI4031
aI8459
aI44849
aI115157
aI8417
asS'd'
I18438
sa(dp36876
S'a'
I84571
sS's'
I18
sS'm'
(lp36877
I1
aI3
aI3
aI5
aI15
aI1
aI105
aI67
aI343
aI333
aI1961
aI649
aI5105
aI11387
aI27437
aI35471
aI26295
aI220309
asS'd'
I18439
sa(dp36878
S'a'
I84580
sS's'
I18
sS'm'
(lp36879
I1
aI3
aI7
aI15
aI1
aI7
aI23
aI113
aI67
aI1019
aI1793
aI3237
aI7223
aI5691
aI6279
aI50231
aI49393
aI84393
asS'd'
I18440
sa(dp36880
S'a'
I84598
sS's'
I18
sS'm'
(lp36881
I1
aI1
aI1
aI7
aI9
aI29
aI125
aI249
aI89
aI813
aI561
aI871
aI1957
aI1095
aI18563
aI5257
aI39563
aI225651
asS'd'
I18441
sa(dp36882
S'a'
I84611
sS's'
I18
sS'm'
(lp36883
I1
aI1
aI1
aI15
aI17
aI11
aI51
aI191
aI217
aI617
aI793
aI3633
aI4673
aI15463
aI10621
aI47221
aI51611
aI155937
asS'd'
I18442
sa(dp36884
S'a'
I84620
sS's'
I18
sS'm'
(lp36885
I1
aI3
aI7
aI15
aI13
aI7
aI63
aI57
aI45
aI1005
aI685
aI2913
aI3597
aI9933
aI14819
aI26015
aI80023
aI60547
asS'd'
I18443
sa(dp36886
S'a'
I84625
sS's'
I18
sS'm'
(lp36887
I1
aI1
aI5
aI5
aI31
aI43
aI69
aI63
aI425
aI439
aI143
aI933
aI675
aI11301
aI31779
aI53445
aI25143
aI213213
asS'd'
I18444
sa(dp36888
S'a'
I84632
sS's'
I18
sS'm'
(lp36889
I1
aI1
aI1
aI5
aI21
aI9
aI91
aI89
aI483
aI153
aI389
aI7
aI633
aI15527
aI21833
aI45171
aI88331
aI150935
asS'd'
I18445
sa(dp36890
S'a'
I84648
sS's'
I18
sS'm'
(lp36891
I1
aI3
aI5
aI5
aI11
aI31
aI49
aI139
aI295
aI289
aI1623
aI3359
aI7551
aI11285
aI25083
aI27699
aI91869
aI237571
asS'd'
I18446
sa(dp36892
S'a'
I84659
sS's'
I18
sS'm'
(lp36893
I1
aI1
aI1
aI9
aI9
aI25
aI55
aI71
aI51
aI603
aI1901
aI2729
aI6803
aI11135
aI5427
aI37285
aI69141
aI262073
asS'd'
I18447
sa(dp36894
S'a'
I84662
sS's'
I18
sS'm'
(lp36895
I1
aI1
aI3
aI11
aI3
aI7
aI81
aI89
aI49
aI303
aI755
aI223
aI603
aI12525
aI26037
aI47867
aI118871
aI238677
asS'd'
I18448
sa(dp36896
S'a'
I84683
sS's'
I18
sS'm'
(lp36897
I1
aI3
aI3
aI11
aI1
aI53
aI55
aI15
aI341
aI151
aI245
aI1979
aI3523
aI15151
aI25075
aI21425
aI48689
aI125391
asS'd'
I18449
sa(dp36898
S'a'
I84697
sS's'
I18
sS'm'
(lp36899
I1
aI3
aI3
aI7
aI15
aI21
aI73
aI247
aI215
aI339
aI1995
aI633
aI2557
aI5625
aI28443
aI16413
aI34615
aI260591
asS'd'
I18450
sa(dp36900
S'a'
I84698
sS's'
I18
sS'm'
(lp36901
I1
aI3
aI5
aI1
aI1
aI59
aI21
aI247
aI403
aI15
aI1129
aI2263
aI3361
aI10675
aI30417
aI31285
aI69913
aI124329
asS'd'
I18451
sa(dp36902
S'a'
I84709
sS's'
I18
sS'm'
(lp36903
I1
aI1
aI5
aI1
aI17
aI3
aI103
aI107
aI333
aI191
aI345
aI3219
aI3845
aI5953
aI26403
aI51115
aI71623
aI52293
asS'd'
I18452
sa(dp36904
S'a'
I84722
sS's'
I18
sS'm'
(lp36905
I1
aI1
aI7
aI5
aI29
aI13
aI59
aI65
aI185
aI91
aI717
aI3179
aI1237
aI1187
aI25485
aI40119
aI6069
aI23567
asS'd'
I18453
sa(dp36906
S'a'
I84753
sS's'
I18
sS'm'
(lp36907
I1
aI1
aI1
aI5
aI17
aI57
aI27
aI39
aI269
aI627
aI1239
aI135
aI623
aI483
aI19229
aI51939
aI114387
aI146431
asS'd'
I18454
sa(dp36908
S'a'
I84775
sS's'
I18
sS'm'
(lp36909
I1
aI3
aI7
aI9
aI21
aI41
aI119
aI129
aI177
aI149
aI1527
aI3639
aI4489
aI11635
aI23007
aI59863
aI85199
aI87795
asS'd'
I18455
sa(dp36910
S'a'
I84789
sS's'
I18
sS'm'
(lp36911
I1
aI3
aI5
aI9
aI17
aI25
aI57
aI237
aI129
aI855
aI199
aI1929
aI2793
aI4277
aI4509
aI46301
aI32905
aI102015
asS'd'
I18456
sa(dp36912
S'a'
I84885
sS's'
I18
sS'm'
(lp36913
I1
aI1
aI7
aI1
aI15
aI23
aI3
aI131
aI475
aI347
aI1301
aI241
aI153
aI2801
aI29271
aI1337
aI107613
aI154105
asS'd'
I18457
sa(dp36914
S'a'
I84890
sS's'
I18
sS'm'
(lp36915
I1
aI1
aI1
aI13
aI5
aI19
aI43
aI47
aI381
aI709
aI637
aI2565
aI7503
aI10027
aI16873
aI23511
aI101785
aI47987
asS'd'
I18458
sa(dp36916
S'a'
I84899
sS's'
I18
sS'm'
(lp36917
I1
aI1
aI5
aI13
aI31
aI15
aI125
aI97
aI361
aI819
aI121
aI2723
aI3395
aI6943
aI5279
aI55977
aI103559
aI134177
asS'd'
I18459
sa(dp36918
S'a'
I84926
sS's'
I18
sS'm'
(lp36919
I1
aI1
aI7
aI13
aI17
aI27
aI105
aI11
aI327
aI203
aI1355
aI1437
aI959
aI10113
aI7405
aI43511
aI114073
aI199463
asS'd'
I18460
sa(dp36920
S'a'
I84946
sS's'
I18
sS'm'
(lp36921
I1
aI1
aI7
aI5
aI29
aI19
aI7
aI151
aI107
aI739
aI1021
aI1287
aI6881
aI2741
aI3407
aI13847
aI75669
aI116015
asS'd'
I18461
sa(dp36922
S'a'
I84958
sS's'
I18
sS'm'
(lp36923
I1
aI1
aI1
aI5
aI5
aI17
aI99
aI67
aI179
aI319
aI149
aI4069
aI7811
aI3055
aI24669
aI21635
aI68057
aI72059
asS'd'
I18462
sa(dp36924
S'a'
I84976
sS's'
I18
sS'm'
(lp36925
I1
aI3
aI1
aI11
aI3
aI3
aI103
aI45
aI431
aI159
aI1693
aI1069
aI3403
aI6121
aI12695
aI16565
aI29787
aI199327
asS'd'
I18463
sa(dp36926
S'a'
I84979
sS's'
I18
sS'm'
(lp36927
I1
aI1
aI5
aI5
aI5
aI35
aI97
aI9
aI7
aI703
aI1533
aI847
aI7693
aI16041
aI13127
aI26829
aI68801
aI205219
asS'd'
I18464
sa(dp36928
S'a'
I85003
sS's'
I18
sS'm'
(lp36929
I1
aI1
aI5
aI9
aI25
aI59
aI3
aI63
aI305
aI71
aI1429
aI1567
aI2377
aI12611
aI9267
aI62381
aI32373
aI187735
asS'd'
I18465
sa(dp36930
S'a'
I85014
sS's'
I18
sS'm'
(lp36931
I1
aI1
aI3
aI15
aI5
aI31
aI21
aI113
aI329
aI573
aI1975
aI1615
aI947
aI987
aI4655
aI46803
aI100251
aI89729
asS'd'
I18466
sa(dp36932
S'a'
I85017
sS's'
I18
sS'm'
(lp36933
I1
aI1
aI5
aI9
aI15
aI63
aI83
aI5
aI71
aI191
aI1127
aI3529
aI7325
aI1169
aI4255
aI6715
aI42765
aI73231
asS'd'
I18467
sa(dp36934
S'a'
I85024
sS's'
I18
sS'm'
(lp36935
I1
aI3
aI3
aI9
aI11
aI1
aI1
aI23
aI97
aI967
aI1465
aI1305
aI2073
aI3143
aI31333
aI1409
aI95321
aI182333
asS'd'
I18468
sa(dp36936
S'a'
I85029
sS's'
I18
sS'm'
(lp36937
I1
aI3
aI5
aI13
aI21
aI53
aI47
aI105
aI75
aI721
aI239
aI3619
aI2581
aI2063
aI21227
aI25579
aI23729
aI20067
asS'd'
I18469
sa(dp36938
S'a'
I85036
sS's'
I18
sS'm'
(lp36939
I1
aI3
aI5
aI13
aI1
aI55
aI55
aI115
aI391
aI539
aI869
aI3347
aI189
aI11087
aI11533
aI18747
aI25387
aI19205
asS'd'
I18470
sa(dp36940
S'a'
I85042
sS's'
I18
sS'm'
(lp36941
I1
aI3
aI1
aI15
aI25
aI57
aI81
aI27
aI379
aI635
aI1697
aI2805
aI8071
aI11407
aI14843
aI17593
aI20819
aI42891
asS'd'
I18471
sa(dp36942
S'a'
I85054
sS's'
I18
sS'm'
(lp36943
I1
aI3
aI5
aI11
aI1
aI59
aI51
aI187
aI11
aI211
aI1425
aI3829
aI3193
aI15743
aI16479
aI4205
aI108205
aI205367
asS'd'
I18472
sa(dp36944
S'a'
I85056
sS's'
I18
sS'm'
(lp36945
I1
aI1
aI1
aI5
aI7
aI7
aI59
aI85
aI63
aI509
aI897
aI2473
aI7345
aI111
aI4431
aI55273
aI114037
aI232541
asS'd'
I18473
sa(dp36946
S'a'
I85061
sS's'
I18
sS'm'
(lp36947
I1
aI1
aI5
aI9
aI29
aI7
aI25
aI41
aI401
aI843
aI115
aI163
aI6835
aI13943
aI5223
aI31033
aI10813
aI250471
asS'd'
I18474
sa(dp36948
S'a'
I85066
sS's'
I18
sS'm'
(lp36949
I1
aI1
aI5
aI11
aI27
aI45
aI43
aI233
aI195
aI151
aI11
aI1539
aI4775
aI15743
aI15507
aI26939
aI30353
aI162929
asS'd'
I18475
sa(dp36950
S'a'
I85073
sS's'
I18
sS'm'
(lp36951
I1
aI3
aI1
aI7
aI7
aI39
aI1
aI87
aI85
aI1019
aI1711
aI2707
aI735
aI5093
aI8231
aI25069
aI102861
aI45751
asS'd'
I18476
sa(dp36952
S'a'
I85096
sS's'
I18
sS'm'
(lp36953
I1
aI3
aI7
aI1
aI19
aI49
aI55
aI249
aI255
aI809
aI1799
aI3475
aI7697
aI5003
aI12437
aI52313
aI96355
aI138537
asS'd'
I18477
sa(dp36954
S'a'
I85101
sS's'
I18
sS'm'
(lp36955
I1
aI1
aI3
aI5
aI5
aI43
aI25
aI95
aI349
aI775
aI213
aI3643
aI1355
aI7745
aI9553
aI53367
aI123655
aI195365
asS'd'
I18478
sa(dp36956
S'a'
I85109
sS's'
I18
sS'm'
(lp36957
I1
aI1
aI1
aI1
aI19
aI53
aI39
aI105
aI449
aI447
aI147
aI2293
aI7817
aI1503
aI31985
aI37193
aI51039
aI209083
asS'd'
I18479
sa(dp36958
S'a'
I85114
sS's'
I18
sS'm'
(lp36959
I1
aI3
aI3
aI15
aI27
aI15
aI59
aI51
aI411
aI543
aI421
aI3595
aI2091
aI7171
aI23595
aI33509
aI37283
aI105719
asS'd'
I18480
sa(dp36960
S'a'
I85119
sS's'
I18
sS'm'
(lp36961
I1
aI3
aI1
aI11
aI5
aI37
aI5
aI203
aI171
aI853
aI1875
aI2735
aI4003
aI15163
aI26193
aI36149
aI31389
aI256631
asS'd'
I18481
sa(dp36962
S'a'
I85123
sS's'
I18
sS'm'
(lp36963
I1
aI3
aI5
aI1
aI21
aI1
aI127
aI41
aI185
aI929
aI1757
aI2711
aI2947
aI9709
aI18401
aI45037
aI1371
aI242397
asS'd'
I18482
sa(dp36964
S'a'
I85149
sS's'
I18
sS'm'
(lp36965
I1
aI1
aI1
aI1
aI13
aI43
aI51
aI187
aI487
aI759
aI1579
aI959
aI2499
aI4781
aI27179
aI6839
aI43869
aI36163
asS'd'
I18483
sa(dp36966
S'a'
I85185
sS's'
I18
sS'm'
(lp36967
I1
aI3
aI1
aI15
aI5
aI3
aI101
aI25
aI181
aI107
aI1105
aI879
aI5341
aI12215
aI21615
aI9619
aI129591
aI108393
asS'd'
I18484
sa(dp36968
S'a'
I85228
sS's'
I18
sS'm'
(lp36969
I1
aI1
aI7
aI1
aI11
aI39
aI55
aI101
aI43
aI935
aI1703
aI1269
aI6751
aI13723
aI7463
aI10055
aI112971
aI72789
asS'd'
I18485
sa(dp36970
S'a'
I85236
sS's'
I18
sS'm'
(lp36971
I1
aI3
aI5
aI13
aI7
aI3
aI81
aI41
aI55
aI375
aI663
aI801
aI5051
aI14583
aI30793
aI63897
aI127255
aI174179
asS'd'
I18486
sa(dp36972
S'a'
I85281
sS's'
I18
sS'm'
(lp36973
I1
aI3
aI1
aI7
aI19
aI31
aI53
aI29
aI313
aI57
aI1411
aI103
aI6863
aI10673
aI4341
aI5587
aI106059
aI222309
asS'd'
I18487
sa(dp36974
S'a'
I85282
sS's'
I18
sS'm'
(lp36975
I1
aI3
aI7
aI3
aI1
aI33
aI89
aI199
aI91
aI557
aI715
aI2715
aI4753
aI5987
aI30355
aI13819
aI57443
aI112179
asS'd'
I18488
sa(dp36976
S'a'
I85293
sS's'
I18
sS'm'
(lp36977
I1
aI3
aI5
aI7
aI13
aI27
aI55
aI11
aI495
aI29
aI1273
aI1727
aI3397
aI2739
aI22907
aI46203
aI16687
aI47385
asS'd'
I18489
sa(dp36978
S'a'
I85302
sS's'
I18
sS'm'
(lp36979
I1
aI3
aI5
aI9
aI13
aI23
aI23
aI107
aI353
aI429
aI359
aI2667
aI6137
aI7213
aI7977
aI35903
aI118507
aI209243
asS'd'
I18490
sa(dp36980
S'a'
I85319
sS's'
I18
sS'm'
(lp36981
I1
aI3
aI1
aI13
aI9
aI7
aI37
aI135
aI377
aI753
aI1819
aI113
aI7379
aI2795
aI10373
aI7131
aI17845
aI246101
asS'd'
I18491
sa(dp36982
S'a'
I85353
sS's'
I18
sS'm'
(lp36983
I1
aI1
aI3
aI11
aI11
aI45
aI53
aI209
aI49
aI385
aI1573
aI1129
aI2939
aI10949
aI413
aI59193
aI15399
aI169355
asS'd'
I18492
sa(dp36984
S'a'
I85354
sS's'
I18
sS'm'
(lp36985
I1
aI1
aI5
aI15
aI29
aI29
aI89
aI139
aI403
aI11
aI1335
aI2601
aI3631
aI15317
aI1707
aI3517
aI1939
aI121997
asS'd'
I18493
sa(dp36986
S'a'
I85356
sS's'
I18
sS'm'
(lp36987
I1
aI3
aI1
aI7
aI17
aI3
aI113
aI97
aI435
aI911
aI1743
aI1649
aI4829
aI9995
aI2873
aI17527
aI46931
aI86199
asS'd'
I18494
sa(dp36988
S'a'
I85389
sS's'
I18
sS'm'
(lp36989
I1
aI1
aI7
aI1
aI13
aI19
aI83
aI51
aI49
aI671
aI1651
aI3443
aI2279
aI5677
aI8859
aI41945
aI110607
aI200469
asS'd'
I18495
sa(dp36990
S'a'
I85418
sS's'
I18
sS'm'
(lp36991
I1
aI1
aI5
aI13
aI19
aI63
aI19
aI73
aI205
aI571
aI507
aI1781
aI1489
aI5909
aI10351
aI64607
aI67023
aI49441
asS'd'
I18496
sa(dp36992
S'a'
I85446
sS's'
I18
sS'm'
(lp36993
I1
aI1
aI3
aI5
aI13
aI35
aI107
aI63
aI489
aI69
aI1541
aI3761
aI17
aI9317
aI20323
aI35401
aI61451
aI116115
asS'd'
I18497
sa(dp36994
S'a'
I85450
sS's'
I18
sS'm'
(lp36995
I1
aI3
aI1
aI5
aI31
aI35
aI39
aI119
aI237
aI533
aI107
aI3235
aI4929
aI15839
aI9309
aI50131
aI110945
aI24739
asS'd'
I18498
sa(dp36996
S'a'
I85463
sS's'
I18
sS'm'
(lp36997
I1
aI3
aI1
aI5
aI3
aI25
aI73
aI145
aI391
aI481
aI1927
aI3071
aI4347
aI13415
aI26723
aI51629
aI3003
aI54575
asS'd'
I18499
sa(dp36998
S'a'
I85497
sS's'
I18
sS'm'
(lp36999
I1
aI3
aI3
aI11
aI29
aI63
aI63
aI183
aI11
aI269
aI153
aI3379
aI5603
aI14279
aI28579
aI4653
aI98179
aI125693
asS'd'
I18500
sa(dp37000
S'a'
I85534
sS's'
I18
sS'm'
(lp37001
I1
aI1
aI5
aI15
aI1
aI51
aI101
aI69
aI177
aI233
aI213
aI2389
aI4963
aI3391
aI13419
aI41283
aI25667
aI187239
asS'd'
I18501
sa(dp37002
S'a'
I85543
sS's'
I18
sS'm'
(lp37003
I1
aI1
aI1
aI11
aI13
aI29
aI95
aI231
aI481
aI283
aI1323
aI521
aI4689
aI5311
aI21949
aI31851
aI115845
aI50433
asS'd'
I18502
sa(dp37004
S'a'
I85549
sS's'
I18
sS'm'
(lp37005
I1
aI3
aI7
aI15
aI11
aI31
aI67
aI207
aI57
aI439
aI1561
aI2167
aI673
aI6467
aI8189
aI31783
aI5051
aI64097
asS'd'
I18503
sa(dp37006
S'a'
I85562
sS's'
I18
sS'm'
(lp37007
I1
aI1
aI7
aI9
aI17
aI57
aI77
aI85
aI119
aI149
aI211
aI2727
aI4921
aI8701
aI23121
aI36355
aI9179
aI68003
asS'd'
I18504
sa(dp37008
S'a'
I85587
sS's'
I18
sS'm'
(lp37009
I1
aI1
aI5
aI15
aI1
aI15
aI123
aI205
aI79
aI299
aI71
aI3413
aI7635
aI5699
aI32393
aI10253
aI86205
aI216015
asS'd'
I18505
sa(dp37010
S'a'
I85590
sS's'
I18
sS'm'
(lp37011
I1
aI1
aI5
aI11
aI23
aI39
aI105
aI187
aI487
aI247
aI333
aI2423
aI5643
aI8111
aI23549
aI50153
aI122859
aI100361
asS'd'
I18506
sa(dp37012
S'a'
I85599
sS's'
I18
sS'm'
(lp37013
I1
aI1
aI1
aI9
aI11
aI33
aI65
aI125
aI67
aI743
aI1331
aI1563
aI6333
aI11375
aI15873
aI18137
aI52765
aI224889
asS'd'
I18507
sa(dp37014
S'a'
I85630
sS's'
I18
sS'm'
(lp37015
I1
aI3
aI5
aI9
aI15
aI17
aI81
aI165
aI3
aI609
aI635
aI2093
aI6635
aI8647
aI25883
aI18907
aI73333
aI80835
asS'd'
I18508
sa(dp37016
S'a'
I85660
sS's'
I18
sS'm'
(lp37017
I1
aI1
aI1
aI11
aI1
aI1
aI25
aI115
aI205
aI941
aI1917
aI1295
aI3659
aI821
aI11355
aI1435
aI40289
aI115627
asS'd'
I18509
sa(dp37018
S'a'
I85669
sS's'
I18
sS'm'
(lp37019
I1
aI3
aI1
aI9
aI13
aI23
aI35
aI105
aI441
aI777
aI1255
aI3315
aI1157
aI8719
aI9939
aI38931
aI120723
aI123201
asS'd'
I18510
sa(dp37020
S'a'
I85693
sS's'
I18
sS'm'
(lp37021
I1
aI1
aI5
aI3
aI1
aI21
aI95
aI143
aI23
aI233
aI73
aI1223
aI5619
aI8583
aI21417
aI61971
aI74565
aI116249
asS'd'
I18511
sa(dp37022
S'a'
I85711
sS's'
I18
sS'm'
(lp37023
I1
aI1
aI7
aI7
aI23
aI35
aI21
aI201
aI441
aI623
aI419
aI2375
aI1189
aI15681
aI29469
aI29437
aI124525
aI241899
asS'd'
I18512
sa(dp37024
S'a'
I85716
sS's'
I18
sS'm'
(lp37025
I1
aI1
aI3
aI1
aI11
aI7
aI23
aI171
aI435
aI467
aI1811
aI63
aI3705
aI9395
aI579
aI58305
aI86165
aI67805
asS'd'
I18513
sa(dp37026
S'a'
I85773
sS's'
I18
sS'm'
(lp37027
I1
aI3
aI1
aI3
aI13
aI11
aI107
aI243
aI163
aI79
aI815
aI1149
aI2247
aI12411
aI30287
aI56915
aI26939
aI85883
asS'd'
I18514
sa(dp37028
S'a'
I85801
sS's'
I18
sS'm'
(lp37029
I1
aI3
aI3
aI1
aI5
aI23
aI105
aI63
aI35
aI57
aI1815
aI3325
aI3727
aI3623
aI7203
aI8301
aI28073
aI190053
asS'd'
I18515
sa(dp37030
S'a'
I85815
sS's'
I18
sS'm'
(lp37031
I1
aI3
aI1
aI11
aI23
aI33
aI121
aI55
aI287
aI139
aI491
aI907
aI4237
aI17
aI20055
aI63729
aI7517
aI151597
asS'd'
I18516
sa(dp37032
S'a'
I85839
sS's'
I18
sS'm'
(lp37033
I1
aI3
aI5
aI5
aI21
aI1
aI37
aI19
aI159
aI1013
aI27
aI2627
aI851
aI14021
aI31311
aI5871
aI77613
aI125311
asS'd'
I18517
sa(dp37034
S'a'
I85842
sS's'
I18
sS'm'
(lp37035
I1
aI3
aI3
aI1
aI27
aI3
aI51
aI133
aI459
aI581
aI383
aI1351
aI6149
aI15611
aI2631
aI20797
aI65955
aI113665
asS'd'
I18518
sa(dp37036
S'a'
I85864
sS's'
I18
sS'm'
(lp37037
I1
aI3
aI1
aI11
aI23
aI61
aI75
aI217
aI283
aI405
aI767
aI1151
aI7501
aI5553
aI113
aI48331
aI49379
aI191473
asS'd'
I18519
sa(dp37038
S'a'
I85867
sS's'
I18
sS'm'
(lp37039
I1
aI3
aI5
aI5
aI15
aI3
aI19
aI27
aI497
aI519
aI1611
aI709
aI405
aI13329
aI27861
aI14981
aI47197
aI173979
asS'd'
I18520
sa(dp37040
S'a'
I85903
sS's'
I18
sS'm'
(lp37041
I1
aI1
aI7
aI7
aI25
aI19
aI99
aI219
aI349
aI713
aI1421
aI3427
aI153
aI13319
aI22415
aI48617
aI119637
aI20835
asS'd'
I18521
sa(dp37042
S'a'
I85917
sS's'
I18
sS'm'
(lp37043
I1
aI1
aI1
aI3
aI13
aI37
aI43
aI9
aI317
aI961
aI1255
aI2975
aI2775
aI12283
aI29941
aI57495
aI77413
aI256695
asS'd'
I18522
sa(dp37044
S'a'
I85921
sS's'
I18
sS'm'
(lp37045
I1
aI1
aI7
aI5
aI29
aI37
aI91
aI199
aI397
aI739
aI877
aI251
aI847
aI2951
aI19497
aI57285
aI76891
aI258711
asS'd'
I18523
sa(dp37046
S'a'
I85928
sS's'
I18
sS'm'
(lp37047
I1
aI1
aI7
aI15
aI21
aI55
aI107
aI115
aI481
aI845
aI2015
aI481
aI3823
aI14071
aI4037
aI39687
aI62867
aI170891
asS'd'
I18524
sa(dp37048
S'a'
I85934
sS's'
I18
sS'm'
(lp37049
I1
aI3
aI3
aI1
aI17
aI51
aI81
aI195
aI189
aI455
aI1343
aI1493
aI351
aI361
aI20289
aI37423
aI7747
aI245861
asS'd'
I18525
sa(dp37050
S'a'
I85974
sS's'
I18
sS'm'
(lp37051
I1
aI1
aI5
aI15
aI9
aI41
aI71
aI155
aI197
aI563
aI1271
aI2227
aI2557
aI6657
aI13565
aI8467
aI96135
aI5903
asS'd'
I18526
sa(dp37052
S'a'
I85994
sS's'
I18
sS'm'
(lp37053
I1
aI1
aI5
aI13
aI29
aI33
aI115
aI131
aI283
aI435
aI1327
aI1113
aI4729
aI14125
aI23743
aI40121
aI119955
aI237453
asS'd'
I18527
sa(dp37054
S'a'
I85999
sS's'
I18
sS'm'
(lp37055
I1
aI1
aI1
aI1
aI13
aI21
aI109
aI91
aI105
aI749
aI1695
aI1123
aI4349
aI9855
aI31565
aI64001
aI7919
aI83591
asS'd'
I18528
sa(dp37056
S'a'
I86004
sS's'
I18
sS'm'
(lp37057
I1
aI3
aI1
aI1
aI13
aI19
aI89
aI83
aI13
aI609
aI731
aI2655
aI1123
aI13415
aI5645
aI10003
aI69381
aI187621
asS'd'
I18529
sa(dp37058
S'a'
I86008
sS's'
I18
sS'm'
(lp37059
I1
aI3
aI1
aI15
aI3
aI37
aI1
aI139
aI11
aI917
aI1191
aI1381
aI6035
aI13851
aI4565
aI5427
aI117703
aI109965
asS'd'
I18530
sa(dp37060
S'a'
I86046
sS's'
I18
sS'm'
(lp37061
I1
aI1
aI5
aI3
aI17
aI37
aI15
aI115
aI19
aI65
aI1807
aI3879
aI2709
aI9819
aI11457
aI53705
aI14821
aI156079
asS'd'
I18531
sa(dp37062
S'a'
I86067
sS's'
I18
sS'm'
(lp37063
I1
aI1
aI7
aI13
aI17
aI29
aI105
aI77
aI127
aI457
aI1287
aI1533
aI6879
aI4001
aI4083
aI29523
aI81175
aI226409
asS'd'
I18532
sa(dp37064
S'a'
I86084
sS's'
I18
sS'm'
(lp37065
I1
aI3
aI1
aI5
aI5
aI59
aI21
aI253
aI459
aI733
aI409
aI3359
aI1913
aI8893
aI16113
aI61063
aI6511
aI22441
asS'd'
I18533
sa(dp37066
S'a'
I86087
sS's'
I18
sS'm'
(lp37067
I1
aI3
aI5
aI11
aI7
aI1
aI121
aI217
aI63
aI83
aI173
aI1869
aI7931
aI655
aI21053
aI20703
aI116853
aI131785
asS'd'
I18534
sa(dp37068
S'a'
I86093
sS's'
I18
sS'm'
(lp37069
I1
aI1
aI7
aI7
aI5
aI13
aI41
aI57
aI1
aI17
aI649
aI233
aI2867
aI5577
aI30553
aI7635
aI45305
aI47979
asS'd'
I18535
sa(dp37070
S'a'
I86121
sS's'
I18
sS'm'
(lp37071
I1
aI1
aI7
aI1
aI1
aI29
aI61
aI241
aI107
aI891
aI49
aI3433
aI5045
aI143
aI22473
aI29243
aI82625
aI184163
asS'd'
I18536
sa(dp37072
S'a'
I86122
sS's'
I18
sS'm'
(lp37073
I1
aI3
aI1
aI11
aI11
aI21
aI119
aI43
aI117
aI429
aI1569
aI637
aI67
aI9475
aI31779
aI2237
aI122037
aI245361
asS'd'
I18537
sa(dp37074
S'a'
I86127
sS's'
I18
sS'm'
(lp37075
I1
aI1
aI7
aI1
aI31
aI61
aI9
aI179
aI467
aI153
aI1913
aI2839
aI6255
aI12715
aI28229
aI20189
aI3617
aI213539
asS'd'
I18538
sa(dp37076
S'a'
I86151
sS's'
I18
sS'm'
(lp37077
I1
aI1
aI7
aI13
aI25
aI61
aI9
aI109
aI331
aI577
aI21
aI1017
aI6521
aI5991
aI26573
aI56881
aI58455
aI169407
asS'd'
I18539
sa(dp37078
S'a'
I86166
sS's'
I18
sS'm'
(lp37079
I1
aI1
aI7
aI9
aI31
aI57
aI51
aI41
aI327
aI859
aI1295
aI1577
aI1071
aI3277
aI11685
aI62129
aI34111
aI174639
asS'd'
I18540
sa(dp37080
S'a'
I86179
sS's'
I18
sS'm'
(lp37081
I1
aI3
aI5
aI1
aI7
aI63
aI35
aI165
aI43
aI943
aI1545
aI3717
aI1471
aI11579
aI29637
aI22913
aI8867
aI12837
asS'd'
I18541
sa(dp37082
S'a'
I86188
sS's'
I18
sS'm'
(lp37083
I1
aI1
aI3
aI1
aI23
aI7
aI19
aI151
aI359
aI347
aI1085
aI3923
aI1039
aI5149
aI6047
aI49811
aI33099
aI247983
asS'd'
I18542
sa(dp37084
S'a'
I86193
sS's'
I18
sS'm'
(lp37085
I1
aI1
aI1
aI5
aI29
aI23
aI73
aI189
aI59
aI865
aI1499
aI1953
aI1261
aI1071
aI26761
aI26145
aI129427
aI223219
asS'd'
I18543
sa(dp37086
S'a'
I86206
sS's'
I18
sS'm'
(lp37087
I1
aI3
aI1
aI9
aI1
aI29
aI107
aI173
aI387
aI703
aI193
aI1965
aI6233
aI10997
aI32697
aI31005
aI15415
aI94345
asS'd'
I18544
sa(dp37088
S'a'
I86214
sS's'
I18
sS'm'
(lp37089
I1
aI1
aI3
aI9
aI31
aI35
aI7
aI15
aI317
aI79
aI2045
aI1455
aI1559
aI15087
aI287
aI46665
aI37225
aI149017
asS'd'
I18545
sa(dp37090
S'a'
I86226
sS's'
I18
sS'm'
(lp37091
I1
aI1
aI5
aI13
aI17
aI27
aI11
aI107
aI47
aI803
aI1487
aI3049
aI1171
aI6237
aI9157
aI10037
aI122349
aI236877
asS'd'
I18546
sa(dp37092
S'a'
I86242
sS's'
I18
sS'm'
(lp37093
I1
aI1
aI7
aI9
aI21
aI35
aI53
aI139
aI165
aI73
aI1405
aI2941
aI3553
aI11945
aI2493
aI5743
aI63749
aI140535
asS'd'
I18547
sa(dp37094
S'a'
I86256
sS's'
I18
sS'm'
(lp37095
I1
aI3
aI7
aI11
aI7
aI57
aI41
aI187
aI483
aI499
aI687
aI117
aI4951
aI14709
aI17025
aI23027
aI94863
aI228465
asS'd'
I18548
sa(dp37096
S'a'
I86259
sS's'
I18
sS'm'
(lp37097
I1
aI3
aI7
aI7
aI27
aI29
aI85
aI117
aI201
aI637
aI823
aI1135
aI7595
aI3323
aI23579
aI40759
aI25087
aI995
asS'd'
I18549
sa(dp37098
S'a'
I86298
sS's'
I18
sS'm'
(lp37099
I1
aI3
aI1
aI9
aI31
aI53
aI101
aI29
aI381
aI101
aI1939
aI1973
aI8191
aI8155
aI13881
aI32309
aI92907
aI239525
asS'd'
I18550
sa(dp37100
S'a'
I86300
sS's'
I18
sS'm'
(lp37101
I1
aI1
aI5
aI7
aI3
aI35
aI15
aI207
aI1
aI47
aI325
aI559
aI3377
aI3909
aI31225
aI28367
aI63891
aI19129
asS'd'
I18551
sa(dp37102
S'a'
I86307
sS's'
I18
sS'm'
(lp37103
I1
aI1
aI1
aI5
aI31
aI61
aI117
aI211
aI127
aI969
aI73
aI1295
aI7167
aI14881
aI9965
aI28143
aI28161
aI131867
asS'd'
I18552
sa(dp37104
S'a'
I86310
sS's'
I18
sS'm'
(lp37105
I1
aI3
aI7
aI5
aI5
aI57
aI37
aI207
aI201
aI79
aI1151
aI3685
aI2071
aI1751
aI5481
aI51447
aI103437
aI154895
asS'd'
I18553
sa(dp37106
S'a'
I86319
sS's'
I18
sS'm'
(lp37107
I1
aI3
aI5
aI13
aI9
aI5
aI57
aI27
aI131
aI211
aI1481
aI2237
aI4227
aI6927
aI18625
aI49773
aI55399
aI15209
asS'd'
I18554
sa(dp37108
S'a'
I86321
sS's'
I18
sS'm'
(lp37109
I1
aI3
aI1
aI7
aI23
aI29
aI3
aI179
aI479
aI787
aI463
aI2041
aI2581
aI6281
aI1657
aI51433
aI93807
aI160047
asS'd'
I18555
sa(dp37110
S'a'
I86328
sS's'
I18
sS'm'
(lp37111
I1
aI1
aI3
aI13
aI25
aI35
aI33
aI231
aI385
aI479
aI335
aI3837
aI5517
aI13603
aI15623
aI46737
aI42507
aI208355
asS'd'
I18556
sa(dp37112
S'a'
I86356
sS's'
I18
sS'm'
(lp37113
I1
aI1
aI7
aI9
aI3
aI25
aI51
aI125
aI213
aI175
aI1575
aI1755
aI1843
aI14361
aI13155
aI22445
aI55435
aI62793
asS'd'
I18557
sa(dp37114
S'a'
I86360
sS's'
I18
sS'm'
(lp37115
I1
aI1
aI1
aI7
aI25
aI3
aI21
aI7
aI309
aI547
aI19
aI471
aI2679
aI16185
aI12149
aI41437
aI47625
aI75113
asS'd'
I18558
sa(dp37116
S'a'
I86375
sS's'
I18
sS'm'
(lp37117
I1
aI1
aI5
aI1
aI9
aI23
aI81
aI95
aI123
aI143
aI1111
aI9
aI3501
aI11897
aI26499
aI10009
aI48073
aI182529
asS'd'
I18559
sa(dp37118
S'a'
I86399
sS's'
I18
sS'm'
(lp37119
I1
aI3
aI1
aI1
aI1
aI17
aI15
aI157
aI129
aI1005
aI543
aI3917
aI3493
aI6537
aI26997
aI33217
aI7987
aI251635
asS'd'
I18560
sa(dp37120
S'a'
I86400
sS's'
I18
sS'm'
(lp37121
I1
aI3
aI1
aI13
aI11
aI41
aI9
aI19
aI173
aI751
aI491
aI1645
aI5205
aI9907
aI28503
aI61137
aI79727
aI200851
asS'd'
I18561
sa(dp37122
S'a'
I86403
sS's'
I18
sS'm'
(lp37123
I1
aI1
aI1
aI1
aI3
aI5
aI105
aI203
aI97
aI903
aI1507
aI2719
aI5275
aI1023
aI29595
aI42507
aI39893
aI151495
asS'd'
I18562
sa(dp37124
S'a'
I86430
sS's'
I18
sS'm'
(lp37125
I1
aI1
aI1
aI11
aI31
aI53
aI51
aI65
aI145
aI671
aI489
aI109
aI803
aI8541
aI4439
aI33893
aI98495
aI114955
asS'd'
I18563
sa(dp37126
S'a'
I86446
sS's'
I18
sS'm'
(lp37127
I1
aI1
aI7
aI11
aI27
aI63
aI117
aI235
aI497
aI841
aI1461
aI3757
aI1077
aI6997
aI9611
aI47453
aI20197
aI176939
asS'd'
I18564
sa(dp37128
S'a'
I86460
sS's'
I18
sS'm'
(lp37129
I1
aI3
aI3
aI3
aI31
aI17
aI85
aI145
aI377
aI225
aI1033
aI3017
aI735
aI5811
aI25503
aI25457
aI124623
aI51713
asS'd'
I18565
sa(dp37130
S'a'
I86465
sS's'
I18
sS'm'
(lp37131
I1
aI3
aI3
aI13
aI19
aI33
aI127
aI57
aI321
aI687
aI1651
aI3321
aI5051
aI8511
aI19609
aI49927
aI30499
aI102613
asS'd'
I18566
sa(dp37132
S'a'
I86486
sS's'
I18
sS'm'
(lp37133
I1
aI1
aI5
aI13
aI7
aI17
aI17
aI147
aI381
aI137
aI1007
aI2607
aI1071
aI8921
aI13955
aI47223
aI130359
aI246265
asS'd'
I18567
sa(dp37134
S'a'
I86502
sS's'
I18
sS'm'
(lp37135
I1
aI3
aI7
aI9
aI13
aI35
aI71
aI129
aI57
aI233
aI357
aI3181
aI2841
aI3707
aI24947
aI57777
aI115133
aI6049
asS'd'
I18568
sa(dp37136
S'a'
I86563
sS's'
I18
sS'm'
(lp37137
I1
aI3
aI7
aI15
aI23
aI1
aI107
aI225
aI57
aI633
aI1515
aI1631
aI4303
aI4221
aI8281
aI59139
aI45023
aI70219
asS'd'
I18569
sa(dp37138
S'a'
I86595
sS's'
I18
sS'm'
(lp37139
I1
aI1
aI7
aI5
aI9
aI9
aI93
aI131
aI41
aI245
aI1261
aI459
aI4811
aI10987
aI10421
aI63839
aI34067
aI196353
asS'd'
I18570
sa(dp37140
S'a'
I86635
sS's'
I18
sS'm'
(lp37141
I1
aI3
aI3
aI13
aI17
aI3
aI87
aI255
aI167
aI701
aI821
aI1965
aI1415
aI4101
aI549
aI6347
aI92421
aI47193
asS'd'
I18571
sa(dp37142
S'a'
I86649
sS's'
I18
sS'm'
(lp37143
I1
aI3
aI7
aI11
aI23
aI17
aI51
aI81
aI71
aI345
aI971
aI917
aI1057
aI3627
aI20361
aI13491
aI12855
aI234215
asS'd'
I18572
sa(dp37144
S'a'
I86661
sS's'
I18
sS'm'
(lp37145
I1
aI1
aI1
aI13
aI17
aI9
aI25
aI155
aI463
aI851
aI243
aI3887
aI2445
aI7459
aI19915
aI21813
aI86969
aI85891
asS'd'
I18573
sa(dp37146
S'a'
I86674
sS's'
I18
sS'm'
(lp37147
I1
aI1
aI7
aI1
aI9
aI15
aI57
aI201
aI193
aI169
aI351
aI1355
aI1089
aI4705
aI15153
aI51359
aI49907
aI66007
asS'd'
I18574
sa(dp37148
S'a'
I86689
sS's'
I18
sS'm'
(lp37149
I1
aI3
aI1
aI9
aI19
aI13
aI69
aI83
aI39
aI667
aI1549
aI1503
aI7167
aI8657
aI17269
aI59357
aI80091
aI194007
asS'd'
I18575
sa(dp37150
S'a'
I86721
sS's'
I18
sS'm'
(lp37151
I1
aI3
aI5
aI5
aI17
aI37
aI125
aI117
aI355
aI685
aI637
aI3159
aI4783
aI3159
aI14953
aI12731
aI126759
aI89149
asS'd'
I18576
sa(dp37152
S'a'
I86734
sS's'
I18
sS'm'
(lp37153
I1
aI3
aI3
aI9
aI15
aI53
aI7
aI41
aI473
aI857
aI511
aI3741
aI6837
aI6167
aI26351
aI9885
aI104819
aI48221
asS'd'
I18577
sa(dp37154
S'a'
I86741
sS's'
I18
sS'm'
(lp37155
I1
aI3
aI7
aI15
aI21
aI21
aI21
aI101
aI465
aI223
aI13
aI1773
aI2763
aI8621
aI23591
aI12633
aI82143
aI134899
asS'd'
I18578
sa(dp37156
S'a'
I86742
sS's'
I18
sS'm'
(lp37157
I1
aI3
aI1
aI1
aI29
aI25
aI67
aI19
aI349
aI503
aI655
aI3567
aI97
aI6967
aI18253
aI42755
aI33041
aI250279
asS'd'
I18579
sa(dp37158
S'a'
I86784
sS's'
I18
sS'm'
(lp37159
I1
aI1
aI7
aI9
aI17
aI1
aI7
aI165
aI255
aI613
aI579
aI127
aI7567
aI13181
aI6255
aI1785
aI21527
aI113815
asS'd'
I18580
sa(dp37160
S'a'
I86799
sS's'
I18
sS'm'
(lp37161
I1
aI3
aI1
aI5
aI27
aI33
aI61
aI235
aI37
aI135
aI1515
aI3611
aI1825
aI9627
aI18805
aI37065
aI126107
aI23223
asS'd'
I18581
sa(dp37162
S'a'
I86808
sS's'
I18
sS'm'
(lp37163
I1
aI3
aI7
aI5
aI23
aI11
aI29
aI121
aI129
aI311
aI429
aI1653
aI5789
aI7693
aI18775
aI18189
aI97203
aI213501
asS'd'
I18582
sa(dp37164
S'a'
I86837
sS's'
I18
sS'm'
(lp37165
I1
aI3
aI7
aI3
aI29
aI61
aI87
aI197
aI43
aI509
aI5
aI3275
aI345
aI7885
aI4381
aI22059
aI1395
aI40125
asS'd'
I18583
sa(dp37166
S'a'
I86869
sS's'
I18
sS'm'
(lp37167
I1
aI3
aI5
aI1
aI1
aI59
aI69
aI125
aI297
aI983
aI641
aI2665
aI7045
aI8591
aI16581
aI58657
aI119189
aI256579
asS'd'
I18584
sa(dp37168
S'a'
I86900
sS's'
I18
sS'm'
(lp37169
I1
aI3
aI7
aI7
aI7
aI53
aI65
aI181
aI149
aI987
aI1377
aI4045
aI971
aI9827
aI17727
aI59357
aI90975
aI27395
asS'd'
I18585
sa(dp37170
S'a'
I86949
sS's'
I18
sS'm'
(lp37171
I1
aI1
aI7
aI7
aI1
aI51
aI109
aI165
aI361
aI515
aI739
aI3709
aI6431
aI113
aI21401
aI41743
aI53071
aI134205
asS'd'
I18586
sa(dp37172
S'a'
I86993
sS's'
I18
sS'm'
(lp37173
I1
aI3
aI1
aI13
aI5
aI51
aI107
aI99
aI135
aI163
aI1705
aI1683
aI6221
aI1377
aI2211
aI13379
aI22801
aI208753
asS'd'
I18587
sa(dp37174
S'a'
I86999
sS's'
I18
sS'm'
(lp37175
I1
aI3
aI5
aI11
aI11
aI39
aI49
aI45
aI503
aI549
aI821
aI4077
aI885
aI13721
aI29673
aI28435
aI6235
aI212071
asS'd'
I18588
sa(dp37176
S'a'
I87030
sS's'
I18
sS'm'
(lp37177
I1
aI3
aI3
aI11
aI15
aI25
aI17
aI67
aI125
aI7
aI1163
aI973
aI5325
aI12707
aI12763
aI9481
aI21363
aI195897
asS'd'
I18589
sa(dp37178
S'a'
I87041
sS's'
I18
sS'm'
(lp37179
I1
aI1
aI7
aI9
aI17
aI19
aI15
aI13
aI107
aI919
aI461
aI343
aI1101
aI8195
aI29293
aI61643
aI64995
aI230469
asS'd'
I18590
sa(dp37180
S'a'
I87044
sS's'
I18
sS'm'
(lp37181
I1
aI3
aI5
aI1
aI9
aI25
aI39
aI65
aI27
aI461
aI669
aI2841
aI7973
aI11565
aI9531
aI52235
aI6741
aI215513
asS'd'
I18591
sa(dp37182
S'a'
I87084
sS's'
I18
sS'm'
(lp37183
I1
aI3
aI3
aI1
aI7
aI57
aI101
aI199
aI37
aI79
aI2033
aI1723
aI6877
aI2733
aI26445
aI62625
aI21671
aI238431
asS'd'
I18592
sa(dp37184
S'a'
I87116
sS's'
I18
sS'm'
(lp37185
I1
aI1
aI1
aI9
aI27
aI31
aI125
aI199
aI331
aI611
aI523
aI407
aI747
aI9499
aI4685
aI17805
aI43717
aI253233
asS'd'
I18593
sa(dp37186
S'a'
I87138
sS's'
I18
sS'm'
(lp37187
I1
aI1
aI3
aI7
aI29
aI7
aI7
aI153
aI339
aI337
aI701
aI2639
aI6311
aI6375
aI26023
aI27693
aI59733
aI260405
asS'd'
I18594
sa(dp37188
S'a'
I87162
sS's'
I18
sS'm'
(lp37189
I1
aI1
aI7
aI7
aI15
aI27
aI23
aI49
aI181
aI433
aI485
aI2915
aI6021
aI9095
aI15951
aI47257
aI104513
aI208089
asS'd'
I18595
sa(dp37190
S'a'
I87188
sS's'
I18
sS'm'
(lp37191
I1
aI1
aI5
aI7
aI19
aI19
aI125
aI153
aI109
aI829
aI1967
aI2567
aI7157
aI6001
aI10151
aI55323
aI92405
aI82549
asS'd'
I18596
sa(dp37192
S'a'
I87211
sS's'
I18
sS'm'
(lp37193
I1
aI1
aI3
aI11
aI13
aI17
aI85
aI215
aI265
aI875
aI311
aI3773
aI8059
aI2115
aI19259
aI63999
aI77411
aI220267
asS'd'
I18597
sa(dp37194
S'a'
I87234
sS's'
I18
sS'm'
(lp37195
I1
aI3
aI1
aI15
aI15
aI35
aI81
aI213
aI411
aI435
aI105
aI1487
aI1991
aI14539
aI8175
aI2115
aI47259
aI45893
asS'd'
I18598
sa(dp37196
S'a'
I87245
sS's'
I18
sS'm'
(lp37197
I1
aI3
aI1
aI11
aI9
aI47
aI27
aI115
aI449
aI521
aI321
aI2463
aI1355
aI5785
aI11269
aI45337
aI29049
aI91675
asS'd'
I18599
sa(dp37198
S'a'
I87246
sS's'
I18
sS'm'
(lp37199
I1
aI3
aI1
aI13
aI21
aI53
aI49
aI83
aI373
aI519
aI757
aI1241
aI577
aI14443
aI449
aI44773
aI116673
aI155209
asS'd'
I18600
sa(dp37200
S'a'
I87260
sS's'
I18
sS'm'
(lp37201
I1
aI3
aI3
aI1
aI29
aI63
aI97
aI145
aI371
aI585
aI1809
aI3997
aI249
aI283
aI28369
aI27325
aI61673
aI12637
asS'd'
I18601
sa(dp37202
S'a'
I87270
sS's'
I18
sS'm'
(lp37203
I1
aI3
aI5
aI9
aI11
aI55
aI77
aI89
aI285
aI297
aI861
aI2791
aI3245
aI15093
aI32489
aI40477
aI97603
aI35347
asS'd'
I18602
sa(dp37204
S'a'
I87284
sS's'
I18
sS'm'
(lp37205
I1
aI1
aI1
aI3
aI27
aI33
aI115
aI209
aI169
aI893
aI393
aI1457
aI6069
aI12511
aI20423
aI11385
aI86711
aI197555
asS'd'
I18603
sa(dp37206
S'a'
I87293
sS's'
I18
sS'm'
(lp37207
I1
aI3
aI5
aI15
aI21
aI25
aI87
aI159
aI477
aI177
aI991
aI495
aI29
aI9347
aI9721
aI4071
aI30145
aI214155
asS'd'
I18604
sa(dp37208
S'a'
I87296
sS's'
I18
sS'm'
(lp37209
I1
aI1
aI5
aI3
aI11
aI45
aI53
aI251
aI177
aI651
aI549
aI3377
aI3247
aI8761
aI20339
aI27743
aI103387
aI159591
asS'd'
I18605
sa(dp37210
S'a'
I87299
sS's'
I18
sS'm'
(lp37211
I1
aI3
aI7
aI11
aI17
aI27
aI43
aI179
aI507
aI553
aI261
aI3939
aI6133
aI6347
aI12987
aI46071
aI42551
aI99225
asS'd'
I18606
sa(dp37212
S'a'
I87305
sS's'
I18
sS'm'
(lp37213
I1
aI3
aI3
aI11
aI3
aI33
aI85
aI51
aI277
aI117
aI1295
aI2435
aI1467
aI13787
aI2209
aI52673
aI53515
aI157625
asS'd'
I18607
sa(dp37214
S'a'
I87308
sS's'
I18
sS'm'
(lp37215
I1
aI3
aI3
aI3
aI9
aI15
aI121
aI229
aI227
aI795
aI541
aI3727
aI4333
aI2251
aI27833
aI43567
aI82505
aI230427
asS'd'
I18608
sa(dp37216
S'a'
I87313
sS's'
I18
sS'm'
(lp37217
I1
aI1
aI5
aI13
aI29
aI19
aI119
aI63
aI207
aI945
aI761
aI2601
aI1391
aI8939
aI11683
aI52433
aI63301
aI82501
asS'd'
I18609
sa(dp37218
S'a'
I87323
sS's'
I18
sS'm'
(lp37219
I1
aI3
aI7
aI1
aI3
aI57
aI127
aI115
aI209
aI31
aI1631
aI347
aI3937
aI4015
aI13313
aI49507
aI15103
aI237071
asS'd'
I18610
sa(dp37220
S'a'
I87332
sS's'
I18
sS'm'
(lp37221
I1
aI3
aI1
aI13
aI3
aI25
aI85
aI151
aI115
aI385
aI303
aI2453
aI2417
aI8051
aI1447
aI59517
aI3711
aI160345
asS'd'
I18611
sa(dp37222
S'a'
I87361
sS's'
I18
sS'm'
(lp37223
I1
aI3
aI1
aI13
aI1
aI23
aI49
aI75
aI117
aI295
aI1737
aI2091
aI6229
aI3157
aI32737
aI13751
aI101667
aI96261
asS'd'
I18612
sa(dp37224
S'a'
I87376
sS's'
I18
sS'm'
(lp37225
I1
aI3
aI3
aI5
aI27
aI19
aI103
aI201
aI65
aI757
aI1847
aI239
aI2185
aI15139
aI8883
aI17737
aI9207
aI147663
asS'd'
I18613
sa(dp37226
S'a'
I87392
sS's'
I18
sS'm'
(lp37227
I1
aI3
aI1
aI9
aI3
aI39
aI51
aI1
aI419
aI929
aI1049
aI2891
aI2585
aI2759
aI27587
aI55711
aI15461
aI46851
asS'd'
I18614
sa(dp37228
S'a'
I87419
sS's'
I18
sS'm'
(lp37229
I1
aI1
aI5
aI3
aI23
aI23
aI23
aI101
aI249
aI997
aI1889
aI2293
aI5693
aI939
aI29619
aI2775
aI49293
aI168895
asS'd'
I18615
sa(dp37230
S'a'
I87432
sS's'
I18
sS'm'
(lp37231
I1
aI3
aI1
aI15
aI9
aI63
aI17
aI97
aI385
aI517
aI1737
aI713
aI157
aI2597
aI20889
aI35209
aI47525
aI14389
asS'd'
I18616
sa(dp37232
S'a'
I87435
sS's'
I18
sS'm'
(lp37233
I1
aI3
aI1
aI1
aI7
aI27
aI9
aI147
aI349
aI493
aI341
aI2699
aI7743
aI4283
aI24691
aI11881
aI78619
aI153899
asS'd'
I18617
sa(dp37234
S'a'
I87445
sS's'
I18
sS'm'
(lp37235
I1
aI3
aI1
aI9
aI15
aI25
aI103
aI177
aI485
aI355
aI1319
aI767
aI6675
aI3425
aI7187
aI53767
aI92023
aI151523
asS'd'
I18618
sa(dp37236
S'a'
I87466
sS's'
I18
sS'm'
(lp37237
I1
aI1
aI3
aI3
aI15
aI35
aI25
aI177
aI295
aI5
aI661
aI3651
aI2597
aI16229
aI1343
aI54941
aI72047
aI169155
asS'd'
I18619
sa(dp37238
S'a'
I87491
sS's'
I18
sS'm'
(lp37239
I1
aI1
aI1
aI1
aI19
aI29
aI63
aI229
aI79
aI551
aI1401
aI2851
aI6935
aI12485
aI9243
aI21671
aI54209
aI105347
asS'd'
I18620
sa(dp37240
S'a'
I87503
sS's'
I18
sS'm'
(lp37241
I1
aI1
aI3
aI1
aI13
aI33
aI53
aI125
aI261
aI623
aI65
aI3863
aI1899
aI2453
aI16483
aI48655
aI50771
aI248555
asS'd'
I18621
sa(dp37242
S'a'
I87558
sS's'
I18
sS'm'
(lp37243
I1
aI3
aI5
aI7
aI25
aI15
aI107
aI149
aI485
aI247
aI1977
aI3125
aI4663
aI4925
aI15749
aI39429
aI52315
aI30545
asS'd'
I18622
sa(dp37244
S'a'
I87570
sS's'
I18
sS'm'
(lp37245
I1
aI1
aI7
aI13
aI23
aI13
aI127
aI111
aI9
aI17
aI1887
aI1341
aI3017
aI14333
aI6003
aI35113
aI14935
aI17593
asS'd'
I18623
sa(dp37246
S'a'
I87581
sS's'
I18
sS'm'
(lp37247
I1
aI1
aI1
aI15
aI23
aI17
aI111
aI43
aI71
aI549
aI1369
aI1711
aI3903
aI13605
aI14573
aI18973
aI28157
aI128421
asS'd'
I18624
sa(dp37248
S'a'
I87603
sS's'
I18
sS'm'
(lp37249
I1
aI1
aI7
aI15
aI27
aI23
aI5
aI99
aI87
aI865
aI1979
aI3287
aI3977
aI14989
aI17439
aI14593
aI92711
aI211259
asS'd'
I18625
sa(dp37250
S'a'
I87615
sS's'
I18
sS'm'
(lp37251
I1
aI3
aI3
aI15
aI5
aI19
aI3
aI231
aI127
aI863
aI1537
aI369
aI7915
aI10281
aI13925
aI12931
aI3905
aI178609
asS'd'
I18626
sa(dp37252
S'a'
I87623
sS's'
I18
sS'm'
(lp37253
I1
aI1
aI3
aI15
aI13
aI5
aI17
aI217
aI383
aI251
aI1701
aI3379
aI8157
aI4991
aI8563
aI24611
aI66081
aI205775
asS'd'
I18627
sa(dp37254
S'a'
I87635
sS's'
I18
sS'm'
(lp37255
I1
aI3
aI7
aI11
aI9
aI25
aI73
aI131
aI217
aI601
aI843
aI807
aI4509
aI16209
aI29581
aI50869
aI56595
aI14283
asS'd'
I18628
sa(dp37256
S'a'
I87647
sS's'
I18
sS'm'
(lp37257
I1
aI3
aI5
aI9
aI5
aI59
aI57
aI153
aI359
aI775
aI859
aI1897
aI6415
aI7389
aI10851
aI64247
aI21627
aI145017
asS'd'
I18629
sa(dp37258
S'a'
I87651
sS's'
I18
sS'm'
(lp37259
I1
aI3
aI7
aI1
aI27
aI11
aI117
aI179
aI175
aI343
aI687
aI2775
aI3655
aI11655
aI2641
aI355
aI83447
aI237799
asS'd'
I18630
sa(dp37260
S'a'
I87657
sS's'
I18
sS'm'
(lp37261
I1
aI1
aI7
aI11
aI17
aI59
aI91
aI21
aI403
aI797
aI1839
aI525
aI3279
aI6575
aI30083
aI12503
aI83057
aI109465
asS'd'
I18631
sa(dp37262
S'a'
I87665
sS's'
I18
sS'm'
(lp37263
I1
aI3
aI7
aI1
aI25
aI27
aI41
aI223
aI169
aI679
aI699
aI3287
aI6305
aI6459
aI23145
aI45519
aI127487
aI183563
asS'd'
I18632
sa(dp37264
S'a'
I87694
sS's'
I18
sS'm'
(lp37265
I1
aI3
aI3
aI1
aI9
aI45
aI105
aI57
aI185
aI97
aI899
aI3113
aI7081
aI7057
aI14559
aI53537
aI105623
aI155399
asS'd'
I18633
sa(dp37266
S'a'
I87701
sS's'
I18
sS'm'
(lp37267
I1
aI3
aI7
aI5
aI17
aI35
aI89
aI13
aI87
aI587
aI451
aI4079
aI1005
aI4311
aI15861
aI49977
aI59653
aI12107
asS'd'
I18634
sa(dp37268
S'a'
I87705
sS's'
I18
sS'm'
(lp37269
I1
aI3
aI7
aI3
aI19
aI55
aI43
aI77
aI317
aI369
aI71
aI937
aI1905
aI5005
aI17715
aI4005
aI55445
aI25159
asS'd'
I18635
sa(dp37270
S'a'
I87722
sS's'
I18
sS'm'
(lp37271
I1
aI1
aI7
aI15
aI15
aI51
aI87
aI37
aI59
aI755
aI763
aI455
aI711
aI13399
aI30999
aI61269
aI66037
aI202793
asS'd'
I18636
sa(dp37272
S'a'
I87739
sS's'
I18
sS'm'
(lp37273
I1
aI1
aI3
aI5
aI11
aI57
aI111
aI135
aI325
aI553
aI273
aI1533
aI3431
aI6427
aI24771
aI42143
aI56711
aI220873
asS'd'
I18637
sa(dp37274
S'a'
I87750
sS's'
I18
sS'm'
(lp37275
I1
aI1
aI1
aI13
aI3
aI49
aI53
aI81
aI491
aI177
aI1543
aI1847
aI7907
aI7817
aI15417
aI9897
aI101597
aI160195
asS'd'
I18638
sa(dp37276
S'a'
I87790
sS's'
I18
sS'm'
(lp37277
I1
aI3
aI1
aI15
aI19
aI15
aI91
aI123
aI365
aI113
aI129
aI3371
aI5789
aI13553
aI6887
aI62317
aI84269
aI44777
asS'd'
I18639
sa(dp37278
S'a'
I87795
sS's'
I18
sS'm'
(lp37279
I1
aI3
aI7
aI13
aI15
aI63
aI43
aI175
aI449
aI437
aI597
aI1371
aI5101
aI13797
aI28025
aI15809
aI7645
aI21169
asS'd'
I18640
sa(dp37280
S'a'
I87797
sS's'
I18
sS'm'
(lp37281
I1
aI1
aI7
aI13
aI23
aI63
aI105
aI69
aI219
aI153
aI1539
aI1537
aI6899
aI9363
aI27459
aI34551
aI62563
aI236679
asS'd'
I18641
sa(dp37282
S'a'
I87816
sS's'
I18
sS'm'
(lp37283
I1
aI3
aI7
aI3
aI9
aI33
aI3
aI101
aI135
aI571
aI127
aI3881
aI7017
aI13403
aI13817
aI55167
aI8645
aI471
asS'd'
I18642
sa(dp37284
S'a'
I87884
sS's'
I18
sS'm'
(lp37285
I1
aI3
aI3
aI13
aI7
aI53
aI29
aI89
aI473
aI135
aI639
aI3137
aI93
aI965
aI4867
aI58265
aI114963
aI175295
asS'd'
I18643
sa(dp37286
S'a'
I87917
sS's'
I18
sS'm'
(lp37287
I1
aI3
aI1
aI1
aI9
aI11
aI45
aI123
aI129
aI441
aI1601
aI39
aI4657
aI3701
aI29581
aI16045
aI57173
aI75195
asS'd'
I18644
sa(dp37288
S'a'
I87930
sS's'
I18
sS'm'
(lp37289
I1
aI1
aI3
aI11
aI7
aI21
aI9
aI73
aI25
aI891
aI1625
aI3019
aI223
aI14351
aI30621
aI3075
aI79051
aI178127
asS'd'
I18645
sa(dp37290
S'a'
I87942
sS's'
I18
sS'm'
(lp37291
I1
aI1
aI7
aI3
aI7
aI43
aI69
aI209
aI9
aI857
aI1539
aI2629
aI5277
aI14583
aI16443
aI28275
aI54143
aI206479
asS'd'
I18646
sa(dp37292
S'a'
I87953
sS's'
I18
sS'm'
(lp37293
I1
aI3
aI5
aI9
aI27
aI15
aI109
aI237
aI323
aI103
aI837
aI597
aI3609
aI6249
aI795
aI37191
aI20997
aI142079
asS'd'
I18647
sa(dp37294
S'a'
I87954
sS's'
I18
sS'm'
(lp37295
I1
aI1
aI3
aI1
aI27
aI33
aI123
aI191
aI55
aI531
aI1707
aI2633
aI6717
aI9645
aI21377
aI51593
aI9017
aI178185
asS'd'
I18648
sa(dp37296
S'a'
I87984
sS's'
I18
sS'm'
(lp37297
I1
aI1
aI5
aI15
aI3
aI3
aI67
aI225
aI229
aI161
aI2039
aI1499
aI873
aI8803
aI29901
aI58809
aI35625
aI207797
asS'd'
I18649
sa(dp37298
S'a'
I87989
sS's'
I18
sS'm'
(lp37299
I1
aI3
aI3
aI7
aI3
aI47
aI99
aI237
aI221
aI31
aI1043
aI1081
aI837
aI1617
aI17323
aI43879
aI55615
aI238537
asS'd'
I18650
sa(dp37300
S'a'
I88008
sS's'
I18
sS'm'
(lp37301
I1
aI3
aI7
aI1
aI17
aI37
aI45
aI211
aI245
aI657
aI221
aI1067
aI1683
aI16127
aI585
aI9067
aI25935
aI162469
asS'd'
I18651
sa(dp37302
S'a'
I88016
sS's'
I18
sS'm'
(lp37303
I1
aI1
aI5
aI9
aI23
aI27
aI5
aI237
aI137
aI227
aI129
aI279
aI4171
aI5963
aI349
aI12387
aI40701
aI177255
asS'd'
I18652
sa(dp37304
S'a'
I88038
sS's'
I18
sS'm'
(lp37305
I1
aI3
aI7
aI15
aI17
aI11
aI95
aI69
aI49
aI901
aI509
aI2541
aI3001
aI15501
aI24235
aI39863
aI95381
aI260793
asS'd'
I18653
sa(dp37306
S'a'
I88042
sS's'
I18
sS'm'
(lp37307
I1
aI3
aI3
aI3
aI29
aI3
aI125
aI3
aI423
aI609
aI1401
aI2337
aI1093
aI11419
aI29735
aI9033
aI115977
aI210201
asS'd'
I18654
sa(dp37308
S'a'
I88044
sS's'
I18
sS'm'
(lp37309
I1
aI1
aI1
aI9
aI11
aI21
aI71
aI33
aI399
aI1005
aI1691
aI1501
aI2585
aI7361
aI21527
aI7535
aI87091
aI192319
asS'd'
I18655
sa(dp37310
S'a'
I88049
sS's'
I18
sS'm'
(lp37311
I1
aI1
aI1
aI11
aI5
aI49
aI65
aI115
aI239
aI255
aI381
aI2803
aI3447
aI5775
aI18243
aI16545
aI108901
aI81355
asS'd'
I18656
sa(dp37312
S'a'
I88083
sS's'
I18
sS'm'
(lp37313
I1
aI3
aI7
aI15
aI31
aI5
aI81
aI213
aI281
aI903
aI189
aI1807
aI3551
aI4423
aI3591
aI27449
aI71659
aI255357
asS'd'
I18657
sa(dp37314
S'a'
I88119
sS's'
I18
sS'm'
(lp37315
I1
aI1
aI1
aI15
aI11
aI1
aI25
aI213
aI259
aI215
aI435
aI3531
aI4889
aI9509
aI21391
aI21589
aI89871
aI85895
asS'd'
I18658
sa(dp37316
S'a'
I88134
sS's'
I18
sS'm'
(lp37317
I1
aI3
aI3
aI9
aI29
aI29
aI37
aI127
aI419
aI631
aI1793
aI1547
aI1463
aI13265
aI17233
aI24627
aI3687
aI174179
asS'd'
I18659
sa(dp37318
S'a'
I88138
sS's'
I18
sS'm'
(lp37319
I1
aI1
aI1
aI7
aI27
aI61
aI105
aI135
aI439
aI161
aI721
aI2779
aI6731
aI14575
aI4565
aI25869
aI38981
aI191683
asS'd'
I18660
sa(dp37320
S'a'
I88143
sS's'
I18
sS'm'
(lp37321
I1
aI1
aI7
aI13
aI21
aI41
aI43
aI5
aI313
aI407
aI505
aI231
aI5023
aI8971
aI15825
aI38461
aI38797
aI136027
asS'd'
I18661
sa(dp37322
S'a'
I88191
sS's'
I18
sS'm'
(lp37323
I1
aI3
aI7
aI3
aI7
aI59
aI11
aI255
aI327
aI843
aI1179
aI889
aI4505
aI10891
aI7901
aI14485
aI72297
aI255985
asS'd'
I18662
sa(dp37324
S'a'
I88216
sS's'
I18
sS'm'
(lp37325
I1
aI1
aI5
aI3
aI9
aI37
aI89
aI59
aI413
aI51
aI515
aI4009
aI6501
aI8443
aI14381
aI60917
aI43567
aI234431
asS'd'
I18663
sa(dp37326
S'a'
I88231
sS's'
I18
sS'm'
(lp37327
I1
aI1
aI3
aI9
aI9
aI57
aI65
aI205
aI367
aI935
aI1975
aI2561
aI225
aI12529
aI4721
aI56659
aI87901
aI219641
asS'd'
I18664
sa(dp37328
S'a'
I88232
sS's'
I18
sS'm'
(lp37329
I1
aI3
aI7
aI15
aI23
aI55
aI61
aI15
aI89
aI267
aI1245
aI2703
aI7471
aI10499
aI19
aI19357
aI72413
aI199289
asS'd'
I18665
sa(dp37330
S'a'
I88297
sS's'
I18
sS'm'
(lp37331
I1
aI1
aI1
aI13
aI1
aI29
aI65
aI11
aI353
aI509
aI1831
aI2181
aI5265
aI14761
aI913
aI17109
aI113613
aI37143
asS'd'
I18666
sa(dp37332
S'a'
I88300
sS's'
I18
sS'm'
(lp37333
I1
aI1
aI7
aI5
aI17
aI37
aI97
aI249
aI169
aI223
aI475
aI2091
aI3101
aI8541
aI325
aI42359
aI16121
aI151739
asS'd'
I18667
sa(dp37334
S'a'
I88308
sS's'
I18
sS'm'
(lp37335
I1
aI1
aI3
aI13
aI1
aI45
aI13
aI209
aI395
aI215
aI15
aI2287
aI5365
aI9887
aI29799
aI51957
aI97483
aI109467
asS'd'
I18668
sa(dp37336
S'a'
I88317
sS's'
I18
sS'm'
(lp37337
I1
aI3
aI7
aI1
aI9
aI35
aI91
aI51
aI387
aI833
aI783
aI2483
aI3743
aI6155
aI5881
aI3047
aI86191
aI151867
asS'd'
I18669
sa(dp37338
S'a'
I88344
sS's'
I18
sS'm'
(lp37339
I1
aI1
aI1
aI7
aI15
aI25
aI13
aI3
aI119
aI333
aI761
aI3459
aI2555
aI15737
aI12945
aI15225
aI45487
aI78235
asS'd'
I18670
sa(dp37340
S'a'
I88350
sS's'
I18
sS'm'
(lp37341
I1
aI3
aI1
aI3
aI13
aI55
aI111
aI45
aI121
aI593
aI633
aI2705
aI1653
aI13275
aI13533
aI3559
aI100573
aI124363
asS'd'
I18671
sa(dp37342
S'a'
I88371
sS's'
I18
sS'm'
(lp37343
I1
aI1
aI3
aI1
aI29
aI49
aI65
aI69
aI133
aI667
aI653
aI2559
aI6335
aI8019
aI9251
aI5415
aI90125
aI197413
asS'd'
I18672
sa(dp37344
S'a'
I88469
sS's'
I18
sS'm'
(lp37345
I1
aI1
aI1
aI7
aI23
aI53
aI99
aI149
aI39
aI453
aI129
aI185
aI1143
aI12799
aI23339
aI41293
aI94023
aI105581
asS'd'
I18673
sa(dp37346
S'a'
I88483
sS's'
I18
sS'm'
(lp37347
I1
aI3
aI3
aI13
aI5
aI3
aI5
aI215
aI425
aI455
aI421
aI3815
aI5983
aI3851
aI19569
aI17363
aI6411
aI60037
asS'd'
I18674
sa(dp37348
S'a'
I88485
sS's'
I18
sS'm'
(lp37349
I1
aI3
aI7
aI1
aI29
aI7
aI63
aI207
aI299
aI17
aI1915
aI2041
aI8129
aI661
aI32481
aI55475
aI72027
aI239683
asS'd'
I18675
sa(dp37350
S'a'
I88518
sS's'
I18
sS'm'
(lp37351
I1
aI1
aI5
aI13
aI29
aI11
aI39
aI177
aI177
aI479
aI1291
aI3931
aI4353
aI327
aI7827
aI9529
aI6967
aI6469
asS'd'
I18676
sa(dp37352
S'a'
I88586
sS's'
I18
sS'm'
(lp37353
I1
aI1
aI5
aI3
aI25
aI39
aI121
aI15
aI7
aI715
aI583
aI3997
aI1373
aI7747
aI1777
aI7269
aI105333
aI201511
asS'd'
I18677
sa(dp37354
S'a'
I88605
sS's'
I18
sS'm'
(lp37355
I1
aI3
aI3
aI9
aI11
aI7
aI81
aI109
aI129
aI359
aI1281
aI1163
aI4895
aI10303
aI17801
aI43461
aI120271
aI173027
asS'd'
I18678
sa(dp37356
S'a'
I88621
sS's'
I18
sS'm'
(lp37357
I1
aI3
aI5
aI1
aI23
aI59
aI123
aI75
aI505
aI925
aI637
aI1713
aI995
aI14031
aI13711
aI62569
aI90553
aI242345
asS'd'
I18679
sa(dp37358
S'a'
I88651
sS's'
I18
sS'm'
(lp37359
I1
aI3
aI3
aI15
aI19
aI11
aI39
aI203
aI229
aI619
aI735
aI1367
aI4963
aI5263
aI30229
aI17847
aI9623
aI3277
asS'd'
I18680
sa(dp37360
S'a'
I88656
sS's'
I18
sS'm'
(lp37361
I1
aI1
aI7
aI13
aI5
aI27
aI23
aI223
aI377
aI335
aI1821
aI2481
aI4111
aI10373
aI18423
aI7237
aI75225
aI223433
asS'd'
I18681
sa(dp37362
S'a'
I88665
sS's'
I18
sS'm'
(lp37363
I1
aI3
aI7
aI1
aI21
aI19
aI71
aI107
aI19
aI703
aI945
aI3831
aI1099
aI6267
aI17489
aI27665
aI8861
aI127499
asS'd'
I18682
sa(dp37364
S'a'
I88668
sS's'
I18
sS'm'
(lp37365
I1
aI3
aI5
aI1
aI19
aI49
aI117
aI181
aI245
aI939
aI1279
aI3127
aI4427
aI3061
aI23399
aI64805
aI43077
aI100789
asS'd'
I18683
sa(dp37366
S'a'
I88712
sS's'
I18
sS'm'
(lp37367
I1
aI1
aI3
aI3
aI31
aI55
aI53
aI205
aI97
aI645
aI215
aI2617
aI7419
aI7159
aI27373
aI62341
aI58121
aI248677
asS'd'
I18684
sa(dp37368
S'a'
I88717
sS's'
I18
sS'm'
(lp37369
I1
aI1
aI7
aI5
aI15
aI41
aI99
aI75
aI201
aI187
aI197
aI3773
aI3097
aI6803
aI5307
aI31375
aI26743
aI142723
asS'd'
I18685
sa(dp37370
S'a'
I88751
sS's'
I18
sS'm'
(lp37371
I1
aI3
aI5
aI11
aI23
aI61
aI127
aI15
aI89
aI245
aI1345
aI1305
aI5937
aI15917
aI23867
aI50319
aI91921
aI248663
asS'd'
I18686
sa(dp37372
S'a'
I88754
sS's'
I18
sS'm'
(lp37373
I1
aI1
aI3
aI9
aI27
aI7
aI1
aI75
aI181
aI155
aI1947
aI577
aI2975
aI8855
aI5295
aI43403
aI112497
aI100679
asS'd'
I18687
sa(dp37374
S'a'
I88756
sS's'
I18
sS'm'
(lp37375
I1
aI1
aI1
aI11
aI29
aI61
aI35
aI241
aI207
aI73
aI1747
aI1797
aI3665
aI14275
aI25359
aI28685
aI79367
aI81819
asS'd'
I18688
sa(dp37376
S'a'
I88785
sS's'
I18
sS'm'
(lp37377
I1
aI3
aI5
aI9
aI11
aI1
aI37
aI79
aI431
aI157
aI1979
aI159
aI3087
aI1731
aI26141
aI31411
aI56457
aI94293
asS'd'
I18689
sa(dp37378
S'a'
I88801
sS's'
I18
sS'm'
(lp37379
I1
aI3
aI7
aI1
aI17
aI35
aI107
aI243
aI279
aI79
aI227
aI1275
aI761
aI11485
aI22181
aI16415
aI68801
aI4577
asS'd'
I18690
sa(dp37380
S'a'
I88807
sS's'
I18
sS'm'
(lp37381
I1
aI3
aI1
aI9
aI21
aI43
aI115
aI131
aI129
aI123
aI1677
aI1875
aI7355
aI15927
aI845
aI24101
aI48985
aI39703
asS'd'
I18691
sa(dp37382
S'a'
I88826
sS's'
I18
sS'm'
(lp37383
I1
aI1
aI7
aI3
aI17
aI25
aI105
aI189
aI317
aI109
aI1629
aI3103
aI615
aI1047
aI621
aI62743
aI43631
aI9811
asS'd'
I18692
sa(dp37384
S'a'
I88848
sS's'
I18
sS'm'
(lp37385
I1
aI3
aI1
aI3
aI3
aI45
aI49
aI73
aI383
aI761
aI685
aI3211
aI3855
aI16307
aI30469
aI1393
aI52535
aI165503
asS'd'
I18693
sa(dp37386
S'a'
I88854
sS's'
I18
sS'm'
(lp37387
I1
aI3
aI5
aI15
aI1
aI41
aI89
aI105
aI213
aI33
aI1477
aI711
aI4823
aI503
aI12533
aI56781
aI42825
aI223399
asS'd'
I18694
sa(dp37388
S'a'
I88864
sS's'
I18
sS'm'
(lp37389
I1
aI3
aI5
aI9
aI17
aI45
aI63
aI113
aI359
aI927
aI1467
aI2811
aI4275
aI5193
aI6023
aI32689
aI87747
aI234697
asS'd'
I18695
sa(dp37390
S'a'
I88879
sS's'
I18
sS'm'
(lp37391
I1
aI1
aI3
aI1
aI21
aI49
aI73
aI157
aI97
aI915
aI1689
aI3289
aI7515
aI10759
aI32253
aI63175
aI66175
aI125813
asS'd'
I18696
sa(dp37392
S'a'
I88939
sS's'
I18
sS'm'
(lp37393
I1
aI1
aI1
aI5
aI19
aI1
aI127
aI229
aI453
aI617
aI511
aI1515
aI3815
aI3125
aI26851
aI31635
aI35389
aI237483
asS'd'
I18697
sa(dp37394
S'a'
I88947
sS's'
I18
sS'm'
(lp37395
I1
aI1
aI1
aI13
aI27
aI61
aI75
aI23
aI289
aI133
aI975
aI3217
aI3777
aI12095
aI15235
aI33845
aI125503
aI88417
asS'd'
I18698
sa(dp37396
S'a'
I88987
sS's'
I18
sS'm'
(lp37397
I1
aI1
aI3
aI13
aI27
aI49
aI29
aI115
aI221
aI995
aI1305
aI2717
aI2243
aI13391
aI20841
aI863
aI63195
aI46829
asS'd'
I18699
sa(dp37398
S'a'
I88990
sS's'
I18
sS'm'
(lp37399
I1
aI3
aI1
aI11
aI31
aI49
aI43
aI163
aI503
aI123
aI657
aI1285
aI3695
aI8401
aI17087
aI48289
aI3947
aI41495
asS'd'
I18700
sa(dp37400
S'a'
I88996
sS's'
I18
sS'm'
(lp37401
I1
aI3
aI7
aI9
aI21
aI17
aI125
aI75
aI395
aI979
aI781
aI2501
aI6511
aI4619
aI28943
aI18295
aI87547
aI196289
asS'd'
I18701
sa(dp37402
S'a'
I89008
sS's'
I18
sS'm'
(lp37403
I1
aI3
aI3
aI5
aI3
aI39
aI107
aI199
aI7
aI331
aI77
aI511
aI5787
aI3155
aI29605
aI44633
aI51041
aI89141
asS'd'
I18702
sa(dp37404
S'a'
I89011
sS's'
I18
sS'm'
(lp37405
I1
aI1
aI3
aI7
aI11
aI37
aI79
aI69
aI181
aI623
aI299
aI2321
aI4371
aI7449
aI3137
aI25753
aI116673
aI30441
asS'd'
I18703
sa(dp37406
S'a'
I89043
sS's'
I18
sS'm'
(lp37407
I1
aI1
aI5
aI9
aI11
aI13
aI49
aI207
aI179
aI671
aI1469
aI1005
aI6887
aI12203
aI9365
aI62455
aI36283
aI42053
asS'd'
I18704
sa(dp37408
S'a'
I89045
sS's'
I18
sS'm'
(lp37409
I1
aI3
aI1
aI1
aI7
aI31
aI115
aI79
aI435
aI101
aI1525
aI3695
aI3229
aI11401
aI23959
aI62055
aI37725
aI219753
asS'd'
I18705
sa(dp37410
S'a'
I89050
sS's'
I18
sS'm'
(lp37411
I1
aI3
aI3
aI7
aI25
aI49
aI123
aI161
aI275
aI291
aI255
aI2247
aI4271
aI10771
aI2449
aI26343
aI61169
aI30691
asS'd'
I18706
sa(dp37412
S'a'
I89097
sS's'
I18
sS'm'
(lp37413
I1
aI1
aI7
aI1
aI15
aI3
aI11
aI15
aI125
aI1021
aI1817
aI417
aI2721
aI13985
aI19039
aI451
aI32559
aI199407
asS'd'
I18707
sa(dp37414
S'a'
I89100
sS's'
I18
sS'm'
(lp37415
I1
aI1
aI5
aI15
aI11
aI25
aI23
aI101
aI427
aI431
aI1353
aI1957
aI3529
aI513
aI11937
aI14469
aI89539
aI242015
asS'd'
I18708
sa(dp37416
S'a'
I89106
sS's'
I18
sS'm'
(lp37417
I1
aI1
aI5
aI13
aI27
aI5
aI107
aI29
aI469
aI3
aI1427
aI1949
aI7007
aI16339
aI3375
aI63545
aI100739
aI229487
asS'd'
I18709
sa(dp37418
S'a'
I89134
sS's'
I18
sS'm'
(lp37419
I1
aI3
aI5
aI13
aI15
aI17
aI59
aI213
aI417
aI557
aI11
aI811
aI5041
aI4133
aI25735
aI46807
aI65669
aI148081
asS'd'
I18710
sa(dp37420
S'a'
I89178
sS's'
I18
sS'm'
(lp37421
I1
aI1
aI5
aI13
aI9
aI47
aI35
aI173
aI277
aI805
aI1249
aI3707
aI1079
aI2833
aI29383
aI58995
aI21005
aI181567
asS'd'
I18711
sa(dp37422
S'a'
I89189
sS's'
I18
sS'm'
(lp37423
I1
aI3
aI5
aI1
aI5
aI25
aI125
aI45
aI157
aI291
aI1329
aI3317
aI2311
aI9919
aI31001
aI65127
aI19451
aI117621
asS'd'
I18712
sa(dp37424
S'a'
I89202
sS's'
I18
sS'm'
(lp37425
I1
aI1
aI5
aI9
aI31
aI21
aI9
aI193
aI23
aI879
aI699
aI1135
aI7151
aI8635
aI20711
aI45207
aI67047
aI4397
asS'd'
I18713
sa(dp37426
S'a'
I89211
sS's'
I18
sS'm'
(lp37427
I1
aI3
aI1
aI11
aI11
aI49
aI119
aI129
aI409
aI491
aI463
aI833
aI3661
aI607
aI25961
aI6061
aI12747
aI160337
asS'd'
I18714
sa(dp37428
S'a'
I89224
sS's'
I18
sS'm'
(lp37429
I1
aI3
aI1
aI15
aI31
aI35
aI93
aI95
aI239
aI695
aI1113
aI2371
aI2625
aI10371
aI10781
aI46209
aI67051
aI109923
asS'd'
I18715
sa(dp37430
S'a'
I89241
sS's'
I18
sS'm'
(lp37431
I1
aI3
aI5
aI9
aI5
aI61
aI25
aI19
aI99
aI159
aI55
aI43
aI3679
aI1023
aI17951
aI44841
aI101653
aI195955
asS'd'
I18716
sa(dp37432
S'a'
I89251
sS's'
I18
sS'm'
(lp37433
I1
aI1
aI3
aI11
aI19
aI45
aI99
aI47
aI407
aI115
aI353
aI3537
aI7147
aI5837
aI27309
aI44539
aI30227
aI93183
asS'd'
I18717
sa(dp37434
S'a'
I89254
sS's'
I18
sS'm'
(lp37435
I1
aI1
aI7
aI15
aI13
aI59
aI83
aI215
aI79
aI865
aI269
aI2999
aI2415
aI10631
aI23655
aI51583
aI46105
aI43965
asS'd'
I18718
sa(dp37436
S'a'
I89258
sS's'
I18
sS'm'
(lp37437
I1
aI3
aI7
aI5
aI17
aI1
aI7
aI119
aI501
aI25
aI1097
aI3639
aI7017
aI381
aI4793
aI37263
aI60431
aI77323
asS'd'
I18719
sa(dp37438
S'a'
I89295
sS's'
I18
sS'm'
(lp37439
I1
aI3
aI7
aI13
aI1
aI37
aI99
aI103
aI459
aI853
aI5
aI3093
aI6167
aI14497
aI7003
aI36979
aI71919
aI64823
asS'd'
I18720
sa(dp37440
S'a'
I89328
sS's'
I18
sS'm'
(lp37441
I1
aI3
aI7
aI5
aI23
aI7
aI37
aI255
aI297
aI115
aI113
aI579
aI4561
aI5245
aI11173
aI28645
aI23989
aI240777
asS'd'
I18721
sa(dp37442
S'a'
I89357
sS's'
I18
sS'm'
(lp37443
I1
aI3
aI5
aI9
aI29
aI47
aI43
aI145
aI481
aI251
aI737
aI2531
aI2425
aI529
aI3953
aI13229
aI35933
aI187855
asS'd'
I18722
sa(dp37444
S'a'
I89403
sS's'
I18
sS'm'
(lp37445
I1
aI3
aI1
aI15
aI15
aI15
aI29
aI73
aI405
aI91
aI1399
aI3599
aI1517
aI11075
aI11265
aI22817
aI26619
aI1183
asS'd'
I18723
sa(dp37446
S'a'
I89420
sS's'
I18
sS'm'
(lp37447
I1
aI1
aI5
aI15
aI21
aI47
aI93
aI33
aI47
aI527
aI877
aI3453
aI3867
aI4007
aI32503
aI11789
aI68333
aI187419
asS'd'
I18724
sa(dp37448
S'a'
I89426
sS's'
I18
sS'm'
(lp37449
I1
aI1
aI3
aI13
aI5
aI3
aI99
aI21
aI17
aI779
aI541
aI3919
aI1339
aI13507
aI28965
aI61145
aI50421
aI192319
asS'd'
I18725
sa(dp37450
S'a'
I89465
sS's'
I18
sS'm'
(lp37451
I1
aI1
aI3
aI7
aI31
aI63
aI17
aI9
aI331
aI681
aI515
aI1067
aI351
aI2471
aI31271
aI36015
aI72911
aI32383
asS'd'
I18726
sa(dp37452
S'a'
I89471
sS's'
I18
sS'm'
(lp37453
I1
aI1
aI3
aI7
aI1
aI55
aI97
aI211
aI409
aI499
aI1207
aI2405
aI2291
aI1373
aI1263
aI65303
aI38655
aI159965
asS'd'
I18727
sa(dp37454
S'a'
I89475
sS's'
I18
sS'm'
(lp37455
I1
aI1
aI1
aI5
aI23
aI11
aI21
aI23
aI35
aI19
aI1699
aI2325
aI6117
aI14971
aI32327
aI31369
aI28061
aI112819
asS'd'
I18728
sa(dp37456
S'a'
I89547
sS's'
I18
sS'm'
(lp37457
I1
aI3
aI3
aI13
aI29
aI63
aI19
aI201
aI173
aI395
aI1437
aI369
aI7045
aI14347
aI5393
aI11311
aI28415
aI161019
asS'd'
I18729
sa(dp37458
S'a'
I89549
sS's'
I18
sS'm'
(lp37459
I1
aI3
aI1
aI5
aI31
aI15
aI101
aI13
aI97
aI865
aI1063
aI2129
aI811
aI3337
aI7585
aI54803
aI122099
aI149531
asS'd'
I18730
sa(dp37460
S'a'
I89558
sS's'
I18
sS'm'
(lp37461
I1
aI3
aI3
aI13
aI19
aI5
aI47
aI13
aI497
aI683
aI1197
aI3509
aI4375
aI3353
aI31847
aI283
aI95281
aI7975
asS'd'
I18731
sa(dp37462
S'a'
I89564
sS's'
I18
sS'm'
(lp37463
I1
aI1
aI7
aI5
aI25
aI19
aI59
aI105
aI167
aI775
aI581
aI2679
aI3003
aI9345
aI20209
aI31487
aI25357
aI226341
asS'd'
I18732
sa(dp37464
S'a'
I89611
sS's'
I18
sS'm'
(lp37465
I1
aI1
aI5
aI5
aI31
aI61
aI13
aI77
aI189
aI141
aI1157
aI609
aI7245
aI15303
aI32743
aI50229
aI67391
aI173977
asS'd'
I18733
sa(dp37466
S'a'
I89616
sS's'
I18
sS'm'
(lp37467
I1
aI3
aI7
aI13
aI17
aI11
aI49
aI135
aI475
aI303
aI1373
aI1437
aI6119
aI1729
aI21347
aI31643
aI86523
aI41223
asS'd'
I18734
sa(dp37468
S'a'
I89626
sS's'
I18
sS'm'
(lp37469
I1
aI3
aI7
aI13
aI3
aI19
aI103
aI53
aI209
aI281
aI77
aI2009
aI7911
aI8549
aI17655
aI33165
aI9685
aI2289
asS'd'
I18735
sa(dp37470
S'a'
I89650
sS's'
I18
sS'm'
(lp37471
I1
aI3
aI3
aI9
aI5
aI31
aI71
aI113
aI93
aI255
aI165
aI3465
aI6769
aI13047
aI20701
aI33669
aI22537
aI175209
asS'd'
I18736
sa(dp37472
S'a'
I89691
sS's'
I18
sS'm'
(lp37473
I1
aI3
aI1
aI1
aI15
aI59
aI89
aI227
aI275
aI451
aI869
aI1865
aI1327
aI3895
aI5459
aI47997
aI34287
aI95343
asS'd'
I18737
sa(dp37474
S'a'
I89718
sS's'
I18
sS'm'
(lp37475
I1
aI3
aI7
aI13
aI27
aI1
aI25
aI99
aI475
aI421
aI693
aI1955
aI4017
aI16037
aI29915
aI52415
aI99913
aI59151
asS'd'
I18738
sa(dp37476
S'a'
I89764
sS's'
I18
sS'm'
(lp37477
I1
aI3
aI5
aI1
aI9
aI61
aI107
aI149
aI465
aI1003
aI891
aI2387
aI407
aI5851
aI6287
aI56401
aI109693
aI72035
asS'd'
I18739
sa(dp37478
S'a'
I89776
sS's'
I18
sS'm'
(lp37479
I1
aI1
aI1
aI11
aI25
aI53
aI101
aI125
aI35
aI949
aI1019
aI3087
aI2785
aI11271
aI25623
aI57313
aI115683
aI101923
asS'd'
I18740
sa(dp37480
S'a'
I89785
sS's'
I18
sS'm'
(lp37481
I1
aI3
aI5
aI13
aI19
aI13
aI97
aI147
aI149
aI483
aI1727
aI1771
aI2089
aI8661
aI28223
aI30437
aI42565
aI13261
asS'd'
I18741
sa(dp37482
S'a'
I89813
sS's'
I18
sS'm'
(lp37483
I1
aI3
aI3
aI11
aI17
aI9
aI13
aI13
aI419
aI531
aI1617
aI1459
aI411
aI9953
aI25581
aI30305
aI120721
aI81113
asS'd'
I18742
sa(dp37484
S'a'
I89817
sS's'
I18
sS'm'
(lp37485
I1
aI3
aI3
aI9
aI9
aI9
aI83
aI35
aI367
aI981
aI911
aI1915
aI4937
aI16187
aI20441
aI30433
aI107605
aI119939
asS'd'
I18743
sa(dp37486
S'a'
I89861
sS's'
I18
sS'm'
(lp37487
I1
aI3
aI7
aI11
aI11
aI47
aI31
aI7
aI141
aI905
aI1753
aI3069
aI47
aI7347
aI10517
aI19515
aI126827
aI68669
asS'd'
I18744
sa(dp37488
S'a'
I89876
sS's'
I18
sS'm'
(lp37489
I1
aI3
aI3
aI9
aI17
aI33
aI121
aI159
aI265
aI389
aI261
aI2479
aI7705
aI6453
aI31963
aI14123
aI100201
aI77235
asS'd'
I18745
sa(dp37490
S'a'
I89889
sS's'
I18
sS'm'
(lp37491
I1
aI1
aI5
aI5
aI13
aI63
aI3
aI1
aI107
aI383
aI633
aI2183
aI1437
aI14525
aI29315
aI3277
aI2153
aI204061
asS'd'
I18746
sa(dp37492
S'a'
I89904
sS's'
I18
sS'm'
(lp37493
I1
aI3
aI3
aI3
aI17
aI9
aI47
aI173
aI17
aI413
aI451
aI1127
aI1807
aI5265
aI32543
aI8215
aI123601
aI138777
asS'd'
I18747
sa(dp37494
S'a'
I89939
sS's'
I18
sS'm'
(lp37495
I1
aI3
aI1
aI1
aI7
aI57
aI93
aI29
aI101
aI955
aI1445
aI4017
aI2853
aI3551
aI22173
aI40355
aI34687
aI133063
asS'd'
I18748
sa(dp37496
S'a'
I89948
sS's'
I18
sS'm'
(lp37497
I1
aI1
aI5
aI11
aI19
aI15
aI95
aI177
aI49
aI971
aI15
aI2293
aI2627
aI7841
aI2103
aI64331
aI60481
aI182431
asS'd'
I18749
sa(dp37498
S'a'
I89952
sS's'
I18
sS'm'
(lp37499
I1
aI3
aI1
aI15
aI25
aI57
aI47
aI85
aI485
aI11
aI1669
aI995
aI6939
aI4125
aI19513
aI62397
aI62645
aI82213
asS'd'
I18750
sa(dp37500
S'a'
I89964
sS's'
I18
sS'm'
(lp37501
I1
aI1
aI1
aI11
aI1
aI37
aI101
aI157
aI17
aI261
aI997
aI817
aI2195
aI4141
aI10505
aI60685
aI98165
aI167391
asS'd'
I18751
sa(dp37502
S'a'
I89969
sS's'
I18
sS'm'
(lp37503
I1
aI1
aI1
aI1
aI31
aI9
aI103
aI97
aI161
aI13
aI1043
aI307
aI7321
aI671
aI12417
aI58661
aI23031
aI207833
asS'd'
I18752
sa(dp37504
S'a'
I89979
sS's'
I18
sS'm'
(lp37505
I1
aI1
aI7
aI9
aI15
aI49
aI69
aI117
aI93
aI95
aI507
aI393
aI6169
aI2111
aI27179
aI47217
aI93699
aI67315
asS'd'
I18753
sa(dp37506
S'a'
I90009
sS's'
I18
sS'm'
(lp37507
I1
aI3
aI7
aI1
aI23
aI41
aI115
aI125
aI343
aI615
aI397
aI1199
aI3041
aI11019
aI1071
aI51069
aI75757
aI245765
asS'd'
I18754
sa(dp37508
S'a'
I90022
sS's'
I18
sS'm'
(lp37509
I1
aI1
aI1
aI13
aI7
aI15
aI111
aI239
aI29
aI419
aI203
aI2395
aI3995
aI13
aI32341
aI17471
aI53259
aI3317
asS'd'
I18755
sa(dp37510
S'a'
I90031
sS's'
I18
sS'm'
(lp37511
I1
aI3
aI1
aI11
aI29
aI27
aI15
aI217
aI17
aI163
aI1847
aI3549
aI4911
aI4539
aI4927
aI57157
aI44893
aI41669
asS'd'
I18756
sa(dp37512
S'a'
I90043
sS's'
I18
sS'm'
(lp37513
I1
aI1
aI7
aI7
aI25
aI15
aI101
aI149
aI433
aI717
aI1827
aI3837
aI4565
aI14521
aI28857
aI27775
aI117429
aI136131
asS'd'
I18757
sa(dp37514
S'a'
I90045
sS's'
I18
sS'm'
(lp37515
I1
aI1
aI1
aI3
aI25
aI35
aI85
aI3
aI381
aI253
aI375
aI3967
aI3101
aI12727
aI31739
aI48885
aI35821
aI92229
asS'd'
I18758
sa(dp37516
S'a'
I90068
sS's'
I18
sS'm'
(lp37517
I1
aI3
aI3
aI7
aI29
aI5
aI51
aI157
aI67
aI467
aI1957
aI3453
aI1353
aI4839
aI25379
aI42731
aI109385
aI52479
asS'd'
I18759
sa(dp37518
S'a'
I90071
sS's'
I18
sS'm'
(lp37519
I1
aI3
aI7
aI13
aI3
aI55
aI61
aI73
aI257
aI313
aI89
aI2557
aI7467
aI2223
aI2951
aI49265
aI126605
aI72007
asS'd'
I18760
sa(dp37520
S'a'
I90082
sS's'
I18
sS'm'
(lp37521
I1
aI3
aI7
aI5
aI5
aI11
aI83
aI3
aI347
aI63
aI479
aI529
aI5059
aI7029
aI20523
aI58387
aI44891
aI168921
asS'd'
I18761
sa(dp37522
S'a'
I90087
sS's'
I18
sS'm'
(lp37523
I1
aI3
aI1
aI11
aI3
aI51
aI99
aI5
aI161
aI279
aI1509
aI3659
aI3107
aI13925
aI5117
aI46153
aI48731
aI69767
asS'd'
I18762
sa(dp37524
S'a'
I90117
sS's'
I18
sS'm'
(lp37525
I1
aI1
aI5
aI5
aI3
aI53
aI49
aI243
aI383
aI401
aI1205
aI975
aI3305
aI12769
aI25533
aI28733
aI115161
aI160885
asS'd'
I18763
sa(dp37526
S'a'
I90127
sS's'
I18
sS'm'
(lp37527
I1
aI1
aI1
aI7
aI15
aI5
aI43
aI143
aI493
aI527
aI1625
aI2115
aI3929
aI12425
aI16127
aI25045
aI55973
aI202359
asS'd'
I18764
sa(dp37528
S'a'
I90163
sS's'
I18
sS'm'
(lp37529
I1
aI3
aI7
aI7
aI11
aI9
aI69
aI79
aI417
aI941
aI473
aI1655
aI5763
aI9889
aI22443
aI12153
aI103489
aI74737
asS'd'
I18765
sa(dp37530
S'a'
I90166
sS's'
I18
sS'm'
(lp37531
I1
aI1
aI7
aI9
aI27
aI31
aI97
aI253
aI199
aI99
aI1955
aI1481
aI2509
aI11923
aI6337
aI15899
aI122515
aI244721
asS'd'
I18766
sa(dp37532
S'a'
I90177
sS's'
I18
sS'm'
(lp37533
I1
aI1
aI5
aI13
aI29
aI15
aI35
aI177
aI261
aI613
aI1279
aI2837
aI2945
aI4501
aI22865
aI36893
aI51979
aI245569
asS'd'
I18767
sa(dp37534
S'a'
I90211
sS's'
I18
sS'm'
(lp37535
I1
aI1
aI5
aI9
aI21
aI5
aI85
aI5
aI303
aI165
aI681
aI3965
aI2575
aI1493
aI10367
aI55845
aI92139
aI92539
asS'd'
I18768
sa(dp37536
S'a'
I90218
sS's'
I18
sS'm'
(lp37537
I1
aI3
aI5
aI1
aI23
aI49
aI49
aI161
aI481
aI181
aI1991
aI1845
aI4541
aI14187
aI10893
aI64931
aI79943
aI57907
asS'd'
I18769
sa(dp37538
S'a'
I90225
sS's'
I18
sS'm'
(lp37539
I1
aI3
aI1
aI5
aI11
aI27
aI19
aI193
aI371
aI463
aI1573
aI271
aI1127
aI15091
aI9967
aI40337
aI104163
aI159339
asS'd'
I18770
sa(dp37540
S'a'
I90231
sS's'
I18
sS'm'
(lp37541
I1
aI3
aI1
aI3
aI13
aI3
aI57
aI149
aI465
aI789
aI1155
aI2223
aI2007
aI13987
aI19057
aI40447
aI5217
aI86191
asS'd'
I18771
sa(dp37542
S'a'
I90235
sS's'
I18
sS'm'
(lp37543
I1
aI3
aI5
aI5
aI9
aI45
aI27
aI155
aI95
aI171
aI489
aI2539
aI843
aI16125
aI7047
aI58541
aI84641
aI212013
asS'd'
I18772
sa(dp37544
S'a'
I90289
sS's'
I18
sS'm'
(lp37545
I1
aI3
aI5
aI13
aI13
aI53
aI101
aI159
aI7
aI481
aI143
aI3869
aI6629
aI3527
aI1555
aI6019
aI155
aI230157
asS'd'
I18773
sa(dp37546
S'a'
I90344
sS's'
I18
sS'm'
(lp37547
I1
aI3
aI1
aI1
aI19
aI59
aI59
aI129
aI107
aI887
aI1595
aI93
aI6577
aI3947
aI14409
aI31081
aI68595
aI226741
asS'd'
I18774
sa(dp37548
S'a'
I90347
sS's'
I18
sS'm'
(lp37549
I1
aI3
aI3
aI11
aI13
aI49
aI81
aI253
aI363
aI875
aI489
aI2181
aI3487
aI1615
aI31157
aI32949
aI44809
aI119421
asS'd'
I18775
sa(dp37550
S'a'
I90379
sS's'
I18
sS'm'
(lp37551
I1
aI1
aI3
aI1
aI3
aI19
aI71
aI93
aI397
aI521
aI2015
aI3829
aI3013
aI3941
aI29437
aI1959
aI70283
aI254361
asS'd'
I18776
sa(dp37552
S'a'
I90387
sS's'
I18
sS'm'
(lp37553
I1
aI3
aI1
aI3
aI27
aI11
aI95
aI39
aI299
aI521
aI389
aI3451
aI3047
aI8637
aI22537
aI11279
aI67407
aI101511
asS'd'
I18777
sa(dp37554
S'a'
I90418
sS's'
I18
sS'm'
(lp37555
I1
aI3
aI1
aI1
aI27
aI43
aI123
aI237
aI315
aI503
aI1059
aI2185
aI3963
aI1593
aI19157
aI13909
aI58025
aI91649
asS'd'
I18778
sa(dp37556
S'a'
I90435
sS's'
I18
sS'm'
(lp37557
I1
aI1
aI7
aI1
aI5
aI19
aI79
aI109
aI459
aI541
aI521
aI89
aI389
aI13499
aI9769
aI1429
aI117357
aI153261
asS'd'
I18779
sa(dp37558
S'a'
I90444
sS's'
I18
sS'm'
(lp37559
I1
aI1
aI5
aI3
aI9
aI29
aI21
aI97
aI219
aI915
aI2013
aI1955
aI1015
aI549
aI9777
aI5005
aI110953
aI16915
asS'd'
I18780
sa(dp37560
S'a'
I90472
sS's'
I18
sS'm'
(lp37561
I1
aI1
aI1
aI9
aI29
aI11
aI103
aI167
aI465
aI515
aI843
aI151
aI769
aI12033
aI9451
aI14949
aI110075
aI113947
asS'd'
I18781
sa(dp37562
S'a'
I90478
sS's'
I18
sS'm'
(lp37563
I1
aI3
aI5
aI1
aI21
aI1
aI105
aI49
aI35
aI737
aI231
aI2761
aI1519
aI9997
aI601
aI20883
aI42575
aI98081
asS'd'
I18782
sa(dp37564
S'a'
I90485
sS's'
I18
sS'm'
(lp37565
I1
aI1
aI1
aI5
aI13
aI63
aI47
aI171
aI187
aI407
aI643
aI1423
aI6325
aI10079
aI23781
aI36353
aI20655
aI10231
asS'd'
I18783
sa(dp37566
S'a'
I90486
sS's'
I18
sS'm'
(lp37567
I1
aI1
aI7
aI13
aI13
aI13
aI91
aI31
aI19
aI305
aI505
aI1937
aI2683
aI10791
aI7719
aI54797
aI9405
aI195819
asS'd'
I18784
sa(dp37568
S'a'
I90513
sS's'
I18
sS'm'
(lp37569
I1
aI1
aI1
aI7
aI17
aI9
aI21
aI211
aI85
aI851
aI211
aI1533
aI4035
aI11
aI26873
aI16755
aI77809
aI44603
asS'd'
I18785
sa(dp37570
S'a'
I90514
sS's'
I18
sS'm'
(lp37571
I1
aI1
aI3
aI1
aI29
aI47
aI31
aI141
aI9
aI881
aI1229
aI1261
aI3747
aI4603
aI22177
aI48937
aI21435
aI157029
asS'd'
I18786
sa(dp37572
S'a'
I90532
sS's'
I18
sS'm'
(lp37573
I1
aI1
aI1
aI9
aI11
aI35
aI109
aI187
aI319
aI863
aI1339
aI2193
aI4147
aI3721
aI7243
aI18295
aI92461
aI88875
asS'd'
I18787
sa(dp37574
S'a'
I90567
sS's'
I18
sS'm'
(lp37575
I1
aI1
aI1
aI1
aI25
aI41
aI79
aI191
aI47
aI819
aI2013
aI3133
aI2763
aI9231
aI10343
aI49693
aI26753
aI97465
asS'd'
I18788
sa(dp37576
S'a'
I90595
sS's'
I18
sS'm'
(lp37577
I1
aI3
aI1
aI3
aI17
aI25
aI63
aI139
aI179
aI113
aI1681
aI1997
aI4561
aI14453
aI30721
aI7053
aI22937
aI183303
asS'd'
I18789
sa(dp37578
S'a'
I90597
sS's'
I18
sS'm'
(lp37579
I1
aI1
aI7
aI3
aI29
aI11
aI41
aI157
aI427
aI887
aI295
aI443
aI5593
aI8633
aI9757
aI37595
aI121655
aI135739
asS'd'
I18790
sa(dp37580
S'a'
I90621
sS's'
I18
sS'm'
(lp37581
I1
aI3
aI1
aI7
aI29
aI5
aI23
aI231
aI85
aI67
aI103
aI1395
aI7821
aI9551
aI17019
aI1825
aI69963
aI254583
asS'd'
I18791
sa(dp37582
S'a'
I90631
sS's'
I18
sS'm'
(lp37583
I1
aI1
aI7
aI9
aI7
aI23
aI7
aI205
aI17
aI111
aI1219
aI3101
aI7485
aI11579
aI11791
aI10203
aI119835
aI175985
asS'd'
I18792
sa(dp37584
S'a'
I90668
sS's'
I18
sS'm'
(lp37585
I1
aI3
aI3
aI13
aI3
aI57
aI101
aI255
aI331
aI911
aI491
aI3929
aI2519
aI2185
aI21107
aI24599
aI92831
aI75001
asS'd'
I18793
sa(dp37586
S'a'
I90706
sS's'
I18
sS'm'
(lp37587
I1
aI3
aI7
aI1
aI23
aI53
aI69
aI229
aI295
aI881
aI905
aI3727
aI3885
aI7967
aI2061
aI53595
aI16033
aI36443
asS'd'
I18794
sa(dp37588
S'a'
I90708
sS's'
I18
sS'm'
(lp37589
I1
aI1
aI7
aI15
aI21
aI1
aI127
aI115
aI191
aI53
aI929
aI1093
aI5447
aI1665
aI4409
aI22611
aI38157
aI139201
asS'd'
I18795
sa(dp37590
S'a'
I90715
sS's'
I18
sS'm'
(lp37591
I1
aI3
aI3
aI13
aI29
aI31
aI113
aI215
aI365
aI41
aI1977
aI2839
aI4147
aI8321
aI1361
aI45717
aI80505
aI176631
asS'd'
I18796
sa(dp37592
S'a'
I90724
sS's'
I18
sS'm'
(lp37593
I1
aI3
aI5
aI3
aI3
aI57
aI25
aI81
aI41
aI229
aI669
aI3371
aI7505
aI1197
aI14921
aI34365
aI67571
aI27355
asS'd'
I18797
sa(dp37594
S'a'
I90769
sS's'
I18
sS'm'
(lp37595
I1
aI3
aI7
aI9
aI17
aI35
aI7
aI169
aI13
aI163
aI2007
aI3697
aI5635
aI8003
aI26105
aI62917
aI19349
aI47029
asS'd'
I18798
sa(dp37596
S'a'
I90791
sS's'
I18
sS'm'
(lp37597
I1
aI1
aI5
aI5
aI29
aI27
aI89
aI137
aI69
aI189
aI871
aI3139
aI6383
aI14955
aI15349
aI60447
aI122291
aI26541
asS'd'
I18799
sa(dp37598
S'a'
I90792
sS's'
I18
sS'm'
(lp37599
I1
aI1
aI7
aI13
aI27
aI49
aI115
aI111
aI441
aI865
aI1397
aI161
aI755
aI14461
aI8601
aI45533
aI105309
aI149799
asS'd'
I18800
sa(dp37600
S'a'
I90805
sS's'
I18
sS'm'
(lp37601
I1
aI1
aI7
aI3
aI9
aI29
aI91
aI5
aI239
aI605
aI491
aI1705
aI4099
aI9111
aI19821
aI56903
aI62815
aI40615
asS'd'
I18801
sa(dp37602
S'a'
I90810
sS's'
I18
sS'm'
(lp37603
I1
aI1
aI7
aI15
aI5
aI45
aI71
aI225
aI211
aI539
aI1881
aI1201
aI5675
aI6061
aI12121
aI13289
aI30455
aI33131
asS'd'
I18802
sa(dp37604
S'a'
I90820
sS's'
I18
sS'm'
(lp37605
I1
aI1
aI5
aI15
aI5
aI59
aI35
aI121
aI185
aI143
aI165
aI2999
aI7907
aI5035
aI8337
aI11951
aI66403
aI219997
asS'd'
I18803
sa(dp37606
S'a'
I90844
sS's'
I18
sS'm'
(lp37607
I1
aI1
aI5
aI9
aI3
aI7
aI27
aI129
aI245
aI93
aI715
aI1249
aI1717
aI13381
aI31255
aI23153
aI22227
aI8077
asS'd'
I18804
sa(dp37608
S'a'
I90907
sS's'
I18
sS'm'
(lp37609
I1
aI3
aI5
aI11
aI19
aI49
aI21
aI163
aI157
aI615
aI1475
aI2453
aI6315
aI12325
aI26565
aI58399
aI49385
aI252127
asS'd'
I18805
sa(dp37610
S'a'
I90925
sS's'
I18
sS'm'
(lp37611
I1
aI3
aI7
aI11
aI23
aI3
aI35
aI61
aI409
aI795
aI1447
aI3461
aI535
aI6533
aI25757
aI31783
aI9509
aI217589
asS'd'
I18806
sa(dp37612
S'a'
I90926
sS's'
I18
sS'm'
(lp37613
I1
aI1
aI5
aI15
aI9
aI21
aI67
aI65
aI399
aI515
aI777
aI3183
aI1155
aI16071
aI7339
aI59985
aI56659
aI200701
asS'd'
I18807
sa(dp37614
S'a'
I90934
sS's'
I18
sS'm'
(lp37615
I1
aI3
aI3
aI1
aI15
aI33
aI119
aI145
aI71
aI517
aI1775
aI163
aI5307
aI1549
aI31071
aI56289
aI128395
aI230381
asS'd'
I18808
sa(dp37616
S'a'
I90946
sS's'
I18
sS'm'
(lp37617
I1
aI3
aI1
aI7
aI7
aI47
aI97
aI187
aI193
aI887
aI515
aI1301
aI4841
aI12069
aI413
aI41503
aI36421
aI45909
asS'd'
I18809
sa(dp37618
S'a'
I90981
sS's'
I18
sS'm'
(lp37619
I1
aI1
aI5
aI3
aI17
aI21
aI53
aI227
aI137
aI865
aI715
aI3601
aI5027
aI2983
aI24113
aI23349
aI106391
aI188193
asS'd'
I18810
sa(dp37620
S'a'
I90999
sS's'
I18
sS'm'
(lp37621
I1
aI3
aI7
aI15
aI29
aI37
aI91
aI235
aI351
aI15
aI425
aI681
aI187
aI1517
aI30079
aI41347
aI49691
aI66369
asS'd'
I18811
sa(dp37622
S'a'
I91003
sS's'
I18
sS'm'
(lp37623
I1
aI1
aI5
aI5
aI15
aI25
aI93
aI245
aI397
aI517
aI1635
aI2475
aI5543
aI9597
aI27721
aI21475
aI79571
aI259011
asS'd'
I18812
sa(dp37624
S'a'
I91019
sS's'
I18
sS'm'
(lp37625
I1
aI1
aI1
aI5
aI21
aI29
aI5
aI161
aI37
aI409
aI1661
aI3371
aI5663
aI4317
aI9951
aI23605
aI7393
aI90593
asS'd'
I18813
sa(dp37626
S'a'
I91022
sS's'
I18
sS'm'
(lp37627
I1
aI1
aI3
aI15
aI21
aI51
aI79
aI159
aI209
aI891
aI1391
aI2895
aI5003
aI6601
aI17983
aI42359
aI104497
aI162181
asS'd'
I18814
sa(dp37628
S'a'
I91050
sS's'
I18
sS'm'
(lp37629
I1
aI3
aI7
aI11
aI23
aI51
aI99
aI145
aI21
aI345
aI1389
aI1035
aI5939
aI8293
aI22765
aI23331
aI7789
aI115149
asS'd'
I18815
sa(dp37630
S'a'
I91055
sS's'
I18
sS'm'
(lp37631
I1
aI1
aI5
aI11
aI27
aI61
aI115
aI123
aI317
aI607
aI463
aI779
aI5121
aI3861
aI18761
aI39407
aI125837
aI244163
asS'd'
I18816
sa(dp37632
S'a'
I91064
sS's'
I18
sS'm'
(lp37633
I1
aI1
aI7
aI3
aI15
aI9
aI33
aI1
aI437
aI621
aI31
aI147
aI8157
aI3451
aI18223
aI61187
aI125297
aI211225
asS'd'
I18817
sa(dp37634
S'a'
I91078
sS's'
I18
sS'm'
(lp37635
I1
aI3
aI1
aI11
aI27
aI55
aI21
aI251
aI5
aI57
aI1889
aI71
aI3745
aI499
aI9043
aI62683
aI21945
aI138615
asS'd'
I18818
sa(dp37636
S'a'
I91117
sS's'
I18
sS'm'
(lp37637
I1
aI1
aI3
aI13
aI25
aI9
aI1
aI31
aI277
aI373
aI507
aI301
aI2341
aI1741
aI11997
aI47661
aI44121
aI183151
asS'd'
I18819
sa(dp37638
S'a'
I91120
sS's'
I18
sS'm'
(lp37639
I1
aI3
aI3
aI5
aI15
aI51
aI5
aI233
aI475
aI397
aI1833
aI1267
aI7025
aI2593
aI15425
aI47053
aI16205
aI208007
asS'd'
I18820
sa(dp37640
S'a'
I91129
sS's'
I18
sS'm'
(lp37641
I1
aI3
aI3
aI9
aI31
aI51
aI123
aI219
aI493
aI789
aI659
aI985
aI7283
aI11545
aI15383
aI25173
aI130423
aI196619
asS'd'
I18821
sa(dp37642
S'a'
I91140
sS's'
I18
sS'm'
(lp37643
I1
aI1
aI1
aI15
aI9
aI17
aI37
aI145
aI297
aI933
aI1019
aI1699
aI2149
aI12391
aI17003
aI42157
aI126283
aI252231
asS'd'
I18822
sa(dp37644
S'a'
I91157
sS's'
I18
sS'm'
(lp37645
I1
aI3
aI5
aI9
aI17
aI1
aI91
aI69
aI335
aI857
aI925
aI3855
aI2225
aI6909
aI19101
aI12191
aI92117
aI229077
asS'd'
I18823
sa(dp37646
S'a'
I91161
sS's'
I18
sS'm'
(lp37647
I1
aI3
aI5
aI13
aI5
aI21
aI67
aI17
aI307
aI879
aI1563
aI3169
aI745
aI6799
aI27237
aI39621
aI1413
aI146295
asS'd'
I18824
sa(dp37648
S'a'
I91240
sS's'
I18
sS'm'
(lp37649
I1
aI3
aI7
aI13
aI17
aI59
aI23
aI253
aI415
aI761
aI451
aI2773
aI3523
aI10985
aI29853
aI7275
aI79521
aI133447
asS'd'
I18825
sa(dp37650
S'a'
I91260
sS's'
I18
sS'm'
(lp37651
I1
aI1
aI1
aI1
aI15
aI43
aI89
aI19
aI85
aI837
aI1335
aI1641
aI105
aI5429
aI8317
aI45555
aI104447
aI102313
asS'd'
I18826
sa(dp37652
S'a'
I91270
sS's'
I18
sS'm'
(lp37653
I1
aI3
aI3
aI5
aI17
aI21
aI67
aI235
aI367
aI265
aI1069
aI835
aI5457
aI177
aI26987
aI39477
aI6895
aI123283
asS'd'
I18827
sa(dp37654
S'a'
I91287
sS's'
I18
sS'm'
(lp37655
I1
aI1
aI5
aI15
aI19
aI61
aI5
aI15
aI487
aI291
aI661
aI825
aI1569
aI8795
aI13035
aI57077
aI112847
aI160267
asS'd'
I18828
sa(dp37656
S'a'
I91321
sS's'
I18
sS'm'
(lp37657
I1
aI1
aI7
aI9
aI27
aI21
aI91
aI133
aI361
aI7
aI447
aI3035
aI3523
aI13167
aI15927
aI35555
aI35713
aI91291
asS'd'
I18829
sa(dp37658
S'a'
I91336
sS's'
I18
sS'm'
(lp37659
I1
aI3
aI7
aI11
aI17
aI23
aI77
aI255
aI437
aI897
aI1185
aI1633
aI6451
aI13081
aI29097
aI61335
aI39671
aI177835
asS'd'
I18830
sa(dp37660
S'a'
I91341
sS's'
I18
sS'm'
(lp37661
I1
aI1
aI7
aI5
aI31
aI49
aI81
aI221
aI167
aI241
aI1895
aI1813
aI1493
aI2475
aI379
aI5685
aI116341
aI121823
asS'd'
I18831
sa(dp37662
S'a'
I91378
sS's'
I18
sS'm'
(lp37663
I1
aI3
aI5
aI9
aI25
aI11
aI117
aI253
aI337
aI381
aI743
aI69
aI1641
aI3649
aI26335
aI59683
aI28729
aI83449
asS'd'
I18832
sa(dp37664
S'a'
I91398
sS's'
I18
sS'm'
(lp37665
I1
aI3
aI1
aI7
aI15
aI17
aI59
aI9
aI129
aI233
aI1905
aI1371
aI6521
aI8953
aI26173
aI9707
aI70817
aI260035
asS'd'
I18833
sa(dp37666
S'a'
I91437
sS's'
I18
sS'm'
(lp37667
I1
aI3
aI1
aI5
aI19
aI53
aI23
aI255
aI305
aI835
aI1387
aI2947
aI3013
aI9117
aI27571
aI47123
aI49881
aI47229
asS'd'
I18834
sa(dp37668
S'a'
I91440
sS's'
I18
sS'm'
(lp37669
I1
aI1
aI3
aI5
aI23
aI37
aI123
aI193
aI365
aI49
aI1211
aI3083
aI8133
aI14205
aI11361
aI55945
aI23225
aI159109
asS'd'
I18835
sa(dp37670
S'a'
I91475
sS's'
I18
sS'm'
(lp37671
I1
aI1
aI7
aI11
aI5
aI49
aI43
aI165
aI97
aI581
aI617
aI3045
aI6187
aI8399
aI24045
aI46713
aI28389
aI156811
asS'd'
I18836
sa(dp37672
S'a'
I91494
sS's'
I18
sS'm'
(lp37673
I1
aI3
aI1
aI5
aI29
aI17
aI75
aI231
aI59
aI143
aI2041
aI2319
aI2289
aI11805
aI4039
aI29895
aI91305
aI179091
asS'd'
I18837
sa(dp37674
S'a'
I91500
sS's'
I18
sS'm'
(lp37675
I1
aI1
aI7
aI7
aI29
aI5
aI13
aI43
aI409
aI81
aI751
aI2157
aI2543
aI13317
aI28275
aI60871
aI119833
aI36743
asS'd'
I18838
sa(dp37676
S'a'
I91511
sS's'
I18
sS'm'
(lp37677
I1
aI1
aI5
aI13
aI5
aI5
aI21
aI101
aI497
aI993
aI157
aI647
aI3587
aI1495
aI20233
aI30889
aI112579
aI172009
asS'd'
I18839
sa(dp37678
S'a'
I91567
sS's'
I18
sS'm'
(lp37679
I1
aI1
aI7
aI13
aI7
aI19
aI101
aI217
aI305
aI897
aI1305
aI1693
aI6881
aI2415
aI17373
aI56327
aI53971
aI19021
asS'd'
I18840
sa(dp37680
S'a'
I91572
sS's'
I18
sS'm'
(lp37681
I1
aI1
aI7
aI15
aI31
aI61
aI93
aI99
aI459
aI999
aI239
aI969
aI2427
aI12295
aI23699
aI4839
aI73707
aI110365
asS'd'
I18841
sa(dp37682
S'a'
I91601
sS's'
I18
sS'm'
(lp37683
I1
aI3
aI5
aI9
aI15
aI59
aI29
aI35
aI15
aI331
aI93
aI529
aI2651
aI5675
aI3039
aI25967
aI2907
aI222053
asS'd'
I18842
sa(dp37684
S'a'
I91617
sS's'
I18
sS'm'
(lp37685
I1
aI3
aI3
aI1
aI1
aI17
aI59
aI81
aI495
aI917
aI1907
aI3
aI1989
aI14339
aI21311
aI60909
aI39393
aI54115
asS'd'
I18843
sa(dp37686
S'a'
I91653
sS's'
I18
sS'm'
(lp37687
I1
aI3
aI1
aI9
aI1
aI37
aI31
aI201
aI251
aI117
aI1753
aI2453
aI5007
aI6935
aI1165
aI49231
aI51495
aI200219
asS'd'
I18844
sa(dp37688
S'a'
I91678
sS's'
I18
sS'm'
(lp37689
I1
aI1
aI3
aI3
aI15
aI47
aI17
aI175
aI77
aI363
aI1455
aI1417
aI1357
aI2295
aI31165
aI53337
aI97891
aI145621
asS'd'
I18845
sa(dp37690
S'a'
I91705
sS's'
I18
sS'm'
(lp37691
I1
aI3
aI3
aI3
aI27
aI31
aI41
aI179
aI47
aI629
aI5
aI2543
aI6817
aI8953
aI7151
aI20715
aI52363
aI251037
asS'd'
I18846
sa(dp37692
S'a'
I91723
sS's'
I18
sS'm'
(lp37693
I1
aI3
aI1
aI1
aI21
aI11
aI103
aI87
aI285
aI189
aI1911
aI2979
aI7563
aI13405
aI2309
aI25695
aI106277
aI179493
asS'd'
I18847
sa(dp37694
S'a'
I91725
sS's'
I18
sS'm'
(lp37695
I1
aI1
aI1
aI5
aI7
aI19
aI1
aI233
aI261
aI825
aI1071
aI3529
aI5617
aI11207
aI24559
aI47461
aI79753
aI41009
asS'd'
I18848
sa(dp37696
S'a'
I91733
sS's'
I18
sS'm'
(lp37697
I1
aI1
aI3
aI1
aI15
aI35
aI65
aI157
aI381
aI509
aI1455
aI3117
aI31
aI2251
aI29729
aI33687
aI74999
aI214765
asS'd'
I18849
sa(dp37698
S'a'
I91761
sS's'
I18
sS'm'
(lp37699
I1
aI1
aI7
aI11
aI31
aI43
aI11
aI147
aI509
aI891
aI1929
aI357
aI3905
aI16251
aI30169
aI27787
aI124003
aI142587
asS'd'
I18850
sa(dp37700
S'a'
I91768
sS's'
I18
sS'm'
(lp37701
I1
aI1
aI7
aI11
aI3
aI31
aI93
aI161
aI311
aI377
aI1119
aI2177
aI4339
aI3889
aI24299
aI35167
aI87583
aI145611
asS'd'
I18851
sa(dp37702
S'a'
I91790
sS's'
I18
sS'm'
(lp37703
I1
aI1
aI5
aI9
aI23
aI9
aI73
aI85
aI233
aI919
aI1319
aI13
aI3353
aI1029
aI31251
aI17731
aI86759
aI11705
asS'd'
I18852
sa(dp37704
S'a'
I91804
sS's'
I18
sS'm'
(lp37705
I1
aI1
aI1
aI7
aI27
aI23
aI67
aI235
aI207
aI161
aI697
aI2433
aI833
aI5305
aI32695
aI29327
aI25285
aI51289
asS'd'
I18853
sa(dp37706
S'a'
I91807
sS's'
I18
sS'm'
(lp37707
I1
aI3
aI3
aI3
aI31
aI55
aI107
aI211
aI61
aI993
aI1443
aI463
aI5029
aI5401
aI8821
aI29721
aI113939
aI194839
asS'd'
I18854
sa(dp37708
S'a'
I91808
sS's'
I18
sS'm'
(lp37709
I1
aI1
aI1
aI11
aI11
aI33
aI39
aI167
aI17
aI863
aI363
aI3967
aI2277
aI4053
aI15403
aI31887
aI98565
aI217953
asS'd'
I18855
sa(dp37710
S'a'
I91818
sS's'
I18
sS'm'
(lp37711
I1
aI1
aI7
aI1
aI21
aI47
aI97
aI147
aI155
aI327
aI1417
aI3531
aI7649
aI8975
aI21221
aI57631
aI72611
aI97745
asS'd'
I18856
sa(dp37712
S'a'
I91840
sS's'
I18
sS'm'
(lp37713
I1
aI3
aI3
aI1
aI1
aI63
aI19
aI91
aI105
aI991
aI819
aI673
aI7845
aI14947
aI1633
aI40517
aI91525
aI151041
asS'd'
I18857
sa(dp37714
S'a'
I91849
sS's'
I18
sS'm'
(lp37715
I1
aI1
aI7
aI11
aI25
aI43
aI57
aI141
aI65
aI415
aI1045
aI3947
aI7099
aI11653
aI29321
aI51591
aI2591
aI44803
asS'd'
I18858
sa(dp37716
S'a'
I91869
sS's'
I18
sS'm'
(lp37717
I1
aI3
aI1
aI9
aI15
aI19
aI105
aI37
aI485
aI3
aI213
aI1217
aI951
aI5637
aI1589
aI25501
aI95073
aI124683
asS'd'
I18859
sa(dp37718
S'a'
I91873
sS's'
I18
sS'm'
(lp37719
I1
aI1
aI5
aI7
aI13
aI19
aI55
aI143
aI507
aI575
aI715
aI1633
aI5201
aI10493
aI26041
aI18407
aI8097
aI152313
asS'd'
I18860
sa(dp37720
S'a'
I91876
sS's'
I18
sS'm'
(lp37721
I1
aI1
aI3
aI5
aI9
aI51
aI5
aI171
aI143
aI877
aI1571
aI2997
aI4209
aI13423
aI9389
aI23015
aI6665
aI254799
asS'd'
I18861
sa(dp37722
S'a'
I91888
sS's'
I18
sS'm'
(lp37723
I1
aI3
aI5
aI15
aI31
aI43
aI87
aI79
aI89
aI463
aI1075
aI1257
aI1631
aI13225
aI13529
aI53267
aI73651
aI89125
asS'd'
I18862
sa(dp37724
S'a'
I91891
sS's'
I18
sS'm'
(lp37725
I1
aI3
aI7
aI13
aI23
aI17
aI93
aI113
aI45
aI225
aI1939
aI3301
aI6031
aI9749
aI16577
aI12857
aI68437
aI169861
asS'd'
I18863
sa(dp37726
S'a'
I91897
sS's'
I18
sS'm'
(lp37727
I1
aI3
aI3
aI15
aI31
aI11
aI91
aI127
aI227
aI813
aI105
aI901
aI6861
aI10627
aI18425
aI2553
aI102503
aI83167
asS'd'
I18864
sa(dp37728
S'a'
I91911
sS's'
I18
sS'm'
(lp37729
I1
aI3
aI3
aI13
aI17
aI63
aI83
aI163
aI451
aI659
aI1995
aI2283
aI6297
aI8097
aI20935
aI6017
aI4977
aI5045
asS'd'
I18865
sa(dp37730
S'a'
I91965
sS's'
I18
sS'm'
(lp37731
I1
aI1
aI5
aI13
aI27
aI47
aI103
aI129
aI259
aI975
aI391
aI2343
aI6639
aI1385
aI30187
aI35401
aI74321
aI24751
asS'd'
I18866
sa(dp37732
S'a'
I91971
sS's'
I18
sS'm'
(lp37733
I1
aI3
aI1
aI13
aI1
aI57
aI37
aI65
aI57
aI413
aI63
aI3819
aI5915
aI3925
aI20777
aI48539
aI3019
aI54965
asS'd'
I18867
sa(dp37734
S'a'
I91988
sS's'
I18
sS'm'
(lp37735
I1
aI1
aI7
aI3
aI7
aI13
aI91
aI33
aI143
aI489
aI657
aI3127
aI707
aI10841
aI11307
aI37855
aI92697
aI119189
asS'd'
I18868
sa(dp37736
S'a'
I91992
sS's'
I18
sS'm'
(lp37737
I1
aI1
aI3
aI11
aI25
aI47
aI11
aI57
aI463
aI693
aI55
aI501
aI3765
aI15443
aI12917
aI61677
aI97145
aI213637
asS'd'
I18869
sa(dp37738
S'a'
I92004
sS's'
I18
sS'm'
(lp37739
I1
aI3
aI1
aI1
aI13
aI49
aI13
aI225
aI101
aI475
aI627
aI1447
aI7587
aI11335
aI3599
aI20795
aI72915
aI174663
asS'd'
I18870
sa(dp37740
S'a'
I92021
sS's'
I18
sS'm'
(lp37741
I1
aI3
aI7
aI5
aI31
aI15
aI115
aI255
aI329
aI365
aI959
aI3399
aI4695
aI14537
aI1447
aI17391
aI88557
aI130213
asS'd'
I18871
sa(dp37742
S'a'
I92035
sS's'
I18
sS'm'
(lp37743
I1
aI1
aI1
aI9
aI25
aI47
aI29
aI173
aI29
aI149
aI291
aI691
aI7621
aI7607
aI20769
aI7149
aI27323
aI57689
asS'd'
I18872
sa(dp37744
S'a'
I92068
sS's'
I18
sS'm'
(lp37745
I1
aI3
aI5
aI15
aI3
aI61
aI119
aI247
aI25
aI495
aI1297
aI1119
aI8011
aI16077
aI21567
aI30559
aI88455
aI68763
asS'd'
I18873
sa(dp37746
S'a'
I92098
sS's'
I18
sS'm'
(lp37747
I1
aI1
aI7
aI1
aI11
aI47
aI109
aI115
aI313
aI517
aI1951
aI3319
aI337
aI11793
aI22345
aI33457
aI47383
aI213893
asS'd'
I18874
sa(dp37748
S'a'
I92104
sS's'
I18
sS'm'
(lp37749
I1
aI1
aI5
aI11
aI11
aI53
aI103
aI237
aI383
aI927
aI421
aI4085
aI3327
aI169
aI9941
aI24753
aI65437
aI108173
asS'd'
I18875
sa(dp37750
S'a'
I92121
sS's'
I18
sS'm'
(lp37751
I1
aI1
aI5
aI11
aI17
aI59
aI107
aI53
aI479
aI143
aI825
aI2667
aI5219
aI6143
aI11573
aI33637
aI124981
aI98195
asS'd'
I18876
sa(dp37752
S'a'
I92127
sS's'
I18
sS'm'
(lp37753
I1
aI3
aI3
aI13
aI3
aI17
aI61
aI129
aI475
aI585
aI1611
aI1791
aI7817
aI4099
aI20437
aI51411
aI130173
aI220085
asS'd'
I18877
sa(dp37754
S'a'
I92151
sS's'
I18
sS'm'
(lp37755
I1
aI3
aI3
aI3
aI27
aI25
aI33
aI255
aI361
aI967
aI1415
aI3213
aI3341
aI15875
aI32359
aI53267
aI27665
aI178301
asS'd'
I18878
sa(dp37756
S'a'
I92162
sS's'
I18
sS'm'
(lp37757
I1
aI1
aI5
aI3
aI13
aI9
aI91
aI187
aI173
aI525
aI1675
aI2217
aI4093
aI2009
aI16917
aI18485
aI104849
aI163233
asS'd'
I18879
sa(dp37758
S'a'
I92186
sS's'
I18
sS'm'
(lp37759
I1
aI1
aI1
aI11
aI3
aI17
aI125
aI157
aI9
aI429
aI1573
aI2257
aI7943
aI9893
aI5611
aI64619
aI4509
aI200181
asS'd'
I18880
sa(dp37760
S'a'
I92191
sS's'
I18
sS'm'
(lp37761
I1
aI3
aI1
aI1
aI3
aI9
aI83
aI53
aI315
aI85
aI1093
aI2621
aI663
aI12369
aI317
aI6089
aI16479
aI225071
asS'd'
I18881
sa(dp37762
S'a'
I92192
sS's'
I18
sS'm'
(lp37763
I1
aI3
aI3
aI1
aI25
aI47
aI57
aI45
aI219
aI45
aI945
aI3989
aI4889
aI8989
aI381
aI52483
aI57029
aI253899
asS'd'
I18882
sa(dp37764
S'a'
I92207
sS's'
I18
sS'm'
(lp37765
I1
aI3
aI3
aI5
aI3
aI61
aI75
aI189
aI53
aI489
aI553
aI2381
aI7485
aI9941
aI29733
aI2611
aI74119
aI203647
asS'd'
I18883
sa(dp37766
S'a'
I92260
sS's'
I18
sS'm'
(lp37767
I1
aI3
aI1
aI11
aI27
aI63
aI53
aI61
aI59
aI613
aI465
aI867
aI1985
aI7605
aI14301
aI53847
aI68547
aI14717
asS'd'
I18884
sa(dp37768
S'a'
I92282
sS's'
I18
sS'm'
(lp37769
I1
aI3
aI5
aI11
aI5
aI41
aI51
aI11
aI59
aI761
aI59
aI267
aI7273
aI3061
aI11223
aI48825
aI117869
aI158551
asS'd'
I18885
sa(dp37770
S'a'
I92288
sS's'
I18
sS'm'
(lp37771
I1
aI3
aI7
aI1
aI31
aI47
aI111
aI43
aI435
aI997
aI135
aI3369
aI6439
aI5637
aI13629
aI13221
aI90607
aI86359
asS'd'
I18886
sa(dp37772
S'a'
I92297
sS's'
I18
sS'm'
(lp37773
I1
aI1
aI1
aI11
aI23
aI51
aI109
aI223
aI495
aI765
aI1557
aI3545
aI305
aI4949
aI23931
aI45115
aI12121
aI14487
asS'd'
I18887
sa(dp37774
S'a'
I92303
sS's'
I18
sS'm'
(lp37775
I1
aI3
aI5
aI9
aI5
aI25
aI41
aI249
aI27
aI375
aI1339
aI3647
aI3529
aI2077
aI21091
aI45523
aI67191
aI1257
asS'd'
I18888
sa(dp37776
S'a'
I92315
sS's'
I18
sS'm'
(lp37777
I1
aI1
aI3
aI7
aI13
aI15
aI63
aI45
aI187
aI761
aI1245
aI3381
aI1817
aI2491
aI16469
aI64417
aI87333
aI143103
asS'd'
I18889
sa(dp37778
S'a'
I92322
sS's'
I18
sS'm'
(lp37779
I1
aI1
aI3
aI13
aI11
aI33
aI87
aI11
aI279
aI689
aI1047
aI3935
aI5359
aI11309
aI19735
aI33259
aI12347
aI183653
asS'd'
I18890
sa(dp37780
S'a'
I92327
sS's'
I18
sS'm'
(lp37781
I1
aI1
aI5
aI7
aI5
aI49
aI109
aI221
aI455
aI167
aI785
aI1859
aI4337
aI14937
aI209
aI23435
aI22923
aI172985
asS'd'
I18891
sa(dp37782
S'a'
I92341
sS's'
I18
sS'm'
(lp37783
I1
aI3
aI7
aI9
aI13
aI7
aI127
aI117
aI147
aI741
aI531
aI2627
aI2565
aI11083
aI30577
aI42471
aI77065
aI120983
asS'd'
I18892
sa(dp37784
S'a'
I92389
sS's'
I18
sS'm'
(lp37785
I1
aI1
aI7
aI1
aI5
aI61
aI115
aI203
aI15
aI305
aI1005
aI2085
aI2597
aI4371
aI11661
aI33219
aI53657
aI40325
asS'd'
I18893
sa(dp37786
S'a'
I92390
sS's'
I18
sS'm'
(lp37787
I1
aI3
aI7
aI13
aI13
aI15
aI69
aI167
aI369
aI747
aI1115
aI1493
aI4881
aI2693
aI32281
aI27089
aI56821
aI121693
asS'd'
I18894
sa(dp37788
S'a'
I92422
sS's'
I18
sS'm'
(lp37789
I1
aI3
aI5
aI3
aI19
aI51
aI101
aI29
aI411
aI509
aI847
aI1033
aI4135
aI15561
aI7045
aI60757
aI48479
aI247295
asS'd'
I18895
sa(dp37790
S'a'
I92439
sS's'
I18
sS'm'
(lp37791
I1
aI1
aI5
aI3
aI27
aI47
aI103
aI123
aI413
aI71
aI689
aI2113
aI4347
aI1983
aI25727
aI20095
aI3271
aI133081
asS'd'
I18896
sa(dp37792
S'a'
I92452
sS's'
I18
sS'm'
(lp37793
I1
aI1
aI3
aI3
aI5
aI39
aI87
aI27
aI505
aI631
aI689
aI2591
aI1955
aI3205
aI12681
aI10821
aI13343
aI101505
asS'd'
I18897
sa(dp37794
S'a'
I92470
sS's'
I18
sS'm'
(lp37795
I1
aI3
aI7
aI9
aI31
aI23
aI103
aI223
aI499
aI721
aI13
aI1399
aI7369
aI3945
aI27727
aI7923
aI60265
aI197793
asS'd'
I18898
sa(dp37796
S'a'
I92473
sS's'
I18
sS'm'
(lp37797
I1
aI3
aI3
aI5
aI27
aI7
aI119
aI23
aI371
aI495
aI1583
aI3913
aI5139
aI12151
aI17477
aI10907
aI121775
aI13369
asS'd'
I18899
sa(dp37798
S'a'
I92493
sS's'
I18
sS'm'
(lp37799
I1
aI3
aI7
aI1
aI19
aI53
aI91
aI235
aI161
aI97
aI37
aI1115
aI5909
aI1943
aI8137
aI1541
aI16253
aI252151
asS'd'
I18900
sa(dp37800
S'a'
I92508
sS's'
I18
sS'm'
(lp37801
I1
aI3
aI7
aI5
aI3
aI1
aI107
aI241
aI187
aI253
aI1225
aI2827
aI4191
aI2749
aI25629
aI47465
aI19969
aI45035
asS'd'
I18901
sa(dp37802
S'a'
I92517
sS's'
I18
sS'm'
(lp37803
I1
aI3
aI1
aI5
aI1
aI29
aI47
aI233
aI175
aI313
aI793
aI2089
aI6235
aI6595
aI27599
aI20505
aI63379
aI8729
asS'd'
I18902
sa(dp37804
S'a'
I92529
sS's'
I18
sS'm'
(lp37805
I1
aI3
aI7
aI11
aI15
aI9
aI87
aI113
aI389
aI1
aI1057
aI3307
aI3455
aI1847
aI1497
aI28115
aI92897
aI2383
asS'd'
I18903
sa(dp37806
S'a'
I92551
sS's'
I18
sS'm'
(lp37807
I1
aI3
aI7
aI13
aI19
aI59
aI45
aI59
aI49
aI273
aI1619
aI1975
aI5949
aI9951
aI7685
aI52559
aI42377
aI29855
asS'd'
I18904
sa(dp37808
S'a'
I92557
sS's'
I18
sS'm'
(lp37809
I1
aI1
aI3
aI9
aI19
aI7
aI119
aI35
aI85
aI37
aI269
aI3443
aI8015
aI8061
aI6001
aI19123
aI70643
aI115513
asS'd'
I18905
sa(dp37810
S'a'
I92596
sS's'
I18
sS'm'
(lp37811
I1
aI1
aI3
aI15
aI3
aI19
aI83
aI171
aI259
aI207
aI1495
aI513
aI5455
aI4071
aI27471
aI15773
aI66301
aI228743
asS'd'
I18906
sa(dp37812
S'a'
I92611
sS's'
I18
sS'm'
(lp37813
I1
aI3
aI7
aI9
aI3
aI27
aI93
aI3
aI471
aI13
aI677
aI4067
aI1941
aI15345
aI26629
aI29419
aI121593
aI108669
asS'd'
I18907
sa(dp37814
S'a'
I92613
sS's'
I18
sS'm'
(lp37815
I1
aI3
aI7
aI15
aI29
aI43
aI97
aI41
aI15
aI181
aI1969
aI1901
aI7237
aI3879
aI19337
aI17659
aI17957
aI164667
asS'd'
I18908
sa(dp37816
S'a'
I92642
sS's'
I18
sS'm'
(lp37817
I1
aI3
aI1
aI1
aI25
aI33
aI7
aI41
aI387
aI469
aI795
aI781
aI113
aI4161
aI29687
aI32225
aI73905
aI137879
asS'd'
I18909
sa(dp37818
S'a'
I92666
sS's'
I18
sS'm'
(lp37819
I1
aI1
aI3
aI13
aI9
aI59
aI89
aI23
aI393
aI111
aI1957
aI719
aI6179
aI16183
aI31331
aI48015
aI32147
aI31691
asS'd'
I18910
sa(dp37820
S'a'
I92684
sS's'
I18
sS'm'
(lp37821
I1
aI3
aI7
aI5
aI9
aI45
aI73
aI219
aI181
aI51
aI717
aI1813
aI2581
aI1395
aI17595
aI23689
aI89709
aI201451
asS'd'
I18911
sa(dp37822
S'a'
I92723
sS's'
I18
sS'm'
(lp37823
I1
aI3
aI7
aI3
aI1
aI21
aI15
aI35
aI131
aI515
aI803
aI1429
aI3855
aI349
aI11795
aI26787
aI6109
aI117745
asS'd'
I18912
sa(dp37824
S'a'
I92732
sS's'
I18
sS'm'
(lp37825
I1
aI3
aI3
aI9
aI7
aI11
aI57
aI15
aI491
aI371
aI1787
aI85
aI577
aI11455
aI27419
aI20687
aI2493
aI209993
asS'd'
I18913
sa(dp37826
S'a'
I92738
sS's'
I18
sS'm'
(lp37827
I1
aI3
aI7
aI15
aI5
aI5
aI87
aI197
aI93
aI643
aI247
aI31
aI357
aI7377
aI10509
aI29883
aI42747
aI248861
asS'd'
I18914
sa(dp37828
S'a'
I92783
sS's'
I18
sS'm'
(lp37829
I1
aI3
aI3
aI3
aI29
aI33
aI47
aI253
aI485
aI25
aI2003
aI2953
aI1629
aI11549
aI5697
aI1135
aI117761
aI96411
asS'd'
I18915
sa(dp37830
S'a'
I92808
sS's'
I18
sS'm'
(lp37831
I1
aI3
aI5
aI1
aI29
aI5
aI27
aI187
aI235
aI423
aI41
aI1855
aI4359
aI15627
aI28409
aI49331
aI37735
aI68823
asS'd'
I18916
sa(dp37832
S'a'
I92816
sS's'
I18
sS'm'
(lp37833
I1
aI1
aI5
aI5
aI21
aI61
aI67
aI85
aI41
aI671
aI1617
aI3867
aI7913
aI1693
aI18487
aI1831
aI100971
aI168191
asS'd'
I18917
sa(dp37834
S'a'
I92835
sS's'
I18
sS'm'
(lp37835
I1
aI1
aI7
aI3
aI1
aI61
aI111
aI87
aI55
aI229
aI217
aI2801
aI563
aI13617
aI9641
aI22247
aI16039
aI113541
asS'd'
I18918
sa(dp37836
S'a'
I92847
sS's'
I18
sS'm'
(lp37837
I1
aI3
aI5
aI7
aI5
aI29
aI67
aI99
aI91
aI561
aI1203
aI643
aI2607
aI13421
aI29695
aI31925
aI82985
aI69031
asS'd'
I18919
sa(dp37838
S'a'
I92879
sS's'
I18
sS'm'
(lp37839
I1
aI1
aI1
aI1
aI27
aI7
aI63
aI107
aI269
aI163
aI1711
aI587
aI5657
aI15077
aI24709
aI10235
aI95483
aI94799
asS'd'
I18920
sa(dp37840
S'a'
I92881
sS's'
I18
sS'm'
(lp37841
I1
aI3
aI3
aI5
aI29
aI5
aI127
aI137
aI67
aI609
aI1657
aI1131
aI959
aI15773
aI17295
aI58575
aI96525
aI80529
asS'd'
I18921
sa(dp37842
S'a'
I92884
sS's'
I18
sS'm'
(lp37843
I1
aI3
aI7
aI7
aI25
aI15
aI89
aI93
aI145
aI695
aI367
aI2853
aI3073
aI4867
aI26823
aI31467
aI94769
aI9145
asS'd'
I18922
sa(dp37844
S'a'
I92932
sS's'
I18
sS'm'
(lp37845
I1
aI1
aI7
aI15
aI15
aI5
aI1
aI225
aI57
aI381
aI1295
aI2525
aI1493
aI2401
aI91
aI19809
aI32803
aI195289
asS'd'
I18923
sa(dp37846
S'a'
I92935
sS's'
I18
sS'm'
(lp37847
I1
aI3
aI3
aI7
aI29
aI51
aI29
aI63
aI249
aI107
aI1689
aI3703
aI7227
aI6967
aI27861
aI39167
aI20043
aI218827
asS'd'
I18924
sa(dp37848
S'a'
I92960
sS's'
I18
sS'm'
(lp37849
I1
aI1
aI3
aI13
aI17
aI1
aI77
aI143
aI167
aI255
aI1709
aI2089
aI7465
aI4805
aI16185
aI15167
aI20493
aI240855
asS'd'
I18925
sa(dp37850
S'a'
I92990
sS's'
I18
sS'm'
(lp37851
I1
aI1
aI5
aI1
aI11
aI43
aI107
aI175
aI93
aI955
aI615
aI2923
aI3637
aI7451
aI18847
aI53467
aI12463
aI127249
asS'd'
I18926
sa(dp37852
S'a'
I93010
sS's'
I18
sS'm'
(lp37853
I1
aI1
aI3
aI13
aI31
aI1
aI61
aI113
aI479
aI777
aI1805
aI3625
aI6299
aI12221
aI29599
aI60175
aI31165
aI122815
asS'd'
I18927
sa(dp37854
S'a'
I93025
sS's'
I18
sS'm'
(lp37855
I1
aI3
aI3
aI11
aI11
aI29
aI89
aI129
aI195
aI337
aI1843
aI2769
aI1747
aI7137
aI9901
aI18459
aI25215
aI70609
asS'd'
I18928
sa(dp37856
S'a'
I93046
sS's'
I18
sS'm'
(lp37857
I1
aI1
aI3
aI7
aI17
aI35
aI55
aI81
aI413
aI25
aI1505
aI2185
aI3121
aI11435
aI17885
aI12543
aI36767
aI64039
asS'd'
I18929
sa(dp37858
S'a'
I93085
sS's'
I18
sS'm'
(lp37859
I1
aI3
aI1
aI13
aI25
aI9
aI83
aI25
aI5
aI49
aI1975
aI3967
aI4135
aI13213
aI26479
aI63913
aI14921
aI96193
asS'd'
I18930
sa(dp37860
S'a'
I93089
sS's'
I18
sS'm'
(lp37861
I1
aI3
aI7
aI5
aI17
aI15
aI101
aI47
aI245
aI821
aI1275
aI3343
aI5471
aI5045
aI31741
aI3319
aI8141
aI95501
asS'd'
I18931
sa(dp37862
S'a'
I93096
sS's'
I18
sS'm'
(lp37863
I1
aI3
aI1
aI13
aI1
aI5
aI105
aI39
aI175
aI439
aI1625
aI249
aI4859
aI12449
aI30529
aI45669
aI49071
aI214037
asS'd'
I18932
sa(dp37864
S'a'
I93145
sS's'
I18
sS'm'
(lp37865
I1
aI1
aI7
aI7
aI17
aI21
aI83
aI123
aI261
aI559
aI1967
aI2933
aI4417
aI8331
aI10119
aI21793
aI128729
aI187247
asS'd'
I18933
sa(dp37866
S'a'
I93155
sS's'
I18
sS'm'
(lp37867
I1
aI1
aI7
aI9
aI15
aI43
aI77
aI231
aI241
aI419
aI503
aI3335
aI927
aI2567
aI31259
aI52453
aI114441
aI257449
asS'd'
I18934
sa(dp37868
S'a'
I93169
sS's'
I18
sS'm'
(lp37869
I1
aI3
aI3
aI1
aI9
aI29
aI21
aI89
aI311
aI185
aI519
aI271
aI3595
aI8951
aI6105
aI64593
aI38209
aI120491
asS'd'
I18935
sa(dp37870
S'a'
I93179
sS's'
I18
sS'm'
(lp37871
I1
aI1
aI7
aI9
aI1
aI57
aI65
aI5
aI275
aI615
aI801
aI2839
aI2851
aI15609
aI28731
aI31223
aI87725
aI437
asS'd'
I18936
sa(dp37872
S'a'
I93184
sS's'
I18
sS'm'
(lp37873
I1
aI3
aI3
aI5
aI29
aI3
aI67
aI53
aI17
aI499
aI263
aI651
aI7963
aI5371
aI11593
aI34761
aI57427
aI84979
asS'd'
I18937
sa(dp37874
S'a'
I93204
sS's'
I18
sS'm'
(lp37875
I1
aI3
aI7
aI11
aI15
aI9
aI33
aI165
aI313
aI659
aI909
aI969
aI2309
aI2197
aI27263
aI35273
aI52887
aI236107
asS'd'
I18938
sa(dp37876
S'a'
I93211
sS's'
I18
sS'm'
(lp37877
I1
aI1
aI7
aI1
aI13
aI17
aI29
aI3
aI329
aI573
aI619
aI1013
aI6947
aI7031
aI30773
aI41129
aI116481
aI184233
asS'd'
I18939
sa(dp37878
S'a'
I93213
sS's'
I18
sS'm'
(lp37879
I1
aI1
aI5
aI9
aI13
aI5
aI87
aI235
aI63
aI759
aI1143
aI1861
aI3783
aI2735
aI26191
aI64387
aI3651
aI119447
asS'd'
I18940
sa(dp37880
S'a'
I93227
sS's'
I18
sS'm'
(lp37881
I1
aI1
aI3
aI7
aI15
aI41
aI117
aI135
aI273
aI655
aI251
aI1859
aI4363
aI14725
aI29385
aI6269
aI91505
aI82679
asS'd'
I18941
sa(dp37882
S'a'
I93285
sS's'
I18
sS'm'
(lp37883
I1
aI3
aI7
aI3
aI13
aI21
aI21
aI9
aI121
aI899
aI199
aI1973
aI7437
aI9771
aI26647
aI30909
aI118573
aI152913
asS'd'
I18942
sa(dp37884
S'a'
I93292
sS's'
I18
sS'm'
(lp37885
I1
aI3
aI3
aI9
aI31
aI43
aI5
aI249
aI109
aI183
aI161
aI1185
aI4025
aI10331
aI20983
aI28549
aI122687
aI183429
asS'd'
I18943
sa(dp37886
S'a'
I93320
sS's'
I18
sS'm'
(lp37887
I1
aI3
aI7
aI7
aI11
aI45
aI111
aI99
aI487
aI971
aI597
aI1555
aI273
aI10403
aI25289
aI45483
aI35845
aI35791
asS'd'
I18944
sa(dp37888
S'a'
I93356
sS's'
I18
sS'm'
(lp37889
I1
aI1
aI1
aI7
aI11
aI49
aI125
aI229
aI279
aI289
aI1945
aI3575
aI5683
aI15659
aI31123
aI12517
aI79303
aI255797
asS'd'
I18945
sa(dp37890
S'a'
I93371
sS's'
I18
sS'm'
(lp37891
I1
aI3
aI1
aI15
aI9
aI23
aI61
aI53
aI383
aI855
aI1743
aI407
aI4401
aI7507
aI26307
aI56205
aI110943
aI184183
asS'd'
I18946
sa(dp37892
S'a'
I93374
sS's'
I18
sS'm'
(lp37893
I1
aI3
aI5
aI13
aI23
aI29
aI101
aI243
aI417
aI925
aI1267
aI257
aI5893
aI4335
aI6309
aI43519
aI126035
aI99205
asS'd'
I18947
sa(dp37894
S'a'
I93385
sS's'
I18
sS'm'
(lp37895
I1
aI3
aI1
aI5
aI5
aI35
aI83
aI25
aI31
aI455
aI1799
aI2919
aI7037
aI11829
aI12239
aI12969
aI108469
aI89513
asS'd'
I18948
sa(dp37896
S'a'
I93403
sS's'
I18
sS'm'
(lp37897
I1
aI3
aI5
aI9
aI17
aI29
aI61
aI217
aI183
aI131
aI425
aI4025
aI7141
aI5445
aI21497
aI10603
aI53423
aI5701
asS'd'
I18949
sa(dp37898
S'a'
I93406
sS's'
I18
sS'm'
(lp37899
I1
aI1
aI1
aI15
aI27
aI35
aI9
aI139
aI261
aI43
aI587
aI3835
aI4627
aI11689
aI15739
aI6031
aI73547
aI134271
asS'd'
I18950
sa(dp37900
S'a'
I93427
sS's'
I18
sS'm'
(lp37901
I1
aI1
aI3
aI3
aI25
aI15
aI7
aI225
aI29
aI785
aI2047
aI2219
aI6083
aI7973
aI17053
aI56167
aI83915
aI87597
asS'd'
I18951
sa(dp37902
S'a'
I93451
sS's'
I18
sS'm'
(lp37903
I1
aI3
aI7
aI15
aI13
aI43
aI85
aI121
aI421
aI867
aI1895
aI2437
aI6003
aI5269
aI8625
aI26877
aI100023
aI110229
asS'd'
I18952
sa(dp37904
S'a'
I93454
sS's'
I18
sS'm'
(lp37905
I1
aI3
aI3
aI3
aI25
aI49
aI121
aI1
aI125
aI353
aI1811
aI1575
aI3925
aI13897
aI26087
aI24977
aI105995
aI242817
asS'd'
I18953
sa(dp37906
S'a'
I93472
sS's'
I18
sS'm'
(lp37907
I1
aI1
aI1
aI9
aI31
aI55
aI71
aI241
aI439
aI927
aI955
aI109
aI7779
aI2397
aI18797
aI34177
aI1255
aI178671
asS'd'
I18954
sa(dp37908
S'a'
I93482
sS's'
I18
sS'm'
(lp37909
I1
aI1
aI7
aI7
aI5
aI15
aI99
aI225
aI49
aI407
aI1711
aI4027
aI4845
aI9209
aI20983
aI33969
aI14205
aI9351
asS'd'
I18955
sa(dp37910
S'a'
I93507
sS's'
I18
sS'm'
(lp37911
I1
aI3
aI5
aI1
aI9
aI13
aI113
aI143
aI97
aI189
aI929
aI1163
aI2261
aI9761
aI30011
aI32911
aI117043
aI169493
asS'd'
I18956
sa(dp37912
S'a'
I93537
sS's'
I18
sS'm'
(lp37913
I1
aI1
aI3
aI5
aI9
aI35
aI95
aI77
aI5
aI95
aI1745
aI2013
aI7009
aI5427
aI18969
aI2771
aI5099
aI52939
asS'd'
I18957
sa(dp37914
S'a'
I93562
sS's'
I18
sS'm'
(lp37915
I1
aI3
aI7
aI9
aI13
aI19
aI31
aI189
aI367
aI569
aI95
aI1665
aI6231
aI2169
aI22589
aI8427
aI116097
aI41077
asS'd'
I18958
sa(dp37916
S'a'
I93564
sS's'
I18
sS'm'
(lp37917
I1
aI3
aI7
aI3
aI31
aI61
aI45
aI233
aI327
aI541
aI87
aI3449
aI2767
aI12237
aI17747
aI53827
aI80389
aI121489
asS'd'
I18959
sa(dp37918
S'a'
I93608
sS's'
I18
sS'm'
(lp37919
I1
aI3
aI7
aI15
aI31
aI1
aI49
aI73
aI157
aI131
aI553
aI3417
aI5283
aI4737
aI31675
aI63213
aI43689
aI261869
asS'd'
I18960
sa(dp37920
S'a'
I93636
sS's'
I18
sS'm'
(lp37921
I1
aI3
aI7
aI1
aI3
aI5
aI113
aI43
aI343
aI39
aI135
aI1555
aI7955
aI9851
aI30983
aI21955
aI34871
aI147649
asS'd'
I18961
sa(dp37922
S'a'
I93640
sS's'
I18
sS'm'
(lp37923
I1
aI3
aI5
aI3
aI5
aI27
aI15
aI179
aI141
aI983
aI265
aI2651
aI5907
aI10501
aI6275
aI29629
aI115965
aI125745
asS'd'
I18962
sa(dp37924
S'a'
I93653
sS's'
I18
sS'm'
(lp37925
I1
aI3
aI7
aI13
aI1
aI1
aI81
aI105
aI309
aI457
aI1817
aI3435
aI4615
aI1181
aI27835
aI26075
aI63447
aI44701
asS'd'
I18963
sa(dp37926
S'a'
I93654
sS's'
I18
sS'm'
(lp37927
I1
aI1
aI5
aI5
aI25
aI19
aI85
aI103
aI409
aI323
aI2001
aI3719
aI3403
aI1301
aI19615
aI47829
aI109905
aI65777
asS'd'
I18964
sa(dp37928
S'a'
I93664
sS's'
I18
sS'm'
(lp37929
I1
aI3
aI5
aI3
aI21
aI15
aI47
aI75
aI467
aI273
aI1885
aI3929
aI1877
aI5209
aI6881
aI34431
aI35663
aI100205
asS'd'
I18965
sa(dp37930
S'a'
I93676
sS's'
I18
sS'm'
(lp37931
I1
aI1
aI5
aI3
aI3
aI9
aI47
aI143
aI471
aI653
aI1011
aI2263
aI3673
aI11921
aI31207
aI50365
aI27177
aI214377
asS'd'
I18966
sa(dp37932
S'a'
I93721
sS's'
I18
sS'm'
(lp37933
I1
aI3
aI1
aI15
aI3
aI43
aI81
aI253
aI495
aI139
aI679
aI2207
aI4603
aI5269
aI27133
aI46461
aI120783
aI185595
asS'd'
I18967
sa(dp37934
S'a'
I93740
sS's'
I18
sS'm'
(lp37935
I1
aI1
aI7
aI3
aI13
aI3
aI109
aI197
aI477
aI101
aI859
aI1035
aI777
aI10153
aI15581
aI22715
aI17493
aI120851
asS'd'
I18968
sa(dp37936
S'a'
I93743
sS's'
I18
sS'm'
(lp37937
I1
aI3
aI1
aI3
aI23
aI5
aI121
aI67
aI265
aI935
aI741
aI3311
aI541
aI1093
aI1639
aI5941
aI5587
aI150345
asS'd'
I18969
sa(dp37938
S'a'
I93745
sS's'
I18
sS'm'
(lp37939
I1
aI3
aI1
aI5
aI3
aI13
aI65
aI173
aI493
aI303
aI359
aI3813
aI6007
aI1105
aI12185
aI10431
aI17117
aI164899
asS'd'
I18970
sa(dp37940
S'a'
I93751
sS's'
I18
sS'm'
(lp37941
I1
aI3
aI3
aI1
aI25
aI33
aI71
aI181
aI149
aI7
aI333
aI1981
aI2981
aI14683
aI10997
aI63373
aI22605
aI119681
asS'd'
I18971
sa(dp37942
S'a'
I93770
sS's'
I18
sS'm'
(lp37943
I1
aI1
aI1
aI15
aI29
aI35
aI89
aI21
aI281
aI175
aI587
aI3117
aI7221
aI8239
aI26399
aI49133
aI65895
aI142175
asS'd'
I18972
sa(dp37944
S'a'
I93796
sS's'
I18
sS'm'
(lp37945
I1
aI3
aI7
aI15
aI9
aI9
aI35
aI161
aI65
aI749
aI421
aI3575
aI6307
aI2029
aI11423
aI63901
aI102049
aI26333
asS'd'
I18973
sa(dp37946
S'a'
I93820
sS's'
I18
sS'm'
(lp37947
I1
aI3
aI1
aI13
aI1
aI45
aI97
aI41
aI231
aI245
aI271
aI1497
aI3119
aI6225
aI21665
aI12113
aI67315
aI62779
asS'd'
I18974
sa(dp37948
S'a'
I93824
sS's'
I18
sS'm'
(lp37949
I1
aI1
aI1
aI7
aI3
aI29
aI119
aI193
aI179
aI353
aI1015
aI2803
aI6869
aI7653
aI22309
aI53421
aI86969
aI115549
asS'd'
I18975
sa(dp37950
S'a'
I93833
sS's'
I18
sS'm'
(lp37951
I1
aI3
aI3
aI5
aI17
aI37
aI49
aI129
aI195
aI537
aI1237
aI2775
aI6683
aI699
aI19181
aI61125
aI27483
aI175645
asS'd'
I18976
sa(dp37952
S'a'
I93841
sS's'
I18
sS'm'
(lp37953
I1
aI3
aI7
aI3
aI7
aI49
aI107
aI41
aI285
aI335
aI1415
aI4015
aI1301
aI6525
aI32429
aI9337
aI87923
aI176751
asS'd'
I18977
sa(dp37954
S'a'
I93847
sS's'
I18
sS'm'
(lp37955
I1
aI1
aI5
aI9
aI21
aI43
aI91
aI25
aI225
aI311
aI417
aI303
aI2629
aI3609
aI29987
aI28647
aI104173
aI52383
asS'd'
I18978
sa(dp37956
S'a'
I93848
sS's'
I18
sS'm'
(lp37957
I1
aI3
aI5
aI9
aI13
aI47
aI75
aI143
aI109
aI173
aI503
aI3843
aI1767
aI9433
aI10009
aI5653
aI87339
aI212975
asS'd'
I18979
sa(dp37958
S'a'
I93854
sS's'
I18
sS'm'
(lp37959
I1
aI3
aI1
aI13
aI13
aI55
aI123
aI95
aI499
aI245
aI1875
aI3661
aI7661
aI6927
aI21003
aI51729
aI88089
aI89063
asS'd'
I18980
sa(dp37960
S'a'
I93882
sS's'
I18
sS'm'
(lp37961
I1
aI1
aI3
aI1
aI31
aI7
aI93
aI35
aI169
aI191
aI1079
aI2137
aI4401
aI1563
aI20021
aI9101
aI66881
aI231589
asS'd'
I18981
sa(dp37962
S'a'
I93943
sS's'
I18
sS'm'
(lp37963
I1
aI3
aI1
aI15
aI21
aI41
aI75
aI231
aI459
aI701
aI1715
aI2581
aI4445
aI5877
aI4765
aI1037
aI15827
aI189529
asS'd'
I18982
sa(dp37964
S'a'
I93958
sS's'
I18
sS'm'
(lp37965
I1
aI3
aI5
aI13
aI17
aI23
aI41
aI133
aI143
aI297
aI1335
aI3907
aI7745
aI5139
aI9397
aI5765
aI5347
aI243091
asS'd'
I18983
sa(dp37966
S'a'
I93972
sS's'
I18
sS'm'
(lp37967
I1
aI1
aI5
aI13
aI7
aI15
aI31
aI183
aI315
aI153
aI785
aI2723
aI97
aI14361
aI10509
aI17717
aI46615
aI133289
asS'd'
I18984
sa(dp37968
S'a'
I93975
sS's'
I18
sS'm'
(lp37969
I1
aI3
aI7
aI9
aI13
aI3
aI75
aI103
aI445
aI409
aI603
aI201
aI1873
aI9277
aI23953
aI6881
aI64327
aI196771
asS'd'
I18985
sa(dp37970
S'a'
I94003
sS's'
I18
sS'm'
(lp37971
I1
aI1
aI7
aI13
aI15
aI21
aI73
aI183
aI419
aI997
aI857
aI1373
aI3855
aI417
aI10175
aI5253
aI66509
aI15731
asS'd'
I18986
sa(dp37972
S'a'
I94009
sS's'
I18
sS'm'
(lp37973
I1
aI3
aI3
aI9
aI19
aI7
aI15
aI119
aI497
aI25
aI1165
aI105
aI2605
aI15097
aI28241
aI2269
aI519
aI235655
asS'd'
I18987
sa(dp37974
S'a'
I94020
sS's'
I18
sS'm'
(lp37975
I1
aI3
aI3
aI9
aI27
aI9
aI103
aI205
aI97
aI317
aI1621
aI971
aI931
aI9099
aI24583
aI12695
aI122399
aI78021
asS'd'
I18988
sa(dp37976
S'a'
I94072
sS's'
I18
sS'm'
(lp37977
I1
aI1
aI3
aI5
aI27
aI45
aI41
aI239
aI87
aI603
aI317
aI3507
aI7677
aI9141
aI26721
aI40225
aI80515
aI205263
asS'd'
I18989
sa(dp37978
S'a'
I94081
sS's'
I18
sS'm'
(lp37979
I1
aI1
aI3
aI1
aI25
aI3
aI63
aI165
aI41
aI783
aI291
aI1997
aI3769
aI1881
aI30613
aI18821
aI86175
aI38837
asS'd'
I18990
sa(dp37980
S'a'
I94093
sS's'
I18
sS'm'
(lp37981
I1
aI1
aI3
aI5
aI17
aI19
aI95
aI17
aI357
aI587
aI689
aI3127
aI6999
aI6703
aI23923
aI55945
aI97629
aI210177
asS'd'
I18991
sa(dp37982
S'a'
I94102
sS's'
I18
sS'm'
(lp37983
I1
aI1
aI3
aI15
aI21
aI55
aI63
aI229
aI397
aI1007
aI779
aI2105
aI681
aI10659
aI26679
aI681
aI115901
aI83627
asS'd'
I18992
sa(dp37984
S'a'
I94122
sS's'
I18
sS'm'
(lp37985
I1
aI1
aI7
aI11
aI25
aI9
aI47
aI133
aI109
aI17
aI697
aI749
aI5529
aI9289
aI29675
aI2631
aI15247
aI13913
asS'd'
I18993
sa(dp37986
S'a'
I94135
sS's'
I18
sS'm'
(lp37987
I1
aI1
aI7
aI7
aI3
aI55
aI29
aI13
aI467
aI889
aI675
aI1187
aI3301
aI13721
aI13783
aI44559
aI78177
aI114219
asS'd'
I18994
sa(dp37988
S'a'
I94136
sS's'
I18
sS'm'
(lp37989
I1
aI3
aI5
aI13
aI15
aI11
aI77
aI71
aI313
aI427
aI1385
aI2007
aI4003
aI1529
aI4797
aI12289
aI24897
aI129513
asS'd'
I18995
sa(dp37990
S'a'
I94150
sS's'
I18
sS'm'
(lp37991
I1
aI1
aI3
aI11
aI9
aI47
aI103
aI253
aI345
aI659
aI1109
aI3493
aI2515
aI5669
aI30551
aI25077
aI97393
aI252689
asS'd'
I18996
sa(dp37992
S'a'
I94184
sS's'
I18
sS'm'
(lp37993
I1
aI3
aI7
aI9
aI25
aI19
aI69
aI161
aI365
aI51
aI1365
aI1045
aI4319
aI10035
aI15529
aI23251
aI44359
aI62163
asS'd'
I18997
sa(dp37994
S'a'
I94187
sS's'
I18
sS'm'
(lp37995
I1
aI3
aI1
aI7
aI3
aI25
aI119
aI33
aI19
aI561
aI659
aI2741
aI6177
aI899
aI30911
aI9627
aI83003
aI12939
asS'd'
I18998
sa(dp37996
S'a'
I94258
sS's'
I18
sS'm'
(lp37997
I1
aI3
aI7
aI1
aI13
aI37
aI19
aI161
aI427
aI621
aI1045
aI1963
aI6067
aI4439
aI32507
aI32775
aI5201
aI144645
asS'd'
I18999
sa(dp37998
S'a'
I94264
sS's'
I18
sS'm'
(lp37999
I1
aI1
aI5
aI7
aI31
aI17
aI89
aI239
aI317
aI109
aI1827
aI1395
aI1587
aI14813
aI29911
aI63545
aI22939
aI235383
asS'd'
I19000
sa(dp38000
S'a'
I94272
sS's'
I18
sS'm'
(lp38001
I1
aI1
aI7
aI15
aI1
aI17
aI41
aI123
aI405
aI539
aI1063
aI1443
aI4611
aI1847
aI24107
aI29365
aI85859
aI218601
asS'd'
I19001
sa(dp38002
S'a'
I94278
sS's'
I18
sS'm'
(lp38003
I1
aI1
aI5
aI13
aI21
aI27
aI101
aI223
aI245
aI705
aI1579
aI679
aI5461
aI8955
aI15031
aI7731
aI31219
aI165033
asS'd'
I19002
sa(dp38004
S'a'
I94281
sS's'
I18
sS'm'
(lp38005
I1
aI1
aI7
aI11
aI19
aI29
aI13
aI223
aI179
aI481
aI761
aI1543
aI3195
aI10695
aI17147
aI37577
aI130901
aI44699
asS'd'
I19003
sa(dp38006
S'a'
I94317
sS's'
I18
sS'm'
(lp38007
I1
aI1
aI7
aI3
aI19
aI53
aI49
aI1
aI393
aI583
aI1183
aI2817
aI1293
aI12949
aI15491
aI44467
aI86261
aI220439
asS'd'
I19004
sa(dp38008
S'a'
I94348
sS's'
I18
sS'm'
(lp38009
I1
aI3
aI7
aI15
aI15
aI47
aI7
aI125
aI467
aI511
aI1207
aI3787
aI5575
aI5359
aI3859
aI29933
aI104627
aI243073
asS'd'
I19005
sa(dp38010
S'a'
I94376
sS's'
I18
sS'm'
(lp38011
I1
aI1
aI1
aI13
aI27
aI25
aI17
aI243
aI477
aI457
aI1835
aI2859
aI1023
aI10107
aI26829
aI49853
aI114569
aI250471
asS'd'
I19006
sa(dp38012
S'a'
I94382
sS's'
I18
sS'm'
(lp38013
I1
aI3
aI1
aI15
aI11
aI43
aI15
aI235
aI431
aI671
aI1935
aI1143
aI4121
aI15403
aI19313
aI15919
aI111961
aI50455
asS'd'
I19007
sa(dp38014
S'a'
I94387
sS's'
I18
sS'm'
(lp38015
I1
aI3
aI3
aI3
aI11
aI45
aI107
aI143
aI353
aI671
aI1259
aI1599
aI6075
aI10645
aI9131
aI28133
aI58679
aI29883
asS'd'
I19008
sa(dp38016
S'a'
I94389
sS's'
I18
sS'm'
(lp38017
I1
aI3
aI5
aI15
aI15
aI43
aI29
aI171
aI303
aI71
aI1751
aI411
aI7615
aI12063
aI26829
aI31469
aI34335
aI3163
asS'd'
I19009
sa(dp38018
S'a'
I94393
sS's'
I18
sS'm'
(lp38019
I1
aI1
aI7
aI7
aI25
aI63
aI25
aI25
aI27
aI671
aI505
aI1235
aI1985
aI2593
aI30031
aI3251
aI94729
aI248911
asS'd'
I19010
sa(dp38020
S'a'
I94402
sS's'
I18
sS'm'
(lp38021
I1
aI1
aI3
aI1
aI19
aI15
aI125
aI133
aI133
aI209
aI1749
aI2091
aI6325
aI1275
aI5675
aI2249
aI22631
aI56293
asS'd'
I19011
sa(dp38022
S'a'
I94421
sS's'
I18
sS'm'
(lp38023
I1
aI1
aI5
aI1
aI19
aI27
aI25
aI99
aI211
aI739
aI565
aI3903
aI7701
aI7547
aI12303
aI5421
aI24663
aI22807
asS'd'
I19012
sa(dp38024
S'a'
I94422
sS's'
I18
sS'm'
(lp38025
I1
aI3
aI5
aI13
aI5
aI45
aI99
aI67
aI21
aI269
aI851
aI3333
aI4555
aI12483
aI14645
aI44757
aI99047
aI198521
asS'd'
I19013
sa(dp38026
S'a'
I94431
sS's'
I18
sS'm'
(lp38027
I1
aI1
aI5
aI13
aI19
aI1
aI123
aI87
aI109
aI799
aI591
aI2997
aI1005
aI16369
aI10329
aI34541
aI100935
aI200397
asS'd'
I19014
sa(dp38028
S'a'
I94473
sS's'
I18
sS'm'
(lp38029
I1
aI3
aI5
aI1
aI13
aI51
aI93
aI23
aI19
aI23
aI965
aI171
aI6865
aI3561
aI23255
aI44295
aI87405
aI222269
asS'd'
I19015
sa(dp38030
S'a'
I94487
sS's'
I18
sS'm'
(lp38031
I1
aI1
aI3
aI5
aI1
aI53
aI25
aI129
aI123
aI737
aI271
aI61
aI113
aI8481
aI27075
aI58633
aI21499
aI156689
asS'd'
I19016
sa(dp38032
S'a'
I94504
sS's'
I18
sS'm'
(lp38033
I1
aI3
aI3
aI11
aI3
aI43
aI11
aI123
aI243
aI1015
aI1389
aI3663
aI1725
aI6933
aI5315
aI7137
aI127705
aI56607
asS'd'
I19017
sa(dp38034
S'a'
I94510
sS's'
I18
sS'm'
(lp38035
I1
aI1
aI5
aI13
aI7
aI23
aI43
aI103
aI503
aI173
aI267
aI1509
aI3311
aI9553
aI28851
aI15771
aI28741
aI236427
asS'd'
I19018
sa(dp38036
S'a'
I94522
sS's'
I18
sS'm'
(lp38037
I1
aI1
aI5
aI15
aI27
aI43
aI119
aI3
aI13
aI107
aI317
aI3725
aI6669
aI4945
aI30485
aI10155
aI96893
aI154081
asS'd'
I19019
sa(dp38038
S'a'
I94532
sS's'
I18
sS'm'
(lp38039
I1
aI3
aI7
aI5
aI11
aI21
aI61
aI99
aI155
aI45
aI569
aI1325
aI673
aI15803
aI12047
aI55431
aI9515
aI106969
asS'd'
I19020
sa(dp38040
S'a'
I94572
sS's'
I18
sS'm'
(lp38041
I1
aI1
aI7
aI11
aI27
aI49
aI121
aI145
aI105
aI223
aI1471
aI1163
aI3889
aI4213
aI21195
aI45649
aI14663
aI82799
asS'd'
I19021
sa(dp38042
S'a'
I94589
sS's'
I18
sS'm'
(lp38043
I1
aI1
aI3
aI3
aI31
aI21
aI17
aI85
aI31
aI695
aI1591
aI2465
aI907
aI11621
aI29681
aI13131
aI77187
aI175913
asS'd'
I19022
sa(dp38044
S'a'
I94599
sS's'
I18
sS'm'
(lp38045
I1
aI3
aI5
aI5
aI21
aI49
aI77
aI229
aI359
aI825
aI1851
aI1223
aI3351
aI5349
aI30971
aI20797
aI26975
aI94425
asS'd'
I19023
sa(dp38046
S'a'
I94603
sS's'
I18
sS'm'
(lp38047
I1
aI1
aI3
aI1
aI3
aI63
aI23
aI219
aI503
aI47
aI1675
aI1641
aI5257
aI8035
aI29793
aI30093
aI44897
aI235691
asS'd'
I19024
sa(dp38048
S'a'
I94647
sS's'
I18
sS'm'
(lp38049
I1
aI1
aI7
aI9
aI27
aI37
aI109
aI33
aI511
aI203
aI1195
aI3281
aI407
aI15237
aI28485
aI21379
aI106325
aI231755
asS'd'
I19025
sa(dp38050
S'a'
I94665
sS's'
I18
sS'm'
(lp38051
I1
aI3
aI1
aI3
aI9
aI45
aI19
aI31
aI255
aI799
aI909
aI767
aI421
aI3301
aI18557
aI15043
aI48505
aI36763
asS'd'
I19026
sa(dp38052
S'a'
I94695
sS's'
I18
sS'm'
(lp38053
I1
aI3
aI7
aI13
aI1
aI45
aI59
aI233
aI319
aI265
aI517
aI1571
aI4593
aI12813
aI30729
aI19517
aI70345
aI142411
asS'd'
I19027
sa(dp38054
S'a'
I94716
sS's'
I18
sS'm'
(lp38055
I1
aI1
aI1
aI13
aI17
aI15
aI79
aI93
aI265
aI381
aI285
aI253
aI919
aI3715
aI30555
aI38801
aI30439
aI51511
asS'd'
I19028
sa(dp38056
S'a'
I94738
sS's'
I18
sS'm'
(lp38057
I1
aI1
aI7
aI15
aI25
aI39
aI71
aI57
aI145
aI487
aI1655
aI2589
aI7655
aI8413
aI24537
aI36761
aI36427
aI88929
asS'd'
I19029
sa(dp38058
S'a'
I94740
sS's'
I18
sS'm'
(lp38059
I1
aI1
aI3
aI3
aI29
aI41
aI61
aI191
aI97
aI849
aI911
aI3269
aI5425
aI13997
aI7749
aI537
aI113705
aI179765
asS'd'
I19030
sa(dp38060
S'a'
I94778
sS's'
I18
sS'm'
(lp38061
I1
aI1
aI5
aI9
aI13
aI55
aI33
aI221
aI27
aI521
aI13
aI2847
aI6035
aI8397
aI6579
aI29353
aI101953
aI88983
asS'd'
I19031
sa(dp38062
S'a'
I94803
sS's'
I18
sS'm'
(lp38063
I1
aI3
aI3
aI13
aI31
aI47
aI97
aI177
aI373
aI353
aI159
aI249
aI4741
aI7427
aI8353
aI38617
aI13857
aI122081
asS'd'
I19032
sa(dp38064
S'a'
I94810
sS's'
I18
sS'm'
(lp38065
I1
aI3
aI5
aI15
aI13
aI21
aI1
aI239
aI369
aI253
aI1009
aI1927
aI5111
aI2219
aI28167
aI32013
aI51487
aI210521
asS'd'
I19033
sa(dp38066
S'a'
I94815
sS's'
I18
sS'm'
(lp38067
I1
aI3
aI5
aI9
aI17
aI21
aI37
aI105
aI405
aI39
aI321
aI1515
aI3759
aI15469
aI13643
aI60157
aI72127
aI233505
asS'd'
I19034
sa(dp38068
S'a'
I94821
sS's'
I18
sS'm'
(lp38069
I1
aI3
aI7
aI5
aI1
aI3
aI3
aI125
aI283
aI757
aI829
aI2303
aI3715
aI6027
aI17795
aI37359
aI54721
aI5891
asS'd'
I19035
sa(dp38070
S'a'
I94862
sS's'
I18
sS'm'
(lp38071
I1
aI3
aI7
aI15
aI27
aI63
aI117
aI101
aI341
aI965
aI1543
aI51
aI3397
aI14051
aI9889
aI64647
aI111169
aI249477
asS'd'
I19036
sa(dp38072
S'a'
I94886
sS's'
I18
sS'm'
(lp38073
I1
aI1
aI5
aI13
aI5
aI29
aI51
aI61
aI233
aI685
aI751
aI163
aI2319
aI14691
aI29881
aI39029
aI57093
aI240147
asS'd'
I19037
sa(dp38074
S'a'
I94900
sS's'
I18
sS'm'
(lp38075
I1
aI3
aI5
aI3
aI9
aI21
aI107
aI147
aI263
aI471
aI621
aI3485
aI197
aI13271
aI24689
aI64341
aI110163
aI142711
asS'd'
I19038
sa(dp38076
S'a'
I94924
sS's'
I18
sS'm'
(lp38077
I1
aI3
aI1
aI7
aI1
aI23
aI17
aI31
aI131
aI631
aI795
aI3751
aI5337
aI9151
aI2873
aI31113
aI65303
aI244969
asS'd'
I19039
sa(dp38078
S'a'
I94952
sS's'
I18
sS'm'
(lp38079
I1
aI1
aI5
aI11
aI3
aI51
aI93
aI155
aI389
aI859
aI1181
aI2711
aI1375
aI6119
aI229
aI47767
aI115521
aI114129
asS'd'
I19040
sa(dp38080
S'a'
I94963
sS's'
I18
sS'm'
(lp38081
I1
aI3
aI3
aI15
aI5
aI7
aI29
aI187
aI259
aI911
aI1537
aI1885
aI6139
aI4549
aI21655
aI58771
aI1003
aI124609
asS'd'
I19041
sa(dp38082
S'a'
I94970
sS's'
I18
sS'm'
(lp38083
I1
aI1
aI3
aI15
aI25
aI45
aI97
aI217
aI331
aI305
aI1105
aI3465
aI3651
aI10171
aI31601
aI6947
aI4545
aI232627
asS'd'
I19042
sa(dp38084
S'a'
I94980
sS's'
I18
sS'm'
(lp38085
I1
aI1
aI5
aI5
aI9
aI53
aI109
aI201
aI473
aI201
aI1113
aI973
aI1825
aI13089
aI1207
aI9947
aI92515
aI216199
asS'd'
I19043
sa(dp38086
S'a'
I94992
sS's'
I18
sS'm'
(lp38087
I1
aI3
aI7
aI3
aI1
aI49
aI25
aI109
aI249
aI489
aI1663
aI3493
aI4615
aI13899
aI27851
aI60711
aI14351
aI41787
asS'd'
I19044
sa(dp38088
S'a'
I95017
sS's'
I18
sS'm'
(lp38089
I1
aI3
aI3
aI7
aI3
aI15
aI29
aI53
aI61
aI669
aI371
aI2187
aI6769
aI4623
aI25785
aI12997
aI52263
aI28387
asS'd'
I19045
sa(dp38090
S'a'
I95028
sS's'
I18
sS'm'
(lp38091
I1
aI1
aI1
aI3
aI9
aI31
aI69
aI3
aI441
aI219
aI285
aI183
aI1971
aI10903
aI8271
aI19389
aI61913
aI203537
asS'd'
I19046
sa(dp38092
S'a'
I95031
sS's'
I18
sS'm'
(lp38093
I1
aI1
aI5
aI3
aI9
aI63
aI117
aI131
aI53
aI525
aI1349
aI2701
aI1317
aI6047
aI1661
aI51785
aI93199
aI158645
asS'd'
I19047
sa(dp38094
S'a'
I95035
sS's'
I18
sS'm'
(lp38095
I1
aI3
aI5
aI3
aI21
aI61
aI11
aI91
aI317
aI635
aI61
aI1919
aI2139
aI12817
aI6587
aI63201
aI52659
aI8971
asS'd'
I19048
sa(dp38096
S'a'
I95040
sS's'
I18
sS'm'
(lp38097
I1
aI3
aI1
aI9
aI11
aI47
aI49
aI35
aI115
aI711
aI511
aI835
aI3787
aI837
aI15737
aI7467
aI53263
aI132047
asS'd'
I19049
sa(dp38098
S'a'
I95058
sS's'
I18
sS'm'
(lp38099
I1
aI1
aI5
aI3
aI27
aI47
aI121
aI211
aI65
aI363
aI1067
aI3813
aI6353
aI13701
aI23943
aI7573
aI112721
aI219587
asS'd'
I19050
sa(dp38100
S'a'
I95079
sS's'
I18
sS'm'
(lp38101
I1
aI1
aI3
aI7
aI21
aI39
aI15
aI199
aI113
aI517
aI1429
aI1399
aI6007
aI1389
aI16425
aI17709
aI1231
aI51803
asS'd'
I19051
sa(dp38102
S'a'
I95122
sS's'
I18
sS'm'
(lp38103
I1
aI3
aI5
aI11
aI5
aI37
aI35
aI97
aI215
aI281
aI517
aI1777
aI4171
aI10161
aI18369
aI23233
aI83005
aI75519
asS'd'
I19052
sa(dp38104
S'a'
I95149
sS's'
I18
sS'm'
(lp38105
I1
aI3
aI7
aI9
aI3
aI9
aI69
aI111
aI135
aI351
aI971
aI3551
aI3739
aI3571
aI22861
aI62669
aI83723
aI10707
asS'd'
I19053
sa(dp38106
S'a'
I95150
sS's'
I18
sS'm'
(lp38107
I1
aI3
aI3
aI5
aI31
aI35
aI103
aI205
aI321
aI553
aI409
aI363
aI4085
aI7735
aI5513
aI64249
aI127883
aI147839
asS'd'
I19054
sa(dp38108
S'a'
I95167
sS's'
I18
sS'm'
(lp38109
I1
aI1
aI7
aI3
aI23
aI35
aI85
aI231
aI251
aI237
aI421
aI757
aI7081
aI11247
aI24941
aI22649
aI51111
aI157383
asS'd'
I19055
sa(dp38110
S'a'
I95175
sS's'
I18
sS'm'
(lp38111
I1
aI3
aI7
aI5
aI23
aI35
aI7
aI101
aI491
aI529
aI1437
aI489
aI5057
aI12955
aI27543
aI60903
aI104151
aI42545
asS'd'
I19056
sa(dp38112
S'a'
I95205
sS's'
I18
sS'm'
(lp38113
I1
aI1
aI3
aI15
aI23
aI53
aI85
aI89
aI247
aI269
aI1555
aI3789
aI467
aI11145
aI11751
aI44343
aI120117
aI9975
asS'd'
I19057
sa(dp38114
S'a'
I95229
sS's'
I18
sS'm'
(lp38115
I1
aI1
aI5
aI3
aI29
aI49
aI123
aI179
aI311
aI45
aI1839
aI2725
aI7307
aI5525
aI32075
aI7979
aI107751
aI133677
asS'd'
I19058
sa(dp38116
S'a'
I95321
sS's'
I18
sS'm'
(lp38117
I1
aI1
aI5
aI3
aI31
aI21
aI65
aI229
aI31
aI597
aI755
aI2653
aI2699
aI2075
aI11693
aI28953
aI55811
aI13653
asS'd'
I19059
sa(dp38118
S'a'
I95345
sS's'
I18
sS'm'
(lp38119
I1
aI1
aI1
aI7
aI25
aI51
aI119
aI21
aI245
aI493
aI407
aI2997
aI4255
aI15487
aI26359
aI24153
aI42955
aI142191
asS'd'
I19060
sa(dp38120
S'a'
I95364
sS's'
I18
sS'm'
(lp38121
I1
aI1
aI5
aI3
aI27
aI61
aI13
aI209
aI13
aI401
aI399
aI2909
aI3623
aI8057
aI21301
aI32273
aI112127
aI210221
asS'd'
I19061
sa(dp38122
S'a'
I95379
sS's'
I18
sS'm'
(lp38123
I1
aI3
aI5
aI13
aI3
aI19
aI121
aI19
aI57
aI583
aI947
aI3591
aI5283
aI10831
aI20429
aI54097
aI7559
aI112465
asS'd'
I19062
sa(dp38124
S'a'
I95386
sS's'
I18
sS'm'
(lp38125
I1
aI3
aI5
aI1
aI21
aI1
aI125
aI245
aI217
aI165
aI1319
aI2119
aI4641
aI9481
aI4147
aI7079
aI119015
aI128401
asS'd'
I19063
sa(dp38126
S'a'
I95395
sS's'
I18
sS'm'
(lp38127
I1
aI1
aI5
aI3
aI3
aI31
aI25
aI63
aI17
aI191
aI497
aI819
aI1515
aI11215
aI24961
aI7679
aI125801
aI239521
asS'd'
I19064
sa(dp38128
S'a'
I95416
sS's'
I18
sS'm'
(lp38129
I1
aI1
aI5
aI1
aI3
aI25
aI27
aI43
aI37
aI863
aI739
aI2585
aI773
aI799
aI17649
aI21171
aI123541
aI164777
asS'd'
I19065
sa(dp38130
S'a'
I95419
sS's'
I18
sS'm'
(lp38131
I1
aI3
aI5
aI7
aI7
aI25
aI15
aI251
aI305
aI159
aI1941
aI3655
aI2881
aI15123
aI10911
aI35541
aI62221
aI175845
asS'd'
I19066
sa(dp38132
S'a'
I95433
sS's'
I18
sS'm'
(lp38133
I1
aI1
aI1
aI9
aI19
aI5
aI103
aI1
aI417
aI951
aI139
aI2413
aI2983
aI15471
aI9495
aI41349
aI110175
aI29501
asS'd'
I19067
sa(dp38134
S'a'
I95464
sS's'
I18
sS'm'
(lp38135
I1
aI1
aI1
aI5
aI29
aI53
aI95
aI173
aI211
aI803
aI1599
aI4093
aI5559
aI15855
aI12271
aI12583
aI102221
aI203453
asS'd'
I19068
sa(dp38136
S'a'
I95490
sS's'
I18
sS'm'
(lp38137
I1
aI3
aI5
aI5
aI19
aI43
aI31
aI175
aI493
aI289
aI1865
aI2925
aI3833
aI11327
aI23337
aI62669
aI99485
aI230583
asS'd'
I19069
sa(dp38138
S'a'
I95496
sS's'
I18
sS'm'
(lp38139
I1
aI3
aI3
aI11
aI11
aI25
aI95
aI215
aI501
aI421
aI725
aI1571
aI2133
aI2761
aI8649
aI45359
aI88851
aI55057
asS'd'
I19070
sa(dp38140
S'a'
I95504
sS's'
I18
sS'm'
(lp38141
I1
aI1
aI7
aI7
aI21
aI45
aI69
aI63
aI399
aI929
aI1431
aI3397
aI3613
aI14595
aI10417
aI62913
aI106283
aI120869
asS'd'
I19071
sa(dp38142
S'a'
I95513
sS's'
I18
sS'm'
(lp38143
I1
aI3
aI7
aI15
aI13
aI45
aI11
aI177
aI125
aI611
aI1115
aI2441
aI2689
aI12517
aI8989
aI34991
aI23789
aI51543
asS'd'
I19072
sa(dp38144
S'a'
I95523
sS's'
I18
sS'm'
(lp38145
I1
aI1
aI3
aI1
aI3
aI15
aI5
aI125
aI511
aI137
aI1919
aI2953
aI5267
aI3543
aI5485
aI7463
aI130407
aI255945
asS'd'
I19073
sa(dp38146
S'a'
I95525
sS's'
I18
sS'm'
(lp38147
I1
aI1
aI3
aI7
aI7
aI21
aI95
aI97
aI51
aI91
aI813
aI2819
aI2839
aI12041
aI26197
aI20143
aI51403
aI171337
asS'd'
I19074
sa(dp38148
S'a'
I95558
sS's'
I18
sS'm'
(lp38149
I1
aI1
aI1
aI1
aI7
aI27
aI15
aI125
aI441
aI387
aI1869
aI2157
aI5863
aI581
aI893
aI58827
aI104063
aI93735
asS'd'
I19075
sa(dp38150
S'a'
I95572
sS's'
I18
sS'm'
(lp38151
I1
aI3
aI3
aI7
aI27
aI9
aI79
aI97
aI465
aI207
aI931
aI2809
aI2225
aI13749
aI18819
aI30605
aI9829
aI130743
asS'd'
I19076
sa(dp38152
S'a'
I95591
sS's'
I18
sS'm'
(lp38153
I1
aI3
aI5
aI13
aI31
aI41
aI19
aI147
aI293
aI725
aI297
aI397
aI1343
aI12669
aI15339
aI58599
aI12113
aI149835
asS'd'
I19077
sa(dp38154
S'a'
I95609
sS's'
I18
sS'm'
(lp38155
I1
aI3
aI3
aI13
aI27
aI13
aI121
aI253
aI349
aI229
aI915
aI1673
aI3819
aI77
aI20691
aI53823
aI78265
aI138743
asS'd'
I19078
sa(dp38156
S'a'
I95619
sS's'
I18
sS'm'
(lp38157
I1
aI1
aI5
aI5
aI29
aI41
aI65
aI235
aI123
aI871
aI1809
aI3013
aI3531
aI1551
aI8441
aI23481
aI58729
aI117639
asS'd'
I19079
sa(dp38158
S'a'
I95625
sS's'
I18
sS'm'
(lp38159
I1
aI1
aI7
aI5
aI23
aI55
aI89
aI81
aI201
aI313
aI1307
aI2427
aI2025
aI8543
aI26631
aI58655
aI122095
aI247579
asS'd'
I19080
sa(dp38160
S'a'
I95633
sS's'
I18
sS'm'
(lp38161
I1
aI3
aI1
aI5
aI3
aI63
aI89
aI219
aI449
aI9
aI1771
aI2915
aI5925
aI13773
aI26119
aI61309
aI65107
aI33001
asS'd'
I19081
sa(dp38162
S'a'
I95649
sS's'
I18
sS'm'
(lp38163
I1
aI3
aI7
aI1
aI27
aI11
aI25
aI221
aI139
aI665
aI1543
aI2157
aI7617
aI9135
aI567
aI64985
aI88749
aI54223
asS'd'
I19082
sa(dp38164
S'a'
I95652
sS's'
I18
sS'm'
(lp38165
I1
aI1
aI3
aI9
aI13
aI41
aI7
aI99
aI483
aI115
aI1499
aI3343
aI7207
aI1805
aI16031
aI63707
aI8555
aI90959
asS'd'
I19083
sa(dp38166
S'a'
I95655
sS's'
I18
sS'm'
(lp38167
I1
aI3
aI1
aI9
aI15
aI53
aI41
aI239
aI295
aI47
aI1645
aI1095
aI5163
aI7739
aI26635
aI28245
aI9315
aI100629
asS'd'
I19084
sa(dp38168
S'a'
I95696
sS's'
I18
sS'm'
(lp38169
I1
aI3
aI1
aI5
aI1
aI19
aI69
aI5
aI171
aI669
aI673
aI633
aI6895
aI7571
aI11539
aI25133
aI99235
aI7991
asS'd'
I19085
sa(dp38170
S'a'
I95721
sS's'
I18
sS'm'
(lp38171
I1
aI3
aI5
aI11
aI21
aI37
aI63
aI77
aI281
aI307
aI1711
aI2671
aI1315
aI14683
aI28757
aI22751
aI56477
aI190805
asS'd'
I19086
sa(dp38172
S'a'
I95766
sS's'
I18
sS'm'
(lp38173
I1
aI3
aI3
aI5
aI15
aI1
aI5
aI21
aI199
aI161
aI655
aI1263
aI3315
aI16051
aI2409
aI773
aI9075
aI121265
asS'd'
I19087
sa(dp38174
S'a'
I95772
sS's'
I18
sS'm'
(lp38175
I1
aI3
aI3
aI3
aI7
aI23
aI71
aI195
aI11
aI263
aI1845
aI165
aI3489
aI447
aI11315
aI23861
aI110949
aI78909
asS'd'
I19088
sa(dp38176
S'a'
I95775
sS's'
I18
sS'm'
(lp38177
I1
aI1
aI7
aI5
aI1
aI53
aI37
aI9
aI439
aI135
aI909
aI457
aI6993
aI11401
aI14065
aI30795
aI56149
aI168013
asS'd'
I19089
sa(dp38178
S'a'
I95794
sS's'
I18
sS'm'
(lp38179
I1
aI1
aI1
aI15
aI23
aI37
aI13
aI87
aI113
aI251
aI233
aI725
aI7757
aI14399
aI3023
aI54277
aI87879
aI54053
asS'd'
I19090
sa(dp38180
S'a'
I95796
sS's'
I18
sS'm'
(lp38181
I1
aI3
aI5
aI11
aI11
aI57
aI109
aI171
aI171
aI17
aI343
aI2749
aI6525
aI9735
aI11715
aI23783
aI54439
aI82819
asS'd'
I19091
sa(dp38182
S'a'
I95818
sS's'
I18
sS'm'
(lp38183
I1
aI1
aI1
aI15
aI3
aI47
aI73
aI237
aI399
aI301
aI947
aI2055
aI1909
aI14105
aI26893
aI47805
aI25
aI172957
asS'd'
I19092
sa(dp38184
S'a'
I95820
sS's'
I18
sS'm'
(lp38185
I1
aI1
aI7
aI7
aI11
aI27
aI93
aI167
aI117
aI637
aI351
aI319
aI4605
aI12897
aI31001
aI39655
aI53551
aI246113
asS'd'
I19093
sa(dp38186
S'a'
I95832
sS's'
I18
sS'm'
(lp38187
I1
aI3
aI5
aI15
aI3
aI37
aI25
aI9
aI421
aI519
aI257
aI3251
aI1649
aI4069
aI999
aI59367
aI112383
aI32095
asS'd'
I19094
sa(dp38188
S'a'
I95842
sS's'
I18
sS'm'
(lp38189
I1
aI3
aI7
aI7
aI25
aI57
aI11
aI37
aI271
aI545
aI1213
aI1927
aI6471
aI5145
aI22995
aI51051
aI126981
aI260457
asS'd'
I19095
sa(dp38190
S'a'
I95851
sS's'
I18
sS'm'
(lp38191
I1
aI3
aI5
aI11
aI1
aI61
aI77
aI201
aI395
aI199
aI477
aI103
aI4069
aI7003
aI26371
aI49145
aI103839
aI195661
asS'd'
I19096
sa(dp38192
S'a'
I95854
sS's'
I18
sS'm'
(lp38193
I1
aI3
aI3
aI9
aI13
aI41
aI25
aI125
aI161
aI371
aI179
aI351
aI7169
aI7179
aI21627
aI57793
aI104679
aI158583
asS'd'
I19097
sa(dp38194
S'a'
I95859
sS's'
I18
sS'm'
(lp38195
I1
aI3
aI7
aI11
aI5
aI7
aI111
aI163
aI201
aI783
aI189
aI273
aI2751
aI13917
aI28501
aI18261
aI12755
aI15521
asS'd'
I19098
sa(dp38196
S'a'
I95868
sS's'
I18
sS'm'
(lp38197
I1
aI1
aI5
aI7
aI3
aI37
aI121
aI209
aI503
aI299
aI1301
aI3703
aI2321
aI99
aI14953
aI28087
aI85059
aI256911
asS'd'
I19099
sa(dp38198
S'a'
I95881
sS's'
I18
sS'm'
(lp38199
I1
aI3
aI3
aI13
aI3
aI29
aI95
aI249
aI383
aI971
aI1291
aI13
aI1587
aI3447
aI26477
aI15837
aI111141
aI73899
asS'd'
I19100
sa(dp38200
S'a'
I95895
sS's'
I18
sS'm'
(lp38201
I1
aI1
aI7
aI1
aI17
aI57
aI31
aI1
aI219
aI329
aI19
aI3841
aI1829
aI5179
aI14945
aI6625
aI3783
aI200583
asS'd'
I19101
sa(dp38202
S'a'
I95943
sS's'
I18
sS'm'
(lp38203
I1
aI3
aI1
aI3
aI1
aI31
aI23
aI17
aI209
aI383
aI297
aI3065
aI4323
aI7847
aI30189
aI56541
aI57535
aI24853
asS'd'
I19102
sa(dp38204
S'a'
I95949
sS's'
I18
sS'm'
(lp38205
I1
aI1
aI3
aI11
aI31
aI35
aI125
aI141
aI251
aI79
aI161
aI775
aI2455
aI6959
aI26433
aI39145
aI26563
aI665
asS'd'
I19103
sa(dp38206
S'a'
I95978
sS's'
I18
sS'm'
(lp38207
I1
aI1
aI7
aI1
aI11
aI9
aI9
aI211
aI231
aI723
aI1337
aI1713
aI3779
aI2001
aI23451
aI27107
aI64297
aI254943
asS'd'
I19104
sa(dp38208
S'a'
I95980
sS's'
I18
sS'm'
(lp38209
I1
aI3
aI7
aI15
aI21
aI55
aI19
aI159
aI449
aI837
aI1259
aI1851
aI5061
aI355
aI21531
aI63479
aI114657
aI139265
asS'd'
I19105
sa(dp38210
S'a'
I96010
sS's'
I18
sS'm'
(lp38211
I1
aI1
aI1
aI3
aI11
aI55
aI103
aI179
aI363
aI567
aI421
aI981
aI7221
aI2077
aI19339
aI1155
aI67019
aI218231
asS'd'
I19106
sa(dp38212
S'a'
I96020
sS's'
I18
sS'm'
(lp38213
I1
aI1
aI7
aI11
aI3
aI43
aI55
aI161
aI347
aI995
aI1555
aI3251
aI1605
aI13313
aI4499
aI19361
aI60145
aI71593
asS'd'
I19107
sa(dp38214
S'a'
I96024
sS's'
I18
sS'm'
(lp38215
I1
aI1
aI5
aI3
aI9
aI15
aI119
aI213
aI455
aI241
aI857
aI683
aI1247
aI13085
aI23919
aI20365
aI16303
aI73263
asS'd'
I19108
sa(dp38216
S'a'
I96063
sS's'
I18
sS'm'
(lp38217
I1
aI1
aI3
aI13
aI25
aI17
aI45
aI193
aI375
aI289
aI1381
aI3629
aI3015
aI15883
aI20633
aI7431
aI108787
aI233297
asS'd'
I19109
sa(dp38218
S'a'
I96108
sS's'
I18
sS'm'
(lp38219
I1
aI1
aI1
aI15
aI21
aI57
aI105
aI91
aI233
aI961
aI1623
aI3849
aI711
aI3857
aI32657
aI5935
aI85113
aI38287
asS'd'
I19110
sa(dp38220
S'a'
I96125
sS's'
I18
sS'm'
(lp38221
I1
aI3
aI3
aI3
aI15
aI31
aI97
aI217
aI335
aI385
aI1661
aI3927
aI6849
aI137
aI28871
aI56485
aI32777
aI260033
asS'd'
I19111
sa(dp38222
S'a'
I96154
sS's'
I18
sS'm'
(lp38223
I1
aI1
aI3
aI13
aI5
aI61
aI19
aI255
aI123
aI481
aI1865
aI1815
aI3047
aI173
aI25363
aI1277
aI6453
aI174405
asS'd'
I19112
sa(dp38224
S'a'
I96172
sS's'
I18
sS'm'
(lp38225
I1
aI3
aI7
aI13
aI27
aI9
aI19
aI21
aI433
aI857
aI1931
aI2927
aI629
aI7733
aI13503
aI48263
aI67517
aI26495
asS'd'
I19113
sa(dp38226
S'a'
I96189
sS's'
I18
sS'm'
(lp38227
I1
aI1
aI1
aI3
aI5
aI43
aI61
aI239
aI81
aI585
aI187
aI1123
aI3319
aI8699
aI20925
aI40815
aI76575
aI169383
asS'd'
I19114
sa(dp38228
S'a'
I96204
sS's'
I18
sS'm'
(lp38229
I1
aI3
aI3
aI3
aI9
aI49
aI71
aI225
aI95
aI365
aI645
aI237
aI7829
aI5727
aI17031
aI58971
aI71415
aI232423
asS'd'
I19115
sa(dp38230
S'a'
I96219
sS's'
I18
sS'm'
(lp38231
I1
aI3
aI5
aI9
aI25
aI49
aI113
aI47
aI105
aI609
aI1557
aI2099
aI2129
aI8663
aI24811
aI25505
aI38153
aI185821
asS'd'
I19116
sa(dp38232
S'a'
I96256
sS's'
I18
sS'm'
(lp38233
I1
aI3
aI5
aI13
aI23
aI55
aI107
aI17
aI309
aI807
aI635
aI1007
aI6207
aI3363
aI7607
aI25013
aI4141
aI171509
asS'd'
I19117
sa(dp38234
S'a'
I96261
sS's'
I18
sS'm'
(lp38235
I1
aI1
aI1
aI13
aI27
aI35
aI31
aI89
aI109
aI879
aI1845
aI3999
aI5415
aI8777
aI9605
aI29703
aI28149
aI36469
asS'd'
I19118
sa(dp38236
S'a'
I96289
sS's'
I18
sS'm'
(lp38237
I1
aI3
aI1
aI3
aI13
aI3
aI51
aI31
aI479
aI549
aI1245
aI2033
aI961
aI13893
aI21829
aI32791
aI109497
aI187425
asS'd'
I19119
sa(dp38238
S'a'
I96292
sS's'
I18
sS'm'
(lp38239
I1
aI1
aI5
aI3
aI19
aI5
aI25
aI187
aI173
aI869
aI201
aI3851
aI7369
aI6229
aI16577
aI45623
aI19859
aI209855
asS'd'
I19120
sa(dp38240
S'a'
I96296
sS's'
I18
sS'm'
(lp38241
I1
aI1
aI7
aI9
aI1
aI9
aI53
aI47
aI289
aI557
aI999
aI141
aI3789
aI3087
aI30217
aI24221
aI81431
aI157507
asS'd'
I19121
sa(dp38242
S'a'
I96324
sS's'
I18
sS'm'
(lp38243
I1
aI1
aI3
aI9
aI1
aI25
aI11
aI73
aI155
aI155
aI621
aI4047
aI6759
aI5641
aI28147
aI8523
aI69439
aI92613
asS'd'
I19122
sa(dp38244
S'a'
I96345
sS's'
I18
sS'm'
(lp38245
I1
aI3
aI1
aI5
aI25
aI23
aI41
aI79
aI71
aI793
aI1381
aI307
aI7863
aI16289
aI28783
aI5299
aI128481
aI222799
asS'd'
I19123
sa(dp38246
S'a'
I96346
sS's'
I18
sS'm'
(lp38247
I1
aI1
aI7
aI1
aI17
aI33
aI117
aI111
aI15
aI249
aI1397
aI1349
aI4883
aI6009
aI3179
aI33509
aI56355
aI31937
asS'd'
I19124
sa(dp38248
S'a'
I96358
sS's'
I18
sS'm'
(lp38249
I1
aI3
aI5
aI13
aI29
aI15
aI41
aI185
aI91
aI501
aI571
aI2889
aI6901
aI3875
aI3737
aI23657
aI101587
aI261181
asS'd'
I19125
sa(dp38250
S'a'
I96386
sS's'
I18
sS'm'
(lp38251
I1
aI1
aI7
aI9
aI21
aI49
aI33
aI143
aI19
aI203
aI75
aI1353
aI585
aI7719
aI11311
aI48989
aI10803
aI51743
asS'd'
I19126
sa(dp38252
S'a'
I96412
sS's'
I18
sS'm'
(lp38253
I1
aI1
aI7
aI13
aI23
aI31
aI103
aI209
aI375
aI817
aI1461
aI3657
aI7931
aI15893
aI15065
aI28721
aI54299
aI71147
asS'd'
I19127
sa(dp38254
S'a'
I96428
sS's'
I18
sS'm'
(lp38255
I1
aI3
aI7
aI7
aI7
aI25
aI37
aI173
aI355
aI499
aI247
aI459
aI7701
aI2219
aI11703
aI20631
aI128857
aI125367
asS'd'
I19128
sa(dp38256
S'a'
I96445
sS's'
I18
sS'm'
(lp38257
I1
aI1
aI3
aI5
aI25
aI61
aI43
aI135
aI451
aI667
aI547
aI443
aI5071
aI12671
aI26975
aI20131
aI101545
aI115281
asS'd'
I19129
sa(dp38258
S'a'
I96446
sS's'
I18
sS'm'
(lp38259
I1
aI3
aI3
aI5
aI9
aI19
aI75
aI133
aI211
aI585
aI1283
aI3397
aI3181
aI65
aI20213
aI47725
aI101883
aI194749
asS'd'
I19130
sa(dp38260
S'a'
I96448
sS's'
I18
sS'm'
(lp38261
I1
aI1
aI1
aI1
aI19
aI13
aI75
aI135
aI111
aI641
aI765
aI1631
aI4711
aI241
aI15125
aI38233
aI95535
aI177965
asS'd'
I19131
sa(dp38262
S'a'
I96458
sS's'
I18
sS'm'
(lp38263
I1
aI1
aI7
aI13
aI31
aI1
aI91
aI61
aI299
aI35
aI1327
aI3903
aI6193
aI5589
aI6331
aI6321
aI105741
aI89639
asS'd'
I19132
sa(dp38264
S'a'
I96475
sS's'
I18
sS'm'
(lp38265
I1
aI3
aI3
aI15
aI1
aI55
aI11
aI39
aI171
aI713
aI973
aI1827
aI3487
aI13057
aI30775
aI16881
aI124989
aI208193
asS'd'
I19133
sa(dp38266
S'a'
I96482
sS's'
I18
sS'm'
(lp38267
I1
aI3
aI7
aI1
aI21
aI29
aI19
aI75
aI397
aI755
aI1601
aI2907
aI6861
aI10377
aI23127
aI2443
aI86545
aI3841
asS'd'
I19134
sa(dp38268
S'a'
I96494
sS's'
I18
sS'm'
(lp38269
I1
aI3
aI1
aI11
aI25
aI33
aI53
aI195
aI343
aI425
aI1523
aI3051
aI3115
aI3205
aI3457
aI20521
aI39187
aI33307
asS'd'
I19135
sa(dp38270
S'a'
I96502
sS's'
I18
sS'm'
(lp38271
I1
aI3
aI5
aI1
aI25
aI23
aI47
aI5
aI133
aI511
aI1549
aI2691
aI7861
aI4987
aI2877
aI38693
aI37491
aI22481
asS'd'
I19136
sa(dp38272
S'a'
I96508
sS's'
I18
sS'm'
(lp38273
I1
aI1
aI5
aI15
aI5
aI55
aI125
aI231
aI11
aI451
aI1443
aI3865
aI4115
aI2379
aI13675
aI29953
aI85721
aI114859
asS'd'
I19137
sa(dp38274
S'a'
I96511
sS's'
I18
sS'm'
(lp38275
I1
aI3
aI1
aI15
aI19
aI37
aI29
aI75
aI483
aI785
aI1933
aI2435
aI1811
aI2787
aI32653
aI23159
aI80993
aI26867
asS'd'
I19138
sa(dp38276
S'a'
I96516
sS's'
I18
sS'm'
(lp38277
I1
aI1
aI1
aI15
aI7
aI27
aI53
aI99
aI11
aI693
aI1085
aI743
aI939
aI6461
aI6391
aI45913
aI94037
aI217039
asS'd'
I19139
sa(dp38278
S'a'
I96520
sS's'
I18
sS'm'
(lp38279
I1
aI3
aI3
aI9
aI19
aI37
aI93
aI77
aI363
aI125
aI1675
aI347
aI5599
aI7771
aI23549
aI39945
aI106931
aI127959
asS'd'
I19140
sa(dp38280
S'a'
I96547
sS's'
I18
sS'm'
(lp38281
I1
aI3
aI1
aI5
aI27
aI47
aI107
aI85
aI31
aI621
aI1529
aI2349
aI7055
aI889
aI4663
aI1705
aI40011
aI214775
asS'd'
I19141
sa(dp38282
S'a'
I96556
sS's'
I18
sS'm'
(lp38283
I1
aI3
aI1
aI5
aI11
aI47
aI35
aI13
aI139
aI783
aI1009
aI845
aI4139
aI14713
aI24191
aI17597
aI124923
aI219657
asS'd'
I19142
sa(dp38284
S'a'
I96561
sS's'
I18
sS'm'
(lp38285
I1
aI1
aI7
aI3
aI3
aI25
aI63
aI207
aI361
aI587
aI763
aI3027
aI6523
aI6783
aI11203
aI57313
aI115397
aI149921
asS'd'
I19143
sa(dp38286
S'a'
I96568
sS's'
I18
sS'm'
(lp38287
I1
aI1
aI7
aI1
aI21
aI55
aI109
aI183
aI487
aI869
aI195
aI83
aI3675
aI13103
aI12383
aI63519
aI48379
aI256443
asS'd'
I19144
sa(dp38288
S'a'
I96581
sS's'
I18
sS'm'
(lp38289
I1
aI3
aI7
aI13
aI9
aI21
aI29
aI163
aI105
aI871
aI747
aI2459
aI7383
aI439
aI5223
aI1655
aI1469
aI50345
asS'd'
I19145
sa(dp38290
S'a'
I96582
sS's'
I18
sS'm'
(lp38291
I1
aI3
aI1
aI1
aI15
aI63
aI37
aI159
aI385
aI795
aI1369
aI1973
aI6119
aI6027
aI23913
aI52475
aI80827
aI198261
asS'd'
I19146
sa(dp38292
S'a'
I96679
sS's'
I18
sS'm'
(lp38293
I1
aI3
aI3
aI11
aI15
aI5
aI121
aI231
aI43
aI907
aI1621
aI3895
aI5075
aI10865
aI3123
aI49657
aI69827
aI215813
asS'd'
I19147
sa(dp38294
S'a'
I96683
sS's'
I18
sS'm'
(lp38295
I1
aI3
aI1
aI15
aI7
aI41
aI75
aI105
aI87
aI899
aI629
aI1699
aI5861
aI9279
aI30107
aI37443
aI7555
aI64461
asS'd'
I19148
sa(dp38296
S'a'
I96717
sS's'
I18
sS'm'
(lp38297
I1
aI3
aI1
aI7
aI9
aI15
aI119
aI127
aI121
aI621
aI1117
aI1659
aI605
aI13705
aI31181
aI40063
aI17257
aI77645
asS'd'
I19149
sa(dp38298
S'a'
I96754
sS's'
I18
sS'm'
(lp38299
I1
aI1
aI5
aI5
aI3
aI37
aI95
aI237
aI379
aI375
aI903
aI257
aI4425
aI14191
aI9185
aI57133
aI82067
aI73521
asS'd'
I19150
sa(dp38300
S'a'
I96769
sS's'
I18
sS'm'
(lp38301
I1
aI1
aI7
aI15
aI1
aI43
aI63
aI45
aI121
aI669
aI1775
aI179
aI7385
aI3557
aI17261
aI379
aI24759
aI214831
asS'd'
I19151
sa(dp38302
S'a'
I96799
sS's'
I18
sS'm'
(lp38303
I1
aI1
aI3
aI9
aI31
aI5
aI43
aI153
aI451
aI573
aI1623
aI2831
aI4483
aI7219
aI27657
aI47111
aI58165
aI145799
asS'd'
I19152
sa(dp38304
S'a'
I96805
sS's'
I18
sS'm'
(lp38305
I1
aI1
aI3
aI11
aI3
aI11
aI111
aI83
aI329
aI807
aI779
aI1223
aI6095
aI7269
aI22425
aI19343
aI11937
aI10173
asS'd'
I19153
sa(dp38306
S'a'
I96809
sS's'
I18
sS'm'
(lp38307
I1
aI1
aI1
aI13
aI27
aI15
aI7
aI111
aI37
aI663
aI51
aI3759
aI6321
aI8253
aI737
aI59501
aI109595
aI177827
asS'd'
I19154
sa(dp38308
S'a'
I96823
sS's'
I18
sS'm'
(lp38309
I1
aI1
aI3
aI3
aI29
aI39
aI79
aI115
aI307
aI765
aI331
aI377
aI1873
aI14491
aI11065
aI11865
aI76717
aI29101
asS'd'
I19155
sa(dp38310
S'a'
I96829
sS's'
I18
sS'm'
(lp38311
I1
aI3
aI7
aI3
aI21
aI45
aI97
aI213
aI309
aI3
aI483
aI3933
aI1043
aI8519
aI22517
aI34675
aI78819
aI172479
asS'd'
I19156
sa(dp38312
S'a'
I96832
sS's'
I18
sS'm'
(lp38313
I1
aI3
aI5
aI3
aI31
aI51
aI27
aI137
aI405
aI427
aI815
aI43
aI6551
aI10971
aI28589
aI53077
aI36639
aI167661
asS'd'
I19157
sa(dp38314
S'a'
I96850
sS's'
I18
sS'm'
(lp38315
I1
aI3
aI1
aI3
aI29
aI5
aI111
aI19
aI343
aI21
aI557
aI4067
aI1525
aI12793
aI11513
aI48869
aI78035
aI171531
asS'd'
I19158
sa(dp38316
S'a'
I96856
sS's'
I18
sS'm'
(lp38317
I1
aI1
aI5
aI7
aI25
aI47
aI53
aI245
aI135
aI137
aI1697
aI2057
aI3147
aI15903
aI26979
aI2157
aI43967
aI207661
asS'd'
I19159
sa(dp38318
S'a'
I96906
sS's'
I18
sS'm'
(lp38319
I1
aI1
aI5
aI3
aI25
aI11
aI15
aI59
aI511
aI307
aI757
aI3275
aI1299
aI10373
aI11943
aI54169
aI32417
aI21645
asS'd'
I19160
sa(dp38320
S'a'
I96962
sS's'
I18
sS'm'
(lp38321
I1
aI3
aI3
aI11
aI15
aI15
aI5
aI137
aI237
aI741
aI1613
aI3565
aI7359
aI6181
aI25953
aI18137
aI59759
aI186693
asS'd'
I19161
sa(dp38322
S'a'
I96974
sS's'
I18
sS'm'
(lp38323
I1
aI3
aI5
aI3
aI19
aI13
aI99
aI167
aI45
aI71
aI1683
aI3635
aI7603
aI14879
aI23903
aI14795
aI58395
aI11853
asS'd'
I19162
sa(dp38324
S'a'
I96979
sS's'
I18
sS'm'
(lp38325
I1
aI3
aI1
aI7
aI15
aI45
aI111
aI111
aI175
aI567
aI1031
aI2255
aI3895
aI11861
aI20195
aI15461
aI88411
aI225713
asS'd'
I19163
sa(dp38326
S'a'
I96997
sS's'
I18
sS'm'
(lp38327
I1
aI1
aI7
aI3
aI5
aI5
aI85
aI65
aI231
aI643
aI1591
aI219
aI2929
aI4845
aI29327
aI14769
aI46629
aI131367
asS'd'
I19164
sa(dp38328
S'a'
I96998
sS's'
I18
sS'm'
(lp38329
I1
aI1
aI5
aI9
aI29
aI21
aI47
aI87
aI113
aI469
aI1647
aI2461
aI3663
aI5865
aI6647
aI41345
aI39539
aI220301
asS'd'
I19165
sa(dp38330
S'a'
I97002
sS's'
I18
sS'm'
(lp38331
I1
aI1
aI5
aI11
aI9
aI55
aI5
aI147
aI141
aI181
aI283
aI1695
aI6537
aI11095
aI10385
aI36013
aI111653
aI182273
asS'd'
I19166
sa(dp38332
S'a'
I97054
sS's'
I18
sS'm'
(lp38333
I1
aI1
aI3
aI5
aI17
aI45
aI103
aI253
aI407
aI151
aI1585
aI1585
aI6661
aI14579
aI5723
aI37641
aI56813
aI258819
asS'd'
I19167
sa(dp38334
S'a'
I97064
sS's'
I18
sS'm'
(lp38335
I1
aI3
aI3
aI3
aI5
aI63
aI85
aI201
aI87
aI419
aI1993
aI737
aI5859
aI6049
aI17393
aI9453
aI65915
aI1731
asS'd'
I19168
sa(dp38336
S'a'
I97067
sS's'
I18
sS'm'
(lp38337
I1
aI1
aI3
aI3
aI3
aI27
aI97
aI135
aI137
aI731
aI1559
aI3409
aI5973
aI15981
aI19833
aI8419
aI33273
aI44155
asS'd'
I19169
sa(dp38338
S'a'
I97110
sS's'
I18
sS'm'
(lp38339
I1
aI3
aI3
aI9
aI31
aI55
aI109
aI191
aI119
aI59
aI645
aI1047
aI7767
aI8379
aI13781
aI52289
aI31605
aI186667
asS'd'
I19170
sa(dp38340
S'a'
I97116
sS's'
I18
sS'm'
(lp38341
I1
aI3
aI1
aI15
aI9
aI1
aI23
aI31
aI23
aI311
aI1879
aI1939
aI5509
aI14573
aI10501
aI38867
aI39131
aI231151
asS'd'
I19171
sa(dp38342
S'a'
I97137
sS's'
I18
sS'm'
(lp38343
I1
aI3
aI7
aI1
aI31
aI33
aI33
aI19
aI475
aI723
aI795
aI1793
aI6639
aI14349
aI16639
aI31473
aI110411
aI95703
asS'd'
I19172
sa(dp38344
S'a'
I97138
sS's'
I18
sS'm'
(lp38345
I1
aI1
aI5
aI9
aI11
aI3
aI39
aI119
aI455
aI839
aI513
aI2423
aI2219
aI6059
aI6125
aI60995
aI117701
aI204057
asS'd'
I19173
sa(dp38346
S'a'
I97143
sS's'
I18
sS'm'
(lp38347
I1
aI1
aI1
aI9
aI5
aI23
aI87
aI33
aI59
aI241
aI1427
aI3867
aI1091
aI14683
aI21651
aI7091
aI38011
aI63809
asS'd'
I19174
sa(dp38348
S'a'
I97183
sS's'
I18
sS'm'
(lp38349
I1
aI1
aI7
aI15
aI15
aI23
aI75
aI227
aI415
aI1015
aI2033
aI1311
aI6659
aI5093
aI14799
aI65331
aI96989
aI170395
asS'd'
I19175
sa(dp38350
S'a'
I97187
sS's'
I18
sS'm'
(lp38351
I1
aI3
aI5
aI15
aI25
aI61
aI33
aI179
aI503
aI875
aI1853
aI257
aI6727
aI9117
aI16777
aI29585
aI110901
aI231617
asS'd'
I19176
sa(dp38352
S'a'
I97190
sS's'
I18
sS'm'
(lp38353
I1
aI1
aI1
aI15
aI13
aI53
aI73
aI151
aI315
aI887
aI669
aI3959
aI5279
aI1461
aI15497
aI40107
aI9595
aI252059
asS'd'
I19177
sa(dp38354
S'a'
I97202
sS's'
I18
sS'm'
(lp38355
I1
aI3
aI7
aI13
aI17
aI45
aI43
aI61
aI99
aI555
aI981
aI3255
aI6385
aI8723
aI24451
aI45243
aI68617
aI171911
asS'd'
I19178
sa(dp38356
S'a'
I97219
sS's'
I18
sS'm'
(lp38357
I1
aI3
aI3
aI11
aI1
aI29
aI97
aI219
aI341
aI597
aI503
aI773
aI3777
aI5431
aI4581
aI37169
aI57269
aI186377
asS'd'
I19179
sa(dp38358
S'a'
I97239
sS's'
I18
sS'm'
(lp38359
I1
aI3
aI1
aI11
aI15
aI49
aI119
aI189
aI279
aI821
aI1541
aI1343
aI4379
aI5833
aI26537
aI29769
aI121125
aI202553
asS'd'
I19180
sa(dp38360
S'a'
I97245
sS's'
I18
sS'm'
(lp38361
I1
aI3
aI5
aI9
aI19
aI23
aI5
aI197
aI323
aI101
aI1155
aI7
aI5933
aI3111
aI19595
aI36807
aI40147
aI153
asS'd'
I19181
sa(dp38362
S'a'
I97246
sS's'
I18
sS'm'
(lp38363
I1
aI1
aI5
aI11
aI17
aI9
aI83
aI51
aI185
aI415
aI367
aI1431
aI7803
aI8253
aI16283
aI54545
aI99733
aI57777
asS'd'
I19182
sa(dp38364
S'a'
I97249
sS's'
I18
sS'm'
(lp38365
I1
aI1
aI5
aI7
aI5
aI31
aI41
aI13
aI33
aI531
aI1381
aI781
aI1699
aI6321
aI18125
aI34567
aI113253
aI104181
asS'd'
I19183
sa(dp38366
S'a'
I97264
sS's'
I18
sS'm'
(lp38367
I1
aI3
aI1
aI5
aI1
aI59
aI37
aI239
aI343
aI395
aI121
aI2181
aI2485
aI13825
aI19127
aI22689
aI103023
aI198213
asS'd'
I19184
sa(dp38368
S'a'
I97267
sS's'
I18
sS'm'
(lp38369
I1
aI3
aI1
aI15
aI29
aI17
aI11
aI27
aI413
aI273
aI1805
aI2845
aI8147
aI10301
aI5423
aI29859
aI85243
aI190379
asS'd'
I19185
sa(dp38370
S'a'
I97269
sS's'
I18
sS'm'
(lp38371
I1
aI3
aI1
aI15
aI7
aI61
aI29
aI135
aI273
aI951
aI725
aI1345
aI4231
aI13651
aI31291
aI6081
aI85735
aI96023
asS'd'
I19186
sa(dp38372
S'a'
I97274
sS's'
I18
sS'm'
(lp38373
I1
aI1
aI3
aI11
aI15
aI29
aI81
aI129
aI245
aI295
aI527
aI3905
aI4323
aI5447
aI21253
aI51177
aI105105
aI48323
asS'd'
I19187
sa(dp38374
S'a'
I97282
sS's'
I18
sS'm'
(lp38375
I1
aI1
aI3
aI13
aI13
aI45
aI71
aI43
aI383
aI95
aI1689
aI639
aI4631
aI15113
aI28053
aI49247
aI128303
aI183999
asS'd'
I19188
sa(dp38376
S'a'
I97287
sS's'
I18
sS'm'
(lp38377
I1
aI1
aI1
aI3
aI19
aI31
aI93
aI35
aI369
aI765
aI1201
aI1625
aI7683
aI8719
aI13843
aI42723
aI62323
aI49431
asS'd'
I19189
sa(dp38378
S'a'
I97294
sS's'
I18
sS'm'
(lp38379
I1
aI3
aI3
aI11
aI5
aI39
aI49
aI217
aI109
aI63
aI1753
aI2489
aI6017
aI403
aI16657
aI59577
aI80255
aI66071
asS'd'
I19190
sa(dp38380
S'a'
I97299
sS's'
I18
sS'm'
(lp38381
I1
aI3
aI5
aI5
aI11
aI1
aI79
aI37
aI261
aI537
aI1845
aI3567
aI3233
aI16249
aI9795
aI2471
aI69661
aI118231
asS'd'
I19191
sa(dp38382
S'a'
I97306
sS's'
I18
sS'm'
(lp38383
I1
aI3
aI3
aI1
aI19
aI61
aI35
aI253
aI31
aI19
aI161
aI2597
aI5733
aI8231
aI26569
aI38613
aI121945
aI137391
asS'd'
I19192
sa(dp38384
S'a'
I97347
sS's'
I18
sS'm'
(lp38385
I1
aI3
aI7
aI3
aI15
aI25
aI125
aI231
aI187
aI797
aI1237
aI1767
aI1557
aI1095
aI13613
aI43325
aI33801
aI127881
asS'd'
I19193
sa(dp38386
S'a'
I97361
sS's'
I18
sS'm'
(lp38387
I1
aI1
aI1
aI9
aI23
aI63
aI75
aI107
aI311
aI493
aI471
aI2985
aI1861
aI4285
aI27125
aI14961
aI122567
aI152033
asS'd'
I19194
sa(dp38388
S'a'
I97371
sS's'
I18
sS'm'
(lp38389
I1
aI3
aI5
aI7
aI9
aI7
aI43
aI117
aI203
aI727
aI101
aI3831
aI3201
aI2327
aI4675
aI12085
aI25131
aI211835
asS'd'
I19195
sa(dp38390
S'a'
I97435
sS's'
I18
sS'm'
(lp38391
I1
aI1
aI7
aI11
aI17
aI1
aI5
aI87
aI291
aI1023
aI1345
aI3879
aI7739
aI9201
aI19573
aI20037
aI128711
aI187263
asS'd'
I19196
sa(dp38392
S'a'
I97466
sS's'
I18
sS'm'
(lp38393
I1
aI1
aI3
aI13
aI25
aI39
aI71
aI251
aI365
aI617
aI1539
aI2121
aI3803
aI8003
aI23393
aI56991
aI56143
aI223453
asS'd'
I19197
sa(dp38394
S'a'
I97473
sS's'
I18
sS'm'
(lp38395
I1
aI3
aI5
aI13
aI25
aI61
aI71
aI139
aI319
aI399
aI903
aI3063
aI3667
aI275
aI13297
aI25285
aI120417
aI169613
asS'd'
I19198
sa(dp38396
S'a'
I97474
sS's'
I18
sS'm'
(lp38397
I1
aI3
aI1
aI9
aI9
aI41
aI59
aI213
aI195
aI705
aI313
aI2313
aI4993
aI323
aI24049
aI30527
aI27287
aI80489
asS'd'
I19199
sa(dp38398
S'a'
I97559
sS's'
I18
sS'm'
(lp38399
I1
aI1
aI5
aI1
aI29
aI57
aI107
aI161
aI217
aI295
aI721
aI3857
aI1935
aI14981
aI12243
aI38541
aI51177
aI248889
asS'd'
I19200
sa(dp38400
S'a'
I97565
sS's'
I18
sS'm'
(lp38401
I1
aI1
aI1
aI15
aI5
aI25
aI95
aI137
aI11
aI215
aI971
aI1573
aI4341
aI4725
aI8201
aI33147
aI87687
aI187405
asS'd'
I19201
sa(dp38402
S'a'
I97587
sS's'
I18
sS'm'
(lp38403
I1
aI3
aI3
aI9
aI9
aI13
aI31
aI3
aI175
aI309
aI145
aI2265
aI4863
aI7199
aI23881
aI15445
aI123753
aI126653
asS'd'
I19202
sa(dp38404
S'a'
I97611
sS's'
I18
sS'm'
(lp38405
I1
aI1
aI7
aI3
aI7
aI43
aI51
aI191
aI21
aI639
aI939
aI691
aI7823
aI10529
aI7757
aI9291
aI115045
aI51539
asS'd'
I19203
sa(dp38406
S'a'
I97622
sS's'
I18
sS'm'
(lp38407
I1
aI1
aI7
aI13
aI7
aI45
aI91
aI173
aI73
aI779
aI1647
aI2059
aI1373
aI16027
aI4611
aI45787
aI699
aI78905
asS'd'
I19204
sa(dp38408
S'a'
I97625
sS's'
I18
sS'm'
(lp38409
I1
aI1
aI5
aI15
aI5
aI23
aI123
aI45
aI265
aI1009
aI235
aI1343
aI5779
aI209
aI23263
aI63163
aI26079
aI240905
asS'd'
I19205
sa(dp38410
S'a'
I97632
sS's'
I18
sS'm'
(lp38411
I1
aI1
aI1
aI11
aI19
aI31
aI75
aI105
aI71
aI21
aI1361
aI2125
aI6949
aI2111
aI10333
aI61881
aI112811
aI85723
asS'd'
I19206
sa(dp38412
S'a'
I97635
sS's'
I18
sS'm'
(lp38413
I1
aI3
aI5
aI7
aI27
aI17
aI95
aI35
aI503
aI181
aI1885
aI1097
aI6019
aI13745
aI15009
aI26343
aI117727
aI93017
asS'd'
I19207
sa(dp38414
S'a'
I97652
sS's'
I18
sS'm'
(lp38415
I1
aI1
aI3
aI11
aI27
aI41
aI109
aI23
aI365
aI283
aI1509
aI3269
aI5969
aI14567
aI27715
aI429
aI65813
aI169391
asS'd'
I19208
sa(dp38416
S'a'
I97672
sS's'
I18
sS'm'
(lp38417
I1
aI1
aI5
aI9
aI11
aI1
aI23
aI143
aI401
aI61
aI993
aI3029
aI1901
aI12947
aI10439
aI48661
aI113863
aI9353
asS'd'
I19209
sa(dp38418
S'a'
I97678
sS's'
I18
sS'm'
(lp38419
I1
aI3
aI7
aI3
aI15
aI27
aI123
aI51
aI403
aI569
aI75
aI3837
aI8167
aI10875
aI29861
aI44133
aI52385
aI185515
asS'd'
I19210
sa(dp38420
S'a'
I97683
sS's'
I18
sS'm'
(lp38421
I1
aI3
aI3
aI15
aI15
aI45
aI3
aI77
aI439
aI265
aI103
aI3715
aI7889
aI9241
aI26511
aI19063
aI108239
aI237233
asS'd'
I19211
sa(dp38422
S'a'
I97695
sS's'
I18
sS'm'
(lp38423
I1
aI1
aI5
aI13
aI7
aI47
aI7
aI185
aI155
aI833
aI1895
aI1103
aI6761
aI4307
aI19551
aI2371
aI41079
aI207663
asS'd'
I19212
sa(dp38424
S'a'
I97705
sS's'
I18
sS'm'
(lp38425
I1
aI1
aI3
aI7
aI1
aI49
aI79
aI127
aI149
aI383
aI919
aI3787
aI6703
aI8823
aI15551
aI28397
aI11497
aI144227
asS'd'
I19213
sa(dp38426
S'a'
I97713
sS's'
I18
sS'm'
(lp38427
I1
aI1
aI7
aI15
aI7
aI5
aI9
aI161
aI425
aI275
aI1943
aI3003
aI3615
aI1417
aI587
aI20949
aI9651
aI101257
asS'd'
I19214
sa(dp38428
S'a'
I97719
sS's'
I18
sS'm'
(lp38429
I1
aI3
aI5
aI11
aI31
aI11
aI113
aI201
aI113
aI889
aI867
aI3537
aI7173
aI3403
aI4713
aI29709
aI50127
aI55893
asS'd'
I19215
sa(dp38430
S'a'
I97757
sS's'
I18
sS'm'
(lp38431
I1
aI3
aI7
aI11
aI11
aI17
aI123
aI97
aI3
aI1009
aI1567
aI3261
aI8053
aI4639
aI24493
aI64085
aI73975
aI123965
asS'd'
I19216
sa(dp38432
S'a'
I97761
sS's'
I18
sS'm'
(lp38433
I1
aI1
aI7
aI1
aI31
aI7
aI111
aI137
aI427
aI615
aI865
aI2243
aI3603
aI5943
aI1639
aI22213
aI81977
aI77283
asS'd'
I19217
sa(dp38434
S'a'
I97762
sS's'
I18
sS'm'
(lp38435
I1
aI1
aI5
aI11
aI25
aI63
aI5
aI19
aI67
aI469
aI621
aI2831
aI1635
aI11859
aI23143
aI29189
aI43955
aI87475
asS'd'
I19218
sa(dp38436
S'a'
I97771
sS's'
I18
sS'm'
(lp38437
I1
aI3
aI7
aI15
aI7
aI61
aI125
aI207
aI401
aI567
aI1943
aI2645
aI641
aI15427
aI24467
aI41767
aI122591
aI48905
asS'd'
I19219
sa(dp38438
S'a'
I97795
sS's'
I18
sS'm'
(lp38439
I1
aI3
aI3
aI5
aI1
aI61
aI65
aI169
aI329
aI489
aI435
aI1719
aI491
aI6189
aI18383
aI34973
aI90611
aI180991
asS'd'
I19220
sa(dp38440
S'a'
I97809
sS's'
I18
sS'm'
(lp38441
I1
aI3
aI7
aI9
aI25
aI43
aI115
aI11
aI289
aI193
aI263
aI3885
aI4881
aI15669
aI19757
aI20073
aI119873
aI67069
asS'd'
I19221
sa(dp38442
S'a'
I97826
sS's'
I18
sS'm'
(lp38443
I1
aI1
aI7
aI11
aI3
aI45
aI93
aI115
aI233
aI891
aI1541
aI2557
aI2115
aI2237
aI4253
aI30445
aI32983
aI86185
asS'd'
I19222
sa(dp38444
S'a'
I97845
sS's'
I18
sS'm'
(lp38445
I1
aI3
aI7
aI3
aI29
aI23
aI105
aI51
aI157
aI505
aI773
aI2403
aI1237
aI5193
aI32725
aI53331
aI66377
aI25745
asS'd'
I19223
sa(dp38446
S'a'
I97877
sS's'
I18
sS'm'
(lp38447
I1
aI3
aI5
aI11
aI31
aI5
aI111
aI251
aI287
aI225
aI913
aI97
aI3429
aI15111
aI10637
aI18843
aI102589
aI229667
asS'd'
I19224
sa(dp38448
S'a'
I97882
sS's'
I18
sS'm'
(lp38449
I1
aI3
aI7
aI13
aI21
aI43
aI27
aI11
aI265
aI991
aI1645
aI1967
aI2675
aI3083
aI2957
aI65275
aI7757
aI201953
asS'd'
I19225
sa(dp38450
S'a'
I97891
sS's'
I18
sS'm'
(lp38451
I1
aI3
aI7
aI7
aI23
aI59
aI37
aI105
aI113
aI961
aI1585
aI855
aI6037
aI8461
aI24057
aI46861
aI42421
aI21061
asS'd'
I19226
sa(dp38452
S'a'
I97903
sS's'
I18
sS'm'
(lp38453
I1
aI1
aI5
aI1
aI7
aI45
aI37
aI147
aI225
aI793
aI737
aI753
aI565
aI5347
aI15393
aI42611
aI39253
aI246455
asS'd'
I19227
sa(dp38454
S'a'
I97961
sS's'
I18
sS'm'
(lp38455
I1
aI3
aI3
aI5
aI29
aI59
aI125
aI69
aI283
aI677
aI1615
aI3341
aI219
aI10753
aI445
aI43343
aI117035
aI137907
asS'd'
I19228
sa(dp38456
S'a'
I97970
sS's'
I18
sS'm'
(lp38457
I1
aI1
aI5
aI1
aI19
aI41
aI93
aI137
aI481
aI93
aI703
aI1211
aI4051
aI5591
aI5913
aI32831
aI62027
aI60519
asS'd'
I19229
sa(dp38458
S'a'
I97975
sS's'
I18
sS'm'
(lp38459
I1
aI1
aI7
aI13
aI17
aI63
aI65
aI147
aI361
aI83
aI1383
aI1761
aI579
aI9493
aI2611
aI6951
aI12197
aI81857
asS'd'
I19230
sa(dp38460
S'a'
I97996
sS's'
I18
sS'm'
(lp38461
I1
aI3
aI3
aI15
aI11
aI3
aI25
aI7
aI221
aI211
aI1745
aI1173
aI5479
aI12063
aI5667
aI43443
aI4865
aI193345
asS'd'
I19231
sa(dp38462
S'a'
I98001
sS's'
I18
sS'm'
(lp38463
I1
aI1
aI5
aI11
aI31
aI11
aI71
aI61
aI57
aI851
aI1089
aI1395
aI4525
aI1223
aI27681
aI14355
aI23125
aI257233
asS'd'
I19232
sa(dp38464
S'a'
I98014
sS's'
I18
sS'm'
(lp38465
I1
aI3
aI1
aI11
aI25
aI59
aI17
aI193
aI229
aI1005
aI387
aI3993
aI2457
aI4185
aI18421
aI1315
aI125155
aI142277
asS'd'
I19233
sa(dp38466
S'a'
I98023
sS's'
I18
sS'm'
(lp38467
I1
aI1
aI5
aI11
aI13
aI55
aI123
aI191
aI5
aI1023
aI705
aI3481
aI367
aI12961
aI11917
aI12131
aI99109
aI105093
asS'd'
I19234
sa(dp38468
S'a'
I98035
sS's'
I18
sS'm'
(lp38469
I1
aI1
aI3
aI11
aI13
aI29
aI57
aI57
aI467
aI19
aI1409
aI971
aI3041
aI13487
aI24737
aI3377
aI97883
aI248893
asS'd'
I19235
sa(dp38470
S'a'
I98052
sS's'
I18
sS'm'
(lp38471
I1
aI3
aI7
aI3
aI3
aI37
aI109
aI77
aI201
aI469
aI39
aI1747
aI2027
aI14781
aI18821
aI34647
aI123865
aI195097
asS'd'
I19236
sa(dp38472
S'a'
I98059
sS's'
I18
sS'm'
(lp38473
I1
aI3
aI3
aI5
aI29
aI27
aI97
aI217
aI249
aI141
aI431
aI1621
aI539
aI8945
aI3443
aI48227
aI27867
aI205355
asS'd'
I19237
sa(dp38474
S'a'
I98061
sS's'
I18
sS'm'
(lp38475
I1
aI1
aI3
aI13
aI7
aI57
aI65
aI167
aI103
aI511
aI239
aI325
aI1793
aI2811
aI14223
aI40999
aI12589
aI149759
asS'd'
I19238
sa(dp38476
S'a'
I98086
sS's'
I18
sS'm'
(lp38477
I1
aI3
aI5
aI11
aI3
aI1
aI61
aI87
aI283
aI29
aI507
aI3473
aI2685
aI13829
aI32337
aI8413
aI12201
aI152309
asS'd'
I19239
sa(dp38478
S'a'
I98098
sS's'
I18
sS'm'
(lp38479
I1
aI3
aI5
aI15
aI1
aI23
aI103
aI173
aI423
aI915
aI1519
aI1859
aI7341
aI8689
aI17141
aI53769
aI81189
aI144305
asS'd'
I19240
sa(dp38480
S'a'
I98117
sS's'
I18
sS'm'
(lp38481
I1
aI1
aI1
aI5
aI31
aI41
aI89
aI117
aI329
aI245
aI381
aI3357
aI1053
aI15079
aI3569
aI27665
aI65645
aI259279
asS'd'
I19241
sa(dp38482
S'a'
I98118
sS's'
I18
sS'm'
(lp38483
I1
aI1
aI7
aI5
aI3
aI55
aI91
aI35
aI463
aI15
aI1195
aI533
aI6013
aI10755
aI1919
aI61169
aI81285
aI82757
asS'd'
I19242
sa(dp38484
S'a'
I98132
sS's'
I18
sS'm'
(lp38485
I1
aI3
aI5
aI11
aI3
aI29
aI85
aI169
aI163
aI733
aI939
aI3401
aI3709
aI3307
aI17329
aI56873
aI10721
aI174235
asS'd'
I19243
sa(dp38486
S'a'
I98135
sS's'
I18
sS'm'
(lp38487
I1
aI1
aI5
aI1
aI11
aI45
aI75
aI247
aI435
aI21
aI1985
aI2261
aI7013
aI4935
aI2457
aI41077
aI53121
aI143269
asS'd'
I19244
sa(dp38488
S'a'
I98145
sS's'
I18
sS'm'
(lp38489
I1
aI3
aI3
aI13
aI17
aI59
aI43
aI149
aI27
aI1
aI367
aI957
aI5607
aI2591
aI22161
aI10095
aI73769
aI52455
asS'd'
I19245
sa(dp38490
S'a'
I98160
sS's'
I18
sS'm'
(lp38491
I1
aI1
aI3
aI13
aI15
aI15
aI121
aI83
aI469
aI819
aI1973
aI3595
aI2313
aI1621
aI3105
aI42971
aI7243
aI98727
asS'd'
I19246
sa(dp38492
S'a'
I98194
sS's'
I18
sS'm'
(lp38493
I1
aI1
aI5
aI7
aI21
aI53
aI123
aI9
aI119
aI437
aI1567
aI431
aI3647
aI10967
aI22037
aI8523
aI81279
aI126473
asS'd'
I19247
sa(dp38494
S'a'
I98205
sS's'
I18
sS'm'
(lp38495
I1
aI1
aI5
aI13
aI5
aI23
aI125
aI119
aI195
aI555
aI341
aI2037
aI313
aI6323
aI27201
aI8377
aI122793
aI197781
asS'd'
I19248
sa(dp38496
S'a'
I98210
sS's'
I18
sS'm'
(lp38497
I1
aI3
aI3
aI5
aI17
aI25
aI67
aI237
aI349
aI443
aI1529
aI3541
aI3105
aI10105
aI13409
aI20165
aI64597
aI244513
asS'd'
I19249
sa(dp38498
S'a'
I98224
sS's'
I18
sS'm'
(lp38499
I1
aI1
aI5
aI1
aI11
aI43
aI77
aI245
aI359
aI625
aI1171
aI597
aI3
aI591
aI2457
aI20275
aI75995
aI204685
asS'd'
I19250
sa(dp38500
S'a'
I98227
sS's'
I18
sS'm'
(lp38501
I1
aI3
aI1
aI11
aI5
aI13
aI99
aI107
aI285
aI617
aI1687
aI2959
aI4439
aI771
aI3103
aI62363
aI89437
aI172221
asS'd'
I19251
sa(dp38502
S'a'
I98254
sS's'
I18
sS'm'
(lp38503
I1
aI3
aI1
aI11
aI1
aI63
aI43
aI85
aI23
aI95
aI501
aI1223
aI669
aI16101
aI1071
aI53175
aI102101
aI419
asS'd'
I19252
sa(dp38504
S'a'
I98271
sS's'
I18
sS'm'
(lp38505
I1
aI3
aI1
aI5
aI19
aI23
aI63
aI105
aI289
aI419
aI885
aI441
aI5107
aI4213
aI8683
aI1847
aI113301
aI240821
asS'd'
I19253
sa(dp38506
S'a'
I98272
sS's'
I18
sS'm'
(lp38507
I1
aI1
aI1
aI9
aI9
aI9
aI111
aI63
aI53
aI531
aI517
aI3463
aI8171
aI2645
aI13883
aI52213
aI40707
aI24637
asS'd'
I19254
sa(dp38508
S'a'
I98302
sS's'
I18
sS'm'
(lp38509
I1
aI3
aI1
aI5
aI15
aI43
aI71
aI215
aI117
aI685
aI1819
aI1105
aI5805
aI8875
aI31093
aI31077
aI93807
aI65631
asS'd'
I19255
sa(dp38510
S'a'
I98320
sS's'
I18
sS'm'
(lp38511
I1
aI1
aI7
aI7
aI17
aI15
aI31
aI87
aI13
aI615
aI2003
aI3461
aI7585
aI1947
aI6693
aI26141
aI95059
aI52229
asS'd'
I19256
sa(dp38512
S'a'
I98346
sS's'
I18
sS'm'
(lp38513
I1
aI1
aI3
aI5
aI5
aI55
aI7
aI41
aI473
aI541
aI545
aI2901
aI763
aI12731
aI24715
aI43301
aI7981
aI123961
asS'd'
I19257
sa(dp38514
S'a'
I98356
sS's'
I18
sS'm'
(lp38515
I1
aI3
aI1
aI11
aI13
aI29
aI65
aI47
aI511
aI931
aI1681
aI3813
aI995
aI4261
aI32243
aI21327
aI33749
aI52607
asS'd'
I19258
sa(dp38516
S'a'
I98360
sS's'
I18
sS'm'
(lp38517
I1
aI1
aI3
aI1
aI27
aI51
aI19
aI119
aI71
aI989
aI485
aI1483
aI4115
aI11743
aI5513
aI32447
aI62599
aI163185
asS'd'
I19259
sa(dp38518
S'a'
I98366
sS's'
I18
sS'm'
(lp38519
I1
aI3
aI7
aI13
aI7
aI5
aI127
aI67
aI221
aI773
aI1641
aI3763
aI2061
aI2025
aI29813
aI64385
aI109219
aI70149
asS'd'
I19260
sa(dp38520
S'a'
I98478
sS's'
I18
sS'm'
(lp38521
I1
aI1
aI5
aI15
aI9
aI29
aI105
aI245
aI333
aI11
aI803
aI1877
aI6735
aI3797
aI1913
aI63837
aI23649
aI234721
asS'd'
I19261
sa(dp38522
S'a'
I98483
sS's'
I18
sS'm'
(lp38523
I1
aI3
aI7
aI13
aI11
aI21
aI113
aI175
aI385
aI885
aI1259
aI983
aI7715
aI11889
aI12515
aI35723
aI9897
aI63415
asS'd'
I19262
sa(dp38524
S'a'
I98486
sS's'
I18
sS'm'
(lp38525
I1
aI1
aI5
aI9
aI31
aI63
aI53
aI51
aI375
aI133
aI2021
aI3173
aI3861
aI9885
aI4117
aI37505
aI73687
aI16411
asS'd'
I19263
sa(dp38526
S'a'
I98497
sS's'
I18
sS'm'
(lp38527
I1
aI3
aI7
aI7
aI11
aI13
aI99
aI235
aI285
aI159
aI489
aI917
aI3033
aI7711
aI6545
aI52893
aI28549
aI68791
asS'd'
I19264
sa(dp38528
S'a'
I98528
sS's'
I18
sS'm'
(lp38529
I1
aI1
aI5
aI11
aI31
aI15
aI89
aI157
aI105
aI347
aI455
aI3391
aI5341
aI16035
aI11819
aI57679
aI48057
aI147673
asS'd'
I19265
sa(dp38530
S'a'
I98537
sS's'
I18
sS'm'
(lp38531
I1
aI3
aI1
aI5
aI21
aI5
aI1
aI41
aI213
aI677
aI1745
aI2591
aI6237
aI14265
aI5963
aI30017
aI47293
aI199411
asS'd'
I19266
sa(dp38532
S'a'
I98551
sS's'
I18
sS'm'
(lp38533
I1
aI3
aI1
aI15
aI19
aI9
aI65
aI103
aI489
aI977
aI579
aI2571
aI2827
aI12971
aI24445
aI17963
aI68829
aI89781
asS'd'
I19267
sa(dp38534
S'a'
I98557
sS's'
I18
sS'm'
(lp38535
I1
aI3
aI5
aI7
aI3
aI45
aI9
aI223
aI137
aI749
aI919
aI2695
aI7569
aI6735
aI16649
aI55899
aI91531
aI10709
asS'd'
I19268
sa(dp38536
S'a'
I98572
sS's'
I18
sS'm'
(lp38537
I1
aI1
aI5
aI11
aI25
aI51
aI81
aI243
aI473
aI85
aI1189
aI2317
aI785
aI9307
aI25555
aI36623
aI66881
aI150945
asS'd'
I19269
sa(dp38538
S'a'
I98575
sS's'
I18
sS'm'
(lp38539
I1
aI1
aI3
aI7
aI9
aI17
aI99
aI57
aI333
aI891
aI71
aI2359
aI2067
aI13265
aI30077
aI17935
aI47343
aI22673
asS'd'
I19270
sa(dp38540
S'a'
I98600
sS's'
I18
sS'm'
(lp38541
I1
aI1
aI5
aI7
aI13
aI17
aI77
aI109
aI427
aI667
aI1367
aI2383
aI7505
aI11239
aI14229
aI35431
aI35473
aI62447
asS'd'
I19271
sa(dp38542
S'a'
I98628
sS's'
I18
sS'm'
(lp38543
I1
aI1
aI1
aI15
aI27
aI5
aI51
aI221
aI471
aI877
aI449
aI3961
aI4197
aI15713
aI2955
aI58985
aI31431
aI241539
asS'd'
I19272
sa(dp38544
S'a'
I98635
sS's'
I18
sS'm'
(lp38545
I1
aI1
aI7
aI1
aI13
aI61
aI55
aI199
aI87
aI679
aI723
aI271
aI1061
aI8043
aI13163
aI8079
aI81501
aI60467
asS'd'
I19273
sa(dp38546
S'a'
I98645
sS's'
I18
sS'm'
(lp38547
I1
aI1
aI3
aI3
aI11
aI1
aI85
aI65
aI445
aI731
aI2017
aI3113
aI8085
aI7133
aI14789
aI2435
aI38459
aI234997
asS'd'
I19274
sa(dp38548
S'a'
I98652
sS's'
I18
sS'm'
(lp38549
I1
aI3
aI3
aI9
aI23
aI31
aI49
aI137
aI349
aI651
aI1975
aI3429
aI7137
aI7841
aI28297
aI58209
aI36493
aI259097
asS'd'
I19275
sa(dp38550
S'a'
I98655
sS's'
I18
sS'm'
(lp38551
I1
aI1
aI7
aI15
aI23
aI11
aI87
aI133
aI245
aI445
aI151
aI4075
aI141
aI15395
aI16649
aI36925
aI98421
aI217265
asS'd'
I19276
sa(dp38552
S'a'
I98665
sS's'
I18
sS'm'
(lp38553
I1
aI3
aI3
aI5
aI25
aI53
aI57
aI177
aI481
aI177
aI671
aI1249
aI2663
aI12855
aI24537
aI31867
aI110323
aI164113
asS'd'
I19277
sa(dp38554
S'a'
I98710
sS's'
I18
sS'm'
(lp38555
I1
aI3
aI5
aI7
aI23
aI25
aI19
aI91
aI447
aI1023
aI373
aI3863
aI4399
aI12973
aI7475
aI37485
aI8567
aI53271
asS'd'
I19278
sa(dp38556
S'a'
I98719
sS's'
I18
sS'm'
(lp38557
I1
aI1
aI5
aI7
aI31
aI33
aI31
aI75
aI223
aI299
aI1549
aI1863
aI353
aI4339
aI8891
aI10365
aI3399
aI185807
asS'd'
I19279
sa(dp38558
S'a'
I98720
sS's'
I18
sS'm'
(lp38559
I1
aI1
aI7
aI9
aI31
aI53
aI23
aI203
aI319
aI915
aI1923
aI205
aI3119
aI7243
aI25251
aI12907
aI101921
aI102695
asS'd'
I19280
sa(dp38560
S'a'
I98786
sS's'
I18
sS'm'
(lp38561
I1
aI1
aI7
aI9
aI15
aI1
aI123
aI173
aI123
aI215
aI263
aI3003
aI5881
aI1117
aI15195
aI47457
aI66663
aI224177
asS'd'
I19281
sa(dp38562
S'a'
I98792
sS's'
I18
sS'm'
(lp38563
I1
aI1
aI7
aI13
aI11
aI25
aI61
aI121
aI173
aI115
aI1897
aI2145
aI7783
aI9673
aI3321
aI1707
aI61475
aI53875
asS'd'
I19282
sa(dp38564
S'a'
I98806
sS's'
I18
sS'm'
(lp38565
I1
aI3
aI7
aI3
aI31
aI21
aI27
aI99
aI421
aI225
aI1565
aI2351
aI2275
aI10583
aI7877
aI43505
aI27629
aI140919
asS'd'
I19283
sa(dp38566
S'a'
I98816
sS's'
I18
sS'm'
(lp38567
I1
aI3
aI5
aI5
aI11
aI45
aI71
aI105
aI487
aI867
aI361
aI3995
aI2039
aI1495
aI27481
aI4753
aI20657
aI67077
asS'd'
I19284
sa(dp38568
S'a'
I98836
sS's'
I18
sS'm'
(lp38569
I1
aI1
aI5
aI1
aI19
aI33
aI1
aI77
aI377
aI353
aI719
aI1463
aI7053
aI7409
aI32165
aI15557
aI117673
aI69887
asS'd'
I19285
sa(dp38570
S'a'
I98859
sS's'
I18
sS'm'
(lp38571
I1
aI1
aI5
aI7
aI25
aI5
aI15
aI231
aI23
aI213
aI1627
aI1801
aI7793
aI651
aI9903
aI51745
aI111611
aI39679
asS'd'
I19286
sa(dp38572
S'a'
I98864
sS's'
I18
sS'm'
(lp38573
I1
aI3
aI3
aI5
aI23
aI43
aI37
aI199
aI437
aI19
aI1853
aI2119
aI461
aI12641
aI15865
aI39941
aI122545
aI213443
asS'd'
I19287
sa(dp38574
S'a'
I98879
sS's'
I18
sS'm'
(lp38575
I1
aI3
aI3
aI11
aI31
aI45
aI19
aI227
aI507
aI909
aI1501
aI2021
aI905
aI1763
aI1897
aI3735
aI81475
aI30005
asS'd'
I19288
sa(dp38576
S'a'
I98905
sS's'
I18
sS'm'
(lp38577
I1
aI1
aI5
aI5
aI29
aI9
aI55
aI25
aI23
aI59
aI593
aI2197
aI6029
aI8235
aI8397
aI27521
aI96221
aI168837
asS'd'
I19289
sa(dp38578
S'a'
I98917
sS's'
I18
sS'm'
(lp38579
I1
aI3
aI1
aI15
aI5
aI33
aI75
aI121
aI433
aI557
aI1011
aI3785
aI2545
aI953
aI17295
aI14407
aI94871
aI60445
asS'd'
I19290
sa(dp38580
S'a'
I98929
sS's'
I18
sS'm'
(lp38581
I1
aI3
aI3
aI7
aI7
aI53
aI29
aI75
aI171
aI587
aI1701
aI3815
aI2761
aI4403
aI39
aI17291
aI34897
aI187257
asS'd'
I19291
sa(dp38582
S'a'
I98969
sS's'
I18
sS'm'
(lp38583
I1
aI3
aI1
aI15
aI17
aI57
aI11
aI95
aI335
aI13
aI265
aI1161
aI7945
aI6419
aI26723
aI31907
aI89995
aI82265
asS'd'
I19292
sa(dp38584
S'a'
I98975
sS's'
I18
sS'm'
(lp38585
I1
aI1
aI7
aI5
aI9
aI59
aI27
aI153
aI37
aI165
aI823
aI3525
aI621
aI4777
aI3485
aI9109
aI116567
aI34691
asS'd'
I19293
sa(dp38586
S'a'
I98976
sS's'
I18
sS'm'
(lp38587
I1
aI1
aI5
aI13
aI23
aI27
aI11
aI63
aI35
aI39
aI995
aI2101
aI2611
aI14139
aI2683
aI63787
aI19813
aI97497
asS'd'
I19294
sa(dp38588
S'a'
I98981
sS's'
I18
sS'm'
(lp38589
I1
aI3
aI7
aI15
aI31
aI15
aI3
aI163
aI167
aI53
aI71
aI1881
aI4213
aI3485
aI21525
aI705
aI122345
aI203549
asS'd'
I19295
sa(dp38590
S'a'
I98999
sS's'
I18
sS'm'
(lp38591
I1
aI3
aI5
aI5
aI21
aI33
aI85
aI133
aI21
aI505
aI1639
aI3989
aI771
aI7171
aI21953
aI34503
aI31247
aI247459
asS'd'
I19296
sa(dp38592
S'a'
I99020
sS's'
I18
sS'm'
(lp38593
I1
aI1
aI7
aI7
aI31
aI1
aI27
aI39
aI469
aI243
aI679
aI4091
aI7137
aI8505
aI13329
aI34139
aI69485
aI259795
asS'd'
I19297
sa(dp38594
S'a'
I99026
sS's'
I18
sS'm'
(lp38595
I1
aI3
aI5
aI5
aI31
aI43
aI31
aI161
aI413
aI657
aI1407
aI1417
aI7349
aI3301
aI7691
aI49355
aI22929
aI68043
asS'd'
I19298
sa(dp38596
S'a'
I99054
sS's'
I18
sS'm'
(lp38597
I1
aI1
aI7
aI11
aI15
aI61
aI73
aI217
aI163
aI503
aI193
aI3795
aI41
aI16251
aI1187
aI65363
aI113211
aI100337
asS'd'
I19299
sa(dp38598
S'a'
I99083
sS's'
I18
sS'm'
(lp38599
I1
aI3
aI1
aI11
aI9
aI15
aI109
aI187
aI109
aI865
aI845
aI1579
aI321
aI1269
aI20613
aI5693
aI58421
aI254959
asS'd'
I19300
sa(dp38600
S'a'
I99093
sS's'
I18
sS'm'
(lp38601
I1
aI3
aI1
aI13
aI11
aI3
aI19
aI135
aI93
aI779
aI1383
aI219
aI2737
aI377
aI1125
aI35663
aI130815
aI103797
asS'd'
I19301
sa(dp38602
S'a'
I99103
sS's'
I18
sS'm'
(lp38603
I1
aI1
aI5
aI11
aI25
aI25
aI71
aI249
aI201
aI679
aI1677
aI1817
aI7619
aI10327
aI14821
aI47847
aI33629
aI250979
asS'd'
I19302
sa(dp38604
S'a'
I99131
sS's'
I18
sS'm'
(lp38605
I1
aI1
aI7
aI15
aI23
aI19
aI69
aI39
aI25
aI843
aI99
aI3499
aI2457
aI11681
aI30009
aI17609
aI46653
aI162427
asS'd'
I19303
sa(dp38606
S'a'
I99156
sS's'
I18
sS'm'
(lp38607
I1
aI3
aI3
aI7
aI23
aI25
aI77
aI135
aI61
aI501
aI1381
aI3977
aI1957
aI11255
aI16053
aI30297
aI58835
aI97589
asS'd'
I19304
sa(dp38608
S'a'
I99159
sS's'
I18
sS'm'
(lp38609
I1
aI3
aI5
aI3
aI9
aI31
aI9
aI55
aI421
aI109
aI1823
aI1921
aI7349
aI2661
aI4503
aI36691
aI48843
aI182631
asS'd'
I19305
sa(dp38610
S'a'
I99165
sS's'
I18
sS'm'
(lp38611
I1
aI1
aI1
aI11
aI7
aI23
aI107
aI125
aI393
aI105
aI1407
aI3461
aI4539
aI6121
aI7881
aI32407
aI83749
aI98831
asS'd'
I19306
sa(dp38612
S'a'
I99170
sS's'
I18
sS'm'
(lp38613
I1
aI3
aI3
aI13
aI5
aI59
aI5
aI3
aI185
aI959
aI241
aI819
aI1443
aI1789
aI12771
aI26703
aI25399
aI182583
asS'd'
I19307
sa(dp38614
S'a'
I99189
sS's'
I18
sS'm'
(lp38615
I1
aI1
aI1
aI3
aI3
aI47
aI7
aI45
aI93
aI373
aI175
aI87
aI649
aI12903
aI5029
aI1945
aI111967
aI140889
asS'd'
I19308
sa(dp38616
S'a'
I99223
sS's'
I18
sS'm'
(lp38617
I1
aI3
aI1
aI11
aI9
aI47
aI25
aI191
aI215
aI845
aI1557
aI9
aI3451
aI5837
aI11763
aI29127
aI113115
aI99039
asS'd'
I19309
sa(dp38618
S'a'
I99227
sS's'
I18
sS'm'
(lp38619
I1
aI1
aI1
aI5
aI23
aI53
aI45
aI1
aI361
aI751
aI807
aI1765
aI685
aI2109
aI28437
aI60489
aI65739
aI234511
asS'd'
I19310
sa(dp38620
S'a'
I99271
sS's'
I18
sS'm'
(lp38621
I1
aI3
aI3
aI7
aI15
aI57
aI71
aI61
aI195
aI123
aI1745
aI3249
aI351
aI14309
aI2017
aI15653
aI110803
aI45937
asS'd'
I19311
sa(dp38622
S'a'
I99277
sS's'
I18
sS'm'
(lp38623
I1
aI3
aI7
aI9
aI25
aI11
aI25
aI29
aI467
aI313
aI1927
aI2423
aI7311
aI14299
aI8145
aI8123
aI115103
aI213881
asS'd'
I19312
sa(dp38624
S'a'
I99278
sS's'
I18
sS'm'
(lp38625
I1
aI1
aI3
aI15
aI1
aI35
aI111
aI99
aI507
aI417
aI1433
aI129
aI5565
aI13365
aI18853
aI8607
aI109739
aI120623
asS'd'
I19313
sa(dp38626
S'a'
I99313
sS's'
I18
sS'm'
(lp38627
I1
aI1
aI1
aI7
aI13
aI31
aI93
aI3
aI327
aI67
aI1101
aI1965
aI5939
aI6505
aI3117
aI3021
aI33707
aI79353
asS'd'
I19314
sa(dp38628
S'a'
I99314
sS's'
I18
sS'm'
(lp38629
I1
aI1
aI3
aI7
aI15
aI21
aI23
aI117
aI367
aI137
aI287
aI903
aI4685
aI13943
aI26779
aI24607
aI70853
aI99743
asS'd'
I19315
sa(dp38630
S'a'
I99345
sS's'
I18
sS'm'
(lp38631
I1
aI1
aI7
aI11
aI25
aI43
aI67
aI181
aI459
aI737
aI1567
aI3491
aI5085
aI6487
aI23115
aI62341
aI102943
aI77301
asS'd'
I19316
sa(dp38632
S'a'
I99361
sS's'
I18
sS'm'
(lp38633
I1
aI1
aI3
aI15
aI7
aI35
aI81
aI199
aI455
aI851
aI835
aI3421
aI4675
aI15173
aI9205
aI7305
aI109849
aI15183
asS'd'
I19317
sa(dp38634
S'a'
I99367
sS's'
I18
sS'm'
(lp38635
I1
aI3
aI5
aI11
aI9
aI55
aI3
aI53
aI235
aI271
aI1265
aI3681
aI3627
aI3485
aI11591
aI53097
aI85949
aI158173
asS'd'
I19318
sa(dp38636
S'a'
I99386
sS's'
I18
sS'm'
(lp38637
I1
aI3
aI7
aI3
aI15
aI27
aI57
aI183
aI487
aI9
aI1797
aI2973
aI3687
aI12987
aI9133
aI14595
aI52067
aI131217
asS'd'
I19319
sa(dp38638
S'a'
I99394
sS's'
I18
sS'm'
(lp38639
I1
aI1
aI5
aI3
aI7
aI25
aI19
aI215
aI291
aI325
aI813
aI577
aI4249
aI10373
aI17233
aI29557
aI72979
aI70721
asS'd'
I19320
sa(dp38640
S'a'
I99417
sS's'
I18
sS'm'
(lp38641
I1
aI3
aI1
aI7
aI25
aI1
aI107
aI167
aI367
aI303
aI883
aI993
aI4189
aI6557
aI13697
aI15251
aI77065
aI116127
asS'd'
I19321
sa(dp38642
S'a'
I99418
sS's'
I18
sS'm'
(lp38643
I1
aI3
aI5
aI11
aI13
aI59
aI9
aI121
aI489
aI593
aI1503
aI601
aI5263
aI13837
aI20991
aI35761
aI45867
aI155905
asS'd'
I19322
sa(dp38644
S'a'
I99453
sS's'
I18
sS'm'
(lp38645
I1
aI1
aI3
aI3
aI19
aI47
aI127
aI115
aI267
aI261
aI969
aI961
aI5919
aI10085
aI29363
aI4935
aI100485
aI75561
asS'd'
I19323
sa(dp38646
S'a'
I99454
sS's'
I18
sS'm'
(lp38647
I1
aI3
aI1
aI15
aI11
aI53
aI39
aI187
aI53
aI11
aI1951
aI913
aI965
aI2565
aI5457
aI3237
aI24923
aI245681
asS'd'
I19324
sa(dp38648
S'a'
I99477
sS's'
I18
sS'm'
(lp38649
I1
aI1
aI7
aI3
aI15
aI5
aI25
aI45
aI17
aI45
aI1317
aI1853
aI6627
aI15879
aI29935
aI24749
aI118149
aI35359
asS'd'
I19325
sa(dp38650
S'a'
I99518
sS's'
I18
sS'm'
(lp38651
I1
aI1
aI7
aI1
aI21
aI45
aI67
aI71
aI25
aI743
aI925
aI3441
aI3013
aI1613
aI6321
aI12491
aI119931
aI164701
asS'd'
I19326
sa(dp38652
S'a'
I99544
sS's'
I18
sS'm'
(lp38653
I1
aI1
aI7
aI1
aI13
aI15
aI35
aI187
aI91
aI995
aI401
aI2443
aI4183
aI10823
aI20589
aI27413
aI117095
aI20359
asS'd'
I19327
sa(dp38654
S'a'
I99559
sS's'
I18
sS'm'
(lp38655
I1
aI3
aI3
aI7
aI15
aI51
aI55
aI167
aI409
aI859
aI719
aI3223
aI2457
aI16013
aI13639
aI4027
aI79339
aI225113
asS'd'
I19328
sa(dp38656
S'a'
I99592
sS's'
I18
sS'm'
(lp38657
I1
aI3
aI1
aI9
aI3
aI29
aI105
aI193
aI279
aI27
aI1093
aI2199
aI6983
aI619
aI10163
aI40365
aI71015
aI102191
asS'd'
I19329
sa(dp38658
S'a'
I99597
sS's'
I18
sS'm'
(lp38659
I1
aI1
aI3
aI9
aI29
aI5
aI33
aI247
aI27
aI299
aI2017
aI379
aI6199
aI15047
aI18329
aI3493
aI47679
aI76703
asS'd'
I19330
sa(dp38660
S'a'
I99603
sS's'
I18
sS'm'
(lp38661
I1
aI3
aI5
aI5
aI9
aI19
aI51
aI129
aI157
aI831
aI1373
aI653
aI7489
aI13125
aI1815
aI10915
aI88679
aI50269
asS'd'
I19331
sa(dp38662
S'a'
I99621
sS's'
I18
sS'm'
(lp38663
I1
aI3
aI3
aI9
aI9
aI49
aI79
aI11
aI181
aI679
aI1697
aI3707
aI205
aI13305
aI6293
aI56653
aI42619
aI116257
asS'd'
I19332
sa(dp38664
S'a'
I99646
sS's'
I18
sS'm'
(lp38665
I1
aI1
aI5
aI9
aI23
aI41
aI17
aI135
aI145
aI715
aI257
aI1561
aI6941
aI2411
aI31459
aI25055
aI35807
aI51579
asS'd'
I19333
sa(dp38666
S'a'
I99693
sS's'
I18
sS'm'
(lp38667
I1
aI1
aI1
aI7
aI11
aI13
aI49
aI155
aI403
aI569
aI751
aI2959
aI425
aI13949
aI22047
aI49829
aI71925
aI101647
asS'd'
I19334
sa(dp38668
S'a'
I99722
sS's'
I18
sS'm'
(lp38669
I1
aI1
aI3
aI15
aI15
aI15
aI17
aI213
aI113
aI395
aI1999
aI2039
aI3623
aI13255
aI24435
aI54487
aI78773
aI202637
asS'd'
I19335
sa(dp38670
S'a'
I99760
sS's'
I18
sS'm'
(lp38671
I1
aI3
aI7
aI9
aI5
aI21
aI61
aI165
aI97
aI349
aI1131
aI2677
aI333
aI13129
aI2137
aI22909
aI95795
aI143081
asS'd'
I19336
sa(dp38672
S'a'
I99780
sS's'
I18
sS'm'
(lp38673
I1
aI1
aI1
aI5
aI31
aI41
aI109
aI179
aI295
aI475
aI639
aI3929
aI1841
aI7545
aI19411
aI52573
aI10173
aI236769
asS'd'
I19337
sa(dp38674
S'a'
I99789
sS's'
I18
sS'm'
(lp38675
I1
aI1
aI1
aI5
aI27
aI51
aI9
aI217
aI393
aI671
aI931
aI433
aI7303
aI16295
aI6727
aI5703
aI88241
aI132665
asS'd'
I19338
sa(dp38676
S'a'
I99804
sS's'
I18
sS'm'
(lp38677
I1
aI1
aI7
aI13
aI21
aI33
aI19
aI241
aI497
aI519
aI1413
aI489
aI4975
aI1345
aI24925
aI40383
aI110815
aI136217
asS'd'
I19339
sa(dp38678
S'a'
I99823
sS's'
I18
sS'm'
(lp38679
I1
aI3
aI1
aI9
aI7
aI51
aI79
aI15
aI15
aI601
aI997
aI3713
aI7829
aI903
aI12393
aI60059
aI42057
aI175141
asS'd'
I19340
sa(dp38680
S'a'
I99826
sS's'
I18
sS'm'
(lp38681
I1
aI1
aI5
aI15
aI9
aI63
aI107
aI63
aI495
aI591
aI207
aI779
aI8069
aI3013
aI23839
aI3075
aI127481
aI193885
asS'd'
I19341
sa(dp38682
S'a'
I99832
sS's'
I18
sS'm'
(lp38683
I1
aI1
aI1
aI7
aI13
aI17
aI121
aI171
aI99
aI59
aI1043
aI1109
aI1337
aI1179
aI27635
aI34063
aI12945
aI1431
asS'd'
I19342
sa(dp38684
S'a'
I99842
sS's'
I18
sS'm'
(lp38685
I1
aI3
aI1
aI3
aI5
aI47
aI101
aI205
aI157
aI595
aI263
aI3887
aI7015
aI4693
aI15211
aI25381
aI128803
aI227233
asS'd'
I19343
sa(dp38686
S'a'
I99851
sS's'
I18
sS'm'
(lp38687
I1
aI1
aI3
aI11
aI17
aI33
aI1
aI19
aI153
aI603
aI119
aI2305
aI4041
aI4011
aI19849
aI761
aI52807
aI129811
asS'd'
I19344
sa(dp38688
S'a'
I99862
sS's'
I18
sS'm'
(lp38689
I1
aI3
aI7
aI15
aI21
aI7
aI13
aI225
aI497
aI459
aI389
aI911
aI6349
aI5059
aI6363
aI41915
aI90687
aI214501
asS'd'
I19345
sa(dp38690
S'a'
I99871
sS's'
I18
sS'm'
(lp38691
I1
aI3
aI1
aI15
aI1
aI39
aI31
aI83
aI147
aI629
aI185
aI1913
aI3217
aI959
aI651
aI65267
aI108613
aI20391
asS'd'
I19346
sa(dp38692
S'a'
I99899
sS's'
I18
sS'm'
(lp38693
I1
aI3
aI1
aI7
aI9
aI11
aI29
aI201
aI245
aI815
aI1869
aI2597
aI5693
aI15669
aI23293
aI30885
aI4029
aI225737
asS'd'
I19347
sa(dp38694
S'a'
I99940
sS's'
I18
sS'm'
(lp38695
I1
aI3
aI5
aI11
aI7
aI29
aI119
aI207
aI499
aI23
aI1563
aI3645
aI3839
aI2509
aI24043
aI64231
aI22509
aI221835
asS'd'
I19348
sa(dp38696
S'a'
I99983
sS's'
I18
sS'm'
(lp38697
I1
aI3
aI3
aI7
aI11
aI49
aI13
aI201
aI353
aI217
aI831
aI2803
aI1521
aI12989
aI25339
aI41035
aI2125
aI165271
asS'd'
I19349
sa(dp38698
S'a'
I99985
sS's'
I18
sS'm'
(lp38699
I1
aI1
aI5
aI11
aI1
aI45
aI93
aI29
aI55
aI1007
aI1919
aI241
aI5931
aI9211
aI17291
aI39849
aI25453
aI96077
asS'd'
I19350
sa(dp38700
S'a'
I100025
sS's'
I18
sS'm'
(lp38701
I1
aI3
aI1
aI1
aI29
aI43
aI11
aI105
aI331
aI693
aI1363
aI291
aI8191
aI7813
aI14135
aI38287
aI15469
aI256913
asS'd'
I19351
sa(dp38702
S'a'
I100043
sS's'
I18
sS'm'
(lp38703
I1
aI3
aI5
aI11
aI19
aI21
aI23
aI117
aI253
aI111
aI733
aI3785
aI5835
aI423
aI30251
aI27283
aI79561
aI162095
asS'd'
I19352
sa(dp38704
S'a'
I100070
sS's'
I18
sS'm'
(lp38705
I1
aI3
aI3
aI1
aI9
aI19
aI83
aI13
aI37
aI725
aI1597
aI1117
aI8067
aI8085
aI1315
aI41813
aI8973
aI175525
asS'd'
I19353
sa(dp38706
S'a'
I100076
sS's'
I18
sS'm'
(lp38707
I1
aI1
aI5
aI3
aI13
aI39
aI3
aI213
aI335
aI907
aI1143
aI1729
aI601
aI11255
aI24351
aI41045
aI11335
aI186221
asS'd'
I19354
sa(dp38708
S'a'
I100084
sS's'
I18
sS'm'
(lp38709
I1
aI3
aI3
aI9
aI17
aI11
aI29
aI33
aI303
aI815
aI1607
aI2403
aI8095
aI4213
aI16697
aI64733
aI24439
aI93081
asS'd'
I19355
sa(dp38710
S'a'
I100096
sS's'
I18
sS'm'
(lp38711
I1
aI1
aI1
aI3
aI13
aI37
aI35
aI181
aI243
aI645
aI1915
aI3521
aI569
aI3005
aI7271
aI32755
aI64575
aI119813
asS'd'
I19356
sa(dp38712
S'a'
I100105
sS's'
I18
sS'm'
(lp38713
I1
aI1
aI5
aI13
aI25
aI59
aI77
aI121
aI459
aI755
aI385
aI1893
aI1227
aI9957
aI5069
aI40063
aI27261
aI4703
asS'd'
I19357
sa(dp38714
S'a'
I100108
sS's'
I18
sS'm'
(lp38715
I1
aI3
aI1
aI3
aI29
aI41
aI95
aI255
aI219
aI21
aI317
aI3021
aI2615
aI5101
aI19413
aI25795
aI6521
aI157749
asS'd'
I19358
sa(dp38716
S'a'
I100123
sS's'
I18
sS'm'
(lp38717
I1
aI3
aI3
aI5
aI7
aI33
aI7
aI205
aI415
aI23
aI1431
aI117
aI1605
aI9541
aI11825
aI49195
aI86341
aI99673
asS'd'
I19359
sa(dp38718
S'a'
I100132
sS's'
I18
sS'm'
(lp38719
I1
aI3
aI3
aI5
aI17
aI37
aI33
aI209
aI49
aI161
aI321
aI3697
aI6483
aI12859
aI895
aI675
aI1899
aI260289
asS'd'
I19360
sa(dp38720
S'a'
I100141
sS's'
I18
sS'm'
(lp38721
I1
aI1
aI3
aI15
aI1
aI27
aI83
aI125
aI309
aI553
aI505
aI2209
aI4931
aI2593
aI28253
aI12879
aI74971
aI9047
asS'd'
I19361
sa(dp38722
S'a'
I100181
sS's'
I18
sS'm'
(lp38723
I1
aI3
aI7
aI3
aI15
aI3
aI105
aI19
aI41
aI119
aI149
aI3847
aI6593
aI875
aI23777
aI4547
aI57717
aI139387
asS'd'
I19362
sa(dp38724
S'a'
I100191
sS's'
I18
sS'm'
(lp38725
I1
aI1
aI1
aI3
aI9
aI43
aI25
aI15
aI67
aI609
aI951
aI273
aI8095
aI621
aI24819
aI17233
aI53423
aI192757
asS'd'
I19363
sa(dp38726
S'a'
I100202
sS's'
I18
sS'm'
(lp38727
I1
aI3
aI5
aI3
aI9
aI49
aI107
aI215
aI245
aI217
aI545
aI2285
aI2075
aI401
aI26599
aI32967
aI130457
aI203755
asS'd'
I19364
sa(dp38728
S'a'
I100216
sS's'
I18
sS'm'
(lp38729
I1
aI1
aI5
aI1
aI7
aI31
aI87
aI181
aI135
aI155
aI9
aI1431
aI307
aI13367
aI31147
aI10327
aI2817
aI63327
asS'd'
I19365
sa(dp38730
S'a'
I100246
sS's'
I18
sS'm'
(lp38731
I1
aI3
aI3
aI3
aI29
aI17
aI55
aI201
aI33
aI275
aI2005
aI3037
aI3439
aI1513
aI7563
aI46491
aI103319
aI5279
asS'd'
I19366
sa(dp38732
S'a'
I100256
sS's'
I18
sS'm'
(lp38733
I1
aI3
aI7
aI11
aI31
aI63
aI105
aI169
aI257
aI225
aI711
aI2041
aI1519
aI11801
aI18543
aI35173
aI92125
aI72729
asS'd'
I19367
sa(dp38734
S'a'
I100271
sS's'
I18
sS'm'
(lp38735
I1
aI1
aI3
aI3
aI11
aI13
aI127
aI231
aI229
aI809
aI303
aI1167
aI47
aI4281
aI2373
aI10455
aI74685
aI131775
asS'd'
I19368
sa(dp38736
S'a'
I100285
sS's'
I18
sS'm'
(lp38737
I1
aI1
aI5
aI15
aI1
aI5
aI97
aI139
aI189
aI39
aI37
aI3513
aI2119
aI1453
aI11477
aI45477
aI75613
aI169915
asS'd'
I19369
sa(dp38738
S'a'
I100294
sS's'
I18
sS'm'
(lp38739
I1
aI3
aI5
aI1
aI15
aI27
aI31
aI87
aI311
aI785
aI489
aI1331
aI5259
aI6963
aI26441
aI41675
aI107187
aI60723
asS'd'
I19370
sa(dp38740
S'a'
I100305
sS's'
I18
sS'm'
(lp38741
I1
aI1
aI3
aI5
aI13
aI9
aI33
aI3
aI273
aI357
aI841
aI1421
aI5993
aI12449
aI6821
aI4283
aI9437
aI215035
asS'd'
I19371
sa(dp38742
S'a'
I100321
sS's'
I18
sS'm'
(lp38743
I1
aI3
aI3
aI1
aI15
aI43
aI101
aI35
aI23
aI743
aI29
aI3953
aI3095
aI14355
aI25977
aI12513
aI54565
aI20199
asS'd'
I19372
sa(dp38744
S'a'
I100331
sS's'
I18
sS'm'
(lp38745
I1
aI3
aI3
aI1
aI11
aI53
aI99
aI95
aI63
aI503
aI1361
aI3231
aI2445
aI5073
aI4667
aI33033
aI4575
aI139475
asS'd'
I19373
sa(dp38746
S'a'
I100334
sS's'
I18
sS'm'
(lp38747
I1
aI3
aI3
aI5
aI19
aI57
aI15
aI115
aI125
aI1017
aI1913
aI907
aI2461
aI3229
aI16591
aI6591
aI26385
aI228661
asS'd'
I19374
sa(dp38748
S'a'
I100345
sS's'
I18
sS'm'
(lp38749
I1
aI3
aI1
aI11
aI27
aI19
aI7
aI121
aI245
aI997
aI1743
aI2571
aI1333
aI9603
aI27811
aI42081
aI44365
aI94943
asS'd'
I19375
sa(dp38750
S'a'
I100346
sS's'
I18
sS'm'
(lp38751
I1
aI3
aI1
aI1
aI13
aI5
aI127
aI217
aI63
aI137
aI989
aI1441
aI1133
aI8273
aI18091
aI22243
aI122931
aI28867
asS'd'
I19376
sa(dp38752
S'a'
I100369
sS's'
I18
sS'm'
(lp38753
I1
aI1
aI7
aI15
aI23
aI39
aI57
aI83
aI321
aI817
aI819
aI223
aI4803
aI6935
aI2027
aI20373
aI119683
aI29781
asS'd'
I19377
sa(dp38754
S'a'
I100372
sS's'
I18
sS'm'
(lp38755
I1
aI1
aI7
aI7
aI9
aI1
aI55
aI211
aI455
aI283
aI1647
aI471
aI4351
aI14119
aI6945
aI63117
aI109687
aI200165
asS'd'
I19378
sa(dp38756
S'a'
I100382
sS's'
I18
sS'm'
(lp38757
I1
aI1
aI7
aI7
aI29
aI45
aI113
aI253
aI135
aI375
aI1091
aI959
aI1423
aI1241
aI31689
aI33751
aI73459
aI91769
asS'd'
I19379
sa(dp38758
S'a'
I100403
sS's'
I18
sS'm'
(lp38759
I1
aI3
aI5
aI11
aI23
aI63
aI55
aI181
aI453
aI267
aI995
aI1309
aI2847
aI3791
aI21683
aI59809
aI81891
aI132635
asS'd'
I19380
sa(dp38760
S'a'
I100451
sS's'
I18
sS'm'
(lp38761
I1
aI1
aI1
aI13
aI17
aI37
aI87
aI17
aI61
aI689
aI1895
aI3877
aI4717
aI6447
aI22329
aI1619
aI30069
aI190887
asS'd'
I19381
sa(dp38762
S'a'
I100488
sS's'
I18
sS'm'
(lp38763
I1
aI3
aI1
aI7
aI27
aI61
aI17
aI51
aI99
aI909
aI85
aI951
aI107
aI1923
aI35
aI63389
aI90273
aI152643
asS'd'
I19382
sa(dp38764
S'a'
I100499
sS's'
I18
sS'm'
(lp38765
I1
aI1
aI1
aI1
aI15
aI39
aI77
aI255
aI191
aI613
aI655
aI1881
aI267
aI3927
aI18025
aI13825
aI15381
aI169193
asS'd'
I19383
sa(dp38766
S'a'
I100501
sS's'
I18
sS'm'
(lp38767
I1
aI3
aI3
aI1
aI31
aI17
aI73
aI69
aI231
aI221
aI501
aI3755
aI1671
aI2049
aI22493
aI16353
aI1775
aI181783
asS'd'
I19384
sa(dp38768
S'a'
I100511
sS's'
I18
sS'm'
(lp38769
I1
aI1
aI7
aI5
aI31
aI45
aI125
aI189
aI287
aI487
aI1911
aI3133
aI3257
aI8967
aI21295
aI1247
aI72297
aI68269
asS'd'
I19385
sa(dp38770
S'a'
I100536
sS's'
I18
sS'm'
(lp38771
I1
aI3
aI3
aI1
aI7
aI9
aI123
aI147
aI7
aI381
aI1597
aI1289
aI7831
aI14493
aI21145
aI15487
aI70353
aI147891
asS'd'
I19386
sa(dp38772
S'a'
I100629
sS's'
I18
sS'm'
(lp38773
I1
aI1
aI5
aI9
aI15
aI17
aI43
aI87
aI101
aI425
aI1819
aI163
aI6741
aI8255
aI2591
aI17719
aI112871
aI110793
asS'd'
I19387
sa(dp38774
S'a'
I100652
sS's'
I18
sS'm'
(lp38775
I1
aI1
aI3
aI9
aI3
aI27
aI7
aI41
aI43
aI743
aI131
aI705
aI2607
aI9963
aI26367
aI41191
aI126347
aI164291
asS'd'
I19388
sa(dp38776
S'a'
I100655
sS's'
I18
sS'm'
(lp38777
I1
aI3
aI3
aI15
aI7
aI17
aI97
aI153
aI283
aI461
aI1723
aI2421
aI4973
aI16369
aI30633
aI62299
aI119425
aI3591
asS'd'
I19389
sa(dp38778
S'a'
I100669
sS's'
I18
sS'm'
(lp38779
I1
aI1
aI1
aI13
aI7
aI33
aI95
aI255
aI429
aI693
aI849
aI3783
aI5985
aI8551
aI23227
aI1015
aI109023
aI42493
asS'd'
I19390
sa(dp38780
S'a'
I100687
sS's'
I18
sS'm'
(lp38781
I1
aI3
aI1
aI7
aI3
aI43
aI53
aI137
aI413
aI265
aI2033
aI1347
aI335
aI529
aI24751
aI16443
aI122195
aI158951
asS'd'
I19391
sa(dp38782
S'a'
I100692
sS's'
I18
sS'm'
(lp38783
I1
aI1
aI5
aI9
aI27
aI21
aI83
aI185
aI325
aI557
aI1247
aI2739
aI6925
aI5459
aI26027
aI62217
aI61113
aI197743
asS'd'
I19392
sa(dp38784
S'a'
I100696
sS's'
I18
sS'm'
(lp38785
I1
aI1
aI3
aI13
aI9
aI57
aI79
aI133
aI137
aI851
aI863
aI1605
aI7839
aI11809
aI29941
aI20473
aI6687
aI164479
asS'd'
I19393
sa(dp38786
S'a'
I100701
sS's'
I18
sS'm'
(lp38787
I1
aI1
aI3
aI9
aI9
aI51
aI123
aI29
aI139
aI43
aI1329
aI2701
aI3413
aI3875
aI19673
aI62369
aI10529
aI226525
asS'd'
I19394
sa(dp38788
S'a'
I100706
sS's'
I18
sS'm'
(lp38789
I1
aI3
aI1
aI5
aI25
aI47
aI7
aI201
aI119
aI623
aI9
aI25
aI1713
aI10817
aI8201
aI5847
aI77477
aI237883
asS'd'
I19395
sa(dp38790
S'a'
I100711
sS's'
I18
sS'm'
(lp38791
I1
aI1
aI5
aI13
aI15
aI53
aI39
aI93
aI235
aI619
aI1695
aI2389
aI2571
aI2389
aI4619
aI45769
aI17245
aI69973
asS'd'
I19396
sa(dp38792
S'a'
I100726
sS's'
I18
sS'm'
(lp38793
I1
aI1
aI7
aI1
aI31
aI9
aI75
aI143
aI1
aI67
aI809
aI1037
aI2801
aI16129
aI22443
aI26021
aI119683
aI14681
asS'd'
I19397
sa(dp38794
S'a'
I100754
sS's'
I18
sS'm'
(lp38795
I1
aI3
aI1
aI9
aI23
aI49
aI69
aI71
aI139
aI953
aI1053
aI3059
aI1061
aI5881
aI26207
aI15907
aI79389
aI95341
asS'd'
I19398
sa(dp38796
S'a'
I100759
sS's'
I18
sS'm'
(lp38797
I1
aI3
aI5
aI1
aI19
aI39
aI69
aI183
aI95
aI289
aI847
aI393
aI1649
aI1275
aI21187
aI34715
aI67553
aI123239
asS'd'
I19399
sa(dp38798
S'a'
I100790
sS's'
I18
sS'm'
(lp38799
I1
aI1
aI1
aI5
aI21
aI39
aI119
aI193
aI347
aI87
aI731
aI3327
aI6089
aI13781
aI20389
aI52303
aI11869
aI48975
asS'd'
I19400
sa(dp38800
S'a'
I100838
sS's'
I18
sS'm'
(lp38801
I1
aI3
aI5
aI1
aI19
aI17
aI93
aI33
aI215
aI457
aI687
aI1325
aI1997
aI2655
aI21195
aI54997
aI36877
aI57991
asS'd'
I19401
sa(dp38802
S'a'
I100878
sS's'
I18
sS'm'
(lp38803
I1
aI1
aI1
aI3
aI17
aI45
aI91
aI45
aI231
aI611
aI413
aI2321
aI7181
aI13765
aI7791
aI6973
aI24497
aI231795
asS'd'
I19402
sa(dp38804
S'a'
I100880
sS's'
I18
sS'm'
(lp38805
I1
aI3
aI3
aI15
aI7
aI29
aI103
aI31
aI295
aI637
aI1775
aI2293
aI8001
aI4175
aI1257
aI16881
aI93695
aI180591
asS'd'
I19403
sa(dp38806
S'a'
I100885
sS's'
I18
sS'm'
(lp38807
I1
aI1
aI1
aI13
aI31
aI7
aI47
aI51
aI267
aI231
aI463
aI939
aI7977
aI8593
aI15329
aI36871
aI50449
aI222341
asS'd'
I19404
sa(dp38808
S'a'
I100886
sS's'
I18
sS'm'
(lp38809
I1
aI3
aI7
aI3
aI27
aI31
aI57
aI135
aI507
aI177
aI1455
aI1963
aI4473
aI6449
aI727
aI49853
aI65275
aI237531
asS'd'
I19405
sa(dp38810
S'a'
I100943
sS's'
I18
sS'm'
(lp38811
I1
aI3
aI7
aI11
aI31
aI21
aI111
aI231
aI269
aI27
aI719
aI3275
aI2489
aI3689
aI3425
aI23763
aI39413
aI64565
asS'd'
I19406
sa(dp38812
S'a'
I100948
sS's'
I18
sS'm'
(lp38813
I1
aI1
aI1
aI11
aI23
aI27
aI31
aI153
aI201
aI985
aI1553
aI3061
aI7663
aI4079
aI13549
aI48765
aI64169
aI68223
asS'd'
I19407
sa(dp38814
S'a'
I100957
sS's'
I18
sS'm'
(lp38815
I1
aI1
aI7
aI15
aI11
aI53
aI125
aI23
aI73
aI799
aI591
aI665
aI127
aI3941
aI11251
aI12649
aI2657
aI230923
asS'd'
I19408
sa(dp38816
S'a'
I100962
sS's'
I18
sS'm'
(lp38817
I1
aI1
aI7
aI15
aI29
aI43
aI95
aI81
aI337
aI367
aI779
aI1237
aI7627
aI997
aI3355
aI1245
aI70477
aI159097
asS'd'
I19409
sa(dp38818
S'a'
I100968
sS's'
I18
sS'm'
(lp38819
I1
aI3
aI1
aI13
aI19
aI31
aI11
aI91
aI179
aI425
aI1395
aI1439
aI2723
aI401
aI26779
aI36723
aI115743
aI179653
asS'd'
I19410
sa(dp38820
S'a'
I100981
sS's'
I18
sS'm'
(lp38821
I1
aI1
aI3
aI1
aI29
aI35
aI11
aI217
aI301
aI421
aI765
aI1949
aI5475
aI2247
aI3953
aI27091
aI16895
aI194821
asS'd'
I19411
sa(dp38822
S'a'
I100986
sS's'
I18
sS'm'
(lp38823
I1
aI1
aI5
aI9
aI11
aI21
aI95
aI135
aI127
aI65
aI609
aI3893
aI7143
aI13231
aI29199
aI36205
aI38711
aI159599
asS'd'
I19412
sa(dp38824
S'a'
I101001
sS's'
I18
sS'm'
(lp38825
I1
aI3
aI5
aI13
aI25
aI25
aI101
aI221
aI495
aI993
aI961
aI2575
aI907
aI5277
aI18415
aI1797
aI22043
aI129889
asS'd'
I19413
sa(dp38826
S'a'
I101012
sS's'
I18
sS'm'
(lp38827
I1
aI1
aI1
aI15
aI21
aI21
aI99
aI3
aI175
aI735
aI659
aI543
aI7573
aI15549
aI14067
aI60009
aI65785
aI112927
asS'd'
I19414
sa(dp38828
S'a'
I101015
sS's'
I18
sS'm'
(lp38829
I1
aI3
aI7
aI7
aI3
aI39
aI61
aI203
aI143
aI581
aI487
aI2097
aI3943
aI6869
aI14435
aI46431
aI101781
aI233067
asS'd'
I19415
sa(dp38830
S'a'
I101019
sS's'
I18
sS'm'
(lp38831
I1
aI1
aI3
aI1
aI13
aI27
aI21
aI147
aI295
aI89
aI1845
aI1017
aI4621
aI10029
aI3517
aI25371
aI104531
aI225179
asS'd'
I19416
sa(dp38832
S'a'
I101037
sS's'
I18
sS'm'
(lp38833
I1
aI1
aI3
aI11
aI25
aI51
aI45
aI179
aI299
aI487
aI2039
aI85
aI4643
aI8211
aI12051
aI64819
aI93481
aI159511
asS'd'
I19417
sa(dp38834
S'a'
I101063
sS's'
I18
sS'm'
(lp38835
I1
aI1
aI5
aI3
aI21
aI7
aI73
aI193
aI415
aI7
aI125
aI2487
aI7369
aI2043
aI7633
aI19265
aI65337
aI57399
asS'd'
I19418
sa(dp38836
S'a'
I101105
sS's'
I18
sS'm'
(lp38837
I1
aI1
aI5
aI3
aI3
aI53
aI51
aI169
aI313
aI973
aI1549
aI243
aI3155
aI13827
aI24971
aI61393
aI15147
aI187397
asS'd'
I19419
sa(dp38838
S'a'
I101144
sS's'
I18
sS'm'
(lp38839
I1
aI3
aI3
aI3
aI29
aI33
aI91
aI41
aI367
aI77
aI1259
aI1703
aI2105
aI14473
aI17763
aI29809
aI47777
aI205553
asS'd'
I19420
sa(dp38840
S'a'
I101147
sS's'
I18
sS'm'
(lp38841
I1
aI3
aI7
aI3
aI23
aI61
aI59
aI235
aI51
aI41
aI417
aI691
aI2953
aI15577
aI32283
aI2793
aI109557
aI64729
asS'd'
I19421
sa(dp38842
S'a'
I101154
sS's'
I18
sS'm'
(lp38843
I1
aI1
aI5
aI9
aI13
aI17
aI93
aI201
aI151
aI323
aI1481
aI3645
aI3039
aI5131
aI503
aI42055
aI114939
aI198755
asS'd'
I19422
sa(dp38844
S'a'
I101163
sS's'
I18
sS'm'
(lp38845
I1
aI3
aI5
aI1
aI21
aI53
aI75
aI29
aI283
aI541
aI499
aI309
aI1923
aI995
aI21479
aI56183
aI103743
aI152113
asS'd'
I19423
sa(dp38846
S'a'
I101166
sS's'
I18
sS'm'
(lp38847
I1
aI3
aI1
aI3
aI21
aI51
aI67
aI97
aI481
aI509
aI805
aI213
aI5157
aI13573
aI16187
aI8199
aI28025
aI208445
asS'd'
I19424
sa(dp38848
S'a'
I101173
sS's'
I18
sS'm'
(lp38849
I1
aI1
aI7
aI7
aI15
aI25
aI107
aI127
aI355
aI249
aI707
aI1287
aI6831
aI5317
aI15613
aI12837
aI48091
aI27611
asS'd'
I19425
sa(dp38850
S'a'
I101174
sS's'
I18
sS'm'
(lp38851
I1
aI3
aI3
aI7
aI31
aI53
aI127
aI239
aI17
aI709
aI979
aI4023
aI7149
aI4239
aI20015
aI44365
aI113245
aI75873
asS'd'
I19426
sa(dp38852
S'a'
I101219
sS's'
I18
sS'm'
(lp38853
I1
aI3
aI5
aI11
aI27
aI37
aI49
aI123
aI137
aI967
aI1857
aI3961
aI7429
aI8355
aI30733
aI64587
aI73903
aI188581
asS'd'
I19427
sa(dp38854
S'a'
I101240
sS's'
I18
sS'm'
(lp38855
I1
aI3
aI3
aI7
aI19
aI51
aI69
aI121
aI345
aI637
aI1987
aI335
aI7071
aI13849
aI22369
aI46895
aI20761
aI148227
asS'd'
I19428
sa(dp38856
S'a'
I101250
sS's'
I18
sS'm'
(lp38857
I1
aI1
aI3
aI11
aI9
aI1
aI93
aI151
aI487
aI889
aI919
aI2429
aI5425
aI15303
aI12583
aI57627
aI42683
aI98265
asS'd'
I19429
sa(dp38858
S'a'
I101259
sS's'
I18
sS'm'
(lp38859
I1
aI3
aI5
aI3
aI1
aI29
aI87
aI189
aI285
aI805
aI933
aI1381
aI2789
aI107
aI14215
aI33171
aI110573
aI250983
asS'd'
I19430
sa(dp38860
S'a'
I101273
sS's'
I18
sS'm'
(lp38861
I1
aI3
aI1
aI1
aI27
aI43
aI63
aI115
aI317
aI401
aI885
aI1029
aI7003
aI10041
aI15299
aI42251
aI58675
aI177545
asS'd'
I19431
sa(dp38862
S'a'
I101274
sS's'
I18
sS'm'
(lp38863
I1
aI3
aI5
aI7
aI11
aI33
aI119
aI5
aI185
aI777
aI1795
aI1585
aI3543
aI1801
aI17681
aI1041
aI44513
aI105435
asS'd'
I19432
sa(dp38864
S'a'
I101285
sS's'
I18
sS'm'
(lp38865
I1
aI3
aI1
aI9
aI19
aI33
aI15
aI253
aI299
aI925
aI241
aI1333
aI615
aI12501
aI499
aI44449
aI16595
aI70357
asS'd'
I19433
sa(dp38866
S'a'
I101300
sS's'
I18
sS'm'
(lp38867
I1
aI1
aI3
aI13
aI13
aI31
aI17
aI123
aI215
aI805
aI1177
aI3683
aI27
aI11881
aI7645
aI25575
aI63057
aI89547
asS'd'
I19434
sa(dp38868
S'a'
I101309
sS's'
I18
sS'm'
(lp38869
I1
aI1
aI5
aI5
aI17
aI1
aI57
aI183
aI267
aI825
aI1987
aI329
aI5603
aI1295
aI425
aI61871
aI27859
aI157109
asS'd'
I19435
sa(dp38870
S'a'
I101310
sS's'
I18
sS'm'
(lp38871
I1
aI3
aI3
aI11
aI19
aI45
aI37
aI79
aI191
aI17
aI17
aI3379
aI7941
aI3159
aI20351
aI26341
aI34687
aI116281
asS'd'
I19436
sa(dp38872
S'a'
I101321
sS's'
I18
sS'm'
(lp38873
I1
aI1
aI3
aI7
aI11
aI7
aI61
aI199
aI459
aI993
aI1729
aI3751
aI1067
aI14965
aI14669
aI40281
aI12579
aI154601
asS'd'
I19437
sa(dp38874
S'a'
I101324
sS's'
I18
sS'm'
(lp38875
I1
aI3
aI3
aI13
aI19
aI53
aI17
aI39
aI137
aI219
aI1645
aI2899
aI505
aI10057
aI22891
aI32317
aI81201
aI126291
asS'd'
I19438
sa(dp38876
S'a'
I101363
sS's'
I18
sS'm'
(lp38877
I1
aI3
aI5
aI7
aI21
aI53
aI65
aI125
aI69
aI781
aI761
aI1683
aI5817
aI11859
aI11543
aI62853
aI57149
aI212261
asS'd'
I19439
sa(dp38878
S'a'
I101366
sS's'
I18
sS'm'
(lp38879
I1
aI3
aI5
aI9
aI13
aI25
aI81
aI119
aI439
aI25
aI239
aI2867
aI421
aI12631
aI22705
aI31039
aI96105
aI79023
asS'd'
I19440
sa(dp38880
S'a'
I101377
sS's'
I18
sS'm'
(lp38881
I1
aI3
aI5
aI3
aI19
aI9
aI51
aI135
aI437
aI393
aI1711
aI1205
aI5195
aI10927
aI28583
aI7513
aI110227
aI139295
asS'd'
I19441
sa(dp38882
S'a'
I101380
sS's'
I18
sS'm'
(lp38883
I1
aI3
aI7
aI9
aI31
aI15
aI5
aI143
aI49
aI13
aI1143
aI2325
aI5437
aI14289
aI31555
aI58777
aI102675
aI64559
asS'd'
I19442
sa(dp38884
S'a'
I101389
sS's'
I18
sS'm'
(lp38885
I1
aI1
aI7
aI3
aI3
aI21
aI65
aI127
aI341
aI109
aI167
aI1835
aI6687
aI1695
aI15631
aI47047
aI127543
aI51413
asS'd'
I19443
sa(dp38886
S'a'
I101407
sS's'
I18
sS'm'
(lp38887
I1
aI1
aI3
aI1
aI31
aI49
aI103
aI147
aI59
aI701
aI1251
aI3391
aI1935
aI5371
aI28585
aI50023
aI73839
aI118205
asS'd'
I19444
sa(dp38888
S'a'
I101411
sS's'
I18
sS'm'
(lp38889
I1
aI3
aI3
aI1
aI5
aI49
aI91
aI23
aI91
aI663
aI1369
aI1437
aI2657
aI11369
aI29857
aI53875
aI127045
aI242323
asS'd'
I19445
sa(dp38890
S'a'
I101432
sS's'
I18
sS'm'
(lp38891
I1
aI3
aI1
aI3
aI23
aI33
aI7
aI105
aI353
aI863
aI1211
aI1175
aI1347
aI12913
aI11973
aI55255
aI27145
aI175539
asS'd'
I19446
sa(dp38892
S'a'
I101476
sS's'
I18
sS'm'
(lp38893
I1
aI1
aI1
aI13
aI9
aI29
aI71
aI71
aI509
aI571
aI2005
aI3125
aI2731
aI6829
aI26863
aI16459
aI113195
aI80247
asS'd'
I19447
sa(dp38894
S'a'
I101488
sS's'
I18
sS'm'
(lp38895
I1
aI3
aI5
aI1
aI7
aI11
aI45
aI177
aI281
aI695
aI1197
aI2035
aI2137
aI11833
aI12417
aI5805
aI77211
aI45553
asS'd'
I19448
sa(dp38896
S'a'
I101494
sS's'
I18
sS'm'
(lp38897
I1
aI1
aI7
aI7
aI27
aI59
aI91
aI183
aI267
aI373
aI677
aI2431
aI903
aI4229
aI31997
aI19843
aI125089
aI242871
asS'd'
I19449
sa(dp38898
S'a'
I101514
sS's'
I18
sS'm'
(lp38899
I1
aI3
aI3
aI13
aI23
aI23
aI53
aI101
aI225
aI247
aI2013
aI853
aI279
aI2161
aI30045
aI28255
aI130907
aI57157
asS'd'
I19450
sa(dp38900
S'a'
I101534
sS's'
I18
sS'm'
(lp38901
I1
aI3
aI5
aI13
aI5
aI11
aI63
aI137
aI219
aI661
aI773
aI1991
aI4081
aI5963
aI25207
aI50461
aI85293
aI159441
asS'd'
I19451
sa(dp38902
S'a'
I101537
sS's'
I18
sS'm'
(lp38903
I1
aI3
aI1
aI15
aI5
aI59
aI43
aI87
aI429
aI77
aI73
aI1275
aI2619
aI16133
aI20009
aI26089
aI38129
aI157267
asS'd'
I19452
sa(dp38904
S'a'
I101543
sS's'
I18
sS'm'
(lp38905
I1
aI1
aI5
aI15
aI9
aI59
aI37
aI127
aI127
aI733
aI1703
aI1331
aI4293
aI3555
aI22739
aI49513
aI34533
aI143225
asS'd'
I19453
sa(dp38906
S'a'
I101547
sS's'
I18
sS'm'
(lp38907
I1
aI1
aI7
aI7
aI11
aI51
aI121
aI59
aI505
aI147
aI1855
aI1661
aI5539
aI3421
aI28863
aI14811
aI39811
aI228927
asS'd'
I19454
sa(dp38908
S'a'
I101584
sS's'
I18
sS'm'
(lp38909
I1
aI1
aI5
aI5
aI1
aI31
aI57
aI167
aI107
aI753
aI1835
aI2491
aI3311
aI8639
aI8743
aI17279
aI77071
aI8673
asS'd'
I19455
sa(dp38910
S'a'
I101589
sS's'
I18
sS'm'
(lp38911
I1
aI1
aI3
aI3
aI15
aI39
aI1
aI223
aI395
aI603
aI1095
aI435
aI1225
aI4045
aI21721
aI40767
aI48971
aI1583
asS'd'
I19456
sa(dp38912
S'a'
I101590
sS's'
I18
sS'm'
(lp38913
I1
aI1
aI3
aI13
aI9
aI17
aI47
aI175
aI229
aI827
aI769
aI2901
aI137
aI9931
aI11115
aI25337
aI105811
aI68413
asS'd'
I19457
sa(dp38914
S'a'
I101629
sS's'
I18
sS'm'
(lp38915
I1
aI1
aI1
aI5
aI29
aI55
aI43
aI39
aI319
aI919
aI749
aI931
aI1973
aI13147
aI913
aI48353
aI40955
aI189783
asS'd'
I19458
sa(dp38916
S'a'
I101678
sS's'
I18
sS'm'
(lp38917
I1
aI3
aI5
aI5
aI13
aI33
aI81
aI213
aI79
aI781
aI1069
aI3117
aI5859
aI9061
aI30963
aI17307
aI6281
aI208707
asS'd'
I19459
sa(dp38918
S'a'
I101683
sS's'
I18
sS'm'
(lp38919
I1
aI1
aI5
aI13
aI27
aI63
aI41
aI91
aI123
aI763
aI1115
aI3193
aI4571
aI4573
aI8235
aI24291
aI40911
aI225985
asS'd'
I19460
sa(dp38920
S'a'
I101700
sS's'
I18
sS'm'
(lp38921
I1
aI1
aI1
aI7
aI13
aI5
aI77
aI215
aI67
aI183
aI1447
aI1571
aI213
aI3481
aI28349
aI22451
aI44951
aI240257
asS'd'
I19461
sa(dp38922
S'a'
I101712
sS's'
I18
sS'm'
(lp38923
I1
aI3
aI7
aI7
aI3
aI21
aI39
aI233
aI263
aI19
aI489
aI1511
aI2313
aI1799
aI25173
aI17913
aI70593
aI2969
asS'd'
I19462
sa(dp38924
S'a'
I101748
sS's'
I18
sS'm'
(lp38925
I1
aI3
aI7
aI15
aI3
aI1
aI93
aI101
aI393
aI911
aI1055
aI953
aI1279
aI11947
aI11963
aI53443
aI29105
aI226057
asS'd'
I19463
sa(dp38926
S'a'
I101757
sS's'
I18
sS'm'
(lp38927
I1
aI1
aI3
aI7
aI7
aI53
aI119
aI39
aI439
aI629
aI1845
aI3411
aI3633
aI16345
aI27501
aI59565
aI39581
aI85373
asS'd'
I19464
sa(dp38928
S'a'
I101762
sS's'
I18
sS'm'
(lp38929
I1
aI1
aI5
aI7
aI19
aI39
aI19
aI191
aI509
aI239
aI359
aI645
aI8107
aI13721
aI21289
aI20763
aI67727
aI45957
asS'd'
I19465
sa(dp38930
S'a'
I101771
sS's'
I18
sS'm'
(lp38931
I1
aI1
aI1
aI13
aI5
aI13
aI113
aI41
aI135
aI351
aI311
aI1099
aI2391
aI16165
aI15805
aI54737
aI102645
aI224417
asS'd'
I19466
sa(dp38932
S'a'
I101785
sS's'
I18
sS'm'
(lp38933
I1
aI3
aI3
aI9
aI17
aI47
aI95
aI249
aI45
aI749
aI313
aI317
aI2413
aI15861
aI27221
aI35537
aI6407
aI50111
asS'd'
I19467
sa(dp38934
S'a'
I101795
sS's'
I18
sS'm'
(lp38935
I1
aI1
aI7
aI5
aI29
aI57
aI81
aI7
aI233
aI393
aI307
aI1089
aI1367
aI12275
aI11861
aI29119
aI27271
aI36351
asS'd'
I19468
sa(dp38936
S'a'
I101822
sS's'
I18
sS'm'
(lp38937
I1
aI3
aI5
aI15
aI9
aI15
aI23
aI241
aI233
aI305
aI1025
aI2035
aI4897
aI10321
aI17345
aI42873
aI109045
aI129533
asS'd'
I19469
sa(dp38938
S'a'
I101851
sS's'
I18
sS'm'
(lp38939
I1
aI3
aI5
aI9
aI21
aI33
aI111
aI81
aI311
aI829
aI1851
aI1437
aI5935
aI7847
aI15493
aI28531
aI74879
aI40567
asS'd'
I19470
sa(dp38940
S'a'
I101863
sS's'
I18
sS'm'
(lp38941
I1
aI3
aI7
aI7
aI13
aI21
aI43
aI119
aI507
aI701
aI1385
aI745
aI799
aI1567
aI13271
aI40267
aI130843
aI59951
asS'd'
I19471
sa(dp38942
S'a'
I101894
sS's'
I18
sS'm'
(lp38943
I1
aI3
aI3
aI1
aI19
aI43
aI45
aI119
aI87
aI263
aI1475
aI3897
aI2811
aI2711
aI4949
aI48043
aI125237
aI230897
asS'd'
I19472
sa(dp38944
S'a'
I101903
sS's'
I18
sS'm'
(lp38945
I1
aI1
aI7
aI1
aI1
aI53
aI17
aI71
aI313
aI373
aI5
aI2359
aI1643
aI9867
aI18365
aI5011
aI40675
aI105371
asS'd'
I19473
sa(dp38946
S'a'
I101924
sS's'
I18
sS'm'
(lp38947
I1
aI3
aI7
aI11
aI23
aI49
aI123
aI255
aI33
aI241
aI473
aI959
aI1859
aI6109
aI5433
aI27625
aI46839
aI90727
asS'd'
I19474
sa(dp38948
S'a'
I101939
sS's'
I18
sS'm'
(lp38949
I1
aI1
aI1
aI3
aI31
aI43
aI33
aI129
aI159
aI445
aI1831
aI1005
aI587
aI2091
aI5749
aI33271
aI50909
aI65057
asS'd'
I19475
sa(dp38950
S'a'
I101980
sS's'
I18
sS'm'
(lp38951
I1
aI1
aI7
aI13
aI3
aI21
aI17
aI125
aI357
aI97
aI1255
aI669
aI1583
aI7433
aI32287
aI61795
aI5915
aI211593
asS'd'
I19476
sa(dp38952
S'a'
I101993
sS's'
I18
sS'm'
(lp38953
I1
aI1
aI7
aI7
aI31
aI19
aI71
aI211
aI213
aI731
aI1491
aI3315
aI3633
aI14953
aI21369
aI4977
aI33533
aI12115
asS'd'
I19477
sa(dp38954
S'a'
I101994
sS's'
I18
sS'm'
(lp38955
I1
aI3
aI1
aI1
aI31
aI1
aI87
aI253
aI211
aI57
aI1431
aI2613
aI4075
aI14463
aI11287
aI38671
aI100129
aI4241
asS'd'
I19478
sa(dp38956
S'a'
I101996
sS's'
I18
sS'm'
(lp38957
I1
aI1
aI1
aI5
aI21
aI13
aI27
aI29
aI31
aI285
aI179
aI3861
aI5319
aI15683
aI2579
aI15663
aI63357
aI81849
asS'd'
I19479
sa(dp38958
S'a'
I102007
sS's'
I18
sS'm'
(lp38959
I1
aI1
aI3
aI7
aI23
aI29
aI29
aI79
aI263
aI865
aI1237
aI3871
aI2097
aI5337
aI2387
aI59277
aI28831
aI57957
asS'd'
I19480
sa(dp38960
S'a'
I102044
sS's'
I18
sS'm'
(lp38961
I1
aI3
aI7
aI15
aI15
aI31
aI83
aI195
aI87
aI691
aI71
aI1025
aI3145
aI675
aI14619
aI22399
aI88885
aI222969
asS'd'
I19481
sa(dp38962
S'a'
I102086
sS's'
I18
sS'm'
(lp38963
I1
aI1
aI5
aI7
aI19
aI3
aI121
aI105
aI383
aI675
aI777
aI2073
aI643
aI14439
aI19467
aI13159
aI115421
aI250561
asS'd'
I19482
sa(dp38964
S'a'
I102097
sS's'
I18
sS'm'
(lp38965
I1
aI3
aI5
aI9
aI13
aI41
aI119
aI23
aI355
aI765
aI579
aI849
aI3313
aI2443
aI29521
aI42965
aI102559
aI227707
asS'd'
I19483
sa(dp38966
S'a'
I102113
sS's'
I18
sS'm'
(lp38967
I1
aI3
aI3
aI7
aI23
aI55
aI115
aI83
aI331
aI873
aI797
aI621
aI1197
aI15299
aI26307
aI34287
aI120459
aI260603
asS'd'
I19484
sa(dp38968
S'a'
I102123
sS's'
I18
sS'm'
(lp38969
I1
aI1
aI3
aI5
aI7
aI11
aI113
aI119
aI65
aI691
aI1169
aI2291
aI7283
aI1391
aI10737
aI3801
aI40649
aI191009
asS'd'
I19485
sa(dp38970
S'a'
I102155
sS's'
I18
sS'm'
(lp38971
I1
aI1
aI7
aI9
aI3
aI23
aI95
aI109
aI457
aI277
aI871
aI3013
aI2833
aI4131
aI21081
aI43635
aI19875
aI162173
asS'd'
I19486
sa(dp38972
S'a'
I102157
sS's'
I18
sS'm'
(lp38973
I1
aI1
aI3
aI15
aI25
aI7
aI45
aI109
aI195
aI935
aI1487
aI1603
aI1663
aI15595
aI10687
aI4073
aI34863
aI45851
asS'd'
I19487
sa(dp38974
S'a'
I102169
sS's'
I18
sS'm'
(lp38975
I1
aI1
aI1
aI5
aI21
aI19
aI37
aI141
aI377
aI553
aI1225
aI2485
aI1235
aI13179
aI2587
aI43659
aI1405
aI52023
asS'd'
I19488
sa(dp38976
S'a'
I102181
sS's'
I18
sS'm'
(lp38977
I1
aI3
aI1
aI9
aI5
aI27
aI1
aI219
aI397
aI555
aI533
aI941
aI2755
aI1295
aI16157
aI30733
aI54279
aI168455
asS'd'
I19489
sa(dp38978
S'a'
I102182
sS's'
I18
sS'm'
(lp38979
I1
aI3
aI3
aI1
aI27
aI41
aI93
aI177
aI119
aI581
aI167
aI3943
aI5765
aI15455
aI555
aI17419
aI33117
aI160599
asS'd'
I19490
sa(dp38980
S'a'
I102185
sS's'
I18
sS'm'
(lp38981
I1
aI1
aI3
aI7
aI15
aI25
aI29
aI177
aI503
aI529
aI229
aI2535
aI1493
aI805
aI30983
aI26309
aI123453
aI118441
asS'd'
I19491
sa(dp38982
S'a'
I102199
sS's'
I18
sS'm'
(lp38983
I1
aI1
aI7
aI15
aI13
aI33
aI83
aI151
aI183
aI433
aI823
aI4003
aI1991
aI6563
aI18743
aI61835
aI56535
aI191193
asS'd'
I19492
sa(dp38984
S'a'
I102208
sS's'
I18
sS'm'
(lp38985
I1
aI1
aI1
aI13
aI9
aI31
aI91
aI63
aI455
aI67
aI243
aI1573
aI3507
aI9491
aI4677
aI13835
aI121603
aI241781
asS'd'
I19493
sa(dp38986
S'a'
I102213
sS's'
I18
sS'm'
(lp38987
I1
aI1
aI5
aI5
aI29
aI13
aI45
aI29
aI289
aI909
aI1923
aI3371
aI3675
aI13119
aI24599
aI58511
aI109467
aI126865
asS'd'
I19494
sa(dp38988
S'a'
I102226
sS's'
I18
sS'm'
(lp38989
I1
aI1
aI5
aI9
aI21
aI41
aI41
aI179
aI25
aI463
aI949
aI819
aI397
aI12329
aI3461
aI34773
aI61337
aI23579
asS'd'
I19495
sa(dp38990
S'a'
I102290
sS's'
I18
sS'm'
(lp38991
I1
aI3
aI3
aI11
aI1
aI25
aI17
aI165
aI77
aI455
aI1983
aI143
aI2763
aI8165
aI14849
aI1601
aI30093
aI54599
asS'd'
I19496
sa(dp38992
S'a'
I102296
sS's'
I18
sS'm'
(lp38993
I1
aI1
aI7
aI3
aI7
aI1
aI117
aI117
aI53
aI611
aI1405
aI3357
aI1717
aI8157
aI247
aI17501
aI30201
aI192485
asS'd'
I19497
sa(dp38994
S'a'
I102306
sS's'
I18
sS'm'
(lp38995
I1
aI1
aI1
aI5
aI23
aI47
aI95
aI173
aI133
aI603
aI1627
aI4039
aI5599
aI10575
aI30275
aI60287
aI3313
aI77551
asS'd'
I19498
sa(dp38996
S'a'
I102317
sS's'
I18
sS'm'
(lp38997
I1
aI1
aI1
aI9
aI31
aI49
aI87
aI169
aI221
aI561
aI1045
aI365
aI1955
aI9905
aI13057
aI33717
aI91203
aI57513
asS'd'
I19499
sa(dp38998
S'a'
I102337
sS's'
I18
sS'm'
(lp38999
I1
aI3
aI3
aI1
aI19
aI25
aI113
aI231
aI415
aI717
aI1767
aI755
aI4825
aI7541
aI10121
aI9351
aI72093
aI255877
asS'd'
I19500
sa(dp39000
S'a'
I102344
sS's'
I18
sS'm'
(lp39001
I1
aI3
aI1
aI7
aI1
aI35
aI3
aI231
aI345
aI375
aI809
aI2753
aI849
aI13915
aI16127
aI7575
aI45259
aI28917
asS'd'
I19501
sa(dp39002
S'a'
I102350
sS's'
I18
sS'm'
(lp39003
I1
aI3
aI3
aI11
aI23
aI43
aI13
aI37
aI305
aI765
aI517
aI333
aI473
aI14949
aI11939
aI35171
aI63963
aI181825
asS'd'
I19502
sa(dp39004
S'a'
I102378
sS's'
I18
sS'm'
(lp39005
I1
aI3
aI3
aI3
aI5
aI43
aI35
aI89
aI293
aI913
aI1325
aI2097
aI603
aI14365
aI4005
aI38935
aI23837
aI34271
asS'd'
I19503
sa(dp39006
S'a'
I102383
sS's'
I18
sS'm'
(lp39007
I1
aI1
aI3
aI15
aI1
aI47
aI67
aI199
aI167
aI909
aI1167
aI1513
aI7087
aI5017
aI23469
aI2621
aI24961
aI226679
asS'd'
I19504
sa(dp39008
S'a'
I102391
sS's'
I18
sS'm'
(lp39009
I1
aI3
aI7
aI15
aI23
aI37
aI111
aI217
aI33
aI423
aI457
aI1767
aI4821
aI10233
aI27045
aI33397
aI85351
aI156751
asS'd'
I19505
sa(dp39010
S'a'
I102400
sS's'
I18
sS'm'
(lp39011
I1
aI3
aI3
aI7
aI31
aI25
aI67
aI65
aI291
aI169
aI1505
aI1707
aI4833
aI11541
aI1189
aI62959
aI59831
aI48729
asS'd'
I19506
sa(dp39012
S'a'
I102409
sS's'
I18
sS'm'
(lp39013
I1
aI1
aI5
aI7
aI27
aI27
aI15
aI23
aI189
aI819
aI709
aI3591
aI2781
aI14807
aI20303
aI27795
aI88349
aI210251
asS'd'
I19507
sa(dp39014
S'a'
I102427
sS's'
I18
sS'm'
(lp39015
I1
aI3
aI7
aI15
aI1
aI19
aI49
aI103
aI219
aI85
aI1281
aI3981
aI7229
aI10427
aI11689
aI1547
aI90747
aI12283
asS'd'
I19508
sa(dp39016
S'a'
I102434
sS's'
I18
sS'm'
(lp39017
I1
aI1
aI1
aI13
aI31
aI49
aI81
aI121
aI287
aI851
aI333
aI353
aI7373
aI10165
aI1157
aI11713
aI89445
aI210709
asS'd'
I19509
sa(dp39018
S'a'
I102453
sS's'
I18
sS'm'
(lp39019
I1
aI3
aI7
aI5
aI31
aI35
aI3
aI229
aI443
aI973
aI1971
aI1861
aI5695
aI6725
aI6405
aI45927
aI115313
aI228667
asS'd'
I19510
sa(dp39020
S'a'
I102472
sS's'
I18
sS'm'
(lp39021
I1
aI3
aI3
aI11
aI29
aI17
aI105
aI203
aI69
aI945
aI1239
aI3213
aI6005
aI10095
aI31233
aI37421
aI62911
aI91371
asS'd'
I19511
sa(dp39022
S'a'
I102502
sS's'
I18
sS'm'
(lp39023
I1
aI1
aI7
aI1
aI21
aI59
aI7
aI43
aI391
aI299
aI1225
aI283
aI3351
aI11495
aI25071
aI16619
aI65127
aI114033
asS'd'
I19512
sa(dp39024
S'a'
I102525
sS's'
I18
sS'm'
(lp39025
I1
aI1
aI3
aI9
aI17
aI11
aI71
aI73
aI377
aI437
aI311
aI1083
aI6941
aI1039
aI1047
aI55647
aI21239
aI209201
asS'd'
I19513
sa(dp39026
S'a'
I102547
sS's'
I18
sS'm'
(lp39027
I1
aI3
aI7
aI5
aI29
aI51
aI113
aI163
aI215
aI511
aI615
aI2427
aI2747
aI14389
aI1005
aI27015
aI31809
aI30603
asS'd'
I19514
sa(dp39028
S'a'
I102554
sS's'
I18
sS'm'
(lp39029
I1
aI1
aI1
aI3
aI9
aI61
aI9
aI201
aI259
aI411
aI175
aI1003
aI401
aI13695
aI13103
aI3075
aI43695
aI177139
asS'd'
I19515
sa(dp39030
S'a'
I102560
sS's'
I18
sS'm'
(lp39031
I1
aI1
aI5
aI3
aI17
aI47
aI125
aI173
aI277
aI17
aI619
aI2295
aI3091
aI5615
aI4529
aI54065
aI81875
aI97279
asS'd'
I19516
sa(dp39032
S'a'
I102578
sS's'
I18
sS'm'
(lp39033
I1
aI3
aI3
aI9
aI3
aI61
aI97
aI151
aI287
aI671
aI1439
aI1129
aI6343
aI8161
aI24593
aI40371
aI109705
aI106279
asS'd'
I19517
sa(dp39034
S'a'
I102580
sS's'
I18
sS'm'
(lp39035
I1
aI3
aI3
aI1
aI5
aI61
aI21
aI31
aI41
aI855
aI1541
aI3351
aI575
aI3195
aI17155
aI46913
aI1797
aI146401
asS'd'
I19518
sa(dp39036
S'a'
I102597
sS's'
I18
sS'm'
(lp39037
I1
aI1
aI7
aI13
aI11
aI43
aI27
aI97
aI479
aI117
aI1873
aI3103
aI1947
aI9273
aI29171
aI35719
aI10601
aI209629
asS'd'
I19519
sa(dp39038
S'a'
I102616
sS's'
I18
sS'm'
(lp39039
I1
aI3
aI1
aI9
aI21
aI47
aI53
aI129
aI85
aI505
aI165
aI3437
aI5687
aI10289
aI6615
aI46719
aI50731
aI25271
asS'd'
I19520
sa(dp39040
S'a'
I102625
sS's'
I18
sS'm'
(lp39041
I1
aI1
aI3
aI1
aI1
aI33
aI37
aI239
aI45
aI565
aI1907
aI3831
aI2177
aI10967
aI12689
aI49395
aI36289
aI247507
asS'd'
I19521
sa(dp39042
S'a'
I102632
sS's'
I18
sS'm'
(lp39043
I1
aI1
aI5
aI1
aI19
aI57
aI75
aI245
aI59
aI967
aI1319
aI3971
aI5267
aI11713
aI15417
aI60503
aI63431
aI157267
asS'd'
I19522
sa(dp39044
S'a'
I102663
sS's'
I18
sS'm'
(lp39045
I1
aI1
aI5
aI7
aI7
aI15
aI99
aI87
aI331
aI261
aI1973
aI219
aI3063
aI4071
aI19273
aI48637
aI128089
aI55511
asS'd'
I19523
sa(dp39046
S'a'
I102669
sS's'
I18
sS'm'
(lp39047
I1
aI3
aI5
aI13
aI23
aI43
aI7
aI171
aI173
aI1023
aI1145
aI3551
aI5127
aI6365
aI18013
aI1613
aI51997
aI107265
asS'd'
I19524
sa(dp39048
S'a'
I102681
sS's'
I18
sS'm'
(lp39049
I1
aI3
aI5
aI3
aI7
aI7
aI23
aI195
aI251
aI387
aI1889
aI3645
aI4221
aI6025
aI27291
aI24831
aI123749
aI231403
asS'd'
I19525
sa(dp39050
S'a'
I102694
sS's'
I18
sS'm'
(lp39051
I1
aI3
aI1
aI1
aI29
aI49
aI107
aI215
aI93
aI211
aI1349
aI1925
aI7149
aI1015
aI27465
aI34139
aI126149
aI121349
asS'd'
I19526
sa(dp39052
S'a'
I102706
sS's'
I18
sS'm'
(lp39053
I1
aI1
aI5
aI5
aI19
aI15
aI121
aI189
aI167
aI801
aI483
aI1955
aI8031
aI4749
aI26665
aI64791
aI18671
aI123221
asS'd'
I19527
sa(dp39054
S'a'
I102715
sS's'
I18
sS'm'
(lp39055
I1
aI3
aI3
aI15
aI23
aI57
aI15
aI249
aI197
aI103
aI2021
aI1897
aI5975
aI12821
aI6441
aI62719
aI81415
aI232417
asS'd'
I19528
sa(dp39056
S'a'
I102723
sS's'
I18
sS'm'
(lp39057
I1
aI1
aI3
aI9
aI17
aI45
aI111
aI225
aI103
aI121
aI1259
aI2849
aI2235
aI2041
aI13261
aI7929
aI76325
aI38677
asS'd'
I19529
sa(dp39058
S'a'
I102730
sS's'
I18
sS'm'
(lp39059
I1
aI1
aI3
aI9
aI25
aI17
aI63
aI63
aI369
aI387
aI31
aI1423
aI5699
aI12519
aI27407
aI53193
aI963
aI123473
asS'd'
I19530
sa(dp39060
S'a'
I102743
sS's'
I18
sS'm'
(lp39061
I1
aI3
aI3
aI13
aI3
aI59
aI59
aI3
aI367
aI357
aI1391
aI1519
aI2619
aI2291
aI1349
aI28275
aI21655
aI8763
asS'd'
I19531
sa(dp39062
S'a'
I102856
sS's'
I18
sS'm'
(lp39063
I1
aI1
aI7
aI1
aI31
aI51
aI121
aI49
aI157
aI509
aI1513
aI3247
aI3439
aI8691
aI20729
aI17585
aI49013
aI228695
asS'd'
I19532
sa(dp39064
S'a'
I102870
sS's'
I18
sS'm'
(lp39065
I1
aI1
aI5
aI9
aI11
aI29
aI103
aI49
aI209
aI35
aI673
aI1409
aI2483
aI4561
aI12435
aI46139
aI31019
aI50929
asS'd'
I19533
sa(dp39066
S'a'
I102898
sS's'
I18
sS'm'
(lp39067
I1
aI1
aI3
aI9
aI11
aI45
aI77
aI143
aI277
aI769
aI1151
aI3705
aI7901
aI3735
aI31155
aI46135
aI84061
aI254357
asS'd'
I19534
sa(dp39068
S'a'
I102926
sS's'
I18
sS'm'
(lp39069
I1
aI3
aI3
aI3
aI5
aI51
aI95
aI7
aI155
aI121
aI1899
aI2261
aI2915
aI6637
aI6557
aI20535
aI20937
aI257275
asS'd'
I19535
sa(dp39070
S'a'
I102931
sS's'
I18
sS'm'
(lp39071
I1
aI1
aI7
aI9
aI13
aI49
aI125
aI135
aI7
aI535
aI1171
aI3501
aI1701
aI5791
aI10121
aI9845
aI21645
aI56451
asS'd'
I19536
sa(dp39072
S'a'
I102938
sS's'
I18
sS'm'
(lp39073
I1
aI1
aI7
aI5
aI9
aI1
aI115
aI27
aI229
aI813
aI133
aI753
aI1959
aI13121
aI30425
aI31059
aI114619
aI132257
asS'd'
I19537
sa(dp39074
S'a'
I102947
sS's'
I18
sS'm'
(lp39075
I1
aI1
aI5
aI3
aI23
aI25
aI41
aI165
aI21
aI551
aI1751
aI3563
aI731
aI15811
aI14777
aI22327
aI82853
aI116699
asS'd'
I19538
sa(dp39076
S'a'
I102968
sS's'
I18
sS'm'
(lp39077
I1
aI3
aI1
aI9
aI3
aI25
aI95
aI211
aI457
aI25
aI349
aI921
aI213
aI5721
aI26725
aI19541
aI102473
aI200845
asS'd'
I19539
sa(dp39078
S'a'
I102976
sS's'
I18
sS'm'
(lp39079
I1
aI3
aI1
aI13
aI5
aI1
aI31
aI49
aI493
aI785
aI61
aI2603
aI5997
aI12545
aI9793
aI32521
aI99859
aI256105
asS'd'
I19540
sa(dp39080
S'a'
I102993
sS's'
I18
sS'm'
(lp39081
I1
aI3
aI3
aI7
aI31
aI19
aI31
aI121
aI507
aI79
aI685
aI197
aI4027
aI12909
aI30533
aI38427
aI38993
aI14581
asS'd'
I19541
sa(dp39082
S'a'
I103005
sS's'
I18
sS'm'
(lp39083
I1
aI3
aI7
aI13
aI13
aI23
aI63
aI15
aI223
aI25
aI1957
aI2527
aI6061
aI11753
aI4835
aI34553
aI45579
aI205805
asS'd'
I19542
sa(dp39084
S'a'
I103019
sS's'
I18
sS'm'
(lp39085
I1
aI3
aI3
aI3
aI5
aI39
aI55
aI123
aI275
aI185
aI749
aI3227
aI8137
aI12959
aI7243
aI20613
aI46247
aI106127
asS'd'
I19543
sa(dp39086
S'a'
I103022
sS's'
I18
sS'm'
(lp39087
I1
aI1
aI7
aI13
aI1
aI21
aI49
aI145
aI237
aI291
aI1721
aI2981
aI267
aI1255
aI21817
aI39553
aI21937
aI115307
asS'd'
I19544
sa(dp39088
S'a'
I103057
sS's'
I18
sS'm'
(lp39089
I1
aI1
aI1
aI15
aI5
aI63
aI3
aI201
aI117
aI991
aI1049
aI1975
aI5117
aI5799
aI28211
aI37907
aI46799
aI240847
asS'd'
I19545
sa(dp39090
S'a'
I103080
sS's'
I18
sS'm'
(lp39091
I1
aI3
aI3
aI5
aI25
aI15
aI29
aI111
aI201
aI857
aI319
aI2695
aI4251
aI4303
aI28495
aI12481
aI31979
aI107503
asS'd'
I19546
sa(dp39092
S'a'
I103100
sS's'
I18
sS'm'
(lp39093
I1
aI1
aI3
aI11
aI3
aI27
aI81
aI55
aI489
aI983
aI293
aI3181
aI4593
aI4759
aI18437
aI51185
aI47701
aI75469
asS'd'
I19547
sa(dp39094
S'a'
I103106
sS's'
I18
sS'm'
(lp39095
I1
aI3
aI1
aI3
aI11
aI25
aI113
aI243
aI365
aI299
aI1717
aI561
aI5173
aI5983
aI7453
aI33563
aI98519
aI162451
asS'd'
I19548
sa(dp39096
S'a'
I103126
sS's'
I18
sS'm'
(lp39097
I1
aI3
aI1
aI5
aI1
aI21
aI3
aI87
aI267
aI483
aI1397
aI791
aI4807
aI4649
aI13713
aI19861
aI110471
aI51443
asS'd'
I19549
sa(dp39098
S'a'
I103173
sS's'
I18
sS'm'
(lp39099
I1
aI3
aI7
aI5
aI31
aI9
aI13
aI251
aI351
aI609
aI841
aI3267
aI4201
aI8771
aI6673
aI44867
aI105221
aI189399
asS'd'
I19550
sa(dp39100
S'a'
I103202
sS's'
I18
sS'm'
(lp39101
I1
aI1
aI3
aI1
aI13
aI43
aI47
aI153
aI331
aI1013
aI705
aI1867
aI563
aI6361
aI7407
aI46243
aI30521
aI213831
asS'd'
I19551
sa(dp39102
S'a'
I103207
sS's'
I18
sS'm'
(lp39103
I1
aI3
aI3
aI7
aI29
aI41
aI3
aI179
aI319
aI877
aI905
aI3803
aI2775
aI9729
aI5673
aI9521
aI117887
aI37605
asS'd'
I19552
sa(dp39104
S'a'
I103219
sS's'
I18
sS'm'
(lp39105
I1
aI1
aI5
aI5
aI11
aI49
aI111
aI195
aI467
aI931
aI849
aI2785
aI7829
aI2291
aI29031
aI52019
aI86493
aI213971
asS'd'
I19553
sa(dp39106
S'a'
I103226
sS's'
I18
sS'm'
(lp39107
I1
aI1
aI3
aI11
aI13
aI11
aI81
aI81
aI419
aI621
aI181
aI1203
aI7305
aI7857
aI16885
aI2531
aI53127
aI35553
asS'd'
I19554
sa(dp39108
S'a'
I103228
sS's'
I18
sS'm'
(lp39109
I1
aI1
aI5
aI1
aI19
aI35
aI75
aI131
aI159
aI211
aI319
aI2805
aI6497
aI14759
aI28997
aI62347
aI4013
aI233821
asS'd'
I19555
sa(dp39110
S'a'
I103248
sS's'
I18
sS'm'
(lp39111
I1
aI3
aI3
aI15
aI1
aI55
aI107
aI105
aI35
aI369
aI1259
aI665
aI6717
aI2555
aI7149
aI10373
aI33153
aI105915
asS'd'
I19556
sa(dp39112
S'a'
I103274
sS's'
I18
sS'm'
(lp39113
I1
aI3
aI7
aI5
aI31
aI13
aI27
aI207
aI359
aI355
aI2047
aI1777
aI5555
aI12659
aI30547
aI3655
aI86189
aI961
asS'd'
I19557
sa(dp39114
S'a'
I103288
sS's'
I18
sS'm'
(lp39115
I1
aI3
aI7
aI9
aI9
aI55
aI7
aI117
aI57
aI115
aI745
aI501
aI2341
aI3899
aI8229
aI10625
aI66905
aI187959
asS'd'
I19558
sa(dp39116
S'a'
I103291
sS's'
I18
sS'm'
(lp39117
I1
aI3
aI3
aI13
aI25
aI43
aI87
aI197
aI303
aI405
aI1459
aI3385
aI4109
aI8325
aI24747
aI18405
aI48845
aI96673
asS'd'
I19559
sa(dp39118
S'a'
I103293
sS's'
I18
sS'm'
(lp39119
I1
aI3
aI5
aI3
aI21
aI37
aI109
aI9
aI183
aI585
aI1287
aI3851
aI4939
aI1057
aI19489
aI42603
aI32447
aI117957
asS'd'
I19560
sa(dp39120
S'a'
I103324
sS's'
I18
sS'm'
(lp39121
I1
aI1
aI5
aI13
aI3
aI47
aI109
aI115
aI73
aI733
aI1189
aI3773
aI7471
aI10339
aI5093
aI50253
aI122665
aI254381
asS'd'
I19561
sa(dp39122
S'a'
I103358
sS's'
I18
sS'm'
(lp39123
I1
aI1
aI5
aI5
aI21
aI33
aI113
aI187
aI51
aI245
aI241
aI3887
aI4075
aI11945
aI20883
aI18817
aI43567
aI851
asS'd'
I19562
sa(dp39124
S'a'
I103372
sS's'
I18
sS'm'
(lp39125
I1
aI3
aI7
aI3
aI1
aI29
aI87
aI101
aI29
aI93
aI643
aI2659
aI1753
aI4797
aI17477
aI16087
aI43453
aI110383
asS'd'
I19563
sa(dp39126
S'a'
I103375
sS's'
I18
sS'm'
(lp39127
I1
aI1
aI7
aI13
aI17
aI51
aI75
aI91
aI445
aI171
aI1369
aI499
aI3753
aI14035
aI4445
aI21437
aI86205
aI231163
asS'd'
I19564
sa(dp39128
S'a'
I103390
sS's'
I18
sS'm'
(lp39129
I1
aI1
aI7
aI13
aI17
aI41
aI67
aI49
aI225
aI659
aI1181
aI1751
aI5211
aI6847
aI20339
aI34271
aI60273
aI52315
asS'd'
I19565
sa(dp39130
S'a'
I103414
sS's'
I18
sS'm'
(lp39131
I1
aI1
aI5
aI5
aI5
aI17
aI87
aI223
aI25
aI773
aI53
aI2447
aI6805
aI6547
aI4429
aI46809
aI51171
aI66457
asS'd'
I19566
sa(dp39132
S'a'
I103432
sS's'
I18
sS'm'
(lp39133
I1
aI1
aI5
aI3
aI29
aI49
aI11
aI205
aI279
aI821
aI725
aI2425
aI443
aI211
aI6347
aI59845
aI90763
aI237911
asS'd'
I19567
sa(dp39134
S'a'
I103446
sS's'
I18
sS'm'
(lp39135
I1
aI1
aI3
aI9
aI31
aI7
aI49
aI1
aI229
aI755
aI517
aI809
aI2955
aI3571
aI2385
aI33097
aI19659
aI55397
asS'd'
I19568
sa(dp39136
S'a'
I103455
sS's'
I18
sS'm'
(lp39137
I1
aI3
aI5
aI3
aI5
aI21
aI67
aI227
aI359
aI401
aI1697
aI2131
aI4585
aI2661
aI3659
aI22621
aI51639
aI245877
asS'd'
I19569
sa(dp39138
S'a'
I103471
sS's'
I18
sS'm'
(lp39139
I1
aI1
aI7
aI5
aI11
aI7
aI57
aI133
aI9
aI917
aI427
aI2777
aI6009
aI11393
aI29473
aI59311
aI77095
aI176215
asS'd'
I19570
sa(dp39140
S'a'
I103474
sS's'
I18
sS'm'
(lp39141
I1
aI1
aI7
aI5
aI13
aI29
aI101
aI15
aI293
aI7
aI797
aI437
aI3739
aI3369
aI16917
aI19047
aI17773
aI219541
asS'd'
I19571
sa(dp39142
S'a'
I103476
sS's'
I18
sS'm'
(lp39143
I1
aI3
aI1
aI9
aI13
aI51
aI15
aI19
aI209
aI991
aI413
aI787
aI3797
aI14029
aI23699
aI8591
aI40429
aI56115
asS'd'
I19572
sa(dp39144
S'a'
I103515
sS's'
I18
sS'm'
(lp39145
I1
aI3
aI7
aI3
aI31
aI9
aI77
aI33
aI487
aI155
aI1903
aI3023
aI8163
aI385
aI4703
aI57511
aI102083
aI85785
asS'd'
I19573
sa(dp39146
S'a'
I103527
sS's'
I18
sS'm'
(lp39147
I1
aI1
aI5
aI11
aI17
aI59
aI115
aI73
aI89
aI723
aI1523
aI2671
aI1755
aI3463
aI19861
aI31573
aI126405
aI90215
asS'd'
I19574
sa(dp39148
S'a'
I103542
sS's'
I18
sS'm'
(lp39149
I1
aI3
aI3
aI13
aI17
aI27
aI123
aI37
aI71
aI173
aI203
aI1245
aI7905
aI8955
aI22589
aI56705
aI120473
aI90129
asS'd'
I19575
sa(dp39150
S'a'
I103562
sS's'
I18
sS'm'
(lp39151
I1
aI3
aI3
aI1
aI17
aI63
aI55
aI225
aI259
aI531
aI1493
aI2639
aI1319
aI10865
aI17993
aI11205
aI13253
aI111261
asS'd'
I19576
sa(dp39152
S'a'
I103569
sS's'
I18
sS'm'
(lp39153
I1
aI1
aI1
aI11
aI25
aI17
aI41
aI45
aI385
aI1009
aI1573
aI1797
aI527
aI543
aI14743
aI35789
aI63871
aI112183
asS'd'
I19577
sa(dp39154
S'a'
I103572
sS's'
I18
sS'm'
(lp39155
I1
aI1
aI7
aI15
aI15
aI57
aI109
aI127
aI143
aI955
aI1091
aI2585
aI3429
aI11763
aI5849
aI53805
aI116865
aI68895
asS'd'
I19578
sa(dp39156
S'a'
I103581
sS's'
I18
sS'm'
(lp39157
I1
aI1
aI7
aI15
aI9
aI11
aI1
aI9
aI491
aI765
aI1835
aI3825
aI5043
aI13091
aI4123
aI19621
aI17687
aI241015
asS'd'
I19579
sa(dp39158
S'a'
I103609
sS's'
I18
sS'm'
(lp39159
I1
aI3
aI5
aI7
aI25
aI51
aI117
aI193
aI91
aI451
aI1871
aI3819
aI1881
aI8065
aI25809
aI36257
aI107955
aI37109
asS'd'
I19580
sa(dp39160
S'a'
I103612
sS's'
I18
sS'm'
(lp39161
I1
aI3
aI1
aI15
aI5
aI5
aI101
aI7
aI59
aI859
aI977
aI2673
aI2825
aI3291
aI26283
aI23467
aI28383
aI257775
asS'd'
I19581
sa(dp39162
S'a'
I103618
sS's'
I18
sS'm'
(lp39163
I1
aI3
aI3
aI15
aI17
aI43
aI87
aI191
aI85
aI829
aI653
aI327
aI1773
aI10101
aI2707
aI18341
aI61435
aI242215
asS'd'
I19582
sa(dp39164
S'a'
I103637
sS's'
I18
sS'm'
(lp39165
I1
aI3
aI7
aI13
aI27
aI43
aI127
aI253
aI403
aI81
aI505
aI2841
aI1509
aI4951
aI23791
aI18099
aI46747
aI192717
asS'd'
I19583
sa(dp39166
S'a'
I103644
sS's'
I18
sS'm'
(lp39167
I1
aI1
aI1
aI9
aI31
aI15
aI127
aI29
aI171
aI999
aI1919
aI4059
aI2781
aI2475
aI8997
aI15459
aI37003
aI217141
asS'd'
I19584
sa(dp39168
S'a'
I103653
sS's'
I18
sS'm'
(lp39169
I1
aI3
aI3
aI1
aI31
aI41
aI67
aI31
aI171
aI719
aI789
aI645
aI3925
aI11117
aI1241
aI63221
aI1087
aI59789
asS'd'
I19585
sa(dp39170
S'a'
I103663
sS's'
I18
sS'm'
(lp39171
I1
aI1
aI1
aI9
aI17
aI41
aI107
aI13
aI405
aI879
aI1955
aI3309
aI1281
aI10601
aI13883
aI43987
aI111691
aI130555
asS'd'
I19586
sa(dp39172
S'a'
I103671
sS's'
I18
sS'm'
(lp39173
I1
aI3
aI7
aI9
aI21
aI17
aI127
aI243
aI51
aI689
aI1945
aI3769
aI7121
aI7703
aI16825
aI34893
aI32167
aI20167
asS'd'
I19587
sa(dp39174
S'a'
I103678
sS's'
I18
sS'm'
(lp39175
I1
aI1
aI1
aI5
aI7
aI7
aI61
aI149
aI75
aI289
aI717
aI1951
aI5917
aI13375
aI15683
aI27507
aI10897
aI199009
asS'd'
I19588
sa(dp39176
S'a'
I103695
sS's'
I18
sS'm'
(lp39177
I1
aI3
aI7
aI9
aI25
aI61
aI43
aI167
aI45
aI299
aI5
aI125
aI3289
aI13685
aI10843
aI25535
aI98383
aI143401
asS'd'
I19589
sa(dp39178
S'a'
I103698
sS's'
I18
sS'm'
(lp39179
I1
aI1
aI3
aI15
aI21
aI27
aI37
aI45
aI233
aI179
aI611
aI3025
aI7887
aI9421
aI16791
aI17147
aI49013
aI33249
asS'd'
I19590
sa(dp39180
S'a'
I103720
sS's'
I18
sS'm'
(lp39181
I1
aI3
aI7
aI15
aI19
aI43
aI43
aI9
aI255
aI295
aI649
aI811
aI1303
aI1989
aI5401
aI53891
aI42679
aI66879
asS'd'
I19591
sa(dp39182
S'a'
I103723
sS's'
I18
sS'm'
(lp39183
I1
aI1
aI7
aI7
aI25
aI31
aI89
aI199
aI455
aI733
aI295
aI1157
aI4375
aI7341
aI7823
aI6025
aI56311
aI257579
asS'd'
I19592
sa(dp39184
S'a'
I103738
sS's'
I18
sS'm'
(lp39185
I1
aI1
aI7
aI3
aI1
aI51
aI15
aI29
aI35
aI917
aI1839
aI741
aI1089
aI8615
aI4967
aI34047
aI32981
aI200693
asS'd'
I19593
sa(dp39186
S'a'
I103743
sS's'
I18
sS'm'
(lp39187
I1
aI3
aI3
aI11
aI21
aI13
aI71
aI53
aI315
aI801
aI1015
aI3829
aI2297
aI6649
aI28501
aI18173
aI83121
aI107195
asS'd'
I19594
sa(dp39188
S'a'
I103745
sS's'
I18
sS'm'
(lp39189
I1
aI1
aI7
aI7
aI1
aI5
aI127
aI103
aI435
aI707
aI7
aI1045
aI187
aI10927
aI32395
aI24999
aI58463
aI94069
asS'd'
I19595
sa(dp39190
S'a'
I103757
sS's'
I18
sS'm'
(lp39191
I1
aI1
aI7
aI9
aI11
aI25
aI73
aI57
aI231
aI455
aI1033
aI2899
aI2861
aI8811
aI21671
aI16115
aI97807
aI221791
asS'd'
I19596
sa(dp39192
S'a'
I103810
sS's'
I18
sS'm'
(lp39193
I1
aI3
aI5
aI11
aI11
aI19
aI105
aI37
aI181
aI1
aI1231
aI2275
aI4789
aI13071
aI24843
aI25901
aI123711
aI145609
asS'd'
I19597
sa(dp39194
S'a'
I103822
sS's'
I18
sS'm'
(lp39195
I1
aI3
aI7
aI13
aI7
aI7
aI21
aI215
aI393
aI43
aI1713
aI2921
aI1959
aI14417
aI17109
aI55793
aI36285
aI81731
asS'd'
I19598
sa(dp39196
S'a'
I103855
sS's'
I18
sS'm'
(lp39197
I1
aI1
aI5
aI15
aI29
aI15
aI5
aI115
aI15
aI795
aI1535
aI2473
aI3663
aI10123
aI20721
aI32739
aI21141
aI217731
asS'd'
I19599
sa(dp39198
S'a'
I103887
sS's'
I18
sS'm'
(lp39199
I1
aI1
aI3
aI1
aI15
aI13
aI21
aI49
aI293
aI689
aI985
aI3949
aI3329
aI7167
aI16925
aI15069
aI47345
aI192749
asS'd'
I19600
sa(dp39200
S'a'
I103905
sS's'
I18
sS'm'
(lp39201
I1
aI3
aI3
aI7
aI5
aI27
aI85
aI11
aI337
aI651
aI777
aI1775
aI5279
aI15229
aI21473
aI36561
aI85293
aI27893
asS'd'
I19601
sa(dp39202
S'a'
I103915
sS's'
I18
sS'm'
(lp39203
I1
aI3
aI7
aI1
aI15
aI29
aI7
aI199
aI71
aI893
aI1587
aI1831
aI3891
aI3299
aI14323
aI23165
aI28199
aI84055
asS'd'
I19602
sa(dp39204
S'a'
I103925
sS's'
I18
sS'm'
(lp39205
I1
aI1
aI1
aI7
aI9
aI27
aI15
aI75
aI497
aI127
aI433
aI1781
aI3711
aI12659
aI3765
aI40827
aI112425
aI36281
asS'd'
I19603
sa(dp39206
S'a'
I103929
sS's'
I18
sS'm'
(lp39207
I1
aI1
aI5
aI1
aI31
aI59
aI9
aI91
aI301
aI217
aI2013
aI2015
aI265
aI3795
aI14609
aI13389
aI5451
aI260169
asS'd'
I19604
sa(dp39208
S'a'
I103945
sS's'
I18
sS'm'
(lp39209
I1
aI1
aI1
aI15
aI19
aI43
aI85
aI37
aI33
aI687
aI1253
aI2615
aI4027
aI3741
aI13971
aI21261
aI106261
aI204233
asS'd'
I19605
sa(dp39210
S'a'
I103954
sS's'
I18
sS'm'
(lp39211
I1
aI1
aI7
aI9
aI31
aI45
aI105
aI111
aI207
aI433
aI633
aI3949
aI8057
aI5049
aI23657
aI12227
aI86251
aI218077
asS'd'
I19606
sa(dp39212
S'a'
I103969
sS's'
I18
sS'm'
(lp39213
I1
aI3
aI5
aI11
aI13
aI33
aI1
aI67
aI117
aI595
aI1835
aI287
aI2509
aI14841
aI1525
aI15761
aI61319
aI182531
asS'd'
I19607
sa(dp39214
S'a'
I104011
sS's'
I18
sS'm'
(lp39215
I1
aI1
aI3
aI9
aI7
aI43
aI105
aI85
aI17
aI349
aI1769
aI3945
aI31
aI4907
aI9373
aI22447
aI70905
aI29189
asS'd'
I19608
sa(dp39216
S'a'
I104055
sS's'
I18
sS'm'
(lp39217
I1
aI1
aI5
aI5
aI9
aI19
aI93
aI179
aI95
aI255
aI1807
aI2717
aI4757
aI15025
aI19479
aI63499
aI42441
aI236519
asS'd'
I19609
sa(dp39218
S'a'
I104111
sS's'
I18
sS'm'
(lp39219
I1
aI1
aI1
aI1
aI21
aI35
aI1
aI101
aI343
aI1023
aI1311
aI347
aI301
aI8419
aI23367
aI31543
aI51661
aI148277
asS'd'
I19610
sa(dp39220
S'a'
I104116
sS's'
I18
sS'm'
(lp39221
I1
aI1
aI7
aI11
aI23
aI37
aI113
aI207
aI147
aI743
aI1905
aI1683
aI3733
aI5659
aI22491
aI62129
aI111671
aI227019
asS'd'
I19611
sa(dp39222
S'a'
I104125
sS's'
I18
sS'm'
(lp39223
I1
aI1
aI3
aI13
aI27
aI5
aI21
aI207
aI445
aI319
aI1355
aI2541
aI2853
aI583
aI1261
aI64477
aI18337
aI91611
asS'd'
I19612
sa(dp39224
S'a'
I104145
sS's'
I18
sS'm'
(lp39225
I1
aI3
aI5
aI3
aI1
aI31
aI51
aI55
aI487
aI735
aI1599
aI523
aI5323
aI10855
aI28717
aI62103
aI18671
aI143885
asS'd'
I19613
sa(dp39226
S'a'
I104162
sS's'
I18
sS'm'
(lp39227
I1
aI3
aI1
aI1
aI15
aI5
aI79
aI107
aI173
aI747
aI1213
aI1151
aI875
aI12759
aI27115
aI16403
aI31349
aI208909
asS'd'
I19614
sa(dp39228
S'a'
I104176
sS's'
I18
sS'm'
(lp39229
I1
aI3
aI5
aI9
aI13
aI57
aI35
aI121
aI135
aI237
aI1707
aI3655
aI8109
aI3623
aI5119
aI27645
aI49401
aI46453
asS'd'
I19615
sa(dp39230
S'a'
I104193
sS's'
I18
sS'm'
(lp39231
I1
aI3
aI7
aI11
aI27
aI3
aI103
aI231
aI43
aI515
aI1279
aI499
aI1355
aI2605
aI11587
aI33641
aI81661
aI29441
asS'd'
I19616
sa(dp39232
S'a'
I104199
sS's'
I18
sS'm'
(lp39233
I1
aI3
aI5
aI7
aI13
aI27
aI55
aI191
aI81
aI185
aI527
aI519
aI4555
aI3349
aI24533
aI60635
aI40009
aI230761
asS'd'
I19617
sa(dp39234
S'a'
I104224
sS's'
I18
sS'm'
(lp39235
I1
aI1
aI5
aI3
aI13
aI35
aI75
aI211
aI67
aI305
aI705
aI543
aI3819
aI16265
aI9867
aI64309
aI35047
aI24873
asS'd'
I19618
sa(dp39236
S'a'
I104236
sS's'
I18
sS'm'
(lp39237
I1
aI3
aI1
aI1
aI11
aI37
aI75
aI3
aI145
aI327
aI1243
aI3291
aI7127
aI5009
aI7757
aI59567
aI90459
aI98035
asS'd'
I19619
sa(dp39238
S'a'
I104247
sS's'
I18
sS'm'
(lp39239
I1
aI3
aI5
aI9
aI9
aI5
aI71
aI29
aI243
aI41
aI447
aI2597
aI1243
aI12899
aI19681
aI30523
aI90799
aI142279
asS'd'
I19620
sa(dp39240
S'a'
I104254
sS's'
I18
sS'm'
(lp39241
I1
aI3
aI3
aI3
aI15
aI21
aI25
aI101
aI451
aI651
aI143
aI3899
aI3377
aI4855
aI23843
aI25047
aI87951
aI239229
asS'd'
I19621
sa(dp39242
S'a'
I104289
sS's'
I18
sS'm'
(lp39243
I1
aI1
aI1
aI1
aI23
aI29
aI25
aI31
aI227
aI43
aI399
aI723
aI693
aI12379
aI11343
aI46123
aI105435
aI224997
asS'd'
I19622
sa(dp39244
S'a'
I104325
sS's'
I18
sS'm'
(lp39245
I1
aI1
aI7
aI3
aI21
aI47
aI77
aI57
aI397
aI689
aI267
aI813
aI1279
aI1727
aI7451
aI34025
aI90273
aI111663
asS'd'
I19623
sa(dp39246
S'a'
I104332
sS's'
I18
sS'm'
(lp39247
I1
aI1
aI5
aI3
aI25
aI61
aI7
aI137
aI271
aI723
aI495
aI2575
aI3695
aI4947
aI31973
aI47835
aI107003
aI221839
asS'd'
I19624
sa(dp39248
S'a'
I104335
sS's'
I18
sS'm'
(lp39249
I1
aI3
aI1
aI7
aI5
aI25
aI21
aI95
aI323
aI3
aI613
aI1721
aI2551
aI8803
aI6803
aI52765
aI34543
aI110945
asS'd'
I19625
sa(dp39250
S'a'
I104353
sS's'
I18
sS'm'
(lp39251
I1
aI3
aI5
aI9
aI27
aI23
aI123
aI193
aI63
aI161
aI1395
aI679
aI161
aI13719
aI29321
aI19243
aI51947
aI229547
asS'd'
I19626
sa(dp39252
S'a'
I104363
sS's'
I18
sS'm'
(lp39253
I1
aI3
aI5
aI15
aI15
aI37
aI127
aI103
aI439
aI513
aI989
aI1393
aI3761
aI9109
aI20649
aI18171
aI69939
aI117447
asS'd'
I19627
sa(dp39254
S'a'
I104371
sS's'
I18
sS'm'
(lp39255
I1
aI1
aI7
aI9
aI7
aI57
aI87
aI159
aI195
aI821
aI57
aI2469
aI7693
aI6759
aI32595
aI41109
aI94785
aI53381
asS'd'
I19628
sa(dp39256
S'a'
I104373
sS's'
I18
sS'm'
(lp39257
I1
aI3
aI3
aI7
aI17
aI63
aI41
aI7
aI437
aI469
aI1453
aI3741
aI7591
aI5563
aI11819
aI23861
aI129777
aI119731
asS'd'
I19629
sa(dp39258
S'a'
I104406
sS's'
I18
sS'm'
(lp39259
I1
aI1
aI3
aI3
aI27
aI55
aI21
aI145
aI243
aI991
aI687
aI909
aI2105
aI4485
aI9095
aI3677
aI53819
aI183027
asS'd'
I19630
sa(dp39260
S'a'
I104412
sS's'
I18
sS'm'
(lp39261
I1
aI1
aI1
aI15
aI13
aI61
aI27
aI81
aI95
aI37
aI1439
aI973
aI7613
aI5749
aI16811
aI26801
aI105745
aI8847
asS'd'
I19631
sa(dp39262
S'a'
I104440
sS's'
I18
sS'm'
(lp39263
I1
aI3
aI1
aI9
aI15
aI17
aI111
aI61
aI27
aI373
aI89
aI2729
aI6397
aI4899
aI11297
aI4403
aI30657
aI207379
asS'd'
I19632
sa(dp39264
S'a'
I104459
sS's'
I18
sS'm'
(lp39265
I1
aI3
aI5
aI15
aI19
aI9
aI19
aI19
aI497
aI667
aI105
aI601
aI6715
aI6355
aI4231
aI19241
aI101771
aI105651
asS'd'
I19633
sa(dp39266
S'a'
I104461
sS's'
I18
sS'm'
(lp39267
I1
aI1
aI3
aI9
aI11
aI11
aI115
aI109
aI177
aI753
aI997
aI2119
aI5969
aI13377
aI13285
aI25373
aI105023
aI158393
asS'd'
I19634
sa(dp39268
S'a'
I104469
sS's'
I18
sS'm'
(lp39269
I1
aI3
aI5
aI15
aI27
aI23
aI19
aI117
aI129
aI457
aI1973
aI2171
aI8071
aI2093
aI13407
aI9295
aI82967
aI184753
asS'd'
I19635
sa(dp39270
S'a'
I104479
sS's'
I18
sS'm'
(lp39271
I1
aI1
aI7
aI5
aI1
aI45
aI9
aI25
aI307
aI629
aI1169
aI2859
aI3987
aI11411
aI13609
aI44993
aI26019
aI107003
asS'd'
I19636
sa(dp39272
S'a'
I104483
sS's'
I18
sS'm'
(lp39273
I1
aI1
aI1
aI13
aI29
aI15
aI13
aI163
aI203
aI885
aI281
aI1605
aI8001
aI899
aI4081
aI65467
aI61283
aI198963
asS'd'
I19637
sa(dp39274
S'a'
I104490
sS's'
I18
sS'm'
(lp39275
I1
aI1
aI3
aI5
aI27
aI9
aI89
aI193
aI375
aI633
aI1807
aI2069
aI3467
aI3167
aI23751
aI39115
aI90093
aI190365
asS'd'
I19638
sa(dp39276
S'a'
I104500
sS's'
I18
sS'm'
(lp39277
I1
aI3
aI3
aI5
aI1
aI1
aI19
aI161
aI21
aI745
aI493
aI2227
aI7993
aI3337
aI27041
aI4817
aI58963
aI237015
asS'd'
I19639
sa(dp39278
S'a'
I104536
sS's'
I18
sS'm'
(lp39279
I1
aI3
aI5
aI15
aI15
aI45
aI51
aI145
aI299
aI787
aI1059
aI2407
aI1143
aI775
aI17473
aI22235
aI18241
aI103897
asS'd'
I19640
sa(dp39280
S'a'
I104560
sS's'
I18
sS'm'
(lp39281
I1
aI1
aI1
aI9
aI11
aI35
aI75
aI195
aI281
aI935
aI113
aI3009
aI4797
aI7221
aI12475
aI18563
aI10315
aI255541
asS'd'
I19641
sa(dp39282
S'a'
I104569
sS's'
I18
sS'm'
(lp39283
I1
aI1
aI5
aI7
aI25
aI9
aI121
aI179
aI303
aI511
aI2041
aI1485
aI529
aI13843
aI29013
aI28139
aI63237
aI21259
asS'd'
I19642
sa(dp39284
S'a'
I104588
sS's'
I18
sS'm'
(lp39285
I1
aI3
aI7
aI11
aI9
aI21
aI61
aI177
aI63
aI179
aI679
aI17
aI4069
aI8929
aI14499
aI53913
aI27925
aI48449
asS'd'
I19643
sa(dp39286
S'a'
I104591
sS's'
I18
sS'm'
(lp39287
I1
aI3
aI5
aI3
aI9
aI27
aI111
aI247
aI253
aI175
aI1591
aI3583
aI3351
aI9039
aI597
aI23859
aI126585
aI157367
asS'd'
I19644
sa(dp39288
S'a'
I104621
sS's'
I18
sS'm'
(lp39289
I1
aI1
aI7
aI9
aI1
aI9
aI29
aI1
aI273
aI89
aI767
aI1873
aI39
aI10487
aI29857
aI23577
aI67457
aI44571
asS'd'
I19645
sa(dp39290
S'a'
I104661
sS's'
I18
sS'm'
(lp39291
I1
aI1
aI7
aI5
aI5
aI23
aI13
aI181
aI283
aI333
aI21
aI1409
aI5937
aI8981
aI7445
aI61267
aI21729
aI32677
asS'd'
I19646
sa(dp39292
S'a'
I104675
sS's'
I18
sS'm'
(lp39293
I1
aI3
aI1
aI15
aI31
aI5
aI45
aI253
aI179
aI91
aI341
aI359
aI4269
aI7567
aI23699
aI30589
aI42909
aI126171
asS'd'
I19647
sa(dp39294
S'a'
I104684
sS's'
I18
sS'm'
(lp39295
I1
aI3
aI7
aI7
aI23
aI45
aI15
aI93
aI115
aI873
aI49
aI845
aI827
aI14357
aI20163
aI41565
aI37105
aI120331
asS'd'
I19648
sa(dp39296
S'a'
I104713
sS's'
I18
sS'm'
(lp39297
I1
aI1
aI7
aI7
aI15
aI25
aI55
aI175
aI11
aI457
aI1537
aI1937
aI807
aI11399
aI27751
aI16935
aI75231
aI204039
asS'd'
I19649
sa(dp39298
S'a'
I104747
sS's'
I18
sS'm'
(lp39299
I1
aI3
aI7
aI13
aI1
aI37
aI67
aI145
aI471
aI1013
aI1273
aI4093
aI4449
aI4433
aI29063
aI205
aI93249
aI140383
asS'd'
I19650
sa(dp39300
S'a'
I104750
sS's'
I18
sS'm'
(lp39301
I1
aI3
aI1
aI13
aI1
aI57
aI85
aI223
aI349
aI125
aI863
aI2179
aI7813
aI8817
aI1767
aI19169
aI41367
aI65883
asS'd'
I19651
sa(dp39302
S'a'
I104757
sS's'
I18
sS'm'
(lp39303
I1
aI3
aI3
aI15
aI21
aI39
aI3
aI31
aI67
aI505
aI271
aI505
aI5495
aI4183
aI3631
aI4567
aI48061
aI149565
asS'd'
I19652
sa(dp39304
S'a'
I104775
sS's'
I18
sS'm'
(lp39305
I1
aI1
aI7
aI11
aI13
aI39
aI109
aI201
aI287
aI1013
aI1505
aI3105
aI3845
aI1963
aI4361
aI61753
aI29145
aI146235
asS'd'
I19653
sa(dp39306
S'a'
I104782
sS's'
I18
sS'm'
(lp39307
I1
aI3
aI5
aI3
aI29
aI39
aI71
aI99
aI501
aI53
aI835
aI3295
aI5335
aI1017
aI25913
aI63637
aI115353
aI210509
asS'd'
I19654
sa(dp39308
S'a'
I104803
sS's'
I18
sS'm'
(lp39309
I1
aI3
aI1
aI13
aI9
aI53
aI33
aI177
aI419
aI63
aI793
aI1329
aI2261
aI11633
aI18805
aI49771
aI47533
aI74949
asS'd'
I19655
sa(dp39310
S'a'
I104820
sS's'
I18
sS'm'
(lp39311
I1
aI3
aI7
aI11
aI7
aI49
aI123
aI237
aI195
aI17
aI1919
aI1911
aI4135
aI11829
aI26307
aI34905
aI114361
aI228655
asS'd'
I19656
sa(dp39312
S'a'
I104824
sS's'
I18
sS'm'
(lp39313
I1
aI3
aI5
aI3
aI5
aI57
aI65
aI161
aI195
aI857
aI1187
aI2303
aI1975
aI2555
aI26065
aI17963
aI57403
aI136193
asS'd'
I19657
sa(dp39314
S'a'
I104829
sS's'
I18
sS'm'
(lp39315
I1
aI1
aI3
aI11
aI1
aI5
aI105
aI217
aI229
aI769
aI1393
aI2419
aI7751
aI14293
aI9407
aI4569
aI30663
aI89345
asS'd'
I19658
sa(dp39316
S'a'
I104830
sS's'
I18
sS'm'
(lp39317
I1
aI3
aI3
aI7
aI23
aI15
aI15
aI189
aI67
aI863
aI485
aI2435
aI5145
aI10537
aI16485
aI50369
aI118245
aI253201
asS'd'
I19659
sa(dp39318
S'a'
I104836
sS's'
I18
sS'm'
(lp39319
I1
aI1
aI1
aI5
aI27
aI27
aI47
aI129
aI383
aI227
aI115
aI3027
aI1575
aI15765
aI10207
aI4885
aI125707
aI184703
asS'd'
I19660
sa(dp39320
S'a'
I104851
sS's'
I18
sS'm'
(lp39321
I1
aI3
aI5
aI5
aI17
aI9
aI45
aI55
aI151
aI751
aI415
aI2139
aI8071
aI2309
aI27641
aI29743
aI47183
aI21437
asS'd'
I19661
sa(dp39322
S'a'
I104882
sS's'
I18
sS'm'
(lp39323
I1
aI3
aI3
aI7
aI9
aI47
aI51
aI31
aI261
aI237
aI1695
aI1065
aI4503
aI7167
aI25791
aI39659
aI90145
aI130481
asS'd'
I19662
sa(dp39324
S'a'
I104920
sS's'
I18
sS'm'
(lp39325
I1
aI1
aI3
aI3
aI21
aI27
aI53
aI249
aI407
aI779
aI1315
aI1005
aI6953
aI14959
aI2265
aI61645
aI118623
aI254067
asS'd'
I19663
sa(dp39326
S'a'
I104941
sS's'
I18
sS'm'
(lp39327
I1
aI3
aI1
aI9
aI13
aI25
aI33
aI225
aI467
aI325
aI1513
aI1237
aI1569
aI8749
aI1587
aI4699
aI19893
aI163597
asS'd'
I19664
sa(dp39328
S'a'
I104963
sS's'
I18
sS'm'
(lp39329
I1
aI3
aI5
aI5
aI29
aI29
aI1
aI55
aI437
aI575
aI149
aI791
aI4243
aI4405
aI22963
aI64125
aI21631
aI25819
asS'd'
I19665
sa(dp39330
S'a'
I104989
sS's'
I18
sS'm'
(lp39331
I1
aI3
aI5
aI3
aI25
aI9
aI19
aI27
aI419
aI139
aI2035
aI2065
aI1925
aI11499
aI20053
aI13161
aI15115
aI120891
asS'd'
I19666
sa(dp39332
S'a'
I104999
sS's'
I18
sS'm'
(lp39333
I1
aI1
aI5
aI9
aI7
aI9
aI113
aI3
aI195
aI555
aI863
aI595
aI6569
aI9819
aI14727
aI38285
aI13737
aI130903
asS'd'
I19667
sa(dp39334
S'a'
I105006
sS's'
I18
sS'm'
(lp39335
I1
aI1
aI3
aI7
aI19
aI17
aI35
aI107
aI489
aI285
aI247
aI3103
aI2919
aI11163
aI2811
aI62081
aI42989
aI24495
asS'd'
I19668
sa(dp39336
S'a'
I105040
sS's'
I18
sS'm'
(lp39337
I1
aI3
aI1
aI15
aI17
aI5
aI123
aI221
aI117
aI689
aI1567
aI3803
aI5801
aI14121
aI23263
aI8851
aI41559
aI226271
asS'd'
I19669
sa(dp39338
S'a'
I105066
sS's'
I18
sS'm'
(lp39339
I1
aI1
aI1
aI15
aI21
aI39
aI87
aI135
aI485
aI59
aI1899
aI2483
aI2599
aI8783
aI6129
aI55407
aI7291
aI217117
asS'd'
I19670
sa(dp39340
S'a'
I105089
sS's'
I18
sS'm'
(lp39341
I1
aI1
aI3
aI11
aI27
aI53
aI45
aI241
aI51
aI829
aI121
aI3047
aI6785
aI15127
aI13923
aI47913
aI9087
aI112005
asS'd'
I19671
sa(dp39342
S'a'
I105120
sS's'
I18
sS'm'
(lp39343
I1
aI1
aI1
aI13
aI5
aI9
aI45
aI37
aI235
aI497
aI871
aI1471
aI2895
aI247
aI9085
aI39611
aI63445
aI218391
asS'd'
I19672
sa(dp39344
S'a'
I105125
sS's'
I18
sS'm'
(lp39345
I1
aI1
aI5
aI3
aI5
aI29
aI23
aI155
aI339
aI293
aI535
aI1569
aI2625
aI2867
aI4639
aI53049
aI88721
aI96903
asS'd'
I19673
sa(dp39346
S'a'
I105137
sS's'
I18
sS'm'
(lp39347
I1
aI1
aI3
aI15
aI5
aI25
aI17
aI45
aI47
aI683
aI949
aI1381
aI5929
aI9539
aI3345
aI59883
aI19071
aI200411
asS'd'
I19674
sa(dp39348
S'a'
I105179
sS's'
I18
sS'm'
(lp39349
I1
aI1
aI1
aI11
aI11
aI5
aI71
aI15
aI469
aI749
aI1763
aI541
aI7109
aI11731
aI8463
aI35145
aI121795
aI219441
asS'd'
I19675
sa(dp39350
S'a'
I105215
sS's'
I18
sS'm'
(lp39351
I1
aI3
aI7
aI11
aI5
aI17
aI63
aI159
aI69
aI993
aI1519
aI1531
aI6913
aI3901
aI22131
aI42909
aI41299
aI261813
asS'd'
I19676
sa(dp39352
S'a'
I105244
sS's'
I18
sS'm'
(lp39353
I1
aI1
aI3
aI5
aI21
aI27
aI27
aI197
aI339
aI275
aI2011
aI2263
aI855
aI1939
aI21561
aI30577
aI6515
aI124417
asS'd'
I19677
sa(dp39354
S'a'
I105258
sS's'
I18
sS'm'
(lp39355
I1
aI3
aI5
aI15
aI25
aI35
aI91
aI31
aI269
aI857
aI327
aI3643
aI3211
aI14401
aI18399
aI9007
aI12897
aI25555
asS'd'
I19678
sa(dp39356
S'a'
I105263
sS's'
I18
sS'm'
(lp39357
I1
aI1
aI7
aI15
aI19
aI47
aI127
aI157
aI407
aI867
aI943
aI1509
aI3113
aI49
aI32131
aI46975
aI130013
aI66457
asS'd'
I19679
sa(dp39358
S'a'
I105280
sS's'
I18
sS'm'
(lp39359
I1
aI3
aI3
aI13
aI31
aI45
aI59
aI135
aI67
aI825
aI157
aI2441
aI2851
aI2355
aI28115
aI14075
aI106317
aI145945
asS'd'
I19680
sa(dp39360
S'a'
I105286
sS's'
I18
sS'm'
(lp39361
I1
aI1
aI1
aI15
aI27
aI27
aI5
aI211
aI85
aI749
aI671
aI1341
aI6865
aI10027
aI14419
aI20159
aI126647
aI147893
asS'd'
I19681
sa(dp39362
S'a'
I105380
sS's'
I18
sS'm'
(lp39363
I1
aI1
aI1
aI7
aI25
aI49
aI115
aI231
aI65
aI161
aI1409
aI573
aI2859
aI639
aI12567
aI58459
aI73781
aI136893
asS'd'
I19682
sa(dp39364
S'a'
I105397
sS's'
I18
sS'm'
(lp39365
I1
aI1
aI1
aI7
aI11
aI57
aI97
aI141
aI327
aI975
aI1799
aI3051
aI365
aI9331
aI14583
aI16723
aI113153
aI224127
asS'd'
I19683
sa(dp39366
S'a'
I105409
sS's'
I18
sS'm'
(lp39367
I1
aI3
aI1
aI13
aI13
aI17
aI105
aI109
aI151
aI41
aI1903
aI1685
aI2285
aI5697
aI16559
aI34133
aI106045
aI203217
asS'd'
I19684
sa(dp39368
S'a'
I105427
sS's'
I18
sS'm'
(lp39369
I1
aI1
aI1
aI3
aI29
aI61
aI43
aI255
aI269
aI277
aI1847
aI3781
aI7991
aI131
aI9833
aI34305
aI10763
aI41869
asS'd'
I19685
sa(dp39370
S'a'
I105452
sS's'
I18
sS'm'
(lp39371
I1
aI1
aI7
aI3
aI11
aI1
aI85
aI89
aI247
aI99
aI419
aI2041
aI3729
aI10301
aI5291
aI36033
aI31749
aI261871
asS'd'
I19686
sa(dp39372
S'a'
I105469
sS's'
I18
sS'm'
(lp39373
I1
aI3
aI7
aI9
aI27
aI27
aI41
aI3
aI303
aI893
aI697
aI1631
aI5015
aI4233
aI22827
aI3913
aI22245
aI121193
asS'd'
I19687
sa(dp39374
S'a'
I105472
sS's'
I18
sS'm'
(lp39375
I1
aI3
aI3
aI13
aI19
aI11
aI15
aI239
aI419
aI123
aI1213
aI185
aI4465
aI4909
aI25421
aI18363
aI72537
aI167939
asS'd'
I19688
sa(dp39376
S'a'
I105477
sS's'
I18
sS'm'
(lp39377
I1
aI1
aI3
aI5
aI13
aI49
aI57
aI197
aI19
aI877
aI1465
aI2933
aI5909
aI7147
aI1039
aI37035
aI91209
aI126013
asS'd'
I19689
sa(dp39378
S'a'
I105490
sS's'
I18
sS'm'
(lp39379
I1
aI1
aI3
aI5
aI23
aI7
aI63
aI179
aI59
aI47
aI1501
aI1863
aI2949
aI13959
aI28131
aI29705
aI107975
aI155251
asS'd'
I19690
sa(dp39380
S'a'
I105505
sS's'
I18
sS'm'
(lp39381
I1
aI1
aI1
aI5
aI25
aI51
aI77
aI169
aI327
aI585
aI1531
aI3367
aI3075
aI6313
aI26725
aI453
aI68635
aI173787
asS'd'
I19691
sa(dp39382
S'a'
I105529
sS's'
I18
sS'm'
(lp39383
I1
aI3
aI7
aI9
aI5
aI15
aI43
aI45
aI311
aI367
aI297
aI2249
aI7507
aI4785
aI22685
aI37279
aI121683
aI141453
asS'd'
I19692
sa(dp39384
S'a'
I105530
sS's'
I18
sS'm'
(lp39385
I1
aI3
aI7
aI15
aI17
aI33
aI43
aI251
aI281
aI345
aI1659
aI3729
aI7629
aI6179
aI26107
aI64255
aI88003
aI2545
asS'd'
I19693
sa(dp39386
S'a'
I105604
sS's'
I18
sS'm'
(lp39387
I1
aI3
aI7
aI9
aI17
aI13
aI71
aI49
aI341
aI495
aI1975
aI173
aI5773
aI3821
aI6615
aI50917
aI53781
aI75557
asS'd'
I19694
sa(dp39388
S'a'
I105614
sS's'
I18
sS'm'
(lp39389
I1
aI1
aI5
aI13
aI17
aI21
aI121
aI35
aI195
aI367
aI1191
aI1331
aI6423
aI8425
aI7705
aI59117
aI44575
aI225431
asS'd'
I19695
sa(dp39390
S'a'
I105622
sS's'
I18
sS'm'
(lp39391
I1
aI1
aI1
aI5
aI25
aI39
aI89
aI65
aI449
aI491
aI211
aI2949
aI4493
aI23
aI22571
aI4801
aI50525
aI222563
asS'd'
I19696
sa(dp39392
S'a'
I105625
sS's'
I18
sS'm'
(lp39393
I1
aI1
aI7
aI5
aI7
aI21
aI15
aI171
aI443
aI225
aI577
aI1841
aI8139
aI15071
aI6095
aI60185
aI71957
aI244753
asS'd'
I19697
sa(dp39394
S'a'
I105652
sS's'
I18
sS'm'
(lp39395
I1
aI1
aI5
aI9
aI31
aI33
aI21
aI195
aI415
aI1003
aI441
aI627
aI2339
aI8245
aI11187
aI55933
aI59045
aI177455
asS'd'
I19698
sa(dp39396
S'a'
I105687
sS's'
I18
sS'm'
(lp39397
I1
aI1
aI3
aI9
aI31
aI29
aI7
aI15
aI31
aI577
aI1435
aI1317
aI2923
aI3807
aI29693
aI45857
aI61787
aI213565
asS'd'
I19699
sa(dp39398
S'a'
I105693
sS's'
I18
sS'm'
(lp39399
I1
aI3
aI7
aI15
aI17
aI51
aI65
aI87
aI295
aI811
aI753
aI1113
aI7695
aI275
aI28331
aI46363
aI53247
aI166993
asS'd'
I19700
sa(dp39400
S'a'
I105707
sS's'
I18
sS'm'
(lp39401
I1
aI1
aI5
aI11
aI15
aI31
aI31
aI47
aI273
aI383
aI1831
aI3821
aI1337
aI13257
aI415
aI35453
aI15293
aI173095
asS'd'
I19701
sa(dp39402
S'a'
I105721
sS's'
I18
sS'm'
(lp39403
I1
aI3
aI1
aI13
aI19
aI21
aI63
aI199
aI159
aI475
aI1257
aI3119
aI7083
aI2861
aI21099
aI16873
aI83583
aI186289
asS'd'
I19702
sa(dp39404
S'a'
I105730
sS's'
I18
sS'm'
(lp39405
I1
aI1
aI7
aI13
aI15
aI23
aI11
aI103
aI387
aI899
aI261
aI2863
aI3681
aI5683
aI5587
aI64655
aI7411
aI148633
asS'd'
I19703
sa(dp39406
S'a'
I105736
sS's'
I18
sS'm'
(lp39407
I1
aI3
aI1
aI13
aI1
aI45
aI5
aI173
aI379
aI287
aI1451
aI2253
aI3811
aI10963
aI20285
aI59681
aI81285
aI48523
asS'd'
I19704
sa(dp39408
S'a'
I105753
sS's'
I18
sS'm'
(lp39409
I1
aI1
aI5
aI9
aI27
aI17
aI57
aI25
aI499
aI289
aI1083
aI3057
aI793
aI5251
aI10519
aI36647
aI67751
aI237487
asS'd'
I19705
sa(dp39410
S'a'
I105759
sS's'
I18
sS'm'
(lp39411
I1
aI1
aI3
aI1
aI17
aI13
aI23
aI73
aI1
aI951
aI111
aI3769
aI3611
aI4827
aI10081
aI55919
aI21411
aI127303
asS'd'
I19706
sa(dp39412
S'a'
I105760
sS's'
I18
sS'm'
(lp39413
I1
aI3
aI1
aI1
aI29
aI35
aI93
aI139
aI179
aI217
aI1839
aI1907
aI4365
aI1813
aI31985
aI28927
aI39319
aI233413
asS'd'
I19707
sa(dp39414
S'a'
I105802
sS's'
I18
sS'm'
(lp39415
I1
aI1
aI1
aI7
aI7
aI47
aI37
aI127
aI449
aI473
aI311
aI3833
aI3263
aI4163
aI15985
aI50159
aI82685
aI73273
asS'd'
I19708
sa(dp39416
S'a'
I105822
sS's'
I18
sS'm'
(lp39417
I1
aI1
aI1
aI3
aI3
aI5
aI69
aI95
aI101
aI115
aI1575
aI63
aI1897
aI13733
aI22149
aI8793
aI82983
aI192553
asS'd'
I19709
sa(dp39418
S'a'
I105832
sS's'
I18
sS'm'
(lp39419
I1
aI3
aI1
aI5
aI3
aI57
aI121
aI13
aI291
aI975
aI505
aI3105
aI6929
aI12737
aI25869
aI29173
aI16757
aI17733
asS'd'
I19710
sa(dp39420
S'a'
I105865
sS's'
I18
sS'm'
(lp39421
I1
aI1
aI3
aI13
aI13
aI23
aI51
aI51
aI239
aI795
aI877
aI1547
aI6533
aI11497
aI14309
aI32941
aI128109
aI187313
asS'd'
I19711
sa(dp39422
S'a'
I105873
sS's'
I18
sS'm'
(lp39423
I1
aI1
aI3
aI5
aI19
aI37
aI57
aI223
aI509
aI379
aI1235
aI1881
aI4133
aI13219
aI5479
aI36781
aI56155
aI231001
asS'd'
I19712
sa(dp39424
S'a'
I105874
sS's'
I18
sS'm'
(lp39425
I1
aI1
aI1
aI5
aI25
aI45
aI111
aI183
aI37
aI875
aI1487
aI3771
aI5593
aI6835
aI10921
aI40697
aI114455
aI259491
asS'd'
I19713
sa(dp39426
S'a'
I105880
sS's'
I18
sS'm'
(lp39427
I1
aI1
aI3
aI11
aI13
aI5
aI43
aI47
aI7
aI435
aI1543
aI3835
aI1631
aI8889
aI23567
aI24705
aI26687
aI14261
asS'd'
I19714
sa(dp39428
S'a'
I105899
sS's'
I18
sS'm'
(lp39429
I1
aI3
aI3
aI11
aI1
aI13
aI51
aI171
aI433
aI1011
aI679
aI1307
aI1683
aI7379
aI7163
aI29727
aI40209
aI259973
asS'd'
I19715
sa(dp39430
S'a'
I105902
sS's'
I18
sS'm'
(lp39431
I1
aI3
aI1
aI5
aI25
aI33
aI111
aI11
aI303
aI815
aI1883
aI263
aI7907
aI12637
aI19203
aI64151
aI55739
aI240509
asS'd'
I19716
sa(dp39432
S'a'
I105909
sS's'
I18
sS'm'
(lp39433
I1
aI3
aI5
aI7
aI23
aI51
aI3
aI183
aI425
aI47
aI231
aI807
aI1891
aI10943
aI17873
aI20209
aI60871
aI30269
asS'd'
I19717
sa(dp39434
S'a'
I105919
sS's'
I18
sS'm'
(lp39435
I1
aI1
aI1
aI1
aI1
aI23
aI11
aI183
aI351
aI5
aI37
aI3883
aI1291
aI15933
aI11379
aI53057
aI61389
aI240547
asS'd'
I19718
sa(dp39436
S'a'
I105922
sS's'
I18
sS'm'
(lp39437
I1
aI3
aI1
aI5
aI7
aI53
aI97
aI29
aI239
aI805
aI1929
aI1001
aI5103
aI11
aI24695
aI7825
aI109755
aI254717
asS'd'
I19719
sa(dp39438
S'a'
I105946
sS's'
I18
sS'm'
(lp39439
I1
aI3
aI3
aI11
aI15
aI37
aI115
aI117
aI391
aI313
aI1761
aI3627
aI3931
aI10277
aI1767
aI25401
aI123827
aI60463
asS'd'
I19720
sa(dp39440
S'a'
I105975
sS's'
I18
sS'm'
(lp39441
I1
aI1
aI7
aI9
aI9
aI61
aI23
aI65
aI473
aI579
aI1979
aI415
aI629
aI3613
aI27409
aI46909
aI3281
aI3883
asS'd'
I19721
sa(dp39442
S'a'
I105992
sS's'
I18
sS'm'
(lp39443
I1
aI3
aI3
aI13
aI25
aI17
aI71
aI247
aI297
aI451
aI153
aI1949
aI2727
aI6427
aI19815
aI54013
aI129363
aI248237
asS'd'
I19722
sa(dp39444
S'a'
I106005
sS's'
I18
sS'm'
(lp39445
I1
aI1
aI7
aI7
aI29
aI9
aI57
aI171
aI159
aI287
aI693
aI1365
aI4665
aI5255
aI15013
aI21225
aI125409
aI5893
asS'd'
I19723
sa(dp39446
S'a'
I106057
sS's'
I18
sS'm'
(lp39447
I1
aI3
aI3
aI1
aI9
aI43
aI13
aI139
aI359
aI267
aI115
aI2025
aI693
aI4789
aI10353
aI60459
aI30835
aI56575
asS'd'
I19724
sa(dp39448
S'a'
I106081
sS's'
I18
sS'm'
(lp39449
I1
aI1
aI1
aI7
aI23
aI35
aI35
aI77
aI245
aI705
aI75
aI2841
aI1683
aI6109
aI19661
aI49021
aI25019
aI7297
asS'd'
I19725
sa(dp39450
S'a'
I106093
sS's'
I18
sS'm'
(lp39451
I1
aI3
aI1
aI5
aI1
aI19
aI49
aI53
aI19
aI435
aI1471
aI409
aI7051
aI7057
aI3621
aI42925
aI59551
aI70941
asS'd'
I19726
sa(dp39452
S'a'
I106117
sS's'
I18
sS'm'
(lp39453
I1
aI3
aI7
aI5
aI21
aI27
aI39
aI221
aI389
aI255
aI537
aI597
aI7729
aI10473
aI6657
aI13261
aI11303
aI112409
asS'd'
I19727
sa(dp39454
S'a'
I106127
sS's'
I18
sS'm'
(lp39455
I1
aI3
aI7
aI13
aI29
aI29
aI81
aI107
aI329
aI25
aI537
aI561
aI2247
aI10371
aI30031
aI20537
aI28437
aI113319
asS'd'
I19728
sa(dp39456
S'a'
I106135
sS's'
I18
sS'm'
(lp39457
I1
aI1
aI5
aI9
aI9
aI7
aI89
aI155
aI337
aI829
aI755
aI3259
aI3563
aI11849
aI31179
aI43297
aI79601
aI187545
asS'd'
I19729
sa(dp39458
S'a'
I106141
sS's'
I18
sS'm'
(lp39459
I1
aI3
aI3
aI5
aI11
aI63
aI101
aI159
aI357
aI627
aI587
aI3233
aI405
aI4083
aI5953
aI44541
aI110723
aI240573
asS'd'
I19730
sa(dp39460
S'a'
I106189
sS's'
I18
sS'm'
(lp39461
I1
aI1
aI3
aI15
aI5
aI39
aI87
aI231
aI455
aI195
aI1603
aI315
aI3869
aI6375
aI745
aI28349
aI56469
aI119033
asS'd'
I19731
sa(dp39462
S'a'
I106190
sS's'
I18
sS'm'
(lp39463
I1
aI1
aI7
aI1
aI11
aI7
aI79
aI47
aI391
aI585
aI1299
aI3237
aI7345
aI15959
aI4293
aI43285
aI111737
aI215923
asS'd'
I19732
sa(dp39464
S'a'
I106197
sS's'
I18
sS'm'
(lp39465
I1
aI1
aI3
aI15
aI21
aI53
aI113
aI73
aI265
aI589
aI299
aI289
aI3983
aI1653
aI17407
aI15287
aI53199
aI44221
asS'd'
I19733
sa(dp39466
S'a'
I106214
sS's'
I18
sS'm'
(lp39467
I1
aI3
aI1
aI3
aI1
aI13
aI31
aI41
aI509
aI523
aI401
aI2647
aI2731
aI5699
aI15649
aI51737
aI81249
aI230865
asS'd'
I19734
sa(dp39468
S'a'
I106237
sS's'
I18
sS'm'
(lp39469
I1
aI3
aI5
aI3
aI15
aI53
aI91
aI209
aI249
aI243
aI1119
aI2531
aI319
aI9259
aI26555
aI59579
aI28767
aI235073
asS'd'
I19735
sa(dp39470
S'a'
I106245
sS's'
I18
sS'm'
(lp39471
I1
aI1
aI1
aI13
aI29
aI25
aI57
aI3
aI413
aI945
aI841
aI1151
aI7167
aI2545
aI7283
aI3947
aI109375
aI148677
asS'd'
I19736
sa(dp39472
S'a'
I106257
sS's'
I18
sS'm'
(lp39473
I1
aI3
aI3
aI7
aI29
aI27
aI81
aI141
aI21
aI771
aI577
aI897
aI73
aI13081
aI30035
aI49213
aI90627
aI7483
asS'd'
I19737
sa(dp39474
S'a'
I106300
sS's'
I18
sS'm'
(lp39475
I1
aI3
aI1
aI13
aI23
aI11
aI19
aI159
aI183
aI789
aI683
aI2071
aI7107
aI3025
aI27333
aI9571
aI69621
aI48529
asS'd'
I19738
sa(dp39476
S'a'
I106335
sS's'
I18
sS'm'
(lp39477
I1
aI3
aI1
aI9
aI5
aI43
aI7
aI123
aI341
aI75
aI1709
aI135
aI7929
aI14563
aI6849
aI32783
aI91971
aI223789
asS'd'
I19739
sa(dp39478
S'a'
I106351
sS's'
I18
sS'm'
(lp39479
I1
aI1
aI7
aI3
aI19
aI13
aI45
aI229
aI7
aI559
aI1895
aI2649
aI7593
aI1063
aI17715
aI45019
aI29743
aI37819
asS'd'
I19740
sa(dp39480
S'a'
I106370
sS's'
I18
sS'm'
(lp39481
I1
aI1
aI5
aI3
aI11
aI25
aI11
aI169
aI249
aI415
aI249
aI209
aI2223
aI5947
aI23381
aI12109
aI37697
aI131729
asS'd'
I19741
sa(dp39482
S'a'
I106375
sS's'
I18
sS'm'
(lp39483
I1
aI1
aI1
aI7
aI19
aI47
aI125
aI117
aI235
aI825
aI801
aI921
aI2363
aI1261
aI20529
aI65445
aI55315
aI173561
asS'd'
I19742
sa(dp39484
S'a'
I106376
sS's'
I18
sS'm'
(lp39485
I1
aI1
aI1
aI1
aI5
aI39
aI31
aI11
aI239
aI333
aI727
aI3991
aI1453
aI2201
aI18129
aI3513
aI112057
aI109673
asS'd'
I19743
sa(dp39486
S'a'
I106384
sS's'
I18
sS'm'
(lp39487
I1
aI1
aI7
aI5
aI21
aI59
aI37
aI255
aI261
aI1
aI1401
aI1101
aI1233
aI3813
aI22809
aI39389
aI66263
aI191623
asS'd'
I19744
sa(dp39488
S'a'
I106403
sS's'
I18
sS'm'
(lp39489
I1
aI3
aI1
aI7
aI5
aI51
aI73
aI85
aI319
aI671
aI1149
aI1631
aI6021
aI10711
aI3813
aI36485
aI106147
aI202021
asS'd'
I19745
sa(dp39490
S'a'
I106405
sS's'
I18
sS'm'
(lp39491
I1
aI3
aI3
aI13
aI7
aI63
aI59
aI253
aI337
aI453
aI61
aI209
aI2809
aI10429
aI28069
aI55971
aI57985
aI244779
asS'd'
I19746
sa(dp39492
S'a'
I106406
sS's'
I18
sS'm'
(lp39493
I1
aI1
aI1
aI9
aI27
aI59
aI45
aI101
aI427
aI713
aI1667
aI2965
aI6161
aI1235
aI8793
aI2387
aI66031
aI85151
asS'd'
I19747
sa(dp39494
S'a'
I106429
sS's'
I18
sS'm'
(lp39495
I1
aI1
aI3
aI1
aI5
aI25
aI101
aI7
aI449
aI149
aI823
aI725
aI6803
aI11949
aI13009
aI14785
aI45633
aI241957
asS'd'
I19748
sa(dp39496
S'a'
I106447
sS's'
I18
sS'm'
(lp39497
I1
aI3
aI7
aI15
aI29
aI45
aI103
aI151
aI159
aI23
aI1353
aI3541
aI5909
aI4173
aI31391
aI16179
aI38289
aI206603
asS'd'
I19749
sa(dp39498
S'a'
I106466
sS's'
I18
sS'm'
(lp39499
I1
aI1
aI7
aI7
aI23
aI3
aI97
aI29
aI141
aI383
aI379
aI3189
aI4399
aI4545
aI30797
aI55827
aI126223
aI97049
asS'd'
I19750
sa(dp39500
S'a'
I106495
sS's'
I18
sS'm'
(lp39501
I1
aI3
aI3
aI7
aI25
aI35
aI61
aI15
aI349
aI929
aI65
aI3697
aI7637
aI12239
aI29051
aI36001
aI114513
aI151069
asS'd'
I19751
sa(dp39502
S'a'
I106532
sS's'
I18
sS'm'
(lp39503
I1
aI3
aI3
aI11
aI19
aI1
aI23
aI245
aI9
aI689
aI1251
aI1043
aI2393
aI15817
aI31561
aI21059
aI3435
aI228091
asS'd'
I19752
sa(dp39504
S'a'
I106542
sS's'
I18
sS'm'
(lp39505
I1
aI3
aI7
aI3
aI23
aI17
aI121
aI147
aI427
aI47
aI905
aI3877
aI2301
aI15709
aI13599
aI48895
aI108955
aI53219
asS'd'
I19753
sa(dp39506
S'a'
I106547
sS's'
I18
sS'm'
(lp39507
I1
aI3
aI7
aI3
aI25
aI31
aI53
aI143
aI1
aI841
aI1691
aI749
aI6713
aI5373
aI23487
aI25749
aI13911
aI240923
asS'd'
I19754
sa(dp39508
S'a'
I106564
sS's'
I18
sS'm'
(lp39509
I1
aI3
aI5
aI3
aI7
aI39
aI101
aI83
aI159
aI187
aI583
aI721
aI7745
aI1119
aI61
aI27319
aI35157
aI241729
asS'd'
I19755
sa(dp39510
S'a'
I106576
sS's'
I18
sS'm'
(lp39511
I1
aI1
aI5
aI11
aI27
aI21
aI79
aI199
aI179
aI463
aI987
aI3909
aI1741
aI503
aI12339
aI15323
aI4063
aI180337
asS'd'
I19756
sa(dp39512
S'a'
I106586
sS's'
I18
sS'm'
(lp39513
I1
aI3
aI7
aI11
aI23
aI43
aI7
aI115
aI489
aI215
aI209
aI3213
aI4057
aI13221
aI27061
aI52037
aI34921
aI36385
asS'd'
I19757
sa(dp39514
S'a'
I106592
sS's'
I18
sS'm'
(lp39515
I1
aI1
aI7
aI1
aI15
aI13
aI113
aI55
aI207
aI839
aI1939
aI4095
aI5629
aI7647
aI12753
aI59739
aI60779
aI196589
asS'd'
I19758
sa(dp39516
S'a'
I106597
sS's'
I18
sS'm'
(lp39517
I1
aI1
aI1
aI5
aI5
aI13
aI113
aI243
aI297
aI513
aI1615
aI1513
aI1247
aI4025
aI20901
aI44775
aI86987
aI75437
asS'd'
I19759
sa(dp39518
S'a'
I106610
sS's'
I18
sS'm'
(lp39519
I1
aI1
aI3
aI7
aI7
aI35
aI77
aI115
aI223
aI929
aI1683
aI949
aI2191
aI15533
aI29471
aI24103
aI102475
aI173027
asS'd'
I19760
sa(dp39520
S'a'
I106612
sS's'
I18
sS'm'
(lp39521
I1
aI3
aI1
aI11
aI23
aI37
aI7
aI179
aI287
aI267
aI319
aI3147
aI1481
aI12297
aI28185
aI51423
aI7505
aI236225
asS'd'
I19761
sa(dp39522
S'a'
I106635
sS's'
I18
sS'm'
(lp39523
I1
aI1
aI3
aI5
aI27
aI5
aI71
aI95
aI289
aI121
aI939
aI3543
aI365
aI4903
aI10091
aI3903
aI111155
aI83517
asS'd'
I19762
sa(dp39524
S'a'
I106640
sS's'
I18
sS'm'
(lp39525
I1
aI1
aI7
aI9
aI29
aI3
aI87
aI245
aI313
aI973
aI1181
aI3389
aI3697
aI13237
aI13703
aI31557
aI17269
aI17329
asS'd'
I19763
sa(dp39526
S'a'
I106673
sS's'
I18
sS'm'
(lp39527
I1
aI1
aI5
aI11
aI5
aI41
aI117
aI51
aI239
aI907
aI809
aI741
aI5327
aI3403
aI11825
aI46981
aI93485
aI38053
asS'd'
I19764
sa(dp39528
S'a'
I106691
sS's'
I18
sS'm'
(lp39529
I1
aI1
aI7
aI1
aI25
aI41
aI5
aI169
aI11
aI599
aI1451
aI2189
aI7255
aI8441
aI11169
aI58313
aI4387
aI69
asS'd'
I19765
sa(dp39530
S'a'
I106724
sS's'
I18
sS'm'
(lp39531
I1
aI3
aI1
aI13
aI5
aI21
aI75
aI229
aI153
aI355
aI1511
aI175
aI4793
aI12111
aI25321
aI39983
aI84205
aI195003
asS'd'
I19766
sa(dp39532
S'a'
I106731
sS's'
I18
sS'm'
(lp39533
I1
aI1
aI7
aI1
aI17
aI61
aI67
aI181
aI69
aI149
aI921
aI1107
aI6319
aI431
aI29481
aI12507
aI13109
aI29527
asS'd'
I19767
sa(dp39534
S'a'
I106741
sS's'
I18
sS'm'
(lp39535
I1
aI1
aI5
aI5
aI27
aI47
aI69
aI119
aI469
aI193
aI513
aI1573
aI7421
aI2723
aI20997
aI59585
aI49645
aI261259
asS'd'
I19768
sa(dp39536
S'a'
I106748
sS's'
I18
sS'm'
(lp39537
I1
aI1
aI3
aI9
aI27
aI41
aI41
aI3
aI181
aI803
aI1281
aI2937
aI6745
aI5629
aI8403
aI18987
aI98411
aI128321
asS'd'
I19769
sa(dp39538
S'a'
I106765
sS's'
I18
sS'm'
(lp39539
I1
aI3
aI3
aI9
aI5
aI29
aI55
aI223
aI309
aI841
aI1049
aI1163
aI3497
aI8935
aI8529
aI51367
aI90693
aI77463
asS'd'
I19770
sa(dp39540
S'a'
I106789
sS's'
I18
sS'm'
(lp39541
I1
aI3
aI5
aI3
aI3
aI59
aI23
aI31
aI309
aI907
aI107
aI3471
aI4365
aI14463
aI24093
aI35435
aI24587
aI151163
asS'd'
I19771
sa(dp39542
S'a'
I106822
sS's'
I18
sS'm'
(lp39543
I1
aI3
aI3
aI1
aI11
aI37
aI99
aI95
aI485
aI601
aI1797
aI891
aI5645
aI8927
aI22085
aI5185
aI18495
aI260455
asS'd'
I19772
sa(dp39544
S'a'
I106856
sS's'
I18
sS'm'
(lp39545
I1
aI3
aI7
aI15
aI31
aI61
aI5
aI177
aI159
aI287
aI311
aI1377
aI1419
aI4387
aI25297
aI22505
aI100937
aI223785
asS'd'
I19773
sa(dp39546
S'a'
I106861
sS's'
I18
sS'm'
(lp39547
I1
aI3
aI3
aI3
aI27
aI31
aI45
aI171
aI95
aI507
aI1475
aI4013
aI2781
aI133
aI6857
aI3367
aI103455
aI69559
asS'd'
I19774
sa(dp39548
S'a'
I106870
sS's'
I18
sS'm'
(lp39549
I1
aI3
aI3
aI7
aI9
aI51
aI47
aI247
aI213
aI665
aI1929
aI2799
aI5513
aI9183
aI20197
aI14831
aI75277
aI187565
asS'd'
I19775
sa(dp39550
S'a'
I106883
sS's'
I18
sS'm'
(lp39551
I1
aI3
aI1
aI9
aI13
aI11
aI15
aI253
aI145
aI31
aI847
aI1579
aI5513
aI9
aI3327
aI46049
aI16709
aI56353
asS'd'
I19776
sa(dp39552
S'a'
I106895
sS's'
I18
sS'm'
(lp39553
I1
aI3
aI3
aI3
aI9
aI17
aI59
aI131
aI3
aI621
aI1209
aI3415
aI2999
aI127
aI629
aI7925
aI6109
aI59743
asS'd'
I19777
sa(dp39554
S'a'
I106928
sS's'
I18
sS'm'
(lp39555
I1
aI1
aI5
aI15
aI29
aI35
aI87
aI197
aI495
aI671
aI51
aI293
aI3943
aI7969
aI4739
aI10161
aI119943
aI97217
asS'd'
I19778
sa(dp39556
S'a'
I106938
sS's'
I18
sS'm'
(lp39557
I1
aI1
aI1
aI9
aI25
aI57
aI61
aI197
aI139
aI899
aI783
aI3835
aI3407
aI16301
aI19033
aI33359
aI56309
aI16237
asS'd'
I19779
sa(dp39558
S'a'
I106979
sS's'
I18
sS'm'
(lp39559
I1
aI1
aI5
aI11
aI25
aI47
aI95
aI121
aI197
aI511
aI227
aI2281
aI5805
aI10581
aI14885
aI19685
aI28075
aI25431
asS'd'
I19780
sa(dp39560
S'a'
I106985
sS's'
I18
sS'm'
(lp39561
I1
aI1
aI3
aI9
aI1
aI43
aI59
aI105
aI319
aI45
aI1567
aI905
aI7641
aI2199
aI3979
aI13717
aI22829
aI44777
asS'd'
I19781
sa(dp39562
S'a'
I107005
sS's'
I18
sS'm'
(lp39563
I1
aI3
aI7
aI1
aI19
aI49
aI105
aI53
aI227
aI293
aI989
aI697
aI2553
aI4561
aI14851
aI8999
aI74815
aI207475
asS'd'
I19782
sa(dp39564
S'a'
I107012
sS's'
I18
sS'm'
(lp39565
I1
aI1
aI5
aI11
aI23
aI3
aI85
aI29
aI419
aI847
aI1385
aI3857
aI641
aI14951
aI25629
aI18019
aI2497
aI24723
asS'd'
I19783
sa(dp39566
S'a'
I107019
sS's'
I18
sS'm'
(lp39567
I1
aI3
aI5
aI15
aI5
aI11
aI51
aI225
aI179
aI695
aI787
aI663
aI7051
aI3595
aI4987
aI53315
aI88693
aI7915
asS'd'
I19784
sa(dp39568
S'a'
I107045
sS's'
I18
sS'm'
(lp39569
I1
aI3
aI1
aI13
aI17
aI51
aI51
aI247
aI431
aI555
aI603
aI3301
aI443
aI629
aI26509
aI32047
aI54433
aI208297
asS'd'
I19785
sa(dp39570
S'a'
I107052
sS's'
I18
sS'm'
(lp39571
I1
aI3
aI7
aI7
aI13
aI41
aI95
aI105
aI135
aI443
aI377
aI1259
aI3301
aI6945
aI11677
aI33869
aI107799
aI186567
asS'd'
I19786
sa(dp39572
S'a'
I107082
sS's'
I18
sS'm'
(lp39573
I1
aI1
aI5
aI7
aI31
aI17
aI85
aI129
aI409
aI781
aI983
aI25
aI6877
aI83
aI12625
aI31919
aI41989
aI55195
asS'd'
I19787
sa(dp39574
S'a'
I107089
sS's'
I18
sS'm'
(lp39575
I1
aI1
aI3
aI3
aI7
aI45
aI37
aI45
aI237
aI967
aI1371
aI657
aI7983
aI3121
aI32707
aI25757
aI49987
aI92683
asS'd'
I19788
sa(dp39576
S'a'
I107118
sS's'
I18
sS'm'
(lp39577
I1
aI3
aI1
aI1
aI19
aI57
aI63
aI25
aI355
aI893
aI2017
aI1671
aI7343
aI4451
aI28243
aI22157
aI103901
aI178017
asS'd'
I19789
sa(dp39578
S'a'
I107163
sS's'
I18
sS'm'
(lp39579
I1
aI3
aI7
aI13
aI17
aI59
aI81
aI99
aI329
aI117
aI1395
aI2565
aI5725
aI2371
aI343
aI4713
aI35077
aI49793
asS'd'
I19790
sa(dp39580
S'a'
I107202
sS's'
I18
sS'm'
(lp39581
I1
aI3
aI7
aI15
aI31
aI19
aI87
aI243
aI411
aI339
aI1063
aI117
aI1827
aI1583
aI12571
aI23153
aI3363
aI81695
asS'd'
I19791
sa(dp39582
S'a'
I107238
sS's'
I18
sS'm'
(lp39583
I1
aI3
aI7
aI1
aI3
aI49
aI95
aI133
aI295
aI433
aI1885
aI843
aI6679
aI13673
aI32277
aI59085
aI46957
aI217037
asS'd'
I19792
sa(dp39584
S'a'
I107242
sS's'
I18
sS'm'
(lp39585
I1
aI1
aI3
aI9
aI23
aI53
aI45
aI183
aI79
aI55
aI1267
aI283
aI3249
aI4101
aI8107
aI54473
aI126141
aI127235
asS'd'
I19793
sa(dp39586
S'a'
I107244
sS's'
I18
sS'm'
(lp39587
I1
aI3
aI5
aI7
aI19
aI57
aI113
aI205
aI37
aI817
aI929
aI3643
aI2231
aI15725
aI1733
aI7877
aI116741
aI254529
asS'd'
I19794
sa(dp39588
S'a'
I107264
sS's'
I18
sS'm'
(lp39589
I1
aI3
aI1
aI9
aI9
aI7
aI75
aI45
aI83
aI203
aI1401
aI445
aI1043
aI239
aI30865
aI32189
aI91081
aI180681
asS'd'
I19795
sa(dp39590
S'a'
I107273
sS's'
I18
sS'm'
(lp39591
I1
aI1
aI7
aI7
aI1
aI3
aI61
aI255
aI483
aI599
aI897
aI1601
aI5189
aI13279
aI4981
aI37235
aI117505
aI66625
asS'd'
I19796
sa(dp39592
S'a'
I107291
sS's'
I18
sS'm'
(lp39593
I1
aI1
aI3
aI13
aI29
aI43
aI65
aI73
aI213
aI307
aI959
aI2735
aI5155
aI16063
aI15745
aI6225
aI50159
aI182445
asS'd'
I19797
sa(dp39594
S'a'
I107293
sS's'
I18
sS'm'
(lp39595
I1
aI1
aI3
aI9
aI31
aI61
aI73
aI185
aI457
aI687
aI1887
aI4041
aI3455
aI4739
aI16399
aI40929
aI32631
aI179031
asS'd'
I19798
sa(dp39596
S'a'
I107312
sS's'
I18
sS'm'
(lp39597
I1
aI3
aI3
aI15
aI5
aI45
aI1
aI241
aI187
aI23
aI63
aI2497
aI7759
aI9175
aI11003
aI40579
aI45769
aI107133
asS'd'
I19799
sa(dp39598
S'a'
I107318
sS's'
I18
sS'm'
(lp39599
I1
aI3
aI3
aI7
aI5
aI63
aI7
aI67
aI31
aI917
aI1825
aI2037
aI2527
aI9767
aI2665
aI18573
aI14289
aI21583
asS'd'
I19800
sa(dp39600
S'a'
I107335
sS's'
I18
sS'm'
(lp39601
I1
aI1
aI5
aI15
aI15
aI25
aI51
aI127
aI261
aI925
aI1651
aI769
aI7779
aI7327
aI7239
aI20437
aI9947
aI144697
asS'd'
I19801
sa(dp39602
S'a'
I107349
sS's'
I18
sS'm'
(lp39603
I1
aI1
aI5
aI1
aI5
aI13
aI99
aI7
aI269
aI135
aI345
aI1851
aI7963
aI457
aI24573
aI32529
aI127359
aI157755
asS'd'
I19802
sa(dp39604
S'a'
I107350
sS's'
I18
sS'm'
(lp39605
I1
aI3
aI5
aI15
aI21
aI17
aI31
aI115
aI321
aI351
aI117
aI1301
aI2455
aI3363
aI14213
aI62903
aI75273
aI261119
asS'd'
I19803
sa(dp39606
S'a'
I107363
sS's'
I18
sS'm'
(lp39607
I1
aI3
aI7
aI1
aI1
aI9
aI53
aI209
aI319
aI489
aI827
aI1365
aI4709
aI7419
aI28213
aI56095
aI9611
aI234877
asS'd'
I19804
sa(dp39608
S'a'
I107390
sS's'
I18
sS'm'
(lp39609
I1
aI1
aI5
aI5
aI1
aI7
aI49
aI15
aI377
aI309
aI1701
aI1775
aI5571
aI12437
aI27521
aI54753
aI68977
aI138549
asS'd'
I19805
sa(dp39610
S'a'
I107405
sS's'
I18
sS'm'
(lp39611
I1
aI1
aI3
aI11
aI9
aI29
aI67
aI21
aI411
aI647
aI983
aI1075
aI2387
aI13355
aI1781
aI56129
aI87235
aI260133
asS'd'
I19806
sa(dp39612
S'a'
I107408
sS's'
I18
sS'm'
(lp39613
I1
aI3
aI1
aI11
aI5
aI57
aI71
aI159
aI345
aI853
aI745
aI3733
aI1607
aI7265
aI20097
aI18911
aI101141
aI70495
asS'd'
I19807
sa(dp39614
S'a'
I107427
sS's'
I18
sS'm'
(lp39615
I1
aI1
aI3
aI9
aI1
aI43
aI127
aI127
aI471
aI465
aI1777
aI1879
aI4655
aI12925
aI24935
aI58445
aI78303
aI200463
asS'd'
I19808
sa(dp39616
S'a'
I107433
sS's'
I18
sS'm'
(lp39617
I1
aI1
aI1
aI15
aI5
aI33
aI63
aI17
aI401
aI831
aI1559
aI3547
aI7869
aI13901
aI18185
aI9399
aI65859
aI185315
asS'd'
I19809
sa(dp39618
S'a'
I107444
sS's'
I18
sS'm'
(lp39619
I1
aI3
aI7
aI1
aI25
aI21
aI17
aI175
aI401
aI833
aI847
aI3593
aI1283
aI14745
aI11827
aI1987
aI89299
aI187369
asS'd'
I19810
sa(dp39620
S'a'
I107514
sS's'
I18
sS'm'
(lp39621
I1
aI1
aI7
aI13
aI29
aI31
aI45
aI219
aI177
aI369
aI1313
aI3015
aI5859
aI1829
aI8793
aI49109
aI97581
aI233179
asS'd'
I19811
sa(dp39622
S'a'
I107522
sS's'
I18
sS'm'
(lp39623
I1
aI3
aI3
aI15
aI17
aI31
aI61
aI215
aI231
aI495
aI1307
aI3067
aI3187
aI8813
aI22505
aI14055
aI30773
aI177955
asS'd'
I19812
sa(dp39624
S'a'
I107545
sS's'
I18
sS'm'
(lp39625
I1
aI3
aI7
aI15
aI9
aI61
aI57
aI105
aI89
aI267
aI233
aI905
aI3727
aI1841
aI31875
aI32499
aI27003
aI208491
asS'd'
I19813
sa(dp39626
S'a'
I107582
sS's'
I18
sS'm'
(lp39627
I1
aI3
aI1
aI11
aI31
aI9
aI63
aI55
aI213
aI209
aI1625
aI2635
aI4335
aI15201
aI6127
aI11097
aI37991
aI62813
asS'd'
I19814
sa(dp39628
S'a'
I107590
sS's'
I18
sS'm'
(lp39629
I1
aI1
aI1
aI5
aI13
aI49
aI127
aI249
aI339
aI525
aI1943
aI2935
aI3255
aI5199
aI6869
aI6325
aI731
aI51085
asS'd'
I19815
sa(dp39630
S'a'
I107593
sS's'
I18
sS'm'
(lp39631
I1
aI1
aI5
aI7
aI7
aI29
aI59
aI187
aI463
aI409
aI1321
aI377
aI7361
aI8303
aI20385
aI39649
aI81379
aI235555
asS'd'
I19816
sa(dp39632
S'a'
I107599
sS's'
I18
sS'm'
(lp39633
I1
aI1
aI5
aI13
aI15
aI45
aI99
aI57
aI217
aI535
aI1747
aI4081
aI4781
aI1005
aI25197
aI23853
aI90587
aI189579
asS'd'
I19817
sa(dp39634
S'a'
I107604
sS's'
I18
sS'm'
(lp39635
I1
aI1
aI7
aI7
aI23
aI31
aI1
aI151
aI339
aI447
aI523
aI2609
aI5917
aI6965
aI25815
aI62797
aI104083
aI245917
asS'd'
I19818
sa(dp39636
S'a'
I107608
sS's'
I18
sS'm'
(lp39637
I1
aI3
aI5
aI3
aI7
aI37
aI79
aI253
aI423
aI511
aI1477
aI3121
aI5557
aI1413
aI31075
aI22249
aI117639
aI243543
asS'd'
I19819
sa(dp39638
S'a'
I107627
sS's'
I18
sS'm'
(lp39639
I1
aI1
aI3
aI7
aI5
aI31
aI93
aI117
aI135
aI235
aI745
aI3287
aI4451
aI2487
aI15179
aI62229
aI18247
aI150277
asS'd'
I19820
sa(dp39640
S'a'
I107638
sS's'
I18
sS'm'
(lp39641
I1
aI1
aI7
aI15
aI31
aI57
aI125
aI219
aI433
aI629
aI1809
aI2499
aI6083
aI7631
aI31495
aI63183
aI28533
aI120579
asS'd'
I19821
sa(dp39642
S'a'
I107666
sS's'
I18
sS'm'
(lp39643
I1
aI1
aI1
aI13
aI31
aI47
aI77
aI73
aI343
aI867
aI1055
aI1121
aI3035
aI15693
aI6971
aI50231
aI16527
aI190795
asS'd'
I19822
sa(dp39644
S'a'
I107678
sS's'
I18
sS'm'
(lp39645
I1
aI3
aI7
aI7
aI11
aI1
aI5
aI215
aI87
aI885
aI1429
aI1277
aI3831
aI9341
aI22011
aI34585
aI56167
aI65301
asS'd'
I19823
sa(dp39646
S'a'
I107701
sS's'
I18
sS'm'
(lp39647
I1
aI3
aI1
aI15
aI13
aI53
aI91
aI59
aI277
aI819
aI453
aI3191
aI757
aI4729
aI20605
aI4283
aI110745
aI233655
asS'd'
I19824
sa(dp39648
S'a'
I107720
sS's'
I18
sS'm'
(lp39649
I1
aI3
aI3
aI5
aI17
aI47
aI69
aI117
aI113
aI775
aI935
aI1879
aI2239
aI5877
aI18337
aI50895
aI44891
aI2759
asS'd'
I19825
sa(dp39650
S'a'
I107723
sS's'
I18
sS'm'
(lp39651
I1
aI3
aI5
aI1
aI27
aI31
aI77
aI65
aI355
aI405
aI825
aI2419
aI3337
aI10179
aI27665
aI35459
aI13721
aI154227
asS'd'
I19826
sa(dp39652
S'a'
I107725
sS's'
I18
sS'm'
(lp39653
I1
aI3
aI3
aI7
aI23
aI61
aI9
aI241
aI239
aI85
aI485
aI2659
aI5371
aI16175
aI28691
aI49109
aI124433
aI167033
asS'd'
I19827
sa(dp39654
S'a'
I107731
sS's'
I18
sS'm'
(lp39655
I1
aI3
aI3
aI3
aI11
aI57
aI37
aI155
aI443
aI249
aI1913
aI1347
aI6569
aI5357
aI4231
aI58273
aI50707
aI57097
asS'd'
I19828
sa(dp39656
S'a'
I107737
sS's'
I18
sS'm'
(lp39657
I1
aI1
aI3
aI3
aI1
aI7
aI87
aI115
aI259
aI807
aI45
aI2997
aI63
aI16313
aI12507
aI39925
aI17699
aI24411
asS'd'
I19829
sa(dp39658
S'a'
I107767
sS's'
I18
sS'm'
(lp39659
I1
aI3
aI3
aI1
aI19
aI59
aI97
aI209
aI247
aI395
aI21
aI1803
aI2547
aI11607
aI15703
aI58099
aI111907
aI196101
asS'd'
I19830
sa(dp39660
S'a'
I107779
sS's'
I18
sS'm'
(lp39661
I1
aI1
aI7
aI13
aI27
aI61
aI73
aI183
aI493
aI953
aI445
aI567
aI7373
aI15275
aI30081
aI539
aI89365
aI3455
asS'd'
I19831
sa(dp39662
S'a'
I107788
sS's'
I18
sS'm'
(lp39663
I1
aI3
aI1
aI3
aI25
aI35
aI3
aI105
aI243
aI781
aI1937
aI2781
aI7849
aI2159
aI2221
aI58005
aI89313
aI182183
asS'd'
I19832
sa(dp39664
S'a'
I107799
sS's'
I18
sS'm'
(lp39665
I1
aI1
aI7
aI15
aI21
aI7
aI67
aI163
aI179
aI453
aI581
aI2245
aI3915
aI8985
aI16809
aI35113
aI93605
aI79009
asS'd'
I19833
sa(dp39666
S'a'
I107819
sS's'
I18
sS'm'
(lp39667
I1
aI1
aI7
aI5
aI29
aI57
aI13
aI75
aI387
aI511
aI331
aI1119
aI307
aI6145
aI3841
aI49987
aI67335
aI120419
asS'd'
I19834
sa(dp39668
S'a'
I107833
sS's'
I18
sS'm'
(lp39669
I1
aI3
aI1
aI5
aI13
aI21
aI119
aI207
aI453
aI943
aI137
aI2245
aI7771
aI5737
aI9541
aI29209
aI106867
aI110513
asS'd'
I19835
sa(dp39670
S'a'
I107881
sS's'
I18
sS'm'
(lp39671
I1
aI3
aI7
aI13
aI7
aI3
aI99
aI129
aI245
aI687
aI883
aI1321
aI1131
aI6661
aI7437
aI1345
aI128247
aI167877
asS'd'
I19836
sa(dp39672
S'a'
I107901
sS's'
I18
sS'm'
(lp39673
I1
aI1
aI3
aI11
aI7
aI57
aI59
aI3
aI217
aI549
aI85
aI2607
aI215
aI2249
aI3963
aI42931
aI33747
aI226265
asS'd'
I19837
sa(dp39674
S'a'
I107911
sS's'
I18
sS'm'
(lp39675
I1
aI1
aI5
aI5
aI29
aI23
aI115
aI215
aI103
aI427
aI1689
aI831
aI3293
aI14055
aI3735
aI49521
aI17703
aI182887
asS'd'
I19838
sa(dp39676
S'a'
I107930
sS's'
I18
sS'm'
(lp39677
I1
aI1
aI1
aI15
aI5
aI43
aI27
aI181
aI217
aI955
aI225
aI2731
aI7347
aI14123
aI26169
aI4371
aI26907
aI15017
asS'd'
I19839
sa(dp39678
S'a'
I107936
sS's'
I18
sS'm'
(lp39679
I1
aI1
aI7
aI7
aI31
aI55
aI63
aI223
aI61
aI63
aI431
aI2779
aI3169
aI14323
aI2945
aI63913
aI85407
aI76511
asS'd'
I19840
sa(dp39680
S'a'
I107980
sS's'
I18
sS'm'
(lp39681
I1
aI1
aI3
aI11
aI5
aI27
aI113
aI75
aI313
aI697
aI13
aI1853
aI7467
aI5701
aI16749
aI17939
aI13475
aI39807
asS'd'
I19841
sa(dp39682
S'a'
I108004
sS's'
I18
sS'm'
(lp39683
I1
aI1
aI7
aI5
aI13
aI53
aI55
aI115
aI125
aI341
aI321
aI3291
aI2675
aI13659
aI16819
aI45397
aI42917
aI104361
asS'd'
I19842
sa(dp39684
S'a'
I108008
sS's'
I18
sS'm'
(lp39685
I1
aI3
aI3
aI15
aI27
aI47
aI19
aI213
aI441
aI605
aI593
aI2287
aI4847
aI10505
aI22185
aI36157
aI10881
aI87799
asS'd'
I19843
sa(dp39686
S'a'
I108016
sS's'
I18
sS'm'
(lp39687
I1
aI1
aI1
aI11
aI15
aI39
aI109
aI3
aI469
aI371
aI1743
aI2789
aI199
aI8703
aI7407
aI3353
aI103417
aI73319
asS'd'
I19844
sa(dp39688
S'a'
I108066
sS's'
I18
sS'm'
(lp39689
I1
aI3
aI7
aI5
aI25
aI61
aI77
aI111
aI263
aI399
aI1579
aI3447
aI6205
aI5945
aI28099
aI39183
aI77003
aI115001
asS'd'
I19845
sa(dp39690
S'a'
I108071
sS's'
I18
sS'm'
(lp39691
I1
aI1
aI1
aI3
aI1
aI7
aI57
aI193
aI379
aI923
aI151
aI2227
aI7285
aI2371
aI24567
aI34037
aI80655
aI125499
asS'd'
I19846
sa(dp39692
S'a'
I108107
sS's'
I18
sS'm'
(lp39693
I1
aI3
aI5
aI7
aI13
aI15
aI5
aI193
aI55
aI319
aI1851
aI2439
aI8071
aI5329
aI3155
aI64669
aI18547
aI238997
asS'd'
I19847
sa(dp39694
S'a'
I108110
sS's'
I18
sS'm'
(lp39695
I1
aI3
aI3
aI11
aI7
aI23
aI1
aI203
aI333
aI951
aI153
aI1249
aI5093
aI407
aI361
aI14175
aI45149
aI186291
asS'd'
I19848
sa(dp39696
S'a'
I108124
sS's'
I18
sS'm'
(lp39697
I1
aI3
aI1
aI11
aI13
aI19
aI89
aI229
aI139
aI981
aI455
aI907
aI5109
aI8513
aI23823
aI54933
aI69985
aI227679
asS'd'
I19849
sa(dp39698
S'a'
I108145
sS's'
I18
sS'm'
(lp39699
I1
aI3
aI5
aI13
aI23
aI47
aI65
aI123
aI115
aI759
aI375
aI1283
aI729
aI11045
aI22015
aI18287
aI112603
aI75911
asS'd'
I19850
sa(dp39700
S'a'
I108148
sS's'
I18
sS'm'
(lp39701
I1
aI1
aI3
aI1
aI5
aI43
aI91
aI123
aI219
aI409
aI517
aI3999
aI1409
aI5949
aI15727
aI62705
aI73573
aI198447
asS'd'
I19851
sa(dp39702
S'a'
I108158
sS's'
I18
sS'm'
(lp39703
I1
aI3
aI3
aI3
aI5
aI23
aI95
aI51
aI275
aI455
aI1831
aI2427
aI3779
aI10209
aI30839
aI23393
aI25681
aI8715
asS'd'
I19852
sa(dp39704
S'a'
I108164
sS's'
I18
sS'm'
(lp39705
I1
aI3
aI5
aI3
aI9
aI5
aI61
aI97
aI497
aI397
aI695
aI517
aI3313
aI4911
aI1961
aI45805
aI99135
aI216657
asS'd'
I19853
sa(dp39706
S'a'
I108197
sS's'
I18
sS'm'
(lp39707
I1
aI1
aI1
aI7
aI13
aI41
aI19
aI31
aI103
aI1005
aI73
aI1855
aI405
aI12395
aI19979
aI17663
aI105183
aI28493
asS'd'
I19854
sa(dp39708
S'a'
I108198
sS's'
I18
sS'm'
(lp39709
I1
aI3
aI1
aI11
aI27
aI27
aI69
aI149
aI1
aI225
aI1809
aI1367
aI3663
aI6545
aI8475
aI40837
aI64459
aI66705
asS'd'
I19855
sa(dp39710
S'a'
I108219
sS's'
I18
sS'm'
(lp39711
I1
aI3
aI3
aI1
aI29
aI21
aI113
aI149
aI215
aI443
aI1069
aI3437
aI1793
aI13573
aI28285
aI33707
aI29127
aI40715
asS'd'
I19856
sa(dp39712
S'a'
I108222
sS's'
I18
sS'm'
(lp39713
I1
aI3
aI3
aI5
aI21
aI9
aI53
aI181
aI455
aI283
aI245
aI3999
aI875
aI9799
aI10963
aI31603
aI34907
aI64977
asS'd'
I19857
sa(dp39714
S'a'
I108224
sS's'
I18
sS'm'
(lp39715
I1
aI3
aI1
aI11
aI31
aI51
aI29
aI103
aI61
aI529
aI777
aI1097
aI445
aI9169
aI6305
aI4513
aI60189
aI164103
asS'd'
I19858
sa(dp39716
S'a'
I108227
sS's'
I18
sS'm'
(lp39717
I1
aI1
aI1
aI3
aI21
aI5
aI87
aI11
aI461
aI637
aI1283
aI1471
aI1429
aI2401
aI12163
aI29401
aI30089
aI41745
asS'd'
I19859
sa(dp39718
S'a'
I108290
sS's'
I18
sS'm'
(lp39719
I1
aI1
aI3
aI3
aI17
aI43
aI13
aI153
aI73
aI419
aI747
aI279
aI7195
aI4383
aI26345
aI28365
aI97753
aI208989
asS'd'
I19860
sa(dp39720
S'a'
I108302
sS's'
I18
sS'm'
(lp39721
I1
aI3
aI7
aI1
aI17
aI55
aI103
aI151
aI327
aI575
aI1923
aI1533
aI4699
aI2171
aI15447
aI64047
aI59007
aI54523
asS'd'
I19861
sa(dp39722
S'a'
I108329
sS's'
I18
sS'm'
(lp39723
I1
aI3
aI7
aI7
aI3
aI51
aI79
aI39
aI403
aI599
aI161
aI2465
aI4911
aI10327
aI23599
aI3267
aI44177
aI184231
asS'd'
I19862
sa(dp39724
S'a'
I108335
sS's'
I18
sS'm'
(lp39725
I1
aI3
aI3
aI15
aI5
aI19
aI57
aI83
aI507
aI927
aI665
aI3471
aI5037
aI2051
aI22923
aI36813
aI97393
aI102715
asS'd'
I19863
sa(dp39726
S'a'
I108338
sS's'
I18
sS'm'
(lp39727
I1
aI3
aI3
aI13
aI11
aI19
aI91
aI179
aI113
aI295
aI855
aI2071
aI3265
aI4089
aI8627
aI7461
aI23855
aI53675
asS'd'
I19864
sa(dp39728
S'a'
I108362
sS's'
I18
sS'm'
(lp39729
I1
aI1
aI3
aI5
aI21
aI41
aI11
aI227
aI87
aI417
aI209
aI2013
aI4849
aI5291
aI22073
aI40235
aI71283
aI140785
asS'd'
I19865
sa(dp39730
S'a'
I108398
sS's'
I18
sS'm'
(lp39731
I1
aI3
aI3
aI5
aI29
aI57
aI95
aI65
aI177
aI177
aI1973
aI563
aI5249
aI337
aI7611
aI45099
aI15443
aI46583
asS'd'
I19866
sa(dp39732
S'a'
I108405
sS's'
I18
sS'm'
(lp39733
I1
aI1
aI7
aI3
aI3
aI5
aI107
aI75
aI31
aI293
aI821
aI1837
aI2363
aI13621
aI8793
aI29841
aI127201
aI131707
asS'd'
I19867
sa(dp39734
S'a'
I108409
sS's'
I18
sS'm'
(lp39735
I1
aI1
aI1
aI5
aI1
aI45
aI69
aI61
aI157
aI999
aI183
aI3431
aI3487
aI9461
aI17545
aI26973
aI115527
aI58419
asS'd'
I19868
sa(dp39736
S'a'
I108431
sS's'
I18
sS'm'
(lp39737
I1
aI1
aI7
aI7
aI27
aI5
aI125
aI153
aI191
aI745
aI125
aI2807
aI5043
aI10657
aI24487
aI19517
aI31735
aI42421
asS'd'
I19869
sa(dp39738
S'a'
I108449
sS's'
I18
sS'm'
(lp39739
I1
aI1
aI7
aI9
aI25
aI37
aI73
aI255
aI141
aI229
aI1723
aI1331
aI6089
aI13109
aI30683
aI2335
aI111517
aI105411
asS'd'
I19870
sa(dp39740
S'a'
I108474
sS's'
I18
sS'm'
(lp39741
I1
aI1
aI5
aI7
aI3
aI61
aI79
aI203
aI423
aI669
aI1757
aI1725
aI4239
aI7013
aI28591
aI61853
aI81103
aI39813
asS'd'
I19871
sa(dp39742
S'a'
I108488
sS's'
I18
sS'm'
(lp39743
I1
aI1
aI3
aI7
aI21
aI27
aI23
aI119
aI441
aI113
aI1019
aI285
aI53
aI8643
aI31689
aI2629
aI126573
aI60835
asS'd'
I19872
sa(dp39744
S'a'
I108530
sS's'
I18
sS'm'
(lp39745
I1
aI1
aI5
aI11
aI29
aI49
aI23
aI55
aI441
aI809
aI1177
aI1371
aI5945
aI6461
aI11537
aI12287
aI85637
aI232065
asS'd'
I19873
sa(dp39746
S'a'
I108545
sS's'
I18
sS'm'
(lp39747
I1
aI3
aI1
aI15
aI5
aI13
aI19
aI209
aI105
aI897
aI565
aI3209
aI487
aI9837
aI22169
aI26317
aI39907
aI185193
asS'd'
I19874
sa(dp39748
S'a'
I108551
sS's'
I18
sS'm'
(lp39749
I1
aI1
aI1
aI5
aI23
aI45
aI49
aI55
aI501
aI213
aI1217
aI3159
aI733
aI5889
aI5475
aI4953
aI37317
aI100369
asS'd'
I19875
sa(dp39750
S'a'
I108558
sS's'
I18
sS'm'
(lp39751
I1
aI1
aI5
aI1
aI19
aI5
aI57
aI137
aI361
aI605
aI1077
aI2015
aI5511
aI4667
aI18457
aI45979
aI120253
aI203397
asS'd'
I19876
sa(dp39752
S'a'
I108585
sS's'
I18
sS'm'
(lp39753
I1
aI1
aI7
aI9
aI19
aI19
aI11
aI187
aI5
aI647
aI275
aI1265
aI7587
aI16183
aI369
aI31885
aI58347
aI36829
asS'd'
I19877
sa(dp39754
S'a'
I108613
sS's'
I18
sS'm'
(lp39755
I1
aI3
aI5
aI3
aI31
aI31
aI7
aI105
aI359
aI839
aI641
aI3215
aI4807
aI13397
aI885
aI52867
aI57125
aI180607
asS'd'
I19878
sa(dp39756
S'a'
I108638
sS's'
I18
sS'm'
(lp39757
I1
aI1
aI7
aI9
aI19
aI45
aI43
aI211
aI429
aI757
aI1637
aI1569
aI935
aI8899
aI24823
aI18599
aI111373
aI190979
asS'd'
I19879
sa(dp39758
S'a'
I108642
sS's'
I18
sS'm'
(lp39759
I1
aI1
aI7
aI7
aI13
aI47
aI3
aI241
aI467
aI209
aI323
aI3467
aI4397
aI15395
aI15373
aI14499
aI92443
aI65931
asS'd'
I19880
sa(dp39760
S'a'
I108653
sS's'
I18
sS'm'
(lp39761
I1
aI3
aI3
aI5
aI5
aI9
aI67
aI115
aI45
aI279
aI1753
aI1575
aI8127
aI9651
aI5169
aI25643
aI29671
aI214383
asS'd'
I19881
sa(dp39762
S'a'
I108671
sS's'
I18
sS'm'
(lp39763
I1
aI3
aI7
aI7
aI29
aI7
aI37
aI205
aI495
aI445
aI1771
aI1439
aI3577
aI10423
aI10865
aI26851
aI15251
aI63373
asS'd'
I19882
sa(dp39764
S'a'
I108675
sS's'
I18
sS'm'
(lp39765
I1
aI3
aI1
aI9
aI29
aI13
aI45
aI61
aI153
aI193
aI1407
aI1075
aI4023
aI2367
aI1147
aI51277
aI52975
aI123061
asS'd'
I19883
sa(dp39766
S'a'
I108684
sS's'
I18
sS'm'
(lp39767
I1
aI3
aI3
aI3
aI31
aI25
aI113
aI173
aI345
aI565
aI145
aI3437
aI7051
aI12557
aI27919
aI41733
aI76717
aI192645
asS'd'
I19884
sa(dp39768
S'a'
I108687
sS's'
I18
sS'm'
(lp39769
I1
aI3
aI3
aI7
aI3
aI55
aI35
aI219
aI55
aI467
aI635
aI3833
aI3753
aI1099
aI15301
aI53121
aI120807
aI70481
asS'd'
I19885
sa(dp39770
S'a'
I108696
sS's'
I18
sS'm'
(lp39771
I1
aI3
aI7
aI7
aI13
aI1
aI121
aI191
aI71
aI193
aI1891
aI2303
aI1401
aI9143
aI31297
aI38979
aI43093
aI138941
asS'd'
I19886
sa(dp39772
S'a'
I108786
sS's'
I18
sS'm'
(lp39773
I1
aI3
aI7
aI5
aI17
aI51
aI83
aI201
aI231
aI423
aI511
aI1301
aI6075
aI475
aI2603
aI49327
aI78565
aI220827
asS'd'
I19887
sa(dp39774
S'a'
I108792
sS's'
I18
sS'm'
(lp39775
I1
aI3
aI1
aI7
aI5
aI61
aI23
aI11
aI9
aI711
aI251
aI1383
aI613
aI6213
aI8921
aI27267
aI66009
aI28575
asS'd'
I19888
sa(dp39776
S'a'
I108795
sS's'
I18
sS'm'
(lp39777
I1
aI3
aI1
aI3
aI13
aI13
aI61
aI211
aI353
aI883
aI343
aI1089
aI2041
aI7781
aI25285
aI9053
aI76801
aI153009
asS'd'
I19889
sa(dp39778
S'a'
I108818
sS's'
I18
sS'm'
(lp39779
I1
aI3
aI7
aI5
aI17
aI61
aI67
aI69
aI361
aI937
aI949
aI1483
aI2825
aI3753
aI16533
aI17277
aI16539
aI140521
asS'd'
I19890
sa(dp39780
S'a'
I108830
sS's'
I18
sS'm'
(lp39781
I1
aI3
aI5
aI9
aI7
aI19
aI31
aI239
aI357
aI561
aI1583
aI3059
aI2023
aI2213
aI11283
aI18603
aI83487
aI162415
asS'd'
I19891
sa(dp39782
S'a'
I108836
sS's'
I18
sS'm'
(lp39783
I1
aI1
aI3
aI15
aI15
aI59
aI9
aI43
aI353
aI203
aI765
aI1907
aI2733
aI3747
aI30617
aI32671
aI119005
aI72131
asS'd'
I19892
sa(dp39784
S'a'
I108839
sS's'
I18
sS'm'
(lp39785
I1
aI3
aI1
aI7
aI17
aI61
aI55
aI113
aI439
aI75
aI703
aI2741
aI1059
aI4561
aI15923
aI17153
aI32563
aI79681
asS'd'
I19893
sa(dp39786
S'a'
I108851
sS's'
I18
sS'm'
(lp39787
I1
aI1
aI1
aI15
aI11
aI57
aI91
aI245
aI187
aI185
aI1859
aI1209
aI3247
aI10863
aI22421
aI47287
aI26831
aI200935
asS'd'
I19894
sa(dp39788
S'a'
I108875
sS's'
I18
sS'm'
(lp39789
I1
aI1
aI3
aI13
aI11
aI39
aI61
aI211
aI51
aI197
aI861
aI2965
aI4691
aI9955
aI2289
aI28795
aI61537
aI235359
asS'd'
I19895
sa(dp39790
S'a'
I108883
sS's'
I18
sS'm'
(lp39791
I1
aI1
aI7
aI1
aI31
aI37
aI121
aI107
aI79
aI521
aI371
aI983
aI6189
aI4515
aI25691
aI26933
aI123189
aI70033
asS'd'
I19896
sa(dp39792
S'a'
I108895
sS's'
I18
sS'm'
(lp39793
I1
aI3
aI3
aI15
aI19
aI33
aI31
aI237
aI35
aI877
aI1013
aI3445
aI573
aI1145
aI27781
aI26327
aI117451
aI142339
asS'd'
I19897
sa(dp39794
S'a'
I108908
sS's'
I18
sS'm'
(lp39795
I1
aI3
aI3
aI15
aI15
aI9
aI17
aI185
aI151
aI499
aI493
aI1331
aI1587
aI12657
aI22577
aI7957
aI126253
aI57971
asS'd'
I19898
sa(dp39796
S'a'
I108953
sS's'
I18
sS'm'
(lp39797
I1
aI3
aI7
aI15
aI3
aI29
aI43
aI105
aI423
aI601
aI793
aI1011
aI6657
aI7287
aI18561
aI46993
aI72945
aI233051
asS'd'
I19899
sa(dp39798
S'a'
I108956
sS's'
I18
sS'm'
(lp39799
I1
aI1
aI7
aI5
aI15
aI17
aI125
aI141
aI75
aI877
aI281
aI2011
aI4217
aI9785
aI7587
aI42863
aI35585
aI212795
asS'd'
I19900
sa(dp39800
S'a'
I108963
sS's'
I18
sS'm'
(lp39801
I1
aI1
aI3
aI3
aI3
aI37
aI67
aI129
aI433
aI233
aI115
aI687
aI6253
aI12805
aI10935
aI49963
aI91675
aI241951
asS'd'
I19901
sa(dp39802
S'a'
I108989
sS's'
I18
sS'm'
(lp39803
I1
aI3
aI1
aI5
aI3
aI41
aI3
aI233
aI317
aI959
aI1407
aI1251
aI743
aI4165
aI15561
aI41165
aI22393
aI111633
asS'd'
I19902
sa(dp39804
S'a'
I109001
sS's'
I18
sS'm'
(lp39805
I1
aI3
aI3
aI15
aI17
aI15
aI53
aI245
aI109
aI413
aI1149
aI35
aI2931
aI11635
aI27091
aI63659
aI33973
aI16867
asS'd'
I19903
sa(dp39806
S'a'
I109009
sS's'
I18
sS'm'
(lp39807
I1
aI3
aI7
aI13
aI3
aI43
aI83
aI201
aI367
aI275
aI235
aI1795
aI4041
aI13539
aI22345
aI31451
aI83985
aI3527
asS'd'
I19904
sa(dp39808
S'a'
I109052
sS's'
I18
sS'm'
(lp39809
I1
aI3
aI3
aI9
aI31
aI47
aI63
aI117
aI195
aI497
aI453
aI1183
aI3857
aI14075
aI28057
aI13205
aI54331
aI54641
asS'd'
I19905
sa(dp39810
S'a'
I109083
sS's'
I18
sS'm'
(lp39811
I1
aI1
aI7
aI11
aI11
aI39
aI37
aI183
aI213
aI537
aI1371
aI901
aI5255
aI11791
aI18993
aI58785
aI114113
aI229815
asS'd'
I19906
sa(dp39812
S'a'
I109090
sS's'
I18
sS'm'
(lp39813
I1
aI3
aI5
aI5
aI13
aI45
aI77
aI165
aI329
aI439
aI2011
aI1845
aI4577
aI12457
aI16959
aI45943
aI37715
aI169775
asS'd'
I19907
sa(dp39814
S'a'
I109104
sS's'
I18
sS'm'
(lp39815
I1
aI1
aI3
aI11
aI5
aI7
aI21
aI245
aI365
aI371
aI1291
aI2569
aI1791
aI6003
aI18619
aI18661
aI98551
aI89645
asS'd'
I19908
sa(dp39816
S'a'
I109121
sS's'
I18
sS'm'
(lp39817
I1
aI3
aI1
aI7
aI3
aI51
aI49
aI245
aI151
aI919
aI489
aI3967
aI3157
aI7159
aI17207
aI19749
aI94455
aI112403
asS'd'
I19909
sa(dp39818
S'a'
I109128
sS's'
I18
sS'm'
(lp39819
I1
aI3
aI7
aI5
aI9
aI15
aI17
aI201
aI273
aI669
aI1571
aI3915
aI1859
aI2569
aI28855
aI27225
aI116711
aI148377
asS'd'
I19910
sa(dp39820
S'a'
I109141
sS's'
I18
sS'm'
(lp39821
I1
aI3
aI7
aI3
aI19
aI45
aI115
aI71
aI201
aI85
aI1349
aI3897
aI4941
aI10839
aI14781
aI36487
aI107037
aI109185
asS'd'
I19911
sa(dp39822
S'a'
I109146
sS's'
I18
sS'm'
(lp39823
I1
aI3
aI5
aI13
aI11
aI11
aI35
aI19
aI213
aI41
aI45
aI4075
aI3163
aI12937
aI17487
aI28911
aI21289
aI198489
asS'd'
I19912
sa(dp39824
S'a'
I109155
sS's'
I18
sS'm'
(lp39825
I1
aI1
aI1
aI11
aI7
aI35
aI67
aI23
aI451
aI235
aI717
aI181
aI6697
aI9359
aI24561
aI36187
aI125179
aI35119
asS'd'
I19913
sa(dp39826
S'a'
I109162
sS's'
I18
sS'm'
(lp39827
I1
aI1
aI7
aI15
aI7
aI47
aI55
aI125
aI465
aI251
aI211
aI3137
aI4101
aI1341
aI2287
aI22149
aI45089
aI94173
asS'd'
I19914
sa(dp39828
S'a'
I109209
sS's'
I18
sS'm'
(lp39829
I1
aI3
aI1
aI15
aI19
aI7
aI15
aI207
aI7
aI759
aI869
aI2337
aI6805
aI8287
aI13447
aI9963
aI7177
aI173505
asS'd'
I19915
sa(dp39830
S'a'
I109216
sS's'
I18
sS'm'
(lp39831
I1
aI3
aI7
aI13
aI15
aI3
aI65
aI143
aI291
aI511
aI939
aI669
aI4095
aI1931
aI26015
aI18253
aI102461
aI93837
asS'd'
I19916
sa(dp39832
S'a'
I109243
sS's'
I18
sS'm'
(lp39833
I1
aI3
aI1
aI5
aI23
aI3
aI19
aI121
aI489
aI583
aI1425
aI627
aI4229
aI5343
aI3759
aI17845
aI105369
aI132239
asS'd'
I19917
sa(dp39834
S'a'
I109265
sS's'
I18
sS'm'
(lp39835
I1
aI3
aI3
aI9
aI29
aI9
aI17
aI153
aI111
aI879
aI1225
aI979
aI2687
aI10477
aI10105
aI18091
aI37825
aI28077
asS'd'
I19918
sa(dp39836
S'a'
I109282
sS's'
I18
sS'm'
(lp39837
I1
aI1
aI7
aI9
aI3
aI51
aI75
aI101
aI197
aI551
aI89
aI1441
aI607
aI14025
aI30411
aI26887
aI3435
aI32977
asS'd'
I19919
sa(dp39838
S'a'
I109306
sS's'
I18
sS'm'
(lp39839
I1
aI1
aI7
aI3
aI25
aI29
aI95
aI123
aI197
aI767
aI1513
aI721
aI4149
aI10873
aI32247
aI4469
aI42297
aI49651
asS'd'
I19920
sa(dp39840
S'a'
I109316
sS's'
I18
sS'm'
(lp39841
I1
aI3
aI5
aI15
aI11
aI3
aI77
aI161
aI309
aI923
aI513
aI161
aI6447
aI9811
aI9209
aI21413
aI8611
aI237557
asS'd'
I19921
sa(dp39842
S'a'
I109334
sS's'
I18
sS'm'
(lp39843
I1
aI3
aI1
aI3
aI23
aI11
aI61
aI107
aI317
aI771
aI1469
aI3367
aI6607
aI11495
aI12271
aI59989
aI52483
aI194761
asS'd'
I19922
sa(dp39844
S'a'
I109359
sS's'
I18
sS'm'
(lp39845
I1
aI3
aI7
aI7
aI23
aI1
aI37
aI47
aI185
aI863
aI1153
aI3517
aI6165
aI3921
aI19311
aI37339
aI112477
aI204915
asS'd'
I19923
sa(dp39846
S'a'
I109373
sS's'
I18
sS'm'
(lp39847
I1
aI3
aI7
aI7
aI15
aI61
aI9
aI113
aI175
aI305
aI1759
aI2933
aI2139
aI3591
aI15303
aI54479
aI126511
aI255205
asS'd'
I19924
sa(dp39848
S'a'
I109386
sS's'
I18
sS'm'
(lp39849
I1
aI1
aI1
aI3
aI29
aI43
aI19
aI183
aI121
aI577
aI1329
aI1737
aI4373
aI5577
aI24023
aI40103
aI22333
aI123423
asS'd'
I19925
sa(dp39850
S'a'
I109394
sS's'
I18
sS'm'
(lp39851
I1
aI3
aI5
aI1
aI29
aI1
aI5
aI177
aI271
aI431
aI139
aI705
aI4319
aI9301
aI15887
aI13253
aI13275
aI233317
asS'd'
I19926
sa(dp39852
S'a'
I109403
sS's'
I18
sS'm'
(lp39853
I1
aI1
aI3
aI7
aI21
aI3
aI13
aI51
aI459
aI359
aI1721
aI193
aI4887
aI7805
aI20615
aI28813
aI82427
aI57853
asS'd'
I19927
sa(dp39854
S'a'
I109409
sS's'
I18
sS'm'
(lp39855
I1
aI3
aI3
aI13
aI3
aI29
aI43
aI95
aI339
aI993
aI979
aI2323
aI7505
aI10203
aI9475
aI16513
aI21651
aI104871
asS'd'
I19928
sa(dp39856
S'a'
I109412
sS's'
I18
sS'm'
(lp39857
I1
aI1
aI3
aI15
aI11
aI5
aI97
aI127
aI397
aI25
aI507
aI1839
aI2313
aI13399
aI899
aI25713
aI94363
aI178287
asS'd'
I19929
sa(dp39858
S'a'
I109427
sS's'
I18
sS'm'
(lp39859
I1
aI1
aI5
aI5
aI9
aI53
aI95
aI171
aI107
aI129
aI1041
aI3583
aI5479
aI943
aI21435
aI36481
aI85519
aI169651
asS'd'
I19930
sa(dp39860
S'a'
I109439
sS's'
I18
sS'm'
(lp39861
I1
aI1
aI1
aI1
aI1
aI11
aI17
aI249
aI277
aI805
aI1827
aI2705
aI3015
aI9033
aI25019
aI38593
aI59235
aI21145
asS'd'
I19931
sa(dp39862
S'a'
I109464
sS's'
I18
sS'm'
(lp39863
I1
aI1
aI7
aI3
aI7
aI59
aI107
aI213
aI109
aI213
aI555
aI3463
aI953
aI3173
aI18947
aI2863
aI27889
aI32493
asS'd'
I19932
sa(dp39864
S'a'
I109467
sS's'
I18
sS'm'
(lp39865
I1
aI3
aI5
aI15
aI21
aI11
aI99
aI131
aI287
aI513
aI1393
aI3327
aI7347
aI4343
aI8805
aI29571
aI97151
aI97313
asS'd'
I19933
sa(dp39866
S'a'
I109476
sS's'
I18
sS'm'
(lp39867
I1
aI3
aI5
aI3
aI17
aI53
aI23
aI21
aI227
aI291
aI111
aI1951
aI6593
aI16325
aI31725
aI31997
aI116907
aI181027
asS'd'
I19934
sa(dp39868
S'a'
I109488
sS's'
I18
sS'm'
(lp39869
I1
aI1
aI5
aI3
aI15
aI57
aI93
aI153
aI345
aI257
aI169
aI795
aI3907
aI5669
aI25447
aI28775
aI1489
aI216417
asS'd'
I19935
sa(dp39870
S'a'
I109497
sS's'
I18
sS'm'
(lp39871
I1
aI1
aI1
aI5
aI19
aI19
aI113
aI55
aI431
aI685
aI1839
aI711
aI4909
aI10211
aI25765
aI37
aI72657
aI79769
asS'd'
I19936
sa(dp39872
S'a'
I109523
sS's'
I18
sS'm'
(lp39873
I1
aI3
aI5
aI7
aI19
aI17
aI93
aI145
aI99
aI799
aI1615
aI1583
aI7705
aI1069
aI30259
aI37951
aI78965
aI16203
asS'd'
I19937
sa(dp39874
S'a'
I109551
sS's'
I18
sS'm'
(lp39875
I1
aI3
aI7
aI13
aI23
aI61
aI119
aI49
aI413
aI1021
aI415
aI465
aI7395
aI31
aI16415
aI1009
aI9843
aI86531
asS'd'
I19938
sa(dp39876
S'a'
I109556
sS's'
I18
sS'm'
(lp39877
I1
aI1
aI3
aI15
aI17
aI21
aI83
aI69
aI411
aI1
aI269
aI1391
aI295
aI13649
aI21649
aI62453
aI11457
aI215375
asS'd'
I19939
sa(dp39878
S'a'
I109566
sS's'
I18
sS'm'
(lp39879
I1
aI3
aI3
aI7
aI25
aI53
aI95
aI99
aI447
aI323
aI27
aI1595
aI3771
aI16099
aI31523
aI14405
aI66999
aI65733
asS'd'
I19940
sa(dp39880
S'a'
I109598
sS's'
I18
sS'm'
(lp39881
I1
aI1
aI3
aI11
aI25
aI29
aI19
aI95
aI101
aI661
aI537
aI641
aI1455
aI16151
aI29087
aI54009
aI89395
aI2223
asS'd'
I19941
sa(dp39882
S'a'
I109628
sS's'
I18
sS'm'
(lp39883
I1
aI1
aI1
aI11
aI11
aI33
aI91
aI227
aI449
aI661
aI1621
aI1743
aI2859
aI9797
aI32397
aI41767
aI116325
aI6839
asS'd'
I19942
sa(dp39884
S'a'
I109643
sS's'
I18
sS'm'
(lp39885
I1
aI3
aI3
aI7
aI31
aI41
aI15
aI139
aI53
aI789
aI25
aI67
aI1131
aI5987
aI14091
aI37259
aI70313
aI6633
asS'd'
I19943
sa(dp39886
S'a'
I109663
sS's'
I18
sS'm'
(lp39887
I1
aI3
aI7
aI13
aI21
aI29
aI71
aI245
aI497
aI493
aI207
aI3221
aI3695
aI3045
aI1497
aI29235
aI65113
aI82015
asS'd'
I19944
sa(dp39888
S'a'
I109674
sS's'
I18
sS'm'
(lp39889
I1
aI1
aI7
aI5
aI29
aI29
aI87
aI197
aI123
aI323
aI773
aI157
aI2925
aI9235
aI31625
aI58187
aI121457
aI25561
asS'd'
I19945
sa(dp39890
S'a'
I109688
sS's'
I18
sS'm'
(lp39891
I1
aI1
aI1
aI9
aI3
aI15
aI99
aI55
aI133
aI757
aI1405
aI997
aI5201
aI8971
aI6095
aI33309
aI35587
aI254545
asS'd'
I19946
sa(dp39892
S'a'
I109691
sS's'
I18
sS'm'
(lp39893
I1
aI3
aI5
aI9
aI19
aI57
aI85
aI45
aI429
aI823
aI1369
aI933
aI1971
aI13753
aI11351
aI45805
aI16527
aI129907
asS'd'
I19947
sa(dp39894
S'a'
I109704
sS's'
I18
sS'm'
(lp39895
I1
aI3
aI3
aI15
aI23
aI15
aI35
aI89
aI477
aI875
aI1087
aI2837
aI1093
aI617
aI18687
aI53269
aI63447
aI226393
asS'd'
I19948
sa(dp39896
S'a'
I109715
sS's'
I18
sS'm'
(lp39897
I1
aI1
aI7
aI9
aI27
aI37
aI23
aI107
aI265
aI485
aI1975
aI3159
aI4065
aI10089
aI26975
aI45067
aI4241
aI49051
asS'd'
I19949
sa(dp39898
S'a'
I109717
sS's'
I18
sS'm'
(lp39899
I1
aI3
aI3
aI1
aI29
aI53
aI89
aI129
aI149
aI717
aI749
aI1481
aI7829
aI15887
aI23185
aI30667
aI11749
aI188963
asS'd'
I19950
sa(dp39900
S'a'
I109746
sS's'
I18
sS'm'
(lp39901
I1
aI3
aI5
aI9
aI13
aI47
aI119
aI135
aI407
aI99
aI1773
aI2307
aI7885
aI4013
aI25723
aI53519
aI37487
aI205671
asS'd'
I19951
sa(dp39902
S'a'
I109789
sS's'
I18
sS'm'
(lp39903
I1
aI1
aI5
aI1
aI11
aI27
aI121
aI213
aI147
aI669
aI799
aI485
aI4275
aI15909
aI30583
aI45925
aI90365
aI226901
asS'd'
I19952
sa(dp39904
S'a'
I109794
sS's'
I18
sS'm'
(lp39905
I1
aI1
aI3
aI5
aI1
aI19
aI81
aI109
aI217
aI949
aI1637
aI3413
aI5957
aI7293
aI17337
aI63857
aI103815
aI80805
asS'd'
I19953
sa(dp39906
S'a'
I109868
sS's'
I18
sS'm'
(lp39907
I1
aI3
aI7
aI3
aI9
aI43
aI119
aI251
aI345
aI3
aI203
aI829
aI3391
aI2575
aI6859
aI50849
aI22221
aI227581
asS'd'
I19954
sa(dp39908
S'a'
I109886
sS's'
I18
sS'm'
(lp39909
I1
aI1
aI7
aI15
aI13
aI43
aI7
aI23
aI101
aI89
aI1747
aI1231
aI1883
aI13363
aI10981
aI59179
aI59555
aI242021
asS'd'
I19955
sa(dp39910
S'a'
I109888
sS's'
I18
sS'm'
(lp39911
I1
aI1
aI5
aI5
aI23
aI15
aI93
aI183
aI231
aI891
aI1745
aI2665
aI1689
aI8515
aI11717
aI35643
aI113067
aI233757
asS'd'
I19956
sa(dp39912
S'a'
I109903
sS's'
I18
sS'm'
(lp39913
I1
aI3
aI7
aI11
aI29
aI21
aI13
aI59
aI103
aI105
aI483
aI863
aI2711
aI7917
aI29279
aI53931
aI7011
aI60075
asS'd'
I19957
sa(dp39914
S'a'
I109908
sS's'
I18
sS'm'
(lp39915
I1
aI1
aI3
aI15
aI5
aI37
aI101
aI163
aI31
aI575
aI2029
aI1625
aI4545
aI12579
aI15175
aI32667
aI59497
aI63653
asS'd'
I19958
sa(dp39916
S'a'
I109912
sS's'
I18
sS'm'
(lp39917
I1
aI3
aI5
aI7
aI23
aI37
aI61
aI101
aI273
aI49
aI1943
aI3381
aI491
aI4079
aI20341
aI26463
aI122261
aI77965
asS'd'
I19959
sa(dp39918
S'a'
I109927
sS's'
I18
sS'm'
(lp39919
I1
aI1
aI3
aI15
aI9
aI51
aI53
aI89
aI175
aI487
aI1583
aI1797
aI4353
aI1339
aI19613
aI26913
aI78955
aI166523
asS'd'
I19960
sa(dp39920
S'a'
I109936
sS's'
I18
sS'm'
(lp39921
I1
aI3
aI3
aI9
aI27
aI57
aI25
aI207
aI233
aI675
aI661
aI2629
aI6971
aI8723
aI31199
aI47215
aI36931
aI28347
asS'd'
I19961
sa(dp39922
S'a'
I109979
sS's'
I18
sS'm'
(lp39923
I1
aI3
aI1
aI3
aI31
aI15
aI123
aI17
aI211
aI883
aI1861
aI2747
aI8075
aI5373
aI23521
aI46217
aI86629
aI171777
asS'd'
I19962
sa(dp39924
S'a'
I110005
sS's'
I18
sS'm'
(lp39925
I1
aI3
aI5
aI9
aI11
aI43
aI45
aI171
aI465
aI835
aI603
aI2121
aI409
aI1643
aI20327
aI63211
aI129479
aI223113
asS'd'
I19963
sa(dp39926
S'a'
I110010
sS's'
I18
sS'm'
(lp39927
I1
aI1
aI5
aI7
aI13
aI57
aI83
aI233
aI307
aI489
aI941
aI1723
aI6149
aI5557
aI2053
aI17377
aI31597
aI176051
asS'd'
I19964
sa(dp39928
S'a'
I110017
sS's'
I18
sS'm'
(lp39929
I1
aI1
aI7
aI3
aI11
aI11
aI85
aI163
aI169
aI989
aI1289
aI2749
aI7681
aI4679
aI645
aI36589
aI85907
aI194713
asS'd'
I19965
sa(dp39930
S'a'
I110041
sS's'
I18
sS'm'
(lp39931
I1
aI1
aI1
aI7
aI3
aI57
aI39
aI149
aI463
aI947
aI481
aI1163
aI7171
aI8539
aI28991
aI61235
aI74487
aI107051
asS'd'
I19966
sa(dp39932
S'a'
I110042
sS's'
I18
sS'm'
(lp39933
I1
aI1
aI3
aI7
aI9
aI57
aI89
aI101
aI231
aI163
aI1355
aI1393
aI5823
aI7565
aI26285
aI13523
aI130329
aI26099
asS'd'
I19967
sa(dp39934
S'a'
I110047
sS's'
I18
sS'm'
(lp39935
I1
aI1
aI7
aI15
aI13
aI59
aI111
aI35
aI265
aI927
aI125
aI1881
aI5397
aI757
aI23845
aI9677
aI76077
aI163361
asS'd'
I19968
sa(dp39936
S'a'
I110048
sS's'
I18
sS'm'
(lp39937
I1
aI1
aI7
aI9
aI1
aI63
aI61
aI157
aI389
aI143
aI1445
aI881
aI3609
aI9955
aI11159
aI59677
aI45831
aI138345
asS'd'
I19969
sa(dp39938
S'a'
I110053
sS's'
I18
sS'm'
(lp39939
I1
aI3
aI5
aI7
aI5
aI45
aI65
aI23
aI257
aI589
aI905
aI3651
aI743
aI117
aI30307
aI7415
aI103331
aI252889
asS'd'
I19970
sa(dp39940
S'a'
I110054
sS's'
I18
sS'm'
(lp39941
I1
aI1
aI7
aI13
aI3
aI57
aI113
aI91
aI217
aI967
aI481
aI989
aI4795
aI3549
aI23717
aI60793
aI80281
aI19977
asS'd'
I19971
sa(dp39942
S'a'
I110057
sS's'
I18
sS'm'
(lp39943
I1
aI3
aI3
aI1
aI15
aI37
aI113
aI245
aI239
aI575
aI197
aI2803
aI7743
aI13447
aI3601
aI56981
aI76381
aI13321
asS'd'
I19972
sa(dp39944
S'a'
I110091
sS's'
I18
sS'm'
(lp39945
I1
aI3
aI1
aI11
aI27
aI33
aI99
aI151
aI43
aI835
aI1951
aI1957
aI2983
aI6781
aI9319
aI2119
aI40533
aI118325
asS'd'
I19973
sa(dp39946
S'a'
I110093
sS's'
I18
sS'm'
(lp39947
I1
aI3
aI3
aI7
aI27
aI19
aI23
aI243
aI347
aI477
aI1661
aI1891
aI2439
aI2485
aI31743
aI1427
aI20317
aI90803
asS'd'
I19974
sa(dp39948
S'a'
I110112
sS's'
I18
sS'm'
(lp39949
I1
aI3
aI1
aI1
aI3
aI59
aI71
aI129
aI21
aI3
aI449
aI629
aI3657
aI4045
aI8305
aI40461
aI60927
aI38529
asS'd'
I19975
sa(dp39950
S'a'
I110127
sS's'
I18
sS'm'
(lp39951
I1
aI3
aI3
aI7
aI9
aI11
aI85
aI185
aI369
aI451
aI887
aI3709
aI6931
aI111
aI1379
aI8741
aI58777
aI188045
asS'd'
I19976
sa(dp39952
S'a'
I110178
sS's'
I18
sS'm'
(lp39953
I1
aI3
aI7
aI11
aI23
aI13
aI77
aI141
aI99
aI543
aI725
aI2439
aI6825
aI1361
aI2785
aI5345
aI5941
aI234751
asS'd'
I19977
sa(dp39954
S'a'
I110183
sS's'
I18
sS'm'
(lp39955
I1
aI1
aI5
aI5
aI17
aI55
aI69
aI9
aI431
aI585
aI1049
aI165
aI1705
aI14907
aI8655
aI12485
aI22783
aI91195
asS'd'
I19978
sa(dp39956
S'a'
I110184
sS's'
I18
sS'm'
(lp39957
I1
aI1
aI7
aI11
aI29
aI41
aI91
aI137
aI17
aI785
aI1151
aI2033
aI7031
aI15077
aI2241
aI21453
aI117947
aI128891
asS'd'
I19979
sa(dp39958
S'a'
I110195
sS's'
I18
sS'm'
(lp39959
I1
aI1
aI1
aI11
aI17
aI53
aI113
aI195
aI409
aI275
aI1757
aI245
aI6263
aI14785
aI29159
aI43827
aI65027
aI248403
asS'd'
I19980
sa(dp39960
S'a'
I110307
sS's'
I18
sS'm'
(lp39961
I1
aI3
aI3
aI3
aI5
aI15
aI23
aI41
aI261
aI891
aI1021
aI1999
aI4883
aI9233
aI10385
aI21953
aI50711
aI4247
asS'd'
I19981
sa(dp39962
S'a'
I110334
sS's'
I18
sS'm'
(lp39963
I1
aI1
aI1
aI13
aI23
aI47
aI33
aI77
aI317
aI251
aI617
aI2265
aI7549
aI327
aI2317
aI41209
aI41063
aI120863
asS'd'
I19982
sa(dp39964
S'a'
I110336
sS's'
I18
sS'm'
(lp39965
I1
aI1
aI7
aI11
aI15
aI17
aI25
aI105
aI253
aI713
aI1147
aI415
aI5777
aI2215
aI4207
aI33857
aI17001
aI260533
asS'd'
I19983
sa(dp39966
S'a'
I110354
sS's'
I18
sS'm'
(lp39967
I1
aI1
aI1
aI11
aI31
aI45
aI3
aI25
aI329
aI45
aI1563
aI121
aI1413
aI16229
aI32485
aI54477
aI101025
aI64847
asS'd'
I19984
sa(dp39968
S'a'
I110356
sS's'
I18
sS'm'
(lp39969
I1
aI3
aI5
aI1
aI15
aI59
aI113
aI203
aI481
aI545
aI371
aI1357
aI5549
aI3043
aI397
aI61483
aI59779
aI58159
asS'd'
I19985
sa(dp39970
S'a'
I110390
sS's'
I18
sS'm'
(lp39971
I1
aI3
aI3
aI3
aI25
aI39
aI29
aI193
aI379
aI293
aI1173
aI3389
aI4231
aI11519
aI6681
aI28813
aI63609
aI13029
asS'd'
I19986
sa(dp39972
S'a'
I110419
sS's'
I18
sS'm'
(lp39973
I1
aI3
aI1
aI1
aI27
aI57
aI37
aI249
aI441
aI905
aI463
aI3543
aI7203
aI10075
aI5373
aI46103
aI6685
aI146943
asS'd'
I19987
sa(dp39974
S'a'
I110452
sS's'
I18
sS'm'
(lp39975
I1
aI3
aI5
aI7
aI31
aI51
aI21
aI139
aI209
aI219
aI1663
aI837
aI3351
aI6291
aI735
aI8715
aI33751
aI199485
asS'd'
I19988
sa(dp39976
S'a'
I110461
sS's'
I18
sS'm'
(lp39977
I1
aI3
aI5
aI11
aI19
aI57
aI41
aI163
aI287
aI327
aI243
aI2891
aI1095
aI3959
aI5067
aI2867
aI16207
aI213089
asS'd'
I19989
sa(dp39978
S'a'
I110465
sS's'
I18
sS'm'
(lp39979
I1
aI1
aI5
aI11
aI21
aI59
aI97
aI239
aI309
aI371
aI797
aI453
aI2595
aI4277
aI3771
aI5665
aI10075
aI56101
asS'd'
I19990
sa(dp39980
S'a'
I110466
sS's'
I18
sS'm'
(lp39981
I1
aI1
aI3
aI15
aI17
aI29
aI103
aI33
aI449
aI525
aI961
aI3551
aI3611
aI14057
aI15971
aI26981
aI35169
aI130213
asS'd'
I19991
sa(dp39982
S'a'
I110468
sS's'
I18
sS'm'
(lp39983
I1
aI1
aI3
aI1
aI5
aI13
aI39
aI211
aI387
aI797
aI1051
aI3287
aI3737
aI12953
aI3387
aI35107
aI78809
aI162907
asS'd'
I19992
sa(dp39984
S'a'
I110486
sS's'
I18
sS'm'
(lp39985
I1
aI1
aI7
aI9
aI7
aI63
aI127
aI115
aI173
aI981
aI623
aI737
aI7625
aI12245
aI4195
aI61873
aI104301
aI250333
asS'd'
I19993
sa(dp39986
S'a'
I110511
sS's'
I18
sS'm'
(lp39987
I1
aI1
aI1
aI5
aI19
aI5
aI123
aI43
aI209
aI741
aI747
aI1057
aI2683
aI15359
aI24121
aI38413
aI5823
aI62213
asS'd'
I19994
sa(dp39988
S'a'
I110531
sS's'
I18
sS'm'
(lp39989
I1
aI1
aI7
aI1
aI23
aI13
aI63
aI165
aI309
aI323
aI247
aI3827
aI5451
aI4823
aI23925
aI56957
aI69765
aI259923
asS'd'
I19995
sa(dp39990
S'a'
I110561
sS's'
I18
sS'm'
(lp39991
I1
aI1
aI5
aI1
aI5
aI49
aI51
aI65
aI31
aI257
aI1363
aI3031
aI5765
aI3645
aI16383
aI6347
aI30429
aI130557
asS'd'
I19996
sa(dp39992
S'a'
I110573
sS's'
I18
sS'm'
(lp39993
I1
aI3
aI5
aI13
aI9
aI63
aI31
aI3
aI317
aI379
aI1345
aI2161
aI7787
aI2055
aI21507
aI20347
aI97021
aI183377
asS'd'
I19997
sa(dp39994
S'a'
I110593
sS's'
I18
sS'm'
(lp39995
I1
aI3
aI5
aI11
aI15
aI59
aI13
aI139
aI415
aI821
aI639
aI1249
aI6349
aI15861
aI21377
aI22813
aI42839
aI76595
asS'd'
I19998
sa(dp39996
S'a'
I110611
sS's'
I18
sS'm'
(lp39997
I1
aI1
aI7
aI9
aI23
aI11
aI25
aI115
aI393
aI153
aI1269
aI871
aI1323
aI1891
aI11619
aI3103
aI79813
aI39811
asS'd'
I19999
sa(dp39998
S'a'
I110614
sS's'
I18
sS'm'
(lp39999
I1
aI3
aI1
aI15
aI27
aI37
aI63
aI253
aI1
aI855
aI1651
aI3111
aI6693
aI1825
aI22549
aI64189
aI18347
aI253425
asS'd'
I20000
sa(dp40000
S'a'
I110651
sS's'
I18
sS'm'
(lp40001
I1
aI3
aI1
aI11
aI23
aI27
aI119
aI59
aI421
aI343
aI831
aI2679
aI5899
aI12087
aI15953
aI18601
aI109551
aI33895
asS'd'
I20001
sa(dp40002
S'a'
I110661
sS's'
I18
sS'm'
(lp40003
I1
aI3
aI1
aI13
aI29
aI3
aI91
aI227
aI301
aI491
aI1045
aI2105
aI5189
aI13717
aI1095
aI6039
aI16229
aI215687
asS'd'
I20002
sa(dp40004
S'a'
I110665
sS's'
I18
sS'm'
(lp40005
I1
aI1
aI7
aI3
aI11
aI53
aI85
aI25
aI23
aI293
aI841
aI3569
aI5335
aI8949
aI28665
aI15139
aI100807
aI155587
asS'd'
I20003
sa(dp40006
S'a'
I110701
sS's'
I18
sS'm'
(lp40007
I1
aI1
aI1
aI9
aI31
aI1
aI77
aI149
aI181
aI5
aI915
aI1155
aI4717
aI2837
aI17545
aI3235
aI26811
aI124901
asS'd'
I20004
sa(dp40008
S'a'
I110730
sS's'
I18
sS'm'
(lp40009
I1
aI1
aI3
aI5
aI25
aI27
aI47
aI215
aI425
aI195
aI1575
aI3961
aI3521
aI4197
aI9565
aI32523
aI125091
aI165543
asS'd'
I20005
sa(dp40010
S'a'
I110735
sS's'
I18
sS'm'
(lp40011
I1
aI3
aI7
aI3
aI17
aI57
aI7
aI137
aI507
aI303
aI1123
aI1511
aI2465
aI4277
aI19959
aI31951
aI83157
aI62843
asS'd'
I20006
sa(dp40012
S'a'
I110738
sS's'
I18
sS'm'
(lp40013
I1
aI3
aI3
aI1
aI27
aI43
aI79
aI191
aI265
aI167
aI665
aI4017
aI6613
aI1175
aI5427
aI47139
aI91517
aI32071
asS'd'
I20007
sa(dp40014
S'a'
I110759
sS's'
I18
sS'm'
(lp40015
I1
aI1
aI1
aI11
aI29
aI63
aI9
aI39
aI303
aI1021
aI415
aI2157
aI5227
aI13557
aI4931
aI12541
aI74101
aI13189
asS'd'
I20008
sa(dp40016
S'a'
I110763
sS's'
I18
sS'm'
(lp40017
I1
aI1
aI3
aI9
aI15
aI41
aI45
aI163
aI301
aI315
aI1433
aI1449
aI3589
aI15783
aI16069
aI16155
aI10527
aI69335
asS'd'
I20009
sa(dp40018
S'a'
I110768
sS's'
I18
sS'm'
(lp40019
I1
aI3
aI1
aI1
aI1
aI5
aI75
aI169
aI215
aI115
aI939
aI1285
aI43
aI1941
aI27847
aI5245
aI51211
aI244817
asS'd'
I20010
sa(dp40020
S'a'
I110774
sS's'
I18
sS'm'
(lp40021
I1
aI1
aI1
aI15
aI7
aI33
aI31
aI23
aI397
aI537
aI1415
aI329
aI6705
aI15015
aI18883
aI62895
aI21435
aI233075
asS'd'
I20011
sa(dp40022
S'a'
I110783
sS's'
I18
sS'm'
(lp40023
I1
aI3
aI3
aI5
aI11
aI49
aI73
aI147
aI183
aI317
aI11
aI1997
aI1045
aI6015
aI29159
aI55195
aI105711
aI134727
asS'd'
I20012
sa(dp40024
S'a'
I110785
sS's'
I18
sS'm'
(lp40025
I1
aI3
aI3
aI9
aI21
aI41
aI5
aI213
aI421
aI539
aI1269
aI1929
aI3701
aI2165
aI14997
aI21933
aI58167
aI239719
asS'd'
I20013
sa(dp40026
S'a'
I110809
sS's'
I18
sS'm'
(lp40027
I1
aI3
aI1
aI5
aI23
aI63
aI37
aI27
aI21
aI547
aI1499
aI1621
aI141
aI5309
aI32257
aI47241
aI123617
aI243853
asS'd'
I20014
sa(dp40028
S'a'
I110812
sS's'
I18
sS'm'
(lp40029
I1
aI1
aI1
aI13
aI31
aI45
aI47
aI91
aI165
aI1007
aI1295
aI4035
aI1461
aI10423
aI7747
aI7329
aI114599
aI169375
asS'd'
I20015
sa(dp40030
S'a'
I110840
sS's'
I18
sS'm'
(lp40031
I1
aI1
aI5
aI3
aI21
aI9
aI75
aI61
aI35
aI745
aI31
aI4085
aI3645
aI16239
aI14979
aI15725
aI108859
aI56745
asS'd'
I20016
sa(dp40032
S'a'
I110853
sS's'
I18
sS'm'
(lp40033
I1
aI1
aI5
aI15
aI11
aI31
aI13
aI229
aI417
aI147
aI1993
aI4043
aI7757
aI13507
aI15297
aI56119
aI2223
aI142275
asS'd'
I20017
sa(dp40034
S'a'
I110905
sS's'
I18
sS'm'
(lp40035
I1
aI3
aI5
aI13
aI3
aI57
aI45
aI109
aI135
aI829
aI159
aI769
aI865
aI2583
aI15755
aI44343
aI84561
aI98621
asS'd'
I20018
sa(dp40036
S'a'
I110933
sS's'
I18
sS'm'
(lp40037
I1
aI1
aI7
aI11
aI17
aI11
aI115
aI45
aI371
aI411
aI863
aI2139
aI3897
aI13981
aI16771
aI4587
aI25195
aI66077
asS'd'
I20019
sa(dp40038
S'a'
I110961
sS's'
I18
sS'm'
(lp40039
I1
aI3
aI7
aI15
aI17
aI47
aI51
aI133
aI133
aI475
aI1363
aI3325
aI4287
aI4837
aI22077
aI60225
aI3097
aI246805
asS'd'
I20020
sa(dp40040
S'a'
I110962
sS's'
I18
sS'm'
(lp40041
I1
aI3
aI5
aI9
aI29
aI51
aI127
aI125
aI353
aI519
aI129
aI1409
aI1497
aI3167
aI14163
aI24921
aI81343
aI129835
asS'd'
I20021
sa(dp40042
S'a'
I110974
sS's'
I18
sS'm'
(lp40043
I1
aI3
aI5
aI5
aI19
aI39
aI95
aI109
aI229
aI1015
aI367
aI2373
aI709
aI6169
aI4089
aI13533
aI22399
aI118977
asS'd'
I20022
sa(dp40044
S'a'
I110987
sS's'
I18
sS'm'
(lp40045
I1
aI1
aI5
aI5
aI17
aI21
aI91
aI131
aI309
aI739
aI1373
aI3723
aI6659
aI1119
aI27521
aI55589
aI34967
aI70831
asS'd'
I20023
sa(dp40046
S'a'
I110989
sS's'
I18
sS'm'
(lp40047
I1
aI3
aI1
aI11
aI5
aI7
aI85
aI215
aI425
aI947
aI589
aI375
aI5943
aI13399
aI18307
aI27007
aI18919
aI200617
asS'd'
I20024
sa(dp40048
S'a'
I111007
sS's'
I18
sS'm'
(lp40049
I1
aI1
aI3
aI15
aI11
aI37
aI111
aI179
aI259
aI517
aI1679
aI225
aI3493
aI15025
aI21751
aI40687
aI73001
aI214477
asS'd'
I20025
sa(dp40050
S'a'
I111011
sS's'
I18
sS'm'
(lp40051
I1
aI1
aI7
aI15
aI17
aI41
aI109
aI129
aI427
aI847
aI1965
aI3269
aI5871
aI12331
aI26899
aI49791
aI103471
aI213789
asS'd'
I20026
sa(dp40052
S'a'
I111043
sS's'
I18
sS'm'
(lp40053
I1
aI3
aI3
aI3
aI9
aI41
aI25
aI115
aI95
aI737
aI717
aI1545
aI841
aI14923
aI7409
aI45365
aI100139
aI77787
asS'd'
I20027
sa(dp40054
S'a'
I111070
sS's'
I18
sS'm'
(lp40055
I1
aI1
aI3
aI15
aI11
aI3
aI63
aI23
aI425
aI433
aI537
aI1599
aI2691
aI11271
aI1695
aI40579
aI53507
aI73033
asS'd'
I20028
sa(dp40056
S'a'
I111088
sS's'
I18
sS'm'
(lp40057
I1
aI3
aI7
aI13
aI9
aI21
aI85
aI31
aI337
aI583
aI1883
aI3877
aI7197
aI7715
aI21525
aI53273
aI11263
aI41907
asS'd'
I20029
sa(dp40058
S'a'
I111127
sS's'
I18
sS'm'
(lp40059
I1
aI3
aI7
aI1
aI7
aI53
aI61
aI45
aI299
aI885
aI1391
aI3109
aI6851
aI6079
aI18857
aI44537
aI95713
aI146125
asS'd'
I20030
sa(dp40060
S'a'
I111140
sS's'
I18
sS'm'
(lp40061
I1
aI1
aI1
aI3
aI21
aI49
aI95
aI105
aI419
aI315
aI365
aI3035
aI4169
aI5723
aI26921
aI62809
aI27019
aI243965
asS'd'
I20031
sa(dp40062
S'a'
I111149
sS's'
I18
sS'm'
(lp40063
I1
aI1
aI3
aI5
aI11
aI39
aI71
aI89
aI249
aI11
aI1395
aI105
aI6637
aI4577
aI22979
aI32405
aI93163
aI58785
asS'd'
I20032
sa(dp40064
S'a'
I111229
sS's'
I18
sS'm'
(lp40065
I1
aI1
aI1
aI1
aI27
aI25
aI125
aI85
aI495
aI697
aI1793
aI301
aI7665
aI12671
aI25359
aI38803
aI58723
aI189837
asS'd'
I20033
sa(dp40066
S'a'
I111230
sS's'
I18
sS'm'
(lp40067
I1
aI3
aI5
aI11
aI17
aI11
aI61
aI211
aI19
aI901
aI1701
aI223
aI2195
aI15571
aI3529
aI34699
aI94607
aI196519
asS'd'
I20034
sa(dp40068
S'a'
I111243
sS's'
I18
sS'm'
(lp40069
I1
aI3
aI7
aI11
aI15
aI1
aI111
aI1
aI381
aI145
aI293
aI3639
aI6931
aI13195
aI19985
aI58491
aI53067
aI184411
asS'd'
I20035
sa(dp40070
S'a'
I111267
sS's'
I18
sS'm'
(lp40071
I1
aI3
aI1
aI15
aI31
aI39
aI39
aI101
aI17
aI431
aI1151
aI2465
aI727
aI12709
aI5377
aI5857
aI49707
aI76439
asS'd'
I20036
sa(dp40072
S'a'
I111287
sS's'
I18
sS'm'
(lp40073
I1
aI3
aI3
aI5
aI7
aI1
aI3
aI39
aI357
aI339
aI415
aI567
aI7245
aI13943
aI7495
aI54133
aI119705
aI160615
asS'd'
I20037
sa(dp40074
S'a'
I111313
sS's'
I18
sS'm'
(lp40075
I1
aI3
aI5
aI15
aI9
aI17
aI13
aI253
aI337
aI673
aI1745
aI2613
aI4635
aI14025
aI11159
aI50001
aI40753
aI172983
asS'd'
I20038
sa(dp40076
S'a'
I111356
sS's'
I18
sS'm'
(lp40077
I1
aI3
aI7
aI11
aI3
aI15
aI19
aI107
aI393
aI1015
aI1441
aI181
aI5721
aI9987
aI15557
aI37263
aI90053
aI205685
asS'd'
I20039
sa(dp40078
S'a'
I111368
sS's'
I18
sS'm'
(lp40079
I1
aI1
aI5
aI1
aI9
aI3
aI69
aI123
aI245
aI111
aI283
aI1581
aI259
aI275
aI813
aI12213
aI19639
aI7335
asS'd'
I20040
sa(dp40080
S'a'
I111409
sS's'
I18
sS'm'
(lp40081
I1
aI3
aI7
aI7
aI1
aI55
aI101
aI63
aI259
aI705
aI653
aI3821
aI2081
aI6447
aI25471
aI15523
aI38827
aI68055
asS'd'
I20041
sa(dp40082
S'a'
I111419
sS's'
I18
sS'm'
(lp40083
I1
aI1
aI5
aI7
aI27
aI9
aI97
aI149
aI445
aI341
aI167
aI3695
aI375
aI853
aI8143
aI36027
aI62729
aI197357
asS'd'
I20042
sa(dp40084
S'a'
I111430
sS's'
I18
sS'm'
(lp40085
I1
aI3
aI7
aI1
aI1
aI37
aI97
aI103
aI493
aI319
aI1287
aI3787
aI4079
aI13049
aI14305
aI6665
aI4631
aI96225
asS'd'
I20043
sa(dp40086
S'a'
I111433
sS's'
I18
sS'm'
(lp40087
I1
aI1
aI5
aI7
aI1
aI5
aI127
aI143
aI251
aI725
aI1759
aI2381
aI181
aI15741
aI9395
aI64441
aI44347
aI221697
asS'd'
I20044
sa(dp40088
S'a'
I111442
sS's'
I18
sS'm'
(lp40089
I1
aI3
aI1
aI3
aI25
aI47
aI29
aI167
aI397
aI827
aI1255
aI3271
aI6307
aI13915
aI3131
aI19123
aI88619
aI62847
asS'd'
I20045
sa(dp40090
S'a'
I111467
sS's'
I18
sS'm'
(lp40091
I1
aI3
aI5
aI1
aI29
aI11
aI59
aI203
aI245
aI553
aI617
aI1287
aI205
aI2291
aI3613
aI39933
aI116981
aI43595
asS'd'
I20046
sa(dp40092
S'a'
I111491
sS's'
I18
sS'm'
(lp40093
I1
aI1
aI3
aI9
aI27
aI33
aI35
aI77
aI437
aI1003
aI119
aI253
aI6643
aI113
aI10587
aI41073
aI55371
aI233307
asS'd'
I20047
sa(dp40094
S'a'
I111503
sS's'
I18
sS'm'
(lp40095
I1
aI3
aI7
aI1
aI29
aI63
aI1
aI231
aI373
aI995
aI1063
aI1385
aI273
aI14115
aI6899
aI62991
aI112003
aI80527
asS'd'
I20048
sa(dp40096
S'a'
I111527
sS's'
I18
sS'm'
(lp40097
I1
aI3
aI3
aI1
aI31
aI21
aI55
aI115
aI393
aI685
aI245
aI1587
aI5617
aI267
aI19639
aI15169
aI14073
aI173091
asS'd'
I20049
sa(dp40098
S'a'
I111531
sS's'
I18
sS'm'
(lp40099
I1
aI3
aI1
aI11
aI29
aI45
aI53
aI21
aI433
aI979
aI1067
aI2999
aI6279
aI4739
aI30835
aI61653
aI112893
aI75839
asS'd'
I20050
sa(dp40100
S'a'
I111541
sS's'
I18
sS'm'
(lp40101
I1
aI1
aI5
aI11
aI23
aI25
aI15
aI107
aI325
aI981
aI1057
aI1181
aI4465
aI16291
aI1523
aI64497
aI129437
aI45067
asS'd'
I20051
sa(dp40102
S'a'
I111546
sS's'
I18
sS'm'
(lp40103
I1
aI1
aI1
aI5
aI1
aI21
aI79
aI135
aI419
aI997
aI1967
aI747
aI2097
aI15397
aI4415
aI15807
aI79583
aI259561
asS'd'
I20052
sa(dp40104
S'a'
I111548
sS's'
I18
sS'm'
(lp40105
I1
aI3
aI1
aI7
aI5
aI49
aI105
aI109
aI243
aI371
aI13
aI2297
aI2845
aI12569
aI13165
aI13551
aI75471
aI168579
asS'd'
I20053
sa(dp40106
S'a'
I111577
sS's'
I18
sS'm'
(lp40107
I1
aI1
aI1
aI1
aI9
aI33
aI7
aI233
aI43
aI773
aI1121
aI3763
aI4047
aI15039
aI8165
aI25407
aI82561
aI215045
asS'd'
I20054
sa(dp40108
S'a'
I111625
sS's'
I18
sS'm'
(lp40109
I1
aI3
aI7
aI1
aI17
aI41
aI105
aI129
aI333
aI687
aI1357
aI1197
aI1271
aI3835
aI15823
aI36777
aI94311
aI192321
asS'd'
I20055
sa(dp40110
S'a'
I111636
sS's'
I18
sS'm'
(lp40111
I1
aI1
aI5
aI15
aI13
aI31
aI93
aI249
aI81
aI167
aI1681
aI1587
aI179
aI5755
aI27741
aI29437
aI100407
aI63287
asS'd'
I20056
sa(dp40112
S'a'
I111650
sS's'
I18
sS'm'
(lp40113
I1
aI3
aI3
aI13
aI11
aI39
aI85
aI23
aI37
aI183
aI547
aI1255
aI1167
aI15961
aI23281
aI59989
aI99393
aI34983
asS'd'
I20057
sa(dp40114
S'a'
I111655
sS's'
I18
sS'm'
(lp40115
I1
aI3
aI7
aI5
aI13
aI33
aI17
aI243
aI321
aI845
aI447
aI1997
aI4639
aI11175
aI24651
aI18281
aI82677
aI244543
asS'd'
I20058
sa(dp40116
S'a'
I111662
sS's'
I18
sS'm'
(lp40117
I1
aI3
aI5
aI13
aI3
aI63
aI75
aI35
aI493
aI207
aI1707
aI1401
aI3687
aI11353
aI5461
aI23433
aI71259
aI93483
asS'd'
I20059
sa(dp40118
S'a'
I111673
sS's'
I18
sS'm'
(lp40119
I1
aI1
aI3
aI1
aI19
aI61
aI81
aI133
aI115
aI957
aI669
aI647
aI347
aI8739
aI18451
aI39641
aI118677
aI136601
asS'd'
I20060
sa(dp40120
S'a'
I111699
sS's'
I18
sS'm'
(lp40121
I1
aI1
aI7
aI13
aI3
aI13
aI119
aI187
aI111
aI181
aI1865
aI1613
aI201
aI3633
aI17805
aI46553
aI8899
aI100407
asS'd'
I20061
sa(dp40122
S'a'
I111727
sS's'
I18
sS'm'
(lp40123
I1
aI3
aI3
aI9
aI19
aI33
aI35
aI215
aI235
aI893
aI877
aI3099
aI7597
aI13521
aI22473
aI65435
aI3205
aI44897
asS'd'
I20062
sa(dp40124
S'a'
I111729
sS's'
I18
sS'm'
(lp40125
I1
aI3
aI3
aI11
aI27
aI3
aI101
aI201
aI215
aI373
aI859
aI1435
aI2637
aI6795
aI21157
aI3333
aI27797
aI199427
asS'd'
I20063
sa(dp40126
S'a'
I111741
sS's'
I18
sS'm'
(lp40127
I1
aI1
aI7
aI13
aI31
aI33
aI85
aI205
aI273
aI565
aI2033
aI3451
aI7581
aI16223
aI1383
aI16297
aI1263
aI49065
asS'd'
I20064
sa(dp40128
S'a'
I111757
sS's'
I18
sS'm'
(lp40129
I1
aI1
aI1
aI13
aI11
aI29
aI65
aI71
aI395
aI179
aI1193
aI3859
aI3075
aI10133
aI6463
aI34617
aI20173
aI203471
asS'd'
I20065
sa(dp40130
S'a'
I111781
sS's'
I18
sS'm'
(lp40131
I1
aI1
aI7
aI3
aI7
aI29
aI11
aI115
aI465
aI695
aI1759
aI3137
aI6337
aI977
aI43
aI62501
aI13297
aI59319
asS'd'
I20066
sa(dp40132
S'a'
I111791
sS's'
I18
sS'm'
(lp40133
I1
aI3
aI5
aI9
aI31
aI59
aI11
aI107
aI109
aI797
aI177
aI2891
aI2535
aI4305
aI19255
aI9591
aI84417
aI87381
asS'd'
I20067
sa(dp40134
S'a'
I111826
sS's'
I18
sS'm'
(lp40135
I1
aI3
aI3
aI3
aI1
aI9
aI45
aI219
aI73
aI573
aI1477
aI3699
aI8145
aI835
aI7123
aI37167
aI53411
aI45397
asS'd'
I20068
sa(dp40136
S'a'
I111838
sS's'
I18
sS'm'
(lp40137
I1
aI3
aI5
aI15
aI15
aI41
aI37
aI63
aI233
aI971
aI1497
aI1223
aI3909
aI11721
aI9217
aI41055
aI9779
aI199339
asS'd'
I20069
sa(dp40138
S'a'
I111854
sS's'
I18
sS'm'
(lp40139
I1
aI3
aI3
aI7
aI17
aI61
aI91
aI237
aI313
aI841
aI7
aI3283
aI4049
aI10403
aI22157
aI4889
aI31903
aI188043
asS'd'
I20070
sa(dp40140
S'a'
I111861
sS's'
I18
sS'm'
(lp40141
I1
aI3
aI7
aI1
aI25
aI3
aI59
aI87
aI191
aI725
aI1615
aI4057
aI3037
aI14597
aI17371
aI42221
aI73919
aI58009
asS'd'
I20071
sa(dp40142
S'a'
I111866
sS's'
I18
sS'm'
(lp40143
I1
aI3
aI5
aI11
aI27
aI7
aI45
aI231
aI315
aI727
aI843
aI2191
aI7909
aI53
aI23309
aI55189
aI96193
aI174017
asS'd'
I20072
sa(dp40144
S'a'
I111868
sS's'
I18
sS'm'
(lp40145
I1
aI3
aI5
aI11
aI27
aI51
aI127
aI187
aI209
aI883
aI429
aI137
aI4585
aI15195
aI16527
aI32571
aI30905
aI8137
asS'd'
I20073
sa(dp40146
S'a'
I111897
sS's'
I18
sS'm'
(lp40147
I1
aI1
aI1
aI11
aI15
aI55
aI13
aI161
aI183
aI671
aI659
aI3669
aI4461
aI13691
aI17119
aI26877
aI52041
aI110103
asS'd'
I20074
sa(dp40148
S'a'
I111903
sS's'
I18
sS'm'
(lp40149
I1
aI3
aI5
aI9
aI5
aI45
aI29
aI19
aI415
aI931
aI683
aI2987
aI3839
aI4529
aI3091
aI54457
aI115537
aI102671
asS'd'
I20075
sa(dp40150
S'a'
I111904
sS's'
I18
sS'm'
(lp40151
I1
aI1
aI7
aI13
aI23
aI31
aI61
aI17
aI327
aI951
aI1333
aI713
aI4309
aI1955
aI22797
aI27007
aI106673
aI47177
asS'd'
I20076
sa(dp40152
S'a'
I111959
sS's'
I18
sS'm'
(lp40153
I1
aI3
aI3
aI9
aI31
aI49
aI19
aI115
aI413
aI257
aI1799
aI3641
aI2075
aI15613
aI14293
aI16123
aI45131
aI209389
asS'd'
I20077
sa(dp40154
S'a'
I111975
sS's'
I18
sS'm'
(lp40155
I1
aI1
aI7
aI15
aI1
aI15
aI27
aI63
aI463
aI825
aI1081
aI991
aI2641
aI5999
aI8551
aI41119
aI80251
aI189263
asS'd'
I20078
sa(dp40156
S'a'
I112018
sS's'
I18
sS'm'
(lp40157
I1
aI1
aI5
aI9
aI7
aI55
aI125
aI97
aI245
aI997
aI457
aI1087
aI1297
aI3433
aI14887
aI24117
aI30689
aI184809
asS'd'
I20079
sa(dp40158
S'a'
I112023
sS's'
I18
sS'm'
(lp40159
I1
aI3
aI1
aI9
aI23
aI59
aI81
aI233
aI341
aI943
aI1335
aI2819
aI2625
aI4957
aI14925
aI7917
aI107383
aI204493
asS'd'
I20080
sa(dp40160
S'a'
I112033
sS's'
I18
sS'm'
(lp40161
I1
aI3
aI5
aI5
aI5
aI25
aI79
aI29
aI191
aI541
aI1295
aI269
aI613
aI5201
aI28639
aI52839
aI52865
aI75181
asS'd'
I20081
sa(dp40162
S'a'
I112036
sS's'
I18
sS'm'
(lp40163
I1
aI3
aI3
aI7
aI19
aI45
aI35
aI201
aI391
aI317
aI1323
aI2733
aI3407
aI10273
aI32689
aI52153
aI108751
aI242493
asS'd'
I20082
sa(dp40164
S'a'
I112043
sS's'
I18
sS'm'
(lp40165
I1
aI3
aI1
aI7
aI15
aI27
aI45
aI21
aI383
aI483
aI1857
aI3443
aI2263
aI11471
aI3697
aI63929
aI116399
aI229733
asS'd'
I20083
sa(dp40166
S'a'
I112095
sS's'
I18
sS'm'
(lp40167
I1
aI1
aI3
aI9
aI13
aI27
aI19
aI37
aI181
aI811
aI1833
aI177
aI7689
aI10073
aI20229
aI31397
aI65415
aI68461
asS'd'
I20084
sa(dp40168
S'a'
I112113
sS's'
I18
sS'm'
(lp40169
I1
aI1
aI1
aI15
aI19
aI53
aI65
aI209
aI433
aI783
aI1647
aI4075
aI3155
aI733
aI25603
aI39033
aI43109
aI151257
asS'd'
I20085
sa(dp40170
S'a'
I112136
sS's'
I18
sS'm'
(lp40171
I1
aI3
aI3
aI5
aI9
aI37
aI61
aI75
aI497
aI825
aI1785
aI3709
aI1731
aI889
aI19325
aI57453
aI39095
aI190855
asS'd'
I20086
sa(dp40172
S'a'
I112153
sS's'
I18
sS'm'
(lp40173
I1
aI3
aI3
aI13
aI3
aI53
aI21
aI39
aI483
aI833
aI1191
aI2829
aI1323
aI1877
aI17257
aI36077
aI47997
aI25349
asS'd'
I20087
sa(dp40174
S'a'
I112165
sS's'
I18
sS'm'
(lp40175
I1
aI1
aI7
aI9
aI29
aI25
aI7
aI91
aI87
aI723
aI777
aI1865
aI7763
aI10995
aI15953
aI36111
aI21313
aI214417
asS'd'
I20088
sa(dp40176
S'a'
I112166
sS's'
I18
sS'm'
(lp40177
I1
aI3
aI5
aI3
aI17
aI15
aI85
aI133
aI245
aI317
aI879
aI3237
aI7049
aI6501
aI13359
aI34063
aI124703
aI118289
asS'd'
I20089
sa(dp40178
S'a'
I112177
sS's'
I18
sS'm'
(lp40179
I1
aI1
aI1
aI7
aI11
aI25
aI111
aI143
aI369
aI593
aI237
aI2787
aI1015
aI13059
aI23275
aI38453
aI90809
aI25803
asS'd'
I20090
sa(dp40180
S'a'
I112189
sS's'
I18
sS'm'
(lp40181
I1
aI1
aI7
aI5
aI13
aI21
aI39
aI235
aI381
aI381
aI949
aI773
aI1123
aI9911
aI18115
aI47657
aI47849
aI197633
asS'd'
I20091
sa(dp40182
S'a'
I112226
sS's'
I18
sS'm'
(lp40183
I1
aI1
aI3
aI13
aI15
aI57
aI47
aI203
aI483
aI341
aI137
aI1283
aI2847
aI5051
aI22593
aI60915
aI45123
aI258733
asS'd'
I20092
sa(dp40184
S'a'
I112237
sS's'
I18
sS'm'
(lp40185
I1
aI1
aI3
aI5
aI29
aI7
aI23
aI127
aI493
aI543
aI747
aI3547
aI4433
aI5847
aI28999
aI18079
aI81205
aI231557
asS'd'
I20093
sa(dp40186
S'a'
I112240
sS's'
I18
sS'm'
(lp40187
I1
aI1
aI3
aI15
aI9
aI63
aI17
aI197
aI75
aI387
aI1679
aI2631
aI1033
aI2757
aI18365
aI11957
aI98915
aI24223
asS'd'
I20094
sa(dp40188
S'a'
I112255
sS's'
I18
sS'm'
(lp40189
I1
aI1
aI3
aI9
aI27
aI55
aI67
aI97
aI345
aI995
aI1151
aI1747
aI4889
aI13847
aI13237
aI9035
aI13461
aI10377
asS'd'
I20095
sa(dp40190
S'a'
I112265
sS's'
I18
sS'm'
(lp40191
I1
aI1
aI5
aI9
aI27
aI13
aI7
aI77
aI399
aI191
aI137
aI2801
aI6379
aI15969
aI1727
aI27503
aI97385
aI147625
asS'd'
I20096
sa(dp40192
S'a'
I112280
sS's'
I18
sS'm'
(lp40193
I1
aI3
aI1
aI7
aI5
aI9
aI103
aI163
aI489
aI615
aI1359
aI2635
aI3115
aI13795
aI18853
aI65225
aI26545
aI212065
asS'd'
I20097
sa(dp40194
S'a'
I112292
sS's'
I18
sS'm'
(lp40195
I1
aI3
aI7
aI3
aI13
aI25
aI125
aI133
aI359
aI423
aI751
aI4045
aI1209
aI7521
aI6653
aI39171
aI125083
aI229399
asS'd'
I20098
sa(dp40196
S'a'
I112307
sS's'
I18
sS'm'
(lp40197
I1
aI3
aI1
aI9
aI15
aI21
aI121
aI223
aI283
aI313
aI1807
aI3829
aI5279
aI10609
aI20113
aI7851
aI23731
aI245327
asS'd'
I20099
sa(dp40198
S'a'
I112316
sS's'
I18
sS'm'
(lp40199
I1
aI1
aI1
aI11
aI15
aI13
aI63
aI253
aI311
aI369
aI1549
aI2803
aI2029
aI14967
aI14217
aI1387
aI104669
aI63375
asS'd'
I20100
sa(dp40200
S'a'
I112342
sS's'
I18
sS'm'
(lp40201
I1
aI1
aI3
aI3
aI31
aI49
aI59
aI189
aI249
aI779
aI275
aI3761
aI3465
aI2205
aI11543
aI16973
aI126249
aI104769
asS'd'
I20101
sa(dp40202
S'a'
I112345
sS's'
I18
sS'm'
(lp40203
I1
aI3
aI1
aI5
aI11
aI39
aI59
aI33
aI121
aI151
aI467
aI1011
aI1379
aI13955
aI20117
aI52537
aI51049
aI50663
asS'd'
I20102
sa(dp40204
S'a'
I112351
sS's'
I18
sS'm'
(lp40205
I1
aI1
aI1
aI7
aI5
aI41
aI121
aI29
aI357
aI33
aI849
aI2441
aI2127
aI1337
aI21147
aI63869
aI80215
aI31211
asS'd'
I20103
sa(dp40206
S'a'
I112361
sS's'
I18
sS'm'
(lp40207
I1
aI3
aI3
aI3
aI25
aI41
aI17
aI101
aI173
aI915
aI463
aI2391
aI1671
aI8789
aI13357
aI42127
aI17599
aI61087
asS'd'
I20104
sa(dp40208
S'a'
I112364
sS's'
I18
sS'm'
(lp40209
I1
aI1
aI3
aI9
aI29
aI23
aI47
aI211
aI435
aI223
aI1421
aI2463
aI4543
aI8569
aI30209
aI46621
aI14367
aI13263
asS'd'
I20105
sa(dp40210
S'a'
I112372
sS's'
I18
sS'm'
(lp40211
I1
aI3
aI7
aI7
aI5
aI9
aI75
aI209
aI299
aI81
aI1705
aI2335
aI6703
aI6309
aI5859
aI57889
aI43219
aI7667
asS'd'
I20106
sa(dp40212
S'a'
I112382
sS's'
I18
sS'm'
(lp40213
I1
aI3
aI3
aI3
aI19
aI31
aI107
aI87
aI413
aI111
aI215
aI2711
aI7053
aI5223
aI25241
aI26675
aI16067
aI122719
asS'd'
I20107
sa(dp40214
S'a'
I112389
sS's'
I18
sS'm'
(lp40215
I1
aI1
aI1
aI15
aI21
aI3
aI15
aI13
aI281
aI63
aI725
aI2025
aI4813
aI14177
aI18577
aI875
aI118623
aI192005
asS'd'
I20108
sa(dp40216
S'a'
I112408
sS's'
I18
sS'm'
(lp40217
I1
aI3
aI5
aI9
aI17
aI21
aI85
aI173
aI59
aI153
aI763
aI3899
aI1985
aI2071
aI10439
aI44911
aI60915
aI122419
asS'd'
I20109
sa(dp40218
S'a'
I112417
sS's'
I18
sS'm'
(lp40219
I1
aI3
aI7
aI13
aI11
aI63
aI59
aI95
aI53
aI927
aI555
aI1897
aI5195
aI13469
aI16973
aI3463
aI125173
aI256021
asS'd'
I20110
sa(dp40220
S'a'
I112423
sS's'
I18
sS'm'
(lp40221
I1
aI1
aI3
aI7
aI9
aI63
aI33
aI193
aI61
aI445
aI1247
aI1379
aI4701
aI5311
aI30709
aI16795
aI69871
aI161113
asS'd'
I20111
sa(dp40222
S'a'
I112476
sS's'
I18
sS'm'
(lp40223
I1
aI3
aI1
aI3
aI21
aI25
aI125
aI111
aI109
aI75
aI455
aI861
aI6483
aI4501
aI19095
aI45601
aI78415
aI30995
asS'd'
I20112
sa(dp40224
S'a'
I112480
sS's'
I18
sS'm'
(lp40225
I1
aI1
aI3
aI5
aI1
aI25
aI15
aI25
aI223
aI961
aI537
aI1453
aI4951
aI5085
aI19801
aI9863
aI108819
aI7319
asS'd'
I20113
sa(dp40226
S'a'
I112483
sS's'
I18
sS'm'
(lp40227
I1
aI3
aI1
aI5
aI29
aI21
aI79
aI113
aI177
aI691
aI219
aI3159
aI3493
aI25
aI30655
aI46257
aI23707
aI243377
asS'd'
I20114
sa(dp40228
S'a'
I112513
sS's'
I18
sS'm'
(lp40229
I1
aI1
aI3
aI3
aI27
aI21
aI11
aI95
aI43
aI161
aI2029
aI4091
aI6695
aI7179
aI9955
aI45195
aI32017
aI128605
asS'd'
I20115
sa(dp40230
S'a'
I112538
sS's'
I18
sS'm'
(lp40231
I1
aI1
aI7
aI5
aI19
aI37
aI47
aI83
aI169
aI143
aI773
aI2127
aI347
aI1887
aI2861
aI8155
aI21437
aI175641
asS'd'
I20116
sa(dp40232
S'a'
I112543
sS's'
I18
sS'm'
(lp40233
I1
aI3
aI1
aI1
aI27
aI63
aI119
aI57
aI77
aI931
aI629
aI1807
aI4469
aI2315
aI3767
aI19207
aI114581
aI125135
asS'd'
I20117
sa(dp40234
S'a'
I112574
sS's'
I18
sS'm'
(lp40235
I1
aI1
aI5
aI11
aI13
aI51
aI51
aI239
aI333
aI369
aI1035
aI3017
aI103
aI1809
aI14579
aI34425
aI123915
aI258397
asS'd'
I20118
sa(dp40236
S'a'
I112599
sS's'
I18
sS'm'
(lp40237
I1
aI1
aI1
aI3
aI3
aI19
aI63
aI237
aI141
aI929
aI943
aI2597
aI3983
aI1043
aI24269
aI12325
aI39013
aI216689
asS'd'
I20119
sa(dp40238
S'a'
I112605
sS's'
I18
sS'm'
(lp40239
I1
aI1
aI3
aI7
aI9
aI61
aI73
aI31
aI287
aI303
aI1415
aI3453
aI2667
aI8625
aI14347
aI51953
aI9181
aI251937
asS'd'
I20120
sa(dp40240
S'a'
I112667
sS's'
I18
sS'm'
(lp40241
I1
aI1
aI7
aI1
aI15
aI41
aI1
aI197
aI87
aI311
aI1147
aI3799
aI2585
aI14027
aI491
aI54203
aI124861
aI227637
asS'd'
I20121
sa(dp40242
S'a'
I112688
sS's'
I18
sS'm'
(lp40243
I1
aI3
aI7
aI3
aI15
aI35
aI97
aI89
aI65
aI493
aI1897
aI3345
aI3807
aI5911
aI12461
aI21393
aI116975
aI212801
asS'd'
I20122
sa(dp40244
S'a'
I112735
sS's'
I18
sS'm'
(lp40245
I1
aI1
aI7
aI11
aI29
aI47
aI61
aI171
aI399
aI929
aI93
aI2815
aI4933
aI9209
aI15053
aI21911
aI117217
aI52539
asS'd'
I20123
sa(dp40246
S'a'
I112746
sS's'
I18
sS'm'
(lp40247
I1
aI3
aI1
aI1
aI19
aI25
aI11
aI41
aI73
aI317
aI215
aI923
aI5153
aI8025
aI18703
aI11513
aI107981
aI2027
asS'd'
I20124
sa(dp40248
S'a'
I112810
sS's'
I18
sS'm'
(lp40249
I1
aI1
aI5
aI7
aI27
aI33
aI47
aI99
aI171
aI259
aI2017
aI2055
aI909
aI4185
aI26689
aI23155
aI109857
aI213957
asS'd'
I20125
sa(dp40250
S'a'
I112817
sS's'
I18
sS'm'
(lp40251
I1
aI3
aI7
aI3
aI31
aI17
aI39
aI203
aI255
aI345
aI1461
aI1561
aI4349
aI6451
aI14763
aI32993
aI74475
aI140557
asS'd'
I20126
sa(dp40252
S'a'
I112827
sS's'
I18
sS'm'
(lp40253
I1
aI1
aI5
aI3
aI21
aI57
aI75
aI201
aI371
aI529
aI1471
aI243
aI3751
aI581
aI18405
aI40933
aI106311
aI745
asS'd'
I20127
sa(dp40254
S'a'
I112835
sS's'
I18
sS'm'
(lp40255
I1
aI1
aI3
aI13
aI7
aI53
aI125
aI15
aI55
aI267
aI1865
aI3297
aI4331
aI2913
aI21675
aI58911
aI28419
aI105585
asS'd'
I20128
sa(dp40256
S'a'
I112892
sS's'
I18
sS'm'
(lp40257
I1
aI3
aI5
aI13
aI7
aI13
aI37
aI37
aI207
aI127
aI785
aI1129
aI8123
aI7655
aI16003
aI18907
aI48883
aI2001
asS'd'
I20129
sa(dp40258
S'a'
I112898
sS's'
I18
sS'm'
(lp40259
I1
aI1
aI5
aI3
aI11
aI3
aI127
aI149
aI503
aI1019
aI887
aI3429
aI7775
aI7113
aI19571
aI34461
aI38889
aI66981
asS'd'
I20130
sa(dp40260
S'a'
I112915
sS's'
I18
sS'm'
(lp40261
I1
aI3
aI7
aI7
aI1
aI55
aI87
aI217
aI465
aI485
aI411
aI2955
aI4899
aI1741
aI7051
aI42885
aI1837
aI68175
asS'd'
I20131
sa(dp40262
S'a'
I112918
sS's'
I18
sS'm'
(lp40263
I1
aI1
aI7
aI1
aI7
aI39
aI25
aI1
aI185
aI523
aI273
aI2409
aI1867
aI3101
aI29823
aI4509
aI81621
aI11815
asS'd'
I20132
sa(dp40264
S'a'
I112937
sS's'
I18
sS'm'
(lp40265
I1
aI1
aI1
aI11
aI13
aI11
aI89
aI237
aI355
aI347
aI91
aI1791
aI5745
aI4181
aI29207
aI39495
aI5275
aI199395
asS'd'
I20133
sa(dp40266
S'a'
I112940
sS's'
I18
sS'm'
(lp40267
I1
aI1
aI7
aI3
aI17
aI37
aI109
aI169
aI191
aI295
aI1001
aI2631
aI1981
aI11821
aI8315
aI40675
aI1293
aI220247
asS'd'
I20134
sa(dp40268
S'a'
I112958
sS's'
I18
sS'm'
(lp40269
I1
aI3
aI1
aI7
aI31
aI25
aI5
aI55
aI1
aI795
aI1663
aI3177
aI6821
aI2073
aI25789
aI23691
aI25015
aI75203
asS'd'
I20135
sa(dp40270
S'a'
I113013
sS's'
I18
sS'm'
(lp40271
I1
aI3
aI5
aI9
aI19
aI9
aI97
aI129
aI351
aI735
aI1897
aI3555
aI1731
aI5413
aI32051
aI12869
aI111973
aI100157
asS'd'
I20136
sa(dp40272
S'a'
I113014
sS's'
I18
sS'm'
(lp40273
I1
aI3
aI3
aI15
aI27
aI1
aI3
aI167
aI7
aI851
aI805
aI713
aI6389
aI1455
aI32371
aI7617
aI107157
aI131299
asS'd'
I20137
sa(dp40274
S'a'
I113027
sS's'
I18
sS'm'
(lp40275
I1
aI3
aI1
aI13
aI31
aI29
aI91
aI123
aI387
aI939
aI223
aI3583
aI2889
aI5307
aI16561
aI6055
aI4437
aI123229
asS'd'
I20138
sa(dp40276
S'a'
I113048
sS's'
I18
sS'm'
(lp40277
I1
aI3
aI5
aI11
aI27
aI17
aI5
aI145
aI369
aI449
aI1677
aI1039
aI3553
aI3057
aI11667
aI51879
aI20519
aI41573
asS'd'
I20139
sa(dp40278
S'a'
I113051
sS's'
I18
sS'm'
(lp40279
I1
aI3
aI1
aI9
aI9
aI1
aI91
aI33
aI379
aI35
aI691
aI375
aI5937
aI15019
aI16177
aI53457
aI52015
aI232257
asS'd'
I20140
sa(dp40280
S'a'
I113053
sS's'
I18
sS'm'
(lp40281
I1
aI1
aI3
aI11
aI23
aI17
aI75
aI217
aI377
aI571
aI1725
aI2719
aI3911
aI12277
aI27799
aI55573
aI21981
aI112529
asS'd'
I20141
sa(dp40282
S'a'
I113102
sS's'
I18
sS'm'
(lp40283
I1
aI3
aI1
aI11
aI9
aI37
aI81
aI95
aI501
aI615
aI327
aI3751
aI7333
aI15407
aI7785
aI29113
aI116335
aI221853
asS'd'
I20142
sa(dp40284
S'a'
I113104
sS's'
I18
sS'm'
(lp40285
I1
aI1
aI1
aI3
aI17
aI1
aI125
aI157
aI461
aI845
aI93
aI107
aI4429
aI2271
aI14445
aI32919
aI81175
aI244557
asS'd'
I20143
sa(dp40286
S'a'
I113114
sS's'
I18
sS'm'
(lp40287
I1
aI3
aI3
aI1
aI27
aI23
aI33
aI15
aI29
aI361
aI409
aI981
aI7819
aI10259
aI21971
aI23317
aI66641
aI54591
asS'd'
I20144
sa(dp40288
S'a'
I113130
sS's'
I18
sS'm'
(lp40289
I1
aI3
aI7
aI13
aI31
aI63
aI11
aI167
aI511
aI81
aI1165
aI3973
aI4275
aI3315
aI10227
aI34973
aI58505
aI2333
asS'd'
I20145
sa(dp40290
S'a'
I113159
sS's'
I18
sS'm'
(lp40291
I1
aI3
aI1
aI9
aI3
aI49
aI111
aI101
aI41
aI775
aI449
aI1349
aI4411
aI8691
aI535
aI60137
aI3269
aI204895
asS'd'
I20146
sa(dp40292
S'a'
I113180
sS's'
I18
sS'm'
(lp40293
I1
aI3
aI7
aI15
aI7
aI43
aI39
aI147
aI309
aI185
aI733
aI1473
aI5467
aI6183
aI17971
aI56805
aI111931
aI163515
asS'd'
I20147
sa(dp40294
S'a'
I113189
sS's'
I18
sS'm'
(lp40295
I1
aI3
aI1
aI3
aI21
aI31
aI17
aI129
aI317
aI587
aI801
aI2517
aI2569
aI765
aI20869
aI16461
aI34425
aI101123
asS'd'
I20148
sa(dp40296
S'a'
I113211
sS's'
I18
sS'm'
(lp40297
I1
aI3
aI1
aI7
aI13
aI63
aI117
aI31
aI25
aI741
aI365
aI687
aI6195
aI2093
aI14679
aI16861
aI123381
aI25263
asS'd'
I20149
sa(dp40298
S'a'
I113245
sS's'
I18
sS'm'
(lp40299
I1
aI1
aI1
aI3
aI13
aI59
aI65
aI131
aI41
aI39
aI1659
aI1491
aI225
aI10277
aI12445
aI4161
aI92119
aI146705
asS'd'
I20150
sa(dp40300
S'a'
I113261
sS's'
I18
sS'm'
(lp40301
I1
aI3
aI5
aI1
aI31
aI11
aI21
aI203
aI345
aI473
aI1643
aI1377
aI555
aI11675
aI15383
aI30855
aI41249
aI231059
asS'd'
I20151
sa(dp40302
S'a'
I113273
sS's'
I18
sS'm'
(lp40303
I1
aI1
aI7
aI15
aI3
aI23
aI33
aI133
aI433
aI407
aI1217
aI3345
aI7455
aI11489
aI21463
aI41621
aI95755
aI86971
asS'd'
I20152
sa(dp40304
S'a'
I113292
sS's'
I18
sS'm'
(lp40305
I1
aI1
aI1
aI3
aI13
aI47
aI45
aI181
aI489
aI89
aI427
aI1915
aI3993
aI10133
aI20437
aI31811
aI48421
aI150009
asS'd'
I20153
sa(dp40306
S'a'
I113314
sS's'
I18
sS'm'
(lp40307
I1
aI3
aI1
aI9
aI9
aI25
aI89
aI195
aI503
aI755
aI59
aI1869
aI6645
aI13841
aI22973
aI17761
aI46759
aI68717
asS'd'
I20154
sa(dp40308
S'a'
I113319
sS's'
I18
sS'm'
(lp40309
I1
aI3
aI1
aI1
aI19
aI21
aI119
aI123
aI481
aI289
aI1009
aI3769
aI3909
aI1123
aI17875
aI17383
aI71533
aI45455
asS'd'
I20155
sa(dp40310
S'a'
I113323
sS's'
I18
sS'm'
(lp40311
I1
aI1
aI1
aI3
aI31
aI33
aI127
aI43
aI467
aI749
aI377
aI3025
aI511
aI13335
aI23987
aI63627
aI50211
aI197253
asS'd'
I20156
sa(dp40312
S'a'
I113326
sS's'
I18
sS'm'
(lp40313
I1
aI1
aI5
aI13
aI29
aI7
aI101
aI43
aI299
aI769
aI1637
aI3731
aI1945
aI9933
aI22263
aI1523
aI127557
aI116867
asS'd'
I20157
sa(dp40314
S'a'
I113337
sS's'
I18
sS'm'
(lp40315
I1
aI1
aI3
aI11
aI1
aI59
aI25
aI45
aI275
aI535
aI1349
aI3625
aI8125
aI727
aI1215
aI15487
aI86229
aI124817
asS'd'
I20158
sa(dp40316
S'a'
I113338
sS's'
I18
sS'm'
(lp40317
I1
aI3
aI3
aI13
aI3
aI11
aI25
aI237
aI213
aI331
aI395
aI1775
aI1225
aI6859
aI16577
aI39105
aI118081
aI74727
asS'd'
I20159
sa(dp40318
S'a'
I113355
sS's'
I18
sS'm'
(lp40319
I1
aI1
aI1
aI9
aI5
aI27
aI117
aI75
aI479
aI757
aI1299
aI2273
aI3221
aI5297
aI249
aI60327
aI48739
aI107023
asS'd'
I20160
sa(dp40320
S'a'
I113365
sS's'
I18
sS'm'
(lp40321
I1
aI1
aI5
aI9
aI27
aI9
aI123
aI49
aI63
aI763
aI121
aI3955
aI2069
aI5999
aI25973
aI64661
aI6321
aI1179
asS'd'
I20161
sa(dp40322
S'a'
I113376
sS's'
I18
sS'm'
(lp40323
I1
aI3
aI7
aI11
aI9
aI51
aI65
aI93
aI51
aI51
aI829
aI3239
aI7431
aI3489
aI7691
aI38777
aI28151
aI96635
asS'd'
I20162
sa(dp40324
S'a'
I113408
sS's'
I18
sS'm'
(lp40325
I1
aI3
aI3
aI13
aI15
aI51
aI13
aI203
aI49
aI73
aI363
aI2173
aI7771
aI11527
aI27683
aI39333
aI2083
aI178623
asS'd'
I20163
sa(dp40326
S'a'
I113462
sS's'
I18
sS'm'
(lp40327
I1
aI1
aI5
aI5
aI15
aI27
aI27
aI127
aI503
aI955
aI427
aI3061
aI6213
aI917
aI889
aI12601
aI72445
aI105383
asS'd'
I20164
sa(dp40328
S'a'
I113476
sS's'
I18
sS'm'
(lp40329
I1
aI3
aI5
aI3
aI27
aI43
aI105
aI187
aI309
aI747
aI1843
aI723
aI539
aI8829
aI19171
aI46009
aI26129
aI173145
asS'd'
I20165
sa(dp40330
S'a'
I113503
sS's'
I18
sS'm'
(lp40331
I1
aI3
aI7
aI7
aI9
aI51
aI121
aI139
aI107
aI453
aI1103
aI2957
aI633
aI1435
aI27275
aI53231
aI51393
aI16847
asS'd'
I20166
sa(dp40332
S'a'
I113550
sS's'
I18
sS'm'
(lp40333
I1
aI3
aI7
aI5
aI25
aI31
aI71
aI191
aI169
aI69
aI1477
aI1413
aI7659
aI11737
aI12365
aI25067
aI21787
aI16225
asS'd'
I20167
sa(dp40334
S'a'
I113578
sS's'
I18
sS'm'
(lp40335
I1
aI1
aI7
aI1
aI9
aI33
aI37
aI123
aI391
aI341
aI829
aI1543
aI7323
aI14695
aI16431
aI20009
aI95821
aI182791
asS'd'
I20168
sa(dp40336
S'a'
I113580
sS's'
I18
sS'm'
(lp40337
I1
aI3
aI1
aI5
aI9
aI59
aI109
aI39
aI301
aI977
aI1963
aI177
aI8107
aI16193
aI5691
aI14157
aI71605
aI250839
asS'd'
I20169
sa(dp40338
S'a'
I113634
sS's'
I18
sS'm'
(lp40339
I1
aI3
aI5
aI9
aI29
aI33
aI33
aI153
aI7
aI217
aI201
aI563
aI6577
aI9605
aI16671
aI63949
aI97937
aI234309
asS'd'
I20170
sa(dp40340
S'a'
I113653
sS's'
I18
sS'm'
(lp40341
I1
aI3
aI7
aI3
aI25
aI11
aI81
aI89
aI275
aI801
aI477
aI1921
aI2279
aI1651
aI13333
aI9127
aI99693
aI83141
asS'd'
I20171
sa(dp40342
S'a'
I113677
sS's'
I18
sS'm'
(lp40343
I1
aI3
aI7
aI5
aI23
aI51
aI23
aI51
aI447
aI689
aI387
aI1845
aI6033
aI2037
aI20139
aI33165
aI56111
aI243353
asS'd'
I20172
sa(dp40344
S'a'
I113713
sS's'
I18
sS'm'
(lp40345
I1
aI3
aI5
aI7
aI5
aI7
aI105
aI121
aI439
aI471
aI721
aI85
aI1627
aI3735
aI29611
aI15537
aI36131
aI30225
asS'd'
I20173
sa(dp40346
S'a'
I113751
sS's'
I18
sS'm'
(lp40347
I1
aI1
aI5
aI5
aI7
aI29
aI31
aI209
aI183
aI217
aI467
aI1287
aI6145
aI14737
aI16249
aI8857
aI101405
aI103355
asS'd'
I20174
sa(dp40348
S'a'
I113771
sS's'
I18
sS'm'
(lp40349
I1
aI3
aI3
aI5
aI19
aI1
aI43
aI15
aI239
aI63
aI617
aI2189
aI3841
aI1223
aI12217
aI4121
aI88047
aI14069
asS'd'
I20175
sa(dp40350
S'a'
I113781
sS's'
I18
sS'm'
(lp40351
I1
aI3
aI7
aI1
aI9
aI49
aI11
aI65
aI297
aI943
aI1739
aI3797
aI6169
aI2057
aI5031
aI2149
aI21439
aI141039
asS'd'
I20176
sa(dp40352
S'a'
I113797
sS's'
I18
sS'm'
(lp40353
I1
aI1
aI1
aI7
aI15
aI59
aI35
aI203
aI347
aI529
aI1741
aI1003
aI6143
aI4979
aI15495
aI48447
aI2139
aI187025
asS'd'
I20177
sa(dp40354
S'a'
I113846
sS's'
I18
sS'm'
(lp40355
I1
aI1
aI7
aI13
aI15
aI17
aI77
aI225
aI461
aI691
aI1067
aI1133
aI6555
aI511
aI25845
aI39835
aI11755
aI142743
asS'd'
I20178
sa(dp40356
S'a'
I113849
sS's'
I18
sS'm'
(lp40357
I1
aI3
aI3
aI11
aI27
aI25
aI49
aI51
aI335
aI1
aI381
aI2703
aI7023
aI14739
aI19335
aI39625
aI82255
aI76277
asS'd'
I20179
sa(dp40358
S'a'
I113855
sS's'
I18
sS'm'
(lp40359
I1
aI3
aI3
aI7
aI19
aI3
aI35
aI95
aI203
aI991
aI515
aI2245
aI6085
aI4129
aI9581
aI38309
aI114203
aI136021
asS'd'
I20180
sa(dp40360
S'a'
I113878
sS's'
I18
sS'm'
(lp40361
I1
aI1
aI7
aI7
aI21
aI61
aI31
aI57
aI459
aI119
aI523
aI1293
aI3647
aI735
aI28849
aI15581
aI123943
aI210069
asS'd'
I20181
sa(dp40362
S'a'
I113884
sS's'
I18
sS'm'
(lp40363
I1
aI1
aI7
aI3
aI9
aI55
aI103
aI23
aI401
aI109
aI23
aI4083
aI6179
aI12817
aI2787
aI43337
aI53647
aI241507
asS'd'
I20182
sa(dp40364
S'a'
I113926
sS's'
I18
sS'm'
(lp40365
I1
aI1
aI5
aI7
aI9
aI51
aI37
aI133
aI97
aI933
aI1509
aI2229
aI1769
aI12901
aI15439
aI25687
aI128823
aI72451
asS'd'
I20183
sa(dp40366
S'a'
I113938
sS's'
I18
sS'm'
(lp40367
I1
aI3
aI1
aI13
aI17
aI19
aI7
aI109
aI299
aI799
aI621
aI3393
aI3645
aI283
aI29889
aI63215
aI97805
aI45795
asS'd'
I20184
sa(dp40368
S'a'
I113956
sS's'
I18
sS'm'
(lp40369
I1
aI3
aI1
aI15
aI21
aI7
aI65
aI237
aI221
aI433
aI1611
aI2591
aI3639
aI3231
aI6025
aI53465
aI88091
aI17657
asS'd'
I20185
sa(dp40370
S'a'
I113980
sS's'
I18
sS'm'
(lp40371
I1
aI1
aI7
aI9
aI27
aI13
aI11
aI185
aI381
aI43
aI961
aI2743
aI2691
aI10531
aI3713
aI61757
aI124011
aI209323
asS'd'
I20186
sa(dp40372
S'a'
I113986
sS's'
I18
sS'm'
(lp40373
I1
aI3
aI5
aI1
aI13
aI7
aI109
aI65
aI359
aI577
aI2001
aI3085
aI3519
aI8577
aI19299
aI40145
aI37159
aI82421
asS'd'
I20187
sa(dp40374
S'a'
I113991
sS's'
I18
sS'm'
(lp40375
I1
aI1
aI3
aI11
aI7
aI5
aI21
aI215
aI391
aI317
aI879
aI1835
aI611
aI7189
aI3887
aI45383
aI41025
aI175701
asS'd'
I20188
sa(dp40376
S'a'
I114005
sS's'
I18
sS'm'
(lp40377
I1
aI1
aI1
aI1
aI5
aI17
aI69
aI115
aI481
aI477
aI2017
aI583
aI8033
aI11349
aI16625
aI213
aI88033
aI31707
asS'd'
I20189
sa(dp40378
S'a'
I114022
sS's'
I18
sS'm'
(lp40379
I1
aI1
aI7
aI15
aI19
aI55
aI121
aI35
aI1
aI71
aI1011
aI3247
aI4133
aI1681
aI29943
aI30149
aI96797
aI177707
asS'd'
I20190
sa(dp40380
S'a'
I114036
sS's'
I18
sS'm'
(lp40381
I1
aI3
aI5
aI13
aI11
aI45
aI83
aI153
aI455
aI223
aI787
aI2025
aI5271
aI229
aI17549
aI5775
aI75311
aI134523
asS'd'
I20191
sa(dp40382
S'a'
I114046
sS's'
I18
sS'm'
(lp40383
I1
aI3
aI5
aI7
aI21
aI43
aI3
aI253
aI395
aI651
aI1111
aI1685
aI539
aI6555
aI25761
aI39477
aI15823
aI261825
asS'd'
I20192
sa(dp40384
S'a'
I114050
sS's'
I18
sS'm'
(lp40385
I1
aI3
aI7
aI15
aI27
aI35
aI43
aI191
aI269
aI247
aI883
aI887
aI1505
aI7433
aI6239
aI5421
aI49583
aI17765
asS'd'
I20193
sa(dp40386
S'a'
I114061
sS's'
I18
sS'm'
(lp40387
I1
aI1
aI5
aI15
aI7
aI19
aI113
aI177
aI63
aI119
aI517
aI3987
aI971
aI12071
aI13107
aI28913
aI85675
aI204921
asS'd'
I20194
sa(dp40388
S'a'
I114067
sS's'
I18
sS'm'
(lp40389
I1
aI3
aI7
aI15
aI31
aI47
aI21
aI129
aI31
aI505
aI661
aI855
aI6135
aI13063
aI27971
aI63801
aI27469
aI75373
asS'd'
I20195
sa(dp40390
S'a'
I114117
sS's'
I18
sS'm'
(lp40391
I1
aI1
aI7
aI5
aI13
aI23
aI111
aI85
aI279
aI969
aI483
aI831
aI483
aI9065
aI10997
aI59031
aI5083
aI150939
asS'd'
I20196
sa(dp40392
S'a'
I114142
sS's'
I18
sS'm'
(lp40393
I1
aI3
aI5
aI7
aI17
aI55
aI11
aI223
aI189
aI209
aI139
aI577
aI5443
aI913
aI19085
aI53113
aI8427
aI11251
asS'd'
I20197
sa(dp40394
S'a'
I114158
sS's'
I18
sS'm'
(lp40395
I1
aI1
aI1
aI7
aI23
aI61
aI95
aI213
aI443
aI803
aI1545
aI3625
aI2195
aI2649
aI10913
aI14339
aI23001
aI16735
asS'd'
I20198
sa(dp40396
S'a'
I114165
sS's'
I18
sS'm'
(lp40397
I1
aI3
aI3
aI3
aI13
aI45
aI15
aI225
aI419
aI445
aI527
aI635
aI2279
aI5097
aI25267
aI199
aI66187
aI156717
asS'd'
I20199
sa(dp40398
S'a'
I114200
sS's'
I18
sS'm'
(lp40399
I1
aI1
aI1
aI7
aI23
aI17
aI113
aI245
aI99
aI159
aI919
aI2961
aI1731
aI6241
aI12749
aI8925
aI44153
aI243249
asS'd'
I20200
sa(dp40400
S'a'
I114219
sS's'
I18
sS'm'
(lp40401
I1
aI3
aI1
aI3
aI29
aI57
aI43
aI245
aI389
aI233
aI135
aI45
aI3771
aI14061
aI10173
aI51939
aI128985
aI81605
asS'd'
I20201
sa(dp40402
S'a'
I114254
sS's'
I18
sS'm'
(lp40403
I1
aI1
aI1
aI15
aI1
aI19
aI25
aI111
aI91
aI193
aI1185
aI3679
aI7155
aI7077
aI13743
aI35631
aI128975
aI196979
asS'd'
I20202
sa(dp40404
S'a'
I114265
sS's'
I18
sS'm'
(lp40405
I1
aI3
aI3
aI13
aI31
aI57
aI25
aI53
aI149
aI331
aI643
aI915
aI1607
aI14429
aI29803
aI23459
aI72915
aI39253
asS'd'
I20203
sa(dp40406
S'a'
I114272
sS's'
I18
sS'm'
(lp40407
I1
aI3
aI3
aI9
aI23
aI45
aI9
aI29
aI383
aI277
aI981
aI1647
aI5217
aI4449
aI26759
aI63849
aI98081
aI37565
asS'd'
I20204
sa(dp40408
S'a'
I114312
sS's'
I18
sS'm'
(lp40409
I1
aI1
aI1
aI15
aI3
aI23
aI9
aI121
aI231
aI27
aI1961
aI2389
aI1689
aI7041
aI8069
aI37973
aI74601
aI15553
asS'd'
I20205
sa(dp40410
S'a'
I114318
sS's'
I18
sS'm'
(lp40411
I1
aI1
aI5
aI15
aI15
aI29
aI11
aI177
aI355
aI47
aI1821
aI393
aI3383
aI10439
aI6357
aI41119
aI60323
aI206253
asS'd'
I20206
sa(dp40412
S'a'
I114348
sS's'
I18
sS'm'
(lp40413
I1
aI1
aI1
aI1
aI21
aI29
aI87
aI149
aI157
aI979
aI1867
aI729
aI1949
aI4409
aI27495
aI6841
aI89033
aI214957
asS'd'
I20207
sa(dp40414
S'a'
I114377
sS's'
I18
sS'm'
(lp40415
I1
aI1
aI3
aI3
aI9
aI7
aI115
aI129
aI141
aI157
aI881
aI109
aI5537
aI303
aI32549
aI1953
aI9903
aI82401
asS'd'
I20208
sa(dp40416
S'a'
I114383
sS's'
I18
sS'm'
(lp40417
I1
aI1
aI5
aI15
aI9
aI19
aI93
aI53
aI319
aI913
aI1341
aI705
aI4639
aI16189
aI11375
aI39155
aI81393
aI115843
asS'd'
I20209
sa(dp40418
S'a'
I114386
sS's'
I18
sS'm'
(lp40419
I1
aI1
aI5
aI7
aI31
aI21
aI3
aI47
aI437
aI799
aI359
aI3291
aI3917
aI12983
aI19283
aI23769
aI34033
aI226041
asS'd'
I20210
sa(dp40420
S'a'
I114431
sS's'
I18
sS'm'
(lp40421
I1
aI3
aI7
aI7
aI27
aI13
aI65
aI31
aI181
aI511
aI1373
aI3871
aI1537
aI6015
aI12103
aI42187
aI121043
aI95715
asS'd'
I20211
sa(dp40422
S'a'
I114448
sS's'
I18
sS'm'
(lp40423
I1
aI1
aI5
aI11
aI1
aI55
aI91
aI11
aI105
aI137
aI1787
aI81
aI5163
aI5793
aI17403
aI59433
aI113439
aI65751
asS'd'
I20212
sa(dp40424
S'a'
I114479
sS's'
I18
sS'm'
(lp40425
I1
aI1
aI3
aI13
aI21
aI57
aI87
aI157
aI379
aI5
aI285
aI3217
aI4557
aI3359
aI28953
aI63397
aI110537
aI230571
asS'd'
I20213
sa(dp40426
S'a'
I114487
sS's'
I18
sS'm'
(lp40427
I1
aI3
aI7
aI7
aI7
aI27
aI25
aI109
aI125
aI337
aI719
aI561
aI5903
aI12913
aI6987
aI17157
aI50655
aI195109
asS'd'
I20214
sa(dp40428
S'a'
I114513
sS's'
I18
sS'm'
(lp40429
I1
aI3
aI3
aI15
aI3
aI11
aI97
aI93
aI441
aI19
aI1435
aI515
aI6129
aI5177
aI28075
aI53495
aI107817
aI78399
asS'd'
I20215
sa(dp40430
S'a'
I114542
sS's'
I18
sS'm'
(lp40431
I1
aI3
aI1
aI9
aI13
aI7
aI89
aI171
aI165
aI479
aI223
aI4001
aI691
aI4033
aI13577
aI33363
aI63447
aI46609
asS'd'
I20216
sa(dp40432
S'a'
I114572
sS's'
I18
sS'm'
(lp40433
I1
aI3
aI7
aI1
aI15
aI47
aI103
aI45
aI209
aI639
aI1465
aI2795
aI6025
aI7981
aI29491
aI47743
aI12861
aI222445
asS'd'
I20217
sa(dp40434
S'a'
I114584
sS's'
I18
sS'm'
(lp40435
I1
aI3
aI3
aI3
aI1
aI25
aI121
aI91
aI253
aI969
aI1259
aI1409
aI1329
aI15995
aI17733
aI24081
aI101747
aI120619
asS'd'
I20218
sa(dp40436
S'a'
I114600
sS's'
I18
sS'm'
(lp40437
I1
aI3
aI7
aI11
aI11
aI5
aI7
aI241
aI469
aI411
aI1733
aI1385
aI7005
aI10977
aI23369
aI10675
aI90341
aI93077
asS'd'
I20219
sa(dp40438
S'a'
I114605
sS's'
I18
sS'm'
(lp40439
I1
aI3
aI3
aI13
aI17
aI35
aI107
aI189
aI437
aI801
aI1761
aI3133
aI3847
aI14079
aI22465
aI45957
aI38449
aI54273
asS'd'
I20220
sa(dp40440
S'a'
I114623
sS's'
I18
sS'm'
(lp40441
I1
aI1
aI7
aI9
aI9
aI47
aI55
aI107
aI491
aI281
aI777
aI2187
aI6179
aI6607
aI2151
aI9093
aI42873
aI104677
asS'd'
I20221
sa(dp40442
S'a'
I114628
sS's'
I18
sS'm'
(lp40443
I1
aI1
aI5
aI3
aI25
aI3
aI37
aI55
aI339
aI619
aI1227
aI3859
aI5593
aI9639
aI31199
aI48155
aI80779
aI6497
asS'd'
I20222
sa(dp40444
S'a'
I114640
sS's'
I18
sS'm'
(lp40445
I1
aI1
aI7
aI1
aI21
aI49
aI105
aI45
aI119
aI635
aI163
aI3821
aI3689
aI11395
aI19265
aI14289
aI89259
aI167433
asS'd'
I20223
sa(dp40446
S'a'
I114650
sS's'
I18
sS'm'
(lp40447
I1
aI3
aI3
aI15
aI29
aI23
aI11
aI255
aI425
aI443
aI1659
aI3965
aI4791
aI10223
aI11113
aI48751
aI7987
aI166605
asS'd'
I20224
sa(dp40448
S'a'
I114668
sS's'
I18
sS'm'
(lp40449
I1
aI1
aI7
aI3
aI7
aI1
aI113
aI153
aI233
aI803
aI539
aI297
aI4847
aI11203
aI29393
aI54319
aI94373
aI173471
asS'd'
I20225
sa(dp40450
S'a'
I114671
sS's'
I18
sS'm'
(lp40451
I1
aI3
aI3
aI5
aI27
aI57
aI23
aI147
aI423
aI617
aI103
aI3369
aI4825
aI13613
aI23635
aI61977
aI5331
aI115243
asS'd'
I20226
sa(dp40452
S'a'
I114674
sS's'
I18
sS'm'
(lp40453
I1
aI3
aI3
aI9
aI11
aI47
aI41
aI27
aI345
aI657
aI1873
aI365
aI1685
aI11181
aI31977
aI60489
aI98741
aI215357
asS'd'
I20227
sa(dp40454
S'a'
I114700
sS's'
I18
sS'm'
(lp40455
I1
aI3
aI1
aI11
aI19
aI33
aI39
aI223
aI151
aI921
aI309
aI3413
aI6735
aI11971
aI25583
aI6927
aI54821
aI125203
asS'd'
I20228
sa(dp40456
S'a'
I114731
sS's'
I18
sS'm'
(lp40457
I1
aI1
aI5
aI1
aI27
aI31
aI61
aI247
aI207
aI895
aI1453
aI3613
aI7097
aI6537
aI29407
aI9903
aI39937
aI98285
asS'd'
I20229
sa(dp40458
S'a'
I114748
sS's'
I18
sS'm'
(lp40459
I1
aI3
aI1
aI5
aI7
aI11
aI119
aI7
aI323
aI27
aI1069
aI2033
aI7387
aI3381
aI19007
aI49039
aI39453
aI115411
asS'd'
I20230
sa(dp40460
S'a'
I114759
sS's'
I18
sS'm'
(lp40461
I1
aI1
aI7
aI3
aI9
aI15
aI51
aI139
aI353
aI857
aI1829
aI3955
aI7669
aI3961
aI22805
aI39879
aI26677
aI66865
asS'd'
I20231
sa(dp40462
S'a'
I114766
sS's'
I18
sS'm'
(lp40463
I1
aI3
aI5
aI7
aI1
aI11
aI59
aI95
aI181
aI645
aI829
aI3119
aI3607
aI5973
aI12381
aI41577
aI79443
aI226945
asS'd'
I20232
sa(dp40464
S'a'
I114768
sS's'
I18
sS'm'
(lp40465
I1
aI3
aI3
aI5
aI3
aI13
aI91
aI119
aI103
aI889
aI703
aI3005
aI541
aI7529
aI12613
aI14267
aI70445
aI217543
asS'd'
I20233
sa(dp40466
S'a'
I114784
sS's'
I18
sS'm'
(lp40467
I1
aI1
aI5
aI7
aI17
aI41
aI5
aI225
aI85
aI759
aI1071
aI2055
aI1655
aI14811
aI25635
aI50803
aI58545
aI105687
asS'd'
I20234
sa(dp40468
S'a'
I114808
sS's'
I18
sS'm'
(lp40469
I1
aI3
aI5
aI13
aI3
aI7
aI77
aI209
aI139
aI717
aI985
aI1085
aI831
aI11011
aI27313
aI46423
aI29435
aI207359
asS'd'
I20235
sa(dp40470
S'a'
I114813
sS's'
I18
sS'm'
(lp40471
I1
aI3
aI1
aI7
aI27
aI45
aI39
aI75
aI311
aI937
aI1593
aI1357
aI4815
aI1997
aI1045
aI48681
aI49301
aI155607
asS'd'
I20236
sa(dp40472
S'a'
I114829
sS's'
I18
sS'm'
(lp40473
I1
aI3
aI5
aI11
aI21
aI9
aI111
aI39
aI447
aI241
aI1613
aI1799
aI4817
aI1861
aI1263
aI63641
aI92081
aI252051
asS'd'
I20237
sa(dp40474
S'a'
I114830
sS's'
I18
sS'm'
(lp40475
I1
aI1
aI1
aI13
aI31
aI13
aI39
aI29
aI349
aI25
aI1227
aI2457
aI3831
aI7965
aI16903
aI25825
aI62381
aI101765
asS'd'
I20238
sa(dp40476
S'a'
I114842
sS's'
I18
sS'm'
(lp40477
I1
aI1
aI3
aI7
aI15
aI17
aI5
aI29
aI83
aI607
aI931
aI261
aI1087
aI16247
aI10129
aI7813
aI5445
aI167723
asS'd'
I20239
sa(dp40478
S'a'
I114875
sS's'
I18
sS'm'
(lp40479
I1
aI3
aI5
aI9
aI15
aI31
aI69
aI191
aI139
aI467
aI1681
aI1951
aI7813
aI4295
aI18191
aI11411
aI15601
aI13025
asS'd'
I20240
sa(dp40480
S'a'
I114898
sS's'
I18
sS'm'
(lp40481
I1
aI1
aI1
aI11
aI29
aI53
aI97
aI205
aI281
aI917
aI1009
aI913
aI1003
aI16085
aI30339
aI55753
aI53099
aI30697
asS'd'
I20241
sa(dp40482
S'a'
I114903
sS's'
I18
sS'm'
(lp40483
I1
aI1
aI3
aI15
aI25
aI35
aI7
aI227
aI63
aI251
aI845
aI843
aI7117
aI6021
aI26917
aI43611
aI108643
aI215471
asS'd'
I20242
sa(dp40484
S'a'
I114913
sS's'
I18
sS'm'
(lp40485
I1
aI1
aI3
aI11
aI19
aI29
aI75
aI5
aI131
aI37
aI1185
aI2387
aI8161
aI1621
aI19887
aI20525
aI33067
aI30869
asS'd'
I20243
sa(dp40486
S'a'
I114928
sS's'
I18
sS'm'
(lp40487
I1
aI1
aI3
aI3
aI7
aI37
aI75
aI159
aI313
aI17
aI479
aI2477
aI7779
aI309
aI26095
aI35693
aI92561
aI143151
asS'd'
I20244
sa(dp40488
S'a'
I114937
sS's'
I18
sS'm'
(lp40489
I1
aI1
aI5
aI9
aI5
aI29
aI65
aI223
aI331
aI1013
aI37
aI1813
aI1379
aI9277
aI14681
aI61687
aI24763
aI124669
asS'd'
I20245
sa(dp40490
S'a'
I114958
sS's'
I18
sS'm'
(lp40491
I1
aI3
aI1
aI1
aI17
aI47
aI7
aI219
aI11
aI13
aI1517
aI2583
aI7483
aI5399
aI6883
aI51387
aI17901
aI108659
asS'd'
I20246
sa(dp40492
S'a'
I114963
sS's'
I18
sS'm'
(lp40493
I1
aI3
aI7
aI11
aI9
aI63
aI81
aI91
aI411
aI535
aI255
aI3683
aI5285
aI1787
aI27205
aI43651
aI15647
aI230651
asS'd'
I20247
sa(dp40494
S'a'
I115000
sS's'
I18
sS'm'
(lp40495
I1
aI3
aI1
aI11
aI7
aI47
aI35
aI255
aI341
aI379
aI421
aI753
aI7821
aI13271
aI13021
aI463
aI48457
aI132521
asS'd'
I20248
sa(dp40496
S'a'
I115018
sS's'
I18
sS'm'
(lp40497
I1
aI1
aI5
aI7
aI21
aI23
aI53
aI229
aI393
aI509
aI1641
aI2245
aI6941
aI10447
aI3231
aI5451
aI18883
aI47401
asS'd'
I20249
sa(dp40498
S'a'
I115023
sS's'
I18
sS'm'
(lp40499
I1
aI3
aI1
aI7
aI13
aI61
aI71
aI49
aI147
aI625
aI299
aI3843
aI4851
aI3483
aI27005
aI23871
aI18855
aI124893
asS'd'
I20250
sa(dp40500
S'a'
I115028
sS's'
I18
sS'm'
(lp40501
I1
aI3
aI1
aI7
aI31
aI13
aI127
aI177
aI259
aI179
aI531
aI1775
aI5481
aI13157
aI23821
aI31773
aI93941
aI237697
asS'd'
I20251
sa(dp40502
S'a'
I115042
sS's'
I18
sS'm'
(lp40503
I1
aI1
aI7
aI1
aI23
aI21
aI111
aI219
aI401
aI455
aI1603
aI2077
aI1537
aI2063
aI17821
aI52087
aI20707
aI29535
asS'd'
I20252
sa(dp40504
S'a'
I115084
sS's'
I18
sS'm'
(lp40505
I1
aI1
aI3
aI11
aI17
aI17
aI13
aI79
aI49
aI353
aI1691
aI361
aI2805
aI7121
aI27013
aI50631
aI108235
aI70513
asS'd'
I20253
sa(dp40506
S'a'
I115096
sS's'
I18
sS'm'
(lp40507
I1
aI1
aI5
aI3
aI15
aI25
aI103
aI73
aI377
aI253
aI1303
aI501
aI555
aI15789
aI16647
aI9019
aI60581
aI157337
asS'd'
I20254
sa(dp40508
S'a'
I115105
sS's'
I18
sS'm'
(lp40509
I1
aI3
aI5
aI9
aI23
aI45
aI3
aI251
aI25
aI559
aI429
aI1091
aI5657
aI15387
aI5113
aI64533
aI131049
aI127587
asS'd'
I20255
sa(dp40510
S'a'
I115117
sS's'
I18
sS'm'
(lp40511
I1
aI1
aI3
aI15
aI1
aI53
aI71
aI141
aI413
aI849
aI737
aI3045
aI7119
aI8049
aI18295
aI31447
aI70735
aI117457
asS'd'
I20256
sa(dp40512
S'a'
I115149
sS's'
I18
sS'm'
(lp40513
I1
aI1
aI1
aI11
aI17
aI11
aI69
aI155
aI211
aI249
aI1869
aI1575
aI6859
aI7045
aI7015
aI20135
aI84157
aI232621
asS'd'
I20257
sa(dp40514
S'a'
I115155
sS's'
I18
sS'm'
(lp40515
I1
aI3
aI7
aI5
aI19
aI55
aI15
aI163
aI457
aI371
aI1665
aI1935
aI601
aI3629
aI21975
aI1191
aI45133
aI111649
asS'd'
I20258
sa(dp40516
S'a'
I115198
sS's'
I18
sS'm'
(lp40517
I1
aI3
aI7
aI11
aI23
aI33
aI5
aI253
aI355
aI379
aI933
aI1781
aI3989
aI6191
aI19081
aI7651
aI74671
aI258799
asS'd'
I20259
sa(dp40518
S'a'
I115221
sS's'
I18
sS'm'
(lp40519
I1
aI1
aI3
aI3
aI23
aI3
aI63
aI123
aI273
aI861
aI369
aI2409
aI1505
aI9059
aI10727
aI189
aI122911
aI44037
asS'd'
I20260
sa(dp40520
S'a'
I115222
sS's'
I18
sS'm'
(lp40521
I1
aI1
aI7
aI13
aI13
aI23
aI19
aI87
aI191
aI397
aI2027
aI1689
aI1143
aI10919
aI27073
aI15013
aI118429
aI119165
asS'd'
I20261
sa(dp40522
S'a'
I115225
sS's'
I18
sS'm'
(lp40523
I1
aI1
aI5
aI9
aI15
aI13
aI29
aI81
aI409
aI955
aI1827
aI1341
aI3473
aI16005
aI29041
aI57527
aI7329
aI167093
asS'd'
I20262
sa(dp40524
S'a'
I115276
sS's'
I18
sS'm'
(lp40525
I1
aI1
aI5
aI3
aI11
aI31
aI47
aI13
aI171
aI995
aI961
aI3885
aI3259
aI2745
aI12405
aI49281
aI2901
aI207591
asS'd'
I20263
sa(dp40526
S'a'
I115294
sS's'
I18
sS'm'
(lp40527
I1
aI3
aI5
aI13
aI31
aI3
aI1
aI215
aI465
aI279
aI1697
aI2449
aI3829
aI2053
aI9877
aI52911
aI126077
aI210515
asS'd'
I20264
sa(dp40528
S'a'
I115297
sS's'
I18
sS'm'
(lp40529
I1
aI1
aI3
aI7
aI11
aI27
aI55
aI115
aI249
aI353
aI407
aI2567
aI8105
aI7747
aI18111
aI3383
aI101875
aI2185
asS'd'
I20265
sa(dp40530
S'a'
I115321
sS's'
I18
sS'm'
(lp40531
I1
aI1
aI3
aI9
aI25
aI5
aI35
aI137
aI405
aI667
aI1671
aI2965
aI5975
aI4999
aI18421
aI43623
aI64621
aI129797
asS'd'
I20266
sa(dp40532
S'a'
I115348
sS's'
I18
sS'm'
(lp40533
I1
aI3
aI7
aI13
aI3
aI17
aI33
aI191
aI463
aI787
aI1795
aI3037
aI1679
aI63
aI12389
aI3983
aI22385
aI84235
asS'd'
I20267
sa(dp40534
S'a'
I115364
sS's'
I18
sS'm'
(lp40535
I1
aI1
aI5
aI9
aI11
aI25
aI85
aI215
aI355
aI553
aI317
aI1637
aI3461
aI15943
aI2619
aI14545
aI125507
aI18659
asS'd'
I20268
sa(dp40536
S'a'
I115376
sS's'
I18
sS'm'
(lp40537
I1
aI1
aI7
aI5
aI3
aI41
aI105
aI179
aI125
aI557
aI1345
aI3631
aI481
aI10621
aI11213
aI40223
aI46581
aI113137
asS'd'
I20269
sa(dp40538
S'a'
I115385
sS's'
I18
sS'm'
(lp40539
I1
aI3
aI3
aI15
aI1
aI63
aI95
aI213
aI89
aI21
aI1249
aI3063
aI413
aI4307
aI26723
aI10225
aI115143
aI144817
asS'd'
I20270
sa(dp40540
S'a'
I115386
sS's'
I18
sS'm'
(lp40541
I1
aI3
aI5
aI15
aI9
aI43
aI41
aI117
aI419
aI143
aI1651
aI377
aI4775
aI8761
aI23793
aI8719
aI76499
aI208119
asS'd'
I20271
sa(dp40542
S'a'
I115400
sS's'
I18
sS'm'
(lp40543
I1
aI3
aI3
aI1
aI21
aI29
aI47
aI117
aI23
aI333
aI1153
aI1067
aI5859
aI9375
aI29997
aI58991
aI55895
aI204933
asS'd'
I20272
sa(dp40544
S'a'
I115414
sS's'
I18
sS'm'
(lp40545
I1
aI1
aI3
aI11
aI11
aI21
aI115
aI85
aI223
aI281
aI701
aI1331
aI1341
aI1149
aI5993
aI10885
aI77353
aI113553
asS'd'
I20273
sa(dp40546
S'a'
I115465
sS's'
I18
sS'm'
(lp40547
I1
aI1
aI5
aI1
aI25
aI1
aI1
aI153
aI449
aI231
aI593
aI3061
aI4157
aI6661
aI21735
aI11361
aI57751
aI129569
asS'd'
I20274
sa(dp40548
S'a'
I115485
sS's'
I18
sS'm'
(lp40549
I1
aI1
aI3
aI7
aI27
aI63
aI81
aI251
aI125
aI197
aI1525
aI1637
aI4643
aI4743
aI17127
aI51217
aI95781
aI973
asS'd'
I20275
sa(dp40550
S'a'
I115492
sS's'
I18
sS'm'
(lp40551
I1
aI1
aI3
aI7
aI11
aI51
aI13
aI139
aI83
aI341
aI543
aI3061
aI7777
aI6705
aI9609
aI28933
aI24669
aI225275
asS'd'
I20276
sa(dp40552
S'a'
I115501
sS's'
I18
sS'm'
(lp40553
I1
aI3
aI1
aI9
aI25
aI39
aI99
aI139
aI5
aI725
aI1759
aI1577
aI1751
aI3197
aI3169
aI39051
aI1743
aI108813
asS'd'
I20277
sa(dp40554
S'a'
I115519
sS's'
I18
sS'm'
(lp40555
I1
aI1
aI7
aI5
aI31
aI15
aI115
aI229
aI499
aI291
aI501
aI3119
aI2293
aI14137
aI625
aI16379
aI111057
aI101643
asS'd'
I20278
sa(dp40556
S'a'
I115527
sS's'
I18
sS'm'
(lp40557
I1
aI3
aI7
aI15
aI31
aI1
aI51
aI73
aI455
aI51
aI1983
aI3687
aI6049
aI3495
aI26247
aI6567
aI28479
aI158909
asS'd'
I20279
sa(dp40558
S'a'
I115531
sS's'
I18
sS'm'
(lp40559
I1
aI3
aI5
aI5
aI9
aI11
aI77
aI181
aI165
aI773
aI1611
aI3945
aI6787
aI3827
aI28597
aI53269
aI34003
aI237291
asS'd'
I20280
sa(dp40560
S'a'
I115567
sS's'
I18
sS'm'
(lp40561
I1
aI1
aI5
aI3
aI31
aI57
aI15
aI9
aI163
aI363
aI1021
aI2193
aI8175
aI3851
aI26059
aI63915
aI114293
aI163637
asS'd'
I20281
sa(dp40562
S'a'
I115572
sS's'
I18
sS'm'
(lp40563
I1
aI1
aI3
aI7
aI27
aI49
aI35
aI121
aI469
aI833
aI879
aI1601
aI6991
aI13271
aI8085
aI45343
aI5189
aI109413
asS'd'
I20282
sa(dp40564
S'a'
I115631
sS's'
I18
sS'm'
(lp40565
I1
aI3
aI1
aI15
aI7
aI11
aI111
aI153
aI129
aI769
aI565
aI2693
aI333
aI7343
aI28535
aI56937
aI85641
aI19871
asS'd'
I20283
sa(dp40566
S'a'
I115648
sS's'
I18
sS'm'
(lp40567
I1
aI1
aI5
aI13
aI7
aI49
aI121
aI223
aI55
aI33
aI19
aI2291
aI1847
aI10173
aI23337
aI23431
aI18181
aI155663
asS'd'
I20284
sa(dp40568
S'a'
I115660
sS's'
I18
sS'm'
(lp40569
I1
aI3
aI1
aI11
aI25
aI9
aI3
aI255
aI425
aI861
aI1025
aI3719
aI6995
aI14687
aI31083
aI60609
aI115375
aI17813
asS'd'
I20285
sa(dp40570
S'a'
I115672
sS's'
I18
sS'm'
(lp40571
I1
aI1
aI5
aI13
aI1
aI55
aI109
aI239
aI13
aI939
aI1077
aI669
aI1643
aI10949
aI25399
aI55055
aI125829
aI253077
asS'd'
I20286
sa(dp40572
S'a'
I115681
sS's'
I18
sS'm'
(lp40573
I1
aI1
aI5
aI3
aI15
aI51
aI13
aI133
aI257
aI387
aI2017
aI2223
aI1479
aI9377
aI12867
aI9833
aI32323
aI6255
asS'd'
I20287
sa(dp40574
S'a'
I115688
sS's'
I18
sS'm'
(lp40575
I1
aI3
aI1
aI9
aI1
aI53
aI121
aI163
aI349
aI491
aI1867
aI3403
aI6859
aI459
aI1483
aI23893
aI66851
aI150843
asS'd'
I20288
sa(dp40576
S'a'
I115694
sS's'
I18
sS'm'
(lp40577
I1
aI1
aI1
aI1
aI1
aI33
aI51
aI33
aI177
aI633
aI449
aI2705
aI663
aI3701
aI8331
aI43895
aI87223
aI48587
asS'd'
I20289
sa(dp40578
S'a'
I115699
sS's'
I18
sS'm'
(lp40579
I1
aI3
aI5
aI7
aI23
aI7
aI99
aI43
aI217
aI31
aI749
aI2831
aI1557
aI3295
aI6797
aI45229
aI46831
aI62183
asS'd'
I20290
sa(dp40580
S'a'
I115719
sS's'
I18
sS'm'
(lp40581
I1
aI1
aI7
aI7
aI1
aI45
aI35
aI51
aI415
aI693
aI479
aI1017
aI6703
aI241
aI30887
aI8953
aI26901
aI2951
asS'd'
I20291
sa(dp40582
S'a'
I115726
sS's'
I18
sS'm'
(lp40583
I1
aI3
aI3
aI7
aI29
aI3
aI25
aI217
aI67
aI769
aI653
aI3983
aI5513
aI15481
aI21399
aI17525
aI81747
aI109843
asS'd'
I20292
sa(dp40584
S'a'
I115733
sS's'
I18
sS'm'
(lp40585
I1
aI3
aI5
aI5
aI29
aI17
aI97
aI187
aI157
aI189
aI1531
aI1123
aI4291
aI14831
aI15493
aI62753
aI53563
aI153679
asS'd'
I20293
sa(dp40586
S'a'
I115796
sS's'
I18
sS'm'
(lp40587
I1
aI3
aI7
aI13
aI15
aI63
aI47
aI5
aI351
aI275
aI1177
aI3947
aI6755
aI1319
aI17053
aI14267
aI98215
aI228795
asS'd'
I20294
sa(dp40588
S'a'
I115879
sS's'
I18
sS'm'
(lp40589
I1
aI3
aI7
aI5
aI19
aI45
aI43
aI223
aI213
aI903
aI539
aI267
aI83
aI6951
aI2979
aI56929
aI58405
aI198373
asS'd'
I20295
sa(dp40590
S'a'
I115880
sS's'
I18
sS'm'
(lp40591
I1
aI1
aI5
aI11
aI21
aI37
aI109
aI103
aI29
aI49
aI17
aI3987
aI5679
aI2559
aI17391
aI46157
aI38743
aI82245
asS'd'
I20296
sa(dp40592
S'a'
I115888
sS's'
I18
sS'm'
(lp40593
I1
aI1
aI3
aI7
aI7
aI35
aI57
aI187
aI113
aI361
aI721
aI1821
aI6473
aI10233
aI22549
aI37725
aI8445
aI220669
asS'd'
I20297
sa(dp40594
S'a'
I115908
sS's'
I18
sS'm'
(lp40595
I1
aI3
aI3
aI9
aI21
aI41
aI73
aI29
aI163
aI701
aI1277
aI3869
aI1529
aI4889
aI10091
aI65507
aI53829
aI191347
asS'd'
I20298
sa(dp40596
S'a'
I115925
sS's'
I18
sS'm'
(lp40597
I1
aI1
aI5
aI15
aI5
aI21
aI39
aI39
aI341
aI271
aI1543
aI3161
aI3935
aI8319
aI24921
aI19575
aI95009
aI256221
asS'd'
I20299
sa(dp40598
S'a'
I115942
sS's'
I18
sS'm'
(lp40599
I1
aI1
aI1
aI3
aI11
aI33
aI63
aI189
aI21
aI773
aI1261
aI3947
aI183
aI6769
aI31337
aI22179
aI57255
aI8323
asS'd'
I20300
sa(dp40600
S'a'
I115978
sS's'
I18
sS'm'
(lp40601
I1
aI1
aI3
aI15
aI29
aI59
aI103
aI251
aI107
aI499
aI915
aI387
aI3127
aI5597
aI3345
aI15657
aI979
aI91685
asS'd'
I20301
sa(dp40602
S'a'
I115986
sS's'
I18
sS'm'
(lp40603
I1
aI3
aI3
aI11
aI13
aI27
aI9
aI137
aI177
aI75
aI567
aI1511
aI7355
aI3087
aI15309
aI51733
aI87329
aI217125
asS'd'
I20302
sa(dp40604
S'a'
I116014
sS's'
I18
sS'm'
(lp40605
I1
aI1
aI1
aI15
aI9
aI43
aI113
aI177
aI507
aI379
aI765
aI75
aI6895
aI7523
aI24611
aI7315
aI49653
aI59263
asS'd'
I20303
sa(dp40606
S'a'
I116019
sS's'
I18
sS'm'
(lp40607
I1
aI3
aI1
aI5
aI29
aI23
aI59
aI215
aI267
aI161
aI1957
aI341
aI4081
aI9635
aI3345
aI12323
aI128751
aI144577
asS'd'
I20304
sa(dp40608
S'a'
I116031
sS's'
I18
sS'm'
(lp40609
I1
aI3
aI3
aI13
aI17
aI55
aI59
aI73
aI65
aI697
aI1209
aI3345
aI5629
aI4545
aI23043
aI37649
aI55015
aI10263
asS'd'
I20305
sa(dp40610
S'a'
I116048
sS's'
I18
sS'm'
(lp40611
I1
aI1
aI7
aI1
aI21
aI3
aI7
aI19
aI445
aI417
aI1677
aI799
aI1241
aI15463
aI19815
aI52845
aI81309
aI256713
asS'd'
I20306
sa(dp40612
S'a'
I116069
sS's'
I18
sS'm'
(lp40613
I1
aI1
aI3
aI13
aI13
aI57
aI17
aI199
aI3
aI377
aI1799
aI2713
aI3937
aI12511
aI7439
aI33605
aI56697
aI168195
asS'd'
I20307
sa(dp40614
S'a'
I116091
sS's'
I18
sS'm'
(lp40615
I1
aI3
aI1
aI7
aI21
aI53
aI115
aI97
aI389
aI83
aI961
aI813
aI1499
aI3411
aI22377
aI33323
aI118405
aI115947
asS'd'
I20308
sa(dp40616
S'a'
I116103
sS's'
I18
sS'm'
(lp40617
I1
aI3
aI7
aI11
aI23
aI43
aI85
aI249
aI151
aI893
aI833
aI901
aI7731
aI13467
aI14721
aI38613
aI104033
aI136097
asS'd'
I20309
sa(dp40618
S'a'
I116107
sS's'
I18
sS'm'
(lp40619
I1
aI3
aI1
aI11
aI23
aI23
aI119
aI129
aI175
aI159
aI1031
aI2379
aI2753
aI6755
aI10979
aI18225
aI52375
aI257003
asS'd'
I20310
sa(dp40620
S'a'
I116145
sS's'
I18
sS'm'
(lp40621
I1
aI3
aI1
aI13
aI1
aI9
aI61
aI255
aI433
aI621
aI1469
aI705
aI5841
aI7421
aI23873
aI30487
aI55823
aI119705
asS'd'
I20311
sa(dp40622
S'a'
I116152
sS's'
I18
sS'm'
(lp40623
I1
aI3
aI1
aI15
aI19
aI31
aI29
aI163
aI87
aI793
aI885
aI2495
aI4609
aI2757
aI5333
aI52937
aI79187
aI228777
asS'd'
I20312
sa(dp40624
S'a'
I116247
sS's'
I18
sS'm'
(lp40625
I1
aI1
aI1
aI3
aI17
aI43
aI69
aI241
aI143
aI173
aI327
aI2747
aI5617
aI16347
aI16155
aI47775
aI25917
aI163663
asS'd'
I20313
sa(dp40626
S'a'
I116289
sS's'
I18
sS'm'
(lp40627
I1
aI1
aI1
aI1
aI19
aI19
aI15
aI27
aI25
aI139
aI691
aI4019
aI3055
aI10301
aI11281
aI10957
aI59117
aI178149
asS'd'
I20314
sa(dp40628
S'a'
I116316
sS's'
I18
sS'm'
(lp40629
I1
aI1
aI1
aI3
aI15
aI15
aI37
aI89
aI103
aI7
aI527
aI2823
aI7205
aI6831
aI25179
aI22249
aI103323
aI31251
asS'd'
I20315
sa(dp40630
S'a'
I116344
sS's'
I18
sS'm'
(lp40631
I1
aI1
aI3
aI3
aI7
aI49
aI7
aI241
aI37
aI11
aI577
aI1987
aI1935
aI14787
aI16411
aI36305
aI65185
aI221253
asS'd'
I20316
sa(dp40632
S'a'
I116354
sS's'
I18
sS'm'
(lp40633
I1
aI1
aI1
aI5
aI31
aI51
aI123
aI169
aI441
aI13
aI721
aI2359
aI5687
aI2641
aI16339
aI8441
aI55967
aI98775
asS'd'
I20317
sa(dp40634
S'a'
I116368
sS's'
I18
sS'm'
(lp40635
I1
aI1
aI7
aI5
aI21
aI23
aI91
aI229
aI23
aI105
aI339
aI2371
aI7803
aI14913
aI12651
aI40573
aI117399
aI134865
asS'd'
I20318
sa(dp40636
S'a'
I116377
sS's'
I18
sS'm'
(lp40637
I1
aI3
aI1
aI15
aI19
aI27
aI127
aI77
aI469
aI343
aI451
aI2251
aI6705
aI7765
aI8623
aI10367
aI100379
aI140899
asS'd'
I20319
sa(dp40638
S'a'
I116383
sS's'
I18
sS'm'
(lp40639
I1
aI3
aI1
aI5
aI1
aI11
aI93
aI231
aI33
aI133
aI1545
aI1015
aI7577
aI8871
aI29975
aI12141
aI130833
aI103123
asS'd'
I20320
sa(dp40640
S'a'
I116387
sS's'
I18
sS'm'
(lp40641
I1
aI3
aI3
aI5
aI7
aI25
aI95
aI93
aI293
aI543
aI1785
aI2097
aI6045
aI4225
aI607
aI443
aI72055
aI32269
asS'd'
I20321
sa(dp40642
S'a'
I116408
sS's'
I18
sS'm'
(lp40643
I1
aI1
aI1
aI1
aI5
aI55
aI47
aI105
aI189
aI359
aI1589
aI765
aI2303
aI11963
aI25565
aI40669
aI98977
aI242089
asS'd'
I20322
sa(dp40644
S'a'
I116428
sS's'
I18
sS'm'
(lp40645
I1
aI1
aI1
aI15
aI13
aI45
aI121
aI235
aI125
aI181
aI1891
aI3265
aI2097
aI3207
aI31647
aI13407
aI22515
aI15155
asS'd'
I20323
sa(dp40646
S'a'
I116445
sS's'
I18
sS'm'
(lp40647
I1
aI1
aI5
aI15
aI13
aI11
aI81
aI233
aI307
aI505
aI221
aI813
aI6483
aI741
aI9819
aI19405
aI74235
aI144761
asS'd'
I20324
sa(dp40648
S'a'
I116476
sS's'
I18
sS'm'
(lp40649
I1
aI3
aI5
aI7
aI9
aI25
aI31
aI209
aI337
aI473
aI1831
aI2711
aI5551
aI13531
aI28747
aI1875
aI6401
aI159995
asS'd'
I20325
sa(dp40650
S'a'
I116482
sS's'
I18
sS'm'
(lp40651
I1
aI1
aI7
aI7
aI29
aI3
aI127
aI207
aI387
aI849
aI1449
aI2741
aI2105
aI885
aI18115
aI5433
aI122119
aI16969
asS'd'
I20326
sa(dp40652
S'a'
I116488
sS's'
I18
sS'm'
(lp40653
I1
aI3
aI7
aI9
aI25
aI17
aI43
aI209
aI41
aI927
aI409
aI1567
aI1609
aI12487
aI16305
aI41365
aI10991
aI172127
asS'd'
I20327
sa(dp40654
S'a'
I116493
sS's'
I18
sS'm'
(lp40655
I1
aI1
aI3
aI7
aI27
aI29
aI63
aI127
aI81
aI283
aI1459
aI143
aI5993
aI14027
aI8055
aI28065
aI128389
aI255307
asS'd'
I20328
sa(dp40656
S'a'
I116502
sS's'
I18
sS'm'
(lp40657
I1
aI3
aI7
aI11
aI13
aI41
aI63
aI223
aI215
aI901
aI1853
aI2881
aI5149
aI7439
aI4519
aI33279
aI127765
aI139431
asS'd'
I20329
sa(dp40658
S'a'
I116518
sS's'
I18
sS'm'
(lp40659
I1
aI3
aI7
aI7
aI15
aI61
aI61
aI173
aI221
aI711
aI191
aI3863
aI2695
aI9663
aI6277
aI8791
aI128019
aI256755
asS'd'
I20330
sa(dp40660
S'a'
I116524
sS's'
I18
sS'm'
(lp40661
I1
aI3
aI1
aI9
aI29
aI45
aI83
aI43
aI297
aI605
aI1887
aI2421
aI2307
aI5199
aI17275
aI39225
aI127215
aI253687
asS'd'
I20331
sa(dp40662
S'a'
I116527
sS's'
I18
sS'm'
(lp40663
I1
aI1
aI5
aI3
aI21
aI23
aI121
aI125
aI497
aI945
aI1367
aI2757
aI3481
aI8607
aI32447
aI62373
aI32171
aI226621
asS'd'
I20332
sa(dp40664
S'a'
I116549
sS's'
I18
sS'm'
(lp40665
I1
aI3
aI1
aI5
aI7
aI1
aI71
aI255
aI465
aI951
aI129
aI1989
aI6053
aI3737
aI6511
aI54519
aI16947
aI124491
asS'd'
I20333
sa(dp40666
S'a'
I116561
sS's'
I18
sS'm'
(lp40667
I1
aI3
aI5
aI1
aI9
aI21
aI127
aI49
aI85
aI615
aI1897
aI1715
aI7923
aI10309
aI16919
aI24131
aI18015
aI140195
asS'd'
I20334
sa(dp40668
S'a'
I116562
sS's'
I18
sS'm'
(lp40669
I1
aI1
aI1
aI1
aI5
aI27
aI3
aI205
aI29
aI319
aI485
aI3941
aI7829
aI789
aI4207
aI39939
aI67761
aI152459
asS'd'
I20335
sa(dp40670
S'a'
I116568
sS's'
I18
sS'm'
(lp40671
I1
aI3
aI7
aI11
aI9
aI41
aI1
aI129
aI511
aI831
aI1007
aI2011
aI6211
aI9179
aI20877
aI62121
aI21879
aI23661
asS'd'
I20336
sa(dp40672
S'a'
I116577
sS's'
I18
sS'm'
(lp40673
I1
aI1
aI7
aI1
aI19
aI53
aI75
aI123
aI181
aI735
aI925
aI1065
aI3317
aI3201
aI27473
aI19379
aI78223
aI45725
asS'd'
I20337
sa(dp40674
S'a'
I116590
sS's'
I18
sS'm'
(lp40675
I1
aI1
aI5
aI9
aI9
aI61
aI3
aI193
aI441
aI815
aI583
aI3235
aI247
aI14091
aI19877
aI33505
aI3477
aI20111
asS'd'
I20338
sa(dp40676
S'a'
I116602
sS's'
I18
sS'm'
(lp40677
I1
aI1
aI5
aI13
aI29
aI53
aI55
aI165
aI359
aI889
aI1833
aI1543
aI7913
aI307
aI22853
aI37839
aI15569
aI140127
asS'd'
I20339
sa(dp40678
S'a'
I116607
sS's'
I18
sS'm'
(lp40679
I1
aI1
aI1
aI15
aI21
aI53
aI63
aI195
aI299
aI1019
aI1371
aI1311
aI5401
aI8015
aI30335
aI56281
aI61011
aI59279
asS'd'
I20340
sa(dp40680
S'a'
I116611
sS's'
I18
sS'm'
(lp40681
I1
aI1
aI3
aI13
aI3
aI57
aI45
aI239
aI445
aI419
aI581
aI3971
aI4621
aI9327
aI27255
aI53069
aI126415
aI250313
asS'd'
I20341
sa(dp40682
S'a'
I116626
sS's'
I18
sS'm'
(lp40683
I1
aI3
aI1
aI9
aI5
aI63
aI21
aI25
aI447
aI961
aI1857
aI3123
aI3029
aI9743
aI26069
aI38251
aI58475
aI108737
asS'd'
I20342
sa(dp40684
S'a'
I116637
sS's'
I18
sS'm'
(lp40685
I1
aI3
aI1
aI15
aI13
aI59
aI5
aI21
aI171
aI107
aI1631
aI2407
aI6695
aI8079
aI2805
aI50995
aI53173
aI104757
asS'd'
I20343
sa(dp40686
S'a'
I116647
sS's'
I18
sS'm'
(lp40687
I1
aI3
aI7
aI7
aI1
aI55
aI103
aI67
aI369
aI533
aI515
aI2363
aI5147
aI11633
aI20435
aI24591
aI68155
aI140029
asS'd'
I20344
sa(dp40688
S'a'
I116665
sS's'
I18
sS'm'
(lp40689
I1
aI3
aI7
aI13
aI19
aI51
aI13
aI149
aI159
aI915
aI1029
aI2825
aI5259
aI5139
aI31325
aI42825
aI119923
aI227811
asS'd'
I20345
sa(dp40690
S'a'
I116674
sS's'
I18
sS'm'
(lp40691
I1
aI3
aI3
aI3
aI23
aI17
aI121
aI25
aI403
aI333
aI491
aI2869
aI881
aI12997
aI5101
aI48351
aI90831
aI143009
asS'd'
I20346
sa(dp40692
S'a'
I116700
sS's'
I18
sS'm'
(lp40693
I1
aI1
aI3
aI15
aI23
aI63
aI93
aI43
aI107
aI393
aI419
aI3509
aI1543
aI10295
aI11019
aI8389
aI73753
aI42681
asS'd'
I20347
sa(dp40694
S'a'
I116714
sS's'
I18
sS'm'
(lp40695
I1
aI1
aI7
aI1
aI29
aI49
aI41
aI189
aI303
aI955
aI1241
aI1623
aI2269
aI3413
aI6261
aI2155
aI90945
aI95117
asS'd'
I20348
sa(dp40696
S'a'
I116719
sS's'
I18
sS'm'
(lp40697
I1
aI1
aI3
aI15
aI31
aI13
aI103
aI241
aI189
aI283
aI1303
aI1693
aI1587
aI16313
aI205
aI43421
aI121799
aI200151
asS'd'
I20349
sa(dp40698
S'a'
I116744
sS's'
I18
sS'm'
(lp40699
I1
aI3
aI5
aI1
aI29
aI27
aI105
aI83
aI345
aI411
aI1197
aI3489
aI5891
aI1137
aI7311
aI681
aI127991
aI69533
asS'd'
I20350
sa(dp40700
S'a'
I116764
sS's'
I18
sS'm'
(lp40701
I1
aI3
aI5
aI15
aI31
aI11
aI105
aI221
aI57
aI39
aI145
aI3233
aI1431
aI16271
aI21225
aI47989
aI72583
aI191327
asS'd'
I20351
sa(dp40702
S'a'
I116774
sS's'
I18
sS'm'
(lp40703
I1
aI3
aI7
aI9
aI25
aI47
aI109
aI61
aI257
aI949
aI981
aI1383
aI8003
aI4661
aI19555
aI20191
aI114641
aI84817
asS'd'
I20352
sa(dp40704
S'a'
I116778
sS's'
I18
sS'm'
(lp40705
I1
aI1
aI5
aI9
aI17
aI9
aI19
aI209
aI73
aI573
aI1039
aI2741
aI1495
aI1615
aI6299
aI20507
aI84729
aI166977
asS'd'
I20353
sa(dp40706
S'a'
I116798
sS's'
I18
sS'm'
(lp40707
I1
aI3
aI5
aI13
aI27
aI51
aI39
aI203
aI437
aI725
aI1479
aI3071
aI621
aI15563
aI28473
aI58403
aI25943
aI116683
asS'd'
I20354
sa(dp40708
S'a'
I116803
sS's'
I18
sS'm'
(lp40709
I1
aI1
aI3
aI9
aI5
aI29
aI63
aI61
aI329
aI305
aI523
aI2243
aI6689
aI11773
aI19319
aI57783
aI24265
aI218153
asS'd'
I20355
sa(dp40710
S'a'
I116806
sS's'
I18
sS'm'
(lp40711
I1
aI3
aI7
aI5
aI17
aI27
aI115
aI9
aI243
aI613
aI679
aI1915
aI7265
aI2989
aI13663
aI15115
aI50779
aI235761
asS'd'
I20356
sa(dp40712
S'a'
I116827
sS's'
I18
sS'm'
(lp40713
I1
aI1
aI5
aI5
aI13
aI35
aI111
aI151
aI255
aI569
aI1209
aI3277
aI4503
aI3797
aI22601
aI19523
aI126339
aI141289
asS'd'
I20357
sa(dp40714
S'a'
I116839
sS's'
I18
sS'm'
(lp40715
I1
aI1
aI3
aI9
aI15
aI51
aI85
aI125
aI233
aI1011
aI231
aI2949
aI1091
aI8605
aI14855
aI62401
aI14143
aI212557
asS'd'
I20358
sa(dp40716
S'a'
I116863
sS's'
I18
sS'm'
(lp40717
I1
aI3
aI5
aI11
aI29
aI53
aI83
aI31
aI201
aI219
aI1083
aI967
aI6913
aI10325
aI1971
aI55841
aI7733
aI208883
asS'd'
I20359
sa(dp40718
S'a'
I116873
sS's'
I18
sS'm'
(lp40719
I1
aI3
aI3
aI1
aI23
aI33
aI51
aI103
aI265
aI285
aI1363
aI2813
aI3327
aI7921
aI13537
aI31483
aI43405
aI189641
asS'd'
I20360
sa(dp40720
S'a'
I116882
sS's'
I18
sS'm'
(lp40721
I1
aI1
aI7
aI15
aI27
aI3
aI5
aI87
aI117
aI437
aI1251
aI189
aI3271
aI15579
aI25025
aI23203
aI39421
aI133581
asS'd'
I20361
sa(dp40722
S'a'
I116887
sS's'
I18
sS'm'
(lp40723
I1
aI1
aI5
aI1
aI9
aI3
aI91
aI45
aI71
aI557
aI2019
aI2355
aI5539
aI2843
aI13025
aI61017
aI3475
aI179891
asS'd'
I20362
sa(dp40724
S'a'
I116915
sS's'
I18
sS'm'
(lp40725
I1
aI1
aI7
aI5
aI17
aI11
aI127
aI241
aI9
aI971
aI1699
aI2719
aI1947
aI109
aI19817
aI13949
aI120247
aI60775
asS'd'
I20363
sa(dp40726
S'a'
I116939
sS's'
I18
sS'm'
(lp40727
I1
aI1
aI5
aI9
aI9
aI39
aI117
aI221
aI197
aI767
aI1691
aI4075
aI3665
aI1271
aI16119
aI64129
aI2681
aI105325
asS'd'
I20364
sa(dp40728
S'a'
I116944
sS's'
I18
sS'm'
(lp40729
I1
aI3
aI3
aI11
aI31
aI51
aI5
aI23
aI419
aI715
aI1985
aI4095
aI7255
aI10491
aI25575
aI6177
aI35917
aI178345
asS'd'
I20365
sa(dp40730
S'a'
I116953
sS's'
I18
sS'm'
(lp40731
I1
aI3
aI5
aI7
aI15
aI23
aI99
aI203
aI461
aI509
aI1501
aI1965
aI1105
aI1341
aI21713
aI21901
aI129905
aI67937
asS'd'
I20366
sa(dp40732
S'a'
I116965
sS's'
I18
sS'm'
(lp40733
I1
aI3
aI3
aI15
aI25
aI5
aI55
aI167
aI477
aI125
aI163
aI2379
aI2433
aI12975
aI26259
aI55825
aI19913
aI202873
asS'd'
I20367
sa(dp40734
S'a'
I117002
sS's'
I18
sS'm'
(lp40735
I1
aI3
aI3
aI7
aI15
aI15
aI67
aI227
aI413
aI905
aI1609
aI2083
aI4011
aI10477
aI22809
aI61873
aI96423
aI119253
asS'd'
I20368
sa(dp40736
S'a'
I117007
sS's'
I18
sS'm'
(lp40737
I1
aI3
aI1
aI11
aI13
aI17
aI37
aI147
aI355
aI445
aI619
aI3181
aI5939
aI6953
aI15859
aI37979
aI24723
aI133037
asS'd'
I20369
sa(dp40738
S'a'
I117015
sS's'
I18
sS'm'
(lp40739
I1
aI1
aI5
aI15
aI5
aI25
aI89
aI3
aI279
aI569
aI343
aI2453
aI5739
aI2901
aI6709
aI43957
aI75791
aI20791
asS'd'
I20370
sa(dp40740
S'a'
I117032
sS's'
I18
sS'm'
(lp40741
I1
aI1
aI3
aI5
aI13
aI39
aI53
aI203
aI75
aI945
aI635
aI349
aI2339
aI2549
aI23827
aI7903
aI128005
aI14949
asS'd'
I20371
sa(dp40742
S'a'
I117035
sS's'
I18
sS'm'
(lp40743
I1
aI1
aI7
aI3
aI7
aI59
aI59
aI77
aI143
aI99
aI1313
aI3957
aI3807
aI15731
aI20919
aI60829
aI105967
aI226767
asS'd'
I20372
sa(dp40744
S'a'
I117052
sS's'
I18
sS'm'
(lp40745
I1
aI1
aI3
aI7
aI17
aI49
aI27
aI245
aI129
aI583
aI1055
aI741
aI5607
aI689
aI20075
aI54837
aI113257
aI222677
asS'd'
I20373
sa(dp40746
S'a'
I117087
sS's'
I18
sS'm'
(lp40747
I1
aI1
aI7
aI13
aI17
aI5
aI19
aI141
aI205
aI749
aI1769
aI2981
aI5787
aI4511
aI135
aI19475
aI113735
aI116859
asS'd'
I20374
sa(dp40748
S'a'
I117139
sS's'
I18
sS'm'
(lp40749
I1
aI3
aI7
aI1
aI9
aI33
aI111
aI139
aI77
aI117
aI363
aI1171
aI2587
aI1539
aI30791
aI10697
aI6879
aI104827
asS'd'
I20375
sa(dp40750
S'a'
I117157
sS's'
I18
sS'm'
(lp40751
I1
aI3
aI3
aI5
aI27
aI47
aI49
aI215
aI65
aI435
aI1601
aI231
aI2047
aI10405
aI28409
aI17013
aI103909
aI232051
asS'd'
I20376
sa(dp40752
S'a'
I117193
sS's'
I18
sS'm'
(lp40753
I1
aI3
aI3
aI3
aI13
aI19
aI3
aI159
aI293
aI675
aI247
aI2829
aI6703
aI6085
aI1935
aI18209
aI15709
aI186669
asS'd'
I20377
sa(dp40754
S'a'
I117211
sS's'
I18
sS'm'
(lp40755
I1
aI3
aI3
aI5
aI21
aI55
aI17
aI237
aI121
aI603
aI953
aI947
aI6973
aI15979
aI11029
aI12381
aI12807
aI131603
asS'd'
I20378
sa(dp40756
S'a'
I117214
sS's'
I18
sS'm'
(lp40757
I1
aI3
aI5
aI3
aI3
aI41
aI121
aI203
aI283
aI349
aI1841
aI115
aI6567
aI2131
aI883
aI50515
aI78381
aI168189
asS'd'
I20379
sa(dp40758
S'a'
I117220
sS's'
I18
sS'm'
(lp40759
I1
aI3
aI7
aI15
aI5
aI55
aI85
aI13
aI77
aI443
aI1711
aI1043
aI1265
aI3701
aI5121
aI41435
aI40637
aI69125
asS'd'
I20380
sa(dp40760
S'a'
I117238
sS's'
I18
sS'm'
(lp40761
I1
aI1
aI5
aI15
aI15
aI33
aI67
aI235
aI3
aI95
aI1685
aI731
aI2187
aI11857
aI7197
aI62113
aI12565
aI127455
asS'd'
I20381
sa(dp40762
S'a'
I117248
sS's'
I18
sS'm'
(lp40763
I1
aI3
aI7
aI7
aI11
aI45
aI125
aI231
aI263
aI611
aI221
aI195
aI6347
aI14029
aI7823
aI52295
aI78879
aI211441
asS'd'
I20382
sa(dp40764
S'a'
I117275
sS's'
I18
sS'm'
(lp40765
I1
aI3
aI7
aI15
aI9
aI63
aI75
aI189
aI187
aI449
aI27
aI3647
aI4705
aI13037
aI3773
aI36441
aI35445
aI181793
asS'd'
I20383
sa(dp40766
S'a'
I117293
sS's'
I18
sS'm'
(lp40767
I1
aI1
aI5
aI3
aI31
aI19
aI123
aI39
aI297
aI1017
aI1191
aI2227
aI6085
aI5117
aI16569
aI64743
aI29329
aI157279
asS'd'
I20384
sa(dp40768
S'a'
I117301
sS's'
I18
sS'm'
(lp40769
I1
aI1
aI5
aI5
aI15
aI47
aI111
aI61
aI435
aI657
aI141
aI3445
aI6921
aI7759
aI30141
aI37631
aI85969
aI227563
asS'd'
I20385
sa(dp40770
S'a'
I117319
sS's'
I18
sS'm'
(lp40771
I1
aI3
aI1
aI13
aI27
aI39
aI15
aI167
aI151
aI185
aI1513
aI211
aI951
aI12705
aI25703
aI29289
aI120993
aI156741
asS'd'
I20386
sa(dp40772
S'a'
I117338
sS's'
I18
sS'm'
(lp40773
I1
aI3
aI7
aI7
aI7
aI39
aI19
aI221
aI351
aI951
aI1231
aI1915
aI3043
aI189
aI18977
aI50149
aI56583
aI122147
asS'd'
I20387
sa(dp40774
S'a'
I117347
sS's'
I18
sS'm'
(lp40775
I1
aI1
aI5
aI15
aI29
aI37
aI77
aI207
aI291
aI851
aI131
aI1041
aI1657
aI4393
aI5023
aI12745
aI32253
aI204431
asS'd'
I20388
sa(dp40776
S'a'
I117361
sS's'
I18
sS'm'
(lp40777
I1
aI1
aI7
aI15
aI11
aI59
aI85
aI255
aI67
aI23
aI1321
aI2153
aI7043
aI417
aI15719
aI59937
aI37619
aI109331
asS'd'
I20389
sa(dp40778
S'a'
I117380
sS's'
I18
sS'm'
(lp40779
I1
aI3
aI7
aI15
aI25
aI37
aI43
aI15
aI385
aI735
aI1741
aI3655
aI4215
aI1097
aI19519
aI44313
aI99851
aI204717
asS'd'
I20390
sa(dp40780
S'a'
I117389
sS's'
I18
sS'm'
(lp40781
I1
aI1
aI3
aI7
aI15
aI17
aI17
aI105
aI399
aI49
aI105
aI159
aI465
aI11991
aI29797
aI23907
aI129609
aI179013
asS'd'
I20391
sa(dp40782
S'a'
I117398
sS's'
I18
sS'm'
(lp40783
I1
aI1
aI7
aI13
aI9
aI17
aI87
aI51
aI391
aI695
aI545
aI3061
aI4499
aI2059
aI10095
aI13847
aI68519
aI60611
asS'd'
I20392
sa(dp40784
S'a'
I117435
sS's'
I18
sS'm'
(lp40785
I1
aI3
aI3
aI1
aI31
aI7
aI11
aI233
aI231
aI189
aI1599
aI1589
aI401
aI8759
aI17273
aI43613
aI48709
aI253521
asS'd'
I20393
sa(dp40786
S'a'
I117458
sS's'
I18
sS'm'
(lp40787
I1
aI1
aI1
aI11
aI5
aI9
aI27
aI77
aI491
aI951
aI579
aI1635
aI3241
aI14497
aI27149
aI45001
aI56769
aI160731
asS'd'
I20394
sa(dp40788
S'a'
I117476
sS's'
I18
sS'm'
(lp40789
I1
aI3
aI7
aI15
aI29
aI11
aI125
aI101
aI19
aI971
aI107
aI1525
aI3939
aI7633
aI16355
aI24727
aI19475
aI157571
asS'd'
I20395
sa(dp40790
S'a'
I117498
sS's'
I18
sS'm'
(lp40791
I1
aI1
aI1
aI9
aI25
aI7
aI35
aI187
aI321
aI483
aI1919
aI1911
aI7869
aI12903
aI26977
aI49419
aI24973
aI214731
asS'd'
I20396
sa(dp40792
S'a'
I117500
sS's'
I18
sS'm'
(lp40793
I1
aI1
aI5
aI3
aI23
aI1
aI11
aI143
aI315
aI1015
aI1367
aI1555
aI1041
aI6655
aI10481
aI49275
aI49575
aI101061
asS'd'
I20397
sa(dp40794
S'a'
I117529
sS's'
I18
sS'm'
(lp40795
I1
aI3
aI5
aI15
aI15
aI59
aI13
aI117
aI217
aI975
aI1821
aI3829
aI1545
aI921
aI20875
aI43305
aI18793
aI158651
asS'd'
I20398
sa(dp40796
S'a'
I117554
sS's'
I18
sS'm'
(lp40797
I1
aI3
aI3
aI11
aI23
aI23
aI91
aI7
aI29
aI613
aI1093
aI3881
aI3301
aI3751
aI16137
aI48277
aI119813
aI177341
asS'd'
I20399
sa(dp40798
S'a'
I117556
sS's'
I18
sS'm'
(lp40799
I1
aI3
aI3
aI15
aI7
aI37
aI115
aI19
aI147
aI585
aI1877
aI2395
aI3343
aI9567
aI16199
aI13969
aI89731
aI124835
asS'd'
I20400
sa(dp40800
S'a'
I117565
sS's'
I18
sS'm'
(lp40801
I1
aI1
aI5
aI7
aI29
aI3
aI59
aI141
aI375
aI527
aI1219
aI409
aI7155
aI2823
aI32497
aI23103
aI73187
aI53089
asS'd'
I20401
sa(dp40802
S'a'
I117580
sS's'
I18
sS'm'
(lp40803
I1
aI1
aI5
aI3
aI13
aI27
aI111
aI63
aI189
aI813
aI643
aI19
aI3461
aI13891
aI26651
aI52395
aI74729
aI148397
asS'd'
I20402
sa(dp40804
S'a'
I117585
sS's'
I18
sS'm'
(lp40805
I1
aI3
aI5
aI1
aI27
aI61
aI97
aI227
aI123
aI829
aI1559
aI2523
aI7737
aI6047
aI213
aI23613
aI61571
aI7093
asS'd'
I20403
sa(dp40806
S'a'
I117592
sS's'
I18
sS'm'
(lp40807
I1
aI1
aI1
aI11
aI17
aI1
aI73
aI203
aI391
aI937
aI321
aI3431
aI7163
aI3547
aI29467
aI65271
aI69775
aI226405
asS'd'
I20404
sa(dp40808
S'a'
I117597
sS's'
I18
sS'm'
(lp40809
I1
aI1
aI5
aI15
aI25
aI7
aI75
aI199
aI511
aI731
aI1547
aI2127
aI1609
aI5623
aI26771
aI29935
aI76671
aI178683
asS'd'
I20405
sa(dp40810
S'a'
I117616
sS's'
I18
sS'm'
(lp40811
I1
aI3
aI5
aI15
aI25
aI49
aI99
aI23
aI281
aI81
aI507
aI1499
aI5235
aI9945
aI14099
aI5993
aI319
aI178581
asS'd'
I20406
sa(dp40812
S'a'
I117637
sS's'
I18
sS'm'
(lp40813
I1
aI3
aI7
aI9
aI5
aI13
aI105
aI7
aI135
aI827
aI927
aI3463
aI839
aI7047
aI19863
aI63859
aI13951
aI221795
asS'd'
I20407
sa(dp40814
S'a'
I117644
sS's'
I18
sS'm'
(lp40815
I1
aI1
aI1
aI13
aI7
aI21
aI59
aI9
aI467
aI299
aI1035
aI1395
aI7413
aI7313
aI24769
aI44043
aI50679
aI72867
asS'd'
I20408
sa(dp40816
S'a'
I117659
sS's'
I18
sS'm'
(lp40817
I1
aI1
aI7
aI1
aI31
aI33
aI95
aI155
aI429
aI413
aI493
aI2025
aI2069
aI551
aI507
aI13515
aI3507
aI93873
asS'd'
I20409
sa(dp40818
S'a'
I117703
sS's'
I18
sS'm'
(lp40819
I1
aI1
aI3
aI5
aI3
aI33
aI109
aI1
aI299
aI727
aI495
aI2981
aI3795
aI11467
aI27173
aI4171
aI6859
aI129961
asS'd'
I20410
sa(dp40820
S'a'
I117704
sS's'
I18
sS'm'
(lp40821
I1
aI1
aI7
aI9
aI23
aI41
aI113
aI103
aI161
aI303
aI1565
aI2637
aI7113
aI11635
aI13707
aI3559
aI21007
aI250107
asS'd'
I20411
sa(dp40822
S'a'
I117727
sS's'
I18
sS'm'
(lp40823
I1
aI1
aI5
aI13
aI1
aI49
aI11
aI87
aI31
aI77
aI1847
aI1137
aI3031
aI1943
aI28755
aI32197
aI96043
aI152447
asS'd'
I20412
sa(dp40824
S'a'
I117738
sS's'
I18
sS'm'
(lp40825
I1
aI1
aI3
aI7
aI3
aI3
aI123
aI65
aI175
aI809
aI681
aI2135
aI5279
aI7119
aI4573
aI19287
aI90235
aI183391
asS'd'
I20413
sa(dp40826
S'a'
I117770
sS's'
I18
sS'm'
(lp40827
I1
aI3
aI1
aI13
aI7
aI5
aI25
aI151
aI437
aI155
aI1841
aI219
aI5641
aI12097
aI6153
aI11
aI60315
aI169293
asS'd'
I20414
sa(dp40828
S'a'
I117796
sS's'
I18
sS'm'
(lp40829
I1
aI1
aI7
aI5
aI21
aI29
aI23
aI83
aI35
aI651
aI1507
aI635
aI3867
aI12133
aI25523
aI55341
aI105741
aI240349
asS'd'
I20415
sa(dp40830
S'a'
I117799
sS's'
I18
sS'm'
(lp40831
I1
aI1
aI1
aI15
aI9
aI29
aI27
aI151
aI463
aI747
aI547
aI577
aI1263
aI15235
aI6695
aI60849
aI72231
aI175671
asS'd'
I20416
sa(dp40832
S'a'
I117820
sS's'
I18
sS'm'
(lp40833
I1
aI3
aI5
aI11
aI11
aI43
aI81
aI37
aI505
aI509
aI1325
aI3295
aI839
aI5855
aI19795
aI1403
aI15711
aI219481
asS'd'
I20417
sa(dp40834
S'a'
I117832
sS's'
I18
sS'm'
(lp40835
I1
aI3
aI3
aI15
aI25
aI61
aI121
aI37
aI201
aI133
aI537
aI1345
aI4213
aI13023
aI18795
aI8949
aI84431
aI105521
asS'd'
I20418
sa(dp40836
S'a'
I117840
sS's'
I18
sS'm'
(lp40837
I1
aI3
aI7
aI11
aI13
aI51
aI87
aI245
aI357
aI7
aI699
aI2003
aI5963
aI1399
aI69
aI19083
aI114585
aI232313
asS'd'
I20419
sa(dp40838
S'a'
I117843
sS's'
I18
sS'm'
(lp40839
I1
aI1
aI1
aI13
aI9
aI29
aI65
aI123
aI37
aI885
aI227
aI2795
aI1037
aI10905
aI21217
aI4081
aI77643
aI254245
asS'd'
I20420
sa(dp40840
S'a'
I117865
sS's'
I18
sS'm'
(lp40841
I1
aI3
aI1
aI3
aI25
aI23
aI71
aI189
aI253
aI785
aI1337
aI1275
aI3285
aI1067
aI8607
aI3883
aI119099
aI116637
asS'd'
I20421
sa(dp40842
S'a'
I117868
sS's'
I18
sS'm'
(lp40843
I1
aI1
aI1
aI15
aI19
aI43
aI17
aI89
aI257
aI175
aI1943
aI207
aI597
aI9279
aI405
aI33209
aI65221
aI39557
asS'd'
I20422
sa(dp40844
S'a'
I117874
sS's'
I18
sS'm'
(lp40845
I1
aI1
aI7
aI5
aI7
aI47
aI127
aI11
aI197
aI871
aI23
aI1951
aI6829
aI7831
aI5223
aI56287
aI115649
aI114283
asS'd'
I20423
sa(dp40846
S'a'
I117886
sS's'
I18
sS'm'
(lp40847
I1
aI3
aI1
aI3
aI1
aI35
aI81
aI189
aI19
aI117
aI1683
aI469
aI8117
aI5449
aI22871
aI5505
aI125111
aI128717
asS'd'
I20424
sa(dp40848
S'a'
I117962
sS's'
I18
sS'm'
(lp40849
I1
aI3
aI7
aI5
aI7
aI31
aI105
aI57
aI387
aI691
aI1293
aI3103
aI2329
aI16247
aI18357
aI55453
aI112633
aI225641
asS'd'
I20425
sa(dp40850
S'a'
I117967
sS's'
I18
sS'm'
(lp40851
I1
aI3
aI1
aI3
aI13
aI3
aI65
aI25
aI47
aI413
aI521
aI3507
aI1793
aI14431
aI22341
aI39813
aI46399
aI204501
asS'd'
I20426
sa(dp40852
S'a'
I117995
sS's'
I18
sS'm'
(lp40853
I1
aI3
aI1
aI1
aI31
aI37
aI21
aI45
aI261
aI665
aI1243
aI1937
aI5001
aI3789
aI26473
aI20153
aI107131
aI75523
asS'd'
I20427
sa(dp40854
S'a'
I118000
sS's'
I18
sS'm'
(lp40855
I1
aI3
aI5
aI15
aI27
aI61
aI109
aI139
aI19
aI583
aI353
aI445
aI53
aI67
aI20753
aI57827
aI116527
aI55109
asS'd'
I20428
sa(dp40856
S'a'
I118017
sS's'
I18
sS'm'
(lp40857
I1
aI1
aI3
aI3
aI1
aI37
aI113
aI21
aI305
aI967
aI1703
aI2095
aI1059
aI2843
aI22381
aI24871
aI24765
aI52425
asS'd'
I20429
sa(dp40858
S'a'
I118037
sS's'
I18
sS'm'
(lp40859
I1
aI3
aI7
aI11
aI27
aI59
aI111
aI111
aI283
aI79
aI1227
aI3631
aI4169
aI5671
aI7769
aI56553
aI75503
aI206259
asS'd'
I20430
sa(dp40860
S'a'
I118038
sS's'
I18
sS'm'
(lp40861
I1
aI1
aI5
aI11
aI11
aI1
aI127
aI13
aI17
aI255
aI1383
aI2879
aI6785
aI289
aI7061
aI53067
aI11539
aI131405
asS'd'
I20431
sa(dp40862
S'a'
I118051
sS's'
I18
sS'm'
(lp40863
I1
aI1
aI5
aI15
aI29
aI27
aI67
aI15
aI247
aI689
aI579
aI3237
aI5279
aI13847
aI20305
aI60237
aI115841
aI144855
asS'd'
I20432
sa(dp40864
S'a'
I118057
sS's'
I18
sS'm'
(lp40865
I1
aI1
aI5
aI1
aI9
aI25
aI75
aI11
aI83
aI1015
aI281
aI1617
aI7449
aI10673
aI7033
aI38839
aI113703
aI233101
asS'd'
I20433
sa(dp40866
S'a'
I118072
sS's'
I18
sS'm'
(lp40867
I1
aI3
aI3
aI3
aI23
aI41
aI81
aI109
aI199
aI969
aI935
aI1793
aI6921
aI4013
aI9625
aI48149
aI54395
aI1193
asS'd'
I20434
sa(dp40868
S'a'
I118080
sS's'
I18
sS'm'
(lp40869
I1
aI3
aI5
aI7
aI19
aI63
aI25
aI201
aI63
aI799
aI765
aI533
aI1417
aI3199
aI7773
aI44247
aI112207
aI11783
asS'd'
I20435
sa(dp40870
S'a'
I118098
sS's'
I18
sS'm'
(lp40871
I1
aI3
aI1
aI11
aI7
aI25
aI87
aI159
aI491
aI749
aI1157
aI667
aI2951
aI12019
aI22259
aI36933
aI124159
aI176041
asS'd'
I20436
sa(dp40872
S'a'
I118107
sS's'
I18
sS'm'
(lp40873
I1
aI1
aI5
aI15
aI21
aI19
aI113
aI175
aI129
aI385
aI2025
aI2685
aI1925
aI8547
aI4835
aI15953
aI128023
aI236341
asS'd'
I20437
sa(dp40874
S'a'
I118109
sS's'
I18
sS'm'
(lp40875
I1
aI1
aI1
aI9
aI13
aI47
aI25
aI81
aI389
aI249
aI1857
aI1061
aI4439
aI3717
aI16299
aI23247
aI95275
aI222701
asS'd'
I20438
sa(dp40876
S'a'
I118114
sS's'
I18
sS'm'
(lp40877
I1
aI3
aI3
aI1
aI3
aI61
aI61
aI117
aI159
aI689
aI43
aI113
aI4203
aI7699
aI27607
aI37195
aI63415
aI90481
asS'd'
I20439
sa(dp40878
S'a'
I118123
sS's'
I18
sS'm'
(lp40879
I1
aI3
aI1
aI5
aI11
aI49
aI73
aI13
aI307
aI655
aI645
aI2765
aI6079
aI12687
aI22417
aI44713
aI5247
aI40265
asS'd'
I20440
sa(dp40880
S'a'
I118150
sS's'
I18
sS'm'
(lp40881
I1
aI1
aI7
aI13
aI5
aI55
aI57
aI237
aI317
aI101
aI481
aI2515
aI707
aI6385
aI9421
aI50557
aI92395
aI193737
asS'd'
I20441
sa(dp40882
S'a'
I118159
sS's'
I18
sS'm'
(lp40883
I1
aI3
aI5
aI1
aI27
aI35
aI65
aI107
aI63
aI57
aI1699
aI4077
aI4279
aI8547
aI15137
aI11533
aI117641
aI64925
asS'd'
I20442
sa(dp40884
S'a'
I118164
sS's'
I18
sS'm'
(lp40885
I1
aI1
aI5
aI1
aI13
aI7
aI7
aI141
aI305
aI191
aI2033
aI2677
aI6025
aI12927
aI4057
aI12047
aI60253
aI90803
asS'd'
I20443
sa(dp40886
S'a'
I118173
sS's'
I18
sS'm'
(lp40887
I1
aI1
aI7
aI3
aI1
aI9
aI63
aI233
aI185
aI97
aI913
aI187
aI4321
aI8951
aI27669
aI27035
aI30029
aI218725
asS'd'
I20444
sa(dp40888
S'a'
I118201
sS's'
I18
sS'm'
(lp40889
I1
aI1
aI7
aI1
aI11
aI59
aI41
aI195
aI335
aI551
aI491
aI3079
aI7777
aI4003
aI24543
aI17165
aI103261
aI167505
asS'd'
I20445
sa(dp40890
S'a'
I118212
sS's'
I18
sS'm'
(lp40891
I1
aI1
aI3
aI3
aI9
aI59
aI37
aI185
aI289
aI845
aI1083
aI63
aI7439
aI4677
aI29245
aI40813
aI16295
aI45499
asS'd'
I20446
sa(dp40892
S'a'
I118246
sS's'
I18
sS'm'
(lp40893
I1
aI3
aI5
aI1
aI1
aI37
aI89
aI5
aI277
aI493
aI155
aI1641
aI5395
aI11389
aI26247
aI2833
aI103803
aI74447
asS'd'
I20447
sa(dp40894
S'a'
I118257
sS's'
I18
sS'm'
(lp40895
I1
aI1
aI3
aI1
aI29
aI55
aI83
aI211
aI377
aI583
aI1075
aI2679
aI7157
aI11719
aI1653
aI5977
aI52263
aI45531
asS'd'
I20448
sa(dp40896
S'a'
I118303
sS's'
I18
sS'm'
(lp40897
I1
aI1
aI5
aI15
aI1
aI31
aI89
aI7
aI239
aI821
aI887
aI1319
aI225
aI14555
aI5443
aI44717
aI99803
aI241577
asS'd'
I20449
sa(dp40898
S'a'
I118310
sS's'
I18
sS'm'
(lp40899
I1
aI3
aI5
aI11
aI7
aI23
aI81
aI161
aI67
aI1011
aI177
aI2837
aI7767
aI14385
aI29415
aI9377
aI7407
aI128403
asS'd'
I20450
sa(dp40900
S'a'
I118313
sS's'
I18
sS'm'
(lp40901
I1
aI1
aI7
aI11
aI3
aI13
aI13
aI237
aI199
aI601
aI481
aI3809
aI6591
aI8497
aI25361
aI22547
aI28317
aI22961
asS'd'
I20451
sa(dp40902
S'a'
I118324
sS's'
I18
sS'm'
(lp40903
I1
aI3
aI1
aI1
aI31
aI29
aI105
aI161
aI483
aI391
aI321
aI1087
aI4149
aI8803
aI22291
aI24611
aI114447
aI33645
asS'd'
I20452
sa(dp40904
S'a'
I118334
sS's'
I18
sS'm'
(lp40905
I1
aI3
aI1
aI11
aI1
aI47
aI41
aI45
aI287
aI503
aI169
aI2265
aI1835
aI6609
aI25245
aI7069
aI61137
aI160653
asS'd'
I20453
sa(dp40906
S'a'
I118336
sS's'
I18
sS'm'
(lp40907
I1
aI1
aI7
aI13
aI11
aI39
aI29
aI39
aI489
aI205
aI741
aI2871
aI377
aI10679
aI11689
aI50947
aI85309
aI95697
asS'd'
I20454
sa(dp40908
S'a'
I118339
sS's'
I18
sS'm'
(lp40909
I1
aI3
aI7
aI7
aI23
aI19
aI103
aI15
aI79
aI425
aI369
aI2009
aI4417
aI11031
aI2113
aI36969
aI73241
aI120903
asS'd'
I20455
sa(dp40910
S'a'
I118353
sS's'
I18
sS'm'
(lp40911
I1
aI3
aI1
aI7
aI27
aI9
aI43
aI33
aI123
aI895
aI223
aI1045
aI2701
aI3339
aI12099
aI24449
aI52973
aI175671
asS'd'
I20456
sa(dp40912
S'a'
I118354
sS's'
I18
sS'm'
(lp40913
I1
aI3
aI5
aI5
aI3
aI1
aI13
aI117
aI429
aI167
aI1361
aI2299
aI7565
aI1153
aI9259
aI29209
aI25747
aI71005
asS'd'
I20457
sa(dp40914
S'a'
I118390
sS's'
I18
sS'm'
(lp40915
I1
aI3
aI1
aI15
aI7
aI7
aI13
aI209
aI73
aI523
aI1549
aI2545
aI5583
aI10209
aI27205
aI41243
aI14217
aI208993
asS'd'
I20458
sa(dp40916
S'a'
I118396
sS's'
I18
sS'm'
(lp40917
I1
aI3
aI5
aI5
aI3
aI29
aI99
aI255
aI479
aI297
aI1319
aI2171
aI7321
aI14425
aI15869
aI44449
aI10917
aI171165
asS'd'
I20459
sa(dp40918
S'a'
I118436
sS's'
I18
sS'm'
(lp40919
I1
aI3
aI3
aI11
aI13
aI49
aI29
aI95
aI79
aI987
aI161
aI859
aI6503
aI8839
aI14131
aI30249
aI16183
aI40257
asS'd'
I20460
sa(dp40920
S'a'
I118453
sS's'
I18
sS'm'
(lp40921
I1
aI3
aI3
aI11
aI27
aI7
aI27
aI33
aI255
aI847
aI789
aI3897
aI2599
aI16107
aI22379
aI1853
aI102713
aI197547
asS'd'
I20461
sa(dp40922
S'a'
I118466
sS's'
I18
sS'm'
(lp40923
I1
aI1
aI3
aI9
aI29
aI11
aI107
aI227
aI35
aI183
aI639
aI1585
aI313
aI1451
aI19789
aI13855
aI94277
aI85569
asS'd'
I20462
sa(dp40924
S'a'
I118492
sS's'
I18
sS'm'
(lp40925
I1
aI3
aI7
aI7
aI25
aI33
aI101
aI49
aI137
aI457
aI2027
aI3317
aI1961
aI6097
aI739
aI12875
aI69503
aI95453
asS'd'
I20463
sa(dp40926
S'a'
I118506
sS's'
I18
sS'm'
(lp40927
I1
aI3
aI3
aI11
aI29
aI13
aI127
aI3
aI31
aI319
aI1341
aI927
aI5067
aI13891
aI31265
aI41381
aI49341
aI160343
asS'd'
I20464
sa(dp40928
S'a'
I118525
sS's'
I18
sS'm'
(lp40929
I1
aI1
aI3
aI15
aI31
aI21
aI93
aI155
aI471
aI707
aI1395
aI2995
aI867
aI10353
aI8137
aI44267
aI24823
aI6113
asS'd'
I20465
sa(dp40930
S'a'
I118557
sS's'
I18
sS'm'
(lp40931
I1
aI3
aI1
aI7
aI7
aI33
aI83
aI79
aI349
aI687
aI1045
aI1183
aI4441
aI15199
aI1953
aI36395
aI84691
aI134939
asS'd'
I20466
sa(dp40932
S'a'
I118561
sS's'
I18
sS'm'
(lp40933
I1
aI1
aI3
aI15
aI13
aI21
aI41
aI105
aI189
aI439
aI1171
aI4005
aI7641
aI1597
aI24317
aI58749
aI35539
aI220647
asS'd'
I20467
sa(dp40934
S'a'
I118586
sS's'
I18
sS'm'
(lp40935
I1
aI1
aI3
aI15
aI13
aI39
aI49
aI5
aI461
aI613
aI1633
aI1951
aI7959
aI5733
aI10061
aI18829
aI49505
aI90033
asS'd'
I20468
sa(dp40936
S'a'
I118606
sS's'
I18
sS'm'
(lp40937
I1
aI3
aI5
aI5
aI29
aI25
aI19
aI227
aI379
aI525
aI687
aI2629
aI7729
aI4791
aI5911
aI14481
aI49063
aI216669
asS'd'
I20469
sa(dp40938
S'a'
I118627
sS's'
I18
sS'm'
(lp40939
I1
aI3
aI7
aI7
aI7
aI63
aI93
aI227
aI79
aI165
aI1971
aI1695
aI4485
aI6009
aI8769
aI12861
aI83653
aI27667
asS'd'
I20470
sa(dp40940
S'a'
I118641
sS's'
I18
sS'm'
(lp40941
I1
aI1
aI5
aI5
aI25
aI11
aI23
aI89
aI363
aI491
aI459
aI4063
aI3787
aI9375
aI28011
aI44757
aI56441
aI116609
asS'd'
I20471
sa(dp40942
S'a'
I118654
sS's'
I18
sS'm'
(lp40943
I1
aI3
aI5
aI9
aI7
aI53
aI43
aI149
aI435
aI35
aI135
aI1759
aI3197
aI7749
aI12731
aI28295
aI25901
aI125847
asS'd'
I20472
sa(dp40944
S'a'
I118687
sS's'
I18
sS'm'
(lp40945
I1
aI1
aI3
aI5
aI15
aI15
aI1
aI215
aI307
aI711
aI1971
aI2795
aI677
aI11921
aI10303
aI37997
aI6653
aI51295
asS'd'
I20473
sa(dp40946
S'a'
I118688
sS's'
I18
sS'm'
(lp40947
I1
aI1
aI5
aI15
aI1
aI59
aI75
aI195
aI51
aI215
aI1303
aI3023
aI8023
aI10951
aI13015
aI23513
aI37029
aI23581
asS'd'
I20474
sa(dp40948
S'a'
I118735
sS's'
I18
sS'm'
(lp40949
I1
aI1
aI1
aI3
aI27
aI3
aI77
aI165
aI97
aI499
aI937
aI1129
aI6649
aI11305
aI27763
aI32849
aI78251
aI210407
asS'd'
I20475
sa(dp40950
S'a'
I118737
sS's'
I18
sS'm'
(lp40951
I1
aI3
aI3
aI3
aI9
aI21
aI19
aI197
aI339
aI53
aI1875
aI1057
aI3485
aI14645
aI13417
aI39307
aI81437
aI45857
asS'd'
I20476
sa(dp40952
S'a'
I118750
sS's'
I18
sS'm'
(lp40953
I1
aI1
aI1
aI5
aI19
aI5
aI95
aI205
aI399
aI699
aI819
aI1927
aI7913
aI8109
aI1223
aI28595
aI397
aI81051
asS'd'
I20477
sa(dp40954
S'a'
I118773
sS's'
I18
sS'm'
(lp40955
I1
aI3
aI7
aI13
aI23
aI11
aI37
aI167
aI189
aI813
aI1199
aI3545
aI655
aI13239
aI10469
aI33895
aI119025
aI185361
asS'd'
I20478
sa(dp40956
S'a'
I118778
sS's'
I18
sS'm'
(lp40957
I1
aI1
aI5
aI11
aI9
aI13
aI41
aI37
aI443
aI269
aI1199
aI1347
aI7081
aI11273
aI14389
aI64083
aI117901
aI51903
asS'd'
I20479
sa(dp40958
S'a'
I118786
sS's'
I18
sS'm'
(lp40959
I1
aI1
aI3
aI13
aI5
aI17
aI117
aI151
aI155
aI637
aI731
aI1839
aI855
aI9749
aI19529
aI18101
aI20341
aI21941
asS'd'
I20480
sa(dp40960
S'a'
I118809
sS's'
I18
sS'm'
(lp40961
I1
aI1
aI1
aI11
aI15
aI17
aI35
aI9
aI91
aI907
aI667
aI853
aI1455
aI10097
aI31277
aI749
aI47089
aI219517
asS'd'
I20481
sa(dp40962
S'a'
I118833
sS's'
I18
sS'm'
(lp40963
I1
aI3
aI7
aI1
aI27
aI31
aI9
aI145
aI309
aI811
aI5
aI2645
aI7851
aI1953
aI21427
aI18805
aI108755
aI77215
asS'd'
I20482
sa(dp40964
S'a'
I118853
sS's'
I18
sS'm'
(lp40965
I1
aI1
aI7
aI3
aI21
aI7
aI25
aI233
aI145
aI1015
aI43
aI2205
aI4735
aI13257
aI24001
aI50469
aI42567
aI253745
asS'd'
I20483
sa(dp40966
S'a'
I118854
sS's'
I18
sS'm'
(lp40967
I1
aI3
aI1
aI15
aI3
aI53
aI51
aI45
aI397
aI379
aI641
aI895
aI7569
aI15783
aI23923
aI6147
aI121395
aI261853
asS'd'
I20484
sa(dp40968
S'a'
I118863
sS's'
I18
sS'm'
(lp40969
I1
aI1
aI1
aI1
aI23
aI47
aI35
aI125
aI289
aI65
aI1875
aI2309
aI519
aI5435
aI5271
aI25319
aI14557
aI19389
asS'd'
I20485
sa(dp40970
S'a'
I118881
sS's'
I18
sS'm'
(lp40971
I1
aI3
aI5
aI7
aI27
aI43
aI15
aI61
aI357
aI791
aI1781
aI3671
aI3911
aI1325
aI5607
aI44107
aI67873
aI119849
asS'd'
I20486
sa(dp40972
S'a'
I118896
sS's'
I18
sS'm'
(lp40973
I1
aI3
aI1
aI5
aI25
aI37
aI105
aI51
aI491
aI407
aI475
aI1763
aI1425
aI14883
aI31435
aI48979
aI120667
aI131089
asS'd'
I20487
sa(dp40974
S'a'
I118917
sS's'
I18
sS'm'
(lp40975
I1
aI3
aI5
aI13
aI1
aI51
aI109
aI161
aI215
aI871
aI185
aI2389
aI7977
aI6705
aI14045
aI45569
aI44557
aI114795
asS'd'
I20488
sa(dp40976
S'a'
I118921
sS's'
I18
sS'm'
(lp40977
I1
aI3
aI7
aI11
aI17
aI21
aI111
aI183
aI343
aI593
aI447
aI3995
aI759
aI3709
aI32655
aI30141
aI127225
aI120899
asS'd'
I20489
sa(dp40978
S'a'
I118939
sS's'
I18
sS'm'
(lp40979
I1
aI1
aI5
aI3
aI17
aI3
aI69
aI25
aI113
aI897
aI1933
aI2717
aI2003
aI5847
aI2541
aI62415
aI50975
aI97903
asS'd'
I20490
sa(dp40980
S'a'
I118948
sS's'
I18
sS'm'
(lp40981
I1
aI1
aI1
aI1
aI9
aI35
aI107
aI81
aI257
aI57
aI1719
aI4049
aI1237
aI10659
aI4689
aI20887
aI90791
aI251911
asS'd'
I20491
sa(dp40982
S'a'
I118960
sS's'
I18
sS'm'
(lp40983
I1
aI1
aI7
aI11
aI1
aI49
aI83
aI213
aI169
aI169
aI825
aI2983
aI5833
aI2413
aI32165
aI47459
aI129021
aI156217
asS'd'
I20492
sa(dp40984
S'a'
I118970
sS's'
I18
sS'm'
(lp40985
I1
aI1
aI5
aI13
aI15
aI39
aI61
aI93
aI407
aI553
aI839
aI4035
aI6609
aI6327
aI945
aI49625
aI127867
aI240161
asS'd'
I20493
sa(dp40986
S'a'
I118983
sS's'
I18
sS'm'
(lp40987
I1
aI1
aI1
aI15
aI23
aI47
aI21
aI235
aI81
aI431
aI1819
aI1141
aI7973
aI4623
aI4539
aI23201
aI83111
aI230857
asS'd'
I20494
sa(dp40988
S'a'
I119025
sS's'
I18
sS'm'
(lp40989
I1
aI3
aI3
aI15
aI9
aI15
aI15
aI173
aI29
aI803
aI1453
aI2621
aI8095
aI6639
aI6607
aI21471
aI44785
aI122271
asS'd'
I20495
sa(dp40990
S'a'
I119026
sS's'
I18
sS'm'
(lp40991
I1
aI3
aI3
aI7
aI21
aI5
aI127
aI203
aI43
aI581
aI1925
aI165
aI4615
aI9141
aI18563
aI54413
aI71559
aI172791
asS'd'
I20496
sa(dp40992
S'a'
I119064
sS's'
I18
sS'm'
(lp40993
I1
aI3
aI3
aI3
aI27
aI31
aI55
aI3
aI381
aI971
aI1087
aI2659
aI139
aI10935
aI9189
aI3445
aI15071
aI218873
asS'd'
I20497
sa(dp40994
S'a'
I119069
sS's'
I18
sS'm'
(lp40995
I1
aI3
aI1
aI3
aI1
aI57
aI47
aI71
aI85
aI335
aI207
aI225
aI2931
aI14721
aI21431
aI17199
aI745
aI177403
asS'd'
I20498
sa(dp40996
S'a'
I119079
sS's'
I18
sS'm'
(lp40997
I1
aI3
aI1
aI7
aI7
aI17
aI21
aI57
aI345
aI29
aI1147
aI1179
aI7371
aI14725
aI27445
aI62061
aI16483
aI112489
asS'd'
I20499
sa(dp40998
S'a'
I119085
sS's'
I18
sS'm'
(lp40999
I1
aI3
aI7
aI15
aI23
aI11
aI91
aI1
aI1
aI117
aI1665
aI3899
aI5683
aI14497
aI25633
aI6233
aI104029
aI22155
asS'd'
I20500
sa(dp41000
S'a'
I119115
sS's'
I18
sS'm'
(lp41001
I1
aI3
aI5
aI9
aI1
aI27
aI15
aI187
aI37
aI329
aI585
aI729
aI5651
aI15715
aI4339
aI1899
aI90611
aI195643
asS'd'
I20501
sa(dp41002
S'a'
I119165
sS's'
I18
sS'm'
(lp41003
I1
aI1
aI3
aI11
aI1
aI45
aI95
aI103
aI13
aI83
aI319
aI2295
aI6333
aI13469
aI19237
aI55985
aI129725
aI141699
asS'd'
I20502
sa(dp41004
S'a'
I119166
sS's'
I18
sS'm'
(lp41005
I1
aI1
aI5
aI11
aI7
aI41
aI97
aI159
aI511
aI617
aI1545
aI3023
aI7919
aI8437
aI8345
aI16701
aI69053
aI105047
asS'd'
I20503
sa(dp41006
S'a'
I119193
sS's'
I18
sS'm'
(lp41007
I1
aI3
aI3
aI11
aI29
aI35
aI23
aI95
aI277
aI931
aI857
aI3887
aI4597
aI10841
aI12947
aI18009
aI61499
aI242827
asS'd'
I20504
sa(dp41008
S'a'
I119194
sS's'
I18
sS'm'
(lp41009
I1
aI1
aI3
aI3
aI7
aI63
aI35
aI161
aI125
aI637
aI149
aI1045
aI3297
aI16213
aI1543
aI8073
aI80373
aI61507
asS'd'
I20505
sa(dp41010
S'a'
I119199
sS's'
I18
sS'm'
(lp41011
I1
aI1
aI7
aI5
aI31
aI43
aI113
aI189
aI181
aI659
aI1971
aI3309
aI4237
aI4279
aI31563
aI29429
aI17443
aI154385
asS'd'
I20506
sa(dp41012
S'a'
I119215
sS's'
I18
sS'm'
(lp41013
I1
aI1
aI1
aI13
aI19
aI43
aI99
aI91
aI47
aI477
aI153
aI3295
aI6281
aI779
aI17169
aI343
aI1723
aI171133
asS'd'
I20507
sa(dp41014
S'a'
I119218
sS's'
I18
sS'm'
(lp41015
I1
aI1
aI1
aI15
aI9
aI35
aI123
aI45
aI417
aI631
aI1415
aI1835
aI3063
aI897
aI18947
aI62477
aI12759
aI97831
asS'd'
I20508
sa(dp41016
S'a'
I119241
sS's'
I18
sS'm'
(lp41017
I1
aI1
aI3
aI13
aI31
aI1
aI31
aI139
aI411
aI605
aI1829
aI303
aI3891
aI15807
aI7335
aI44833
aI87427
aI62183
asS'd'
I20509
sa(dp41018
S'a'
I119277
sS's'
I18
sS'm'
(lp41019
I1
aI1
aI3
aI13
aI29
aI55
aI97
aI107
aI241
aI981
aI1281
aI3295
aI6825
aI15865
aI4221
aI24695
aI54203
aI252069
asS'd'
I20510
sa(dp41020
S'a'
I119301
sS's'
I18
sS'm'
(lp41021
I1
aI3
aI7
aI13
aI1
aI3
aI33
aI165
aI483
aI813
aI127
aI1717
aI8077
aI3521
aI23465
aI41705
aI2769
aI173233
asS'd'
I20511
sa(dp41022
S'a'
I119306
sS's'
I18
sS'm'
(lp41023
I1
aI3
aI1
aI11
aI11
aI51
aI39
aI125
aI375
aI825
aI1775
aI2923
aI4903
aI4779
aI907
aI47787
aI22293
aI169631
asS'd'
I20512
sa(dp41024
S'a'
I119335
sS's'
I18
sS'm'
(lp41025
I1
aI3
aI5
aI5
aI13
aI27
aI43
aI229
aI267
aI153
aI567
aI3403
aI2103
aI6203
aI29629
aI29715
aI116735
aI122515
asS'd'
I20513
sa(dp41026
S'a'
I119344
sS's'
I18
sS'm'
(lp41027
I1
aI3
aI5
aI13
aI3
aI3
aI15
aI5
aI345
aI343
aI691
aI3703
aI361
aI2019
aI9309
aI26909
aI22897
aI103555
asS'd'
I20514
sa(dp41028
S'a'
I119364
sS's'
I18
sS'm'
(lp41029
I1
aI3
aI1
aI11
aI9
aI25
aI123
aI235
aI131
aI469
aI1749
aI3681
aI3841
aI10157
aI15183
aI61413
aI42207
aI170359
asS'd'
I20515
sa(dp41030
S'a'
I119373
sS's'
I18
sS'm'
(lp41031
I1
aI1
aI7
aI3
aI13
aI9
aI7
aI191
aI239
aI417
aI817
aI1381
aI1179
aI2719
aI21025
aI17429
aI50295
aI196485
asS'd'
I20516
sa(dp41032
S'a'
I119382
sS's'
I18
sS'm'
(lp41033
I1
aI1
aI7
aI9
aI9
aI31
aI123
aI229
aI381
aI569
aI513
aI1617
aI6141
aI9717
aI31769
aI30159
aI113697
aI254237
asS'd'
I20517
sa(dp41034
S'a'
I119410
sS's'
I18
sS'm'
(lp41035
I1
aI1
aI1
aI15
aI31
aI43
aI37
aI17
aI283
aI905
aI297
aI1317
aI1883
aI11313
aI5653
aI55655
aI121029
aI149831
asS'd'
I20518
sa(dp41036
S'a'
I119425
sS's'
I18
sS'm'
(lp41037
I1
aI1
aI1
aI7
aI7
aI47
aI83
aI101
aI497
aI465
aI1133
aI3877
aI5371
aI3355
aI17161
aI50185
aI120837
aI255103
asS'd'
I20519
sa(dp41038
S'a'
I119459
sS's'
I18
sS'm'
(lp41039
I1
aI1
aI3
aI5
aI5
aI19
aI3
aI251
aI433
aI303
aI1193
aI1263
aI2139
aI473
aI10725
aI57725
aI111411
aI133687
asS'd'
I20520
sa(dp41040
S'a'
I119473
sS's'
I18
sS'm'
(lp41041
I1
aI1
aI1
aI7
aI3
aI1
aI99
aI115
aI481
aI395
aI115
aI1699
aI953
aI2807
aI7227
aI52781
aI2855
aI161159
asS'd'
I20521
sa(dp41042
S'a'
I119486
sS's'
I18
sS'm'
(lp41043
I1
aI1
aI5
aI9
aI25
aI25
aI85
aI3
aI451
aI847
aI837
aI3669
aI4717
aI3661
aI29111
aI43735
aI49445
aI100379
asS'd'
I20522
sa(dp41044
S'a'
I119500
sS's'
I18
sS'm'
(lp41045
I1
aI1
aI7
aI9
aI19
aI61
aI67
aI123
aI195
aI483
aI1741
aI2719
aI7809
aI5035
aI30689
aI21325
aI56191
aI46127
asS'd'
I20523
sa(dp41046
S'a'
I119508
sS's'
I18
sS'm'
(lp41047
I1
aI1
aI3
aI11
aI15
aI39
aI101
aI27
aI103
aI807
aI1557
aI1647
aI1285
aI16169
aI20203
aI57153
aI60749
aI71361
asS'd'
I20524
sa(dp41048
S'a'
I119511
sS's'
I18
sS'm'
(lp41049
I1
aI3
aI3
aI3
aI19
aI1
aI93
aI31
aI105
aI925
aI689
aI3061
aI7451
aI12667
aI27179
aI36295
aI61011
aI90321
asS'd'
I20525
sa(dp41050
S'a'
I119517
sS's'
I18
sS'm'
(lp41051
I1
aI3
aI7
aI15
aI19
aI35
aI47
aI241
aI261
aI935
aI1033
aI751
aI6519
aI6911
aI13519
aI2539
aI40285
aI81535
asS'd'
I20526
sa(dp41052
S'a'
I119528
sS's'
I18
sS'm'
(lp41053
I1
aI1
aI7
aI13
aI17
aI31
aI71
aI135
aI167
aI5
aI673
aI2909
aI4377
aI3453
aI31289
aI38081
aI21993
aI192933
asS'd'
I20527
sa(dp41054
S'a'
I119533
sS's'
I18
sS'm'
(lp41055
I1
aI1
aI5
aI15
aI23
aI49
aI85
aI127
aI13
aI849
aI1661
aI2099
aI3479
aI3613
aI21723
aI58147
aI56321
aI203171
asS'd'
I20528
sa(dp41056
S'a'
I119539
sS's'
I18
sS'm'
(lp41057
I1
aI1
aI7
aI15
aI23
aI63
aI3
aI207
aI445
aI573
aI1419
aI1161
aI2237
aI1251
aI23387
aI65259
aI81447
aI74555
asS'd'
I20529
sa(dp41058
S'a'
I119542
sS's'
I18
sS'm'
(lp41059
I1
aI3
aI5
aI3
aI25
aI61
aI37
aI25
aI287
aI969
aI37
aI1615
aI7923
aI4457
aI27611
aI8519
aI113957
aI237427
asS'd'
I20530
sa(dp41060
S'a'
I119546
sS's'
I18
sS'm'
(lp41061
I1
aI3
aI5
aI11
aI11
aI25
aI117
aI169
aI149
aI701
aI139
aI2835
aI6029
aI9067
aI841
aI51707
aI9287
aI115825
asS'd'
I20531
sa(dp41062
S'a'
I119566
sS's'
I18
sS'm'
(lp41063
I1
aI1
aI1
aI13
aI31
aI5
aI9
aI5
aI313
aI1023
aI551
aI3635
aI6765
aI13379
aI29135
aI39737
aI80913
aI256355
asS'd'
I20532
sa(dp41064
S'a'
I119587
sS's'
I18
sS'm'
(lp41065
I1
aI3
aI1
aI13
aI9
aI23
aI105
aI117
aI181
aI211
aI755
aI555
aI2763
aI13965
aI14743
aI63725
aI16377
aI203435
asS'd'
I20533
sa(dp41066
S'a'
I119590
sS's'
I18
sS'm'
(lp41067
I1
aI3
aI1
aI15
aI11
aI45
aI111
aI147
aI471
aI321
aI381
aI2921
aI6423
aI629
aI25117
aI51213
aI126941
aI181931
asS'd'
I20534
sa(dp41068
S'a'
I119607
sS's'
I18
sS'm'
(lp41069
I1
aI1
aI1
aI11
aI1
aI49
aI127
aI105
aI315
aI1
aI859
aI1223
aI5967
aI2521
aI14491
aI58399
aI45155
aI192567
asS'd'
I20535
sa(dp41070
S'a'
I119614
sS's'
I18
sS'm'
(lp41071
I1
aI1
aI3
aI9
aI19
aI21
aI73
aI93
aI95
aI307
aI293
aI3243
aI4765
aI2253
aI16775
aI29861
aI3785
aI90357
asS'd'
I20536
sa(dp41072
S'a'
I119625
sS's'
I18
sS'm'
(lp41073
I1
aI1
aI3
aI7
aI3
aI53
aI33
aI167
aI165
aI509
aI1133
aI169
aI6951
aI7715
aI26317
aI5249
aI86235
aI39649
asS'd'
I20537
sa(dp41074
S'a'
I119633
sS's'
I18
sS'm'
(lp41075
I1
aI1
aI5
aI5
aI3
aI47
aI105
aI89
aI201
aI1003
aI877
aI635
aI2225
aI6391
aI21247
aI5707
aI1233
aI87055
asS'd'
I20538
sa(dp41076
S'a'
I119634
sS's'
I18
sS'm'
(lp41077
I1
aI1
aI1
aI1
aI9
aI39
aI61
aI201
aI435
aI843
aI1245
aI533
aI1757
aI1117
aI19687
aI54817
aI32495
aI228865
asS'd'
I20539
sa(dp41078
S'a'
I119643
sS's'
I18
sS'm'
(lp41079
I1
aI3
aI1
aI1
aI3
aI61
aI63
aI117
aI143
aI217
aI435
aI1977
aI2647
aI7631
aI12969
aI50211
aI26483
aI256329
asS'd'
I20540
sa(dp41080
S'a'
I119683
sS's'
I18
sS'm'
(lp41081
I1
aI3
aI3
aI9
aI21
aI25
aI113
aI243
aI457
aI143
aI833
aI1505
aI3071
aI1845
aI17867
aI58205
aI103819
aI185215
asS'd'
I20541
sa(dp41082
S'a'
I119704
sS's'
I18
sS'm'
(lp41083
I1
aI1
aI5
aI7
aI23
aI51
aI11
aI117
aI195
aI535
aI685
aI31
aI3037
aI9719
aI22811
aI42959
aI21021
aI126297
asS'd'
I20542
sa(dp41084
S'a'
I119710
sS's'
I18
sS'm'
(lp41085
I1
aI3
aI3
aI1
aI1
aI17
aI107
aI245
aI257
aI547
aI887
aI45
aI5243
aI2439
aI22191
aI19503
aI2143
aI75187
asS'd'
I20543
sa(dp41086
S'a'
I119734
sS's'
I18
sS'm'
(lp41087
I1
aI3
aI7
aI9
aI11
aI27
aI123
aI197
aI353
aI151
aI1115
aI403
aI1105
aI7425
aI7463
aI42065
aI116187
aI154537
asS'd'
I20544
sa(dp41088
S'a'
I119752
sS's'
I18
sS'm'
(lp41089
I1
aI3
aI5
aI15
aI23
aI5
aI49
aI177
aI223
aI615
aI1255
aI2081
aI321
aI8733
aI19549
aI53027
aI275
aI62739
asS'd'
I20545
sa(dp41090
S'a'
I119760
sS's'
I18
sS'm'
(lp41091
I1
aI3
aI7
aI1
aI17
aI61
aI3
aI233
aI249
aI763
aI369
aI555
aI1621
aI7221
aI22575
aI13295
aI99793
aI233635
asS'd'
I20546
sa(dp41092
S'a'
I119826
sS's'
I18
sS'm'
(lp41093
I1
aI1
aI3
aI1
aI27
aI41
aI125
aI113
aI47
aI583
aI543
aI453
aI1213
aI14187
aI1645
aI35761
aI110051
aI197081
asS'd'
I20547
sa(dp41094
S'a'
I119922
sS's'
I18
sS'm'
(lp41095
I1
aI1
aI7
aI9
aI29
aI43
aI105
aI125
aI489
aI135
aI153
aI2279
aI4079
aI6731
aI12055
aI60181
aI82563
aI173991
asS'd'
I20548
sa(dp41096
S'a'
I119931
sS's'
I18
sS'm'
(lp41097
I1
aI3
aI7
aI3
aI29
aI1
aI67
aI151
aI127
aI625
aI113
aI2127
aI6723
aI12359
aI28609
aI60605
aI20375
aI120129
asS'd'
I20549
sa(dp41098
S'a'
I119977
sS's'
I18
sS'm'
(lp41099
I1
aI1
aI5
aI7
aI9
aI13
aI27
aI171
aI129
aI199
aI303
aI4045
aI2047
aI8887
aI22233
aI57571
aI40545
aI36479
asS'd'
I20550
sa(dp41100
S'a'
I120015
sS's'
I18
sS'm'
(lp41101
I1
aI3
aI1
aI7
aI1
aI5
aI127
aI203
aI213
aI691
aI1155
aI27
aI5409
aI13519
aI6747
aI42371
aI37089
aI145855
asS'd'
I20551
sa(dp41102
S'a'
I120020
sS's'
I18
sS'm'
(lp41103
I1
aI3
aI3
aI9
aI3
aI33
aI119
aI25
aI337
aI715
aI1093
aI987
aI7157
aI14975
aI28595
aI19021
aI1243
aI148707
asS'd'
I20552
sa(dp41104
S'a'
I120083
sS's'
I18
sS'm'
(lp41105
I1
aI3
aI1
aI13
aI3
aI41
aI81
aI151
aI23
aI787
aI181
aI2357
aI5077
aI1997
aI6451
aI25505
aI44875
aI198341
asS'd'
I20553
sa(dp41106
S'a'
I120086
sS's'
I18
sS'm'
(lp41107
I1
aI1
aI7
aI3
aI5
aI15
aI9
aI169
aI337
aI487
aI1325
aI1505
aI465
aI2339
aI20747
aI8269
aI96875
aI108985
asS'd'
I20554
sa(dp41108
S'a'
I120101
sS's'
I18
sS'm'
(lp41109
I1
aI3
aI1
aI7
aI23
aI7
aI113
aI181
aI25
aI989
aI1649
aI1
aI823
aI6793
aI18729
aI3599
aI97951
aI239609
asS'd'
I20555
sa(dp41110
S'a'
I120105
sS's'
I18
sS'm'
(lp41111
I1
aI3
aI7
aI7
aI3
aI63
aI63
aI207
aI419
aI355
aI1133
aI2979
aI2071
aI11699
aI32565
aI61347
aI106475
aI16893
asS'd'
I20556
sa(dp41112
S'a'
I120155
sS's'
I18
sS'm'
(lp41113
I1
aI1
aI7
aI5
aI7
aI3
aI119
aI133
aI189
aI341
aI1571
aI1559
aI4309
aI16203
aI22459
aI21019
aI80375
aI3453
asS'd'
I20557
sa(dp41114
S'a'
I120188
sS's'
I18
sS'm'
(lp41115
I1
aI1
aI3
aI11
aI21
aI21
aI45
aI69
aI485
aI21
aI727
aI703
aI5209
aI14745
aI3437
aI54603
aI104357
aI151207
asS'd'
I20558
sa(dp41116
S'a'
I120195
sS's'
I18
sS'm'
(lp41117
I1
aI3
aI7
aI5
aI25
aI17
aI13
aI147
aI329
aI93
aI121
aI315
aI2779
aI6921
aI425
aI50441
aI1133
aI252291
asS'd'
I20559
sa(dp41118
S'a'
I120216
sS's'
I18
sS'm'
(lp41119
I1
aI1
aI3
aI5
aI5
aI51
aI15
aI135
aI323
aI841
aI409
aI1067
aI3243
aI4207
aI6833
aI59329
aI90545
aI116661
asS'd'
I20560
sa(dp41120
S'a'
I120228
sS's'
I18
sS'm'
(lp41121
I1
aI1
aI3
aI15
aI25
aI7
aI43
aI147
aI153
aI947
aI79
aI1897
aI4519
aI14441
aI27181
aI38517
aI71673
aI158597
asS'd'
I20561
sa(dp41122
S'a'
I120238
sS's'
I18
sS'm'
(lp41123
I1
aI3
aI3
aI3
aI3
aI47
aI109
aI111
aI273
aI271
aI741
aI3999
aI649
aI7367
aI14933
aI11785
aI92709
aI133815
asS'd'
I20562
sa(dp41124
S'a'
I120255
sS's'
I18
sS'm'
(lp41125
I1
aI1
aI3
aI13
aI5
aI41
aI55
aI15
aI409
aI355
aI1255
aI3043
aI7503
aI3523
aI4261
aI48927
aI119901
aI149411
asS'd'
I20563
sa(dp41126
S'a'
I120317
sS's'
I18
sS'm'
(lp41127
I1
aI1
aI5
aI1
aI11
aI57
aI3
aI149
aI409
aI287
aI909
aI3541
aI4243
aI4485
aI3611
aI63213
aI102575
aI49863
asS'd'
I20564
sa(dp41128
S'a'
I120327
sS's'
I18
sS'm'
(lp41129
I1
aI1
aI3
aI7
aI15
aI55
aI119
aI185
aI511
aI301
aI237
aI3701
aI3195
aI1323
aI27511
aI44635
aI45363
aI117683
asS'd'
I20565
sa(dp41130
S'a'
I120369
sS's'
I18
sS'm'
(lp41131
I1
aI3
aI3
aI11
aI3
aI61
aI125
aI15
aI235
aI819
aI467
aI1097
aI1055
aI16343
aI8329
aI37807
aI38663
aI145625
asS'd'
I20566
sa(dp41132
S'a'
I120382
sS's'
I18
sS'm'
(lp41133
I1
aI1
aI7
aI1
aI25
aI49
aI117
aI163
aI413
aI509
aI63
aI1313
aI5113
aI6505
aI25475
aI12059
aI88021
aI168037
asS'd'
I20567
sa(dp41134
S'a'
I120404
sS's'
I18
sS'm'
(lp41135
I1
aI3
aI3
aI9
aI9
aI37
aI15
aI185
aI359
aI231
aI1483
aI2999
aI773
aI10375
aI27103
aI39899
aI100187
aI54485
asS'd'
I20568
sa(dp41136
S'a'
I120414
sS's'
I18
sS'm'
(lp41137
I1
aI3
aI3
aI1
aI13
aI63
aI111
aI79
aI133
aI425
aI1491
aI3735
aI533
aI13417
aI5161
aI11455
aI16907
aI132267
asS'd'
I20569
sa(dp41138
S'a'
I120432
sS's'
I18
sS'm'
(lp41139
I1
aI3
aI1
aI15
aI15
aI47
aI57
aI157
aI453
aI681
aI1811
aI1685
aI4329
aI131
aI22763
aI1017
aI66637
aI1673
asS'd'
I20570
sa(dp41140
S'a'
I120435
sS's'
I18
sS'm'
(lp41141
I1
aI3
aI7
aI3
aI31
aI5
aI105
aI101
aI431
aI983
aI1333
aI845
aI7369
aI15041
aI6527
aI8617
aI61911
aI137513
asS'd'
I20571
sa(dp41142
S'a'
I120454
sS's'
I18
sS'm'
(lp41143
I1
aI3
aI7
aI15
aI3
aI45
aI57
aI23
aI393
aI495
aI769
aI215
aI3611
aI12907
aI20637
aI52997
aI88345
aI37961
asS'd'
I20572
sa(dp41144
S'a'
I120463
sS's'
I18
sS'm'
(lp41145
I1
aI3
aI5
aI7
aI31
aI3
aI85
aI249
aI353
aI559
aI1803
aI959
aI4625
aI9413
aI22339
aI6071
aI124765
aI43973
asS'd'
I20573
sa(dp41146
S'a'
I120471
sS's'
I18
sS'm'
(lp41147
I1
aI3
aI1
aI9
aI31
aI15
aI109
aI197
aI179
aI293
aI457
aI709
aI627
aI7743
aI1997
aI59625
aI36919
aI252849
asS'd'
I20574
sa(dp41148
S'a'
I120505
sS's'
I18
sS'm'
(lp41149
I1
aI1
aI7
aI11
aI17
aI21
aI89
aI11
aI431
aI617
aI1029
aI2649
aI1725
aI2723
aI18367
aI46103
aI108063
aI221855
asS'd'
I20575
sa(dp41150
S'a'
I120528
sS's'
I18
sS'm'
(lp41151
I1
aI3
aI7
aI1
aI29
aI7
aI81
aI187
aI353
aI807
aI965
aI3655
aI3255
aI9139
aI28619
aI32127
aI107901
aI139099
asS'd'
I20576
sa(dp41152
S'a'
I120588
sS's'
I18
sS'm'
(lp41153
I1
aI1
aI5
aI3
aI17
aI9
aI39
aI59
aI431
aI829
aI525
aI1885
aI2387
aI13381
aI5271
aI29739
aI80413
aI240595
asS'd'
I20577
sa(dp41154
S'a'
I120616
sS's'
I18
sS'm'
(lp41155
I1
aI3
aI1
aI5
aI1
aI5
aI25
aI9
aI193
aI971
aI681
aI2921
aI3271
aI3891
aI20123
aI37477
aI33141
aI82481
asS'd'
I20578
sa(dp41156
S'a'
I120630
sS's'
I18
sS'm'
(lp41157
I1
aI1
aI3
aI5
aI1
aI63
aI77
aI93
aI63
aI399
aI1945
aI387
aI5457
aI7339
aI13279
aI34119
aI129903
aI12621
asS'd'
I20579
sa(dp41158
S'a'
I120644
sS's'
I18
sS'm'
(lp41159
I1
aI3
aI3
aI11
aI17
aI11
aI117
aI189
aI473
aI23
aI13
aI1261
aI2153
aI12181
aI21075
aI33179
aI43355
aI168293
asS'd'
I20580
sa(dp41160
S'a'
I120672
sS's'
I18
sS'm'
(lp41161
I1
aI3
aI5
aI11
aI21
aI13
aI73
aI227
aI211
aI939
aI885
aI2091
aI1123
aI14809
aI15705
aI56675
aI58087
aI1451
asS'd'
I20581
sa(dp41162
S'a'
I120677
sS's'
I18
sS'm'
(lp41163
I1
aI3
aI7
aI1
aI15
aI15
aI101
aI235
aI287
aI675
aI1741
aI3885
aI6211
aI14817
aI10235
aI29289
aI60401
aI27639
asS'd'
I20582
sa(dp41164
S'a'
I120718
sS's'
I18
sS'm'
(lp41165
I1
aI3
aI5
aI13
aI13
aI7
aI117
aI31
aI143
aI505
aI1823
aI2841
aI2133
aI7305
aI14093
aI14229
aI85179
aI136793
asS'd'
I20583
sa(dp41166
S'a'
I120725
sS's'
I18
sS'm'
(lp41167
I1
aI1
aI3
aI15
aI27
aI15
aI67
aI237
aI315
aI793
aI207
aI3781
aI5201
aI13191
aI9601
aI44041
aI116097
aI178543
asS'd'
I20584
sa(dp41168
S'a'
I120730
sS's'
I18
sS'm'
(lp41169
I1
aI3
aI1
aI9
aI9
aI47
aI53
aI225
aI109
aI831
aI1757
aI349
aI6353
aI15417
aI16395
aI36295
aI10901
aI122349
asS'd'
I20585
sa(dp41170
S'a'
I120756
sS's'
I18
sS'm'
(lp41171
I1
aI1
aI1
aI1
aI1
aI17
aI109
aI13
aI123
aI537
aI1859
aI3717
aI4441
aI4271
aI29017
aI13601
aI18533
aI216695
asS'd'
I20586
sa(dp41172
S'a'
I120759
sS's'
I18
sS'm'
(lp41173
I1
aI3
aI3
aI13
aI31
aI23
aI45
aI233
aI29
aI295
aI761
aI757
aI777
aI3499
aI26715
aI24153
aI113777
aI256337
asS'd'
I20587
sa(dp41174
S'a'
I120765
sS's'
I18
sS'm'
(lp41175
I1
aI3
aI5
aI7
aI23
aI57
aI37
aI179
aI511
aI897
aI2031
aI1285
aI3957
aI15085
aI19993
aI28819
aI39959
aI187445
asS'd'
I20588
sa(dp41176
S'a'
I120795
sS's'
I18
sS'm'
(lp41177
I1
aI3
aI1
aI15
aI25
aI41
aI121
aI227
aI433
aI859
aI357
aI3107
aI3241
aI879
aI10763
aI59473
aI73145
aI258493
asS'd'
I20589
sa(dp41178
S'a'
I120821
sS's'
I18
sS'm'
(lp41179
I1
aI3
aI3
aI5
aI13
aI9
aI85
aI117
aI347
aI161
aI995
aI549
aI5443
aI9057
aI28931
aI57549
aI27523
aI54717
asS'd'
I20590
sa(dp41180
S'a'
I120850
sS's'
I18
sS'm'
(lp41181
I1
aI1
aI5
aI13
aI13
aI33
aI97
aI37
aI55
aI367
aI403
aI2361
aI5717
aI4433
aI26921
aI14227
aI69445
aI100337
asS'd'
I20591
sa(dp41182
S'a'
I120868
sS's'
I18
sS'm'
(lp41183
I1
aI3
aI5
aI7
aI1
aI41
aI89
aI165
aI163
aI1
aI685
aI3577
aI1079
aI1057
aI125
aI35853
aI8387
aI113035
asS'd'
I20592
sa(dp41184
S'a'
I120880
sS's'
I18
sS'm'
(lp41185
I1
aI3
aI3
aI13
aI3
aI59
aI119
aI51
aI325
aI205
aI821
aI1417
aI2097
aI13725
aI31785
aI53803
aI15737
aI2013
asS'd'
I20593
sa(dp41186
S'a'
I120885
sS's'
I18
sS'm'
(lp41187
I1
aI3
aI7
aI7
aI1
aI45
aI95
aI221
aI249
aI65
aI1479
aI2163
aI5761
aI15321
aI8013
aI25771
aI110897
aI214127
asS'd'
I20594
sa(dp41188
S'a'
I120892
sS's'
I18
sS'm'
(lp41189
I1
aI3
aI3
aI5
aI25
aI11
aI27
aI29
aI95
aI955
aI1989
aI3775
aI609
aI7073
aI4571
aI38857
aI92205
aI156209
asS'd'
I20595
sa(dp41190
S'a'
I120897
sS's'
I18
sS'm'
(lp41191
I1
aI3
aI1
aI13
aI29
aI57
aI97
aI47
aI499
aI641
aI587
aI2125
aI2257
aI13911
aI13993
aI1715
aI46233
aI181279
asS'd'
I20596
sa(dp41192
S'a'
I120912
sS's'
I18
sS'm'
(lp41193
I1
aI1
aI7
aI13
aI7
aI55
aI69
aI235
aI379
aI269
aI1969
aI2733
aI3677
aI1707
aI26999
aI64041
aI98111
aI138691
asS'd'
I20597
sa(dp41194
S'a'
I120934
sS's'
I18
sS'm'
(lp41195
I1
aI1
aI3
aI5
aI27
aI63
aI65
aI251
aI413
aI903
aI1307
aI2941
aI5649
aI11271
aI1935
aI49389
aI37995
aI218197
asS'd'
I20598
sa(dp41196
S'a'
I120991
sS's'
I18
sS'm'
(lp41197
I1
aI3
aI3
aI13
aI7
aI1
aI45
aI225
aI13
aI169
aI581
aI1657
aI117
aI10251
aI23435
aI40379
aI127085
aI88185
asS'd'
I20599
sa(dp41198
S'a'
I121004
sS's'
I18
sS'm'
(lp41199
I1
aI1
aI7
aI5
aI21
aI63
aI69
aI37
aI459
aI115
aI1403
aI1939
aI6437
aI13149
aI3597
aI50115
aI129075
aI260613
asS'd'
I20600
sa(dp41200
S'a'
I121021
sS's'
I18
sS'm'
(lp41201
I1
aI1
aI3
aI5
aI29
aI57
aI15
aI223
aI131
aI907
aI1561
aI1103
aI4355
aI2763
aI6359
aI55401
aI1751
aI53143
asS'd'
I20601
sa(dp41202
S'a'
I121022
sS's'
I18
sS'm'
(lp41203
I1
aI1
aI7
aI11
aI15
aI5
aI17
aI159
aI505
aI407
aI1873
aI2501
aI2203
aI12559
aI5123
aI53281
aI29307
aI16215
asS'd'
I20602
sa(dp41204
S'a'
I121036
sS's'
I18
sS'm'
(lp41205
I1
aI1
aI1
aI15
aI9
aI41
aI77
aI145
aI325
aI529
aI1939
aI3835
aI3109
aI12215
aI18323
aI2551
aI89793
aI94745
asS'd'
I20603
sa(dp41206
S'a'
I121041
sS's'
I18
sS'm'
(lp41207
I1
aI3
aI1
aI9
aI21
aI5
aI7
aI57
aI99
aI1005
aI1211
aI4063
aI6851
aI7653
aI29283
aI15463
aI121289
aI187055
asS'd'
I20604
sa(dp41208
S'a'
I121069
sS's'
I18
sS'm'
(lp41209
I1
aI1
aI3
aI3
aI13
aI11
aI17
aI125
aI173
aI283
aI1419
aI2533
aI6875
aI16031
aI26633
aI51027
aI27343
aI74257
asS'd'
I20605
sa(dp41210
S'a'
I121078
sS's'
I18
sS'm'
(lp41211
I1
aI1
aI3
aI11
aI29
aI49
aI87
aI19
aI367
aI941
aI983
aI1041
aI8099
aI8735
aI30123
aI62665
aI87051
aI98745
asS'd'
I20606
sa(dp41212
S'a'
I121126
sS's'
I18
sS'm'
(lp41213
I1
aI1
aI1
aI15
aI25
aI57
aI97
aI177
aI467
aI181
aI923
aI3833
aI5405
aI14335
aI23495
aI48323
aI70331
aI136825
asS'd'
I20607
sa(dp41214
S'a'
I121135
sS's'
I18
sS'm'
(lp41215
I1
aI1
aI1
aI13
aI21
aI17
aI29
aI237
aI305
aI117
aI1077
aI2999
aI1879
aI6875
aI19321
aI10999
aI130513
aI160883
asS'd'
I20608
sa(dp41216
S'a'
I121138
sS's'
I18
sS'm'
(lp41217
I1
aI3
aI7
aI15
aI19
aI47
aI89
aI153
aI287
aI7
aI1429
aI1507
aI2853
aI4197
aI11195
aI33891
aI59063
aI189601
asS'd'
I20609
sa(dp41218
S'a'
I121162
sS's'
I18
sS'm'
(lp41219
I1
aI1
aI3
aI15
aI31
aI5
aI113
aI71
aI13
aI925
aI147
aI451
aI5701
aI13671
aI13943
aI13799
aI59627
aI115715
asS'd'
I20610
sa(dp41220
S'a'
I121176
sS's'
I18
sS'm'
(lp41221
I1
aI1
aI1
aI5
aI13
aI25
aI51
aI11
aI409
aI393
aI1479
aI2583
aI3101
aI303
aI17609
aI10653
aI69107
aI150459
asS'd'
I20611
sa(dp41222
S'a'
I121186
sS's'
I18
sS'm'
(lp41223
I1
aI1
aI5
aI7
aI21
aI27
aI107
aI77
aI61
aI467
aI1723
aI1247
aI287
aI1039
aI25347
aI18111
aI24837
aI42903
asS'd'
I20612
sa(dp41224
S'a'
I121209
sS's'
I18
sS'm'
(lp41225
I1
aI3
aI5
aI5
aI1
aI29
aI57
aI177
aI311
aI9
aI819
aI3235
aI3887
aI9679
aI11849
aI31755
aI68467
aI135587
asS'd'
I20613
sa(dp41226
S'a'
I121212
sS's'
I18
sS'm'
(lp41227
I1
aI1
aI5
aI5
aI29
aI15
aI13
aI69
aI163
aI709
aI1405
aI3835
aI777
aI7567
aI10153
aI5043
aI129465
aI59113
asS'd'
I20614
sa(dp41228
S'a'
I121221
sS's'
I18
sS'm'
(lp41229
I1
aI1
aI1
aI13
aI3
aI47
aI97
aI101
aI179
aI53
aI1919
aI17
aI3597
aI11769
aI17971
aI39257
aI76167
aI255653
asS'd'
I20615
sa(dp41230
S'a'
I121225
sS's'
I18
sS'm'
(lp41231
I1
aI1
aI7
aI5
aI1
aI45
aI121
aI223
aI299
aI271
aI1857
aI1955
aI5509
aI1245
aI9519
aI60547
aI78497
aI191251
asS'd'
I20616
sa(dp41232
S'a'
I121228
sS's'
I18
sS'm'
(lp41233
I1
aI1
aI1
aI11
aI5
aI57
aI87
aI113
aI59
aI547
aI1591
aI1905
aI475
aI4687
aI27591
aI43807
aI9617
aI70769
asS'd'
I20617
sa(dp41234
S'a'
I121239
sS's'
I18
sS'm'
(lp41235
I1
aI1
aI7
aI1
aI1
aI35
aI35
aI11
aI107
aI401
aI827
aI2271
aI2131
aI12751
aI25771
aI51311
aI46897
aI198655
asS'd'
I20618
sa(dp41236
S'a'
I121245
sS's'
I18
sS'm'
(lp41237
I1
aI3
aI3
aI3
aI15
aI29
aI91
aI147
aI9
aI611
aI1739
aI3211
aI3883
aI14205
aI24073
aI37445
aI54451
aI208123
asS'd'
I20619
sa(dp41238
S'a'
I121264
sS's'
I18
sS'm'
(lp41239
I1
aI3
aI7
aI1
aI25
aI53
aI33
aI161
aI365
aI651
aI1263
aI1623
aI1767
aI3789
aI18013
aI52133
aI60119
aI100859
asS'd'
I20620
sa(dp41240
S'a'
I121270
sS's'
I18
sS'm'
(lp41241
I1
aI3
aI5
aI9
aI15
aI9
aI97
aI67
aI379
aI981
aI1323
aI1095
aI3701
aI3257
aI13647
aI12511
aI53375
aI156689
asS'd'
I20621
sa(dp41242
S'a'
I121299
sS's'
I18
sS'm'
(lp41243
I1
aI1
aI3
aI5
aI9
aI23
aI97
aI239
aI171
aI501
aI1549
aI1029
aI8019
aI5023
aI9439
aI14223
aI54433
aI152855
asS'd'
I20622
sa(dp41244
S'a'
I121306
sS's'
I18
sS'm'
(lp41245
I1
aI1
aI7
aI13
aI17
aI35
aI111
aI169
aI81
aI599
aI1673
aI3461
aI7905
aI7925
aI16311
aI20327
aI57109
aI158719
asS'd'
I20623
sa(dp41246
S'a'
I121315
sS's'
I18
sS'm'
(lp41247
I1
aI1
aI3
aI15
aI7
aI33
aI23
aI33
aI273
aI987
aI1489
aI363
aI4017
aI8919
aI28839
aI10143
aI114179
aI218155
asS'd'
I20624
sa(dp41248
S'a'
I121317
sS's'
I18
sS'm'
(lp41249
I1
aI3
aI5
aI13
aI29
aI5
aI61
aI133
aI65
aI933
aI509
aI551
aI7365
aI3703
aI15003
aI27849
aI64211
aI140383
asS'd'
I20625
sa(dp41250
S'a'
I121321
sS's'
I18
sS'm'
(lp41251
I1
aI1
aI1
aI13
aI7
aI25
aI49
aI1
aI269
aI601
aI251
aI33
aI2443
aI13725
aI5805
aI63347
aI109489
aI111491
asS'd'
I20626
sa(dp41252
S'a'
I121324
sS's'
I18
sS'm'
(lp41253
I1
aI1
aI7
aI9
aI31
aI23
aI117
aI71
aI371
aI733
aI605
aI4019
aI4577
aI3887
aI31061
aI24939
aI57905
aI148331
asS'd'
I20627
sa(dp41254
S'a'
I121330
sS's'
I18
sS'm'
(lp41255
I1
aI3
aI5
aI1
aI3
aI35
aI35
aI227
aI355
aI27
aI1673
aI2173
aI5001
aI14613
aI6343
aI40775
aI72349
aI101287
asS'd'
I20628
sa(dp41256
S'a'
I121339
sS's'
I18
sS'm'
(lp41257
I1
aI1
aI7
aI15
aI29
aI3
aI43
aI77
aI43
aI51
aI1495
aI2577
aI2093
aI7515
aI20151
aI44533
aI32223
aI6355
asS'd'
I20629
sa(dp41258
S'a'
I121346
sS's'
I18
sS'm'
(lp41259
I1
aI1
aI5
aI5
aI11
aI47
aI91
aI221
aI35
aI969
aI343
aI3287
aI857
aI4851
aI12599
aI939
aI53615
aI262125
asS'd'
I20630
sa(dp41260
S'a'
I121370
sS's'
I18
sS'm'
(lp41261
I1
aI1
aI5
aI7
aI29
aI11
aI67
aI155
aI317
aI629
aI211
aI583
aI1061
aI13243
aI13999
aI45405
aI18187
aI99021
asS'd'
I20631
sa(dp41262
S'a'
I121372
sS's'
I18
sS'm'
(lp41263
I1
aI3
aI7
aI13
aI3
aI39
aI5
aI207
aI175
aI515
aI1181
aI739
aI379
aI9919
aI12079
aI18903
aI62475
aI239383
asS'd'
I20632
sa(dp41264
S'a'
I121393
sS's'
I18
sS'm'
(lp41265
I1
aI1
aI1
aI15
aI25
aI15
aI113
aI215
aI281
aI861
aI1055
aI2577
aI5545
aI12365
aI16097
aI35775
aI8331
aI119353
asS'd'
I20633
sa(dp41266
S'a'
I121396
sS's'
I18
sS'm'
(lp41267
I1
aI1
aI1
aI1
aI25
aI55
aI111
aI185
aI485
aI361
aI155
aI4077
aI5517
aI16057
aI19069
aI40129
aI38959
aI211233
asS'd'
I20634
sa(dp41268
S'a'
I121414
sS's'
I18
sS'm'
(lp41269
I1
aI1
aI3
aI3
aI13
aI1
aI93
aI129
aI243
aI813
aI115
aI177
aI53
aI8251
aI32351
aI63847
aI54537
aI25527
asS'd'
I20635
sa(dp41270
S'a'
I121442
sS's'
I18
sS'm'
(lp41271
I1
aI3
aI5
aI11
aI25
aI47
aI113
aI69
aI285
aI451
aI2011
aI81
aI6535
aI3409
aI8647
aI56575
aI975
aI149571
asS'd'
I20636
sa(dp41272
S'a'
I121475
sS's'
I18
sS'm'
(lp41273
I1
aI1
aI7
aI7
aI19
aI1
aI75
aI123
aI413
aI697
aI41
aI3179
aI4075
aI15967
aI2477
aI17549
aI54193
aI258657
asS'd'
I20637
sa(dp41274
S'a'
I121499
sS's'
I18
sS'm'
(lp41275
I1
aI1
aI5
aI5
aI11
aI23
aI19
aI253
aI303
aI255
aI901
aI875
aI1517
aI6953
aI25189
aI26763
aI28843
aI167705
asS'd'
I20638
sa(dp41276
S'a'
I121512
sS's'
I18
sS'm'
(lp41277
I1
aI3
aI7
aI7
aI17
aI45
aI31
aI79
aI279
aI965
aI1869
aI1201
aI1627
aI14035
aI11651
aI45021
aI76171
aI49137
asS'd'
I20639
sa(dp41278
S'a'
I121518
sS's'
I18
sS'm'
(lp41279
I1
aI3
aI1
aI15
aI9
aI55
aI83
aI59
aI437
aI915
aI1667
aI89
aI2797
aI1841
aI29261
aI23497
aI55785
aI102265
asS'd'
I20640
sa(dp41280
S'a'
I121588
sS's'
I18
sS'm'
(lp41281
I1
aI1
aI5
aI5
aI3
aI59
aI17
aI131
aI199
aI541
aI1647
aI2175
aI4449
aI6081
aI10609
aI39467
aI72945
aI32423
asS'd'
I20641
sa(dp41282
S'a'
I121597
sS's'
I18
sS'm'
(lp41283
I1
aI3
aI1
aI7
aI5
aI7
aI85
aI11
aI255
aI397
aI87
aI1661
aI6523
aI5699
aI29407
aI28015
aI50783
aI246625
asS'd'
I20642
sa(dp41284
S'a'
I121612
sS's'
I18
sS'm'
(lp41285
I1
aI3
aI3
aI13
aI5
aI61
aI123
aI147
aI295
aI37
aI301
aI2549
aI7615
aI5725
aI32477
aI18121
aI69353
aI242579
asS'd'
I20643
sa(dp41286
S'a'
I121629
sS's'
I18
sS'm'
(lp41287
I1
aI3
aI5
aI7
aI9
aI45
aI83
aI211
aI475
aI281
aI743
aI3955
aI7811
aI6043
aI30547
aI5315
aI53345
aI25775
asS'd'
I20644
sa(dp41288
S'a'
I121630
sS's'
I18
sS'm'
(lp41289
I1
aI3
aI3
aI5
aI7
aI63
aI125
aI43
aI131
aI353
aI345
aI1689
aI5483
aI5467
aI13445
aI13041
aI68381
aI134567
asS'd'
I20645
sa(dp41290
S'a'
I121634
sS's'
I18
sS'm'
(lp41291
I1
aI3
aI5
aI11
aI31
aI9
aI123
aI53
aI237
aI911
aI349
aI3737
aI1867
aI7375
aI3031
aI4191
aI8697
aI182255
asS'd'
I20646
sa(dp41292
S'a'
I121680
sS's'
I18
sS'm'
(lp41293
I1
aI1
aI3
aI3
aI11
aI11
aI89
aI251
aI69
aI93
aI1241
aI1719
aI2227
aI1793
aI21683
aI58099
aI110831
aI24835
asS'd'
I20647
sa(dp41294
S'a'
I121725
sS's'
I18
sS'm'
(lp41295
I1
aI1
aI5
aI9
aI3
aI57
aI3
aI17
aI11
aI217
aI923
aI3623
aI727
aI1837
aI21203
aI63007
aI33691
aI216259
asS'd'
I20648
sa(dp41296
S'a'
I121798
sS's'
I18
sS'm'
(lp41297
I1
aI3
aI3
aI9
aI21
aI25
aI83
aI115
aI325
aI921
aI811
aI303
aI3555
aI10669
aI5837
aI45585
aI61923
aI159061
asS'd'
I20649
sa(dp41298
S'a'
I121815
sS's'
I18
sS'm'
(lp41299
I1
aI3
aI3
aI15
aI17
aI29
aI77
aI75
aI509
aI363
aI199
aI317
aI7375
aI11971
aI15679
aI17135
aI101925
aI103375
asS'd'
I20650
sa(dp41300
S'a'
I121822
sS's'
I18
sS'm'
(lp41301
I1
aI3
aI5
aI5
aI29
aI45
aI85
aI151
aI73
aI329
aI911
aI3055
aI2381
aI4717
aI5133
aI58987
aI59885
aI226689
asS'd'
I20651
sa(dp41302
S'a'
I121837
sS's'
I18
sS'm'
(lp41303
I1
aI3
aI5
aI1
aI11
aI59
aI17
aI83
aI385
aI867
aI215
aI2275
aI7247
aI10613
aI6493
aI63843
aI74483
aI134271
asS'd'
I20652
sa(dp41304
S'a'
I121840
sS's'
I18
sS'm'
(lp41305
I1
aI1
aI7
aI13
aI29
aI25
aI77
aI61
aI281
aI439
aI353
aI2213
aI697
aI14741
aI1597
aI7515
aI7703
aI149123
asS'd'
I20653
sa(dp41306
S'a'
I121869
sS's'
I18
sS'm'
(lp41307
I1
aI3
aI7
aI7
aI7
aI53
aI77
aI21
aI483
aI793
aI969
aI123
aI3581
aI12489
aI22943
aI54573
aI25785
aI178419
asS'd'
I20654
sa(dp41308
S'a'
I121878
sS's'
I18
sS'm'
(lp41309
I1
aI3
aI5
aI3
aI17
aI59
aI75
aI55
aI125
aI569
aI1625
aI77
aI4593
aI8493
aI5259
aI54537
aI100479
aI107509
asS'd'
I20655
sa(dp41310
S'a'
I121897
sS's'
I18
sS'm'
(lp41311
I1
aI1
aI1
aI15
aI5
aI7
aI11
aI169
aI349
aI133
aI1113
aI2877
aI6109
aI16275
aI9755
aI1385
aI55005
aI36095
asS'd'
I20656
sa(dp41312
S'a'
I121920
sS's'
I18
sS'm'
(lp41313
I1
aI1
aI7
aI7
aI21
aI25
aI41
aI161
aI11
aI321
aI343
aI705
aI4601
aI12867
aI21997
aI25283
aI78467
aI159089
asS'd'
I20657
sa(dp41314
S'a'
I121929
sS's'
I18
sS'm'
(lp41315
I1
aI3
aI7
aI1
aI15
aI3
aI71
aI227
aI427
aI883
aI1021
aI1405
aI7791
aI12669
aI9159
aI30931
aI105993
aI40917
asS'd'
I20658
sa(dp41316
S'a'
I121944
sS's'
I18
sS'm'
(lp41317
I1
aI1
aI5
aI5
aI25
aI41
aI93
aI57
aI125
aI915
aI701
aI2589
aI7147
aI15369
aI28307
aI54635
aI13253
aI97177
asS'd'
I20659
sa(dp41318
S'a'
I121999
sS's'
I18
sS'm'
(lp41319
I1
aI3
aI1
aI13
aI11
aI63
aI103
aI241
aI317
aI927
aI965
aI3179
aI5213
aI13849
aI11509
aI52665
aI1637
aI235647
asS'd'
I20660
sa(dp41320
S'a'
I122008
sS's'
I18
sS'm'
(lp41321
I1
aI3
aI5
aI3
aI31
aI31
aI111
aI9
aI339
aI1017
aI1715
aI101
aI6849
aI14329
aI31607
aI40741
aI73067
aI119001
asS'd'
I20661
sa(dp41322
S'a'
I122056
sS's'
I18
sS'm'
(lp41323
I1
aI3
aI1
aI13
aI3
aI55
aI53
aI15
aI185
aI717
aI1447
aI3029
aI4899
aI14217
aI19949
aI32817
aI24829
aI206829
asS'd'
I20662
sa(dp41324
S'a'
I122062
sS's'
I18
sS'm'
(lp41325
I1
aI3
aI7
aI9
aI21
aI11
aI33
aI213
aI5
aI769
aI1807
aI2179
aI63
aI5167
aI23235
aI25495
aI113299
aI129419
asS'd'
I20663
sa(dp41326
S'a'
I122067
sS's'
I18
sS'm'
(lp41327
I1
aI3
aI7
aI11
aI11
aI33
aI23
aI125
aI21
aI609
aI595
aI1329
aI6175
aI15837
aI3889
aI57797
aI81453
aI211413
asS'd'
I20664
sa(dp41328
S'a'
I122083
sS's'
I18
sS'm'
(lp41329
I1
aI1
aI5
aI11
aI31
aI7
aI13
aI73
aI143
aI559
aI1541
aI275
aI3349
aI2987
aI21797
aI32921
aI125395
aI247667
asS'd'
I20665
sa(dp41330
S'a'
I122095
sS's'
I18
sS'm'
(lp41331
I1
aI1
aI3
aI11
aI19
aI1
aI23
aI167
aI337
aI75
aI1597
aI3591
aI2705
aI7323
aI5957
aI7317
aI58945
aI44625
asS'd'
I20666
sa(dp41332
S'a'
I122110
sS's'
I18
sS'm'
(lp41333
I1
aI1
aI5
aI11
aI27
aI19
aI63
aI231
aI353
aI645
aI531
aI3861
aI1681
aI6901
aI16217
aI20639
aI70077
aI220233
asS'd'
I20667
sa(dp41334
S'a'
I122118
sS's'
I18
sS'm'
(lp41335
I1
aI1
aI5
aI15
aI31
aI41
aI19
aI253
aI147
aI365
aI509
aI1199
aI6699
aI14633
aI1339
aI48203
aI58707
aI83315
asS'd'
I20668
sa(dp41336
S'a'
I122127
sS's'
I18
sS'm'
(lp41337
I1
aI1
aI5
aI7
aI19
aI47
aI47
aI17
aI267
aI139
aI549
aI803
aI4625
aI6851
aI32141
aI12891
aI43785
aI211361
asS'd'
I20669
sa(dp41338
S'a'
I122189
sS's'
I18
sS'm'
(lp41339
I1
aI3
aI1
aI13
aI19
aI35
aI13
aI45
aI167
aI627
aI1449
aI3041
aI5043
aI9279
aI15889
aI41675
aI25769
aI13835
asS'd'
I20670
sa(dp41340
S'a'
I122207
sS's'
I18
sS'm'
(lp41341
I1
aI3
aI7
aI7
aI19
aI47
aI13
aI117
aI403
aI79
aI1623
aI3741
aI3255
aI2301
aI25
aI2311
aI5237
aI150879
asS'd'
I20671
sa(dp41342
S'a'
I122226
sS's'
I18
sS'm'
(lp41343
I1
aI1
aI1
aI15
aI29
aI43
aI75
aI21
aI237
aI809
aI129
aI2637
aI181
aI15921
aI30709
aI61281
aI82405
aI232885
asS'd'
I20672
sa(dp41344
S'a'
I122253
sS's'
I18
sS'm'
(lp41345
I1
aI1
aI3
aI15
aI15
aI15
aI55
aI217
aI243
aI579
aI945
aI3993
aI1875
aI2425
aI25045
aI36729
aI42935
aI213703
asS'd'
I20673
sa(dp41346
S'a'
I122301
sS's'
I18
sS'm'
(lp41347
I1
aI3
aI3
aI1
aI5
aI59
aI115
aI71
aI483
aI327
aI701
aI2893
aI1815
aI4611
aI3843
aI5893
aI126479
aI167807
asS'd'
I20674
sa(dp41348
S'a'
I122324
sS's'
I18
sS'm'
(lp41349
I1
aI3
aI1
aI7
aI17
aI59
aI115
aI191
aI3
aI615
aI215
aI2121
aI5085
aI15233
aI16661
aI6215
aI31061
aI192847
asS'd'
I20675
sa(dp41350
S'a'
I122337
sS's'
I18
sS'm'
(lp41351
I1
aI3
aI1
aI1
aI21
aI25
aI41
aI195
aI151
aI905
aI1587
aI439
aI3317
aI2275
aI4743
aI33505
aI1185
aI254873
asS'd'
I20676
sa(dp41352
S'a'
I122362
sS's'
I18
sS'm'
(lp41353
I1
aI1
aI3
aI9
aI11
aI49
aI37
aI117
aI419
aI1007
aI789
aI1323
aI345
aI2047
aI20697
aI57063
aI69167
aI219393
asS'd'
I20677
sa(dp41354
S'a'
I122368
sS's'
I18
sS'm'
(lp41355
I1
aI3
aI1
aI1
aI25
aI49
aI81
aI79
aI117
aI1015
aI1777
aI2427
aI527
aI10139
aI16261
aI42587
aI33933
aI19749
asS'd'
I20678
sa(dp41356
S'a'
I122371
sS's'
I18
sS'm'
(lp41357
I1
aI1
aI5
aI15
aI31
aI43
aI19
aI17
aI97
aI591
aI891
aI177
aI7835
aI3979
aI15473
aI2173
aI65555
aI182773
asS'd'
I20679
sa(dp41358
S'a'
I122378
sS's'
I18
sS'm'
(lp41359
I1
aI1
aI3
aI15
aI13
aI53
aI31
aI179
aI247
aI635
aI683
aI423
aI2981
aI14401
aI26385
aI10935
aI68497
aI181703
asS'd'
I20680
sa(dp41360
S'a'
I122404
sS's'
I18
sS'm'
(lp41361
I1
aI3
aI3
aI9
aI25
aI55
aI43
aI29
aI487
aI969
aI989
aI3561
aI6425
aI11619
aI5773
aI56515
aI17461
aI151239
asS'd'
I20681
sa(dp41362
S'a'
I122443
sS's'
I18
sS'm'
(lp41363
I1
aI1
aI5
aI13
aI3
aI31
aI73
aI51
aI213
aI215
aI297
aI783
aI697
aI14197
aI7277
aI60697
aI985
aI189995
asS'd'
I20682
sa(dp41364
S'a'
I122451
sS's'
I18
sS'm'
(lp41365
I1
aI3
aI3
aI5
aI9
aI47
aI69
aI27
aI15
aI407
aI1029
aI2541
aI183
aI4413
aI5143
aI33903
aI49509
aI49007
asS'd'
I20683
sa(dp41366
S'a'
I122457
sS's'
I18
sS'm'
(lp41367
I1
aI1
aI1
aI3
aI5
aI3
aI9
aI165
aI215
aI577
aI1657
aI363
aI737
aI5483
aI5955
aI34533
aI45861
aI104645
asS'd'
I20684
sa(dp41368
S'a'
I122458
sS's'
I18
sS'm'
(lp41369
I1
aI1
aI5
aI7
aI31
aI27
aI37
aI215
aI125
aI915
aI1297
aI3095
aI1529
aI12737
aI25675
aI29355
aI83939
aI106765
asS'd'
I20685
sa(dp41370
S'a'
I122479
sS's'
I18
sS'm'
(lp41371
I1
aI1
aI3
aI3
aI1
aI49
aI29
aI115
aI395
aI647
aI147
aI3905
aI1025
aI8873
aI10587
aI25471
aI72089
aI171467
asS'd'
I20686
sa(dp41372
S'a'
I122484
sS's'
I18
sS'm'
(lp41373
I1
aI3
aI7
aI1
aI21
aI55
aI57
aI233
aI487
aI883
aI439
aI929
aI1405
aI13709
aI2389
aI20205
aI17579
aI9129
asS'd'
I20687
sa(dp41374
S'a'
I122521
sS's'
I18
sS'm'
(lp41375
I1
aI3
aI5
aI5
aI23
aI51
aI55
aI45
aI307
aI855
aI933
aI1443
aI4757
aI8719
aI28401
aI35189
aI105329
aI9211
asS'd'
I20688
sa(dp41376
S'a'
I122540
sS's'
I18
sS'm'
(lp41377
I1
aI3
aI7
aI11
aI29
aI17
aI17
aI147
aI221
aI997
aI1433
aI59
aI8027
aI231
aI30335
aI2153
aI21393
aI116661
asS'd'
I20689
sa(dp41378
S'a'
I122545
sS's'
I18
sS'm'
(lp41379
I1
aI1
aI7
aI3
aI5
aI43
aI47
aI155
aI357
aI915
aI1923
aI3315
aI4107
aI9785
aI4847
aI57683
aI87569
aI179583
asS'd'
I20690
sa(dp41380
S'a'
I122552
sS's'
I18
sS'm'
(lp41381
I1
aI1
aI7
aI7
aI27
aI5
aI37
aI95
aI265
aI113
aI143
aI3755
aI5793
aI5601
aI16621
aI54777
aI15989
aI158933
asS'd'
I20691
sa(dp41382
S'a'
I122560
sS's'
I18
sS'm'
(lp41383
I1
aI3
aI1
aI1
aI13
aI31
aI113
aI255
aI367
aI559
aI1777
aI4065
aI8061
aI15785
aI10345
aI54833
aI95277
aI159347
asS'd'
I20692
sa(dp41384
S'a'
I122575
sS's'
I18
sS'm'
(lp41385
I1
aI1
aI1
aI15
aI3
aI43
aI85
aI251
aI193
aI19
aI1685
aI271
aI1779
aI11901
aI18983
aI65361
aI128217
aI248051
asS'd'
I20693
sa(dp41386
S'a'
I122578
sS's'
I18
sS'm'
(lp41387
I1
aI3
aI5
aI7
aI29
aI21
aI5
aI47
aI263
aI913
aI83
aI3233
aI113
aI8341
aI14473
aI37405
aI2363
aI155931
asS'd'
I20694
sa(dp41388
S'a'
I122605
sS's'
I18
sS'm'
(lp41389
I1
aI1
aI7
aI15
aI7
aI25
aI41
aI39
aI315
aI323
aI827
aI1277
aI1211
aI4465
aI21161
aI36865
aI6689
aI139147
asS'd'
I20695
sa(dp41390
S'a'
I122661
sS's'
I18
sS'm'
(lp41391
I1
aI1
aI1
aI9
aI11
aI45
aI81
aI235
aI31
aI247
aI77
aI1877
aI7119
aI16007
aI2225
aI65
aI85537
aI99251
asS'd'
I20696
sa(dp41392
S'a'
I122673
sS's'
I18
sS'm'
(lp41393
I1
aI1
aI3
aI11
aI19
aI5
aI49
aI179
aI345
aI961
aI349
aI2099
aI2317
aI12771
aI27169
aI59389
aI116071
aI68333
asS'd'
I20697
sa(dp41394
S'a'
I122722
sS's'
I18
sS'm'
(lp41395
I1
aI3
aI3
aI5
aI15
aI53
aI47
aI177
aI103
aI941
aI87
aI2813
aI7729
aI8003
aI7717
aI40095
aI74569
aI106617
asS'd'
I20698
sa(dp41396
S'a'
I122728
sS's'
I18
sS'm'
(lp41397
I1
aI1
aI7
aI7
aI31
aI23
aI105
aI205
aI325
aI855
aI1529
aI3601
aI7151
aI15827
aI16241
aI18221
aI55771
aI139225
asS'd'
I20699
sa(dp41398
S'a'
I122739
sS's'
I18
sS'm'
(lp41399
I1
aI3
aI5
aI13
aI13
aI5
aI95
aI167
aI25
aI779
aI1147
aI221
aI5055
aI10943
aI28077
aI15131
aI89501
aI137407
asS'd'
I20700
sa(dp41400
S'a'
I122751
sS's'
I18
sS'm'
(lp41401
I1
aI3
aI7
aI13
aI31
aI3
aI105
aI163
aI41
aI823
aI1493
aI2985
aI5589
aI3543
aI24683
aI34469
aI40595
aI200875
asS'd'
I20701
sa(dp41402
S'a'
I122762
sS's'
I18
sS'm'
(lp41403
I1
aI3
aI5
aI3
aI27
aI35
aI105
aI163
aI477
aI667
aI45
aI319
aI3201
aI11535
aI19349
aI55253
aI60275
aI209597
asS'd'
I20702
sa(dp41404
S'a'
I122769
sS's'
I18
sS'm'
(lp41405
I1
aI3
aI1
aI11
aI5
aI13
aI25
aI225
aI169
aI925
aI1617
aI537
aI891
aI5583
aI7181
aI39953
aI97537
aI104019
asS'd'
I20703
sa(dp41406
S'a'
I122772
sS's'
I18
sS'm'
(lp41407
I1
aI3
aI5
aI9
aI13
aI1
aI69
aI147
aI465
aI259
aI1219
aI2407
aI4015
aI4883
aI4333
aI40441
aI31289
aI52989
asS'd'
I20704
sa(dp41408
S'a'
I122782
sS's'
I18
sS'm'
(lp41409
I1
aI3
aI5
aI5
aI17
aI29
aI105
aI233
aI307
aI807
aI1535
aI251
aI135
aI925
aI6865
aI59739
aI112757
aI208275
asS'd'
I20705
sa(dp41410
S'a'
I122798
sS's'
I18
sS'm'
(lp41411
I1
aI1
aI1
aI7
aI9
aI21
aI49
aI247
aI33
aI127
aI1277
aI1745
aI139
aI12165
aI23517
aI50235
aI101003
aI109031
asS'd'
I20706
sa(dp41412
S'a'
I122810
sS's'
I18
sS'm'
(lp41413
I1
aI3
aI5
aI9
aI19
aI7
aI9
aI139
aI511
aI901
aI551
aI2717
aI6091
aI3213
aI819
aI51381
aI108333
aI119681
asS'd'
I20707
sa(dp41414
S'a'
I122812
sS's'
I18
sS'm'
(lp41415
I1
aI1
aI1
aI3
aI19
aI17
aI73
aI19
aI313
aI589
aI1965
aI1745
aI921
aI4237
aI12527
aI10735
aI110139
aI171513
asS'd'
I20708
sa(dp41416
S'a'
I122818
sS's'
I18
sS'm'
(lp41417
I1
aI1
aI5
aI15
aI25
aI29
aI15
aI217
aI87
aI793
aI419
aI915
aI1359
aI10507
aI25343
aI62977
aI100913
aI110041
asS'd'
I20709
sa(dp41418
S'a'
I122829
sS's'
I18
sS'm'
(lp41419
I1
aI3
aI7
aI15
aI15
aI15
aI91
aI243
aI441
aI437
aI1759
aI2659
aI2319
aI7783
aI16857
aI19051
aI15463
aI253115
asS'd'
I20710
sa(dp41420
S'a'
I122851
sS's'
I18
sS'm'
(lp41421
I1
aI1
aI5
aI13
aI15
aI5
aI7
aI165
aI355
aI559
aI217
aI235
aI3565
aI12047
aI2387
aI62285
aI73363
aI238551
asS'd'
I20711
sa(dp41422
S'a'
I122857
sS's'
I18
sS'm'
(lp41423
I1
aI3
aI7
aI11
aI5
aI5
aI69
aI205
aI179
aI815
aI335
aI979
aI2129
aI6221
aI31987
aI13623
aI23103
aI24373
asS'd'
I20712
sa(dp41424
S'a'
I122868
sS's'
I18
sS'm'
(lp41425
I1
aI1
aI7
aI11
aI31
aI51
aI55
aI93
aI195
aI219
aI825
aI2919
aI4495
aI5927
aI11813
aI16415
aI121595
aI188613
asS'd'
I20713
sa(dp41426
S'a'
I122893
sS's'
I18
sS'm'
(lp41427
I1
aI1
aI3
aI1
aI3
aI11
aI73
aI45
aI141
aI219
aI337
aI2569
aI6549
aI3699
aI2417
aI2945
aI19389
aI82561
asS'd'
I20714
sa(dp41428
S'a'
I122894
sS's'
I18
sS'm'
(lp41429
I1
aI3
aI1
aI13
aI5
aI47
aI47
aI81
aI11
aI57
aI1965
aI2173
aI3209
aI10617
aI19887
aI5571
aI61403
aI37401
asS'd'
I20715
sa(dp41430
S'a'
I122902
sS's'
I18
sS'm'
(lp41431
I1
aI3
aI3
aI7
aI3
aI5
aI99
aI253
aI287
aI655
aI813
aI3365
aI2387
aI8951
aI1561
aI37637
aI97625
aI148699
asS'd'
I20716
sa(dp41432
S'a'
I122921
sS's'
I18
sS'm'
(lp41433
I1
aI1
aI3
aI5
aI9
aI31
aI49
aI7
aI55
aI607
aI1489
aI3229
aI5871
aI1271
aI22751
aI32309
aI16125
aI93409
asS'd'
I20717
sa(dp41434
S'a'
I122959
sS's'
I18
sS'm'
(lp41435
I1
aI3
aI1
aI5
aI19
aI35
aI29
aI233
aI407
aI297
aI1465
aI3089
aI7535
aI7221
aI24469
aI42653
aI65719
aI196771
asS'd'
I20718
sa(dp41436
S'a'
I122984
sS's'
I18
sS'm'
(lp41437
I1
aI1
aI7
aI13
aI17
aI19
aI37
aI45
aI211
aI545
aI963
aI79
aI13
aI8319
aI8045
aI24975
aI122749
aI25845
asS'd'
I20719
sa(dp41438
S'a'
I123001
sS's'
I18
sS'm'
(lp41439
I1
aI3
aI5
aI3
aI23
aI23
aI53
aI101
aI363
aI49
aI1351
aI3419
aI1603
aI10795
aI5289
aI63695
aI113911
aI228301
asS'd'
I20720
sa(dp41440
S'a'
I123038
sS's'
I18
sS'm'
(lp41441
I1
aI3
aI7
aI9
aI15
aI59
aI125
aI59
aI11
aI397
aI693
aI397
aI3829
aI14349
aI13973
aI54739
aI22093
aI216009
asS'd'
I20721
sa(dp41442
S'a'
I123076
sS's'
I18
sS'm'
(lp41443
I1
aI1
aI1
aI13
aI5
aI63
aI21
aI23
aI421
aI35
aI589
aI803
aI6193
aI11375
aI9501
aI34441
aI68421
aI120109
asS'd'
I20722
sa(dp41444
S'a'
I123100
sS's'
I18
sS'm'
(lp41445
I1
aI3
aI3
aI7
aI23
aI25
aI79
aI203
aI177
aI173
aI175
aI809
aI4331
aI6953
aI4999
aI34345
aI94481
aI88683
asS'd'
I20723
sa(dp41446
S'a'
I123107
sS's'
I18
sS'm'
(lp41447
I1
aI3
aI3
aI3
aI1
aI51
aI109
aI195
aI83
aI747
aI63
aI325
aI927
aI1757
aI32055
aI37185
aI22697
aI41509
asS'd'
I20724
sa(dp41448
S'a'
I123122
sS's'
I18
sS'm'
(lp41449
I1
aI3
aI5
aI13
aI1
aI63
aI45
aI73
aI121
aI445
aI1935
aI3373
aI563
aI7503
aI17941
aI60313
aI42219
aI220917
asS'd'
I20725
sa(dp41450
S'a'
I123139
sS's'
I18
sS'm'
(lp41451
I1
aI3
aI5
aI13
aI27
aI31
aI7
aI83
aI109
aI387
aI447
aI1691
aI1301
aI7449
aI8075
aI30713
aI87207
aI84855
asS'd'
I20726
sa(dp41452
S'a'
I123175
sS's'
I18
sS'm'
(lp41453
I1
aI3
aI1
aI5
aI25
aI51
aI83
aI11
aI261
aI977
aI1415
aI2973
aI1789
aI12641
aI16279
aI4225
aI44237
aI173561
asS'd'
I20727
sa(dp41454
S'a'
I123190
sS's'
I18
sS'm'
(lp41455
I1
aI1
aI5
aI15
aI17
aI47
aI105
aI199
aI83
aI705
aI1215
aI2759
aI7509
aI10407
aI4005
aI4575
aI65961
aI209933
asS'd'
I20728
sa(dp41456
S'a'
I123201
sS's'
I18
sS'm'
(lp41457
I1
aI3
aI1
aI13
aI27
aI3
aI5
aI199
aI405
aI515
aI291
aI3399
aI1497
aI14755
aI30229
aI35075
aI111585
aI16633
asS'd'
I20729
sa(dp41458
S'a'
I123219
sS's'
I18
sS'm'
(lp41459
I1
aI1
aI1
aI13
aI17
aI53
aI93
aI161
aI447
aI903
aI947
aI1871
aI3597
aI10575
aI18389
aI48551
aI65229
aI32591
asS'd'
I20730
sa(dp41460
S'a'
I123232
sS's'
I18
sS'm'
(lp41461
I1
aI3
aI5
aI11
aI9
aI7
aI53
aI155
aI299
aI523
aI1653
aI3517
aI2725
aI9485
aI27099
aI47895
aI30169
aI260463
asS'd'
I20731
sa(dp41462
S'a'
I123255
sS's'
I18
sS'm'
(lp41463
I1
aI1
aI3
aI13
aI11
aI45
aI31
aI183
aI445
aI21
aI313
aI2597
aI195
aI16053
aI7323
aI52951
aI25919
aI9323
asS'd'
I20732
sa(dp41464
S'a'
I123285
sS's'
I18
sS'm'
(lp41465
I1
aI1
aI1
aI15
aI1
aI35
aI15
aI115
aI53
aI561
aI1141
aI3261
aI83
aI2547
aI8925
aI43455
aI112755
aI94157
asS'd'
I20733
sa(dp41466
S'a'
I123328
sS's'
I18
sS'm'
(lp41467
I1
aI1
aI3
aI5
aI27
aI25
aI81
aI51
aI209
aI87
aI379
aI3167
aI4953
aI13885
aI20159
aI103
aI115363
aI123585
asS'd'
I20734
sa(dp41468
S'a'
I123385
sS's'
I18
sS'm'
(lp41469
I1
aI3
aI5
aI1
aI25
aI29
aI107
aI225
aI77
aI435
aI2009
aI4069
aI3703
aI8855
aI14101
aI61683
aI16993
aI110823
asS'd'
I20735
sa(dp41470
S'a'
I123397
sS's'
I18
sS'm'
(lp41471
I1
aI3
aI7
aI3
aI31
aI25
aI51
aI117
aI397
aI271
aI89
aI3571
aI2357
aI4923
aI16303
aI2357
aI107775
aI73809
asS'd'
I20736
sa(dp41472
S'a'
I123404
sS's'
I18
sS'm'
(lp41473
I1
aI1
aI3
aI1
aI29
aI21
aI33
aI67
aI363
aI753
aI915
aI3715
aI2013
aI8439
aI5779
aI267
aI32687
aI104283
asS'd'
I20737
sa(dp41474
S'a'
I123419
sS's'
I18
sS'm'
(lp41475
I1
aI3
aI1
aI11
aI11
aI7
aI31
aI43
aI339
aI917
aI2005
aI759
aI4285
aI1933
aI4341
aI19111
aI130651
aI122853
asS'd'
I20738
sa(dp41476
S'a'
I123435
sS's'
I18
sS'm'
(lp41477
I1
aI1
aI5
aI1
aI17
aI63
aI85
aI63
aI387
aI127
aI1313
aI619
aI6525
aI9003
aI10915
aI64507
aI13175
aI45219
asS'd'
I20739
sa(dp41478
S'a'
I123437
sS's'
I18
sS'm'
(lp41479
I1
aI3
aI7
aI9
aI25
aI15
aI43
aI13
aI411
aI391
aI571
aI527
aI8175
aI10849
aI20093
aI2987
aI29869
aI77207
asS'd'
I20740
sa(dp41480
S'a'
I123443
sS's'
I18
sS'm'
(lp41481
I1
aI1
aI5
aI3
aI29
aI53
aI105
aI57
aI91
aI977
aI1103
aI2977
aI5617
aI7203
aI26717
aI28463
aI55909
aI59943
asS'd'
I20741
sa(dp41482
S'a'
I123500
sS's'
I18
sS'm'
(lp41483
I1
aI3
aI1
aI7
aI1
aI47
aI111
aI215
aI189
aI377
aI11
aI871
aI2267
aI5705
aI8165
aI38895
aI71025
aI10921
asS'd'
I20742
sa(dp41484
S'a'
I123506
sS's'
I18
sS'm'
(lp41485
I1
aI1
aI7
aI9
aI29
aI55
aI1
aI103
aI505
aI697
aI317
aI3209
aI3643
aI13689
aI8499
aI14671
aI67937
aI100467
asS'd'
I20743
sa(dp41486
S'a'
I123511
sS's'
I18
sS'm'
(lp41487
I1
aI3
aI1
aI5
aI25
aI15
aI99
aI199
aI117
aI957
aI1421
aI1719
aI5185
aI15247
aI28615
aI2657
aI46867
aI190135
asS'd'
I20744
sa(dp41488
S'a'
I123512
sS's'
I18
sS'm'
(lp41489
I1
aI1
aI7
aI9
aI27
aI55
aI81
aI47
aI15
aI497
aI537
aI857
aI2905
aI1909
aI3341
aI32625
aI123189
aI21875
asS'd'
I20745
sa(dp41490
S'a'
I123518
sS's'
I18
sS'm'
(lp41491
I1
aI3
aI1
aI3
aI19
aI51
aI97
aI143
aI305
aI1021
aI543
aI829
aI7593
aI8101
aI6337
aI4869
aI19177
aI38981
asS'd'
I20746
sa(dp41492
S'a'
I123522
sS's'
I18
sS'm'
(lp41493
I1
aI3
aI3
aI15
aI9
aI37
aI51
aI193
aI295
aI731
aI809
aI4065
aI3377
aI15303
aI12505
aI11327
aI76191
aI139899
asS'd'
I20747
sa(dp41494
S'a'
I123545
sS's'
I18
sS'm'
(lp41495
I1
aI1
aI5
aI1
aI3
aI19
aI5
aI223
aI379
aI7
aI755
aI1127
aI505
aI9429
aI27409
aI50817
aI97599
aI179019
asS'd'
I20748
sa(dp41496
S'a'
I123572
sS's'
I18
sS'm'
(lp41497
I1
aI3
aI5
aI3
aI5
aI11
aI83
aI119
aI3
aI311
aI405
aI2401
aI1821
aI1381
aI4567
aI44079
aI61903
aI183583
asS'd'
I20749
sa(dp41498
S'a'
I123587
sS's'
I18
sS'm'
(lp41499
I1
aI3
aI7
aI9
aI7
aI11
aI107
aI95
aI271
aI537
aI335
aI3079
aI6695
aI1163
aI32055
aI44985
aI29075
aI94235
asS'd'
I20750
sa(dp41500
S'a'
I123594
sS's'
I18
sS'm'
(lp41501
I1
aI3
aI7
aI7
aI7
aI57
aI59
aI85
aI199
aI563
aI1835
aI351
aI7675
aI2601
aI3717
aI57975
aI92529
aI101511
asS'd'
I20751
sa(dp41502
S'a'
I123607
sS's'
I18
sS'm'
(lp41503
I1
aI1
aI1
aI5
aI17
aI5
aI97
aI43
aI101
aI141
aI1511
aI199
aI7157
aI3169
aI24815
aI55653
aI104195
aI37951
asS'd'
I20752
sa(dp41504
S'a'
I123608
sS's'
I18
sS'm'
(lp41505
I1
aI1
aI1
aI11
aI23
aI29
aI41
aI47
aI447
aI583
aI773
aI859
aI1657
aI8707
aI16709
aI53477
aI42037
aI186809
asS'd'
I20753
sa(dp41506
S'a'
I123630
sS's'
I18
sS'm'
(lp41507
I1
aI1
aI1
aI13
aI13
aI5
aI85
aI9
aI213
aI511
aI1003
aI811
aI2271
aI14715
aI21423
aI48127
aI50613
aI214031
asS'd'
I20754
sa(dp41508
S'a'
I123635
sS's'
I18
sS'm'
(lp41509
I1
aI1
aI3
aI5
aI7
aI31
aI87
aI101
aI297
aI853
aI1599
aI1521
aI4965
aI9655
aI23543
aI62277
aI11231
aI49931
asS'd'
I20755
sa(dp41510
S'a'
I123659
sS's'
I18
sS'm'
(lp41511
I1
aI1
aI7
aI7
aI19
aI31
aI9
aI165
aI207
aI919
aI739
aI3849
aI2121
aI867
aI3233
aI40867
aI75721
aI5327
asS'd'
I20756
sa(dp41512
S'a'
I123692
sS's'
I18
sS'm'
(lp41513
I1
aI3
aI7
aI1
aI17
aI3
aI27
aI13
aI431
aI283
aI465
aI1427
aI1937
aI15601
aI21793
aI9315
aI54285
aI196453
asS'd'
I20757
sa(dp41514
S'a'
I123695
sS's'
I18
sS'm'
(lp41515
I1
aI1
aI7
aI3
aI17
aI39
aI65
aI137
aI511
aI19
aI1357
aI3373
aI5227
aI2485
aI1151
aI25061
aI117507
aI119219
asS'd'
I20758
sa(dp41516
S'a'
I123700
sS's'
I18
sS'm'
(lp41517
I1
aI3
aI3
aI9
aI9
aI23
aI13
aI235
aI505
aI625
aI115
aI3859
aI6943
aI14719
aI6363
aI14957
aI28241
aI187989
asS'd'
I20759
sa(dp41518
S'a'
I123730
sS's'
I18
sS'm'
(lp41519
I1
aI3
aI7
aI13
aI11
aI31
aI23
aI39
aI463
aI441
aI1145
aI417
aI4177
aI1655
aI26491
aI16895
aI26263
aI198157
asS'd'
I20760
sa(dp41520
S'a'
I123748
sS's'
I18
sS'm'
(lp41521
I1
aI3
aI3
aI7
aI25
aI55
aI121
aI157
aI131
aI537
aI1891
aI2367
aI1717
aI2331
aI20251
aI8679
aI62657
aI121957
asS'd'
I20761
sa(dp41522
S'a'
I123751
sS's'
I18
sS'm'
(lp41523
I1
aI1
aI1
aI11
aI7
aI43
aI59
aI101
aI333
aI961
aI569
aI1603
aI3009
aI6539
aI9627
aI5759
aI44401
aI127613
asS'd'
I20762
sa(dp41524
S'a'
I123770
sS's'
I18
sS'm'
(lp41525
I1
aI3
aI3
aI3
aI1
aI27
aI15
aI135
aI235
aI215
aI627
aI2427
aI2647
aI3201
aI22873
aI64445
aI32635
aI16587
asS'd'
I20763
sa(dp41526
S'a'
I123793
sS's'
I18
sS'm'
(lp41527
I1
aI3
aI7
aI7
aI5
aI49
aI63
aI103
aI467
aI897
aI117
aI1149
aI6045
aI5003
aI5005
aI6183
aI90815
aI190909
asS'd'
I20764
sa(dp41528
S'a'
I123800
sS's'
I18
sS'm'
(lp41529
I1
aI1
aI3
aI3
aI31
aI43
aI45
aI227
aI363
aI409
aI1097
aI3155
aI1519
aI14461
aI29377
aI19577
aI52595
aI94041
asS'd'
I20765
sa(dp41530
S'a'
I123803
sS's'
I18
sS'm'
(lp41531
I1
aI3
aI1
aI1
aI3
aI25
aI83
aI243
aI57
aI243
aI389
aI1427
aI5197
aI13125
aI18571
aI17845
aI74961
aI125569
asS'd'
I20766
sa(dp41532
S'a'
I123851
sS's'
I18
sS'm'
(lp41533
I1
aI3
aI5
aI5
aI1
aI39
aI11
aI175
aI97
aI1001
aI143
aI3653
aI5887
aI5845
aI5691
aI5433
aI62629
aI176261
asS'd'
I20767
sa(dp41534
S'a'
I123859
sS's'
I18
sS'm'
(lp41535
I1
aI1
aI1
aI13
aI11
aI41
aI41
aI155
aI17
aI823
aI1507
aI733
aI5663
aI13657
aI24133
aI9971
aI27179
aI108075
asS'd'
I20768
sa(dp41536
S'a'
I123866
sS's'
I18
sS'm'
(lp41537
I1
aI1
aI7
aI9
aI19
aI9
aI103
aI29
aI427
aI363
aI931
aI3959
aI1629
aI5127
aI14807
aI61937
aI127175
aI237233
asS'd'
I20769
sa(dp41538
S'a'
I123868
sS's'
I18
sS'm'
(lp41539
I1
aI1
aI1
aI5
aI29
aI53
aI51
aI229
aI123
aI1001
aI697
aI411
aI4669
aI5051
aI18447
aI55437
aI129269
aI72613
asS'd'
I20770
sa(dp41540
S'a'
I123877
sS's'
I18
sS'm'
(lp41541
I1
aI3
aI7
aI3
aI19
aI53
aI111
aI131
aI255
aI547
aI653
aI2839
aI1447
aI14397
aI5707
aI23773
aI127897
aI135177
asS'd'
I20771
sa(dp41542
S'a'
I123884
sS's'
I18
sS'm'
(lp41543
I1
aI3
aI3
aI5
aI21
aI37
aI53
aI219
aI359
aI341
aI489
aI2477
aI3383
aI6931
aI7753
aI2619
aI114267
aI63271
asS'd'
I20772
sa(dp41544
S'a'
I123909
sS's'
I18
sS'm'
(lp41545
I1
aI1
aI3
aI15
aI15
aI15
aI95
aI171
aI301
aI563
aI603
aI593
aI4037
aI7305
aI10849
aI11753
aI103087
aI74887
asS'd'
I20773
sa(dp41546
S'a'
I123910
sS's'
I18
sS'm'
(lp41547
I1
aI1
aI5
aI3
aI11
aI51
aI105
aI35
aI155
aI643
aI569
aI1697
aI1679
aI7547
aI19289
aI7065
aI57359
aI142855
asS'd'
I20774
sa(dp41548
S'a'
I123919
sS's'
I18
sS'm'
(lp41549
I1
aI1
aI7
aI15
aI23
aI55
aI17
aI67
aI375
aI559
aI355
aI165
aI93
aI973
aI22831
aI48027
aI98435
aI59945
asS'd'
I20775
sa(dp41550
S'a'
I123927
sS's'
I18
sS'm'
(lp41551
I1
aI3
aI7
aI3
aI5
aI43
aI9
aI253
aI111
aI819
aI45
aI3461
aI4821
aI14735
aI14469
aI29793
aI30681
aI26359
asS'd'
I20776
sa(dp41552
S'a'
I123937
sS's'
I18
sS'm'
(lp41553
I1
aI3
aI3
aI5
aI15
aI1
aI39
aI233
aI63
aI287
aI131
aI3453
aI427
aI4929
aI30085
aI18583
aI50119
aI262101
asS'd'
I20777
sa(dp41554
S'a'
I123950
sS's'
I18
sS'm'
(lp41555
I1
aI3
aI7
aI9
aI31
aI7
aI127
aI157
aI287
aI57
aI1091
aI1989
aI5045
aI13071
aI27705
aI58125
aI85317
aI66649
asS'd'
I20778
sa(dp41556
S'a'
I123957
sS's'
I18
sS'm'
(lp41557
I1
aI1
aI1
aI15
aI17
aI29
aI25
aI223
aI311
aI489
aI1901
aI3197
aI1813
aI10097
aI31915
aI54871
aI32289
aI227001
asS'd'
I20779
sa(dp41558
S'a'
I123979
sS's'
I18
sS'm'
(lp41559
I1
aI3
aI7
aI5
aI31
aI35
aI69
aI87
aI131
aI963
aI1125
aI1109
aI8037
aI3257
aI27655
aI50999
aI3715
aI57851
asS'd'
I20780
sa(dp41560
S'a'
I123996
sS's'
I18
sS'm'
(lp41561
I1
aI1
aI5
aI5
aI27
aI7
aI119
aI29
aI425
aI721
aI541
aI3069
aI3349
aI13623
aI12293
aI51395
aI14033
aI61545
asS'd'
I20781
sa(dp41562
S'a'
I124040
sS's'
I18
sS'm'
(lp41563
I1
aI3
aI7
aI3
aI17
aI49
aI103
aI115
aI387
aI729
aI1389
aI2257
aI3273
aI3375
aI23143
aI2835
aI28071
aI79533
asS'd'
I20782
sa(dp41564
S'a'
I124054
sS's'
I18
sS'm'
(lp41565
I1
aI1
aI3
aI7
aI9
aI19
aI55
aI159
aI261
aI467
aI17
aI2595
aI3947
aI7045
aI193
aI23629
aI89067
aI81197
asS'd'
I20783
sa(dp41566
S'a'
I124063
sS's'
I18
sS'm'
(lp41567
I1
aI1
aI1
aI15
aI13
aI39
aI103
aI195
aI251
aI769
aI1003
aI2707
aI3263
aI8451
aI8007
aI53789
aI112653
aI258717
asS'd'
I20784
sa(dp41568
S'a'
I124067
sS's'
I18
sS'm'
(lp41569
I1
aI3
aI5
aI1
aI3
aI21
aI29
aI217
aI125
aI779
aI1597
aI513
aI2677
aI3979
aI31903
aI64813
aI69963
aI92887
asS'd'
I20785
sa(dp41570
S'a'
I124088
sS's'
I18
sS'm'
(lp41571
I1
aI1
aI3
aI11
aI11
aI59
aI81
aI237
aI447
aI703
aI41
aI3369
aI3547
aI14935
aI31693
aI52005
aI74149
aI131039
asS'd'
I20786
sa(dp41572
S'a'
I124101
sS's'
I18
sS'm'
(lp41573
I1
aI1
aI3
aI1
aI23
aI1
aI33
aI93
aI275
aI847
aI921
aI2745
aI533
aI8975
aI30529
aI46809
aI98975
aI75541
asS'd'
I20787
sa(dp41574
S'a'
I124154
sS's'
I18
sS'm'
(lp41575
I1
aI1
aI7
aI7
aI15
aI1
aI3
aI213
aI71
aI1009
aI1951
aI3015
aI713
aI9365
aI21949
aI60983
aI117633
aI225387
asS'd'
I20788
sa(dp41576
S'a'
I124162
sS's'
I18
sS'm'
(lp41577
I1
aI1
aI1
aI11
aI31
aI27
aI59
aI179
aI231
aI635
aI1555
aI2765
aI31
aI15065
aI22719
aI59251
aI84733
aI96769
asS'd'
I20789
sa(dp41578
S'a'
I124167
sS's'
I18
sS'm'
(lp41579
I1
aI3
aI7
aI3
aI19
aI5
aI35
aI207
aI317
aI735
aI943
aI3987
aI4021
aI11229
aI13015
aI713
aI125167
aI55887
asS'd'
I20790
sa(dp41580
S'a'
I124176
sS's'
I18
sS'm'
(lp41581
I1
aI1
aI3
aI1
aI27
aI49
aI81
aI17
aI253
aI633
aI43
aI2953
aI3151
aI8429
aI30625
aI28551
aI126683
aI175087
asS'd'
I20791
sa(dp41582
S'a'
I124195
sS's'
I18
sS'm'
(lp41583
I1
aI3
aI3
aI3
aI9
aI53
aI111
aI245
aI57
aI557
aI945
aI2957
aI7669
aI12537
aI17291
aI9713
aI87727
aI44739
asS'd'
I20792
sa(dp41584
S'a'
I124230
sS's'
I18
sS'm'
(lp41585
I1
aI3
aI1
aI1
aI5
aI39
aI77
aI127
aI87
aI687
aI1485
aI1555
aI2567
aI13551
aI17075
aI24003
aI47627
aI129813
asS'd'
I20793
sa(dp41586
S'a'
I124267
sS's'
I18
sS'm'
(lp41587
I1
aI3
aI5
aI9
aI25
aI35
aI87
aI233
aI439
aI563
aI1719
aI419
aI4459
aI4285
aI25157
aI943
aI111543
aI232107
asS'd'
I20794
sa(dp41588
S'a'
I124346
sS's'
I18
sS'm'
(lp41589
I1
aI3
aI5
aI11
aI1
aI13
aI97
aI153
aI459
aI551
aI73
aI2087
aI3985
aI4661
aI15603
aI22211
aI123163
aI187233
asS'd'
I20795
sa(dp41590
S'a'
I124389
sS's'
I18
sS'm'
(lp41591
I1
aI1
aI7
aI13
aI17
aI59
aI5
aI219
aI353
aI441
aI387
aI441
aI3009
aI485
aI20081
aI38023
aI50659
aI159243
asS'd'
I20796
sa(dp41592
S'a'
I124418
sS's'
I18
sS'm'
(lp41593
I1
aI3
aI1
aI7
aI21
aI31
aI117
aI49
aI227
aI677
aI417
aI1153
aI1611
aI1669
aI25161
aI52223
aI15109
aI114759
asS'd'
I20797
sa(dp41594
S'a'
I124420
sS's'
I18
sS'm'
(lp41595
I1
aI1
aI7
aI9
aI19
aI13
aI111
aI7
aI1
aI701
aI731
aI2075
aI685
aI15679
aI19149
aI44315
aI41719
aI243975
asS'd'
I20798
sa(dp41596
S'a'
I124454
sS's'
I18
sS'm'
(lp41597
I1
aI3
aI3
aI7
aI15
aI63
aI59
aI105
aI327
aI39
aI1497
aI2407
aI2865
aI7065
aI9957
aI20031
aI45359
aI73657
asS'd'
I20799
sa(dp41598
S'a'
I124490
sS's'
I18
sS'm'
(lp41599
I1
aI3
aI1
aI13
aI7
aI3
aI55
aI221
aI443
aI953
aI15
aI2455
aI4681
aI16247
aI18179
aI44731
aI41323
aI172621
asS'd'
I20800
sa(dp41600
S'a'
I124503
sS's'
I18
sS'm'
(lp41601
I1
aI1
aI3
aI9
aI1
aI27
aI65
aI167
aI115
aI137
aI819
aI2129
aI3393
aI5901
aI11735
aI62753
aI14941
aI21425
asS'd'
I20801
sa(dp41602
S'a'
I124519
sS's'
I18
sS'm'
(lp41603
I1
aI3
aI7
aI11
aI7
aI9
aI41
aI175
aI237
aI481
aI59
aI265
aI2135
aI9419
aI3937
aI55959
aI48343
aI172549
asS'd'
I20802
sa(dp41604
S'a'
I124533
sS's'
I18
sS'm'
(lp41605
I1
aI1
aI1
aI13
aI23
aI33
aI105
aI87
aI461
aI297
aI1345
aI3715
aI7715
aI16369
aI19017
aI15141
aI10873
aI109641
asS'd'
I20803
sa(dp41606
S'a'
I124544
sS's'
I18
sS'm'
(lp41607
I1
aI3
aI5
aI1
aI11
aI1
aI13
aI41
aI447
aI511
aI447
aI2295
aI2401
aI14171
aI16269
aI50453
aI40361
aI205857
asS'd'
I20804
sa(dp41608
S'a'
I124589
sS's'
I18
sS'm'
(lp41609
I1
aI1
aI3
aI3
aI27
aI7
aI35
aI193
aI113
aI341
aI335
aI2113
aI343
aI4575
aI20863
aI40383
aI86787
aI142603
asS'd'
I20805
sa(dp41610
S'a'
I124616
sS's'
I18
sS'm'
(lp41611
I1
aI1
aI7
aI11
aI17
aI25
aI51
aI89
aI341
aI237
aI1233
aI1505
aI7401
aI6887
aI26897
aI20127
aI51077
aI107559
asS'd'
I20806
sa(dp41612
S'a'
I124630
sS's'
I18
sS'm'
(lp41613
I1
aI1
aI1
aI15
aI9
aI59
aI29
aI115
aI339
aI785
aI201
aI947
aI1501
aI6883
aI169
aI44059
aI17527
aI197623
asS'd'
I20807
sa(dp41614
S'a'
I124636
sS's'
I18
sS'm'
(lp41615
I1
aI3
aI7
aI3
aI17
aI59
aI15
aI5
aI379
aI347
aI821
aI4047
aI3565
aI13689
aI23275
aI27901
aI121401
aI43077
asS'd'
I20808
sa(dp41616
S'a'
I124645
sS's'
I18
sS'm'
(lp41617
I1
aI1
aI1
aI7
aI15
aI11
aI35
aI29
aI59
aI99
aI181
aI3035
aI3239
aI1553
aI32319
aI64195
aI115247
aI149211
asS'd'
I20809
sa(dp41618
S'a'
I124652
sS's'
I18
sS'm'
(lp41619
I1
aI1
aI5
aI3
aI31
aI49
aI87
aI177
aI231
aI167
aI373
aI1125
aI5919
aI4805
aI31983
aI42873
aI30169
aI91853
asS'd'
I20810
sa(dp41620
S'a'
I124663
sS's'
I18
sS'm'
(lp41621
I1
aI1
aI5
aI15
aI7
aI63
aI37
aI15
aI459
aI449
aI1835
aI1769
aI2527
aI2577
aI4251
aI62459
aI76699
aI81721
asS'd'
I20811
sa(dp41622
S'a'
I124667
sS's'
I18
sS'm'
(lp41623
I1
aI1
aI1
aI3
aI21
aI15
aI19
aI183
aI423
aI827
aI341
aI2101
aI3797
aI7103
aI30845
aI24511
aI115337
aI117019
asS'd'
I20812
sa(dp41624
S'a'
I124690
sS's'
I18
sS'm'
(lp41625
I1
aI3
aI5
aI5
aI19
aI1
aI97
aI13
aI491
aI51
aI445
aI1987
aI7481
aI2613
aI23141
aI36603
aI26917
aI177397
asS'd'
I20813
sa(dp41626
S'a'
I124755
sS's'
I18
sS'm'
(lp41627
I1
aI1
aI1
aI5
aI15
aI43
aI21
aI167
aI259
aI989
aI1937
aI1519
aI2201
aI13973
aI18031
aI31583
aI57557
aI252737
asS'd'
I20814
sa(dp41628
S'a'
I124757
sS's'
I18
sS'm'
(lp41629
I1
aI3
aI3
aI7
aI31
aI11
aI89
aI53
aI279
aI83
aI1247
aI1221
aI5499
aI1199
aI361
aI22269
aI88633
aI134975
asS'd'
I20815
sa(dp41630
S'a'
I124819
sS's'
I18
sS'm'
(lp41631
I1
aI1
aI3
aI5
aI9
aI5
aI113
aI41
aI143
aI655
aI1147
aI2043
aI4229
aI10523
aI1453
aI1735
aI76259
aI30607
asS'd'
I20816
sa(dp41632
S'a'
I124862
sS's'
I18
sS'm'
(lp41633
I1
aI1
aI1
aI11
aI5
aI7
aI9
aI155
aI221
aI619
aI813
aI3111
aI6039
aI10789
aI10905
aI33285
aI62841
aI229217
asS'd'
I20817
sa(dp41634
S'a'
I124864
sS's'
I18
sS'm'
(lp41635
I1
aI3
aI5
aI9
aI19
aI19
aI123
aI65
aI419
aI597
aI87
aI3843
aI4857
aI15903
aI23655
aI13023
aI8389
aI230803
asS'd'
I20818
sa(dp41636
S'a'
I124874
sS's'
I18
sS'm'
(lp41637
I1
aI3
aI3
aI15
aI9
aI19
aI51
aI223
aI197
aI759
aI139
aI59
aI6547
aI1043
aI5077
aI55267
aI23681
aI17099
asS'd'
I20819
sa(dp41638
S'a'
I124882
sS's'
I18
sS'm'
(lp41639
I1
aI1
aI3
aI3
aI15
aI43
aI63
aI85
aI227
aI961
aI1043
aI1069
aI6557
aI7499
aI31639
aI4345
aI26991
aI132783
asS'd'
I20820
sa(dp41640
S'a'
I124903
sS's'
I18
sS'm'
(lp41641
I1
aI3
aI1
aI1
aI25
aI33
aI59
aI219
aI435
aI105
aI1001
aI323
aI2729
aI12517
aI16607
aI57533
aI101167
aI53829
asS'd'
I20821
sa(dp41642
S'a'
I124915
sS's'
I18
sS'm'
(lp41643
I1
aI3
aI7
aI7
aI9
aI55
aI3
aI247
aI303
aI453
aI861
aI817
aI705
aI14337
aI15965
aI28867
aI126763
aI204005
asS'd'
I20822
sa(dp41644
S'a'
I124938
sS's'
I18
sS'm'
(lp41645
I1
aI1
aI3
aI7
aI1
aI39
aI15
aI39
aI359
aI313
aI1753
aI2835
aI387
aI16223
aI10945
aI19481
aI19995
aI29989
asS'd'
I20823
sa(dp41646
S'a'
I124945
sS's'
I18
sS'm'
(lp41647
I1
aI1
aI1
aI5
aI3
aI31
aI103
aI123
aI493
aI1023
aI119
aI2175
aI2273
aI11637
aI21605
aI23349
aI100759
aI41227
asS'd'
I20824
sa(dp41648
S'a'
I124973
sS's'
I18
sS'm'
(lp41649
I1
aI3
aI7
aI13
aI5
aI13
aI17
aI7
aI341
aI945
aI621
aI1421
aI3893
aI5825
aI26777
aI35497
aI13791
aI25415
asS'd'
I20825
sa(dp41650
S'a'
I124986
sS's'
I18
sS'm'
(lp41651
I1
aI1
aI5
aI1
aI19
aI27
aI39
aI7
aI177
aI749
aI1217
aI2133
aI6913
aI13489
aI23713
aI1085
aI31529
aI179741
asS'd'
I20826
sa(dp41652
S'a'
I124991
sS's'
I18
sS'm'
(lp41653
I1
aI3
aI3
aI3
aI15
aI21
aI5
aI145
aI281
aI131
aI1347
aI19
aI4917
aI8655
aI2515
aI36927
aI56551
aI202039
asS'd'
I20827
sa(dp41654
S'a'
I125011
sS's'
I18
sS'm'
(lp41655
I1
aI1
aI3
aI7
aI27
aI31
aI33
aI255
aI511
aI195
aI1493
aI2221
aI2157
aI9303
aI3957
aI14163
aI70435
aI215763
asS'd'
I20828
sa(dp41656
S'a'
I125014
sS's'
I18
sS'm'
(lp41657
I1
aI3
aI3
aI15
aI31
aI41
aI33
aI37
aI239
aI865
aI375
aI2217
aI809
aI11961
aI29393
aI52145
aI76223
aI202623
asS'd'
I20829
sa(dp41658
S'a'
I125082
sS's'
I18
sS'm'
(lp41659
I1
aI3
aI5
aI11
aI5
aI57
aI73
aI193
aI341
aI843
aI1817
aI231
aI65
aI5941
aI29693
aI31751
aI57081
aI180977
asS'd'
I20830
sa(dp41660
S'a'
I125105
sS's'
I18
sS'm'
(lp41661
I1
aI3
aI1
aI9
aI29
aI57
aI85
aI107
aI343
aI891
aI465
aI2413
aI1965
aI7303
aI7461
aI25857
aI110517
aI16995
asS'd'
I20831
sa(dp41662
S'a'
I125163
sS's'
I18
sS'm'
(lp41663
I1
aI1
aI7
aI15
aI7
aI49
aI121
aI211
aI253
aI511
aI1385
aI1205
aI33
aI7713
aI20059
aI47353
aI3267
aI215759
asS'd'
I20832
sa(dp41664
S'a'
I125165
sS's'
I18
sS'm'
(lp41665
I1
aI1
aI3
aI9
aI15
aI63
aI45
aI155
aI415
aI589
aI651
aI3707
aI5429
aI4497
aI13733
aI21231
aI18953
aI28671
asS'd'
I20833
sa(dp41666
S'a'
I125205
sS's'
I18
sS'm'
(lp41667
I1
aI1
aI5
aI15
aI7
aI57
aI31
aI109
aI427
aI921
aI629
aI3439
aI7615
aI4535
aI1507
aI58931
aI49597
aI214397
asS'd'
I20834
sa(dp41668
S'a'
I125219
sS's'
I18
sS'm'
(lp41669
I1
aI3
aI3
aI5
aI21
aI35
aI9
aI241
aI67
aI767
aI659
aI1639
aI7797
aI5209
aI14851
aI55311
aI108549
aI175937
asS'd'
I20835
sa(dp41670
S'a'
I125239
sS's'
I18
sS'm'
(lp41671
I1
aI1
aI7
aI15
aI21
aI61
aI13
aI47
aI267
aI269
aI1169
aI257
aI2481
aI8345
aI1061
aI28119
aI127197
aI95379
asS'd'
I20836
sa(dp41672
S'a'
I125240
sS's'
I18
sS'm'
(lp41673
I1
aI1
aI3
aI7
aI21
aI35
aI121
aI223
aI145
aI665
aI1389
aI2105
aI5499
aI1377
aI32417
aI39027
aI5335
aI248315
asS'd'
I20837
sa(dp41674
S'a'
I125257
sS's'
I18
sS'm'
(lp41675
I1
aI3
aI3
aI9
aI13
aI33
aI57
aI225
aI123
aI703
aI1049
aI709
aI7347
aI11317
aI12339
aI23247
aI62157
aI85931
asS'd'
I20838
sa(dp41676
S'a'
I125284
sS's'
I18
sS'm'
(lp41677
I1
aI3
aI3
aI3
aI5
aI3
aI105
aI115
aI13
aI675
aI757
aI987
aI6429
aI13017
aI21347
aI38829
aI82153
aI220677
asS'd'
I20839
sa(dp41678
S'a'
I125302
sS's'
I18
sS'm'
(lp41679
I1
aI3
aI1
aI13
aI5
aI51
aI75
aI205
aI7
aI561
aI207
aI1133
aI3303
aI1889
aI17093
aI5933
aI48109
aI244387
asS'd'
I20840
sa(dp41680
S'a'
I125327
sS's'
I18
sS'm'
(lp41681
I1
aI3
aI3
aI7
aI11
aI11
aI71
aI243
aI235
aI941
aI1875
aI2387
aI1139
aI16275
aI4537
aI18791
aI67927
aI156759
asS'd'
I20841
sa(dp41682
S'a'
I125336
sS's'
I18
sS'm'
(lp41683
I1
aI3
aI5
aI5
aI11
aI35
aI29
aI5
aI351
aI223
aI847
aI1539
aI4903
aI7619
aI24907
aI37071
aI57899
aI32981
asS'd'
I20842
sa(dp41684
S'a'
I125355
sS's'
I18
sS'm'
(lp41685
I1
aI3
aI1
aI7
aI23
aI7
aI119
aI17
aI123
aI845
aI1423
aI2995
aI3595
aI13287
aI31217
aI33939
aI6891
aI156477
asS'd'
I20843
sa(dp41686
S'a'
I125363
sS's'
I18
sS'm'
(lp41687
I1
aI1
aI1
aI9
aI11
aI55
aI121
aI85
aI175
aI757
aI1093
aI513
aI1117
aI14049
aI22377
aI4623
aI38511
aI51391
asS'd'
I20844
sa(dp41688
S'a'
I125365
sS's'
I18
sS'm'
(lp41689
I1
aI1
aI7
aI1
aI17
aI1
aI7
aI185
aI173
aI841
aI61
aI2735
aI6679
aI7617
aI17309
aI58047
aI11791
aI228635
asS'd'
I20845
sa(dp41690
S'a'
I125372
sS's'
I18
sS'm'
(lp41691
I1
aI1
aI5
aI5
aI1
aI1
aI61
aI57
aI37
aI857
aI531
aI2655
aI1907
aI4245
aI3047
aI50489
aI93447
aI116637
asS'd'
I20846
sa(dp41692
S'a'
I125378
sS's'
I18
sS'm'
(lp41693
I1
aI3
aI5
aI7
aI27
aI47
aI121
aI95
aI73
aI63
aI539
aI137
aI1765
aI4659
aI31141
aI24495
aI109541
aI16421
asS'd'
I20847
sa(dp41694
S'a'
I125383
sS's'
I18
sS'm'
(lp41695
I1
aI3
aI7
aI13
aI13
aI25
aI47
aI45
aI97
aI243
aI1509
aI3539
aI4791
aI5627
aI31981
aI57663
aI65359
aI32183
asS'd'
I20848
sa(dp41696
S'a'
I125387
sS's'
I18
sS'm'
(lp41697
I1
aI1
aI1
aI7
aI15
aI57
aI117
aI249
aI183
aI849
aI557
aI833
aI5751
aI8035
aI3371
aI11389
aI125581
aI248799
asS'd'
I20849
sa(dp41698
S'a'
I125390
sS's'
I18
sS'm'
(lp41699
I1
aI3
aI1
aI15
aI13
aI31
aI59
aI243
aI17
aI473
aI289
aI1527
aI649
aI2807
aI6183
aI6173
aI74381
aI261673
asS'd'
I20850
sa(dp41700
S'a'
I125414
sS's'
I18
sS'm'
(lp41701
I1
aI1
aI7
aI7
aI19
aI17
aI95
aI81
aI191
aI487
aI2023
aI307
aI3261
aI13885
aI9285
aI30831
aI114009
aI26483
asS'd'
I20851
sa(dp41702
S'a'
I125435
sS's'
I18
sS'm'
(lp41703
I1
aI1
aI3
aI11
aI17
aI21
aI85
aI193
aI477
aI267
aI393
aI39
aI5793
aI8621
aI25379
aI9721
aI13947
aI44235
asS'd'
I20852
sa(dp41704
S'a'
I125466
sS's'
I18
sS'm'
(lp41705
I1
aI1
aI7
aI5
aI27
aI33
aI93
aI9
aI471
aI751
aI1279
aI695
aI2625
aI7061
aI29577
aI5403
aI80705
aI77895
asS'd'
I20853
sa(dp41706
S'a'
I125481
sS's'
I18
sS'm'
(lp41707
I1
aI1
aI3
aI11
aI7
aI25
aI125
aI9
aI233
aI935
aI1897
aI3685
aI595
aI15499
aI43
aI29251
aI18029
aI250231
asS'd'
I20854
sa(dp41708
S'a'
I125522
sS's'
I18
sS'm'
(lp41709
I1
aI1
aI5
aI1
aI9
aI55
aI125
aI179
aI287
aI371
aI233
aI149
aI5639
aI5737
aI25251
aI103
aI117015
aI35579
asS'd'
I20855
sa(dp41710
S'a'
I125531
sS's'
I18
sS'm'
(lp41711
I1
aI1
aI5
aI9
aI17
aI33
aI69
aI15
aI209
aI521
aI1083
aI2469
aI679
aI9307
aI31539
aI63889
aI48825
aI126327
asS'd'
I20856
sa(dp41712
S'a'
I125544
sS's'
I18
sS'm'
(lp41713
I1
aI3
aI5
aI15
aI15
aI41
aI105
aI121
aI21
aI935
aI721
aI445
aI6759
aI4227
aI15227
aI54933
aI69589
aI2689
asS'd'
I20857
sa(dp41714
S'a'
I125564
sS's'
I18
sS'm'
(lp41715
I1
aI1
aI7
aI15
aI1
aI15
aI111
aI141
aI279
aI1013
aI825
aI1069
aI3793
aI12929
aI153
aI11463
aI87759
aI179987
asS'd'
I20858
sa(dp41716
S'a'
I125588
sS's'
I18
sS'm'
(lp41717
I1
aI1
aI1
aI9
aI15
aI59
aI109
aI103
aI223
aI695
aI1979
aI1241
aI2559
aI8627
aI10559
aI53319
aI94311
aI245193
asS'd'
I20859
sa(dp41718
S'a'
I125602
sS's'
I18
sS'm'
(lp41719
I1
aI1
aI5
aI3
aI1
aI41
aI51
aI129
aI297
aI15
aI637
aI2489
aI343
aI13549
aI7707
aI36757
aI55703
aI161043
asS'd'
I20860
sa(dp41720
S'a'
I125613
sS's'
I18
sS'm'
(lp41721
I1
aI3
aI5
aI1
aI13
aI35
aI119
aI219
aI319
aI733
aI789
aI1343
aI8035
aI15049
aI981
aI14477
aI13717
aI177481
asS'd'
I20861
sa(dp41722
S'a'
I125621
sS's'
I18
sS'm'
(lp41723
I1
aI1
aI1
aI11
aI5
aI17
aI3
aI43
aI129
aI705
aI1701
aI3635
aI1201
aI12283
aI27443
aI54257
aI102281
aI211859
asS'd'
I20862
sa(dp41724
S'a'
I125648
sS's'
I18
sS'm'
(lp41725
I1
aI3
aI1
aI13
aI7
aI63
aI9
aI45
aI283
aI41
aI801
aI131
aI2797
aI13329
aI19011
aI21055
aI122965
aI7961
asS'd'
I20863
sa(dp41726
S'a'
I125651
sS's'
I18
sS'm'
(lp41727
I1
aI3
aI3
aI5
aI9
aI27
aI99
aI129
aI499
aI523
aI1939
aI3661
aI455
aI12601
aI11723
aI3727
aI32671
aI78251
asS'd'
I20864
sa(dp41728
S'a'
I125681
sS's'
I18
sS'm'
(lp41729
I1
aI3
aI7
aI5
aI9
aI57
aI63
aI47
aI49
aI745
aI945
aI2927
aI6659
aI1023
aI9991
aI55379
aI105295
aI259901
asS'd'
I20865
sa(dp41730
S'a'
I125719
sS's'
I18
sS'm'
(lp41731
I1
aI1
aI3
aI3
aI3
aI47
aI15
aI237
aI193
aI409
aI1165
aI3581
aI719
aI3049
aI14679
aI31559
aI7825
aI96083
asS'd'
I20866
sa(dp41732
S'a'
I125773
sS's'
I18
sS'm'
(lp41733
I1
aI1
aI7
aI11
aI9
aI55
aI29
aI123
aI163
aI415
aI2013
aI97
aI1471
aI1409
aI28867
aI50405
aI99417
aI57113
asS'd'
I20867
sa(dp41734
S'a'
I125774
sS's'
I18
sS'm'
(lp41735
I1
aI1
aI7
aI15
aI5
aI17
aI57
aI123
aI25
aI119
aI1699
aI1289
aI3139
aI7177
aI13465
aI33583
aI34517
aI182669
asS'd'
I20868
sa(dp41736
S'a'
I125779
sS's'
I18
sS'm'
(lp41737
I1
aI1
aI3
aI1
aI15
aI45
aI79
aI7
aI461
aI223
aI691
aI3071
aI6233
aI14997
aI4083
aI65391
aI60571
aI82929
asS'd'
I20869
sa(dp41738
S'a'
I125786
sS's'
I18
sS'm'
(lp41739
I1
aI3
aI5
aI1
aI25
aI39
aI89
aI105
aI497
aI685
aI1921
aI133
aI4849
aI8467
aI609
aI62183
aI123787
aI223025
asS'd'
I20870
sa(dp41740
S'a'
I125792
sS's'
I18
sS'm'
(lp41741
I1
aI1
aI1
aI13
aI3
aI57
aI117
aI241
aI501
aI107
aI1253
aI3097
aI603
aI10645
aI3395
aI13997
aI112527
aI208263
asS'd'
I20871
sa(dp41742
S'a'
I125822
sS's'
I18
sS'm'
(lp41743
I1
aI1
aI7
aI9
aI7
aI31
aI25
aI97
aI205
aI785
aI517
aI549
aI6841
aI7097
aI9635
aI17151
aI57135
aI105469
asS'd'
I20872
sa(dp41744
S'a'
I125828
sS's'
I18
sS'm'
(lp41745
I1
aI3
aI7
aI1
aI25
aI35
aI43
aI75
aI179
aI1023
aI1921
aI1529
aI2791
aI6747
aI9135
aI61801
aI46729
aI26821
asS'd'
I20873
sa(dp41746
S'a'
I125831
sS's'
I18
sS'm'
(lp41747
I1
aI1
aI3
aI1
aI31
aI49
aI107
aI219
aI285
aI501
aI1503
aI3103
aI5257
aI14561
aI31493
aI7753
aI34375
aI260357
asS'd'
I20874
sa(dp41748
S'a'
I125837
sS's'
I18
sS'm'
(lp41749
I1
aI3
aI7
aI13
aI15
aI7
aI75
aI57
aI329
aI67
aI1541
aI2445
aI3069
aI6723
aI10189
aI22913
aI110781
aI243765
asS'd'
I20875
sa(dp41750
S'a'
I125880
sS's'
I18
sS'm'
(lp41751
I1
aI1
aI5
aI3
aI5
aI45
aI47
aI39
aI493
aI787
aI1019
aI3933
aI535
aI1763
aI139
aI45967
aI123167
aI115019
asS'd'
I20876
sa(dp41752
S'a'
I125891
sS's'
I18
sS'm'
(lp41753
I1
aI3
aI1
aI3
aI23
aI59
aI93
aI139
aI349
aI973
aI1401
aI2109
aI701
aI461
aI19199
aI21733
aI80009
aI37239
asS'd'
I20877
sa(dp41754
S'a'
I125922
sS's'
I18
sS'm'
(lp41755
I1
aI1
aI3
aI5
aI3
aI39
aI117
aI77
aI257
aI117
aI1991
aI3371
aI509
aI3963
aI14579
aI62459
aI52281
aI99209
asS'd'
I20878
sa(dp41756
S'a'
I125928
sS's'
I18
sS'm'
(lp41757
I1
aI3
aI1
aI15
aI29
aI47
aI73
aI157
aI429
aI497
aI39
aI123
aI4851
aI12871
aI4567
aI29453
aI90777
aI188683
asS'd'
I20879
sa(dp41758
S'a'
I125931
sS's'
I18
sS'm'
(lp41759
I1
aI3
aI3
aI1
aI17
aI25
aI5
aI247
aI61
aI81
aI1555
aI2167
aI6003
aI15911
aI16023
aI7841
aI50731
aI229163
asS'd'
I20880
sa(dp41760
S'a'
I125941
sS's'
I18
sS'm'
(lp41761
I1
aI1
aI5
aI9
aI9
aI39
aI89
aI37
aI105
aI133
aI333
aI2863
aI7249
aI2355
aI9407
aI28145
aI25923
aI68827
asS'd'
I20881
sa(dp41762
S'a'
I125948
sS's'
I18
sS'm'
(lp41763
I1
aI3
aI7
aI13
aI15
aI55
aI11
aI93
aI197
aI447
aI1793
aI1793
aI4639
aI1869
aI1711
aI1439
aI15899
aI106931
asS'd'
I20882
sa(dp41764
S'a'
I125960
sS's'
I18
sS'm'
(lp41765
I1
aI1
aI5
aI13
aI23
aI11
aI67
aI155
aI511
aI363
aI1073
aI2249
aI719
aI11167
aI7953
aI21699
aI55735
aI47353
asS'd'
I20883
sa(dp41766
S'a'
I125973
sS's'
I18
sS'm'
(lp41767
I1
aI1
aI3
aI3
aI21
aI59
aI123
aI227
aI65
aI695
aI1769
aI4057
aI7071
aI1827
aI13639
aI45711
aI84019
aI96897
asS'd'
I20884
sa(dp41768
S'a'
I126014
sS's'
I18
sS'm'
(lp41769
I1
aI3
aI1
aI11
aI25
aI51
aI57
aI139
aI147
aI589
aI1565
aI511
aI3629
aI7329
aI9565
aI62893
aI85789
aI112047
asS'd'
I20885
sa(dp41770
S'a'
I126016
sS's'
I18
sS'm'
(lp41771
I1
aI3
aI5
aI5
aI23
aI11
aI93
aI75
aI25
aI947
aI1489
aI4081
aI3395
aI4655
aI27853
aI41299
aI89447
aI100971
asS'd'
I20886
sa(dp41772
S'a'
I126036
sS's'
I18
sS'm'
(lp41773
I1
aI3
aI5
aI11
aI25
aI63
aI93
aI227
aI411
aI49
aI403
aI437
aI1739
aI8453
aI31693
aI51439
aI89729
aI113405
asS'd'
I20887
sa(dp41774
S'a'
I126043
sS's'
I18
sS'm'
(lp41775
I1
aI3
aI7
aI9
aI29
aI13
aI51
aI155
aI403
aI627
aI173
aI2111
aI833
aI11453
aI17673
aI7121
aI52943
aI114835
asS'd'
I20888
sa(dp41776
S'a'
I126062
sS's'
I18
sS'm'
(lp41777
I1
aI1
aI5
aI13
aI11
aI11
aI105
aI101
aI309
aI577
aI1003
aI3667
aI3489
aI11807
aI6119
aI13773
aI89879
aI12391
asS'd'
I20889
sa(dp41778
S'a'
I126076
sS's'
I18
sS'm'
(lp41779
I1
aI3
aI7
aI9
aI17
aI11
aI111
aI239
aI225
aI723
aI933
aI3353
aI2003
aI5273
aI207
aI38539
aI82539
aI209781
asS'd'
I20890
sa(dp41780
S'a'
I126116
sS's'
I18
sS'm'
(lp41781
I1
aI1
aI7
aI13
aI15
aI53
aI69
aI105
aI155
aI445
aI353
aI617
aI5625
aI13439
aI29223
aI60439
aI119635
aI49643
asS'd'
I20891
sa(dp41782
S'a'
I126125
sS's'
I18
sS'm'
(lp41783
I1
aI1
aI3
aI11
aI13
aI25
aI107
aI27
aI109
aI313
aI1721
aI2647
aI1861
aI10631
aI17131
aI31365
aI65319
aI102905
asS'd'
I20892
sa(dp41784
S'a'
I126145
sS's'
I18
sS'm'
(lp41785
I1
aI1
aI1
aI7
aI21
aI29
aI39
aI167
aI341
aI115
aI1523
aI2209
aI95
aI4399
aI3881
aI38875
aI107691
aI132471
asS'd'
I20893
sa(dp41786
S'a'
I126152
sS's'
I18
sS'm'
(lp41787
I1
aI1
aI1
aI13
aI15
aI61
aI125
aI23
aI301
aI407
aI1497
aI3731
aI7013
aI5405
aI31233
aI51701
aI45619
aI107407
asS'd'
I20894
sa(dp41788
S'a'
I126175
sS's'
I18
sS'm'
(lp41789
I1
aI3
aI5
aI5
aI31
aI27
aI77
aI21
aI339
aI1013
aI371
aI19
aI5733
aI2177
aI15547
aI27595
aI6805
aI172695
asS'd'
I20895
sa(dp41790
S'a'
I126188
sS's'
I18
sS'm'
(lp41791
I1
aI3
aI5
aI11
aI5
aI15
aI71
aI23
aI441
aI169
aI1715
aI437
aI1791
aI293
aI13441
aI11225
aI119119
aI223035
asS'd'
I20896
sa(dp41792
S'a'
I126199
sS's'
I18
sS'm'
(lp41793
I1
aI3
aI1
aI5
aI11
aI15
aI95
aI127
aI433
aI789
aI899
aI2591
aI2339
aI8237
aI20765
aI32897
aI51511
aI58437
asS'd'
I20897
sa(dp41794
S'a'
I126205
sS's'
I18
sS'm'
(lp41795
I1
aI3
aI5
aI9
aI11
aI57
aI103
aI5
aI401
aI51
aI1813
aI923
aI1983
aI1853
aI21913
aI3051
aI56309
aI19423
asS'd'
I20898
sa(dp41796
S'a'
I126213
sS's'
I18
sS'm'
(lp41797
I1
aI1
aI3
aI15
aI3
aI3
aI41
aI5
aI231
aI35
aI391
aI185
aI7585
aI1005
aI20311
aI11193
aI18275
aI114131
asS'd'
I20899
sa(dp41798
S'a'
I126220
sS's'
I18
sS'm'
(lp41799
I1
aI1
aI3
aI9
aI13
aI13
aI115
aI203
aI223
aI575
aI459
aI1839
aI3949
aI16027
aI23137
aI13723
aI19195
aI249337
asS'd'
I20900
sa(dp41800
S'a'
I126286
sS's'
I18
sS'm'
(lp41801
I1
aI1
aI1
aI9
aI21
aI51
aI39
aI245
aI187
aI609
aI319
aI2927
aI3625
aI10789
aI31291
aI45557
aI45935
aI132447
asS'd'
I20901
sa(dp41802
S'a'
I126294
sS's'
I18
sS'm'
(lp41803
I1
aI3
aI3
aI9
aI25
aI37
aI47
aI55
aI219
aI409
aI1927
aI553
aI3953
aI6209
aI11807
aI11133
aI48047
aI132437
asS'd'
I20902
sa(dp41804
S'a'
I126300
sS's'
I18
sS'm'
(lp41805
I1
aI3
aI1
aI13
aI23
aI31
aI109
aI167
aI281
aI179
aI233
aI1603
aI7391
aI9091
aI27021
aI31213
aI13093
aI86017
asS'd'
I20903
sa(dp41806
S'a'
I126322
sS's'
I18
sS'm'
(lp41807
I1
aI3
aI1
aI13
aI13
aI15
aI7
aI29
aI409
aI59
aI505
aI1307
aI6247
aI6055
aI5531
aI59727
aI58069
aI84049
asS'd'
I20904
sa(dp41808
S'a'
I126324
sS's'
I18
sS'm'
(lp41809
I1
aI1
aI5
aI7
aI19
aI7
aI103
aI245
aI133
aI609
aI1087
aI2365
aI3341
aI1689
aI18841
aI19625
aI47413
aI63445
asS'd'
I20905
sa(dp41810
S'a'
I126343
sS's'
I18
sS'm'
(lp41811
I1
aI1
aI7
aI3
aI11
aI27
aI79
aI75
aI97
aI355
aI493
aI2035
aI3413
aI11835
aI9157
aI51173
aI1
aI71797
asS'd'
I20906
sa(dp41812
S'a'
I126344
sS's'
I18
sS'm'
(lp41813
I1
aI1
aI5
aI5
aI29
aI1
aI43
aI135
aI473
aI329
aI1197
aI1693
aI3823
aI7723
aI24771
aI22349
aI94383
aI41461
asS'd'
I20907
sa(dp41814
S'a'
I126430
sS's'
I18
sS'm'
(lp41815
I1
aI3
aI3
aI15
aI25
aI9
aI79
aI213
aI317
aI615
aI541
aI441
aI505
aI13665
aI3691
aI17825
aI49303
aI91783
asS'd'
I20908
sa(dp41816
S'a'
I126433
sS's'
I18
sS'm'
(lp41817
I1
aI3
aI1
aI11
aI27
aI33
aI65
aI31
aI469
aI799
aI1251
aI3357
aI5239
aI5651
aI13317
aI28553
aI64225
aI9805
asS'd'
I20909
sa(dp41818
S'a'
I126448
sS's'
I18
sS'm'
(lp41819
I1
aI1
aI7
aI13
aI9
aI49
aI77
aI43
aI363
aI719
aI1943
aI1285
aI1587
aI1047
aI29419
aI24025
aI89901
aI229095
asS'd'
I20910
sa(dp41820
S'a'
I126451
sS's'
I18
sS'm'
(lp41821
I1
aI1
aI1
aI15
aI19
aI57
aI33
aI243
aI111
aI183
aI497
aI603
aI923
aI1957
aI6493
aI11833
aI7331
aI229975
asS'd'
I20911
sa(dp41822
S'a'
I126457
sS's'
I18
sS'm'
(lp41823
I1
aI3
aI1
aI1
aI13
aI43
aI31
aI25
aI169
aI303
aI69
aI723
aI1745
aI1025
aI14301
aI2523
aI111887
aI179519
asS'd'
I20912
sa(dp41824
S'a'
I126458
sS's'
I18
sS'm'
(lp41825
I1
aI3
aI1
aI5
aI31
aI55
aI11
aI103
aI391
aI881
aI1885
aI3923
aI7507
aI377
aI29331
aI32167
aI56915
aI44211
asS'd'
I20913
sa(dp41826
S'a'
I126463
sS's'
I18
sS'm'
(lp41827
I1
aI1
aI7
aI1
aI27
aI55
aI33
aI141
aI393
aI73
aI701
aI3173
aI973
aI15553
aI10219
aI51441
aI55201
aI131055
asS'd'
I20914
sa(dp41828
S'a'
I126470
sS's'
I18
sS'm'
(lp41829
I1
aI1
aI3
aI9
aI1
aI31
aI115
aI85
aI173
aI227
aI163
aI157
aI5569
aI8291
aI27163
aI7581
aI8699
aI104523
asS'd'
I20915
sa(dp41830
S'a'
I126484
sS's'
I18
sS'm'
(lp41831
I1
aI1
aI1
aI5
aI5
aI19
aI11
aI55
aI217
aI571
aI1001
aI945
aI6237
aI1993
aI11809
aI63893
aI60081
aI102997
asS'd'
I20916
sa(dp41832
S'a'
I126518
sS's'
I18
sS'm'
(lp41833
I1
aI3
aI3
aI15
aI15
aI7
aI51
aI161
aI147
aI263
aI1701
aI1079
aI3027
aI11779
aI24885
aI16127
aI68985
aI162975
asS'd'
I20917
sa(dp41834
S'a'
I126521
sS's'
I18
sS'm'
(lp41835
I1
aI3
aI1
aI9
aI25
aI1
aI47
aI107
aI149
aI997
aI1779
aI2905
aI4951
aI10345
aI31059
aI63831
aI117219
aI251935
asS'd'
I20918
sa(dp41836
S'a'
I126547
sS's'
I18
sS'm'
(lp41837
I1
aI3
aI3
aI9
aI5
aI23
aI83
aI95
aI399
aI343
aI1597
aI1733
aI5959
aI9685
aI4721
aI59109
aI113633
aI80365
asS'd'
I20919
sa(dp41838
S'a'
I126556
sS's'
I18
sS'm'
(lp41839
I1
aI1
aI7
aI11
aI31
aI43
aI71
aI133
aI305
aI529
aI645
aI3095
aI6273
aI4019
aI14433
aI41609
aI64093
aI79051
asS'd'
I20920
sa(dp41840
S'a'
I126583
sS's'
I18
sS'm'
(lp41841
I1
aI3
aI5
aI11
aI1
aI63
aI123
aI25
aI245
aI583
aI1013
aI3275
aI2997
aI13021
aI27515
aI16233
aI113093
aI249101
asS'd'
I20921
sa(dp41842
S'a'
I126584
sS's'
I18
sS'm'
(lp41843
I1
aI1
aI1
aI9
aI3
aI59
aI113
aI155
aI125
aI423
aI259
aI1559
aI3745
aI9105
aI27673
aI36601
aI36117
aI47953
asS'd'
I20922
sa(dp41844
S'a'
I126633
sS's'
I18
sS'm'
(lp41845
I1
aI3
aI3
aI7
aI19
aI41
aI55
aI87
aI53
aI801
aI661
aI329
aI3391
aI7581
aI25487
aI25751
aI120171
aI35953
asS'd'
I20923
sa(dp41846
S'a'
I126673
sS's'
I18
sS'm'
(lp41847
I1
aI1
aI7
aI15
aI31
aI57
aI49
aI179
aI147
aI139
aI957
aI289
aI4321
aI8747
aI53
aI46003
aI40219
aI96855
asS'd'
I20924
sa(dp41848
S'a'
I126680
sS's'
I18
sS'm'
(lp41849
I1
aI1
aI7
aI9
aI29
aI49
aI71
aI101
aI389
aI793
aI1355
aI3263
aI6331
aI4869
aI28479
aI8335
aI74653
aI8519
asS'd'
I20925
sa(dp41850
S'a'
I126686
sS's'
I18
sS'm'
(lp41851
I1
aI1
aI1
aI13
aI1
aI19
aI31
aI161
aI261
aI679
aI1115
aI985
aI2855
aI4395
aI15087
aI18593
aI98535
aI52537
asS'd'
I20926
sa(dp41852
S'a'
I126689
sS's'
I18
sS'm'
(lp41853
I1
aI1
aI3
aI11
aI15
aI51
aI79
aI7
aI75
aI75
aI753
aI2637
aI7193
aI7961
aI21411
aI24273
aI7543
aI6277
asS'd'
I20927
sa(dp41854
S'a'
I126702
sS's'
I18
sS'm'
(lp41855
I1
aI3
aI1
aI5
aI5
aI51
aI67
aI191
aI201
aI777
aI587
aI1439
aI1027
aI3759
aI31141
aI42159
aI58475
aI7355
asS'd'
I20928
sa(dp41856
S'a'
I126710
sS's'
I18
sS'm'
(lp41857
I1
aI3
aI5
aI3
aI15
aI37
aI11
aI251
aI53
aI799
aI739
aI2225
aI6985
aI9183
aI12341
aI29963
aI44101
aI23889
asS'd'
I20929
sa(dp41858
S'a'
I126722
sS's'
I18
sS'm'
(lp41859
I1
aI1
aI3
aI3
aI5
aI33
aI81
aI223
aI89
aI531
aI301
aI305
aI2401
aI4015
aI18607
aI65041
aI82447
aI228487
asS'd'
I20930
sa(dp41860
S'a'
I126787
sS's'
I18
sS'm'
(lp41861
I1
aI3
aI1
aI7
aI15
aI29
aI71
aI247
aI445
aI1005
aI1229
aI3897
aI899
aI11175
aI6349
aI29145
aI103153
aI90275
asS'd'
I20931
sa(dp41862
S'a'
I126789
sS's'
I18
sS'm'
(lp41863
I1
aI3
aI3
aI7
aI15
aI7
aI75
aI13
aI417
aI719
aI121
aI1345
aI3737
aI4119
aI15259
aI33579
aI57727
aI111517
asS'd'
I20932
sa(dp41864
S'a'
I126794
sS's'
I18
sS'm'
(lp41865
I1
aI3
aI7
aI3
aI31
aI11
aI49
aI61
aI405
aI741
aI1607
aI1561
aI4655
aI9775
aI14349
aI27431
aI91791
aI228607
asS'd'
I20933
sa(dp41866
S'a'
I126801
sS's'
I18
sS'm'
(lp41867
I1
aI3
aI5
aI13
aI17
aI27
aI93
aI153
aI99
aI683
aI219
aI3783
aI6963
aI1633
aI6621
aI8133
aI5111
aI57333
asS'd'
I20934
sa(dp41868
S'a'
I126857
sS's'
I18
sS'm'
(lp41869
I1
aI3
aI3
aI3
aI25
aI29
aI75
aI155
aI159
aI25
aI637
aI3053
aI4737
aI5831
aI9651
aI45331
aI100407
aI188607
asS'd'
I20935
sa(dp41870
S'a'
I126891
sS's'
I18
sS'm'
(lp41871
I1
aI3
aI1
aI9
aI21
aI55
aI123
aI27
aI509
aI227
aI1569
aI1379
aI7137
aI11749
aI14257
aI38349
aI459
aI54873
asS'd'
I20936
sa(dp41872
S'a'
I126896
sS's'
I18
sS'm'
(lp41873
I1
aI1
aI7
aI15
aI29
aI35
aI81
aI79
aI237
aI155
aI1551
aI343
aI5127
aI3233
aI3691
aI59917
aI7367
aI181979
asS'd'
I20937
sa(dp41874
S'a'
I126902
sS's'
I18
sS'm'
(lp41875
I1
aI3
aI1
aI15
aI1
aI57
aI111
aI121
aI375
aI635
aI1529
aI635
aI2337
aI6553
aI22067
aI36047
aI80099
aI116411
asS'd'
I20938
sa(dp41876
S'a'
I126911
sS's'
I18
sS'm'
(lp41877
I1
aI1
aI3
aI3
aI29
aI27
aI19
aI3
aI69
aI197
aI1829
aI1907
aI5901
aI12651
aI3295
aI7805
aI57871
aI47571
asS'd'
I20939
sa(dp41878
S'a'
I126919
sS's'
I18
sS'm'
(lp41879
I1
aI1
aI1
aI13
aI21
aI5
aI115
aI51
aI203
aI991
aI1731
aI1101
aI1607
aI14323
aI12233
aI48047
aI33969
aI147621
asS'd'
I20940
sa(dp41880
S'a'
I126943
sS's'
I18
sS'm'
(lp41881
I1
aI3
aI3
aI1
aI25
aI55
aI41
aI31
aI101
aI551
aI1519
aI1915
aI6961
aI13919
aI15339
aI13141
aI107625
aI9247
asS'd'
I20941
sa(dp41882
S'a'
I126947
sS's'
I18
sS'm'
(lp41883
I1
aI1
aI7
aI11
aI17
aI25
aI87
aI175
aI341
aI445
aI1813
aI2995
aI3217
aI6015
aI1637
aI65243
aI72743
aI248715
asS'd'
I20942
sa(dp41884
S'a'
I126949
sS's'
I18
sS'm'
(lp41885
I1
aI1
aI5
aI13
aI11
aI1
aI61
aI139
aI233
aI167
aI1363
aI1991
aI7999
aI16289
aI25595
aI8915
aI32205
aI169963
asS'd'
I20943
sa(dp41886
S'a'
I126953
sS's'
I18
sS'm'
(lp41887
I1
aI1
aI1
aI13
aI31
aI33
aI67
aI57
aI121
aI477
aI755
aI2035
aI3683
aI16205
aI5511
aI25615
aI5169
aI128843
asS'd'
I20944
sa(dp41888
S'a'
I126971
sS's'
I18
sS'm'
(lp41889
I1
aI3
aI5
aI15
aI31
aI1
aI89
aI29
aI493
aI379
aI1627
aI491
aI2503
aI8105
aI30275
aI27379
aI43905
aI46397
asS'd'
I20945
sa(dp41890
S'a'
I126976
sS's'
I18
sS'm'
(lp41891
I1
aI1
aI7
aI9
aI31
aI41
aI127
aI121
aI61
aI561
aI223
aI3231
aI7321
aI3683
aI15455
aI8019
aI116739
aI96557
asS'd'
I20946
sa(dp41892
S'a'
I126982
sS's'
I18
sS'm'
(lp41893
I1
aI1
aI5
aI9
aI19
aI41
aI25
aI65
aI111
aI535
aI611
aI1631
aI4251
aI107
aI19787
aI40749
aI65701
aI48749
asS'd'
I20947
sa(dp41894
S'a'
I127009
sS's'
I18
sS'm'
(lp41895
I1
aI3
aI3
aI11
aI27
aI49
aI75
aI93
aI15
aI937
aI1517
aI1577
aI7485
aI8713
aI15979
aI13799
aI103057
aI144799
asS'd'
I20948
sa(dp41896
S'a'
I127021
sS's'
I18
sS'm'
(lp41897
I1
aI1
aI7
aI7
aI7
aI29
aI3
aI63
aI177
aI781
aI2001
aI315
aI6703
aI8055
aI19081
aI33641
aI44279
aI87597
asS'd'
I20949
sa(dp41898
S'a'
I127027
sS's'
I18
sS'm'
(lp41899
I1
aI1
aI7
aI11
aI29
aI53
aI83
aI223
aI275
aI951
aI1883
aI2447
aI1815
aI9313
aI9247
aI17185
aI8143
aI135247
asS'd'
I20950
sa(dp41900
S'a'
I127039
sS's'
I18
sS'm'
(lp41901
I1
aI1
aI5
aI13
aI21
aI47
aI87
aI55
aI169
aI95
aI777
aI2787
aI7227
aI11373
aI16707
aI28237
aI30789
aI64589
asS'd'
I20951
sa(dp41902
S'a'
I127041
sS's'
I18
sS'm'
(lp41903
I1
aI1
aI1
aI11
aI15
aI41
aI1
aI55
aI337
aI493
aI1379
aI2505
aI6831
aI10955
aI1875
aI21821
aI54101
aI9379
asS'd'
I20952
sa(dp41904
S'a'
I127044
sS's'
I18
sS'm'
(lp41905
I1
aI3
aI5
aI15
aI27
aI9
aI121
aI49
aI439
aI781
aI1457
aI1341
aI7433
aI5879
aI13039
aI24001
aI64059
aI157077
asS'd'
I20953
sa(dp41906
S'a'
I127059
sS's'
I18
sS'm'
(lp41907
I1
aI1
aI1
aI9
aI19
aI55
aI89
aI37
aI255
aI345
aI215
aI4067
aI8151
aI14253
aI12121
aI3637
aI29185
aI60643
asS'd'
I20954
sa(dp41908
S'a'
I127071
sS's'
I18
sS'm'
(lp41909
I1
aI3
aI1
aI7
aI31
aI39
aI71
aI29
aI71
aI83
aI1249
aI871
aI8037
aI1001
aI25245
aI26651
aI34509
aI123607
asS'd'
I20955
sa(dp41910
S'a'
I127075
sS's'
I18
sS'm'
(lp41911
I1
aI1
aI3
aI11
aI13
aI21
aI15
aI171
aI255
aI373
aI429
aI2179
aI4431
aI16087
aI17949
aI16307
aI129877
aI186495
asS'd'
I20956
sa(dp41912
S'a'
I127126
sS's'
I18
sS'm'
(lp41913
I1
aI3
aI3
aI9
aI17
aI45
aI75
aI175
aI3
aI403
aI215
aI1781
aI7875
aI14113
aI6967
aI65263
aI125885
aI232983
asS'd'
I20957
sa(dp41914
S'a'
I127129
sS's'
I18
sS'm'
(lp41915
I1
aI1
aI7
aI9
aI21
aI57
aI73
aI105
aI163
aI583
aI587
aI2743
aI2199
aI5187
aI5571
aI56399
aI797
aI192405
asS'd'
I20958
sa(dp41916
S'a'
I127145
sS's'
I18
sS'm'
(lp41917
I1
aI1
aI1
aI3
aI29
aI27
aI71
aI145
aI11
aI455
aI1505
aI2789
aI4083
aI12345
aI14785
aI4981
aI95121
aI134977
asS'd'
I20959
sa(dp41918
S'a'
I127154
sS's'
I18
sS'm'
(lp41919
I1
aI1
aI1
aI1
aI1
aI27
aI1
aI145
aI473
aI483
aI83
aI3009
aI7241
aI13633
aI15071
aI30767
aI128103
aI94727
asS'd'
I20960
sa(dp41920
S'a'
I127165
sS's'
I18
sS'm'
(lp41921
I1
aI3
aI5
aI15
aI17
aI51
aI71
aI21
aI237
aI65
aI901
aI3365
aI7831
aI3027
aI8751
aI14435
aI79445
aI172587
asS'd'
I20961
sa(dp41922
S'a'
I127171
sS's'
I18
sS'm'
(lp41923
I1
aI1
aI5
aI1
aI9
aI49
aI49
aI31
aI395
aI339
aI343
aI1813
aI2607
aI9347
aI11239
aI6761
aI127623
aI43459
asS'd'
I20962
sa(dp41924
S'a'
I127197
sS's'
I18
sS'm'
(lp41925
I1
aI1
aI1
aI1
aI13
aI23
aI71
aI131
aI225
aI229
aI117
aI889
aI8145
aI5953
aI10679
aI38687
aI80029
aI63689
asS'd'
I20963
sa(dp41926
S'a'
I127198
sS's'
I18
sS'm'
(lp41927
I1
aI1
aI5
aI1
aI29
aI1
aI87
aI181
aI441
aI353
aI257
aI335
aI203
aI10897
aI24085
aI26967
aI62573
aI170285
asS'd'
I20964
sa(dp41928
S'a'
I127207
sS's'
I18
sS'm'
(lp41929
I1
aI1
aI7
aI3
aI3
aI39
aI47
aI135
aI353
aI977
aI89
aI259
aI6411
aI5511
aI10697
aI57623
aI27367
aI108451
asS'd'
I20965
sa(dp41930
S'a'
I127226
sS's'
I18
sS'm'
(lp41931
I1
aI3
aI3
aI11
aI9
aI57
aI95
aI211
aI237
aI281
aI1703
aI2107
aI2179
aI3411
aI32621
aI5387
aI29971
aI102889
asS'd'
I20966
sa(dp41932
S'a'
I127240
sS's'
I18
sS'm'
(lp41933
I1
aI1
aI1
aI13
aI27
aI49
aI47
aI49
aI413
aI985
aI649
aI1245
aI807
aI13637
aI21741
aI32565
aI80135
aI127971
asS'd'
I20967
sa(dp41934
S'a'
I127254
sS's'
I18
sS'm'
(lp41935
I1
aI3
aI7
aI13
aI3
aI19
aI57
aI97
aI493
aI597
aI135
aI1689
aI5011
aI4579
aI6093
aI28341
aI37279
aI142197
asS'd'
I20968
sa(dp41936
S'a'
I127269
sS's'
I18
sS'm'
(lp41937
I1
aI3
aI1
aI15
aI15
aI31
aI3
aI89
aI327
aI107
aI827
aI1111
aI261
aI6211
aI4359
aI38553
aI43297
aI75057
asS'd'
I20969
sa(dp41938
S'a'
I127305
sS's'
I18
sS'm'
(lp41939
I1
aI1
aI1
aI9
aI19
aI19
aI53
aI195
aI141
aI297
aI141
aI3859
aI4173
aI12243
aI31399
aI6353
aI110505
aI172219
asS'd'
I20970
sa(dp41940
S'a'
I127323
sS's'
I18
sS'm'
(lp41941
I1
aI3
aI3
aI9
aI31
aI51
aI59
aI53
aI55
aI723
aI1575
aI3399
aI8057
aI12317
aI8393
aI1719
aI96987
aI228955
asS'd'
I20971
sa(dp41942
S'a'
I127342
sS's'
I18
sS'm'
(lp41943
I1
aI1
aI7
aI11
aI19
aI59
aI41
aI9
aI217
aI267
aI629
aI2977
aI4515
aI463
aI31773
aI61765
aI78827
aI51331
asS'd'
I20972
sa(dp41944
S'a'
I127347
sS's'
I18
sS'm'
(lp41945
I1
aI3
aI3
aI13
aI9
aI55
aI51
aI177
aI183
aI431
aI555
aI3573
aI7977
aI3067
aI21111
aI12971
aI78283
aI260721
asS'd'
I20973
sa(dp41946
S'a'
I127363
sS's'
I18
sS'm'
(lp41947
I1
aI1
aI1
aI11
aI27
aI5
aI89
aI69
aI435
aI199
aI221
aI1017
aI7703
aI7469
aI7755
aI46319
aI37941
aI55285
asS'd'
I20974
sa(dp41948
S'a'
I127394
sS's'
I18
sS'm'
(lp41949
I1
aI1
aI7
aI13
aI19
aI55
aI53
aI207
aI367
aI177
aI1483
aI2857
aI3753
aI5493
aI13349
aI14033
aI7933
aI93457
asS'd'
I20975
sa(dp41950
S'a'
I127426
sS's'
I18
sS'm'
(lp41951
I1
aI1
aI7
aI3
aI27
aI35
aI19
aI223
aI341
aI137
aI1195
aI1263
aI5937
aI13517
aI55
aI6391
aI106173
aI176503
asS'd'
I20976
sa(dp41952
S'a'
I127432
sS's'
I18
sS'm'
(lp41953
I1
aI3
aI7
aI11
aI23
aI25
aI37
aI103
aI351
aI945
aI1205
aI2543
aI3875
aI155
aI27777
aI36647
aI47979
aI25113
asS'd'
I20977
sa(dp41954
S'a'
I127440
sS's'
I18
sS'm'
(lp41955
I1
aI1
aI3
aI15
aI25
aI59
aI79
aI39
aI17
aI553
aI1119
aI3353
aI2619
aI3851
aI5945
aI47501
aI17369
aI89355
asS'd'
I20978
sa(dp41956
S'a'
I127462
sS's'
I18
sS'm'
(lp41957
I1
aI1
aI1
aI13
aI9
aI55
aI13
aI173
aI207
aI925
aI1855
aI1871
aI7851
aI1361
aI20117
aI51677
aI77703
aI51309
asS'd'
I20979
sa(dp41958
S'a'
I127468
sS's'
I18
sS'm'
(lp41959
I1
aI3
aI1
aI1
aI31
aI57
aI3
aI25
aI329
aI927
aI1683
aI1447
aI6853
aI103
aI9549
aI21393
aI415
aI122749
asS'd'
I20980
sa(dp41960
S'a'
I127473
sS's'
I18
sS'm'
(lp41961
I1
aI1
aI5
aI5
aI31
aI61
aI31
aI213
aI85
aI531
aI931
aI999
aI1189
aI5189
aI15127
aI47799
aI70769
aI81901
asS'd'
I20981
sa(dp41962
S'a'
I127486
sS's'
I18
sS'm'
(lp41963
I1
aI1
aI3
aI1
aI5
aI59
aI89
aI53
aI105
aI761
aI313
aI3013
aI4093
aI9595
aI4287
aI51505
aI20095
aI232933
asS'd'
I20982
sa(dp41964
S'a'
I127502
sS's'
I18
sS'm'
(lp41965
I1
aI1
aI7
aI7
aI23
aI9
aI41
aI29
aI399
aI395
aI759
aI2541
aI2373
aI15365
aI12083
aI49579
aI34401
aI168121
asS'd'
I20983
sa(dp41966
S'a'
I127510
sS's'
I18
sS'm'
(lp41967
I1
aI1
aI3
aI1
aI7
aI23
aI37
aI183
aI205
aI377
aI1081
aI1081
aI7767
aI363
aI14571
aI16265
aI18267
aI102155
asS'd'
I20984
sa(dp41968
S'a'
I127523
sS's'
I18
sS'm'
(lp41969
I1
aI3
aI3
aI15
aI19
aI11
aI59
aI59
aI465
aI437
aI965
aI3707
aI3505
aI14785
aI23605
aI12505
aI130607
aI40693
asS'd'
I20985
sa(dp41970
S'a'
I127543
sS's'
I18
sS'm'
(lp41971
I1
aI1
aI3
aI13
aI5
aI15
aI91
aI33
aI235
aI215
aI1997
aI2035
aI7407
aI3203
aI27143
aI14007
aI96411
aI593
asS'd'
I20986
sa(dp41972
S'a'
I127567
sS's'
I18
sS'm'
(lp41973
I1
aI3
aI7
aI1
aI19
aI51
aI1
aI69
aI489
aI629
aI1731
aI393
aI6807
aI10521
aI23971
aI45649
aI105183
aI207351
asS'd'
I20987
sa(dp41974
S'a'
I127570
sS's'
I18
sS'm'
(lp41975
I1
aI3
aI5
aI3
aI5
aI41
aI89
aI141
aI469
aI177
aI109
aI2439
aI7155
aI2083
aI31993
aI13933
aI100557
aI137255
asS'd'
I20988
sa(dp41976
S'a'
I127585
sS's'
I18
sS'm'
(lp41977
I1
aI1
aI7
aI15
aI21
aI45
aI41
aI197
aI365
aI177
aI61
aI811
aI2535
aI5219
aI3689
aI53129
aI42063
aI60759
asS'd'
I20989
sa(dp41978
S'a'
I127588
sS's'
I18
sS'm'
(lp41979
I1
aI3
aI5
aI1
aI23
aI7
aI19
aI193
aI253
aI793
aI539
aI3747
aI2611
aI16211
aI17199
aI14875
aI95377
aI6999
asS'd'
I20990
sa(dp41980
S'a'
I127619
sS's'
I18
sS'm'
(lp41981
I1
aI1
aI3
aI5
aI9
aI5
aI9
aI129
aI217
aI473
aI151
aI3053
aI6981
aI8075
aI32121
aI31995
aI41271
aI208927
asS'd'
I20991
sa(dp41982
S'a'
I127621
sS's'
I18
sS'm'
(lp41983
I1
aI1
aI5
aI9
aI9
aI9
aI89
aI139
aI381
aI937
aI1937
aI1879
aI8191
aI2237
aI25629
aI51471
aI87639
aI173697
asS'd'
I20992
sa(dp41984
S'a'
I127626
sS's'
I18
sS'm'
(lp41985
I1
aI1
aI5
aI9
aI17
aI35
aI81
aI223
aI161
aI315
aI139
aI2597
aI2599
aI16191
aI2567
aI54947
aI8603
aI121589
asS'd'
I20993
sa(dp41986
S'a'
I127628
sS's'
I18
sS'm'
(lp41987
I1
aI1
aI5
aI5
aI9
aI7
aI33
aI49
aI49
aI723
aI1013
aI1055
aI4025
aI1471
aI30081
aI17475
aI127931
aI63723
asS'd'
I20994
sa(dp41988
S'a'
I127640
sS's'
I18
sS'm'
(lp41989
I1
aI3
aI7
aI7
aI9
aI49
aI107
aI17
aI335
aI119
aI1959
aI3613
aI8129
aI11033
aI12197
aI23803
aI112595
aI131655
asS'd'
I20995
sa(dp41990
S'a'
I127662
sS's'
I18
sS'm'
(lp41991
I1
aI3
aI5
aI11
aI3
aI45
aI91
aI17
aI181
aI1005
aI985
aI3045
aI853
aI8181
aI5517
aI48515
aI16225
aI237151
asS'd'
I20996
sa(dp41992
S'a'
I127684
sS's'
I18
sS'm'
(lp41993
I1
aI1
aI3
aI1
aI3
aI63
aI35
aI135
aI61
aI383
aI1233
aI675
aI151
aI2157
aI18711
aI37113
aI40353
aI61783
asS'd'
I20997
sa(dp41994
S'a'
I127687
sS's'
I18
sS'm'
(lp41995
I1
aI3
aI5
aI5
aI29
aI3
aI105
aI11
aI351
aI761
aI165
aI911
aI6903
aI10111
aI1779
aI24601
aI3177
aI110301
asS'd'
I20998
sa(dp41996
S'a'
I127693
sS's'
I18
sS'm'
(lp41997
I1
aI1
aI3
aI15
aI25
aI19
aI73
aI237
aI263
aI161
aI731
aI3853
aI7705
aI14497
aI30799
aI32979
aI100729
aI21761
asS'd'
I20999
sa(dp41998
S'a'
I127696
sS's'
I18
sS'm'
(lp41999
I1
aI1
aI3
aI5
aI27
aI9
aI3
aI149
aI207
aI715
aI1435
aI2563
aI2451
aI7951
aI26313
aI55115
aI99423
aI231639
asS'd'
I21000
sa(dp42000
S'a'
I127708
sS's'
I18
sS'm'
(lp42001
I1
aI1
aI5
aI15
aI11
aI51
aI13
aI47
aI311
aI969
aI2013
aI357
aI4847
aI1831
aI2235
aI22779
aI32375
aI40893
asS'd'
I21001
sa(dp42002
S'a'
I127711
sS's'
I18
sS'm'
(lp42003
I1
aI1
aI3
aI9
aI21
aI45
aI11
aI99
aI275
aI849
aI443
aI1257
aI7855
aI9121
aI6549
aI20289
aI101337
aI13869
asS'd'
I21002
sa(dp42004
S'a'
I127722
sS's'
I18
sS'm'
(lp42005
I1
aI1
aI1
aI15
aI25
aI27
aI15
aI111
aI215
aI437
aI1923
aI1985
aI4603
aI15469
aI6667
aI17941
aI50433
aI152759
asS'd'
I21003
sa(dp42006
S'a'
I127732
sS's'
I18
sS'm'
(lp42007
I1
aI3
aI7
aI15
aI7
aI37
aI119
aI53
aI337
aI853
aI1785
aI3507
aI3743
aI14303
aI22757
aI5149
aI1539
aI227051
asS'd'
I21004
sa(dp42008
S'a'
I127747
sS's'
I18
sS'm'
(lp42009
I1
aI1
aI3
aI13
aI11
aI23
aI55
aI19
aI495
aI531
aI1021
aI3831
aI5993
aI15819
aI2121
aI52773
aI19775
aI94643
asS'd'
I21005
sa(dp42010
S'a'
I127767
sS's'
I18
sS'm'
(lp42011
I1
aI1
aI3
aI3
aI23
aI55
aI55
aI69
aI457
aI755
aI1187
aI3993
aI613
aI12691
aI1779
aI21251
aI2293
aI236725
asS'd'
I21006
sa(dp42012
S'a'
I127771
sS's'
I18
sS'm'
(lp42013
I1
aI1
aI5
aI9
aI23
aI27
aI61
aI125
aI113
aI99
aI503
aI699
aI6873
aI13141
aI10649
aI65209
aI21773
aI162749
asS'd'
I21007
sa(dp42014
S'a'
I127790
sS's'
I18
sS'm'
(lp42015
I1
aI1
aI1
aI11
aI15
aI27
aI111
aI227
aI493
aI361
aI1071
aI607
aI1409
aI9281
aI24515
aI26739
aI82421
aI30463
asS'd'
I21008
sa(dp42016
S'a'
I127816
sS's'
I18
sS'm'
(lp42017
I1
aI3
aI5
aI1
aI11
aI57
aI23
aI239
aI265
aI675
aI441
aI4031
aI5163
aI15729
aI2741
aI26037
aI32533
aI140645
asS'd'
I21009
sa(dp42018
S'a'
I127824
sS's'
I18
sS'm'
(lp42019
I1
aI3
aI3
aI7
aI3
aI45
aI105
aI135
aI493
aI579
aI1707
aI2933
aI1135
aI11891
aI3171
aI45401
aI24993
aI175681
asS'd'
I21010
sa(dp42020
S'a'
I127833
sS's'
I18
sS'm'
(lp42021
I1
aI1
aI1
aI11
aI11
aI3
aI67
aI213
aI483
aI9
aI1053
aI213
aI3205
aI8487
aI16093
aI7305
aI122591
aI31811
asS'd'
I21011
sa(dp42022
S'a'
I127869
sS's'
I18
sS'm'
(lp42023
I1
aI1
aI5
aI13
aI19
aI31
aI13
aI65
aI29
aI929
aI343
aI463
aI1885
aI13467
aI14997
aI22737
aI42869
aI128239
asS'd'
I21012
sa(dp42024
S'a'
I127910
sS's'
I18
sS'm'
(lp42025
I1
aI3
aI3
aI13
aI9
aI47
aI125
aI33
aI475
aI285
aI1901
aI2525
aI305
aI11587
aI27309
aI30037
aI70681
aI180425
asS'd'
I21013
sa(dp42026
S'a'
I127928
sS's'
I18
sS'm'
(lp42027
I1
aI3
aI5
aI15
aI9
aI37
aI45
aI149
aI19
aI135
aI555
aI4037
aI5173
aI12473
aI983
aI40923
aI28561
aI185941
asS'd'
I21014
sa(dp42028
S'a'
I127933
sS's'
I18
sS'm'
(lp42029
I1
aI3
aI3
aI9
aI23
aI35
aI35
aI151
aI113
aI885
aI1553
aI2233
aI351
aI4071
aI28127
aI26109
aI12299
aI163973
asS'd'
I21015
sa(dp42030
S'a'
I127963
sS's'
I18
sS'm'
(lp42031
I1
aI3
aI3
aI11
aI17
aI55
aI125
aI87
aI315
aI917
aI383
aI2397
aI1573
aI9255
aI10499
aI16051
aI99487
aI139415
asS'd'
I21016
sa(dp42032
S'a'
I127975
sS's'
I18
sS'm'
(lp42033
I1
aI1
aI3
aI1
aI29
aI21
aI101
aI153
aI5
aI705
aI1965
aI1447
aI8163
aI13547
aI25929
aI28569
aI57897
aI173229
asS'd'
I21017
sa(dp42034
S'a'
I128016
sS's'
I18
sS'm'
(lp42035
I1
aI1
aI7
aI15
aI3
aI37
aI113
aI213
aI495
aI935
aI529
aI2299
aI6901
aI1765
aI4255
aI14579
aI14175
aI112333
asS'd'
I21018
sa(dp42036
S'a'
I128031
sS's'
I18
sS'm'
(lp42037
I1
aI3
aI3
aI9
aI11
aI53
aI89
aI27
aI461
aI235
aI1525
aI3533
aI3061
aI4351
aI12847
aI21649
aI10843
aI60901
asS'd'
I21019
sa(dp42038
S'a'
I128059
sS's'
I18
sS'm'
(lp42039
I1
aI1
aI5
aI1
aI17
aI11
aI17
aI157
aI387
aI887
aI2017
aI3641
aI923
aI12659
aI19691
aI18657
aI3127
aI218819
asS'd'
I21020
sa(dp42040
S'a'
I128074
sS's'
I18
sS'm'
(lp42041
I1
aI3
aI1
aI9
aI5
aI3
aI49
aI215
aI379
aI765
aI1375
aI345
aI2285
aI8197
aI9531
aI6725
aI22475
aI203883
asS'd'
I21021
sa(dp42042
S'a'
I128107
sS's'
I18
sS'm'
(lp42043
I1
aI1
aI3
aI9
aI19
aI41
aI13
aI233
aI97
aI755
aI249
aI2011
aI5815
aI6317
aI4121
aI63637
aI43353
aI154753
asS'd'
I21022
sa(dp42044
S'a'
I128122
sS's'
I18
sS'm'
(lp42045
I1
aI3
aI7
aI15
aI31
aI25
aI93
aI197
aI455
aI979
aI1805
aI2619
aI803
aI5705
aI1679
aI29317
aI66477
aI159187
asS'd'
I21023
sa(dp42046
S'a'
I128157
sS's'
I18
sS'm'
(lp42047
I1
aI1
aI1
aI13
aI11
aI25
aI61
aI233
aI339
aI171
aI559
aI427
aI3239
aI8889
aI3711
aI19743
aI18099
aI49201
asS'd'
I21024
sa(dp42048
S'a'
I128161
sS's'
I18
sS'm'
(lp42049
I1
aI3
aI7
aI13
aI19
aI5
aI9
aI183
aI355
aI137
aI1767
aI1113
aI1149
aI5791
aI4099
aI37911
aI75945
aI115397
asS'd'
I21025
sa(dp42050
S'a'
I128200
sS's'
I18
sS'm'
(lp42051
I1
aI1
aI3
aI13
aI27
aI25
aI121
aI3
aI337
aI195
aI1841
aI2009
aI4181
aI3197
aI20275
aI42493
aI7495
aI24407
asS'd'
I21026
sa(dp42052
S'a'
I128213
sS's'
I18
sS'm'
(lp42053
I1
aI1
aI1
aI15
aI3
aI43
aI39
aI25
aI57
aI829
aI565
aI1977
aI4027
aI11053
aI13961
aI13965
aI4207
aI1663
asS'd'
I21027
sa(dp42054
S'a'
I128229
sS's'
I18
sS'm'
(lp42055
I1
aI3
aI5
aI9
aI11
aI7
aI107
aI205
aI479
aI961
aI1549
aI1701
aI6305
aI15419
aI23331
aI46443
aI55171
aI235109
asS'd'
I21028
sa(dp42056
S'a'
I128233
sS's'
I18
sS'm'
(lp42057
I1
aI3
aI5
aI5
aI19
aI3
aI39
aI211
aI429
aI363
aI765
aI283
aI2469
aI1947
aI10481
aI1969
aI95545
aI187671
asS'd'
I21029
sa(dp42058
S'a'
I128241
sS's'
I18
sS'm'
(lp42059
I1
aI1
aI1
aI1
aI11
aI55
aI47
aI121
aI251
aI63
aI767
aI3673
aI3233
aI14865
aI25713
aI48443
aI79139
aI225021
asS'd'
I21030
sa(dp42060
S'a'
I128259
sS's'
I18
sS'm'
(lp42061
I1
aI1
aI3
aI11
aI25
aI35
aI57
aI103
aI385
aI155
aI173
aI4023
aI489
aI1733
aI14423
aI61843
aI24793
aI9871
asS'd'
I21031
sa(dp42062
S'a'
I128295
sS's'
I18
sS'm'
(lp42063
I1
aI3
aI1
aI15
aI9
aI29
aI99
aI187
aI471
aI877
aI1321
aI2489
aI7439
aI4259
aI32703
aI1459
aI42093
aI261097
asS'd'
I21032
sa(dp42064
S'a'
I128299
sS's'
I18
sS'm'
(lp42065
I1
aI1
aI3
aI11
aI9
aI25
aI113
aI251
aI337
aI405
aI847
aI2451
aI5649
aI3449
aI11703
aI18271
aI108005
aI208789
asS'd'
I21033
sa(dp42066
S'a'
I128313
sS's'
I18
sS'm'
(lp42067
I1
aI3
aI3
aI7
aI13
aI53
aI61
aI251
aI461
aI461
aI1557
aI1215
aI6731
aI13349
aI21003
aI11573
aI66751
aI79733
asS'd'
I21034
sa(dp42068
S'a'
I128345
sS's'
I18
sS'm'
(lp42069
I1
aI1
aI5
aI1
aI23
aI49
aI101
aI175
aI251
aI577
aI1667
aI2561
aI6545
aI16305
aI18457
aI65067
aI35843
aI123445
asS'd'
I21035
sa(dp42070
S'a'
I128362
sS's'
I18
sS'm'
(lp42071
I1
aI1
aI5
aI3
aI7
aI9
aI61
aI107
aI395
aI137
aI559
aI2315
aI2559
aI11929
aI4843
aI41661
aI61361
aI146163
asS'd'
I21036
sa(dp42072
S'a'
I128375
sS's'
I18
sS'm'
(lp42073
I1
aI3
aI1
aI5
aI1
aI3
aI43
aI251
aI329
aI289
aI323
aI2201
aI4129
aI4963
aI27477
aI18743
aI46551
aI93061
asS'd'
I21037
sa(dp42074
S'a'
I128376
sS's'
I18
sS'm'
(lp42075
I1
aI1
aI7
aI3
aI17
aI63
aI21
aI159
aI447
aI377
aI69
aI2517
aI8181
aI6043
aI3039
aI7747
aI72465
aI41027
asS'd'
I21038
sa(dp42076
S'a'
I128406
sS's'
I18
sS'm'
(lp42077
I1
aI1
aI1
aI1
aI13
aI3
aI45
aI93
aI391
aI509
aI867
aI1561
aI5017
aI11851
aI24891
aI22531
aI18993
aI129421
asS'd'
I21039
sa(dp42078
S'a'
I128410
sS's'
I18
sS'm'
(lp42079
I1
aI3
aI3
aI11
aI15
aI1
aI127
aI9
aI161
aI321
aI2003
aI239
aI1379
aI11903
aI13503
aI26529
aI57725
aI214797
asS'd'
I21040
sa(dp42080
S'a'
I128416
sS's'
I18
sS'm'
(lp42081
I1
aI1
aI1
aI13
aI31
aI11
aI17
aI25
aI1
aI645
aI675
aI735
aI2083
aI1919
aI18977
aI4995
aI91559
aI230463
asS'd'
I21041
sa(dp42082
S'a'
I128443
sS's'
I18
sS'm'
(lp42083
I1
aI3
aI1
aI13
aI17
aI21
aI107
aI167
aI135
aI797
aI715
aI3275
aI5437
aI4253
aI11671
aI14867
aI36041
aI71751
asS'd'
I21042
sa(dp42084
S'a'
I128451
sS's'
I18
sS'm'
(lp42085
I1
aI3
aI5
aI5
aI11
aI49
aI93
aI231
aI431
aI567
aI1605
aI3281
aI7049
aI2947
aI863
aI39593
aI117167
aI167301
asS'd'
I21043
sa(dp42086
S'a'
I128465
sS's'
I18
sS'm'
(lp42087
I1
aI1
aI1
aI1
aI5
aI13
aI61
aI91
aI127
aI189
aI1879
aI3921
aI4303
aI4831
aI6765
aI31005
aI107627
aI80693
asS'd'
I21044
sa(dp42088
S'a'
I128466
sS's'
I18
sS'm'
(lp42089
I1
aI3
aI1
aI3
aI1
aI49
aI61
aI9
aI467
aI891
aI105
aI317
aI137
aI12789
aI12367
aI57455
aI39777
aI88047
asS'd'
I21045
sa(dp42090
S'a'
I128484
sS's'
I18
sS'm'
(lp42091
I1
aI1
aI3
aI13
aI23
aI63
aI37
aI103
aI23
aI223
aI647
aI2523
aI3211
aI14551
aI22663
aI48237
aI54777
aI180297
asS'd'
I21046
sa(dp42092
S'a'
I128499
sS's'
I18
sS'm'
(lp42093
I1
aI3
aI3
aI7
aI29
aI51
aI85
aI179
aI441
aI431
aI535
aI2975
aI8083
aI8619
aI30229
aI31421
aI54063
aI163601
asS'd'
I21047
sa(dp42094
S'a'
I128502
sS's'
I18
sS'm'
(lp42095
I1
aI3
aI1
aI1
aI27
aI39
aI125
aI171
aI57
aI729
aI511
aI957
aI7541
aI2347
aI1669
aI32323
aI108531
aI69943
asS'd'
I21048
sa(dp42096
S'a'
I128524
sS's'
I18
sS'm'
(lp42097
I1
aI3
aI3
aI7
aI1
aI33
aI89
aI245
aI95
aI21
aI699
aI1441
aI2659
aI501
aI32323
aI39145
aI82311
aI155479
asS'd'
I21049
sa(dp42098
S'a'
I128536
sS's'
I18
sS'm'
(lp42099
I1
aI1
aI3
aI11
aI29
aI13
aI87
aI251
aI329
aI667
aI325
aI2411
aI7959
aI8069
aI20817
aI42445
aI121675
aI113421
asS'd'
I21050
sa(dp42100
S'a'
I128541
sS's'
I18
sS'm'
(lp42101
I1
aI1
aI1
aI7
aI9
aI57
aI109
aI237
aI325
aI535
aI89
aI1285
aI5649
aI13673
aI29375
aI51553
aI81723
aI11003
asS'd'
I21051
sa(dp42102
S'a'
I128542
sS's'
I18
sS'm'
(lp42103
I1
aI3
aI1
aI1
aI13
aI5
aI31
aI109
aI157
aI817
aI1303
aI725
aI1841
aI5503
aI2255
aI34637
aI93603
aI82825
asS'd'
I21052
sa(dp42104
S'a'
I128592
sS's'
I18
sS'm'
(lp42105
I1
aI3
aI7
aI7
aI5
aI33
aI39
aI233
aI217
aI157
aI357
aI2727
aI3565
aI1539
aI5317
aI23967
aI30375
aI260381
asS'd'
I21053
sa(dp42106
S'a'
I128598
sS's'
I18
sS'm'
(lp42107
I1
aI1
aI5
aI3
aI23
aI51
aI45
aI181
aI353
aI519
aI949
aI3043
aI1517
aI3387
aI15081
aI5997
aI31523
aI80007
asS'd'
I21054
sa(dp42108
S'a'
I128623
sS's'
I18
sS'm'
(lp42109
I1
aI1
aI3
aI5
aI23
aI21
aI83
aI51
aI275
aI629
aI1433
aI1821
aI3761
aI2367
aI32089
aI13813
aI99629
aI64603
asS'd'
I21055
sa(dp42110
S'a'
I128632
sS's'
I18
sS'm'
(lp42111
I1
aI1
aI5
aI15
aI11
aI49
aI69
aI197
aI193
aI459
aI1915
aI787
aI3631
aI5219
aI11109
aI12311
aI56625
aI117439
asS'd'
I21056
sa(dp42112
S'a'
I128671
sS's'
I18
sS'm'
(lp42113
I1
aI3
aI3
aI3
aI31
aI29
aI57
aI27
aI43
aI231
aI777
aI2139
aI2609
aI12273
aI23777
aI4151
aI51749
aI110013
asS'd'
I21057
sa(dp42114
S'a'
I128713
sS's'
I18
sS'm'
(lp42115
I1
aI1
aI5
aI13
aI9
aI63
aI83
aI69
aI225
aI913
aI99
aI1167
aI5279
aI14163
aI3979
aI55151
aI84387
aI234583
asS'd'
I21058
sa(dp42116
S'a'
I128724
sS's'
I18
sS'm'
(lp42117
I1
aI1
aI7
aI5
aI9
aI57
aI87
aI23
aI335
aI403
aI1843
aI725
aI5187
aI4137
aI24299
aI44807
aI98523
aI217815
asS'd'
I21059
sa(dp42118
S'a'
I128731
sS's'
I18
sS'm'
(lp42119
I1
aI3
aI3
aI3
aI3
aI23
aI115
aI229
aI193
aI655
aI1205
aI3159
aI1935
aI113
aI20943
aI32917
aI69633
aI2133
asS'd'
I21060
sa(dp42120
S'a'
I128761
sS's'
I18
sS'm'
(lp42121
I1
aI3
aI3
aI1
aI17
aI5
aI59
aI139
aI75
aI185
aI1951
aI3689
aI4997
aI2761
aI8673
aI41783
aI75075
aI101063
asS'd'
I21061
sa(dp42122
S'a'
I128767
sS's'
I18
sS'm'
(lp42123
I1
aI3
aI1
aI13
aI1
aI51
aI63
aI127
aI67
aI743
aI1049
aI2055
aI4249
aI131
aI8153
aI50237
aI28135
aI76059
asS'd'
I21062
sa(dp42124
S'a'
I128782
sS's'
I18
sS'm'
(lp42125
I1
aI3
aI7
aI13
aI5
aI39
aI83
aI63
aI429
aI573
aI1915
aI3801
aI2223
aI1585
aI16997
aI45571
aI23311
aI108099
asS'd'
I21063
sa(dp42126
S'a'
I128793
sS's'
I18
sS'm'
(lp42127
I1
aI3
aI1
aI3
aI15
aI49
aI19
aI65
aI433
aI401
aI1901
aI3653
aI2399
aI15171
aI9695
aI30257
aI104877
aI181221
asS'd'
I21064
sa(dp42128
S'a'
I128805
sS's'
I18
sS'm'
(lp42129
I1
aI3
aI1
aI1
aI25
aI37
aI89
aI7
aI81
aI343
aI949
aI3535
aI1681
aI10089
aI23513
aI3897
aI127083
aI214005
asS'd'
I21065
sa(dp42130
S'a'
I128820
sS's'
I18
sS'm'
(lp42131
I1
aI3
aI7
aI13
aI1
aI1
aI123
aI89
aI433
aI541
aI1579
aI931
aI3459
aI11095
aI20729
aI13117
aI59323
aI90309
asS'd'
I21066
sa(dp42132
S'a'
I128830
sS's'
I18
sS'm'
(lp42133
I1
aI1
aI7
aI1
aI19
aI9
aI31
aI211
aI271
aI25
aI1053
aI2249
aI6549
aI12785
aI16947
aI55633
aI70155
aI253741
asS'd'
I21067
sa(dp42134
S'a'
I128842
sS's'
I18
sS'm'
(lp42135
I1
aI3
aI5
aI9
aI7
aI49
aI11
aI251
aI101
aI795
aI1015
aI2037
aI1239
aI10151
aI22179
aI749
aI2373
aI224517
asS'd'
I21068
sa(dp42136
S'a'
I128852
sS's'
I18
sS'm'
(lp42137
I1
aI3
aI7
aI15
aI21
aI19
aI15
aI59
aI439
aI621
aI1081
aI3041
aI1587
aI3077
aI2319
aI51135
aI110513
aI222551
asS'd'
I21069
sa(dp42138
S'a'
I128859
sS's'
I18
sS'm'
(lp42139
I1
aI1
aI5
aI5
aI9
aI61
aI49
aI97
aI361
aI647
aI351
aI1977
aI3023
aI10213
aI6889
aI8753
aI72203
aI37521
asS'd'
I21070
sa(dp42140
S'a'
I128866
sS's'
I18
sS'm'
(lp42141
I1
aI3
aI3
aI1
aI7
aI29
aI51
aI117
aI259
aI81
aI1263
aI1829
aI6541
aI5699
aI30367
aI61325
aI78795
aI3491
asS'd'
I21071
sa(dp42142
S'a'
I128875
sS's'
I18
sS'm'
(lp42143
I1
aI1
aI1
aI5
aI5
aI23
aI19
aI255
aI267
aI251
aI239
aI3561
aI6771
aI10647
aI4129
aI40285
aI11041
aI27023
asS'd'
I21072
sa(dp42144
S'a'
I128892
sS's'
I18
sS'm'
(lp42145
I1
aI1
aI7
aI5
aI29
aI17
aI121
aI91
aI427
aI51
aI243
aI1617
aI5389
aI3633
aI14105
aI5329
aI109507
aI93719
asS'd'
I21073
sa(dp42146
S'a'
I128896
sS's'
I18
sS'm'
(lp42147
I1
aI3
aI7
aI5
aI7
aI59
aI107
aI89
aI181
aI719
aI1029
aI585
aI2415
aI9175
aI11605
aI9271
aI12105
aI42503
asS'd'
I21074
sa(dp42148
S'a'
I128914
sS's'
I18
sS'm'
(lp42149
I1
aI3
aI5
aI15
aI27
aI15
aI83
aI223
aI489
aI901
aI1823
aI1515
aI6295
aI12509
aI27179
aI181
aI29813
aI66163
asS'd'
I21075
sa(dp42150
S'a'
I128935
sS's'
I18
sS'm'
(lp42151
I1
aI1
aI7
aI15
aI5
aI9
aI79
aI29
aI201
aI391
aI609
aI935
aI4025
aI201
aI8333
aI24557
aI33739
aI257979
asS'd'
I21076
sa(dp42152
S'a'
I128964
sS's'
I18
sS'm'
(lp42153
I1
aI3
aI3
aI9
aI9
aI19
aI55
aI211
aI347
aI943
aI559
aI467
aI1363
aI10249
aI7109
aI41293
aI28035
aI205889
asS'd'
I21077
sa(dp42154
S'a'
I128973
sS's'
I18
sS'm'
(lp42155
I1
aI1
aI5
aI3
aI21
aI25
aI25
aI163
aI95
aI119
aI789
aI1679
aI3845
aI1427
aI25531
aI13375
aI121029
aI194845
asS'd'
I21078
sa(dp42156
S'a'
I129001
sS's'
I18
sS'm'
(lp42157
I1
aI1
aI5
aI3
aI31
aI21
aI83
aI27
aI17
aI59
aI885
aI3889
aI4795
aI4383
aI28739
aI55129
aI10387
aI176437
asS'd'
I21079
sa(dp42158
S'a'
I129007
sS's'
I18
sS'm'
(lp42159
I1
aI3
aI1
aI5
aI31
aI39
aI37
aI79
aI433
aI313
aI1155
aI3025
aI6141
aI10695
aI27819
aI28227
aI32161
aI250515
asS'd'
I21080
sa(dp42160
S'a'
I129012
sS's'
I18
sS'm'
(lp42161
I1
aI1
aI5
aI9
aI27
aI41
aI3
aI129
aI235
aI621
aI1171
aI3305
aI6309
aI5323
aI15049
aI16301
aI13817
aI238521
asS'd'
I21081
sa(dp42162
S'a'
I129021
sS's'
I18
sS'm'
(lp42163
I1
aI3
aI7
aI7
aI27
aI31
aI63
aI143
aI183
aI625
aI1627
aI3093
aI6597
aI14089
aI30197
aI60411
aI66221
aI221691
asS'd'
I21082
sa(dp42164
S'a'
I129026
sS's'
I18
sS'm'
(lp42165
I1
aI3
aI7
aI13
aI21
aI15
aI59
aI67
aI441
aI113
aI1229
aI1587
aI5889
aI6691
aI10641
aI11865
aI89791
aI82867
asS'd'
I21083
sa(dp42166
S'a'
I129055
sS's'
I18
sS'm'
(lp42167
I1
aI1
aI7
aI9
aI13
aI21
aI53
aI145
aI235
aI877
aI2005
aI1005
aI7137
aI6091
aI19611
aI25959
aI124019
aI216269
asS'd'
I21084
sa(dp42168
S'a'
I129071
sS's'
I18
sS'm'
(lp42169
I1
aI3
aI7
aI9
aI17
aI63
aI5
aI245
aI397
aI351
aI1613
aI4079
aI7235
aI4397
aI18951
aI11609
aI71593
aI148615
asS'd'
I21085
sa(dp42170
S'a'
I129093
sS's'
I18
sS'm'
(lp42171
I1
aI3
aI3
aI11
aI5
aI59
aI65
aI221
aI237
aI527
aI861
aI397
aI249
aI15273
aI8415
aI61185
aI59419
aI98115
asS'd'
I21086
sa(dp42172
S'a'
I129105
sS's'
I18
sS'm'
(lp42173
I1
aI1
aI3
aI5
aI5
aI59
aI17
aI247
aI3
aI765
aI835
aI1131
aI3985
aI9021
aI18067
aI28525
aI86513
aI250227
asS'd'
I21087
sa(dp42174
S'a'
I129124
sS's'
I18
sS'm'
(lp42175
I1
aI3
aI7
aI15
aI25
aI47
aI119
aI143
aI143
aI283
aI1791
aI59
aI8171
aI12577
aI17079
aI9809
aI100299
aI63977
asS'd'
I21088
sa(dp42176
S'a'
I129148
sS's'
I18
sS'm'
(lp42177
I1
aI3
aI5
aI13
aI1
aI47
aI93
aI159
aI199
aI863
aI1279
aI77
aI4719
aI3623
aI30713
aI39271
aI126299
aI130297
asS'd'
I21089
sa(dp42178
S'a'
I129151
sS's'
I18
sS'm'
(lp42179
I1
aI1
aI5
aI3
aI23
aI11
aI119
aI187
aI57
aI373
aI747
aI1507
aI5165
aI12929
aI903
aI49041
aI70215
aI117113
asS'd'
I21090
sa(dp42180
S'a'
I129155
sS's'
I18
sS'm'
(lp42181
I1
aI1
aI5
aI1
aI3
aI59
aI23
aI77
aI151
aI77
aI627
aI2865
aI7055
aI10469
aI12095
aI20481
aI13429
aI47573
asS'd'
I21091
sa(dp42182
S'a'
I129161
sS's'
I18
sS'm'
(lp42183
I1
aI3
aI1
aI13
aI27
aI13
aI115
aI233
aI343
aI407
aI1321
aI4011
aI5589
aI15369
aI23495
aI4435
aI75421
aI229325
asS'd'
I21092
sa(dp42184
S'a'
I129200
sS's'
I18
sS'm'
(lp42185
I1
aI3
aI3
aI3
aI5
aI51
aI89
aI53
aI275
aI279
aI203
aI2829
aI4415
aI4735
aI25417
aI17633
aI99445
aI183945
asS'd'
I21093
sa(dp42186
S'a'
I129224
sS's'
I18
sS'm'
(lp42187
I1
aI3
aI3
aI15
aI7
aI9
aI91
aI63
aI143
aI945
aI453
aI4001
aI3943
aI7285
aI9359
aI27507
aI8571
aI31827
asS'd'
I21094
sa(dp42188
S'a'
I129230
sS's'
I18
sS'm'
(lp42189
I1
aI3
aI3
aI11
aI15
aI49
aI103
aI25
aI273
aI791
aI145
aI2203
aI4721
aI7709
aI25085
aI33937
aI98693
aI97445
asS'd'
I21095
sa(dp42190
S'a'
I129237
sS's'
I18
sS'm'
(lp42191
I1
aI1
aI5
aI15
aI9
aI13
aI87
aI27
aI331
aI137
aI1031
aI585
aI7841
aI12213
aI32259
aI46953
aI17813
aI203379
asS'd'
I21096
sa(dp42192
S'a'
I129298
sS's'
I18
sS'm'
(lp42193
I1
aI3
aI1
aI5
aI29
aI53
aI121
aI179
aI21
aI311
aI991
aI2145
aI6577
aI12889
aI8763
aI46629
aI128093
aI105033
asS'd'
I21097
sa(dp42194
S'a'
I129300
sS's'
I18
sS'm'
(lp42195
I1
aI3
aI1
aI9
aI7
aI29
aI57
aI137
aI333
aI109
aI615
aI749
aI2665
aI13087
aI13989
aI41857
aI102937
aI125183
asS'd'
I21098
sa(dp42196
S'a'
I129316
sS's'
I18
sS'm'
(lp42197
I1
aI1
aI5
aI5
aI3
aI23
aI107
aI5
aI319
aI503
aI1209
aI47
aI349
aI11681
aI28521
aI44707
aI112887
aI232275
asS'd'
I21099
sa(dp42198
S'a'
I129345
sS's'
I18
sS'm'
(lp42199
I1
aI3
aI3
aI13
aI13
aI51
aI13
aI5
aI293
aI15
aI555
aI135
aI2565
aI13325
aI30411
aI14837
aI65591
aI249205
asS'd'
I21100
sa(dp42200
S'a'
I129351
sS's'
I18
sS'm'
(lp42201
I1
aI3
aI5
aI13
aI17
aI3
aI73
aI255
aI447
aI699
aI503
aI3655
aI7735
aI12163
aI6167
aI15027
aI103831
aI146395
asS'd'
I21101
sa(dp42202
S'a'
I129352
sS's'
I18
sS'm'
(lp42203
I1
aI3
aI1
aI13
aI5
aI9
aI27
aI45
aI397
aI463
aI1739
aI3193
aI6731
aI7533
aI11217
aI22359
aI82603
aI231613
asS'd'
I21102
sa(dp42204
S'a'
I129363
sS's'
I18
sS'm'
(lp42205
I1
aI1
aI3
aI15
aI5
aI43
aI73
aI191
aI53
aI187
aI1905
aI745
aI1571
aI9013
aI8515
aI59527
aI104671
aI227063
asS'd'
I21103
sa(dp42206
S'a'
I129393
sS's'
I18
sS'm'
(lp42207
I1
aI1
aI3
aI1
aI5
aI47
aI57
aI179
aI433
aI979
aI147
aI1701
aI4019
aI6855
aI24487
aI65495
aI69919
aI6659
asS'd'
I21104
sa(dp42208
S'a'
I129410
sS's'
I18
sS'm'
(lp42209
I1
aI3
aI3
aI1
aI17
aI17
aI13
aI75
aI163
aI781
aI421
aI1573
aI2519
aI9243
aI20693
aI60909
aI65661
aI208125
asS'd'
I21105
sa(dp42210
S'a'
I129415
sS's'
I18
sS'm'
(lp42211
I1
aI3
aI5
aI7
aI27
aI57
aI39
aI79
aI157
aI415
aI729
aI3651
aI3581
aI9443
aI6409
aI45993
aI99051
aI140977
asS'd'
I21106
sa(dp42212
S'a'
I129449
sS's'
I18
sS'm'
(lp42213
I1
aI3
aI3
aI13
aI1
aI7
aI109
aI77
aI423
aI185
aI97
aI3719
aI2355
aI10593
aI2421
aI37339
aI24961
aI24305
asS'd'
I21107
sa(dp42214
S'a'
I129477
sS's'
I18
sS'm'
(lp42215
I1
aI3
aI5
aI13
aI17
aI7
aI125
aI43
aI453
aI43
aI643
aI3757
aI3721
aI16083
aI20871
aI26451
aI95201
aI29153
asS'd'
I21108
sa(dp42216
S'a'
I129501
sS's'
I18
sS'm'
(lp42217
I1
aI1
aI7
aI3
aI13
aI49
aI99
aI253
aI59
aI21
aI445
aI3677
aI6683
aI2165
aI32367
aI55249
aI5991
aI155033
asS'd'
I21109
sa(dp42218
S'a'
I129518
sS's'
I18
sS'm'
(lp42219
I1
aI1
aI3
aI9
aI21
aI9
aI15
aI219
aI175
aI631
aI665
aI2455
aI4701
aI10639
aI13907
aI26937
aI58867
aI259861
asS'd'
I21110
sa(dp42220
S'a'
I129520
sS's'
I18
sS'm'
(lp42221
I1
aI3
aI5
aI5
aI23
aI5
aI39
aI233
aI27
aI811
aI1435
aI625
aI4703
aI3699
aI20763
aI50047
aI123875
aI10129
asS'd'
I21111
sa(dp42222
S'a'
I129545
sS's'
I18
sS'm'
(lp42223
I1
aI1
aI7
aI5
aI23
aI1
aI49
aI223
aI309
aI691
aI953
aI575
aI5279
aI10515
aI11519
aI35387
aI48417
aI134001
asS'd'
I21112
sa(dp42224
S'a'
I129563
sS's'
I18
sS'm'
(lp42225
I1
aI1
aI7
aI11
aI3
aI15
aI125
aI109
aI39
aI713
aI1823
aI1613
aI4347
aI6839
aI29511
aI26865
aI102077
aI31425
asS'd'
I21113
sa(dp42226
S'a'
I129594
sS's'
I18
sS'm'
(lp42227
I1
aI1
aI1
aI7
aI31
aI43
aI13
aI221
aI115
aI993
aI1155
aI1641
aI1063
aI2065
aI18909
aI45769
aI65331
aI188455
asS'd'
I21114
sa(dp42228
S'a'
I129608
sS's'
I18
sS'm'
(lp42229
I1
aI1
aI7
aI13
aI21
aI9
aI59
aI7
aI79
aI217
aI2009
aI667
aI7685
aI14761
aI20149
aI44133
aI41037
aI78369
asS'd'
I21115
sa(dp42230
S'a'
I129616
sS's'
I18
sS'm'
(lp42231
I1
aI3
aI1
aI9
aI23
aI57
aI1
aI193
aI77
aI681
aI1135
aI3657
aI8149
aI3559
aI25011
aI55027
aI121903
aI240157
asS'd'
I21116
sa(dp42232
S'a'
I129652
sS's'
I18
sS'm'
(lp42233
I1
aI3
aI3
aI5
aI31
aI41
aI59
aI5
aI159
aI627
aI1569
aI23
aI2311
aI2239
aI20811
aI54931
aI130949
aI193071
asS'd'
I21117
sa(dp42234
S'a'
I129659
sS's'
I18
sS'm'
(lp42235
I1
aI1
aI1
aI1
aI27
aI45
aI43
aI1
aI381
aI801
aI451
aI1361
aI1611
aI5379
aI27819
aI8949
aI4953
aI222335
asS'd'
I21118
sa(dp42236
S'a'
I129680
sS's'
I18
sS'm'
(lp42237
I1
aI1
aI7
aI7
aI7
aI11
aI101
aI17
aI197
aI561
aI297
aI159
aI7443
aI7273
aI819
aI23487
aI24927
aI151781
asS'd'
I21119
sa(dp42238
S'a'
I129692
sS's'
I18
sS'm'
(lp42239
I1
aI3
aI1
aI3
aI15
aI43
aI119
aI193
aI205
aI835
aI7
aI689
aI8045
aI11167
aI19521
aI65075
aI87265
aI53669
asS'd'
I21120
sa(dp42240
S'a'
I129738
sS's'
I18
sS'm'
(lp42241
I1
aI3
aI7
aI7
aI9
aI51
aI43
aI209
aI239
aI415
aI995
aI4037
aI1219
aI2683
aI30459
aI36161
aI111157
aI184551
asS'd'
I21121
sa(dp42242
S'a'
I129764
sS's'
I18
sS'm'
(lp42243
I1
aI3
aI7
aI11
aI27
aI3
aI81
aI43
aI407
aI463
aI231
aI3545
aI2691
aI5235
aI22053
aI37233
aI98757
aI149111
asS'd'
I21122
sa(dp42244
S'a'
I129782
sS's'
I18
sS'm'
(lp42245
I1
aI3
aI1
aI5
aI17
aI25
aI47
aI185
aI487
aI403
aI1063
aI1445
aI4457
aI15443
aI11693
aI54823
aI131001
aI9813
asS'd'
I21123
sa(dp42246
S'a'
I129793
sS's'
I18
sS'm'
(lp42247
I1
aI3
aI5
aI3
aI5
aI35
aI127
aI253
aI173
aI491
aI133
aI3575
aI1981
aI12735
aI26021
aI61615
aI74615
aI159829
asS'd'
I21124
sa(dp42248
S'a'
I129820
sS's'
I18
sS'm'
(lp42249
I1
aI1
aI5
aI9
aI13
aI37
aI67
aI155
aI317
aI389
aI603
aI4061
aI3527
aI9315
aI32331
aI43145
aI82511
aI240133
asS'd'
I21125
sa(dp42250
S'a'
I129824
sS's'
I18
sS'm'
(lp42251
I1
aI1
aI5
aI1
aI21
aI41
aI89
aI3
aI61
aI627
aI1301
aI2073
aI447
aI8139
aI2509
aI52075
aI50687
aI240239
asS'd'
I21126
sa(dp42252
S'a'
I129829
sS's'
I18
sS'm'
(lp42253
I1
aI3
aI7
aI13
aI25
aI61
aI117
aI107
aI175
aI7
aI1173
aI561
aI5777
aI10525
aI20713
aI34987
aI48005
aI214361
asS'd'
I21127
sa(dp42254
S'a'
I129848
sS's'
I18
sS'm'
(lp42255
I1
aI3
aI7
aI15
aI25
aI31
aI127
aI147
aI177
aI881
aI95
aI2115
aI4765
aI10485
aI9253
aI721
aI193
aI222459
asS'd'
I21128
sa(dp42256
S'a'
I129854
sS's'
I18
sS'm'
(lp42257
I1
aI3
aI5
aI3
aI3
aI13
aI47
aI77
aI441
aI1001
aI215
aI2365
aI3603
aI405
aI11401
aI14523
aI65755
aI258229
asS'd'
I21129
sa(dp42258
S'a'
I129861
sS's'
I18
sS'm'
(lp42259
I1
aI1
aI3
aI9
aI19
aI29
aI71
aI153
aI77
aI613
aI1815
aI2033
aI1821
aI15497
aI18805
aI28851
aI88247
aI143115
asS'd'
I21130
sa(dp42260
S'a'
I129879
sS's'
I18
sS'm'
(lp42261
I1
aI3
aI1
aI11
aI21
aI19
aI37
aI35
aI427
aI887
aI1977
aI1961
aI3619
aI10739
aI30115
aI55937
aI102045
aI110929
asS'd'
I21131
sa(dp42262
S'a'
I129886
sS's'
I18
sS'm'
(lp42263
I1
aI3
aI3
aI13
aI21
aI27
aI49
aI15
aI405
aI629
aI2015
aI867
aI2121
aI13789
aI19225
aI22343
aI105629
aI123113
asS'd'
I21132
sa(dp42264
S'a'
I129907
sS's'
I18
sS'm'
(lp42265
I1
aI1
aI1
aI1
aI17
aI19
aI55
aI207
aI507
aI1001
aI1753
aI315
aI2799
aI8643
aI1519
aI4057
aI16599
aI222223
asS'd'
I21133
sa(dp42266
S'a'
I129949
sS's'
I18
sS'm'
(lp42267
I1
aI1
aI5
aI7
aI21
aI37
aI63
aI53
aI103
aI261
aI595
aI389
aI6041
aI11127
aI23625
aI61683
aI80953
aI255891
asS'd'
I21134
sa(dp42268
S'a'
I129953
sS's'
I18
sS'm'
(lp42269
I1
aI1
aI1
aI5
aI25
aI21
aI81
aI233
aI79
aI57
aI1311
aI3965
aI7747
aI687
aI32149
aI397
aI4551
aI37657
asS'd'
I21135
sa(dp42270
S'a'
I129980
sS's'
I18
sS'm'
(lp42271
I1
aI1
aI1
aI5
aI9
aI19
aI87
aI67
aI325
aI157
aI317
aI591
aI1401
aI8275
aI20413
aI39529
aI75349
aI183679
asS'd'
I21136
sa(dp42272
S'a'
I129998
sS's'
I18
sS'm'
(lp42273
I1
aI3
aI3
aI15
aI9
aI3
aI83
aI205
aI195
aI599
aI829
aI3109
aI3705
aI13991
aI8781
aI41555
aI31689
aI86933
asS'd'
I21137
sa(dp42274
S'a'
I130031
sS's'
I18
sS'm'
(lp42275
I1
aI3
aI5
aI1
aI3
aI9
aI37
aI235
aI271
aI883
aI561
aI1473
aI7693
aI177
aI14113
aI19507
aI75221
aI67517
asS'd'
I21138
sa(dp42276
S'a'
I130043
sS's'
I18
sS'm'
(lp42277
I1
aI1
aI1
aI9
aI7
aI29
aI87
aI189
aI239
aI429
aI537
aI1657
aI6373
aI2449
aI17621
aI19649
aI77235
aI102775
asS'd'
I21139
sa(dp42278
S'a'
I130057
sS's'
I18
sS'm'
(lp42279
I1
aI1
aI7
aI7
aI1
aI43
aI69
aI207
aI241
aI561
aI1809
aI3119
aI4657
aI15797
aI18751
aI52169
aI105005
aI172657
asS'd'
I21140
sa(dp42280
S'a'
I130065
sS's'
I18
sS'm'
(lp42281
I1
aI1
aI7
aI5
aI5
aI59
aI67
aI231
aI27
aI435
aI1073
aI2689
aI229
aI733
aI1579
aI52289
aI110285
aI76721
asS'd'
I21141
sa(dp42282
S'a'
I130087
sS's'
I18
sS'm'
(lp42283
I1
aI3
aI5
aI9
aI31
aI19
aI87
aI41
aI489
aI705
aI1363
aI963
aI5865
aI8237
aI10295
aI43169
aI81561
aI177209
asS'd'
I21142
sa(dp42284
S'a'
I130091
sS's'
I18
sS'm'
(lp42285
I1
aI3
aI3
aI1
aI25
aI39
aI63
aI255
aI403
aI625
aI1601
aI71
aI6609
aI4165
aI21987
aI31269
aI25473
aI17063
asS'd'
I21143
sa(dp42286
S'a'
I130096
sS's'
I18
sS'm'
(lp42287
I1
aI3
aI3
aI11
aI19
aI13
aI101
aI245
aI17
aI687
aI1037
aI3345
aI7257
aI13081
aI5131
aI29003
aI72319
aI223505
asS'd'
I21144
sa(dp42288
S'a'
I130101
sS's'
I18
sS'm'
(lp42289
I1
aI3
aI3
aI3
aI11
aI49
aI107
aI29
aI463
aI465
aI977
aI4007
aI2121
aI4821
aI1465
aI53725
aI36783
aI247057
asS'd'
I21145
sa(dp42290
S'a'
I130111
sS's'
I18
sS'm'
(lp42291
I1
aI1
aI1
aI3
aI1
aI43
aI71
aI49
aI261
aI965
aI1041
aI3951
aI3791
aI2503
aI26009
aI52039
aI4639
aI141281
asS'd'
I21146
sa(dp42292
S'a'
I130126
sS's'
I18
sS'm'
(lp42293
I1
aI1
aI3
aI5
aI29
aI45
aI79
aI33
aI119
aI491
aI1403
aI1637
aI853
aI5609
aI29853
aI16435
aI117877
aI58443
asS'd'
I21147
sa(dp42294
S'a'
I130137
sS's'
I18
sS'm'
(lp42295
I1
aI1
aI3
aI5
aI13
aI17
aI109
aI187
aI201
aI705
aI235
aI1485
aI7673
aI6335
aI3341
aI20451
aI64697
aI129519
asS'd'
I21148
sa(dp42296
S'a'
I130138
sS's'
I18
sS'm'
(lp42297
I1
aI1
aI7
aI13
aI11
aI41
aI95
aI81
aI135
aI783
aI1293
aI2095
aI3599
aI10175
aI3205
aI56915
aI131
aI19281
asS'd'
I21149
sa(dp42298
S'a'
I130149
sS's'
I18
sS'm'
(lp42299
I1
aI3
aI5
aI9
aI13
aI19
aI53
aI223
aI283
aI733
aI1915
aI3029
aI2779
aI8133
aI28163
aI37263
aI91245
aI1927
asS'd'
I21150
sa(dp42300
S'a'
I130177
sS's'
I18
sS'm'
(lp42301
I1
aI3
aI3
aI3
aI1
aI55
aI41
aI123
aI209
aI195
aI1423
aI2467
aI3809
aI11169
aI23593
aI8703
aI40975
aI175651
asS'd'
I21151
sa(dp42302
S'a'
I130180
sS's'
I18
sS'm'
(lp42303
I1
aI3
aI7
aI9
aI31
aI57
aI31
aI115
aI415
aI445
aI557
aI3971
aI1565
aI15223
aI7799
aI10463
aI117387
aI225127
asS'd'
I21152
sa(dp42304
S'a'
I130214
sS's'
I18
sS'm'
(lp42305
I1
aI3
aI5
aI11
aI31
aI19
aI3
aI63
aI315
aI501
aI903
aI1925
aI3393
aI16149
aI11013
aI15483
aI70765
aI279
asS'd'
I21153
sa(dp42306
S'a'
I130228
sS's'
I18
sS'm'
(lp42307
I1
aI3
aI3
aI9
aI29
aI21
aI13
aI227
aI263
aI815
aI1259
aI2549
aI955
aI9237
aI16083
aI38891
aI31145
aI731
asS'd'
I21154
sa(dp42308
S'a'
I130240
sS's'
I18
sS'm'
(lp42309
I1
aI3
aI3
aI5
aI27
aI23
aI33
aI189
aI107
aI655
aI889
aI1549
aI7315
aI13341
aI12721
aI59339
aI54503
aI91679
asS'd'
I21155
sa(dp42310
S'a'
I130267
sS's'
I18
sS'm'
(lp42311
I1
aI3
aI7
aI5
aI15
aI9
aI1
aI255
aI451
aI91
aI1279
aI2359
aI5913
aI5215
aI23161
aI29327
aI45275
aI206709
asS'd'
I21156
sa(dp42312
S'a'
I130280
sS's'
I18
sS'm'
(lp42313
I1
aI3
aI3
aI9
aI9
aI41
aI75
aI91
aI87
aI695
aI335
aI3375
aI7307
aI14095
aI5359
aI7815
aI9339
aI46387
asS'd'
I21157
sa(dp42314
S'a'
I130294
sS's'
I18
sS'm'
(lp42315
I1
aI1
aI3
aI15
aI5
aI47
aI69
aI231
aI423
aI255
aI1335
aI3395
aI2799
aI8955
aI31445
aI59849
aI104955
aI240587
asS'd'
I21158
sa(dp42316
S'a'
I130306
sS's'
I18
sS'm'
(lp42317
I1
aI3
aI5
aI7
aI7
aI9
aI21
aI209
aI321
aI5
aI653
aI2199
aI3657
aI6397
aI20229
aI32349
aI54543
aI47971
asS'd'
I21159
sa(dp42318
S'a'
I130325
sS's'
I18
sS'm'
(lp42319
I1
aI3
aI7
aI11
aI31
aI21
aI85
aI49
aI197
aI865
aI53
aI609
aI1867
aI14503
aI12671
aI61703
aI39245
aI8493
asS'd'
I21160
sa(dp42320
S'a'
I130398
sS's'
I18
sS'm'
(lp42321
I1
aI3
aI1
aI13
aI27
aI31
aI119
aI247
aI209
aI65
aI1729
aI1563
aI1597
aI1617
aI26597
aI50139
aI108667
aI77035
asS'd'
I21161
sa(dp42322
S'a'
I130402
sS's'
I18
sS'm'
(lp42323
I1
aI1
aI5
aI13
aI3
aI49
aI53
aI219
aI71
aI1013
aI1239
aI3725
aI117
aI9273
aI8277
aI32619
aI45933
aI71509
asS'd'
I21162
sa(dp42324
S'a'
I130421
sS's'
I18
sS'm'
(lp42325
I1
aI3
aI7
aI13
aI1
aI3
aI119
aI153
aI79
aI555
aI429
aI1221
aI3725
aI6073
aI1295
aI7187
aI117709
aI258911
asS'd'
I21163
sa(dp42326
S'a'
I130438
sS's'
I18
sS'm'
(lp42327
I1
aI3
aI3
aI13
aI1
aI13
aI105
aI185
aI81
aI989
aI563
aI3761
aI6725
aI4699
aI10539
aI50247
aI95307
aI211927
asS'd'
I21164
sa(dp42328
S'a'
I130441
sS's'
I18
sS'm'
(lp42329
I1
aI3
aI7
aI3
aI21
aI11
aI45
aI81
aI495
aI391
aI1437
aI3495
aI3789
aI13701
aI9479
aI42505
aI22561
aI135019
asS'd'
I21165
sa(dp42330
S'a'
I130475
sS's'
I18
sS'm'
(lp42331
I1
aI3
aI3
aI11
aI7
aI61
aI65
aI211
aI269
aI997
aI385
aI3843
aI4905
aI2939
aI28551
aI19515
aI25177
aI68137
asS'd'
I21166
sa(dp42332
S'a'
I130486
sS's'
I18
sS'm'
(lp42333
I1
aI1
aI3
aI3
aI3
aI47
aI73
aI127
aI15
aI977
aI209
aI1791
aI4711
aI6733
aI29093
aI36311
aI13665
aI240603
asS'd'
I21167
sa(dp42334
S'a'
I130503
sS's'
I18
sS'm'
(lp42335
I1
aI3
aI5
aI5
aI19
aI39
aI29
aI211
aI463
aI755
aI1723
aI397
aI213
aI14009
aI22701
aI7131
aI35587
aI183885
asS'd'
I21168
sa(dp42336
S'a'
I130551
sS's'
I18
sS'm'
(lp42337
I1
aI3
aI5
aI9
aI11
aI29
aI7
aI25
aI381
aI631
aI1343
aI2255
aI2535
aI3239
aI7287
aI14161
aI69295
aI85245
asS'd'
I21169
sa(dp42338
S'a'
I130574
sS's'
I18
sS'm'
(lp42339
I1
aI1
aI5
aI5
aI17
aI47
aI19
aI217
aI289
aI411
aI1855
aI323
aI4109
aI2601
aI5835
aI61909
aI99333
aI99959
asS'd'
I21170
sa(dp42340
S'a'
I130602
sS's'
I18
sS'm'
(lp42341
I1
aI1
aI3
aI11
aI1
aI51
aI121
aI207
aI403
aI993
aI1171
aI3451
aI3389
aI957
aI22125
aI9333
aI110775
aI54125
asS'd'
I21171
sa(dp42342
S'a'
I130612
sS's'
I18
sS'm'
(lp42343
I1
aI3
aI5
aI15
aI9
aI51
aI13
aI251
aI203
aI861
aI321
aI2017
aI6933
aI10785
aI20089
aI65213
aI105451
aI117319
asS'd'
I21172
sa(dp42344
S'a'
I130621
sS's'
I18
sS'm'
(lp42345
I1
aI3
aI3
aI15
aI19
aI63
aI89
aI217
aI269
aI723
aI57
aI1923
aI4267
aI4895
aI2191
aI21605
aI62401
aI11063
asS'd'
I21173
sa(dp42346
S'a'
I130633
sS's'
I18
sS'm'
(lp42347
I1
aI3
aI1
aI3
aI21
aI47
aI103
aI75
aI167
aI989
aI1401
aI575
aI3717
aI10373
aI21321
aI5487
aI36063
aI140411
asS'd'
I21174
sa(dp42348
S'a'
I130684
sS's'
I18
sS'm'
(lp42349
I1
aI1
aI7
aI15
aI19
aI29
aI121
aI197
aI429
aI773
aI901
aI1875
aI291
aI11395
aI31459
aI55041
aI49263
aI185143
asS'd'
I21175
sa(dp42350
S'a'
I130688
sS's'
I18
sS'm'
(lp42351
I1
aI3
aI3
aI1
aI19
aI17
aI19
aI21
aI41
aI885
aI1665
aI547
aI5887
aI6205
aI3317
aI59399
aI125559
aI82721
asS'd'
I21176
sa(dp42352
S'a'
I130698
sS's'
I18
sS'm'
(lp42353
I1
aI3
aI1
aI9
aI15
aI39
aI81
aI9
aI279
aI33
aI1287
aI3035
aI5759
aI10647
aI3933
aI20953
aI3137
aI30693
asS'd'
I21177
sa(dp42354
S'a'
I130700
sS's'
I18
sS'm'
(lp42355
I1
aI3
aI5
aI13
aI3
aI33
aI33
aI169
aI233
aI83
aI467
aI3719
aI5617
aI6165
aI15631
aI56059
aI95541
aI245233
asS'd'
I21178
sa(dp42356
S'a'
I130712
sS's'
I18
sS'm'
(lp42357
I1
aI3
aI5
aI13
aI5
aI21
aI81
aI9
aI413
aI247
aI1307
aI3363
aI3383
aI11525
aI1259
aI8735
aI36507
aI98359
asS'd'
I21179
sa(dp42358
S'a'
I130739
sS's'
I18
sS'm'
(lp42359
I1
aI1
aI1
aI13
aI17
aI49
aI105
aI131
aI385
aI309
aI1295
aI565
aI8031
aI15391
aI31263
aI52657
aI102721
aI212195
asS'd'
I21180
sa(dp42360
S'a'
I130748
sS's'
I18
sS'm'
(lp42361
I1
aI3
aI1
aI7
aI13
aI41
aI21
aI103
aI237
aI649
aI55
aI1565
aI6327
aI8743
aI15457
aI29975
aI34165
aI80839
asS'd'
I21181
sa(dp42362
S'a'
I130774
sS's'
I18
sS'm'
(lp42363
I1
aI3
aI5
aI11
aI15
aI31
aI121
aI219
aI375
aI159
aI731
aI59
aI3205
aI15039
aI10023
aI46209
aI34619
aI110253
asS'd'
I21182
sa(dp42364
S'a'
I130783
sS's'
I18
sS'm'
(lp42365
I1
aI3
aI3
aI11
aI31
aI19
aI79
aI185
aI363
aI635
aI463
aI987
aI2681
aI6405
aI30077
aI21173
aI14213
aI58095
asS'd'
I21183
sa(dp42366
S'a'
I130802
sS's'
I18
sS'm'
(lp42367
I1
aI1
aI5
aI13
aI23
aI37
aI57
aI111
aI293
aI553
aI269
aI3393
aI345
aI1983
aI1097
aI47217
aI22281
aI212607
asS'd'
I21184
sa(dp42368
S'a'
I130811
sS's'
I18
sS'm'
(lp42369
I1
aI3
aI5
aI7
aI7
aI33
aI65
aI61
aI185
aI411
aI187
aI641
aI6437
aI4625
aI17547
aI38941
aI81119
aI48651
asS'd'
I21185
sa(dp42370
S'a'
I130836
sS's'
I18
sS'm'
(lp42371
I1
aI1
aI7
aI3
aI19
aI25
aI39
aI243
aI139
aI465
aI691
aI713
aI7879
aI14539
aI31669
aI35871
aI130681
aI255929
asS'd'
I21186
sa(dp42372
S'a'
I130840
sS's'
I18
sS'm'
(lp42373
I1
aI3
aI1
aI1
aI3
aI43
aI87
aI13
aI179
aI835
aI719
aI1189
aI7207
aI5863
aI6077
aI20669
aI35469
aI211155
asS'd'
I21187
sa(dp42374
S'a'
I130856
sS's'
I18
sS'm'
(lp42375
I1
aI3
aI7
aI13
aI25
aI59
aI97
aI129
aI151
aI985
aI739
aI1919
aI7729
aI14057
aI21721
aI17603
aI82797
aI181319
asS'd'
I21188
sa(dp42376
S'a'
I130864
sS's'
I18
sS'm'
(lp42377
I1
aI1
aI7
aI7
aI5
aI3
aI21
aI141
aI379
aI257
aI207
aI597
aI4051
aI7563
aI25481
aI59427
aI45449
aI61159
asS'd'
I21189
sa(dp42378
S'a'
I130873
sS's'
I18
sS'm'
(lp42379
I1
aI3
aI3
aI9
aI11
aI25
aI5
aI29
aI131
aI603
aI637
aI189
aI4033
aI13099
aI15219
aI4447
aI73501
aI135795
asS'd'
I21190
sa(dp42380
S'a'
I130918
sS's'
I18
sS'm'
(lp42381
I1
aI3
aI1
aI9
aI1
aI49
aI57
aI227
aI141
aI543
aI1499
aI3525
aI3127
aI11191
aI4071
aI47003
aI7431
aI155137
asS'd'
I21191
sa(dp42382
S'a'
I130927
sS's'
I18
sS'm'
(lp42383
I1
aI3
aI1
aI11
aI27
aI31
aI15
aI31
aI113
aI135
aI1251
aI245
aI6965
aI14263
aI5679
aI55201
aI121453
aI132503
asS'd'
I21192
sa(dp42384
S'a'
I130929
sS's'
I18
sS'm'
(lp42385
I1
aI1
aI5
aI15
aI7
aI23
aI67
aI163
aI57
aI513
aI1809
aI1343
aI6165
aI199
aI31169
aI30803
aI86705
aI71103
asS'd'
I21193
sa(dp42386
S'a'
I130958
sS's'
I18
sS'm'
(lp42387
I1
aI1
aI3
aI1
aI15
aI9
aI75
aI143
aI273
aI797
aI819
aI4037
aI2305
aI4841
aI15697
aI41191
aI38187
aI174131
asS'd'
I21194
sa(dp42388
S'a'
I130966
sS's'
I18
sS'm'
(lp42389
I1
aI3
aI7
aI7
aI3
aI55
aI65
aI135
aI423
aI185
aI299
aI2221
aI7987
aI4223
aI28183
aI32273
aI95941
aI260297
asS'd'
I21195
sa(dp42390
S'a'
I130970
sS's'
I18
sS'm'
(lp42391
I1
aI1
aI7
aI7
aI7
aI11
aI67
aI109
aI507
aI673
aI1555
aI2537
aI7553
aI4659
aI3945
aI20839
aI32539
aI43053
asS'd'
I21196
sa(dp42392
S'a'
I130976
sS's'
I18
sS'm'
(lp42393
I1
aI1
aI7
aI15
aI1
aI47
aI61
aI73
aI211
aI397
aI1785
aI4063
aI6461
aI13725
aI11299
aI17565
aI80063
aI118271
asS'd'
I21197
sa(dp42394
S'a'
I131006
sS's'
I18
sS'm'
(lp42395
I1
aI1
aI7
aI5
aI29
aI27
aI97
aI105
aI379
aI153
aI915
aI2795
aI4933
aI6729
aI21207
aI9995
aI70241
aI85641
asS'd'
I21198
sa(dp42396
S'a'
I131008
sS's'
I18
sS'm'
(lp42397
I1
aI3
aI5
aI5
aI23
aI13
aI41
aI67
aI127
aI649
aI1351
aI3597
aI7077
aI4989
aI14649
aI17401
aI70883
aI239841
asS'd'
I21199
sa(dp42398
S'a'
I131020
sS's'
I18
sS'm'
(lp42399
I1
aI1
aI5
aI1
aI19
aI1
aI83
aI3
aI425
aI873
aI1943
aI3935
aI4257
aI14587
aI11829
aI55217
aI21963
aI39683
asS'd'
I21200
sa(dp42400
S'a'
I131059
sS's'
I18
sS'm'
(lp42401
I1
aI1
aI7
aI11
aI15
aI7
aI37
aI239
aI337
aI245
aI1557
aI3681
aI7357
aI9639
aI27367
aI26869
aI114603
aI86317
asS'd'
I21201
sa."""

def get_params():
    return pickle.loads(params)

if __name__ == '__main__':

    sobol(1000, 4)


